version = (1, 1)

data = ({'ATOMTemplate': {'type': 'runtime'},
  'AbortCurrentTool': {'type': 'runtime'},
  'ActivateGlobalScreenSlider': {'type': 'runtime'},
  'ActivateGlobalScreenSliderModeMarkingMenu': {'type': 'runtime'},
  'ActivateViewport20': {'type': 'runtime'},
  'AddAnimationOffset': {'type': 'runtime'},
  'AddAnimationOffsetOptions': {'type': 'runtime'},
  'AddAttribute': {'type': 'runtime'},
  'AddBlendShape': {'type': 'runtime'},
  'AddBlendShapeOptions': {'type': 'runtime'},
  'AddBoatLocator': {'type': 'runtime'},
  'AddBoatLocatorOptions': {'type': 'runtime'},
  'AddCombinationTarget': {'type': 'runtime'},
  'AddCombinationTargetOptions': {'type': 'runtime'},
  'AddCurvesToHairSystem': {'type': 'runtime'},
  'AddDivisions': {'type': 'runtime'},
  'AddDivisionsOptions': {'type': 'runtime'},
  'AddDynamicBuoy': {'type': 'runtime'},
  'AddDynamicBuoyOptions': {'type': 'runtime'},
  'AddEdgeDivisions': {'type': 'runtime'},
  'AddEdgeDivisionsOptions': {'type': 'runtime'},
  'AddFaceDivisions': {'type': 'runtime'},
  'AddFaceDivisionsOptions': {'type': 'runtime'},
  'AddFloorContactPlane': {'type': 'runtime'},
  'AddHolder': {'type': 'runtime'},
  'AddHolderOptions': {'type': 'runtime'},
  'AddInBetweenTargetShape': {'type': 'runtime'},
  'AddInBetweenTargetShapeOptions': {'type': 'runtime'},
  'AddInbetween': {'type': 'runtime'},
  'AddInfluence': {'type': 'runtime'},
  'AddInfluenceOptions': {'type': 'runtime'},
  'AddKeyToolActivate': {'type': 'runtime'},
  'AddKeyToolDeactivate': {'type': 'runtime'},
  'AddKeysTool': {'type': 'runtime'},
  'AddKeysToolOptions': {'type': 'runtime'},
  'AddOceanDynamicLocator': {'type': 'runtime'},
  'AddOceanDynamicLocatorOptions': {'type': 'runtime'},
  'AddOceanPreviewPlane': {'type': 'runtime'},
  'AddOceanSurfaceLocator': {'type': 'runtime'},
  'AddPfxToHairSystem': {'type': 'runtime'},
  'AddPointsTool': {'type': 'runtime'},
  'AddPondBoatLocator': {'type': 'runtime'},
  'AddPondBoatLocatorOptions': {'type': 'runtime'},
  'AddPondDynamicBuoy': {'type': 'runtime'},
  'AddPondDynamicBuoyOptions': {'type': 'runtime'},
  'AddPondDynamicLocator': {'type': 'runtime'},
  'AddPondDynamicLocatorOptions': {'type': 'runtime'},
  'AddPondSurfaceLocator': {'type': 'runtime'},
  'AddSelectionAsCombinationTarget': {'type': 'runtime'},
  'AddSelectionAsCombinationTargetOptions': {'type': 'runtime'},
  'AddSelectionAsInBetweenTargetShape': {'type': 'runtime'},
  'AddSelectionAsInBetweenTargetShapeOptions': {'type': 'runtime'},
  'AddSelectionAsTargetShape': {'type': 'runtime'},
  'AddSelectionAsTargetShapeOptions': {'type': 'runtime'},
  'AddShrinkWrapSurfaces': {'type': 'runtime'},
  'AddTargetShape': {'type': 'runtime'},
  'AddTargetShapeOptions': {'type': 'runtime'},
  'AddTimeWarp': {'type': 'runtime'},
  'AddToCharacterSet': {'type': 'runtime'},
  'AddToContainer': {'type': 'runtime'},
  'AddToContainerOptions': {'type': 'runtime'},
  'AddToCurrentScene3dsMax': {'type': 'runtime'},
  'AddToCurrentSceneMotionBuilder': {'type': 'runtime'},
  'AddToCurrentSceneMudbox': {'type': 'runtime'},
  'AddWire': {'type': 'runtime'},
  'AddWireOptions': {'type': 'runtime'},
  'AddWrapInfluence': {'type': 'runtime'},
  'AffectSelectedObject': {'type': 'runtime'},
  'AimConstraint': {'type': 'runtime'},
  'AimConstraintOptions': {'type': 'runtime'},
  'Air': {'type': 'runtime'},
  'AirOptions': {'type': 'runtime'},
  'AlignCameraToPolygon': {'type': 'runtime'},
  'AlignCurve': {'type': 'runtime'},
  'AlignCurveOptions': {'type': 'runtime'},
  'AlignObjects': {'type': 'runtime'},
  'AlignSurfaces': {'type': 'runtime'},
  'AlignSurfacesOptions': {'type': 'runtime'},
  'AlignUV': {'type': 'runtime'},
  'AlignUVOptions': {'type': 'runtime'},
  'AnimLayerRelationshipEditor': {'type': 'runtime'},
  'AnimationSnapshot': {'type': 'runtime'},
  'AnimationSnapshotOptions': {'type': 'runtime'},
  'AnimationSweep': {'type': 'runtime'},
  'AnimationSweepOptions': {'type': 'runtime'},
  'AnimationTurntable': {'type': 'runtime'},
  'AnimationTurntableOptions': {'type': 'runtime'},
  'AppendToHairCache': {'type': 'runtime'},
  'AppendToHairCacheOptions': {'type': 'runtime'},
  'AppendToPolygonTool': {'type': 'runtime'},
  'AppendToPolygonToolOptions': {'type': 'runtime'},
  'ApplySettingsToLastStroke': {'type': 'runtime'},
  'ApplySettingsToSelectedStroke': {'type': 'runtime'},
  'ArcLengthTool': {'type': 'runtime'},
  'ArchiveScene': {'type': 'runtime'},
  'ArchiveSceneOptions': {'type': 'runtime'},
  'Art3dPaintTool': {'type': 'runtime'},
  'Art3dPaintToolOptions': {'type': 'runtime'},
  'ArtPaintAttrTool': {'type': 'runtime'},
  'ArtPaintAttrToolOptions': {'type': 'runtime'},
  'ArtPaintBlendShapeWeightsTool': {'type': 'runtime'},
  'ArtPaintBlendShapeWeightsToolOptions': {'type': 'runtime'},
  'ArtPaintSelectTool': {'type': 'runtime'},
  'ArtPaintSelectToolOptions': {'type': 'runtime'},
  'ArtPaintSkinWeightsTool': {'type': 'runtime'},
  'ArtPaintSkinWeightsToolOptions': {'type': 'runtime'},
  'AssetEditor': {'type': 'runtime'},
  'AssignBrushToHairSystem': {'type': 'runtime'},
  'AssignBrushToPfxToon': {'type': 'runtime'},
  'AssignHairConstraint': {'type': 'runtime'},
  'AssignHairConstraintOptions': {'type': 'runtime'},
  'AssignNewMaterial': {'type': 'runtime'},
  'AssignNewPfxToon': {'type': 'runtime'},
  'AssignNewSet': {'type': 'runtime'},
  'AssignOfflineFile': {'type': 'runtime'},
  'AssignOfflineFileFromRefEd': {'type': 'runtime'},
  'AssignOfflineFileFromRefEdOptions': {'type': 'runtime'},
  'AssignOfflineFileOptions': {'type': 'runtime'},
  'AssignTemplate': {'type': 'runtime'},
  'AssignTemplateOptions': {'type': 'runtime'},
  'AssignToonShaderCircleHighlight': {'type': 'runtime'},
  'AssignToonShaderDarkProfile': {'type': 'runtime'},
  'AssignToonShaderLightAngle': {'type': 'runtime'},
  'AssignToonShaderRimLight': {'type': 'runtime'},
  'AssignToonShaderShadedBrightness': {'type': 'runtime'},
  'AssignToonShaderSolid': {'type': 'runtime'},
  'AssignToonShaderThreeToneBrightness': {'type': 'runtime'},
  'AssumePreferredAngle': {'type': 'runtime'},
  'AssumePreferredAngleOptions': {'type': 'runtime'},
  'AttachBrushToCurves': {'type': 'runtime'},
  'AttachCurve': {'type': 'runtime'},
  'AttachCurveOptions': {'type': 'runtime'},
  'AttachSelectedAsSourceField': {'type': 'runtime'},
  'AttachSubdivSurface': {'type': 'runtime'},
  'AttachSubdivSurfaceOptions': {'type': 'runtime'},
  'AttachSurfaceWithoutMoving': {'type': 'runtime'},
  'AttachSurfaces': {'type': 'runtime'},
  'AttachSurfacesOptions': {'type': 'runtime'},
  'AttachToPath': {'type': 'runtime'},
  'AttachToPathOptions': {'type': 'runtime'},
  'AttributeEditor': {'type': 'runtime'},
  'AutoPaintMarkingMenu': {'type': 'runtime'},
  'AutoPaintMarkingMenuPopDown': {'type': 'runtime'},
  'AutoProjection': {'type': 'runtime'},
  'AutoProjectionOptions': {'type': 'runtime'},
  'AutoSeamUVs': {'type': 'runtime'},
  'AutoSeamUVsOptions': {'type': 'runtime'},
  'AutobindContainer': {'type': 'runtime'},
  'AutobindContainerOptions': {'type': 'runtime'},
  'AveragePolygonNormals': {'type': 'runtime'},
  'AveragePolygonNormalsOptions': {'type': 'runtime'},
  'AverageVertex': {'type': 'runtime'},
  'BakeAllNonDefHistory': {'type': 'runtime'},
  'BakeChannel': {'type': 'runtime'},
  'BakeChannelOptions': {'type': 'runtime'},
  'BakeCustomPivot': {'type': 'runtime'},
  'BakeCustomPivotOptions': {'type': 'runtime'},
  'BakeDeformerTool': {'type': 'runtime'},
  'BakeNonDefHistory': {'type': 'runtime'},
  'BakeNonDefHistoryOptions': {'type': 'runtime'},
  'BakeSimulation': {'type': 'runtime'},
  'BakeSimulationOptions': {'type': 'runtime'},
  'BakeSpringAnimation': {'type': 'runtime'},
  'BakeSpringAnimationOptions': {'type': 'runtime'},
  'BakeSurfaceToTexture': {'type': 'runtime'},
  'BakeTopologyToTargets': {'type': 'runtime'},
  'BaseLevelComponentDisplay': {'type': 'runtime'},
  'BatchBake': {'type': 'runtime'},
  'BatchBakeOptions': {'type': 'runtime'},
  'BatchRender': {'type': 'runtime'},
  'BatchRenderOptions': {'type': 'runtime'},
  'Bend': {'type': 'runtime'},
  'BendCurves': {'type': 'runtime'},
  'BendCurvesOptions': {'type': 'runtime'},
  'BendOptions': {'type': 'runtime'},
  'BestPlaneTexturingTool': {'type': 'runtime'},
  'Bevel': {'type': 'runtime'},
  'BevelOptions': {'type': 'runtime'},
  'BevelPlus': {'type': 'runtime'},
  'BevelPlusOptions': {'type': 'runtime'},
  'BevelPolygon': {'type': 'runtime'},
  'BevelPolygonOptions': {'type': 'runtime'},
  'BezierCurveToNurbs': {'type': 'runtime'},
  'BezierPresetBezier': {'type': 'runtime'},
  'BezierPresetBezierCorner': {'type': 'runtime'},
  'BezierPresetCorner': {'type': 'runtime'},
  'BezierSetAnchorBroken': {'type': 'runtime'},
  'BezierSetAnchorEven': {'type': 'runtime'},
  'BezierSetAnchorSmooth': {'type': 'runtime'},
  'BezierSetAnchorUneven': {'type': 'runtime'},
  'Birail1': {'type': 'runtime'},
  'Birail1Options': {'type': 'runtime'},
  'Birail2': {'type': 'runtime'},
  'Birail2Options': {'type': 'runtime'},
  'Birail3': {'type': 'runtime'},
  'Birail3Options': {'type': 'runtime'},
  'BlendShapeEditor': {'type': 'runtime'},
  'BlindDataEditor': {'type': 'runtime'},
  'BothProxySubdivDisplay': {'type': 'runtime'},
  'Boundary': {'type': 'runtime'},
  'BoundaryOptions': {'type': 'runtime'},
  'BreakLightLinks': {'type': 'runtime'},
  'BreakRigidBodyConnection': {'type': 'runtime'},
  'BreakShadowLinks': {'type': 'runtime'},
  'BreakStereoRigs': {'type': 'runtime'},
  'BreakTangent': {'type': 'runtime'},
  'BreakTangents': {'type': 'runtime'},
  'BridgeEdge': {'type': 'runtime'},
  'BridgeEdgeOptions': {'type': 'runtime'},
  'BridgeOrFill': {'type': 'runtime'},
  'BrushAnimationMarkingMenu': {'type': 'runtime'},
  'BrushAnimationMarkingMenuPopDown': {'type': 'runtime'},
  'BrushPresetBlend': {'type': 'runtime'},
  'BrushPresetBlendOff': {'type': 'runtime'},
  'BrushPresetBlendShading': {'type': 'runtime'},
  'BrushPresetBlendShadingOff': {'type': 'runtime'},
  'BrushPresetBlendShape': {'type': 'runtime'},
  'BrushPresetBlendShapeOff': {'type': 'runtime'},
  'BrushPresetReplaceShading': {'type': 'runtime'},
  'BrushPresetReplaceShadingOff': {'type': 'runtime'},
  'BufferCurveSnapshot': {'type': 'runtime'},
  'CVCurveTool': {'type': 'runtime'},
  'CVCurveToolOptions': {'type': 'runtime'},
  'CVHardness': {'type': 'runtime'},
  'CVHardnessOptions': {'type': 'runtime'},
  'CameraModeOrthographic': {'type': 'runtime'},
  'CameraModePerspective': {'type': 'runtime'},
  'CameraModeToggle': {'type': 'runtime'},
  'CameraRemoveAll': {'type': 'runtime'},
  'CameraRemoveAllForAll': {'type': 'runtime'},
  'CameraRemoveFromExclusive': {'type': 'runtime'},
  'CameraRemoveFromHidden': {'type': 'runtime'},
  'CameraSetEditor': {'type': 'runtime'},
  'CancelBatchRender': {'type': 'runtime'},
  'CenterPivot': {'type': 'runtime'},
  'CenterViewOfSelection': {'type': 'runtime'},
  'ChamferVertex': {'type': 'runtime'},
  'ChamferVertexOptions': {'type': 'runtime'},
  'ChangeAnimPrefs': {'type': 'runtime'},
  'ChangeColorPrefs': {'type': 'runtime'},
  'ChangeEdgeWidth': {'type': 'runtime'},
  'ChangeNormalSize': {'type': 'runtime'},
  'ChangeUIPrefs': {'type': 'runtime'},
  'ChangeUVSize': {'type': 'runtime'},
  'ChangeVertexSize': {'type': 'runtime'},
  'ChannelControlEditor': {'type': 'runtime'},
  'CharacterAnimationEditor': {'type': 'runtime'},
  'CharacterMapper': {'type': 'runtime'},
  'CharacterSetEditor': {'type': 'runtime'},
  'CircularFillet': {'type': 'runtime'},
  'CircularFilletOptions': {'type': 'runtime'},
  'CleanupPolygon': {'type': 'runtime'},
  'CleanupPolygonOptions': {'type': 'runtime'},
  'ClearCurrentCharacterList': {'type': 'runtime'},
  'ClearCurrentContainer': {'type': 'runtime'},
  'ClearInitialState': {'type': 'runtime'},
  'ClearPaintEffectsView': {'type': 'runtime'},
  'CloseFrontWindow': {'type': 'runtime'},
  'ClosestPointOn': {'type': 'runtime'},
  'ClosestPointOnOptions': {'type': 'runtime'},
  'ClusterCurve': {'type': 'runtime'},
  'CoarseLevelComponentDisplay': {'type': 'runtime'},
  'CoarsenSelectedComponents': {'type': 'runtime'},
  'CoarserSubdivLevel': {'type': 'runtime'},
  'CollapseSubdivSurfaceHierarchy': {'type': 'runtime'},
  'CollapseSubdivSurfaceHierarchyOptions': {'type': 'runtime'},
  'ColorPreferencesWindow': {'type': 'runtime'},
  'CombinePolygons': {'type': 'runtime'},
  'CombinePolygonsOptions': {'type': 'runtime'},
  'CommandShell': {'type': 'runtime'},
  'CommandWindow': {'type': 'runtime'},
  'CompleteCurrentTool': {'type': 'runtime'},
  'ComponentEditor': {'type': 'runtime'},
  'ConformPolygon': {'type': 'runtime'},
  'ConformPolygonNormals': {'type': 'runtime'},
  'ConformPolygonOptions': {'type': 'runtime'},
  'ConnectComponents': {'type': 'runtime'},
  'ConnectComponentsOptions': {'type': 'runtime'},
  'ConnectJoint': {'type': 'runtime'},
  'ConnectJointOptions': {'type': 'runtime'},
  'ConnectNodeToIKFK': {'type': 'runtime'},
  'ConnectToTime': {'type': 'runtime'},
  'ConnectionEditor': {'type': 'runtime'},
  'ContentBrowserLayout': {'type': 'runtime'},
  'ContentBrowserWindow': {'type': 'runtime'},
  'ConvertHairSelectionToConstraints': {'type': 'runtime'},
  'ConvertHairSelectionToCurrentPositions': {'type': 'runtime'},
  'ConvertHairSelectionToFollicles': {'type': 'runtime'},
  'ConvertHairSelectionToHairSystems': {'type': 'runtime'},
  'ConvertHairSelectionToRestCurveEndCVs': {'type': 'runtime'},
  'ConvertHairSelectionToRestCurves': {'type': 'runtime'},
  'ConvertHairSelectionToStartAndRestCurveEndCVs': {'type': 'runtime'},
  'ConvertHairSelectionToStartCurveEndCVs': {'type': 'runtime'},
  'ConvertHairSelectionToStartCurves': {'type': 'runtime'},
  'ConvertInstanceToObject': {'type': 'runtime'},
  'ConvertPaintEffectsToPoly': {'type': 'runtime'},
  'ConvertPaintEffectsToPolyOptions': {'type': 'runtime'},
  'ConvertSelectionToContainedEdges': {'type': 'runtime'},
  'ConvertSelectionToContainedFaces': {'type': 'runtime'},
  'ConvertSelectionToEdgePerimeter': {'type': 'runtime'},
  'ConvertSelectionToEdges': {'type': 'runtime'},
  'ConvertSelectionToFacePerimeter': {'type': 'runtime'},
  'ConvertSelectionToFaces': {'type': 'runtime'},
  'ConvertSelectionToShell': {'type': 'runtime'},
  'ConvertSelectionToShellBorder': {'type': 'runtime'},
  'ConvertSelectionToUVBorder': {'type': 'runtime'},
  'ConvertSelectionToUVEdgeLoop': {'type': 'runtime'},
  'ConvertSelectionToUVPerimeter': {'type': 'runtime'},
  'ConvertSelectionToUVShell': {'type': 'runtime'},
  'ConvertSelectionToUVShellBorder': {'type': 'runtime'},
  'ConvertSelectionToUVs': {'type': 'runtime'},
  'ConvertSelectionToVertexFaces': {'type': 'runtime'},
  'ConvertSelectionToVertexPerimeter': {'type': 'runtime'},
  'ConvertSelectionToVertices': {'type': 'runtime'},
  'ConvertToBreakdown': {'type': 'runtime'},
  'ConvertToFrozen': {'type': 'runtime'},
  'ConvertToKey': {'type': 'runtime'},
  'ConvertTypeCapsToCurves': {'type': 'runtime'},
  'CopyFlexor': {'type': 'runtime'},
  'CopyKeys': {'type': 'runtime'},
  'CopyKeysOptions': {'type': 'runtime'},
  'CopyMeshAttributes': {'type': 'runtime'},
  'CopySelected': {'type': 'runtime'},
  'CopySkinWeights': {'type': 'runtime'},
  'CopySkinWeightsOptions': {'type': 'runtime'},
  'CopyUVs': {'type': 'runtime'},
  'CopyUVsToUVSet': {'type': 'runtime'},
  'CopyUVsToUVSetOptions': {'type': 'runtime'},
  'CopyVertexSkinWeights': {'type': 'runtime'},
  'CopyVertexWeights': {'type': 'runtime'},
  'CreaseProxyEdgeTool': {'type': 'runtime'},
  'CreaseProxyEdgeToolOptions': {'type': 'runtime'},
  'Create2DContainer': {'type': 'runtime'},
  'Create2DContainerEmitter': {'type': 'runtime'},
  'Create2DContainerEmitterOptions': {'type': 'runtime'},
  'Create2DContainerOptions': {'type': 'runtime'},
  'Create3DContainer': {'type': 'runtime'},
  'Create3DContainerEmitter': {'type': 'runtime'},
  'Create3DContainerEmitterOptions': {'type': 'runtime'},
  'Create3DContainerOptions': {'type': 'runtime'},
  'CreateActiveRigidBody': {'type': 'runtime'},
  'CreateActiveRigidBodyOptions': {'type': 'runtime'},
  'CreateAmbientLight': {'type': 'runtime'},
  'CreateAmbientLightOptions': {'type': 'runtime'},
  'CreateAnnotateNode': {'type': 'runtime'},
  'CreateAreaLight': {'type': 'runtime'},
  'CreateAreaLightOptions': {'type': 'runtime'},
  'CreateBarrierConstraint': {'type': 'runtime'},
  'CreateBarrierConstraintOptions': {'type': 'runtime'},
  'CreateBezierCurveTool': {'type': 'runtime'},
  'CreateBezierCurveToolOptions': {'type': 'runtime'},
  'CreateBindingSet': {'type': 'runtime'},
  'CreateBlendShape': {'type': 'runtime'},
  'CreateBlendShapeOptions': {'type': 'runtime'},
  'CreateCameraAim': {'type': 'runtime'},
  'CreateCameraAimOptions': {'type': 'runtime'},
  'CreateCameraAimUp': {'type': 'runtime'},
  'CreateCameraAimUpOptions': {'type': 'runtime'},
  'CreateCameraFromView': {'type': 'runtime'},
  'CreateCameraOnly': {'type': 'runtime'},
  'CreateCameraOnlyOptions': {'type': 'runtime'},
  'CreateCharacter': {'type': 'runtime'},
  'CreateCharacterOptions': {'type': 'runtime'},
  'CreateClip': {'type': 'runtime'},
  'CreateClipOptions': {'type': 'runtime'},
  'CreateCluster': {'type': 'runtime'},
  'CreateClusterOptions': {'type': 'runtime'},
  'CreateConstraint': {'type': 'runtime'},
  'CreateConstraintClip': {'type': 'runtime'},
  'CreateConstraintClipOptions': {'type': 'runtime'},
  'CreateConstraintOptions': {'type': 'runtime'},
  'CreateConstructionPlane': {'type': 'runtime'},
  'CreateConstructionPlaneOptions': {'type': 'runtime'},
  'CreateContainer': {'type': 'runtime'},
  'CreateContainerOptions': {'type': 'runtime'},
  'CreateControlRig': {'type': 'runtime'},
  'CreateCreaseSet': {'type': 'runtime'},
  'CreateCreaseSetOptions': {'type': 'runtime'},
  'CreateCurveField': {'type': 'runtime'},
  'CreateCurveFromPoly': {'type': 'runtime'},
  'CreateCurveFromPolyOptions': {'type': 'runtime'},
  'CreateDagContainer': {'type': 'runtime'},
  'CreateDagContainerOptions': {'type': 'runtime'},
  'CreateDirectionalLight': {'type': 'runtime'},
  'CreateDirectionalLightOptions': {'type': 'runtime'},
  'CreateDiskCache': {'type': 'runtime'},
  'CreateDiskCacheOptions': {'type': 'runtime'},
  'CreateEmitter': {'type': 'runtime'},
  'CreateEmitterOptions': {'type': 'runtime'},
  'CreateEmptyGroup': {'type': 'runtime'},
  'CreateEmptySet': {'type': 'runtime'},
  'CreateEmptySetOptions': {'type': 'runtime'},
  'CreateEmptyUVSet': {'type': 'runtime'},
  'CreateEmptyUVSetOptions': {'type': 'runtime'},
  'CreateExpressionClip': {'type': 'runtime'},
  'CreateExpressionClipOptions': {'type': 'runtime'},
  'CreateFlexorWindow': {'type': 'runtime'},
  'CreateFluidCache': {'type': 'runtime'},
  'CreateFluidCacheOptions': {'type': 'runtime'},
  'CreateGhost': {'type': 'runtime'},
  'CreateGhostOptions': {'type': 'runtime'},
  'CreateHair': {'type': 'runtime'},
  'CreateHairCache': {'type': 'runtime'},
  'CreateHairCacheOptions': {'type': 'runtime'},
  'CreateHairOptions': {'type': 'runtime'},
  'CreateHingeConstraint': {'type': 'runtime'},
  'CreateHingeConstraintOptions': {'type': 'runtime'},
  'CreateIllustratorCurves': {'type': 'runtime'},
  'CreateIllustratorCurvesOptions': {'type': 'runtime'},
  'CreateImagePlane': {'type': 'runtime'},
  'CreateImagePlaneOptions': {'type': 'runtime'},
  'CreateJiggleDeformer': {'type': 'runtime'},
  'CreateJiggleOptions': {'type': 'runtime'},
  'CreateLattice': {'type': 'runtime'},
  'CreateLatticeOptions': {'type': 'runtime'},
  'CreateLineModifier': {'type': 'runtime'},
  'CreateLocator': {'type': 'runtime'},
  'CreateMotionTrail': {'type': 'runtime'},
  'CreateMotionTrailOptions': {'type': 'runtime'},
  'CreateMultiStereoRig': {'type': 'runtime'},
  'CreateNSoftBody': {'type': 'runtime'},
  'CreateNSoftBodyOptions': {'type': 'runtime'},
  'CreateNURBSCircle': {'type': 'runtime'},
  'CreateNURBSCircleOptions': {'type': 'runtime'},
  'CreateNURBSCone': {'type': 'runtime'},
  'CreateNURBSConeOptions': {'type': 'runtime'},
  'CreateNURBSCube': {'type': 'runtime'},
  'CreateNURBSCubeOptions': {'type': 'runtime'},
  'CreateNURBSCylinder': {'type': 'runtime'},
  'CreateNURBSCylinderOptions': {'type': 'runtime'},
  'CreateNURBSPlane': {'type': 'runtime'},
  'CreateNURBSPlaneOptions': {'type': 'runtime'},
  'CreateNURBSSphere': {'type': 'runtime'},
  'CreateNURBSSphereOptions': {'type': 'runtime'},
  'CreateNURBSSquare': {'type': 'runtime'},
  'CreateNURBSSquareOptions': {'type': 'runtime'},
  'CreateNURBSTorus': {'type': 'runtime'},
  'CreateNURBSTorusOptions': {'type': 'runtime'},
  'CreateNailConstraint': {'type': 'runtime'},
  'CreateNailConstraintOptions': {'type': 'runtime'},
  'CreateNodeWindow': {'type': 'runtime'},
  'CreateOcean': {'type': 'runtime'},
  'CreateOceanOptions': {'type': 'runtime'},
  'CreateOceanWake': {'type': 'runtime'},
  'CreateOceanWakeOptions': {'type': 'runtime'},
  'CreatePSDTextureItem': {'type': 'runtime'},
  'CreateParticleDiskCache': {'type': 'runtime'},
  'CreateParticleDiskCacheOptions': {'type': 'runtime'},
  'CreatePartition': {'type': 'runtime'},
  'CreatePartitionOptions': {'type': 'runtime'},
  'CreatePassiveRigidBody': {'type': 'runtime'},
  'CreatePassiveRigidBodyOptions': {'type': 'runtime'},
  'CreatePinConstraint': {'type': 'runtime'},
  'CreatePinConstraintOptions': {'type': 'runtime'},
  'CreatePlatonicSolid': {'type': 'runtime'},
  'CreatePlatonicSolidOptions': {'type': 'runtime'},
  'CreatePointLight': {'type': 'runtime'},
  'CreatePointLightOptions': {'type': 'runtime'},
  'CreatePolyFromPreview': {'type': 'runtime'},
  'CreatePolygonCone': {'type': 'runtime'},
  'CreatePolygonConeOptions': {'type': 'runtime'},
  'CreatePolygonCube': {'type': 'runtime'},
  'CreatePolygonCubeOptions': {'type': 'runtime'},
  'CreatePolygonCylinder': {'type': 'runtime'},
  'CreatePolygonCylinderOptions': {'type': 'runtime'},
  'CreatePolygonDisc': {'type': 'runtime'},
  'CreatePolygonDiscOptions': {'type': 'runtime'},
  'CreatePolygonGear': {'type': 'runtime'},
  'CreatePolygonGearOptions': {'type': 'runtime'},
  'CreatePolygonHelix': {'type': 'runtime'},
  'CreatePolygonHelixOptions': {'type': 'runtime'},
  'CreatePolygonPipe': {'type': 'runtime'},
  'CreatePolygonPipeOptions': {'type': 'runtime'},
  'CreatePolygonPlane': {'type': 'runtime'},
  'CreatePolygonPlaneOptions': {'type': 'runtime'},
  'CreatePolygonPlatonic': {'type': 'runtime'},
  'CreatePolygonPlatonicOptions': {'type': 'runtime'},
  'CreatePolygonPrism': {'type': 'runtime'},
  'CreatePolygonPrismOptions': {'type': 'runtime'},
  'CreatePolygonPyramid': {'type': 'runtime'},
  'CreatePolygonPyramidOptions': {'type': 'runtime'},
  'CreatePolygonSVG': {'type': 'runtime'},
  'CreatePolygonSoccerBall': {'type': 'runtime'},
  'CreatePolygonSoccerBallOptions': {'type': 'runtime'},
  'CreatePolygonSphere': {'type': 'runtime'},
  'CreatePolygonSphereOptions': {'type': 'runtime'},
  'CreatePolygonSphericalHarmonics': {'type': 'runtime'},
  'CreatePolygonSphericalHarmonicsOptions': {'type': 'runtime'},
  'CreatePolygonSuperEllipse': {'type': 'runtime'},
  'CreatePolygonSuperEllipseOptions': {'type': 'runtime'},
  'CreatePolygonTool': {'type': 'runtime'},
  'CreatePolygonToolOptions': {'type': 'runtime'},
  'CreatePolygonTorus': {'type': 'runtime'},
  'CreatePolygonTorusOptions': {'type': 'runtime'},
  'CreatePolygonType': {'type': 'runtime'},
  'CreatePolygonUltraShape': {'type': 'runtime'},
  'CreatePolygonUltraShapeOptions': {'type': 'runtime'},
  'CreatePond': {'type': 'runtime'},
  'CreatePondOptions': {'type': 'runtime'},
  'CreatePose': {'type': 'runtime'},
  'CreatePoseInterpolator': {'type': 'runtime'},
  'CreatePoseInterpolatorOptions': {'type': 'runtime'},
  'CreatePoseOptions': {'type': 'runtime'},
  'CreateQuickSelectSet': {'type': 'runtime'},
  'CreateReference': {'type': 'runtime'},
  'CreateReferenceOptions': {'type': 'runtime'},
  'CreateRigidBodySolver': {'type': 'runtime'},
  'CreateSculptDeformer': {'type': 'runtime'},
  'CreateSculptDeformerOptions': {'type': 'runtime'},
  'CreateSet': {'type': 'runtime'},
  'CreateSetOptions': {'type': 'runtime'},
  'CreateShot': {'type': 'runtime'},
  'CreateShotOptions': {'type': 'runtime'},
  'CreateShrinkWrap': {'type': 'runtime'},
  'CreateShrinkWrapOptions': {'type': 'runtime'},
  'CreateSoftBody': {'type': 'runtime'},
  'CreateSoftBodyOptions': {'type': 'runtime'},
  'CreateSpotLight': {'type': 'runtime'},
  'CreateSpotLightOptions': {'type': 'runtime'},
  'CreateSpring': {'type': 'runtime'},
  'CreateSpringConstraint': {'type': 'runtime'},
  'CreateSpringConstraintOptions': {'type': 'runtime'},
  'CreateSpringOptions': {'type': 'runtime'},
  'CreateStereoRig': {'type': 'runtime'},
  'CreateSubCharacter': {'type': 'runtime'},
  'CreateSubCharacterOptions': {'type': 'runtime'},
  'CreateSubdivCone': {'type': 'runtime'},
  'CreateSubdivCube': {'type': 'runtime'},
  'CreateSubdivCylinder': {'type': 'runtime'},
  'CreateSubdivPlane': {'type': 'runtime'},
  'CreateSubdivRegion': {'type': 'runtime'},
  'CreateSubdivSphere': {'type': 'runtime'},
  'CreateSubdivSurface': {'type': 'runtime'},
  'CreateSubdivSurfaceOptions': {'type': 'runtime'},
  'CreateSubdivSurfacePoly': {'type': 'runtime'},
  'CreateSubdivSurfacePolyOptions': {'type': 'runtime'},
  'CreateSubdivTorus': {'type': 'runtime'},
  'CreateText': {'type': 'runtime'},
  'CreateTextOptions': {'type': 'runtime'},
  'CreateTextureDeformer': {'type': 'runtime'},
  'CreateTextureDeformerOptions': {'type': 'runtime'},
  'CreateTextureReferenceObject': {'type': 'runtime'},
  'CreateUVShellAlongBorder': {'type': 'runtime'},
  'CreateUVsBasedOnCamera': {'type': 'runtime'},
  'CreateUVsBasedOnCameraOptions': {'type': 'runtime'},
  'CreateVolumeCone': {'type': 'runtime'},
  'CreateVolumeCube': {'type': 'runtime'},
  'CreateVolumeLight': {'type': 'runtime'},
  'CreateVolumeLightOptions': {'type': 'runtime'},
  'CreateVolumeSphere': {'type': 'runtime'},
  'CreateWake': {'type': 'runtime'},
  'CreateWakeOptions': {'type': 'runtime'},
  'CreateWrap': {'type': 'runtime'},
  'CreateWrapOptions': {'type': 'runtime'},
  'CurlCurves': {'type': 'runtime'},
  'CurlCurvesOptions': {'type': 'runtime'},
  'CurveEditTool': {'type': 'runtime'},
  'CurveFillet': {'type': 'runtime'},
  'CurveFilletOptions': {'type': 'runtime'},
  'CurveFlow': {'type': 'runtime'},
  'CurveFlowOptions': {'type': 'runtime'},
  'CurveSmoothnessCoarse': {'type': 'runtime'},
  'CurveSmoothnessFine': {'type': 'runtime'},
  'CurveSmoothnessMedium': {'type': 'runtime'},
  'CurveSmoothnessRough': {'type': 'runtime'},
  'CurveUtilitiesMarkingMenu': {'type': 'runtime'},
  'CurveUtilitiesMarkingMenuPopDown': {'type': 'runtime'},
  'CurveWarp': {'type': 'runtime'},
  'CustomNURBSComponentsOptions': {'type': 'runtime'},
  'CustomNURBSSmoothness': {'type': 'runtime'},
  'CustomNURBSSmoothnessOptions': {'type': 'runtime'},
  'CustomPolygonDisplay': {'type': 'runtime'},
  'CustomPolygonDisplayOptions': {'type': 'runtime'},
  'CutCurve': {'type': 'runtime'},
  'CutCurveOptions': {'type': 'runtime'},
  'CutKeys': {'type': 'runtime'},
  'CutKeysOptions': {'type': 'runtime'},
  'CutPolygon': {'type': 'runtime'},
  'CutPolygonOptions': {'type': 'runtime'},
  'CutSelected': {'type': 'runtime'},
  'CutUVs': {'type': 'runtime'},
  'CutUVs3D': {'type': 'runtime'},
  'CutUVsWithoutHotkey': {'type': 'runtime'},
  'CycleBackgroundColor': {'type': 'runtime'},
  'CycleIKHandleStickyState': {'type': 'runtime'},
  'CycleThroughCameras': {'type': 'runtime'},
  'DeactivateGlobalScreenSlider': {'type': 'runtime'},
  'DeactivateGlobalScreenSliderModeMarkingMenu': {'type': 'runtime'},
  'DecreaseCheckerDensity': {'type': 'runtime'},
  'DecreaseExposureCoarse': {'type': 'runtime'},
  'DecreaseExposureFine': {'type': 'runtime'},
  'DecreaseGammaCoarse': {'type': 'runtime'},
  'DecreaseGammaFine': {'type': 'runtime'},
  'DecreaseManipulatorSize': {'type': 'runtime'},
  'DecrementFluidCenter': {'type': 'runtime'},
  'DefaultQualityDisplay': {'type': 'runtime'},
  'DeformerSetEditor': {'type': 'runtime'},
  'Delete': {'type': 'runtime'},
  'DeleteAllCameras': {'type': 'runtime'},
  'DeleteAllChannels': {'type': 'runtime'},
  'DeleteAllClips': {'type': 'runtime'},
  'DeleteAllClusters': {'type': 'runtime'},
  'DeleteAllConstraints': {'type': 'runtime'},
  'DeleteAllContainers': {'type': 'runtime'},
  'DeleteAllControllers': {'type': 'runtime'},
  'DeleteAllDynamicConstraints': {'type': 'runtime'},
  'DeleteAllExpressions': {'type': 'runtime'},
  'DeleteAllFluids': {'type': 'runtime'},
  'DeleteAllFurs': {'type': 'runtime'},
  'DeleteAllHistory': {'type': 'runtime'},
  'DeleteAllIKHandles': {'type': 'runtime'},
  'DeleteAllImagePlanes': {'type': 'runtime'},
  'DeleteAllJoints': {'type': 'runtime'},
  'DeleteAllLattices': {'type': 'runtime'},
  'DeleteAllLights': {'type': 'runtime'},
  'DeleteAllMotionPaths': {'type': 'runtime'},
  'DeleteAllNCloths': {'type': 'runtime'},
  'DeleteAllNParticles': {'type': 'runtime'},
  'DeleteAllNRigids': {'type': 'runtime'},
  'DeleteAllNonLinearDeformers': {'type': 'runtime'},
  'DeleteAllParticles': {'type': 'runtime'},
  'DeleteAllPoses': {'type': 'runtime'},
  'DeleteAllRigidBodies': {'type': 'runtime'},
  'DeleteAllRigidConstraints': {'type': 'runtime'},
  'DeleteAllSculptObjects': {'type': 'runtime'},
  'DeleteAllShadingGroupsAndMaterials': {'type': 'runtime'},
  'DeleteAllSounds': {'type': 'runtime'},
  'DeleteAllStaticChannels': {'type': 'runtime'},
  'DeleteAllStrokes': {'type': 'runtime'},
  'DeleteAllWires': {'type': 'runtime'},
  'DeleteAttribute': {'type': 'runtime'},
  'DeleteChannels': {'type': 'runtime'},
  'DeleteChannelsOptions': {'type': 'runtime'},
  'DeleteConstraints': {'type': 'runtime'},
  'DeleteCurrentColorSet': {'type': 'runtime'},
  'DeleteCurrentSet': {'type': 'runtime'},
  'DeleteCurrentUVSet': {'type': 'runtime'},
  'DeleteCurrentWorkspace': {'type': 'runtime'},
  'DeleteEdge': {'type': 'runtime'},
  'DeleteEntireHairSystem': {'type': 'runtime'},
  'DeleteExpressions': {'type': 'runtime'},
  'DeleteExpressionsOptions': {'type': 'runtime'},
  'DeleteHair': {'type': 'runtime'},
  'DeleteHairCache': {'type': 'runtime'},
  'DeleteHistory': {'type': 'runtime'},
  'DeleteKeys': {'type': 'runtime'},
  'DeleteKeysOptions': {'type': 'runtime'},
  'DeleteMemoryCaching': {'type': 'runtime'},
  'DeleteMotionPaths': {'type': 'runtime'},
  'DeletePolyElements': {'type': 'runtime'},
  'DeleteRigidBodies': {'type': 'runtime'},
  'DeleteSelectedContainers': {'type': 'runtime'},
  'DeleteStaticChannels': {'type': 'runtime'},
  'DeleteStaticChannelsOptions': {'type': 'runtime'},
  'DeleteSurfaceFlow': {'type': 'runtime'},
  'DeleteSurfaceFlowOptions': {'type': 'runtime'},
  'DeleteTextureReferenceObject': {'type': 'runtime'},
  'DeleteTimeWarp': {'type': 'runtime'},
  'DeleteUVs': {'type': 'runtime'},
  'DeleteUVsWithoutHotkey': {'type': 'runtime'},
  'DeleteUnknownNodes': {'type': 'runtime'},
  'DeleteVertex': {'type': 'runtime'},
  'DeltaMush': {'type': 'runtime'},
  'DeltaMushOptions': {'type': 'runtime'},
  'DetachComponent': {'type': 'runtime'},
  'DetachCurve': {'type': 'runtime'},
  'DetachCurveOptions': {'type': 'runtime'},
  'DetachEdgeComponent': {'type': 'runtime'},
  'DetachSkeleton': {'type': 'runtime'},
  'DetachSkeletonJoints': {'type': 'runtime'},
  'DetachSkin': {'type': 'runtime'},
  'DetachSkinOptions': {'type': 'runtime'},
  'DetachSurfaces': {'type': 'runtime'},
  'DetachSurfacesOptions': {'type': 'runtime'},
  'DetachVertexComponent': {'type': 'runtime'},
  'DeviceEditor': {'type': 'runtime'},
  'DisableAll': {'type': 'runtime'},
  'DisableAllCaches': {'type': 'runtime'},
  'DisableAllGeometryCache': {'type': 'runtime'},
  'DisableConstraints': {'type': 'runtime'},
  'DisableExpressions': {'type': 'runtime'},
  'DisableFluids': {'type': 'runtime'},
  'DisableGlobalStitch': {'type': 'runtime'},
  'DisableIKSolvers': {'type': 'runtime'},
  'DisableMemoryCaching': {'type': 'runtime'},
  'DisableParticles': {'type': 'runtime'},
  'DisableRigidBodies': {'type': 'runtime'},
  'DisableSelectedIKHandles': {'type': 'runtime'},
  'DisableSnapshots': {'type': 'runtime'},
  'DisableTimeChangeUndoConsolidation': {'type': 'runtime'},
  'DisableWeightNrm': {'type': 'runtime'},
  'DisconnectJoint': {'type': 'runtime'},
  'DisplacementToPolygon': {'type': 'runtime'},
  'DisplayCurrentHairCurves': {'type': 'runtime'},
  'DisplayHairCurves': {'type': 'runtime'},
  'DisplayHairCurvesCurrentAndRest': {'type': 'runtime'},
  'DisplayHairCurvesCurrentAndStart': {'type': 'runtime'},
  'DisplayHairCurvesRestPosition': {'type': 'runtime'},
  'DisplayHairCurvesStart': {'type': 'runtime'},
  'DisplayIntermediateObjects': {'type': 'runtime'},
  'DisplayLayerEditorWindow': {'type': 'runtime'},
  'DisplayLight': {'type': 'runtime'},
  'DisplayShaded': {'type': 'runtime'},
  'DisplayShadedAndTextured': {'type': 'runtime'},
  'DisplayShadingMarkingMenu': {'type': 'runtime'},
  'DisplayShadingMarkingMenuPopDown': {'type': 'runtime'},
  'DisplaySmoothShaded': {'type': 'runtime'},
  'DisplayUVShaded': {'type': 'runtime'},
  'DisplayUVWireframe': {'type': 'runtime'},
  'DisplayViewport': {'type': 'runtime'},
  'DisplayWireframe': {'type': 'runtime'},
  'DistanceTool': {'type': 'runtime'},
  'DistributeShells': {'type': 'runtime'},
  'DistributeShellsOptions': {'type': 'runtime'},
  'DistributeUVs': {'type': 'runtime'},
  'DistributeUVsOptions': {'type': 'runtime'},
  'DoUnghost': {'type': 'runtime'},
  'DoUnghostOptions': {'type': 'runtime'},
  'DollyTool': {'type': 'runtime'},
  'DopeSheetEditor': {'type': 'runtime'},
  'DownloadBonusTools': {'type': 'runtime'},
  'Drag': {'type': 'runtime'},
  'DragOptions': {'type': 'runtime'},
  'Duplicate': {'type': 'runtime'},
  'DuplicateCurve': {'type': 'runtime'},
  'DuplicateCurveOptions': {'type': 'runtime'},
  'DuplicateEdges': {'type': 'runtime'},
  'DuplicateEdgesOptions': {'type': 'runtime'},
  'DuplicateFace': {'type': 'runtime'},
  'DuplicateFaceOptions': {'type': 'runtime'},
  'DuplicateNURBSPatches': {'type': 'runtime'},
  'DuplicateNURBSPatchesOptions': {'type': 'runtime'},
  'DuplicateSpecial': {'type': 'runtime'},
  'DuplicateSpecialOptions': {'type': 'runtime'},
  'DuplicateWithTransform': {'type': 'runtime'},
  'DynamicRelationshipEditor': {'type': 'runtime'},
  'EPCurveTool': {'type': 'runtime'},
  'EPCurveToolOptions': {'type': 'runtime'},
  'EditAssignedSet': {'type': 'runtime'},
  'EditCharacterAttributes': {'type': 'runtime'},
  'EditFluidResolution': {'type': 'runtime'},
  'EditFluidResolutionOptions': {'type': 'runtime'},
  'EditMembershipTool': {'type': 'runtime'},
  'EditNormalizationGroups': {'type': 'runtime'},
  'EditOversamplingForCacheSettings': {'type': 'runtime'},
  'EditPSDTextureItem': {'type': 'runtime'},
  'EditPolygonType': {'type': 'runtime'},
  'EditTexture': {'type': 'runtime'},
  'EmitFluidFromObject': {'type': 'runtime'},
  'EmitFluidFromObjectOptions': {'type': 'runtime'},
  'EmitFromObject': {'type': 'runtime'},
  'EmitFromObjectOptions': {'type': 'runtime'},
  'EmptyAnimLayer': {'type': 'runtime'},
  'EnableAll': {'type': 'runtime'},
  'EnableAllCaches': {'type': 'runtime'},
  'EnableAllGeometryCache': {'type': 'runtime'},
  'EnableConstraints': {'type': 'runtime'},
  'EnableDynamicConstraints': {'type': 'runtime'},
  'EnableExpressions': {'type': 'runtime'},
  'EnableFluids': {'type': 'runtime'},
  'EnableGlobalStitch': {'type': 'runtime'},
  'EnableIKSolvers': {'type': 'runtime'},
  'EnableMemoryCaching': {'type': 'runtime'},
  'EnableNCloths': {'type': 'runtime'},
  'EnableNParticles': {'type': 'runtime'},
  'EnableNRigids': {'type': 'runtime'},
  'EnableNucleuses': {'type': 'runtime'},
  'EnableParticles': {'type': 'runtime'},
  'EnableRigidBodies': {'type': 'runtime'},
  'EnableSelectTool': {'type': 'runtime'},
  'EnableSelectedIKHandles': {'type': 'runtime'},
  'EnableSnapshots': {'type': 'runtime'},
  'EnableTimeChangeUndoConsolidation': {'type': 'runtime'},
  'EnableTimeWarp': {'type': 'runtime'},
  'EnableWeightNrm': {'type': 'runtime'},
  'EnableWeightPostNrm': {'type': 'runtime'},
  'EnterConnectTool': {'type': 'runtime'},
  'EnterEditMode': {'type': 'runtime'},
  'EnterEditModePress': {'type': 'runtime'},
  'EnterEditModeRelease': {'type': 'runtime'},
  'EvaluationToolkit': {'type': 'runtime'},
  'ExpandSelectedComponents': {'type': 'runtime'},
  'Export': {'type': 'runtime'},
  'ExportAnim': {'type': 'runtime'},
  'ExportDeformerWeights': {'type': 'runtime'},
  'ExportDeformerWeightsOptions': {'type': 'runtime'},
  'ExportOfflineFile': {'type': 'runtime'},
  'ExportOfflineFileFromRefEd': {'type': 'runtime'},
  'ExportOfflineFileFromRefEdOptions': {'type': 'runtime'},
  'ExportOfflineFileOptions': {'type': 'runtime'},
  'ExportOptions': {'type': 'runtime'},
  'ExportProxyContainer': {'type': 'runtime'},
  'ExportProxyContainerOptions': {'type': 'runtime'},
  'ExportSelection': {'type': 'runtime'},
  'ExportSelectionOptions': {'type': 'runtime'},
  'ExportSkinWeightMaps': {'type': 'runtime'},
  'ExportSkinWeightMapsOptions': {'type': 'runtime'},
  'ExpressionEditor': {'type': 'runtime'},
  'ExtendCurve': {'type': 'runtime'},
  'ExtendCurveOnSurface': {'type': 'runtime'},
  'ExtendCurveOnSurfaceOptions': {'type': 'runtime'},
  'ExtendCurveOptions': {'type': 'runtime'},
  'ExtendFluid': {'type': 'runtime'},
  'ExtendFluidOptions': {'type': 'runtime'},
  'ExtendSurfaces': {'type': 'runtime'},
  'ExtendSurfacesOptions': {'type': 'runtime'},
  'ExtractFace': {'type': 'runtime'},
  'ExtractFaceOptions': {'type': 'runtime'},
  'ExtractSubdivSurfaceVertices': {'type': 'runtime'},
  'ExtractSubdivSurfaceVerticesOptions': {'type': 'runtime'},
  'Extrude': {'type': 'runtime'},
  'ExtrudeEdge': {'type': 'runtime'},
  'ExtrudeEdgeOptions': {'type': 'runtime'},
  'ExtrudeFace': {'type': 'runtime'},
  'ExtrudeFaceOptions': {'type': 'runtime'},
  'ExtrudeOptions': {'type': 'runtime'},
  'ExtrudeVertex': {'type': 'runtime'},
  'ExtrudeVertexOptions': {'type': 'runtime'},
  'FBIKLabelArm': {'type': 'runtime'},
  'FBIKLabelBigToe': {'type': 'runtime'},
  'FBIKLabelCenter': {'type': 'runtime'},
  'FBIKLabelCollar': {'type': 'runtime'},
  'FBIKLabelElbow': {'type': 'runtime'},
  'FBIKLabelExtraFinger': {'type': 'runtime'},
  'FBIKLabelFingerBase': {'type': 'runtime'},
  'FBIKLabelFoot': {'type': 'runtime'},
  'FBIKLabelFootThumb': {'type': 'runtime'},
  'FBIKLabelHand': {'type': 'runtime'},
  'FBIKLabelHead': {'type': 'runtime'},
  'FBIKLabelHip': {'type': 'runtime'},
  'FBIKLabelIndex': {'type': 'runtime'},
  'FBIKLabelIndexToe': {'type': 'runtime'},
  'FBIKLabelKnee': {'type': 'runtime'},
  'FBIKLabelLeft': {'type': 'runtime'},
  'FBIKLabelLeg': {'type': 'runtime'},
  'FBIKLabelMiddleFinger': {'type': 'runtime'},
  'FBIKLabelMiddleToe': {'type': 'runtime'},
  'FBIKLabelNeck': {'type': 'runtime'},
  'FBIKLabelNone': {'type': 'runtime'},
  'FBIKLabelPinky': {'type': 'runtime'},
  'FBIKLabelPinkyToe': {'type': 'runtime'},
  'FBIKLabelRight': {'type': 'runtime'},
  'FBIKLabelRingFinger': {'type': 'runtime'},
  'FBIKLabelRingToe': {'type': 'runtime'},
  'FBIKLabelRoot': {'type': 'runtime'},
  'FBIKLabelShoulder': {'type': 'runtime'},
  'FBIKLabelSpine': {'type': 'runtime'},
  'FBIKLabelThumb': {'type': 'runtime'},
  'FBIKLabelToeBase': {'type': 'runtime'},
  'FilePathEditor': {'type': 'runtime'},
  'FillHole': {'type': 'runtime'},
  'FilletBlendTool': {'type': 'runtime'},
  'FilletBlendToolOptions': {'type': 'runtime'},
  'FineLevelComponentDisplay': {'type': 'runtime'},
  'Fire': {'type': 'runtime'},
  'FireOptions': {'type': 'runtime'},
  'Fireworks': {'type': 'runtime'},
  'FireworksOptions': {'type': 'runtime'},
  'FitBSpline': {'type': 'runtime'},
  'FitBSplineOptions': {'type': 'runtime'},
  'Flare': {'type': 'runtime'},
  'FlareOptions': {'type': 'runtime'},
  'FlipMesh': {'type': 'runtime'},
  'FlipTriangleEdge': {'type': 'runtime'},
  'FlipTubeDirection': {'type': 'runtime'},
  'FlipUVs': {'type': 'runtime'},
  'FlipUVsOptions': {'type': 'runtime'},
  'FloatSelectedObjects': {'type': 'runtime'},
  'FloatSelectedObjectsOptions': {'type': 'runtime'},
  'FloatSelectedPondObjects': {'type': 'runtime'},
  'FloatSelectedPondObjectsOptions': {'type': 'runtime'},
  'FloodSurfaces': {'type': 'runtime'},
  'FlowPathObject': {'type': 'runtime'},
  'FlowPathObjectOptions': {'type': 'runtime'},
  'FluidEmitter': {'type': 'runtime'},
  'FluidEmitterOptions': {'type': 'runtime'},
  'FluidGradients': {'type': 'runtime'},
  'FluidGradientsOptions': {'type': 'runtime'},
  'FluidsToPolygons': {'type': 'runtime'},
  'FourViewArrangement': {'type': 'runtime'},
  'FourViewLayout': {'type': 'runtime'},
  'FrameAll': {'type': 'runtime'},
  'FrameAllInAllViews': {'type': 'runtime'},
  'FrameSelected': {'type': 'runtime'},
  'FrameSelected2': {'type': 'runtime'},
  'FrameSelectedInAllViews': {'type': 'runtime'},
  'FrameSelectedWithoutChildren': {'type': 'runtime'},
  'FrameSelectedWithoutChildrenInAllViews': {'type': 'runtime'},
  'FreeTangentWeight': {'type': 'runtime'},
  'FreeformFillet': {'type': 'runtime'},
  'FreeformFilletOptions': {'type': 'runtime'},
  'FreezeTransformations': {'type': 'runtime'},
  'FreezeTransformationsOptions': {'type': 'runtime'},
  'FrontPerspViewLayout': {'type': 'runtime'},
  'FullCreaseSubdivSurface': {'type': 'runtime'},
  'FullHotboxDisplay': {'type': 'runtime'},
  'GameExporterWnd': {'type': 'runtime'},
  'GeometryConstraint': {'type': 'runtime'},
  'GeometryConstraintOptions': {'type': 'runtime'},
  'GeometryToBoundingBox': {'type': 'runtime'},
  'GeometryToBoundingBoxOptions': {'type': 'runtime'},
  'GetCartoonExample': {'type': 'runtime'},
  'GetFluidExample': {'type': 'runtime'},
  'GetHairExample': {'type': 'runtime'},
  'GetOceanPondExample': {'type': 'runtime'},
  'GetSettingsFromSelectedStroke': {'type': 'runtime'},
  'GetToonExample': {'type': 'runtime'},
  'GhostObject': {'type': 'runtime'},
  'GlobalDiskCacheControl': {'type': 'runtime'},
  'GlobalStitch': {'type': 'runtime'},
  'GlobalStitchOptions': {'type': 'runtime'},
  'GoToBindPose': {'type': 'runtime'},
  'GoToDefaultView': {'type': 'runtime'},
  'GoToMaxFrame': {'type': 'runtime'},
  'GoToMinFrame': {'type': 'runtime'},
  'GoToNextDrivenKey': {'type': 'runtime'},
  'GoToPreviousDrivenKey': {'type': 'runtime'},
  'GoToWorkingFrame': {'type': 'runtime'},
  'Goal': {'type': 'runtime'},
  'GoalOptions': {'type': 'runtime'},
  'GraphCopy': {'type': 'runtime'},
  'GraphCopyOptions': {'type': 'runtime'},
  'GraphCut': {'type': 'runtime'},
  'GraphCutOptions': {'type': 'runtime'},
  'GraphDelete': {'type': 'runtime'},
  'GraphDeleteOptions': {'type': 'runtime'},
  'GraphEditor': {'type': 'runtime'},
  'GraphEditorAbsoluteView': {'type': 'runtime'},
  'GraphEditorAlwaysDisplayTangents': {'type': 'runtime'},
  'GraphEditorDisableCurveSelection': {'type': 'runtime'},
  'GraphEditorDisplayTangentActive': {'type': 'runtime'},
  'GraphEditorDisplayValues': {'type': 'runtime'},
  'GraphEditorEnableCurveSelection': {'type': 'runtime'},
  'GraphEditorFrameAll': {'type': 'runtime'},
  'GraphEditorFrameCenterView': {'type': 'runtime'},
  'GraphEditorFramePlaybackRange': {'type': 'runtime'},
  'GraphEditorFrameSelected': {'type': 'runtime'},
  'GraphEditorLockChannel': {'type': 'runtime'},
  'GraphEditorNeverDisplayTangents': {'type': 'runtime'},
  'GraphEditorNormalizedView': {'type': 'runtime'},
  'GraphEditorStackedView': {'type': 'runtime'},
  'GraphEditorUnlockChannel': {'type': 'runtime'},
  'GraphEditorValueLinesToggle': {'type': 'runtime'},
  'GraphPaste': {'type': 'runtime'},
  'GraphPasteOptions': {'type': 'runtime'},
  'GraphSnap': {'type': 'runtime'},
  'GraphSnapOptions': {'type': 'runtime'},
  'Gravity': {'type': 'runtime'},
  'GravityOptions': {'type': 'runtime'},
  'GreasePencilTool': {'type': 'runtime'},
  'GridOptions': {'type': 'runtime'},
  'GridUV': {'type': 'runtime'},
  'GridUVOptions': {'type': 'runtime'},
  'Group': {'type': 'runtime'},
  'GroupOptions': {'type': 'runtime'},
  'GrowLoopPolygonSelectionRegion': {'type': 'runtime'},
  'GrowPolygonSelectionRegion': {'type': 'runtime'},
  'HIKBodyPartMode': {'type': 'runtime'},
  'HIKCharacterControlsTool': {'type': 'runtime'},
  'HIKCycleMode': {'type': 'runtime'},
  'HIKFullBodyMode': {'type': 'runtime'},
  'HIKLiveConnectionTool': {'type': 'runtime'},
  'HIKPinRotate': {'type': 'runtime'},
  'HIKPinTranslate': {'type': 'runtime'},
  'HIKSelectedMode': {'type': 'runtime'},
  'HIKSetBodyPartKey': {'type': 'runtime'},
  'HIKSetFullBodyKey': {'type': 'runtime'},
  'HIKSetSelectionKey': {'type': 'runtime'},
  'HIKToggleReleasePinning': {'type': 'runtime'},
  'HairScaleTool': {'type': 'runtime'},
  'HairUVSetLinkingEditor': {'type': 'runtime'},
  'HardwareRenderBuffer': {'type': 'runtime'},
  'Help': {'type': 'runtime'},
  'HideAll': {'type': 'runtime'},
  'HideAllLabels': {'type': 'runtime'},
  'HideBoundingBox': {'type': 'runtime'},
  'HideCameraManipulators': {'type': 'runtime'},
  'HideCameras': {'type': 'runtime'},
  'HideClusters': {'type': 'runtime'},
  'HideControllers': {'type': 'runtime'},
  'HideDeformers': {'type': 'runtime'},
  'HideDeformingGeometry': {'type': 'runtime'},
  'HideDynamicConstraints': {'type': 'runtime'},
  'HideFluids': {'type': 'runtime'},
  'HideFollicles': {'type': 'runtime'},
  'HideFur': {'type': 'runtime'},
  'HideGeometry': {'type': 'runtime'},
  'HideHairSystems': {'type': 'runtime'},
  'HideHotbox': {'type': 'runtime'},
  'HideIKHandles': {'type': 'runtime'},
  'HideIntermediateObjects': {'type': 'runtime'},
  'HideJoints': {'type': 'runtime'},
  'HideKinematics': {'type': 'runtime'},
  'HideLattices': {'type': 'runtime'},
  'HideLightManipulators': {'type': 'runtime'},
  'HideLights': {'type': 'runtime'},
  'HideManipulators': {'type': 'runtime'},
  'HideMarkers': {'type': 'runtime'},
  'HideNCloths': {'type': 'runtime'},
  'HideNParticles': {'type': 'runtime'},
  'HideNRigids': {'type': 'runtime'},
  'HideNURBSCurves': {'type': 'runtime'},
  'HideNURBSSurfaces': {'type': 'runtime'},
  'HideNonlinears': {'type': 'runtime'},
  'HideObjectGeometry': {'type': 'runtime'},
  'HidePlanes': {'type': 'runtime'},
  'HidePolygonSurfaces': {'type': 'runtime'},
  'HideSculptObjects': {'type': 'runtime'},
  'HideSelectedObjects': {'type': 'runtime'},
  'HideSmoothSkinInfluences': {'type': 'runtime'},
  'HideStrokeControlCurves': {'type': 'runtime'},
  'HideStrokePathCurves': {'type': 'runtime'},
  'HideStrokes': {'type': 'runtime'},
  'HideSubdivSurfaces': {'type': 'runtime'},
  'HideTexturePlacements': {'type': 'runtime'},
  'HideUIElements': {'type': 'runtime'},
  'HideUnselectedCVs': {'type': 'runtime'},
  'HideUnselectedObjects': {'type': 'runtime'},
  'HideWrapInfluences': {'type': 'runtime'},
  'HighQualityDisplay': {'type': 'runtime'},
  'HighlightWhatsNew': {'type': 'runtime'},
  'HoldCurrentKeys': {'type': 'runtime'},
  'HotkeyPreferencesWindow': {'type': 'runtime'},
  'HyperGraphPanelRedoViewChange': {'type': 'runtime'},
  'HyperGraphPanelUndoViewChange': {'type': 'runtime'},
  'HypergraphDGWindow': {'type': 'runtime'},
  'HypergraphDecreaseDepth': {'type': 'runtime'},
  'HypergraphHierarchyWindow': {'type': 'runtime'},
  'HypergraphIncreaseDepth': {'type': 'runtime'},
  'HypergraphWindow': {'type': 'runtime'},
  'HypershadeAddOnNodeCreate': {'type': 'runtime'},
  'HypershadeAdditiveGraphingMode': {'type': 'runtime'},
  'HypershadeAutoSizeNodes': {'type': 'runtime'},
  'HypershadeCloseActiveTab': {'type': 'runtime'},
  'HypershadeCloseAllTabs': {'type': 'runtime'},
  'HypershadeCollapseAsset': {'type': 'runtime'},
  'HypershadeConnectSelected': {'type': 'runtime'},
  'HypershadeConvertPSDToFileTexture': {'type': 'runtime'},
  'HypershadeConvertPSDToLayeredTexture': {'type': 'runtime'},
  'HypershadeConvertToFileTexture': {'type': 'runtime'},
  'HypershadeConvertToFileTextureOptionBox': {'type': 'runtime'},
  'HypershadeCreateAsset': {'type': 'runtime'},
  'HypershadeCreateContainerOptions': {'type': 'runtime'},
  'HypershadeCreateNewTab': {'type': 'runtime'},
  'HypershadeCreatePSDFile': {'type': 'runtime'},
  'HypershadeCreateTab': {'type': 'runtime'},
  'HypershadeDeleteAllBakeSets': {'type': 'runtime'},
  'HypershadeDeleteAllCamerasAndImagePlanes': {'type': 'runtime'},
  'HypershadeDeleteAllLights': {'type': 'runtime'},
  'HypershadeDeleteAllShadingGroupsAndMaterials': {'type': 'runtime'},
  'HypershadeDeleteAllTextures': {'type': 'runtime'},
  'HypershadeDeleteAllUtilities': {'type': 'runtime'},
  'HypershadeDeleteDuplicateShadingNetworks': {'type': 'runtime'},
  'HypershadeDeleteNodes': {'type': 'runtime'},
  'HypershadeDeleteSelected': {'type': 'runtime'},
  'HypershadeDeleteUnusedNodes': {'type': 'runtime'},
  'HypershadeDisplayAllShapes': {'type': 'runtime'},
  'HypershadeDisplayAsExtraLargeSwatches': {'type': 'runtime'},
  'HypershadeDisplayAsIcons': {'type': 'runtime'},
  'HypershadeDisplayAsLargeSwatches': {'type': 'runtime'},
  'HypershadeDisplayAsList': {'type': 'runtime'},
  'HypershadeDisplayAsMediumSwatches': {'type': 'runtime'},
  'HypershadeDisplayAsSmallSwatches': {'type': 'runtime'},
  'HypershadeDisplayInterestingShapes': {'type': 'runtime'},
  'HypershadeDisplayNoShapes': {'type': 'runtime'},
  'HypershadeDuplicateShadingNetwork': {'type': 'runtime'},
  'HypershadeDuplicateWithConnections': {'type': 'runtime'},
  'HypershadeDuplicateWithoutNetwork': {'type': 'runtime'},
  'HypershadeEditPSDFile': {'type': 'runtime'},
  'HypershadeEditTexture': {'type': 'runtime'},
  'HypershadeExpandAsset': {'type': 'runtime'},
  'HypershadeExportSelectedNetwork': {'type': 'runtime'},
  'HypershadeFrameAll': {'type': 'runtime'},
  'HypershadeFrameSelected': {'type': 'runtime'},
  'HypershadeGraphAddSelected': {'type': 'runtime'},
  'HypershadeGraphClearGraph': {'type': 'runtime'},
  'HypershadeGraphDownstream': {'type': 'runtime'},
  'HypershadeGraphMaterialsOnSelectedObjects': {'type': 'runtime'},
  'HypershadeGraphRearrange': {'type': 'runtime'},
  'HypershadeGraphRemoveDownstream': {'type': 'runtime'},
  'HypershadeGraphRemoveSelected': {'type': 'runtime'},
  'HypershadeGraphRemoveUnselected': {'type': 'runtime'},
  'HypershadeGraphRemoveUpstream': {'type': 'runtime'},
  'HypershadeGraphUpDownstream': {'type': 'runtime'},
  'HypershadeGraphUpstream': {'type': 'runtime'},
  'HypershadeGridToggleSnap': {'type': 'runtime'},
  'HypershadeGridToggleVisibility': {'type': 'runtime'},
  'HypershadeHideAttributes': {'type': 'runtime'},
  'HypershadeImport': {'type': 'runtime'},
  'HypershadeIncreaseTraversalDepth': {'type': 'runtime'},
  'HypershadeMoveTabDown': {'type': 'runtime'},
  'HypershadeMoveTabLeft': {'type': 'runtime'},
  'HypershadeMoveTabRight': {'type': 'runtime'},
  'HypershadeMoveTabUp': {'type': 'runtime'},
  'HypershadeOpenBinsWindow': {'type': 'runtime'},
  'HypershadeOpenBrowserWindow': {'type': 'runtime'},
  'HypershadeOpenConnectWindow': {'type': 'runtime'},
  'HypershadeOpenCreateWindow': {'type': 'runtime'},
  'HypershadeOpenGraphEditorWindow': {'type': 'runtime'},
  'HypershadeOpenMaterialViewerWindow': {'type': 'runtime'},
  'HypershadeOpenModelEditorWindow': {'type': 'runtime'},
  'HypershadeOpenOutlinerWindow': {'type': 'runtime'},
  'HypershadeOpenPropertyEditorWindow': {'type': 'runtime'},
  'HypershadeOpenRenderViewWindow': {'type': 'runtime'},
  'HypershadeOpenSpreadSheetWindow': {'type': 'runtime'},
  'HypershadeOpenUVEditorWindow': {'type': 'runtime'},
  'HypershadeOutlinerPerspLayout': {'type': 'runtime'},
  'HypershadePerspLayout': {'type': 'runtime'},
  'HypershadePickWalkDown': {'type': 'runtime'},
  'HypershadePickWalkLeft': {'type': 'runtime'},
  'HypershadePickWalkRight': {'type': 'runtime'},
  'HypershadePickWalkUp': {'type': 'runtime'},
  'HypershadePinByDefault': {'type': 'runtime'},
  'HypershadePinSelected': {'type': 'runtime'},
  'HypershadePublishConnections': {'type': 'runtime'},
  'HypershadeReduceTraversalDepth': {'type': 'runtime'},
  'HypershadeRefreshAllSwatchesOnDisk': {'type': 'runtime'},
  'HypershadeRefreshFileListing': {'type': 'runtime'},
  'HypershadeRefreshSelectedSwatches': {'type': 'runtime'},
  'HypershadeRefreshSelectedSwatchesOnDisk': {'type': 'runtime'},
  'HypershadeRemoveAsset': {'type': 'runtime'},
  'HypershadeRemoveTab': {'type': 'runtime'},
  'HypershadeRenameActiveTab': {'type': 'runtime'},
  'HypershadeRenameTab': {'type': 'runtime'},
  'HypershadeRenderPerspLayout': {'type': 'runtime'},
  'HypershadeRenderTextureRange': {'type': 'runtime'},
  'HypershadeRenderTextureRangeOptions': {'type': 'runtime'},
  'HypershadeRestoreLastClosedTab': {'type': 'runtime'},
  'HypershadeRevertSelectedSwatches': {'type': 'runtime'},
  'HypershadeRevertToDefaultTabs': {'type': 'runtime'},
  'HypershadeSaveSwatchesToDisk': {'type': 'runtime'},
  'HypershadeSelectBakeSets': {'type': 'runtime'},
  'HypershadeSelectCamerasAndImagePlanes': {'type': 'runtime'},
  'HypershadeSelectConnected': {'type': 'runtime'},
  'HypershadeSelectDownStream': {'type': 'runtime'},
  'HypershadeSelectLights': {'type': 'runtime'},
  'HypershadeSelectMaterialsFromObjects': {'type': 'runtime'},
  'HypershadeSelectObjectsWithMaterials': {'type': 'runtime'},
  'HypershadeSelectShadingGroupsAndMaterials': {'type': 'runtime'},
  'HypershadeSelectTextures': {'type': 'runtime'},
  'HypershadeSelectUpStream': {'type': 'runtime'},
  'HypershadeSelectUtilities': {'type': 'runtime'},
  'HypershadeSetLargeNodeSwatchSize': {'type': 'runtime'},
  'HypershadeSetSmallNodeSwatchSize': {'type': 'runtime'},
  'HypershadeSetTraversalDepthUnlim': {'type': 'runtime'},
  'HypershadeSetTraversalDepthZero': {'type': 'runtime'},
  'HypershadeShapeMenuStateAll': {'type': 'runtime'},
  'HypershadeShapeMenuStateAllExceptShadingGroupMembers': {'type': 'runtime'},
  'HypershadeShapeMenuStateNoShapes': {'type': 'runtime'},
  'HypershadeShowAllAttrs': {'type': 'runtime'},
  'HypershadeShowConnectedAttrs': {'type': 'runtime'},
  'HypershadeShowCustomAttrs': {'type': 'runtime'},
  'HypershadeShowDirectoriesAndFiles': {'type': 'runtime'},
  'HypershadeShowDirectoriesOnly': {'type': 'runtime'},
  'HypershadeSortByName': {'type': 'runtime'},
  'HypershadeSortByTime': {'type': 'runtime'},
  'HypershadeSortByType': {'type': 'runtime'},
  'HypershadeSortReverseOrder': {'type': 'runtime'},
  'HypershadeTestTexture': {'type': 'runtime'},
  'HypershadeTestTextureOptions': {'type': 'runtime'},
  'HypershadeToggleAttrFilter': {'type': 'runtime'},
  'HypershadeToggleCrosshairOnEdgeDragging': {'type': 'runtime'},
  'HypershadeToggleNodeSwatchSize': {'type': 'runtime'},
  'HypershadeToggleNodeTitleMode': {'type': 'runtime'},
  'HypershadeToggleShowNamespace': {'type': 'runtime'},
  'HypershadeToggleTransformDisplay': {'type': 'runtime'},
  'HypershadeToggleUseAssetsAndPublishedAttributes': {'type': 'runtime'},
  'HypershadeToggleZoomIn': {'type': 'runtime'},
  'HypershadeToggleZoomOut': {'type': 'runtime'},
  'HypershadeTransferAttributeValues': {'type': 'runtime'},
  'HypershadeTransferAttributeValuesOptions': {'type': 'runtime'},
  'HypershadeUnpinSelected': {'type': 'runtime'},
  'HypershadeUpdatePSDNetworks': {'type': 'runtime'},
  'HypershadeWindow': {'type': 'runtime'},
  'IKHandleTool': {'type': 'runtime'},
  'IKHandleToolOptions': {'type': 'runtime'},
  'IKSplineHandleTool': {'type': 'runtime'},
  'IKSplineHandleToolOptions': {'type': 'runtime'},
  'IPROptions': {'type': 'runtime'},
  'IPRRenderIntoNewWindow': {'type': 'runtime'},
  'IgnoreHardwareShader': {'type': 'runtime'},
  'IkHdsWin': {'type': 'runtime'},
  'IkfkjdsWin': {'type': 'runtime'},
  'Import': {'type': 'runtime'},
  'ImportAnim': {'type': 'runtime'},
  'ImportDeformerWeights': {'type': 'runtime'},
  'ImportDeformerWeightsOptions': {'type': 'runtime'},
  'ImportOptions': {'type': 'runtime'},
  'ImportSkinWeightMaps': {'type': 'runtime'},
  'ImportWorkspaceFiles': {'type': 'runtime'},
  'InTangentAuto': {'type': 'runtime'},
  'InTangentClamped': {'type': 'runtime'},
  'InTangentFixed': {'type': 'runtime'},
  'InTangentFlat': {'type': 'runtime'},
  'InTangentLinear': {'type': 'runtime'},
  'InTangentPlateau': {'type': 'runtime'},
  'InTangentSpline': {'type': 'runtime'},
  'IncreaseCheckerDensity': {'type': 'runtime'},
  'IncreaseExposureCoarse': {'type': 'runtime'},
  'IncreaseExposureFine': {'type': 'runtime'},
  'IncreaseGammaCoarse': {'type': 'runtime'},
  'IncreaseGammaFine': {'type': 'runtime'},
  'IncreaseManipulatorSize': {'type': 'runtime'},
  'IncrementAndSave': {'type': 'runtime'},
  'IncrementFluidCenter': {'type': 'runtime'},
  'InitialFluidStates': {'type': 'runtime'},
  'InitialFluidStatesOptions': {'type': 'runtime'},
  'InsertEdgeLoopTool': {'type': 'runtime'},
  'InsertEdgeLoopToolOptions': {'type': 'runtime'},
  'InsertIsoparms': {'type': 'runtime'},
  'InsertIsoparmsOptions': {'type': 'runtime'},
  'InsertJointTool': {'type': 'runtime'},
  'InsertKeyToolActivate': {'type': 'runtime'},
  'InsertKeyToolDeactivate': {'type': 'runtime'},
  'InsertKeysTool': {'type': 'runtime'},
  'InsertKeysToolOptions': {'type': 'runtime'},
  'InsertKnot': {'type': 'runtime'},
  'InsertKnotOptions': {'type': 'runtime'},
  'InteractiveBindSkin': {'type': 'runtime'},
  'InteractiveBindSkinOptions': {'type': 'runtime'},
  'InteractivePlayback': {'type': 'runtime'},
  'InteractiveSplitTool': {'type': 'runtime'},
  'InteractiveSplitToolOptions': {'type': 'runtime'},
  'IntersectCurve': {'type': 'runtime'},
  'IntersectCurveOptions': {'type': 'runtime'},
  'IntersectSurfaces': {'type': 'runtime'},
  'IntersectSurfacesOptions': {'type': 'runtime'},
  'InvertSelection': {'type': 'runtime'},
  'JdsWin': {'type': 'runtime'},
  'JointTool': {'type': 'runtime'},
  'JointToolOptions': {'type': 'runtime'},
  'KeyBlendShapeTargetsWeight': {'type': 'runtime'},
  'KeyframeTangentMarkingMenu': {'type': 'runtime'},
  'KeyframeTangentMarkingMenuPopDown': {'type': 'runtime'},
  'LODGenerateMeshes': {'type': 'runtime'},
  'LODGenerateMeshesOptions': {'type': 'runtime'},
  'LabelBasedOnJointNames': {'type': 'runtime'},
  'LassoTool': {'type': 'runtime'},
  'LastActionTool': {'type': 'runtime'},
  'LatticeDeformKeysTool': {'type': 'runtime'},
  'LatticeDeformKeysToolOptions': {'type': 'runtime'},
  'LatticeUVTool': {'type': 'runtime'},
  'LatticeUVToolOptions': {'type': 'runtime'},
  'LayerRelationshipEditor': {'type': 'runtime'},
  'LayoutUV': {'type': 'runtime'},
  'LayoutUVAlong': {'type': 'runtime'},
  'LayoutUVAlongOptions': {'type': 'runtime'},
  'LayoutUVOptions': {'type': 'runtime'},
  'LayoutUVRectangle': {'type': 'runtime'},
  'LevelOfDetailGroup': {'type': 'runtime'},
  'LevelOfDetailGroupOptions': {'type': 'runtime'},
  'LevelOfDetailUngroup': {'type': 'runtime'},
  'LightCentricLightLinkingEditor': {'type': 'runtime'},
  'Lightning': {'type': 'runtime'},
  'LightningOptions': {'type': 'runtime'},
  'LockCamera': {'type': 'runtime'},
  'LockContainer': {'type': 'runtime'},
  'LockCurveLength': {'type': 'runtime'},
  'LockNormals': {'type': 'runtime'},
  'LockTangentWeight': {'type': 'runtime'},
  'Loft': {'type': 'runtime'},
  'LoftOptions': {'type': 'runtime'},
  'LongPolygonNormals': {'type': 'runtime'},
  'LookAtSelection': {'type': 'runtime'},
  'LoopBrushAnimation': {'type': 'runtime'},
  'LoopBrushAnimationOptions': {'type': 'runtime'},
  'LowQualityDisplay': {'type': 'runtime'},
  'MakeBoats': {'type': 'runtime'},
  'MakeBoatsOptions': {'type': 'runtime'},
  'MakeBrushSpring': {'type': 'runtime'},
  'MakeBrushSpringOptions': {'type': 'runtime'},
  'MakeCollide': {'type': 'runtime'},
  'MakeCollideHair': {'type': 'runtime'},
  'MakeCollideOptions': {'type': 'runtime'},
  'MakeCurvesDynamic': {'type': 'runtime'},
  'MakeCurvesDynamicOptions': {'type': 'runtime'},
  'MakeFluidCollide': {'type': 'runtime'},
  'MakeFluidCollideOptions': {'type': 'runtime'},
  'MakeHoleTool': {'type': 'runtime'},
  'MakeHoleToolOptions': {'type': 'runtime'},
  'MakeLightLinks': {'type': 'runtime'},
  'MakeLive': {'type': 'runtime'},
  'MakeMotionField': {'type': 'runtime'},
  'MakeMotorBoats': {'type': 'runtime'},
  'MakeMotorBoatsOptions': {'type': 'runtime'},
  'MakePaintable': {'type': 'runtime'},
  'MakePondBoats': {'type': 'runtime'},
  'MakePondBoatsOptions': {'type': 'runtime'},
  'MakePondMotorBoats': {'type': 'runtime'},
  'MakePondMotorBoatsOptions': {'type': 'runtime'},
  'MakePressureCurve': {'type': 'runtime'},
  'MakePressureCurveOptions': {'type': 'runtime'},
  'MakeShadowLinks': {'type': 'runtime'},
  'MakeStereoLinks': {'type': 'runtime'},
  'MakeUVInstanceCurrent': {'type': 'runtime'},
  'MapUVBorder': {'type': 'runtime'},
  'MapUVBorderOptions': {'type': 'runtime'},
  'MarkingMenuPopDown': {'type': 'runtime'},
  'MarkingMenuPreferencesWindow': {'type': 'runtime'},
  'MatchPivots': {'type': 'runtime'},
  'MatchRotation': {'type': 'runtime'},
  'MatchScaling': {'type': 'runtime'},
  'MatchTransform': {'type': 'runtime'},
  'MatchTranslation': {'type': 'runtime'},
  'MatchUVs': {'type': 'runtime'},
  'MatchUVsOptions': {'type': 'runtime'},
  'MediumPolygonNormals': {'type': 'runtime'},
  'MediumQualityDisplay': {'type': 'runtime'},
  'MergeCharacterSet': {'type': 'runtime'},
  'MergeEdgeTool': {'type': 'runtime'},
  'MergeEdgeToolOptions': {'type': 'runtime'},
  'MergeMultipleEdges': {'type': 'runtime'},
  'MergeMultipleEdgesOptions': {'type': 'runtime'},
  'MergeToCenter': {'type': 'runtime'},
  'MergeUV': {'type': 'runtime'},
  'MergeUVOptions': {'type': 'runtime'},
  'MergeVertexTool': {'type': 'runtime'},
  'MergeVertexToolOptions': {'type': 'runtime'},
  'MergeVertices': {'type': 'runtime'},
  'MergeVerticesOptions': {'type': 'runtime'},
  'MinimizeApplication': {'type': 'runtime'},
  'MirrorCutPolygonGeometry': {'type': 'runtime'},
  'MirrorCutPolygonGeometryOptions': {'type': 'runtime'},
  'MirrorDeformerWeights': {'type': 'runtime'},
  'MirrorDeformerWeightsOptions': {'type': 'runtime'},
  'MirrorJoint': {'type': 'runtime'},
  'MirrorJointOptions': {'type': 'runtime'},
  'MirrorPolygonGeometry': {'type': 'runtime'},
  'MirrorPolygonGeometryOptions': {'type': 'runtime'},
  'MirrorSkinWeights': {'type': 'runtime'},
  'MirrorSkinWeightsOptions': {'type': 'runtime'},
  'MirrorSubdivSurface': {'type': 'runtime'},
  'MirrorSubdivSurfaceOptions': {'type': 'runtime'},
  'ModelingPanelRedoViewChange': {'type': 'runtime'},
  'ModelingPanelUndoViewChange': {'type': 'runtime'},
  'ModifyConstraintAxis': {'type': 'runtime'},
  'ModifyConstraintAxisOptions': {'type': 'runtime'},
  'ModifyCurrentSet': {'type': 'runtime'},
  'ModifyDisplacementPress': {'type': 'runtime'},
  'ModifyDisplacementRelease': {'type': 'runtime'},
  'ModifyLowerRadiusPress': {'type': 'runtime'},
  'ModifyLowerRadiusRelease': {'type': 'runtime'},
  'ModifyOpacityPress': {'type': 'runtime'},
  'ModifyOpacityRelease': {'type': 'runtime'},
  'ModifyPaintValuePress': {'type': 'runtime'},
  'ModifyPaintValueRelease': {'type': 'runtime'},
  'ModifyStampDepthPress': {'type': 'runtime'},
  'ModifyStampDepthRelease': {'type': 'runtime'},
  'ModifyUVVectorPress': {'type': 'runtime'},
  'ModifyUVVectorRelease': {'type': 'runtime'},
  'ModifyUpperRadiusPress': {'type': 'runtime'},
  'ModifyUpperRadiusRelease': {'type': 'runtime'},
  'MoveCacheToInput': {'type': 'runtime'},
  'MoveCurveSeam': {'type': 'runtime'},
  'MoveDown': {'type': 'runtime'},
  'MoveIKtoFK': {'type': 'runtime'},
  'MoveInfluence': {'type': 'runtime'},
  'MoveLeft': {'type': 'runtime'},
  'MoveNearestPickedKeyToolActivate': {'type': 'runtime'},
  'MoveNearestPickedKeyToolDeactivate': {'type': 'runtime'},
  'MoveNormalTool': {'type': 'runtime'},
  'MoveNormalToolOptions': {'type': 'runtime'},
  'MovePolygonComponent': {'type': 'runtime'},
  'MovePolygonComponentOptions': {'type': 'runtime'},
  'MoveRight': {'type': 'runtime'},
  'MoveRotateScaleTool': {'type': 'runtime'},
  'MoveRotateScaleToolToggleSnapMode': {'type': 'runtime'},
  'MoveRotateScaleToolToggleSnapRelativeMode': {'type': 'runtime'},
  'MoveSewUVs': {'type': 'runtime'},
  'MoveSkinJointsTool': {'type': 'runtime'},
  'MoveSkinJointsToolOptions': {'type': 'runtime'},
  'MoveSurfaceSeam': {'type': 'runtime'},
  'MoveTool': {'type': 'runtime'},
  'MoveToolOptions': {'type': 'runtime'},
  'MoveUVTool': {'type': 'runtime'},
  'MoveUp': {'type': 'runtime'},
  'MultiCutTool': {'type': 'runtime'},
  'NCreateEmitter': {'type': 'runtime'},
  'NCreateEmitterOptions': {'type': 'runtime'},
  'NEmitFromObject': {'type': 'runtime'},
  'NEmitFromObjectOptions': {'type': 'runtime'},
  'NParticleStyleBalls': {'type': 'runtime'},
  'NParticleStyleCloud': {'type': 'runtime'},
  'NParticleStylePoints': {'type': 'runtime'},
  'NParticleStyleThickCloud': {'type': 'runtime'},
  'NParticleStyleWater': {'type': 'runtime'},
  'NParticleToPolygons': {'type': 'runtime'},
  'NParticleTool': {'type': 'runtime'},
  'NParticleToolOptions': {'type': 'runtime'},
  'NURBSSmoothnessFine': {'type': 'runtime'},
  'NURBSSmoothnessFineOptions': {'type': 'runtime'},
  'NURBSSmoothnessHull': {'type': 'runtime'},
  'NURBSSmoothnessHullOptions': {'type': 'runtime'},
  'NURBSSmoothnessMedium': {'type': 'runtime'},
  'NURBSSmoothnessMediumOptions': {'type': 'runtime'},
  'NURBSSmoothnessRough': {'type': 'runtime'},
  'NURBSSmoothnessRoughOptions': {'type': 'runtime'},
  'NURBSTexturePlacementTool': {'type': 'runtime'},
  'NURBSTexturePlacementToolOptions': {'type': 'runtime'},
  'NURBSToPolygons': {'type': 'runtime'},
  'NURBSToPolygonsOptions': {'type': 'runtime'},
  'NamespaceEditor': {'type': 'runtime'},
  'NewScene': {'type': 'runtime'},
  'NewSceneOptions': {'type': 'runtime'},
  'Newton': {'type': 'runtime'},
  'NewtonOptions': {'type': 'runtime'},
  'NextFrame': {'type': 'runtime'},
  'NextGreasePencilFrame': {'type': 'runtime'},
  'NextKey': {'type': 'runtime'},
  'NextManipulatorHandle': {'type': 'runtime'},
  'NextSkinPaintMode': {'type': 'runtime'},
  'NextViewArrangement': {'type': 'runtime'},
  'NodeEditorAddIterationStatePorts': {'type': 'runtime'},
  'NodeEditorAddOnNodeCreate': {'type': 'runtime'},
  'NodeEditorAdditiveGraphingMode': {'type': 'runtime'},
  'NodeEditorAutoSizeNodes': {'type': 'runtime'},
  'NodeEditorBackToParent': {'type': 'runtime'},
  'NodeEditorCloseActiveTab': {'type': 'runtime'},
  'NodeEditorCloseAllTabs': {'type': 'runtime'},
  'NodeEditorConnectNodeOnCreation': {'type': 'runtime'},
  'NodeEditorConnectOnDrop': {'type': 'runtime'},
  'NodeEditorConnectSelectedNodes': {'type': 'runtime'},
  'NodeEditorConnectionStyleBezier': {'type': 'runtime'},
  'NodeEditorConnectionStyleCorner': {'type': 'runtime'},
  'NodeEditorConnectionStyleSShape': {'type': 'runtime'},
  'NodeEditorConnectionStyleStraight': {'type': 'runtime'},
  'NodeEditorCopy': {'type': 'runtime'},
  'NodeEditorCopyConnectionsOnPaste': {'type': 'runtime'},
  'NodeEditorCreateCompound': {'type': 'runtime'},
  'NodeEditorCreateDoWhileCompound': {'type': 'runtime'},
  'NodeEditorCreateForEachCompound': {'type': 'runtime'},
  'NodeEditorCreateIterateCompound': {'type': 'runtime'},
  'NodeEditorCreateNodePopup': {'type': 'runtime'},
  'NodeEditorCreateTab': {'type': 'runtime'},
  'NodeEditorCut': {'type': 'runtime'},
  'NodeEditorDeleteNodes': {'type': 'runtime'},
  'NodeEditorDiveIntoCompound': {'type': 'runtime'},
  'NodeEditorExplodeCompound': {'type': 'runtime'},
  'NodeEditorExtendToShapes': {'type': 'runtime'},
  'NodeEditorGraphAddSelected': {'type': 'runtime'},
  'NodeEditorGraphAllShapes': {'type': 'runtime'},
  'NodeEditorGraphAllShapesExceptShading': {'type': 'runtime'},
  'NodeEditorGraphClearGraph': {'type': 'runtime'},
  'NodeEditorGraphDownstream': {'type': 'runtime'},
  'NodeEditorGraphNoShapes': {'type': 'runtime'},
  'NodeEditorGraphRearrange': {'type': 'runtime'},
  'NodeEditorGraphRemoveDownstream': {'type': 'runtime'},
  'NodeEditorGraphRemoveSelected': {'type': 'runtime'},
  'NodeEditorGraphRemoveUnselected': {'type': 'runtime'},
  'NodeEditorGraphRemoveUpstream': {'type': 'runtime'},
  'NodeEditorGraphUpDownstream': {'type': 'runtime'},
  'NodeEditorGraphUpstream': {'type': 'runtime'},
  'NodeEditorGridToggleCrosshairOnEdgeDragging': {'type': 'runtime'},
  'NodeEditorGridToggleSnap': {'type': 'runtime'},
  'NodeEditorGridToggleVisibility': {'type': 'runtime'},
  'NodeEditorHideAttributes': {'type': 'runtime'},
  'NodeEditorHighlightConnectionsOnNodeSelection': {'type': 'runtime'},
  'NodeEditorIncreaseTraversalDepth': {'type': 'runtime'},
  'NodeEditorLayout': {'type': 'runtime'},
  'NodeEditorPaste': {'type': 'runtime'},
  'NodeEditorPickWalkDown': {'type': 'runtime'},
  'NodeEditorPickWalkLeft': {'type': 'runtime'},
  'NodeEditorPickWalkRight': {'type': 'runtime'},
  'NodeEditorPickWalkUp': {'type': 'runtime'},
  'NodeEditorPinByDefault': {'type': 'runtime'},
  'NodeEditorPinSelected': {'type': 'runtime'},
  'NodeEditorPublishCompound': {'type': 'runtime'},
  'NodeEditorRedockTornOffTab': {'type': 'runtime'},
  'NodeEditorReduceTraversalDepth': {'type': 'runtime'},
  'NodeEditorRenameActiveTab': {'type': 'runtime'},
  'NodeEditorRenderSwatches': {'type': 'runtime'},
  'NodeEditorRestoreLastClosedTab': {'type': 'runtime'},
  'NodeEditorSelectConnected': {'type': 'runtime'},
  'NodeEditorSelectDownStream': {'type': 'runtime'},
  'NodeEditorSelectUpStream': {'type': 'runtime'},
  'NodeEditorSetLargeNodeSwatchSize': {'type': 'runtime'},
  'NodeEditorSetSmallNodeSwatchSize': {'type': 'runtime'},
  'NodeEditorSetTraversalDepthUnlim': {'type': 'runtime'},
  'NodeEditorSetTraversalDepthZero': {'type': 'runtime'},
  'NodeEditorShapeMenuStateAll': {'type': 'runtime'},
  'NodeEditorShapeMenuStateAllExceptShadingGroupMembers': {'type': 'runtime'},
  'NodeEditorShapeMenuStateNoShapes': {'type': 'runtime'},
  'NodeEditorShowAllAttrs': {'type': 'runtime'},
  'NodeEditorShowAllCustomAttrs': {'type': 'runtime'},
  'NodeEditorShowConnectedAttrs': {'type': 'runtime'},
  'NodeEditorShowCustomAttrs': {'type': 'runtime'},
  'NodeEditorToggleAttrFilter': {'type': 'runtime'},
  'NodeEditorToggleConsistentNodeNameSize': {'type': 'runtime'},
  'NodeEditorToggleCreateNodePane': {'type': 'runtime'},
  'NodeEditorToggleLockUnlock': {'type': 'runtime'},
  'NodeEditorToggleNodeSelectedPins': {'type': 'runtime'},
  'NodeEditorToggleNodeSwatchSize': {'type': 'runtime'},
  'NodeEditorToggleNodeTitleMode': {'type': 'runtime'},
  'NodeEditorToggleShowNamespace': {'type': 'runtime'},
  'NodeEditorToggleSyncedSelection': {'type': 'runtime'},
  'NodeEditorToggleUseAssetsAndPublishedAttributes': {'type': 'runtime'},
  'NodeEditorToggleZoomIn': {'type': 'runtime'},
  'NodeEditorToggleZoomOut': {'type': 'runtime'},
  'NodeEditorTransforms': {'type': 'runtime'},
  'NodeEditorUnpinSelected': {'type': 'runtime'},
  'NodeEditorWindow': {'type': 'runtime'},
  'NonSacredTool': {'type': 'runtime'},
  'NonWeightedTangents': {'type': 'runtime'},
  'NormalConstraint': {'type': 'runtime'},
  'NormalConstraintOptions': {'type': 'runtime'},
  'NormalizeUVs': {'type': 'runtime'},
  'NormalizeUVsOptions': {'type': 'runtime'},
  'NormalizeWeights': {'type': 'runtime'},
  'NurbsCurveToBezier': {'type': 'runtime'},
  'ObjectCentricLightLinkingEditor': {'type': 'runtime'},
  'OffsetCurve': {'type': 'runtime'},
  'OffsetCurveOnSurface': {'type': 'runtime'},
  'OffsetCurveOnSurfaceOptions': {'type': 'runtime'},
  'OffsetCurveOptions': {'type': 'runtime'},
  'OffsetEdgeLoopTool': {'type': 'runtime'},
  'OffsetEdgeLoopToolOptions': {'type': 'runtime'},
  'OffsetSurfaces': {'type': 'runtime'},
  'OffsetSurfacesOptions': {'type': 'runtime'},
  'OpenAELiveLink': {'type': 'runtime'},
  'OpenAREACommunity': {'type': 'runtime'},
  'OpenAREAForums': {'type': 'runtime'},
  'OpenAutodeskAccount': {'type': 'runtime'},
  'OpenAutodeskExchange': {'type': 'runtime'},
  'OpenAutodeskStore': {'type': 'runtime'},
  'OpenBifContentBrowser': {'type': 'runtime'},
  'OpenBrowserSetupAssistant': {'type': 'runtime'},
  'OpenBugReport': {'type': 'runtime'},
  'OpenChannelBox': {'type': 'runtime'},
  'OpenChannelsLayers': {'type': 'runtime'},
  'OpenCharacterGenerator': {'type': 'runtime'},
  'OpenCloseCurve': {'type': 'runtime'},
  'OpenCloseCurveOptions': {'type': 'runtime'},
  'OpenCloseSurfaces': {'type': 'runtime'},
  'OpenCloseSurfacesOptions': {'type': 'runtime'},
  'OpenColorSetEditor': {'type': 'runtime'},
  'OpenContentBrowser': {'type': 'runtime'},
  'OpenCreaseEditor': {'type': 'runtime'},
  'OpenDGProfiler': {'type': 'runtime'},
  'OpenDesktopAnalytics': {'type': 'runtime'},
  'OpenFBXReview': {'type': 'runtime'},
  'OpenFacebook': {'type': 'runtime'},
  'OpenFeatureRequest': {'type': 'runtime'},
  'OpenHomePage': {'type': 'runtime'},
  'OpenLayerEditor': {'type': 'runtime'},
  'OpenLearningChannel': {'type': 'runtime'},
  'OpenLearningPath': {'type': 'runtime'},
  'OpenLightEditor': {'type': 'runtime'},
  'OpenMASHContentBrowser': {'type': 'runtime'},
  'OpenMelCmdRef': {'type': 'runtime'},
  'OpenMenuFinder': {'type': 'runtime'},
  'OpenModelingToolkit': {'type': 'runtime'},
  'OpenNodeAttrRef': {'type': 'runtime'},
  'OpenProductResearch': {'type': 'runtime'},
  'OpenPyCmdRef': {'type': 'runtime'},
  'OpenReleaseNotes': {'type': 'runtime'},
  'OpenScene': {'type': 'runtime'},
  'OpenSceneOptions': {'type': 'runtime'},
  'OpenStartupMovies': {'type': 'runtime'},
  'OpenStereoRigManager': {'type': 'runtime'},
  'OpenSupportCenter': {'type': 'runtime'},
  'OpenTimeSliderPrefs': {'type': 'runtime'},
  'OpenTinkercad': {'type': 'runtime'},
  'OpenTrialTutorials': {'type': 'runtime'},
  'OpenTutorials': {'type': 'runtime'},
  'OpenViewportPrefs': {'type': 'runtime'},
  'OpenVisorForMeshes': {'type': 'runtime'},
  'OpenXGenEditor': {'type': 'runtime'},
  'OptimizeScene': {'type': 'runtime'},
  'OptimizeSceneOptions': {'type': 'runtime'},
  'OptimzeUVs': {'type': 'runtime'},
  'OptimzeUVsOptions': {'type': 'runtime'},
  'OrientConstraint': {'type': 'runtime'},
  'OrientConstraintOptions': {'type': 'runtime'},
  'OrientJoint': {'type': 'runtime'},
  'OrientJointOptions': {'type': 'runtime'},
  'OutTangentAuto': {'type': 'runtime'},
  'OutTangentClamped': {'type': 'runtime'},
  'OutTangentFixed': {'type': 'runtime'},
  'OutTangentFlat': {'type': 'runtime'},
  'OutTangentLinear': {'type': 'runtime'},
  'OutTangentPlateau': {'type': 'runtime'},
  'OutTangentSpline': {'type': 'runtime'},
  'OutlinerCollapseAllItems': {'type': 'runtime'},
  'OutlinerCollapseAllSelectedItems': {'type': 'runtime'},
  'OutlinerDoHide': {'type': 'runtime'},
  'OutlinerExpandAllItems': {'type': 'runtime'},
  'OutlinerExpandAllSelectedItems': {'type': 'runtime'},
  'OutlinerRenameSelectedItem': {'type': 'runtime'},
  'OutlinerRevealSelected': {'type': 'runtime'},
  'OutlinerToggleAssignedMaterials': {'type': 'runtime'},
  'OutlinerToggleAttributes': {'type': 'runtime'},
  'OutlinerToggleAutoExpandLayers': {'type': 'runtime'},
  'OutlinerToggleConnected': {'type': 'runtime'},
  'OutlinerToggleDAGOnly': {'type': 'runtime'},
  'OutlinerToggleIgnoreHidden': {'type': 'runtime'},
  'OutlinerToggleIgnoreUseColor': {'type': 'runtime'},
  'OutlinerToggleNamespace': {'type': 'runtime'},
  'OutlinerToggleOrganizeByClip': {'type': 'runtime'},
  'OutlinerToggleOrganizeByLayer': {'type': 'runtime'},
  'OutlinerToggleReferenceMembers': {'type': 'runtime'},
  'OutlinerToggleReferenceNodes': {'type': 'runtime'},
  'OutlinerToggleSetMembers': {'type': 'runtime'},
  'OutlinerToggleShapes': {'type': 'runtime'},
  'OutlinerToggleShowMuteInformation': {'type': 'runtime'},
  'OutlinerToggleTimeEditor': {'type': 'runtime'},
  'OutlinerUnhide': {'type': 'runtime'},
  'OutlinerWindow': {'type': 'runtime'},
  'PFXUVSetLinkingEditor': {'type': 'runtime'},
  'PaintCacheTool': {'type': 'runtime'},
  'PaintCacheToolOptions': {'type': 'runtime'},
  'PaintClusterWeightsTool': {'type': 'runtime'},
  'PaintClusterWeightsToolOptions': {'type': 'runtime'},
  'PaintDeltaMushWeightsTool': {'type': 'runtime'},
  'PaintDeltaMushWeightsToolOptions': {'type': 'runtime'},
  'PaintEffectPanelActivate': {'type': 'runtime'},
  'PaintEffectPanelDeactivate': {'type': 'runtime'},
  'PaintEffectsGlobalSettings': {'type': 'runtime'},
  'PaintEffectsMeshQuality': {'type': 'runtime'},
  'PaintEffectsPanel': {'type': 'runtime'},
  'PaintEffectsToCurve': {'type': 'runtime'},
  'PaintEffectsToCurveOptions': {'type': 'runtime'},
  'PaintEffectsToNurbs': {'type': 'runtime'},
  'PaintEffectsToNurbsOptions': {'type': 'runtime'},
  'PaintEffectsToPoly': {'type': 'runtime'},
  'PaintEffectsToPolyOptions': {'type': 'runtime'},
  'PaintEffectsTool': {'type': 'runtime'},
  'PaintEffectsToolOptions': {'type': 'runtime'},
  'PaintEffectsWindow': {'type': 'runtime'},
  'PaintFluidsTool': {'type': 'runtime'},
  'PaintFluidsToolOptions': {'type': 'runtime'},
  'PaintGeomCacheTool': {'type': 'runtime'},
  'PaintGeomCacheToolOptions': {'type': 'runtime'},
  'PaintGrid': {'type': 'runtime'},
  'PaintGridOptions': {'type': 'runtime'},
  'PaintHairBaldness': {'type': 'runtime'},
  'PaintHairColor': {'type': 'runtime'},
  'PaintHairFollicles': {'type': 'runtime'},
  'PaintHairFolliclesOptions': {'type': 'runtime'},
  'PaintHairSpecularColor': {'type': 'runtime'},
  'PaintJiggleWeightsTool': {'type': 'runtime'},
  'PaintJiggleWeightsToolOptions': {'type': 'runtime'},
  'PaintLatticeWeightsTool': {'type': 'runtime'},
  'PaintLatticeWeightsToolOptions': {'type': 'runtime'},
  'PaintNonlinearWeightsTool': {'type': 'runtime'},
  'PaintNonlinearWeightsToolOptions': {'type': 'runtime'},
  'PaintOnPaintableObjects': {'type': 'runtime'},
  'PaintOnViewPlane': {'type': 'runtime'},
  'PaintOperationMarkingMenuPress': {'type': 'runtime'},
  'PaintOperationMarkingMenuRelease': {'type': 'runtime'},
  'PaintRandom': {'type': 'runtime'},
  'PaintRandomOptions': {'type': 'runtime'},
  'PaintReduceWeightsTool': {'type': 'runtime'},
  'PaintReduceWeightsToolOptions': {'type': 'runtime'},
  'PaintSetMembershipTool': {'type': 'runtime'},
  'PaintSetMembershipToolOptions': {'type': 'runtime'},
  'PaintShrinkWrapWeightsTool': {'type': 'runtime'},
  'PaintShrinkWrapWeightsToolOptions': {'type': 'runtime'},
  'PaintSoftWeights': {'type': 'runtime'},
  'PaintSoftWeightsOptions': {'type': 'runtime'},
  'PaintTensionWeightsTool': {'type': 'runtime'},
  'PaintTensionWeightsToolOptions': {'type': 'runtime'},
  'PaintTextureDeformerWeightsTool': {'type': 'runtime'},
  'PaintTextureDeformerWeightsToolOptions': {'type': 'runtime'},
  'PaintToonBorderColor': {'type': 'runtime'},
  'PaintToonCreaseColor': {'type': 'runtime'},
  'PaintToonLineOffset': {'type': 'runtime'},
  'PaintToonLineOpacity': {'type': 'runtime'},
  'PaintToonLineWidth': {'type': 'runtime'},
  'PaintToonProfileColor': {'type': 'runtime'},
  'PaintTransferAttributes': {'type': 'runtime'},
  'PaintVertexColorTool': {'type': 'runtime'},
  'PaintVertexColorToolOptions': {'type': 'runtime'},
  'PaintWireWeightsTool': {'type': 'runtime'},
  'PaintWireWeightsToolOptions': {'type': 'runtime'},
  'PanZoomTool': {'type': 'runtime'},
  'PanePop': {'type': 'runtime'},
  'PanelPreferencesWindow': {'type': 'runtime'},
  'ParameterTool': {'type': 'runtime'},
  'Parent': {'type': 'runtime'},
  'ParentBaseWire': {'type': 'runtime'},
  'ParentBaseWireOptions': {'type': 'runtime'},
  'ParentConstraint': {'type': 'runtime'},
  'ParentConstraintOptions': {'type': 'runtime'},
  'ParentOptions': {'type': 'runtime'},
  'PartSpriteWizard': {'type': 'runtime'},
  'PartialCreaseSubdivSurface': {'type': 'runtime'},
  'ParticleCollisionEvents': {'type': 'runtime'},
  'ParticleFill': {'type': 'runtime'},
  'ParticleFillOptions': {'type': 'runtime'},
  'ParticleInstancer': {'type': 'runtime'},
  'ParticleInstancerOptions': {'type': 'runtime'},
  'ParticleTool': {'type': 'runtime'},
  'ParticleToolOptions': {'type': 'runtime'},
  'PartitionEditor': {'type': 'runtime'},
  'PasteKeys': {'type': 'runtime'},
  'PasteKeysOptions': {'type': 'runtime'},
  'PasteSelected': {'type': 'runtime'},
  'PasteUVs': {'type': 'runtime'},
  'PasteVertexSkinWeights': {'type': 'runtime'},
  'PasteVertexWeights': {'type': 'runtime'},
  'PauseViewportEval': {'type': 'runtime'},
  'PencilCurveTool': {'type': 'runtime'},
  'PencilCurveToolOptions': {'type': 'runtime'},
  'PerPointEmissionRates': {'type': 'runtime'},
  'PerformExportToBackburner': {'type': 'runtime'},
  'PerformPrecompExport': {'type': 'runtime'},
  'PerformPrecompExportOptions': {'type': 'runtime'},
  'PerformTessellationSetup': {'type': 'runtime'},
  'PerformTessellationSetupOptions': {'type': 'runtime'},
  'PerformanceSettingsWindow': {'type': 'runtime'},
  'PerspGraphHypergraphLayout': {'type': 'runtime'},
  'PerspGraphLayout': {'type': 'runtime'},
  'PerspGraphOutlinerLayout': {'type': 'runtime'},
  'PerspRelationshipEditorLayout': {'type': 'runtime'},
  'PerspTextureLayout': {'type': 'runtime'},
  'PfxBrushTransfer': {'type': 'runtime'},
  'PfxFlipTubeDir': {'type': 'runtime'},
  'PfxGetBrush': {'type': 'runtime'},
  'PfxMakeCollide': {'type': 'runtime'},
  'PfxPresetBlend': {'type': 'runtime'},
  'PfxSetLineModifierObject': {'type': 'runtime'},
  'PickColorActivate': {'type': 'runtime'},
  'PickColorDeactivate': {'type': 'runtime'},
  'PickWalkDown': {'type': 'runtime'},
  'PickWalkDownSelect': {'type': 'runtime'},
  'PickWalkIn': {'type': 'runtime'},
  'PickWalkLeft': {'type': 'runtime'},
  'PickWalkLeftSelect': {'type': 'runtime'},
  'PickWalkOut': {'type': 'runtime'},
  'PickWalkRight': {'type': 'runtime'},
  'PickWalkRightSelect': {'type': 'runtime'},
  'PickWalkStopAtTransform': {'type': 'runtime'},
  'PickWalkUp': {'type': 'runtime'},
  'PickWalkUpSelect': {'type': 'runtime'},
  'PickWalkUseController': {'type': 'runtime'},
  'PinSelection': {'type': 'runtime'},
  'PinSelectionOptions': {'type': 'runtime'},
  'PixelMoveDown': {'type': 'runtime'},
  'PixelMoveLeft': {'type': 'runtime'},
  'PixelMoveRight': {'type': 'runtime'},
  'PixelMoveUp': {'type': 'runtime'},
  'Planar': {'type': 'runtime'},
  'PlanarOptions': {'type': 'runtime'},
  'PlaybackBackward': {'type': 'runtime'},
  'PlaybackForward': {'type': 'runtime'},
  'PlaybackFree': {'type': 'runtime'},
  'PlaybackLoopContinuous': {'type': 'runtime'},
  'PlaybackLoopOnce': {'type': 'runtime'},
  'PlaybackLoopOscillate': {'type': 'runtime'},
  'PlaybackPefRealtime': {'type': 'runtime'},
  'PlaybackRangeEnabledClips': {'type': 'runtime'},
  'PlaybackRangeHighlight': {'type': 'runtime'},
  'PlaybackRangeMinMax': {'type': 'runtime'},
  'PlaybackRangePrefs': {'type': 'runtime'},
  'PlaybackRangeSound': {'type': 'runtime'},
  'PlaybackRangeStartEnd': {'type': 'runtime'},
  'PlaybackRealtime': {'type': 'runtime'},
  'PlaybackSteppedPreview': {'type': 'runtime'},
  'PlaybackStop': {'type': 'runtime'},
  'PlaybackToggle': {'type': 'runtime'},
  'PlayblastOptions': {'type': 'runtime'},
  'PlayblastWindow': {'type': 'runtime'},
  'PlayblastWindowOptions': {'type': 'runtime'},
  'PluginManager': {'type': 'runtime'},
  'PointConstraint': {'type': 'runtime'},
  'PointConstraintOptions': {'type': 'runtime'},
  'PointOnCurve': {'type': 'runtime'},
  'PointOnCurveOptions': {'type': 'runtime'},
  'PointOnPolyConstraint': {'type': 'runtime'},
  'PointOnPolyConstraintOptions': {'type': 'runtime'},
  'PokePolygon': {'type': 'runtime'},
  'PokePolygonOptions': {'type': 'runtime'},
  'PoleVectorConstraint': {'type': 'runtime'},
  'PoleVectorConstraintOptions': {'type': 'runtime'},
  'PolyAssignSubdivHole': {'type': 'runtime'},
  'PolyAssignSubdivHoleOptions': {'type': 'runtime'},
  'PolyBrushMarkingMenu': {'type': 'runtime'},
  'PolyBrushMarkingMenuPopDown': {'type': 'runtime'},
  'PolyCircularize': {'type': 'runtime'},
  'PolyCircularizeOptions': {'type': 'runtime'},
  'PolyConvertToLoopAndDelete': {'type': 'runtime'},
  'PolyConvertToLoopAndDuplicate': {'type': 'runtime'},
  'PolyConvertToRingAndCollapse': {'type': 'runtime'},
  'PolyConvertToRingAndSplit': {'type': 'runtime'},
  'PolyCreaseTool': {'type': 'runtime'},
  'PolyCreaseToolOptions': {'type': 'runtime'},
  'PolyDisplayReset': {'type': 'runtime'},
  'PolyEditEdgeFlow': {'type': 'runtime'},
  'PolyEditEdgeFlowOptions': {'type': 'runtime'},
  'PolyExtrude': {'type': 'runtime'},
  'PolyExtrudeEdges': {'type': 'runtime'},
  'PolyExtrudeEdgesOptions': {'type': 'runtime'},
  'PolyExtrudeFaces': {'type': 'runtime'},
  'PolyExtrudeFacesOptions': {'type': 'runtime'},
  'PolyExtrudeOptions': {'type': 'runtime'},
  'PolyExtrudeVertices': {'type': 'runtime'},
  'PolyExtrudeVerticesOptions': {'type': 'runtime'},
  'PolyMerge': {'type': 'runtime'},
  'PolyMergeEdges': {'type': 'runtime'},
  'PolyMergeEdgesOptions': {'type': 'runtime'},
  'PolyMergeOptions': {'type': 'runtime'},
  'PolyMergeVertices': {'type': 'runtime'},
  'PolyMergeVerticesOptions': {'type': 'runtime'},
  'PolyRemoveAllCrease': {'type': 'runtime'},
  'PolyRemoveCrease': {'type': 'runtime'},
  'PolySelectTool': {'type': 'runtime'},
  'PolySelectToolOptions': {'type': 'runtime'},
  'PolySpinEdgeBackward': {'type': 'runtime'},
  'PolySpinEdgeForward': {'type': 'runtime'},
  'PolygonApplyColor': {'type': 'runtime'},
  'PolygonApplyColorOptions': {'type': 'runtime'},
  'PolygonBooleanDifference': {'type': 'runtime'},
  'PolygonBooleanDifferenceOptions': {'type': 'runtime'},
  'PolygonBooleanIntersection': {'type': 'runtime'},
  'PolygonBooleanIntersectionOptions': {'type': 'runtime'},
  'PolygonBooleanUnion': {'type': 'runtime'},
  'PolygonBooleanUnionOptions': {'type': 'runtime'},
  'PolygonClearClipboard': {'type': 'runtime'},
  'PolygonClearClipboardOptions': {'type': 'runtime'},
  'PolygonCollapse': {'type': 'runtime'},
  'PolygonCollapseEdges': {'type': 'runtime'},
  'PolygonCollapseFaces': {'type': 'runtime'},
  'PolygonCopy': {'type': 'runtime'},
  'PolygonCopyOptions': {'type': 'runtime'},
  'PolygonHardenEdge': {'type': 'runtime'},
  'PolygonNormalEditTool': {'type': 'runtime'},
  'PolygonPaste': {'type': 'runtime'},
  'PolygonPasteOptions': {'type': 'runtime'},
  'PolygonSelectionConstraints': {'type': 'runtime'},
  'PolygonSoftenEdge': {'type': 'runtime'},
  'PolygonSoftenHarden': {'type': 'runtime'},
  'PolygonSoftenHardenOptions': {'type': 'runtime'},
  'PoseEditor': {'type': 'runtime'},
  'PoseInterpolatorNewGroup': {'type': 'runtime'},
  'PositionAlongCurve': {'type': 'runtime'},
  'PostInfinityConstant': {'type': 'runtime'},
  'PostInfinityCycle': {'type': 'runtime'},
  'PostInfinityCycleOffset': {'type': 'runtime'},
  'PostInfinityLinear': {'type': 'runtime'},
  'PostInfinityOscillate': {'type': 'runtime'},
  'PreInfinityConstant': {'type': 'runtime'},
  'PreInfinityCycle': {'type': 'runtime'},
  'PreInfinityCycleOffset': {'type': 'runtime'},
  'PreInfinityLinear': {'type': 'runtime'},
  'PreInfinityOscillate': {'type': 'runtime'},
  'PreferencesWindow': {'type': 'runtime'},
  'PrefixHierarchyNames': {'type': 'runtime'},
  'PreflightPolygon': {'type': 'runtime'},
  'PreflightPolygonOptions': {'type': 'runtime'},
  'PrelightPolygon': {'type': 'runtime'},
  'PrelightPolygonOptions': {'type': 'runtime'},
  'PreloadReferenceEditor': {'type': 'runtime'},
  'PresetBlendingWindow': {'type': 'runtime'},
  'PrevSkinPaintMode': {'type': 'runtime'},
  'PreviousFrame': {'type': 'runtime'},
  'PreviousGreasePencilFrame': {'type': 'runtime'},
  'PreviousKey': {'type': 'runtime'},
  'PreviousManipulatorHandle': {'type': 'runtime'},
  'PreviousViewArrangement': {'type': 'runtime'},
  'ProductInformation': {'type': 'runtime'},
  'ProfilerTool': {'type': 'runtime'},
  'ProfilerToolCategoryView': {'type': 'runtime'},
  'ProfilerToolCpuView': {'type': 'runtime'},
  'ProfilerToolHideSelected': {'type': 'runtime'},
  'ProfilerToolHideSelectedRepetition': {'type': 'runtime'},
  'ProfilerToolReset': {'type': 'runtime'},
  'ProfilerToolShowAll': {'type': 'runtime'},
  'ProfilerToolShowSelected': {'type': 'runtime'},
  'ProfilerToolShowSelectedRepetition': {'type': 'runtime'},
  'ProfilerToolThreadView': {'type': 'runtime'},
  'ProfilerToolToggleRecording': {'type': 'runtime'},
  'ProjectCurveOnMesh': {'type': 'runtime'},
  'ProjectCurveOnMeshOptions': {'type': 'runtime'},
  'ProjectCurveOnSurface': {'type': 'runtime'},
  'ProjectCurveOnSurfaceOptions': {'type': 'runtime'},
  'ProjectTangent': {'type': 'runtime'},
  'ProjectTangentOptions': {'type': 'runtime'},
  'ProjectWindow': {'type': 'runtime'},
  'ProportionalModificationTool': {'type': 'runtime'},
  'PruneCluster': {'type': 'runtime'},
  'PruneLattice': {'type': 'runtime'},
  'PruneSculpt': {'type': 'runtime'},
  'PruneSmallWeights': {'type': 'runtime'},
  'PruneSmallWeightsOptions': {'type': 'runtime'},
  'PruneWire': {'type': 'runtime'},
  'PublishAttributes': {'type': 'runtime'},
  'PublishAttributesOptions': {'type': 'runtime'},
  'PublishChildAnchor': {'type': 'runtime'},
  'PublishChildAnchorOptions': {'type': 'runtime'},
  'PublishConnections': {'type': 'runtime'},
  'PublishConnectionsOptions': {'type': 'runtime'},
  'PublishNode': {'type': 'runtime'},
  'PublishParentAnchor': {'type': 'runtime'},
  'PublishParentAnchorOptions': {'type': 'runtime'},
  'PublishRootTransform': {'type': 'runtime'},
  'PublishRootTransformOptions': {'type': 'runtime'},
  'QuadDrawTool': {'type': 'runtime'},
  'Quadrangulate': {'type': 'runtime'},
  'QuadrangulateOptions': {'type': 'runtime'},
  'QualityDisplayMarkingMenu': {'type': 'runtime'},
  'QualityDisplayMarkingMenuPopDown': {'type': 'runtime'},
  'QuickRigEditor': {'type': 'runtime'},
  'Quit': {'type': 'runtime'},
  'Radial': {'type': 'runtime'},
  'RadialOptions': {'type': 'runtime'},
  'RaiseApplicationWindows': {'type': 'runtime'},
  'RaiseMainWindow': {'type': 'runtime'},
  'RandomizeFollicles': {'type': 'runtime'},
  'RandomizeFolliclesOptions': {'type': 'runtime'},
  'RandomizeShells': {'type': 'runtime'},
  'RandomizeShellsOptions': {'type': 'runtime'},
  'ReassignBoneLatticeJoint': {'type': 'runtime'},
  'ReattachSkeleton': {'type': 'runtime'},
  'ReattachSkeletonJoints': {'type': 'runtime'},
  'RebuildCurve': {'type': 'runtime'},
  'RebuildCurveOptions': {'type': 'runtime'},
  'RebuildSurfaces': {'type': 'runtime'},
  'RebuildSurfacesOptions': {'type': 'runtime'},
  'RecentCommandsWindow': {'type': 'runtime'},
  'Redo': {'type': 'runtime'},
  'RedoPreviousIPRRender': {'type': 'runtime'},
  'RedoPreviousRender': {'type': 'runtime'},
  'RedoViewChange': {'type': 'runtime'},
  'ReducePolygon': {'type': 'runtime'},
  'ReducePolygonOptions': {'type': 'runtime'},
  'ReferenceEditor': {'type': 'runtime'},
  'RefineSelectedComponents': {'type': 'runtime'},
  'RegionKeysTool': {'type': 'runtime'},
  'RelaxInitialState': {'type': 'runtime'},
  'RelaxInitialStateOptions': {'type': 'runtime'},
  'RelaxUVShell': {'type': 'runtime'},
  'RelaxUVShellOptions': {'type': 'runtime'},
  'RemoveBindingSet': {'type': 'runtime'},
  'RemoveBlendShape': {'type': 'runtime'},
  'RemoveBlendShapeOptions': {'type': 'runtime'},
  'RemoveBrushSharing': {'type': 'runtime'},
  'RemoveConstraintTarget': {'type': 'runtime'},
  'RemoveConstraintTargetOptions': {'type': 'runtime'},
  'RemoveFromCharacterSet': {'type': 'runtime'},
  'RemoveFromContainer': {'type': 'runtime'},
  'RemoveFromContainerOptions': {'type': 'runtime'},
  'RemoveInbetween': {'type': 'runtime'},
  'RemoveInfluence': {'type': 'runtime'},
  'RemoveJoint': {'type': 'runtime'},
  'RemoveLatticeTweaks': {'type': 'runtime'},
  'RemoveMaterialSoloing': {'type': 'runtime'},
  'RemoveNewPfxToon': {'type': 'runtime'},
  'RemoveShrinkWrapInnerObject': {'type': 'runtime'},
  'RemoveShrinkWrapSurfaces': {'type': 'runtime'},
  'RemoveShrinkWrapTarget': {'type': 'runtime'},
  'RemoveSubdivProxyMirror': {'type': 'runtime'},
  'RemoveSubdivProxyMirrorOptions': {'type': 'runtime'},
  'RemoveUnusedInfluences': {'type': 'runtime'},
  'RemoveWire': {'type': 'runtime'},
  'RemoveWireOptions': {'type': 'runtime'},
  'RemoveWrapInfluence': {'type': 'runtime'},
  'RenameAttribute': {'type': 'runtime'},
  'RenameCurrentColorSet': {'type': 'runtime'},
  'RenameCurrentSet': {'type': 'runtime'},
  'RenameCurrentUVSet': {'type': 'runtime'},
  'RenameJointsFromLabels': {'type': 'runtime'},
  'RenderDiagnostics': {'type': 'runtime'},
  'RenderFlagsWindow': {'type': 'runtime'},
  'RenderGlobalsWindow': {'type': 'runtime'},
  'RenderIntoNewWindow': {'type': 'runtime'},
  'RenderLayerEditorWindow': {'type': 'runtime'},
  'RenderLayerRelationshipEditor': {'type': 'runtime'},
  'RenderOptions': {'type': 'runtime'},
  'RenderPassSetEditor': {'type': 'runtime'},
  'RenderSequence': {'type': 'runtime'},
  'RenderSequenceOptions': {'type': 'runtime'},
  'RenderSetupWindow': {'type': 'runtime'},
  'RenderTextureRange': {'type': 'runtime'},
  'RenderTextureRangeOptions': {'type': 'runtime'},
  'RenderViewNextImage': {'type': 'runtime'},
  'RenderViewPrevImage': {'type': 'runtime'},
  'RenderViewWindow': {'type': 'runtime'},
  'ReorderVertex': {'type': 'runtime'},
  'RepeatLast': {'type': 'runtime'},
  'RepeatLastActionAtMousePosition': {'type': 'runtime'},
  'ReplaceObjects': {'type': 'runtime'},
  'ReplaceObjectsOptions': {'type': 'runtime'},
  'RerootSkeleton': {'type': 'runtime'},
  'ResampleCurve': {'type': 'runtime'},
  'ResampleCurveOptions': {'type': 'runtime'},
  'ResetCurrentWorkspace': {'type': 'runtime'},
  'ResetDisplay': {'type': 'runtime'},
  'ResetLattice': {'type': 'runtime'},
  'ResetReflectionOptions': {'type': 'runtime'},
  'ResetSoftSelectOptions': {'type': 'runtime'},
  'ResetTemplateBrush': {'type': 'runtime'},
  'ResetTransformations': {'type': 'runtime'},
  'ResetTransformationsOptions': {'type': 'runtime'},
  'ResetViewport': {'type': 'runtime'},
  'ResetWeightsToDefault': {'type': 'runtime'},
  'ResetWire': {'type': 'runtime'},
  'ResetWireOptions': {'type': 'runtime'},
  'ResolveInterpenetration': {'type': 'runtime'},
  'ResolveInterpenetrationOptions': {'type': 'runtime'},
  'RestoreUIElements': {'type': 'runtime'},
  'RetimeKeysTool': {'type': 'runtime'},
  'RetimeKeysToolOptions': {'type': 'runtime'},
  'ReverseCurve': {'type': 'runtime'},
  'ReverseCurveOptions': {'type': 'runtime'},
  'ReversePolygonNormals': {'type': 'runtime'},
  'ReversePolygonNormalsOptions': {'type': 'runtime'},
  'ReverseSurfaceDirection': {'type': 'runtime'},
  'ReverseSurfaceDirectionOptions': {'type': 'runtime'},
  'ReverseToonObjects': {'type': 'runtime'},
  'Revolve': {'type': 'runtime'},
  'RevolveOptions': {'type': 'runtime'},
  'RigidBindSkin': {'type': 'runtime'},
  'RigidBindSkinOptions': {'type': 'runtime'},
  'RigidBodySolver': {'type': 'runtime'},
  'RotateTool': {'type': 'runtime'},
  'RotateToolMarkingMenu': {'type': 'runtime'},
  'RotateToolMarkingMenuPopDown': {'type': 'runtime'},
  'RotateToolOptions': {'type': 'runtime'},
  'RotateToolWithSnapMarkingMenu': {'type': 'runtime'},
  'RotateToolWithSnapMarkingMenuPopDown': {'type': 'runtime'},
  'RotateUVTool': {'type': 'runtime'},
  'RotateUVs': {'type': 'runtime'},
  'RotateUVsOptions': {'type': 'runtime'},
  'RoundTool': {'type': 'runtime'},
  'RoundToolOptions': {'type': 'runtime'},
  'STRSTweakModeOff': {'type': 'runtime'},
  'STRSTweakModeOn': {'type': 'runtime'},
  'STRSTweakModeToggle': {'type': 'runtime'},
  'SaveBrushPreset': {'type': 'runtime'},
  'SaveCurrentLayout': {'type': 'runtime'},
  'SaveCurrentWorkspace': {'type': 'runtime'},
  'SaveFluidStateAs': {'type': 'runtime'},
  'SaveInitialStateAll': {'type': 'runtime'},
  'SavePreferences': {'type': 'runtime'},
  'SaveScene': {'type': 'runtime'},
  'SaveSceneAs': {'type': 'runtime'},
  'SaveSceneAsOptions': {'type': 'runtime'},
  'SaveSceneOptions': {'type': 'runtime'},
  'ScaleConstraint': {'type': 'runtime'},
  'ScaleConstraintOptions': {'type': 'runtime'},
  'ScaleCurvature': {'type': 'runtime'},
  'ScaleCurvatureOptions': {'type': 'runtime'},
  'ScaleKeys': {'type': 'runtime'},
  'ScaleKeysOptions': {'type': 'runtime'},
  'ScaleTool': {'type': 'runtime'},
  'ScaleToolMarkingMenu': {'type': 'runtime'},
  'ScaleToolMarkingMenuPopDown': {'type': 'runtime'},
  'ScaleToolOptions': {'type': 'runtime'},
  'ScaleToolWithSnapMarkingMenu': {'type': 'runtime'},
  'ScaleToolWithSnapMarkingMenuPopDown': {'type': 'runtime'},
  'ScaleUVTool': {'type': 'runtime'},
  'ScriptEditor': {'type': 'runtime'},
  'ScriptPaintTool': {'type': 'runtime'},
  'ScriptPaintToolOptions': {'type': 'runtime'},
  'SculptGeometryTool': {'type': 'runtime'},
  'SculptGeometryToolOptions': {'type': 'runtime'},
  'SculptMeshActivateBrushSize': {'type': 'runtime'},
  'SculptMeshActivateBrushStrength': {'type': 'runtime'},
  'SculptMeshDeactivateBrushSize': {'type': 'runtime'},
  'SculptMeshDeactivateBrushStrength': {'type': 'runtime'},
  'SculptMeshFrame': {'type': 'runtime'},
  'SculptMeshInvertFreeze': {'type': 'runtime'},
  'SculptMeshUnfreezeAll': {'type': 'runtime'},
  'SculptPolygonsTool': {'type': 'runtime'},
  'SculptPolygonsToolOptions': {'type': 'runtime'},
  'SculptReferenceVectorMarkingMenuPress': {'type': 'runtime'},
  'SculptReferenceVectorMarkingMenuRelease': {'type': 'runtime'},
  'SculptSubdivsTool': {'type': 'runtime'},
  'SculptSubdivsToolOptions': {'type': 'runtime'},
  'SculptSurfacesTool': {'type': 'runtime'},
  'SculptSurfacesToolOptions': {'type': 'runtime'},
  'SearchAndReplaceNames': {'type': 'runtime'},
  'SelectAll': {'type': 'runtime'},
  'SelectAllAssets': {'type': 'runtime'},
  'SelectAllBrushes': {'type': 'runtime'},
  'SelectAllCameras': {'type': 'runtime'},
  'SelectAllClusters': {'type': 'runtime'},
  'SelectAllDynamicConstraints': {'type': 'runtime'},
  'SelectAllFluids': {'type': 'runtime'},
  'SelectAllFollicles': {'type': 'runtime'},
  'SelectAllFurs': {'type': 'runtime'},
  'SelectAllGeometry': {'type': 'runtime'},
  'SelectAllHairSystem': {'type': 'runtime'},
  'SelectAllIKHandles': {'type': 'runtime'},
  'SelectAllImagePlanes': {'type': 'runtime'},
  'SelectAllInput': {'type': 'runtime'},
  'SelectAllJoints': {'type': 'runtime'},
  'SelectAllLattices': {'type': 'runtime'},
  'SelectAllLights': {'type': 'runtime'},
  'SelectAllMarkingMenu': {'type': 'runtime'},
  'SelectAllMarkingMenuPopDown': {'type': 'runtime'},
  'SelectAllNCloths': {'type': 'runtime'},
  'SelectAllNParticles': {'type': 'runtime'},
  'SelectAllNRigids': {'type': 'runtime'},
  'SelectAllNURBSCurves': {'type': 'runtime'},
  'SelectAllNURBSSurfaces': {'type': 'runtime'},
  'SelectAllOutput': {'type': 'runtime'},
  'SelectAllParticles': {'type': 'runtime'},
  'SelectAllPolygonGeometry': {'type': 'runtime'},
  'SelectAllRigidBodies': {'type': 'runtime'},
  'SelectAllRigidConstraints': {'type': 'runtime'},
  'SelectAllSculptObjects': {'type': 'runtime'},
  'SelectAllStrokes': {'type': 'runtime'},
  'SelectAllSubdivGeometry': {'type': 'runtime'},
  'SelectAllTransforms': {'type': 'runtime'},
  'SelectAllWires': {'type': 'runtime'},
  'SelectBorderEdgeTool': {'type': 'runtime'},
  'SelectBrushNames': {'type': 'runtime'},
  'SelectCVSelectionBoundary': {'type': 'runtime'},
  'SelectCVsMask': {'type': 'runtime'},
  'SelectComponentToolMarkingMenu': {'type': 'runtime'},
  'SelectComponentToolMarkingMenuPopDown': {'type': 'runtime'},
  'SelectContainerContents': {'type': 'runtime'},
  'SelectContiguousEdges': {'type': 'runtime'},
  'SelectContiguousEdgesOptions': {'type': 'runtime'},
  'SelectCurveCVsAll': {'type': 'runtime'},
  'SelectCurveCVsFirst': {'type': 'runtime'},
  'SelectCurveCVsLast': {'type': 'runtime'},
  'SelectCurvePointsMask': {'type': 'runtime'},
  'SelectEdgeLoop': {'type': 'runtime'},
  'SelectEdgeLoopSp': {'type': 'runtime'},
  'SelectEdgeMask': {'type': 'runtime'},
  'SelectEdgeRing': {'type': 'runtime'},
  'SelectEdgeRingSp': {'type': 'runtime'},
  'SelectFacePath': {'type': 'runtime'},
  'SelectFacetMask': {'type': 'runtime'},
  'SelectHierarchy': {'type': 'runtime'},
  'SelectHullsMask': {'type': 'runtime'},
  'SelectIsolate': {'type': 'runtime'},
  'SelectLightsIlluminatingObject': {'type': 'runtime'},
  'SelectLightsShadowingObject': {'type': 'runtime'},
  'SelectLinesMask': {'type': 'runtime'},
  'SelectMaskToolMarkingMenu': {'type': 'runtime'},
  'SelectMaskToolMarkingMenuPopDown': {'type': 'runtime'},
  'SelectMeshUVShell': {'type': 'runtime'},
  'SelectMultiComponentMask': {'type': 'runtime'},
  'SelectNextIntermediatObject': {'type': 'runtime'},
  'SelectNone': {'type': 'runtime'},
  'SelectObjectsIlluminatedByLight': {'type': 'runtime'},
  'SelectObjectsShadowedByLight': {'type': 'runtime'},
  'SelectPointsMask': {'type': 'runtime'},
  'SelectPolygonSelectionBoundary': {'type': 'runtime'},
  'SelectPolygonToolMarkingMenu': {'type': 'runtime'},
  'SelectPolygonToolMarkingMenuPopDown': {'type': 'runtime'},
  'SelectPreviousObjects3dsMax': {'type': 'runtime'},
  'SelectPreviousObjectsMotionBuilder': {'type': 'runtime'},
  'SelectPreviousObjectsMudbox': {'type': 'runtime'},
  'SelectSharedColorInstances': {'type': 'runtime'},
  'SelectSharedUVInstances': {'type': 'runtime'},
  'SelectShortestEdgePathTool': {'type': 'runtime'},
  'SelectSimilar': {'type': 'runtime'},
  'SelectSimilarOptions': {'type': 'runtime'},
  'SelectSurfaceBorder': {'type': 'runtime'},
  'SelectSurfaceBorderOptions': {'type': 'runtime'},
  'SelectSurfacePointsMask': {'type': 'runtime'},
  'SelectTextureReferenceObject': {'type': 'runtime'},
  'SelectTimeWarp': {'type': 'runtime'},
  'SelectToggleMode': {'type': 'runtime'},
  'SelectTool': {'type': 'runtime'},
  'SelectToolMarkingMenu': {'type': 'runtime'},
  'SelectToolMarkingMenuPopDown': {'type': 'runtime'},
  'SelectToolOptionsMarkingMenu': {'type': 'runtime'},
  'SelectToolOptionsMarkingMenuPopDown': {'type': 'runtime'},
  'SelectUVBackFacingComponents': {'type': 'runtime'},
  'SelectUVBorder': {'type': 'runtime'},
  'SelectUVBorderComponents': {'type': 'runtime'},
  'SelectUVFrontFacingComponents': {'type': 'runtime'},
  'SelectUVMask': {'type': 'runtime'},
  'SelectUVNonOverlappingComponents': {'type': 'runtime'},
  'SelectUVNonOverlappingComponentsPerObject': {'type': 'runtime'},
  'SelectUVOverlappingComponents': {'type': 'runtime'},
  'SelectUVOverlappingComponentsPerObject': {'type': 'runtime'},
  'SelectUVShell': {'type': 'runtime'},
  'SelectUVTool': {'type': 'runtime'},
  'SelectUnmappedFaces': {'type': 'runtime'},
  'SelectVertexFaceMask': {'type': 'runtime'},
  'SelectVertexMask': {'type': 'runtime'},
  'SelectedAnimLayer': {'type': 'runtime'},
  'SendAsNewScene3dsMax': {'type': 'runtime'},
  'SendAsNewSceneMotionBuilder': {'type': 'runtime'},
  'SendAsNewSceneMudbox': {'type': 'runtime'},
  'SeparatePolygon': {'type': 'runtime'},
  'SequenceEditor': {'type': 'runtime'},
  'SetActiveKey': {'type': 'runtime'},
  'SetAlignTool': {'type': 'runtime'},
  'SetAsCombinationTarget': {'type': 'runtime'},
  'SetAsCombinationTargetOptions': {'type': 'runtime'},
  'SetBreakdownKey': {'type': 'runtime'},
  'SetBreakdownKeyOptions': {'type': 'runtime'},
  'SetCMCAmbient': {'type': 'runtime'},
  'SetCMCAmbientDiffuse': {'type': 'runtime'},
  'SetCMCDiffuse': {'type': 'runtime'},
  'SetCMCEmission': {'type': 'runtime'},
  'SetCMCNone': {'type': 'runtime'},
  'SetCMCSpecular': {'type': 'runtime'},
  'SetCurrentColorSet': {'type': 'runtime'},
  'SetCurrentUVSet': {'type': 'runtime'},
  'SetCutSewUVTool': {'type': 'runtime'},
  'SetDefaultManipMove': {'type': 'runtime'},
  'SetDefaultManipNone': {'type': 'runtime'},
  'SetDefaultManipRotate': {'type': 'runtime'},
  'SetDefaultManipScale': {'type': 'runtime'},
  'SetDefaultManipTransform': {'type': 'runtime'},
  'SetDrivenKey': {'type': 'runtime'},
  'SetDrivenKeyOptions': {'type': 'runtime'},
  'SetEditor': {'type': 'runtime'},
  'SetExclusiveToCamera': {'type': 'runtime'},
  'SetFluidAttrFromCurve': {'type': 'runtime'},
  'SetFluidAttrFromCurveOptions': {'type': 'runtime'},
  'SetFocusToCommandLine': {'type': 'runtime'},
  'SetFocusToNumericInputLine': {'type': 'runtime'},
  'SetFullBodyIKKeys': {'type': 'runtime'},
  'SetFullBodyIKKeysAll': {'type': 'runtime'},
  'SetFullBodyIKKeysBodyPart': {'type': 'runtime'},
  'SetFullBodyIKKeysKeyToPin': {'type': 'runtime'},
  'SetFullBodyIKKeysOptions': {'type': 'runtime'},
  'SetFullBodyIKKeysSelected': {'type': 'runtime'},
  'SetHairRestPositionFromCurrent': {'type': 'runtime'},
  'SetHairRestPositionFromStart': {'type': 'runtime'},
  'SetHairStartPositionFromCurrent': {'type': 'runtime'},
  'SetHairStartPositionFromRest': {'type': 'runtime'},
  'SetHiddenFromCamera': {'type': 'runtime'},
  'SetIKFKKeyframe': {'type': 'runtime'},
  'SetInitialState': {'type': 'runtime'},
  'SetInitialStateOptions': {'type': 'runtime'},
  'SetKey': {'type': 'runtime'},
  'SetKeyAnimated': {'type': 'runtime'},
  'SetKeyOptions': {'type': 'runtime'},
  'SetKeyPath': {'type': 'runtime'},
  'SetKeyRotate': {'type': 'runtime'},
  'SetKeyScale': {'type': 'runtime'},
  'SetKeyTranslate': {'type': 'runtime'},
  'SetKeyVertexColor': {'type': 'runtime'},
  'SetKeyframeForVertexColor': {'type': 'runtime'},
  'SetMBSAdd': {'type': 'runtime'},
  'SetMBSAverage': {'type': 'runtime'},
  'SetMBSDivide': {'type': 'runtime'},
  'SetMBSModulate2': {'type': 'runtime'},
  'SetMBSMultiply': {'type': 'runtime'},
  'SetMBSOverwrite': {'type': 'runtime'},
  'SetMBSSubtract': {'type': 'runtime'},
  'SetMaxInfluences': {'type': 'runtime'},
  'SetMeshAmplifyTool': {'type': 'runtime'},
  'SetMeshBulgeTool': {'type': 'runtime'},
  'SetMeshCloneTargetTool': {'type': 'runtime'},
  'SetMeshEraseTool': {'type': 'runtime'},
  'SetMeshFillTool': {'type': 'runtime'},
  'SetMeshFlattenTool': {'type': 'runtime'},
  'SetMeshFoamyTool': {'type': 'runtime'},
  'SetMeshFreezeTool': {'type': 'runtime'},
  'SetMeshGrabTool': {'type': 'runtime'},
  'SetMeshGrabUVTool': {'type': 'runtime'},
  'SetMeshImprintTool': {'type': 'runtime'},
  'SetMeshKnifeTool': {'type': 'runtime'},
  'SetMeshMaskTool': {'type': 'runtime'},
  'SetMeshPinchTool': {'type': 'runtime'},
  'SetMeshRelaxTool': {'type': 'runtime'},
  'SetMeshRepeatTool': {'type': 'runtime'},
  'SetMeshScrapeTool': {'type': 'runtime'},
  'SetMeshSculptTool': {'type': 'runtime'},
  'SetMeshSmearTool': {'type': 'runtime'},
  'SetMeshSmoothTargetTool': {'type': 'runtime'},
  'SetMeshSmoothTool': {'type': 'runtime'},
  'SetMeshSprayTool': {'type': 'runtime'},
  'SetMeshWaxTool': {'type': 'runtime'},
  'SetNClothStartFromMesh': {'type': 'runtime'},
  'SetNormalAngle': {'type': 'runtime'},
  'SetPassiveKey': {'type': 'runtime'},
  'SetPreferredAngle': {'type': 'runtime'},
  'SetPreferredAngleOptions': {'type': 'runtime'},
  'SetProject': {'type': 'runtime'},
  'SetReFormTool': {'type': 'runtime'},
  'SetRestPosition': {'type': 'runtime'},
  'SetRigidBodyCollision': {'type': 'runtime'},
  'SetRigidBodyInterpenetration': {'type': 'runtime'},
  'SetShrinkWrapInnerObject': {'type': 'runtime'},
  'SetShrinkWrapTarget': {'type': 'runtime'},
  'SetSnapTogetherTool': {'type': 'runtime'},
  'SetSnapTogetherToolOptions': {'type': 'runtime'},
  'SetStrokeControlCurves': {'type': 'runtime'},
  'SetTimecode': {'type': 'runtime'},
  'SetToFaceNormals': {'type': 'runtime'},
  'SetToFaceNormalsOptions': {'type': 'runtime'},
  'SetVertexNormal': {'type': 'runtime'},
  'SetVertexNormalOptions': {'type': 'runtime'},
  'SetWireframeColor': {'type': 'runtime'},
  'SetWorkingFrame': {'type': 'runtime'},
  'SetupAnimatedDisplacement': {'type': 'runtime'},
  'SewUVs': {'type': 'runtime'},
  'SewUVs3D': {'type': 'runtime'},
  'SewUVsWithoutHotkey': {'type': 'runtime'},
  'ShadingGroupAttributeEditor': {'type': 'runtime'},
  'ShapeEditor': {'type': 'runtime'},
  'ShapeEditorDuplicateTarget': {'type': 'runtime'},
  'ShapeEditorNewGroup': {'type': 'runtime'},
  'ShapeEditorSelectNone': {'type': 'runtime'},
  'ShareColorInstances': {'type': 'runtime'},
  'ShareOneBrush': {'type': 'runtime'},
  'ShareUVInstances': {'type': 'runtime'},
  'Shatter': {'type': 'runtime'},
  'ShatterOptions': {'type': 'runtime'},
  'ShelfPreferencesWindow': {'type': 'runtime'},
  'ShortPolygonNormals': {'type': 'runtime'},
  'ShotPlaylistEditor': {'type': 'runtime'},
  'ShowAll': {'type': 'runtime'},
  'ShowAllComponents': {'type': 'runtime'},
  'ShowAllEditedComponents': {'type': 'runtime'},
  'ShowAllLabels': {'type': 'runtime'},
  'ShowAllPolyComponents': {'type': 'runtime'},
  'ShowAllUI': {'type': 'runtime'},
  'ShowAnimationUI': {'type': 'runtime'},
  'ShowAttributeEditorOrChannelBox': {'type': 'runtime'},
  'ShowBaseWire': {'type': 'runtime'},
  'ShowBatchRender': {'type': 'runtime'},
  'ShowBoundingBox': {'type': 'runtime'},
  'ShowCameraManipulators': {'type': 'runtime'},
  'ShowCameras': {'type': 'runtime'},
  'ShowClusters': {'type': 'runtime'},
  'ShowControllers': {'type': 'runtime'},
  'ShowDeformers': {'type': 'runtime'},
  'ShowDeformingGeometry': {'type': 'runtime'},
  'ShowDynamicConstraints': {'type': 'runtime'},
  'ShowDynamicsUI': {'type': 'runtime'},
  'ShowFluids': {'type': 'runtime'},
  'ShowFollicles': {'type': 'runtime'},
  'ShowFur': {'type': 'runtime'},
  'ShowGeometry': {'type': 'runtime'},
  'ShowHairSystems': {'type': 'runtime'},
  'ShowHotbox': {'type': 'runtime'},
  'ShowIKHandles': {'type': 'runtime'},
  'ShowJoints': {'type': 'runtime'},
  'ShowKinematics': {'type': 'runtime'},
  'ShowLastHidden': {'type': 'runtime'},
  'ShowLattices': {'type': 'runtime'},
  'ShowLightManipulators': {'type': 'runtime'},
  'ShowLights': {'type': 'runtime'},
  'ShowManipulatorTool': {'type': 'runtime'},
  'ShowManipulators': {'type': 'runtime'},
  'ShowMarkers': {'type': 'runtime'},
  'ShowMeshAmplifyToolOptions': {'type': 'runtime'},
  'ShowMeshBulgeToolOptions': {'type': 'runtime'},
  'ShowMeshCloneTargetToolOptions': {'type': 'runtime'},
  'ShowMeshEraseToolOptions': {'type': 'runtime'},
  'ShowMeshFillToolOptions': {'type': 'runtime'},
  'ShowMeshFlattenToolOptions': {'type': 'runtime'},
  'ShowMeshFoamyToolOptions': {'type': 'runtime'},
  'ShowMeshFreezeToolOptions': {'type': 'runtime'},
  'ShowMeshGrabToolOptions': {'type': 'runtime'},
  'ShowMeshGrabUVToolOptions': {'type': 'runtime'},
  'ShowMeshImprintToolOptions': {'type': 'runtime'},
  'ShowMeshKnifeToolOptions': {'type': 'runtime'},
  'ShowMeshMaskToolOptions': {'type': 'runtime'},
  'ShowMeshPinchToolOptions': {'type': 'runtime'},
  'ShowMeshRelaxToolOptions': {'type': 'runtime'},
  'ShowMeshRepeatToolOptions': {'type': 'runtime'},
  'ShowMeshScrapeToolOptions': {'type': 'runtime'},
  'ShowMeshSculptToolOptions': {'type': 'runtime'},
  'ShowMeshSmearToolOptions': {'type': 'runtime'},
  'ShowMeshSmoothTargetToolOptions': {'type': 'runtime'},
  'ShowMeshSmoothToolOptions': {'type': 'runtime'},
  'ShowMeshSprayToolOptions': {'type': 'runtime'},
  'ShowMeshWaxToolOptions': {'type': 'runtime'},
  'ShowModelingUI': {'type': 'runtime'},
  'ShowNCloths': {'type': 'runtime'},
  'ShowNParticles': {'type': 'runtime'},
  'ShowNRigids': {'type': 'runtime'},
  'ShowNURBSCurves': {'type': 'runtime'},
  'ShowNURBSSurfaces': {'type': 'runtime'},
  'ShowNonlinears': {'type': 'runtime'},
  'ShowObjectGeometry': {'type': 'runtime'},
  'ShowPlanes': {'type': 'runtime'},
  'ShowPolygonSurfaces': {'type': 'runtime'},
  'ShowRenderingUI': {'type': 'runtime'},
  'ShowResultsOptions': {'type': 'runtime'},
  'ShowRiggingUI': {'type': 'runtime'},
  'ShowSculptObjects': {'type': 'runtime'},
  'ShowSelectedObjects': {'type': 'runtime'},
  'ShowShadingGroupAttributeEditor': {'type': 'runtime'},
  'ShowSmoothSkinInfluences': {'type': 'runtime'},
  'ShowStrokeControlCurves': {'type': 'runtime'},
  'ShowStrokePathCurves': {'type': 'runtime'},
  'ShowStrokes': {'type': 'runtime'},
  'ShowSubdivSurfaces': {'type': 'runtime'},
  'ShowSurfaceCVs': {'type': 'runtime'},
  'ShowTexturePlacements': {'type': 'runtime'},
  'ShowUIElements': {'type': 'runtime'},
  'ShowWhatsNew': {'type': 'runtime'},
  'ShowWrapInfluences': {'type': 'runtime'},
  'ShrinkLoopPolygonSelectionRegion': {'type': 'runtime'},
  'ShrinkPolygonSelectionRegion': {'type': 'runtime'},
  'SimplifyCurve': {'type': 'runtime'},
  'SimplifyCurveOptions': {'type': 'runtime'},
  'SimplifyStrokePathCurves': {'type': 'runtime'},
  'Sine': {'type': 'runtime'},
  'SineOptions': {'type': 'runtime'},
  'SinglePerspectiveViewLayout': {'type': 'runtime'},
  'SingleViewArrangement': {'type': 'runtime'},
  'SlideEdgeTool': {'type': 'runtime'},
  'SlideEdgeToolOptions': {'type': 'runtime'},
  'Smoke': {'type': 'runtime'},
  'SmokeOptions': {'type': 'runtime'},
  'SmoothBindSkin': {'type': 'runtime'},
  'SmoothBindSkinOptions': {'type': 'runtime'},
  'SmoothCurve': {'type': 'runtime'},
  'SmoothCurveOptions': {'type': 'runtime'},
  'SmoothHairCurves': {'type': 'runtime'},
  'SmoothHairCurvesOptions': {'type': 'runtime'},
  'SmoothPolygon': {'type': 'runtime'},
  'SmoothPolygonOptions': {'type': 'runtime'},
  'SmoothProxy': {'type': 'runtime'},
  'SmoothProxyOptions': {'type': 'runtime'},
  'SmoothSkinWeights': {'type': 'runtime'},
  'SmoothSkinWeightsOptions': {'type': 'runtime'},
  'SmoothTangent': {'type': 'runtime'},
  'SmoothingDisplayShowBoth': {'type': 'runtime'},
  'SmoothingDisplayToggle': {'type': 'runtime'},
  'SmoothingLevelDecrease': {'type': 'runtime'},
  'SmoothingLevelIncrease': {'type': 'runtime'},
  'Snap2PointsTo2Points': {'type': 'runtime'},
  'Snap2PointsTo2PointsOptions': {'type': 'runtime'},
  'Snap3PointsTo3Points': {'type': 'runtime'},
  'Snap3PointsTo3PointsOptions': {'type': 'runtime'},
  'SnapKeys': {'type': 'runtime'},
  'SnapKeysOptions': {'type': 'runtime'},
  'SnapPointToPoint': {'type': 'runtime'},
  'SnapPointToPointOptions': {'type': 'runtime'},
  'SnapRotation': {'type': 'runtime'},
  'SnapToCurve': {'type': 'runtime'},
  'SnapToCurvePress': {'type': 'runtime'},
  'SnapToCurveRelease': {'type': 'runtime'},
  'SnapToGrid': {'type': 'runtime'},
  'SnapToGridPress': {'type': 'runtime'},
  'SnapToGridRelease': {'type': 'runtime'},
  'SnapToMeshCenter': {'type': 'runtime'},
  'SnapToMeshCenterPress': {'type': 'runtime'},
  'SnapToMeshCenterRelease': {'type': 'runtime'},
  'SnapToPixel': {'type': 'runtime'},
  'SnapToPoint': {'type': 'runtime'},
  'SnapToPointPress': {'type': 'runtime'},
  'SnapToPointRelease': {'type': 'runtime'},
  'SoftModDeformer': {'type': 'runtime'},
  'SoftModDeformerOptions': {'type': 'runtime'},
  'SoftModTool': {'type': 'runtime'},
  'SoftModToolOptions': {'type': 'runtime'},
  'SoloLastOutput': {'type': 'runtime'},
  'SoloMaterial': {'type': 'runtime'},
  'SplitEdge': {'type': 'runtime'},
  'SplitEdgeRingTool': {'type': 'runtime'},
  'SplitEdgeRingToolOptions': {'type': 'runtime'},
  'SplitMeshWithProjectedCurve': {'type': 'runtime'},
  'SplitMeshWithProjectedCurveOptions': {'type': 'runtime'},
  'SplitPolygonTool': {'type': 'runtime'},
  'SplitPolygonToolOptions': {'type': 'runtime'},
  'SplitUV': {'type': 'runtime'},
  'SplitVertex': {'type': 'runtime'},
  'SpreadSheetEditor': {'type': 'runtime'},
  'SquareSurface': {'type': 'runtime'},
  'SquareSurfaceOptions': {'type': 'runtime'},
  'Squash': {'type': 'runtime'},
  'SquashOptions': {'type': 'runtime'},
  'StitchEdgesTool': {'type': 'runtime'},
  'StitchEdgesToolOptions': {'type': 'runtime'},
  'StitchSurfacePoints': {'type': 'runtime'},
  'StitchSurfacePointsOptions': {'type': 'runtime'},
  'StitchTogether': {'type': 'runtime'},
  'StitchTogetherOptions': {'type': 'runtime'},
  'StraightenCurves': {'type': 'runtime'},
  'StraightenCurvesOptions': {'type': 'runtime'},
  'StraightenUVBorder': {'type': 'runtime'},
  'StraightenUVBorderOptions': {'type': 'runtime'},
  'SubdCutUVs': {'type': 'runtime'},
  'SubdivProxy': {'type': 'runtime'},
  'SubdivProxyOptions': {'type': 'runtime'},
  'SubdivSmoothnessFine': {'type': 'runtime'},
  'SubdivSmoothnessFineOptions': {'type': 'runtime'},
  'SubdivSmoothnessHull': {'type': 'runtime'},
  'SubdivSmoothnessHullOptions': {'type': 'runtime'},
  'SubdivSmoothnessMedium': {'type': 'runtime'},
  'SubdivSmoothnessMediumOptions': {'type': 'runtime'},
  'SubdivSmoothnessRough': {'type': 'runtime'},
  'SubdivSmoothnessRoughOptions': {'type': 'runtime'},
  'SubdivSurfaceCleanTopology': {'type': 'runtime'},
  'SubdivSurfaceHierarchyMode': {'type': 'runtime'},
  'SubdivSurfaceMatchTopology': {'type': 'runtime'},
  'SubdivSurfacePolygonProxyMode': {'type': 'runtime'},
  'SubdivToNURBS': {'type': 'runtime'},
  'SubdivToNURBSOptions': {'type': 'runtime'},
  'SubdividePolygon': {'type': 'runtime'},
  'SubdividePolygonOptions': {'type': 'runtime'},
  'SubstituteGeometry': {'type': 'runtime'},
  'SubstituteGeometryOptions': {'type': 'runtime'},
  'SurfaceBooleanIntersectTool': {'type': 'runtime'},
  'SurfaceBooleanIntersectToolOptions': {'type': 'runtime'},
  'SurfaceBooleanSubtractTool': {'type': 'runtime'},
  'SurfaceBooleanSubtractToolOptions': {'type': 'runtime'},
  'SurfaceBooleanUnionTool': {'type': 'runtime'},
  'SurfaceBooleanUnionToolOptions': {'type': 'runtime'},
  'SurfaceEditingTool': {'type': 'runtime'},
  'SurfaceEditingToolOptions': {'type': 'runtime'},
  'SurfaceFlow': {'type': 'runtime'},
  'SurfaceFlowOptions': {'type': 'runtime'},
  'SwapBlendShape': {'type': 'runtime'},
  'SwapBlendShapeOptions': {'type': 'runtime'},
  'SwapBufferCurve': {'type': 'runtime'},
  'Symmetrize': {'type': 'runtime'},
  'SymmetrizeSelection': {'type': 'runtime'},
  'SymmetrizeUV': {'type': 'runtime'},
  'SymmetrizeUVBrushSizeOff': {'type': 'runtime'},
  'SymmetrizeUVBrushSizeOn': {'type': 'runtime'},
  'SymmetrizeUVOptions': {'type': 'runtime'},
  'TagAsController': {'type': 'runtime'},
  'TagAsControllerParent': {'type': 'runtime'},
  'TangentConstraint': {'type': 'runtime'},
  'TangentConstraintOptions': {'type': 'runtime'},
  'TangentsAuto': {'type': 'runtime'},
  'TangentsClamped': {'type': 'runtime'},
  'TangentsFixed': {'type': 'runtime'},
  'TangentsFlat': {'type': 'runtime'},
  'TangentsLinear': {'type': 'runtime'},
  'TangentsPlateau': {'type': 'runtime'},
  'TangentsSpline': {'type': 'runtime'},
  'TangentsStepped': {'type': 'runtime'},
  'TanimLayer': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'TemplateBrushSettings': {'type': 'runtime'},
  'TemplateObject': {'type': 'runtime'},
  'Tension': {'type': 'runtime'},
  'TensionOptions': {'type': 'runtime'},
  'TesselateSubdivSurface': {'type': 'runtime'},
  'TesselateSubdivSurfaceOptions': {'type': 'runtime'},
  'TestTexture': {'type': 'runtime'},
  'TestTextureOptions': {'type': 'runtime'},
  'TexSculptActivateBrushSize': {'type': 'runtime'},
  'TexSculptActivateBrushStrength': {'type': 'runtime'},
  'TexSculptDeactivateBrushSize': {'type': 'runtime'},
  'TexSculptDeactivateBrushStrength': {'type': 'runtime'},
  'TexSculptInvertPin': {'type': 'runtime'},
  'TexSculptUnpinAll': {'type': 'runtime'},
  'TexSewActivateBrushSize': {'type': 'runtime'},
  'TexSewDeactivateBrushSize': {'type': 'runtime'},
  'TextureCentricUVLinkingEditor': {'type': 'runtime'},
  'TextureToGeometry': {'type': 'runtime'},
  'TextureToGeometryOptions': {'type': 'runtime'},
  'TextureViewWindow': {'type': 'runtime'},
  'ThreeBottomSplitViewArrangement': {'type': 'runtime'},
  'ThreeLeftSplitViewArrangement': {'type': 'runtime'},
  'ThreePointArcTool': {'type': 'runtime'},
  'ThreePointArcToolOptions': {'type': 'runtime'},
  'ThreeRightSplitViewArrangement': {'type': 'runtime'},
  'ThreeTopSplitViewArrangement': {'type': 'runtime'},
  'TimeDraggerToolActivate': {'type': 'runtime'},
  'TimeDraggerToolDeactivate': {'type': 'runtime'},
  'TimeEditorAddToSoloSelectedTracks': {'type': 'runtime'},
  'TimeEditorClipHoldToggle': {'type': 'runtime'},
  'TimeEditorClipLoopToggle': {'type': 'runtime'},
  'TimeEditorClipRazor': {'type': 'runtime'},
  'TimeEditorClipResetTiming': {'type': 'runtime'},
  'TimeEditorClipScaleEnd': {'type': 'runtime'},
  'TimeEditorClipScaleStart': {'type': 'runtime'},
  'TimeEditorClipScaleToggle': {'type': 'runtime'},
  'TimeEditorClipTrimEnd': {'type': 'runtime'},
  'TimeEditorClipTrimStart': {'type': 'runtime'},
  'TimeEditorClipTrimToggle': {'type': 'runtime'},
  'TimeEditorCopyClips': {'type': 'runtime'},
  'TimeEditorCreateAdditiveLayer': {'type': 'runtime'},
  'TimeEditorCreateAnimTracksAtEnd': {'type': 'runtime'},
  'TimeEditorCreateAudioClip': {'type': 'runtime'},
  'TimeEditorCreateAudioTracksAtEnd': {'type': 'runtime'},
  'TimeEditorCreateClip': {'type': 'runtime'},
  'TimeEditorCreateClipOptions': {'type': 'runtime'},
  'TimeEditorCreateGroupFromSelection': {'type': 'runtime'},
  'TimeEditorCreateOverrideLayer': {'type': 'runtime'},
  'TimeEditorCreatePoseClip': {'type': 'runtime'},
  'TimeEditorCreatePoseClipOptions': {'type': 'runtime'},
  'TimeEditorCutClips': {'type': 'runtime'},
  'TimeEditorDeleteClips': {'type': 'runtime'},
  'TimeEditorDeleteSelectedTracks': {'type': 'runtime'},
  'TimeEditorExplodeGroup': {'type': 'runtime'},
  'TimeEditorExportSelection': {'type': 'runtime'},
  'TimeEditorExportSelectionOpt': {'type': 'runtime'},
  'TimeEditorFbxExportAll': {'type': 'runtime'},
  'TimeEditorFrameAll': {'type': 'runtime'},
  'TimeEditorFrameCenterView': {'type': 'runtime'},
  'TimeEditorFramePlaybackRange': {'type': 'runtime'},
  'TimeEditorFrameSelected': {'type': 'runtime'},
  'TimeEditorGhostTrackToggle': {'type': 'runtime'},
  'TimeEditorImportAnimation': {'type': 'runtime'},
  'TimeEditorKeepTransitionsTogglePress': {'type': 'runtime'},
  'TimeEditorKeepTransitionsToggleRelease': {'type': 'runtime'},
  'TimeEditorMuteSelectedTracks': {'type': 'runtime'},
  'TimeEditorOpenContentBrowser': {'type': 'runtime'},
  'TimeEditorPasteClips': {'type': 'runtime'},
  'TimeEditorRealTimeRefreshToggle': {'type': 'runtime'},
  'TimeEditorRippleEditTogglePress': {'type': 'runtime'},
  'TimeEditorRippleEditToggleRelease': {'type': 'runtime'},
  'TimeEditorSceneAuthoringToggle': {'type': 'runtime'},
  'TimeEditorSetKey': {'type': 'runtime'},
  'TimeEditorSetZeroKey': {'type': 'runtime'},
  'TimeEditorSoloSelectedTracks': {'type': 'runtime'},
  'TimeEditorToggleMuteSelectedTracks': {'type': 'runtime'},
  'TimeEditorToggleSnapToClipPress': {'type': 'runtime'},
  'TimeEditorToggleSnapToClipRelease': {'type': 'runtime'},
  'TimeEditorToggleSoloSelectedTracks': {'type': 'runtime'},
  'TimeEditorToggleTimeCursorPress': {'type': 'runtime'},
  'TimeEditorToggleTimeCursorRelease': {'type': 'runtime'},
  'TimeEditorUnmuteAllTracks': {'type': 'runtime'},
  'TimeEditorUnmuteSelectedTracks': {'type': 'runtime'},
  'TimeEditorUnsoloAllTracks': {'type': 'runtime'},
  'TimeEditorWindow': {'type': 'runtime'},
  'ToggleAnimationDetails': {'type': 'runtime'},
  'ToggleAttributeEditor': {'type': 'runtime'},
  'ToggleAutoActivateBodyPart': {'type': 'runtime'},
  'ToggleAutoFrame': {'type': 'runtime'},
  'ToggleAutoFrameTime': {'type': 'runtime'},
  'ToggleAutoSmooth': {'type': 'runtime'},
  'ToggleBackfaceCulling': {'type': 'runtime'},
  'ToggleBackfaceGeometry': {'type': 'runtime'},
  'ToggleBorderEdges': {'type': 'runtime'},
  'ToggleCVs': {'type': 'runtime'},
  'ToggleCacheVisibility': {'type': 'runtime'},
  'ToggleCameraNames': {'type': 'runtime'},
  'ToggleCapsLockDisplay': {'type': 'runtime'},
  'ToggleChannelBox': {'type': 'runtime'},
  'ToggleChannelsLayers': {'type': 'runtime'},
  'ToggleCharacterControls': {'type': 'runtime'},
  'ToggleColorFeedback': {'type': 'runtime'},
  'ToggleCommandLine': {'type': 'runtime'},
  'ToggleCompIDs': {'type': 'runtime'},
  'ToggleContainerCentric': {'type': 'runtime'},
  'ToggleCreaseEdges': {'type': 'runtime'},
  'ToggleCreaseVertices': {'type': 'runtime'},
  'ToggleCreateNurbsPrimitivesAsTool': {'type': 'runtime'},
  'ToggleCreatePolyPrimitivesAsTool': {'type': 'runtime'},
  'ToggleCullingVertices': {'type': 'runtime'},
  'ToggleCurrentContainerHud': {'type': 'runtime'},
  'ToggleCurrentFrame': {'type': 'runtime'},
  'ToggleCustomNURBSComponents': {'type': 'runtime'},
  'ToggleDisplacement': {'type': 'runtime'},
  'ToggleDisplayColorsAttr': {'type': 'runtime'},
  'ToggleDisplayGradient': {'type': 'runtime'},
  'ToggleEdgeIDs': {'type': 'runtime'},
  'ToggleEdgeMetadata': {'type': 'runtime'},
  'ToggleEditPivot': {'type': 'runtime'},
  'ToggleEditPoints': {'type': 'runtime'},
  'ToggleEffectsMeshDisplay': {'type': 'runtime'},
  'ToggleEvaluationManagerVisibility': {'type': 'runtime'},
  'ToggleFaceIDs': {'type': 'runtime'},
  'ToggleFaceMetadata': {'type': 'runtime'},
  'ToggleFaceNormalDisplay': {'type': 'runtime'},
  'ToggleFaceNormals': {'type': 'runtime'},
  'ToggleFastInteraction': {'type': 'runtime'},
  'ToggleFkIk': {'type': 'runtime'},
  'ToggleFocalLength': {'type': 'runtime'},
  'ToggleFrameRate': {'type': 'runtime'},
  'ToggleFullScreenMode': {'type': 'runtime'},
  'ToggleGrid': {'type': 'runtime'},
  'ToggleHelpLine': {'type': 'runtime'},
  'ToggleHikDetails': {'type': 'runtime'},
  'ToggleHoleFaces': {'type': 'runtime'},
  'ToggleHulls': {'type': 'runtime'},
  'ToggleIKAllowRotation': {'type': 'runtime'},
  'ToggleIKHandleSnap': {'type': 'runtime'},
  'ToggleIKSolvers': {'type': 'runtime'},
  'ToggleInViewEditor': {'type': 'runtime'},
  'ToggleInViewMessage': {'type': 'runtime'},
  'ToggleIsolateSelect': {'type': 'runtime'},
  'ToggleJointLabels': {'type': 'runtime'},
  'ToggleKeepHardEdgeCulling': {'type': 'runtime'},
  'ToggleKeepWireCulling': {'type': 'runtime'},
  'ToggleLatticePoints': {'type': 'runtime'},
  'ToggleLatticeShape': {'type': 'runtime'},
  'ToggleLayerBar': {'type': 'runtime'},
  'ToggleLocalRotationAxes': {'type': 'runtime'},
  'ToggleMainMenubar': {'type': 'runtime'},
  'ToggleMaterialLoadingDetailsVisibility': {'type': 'runtime'},
  'ToggleMeshEdges': {'type': 'runtime'},
  'ToggleMeshFaces': {'type': 'runtime'},
  'ToggleMeshMaps': {'type': 'runtime'},
  'ToggleMeshPoints': {'type': 'runtime'},
  'ToggleMeshUVBorders': {'type': 'runtime'},
  'ToggleMetadata': {'type': 'runtime'},
  'ToggleModelEditorBars': {'type': 'runtime'},
  'ToggleModelingToolkit': {'type': 'runtime'},
  'ToggleMultiColorFeedback': {'type': 'runtime'},
  'ToggleNormals': {'type': 'runtime'},
  'ToggleNurbsPrimitivesAsToolExitOnComplete': {'type': 'runtime'},
  'ToggleObjectDetails': {'type': 'runtime'},
  'ToggleOppositeFlagOfSelectedShapes': {'type': 'runtime'},
  'ToggleOriginAxis': {'type': 'runtime'},
  'ToggleOutliner': {'type': 'runtime'},
  'TogglePaintAtDepth': {'type': 'runtime'},
  'TogglePaintOnPaintableObjects': {'type': 'runtime'},
  'TogglePanZoomPress': {'type': 'runtime'},
  'TogglePanZoomRelease': {'type': 'runtime'},
  'TogglePanelMenubar': {'type': 'runtime'},
  'ToggleParticleCount': {'type': 'runtime'},
  'TogglePolyCount': {'type': 'runtime'},
  'TogglePolyDisplayEdges': {'type': 'runtime'},
  'TogglePolyDisplayHardEdges': {'type': 'runtime'},
  'TogglePolyDisplayHardEdgesColor': {'type': 'runtime'},
  'TogglePolyDisplayLimitToSelected': {'type': 'runtime'},
  'TogglePolyDisplaySoftEdges': {'type': 'runtime'},
  'TogglePolyNonPlanarFaceDisplay': {'type': 'runtime'},
  'TogglePolyPrimitivesAsToolExitOnComplete': {'type': 'runtime'},
  'TogglePolyUVsCreateShader': {'type': 'runtime'},
  'TogglePolygonFaceCenters': {'type': 'runtime'},
  'TogglePolygonFaceTriangles': {'type': 'runtime'},
  'TogglePolygonFaceTrianglesDisplay': {'type': 'runtime'},
  'ToggleProxyDisplay': {'type': 'runtime'},
  'ToggleRangeSlider': {'type': 'runtime'},
  'ToggleReflection': {'type': 'runtime'},
  'ToggleRotationPivots': {'type': 'runtime'},
  'ToggleScalePivots': {'type': 'runtime'},
  'ToggleSceneTimecode': {'type': 'runtime'},
  'ToggleSelectDetails': {'type': 'runtime'},
  'ToggleSelectedLabels': {'type': 'runtime'},
  'ToggleSelectionHandles': {'type': 'runtime'},
  'ToggleShelf': {'type': 'runtime'},
  'ToggleShowBufferCurves': {'type': 'runtime'},
  'ToggleShowResults': {'type': 'runtime'},
  'ToggleSoftEdges': {'type': 'runtime'},
  'ToggleStatusLine': {'type': 'runtime'},
  'ToggleSubdDetails': {'type': 'runtime'},
  'ToggleSurfaceFaceCenters': {'type': 'runtime'},
  'ToggleSurfaceOrigin': {'type': 'runtime'},
  'ToggleSymmetryDisplay': {'type': 'runtime'},
  'ToggleTangentDisplay': {'type': 'runtime'},
  'ToggleTextureBorder': {'type': 'runtime'},
  'ToggleTextureBorderEdges': {'type': 'runtime'},
  'ToggleTimeSlider': {'type': 'runtime'},
  'ToggleToolMessage': {'type': 'runtime'},
  'ToggleToolSettings': {'type': 'runtime'},
  'ToggleToolbox': {'type': 'runtime'},
  'ToggleUIElements': {'type': 'runtime'},
  'ToggleUVDistortion': {'type': 'runtime'},
  'ToggleUVEditorIsolateSelectHUD': {'type': 'runtime'},
  'ToggleUVEditorUVStatisticsHUD': {'type': 'runtime'},
  'ToggleUVEditorUVStatisticsHUDOptions': {'type': 'runtime'},
  'ToggleUVIsolateViewSelected': {'type': 'runtime'},
  'ToggleUVShellBorder': {'type': 'runtime'},
  'ToggleUVTextureImage': {'type': 'runtime'},
  'ToggleUVs': {'type': 'runtime'},
  'ToggleUnsharedUVs': {'type': 'runtime'},
  'ToggleUseDefaultMaterial': {'type': 'runtime'},
  'ToggleVertIDs': {'type': 'runtime'},
  'ToggleVertMetadata': {'type': 'runtime'},
  'ToggleVertexNormalDisplay': {'type': 'runtime'},
  'ToggleVertices': {'type': 'runtime'},
  'ToggleViewAxis': {'type': 'runtime'},
  'ToggleViewCube': {'type': 'runtime'},
  'ToggleViewportRenderer': {'type': 'runtime'},
  'ToggleVisibilityAndKeepSelection': {'type': 'runtime'},
  'ToggleVisibilityAndKeepSelectionOptions': {'type': 'runtime'},
  'ToggleWireframeInArtisan': {'type': 'runtime'},
  'ToggleZoomInMode': {'type': 'runtime'},
  'ToolSettingsWindow': {'type': 'runtime'},
  'TrackTool': {'type': 'runtime'},
  'TransferAttributeValues': {'type': 'runtime'},
  'TransferAttributeValuesOptions': {'type': 'runtime'},
  'TransferAttributes': {'type': 'runtime'},
  'TransferShadingSets': {'type': 'runtime'},
  'TransferVertexOrder': {'type': 'runtime'},
  'TransformNoSelectOffTool': {'type': 'runtime'},
  'TransformNoSelectOnTool': {'type': 'runtime'},
  'TransformPolygonComponent': {'type': 'runtime'},
  'TransformPolygonComponentOptions': {'type': 'runtime'},
  'TranslateToolMarkingMenu': {'type': 'runtime'},
  'TranslateToolMarkingMenuPopDown': {'type': 'runtime'},
  'TranslateToolWithSnapMarkingMenu': {'type': 'runtime'},
  'TranslateToolWithSnapMarkingMenuPopDown': {'type': 'runtime'},
  'TransplantHair': {'type': 'runtime'},
  'TransplantHairOptions': {'type': 'runtime'},
  'TrenderSetupStates': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'Triangulate': {'type': 'runtime'},
  'TrimTool': {'type': 'runtime'},
  'TrimToolOptions': {'type': 'runtime'},
  'TruncateHairCache': {'type': 'runtime'},
  'TumbleTool': {'type': 'runtime'},
  'Turbulence': {'type': 'runtime'},
  'TurbulenceOptions': {'type': 'runtime'},
  'Twist': {'type': 'runtime'},
  'TwistOptions': {'type': 'runtime'},
  'TwoPointArcTool': {'type': 'runtime'},
  'TwoPointArcToolOptions': {'type': 'runtime'},
  'TwoSideBySideViewArrangement': {'type': 'runtime'},
  'TwoStackedViewArrangement': {'type': 'runtime'},
  'U3DBrushPressureOff': {'type': 'runtime'},
  'U3DBrushPressureOn': {'type': 'runtime'},
  'U3DBrushSizeOff': {'type': 'runtime'},
  'U3DBrushSizeOn': {'type': 'runtime'},
  'UIModeMarkingMenu': {'type': 'runtime'},
  'UIModeMarkingMenuPopDown': {'type': 'runtime'},
  'UVAutomaticProjection': {'type': 'runtime'},
  'UVAutomaticProjectionOptions': {'type': 'runtime'},
  'UVCameraBasedProjection': {'type': 'runtime'},
  'UVCameraBasedProjectionOptions': {'type': 'runtime'},
  'UVCentricUVLinkingEditor': {'type': 'runtime'},
  'UVContourStretchProjection': {'type': 'runtime'},
  'UVContourStretchProjectionOptions': {'type': 'runtime'},
  'UVCreateSnapshot': {'type': 'runtime'},
  'UVCylindricProjection': {'type': 'runtime'},
  'UVCylindricProjectionOptions': {'type': 'runtime'},
  'UVEditorFrameAll': {'type': 'runtime'},
  'UVEditorFrameSelected': {'type': 'runtime'},
  'UVEditorInvertPin': {'type': 'runtime'},
  'UVEditorResetAllToDefault': {'type': 'runtime'},
  'UVEditorToggleTextureBorderDisplay': {'type': 'runtime'},
  'UVEditorUnpinAll': {'type': 'runtime'},
  'UVGatherShells': {'type': 'runtime'},
  'UVIsolateLoadSet': {'type': 'runtime'},
  'UVNormalBasedProjection': {'type': 'runtime'},
  'UVNormalBasedProjectionOptions': {'type': 'runtime'},
  'UVOrientShells': {'type': 'runtime'},
  'UVPlanarProjection': {'type': 'runtime'},
  'UVPlanarProjectionOptions': {'type': 'runtime'},
  'UVSetEditor': {'type': 'runtime'},
  'UVSnapTogether': {'type': 'runtime'},
  'UVSnapTogetherOptions': {'type': 'runtime'},
  'UVSphericalProjection': {'type': 'runtime'},
  'UVSphericalProjectionOptions': {'type': 'runtime'},
  'UVStackSimilarShells': {'type': 'runtime'},
  'UVStackSimilarShellsOptions': {'type': 'runtime'},
  'UVStraighten': {'type': 'runtime'},
  'UVStraightenOptions': {'type': 'runtime'},
  'UVUnstackShells': {'type': 'runtime'},
  'UVUnstackShellsOptions': {'type': 'runtime'},
  'UncreaseSubdivSurface': {'type': 'runtime'},
  'Undo': {'type': 'runtime'},
  'UndoCanvas': {'type': 'runtime'},
  'UndoViewChange': {'type': 'runtime'},
  'UnfoldPackUVs3DInCurrentTile': {'type': 'runtime'},
  'UnfoldPackUVs3DInEmptyTile': {'type': 'runtime'},
  'UnfoldUV': {'type': 'runtime'},
  'UnfoldUVOptions': {'type': 'runtime'},
  'UnghostAll': {'type': 'runtime'},
  'UnghostObject': {'type': 'runtime'},
  'Ungroup': {'type': 'runtime'},
  'UngroupOptions': {'type': 'runtime'},
  'Uniform': {'type': 'runtime'},
  'UniformOptions': {'type': 'runtime'},
  'UnifyTangents': {'type': 'runtime'},
  'UnitizeUVs': {'type': 'runtime'},
  'UnitizeUVsOptions': {'type': 'runtime'},
  'UniversalManip': {'type': 'runtime'},
  'UniversalManipOptions': {'type': 'runtime'},
  'UnlockContainer': {'type': 'runtime'},
  'UnlockCurveLength': {'type': 'runtime'},
  'UnlockNormals': {'type': 'runtime'},
  'UnmirrorSmoothProxy': {'type': 'runtime'},
  'UnmirrorSmoothProxyOptions': {'type': 'runtime'},
  'Unparent': {'type': 'runtime'},
  'UnparentOptions': {'type': 'runtime'},
  'UnpinSelection': {'type': 'runtime'},
  'UnpublishAttributes': {'type': 'runtime'},
  'UnpublishChildAnchor': {'type': 'runtime'},
  'UnpublishNode': {'type': 'runtime'},
  'UnpublishParentAnchor': {'type': 'runtime'},
  'UnpublishRootTransform': {'type': 'runtime'},
  'UntemplateObject': {'type': 'runtime'},
  'UntrimSurfaces': {'type': 'runtime'},
  'UntrimSurfacesOptions': {'type': 'runtime'},
  'UpdateBindingSet': {'type': 'runtime'},
  'UpdateBindingSetOptions': {'type': 'runtime'},
  'UpdateCurrentScene3dsMax': {'type': 'runtime'},
  'UpdateCurrentSceneMotionBuilder': {'type': 'runtime'},
  'UpdateCurrentSceneMudbox': {'type': 'runtime'},
  'UpdateEraseSurface': {'type': 'runtime'},
  'UpdatePSDTextureItem': {'type': 'runtime'},
  'UpdateReferenceSurface': {'type': 'runtime'},
  'UpdateSnapshot': {'type': 'runtime'},
  'UseHardwareShader': {'type': 'runtime'},
  'UseSelectedEmitter': {'type': 'runtime'},
  'VertexNormalEditTool': {'type': 'runtime'},
  'ViewAlongAxisNegativeX': {'type': 'runtime'},
  'ViewAlongAxisNegativeY': {'type': 'runtime'},
  'ViewAlongAxisNegativeZ': {'type': 'runtime'},
  'ViewAlongAxisX': {'type': 'runtime'},
  'ViewAlongAxisY': {'type': 'runtime'},
  'ViewAlongAxisZ': {'type': 'runtime'},
  'ViewImage': {'type': 'runtime'},
  'ViewSequence': {'type': 'runtime'},
  'ViewportEnableSmoothing': {'type': 'runtime'},
  'VisorWindow': {'type': 'runtime'},
  'VisualizeMetadataOptions': {'type': 'runtime'},
  'VolumeAxis': {'type': 'runtime'},
  'VolumeAxisOptions': {'type': 'runtime'},
  'VolumeSkinBinding': {'type': 'runtime'},
  'VolumeSkinBindingOptions': {'type': 'runtime'},
  'Vortex': {'type': 'runtime'},
  'VortexOptions': {'type': 'runtime'},
  'WalkTool': {'type': 'runtime'},
  'WarpImage': {'type': 'runtime'},
  'WarpImageOptions': {'type': 'runtime'},
  'Wave': {'type': 'runtime'},
  'WaveOptions': {'type': 'runtime'},
  'WedgePolygon': {'type': 'runtime'},
  'WedgePolygonOptions': {'type': 'runtime'},
  'WeightHammer': {'type': 'runtime'},
  'WeightedTangents': {'type': 'runtime'},
  'WhatsNewHighlightingOff': {'type': 'runtime'},
  'WhatsNewHighlightingOn': {'type': 'runtime'},
  'WhatsNewStartupDialogOff': {'type': 'runtime'},
  'WhatsNewStartupDialogOn': {'type': 'runtime'},
  'WireDropoffLocator': {'type': 'runtime'},
  'WireDropoffLocatorOptions': {'type': 'runtime'},
  'WireTool': {'type': 'runtime'},
  'WireToolOptions': {'type': 'runtime'},
  'WrinkleTool': {'type': 'runtime'},
  'WrinkleToolOptions': {'type': 'runtime'},
  'XgCreateIgSplineEditor': {'type': 'runtime'},
  'XgmCreateInteractiveGroomSplines': {'type': 'runtime'},
  'XgmCreateInteractiveGroomSplinesOption': {'type': 'runtime'},
  'XgmSetClumpBrushTool': {'type': 'runtime'},
  'XgmSetClumpBrushToolOption': {'type': 'runtime'},
  'XgmSetCombBrushTool': {'type': 'runtime'},
  'XgmSetCombBrushToolOption': {'type': 'runtime'},
  'XgmSetCutBrushTool': {'type': 'runtime'},
  'XgmSetCutBrushToolOption': {'type': 'runtime'},
  'XgmSetDensityBrushTool': {'type': 'runtime'},
  'XgmSetDensityBrushToolOption': {'type': 'runtime'},
  'XgmSetDirectionBrushTool': {'type': 'runtime'},
  'XgmSetDirectionBrushToolOption': {'type': 'runtime'},
  'XgmSetFreezeBrushTool': {'type': 'runtime'},
  'XgmSetFreezeBrushToolOption': {'type': 'runtime'},
  'XgmSetGrabBrushTool': {'type': 'runtime'},
  'XgmSetGrabBrushToolOption': {'type': 'runtime'},
  'XgmSetLengthBrushTool': {'type': 'runtime'},
  'XgmSetLengthBrushToolOption': {'type': 'runtime'},
  'XgmSetNoiseBrushTool': {'type': 'runtime'},
  'XgmSetNoiseBrushToolOption': {'type': 'runtime'},
  'XgmSetPartBrushTool': {'type': 'runtime'},
  'XgmSetPartBrushToolOption': {'type': 'runtime'},
  'XgmSetPlaceBrushTool': {'type': 'runtime'},
  'XgmSetPlaceBrushToolOption': {'type': 'runtime'},
  'XgmSetSelectBrushTool': {'type': 'runtime'},
  'XgmSetSelectBrushToolOption': {'type': 'runtime'},
  'XgmSetSmoothBrushTool': {'type': 'runtime'},
  'XgmSetSmoothBrushToolOption': {'type': 'runtime'},
  'XgmSetWidthBrushTool': {'type': 'runtime'},
  'XgmSetWidthBrushToolOption': {'type': 'runtime'},
  'XgmSplineCacheCreate': {'type': 'runtime'},
  'XgmSplineCacheCreateOptions': {'type': 'runtime'},
  'XgmSplineCacheDelete': {'type': 'runtime'},
  'XgmSplineCacheDeleteNodesAhead': {'type': 'runtime'},
  'XgmSplineCacheDeleteOptions': {'type': 'runtime'},
  'XgmSplineCacheDisableSelectedCache': {'type': 'runtime'},
  'XgmSplineCacheEnableSelectedCache': {'type': 'runtime'},
  'XgmSplineCacheExport': {'type': 'runtime'},
  'XgmSplineCacheExportOptions': {'type': 'runtime'},
  'XgmSplineCacheImport': {'type': 'runtime'},
  'XgmSplineCacheImportOptions': {'type': 'runtime'},
  'XgmSplineCacheReplace': {'type': 'runtime'},
  'XgmSplineCacheReplaceOptions': {'type': 'runtime'},
  'XgmSplineGeometryConvert': {'type': 'runtime'},
  'XgmSplinePresetExport': {'type': 'runtime'},
  'XgmSplinePresetImport': {'type': 'runtime'},
  'XgmSplineSelectConvertToFreeze': {'type': 'runtime'},
  'XgmSplineSelectReplaceBySelectedFaces': {'type': 'runtime'},
  'ZoomTool': {'type': 'runtime'},
  'aaf2fcp': {'flags': {'deleteFile': {'args': '<type __builtin__.bool>',
                                       'longname': 'deleteFile',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'df'},
                        'dstPath': {'args': '<type __builtin__.str>',
                                    'longname': 'dstPath',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'dst'},
                        'getFileName': {'args': '<type __builtin__.int>',
                                        'longname': 'getFileName',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'gfn'},
                        'progress': {'args': '<type __builtin__.int>',
                                     'longname': 'progress',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'pr'},
                        'srcFile': {'args': '<type __builtin__.str>',
                                    'longname': 'srcFile',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'src'},
                        'terminate': {'args': '<type __builtin__.int>',
                                      'longname': 'terminate',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 't'},
                        'waitCompletion': {'args': '<type __builtin__.int>',
                                           'longname': 'waitCompletion',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'wc'}},
              'shortFlags': {'df': 'deleteFile',
                             'dst': 'dstPath',
                             'gfn': 'getFileName',
                             'pr': 'progress',
                             'src': 'srcFile',
                             't': 'terminate',
                             'wc': 'waitCompletion'},
              'type': 'system'},
  'about': {'flags': {'apiVersion': {'args': '<type __builtin__.bool>',
                                     'longname': 'apiVersion',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'api'},
                      'application': {'args': '<type __builtin__.bool>',
                                      'longname': 'application',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'a'},
                      'batch': {'args': '<type __builtin__.bool>',
                                'longname': 'batch',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'b'},
                      'buildDirectory': {'args': '<type __builtin__.bool>',
                                         'longname': 'buildDirectory',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'bd'},
                      'buildVariant': {'args': '<type __builtin__.bool>',
                                       'longname': 'buildVariant',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'bv'},
                      'codeset': {'args': '<type __builtin__.bool>',
                                  'longname': 'codeset',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'cs'},
                      'compositingManager': {'args': '<type __builtin__.bool>',
                                             'longname': 'compositingManager',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'cm'},
                      'connected': {'args': '<type __builtin__.bool>',
                                    'longname': 'connected',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'cnt'},
                      'ctime': {'args': '<type __builtin__.bool>',
                                'longname': 'ctime',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'cti'},
                      'currentDate': {'args': '<type __builtin__.bool>',
                                      'longname': 'currentDate',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'cd'},
                      'currentTime': {'args': '<type __builtin__.bool>',
                                      'longname': 'currentTime',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ct'},
                      'cutIdentifier': {'args': '<type __builtin__.bool>',
                                        'longname': 'cutIdentifier',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'c'},
                      'date': {'args': '<type __builtin__.bool>',
                               'longname': 'date',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'd'},
                      'environmentFile': {'args': '<type __builtin__.bool>',
                                          'longname': 'environmentFile',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'env'},
                      'evalVersion': {'args': '<type __builtin__.bool>',
                                      'longname': 'evalVersion',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ev'},
                      'file': {'args': '<type __builtin__.bool>',
                               'longname': 'file',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'f'},
                      'fontInfo': {'args': '<type __builtin__.bool>',
                                   'longname': 'fontInfo',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'foi'},
                      'helpDataDirectory': {'args': '<type __builtin__.bool>',
                                            'longname': 'helpDataDirectory',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'hdd'},
                      'installedVersion': {'args': '<type __builtin__.bool>',
                                           'longname': 'installedVersion',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'iv'},
                      'ioVersion': {'args': '<type __builtin__.bool>',
                                    'longname': 'ioVersion',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'io'},
                      'irix': {'args': '<type __builtin__.bool>',
                               'longname': 'irix',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'ir'},
                      'is64': {'args': '<type __builtin__.bool>',
                               'longname': 'is64',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'x64'},
                      'languageResources': {'args': '<type __builtin__.bool>',
                                            'longname': 'languageResources',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'lr'},
                      'linux': {'args': '<type __builtin__.bool>',
                                'longname': 'linux',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'li'},
                      'linux64': {'args': '<type __builtin__.bool>',
                                  'longname': 'linux64',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'l64'},
                      'liveUpdate': {'args': '<type __builtin__.bool>',
                                     'longname': 'liveUpdate',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'lu'},
                      'localizedResourceLocation': {'args': '<type __builtin__.bool>',
                                                    'longname': 'localizedResourceLocation',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'lrl'},
                      'ltVersion': {'args': '<type __builtin__.bool>',
                                    'longname': 'ltVersion',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'lt'},
                      'macOS': {'args': '<type __builtin__.bool>',
                                'longname': 'macOS',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'mac'},
                      'macOSppc': {'args': '<type __builtin__.bool>',
                                   'longname': 'macOSppc',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'ppc'},
                      'macOSx86': {'args': '<type __builtin__.bool>',
                                   'longname': 'macOSx86',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'x86'},
                      'ntOS': {'args': '<type __builtin__.bool>',
                               'longname': 'ntOS',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'nt'},
                      'operatingSystem': {'args': '<type __builtin__.bool>',
                                          'longname': 'operatingSystem',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'os'},
                      'operatingSystemVersion': {'args': '<type __builtin__.bool>',
                                                 'longname': 'operatingSystemVersion',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'osv'},
                      'preferences': {'args': '<type __builtin__.bool>',
                                      'longname': 'preferences',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'pd'},
                      'product': {'args': '<type __builtin__.bool>',
                                  'longname': 'product',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'p'},
                      'qtVersion': {'args': '<type __builtin__.bool>',
                                    'longname': 'qtVersion',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'qt'},
                      'query': {'args': '<type __builtin__.bool>',
                                'longname': 'query',
                                'numArgs': 0,
                                'shortname': 'q'},
                      'tablet': {'args': '<type __builtin__.bool>',
                                 'longname': 'tablet',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'tab'},
                      'tabletMode': {'args': '<type __builtin__.bool>',
                                     'longname': 'tabletMode',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'tm'},
                      'uiLanguage': {'args': '<type __builtin__.bool>',
                                     'longname': 'uiLanguage',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'uil'},
                      'uiLanguageForStartup': {'args': '<type __builtin__.bool>',
                                               'longname': 'uiLanguageForStartup',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'uis'},
                      'uiLanguageIsLocalized': {'args': '<type __builtin__.bool>',
                                                'longname': 'uiLanguageIsLocalized',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'uii'},
                      'uiLocaleLanguage': {'args': '<type __builtin__.bool>',
                                           'longname': 'uiLocaleLanguage',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ull'},
                      'version': {'args': '<type __builtin__.bool>',
                                  'longname': 'version',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'v'},
                      'win64': {'args': '<type __builtin__.bool>',
                                'longname': 'win64',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'w64'},
                      'windowManager': {'args': '<type __builtin__.bool>',
                                        'longname': 'windowManager',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'wm'},
                      'windows': {'args': '<type __builtin__.bool>',
                                  'longname': 'windows',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'win'}},
            'shortFlags': {'a': 'application',
                           'api': 'apiVersion',
                           'b': 'batch',
                           'bd': 'buildDirectory',
                           'bv': 'buildVariant',
                           'c': 'cutIdentifier',
                           'cd': 'currentDate',
                           'cm': 'compositingManager',
                           'cnt': 'connected',
                           'cs': 'codeset',
                           'ct': 'currentTime',
                           'cti': 'ctime',
                           'd': 'date',
                           'env': 'environmentFile',
                           'ev': 'evalVersion',
                           'f': 'file',
                           'foi': 'fontInfo',
                           'hdd': 'helpDataDirectory',
                           'io': 'ioVersion',
                           'ir': 'irix',
                           'iv': 'installedVersion',
                           'l64': 'linux64',
                           'li': 'linux',
                           'lr': 'languageResources',
                           'lrl': 'localizedResourceLocation',
                           'lt': 'ltVersion',
                           'lu': 'liveUpdate',
                           'mac': 'macOS',
                           'nt': 'ntOS',
                           'os': 'operatingSystem',
                           'osv': 'operatingSystemVersion',
                           'p': 'product',
                           'pd': 'preferences',
                           'ppc': 'macOSppc',
                           'q': 'query',
                           'qt': 'qtVersion',
                           'tab': 'tablet',
                           'tm': 'tabletMode',
                           'uii': 'uiLanguageIsLocalized',
                           'uil': 'uiLanguage',
                           'uis': 'uiLanguageForStartup',
                           'ull': 'uiLocaleLanguage',
                           'v': 'version',
                           'w64': 'win64',
                           'win': 'windows',
                           'wm': 'windowManager',
                           'x64': 'is64',
                           'x86': 'macOSx86'},
            'type': 'general'},
  'addAttr': {'flags': {'attributeType': {'args': '<type __builtin__.str>',
                                          'longname': 'attributeType',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'at'},
                        'binaryTag': {'args': '<type __builtin__.str>',
                                      'longname': 'binaryTag',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'bt'},
                        'cachedInternally': {'args': '<type __builtin__.bool>',
                                             'longname': 'cachedInternally',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'ci'},
                        'category': {'args': '<type __builtin__.str>',
                                     'longname': 'category',
                                     'modes': ['create',
                                               'query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'ct'},
                        'dataType': {'args': '<type __builtin__.str>',
                                     'longname': 'dataType',
                                     'modes': ['create',
                                               'query',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'dt'},
                        'defaultValue': {'args': '<type __builtin__.float>',
                                         'longname': 'defaultValue',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dv'},
                        'disconnectBehaviour': {'args': '<type __builtin__.int>',
                                                'longname': 'disconnectBehaviour',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'dcb'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'enumName': {'args': '<type __builtin__.str>',
                                     'longname': 'enumName',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'en'},
                        'exists': {'args': '<type __builtin__.bool>',
                                   'longname': 'exists',
                                   'modes': ['create', 'query'],
                                   'numArgs': 0,
                                   'shortname': 'ex'},
                        'fromPlugin': {'args': '<type __builtin__.bool>',
                                       'longname': 'fromPlugin',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'fp'},
                        'hasMaxValue': {'args': '<type __builtin__.bool>',
                                        'longname': 'hasMaxValue',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'hxv'},
                        'hasMinValue': {'args': '<type __builtin__.bool>',
                                        'longname': 'hasMinValue',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'hnv'},
                        'hasSoftMaxValue': {'args': '<type __builtin__.bool>',
                                            'longname': 'hasSoftMaxValue',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'hsx'},
                        'hasSoftMinValue': {'args': '<type __builtin__.bool>',
                                            'longname': 'hasSoftMinValue',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'hsn'},
                        'hidden': {'args': '<type __builtin__.bool>',
                                   'longname': 'hidden',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'h'},
                        'indexMatters': {'args': '<type __builtin__.bool>',
                                         'longname': 'indexMatters',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'im'},
                        'internalSet': {'args': '<type __builtin__.bool>',
                                        'longname': 'internalSet',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'internalSet'},
                        'keyable': {'args': '<type __builtin__.bool>',
                                    'longname': 'keyable',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'k'},
                        'longName': {'args': '<type __builtin__.str>',
                                     'longname': 'longName',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'ln'},
                        'maxValue': {'args': '<type __builtin__.float>',
                                     'longname': 'maxValue',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'max'},
                        'minValue': {'args': '<type __builtin__.float>',
                                     'longname': 'minValue',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'min'},
                        'multi': {'args': '<type __builtin__.bool>',
                                  'longname': 'multi',
                                  'modes': ['create', 'query'],
                                  'numArgs': 0,
                                  'shortname': 'm'},
                        'niceName': {'args': '<type __builtin__.str>',
                                     'longname': 'niceName',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'nn'},
                        'numberOfChildren': {'args': '<type __builtin__.int>',
                                             'longname': 'numberOfChildren',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'nc'},
                        'parent': {'args': '<type __builtin__.str>',
                                   'longname': 'parent',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'p'},
                        'proxy': {'args': '<type __builtin__.str>',
                                  'longname': 'proxy',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'pxy'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'readable': {'args': '<type __builtin__.bool>',
                                     'longname': 'readable',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'r'},
                        'shortName': {'args': '<type __builtin__.str>',
                                      'longname': 'shortName',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'sn'},
                        'softMaxValue': {'args': '<type __builtin__.float>',
                                         'longname': 'softMaxValue',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'smx'},
                        'softMinValue': {'args': '<type __builtin__.float>',
                                         'longname': 'softMinValue',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'smn'},
                        'storable': {'args': '<type __builtin__.bool>',
                                     'longname': 'storable',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 's'},
                        'usedAsColor': {'args': '<type __builtin__.bool>',
                                        'longname': 'usedAsColor',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'uac'},
                        'usedAsFilename': {'args': '<type __builtin__.bool>',
                                           'longname': 'usedAsFilename',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'uaf'},
                        'usedAsProxy': {'args': '<type __builtin__.bool>',
                                        'longname': 'usedAsProxy',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'uap'},
                        'writable': {'args': '<type __builtin__.bool>',
                                     'longname': 'writable',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'w'}},
              'removedFlags': {'is': 'internalSet'},
              'shortFlags': {'at': 'attributeType',
                             'bt': 'binaryTag',
                             'ci': 'cachedInternally',
                             'ct': 'category',
                             'dcb': 'disconnectBehaviour',
                             'dt': 'dataType',
                             'dv': 'defaultValue',
                             'e': 'edit',
                             'en': 'enumName',
                             'ex': 'exists',
                             'fp': 'fromPlugin',
                             'h': 'hidden',
                             'hnv': 'hasMinValue',
                             'hsn': 'hasSoftMinValue',
                             'hsx': 'hasSoftMaxValue',
                             'hxv': 'hasMaxValue',
                             'im': 'indexMatters',
                             'internalSet': 'internalSet',
                             'k': 'keyable',
                             'ln': 'longName',
                             'm': 'multi',
                             'max': 'maxValue',
                             'min': 'minValue',
                             'nc': 'numberOfChildren',
                             'nn': 'niceName',
                             'p': 'parent',
                             'pxy': 'proxy',
                             'q': 'query',
                             'r': 'readable',
                             's': 'storable',
                             'smn': 'softMinValue',
                             'smx': 'softMaxValue',
                             'sn': 'shortName',
                             'uac': 'usedAsColor',
                             'uaf': 'usedAsFilename',
                             'uap': 'usedAsProxy',
                             'w': 'writable'},
              'type': 'general'},
  'addDynamic': {'flags': {}, 'shortFlags': {}, 'type': 'effects'},
  'addExtension': {'flags': {'attributeType': {'args': '<type __builtin__.str>',
                                               'longname': 'attributeType',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'at'},
                             'binaryTag': {'args': '<type __builtin__.str>',
                                           'longname': 'binaryTag',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'bt'},
                             'cachedInternally': {'args': '<type __builtin__.bool>',
                                                  'longname': 'cachedInternally',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'ci'},
                             'category': {'args': '<type __builtin__.str>',
                                          'longname': 'category',
                                          'modes': ['create',
                                                    'query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'ct'},
                             'dataType': {'args': '<type __builtin__.str>',
                                          'longname': 'dataType',
                                          'modes': ['create',
                                                    'query',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'dt'},
                             'defaultValue': {'args': '<type __builtin__.float>',
                                              'longname': 'defaultValue',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dv'},
                             'disconnectBehaviour': {'args': '<type __builtin__.int>',
                                                     'longname': 'disconnectBehaviour',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'dcb'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enumName': {'args': '<type __builtin__.str>',
                                          'longname': 'enumName',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fromPlugin': {'args': '<type __builtin__.bool>',
                                            'longname': 'fromPlugin',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'fp'},
                             'hasMaxValue': {'args': '<type __builtin__.bool>',
                                             'longname': 'hasMaxValue',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'hxv'},
                             'hasMinValue': {'args': '<type __builtin__.bool>',
                                             'longname': 'hasMinValue',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'hnv'},
                             'hasSoftMaxValue': {'args': '<type __builtin__.bool>',
                                                 'longname': 'hasSoftMaxValue',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'hsx'},
                             'hasSoftMinValue': {'args': '<type __builtin__.bool>',
                                                 'longname': 'hasSoftMinValue',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'hsn'},
                             'hidden': {'args': '<type __builtin__.bool>',
                                        'longname': 'hidden',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'indexMatters': {'args': '<type __builtin__.bool>',
                                              'longname': 'indexMatters',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'im'},
                             'internalSet': {'args': '<type __builtin__.bool>',
                                             'longname': 'internalSet',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'internalSet'},
                             'keyable': {'args': '<type __builtin__.bool>',
                                         'longname': 'keyable',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'k'},
                             'longName': {'args': '<type __builtin__.str>',
                                          'longname': 'longName',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'ln'},
                             'maxValue': {'args': '<type __builtin__.float>',
                                          'longname': 'maxValue',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'max'},
                             'minValue': {'args': '<type __builtin__.float>',
                                          'longname': 'minValue',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'min'},
                             'multi': {'args': '<type __builtin__.bool>',
                                       'longname': 'multi',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'm'},
                             'niceName': {'args': '<type __builtin__.str>',
                                          'longname': 'niceName',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nn'},
                             'nodeType': {'args': '<type __builtin__.str>',
                                          'longname': 'nodeType',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nt'},
                             'numberOfChildren': {'args': '<type __builtin__.int>',
                                                  'longname': 'numberOfChildren',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'nc'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'proxy': {'args': '<type __builtin__.str>',
                                       'longname': 'proxy',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'pxy'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'readable': {'args': '<type __builtin__.bool>',
                                          'longname': 'readable',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'r'},
                             'shortName': {'args': '<type __builtin__.str>',
                                           'longname': 'shortName',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'sn'},
                             'softMaxValue': {'args': '<type __builtin__.float>',
                                              'longname': 'softMaxValue',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'smx'},
                             'softMinValue': {'args': '<type __builtin__.float>',
                                              'longname': 'softMinValue',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'smn'},
                             'storable': {'args': '<type __builtin__.bool>',
                                          'longname': 'storable',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 's'},
                             'usedAsColor': {'args': '<type __builtin__.bool>',
                                             'longname': 'usedAsColor',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'uac'},
                             'usedAsFilename': {'args': '<type __builtin__.bool>',
                                                'longname': 'usedAsFilename',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 0,
                                                'shortname': 'uaf'},
                             'usedAsProxy': {'args': '<type __builtin__.bool>',
                                             'longname': 'usedAsProxy',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'uap'},
                             'writable': {'args': '<type __builtin__.bool>',
                                          'longname': 'writable',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'w'}},
                   'removedFlags': {'is': 'internalSet'},
                   'shortFlags': {'at': 'attributeType',
                                  'bt': 'binaryTag',
                                  'ci': 'cachedInternally',
                                  'ct': 'category',
                                  'dcb': 'disconnectBehaviour',
                                  'dt': 'dataType',
                                  'dv': 'defaultValue',
                                  'e': 'edit',
                                  'en': 'enumName',
                                  'ex': 'exists',
                                  'fp': 'fromPlugin',
                                  'h': 'hidden',
                                  'hnv': 'hasMinValue',
                                  'hsn': 'hasSoftMinValue',
                                  'hsx': 'hasSoftMaxValue',
                                  'hxv': 'hasMaxValue',
                                  'im': 'indexMatters',
                                  'internalSet': 'internalSet',
                                  'k': 'keyable',
                                  'ln': 'longName',
                                  'm': 'multi',
                                  'max': 'maxValue',
                                  'min': 'minValue',
                                  'nc': 'numberOfChildren',
                                  'nn': 'niceName',
                                  'nt': 'nodeType',
                                  'p': 'parent',
                                  'pxy': 'proxy',
                                  'q': 'query',
                                  'r': 'readable',
                                  's': 'storable',
                                  'smn': 'softMinValue',
                                  'smx': 'softMaxValue',
                                  'sn': 'shortName',
                                  'uac': 'usedAsColor',
                                  'uaf': 'usedAsFilename',
                                  'uap': 'usedAsProxy',
                                  'w': 'writable'},
                   'type': 'general'},
  'addMetadata': {'flags': {'channelName': {'args': '<type __builtin__.str>',
                                            'longname': 'channelName',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'cn'},
                            'channelType': {'args': '<type __builtin__.str>',
                                            'longname': 'channelType',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'cht'},
                            'indexType': {'args': '<type __builtin__.str>',
                                          'longname': 'indexType',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'idt'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'scene': {'args': '<type __builtin__.bool>',
                                      'longname': 'scene',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'scn'},
                            'streamName': {'args': '<type __builtin__.str>',
                                           'longname': 'streamName',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'stn'},
                            'structure': {'args': '<type __builtin__.str>',
                                          'longname': 'structure',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'str'}},
                  'shortFlags': {'cht': 'channelType',
                                 'cn': 'channelName',
                                 'idt': 'indexType',
                                 'q': 'query',
                                 'scn': 'scene',
                                 'stn': 'streamName',
                                 'str': 'structure'},
                  'type': 'modeling'},
  'addPP': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                    'longname': 'attribute',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'atr'}},
            'shortFlags': {'atr': 'attribute'},
            'type': 'effects'},
  'adpAnalyticsDialog': {'flags': {'desktopAnalytics': {'args': '<type __builtin__.bool>',
                                                        'longname': 'desktopAnalytics',
                                                        'numArgs': 0,
                                                        'shortname': 'da'}},
                         'shortFlags': {'da': 'desktopAnalytics'},
                         'type': 'other'},
  'adskAsset': {'flags': {'assetID': {'args': '<type __builtin__.str>',
                                      'longname': 'assetID',
                                      'numArgs': 1,
                                      'shortname': 'a'},
                          'library': {'args': '<type __builtin__.str>',
                                      'longname': 'library',
                                      'numArgs': 1,
                                      'shortname': 'l'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'resolved': {'args': '<type __builtin__.bool>',
                                       'longname': 'resolved',
                                       'numArgs': 0,
                                       'shortname': 'r'}},
                'shortFlags': {'a': 'assetID',
                               'l': 'library',
                               'q': 'query',
                               'r': 'resolved'},
                'type': 'other'},
  'adskAssetLibrary': {'flags': {'unload': {'args': '<type __builtin__.bool>',
                                            'longname': 'unload',
                                            'numArgs': 0,
                                            'shortname': 'ul'},
                                 'unloadAll': {'args': '<type __builtin__.bool>',
                                               'longname': 'unloadAll',
                                               'numArgs': 0,
                                               'shortname': 'ua'}},
                       'shortFlags': {'ua': 'unloadAll', 'ul': 'unload'},
                       'type': 'other'},
  'adskAssetList': {'flags': {'infoType': {'args': '<type __builtin__.str>',
                                           'longname': 'infoType',
                                           'numArgs': 1,
                                           'shortname': 'it'}},
                    'shortFlags': {'it': 'infoType'},
                    'type': 'other'},
  'adskAssetListUI': {'callbackFlags': ['uiCommand',
                                        'uiC',
                                        'commandSuffix',
                                        'cms'],
                      'flags': {'commandSuffix': {'args': '<type __builtin__.str>',
                                                  'longname': 'commandSuffix',
                                                  'numArgs': 1,
                                                  'shortname': 'cms'},
                                'materialLoaded': {'args': '<type __builtin__.bool>',
                                                   'longname': 'materialLoaded',
                                                   'numArgs': 0,
                                                   'shortname': 'mld'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'uiCommand': {'args': '<type __builtin__.str>',
                                              'longname': 'uiCommand',
                                              'numArgs': 1,
                                              'shortname': 'uiC'}},
                      'shortFlags': {'cms': 'commandSuffix',
                                     'mld': 'materialLoaded',
                                     'q': 'query',
                                     'uiC': 'uiCommand'},
                      'type': 'other'},
  'affectedNet': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'type': {'args': '<type __builtin__.str>',
                                     'longname': 'type',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 't'}},
                  'shortFlags': {'e': 'edit',
                                 'n': 'name',
                                 'q': 'query',
                                 't': 'type'},
                  'type': 'general'},
  'affects': {'flags': {'by': {'args': 'Show attributes that are affected by the given one rather than the\nones that affect it.',
                               'longname': 'by',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'boolean'},
                        'type': {'args': '<type __builtin__.str>',
                                 'longname': 'type',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 't'}},
              'shortFlags': {'by': 'by', 't': 'type'},
              'type': 'general'},
  'agFormatIn': {'flags': {'file': {'args': '<type __builtin__.str>',
                                    'longname': 'file',
                                    'numArgs': 1,
                                    'shortname': 'f'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'numArgs': 1,
                                    'shortname': 'n'}},
                 'shortFlags': {'f': 'file', 'n': 'name'},
                 'type': 'other'},
  'agFormatOut': {'flags': {'file': {'args': '<type __builtin__.str>',
                                     'longname': 'file',
                                     'numArgs': 1,
                                     'shortname': 'f'}},
                  'shortFlags': {'f': 'file'},
                  'type': 'other'},
  'aimConstraint': {'flags': {'aimVector': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'aimVector',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'aim'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'layer': {'args': '<type __builtin__.str>',
                                        'longname': 'layer',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'l'},
                              'maintainOffset': {'args': '<type __builtin__.bool>',
                                                 'longname': 'maintainOffset',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'mo'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'offset': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'offset',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 3,
                                         'shortname': 'o'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'remove': {'args': '<type __builtin__.bool>',
                                         'longname': 'remove',
                                         'modes': ['edit'],
                                         'numArgs': 0,
                                         'shortname': 'rm'},
                              'skip': {'args': '<type __builtin__.str>',
                                       'longname': 'skip',
                                       'modes': ['create',
                                                 'edit',
                                                 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'sk'},
                              'targetList': {'args': '<type __builtin__.bool>',
                                             'longname': 'targetList',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'tl'},
                              'upVector': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'upVector',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'u'},
                              'weight': {'args': '<type __builtin__.float>',
                                         'longname': 'weight',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'w'},
                              'weightAliasList': {'args': '<type __builtin__.bool>',
                                                  'longname': 'weightAliasList',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'wal'},
                              'worldUpObject': {'args': 'PyNode',
                                                'longname': 'worldUpObject',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'wuo'},
                              'worldUpType': {'args': '<type __builtin__.str>',
                                              'longname': 'worldUpType',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'wut'},
                              'worldUpVector': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'worldUpVector',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'wu'}},
                    'resultNeedsUnpacking': True,
                    'shortFlags': {'aim': 'aimVector',
                                   'e': 'edit',
                                   'l': 'layer',
                                   'mo': 'maintainOffset',
                                   'n': 'name',
                                   'o': 'offset',
                                   'q': 'query',
                                   'rm': 'remove',
                                   'sk': 'skip',
                                   'tl': 'targetList',
                                   'u': 'upVector',
                                   'w': 'weight',
                                   'wal': 'weightAliasList',
                                   'wu': 'worldUpVector',
                                   'wuo': 'worldUpObject',
                                   'wut': 'worldUpType'},
                    'type': 'animation'},
  'air': {'flags': {'attenuation': {'args': '<type __builtin__.float>',
                                    'longname': 'attenuation',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'att'},
                    'directionX': {'args': '<type __builtin__.float>',
                                   'longname': 'directionX',
                                   'modes': ['query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'dx'},
                    'directionY': {'args': '<type __builtin__.float>',
                                   'longname': 'directionY',
                                   'modes': ['query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'dy'},
                    'directionZ': {'args': '<type __builtin__.float>',
                                   'longname': 'directionZ',
                                   'modes': ['query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'dz'},
                    'edit': {'args': '<type __builtin__.bool>',
                             'longname': 'edit',
                             'numArgs': 0,
                             'shortname': 'e'},
                    'enableSpread': {'args': '<type __builtin__.bool>',
                                     'longname': 'enableSpread',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'es'},
                    'fanSetup': {'args': '<type __builtin__.bool>',
                                 'longname': 'fanSetup',
                                 'modes': ['edit'],
                                 'numArgs': 0,
                                 'shortname': 'fs'},
                    'inheritRotation': {'args': '<type __builtin__.bool>',
                                        'longname': 'inheritRotation',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'iro'},
                    'inheritVelocity': {'args': '<type __builtin__.float>',
                                        'longname': 'inheritVelocity',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'iv'},
                    'magnitude': {'args': '<type __builtin__.float>',
                                  'longname': 'magnitude',
                                  'modes': ['query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'm'},
                    'maxDistance': {'args': '<type __builtin__.float>',
                                    'longname': 'maxDistance',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'mxd'},
                    'name': {'args': '<type __builtin__.str>',
                             'longname': 'name',
                             'modes': ['query', 'edit'],
                             'numArgs': 1,
                             'shortname': 'n'},
                    'perVertex': {'args': '<type __builtin__.bool>',
                                  'longname': 'perVertex',
                                  'modes': ['query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'pv'},
                    'position': {'args': ['<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>'],
                                 'longname': 'position',
                                 'modes': ['query', 'edit', 'multiuse'],
                                 'numArgs': 3,
                                 'shortname': 'pos'},
                    'query': {'args': '<type __builtin__.bool>',
                              'longname': 'query',
                              'numArgs': 0,
                              'shortname': 'q'},
                    'speed': {'args': '<type __builtin__.float>',
                              'longname': 'speed',
                              'modes': ['query', 'edit'],
                              'numArgs': 1,
                              'shortname': 's'},
                    'spread': {'args': '<type __builtin__.float>',
                               'longname': 'spread',
                               'modes': ['query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'sp'},
                    'torusSectionRadius': {'args': '<type __builtin__.float>',
                                           'longname': 'torusSectionRadius',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'tsr'},
                    'velocityComponentOnly': {'args': '<type __builtin__.bool>',
                                              'longname': 'velocityComponentOnly',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'vco'},
                    'volumeExclusion': {'args': '<type __builtin__.bool>',
                                        'longname': 'volumeExclusion',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vex'},
                    'volumeOffset': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'volumeOffset',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 3,
                                     'shortname': 'vof'},
                    'volumeShape': {'args': '<type __builtin__.str>',
                                    'longname': 'volumeShape',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'vsh'},
                    'volumeSweep': {'args': '<type __builtin__.float>',
                                    'longname': 'volumeSweep',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'vsw'},
                    'wakeSetup': {'args': '<type __builtin__.bool>',
                                  'longname': 'wakeSetup',
                                  'modes': ['edit'],
                                  'numArgs': 0,
                                  'shortname': 'wks'},
                    'windSetup': {'args': '<type __builtin__.bool>',
                                  'longname': 'windSetup',
                                  'modes': ['edit'],
                                  'numArgs': 0,
                                  'shortname': 'wns'}},
          'resultNeedsUnpacking': True,
          'shortFlags': {'att': 'attenuation',
                         'dx': 'directionX',
                         'dy': 'directionY',
                         'dz': 'directionZ',
                         'e': 'edit',
                         'es': 'enableSpread',
                         'fs': 'fanSetup',
                         'iro': 'inheritRotation',
                         'iv': 'inheritVelocity',
                         'm': 'magnitude',
                         'mxd': 'maxDistance',
                         'n': 'name',
                         'pos': 'position',
                         'pv': 'perVertex',
                         'q': 'query',
                         's': 'speed',
                         'sp': 'spread',
                         'tsr': 'torusSectionRadius',
                         'vco': 'velocityComponentOnly',
                         'vex': 'volumeExclusion',
                         'vof': 'volumeOffset',
                         'vsh': 'volumeShape',
                         'vsw': 'volumeSweep',
                         'wks': 'wakeSetup',
                         'wns': 'windSetup'},
          'type': 'effects'},
  'aliasAttr': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'remove': {'args': '<type __builtin__.bool>',
                                     'longname': 'remove',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'rm'}},
                'shortFlags': {'e': 'edit', 'q': 'query', 'rm': 'remove'},
                'type': 'general'},
  'align': {'flags': {'alignToLead': {'args': '<type __builtin__.bool>',
                                      'longname': 'alignToLead',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'atl'},
                      'coordinateSystem': {'args': 'PyNode',
                                           'longname': 'coordinateSystem',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'cs'},
                      'xAxis': {'args': '<type __builtin__.str>',
                                'longname': 'xAxis',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'x'},
                      'yAxis': {'args': '<type __builtin__.str>',
                                'longname': 'yAxis',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'y'},
                      'zAxis': {'args': '<type __builtin__.str>',
                                'longname': 'zAxis',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'z'}},
            'shortFlags': {'atl': 'alignToLead',
                           'cs': 'coordinateSystem',
                           'x': 'xAxis',
                           'y': 'yAxis',
                           'z': 'zAxis'},
            'type': 'general'},
  'alignCtx': {'flags': {'align': {'args': '<type __builtin__.bool>',
                                   'longname': 'align',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 0,
                                   'shortname': 'a'},
                         'anchorFirstObject': {'args': '<type __builtin__.bool>',
                                               'longname': 'anchorFirstObject',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'afo'},
                         'distribute': {'args': '<type __builtin__.bool>',
                                        'longname': 'distribute',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 0,
                                        'shortname': 'd'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'history': {'args': '<type __builtin__.bool>',
                                     'longname': 'history',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'ch'},
                         'image1': {'args': '<type __builtin__.str>',
                                    'longname': 'image1',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i1'},
                         'image2': {'args': '<type __builtin__.str>',
                                    'longname': 'image2',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i2'},
                         'image3': {'args': '<type __builtin__.str>',
                                    'longname': 'image3',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i3'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'showAlignTouch': {'args': '<type __builtin__.bool>',
                                            'longname': 'showAlignTouch',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sat'}},
               'shortFlags': {'a': 'align',
                              'afo': 'anchorFirstObject',
                              'ch': 'history',
                              'd': 'distribute',
                              'e': 'edit',
                              'ex': 'exists',
                              'i1': 'image1',
                              'i2': 'image2',
                              'i3': 'image3',
                              'n': 'name',
                              'q': 'query',
                              'sat': 'showAlignTouch'},
               'type': 'context'},
  'alignCurve': {'flags': {'attach': {'args': '<type __builtin__.bool>',
                                      'longname': 'attach',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'at'},
                           'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'curvatureContinuity': {'args': '<type __builtin__.bool>',
                                                   'longname': 'curvatureContinuity',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cc'},
                           'curvatureScale1': {'args': '<type __builtin__.float>',
                                               'longname': 'curvatureScale1',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cs1'},
                           'curvatureScale2': {'args': '<type __builtin__.float>',
                                               'longname': 'curvatureScale2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cs2'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'joinParameter': {'args': '<type __builtin__.float>',
                                             'longname': 'joinParameter',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'jnp'},
                           'keepMultipleKnots': {'args': '<type __builtin__.bool>',
                                                 'longname': 'keepMultipleKnots',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'kmk'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'object': {'args': '<type __builtin__.bool>',
                                      'longname': 'object',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'positionalContinuity': {'args': '<type __builtin__.bool>',
                                                    'longname': 'positionalContinuity',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'pc'},
                           'positionalContinuityType': {'args': '<type __builtin__.int>',
                                                        'longname': 'positionalContinuityType',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'pct'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'replaceOriginal': {'args': '<type __builtin__.bool>',
                                               'longname': 'replaceOriginal',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'rpo'},
                           'reverse1': {'args': '<type __builtin__.bool>',
                                        'longname': 'reverse1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'rv1'},
                           'reverse2': {'args': '<type __builtin__.bool>',
                                        'longname': 'reverse2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'rv2'},
                           'tangentContinuity': {'args': '<type __builtin__.bool>',
                                                 'longname': 'tangentContinuity',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'tc'},
                           'tangentContinuityType': {'args': '<type __builtin__.int>',
                                                     'longname': 'tangentContinuityType',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'tct'},
                           'tangentScale1': {'args': '<type __builtin__.float>',
                                             'longname': 'tangentScale1',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ts1'},
                           'tangentScale2': {'args': '<type __builtin__.float>',
                                             'longname': 'tangentScale2',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ts2'}},
                 'shortFlags': {'at': 'attach',
                                'cc': 'curvatureContinuity',
                                'cch': 'caching',
                                'ch': 'constructionHistory',
                                'cs1': 'curvatureScale1',
                                'cs2': 'curvatureScale2',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'jnp': 'joinParameter',
                                'kmk': 'keepMultipleKnots',
                                'n': 'name',
                                'nds': 'nodeState',
                                'o': 'object',
                                'pc': 'positionalContinuity',
                                'pct': 'positionalContinuityType',
                                'q': 'query',
                                'rpo': 'replaceOriginal',
                                'rv1': 'reverse1',
                                'rv2': 'reverse2',
                                'tc': 'tangentContinuity',
                                'tct': 'tangentContinuityType',
                                'ts1': 'tangentScale1',
                                'ts2': 'tangentScale2'},
                 'type': 'modeling'},
  'alignSurface': {'flags': {'attach': {'args': '<type __builtin__.bool>',
                                        'longname': 'attach',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'at'},
                             'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'curvatureContinuity': {'args': '<type __builtin__.bool>',
                                                     'longname': 'curvatureContinuity',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'cc'},
                             'curvatureScale1': {'args': '<type __builtin__.float>',
                                                 'longname': 'curvatureScale1',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cs1'},
                             'curvatureScale2': {'args': '<type __builtin__.float>',
                                                 'longname': 'curvatureScale2',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cs2'},
                             'directionU': {'args': '<type __builtin__.bool>',
                                            'longname': 'directionU',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'du'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'joinParameter': {'args': '<type __builtin__.float>',
                                               'longname': 'joinParameter',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'jnp'},
                             'keepMultipleKnots': {'args': '<type __builtin__.bool>',
                                                   'longname': 'keepMultipleKnots',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'kmk'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'object': {'args': '<type __builtin__.bool>',
                                        'longname': 'object',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                             'positionalContinuity': {'args': '<type __builtin__.bool>',
                                                      'longname': 'positionalContinuity',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'pc'},
                             'positionalContinuityType': {'args': '<type __builtin__.int>',
                                                          'longname': 'positionalContinuityType',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'pct'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                 'longname': 'replaceOriginal',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'rpo'},
                             'reverse1': {'args': '<type __builtin__.bool>',
                                          'longname': 'reverse1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'rv1'},
                             'reverse2': {'args': '<type __builtin__.bool>',
                                          'longname': 'reverse2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'rv2'},
                             'swap1': {'args': '<type __builtin__.bool>',
                                       'longname': 'swap1',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sw1'},
                             'swap2': {'args': '<type __builtin__.bool>',
                                       'longname': 'swap2',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sw2'},
                             'tangentContinuity': {'args': '<type __builtin__.bool>',
                                                   'longname': 'tangentContinuity',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'tc'},
                             'tangentContinuityType': {'args': '<type __builtin__.int>',
                                                       'longname': 'tangentContinuityType',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'tct'},
                             'tangentScale1': {'args': '<type __builtin__.float>',
                                               'longname': 'tangentScale1',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ts1'},
                             'tangentScale2': {'args': '<type __builtin__.float>',
                                               'longname': 'tangentScale2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ts2'},
                             'twist': {'args': '<type __builtin__.bool>',
                                       'longname': 'twist',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'tw'}},
                   'shortFlags': {'at': 'attach',
                                  'cc': 'curvatureContinuity',
                                  'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'cs1': 'curvatureScale1',
                                  'cs2': 'curvatureScale2',
                                  'du': 'directionU',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'jnp': 'joinParameter',
                                  'kmk': 'keepMultipleKnots',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'o': 'object',
                                  'pc': 'positionalContinuity',
                                  'pct': 'positionalContinuityType',
                                  'q': 'query',
                                  'rpo': 'replaceOriginal',
                                  'rv1': 'reverse1',
                                  'rv2': 'reverse2',
                                  'sw1': 'swap1',
                                  'sw2': 'swap2',
                                  'tc': 'tangentContinuity',
                                  'tct': 'tangentContinuityType',
                                  'ts1': 'tangentScale1',
                                  'ts2': 'tangentScale2',
                                  'tw': 'twist'},
                   'type': 'modeling'},
  'allNodeTypes': {'flags': {'includeAbstract': {'args': '<type __builtin__.bool>',
                                                 'longname': 'includeAbstract',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ia'}},
                   'shortFlags': {'ia': 'includeAbstract'},
                   'type': 'system'},
  'ambientLight': {'flags': {'ambientShade': {'args': '<type __builtin__.float>',
                                              'longname': 'ambientShade',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ambientShade'},
                             'discRadius': {'args': '<type __builtin__.float>',
                                            'longname': 'discRadius',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'drs'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exclusive': {'args': '<type __builtin__.bool>',
                                           'longname': 'exclusive',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'exc'},
                             'intensity': {'args': '<type __builtin__.float>',
                                           'longname': 'intensity',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'i'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'position': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'position',
                                          'modes': ['create', 'query'],
                                          'numArgs': 3,
                                          'shortname': 'pos'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'rgb': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'rgb',
                                     'modes': ['create', 'query'],
                                     'numArgs': 3,
                                     'shortname': 'rgb'},
                             'rotation': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'rotation',
                                          'modes': ['create', 'query'],
                                          'numArgs': 3,
                                          'shortname': 'rot'},
                             'shadowColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'shadowColor',
                                             'modes': ['create', 'query'],
                                             'numArgs': 3,
                                             'shortname': 'sc'},
                             'shadowDither': {'args': '<type __builtin__.float>',
                                              'longname': 'shadowDither',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sd'},
                             'shadowSamples': {'args': '<type __builtin__.int>',
                                               'longname': 'shadowSamples',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sh'},
                             'softShadow': {'args': '<type __builtin__.bool>',
                                            'longname': 'softShadow',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ss'},
                             'useRayTraceShadows': {'args': '<type __builtin__.bool>',
                                                    'longname': 'useRayTraceShadows',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'rs'}},
                   'removedFlags': {'as': 'ambientShade'},
                   'shortFlags': {'ambientShade': 'ambientShade',
                                  'drs': 'discRadius',
                                  'e': 'edit',
                                  'exc': 'exclusive',
                                  'i': 'intensity',
                                  'n': 'name',
                                  'pos': 'position',
                                  'q': 'query',
                                  'rgb': 'rgb',
                                  'rot': 'rotation',
                                  'rs': 'useRayTraceShadows',
                                  'sc': 'shadowColor',
                                  'sd': 'shadowDither',
                                  'sh': 'shadowSamples',
                                  'ss': 'softShadow'},
                   'type': 'rendering'},
  'angleBetween': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'euler': {'args': '<type __builtin__.bool>',
                                       'longname': 'euler',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'er'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'vector1': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'vector1',
                                         'modes': ['create'],
                                         'numArgs': 3,
                                         'shortname': 'v1'},
                             'vector1X': {'args': '<type __builtin__.float>',
                                          'longname': 'vector1X',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'v1x'},
                             'vector1Y': {'args': '<type __builtin__.float>',
                                          'longname': 'vector1Y',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'v1y'},
                             'vector1Z': {'args': '<type __builtin__.float>',
                                          'longname': 'vector1Z',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'v1z'},
                             'vector2': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'vector2',
                                         'modes': ['create'],
                                         'numArgs': 3,
                                         'shortname': 'v2'},
                             'vector2X': {'args': '<type __builtin__.float>',
                                          'longname': 'vector2X',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'v2x'},
                             'vector2Y': {'args': '<type __builtin__.float>',
                                          'longname': 'vector2Y',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'v2y'},
                             'vector2Z': {'args': '<type __builtin__.float>',
                                          'longname': 'vector2Z',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'v2z'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'er': 'euler',
                                  'fzn': 'frozen',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'v1': 'vector1',
                                  'v1x': 'vector1X',
                                  'v1y': 'vector1Y',
                                  'v1z': 'vector1Z',
                                  'v2': 'vector2',
                                  'v2x': 'vector2X',
                                  'v2y': 'vector2Y',
                                  'v2z': 'vector2Z'},
                   'type': 'modeling'},
  'animCurveEditor': {'callbackFlags': ['denormalizeCurvesCommand',
                                        'dcc',
                                        'menu',
                                        'm',
                                        'normalizeCurvesCommand',
                                        'ncc'],
                      'flags': {'areCurvesSelected': {'args': '<type __builtin__.bool>',
                                                      'longname': 'areCurvesSelected',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'acs'},
                                'autoFit': {'args': '<type __builtin__.str>',
                                            'longname': 'autoFit',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'af'},
                                'autoFitTime': {'args': '<type __builtin__.str>',
                                                'longname': 'autoFitTime',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'aft'},
                                'classicMode': {'args': '<type __builtin__.bool>',
                                                'longname': 'classicMode',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cm'},
                                'clipTime': {'args': '<type __builtin__.str>',
                                             'longname': 'clipTime',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ct'},
                                'constrainDrag': {'args': '<type __builtin__.int>',
                                                  'longname': 'constrainDrag',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cd'},
                                'control': {'args': '<type __builtin__.bool>',
                                            'longname': 'control',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ctl'},
                                'curvesShown': {'args': '<type __builtin__.bool>',
                                                'longname': 'curvesShown',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'cs'},
                                'curvesShownForceUpdate': {'args': '<type __builtin__.bool>',
                                                           'longname': 'curvesShownForceUpdate',
                                                           'modes': ['query'],
                                                           'numArgs': 0,
                                                           'shortname': 'csf'},
                                'defineTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'defineTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'dt'},
                                'denormalizeCurvesCommand': {'args': '<type __builtin__.str>',
                                                             'longname': 'denormalizeCurvesCommand',
                                                             'modes': ['create',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'dcc'},
                                'displayActiveKeyTangents': {'args': '<type __builtin__.str>',
                                                             'longname': 'displayActiveKeyTangents',
                                                             'modes': ['edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'dat'},
                                'displayActiveKeys': {'args': '<type __builtin__.str>',
                                                      'longname': 'displayActiveKeys',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dak'},
                                'displayInfinities': {'args': '<type __builtin__.str>',
                                                      'longname': 'displayInfinities',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'di'},
                                'displayKeys': {'args': '<type __builtin__.str>',
                                                'longname': 'displayKeys',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'dk'},
                                'displayNormalized': {'args': '<type __builtin__.bool>',
                                                      'longname': 'displayNormalized',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dn'},
                                'displayTangents': {'args': '<type __builtin__.str>',
                                                    'longname': 'displayTangents',
                                                    'modes': ['edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dtn'},
                                'displayValues': {'args': '<type __builtin__.str>',
                                                  'longname': 'displayValues',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dv'},
                                'docTag': {'args': '<type __builtin__.str>',
                                           'longname': 'docTag',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dtg'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'filter': {'args': '<type __builtin__.str>',
                                           'longname': 'filter',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'f'},
                                'forceMainConnection': {'args': '<type __builtin__.str>',
                                                        'longname': 'forceMainConnection',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'fmc'},
                                'highlightConnection': {'args': '<type __builtin__.str>',
                                                        'longname': 'highlightConnection',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'hlc'},
                                'keyingTime': {'args': '<type __builtin__.str>',
                                               'longname': 'keyingTime',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'kt'},
                                'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                       'longname': 'lockMainConnection',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'lck'},
                                'lookAt': {'args': '<type __builtin__.str>',
                                           'longname': 'lookAt',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'la'},
                                'mainListConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'mainListConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'mlc'},
                                'menu': {'args': 'script',
                                         'longname': 'menu',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                                'normalizeCurvesCommand': {'args': '<type __builtin__.str>',
                                                           'longname': 'normalizeCurvesCommand',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ncc'},
                                'outliner': {'args': '<type __builtin__.str>',
                                             'longname': 'outliner',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'o'},
                                'panel': {'args': '<type __builtin__.str>',
                                          'longname': 'panel',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'pnl'},
                                'parent': {'args': '<type __builtin__.str>',
                                           'longname': 'parent',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'p'},
                                'preSelectionHighlight': {'args': '<type __builtin__.bool>',
                                                          'longname': 'preSelectionHighlight',
                                                          'modes': ['query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'psh'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'renormalizeCurves': {'args': '<type __builtin__.bool>',
                                                      'longname': 'renormalizeCurves',
                                                      'modes': ['edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'rnc'},
                                'resultSamples': {'args': 'time',
                                                  'longname': 'resultSamples',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'rs'},
                                'resultScreenSamples': {'args': '<type __builtin__.int>',
                                                        'longname': 'resultScreenSamples',
                                                        'modes': ['query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'rss'},
                                'resultUpdate': {'args': '<type __builtin__.str>',
                                                 'longname': 'resultUpdate',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ru'},
                                'selectionConnection': {'args': '<type __builtin__.str>',
                                                        'longname': 'selectionConnection',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'slc'},
                                'showActiveCurveNames': {'args': '<type __builtin__.bool>',
                                                         'longname': 'showActiveCurveNames',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'acn'},
                                'showBufferCurves': {'args': '<type __builtin__.str>',
                                                     'longname': 'showBufferCurves',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sb'},
                                'showCurveNames': {'args': '<type __builtin__.bool>',
                                                   'longname': 'showCurveNames',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'scn'},
                                'showResults': {'args': '<type __builtin__.str>',
                                                'longname': 'showResults',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sr'},
                                'showUpstreamCurves': {'args': '<type __builtin__.bool>',
                                                       'longname': 'showUpstreamCurves',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'suc'},
                                'smoothness': {'args': '<type __builtin__.str>',
                                               'longname': 'smoothness',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 's'},
                                'snapTime': {'args': '<type __builtin__.str>',
                                             'longname': 'snapTime',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'st'},
                                'snapValue': {'args': '<type __builtin__.str>',
                                              'longname': 'snapValue',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sv'},
                                'stackedCurves': {'args': '<type __builtin__.bool>',
                                                  'longname': 'stackedCurves',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sc'},
                                'stackedCurvesMax': {'args': '<type __builtin__.float>',
                                                     'longname': 'stackedCurvesMax',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'scx'},
                                'stackedCurvesMin': {'args': '<type __builtin__.float>',
                                                     'longname': 'stackedCurvesMin',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'scm'},
                                'stackedCurvesSpace': {'args': '<type __builtin__.float>',
                                                       'longname': 'stackedCurvesSpace',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'scs'},
                                'stateString': {'args': '<type __builtin__.bool>',
                                                'longname': 'stateString',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'sts'},
                                'unParent': {'args': '<type __builtin__.bool>',
                                             'longname': 'unParent',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'up'},
                                'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'unlockMainConnection',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'ulk'},
                                'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'updateMainConnection',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'upd'},
                                'useTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'useTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ut'},
                                'valueLinesToggle': {'args': '<type __builtin__.str>',
                                                     'longname': 'valueLinesToggle',
                                                     'modes': ['edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vlt'},
                                'viewLeft': {'args': '<type __builtin__.float>',
                                             'longname': 'viewLeft',
                                             'numArgs': 1,
                                             'shortname': 'vl'},
                                'viewRight': {'args': '<type __builtin__.float>',
                                              'longname': 'viewRight',
                                              'numArgs': 1,
                                              'shortname': 'vr'}},
                      'shortFlags': {'acn': 'showActiveCurveNames',
                                     'acs': 'areCurvesSelected',
                                     'af': 'autoFit',
                                     'aft': 'autoFitTime',
                                     'cd': 'constrainDrag',
                                     'cm': 'classicMode',
                                     'cs': 'curvesShown',
                                     'csf': 'curvesShownForceUpdate',
                                     'ct': 'clipTime',
                                     'ctl': 'control',
                                     'dak': 'displayActiveKeys',
                                     'dat': 'displayActiveKeyTangents',
                                     'dcc': 'denormalizeCurvesCommand',
                                     'di': 'displayInfinities',
                                     'dk': 'displayKeys',
                                     'dn': 'displayNormalized',
                                     'dt': 'defineTemplate',
                                     'dtg': 'docTag',
                                     'dtn': 'displayTangents',
                                     'dv': 'displayValues',
                                     'e': 'edit',
                                     'ex': 'exists',
                                     'f': 'filter',
                                     'fmc': 'forceMainConnection',
                                     'hlc': 'highlightConnection',
                                     'kt': 'keyingTime',
                                     'la': 'lookAt',
                                     'lck': 'lockMainConnection',
                                     'm': 'menu',
                                     'mlc': 'mainListConnection',
                                     'ncc': 'normalizeCurvesCommand',
                                     'o': 'outliner',
                                     'p': 'parent',
                                     'pnl': 'panel',
                                     'psh': 'preSelectionHighlight',
                                     'q': 'query',
                                     'rnc': 'renormalizeCurves',
                                     'rs': 'resultSamples',
                                     'rss': 'resultScreenSamples',
                                     'ru': 'resultUpdate',
                                     's': 'smoothness',
                                     'sb': 'showBufferCurves',
                                     'sc': 'stackedCurves',
                                     'scm': 'stackedCurvesMin',
                                     'scn': 'showCurveNames',
                                     'scs': 'stackedCurvesSpace',
                                     'scx': 'stackedCurvesMax',
                                     'slc': 'selectionConnection',
                                     'sr': 'showResults',
                                     'st': 'snapTime',
                                     'sts': 'stateString',
                                     'suc': 'showUpstreamCurves',
                                     'sv': 'snapValue',
                                     'ulk': 'unlockMainConnection',
                                     'up': 'unParent',
                                     'upd': 'updateMainConnection',
                                     'ut': 'useTemplate',
                                     'vl': 'viewLeft',
                                     'vlt': 'valueLinesToggle',
                                     'vr': 'viewRight'},
                      'type': 'animation'},
  'animDisplay': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'modelUpdate': {'args': '<type __builtin__.str>',
                                            'longname': 'modelUpdate',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'upd'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'refAnimCurvesEditable': {'args': '<type __builtin__.bool>',
                                                      'longname': 'refAnimCurvesEditable',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'rae'},
                            'timeCode': {'args': '<type __builtin__.str>',
                                         'longname': 'timeCode',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'tc'},
                            'timeCodeOffset': {'args': '<type __builtin__.str>',
                                               'longname': 'timeCodeOffset',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tco'}},
                  'shortFlags': {'e': 'edit',
                                 'q': 'query',
                                 'rae': 'refAnimCurvesEditable',
                                 'tc': 'timeCode',
                                 'tco': 'timeCodeOffset',
                                 'upd': 'modelUpdate'},
                  'type': 'animation'},
  'animLayer': {'flags': {'addRelatedKG': {'args': '<type __builtin__.bool>',
                                           'longname': 'addRelatedKG',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'akg'},
                          'addSelectedObjects': {'args': '<type __builtin__.bool>',
                                                 'longname': 'addSelectedObjects',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'aso'},
                          'affectedLayers': {'args': '<type __builtin__.bool>',
                                             'longname': 'affectedLayers',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'afl'},
                          'animCurves': {'args': '<type __builtin__.bool>',
                                         'longname': 'animCurves',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 0,
                                         'shortname': 'anc'},
                          'attribute': {'args': '<type __builtin__.str>',
                                        'longname': 'attribute',
                                        'modes': ['create',
                                                  'query',
                                                  'edit',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'at'},
                          'baseAnimCurves': {'args': '<type __builtin__.bool>',
                                             'longname': 'baseAnimCurves',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 0,
                                             'shortname': 'bac'},
                          'bestAnimLayer': {'args': '<type __builtin__.bool>',
                                            'longname': 'bestAnimLayer',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 0,
                                            'shortname': 'blr'},
                          'bestLayer': {'args': '<type __builtin__.bool>',
                                        'longname': 'bestLayer',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'bl'},
                          'blendNodes': {'args': '<type __builtin__.bool>',
                                         'longname': 'blendNodes',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 0,
                                         'shortname': 'bld'},
                          'children': {'args': '<type __builtin__.str>',
                                       'longname': 'children',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'c'},
                          'collapse': {'args': '<type __builtin__.bool>',
                                       'longname': 'collapse',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'col'},
                          'copy': {'args': '<type __builtin__.str>',
                                   'longname': 'copy',
                                   'modes': ['edit'],
                                   'numArgs': 1,
                                   'shortname': 'cp'},
                          'copyAnimation': {'args': '<type __builtin__.str>',
                                            'longname': 'copyAnimation',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ca'},
                          'copyNoAnimation': {'args': '<type __builtin__.str>',
                                              'longname': 'copyNoAnimation',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'cna'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'excludeBoolean': {'args': '<type __builtin__.bool>',
                                             'longname': 'excludeBoolean',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 0,
                                             'shortname': 'ebl'},
                          'excludeDynamic': {'args': '<type __builtin__.bool>',
                                             'longname': 'excludeDynamic',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 0,
                                             'shortname': 'edn'},
                          'excludeEnum': {'args': '<type __builtin__.bool>',
                                          'longname': 'excludeEnum',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'een'},
                          'excludeRotate': {'args': '<type __builtin__.bool>',
                                            'longname': 'excludeRotate',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 0,
                                            'shortname': 'ert'},
                          'excludeScale': {'args': '<type __builtin__.bool>',
                                           'longname': 'excludeScale',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 0,
                                           'shortname': 'esc'},
                          'excludeTranslate': {'args': '<type __builtin__.bool>',
                                               'longname': 'excludeTranslate',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 0,
                                               'shortname': 'etr'},
                          'excludeVisibility': {'args': '<type __builtin__.bool>',
                                                'longname': 'excludeVisibility',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'evs'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['query'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'extractAnimation': {'args': '<type __builtin__.str>',
                                               'longname': 'extractAnimation',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ea'},
                          'findCurveForPlug': {'args': '<type __builtin__.str>',
                                               'longname': 'findCurveForPlug',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'fcv'},
                          'forceUIRebuild': {'args': '<type __builtin__.bool>',
                                             'longname': 'forceUIRebuild',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'fur'},
                          'forceUIRefresh': {'args': '<type __builtin__.bool>',
                                             'longname': 'forceUIRefresh',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'uir'},
                          'layeredPlug': {'args': '<type __builtin__.str>',
                                          'longname': 'layeredPlug',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'lp'},
                          'lock': {'args': '<type __builtin__.bool>',
                                   'longname': 'lock',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'l'},
                          'maxLayers': {'args': '<type __builtin__.bool>',
                                        'longname': 'maxLayers',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'ml'},
                          'moveLayerAfter': {'args': '<type __builtin__.str>',
                                             'longname': 'moveLayerAfter',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'mva'},
                          'moveLayerBefore': {'args': '<type __builtin__.str>',
                                              'longname': 'moveLayerBefore',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'mvb'},
                          'mute': {'args': '<type __builtin__.bool>',
                                   'longname': 'mute',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'm'},
                          'override': {'args': '<type __builtin__.bool>',
                                       'longname': 'override',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                          'passthrough': {'args': '<type __builtin__.bool>',
                                          'longname': 'passthrough',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pth'},
                          'preferred': {'args': '<type __builtin__.bool>',
                                        'longname': 'preferred',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'prf'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'removeAllAttributes': {'args': '<type __builtin__.bool>',
                                                  'longname': 'removeAllAttributes',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'raa'},
                          'removeAttribute': {'args': '<type __builtin__.str>',
                                              'longname': 'removeAttribute',
                                              'modes': ['edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'ra'},
                          'root': {'args': '<type __builtin__.str>',
                                   'longname': 'root',
                                   'modes': ['query'],
                                   'numArgs': 1,
                                   'shortname': 'r'},
                          'selected': {'args': '<type __builtin__.bool>',
                                       'longname': 'selected',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sel'},
                          'solo': {'args': '<type __builtin__.bool>',
                                   'longname': 'solo',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 's'},
                          'weight': {'args': '<type __builtin__.float>',
                                     'longname': 'weight',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'},
                          'writeBlendnodeDestinations': {'args': '<type __builtin__.bool>',
                                                         'longname': 'writeBlendnodeDestinations',
                                                         'modes': ['edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'wbd'}},
                'shortFlags': {'afl': 'affectedLayers',
                               'akg': 'addRelatedKG',
                               'anc': 'animCurves',
                               'aso': 'addSelectedObjects',
                               'at': 'attribute',
                               'bac': 'baseAnimCurves',
                               'bl': 'bestLayer',
                               'bld': 'blendNodes',
                               'blr': 'bestAnimLayer',
                               'c': 'children',
                               'ca': 'copyAnimation',
                               'cna': 'copyNoAnimation',
                               'col': 'collapse',
                               'cp': 'copy',
                               'e': 'edit',
                               'ea': 'extractAnimation',
                               'ebl': 'excludeBoolean',
                               'edn': 'excludeDynamic',
                               'een': 'excludeEnum',
                               'ert': 'excludeRotate',
                               'esc': 'excludeScale',
                               'etr': 'excludeTranslate',
                               'evs': 'excludeVisibility',
                               'ex': 'exists',
                               'fcv': 'findCurveForPlug',
                               'fur': 'forceUIRebuild',
                               'l': 'lock',
                               'lp': 'layeredPlug',
                               'm': 'mute',
                               'ml': 'maxLayers',
                               'mva': 'moveLayerAfter',
                               'mvb': 'moveLayerBefore',
                               'o': 'override',
                               'p': 'parent',
                               'prf': 'preferred',
                               'pth': 'passthrough',
                               'q': 'query',
                               'r': 'root',
                               'ra': 'removeAttribute',
                               'raa': 'removeAllAttributes',
                               's': 'solo',
                               'sel': 'selected',
                               'uir': 'forceUIRefresh',
                               'w': 'weight',
                               'wbd': 'writeBlendnodeDestinations'},
                'type': 'animation'},
  'animView': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'endTime': {'args': 'time',
                                     'longname': 'endTime',
                                     'modes': [],
                                     'numArgs': 1,
                                     'shortname': 'et'},
                         'maxValue': {'args': '<type __builtin__.float>',
                                      'longname': 'maxValue',
                                      'modes': [],
                                      'numArgs': 1,
                                      'shortname': 'max'},
                         'minValue': {'args': '<type __builtin__.float>',
                                      'longname': 'minValue',
                                      'modes': [],
                                      'numArgs': 1,
                                      'shortname': 'min'},
                         'nextView': {'args': '<type __builtin__.bool>',
                                      'longname': 'nextView',
                                      'modes': ['edit'],
                                      'numArgs': 0,
                                      'shortname': 'nv'},
                         'previousView': {'args': '<type __builtin__.bool>',
                                          'longname': 'previousView',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'pv'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'startTime': {'args': 'time',
                                       'longname': 'startTime',
                                       'modes': [],
                                       'numArgs': 1,
                                       'shortname': 'st'}},
               'shortFlags': {'e': 'edit',
                              'et': 'endTime',
                              'max': 'maxValue',
                              'min': 'minValue',
                              'nv': 'nextView',
                              'pv': 'previousView',
                              'q': 'query',
                              'st': 'startTime'},
               'type': 'animation'},
  'annotate': {'flags': {'point': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'point',
                                   'modes': ['create'],
                                   'numArgs': 3,
                                   'shortname': 'p'},
                         'text': {'args': '<type __builtin__.str>',
                                  'longname': 'text',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'tx'}},
               'shortFlags': {'p': 'point', 'tx': 'text'},
               'type': 'windows'},
  'applyAttrPattern': {'flags': {'nodeType': {'args': '<type __builtin__.str>',
                                              'longname': 'nodeType',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'nt'},
                                 'patternName': {'args': '<type __builtin__.str>',
                                                 'longname': 'patternName',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'pn'}},
                       'shortFlags': {'nt': 'nodeType', 'pn': 'patternName'},
                       'type': 'general'},
  'applyMetadata': {'flags': {'format': {'args': '<type __builtin__.str>',
                                         'longname': 'format',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'fmt'},
                              'scene': {'args': '<type __builtin__.bool>',
                                        'longname': 'scene',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'scn'},
                              'value': {'args': '<type __builtin__.str>',
                                        'longname': 'value',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'v'}},
                    'shortFlags': {'fmt': 'format',
                                   'scn': 'scene',
                                   'v': 'value'},
                    'type': 'modeling'},
  'applyTake': {'flags': {'channel': {'args': '<type __builtin__.str>',
                                      'longname': 'channel',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'c'},
                          'device': {'args': '<type __builtin__.str>',
                                     'longname': 'device',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'd'},
                          'filter': {'args': '<type __builtin__.str>',
                                     'longname': 'filter',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'f'},
                          'preview': {'args': '<type __builtin__.bool>',
                                      'longname': 'preview',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                          'recurseChannel': {'args': '<type __builtin__.bool>',
                                             'longname': 'recurseChannel',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'rc'},
                          'reset': {'args': '<type __builtin__.bool>',
                                    'longname': 'reset',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'r'},
                          'specifyChannel': {'args': '<type __builtin__.bool>',
                                             'longname': 'specifyChannel',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'sc'},
                          'startTime': {'args': 'time',
                                        'longname': 'startTime',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'st'}},
                'shortFlags': {},
                'type': 'animation'},
  'arcLenDimContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'history': {'args': '<type __builtin__.bool>',
                                             'longname': 'history',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ch'},
                                 'image1': {'args': '<type __builtin__.str>',
                                            'longname': 'image1',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i1'},
                                 'image2': {'args': '<type __builtin__.str>',
                                            'longname': 'image2',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i2'},
                                 'image3': {'args': '<type __builtin__.str>',
                                            'longname': 'image3',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i3'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'}},
                       'shortFlags': {'ch': 'history',
                                      'e': 'edit',
                                      'ex': 'exists',
                                      'i1': 'image1',
                                      'i2': 'image2',
                                      'i3': 'image3',
                                      'n': 'name',
                                      'q': 'query'},
                       'type': 'context'},
  'arcLengthDimension': {'flags': {}, 'shortFlags': {}, 'type': 'modeling'},
  'arclen': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                   'longname': 'caching',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'cch'},
                       'constructionHistory': {'args': '<type __builtin__.bool>',
                                               'longname': 'constructionHistory',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'frozen': {'args': '<type __builtin__.bool>',
                                  'longname': 'frozen',
                                  'numArgs': 1,
                                  'shortname': 'fzn'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'numArgs': 1,
                                'shortname': 'n'},
                       'nodeState': {'args': '<type __builtin__.int>',
                                     'longname': 'nodeState',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'nds'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'}},
             'shortFlags': {'cch': 'caching',
                            'ch': 'constructionHistory',
                            'e': 'edit',
                            'fzn': 'frozen',
                            'n': 'name',
                            'nds': 'nodeState',
                            'q': 'query'},
             'type': 'modeling'},
  'arrayMapper': {'flags': {'destAttr': {'args': '<type __builtin__.str>',
                                         'longname': 'destAttr',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'da'},
                            'inputU': {'args': '<type __builtin__.str>',
                                       'longname': 'inputU',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'iu'},
                            'inputV': {'args': '<type __builtin__.str>',
                                       'longname': 'inputV',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'iv'},
                            'mapTo': {'args': '<type __builtin__.str>',
                                      'longname': 'mapTo',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'mt'},
                            'target': {'args': '<type __builtin__.str>',
                                       'longname': 'target',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 't'},
                            'type': {'args': '<type __builtin__.str>',
                                     'longname': 'type',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'ty'}},
                  'resultNeedsUnpacking': True,
                  'shortFlags': {'da': 'destAttr',
                                 'iu': 'inputU',
                                 'iv': 'inputV',
                                 'mt': 'mapTo',
                                 't': 'target',
                                 'ty': 'type'},
                  'type': 'effects'},
  'art3dPaintCtx': {'flags': {'accopacity': {'args': '<type __builtin__.bool>',
                                             'longname': 'accopacity',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'aco'},
                              'afterStrokeCmd': {'args': '<type __builtin__.str>',
                                                 'longname': 'afterStrokeCmd',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'asc'},
                              'alphablendmode': {'args': '<type __builtin__.str>',
                                                 'longname': 'alphablendmode',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'abm'},
                              'alwaysKeepFile': {'args': '<type __builtin__.bool>',
                                                 'longname': 'alwaysKeepFile',
                                                 'numArgs': 1,
                                                 'shortname': 'akf'},
                              'assigntxt': {'args': '<type __builtin__.bool>',
                                            'longname': 'assigntxt',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'ast'},
                              'attrnames': {'args': '<type __builtin__.str>',
                                            'longname': 'attrnames',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'atn'},
                              'beforeStrokeCmd': {'args': '<type __builtin__.str>',
                                                  'longname': 'beforeStrokeCmd',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'bsc'},
                              'brushalignment': {'args': '<type __builtin__.bool>',
                                                 'longname': 'brushalignment',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'bra'},
                              'brushdepth': {'args': '<type __builtin__.float>',
                                             'longname': 'brushdepth',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'bd'},
                              'brushfeedback': {'args': '<type __builtin__.bool>',
                                                'longname': 'brushfeedback',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'brf'},
                              'brushtype': {'args': '<type __builtin__.str>',
                                            'longname': 'brushtype',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'brt'},
                              'clear': {'args': '<type __builtin__.bool>',
                                        'longname': 'clear',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'clr'},
                              'commonattr': {'args': '<type __builtin__.str>',
                                             'longname': 'commonattr',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'cat'},
                              'dragSlider': {'args': '<type __builtin__.str>',
                                             'longname': 'dragSlider',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dsl'},
                              'dynclonemode': {'args': '<type __builtin__.bool>',
                                               'longname': 'dynclonemode',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dcm'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'expandfilename': {'args': '<type __builtin__.bool>',
                                                 'longname': 'expandfilename',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'eef'},
                              'exportaspectratio': {'args': '<type __builtin__.float>',
                                                    'longname': 'exportaspectratio',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ear'},
                              'exportfilemode': {'args': '<type __builtin__.str>',
                                                 'longname': 'exportfilemode',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'efm'},
                              'exportfilesave': {'args': '<type __builtin__.str>',
                                                 'longname': 'exportfilesave',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'esf'},
                              'exportfilesizex': {'args': '<type __builtin__.int>',
                                                  'longname': 'exportfilesizex',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'fsx'},
                              'exportfilesizey': {'args': '<type __builtin__.int>',
                                                  'longname': 'exportfilesizey',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'fsy'},
                              'exportfiletype': {'args': '<type __builtin__.str>',
                                                 'longname': 'exportfiletype',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'eft'},
                              'extendFillColor': {'args': '<type __builtin__.bool>',
                                                  'longname': 'extendFillColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'efc'},
                              'fileformat': {'args': '<type __builtin__.str>',
                                             'longname': 'fileformat',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'eff'},
                              'filetxtaspectratio': {'args': '<type __builtin__.float>',
                                                     'longname': 'filetxtaspectratio',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'far'},
                              'filetxtsizex': {'args': '<type __builtin__.int>',
                                               'longname': 'filetxtsizex',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ftx'},
                              'filetxtsizey': {'args': '<type __builtin__.int>',
                                               'longname': 'filetxtsizey',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'fty'},
                              'floodOpacity': {'args': '<type __builtin__.float>',
                                               'longname': 'floodOpacity',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'fop'},
                              'floodall': {'args': '<type __builtin__.bool>',
                                           'longname': 'floodall',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 0,
                                           'shortname': 'fal'},
                              'floodselect': {'args': '<type __builtin__.bool>',
                                              'longname': 'floodselect',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 0,
                                              'shortname': 'fsl'},
                              'history': {'args': '<type __builtin__.bool>',
                                          'longname': 'history',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ch'},
                              'image1': {'args': '<type __builtin__.str>',
                                         'longname': 'image1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i1'},
                              'image2': {'args': '<type __builtin__.str>',
                                         'longname': 'image2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i2'},
                              'image3': {'args': '<type __builtin__.str>',
                                         'longname': 'image3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i3'},
                              'importfileload': {'args': '<type __builtin__.str>',
                                                 'longname': 'importfileload',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ifl'},
                              'importfilemode': {'args': '<type __builtin__.str>',
                                                 'longname': 'importfilemode',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ifm'},
                              'importreassign': {'args': '<type __builtin__.bool>',
                                                 'longname': 'importreassign',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'irm'},
                              'keepaspectratio': {'args': '<type __builtin__.bool>',
                                                  'longname': 'keepaspectratio',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'kar'},
                              'lastRecorderCmd': {'args': '<type __builtin__.str>',
                                                  'longname': 'lastRecorderCmd',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'lrc'},
                              'lastStampName': {'args': '<type __builtin__.str>',
                                                'longname': 'lastStampName',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'lsn'},
                              'lowerradius': {'args': '<type __builtin__.float>',
                                              'longname': 'lowerradius',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lr'},
                              'makeStroke': {'args': '<type __builtin__.int>',
                                             'longname': 'makeStroke',
                                             'modes': ['create',
                                                       'query',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'mst'},
                              'mappressure': {'args': '<type __builtin__.str>',
                                              'longname': 'mappressure',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'mp'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'opacity': {'args': '<type __builtin__.float>',
                                          'longname': 'opacity',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'op'},
                              'outline': {'args': '<type __builtin__.bool>',
                                          'longname': 'outline',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'o'},
                              'outwhilepaint': {'args': '<type __builtin__.bool>',
                                                'longname': 'outwhilepaint',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'owp'},
                              'paintmode': {'args': '<type __builtin__.str>',
                                            'longname': 'paintmode',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pm'},
                              'paintoperationtype': {'args': '<type __builtin__.str>',
                                                     'longname': 'paintoperationtype',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'pot'},
                              'painttxtattr': {'args': '<type __builtin__.str>',
                                               'longname': 'painttxtattr',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'pta'},
                              'painttxtattrname': {'args': '<type __builtin__.str>',
                                                   'longname': 'painttxtattrname',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ptn'},
                              'pfxScale': {'args': '<type __builtin__.float>',
                                           'longname': 'pfxScale',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'psc'},
                              'pfxWidth': {'args': '<type __builtin__.float>',
                                           'longname': 'pfxWidth',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pwd'},
                              'pickColor': {'args': '<type __builtin__.bool>',
                                            'longname': 'pickColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pcm'},
                              'pickValue': {'args': '<type __builtin__.bool>',
                                            'longname': 'pickValue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 0,
                                            'shortname': 'pv'},
                              'playbackCursor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'playbackCursor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit',
                                                           'multiuse'],
                                                 'numArgs': 2,
                                                 'shortname': 'plc'},
                              'playbackPressure': {'args': '<type __builtin__.float>',
                                                   'longname': 'playbackPressure',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'plp'},
                              'preserveclonesource': {'args': '<type __builtin__.bool>',
                                                      'longname': 'preserveclonesource',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'pcs'},
                              'pressureMapping1': {'args': '<type __builtin__.int>',
                                                   'longname': 'pressureMapping1',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'pm1'},
                              'pressureMapping2': {'args': '<type __builtin__.int>',
                                                   'longname': 'pressureMapping2',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'pm2'},
                              'pressureMapping3': {'args': '<type __builtin__.int>',
                                                   'longname': 'pressureMapping3',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'pm3'},
                              'pressureMax1': {'args': '<type __builtin__.float>',
                                               'longname': 'pressureMax1',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'px1'},
                              'pressureMax2': {'args': '<type __builtin__.float>',
                                               'longname': 'pressureMax2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'px2'},
                              'pressureMax3': {'args': '<type __builtin__.float>',
                                               'longname': 'pressureMax3',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'px3'},
                              'pressureMin1': {'args': '<type __builtin__.float>',
                                               'longname': 'pressureMin1',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ps1'},
                              'pressureMin2': {'args': '<type __builtin__.float>',
                                               'longname': 'pressureMin2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ps2'},
                              'pressureMin3': {'args': '<type __builtin__.float>',
                                               'longname': 'pressureMin3',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ps3'},
                              'profileShapeFile': {'args': '<type __builtin__.str>',
                                                   'longname': 'profileShapeFile',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'psf'},
                              'projective': {'args': '<type __builtin__.bool>',
                                             'longname': 'projective',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'prm'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'radius': {'args': '<type __builtin__.float>',
                                         'longname': 'radius',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'r'},
                              'record': {'args': '<type __builtin__.bool>',
                                         'longname': 'record',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'rec'},
                              'reflection': {'args': '<type __builtin__.bool>',
                                             'longname': 'reflection',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'rn'},
                              'reflectionaboutorigin': {'args': '<type __builtin__.bool>',
                                                        'longname': 'reflectionaboutorigin',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'rno'},
                              'reflectionaxis': {'args': '<type __builtin__.str>',
                                                 'longname': 'reflectionaxis',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ra'},
                              'reloadtexfile': {'args': '<type __builtin__.bool>',
                                                'longname': 'reloadtexfile',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'rtf'},
                              'resizeratio': {'args': '<type __builtin__.float>',
                                              'longname': 'resizeratio',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'rr'},
                              'resizetxt': {'args': '<type __builtin__.bool>',
                                            'longname': 'resizetxt',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'rft'},
                              'rgbcolor': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'rgbcolor',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'rgb'},
                              'rgbflood': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'rgbflood',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'fc'},
                              'saveTextureOnStroke': {'args': '<type __builtin__.bool>',
                                                      'longname': 'saveTextureOnStroke',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'sts'},
                              'saveonstroke': {'args': '<type __builtin__.bool>',
                                               'longname': 'saveonstroke',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sos'},
                              'savetexture': {'args': '<type __builtin__.bool>',
                                              'longname': 'savetexture',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'stx'},
                              'screenRadius': {'args': '<type __builtin__.float>',
                                               'longname': 'screenRadius',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'scR'},
                              'selectclonesource': {'args': '<type __builtin__.bool>',
                                                    'longname': 'selectclonesource',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'scs'},
                              'shadernames': {'args': '<type __builtin__.str>',
                                              'longname': 'shadernames',
                                              'modes': ['query'],
                                              'numArgs': 1,
                                              'shortname': 'hnm'},
                              'shapeattr': {'args': '<type __builtin__.bool>',
                                            'longname': 'shapeattr',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'spa'},
                              'shapenames': {'args': '<type __builtin__.str>',
                                             'longname': 'shapenames',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'shn'},
                              'showactive': {'args': '<type __builtin__.bool>',
                                             'longname': 'showactive',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sa'},
                              'soloAsDiffuse': {'args': '<type __builtin__.bool>',
                                                'longname': 'soloAsDiffuse',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sod'},
                              'stampDepth': {'args': '<type __builtin__.float>',
                                             'longname': 'stampDepth',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'stD'},
                              'stampProfile': {'args': '<type __builtin__.str>',
                                               'longname': 'stampProfile',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'stP'},
                              'stampSpacing': {'args': '<type __builtin__.float>',
                                               'longname': 'stampSpacing',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'stS'},
                              'strokesmooth': {'args': '<type __builtin__.str>',
                                               'longname': 'strokesmooth',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ssm'},
                              'surfaceConformedBrushVertices': {'args': '<type __builtin__.bool>',
                                                                'longname': 'surfaceConformedBrushVertices',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'scv'},
                              'tablet': {'args': '<type __builtin__.bool>',
                                         'longname': 'tablet',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'tab'},
                              'tangentOutline': {'args': '<type __builtin__.bool>',
                                                 'longname': 'tangentOutline',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'to'},
                              'textureFilenames': {'args': '<type __builtin__.bool>',
                                                   'longname': 'textureFilenames',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'tfn'},
                              'updateEraseTex': {'args': '<type __builtin__.bool>',
                                                 'longname': 'updateEraseTex',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'uet'},
                              'usepressure': {'args': '<type __builtin__.bool>',
                                              'longname': 'usepressure',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'up'},
                              'worldRadius': {'args': '<type __builtin__.float>',
                                              'longname': 'worldRadius',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'wlR'}},
                    'shortFlags': {'abm': 'alphablendmode',
                                   'aco': 'accopacity',
                                   'akf': 'alwaysKeepFile',
                                   'asc': 'afterStrokeCmd',
                                   'ast': 'assigntxt',
                                   'atn': 'attrnames',
                                   'bd': 'brushdepth',
                                   'bra': 'brushalignment',
                                   'brf': 'brushfeedback',
                                   'brt': 'brushtype',
                                   'bsc': 'beforeStrokeCmd',
                                   'cat': 'commonattr',
                                   'ch': 'history',
                                   'clr': 'clear',
                                   'dcm': 'dynclonemode',
                                   'dsl': 'dragSlider',
                                   'e': 'edit',
                                   'ear': 'exportaspectratio',
                                   'eef': 'expandfilename',
                                   'efc': 'extendFillColor',
                                   'eff': 'fileformat',
                                   'efm': 'exportfilemode',
                                   'eft': 'exportfiletype',
                                   'esf': 'exportfilesave',
                                   'ex': 'exists',
                                   'fal': 'floodall',
                                   'far': 'filetxtaspectratio',
                                   'fc': 'rgbflood',
                                   'fop': 'floodOpacity',
                                   'fsl': 'floodselect',
                                   'fsx': 'exportfilesizex',
                                   'fsy': 'exportfilesizey',
                                   'ftx': 'filetxtsizex',
                                   'fty': 'filetxtsizey',
                                   'hnm': 'shadernames',
                                   'i1': 'image1',
                                   'i2': 'image2',
                                   'i3': 'image3',
                                   'ifl': 'importfileload',
                                   'ifm': 'importfilemode',
                                   'irm': 'importreassign',
                                   'kar': 'keepaspectratio',
                                   'lr': 'lowerradius',
                                   'lrc': 'lastRecorderCmd',
                                   'lsn': 'lastStampName',
                                   'mp': 'mappressure',
                                   'mst': 'makeStroke',
                                   'n': 'name',
                                   'o': 'outline',
                                   'op': 'opacity',
                                   'owp': 'outwhilepaint',
                                   'pcm': 'pickColor',
                                   'pcs': 'preserveclonesource',
                                   'plc': 'playbackCursor',
                                   'plp': 'playbackPressure',
                                   'pm': 'paintmode',
                                   'pm1': 'pressureMapping1',
                                   'pm2': 'pressureMapping2',
                                   'pm3': 'pressureMapping3',
                                   'pot': 'paintoperationtype',
                                   'prm': 'projective',
                                   'ps1': 'pressureMin1',
                                   'ps2': 'pressureMin2',
                                   'ps3': 'pressureMin3',
                                   'psc': 'pfxScale',
                                   'psf': 'profileShapeFile',
                                   'pta': 'painttxtattr',
                                   'ptn': 'painttxtattrname',
                                   'pv': 'pickValue',
                                   'pwd': 'pfxWidth',
                                   'px1': 'pressureMax1',
                                   'px2': 'pressureMax2',
                                   'px3': 'pressureMax3',
                                   'q': 'query',
                                   'r': 'radius',
                                   'ra': 'reflectionaxis',
                                   'rec': 'record',
                                   'rft': 'resizetxt',
                                   'rgb': 'rgbcolor',
                                   'rn': 'reflection',
                                   'rno': 'reflectionaboutorigin',
                                   'rr': 'resizeratio',
                                   'rtf': 'reloadtexfile',
                                   'sa': 'showactive',
                                   'scR': 'screenRadius',
                                   'scs': 'selectclonesource',
                                   'scv': 'surfaceConformedBrushVertices',
                                   'shn': 'shapenames',
                                   'sod': 'soloAsDiffuse',
                                   'sos': 'saveonstroke',
                                   'spa': 'shapeattr',
                                   'ssm': 'strokesmooth',
                                   'stD': 'stampDepth',
                                   'stP': 'stampProfile',
                                   'stS': 'stampSpacing',
                                   'sts': 'saveTextureOnStroke',
                                   'stx': 'savetexture',
                                   'tab': 'tablet',
                                   'tfn': 'textureFilenames',
                                   'to': 'tangentOutline',
                                   'uet': 'updateEraseTex',
                                   'up': 'usepressure',
                                   'wlR': 'worldRadius'},
                    'type': 'context'},
  'artAttr': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'artAttrCtx': {'flags': {'accopacity': {'args': '<type __builtin__.bool>',
                                          'longname': 'accopacity',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'aco'},
                           'activeListChangedProc': {'args': '<type __builtin__.str>',
                                                     'longname': 'activeListChangedProc',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'alp'},
                           'afterStrokeCmd': {'args': '<type __builtin__.str>',
                                              'longname': 'afterStrokeCmd',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'asc'},
                           'alphaclamp': {'args': '<type __builtin__.str>',
                                          'longname': 'alphaclamp',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'alc'},
                           'alphaclamplower': {'args': '<type __builtin__.float>',
                                               'longname': 'alphaclamplower',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'acl'},
                           'alphaclampupper': {'args': '<type __builtin__.float>',
                                               'longname': 'alphaclampupper',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'acu'},
                           'attrSelected': {'args': '<type __builtin__.str>',
                                            'longname': 'attrSelected',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'asl'},
                           'beforeStrokeCmd': {'args': '<type __builtin__.str>',
                                               'longname': 'beforeStrokeCmd',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'bsc'},
                           'brushalignment': {'args': '<type __builtin__.bool>',
                                              'longname': 'brushalignment',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'bra'},
                           'brushfeedback': {'args': '<type __builtin__.bool>',
                                             'longname': 'brushfeedback',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'brf'},
                           'clamp': {'args': '<type __builtin__.str>',
                                     'longname': 'clamp',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'cl'},
                           'clamplower': {'args': '<type __builtin__.float>',
                                          'longname': 'clamplower',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cll'},
                           'clampupper': {'args': '<type __builtin__.float>',
                                          'longname': 'clampupper',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'clu'},
                           'clear': {'args': '<type __builtin__.bool>',
                                     'longname': 'clear',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'clr'},
                           'colorAlphaValue': {'args': '<type __builtin__.float>',
                                               'longname': 'colorAlphaValue',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cl1'},
                           'colorRGBAValue': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'colorRGBAValue',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 4,
                                              'shortname': 'cl4'},
                           'colorRGBValue': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'colorRGBValue',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'cl3'},
                           'colorRamp': {'args': '<type __builtin__.str>',
                                         'longname': 'colorRamp',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cr'},
                           'colorfeedback': {'args': '<type __builtin__.bool>',
                                             'longname': 'colorfeedback',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cf'},
                           'colorfeedbackOverride': {'args': '<type __builtin__.bool>',
                                                     'longname': 'colorfeedbackOverride',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'cfo'},
                           'colorrangelower': {'args': '<type __builtin__.float>',
                                               'longname': 'colorrangelower',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'crl'},
                           'colorrangeupper': {'args': '<type __builtin__.float>',
                                               'longname': 'colorrangeupper',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cru'},
                           'dataTypeIndex': {'args': '<type __builtin__.int>',
                                             'longname': 'dataTypeIndex',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dti'},
                           'disablelighting': {'args': '<type __builtin__.bool>',
                                               'longname': 'disablelighting',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dl'},
                           'dragSlider': {'args': '<type __builtin__.str>',
                                          'longname': 'dragSlider',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dsl'},
                           'duringStrokeCmd': {'args': '<type __builtin__.str>',
                                               'longname': 'duringStrokeCmd',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dsk'},
                           'dynclonemode': {'args': '<type __builtin__.bool>',
                                            'longname': 'dynclonemode',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dcm'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'expandfilename': {'args': '<type __builtin__.bool>',
                                              'longname': 'expandfilename',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'eef'},
                           'exportaspectratio': {'args': '<type __builtin__.float>',
                                                 'longname': 'exportaspectratio',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ear'},
                           'exportfilemode': {'args': '<type __builtin__.str>',
                                              'longname': 'exportfilemode',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'efm'},
                           'exportfilesave': {'args': '<type __builtin__.str>',
                                              'longname': 'exportfilesave',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'esf'},
                           'exportfilesizex': {'args': '<type __builtin__.int>',
                                               'longname': 'exportfilesizex',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'fsx'},
                           'exportfilesizey': {'args': '<type __builtin__.int>',
                                               'longname': 'exportfilesizey',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'fsy'},
                           'exportfiletype': {'args': '<type __builtin__.str>',
                                              'longname': 'exportfiletype',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'eft'},
                           'filterNodes': {'args': '<type __builtin__.bool>',
                                           'longname': 'filterNodes',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'fon'},
                           'history': {'args': '<type __builtin__.bool>',
                                       'longname': 'history',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ch'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'importfileload': {'args': '<type __builtin__.str>',
                                              'longname': 'importfileload',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'ifl'},
                           'importfilemode': {'args': '<type __builtin__.str>',
                                              'longname': 'importfilemode',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ifm'},
                           'importreassign': {'args': '<type __builtin__.bool>',
                                              'longname': 'importreassign',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'irm'},
                           'interactiveUpdate': {'args': '<type __builtin__.bool>',
                                                 'longname': 'interactiveUpdate',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'iu'},
                           'lastRecorderCmd': {'args': '<type __builtin__.str>',
                                               'longname': 'lastRecorderCmd',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'lrc'},
                           'lastStampName': {'args': '<type __builtin__.str>',
                                             'longname': 'lastStampName',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lsn'},
                           'lowerradius': {'args': '<type __builtin__.float>',
                                           'longname': 'lowerradius',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'lr'},
                           'makeStroke': {'args': '<type __builtin__.int>',
                                          'longname': 'makeStroke',
                                          'modes': ['create',
                                                    'query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'mst'},
                           'mappressure': {'args': '<type __builtin__.str>',
                                           'longname': 'mappressure',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'mp'},
                           'maxvalue': {'args': '<type __builtin__.float>',
                                        'longname': 'maxvalue',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'mxv'},
                           'minvalue': {'args': '<type __builtin__.float>',
                                        'longname': 'minvalue',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'miv'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'objattrArray': {'args': '<type __builtin__.str>',
                                            'longname': 'objattrArray',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'oaa'},
                           'opacity': {'args': '<type __builtin__.float>',
                                       'longname': 'opacity',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'op'},
                           'outline': {'args': '<type __builtin__.bool>',
                                       'longname': 'outline',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                           'outwhilepaint': {'args': '<type __builtin__.bool>',
                                             'longname': 'outwhilepaint',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'owp'},
                           'paintNodeArray': {'args': '<type __builtin__.str>',
                                              'longname': 'paintNodeArray',
                                              'modes': ['query'],
                                              'numArgs': 1,
                                              'shortname': 'pna'},
                           'paintattrselected': {'args': '<type __builtin__.str>',
                                                 'longname': 'paintattrselected',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'pas'},
                           'paintmode': {'args': '<type __builtin__.str>',
                                         'longname': 'paintmode',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'pm'},
                           'paintoperationtype': {'args': '<type __builtin__.str>',
                                                  'longname': 'paintoperationtype',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pot'},
                           'pickColor': {'args': '<type __builtin__.bool>',
                                         'longname': 'pickColor',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'pcm'},
                           'pickValue': {'args': '<type __builtin__.bool>',
                                         'longname': 'pickValue',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 0,
                                         'shortname': 'pv'},
                           'playbackCursor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'playbackCursor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'plc'},
                           'playbackPressure': {'args': '<type __builtin__.float>',
                                                'longname': 'playbackPressure',
                                                'modes': ['create',
                                                          'query',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'plp'},
                           'preserveclonesource': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preserveclonesource',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'pcs'},
                           'profileShapeFile': {'args': '<type __builtin__.str>',
                                                'longname': 'profileShapeFile',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'psf'},
                           'projective': {'args': '<type __builtin__.bool>',
                                          'longname': 'projective',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'prm'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'radius': {'args': '<type __builtin__.float>',
                                      'longname': 'radius',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'r'},
                           'rampMaxColor': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'rampMaxColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'rxc'},
                           'rampMinColor': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'rampMinColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'rmc'},
                           'record': {'args': '<type __builtin__.bool>',
                                      'longname': 'record',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'rec'},
                           'reflection': {'args': '<type __builtin__.bool>',
                                          'longname': 'reflection',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'rn'},
                           'reflectionaboutorigin': {'args': '<type __builtin__.bool>',
                                                     'longname': 'reflectionaboutorigin',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'rno'},
                           'reflectionaxis': {'args': '<type __builtin__.str>',
                                              'longname': 'reflectionaxis',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ra'},
                           'screenRadius': {'args': '<type __builtin__.float>',
                                            'longname': 'screenRadius',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'scR'},
                           'selectclonesource': {'args': '<type __builtin__.bool>',
                                                 'longname': 'selectclonesource',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'scs'},
                           'selectedattroper': {'args': '<type __builtin__.str>',
                                                'longname': 'selectedattroper',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sao'},
                           'showactive': {'args': '<type __builtin__.bool>',
                                          'longname': 'showactive',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sa'},
                           'stampDepth': {'args': '<type __builtin__.float>',
                                          'longname': 'stampDepth',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'stD'},
                           'stampProfile': {'args': '<type __builtin__.str>',
                                            'longname': 'stampProfile',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'stP'},
                           'stampSpacing': {'args': '<type __builtin__.float>',
                                            'longname': 'stampSpacing',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'stS'},
                           'strokesmooth': {'args': '<type __builtin__.str>',
                                            'longname': 'strokesmooth',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ssm'},
                           'surfaceConformedBrushVertices': {'args': '<type __builtin__.bool>',
                                                             'longname': 'surfaceConformedBrushVertices',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'scv'},
                           'tablet': {'args': '<type __builtin__.bool>',
                                      'longname': 'tablet',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'tab'},
                           'tangentOutline': {'args': '<type __builtin__.bool>',
                                              'longname': 'tangentOutline',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'to'},
                           'toolOffProc': {'args': '<type __builtin__.str>',
                                           'longname': 'toolOffProc',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'tfp'},
                           'toolOnProc': {'args': '<type __builtin__.str>',
                                          'longname': 'toolOnProc',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'top'},
                           'useColorRamp': {'args': '<type __builtin__.bool>',
                                            'longname': 'useColorRamp',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ucr'},
                           'useMaxMinColor': {'args': '<type __builtin__.bool>',
                                              'longname': 'useMaxMinColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'umc'},
                           'usepressure': {'args': '<type __builtin__.bool>',
                                           'longname': 'usepressure',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'up'},
                           'value': {'args': '<type __builtin__.float>',
                                     'longname': 'value',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'val'},
                           'whichTool': {'args': '<type __builtin__.str>',
                                         'longname': 'whichTool',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'wst'},
                           'worldRadius': {'args': '<type __builtin__.float>',
                                           'longname': 'worldRadius',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'wlR'}},
                 'shortFlags': {'acl': 'alphaclamplower',
                                'aco': 'accopacity',
                                'acu': 'alphaclampupper',
                                'alc': 'alphaclamp',
                                'alp': 'activeListChangedProc',
                                'asc': 'afterStrokeCmd',
                                'asl': 'attrSelected',
                                'bra': 'brushalignment',
                                'brf': 'brushfeedback',
                                'bsc': 'beforeStrokeCmd',
                                'cf': 'colorfeedback',
                                'cfo': 'colorfeedbackOverride',
                                'ch': 'history',
                                'cl': 'clamp',
                                'cl1': 'colorAlphaValue',
                                'cl3': 'colorRGBValue',
                                'cl4': 'colorRGBAValue',
                                'cll': 'clamplower',
                                'clr': 'clear',
                                'clu': 'clampupper',
                                'cr': 'colorRamp',
                                'crl': 'colorrangelower',
                                'cru': 'colorrangeupper',
                                'dcm': 'dynclonemode',
                                'dl': 'disablelighting',
                                'dsk': 'duringStrokeCmd',
                                'dsl': 'dragSlider',
                                'dti': 'dataTypeIndex',
                                'e': 'edit',
                                'ear': 'exportaspectratio',
                                'eef': 'expandfilename',
                                'efm': 'exportfilemode',
                                'eft': 'exportfiletype',
                                'esf': 'exportfilesave',
                                'ex': 'exists',
                                'fon': 'filterNodes',
                                'fsx': 'exportfilesizex',
                                'fsy': 'exportfilesizey',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'ifl': 'importfileload',
                                'ifm': 'importfilemode',
                                'irm': 'importreassign',
                                'iu': 'interactiveUpdate',
                                'lr': 'lowerradius',
                                'lrc': 'lastRecorderCmd',
                                'lsn': 'lastStampName',
                                'miv': 'minvalue',
                                'mp': 'mappressure',
                                'mst': 'makeStroke',
                                'mxv': 'maxvalue',
                                'n': 'name',
                                'o': 'outline',
                                'oaa': 'objattrArray',
                                'op': 'opacity',
                                'owp': 'outwhilepaint',
                                'pas': 'paintattrselected',
                                'pcm': 'pickColor',
                                'pcs': 'preserveclonesource',
                                'plc': 'playbackCursor',
                                'plp': 'playbackPressure',
                                'pm': 'paintmode',
                                'pna': 'paintNodeArray',
                                'pot': 'paintoperationtype',
                                'prm': 'projective',
                                'psf': 'profileShapeFile',
                                'pv': 'pickValue',
                                'q': 'query',
                                'r': 'radius',
                                'ra': 'reflectionaxis',
                                'rec': 'record',
                                'rmc': 'rampMinColor',
                                'rn': 'reflection',
                                'rno': 'reflectionaboutorigin',
                                'rxc': 'rampMaxColor',
                                'sa': 'showactive',
                                'sao': 'selectedattroper',
                                'scR': 'screenRadius',
                                'scs': 'selectclonesource',
                                'scv': 'surfaceConformedBrushVertices',
                                'ssm': 'strokesmooth',
                                'stD': 'stampDepth',
                                'stP': 'stampProfile',
                                'stS': 'stampSpacing',
                                'tab': 'tablet',
                                'tfp': 'toolOffProc',
                                'to': 'tangentOutline',
                                'top': 'toolOnProc',
                                'ucr': 'useColorRamp',
                                'umc': 'useMaxMinColor',
                                'up': 'usepressure',
                                'val': 'value',
                                'wlR': 'worldRadius',
                                'wst': 'whichTool'},
                 'type': 'context'},
  'artAttrPaintVertexCtx': {'flags': {'accopacity': {'args': '<type __builtin__.bool>',
                                                     'longname': 'accopacity',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'aco'},
                                      'activeListChangedProc': {'args': '<type __builtin__.str>',
                                                                'longname': 'activeListChangedProc',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'alp'},
                                      'afterStrokeCmd': {'args': '<type __builtin__.str>',
                                                         'longname': 'afterStrokeCmd',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'asc'},
                                      'alphaclamp': {'args': '<type __builtin__.str>',
                                                     'longname': 'alphaclamp',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'alc'},
                                      'alphaclamplower': {'args': '<type __builtin__.float>',
                                                          'longname': 'alphaclamplower',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'acl'},
                                      'alphaclampupper': {'args': '<type __builtin__.float>',
                                                          'longname': 'alphaclampupper',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'acu'},
                                      'attrSelected': {'args': '<type __builtin__.str>',
                                                       'longname': 'attrSelected',
                                                       'modes': ['query'],
                                                       'numArgs': 1,
                                                       'shortname': 'asl'},
                                      'beforeStrokeCmd': {'args': '<type __builtin__.str>',
                                                          'longname': 'beforeStrokeCmd',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'bsc'},
                                      'brushalignment': {'args': '<type __builtin__.bool>',
                                                         'longname': 'brushalignment',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'bra'},
                                      'brushfeedback': {'args': '<type __builtin__.bool>',
                                                        'longname': 'brushfeedback',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'brf'},
                                      'clamp': {'args': '<type __builtin__.str>',
                                                'longname': 'clamp',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cl'},
                                      'clamplower': {'args': '<type __builtin__.float>',
                                                     'longname': 'clamplower',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'cll'},
                                      'clampupper': {'args': '<type __builtin__.float>',
                                                     'longname': 'clampupper',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'clu'},
                                      'clear': {'args': '<type __builtin__.bool>',
                                                'longname': 'clear',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'clr'},
                                      'colorAlphaValue': {'args': '<type __builtin__.float>',
                                                          'longname': 'colorAlphaValue',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'cl1'},
                                      'colorRGBAValue': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'colorRGBAValue',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 4,
                                                         'shortname': 'cl4'},
                                      'colorRGBValue': {'args': ['<type __builtin__.float>',
                                                                 '<type __builtin__.float>',
                                                                 '<type __builtin__.float>'],
                                                        'longname': 'colorRGBValue',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 3,
                                                        'shortname': 'cl3'},
                                      'colorRamp': {'args': '<type __builtin__.str>',
                                                    'longname': 'colorRamp',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'cr'},
                                      'colorfeedback': {'args': '<type __builtin__.bool>',
                                                        'longname': 'colorfeedback',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'cf'},
                                      'colorfeedbackOverride': {'args': '<type __builtin__.bool>',
                                                                'longname': 'colorfeedbackOverride',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'cfo'},
                                      'colorrangelower': {'args': '<type __builtin__.float>',
                                                          'longname': 'colorrangelower',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'crl'},
                                      'colorrangeupper': {'args': '<type __builtin__.float>',
                                                          'longname': 'colorrangeupper',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'cru'},
                                      'dataTypeIndex': {'args': '<type __builtin__.int>',
                                                        'longname': 'dataTypeIndex',
                                                        'modes': ['query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'dti'},
                                      'disablelighting': {'args': '<type __builtin__.bool>',
                                                          'longname': 'disablelighting',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'dl'},
                                      'dragSlider': {'args': '<type __builtin__.str>',
                                                     'longname': 'dragSlider',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dsl'},
                                      'duringStrokeCmd': {'args': '<type __builtin__.str>',
                                                          'longname': 'duringStrokeCmd',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'dsk'},
                                      'dynclonemode': {'args': '<type __builtin__.bool>',
                                                       'longname': 'dynclonemode',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'dcm'},
                                      'edit': {'args': '<type __builtin__.bool>',
                                               'longname': 'edit',
                                               'numArgs': 0,
                                               'shortname': 'e'},
                                      'exists': {'args': '<type __builtin__.bool>',
                                                 'longname': 'exists',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ex'},
                                      'expandfilename': {'args': '<type __builtin__.bool>',
                                                         'longname': 'expandfilename',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'eef'},
                                      'exportaspectratio': {'args': '<type __builtin__.float>',
                                                            'longname': 'exportaspectratio',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ear'},
                                      'exportfilemode': {'args': '<type __builtin__.str>',
                                                         'longname': 'exportfilemode',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'efm'},
                                      'exportfilesave': {'args': '<type __builtin__.str>',
                                                         'longname': 'exportfilesave',
                                                         'modes': ['edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'esf'},
                                      'exportfilesizex': {'args': '<type __builtin__.int>',
                                                          'longname': 'exportfilesizex',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'fsx'},
                                      'exportfilesizey': {'args': '<type __builtin__.int>',
                                                          'longname': 'exportfilesizey',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'fsy'},
                                      'exportfiletype': {'args': '<type __builtin__.str>',
                                                         'longname': 'exportfiletype',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'eft'},
                                      'filterNodes': {'args': '<type __builtin__.bool>',
                                                      'longname': 'filterNodes',
                                                      'modes': ['edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'fon'},
                                      'history': {'args': '<type __builtin__.bool>',
                                                  'longname': 'history',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                                      'image1': {'args': '<type __builtin__.str>',
                                                 'longname': 'image1',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'i1'},
                                      'image2': {'args': '<type __builtin__.str>',
                                                 'longname': 'image2',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'i2'},
                                      'image3': {'args': '<type __builtin__.str>',
                                                 'longname': 'image3',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'i3'},
                                      'importfileload': {'args': '<type __builtin__.str>',
                                                         'longname': 'importfileload',
                                                         'modes': ['edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ifl'},
                                      'importfilemode': {'args': '<type __builtin__.str>',
                                                         'longname': 'importfilemode',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ifm'},
                                      'importreassign': {'args': '<type __builtin__.bool>',
                                                         'longname': 'importreassign',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'irm'},
                                      'interactiveUpdate': {'args': '<type __builtin__.bool>',
                                                            'longname': 'interactiveUpdate',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'iu'},
                                      'lastRecorderCmd': {'args': '<type __builtin__.str>',
                                                          'longname': 'lastRecorderCmd',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'lrc'},
                                      'lastStampName': {'args': '<type __builtin__.str>',
                                                        'longname': 'lastStampName',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'lsn'},
                                      'lowerradius': {'args': '<type __builtin__.float>',
                                                      'longname': 'lowerradius',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'lr'},
                                      'makeStroke': {'args': '<type __builtin__.int>',
                                                     'longname': 'makeStroke',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'mst'},
                                      'mappressure': {'args': '<type __builtin__.str>',
                                                      'longname': 'mappressure',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'mp'},
                                      'maxvalue': {'args': '<type __builtin__.float>',
                                                   'longname': 'maxvalue',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'mxv'},
                                      'minvalue': {'args': '<type __builtin__.float>',
                                                   'longname': 'minvalue',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'miv'},
                                      'name': {'args': '<type __builtin__.str>',
                                               'longname': 'name',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'n'},
                                      'objattrArray': {'args': '<type __builtin__.str>',
                                                       'longname': 'objattrArray',
                                                       'modes': ['query'],
                                                       'numArgs': 1,
                                                       'shortname': 'oaa'},
                                      'opacity': {'args': '<type __builtin__.float>',
                                                  'longname': 'opacity',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'op'},
                                      'outline': {'args': '<type __builtin__.bool>',
                                                  'longname': 'outline',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'o'},
                                      'outwhilepaint': {'args': '<type __builtin__.bool>',
                                                        'longname': 'outwhilepaint',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'owp'},
                                      'paintComponent': {'args': '<type __builtin__.int>',
                                                         'longname': 'paintComponent',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'pc'},
                                      'paintNodeArray': {'args': '<type __builtin__.str>',
                                                         'longname': 'paintNodeArray',
                                                         'modes': ['query'],
                                                         'numArgs': 1,
                                                         'shortname': 'pna'},
                                      'paintNumChannels': {'args': '<type __builtin__.int>',
                                                           'longname': 'paintNumChannels',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'pnc'},
                                      'paintRGBA': {'args': '<type __builtin__.bool>',
                                                    'longname': 'paintRGBA',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'pc4'},
                                      'paintVertexFace': {'args': '<type __builtin__.bool>',
                                                          'longname': 'paintVertexFace',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'pvf'},
                                      'paintattrselected': {'args': '<type __builtin__.str>',
                                                            'longname': 'paintattrselected',
                                                            'modes': ['edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'pas'},
                                      'paintmode': {'args': '<type __builtin__.str>',
                                                    'longname': 'paintmode',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'pm'},
                                      'paintoperationtype': {'args': '<type __builtin__.str>',
                                                             'longname': 'paintoperationtype',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'pot'},
                                      'pickColor': {'args': '<type __builtin__.bool>',
                                                    'longname': 'pickColor',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'pcm'},
                                      'pickValue': {'args': '<type __builtin__.bool>',
                                                    'longname': 'pickValue',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'pv'},
                                      'playbackCursor': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'playbackCursor',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit',
                                                                   'multiuse'],
                                                         'numArgs': 2,
                                                         'shortname': 'plc'},
                                      'playbackPressure': {'args': '<type __builtin__.float>',
                                                           'longname': 'playbackPressure',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit',
                                                                     'multiuse'],
                                                           'numArgs': 1,
                                                           'shortname': 'plp'},
                                      'preserveclonesource': {'args': '<type __builtin__.bool>',
                                                              'longname': 'preserveclonesource',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'pcs'},
                                      'profileShapeFile': {'args': '<type __builtin__.str>',
                                                           'longname': 'profileShapeFile',
                                                           'modes': ['query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'psf'},
                                      'projective': {'args': '<type __builtin__.bool>',
                                                     'longname': 'projective',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'prm'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'radius': {'args': '<type __builtin__.float>',
                                                 'longname': 'radius',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'r'},
                                      'rampMaxColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'rampMaxColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'rxc'},
                                      'rampMinColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'rampMinColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'rmc'},
                                      'record': {'args': '<type __builtin__.bool>',
                                                 'longname': 'record',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rec'},
                                      'reflection': {'args': '<type __builtin__.bool>',
                                                     'longname': 'reflection',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'rn'},
                                      'reflectionaboutorigin': {'args': '<type __builtin__.bool>',
                                                                'longname': 'reflectionaboutorigin',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'rno'},
                                      'reflectionaxis': {'args': '<type __builtin__.str>',
                                                         'longname': 'reflectionaxis',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ra'},
                                      'screenRadius': {'args': '<type __builtin__.float>',
                                                       'longname': 'screenRadius',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'scR'},
                                      'selectclonesource': {'args': '<type __builtin__.bool>',
                                                            'longname': 'selectclonesource',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'scs'},
                                      'selectedattroper': {'args': '<type __builtin__.str>',
                                                           'longname': 'selectedattroper',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'sao'},
                                      'showactive': {'args': '<type __builtin__.bool>',
                                                     'longname': 'showactive',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sa'},
                                      'stampDepth': {'args': '<type __builtin__.float>',
                                                     'longname': 'stampDepth',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'stD'},
                                      'stampProfile': {'args': '<type __builtin__.str>',
                                                       'longname': 'stampProfile',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'stP'},
                                      'stampSpacing': {'args': '<type __builtin__.float>',
                                                       'longname': 'stampSpacing',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'stS'},
                                      'strokesmooth': {'args': '<type __builtin__.str>',
                                                       'longname': 'strokesmooth',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ssm'},
                                      'surfaceConformedBrushVertices': {'args': '<type __builtin__.bool>',
                                                                        'longname': 'surfaceConformedBrushVertices',
                                                                        'modes': ['create',
                                                                                  'query',
                                                                                  'edit'],
                                                                        'numArgs': 1,
                                                                        'shortname': 'scv'},
                                      'tablet': {'args': '<type __builtin__.bool>',
                                                 'longname': 'tablet',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'tab'},
                                      'tangentOutline': {'args': '<type __builtin__.bool>',
                                                         'longname': 'tangentOutline',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'to'},
                                      'toolOffProc': {'args': '<type __builtin__.str>',
                                                      'longname': 'toolOffProc',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'tfp'},
                                      'toolOnProc': {'args': '<type __builtin__.str>',
                                                     'longname': 'toolOnProc',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'top'},
                                      'useColorRamp': {'args': '<type __builtin__.bool>',
                                                       'longname': 'useColorRamp',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ucr'},
                                      'useMaxMinColor': {'args': '<type __builtin__.bool>',
                                                         'longname': 'useMaxMinColor',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'umc'},
                                      'usepressure': {'args': '<type __builtin__.bool>',
                                                      'longname': 'usepressure',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'up'},
                                      'value': {'args': '<type __builtin__.float>',
                                                'longname': 'value',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'val'},
                                      'vertexColorRange': {'args': '<type __builtin__.bool>',
                                                           'longname': 'vertexColorRange',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'vcr'},
                                      'vertexColorRangeLower': {'args': '<type __builtin__.float>',
                                                                'longname': 'vertexColorRangeLower',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'vcl'},
                                      'vertexColorRangeUpper': {'args': '<type __builtin__.float>',
                                                                'longname': 'vertexColorRangeUpper',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'vcu'},
                                      'whichTool': {'args': '<type __builtin__.str>',
                                                    'longname': 'whichTool',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'wst'},
                                      'worldRadius': {'args': '<type __builtin__.float>',
                                                      'longname': 'worldRadius',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'wlR'}},
                            'shortFlags': {'acl': 'alphaclamplower',
                                           'aco': 'accopacity',
                                           'acu': 'alphaclampupper',
                                           'alc': 'alphaclamp',
                                           'alp': 'activeListChangedProc',
                                           'asc': 'afterStrokeCmd',
                                           'asl': 'attrSelected',
                                           'bra': 'brushalignment',
                                           'brf': 'brushfeedback',
                                           'bsc': 'beforeStrokeCmd',
                                           'cf': 'colorfeedback',
                                           'cfo': 'colorfeedbackOverride',
                                           'ch': 'history',
                                           'cl': 'clamp',
                                           'cl1': 'colorAlphaValue',
                                           'cl3': 'colorRGBValue',
                                           'cl4': 'colorRGBAValue',
                                           'cll': 'clamplower',
                                           'clr': 'clear',
                                           'clu': 'clampupper',
                                           'cr': 'colorRamp',
                                           'crl': 'colorrangelower',
                                           'cru': 'colorrangeupper',
                                           'dcm': 'dynclonemode',
                                           'dl': 'disablelighting',
                                           'dsk': 'duringStrokeCmd',
                                           'dsl': 'dragSlider',
                                           'dti': 'dataTypeIndex',
                                           'e': 'edit',
                                           'ear': 'exportaspectratio',
                                           'eef': 'expandfilename',
                                           'efm': 'exportfilemode',
                                           'eft': 'exportfiletype',
                                           'esf': 'exportfilesave',
                                           'ex': 'exists',
                                           'fon': 'filterNodes',
                                           'fsx': 'exportfilesizex',
                                           'fsy': 'exportfilesizey',
                                           'i1': 'image1',
                                           'i2': 'image2',
                                           'i3': 'image3',
                                           'ifl': 'importfileload',
                                           'ifm': 'importfilemode',
                                           'irm': 'importreassign',
                                           'iu': 'interactiveUpdate',
                                           'lr': 'lowerradius',
                                           'lrc': 'lastRecorderCmd',
                                           'lsn': 'lastStampName',
                                           'miv': 'minvalue',
                                           'mp': 'mappressure',
                                           'mst': 'makeStroke',
                                           'mxv': 'maxvalue',
                                           'n': 'name',
                                           'o': 'outline',
                                           'oaa': 'objattrArray',
                                           'op': 'opacity',
                                           'owp': 'outwhilepaint',
                                           'pas': 'paintattrselected',
                                           'pc': 'paintComponent',
                                           'pc4': 'paintRGBA',
                                           'pcm': 'pickColor',
                                           'pcs': 'preserveclonesource',
                                           'plc': 'playbackCursor',
                                           'plp': 'playbackPressure',
                                           'pm': 'paintmode',
                                           'pna': 'paintNodeArray',
                                           'pnc': 'paintNumChannels',
                                           'pot': 'paintoperationtype',
                                           'prm': 'projective',
                                           'psf': 'profileShapeFile',
                                           'pv': 'pickValue',
                                           'pvf': 'paintVertexFace',
                                           'q': 'query',
                                           'r': 'radius',
                                           'ra': 'reflectionaxis',
                                           'rec': 'record',
                                           'rmc': 'rampMinColor',
                                           'rn': 'reflection',
                                           'rno': 'reflectionaboutorigin',
                                           'rxc': 'rampMaxColor',
                                           'sa': 'showactive',
                                           'sao': 'selectedattroper',
                                           'scR': 'screenRadius',
                                           'scs': 'selectclonesource',
                                           'scv': 'surfaceConformedBrushVertices',
                                           'ssm': 'strokesmooth',
                                           'stD': 'stampDepth',
                                           'stP': 'stampProfile',
                                           'stS': 'stampSpacing',
                                           'tab': 'tablet',
                                           'tfp': 'toolOffProc',
                                           'to': 'tangentOutline',
                                           'top': 'toolOnProc',
                                           'ucr': 'useColorRamp',
                                           'umc': 'useMaxMinColor',
                                           'up': 'usepressure',
                                           'val': 'value',
                                           'vcl': 'vertexColorRangeLower',
                                           'vcr': 'vertexColorRange',
                                           'vcu': 'vertexColorRangeUpper',
                                           'wlR': 'worldRadius',
                                           'wst': 'whichTool'},
                            'type': 'context'},
  'artAttrSkinPaint': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'artAttrSkinPaintCmd': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'artAttrSkinPaintCtx': {'flags': {'accopacity': {'args': '<type __builtin__.bool>',
                                                   'longname': 'accopacity',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'aco'},
                                    'activeListChangedProc': {'args': '<type __builtin__.str>',
                                                              'longname': 'activeListChangedProc',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'alp'},
                                    'afterStrokeCmd': {'args': '<type __builtin__.str>',
                                                       'longname': 'afterStrokeCmd',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'asc'},
                                    'alphaclamp': {'args': '<type __builtin__.str>',
                                                   'longname': 'alphaclamp',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'alc'},
                                    'alphaclamplower': {'args': '<type __builtin__.float>',
                                                        'longname': 'alphaclamplower',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'acl'},
                                    'alphaclampupper': {'args': '<type __builtin__.float>',
                                                        'longname': 'alphaclampupper',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'acu'},
                                    'attrSelected': {'args': '<type __builtin__.str>',
                                                     'longname': 'attrSelected',
                                                     'modes': ['query'],
                                                     'numArgs': 1,
                                                     'shortname': 'asl'},
                                    'beforeStrokeCmd': {'args': '<type __builtin__.str>',
                                                        'longname': 'beforeStrokeCmd',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'bsc'},
                                    'brushalignment': {'args': '<type __builtin__.bool>',
                                                       'longname': 'brushalignment',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'bra'},
                                    'brushfeedback': {'args': '<type __builtin__.bool>',
                                                      'longname': 'brushfeedback',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'brf'},
                                    'clamp': {'args': '<type __builtin__.str>',
                                              'longname': 'clamp',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cl'},
                                    'clamplower': {'args': '<type __builtin__.float>',
                                                   'longname': 'clamplower',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cll'},
                                    'clampupper': {'args': '<type __builtin__.float>',
                                                   'longname': 'clampupper',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'clu'},
                                    'clear': {'args': '<type __builtin__.bool>',
                                              'longname': 'clear',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'clr'},
                                    'colorAlphaValue': {'args': '<type __builtin__.float>',
                                                        'longname': 'colorAlphaValue',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'cl1'},
                                    'colorRGBAValue': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'colorRGBAValue',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 4,
                                                       'shortname': 'cl4'},
                                    'colorRGBValue': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'colorRGBValue',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'cl3'},
                                    'colorRamp': {'args': '<type __builtin__.str>',
                                                  'longname': 'colorRamp',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cr'},
                                    'colorfeedback': {'args': '<type __builtin__.bool>',
                                                      'longname': 'colorfeedback',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cf'},
                                    'colorfeedbackOverride': {'args': '<type __builtin__.bool>',
                                                              'longname': 'colorfeedbackOverride',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'cfo'},
                                    'colorrangelower': {'args': '<type __builtin__.float>',
                                                        'longname': 'colorrangelower',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'crl'},
                                    'colorrangeupper': {'args': '<type __builtin__.float>',
                                                        'longname': 'colorrangeupper',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'cru'},
                                    'dataTypeIndex': {'args': '<type __builtin__.int>',
                                                      'longname': 'dataTypeIndex',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dti'},
                                    'disablelighting': {'args': '<type __builtin__.bool>',
                                                        'longname': 'disablelighting',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'dl'},
                                    'dragSlider': {'args': '<type __builtin__.str>',
                                                   'longname': 'dragSlider',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dsl'},
                                    'duringStrokeCmd': {'args': '<type __builtin__.str>',
                                                        'longname': 'duringStrokeCmd',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'dsk'},
                                    'dynclonemode': {'args': '<type __builtin__.bool>',
                                                     'longname': 'dynclonemode',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dcm'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'expandfilename': {'args': '<type __builtin__.bool>',
                                                       'longname': 'expandfilename',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'eef'},
                                    'exportaspectratio': {'args': '<type __builtin__.float>',
                                                          'longname': 'exportaspectratio',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ear'},
                                    'exportfilemode': {'args': '<type __builtin__.str>',
                                                       'longname': 'exportfilemode',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'efm'},
                                    'exportfilesave': {'args': '<type __builtin__.str>',
                                                       'longname': 'exportfilesave',
                                                       'modes': ['edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'esf'},
                                    'exportfilesizex': {'args': '<type __builtin__.int>',
                                                        'longname': 'exportfilesizex',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'fsx'},
                                    'exportfilesizey': {'args': '<type __builtin__.int>',
                                                        'longname': 'exportfilesizey',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'fsy'},
                                    'exportfiletype': {'args': '<type __builtin__.str>',
                                                       'longname': 'exportfiletype',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'eft'},
                                    'filterNodes': {'args': '<type __builtin__.bool>',
                                                    'longname': 'filterNodes',
                                                    'modes': ['edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'fon'},
                                    'history': {'args': '<type __builtin__.bool>',
                                                'longname': 'history',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'importfileload': {'args': '<type __builtin__.str>',
                                                       'longname': 'importfileload',
                                                       'modes': ['edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ifl'},
                                    'importfilemode': {'args': '<type __builtin__.str>',
                                                       'longname': 'importfilemode',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ifm'},
                                    'importreassign': {'args': '<type __builtin__.bool>',
                                                       'longname': 'importreassign',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'irm'},
                                    'influence': {'args': '<type __builtin__.str>',
                                                  'longname': 'influence',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'inf'},
                                    'interactiveUpdate': {'args': '<type __builtin__.bool>',
                                                          'longname': 'interactiveUpdate',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'iu'},
                                    'lastRecorderCmd': {'args': '<type __builtin__.str>',
                                                        'longname': 'lastRecorderCmd',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'lrc'},
                                    'lastStampName': {'args': '<type __builtin__.str>',
                                                      'longname': 'lastStampName',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'lsn'},
                                    'lowerradius': {'args': '<type __builtin__.float>',
                                                    'longname': 'lowerradius',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'lr'},
                                    'makeStroke': {'args': '<type __builtin__.int>',
                                                   'longname': 'makeStroke',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'mst'},
                                    'mappressure': {'args': '<type __builtin__.str>',
                                                    'longname': 'mappressure',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'mp'},
                                    'maxvalue': {'args': '<type __builtin__.float>',
                                                 'longname': 'maxvalue',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'mxv'},
                                    'minvalue': {'args': '<type __builtin__.float>',
                                                 'longname': 'minvalue',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'miv'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'objattrArray': {'args': '<type __builtin__.str>',
                                                     'longname': 'objattrArray',
                                                     'modes': ['query'],
                                                     'numArgs': 1,
                                                     'shortname': 'oaa'},
                                    'opacity': {'args': '<type __builtin__.float>',
                                                'longname': 'opacity',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'op'},
                                    'outline': {'args': '<type __builtin__.bool>',
                                                'longname': 'outline',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'o'},
                                    'outwhilepaint': {'args': '<type __builtin__.bool>',
                                                      'longname': 'outwhilepaint',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'owp'},
                                    'paintNodeArray': {'args': '<type __builtin__.str>',
                                                       'longname': 'paintNodeArray',
                                                       'modes': ['query'],
                                                       'numArgs': 1,
                                                       'shortname': 'pna'},
                                    'paintSelectMode': {'args': '<type __builtin__.int>',
                                                        'longname': 'paintSelectMode',
                                                        'modes': ['query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'psm'},
                                    'paintattrselected': {'args': '<type __builtin__.str>',
                                                          'longname': 'paintattrselected',
                                                          'modes': ['edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'pas'},
                                    'paintmode': {'args': '<type __builtin__.str>',
                                                  'longname': 'paintmode',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pm'},
                                    'paintoperationtype': {'args': '<type __builtin__.str>',
                                                           'longname': 'paintoperationtype',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'pot'},
                                    'pickColor': {'args': '<type __builtin__.bool>',
                                                  'longname': 'pickColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pcm'},
                                    'pickValue': {'args': '<type __builtin__.bool>',
                                                  'longname': 'pickValue',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'pv'},
                                    'playbackCursor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'playbackCursor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit',
                                                                 'multiuse'],
                                                       'numArgs': 2,
                                                       'shortname': 'plc'},
                                    'playbackPressure': {'args': '<type __builtin__.float>',
                                                         'longname': 'playbackPressure',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit',
                                                                   'multiuse'],
                                                         'numArgs': 1,
                                                         'shortname': 'plp'},
                                    'preserveclonesource': {'args': '<type __builtin__.bool>',
                                                            'longname': 'preserveclonesource',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'pcs'},
                                    'profileShapeFile': {'args': '<type __builtin__.str>',
                                                         'longname': 'profileShapeFile',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'psf'},
                                    'projective': {'args': '<type __builtin__.bool>',
                                                   'longname': 'projective',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'prm'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'radius': {'args': '<type __builtin__.float>',
                                               'longname': 'radius',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'r'},
                                    'rampMaxColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'rampMaxColor',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'rxc'},
                                    'rampMinColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'rampMinColor',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'rmc'},
                                    'record': {'args': '<type __builtin__.bool>',
                                               'longname': 'record',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rec'},
                                    'reflection': {'args': '<type __builtin__.bool>',
                                                   'longname': 'reflection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'rn'},
                                    'reflectionaboutorigin': {'args': '<type __builtin__.bool>',
                                                              'longname': 'reflectionaboutorigin',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'rno'},
                                    'reflectionaxis': {'args': '<type __builtin__.str>',
                                                       'longname': 'reflectionaxis',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ra'},
                                    'screenRadius': {'args': '<type __builtin__.float>',
                                                     'longname': 'screenRadius',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'scR'},
                                    'selectclonesource': {'args': '<type __builtin__.bool>',
                                                          'longname': 'selectclonesource',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'scs'},
                                    'selectedattroper': {'args': '<type __builtin__.str>',
                                                         'longname': 'selectedattroper',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'sao'},
                                    'showactive': {'args': '<type __builtin__.bool>',
                                                   'longname': 'showactive',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sa'},
                                    'skinPaintMode': {'args': '<type __builtin__.int>',
                                                      'longname': 'skinPaintMode',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'spm'},
                                    'stampDepth': {'args': '<type __builtin__.float>',
                                                   'longname': 'stampDepth',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'stD'},
                                    'stampProfile': {'args': '<type __builtin__.str>',
                                                     'longname': 'stampProfile',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'stP'},
                                    'stampSpacing': {'args': '<type __builtin__.float>',
                                                     'longname': 'stampSpacing',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'stS'},
                                    'strokesmooth': {'args': '<type __builtin__.str>',
                                                     'longname': 'strokesmooth',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ssm'},
                                    'surfaceConformedBrushVertices': {'args': '<type __builtin__.bool>',
                                                                      'longname': 'surfaceConformedBrushVertices',
                                                                      'modes': ['create',
                                                                                'query',
                                                                                'edit'],
                                                                      'numArgs': 1,
                                                                      'shortname': 'scv'},
                                    'tablet': {'args': '<type __builtin__.bool>',
                                               'longname': 'tablet',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'tab'},
                                    'tangentOutline': {'args': '<type __builtin__.bool>',
                                                       'longname': 'tangentOutline',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'to'},
                                    'toolOffProc': {'args': '<type __builtin__.str>',
                                                    'longname': 'toolOffProc',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'tfp'},
                                    'toolOnProc': {'args': '<type __builtin__.str>',
                                                   'longname': 'toolOnProc',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'top'},
                                    'useColorRamp': {'args': '<type __builtin__.bool>',
                                                     'longname': 'useColorRamp',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ucr'},
                                    'useMaxMinColor': {'args': '<type __builtin__.bool>',
                                                       'longname': 'useMaxMinColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'umc'},
                                    'usepressure': {'args': '<type __builtin__.bool>',
                                                    'longname': 'usepressure',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'up'},
                                    'value': {'args': '<type __builtin__.float>',
                                              'longname': 'value',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'val'},
                                    'whichTool': {'args': '<type __builtin__.str>',
                                                  'longname': 'whichTool',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'wst'},
                                    'worldRadius': {'args': '<type __builtin__.float>',
                                                    'longname': 'worldRadius',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'wlR'},
                                    'xrayJoints': {'args': '<type __builtin__.bool>',
                                                   'longname': 'xrayJoints',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'xry'}},
                          'shortFlags': {'acl': 'alphaclamplower',
                                         'aco': 'accopacity',
                                         'acu': 'alphaclampupper',
                                         'alc': 'alphaclamp',
                                         'alp': 'activeListChangedProc',
                                         'asc': 'afterStrokeCmd',
                                         'asl': 'attrSelected',
                                         'bra': 'brushalignment',
                                         'brf': 'brushfeedback',
                                         'bsc': 'beforeStrokeCmd',
                                         'cf': 'colorfeedback',
                                         'cfo': 'colorfeedbackOverride',
                                         'ch': 'history',
                                         'cl': 'clamp',
                                         'cl1': 'colorAlphaValue',
                                         'cl3': 'colorRGBValue',
                                         'cl4': 'colorRGBAValue',
                                         'cll': 'clamplower',
                                         'clr': 'clear',
                                         'clu': 'clampupper',
                                         'cr': 'colorRamp',
                                         'crl': 'colorrangelower',
                                         'cru': 'colorrangeupper',
                                         'dcm': 'dynclonemode',
                                         'dl': 'disablelighting',
                                         'dsk': 'duringStrokeCmd',
                                         'dsl': 'dragSlider',
                                         'dti': 'dataTypeIndex',
                                         'e': 'edit',
                                         'ear': 'exportaspectratio',
                                         'eef': 'expandfilename',
                                         'efm': 'exportfilemode',
                                         'eft': 'exportfiletype',
                                         'esf': 'exportfilesave',
                                         'ex': 'exists',
                                         'fon': 'filterNodes',
                                         'fsx': 'exportfilesizex',
                                         'fsy': 'exportfilesizey',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'ifl': 'importfileload',
                                         'ifm': 'importfilemode',
                                         'inf': 'influence',
                                         'irm': 'importreassign',
                                         'iu': 'interactiveUpdate',
                                         'lr': 'lowerradius',
                                         'lrc': 'lastRecorderCmd',
                                         'lsn': 'lastStampName',
                                         'miv': 'minvalue',
                                         'mp': 'mappressure',
                                         'mst': 'makeStroke',
                                         'mxv': 'maxvalue',
                                         'n': 'name',
                                         'o': 'outline',
                                         'oaa': 'objattrArray',
                                         'op': 'opacity',
                                         'owp': 'outwhilepaint',
                                         'pas': 'paintattrselected',
                                         'pcm': 'pickColor',
                                         'pcs': 'preserveclonesource',
                                         'plc': 'playbackCursor',
                                         'plp': 'playbackPressure',
                                         'pm': 'paintmode',
                                         'pna': 'paintNodeArray',
                                         'pot': 'paintoperationtype',
                                         'prm': 'projective',
                                         'psf': 'profileShapeFile',
                                         'psm': 'paintSelectMode',
                                         'pv': 'pickValue',
                                         'q': 'query',
                                         'r': 'radius',
                                         'ra': 'reflectionaxis',
                                         'rec': 'record',
                                         'rmc': 'rampMinColor',
                                         'rn': 'reflection',
                                         'rno': 'reflectionaboutorigin',
                                         'rxc': 'rampMaxColor',
                                         'sa': 'showactive',
                                         'sao': 'selectedattroper',
                                         'scR': 'screenRadius',
                                         'scs': 'selectclonesource',
                                         'scv': 'surfaceConformedBrushVertices',
                                         'spm': 'skinPaintMode',
                                         'ssm': 'strokesmooth',
                                         'stD': 'stampDepth',
                                         'stP': 'stampProfile',
                                         'stS': 'stampSpacing',
                                         'tab': 'tablet',
                                         'tfp': 'toolOffProc',
                                         'to': 'tangentOutline',
                                         'top': 'toolOnProc',
                                         'ucr': 'useColorRamp',
                                         'umc': 'useMaxMinColor',
                                         'up': 'usepressure',
                                         'val': 'value',
                                         'wlR': 'worldRadius',
                                         'wst': 'whichTool',
                                         'xry': 'xrayJoints'},
                          'type': 'context'},
  'artAttrTool': {'flags': {'add': {'args': 'Adds the named tool to the internal list of tools.',
                                    'longname': 'add',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'string'},
                            'exists': {'args': '<type __builtin__.str>',
                                       'longname': 'exists',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'ex'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'remove': {'args': '<type __builtin__.str>',
                                       'longname': 'remove',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'rm'}},
                  'shortFlags': {'add': 'add',
                                 'ex': 'exists',
                                 'q': 'query',
                                 'rm': 'remove'},
                  'type': 'general'},
  'artBaseCtx': {'flags': {'accopacity': {'args': '<type __builtin__.bool>',
                                          'longname': 'accopacity',
                                          'numArgs': 1,
                                          'shortname': 'aco'},
                           'afterStrokeCmd': {'args': '<type __builtin__.str>',
                                              'longname': 'afterStrokeCmd',
                                              'numArgs': 1,
                                              'shortname': 'asc'},
                           'beforeStrokeCmd': {'args': '<type __builtin__.str>',
                                               'longname': 'beforeStrokeCmd',
                                               'numArgs': 1,
                                               'shortname': 'bsc'},
                           'brushalignment': {'args': '<type __builtin__.bool>',
                                              'longname': 'brushalignment',
                                              'numArgs': 1,
                                              'shortname': 'bra'},
                           'brushfeedback': {'args': '<type __builtin__.bool>',
                                             'longname': 'brushfeedback',
                                             'numArgs': 1,
                                             'shortname': 'brf'},
                           'clear': {'args': '<type __builtin__.bool>',
                                     'longname': 'clear',
                                     'numArgs': 0,
                                     'shortname': 'clr'},
                           'dragSlider': {'args': '<type __builtin__.str>',
                                          'longname': 'dragSlider',
                                          'numArgs': 1,
                                          'shortname': 'dsl'},
                           'dynclonemode': {'args': '<type __builtin__.bool>',
                                            'longname': 'dynclonemode',
                                            'numArgs': 1,
                                            'shortname': 'dcm'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'expandfilename': {'args': '<type __builtin__.bool>',
                                              'longname': 'expandfilename',
                                              'numArgs': 1,
                                              'shortname': 'eef'},
                           'exportaspectratio': {'args': '<type __builtin__.float>',
                                                 'longname': 'exportaspectratio',
                                                 'numArgs': 1,
                                                 'shortname': 'ear'},
                           'exportfilemode': {'args': '<type __builtin__.str>',
                                              'longname': 'exportfilemode',
                                              'numArgs': 1,
                                              'shortname': 'efm'},
                           'exportfilesave': {'args': '<type __builtin__.str>',
                                              'longname': 'exportfilesave',
                                              'numArgs': 1,
                                              'shortname': 'esf'},
                           'exportfilesizex': {'args': '<type __builtin__.int>',
                                               'longname': 'exportfilesizex',
                                               'numArgs': 1,
                                               'shortname': 'fsx'},
                           'exportfilesizey': {'args': '<type __builtin__.int>',
                                               'longname': 'exportfilesizey',
                                               'numArgs': 1,
                                               'shortname': 'fsy'},
                           'exportfiletype': {'args': '<type __builtin__.str>',
                                              'longname': 'exportfiletype',
                                              'numArgs': 1,
                                              'shortname': 'eft'},
                           'history': {'args': '<type __builtin__.bool>',
                                       'longname': 'history',
                                       'numArgs': 1,
                                       'shortname': 'ch'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'importfileload': {'args': '<type __builtin__.str>',
                                              'longname': 'importfileload',
                                              'numArgs': 1,
                                              'shortname': 'ifl'},
                           'importfilemode': {'args': '<type __builtin__.str>',
                                              'longname': 'importfilemode',
                                              'numArgs': 1,
                                              'shortname': 'ifm'},
                           'importreassign': {'args': '<type __builtin__.bool>',
                                              'longname': 'importreassign',
                                              'numArgs': 1,
                                              'shortname': 'irm'},
                           'lastRecorderCmd': {'args': '<type __builtin__.str>',
                                               'longname': 'lastRecorderCmd',
                                               'numArgs': 1,
                                               'shortname': 'lrc'},
                           'lastStampName': {'args': '<type __builtin__.str>',
                                             'longname': 'lastStampName',
                                             'numArgs': 1,
                                             'shortname': 'lsn'},
                           'lowerradius': {'args': '<type __builtin__.float>',
                                           'longname': 'lowerradius',
                                           'numArgs': 1,
                                           'shortname': 'lr'},
                           'makeStroke': {'args': '<type __builtin__.int>',
                                          'longname': 'makeStroke',
                                          'modes': ['multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'mst'},
                           'mappressure': {'args': '<type __builtin__.str>',
                                           'longname': 'mappressure',
                                           'numArgs': 1,
                                           'shortname': 'mp'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'opacity': {'args': '<type __builtin__.float>',
                                       'longname': 'opacity',
                                       'numArgs': 1,
                                       'shortname': 'op'},
                           'outline': {'args': '<type __builtin__.bool>',
                                       'longname': 'outline',
                                       'numArgs': 1,
                                       'shortname': 'o'},
                           'outwhilepaint': {'args': '<type __builtin__.bool>',
                                             'longname': 'outwhilepaint',
                                             'numArgs': 1,
                                             'shortname': 'owp'},
                           'paintmode': {'args': '<type __builtin__.str>',
                                         'longname': 'paintmode',
                                         'numArgs': 1,
                                         'shortname': 'pm'},
                           'paintoperationtype': {'args': '<type __builtin__.str>',
                                                  'longname': 'paintoperationtype',
                                                  'numArgs': 1,
                                                  'shortname': 'pot'},
                           'pickColor': {'args': '<type __builtin__.bool>',
                                         'longname': 'pickColor',
                                         'numArgs': 1,
                                         'shortname': 'pcm'},
                           'pickValue': {'args': '<type __builtin__.bool>',
                                         'longname': 'pickValue',
                                         'numArgs': 0,
                                         'shortname': 'pv'},
                           'playbackCursor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'playbackCursor',
                                              'modes': ['multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'plc'},
                           'playbackPressure': {'args': '<type __builtin__.float>',
                                                'longname': 'playbackPressure',
                                                'modes': ['multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'plp'},
                           'preserveclonesource': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preserveclonesource',
                                                   'numArgs': 1,
                                                   'shortname': 'pcs'},
                           'profileShapeFile': {'args': '<type __builtin__.str>',
                                                'longname': 'profileShapeFile',
                                                'numArgs': 1,
                                                'shortname': 'psf'},
                           'projective': {'args': '<type __builtin__.bool>',
                                          'longname': 'projective',
                                          'numArgs': 1,
                                          'shortname': 'prm'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'radius': {'args': '<type __builtin__.float>',
                                      'longname': 'radius',
                                      'numArgs': 1,
                                      'shortname': 'r'},
                           'record': {'args': '<type __builtin__.bool>',
                                      'longname': 'record',
                                      'numArgs': 1,
                                      'shortname': 'rec'},
                           'reflection': {'args': '<type __builtin__.bool>',
                                          'longname': 'reflection',
                                          'numArgs': 1,
                                          'shortname': 'rn'},
                           'reflectionaboutorigin': {'args': '<type __builtin__.bool>',
                                                     'longname': 'reflectionaboutorigin',
                                                     'numArgs': 1,
                                                     'shortname': 'rno'},
                           'reflectionaxis': {'args': '<type __builtin__.str>',
                                              'longname': 'reflectionaxis',
                                              'numArgs': 1,
                                              'shortname': 'ra'},
                           'screenRadius': {'args': '<type __builtin__.float>',
                                            'longname': 'screenRadius',
                                            'numArgs': 1,
                                            'shortname': 'scR'},
                           'selectclonesource': {'args': '<type __builtin__.bool>',
                                                 'longname': 'selectclonesource',
                                                 'numArgs': 1,
                                                 'shortname': 'scs'},
                           'showactive': {'args': '<type __builtin__.bool>',
                                          'longname': 'showactive',
                                          'numArgs': 1,
                                          'shortname': 'sa'},
                           'stampDepth': {'args': '<type __builtin__.float>',
                                          'longname': 'stampDepth',
                                          'numArgs': 1,
                                          'shortname': 'stD'},
                           'stampProfile': {'args': '<type __builtin__.str>',
                                            'longname': 'stampProfile',
                                            'numArgs': 1,
                                            'shortname': 'stP'},
                           'stampSpacing': {'args': '<type __builtin__.float>',
                                            'longname': 'stampSpacing',
                                            'numArgs': 1,
                                            'shortname': 'stS'},
                           'strokesmooth': {'args': '<type __builtin__.str>',
                                            'longname': 'strokesmooth',
                                            'numArgs': 1,
                                            'shortname': 'ssm'},
                           'surfaceConformedBrushVertices': {'args': '<type __builtin__.bool>',
                                                             'longname': 'surfaceConformedBrushVertices',
                                                             'numArgs': 1,
                                                             'shortname': 'scv'},
                           'tablet': {'args': '<type __builtin__.bool>',
                                      'longname': 'tablet',
                                      'numArgs': 0,
                                      'shortname': 'tab'},
                           'tangentOutline': {'args': '<type __builtin__.bool>',
                                              'longname': 'tangentOutline',
                                              'numArgs': 1,
                                              'shortname': 'to'},
                           'usepressure': {'args': '<type __builtin__.bool>',
                                           'longname': 'usepressure',
                                           'numArgs': 1,
                                           'shortname': 'up'},
                           'worldRadius': {'args': '<type __builtin__.float>',
                                           'longname': 'worldRadius',
                                           'numArgs': 1,
                                           'shortname': 'wlR'}},
                 'shortFlags': {'aco': 'accopacity',
                                'asc': 'afterStrokeCmd',
                                'bra': 'brushalignment',
                                'brf': 'brushfeedback',
                                'bsc': 'beforeStrokeCmd',
                                'ch': 'history',
                                'clr': 'clear',
                                'dcm': 'dynclonemode',
                                'dsl': 'dragSlider',
                                'e': 'edit',
                                'ear': 'exportaspectratio',
                                'eef': 'expandfilename',
                                'efm': 'exportfilemode',
                                'eft': 'exportfiletype',
                                'esf': 'exportfilesave',
                                'ex': 'exists',
                                'fsx': 'exportfilesizex',
                                'fsy': 'exportfilesizey',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'ifl': 'importfileload',
                                'ifm': 'importfilemode',
                                'irm': 'importreassign',
                                'lr': 'lowerradius',
                                'lrc': 'lastRecorderCmd',
                                'lsn': 'lastStampName',
                                'mp': 'mappressure',
                                'mst': 'makeStroke',
                                'n': 'name',
                                'o': 'outline',
                                'op': 'opacity',
                                'owp': 'outwhilepaint',
                                'pcm': 'pickColor',
                                'pcs': 'preserveclonesource',
                                'plc': 'playbackCursor',
                                'plp': 'playbackPressure',
                                'pm': 'paintmode',
                                'pot': 'paintoperationtype',
                                'prm': 'projective',
                                'psf': 'profileShapeFile',
                                'pv': 'pickValue',
                                'q': 'query',
                                'r': 'radius',
                                'ra': 'reflectionaxis',
                                'rec': 'record',
                                'rn': 'reflection',
                                'rno': 'reflectionaboutorigin',
                                'sa': 'showactive',
                                'scR': 'screenRadius',
                                'scs': 'selectclonesource',
                                'scv': 'surfaceConformedBrushVertices',
                                'ssm': 'strokesmooth',
                                'stD': 'stampDepth',
                                'stP': 'stampProfile',
                                'stS': 'stampSpacing',
                                'tab': 'tablet',
                                'to': 'tangentOutline',
                                'up': 'usepressure',
                                'wlR': 'worldRadius'},
                 'type': 'context'},
  'artBuildPaintMenu': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'artFluidAttr': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'artFluidAttrCtx': {'flags': {'accopacity': {'args': '<type __builtin__.bool>',
                                               'longname': 'accopacity',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'aco'},
                                'activeListChangedProc': {'args': '<type __builtin__.str>',
                                                          'longname': 'activeListChangedProc',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'alp'},
                                'afterStrokeCmd': {'args': '<type __builtin__.str>',
                                                   'longname': 'afterStrokeCmd',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'asc'},
                                'alphaclamp': {'args': '<type __builtin__.str>',
                                               'longname': 'alphaclamp',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'alc'},
                                'alphaclamplower': {'args': '<type __builtin__.float>',
                                                    'longname': 'alphaclamplower',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'acl'},
                                'alphaclampupper': {'args': '<type __builtin__.float>',
                                                    'longname': 'alphaclampupper',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'acu'},
                                'attrSelected': {'args': '<type __builtin__.str>',
                                                 'longname': 'attrSelected',
                                                 'modes': ['query'],
                                                 'numArgs': 1,
                                                 'shortname': 'asl'},
                                'autoSave': {'args': '<type __builtin__.str>',
                                             'longname': 'autoSave',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'autoSave'},
                                'beforeStrokeCmd': {'args': '<type __builtin__.str>',
                                                    'longname': 'beforeStrokeCmd',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'bsc'},
                                'brushalignment': {'args': '<type __builtin__.bool>',
                                                   'longname': 'brushalignment',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'bra'},
                                'brushfeedback': {'args': '<type __builtin__.bool>',
                                                  'longname': 'brushfeedback',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'brf'},
                                'clamp': {'args': '<type __builtin__.str>',
                                          'longname': 'clamp',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cl'},
                                'clamplower': {'args': '<type __builtin__.float>',
                                               'longname': 'clamplower',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cll'},
                                'clampupper': {'args': '<type __builtin__.float>',
                                               'longname': 'clampupper',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'clu'},
                                'clear': {'args': '<type __builtin__.bool>',
                                          'longname': 'clear',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'clr'},
                                'colorAlphaValue': {'args': '<type __builtin__.float>',
                                                    'longname': 'colorAlphaValue',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'cl1'},
                                'colorRGBAValue': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'colorRGBAValue',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 4,
                                                   'shortname': 'cl4'},
                                'colorRGBValue': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'colorRGBValue',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'cl3'},
                                'colorRamp': {'args': '<type __builtin__.str>',
                                              'longname': 'colorRamp',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cr'},
                                'colorfeedback': {'args': '<type __builtin__.bool>',
                                                  'longname': 'colorfeedback',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cf'},
                                'colorfeedbackOverride': {'args': '<type __builtin__.bool>',
                                                          'longname': 'colorfeedbackOverride',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'cfo'},
                                'colorrangelower': {'args': '<type __builtin__.float>',
                                                    'longname': 'colorrangelower',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'crl'},
                                'colorrangeupper': {'args': '<type __builtin__.float>',
                                                    'longname': 'colorrangeupper',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'cru'},
                                'currentPaintableFluid': {'args': '<type __builtin__.str>',
                                                          'longname': 'currentPaintableFluid',
                                                          'modes': ['query'],
                                                          'numArgs': 1,
                                                          'shortname': 'cpf'},
                                'dataTypeIndex': {'args': '<type __builtin__.int>',
                                                  'longname': 'dataTypeIndex',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dti'},
                                'delaySelectionChanged': {'args': '<type __builtin__.bool>',
                                                          'longname': 'delaySelectionChanged',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'dsc'},
                                'disablelighting': {'args': '<type __builtin__.bool>',
                                                    'longname': 'disablelighting',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dl'},
                                'displayAsRender': {'args': '<type __builtin__.bool>',
                                                    'longname': 'displayAsRender',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dar'},
                                'displayVelocity': {'args': '<type __builtin__.bool>',
                                                    'longname': 'displayVelocity',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dv'},
                                'doAutoSave': {'args': '<type __builtin__.bool>',
                                               'longname': 'doAutoSave',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'das'},
                                'dragSlider': {'args': '<type __builtin__.str>',
                                               'longname': 'dragSlider',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dsl'},
                                'duringStrokeCmd': {'args': '<type __builtin__.str>',
                                                    'longname': 'duringStrokeCmd',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dsk'},
                                'dynclonemode': {'args': '<type __builtin__.bool>',
                                                 'longname': 'dynclonemode',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dcm'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'expandfilename': {'args': '<type __builtin__.bool>',
                                                   'longname': 'expandfilename',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'eef'},
                                'exportaspectratio': {'args': '<type __builtin__.float>',
                                                      'longname': 'exportaspectratio',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ear'},
                                'exportfilemode': {'args': '<type __builtin__.str>',
                                                   'longname': 'exportfilemode',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'efm'},
                                'exportfilesave': {'args': '<type __builtin__.str>',
                                                   'longname': 'exportfilesave',
                                                   'modes': ['edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'esf'},
                                'exportfilesizex': {'args': '<type __builtin__.int>',
                                                    'longname': 'exportfilesizex',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'fsx'},
                                'exportfilesizey': {'args': '<type __builtin__.int>',
                                                    'longname': 'exportfilesizey',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'fsy'},
                                'exportfiletype': {'args': '<type __builtin__.str>',
                                                   'longname': 'exportfiletype',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'eft'},
                                'filterNodes': {'args': '<type __builtin__.bool>',
                                                'longname': 'filterNodes',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'fon'},
                                'history': {'args': '<type __builtin__.bool>',
                                            'longname': 'history',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ch'},
                                'image1': {'args': '<type __builtin__.str>',
                                           'longname': 'image1',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i1'},
                                'image2': {'args': '<type __builtin__.str>',
                                           'longname': 'image2',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i2'},
                                'image3': {'args': '<type __builtin__.str>',
                                           'longname': 'image3',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i3'},
                                'importfileload': {'args': '<type __builtin__.str>',
                                                   'longname': 'importfileload',
                                                   'modes': ['edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ifl'},
                                'importfilemode': {'args': '<type __builtin__.str>',
                                                   'longname': 'importfilemode',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ifm'},
                                'importreassign': {'args': '<type __builtin__.bool>',
                                                   'longname': 'importreassign',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'irm'},
                                'interactiveUpdate': {'args': '<type __builtin__.bool>',
                                                      'longname': 'interactiveUpdate',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'iu'},
                                'lastRecorderCmd': {'args': '<type __builtin__.str>',
                                                    'longname': 'lastRecorderCmd',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'lrc'},
                                'lastStampName': {'args': '<type __builtin__.str>',
                                                  'longname': 'lastStampName',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'lsn'},
                                'lowerradius': {'args': '<type __builtin__.float>',
                                                'longname': 'lowerradius',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'lr'},
                                'makeStroke': {'args': '<type __builtin__.int>',
                                               'longname': 'makeStroke',
                                               'modes': ['create',
                                                         'query',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'mst'},
                                'mappressure': {'args': '<type __builtin__.str>',
                                                'longname': 'mappressure',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'mp'},
                                'maxvalue': {'args': '<type __builtin__.float>',
                                             'longname': 'maxvalue',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'mxv'},
                                'minvalue': {'args': '<type __builtin__.float>',
                                             'longname': 'minvalue',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'miv'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'objattrArray': {'args': '<type __builtin__.str>',
                                                 'longname': 'objattrArray',
                                                 'modes': ['query'],
                                                 'numArgs': 1,
                                                 'shortname': 'oaa'},
                                'opacity': {'args': '<type __builtin__.float>',
                                            'longname': 'opacity',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'op'},
                                'outline': {'args': '<type __builtin__.bool>',
                                            'longname': 'outline',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'o'},
                                'outwhilepaint': {'args': '<type __builtin__.bool>',
                                                  'longname': 'outwhilepaint',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'owp'},
                                'paintNodeArray': {'args': '<type __builtin__.str>',
                                                   'longname': 'paintNodeArray',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'pna'},
                                'paintattrselected': {'args': '<type __builtin__.str>',
                                                      'longname': 'paintattrselected',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'pas'},
                                'paintmode': {'args': '<type __builtin__.str>',
                                              'longname': 'paintmode',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pm'},
                                'paintoperationtype': {'args': '<type __builtin__.str>',
                                                       'longname': 'paintoperationtype',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'pot'},
                                'pickColor': {'args': '<type __builtin__.bool>',
                                              'longname': 'pickColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pcm'},
                                'pickValue': {'args': '<type __builtin__.bool>',
                                              'longname': 'pickValue',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 0,
                                              'shortname': 'pv'},
                                'playbackCursor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'playbackCursor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 2,
                                                   'shortname': 'plc'},
                                'playbackPressure': {'args': '<type __builtin__.float>',
                                                     'longname': 'playbackPressure',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'plp'},
                                'preserveclonesource': {'args': '<type __builtin__.bool>',
                                                        'longname': 'preserveclonesource',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'pcs'},
                                'profileShapeFile': {'args': '<type __builtin__.str>',
                                                     'longname': 'profileShapeFile',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'psf'},
                                'projective': {'args': '<type __builtin__.bool>',
                                               'longname': 'projective',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'prm'},
                                'property': {'args': '<type __builtin__.str>',
                                             'longname': 'property',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'p'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'radius': {'args': '<type __builtin__.float>',
                                           'longname': 'radius',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'r'},
                                'rampMaxColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'rampMaxColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'rxc'},
                                'rampMinColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'rampMinColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'rmc'},
                                'record': {'args': '<type __builtin__.bool>',
                                           'longname': 'record',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'rec'},
                                'reflection': {'args': '<type __builtin__.bool>',
                                               'longname': 'reflection',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rn'},
                                'reflectionaboutorigin': {'args': '<type __builtin__.bool>',
                                                          'longname': 'reflectionaboutorigin',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'rno'},
                                'reflectionaxis': {'args': '<type __builtin__.str>',
                                                   'longname': 'reflectionaxis',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ra'},
                                'rgbValue': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'rgbValue',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'rgb'},
                                'screenRadius': {'args': '<type __builtin__.float>',
                                                 'longname': 'screenRadius',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'scR'},
                                'selectclonesource': {'args': '<type __builtin__.bool>',
                                                      'longname': 'selectclonesource',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'scs'},
                                'selectedattroper': {'args': '<type __builtin__.str>',
                                                     'longname': 'selectedattroper',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sao'},
                                'showactive': {'args': '<type __builtin__.bool>',
                                               'longname': 'showactive',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sa'},
                                'stampDepth': {'args': '<type __builtin__.float>',
                                               'longname': 'stampDepth',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'stD'},
                                'stampProfile': {'args': '<type __builtin__.str>',
                                                 'longname': 'stampProfile',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'stP'},
                                'stampSpacing': {'args': '<type __builtin__.float>',
                                                 'longname': 'stampSpacing',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'stS'},
                                'strokesmooth': {'args': '<type __builtin__.str>',
                                                 'longname': 'strokesmooth',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ssm'},
                                'surfaceConformedBrushVertices': {'args': '<type __builtin__.bool>',
                                                                  'longname': 'surfaceConformedBrushVertices',
                                                                  'modes': ['create',
                                                                            'query',
                                                                            'edit'],
                                                                  'numArgs': 1,
                                                                  'shortname': 'scv'},
                                'tablet': {'args': '<type __builtin__.bool>',
                                           'longname': 'tablet',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'tab'},
                                'tangentOutline': {'args': '<type __builtin__.bool>',
                                                   'longname': 'tangentOutline',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'to'},
                                'toolOffProc': {'args': '<type __builtin__.str>',
                                                'longname': 'toolOffProc',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'tfp'},
                                'toolOnProc': {'args': '<type __builtin__.str>',
                                               'longname': 'toolOnProc',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'top'},
                                'useColorRamp': {'args': '<type __builtin__.bool>',
                                                 'longname': 'useColorRamp',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ucr'},
                                'useMaxMinColor': {'args': '<type __builtin__.bool>',
                                                   'longname': 'useMaxMinColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'umc'},
                                'useStrokeDirection': {'args': '<type __builtin__.bool>',
                                                       'longname': 'useStrokeDirection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'usd'},
                                'usepressure': {'args': '<type __builtin__.bool>',
                                                'longname': 'usepressure',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'up'},
                                'value': {'args': '<type __builtin__.float>',
                                          'longname': 'value',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'val'},
                                'velocity': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'velocity',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'v'},
                                'whichTool': {'args': '<type __builtin__.str>',
                                              'longname': 'whichTool',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'wst'},
                                'worldRadius': {'args': '<type __builtin__.float>',
                                                'longname': 'worldRadius',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'wlR'}},
                      'removedFlags': {'as': 'autoSave'},
                      'shortFlags': {'acl': 'alphaclamplower',
                                     'aco': 'accopacity',
                                     'acu': 'alphaclampupper',
                                     'alc': 'alphaclamp',
                                     'alp': 'activeListChangedProc',
                                     'asc': 'afterStrokeCmd',
                                     'asl': 'attrSelected',
                                     'autoSave': 'autoSave',
                                     'bra': 'brushalignment',
                                     'brf': 'brushfeedback',
                                     'bsc': 'beforeStrokeCmd',
                                     'cf': 'colorfeedback',
                                     'cfo': 'colorfeedbackOverride',
                                     'ch': 'history',
                                     'cl': 'clamp',
                                     'cl1': 'colorAlphaValue',
                                     'cl3': 'colorRGBValue',
                                     'cl4': 'colorRGBAValue',
                                     'cll': 'clamplower',
                                     'clr': 'clear',
                                     'clu': 'clampupper',
                                     'cpf': 'currentPaintableFluid',
                                     'cr': 'colorRamp',
                                     'crl': 'colorrangelower',
                                     'cru': 'colorrangeupper',
                                     'dar': 'displayAsRender',
                                     'das': 'doAutoSave',
                                     'dcm': 'dynclonemode',
                                     'dl': 'disablelighting',
                                     'dsc': 'delaySelectionChanged',
                                     'dsk': 'duringStrokeCmd',
                                     'dsl': 'dragSlider',
                                     'dti': 'dataTypeIndex',
                                     'dv': 'displayVelocity',
                                     'e': 'edit',
                                     'ear': 'exportaspectratio',
                                     'eef': 'expandfilename',
                                     'efm': 'exportfilemode',
                                     'eft': 'exportfiletype',
                                     'esf': 'exportfilesave',
                                     'ex': 'exists',
                                     'fon': 'filterNodes',
                                     'fsx': 'exportfilesizex',
                                     'fsy': 'exportfilesizey',
                                     'i1': 'image1',
                                     'i2': 'image2',
                                     'i3': 'image3',
                                     'ifl': 'importfileload',
                                     'ifm': 'importfilemode',
                                     'irm': 'importreassign',
                                     'iu': 'interactiveUpdate',
                                     'lr': 'lowerradius',
                                     'lrc': 'lastRecorderCmd',
                                     'lsn': 'lastStampName',
                                     'miv': 'minvalue',
                                     'mp': 'mappressure',
                                     'mst': 'makeStroke',
                                     'mxv': 'maxvalue',
                                     'n': 'name',
                                     'o': 'outline',
                                     'oaa': 'objattrArray',
                                     'op': 'opacity',
                                     'owp': 'outwhilepaint',
                                     'p': 'property',
                                     'pas': 'paintattrselected',
                                     'pcm': 'pickColor',
                                     'pcs': 'preserveclonesource',
                                     'plc': 'playbackCursor',
                                     'plp': 'playbackPressure',
                                     'pm': 'paintmode',
                                     'pna': 'paintNodeArray',
                                     'pot': 'paintoperationtype',
                                     'prm': 'projective',
                                     'psf': 'profileShapeFile',
                                     'pv': 'pickValue',
                                     'q': 'query',
                                     'r': 'radius',
                                     'ra': 'reflectionaxis',
                                     'rec': 'record',
                                     'rgb': 'rgbValue',
                                     'rmc': 'rampMinColor',
                                     'rn': 'reflection',
                                     'rno': 'reflectionaboutorigin',
                                     'rxc': 'rampMaxColor',
                                     'sa': 'showactive',
                                     'sao': 'selectedattroper',
                                     'scR': 'screenRadius',
                                     'scs': 'selectclonesource',
                                     'scv': 'surfaceConformedBrushVertices',
                                     'ssm': 'strokesmooth',
                                     'stD': 'stampDepth',
                                     'stP': 'stampProfile',
                                     'stS': 'stampSpacing',
                                     'tab': 'tablet',
                                     'tfp': 'toolOffProc',
                                     'to': 'tangentOutline',
                                     'top': 'toolOnProc',
                                     'ucr': 'useColorRamp',
                                     'umc': 'useMaxMinColor',
                                     'up': 'usepressure',
                                     'usd': 'useStrokeDirection',
                                     'v': 'velocity',
                                     'val': 'value',
                                     'wlR': 'worldRadius',
                                     'wst': 'whichTool'},
                      'type': 'context'},
  'artPuttyCtx': {'flags': {'accopacity': {'args': '<type __builtin__.bool>',
                                           'longname': 'accopacity',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'aco'},
                            'activeListChangedProc': {'args': '<type __builtin__.str>',
                                                      'longname': 'activeListChangedProc',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'alp'},
                            'afterStrokeCmd': {'args': '<type __builtin__.str>',
                                               'longname': 'afterStrokeCmd',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'asc'},
                            'alphaclamp': {'args': '<type __builtin__.str>',
                                           'longname': 'alphaclamp',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'alc'},
                            'alphaclamplower': {'args': '<type __builtin__.float>',
                                                'longname': 'alphaclamplower',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'acl'},
                            'alphaclampupper': {'args': '<type __builtin__.float>',
                                                'longname': 'alphaclampupper',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'acu'},
                            'attrSelected': {'args': '<type __builtin__.str>',
                                             'longname': 'attrSelected',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'asl'},
                            'autosmooth': {'args': '<type __builtin__.bool>',
                                           'longname': 'autosmooth',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'asm'},
                            'beforeStrokeCmd': {'args': '<type __builtin__.str>',
                                                'longname': 'beforeStrokeCmd',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'bsc'},
                            'brushStrength': {'args': '<type __builtin__.float>',
                                              'longname': 'brushStrength',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'bs'},
                            'brushalignment': {'args': '<type __builtin__.bool>',
                                               'longname': 'brushalignment',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'bra'},
                            'brushfeedback': {'args': '<type __builtin__.bool>',
                                              'longname': 'brushfeedback',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'brf'},
                            'clamp': {'args': '<type __builtin__.str>',
                                      'longname': 'clamp',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'cl'},
                            'clamplower': {'args': '<type __builtin__.float>',
                                           'longname': 'clamplower',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cll'},
                            'clampupper': {'args': '<type __builtin__.float>',
                                           'longname': 'clampupper',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'clu'},
                            'clear': {'args': '<type __builtin__.bool>',
                                      'longname': 'clear',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'clr'},
                            'collapsecvtol': {'args': '<type __builtin__.float>',
                                              'longname': 'collapsecvtol',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'clc'},
                            'colorAlphaValue': {'args': '<type __builtin__.float>',
                                                'longname': 'colorAlphaValue',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cl1'},
                            'colorRGBAValue': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'colorRGBAValue',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 4,
                                               'shortname': 'cl4'},
                            'colorRGBValue': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'colorRGBValue',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'cl3'},
                            'colorRamp': {'args': '<type __builtin__.str>',
                                          'longname': 'colorRamp',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cr'},
                            'colorfeedback': {'args': '<type __builtin__.bool>',
                                              'longname': 'colorfeedback',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cf'},
                            'colorfeedbackOverride': {'args': '<type __builtin__.bool>',
                                                      'longname': 'colorfeedbackOverride',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cfo'},
                            'colorrangelower': {'args': '<type __builtin__.float>',
                                                'longname': 'colorrangelower',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'crl'},
                            'colorrangeupper': {'args': '<type __builtin__.float>',
                                                'longname': 'colorrangeupper',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cru'},
                            'dataTypeIndex': {'args': '<type __builtin__.int>',
                                              'longname': 'dataTypeIndex',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dti'},
                            'disablelighting': {'args': '<type __builtin__.bool>',
                                                'longname': 'disablelighting',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dl'},
                            'dispdecr': {'args': '<type __builtin__.bool>',
                                         'longname': 'dispdecr',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'dde'},
                            'dispincr': {'args': '<type __builtin__.bool>',
                                         'longname': 'dispincr',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'din'},
                            'dragSlider': {'args': '<type __builtin__.str>',
                                           'longname': 'dragSlider',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dsl'},
                            'duringStrokeCmd': {'args': '<type __builtin__.str>',
                                                'longname': 'duringStrokeCmd',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dsk'},
                            'dynclonemode': {'args': '<type __builtin__.bool>',
                                             'longname': 'dynclonemode',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dcm'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'erasesrfupd': {'args': '<type __builtin__.bool>',
                                            'longname': 'erasesrfupd',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'eut'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'expandfilename': {'args': '<type __builtin__.bool>',
                                               'longname': 'expandfilename',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'eef'},
                            'exportaspectratio': {'args': '<type __builtin__.float>',
                                                  'longname': 'exportaspectratio',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ear'},
                            'exportfilemode': {'args': '<type __builtin__.str>',
                                               'longname': 'exportfilemode',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'efm'},
                            'exportfilesave': {'args': '<type __builtin__.str>',
                                               'longname': 'exportfilesave',
                                               'modes': ['edit'],
                                               'numArgs': 1,
                                               'shortname': 'esf'},
                            'exportfilesizex': {'args': '<type __builtin__.int>',
                                                'longname': 'exportfilesizex',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'fsx'},
                            'exportfilesizey': {'args': '<type __builtin__.int>',
                                                'longname': 'exportfilesizey',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'fsy'},
                            'exportfiletype': {'args': '<type __builtin__.str>',
                                               'longname': 'exportfiletype',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'eft'},
                            'filterNodes': {'args': '<type __builtin__.bool>',
                                            'longname': 'filterNodes',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'fon'},
                            'history': {'args': '<type __builtin__.bool>',
                                        'longname': 'history',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ch'},
                            'image1': {'args': '<type __builtin__.str>',
                                       'longname': 'image1',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i1'},
                            'image2': {'args': '<type __builtin__.str>',
                                       'longname': 'image2',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i2'},
                            'image3': {'args': '<type __builtin__.str>',
                                       'longname': 'image3',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i3'},
                            'importfileload': {'args': '<type __builtin__.str>',
                                               'longname': 'importfileload',
                                               'modes': ['edit'],
                                               'numArgs': 1,
                                               'shortname': 'ifl'},
                            'importfilemode': {'args': '<type __builtin__.str>',
                                               'longname': 'importfilemode',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ifm'},
                            'importreassign': {'args': '<type __builtin__.bool>',
                                               'longname': 'importreassign',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'irm'},
                            'interactiveUpdate': {'args': '<type __builtin__.bool>',
                                                  'longname': 'interactiveUpdate',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'iu'},
                            'invertrefvector': {'args': '<type __builtin__.bool>',
                                                'longname': 'invertrefvector',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'irv'},
                            'lastRecorderCmd': {'args': '<type __builtin__.str>',
                                                'longname': 'lastRecorderCmd',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'lrc'},
                            'lastStampName': {'args': '<type __builtin__.str>',
                                              'longname': 'lastStampName',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lsn'},
                            'lowerradius': {'args': '<type __builtin__.float>',
                                            'longname': 'lowerradius',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'lr'},
                            'makeStroke': {'args': '<type __builtin__.int>',
                                           'longname': 'makeStroke',
                                           'modes': ['create',
                                                     'query',
                                                     'edit',
                                                     'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'mst'},
                            'mappressure': {'args': '<type __builtin__.str>',
                                            'longname': 'mappressure',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'mp'},
                            'maxdisp': {'args': '<type __builtin__.float>',
                                        'longname': 'maxdisp',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'md'},
                            'maxvalue': {'args': '<type __builtin__.float>',
                                         'longname': 'maxvalue',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'mxv'},
                            'minvalue': {'args': '<type __builtin__.float>',
                                         'longname': 'minvalue',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'miv'},
                            'mouldtypehead': {'args': '<type __builtin__.str>',
                                              'longname': 'mouldtypehead',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'mth'},
                            'mouldtypemouse': {'args': '<type __builtin__.str>',
                                               'longname': 'mouldtypemouse',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'mtm'},
                            'mouldtypetail': {'args': '<type __builtin__.str>',
                                              'longname': 'mouldtypetail',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'mtt'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'objattrArray': {'args': '<type __builtin__.str>',
                                             'longname': 'objattrArray',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'oaa'},
                            'opacity': {'args': '<type __builtin__.float>',
                                        'longname': 'opacity',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'op'},
                            'outline': {'args': '<type __builtin__.bool>',
                                        'longname': 'outline',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                            'outwhilepaint': {'args': '<type __builtin__.bool>',
                                              'longname': 'outwhilepaint',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'owp'},
                            'paintNodeArray': {'args': '<type __builtin__.str>',
                                               'longname': 'paintNodeArray',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'pna'},
                            'paintattrselected': {'args': '<type __builtin__.str>',
                                                  'longname': 'paintattrselected',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pas'},
                            'paintmode': {'args': '<type __builtin__.str>',
                                          'longname': 'paintmode',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pm'},
                            'paintoperationtype': {'args': '<type __builtin__.str>',
                                                   'longname': 'paintoperationtype',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'pot'},
                            'pickColor': {'args': '<type __builtin__.bool>',
                                          'longname': 'pickColor',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pcm'},
                            'pickValue': {'args': '<type __builtin__.bool>',
                                          'longname': 'pickValue',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'pv'},
                            'playbackCursor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'playbackCursor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 2,
                                               'shortname': 'plc'},
                            'playbackPressure': {'args': '<type __builtin__.float>',
                                                 'longname': 'playbackPressure',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit',
                                                           'multiuse'],
                                                 'numArgs': 1,
                                                 'shortname': 'plp'},
                            'polecv': {'args': '<type __builtin__.bool>',
                                       'longname': 'polecv',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'pcv'},
                            'preserveclonesource': {'args': '<type __builtin__.bool>',
                                                    'longname': 'preserveclonesource',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'pcs'},
                            'profileShapeFile': {'args': '<type __builtin__.str>',
                                                 'longname': 'profileShapeFile',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'psf'},
                            'projective': {'args': '<type __builtin__.bool>',
                                           'longname': 'projective',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'prm'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'radius': {'args': '<type __builtin__.float>',
                                       'longname': 'radius',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'r'},
                            'rampMaxColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'rampMaxColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'rxc'},
                            'rampMinColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'rampMinColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'rmc'},
                            'record': {'args': '<type __builtin__.bool>',
                                       'longname': 'record',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'rec'},
                            'reflection': {'args': '<type __builtin__.bool>',
                                           'longname': 'reflection',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'rn'},
                            'reflectionaboutorigin': {'args': '<type __builtin__.bool>',
                                                      'longname': 'reflectionaboutorigin',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'rno'},
                            'reflectionaxis': {'args': '<type __builtin__.str>',
                                               'longname': 'reflectionaxis',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ra'},
                            'refsurface': {'args': '<type __builtin__.bool>',
                                           'longname': 'refsurface',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'rs'},
                            'refvector': {'args': '<type __builtin__.str>',
                                          'longname': 'refvector',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'rv'},
                            'refvectoru': {'args': '<type __builtin__.float>',
                                           'longname': 'refvectoru',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'rvu'},
                            'refvectorv': {'args': '<type __builtin__.float>',
                                           'longname': 'refvectorv',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'rvv'},
                            'screenRadius': {'args': '<type __builtin__.float>',
                                             'longname': 'screenRadius',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'scR'},
                            'selectclonesource': {'args': '<type __builtin__.bool>',
                                                  'longname': 'selectclonesource',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'scs'},
                            'selectedattroper': {'args': '<type __builtin__.str>',
                                                 'longname': 'selectedattroper',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sao'},
                            'showactive': {'args': '<type __builtin__.bool>',
                                           'longname': 'showactive',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sa'},
                            'smoothiters': {'args': '<type __builtin__.int>',
                                            'longname': 'smoothiters',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'si'},
                            'stampDepth': {'args': '<type __builtin__.float>',
                                           'longname': 'stampDepth',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'stD'},
                            'stampProfile': {'args': '<type __builtin__.str>',
                                             'longname': 'stampProfile',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'stP'},
                            'stampSpacing': {'args': '<type __builtin__.float>',
                                             'longname': 'stampSpacing',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'stS'},
                            'stitchcorner': {'args': '<type __builtin__.bool>',
                                             'longname': 'stitchcorner',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'stc'},
                            'stitchedgeflood': {'args': '<type __builtin__.bool>',
                                                'longname': 'stitchedgeflood',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'sef'},
                            'stitchtype': {'args': '<type __builtin__.str>',
                                           'longname': 'stitchtype',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'stt'},
                            'strokesmooth': {'args': '<type __builtin__.str>',
                                             'longname': 'strokesmooth',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ssm'},
                            'surfaceConformedBrushVertices': {'args': '<type __builtin__.bool>',
                                                              'longname': 'surfaceConformedBrushVertices',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'scv'},
                            'tablet': {'args': '<type __builtin__.bool>',
                                       'longname': 'tablet',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'tab'},
                            'tangentOutline': {'args': '<type __builtin__.bool>',
                                               'longname': 'tangentOutline',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'to'},
                            'toolOffProc': {'args': '<type __builtin__.str>',
                                            'longname': 'toolOffProc',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'tfp'},
                            'toolOnProc': {'args': '<type __builtin__.str>',
                                           'longname': 'toolOnProc',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'top'},
                            'updateerasesrf': {'args': '<type __builtin__.bool>',
                                               'longname': 'updateerasesrf',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'ues'},
                            'updaterefsrf': {'args': '<type __builtin__.bool>',
                                             'longname': 'updaterefsrf',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'urs'},
                            'useColorRamp': {'args': '<type __builtin__.bool>',
                                             'longname': 'useColorRamp',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ucr'},
                            'useMaxMinColor': {'args': '<type __builtin__.bool>',
                                               'longname': 'useMaxMinColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'umc'},
                            'usepressure': {'args': '<type __builtin__.bool>',
                                            'longname': 'usepressure',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'up'},
                            'value': {'args': '<type __builtin__.float>',
                                      'longname': 'value',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'val'},
                            'whichTool': {'args': '<type __builtin__.str>',
                                          'longname': 'whichTool',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'wst'},
                            'worldRadius': {'args': '<type __builtin__.float>',
                                            'longname': 'worldRadius',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'wlR'}},
                  'shortFlags': {'acl': 'alphaclamplower',
                                 'aco': 'accopacity',
                                 'acu': 'alphaclampupper',
                                 'alc': 'alphaclamp',
                                 'alp': 'activeListChangedProc',
                                 'asc': 'afterStrokeCmd',
                                 'asl': 'attrSelected',
                                 'asm': 'autosmooth',
                                 'bra': 'brushalignment',
                                 'brf': 'brushfeedback',
                                 'bs': 'brushStrength',
                                 'bsc': 'beforeStrokeCmd',
                                 'cf': 'colorfeedback',
                                 'cfo': 'colorfeedbackOverride',
                                 'ch': 'history',
                                 'cl': 'clamp',
                                 'cl1': 'colorAlphaValue',
                                 'cl3': 'colorRGBValue',
                                 'cl4': 'colorRGBAValue',
                                 'clc': 'collapsecvtol',
                                 'cll': 'clamplower',
                                 'clr': 'clear',
                                 'clu': 'clampupper',
                                 'cr': 'colorRamp',
                                 'crl': 'colorrangelower',
                                 'cru': 'colorrangeupper',
                                 'dcm': 'dynclonemode',
                                 'dde': 'dispdecr',
                                 'din': 'dispincr',
                                 'dl': 'disablelighting',
                                 'dsk': 'duringStrokeCmd',
                                 'dsl': 'dragSlider',
                                 'dti': 'dataTypeIndex',
                                 'e': 'edit',
                                 'ear': 'exportaspectratio',
                                 'eef': 'expandfilename',
                                 'efm': 'exportfilemode',
                                 'eft': 'exportfiletype',
                                 'esf': 'exportfilesave',
                                 'eut': 'erasesrfupd',
                                 'ex': 'exists',
                                 'fon': 'filterNodes',
                                 'fsx': 'exportfilesizex',
                                 'fsy': 'exportfilesizey',
                                 'i1': 'image1',
                                 'i2': 'image2',
                                 'i3': 'image3',
                                 'ifl': 'importfileload',
                                 'ifm': 'importfilemode',
                                 'irm': 'importreassign',
                                 'irv': 'invertrefvector',
                                 'iu': 'interactiveUpdate',
                                 'lr': 'lowerradius',
                                 'lrc': 'lastRecorderCmd',
                                 'lsn': 'lastStampName',
                                 'md': 'maxdisp',
                                 'miv': 'minvalue',
                                 'mp': 'mappressure',
                                 'mst': 'makeStroke',
                                 'mth': 'mouldtypehead',
                                 'mtm': 'mouldtypemouse',
                                 'mtt': 'mouldtypetail',
                                 'mxv': 'maxvalue',
                                 'n': 'name',
                                 'o': 'outline',
                                 'oaa': 'objattrArray',
                                 'op': 'opacity',
                                 'owp': 'outwhilepaint',
                                 'pas': 'paintattrselected',
                                 'pcm': 'pickColor',
                                 'pcs': 'preserveclonesource',
                                 'pcv': 'polecv',
                                 'plc': 'playbackCursor',
                                 'plp': 'playbackPressure',
                                 'pm': 'paintmode',
                                 'pna': 'paintNodeArray',
                                 'pot': 'paintoperationtype',
                                 'prm': 'projective',
                                 'psf': 'profileShapeFile',
                                 'pv': 'pickValue',
                                 'q': 'query',
                                 'r': 'radius',
                                 'ra': 'reflectionaxis',
                                 'rec': 'record',
                                 'rmc': 'rampMinColor',
                                 'rn': 'reflection',
                                 'rno': 'reflectionaboutorigin',
                                 'rs': 'refsurface',
                                 'rv': 'refvector',
                                 'rvu': 'refvectoru',
                                 'rvv': 'refvectorv',
                                 'rxc': 'rampMaxColor',
                                 'sa': 'showactive',
                                 'sao': 'selectedattroper',
                                 'scR': 'screenRadius',
                                 'scs': 'selectclonesource',
                                 'scv': 'surfaceConformedBrushVertices',
                                 'sef': 'stitchedgeflood',
                                 'si': 'smoothiters',
                                 'ssm': 'strokesmooth',
                                 'stD': 'stampDepth',
                                 'stP': 'stampProfile',
                                 'stS': 'stampSpacing',
                                 'stc': 'stitchcorner',
                                 'stt': 'stitchtype',
                                 'tab': 'tablet',
                                 'tfp': 'toolOffProc',
                                 'to': 'tangentOutline',
                                 'top': 'toolOnProc',
                                 'ucr': 'useColorRamp',
                                 'ues': 'updateerasesrf',
                                 'umc': 'useMaxMinColor',
                                 'up': 'usepressure',
                                 'urs': 'updaterefsrf',
                                 'val': 'value',
                                 'wlR': 'worldRadius',
                                 'wst': 'whichTool'},
                  'type': 'context'},
  'artSelect': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'artSelectCtx': {'flags': {'accopacity': {'args': '<type __builtin__.bool>',
                                            'longname': 'accopacity',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'aco'},
                             'addselection': {'args': '<type __builtin__.bool>',
                                              'longname': 'addselection',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ads'},
                             'afterStrokeCmd': {'args': '<type __builtin__.str>',
                                                'longname': 'afterStrokeCmd',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'asc'},
                             'beforeStrokeCmd': {'args': '<type __builtin__.str>',
                                                 'longname': 'beforeStrokeCmd',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'bsc'},
                             'brushalignment': {'args': '<type __builtin__.bool>',
                                                'longname': 'brushalignment',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'bra'},
                             'brushfeedback': {'args': '<type __builtin__.bool>',
                                               'longname': 'brushfeedback',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'brf'},
                             'clear': {'args': '<type __builtin__.bool>',
                                       'longname': 'clear',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'clr'},
                             'dragSlider': {'args': '<type __builtin__.str>',
                                            'longname': 'dragSlider',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dsl'},
                             'dynclonemode': {'args': '<type __builtin__.bool>',
                                              'longname': 'dynclonemode',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dcm'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'expandfilename': {'args': '<type __builtin__.bool>',
                                                'longname': 'expandfilename',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'eef'},
                             'exportaspectratio': {'args': '<type __builtin__.float>',
                                                   'longname': 'exportaspectratio',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ear'},
                             'exportfilemode': {'args': '<type __builtin__.str>',
                                                'longname': 'exportfilemode',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'efm'},
                             'exportfilesave': {'args': '<type __builtin__.str>',
                                                'longname': 'exportfilesave',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'esf'},
                             'exportfilesizex': {'args': '<type __builtin__.int>',
                                                 'longname': 'exportfilesizex',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'fsx'},
                             'exportfilesizey': {'args': '<type __builtin__.int>',
                                                 'longname': 'exportfilesizey',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'fsy'},
                             'exportfiletype': {'args': '<type __builtin__.str>',
                                                'longname': 'exportfiletype',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'eft'},
                             'history': {'args': '<type __builtin__.bool>',
                                         'longname': 'history',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ch'},
                             'image1': {'args': '<type __builtin__.str>',
                                        'longname': 'image1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i1'},
                             'image2': {'args': '<type __builtin__.str>',
                                        'longname': 'image2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i2'},
                             'image3': {'args': '<type __builtin__.str>',
                                        'longname': 'image3',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i3'},
                             'importfileload': {'args': '<type __builtin__.str>',
                                                'longname': 'importfileload',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'ifl'},
                             'importfilemode': {'args': '<type __builtin__.str>',
                                                'longname': 'importfilemode',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ifm'},
                             'importreassign': {'args': '<type __builtin__.bool>',
                                                'longname': 'importreassign',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'irm'},
                             'importthreshold': {'args': '<type __builtin__.float>',
                                                 'longname': 'importthreshold',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ift'},
                             'lastRecorderCmd': {'args': '<type __builtin__.str>',
                                                 'longname': 'lastRecorderCmd',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'lrc'},
                             'lastStampName': {'args': '<type __builtin__.str>',
                                               'longname': 'lastStampName',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'lsn'},
                             'lowerradius': {'args': '<type __builtin__.float>',
                                             'longname': 'lowerradius',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lr'},
                             'makeStroke': {'args': '<type __builtin__.int>',
                                            'longname': 'makeStroke',
                                            'modes': ['create',
                                                      'query',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'mst'},
                             'mappressure': {'args': '<type __builtin__.str>',
                                             'longname': 'mappressure',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'mp'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'opacity': {'args': '<type __builtin__.float>',
                                         'longname': 'opacity',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'op'},
                             'outline': {'args': '<type __builtin__.bool>',
                                         'longname': 'outline',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'o'},
                             'outwhilepaint': {'args': '<type __builtin__.bool>',
                                               'longname': 'outwhilepaint',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'owp'},
                             'paintmode': {'args': '<type __builtin__.str>',
                                           'longname': 'paintmode',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pm'},
                             'paintoperationtype': {'args': '<type __builtin__.str>',
                                                    'longname': 'paintoperationtype',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'pot'},
                             'pickColor': {'args': '<type __builtin__.bool>',
                                           'longname': 'pickColor',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pcm'},
                             'pickValue': {'args': '<type __builtin__.bool>',
                                           'longname': 'pickValue',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 0,
                                           'shortname': 'pv'},
                             'playbackCursor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'playbackCursor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 2,
                                                'shortname': 'plc'},
                             'playbackPressure': {'args': '<type __builtin__.float>',
                                                  'longname': 'playbackPressure',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'plp'},
                             'preserveclonesource': {'args': '<type __builtin__.bool>',
                                                     'longname': 'preserveclonesource',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'pcs'},
                             'profileShapeFile': {'args': '<type __builtin__.str>',
                                                  'longname': 'profileShapeFile',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'psf'},
                             'projective': {'args': '<type __builtin__.bool>',
                                            'longname': 'projective',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'prm'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'radius': {'args': '<type __builtin__.float>',
                                        'longname': 'radius',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'r'},
                             'record': {'args': '<type __builtin__.bool>',
                                        'longname': 'record',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'rec'},
                             'reflection': {'args': '<type __builtin__.bool>',
                                            'longname': 'reflection',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'rn'},
                             'reflectionaboutorigin': {'args': '<type __builtin__.bool>',
                                                       'longname': 'reflectionaboutorigin',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'rno'},
                             'reflectionaxis': {'args': '<type __builtin__.str>',
                                                'longname': 'reflectionaxis',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ra'},
                             'screenRadius': {'args': '<type __builtin__.float>',
                                              'longname': 'screenRadius',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'scR'},
                             'selectall': {'args': '<type __builtin__.bool>',
                                           'longname': 'selectall',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'sal'},
                             'selectclonesource': {'args': '<type __builtin__.bool>',
                                                   'longname': 'selectclonesource',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'scs'},
                             'selectop': {'args': '<type __builtin__.str>',
                                          'longname': 'selectop',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sop'},
                             'showactive': {'args': '<type __builtin__.bool>',
                                            'longname': 'showactive',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sa'},
                             'stampDepth': {'args': '<type __builtin__.float>',
                                            'longname': 'stampDepth',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'stD'},
                             'stampProfile': {'args': '<type __builtin__.str>',
                                              'longname': 'stampProfile',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'stP'},
                             'stampSpacing': {'args': '<type __builtin__.float>',
                                              'longname': 'stampSpacing',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'stS'},
                             'strokesmooth': {'args': '<type __builtin__.str>',
                                              'longname': 'strokesmooth',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ssm'},
                             'surfaceConformedBrushVertices': {'args': '<type __builtin__.bool>',
                                                               'longname': 'surfaceConformedBrushVertices',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'scv'},
                             'tablet': {'args': '<type __builtin__.bool>',
                                        'longname': 'tablet',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'tab'},
                             'tangentOutline': {'args': '<type __builtin__.bool>',
                                                'longname': 'tangentOutline',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'to'},
                             'toggleall': {'args': '<type __builtin__.bool>',
                                           'longname': 'toggleall',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'tal'},
                             'unselectall': {'args': '<type __builtin__.bool>',
                                             'longname': 'unselectall',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'ual'},
                             'usepressure': {'args': '<type __builtin__.bool>',
                                             'longname': 'usepressure',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'up'},
                             'worldRadius': {'args': '<type __builtin__.float>',
                                             'longname': 'worldRadius',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'wlR'}},
                   'shortFlags': {'aco': 'accopacity',
                                  'ads': 'addselection',
                                  'asc': 'afterStrokeCmd',
                                  'bra': 'brushalignment',
                                  'brf': 'brushfeedback',
                                  'bsc': 'beforeStrokeCmd',
                                  'ch': 'history',
                                  'clr': 'clear',
                                  'dcm': 'dynclonemode',
                                  'dsl': 'dragSlider',
                                  'e': 'edit',
                                  'ear': 'exportaspectratio',
                                  'eef': 'expandfilename',
                                  'efm': 'exportfilemode',
                                  'eft': 'exportfiletype',
                                  'esf': 'exportfilesave',
                                  'ex': 'exists',
                                  'fsx': 'exportfilesizex',
                                  'fsy': 'exportfilesizey',
                                  'i1': 'image1',
                                  'i2': 'image2',
                                  'i3': 'image3',
                                  'ifl': 'importfileload',
                                  'ifm': 'importfilemode',
                                  'ift': 'importthreshold',
                                  'irm': 'importreassign',
                                  'lr': 'lowerradius',
                                  'lrc': 'lastRecorderCmd',
                                  'lsn': 'lastStampName',
                                  'mp': 'mappressure',
                                  'mst': 'makeStroke',
                                  'n': 'name',
                                  'o': 'outline',
                                  'op': 'opacity',
                                  'owp': 'outwhilepaint',
                                  'pcm': 'pickColor',
                                  'pcs': 'preserveclonesource',
                                  'plc': 'playbackCursor',
                                  'plp': 'playbackPressure',
                                  'pm': 'paintmode',
                                  'pot': 'paintoperationtype',
                                  'prm': 'projective',
                                  'psf': 'profileShapeFile',
                                  'pv': 'pickValue',
                                  'q': 'query',
                                  'r': 'radius',
                                  'ra': 'reflectionaxis',
                                  'rec': 'record',
                                  'rn': 'reflection',
                                  'rno': 'reflectionaboutorigin',
                                  'sa': 'showactive',
                                  'sal': 'selectall',
                                  'scR': 'screenRadius',
                                  'scs': 'selectclonesource',
                                  'scv': 'surfaceConformedBrushVertices',
                                  'sop': 'selectop',
                                  'ssm': 'strokesmooth',
                                  'stD': 'stampDepth',
                                  'stP': 'stampProfile',
                                  'stS': 'stampSpacing',
                                  'tab': 'tablet',
                                  'tal': 'toggleall',
                                  'to': 'tangentOutline',
                                  'ual': 'unselectall',
                                  'up': 'usepressure',
                                  'wlR': 'worldRadius'},
                   'type': 'context'},
  'artSetPaint': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'artSetPaintCtx': {'flags': {'accopacity': {'args': '<type __builtin__.bool>',
                                              'longname': 'accopacity',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'aco'},
                               'afterStrokeCmd': {'args': '<type __builtin__.str>',
                                                  'longname': 'afterStrokeCmd',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'asc'},
                               'beforeStrokeCmd': {'args': '<type __builtin__.str>',
                                                   'longname': 'beforeStrokeCmd',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'bsc'},
                               'brushalignment': {'args': '<type __builtin__.bool>',
                                                  'longname': 'brushalignment',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'bra'},
                               'brushfeedback': {'args': '<type __builtin__.bool>',
                                                 'longname': 'brushfeedback',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'brf'},
                               'clear': {'args': '<type __builtin__.bool>',
                                         'longname': 'clear',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'clr'},
                               'dragSlider': {'args': '<type __builtin__.str>',
                                              'longname': 'dragSlider',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dsl'},
                               'dynclonemode': {'args': '<type __builtin__.bool>',
                                                'longname': 'dynclonemode',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dcm'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'expandfilename': {'args': '<type __builtin__.bool>',
                                                  'longname': 'expandfilename',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'eef'},
                               'exportaspectratio': {'args': '<type __builtin__.float>',
                                                     'longname': 'exportaspectratio',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ear'},
                               'exportfilemode': {'args': '<type __builtin__.str>',
                                                  'longname': 'exportfilemode',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'efm'},
                               'exportfilesave': {'args': '<type __builtin__.str>',
                                                  'longname': 'exportfilesave',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'esf'},
                               'exportfilesizex': {'args': '<type __builtin__.int>',
                                                   'longname': 'exportfilesizex',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'fsx'},
                               'exportfilesizey': {'args': '<type __builtin__.int>',
                                                   'longname': 'exportfilesizey',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'fsy'},
                               'exportfiletype': {'args': '<type __builtin__.str>',
                                                  'longname': 'exportfiletype',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'eft'},
                               'history': {'args': '<type __builtin__.bool>',
                                           'longname': 'history',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ch'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'importfileload': {'args': '<type __builtin__.str>',
                                                  'longname': 'importfileload',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ifl'},
                               'importfilemode': {'args': '<type __builtin__.str>',
                                                  'longname': 'importfilemode',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ifm'},
                               'importreassign': {'args': '<type __builtin__.bool>',
                                                  'longname': 'importreassign',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'irm'},
                               'lastRecorderCmd': {'args': '<type __builtin__.str>',
                                                   'longname': 'lastRecorderCmd',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'lrc'},
                               'lastStampName': {'args': '<type __builtin__.str>',
                                                 'longname': 'lastStampName',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'lsn'},
                               'lowerradius': {'args': '<type __builtin__.float>',
                                               'longname': 'lowerradius',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'lr'},
                               'makeStroke': {'args': '<type __builtin__.int>',
                                              'longname': 'makeStroke',
                                              'modes': ['create',
                                                        'query',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'mst'},
                               'mappressure': {'args': '<type __builtin__.str>',
                                               'longname': 'mappressure',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'mp'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'objectsetnames': {'args': '<type __builtin__.str>',
                                                  'longname': 'objectsetnames',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'osn'},
                               'opacity': {'args': '<type __builtin__.float>',
                                           'longname': 'opacity',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'op'},
                               'outline': {'args': '<type __builtin__.bool>',
                                           'longname': 'outline',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'o'},
                               'outwhilepaint': {'args': '<type __builtin__.bool>',
                                                 'longname': 'outwhilepaint',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'owp'},
                               'paintmode': {'args': '<type __builtin__.str>',
                                             'longname': 'paintmode',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'pm'},
                               'paintoperationtype': {'args': '<type __builtin__.str>',
                                                      'longname': 'paintoperationtype',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'pot'},
                               'pickColor': {'args': '<type __builtin__.bool>',
                                             'longname': 'pickColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'pcm'},
                               'pickValue': {'args': '<type __builtin__.bool>',
                                             'longname': 'pickValue',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 0,
                                             'shortname': 'pv'},
                               'playbackCursor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'playbackCursor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 2,
                                                  'shortname': 'plc'},
                               'playbackPressure': {'args': '<type __builtin__.float>',
                                                    'longname': 'playbackPressure',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 1,
                                                    'shortname': 'plp'},
                               'preserveclonesource': {'args': '<type __builtin__.bool>',
                                                       'longname': 'preserveclonesource',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'pcs'},
                               'profileShapeFile': {'args': '<type __builtin__.str>',
                                                    'longname': 'profileShapeFile',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'psf'},
                               'projective': {'args': '<type __builtin__.bool>',
                                              'longname': 'projective',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'prm'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'radius': {'args': '<type __builtin__.float>',
                                          'longname': 'radius',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'r'},
                               'record': {'args': '<type __builtin__.bool>',
                                          'longname': 'record',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'rec'},
                               'reflection': {'args': '<type __builtin__.bool>',
                                              'longname': 'reflection',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'rn'},
                               'reflectionaboutorigin': {'args': '<type __builtin__.bool>',
                                                         'longname': 'reflectionaboutorigin',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'rno'},
                               'reflectionaxis': {'args': '<type __builtin__.str>',
                                                  'longname': 'reflectionaxis',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ra'},
                               'screenRadius': {'args': '<type __builtin__.float>',
                                                'longname': 'screenRadius',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'scR'},
                               'selectclonesource': {'args': '<type __builtin__.bool>',
                                                     'longname': 'selectclonesource',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'scs'},
                               'setcolorfeedback': {'args': '<type __builtin__.bool>',
                                                    'longname': 'setcolorfeedback',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'scf'},
                               'setdisplaycvs': {'args': '<type __builtin__.bool>',
                                                 'longname': 'setdisplaycvs',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dcv'},
                               'setopertype': {'args': '<type __builtin__.str>',
                                               'longname': 'setopertype',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sot'},
                               'settomodify': {'args': '<type __builtin__.str>',
                                               'longname': 'settomodify',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'stm'},
                               'showactive': {'args': '<type __builtin__.bool>',
                                              'longname': 'showactive',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sa'},
                               'stampDepth': {'args': '<type __builtin__.float>',
                                              'longname': 'stampDepth',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'stD'},
                               'stampProfile': {'args': '<type __builtin__.str>',
                                                'longname': 'stampProfile',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'stP'},
                               'stampSpacing': {'args': '<type __builtin__.float>',
                                                'longname': 'stampSpacing',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'stS'},
                               'strokesmooth': {'args': '<type __builtin__.str>',
                                                'longname': 'strokesmooth',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ssm'},
                               'surfaceConformedBrushVertices': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'surfaceConformedBrushVertices',
                                                                 'modes': ['create',
                                                                           'query',
                                                                           'edit'],
                                                                 'numArgs': 1,
                                                                 'shortname': 'scv'},
                               'tablet': {'args': '<type __builtin__.bool>',
                                          'longname': 'tablet',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'tab'},
                               'tangentOutline': {'args': '<type __builtin__.bool>',
                                                  'longname': 'tangentOutline',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'to'},
                               'usepressure': {'args': '<type __builtin__.bool>',
                                               'longname': 'usepressure',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'up'},
                               'worldRadius': {'args': '<type __builtin__.float>',
                                               'longname': 'worldRadius',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'wlR'}},
                     'shortFlags': {'aco': 'accopacity',
                                    'asc': 'afterStrokeCmd',
                                    'bra': 'brushalignment',
                                    'brf': 'brushfeedback',
                                    'bsc': 'beforeStrokeCmd',
                                    'ch': 'history',
                                    'clr': 'clear',
                                    'dcm': 'dynclonemode',
                                    'dcv': 'setdisplaycvs',
                                    'dsl': 'dragSlider',
                                    'e': 'edit',
                                    'ear': 'exportaspectratio',
                                    'eef': 'expandfilename',
                                    'efm': 'exportfilemode',
                                    'eft': 'exportfiletype',
                                    'esf': 'exportfilesave',
                                    'ex': 'exists',
                                    'fsx': 'exportfilesizex',
                                    'fsy': 'exportfilesizey',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'ifl': 'importfileload',
                                    'ifm': 'importfilemode',
                                    'irm': 'importreassign',
                                    'lr': 'lowerradius',
                                    'lrc': 'lastRecorderCmd',
                                    'lsn': 'lastStampName',
                                    'mp': 'mappressure',
                                    'mst': 'makeStroke',
                                    'n': 'name',
                                    'o': 'outline',
                                    'op': 'opacity',
                                    'osn': 'objectsetnames',
                                    'owp': 'outwhilepaint',
                                    'pcm': 'pickColor',
                                    'pcs': 'preserveclonesource',
                                    'plc': 'playbackCursor',
                                    'plp': 'playbackPressure',
                                    'pm': 'paintmode',
                                    'pot': 'paintoperationtype',
                                    'prm': 'projective',
                                    'psf': 'profileShapeFile',
                                    'pv': 'pickValue',
                                    'q': 'query',
                                    'r': 'radius',
                                    'ra': 'reflectionaxis',
                                    'rec': 'record',
                                    'rn': 'reflection',
                                    'rno': 'reflectionaboutorigin',
                                    'sa': 'showactive',
                                    'scR': 'screenRadius',
                                    'scf': 'setcolorfeedback',
                                    'scs': 'selectclonesource',
                                    'scv': 'surfaceConformedBrushVertices',
                                    'sot': 'setopertype',
                                    'ssm': 'strokesmooth',
                                    'stD': 'stampDepth',
                                    'stP': 'stampProfile',
                                    'stS': 'stampSpacing',
                                    'stm': 'settomodify',
                                    'tab': 'tablet',
                                    'to': 'tangentOutline',
                                    'up': 'usepressure',
                                    'wlR': 'worldRadius'},
                     'type': 'context'},
  'artUserPaintCtx': {'callbackFlags': ['setArrayValueCommand',
                                        'sac',
                                        'getValueCommand',
                                        'gvc',
                                        'getSurfaceCommand',
                                        'gsc',
                                        'chunkCommand',
                                        'cc',
                                        'getArrayAttrCommand',
                                        'gac',
                                        'setValueCommand',
                                        'svc'],
                      'flags': {'accopacity': {'args': '<type __builtin__.bool>',
                                               'longname': 'accopacity',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'aco'},
                                'activeListChangedProc': {'args': '<type __builtin__.str>',
                                                          'longname': 'activeListChangedProc',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'alp'},
                                'afterStrokeCmd': {'args': '<type __builtin__.str>',
                                                   'longname': 'afterStrokeCmd',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'asc'},
                                'alphaclamp': {'args': '<type __builtin__.str>',
                                               'longname': 'alphaclamp',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'alc'},
                                'alphaclamplower': {'args': '<type __builtin__.float>',
                                                    'longname': 'alphaclamplower',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'acl'},
                                'alphaclampupper': {'args': '<type __builtin__.float>',
                                                    'longname': 'alphaclampupper',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'acu'},
                                'attrSelected': {'args': '<type __builtin__.str>',
                                                 'longname': 'attrSelected',
                                                 'modes': ['query'],
                                                 'numArgs': 1,
                                                 'shortname': 'asl'},
                                'beforeStrokeCmd': {'args': '<type __builtin__.str>',
                                                    'longname': 'beforeStrokeCmd',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'bsc'},
                                'brushalignment': {'args': '<type __builtin__.bool>',
                                                   'longname': 'brushalignment',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'bra'},
                                'brushfeedback': {'args': '<type __builtin__.bool>',
                                                  'longname': 'brushfeedback',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'brf'},
                                'chunkCommand': {'args': '<type __builtin__.str>',
                                                 'longname': 'chunkCommand',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cc'},
                                'clamp': {'args': '<type __builtin__.str>',
                                          'longname': 'clamp',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cl'},
                                'clamplower': {'args': '<type __builtin__.float>',
                                               'longname': 'clamplower',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cll'},
                                'clampupper': {'args': '<type __builtin__.float>',
                                               'longname': 'clampupper',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'clu'},
                                'clear': {'args': '<type __builtin__.bool>',
                                          'longname': 'clear',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'clr'},
                                'colorAlphaValue': {'args': '<type __builtin__.float>',
                                                    'longname': 'colorAlphaValue',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'cl1'},
                                'colorRGBAValue': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'colorRGBAValue',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 4,
                                                   'shortname': 'cl4'},
                                'colorRGBValue': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'colorRGBValue',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'cl3'},
                                'colorRamp': {'args': '<type __builtin__.str>',
                                              'longname': 'colorRamp',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cr'},
                                'colorfeedback': {'args': '<type __builtin__.bool>',
                                                  'longname': 'colorfeedback',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cf'},
                                'colorfeedbackOverride': {'args': '<type __builtin__.bool>',
                                                          'longname': 'colorfeedbackOverride',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'cfo'},
                                'colorrangelower': {'args': '<type __builtin__.float>',
                                                    'longname': 'colorrangelower',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'crl'},
                                'colorrangeupper': {'args': '<type __builtin__.float>',
                                                    'longname': 'colorrangeupper',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'cru'},
                                'dataTypeIndex': {'args': '<type __builtin__.int>',
                                                  'longname': 'dataTypeIndex',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dti'},
                                'disablelighting': {'args': '<type __builtin__.bool>',
                                                    'longname': 'disablelighting',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dl'},
                                'dragSlider': {'args': '<type __builtin__.str>',
                                               'longname': 'dragSlider',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dsl'},
                                'duringStrokeCmd': {'args': '<type __builtin__.str>',
                                                    'longname': 'duringStrokeCmd',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dsk'},
                                'dynclonemode': {'args': '<type __builtin__.bool>',
                                                 'longname': 'dynclonemode',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dcm'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'expandfilename': {'args': '<type __builtin__.bool>',
                                                   'longname': 'expandfilename',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'eef'},
                                'exportaspectratio': {'args': '<type __builtin__.float>',
                                                      'longname': 'exportaspectratio',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ear'},
                                'exportfilemode': {'args': '<type __builtin__.str>',
                                                   'longname': 'exportfilemode',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'efm'},
                                'exportfilesave': {'args': '<type __builtin__.str>',
                                                   'longname': 'exportfilesave',
                                                   'modes': ['edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'esf'},
                                'exportfilesizex': {'args': '<type __builtin__.int>',
                                                    'longname': 'exportfilesizex',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'fsx'},
                                'exportfilesizey': {'args': '<type __builtin__.int>',
                                                    'longname': 'exportfilesizey',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'fsy'},
                                'exportfiletype': {'args': '<type __builtin__.str>',
                                                   'longname': 'exportfiletype',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'eft'},
                                'filterNodes': {'args': '<type __builtin__.bool>',
                                                'longname': 'filterNodes',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'fon'},
                                'finalizeCmd': {'args': '<type __builtin__.str>',
                                                'longname': 'finalizeCmd',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'fc'},
                                'fullpaths': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullpaths',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fp'},
                                'getArrayAttrCommand': {'args': '<type __builtin__.str>',
                                                        'longname': 'getArrayAttrCommand',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'gac'},
                                'getSurfaceCommand': {'args': '<type __builtin__.str>',
                                                      'longname': 'getSurfaceCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'gsc'},
                                'getValueCommand': {'args': '<type __builtin__.str>',
                                                    'longname': 'getValueCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'gvc'},
                                'history': {'args': '<type __builtin__.bool>',
                                            'longname': 'history',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ch'},
                                'image1': {'args': '<type __builtin__.str>',
                                           'longname': 'image1',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i1'},
                                'image2': {'args': '<type __builtin__.str>',
                                           'longname': 'image2',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i2'},
                                'image3': {'args': '<type __builtin__.str>',
                                           'longname': 'image3',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i3'},
                                'importfileload': {'args': '<type __builtin__.str>',
                                                   'longname': 'importfileload',
                                                   'modes': ['edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ifl'},
                                'importfilemode': {'args': '<type __builtin__.str>',
                                                   'longname': 'importfilemode',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ifm'},
                                'importreassign': {'args': '<type __builtin__.bool>',
                                                   'longname': 'importreassign',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'irm'},
                                'initializeCmd': {'args': '<type __builtin__.str>',
                                                  'longname': 'initializeCmd',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ic'},
                                'interactiveUpdate': {'args': '<type __builtin__.bool>',
                                                      'longname': 'interactiveUpdate',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'iu'},
                                'lastRecorderCmd': {'args': '<type __builtin__.str>',
                                                    'longname': 'lastRecorderCmd',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'lrc'},
                                'lastStampName': {'args': '<type __builtin__.str>',
                                                  'longname': 'lastStampName',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'lsn'},
                                'lowerradius': {'args': '<type __builtin__.float>',
                                                'longname': 'lowerradius',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'lr'},
                                'makeStroke': {'args': '<type __builtin__.int>',
                                               'longname': 'makeStroke',
                                               'modes': ['create',
                                                         'query',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'mst'},
                                'mappressure': {'args': '<type __builtin__.str>',
                                                'longname': 'mappressure',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'mp'},
                                'maxvalue': {'args': '<type __builtin__.float>',
                                             'longname': 'maxvalue',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'mxv'},
                                'minvalue': {'args': '<type __builtin__.float>',
                                             'longname': 'minvalue',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'miv'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'objattrArray': {'args': '<type __builtin__.str>',
                                                 'longname': 'objattrArray',
                                                 'modes': ['query'],
                                                 'numArgs': 1,
                                                 'shortname': 'oaa'},
                                'opacity': {'args': '<type __builtin__.float>',
                                            'longname': 'opacity',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'op'},
                                'outline': {'args': '<type __builtin__.bool>',
                                            'longname': 'outline',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'o'},
                                'outwhilepaint': {'args': '<type __builtin__.bool>',
                                                  'longname': 'outwhilepaint',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'owp'},
                                'paintNodeArray': {'args': '<type __builtin__.str>',
                                                   'longname': 'paintNodeArray',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'pna'},
                                'paintattrselected': {'args': '<type __builtin__.str>',
                                                      'longname': 'paintattrselected',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'pas'},
                                'paintmode': {'args': '<type __builtin__.str>',
                                              'longname': 'paintmode',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pm'},
                                'paintoperationtype': {'args': '<type __builtin__.str>',
                                                       'longname': 'paintoperationtype',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'pot'},
                                'pickColor': {'args': '<type __builtin__.bool>',
                                              'longname': 'pickColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pcm'},
                                'pickValue': {'args': '<type __builtin__.bool>',
                                              'longname': 'pickValue',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 0,
                                              'shortname': 'pv'},
                                'playbackCursor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'playbackCursor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 2,
                                                   'shortname': 'plc'},
                                'playbackPressure': {'args': '<type __builtin__.float>',
                                                     'longname': 'playbackPressure',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'plp'},
                                'preserveclonesource': {'args': '<type __builtin__.bool>',
                                                        'longname': 'preserveclonesource',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'pcs'},
                                'profileShapeFile': {'args': '<type __builtin__.str>',
                                                     'longname': 'profileShapeFile',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'psf'},
                                'projective': {'args': '<type __builtin__.bool>',
                                               'longname': 'projective',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'prm'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'radius': {'args': '<type __builtin__.float>',
                                           'longname': 'radius',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'r'},
                                'rampMaxColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'rampMaxColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'rxc'},
                                'rampMinColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'rampMinColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'rmc'},
                                'record': {'args': '<type __builtin__.bool>',
                                           'longname': 'record',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'rec'},
                                'reflection': {'args': '<type __builtin__.bool>',
                                               'longname': 'reflection',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rn'},
                                'reflectionaboutorigin': {'args': '<type __builtin__.bool>',
                                                          'longname': 'reflectionaboutorigin',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'rno'},
                                'reflectionaxis': {'args': '<type __builtin__.str>',
                                                   'longname': 'reflectionaxis',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ra'},
                                'screenRadius': {'args': '<type __builtin__.float>',
                                                 'longname': 'screenRadius',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'scR'},
                                'selectclonesource': {'args': '<type __builtin__.bool>',
                                                      'longname': 'selectclonesource',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'scs'},
                                'selectedattroper': {'args': '<type __builtin__.str>',
                                                     'longname': 'selectedattroper',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sao'},
                                'setArrayValueCommand': {'args': '<type __builtin__.str>',
                                                         'longname': 'setArrayValueCommand',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'sac'},
                                'setValueCommand': {'args': '<type __builtin__.str>',
                                                    'longname': 'setValueCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'svc'},
                                'showactive': {'args': '<type __builtin__.bool>',
                                               'longname': 'showactive',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sa'},
                                'stampDepth': {'args': '<type __builtin__.float>',
                                               'longname': 'stampDepth',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'stD'},
                                'stampProfile': {'args': '<type __builtin__.str>',
                                                 'longname': 'stampProfile',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'stP'},
                                'stampSpacing': {'args': '<type __builtin__.float>',
                                                 'longname': 'stampSpacing',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'stS'},
                                'strokesmooth': {'args': '<type __builtin__.str>',
                                                 'longname': 'strokesmooth',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ssm'},
                                'surfaceConformedBrushVertices': {'args': '<type __builtin__.bool>',
                                                                  'longname': 'surfaceConformedBrushVertices',
                                                                  'modes': ['create',
                                                                            'query',
                                                                            'edit'],
                                                                  'numArgs': 1,
                                                                  'shortname': 'scv'},
                                'tablet': {'args': '<type __builtin__.bool>',
                                           'longname': 'tablet',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'tab'},
                                'tangentOutline': {'args': '<type __builtin__.bool>',
                                                   'longname': 'tangentOutline',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'to'},
                                'toolCleanupCmd': {'args': '<type __builtin__.str>',
                                                   'longname': 'toolCleanupCmd',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'tcc'},
                                'toolOffProc': {'args': '<type __builtin__.str>',
                                                'longname': 'toolOffProc',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'tfp'},
                                'toolOnProc': {'args': '<type __builtin__.str>',
                                               'longname': 'toolOnProc',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'top'},
                                'toolSetupCmd': {'args': '<type __builtin__.str>',
                                                 'longname': 'toolSetupCmd',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'tsc'},
                                'useColorRamp': {'args': '<type __builtin__.bool>',
                                                 'longname': 'useColorRamp',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ucr'},
                                'useMaxMinColor': {'args': '<type __builtin__.bool>',
                                                   'longname': 'useMaxMinColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'umc'},
                                'usepressure': {'args': '<type __builtin__.bool>',
                                                'longname': 'usepressure',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'up'},
                                'value': {'args': '<type __builtin__.float>',
                                          'longname': 'value',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'val'},
                                'whichTool': {'args': '<type __builtin__.str>',
                                              'longname': 'whichTool',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'wst'},
                                'worldRadius': {'args': '<type __builtin__.float>',
                                                'longname': 'worldRadius',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'wlR'}},
                      'shortFlags': {'acl': 'alphaclamplower',
                                     'aco': 'accopacity',
                                     'acu': 'alphaclampupper',
                                     'alc': 'alphaclamp',
                                     'alp': 'activeListChangedProc',
                                     'asc': 'afterStrokeCmd',
                                     'asl': 'attrSelected',
                                     'bra': 'brushalignment',
                                     'brf': 'brushfeedback',
                                     'bsc': 'beforeStrokeCmd',
                                     'cc': 'chunkCommand',
                                     'cf': 'colorfeedback',
                                     'cfo': 'colorfeedbackOverride',
                                     'ch': 'history',
                                     'cl': 'clamp',
                                     'cl1': 'colorAlphaValue',
                                     'cl3': 'colorRGBValue',
                                     'cl4': 'colorRGBAValue',
                                     'cll': 'clamplower',
                                     'clr': 'clear',
                                     'clu': 'clampupper',
                                     'cr': 'colorRamp',
                                     'crl': 'colorrangelower',
                                     'cru': 'colorrangeupper',
                                     'dcm': 'dynclonemode',
                                     'dl': 'disablelighting',
                                     'dsk': 'duringStrokeCmd',
                                     'dsl': 'dragSlider',
                                     'dti': 'dataTypeIndex',
                                     'e': 'edit',
                                     'ear': 'exportaspectratio',
                                     'eef': 'expandfilename',
                                     'efm': 'exportfilemode',
                                     'eft': 'exportfiletype',
                                     'esf': 'exportfilesave',
                                     'ex': 'exists',
                                     'fc': 'finalizeCmd',
                                     'fon': 'filterNodes',
                                     'fp': 'fullpaths',
                                     'fsx': 'exportfilesizex',
                                     'fsy': 'exportfilesizey',
                                     'gac': 'getArrayAttrCommand',
                                     'gsc': 'getSurfaceCommand',
                                     'gvc': 'getValueCommand',
                                     'i1': 'image1',
                                     'i2': 'image2',
                                     'i3': 'image3',
                                     'ic': 'initializeCmd',
                                     'ifl': 'importfileload',
                                     'ifm': 'importfilemode',
                                     'irm': 'importreassign',
                                     'iu': 'interactiveUpdate',
                                     'lr': 'lowerradius',
                                     'lrc': 'lastRecorderCmd',
                                     'lsn': 'lastStampName',
                                     'miv': 'minvalue',
                                     'mp': 'mappressure',
                                     'mst': 'makeStroke',
                                     'mxv': 'maxvalue',
                                     'n': 'name',
                                     'o': 'outline',
                                     'oaa': 'objattrArray',
                                     'op': 'opacity',
                                     'owp': 'outwhilepaint',
                                     'pas': 'paintattrselected',
                                     'pcm': 'pickColor',
                                     'pcs': 'preserveclonesource',
                                     'plc': 'playbackCursor',
                                     'plp': 'playbackPressure',
                                     'pm': 'paintmode',
                                     'pna': 'paintNodeArray',
                                     'pot': 'paintoperationtype',
                                     'prm': 'projective',
                                     'psf': 'profileShapeFile',
                                     'pv': 'pickValue',
                                     'q': 'query',
                                     'r': 'radius',
                                     'ra': 'reflectionaxis',
                                     'rec': 'record',
                                     'rmc': 'rampMinColor',
                                     'rn': 'reflection',
                                     'rno': 'reflectionaboutorigin',
                                     'rxc': 'rampMaxColor',
                                     'sa': 'showactive',
                                     'sac': 'setArrayValueCommand',
                                     'sao': 'selectedattroper',
                                     'scR': 'screenRadius',
                                     'scs': 'selectclonesource',
                                     'scv': 'surfaceConformedBrushVertices',
                                     'ssm': 'strokesmooth',
                                     'stD': 'stampDepth',
                                     'stP': 'stampProfile',
                                     'stS': 'stampSpacing',
                                     'svc': 'setValueCommand',
                                     'tab': 'tablet',
                                     'tcc': 'toolCleanupCmd',
                                     'tfp': 'toolOffProc',
                                     'to': 'tangentOutline',
                                     'top': 'toolOnProc',
                                     'tsc': 'toolSetupCmd',
                                     'ucr': 'useColorRamp',
                                     'umc': 'useMaxMinColor',
                                     'up': 'usepressure',
                                     'val': 'value',
                                     'wlR': 'worldRadius',
                                     'wst': 'whichTool'},
                      'type': 'context'},
  'arubaNurbsToPoly': {'flags': {'adaptive': {'args': '<type __builtin__.bool>',
                                              'longname': 'adaptive',
                                              'numArgs': 1,
                                              'shortname': 'adp'},
                                 'caching': {'args': '<type __builtin__.bool>',
                                             'longname': 'caching',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cch'},
                                 'chordalDeviation': {'args': '<type __builtin__.float>',
                                                      'longname': 'chordalDeviation',
                                                      'numArgs': 1,
                                                      'shortname': 'cd'},
                                 'constructionHistory': {'args': '<type __builtin__.bool>',
                                                         'longname': 'constructionHistory',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'ch'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'frozen': {'args': '<type __builtin__.bool>',
                                            'longname': 'frozen',
                                            'numArgs': 1,
                                            'shortname': 'fzn'},
                                 'localSpace': {'args': '<type __builtin__.bool>',
                                                'longname': 'localSpace',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ls'},
                                 'maxChordLength': {'args': '<type __builtin__.float>',
                                                    'longname': 'maxChordLength',
                                                    'numArgs': 1,
                                                    'shortname': 'mcl'},
                                 'minChordLength': {'args': '<type __builtin__.float>',
                                                    'longname': 'minChordLength',
                                                    'numArgs': 1,
                                                    'shortname': 'icl'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'nodeState': {'args': '<type __builtin__.int>',
                                               'longname': 'nodeState',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nds'},
                                 'normalTolerance': {'args': '<type __builtin__.float>',
                                                     'longname': 'normalTolerance',
                                                     'numArgs': 1,
                                                     'shortname': 'ntl'},
                                 'object': {'args': '<type __builtin__.bool>',
                                            'longname': 'object',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'o'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'sampleType': {'args': '<type __builtin__.int>',
                                                'longname': 'sampleType',
                                                'numArgs': 1,
                                                'shortname': 'st'},
                                 'samples': {'args': '<type __builtin__.int>',
                                             'longname': 'samples',
                                             'numArgs': 1,
                                             'shortname': 'smp'},
                                 'tolerance': {'args': '<type __builtin__.float>',
                                               'longname': 'tolerance',
                                               'numArgs': 1,
                                               'shortname': 'tol'},
                                 'useSurfaceShader': {'args': '<type __builtin__.bool>',
                                                      'longname': 'useSurfaceShader',
                                                      'numArgs': 1,
                                                      'shortname': 'uss'},
                                 'worldSpace': {'args': '<type __builtin__.bool>',
                                                'longname': 'worldSpace',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ws'}},
                       'shortFlags': {'adp': 'adaptive',
                                      'cch': 'caching',
                                      'cd': 'chordalDeviation',
                                      'ch': 'constructionHistory',
                                      'e': 'edit',
                                      'fzn': 'frozen',
                                      'icl': 'minChordLength',
                                      'ls': 'localSpace',
                                      'mcl': 'maxChordLength',
                                      'n': 'name',
                                      'nds': 'nodeState',
                                      'ntl': 'normalTolerance',
                                      'o': 'object',
                                      'q': 'query',
                                      'smp': 'samples',
                                      'st': 'sampleType',
                                      'tol': 'tolerance',
                                      'uss': 'useSurfaceShader',
                                      'ws': 'worldSpace'},
                       'type': 'modeling'},
  'assembly': {'callbackFlags': ['createOptionBoxProc',
                                 'cob',
                                 'repTypeLabelProc',
                                 'rtp',
                                 'listRepTypesProc',
                                 'lrp',
                                 'proc',
                                 'prc',
                                 'postCreateUIProc',
                                 'aoc'],
               'flags': {'active': {'args': '<type __builtin__.str>',
                                    'longname': 'active',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'a'},
                         'activeLabel': {'args': '<type __builtin__.str>',
                                         'longname': 'activeLabel',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'al'},
                         'canCreate': {'args': '<type __builtin__.str>',
                                       'longname': 'canCreate',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'cc'},
                         'createOptionBoxProc': {'args': 'script',
                                                 'longname': 'createOptionBoxProc',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cob'},
                         'createRepresentation': {'args': '<type __builtin__.str>',
                                                  'longname': 'createRepresentation',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cr'},
                         'defaultType': {'args': '<type __builtin__.str>',
                                         'longname': 'defaultType',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dt'},
                         'deleteRepresentation': {'args': '<type __builtin__.str>',
                                                  'longname': 'deleteRepresentation',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dr'},
                         'deregister': {'args': '<type __builtin__.str>',
                                        'longname': 'deregister',
                                        'modes': ['edit'],
                                        'numArgs': 1,
                                        'shortname': 'd'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'input': {'args': '<type __builtin__.str>',
                                   'longname': 'input',
                                   'modes': ['edit'],
                                   'numArgs': 1,
                                   'shortname': 'input'},
                         'isAType': {'args': '<type __builtin__.str>',
                                     'longname': 'isAType',
                                     'modes': ['query'],
                                     'numArgs': 1,
                                     'shortname': 'isa'},
                         'isTrackingMemberEdits': {'args': '<type __builtin__.str>',
                                                   'longname': 'isTrackingMemberEdits',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ite'},
                         'label': {'args': '<type __builtin__.str>',
                                   'longname': 'label',
                                   'modes': ['query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'lbl'},
                         'listRepTypes': {'args': '<type __builtin__.bool>',
                                          'longname': 'listRepTypes',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'lrt'},
                         'listRepTypesProc': {'args': 'script',
                                              'longname': 'listRepTypesProc',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lrp'},
                         'listRepresentations': {'args': '<type __builtin__.bool>',
                                                 'longname': 'listRepresentations',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'lr'},
                         'listTypes': {'args': '<type __builtin__.bool>',
                                       'longname': 'listTypes',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'lt'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'newRepLabel': {'args': '<type __builtin__.str>',
                                         'longname': 'newRepLabel',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 'nrl'},
                         'postCreateUIProc': {'args': 'script',
                                              'longname': 'postCreateUIProc',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'aoc'},
                         'proc': {'args': 'script',
                                  'longname': 'proc',
                                  'modes': ['edit'],
                                  'numArgs': 1,
                                  'shortname': 'prc'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'renameRepresentation': {'args': '<type __builtin__.str>',
                                                  'longname': 'renameRepresentation',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'rnr'},
                         'repLabel': {'args': '<type __builtin__.str>',
                                      'longname': 'repLabel',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'rl'},
                         'repName': {'args': '<type __builtin__.str>',
                                     'longname': 'repName',
                                     'modes': ['edit'],
                                     'numArgs': 1,
                                     'shortname': 'rnm'},
                         'repNamespace': {'args': '<type __builtin__.str>',
                                          'longname': 'repNamespace',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'rns'},
                         'repPostCreateUIProc': {'args': '<type __builtin__.str>',
                                                 'longname': 'repPostCreateUIProc',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'poc'},
                         'repPreCreateUIProc': {'args': '<type __builtin__.str>',
                                                'longname': 'repPreCreateUIProc',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'pec'},
                         'repType': {'args': '<type __builtin__.str>',
                                     'longname': 'repType',
                                     'modes': ['query'],
                                     'numArgs': 1,
                                     'shortname': 'rt'},
                         'repTypeLabel': {'args': '<type __builtin__.str>',
                                          'longname': 'repTypeLabel',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'rtl'},
                         'repTypeLabelProc': {'args': 'script',
                                              'longname': 'repTypeLabelProc',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'rtp'},
                         'type': {'args': '<type __builtin__.str>',
                                  'longname': 'type',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'typ'}},
               'removedFlags': {'in': 'input'},
               'shortFlags': {'a': 'active',
                              'al': 'activeLabel',
                              'aoc': 'postCreateUIProc',
                              'cc': 'canCreate',
                              'cob': 'createOptionBoxProc',
                              'cr': 'createRepresentation',
                              'd': 'deregister',
                              'dr': 'deleteRepresentation',
                              'dt': 'defaultType',
                              'e': 'edit',
                              'input': 'input',
                              'isa': 'isAType',
                              'ite': 'isTrackingMemberEdits',
                              'lbl': 'label',
                              'lr': 'listRepresentations',
                              'lrp': 'listRepTypesProc',
                              'lrt': 'listRepTypes',
                              'lt': 'listTypes',
                              'n': 'name',
                              'nrl': 'newRepLabel',
                              'pec': 'repPreCreateUIProc',
                              'poc': 'repPostCreateUIProc',
                              'prc': 'proc',
                              'q': 'query',
                              'rl': 'repLabel',
                              'rnm': 'repName',
                              'rnr': 'renameRepresentation',
                              'rns': 'repNamespace',
                              'rt': 'repType',
                              'rtl': 'repTypeLabel',
                              'rtp': 'repTypeLabelProc',
                              'typ': 'type'},
               'type': 'general'},
  'assignCommand': {'callbackFlags': ['enableCommandRepeat',
                                      'ecr',
                                      'sourceUserCommands',
                                      'suc',
                                      'commandModifier',
                                      'cmd',
                                      'command',
                                      'c'],
                    'flags': {'addDivider': {'args': '<type __builtin__.str>',
                                             'longname': 'addDivider',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'ad'},
                              'altModifier': {'args': '<type __builtin__.bool>',
                                              'longname': 'altModifier',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'alt'},
                              'annotation': {'args': '<type __builtin__.str>',
                                             'longname': 'annotation',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ann'},
                              'command': {'args': 'script',
                                          'longname': 'command',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'c'},
                              'commandModifier': {'args': '<type __builtin__.bool>',
                                                  'longname': 'commandModifier',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'cmd'},
                              'ctrlModifier': {'args': '<type __builtin__.bool>',
                                               'longname': 'ctrlModifier',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'ctl'},
                              'data1': {'args': '<type __builtin__.str>',
                                        'longname': 'data1',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'da1'},
                              'data2': {'args': '<type __builtin__.str>',
                                        'longname': 'data2',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'da2'},
                              'data3': {'args': '<type __builtin__.str>',
                                        'longname': 'data3',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'da3'},
                              'delete': {'args': '<type __builtin__.int>',
                                         'longname': 'delete',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 'd'},
                              'dividerString': {'args': '<type __builtin__.str>',
                                                'longname': 'dividerString',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'ds'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'enableCommandRepeat': {'args': '<type __builtin__.bool>',
                                                      'longname': 'enableCommandRepeat',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ecr'},
                              'factorySettings': {'args': '<type __builtin__.bool>',
                                                  'longname': 'factorySettings',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'fs'},
                              'index': {'args': '<type __builtin__.int>',
                                        'longname': 'index',
                                        'modes': ['edit'],
                                        'numArgs': 1,
                                        'shortname': 'i'},
                              'keyArray': {'args': '<type __builtin__.bool>',
                                           'longname': 'keyArray',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'ka'},
                              'keyString': {'args': '<type __builtin__.str>',
                                            'longname': 'keyString',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'k'},
                              'keyUp': {'args': '<type __builtin__.bool>',
                                        'longname': 'keyUp',
                                        'modes': ['edit'],
                                        'numArgs': 0,
                                        'shortname': 'kup'},
                              'name': {'args': '<type __builtin__.bool>',
                                       'longname': 'name',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'n'},
                              'numDividersPreceding': {'args': '<type __builtin__.int>',
                                                       'longname': 'numDividersPreceding',
                                                       'modes': ['query'],
                                                       'numArgs': 1,
                                                       'shortname': 'ndp'},
                              'numElements': {'args': '<type __builtin__.bool>',
                                              'longname': 'numElements',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'num'},
                              'optionModifier': {'args': '<type __builtin__.bool>',
                                                 'longname': 'optionModifier',
                                                 'modes': ['edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'opt'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'sortByKey': {'args': '<type __builtin__.bool>',
                                            'longname': 'sortByKey',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sbk'},
                              'sourceUserCommands': {'args': '<type __builtin__.bool>',
                                                     'longname': 'sourceUserCommands',
                                                     'modes': ['edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'suc'}},
                    'shortFlags': {'ad': 'addDivider',
                                   'alt': 'altModifier',
                                   'ann': 'annotation',
                                   'c': 'command',
                                   'cmd': 'commandModifier',
                                   'ctl': 'ctrlModifier',
                                   'd': 'delete',
                                   'da1': 'data1',
                                   'da2': 'data2',
                                   'da3': 'data3',
                                   'ds': 'dividerString',
                                   'e': 'edit',
                                   'ecr': 'enableCommandRepeat',
                                   'fs': 'factorySettings',
                                   'i': 'index',
                                   'k': 'keyString',
                                   'ka': 'keyArray',
                                   'kup': 'keyUp',
                                   'n': 'name',
                                   'ndp': 'numDividersPreceding',
                                   'num': 'numElements',
                                   'opt': 'optionModifier',
                                   'q': 'query',
                                   'sbk': 'sortByKey',
                                   'suc': 'sourceUserCommands'},
                    'type': 'language'},
  'assignInputDevice': {'flags': {'clutch': {'args': '<type __builtin__.str>',
                                             'longname': 'clutch',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'c'},
                                  'continuous': {'args': '<type __builtin__.bool>',
                                                 'longname': 'continuous',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ct'},
                                  'device': {'args': '<type __builtin__.str>',
                                             'longname': 'device',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'd'},
                                  'immediate': {'args': '<type __builtin__.bool>',
                                                'longname': 'immediate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'im'},
                                  'multiple': {'args': '<type __builtin__.bool>',
                                               'longname': 'multiple',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'm'}},
                        'shortFlags': {},
                        'type': 'system'},
  'assignViewportFactories': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                 'longname': 'edit',
                                                 'numArgs': 0,
                                                 'shortname': 'e'},
                                        'materialFactory': {'args': '<type __builtin__.str>',
                                                            'longname': 'materialFactory',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'mf'},
                                        'nodeType': {'args': '<type __builtin__.str>',
                                                     'longname': 'nodeType',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'nt'},
                                        'query': {'args': '<type __builtin__.bool>',
                                                  'longname': 'query',
                                                  'numArgs': 0,
                                                  'shortname': 'q'},
                                        'textureFactory': {'args': '<type __builtin__.str>',
                                                           'longname': 'textureFactory',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'tf'}},
                              'shortFlags': {'e': 'edit',
                                             'mf': 'materialFactory',
                                             'nt': 'nodeType',
                                             'q': 'query',
                                             'tf': 'textureFactory'},
                              'type': 'rendering'},
  'attachCache': {'type': 'runtime'},
  'attachCurve': {'flags': {'blendBias': {'args': '<type __builtin__.float>',
                                          'longname': 'blendBias',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'bb'},
                            'blendKnotInsertion': {'args': '<type __builtin__.bool>',
                                                   'longname': 'blendKnotInsertion',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'bki'},
                            'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'keepMultipleKnots': {'args': '<type __builtin__.bool>',
                                                  'longname': 'keepMultipleKnots',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'kmk'},
                            'method': {'args': '<type __builtin__.int>',
                                       'longname': 'method',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'parameter': {'args': '<type __builtin__.float>',
                                          'longname': 'parameter',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                'longname': 'replaceOriginal',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'rpo'},
                            'reverse1': {'args': '<type __builtin__.bool>',
                                         'longname': 'reverse1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'rv1'},
                            'reverse2': {'args': '<type __builtin__.bool>',
                                         'longname': 'reverse2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'rv2'}},
                  'shortFlags': {'bb': 'blendBias',
                                 'bki': 'blendKnotInsertion',
                                 'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'e': 'edit',
                                 'fzn': 'frozen',
                                 'kmk': 'keepMultipleKnots',
                                 'm': 'method',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'o': 'object',
                                 'p': 'parameter',
                                 'q': 'query',
                                 'rpo': 'replaceOriginal',
                                 'rv1': 'reverse1',
                                 'rv2': 'reverse2'},
                  'type': 'modeling'},
  'attachDeviceAttr': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                               'longname': 'attribute',
                                               'modes': ['create',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'at'},
                                 'axis': {'args': '<type __builtin__.str>',
                                          'longname': 'axis',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ax'},
                                 'camera': {'args': '<type __builtin__.bool>',
                                            'longname': 'camera',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'cam'},
                                 'cameraRotate': {'args': '<type __builtin__.bool>',
                                                  'longname': 'cameraRotate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'cr'},
                                 'cameraTranslate': {'args': '<type __builtin__.bool>',
                                                     'longname': 'cameraTranslate',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ct'},
                                 'clutch': {'args': '<type __builtin__.str>',
                                            'longname': 'clutch',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'c'},
                                 'device': {'args': '<type __builtin__.str>',
                                            'longname': 'device',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'd'},
                                 'selection': {'args': '<type __builtin__.bool>',
                                               'longname': 'selection',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'sl'}},
                       'shortFlags': {},
                       'type': 'system'},
  'attachFluidCache': {'type': 'runtime'},
  'attachGeometryCache': {'type': 'runtime'},
  'attachNclothCache': {'type': 'runtime'},
  'attachSurface': {'flags': {'blendBias': {'args': '<type __builtin__.float>',
                                            'longname': 'blendBias',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'bb'},
                              'blendKnotInsertion': {'args': '<type __builtin__.bool>',
                                                     'longname': 'blendKnotInsertion',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'bki'},
                              'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cch'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'directionU': {'args': '<type __builtin__.bool>',
                                             'longname': 'directionU',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'du'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'keepMultipleKnots': {'args': '<type __builtin__.bool>',
                                                    'longname': 'keepMultipleKnots',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'kmk'},
                              'method': {'args': '<type __builtin__.int>',
                                         'longname': 'method',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nds'},
                              'object': {'args': '<type __builtin__.bool>',
                                         'longname': 'object',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'o'},
                              'parameter': {'args': '<type __builtin__.float>',
                                            'longname': 'parameter',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'p'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                  'longname': 'replaceOriginal',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'rpo'},
                              'reverse1': {'args': '<type __builtin__.bool>',
                                           'longname': 'reverse1',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'rv1'},
                              'reverse2': {'args': '<type __builtin__.bool>',
                                           'longname': 'reverse2',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'rv2'},
                              'swap1': {'args': '<type __builtin__.bool>',
                                        'longname': 'swap1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'sw1'},
                              'swap2': {'args': '<type __builtin__.bool>',
                                        'longname': 'swap2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'sw2'},
                              'twist': {'args': '<type __builtin__.bool>',
                                        'longname': 'twist',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'tw'}},
                    'shortFlags': {'bb': 'blendBias',
                                   'bki': 'blendKnotInsertion',
                                   'cch': 'caching',
                                   'ch': 'constructionHistory',
                                   'du': 'directionU',
                                   'e': 'edit',
                                   'fzn': 'frozen',
                                   'kmk': 'keepMultipleKnots',
                                   'm': 'method',
                                   'n': 'name',
                                   'nds': 'nodeState',
                                   'o': 'object',
                                   'p': 'parameter',
                                   'q': 'query',
                                   'rpo': 'replaceOriginal',
                                   'rv1': 'reverse1',
                                   'rv2': 'reverse2',
                                   'sw1': 'swap1',
                                   'sw2': 'swap2',
                                   'tw': 'twist'},
                    'type': 'modeling'},
  'attrColorSliderGrp': {'callbackFlags': ['visibleChangeCommand',
                                           'vcc',
                                           'dragCallback',
                                           'dgc',
                                           'dropCallback',
                                           'dpc'],
                         'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                        'longname': 'adjustableColumn',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'adj'},
                                   'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn2',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad2'},
                                   'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn3',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad3'},
                                   'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn4',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad4'},
                                   'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn5',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad5'},
                                   'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn6',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad6'},
                                   'annotation': {'args': '<type __builtin__.str>',
                                                  'longname': 'annotation',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ann'},
                                   'attrNavDecision': {'args': ['PyNode',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'attrNavDecision',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 2,
                                                       'shortname': 'attrNavDecision'},
                                   'attribute': {'args': '<type __builtin__.str>',
                                                 'longname': 'attribute',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'at'},
                                   'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'backgroundColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'bgc'},
                                   'columnAlign': {'args': ['<type __builtin__.int>',
                                                            '<type __builtin__.str>'],
                                                   'longname': 'columnAlign',
                                                   'modes': ['create',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 2,
                                                   'shortname': 'cal'},
                                   'columnAlign2': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign2',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 2,
                                                    'shortname': 'cl2'},
                                   'columnAlign3': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign3',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'cl3'},
                                   'columnAlign4': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign4',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 4,
                                                    'shortname': 'cl4'},
                                   'columnAlign5': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign5',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 5,
                                                    'shortname': 'cl5'},
                                   'columnAlign6': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign6',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 6,
                                                    'shortname': 'cl6'},
                                   'columnAttach': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnAttach',
                                                    'modes': ['create',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 3,
                                                    'shortname': 'cat'},
                                   'columnAttach2': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach2',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'ct2'},
                                   'columnAttach3': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach3',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'ct3'},
                                   'columnAttach4': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach4',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 4,
                                                     'shortname': 'ct4'},
                                   'columnAttach5': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach5',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 5,
                                                     'shortname': 'ct5'},
                                   'columnAttach6': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach6',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 6,
                                                     'shortname': 'ct6'},
                                   'columnOffset2': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset2',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'co2'},
                                   'columnOffset3': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset3',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'co3'},
                                   'columnOffset4': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset4',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 4,
                                                     'shortname': 'co4'},
                                   'columnOffset5': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset5',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 5,
                                                     'shortname': 'co5'},
                                   'columnOffset6': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset6',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 6,
                                                     'shortname': 'co6'},
                                   'columnWidth': {'args': ['<type __builtin__.int>',
                                                            '<type __builtin__.int>'],
                                                   'longname': 'columnWidth',
                                                   'modes': ['create',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 2,
                                                   'shortname': 'cw'},
                                   'columnWidth1': {'args': '<type __builtin__.int>',
                                                    'longname': 'columnWidth1',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'cw1'},
                                   'columnWidth2': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth2',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 2,
                                                    'shortname': 'cw2'},
                                   'columnWidth3': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth3',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'cw3'},
                                   'columnWidth4': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth4',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 4,
                                                    'shortname': 'cw4'},
                                   'columnWidth5': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth5',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 5,
                                                    'shortname': 'cw5'},
                                   'columnWidth6': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth6',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 6,
                                                    'shortname': 'cw6'},
                                   'defineTemplate': {'args': '<type __builtin__.str>',
                                                      'longname': 'defineTemplate',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'dt'},
                                   'docTag': {'args': '<type __builtin__.str>',
                                              'longname': 'docTag',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dtg'},
                                   'dragCallback': {'args': 'script',
                                                    'longname': 'dragCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dgc'},
                                   'dropCallback': {'args': 'script',
                                                    'longname': 'dropCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dpc'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'enable': {'args': '<type __builtin__.bool>',
                                              'longname': 'enable',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'en'},
                                   'enableBackground': {'args': '<type __builtin__.bool>',
                                                        'longname': 'enableBackground',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ebg'},
                                   'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                           'longname': 'enableKeyboardFocus',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ekf'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'fullPathName': {'args': '<type __builtin__.bool>',
                                                    'longname': 'fullPathName',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'fpn'},
                                   'height': {'args': '<type __builtin__.int>',
                                              'longname': 'height',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'h'},
                                   'highlightColor': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'highlightColor',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'hlc'},
                                   'hsvValue': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'hsvValue',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hsv'},
                                   'isObscured': {'args': '<type __builtin__.bool>',
                                                  'longname': 'isObscured',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'io'},
                                   'label': {'args': '<type __builtin__.str>',
                                             'longname': 'label',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'l'},
                                   'manage': {'args': '<type __builtin__.bool>',
                                              'longname': 'manage',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'm'},
                                   'noBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'noBackground',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'nbg'},
                                   'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                          'longname': 'numberOfPopupMenus',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'npm'},
                                   'parent': {'args': '<type __builtin__.str>',
                                              'longname': 'parent',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'p'},
                                   'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                      'longname': 'popupMenuArray',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'pma'},
                                   'preventOverride': {'args': '<type __builtin__.bool>',
                                                       'longname': 'preventOverride',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'po'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'rgbValue': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'rgbValue',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'rgb'},
                                   'rowAttach': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'rowAttach',
                                                 'modes': ['create',
                                                           'edit',
                                                           'multiuse'],
                                                 'numArgs': 3,
                                                 'shortname': 'rat'},
                                   'showButton': {'args': '<type __builtin__.bool>',
                                                  'longname': 'showButton',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sb'},
                                   'statusBarMessage': {'args': '<type __builtin__.str>',
                                                        'longname': 'statusBarMessage',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'sbm'},
                                   'useTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'useTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ut'},
                                   'visible': {'args': '<type __builtin__.bool>',
                                               'longname': 'visible',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'vis'},
                                   'visibleChangeCommand': {'args': 'script',
                                                            'longname': 'visibleChangeCommand',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'vcc'},
                                   'width': {'args': '<type __builtin__.int>',
                                             'longname': 'width',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'w'}},
                         'removedFlags': {'and': 'attrNavDecision'},
                         'shortFlags': {'ad2': 'adjustableColumn2',
                                        'ad3': 'adjustableColumn3',
                                        'ad4': 'adjustableColumn4',
                                        'ad5': 'adjustableColumn5',
                                        'ad6': 'adjustableColumn6',
                                        'adj': 'adjustableColumn',
                                        'ann': 'annotation',
                                        'at': 'attribute',
                                        'attrNavDecision': 'attrNavDecision',
                                        'bgc': 'backgroundColor',
                                        'cal': 'columnAlign',
                                        'cat': 'columnAttach',
                                        'cl2': 'columnAlign2',
                                        'cl3': 'columnAlign3',
                                        'cl4': 'columnAlign4',
                                        'cl5': 'columnAlign5',
                                        'cl6': 'columnAlign6',
                                        'co2': 'columnOffset2',
                                        'co3': 'columnOffset3',
                                        'co4': 'columnOffset4',
                                        'co5': 'columnOffset5',
                                        'co6': 'columnOffset6',
                                        'ct2': 'columnAttach2',
                                        'ct3': 'columnAttach3',
                                        'ct4': 'columnAttach4',
                                        'ct5': 'columnAttach5',
                                        'ct6': 'columnAttach6',
                                        'cw': 'columnWidth',
                                        'cw1': 'columnWidth1',
                                        'cw2': 'columnWidth2',
                                        'cw3': 'columnWidth3',
                                        'cw4': 'columnWidth4',
                                        'cw5': 'columnWidth5',
                                        'cw6': 'columnWidth6',
                                        'dgc': 'dragCallback',
                                        'dpc': 'dropCallback',
                                        'dt': 'defineTemplate',
                                        'dtg': 'docTag',
                                        'e': 'edit',
                                        'ebg': 'enableBackground',
                                        'ekf': 'enableKeyboardFocus',
                                        'en': 'enable',
                                        'ex': 'exists',
                                        'fpn': 'fullPathName',
                                        'h': 'height',
                                        'hlc': 'highlightColor',
                                        'hsv': 'hsvValue',
                                        'io': 'isObscured',
                                        'l': 'label',
                                        'm': 'manage',
                                        'nbg': 'noBackground',
                                        'npm': 'numberOfPopupMenus',
                                        'p': 'parent',
                                        'pma': 'popupMenuArray',
                                        'po': 'preventOverride',
                                        'q': 'query',
                                        'rat': 'rowAttach',
                                        'rgb': 'rgbValue',
                                        'sb': 'showButton',
                                        'sbm': 'statusBarMessage',
                                        'ut': 'useTemplate',
                                        'vcc': 'visibleChangeCommand',
                                        'vis': 'visible',
                                        'w': 'width'},
                         'type': 'windows'},
  'attrCompatibility': {'flags': {'addAttr': {'args': '<type __builtin__.bool>',
                                              'longname': 'addAttr',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'a'},
                                  'clear': {'args': '<type __builtin__.bool>',
                                            'longname': 'clear',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'clr'},
                                  'dumpTable': {'args': '<type __builtin__.bool>',
                                                'longname': 'dumpTable',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'dmp'},
                                  'enable': {'args': '<type __builtin__.bool>',
                                             'longname': 'enable',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'e'},
                                  'nodeRename': {'args': '<type __builtin__.str>',
                                                 'longname': 'nodeRename',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'nr'},
                                  'pluginNode': {'args': '<type __builtin__.str>',
                                                 'longname': 'pluginNode',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'pn'},
                                  'removeAttr': {'args': '<type __builtin__.bool>',
                                                 'longname': 'removeAttr',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'rm'},
                                  'renameAttr': {'args': '<type __builtin__.str>',
                                                 'longname': 'renameAttr',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'r'},
                                  'type': {'args': '<type __builtin__.str>',
                                           'longname': 'type',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'typ'},
                                  'version': {'args': '<type __builtin__.str>',
                                              'longname': 'version',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'v'}},
                        'shortFlags': {'a': 'addAttr',
                                       'clr': 'clear',
                                       'dmp': 'dumpTable',
                                       'e': 'enable',
                                       'nr': 'nodeRename',
                                       'pn': 'pluginNode',
                                       'r': 'renameAttr',
                                       'rm': 'removeAttr',
                                       'typ': 'type',
                                       'v': 'version'},
                        'type': 'system'},
  'attrControlGrp': {'callbackFlags': ['changeCommand', 'cc'],
                     'flags': {'annotation': {'args': '<type __builtin__.str>',
                                              'longname': 'annotation',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ann'},
                               'attribute': {'args': 'PyNode',
                                             'longname': 'attribute',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'a'},
                               'changeCommand': {'args': 'script',
                                                 'longname': 'changeCommand',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cc'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'enable': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'handlesAttribute': {'args': 'PyNode',
                                                    'longname': 'handlesAttribute',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ha'},
                               'hideMapButton': {'args': '<type __builtin__.bool>',
                                                 'longname': 'hideMapButton',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'hmb'},
                               'label': {'args': '<type __builtin__.str>',
                                         'longname': 'label',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'l'},
                               'preventOverride': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preventOverride',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'po'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'}},
                     'shortFlags': {'a': 'attribute',
                                    'ann': 'annotation',
                                    'cc': 'changeCommand',
                                    'e': 'edit',
                                    'en': 'enable',
                                    'ex': 'exists',
                                    'ha': 'handlesAttribute',
                                    'hmb': 'hideMapButton',
                                    'l': 'label',
                                    'po': 'preventOverride',
                                    'q': 'query'},
                     'type': 'windows'},
  'attrEnumOptionMenu': {'callbackFlags': ['dropCallback',
                                           'dpc',
                                           'visibleChangeCommand',
                                           'vcc',
                                           'dragCallback',
                                           'dgc',
                                           'changeCommand',
                                           'cc'],
                         'flags': {'annotation': {'args': '<type __builtin__.str>',
                                                  'longname': 'annotation',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ann'},
                                   'attribute': {'args': 'PyNode',
                                                 'longname': 'attribute',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'at'},
                                   'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'backgroundColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'bgc'},
                                   'changeCommand': {'args': 'script',
                                                     'longname': 'changeCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'cc'},
                                   'defineTemplate': {'args': '<type __builtin__.str>',
                                                      'longname': 'defineTemplate',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'dt'},
                                   'docTag': {'args': '<type __builtin__.str>',
                                              'longname': 'docTag',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dtg'},
                                   'dragCallback': {'args': 'script',
                                                    'longname': 'dragCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dgc'},
                                   'dropCallback': {'args': 'script',
                                                    'longname': 'dropCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dpc'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'enable': {'args': '<type __builtin__.bool>',
                                              'longname': 'enable',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'en'},
                                   'enableBackground': {'args': '<type __builtin__.bool>',
                                                        'longname': 'enableBackground',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ebg'},
                                   'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                           'longname': 'enableKeyboardFocus',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ekf'},
                                   'enumeratedItem': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'enumeratedItem',
                                                      'modes': ['create',
                                                                'multiuse'],
                                                      'numArgs': 2,
                                                      'shortname': 'ei'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'fullPathName': {'args': '<type __builtin__.bool>',
                                                    'longname': 'fullPathName',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'fpn'},
                                   'height': {'args': '<type __builtin__.int>',
                                              'longname': 'height',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'h'},
                                   'highlightColor': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'highlightColor',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'hlc'},
                                   'isObscured': {'args': '<type __builtin__.bool>',
                                                  'longname': 'isObscured',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'io'},
                                   'label': {'args': '<type __builtin__.str>',
                                             'longname': 'label',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'l'},
                                   'manage': {'args': '<type __builtin__.bool>',
                                              'longname': 'manage',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'm'},
                                   'noBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'noBackground',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'nbg'},
                                   'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                          'longname': 'numberOfPopupMenus',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'npm'},
                                   'parent': {'args': '<type __builtin__.str>',
                                              'longname': 'parent',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'p'},
                                   'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                      'longname': 'popupMenuArray',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'pma'},
                                   'preventOverride': {'args': '<type __builtin__.bool>',
                                                       'longname': 'preventOverride',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'po'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'statusBarMessage': {'args': '<type __builtin__.str>',
                                                        'longname': 'statusBarMessage',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'sbm'},
                                   'useTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'useTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ut'},
                                   'visible': {'args': '<type __builtin__.bool>',
                                               'longname': 'visible',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'vis'},
                                   'visibleChangeCommand': {'args': 'script',
                                                            'longname': 'visibleChangeCommand',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'vcc'},
                                   'width': {'args': '<type __builtin__.int>',
                                             'longname': 'width',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'w'}},
                         'shortFlags': {'ann': 'annotation',
                                        'at': 'attribute',
                                        'bgc': 'backgroundColor',
                                        'cc': 'changeCommand',
                                        'dgc': 'dragCallback',
                                        'dpc': 'dropCallback',
                                        'dt': 'defineTemplate',
                                        'dtg': 'docTag',
                                        'e': 'edit',
                                        'ebg': 'enableBackground',
                                        'ei': 'enumeratedItem',
                                        'ekf': 'enableKeyboardFocus',
                                        'en': 'enable',
                                        'ex': 'exists',
                                        'fpn': 'fullPathName',
                                        'h': 'height',
                                        'hlc': 'highlightColor',
                                        'io': 'isObscured',
                                        'l': 'label',
                                        'm': 'manage',
                                        'nbg': 'noBackground',
                                        'npm': 'numberOfPopupMenus',
                                        'p': 'parent',
                                        'pma': 'popupMenuArray',
                                        'po': 'preventOverride',
                                        'q': 'query',
                                        'sbm': 'statusBarMessage',
                                        'ut': 'useTemplate',
                                        'vcc': 'visibleChangeCommand',
                                        'vis': 'visible',
                                        'w': 'width'},
                         'type': 'windows'},
  'attrEnumOptionMenuGrp': {'callbackFlags': ['dropCallback',
                                              'dpc',
                                              'visibleChangeCommand',
                                              'vcc',
                                              'dragCallback',
                                              'dgc'],
                            'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                           'longname': 'adjustableColumn',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'adj'},
                                      'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                            'longname': 'adjustableColumn2',
                                                            'modes': ['create',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ad2'},
                                      'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                            'longname': 'adjustableColumn3',
                                                            'modes': ['create',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ad3'},
                                      'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                            'longname': 'adjustableColumn4',
                                                            'modes': ['create',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ad4'},
                                      'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                            'longname': 'adjustableColumn5',
                                                            'modes': ['create',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ad5'},
                                      'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                            'longname': 'adjustableColumn6',
                                                            'modes': ['create',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ad6'},
                                      'annotation': {'args': '<type __builtin__.str>',
                                                     'longname': 'annotation',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ann'},
                                      'attribute': {'args': 'PyNode',
                                                    'longname': 'attribute',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'at'},
                                      'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                   '<type __builtin__.float>',
                                                                   '<type __builtin__.float>'],
                                                          'longname': 'backgroundColor',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 3,
                                                          'shortname': 'bgc'},
                                      'columnAlign': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAlign',
                                                      'modes': ['create',
                                                                'edit',
                                                                'multiuse'],
                                                      'numArgs': 2,
                                                      'shortname': 'cal'},
                                      'columnAlign2': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAlign2',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 2,
                                                       'shortname': 'cl2'},
                                      'columnAlign3': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAlign3',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'cl3'},
                                      'columnAlign4': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAlign4',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 4,
                                                       'shortname': 'cl4'},
                                      'columnAlign5': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAlign5',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 5,
                                                       'shortname': 'cl5'},
                                      'columnAlign6': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAlign6',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 6,
                                                       'shortname': 'cl6'},
                                      'columnAttach': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnAttach',
                                                       'modes': ['create',
                                                                 'edit',
                                                                 'multiuse'],
                                                       'numArgs': 3,
                                                       'shortname': 'cat'},
                                      'columnAttach2': {'args': ['<type __builtin__.str>',
                                                                 '<type __builtin__.str>'],
                                                        'longname': 'columnAttach2',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 2,
                                                        'shortname': 'ct2'},
                                      'columnAttach3': {'args': ['<type __builtin__.str>',
                                                                 '<type __builtin__.str>',
                                                                 '<type __builtin__.str>'],
                                                        'longname': 'columnAttach3',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 3,
                                                        'shortname': 'ct3'},
                                      'columnAttach4': {'args': ['<type __builtin__.str>',
                                                                 '<type __builtin__.str>',
                                                                 '<type __builtin__.str>',
                                                                 '<type __builtin__.str>'],
                                                        'longname': 'columnAttach4',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 4,
                                                        'shortname': 'ct4'},
                                      'columnAttach5': {'args': ['<type __builtin__.str>',
                                                                 '<type __builtin__.str>',
                                                                 '<type __builtin__.str>',
                                                                 '<type __builtin__.str>',
                                                                 '<type __builtin__.str>'],
                                                        'longname': 'columnAttach5',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 5,
                                                        'shortname': 'ct5'},
                                      'columnAttach6': {'args': ['<type __builtin__.str>',
                                                                 '<type __builtin__.str>',
                                                                 '<type __builtin__.str>',
                                                                 '<type __builtin__.str>',
                                                                 '<type __builtin__.str>',
                                                                 '<type __builtin__.str>'],
                                                        'longname': 'columnAttach6',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 6,
                                                        'shortname': 'ct6'},
                                      'columnOffset2': {'args': ['<type __builtin__.int>',
                                                                 '<type __builtin__.int>'],
                                                        'longname': 'columnOffset2',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 2,
                                                        'shortname': 'co2'},
                                      'columnOffset3': {'args': ['<type __builtin__.int>',
                                                                 '<type __builtin__.int>',
                                                                 '<type __builtin__.int>'],
                                                        'longname': 'columnOffset3',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 3,
                                                        'shortname': 'co3'},
                                      'columnOffset4': {'args': ['<type __builtin__.int>',
                                                                 '<type __builtin__.int>',
                                                                 '<type __builtin__.int>',
                                                                 '<type __builtin__.int>'],
                                                        'longname': 'columnOffset4',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 4,
                                                        'shortname': 'co4'},
                                      'columnOffset5': {'args': ['<type __builtin__.int>',
                                                                 '<type __builtin__.int>',
                                                                 '<type __builtin__.int>',
                                                                 '<type __builtin__.int>',
                                                                 '<type __builtin__.int>'],
                                                        'longname': 'columnOffset5',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 5,
                                                        'shortname': 'co5'},
                                      'columnOffset6': {'args': ['<type __builtin__.int>',
                                                                 '<type __builtin__.int>',
                                                                 '<type __builtin__.int>',
                                                                 '<type __builtin__.int>',
                                                                 '<type __builtin__.int>',
                                                                 '<type __builtin__.int>'],
                                                        'longname': 'columnOffset6',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 6,
                                                        'shortname': 'co6'},
                                      'columnWidth': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnWidth',
                                                      'modes': ['create',
                                                                'edit',
                                                                'multiuse'],
                                                      'numArgs': 2,
                                                      'shortname': 'cw'},
                                      'columnWidth1': {'args': '<type __builtin__.int>',
                                                       'longname': 'columnWidth1',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'cw1'},
                                      'columnWidth2': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnWidth2',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 2,
                                                       'shortname': 'cw2'},
                                      'columnWidth3': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnWidth3',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'cw3'},
                                      'columnWidth4': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnWidth4',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 4,
                                                       'shortname': 'cw4'},
                                      'columnWidth5': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnWidth5',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 5,
                                                       'shortname': 'cw5'},
                                      'columnWidth6': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnWidth6',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 6,
                                                       'shortname': 'cw6'},
                                      'defineTemplate': {'args': '<type __builtin__.str>',
                                                         'longname': 'defineTemplate',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'dt'},
                                      'docTag': {'args': '<type __builtin__.str>',
                                                 'longname': 'docTag',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dtg'},
                                      'dragCallback': {'args': 'script',
                                                       'longname': 'dragCallback',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'dgc'},
                                      'dropCallback': {'args': 'script',
                                                       'longname': 'dropCallback',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'dpc'},
                                      'edit': {'args': '<type __builtin__.bool>',
                                               'longname': 'edit',
                                               'numArgs': 0,
                                               'shortname': 'e'},
                                      'enable': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enable',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'en'},
                                      'enableBackground': {'args': '<type __builtin__.bool>',
                                                           'longname': 'enableBackground',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ebg'},
                                      'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                              'longname': 'enableKeyboardFocus',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'ekf'},
                                      'enumeratedItem': {'args': ['<type __builtin__.int>',
                                                                  '<type __builtin__.str>'],
                                                         'longname': 'enumeratedItem',
                                                         'modes': ['create',
                                                                   'multiuse'],
                                                         'numArgs': 2,
                                                         'shortname': 'ei'},
                                      'exists': {'args': '<type __builtin__.bool>',
                                                 'longname': 'exists',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ex'},
                                      'fullPathName': {'args': '<type __builtin__.bool>',
                                                       'longname': 'fullPathName',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'fpn'},
                                      'height': {'args': '<type __builtin__.int>',
                                                 'longname': 'height',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'h'},
                                      'highlightColor': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'highlightColor',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 3,
                                                         'shortname': 'hlc'},
                                      'isObscured': {'args': '<type __builtin__.bool>',
                                                     'longname': 'isObscured',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'io'},
                                      'label': {'args': '<type __builtin__.str>',
                                                'longname': 'label',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'l'},
                                      'manage': {'args': '<type __builtin__.bool>',
                                                 'longname': 'manage',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'm'},
                                      'noBackground': {'args': '<type __builtin__.bool>',
                                                       'longname': 'noBackground',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'nbg'},
                                      'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                             'longname': 'numberOfPopupMenus',
                                                             'modes': ['query'],
                                                             'numArgs': 0,
                                                             'shortname': 'npm'},
                                      'parent': {'args': '<type __builtin__.str>',
                                                 'longname': 'parent',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'p'},
                                      'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                         'longname': 'popupMenuArray',
                                                         'modes': ['query'],
                                                         'numArgs': 0,
                                                         'shortname': 'pma'},
                                      'preventOverride': {'args': '<type __builtin__.bool>',
                                                          'longname': 'preventOverride',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'po'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'rowAttach': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'rowAttach',
                                                    'modes': ['create',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 3,
                                                    'shortname': 'rat'},
                                      'statusBarMessage': {'args': '<type __builtin__.str>',
                                                           'longname': 'statusBarMessage',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'sbm'},
                                      'useTemplate': {'args': '<type __builtin__.str>',
                                                      'longname': 'useTemplate',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ut'},
                                      'visible': {'args': '<type __builtin__.bool>',
                                                  'longname': 'visible',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'vis'},
                                      'visibleChangeCommand': {'args': 'script',
                                                               'longname': 'visibleChangeCommand',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'vcc'},
                                      'width': {'args': '<type __builtin__.int>',
                                                'longname': 'width',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'w'}},
                            'shortFlags': {'ad2': 'adjustableColumn2',
                                           'ad3': 'adjustableColumn3',
                                           'ad4': 'adjustableColumn4',
                                           'ad5': 'adjustableColumn5',
                                           'ad6': 'adjustableColumn6',
                                           'adj': 'adjustableColumn',
                                           'ann': 'annotation',
                                           'at': 'attribute',
                                           'bgc': 'backgroundColor',
                                           'cal': 'columnAlign',
                                           'cat': 'columnAttach',
                                           'cl2': 'columnAlign2',
                                           'cl3': 'columnAlign3',
                                           'cl4': 'columnAlign4',
                                           'cl5': 'columnAlign5',
                                           'cl6': 'columnAlign6',
                                           'co2': 'columnOffset2',
                                           'co3': 'columnOffset3',
                                           'co4': 'columnOffset4',
                                           'co5': 'columnOffset5',
                                           'co6': 'columnOffset6',
                                           'ct2': 'columnAttach2',
                                           'ct3': 'columnAttach3',
                                           'ct4': 'columnAttach4',
                                           'ct5': 'columnAttach5',
                                           'ct6': 'columnAttach6',
                                           'cw': 'columnWidth',
                                           'cw1': 'columnWidth1',
                                           'cw2': 'columnWidth2',
                                           'cw3': 'columnWidth3',
                                           'cw4': 'columnWidth4',
                                           'cw5': 'columnWidth5',
                                           'cw6': 'columnWidth6',
                                           'dgc': 'dragCallback',
                                           'dpc': 'dropCallback',
                                           'dt': 'defineTemplate',
                                           'dtg': 'docTag',
                                           'e': 'edit',
                                           'ebg': 'enableBackground',
                                           'ei': 'enumeratedItem',
                                           'ekf': 'enableKeyboardFocus',
                                           'en': 'enable',
                                           'ex': 'exists',
                                           'fpn': 'fullPathName',
                                           'h': 'height',
                                           'hlc': 'highlightColor',
                                           'io': 'isObscured',
                                           'l': 'label',
                                           'm': 'manage',
                                           'nbg': 'noBackground',
                                           'npm': 'numberOfPopupMenus',
                                           'p': 'parent',
                                           'pma': 'popupMenuArray',
                                           'po': 'preventOverride',
                                           'q': 'query',
                                           'rat': 'rowAttach',
                                           'sbm': 'statusBarMessage',
                                           'ut': 'useTemplate',
                                           'vcc': 'visibleChangeCommand',
                                           'vis': 'visible',
                                           'w': 'width'},
                            'type': 'windows'},
  'attrFieldGrp': {'callbackFlags': ['visibleChangeCommand',
                                     'vcc',
                                     'dragCallback',
                                     'dgc',
                                     'extraButtonCommand',
                                     'ebc',
                                     'changeCommand',
                                     'cc',
                                     'dropCallback',
                                     'dpc'],
                   'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'adj'},
                             'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn2',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad2'},
                             'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn3',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad3'},
                             'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn4',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad4'},
                             'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn5',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad5'},
                             'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn6',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad6'},
                             'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'attribute': {'args': '<type __builtin__.str>',
                                           'longname': 'attribute',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'at'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'changeCommand': {'args': 'script',
                                               'longname': 'changeCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cc'},
                             'columnAlign': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 2,
                                             'shortname': 'cal'},
                             'columnAlign2': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign2',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'cl2'},
                             'columnAlign3': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign3',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'cl3'},
                             'columnAlign4': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign4',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 4,
                                              'shortname': 'cl4'},
                             'columnAlign5': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign5',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 5,
                                              'shortname': 'cl5'},
                             'columnAlign6': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign6',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 6,
                                              'shortname': 'cl6'},
                             'columnAttach': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnAttach',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 3,
                                              'shortname': 'cat'},
                             'columnAttach2': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'ct2'},
                             'columnAttach3': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'ct3'},
                             'columnAttach4': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'ct4'},
                             'columnAttach5': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'ct5'},
                             'columnAttach6': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'ct6'},
                             'columnOffset2': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'co2'},
                             'columnOffset3': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'co3'},
                             'columnOffset4': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'co4'},
                             'columnOffset5': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'co5'},
                             'columnOffset6': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'co6'},
                             'columnWidth': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 2,
                                             'shortname': 'cw'},
                             'columnWidth1': {'args': '<type __builtin__.int>',
                                              'longname': 'columnWidth1',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cw1'},
                             'columnWidth2': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth2',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'cw2'},
                             'columnWidth3': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth3',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'cw3'},
                             'columnWidth4': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth4',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 4,
                                              'shortname': 'cw4'},
                             'columnWidth5': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth5',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 5,
                                              'shortname': 'cw5'},
                             'columnWidth6': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth6',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 6,
                                              'shortname': 'cw6'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'extraButton': {'args': '<type __builtin__.bool>',
                                             'longname': 'extraButton',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'eb'},
                             'extraButtonCommand': {'args': 'script',
                                                    'longname': 'extraButtonCommand',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ebc'},
                             'extraButtonIcon': {'args': '<type __builtin__.str>',
                                                 'longname': 'extraButtonIcon',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebi'},
                             'extraLabel': {'args': '<type __builtin__.str>',
                                            'longname': 'extraLabel',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'el'},
                             'forceAddMapButton': {'args': '<type __builtin__.bool>',
                                                   'longname': 'forceAddMapButton',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'fmb'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'hideMapButton': {'args': '<type __builtin__.bool>',
                                               'longname': 'hideMapButton',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'hmb'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'label': {'args': '<type __builtin__.str>',
                                       'longname': 'label',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'l'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'maxValue': {'args': '<type __builtin__.float>',
                                          'longname': 'maxValue',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'max'},
                             'minValue': {'args': '<type __builtin__.float>',
                                          'longname': 'minValue',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'min'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'numberOfFields': {'args': '<type __builtin__.int>',
                                                'longname': 'numberOfFields',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'nf'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'precision': {'args': '<type __builtin__.int>',
                                           'longname': 'precision',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pre'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'rowAttach': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'rowAttach',
                                           'modes': ['create',
                                                     'edit',
                                                     'multiuse'],
                                           'numArgs': 3,
                                           'shortname': 'rat'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'step': {'args': '<type __builtin__.float>',
                                      'longname': 'step',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'shortFlags': {'ad2': 'adjustableColumn2',
                                  'ad3': 'adjustableColumn3',
                                  'ad4': 'adjustableColumn4',
                                  'ad5': 'adjustableColumn5',
                                  'ad6': 'adjustableColumn6',
                                  'adj': 'adjustableColumn',
                                  'ann': 'annotation',
                                  'at': 'attribute',
                                  'bgc': 'backgroundColor',
                                  'cal': 'columnAlign',
                                  'cat': 'columnAttach',
                                  'cc': 'changeCommand',
                                  'cl2': 'columnAlign2',
                                  'cl3': 'columnAlign3',
                                  'cl4': 'columnAlign4',
                                  'cl5': 'columnAlign5',
                                  'cl6': 'columnAlign6',
                                  'co2': 'columnOffset2',
                                  'co3': 'columnOffset3',
                                  'co4': 'columnOffset4',
                                  'co5': 'columnOffset5',
                                  'co6': 'columnOffset6',
                                  'ct2': 'columnAttach2',
                                  'ct3': 'columnAttach3',
                                  'ct4': 'columnAttach4',
                                  'ct5': 'columnAttach5',
                                  'ct6': 'columnAttach6',
                                  'cw': 'columnWidth',
                                  'cw1': 'columnWidth1',
                                  'cw2': 'columnWidth2',
                                  'cw3': 'columnWidth3',
                                  'cw4': 'columnWidth4',
                                  'cw5': 'columnWidth5',
                                  'cw6': 'columnWidth6',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'eb': 'extraButton',
                                  'ebc': 'extraButtonCommand',
                                  'ebg': 'enableBackground',
                                  'ebi': 'extraButtonIcon',
                                  'ekf': 'enableKeyboardFocus',
                                  'el': 'extraLabel',
                                  'en': 'enable',
                                  'ex': 'exists',
                                  'fmb': 'forceAddMapButton',
                                  'fpn': 'fullPathName',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'hmb': 'hideMapButton',
                                  'io': 'isObscured',
                                  'l': 'label',
                                  'm': 'manage',
                                  'max': 'maxValue',
                                  'min': 'minValue',
                                  'nbg': 'noBackground',
                                  'nf': 'numberOfFields',
                                  'npm': 'numberOfPopupMenus',
                                  'p': 'parent',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'pre': 'precision',
                                  'q': 'query',
                                  'rat': 'rowAttach',
                                  's': 'step',
                                  'sbm': 'statusBarMessage',
                                  'ut': 'useTemplate',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'w': 'width'},
                   'type': 'windows'},
  'attrFieldSliderGrp': {'callbackFlags': ['visibleChangeCommand',
                                           'vcc',
                                           'dragCallback',
                                           'dgc',
                                           'extraButtonCommand',
                                           'ebc',
                                           'changeCommand',
                                           'cc',
                                           'dropCallback',
                                           'dpc'],
                         'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                        'longname': 'adjustableColumn',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'adj'},
                                   'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn2',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad2'},
                                   'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn3',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad3'},
                                   'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn4',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad4'},
                                   'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn5',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad5'},
                                   'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn6',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad6'},
                                   'annotation': {'args': '<type __builtin__.str>',
                                                  'longname': 'annotation',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ann'},
                                   'attribute': {'args': '<type __builtin__.str>',
                                                 'longname': 'attribute',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'at'},
                                   'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'backgroundColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'bgc'},
                                   'changeCommand': {'args': 'script',
                                                     'longname': 'changeCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'cc'},
                                   'columnAlign': {'args': ['<type __builtin__.int>',
                                                            '<type __builtin__.str>'],
                                                   'longname': 'columnAlign',
                                                   'modes': ['create',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 2,
                                                   'shortname': 'cal'},
                                   'columnAlign2': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign2',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 2,
                                                    'shortname': 'cl2'},
                                   'columnAlign3': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign3',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'cl3'},
                                   'columnAlign4': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign4',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 4,
                                                    'shortname': 'cl4'},
                                   'columnAlign5': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign5',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 5,
                                                    'shortname': 'cl5'},
                                   'columnAlign6': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign6',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 6,
                                                    'shortname': 'cl6'},
                                   'columnAttach': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnAttach',
                                                    'modes': ['create',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 3,
                                                    'shortname': 'cat'},
                                   'columnAttach2': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach2',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'ct2'},
                                   'columnAttach3': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach3',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'ct3'},
                                   'columnAttach4': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach4',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 4,
                                                     'shortname': 'ct4'},
                                   'columnAttach5': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach5',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 5,
                                                     'shortname': 'ct5'},
                                   'columnAttach6': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach6',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 6,
                                                     'shortname': 'ct6'},
                                   'columnOffset2': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset2',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'co2'},
                                   'columnOffset3': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset3',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'co3'},
                                   'columnOffset4': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset4',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 4,
                                                     'shortname': 'co4'},
                                   'columnOffset5': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset5',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 5,
                                                     'shortname': 'co5'},
                                   'columnOffset6': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset6',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 6,
                                                     'shortname': 'co6'},
                                   'columnWidth': {'args': ['<type __builtin__.int>',
                                                            '<type __builtin__.int>'],
                                                   'longname': 'columnWidth',
                                                   'modes': ['create',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 2,
                                                   'shortname': 'cw'},
                                   'columnWidth1': {'args': '<type __builtin__.int>',
                                                    'longname': 'columnWidth1',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'cw1'},
                                   'columnWidth2': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth2',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 2,
                                                    'shortname': 'cw2'},
                                   'columnWidth3': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth3',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'cw3'},
                                   'columnWidth4': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth4',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 4,
                                                    'shortname': 'cw4'},
                                   'columnWidth5': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth5',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 5,
                                                    'shortname': 'cw5'},
                                   'columnWidth6': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth6',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 6,
                                                    'shortname': 'cw6'},
                                   'defineTemplate': {'args': '<type __builtin__.str>',
                                                      'longname': 'defineTemplate',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'dt'},
                                   'docTag': {'args': '<type __builtin__.str>',
                                              'longname': 'docTag',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dtg'},
                                   'dragCallback': {'args': 'script',
                                                    'longname': 'dragCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dgc'},
                                   'dropCallback': {'args': 'script',
                                                    'longname': 'dropCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dpc'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'enable': {'args': '<type __builtin__.bool>',
                                              'longname': 'enable',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'en'},
                                   'enableBackground': {'args': '<type __builtin__.bool>',
                                                        'longname': 'enableBackground',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ebg'},
                                   'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                           'longname': 'enableKeyboardFocus',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ekf'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'extraButton': {'args': '<type __builtin__.bool>',
                                                   'longname': 'extraButton',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'eb'},
                                   'extraButtonCommand': {'args': 'script',
                                                          'longname': 'extraButtonCommand',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ebc'},
                                   'extraButtonIcon': {'args': '<type __builtin__.str>',
                                                       'longname': 'extraButtonIcon',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ebi'},
                                   'fieldMaxValue': {'args': '<type __builtin__.float>',
                                                     'longname': 'fieldMaxValue',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'fmx'},
                                   'fieldMinValue': {'args': '<type __builtin__.float>',
                                                     'longname': 'fieldMinValue',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'fmn'},
                                   'fieldStep': {'args': '<type __builtin__.float>',
                                                 'longname': 'fieldStep',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'fs'},
                                   'forceAddMapButton': {'args': '<type __builtin__.bool>',
                                                         'longname': 'forceAddMapButton',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'fmb'},
                                   'fullPathName': {'args': '<type __builtin__.bool>',
                                                    'longname': 'fullPathName',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'fpn'},
                                   'height': {'args': '<type __builtin__.int>',
                                              'longname': 'height',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'h'},
                                   'hideMapButton': {'args': '<type __builtin__.bool>',
                                                     'longname': 'hideMapButton',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'hmb'},
                                   'highlightColor': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'highlightColor',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'hlc'},
                                   'isObscured': {'args': '<type __builtin__.bool>',
                                                  'longname': 'isObscured',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'io'},
                                   'label': {'args': '<type __builtin__.str>',
                                             'longname': 'label',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'l'},
                                   'manage': {'args': '<type __builtin__.bool>',
                                              'longname': 'manage',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'm'},
                                   'maxValue': {'args': '<type __builtin__.float>',
                                                'longname': 'maxValue',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'max'},
                                   'minValue': {'args': '<type __builtin__.float>',
                                                'longname': 'minValue',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'min'},
                                   'noBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'noBackground',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'nbg'},
                                   'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                          'longname': 'numberOfPopupMenus',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'npm'},
                                   'parent': {'args': '<type __builtin__.str>',
                                              'longname': 'parent',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'p'},
                                   'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                      'longname': 'popupMenuArray',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'pma'},
                                   'precision': {'args': '<type __builtin__.int>',
                                                 'longname': 'precision',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'pre'},
                                   'preventOverride': {'args': '<type __builtin__.bool>',
                                                       'longname': 'preventOverride',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'po'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'rowAttach': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'rowAttach',
                                                 'modes': ['create',
                                                           'edit',
                                                           'multiuse'],
                                                 'numArgs': 3,
                                                 'shortname': 'rat'},
                                   'sliderMaxValue': {'args': '<type __builtin__.float>',
                                                      'longname': 'sliderMaxValue',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'smx'},
                                   'sliderMinValue': {'args': '<type __builtin__.float>',
                                                      'longname': 'sliderMinValue',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'smn'},
                                   'sliderStep': {'args': '<type __builtin__.float>',
                                                  'longname': 'sliderStep',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ss'},
                                   'statusBarMessage': {'args': '<type __builtin__.str>',
                                                        'longname': 'statusBarMessage',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'sbm'},
                                   'step': {'args': '<type __builtin__.float>',
                                            'longname': 'step',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 's'},
                                   'useTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'useTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ut'},
                                   'vertical': {'args': '<type __builtin__.bool>',
                                                'longname': 'vertical',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 0,
                                                'shortname': 'vr'},
                                   'visible': {'args': '<type __builtin__.bool>',
                                               'longname': 'visible',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'vis'},
                                   'visibleChangeCommand': {'args': 'script',
                                                            'longname': 'visibleChangeCommand',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'vcc'},
                                   'width': {'args': '<type __builtin__.int>',
                                             'longname': 'width',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'w'}},
                         'shortFlags': {'ad2': 'adjustableColumn2',
                                        'ad3': 'adjustableColumn3',
                                        'ad4': 'adjustableColumn4',
                                        'ad5': 'adjustableColumn5',
                                        'ad6': 'adjustableColumn6',
                                        'adj': 'adjustableColumn',
                                        'ann': 'annotation',
                                        'at': 'attribute',
                                        'bgc': 'backgroundColor',
                                        'cal': 'columnAlign',
                                        'cat': 'columnAttach',
                                        'cc': 'changeCommand',
                                        'cl2': 'columnAlign2',
                                        'cl3': 'columnAlign3',
                                        'cl4': 'columnAlign4',
                                        'cl5': 'columnAlign5',
                                        'cl6': 'columnAlign6',
                                        'co2': 'columnOffset2',
                                        'co3': 'columnOffset3',
                                        'co4': 'columnOffset4',
                                        'co5': 'columnOffset5',
                                        'co6': 'columnOffset6',
                                        'ct2': 'columnAttach2',
                                        'ct3': 'columnAttach3',
                                        'ct4': 'columnAttach4',
                                        'ct5': 'columnAttach5',
                                        'ct6': 'columnAttach6',
                                        'cw': 'columnWidth',
                                        'cw1': 'columnWidth1',
                                        'cw2': 'columnWidth2',
                                        'cw3': 'columnWidth3',
                                        'cw4': 'columnWidth4',
                                        'cw5': 'columnWidth5',
                                        'cw6': 'columnWidth6',
                                        'dgc': 'dragCallback',
                                        'dpc': 'dropCallback',
                                        'dt': 'defineTemplate',
                                        'dtg': 'docTag',
                                        'e': 'edit',
                                        'eb': 'extraButton',
                                        'ebc': 'extraButtonCommand',
                                        'ebg': 'enableBackground',
                                        'ebi': 'extraButtonIcon',
                                        'ekf': 'enableKeyboardFocus',
                                        'en': 'enable',
                                        'ex': 'exists',
                                        'fmb': 'forceAddMapButton',
                                        'fmn': 'fieldMinValue',
                                        'fmx': 'fieldMaxValue',
                                        'fpn': 'fullPathName',
                                        'fs': 'fieldStep',
                                        'h': 'height',
                                        'hlc': 'highlightColor',
                                        'hmb': 'hideMapButton',
                                        'io': 'isObscured',
                                        'l': 'label',
                                        'm': 'manage',
                                        'max': 'maxValue',
                                        'min': 'minValue',
                                        'nbg': 'noBackground',
                                        'npm': 'numberOfPopupMenus',
                                        'p': 'parent',
                                        'pma': 'popupMenuArray',
                                        'po': 'preventOverride',
                                        'pre': 'precision',
                                        'q': 'query',
                                        'rat': 'rowAttach',
                                        's': 'step',
                                        'sbm': 'statusBarMessage',
                                        'smn': 'sliderMinValue',
                                        'smx': 'sliderMaxValue',
                                        'ss': 'sliderStep',
                                        'ut': 'useTemplate',
                                        'vcc': 'visibleChangeCommand',
                                        'vis': 'visible',
                                        'vr': 'vertical',
                                        'w': 'width'},
                         'type': 'windows'},
  'attrNavigationControlGrp': {'callbackFlags': ['connectAttrToDropped',
                                                 'cad',
                                                 'visibleChangeCommand',
                                                 'vcc',
                                                 'dragCallback',
                                                 'dgc',
                                                 'defaultTraversal',
                                                 'dtv',
                                                 'ignore',
                                                 'i',
                                                 'connectNodeToDropped',
                                                 'cnd',
                                                 'createNew',
                                                 'cn',
                                                 'extraButtonCommand',
                                                 'ebc',
                                                 'unignore',
                                                 'u',
                                                 'disconnect',
                                                 'd',
                                                 'connectToExisting',
                                                 'ce',
                                                 'relatedNodes',
                                                 'ren',
                                                 'dropCallback',
                                                 'dpc'],
                               'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                              'longname': 'adjustableColumn',
                                                              'modes': ['create',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'adj'},
                                         'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                               'longname': 'adjustableColumn2',
                                                               'modes': ['create',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'ad2'},
                                         'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                               'longname': 'adjustableColumn3',
                                                               'modes': ['create',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'ad3'},
                                         'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                               'longname': 'adjustableColumn4',
                                                               'modes': ['create',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'ad4'},
                                         'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                               'longname': 'adjustableColumn5',
                                                               'modes': ['create',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'ad5'},
                                         'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                               'longname': 'adjustableColumn6',
                                                               'modes': ['create',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'ad6'},
                                         'annotation': {'args': '<type __builtin__.str>',
                                                        'longname': 'annotation',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ann'},
                                         'attrNavDecision': {'args': ['PyNode',
                                                                      '<type __builtin__.str>'],
                                                             'longname': 'attrNavDecision',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 2,
                                                             'shortname': 'attrNavDecision'},
                                         'attribute': {'args': 'PyNode',
                                                       'longname': 'attribute',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'at'},
                                         'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                      '<type __builtin__.float>',
                                                                      '<type __builtin__.float>'],
                                                             'longname': 'backgroundColor',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 3,
                                                             'shortname': 'bgc'},
                                         'columnAlign': {'args': ['<type __builtin__.int>',
                                                                  '<type __builtin__.str>'],
                                                         'longname': 'columnAlign',
                                                         'modes': ['create',
                                                                   'edit',
                                                                   'multiuse'],
                                                         'numArgs': 2,
                                                         'shortname': 'cal'},
                                         'columnAlign2': {'args': ['<type __builtin__.str>',
                                                                   '<type __builtin__.str>'],
                                                          'longname': 'columnAlign2',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 2,
                                                          'shortname': 'cl2'},
                                         'columnAlign3': {'args': ['<type __builtin__.str>',
                                                                   '<type __builtin__.str>',
                                                                   '<type __builtin__.str>'],
                                                          'longname': 'columnAlign3',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 3,
                                                          'shortname': 'cl3'},
                                         'columnAlign4': {'args': ['<type __builtin__.str>',
                                                                   '<type __builtin__.str>',
                                                                   '<type __builtin__.str>',
                                                                   '<type __builtin__.str>'],
                                                          'longname': 'columnAlign4',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 4,
                                                          'shortname': 'cl4'},
                                         'columnAlign5': {'args': ['<type __builtin__.str>',
                                                                   '<type __builtin__.str>',
                                                                   '<type __builtin__.str>',
                                                                   '<type __builtin__.str>',
                                                                   '<type __builtin__.str>'],
                                                          'longname': 'columnAlign5',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 5,
                                                          'shortname': 'cl5'},
                                         'columnAlign6': {'args': ['<type __builtin__.str>',
                                                                   '<type __builtin__.str>',
                                                                   '<type __builtin__.str>',
                                                                   '<type __builtin__.str>',
                                                                   '<type __builtin__.str>',
                                                                   '<type __builtin__.str>'],
                                                          'longname': 'columnAlign6',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 6,
                                                          'shortname': 'cl6'},
                                         'columnAttach': {'args': ['<type __builtin__.int>',
                                                                   '<type __builtin__.str>',
                                                                   '<type __builtin__.int>'],
                                                          'longname': 'columnAttach',
                                                          'modes': ['create',
                                                                    'edit',
                                                                    'multiuse'],
                                                          'numArgs': 3,
                                                          'shortname': 'cat'},
                                         'columnAttach2': {'args': ['<type __builtin__.str>',
                                                                    '<type __builtin__.str>'],
                                                           'longname': 'columnAttach2',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 2,
                                                           'shortname': 'ct2'},
                                         'columnAttach3': {'args': ['<type __builtin__.str>',
                                                                    '<type __builtin__.str>',
                                                                    '<type __builtin__.str>'],
                                                           'longname': 'columnAttach3',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 3,
                                                           'shortname': 'ct3'},
                                         'columnAttach4': {'args': ['<type __builtin__.str>',
                                                                    '<type __builtin__.str>',
                                                                    '<type __builtin__.str>',
                                                                    '<type __builtin__.str>'],
                                                           'longname': 'columnAttach4',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 4,
                                                           'shortname': 'ct4'},
                                         'columnAttach5': {'args': ['<type __builtin__.str>',
                                                                    '<type __builtin__.str>',
                                                                    '<type __builtin__.str>',
                                                                    '<type __builtin__.str>',
                                                                    '<type __builtin__.str>'],
                                                           'longname': 'columnAttach5',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 5,
                                                           'shortname': 'ct5'},
                                         'columnAttach6': {'args': ['<type __builtin__.str>',
                                                                    '<type __builtin__.str>',
                                                                    '<type __builtin__.str>',
                                                                    '<type __builtin__.str>',
                                                                    '<type __builtin__.str>',
                                                                    '<type __builtin__.str>'],
                                                           'longname': 'columnAttach6',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 6,
                                                           'shortname': 'ct6'},
                                         'columnOffset2': {'args': ['<type __builtin__.int>',
                                                                    '<type __builtin__.int>'],
                                                           'longname': 'columnOffset2',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 2,
                                                           'shortname': 'co2'},
                                         'columnOffset3': {'args': ['<type __builtin__.int>',
                                                                    '<type __builtin__.int>',
                                                                    '<type __builtin__.int>'],
                                                           'longname': 'columnOffset3',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 3,
                                                           'shortname': 'co3'},
                                         'columnOffset4': {'args': ['<type __builtin__.int>',
                                                                    '<type __builtin__.int>',
                                                                    '<type __builtin__.int>',
                                                                    '<type __builtin__.int>'],
                                                           'longname': 'columnOffset4',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 4,
                                                           'shortname': 'co4'},
                                         'columnOffset5': {'args': ['<type __builtin__.int>',
                                                                    '<type __builtin__.int>',
                                                                    '<type __builtin__.int>',
                                                                    '<type __builtin__.int>',
                                                                    '<type __builtin__.int>'],
                                                           'longname': 'columnOffset5',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 5,
                                                           'shortname': 'co5'},
                                         'columnOffset6': {'args': ['<type __builtin__.int>',
                                                                    '<type __builtin__.int>',
                                                                    '<type __builtin__.int>',
                                                                    '<type __builtin__.int>',
                                                                    '<type __builtin__.int>',
                                                                    '<type __builtin__.int>'],
                                                           'longname': 'columnOffset6',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 6,
                                                           'shortname': 'co6'},
                                         'columnWidth': {'args': ['<type __builtin__.int>',
                                                                  '<type __builtin__.int>'],
                                                         'longname': 'columnWidth',
                                                         'modes': ['create',
                                                                   'edit',
                                                                   'multiuse'],
                                                         'numArgs': 2,
                                                         'shortname': 'cw'},
                                         'columnWidth1': {'args': '<type __builtin__.int>',
                                                          'longname': 'columnWidth1',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'cw1'},
                                         'columnWidth2': {'args': ['<type __builtin__.int>',
                                                                   '<type __builtin__.int>'],
                                                          'longname': 'columnWidth2',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 2,
                                                          'shortname': 'cw2'},
                                         'columnWidth3': {'args': ['<type __builtin__.int>',
                                                                   '<type __builtin__.int>',
                                                                   '<type __builtin__.int>'],
                                                          'longname': 'columnWidth3',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 3,
                                                          'shortname': 'cw3'},
                                         'columnWidth4': {'args': ['<type __builtin__.int>',
                                                                   '<type __builtin__.int>',
                                                                   '<type __builtin__.int>',
                                                                   '<type __builtin__.int>'],
                                                          'longname': 'columnWidth4',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 4,
                                                          'shortname': 'cw4'},
                                         'columnWidth5': {'args': ['<type __builtin__.int>',
                                                                   '<type __builtin__.int>',
                                                                   '<type __builtin__.int>',
                                                                   '<type __builtin__.int>',
                                                                   '<type __builtin__.int>'],
                                                          'longname': 'columnWidth5',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 5,
                                                          'shortname': 'cw5'},
                                         'columnWidth6': {'args': ['<type __builtin__.int>',
                                                                   '<type __builtin__.int>',
                                                                   '<type __builtin__.int>',
                                                                   '<type __builtin__.int>',
                                                                   '<type __builtin__.int>',
                                                                   '<type __builtin__.int>'],
                                                          'longname': 'columnWidth6',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 6,
                                                          'shortname': 'cw6'},
                                         'connectAttrToDropped': {'args': 'script',
                                                                  'longname': 'connectAttrToDropped',
                                                                  'modes': ['create',
                                                                            'query',
                                                                            'edit'],
                                                                  'numArgs': 1,
                                                                  'shortname': 'cad'},
                                         'connectNodeToDropped': {'args': 'script',
                                                                  'longname': 'connectNodeToDropped',
                                                                  'modes': ['create',
                                                                            'query',
                                                                            'edit'],
                                                                  'numArgs': 1,
                                                                  'shortname': 'cnd'},
                                         'connectToExisting': {'args': 'script',
                                                               'longname': 'connectToExisting',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'ce'},
                                         'createNew': {'args': 'script',
                                                       'longname': 'createNew',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'cn'},
                                         'defaultTraversal': {'args': 'script',
                                                              'longname': 'defaultTraversal',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'dtv'},
                                         'defineTemplate': {'args': '<type __builtin__.str>',
                                                            'longname': 'defineTemplate',
                                                            'modes': ['create'],
                                                            'numArgs': 1,
                                                            'shortname': 'dt'},
                                         'delete': {'args': '<type __builtin__.str>',
                                                    'longname': 'delete',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'delete'},
                                         'disconnect': {'args': 'script',
                                                        'longname': 'disconnect',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'd'},
                                         'docTag': {'args': '<type __builtin__.str>',
                                                    'longname': 'docTag',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dtg'},
                                         'dragCallback': {'args': 'script',
                                                          'longname': 'dragCallback',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'dgc'},
                                         'dropCallback': {'args': 'script',
                                                          'longname': 'dropCallback',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'dpc'},
                                         'edit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'edit',
                                                  'numArgs': 0,
                                                  'shortname': 'e'},
                                         'enable': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enable',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'en'},
                                         'enableBackground': {'args': '<type __builtin__.bool>',
                                                              'longname': 'enableBackground',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'ebg'},
                                         'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'enableKeyboardFocus',
                                                                 'modes': ['create',
                                                                           'query',
                                                                           'edit'],
                                                                 'numArgs': 1,
                                                                 'shortname': 'ekf'},
                                         'exists': {'args': '<type __builtin__.bool>',
                                                    'longname': 'exists',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ex'},
                                         'extraButton': {'args': '<type __builtin__.bool>',
                                                         'longname': 'extraButton',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'eb'},
                                         'extraButtonCommand': {'args': 'script',
                                                                'longname': 'extraButtonCommand',
                                                                'modes': ['create',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'ebc'},
                                         'extraButtonIcon': {'args': '<type __builtin__.str>',
                                                             'longname': 'extraButtonIcon',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'ebi'},
                                         'fullPathName': {'args': '<type __builtin__.bool>',
                                                          'longname': 'fullPathName',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'fpn'},
                                         'height': {'args': '<type __builtin__.int>',
                                                    'longname': 'height',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'h'},
                                         'highlightColor': {'args': ['<type __builtin__.float>',
                                                                     '<type __builtin__.float>',
                                                                     '<type __builtin__.float>'],
                                                            'longname': 'highlightColor',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 3,
                                                            'shortname': 'hlc'},
                                         'ignore': {'args': 'script',
                                                    'longname': 'ignore',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'i'},
                                         'ignoreNotSupported': {'args': '<type __builtin__.bool>',
                                                                'longname': 'ignoreNotSupported',
                                                                'modes': [],
                                                                'numArgs': 0,
                                                                'shortname': 'ins'},
                                         'isObscured': {'args': '<type __builtin__.bool>',
                                                        'longname': 'isObscured',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'io'},
                                         'label': {'args': '<type __builtin__.str>',
                                                   'longname': 'label',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'l'},
                                         'manage': {'args': '<type __builtin__.bool>',
                                                    'longname': 'manage',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'm'},
                                         'noBackground': {'args': '<type __builtin__.bool>',
                                                          'longname': 'noBackground',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'nbg'},
                                         'noIgnorableMenu': {'args': '<type __builtin__.bool>',
                                                             'longname': 'noIgnorableMenu',
                                                             'modes': ['create',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'nim'},
                                         'noKeyableMenu': {'args': '<type __builtin__.bool>',
                                                           'longname': 'noKeyableMenu',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'nkm'},
                                         'notIgnorableMenu': {'args': '<type __builtin__.bool>',
                                                              'longname': 'notIgnorableMenu',
                                                              'numArgs': 1,
                                                              'shortname': 'nim'},
                                         'notKeyableMenu': {'args': '<type __builtin__.bool>',
                                                            'longname': 'notKeyableMenu',
                                                            'numArgs': 1,
                                                            'shortname': 'nkm'},
                                         'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                                'longname': 'numberOfPopupMenus',
                                                                'modes': ['query'],
                                                                'numArgs': 0,
                                                                'shortname': 'npm'},
                                         'parent': {'args': '<type __builtin__.str>',
                                                    'longname': 'parent',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'p'},
                                         'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                            'longname': 'popupMenuArray',
                                                            'modes': ['query'],
                                                            'numArgs': 0,
                                                            'shortname': 'pma'},
                                         'preventOverride': {'args': '<type __builtin__.bool>',
                                                             'longname': 'preventOverride',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'po'},
                                         'query': {'args': '<type __builtin__.bool>',
                                                   'longname': 'query',
                                                   'numArgs': 0,
                                                   'shortname': 'q'},
                                         'relatedNodes': {'args': 'script',
                                                          'longname': 'relatedNodes',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ren'},
                                         'rowAttach': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'rowAttach',
                                                       'modes': ['create',
                                                                 'edit',
                                                                 'multiuse'],
                                                       'numArgs': 3,
                                                       'shortname': 'rat'},
                                         'statusBarMessage': {'args': '<type __builtin__.str>',
                                                              'longname': 'statusBarMessage',
                                                              'modes': ['create',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'sbm'},
                                         'unignore': {'args': 'script',
                                                      'longname': 'unignore',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'u'},
                                         'useTemplate': {'args': '<type __builtin__.str>',
                                                         'longname': 'useTemplate',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'ut'},
                                         'visible': {'args': '<type __builtin__.bool>',
                                                     'longname': 'visible',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vis'},
                                         'visibleChangeCommand': {'args': 'script',
                                                                  'longname': 'visibleChangeCommand',
                                                                  'modes': ['create',
                                                                            'query',
                                                                            'edit'],
                                                                  'numArgs': 1,
                                                                  'shortname': 'vcc'},
                                         'width': {'args': '<type __builtin__.int>',
                                                   'longname': 'width',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'w'}},
                               'removedFlags': {'and': 'attrNavDecision',
                                                'del': 'delete'},
                               'shortFlags': {'ad2': 'adjustableColumn2',
                                              'ad3': 'adjustableColumn3',
                                              'ad4': 'adjustableColumn4',
                                              'ad5': 'adjustableColumn5',
                                              'ad6': 'adjustableColumn6',
                                              'adj': 'adjustableColumn',
                                              'ann': 'annotation',
                                              'at': 'attribute',
                                              'attrNavDecision': 'attrNavDecision',
                                              'bgc': 'backgroundColor',
                                              'cad': 'connectAttrToDropped',
                                              'cal': 'columnAlign',
                                              'cat': 'columnAttach',
                                              'ce': 'connectToExisting',
                                              'cl2': 'columnAlign2',
                                              'cl3': 'columnAlign3',
                                              'cl4': 'columnAlign4',
                                              'cl5': 'columnAlign5',
                                              'cl6': 'columnAlign6',
                                              'cn': 'createNew',
                                              'cnd': 'connectNodeToDropped',
                                              'co2': 'columnOffset2',
                                              'co3': 'columnOffset3',
                                              'co4': 'columnOffset4',
                                              'co5': 'columnOffset5',
                                              'co6': 'columnOffset6',
                                              'ct2': 'columnAttach2',
                                              'ct3': 'columnAttach3',
                                              'ct4': 'columnAttach4',
                                              'ct5': 'columnAttach5',
                                              'ct6': 'columnAttach6',
                                              'cw': 'columnWidth',
                                              'cw1': 'columnWidth1',
                                              'cw2': 'columnWidth2',
                                              'cw3': 'columnWidth3',
                                              'cw4': 'columnWidth4',
                                              'cw5': 'columnWidth5',
                                              'cw6': 'columnWidth6',
                                              'd': 'disconnect',
                                              'delete': 'delete',
                                              'dgc': 'dragCallback',
                                              'dpc': 'dropCallback',
                                              'dt': 'defineTemplate',
                                              'dtg': 'docTag',
                                              'dtv': 'defaultTraversal',
                                              'e': 'edit',
                                              'eb': 'extraButton',
                                              'ebc': 'extraButtonCommand',
                                              'ebg': 'enableBackground',
                                              'ebi': 'extraButtonIcon',
                                              'ekf': 'enableKeyboardFocus',
                                              'en': 'enable',
                                              'ex': 'exists',
                                              'fpn': 'fullPathName',
                                              'h': 'height',
                                              'hlc': 'highlightColor',
                                              'i': 'ignore',
                                              'ins': 'ignoreNotSupported',
                                              'io': 'isObscured',
                                              'l': 'label',
                                              'm': 'manage',
                                              'nbg': 'noBackground',
                                              'nim': 'notIgnorableMenu',
                                              'nkm': 'notKeyableMenu',
                                              'npm': 'numberOfPopupMenus',
                                              'p': 'parent',
                                              'pma': 'popupMenuArray',
                                              'po': 'preventOverride',
                                              'q': 'query',
                                              'rat': 'rowAttach',
                                              'ren': 'relatedNodes',
                                              'sbm': 'statusBarMessage',
                                              'u': 'unignore',
                                              'ut': 'useTemplate',
                                              'vcc': 'visibleChangeCommand',
                                              'vis': 'visible',
                                              'w': 'width'},
                               'type': 'windows'},
  'attributeInfo': {'flags': {'allAttributes': {'args': '<type __builtin__.bool>',
                                                'longname': 'allAttributes',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'all'},
                              'bool': {'args': '<type __builtin__.bool>',
                                       'longname': 'bool',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'b'},
                              'enumerated': {'args': '<type __builtin__.bool>',
                                             'longname': 'enumerated',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'e'},
                              'hidden': {'args': '<type __builtin__.bool>',
                                         'longname': 'hidden',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'h'},
                              'inherited': {'args': '<type __builtin__.bool>',
                                            'longname': 'inherited',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'inherited'},
                              'internal': {'args': '<type __builtin__.bool>',
                                           'longname': 'internal',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'i'},
                              'leaf': {'args': '<type __builtin__.bool>',
                                       'longname': 'leaf',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'l'},
                              'logicalAnd': {'args': '<type __builtin__.bool>',
                                             'longname': 'logicalAnd',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'logicalAnd'},
                              'multi': {'args': '<type __builtin__.bool>',
                                        'longname': 'multi',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                              'short': {'args': '<type __builtin__.bool>',
                                        'longname': 'short',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 's'},
                              'type': {'args': '<type __builtin__.str>',
                                       'longname': 'type',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 't'},
                              'userInterface': {'args': '<type __builtin__.bool>',
                                                'longname': 'userInterface',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ui'},
                              'writable': {'args': '<type __builtin__.bool>',
                                           'longname': 'writable',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'w'}},
                    'removedFlags': {'and': 'logicalAnd', 'in': 'inherited'},
                    'shortFlags': {'all': 'allAttributes',
                                   'b': 'bool',
                                   'e': 'enumerated',
                                   'h': 'hidden',
                                   'i': 'internal',
                                   'inherited': 'inherited',
                                   'l': 'leaf',
                                   'logicalAnd': 'logicalAnd',
                                   'm': 'multi',
                                   's': 'short',
                                   't': 'type',
                                   'ui': 'userInterface',
                                   'w': 'writable'},
                    'type': 'general'},
  'attributeMenu': {'callbackFlags': ['unregPulldownMenuCommand',
                                      'upm',
                                      'regPulldownMenuCommand',
                                      'rpm'],
                    'flags': {'beginMenu': {'args': '<type __builtin__.bool>',
                                            'longname': 'beginMenu',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'beg'},
                              'editor': {'args': '<type __builtin__.str>',
                                         'longname': 'editor',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'edt'},
                              'finishMenu': {'args': '<type __builtin__.bool>',
                                             'longname': 'finishMenu',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'fsh'},
                              'inputs': {'args': '<type __builtin__.bool>',
                                         'longname': 'inputs',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'inp'},
                              'plug': {'args': 'PyNode',
                                       'longname': 'plug',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                              'regPulldownMenuCommand': {'args': '<type __builtin__.str>',
                                                         'longname': 'regPulldownMenuCommand',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'rpm'},
                              'unregPulldownMenuCommand': {'args': '<type __builtin__.int>',
                                                           'longname': 'unregPulldownMenuCommand',
                                                           'modes': ['create'],
                                                           'numArgs': 1,
                                                           'shortname': 'upm'}},
                    'shortFlags': {'beg': 'beginMenu',
                                   'edt': 'editor',
                                   'fsh': 'finishMenu',
                                   'inp': 'inputs',
                                   'p': 'plug',
                                   'rpm': 'regPulldownMenuCommand',
                                   'upm': 'unregPulldownMenuCommand'},
                    'type': 'windows'},
  'attributeName': {'flags': {'leaf': {'args': '<type __builtin__.bool>',
                                       'longname': 'leaf',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'lf'},
                              'long': {'args': '<type __builtin__.bool>',
                                       'longname': 'long',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'l'},
                              'nice': {'args': '<type __builtin__.bool>',
                                       'longname': 'nice',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'n'},
                              'short': {'args': '<type __builtin__.bool>',
                                        'longname': 'short',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 's'}},
                    'shortFlags': {'l': 'long',
                                   'lf': 'leaf',
                                   'n': 'nice',
                                   's': 'short'},
                    'type': 'general'},
  'attributeQuery': {'flags': {'affectsAppearance': {'args': '<type __builtin__.bool>',
                                                     'longname': 'affectsAppearance',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'aa'},
                               'affectsWorldspace': {'args': '<type __builtin__.bool>',
                                                     'longname': 'affectsWorldspace',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'aws'},
                               'attributeType': {'args': '<type __builtin__.bool>',
                                                 'longname': 'attributeType',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'at'},
                               'cachedInternally': {'args': '<type __builtin__.bool>',
                                                    'longname': 'cachedInternally',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ci'},
                               'categories': {'args': '<type __builtin__.bool>',
                                              'longname': 'categories',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ct'},
                               'channelBox': {'args': '<type __builtin__.bool>',
                                              'longname': 'channelBox',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ch'},
                               'connectable': {'args': '<type __builtin__.bool>',
                                               'longname': 'connectable',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'c'},
                               'enum': {'args': '<type __builtin__.bool>',
                                        'longname': 'enum',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'hidden': {'args': '<type __builtin__.bool>',
                                          'longname': 'hidden',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'h'},
                               'indeterminant': {'args': '<type __builtin__.bool>',
                                                 'longname': 'indeterminant',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'idt'},
                               'indexMatters': {'args': '<type __builtin__.bool>',
                                                'longname': 'indexMatters',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'im'},
                               'internal': {'args': '<type __builtin__.bool>',
                                            'longname': 'internal',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'i'},
                               'internalGet': {'args': '<type __builtin__.bool>',
                                               'longname': 'internalGet',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ig'},
                               'internalSet': {'args': '<type __builtin__.bool>',
                                               'longname': 'internalSet',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'internalSet'},
                               'keyable': {'args': '<type __builtin__.bool>',
                                           'longname': 'keyable',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'k'},
                               'listChildren': {'args': '<type __builtin__.bool>',
                                                'longname': 'listChildren',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'lc'},
                               'listDefault': {'args': '<type __builtin__.bool>',
                                               'longname': 'listDefault',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ld'},
                               'listEnum': {'args': '<type __builtin__.bool>',
                                            'longname': 'listEnum',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'le'},
                               'listParent': {'args': '<type __builtin__.bool>',
                                              'longname': 'listParent',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'lp'},
                               'listSiblings': {'args': '<type __builtin__.bool>',
                                                'longname': 'listSiblings',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ls'},
                               'longName': {'args': '<type __builtin__.bool>',
                                            'longname': 'longName',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ln'},
                               'maxExists': {'args': '<type __builtin__.bool>',
                                             'longname': 'maxExists',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'mxe'},
                               'maximum': {'args': '<type __builtin__.bool>',
                                           'longname': 'maximum',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'max'},
                               'message': {'args': '<type __builtin__.bool>',
                                           'longname': 'message',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'msg'},
                               'minExists': {'args': '<type __builtin__.bool>',
                                             'longname': 'minExists',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'mne'},
                               'minimum': {'args': '<type __builtin__.bool>',
                                           'longname': 'minimum',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'min'},
                               'multi': {'args': '<type __builtin__.bool>',
                                         'longname': 'multi',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'm'},
                               'niceName': {'args': '<type __builtin__.bool>',
                                            'longname': 'niceName',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'nn'},
                               'node': {'args': 'PyNode',
                                        'longname': 'node',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'numberOfChildren': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfChildren',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'nc'},
                               'range': {'args': '<type __builtin__.bool>',
                                         'longname': 'range',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'r'},
                               'rangeExists': {'args': '<type __builtin__.bool>',
                                               'longname': 'rangeExists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 're'},
                               'readable': {'args': '<type __builtin__.bool>',
                                            'longname': 'readable',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'rd'},
                               'renderSource': {'args': '<type __builtin__.bool>',
                                                'longname': 'renderSource',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'rs'},
                               'shortName': {'args': '<type __builtin__.bool>',
                                             'longname': 'shortName',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'sn'},
                               'softMax': {'args': '<type __builtin__.bool>',
                                           'longname': 'softMax',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'smx'},
                               'softMaxExists': {'args': '<type __builtin__.bool>',
                                                 'longname': 'softMaxExists',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'sxe'},
                               'softMin': {'args': '<type __builtin__.bool>',
                                           'longname': 'softMin',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'smn'},
                               'softMinExists': {'args': '<type __builtin__.bool>',
                                                 'longname': 'softMinExists',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'sme'},
                               'softRange': {'args': '<type __builtin__.bool>',
                                             'longname': 'softRange',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 's'},
                               'softRangeExists': {'args': '<type __builtin__.bool>',
                                                   'longname': 'softRangeExists',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'se'},
                               'storable': {'args': '<type __builtin__.bool>',
                                            'longname': 'storable',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'st'},
                               'type': {'args': '<type __builtin__.str>',
                                        'longname': 'type',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'typ'},
                               'typeExact': {'args': '<type __builtin__.str>',
                                             'longname': 'typeExact',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'tex'},
                               'usedAsColor': {'args': '<type __builtin__.bool>',
                                               'longname': 'usedAsColor',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'uac'},
                               'usedAsFilename': {'args': '<type __builtin__.bool>',
                                                  'longname': 'usedAsFilename',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'uaf'},
                               'usesMultiBuilder': {'args': '<type __builtin__.bool>',
                                                    'longname': 'usesMultiBuilder',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'umb'},
                               'worldspace': {'args': '<type __builtin__.bool>',
                                              'longname': 'worldspace',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ws'},
                               'writable': {'args': '<type __builtin__.bool>',
                                            'longname': 'writable',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'w'}},
                     'removedFlags': {'is': 'internalSet'},
                     'shortFlags': {'aa': 'affectsAppearance',
                                    'at': 'attributeType',
                                    'aws': 'affectsWorldspace',
                                    'c': 'connectable',
                                    'ch': 'channelBox',
                                    'ci': 'cachedInternally',
                                    'ct': 'categories',
                                    'e': 'enum',
                                    'ex': 'exists',
                                    'h': 'hidden',
                                    'i': 'internal',
                                    'idt': 'indeterminant',
                                    'ig': 'internalGet',
                                    'im': 'indexMatters',
                                    'internalSet': 'internalSet',
                                    'k': 'keyable',
                                    'lc': 'listChildren',
                                    'ld': 'listDefault',
                                    'le': 'listEnum',
                                    'ln': 'longName',
                                    'lp': 'listParent',
                                    'ls': 'listSiblings',
                                    'm': 'multi',
                                    'max': 'maximum',
                                    'min': 'minimum',
                                    'mne': 'minExists',
                                    'msg': 'message',
                                    'mxe': 'maxExists',
                                    'n': 'node',
                                    'nc': 'numberOfChildren',
                                    'nn': 'niceName',
                                    'r': 'range',
                                    'rd': 'readable',
                                    're': 'rangeExists',
                                    'rs': 'renderSource',
                                    's': 'softRange',
                                    'se': 'softRangeExists',
                                    'sme': 'softMinExists',
                                    'smn': 'softMin',
                                    'smx': 'softMax',
                                    'sn': 'shortName',
                                    'st': 'storable',
                                    'sxe': 'softMaxExists',
                                    'tex': 'typeExact',
                                    'typ': 'type',
                                    'uac': 'usedAsColor',
                                    'uaf': 'usedAsFilename',
                                    'umb': 'usesMultiBuilder',
                                    'w': 'writable',
                                    'ws': 'worldspace'},
                     'type': 'general'},
  'audioTrack': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'insertTrack': {'args': '<type __builtin__.int>',
                                           'longname': 'insertTrack',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'it'},
                           'lock': {'args': '<type __builtin__.bool>',
                                    'longname': 'lock',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'l'},
                           'mute': {'args': '<type __builtin__.bool>',
                                    'longname': 'mute',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'm'},
                           'numTracks': {'args': '<type __builtin__.int>',
                                         'longname': 'numTracks',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'nt'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'removeEmptyTracks': {'args': '<type __builtin__.bool>',
                                                 'longname': 'removeEmptyTracks',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ret'},
                           'removeTrack': {'args': '<type __builtin__.int>',
                                           'longname': 'removeTrack',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'rt'},
                           'solo': {'args': '<type __builtin__.bool>',
                                    'longname': 'solo',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'so'},
                           'swapTracks': {'args': ['<type __builtin__.int>',
                                                   '<type __builtin__.int>'],
                                          'longname': 'swapTracks',
                                          'modes': ['create'],
                                          'numArgs': 2,
                                          'shortname': 'st'},
                           'title': {'args': '<type __builtin__.str>',
                                     'longname': 'title',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 't'},
                           'track': {'args': '<type __builtin__.int>',
                                     'longname': 'track',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'tr'}},
                 'shortFlags': {'e': 'edit',
                                'it': 'insertTrack',
                                'l': 'lock',
                                'm': 'mute',
                                'nt': 'numTracks',
                                'q': 'query',
                                'ret': 'removeEmptyTracks',
                                'rt': 'removeTrack',
                                'so': 'solo',
                                'st': 'swapTracks',
                                't': 'title',
                                'tr': 'track'},
                 'type': 'system'},
  'autoKeyframe': {'flags': {'addAttr': {'args': 'PyNode',
                                         'longname': 'addAttr',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 'aa'},
                             'characterOption': {'args': '<type __builtin__.str>',
                                                 'longname': 'characterOption',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'co'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'listAttr': {'args': '<type __builtin__.bool>',
                                          'longname': 'listAttr',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'lsa'},
                             'noReset': {'args': '<type __builtin__.bool>',
                                         'longname': 'noReset',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'nr'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'state': {'args': '<type __builtin__.bool>',
                                       'longname': 'state',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'st'}},
                   'shortFlags': {'aa': 'addAttr',
                                  'co': 'characterOption',
                                  'e': 'edit',
                                  'lsa': 'listAttr',
                                  'nr': 'noReset',
                                  'q': 'query',
                                  'st': 'state'},
                   'type': 'animation'},
  'autoPlace': {'flags': {'useMouse': {'args': '<type __builtin__.bool>',
                                       'longname': 'useMouse',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'um'}},
                'shortFlags': {'um': 'useMouse'},
                'type': 'windows'},
  'autoSave': {'flags': {'destination': {'args': '<type __builtin__.int>',
                                         'longname': 'destination',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'dst'},
                         'destinationFolder': {'args': '<type __builtin__.bool>',
                                               'longname': 'destinationFolder',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'df'},
                         'enable': {'args': '<type __builtin__.bool>',
                                    'longname': 'enable',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'en'},
                         'folder': {'args': '<type __builtin__.str>',
                                    'longname': 'folder',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'fol'},
                         'interval': {'args': '<type __builtin__.float>',
                                      'longname': 'interval',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'int'},
                         'limitBackups': {'args': '<type __builtin__.bool>',
                                          'longname': 'limitBackups',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'lim'},
                         'maxBackups': {'args': '<type __builtin__.int>',
                                        'longname': 'maxBackups',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'max'},
                         'perform': {'args': '<type __builtin__.bool>',
                                     'longname': 'perform',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'p'},
                         'prompt': {'args': '<type __builtin__.bool>',
                                    'longname': 'prompt',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'prm'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'}},
               'shortFlags': {'df': 'destinationFolder',
                              'dst': 'destination',
                              'en': 'enable',
                              'fol': 'folder',
                              'int': 'interval',
                              'lim': 'limitBackups',
                              'max': 'maxBackups',
                              'p': 'perform',
                              'prm': 'prompt',
                              'q': 'query'},
               'type': 'system'},
  'backgroundEvaluationManager': {'flags': {'interrupt': {'args': '<type __builtin__.bool>',
                                                          'longname': 'interrupt',
                                                          'modes': ['create',
                                                                    'query'],
                                                          'numArgs': 1,
                                                          'shortname': 'i'},
                                            'mode': {'args': '<type __builtin__.str>',
                                                     'longname': 'mode',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'm'},
                                            'pause': {'args': '<type __builtin__.bool>',
                                                      'longname': 'pause',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 0,
                                                      'shortname': 'p'},
                                            'query': {'args': '<type __builtin__.bool>',
                                                      'longname': 'query',
                                                      'numArgs': 0,
                                                      'shortname': 'q'},
                                            'resume': {'args': '<type __builtin__.bool>',
                                                       'longname': 'resume',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'r'}},
                                  'shortFlags': {'i': 'interrupt',
                                                 'm': 'mode',
                                                 'p': 'pause',
                                                 'q': 'query',
                                                 'r': 'resume'},
                                  'type': 'animation'},
  'bakeClip': {'flags': {'blend': {'args': ['<type __builtin__.int>',
                                            '<type __builtin__.int>'],
                                   'longname': 'blend',
                                   'modes': ['create'],
                                   'numArgs': 2,
                                   'shortname': 'b'},
                         'clipIndex': {'args': '<type __builtin__.int>',
                                       'longname': 'clipIndex',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'ci'},
                         'keepOriginals': {'args': '<type __builtin__.bool>',
                                           'longname': 'keepOriginals',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'k'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'}},
               'shortFlags': {'b': 'blend',
                              'ci': 'clipIndex',
                              'k': 'keepOriginals',
                              'n': 'name'},
               'type': 'animation'},
  'bakeDeformer': {'flags': {'colorizeSkeleton': {'args': '<type __builtin__.bool>',
                                                  'longname': 'colorizeSkeleton',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'cs'},
                             'customRangeOfMotion': {'args': 'timerange',
                                                     'longname': 'customRangeOfMotion',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'rom'},
                             'dstMeshName': {'args': '<type __builtin__.str>',
                                             'longname': 'dstMeshName',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dm'},
                             'dstSkeletonName': {'args': '<type __builtin__.str>',
                                                 'longname': 'dstSkeletonName',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ds'},
                             'hierarchy': {'args': '<type __builtin__.bool>',
                                           'longname': 'hierarchy',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'hi'},
                             'influences': {'args': '<type __builtin__.str>',
                                            'longname': 'influences',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'i'},
                             'maxInfluences': {'args': '<type __builtin__.int>',
                                               'longname': 'maxInfluences',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'mi'},
                             'pruneWeights': {'args': '<type __builtin__.float>',
                                              'longname': 'pruneWeights',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'pw'},
                             'smoothWeights': {'args': '<type __builtin__.int>',
                                               'longname': 'smoothWeights',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'sw'},
                             'srcMeshName': {'args': '<type __builtin__.str>',
                                             'longname': 'srcMeshName',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'sm'},
                             'srcSkeletonName': {'args': '<type __builtin__.str>',
                                                 'longname': 'srcSkeletonName',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ss'}},
                   'shortFlags': {'cs': 'colorizeSkeleton',
                                  'dm': 'dstMeshName',
                                  'ds': 'dstSkeletonName',
                                  'hi': 'hierarchy',
                                  'i': 'influences',
                                  'mi': 'maxInfluences',
                                  'pw': 'pruneWeights',
                                  'rom': 'customRangeOfMotion',
                                  'sm': 'srcMeshName',
                                  'ss': 'srcSkeletonName',
                                  'sw': 'smoothWeights'},
                   'type': 'animation'},
  'bakePartialHistory': {'flags': {'allShapes': {'args': '<type __builtin__.bool>',
                                                 'longname': 'allShapes',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 0,
                                                 'shortname': 'all'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'postSmooth': {'args': '<type __builtin__.bool>',
                                                  'longname': 'postSmooth',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'nps'},
                                   'preCache': {'args': '<type __builtin__.bool>',
                                                'longname': 'preCache',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 0,
                                                'shortname': 'pc'},
                                   'preDeformers': {'args': '<type __builtin__.bool>',
                                                    'longname': 'preDeformers',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 0,
                                                    'shortname': 'pre'},
                                   'prePostDeformers': {'args': '<type __builtin__.bool>',
                                                        'longname': 'prePostDeformers',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 0,
                                                        'shortname': 'ppt'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'}},
                         'shortFlags': {'all': 'allShapes',
                                        'e': 'edit',
                                        'nps': 'postSmooth',
                                        'pc': 'preCache',
                                        'ppt': 'prePostDeformers',
                                        'pre': 'preDeformers',
                                        'q': 'query'},
                         'type': 'general'},
  'bakeResults': {'flags': {'animation': {'args': '<type __builtin__.str>',
                                          'longname': 'animation',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'an'},
                            'attribute': {'args': '<type __builtin__.str>',
                                          'longname': 'attribute',
                                          'modes': ['create', 'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'at'},
                            'bakeOnOverrideLayer': {'args': '<type __builtin__.bool>',
                                                    'longname': 'bakeOnOverrideLayer',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'bol'},
                            'controlPoints': {'args': '<type __builtin__.bool>',
                                              'longname': 'controlPoints',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'cp'},
                            'destinationLayer': {'args': '<type __builtin__.str>',
                                                 'longname': 'destinationLayer',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'dl'},
                            'disableImplicitControl': {'args': '<type __builtin__.bool>',
                                                       'longname': 'disableImplicitControl',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'dic'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'float': {'args': 'floatrange',
                                      'longname': 'float',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'f'},
                            'hierarchy': {'args': '<type __builtin__.str>',
                                          'longname': 'hierarchy',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'hi'},
                            'includeUpperBound': {'args': '<type __builtin__.bool>',
                                                  'longname': 'includeUpperBound',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'iub'},
                            'index': {'args': '<type __builtin__.int>',
                                      'longname': 'index',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'index'},
                            'minimizeRotation': {'args': '<type __builtin__.bool>',
                                                 'longname': 'minimizeRotation',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'mr'},
                            'oversamplingRate': {'args': '<type __builtin__.int>',
                                                 'longname': 'oversamplingRate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'osr'},
                            'preserveOutsideKeys': {'args': '<type __builtin__.bool>',
                                                    'longname': 'preserveOutsideKeys',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'pok'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'removeBakedAnimFromLayer': {'args': '<type __builtin__.bool>',
                                                         'longname': 'removeBakedAnimFromLayer',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'rba'},
                            'removeBakedAttributeFromLayer': {'args': '<type __builtin__.bool>',
                                                              'longname': 'removeBakedAttributeFromLayer',
                                                              'modes': ['create'],
                                                              'numArgs': 1,
                                                              'shortname': 'ral'},
                            'resolveWithoutLayer': {'args': '<type __builtin__.str>',
                                                    'longname': 'resolveWithoutLayer',
                                                    'modes': ['create',
                                                              'multiuse'],
                                                    'numArgs': 1,
                                                    'shortname': 'rwl'},
                            'sampleBy': {'args': 'time',
                                         'longname': 'sampleBy',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'sb'},
                            'shape': {'args': '<type __builtin__.bool>',
                                      'longname': 'shape',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                            'simulation': {'args': '<type __builtin__.bool>',
                                           'longname': 'simulation',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'sm'},
                            'smart': {'args': ['<type __builtin__.bool>',
                                               '<type __builtin__.float>'],
                                      'longname': 'smart',
                                      'modes': ['create'],
                                      'numArgs': 2,
                                      'shortname': 'sr'},
                            'sparseAnimCurveBake': {'args': '<type __builtin__.bool>',
                                                    'longname': 'sparseAnimCurveBake',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'sac'},
                            'time': {'args': 'timerange',
                                     'longname': 'time',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 't'}},
                  'removedFlags': {'in': 'index'},
                  'shortFlags': {'an': 'animation',
                                 'at': 'attribute',
                                 'bol': 'bakeOnOverrideLayer',
                                 'cp': 'controlPoints',
                                 'dic': 'disableImplicitControl',
                                 'dl': 'destinationLayer',
                                 'e': 'edit',
                                 'f': 'float',
                                 'hi': 'hierarchy',
                                 'index': 'index',
                                 'iub': 'includeUpperBound',
                                 'mr': 'minimizeRotation',
                                 'osr': 'oversamplingRate',
                                 'pok': 'preserveOutsideKeys',
                                 'q': 'query',
                                 'ral': 'removeBakedAttributeFromLayer',
                                 'rba': 'removeBakedAnimFromLayer',
                                 'rwl': 'resolveWithoutLayer',
                                 's': 'shape',
                                 'sac': 'sparseAnimCurveBake',
                                 'sb': 'sampleBy',
                                 'sm': 'simulation',
                                 'sr': 'smart',
                                 't': 'time'},
                  'type': 'animation'},
  'bakeSimulation': {'flags': {'animation': {'args': '<type __builtin__.str>',
                                             'longname': 'animation',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'an'},
                               'attribute': {'args': '<type __builtin__.str>',
                                             'longname': 'attribute',
                                             'modes': ['create',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'at'},
                               'bakeOnOverrideLayer': {'args': '<type __builtin__.bool>',
                                                       'longname': 'bakeOnOverrideLayer',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'bol'},
                               'controlPoints': {'args': '<type __builtin__.bool>',
                                                 'longname': 'controlPoints',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'cp'},
                               'destinationLayer': {'args': '<type __builtin__.str>',
                                                    'longname': 'destinationLayer',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'dl'},
                               'disableImplicitControl': {'args': '<type __builtin__.bool>',
                                                          'longname': 'disableImplicitControl',
                                                          'modes': ['create'],
                                                          'numArgs': 1,
                                                          'shortname': 'dic'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'float': {'args': 'floatrange',
                                         'longname': 'float',
                                         'modes': ['create', 'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'f'},
                               'hierarchy': {'args': '<type __builtin__.str>',
                                             'longname': 'hierarchy',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'hi'},
                               'includeUpperBound': {'args': '<type __builtin__.bool>',
                                                     'longname': 'includeUpperBound',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'iub'},
                               'index': {'args': '<type __builtin__.int>',
                                         'longname': 'index',
                                         'modes': ['create', 'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'index'},
                               'minimizeRotation': {'args': '<type __builtin__.bool>',
                                                    'longname': 'minimizeRotation',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'mr'},
                               'oversamplingRate': {'args': '<type __builtin__.int>',
                                                    'longname': 'oversamplingRate',
                                                    'numArgs': 1,
                                                    'shortname': 'osr'},
                               'preserveOutsideKeys': {'args': '<type __builtin__.bool>',
                                                       'longname': 'preserveOutsideKeys',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'pok'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'removeBakedAnimFromLayer': {'args': '<type __builtin__.bool>',
                                                            'longname': 'removeBakedAnimFromLayer',
                                                            'modes': ['create'],
                                                            'numArgs': 1,
                                                            'shortname': 'rba'},
                               'removeBakedAttributeFromLayer': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'removeBakedAttributeFromLayer',
                                                                 'modes': ['create'],
                                                                 'numArgs': 1,
                                                                 'shortname': 'ral'},
                               'resolveWithoutLayer': {'args': '<type __builtin__.str>',
                                                       'longname': 'resolveWithoutLayer',
                                                       'modes': ['create',
                                                                 'multiuse'],
                                                       'numArgs': 1,
                                                       'shortname': 'rwl'},
                               'sampleBy': {'args': 'time',
                                            'longname': 'sampleBy',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'sb'},
                               'shape': {'args': '<type __builtin__.bool>',
                                         'longname': 'shape',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 's'},
                               'simulation': {'args': '<type __builtin__.bool>',
                                              'longname': 'simulation',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'sm'},
                               'smart': {'args': ['<type __builtin__.bool>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'smart',
                                         'modes': ['create'],
                                         'numArgs': 2,
                                         'shortname': 'sr'},
                               'sparseAnimCurveBake': {'args': '<type __builtin__.bool>',
                                                       'longname': 'sparseAnimCurveBake',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'sac'},
                               'time': {'args': 'timerange',
                                        'longname': 'time',
                                        'modes': ['create', 'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 't'}},
                     'removedFlags': {'in': 'index'},
                     'shortFlags': {'an': 'animation',
                                    'at': 'attribute',
                                    'bol': 'bakeOnOverrideLayer',
                                    'cp': 'controlPoints',
                                    'dic': 'disableImplicitControl',
                                    'dl': 'destinationLayer',
                                    'e': 'edit',
                                    'f': 'float',
                                    'hi': 'hierarchy',
                                    'index': 'index',
                                    'iub': 'includeUpperBound',
                                    'mr': 'minimizeRotation',
                                    'osr': 'oversamplingRate',
                                    'pok': 'preserveOutsideKeys',
                                    'q': 'query',
                                    'ral': 'removeBakedAttributeFromLayer',
                                    'rba': 'removeBakedAnimFromLayer',
                                    'rwl': 'resolveWithoutLayer',
                                    's': 'shape',
                                    'sac': 'sparseAnimCurveBake',
                                    'sb': 'sampleBy',
                                    'sm': 'simulation',
                                    'sr': 'smart',
                                    't': 'time'},
                     'type': 'animation'},
  'baseTemplate': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fileName': {'args': '<type __builtin__.str>',
                                          'longname': 'fileName',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'fn'},
                             'force': {'args': '<type __builtin__.bool>',
                                       'longname': 'force',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'f'},
                             'load': {'args': '<type __builtin__.bool>',
                                      'longname': 'load',
                                      'modes': [],
                                      'numArgs': 0,
                                      'shortname': 'l'},
                             'matchFile': {'args': '<type __builtin__.str>',
                                           'longname': 'matchFile',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 'mf'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'silent': {'args': '<type __builtin__.bool>',
                                        'longname': 'silent',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 0,
                                        'shortname': 'si'},
                             'unload': {'args': '<type __builtin__.bool>',
                                        'longname': 'unload',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'u'},
                             'viewList': {'args': '<type __builtin__.str>',
                                          'longname': 'viewList',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'vl'}},
                   'shortFlags': {'e': 'edit',
                                  'ex': 'exists',
                                  'f': 'force',
                                  'fn': 'fileName',
                                  'l': 'load',
                                  'mf': 'matchFile',
                                  'q': 'query',
                                  'si': 'silent',
                                  'u': 'unload',
                                  'vl': 'viewList'},
                   'type': 'general'},
  'baseView': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'itemInfo': {'args': '<type __builtin__.str>',
                                      'longname': 'itemInfo',
                                      'modes': ['query'],
                                      'numArgs': 1,
                                      'shortname': 'ii'},
                         'itemList': {'args': '<type __builtin__.bool>',
                                      'longname': 'itemList',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'il'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'viewDescription': {'args': '<type __builtin__.bool>',
                                             'longname': 'viewDescription',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'vd'},
                         'viewLabel': {'args': '<type __builtin__.bool>',
                                       'longname': 'viewLabel',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'vb'},
                         'viewList': {'args': '<type __builtin__.bool>',
                                      'longname': 'viewList',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'vl'},
                         'viewName': {'args': '<type __builtin__.str>',
                                      'longname': 'viewName',
                                      'modes': ['query'],
                                      'numArgs': 1,
                                      'shortname': 'vn'}},
               'shortFlags': {'e': 'edit',
                              'ii': 'itemInfo',
                              'il': 'itemList',
                              'q': 'query',
                              'vb': 'viewLabel',
                              'vd': 'viewDescription',
                              'vl': 'viewList',
                              'vn': 'viewName'},
               'type': 'general'},
  'batchRender': {'callbackFlags': ['preRenderCommand',
                                    'prc',
                                    'melCommand',
                                    'mc',
                                    'renderCommandOptions',
                                    'rco'],
                  'flags': {'filename': {'args': '<type __builtin__.str>',
                                         'longname': 'filename',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'f'},
                            'melCommand': {'args': '<type __builtin__.str>',
                                           'longname': 'melCommand',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'mc'},
                            'numProcs': {'args': '<type __builtin__.int>',
                                         'longname': 'numProcs',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                            'preRenderCommand': {'args': '<type __builtin__.str>',
                                                 'longname': 'preRenderCommand',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'prc'},
                            'remoteRenderMachine': {'args': '<type __builtin__.str>',
                                                    'longname': 'remoteRenderMachine',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'rm'},
                            'renderCommandOptions': {'args': '<type __builtin__.str>',
                                                     'longname': 'renderCommandOptions',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'rco'},
                            'showImage': {'args': '<type __builtin__.bool>',
                                          'longname': 'showImage',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'si'},
                            'status': {'args': '<type __builtin__.str>',
                                       'longname': 'status',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'st'},
                            'useRemoteRender': {'args': '<type __builtin__.bool>',
                                                'longname': 'useRemoteRender',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'um'},
                            'useStandalone': {'args': '<type __builtin__.bool>',
                                              'longname': 'useStandalone',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'us'},
                            'verbosity': {'args': '<type __builtin__.int>',
                                          'longname': 'verbosity',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'v'}},
                  'shortFlags': {'f': 'filename',
                                 'mc': 'melCommand',
                                 'n': 'numProcs',
                                 'prc': 'preRenderCommand',
                                 'rco': 'renderCommandOptions',
                                 'rm': 'remoteRenderMachine',
                                 'si': 'showImage',
                                 'st': 'status',
                                 'um': 'useRemoteRender',
                                 'us': 'useStandalone',
                                 'v': 'verbosity'},
                  'type': 'rendering'},
  'bevel': {'flags': {'bevelShapeType': {'args': '<type __builtin__.int>',
                                         'longname': 'bevelShapeType',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'bst'},
                      'caching': {'args': '<type __builtin__.bool>',
                                  'longname': 'caching',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'cch'},
                      'constructionHistory': {'args': '<type __builtin__.bool>',
                                              'longname': 'constructionHistory',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ch'},
                      'cornerType': {'args': '<type __builtin__.int>',
                                     'longname': 'cornerType',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'ct'},
                      'depth': {'args': '<type __builtin__.float>',
                                'longname': 'depth',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'd'},
                      'edit': {'args': '<type __builtin__.bool>',
                               'longname': 'edit',
                               'numArgs': 0,
                               'shortname': 'e'},
                      'extrudeDepth': {'args': '<type __builtin__.float>',
                                       'longname': 'extrudeDepth',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ed'},
                      'frozen': {'args': '<type __builtin__.bool>',
                                 'longname': 'frozen',
                                 'numArgs': 1,
                                 'shortname': 'fzn'},
                      'joinSurfaces': {'args': '<type __builtin__.bool>',
                                       'longname': 'joinSurfaces',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'js'},
                      'name': {'args': '<type __builtin__.str>',
                               'longname': 'name',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'n'},
                      'nodeState': {'args': '<type __builtin__.int>',
                                    'longname': 'nodeState',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'nds'},
                      'numberOfSides': {'args': '<type __builtin__.int>',
                                        'longname': 'numberOfSides',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ns'},
                      'object': {'args': '<type __builtin__.bool>',
                                 'longname': 'object',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'o'},
                      'polygon': {'args': '<type __builtin__.int>',
                                  'longname': 'polygon',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'po'},
                      'query': {'args': '<type __builtin__.bool>',
                                'longname': 'query',
                                'numArgs': 0,
                                'shortname': 'q'},
                      'range': {'args': '<type __builtin__.bool>',
                                'longname': 'range',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'rn'},
                      'tolerance': {'args': '<type __builtin__.float>',
                                    'longname': 'tolerance',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'tol'},
                      'width': {'args': '<type __builtin__.float>',
                                'longname': 'width',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'w'}},
            'shortFlags': {'bst': 'bevelShapeType',
                           'cch': 'caching',
                           'ch': 'constructionHistory',
                           'ct': 'cornerType',
                           'd': 'depth',
                           'e': 'edit',
                           'ed': 'extrudeDepth',
                           'fzn': 'frozen',
                           'js': 'joinSurfaces',
                           'n': 'name',
                           'nds': 'nodeState',
                           'ns': 'numberOfSides',
                           'o': 'object',
                           'po': 'polygon',
                           'q': 'query',
                           'rn': 'range',
                           'tol': 'tolerance',
                           'w': 'width'},
            'type': 'modeling'},
  'bevelPlus': {'flags': {'bevelInside': {'args': '<type __builtin__.bool>',
                                          'longname': 'bevelInside',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'bin'},
                          'caching': {'args': '<type __builtin__.bool>',
                                      'longname': 'caching',
                                      'numArgs': 1,
                                      'shortname': 'cch'},
                          'capSides': {'args': '<type __builtin__.int>',
                                       'longname': 'capSides',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'cap'},
                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'constructionHistory',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                          'depth': {'args': '<type __builtin__.float>',
                                    'longname': 'depth',
                                    'numArgs': 1,
                                    'shortname': 'd'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'extrudeDepth': {'args': '<type __builtin__.float>',
                                           'longname': 'extrudeDepth',
                                           'numArgs': 1,
                                           'shortname': 'ed'},
                          'frozen': {'args': '<type __builtin__.bool>',
                                     'longname': 'frozen',
                                     'numArgs': 1,
                                     'shortname': 'fzn'},
                          'innerStyle': {'args': '<type __builtin__.int>',
                                         'longname': 'innerStyle',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'innerStyle'},
                          'joinSurfaces': {'args': '<type __builtin__.bool>',
                                           'longname': 'joinSurfaces',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'js'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nodeState': {'args': '<type __builtin__.int>',
                                        'longname': 'nodeState',
                                        'numArgs': 1,
                                        'shortname': 'nds'},
                          'normalsOutwards': {'args': '<type __builtin__.bool>',
                                              'longname': 'normalsOutwards',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'no'},
                          'numberOfSides': {'args': '<type __builtin__.int>',
                                            'longname': 'numberOfSides',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ns'},
                          'outerStyle': {'args': '<type __builtin__.int>',
                                         'longname': 'outerStyle',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'os'},
                          'polyOutChordHeight': {'args': '<type __builtin__.float>',
                                                 'longname': 'polyOutChordHeight',
                                                 'numArgs': 1,
                                                 'shortname': 'cht'},
                          'polyOutChordHeightRatio': {'args': '<type __builtin__.float>',
                                                      'longname': 'polyOutChordHeightRatio',
                                                      'numArgs': 1,
                                                      'shortname': 'chr'},
                          'polyOutCount': {'args': '<type __builtin__.int>',
                                           'longname': 'polyOutCount',
                                           'numArgs': 1,
                                           'shortname': 'poc'},
                          'polyOutCurveSamples': {'args': '<type __builtin__.int>',
                                                  'longname': 'polyOutCurveSamples',
                                                  'numArgs': 1,
                                                  'shortname': 'pcs'},
                          'polyOutCurveType': {'args': '<type __builtin__.int>',
                                               'longname': 'polyOutCurveType',
                                               'numArgs': 1,
                                               'shortname': 'pct'},
                          'polyOutExtrusionSamples': {'args': '<type __builtin__.int>',
                                                      'longname': 'polyOutExtrusionSamples',
                                                      'numArgs': 1,
                                                      'shortname': 'pes'},
                          'polyOutExtrusionType': {'args': '<type __builtin__.int>',
                                                   'longname': 'polyOutExtrusionType',
                                                   'numArgs': 1,
                                                   'shortname': 'pet'},
                          'polyOutMethod': {'args': '<type __builtin__.int>',
                                            'longname': 'polyOutMethod',
                                            'numArgs': 1,
                                            'shortname': 'pom'},
                          'polyOutUseChordHeight': {'args': '<type __builtin__.bool>',
                                                    'longname': 'polyOutUseChordHeight',
                                                    'numArgs': 1,
                                                    'shortname': 'uch'},
                          'polyOutUseChordHeightRatio': {'args': '<type __builtin__.bool>',
                                                         'longname': 'polyOutUseChordHeightRatio',
                                                         'numArgs': 1,
                                                         'shortname': 'ucr'},
                          'polygon': {'args': '<type __builtin__.int>',
                                      'longname': 'polygon',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'po'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'range': {'args': '<type __builtin__.bool>',
                                    'longname': 'range',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'rn'},
                          'tolerance': {'args': '<type __builtin__.float>',
                                        'longname': 'tolerance',
                                        'numArgs': 1,
                                        'shortname': 'tol'},
                          'width': {'args': '<type __builtin__.float>',
                                    'longname': 'width',
                                    'numArgs': 1,
                                    'shortname': 'w'}},
                'removedFlags': {'is': 'innerStyle'},
                'shortFlags': {'bin': 'bevelInside',
                               'cap': 'capSides',
                               'cch': 'caching',
                               'ch': 'constructionHistory',
                               'chr': 'polyOutChordHeightRatio',
                               'cht': 'polyOutChordHeight',
                               'd': 'depth',
                               'e': 'edit',
                               'ed': 'extrudeDepth',
                               'fzn': 'frozen',
                               'innerStyle': 'innerStyle',
                               'js': 'joinSurfaces',
                               'n': 'name',
                               'nds': 'nodeState',
                               'no': 'normalsOutwards',
                               'ns': 'numberOfSides',
                               'os': 'outerStyle',
                               'pcs': 'polyOutCurveSamples',
                               'pct': 'polyOutCurveType',
                               'pes': 'polyOutExtrusionSamples',
                               'pet': 'polyOutExtrusionType',
                               'po': 'polygon',
                               'poc': 'polyOutCount',
                               'pom': 'polyOutMethod',
                               'q': 'query',
                               'rn': 'range',
                               'tol': 'tolerance',
                               'uch': 'polyOutUseChordHeight',
                               'ucr': 'polyOutUseChordHeightRatio',
                               'w': 'width'},
                'type': 'modeling'},
  'bezierAnchorPreset': {'flags': {'preset': {'args': '<type __builtin__.int>',
                                              'longname': 'preset',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'p'}},
                         'shortFlags': {'p': 'preset'},
                         'type': 'modeling'},
  'bezierAnchorState': {'flags': {'even': {'args': '<type __builtin__.bool>',
                                           'longname': 'even',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ev'},
                                  'smooth': {'args': '<type __builtin__.bool>',
                                             'longname': 'smooth',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'sm'}},
                        'shortFlags': {'ev': 'even', 'sm': 'smooth'},
                        'type': 'modeling'},
  'bezierCurveToNurbs': {'flags': {}, 'shortFlags': {}, 'type': 'modeling'},
  'bezierInfo': {'flags': {'anchorFromCV': {'args': '<type __builtin__.int>',
                                            'longname': 'anchorFromCV',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'afc'},
                           'cvFromAnchor': {'args': '<type __builtin__.int>',
                                            'longname': 'cvFromAnchor',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'cfa'},
                           'isAnchorSelected': {'args': '<type __builtin__.bool>',
                                                'longname': 'isAnchorSelected',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ias'},
                           'isTangentSelected': {'args': '<type __builtin__.bool>',
                                                 'longname': 'isTangentSelected',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'its'},
                           'onlyAnchorsSelected': {'args': '<type __builtin__.bool>',
                                                   'longname': 'onlyAnchorsSelected',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'oas'},
                           'onlyTangentsSelected': {'args': '<type __builtin__.bool>',
                                                    'longname': 'onlyTangentsSelected',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ots'}},
                 'shortFlags': {'afc': 'anchorFromCV',
                                'cfa': 'cvFromAnchor',
                                'ias': 'isAnchorSelected',
                                'its': 'isTangentSelected',
                                'oas': 'onlyAnchorsSelected',
                                'ots': 'onlyTangentsSelected'},
                 'type': 'modeling'},
  'binMembership': {'flags': {'addToBin': {'args': '<type __builtin__.str>',
                                           'longname': 'addToBin',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'add'},
                              'exists': {'args': '<type __builtin__.str>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ex'},
                              'inheritBinsFromNodes': {'args': 'PyNode',
                                                       'longname': 'inheritBinsFromNodes',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ibn'},
                              'isValidBinName': {'args': '<type __builtin__.str>',
                                                 'longname': 'isValidBinName',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ivn'},
                              'listBins': {'args': '<type __builtin__.bool>',
                                           'longname': 'listBins',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'lb'},
                              'makeExclusive': {'args': '<type __builtin__.str>',
                                                'longname': 'makeExclusive',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'mke'},
                              'notifyChanged': {'args': '<type __builtin__.bool>',
                                                'longname': 'notifyChanged',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'nfc'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'removeFromBin': {'args': '<type __builtin__.str>',
                                                'longname': 'removeFromBin',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'rm'}},
                    'shortFlags': {'add': 'addToBin',
                                   'ex': 'exists',
                                   'ibn': 'inheritBinsFromNodes',
                                   'ivn': 'isValidBinName',
                                   'lb': 'listBins',
                                   'mke': 'makeExclusive',
                                   'nfc': 'notifyChanged',
                                   'q': 'query',
                                   'rm': 'removeFromBin'},
                    'type': 'rendering'},
  'bindSkin': {'flags': {'byClosestPoint': {'args': '<type __builtin__.bool>',
                                            'longname': 'byClosestPoint',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'bcp'},
                         'byPartition': {'args': '<type __builtin__.bool>',
                                         'longname': 'byPartition',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'bp'},
                         'colorJoints': {'args': '<type __builtin__.bool>',
                                         'longname': 'colorJoints',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'cj'},
                         'delete': {'args': '<type __builtin__.bool>',
                                    'longname': 'delete',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'd'},
                         'doNotDescend': {'args': '<type __builtin__.bool>',
                                          'longname': 'doNotDescend',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'dnd'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'enable': {'args': '<type __builtin__.bool>',
                                    'longname': 'enable',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'en'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'partition': {'args': '<type __builtin__.str>',
                                       'longname': 'partition',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'toAll': {'args': '<type __builtin__.bool>',
                                   'longname': 'toAll',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'ta'},
                         'toSelectedBones': {'args': '<type __builtin__.bool>',
                                             'longname': 'toSelectedBones',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'tsb'},
                         'toSkeleton': {'args': '<type __builtin__.bool>',
                                        'longname': 'toSkeleton',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ts'},
                         'unbind': {'args': '<type __builtin__.bool>',
                                    'longname': 'unbind',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ub'},
                         'unbindKeepHistory': {'args': '<type __builtin__.bool>',
                                               'longname': 'unbindKeepHistory',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ubk'},
                         'unlock': {'args': '<type __builtin__.bool>',
                                    'longname': 'unlock',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ul'}},
               'shortFlags': {'bcp': 'byClosestPoint',
                              'bp': 'byPartition',
                              'cj': 'colorJoints',
                              'd': 'delete',
                              'dnd': 'doNotDescend',
                              'e': 'edit',
                              'en': 'enable',
                              'n': 'name',
                              'p': 'partition',
                              'q': 'query',
                              'ta': 'toAll',
                              'ts': 'toSkeleton',
                              'tsb': 'toSelectedBones',
                              'ub': 'unbind',
                              'ubk': 'unbindKeepHistory',
                              'ul': 'unlock'},
               'type': 'animation'},
  'blend': {'flags': {'autoDirection': {'args': '<type __builtin__.bool>',
                                        'longname': 'autoDirection',
                                        'numArgs': 1,
                                        'shortname': 'ad'},
                      'caching': {'args': '<type __builtin__.bool>',
                                  'longname': 'caching',
                                  'numArgs': 1,
                                  'shortname': 'cch'},
                      'constructionHistory': {'args': '<type __builtin__.bool>',
                                              'longname': 'constructionHistory',
                                              'numArgs': 1,
                                              'shortname': 'ch'},
                      'crvsInFirstRail': {'args': '<type __builtin__.int>',
                                          'longname': 'crvsInFirstRail',
                                          'numArgs': 1,
                                          'shortname': 'cfr'},
                      'edit': {'args': '<type __builtin__.bool>',
                               'longname': 'edit',
                               'numArgs': 0,
                               'shortname': 'e'},
                      'flipLeft': {'args': '<type __builtin__.bool>',
                                   'longname': 'flipLeft',
                                   'numArgs': 1,
                                   'shortname': 'fl'},
                      'flipRight': {'args': '<type __builtin__.bool>',
                                    'longname': 'flipRight',
                                    'numArgs': 1,
                                    'shortname': 'fr'},
                      'frozen': {'args': '<type __builtin__.bool>',
                                 'longname': 'frozen',
                                 'numArgs': 1,
                                 'shortname': 'fzn'},
                      'leftParameter': {'args': '<type __builtin__.float>',
                                        'longname': 'leftParameter',
                                        'numArgs': 1,
                                        'shortname': 'lp'},
                      'multipleKnots': {'args': '<type __builtin__.bool>',
                                        'longname': 'multipleKnots',
                                        'numArgs': 1,
                                        'shortname': 'mk'},
                      'name': {'args': '<type __builtin__.str>',
                               'longname': 'name',
                               'numArgs': 1,
                               'shortname': 'n'},
                      'nodeState': {'args': '<type __builtin__.int>',
                                    'longname': 'nodeState',
                                    'numArgs': 1,
                                    'shortname': 'nds'},
                      'object': {'args': '<type __builtin__.bool>',
                                 'longname': 'object',
                                 'numArgs': 1,
                                 'shortname': 'o'},
                      'polygon': {'args': '<type __builtin__.int>',
                                  'longname': 'polygon',
                                  'numArgs': 1,
                                  'shortname': 'po'},
                      'positionTolerance': {'args': '<type __builtin__.float>',
                                            'longname': 'positionTolerance',
                                            'numArgs': 1,
                                            'shortname': 'pt'},
                      'query': {'args': '<type __builtin__.bool>',
                                'longname': 'query',
                                'numArgs': 0,
                                'shortname': 'q'},
                      'rightParameter': {'args': '<type __builtin__.float>',
                                         'longname': 'rightParameter',
                                         'numArgs': 1,
                                         'shortname': 'rp'},
                      'tangentTolerance': {'args': '<type __builtin__.float>',
                                           'longname': 'tangentTolerance',
                                           'numArgs': 1,
                                           'shortname': 'tt'}},
            'shortFlags': {'ad': 'autoDirection',
                           'cch': 'caching',
                           'cfr': 'crvsInFirstRail',
                           'ch': 'constructionHistory',
                           'e': 'edit',
                           'fl': 'flipLeft',
                           'fr': 'flipRight',
                           'fzn': 'frozen',
                           'lp': 'leftParameter',
                           'mk': 'multipleKnots',
                           'n': 'name',
                           'nds': 'nodeState',
                           'o': 'object',
                           'po': 'polygon',
                           'pt': 'positionTolerance',
                           'q': 'query',
                           'rp': 'rightParameter',
                           'tt': 'tangentTolerance'},
            'type': 'other'},
  'blend2': {'flags': {'autoAnchor': {'args': '<type __builtin__.bool>',
                                      'longname': 'autoAnchor',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'aa'},
                       'autoNormal': {'args': '<type __builtin__.bool>',
                                      'longname': 'autoNormal',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'an'},
                       'caching': {'args': '<type __builtin__.bool>',
                                   'longname': 'caching',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'cch'},
                       'constructionHistory': {'args': '<type __builtin__.bool>',
                                               'longname': 'constructionHistory',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                       'crvsInFirstRail': {'args': '<type __builtin__.int>',
                                           'longname': 'crvsInFirstRail',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cfr'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'flipLeftNormal': {'args': '<type __builtin__.bool>',
                                          'longname': 'flipLeftNormal',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'fln'},
                       'flipRightNormal': {'args': '<type __builtin__.bool>',
                                           'longname': 'flipRightNormal',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'frn'},
                       'frozen': {'args': '<type __builtin__.bool>',
                                  'longname': 'frozen',
                                  'numArgs': 1,
                                  'shortname': 'fzn'},
                       'leftAnchor': {'args': '<type __builtin__.float>',
                                      'longname': 'leftAnchor',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'la'},
                       'leftEnd': {'args': '<type __builtin__.float>',
                                   'longname': 'leftEnd',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'le'},
                       'leftStart': {'args': '<type __builtin__.float>',
                                     'longname': 'leftStart',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'ls'},
                       'multipleKnots': {'args': '<type __builtin__.bool>',
                                         'longname': 'multipleKnots',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'mk'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'nodeState': {'args': '<type __builtin__.int>',
                                     'longname': 'nodeState',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'nds'},
                       'object': {'args': '<type __builtin__.bool>',
                                  'longname': 'object',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'o'},
                       'polygon': {'args': '<type __builtin__.int>',
                                   'longname': 'polygon',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'po'},
                       'positionTolerance': {'args': '<type __builtin__.float>',
                                             'longname': 'positionTolerance',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'pt'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'reverseLeft': {'args': '<type __builtin__.bool>',
                                       'longname': 'reverseLeft',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'rvl'},
                       'reverseRight': {'args': '<type __builtin__.bool>',
                                        'longname': 'reverseRight',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'rvr'},
                       'rightAnchor': {'args': '<type __builtin__.float>',
                                       'longname': 'rightAnchor',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ra'},
                       'rightEnd': {'args': '<type __builtin__.float>',
                                    'longname': 'rightEnd',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 're'},
                       'rightStart': {'args': '<type __builtin__.float>',
                                      'longname': 'rightStart',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'rs'},
                       'tangentTolerance': {'args': '<type __builtin__.float>',
                                            'longname': 'tangentTolerance',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'tt'}},
             'shortFlags': {'aa': 'autoAnchor',
                            'an': 'autoNormal',
                            'cch': 'caching',
                            'cfr': 'crvsInFirstRail',
                            'ch': 'constructionHistory',
                            'e': 'edit',
                            'fln': 'flipLeftNormal',
                            'frn': 'flipRightNormal',
                            'fzn': 'frozen',
                            'la': 'leftAnchor',
                            'le': 'leftEnd',
                            'ls': 'leftStart',
                            'mk': 'multipleKnots',
                            'n': 'name',
                            'nds': 'nodeState',
                            'o': 'object',
                            'po': 'polygon',
                            'pt': 'positionTolerance',
                            'q': 'query',
                            'ra': 'rightAnchor',
                            're': 'rightEnd',
                            'rs': 'rightStart',
                            'rvl': 'reverseLeft',
                            'rvr': 'reverseRight',
                            'tt': 'tangentTolerance'},
             'type': 'modeling'},
  'blendCtx': {'flags': {'activeNodes': {'args': '<type __builtin__.bool>',
                                         'longname': 'activeNodes',
                                         'numArgs': 0,
                                         'shortname': 'anq'},
                         'autoCreate': {'args': '<type __builtin__.bool>',
                                        'longname': 'autoCreate',
                                        'numArgs': 1,
                                        'shortname': 'ac'},
                         'autoDirection': {'args': '<type __builtin__.bool>',
                                           'longname': 'autoDirection',
                                           'numArgs': 1,
                                           'shortname': 'ad'},
                         'caching': {'args': '<type __builtin__.bool>',
                                     'longname': 'caching',
                                     'numArgs': 1,
                                     'shortname': 'cch'},
                         'constructionHistory': {'args': '<type __builtin__.bool>',
                                                 'longname': 'constructionHistory',
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                         'crvsInFirstRail': {'args': '<type __builtin__.int>',
                                             'longname': 'crvsInFirstRail',
                                             'numArgs': 1,
                                             'shortname': 'cfr'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'flipLeft': {'args': '<type __builtin__.bool>',
                                      'longname': 'flipLeft',
                                      'numArgs': 1,
                                      'shortname': 'fl'},
                         'flipRight': {'args': '<type __builtin__.bool>',
                                       'longname': 'flipRight',
                                       'numArgs': 1,
                                       'shortname': 'fr'},
                         'frozen': {'args': '<type __builtin__.bool>',
                                    'longname': 'frozen',
                                    'numArgs': 1,
                                    'shortname': 'fzn'},
                         'image1': {'args': '<type __builtin__.str>',
                                    'longname': 'image1',
                                    'numArgs': 1,
                                    'shortname': 'i1'},
                         'image2': {'args': '<type __builtin__.str>',
                                    'longname': 'image2',
                                    'numArgs': 1,
                                    'shortname': 'i2'},
                         'image3': {'args': '<type __builtin__.str>',
                                    'longname': 'image3',
                                    'numArgs': 1,
                                    'shortname': 'i3'},
                         'immediate': {'args': '<type __builtin__.bool>',
                                       'longname': 'immediate',
                                       'numArgs': 0,
                                       'shortname': 'im'},
                         'leftParameter': {'args': '<type __builtin__.float>',
                                           'longname': 'leftParameter',
                                           'numArgs': 1,
                                           'shortname': 'lp'},
                         'multipleKnots': {'args': '<type __builtin__.bool>',
                                           'longname': 'multipleKnots',
                                           'numArgs': 1,
                                           'shortname': 'mk'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'nodeState': {'args': '<type __builtin__.int>',
                                       'longname': 'nodeState',
                                       'numArgs': 1,
                                       'shortname': 'nds'},
                         'object': {'args': '<type __builtin__.bool>',
                                    'longname': 'object',
                                    'numArgs': 1,
                                    'shortname': 'o'},
                         'polygon': {'args': '<type __builtin__.int>',
                                     'longname': 'polygon',
                                     'numArgs': 1,
                                     'shortname': 'po'},
                         'positionTolerance': {'args': '<type __builtin__.float>',
                                               'longname': 'positionTolerance',
                                               'numArgs': 1,
                                               'shortname': 'pt'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'reset': {'args': '<type __builtin__.bool>',
                                   'longname': 'reset',
                                   'numArgs': 0,
                                   'shortname': 'rs'},
                         'rightParameter': {'args': '<type __builtin__.float>',
                                            'longname': 'rightParameter',
                                            'numArgs': 1,
                                            'shortname': 'rp'},
                         'tangentTolerance': {'args': '<type __builtin__.float>',
                                              'longname': 'tangentTolerance',
                                              'numArgs': 1,
                                              'shortname': 'tt'},
                         'toolNode': {'args': '<type __builtin__.bool>',
                                      'longname': 'toolNode',
                                      'numArgs': 0,
                                      'shortname': 'tnq'},
                         'useGlobalTol': {'args': '<type __builtin__.bool>',
                                          'longname': 'useGlobalTol',
                                          'numArgs': 1,
                                          'shortname': 'ugt'}},
               'shortFlags': {'ac': 'autoCreate',
                              'ad': 'autoDirection',
                              'anq': 'activeNodes',
                              'cch': 'caching',
                              'cfr': 'crvsInFirstRail',
                              'ch': 'constructionHistory',
                              'e': 'edit',
                              'ex': 'exists',
                              'fl': 'flipLeft',
                              'fr': 'flipRight',
                              'fzn': 'frozen',
                              'i1': 'image1',
                              'i2': 'image2',
                              'i3': 'image3',
                              'im': 'immediate',
                              'lp': 'leftParameter',
                              'mk': 'multipleKnots',
                              'n': 'name',
                              'nds': 'nodeState',
                              'o': 'object',
                              'po': 'polygon',
                              'pt': 'positionTolerance',
                              'q': 'query',
                              'rp': 'rightParameter',
                              'rs': 'reset',
                              'tnq': 'toolNode',
                              'tt': 'tangentTolerance',
                              'ugt': 'useGlobalTol'},
               'type': 'context'},
  'blendShape': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                     'longname': 'after',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'af'},
                           'afterReference': {'args': '<type __builtin__.bool>',
                                              'longname': 'afterReference',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'ar'},
                           'automatic': {'args': '<type __builtin__.bool>',
                                         'longname': 'automatic',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'at'},
                           'before': {'args': '<type __builtin__.bool>',
                                      'longname': 'before',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'bf'},
                           'copyDelta': {'args': ['<type __builtin__.int>',
                                                  '<type __builtin__.int>',
                                                  '<type __builtin__.int>'],
                                         'longname': 'copyDelta',
                                         'modes': ['edit'],
                                         'numArgs': 3,
                                         'shortname': 'cd'},
                           'copyInBetweenDelta': {'args': ['<type __builtin__.int>',
                                                           '<type __builtin__.int>',
                                                           '<type __builtin__.int>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'copyInBetweenDelta',
                                                  'modes': ['edit'],
                                                  'numArgs': 4,
                                                  'shortname': 'cid'},
                           'deformerTools': {'args': '<type __builtin__.bool>',
                                             'longname': 'deformerTools',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'dt'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'envelope': {'args': '<type __builtin__.float>',
                                        'longname': 'envelope',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                           'exclusive': {'args': '<type __builtin__.str>',
                                         'longname': 'exclusive',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'ex'},
                           'export': {'args': '<type __builtin__.str>',
                                      'longname': 'export',
                                      'modes': ['edit'],
                                      'numArgs': 1,
                                      'shortname': 'ep'},
                           'exportTarget': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'exportTarget',
                                            'modes': ['edit', 'multiuse'],
                                            'numArgs': 2,
                                            'shortname': 'et'},
                           'flipTarget': {'args': ['<type __builtin__.int>',
                                                   '<type __builtin__.int>'],
                                          'longname': 'flipTarget',
                                          'modes': ['edit', 'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'ft'},
                           'frontOfChain': {'args': '<type __builtin__.bool>',
                                            'longname': 'frontOfChain',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'foc'},
                           'geometry': {'args': '<type __builtin__.str>',
                                        'longname': 'geometry',
                                        'modes': ['query',
                                                  'edit',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'g'},
                           'geometryIndices': {'args': '<type __builtin__.bool>',
                                               'longname': 'geometryIndices',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'gi'},
                           'ignoreSelected': {'args': '<type __builtin__.bool>',
                                              'longname': 'ignoreSelected',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ignoreSelected'},
                           'inBetween': {'args': '<type __builtin__.bool>',
                                         'longname': 'inBetween',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'ib'},
                           'inBetweenIndex': {'args': '<type __builtin__.int>',
                                              'longname': 'inBetweenIndex',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'ibi'},
                           'inBetweenType': {'args': '<type __builtin__.str>',
                                             'longname': 'inBetweenType',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ibt'},
                           'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                       'longname': 'includeHiddenSelections',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'ihs'},
                           'ip': {'args': '<type __builtin__.str>',
                                  'longname': 'ip',
                                  'modes': ['edit'],
                                  'numArgs': 1,
                                  'shortname': 'ip'},
                           'mergeSource': {'args': '<type __builtin__.int>',
                                           'longname': 'mergeSource',
                                           'modes': ['edit', 'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'mgs'},
                           'mergeTarget': {'args': '<type __builtin__.int>',
                                           'longname': 'mergeTarget',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'mgt'},
                           'mirrorDirection': {'args': '<type __builtin__.int>',
                                               'longname': 'mirrorDirection',
                                               'modes': ['edit'],
                                               'numArgs': 1,
                                               'shortname': 'md'},
                           'mirrorTarget': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'mirrorTarget',
                                            'modes': ['edit', 'multiuse'],
                                            'numArgs': 2,
                                            'shortname': 'mt'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'normalizationGroups': {'args': '<type __builtin__.bool>',
                                                   'longname': 'normalizationGroups',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'ng'},
                           'origin': {'args': '<type __builtin__.str>',
                                      'longname': 'origin',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'parallel': {'args': '<type __builtin__.bool>',
                                        'longname': 'parallel',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'par'},
                           'prune': {'args': '<type __builtin__.bool>',
                                     'longname': 'prune',
                                     'modes': ['edit'],
                                     'numArgs': 0,
                                     'shortname': 'pr'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'remove': {'args': '<type __builtin__.bool>',
                                      'longname': 'remove',
                                      'modes': ['edit', 'multiuse'],
                                      'numArgs': 0,
                                      'shortname': 'rm'},
                           'resetTargetDelta': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'resetTargetDelta',
                                                'modes': ['edit',
                                                          'multiuse'],
                                                'numArgs': 2,
                                                'shortname': 'rtd'},
                           'split': {'args': '<type __builtin__.bool>',
                                     'longname': 'split',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'sp'},
                           'suppressDialog': {'args': '<type __builtin__.bool>',
                                              'longname': 'suppressDialog',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'sd'},
                           'symmetryAxis': {'args': '<type __builtin__.str>',
                                            'longname': 'symmetryAxis',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sa'},
                           'symmetryEdge': {'args': '<type __builtin__.str>',
                                            'longname': 'symmetryEdge',
                                            'modes': ['query',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'se'},
                           'symmetrySpace': {'args': '<type __builtin__.int>',
                                             'longname': 'symmetrySpace',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ss'},
                           'tangentSpace': {'args': '<type __builtin__.bool>',
                                            'longname': 'tangentSpace',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'ts'},
                           'target': {'args': ['<type __builtin__.str>',
                                               '<type __builtin__.int>',
                                               '<type __builtin__.str>',
                                               '<type __builtin__.float>'],
                                      'longname': 'target',
                                      'modes': ['create',
                                                'query',
                                                'edit',
                                                'multiuse'],
                                      'numArgs': 4,
                                      'shortname': 't'},
                           'topologyCheck': {'args': '<type __builtin__.bool>',
                                             'longname': 'topologyCheck',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'tc'},
                           'transform': {'args': '<type __builtin__.str>',
                                         'longname': 'transform',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'tr'},
                           'weight': {'args': ['<type __builtin__.int>',
                                               '<type __builtin__.float>'],
                                      'longname': 'weight',
                                      'modes': ['create',
                                                'query',
                                                'edit',
                                                'multiuse'],
                                      'numArgs': 2,
                                      'shortname': 'w'},
                           'weightCount': {'args': '<type __builtin__.int>',
                                           'longname': 'weightCount',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'wc'}},
                 'removedFlags': {'import': 'ip', 'is': 'ignoreSelected'},
                 'shortFlags': {'af': 'after',
                                'ar': 'afterReference',
                                'at': 'automatic',
                                'bf': 'before',
                                'cd': 'copyDelta',
                                'cid': 'copyInBetweenDelta',
                                'dt': 'deformerTools',
                                'e': 'edit',
                                'en': 'envelope',
                                'ep': 'export',
                                'et': 'exportTarget',
                                'ex': 'exclusive',
                                'foc': 'frontOfChain',
                                'ft': 'flipTarget',
                                'g': 'geometry',
                                'gi': 'geometryIndices',
                                'ib': 'inBetween',
                                'ibi': 'inBetweenIndex',
                                'ibt': 'inBetweenType',
                                'ignoreSelected': 'ignoreSelected',
                                'ihs': 'includeHiddenSelections',
                                'ip': 'ip',
                                'md': 'mirrorDirection',
                                'mgs': 'mergeSource',
                                'mgt': 'mergeTarget',
                                'mt': 'mirrorTarget',
                                'n': 'name',
                                'ng': 'normalizationGroups',
                                'o': 'origin',
                                'par': 'parallel',
                                'pr': 'prune',
                                'q': 'query',
                                'rm': 'remove',
                                'rtd': 'resetTargetDelta',
                                'sa': 'symmetryAxis',
                                'sd': 'suppressDialog',
                                'se': 'symmetryEdge',
                                'sp': 'split',
                                'ss': 'symmetrySpace',
                                't': 'target',
                                'tc': 'topologyCheck',
                                'tr': 'transform',
                                'ts': 'tangentSpace',
                                'w': 'weight',
                                'wc': 'weightCount'},
                 'type': 'animation'},
  'blendShapeEditor': {'flags': {'control': {'args': '<type __builtin__.bool>',
                                             'longname': 'control',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'ctl'},
                                 'defineTemplate': {'args': '<type __builtin__.str>',
                                                    'longname': 'defineTemplate',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'dt'},
                                 'docTag': {'args': '<type __builtin__.str>',
                                            'longname': 'docTag',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dtg'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'filter': {'args': '<type __builtin__.str>',
                                            'longname': 'filter',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'f'},
                                 'forceMainConnection': {'args': '<type __builtin__.str>',
                                                         'longname': 'forceMainConnection',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'fmc'},
                                 'highlightConnection': {'args': '<type __builtin__.str>',
                                                         'longname': 'highlightConnection',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'hlc'},
                                 'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                        'longname': 'lockMainConnection',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'lck'},
                                 'mainListConnection': {'args': '<type __builtin__.str>',
                                                        'longname': 'mainListConnection',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'mlc'},
                                 'panel': {'args': '<type __builtin__.str>',
                                           'longname': 'panel',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'pnl'},
                                 'parent': {'args': '<type __builtin__.str>',
                                            'longname': 'parent',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'p'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'selectionConnection': {'args': '<type __builtin__.str>',
                                                         'longname': 'selectionConnection',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'slc'},
                                 'stateString': {'args': '<type __builtin__.bool>',
                                                 'longname': 'stateString',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'sts'},
                                 'targetControlList': {'args': '<type __builtin__.bool>',
                                                       'longname': 'targetControlList',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'tcl'},
                                 'targetList': {'args': '<type __builtin__.bool>',
                                                'longname': 'targetList',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'tl'},
                                 'unParent': {'args': '<type __builtin__.bool>',
                                              'longname': 'unParent',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'up'},
                                 'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                          'longname': 'unlockMainConnection',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 0,
                                                          'shortname': 'ulk'},
                                 'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                          'longname': 'updateMainConnection',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 0,
                                                          'shortname': 'upd'},
                                 'useTemplate': {'args': '<type __builtin__.str>',
                                                 'longname': 'useTemplate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ut'},
                                 'verticalSliders': {'args': '<type __builtin__.bool>',
                                                     'longname': 'verticalSliders',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vs'}},
                       'shortFlags': {'ctl': 'control',
                                      'dt': 'defineTemplate',
                                      'dtg': 'docTag',
                                      'e': 'edit',
                                      'ex': 'exists',
                                      'f': 'filter',
                                      'fmc': 'forceMainConnection',
                                      'hlc': 'highlightConnection',
                                      'lck': 'lockMainConnection',
                                      'mlc': 'mainListConnection',
                                      'p': 'parent',
                                      'pnl': 'panel',
                                      'q': 'query',
                                      'slc': 'selectionConnection',
                                      'sts': 'stateString',
                                      'tcl': 'targetControlList',
                                      'tl': 'targetList',
                                      'ulk': 'unlockMainConnection',
                                      'up': 'unParent',
                                      'upd': 'updateMainConnection',
                                      'ut': 'useTemplate',
                                      'vs': 'verticalSliders'},
                       'type': 'animation'},
  'blendShapePanel': {'callbackFlags': ['popupMenuProcedure', 'pmp'],
                      'flags': {'blendShapeEditor': {'args': '<type __builtin__.bool>',
                                                     'longname': 'blendShapeEditor',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'be'},
                                'control': {'args': '<type __builtin__.bool>',
                                            'longname': 'control',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ctl'},
                                'copy': {'args': '<type __builtin__.str>',
                                         'longname': 'copy',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 'cp'},
                                'createString': {'args': '<type __builtin__.bool>',
                                                 'longname': 'createString',
                                                 'modes': ['edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'cs'},
                                'defineTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'defineTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'dt'},
                                'docTag': {'args': '<type __builtin__.str>',
                                           'longname': 'docTag',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dtg'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'editString': {'args': '<type __builtin__.bool>',
                                               'longname': 'editString',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'es'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'init': {'args': '<type __builtin__.bool>',
                                         'longname': 'init',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'init'},
                                'isUnique': {'args': '<type __builtin__.bool>',
                                             'longname': 'isUnique',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'iu'},
                                'label': {'args': '<type __builtin__.str>',
                                          'longname': 'label',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'l'},
                                'menuBarRepeatLast': {'args': '<type __builtin__.bool>',
                                                      'longname': 'menuBarRepeatLast',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'mrl'},
                                'menuBarVisible': {'args': '<type __builtin__.bool>',
                                                   'longname': 'menuBarVisible',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'mbv'},
                                'needsInit': {'args': '<type __builtin__.bool>',
                                              'longname': 'needsInit',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'ni'},
                                'parent': {'args': '<type __builtin__.str>',
                                           'longname': 'parent',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'p'},
                                'popupMenuProcedure': {'args': 'script',
                                                       'longname': 'popupMenuProcedure',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'pmp'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'replacePanel': {'args': '<type __builtin__.str>',
                                                 'longname': 'replacePanel',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rp'},
                                'tearOff': {'args': '<type __builtin__.bool>',
                                            'longname': 'tearOff',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'to'},
                                'tearOffCopy': {'args': '<type __builtin__.str>',
                                                'longname': 'tearOffCopy',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'toc'},
                                'tearOffRestore': {'args': '<type __builtin__.bool>',
                                                   'longname': 'tearOffRestore',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'tor'},
                                'unParent': {'args': '<type __builtin__.bool>',
                                             'longname': 'unParent',
                                             'modes': ['edit'],
                                             'numArgs': 0,
                                             'shortname': 'up'},
                                'useTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'useTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ut'}},
                      'removedFlags': {'in': 'init'},
                      'shortFlags': {'be': 'blendShapeEditor',
                                     'cp': 'copy',
                                     'cs': 'createString',
                                     'ctl': 'control',
                                     'dt': 'defineTemplate',
                                     'dtg': 'docTag',
                                     'e': 'edit',
                                     'es': 'editString',
                                     'ex': 'exists',
                                     'init': 'init',
                                     'iu': 'isUnique',
                                     'l': 'label',
                                     'mbv': 'menuBarVisible',
                                     'mrl': 'menuBarRepeatLast',
                                     'ni': 'needsInit',
                                     'p': 'parent',
                                     'pmp': 'popupMenuProcedure',
                                     'q': 'query',
                                     'rp': 'replacePanel',
                                     'to': 'tearOff',
                                     'toc': 'tearOffCopy',
                                     'tor': 'tearOffRestore',
                                     'up': 'unParent',
                                     'ut': 'useTemplate'},
                      'type': 'animation'},
  'blendTwoAttr': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                           'longname': 'attribute',
                                           'modes': ['create', 'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'at'},
                             'attribute0': {'args': 'PyNode',
                                            'longname': 'attribute0',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'at0'},
                             'attribute1': {'args': 'PyNode',
                                            'longname': 'attribute1',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'at1'},
                             'blender': {'args': 'PyNode',
                                         'longname': 'blender',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'bl'},
                             'controlPoints': {'args': '<type __builtin__.bool>',
                                               'longname': 'controlPoints',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'cp'},
                             'driver': {'args': '<type __builtin__.int>',
                                        'longname': 'driver',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'd'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'shape': {'args': '<type __builtin__.bool>',
                                       'longname': 'shape',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 's'},
                             'time': {'args': 'timerange',
                                      'longname': 'time',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 't'}},
                   'shortFlags': {'at': 'attribute',
                                  'at0': 'attribute0',
                                  'at1': 'attribute1',
                                  'bl': 'blender',
                                  'cp': 'controlPoints',
                                  'd': 'driver',
                                  'e': 'edit',
                                  'n': 'name',
                                  'q': 'query',
                                  's': 'shape',
                                  't': 'time'},
                   'type': 'animation'},
  'blindDataType': {'flags': {'dataType': {'args': '<type __builtin__.str>',
                                           'longname': 'dataType',
                                           'modes': ['create', 'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'dt'},
                              'longDataName': {'args': '<type __builtin__.str>',
                                               'longname': 'longDataName',
                                               'modes': ['create',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'ldn'},
                              'longNames': {'args': '<type __builtin__.bool>',
                                            'longname': 'longNames',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ln'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'shortDataName': {'args': '<type __builtin__.str>',
                                                'longname': 'shortDataName',
                                                'modes': ['create',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'sdn'},
                              'shortNames': {'args': '<type __builtin__.bool>',
                                             'longname': 'shortNames',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'sn'},
                              'typeId': {'args': '<type __builtin__.int>',
                                         'longname': 'typeId',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'id'},
                              'typeNames': {'args': '<type __builtin__.bool>',
                                            'longname': 'typeNames',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'tn'}},
                    'shortFlags': {'dt': 'dataType',
                                   'id': 'typeId',
                                   'ldn': 'longDataName',
                                   'ln': 'longNames',
                                   'q': 'query',
                                   'sdn': 'shortDataName',
                                   'sn': 'shortNames',
                                   'tn': 'typeNames'},
                    'type': 'modeling'},
  'boneLattice': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                      'longname': 'after',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'af'},
                            'afterReference': {'args': '<type __builtin__.bool>',
                                               'longname': 'afterReference',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'ar'},
                            'before': {'args': '<type __builtin__.bool>',
                                       'longname': 'before',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'bf'},
                            'bicep': {'args': '<type __builtin__.float>',
                                      'longname': 'bicep',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'bi'},
                            'deformerTools': {'args': '<type __builtin__.bool>',
                                              'longname': 'deformerTools',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'dt'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'exclusive': {'args': '<type __builtin__.str>',
                                          'longname': 'exclusive',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'ex'},
                            'frontOfChain': {'args': '<type __builtin__.bool>',
                                             'longname': 'frontOfChain',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'foc'},
                            'geometry': {'args': '<type __builtin__.str>',
                                         'longname': 'geometry',
                                         'modes': ['query',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'g'},
                            'geometryIndices': {'args': '<type __builtin__.bool>',
                                                'longname': 'geometryIndices',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'gi'},
                            'ignoreSelected': {'args': '<type __builtin__.bool>',
                                               'longname': 'ignoreSelected',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ignoreSelected'},
                            'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                        'longname': 'includeHiddenSelections',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'ihs'},
                            'joint': {'args': '<type __builtin__.str>',
                                      'longname': 'joint',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'j'},
                            'lengthIn': {'args': '<type __builtin__.float>',
                                         'longname': 'lengthIn',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'li'},
                            'lengthOut': {'args': '<type __builtin__.float>',
                                          'longname': 'lengthOut',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'lo'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'parallel': {'args': '<type __builtin__.bool>',
                                         'longname': 'parallel',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'par'},
                            'prune': {'args': '<type __builtin__.bool>',
                                      'longname': 'prune',
                                      'modes': ['edit'],
                                      'numArgs': 0,
                                      'shortname': 'pr'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'remove': {'args': '<type __builtin__.bool>',
                                       'longname': 'remove',
                                       'modes': ['edit', 'multiuse'],
                                       'numArgs': 0,
                                       'shortname': 'rm'},
                            'split': {'args': '<type __builtin__.bool>',
                                      'longname': 'split',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'sp'},
                            'transform': {'args': '<type __builtin__.str>',
                                          'longname': 'transform',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 't'},
                            'tricep': {'args': '<type __builtin__.float>',
                                       'longname': 'tricep',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'tr'},
                            'widthLeft': {'args': '<type __builtin__.float>',
                                          'longname': 'widthLeft',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'wl'},
                            'widthRight': {'args': '<type __builtin__.float>',
                                           'longname': 'widthRight',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'wr'}},
                  'removedFlags': {'is': 'ignoreSelected'},
                  'shortFlags': {'af': 'after',
                                 'ar': 'afterReference',
                                 'bf': 'before',
                                 'bi': 'bicep',
                                 'dt': 'deformerTools',
                                 'e': 'edit',
                                 'ex': 'exclusive',
                                 'foc': 'frontOfChain',
                                 'g': 'geometry',
                                 'gi': 'geometryIndices',
                                 'ignoreSelected': 'ignoreSelected',
                                 'ihs': 'includeHiddenSelections',
                                 'j': 'joint',
                                 'li': 'lengthIn',
                                 'lo': 'lengthOut',
                                 'n': 'name',
                                 'par': 'parallel',
                                 'pr': 'prune',
                                 'q': 'query',
                                 'rm': 'remove',
                                 'sp': 'split',
                                 't': 'transform',
                                 'tr': 'tricep',
                                 'wl': 'widthLeft',
                                 'wr': 'widthRight'},
                  'type': 'animation'},
  'boundary': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                     'longname': 'caching',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'cch'},
                         'constructionHistory': {'args': '<type __builtin__.bool>',
                                                 'longname': 'constructionHistory',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'endPoint': {'args': '<type __builtin__.bool>',
                                      'longname': 'endPoint',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'ep'},
                         'endPointTolerance': {'args': '<type __builtin__.float>',
                                               'longname': 'endPointTolerance',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ept'},
                         'frozen': {'args': '<type __builtin__.bool>',
                                    'longname': 'frozen',
                                    'numArgs': 1,
                                    'shortname': 'fzn'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'nodeState': {'args': '<type __builtin__.int>',
                                       'longname': 'nodeState',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'nds'},
                         'object': {'args': '<type __builtin__.bool>',
                                    'longname': 'object',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'o'},
                         'order': {'args': '<type __builtin__.bool>',
                                   'longname': 'order',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'order'},
                         'polygon': {'args': '<type __builtin__.int>',
                                     'longname': 'polygon',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'po'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'range': {'args': '<type __builtin__.bool>',
                                   'longname': 'range',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'rn'}},
               'removedFlags': {'or': 'order'},
               'shortFlags': {'cch': 'caching',
                              'ch': 'constructionHistory',
                              'e': 'edit',
                              'ep': 'endPoint',
                              'ept': 'endPointTolerance',
                              'fzn': 'frozen',
                              'n': 'name',
                              'nds': 'nodeState',
                              'o': 'object',
                              'order': 'order',
                              'po': 'polygon',
                              'q': 'query',
                              'rn': 'range'},
               'type': 'modeling'},
  'boxDollyCtx': {'flags': {'alternateContext': {'args': '<type __builtin__.bool>',
                                                 'longname': 'alternateContext',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ac'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'history': {'args': '<type __builtin__.bool>',
                                        'longname': 'history',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ch'},
                            'image1': {'args': '<type __builtin__.str>',
                                       'longname': 'image1',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i1'},
                            'image2': {'args': '<type __builtin__.str>',
                                       'longname': 'image2',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i2'},
                            'image3': {'args': '<type __builtin__.str>',
                                       'longname': 'image3',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i3'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'toolName': {'args': '<type __builtin__.str>',
                                         'longname': 'toolName',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'tn'}},
                  'shortFlags': {'ac': 'alternateContext',
                                 'ch': 'history',
                                 'e': 'edit',
                                 'ex': 'exists',
                                 'i1': 'image1',
                                 'i2': 'image2',
                                 'i3': 'image3',
                                 'n': 'name',
                                 'q': 'query',
                                 'tn': 'toolName'},
                  'type': 'context'},
  'boxZoomCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'history': {'args': '<type __builtin__.bool>',
                                       'longname': 'history',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ch'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'zoomScale': {'args': '<type __builtin__.float>',
                                         'longname': 'zoomScale',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'zs'}},
                 'shortFlags': {'ch': 'history',
                                'e': 'edit',
                                'ex': 'exists',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'n': 'name',
                                'q': 'query',
                                'zs': 'zoomScale'},
                 'type': 'context'},
  'bufferCurve': {'flags': {'animation': {'args': '<type __builtin__.str>',
                                          'longname': 'animation',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'an'},
                            'attribute': {'args': '<type __builtin__.str>',
                                          'longname': 'attribute',
                                          'modes': ['create', 'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'at'},
                            'controlPoints': {'args': '<type __builtin__.bool>',
                                              'longname': 'controlPoints',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'cp'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'float': {'args': 'floatrange',
                                      'longname': 'float',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'f'},
                            'hierarchy': {'args': '<type __builtin__.str>',
                                          'longname': 'hierarchy',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'hi'},
                            'includeUpperBound': {'args': '<type __builtin__.bool>',
                                                  'longname': 'includeUpperBound',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'iub'},
                            'index': {'args': '<type __builtin__.int>',
                                      'longname': 'index',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'index'},
                            'overwrite': {'args': '<type __builtin__.bool>',
                                          'longname': 'overwrite',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ov'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'shape': {'args': '<type __builtin__.bool>',
                                      'longname': 'shape',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                            'swap': {'args': '<type __builtin__.bool>',
                                     'longname': 'swap',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'sw'},
                            'time': {'args': 'timerange',
                                     'longname': 'time',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 't'},
                            'useReferencedCurve': {'args': '<type __builtin__.bool>',
                                                   'longname': 'useReferencedCurve',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 0,
                                                   'shortname': 'urc'}},
                  'removedFlags': {'in': 'index'},
                  'shortFlags': {'an': 'animation',
                                 'at': 'attribute',
                                 'cp': 'controlPoints',
                                 'ex': 'exists',
                                 'f': 'float',
                                 'hi': 'hierarchy',
                                 'index': 'index',
                                 'iub': 'includeUpperBound',
                                 'ov': 'overwrite',
                                 'q': 'query',
                                 's': 'shape',
                                 'sw': 'swap',
                                 't': 'time',
                                 'urc': 'useReferencedCurve'},
                  'type': 'animation'},
  'buildBookmarkMenu': {'flags': {'editor': {'args': '<type __builtin__.str>',
                                             'longname': 'editor',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ed'},
                                  'type': {'args': '<type __builtin__.str>',
                                           'longname': 'type',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'typ'}},
                        'shortFlags': {'ed': 'editor', 'typ': 'type'},
                        'type': 'animation'},
  'buildKeyframeMenu': {'flags': {}, 'shortFlags': {}, 'type': 'animation'},
  'buildSendToBackburnerDialog': {'type': 'runtime'},
  'button': {'callbackFlags': ['visibleChangeCommand',
                               'vcc',
                               'dragCallback',
                               'dgc',
                               'command',
                               'c',
                               'dropCallback',
                               'dpc'],
             'flags': {'actOnPress': {'args': '<type __builtin__.bool>',
                                      'longname': 'actOnPress',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'aop'},
                       'actionIsSubstitute': {'args': '<type __builtin__.bool>',
                                              'longname': 'actionIsSubstitute',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ais'},
                       'align': {'args': '<type __builtin__.str>',
                                 'longname': 'align',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'al'},
                       'annotation': {'args': '<type __builtin__.str>',
                                      'longname': 'annotation',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'ann'},
                       'backgroundColor': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'backgroundColor',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'bgc'},
                       'command': {'args': 'script',
                                   'longname': 'command',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'c'},
                       'defineTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'defineTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'dt'},
                       'docTag': {'args': '<type __builtin__.str>',
                                  'longname': 'docTag',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'dtg'},
                       'dragCallback': {'args': 'script',
                                        'longname': 'dragCallback',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dgc'},
                       'dropCallback': {'args': 'script',
                                        'longname': 'dropCallback',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dpc'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'enable': {'args': '<type __builtin__.bool>',
                                  'longname': 'enable',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'en'},
                       'enableBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'enableBackground',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ebg'},
                       'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                               'longname': 'enableKeyboardFocus',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ekf'},
                       'exists': {'args': '<type __builtin__.bool>',
                                  'longname': 'exists',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'ex'},
                       'fullPathName': {'args': '<type __builtin__.bool>',
                                        'longname': 'fullPathName',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'fpn'},
                       'height': {'args': '<type __builtin__.int>',
                                  'longname': 'height',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'h'},
                       'highlightColor': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'highlightColor',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 3,
                                          'shortname': 'hlc'},
                       'isObscured': {'args': '<type __builtin__.bool>',
                                      'longname': 'isObscured',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'io'},
                       'label': {'args': '<type __builtin__.str>',
                                 'longname': 'label',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'l'},
                       'manage': {'args': '<type __builtin__.bool>',
                                  'longname': 'manage',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'm'},
                       'noBackground': {'args': '<type __builtin__.bool>',
                                        'longname': 'noBackground',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'nbg'},
                       'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                              'longname': 'numberOfPopupMenus',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'npm'},
                       'parent': {'args': '<type __builtin__.str>',
                                  'longname': 'parent',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'p'},
                       'popupMenuArray': {'args': '<type __builtin__.bool>',
                                          'longname': 'popupMenuArray',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'pma'},
                       'preventOverride': {'args': '<type __builtin__.bool>',
                                           'longname': 'preventOverride',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'po'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'recomputeSize': {'args': '<type __builtin__.bool>',
                                         'longname': 'recomputeSize',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'rs'},
                       'statusBarMessage': {'args': '<type __builtin__.str>',
                                            'longname': 'statusBarMessage',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sbm'},
                       'useTemplate': {'args': '<type __builtin__.str>',
                                       'longname': 'useTemplate',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ut'},
                       'visible': {'args': '<type __builtin__.bool>',
                                   'longname': 'visible',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'vis'},
                       'visibleChangeCommand': {'args': 'script',
                                                'longname': 'visibleChangeCommand',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'vcc'},
                       'width': {'args': '<type __builtin__.int>',
                                 'longname': 'width',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'w'}},
             'shortFlags': {'ais': 'actionIsSubstitute',
                            'al': 'align',
                            'ann': 'annotation',
                            'aop': 'actOnPress',
                            'bgc': 'backgroundColor',
                            'c': 'command',
                            'dgc': 'dragCallback',
                            'dpc': 'dropCallback',
                            'dt': 'defineTemplate',
                            'dtg': 'docTag',
                            'e': 'edit',
                            'ebg': 'enableBackground',
                            'ekf': 'enableKeyboardFocus',
                            'en': 'enable',
                            'ex': 'exists',
                            'fpn': 'fullPathName',
                            'h': 'height',
                            'hlc': 'highlightColor',
                            'io': 'isObscured',
                            'l': 'label',
                            'm': 'manage',
                            'nbg': 'noBackground',
                            'npm': 'numberOfPopupMenus',
                            'p': 'parent',
                            'pma': 'popupMenuArray',
                            'po': 'preventOverride',
                            'q': 'query',
                            'rs': 'recomputeSize',
                            'sbm': 'statusBarMessage',
                            'ut': 'useTemplate',
                            'vcc': 'visibleChangeCommand',
                            'vis': 'visible',
                            'w': 'width'},
             'type': 'windows'},
  'buttonManip': {'flags': {'icon': {'args': '<type __builtin__.str>',
                                     'longname': 'icon',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'i'}},
                  'shortFlags': {'i': 'icon'},
                  'type': 'windows'},
  'cacheAppend': {'type': 'runtime'},
  'cacheAppendOpt': {'type': 'runtime'},
  'cacheFile': {'flags': {'appendFrame': {'args': '<type __builtin__.bool>',
                                          'longname': 'appendFrame',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'apf'},
                          'attachFile': {'args': '<type __builtin__.bool>',
                                         'longname': 'attachFile',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'af'},
                          'cacheFileNode': {'args': '<type __builtin__.str>',
                                            'longname': 'cacheFileNode',
                                            'modes': ['create',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'cfn'},
                          'cacheFormat': {'args': '<type __builtin__.str>',
                                          'longname': 'cacheFormat',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'cf'},
                          'cacheInfo': {'args': '<type __builtin__.str>',
                                        'longname': 'cacheInfo',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'ci'},
                          'cacheableAttrs': {'args': '<type __builtin__.str>',
                                             'longname': 'cacheableAttrs',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'cat'},
                          'cacheableNode': {'args': '<type __builtin__.str>',
                                            'longname': 'cacheableNode',
                                            'modes': ['create',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'cnd'},
                          'channelIndex': {'args': '<type __builtin__.bool>',
                                           'longname': 'channelIndex',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'chi'},
                          'channelName': {'args': '<type __builtin__.str>',
                                          'longname': 'channelName',
                                          'modes': ['create',
                                                    'query',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'cnm'},
                          'convertPc2': {'args': '<type __builtin__.bool>',
                                         'longname': 'convertPc2',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'pc2'},
                          'createCacheNode': {'args': '<type __builtin__.bool>',
                                              'longname': 'createCacheNode',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ccn'},
                          'creationChannelName': {'args': '<type __builtin__.str>',
                                                  'longname': 'creationChannelName',
                                                  'modes': ['create',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'cch'},
                          'dataSize': {'args': '<type __builtin__.bool>',
                                       'longname': 'dataSize',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'dsz'},
                          'deleteCachedFrame': {'args': '<type __builtin__.bool>',
                                                'longname': 'deleteCachedFrame',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'dcf'},
                          'descriptionFileName': {'args': '<type __builtin__.bool>',
                                                  'longname': 'descriptionFileName',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'dfn'},
                          'directory': {'args': '<type __builtin__.str>',
                                        'longname': 'directory',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'dir'},
                          'doubleToFloat': {'args': '<type __builtin__.bool>',
                                            'longname': 'doubleToFloat',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'dtf'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'endTime': {'args': 'time',
                                      'longname': 'endTime',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'et'},
                          'fileName': {'args': '<type __builtin__.str>',
                                       'longname': 'fileName',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'f'},
                          'format': {'args': '<type __builtin__.str>',
                                     'longname': 'format',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'fm'},
                          'geometry': {'args': '<type __builtin__.bool>',
                                       'longname': 'geometry',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'gm'},
                          'inAttr': {'args': '<type __builtin__.str>',
                                     'longname': 'inAttr',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'ia'},
                          'inTangent': {'args': '<type __builtin__.str>',
                                        'longname': 'inTangent',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'it'},
                          'interpEndTime': {'args': 'time',
                                            'longname': 'interpEndTime',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'iet'},
                          'interpStartTime': {'args': 'time',
                                              'longname': 'interpStartTime',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ist'},
                          'noBackup': {'args': '<type __builtin__.bool>',
                                       'longname': 'noBackup',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'nb'},
                          'outAttr': {'args': '<type __builtin__.str>',
                                      'longname': 'outAttr',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'oa'},
                          'outTangent': {'args': '<type __builtin__.str>',
                                         'longname': 'outTangent',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ot'},
                          'pc2File': {'args': '<type __builtin__.str>',
                                      'longname': 'pc2File',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'pcf'},
                          'pointCount': {'args': '<type __builtin__.bool>',
                                         'longname': 'pointCount',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'pc'},
                          'points': {'args': '<type __builtin__.str>',
                                     'longname': 'points',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'pts'},
                          'pointsAndNormals': {'args': '<type __builtin__.str>',
                                               'longname': 'pointsAndNormals',
                                               'modes': ['create',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'pan'},
                          'prefix': {'args': '<type __builtin__.bool>',
                                     'longname': 'prefix',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'p'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'refresh': {'args': '<type __builtin__.bool>',
                                      'longname': 'refresh',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'r'},
                          'replaceCachedFrame': {'args': '<type __builtin__.bool>',
                                                 'longname': 'replaceCachedFrame',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'rcf'},
                          'replaceWithoutSimulating': {'args': '<type __builtin__.bool>',
                                                       'longname': 'replaceWithoutSimulating',
                                                       'modes': ['edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'rws'},
                          'runupFrames': {'args': '<type __builtin__.int>',
                                          'longname': 'runupFrames',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'rf'},
                          'sampleMultiplier': {'args': '<type __builtin__.int>',
                                               'longname': 'sampleMultiplier',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'spm'},
                          'simulationRate': {'args': 'time',
                                             'longname': 'simulationRate',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'smr'},
                          'singleCache': {'args': '<type __builtin__.bool>',
                                          'longname': 'singleCache',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'sch'},
                          'startTime': {'args': 'time',
                                        'longname': 'startTime',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'st'},
                          'staticCache': {'args': '<type __builtin__.bool>',
                                          'longname': 'staticCache',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'sc'},
                          'worldSpace': {'args': '<type __builtin__.bool>',
                                         'longname': 'worldSpace',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ws'}},
                'shortFlags': {'af': 'attachFile',
                               'apf': 'appendFrame',
                               'cat': 'cacheableAttrs',
                               'cch': 'creationChannelName',
                               'ccn': 'createCacheNode',
                               'cf': 'cacheFormat',
                               'cfn': 'cacheFileNode',
                               'chi': 'channelIndex',
                               'ci': 'cacheInfo',
                               'cnd': 'cacheableNode',
                               'cnm': 'channelName',
                               'dcf': 'deleteCachedFrame',
                               'dfn': 'descriptionFileName',
                               'dir': 'directory',
                               'dsz': 'dataSize',
                               'dtf': 'doubleToFloat',
                               'e': 'edit',
                               'et': 'endTime',
                               'f': 'fileName',
                               'fm': 'format',
                               'gm': 'geometry',
                               'ia': 'inAttr',
                               'iet': 'interpEndTime',
                               'ist': 'interpStartTime',
                               'it': 'inTangent',
                               'nb': 'noBackup',
                               'oa': 'outAttr',
                               'ot': 'outTangent',
                               'p': 'prefix',
                               'pan': 'pointsAndNormals',
                               'pc': 'pointCount',
                               'pc2': 'convertPc2',
                               'pcf': 'pc2File',
                               'pts': 'points',
                               'q': 'query',
                               'r': 'refresh',
                               'rcf': 'replaceCachedFrame',
                               'rf': 'runupFrames',
                               'rws': 'replaceWithoutSimulating',
                               'sc': 'staticCache',
                               'sch': 'singleCache',
                               'smr': 'simulationRate',
                               'spm': 'sampleMultiplier',
                               'st': 'startTime',
                               'ws': 'worldSpace'},
                'type': 'system'},
  'cacheFileCombine': {'flags': {'cacheIndex': {'args': '<type __builtin__.bool>',
                                                'longname': 'cacheIndex',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'ci'},
                                 'channelName': {'args': '<type __builtin__.str>',
                                                 'longname': 'channelName',
                                                 'modes': ['edit',
                                                           'multiuse'],
                                                 'numArgs': 1,
                                                 'shortname': 'cnm'},
                                 'connectCache': {'args': '<type __builtin__.str>',
                                                  'longname': 'connectCache',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cc'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'keepWeights': {'args': '<type __builtin__.bool>',
                                                 'longname': 'keepWeights',
                                                 'modes': ['edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'kw'},
                                 'layerNode': {'args': '<type __builtin__.bool>',
                                               'longname': 'layerNode',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'ln'},
                                 'nextAvailable': {'args': '<type __builtin__.bool>',
                                                   'longname': 'nextAvailable',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'na'},
                                 'object': {'args': '<type __builtin__.str>',
                                            'longname': 'object',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'obj'},
                                 'objectIndex': {'args': '<type __builtin__.int>',
                                                 'longname': 'objectIndex',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'oi'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'}},
                       'shortFlags': {'cc': 'connectCache',
                                      'ci': 'cacheIndex',
                                      'cnm': 'channelName',
                                      'e': 'edit',
                                      'kw': 'keepWeights',
                                      'ln': 'layerNode',
                                      'na': 'nextAvailable',
                                      'obj': 'object',
                                      'oi': 'objectIndex',
                                      'q': 'query'},
                       'type': 'system'},
  'cacheFileMerge': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'endTime': {'args': 'time',
                                           'longname': 'endTime',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'et'},
                               'geometry': {'args': '<type __builtin__.bool>',
                                            'longname': 'geometry',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'g'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'startTime': {'args': 'time',
                                             'longname': 'startTime',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'st'}},
                     'shortFlags': {'e': 'edit',
                                    'et': 'endTime',
                                    'g': 'geometry',
                                    'q': 'query',
                                    'st': 'startTime'},
                     'type': 'system'},
  'cacheFileTrack': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'insertTrack': {'args': '<type __builtin__.int>',
                                               'longname': 'insertTrack',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'it'},
                               'lock': {'args': '<type __builtin__.bool>',
                                        'longname': 'lock',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'l'},
                               'mute': {'args': '<type __builtin__.bool>',
                                        'longname': 'mute',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'removeEmptyTracks': {'args': '<type __builtin__.bool>',
                                                     'longname': 'removeEmptyTracks',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'ret'},
                               'removeTrack': {'args': '<type __builtin__.int>',
                                               'longname': 'removeTrack',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'rt'},
                               'solo': {'args': '<type __builtin__.bool>',
                                        'longname': 'solo',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'so'},
                               'track': {'args': '<type __builtin__.int>',
                                         'longname': 'track',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 't'}},
                     'shortFlags': {'e': 'edit',
                                    'it': 'insertTrack',
                                    'l': 'lock',
                                    'm': 'mute',
                                    'q': 'query',
                                    'ret': 'removeEmptyTracks',
                                    'rt': 'removeTrack',
                                    'so': 'solo',
                                    't': 'track'},
                     'type': 'system'},
  'caddyManip': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'callbacks': {'callbackFlags': ['addCallback',
                                  'ac',
                                  'removeCallback',
                                  'rc'],
                'flags': {'addCallback': {'args': 'script',
                                          'longname': 'addCallback',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ac'},
                          'clearAllCallbacks': {'args': '<type __builtin__.bool>',
                                                'longname': 'clearAllCallbacks',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'cac'},
                          'clearCallbacks': {'args': '<type __builtin__.bool>',
                                             'longname': 'clearCallbacks',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'cc'},
                          'describeHooks': {'args': '<type __builtin__.bool>',
                                            'longname': 'describeHooks',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'dh'},
                          'dumpCallbacks': {'args': '<type __builtin__.bool>',
                                            'longname': 'dumpCallbacks',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'dc'},
                          'executeCallbacks': {'args': '<type __builtin__.bool>',
                                               'longname': 'executeCallbacks',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ec'},
                          'hook': {'args': '<type __builtin__.str>',
                                   'longname': 'hook',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'h'},
                          'listCallbacks': {'args': '<type __builtin__.bool>',
                                            'longname': 'listCallbacks',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'lc'},
                          'owner': {'args': '<type __builtin__.str>',
                                    'longname': 'owner',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'o'},
                          'removeCallback': {'args': 'script',
                                             'longname': 'removeCallback',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'rc'}},
                'shortFlags': {'ac': 'addCallback',
                               'cac': 'clearAllCallbacks',
                               'cc': 'clearCallbacks',
                               'dc': 'dumpCallbacks',
                               'dh': 'describeHooks',
                               'ec': 'executeCallbacks',
                               'h': 'hook',
                               'lc': 'listCallbacks',
                               'o': 'owner',
                               'rc': 'removeCallback'},
                'type': 'language'},
  'camera': {'callbackFlags': ['homeCommand', 'hc', 'journalCommand', 'jc'],
             'flags': {'aspectRatio': {'args': '<type __builtin__.float>',
                                       'longname': 'aspectRatio',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ar'},
                       'cameraScale': {'args': '<type __builtin__.float>',
                                       'longname': 'cameraScale',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cs'},
                       'centerOfInterest': {'args': '<type __builtin__.float>',
                                            'longname': 'centerOfInterest',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'coi'},
                       'clippingPlanes': {'args': '<type __builtin__.bool>',
                                          'longname': 'clippingPlanes',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cp'},
                       'depthOfField': {'args': '<type __builtin__.bool>',
                                        'longname': 'depthOfField',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dof'},
                       'displayFieldChart': {'args': '<type __builtin__.bool>',
                                             'longname': 'displayFieldChart',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dfc'},
                       'displayFilmGate': {'args': '<type __builtin__.bool>',
                                           'longname': 'displayFilmGate',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dfg'},
                       'displayFilmOrigin': {'args': '<type __builtin__.bool>',
                                             'longname': 'displayFilmOrigin',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dfo'},
                       'displayFilmPivot': {'args': '<type __builtin__.bool>',
                                            'longname': 'displayFilmPivot',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dfp'},
                       'displayGateMask': {'args': '<type __builtin__.bool>',
                                           'longname': 'displayGateMask',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dgm'},
                       'displayResolution': {'args': '<type __builtin__.bool>',
                                             'longname': 'displayResolution',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dr'},
                       'displaySafeAction': {'args': '<type __builtin__.bool>',
                                             'longname': 'displaySafeAction',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dsa'},
                       'displaySafeTitle': {'args': '<type __builtin__.bool>',
                                            'longname': 'displaySafeTitle',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dst'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'fStop': {'args': '<type __builtin__.float>',
                                 'longname': 'fStop',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'fs'},
                       'farClipPlane': {'args': '<type __builtin__.float>',
                                        'longname': 'farClipPlane',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'fcp'},
                       'farFocusDistance': {'args': '<type __builtin__.float>',
                                            'longname': 'farFocusDistance',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ffd'},
                       'filmFit': {'args': '<type __builtin__.str>',
                                   'longname': 'filmFit',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'ff'},
                       'filmFitOffset': {'args': '<type __builtin__.float>',
                                         'longname': 'filmFitOffset',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ffo'},
                       'filmRollOrder': {'args': '<type __builtin__.str>',
                                         'longname': 'filmRollOrder',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'fro'},
                       'filmRollValue': {'args': '<type __builtin__.float>',
                                         'longname': 'filmRollValue',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'frv'},
                       'filmTranslateH': {'args': '<type __builtin__.float>',
                                          'longname': 'filmTranslateH',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'fth'},
                       'filmTranslateV': {'args': '<type __builtin__.float>',
                                          'longname': 'filmTranslateV',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ftv'},
                       'focalLength': {'args': '<type __builtin__.float>',
                                       'longname': 'focalLength',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'fl'},
                       'focusDistance': {'args': '<type __builtin__.float>',
                                         'longname': 'focusDistance',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'fd'},
                       'homeCommand': {'args': '<type __builtin__.str>',
                                       'longname': 'homeCommand',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'hc'},
                       'horizontalFieldOfView': {'args': '<type __builtin__.float>',
                                                 'longname': 'horizontalFieldOfView',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'hfv'},
                       'horizontalFilmAperture': {'args': '<type __builtin__.float>',
                                                  'longname': 'horizontalFilmAperture',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'hfa'},
                       'horizontalFilmOffset': {'args': '<type __builtin__.float>',
                                                'longname': 'horizontalFilmOffset',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'hfo'},
                       'horizontalPan': {'args': '<type __builtin__.float>',
                                         'longname': 'horizontalPan',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'hpn'},
                       'horizontalRollPivot': {'args': '<type __builtin__.float>',
                                               'longname': 'horizontalRollPivot',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'hrp'},
                       'horizontalShake': {'args': '<type __builtin__.float>',
                                           'longname': 'horizontalShake',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'hs'},
                       'journalCommand': {'args': '<type __builtin__.bool>',
                                          'longname': 'journalCommand',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'jc'},
                       'lensSqueezeRatio': {'args': '<type __builtin__.float>',
                                            'longname': 'lensSqueezeRatio',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'lsr'},
                       'lockTransform': {'args': '<type __builtin__.bool>',
                                         'longname': 'lockTransform',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'lt'},
                       'motionBlur': {'args': '<type __builtin__.bool>',
                                      'longname': 'motionBlur',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'mb'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'nearClipPlane': {'args': '<type __builtin__.float>',
                                         'longname': 'nearClipPlane',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ncp'},
                       'nearFocusDistance': {'args': '<type __builtin__.float>',
                                             'longname': 'nearFocusDistance',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nfd'},
                       'orthographic': {'args': '<type __builtin__.bool>',
                                        'longname': 'orthographic',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                       'orthographicWidth': {'args': '<type __builtin__.float>',
                                             'longname': 'orthographicWidth',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ow'},
                       'overscan': {'args': '<type __builtin__.float>',
                                    'longname': 'overscan',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'ovr'},
                       'panZoomEnabled': {'args': '<type __builtin__.bool>',
                                          'longname': 'panZoomEnabled',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pze'},
                       'position': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'position',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 3,
                                    'shortname': 'p'},
                       'postScale': {'args': '<type __builtin__.float>',
                                     'longname': 'postScale',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'pts'},
                       'preScale': {'args': '<type __builtin__.float>',
                                    'longname': 'preScale',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'prs'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'renderPanZoom': {'args': '<type __builtin__.bool>',
                                         'longname': 'renderPanZoom',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'rpz'},
                       'rotation': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'rotation',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 3,
                                    'shortname': 'rot'},
                       'shakeEnabled': {'args': '<type __builtin__.bool>',
                                        'longname': 'shakeEnabled',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'se'},
                       'shakeOverscan': {'args': '<type __builtin__.float>',
                                         'longname': 'shakeOverscan',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'so'},
                       'shakeOverscanEnabled': {'args': '<type __builtin__.bool>',
                                                'longname': 'shakeOverscanEnabled',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'soe'},
                       'shutterAngle': {'args': '<type __builtin__.float>',
                                        'longname': 'shutterAngle',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'sa'},
                       'startupCamera': {'args': '<type __builtin__.bool>',
                                         'longname': 'startupCamera',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'sc'},
                       'stereoHorizontalImageTranslate': {'args': '<type __builtin__.float>',
                                                          'longname': 'stereoHorizontalImageTranslate',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'hit'},
                       'stereoHorizontalImageTranslateEnabled': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'stereoHorizontalImageTranslateEnabled',
                                                                 'modes': ['create',
                                                                           'query',
                                                                           'edit'],
                                                                 'numArgs': 1,
                                                                 'shortname': 'she'},
                       'verticalFieldOfView': {'args': '<type __builtin__.float>',
                                               'longname': 'verticalFieldOfView',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'vfv'},
                       'verticalFilmAperture': {'args': '<type __builtin__.float>',
                                                'longname': 'verticalFilmAperture',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'vfa'},
                       'verticalFilmOffset': {'args': '<type __builtin__.float>',
                                              'longname': 'verticalFilmOffset',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'vfo'},
                       'verticalLock': {'args': '<type __builtin__.bool>',
                                        'longname': 'verticalLock',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vl'},
                       'verticalPan': {'args': '<type __builtin__.float>',
                                       'longname': 'verticalPan',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vpn'},
                       'verticalRollPivot': {'args': '<type __builtin__.float>',
                                             'longname': 'verticalRollPivot',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'vrp'},
                       'verticalShake': {'args': '<type __builtin__.float>',
                                         'longname': 'verticalShake',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vs'},
                       'worldCenterOfInterest': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'worldCenterOfInterest',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'wci'},
                       'worldUp': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'worldUp',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 3,
                                   'shortname': 'wup'},
                       'zoom': {'args': '<type __builtin__.float>',
                                'longname': 'zoom',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'zom'}},
             'shortFlags': {'ar': 'aspectRatio',
                            'coi': 'centerOfInterest',
                            'cp': 'clippingPlanes',
                            'cs': 'cameraScale',
                            'dfc': 'displayFieldChart',
                            'dfg': 'displayFilmGate',
                            'dfo': 'displayFilmOrigin',
                            'dfp': 'displayFilmPivot',
                            'dgm': 'displayGateMask',
                            'dof': 'depthOfField',
                            'dr': 'displayResolution',
                            'dsa': 'displaySafeAction',
                            'dst': 'displaySafeTitle',
                            'e': 'edit',
                            'fcp': 'farClipPlane',
                            'fd': 'focusDistance',
                            'ff': 'filmFit',
                            'ffd': 'farFocusDistance',
                            'ffo': 'filmFitOffset',
                            'fl': 'focalLength',
                            'fro': 'filmRollOrder',
                            'frv': 'filmRollValue',
                            'fs': 'fStop',
                            'fth': 'filmTranslateH',
                            'ftv': 'filmTranslateV',
                            'hc': 'homeCommand',
                            'hfa': 'horizontalFilmAperture',
                            'hfo': 'horizontalFilmOffset',
                            'hfv': 'horizontalFieldOfView',
                            'hit': 'stereoHorizontalImageTranslate',
                            'hpn': 'horizontalPan',
                            'hrp': 'horizontalRollPivot',
                            'hs': 'horizontalShake',
                            'jc': 'journalCommand',
                            'lsr': 'lensSqueezeRatio',
                            'lt': 'lockTransform',
                            'mb': 'motionBlur',
                            'n': 'name',
                            'ncp': 'nearClipPlane',
                            'nfd': 'nearFocusDistance',
                            'o': 'orthographic',
                            'ovr': 'overscan',
                            'ow': 'orthographicWidth',
                            'p': 'position',
                            'prs': 'preScale',
                            'pts': 'postScale',
                            'pze': 'panZoomEnabled',
                            'q': 'query',
                            'rot': 'rotation',
                            'rpz': 'renderPanZoom',
                            'sa': 'shutterAngle',
                            'sc': 'startupCamera',
                            'se': 'shakeEnabled',
                            'she': 'stereoHorizontalImageTranslateEnabled',
                            'so': 'shakeOverscan',
                            'soe': 'shakeOverscanEnabled',
                            'vfa': 'verticalFilmAperture',
                            'vfo': 'verticalFilmOffset',
                            'vfv': 'verticalFieldOfView',
                            'vl': 'verticalLock',
                            'vpn': 'verticalPan',
                            'vrp': 'verticalRollPivot',
                            'vs': 'verticalShake',
                            'wci': 'worldCenterOfInterest',
                            'wup': 'worldUp',
                            'zom': 'zoom'},
             'type': 'rendering'},
  'cameraSet': {'flags': {'active': {'args': '<type __builtin__.bool>',
                                     'longname': 'active',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'a'},
                          'appendTo': {'args': '<type __builtin__.bool>',
                                       'longname': 'appendTo',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'atl'},
                          'camera': {'args': '<type __builtin__.str>',
                                     'longname': 'camera',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'cam'},
                          'clearDepth': {'args': '<type __builtin__.bool>',
                                         'longname': 'clearDepth',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cd'},
                          'deleteAll': {'args': '<type __builtin__.bool>',
                                        'longname': 'deleteAll',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'da'},
                          'deleteLayer': {'args': '<type __builtin__.bool>',
                                          'longname': 'deleteLayer',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'd'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'insertAt': {'args': '<type __builtin__.bool>',
                                       'longname': 'insertAt',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'ins'},
                          'layer': {'args': '<type __builtin__.int>',
                                    'longname': 'layer',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'l'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'numLayers': {'args': '<type __builtin__.bool>',
                                        'longname': 'numLayers',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'nl'},
                          'objectSet': {'args': '<type __builtin__.str>',
                                        'longname': 'objectSet',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'os'},
                          'order': {'args': '<type __builtin__.int>',
                                    'longname': 'order',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'o'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'}},
                'shortFlags': {'a': 'active',
                               'atl': 'appendTo',
                               'cam': 'camera',
                               'cd': 'clearDepth',
                               'd': 'deleteLayer',
                               'da': 'deleteAll',
                               'e': 'edit',
                               'ins': 'insertAt',
                               'l': 'layer',
                               'n': 'name',
                               'nl': 'numLayers',
                               'o': 'order',
                               'os': 'objectSet',
                               'q': 'query'},
                'type': 'rendering'},
  'cameraView': {'flags': {'addBookmark': {'args': '<type __builtin__.bool>',
                                           'longname': 'addBookmark',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'ab'},
                           'animate': {'args': '<type __builtin__.bool>',
                                       'longname': 'animate',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'an'},
                           'bookmarkType': {'args': '<type __builtin__.int>',
                                            'longname': 'bookmarkType',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'typ'},
                           'camera': {'args': 'PyNode',
                                      'longname': 'camera',
                                      'modes': ['edit'],
                                      'numArgs': 1,
                                      'shortname': 'c'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'removeBookmark': {'args': '<type __builtin__.bool>',
                                              'longname': 'removeBookmark',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'rb'},
                           'setCamera': {'args': '<type __builtin__.bool>',
                                         'longname': 'setCamera',
                                         'modes': ['edit'],
                                         'numArgs': 0,
                                         'shortname': 'sc'},
                           'setView': {'args': '<type __builtin__.bool>',
                                       'longname': 'setView',
                                       'modes': ['edit'],
                                       'numArgs': 0,
                                       'shortname': 'sv'}},
                 'shortFlags': {'ab': 'addBookmark',
                                'an': 'animate',
                                'c': 'camera',
                                'e': 'edit',
                                'n': 'name',
                                'rb': 'removeBookmark',
                                'sc': 'setCamera',
                                'sv': 'setView',
                                'typ': 'bookmarkType'},
                 'type': 'rendering'},
  'canCreateCaddyManip': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'canCreateManip': {'flags': {}, 'shortFlags': {}, 'type': 'modeling'},
  'canvas': {'callbackFlags': ['pressCommand',
                               'pc',
                               'visibleChangeCommand',
                               'vcc',
                               'dragCallback',
                               'dgc',
                               'dropCallback',
                               'dpc'],
             'flags': {'annotation': {'args': '<type __builtin__.str>',
                                      'longname': 'annotation',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'ann'},
                       'backgroundColor': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'backgroundColor',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'bgc'},
                       'defineTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'defineTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'dt'},
                       'docTag': {'args': '<type __builtin__.str>',
                                  'longname': 'docTag',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'dtg'},
                       'dragCallback': {'args': 'script',
                                        'longname': 'dragCallback',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dgc'},
                       'dropCallback': {'args': 'script',
                                        'longname': 'dropCallback',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dpc'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'enable': {'args': '<type __builtin__.bool>',
                                  'longname': 'enable',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'en'},
                       'enableBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'enableBackground',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ebg'},
                       'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                               'longname': 'enableKeyboardFocus',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ekf'},
                       'exists': {'args': '<type __builtin__.bool>',
                                  'longname': 'exists',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'ex'},
                       'fullPathName': {'args': '<type __builtin__.bool>',
                                        'longname': 'fullPathName',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'fpn'},
                       'height': {'args': '<type __builtin__.int>',
                                  'longname': 'height',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'h'},
                       'highlightColor': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'highlightColor',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 3,
                                          'shortname': 'hlc'},
                       'hsvValue': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'hsvValue',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 3,
                                    'shortname': 'hsv'},
                       'isObscured': {'args': '<type __builtin__.bool>',
                                      'longname': 'isObscured',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'io'},
                       'manage': {'args': '<type __builtin__.bool>',
                                  'longname': 'manage',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'm'},
                       'noBackground': {'args': '<type __builtin__.bool>',
                                        'longname': 'noBackground',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'nbg'},
                       'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                              'longname': 'numberOfPopupMenus',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'npm'},
                       'parent': {'args': '<type __builtin__.str>',
                                  'longname': 'parent',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'p'},
                       'popupMenuArray': {'args': '<type __builtin__.bool>',
                                          'longname': 'popupMenuArray',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'pma'},
                       'pressCommand': {'args': 'script',
                                        'longname': 'pressCommand',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'pc'},
                       'preventOverride': {'args': '<type __builtin__.bool>',
                                           'longname': 'preventOverride',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'po'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'rgbValue': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'rgbValue',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 3,
                                    'shortname': 'rgb'},
                       'statusBarMessage': {'args': '<type __builtin__.str>',
                                            'longname': 'statusBarMessage',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sbm'},
                       'useTemplate': {'args': '<type __builtin__.str>',
                                       'longname': 'useTemplate',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ut'},
                       'visible': {'args': '<type __builtin__.bool>',
                                   'longname': 'visible',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'vis'},
                       'visibleChangeCommand': {'args': 'script',
                                                'longname': 'visibleChangeCommand',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'vcc'},
                       'width': {'args': '<type __builtin__.int>',
                                 'longname': 'width',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'w'}},
             'shortFlags': {'ann': 'annotation',
                            'bgc': 'backgroundColor',
                            'dgc': 'dragCallback',
                            'dpc': 'dropCallback',
                            'dt': 'defineTemplate',
                            'dtg': 'docTag',
                            'e': 'edit',
                            'ebg': 'enableBackground',
                            'ekf': 'enableKeyboardFocus',
                            'en': 'enable',
                            'ex': 'exists',
                            'fpn': 'fullPathName',
                            'h': 'height',
                            'hlc': 'highlightColor',
                            'hsv': 'hsvValue',
                            'io': 'isObscured',
                            'm': 'manage',
                            'nbg': 'noBackground',
                            'npm': 'numberOfPopupMenus',
                            'p': 'parent',
                            'pc': 'pressCommand',
                            'pma': 'popupMenuArray',
                            'po': 'preventOverride',
                            'q': 'query',
                            'rgb': 'rgbValue',
                            'sbm': 'statusBarMessage',
                            'ut': 'useTemplate',
                            'vcc': 'visibleChangeCommand',
                            'vis': 'visible',
                            'w': 'width'},
             'type': 'windows'},
  'changeSubdivComponentDisplayLevel': {'flags': {'level': {'args': '<type __builtin__.int>',
                                                            'longname': 'level',
                                                            'modes': ['create',
                                                                      'query'],
                                                            'numArgs': 1,
                                                            'shortname': 'l'},
                                                  'query': {'args': '<type __builtin__.bool>',
                                                            'longname': 'query',
                                                            'numArgs': 0,
                                                            'shortname': 'q'},
                                                  'relative': {'args': '<type __builtin__.bool>',
                                                               'longname': 'relative',
                                                               'modes': ['create',
                                                                         'query'],
                                                               'numArgs': 1,
                                                               'shortname': 'r'}},
                                        'shortFlags': {'l': 'level',
                                                       'q': 'query',
                                                       'r': 'relative'},
                                        'type': 'modeling'},
  'changeSubdivRegion': {'flags': {'action': {'args': '<type __builtin__.int>',
                                              'longname': 'action',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'a'},
                                   'level': {'args': '<type __builtin__.int>',
                                             'longname': 'level',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'l'}},
                         'shortFlags': {'a': 'action', 'l': 'level'},
                         'type': 'modeling'},
  'channelBox': {'callbackFlags': ['visibleChangeCommand',
                                   'vcc',
                                   'dragCallback',
                                   'dgc',
                                   'dropCallback',
                                   'dpc'],
                 'flags': {'annotation': {'args': '<type __builtin__.str>',
                                          'longname': 'annotation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ann'},
                           'attrBgColor': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'attrBgColor',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'bc'},
                           'attrColor': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'attrColor',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 3,
                                         'shortname': 'ac'},
                           'attrFilter': {'args': '<type __builtin__.str>',
                                          'longname': 'attrFilter',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'af'},
                           'attrRegex': {'args': '<type __builtin__.str>',
                                         'longname': 'attrRegex',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ar'},
                           'attributeEditorMode': {'args': '<type __builtin__.bool>',
                                                   'longname': 'attributeEditorMode',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'aem'},
                           'backgroundColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'backgroundColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'bgc'},
                           'containerAtTop': {'args': '<type __builtin__.bool>',
                                              'longname': 'containerAtTop',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cat'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'dragCallback': {'args': 'script',
                                            'longname': 'dragCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dgc'},
                           'dropCallback': {'args': 'script',
                                            'longname': 'dropCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dpc'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'enable': {'args': '<type __builtin__.bool>',
                                      'longname': 'enable',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'en'},
                           'enableBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableBackground',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ebg'},
                           'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableKeyboardFocus',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ekf'},
                           'enableLabelSelection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableLabelSelection',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'els'},
                           'execute': {'args': ['<type __builtin__.str>',
                                                '<type __builtin__.bool>'],
                                       'longname': 'execute',
                                       'modes': ['edit'],
                                       'numArgs': 2,
                                       'shortname': 'exe'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'fieldWidth': {'args': '<type __builtin__.int>',
                                          'longname': 'fieldWidth',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'fw'},
                           'fixedAttrList': {'args': '<type __builtin__.str>',
                                             'longname': 'fixedAttrList',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'fal'},
                           'fullPathName': {'args': '<type __builtin__.bool>',
                                            'longname': 'fullPathName',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fpn'},
                           'height': {'args': '<type __builtin__.int>',
                                      'longname': 'height',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                           'highlightColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'highlightColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'hlc'},
                           'historyObjectList': {'args': '<type __builtin__.bool>',
                                                 'longname': 'historyObjectList',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'hol'},
                           'hyperbolic': {'args': '<type __builtin__.bool>',
                                          'longname': 'hyperbolic',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'hyp'},
                           'inputs': {'args': '<type __builtin__.bool>',
                                      'longname': 'inputs',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'inputs'},
                           'isObscured': {'args': '<type __builtin__.bool>',
                                          'longname': 'isObscured',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'io'},
                           'labelWidth': {'args': '<type __builtin__.int>',
                                          'longname': 'labelWidth',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'lw'},
                           'longNames': {'args': '<type __builtin__.bool>',
                                         'longname': 'longNames',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ln'},
                           'mainListConnection': {'args': '<type __builtin__.str>',
                                                  'longname': 'mainListConnection',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mlc'},
                           'mainObjectList': {'args': '<type __builtin__.bool>',
                                              'longname': 'mainObjectList',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'mol'},
                           'manage': {'args': '<type __builtin__.bool>',
                                      'longname': 'manage',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                           'maxHeight': {'args': '<type __builtin__.int>',
                                         'longname': 'maxHeight',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'mh'},
                           'maxWidth': {'args': '<type __builtin__.int>',
                                        'longname': 'maxWidth',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'mw'},
                           'niceNames': {'args': '<type __builtin__.bool>',
                                         'longname': 'niceNames',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nn'},
                           'noBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'noBackground',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nbg'},
                           'nodeRegex': {'args': '<type __builtin__.str>',
                                         'longname': 'nodeRegex',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nr'},
                           'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'numberOfPopupMenus',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'npm'},
                           'outputObjectList': {'args': '<type __builtin__.bool>',
                                                'longname': 'outputObjectList',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'ool'},
                           'outputs': {'args': '<type __builtin__.bool>',
                                       'longname': 'outputs',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'out'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuArray': {'args': '<type __builtin__.bool>',
                                              'longname': 'popupMenuArray',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'pma'},
                           'precision': {'args': '<type __builtin__.int>',
                                         'longname': 'precision',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'pre'},
                           'preventOverride': {'args': '<type __builtin__.bool>',
                                               'longname': 'preventOverride',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'po'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'select': {'args': '<type __builtin__.str>',
                                      'longname': 'select',
                                      'modes': ['edit', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                           'selectedHistoryAttributes': {'args': '<type __builtin__.bool>',
                                                         'longname': 'selectedHistoryAttributes',
                                                         'modes': ['query'],
                                                         'numArgs': 0,
                                                         'shortname': 'sha'},
                           'selectedMainAttributes': {'args': '<type __builtin__.bool>',
                                                      'longname': 'selectedMainAttributes',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'sma'},
                           'selectedOutputAttributes': {'args': '<type __builtin__.bool>',
                                                        'longname': 'selectedOutputAttributes',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'soa'},
                           'selectedShapeAttributes': {'args': '<type __builtin__.bool>',
                                                       'longname': 'selectedShapeAttributes',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'ssa'},
                           'shapeObjectList': {'args': '<type __builtin__.bool>',
                                               'longname': 'shapeObjectList',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'sol'},
                           'shapes': {'args': '<type __builtin__.bool>',
                                      'longname': 'shapes',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'shp'},
                           'showNamespace': {'args': '<type __builtin__.bool>',
                                             'longname': 'showNamespace',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sn'},
                           'showTransforms': {'args': '<type __builtin__.bool>',
                                              'longname': 'showTransforms',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'st'},
                           'speed': {'args': '<type __builtin__.float>',
                                     'longname': 'speed',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'spd'},
                           'statusBarMessage': {'args': '<type __builtin__.str>',
                                                'longname': 'statusBarMessage',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sbm'},
                           'takeFocus': {'args': '<type __builtin__.bool>',
                                         'longname': 'takeFocus',
                                         'modes': ['edit'],
                                         'numArgs': 0,
                                         'shortname': 'tf'},
                           'update': {'args': '<type __builtin__.bool>',
                                      'longname': 'update',
                                      'modes': ['edit'],
                                      'numArgs': 0,
                                      'shortname': 'u'},
                           'useManips': {'args': '<type __builtin__.str>',
                                         'longname': 'useManips',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'mnp'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'},
                           'visible': {'args': '<type __builtin__.bool>',
                                       'longname': 'visible',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vis'},
                           'visibleChangeCommand': {'args': 'script',
                                                    'longname': 'visibleChangeCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'vcc'},
                           'width': {'args': '<type __builtin__.int>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'}},
                 'removedFlags': {'in': 'inputs'},
                 'shortFlags': {'ac': 'attrColor',
                                'aem': 'attributeEditorMode',
                                'af': 'attrFilter',
                                'ann': 'annotation',
                                'ar': 'attrRegex',
                                'bc': 'attrBgColor',
                                'bgc': 'backgroundColor',
                                'cat': 'containerAtTop',
                                'dgc': 'dragCallback',
                                'dpc': 'dropCallback',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'ebg': 'enableBackground',
                                'ekf': 'enableKeyboardFocus',
                                'els': 'enableLabelSelection',
                                'en': 'enable',
                                'ex': 'exists',
                                'exe': 'execute',
                                'fal': 'fixedAttrList',
                                'fpn': 'fullPathName',
                                'fw': 'fieldWidth',
                                'h': 'height',
                                'hlc': 'highlightColor',
                                'hol': 'historyObjectList',
                                'hyp': 'hyperbolic',
                                'inputs': 'inputs',
                                'io': 'isObscured',
                                'ln': 'longNames',
                                'lw': 'labelWidth',
                                'm': 'manage',
                                'mh': 'maxHeight',
                                'mlc': 'mainListConnection',
                                'mnp': 'useManips',
                                'mol': 'mainObjectList',
                                'mw': 'maxWidth',
                                'nbg': 'noBackground',
                                'nn': 'niceNames',
                                'npm': 'numberOfPopupMenus',
                                'nr': 'nodeRegex',
                                'ool': 'outputObjectList',
                                'out': 'outputs',
                                'p': 'parent',
                                'pma': 'popupMenuArray',
                                'po': 'preventOverride',
                                'pre': 'precision',
                                'q': 'query',
                                's': 'select',
                                'sbm': 'statusBarMessage',
                                'sha': 'selectedHistoryAttributes',
                                'shp': 'shapes',
                                'sma': 'selectedMainAttributes',
                                'sn': 'showNamespace',
                                'soa': 'selectedOutputAttributes',
                                'sol': 'shapeObjectList',
                                'spd': 'speed',
                                'ssa': 'selectedShapeAttributes',
                                'st': 'showTransforms',
                                'tf': 'takeFocus',
                                'u': 'update',
                                'ut': 'useTemplate',
                                'vcc': 'visibleChangeCommand',
                                'vis': 'visible',
                                'w': 'width'},
                 'type': 'windows'},
  'character': {'flags': {'addElement': {'args': 'PyNode',
                                         'longname': 'addElement',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 'add'},
                          'addOffsetObject': {'args': '<type __builtin__.str>',
                                              'longname': 'addOffsetObject',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'aoo'},
                          'characterPlug': {'args': '<type __builtin__.bool>',
                                            'longname': 'characterPlug',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'cp'},
                          'clear': {'args': 'PyNode',
                                    'longname': 'clear',
                                    'modes': ['edit'],
                                    'numArgs': 1,
                                    'shortname': 'cl'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'empty': {'args': '<type __builtin__.bool>',
                                    'longname': 'empty',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'em'},
                          'excludeDynamic': {'args': '<type __builtin__.bool>',
                                             'longname': 'excludeDynamic',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ed'},
                          'excludeRotate': {'args': '<type __builtin__.bool>',
                                            'longname': 'excludeRotate',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'er'},
                          'excludeScale': {'args': '<type __builtin__.bool>',
                                           'longname': 'excludeScale',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'es'},
                          'excludeTranslate': {'args': '<type __builtin__.bool>',
                                               'longname': 'excludeTranslate',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'et'},
                          'excludeVisibility': {'args': '<type __builtin__.bool>',
                                                'longname': 'excludeVisibility',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ev'},
                          'flatten': {'args': 'PyNode',
                                      'longname': 'flatten',
                                      'modes': ['edit'],
                                      'numArgs': 1,
                                      'shortname': 'fl'},
                          'forceElement': {'args': 'PyNode',
                                           'longname': 'forceElement',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'fe'},
                          'include': {'args': 'PyNode',
                                      'longname': 'include',
                                      'modes': ['edit'],
                                      'numArgs': 1,
                                      'shortname': 'include'},
                          'intersection': {'args': 'PyNode',
                                           'longname': 'intersection',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 'int'},
                          'isIntersecting': {'args': 'PyNode',
                                             'longname': 'isIntersecting',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'ii'},
                          'isMember': {'args': 'PyNode',
                                       'longname': 'isMember',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'im'},
                          'library': {'args': '<type __builtin__.bool>',
                                      'longname': 'library',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'lib'},
                          'memberIndex': {'args': '<type __builtin__.int>',
                                          'longname': 'memberIndex',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'mi'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'noWarnings': {'args': '<type __builtin__.bool>',
                                         'longname': 'noWarnings',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'nw'},
                          'nodesOnly': {'args': '<type __builtin__.bool>',
                                        'longname': 'nodesOnly',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'no'},
                          'offsetNode': {'args': '<type __builtin__.bool>',
                                         'longname': 'offsetNode',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'ofs'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'remove': {'args': 'PyNode',
                                     'longname': 'remove',
                                     'modes': ['edit'],
                                     'numArgs': 1,
                                     'shortname': 'rm'},
                          'removeOffsetObject': {'args': '<type __builtin__.str>',
                                                 'longname': 'removeOffsetObject',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'roo'},
                          'root': {'args': '<type __builtin__.str>',
                                   'longname': 'root',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'rt'},
                          'scheduler': {'args': '<type __builtin__.bool>',
                                        'longname': 'scheduler',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'sc'},
                          'split': {'args': 'PyNode',
                                    'longname': 'split',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'sp'},
                          'subtract': {'args': 'PyNode',
                                       'longname': 'subtract',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'sub'},
                          'text': {'args': '<type __builtin__.str>',
                                   'longname': 'text',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 't'},
                          'union': {'args': 'PyNode',
                                    'longname': 'union',
                                    'modes': ['query'],
                                    'numArgs': 1,
                                    'shortname': 'un'},
                          'userAlias': {'args': 'PyNode',
                                        'longname': 'userAlias',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'ua'}},
                'removedFlags': {'in': 'include'},
                'shortFlags': {'add': 'addElement',
                               'aoo': 'addOffsetObject',
                               'cl': 'clear',
                               'cp': 'characterPlug',
                               'e': 'edit',
                               'ed': 'excludeDynamic',
                               'em': 'empty',
                               'er': 'excludeRotate',
                               'es': 'excludeScale',
                               'et': 'excludeTranslate',
                               'ev': 'excludeVisibility',
                               'fe': 'forceElement',
                               'fl': 'flatten',
                               'ii': 'isIntersecting',
                               'im': 'isMember',
                               'include': 'include',
                               'int': 'intersection',
                               'lib': 'library',
                               'mi': 'memberIndex',
                               'n': 'name',
                               'no': 'nodesOnly',
                               'nw': 'noWarnings',
                               'ofs': 'offsetNode',
                               'q': 'query',
                               'rm': 'remove',
                               'roo': 'removeOffsetObject',
                               'rt': 'root',
                               'sc': 'scheduler',
                               'sp': 'split',
                               'sub': 'subtract',
                               't': 'text',
                               'ua': 'userAlias',
                               'un': 'union'},
                'type': 'animation'},
  'characterMap': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'mapAttr': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'mapAttr',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 2,
                                         'shortname': 'ma'},
                             'mapMethod': {'args': '<type __builtin__.str>',
                                           'longname': 'mapMethod',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'mm'},
                             'mapNode': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'mapNode',
                                         'modes': ['create', 'query'],
                                         'numArgs': 2,
                                         'shortname': 'mn'},
                             'mapping': {'args': '<type __builtin__.str>',
                                         'longname': 'mapping',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                             'proposedMapping': {'args': '<type __builtin__.bool>',
                                                 'longname': 'proposedMapping',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'pm'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'unmapAttr': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'unmapAttr',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 2,
                                           'shortname': 'ua'},
                             'unmapNode': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'unmapNode',
                                           'modes': ['create'],
                                           'numArgs': 2,
                                           'shortname': 'umn'}},
                   'shortFlags': {'e': 'edit',
                                  'm': 'mapping',
                                  'ma': 'mapAttr',
                                  'mm': 'mapMethod',
                                  'mn': 'mapNode',
                                  'pm': 'proposedMapping',
                                  'q': 'query',
                                  'ua': 'unmapAttr',
                                  'umn': 'unmapNode'},
                   'type': 'animation'},
  'characterize': {'flags': {'activatePivot': {'args': '<type __builtin__.bool>',
                                               'longname': 'activatePivot',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'apv'},
                             'addAuxEffector': {'args': '<type __builtin__.bool>',
                                                'longname': 'addAuxEffector',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'aae'},
                             'addFloorContactPlane': {'args': '<type __builtin__.bool>',
                                                      'longname': 'addFloorContactPlane',
                                                      'modes': ['edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'afp'},
                             'addMissingEffectors': {'args': '<type __builtin__.bool>',
                                                     'longname': 'addMissingEffectors',
                                                     'modes': ['edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'ame'},
                             'attributeFromHIKProperty': {'args': '<type __builtin__.str>',
                                                          'longname': 'attributeFromHIKProperty',
                                                          'modes': ['query'],
                                                          'numArgs': 1,
                                                          'shortname': 'ahk'},
                             'attributeFromHIKPropertyMode': {'args': '<type __builtin__.str>',
                                                              'longname': 'attributeFromHIKPropertyMode',
                                                              'modes': ['query'],
                                                              'numArgs': 1,
                                                              'shortname': 'mhk'},
                             'autoActivateBodyPart': {'args': '<type __builtin__.bool>',
                                                      'longname': 'autoActivateBodyPart',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'aab'},
                             'changePivotPlacement': {'args': '<type __builtin__.bool>',
                                                      'longname': 'changePivotPlacement',
                                                      'modes': ['edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'cpp'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'effectors': {'args': '<type __builtin__.str>',
                                           'longname': 'effectors',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ef'},
                             'fkSkeleton': {'args': '<type __builtin__.str>',
                                            'longname': 'fkSkeleton',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'fk'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'nm'},
                             'pinHandFeet': {'args': '<type __builtin__.bool>',
                                             'longname': 'pinHandFeet',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'phf'},
                             'placeNewPivot': {'args': '<type __builtin__.bool>',
                                               'longname': 'placeNewPivot',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'pnp'},
                             'posture': {'args': '<type __builtin__.str>',
                                         'longname': 'posture',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'pos'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'sourceSkeleton': {'args': '<type __builtin__.str>',
                                                'longname': 'sourceSkeleton',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sk'},
                             'stancePose': {'args': '<type __builtin__.str>',
                                            'longname': 'stancePose',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'sp'},
                             'type': {'args': '<type __builtin__.str>',
                                      'longname': 'type',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'typ'}},
                   'shortFlags': {'aab': 'autoActivateBodyPart',
                                  'aae': 'addAuxEffector',
                                  'afp': 'addFloorContactPlane',
                                  'ahk': 'attributeFromHIKProperty',
                                  'ame': 'addMissingEffectors',
                                  'apv': 'activatePivot',
                                  'cpp': 'changePivotPlacement',
                                  'e': 'edit',
                                  'ef': 'effectors',
                                  'fk': 'fkSkeleton',
                                  'mhk': 'attributeFromHIKPropertyMode',
                                  'nm': 'name',
                                  'phf': 'pinHandFeet',
                                  'pnp': 'placeNewPivot',
                                  'pos': 'posture',
                                  'q': 'query',
                                  'sk': 'sourceSkeleton',
                                  'sp': 'stancePose',
                                  'typ': 'type'},
                   'type': 'animation'},
  'checkBox': {'callbackFlags': ['dropCallback',
                                 'dpc',
                                 'visibleChangeCommand',
                                 'vcc',
                                 'onCommand',
                                 'onc',
                                 'dragCallback',
                                 'dgc',
                                 'offCommand',
                                 'ofc',
                                 'changeCommand',
                                 'cc'],
               'flags': {'align': {'args': '<type __builtin__.str>',
                                   'longname': 'align',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'al'},
                         'annotation': {'args': '<type __builtin__.str>',
                                        'longname': 'annotation',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ann'},
                         'backgroundColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'backgroundColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'bgc'},
                         'changeCommand': {'args': 'script',
                                           'longname': 'changeCommand',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cc'},
                         'defineTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'defineTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'dt'},
                         'docTag': {'args': '<type __builtin__.str>',
                                    'longname': 'docTag',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'dtg'},
                         'dragCallback': {'args': 'script',
                                          'longname': 'dragCallback',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dgc'},
                         'dropCallback': {'args': 'script',
                                          'longname': 'dropCallback',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dpc'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'editable': {'args': '<type __builtin__.bool>',
                                      'longname': 'editable',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'ed'},
                         'enable': {'args': '<type __builtin__.bool>',
                                    'longname': 'enable',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'en'},
                         'enableBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'enableBackground',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ebg'},
                         'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableKeyboardFocus',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ekf'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'fullPathName': {'args': '<type __builtin__.bool>',
                                          'longname': 'fullPathName',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'fpn'},
                         'height': {'args': '<type __builtin__.int>',
                                    'longname': 'height',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'h'},
                         'highlightColor': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'highlightColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'hlc'},
                         'isObscured': {'args': '<type __builtin__.bool>',
                                        'longname': 'isObscured',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'io'},
                         'label': {'args': '<type __builtin__.str>',
                                   'longname': 'label',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'l'},
                         'manage': {'args': '<type __builtin__.bool>',
                                    'longname': 'manage',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'm'},
                         'noBackground': {'args': '<type __builtin__.bool>',
                                          'longname': 'noBackground',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nbg'},
                         'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                'longname': 'numberOfPopupMenus',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'npm'},
                         'offCommand': {'args': 'script',
                                        'longname': 'offCommand',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ofc'},
                         'onCommand': {'args': 'script',
                                       'longname': 'onCommand',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'onc'},
                         'parent': {'args': '<type __builtin__.str>',
                                    'longname': 'parent',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'p'},
                         'popupMenuArray': {'args': '<type __builtin__.bool>',
                                            'longname': 'popupMenuArray',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'pma'},
                         'preventOverride': {'args': '<type __builtin__.bool>',
                                             'longname': 'preventOverride',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'po'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'recomputeSize': {'args': '<type __builtin__.bool>',
                                           'longname': 'recomputeSize',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'rs'},
                         'statusBarMessage': {'args': '<type __builtin__.str>',
                                              'longname': 'statusBarMessage',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sbm'},
                         'useTemplate': {'args': '<type __builtin__.str>',
                                         'longname': 'useTemplate',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ut'},
                         'value': {'args': '<type __builtin__.bool>',
                                   'longname': 'value',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'v'},
                         'visible': {'args': '<type __builtin__.bool>',
                                     'longname': 'visible',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'vis'},
                         'visibleChangeCommand': {'args': 'script',
                                                  'longname': 'visibleChangeCommand',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'vcc'},
                         'width': {'args': '<type __builtin__.int>',
                                   'longname': 'width',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'w'}},
               'shortFlags': {'al': 'align',
                              'ann': 'annotation',
                              'bgc': 'backgroundColor',
                              'cc': 'changeCommand',
                              'dgc': 'dragCallback',
                              'dpc': 'dropCallback',
                              'dt': 'defineTemplate',
                              'dtg': 'docTag',
                              'e': 'edit',
                              'ebg': 'enableBackground',
                              'ed': 'editable',
                              'ekf': 'enableKeyboardFocus',
                              'en': 'enable',
                              'ex': 'exists',
                              'fpn': 'fullPathName',
                              'h': 'height',
                              'hlc': 'highlightColor',
                              'io': 'isObscured',
                              'l': 'label',
                              'm': 'manage',
                              'nbg': 'noBackground',
                              'npm': 'numberOfPopupMenus',
                              'ofc': 'offCommand',
                              'onc': 'onCommand',
                              'p': 'parent',
                              'pma': 'popupMenuArray',
                              'po': 'preventOverride',
                              'q': 'query',
                              'rs': 'recomputeSize',
                              'sbm': 'statusBarMessage',
                              'ut': 'useTemplate',
                              'v': 'value',
                              'vcc': 'visibleChangeCommand',
                              'vis': 'visible',
                              'w': 'width'},
               'type': 'windows'},
  'checkBoxGrp': {'callbackFlags': ['changeCommand4',
                                    'cc4',
                                    'dragCallback',
                                    'dgc',
                                    'changeCommand3',
                                    'cc3',
                                    'changeCommand2',
                                    'cc2',
                                    'changeCommand1',
                                    'cc1',
                                    'offCommand',
                                    'ofc',
                                    'visibleChangeCommand',
                                    'vcc',
                                    'offCommand1',
                                    'of1',
                                    'offCommand3',
                                    'of3',
                                    'offCommand2',
                                    'of2',
                                    'offCommand4',
                                    'of4',
                                    'dropCallback',
                                    'dpc',
                                    'onCommand',
                                    'onc',
                                    'changeCommand',
                                    'cc',
                                    'onCommand4',
                                    'on4',
                                    'onCommand1',
                                    'on1',
                                    'onCommand2',
                                    'on2',
                                    'onCommand3',
                                    'on3'],
                  'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                 'longname': 'adjustableColumn',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'adj'},
                            'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn2',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ad2'},
                            'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn3',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ad3'},
                            'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn4',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ad4'},
                            'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn5',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ad5'},
                            'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn6',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ad6'},
                            'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'changeCommand': {'args': 'script',
                                              'longname': 'changeCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cc'},
                            'changeCommand1': {'args': 'script',
                                               'longname': 'changeCommand1',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cc1'},
                            'changeCommand2': {'args': 'script',
                                               'longname': 'changeCommand2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cc2'},
                            'changeCommand3': {'args': 'script',
                                               'longname': 'changeCommand3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cc3'},
                            'changeCommand4': {'args': 'script',
                                               'longname': 'changeCommand4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cc4'},
                            'columnAlign': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'columnAlign',
                                            'modes': ['create',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 2,
                                            'shortname': 'cal'},
                            'columnAlign2': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign2',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 2,
                                             'shortname': 'cl2'},
                            'columnAlign3': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign3',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 3,
                                             'shortname': 'cl3'},
                            'columnAlign4': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign4',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 4,
                                             'shortname': 'cl4'},
                            'columnAlign5': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign5',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 5,
                                             'shortname': 'cl5'},
                            'columnAlign6': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign6',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 6,
                                             'shortname': 'cl6'},
                            'columnAttach': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnAttach',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 3,
                                             'shortname': 'cat'},
                            'columnAttach2': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAttach2',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'ct2'},
                            'columnAttach3': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAttach3',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'ct3'},
                            'columnAttach4': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAttach4',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 4,
                                              'shortname': 'ct4'},
                            'columnAttach5': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAttach5',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 5,
                                              'shortname': 'ct5'},
                            'columnAttach6': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAttach6',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 6,
                                              'shortname': 'ct6'},
                            'columnOffset2': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnOffset2',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'co2'},
                            'columnOffset3': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnOffset3',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'co3'},
                            'columnOffset4': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnOffset4',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 4,
                                              'shortname': 'co4'},
                            'columnOffset5': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnOffset5',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 5,
                                              'shortname': 'co5'},
                            'columnOffset6': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnOffset6',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 6,
                                              'shortname': 'co6'},
                            'columnWidth': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'columnWidth',
                                            'modes': ['create',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 2,
                                            'shortname': 'cw'},
                            'columnWidth1': {'args': '<type __builtin__.int>',
                                             'longname': 'columnWidth1',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cw1'},
                            'columnWidth2': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth2',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 2,
                                             'shortname': 'cw2'},
                            'columnWidth3': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth3',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 3,
                                             'shortname': 'cw3'},
                            'columnWidth4': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth4',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 4,
                                             'shortname': 'cw4'},
                            'columnWidth5': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth5',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 5,
                                             'shortname': 'cw5'},
                            'columnWidth6': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth6',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 6,
                                             'shortname': 'cw6'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'editable': {'args': '<type __builtin__.bool>',
                                         'longname': 'editable',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ed'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enable1': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en1'},
                            'enable2': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en2'},
                            'enable3': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable3',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en3'},
                            'enable4': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable4',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en4'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'label': {'args': '<type __builtin__.str>',
                                      'longname': 'label',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'l'},
                            'label1': {'args': '<type __builtin__.str>',
                                       'longname': 'label1',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'l1'},
                            'label2': {'args': '<type __builtin__.str>',
                                       'longname': 'label2',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'l2'},
                            'label3': {'args': '<type __builtin__.str>',
                                       'longname': 'label3',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'l3'},
                            'label4': {'args': '<type __builtin__.str>',
                                       'longname': 'label4',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'l4'},
                            'labelArray2': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'labelArray2',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 2,
                                            'shortname': 'la2'},
                            'labelArray3': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'labelArray3',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'la3'},
                            'labelArray4': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'labelArray4',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 4,
                                            'shortname': 'la4'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfCheckBoxes': {'args': '<type __builtin__.int>',
                                                   'longname': 'numberOfCheckBoxes',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ncb'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'offCommand': {'args': 'script',
                                           'longname': 'offCommand',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ofc'},
                            'offCommand1': {'args': 'script',
                                            'longname': 'offCommand1',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'of1'},
                            'offCommand2': {'args': 'script',
                                            'longname': 'offCommand2',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'of2'},
                            'offCommand3': {'args': 'script',
                                            'longname': 'offCommand3',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'of3'},
                            'offCommand4': {'args': 'script',
                                            'longname': 'offCommand4',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'of4'},
                            'onCommand': {'args': 'script',
                                          'longname': 'onCommand',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'onc'},
                            'onCommand1': {'args': 'script',
                                           'longname': 'onCommand1',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'on1'},
                            'onCommand2': {'args': 'script',
                                           'longname': 'onCommand2',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'on2'},
                            'onCommand3': {'args': 'script',
                                           'longname': 'onCommand3',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'on3'},
                            'onCommand4': {'args': 'script',
                                           'longname': 'onCommand4',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'on4'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'rowAttach': {'args': ['<type __builtin__.int>',
                                                   '<type __builtin__.str>',
                                                   '<type __builtin__.int>'],
                                          'longname': 'rowAttach',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 3,
                                          'shortname': 'rat'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'value1': {'args': '<type __builtin__.bool>',
                                       'longname': 'value1',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'v1'},
                            'value2': {'args': '<type __builtin__.bool>',
                                       'longname': 'value2',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'v2'},
                            'value3': {'args': '<type __builtin__.bool>',
                                       'longname': 'value3',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'v3'},
                            'value4': {'args': '<type __builtin__.bool>',
                                       'longname': 'value4',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'v4'},
                            'valueArray2': {'args': ['<type __builtin__.bool>',
                                                     '<type __builtin__.bool>'],
                                            'longname': 'valueArray2',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 2,
                                            'shortname': 'va2'},
                            'valueArray3': {'args': ['<type __builtin__.bool>',
                                                     '<type __builtin__.bool>',
                                                     '<type __builtin__.bool>'],
                                            'longname': 'valueArray3',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'va3'},
                            'valueArray4': {'args': ['<type __builtin__.bool>',
                                                     '<type __builtin__.bool>',
                                                     '<type __builtin__.bool>',
                                                     '<type __builtin__.bool>'],
                                            'longname': 'valueArray4',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 4,
                                            'shortname': 'va4'},
                            'vertical': {'args': '<type __builtin__.bool>',
                                         'longname': 'vertical',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'vr'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'ad2': 'adjustableColumn2',
                                 'ad3': 'adjustableColumn3',
                                 'ad4': 'adjustableColumn4',
                                 'ad5': 'adjustableColumn5',
                                 'ad6': 'adjustableColumn6',
                                 'adj': 'adjustableColumn',
                                 'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'cal': 'columnAlign',
                                 'cat': 'columnAttach',
                                 'cc': 'changeCommand',
                                 'cc1': 'changeCommand1',
                                 'cc2': 'changeCommand2',
                                 'cc3': 'changeCommand3',
                                 'cc4': 'changeCommand4',
                                 'cl2': 'columnAlign2',
                                 'cl3': 'columnAlign3',
                                 'cl4': 'columnAlign4',
                                 'cl5': 'columnAlign5',
                                 'cl6': 'columnAlign6',
                                 'co2': 'columnOffset2',
                                 'co3': 'columnOffset3',
                                 'co4': 'columnOffset4',
                                 'co5': 'columnOffset5',
                                 'co6': 'columnOffset6',
                                 'ct2': 'columnAttach2',
                                 'ct3': 'columnAttach3',
                                 'ct4': 'columnAttach4',
                                 'ct5': 'columnAttach5',
                                 'ct6': 'columnAttach6',
                                 'cw': 'columnWidth',
                                 'cw1': 'columnWidth1',
                                 'cw2': 'columnWidth2',
                                 'cw3': 'columnWidth3',
                                 'cw4': 'columnWidth4',
                                 'cw5': 'columnWidth5',
                                 'cw6': 'columnWidth6',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ed': 'editable',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'en1': 'enable1',
                                 'en2': 'enable2',
                                 'en3': 'enable3',
                                 'en4': 'enable4',
                                 'ex': 'exists',
                                 'fpn': 'fullPathName',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'io': 'isObscured',
                                 'l': 'label',
                                 'l1': 'label1',
                                 'l2': 'label2',
                                 'l3': 'label3',
                                 'l4': 'label4',
                                 'la2': 'labelArray2',
                                 'la3': 'labelArray3',
                                 'la4': 'labelArray4',
                                 'm': 'manage',
                                 'nbg': 'noBackground',
                                 'ncb': 'numberOfCheckBoxes',
                                 'npm': 'numberOfPopupMenus',
                                 'of1': 'offCommand1',
                                 'of2': 'offCommand2',
                                 'of3': 'offCommand3',
                                 'of4': 'offCommand4',
                                 'ofc': 'offCommand',
                                 'on1': 'onCommand1',
                                 'on2': 'onCommand2',
                                 'on3': 'onCommand3',
                                 'on4': 'onCommand4',
                                 'onc': 'onCommand',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'rat': 'rowAttach',
                                 'sbm': 'statusBarMessage',
                                 'ut': 'useTemplate',
                                 'v1': 'value1',
                                 'v2': 'value2',
                                 'v3': 'value3',
                                 'v4': 'value4',
                                 'va2': 'valueArray2',
                                 'va3': 'valueArray3',
                                 'va4': 'valueArray4',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'vr': 'vertical',
                                 'w': 'width'},
                  'type': 'windows'},
  'checkDefaultRenderGlobals': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                   'longname': 'edit',
                                                   'numArgs': 0,
                                                   'shortname': 'e'},
                                          'query': {'args': '<type __builtin__.bool>',
                                                    'longname': 'query',
                                                    'numArgs': 0,
                                                    'shortname': 'q'}},
                                'shortFlags': {'e': 'edit', 'q': 'query'},
                                'type': 'rendering'},
  'choice': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                     'longname': 'attribute',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'at'},
                       'controlPoints': {'args': '<type __builtin__.bool>',
                                         'longname': 'controlPoints',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'cp'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'index': {'args': '<type __builtin__.int>',
                                 'longname': 'index',
                                 'modes': ['create', 'query'],
                                 'numArgs': 1,
                                 'shortname': 'index'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['create', 'query'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'selector': {'args': 'PyNode',
                                    'longname': 'selector',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'sl'},
                       'shape': {'args': '<type __builtin__.bool>',
                                 'longname': 'shape',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 's'},
                       'sourceAttribute': {'args': 'PyNode',
                                           'longname': 'sourceAttribute',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'sa'},
                       'time': {'args': 'time',
                                'longname': 'time',
                                'modes': ['create', 'multiuse'],
                                'numArgs': 1,
                                'shortname': 't'}},
             'removedFlags': {'in': 'index'},
             'shortFlags': {'at': 'attribute',
                            'cp': 'controlPoints',
                            'e': 'edit',
                            'index': 'index',
                            'n': 'name',
                            'q': 'query',
                            's': 'shape',
                            'sa': 'sourceAttribute',
                            'sl': 'selector',
                            't': 'time'},
             'type': 'animation'},
  'circle': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                   'longname': 'caching',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'resultNeedsCasting': True,
                                   'shortname': 'cch'},
                       'center': {'args': ['<type __builtin__.float>',
                                           '<type __builtin__.float>',
                                           '<type __builtin__.float>'],
                                  'longname': 'center',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 3,
                                  'shortname': 'c'},
                       'centerX': {'args': '<type __builtin__.float>',
                                   'longname': 'centerX',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'cx'},
                       'centerY': {'args': '<type __builtin__.float>',
                                   'longname': 'centerY',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'cy'},
                       'centerZ': {'args': '<type __builtin__.float>',
                                   'longname': 'centerZ',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'cz'},
                       'constructionHistory': {'args': '<type __builtin__.bool>',
                                               'longname': 'constructionHistory',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                       'degree': {'args': '<type __builtin__.int>',
                                  'longname': 'degree',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'resultNeedsCasting': True,
                                  'shortname': 'd'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'first': {'args': ['<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>'],
                                 'longname': 'first',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 3,
                                 'shortname': 'fp'},
                       'firstPointX': {'args': '<type __builtin__.float>',
                                       'longname': 'firstPointX',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'fpx'},
                       'firstPointY': {'args': '<type __builtin__.float>',
                                       'longname': 'firstPointY',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'fpy'},
                       'firstPointZ': {'args': '<type __builtin__.float>',
                                       'longname': 'firstPointZ',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'fpz'},
                       'fixCenter': {'args': '<type __builtin__.bool>',
                                     'longname': 'fixCenter',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsCasting': True,
                                     'shortname': 'fc'},
                       'frozen': {'args': '<type __builtin__.bool>',
                                  'longname': 'frozen',
                                  'modes': ['query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'fzn'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'nodeState': {'args': '<type __builtin__.int>',
                                     'longname': 'nodeState',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsCasting': True,
                                     'shortname': 'nds'},
                       'normal': {'args': ['<type __builtin__.float>',
                                           '<type __builtin__.float>',
                                           '<type __builtin__.float>'],
                                  'longname': 'normal',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 3,
                                  'shortname': 'nr'},
                       'normalX': {'args': '<type __builtin__.float>',
                                   'longname': 'normalX',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'nrx'},
                       'normalY': {'args': '<type __builtin__.float>',
                                   'longname': 'normalY',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'nry'},
                       'normalZ': {'args': '<type __builtin__.float>',
                                   'longname': 'normalZ',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'nrz'},
                       'object': {'args': '<type __builtin__.bool>',
                                  'longname': 'object',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'o'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'radius': {'args': '<type __builtin__.float>',
                                  'longname': 'radius',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'r'},
                       'sections': {'args': '<type __builtin__.int>',
                                    'longname': 'sections',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'resultNeedsCasting': True,
                                    'shortname': 's'},
                       'sweep': {'args': '<type __builtin__.float>',
                                 'longname': 'sweep',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'sw'},
                       'tolerance': {'args': '<type __builtin__.float>',
                                     'longname': 'tolerance',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'tol'},
                       'useTolerance': {'args': '<type __builtin__.bool>',
                                        'longname': 'useTolerance',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'ut'}},
             'shortFlags': {'c': 'center',
                            'cch': 'caching',
                            'ch': 'constructionHistory',
                            'cx': 'centerX',
                            'cy': 'centerY',
                            'cz': 'centerZ',
                            'd': 'degree',
                            'e': 'edit',
                            'fc': 'fixCenter',
                            'fp': 'first',
                            'fpx': 'firstPointX',
                            'fpy': 'firstPointY',
                            'fpz': 'firstPointZ',
                            'fzn': 'frozen',
                            'n': 'name',
                            'nds': 'nodeState',
                            'nr': 'normal',
                            'nrx': 'normalX',
                            'nry': 'normalY',
                            'nrz': 'normalZ',
                            'o': 'object',
                            'q': 'query',
                            'r': 'radius',
                            's': 'sections',
                            'sw': 'sweep',
                            'tol': 'tolerance',
                            'ut': 'useTolerance'},
             'type': 'modeling'},
  'circularFillet': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'curveOnSurface': {'args': '<type __builtin__.bool>',
                                                  'longname': 'curveOnSurface',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'cos'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'object': {'args': '<type __builtin__.bool>',
                                          'longname': 'object',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'o'},
                               'positionTolerance': {'args': '<type __builtin__.float>',
                                                     'longname': 'positionTolerance',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'pt'},
                               'primaryRadius': {'args': '<type __builtin__.float>',
                                                 'longname': 'primaryRadius',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'pr'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'secondaryRadius': {'args': '<type __builtin__.float>',
                                                   'longname': 'secondaryRadius',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sr'},
                               'tangentTolerance': {'args': '<type __builtin__.float>',
                                                    'longname': 'tangentTolerance',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'tt'}},
                     'shortFlags': {'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'cos': 'curveOnSurface',
                                    'e': 'edit',
                                    'fzn': 'frozen',
                                    'n': 'name',
                                    'nds': 'nodeState',
                                    'o': 'object',
                                    'pr': 'primaryRadius',
                                    'pt': 'positionTolerance',
                                    'q': 'query',
                                    'sr': 'secondaryRadius',
                                    'tt': 'tangentTolerance'},
                     'type': 'modeling'},
  'clearCache': {'flags': {'allNodes': {'args': '<type __builtin__.bool>',
                                        'longname': 'allNodes',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'all'},
                           'computed': {'args': '<type __builtin__.bool>',
                                        'longname': 'computed',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'c'},
                           'dirty': {'args': '<type __builtin__.bool>',
                                     'longname': 'dirty',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'd'}},
                 'shortFlags': {'all': 'allNodes',
                                'c': 'computed',
                                'd': 'dirty'},
                 'type': 'system'},
  'clearDynStartState': {'type': 'runtime'},
  'clearNClothStartState': {'type': 'runtime'},
  'clearShear': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'clip': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                  'longname': 'absolute',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'abs'},
                     'absoluteRotations': {'args': '<type __builtin__.bool>',
                                           'longname': 'absoluteRotations',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'abr'},
                     'active': {'args': '<type __builtin__.str>',
                                'longname': 'active',
                                'modes': ['query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'a'},
                     'addTrack': {'args': '<type __builtin__.bool>',
                                  'longname': 'addTrack',
                                  'modes': [],
                                  'numArgs': 0,
                                  'shortname': 'at'},
                     'allAbsolute': {'args': '<type __builtin__.bool>',
                                     'longname': 'allAbsolute',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'aa'},
                     'allClips': {'args': '<type __builtin__.bool>',
                                  'longname': 'allClips',
                                  'modes': ['query'],
                                  'numArgs': 0,
                                  'shortname': 'ac'},
                     'allRelative': {'args': '<type __builtin__.bool>',
                                     'longname': 'allRelative',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ar'},
                     'allSourceClips': {'args': '<type __builtin__.bool>',
                                        'longname': 'allSourceClips',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'asc'},
                     'animCurveRange': {'args': '<type __builtin__.bool>',
                                        'longname': 'animCurveRange',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'acr'},
                     'character': {'args': '<type __builtin__.bool>',
                                   'longname': 'character',
                                   'modes': ['query'],
                                   'numArgs': 0,
                                   'shortname': 'ch'},
                     'constraint': {'args': '<type __builtin__.bool>',
                                    'longname': 'constraint',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'cn'},
                     'copy': {'args': '<type __builtin__.bool>',
                              'longname': 'copy',
                              'modes': ['create', 'query'],
                              'numArgs': 0,
                              'shortname': 'c'},
                     'defaultAbsolute': {'args': '<type __builtin__.bool>',
                                         'longname': 'defaultAbsolute',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'da'},
                     'duplicate': {'args': '<type __builtin__.bool>',
                                   'longname': 'duplicate',
                                   'modes': ['query'],
                                   'numArgs': 0,
                                   'shortname': 'd'},
                     'edit': {'args': '<type __builtin__.bool>',
                              'longname': 'edit',
                              'numArgs': 0,
                              'shortname': 'e'},
                     'endTime': {'args': 'time',
                                 'longname': 'endTime',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'end'},
                     'expression': {'args': '<type __builtin__.bool>',
                                    'longname': 'expression',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                     'ignoreSubcharacters': {'args': '<type __builtin__.bool>',
                                             'longname': 'ignoreSubcharacters',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ignoreSubcharacters'},
                     'isolate': {'args': '<type __builtin__.bool>',
                                 'longname': 'isolate',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'i'},
                     'leaveOriginal': {'args': '<type __builtin__.bool>',
                                       'longname': 'leaveOriginal',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'lo'},
                     'mapMethod': {'args': '<type __builtin__.str>',
                                   'longname': 'mapMethod',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'mm'},
                     'name': {'args': '<type __builtin__.str>',
                              'longname': 'name',
                              'modes': ['create', 'query', 'multiuse'],
                              'numArgs': 1,
                              'shortname': 'n'},
                     'newName': {'args': '<type __builtin__.str>',
                                 'longname': 'newName',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'nn'},
                     'paste': {'args': '<type __builtin__.bool>',
                               'longname': 'paste',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'p'},
                     'pasteInstance': {'args': '<type __builtin__.bool>',
                                       'longname': 'pasteInstance',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'pi'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'remove': {'args': '<type __builtin__.bool>',
                                'longname': 'remove',
                                'modes': ['query'],
                                'numArgs': 0,
                                'shortname': 'rm'},
                     'removeTrack': {'args': '<type __builtin__.bool>',
                                     'longname': 'removeTrack',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'rt'},
                     'rotationOffset': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'rotationOffset',
                                        'modes': ['create', 'query'],
                                        'numArgs': 3,
                                        'shortname': 'rof'},
                     'rotationsAbsolute': {'args': '<type __builtin__.bool>',
                                           'longname': 'rotationsAbsolute',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ra'},
                     'scheduleClip': {'args': '<type __builtin__.bool>',
                                      'longname': 'scheduleClip',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'sc'},
                     'sourceClipName': {'args': '<type __builtin__.bool>',
                                        'longname': 'sourceClipName',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'scn'},
                     'split': {'args': 'time',
                               'longname': 'split',
                               'modes': ['create', 'edit'],
                               'numArgs': 1,
                               'shortname': 'sp'},
                     'startTime': {'args': 'time',
                                   'longname': 'startTime',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 's'},
                     'translationOffset': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'translationOffset',
                                           'modes': ['create', 'query'],
                                           'numArgs': 3,
                                           'shortname': 'tof'},
                     'useChannel': {'args': '<type __builtin__.str>',
                                    'longname': 'useChannel',
                                    'modes': ['create', 'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'uc'}},
           'removedFlags': {'is': 'ignoreSubcharacters'},
           'shortFlags': {'a': 'active',
                          'aa': 'allAbsolute',
                          'abr': 'absoluteRotations',
                          'abs': 'absolute',
                          'ac': 'allClips',
                          'acr': 'animCurveRange',
                          'ar': 'allRelative',
                          'asc': 'allSourceClips',
                          'at': 'addTrack',
                          'c': 'copy',
                          'ch': 'character',
                          'cn': 'constraint',
                          'd': 'duplicate',
                          'da': 'defaultAbsolute',
                          'e': 'edit',
                          'end': 'endTime',
                          'ex': 'expression',
                          'i': 'isolate',
                          'ignoreSubcharacters': 'ignoreSubcharacters',
                          'lo': 'leaveOriginal',
                          'mm': 'mapMethod',
                          'n': 'name',
                          'nn': 'newName',
                          'p': 'paste',
                          'pi': 'pasteInstance',
                          'q': 'query',
                          'ra': 'rotationsAbsolute',
                          'rm': 'remove',
                          'rof': 'rotationOffset',
                          'rt': 'removeTrack',
                          's': 'startTime',
                          'sc': 'scheduleClip',
                          'scn': 'sourceClipName',
                          'sp': 'split',
                          'tof': 'translationOffset',
                          'uc': 'useChannel'},
           'type': 'animation'},
  'clipEditor': {'flags': {'allTrackHeights': {'args': '<type __builtin__.int>',
                                               'longname': 'allTrackHeights',
                                               'modes': [],
                                               'numArgs': 1,
                                               'shortname': 'th'},
                           'autoFit': {'args': '<type __builtin__.str>',
                                       'longname': 'autoFit',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'af'},
                           'autoFitTime': {'args': '<type __builtin__.str>',
                                           'longname': 'autoFitTime',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'aft'},
                           'characterOutline': {'args': '<type __builtin__.str>',
                                                'longname': 'characterOutline',
                                                'numArgs': 1,
                                                'shortname': 'co'},
                           'clipDropCmd': {'args': '<type __builtin__.str>',
                                           'longname': 'clipDropCmd',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'cd'},
                           'clipStyle': {'args': '<type __builtin__.int>',
                                         'longname': 'clipStyle',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cs'},
                           'control': {'args': '<type __builtin__.bool>',
                                       'longname': 'control',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'ctl'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'deleteCmd': {'args': '<type __builtin__.str>',
                                         'longname': 'deleteCmd',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 'dc'},
                           'deselectAll': {'args': '<type __builtin__.bool>',
                                           'longname': 'deselectAll',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'da'},
                           'displayActiveKeyTangents': {'args': '<type __builtin__.str>',
                                                        'longname': 'displayActiveKeyTangents',
                                                        'modes': ['edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'dat'},
                           'displayActiveKeys': {'args': '<type __builtin__.str>',
                                                 'longname': 'displayActiveKeys',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dak'},
                           'displayInfinities': {'args': '<type __builtin__.str>',
                                                 'longname': 'displayInfinities',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'di'},
                           'displayKeys': {'args': '<type __builtin__.str>',
                                           'longname': 'displayKeys',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'dk'},
                           'displayTangents': {'args': '<type __builtin__.str>',
                                               'longname': 'displayTangents',
                                               'modes': ['edit'],
                                               'numArgs': 1,
                                               'shortname': 'dtn'},
                           'displayValues': {'args': '<type __builtin__.str>',
                                             'longname': 'displayValues',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'dv'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'filter': {'args': '<type __builtin__.str>',
                                      'longname': 'filter',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'f'},
                           'forceMainConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'forceMainConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'fmc'},
                           'frameAll': {'args': '<type __builtin__.bool>',
                                        'longname': 'frameAll',
                                        'modes': ['edit'],
                                        'numArgs': 0,
                                        'shortname': 'fa'},
                           'frameRange': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'frameRange',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 2,
                                          'shortname': 'fr'},
                           'highlightConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'highlightConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'hlc'},
                           'highlightedBlend': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'highlightedBlend',
                                                'modes': ['query'],
                                                'numArgs': 2,
                                                'shortname': 'hb'},
                           'highlightedClip': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'highlightedClip',
                                               'modes': ['query'],
                                               'numArgs': 2,
                                               'shortname': 'hc'},
                           'initialized': {'args': '<type __builtin__.bool>',
                                           'longname': 'initialized',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 'it'},
                           'listAllCharacters': {'args': '<type __builtin__.bool>',
                                                 'longname': 'listAllCharacters',
                                                 'modes': ['edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'lac'},
                           'listCurrentCharacters': {'args': '<type __builtin__.bool>',
                                                     'longname': 'listCurrentCharacters',
                                                     'modes': ['edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'lc'},
                           'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                  'longname': 'lockMainConnection',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'lck'},
                           'lookAt': {'args': '<type __builtin__.str>',
                                      'longname': 'lookAt',
                                      'modes': ['edit'],
                                      'numArgs': 1,
                                      'shortname': 'la'},
                           'mainListConnection': {'args': '<type __builtin__.str>',
                                                  'longname': 'mainListConnection',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mlc'},
                           'manageSequencer': {'args': '<type __builtin__.bool>',
                                               'longname': 'manageSequencer',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ms'},
                           'menuContext': {'args': '<type __builtin__.str>',
                                           'longname': 'menuContext',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 'mc'},
                           'panel': {'args': '<type __builtin__.str>',
                                     'longname': 'panel',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'pnl'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'selectBlend': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'selectBlend',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 3,
                                           'shortname': 'sb'},
                           'selectClip': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'selectClip',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 2,
                                          'shortname': 'sc'},
                           'selectionConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'selectionConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'slc'},
                           'snapTime': {'args': '<type __builtin__.str>',
                                        'longname': 'snapTime',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'st'},
                           'snapValue': {'args': '<type __builtin__.str>',
                                         'longname': 'snapValue',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'sv'},
                           'stateString': {'args': '<type __builtin__.bool>',
                                           'longname': 'stateString',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'sts'},
                           'unParent': {'args': '<type __builtin__.bool>',
                                        'longname': 'unParent',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'up'},
                           'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'unlockMainConnection',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'ulk'},
                           'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'updateMainConnection',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'upd'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'}},
                 'shortFlags': {'af': 'autoFit',
                                'aft': 'autoFitTime',
                                'cd': 'clipDropCmd',
                                'co': 'characterOutline',
                                'cs': 'clipStyle',
                                'ctl': 'control',
                                'da': 'deselectAll',
                                'dak': 'displayActiveKeys',
                                'dat': 'displayActiveKeyTangents',
                                'dc': 'deleteCmd',
                                'di': 'displayInfinities',
                                'dk': 'displayKeys',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'dtn': 'displayTangents',
                                'dv': 'displayValues',
                                'e': 'edit',
                                'ex': 'exists',
                                'f': 'filter',
                                'fa': 'frameAll',
                                'fmc': 'forceMainConnection',
                                'fr': 'frameRange',
                                'hb': 'highlightedBlend',
                                'hc': 'highlightedClip',
                                'hlc': 'highlightConnection',
                                'it': 'initialized',
                                'la': 'lookAt',
                                'lac': 'listAllCharacters',
                                'lc': 'listCurrentCharacters',
                                'lck': 'lockMainConnection',
                                'mc': 'menuContext',
                                'mlc': 'mainListConnection',
                                'ms': 'manageSequencer',
                                'p': 'parent',
                                'pnl': 'panel',
                                'q': 'query',
                                'sb': 'selectBlend',
                                'sc': 'selectClip',
                                'slc': 'selectionConnection',
                                'st': 'snapTime',
                                'sts': 'stateString',
                                'sv': 'snapValue',
                                'ulk': 'unlockMainConnection',
                                'up': 'unParent',
                                'upd': 'updateMainConnection',
                                'ut': 'useTemplate'},
                 'type': 'animation'},
  'clipEditorCurrentTimeCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'edit',
                                                  'numArgs': 0,
                                                  'shortname': 'e'},
                                         'exists': {'args': '<type __builtin__.bool>',
                                                    'longname': 'exists',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ex'},
                                         'history': {'args': '<type __builtin__.bool>',
                                                     'longname': 'history',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                                         'image1': {'args': '<type __builtin__.str>',
                                                    'longname': 'image1',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'i1'},
                                         'image2': {'args': '<type __builtin__.str>',
                                                    'longname': 'image2',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'i2'},
                                         'image3': {'args': '<type __builtin__.str>',
                                                    'longname': 'image3',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'i3'},
                                         'name': {'args': '<type __builtin__.str>',
                                                  'longname': 'name',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'n'},
                                         'query': {'args': '<type __builtin__.bool>',
                                                   'longname': 'query',
                                                   'numArgs': 0,
                                                   'shortname': 'q'}},
                               'shortFlags': {'ch': 'history',
                                              'e': 'edit',
                                              'ex': 'exists',
                                              'i1': 'image1',
                                              'i2': 'image2',
                                              'i3': 'image3',
                                              'n': 'name',
                                              'q': 'query'},
                               'type': 'context'},
  'clipMatching': {'flags': {'clipDst': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'clipDst',
                                         'modes': ['create'],
                                         'numArgs': 2,
                                         'shortname': 'cd'},
                             'clipSrc': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'clipSrc',
                                         'modes': ['create'],
                                         'numArgs': 2,
                                         'shortname': 'cs'},
                             'matchRotation': {'args': '<type __builtin__.int>',
                                               'longname': 'matchRotation',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'mr'},
                             'matchTranslation': {'args': '<type __builtin__.int>',
                                                  'longname': 'matchTranslation',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'mt'}},
                   'shortFlags': {'cd': 'clipDst',
                                  'cs': 'clipSrc',
                                  'mr': 'matchRotation',
                                  'mt': 'matchTranslation'},
                   'type': 'animation'},
  'clipSchedule': {'flags': {'allAbsolute': {'args': '<type __builtin__.bool>',
                                             'longname': 'allAbsolute',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'aa'},
                             'allRelative': {'args': '<type __builtin__.bool>',
                                             'longname': 'allRelative',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'ar'},
                             'blend': {'args': ['<type __builtin__.int>',
                                                '<type __builtin__.int>'],
                                       'longname': 'blend',
                                       'modes': ['create', 'query'],
                                       'numArgs': 2,
                                       'shortname': 'b'},
                             'blendNode': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'blendNode',
                                           'modes': ['query'],
                                           'numArgs': 2,
                                           'shortname': 'bn'},
                             'blendUsingNode': {'args': '<type __builtin__.str>',
                                                'longname': 'blendUsingNode',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'bun'},
                             'character': {'args': '<type __builtin__.bool>',
                                           'longname': 'character',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'ch'},
                             'clipIndex': {'args': '<type __builtin__.int>',
                                           'longname': 'clipIndex',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'ci'},
                             'cycle': {'args': '<type __builtin__.float>',
                                       'longname': 'cycle',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'c'},
                             'defaultAbsolute': {'args': '<type __builtin__.bool>',
                                                 'longname': 'defaultAbsolute',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'da'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'group': {'args': '<type __builtin__.bool>',
                                       'longname': 'group',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'grp'},
                             'groupIndex': {'args': '<type __builtin__.int>',
                                            'longname': 'groupIndex',
                                            'modes': ['create',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'gri'},
                             'groupName': {'args': '<type __builtin__.str>',
                                           'longname': 'groupName',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'gn'},
                             'hold': {'args': 'time',
                                      'longname': 'hold',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'ph'},
                             'insertTrack': {'args': '<type __builtin__.int>',
                                             'longname': 'insertTrack',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'it'},
                             'instance': {'args': '<type __builtin__.str>',
                                          'longname': 'instance',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'instance'},
                             'listCurves': {'args': '<type __builtin__.bool>',
                                            'longname': 'listCurves',
                                            'modes': ['create', 'query'],
                                            'numArgs': 0,
                                            'shortname': 'lc'},
                             'listPairs': {'args': '<type __builtin__.bool>',
                                           'longname': 'listPairs',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'lp'},
                             'lock': {'args': '<type __builtin__.bool>',
                                      'longname': 'lock',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'l'},
                             'mute': {'args': '<type __builtin__.bool>',
                                      'longname': 'mute',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'postCycle': {'args': '<type __builtin__.float>',
                                           'longname': 'postCycle',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'poc'},
                             'preCycle': {'args': '<type __builtin__.float>',
                                          'longname': 'preCycle',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'prc'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'remove': {'args': '<type __builtin__.bool>',
                                        'longname': 'remove',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'rm'},
                             'removeBlend': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'removeBlend',
                                             'modes': ['create'],
                                             'numArgs': 2,
                                             'shortname': 'rb'},
                             'removeEmptyTracks': {'args': '<type __builtin__.bool>',
                                                   'longname': 'removeEmptyTracks',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'ret'},
                             'removeTrack': {'args': '<type __builtin__.int>',
                                             'longname': 'removeTrack',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'rt'},
                             'rotationsAbsolute': {'args': '<type __builtin__.bool>',
                                                   'longname': 'rotationsAbsolute',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'ra'},
                             'scale': {'args': '<type __builtin__.float>',
                                       'longname': 'scale',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'sc'},
                             'shift': {'args': '<type __builtin__.int>',
                                       'longname': 'shift',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'sh'},
                             'shiftIndex': {'args': '<type __builtin__.int>',
                                            'longname': 'shiftIndex',
                                            'modes': ['create',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'shi'},
                             'solo': {'args': '<type __builtin__.bool>',
                                      'longname': 'solo',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'so'},
                             'sourceClipName': {'args': '<type __builtin__.bool>',
                                                'longname': 'sourceClipName',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 0,
                                                'shortname': 'scn'},
                             'sourceEnd': {'args': 'time',
                                           'longname': 'sourceEnd',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'se'},
                             'sourceStart': {'args': 'time',
                                             'longname': 'sourceStart',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'ss'},
                             'start': {'args': 'time',
                                       'longname': 'start',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 's'},
                             'track': {'args': '<type __builtin__.int>',
                                       'longname': 'track',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 't'},
                             'weight': {'args': '<type __builtin__.float>',
                                        'longname': 'weight',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'w'},
                             'weightStyle': {'args': '<type __builtin__.int>',
                                             'longname': 'weightStyle',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'ws'}},
                   'removedFlags': {'in': 'instance'},
                   'shortFlags': {'aa': 'allAbsolute',
                                  'ar': 'allRelative',
                                  'b': 'blend',
                                  'bn': 'blendNode',
                                  'bun': 'blendUsingNode',
                                  'c': 'cycle',
                                  'ch': 'character',
                                  'ci': 'clipIndex',
                                  'da': 'defaultAbsolute',
                                  'e': 'edit',
                                  'en': 'enable',
                                  'gn': 'groupName',
                                  'gri': 'groupIndex',
                                  'grp': 'group',
                                  'instance': 'instance',
                                  'it': 'insertTrack',
                                  'l': 'lock',
                                  'lc': 'listCurves',
                                  'lp': 'listPairs',
                                  'm': 'mute',
                                  'n': 'name',
                                  'ph': 'hold',
                                  'poc': 'postCycle',
                                  'prc': 'preCycle',
                                  'q': 'query',
                                  'ra': 'rotationsAbsolute',
                                  'rb': 'removeBlend',
                                  'ret': 'removeEmptyTracks',
                                  'rm': 'remove',
                                  'rt': 'removeTrack',
                                  's': 'start',
                                  'sc': 'scale',
                                  'scn': 'sourceClipName',
                                  'se': 'sourceEnd',
                                  'sh': 'shift',
                                  'shi': 'shiftIndex',
                                  'so': 'solo',
                                  'ss': 'sourceStart',
                                  't': 'track',
                                  'w': 'weight',
                                  'ws': 'weightStyle'},
                   'type': 'animation'},
  'clipSchedulerOutliner': {'callbackFlags': ['visibleChangeCommand',
                                              'vcc',
                                              'dragCallback',
                                              'dgc',
                                              'dropCallback',
                                              'dpc'],
                            'flags': {'annotation': {'args': '<type __builtin__.str>',
                                                     'longname': 'annotation',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ann'},
                                      'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                   '<type __builtin__.float>',
                                                                   '<type __builtin__.float>'],
                                                          'longname': 'backgroundColor',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 3,
                                                          'shortname': 'bgc'},
                                      'clipScheduler': {'args': '<type __builtin__.str>',
                                                        'longname': 'clipScheduler',
                                                        'modes': ['edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'cs'},
                                      'defineTemplate': {'args': '<type __builtin__.str>',
                                                         'longname': 'defineTemplate',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'dt'},
                                      'docTag': {'args': '<type __builtin__.str>',
                                                 'longname': 'docTag',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dtg'},
                                      'dragCallback': {'args': 'script',
                                                       'longname': 'dragCallback',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'dgc'},
                                      'dropCallback': {'args': 'script',
                                                       'longname': 'dropCallback',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'dpc'},
                                      'edit': {'args': '<type __builtin__.bool>',
                                               'longname': 'edit',
                                               'numArgs': 0,
                                               'shortname': 'e'},
                                      'enable': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enable',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'en'},
                                      'enableBackground': {'args': '<type __builtin__.bool>',
                                                           'longname': 'enableBackground',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ebg'},
                                      'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                              'longname': 'enableKeyboardFocus',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'ekf'},
                                      'exists': {'args': '<type __builtin__.bool>',
                                                 'longname': 'exists',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ex'},
                                      'fullPathName': {'args': '<type __builtin__.bool>',
                                                       'longname': 'fullPathName',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'fpn'},
                                      'height': {'args': '<type __builtin__.int>',
                                                 'longname': 'height',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'h'},
                                      'highlightColor': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'highlightColor',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 3,
                                                         'shortname': 'hlc'},
                                      'isObscured': {'args': '<type __builtin__.bool>',
                                                     'longname': 'isObscured',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'io'},
                                      'manage': {'args': '<type __builtin__.bool>',
                                                 'longname': 'manage',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'm'},
                                      'noBackground': {'args': '<type __builtin__.bool>',
                                                       'longname': 'noBackground',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'nbg'},
                                      'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                             'longname': 'numberOfPopupMenus',
                                                             'modes': ['query'],
                                                             'numArgs': 0,
                                                             'shortname': 'npm'},
                                      'parent': {'args': '<type __builtin__.str>',
                                                 'longname': 'parent',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'p'},
                                      'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                         'longname': 'popupMenuArray',
                                                         'modes': ['query'],
                                                         'numArgs': 0,
                                                         'shortname': 'pma'},
                                      'preventOverride': {'args': '<type __builtin__.bool>',
                                                          'longname': 'preventOverride',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'po'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'statusBarMessage': {'args': '<type __builtin__.str>',
                                                           'longname': 'statusBarMessage',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'sbm'},
                                      'useTemplate': {'args': '<type __builtin__.str>',
                                                      'longname': 'useTemplate',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ut'},
                                      'visible': {'args': '<type __builtin__.bool>',
                                                  'longname': 'visible',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'vis'},
                                      'visibleChangeCommand': {'args': 'script',
                                                               'longname': 'visibleChangeCommand',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'vcc'},
                                      'width': {'args': '<type __builtin__.int>',
                                                'longname': 'width',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'w'}},
                            'shortFlags': {'ann': 'annotation',
                                           'bgc': 'backgroundColor',
                                           'cs': 'clipScheduler',
                                           'dgc': 'dragCallback',
                                           'dpc': 'dropCallback',
                                           'dt': 'defineTemplate',
                                           'dtg': 'docTag',
                                           'e': 'edit',
                                           'ebg': 'enableBackground',
                                           'ekf': 'enableKeyboardFocus',
                                           'en': 'enable',
                                           'ex': 'exists',
                                           'fpn': 'fullPathName',
                                           'h': 'height',
                                           'hlc': 'highlightColor',
                                           'io': 'isObscured',
                                           'm': 'manage',
                                           'nbg': 'noBackground',
                                           'npm': 'numberOfPopupMenus',
                                           'p': 'parent',
                                           'pma': 'popupMenuArray',
                                           'po': 'preventOverride',
                                           'q': 'query',
                                           'sbm': 'statusBarMessage',
                                           'ut': 'useTemplate',
                                           'vcc': 'visibleChangeCommand',
                                           'vis': 'visible',
                                           'w': 'width'},
                            'type': 'animation'},
  'closeCurve': {'flags': {'blendBias': {'args': '<type __builtin__.float>',
                                         'longname': 'blendBias',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'bb'},
                           'blendKnotInsertion': {'args': '<type __builtin__.bool>',
                                                  'longname': 'blendKnotInsertion',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'bki'},
                           'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'curveOnSurface': {'args': '<type __builtin__.bool>',
                                              'longname': 'curveOnSurface',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'cos'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'object': {'args': '<type __builtin__.bool>',
                                      'longname': 'object',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'parameter': {'args': '<type __builtin__.float>',
                                         'longname': 'parameter',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'p'},
                           'preserveShape': {'args': '<type __builtin__.int>',
                                             'longname': 'preserveShape',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ps'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'replaceOriginal': {'args': '<type __builtin__.bool>',
                                               'longname': 'replaceOriginal',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'rpo'}},
                 'shortFlags': {'bb': 'blendBias',
                                'bki': 'blendKnotInsertion',
                                'cch': 'caching',
                                'ch': 'constructionHistory',
                                'cos': 'curveOnSurface',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'n': 'name',
                                'nds': 'nodeState',
                                'o': 'object',
                                'p': 'parameter',
                                'ps': 'preserveShape',
                                'q': 'query',
                                'rpo': 'replaceOriginal'},
                 'type': 'modeling'},
  'closeSurface': {'flags': {'blendBias': {'args': '<type __builtin__.float>',
                                           'longname': 'blendBias',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'bb'},
                             'blendKnotInsertion': {'args': '<type __builtin__.bool>',
                                                    'longname': 'blendKnotInsertion',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'bki'},
                             'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'direction': {'args': '<type __builtin__.int>',
                                           'longname': 'direction',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'd'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'object': {'args': '<type __builtin__.bool>',
                                        'longname': 'object',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                             'parameter': {'args': '<type __builtin__.float>',
                                           'longname': 'parameter',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'p'},
                             'preserveShape': {'args': '<type __builtin__.int>',
                                               'longname': 'preserveShape',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ps'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                 'longname': 'replaceOriginal',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'rpo'}},
                   'shortFlags': {'bb': 'blendBias',
                                  'bki': 'blendKnotInsertion',
                                  'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'd': 'direction',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'o': 'object',
                                  'p': 'parameter',
                                  'ps': 'preserveShape',
                                  'q': 'query',
                                  'rpo': 'replaceOriginal'},
                   'type': 'modeling'},
  'cluster': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                  'longname': 'after',
                                  'modes': ['create', 'edit'],
                                  'numArgs': 0,
                                  'shortname': 'af'},
                        'afterReference': {'args': '<type __builtin__.bool>',
                                           'longname': 'afterReference',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'ar'},
                        'before': {'args': '<type __builtin__.bool>',
                                   'longname': 'before',
                                   'modes': ['create', 'edit'],
                                   'numArgs': 0,
                                   'shortname': 'bf'},
                        'bindState': {'args': '<type __builtin__.bool>',
                                      'longname': 'bindState',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'bs'},
                        'deformerTools': {'args': '<type __builtin__.bool>',
                                          'longname': 'deformerTools',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'dt'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'envelope': {'args': '<type __builtin__.float>',
                                     'longname': 'envelope',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'en'},
                        'exclusive': {'args': '<type __builtin__.str>',
                                      'longname': 'exclusive',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'ex'},
                        'frontOfChain': {'args': '<type __builtin__.bool>',
                                         'longname': 'frontOfChain',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'foc'},
                        'geometry': {'args': '<type __builtin__.str>',
                                     'longname': 'geometry',
                                     'modes': ['query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'g'},
                        'geometryIndices': {'args': '<type __builtin__.bool>',
                                            'longname': 'geometryIndices',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'gi'},
                        'ignoreSelected': {'args': '<type __builtin__.bool>',
                                           'longname': 'ignoreSelected',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ignoreSelected'},
                        'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                    'longname': 'includeHiddenSelections',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ihs'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'parallel': {'args': '<type __builtin__.bool>',
                                     'longname': 'parallel',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'par'},
                        'prune': {'args': '<type __builtin__.bool>',
                                  'longname': 'prune',
                                  'modes': ['edit'],
                                  'numArgs': 0,
                                  'shortname': 'pr'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'relative': {'args': '<type __builtin__.bool>',
                                     'longname': 'relative',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'rel'},
                        'remove': {'args': '<type __builtin__.bool>',
                                   'longname': 'remove',
                                   'modes': ['edit', 'multiuse'],
                                   'numArgs': 0,
                                   'shortname': 'rm'},
                        'resetGeometry': {'args': '<type __builtin__.bool>',
                                          'longname': 'resetGeometry',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'rg'},
                        'split': {'args': '<type __builtin__.bool>',
                                  'longname': 'split',
                                  'modes': ['create', 'edit'],
                                  'numArgs': 0,
                                  'shortname': 'sp'},
                        'weightedNode': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'weightedNode',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 2,
                                         'shortname': 'wn'}},
              'removedFlags': {'is': 'ignoreSelected'},
              'shortFlags': {'af': 'after',
                             'ar': 'afterReference',
                             'bf': 'before',
                             'bs': 'bindState',
                             'dt': 'deformerTools',
                             'e': 'edit',
                             'en': 'envelope',
                             'ex': 'exclusive',
                             'foc': 'frontOfChain',
                             'g': 'geometry',
                             'gi': 'geometryIndices',
                             'ignoreSelected': 'ignoreSelected',
                             'ihs': 'includeHiddenSelections',
                             'n': 'name',
                             'par': 'parallel',
                             'pr': 'prune',
                             'q': 'query',
                             'rel': 'relative',
                             'rg': 'resetGeometry',
                             'rm': 'remove',
                             'sp': 'split',
                             'wn': 'weightedNode'},
              'type': 'animation'},
  'cmdFileOutput': {'flags': {'close': {'args': '<type __builtin__.int>',
                                        'longname': 'close',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'c'},
                              'closeAll': {'args': '<type __builtin__.bool>',
                                           'longname': 'closeAll',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ca'},
                              'open': {'args': '<type __builtin__.str>',
                                       'longname': 'open',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'status': {'args': '<type __builtin__.int>',
                                         'longname': 'status',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 's'}},
                    'shortFlags': {'c': 'close',
                                   'ca': 'closeAll',
                                   'o': 'open',
                                   'q': 'query',
                                   's': 'status'},
                    'type': 'system'},
  'cmdScrollFieldExecuter': {'callbackFlags': ['fileChangedCommand',
                                               'fcc',
                                               'visibleChangeCommand',
                                               'vcc',
                                               'dragCallback',
                                               'dgc',
                                               'receiveFocusCommand',
                                               'rfc',
                                               'modificationChangedCommand',
                                               'mcc',
                                               'filterKeyPress',
                                               'fkp',
                                               'commandCompletion',
                                               'cco',
                                               'dropCallback',
                                               'dpc'],
                             'flags': {'annotation': {'args': '<type __builtin__.str>',
                                                      'longname': 'annotation',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ann'},
                                       'appendText': {'args': '<type __builtin__.str>',
                                                      'longname': 'appendText',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'at'},
                                       'autoCloseBraces': {'args': '<type __builtin__.bool>',
                                                           'longname': 'autoCloseBraces',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'acb'},
                                       'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                    '<type __builtin__.float>',
                                                                    '<type __builtin__.float>'],
                                                           'longname': 'backgroundColor',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 3,
                                                           'shortname': 'bgc'},
                                       'clear': {'args': '<type __builtin__.bool>',
                                                 'longname': 'clear',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'clr'},
                                       'commandCompletion': {'args': '<type __builtin__.bool>',
                                                             'longname': 'commandCompletion',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'cco'},
                                       'copySelection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'copySelection',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'cp'},
                                       'currentLine': {'args': '<type __builtin__.int>',
                                                       'longname': 'currentLine',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'cl'},
                                       'cutSelection': {'args': '<type __builtin__.bool>',
                                                        'longname': 'cutSelection',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'ct'},
                                       'defineTemplate': {'args': '<type __builtin__.str>',
                                                          'longname': 'defineTemplate',
                                                          'modes': ['create'],
                                                          'numArgs': 1,
                                                          'shortname': 'dt'},
                                       'docTag': {'args': '<type __builtin__.str>',
                                                  'longname': 'docTag',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dtg'},
                                       'dragCallback': {'args': 'script',
                                                        'longname': 'dragCallback',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'dgc'},
                                       'dropCallback': {'args': 'script',
                                                        'longname': 'dropCallback',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'dpc'},
                                       'edit': {'args': '<type __builtin__.bool>',
                                                'longname': 'edit',
                                                'numArgs': 0,
                                                'shortname': 'e'},
                                       'enable': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enable',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'en'},
                                       'enableBackground': {'args': '<type __builtin__.bool>',
                                                            'longname': 'enableBackground',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ebg'},
                                       'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                               'longname': 'enableKeyboardFocus',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'ekf'},
                                       'execute': {'args': '<type __builtin__.bool>',
                                                   'longname': 'execute',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'exc'},
                                       'executeAll': {'args': '<type __builtin__.bool>',
                                                      'longname': 'executeAll',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'exa'},
                                       'exists': {'args': '<type __builtin__.bool>',
                                                  'longname': 'exists',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'ex'},
                                       'fileChangedCommand': {'args': 'script',
                                                              'longname': 'fileChangedCommand',
                                                              'modes': ['create',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'fcc'},
                                       'filename': {'args': '<type __builtin__.bool>',
                                                    'longname': 'filename',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'fn'},
                                       'filterKeyPress': {'args': 'script',
                                                          'longname': 'filterKeyPress',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'fkp'},
                                       'fullPathName': {'args': '<type __builtin__.bool>',
                                                        'longname': 'fullPathName',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'fpn'},
                                       'hasFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'hasFocus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'hf'},
                                       'hasSelection': {'args': '<type __builtin__.bool>',
                                                        'longname': 'hasSelection',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'hsl'},
                                       'height': {'args': '<type __builtin__.int>',
                                                  'longname': 'height',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'h'},
                                       'highlightColor': {'args': ['<type __builtin__.float>',
                                                                   '<type __builtin__.float>',
                                                                   '<type __builtin__.float>'],
                                                          'longname': 'highlightColor',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 3,
                                                          'shortname': 'hlc'},
                                       'indentSelection': {'args': '<type __builtin__.bool>',
                                                           'longname': 'indentSelection',
                                                           'numArgs': 0,
                                                           'shortname': 'isl'},
                                       'insertText': {'args': '<type __builtin__.str>',
                                                      'longname': 'insertText',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'it'},
                                       'isObscured': {'args': '<type __builtin__.bool>',
                                                      'longname': 'isObscured',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'io'},
                                       'load': {'args': '<type __builtin__.bool>',
                                                'longname': 'load',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'ld'},
                                       'loadContents': {'args': '<type __builtin__.str>',
                                                        'longname': 'loadContents',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ldc'},
                                       'loadFile': {'args': '<type __builtin__.str>',
                                                    'longname': 'loadFile',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ldf'},
                                       'manage': {'args': '<type __builtin__.bool>',
                                                  'longname': 'manage',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'm'},
                                       'modificationChangedCommand': {'args': 'script',
                                                                      'longname': 'modificationChangedCommand',
                                                                      'modes': ['create',
                                                                                'edit'],
                                                                      'numArgs': 1,
                                                                      'shortname': 'mcc'},
                                       'modified': {'args': '<type __builtin__.bool>',
                                                    'longname': 'modified',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'mod'},
                                       'noBackground': {'args': '<type __builtin__.bool>',
                                                        'longname': 'noBackground',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'nbg'},
                                       'numberOfLines': {'args': '<type __builtin__.int>',
                                                         'longname': 'numberOfLines',
                                                         'modes': ['query'],
                                                         'numArgs': 1,
                                                         'shortname': 'nl'},
                                       'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                              'longname': 'numberOfPopupMenus',
                                                              'modes': ['query'],
                                                              'numArgs': 0,
                                                              'shortname': 'npm'},
                                       'objectPathCompletion': {'args': '<type __builtin__.bool>',
                                                                'longname': 'objectPathCompletion',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'opc'},
                                       'parent': {'args': '<type __builtin__.str>',
                                                  'longname': 'parent',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'p'},
                                       'pasteSelection': {'args': '<type __builtin__.bool>',
                                                          'longname': 'pasteSelection',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 0,
                                                          'shortname': 'pst'},
                                       'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                          'longname': 'popupMenuArray',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'pma'},
                                       'preventOverride': {'args': '<type __builtin__.bool>',
                                                           'longname': 'preventOverride',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'po'},
                                       'query': {'args': '<type __builtin__.bool>',
                                                 'longname': 'query',
                                                 'numArgs': 0,
                                                 'shortname': 'q'},
                                       'receiveFocusCommand': {'args': '<type __builtin__.callable>',
                                                               'longname': 'receiveFocusCommand',
                                                               'numArgs': 1,
                                                               'shortname': 'rfc'},
                                       'redo': {'args': '<type __builtin__.bool>',
                                                'longname': 'redo',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'rd'},
                                       'removeStoredContents': {'args': '<type __builtin__.str>',
                                                                'longname': 'removeStoredContents',
                                                                'modes': ['create',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'rsc'},
                                       'replaceAll': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'replaceAll',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'rpa'},
                                       'saveFile': {'args': '<type __builtin__.str>',
                                                    'longname': 'saveFile',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'svf'},
                                       'saveSelection': {'args': '<type __builtin__.str>',
                                                         'longname': 'saveSelection',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'sv'},
                                       'saveSelectionToShelf': {'args': '<type __builtin__.bool>',
                                                                'longname': 'saveSelectionToShelf',
                                                                'modes': ['create',
                                                                          'edit'],
                                                                'numArgs': 0,
                                                                'shortname': 'svs'},
                                       'searchAndSelect': {'args': '<type __builtin__.bool>',
                                                           'longname': 'searchAndSelect',
                                                           'modes': ['query'],
                                                           'numArgs': 0,
                                                           'shortname': 'sas'},
                                       'searchDown': {'args': '<type __builtin__.bool>',
                                                      'longname': 'searchDown',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'sd'},
                                       'searchMatchCase': {'args': '<type __builtin__.bool>',
                                                           'longname': 'searchMatchCase',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'smc'},
                                       'searchString': {'args': '<type __builtin__.str>',
                                                        'longname': 'searchString',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ss'},
                                       'searchWraps': {'args': '<type __builtin__.bool>',
                                                       'longname': 'searchWraps',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'sw'},
                                       'select': {'args': ['<type __builtin__.int>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'select',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 2,
                                                  'shortname': 'sl'},
                                       'selectAll': {'args': '<type __builtin__.bool>',
                                                     'longname': 'selectAll',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'sla'},
                                       'selectedText': {'args': '<type __builtin__.bool>',
                                                        'longname': 'selectedText',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'slt'},
                                       'showLineNumbers': {'args': '<type __builtin__.bool>',
                                                           'longname': 'showLineNumbers',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'sln'},
                                       'showTooltipHelp': {'args': '<type __builtin__.bool>',
                                                           'longname': 'showTooltipHelp',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'sth'},
                                       'source': {'args': '<type __builtin__.bool>',
                                                  'longname': 'source',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'src'},
                                       'sourceType': {'args': '<type __builtin__.str>',
                                                      'longname': 'sourceType',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'st'},
                                       'spacesPerTab': {'args': '<type __builtin__.int>',
                                                        'longname': 'spacesPerTab',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'spt'},
                                       'statusBarMessage': {'args': '<type __builtin__.str>',
                                                            'longname': 'statusBarMessage',
                                                            'modes': ['create',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'sbm'},
                                       'storeContents': {'args': '<type __builtin__.str>',
                                                         'longname': 'storeContents',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'stc'},
                                       'tabsForIndent': {'args': '<type __builtin__.bool>',
                                                         'longname': 'tabsForIndent',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'tfi'},
                                       'text': {'args': '<type __builtin__.str>',
                                                'longname': 'text',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 't'},
                                       'textLength': {'args': '<type __builtin__.bool>',
                                                      'longname': 'textLength',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'tl'},
                                       'undo': {'args': '<type __builtin__.bool>',
                                                'longname': 'undo',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'ud'},
                                       'unindentSelection': {'args': '<type __builtin__.bool>',
                                                             'longname': 'unindentSelection',
                                                             'numArgs': 0,
                                                             'shortname': 'usl'},
                                       'useTemplate': {'args': '<type __builtin__.str>',
                                                       'longname': 'useTemplate',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ut'},
                                       'visible': {'args': '<type __builtin__.bool>',
                                                   'longname': 'visible',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'vis'},
                                       'visibleChangeCommand': {'args': 'script',
                                                                'longname': 'visibleChangeCommand',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'vcc'},
                                       'width': {'args': '<type __builtin__.int>',
                                                 'longname': 'width',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'w'}},
                             'shortFlags': {'acb': 'autoCloseBraces',
                                            'ann': 'annotation',
                                            'at': 'appendText',
                                            'bgc': 'backgroundColor',
                                            'cco': 'commandCompletion',
                                            'cl': 'currentLine',
                                            'clr': 'clear',
                                            'cp': 'copySelection',
                                            'ct': 'cutSelection',
                                            'dgc': 'dragCallback',
                                            'dpc': 'dropCallback',
                                            'dt': 'defineTemplate',
                                            'dtg': 'docTag',
                                            'e': 'edit',
                                            'ebg': 'enableBackground',
                                            'ekf': 'enableKeyboardFocus',
                                            'en': 'enable',
                                            'ex': 'exists',
                                            'exa': 'executeAll',
                                            'exc': 'execute',
                                            'fcc': 'fileChangedCommand',
                                            'fkp': 'filterKeyPress',
                                            'fn': 'filename',
                                            'fpn': 'fullPathName',
                                            'h': 'height',
                                            'hf': 'hasFocus',
                                            'hlc': 'highlightColor',
                                            'hsl': 'hasSelection',
                                            'io': 'isObscured',
                                            'isl': 'indentSelection',
                                            'it': 'insertText',
                                            'ld': 'load',
                                            'ldc': 'loadContents',
                                            'ldf': 'loadFile',
                                            'm': 'manage',
                                            'mcc': 'modificationChangedCommand',
                                            'mod': 'modified',
                                            'nbg': 'noBackground',
                                            'nl': 'numberOfLines',
                                            'npm': 'numberOfPopupMenus',
                                            'opc': 'objectPathCompletion',
                                            'p': 'parent',
                                            'pma': 'popupMenuArray',
                                            'po': 'preventOverride',
                                            'pst': 'pasteSelection',
                                            'q': 'query',
                                            'rd': 'redo',
                                            'rfc': 'receiveFocusCommand',
                                            'rpa': 'replaceAll',
                                            'rsc': 'removeStoredContents',
                                            'sas': 'searchAndSelect',
                                            'sbm': 'statusBarMessage',
                                            'sd': 'searchDown',
                                            'sl': 'select',
                                            'sla': 'selectAll',
                                            'sln': 'showLineNumbers',
                                            'slt': 'selectedText',
                                            'smc': 'searchMatchCase',
                                            'spt': 'spacesPerTab',
                                            'src': 'source',
                                            'ss': 'searchString',
                                            'st': 'sourceType',
                                            'stc': 'storeContents',
                                            'sth': 'showTooltipHelp',
                                            'sv': 'saveSelection',
                                            'svf': 'saveFile',
                                            'svs': 'saveSelectionToShelf',
                                            'sw': 'searchWraps',
                                            't': 'text',
                                            'tfi': 'tabsForIndent',
                                            'tl': 'textLength',
                                            'ud': 'undo',
                                            'usl': 'unindentSelection',
                                            'ut': 'useTemplate',
                                            'vcc': 'visibleChangeCommand',
                                            'vis': 'visible',
                                            'w': 'width'},
                             'type': 'windows'},
  'cmdScrollFieldReporter': {'callbackFlags': ['visibleChangeCommand',
                                               'vcc',
                                               'dragCallback',
                                               'dgc',
                                               'receiveFocusCommand',
                                               'rfc',
                                               'echoAllCommands',
                                               'eac',
                                               'dropCallback',
                                               'dpc'],
                             'flags': {'annotation': {'args': '<type __builtin__.str>',
                                                      'longname': 'annotation',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ann'},
                                       'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                    '<type __builtin__.float>',
                                                                    '<type __builtin__.float>'],
                                                           'longname': 'backgroundColor',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 3,
                                                           'shortname': 'bgc'},
                                       'clear': {'args': '<type __builtin__.bool>',
                                                 'longname': 'clear',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'clr'},
                                       'copySelection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'copySelection',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'cp'},
                                       'cutSelection': {'args': '<type __builtin__.bool>',
                                                        'longname': 'cutSelection',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'ct'},
                                       'defineTemplate': {'args': '<type __builtin__.str>',
                                                          'longname': 'defineTemplate',
                                                          'modes': ['create'],
                                                          'numArgs': 1,
                                                          'shortname': 'dt'},
                                       'docTag': {'args': '<type __builtin__.str>',
                                                  'longname': 'docTag',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dtg'},
                                       'dragCallback': {'args': 'script',
                                                        'longname': 'dragCallback',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'dgc'},
                                       'dropCallback': {'args': 'script',
                                                        'longname': 'dropCallback',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'dpc'},
                                       'echoAllCommands': {'args': '<type __builtin__.bool>',
                                                           'longname': 'echoAllCommands',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'eac'},
                                       'edit': {'args': '<type __builtin__.bool>',
                                                'longname': 'edit',
                                                'numArgs': 0,
                                                'shortname': 'e'},
                                       'enable': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enable',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'en'},
                                       'enableBackground': {'args': '<type __builtin__.bool>',
                                                            'longname': 'enableBackground',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ebg'},
                                       'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                               'longname': 'enableKeyboardFocus',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'ekf'},
                                       'exists': {'args': '<type __builtin__.bool>',
                                                  'longname': 'exists',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'ex'},
                                       'filterSourceType': {'args': '<type __builtin__.str>',
                                                            'longname': 'filterSourceType',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'fst'},
                                       'fullPathName': {'args': '<type __builtin__.bool>',
                                                        'longname': 'fullPathName',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'fpn'},
                                       'hasFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'hasFocus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'hf'},
                                       'height': {'args': '<type __builtin__.int>',
                                                  'longname': 'height',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'h'},
                                       'highlightColor': {'args': ['<type __builtin__.float>',
                                                                   '<type __builtin__.float>',
                                                                   '<type __builtin__.float>'],
                                                          'longname': 'highlightColor',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 3,
                                                          'shortname': 'hlc'},
                                       'isObscured': {'args': '<type __builtin__.bool>',
                                                      'longname': 'isObscured',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'io'},
                                       'lineNumbers': {'args': '<type __builtin__.bool>',
                                                       'longname': 'lineNumbers',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ln'},
                                       'manage': {'args': '<type __builtin__.bool>',
                                                  'longname': 'manage',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'm'},
                                       'noBackground': {'args': '<type __builtin__.bool>',
                                                        'longname': 'noBackground',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'nbg'},
                                       'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                              'longname': 'numberOfPopupMenus',
                                                              'modes': ['query'],
                                                              'numArgs': 0,
                                                              'shortname': 'npm'},
                                       'parent': {'args': '<type __builtin__.str>',
                                                  'longname': 'parent',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'p'},
                                       'pasteSelection': {'args': '<type __builtin__.bool>',
                                                          'longname': 'pasteSelection',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 0,
                                                          'shortname': 'pst'},
                                       'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                          'longname': 'popupMenuArray',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'pma'},
                                       'preventOverride': {'args': '<type __builtin__.bool>',
                                                           'longname': 'preventOverride',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'po'},
                                       'query': {'args': '<type __builtin__.bool>',
                                                 'longname': 'query',
                                                 'numArgs': 0,
                                                 'shortname': 'q'},
                                       'receiveFocusCommand': {'args': 'script',
                                                               'longname': 'receiveFocusCommand',
                                                               'modes': ['create',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'rfc'},
                                       'saveSelection': {'args': '<type __builtin__.str>',
                                                         'longname': 'saveSelection',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'sv'},
                                       'saveSelectionToShelf': {'args': '<type __builtin__.bool>',
                                                                'longname': 'saveSelectionToShelf',
                                                                'modes': ['create',
                                                                          'edit'],
                                                                'numArgs': 0,
                                                                'shortname': 'svs'},
                                       'select': {'args': ['<type __builtin__.int>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'select',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 2,
                                                  'shortname': 'sl'},
                                       'selectAll': {'args': '<type __builtin__.bool>',
                                                     'longname': 'selectAll',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'sla'},
                                       'stackTrace': {'args': '<type __builtin__.bool>',
                                                      'longname': 'stackTrace',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'st'},
                                       'statusBarMessage': {'args': '<type __builtin__.str>',
                                                            'longname': 'statusBarMessage',
                                                            'modes': ['create',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'sbm'},
                                       'suppressErrors': {'args': '<type __builtin__.bool>',
                                                          'longname': 'suppressErrors',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'se'},
                                       'suppressInfo': {'args': '<type __builtin__.bool>',
                                                        'longname': 'suppressInfo',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'si'},
                                       'suppressResults': {'args': '<type __builtin__.bool>',
                                                           'longname': 'suppressResults',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'sr'},
                                       'suppressStackTrace': {'args': '<type __builtin__.bool>',
                                                              'longname': 'suppressStackTrace',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'sst'},
                                       'suppressWarnings': {'args': '<type __builtin__.bool>',
                                                            'longname': 'suppressWarnings',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'sw'},
                                       'text': {'args': '<type __builtin__.str>',
                                                'longname': 'text',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 't'},
                                       'textLength': {'args': '<type __builtin__.bool>',
                                                      'longname': 'textLength',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'tl'},
                                       'useTemplate': {'args': '<type __builtin__.str>',
                                                       'longname': 'useTemplate',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ut'},
                                       'visible': {'args': '<type __builtin__.bool>',
                                                   'longname': 'visible',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'vis'},
                                       'visibleChangeCommand': {'args': 'script',
                                                                'longname': 'visibleChangeCommand',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'vcc'},
                                       'width': {'args': '<type __builtin__.int>',
                                                 'longname': 'width',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'w'}},
                             'shortFlags': {'ann': 'annotation',
                                            'bgc': 'backgroundColor',
                                            'clr': 'clear',
                                            'cp': 'copySelection',
                                            'ct': 'cutSelection',
                                            'dgc': 'dragCallback',
                                            'dpc': 'dropCallback',
                                            'dt': 'defineTemplate',
                                            'dtg': 'docTag',
                                            'e': 'edit',
                                            'eac': 'echoAllCommands',
                                            'ebg': 'enableBackground',
                                            'ekf': 'enableKeyboardFocus',
                                            'en': 'enable',
                                            'ex': 'exists',
                                            'fpn': 'fullPathName',
                                            'fst': 'filterSourceType',
                                            'h': 'height',
                                            'hf': 'hasFocus',
                                            'hlc': 'highlightColor',
                                            'io': 'isObscured',
                                            'ln': 'lineNumbers',
                                            'm': 'manage',
                                            'nbg': 'noBackground',
                                            'npm': 'numberOfPopupMenus',
                                            'p': 'parent',
                                            'pma': 'popupMenuArray',
                                            'po': 'preventOverride',
                                            'pst': 'pasteSelection',
                                            'q': 'query',
                                            'rfc': 'receiveFocusCommand',
                                            'sbm': 'statusBarMessage',
                                            'se': 'suppressErrors',
                                            'si': 'suppressInfo',
                                            'sl': 'select',
                                            'sla': 'selectAll',
                                            'sr': 'suppressResults',
                                            'sst': 'suppressStackTrace',
                                            'st': 'stackTrace',
                                            'sv': 'saveSelection',
                                            'svs': 'saveSelectionToShelf',
                                            'sw': 'suppressWarnings',
                                            't': 'text',
                                            'tl': 'textLength',
                                            'ut': 'useTemplate',
                                            'vcc': 'visibleChangeCommand',
                                            'vis': 'visible',
                                            'w': 'width'},
                             'type': 'windows'},
  'cmdShell': {'callbackFlags': ['visibleChangeCommand',
                                 'vcc',
                                 'dragCallback',
                                 'dgc',
                                 'command',
                                 'c',
                                 'dropCallback',
                                 'dpc'],
               'flags': {'annotation': {'args': '<type __builtin__.str>',
                                        'longname': 'annotation',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ann'},
                         'backgroundColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'backgroundColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'bgc'},
                         'clear': {'args': '<type __builtin__.bool>',
                                   'longname': 'clear',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'cl'},
                         'command': {'args': '<type __builtin__.str>',
                                     'longname': 'command',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'c'},
                         'defineTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'defineTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'dt'},
                         'docTag': {'args': '<type __builtin__.str>',
                                    'longname': 'docTag',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'dtg'},
                         'dragCallback': {'args': 'script',
                                          'longname': 'dragCallback',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dgc'},
                         'dropCallback': {'args': 'script',
                                          'longname': 'dropCallback',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dpc'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'enable': {'args': '<type __builtin__.bool>',
                                    'longname': 'enable',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'en'},
                         'enableBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'enableBackground',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ebg'},
                         'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableKeyboardFocus',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ekf'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'fullPathName': {'args': '<type __builtin__.bool>',
                                          'longname': 'fullPathName',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'fpn'},
                         'height': {'args': '<type __builtin__.int>',
                                    'longname': 'height',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'h'},
                         'highlightColor': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'highlightColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'hlc'},
                         'isObscured': {'args': '<type __builtin__.bool>',
                                        'longname': 'isObscured',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'io'},
                         'manage': {'args': '<type __builtin__.bool>',
                                    'longname': 'manage',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'm'},
                         'noBackground': {'args': '<type __builtin__.bool>',
                                          'longname': 'noBackground',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nbg'},
                         'numberOfHistoryLines': {'args': '<type __builtin__.int>',
                                                  'longname': 'numberOfHistoryLines',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'nhl'},
                         'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                'longname': 'numberOfPopupMenus',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'npm'},
                         'numberOfSavedLines': {'args': '<type __builtin__.int>',
                                                'longname': 'numberOfSavedLines',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nsl'},
                         'parent': {'args': '<type __builtin__.str>',
                                    'longname': 'parent',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'p'},
                         'popupMenuArray': {'args': '<type __builtin__.bool>',
                                            'longname': 'popupMenuArray',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'pma'},
                         'preventOverride': {'args': '<type __builtin__.bool>',
                                             'longname': 'preventOverride',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'po'},
                         'prompt': {'args': '<type __builtin__.str>',
                                    'longname': 'prompt',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'pr'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'statusBarMessage': {'args': '<type __builtin__.str>',
                                              'longname': 'statusBarMessage',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sbm'},
                         'useTemplate': {'args': '<type __builtin__.str>',
                                         'longname': 'useTemplate',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ut'},
                         'visible': {'args': '<type __builtin__.bool>',
                                     'longname': 'visible',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'vis'},
                         'visibleChangeCommand': {'args': 'script',
                                                  'longname': 'visibleChangeCommand',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'vcc'},
                         'width': {'args': '<type __builtin__.int>',
                                   'longname': 'width',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'w'}},
               'shortFlags': {'ann': 'annotation',
                              'bgc': 'backgroundColor',
                              'c': 'command',
                              'cl': 'clear',
                              'dgc': 'dragCallback',
                              'dpc': 'dropCallback',
                              'dt': 'defineTemplate',
                              'dtg': 'docTag',
                              'e': 'edit',
                              'ebg': 'enableBackground',
                              'ekf': 'enableKeyboardFocus',
                              'en': 'enable',
                              'ex': 'exists',
                              'fpn': 'fullPathName',
                              'h': 'height',
                              'hlc': 'highlightColor',
                              'io': 'isObscured',
                              'm': 'manage',
                              'nbg': 'noBackground',
                              'nhl': 'numberOfHistoryLines',
                              'npm': 'numberOfPopupMenus',
                              'nsl': 'numberOfSavedLines',
                              'p': 'parent',
                              'pma': 'popupMenuArray',
                              'po': 'preventOverride',
                              'pr': 'prompt',
                              'q': 'query',
                              'sbm': 'statusBarMessage',
                              'ut': 'useTemplate',
                              'vcc': 'visibleChangeCommand',
                              'vis': 'visible',
                              'w': 'width'},
               'type': 'windows'},
  'coarsenSubdivSelectionList': {'flags': {},
                                 'shortFlags': {},
                                 'type': 'modeling'},
  'collision': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'friction': {'args': '<type __builtin__.float>',
                                       'longname': 'friction',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'f'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'offset': {'args': '<type __builtin__.float>',
                                     'longname': 'offset',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'o'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'resilience': {'args': '<type __builtin__.float>',
                                         'longname': 'resilience',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'r'}},
                'shortFlags': {'e': 'edit',
                               'f': 'friction',
                               'n': 'name',
                               'o': 'offset',
                               'q': 'query',
                               'r': 'resilience'},
                'type': 'effects'},
  'color': {'flags': {'rgbColor': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'rgbColor',
                                   'modes': ['create'],
                                   'numArgs': 3,
                                   'shortname': 'rgb'},
                      'userDefined': {'args': '<type __builtin__.int>',
                                      'longname': 'userDefined',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'ud'}},
            'shortFlags': {'rgb': 'rgbColor', 'ud': 'userDefined'},
            'type': 'general'},
  'colorAtPoint': {'flags': {'coordU': {'args': '<type __builtin__.float>',
                                        'longname': 'coordU',
                                        'modes': ['create', 'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'u'},
                             'coordV': {'args': '<type __builtin__.float>',
                                        'longname': 'coordV',
                                        'modes': ['create', 'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'v'},
                             'maxU': {'args': '<type __builtin__.float>',
                                      'longname': 'maxU',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'xu'},
                             'maxV': {'args': '<type __builtin__.float>',
                                      'longname': 'maxV',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'xv'},
                             'minU': {'args': '<type __builtin__.float>',
                                      'longname': 'minU',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'mu'},
                             'minV': {'args': '<type __builtin__.float>',
                                      'longname': 'minV',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'mv'},
                             'output': {'args': '<type __builtin__.str>',
                                        'longname': 'output',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                             'samplesU': {'args': '<type __builtin__.int>',
                                          'longname': 'samplesU',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'su'},
                             'samplesV': {'args': '<type __builtin__.int>',
                                          'longname': 'samplesV',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'sv'}},
                   'shortFlags': {'mu': 'minU',
                                  'mv': 'minV',
                                  'o': 'output',
                                  'su': 'samplesU',
                                  'sv': 'samplesV',
                                  'u': 'coordU',
                                  'v': 'coordV',
                                  'xu': 'maxU',
                                  'xv': 'maxV'},
                   'type': 'effects'},
  'colorEditor': {'flags': {'alpha': {'args': '<type __builtin__.float>',
                                      'longname': 'alpha',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'a'},
                            'hsvValue': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'hsvValue',
                                         'modes': ['create', 'query'],
                                         'numArgs': 3,
                                         'shortname': 'hsv'},
                            'mini': {'args': '<type __builtin__.bool>',
                                     'longname': 'mini',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'm'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'position': {'args': ['<type __builtin__.int>',
                                                  '<type __builtin__.int>'],
                                         'longname': 'position',
                                         'modes': ['create'],
                                         'numArgs': 2,
                                         'shortname': 'pos'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'result': {'args': '<type __builtin__.bool>',
                                       'longname': 'result',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'r'},
                            'rgbValue': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'rgbValue',
                                         'modes': ['create', 'query'],
                                         'numArgs': 3,
                                         'shortname': 'rgb'}},
                  'shortFlags': {'a': 'alpha',
                                 'hsv': 'hsvValue',
                                 'm': 'mini',
                                 'p': 'parent',
                                 'pos': 'position',
                                 'q': 'query',
                                 'r': 'result',
                                 'rgb': 'rgbValue'},
                  'type': 'windows'},
  'colorIndex': {'flags': {'hueSaturationValue': {'args': '<type __builtin__.bool>',
                                                  'longname': 'hueSaturationValue',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 0,
                                                  'shortname': 'hsv'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'resetToFactory': {'args': '<type __builtin__.bool>',
                                              'longname': 'resetToFactory',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'rf'},
                           'resetToSaved': {'args': '<type __builtin__.bool>',
                                            'longname': 'resetToSaved',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'rs'}},
                 'shortFlags': {'hsv': 'hueSaturationValue',
                                'q': 'query',
                                'rf': 'resetToFactory',
                                'rs': 'resetToSaved'},
                 'type': 'general'},
  'colorIndexSliderGrp': {'callbackFlags': ['dragCommand',
                                            'dc',
                                            'visibleChangeCommand',
                                            'vcc',
                                            'dragCallback',
                                            'dgc',
                                            'changeCommand',
                                            'cc',
                                            'dropCallback',
                                            'dpc'],
                          'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'adj'},
                                    'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                          'longname': 'adjustableColumn2',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ad2'},
                                    'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                          'longname': 'adjustableColumn3',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ad3'},
                                    'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                          'longname': 'adjustableColumn4',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ad4'},
                                    'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                          'longname': 'adjustableColumn5',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ad5'},
                                    'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                          'longname': 'adjustableColumn6',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ad6'},
                                    'annotation': {'args': '<type __builtin__.str>',
                                                   'longname': 'annotation',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ann'},
                                    'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                 '<type __builtin__.float>',
                                                                 '<type __builtin__.float>'],
                                                        'longname': 'backgroundColor',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 3,
                                                        'shortname': 'bgc'},
                                    'changeCommand': {'args': 'script',
                                                      'longname': 'changeCommand',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cc'},
                                    'columnAlign': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign',
                                                    'modes': ['create',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 2,
                                                    'shortname': 'cal'},
                                    'columnAlign2': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAlign2',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'cl2'},
                                    'columnAlign3': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAlign3',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'cl3'},
                                    'columnAlign4': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAlign4',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 4,
                                                     'shortname': 'cl4'},
                                    'columnAlign5': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAlign5',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 5,
                                                     'shortname': 'cl5'},
                                    'columnAlign6': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAlign6',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 6,
                                                     'shortname': 'cl6'},
                                    'columnAttach': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnAttach',
                                                     'modes': ['create',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 3,
                                                     'shortname': 'cat'},
                                    'columnAttach2': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAttach2',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'ct2'},
                                    'columnAttach3': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAttach3',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'ct3'},
                                    'columnAttach4': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAttach4',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 4,
                                                      'shortname': 'ct4'},
                                    'columnAttach5': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAttach5',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 5,
                                                      'shortname': 'ct5'},
                                    'columnAttach6': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAttach6',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 6,
                                                      'shortname': 'ct6'},
                                    'columnOffset2': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnOffset2',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'co2'},
                                    'columnOffset3': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnOffset3',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'co3'},
                                    'columnOffset4': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnOffset4',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 4,
                                                      'shortname': 'co4'},
                                    'columnOffset5': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnOffset5',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 5,
                                                      'shortname': 'co5'},
                                    'columnOffset6': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnOffset6',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 6,
                                                      'shortname': 'co6'},
                                    'columnWidth': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth',
                                                    'modes': ['create',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 2,
                                                    'shortname': 'cw'},
                                    'columnWidth1': {'args': '<type __builtin__.int>',
                                                     'longname': 'columnWidth1',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'cw1'},
                                    'columnWidth2': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnWidth2',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'cw2'},
                                    'columnWidth3': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnWidth3',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'cw3'},
                                    'columnWidth4': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnWidth4',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 4,
                                                     'shortname': 'cw4'},
                                    'columnWidth5': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnWidth5',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 5,
                                                     'shortname': 'cw5'},
                                    'columnWidth6': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnWidth6',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 6,
                                                     'shortname': 'cw6'},
                                    'defineTemplate': {'args': '<type __builtin__.str>',
                                                       'longname': 'defineTemplate',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'dt'},
                                    'docTag': {'args': '<type __builtin__.str>',
                                               'longname': 'docTag',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dtg'},
                                    'dragCallback': {'args': 'script',
                                                     'longname': 'dragCallback',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dgc'},
                                    'dragCommand': {'args': 'script',
                                                    'longname': 'dragCommand',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dc'},
                                    'dropCallback': {'args': 'script',
                                                     'longname': 'dropCallback',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dpc'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'enable': {'args': '<type __builtin__.bool>',
                                               'longname': 'enable',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'en'},
                                    'enableBackground': {'args': '<type __builtin__.bool>',
                                                         'longname': 'enableBackground',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ebg'},
                                    'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                            'longname': 'enableKeyboardFocus',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ekf'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'extraLabel': {'args': '<type __builtin__.str>',
                                                   'longname': 'extraLabel',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'el'},
                                    'forceDragRefresh': {'args': '<type __builtin__.bool>',
                                                         'longname': 'forceDragRefresh',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'fdr'},
                                    'fullPathName': {'args': '<type __builtin__.bool>',
                                                     'longname': 'fullPathName',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'fpn'},
                                    'height': {'args': '<type __builtin__.int>',
                                               'longname': 'height',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'h'},
                                    'highlightColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'highlightColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'hlc'},
                                    'invisible': {'args': '<type __builtin__.int>',
                                                  'longname': 'invisible',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'inv'},
                                    'isObscured': {'args': '<type __builtin__.bool>',
                                                   'longname': 'isObscured',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'io'},
                                    'label': {'args': '<type __builtin__.str>',
                                              'longname': 'label',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'l'},
                                    'manage': {'args': '<type __builtin__.bool>',
                                               'longname': 'manage',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'm'},
                                    'maxValue': {'args': '<type __builtin__.int>',
                                                 'longname': 'maxValue',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'max'},
                                    'minValue': {'args': '<type __builtin__.int>',
                                                 'longname': 'minValue',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'min'},
                                    'noBackground': {'args': '<type __builtin__.bool>',
                                                     'longname': 'noBackground',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'nbg'},
                                    'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                           'longname': 'numberOfPopupMenus',
                                                           'modes': ['query'],
                                                           'numArgs': 0,
                                                           'shortname': 'npm'},
                                    'parent': {'args': '<type __builtin__.str>',
                                               'longname': 'parent',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'p'},
                                    'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                       'longname': 'popupMenuArray',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'pma'},
                                    'preventOverride': {'args': '<type __builtin__.bool>',
                                                        'longname': 'preventOverride',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'po'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'rowAttach': {'args': ['<type __builtin__.int>',
                                                           '<type __builtin__.str>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'rowAttach',
                                                  'modes': ['create',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 3,
                                                  'shortname': 'rat'},
                                    'statusBarMessage': {'args': '<type __builtin__.str>',
                                                         'longname': 'statusBarMessage',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'sbm'},
                                    'useTemplate': {'args': '<type __builtin__.str>',
                                                    'longname': 'useTemplate',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ut'},
                                    'value': {'args': '<type __builtin__.int>',
                                              'longname': 'value',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'v'},
                                    'visible': {'args': '<type __builtin__.bool>',
                                                'longname': 'visible',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'vis'},
                                    'visibleChangeCommand': {'args': 'script',
                                                             'longname': 'visibleChangeCommand',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'vcc'},
                                    'width': {'args': '<type __builtin__.int>',
                                              'longname': 'width',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'w'}},
                          'shortFlags': {'ad2': 'adjustableColumn2',
                                         'ad3': 'adjustableColumn3',
                                         'ad4': 'adjustableColumn4',
                                         'ad5': 'adjustableColumn5',
                                         'ad6': 'adjustableColumn6',
                                         'adj': 'adjustableColumn',
                                         'ann': 'annotation',
                                         'bgc': 'backgroundColor',
                                         'cal': 'columnAlign',
                                         'cat': 'columnAttach',
                                         'cc': 'changeCommand',
                                         'cl2': 'columnAlign2',
                                         'cl3': 'columnAlign3',
                                         'cl4': 'columnAlign4',
                                         'cl5': 'columnAlign5',
                                         'cl6': 'columnAlign6',
                                         'co2': 'columnOffset2',
                                         'co3': 'columnOffset3',
                                         'co4': 'columnOffset4',
                                         'co5': 'columnOffset5',
                                         'co6': 'columnOffset6',
                                         'ct2': 'columnAttach2',
                                         'ct3': 'columnAttach3',
                                         'ct4': 'columnAttach4',
                                         'ct5': 'columnAttach5',
                                         'ct6': 'columnAttach6',
                                         'cw': 'columnWidth',
                                         'cw1': 'columnWidth1',
                                         'cw2': 'columnWidth2',
                                         'cw3': 'columnWidth3',
                                         'cw4': 'columnWidth4',
                                         'cw5': 'columnWidth5',
                                         'cw6': 'columnWidth6',
                                         'dc': 'dragCommand',
                                         'dgc': 'dragCallback',
                                         'dpc': 'dropCallback',
                                         'dt': 'defineTemplate',
                                         'dtg': 'docTag',
                                         'e': 'edit',
                                         'ebg': 'enableBackground',
                                         'ekf': 'enableKeyboardFocus',
                                         'el': 'extraLabel',
                                         'en': 'enable',
                                         'ex': 'exists',
                                         'fdr': 'forceDragRefresh',
                                         'fpn': 'fullPathName',
                                         'h': 'height',
                                         'hlc': 'highlightColor',
                                         'inv': 'invisible',
                                         'io': 'isObscured',
                                         'l': 'label',
                                         'm': 'manage',
                                         'max': 'maxValue',
                                         'min': 'minValue',
                                         'nbg': 'noBackground',
                                         'npm': 'numberOfPopupMenus',
                                         'p': 'parent',
                                         'pma': 'popupMenuArray',
                                         'po': 'preventOverride',
                                         'q': 'query',
                                         'rat': 'rowAttach',
                                         'sbm': 'statusBarMessage',
                                         'ut': 'useTemplate',
                                         'v': 'value',
                                         'vcc': 'visibleChangeCommand',
                                         'vis': 'visible',
                                         'w': 'width'},
                          'type': 'windows'},
  'colorInputWidgetGrp': {'callbackFlags': ['dropCallback',
                                            'dpc',
                                            'visibleChangeCommand',
                                            'vcc',
                                            'dragCallback',
                                            'dgc',
                                            'changeCommand',
                                            'cc'],
                          'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'adj'},
                                    'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                          'longname': 'adjustableColumn2',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ad2'},
                                    'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                          'longname': 'adjustableColumn3',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ad3'},
                                    'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                          'longname': 'adjustableColumn4',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ad4'},
                                    'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                          'longname': 'adjustableColumn5',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ad5'},
                                    'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                          'longname': 'adjustableColumn6',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ad6'},
                                    'annotation': {'args': '<type __builtin__.str>',
                                                   'longname': 'annotation',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ann'},
                                    'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                 '<type __builtin__.float>',
                                                                 '<type __builtin__.float>'],
                                                        'longname': 'backgroundColor',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 3,
                                                        'shortname': 'bgc'},
                                    'changeCommand': {'args': 'script',
                                                      'longname': 'changeCommand',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cc'},
                                    'columnAlign': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign',
                                                    'modes': ['create',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 2,
                                                    'shortname': 'cal'},
                                    'columnAlign2': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAlign2',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'cl2'},
                                    'columnAlign3': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAlign3',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'cl3'},
                                    'columnAlign4': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAlign4',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 4,
                                                     'shortname': 'cl4'},
                                    'columnAlign5': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAlign5',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 5,
                                                     'shortname': 'cl5'},
                                    'columnAlign6': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAlign6',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 6,
                                                     'shortname': 'cl6'},
                                    'columnAttach': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnAttach',
                                                     'modes': ['create',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 3,
                                                     'shortname': 'cat'},
                                    'columnAttach2': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAttach2',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'ct2'},
                                    'columnAttach3': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAttach3',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'ct3'},
                                    'columnAttach4': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAttach4',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 4,
                                                      'shortname': 'ct4'},
                                    'columnAttach5': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAttach5',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 5,
                                                      'shortname': 'ct5'},
                                    'columnAttach6': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAttach6',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 6,
                                                      'shortname': 'ct6'},
                                    'columnOffset2': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnOffset2',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'co2'},
                                    'columnOffset3': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnOffset3',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'co3'},
                                    'columnOffset4': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnOffset4',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 4,
                                                      'shortname': 'co4'},
                                    'columnOffset5': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnOffset5',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 5,
                                                      'shortname': 'co5'},
                                    'columnOffset6': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnOffset6',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 6,
                                                      'shortname': 'co6'},
                                    'columnWidth': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth',
                                                    'modes': ['create',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 2,
                                                    'shortname': 'cw'},
                                    'columnWidth1': {'args': '<type __builtin__.int>',
                                                     'longname': 'columnWidth1',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'cw1'},
                                    'columnWidth2': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnWidth2',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'cw2'},
                                    'columnWidth3': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnWidth3',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'cw3'},
                                    'columnWidth4': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnWidth4',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 4,
                                                     'shortname': 'cw4'},
                                    'columnWidth5': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnWidth5',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 5,
                                                     'shortname': 'cw5'},
                                    'columnWidth6': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnWidth6',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 6,
                                                     'shortname': 'cw6'},
                                    'defineTemplate': {'args': '<type __builtin__.str>',
                                                       'longname': 'defineTemplate',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'dt'},
                                    'docTag': {'args': '<type __builtin__.str>',
                                               'longname': 'docTag',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dtg'},
                                    'dragCallback': {'args': 'script',
                                                     'longname': 'dragCallback',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dgc'},
                                    'dropCallback': {'args': 'script',
                                                     'longname': 'dropCallback',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dpc'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'enable': {'args': '<type __builtin__.bool>',
                                               'longname': 'enable',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'en'},
                                    'enableBackground': {'args': '<type __builtin__.bool>',
                                                         'longname': 'enableBackground',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ebg'},
                                    'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                            'longname': 'enableKeyboardFocus',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ekf'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'forceDragRefresh': {'args': '<type __builtin__.bool>',
                                                         'longname': 'forceDragRefresh',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'fdr'},
                                    'fullPathName': {'args': '<type __builtin__.bool>',
                                                     'longname': 'fullPathName',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'fpn'},
                                    'height': {'args': '<type __builtin__.int>',
                                               'longname': 'height',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'h'},
                                    'highlightColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'highlightColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'hlc'},
                                    'hsvValue': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'hsvValue',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'hsv'},
                                    'isObscured': {'args': '<type __builtin__.bool>',
                                                   'longname': 'isObscured',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'io'},
                                    'label': {'args': '<type __builtin__.str>',
                                              'longname': 'label',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'l'},
                                    'manage': {'args': '<type __builtin__.bool>',
                                               'longname': 'manage',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'm'},
                                    'noBackground': {'args': '<type __builtin__.bool>',
                                                     'longname': 'noBackground',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'nbg'},
                                    'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                           'longname': 'numberOfPopupMenus',
                                                           'modes': ['query'],
                                                           'numArgs': 0,
                                                           'shortname': 'npm'},
                                    'parent': {'args': '<type __builtin__.str>',
                                               'longname': 'parent',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'p'},
                                    'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                       'longname': 'popupMenuArray',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'pma'},
                                    'preventOverride': {'args': '<type __builtin__.bool>',
                                                        'longname': 'preventOverride',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'po'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'rgbValue': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'rgbValue',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'rgb'},
                                    'rowAttach': {'args': ['<type __builtin__.int>',
                                                           '<type __builtin__.str>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'rowAttach',
                                                  'modes': ['create',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 3,
                                                  'shortname': 'rat'},
                                    'statusBarMessage': {'args': '<type __builtin__.str>',
                                                         'longname': 'statusBarMessage',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'sbm'},
                                    'useTemplate': {'args': '<type __builtin__.str>',
                                                    'longname': 'useTemplate',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ut'},
                                    'visible': {'args': '<type __builtin__.bool>',
                                                'longname': 'visible',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'vis'},
                                    'visibleChangeCommand': {'args': 'script',
                                                             'longname': 'visibleChangeCommand',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'vcc'},
                                    'width': {'args': '<type __builtin__.int>',
                                              'longname': 'width',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'w'}},
                          'shortFlags': {'ad2': 'adjustableColumn2',
                                         'ad3': 'adjustableColumn3',
                                         'ad4': 'adjustableColumn4',
                                         'ad5': 'adjustableColumn5',
                                         'ad6': 'adjustableColumn6',
                                         'adj': 'adjustableColumn',
                                         'ann': 'annotation',
                                         'bgc': 'backgroundColor',
                                         'cal': 'columnAlign',
                                         'cat': 'columnAttach',
                                         'cc': 'changeCommand',
                                         'cl2': 'columnAlign2',
                                         'cl3': 'columnAlign3',
                                         'cl4': 'columnAlign4',
                                         'cl5': 'columnAlign5',
                                         'cl6': 'columnAlign6',
                                         'co2': 'columnOffset2',
                                         'co3': 'columnOffset3',
                                         'co4': 'columnOffset4',
                                         'co5': 'columnOffset5',
                                         'co6': 'columnOffset6',
                                         'ct2': 'columnAttach2',
                                         'ct3': 'columnAttach3',
                                         'ct4': 'columnAttach4',
                                         'ct5': 'columnAttach5',
                                         'ct6': 'columnAttach6',
                                         'cw': 'columnWidth',
                                         'cw1': 'columnWidth1',
                                         'cw2': 'columnWidth2',
                                         'cw3': 'columnWidth3',
                                         'cw4': 'columnWidth4',
                                         'cw5': 'columnWidth5',
                                         'cw6': 'columnWidth6',
                                         'dgc': 'dragCallback',
                                         'dpc': 'dropCallback',
                                         'dt': 'defineTemplate',
                                         'dtg': 'docTag',
                                         'e': 'edit',
                                         'ebg': 'enableBackground',
                                         'ekf': 'enableKeyboardFocus',
                                         'en': 'enable',
                                         'ex': 'exists',
                                         'fpn': 'fullPathName',
                                         'h': 'height',
                                         'hlc': 'highlightColor',
                                         'hsv': 'hsvValue',
                                         'io': 'isObscured',
                                         'l': 'label',
                                         'm': 'manage',
                                         'nbg': 'noBackground',
                                         'npm': 'numberOfPopupMenus',
                                         'p': 'parent',
                                         'pma': 'popupMenuArray',
                                         'po': 'preventOverride',
                                         'q': 'query',
                                         'rat': 'rowAttach',
                                         'rgb': 'rgbValue',
                                         'sbm': 'statusBarMessage',
                                         'ut': 'useTemplate',
                                         'vcc': 'visibleChangeCommand',
                                         'vis': 'visible',
                                         'w': 'width'},
                          'type': 'windows'},
  'colorManagementCatalog': {'flags': {'addTransform': {'args': '<type __builtin__.str>',
                                                        'longname': 'addTransform',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'adt'},
                                       'editUserTransformPath': {'args': '<type __builtin__.str>',
                                                                 'longname': 'editUserTransformPath',
                                                                 'modes': ['create'],
                                                                 'numArgs': 1,
                                                                 'shortname': 'eut'},
                                       'listSupportedExtensions': {'args': '<type __builtin__.bool>',
                                                                   'longname': 'listSupportedExtensions',
                                                                   'modes': ['create'],
                                                                   'numArgs': 0,
                                                                   'shortname': 'lse'},
                                       'listTransformConnections': {'args': '<type __builtin__.bool>',
                                                                    'longname': 'listTransformConnections',
                                                                    'modes': ['create'],
                                                                    'numArgs': 0,
                                                                    'shortname': 'ltc'},
                                       'path': {'args': '<type __builtin__.str>',
                                                'longname': 'path',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'pth'},
                                       'queryUserTransformPath': {'args': '<type __builtin__.bool>',
                                                                  'longname': 'queryUserTransformPath',
                                                                  'modes': ['create'],
                                                                  'numArgs': 0,
                                                                  'shortname': 'qut'},
                                       'removeTransform': {'args': '<type __builtin__.str>',
                                                           'longname': 'removeTransform',
                                                           'modes': ['create'],
                                                           'numArgs': 1,
                                                           'shortname': 'rmt'},
                                       'transformConnection': {'args': '<type __builtin__.str>',
                                                               'longname': 'transformConnection',
                                                               'modes': ['create'],
                                                               'numArgs': 1,
                                                               'shortname': 'tcn'},
                                       'type': {'args': '<type __builtin__.str>',
                                                'longname': 'type',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'typ'}},
                             'shortFlags': {'adt': 'addTransform',
                                            'eut': 'editUserTransformPath',
                                            'lse': 'listSupportedExtensions',
                                            'ltc': 'listTransformConnections',
                                            'pth': 'path',
                                            'qut': 'queryUserTransformPath',
                                            'rmt': 'removeTransform',
                                            'tcn': 'transformConnection',
                                            'typ': 'type'},
                             'type': 'general'},
  'colorManagementConvert': {'flags': {'toDisplaySpace': {'args': ['<type __builtin__.float>',
                                                                   '<type __builtin__.float>',
                                                                   '<type __builtin__.float>'],
                                                          'longname': 'toDisplaySpace',
                                                          'modes': ['create'],
                                                          'numArgs': 3,
                                                          'shortname': 'tds'}},
                             'shortFlags': {'tds': 'toDisplaySpace'},
                             'type': 'general'},
  'colorManagementFileRules': {'flags': {'addRule': {'args': '<type __builtin__.str>',
                                                     'longname': 'addRule',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'add'},
                                         'colorSpace': {'args': '<type __builtin__.str>',
                                                        'longname': 'colorSpace',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'cs'},
                                         'down': {'args': '<type __builtin__.str>',
                                                  'longname': 'down',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dwn'},
                                         'edit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'edit',
                                                  'numArgs': 0,
                                                  'shortname': 'e'},
                                         'evaluate': {'args': '<type __builtin__.str>',
                                                      'longname': 'evaluate',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ev'},
                                         'extension': {'args': '<type __builtin__.str>',
                                                       'longname': 'extension',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ext'},
                                         'listRules': {'args': '<type __builtin__.bool>',
                                                       'longname': 'listRules',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'lsr'},
                                         'load': {'args': '<type __builtin__.bool>',
                                                  'longname': 'load',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'ld'},
                                         'moveUp': {'args': '<type __builtin__.str>',
                                                    'longname': 'moveUp',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'up'},
                                         'pattern': {'args': '<type __builtin__.str>',
                                                     'longname': 'pattern',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'pat'},
                                         'query': {'args': '<type __builtin__.bool>',
                                                   'longname': 'query',
                                                   'numArgs': 0,
                                                   'shortname': 'q'},
                                         'remove': {'args': '<type __builtin__.str>',
                                                    'longname': 'remove',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'rm'},
                                         'restoreDefaults': {'args': '<type __builtin__.bool>',
                                                             'longname': 'restoreDefaults',
                                                             'modes': ['create',
                                                                       'edit'],
                                                             'numArgs': 0,
                                                             'shortname': 'rde'},
                                         'save': {'args': '<type __builtin__.bool>',
                                                  'longname': 'save',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'sav'}},
                               'shortFlags': {'add': 'addRule',
                                              'cs': 'colorSpace',
                                              'dwn': 'down',
                                              'e': 'edit',
                                              'ev': 'evaluate',
                                              'ext': 'extension',
                                              'ld': 'load',
                                              'lsr': 'listRules',
                                              'pat': 'pattern',
                                              'q': 'query',
                                              'rde': 'restoreDefaults',
                                              'rm': 'remove',
                                              'sav': 'save',
                                              'up': 'moveUp'},
                               'type': 'general'},
  'colorManagementPrefs': {'flags': {'cmConfigFileEnabled': {'args': '<type __builtin__.bool>',
                                                             'longname': 'cmConfigFileEnabled',
                                                             'modes': ['query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'cfe'},
                                     'cmEnabled': {'args': '<type __builtin__.bool>',
                                                   'longname': 'cmEnabled',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cme'},
                                     'colorManageAllNodes': {'args': '<type __builtin__.bool>',
                                                             'longname': 'colorManageAllNodes',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 0,
                                                             'shortname': 'cma'},
                                     'colorManagePots': {'args': '<type __builtin__.bool>',
                                                         'longname': 'colorManagePots',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'cmp'},
                                     'colorManagedNodes': {'args': '<type __builtin__.bool>',
                                                           'longname': 'colorManagedNodes',
                                                           'modes': ['query',
                                                                     'edit'],
                                                           'numArgs': 0,
                                                           'shortname': 'cmn'},
                                     'colorManagementSDKVersion': {'args': '<type __builtin__.str>',
                                                                   'longname': 'colorManagementSDKVersion',
                                                                   'modes': ['query',
                                                                             'edit'],
                                                                   'numArgs': 1,
                                                                   'shortname': 'cmv'},
                                     'configFilePath': {'args': '<type __builtin__.str>',
                                                        'longname': 'configFilePath',
                                                        'modes': ['query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'cfp'},
                                     'defaultInputSpaceName': {'args': '<type __builtin__.str>',
                                                               'longname': 'defaultInputSpaceName',
                                                               'modes': ['query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'din'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'equalsToPolicyFile': {'args': '<type __builtin__.str>',
                                                            'longname': 'equalsToPolicyFile',
                                                            'modes': ['query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'etp'},
                                     'exportPolicy': {'args': '<type __builtin__.str>',
                                                      'longname': 'exportPolicy',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'epy'},
                                     'inhibitEvents': {'args': '<type __builtin__.bool>',
                                                       'longname': 'inhibitEvents',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'ie'},
                                     'inputSpaceNames': {'args': '<type __builtin__.bool>',
                                                         'longname': 'inputSpaceNames',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'iss'},
                                     'loadPolicy': {'args': '<type __builtin__.str>',
                                                    'longname': 'loadPolicy',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'lpy'},
                                     'loadedDefaultInputSpaceName': {'args': '<type __builtin__.str>',
                                                                     'longname': 'loadedDefaultInputSpaceName',
                                                                     'modes': ['query',
                                                                               'edit'],
                                                                     'numArgs': 1,
                                                                     'shortname': 'ldn'},
                                     'loadedOutputTransformName': {'args': '<type __builtin__.str>',
                                                                   'longname': 'loadedOutputTransformName',
                                                                   'modes': ['query',
                                                                             'edit'],
                                                                   'numArgs': 1,
                                                                   'shortname': 'lon'},
                                     'loadedRenderingSpaceName': {'args': '<type __builtin__.str>',
                                                                  'longname': 'loadedRenderingSpaceName',
                                                                  'modes': ['query',
                                                                            'edit'],
                                                                  'numArgs': 1,
                                                                  'shortname': 'lrn'},
                                     'loadedViewTransformName': {'args': '<type __builtin__.str>',
                                                                 'longname': 'loadedViewTransformName',
                                                                 'modes': ['query',
                                                                           'edit'],
                                                                 'numArgs': 1,
                                                                 'shortname': 'lvn'},
                                     'missingColorSpaceNodes': {'args': '<type __builtin__.bool>',
                                                                'longname': 'missingColorSpaceNodes',
                                                                'modes': ['query',
                                                                          'edit'],
                                                                'numArgs': 0,
                                                                'shortname': 'mcn'},
                                     'ocioRulesEnabled': {'args': '<type __builtin__.bool>',
                                                          'longname': 'ocioRulesEnabled',
                                                          'modes': ['query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ore'},
                                     'outputTarget': {'args': '<type __builtin__.str>',
                                                      'longname': 'outputTarget',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ott'},
                                     'outputTransformEnabled': {'args': '<type __builtin__.bool>',
                                                                'longname': 'outputTransformEnabled',
                                                                'modes': ['query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'ote'},
                                     'outputTransformName': {'args': '<type __builtin__.str>',
                                                             'longname': 'outputTransformName',
                                                             'modes': ['query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'otn'},
                                     'outputTransformNames': {'args': '<type __builtin__.bool>',
                                                              'longname': 'outputTransformNames',
                                                              'modes': ['query',
                                                                        'edit'],
                                                              'numArgs': 0,
                                                              'shortname': 'ots'},
                                     'outputTransformUseColorConversion': {'args': '<type __builtin__.bool>',
                                                                           'longname': 'outputTransformUseColorConversion',
                                                                           'modes': ['query',
                                                                                     'edit'],
                                                                           'numArgs': 1,
                                                                           'shortname': 'otc'},
                                     'outputUseViewTransform': {'args': '<type __builtin__.bool>',
                                                                'longname': 'outputUseViewTransform',
                                                                'modes': ['query',
                                                                          'edit'],
                                                                'numArgs': 0,
                                                                'shortname': 'ovt'},
                                     'policyFileName': {'args': '<type __builtin__.str>',
                                                        'longname': 'policyFileName',
                                                        'modes': ['query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'pfn'},
                                     'popupOnError': {'args': '<type __builtin__.bool>',
                                                      'longname': 'popupOnError',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'poe'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'refresh': {'args': '<type __builtin__.bool>',
                                                 'longname': 'refresh',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'rfr'},
                                     'renderingSpaceName': {'args': '<type __builtin__.str>',
                                                            'longname': 'renderingSpaceName',
                                                            'modes': ['query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'rsn'},
                                     'renderingSpaceNames': {'args': '<type __builtin__.bool>',
                                                             'longname': 'renderingSpaceNames',
                                                             'modes': ['query',
                                                                       'edit'],
                                                             'numArgs': 0,
                                                             'shortname': 'rss'},
                                     'restoreDefaults': {'args': '<type __builtin__.bool>',
                                                         'longname': 'restoreDefaults',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'rde'},
                                     'viewTransformName': {'args': '<type __builtin__.str>',
                                                           'longname': 'viewTransformName',
                                                           'modes': ['query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'vtn'},
                                     'viewTransformNames': {'args': '<type __builtin__.bool>',
                                                            'longname': 'viewTransformNames',
                                                            'modes': ['query',
                                                                      'edit'],
                                                            'numArgs': 0,
                                                            'shortname': 'vts'}},
                           'shortFlags': {'cfe': 'cmConfigFileEnabled',
                                          'cfp': 'configFilePath',
                                          'cma': 'colorManageAllNodes',
                                          'cme': 'cmEnabled',
                                          'cmn': 'colorManagedNodes',
                                          'cmp': 'colorManagePots',
                                          'cmv': 'colorManagementSDKVersion',
                                          'din': 'defaultInputSpaceName',
                                          'e': 'edit',
                                          'epy': 'exportPolicy',
                                          'etp': 'equalsToPolicyFile',
                                          'ie': 'inhibitEvents',
                                          'iss': 'inputSpaceNames',
                                          'ldn': 'loadedDefaultInputSpaceName',
                                          'lon': 'loadedOutputTransformName',
                                          'lpy': 'loadPolicy',
                                          'lrn': 'loadedRenderingSpaceName',
                                          'lvn': 'loadedViewTransformName',
                                          'mcn': 'missingColorSpaceNodes',
                                          'ore': 'ocioRulesEnabled',
                                          'otc': 'outputTransformUseColorConversion',
                                          'ote': 'outputTransformEnabled',
                                          'otn': 'outputTransformName',
                                          'ots': 'outputTransformNames',
                                          'ott': 'outputTarget',
                                          'ovt': 'outputUseViewTransform',
                                          'pfn': 'policyFileName',
                                          'poe': 'popupOnError',
                                          'q': 'query',
                                          'rde': 'restoreDefaults',
                                          'rfr': 'refresh',
                                          'rsn': 'renderingSpaceName',
                                          'rss': 'renderingSpaceNames',
                                          'vtn': 'viewTransformName',
                                          'vts': 'viewTransformNames'},
                           'type': 'general'},
  'colorSliderButtonGrp': {'callbackFlags': ['dragCommand',
                                             'dc',
                                             'visibleChangeCommand',
                                             'vcc',
                                             'dragCallback',
                                             'dgc',
                                             'buttonCommand',
                                             'bc',
                                             'symbolButtonCommand',
                                             'sbc',
                                             'changeCommand',
                                             'cc',
                                             'dropCallback',
                                             'dpc'],
                           'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                          'longname': 'adjustableColumn',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'adj'},
                                     'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                           'longname': 'adjustableColumn2',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ad2'},
                                     'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                           'longname': 'adjustableColumn3',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ad3'},
                                     'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                           'longname': 'adjustableColumn4',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ad4'},
                                     'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                           'longname': 'adjustableColumn5',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ad5'},
                                     'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                           'longname': 'adjustableColumn6',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ad6'},
                                     'annotation': {'args': '<type __builtin__.str>',
                                                    'longname': 'annotation',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ann'},
                                     'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'backgroundColor',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 3,
                                                         'shortname': 'bgc'},
                                     'buttonCommand': {'args': 'script',
                                                       'longname': 'buttonCommand',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'bc'},
                                     'buttonLabel': {'args': '<type __builtin__.str>',
                                                     'longname': 'buttonLabel',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'bl'},
                                     'changeCommand': {'args': 'script',
                                                       'longname': 'changeCommand',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'cc'},
                                     'columnAlign': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAlign',
                                                     'modes': ['create',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 2,
                                                     'shortname': 'cal'},
                                     'columnAlign2': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAlign2',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'cl2'},
                                     'columnAlign3': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAlign3',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'cl3'},
                                     'columnAlign4': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAlign4',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 4,
                                                      'shortname': 'cl4'},
                                     'columnAlign5': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAlign5',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 5,
                                                      'shortname': 'cl5'},
                                     'columnAlign6': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAlign6',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 6,
                                                      'shortname': 'cl6'},
                                     'columnAttach': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnAttach',
                                                      'modes': ['create',
                                                                'edit',
                                                                'multiuse'],
                                                      'numArgs': 3,
                                                      'shortname': 'cat'},
                                     'columnAttach2': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAttach2',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 2,
                                                       'shortname': 'ct2'},
                                     'columnAttach3': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAttach3',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'ct3'},
                                     'columnAttach4': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAttach4',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 4,
                                                       'shortname': 'ct4'},
                                     'columnAttach5': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAttach5',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 5,
                                                       'shortname': 'ct5'},
                                     'columnAttach6': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAttach6',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 6,
                                                       'shortname': 'ct6'},
                                     'columnOffset2': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnOffset2',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 2,
                                                       'shortname': 'co2'},
                                     'columnOffset3': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnOffset3',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'co3'},
                                     'columnOffset4': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnOffset4',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 4,
                                                       'shortname': 'co4'},
                                     'columnOffset5': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnOffset5',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 5,
                                                       'shortname': 'co5'},
                                     'columnOffset6': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnOffset6',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 6,
                                                       'shortname': 'co6'},
                                     'columnWidth': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnWidth',
                                                     'modes': ['create',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 2,
                                                     'shortname': 'cw'},
                                     'columnWidth1': {'args': '<type __builtin__.int>',
                                                      'longname': 'columnWidth1',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cw1'},
                                     'columnWidth2': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnWidth2',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'cw2'},
                                     'columnWidth3': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnWidth3',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'cw3'},
                                     'columnWidth4': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnWidth4',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 4,
                                                      'shortname': 'cw4'},
                                     'columnWidth5': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnWidth5',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 5,
                                                      'shortname': 'cw5'},
                                     'columnWidth6': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnWidth6',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 6,
                                                      'shortname': 'cw6'},
                                     'defineTemplate': {'args': '<type __builtin__.str>',
                                                        'longname': 'defineTemplate',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'dt'},
                                     'docTag': {'args': '<type __builtin__.str>',
                                                'longname': 'docTag',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dtg'},
                                     'dragCallback': {'args': 'script',
                                                      'longname': 'dragCallback',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dgc'},
                                     'dragCommand': {'args': 'script',
                                                     'longname': 'dragCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dc'},
                                     'dropCallback': {'args': 'script',
                                                      'longname': 'dropCallback',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dpc'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'enable': {'args': '<type __builtin__.bool>',
                                                'longname': 'enable',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'en'},
                                     'enableBackground': {'args': '<type __builtin__.bool>',
                                                          'longname': 'enableBackground',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ebg'},
                                     'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                             'longname': 'enableKeyboardFocus',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'ekf'},
                                     'exists': {'args': '<type __builtin__.bool>',
                                                'longname': 'exists',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ex'},
                                     'forceDragRefresh': {'args': '<type __builtin__.bool>',
                                                          'longname': 'forceDragRefresh',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 0,
                                                          'shortname': 'fdr'},
                                     'fullPathName': {'args': '<type __builtin__.bool>',
                                                      'longname': 'fullPathName',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'fpn'},
                                     'height': {'args': '<type __builtin__.int>',
                                                'longname': 'height',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'h'},
                                     'highlightColor': {'args': ['<type __builtin__.float>',
                                                                 '<type __builtin__.float>',
                                                                 '<type __builtin__.float>'],
                                                        'longname': 'highlightColor',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 3,
                                                        'shortname': 'hlc'},
                                     'hsvValue': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'hsvValue',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'hsv'},
                                     'image': {'args': '<type __builtin__.str>',
                                               'longname': 'image',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i'},
                                     'isObscured': {'args': '<type __builtin__.bool>',
                                                    'longname': 'isObscured',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'io'},
                                     'label': {'args': '<type __builtin__.str>',
                                               'longname': 'label',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'l'},
                                     'manage': {'args': '<type __builtin__.bool>',
                                                'longname': 'manage',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'm'},
                                     'noBackground': {'args': '<type __builtin__.bool>',
                                                      'longname': 'noBackground',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'nbg'},
                                     'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                            'longname': 'numberOfPopupMenus',
                                                            'modes': ['query'],
                                                            'numArgs': 0,
                                                            'shortname': 'npm'},
                                     'parent': {'args': '<type __builtin__.str>',
                                                'longname': 'parent',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'p'},
                                     'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                        'longname': 'popupMenuArray',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'pma'},
                                     'preventOverride': {'args': '<type __builtin__.bool>',
                                                         'longname': 'preventOverride',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'po'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'rgbValue': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'rgbValue',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'rgb'},
                                     'rowAttach': {'args': ['<type __builtin__.int>',
                                                            '<type __builtin__.str>',
                                                            '<type __builtin__.int>'],
                                                   'longname': 'rowAttach',
                                                   'modes': ['create',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 3,
                                                   'shortname': 'rat'},
                                     'statusBarMessage': {'args': '<type __builtin__.str>',
                                                          'longname': 'statusBarMessage',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'sbm'},
                                     'symbolButtonCommand': {'args': 'script',
                                                             'longname': 'symbolButtonCommand',
                                                             'modes': ['create',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'sbc'},
                                     'symbolButtonDisplay': {'args': '<type __builtin__.bool>',
                                                             'longname': 'symbolButtonDisplay',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'sbd'},
                                     'useTemplate': {'args': '<type __builtin__.str>',
                                                     'longname': 'useTemplate',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ut'},
                                     'visible': {'args': '<type __builtin__.bool>',
                                                 'longname': 'visible',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'vis'},
                                     'visibleChangeCommand': {'args': 'script',
                                                              'longname': 'visibleChangeCommand',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'vcc'},
                                     'width': {'args': '<type __builtin__.int>',
                                               'longname': 'width',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'w'}},
                           'shortFlags': {'ad2': 'adjustableColumn2',
                                          'ad3': 'adjustableColumn3',
                                          'ad4': 'adjustableColumn4',
                                          'ad5': 'adjustableColumn5',
                                          'ad6': 'adjustableColumn6',
                                          'adj': 'adjustableColumn',
                                          'ann': 'annotation',
                                          'bc': 'buttonCommand',
                                          'bgc': 'backgroundColor',
                                          'bl': 'buttonLabel',
                                          'cal': 'columnAlign',
                                          'cat': 'columnAttach',
                                          'cc': 'changeCommand',
                                          'cl2': 'columnAlign2',
                                          'cl3': 'columnAlign3',
                                          'cl4': 'columnAlign4',
                                          'cl5': 'columnAlign5',
                                          'cl6': 'columnAlign6',
                                          'co2': 'columnOffset2',
                                          'co3': 'columnOffset3',
                                          'co4': 'columnOffset4',
                                          'co5': 'columnOffset5',
                                          'co6': 'columnOffset6',
                                          'ct2': 'columnAttach2',
                                          'ct3': 'columnAttach3',
                                          'ct4': 'columnAttach4',
                                          'ct5': 'columnAttach5',
                                          'ct6': 'columnAttach6',
                                          'cw': 'columnWidth',
                                          'cw1': 'columnWidth1',
                                          'cw2': 'columnWidth2',
                                          'cw3': 'columnWidth3',
                                          'cw4': 'columnWidth4',
                                          'cw5': 'columnWidth5',
                                          'cw6': 'columnWidth6',
                                          'dc': 'dragCommand',
                                          'dgc': 'dragCallback',
                                          'dpc': 'dropCallback',
                                          'dt': 'defineTemplate',
                                          'dtg': 'docTag',
                                          'e': 'edit',
                                          'ebg': 'enableBackground',
                                          'ekf': 'enableKeyboardFocus',
                                          'en': 'enable',
                                          'ex': 'exists',
                                          'fdr': 'forceDragRefresh',
                                          'fpn': 'fullPathName',
                                          'h': 'height',
                                          'hlc': 'highlightColor',
                                          'hsv': 'hsvValue',
                                          'i': 'image',
                                          'io': 'isObscured',
                                          'l': 'label',
                                          'm': 'manage',
                                          'nbg': 'noBackground',
                                          'npm': 'numberOfPopupMenus',
                                          'p': 'parent',
                                          'pma': 'popupMenuArray',
                                          'po': 'preventOverride',
                                          'q': 'query',
                                          'rat': 'rowAttach',
                                          'rgb': 'rgbValue',
                                          'sbc': 'symbolButtonCommand',
                                          'sbd': 'symbolButtonDisplay',
                                          'sbm': 'statusBarMessage',
                                          'ut': 'useTemplate',
                                          'vcc': 'visibleChangeCommand',
                                          'vis': 'visible',
                                          'w': 'width'},
                           'type': 'windows'},
  'colorSliderGrp': {'callbackFlags': ['dragCommand',
                                       'dc',
                                       'visibleChangeCommand',
                                       'vcc',
                                       'dragCallback',
                                       'dgc',
                                       'changeCommand',
                                       'cc',
                                       'dropCallback',
                                       'dpc'],
                     'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'adj'},
                               'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn2',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad2'},
                               'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn3',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad3'},
                               'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn4',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad4'},
                               'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn5',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad5'},
                               'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn6',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad6'},
                               'annotation': {'args': '<type __builtin__.str>',
                                              'longname': 'annotation',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ann'},
                               'backgroundColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'backgroundColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'bgc'},
                               'changeCommand': {'args': 'script',
                                                 'longname': 'changeCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cc'},
                               'columnAlign': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 2,
                                               'shortname': 'cal'},
                               'columnAlign2': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign2',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'cl2'},
                               'columnAlign3': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign3',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'cl3'},
                               'columnAlign4': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign4',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 4,
                                                'shortname': 'cl4'},
                               'columnAlign5': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign5',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 5,
                                                'shortname': 'cl5'},
                               'columnAlign6': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign6',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 6,
                                                'shortname': 'cl6'},
                               'columnAttach': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnAttach',
                                                'modes': ['create',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 3,
                                                'shortname': 'cat'},
                               'columnAttach2': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach2',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 2,
                                                 'shortname': 'ct2'},
                               'columnAttach3': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach3',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'ct3'},
                               'columnAttach4': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach4',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 4,
                                                 'shortname': 'ct4'},
                               'columnAttach5': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach5',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 5,
                                                 'shortname': 'ct5'},
                               'columnAttach6': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach6',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 6,
                                                 'shortname': 'ct6'},
                               'columnOffset2': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset2',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 2,
                                                 'shortname': 'co2'},
                               'columnOffset3': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset3',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'co3'},
                               'columnOffset4': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset4',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 4,
                                                 'shortname': 'co4'},
                               'columnOffset5': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset5',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 5,
                                                 'shortname': 'co5'},
                               'columnOffset6': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset6',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 6,
                                                 'shortname': 'co6'},
                               'columnWidth': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 2,
                                               'shortname': 'cw'},
                               'columnWidth1': {'args': '<type __builtin__.int>',
                                                'longname': 'columnWidth1',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cw1'},
                               'columnWidth2': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth2',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'cw2'},
                               'columnWidth3': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth3',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'cw3'},
                               'columnWidth4': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth4',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 4,
                                                'shortname': 'cw4'},
                               'columnWidth5': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth5',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 5,
                                                'shortname': 'cw5'},
                               'columnWidth6': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth6',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 6,
                                                'shortname': 'cw6'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'dragCallback': {'args': 'script',
                                                'longname': 'dragCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dgc'},
                               'dragCommand': {'args': 'script',
                                               'longname': 'dragCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dc'},
                               'dropCallback': {'args': 'script',
                                                'longname': 'dropCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dpc'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'enable': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en'},
                               'enableBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableBackground',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ebg'},
                               'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                       'longname': 'enableKeyboardFocus',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ekf'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'forceDragRefresh': {'args': '<type __builtin__.bool>',
                                                    'longname': 'forceDragRefresh',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'fdr'},
                               'fullPathName': {'args': '<type __builtin__.bool>',
                                                'longname': 'fullPathName',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'fpn'},
                               'height': {'args': '<type __builtin__.int>',
                                          'longname': 'height',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'h'},
                               'highlightColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'highlightColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'hlc'},
                               'hsvValue': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'hsvValue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'hsv'},
                               'isObscured': {'args': '<type __builtin__.bool>',
                                              'longname': 'isObscured',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'io'},
                               'label': {'args': '<type __builtin__.str>',
                                         'longname': 'label',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'l'},
                               'manage': {'args': '<type __builtin__.bool>',
                                          'longname': 'manage',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'm'},
                               'noBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'noBackground',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nbg'},
                               'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'numberOfPopupMenus',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'npm'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                  'longname': 'popupMenuArray',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'pma'},
                               'preventOverride': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preventOverride',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'po'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'rgbValue': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'rgbValue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'rgb'},
                               'rowAttach': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'rowAttach',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 3,
                                             'shortname': 'rat'},
                               'statusBarMessage': {'args': '<type __builtin__.str>',
                                                    'longname': 'statusBarMessage',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sbm'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'},
                               'visible': {'args': '<type __builtin__.bool>',
                                           'longname': 'visible',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vis'},
                               'visibleChangeCommand': {'args': 'script',
                                                        'longname': 'visibleChangeCommand',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'vcc'},
                               'width': {'args': '<type __builtin__.int>',
                                         'longname': 'width',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'w'}},
                     'shortFlags': {'ad2': 'adjustableColumn2',
                                    'ad3': 'adjustableColumn3',
                                    'ad4': 'adjustableColumn4',
                                    'ad5': 'adjustableColumn5',
                                    'ad6': 'adjustableColumn6',
                                    'adj': 'adjustableColumn',
                                    'ann': 'annotation',
                                    'bgc': 'backgroundColor',
                                    'cal': 'columnAlign',
                                    'cat': 'columnAttach',
                                    'cc': 'changeCommand',
                                    'cl2': 'columnAlign2',
                                    'cl3': 'columnAlign3',
                                    'cl4': 'columnAlign4',
                                    'cl5': 'columnAlign5',
                                    'cl6': 'columnAlign6',
                                    'co2': 'columnOffset2',
                                    'co3': 'columnOffset3',
                                    'co4': 'columnOffset4',
                                    'co5': 'columnOffset5',
                                    'co6': 'columnOffset6',
                                    'ct2': 'columnAttach2',
                                    'ct3': 'columnAttach3',
                                    'ct4': 'columnAttach4',
                                    'ct5': 'columnAttach5',
                                    'ct6': 'columnAttach6',
                                    'cw': 'columnWidth',
                                    'cw1': 'columnWidth1',
                                    'cw2': 'columnWidth2',
                                    'cw3': 'columnWidth3',
                                    'cw4': 'columnWidth4',
                                    'cw5': 'columnWidth5',
                                    'cw6': 'columnWidth6',
                                    'dc': 'dragCommand',
                                    'dgc': 'dragCallback',
                                    'dpc': 'dropCallback',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'e': 'edit',
                                    'ebg': 'enableBackground',
                                    'ekf': 'enableKeyboardFocus',
                                    'en': 'enable',
                                    'ex': 'exists',
                                    'fdr': 'forceDragRefresh',
                                    'fpn': 'fullPathName',
                                    'h': 'height',
                                    'hlc': 'highlightColor',
                                    'hsv': 'hsvValue',
                                    'io': 'isObscured',
                                    'l': 'label',
                                    'm': 'manage',
                                    'nbg': 'noBackground',
                                    'npm': 'numberOfPopupMenus',
                                    'p': 'parent',
                                    'pma': 'popupMenuArray',
                                    'po': 'preventOverride',
                                    'q': 'query',
                                    'rat': 'rowAttach',
                                    'rgb': 'rgbValue',
                                    'sbm': 'statusBarMessage',
                                    'ut': 'useTemplate',
                                    'vcc': 'visibleChangeCommand',
                                    'vis': 'visible',
                                    'w': 'width'},
                     'type': 'windows'},
  'columnLayout': {'callbackFlags': ['visibleChangeCommand',
                                     'vcc',
                                     'dragCallback',
                                     'dgc',
                                     'dropCallback',
                                     'dpc'],
                   'flags': {'adjustableColumn': {'args': '<type __builtin__.bool>',
                                                  'longname': 'adjustableColumn',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'adj'},
                             'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'childArray': {'args': '<type __builtin__.bool>',
                                            'longname': 'childArray',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ca'},
                             'columnAlign': {'args': '<type __builtin__.str>',
                                             'longname': 'columnAlign',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cal'},
                             'columnAttach': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnAttach',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'cat'},
                             'columnOffset': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnOffset',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'co'},
                             'columnWidth': {'args': '<type __builtin__.int>',
                                             'longname': 'columnWidth',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cw'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'numberOfChildren': {'args': '<type __builtin__.bool>',
                                                  'longname': 'numberOfChildren',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'nch'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'rowSpacing': {'args': '<type __builtin__.int>',
                                            'longname': 'rowSpacing',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'rs'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'shortFlags': {'adj': 'adjustableColumn',
                                  'ann': 'annotation',
                                  'bgc': 'backgroundColor',
                                  'ca': 'childArray',
                                  'cal': 'columnAlign',
                                  'cat': 'columnAttach',
                                  'co': 'columnOffset',
                                  'cw': 'columnWidth',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ebg': 'enableBackground',
                                  'ekf': 'enableKeyboardFocus',
                                  'en': 'enable',
                                  'ex': 'exists',
                                  'fpn': 'fullPathName',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'io': 'isObscured',
                                  'm': 'manage',
                                  'nbg': 'noBackground',
                                  'nch': 'numberOfChildren',
                                  'npm': 'numberOfPopupMenus',
                                  'p': 'parent',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'q': 'query',
                                  'rs': 'rowSpacing',
                                  'sbm': 'statusBarMessage',
                                  'ut': 'useTemplate',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'w': 'width'},
                   'type': 'windows'},
  'combinationShape': {'flags': {'addDriver': {'args': '<type __builtin__.bool>',
                                               'longname': 'addDriver',
                                               'modes': [],
                                               'numArgs': 0,
                                               'shortname': 'add'},
                                 'allDrivers': {'args': '<type __builtin__.bool>',
                                                'longname': 'allDrivers',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'ald'},
                                 'blendShape': {'args': '<type __builtin__.str>',
                                                'longname': 'blendShape',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'bs'},
                                 'combinationTargetIndex': {'args': '<type __builtin__.int>',
                                                            'longname': 'combinationTargetIndex',
                                                            'modes': ['create'],
                                                            'numArgs': 1,
                                                            'shortname': 'cti'},
                                 'combinationTargetName': {'args': '<type __builtin__.str>',
                                                           'longname': 'combinationTargetName',
                                                           'modes': ['create'],
                                                           'numArgs': 1,
                                                           'shortname': 'ctn'},
                                 'combineMethod': {'args': '<type __builtin__.int>',
                                                   'longname': 'combineMethod',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cm'},
                                 'driverTargetIndex': {'args': '<type __builtin__.int>',
                                                       'longname': 'driverTargetIndex',
                                                       'modes': ['create',
                                                                 'multiuse'],
                                                       'numArgs': 1,
                                                       'shortname': 'dti'},
                                 'driverTargetName': {'args': '<type __builtin__.str>',
                                                      'longname': 'driverTargetName',
                                                      'modes': ['create',
                                                                'multiuse'],
                                                      'numArgs': 1,
                                                      'shortname': 'dtn'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'exist': {'args': '<type __builtin__.bool>',
                                           'longname': 'exist',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'removeDriver': {'args': '<type __builtin__.bool>',
                                                  'longname': 'removeDriver',
                                                  'modes': [],
                                                  'numArgs': 0,
                                                  'shortname': 'rmd'}},
                       'shortFlags': {'add': 'addDriver',
                                      'ald': 'allDrivers',
                                      'bs': 'blendShape',
                                      'cm': 'combineMethod',
                                      'cti': 'combinationTargetIndex',
                                      'ctn': 'combinationTargetName',
                                      'dti': 'driverTargetIndex',
                                      'dtn': 'driverTargetName',
                                      'e': 'edit',
                                      'ex': 'exist',
                                      'q': 'query',
                                      'rmd': 'removeDriver'},
                       'type': 'animation'},
  'commandEcho': {'flags': {'addFilter': {'args': '<type __builtin__.str>',
                                          'longname': 'addFilter',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'af'},
                            'filter': {'args': '<type __builtin__.str>',
                                       'longname': 'filter',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'f'},
                            'lineNumbers': {'args': '<type __builtin__.bool>',
                                            'longname': 'lineNumbers',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'ln'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'state': {'args': '<type __builtin__.bool>',
                                      'longname': 'state',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'st'}},
                  'shortFlags': {'af': 'addFilter',
                                 'f': 'filter',
                                 'ln': 'lineNumbers',
                                 'q': 'query',
                                 'st': 'state'},
                  'type': 'language'},
  'commandLine': {'callbackFlags': ['visibleChangeCommand',
                                    'vcc',
                                    'dragCallback',
                                    'dgc',
                                    'enterCommand',
                                    'ec',
                                    'command',
                                    'c',
                                    'dropCallback',
                                    'dpc'],
                  'flags': {'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'command': {'args': 'script',
                                        'longname': 'command',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'c'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'enterCommand': {'args': 'script',
                                             'longname': 'enterCommand',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ec'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'holdFocus': {'args': '<type __builtin__.bool>',
                                          'longname': 'holdFocus',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'hf'},
                            'inputAnnotation': {'args': '<type __builtin__.str>',
                                                'longname': 'inputAnnotation',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ian'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfHistoryLines': {'args': '<type __builtin__.int>',
                                                     'longname': 'numberOfHistoryLines',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'nhl'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'outputAnnotation': {'args': '<type __builtin__.str>',
                                                 'longname': 'outputAnnotation',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'oan'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'sourceType': {'args': '<type __builtin__.str>',
                                           'longname': 'sourceType',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'st'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'c': 'command',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ec': 'enterCommand',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'ex': 'exists',
                                 'fpn': 'fullPathName',
                                 'h': 'height',
                                 'hf': 'holdFocus',
                                 'hlc': 'highlightColor',
                                 'ian': 'inputAnnotation',
                                 'io': 'isObscured',
                                 'm': 'manage',
                                 'nbg': 'noBackground',
                                 'nhl': 'numberOfHistoryLines',
                                 'npm': 'numberOfPopupMenus',
                                 'oan': 'outputAnnotation',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'sbm': 'statusBarMessage',
                                 'st': 'sourceType',
                                 'ut': 'useTemplate',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'commandLogging': {'callbackFlags': ['recordCommands',
                                       'rc',
                                       'logCommands',
                                       'lc'],
                     'flags': {'historySize': {'args': '<type __builtin__.int>',
                                               'longname': 'historySize',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'hs'},
                               'logCommands': {'args': '<type __builtin__.bool>',
                                               'longname': 'logCommands',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'lc'},
                               'logFile': {'args': '<type __builtin__.str>',
                                           'longname': 'logFile',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'lf'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'recordCommands': {'args': '<type __builtin__.bool>',
                                                  'longname': 'recordCommands',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'rc'},
                               'resetLogFile': {'args': '<type __builtin__.bool>',
                                                'longname': 'resetLogFile',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 0,
                                                'shortname': 'rl'}},
                     'shortFlags': {'hs': 'historySize',
                                    'lc': 'logCommands',
                                    'lf': 'logFile',
                                    'q': 'query',
                                    'rc': 'recordCommands',
                                    'rl': 'resetLogFile'},
                     'type': 'general'},
  'commandPort': {'callbackFlags': ['returnNumCommands', 'rnc'],
                  'flags': {'bufferSize': {'args': '<type __builtin__.int>',
                                           'longname': 'bufferSize',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'bs'},
                            'close': {'args': '<type __builtin__.bool>',
                                      'longname': 'close',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'cl'},
                            'echoOutput': {'args': '<type __builtin__.bool>',
                                           'longname': 'echoOutput',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'eo'},
                            'listPorts': {'args': '<type __builtin__.bool>',
                                          'longname': 'listPorts',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'lp'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'noreturn': {'args': '<type __builtin__.bool>',
                                         'longname': 'noreturn',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'nr'},
                            'pickleOutput': {'args': '<type __builtin__.bool>',
                                             'longname': 'pickleOutput',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'po'},
                            'prefix': {'args': '<type __builtin__.str>',
                                       'longname': 'prefix',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'pre'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'returnNumCommands': {'args': '<type __builtin__.bool>',
                                                  'longname': 'returnNumCommands',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'rnc'},
                            'securityWarning': {'args': '<type __builtin__.bool>',
                                                'longname': 'securityWarning',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'sw'},
                            'sourceType': {'args': '<type __builtin__.str>',
                                           'longname': 'sourceType',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'stp'}},
                  'shortFlags': {'bs': 'bufferSize',
                                 'cl': 'close',
                                 'eo': 'echoOutput',
                                 'lp': 'listPorts',
                                 'n': 'name',
                                 'nr': 'noreturn',
                                 'po': 'pickleOutput',
                                 'pre': 'prefix',
                                 'q': 'query',
                                 'rnc': 'returnNumCommands',
                                 'stp': 'sourceType',
                                 'sw': 'securityWarning'},
                  'type': 'general'},
  'componentBox': {'callbackFlags': ['visibleChangeCommand',
                                     'vcc',
                                     'dragCallback',
                                     'dgc',
                                     'dropCallback',
                                     'dpc'],
                   'flags': {'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'execute': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.bool>'],
                                         'longname': 'execute',
                                         'modes': ['edit'],
                                         'numArgs': 2,
                                         'shortname': 'exe'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'labelWidth': {'args': '<type __builtin__.int>',
                                            'longname': 'labelWidth',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'lw'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'maxHeight': {'args': '<type __builtin__.int>',
                                           'longname': 'maxHeight',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'mh'},
                             'maxWidth': {'args': '<type __builtin__.int>',
                                          'longname': 'maxWidth',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'mw'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'precision': {'args': '<type __builtin__.int>',
                                           'longname': 'precision',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pre'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'rowHeight': {'args': '<type __builtin__.int>',
                                           'longname': 'rowHeight',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'rh'},
                             'selectedAttr': {'args': '<type __builtin__.bool>',
                                              'longname': 'selectedAttr',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'sla'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'shortFlags': {'ann': 'annotation',
                                  'bgc': 'backgroundColor',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ebg': 'enableBackground',
                                  'ekf': 'enableKeyboardFocus',
                                  'en': 'enable',
                                  'ex': 'exists',
                                  'exe': 'execute',
                                  'fpn': 'fullPathName',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'io': 'isObscured',
                                  'lw': 'labelWidth',
                                  'm': 'manage',
                                  'mh': 'maxHeight',
                                  'mw': 'maxWidth',
                                  'nbg': 'noBackground',
                                  'npm': 'numberOfPopupMenus',
                                  'p': 'parent',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'pre': 'precision',
                                  'q': 'query',
                                  'rh': 'rowHeight',
                                  'sbm': 'statusBarMessage',
                                  'sla': 'selectedAttr',
                                  'ut': 'useTemplate',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'w': 'width'},
                   'type': 'windows'},
  'componentEditor': {'flags': {'control': {'args': '<type __builtin__.bool>',
                                            'longname': 'control',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ctl'},
                                'defineTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'defineTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'dt'},
                                'docTag': {'args': '<type __builtin__.str>',
                                           'longname': 'docTag',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dtg'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'filter': {'args': '<type __builtin__.str>',
                                           'longname': 'filter',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'f'},
                                'floatField': {'args': '<type __builtin__.str>',
                                               'longname': 'floatField',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ff'},
                                'floatSlider': {'args': '<type __builtin__.str>',
                                                'longname': 'floatSlider',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'fs'},
                                'forceMainConnection': {'args': '<type __builtin__.str>',
                                                        'longname': 'forceMainConnection',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'fmc'},
                                'hidePathName': {'args': '<type __builtin__.bool>',
                                                 'longname': 'hidePathName',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'hpn'},
                                'hideZeroColumns': {'args': '<type __builtin__.bool>',
                                                    'longname': 'hideZeroColumns',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'hzc'},
                                'highlightConnection': {'args': '<type __builtin__.str>',
                                                        'longname': 'highlightConnection',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'hlc'},
                                'lockInput': {'args': '<type __builtin__.bool>',
                                              'longname': 'lockInput',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'li'},
                                'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                       'longname': 'lockMainConnection',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'lck'},
                                'mainListConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'mainListConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'mlc'},
                                'newTab': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'newTab',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 3,
                                           'shortname': 'nt'},
                                'operationCount': {'args': '<type __builtin__.bool>',
                                                   'longname': 'operationCount',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'oc'},
                                'operationLabels': {'args': '<type __builtin__.bool>',
                                                    'longname': 'operationLabels',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'ol'},
                                'operationType': {'args': '<type __builtin__.int>',
                                                  'longname': 'operationType',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ot'},
                                'panel': {'args': '<type __builtin__.str>',
                                          'longname': 'panel',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'pnl'},
                                'parent': {'args': '<type __builtin__.str>',
                                           'longname': 'parent',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'p'},
                                'precision': {'args': '<type __builtin__.int>',
                                              'longname': 'precision',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pre'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'removeTab': {'args': '<type __builtin__.str>',
                                              'longname': 'removeTab',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'rt'},
                                'selected': {'args': '<type __builtin__.bool>',
                                             'longname': 'selected',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'sl'},
                                'selectionConnection': {'args': '<type __builtin__.str>',
                                                        'longname': 'selectionConnection',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'slc'},
                                'setOperationLabel': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'setOperationLabel',
                                                      'modes': ['edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'sol'},
                                'showObjects': {'args': '<type __builtin__.bool>',
                                                'longname': 'showObjects',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'so'},
                                'showSelected': {'args': '<type __builtin__.bool>',
                                                 'longname': 'showSelected',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ss'},
                                'sortAlpha': {'args': '<type __builtin__.bool>',
                                              'longname': 'sortAlpha',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sa'},
                                'stateString': {'args': '<type __builtin__.bool>',
                                                'longname': 'stateString',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'sts'},
                                'unParent': {'args': '<type __builtin__.bool>',
                                             'longname': 'unParent',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'up'},
                                'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'unlockMainConnection',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'ulk'},
                                'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'updateMainConnection',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'upd'},
                                'useTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'useTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ut'}},
                      'shortFlags': {'ctl': 'control',
                                     'dt': 'defineTemplate',
                                     'dtg': 'docTag',
                                     'e': 'edit',
                                     'ex': 'exists',
                                     'f': 'filter',
                                     'ff': 'floatField',
                                     'fmc': 'forceMainConnection',
                                     'fs': 'floatSlider',
                                     'hlc': 'highlightConnection',
                                     'hpn': 'hidePathName',
                                     'hzc': 'hideZeroColumns',
                                     'lck': 'lockMainConnection',
                                     'li': 'lockInput',
                                     'mlc': 'mainListConnection',
                                     'nt': 'newTab',
                                     'oc': 'operationCount',
                                     'ol': 'operationLabels',
                                     'ot': 'operationType',
                                     'p': 'parent',
                                     'pnl': 'panel',
                                     'pre': 'precision',
                                     'q': 'query',
                                     'rt': 'removeTab',
                                     'sa': 'sortAlpha',
                                     'sl': 'selected',
                                     'slc': 'selectionConnection',
                                     'so': 'showObjects',
                                     'sol': 'setOperationLabel',
                                     'ss': 'showSelected',
                                     'sts': 'stateString',
                                     'ulk': 'unlockMainConnection',
                                     'up': 'unParent',
                                     'upd': 'updateMainConnection',
                                     'ut': 'useTemplate'},
                      'type': 'windows'},
  'condition': {'flags': {'delete': {'args': '<type __builtin__.bool>',
                                     'longname': 'delete',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'delete'},
                          'dependency': {'args': '<type __builtin__.str>',
                                         'longname': 'dependency',
                                         'modes': ['create', 'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'd'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'initialize': {'args': '<type __builtin__.bool>',
                                         'longname': 'initialize',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'i'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'script': {'args': '<type __builtin__.str>',
                                     'longname': 'script',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 's'},
                          'state': {'args': '<type __builtin__.bool>',
                                    'longname': 'state',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'st'}},
                'removedFlags': {'del': 'delete'},
                'shortFlags': {'d': 'dependency',
                               'delete': 'delete',
                               'e': 'edit',
                               'i': 'initialize',
                               'q': 'query',
                               's': 'script',
                               'st': 'state'},
                'type': 'language'},
  'cone': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                       '<type __builtin__.float>',
                                       '<type __builtin__.float>'],
                              'longname': 'axis',
                              'modes': ['create', 'query', 'edit'],
                              'numArgs': 3,
                              'shortname': 'ax'},
                     'caching': {'args': '<type __builtin__.bool>',
                                 'longname': 'caching',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'resultNeedsCasting': True,
                                 'shortname': 'cch'},
                     'constructionHistory': {'args': '<type __builtin__.bool>',
                                             'longname': 'constructionHistory',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ch'},
                     'degree': {'args': '<type __builtin__.int>',
                                'longname': 'degree',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'resultNeedsCasting': True,
                                'shortname': 'd'},
                     'edit': {'args': '<type __builtin__.bool>',
                              'longname': 'edit',
                              'numArgs': 0,
                              'shortname': 'e'},
                     'endSweep': {'args': '<type __builtin__.float>',
                                  'longname': 'endSweep',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'esw'},
                     'frozen': {'args': '<type __builtin__.bool>',
                                'longname': 'frozen',
                                'modes': ['query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'fzn'},
                     'heightRatio': {'args': '<type __builtin__.float>',
                                     'longname': 'heightRatio',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'hr'},
                     'name': {'args': '<type __builtin__.str>',
                              'longname': 'name',
                              'modes': ['create'],
                              'numArgs': 1,
                              'shortname': 'n'},
                     'nodeState': {'args': '<type __builtin__.int>',
                                   'longname': 'nodeState',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'resultNeedsCasting': True,
                                   'shortname': 'nds'},
                     'object': {'args': '<type __builtin__.bool>',
                                'longname': 'object',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'o'},
                     'pivot': {'args': ['<type __builtin__.float>',
                                        '<type __builtin__.float>',
                                        '<type __builtin__.float>'],
                               'longname': 'pivot',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 3,
                               'shortname': 'p'},
                     'polygon': {'args': '<type __builtin__.int>',
                                 'longname': 'polygon',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'po'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'radius': {'args': '<type __builtin__.float>',
                                'longname': 'radius',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'r'},
                     'sections': {'args': '<type __builtin__.int>',
                                  'longname': 'sections',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'resultNeedsCasting': True,
                                  'shortname': 's'},
                     'spans': {'args': '<type __builtin__.int>',
                               'longname': 'spans',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'resultNeedsCasting': True,
                               'shortname': 'nsp'},
                     'startSweep': {'args': '<type __builtin__.float>',
                                    'longname': 'startSweep',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'ssw'},
                     'tolerance': {'args': '<type __builtin__.float>',
                                   'longname': 'tolerance',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'tol'},
                     'useOldInitBehaviour': {'args': '<type __builtin__.bool>',
                                             'longname': 'useOldInitBehaviour',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'resultNeedsCasting': True,
                                             'shortname': 'oib'},
                     'useTolerance': {'args': '<type __builtin__.bool>',
                                      'longname': 'useTolerance',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsCasting': True,
                                      'shortname': 'ut'}},
           'shortFlags': {'ax': 'axis',
                          'cch': 'caching',
                          'ch': 'constructionHistory',
                          'd': 'degree',
                          'e': 'edit',
                          'esw': 'endSweep',
                          'fzn': 'frozen',
                          'hr': 'heightRatio',
                          'n': 'name',
                          'nds': 'nodeState',
                          'nsp': 'spans',
                          'o': 'object',
                          'oib': 'useOldInitBehaviour',
                          'p': 'pivot',
                          'po': 'polygon',
                          'q': 'query',
                          'r': 'radius',
                          's': 'sections',
                          'ssw': 'startSweep',
                          'tol': 'tolerance',
                          'ut': 'useTolerance'},
           'type': 'modeling'},
  'confirmDialog': {'flags': {'annotation': {'args': '<type __builtin__.str>',
                                             'longname': 'annotation',
                                             'modes': ['create',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'ann'},
                              'backgroundColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'backgroundColor',
                                                  'modes': ['create'],
                                                  'numArgs': 3,
                                                  'shortname': 'bgc'},
                              'button': {'args': '<type __builtin__.str>',
                                         'longname': 'button',
                                         'modes': ['create', 'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'b'},
                              'cancelButton': {'args': '<type __builtin__.str>',
                                               'longname': 'cancelButton',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'cb'},
                              'defaultButton': {'args': '<type __builtin__.str>',
                                                'longname': 'defaultButton',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'db'},
                              'dismissString': {'args': '<type __builtin__.str>',
                                                'longname': 'dismissString',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ds'},
                              'icon': {'args': '<type __builtin__.str>',
                                       'longname': 'icon',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'icn'},
                              'message': {'args': '<type __builtin__.str>',
                                          'longname': 'message',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'm'},
                              'messageAlign': {'args': '<type __builtin__.str>',
                                               'longname': 'messageAlign',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ma'},
                              'parent': {'args': '<type __builtin__.str>',
                                         'longname': 'parent',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'p'},
                              'title': {'args': '<type __builtin__.str>',
                                        'longname': 'title',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 't'}},
                    'shortFlags': {'ann': 'annotation',
                                   'b': 'button',
                                   'bgc': 'backgroundColor',
                                   'cb': 'cancelButton',
                                   'db': 'defaultButton',
                                   'ds': 'dismissString',
                                   'icn': 'icon',
                                   'm': 'message',
                                   'ma': 'messageAlign',
                                   'p': 'parent',
                                   't': 'title'},
                    'type': 'windows'},
  'connectAttr': {'flags': {'force': {'args': '<type __builtin__.bool>',
                                      'longname': 'force',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'f'},
                            'lock': {'args': '<type __builtin__.bool>',
                                     'longname': 'lock',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'l'},
                            'nextAvailable': {'args': '<type __builtin__.bool>',
                                              'longname': 'nextAvailable',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'na'},
                            'referenceDest': {'args': '<type __builtin__.str>',
                                              'longname': 'referenceDest',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'rd'}},
                  'shortFlags': {'f': 'force',
                                 'l': 'lock',
                                 'na': 'nextAvailable',
                                 'rd': 'referenceDest'},
                  'type': 'general'},
  'connectControl': {'flags': {'fileName': {'args': '<type __builtin__.bool>',
                                            'longname': 'fileName',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'fi'},
                               'index': {'args': '<type __builtin__.int>',
                                         'longname': 'index',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'index'},
                               'preventContextualMenu': {'args': '<type __builtin__.bool>',
                                                         'longname': 'preventContextualMenu',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'pcm'},
                               'preventOverride': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preventOverride',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'po'}},
                     'removedFlags': {'in': 'index'},
                     'shortFlags': {'fi': 'fileName',
                                    'index': 'index',
                                    'pcm': 'preventContextualMenu',
                                    'po': 'preventOverride'},
                     'type': 'language'},
  'connectDynamic': {'callbackFlags': ['addScriptHandler', 'ash'],
                     'flags': {'addScriptHandler': {'args': 'script',
                                                    'longname': 'addScriptHandler',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ash'},
                               'collisions': {'args': '<type __builtin__.str>',
                                              'longname': 'collisions',
                                              'modes': ['create',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'c'},
                               'delete': {'args': '<type __builtin__.bool>',
                                          'longname': 'delete',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'd'},
                               'emitters': {'args': '<type __builtin__.str>',
                                            'longname': 'emitters',
                                            'modes': ['create',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'em'},
                               'fields': {'args': '<type __builtin__.str>',
                                          'longname': 'fields',
                                          'modes': ['create', 'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'f'},
                               'removeScriptHandler': {'args': '<type __builtin__.int>',
                                                       'longname': 'removeScriptHandler',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'rsh'}},
                     'shortFlags': {'ash': 'addScriptHandler',
                                    'c': 'collisions',
                                    'd': 'delete',
                                    'em': 'emitters',
                                    'f': 'fields',
                                    'rsh': 'removeScriptHandler'},
                     'type': 'effects'},
  'connectJoint': {'flags': {'connectMode': {'args': '<type __builtin__.bool>',
                                             'longname': 'connectMode',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'cm'},
                             'parentMode': {'args': '<type __builtin__.bool>',
                                            'longname': 'parentMode',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'pm'}},
                   'shortFlags': {'cm': 'connectMode', 'pm': 'parentMode'},
                   'type': 'animation'},
  'connectionInfo': {'flags': {'destinationFromSource': {'args': '<type __builtin__.bool>',
                                                         'longname': 'destinationFromSource',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'dfs'},
                               'getExactDestination': {'args': '<type __builtin__.bool>',
                                                       'longname': 'getExactDestination',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'ged'},
                               'getExactSource': {'args': '<type __builtin__.bool>',
                                                  'longname': 'getExactSource',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'ges'},
                               'getLockedAncestor': {'args': '<type __builtin__.bool>',
                                                     'longname': 'getLockedAncestor',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'gla'},
                               'isDestination': {'args': '<type __builtin__.bool>',
                                                 'longname': 'isDestination',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'id'},
                               'isExactDestination': {'args': '<type __builtin__.bool>',
                                                      'longname': 'isExactDestination',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'ied'},
                               'isExactSource': {'args': '<type __builtin__.bool>',
                                                 'longname': 'isExactSource',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ies'},
                               'isLocked': {'args': '<type __builtin__.bool>',
                                            'longname': 'isLocked',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'il'},
                               'isSource': {'args': '<type __builtin__.bool>',
                                            'longname': 'isSource',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'isSource'},
                               'sourceFromDestination': {'args': '<type __builtin__.bool>',
                                                         'longname': 'sourceFromDestination',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'sfd'}},
                     'removedFlags': {'is': 'isSource'},
                     'shortFlags': {'dfs': 'destinationFromSource',
                                    'ged': 'getExactDestination',
                                    'ges': 'getExactSource',
                                    'gla': 'getLockedAncestor',
                                    'id': 'isDestination',
                                    'ied': 'isExactDestination',
                                    'ies': 'isExactSource',
                                    'il': 'isLocked',
                                    'isSource': 'isSource',
                                    'sfd': 'sourceFromDestination'},
                     'type': 'general'},
  'constrain': {'flags': {'barrier': {'args': '<type __builtin__.bool>',
                                      'longname': 'barrier',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'br'},
                          'damping': {'args': '<type __builtin__.float>',
                                      'longname': 'damping',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'd'},
                          'directionalHinge': {'args': '<type __builtin__.bool>',
                                               'longname': 'directionalHinge',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 0,
                                               'shortname': 'dhi'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'hinge': {'args': '<type __builtin__.bool>',
                                    'longname': 'hinge',
                                    'modes': ['create', 'query'],
                                    'numArgs': 0,
                                    'shortname': 'hi'},
                          'interpenetrate': {'args': '<type __builtin__.bool>',
                                             'longname': 'interpenetrate',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i'},
                          'nail': {'args': '<type __builtin__.bool>',
                                   'longname': 'nail',
                                   'modes': ['create', 'query'],
                                   'numArgs': 0,
                                   'shortname': 'na'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'orientation': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'orientation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 3,
                                          'shortname': 'o'},
                          'pinConstraint': {'args': '<type __builtin__.bool>',
                                            'longname': 'pinConstraint',
                                            'modes': ['create', 'query'],
                                            'numArgs': 0,
                                            'shortname': 'pin'},
                          'position': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'position',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 3,
                                       'shortname': 'p'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'restLength': {'args': '<type __builtin__.float>',
                                         'longname': 'restLength',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'rl'},
                          'spring': {'args': '<type __builtin__.bool>',
                                     'longname': 'spring',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 's'},
                          'stiffness': {'args': '<type __builtin__.float>',
                                        'longname': 'stiffness',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'st'}},
                'shortFlags': {'br': 'barrier',
                               'd': 'damping',
                               'dhi': 'directionalHinge',
                               'e': 'edit',
                               'hi': 'hinge',
                               'i': 'interpenetrate',
                               'n': 'name',
                               'na': 'nail',
                               'o': 'orientation',
                               'p': 'position',
                               'pin': 'pinConstraint',
                               'q': 'query',
                               'rl': 'restLength',
                               's': 'spring',
                               'st': 'stiffness'},
                'type': 'effects'},
  'constructionHistory': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'toggle': {'args': '<type __builtin__.bool>',
                                               'longname': 'toggle',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'tgl'}},
                          'shortFlags': {'q': 'query', 'tgl': 'toggle'},
                          'type': 'modeling'},
  'container': {'flags': {'addNode': {'args': '<type __builtin__.str>',
                                      'longname': 'addNode',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'an'},
                          'asset': {'args': '<type __builtin__.str>',
                                    'longname': 'asset',
                                    'modes': ['query'],
                                    'numArgs': 1,
                                    'shortname': 'a'},
                          'assetMember': {'args': '<type __builtin__.str>',
                                          'longname': 'assetMember',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'am'},
                          'bindAttr': {'args': ['<type __builtin__.str>',
                                                '<type __builtin__.str>'],
                                       'longname': 'bindAttr',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 2,
                                       'shortname': 'ba'},
                          'connectionList': {'args': '<type __builtin__.bool>',
                                             'longname': 'connectionList',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'cl'},
                          'current': {'args': '<type __builtin__.bool>',
                                      'longname': 'current',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'c'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'fileName': {'args': '<type __builtin__.str>',
                                       'longname': 'fileName',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'fn'},
                          'findContainer': {'args': '<type __builtin__.str>',
                                            'longname': 'findContainer',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'fc'},
                          'force': {'args': '<type __builtin__.bool>',
                                    'longname': 'force',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'f'},
                          'includeHierarchyAbove': {'args': '<type __builtin__.bool>',
                                                    'longname': 'includeHierarchyAbove',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'iha'},
                          'includeHierarchyBelow': {'args': '<type __builtin__.bool>',
                                                    'longname': 'includeHierarchyBelow',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'ihb'},
                          'includeNetwork': {'args': '<type __builtin__.bool>',
                                             'longname': 'includeNetwork',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'inc'},
                          'includeNetworkDetails': {'args': '<type __builtin__.str>',
                                                    'longname': 'includeNetworkDetails',
                                                    'modes': ['create',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 1,
                                                    'shortname': 'ind'},
                          'includeShaders': {'args': '<type __builtin__.bool>',
                                             'longname': 'includeShaders',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'isd'},
                          'includeShapes': {'args': '<type __builtin__.bool>',
                                            'longname': 'includeShapes',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'ish'},
                          'includeTransform': {'args': '<type __builtin__.bool>',
                                               'longname': 'includeTransform',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'it'},
                          'isContainer': {'args': '<type __builtin__.bool>',
                                          'longname': 'isContainer',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'isc'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nodeList': {'args': '<type __builtin__.bool>',
                                       'longname': 'nodeList',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'nl'},
                          'nodeNamePrefix': {'args': '<type __builtin__.bool>',
                                             'longname': 'nodeNamePrefix',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'nnp'},
                          'parentContainer': {'args': '<type __builtin__.bool>',
                                              'longname': 'parentContainer',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'par'},
                          'preview': {'args': '<type __builtin__.bool>',
                                      'longname': 'preview',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'p'},
                          'publishAndBind': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'publishAndBind',
                                             'modes': ['edit'],
                                             'numArgs': 2,
                                             'shortname': 'pb'},
                          'publishAsChild': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'publishAsChild',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 2,
                                             'shortname': 'pac'},
                          'publishAsParent': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'publishAsParent',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'pap'},
                          'publishAsRoot': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.bool>'],
                                            'longname': 'publishAsRoot',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 2,
                                            'shortname': 'pro'},
                          'publishAttr': {'args': '<type __builtin__.str>',
                                          'longname': 'publishAttr',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'pa'},
                          'publishConnections': {'args': '<type __builtin__.bool>',
                                                 'longname': 'publishConnections',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'pc'},
                          'publishName': {'args': '<type __builtin__.str>',
                                          'longname': 'publishName',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pn'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'removeContainer': {'args': '<type __builtin__.bool>',
                                              'longname': 'removeContainer',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'rc'},
                          'removeNode': {'args': '<type __builtin__.str>',
                                         'longname': 'removeNode',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 'rn'},
                          'type': {'args': '<type __builtin__.str>',
                                   'longname': 'type',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'typ'},
                          'unbindAndUnpublish': {'args': '<type __builtin__.str>',
                                                 'longname': 'unbindAndUnpublish',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ubp'},
                          'unbindAttr': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'unbindAttr',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 2,
                                         'shortname': 'ua'},
                          'unbindChild': {'args': '<type __builtin__.str>',
                                          'longname': 'unbindChild',
                                          'modes': ['edit'],
                                          'numArgs': 1,
                                          'shortname': 'unc'},
                          'unbindParent': {'args': '<type __builtin__.str>',
                                           'longname': 'unbindParent',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'unp'},
                          'unpublishChild': {'args': '<type __builtin__.str>',
                                             'longname': 'unpublishChild',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'upc'},
                          'unpublishName': {'args': '<type __builtin__.str>',
                                            'longname': 'unpublishName',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'un'},
                          'unpublishParent': {'args': '<type __builtin__.str>',
                                              'longname': 'unpublishParent',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'upp'},
                          'unsortedOrder': {'args': '<type __builtin__.bool>',
                                            'longname': 'unsortedOrder',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'uso'}},
                'shortFlags': {'a': 'asset',
                               'am': 'assetMember',
                               'an': 'addNode',
                               'ba': 'bindAttr',
                               'c': 'current',
                               'cl': 'connectionList',
                               'e': 'edit',
                               'f': 'force',
                               'fc': 'findContainer',
                               'fn': 'fileName',
                               'iha': 'includeHierarchyAbove',
                               'ihb': 'includeHierarchyBelow',
                               'inc': 'includeNetwork',
                               'ind': 'includeNetworkDetails',
                               'isc': 'isContainer',
                               'isd': 'includeShaders',
                               'ish': 'includeShapes',
                               'it': 'includeTransform',
                               'n': 'name',
                               'nl': 'nodeList',
                               'nnp': 'nodeNamePrefix',
                               'p': 'preview',
                               'pa': 'publishAttr',
                               'pac': 'publishAsChild',
                               'pap': 'publishAsParent',
                               'par': 'parentContainer',
                               'pb': 'publishAndBind',
                               'pc': 'publishConnections',
                               'pn': 'publishName',
                               'pro': 'publishAsRoot',
                               'q': 'query',
                               'rc': 'removeContainer',
                               'rn': 'removeNode',
                               'typ': 'type',
                               'ua': 'unbindAttr',
                               'ubp': 'unbindAndUnpublish',
                               'un': 'unpublishName',
                               'unc': 'unbindChild',
                               'unp': 'unbindParent',
                               'upc': 'unpublishChild',
                               'upp': 'unpublishParent',
                               'uso': 'unsortedOrder'},
                'type': 'general'},
  'containerBind': {'flags': {'allNames': {'args': '<type __builtin__.bool>',
                                           'longname': 'allNames',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'all'},
                              'bindingSet': {'args': '<type __builtin__.str>',
                                             'longname': 'bindingSet',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'bs'},
                              'bindingSetConditions': {'args': '<type __builtin__.bool>',
                                                       'longname': 'bindingSetConditions',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'bsc'},
                              'bindingSetList': {'args': '<type __builtin__.bool>',
                                                 'longname': 'bindingSetList',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'bsl'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'force': {'args': '<type __builtin__.bool>',
                                        'longname': 'force',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'f'},
                              'preview': {'args': '<type __builtin__.bool>',
                                          'longname': 'preview',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'p'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'}},
                    'shortFlags': {'all': 'allNames',
                                   'bs': 'bindingSet',
                                   'bsc': 'bindingSetConditions',
                                   'bsl': 'bindingSetList',
                                   'e': 'edit',
                                   'f': 'force',
                                   'p': 'preview',
                                   'q': 'query'},
                    'type': 'general'},
  'containerProxy': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'fromTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'fromTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ft'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'type': {'args': '<type __builtin__.str>',
                                        'longname': 'type',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'typ'}},
                     'shortFlags': {'e': 'edit',
                                    'ft': 'fromTemplate',
                                    'q': 'query',
                                    'typ': 'type'},
                     'type': 'general'},
  'containerPublish': {'flags': {'bindNode': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'bindNode',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 2,
                                              'shortname': 'bn'},
                                 'bindTemplateStandins': {'args': '<type __builtin__.bool>',
                                                          'longname': 'bindTemplateStandins',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 0,
                                                          'shortname': 'bts'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'inConnections': {'args': '<type __builtin__.bool>',
                                                   'longname': 'inConnections',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'ic'},
                                 'mergeShared': {'args': '<type __builtin__.bool>',
                                                 'longname': 'mergeShared',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ms'},
                                 'outConnections': {'args': '<type __builtin__.bool>',
                                                    'longname': 'outConnections',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'oc'},
                                 'publishNode': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'publishNode',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 2,
                                                 'shortname': 'pn'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'unbindNode': {'args': '<type __builtin__.str>',
                                                'longname': 'unbindNode',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ubn'},
                                 'unpublishNode': {'args': '<type __builtin__.str>',
                                                   'longname': 'unpublishNode',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'upn'}},
                       'shortFlags': {'bn': 'bindNode',
                                      'bts': 'bindTemplateStandins',
                                      'e': 'edit',
                                      'ic': 'inConnections',
                                      'ms': 'mergeShared',
                                      'oc': 'outConnections',
                                      'pn': 'publishNode',
                                      'q': 'query',
                                      'ubn': 'unbindNode',
                                      'upn': 'unpublishNode'},
                       'type': 'general'},
  'containerTemplate': {'flags': {'addBindingSet': {'args': '<type __builtin__.str>',
                                                    'longname': 'addBindingSet',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'abs'},
                                  'addNames': {'args': '<type __builtin__.bool>',
                                               'longname': 'addNames',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'an'},
                                  'addView': {'args': '<type __builtin__.str>',
                                              'longname': 'addView',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'av'},
                                  'allKeyable': {'args': '<type __builtin__.bool>',
                                                 'longname': 'allKeyable',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'ak'},
                                  'attribute': {'args': '<type __builtin__.str>',
                                                'longname': 'attribute',
                                                'modes': ['create',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'at'},
                                  'attributeList': {'args': '<type __builtin__.str>',
                                                    'longname': 'attributeList',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'al'},
                                  'baseName': {'args': '<type __builtin__.str>',
                                               'longname': 'baseName',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'bn'},
                                  'bindingSetList': {'args': '<type __builtin__.str>',
                                                     'longname': 'bindingSetList',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'bsl'},
                                  'childAnchor': {'args': '<type __builtin__.bool>',
                                                  'longname': 'childAnchor',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 0,
                                                  'shortname': 'can'},
                                  'delete': {'args': '<type __builtin__.bool>',
                                             'longname': 'delete',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'd'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'expandCompounds': {'args': '<type __builtin__.bool>',
                                                      'longname': 'expandCompounds',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ec'},
                                  'fileName': {'args': '<type __builtin__.str>',
                                               'longname': 'fileName',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'fn'},
                                  'force': {'args': '<type __builtin__.bool>',
                                            'longname': 'force',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'f'},
                                  'fromContainer': {'args': '<type __builtin__.str>',
                                                    'longname': 'fromContainer',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'fc'},
                                  'fromSelection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'fromSelection',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'fs'},
                                  'layoutMode': {'args': '<type __builtin__.int>',
                                                 'longname': 'layoutMode',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'lm'},
                                  'load': {'args': '<type __builtin__.bool>',
                                           'longname': 'load',
                                           'modes': [],
                                           'numArgs': 0,
                                           'shortname': 'l'},
                                  'matchFile': {'args': '<type __builtin__.str>',
                                                'longname': 'matchFile',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'mf'},
                                  'matchName': {'args': '<type __builtin__.str>',
                                                'longname': 'matchName',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'mn'},
                                  'parentAnchor': {'args': '<type __builtin__.bool>',
                                                   'longname': 'parentAnchor',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 0,
                                                   'shortname': 'pan'},
                                  'publishedNodeList': {'args': '<type __builtin__.str>',
                                                        'longname': 'publishedNodeList',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'pnl'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'removeBindingSet': {'args': '<type __builtin__.str>',
                                                       'longname': 'removeBindingSet',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'rbs'},
                                  'removeView': {'args': '<type __builtin__.str>',
                                                 'longname': 'removeView',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rv'},
                                  'rootTransform': {'args': '<type __builtin__.bool>',
                                                    'longname': 'rootTransform',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 0,
                                                    'shortname': 'rtn'},
                                  'save': {'args': '<type __builtin__.bool>',
                                           'longname': 'save',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 's'},
                                  'searchPath': {'args': '<type __builtin__.str>',
                                                 'longname': 'searchPath',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sp'},
                                  'silent': {'args': '<type __builtin__.bool>',
                                             'longname': 'silent',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 0,
                                             'shortname': 'si'},
                                  'templateList': {'args': '<type __builtin__.str>',
                                                   'longname': 'templateList',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'tl'},
                                  'unload': {'args': '<type __builtin__.bool>',
                                             'longname': 'unload',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'u'},
                                  'updateBindingSet': {'args': '<type __builtin__.str>',
                                                       'longname': 'updateBindingSet',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ubs'},
                                  'useHierarchy': {'args': '<type __builtin__.bool>',
                                                   'longname': 'useHierarchy',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'uh'},
                                  'viewList': {'args': '<type __builtin__.str>',
                                               'longname': 'viewList',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'vl'}},
                        'shortFlags': {'abs': 'addBindingSet',
                                       'ak': 'allKeyable',
                                       'al': 'attributeList',
                                       'an': 'addNames',
                                       'at': 'attribute',
                                       'av': 'addView',
                                       'bn': 'baseName',
                                       'bsl': 'bindingSetList',
                                       'can': 'childAnchor',
                                       'd': 'delete',
                                       'e': 'edit',
                                       'ec': 'expandCompounds',
                                       'ex': 'exists',
                                       'f': 'force',
                                       'fc': 'fromContainer',
                                       'fn': 'fileName',
                                       'fs': 'fromSelection',
                                       'l': 'load',
                                       'lm': 'layoutMode',
                                       'mf': 'matchFile',
                                       'mn': 'matchName',
                                       'pan': 'parentAnchor',
                                       'pnl': 'publishedNodeList',
                                       'q': 'query',
                                       'rbs': 'removeBindingSet',
                                       'rtn': 'rootTransform',
                                       'rv': 'removeView',
                                       's': 'save',
                                       'si': 'silent',
                                       'sp': 'searchPath',
                                       'tl': 'templateList',
                                       'u': 'unload',
                                       'ubs': 'updateBindingSet',
                                       'uh': 'useHierarchy',
                                       'vl': 'viewList'},
                        'type': 'general'},
  'containerView': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'itemInfo': {'args': '<type __builtin__.str>',
                                           'longname': 'itemInfo',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 'ii'},
                              'itemList': {'args': '<type __builtin__.bool>',
                                           'longname': 'itemList',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'il'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'viewDescription': {'args': '<type __builtin__.bool>',
                                                  'longname': 'viewDescription',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'vd'},
                              'viewLabel': {'args': '<type __builtin__.bool>',
                                            'longname': 'viewLabel',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'vb'},
                              'viewList': {'args': '<type __builtin__.bool>',
                                           'longname': 'viewList',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'vl'},
                              'viewName': {'args': '<type __builtin__.str>',
                                           'longname': 'viewName',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 'vn'}},
                    'shortFlags': {'e': 'edit',
                                   'ii': 'itemInfo',
                                   'il': 'itemList',
                                   'q': 'query',
                                   'vb': 'viewLabel',
                                   'vd': 'viewDescription',
                                   'vl': 'viewList',
                                   'vn': 'viewName'},
                    'type': 'general'},
  'contentBrowser': {'flags': {'addContentPath': {'args': '<type __builtin__.str>',
                                                  'longname': 'addContentPath',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'acp'},
                               'context': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'context',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 3,
                                           'shortname': 'ctx'},
                               'control': {'args': '<type __builtin__.bool>',
                                           'longname': 'control',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'ctl'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'filter': {'args': '<type __builtin__.str>',
                                          'longname': 'filter',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'f'},
                               'forceMainConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'forceMainConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'fmc'},
                               'highlightConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'highlightConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'hlc'},
                               'location': {'args': '<type __builtin__.str>',
                                            'longname': 'location',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'loc'},
                               'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                      'longname': 'lockMainConnection',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'lck'},
                               'mainListConnection': {'args': '<type __builtin__.str>',
                                                      'longname': 'mainListConnection',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'mlc'},
                               'panel': {'args': '<type __builtin__.str>',
                                         'longname': 'panel',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'pnl'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'preview': {'args': '<type __builtin__.bool>',
                                           'longname': 'preview',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'pv'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'refreshTreeView': {'args': '<type __builtin__.bool>',
                                                   'longname': 'refreshTreeView',
                                                   'modes': ['edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'rtv'},
                               'removeContentPath': {'args': '<type __builtin__.str>',
                                                     'longname': 'removeContentPath',
                                                     'modes': ['edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'rcp'},
                               'saveCurrentContext': {'args': '<type __builtin__.bool>',
                                                      'longname': 'saveCurrentContext',
                                                      'modes': ['edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'scc'},
                               'selectionConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'selectionConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'slc'},
                               'stateString': {'args': '<type __builtin__.bool>',
                                               'longname': 'stateString',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'sts'},
                               'thumbnailView': {'args': '<type __builtin__.bool>',
                                                 'longname': 'thumbnailView',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'th'},
                               'treeView': {'args': '<type __builtin__.bool>',
                                            'longname': 'treeView',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'tr'},
                               'unParent': {'args': '<type __builtin__.bool>',
                                            'longname': 'unParent',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'up'},
                               'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                        'longname': 'unlockMainConnection',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'ulk'},
                               'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                        'longname': 'updateMainConnection',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'upd'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'}},
                     'shortFlags': {'acp': 'addContentPath',
                                    'ctl': 'control',
                                    'ctx': 'context',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'f': 'filter',
                                    'fmc': 'forceMainConnection',
                                    'hlc': 'highlightConnection',
                                    'lck': 'lockMainConnection',
                                    'loc': 'location',
                                    'mlc': 'mainListConnection',
                                    'p': 'parent',
                                    'pnl': 'panel',
                                    'pv': 'preview',
                                    'q': 'query',
                                    'rcp': 'removeContentPath',
                                    'rtv': 'refreshTreeView',
                                    'scc': 'saveCurrentContext',
                                    'slc': 'selectionConnection',
                                    'sts': 'stateString',
                                    'th': 'thumbnailView',
                                    'tr': 'treeView',
                                    'ulk': 'unlockMainConnection',
                                    'up': 'unParent',
                                    'upd': 'updateMainConnection',
                                    'ut': 'useTemplate'},
                     'type': 'windows'},
  'contextInfo': {'flags': {'apiImage1': {'args': '<type __builtin__.str>',
                                          'longname': 'apiImage1',
                                          'numArgs': 1,
                                          'shortname': 'ip1'},
                            'c': {'args': '<type __builtin__.bool>',
                                  'longname': 'c',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'c'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'escapeContext': {'args': '<type __builtin__.bool>',
                                              'longname': 'escapeContext',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'esc'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'image1': {'args': '<type __builtin__.bool>',
                                       'longname': 'image1',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'i1'},
                            'image2': {'args': '<type __builtin__.bool>',
                                       'longname': 'image2',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'i2'},
                            'image3': {'args': '<type __builtin__.bool>',
                                       'longname': 'image3',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'i3'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'title': {'args': '<type __builtin__.bool>',
                                      'longname': 'title',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 't'}},
                  'removedFlags': {'class': 'c'},
                  'shortFlags': {'c': 'c',
                                 'e': 'edit',
                                 'esc': 'escapeContext',
                                 'ex': 'exists',
                                 'i1': 'image1',
                                 'i2': 'image2',
                                 'i3': 'image3',
                                 'ip1': 'apiImage1',
                                 'q': 'query',
                                 't': 'title'},
                  'type': 'general'},
  'control': {'callbackFlags': ['visibleChangeCommand',
                                'vcc',
                                'dragCallback',
                                'dgc',
                                'dropCallback',
                                'dpc'],
              'flags': {'annotation': {'args': '<type __builtin__.str>',
                                       'longname': 'annotation',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ann'},
                        'backgroundColor': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'backgroundColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'bgc'},
                        'defineTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'defineTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'dt'},
                        'docTag': {'args': '<type __builtin__.str>',
                                   'longname': 'docTag',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'dtg'},
                        'dragCallback': {'args': 'script',
                                         'longname': 'dragCallback',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dgc'},
                        'dropCallback': {'args': 'script',
                                         'longname': 'dropCallback',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dpc'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'enable': {'args': '<type __builtin__.bool>',
                                   'longname': 'enable',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'en'},
                        'enableBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'enableBackground',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ebg'},
                        'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableKeyboardFocus',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ekf'},
                        'exists': {'args': '<type __builtin__.bool>',
                                   'longname': 'exists',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'ex'},
                        'fullPathName': {'args': '<type __builtin__.bool>',
                                         'longname': 'fullPathName',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'fpn'},
                        'height': {'args': '<type __builtin__.int>',
                                   'longname': 'height',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'h'},
                        'highlightColor': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'highlightColor',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'hlc'},
                        'isObscured': {'args': '<type __builtin__.bool>',
                                       'longname': 'isObscured',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'io'},
                        'manage': {'args': '<type __builtin__.bool>',
                                   'longname': 'manage',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'm'},
                        'noBackground': {'args': '<type __builtin__.bool>',
                                         'longname': 'noBackground',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nbg'},
                        'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                               'longname': 'numberOfPopupMenus',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'npm'},
                        'parent': {'args': '<type __builtin__.str>',
                                   'longname': 'parent',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'p'},
                        'popupMenuArray': {'args': '<type __builtin__.bool>',
                                           'longname': 'popupMenuArray',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'pma'},
                        'preventOverride': {'args': '<type __builtin__.bool>',
                                            'longname': 'preventOverride',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'po'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'statusBarMessage': {'args': '<type __builtin__.str>',
                                             'longname': 'statusBarMessage',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sbm'},
                        'useTemplate': {'args': '<type __builtin__.str>',
                                        'longname': 'useTemplate',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ut'},
                        'visible': {'args': '<type __builtin__.bool>',
                                    'longname': 'visible',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'vis'},
                        'visibleChangeCommand': {'args': 'script',
                                                 'longname': 'visibleChangeCommand',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'vcc'},
                        'width': {'args': '<type __builtin__.int>',
                                  'longname': 'width',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'w'}},
              'shortFlags': {'ann': 'annotation',
                             'bgc': 'backgroundColor',
                             'dgc': 'dragCallback',
                             'dpc': 'dropCallback',
                             'dt': 'defineTemplate',
                             'dtg': 'docTag',
                             'e': 'edit',
                             'ebg': 'enableBackground',
                             'ekf': 'enableKeyboardFocus',
                             'en': 'enable',
                             'ex': 'exists',
                             'fpn': 'fullPathName',
                             'h': 'height',
                             'hlc': 'highlightColor',
                             'io': 'isObscured',
                             'm': 'manage',
                             'nbg': 'noBackground',
                             'npm': 'numberOfPopupMenus',
                             'p': 'parent',
                             'pma': 'popupMenuArray',
                             'po': 'preventOverride',
                             'q': 'query',
                             'sbm': 'statusBarMessage',
                             'ut': 'useTemplate',
                             'vcc': 'visibleChangeCommand',
                             'vis': 'visible',
                             'w': 'width'},
              'type': 'windows'},
  'controller': {'flags': {'allControllers': {'args': '<type __builtin__.bool>',
                                              'longname': 'allControllers',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 0,
                                              'shortname': 'ac'},
                           'children': {'args': '<type __builtin__.bool>',
                                        'longname': 'children',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'cld'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'group': {'args': '<type __builtin__.bool>',
                                     'longname': 'group',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'g'},
                           'index': {'args': '<type __builtin__.int>',
                                     'longname': 'index',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'idx'},
                           'isController': {'args': '<type __builtin__.str>',
                                            'longname': 'isController',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ic'},
                           'parent': {'args': '<type __builtin__.bool>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'p'},
                           'pickWalkDown': {'args': '<type __builtin__.bool>',
                                            'longname': 'pickWalkDown',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'pwd'},
                           'pickWalkLeft': {'args': '<type __builtin__.bool>',
                                            'longname': 'pickWalkLeft',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'pwl'},
                           'pickWalkRight': {'args': '<type __builtin__.bool>',
                                             'longname': 'pickWalkRight',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'pwr'},
                           'pickWalkUp': {'args': '<type __builtin__.bool>',
                                          'longname': 'pickWalkUp',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'pwu'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'unparent': {'args': '<type __builtin__.bool>',
                                        'longname': 'unparent',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'unp'}},
                 'shortFlags': {'ac': 'allControllers',
                                'cld': 'children',
                                'e': 'edit',
                                'g': 'group',
                                'ic': 'isController',
                                'idx': 'index',
                                'p': 'parent',
                                'pwd': 'pickWalkDown',
                                'pwl': 'pickWalkLeft',
                                'pwr': 'pickWalkRight',
                                'pwu': 'pickWalkUp',
                                'q': 'query',
                                'unp': 'unparent'},
                 'type': 'animation'},
  'convertIffToPsd': {'flags': {'iffFileName': {'args': '<type __builtin__.str>',
                                                'longname': 'iffFileName',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'ifn'},
                                'psdFileName': {'args': '<type __builtin__.str>',
                                                'longname': 'psdFileName',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'pfn'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'xResolution': {'args': '<type __builtin__.int>',
                                                'longname': 'xResolution',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'xr'},
                                'yResolution': {'args': '<type __builtin__.int>',
                                                'longname': 'yResolution',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'yr'}},
                      'shortFlags': {'ifn': 'iffFileName',
                                     'pfn': 'psdFileName',
                                     'q': 'query',
                                     'xr': 'xResolution',
                                     'yr': 'yResolution'},
                      'type': 'rendering'},
  'convertSolidTx': {'flags': {'alpha': {'args': '<type __builtin__.bool>',
                                         'longname': 'alpha',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'al'},
                               'antiAlias': {'args': '<type __builtin__.bool>',
                                             'longname': 'antiAlias',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'aa'},
                               'backgroundColor': {'args': ['<type __builtin__.int>',
                                                            '<type __builtin__.int>',
                                                            '<type __builtin__.int>'],
                                                   'longname': 'backgroundColor',
                                                   'modes': ['create'],
                                                   'numArgs': 3,
                                                   'shortname': 'bc'},
                               'backgroundMode': {'args': '<type __builtin__.str>',
                                                  'longname': 'backgroundMode',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'bm'},
                               'camera': {'args': 'PyNode',
                                          'longname': 'camera',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'cam'},
                               'componentRange': {'args': '<type __builtin__.bool>',
                                                  'longname': 'componentRange',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'cr'},
                               'doubleSided': {'args': '<type __builtin__.bool>',
                                               'longname': 'doubleSided',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ds'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'fileFormat': {'args': '<type __builtin__.str>',
                                              'longname': 'fileFormat',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'fil'},
                               'fileImageName': {'args': '<type __builtin__.str>',
                                                 'longname': 'fileImageName',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'fin'},
                               'fillTextureSeams': {'args': '<type __builtin__.bool>',
                                                    'longname': 'fillTextureSeams',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'fts'},
                               'force': {'args': '<type __builtin__.bool>',
                                         'longname': 'force',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'f'},
                               'fullUvRange': {'args': '<type __builtin__.bool>',
                                               'longname': 'fullUvRange',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'fur'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'pixelFormat': {'args': '<type __builtin__.str>',
                                               'longname': 'pixelFormat',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'pf'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'resolutionX': {'args': '<type __builtin__.int>',
                                               'longname': 'resolutionX',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'rx'},
                               'resolutionY': {'args': '<type __builtin__.int>',
                                               'longname': 'resolutionY',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ry'},
                               'reuseDepthMap': {'args': '<type __builtin__.bool>',
                                                 'longname': 'reuseDepthMap',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'rdm'},
                               'samplePlane': {'args': '<type __builtin__.bool>',
                                               'longname': 'samplePlane',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'sp'},
                               'samplePlaneRange': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'samplePlaneRange',
                                                    'modes': ['create'],
                                                    'numArgs': 4,
                                                    'shortname': 'spr'},
                               'shadows': {'args': '<type __builtin__.bool>',
                                           'longname': 'shadows',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'sh'},
                               'uvBBoxIntersect': {'args': '<type __builtin__.bool>',
                                                   'longname': 'uvBBoxIntersect',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ubi'},
                               'uvRange': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'uvRange',
                                           'modes': ['create'],
                                           'numArgs': 4,
                                           'shortname': 'uvr'},
                               'uvSetName': {'args': '<type __builtin__.str>',
                                             'longname': 'uvSetName',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'uv'}},
                     'shortFlags': {'aa': 'antiAlias',
                                    'al': 'alpha',
                                    'bc': 'backgroundColor',
                                    'bm': 'backgroundMode',
                                    'cam': 'camera',
                                    'cr': 'componentRange',
                                    'ds': 'doubleSided',
                                    'e': 'edit',
                                    'f': 'force',
                                    'fil': 'fileFormat',
                                    'fin': 'fileImageName',
                                    'fts': 'fillTextureSeams',
                                    'fur': 'fullUvRange',
                                    'n': 'name',
                                    'pf': 'pixelFormat',
                                    'q': 'query',
                                    'rdm': 'reuseDepthMap',
                                    'rx': 'resolutionX',
                                    'ry': 'resolutionY',
                                    'sh': 'shadows',
                                    'sp': 'samplePlane',
                                    'spr': 'samplePlaneRange',
                                    'ubi': 'uvBBoxIntersect',
                                    'uv': 'uvSetName',
                                    'uvr': 'uvRange'},
                     'type': 'rendering'},
  'convertTessellation': {'flags': {'allCameras': {'args': '<type __builtin__.bool>',
                                                   'longname': 'allCameras',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'acm'},
                                    'camera': {'args': 'PyNode',
                                               'longname': 'camera',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'cam'}},
                          'shortFlags': {'acm': 'allCameras',
                                         'cam': 'camera'},
                          'type': 'rendering'},
  'convertUnit': {'flags': {'fromUnit': {'args': '<type __builtin__.str>',
                                         'longname': 'fromUnit',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'f'},
                            'toUnit': {'args': '<type __builtin__.str>',
                                       'longname': 'toUnit',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 't'}},
                  'shortFlags': {'f': 'fromUnit', 't': 'toUnit'},
                  'type': 'system'},
  'copyAttr': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                       'longname': 'attribute',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'at'},
                         'containerParentChild': {'args': '<type __builtin__.bool>',
                                                  'longname': 'containerParentChild',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'cpc'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'inConnections': {'args': '<type __builtin__.bool>',
                                           'longname': 'inConnections',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ic'},
                         'keepSourceConnections': {'args': '<type __builtin__.bool>',
                                                   'longname': 'keepSourceConnections',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'ksc'},
                         'outConnections': {'args': '<type __builtin__.bool>',
                                            'longname': 'outConnections',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'oc'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'renameTargetContainer': {'args': '<type __builtin__.bool>',
                                                   'longname': 'renameTargetContainer',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'rtc'},
                         'values': {'args': '<type __builtin__.bool>',
                                    'longname': 'values',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'v'}},
               'shortFlags': {'at': 'attribute',
                              'cpc': 'containerParentChild',
                              'e': 'edit',
                              'ic': 'inConnections',
                              'ksc': 'keepSourceConnections',
                              'oc': 'outConnections',
                              'q': 'query',
                              'rtc': 'renameTargetContainer',
                              'v': 'values'},
               'type': 'general'},
  'copyDeformerWeights': {'flags': {'destinationDeformer': {'args': '<type __builtin__.str>',
                                                            'longname': 'destinationDeformer',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'dd'},
                                    'destinationShape': {'args': '<type __builtin__.str>',
                                                         'longname': 'destinationShape',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ds'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'mirrorInverse': {'args': '<type __builtin__.bool>',
                                                      'longname': 'mirrorInverse',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'mi'},
                                    'mirrorMode': {'args': '<type __builtin__.str>',
                                                   'longname': 'mirrorMode',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'mm'},
                                    'noMirror': {'args': '<type __builtin__.bool>',
                                                 'longname': 'noMirror',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'nm'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'smooth': {'args': '<type __builtin__.bool>',
                                               'longname': 'smooth',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 0,
                                               'shortname': 'sm'},
                                    'sourceDeformer': {'args': '<type __builtin__.str>',
                                                       'longname': 'sourceDeformer',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'sd'},
                                    'sourceShape': {'args': '<type __builtin__.str>',
                                                    'longname': 'sourceShape',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ss'},
                                    'surfaceAssociation': {'args': '<type __builtin__.str>',
                                                           'longname': 'surfaceAssociation',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'sa'},
                                    'uvSpace': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'uvSpace',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'uv'}},
                          'shortFlags': {'dd': 'destinationDeformer',
                                         'ds': 'destinationShape',
                                         'e': 'edit',
                                         'mi': 'mirrorInverse',
                                         'mm': 'mirrorMode',
                                         'nm': 'noMirror',
                                         'q': 'query',
                                         'sa': 'surfaceAssociation',
                                         'sd': 'sourceDeformer',
                                         'sm': 'smooth',
                                         'ss': 'sourceShape',
                                         'uv': 'uvSpace'},
                          'type': 'animation'},
  'copyFlexor': {'flags': {}, 'shortFlags': {}, 'type': 'animation'},
  'copyKey': {'flags': {'animLayer': {'args': '<type __builtin__.str>',
                                      'longname': 'animLayer',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'al'},
                        'animation': {'args': '<type __builtin__.str>',
                                      'longname': 'animation',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'an'},
                        'attribute': {'args': '<type __builtin__.str>',
                                      'longname': 'attribute',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'at'},
                        'clipboard': {'args': '<type __builtin__.str>',
                                      'longname': 'clipboard',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'cb'},
                        'controlPoints': {'args': '<type __builtin__.bool>',
                                          'longname': 'controlPoints',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'cp'},
                        'float': {'args': 'floatrange',
                                  'longname': 'float',
                                  'modes': ['create', 'multiuse'],
                                  'numArgs': 1,
                                  'shortname': 'f'},
                        'forceIndependentEulerAngles': {'args': '<type __builtin__.bool>',
                                                        'longname': 'forceIndependentEulerAngles',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'fea'},
                        'hierarchy': {'args': '<type __builtin__.str>',
                                      'longname': 'hierarchy',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'hi'},
                        'includeUpperBound': {'args': '<type __builtin__.bool>',
                                              'longname': 'includeUpperBound',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'iub'},
                        'index': {'args': '<type __builtin__.int>',
                                  'longname': 'index',
                                  'modes': ['create', 'multiuse'],
                                  'numArgs': 1,
                                  'shortname': 'index'},
                        'option': {'args': '<type __builtin__.str>',
                                   'longname': 'option',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'o'},
                        'shape': {'args': '<type __builtin__.bool>',
                                  'longname': 'shape',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 's'},
                        'time': {'args': 'timerange',
                                 'longname': 'time',
                                 'modes': ['create', 'multiuse'],
                                 'numArgs': 1,
                                 'shortname': 't'}},
              'removedFlags': {'in': 'index'},
              'shortFlags': {'al': 'animLayer',
                             'an': 'animation',
                             'at': 'attribute',
                             'cb': 'clipboard',
                             'cp': 'controlPoints',
                             'f': 'float',
                             'fea': 'forceIndependentEulerAngles',
                             'hi': 'hierarchy',
                             'index': 'index',
                             'iub': 'includeUpperBound',
                             'o': 'option',
                             's': 'shape',
                             't': 'time'},
              'type': 'animation'},
  'copyNode': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'copySkinWeights': {'flags': {'destinationSkin': {'args': '<type __builtin__.str>',
                                                    'longname': 'destinationSkin',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ds'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'influenceAssociation': {'args': '<type __builtin__.str>',
                                                         'longname': 'influenceAssociation',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit',
                                                                   'multiuse'],
                                                         'numArgs': 1,
                                                         'shortname': 'ia'},
                                'mirrorInverse': {'args': '<type __builtin__.bool>',
                                                  'longname': 'mirrorInverse',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'mi'},
                                'mirrorMode': {'args': '<type __builtin__.str>',
                                               'longname': 'mirrorMode',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'mm'},
                                'noBlendWeight': {'args': '<type __builtin__.bool>',
                                                  'longname': 'noBlendWeight',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'nbw'},
                                'noMirror': {'args': '<type __builtin__.bool>',
                                             'longname': 'noMirror',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 0,
                                             'shortname': 'nm'},
                                'normalize': {'args': '<type __builtin__.bool>',
                                              'longname': 'normalize',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 0,
                                              'shortname': 'nr'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'sampleSpace': {'args': '<type __builtin__.int>',
                                                'longname': 'sampleSpace',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'spa'},
                                'smooth': {'args': '<type __builtin__.bool>',
                                           'longname': 'smooth',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 0,
                                           'shortname': 'sm'},
                                'sourceSkin': {'args': '<type __builtin__.str>',
                                               'longname': 'sourceSkin',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ss'},
                                'surfaceAssociation': {'args': '<type __builtin__.str>',
                                                       'longname': 'surfaceAssociation',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'sa'},
                                'uvSpace': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'uvSpace',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 2,
                                            'shortname': 'uv'}},
                      'shortFlags': {'ds': 'destinationSkin',
                                     'e': 'edit',
                                     'ia': 'influenceAssociation',
                                     'mi': 'mirrorInverse',
                                     'mm': 'mirrorMode',
                                     'nbw': 'noBlendWeight',
                                     'nm': 'noMirror',
                                     'nr': 'normalize',
                                     'q': 'query',
                                     'sa': 'surfaceAssociation',
                                     'sm': 'smooth',
                                     'spa': 'sampleSpace',
                                     'ss': 'sourceSkin',
                                     'uv': 'uvSpace'},
                      'type': 'animation'},
  'crashInfo': {'flags': {'crashFile': {'args': '<type __builtin__.bool>',
                                        'longname': 'crashFile',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'cf'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'savedBeforeCrash': {'args': '<type __builtin__.bool>',
                                               'longname': 'savedBeforeCrash',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'sbc'}},
                'shortFlags': {'cf': 'crashFile',
                               'q': 'query',
                               'sbc': 'savedBeforeCrash'},
                'type': 'system'},
  'crashInfoCmd': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'createAttrPatterns': {'flags': {'patternDefinition': {'args': '<type __builtin__.str>',
                                                         'longname': 'patternDefinition',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'pd'},
                                   'patternFile': {'args': '<type __builtin__.str>',
                                                   'longname': 'patternFile',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'pf'},
                                   'patternType': {'args': '<type __builtin__.str>',
                                                   'longname': 'patternType',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'pt'}},
                         'shortFlags': {'pd': 'patternDefinition',
                                        'pf': 'patternFile',
                                        'pt': 'patternType'},
                         'type': 'general'},
  'createDisplayLayer': {'flags': {'empty': {'args': '<type __builtin__.bool>',
                                             'longname': 'empty',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                   'makeCurrent': {'args': '<type __builtin__.bool>',
                                                   'longname': 'makeCurrent',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'mc'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'noRecurse': {'args': '<type __builtin__.bool>',
                                                 'longname': 'noRecurse',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'nr'},
                                   'number': {'args': '<type __builtin__.int>',
                                              'longname': 'number',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'num'}},
                         'shortFlags': {'e': 'empty',
                                        'mc': 'makeCurrent',
                                        'n': 'name',
                                        'nr': 'noRecurse',
                                        'num': 'number'},
                         'type': 'general'},
  'createEditor': {'flags': {'noCloseOnDelete': {'args': '<type __builtin__.bool>',
                                                 'longname': 'noCloseOnDelete',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'nc'},
                             'queueForDelete': {'args': '<type __builtin__.bool>',
                                                'longname': 'queueForDelete',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'qfd'}},
                   'shortFlags': {'nc': 'noCloseOnDelete',
                                  'qfd': 'queueForDelete'},
                   'type': 'windows'},
  'createLayeredPsdFile': {'flags': {'imageFileName': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'imageFileName',
                                                       'modes': ['create',
                                                                 'multiuse'],
                                                       'numArgs': 3,
                                                       'shortname': 'ifn'},
                                     'psdFileName': {'args': '<type __builtin__.str>',
                                                     'longname': 'psdFileName',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'psf'},
                                     'xResolution': {'args': '<type __builtin__.int>',
                                                     'longname': 'xResolution',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'xr'},
                                     'yResolution': {'args': '<type __builtin__.int>',
                                                     'longname': 'yResolution',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'yr'}},
                           'shortFlags': {},
                           'type': 'rendering'},
  'createNode': {'flags': {'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'shared': {'args': '<type __builtin__.bool>',
                                      'longname': 'shared',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 's'},
                           'skipSelect': {'args': '<type __builtin__.bool>',
                                          'longname': 'skipSelect',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ss'}},
                 'shortFlags': {'n': 'name',
                                'p': 'parent',
                                's': 'shared',
                                'ss': 'skipSelect'},
                 'type': 'general'},
  'createNurbsCircleCtx': {'flags': {'attachToSections': {'args': '<type __builtin__.bool>',
                                                          'longname': 'attachToSections',
                                                          'numArgs': 1,
                                                          'shortname': 'attachToSections'},
                                     'degree': {'args': '<type __builtin__.int>',
                                                'longname': 'degree',
                                                'numArgs': 1,
                                                'shortname': 'd'},
                                     'doDragEdit': {'args': '<type __builtin__.bool>',
                                                    'longname': 'doDragEdit',
                                                    'numArgs': 1,
                                                    'shortname': 'dde'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'exists': {'args': '<type __builtin__.bool>',
                                                'longname': 'exists',
                                                'numArgs': 0,
                                                'shortname': 'ex'},
                                     'history': {'args': '<type __builtin__.bool>',
                                                 'longname': 'history',
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                                     'image1': {'args': '<type __builtin__.str>',
                                                'longname': 'image1',
                                                'numArgs': 1,
                                                'shortname': 'i1'},
                                     'image2': {'args': '<type __builtin__.str>',
                                                'longname': 'image2',
                                                'numArgs': 1,
                                                'shortname': 'i2'},
                                     'image3': {'args': '<type __builtin__.str>',
                                                'longname': 'image3',
                                                'numArgs': 1,
                                                'shortname': 'i3'},
                                     'name': {'args': '<type __builtin__.str>',
                                              'longname': 'name',
                                              'numArgs': 1,
                                              'shortname': 'n'},
                                     'normal': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'normal',
                                                'numArgs': 3,
                                                'shortname': 'nr'},
                                     'normalType': {'args': '<type __builtin__.int>',
                                                    'longname': 'normalType',
                                                    'numArgs': 1,
                                                    'shortname': 'nrt'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'radius': {'args': '<type __builtin__.float>',
                                                'longname': 'radius',
                                                'numArgs': 1,
                                                'shortname': 'r'},
                                     'sections': {'args': '<type __builtin__.int>',
                                                  'longname': 'sections',
                                                  'numArgs': 1,
                                                  'shortname': 'sc'},
                                     'sweep': {'args': '<type __builtin__.float>',
                                               'longname': 'sweep',
                                               'numArgs': 1,
                                               'shortname': 'sw'},
                                     'tolerance': {'args': '<type __builtin__.float>',
                                                   'longname': 'tolerance',
                                                   'numArgs': 1,
                                                   'shortname': 'tol'},
                                     'toleranceType': {'args': '<type __builtin__.int>',
                                                       'longname': 'toleranceType',
                                                       'numArgs': 1,
                                                       'shortname': 'tlt'},
                                     'useTolerance': {'args': '<type __builtin__.bool>',
                                                      'longname': 'useTolerance',
                                                      'numArgs': 1,
                                                      'shortname': 'ut'}},
                           'removedFlags': {'as': 'attachToSections'},
                           'shortFlags': {'attachToSections': 'attachToSections',
                                          'ch': 'history',
                                          'd': 'degree',
                                          'dde': 'doDragEdit',
                                          'e': 'edit',
                                          'ex': 'exists',
                                          'i1': 'image1',
                                          'i2': 'image2',
                                          'i3': 'image3',
                                          'n': 'name',
                                          'nr': 'normal',
                                          'nrt': 'normalType',
                                          'q': 'query',
                                          'r': 'radius',
                                          'sc': 'sections',
                                          'sw': 'sweep',
                                          'tlt': 'toleranceType',
                                          'tol': 'tolerance',
                                          'ut': 'useTolerance'},
                           'type': 'context'},
  'createNurbsConeCtx': {'flags': {'attachToHeightRatio': {'args': '<type __builtin__.bool>',
                                                           'longname': 'attachToHeightRatio',
                                                           'numArgs': 1,
                                                           'shortname': 'ahr'},
                                   'attachToSections': {'args': '<type __builtin__.bool>',
                                                        'longname': 'attachToSections',
                                                        'numArgs': 1,
                                                        'shortname': 'attachToSections'},
                                   'attachToSpans': {'args': '<type __builtin__.bool>',
                                                     'longname': 'attachToSpans',
                                                     'numArgs': 1,
                                                     'shortname': 'asp'},
                                   'axis': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'axis',
                                            'numArgs': 3,
                                            'shortname': 'ax'},
                                   'axisType': {'args': '<type __builtin__.int>',
                                                'longname': 'axisType',
                                                'numArgs': 1,
                                                'shortname': 'axt'},
                                   'caps': {'args': '<type __builtin__.int>',
                                            'longname': 'caps',
                                            'numArgs': 1,
                                            'shortname': 'cp'},
                                   'doDragEdit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'doDragEdit',
                                                  'numArgs': 1,
                                                  'shortname': 'dde'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'endSweep': {'args': '<type __builtin__.float>',
                                                'longname': 'endSweep',
                                                'numArgs': 1,
                                                'shortname': 'esw'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'extraTransformOnCaps': {'args': '<type __builtin__.bool>',
                                                            'longname': 'extraTransformOnCaps',
                                                            'numArgs': 1,
                                                            'shortname': 'xtc'},
                                   'height': {'args': '<type __builtin__.float>',
                                              'longname': 'height',
                                              'numArgs': 1,
                                              'shortname': 'h'},
                                   'history': {'args': '<type __builtin__.bool>',
                                               'longname': 'history',
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'radius': {'args': '<type __builtin__.float>',
                                              'longname': 'radius',
                                              'numArgs': 1,
                                              'shortname': 'r'},
                                   'sections': {'args': '<type __builtin__.int>',
                                                'longname': 'sections',
                                                'numArgs': 1,
                                                'shortname': 'sc'},
                                   'spans': {'args': '<type __builtin__.int>',
                                             'longname': 'spans',
                                             'numArgs': 1,
                                             'shortname': 'sp'},
                                   'startSweep': {'args': '<type __builtin__.float>',
                                                  'longname': 'startSweep',
                                                  'numArgs': 1,
                                                  'shortname': 'ssw'},
                                   'surfaceDegree': {'args': '<type __builtin__.int>',
                                                     'longname': 'surfaceDegree',
                                                     'numArgs': 1,
                                                     'shortname': 'sd'},
                                   'tolerance': {'args': '<type __builtin__.float>',
                                                 'longname': 'tolerance',
                                                 'numArgs': 1,
                                                 'shortname': 'tol'},
                                   'toleranceType': {'args': '<type __builtin__.int>',
                                                     'longname': 'toleranceType',
                                                     'numArgs': 1,
                                                     'shortname': 'tlt'},
                                   'useTolerance': {'args': '<type __builtin__.bool>',
                                                    'longname': 'useTolerance',
                                                    'numArgs': 1,
                                                    'shortname': 'ut'}},
                         'removedFlags': {'as': 'attachToSections'},
                         'shortFlags': {'ahr': 'attachToHeightRatio',
                                        'asp': 'attachToSpans',
                                        'attachToSections': 'attachToSections',
                                        'ax': 'axis',
                                        'axt': 'axisType',
                                        'ch': 'history',
                                        'cp': 'caps',
                                        'dde': 'doDragEdit',
                                        'e': 'edit',
                                        'esw': 'endSweep',
                                        'ex': 'exists',
                                        'h': 'height',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'n': 'name',
                                        'q': 'query',
                                        'r': 'radius',
                                        'sc': 'sections',
                                        'sd': 'surfaceDegree',
                                        'sp': 'spans',
                                        'ssw': 'startSweep',
                                        'tlt': 'toleranceType',
                                        'tol': 'tolerance',
                                        'ut': 'useTolerance',
                                        'xtc': 'extraTransformOnCaps'},
                         'type': 'context'},
  'createNurbsCubeCtx': {'flags': {'attachToPatchesU': {'args': '<type __builtin__.bool>',
                                                        'longname': 'attachToPatchesU',
                                                        'numArgs': 1,
                                                        'shortname': 'apu'},
                                   'attachToPatchesV': {'args': '<type __builtin__.bool>',
                                                        'longname': 'attachToPatchesV',
                                                        'numArgs': 1,
                                                        'shortname': 'apv'},
                                   'axis': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'axis',
                                            'numArgs': 3,
                                            'shortname': 'ax'},
                                   'axisType': {'args': '<type __builtin__.int>',
                                                'longname': 'axisType',
                                                'numArgs': 1,
                                                'shortname': 'axt'},
                                   'depth': {'args': '<type __builtin__.float>',
                                             'longname': 'depth',
                                             'numArgs': 1,
                                             'shortname': 'd'},
                                   'doDragEdit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'doDragEdit',
                                                  'numArgs': 1,
                                                  'shortname': 'dde'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'height': {'args': '<type __builtin__.float>',
                                              'longname': 'height',
                                              'numArgs': 1,
                                              'shortname': 'h'},
                                   'history': {'args': '<type __builtin__.bool>',
                                               'longname': 'history',
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'surfaceDegree': {'args': '<type __builtin__.int>',
                                                     'longname': 'surfaceDegree',
                                                     'numArgs': 1,
                                                     'shortname': 'sd'},
                                   'uPatches': {'args': '<type __builtin__.int>',
                                                'longname': 'uPatches',
                                                'numArgs': 1,
                                                'shortname': 'up'},
                                   'vPatches': {'args': '<type __builtin__.int>',
                                                'longname': 'vPatches',
                                                'numArgs': 1,
                                                'shortname': 'vp'},
                                   'width': {'args': '<type __builtin__.float>',
                                             'longname': 'width',
                                             'numArgs': 1,
                                             'shortname': 'w'}},
                         'shortFlags': {'apu': 'attachToPatchesU',
                                        'apv': 'attachToPatchesV',
                                        'ax': 'axis',
                                        'axt': 'axisType',
                                        'ch': 'history',
                                        'd': 'depth',
                                        'dde': 'doDragEdit',
                                        'e': 'edit',
                                        'ex': 'exists',
                                        'h': 'height',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'n': 'name',
                                        'q': 'query',
                                        'sd': 'surfaceDegree',
                                        'up': 'uPatches',
                                        'vp': 'vPatches',
                                        'w': 'width'},
                         'type': 'context'},
  'createNurbsCylinderCtx': {'flags': {'attachToHeightRatio': {'args': '<type __builtin__.bool>',
                                                               'longname': 'attachToHeightRatio',
                                                               'numArgs': 1,
                                                               'shortname': 'ahr'},
                                       'attachToSections': {'args': '<type __builtin__.bool>',
                                                            'longname': 'attachToSections',
                                                            'numArgs': 1,
                                                            'shortname': 'attachToSections'},
                                       'attachToSpans': {'args': '<type __builtin__.bool>',
                                                         'longname': 'attachToSpans',
                                                         'numArgs': 1,
                                                         'shortname': 'asp'},
                                       'axis': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'axis',
                                                'numArgs': 3,
                                                'shortname': 'ax'},
                                       'axisType': {'args': '<type __builtin__.int>',
                                                    'longname': 'axisType',
                                                    'numArgs': 1,
                                                    'shortname': 'axt'},
                                       'caps': {'args': '<type __builtin__.int>',
                                                'longname': 'caps',
                                                'numArgs': 1,
                                                'shortname': 'cp'},
                                       'doDragEdit': {'args': '<type __builtin__.bool>',
                                                      'longname': 'doDragEdit',
                                                      'numArgs': 1,
                                                      'shortname': 'dde'},
                                       'edit': {'args': '<type __builtin__.bool>',
                                                'longname': 'edit',
                                                'numArgs': 0,
                                                'shortname': 'e'},
                                       'endSweep': {'args': '<type __builtin__.float>',
                                                    'longname': 'endSweep',
                                                    'numArgs': 1,
                                                    'shortname': 'esw'},
                                       'exists': {'args': '<type __builtin__.bool>',
                                                  'longname': 'exists',
                                                  'numArgs': 0,
                                                  'shortname': 'ex'},
                                       'extraTransformOnCaps': {'args': '<type __builtin__.bool>',
                                                                'longname': 'extraTransformOnCaps',
                                                                'numArgs': 1,
                                                                'shortname': 'xtc'},
                                       'height': {'args': '<type __builtin__.float>',
                                                  'longname': 'height',
                                                  'numArgs': 1,
                                                  'shortname': 'h'},
                                       'history': {'args': '<type __builtin__.bool>',
                                                   'longname': 'history',
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                                       'image1': {'args': '<type __builtin__.str>',
                                                  'longname': 'image1',
                                                  'numArgs': 1,
                                                  'shortname': 'i1'},
                                       'image2': {'args': '<type __builtin__.str>',
                                                  'longname': 'image2',
                                                  'numArgs': 1,
                                                  'shortname': 'i2'},
                                       'image3': {'args': '<type __builtin__.str>',
                                                  'longname': 'image3',
                                                  'numArgs': 1,
                                                  'shortname': 'i3'},
                                       'name': {'args': '<type __builtin__.str>',
                                                'longname': 'name',
                                                'numArgs': 1,
                                                'shortname': 'n'},
                                       'query': {'args': '<type __builtin__.bool>',
                                                 'longname': 'query',
                                                 'numArgs': 0,
                                                 'shortname': 'q'},
                                       'radius': {'args': '<type __builtin__.float>',
                                                  'longname': 'radius',
                                                  'numArgs': 1,
                                                  'shortname': 'r'},
                                       'sections': {'args': '<type __builtin__.int>',
                                                    'longname': 'sections',
                                                    'numArgs': 1,
                                                    'shortname': 'sc'},
                                       'spans': {'args': '<type __builtin__.int>',
                                                 'longname': 'spans',
                                                 'numArgs': 1,
                                                 'shortname': 'sp'},
                                       'startSweep': {'args': '<type __builtin__.float>',
                                                      'longname': 'startSweep',
                                                      'numArgs': 1,
                                                      'shortname': 'ssw'},
                                       'surfaceDegree': {'args': '<type __builtin__.int>',
                                                         'longname': 'surfaceDegree',
                                                         'numArgs': 1,
                                                         'shortname': 'sd'},
                                       'tolerance': {'args': '<type __builtin__.float>',
                                                     'longname': 'tolerance',
                                                     'numArgs': 1,
                                                     'shortname': 'tol'},
                                       'toleranceType': {'args': '<type __builtin__.int>',
                                                         'longname': 'toleranceType',
                                                         'numArgs': 1,
                                                         'shortname': 'tlt'},
                                       'useTolerance': {'args': '<type __builtin__.bool>',
                                                        'longname': 'useTolerance',
                                                        'numArgs': 1,
                                                        'shortname': 'ut'}},
                             'removedFlags': {'as': 'attachToSections'},
                             'shortFlags': {'ahr': 'attachToHeightRatio',
                                            'asp': 'attachToSpans',
                                            'attachToSections': 'attachToSections',
                                            'ax': 'axis',
                                            'axt': 'axisType',
                                            'ch': 'history',
                                            'cp': 'caps',
                                            'dde': 'doDragEdit',
                                            'e': 'edit',
                                            'esw': 'endSweep',
                                            'ex': 'exists',
                                            'h': 'height',
                                            'i1': 'image1',
                                            'i2': 'image2',
                                            'i3': 'image3',
                                            'n': 'name',
                                            'q': 'query',
                                            'r': 'radius',
                                            'sc': 'sections',
                                            'sd': 'surfaceDegree',
                                            'sp': 'spans',
                                            'ssw': 'startSweep',
                                            'tlt': 'toleranceType',
                                            'tol': 'tolerance',
                                            'ut': 'useTolerance',
                                            'xtc': 'extraTransformOnCaps'},
                             'type': 'context'},
  'createNurbsPlaneCtx': {'flags': {'attachToPatchesU': {'args': '<type __builtin__.bool>',
                                                         'longname': 'attachToPatchesU',
                                                         'numArgs': 1,
                                                         'shortname': 'apu'},
                                    'attachToPatchesV': {'args': '<type __builtin__.bool>',
                                                         'longname': 'attachToPatchesV',
                                                         'numArgs': 1,
                                                         'shortname': 'apv'},
                                    'axis': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'axis',
                                             'numArgs': 3,
                                             'shortname': 'ax'},
                                    'axisType': {'args': '<type __builtin__.int>',
                                                 'longname': 'axisType',
                                                 'numArgs': 1,
                                                 'shortname': 'axt'},
                                    'doDragEdit': {'args': '<type __builtin__.bool>',
                                                   'longname': 'doDragEdit',
                                                   'numArgs': 1,
                                                   'shortname': 'dde'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'height': {'args': '<type __builtin__.float>',
                                               'longname': 'height',
                                               'numArgs': 1,
                                               'shortname': 'h'},
                                    'history': {'args': '<type __builtin__.bool>',
                                                'longname': 'history',
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'surfaceDegree': {'args': '<type __builtin__.int>',
                                                      'longname': 'surfaceDegree',
                                                      'numArgs': 1,
                                                      'shortname': 'sd'},
                                    'uPatches': {'args': '<type __builtin__.int>',
                                                 'longname': 'uPatches',
                                                 'numArgs': 1,
                                                 'shortname': 'up'},
                                    'vPatches': {'args': '<type __builtin__.int>',
                                                 'longname': 'vPatches',
                                                 'numArgs': 1,
                                                 'shortname': 'vp'},
                                    'width': {'args': '<type __builtin__.float>',
                                              'longname': 'width',
                                              'numArgs': 1,
                                              'shortname': 'w'}},
                          'shortFlags': {'apu': 'attachToPatchesU',
                                         'apv': 'attachToPatchesV',
                                         'ax': 'axis',
                                         'axt': 'axisType',
                                         'ch': 'history',
                                         'dde': 'doDragEdit',
                                         'e': 'edit',
                                         'ex': 'exists',
                                         'h': 'height',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'n': 'name',
                                         'q': 'query',
                                         'sd': 'surfaceDegree',
                                         'up': 'uPatches',
                                         'vp': 'vPatches',
                                         'w': 'width'},
                          'type': 'context'},
  'createNurbsSphereCtx': {'flags': {'attachToHeightRatio': {'args': '<type __builtin__.bool>',
                                                             'longname': 'attachToHeightRatio',
                                                             'numArgs': 1,
                                                             'shortname': 'ahr'},
                                     'attachToSections': {'args': '<type __builtin__.bool>',
                                                          'longname': 'attachToSections',
                                                          'numArgs': 1,
                                                          'shortname': 'attachToSections'},
                                     'attachToSpans': {'args': '<type __builtin__.bool>',
                                                       'longname': 'attachToSpans',
                                                       'numArgs': 1,
                                                       'shortname': 'asp'},
                                     'axis': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'axis',
                                              'numArgs': 3,
                                              'shortname': 'ax'},
                                     'axisType': {'args': '<type __builtin__.int>',
                                                  'longname': 'axisType',
                                                  'numArgs': 1,
                                                  'shortname': 'axt'},
                                     'degree': {'args': '<type __builtin__.int>',
                                                'longname': 'degree',
                                                'numArgs': 1,
                                                'shortname': 'd'},
                                     'doDragEdit': {'args': '<type __builtin__.bool>',
                                                    'longname': 'doDragEdit',
                                                    'numArgs': 1,
                                                    'shortname': 'dde'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'endSweep': {'args': '<type __builtin__.float>',
                                                  'longname': 'endSweep',
                                                  'numArgs': 1,
                                                  'shortname': 'esw'},
                                     'exists': {'args': '<type __builtin__.bool>',
                                                'longname': 'exists',
                                                'numArgs': 0,
                                                'shortname': 'ex'},
                                     'history': {'args': '<type __builtin__.bool>',
                                                 'longname': 'history',
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                                     'image1': {'args': '<type __builtin__.str>',
                                                'longname': 'image1',
                                                'numArgs': 1,
                                                'shortname': 'i1'},
                                     'image2': {'args': '<type __builtin__.str>',
                                                'longname': 'image2',
                                                'numArgs': 1,
                                                'shortname': 'i2'},
                                     'image3': {'args': '<type __builtin__.str>',
                                                'longname': 'image3',
                                                'numArgs': 1,
                                                'shortname': 'i3'},
                                     'name': {'args': '<type __builtin__.str>',
                                              'longname': 'name',
                                              'numArgs': 1,
                                              'shortname': 'n'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'radius': {'args': '<type __builtin__.float>',
                                                'longname': 'radius',
                                                'numArgs': 1,
                                                'shortname': 'r'},
                                     'sections': {'args': '<type __builtin__.int>',
                                                  'longname': 'sections',
                                                  'numArgs': 1,
                                                  'shortname': 'sc'},
                                     'spans': {'args': '<type __builtin__.int>',
                                               'longname': 'spans',
                                               'numArgs': 1,
                                               'shortname': 'sp'},
                                     'startSweep': {'args': '<type __builtin__.float>',
                                                    'longname': 'startSweep',
                                                    'numArgs': 1,
                                                    'shortname': 'ssw'},
                                     'tolerance': {'args': '<type __builtin__.float>',
                                                   'longname': 'tolerance',
                                                   'numArgs': 1,
                                                   'shortname': 'tol'},
                                     'toleranceType': {'args': '<type __builtin__.int>',
                                                       'longname': 'toleranceType',
                                                       'numArgs': 1,
                                                       'shortname': 'tlt'},
                                     'useTolerance': {'args': '<type __builtin__.bool>',
                                                      'longname': 'useTolerance',
                                                      'numArgs': 1,
                                                      'shortname': 'ut'}},
                           'removedFlags': {'as': 'attachToSections'},
                           'shortFlags': {'ahr': 'attachToHeightRatio',
                                          'asp': 'attachToSpans',
                                          'attachToSections': 'attachToSections',
                                          'ax': 'axis',
                                          'axt': 'axisType',
                                          'ch': 'history',
                                          'd': 'degree',
                                          'dde': 'doDragEdit',
                                          'e': 'edit',
                                          'esw': 'endSweep',
                                          'ex': 'exists',
                                          'i1': 'image1',
                                          'i2': 'image2',
                                          'i3': 'image3',
                                          'n': 'name',
                                          'q': 'query',
                                          'r': 'radius',
                                          'sc': 'sections',
                                          'sp': 'spans',
                                          'ssw': 'startSweep',
                                          'tlt': 'toleranceType',
                                          'tol': 'tolerance',
                                          'ut': 'useTolerance'},
                           'type': 'context'},
  'createNurbsSquareCtx': {'flags': {'attachToSpans': {'args': '<type __builtin__.bool>',
                                                       'longname': 'attachToSpans',
                                                       'numArgs': 1,
                                                       'shortname': 'asp'},
                                     'axis': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'axis',
                                              'numArgs': 3,
                                              'shortname': 'ax'},
                                     'axisType': {'args': '<type __builtin__.int>',
                                                  'longname': 'axisType',
                                                  'numArgs': 1,
                                                  'shortname': 'axt'},
                                     'doDragEdit': {'args': '<type __builtin__.bool>',
                                                    'longname': 'doDragEdit',
                                                    'numArgs': 1,
                                                    'shortname': 'dde'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'exists': {'args': '<type __builtin__.bool>',
                                                'longname': 'exists',
                                                'numArgs': 0,
                                                'shortname': 'ex'},
                                     'height': {'args': '<type __builtin__.float>',
                                                'longname': 'height',
                                                'numArgs': 1,
                                                'shortname': 'h'},
                                     'history': {'args': '<type __builtin__.bool>',
                                                 'longname': 'history',
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                                     'image1': {'args': '<type __builtin__.str>',
                                                'longname': 'image1',
                                                'numArgs': 1,
                                                'shortname': 'i1'},
                                     'image2': {'args': '<type __builtin__.str>',
                                                'longname': 'image2',
                                                'numArgs': 1,
                                                'shortname': 'i2'},
                                     'image3': {'args': '<type __builtin__.str>',
                                                'longname': 'image3',
                                                'numArgs': 1,
                                                'shortname': 'i3'},
                                     'name': {'args': '<type __builtin__.str>',
                                              'longname': 'name',
                                              'numArgs': 1,
                                              'shortname': 'n'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'spans': {'args': '<type __builtin__.int>',
                                               'longname': 'spans',
                                               'numArgs': 1,
                                               'shortname': 'sp'},
                                     'surfaceDegree': {'args': '<type __builtin__.int>',
                                                       'longname': 'surfaceDegree',
                                                       'numArgs': 1,
                                                       'shortname': 'sd'},
                                     'width': {'args': '<type __builtin__.float>',
                                               'longname': 'width',
                                               'numArgs': 1,
                                               'shortname': 'w'}},
                           'shortFlags': {'asp': 'attachToSpans',
                                          'ax': 'axis',
                                          'axt': 'axisType',
                                          'ch': 'history',
                                          'dde': 'doDragEdit',
                                          'e': 'edit',
                                          'ex': 'exists',
                                          'h': 'height',
                                          'i1': 'image1',
                                          'i2': 'image2',
                                          'i3': 'image3',
                                          'n': 'name',
                                          'q': 'query',
                                          'sd': 'surfaceDegree',
                                          'sp': 'spans',
                                          'w': 'width'},
                           'type': 'context'},
  'createNurbsTorusCtx': {'flags': {'attachToHeightRatio': {'args': '<type __builtin__.bool>',
                                                            'longname': 'attachToHeightRatio',
                                                            'numArgs': 1,
                                                            'shortname': 'ahr'},
                                    'attachToSections': {'args': '<type __builtin__.bool>',
                                                         'longname': 'attachToSections',
                                                         'numArgs': 1,
                                                         'shortname': 'attachToSections'},
                                    'attachToSpans': {'args': '<type __builtin__.bool>',
                                                      'longname': 'attachToSpans',
                                                      'numArgs': 1,
                                                      'shortname': 'asp'},
                                    'axis': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'axis',
                                             'numArgs': 3,
                                             'shortname': 'ax'},
                                    'axisType': {'args': '<type __builtin__.int>',
                                                 'longname': 'axisType',
                                                 'numArgs': 1,
                                                 'shortname': 'axt'},
                                    'doDragEdit': {'args': '<type __builtin__.bool>',
                                                   'longname': 'doDragEdit',
                                                   'numArgs': 1,
                                                   'shortname': 'dde'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'endSweep': {'args': '<type __builtin__.float>',
                                                 'longname': 'endSweep',
                                                 'numArgs': 1,
                                                 'shortname': 'esw'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'history': {'args': '<type __builtin__.bool>',
                                                'longname': 'history',
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'minorRadius': {'args': '<type __builtin__.float>',
                                                    'longname': 'minorRadius',
                                                    'numArgs': 1,
                                                    'shortname': 'mr'},
                                    'minorSweep': {'args': '<type __builtin__.float>',
                                                   'longname': 'minorSweep',
                                                   'numArgs': 1,
                                                   'shortname': 'msw'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'radius': {'args': '<type __builtin__.float>',
                                               'longname': 'radius',
                                               'numArgs': 1,
                                               'shortname': 'r'},
                                    'sections': {'args': '<type __builtin__.int>',
                                                 'longname': 'sections',
                                                 'numArgs': 1,
                                                 'shortname': 'sc'},
                                    'spans': {'args': '<type __builtin__.int>',
                                              'longname': 'spans',
                                              'numArgs': 1,
                                              'shortname': 'sp'},
                                    'startSweep': {'args': '<type __builtin__.float>',
                                                   'longname': 'startSweep',
                                                   'numArgs': 1,
                                                   'shortname': 'ssw'},
                                    'surfaceDegree': {'args': '<type __builtin__.int>',
                                                      'longname': 'surfaceDegree',
                                                      'numArgs': 1,
                                                      'shortname': 'sd'},
                                    'tolerance': {'args': '<type __builtin__.float>',
                                                  'longname': 'tolerance',
                                                  'numArgs': 1,
                                                  'shortname': 'tol'},
                                    'toleranceType': {'args': '<type __builtin__.int>',
                                                      'longname': 'toleranceType',
                                                      'numArgs': 1,
                                                      'shortname': 'tlt'},
                                    'useTolerance': {'args': '<type __builtin__.bool>',
                                                     'longname': 'useTolerance',
                                                     'numArgs': 1,
                                                     'shortname': 'ut'}},
                          'removedFlags': {'as': 'attachToSections'},
                          'shortFlags': {'ahr': 'attachToHeightRatio',
                                         'asp': 'attachToSpans',
                                         'attachToSections': 'attachToSections',
                                         'ax': 'axis',
                                         'axt': 'axisType',
                                         'ch': 'history',
                                         'dde': 'doDragEdit',
                                         'e': 'edit',
                                         'esw': 'endSweep',
                                         'ex': 'exists',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'mr': 'minorRadius',
                                         'msw': 'minorSweep',
                                         'n': 'name',
                                         'q': 'query',
                                         'r': 'radius',
                                         'sc': 'sections',
                                         'sd': 'surfaceDegree',
                                         'sp': 'spans',
                                         'ssw': 'startSweep',
                                         'tlt': 'toleranceType',
                                         'tol': 'tolerance',
                                         'ut': 'useTolerance'},
                          'type': 'context'},
  'createPolyConeCtx': {'flags': {'attachToSubdivisionsAxis': {'args': '<type __builtin__.bool>',
                                                               'longname': 'attachToSubdivisionsAxis',
                                                               'numArgs': 1,
                                                               'shortname': 'asa'},
                                  'attachToSubdivisionsCap': {'args': '<type __builtin__.bool>',
                                                              'longname': 'attachToSubdivisionsCap',
                                                              'numArgs': 1,
                                                              'shortname': 'asc'},
                                  'attachToSubdivisionsHeight': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'attachToSubdivisionsHeight',
                                                                 'numArgs': 1,
                                                                 'shortname': 'ash'},
                                  'axis': {'args': '<type __builtin__.int>',
                                           'longname': 'axis',
                                           'numArgs': 1,
                                           'shortname': 'ax'},
                                  'createUVs': {'args': '<type __builtin__.int>',
                                                'longname': 'createUVs',
                                                'numArgs': 1,
                                                'shortname': 'cuv'},
                                  'doDragEdit': {'args': '<type __builtin__.bool>',
                                                 'longname': 'doDragEdit',
                                                 'numArgs': 1,
                                                 'shortname': 'dde'},
                                  'doSubdivisionsCapsEdit': {'args': '<type __builtin__.bool>',
                                                             'longname': 'doSubdivisionsCapsEdit',
                                                             'numArgs': 1,
                                                             'shortname': 'dsc'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'height': {'args': '<type __builtin__.float>',
                                             'longname': 'height',
                                             'numArgs': 1,
                                             'shortname': 'h'},
                                  'history': {'args': '<type __builtin__.bool>',
                                              'longname': 'history',
                                              'numArgs': 1,
                                              'shortname': 'ch'},
                                  'image1': {'args': '<type __builtin__.str>',
                                             'longname': 'image1',
                                             'numArgs': 1,
                                             'shortname': 'i1'},
                                  'image2': {'args': '<type __builtin__.str>',
                                             'longname': 'image2',
                                             'numArgs': 1,
                                             'shortname': 'i2'},
                                  'image3': {'args': '<type __builtin__.str>',
                                             'longname': 'image3',
                                             'numArgs': 1,
                                             'shortname': 'i3'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'radius': {'args': '<type __builtin__.float>',
                                             'longname': 'radius',
                                             'numArgs': 1,
                                             'shortname': 'r'},
                                  'roundCap': {'args': '<type __builtin__.bool>',
                                               'longname': 'roundCap',
                                               'numArgs': 1,
                                               'shortname': 'rc'},
                                  'subdivisionsDepth': {'args': '<type __builtin__.int>',
                                                        'longname': 'subdivisionsDepth',
                                                        'numArgs': 1,
                                                        'shortname': 'sd'},
                                  'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                         'longname': 'subdivisionsHeight',
                                                         'numArgs': 1,
                                                         'shortname': 'sh'},
                                  'subdivisionsWidth': {'args': '<type __builtin__.int>',
                                                        'longname': 'subdivisionsWidth',
                                                        'numArgs': 1,
                                                        'shortname': 'sw'}},
                        'shortFlags': {'asa': 'attachToSubdivisionsAxis',
                                       'asc': 'attachToSubdivisionsCap',
                                       'ash': 'attachToSubdivisionsHeight',
                                       'ax': 'axis',
                                       'ch': 'history',
                                       'cuv': 'createUVs',
                                       'dde': 'doDragEdit',
                                       'dsc': 'doSubdivisionsCapsEdit',
                                       'e': 'edit',
                                       'ex': 'exists',
                                       'h': 'height',
                                       'i1': 'image1',
                                       'i2': 'image2',
                                       'i3': 'image3',
                                       'n': 'name',
                                       'q': 'query',
                                       'r': 'radius',
                                       'rc': 'roundCap',
                                       'sd': 'subdivisionsDepth',
                                       'sh': 'subdivisionsHeight',
                                       'sw': 'subdivisionsWidth'},
                        'type': 'context'},
  'createPolyCubeCtx': {'flags': {'attachToSubdivisionsAll': {'args': '<type __builtin__.bool>',
                                                              'longname': 'attachToSubdivisionsAll',
                                                              'numArgs': 1,
                                                              'shortname': 'asa'},
                                  'attachToSubdivisionsDepth': {'args': '<type __builtin__.bool>',
                                                                'longname': 'attachToSubdivisionsDepth',
                                                                'numArgs': 1,
                                                                'shortname': 'asd'},
                                  'attachToSubdivisionsHeight': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'attachToSubdivisionsHeight',
                                                                 'numArgs': 1,
                                                                 'shortname': 'ash'},
                                  'attachToSubdivisionsWidth': {'args': '<type __builtin__.bool>',
                                                                'longname': 'attachToSubdivisionsWidth',
                                                                'numArgs': 1,
                                                                'shortname': 'asw'},
                                  'axis': {'args': '<type __builtin__.int>',
                                           'longname': 'axis',
                                           'numArgs': 1,
                                           'shortname': 'ax'},
                                  'createUVs': {'args': '<type __builtin__.int>',
                                                'longname': 'createUVs',
                                                'numArgs': 1,
                                                'shortname': 'cuv'},
                                  'depth': {'args': '<type __builtin__.float>',
                                            'longname': 'depth',
                                            'numArgs': 1,
                                            'shortname': 'd'},
                                  'doDragEdit': {'args': '<type __builtin__.bool>',
                                                 'longname': 'doDragEdit',
                                                 'numArgs': 1,
                                                 'shortname': 'dde'},
                                  'doSubdivisionsCapsEdit': {'args': '<type __builtin__.bool>',
                                                             'longname': 'doSubdivisionsCapsEdit',
                                                             'numArgs': 1,
                                                             'shortname': 'dsc'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'height': {'args': '<type __builtin__.float>',
                                             'longname': 'height',
                                             'numArgs': 1,
                                             'shortname': 'h'},
                                  'history': {'args': '<type __builtin__.bool>',
                                              'longname': 'history',
                                              'numArgs': 1,
                                              'shortname': 'ch'},
                                  'image1': {'args': '<type __builtin__.str>',
                                             'longname': 'image1',
                                             'numArgs': 1,
                                             'shortname': 'i1'},
                                  'image2': {'args': '<type __builtin__.str>',
                                             'longname': 'image2',
                                             'numArgs': 1,
                                             'shortname': 'i2'},
                                  'image3': {'args': '<type __builtin__.str>',
                                             'longname': 'image3',
                                             'numArgs': 1,
                                             'shortname': 'i3'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'subdivisionsDepth': {'args': '<type __builtin__.int>',
                                                        'longname': 'subdivisionsDepth',
                                                        'numArgs': 1,
                                                        'shortname': 'sd'},
                                  'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                         'longname': 'subdivisionsHeight',
                                                         'numArgs': 1,
                                                         'shortname': 'sh'},
                                  'subdivisionsWidth': {'args': '<type __builtin__.int>',
                                                        'longname': 'subdivisionsWidth',
                                                        'numArgs': 1,
                                                        'shortname': 'sw'},
                                  'width': {'args': '<type __builtin__.float>',
                                            'longname': 'width',
                                            'numArgs': 1,
                                            'shortname': 'w'}},
                        'shortFlags': {'asa': 'attachToSubdivisionsAll',
                                       'asd': 'attachToSubdivisionsDepth',
                                       'ash': 'attachToSubdivisionsHeight',
                                       'asw': 'attachToSubdivisionsWidth',
                                       'ax': 'axis',
                                       'ch': 'history',
                                       'cuv': 'createUVs',
                                       'd': 'depth',
                                       'dde': 'doDragEdit',
                                       'dsc': 'doSubdivisionsCapsEdit',
                                       'e': 'edit',
                                       'ex': 'exists',
                                       'h': 'height',
                                       'i1': 'image1',
                                       'i2': 'image2',
                                       'i3': 'image3',
                                       'n': 'name',
                                       'q': 'query',
                                       'sd': 'subdivisionsDepth',
                                       'sh': 'subdivisionsHeight',
                                       'sw': 'subdivisionsWidth',
                                       'w': 'width'},
                        'type': 'context'},
  'createPolyCylinderCtx': {'flags': {'attachToSubdivisionsAxis': {'args': '<type __builtin__.bool>',
                                                                   'longname': 'attachToSubdivisionsAxis',
                                                                   'numArgs': 1,
                                                                   'shortname': 'asa'},
                                      'attachToSubdivisionsCap': {'args': '<type __builtin__.bool>',
                                                                  'longname': 'attachToSubdivisionsCap',
                                                                  'numArgs': 1,
                                                                  'shortname': 'asc'},
                                      'attachToSubdivisionsHeight': {'args': '<type __builtin__.bool>',
                                                                     'longname': 'attachToSubdivisionsHeight',
                                                                     'numArgs': 1,
                                                                     'shortname': 'ash'},
                                      'axis': {'args': '<type __builtin__.int>',
                                               'longname': 'axis',
                                               'numArgs': 1,
                                               'shortname': 'ax'},
                                      'createUVs': {'args': '<type __builtin__.int>',
                                                    'longname': 'createUVs',
                                                    'numArgs': 1,
                                                    'shortname': 'cuv'},
                                      'doDragEdit': {'args': '<type __builtin__.bool>',
                                                     'longname': 'doDragEdit',
                                                     'numArgs': 1,
                                                     'shortname': 'dde'},
                                      'doSubdivisionsCapsEdit': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'doSubdivisionsCapsEdit',
                                                                 'numArgs': 1,
                                                                 'shortname': 'dsc'},
                                      'edit': {'args': '<type __builtin__.bool>',
                                               'longname': 'edit',
                                               'numArgs': 0,
                                               'shortname': 'e'},
                                      'exists': {'args': '<type __builtin__.bool>',
                                                 'longname': 'exists',
                                                 'numArgs': 0,
                                                 'shortname': 'ex'},
                                      'height': {'args': '<type __builtin__.float>',
                                                 'longname': 'height',
                                                 'numArgs': 1,
                                                 'shortname': 'h'},
                                      'history': {'args': '<type __builtin__.bool>',
                                                  'longname': 'history',
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                                      'image1': {'args': '<type __builtin__.str>',
                                                 'longname': 'image1',
                                                 'numArgs': 1,
                                                 'shortname': 'i1'},
                                      'image2': {'args': '<type __builtin__.str>',
                                                 'longname': 'image2',
                                                 'numArgs': 1,
                                                 'shortname': 'i2'},
                                      'image3': {'args': '<type __builtin__.str>',
                                                 'longname': 'image3',
                                                 'numArgs': 1,
                                                 'shortname': 'i3'},
                                      'name': {'args': '<type __builtin__.str>',
                                               'longname': 'name',
                                               'numArgs': 1,
                                               'shortname': 'n'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'radius': {'args': '<type __builtin__.float>',
                                                 'longname': 'radius',
                                                 'numArgs': 1,
                                                 'shortname': 'r'},
                                      'roundCap': {'args': '<type __builtin__.bool>',
                                                   'longname': 'roundCap',
                                                   'numArgs': 1,
                                                   'shortname': 'rc'},
                                      'subdivisionsAxis': {'args': '<type __builtin__.int>',
                                                           'longname': 'subdivisionsAxis',
                                                           'numArgs': 1,
                                                           'shortname': 'sa'},
                                      'subdivisionsCap': {'args': '<type __builtin__.int>',
                                                          'longname': 'subdivisionsCap',
                                                          'numArgs': 1,
                                                          'shortname': 'sc'},
                                      'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                             'longname': 'subdivisionsHeight',
                                                             'numArgs': 1,
                                                             'shortname': 'sh'}},
                            'shortFlags': {'asa': 'attachToSubdivisionsAxis',
                                           'asc': 'attachToSubdivisionsCap',
                                           'ash': 'attachToSubdivisionsHeight',
                                           'ax': 'axis',
                                           'ch': 'history',
                                           'cuv': 'createUVs',
                                           'dde': 'doDragEdit',
                                           'dsc': 'doSubdivisionsCapsEdit',
                                           'e': 'edit',
                                           'ex': 'exists',
                                           'h': 'height',
                                           'i1': 'image1',
                                           'i2': 'image2',
                                           'i3': 'image3',
                                           'n': 'name',
                                           'q': 'query',
                                           'r': 'radius',
                                           'rc': 'roundCap',
                                           'sa': 'subdivisionsAxis',
                                           'sc': 'subdivisionsCap',
                                           'sh': 'subdivisionsHeight'},
                            'type': 'context'},
  'createPolyHelixCtx': {'flags': {'attachToNumberCoils': {'args': '<type __builtin__.bool>',
                                                           'longname': 'attachToNumberCoils',
                                                           'numArgs': 1,
                                                           'shortname': 'anc'},
                                   'attachToRadius': {'args': '<type __builtin__.bool>',
                                                      'longname': 'attachToRadius',
                                                      'numArgs': 1,
                                                      'shortname': 'atr'},
                                   'attachToSubdivisionsAxis': {'args': '<type __builtin__.bool>',
                                                                'longname': 'attachToSubdivisionsAxis',
                                                                'numArgs': 1,
                                                                'shortname': 'asa'},
                                   'attachToSubdivisionsCap': {'args': '<type __builtin__.bool>',
                                                               'longname': 'attachToSubdivisionsCap',
                                                               'numArgs': 1,
                                                               'shortname': 'asc'},
                                   'attachToSubdivisionsCoil': {'args': '<type __builtin__.bool>',
                                                                'longname': 'attachToSubdivisionsCoil',
                                                                'numArgs': 1,
                                                                'shortname': 'aso'},
                                   'axis': {'args': '<type __builtin__.int>',
                                            'longname': 'axis',
                                            'numArgs': 1,
                                            'shortname': 'ax'},
                                   'coils': {'args': '<type __builtin__.float>',
                                             'longname': 'coils',
                                             'numArgs': 1,
                                             'shortname': 'c'},
                                   'createUVs': {'args': '<type __builtin__.int>',
                                                 'longname': 'createUVs',
                                                 'numArgs': 1,
                                                 'shortname': 'cuv'},
                                   'direction': {'args': '<type __builtin__.bool>',
                                                 'longname': 'direction',
                                                 'numArgs': 1,
                                                 'shortname': 'dir'},
                                   'doDragEdit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'doDragEdit',
                                                  'numArgs': 1,
                                                  'shortname': 'dde'},
                                   'doSubdivisionsCapsEdit': {'args': '<type __builtin__.bool>',
                                                              'longname': 'doSubdivisionsCapsEdit',
                                                              'numArgs': 1,
                                                              'shortname': 'dsc'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'height': {'args': '<type __builtin__.float>',
                                              'longname': 'height',
                                              'numArgs': 1,
                                              'shortname': 'h'},
                                   'history': {'args': '<type __builtin__.bool>',
                                               'longname': 'history',
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'radius': {'args': '<type __builtin__.float>',
                                              'longname': 'radius',
                                              'numArgs': 1,
                                              'shortname': 'r'},
                                   'roundCap': {'args': '<type __builtin__.bool>',
                                                'longname': 'roundCap',
                                                'numArgs': 1,
                                                'shortname': 'rc'},
                                   'subdivisionsAxis': {'args': '<type __builtin__.int>',
                                                        'longname': 'subdivisionsAxis',
                                                        'numArgs': 1,
                                                        'shortname': 'sa'},
                                   'subdivisionsCap': {'args': '<type __builtin__.int>',
                                                       'longname': 'subdivisionsCap',
                                                       'numArgs': 1,
                                                       'shortname': 'sc'},
                                   'subdivisionsCoil': {'args': '<type __builtin__.int>',
                                                        'longname': 'subdivisionsCoil',
                                                        'numArgs': 1,
                                                        'shortname': 'sl'},
                                   'width': {'args': '<type __builtin__.float>',
                                             'longname': 'width',
                                             'numArgs': 1,
                                             'shortname': 'w'}},
                         'shortFlags': {'anc': 'attachToNumberCoils',
                                        'asa': 'attachToSubdivisionsAxis',
                                        'asc': 'attachToSubdivisionsCap',
                                        'aso': 'attachToSubdivisionsCoil',
                                        'atr': 'attachToRadius',
                                        'ax': 'axis',
                                        'c': 'coils',
                                        'ch': 'history',
                                        'cuv': 'createUVs',
                                        'dde': 'doDragEdit',
                                        'dir': 'direction',
                                        'dsc': 'doSubdivisionsCapsEdit',
                                        'e': 'edit',
                                        'ex': 'exists',
                                        'h': 'height',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'n': 'name',
                                        'q': 'query',
                                        'r': 'radius',
                                        'rc': 'roundCap',
                                        'sa': 'subdivisionsAxis',
                                        'sc': 'subdivisionsCap',
                                        'sl': 'subdivisionsCoil',
                                        'w': 'width'},
                         'type': 'context'},
  'createPolyPipeCtx': {'flags': {'attachToSubdivisionsAxis': {'args': '<type __builtin__.bool>',
                                                               'longname': 'attachToSubdivisionsAxis',
                                                               'numArgs': 1,
                                                               'shortname': 'asa'},
                                  'attachToSubdivisionsCap': {'args': '<type __builtin__.bool>',
                                                              'longname': 'attachToSubdivisionsCap',
                                                              'numArgs': 1,
                                                              'shortname': 'asc'},
                                  'attachToSubdivisionsHeight': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'attachToSubdivisionsHeight',
                                                                 'numArgs': 1,
                                                                 'shortname': 'ash'},
                                  'attachToThickness': {'args': '<type __builtin__.bool>',
                                                        'longname': 'attachToThickness',
                                                        'numArgs': 1,
                                                        'shortname': 'att'},
                                  'axis': {'args': '<type __builtin__.int>',
                                           'longname': 'axis',
                                           'numArgs': 1,
                                           'shortname': 'ax'},
                                  'createUVs': {'args': '<type __builtin__.int>',
                                                'longname': 'createUVs',
                                                'numArgs': 1,
                                                'shortname': 'cuv'},
                                  'doDragEdit': {'args': '<type __builtin__.bool>',
                                                 'longname': 'doDragEdit',
                                                 'numArgs': 1,
                                                 'shortname': 'dde'},
                                  'doSubdivisionsCapsEdit': {'args': '<type __builtin__.bool>',
                                                             'longname': 'doSubdivisionsCapsEdit',
                                                             'numArgs': 1,
                                                             'shortname': 'dsc'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'height': {'args': '<type __builtin__.float>',
                                             'longname': 'height',
                                             'numArgs': 1,
                                             'shortname': 'h'},
                                  'history': {'args': '<type __builtin__.bool>',
                                              'longname': 'history',
                                              'numArgs': 1,
                                              'shortname': 'ch'},
                                  'image1': {'args': '<type __builtin__.str>',
                                             'longname': 'image1',
                                             'numArgs': 1,
                                             'shortname': 'i1'},
                                  'image2': {'args': '<type __builtin__.str>',
                                             'longname': 'image2',
                                             'numArgs': 1,
                                             'shortname': 'i2'},
                                  'image3': {'args': '<type __builtin__.str>',
                                             'longname': 'image3',
                                             'numArgs': 1,
                                             'shortname': 'i3'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'radius': {'args': '<type __builtin__.float>',
                                             'longname': 'radius',
                                             'numArgs': 1,
                                             'shortname': 'r'},
                                  'roundCap': {'args': '<type __builtin__.bool>',
                                               'longname': 'roundCap',
                                               'numArgs': 1,
                                               'shortname': 'rc'},
                                  'subdivisionsAxis': {'args': '<type __builtin__.int>',
                                                       'longname': 'subdivisionsAxis',
                                                       'numArgs': 1,
                                                       'shortname': 'sa'},
                                  'subdivisionsCap': {'args': '<type __builtin__.int>',
                                                      'longname': 'subdivisionsCap',
                                                      'numArgs': 1,
                                                      'shortname': 'sc'},
                                  'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                         'longname': 'subdivisionsHeight',
                                                         'numArgs': 1,
                                                         'shortname': 'sh'},
                                  'thickness': {'args': '<type __builtin__.float>',
                                                'longname': 'thickness',
                                                'numArgs': 1,
                                                'shortname': 'th'}},
                        'shortFlags': {'asa': 'attachToSubdivisionsAxis',
                                       'asc': 'attachToSubdivisionsCap',
                                       'ash': 'attachToSubdivisionsHeight',
                                       'att': 'attachToThickness',
                                       'ax': 'axis',
                                       'ch': 'history',
                                       'cuv': 'createUVs',
                                       'dde': 'doDragEdit',
                                       'dsc': 'doSubdivisionsCapsEdit',
                                       'e': 'edit',
                                       'ex': 'exists',
                                       'h': 'height',
                                       'i1': 'image1',
                                       'i2': 'image2',
                                       'i3': 'image3',
                                       'n': 'name',
                                       'q': 'query',
                                       'r': 'radius',
                                       'rc': 'roundCap',
                                       'sa': 'subdivisionsAxis',
                                       'sc': 'subdivisionsCap',
                                       'sh': 'subdivisionsHeight',
                                       'th': 'thickness'},
                        'type': 'context'},
  'createPolyPlaneCtx': {'flags': {'attachToSubdivisionsAll': {'args': '<type __builtin__.bool>',
                                                               'longname': 'attachToSubdivisionsAll',
                                                               'numArgs': 1,
                                                               'shortname': 'asa'},
                                   'attachToSubdivisionsHeight': {'args': '<type __builtin__.bool>',
                                                                  'longname': 'attachToSubdivisionsHeight',
                                                                  'numArgs': 1,
                                                                  'shortname': 'ash'},
                                   'attachToSubdivisionsWidth': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'attachToSubdivisionsWidth',
                                                                 'numArgs': 1,
                                                                 'shortname': 'asw'},
                                   'axis': {'args': '<type __builtin__.int>',
                                            'longname': 'axis',
                                            'numArgs': 1,
                                            'shortname': 'ax'},
                                   'createUVs': {'args': '<type __builtin__.int>',
                                                 'longname': 'createUVs',
                                                 'numArgs': 1,
                                                 'shortname': 'cuv'},
                                   'doDragEdit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'doDragEdit',
                                                  'numArgs': 1,
                                                  'shortname': 'dde'},
                                   'doSubdivisionsCapsEdit': {'args': '<type __builtin__.bool>',
                                                              'longname': 'doSubdivisionsCapsEdit',
                                                              'numArgs': 1,
                                                              'shortname': 'dsc'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'height': {'args': '<type __builtin__.float>',
                                              'longname': 'height',
                                              'numArgs': 1,
                                              'shortname': 'h'},
                                   'history': {'args': '<type __builtin__.bool>',
                                               'longname': 'history',
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                          'longname': 'subdivisionsHeight',
                                                          'numArgs': 1,
                                                          'shortname': 'sh'},
                                   'subdivisionsWidth': {'args': '<type __builtin__.int>',
                                                         'longname': 'subdivisionsWidth',
                                                         'numArgs': 1,
                                                         'shortname': 'sw'},
                                   'width': {'args': '<type __builtin__.float>',
                                             'longname': 'width',
                                             'numArgs': 1,
                                             'shortname': 'w'}},
                         'shortFlags': {'asa': 'attachToSubdivisionsAll',
                                        'ash': 'attachToSubdivisionsHeight',
                                        'asw': 'attachToSubdivisionsWidth',
                                        'ax': 'axis',
                                        'ch': 'history',
                                        'cuv': 'createUVs',
                                        'dde': 'doDragEdit',
                                        'dsc': 'doSubdivisionsCapsEdit',
                                        'e': 'edit',
                                        'ex': 'exists',
                                        'h': 'height',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'n': 'name',
                                        'q': 'query',
                                        'sh': 'subdivisionsHeight',
                                        'sw': 'subdivisionsWidth',
                                        'w': 'width'},
                         'type': 'context'},
  'createPolyPlatonicSolidCtx': {'flags': {'axis': {'args': '<type __builtin__.int>',
                                                    'longname': 'axis',
                                                    'numArgs': 1,
                                                    'shortname': 'ax'},
                                           'createUVs': {'args': '<type __builtin__.int>',
                                                         'longname': 'createUVs',
                                                         'numArgs': 1,
                                                         'shortname': 'cuv'},
                                           'doDragEdit': {'args': '<type __builtin__.bool>',
                                                          'longname': 'doDragEdit',
                                                          'numArgs': 1,
                                                          'shortname': 'dde'},
                                           'doSubdivisionsCapsEdit': {'args': '<type __builtin__.bool>',
                                                                      'longname': 'doSubdivisionsCapsEdit',
                                                                      'numArgs': 1,
                                                                      'shortname': 'dsc'},
                                           'edit': {'args': '<type __builtin__.bool>',
                                                    'longname': 'edit',
                                                    'numArgs': 0,
                                                    'shortname': 'e'},
                                           'exists': {'args': '<type __builtin__.bool>',
                                                      'longname': 'exists',
                                                      'numArgs': 0,
                                                      'shortname': 'ex'},
                                           'history': {'args': '<type __builtin__.bool>',
                                                       'longname': 'history',
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                                           'image1': {'args': '<type __builtin__.str>',
                                                      'longname': 'image1',
                                                      'numArgs': 1,
                                                      'shortname': 'i1'},
                                           'image2': {'args': '<type __builtin__.str>',
                                                      'longname': 'image2',
                                                      'numArgs': 1,
                                                      'shortname': 'i2'},
                                           'image3': {'args': '<type __builtin__.str>',
                                                      'longname': 'image3',
                                                      'numArgs': 1,
                                                      'shortname': 'i3'},
                                           'name': {'args': '<type __builtin__.str>',
                                                    'longname': 'name',
                                                    'numArgs': 1,
                                                    'shortname': 'n'},
                                           'query': {'args': '<type __builtin__.bool>',
                                                     'longname': 'query',
                                                     'numArgs': 0,
                                                     'shortname': 'q'},
                                           'radius': {'args': '<type __builtin__.float>',
                                                      'longname': 'radius',
                                                      'numArgs': 1,
                                                      'shortname': 'r'},
                                           'sideLength': {'args': '<type __builtin__.float>',
                                                          'longname': 'sideLength',
                                                          'numArgs': 1,
                                                          'shortname': 'sl'},
                                           'solidType': {'args': '<type __builtin__.int>',
                                                         'longname': 'solidType',
                                                         'numArgs': 1,
                                                         'shortname': 'st'}},
                                 'shortFlags': {'ax': 'axis',
                                                'ch': 'history',
                                                'cuv': 'createUVs',
                                                'dde': 'doDragEdit',
                                                'dsc': 'doSubdivisionsCapsEdit',
                                                'e': 'edit',
                                                'ex': 'exists',
                                                'i1': 'image1',
                                                'i2': 'image2',
                                                'i3': 'image3',
                                                'n': 'name',
                                                'q': 'query',
                                                'r': 'radius',
                                                'sl': 'sideLength',
                                                'st': 'solidType'},
                                 'type': 'context'},
  'createPolyPrismCtx': {'flags': {'attachToNumSides': {'args': '<type __builtin__.bool>',
                                                        'longname': 'attachToNumSides',
                                                        'numArgs': 1,
                                                        'shortname': 'ans'},
                                   'attachToSubdivisionsCap': {'args': '<type __builtin__.bool>',
                                                               'longname': 'attachToSubdivisionsCap',
                                                               'numArgs': 1,
                                                               'shortname': 'asc'},
                                   'attachToSubdivisionsHeight': {'args': '<type __builtin__.bool>',
                                                                  'longname': 'attachToSubdivisionsHeight',
                                                                  'numArgs': 1,
                                                                  'shortname': 'ash'},
                                   'axis': {'args': '<type __builtin__.int>',
                                            'longname': 'axis',
                                            'numArgs': 1,
                                            'shortname': 'ax'},
                                   'createUVs': {'args': '<type __builtin__.int>',
                                                 'longname': 'createUVs',
                                                 'numArgs': 1,
                                                 'shortname': 'cuv'},
                                   'doDragEdit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'doDragEdit',
                                                  'numArgs': 1,
                                                  'shortname': 'dde'},
                                   'doSubdivisionsCapsEdit': {'args': '<type __builtin__.bool>',
                                                              'longname': 'doSubdivisionsCapsEdit',
                                                              'numArgs': 1,
                                                              'shortname': 'dsc'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'history': {'args': '<type __builtin__.bool>',
                                               'longname': 'history',
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'length': {'args': '<type __builtin__.float>',
                                              'longname': 'length',
                                              'numArgs': 1,
                                              'shortname': 'ln'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'numberOfSides': {'args': '<type __builtin__.int>',
                                                     'longname': 'numberOfSides',
                                                     'numArgs': 1,
                                                     'shortname': 'ns'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'sideLength': {'args': '<type __builtin__.float>',
                                                  'longname': 'sideLength',
                                                  'numArgs': 1,
                                                  'shortname': 'sl'},
                                   'subdivisionsCap': {'args': '<type __builtin__.int>',
                                                       'longname': 'subdivisionsCap',
                                                       'numArgs': 1,
                                                       'shortname': 'sc'},
                                   'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                          'longname': 'subdivisionsHeight',
                                                          'numArgs': 1,
                                                          'shortname': 'sh'}},
                         'shortFlags': {'ans': 'attachToNumSides',
                                        'asc': 'attachToSubdivisionsCap',
                                        'ash': 'attachToSubdivisionsHeight',
                                        'ax': 'axis',
                                        'ch': 'history',
                                        'cuv': 'createUVs',
                                        'dde': 'doDragEdit',
                                        'dsc': 'doSubdivisionsCapsEdit',
                                        'e': 'edit',
                                        'ex': 'exists',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'ln': 'length',
                                        'n': 'name',
                                        'ns': 'numberOfSides',
                                        'q': 'query',
                                        'sc': 'subdivisionsCap',
                                        'sh': 'subdivisionsHeight',
                                        'sl': 'sideLength'},
                         'type': 'context'},
  'createPolyPyramidCtx': {'flags': {'attachToSubdivisionsCap': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'attachToSubdivisionsCap',
                                                                 'numArgs': 1,
                                                                 'shortname': 'asc'},
                                     'attachToSubdivisionsHeight': {'args': '<type __builtin__.bool>',
                                                                    'longname': 'attachToSubdivisionsHeight',
                                                                    'numArgs': 1,
                                                                    'shortname': 'ash'},
                                     'axis': {'args': '<type __builtin__.int>',
                                              'longname': 'axis',
                                              'numArgs': 1,
                                              'shortname': 'ax'},
                                     'createUVs': {'args': '<type __builtin__.int>',
                                                   'longname': 'createUVs',
                                                   'numArgs': 1,
                                                   'shortname': 'cuv'},
                                     'doDragEdit': {'args': '<type __builtin__.bool>',
                                                    'longname': 'doDragEdit',
                                                    'numArgs': 1,
                                                    'shortname': 'dde'},
                                     'doSubdivisionsCapsEdit': {'args': '<type __builtin__.bool>',
                                                                'longname': 'doSubdivisionsCapsEdit',
                                                                'numArgs': 1,
                                                                'shortname': 'dsc'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'exists': {'args': '<type __builtin__.bool>',
                                                'longname': 'exists',
                                                'numArgs': 0,
                                                'shortname': 'ex'},
                                     'history': {'args': '<type __builtin__.bool>',
                                                 'longname': 'history',
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                                     'image1': {'args': '<type __builtin__.str>',
                                                'longname': 'image1',
                                                'numArgs': 1,
                                                'shortname': 'i1'},
                                     'image2': {'args': '<type __builtin__.str>',
                                                'longname': 'image2',
                                                'numArgs': 1,
                                                'shortname': 'i2'},
                                     'image3': {'args': '<type __builtin__.str>',
                                                'longname': 'image3',
                                                'numArgs': 1,
                                                'shortname': 'i3'},
                                     'name': {'args': '<type __builtin__.str>',
                                              'longname': 'name',
                                              'numArgs': 1,
                                              'shortname': 'n'},
                                     'numberOfSides': {'args': '<type __builtin__.int>',
                                                       'longname': 'numberOfSides',
                                                       'numArgs': 1,
                                                       'shortname': 'ns'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'sideLength': {'args': '<type __builtin__.float>',
                                                    'longname': 'sideLength',
                                                    'numArgs': 1,
                                                    'shortname': 'sl'},
                                     'subdivisionsDepth': {'args': '<type __builtin__.int>',
                                                           'longname': 'subdivisionsDepth',
                                                           'numArgs': 1,
                                                           'shortname': 'sd'},
                                     'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                            'longname': 'subdivisionsHeight',
                                                            'numArgs': 1,
                                                            'shortname': 'sh'}},
                           'shortFlags': {'asc': 'attachToSubdivisionsCap',
                                          'ash': 'attachToSubdivisionsHeight',
                                          'ax': 'axis',
                                          'ch': 'history',
                                          'cuv': 'createUVs',
                                          'dde': 'doDragEdit',
                                          'dsc': 'doSubdivisionsCapsEdit',
                                          'e': 'edit',
                                          'ex': 'exists',
                                          'i1': 'image1',
                                          'i2': 'image2',
                                          'i3': 'image3',
                                          'n': 'name',
                                          'ns': 'numberOfSides',
                                          'q': 'query',
                                          'sd': 'subdivisionsDepth',
                                          'sh': 'subdivisionsHeight',
                                          'sl': 'sideLength'},
                           'type': 'context'},
  'createPolySoccerBallCtx': {'flags': {'axis': {'args': '<type __builtin__.int>',
                                                 'longname': 'axis',
                                                 'numArgs': 1,
                                                 'shortname': 'ax'},
                                        'createUVs': {'args': '<type __builtin__.int>',
                                                      'longname': 'createUVs',
                                                      'numArgs': 1,
                                                      'shortname': 'cuv'},
                                        'doDragEdit': {'args': '<type __builtin__.bool>',
                                                       'longname': 'doDragEdit',
                                                       'numArgs': 1,
                                                       'shortname': 'dde'},
                                        'doSubdivisionsCapsEdit': {'args': '<type __builtin__.bool>',
                                                                   'longname': 'doSubdivisionsCapsEdit',
                                                                   'numArgs': 1,
                                                                   'shortname': 'dsc'},
                                        'edit': {'args': '<type __builtin__.bool>',
                                                 'longname': 'edit',
                                                 'numArgs': 0,
                                                 'shortname': 'e'},
                                        'exists': {'args': '<type __builtin__.bool>',
                                                   'longname': 'exists',
                                                   'numArgs': 0,
                                                   'shortname': 'ex'},
                                        'history': {'args': '<type __builtin__.bool>',
                                                    'longname': 'history',
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                                        'image1': {'args': '<type __builtin__.str>',
                                                   'longname': 'image1',
                                                   'numArgs': 1,
                                                   'shortname': 'i1'},
                                        'image2': {'args': '<type __builtin__.str>',
                                                   'longname': 'image2',
                                                   'numArgs': 1,
                                                   'shortname': 'i2'},
                                        'image3': {'args': '<type __builtin__.str>',
                                                   'longname': 'image3',
                                                   'numArgs': 1,
                                                   'shortname': 'i3'},
                                        'name': {'args': '<type __builtin__.str>',
                                                 'longname': 'name',
                                                 'numArgs': 1,
                                                 'shortname': 'n'},
                                        'query': {'args': '<type __builtin__.bool>',
                                                  'longname': 'query',
                                                  'numArgs': 0,
                                                  'shortname': 'q'},
                                        'radius': {'args': '<type __builtin__.float>',
                                                   'longname': 'radius',
                                                   'numArgs': 1,
                                                   'shortname': 'r'},
                                        'sideLength': {'args': '<type __builtin__.float>',
                                                       'longname': 'sideLength',
                                                       'numArgs': 1,
                                                       'shortname': 'sl'}},
                              'shortFlags': {'ax': 'axis',
                                             'ch': 'history',
                                             'cuv': 'createUVs',
                                             'dde': 'doDragEdit',
                                             'dsc': 'doSubdivisionsCapsEdit',
                                             'e': 'edit',
                                             'ex': 'exists',
                                             'i1': 'image1',
                                             'i2': 'image2',
                                             'i3': 'image3',
                                             'n': 'name',
                                             'q': 'query',
                                             'r': 'radius',
                                             'sl': 'sideLength'},
                              'type': 'context'},
  'createPolySphereCtx': {'flags': {'attachToSubdivisionsAll': {'args': '<type __builtin__.bool>',
                                                                'longname': 'attachToSubdivisionsAll',
                                                                'numArgs': 1,
                                                                'shortname': 'asa'},
                                    'attachToSubdivisionsAxis': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'attachToSubdivisionsAxis',
                                                                 'numArgs': 1,
                                                                 'shortname': 'asx'},
                                    'attachToSubdivisionsHeight': {'args': '<type __builtin__.bool>',
                                                                   'longname': 'attachToSubdivisionsHeight',
                                                                   'numArgs': 1,
                                                                   'shortname': 'ash'},
                                    'axis': {'args': '<type __builtin__.int>',
                                             'longname': 'axis',
                                             'numArgs': 1,
                                             'shortname': 'ax'},
                                    'createUVs': {'args': '<type __builtin__.int>',
                                                  'longname': 'createUVs',
                                                  'numArgs': 1,
                                                  'shortname': 'cuv'},
                                    'doDragEdit': {'args': '<type __builtin__.bool>',
                                                   'longname': 'doDragEdit',
                                                   'numArgs': 1,
                                                   'shortname': 'dde'},
                                    'doSubdivisionsCapsEdit': {'args': '<type __builtin__.bool>',
                                                               'longname': 'doSubdivisionsCapsEdit',
                                                               'numArgs': 1,
                                                               'shortname': 'dsc'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'history': {'args': '<type __builtin__.bool>',
                                                'longname': 'history',
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'radius': {'args': '<type __builtin__.float>',
                                               'longname': 'radius',
                                               'numArgs': 1,
                                               'shortname': 'r'},
                                    'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                           'longname': 'subdivisionsHeight',
                                                           'numArgs': 1,
                                                           'shortname': 'sh'},
                                    'subdivisionsWidth': {'args': '<type __builtin__.int>',
                                                          'longname': 'subdivisionsWidth',
                                                          'numArgs': 1,
                                                          'shortname': 'sw'}},
                          'shortFlags': {'asa': 'attachToSubdivisionsAll',
                                         'ash': 'attachToSubdivisionsHeight',
                                         'asx': 'attachToSubdivisionsAxis',
                                         'ax': 'axis',
                                         'ch': 'history',
                                         'cuv': 'createUVs',
                                         'dde': 'doDragEdit',
                                         'dsc': 'doSubdivisionsCapsEdit',
                                         'e': 'edit',
                                         'ex': 'exists',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'n': 'name',
                                         'q': 'query',
                                         'r': 'radius',
                                         'sh': 'subdivisionsHeight',
                                         'sw': 'subdivisionsWidth'},
                          'type': 'context'},
  'createPolyTorusCtx': {'flags': {'attachToSectionRadius': {'args': '<type __builtin__.bool>',
                                                             'longname': 'attachToSectionRadius',
                                                             'numArgs': 1,
                                                             'shortname': 'asr'},
                                   'attachToSubdivisionsAxis': {'args': '<type __builtin__.bool>',
                                                                'longname': 'attachToSubdivisionsAxis',
                                                                'numArgs': 1,
                                                                'shortname': 'asa'},
                                   'attachToSubdivisionsHeight': {'args': '<type __builtin__.bool>',
                                                                  'longname': 'attachToSubdivisionsHeight',
                                                                  'numArgs': 1,
                                                                  'shortname': 'ash'},
                                   'axis': {'args': '<type __builtin__.int>',
                                            'longname': 'axis',
                                            'numArgs': 1,
                                            'shortname': 'ax'},
                                   'createUVs': {'args': '<type __builtin__.int>',
                                                 'longname': 'createUVs',
                                                 'numArgs': 1,
                                                 'shortname': 'cuv'},
                                   'doDragEdit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'doDragEdit',
                                                  'numArgs': 1,
                                                  'shortname': 'dde'},
                                   'doSubdivisionsCapsEdit': {'args': '<type __builtin__.bool>',
                                                              'longname': 'doSubdivisionsCapsEdit',
                                                              'numArgs': 1,
                                                              'shortname': 'dsc'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'history': {'args': '<type __builtin__.bool>',
                                               'longname': 'history',
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'radius': {'args': '<type __builtin__.float>',
                                              'longname': 'radius',
                                              'numArgs': 1,
                                              'shortname': 'r'},
                                   'sectionRadius': {'args': '<type __builtin__.float>',
                                                     'longname': 'sectionRadius',
                                                     'numArgs': 1,
                                                     'shortname': 'sr'},
                                   'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                          'longname': 'subdivisionsHeight',
                                                          'numArgs': 1,
                                                          'shortname': 'sh'},
                                   'subdivisionsWidth': {'args': '<type __builtin__.int>',
                                                         'longname': 'subdivisionsWidth',
                                                         'numArgs': 1,
                                                         'shortname': 'sw'},
                                   'twist': {'args': '<type __builtin__.float>',
                                             'longname': 'twist',
                                             'numArgs': 1,
                                             'shortname': 'tw'}},
                         'shortFlags': {'asa': 'attachToSubdivisionsAxis',
                                        'ash': 'attachToSubdivisionsHeight',
                                        'asr': 'attachToSectionRadius',
                                        'ax': 'axis',
                                        'ch': 'history',
                                        'cuv': 'createUVs',
                                        'dde': 'doDragEdit',
                                        'dsc': 'doSubdivisionsCapsEdit',
                                        'e': 'edit',
                                        'ex': 'exists',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'n': 'name',
                                        'q': 'query',
                                        'r': 'radius',
                                        'sh': 'subdivisionsHeight',
                                        'sr': 'sectionRadius',
                                        'sw': 'subdivisionsWidth',
                                        'tw': 'twist'},
                         'type': 'context'},
  'createRenderLayer': {'flags': {'empty': {'args': '<type __builtin__.bool>',
                                            'longname': 'empty',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                  'g': {'args': '<type __builtin__.bool>',
                                        'longname': 'g',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'g'},
                                  'makeCurrent': {'args': '<type __builtin__.bool>',
                                                  'longname': 'makeCurrent',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'mc'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'noRecurse': {'args': '<type __builtin__.bool>',
                                                'longname': 'noRecurse',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'nr'},
                                  'number': {'args': '<type __builtin__.int>',
                                             'longname': 'number',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'num'}},
                        'removedFlags': {'global': 'g'},
                        'shortFlags': {'e': 'empty',
                                       'g': 'g',
                                       'mc': 'makeCurrent',
                                       'n': 'name',
                                       'nr': 'noRecurse',
                                       'num': 'number'},
                        'type': 'rendering'},
  'createSubdivRegion': {'flags': {}, 'shortFlags': {}, 'type': 'modeling'},
  'ctxAbort': {'flags': {}, 'shortFlags': {}, 'type': 'context'},
  'ctxCompletion': {'flags': {}, 'shortFlags': {}, 'type': 'context'},
  'ctxData': {'flags': {}, 'shortFlags': {}, 'type': 'context'},
  'ctxEditMode': {'flags': {'buttonDown': {'args': '<type __builtin__.bool>',
                                           'longname': 'buttonDown',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'btd'},
                            'buttonUp': {'args': '<type __builtin__.bool>',
                                         'longname': 'buttonUp',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'btu'}},
                  'shortFlags': {'btd': 'buttonDown', 'btu': 'buttonUp'},
                  'type': 'context'},
  'ctxTraverse': {'flags': {'down': {'args': '<type __builtin__.bool>',
                                     'longname': 'down',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'd'},
                            'left': {'args': '<type __builtin__.bool>',
                                     'longname': 'left',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'l'},
                            'right': {'args': '<type __builtin__.bool>',
                                      'longname': 'right',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'r'},
                            'up': {'args': '<type __builtin__.bool>',
                                   'longname': 'up',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'up'}},
                  'shortFlags': {'d': 'down',
                                 'l': 'left',
                                 'r': 'right',
                                 'up': 'up'},
                  'type': 'context'},
  'currentCtx': {'flags': {}, 'shortFlags': {}, 'type': 'context'},
  'currentTime': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'update': {'args': '<type __builtin__.bool>',
                                       'longname': 'update',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'u'}},
                  'shortFlags': {'e': 'edit', 'q': 'query', 'u': 'update'},
                  'type': 'animation'},
  'currentTimeCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'history': {'args': '<type __builtin__.bool>',
                                           'longname': 'history',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ch'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'}},
                     'shortFlags': {'ch': 'history',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'n': 'name',
                                    'q': 'query'},
                     'type': 'context'},
  'currentUnit': {'flags': {'angle': {'args': '<type __builtin__.str>',
                                      'longname': 'angle',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'a'},
                            'fullName': {'args': '<type __builtin__.bool>',
                                         'longname': 'fullName',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'f'},
                            'linear': {'args': '<type __builtin__.str>',
                                       'longname': 'linear',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'l'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'time': {'args': '<type __builtin__.str>',
                                     'longname': 'time',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 't'},
                            'updateAnimation': {'args': '<type __builtin__.bool>',
                                                'longname': 'updateAnimation',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ua'}},
                  'shortFlags': {'a': 'angle',
                                 'f': 'fullName',
                                 'l': 'linear',
                                 'q': 'query',
                                 't': 'time',
                                 'ua': 'updateAnimation'},
                  'type': 'general'},
  'curve': {'flags': {'append': {'args': '<type __builtin__.bool>',
                                 'longname': 'append',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'a'},
                      'bezier': {'args': '<type __builtin__.bool>',
                                 'longname': 'bezier',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'bez'},
                      'degree': {'args': '<type __builtin__.float>',
                                 'longname': 'degree',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'd'},
                      'editPoint': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'editPoint',
                                    'modes': ['create', 'multiuse'],
                                    'numArgs': 3,
                                    'shortname': 'ep'},
                      'knot': {'args': '<type __builtin__.float>',
                               'longname': 'knot',
                               'modes': ['create', 'multiuse'],
                               'numArgs': 1,
                               'shortname': 'k'},
                      'name': {'args': '<type __builtin__.str>',
                               'longname': 'name',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'n'},
                      'objectSpace': {'args': '<type __builtin__.bool>',
                                      'longname': 'objectSpace',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'os'},
                      'periodic': {'args': '<type __builtin__.bool>',
                                   'longname': 'periodic',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'per'},
                      'point': {'args': ['<type __builtin__.float>',
                                         '<type __builtin__.float>',
                                         '<type __builtin__.float>'],
                                'longname': 'point',
                                'modes': ['create', 'multiuse'],
                                'numArgs': 3,
                                'shortname': 'p'},
                      'pointWeight': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'pointWeight',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 4,
                                      'shortname': 'pw'},
                      'replace': {'args': '<type __builtin__.bool>',
                                  'longname': 'replace',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'r'},
                      'worldSpace': {'args': '<type __builtin__.bool>',
                                     'longname': 'worldSpace',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ws'}},
            'shortFlags': {'a': 'append',
                           'bez': 'bezier',
                           'd': 'degree',
                           'ep': 'editPoint',
                           'k': 'knot',
                           'n': 'name',
                           'os': 'objectSpace',
                           'p': 'point',
                           'per': 'periodic',
                           'pw': 'pointWeight',
                           'r': 'replace',
                           'ws': 'worldSpace'},
            'type': 'modeling'},
  'curveAddPtCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'image1': {'args': '<type __builtin__.str>',
                                         'longname': 'image1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i1'},
                              'image2': {'args': '<type __builtin__.str>',
                                         'longname': 'image2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i2'},
                              'image3': {'args': '<type __builtin__.str>',
                                         'longname': 'image3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i3'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'}},
                    'shortFlags': {'e': 'edit',
                                   'ex': 'exists',
                                   'i1': 'image1',
                                   'i2': 'image2',
                                   'i3': 'image3',
                                   'q': 'query'},
                    'type': 'context'},
  'curveBezierCtx': {'flags': {'degree': {'args': '<type __builtin__.int>',
                                          'longname': 'degree',
                                          'numArgs': 1,
                                          'shortname': 'd'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'history': {'args': '<type __builtin__.bool>',
                                           'longname': 'history',
                                           'numArgs': 1,
                                           'shortname': 'ch'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'manipMode': {'args': '<type __builtin__.str>',
                                             'longname': 'manipMode',
                                             'numArgs': 1,
                                             'shortname': 'mm'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'preserveShape': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preserveShape',
                                                 'numArgs': 0,
                                                 'shortname': 'ps'},
                               'preserveShapeFraction': {'args': '<type __builtin__.float>',
                                                         'longname': 'preserveShapeFraction',
                                                         'numArgs': 1,
                                                         'shortname': 'pf'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'refit': {'args': '<type __builtin__.bool>',
                                         'longname': 'refit',
                                         'numArgs': 0,
                                         'shortname': 'rf'},
                               'selectMode': {'args': '<type __builtin__.str>',
                                              'longname': 'selectMode',
                                              'numArgs': 1,
                                              'shortname': 'slm'},
                               'uniform': {'args': '<type __builtin__.bool>',
                                           'longname': 'uniform',
                                           'numArgs': 1,
                                           'shortname': 'un'}},
                     'shortFlags': {'ch': 'history',
                                    'd': 'degree',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'mm': 'manipMode',
                                    'n': 'name',
                                    'pf': 'preserveShapeFraction',
                                    'ps': 'preserveShape',
                                    'q': 'query',
                                    'rf': 'refit',
                                    'slm': 'selectMode',
                                    'un': 'uniform'},
                     'type': 'context'},
  'curveCVCtx': {'flags': {'bezier': {'args': '<type __builtin__.bool>',
                                      'longname': 'bezier',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'bez'},
                           'degree': {'args': '<type __builtin__.int>',
                                      'longname': 'degree',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'd'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'history': {'args': '<type __builtin__.bool>',
                                       'longname': 'history',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ch'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'multEndKnots': {'args': '<type __builtin__.bool>',
                                            'longname': 'multEndKnots',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'me'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'preserveShape': {'args': '<type __builtin__.bool>',
                                             'longname': 'preserveShape',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 0,
                                             'shortname': 'ps'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'rational': {'args': '<type __builtin__.bool>',
                                        'longname': 'rational',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'rl'},
                           'refit': {'args': '<type __builtin__.bool>',
                                     'longname': 'refit',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'rf'},
                           'symmetry': {'args': '<type __builtin__.bool>',
                                        'longname': 'symmetry',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'sm'},
                           'uniform': {'args': '<type __builtin__.bool>',
                                       'longname': 'uniform',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'un'}},
                 'shortFlags': {'bez': 'bezier',
                                'ch': 'history',
                                'd': 'degree',
                                'e': 'edit',
                                'ex': 'exists',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'me': 'multEndKnots',
                                'n': 'name',
                                'ps': 'preserveShape',
                                'q': 'query',
                                'rf': 'refit',
                                'rl': 'rational',
                                'sm': 'symmetry',
                                'un': 'uniform'},
                 'type': 'context'},
  'curveEPCtx': {'flags': {'bezier': {'args': '<type __builtin__.bool>',
                                      'longname': 'bezier',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'bez'},
                           'degree': {'args': '<type __builtin__.int>',
                                      'longname': 'degree',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'd'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'history': {'args': '<type __builtin__.bool>',
                                       'longname': 'history',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ch'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'preserveShape': {'args': '<type __builtin__.bool>',
                                             'longname': 'preserveShape',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 0,
                                             'shortname': 'ps'},
                           'preserveShapeFraction': {'args': '<type __builtin__.float>',
                                                     'longname': 'preserveShapeFraction',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'pf'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'refit': {'args': '<type __builtin__.bool>',
                                     'longname': 'refit',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'rf'},
                           'uniform': {'args': '<type __builtin__.bool>',
                                       'longname': 'uniform',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'un'}},
                 'shortFlags': {'bez': 'bezier',
                                'ch': 'history',
                                'd': 'degree',
                                'e': 'edit',
                                'ex': 'exists',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'n': 'name',
                                'pf': 'preserveShapeFraction',
                                'ps': 'preserveShape',
                                'q': 'query',
                                'rf': 'refit',
                                'un': 'uniform'},
                 'type': 'context'},
  'curveEditorCtx': {'flags': {'direction': {'args': '<type __builtin__.int>',
                                             'longname': 'direction',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'dir'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'history': {'args': '<type __builtin__.bool>',
                                           'longname': 'history',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ch'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'relativeTangentSize': {'args': '<type __builtin__.float>',
                                                       'longname': 'relativeTangentSize',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'rts'},
                               'title': {'args': '<type __builtin__.str>',
                                         'longname': 'title',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 't'}},
                     'shortFlags': {'ch': 'history',
                                    'dir': 'direction',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'n': 'name',
                                    'q': 'query',
                                    'rts': 'relativeTangentSize',
                                    't': 'title'},
                     'type': 'context'},
  'curveIntersect': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'direction': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'direction',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 3,
                                             'shortname': 'd'},
                               'directionX': {'args': '<type __builtin__.float>',
                                              'longname': 'directionX',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dx'},
                               'directionY': {'args': '<type __builtin__.float>',
                                              'longname': 'directionY',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dy'},
                               'directionZ': {'args': '<type __builtin__.float>',
                                              'longname': 'directionZ',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dz'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'tolerance': {'args': '<type __builtin__.float>',
                                             'longname': 'tolerance',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'tol'},
                               'useDirection': {'args': '<type __builtin__.bool>',
                                                'longname': 'useDirection',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ud'}},
                     'shortFlags': {'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'd': 'direction',
                                    'dx': 'directionX',
                                    'dy': 'directionY',
                                    'dz': 'directionZ',
                                    'e': 'edit',
                                    'fzn': 'frozen',
                                    'nds': 'nodeState',
                                    'q': 'query',
                                    'tol': 'tolerance',
                                    'ud': 'useDirection'},
                     'type': 'modeling'},
  'curveMoveEPCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'}},
                     'shortFlags': {'e': 'edit',
                                    'ex': 'exists',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'q': 'query'},
                     'type': 'context'},
  'curveOnSurface': {'flags': {'append': {'args': '<type __builtin__.bool>',
                                          'longname': 'append',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'a'},
                               'degree': {'args': '<type __builtin__.float>',
                                          'longname': 'degree',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'd'},
                               'knot': {'args': '<type __builtin__.float>',
                                        'longname': 'knot',
                                        'modes': ['create', 'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'k'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'periodic': {'args': '<type __builtin__.bool>',
                                            'longname': 'periodic',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'per'},
                               'positionUV': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'positionUV',
                                              'modes': ['create',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'uv'},
                               'replace': {'args': '<type __builtin__.bool>',
                                           'longname': 'replace',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'r'}},
                     'shortFlags': {'a': 'append',
                                    'd': 'degree',
                                    'k': 'knot',
                                    'n': 'name',
                                    'per': 'periodic',
                                    'r': 'replace',
                                    'uv': 'positionUV'},
                     'type': 'modeling'},
  'curveRGBColor': {'flags': {'hueSaturationValue': {'args': '<type __builtin__.bool>',
                                                     'longname': 'hueSaturationValue',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 0,
                                                     'shortname': 'hsv'},
                              'list': {'args': '<type __builtin__.bool>',
                                       'longname': 'list',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'l'},
                              'listNames': {'args': '<type __builtin__.bool>',
                                            'longname': 'listNames',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ln'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'remove': {'args': '<type __builtin__.bool>',
                                         'longname': 'remove',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'r'},
                              'resetToFactory': {'args': '<type __builtin__.bool>',
                                                 'longname': 'resetToFactory',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'rf'},
                              'resetToSaved': {'args': '<type __builtin__.bool>',
                                               'longname': 'resetToSaved',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'rs'}},
                    'shortFlags': {'hsv': 'hueSaturationValue',
                                   'l': 'list',
                                   'ln': 'listNames',
                                   'q': 'query',
                                   'r': 'remove',
                                   'rf': 'resetToFactory',
                                   'rs': 'resetToSaved'},
                    'type': 'general'},
  'curveSketchCtx': {'flags': {'degree': {'args': '<type __builtin__.int>',
                                          'longname': 'degree',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'd'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'history': {'args': '<type __builtin__.bool>',
                                           'longname': 'history',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ch'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'}},
                     'shortFlags': {'ch': 'history',
                                    'd': 'degree',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'n': 'name',
                                    'q': 'query'},
                     'type': 'context'},
  'cutKey': {'flags': {'animation': {'args': '<type __builtin__.str>',
                                     'longname': 'animation',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'an'},
                       'attribute': {'args': '<type __builtin__.str>',
                                     'longname': 'attribute',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'at'},
                       'clear': {'args': '<type __builtin__.bool>',
                                 'longname': 'clear',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'cl'},
                       'controlPoints': {'args': '<type __builtin__.bool>',
                                         'longname': 'controlPoints',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'cp'},
                       'float': {'args': 'floatrange',
                                 'longname': 'float',
                                 'modes': ['create', 'multiuse'],
                                 'numArgs': 1,
                                 'shortname': 'f'},
                       'hierarchy': {'args': '<type __builtin__.str>',
                                     'longname': 'hierarchy',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'hi'},
                       'includeUpperBound': {'args': '<type __builtin__.bool>',
                                             'longname': 'includeUpperBound',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'iub'},
                       'index': {'args': '<type __builtin__.int>',
                                 'longname': 'index',
                                 'modes': ['create', 'multiuse'],
                                 'numArgs': 1,
                                 'shortname': 'index'},
                       'option': {'args': '<type __builtin__.str>',
                                  'longname': 'option',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'o'},
                       'selectKey': {'args': '<type __builtin__.bool>',
                                     'longname': 'selectKey',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'sl'},
                       'shape': {'args': '<type __builtin__.bool>',
                                 'longname': 'shape',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 's'},
                       'time': {'args': 'timerange',
                                'longname': 'time',
                                'modes': ['create', 'multiuse'],
                                'numArgs': 1,
                                'shortname': 't'}},
             'removedFlags': {'in': 'index'},
             'shortFlags': {'an': 'animation',
                            'at': 'attribute',
                            'cl': 'clear',
                            'cp': 'controlPoints',
                            'f': 'float',
                            'hi': 'hierarchy',
                            'index': 'index',
                            'iub': 'includeUpperBound',
                            'o': 'option',
                            's': 'shape',
                            'sl': 'selectKey',
                            't': 'time'},
             'type': 'animation'},
  'cycleCheck': {'flags': {'all': {'args': '<type __builtin__.bool>',
                                   'longname': 'all',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'all'},
                           'children': {'args': '<type __builtin__.bool>',
                                        'longname': 'children',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'c'},
                           'dag': {'args': '<type __builtin__.bool>',
                                   'longname': 'dag',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'dag'},
                           'evaluation': {'args': '<type __builtin__.bool>',
                                          'longname': 'evaluation',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'e'},
                           'firstCycleOnly': {'args': '<type __builtin__.bool>',
                                              'longname': 'firstCycleOnly',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'fco'},
                           'firstPlugPerNode': {'args': '<type __builtin__.bool>',
                                                'longname': 'firstPlugPerNode',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'fpn'},
                           'lastPlugPerNode': {'args': '<type __builtin__.bool>',
                                               'longname': 'lastPlugPerNode',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'lpn'},
                           'list': {'args': '<type __builtin__.bool>',
                                    'longname': 'list',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'l'},
                           'listSeparator': {'args': '<type __builtin__.str>',
                                             'longname': 'listSeparator',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ls'},
                           'parents': {'args': '<type __builtin__.bool>',
                                       'longname': 'parents',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'p'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'secondary': {'args': '<type __builtin__.bool>',
                                         'longname': 'secondary',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 's'},
                           'timeLimit': {'args': 'time',
                                         'longname': 'timeLimit',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'tl'}},
                 'shortFlags': {'all': 'all',
                                'c': 'children',
                                'dag': 'dag',
                                'e': 'evaluation',
                                'fco': 'firstCycleOnly',
                                'fpn': 'firstPlugPerNode',
                                'l': 'list',
                                'lpn': 'lastPlugPerNode',
                                'ls': 'listSeparator',
                                'p': 'parents',
                                'q': 'query',
                                's': 'secondary',
                                'tl': 'timeLimit'},
                 'type': 'general'},
  'cylinder': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                           '<type __builtin__.float>',
                                           '<type __builtin__.float>'],
                                  'longname': 'axis',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 3,
                                  'shortname': 'ax'},
                         'caching': {'args': '<type __builtin__.bool>',
                                     'longname': 'caching',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsCasting': True,
                                     'shortname': 'cch'},
                         'constructionHistory': {'args': '<type __builtin__.bool>',
                                                 'longname': 'constructionHistory',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                         'degree': {'args': '<type __builtin__.int>',
                                    'longname': 'degree',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'resultNeedsCasting': True,
                                    'shortname': 'd'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'endSweep': {'args': '<type __builtin__.float>',
                                      'longname': 'endSweep',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'esw'},
                         'frozen': {'args': '<type __builtin__.bool>',
                                    'longname': 'frozen',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'fzn'},
                         'heightRatio': {'args': '<type __builtin__.float>',
                                         'longname': 'heightRatio',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'hr'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'nodeState': {'args': '<type __builtin__.int>',
                                       'longname': 'nodeState',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'nds'},
                         'object': {'args': '<type __builtin__.bool>',
                                    'longname': 'object',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'o'},
                         'pivot': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'pivot',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 3,
                                   'shortname': 'p'},
                         'polygon': {'args': '<type __builtin__.int>',
                                     'longname': 'polygon',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'po'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'radius': {'args': '<type __builtin__.float>',
                                    'longname': 'radius',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'r'},
                         'sections': {'args': '<type __builtin__.int>',
                                      'longname': 'sections',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsCasting': True,
                                      'shortname': 's'},
                         'spans': {'args': '<type __builtin__.int>',
                                   'longname': 'spans',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'resultNeedsCasting': True,
                                   'shortname': 'nsp'},
                         'startSweep': {'args': '<type __builtin__.float>',
                                        'longname': 'startSweep',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ssw'},
                         'tolerance': {'args': '<type __builtin__.float>',
                                       'longname': 'tolerance',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'tol'},
                         'useTolerance': {'args': '<type __builtin__.bool>',
                                          'longname': 'useTolerance',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'resultNeedsCasting': True,
                                          'shortname': 'ut'}},
               'shortFlags': {'ax': 'axis',
                              'cch': 'caching',
                              'ch': 'constructionHistory',
                              'd': 'degree',
                              'e': 'edit',
                              'esw': 'endSweep',
                              'fzn': 'frozen',
                              'hr': 'heightRatio',
                              'n': 'name',
                              'nds': 'nodeState',
                              'nsp': 'spans',
                              'o': 'object',
                              'p': 'pivot',
                              'po': 'polygon',
                              'q': 'query',
                              'r': 'radius',
                              's': 'sections',
                              'ssw': 'startSweep',
                              'tol': 'tolerance',
                              'ut': 'useTolerance'},
               'type': 'modeling'},
  'dagCommandWrapper': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'dagObjectCompare': {'flags': {'attribute': {'args': '<type __builtin__.bool>',
                                               'longname': 'attribute',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'a'},
                                 'bail': {'args': '<type __builtin__.str>',
                                          'longname': 'bail',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'b'},
                                 'connection': {'args': '<type __builtin__.bool>',
                                                'longname': 'connection',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'c'},
                                 'namespace': {'args': '<type __builtin__.str>',
                                               'longname': 'namespace',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'n'},
                                 'relative': {'args': '<type __builtin__.bool>',
                                              'longname': 'relative',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'r'},
                                 'short': {'args': '<type __builtin__.bool>',
                                           'longname': 'short',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 's'},
                                 'type': {'args': '<type __builtin__.bool>',
                                          'longname': 'type',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 't'}},
                       'shortFlags': {'a': 'attribute',
                                      'b': 'bail',
                                      'c': 'connection',
                                      'n': 'namespace',
                                      'r': 'relative',
                                      's': 'short',
                                      't': 'type'},
                       'type': 'system'},
  'dagObjectHit': {'flags': {'cache': {'args': '<type __builtin__.bool>',
                                       'longname': 'cache',
                                       'numArgs': 0,
                                       'shortname': 'ch'},
                             'menu': {'args': '<type __builtin__.str>',
                                      'longname': 'menu',
                                      'numArgs': 1,
                                      'shortname': 'mn'},
                             'multiple': {'args': '<type __builtin__.bool>',
                                          'longname': 'multiple',
                                          'numArgs': 0,
                                          'shortname': 'm'},
                             'targetSize': {'args': '<type __builtin__.int>',
                                            'longname': 'targetSize',
                                            'numArgs': 1,
                                            'shortname': 'ts'}},
                   'shortFlags': {'ch': 'cache',
                                  'm': 'multiple',
                                  'mn': 'menu',
                                  'ts': 'targetSize'},
                   'type': 'other'},
  'dagPose': {'flags': {'addToPose': {'args': '<type __builtin__.bool>',
                                      'longname': 'addToPose',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'a'},
                        'atPose': {'args': '<type __builtin__.bool>',
                                   'longname': 'atPose',
                                   'modes': ['query'],
                                   'numArgs': 0,
                                   'shortname': 'ap'},
                        'bindPose': {'args': '<type __builtin__.bool>',
                                     'longname': 'bindPose',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 'bp'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'g': {'args': '<type __builtin__.bool>',
                              'longname': 'g',
                              'modes': ['create'],
                              'numArgs': 0,
                              'shortname': 'g'},
                        'members': {'args': '<type __builtin__.bool>',
                                    'longname': 'members',
                                    'modes': ['query'],
                                    'numArgs': 0,
                                    'shortname': 'm'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['create', 'query'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'remove': {'args': '<type __builtin__.bool>',
                                   'longname': 'remove',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'rm'},
                        'reset': {'args': '<type __builtin__.bool>',
                                  'longname': 'reset',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'rs'},
                        'restore': {'args': '<type __builtin__.bool>',
                                    'longname': 'restore',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'r'},
                        'save': {'args': '<type __builtin__.bool>',
                                 'longname': 'save',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 's'},
                        'selection': {'args': '<type __builtin__.bool>',
                                      'longname': 'selection',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'sl'},
                        'worldParent': {'args': '<type __builtin__.bool>',
                                        'longname': 'worldParent',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'wp'}},
              'removedFlags': {'global': 'g'},
              'shortFlags': {'a': 'addToPose',
                             'ap': 'atPose',
                             'bp': 'bindPose',
                             'e': 'edit',
                             'g': 'g',
                             'm': 'members',
                             'n': 'name',
                             'q': 'query',
                             'r': 'restore',
                             'rm': 'remove',
                             'rs': 'reset',
                             's': 'save',
                             'sl': 'selection',
                             'wp': 'worldParent'},
              'type': 'animation'},
  'dataStructure': {'flags': {'asFile': {'args': '<type __builtin__.str>',
                                         'longname': 'asFile',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'af'},
                              'asString': {'args': '<type __builtin__.str>',
                                           'longname': 'asString',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'asString'},
                              'dataType': {'args': '<type __builtin__.bool>',
                                           'longname': 'dataType',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'dt'},
                              'format': {'args': '<type __builtin__.str>',
                                         'longname': 'format',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'fmt'},
                              'listMemberNames': {'args': '<type __builtin__.str>',
                                                  'longname': 'listMemberNames',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'lmn'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'remove': {'args': '<type __builtin__.bool>',
                                         'longname': 'remove',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'rem'},
                              'removeAll': {'args': '<type __builtin__.bool>',
                                            'longname': 'removeAll',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ral'}},
                    'removedFlags': {'as': 'asString'},
                    'shortFlags': {'af': 'asFile',
                                   'asString': 'asString',
                                   'dt': 'dataType',
                                   'fmt': 'format',
                                   'lmn': 'listMemberNames',
                                   'n': 'name',
                                   'q': 'query',
                                   'ral': 'removeAll',
                                   'rem': 'remove'},
                    'type': 'modeling'},
  'date': {'flags': {'date': {'args': '<type __builtin__.bool>',
                              'longname': 'date',
                              'modes': ['create'],
                              'numArgs': 0,
                              'shortname': 'd'},
                     'format': {'args': '<type __builtin__.str>',
                                'longname': 'format',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'f'},
                     'shortDate': {'args': '<type __builtin__.bool>',
                                   'longname': 'shortDate',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'sd'},
                     'shortTime': {'args': '<type __builtin__.bool>',
                                   'longname': 'shortTime',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'st'},
                     'time': {'args': '<type __builtin__.bool>',
                              'longname': 'time',
                              'modes': ['create'],
                              'numArgs': 0,
                              'shortname': 't'}},
           'shortFlags': {'d': 'date',
                          'f': 'format',
                          'sd': 'shortDate',
                          'st': 'shortTime',
                          't': 'time'},
           'type': 'system'},
  'dbcount': {'flags': {'enabled': {'args': '<type __builtin__.bool>',
                                    'longname': 'enabled',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'e'},
                        'file': {'args': '<type __builtin__.str>',
                                 'longname': 'file',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'f'},
                        'keyword': {'args': '<type __builtin__.str>',
                                    'longname': 'keyword',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'k'},
                        'list': {'args': '<type __builtin__.bool>',
                                 'longname': 'list',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'l'},
                        'maxdepth': {'args': '<type __builtin__.int>',
                                     'longname': 'maxdepth',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'md'},
                        'quick': {'args': '<type __builtin__.bool>',
                                  'longname': 'quick',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'reset': {'args': '<type __builtin__.bool>',
                                  'longname': 'reset',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'r'},
                        'spreadsheet': {'args': '<type __builtin__.bool>',
                                        'longname': 'spreadsheet',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 's'}},
              'shortFlags': {'e': 'enabled',
                             'f': 'file',
                             'k': 'keyword',
                             'l': 'list',
                             'md': 'maxdepth',
                             'q': 'quick',
                             'r': 'reset',
                             's': 'spreadsheet'},
              'type': 'system'},
  'dbfootprint': {'flags': {'allObjects': {'args': '<type __builtin__.bool>',
                                           'longname': 'allObjects',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'all'},
                            'outputFile': {'args': '<type __builtin__.str>',
                                           'longname': 'outputFile',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'of'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'type': {'args': '<type __builtin__.str>',
                                     'longname': 'type',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 't'}},
                  'shortFlags': {'all': 'allObjects',
                                 'of': 'outputFile',
                                 'q': 'query',
                                 't': 'type'},
                  'type': 'system'},
  'dbmessage': {'flags': {'file': {'args': '<type __builtin__.str>',
                                   'longname': 'file',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'f'},
                          'list': {'args': '<type __builtin__.bool>',
                                   'longname': 'list',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'l'},
                          'monitor': {'args': '<type __builtin__.bool>',
                                      'longname': 'monitor',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                          'type': {'args': '<type __builtin__.str>',
                                   'longname': 'type',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 't'}},
                'shortFlags': {'f': 'file',
                               'l': 'list',
                               'm': 'monitor',
                               't': 'type'},
                'type': 'system'},
  'dbpeek': {'flags': {'allObjects': {'args': '<type __builtin__.bool>',
                                      'longname': 'allObjects',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'all'},
                       'argument': {'args': '<type __builtin__.str>',
                                    'longname': 'argument',
                                    'modes': ['create',
                                              'query',
                                              'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'a'},
                       'count': {'args': '<type __builtin__.int>',
                                 'longname': 'count',
                                 'modes': ['create', 'query'],
                                 'numArgs': 1,
                                 'shortname': 'c'},
                       'evaluationGraph': {'args': '<type __builtin__.bool>',
                                           'longname': 'evaluationGraph',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'eg'},
                       'operation': {'args': '<type __builtin__.str>',
                                     'longname': 'operation',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'op'},
                       'outputFile': {'args': '<type __builtin__.str>',
                                      'longname': 'outputFile',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'of'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'}},
             'shortFlags': {'a': 'argument',
                            'all': 'allObjects',
                            'c': 'count',
                            'eg': 'evaluationGraph',
                            'of': 'outputFile',
                            'op': 'operation',
                            'q': 'query'},
             'type': 'system'},
  'dbtrace': {'flags': {'filter': {'args': '<type __builtin__.str>',
                                   'longname': 'filter',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'f'},
                        'info': {'args': '<type __builtin__.bool>',
                                 'longname': 'info',
                                 'modes': ['query'],
                                 'numArgs': 0,
                                 'shortname': 'i'},
                        'keyword': {'args': '<type __builtin__.str>',
                                    'longname': 'keyword',
                                    'modes': ['create',
                                              'query',
                                              'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'k'},
                        'mark': {'args': '<type __builtin__.bool>',
                                 'longname': 'mark',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'm'},
                        'off': {'args': 'Toggle the traces off.  If omitted it will turn them on.',
                                'longname': 'off',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'boolean'},
                        'output': {'args': '<type __builtin__.str>',
                                   'longname': 'output',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'o'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'timed': {'args': '<type __builtin__.bool>',
                                  'longname': 'timed',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'tm'},
                        'title': {'args': '<type __builtin__.str>',
                                  'longname': 'title',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 't'},
                        'verbose': {'args': '<type __builtin__.bool>',
                                    'longname': 'verbose',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'v'}},
              'shortFlags': {'f': 'filter',
                             'i': 'info',
                             'k': 'keyword',
                             'm': 'mark',
                             'o': 'output',
                             'off': 'off',
                             'q': 'query',
                             't': 'title',
                             'tm': 'timed',
                             'v': 'verbose'},
              'type': 'system'},
  'debug': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'debugNamespace': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'debugVar': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'defaultLightListCheckBox': {'callbackFlags': ['visibleChangeCommand',
                                                 'vcc',
                                                 'dragCallback',
                                                 'dgc',
                                                 'dropCallback',
                                                 'dpc'],
                               'flags': {'annotation': {'args': '<type __builtin__.str>',
                                                        'longname': 'annotation',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ann'},
                                         'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                      '<type __builtin__.float>',
                                                                      '<type __builtin__.float>'],
                                                             'longname': 'backgroundColor',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 3,
                                                             'shortname': 'bgc'},
                                         'defineTemplate': {'args': '<type __builtin__.str>',
                                                            'longname': 'defineTemplate',
                                                            'modes': ['create'],
                                                            'numArgs': 1,
                                                            'shortname': 'dt'},
                                         'docTag': {'args': '<type __builtin__.str>',
                                                    'longname': 'docTag',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dtg'},
                                         'dragCallback': {'args': 'script',
                                                          'longname': 'dragCallback',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'dgc'},
                                         'dropCallback': {'args': 'script',
                                                          'longname': 'dropCallback',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'dpc'},
                                         'edit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'edit',
                                                  'numArgs': 0,
                                                  'shortname': 'e'},
                                         'enable': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enable',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'en'},
                                         'enableBackground': {'args': '<type __builtin__.bool>',
                                                              'longname': 'enableBackground',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'ebg'},
                                         'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'enableKeyboardFocus',
                                                                 'modes': ['create',
                                                                           'query',
                                                                           'edit'],
                                                                 'numArgs': 1,
                                                                 'shortname': 'ekf'},
                                         'exists': {'args': '<type __builtin__.bool>',
                                                    'longname': 'exists',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ex'},
                                         'fullPathName': {'args': '<type __builtin__.bool>',
                                                          'longname': 'fullPathName',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'fpn'},
                                         'height': {'args': '<type __builtin__.int>',
                                                    'longname': 'height',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'h'},
                                         'highlightColor': {'args': ['<type __builtin__.float>',
                                                                     '<type __builtin__.float>',
                                                                     '<type __builtin__.float>'],
                                                            'longname': 'highlightColor',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 3,
                                                            'shortname': 'hlc'},
                                         'isObscured': {'args': '<type __builtin__.bool>',
                                                        'longname': 'isObscured',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'io'},
                                         'label': {'args': '<type __builtin__.str>',
                                                   'longname': 'label',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'l'},
                                         'manage': {'args': '<type __builtin__.bool>',
                                                    'longname': 'manage',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'm'},
                                         'noBackground': {'args': '<type __builtin__.bool>',
                                                          'longname': 'noBackground',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'nbg'},
                                         'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                                'longname': 'numberOfPopupMenus',
                                                                'modes': ['query'],
                                                                'numArgs': 0,
                                                                'shortname': 'npm'},
                                         'parent': {'args': '<type __builtin__.str>',
                                                    'longname': 'parent',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'p'},
                                         'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                            'longname': 'popupMenuArray',
                                                            'modes': ['query'],
                                                            'numArgs': 0,
                                                            'shortname': 'pma'},
                                         'preventOverride': {'args': '<type __builtin__.bool>',
                                                             'longname': 'preventOverride',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'po'},
                                         'query': {'args': '<type __builtin__.bool>',
                                                   'longname': 'query',
                                                   'numArgs': 0,
                                                   'shortname': 'q'},
                                         'shadingGroup': {'args': 'PyNode',
                                                          'longname': 'shadingGroup',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'sg'},
                                         'statusBarMessage': {'args': '<type __builtin__.str>',
                                                              'longname': 'statusBarMessage',
                                                              'modes': ['create',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'sbm'},
                                         'useTemplate': {'args': '<type __builtin__.str>',
                                                         'longname': 'useTemplate',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'ut'},
                                         'visible': {'args': '<type __builtin__.bool>',
                                                     'longname': 'visible',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vis'},
                                         'visibleChangeCommand': {'args': 'script',
                                                                  'longname': 'visibleChangeCommand',
                                                                  'modes': ['create',
                                                                            'query',
                                                                            'edit'],
                                                                  'numArgs': 1,
                                                                  'shortname': 'vcc'},
                                         'width': {'args': '<type __builtin__.int>',
                                                   'longname': 'width',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'w'}},
                               'shortFlags': {'ann': 'annotation',
                                              'bgc': 'backgroundColor',
                                              'dgc': 'dragCallback',
                                              'dpc': 'dropCallback',
                                              'dt': 'defineTemplate',
                                              'dtg': 'docTag',
                                              'e': 'edit',
                                              'ebg': 'enableBackground',
                                              'ekf': 'enableKeyboardFocus',
                                              'en': 'enable',
                                              'ex': 'exists',
                                              'fpn': 'fullPathName',
                                              'h': 'height',
                                              'hlc': 'highlightColor',
                                              'io': 'isObscured',
                                              'l': 'label',
                                              'm': 'manage',
                                              'nbg': 'noBackground',
                                              'npm': 'numberOfPopupMenus',
                                              'p': 'parent',
                                              'pma': 'popupMenuArray',
                                              'po': 'preventOverride',
                                              'q': 'query',
                                              'sbm': 'statusBarMessage',
                                              'sg': 'shadingGroup',
                                              'ut': 'useTemplate',
                                              'vcc': 'visibleChangeCommand',
                                              'vis': 'visible',
                                              'w': 'width'},
                               'type': 'rendering'},
  'defaultNavigation': {'flags': {'connectToExisting': {'args': '<type __builtin__.bool>',
                                                        'longname': 'connectToExisting',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'ce'},
                                  'createNew': {'args': '<type __builtin__.bool>',
                                                'longname': 'createNew',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'cn'},
                                  'defaultAttribute': {'args': '<type __builtin__.bool>',
                                                       'longname': 'defaultAttribute',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'da'},
                                  'defaultTraversal': {'args': '<type __builtin__.bool>',
                                                       'longname': 'defaultTraversal',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'dtv'},
                                  'defaultWorkingNode': {'args': '<type __builtin__.bool>',
                                                         'longname': 'defaultWorkingNode',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'dwn'},
                                  'delete': {'args': '<type __builtin__.bool>',
                                             'longname': 'delete',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'delete'},
                                  'destination': {'args': 'PyNode',
                                                  'longname': 'destination',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'd'},
                                  'disconnect': {'args': '<type __builtin__.bool>',
                                                 'longname': 'disconnect',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'dis'},
                                  'force': {'args': '<type __builtin__.bool>',
                                            'longname': 'force',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'f'},
                                  'ignore': {'args': '<type __builtin__.bool>',
                                             'longname': 'ignore',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'i'},
                                  'navigatorDecisionString': {'args': '<type __builtin__.str>',
                                                              'longname': 'navigatorDecisionString',
                                                              'modes': ['create'],
                                                              'numArgs': 1,
                                                              'shortname': 'nds'},
                                  'quiet': {'args': '<type __builtin__.bool>',
                                            'longname': 'quiet',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'qt'},
                                  'relatedNodes': {'args': '<type __builtin__.bool>',
                                                   'longname': 'relatedNodes',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'ren'},
                                  'source': {'args': 'PyNode',
                                             'longname': 'source',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 's'},
                                  'unignore': {'args': '<type __builtin__.bool>',
                                               'longname': 'unignore',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'u'}},
                        'removedFlags': {'del': 'delete'},
                        'shortFlags': {'ce': 'connectToExisting',
                                       'cn': 'createNew',
                                       'd': 'destination',
                                       'da': 'defaultAttribute',
                                       'delete': 'delete',
                                       'dis': 'disconnect',
                                       'dtv': 'defaultTraversal',
                                       'dwn': 'defaultWorkingNode',
                                       'f': 'force',
                                       'i': 'ignore',
                                       'nds': 'navigatorDecisionString',
                                       'qt': 'quiet',
                                       'ren': 'relatedNodes',
                                       's': 'source',
                                       'u': 'unignore'},
                        'type': 'windows'},
  'defineDataServer': {'flags': {'device': {'args': '<type __builtin__.str>',
                                            'longname': 'device',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'd'},
                                 'server': {'args': '<type __builtin__.str>',
                                            'longname': 'server',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 's'},
                                 'undefine': {'args': '<type __builtin__.bool>',
                                              'longname': 'undefine',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'u'}},
                       'shortFlags': {},
                       'type': 'animation'},
  'defineVirtualDevice': {'flags': {'axis': {'args': '<type __builtin__.int>',
                                             'longname': 'axis',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ax'},
                                    'channel': {'args': '<type __builtin__.str>',
                                                'longname': 'channel',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'c'},
                                    'clear': {'args': '<type __builtin__.bool>',
                                              'longname': 'clear',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'cl'},
                                    'create': {'args': '<type __builtin__.bool>',
                                               'longname': 'create',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'cr'},
                                    'device': {'args': '<type __builtin__.str>',
                                               'longname': 'device',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'd'},
                                    'parent': {'args': '<type __builtin__.str>',
                                               'longname': 'parent',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'p'},
                                    'undefine': {'args': '<type __builtin__.bool>',
                                                 'longname': 'undefine',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'u'},
                                    'usage': {'args': '<type __builtin__.str>',
                                              'longname': 'usage',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'use'}},
                          'shortFlags': {},
                          'type': 'animation'},
  'deformer': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                   'longname': 'after',
                                   'modes': ['create', 'edit'],
                                   'numArgs': 0,
                                   'shortname': 'af'},
                         'afterReference': {'args': '<type __builtin__.bool>',
                                            'longname': 'afterReference',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'ar'},
                         'before': {'args': '<type __builtin__.bool>',
                                    'longname': 'before',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'bf'},
                         'deformerTools': {'args': '<type __builtin__.bool>',
                                           'longname': 'deformerTools',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'dt'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'exclusive': {'args': '<type __builtin__.str>',
                                       'longname': 'exclusive',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'ex'},
                         'frontOfChain': {'args': '<type __builtin__.bool>',
                                          'longname': 'frontOfChain',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'foc'},
                         'geometry': {'args': '<type __builtin__.str>',
                                      'longname': 'geometry',
                                      'modes': ['query',
                                                'edit',
                                                'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'g'},
                         'geometryIndices': {'args': '<type __builtin__.bool>',
                                             'longname': 'geometryIndices',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'gi'},
                         'ignoreSelected': {'args': '<type __builtin__.bool>',
                                            'longname': 'ignoreSelected',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ignoreSelected'},
                         'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                     'longname': 'includeHiddenSelections',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'ihs'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'parallel': {'args': '<type __builtin__.bool>',
                                      'longname': 'parallel',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'par'},
                         'prune': {'args': '<type __builtin__.bool>',
                                   'longname': 'prune',
                                   'modes': ['edit'],
                                   'numArgs': 0,
                                   'shortname': 'pr'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'remove': {'args': '<type __builtin__.bool>',
                                    'longname': 'remove',
                                    'modes': ['edit', 'multiuse'],
                                    'numArgs': 0,
                                    'shortname': 'rm'},
                         'split': {'args': '<type __builtin__.bool>',
                                   'longname': 'split',
                                   'modes': ['create', 'edit'],
                                   'numArgs': 0,
                                   'shortname': 'sp'},
                         'type': {'args': '<type __builtin__.str>',
                                  'longname': 'type',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'typ'}},
               'removedFlags': {'is': 'ignoreSelected'},
               'shortFlags': {'af': 'after',
                              'ar': 'afterReference',
                              'bf': 'before',
                              'dt': 'deformerTools',
                              'e': 'edit',
                              'ex': 'exclusive',
                              'foc': 'frontOfChain',
                              'g': 'geometry',
                              'gi': 'geometryIndices',
                              'ignoreSelected': 'ignoreSelected',
                              'ihs': 'includeHiddenSelections',
                              'n': 'name',
                              'par': 'parallel',
                              'pr': 'prune',
                              'q': 'query',
                              'rm': 'remove',
                              'sp': 'split',
                              'typ': 'type'},
               'type': 'animation'},
  'deformerWeights': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                              'longname': 'attribute',
                                              'modes': ['create',
                                                        'query',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'at'},
                                'defaultValue': {'args': '<type __builtin__.float>',
                                                 'longname': 'defaultValue',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dv'},
                                'deformer': {'args': '<type __builtin__.str>',
                                             'longname': 'deformer',
                                             'modes': ['create',
                                                       'query',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'df'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'export': {'args': '<type __builtin__.bool>',
                                           'longname': 'export',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'format': {'args': '<type __builtin__.str>',
                                           'longname': 'format',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'fm'},
                                'ignoreName': {'args': '<type __builtin__.bool>',
                                               'longname': 'ignoreName',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 0,
                                               'shortname': 'ig'},
                                'im': {'args': '<type __builtin__.bool>',
                                       'longname': 'im',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'im'},
                                'method': {'args': '<type __builtin__.str>',
                                           'longname': 'method',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'm'},
                                'path': {'args': '<type __builtin__.str>',
                                         'longname': 'path',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'p'},
                                'positionTolerance': {'args': '<type __builtin__.float>',
                                                      'longname': 'positionTolerance',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'pt'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'remap': {'args': '<type __builtin__.str>',
                                          'longname': 'remap',
                                          'modes': ['create',
                                                    'query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'r'},
                                'shape': {'args': '<type __builtin__.str>',
                                          'longname': 'shape',
                                          'modes': ['create',
                                                    'query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'sh'},
                                'skip': {'args': '<type __builtin__.str>',
                                         'longname': 'skip',
                                         'modes': ['create',
                                                   'query',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'sk'},
                                'vertexConnections': {'args': '<type __builtin__.bool>',
                                                      'longname': 'vertexConnections',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'vc'},
                                'weightPrecision': {'args': '<type __builtin__.int>',
                                                    'longname': 'weightPrecision',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'wp'},
                                'weightTolerance': {'args': '<type __builtin__.float>',
                                                    'longname': 'weightTolerance',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'wt'},
                                'worldSpace': {'args': '<type __builtin__.bool>',
                                               'longname': 'worldSpace',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 0,
                                               'shortname': 'ws'}},
                      'removedFlags': {'import': 'im'},
                      'shortFlags': {'at': 'attribute',
                                     'df': 'deformer',
                                     'dv': 'defaultValue',
                                     'e': 'edit',
                                     'ex': 'export',
                                     'fm': 'format',
                                     'ig': 'ignoreName',
                                     'im': 'im',
                                     'm': 'method',
                                     'p': 'path',
                                     'pt': 'positionTolerance',
                                     'q': 'query',
                                     'r': 'remap',
                                     'sh': 'shape',
                                     'sk': 'skip',
                                     'vc': 'vertexConnections',
                                     'wp': 'weightPrecision',
                                     'ws': 'worldSpace',
                                     'wt': 'weightTolerance'},
                      'type': 'animation'},
  'delete': {'flags': {'all': {'args': '<type __builtin__.bool>',
                               'longname': 'all',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'all'},
                       'attribute': {'args': '<type __builtin__.str>',
                                     'longname': 'attribute',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'at'},
                       'channels': {'args': '<type __builtin__.bool>',
                                    'longname': 'channels',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'c'},
                       'constraints': {'args': '<type __builtin__.bool>',
                                       'longname': 'constraints',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'cn'},
                       'constructionHistory': {'args': '<type __builtin__.bool>',
                                               'longname': 'constructionHistory',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ch'},
                       'controlPoints': {'args': '<type __builtin__.bool>',
                                         'longname': 'controlPoints',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'cp'},
                       'expressions': {'args': '<type __builtin__.bool>',
                                       'longname': 'expressions',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'e'},
                       'hierarchy': {'args': '<type __builtin__.str>',
                                     'longname': 'hierarchy',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'hi'},
                       'inputConnectionsAndNodes': {'args': '<type __builtin__.bool>',
                                                    'longname': 'inputConnectionsAndNodes',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'icn'},
                       'motionPaths': {'args': '<type __builtin__.bool>',
                                       'longname': 'motionPaths',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'mp'},
                       'shape': {'args': '<type __builtin__.bool>',
                                 'longname': 'shape',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 's'},
                       'staticChannels': {'args': '<type __builtin__.bool>',
                                          'longname': 'staticChannels',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'sc'},
                       'timeAnimationCurves': {'args': '<type __builtin__.bool>',
                                               'longname': 'timeAnimationCurves',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'tac'},
                       'unitlessAnimationCurves': {'args': '<type __builtin__.bool>',
                                                   'longname': 'unitlessAnimationCurves',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'uac'}},
             'shortFlags': {'all': 'all',
                            'at': 'attribute',
                            'c': 'channels',
                            'ch': 'constructionHistory',
                            'cn': 'constraints',
                            'cp': 'controlPoints',
                            'e': 'expressions',
                            'hi': 'hierarchy',
                            'icn': 'inputConnectionsAndNodes',
                            'mp': 'motionPaths',
                            's': 'shape',
                            'sc': 'staticChannels',
                            'tac': 'timeAnimationCurves',
                            'uac': 'unitlessAnimationCurves'},
             'type': 'general'},
  'deleteAttr': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                         'longname': 'attribute',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'at'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'at': 'attribute',
                                'e': 'edit',
                                'n': 'name',
                                'q': 'query'},
                 'type': 'general'},
  'deleteAttrPattern': {'flags': {'allPatterns': {'args': '<type __builtin__.bool>',
                                                  'longname': 'allPatterns',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'all'},
                                  'patternName': {'args': '<type __builtin__.str>',
                                                  'longname': 'patternName',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'pn'},
                                  'patternType': {'args': '<type __builtin__.str>',
                                                  'longname': 'patternType',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'pt'}},
                        'shortFlags': {'all': 'allPatterns',
                                       'pn': 'patternName',
                                       'pt': 'patternType'},
                        'type': 'general'},
  'deleteExtension': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                              'longname': 'attribute',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'at'},
                                'forceDelete': {'args': '<type __builtin__.bool>',
                                                'longname': 'forceDelete',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'fd'},
                                'nodeType': {'args': '<type __builtin__.str>',
                                             'longname': 'nodeType',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'nt'}},
                      'shortFlags': {'at': 'attribute',
                                     'fd': 'forceDelete',
                                     'nt': 'nodeType'},
                      'type': 'general'},
  'deleteGeometryCache': {'type': 'runtime'},
  'deleteHistoryAheadOfGeomCache': {'type': 'runtime'},
  'deleteNclothCache': {'type': 'runtime'},
  'deleteUI': {'flags': {'collection': {'args': '<type __builtin__.bool>',
                                        'longname': 'collection',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'cl'},
                         'control': {'args': '<type __builtin__.bool>',
                                     'longname': 'control',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ctl'},
                         'editor': {'args': '<type __builtin__.bool>',
                                    'longname': 'editor',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ed'},
                         'layout': {'args': '<type __builtin__.bool>',
                                    'longname': 'layout',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'lay'},
                         'menu': {'args': '<type __builtin__.bool>',
                                  'longname': 'menu',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'm'},
                         'menuItem': {'args': '<type __builtin__.bool>',
                                      'longname': 'menuItem',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'mi'},
                         'panel': {'args': '<type __builtin__.bool>',
                                   'longname': 'panel',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'pnl'},
                         'panelConfig': {'args': '<type __builtin__.bool>',
                                         'longname': 'panelConfig',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'pc'},
                         'radioMenuItemCollection': {'args': '<type __builtin__.bool>',
                                                     'longname': 'radioMenuItemCollection',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'ric'},
                         'toolContext': {'args': '<type __builtin__.bool>',
                                         'longname': 'toolContext',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'tc'},
                         'uiTemplate': {'args': '<type __builtin__.bool>',
                                        'longname': 'uiTemplate',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'uit'},
                         'window': {'args': '<type __builtin__.bool>',
                                    'longname': 'window',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'wnd'}},
               'shortFlags': {'cl': 'collection',
                              'ctl': 'control',
                              'ed': 'editor',
                              'lay': 'layout',
                              'm': 'menu',
                              'mi': 'menuItem',
                              'pc': 'panelConfig',
                              'pnl': 'panel',
                              'ric': 'radioMenuItemCollection',
                              'tc': 'toolContext',
                              'uit': 'uiTemplate',
                              'wnd': 'window'},
               'type': 'language'},
  'deltaMush': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                    'longname': 'after',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'af'},
                          'afterReference': {'args': '<type __builtin__.bool>',
                                             'longname': 'afterReference',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'ar'},
                          'before': {'args': '<type __builtin__.bool>',
                                     'longname': 'before',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'bf'},
                          'deformerTools': {'args': '<type __builtin__.bool>',
                                            'longname': 'deformerTools',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'dt'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'envelope': {'args': '<type __builtin__.float>',
                                       'longname': 'envelope',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsUnpacking': True,
                                       'shortname': 'en'},
                          'exclusive': {'args': '<type __builtin__.str>',
                                        'longname': 'exclusive',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'ex'},
                          'frontOfChain': {'args': '<type __builtin__.bool>',
                                           'longname': 'frontOfChain',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'foc'},
                          'geometry': {'args': '<type __builtin__.str>',
                                       'longname': 'geometry',
                                       'modes': ['query',
                                                 'edit',
                                                 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'g'},
                          'geometryIndices': {'args': '<type __builtin__.bool>',
                                              'longname': 'geometryIndices',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'gi'},
                          'ignoreSelected': {'args': '<type __builtin__.bool>',
                                             'longname': 'ignoreSelected',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ignoreSelected'},
                          'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                      'longname': 'includeHiddenSelections',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'ihs'},
                          'inwardConstraint': {'args': '<type __builtin__.float>',
                                               'longname': 'inwardConstraint',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'resultNeedsUnpacking': True,
                                               'shortname': 'iwc'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'outwardConstraint': {'args': '<type __builtin__.float>',
                                                'longname': 'outwardConstraint',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'resultNeedsUnpacking': True,
                                                'shortname': 'owc'},
                          'parallel': {'args': '<type __builtin__.bool>',
                                       'longname': 'parallel',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'par'},
                          'pinBorderVertices': {'args': '<type __builtin__.bool>',
                                                'longname': 'pinBorderVertices',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'resultNeedsUnpacking': True,
                                                'shortname': 'pbv'},
                          'prune': {'args': '<type __builtin__.bool>',
                                    'longname': 'prune',
                                    'modes': ['edit'],
                                    'numArgs': 0,
                                    'shortname': 'pr'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'remove': {'args': '<type __builtin__.bool>',
                                     'longname': 'remove',
                                     'modes': ['edit', 'multiuse'],
                                     'numArgs': 0,
                                     'shortname': 'rm'},
                          'smoothingIterations': {'args': '<type __builtin__.int>',
                                                  'longname': 'smoothingIterations',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'si'},
                          'smoothingStep': {'args': '<type __builtin__.float>',
                                            'longname': 'smoothingStep',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'resultNeedsUnpacking': True,
                                            'shortname': 'ss'},
                          'split': {'args': '<type __builtin__.bool>',
                                    'longname': 'split',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'sp'}},
                'removedFlags': {'is': 'ignoreSelected'},
                'resultNeedsUnpacking': True,
                'shortFlags': {'af': 'after',
                               'ar': 'afterReference',
                               'bf': 'before',
                               'dt': 'deformerTools',
                               'e': 'edit',
                               'en': 'envelope',
                               'ex': 'exclusive',
                               'foc': 'frontOfChain',
                               'g': 'geometry',
                               'gi': 'geometryIndices',
                               'ignoreSelected': 'ignoreSelected',
                               'ihs': 'includeHiddenSelections',
                               'iwc': 'inwardConstraint',
                               'n': 'name',
                               'owc': 'outwardConstraint',
                               'par': 'parallel',
                               'pbv': 'pinBorderVertices',
                               'pr': 'prune',
                               'q': 'query',
                               'rm': 'remove',
                               'si': 'smoothingIterations',
                               'sp': 'split',
                               'ss': 'smoothingStep'},
                'type': 'animation'},
  'detachCurve': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'curveOnSurface': {'args': '<type __builtin__.bool>',
                                               'longname': 'curveOnSurface',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'cos'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'keep': {'args': '<type __builtin__.bool>',
                                     'longname': 'keep',
                                     'modes': ['create',
                                               'query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'k'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'parameter': {'args': '<type __builtin__.float>',
                                          'longname': 'parameter',
                                          'modes': ['create',
                                                    'query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                'longname': 'replaceOriginal',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'rpo'}},
                  'shortFlags': {'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'cos': 'curveOnSurface',
                                 'e': 'edit',
                                 'fzn': 'frozen',
                                 'k': 'keep',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'o': 'object',
                                 'p': 'parameter',
                                 'q': 'query',
                                 'rpo': 'replaceOriginal'},
                  'type': 'modeling'},
  'detachDeviceAttr': {'flags': {'all': {'args': '<type __builtin__.bool>',
                                         'longname': 'all',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'all'},
                                 'attribute': {'args': '<type __builtin__.str>',
                                               'longname': 'attribute',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'at'},
                                 'axis': {'args': '<type __builtin__.str>',
                                          'longname': 'axis',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ax'},
                                 'device': {'args': '<type __builtin__.str>',
                                            'longname': 'device',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'd'},
                                 'selection': {'args': '<type __builtin__.bool>',
                                               'longname': 'selection',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'sl'}},
                       'shortFlags': {},
                       'type': 'system'},
  'detachSurface': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cch'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'direction': {'args': '<type __builtin__.int>',
                                            'longname': 'direction',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'd'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'keep': {'args': '<type __builtin__.bool>',
                                       'longname': 'keep',
                                       'modes': ['create',
                                                 'query',
                                                 'edit',
                                                 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'k'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nds'},
                              'object': {'args': '<type __builtin__.bool>',
                                         'longname': 'object',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'o'},
                              'parameter': {'args': '<type __builtin__.float>',
                                            'longname': 'parameter',
                                            'modes': ['create',
                                                      'query',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'p'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                  'longname': 'replaceOriginal',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'rpo'}},
                    'shortFlags': {'cch': 'caching',
                                   'ch': 'constructionHistory',
                                   'd': 'direction',
                                   'e': 'edit',
                                   'fzn': 'frozen',
                                   'k': 'keep',
                                   'n': 'name',
                                   'nds': 'nodeState',
                                   'o': 'object',
                                   'p': 'parameter',
                                   'q': 'query',
                                   'rpo': 'replaceOriginal'},
                    'type': 'modeling'},
  'deviceEditor': {'flags': {'control': {'args': '<type __builtin__.bool>',
                                         'longname': 'control',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'ctl'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ex'},
                             'filter': {'args': '<type __builtin__.str>',
                                        'longname': 'filter',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'f'},
                             'forceMainConnection': {'args': '<type __builtin__.str>',
                                                     'longname': 'forceMainConnection',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'fmc'},
                             'highlightConnection': {'args': '<type __builtin__.str>',
                                                     'longname': 'highlightConnection',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'hlc'},
                             'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'lockMainConnection',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'lck'},
                             'mainListConnection': {'args': '<type __builtin__.str>',
                                                    'longname': 'mainListConnection',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'mlc'},
                             'panel': {'args': '<type __builtin__.str>',
                                       'longname': 'panel',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'pnl'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'selectionConnection': {'args': '<type __builtin__.str>',
                                                     'longname': 'selectionConnection',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'slc'},
                             'stateString': {'args': '<type __builtin__.bool>',
                                             'longname': 'stateString',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'sts'},
                             'takePath': {'args': '<type __builtin__.str>',
                                          'longname': 'takePath',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'tp'},
                             'unParent': {'args': '<type __builtin__.bool>',
                                          'longname': 'unParent',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'up'},
                             'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                      'longname': 'unlockMainConnection',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ulk'},
                             'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                      'longname': 'updateMainConnection',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'upd'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'}},
                   'shortFlags': {},
                   'type': 'system'},
  'deviceManager': {'flags': {'attachment': {'args': '<type __builtin__.bool>',
                                             'longname': 'attachment',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'att'},
                              'axisCoordChanges': {'args': '<type __builtin__.bool>',
                                                   'longname': 'axisCoordChanges',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'acc'},
                              'axisIndex': {'args': '<type __builtin__.int>',
                                            'longname': 'axisIndex',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'axi'},
                              'axisName': {'args': '<type __builtin__.bool>',
                                           'longname': 'axisName',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 'axn'},
                              'axisOffset': {'args': '<type __builtin__.bool>',
                                             'longname': 'axisOffset',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'axo'},
                              'axisScale': {'args': '<type __builtin__.bool>',
                                            'longname': 'axisScale',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'axs'},
                              'deviceIndex': {'args': '<type __builtin__.int>',
                                              'longname': 'deviceIndex',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dvi'},
                              'deviceNameFromIndex': {'args': '<type __builtin__.int>',
                                                      'longname': 'deviceNameFromIndex',
                                                      'modes': ['query'],
                                                      'numArgs': 1,
                                                      'shortname': 'dni'},
                              'numAxis': {'args': '<type __builtin__.bool>',
                                          'longname': 'numAxis',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'nax'},
                              'numDevices': {'args': '<type __builtin__.bool>',
                                             'longname': 'numDevices',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'ndv'}},
                    'shortFlags': {},
                    'type': 'animation'},
  'devicePanel': {'callbackFlags': ['popupMenuProcedure', 'pmp'],
                  'flags': {'control': {'args': '<type __builtin__.bool>',
                                        'longname': 'control',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'ctl'},
                            'copy': {'args': '<type __builtin__.str>',
                                     'longname': 'copy',
                                     'modes': ['edit'],
                                     'numArgs': 1,
                                     'shortname': 'cp'},
                            'createString': {'args': '<type __builtin__.bool>',
                                             'longname': 'createString',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'cs'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'editString': {'args': '<type __builtin__.bool>',
                                           'longname': 'editString',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'es'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ex'},
                            'init': {'args': '<type __builtin__.bool>',
                                     'longname': 'init',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'init'},
                            'isUnique': {'args': '<type __builtin__.bool>',
                                         'longname': 'isUnique',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'iu'},
                            'label': {'args': '<type __builtin__.str>',
                                      'longname': 'label',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'l'},
                            'menuBarRepeatLast': {'args': '<type __builtin__.bool>',
                                                  'longname': 'menuBarRepeatLast',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mrl'},
                            'menuBarVisible': {'args': '<type __builtin__.bool>',
                                               'longname': 'menuBarVisible',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'mbv'},
                            'needsInit': {'args': '<type __builtin__.bool>',
                                          'longname': 'needsInit',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ni'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuProcedure': {'args': 'script',
                                                   'longname': 'popupMenuProcedure',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'pmp'},
                            'replacePanel': {'args': '<type __builtin__.str>',
                                             'longname': 'replacePanel',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'rp'},
                            'tearOff': {'args': '<type __builtin__.bool>',
                                        'longname': 'tearOff',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'to'},
                            'tearOffCopy': {'args': '<type __builtin__.str>',
                                            'longname': 'tearOffCopy',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'toc'},
                            'tearOffRestore': {'args': '<type __builtin__.bool>',
                                               'longname': 'tearOffRestore',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tor'},
                            'unParent': {'args': '<type __builtin__.bool>',
                                         'longname': 'unParent',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 'up'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'}},
                  'shortFlags': {},
                  'type': 'system'},
  'dgControl': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'dgInfo': {'flags': {'allNodes': {'args': '<type __builtin__.bool>',
                                    'longname': 'allNodes',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'all'},
                       'connections': {'args': '<type __builtin__.bool>',
                                       'longname': 'connections',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'c'},
                       'dirty': {'args': '<type __builtin__.bool>',
                                 'longname': 'dirty',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'd'},
                       'nodes': {'args': '<type __builtin__.bool>',
                                 'longname': 'nodes',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'n'},
                       'nonDeletable': {'args': '<type __builtin__.bool>',
                                        'longname': 'nonDeletable',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'nd'},
                       'outputFile': {'args': '<type __builtin__.str>',
                                      'longname': 'outputFile',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'of'},
                       'propagation': {'args': '<type __builtin__.bool>',
                                       'longname': 'propagation',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                       'short': {'args': '<type __builtin__.bool>',
                                 'longname': 'short',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 's'},
                       'size': {'args': '<type __builtin__.bool>',
                                'longname': 'size',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'sz'},
                       'subgraph': {'args': '<type __builtin__.bool>',
                                    'longname': 'subgraph',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'sub'},
                       'type': {'args': '<type __builtin__.str>',
                                'longname': 'type',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'nt'}},
             'shortFlags': {'all': 'allNodes',
                            'c': 'connections',
                            'd': 'dirty',
                            'n': 'nodes',
                            'nd': 'nonDeletable',
                            'nt': 'type',
                            'of': 'outputFile',
                            'p': 'propagation',
                            's': 'short',
                            'sub': 'subgraph',
                            'sz': 'size'},
             'type': 'system'},
  'dgPerformance': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'dgValidateCurve': {'flags': {'allCurves': {'args': '<type __builtin__.bool>',
                                              'longname': 'allCurves',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'a'},
                                'verbose': {'args': '<type __builtin__.bool>',
                                            'longname': 'verbose',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'v'}},
                      'shortFlags': {'a': 'allCurves', 'v': 'verbose'},
                      'type': 'system'},
  'dgcontrol': {'flags': {'iomode': {'args': '<type __builtin__.bool>',
                                     'longname': 'iomode',
                                     'numArgs': 1,
                                     'shortname': 'iom'}},
                'shortFlags': {'iom': 'iomode'},
                'type': 'other'},
  'dgdebug': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'dgdirty': {'flags': {'allPlugs': {'args': '<type __builtin__.bool>',
                                     'longname': 'allPlugs',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 'a'},
                        'clean': {'args': '<type __builtin__.bool>',
                                  'longname': 'clean',
                                  'modes': ['create', 'query'],
                                  'numArgs': 0,
                                  'shortname': 'c'},
                        'implicit': {'args': '<type __builtin__.bool>',
                                     'longname': 'implicit',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 'i'},
                        'list': {'args': '<type __builtin__.str>',
                                 'longname': 'list',
                                 'modes': ['create', 'query'],
                                 'numArgs': 1,
                                 'shortname': 'l'},
                        'propagation': {'args': '<type __builtin__.bool>',
                                        'longname': 'propagation',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'p'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'showTiming': {'args': '<type __builtin__.bool>',
                                       'longname': 'showTiming',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'st'},
                        'verbose': {'args': '<type __builtin__.bool>',
                                    'longname': 'verbose',
                                    'modes': ['create', 'query'],
                                    'numArgs': 0,
                                    'shortname': 'v'}},
              'shortFlags': {'a': 'allPlugs',
                             'c': 'clean',
                             'i': 'implicit',
                             'l': 'list',
                             'p': 'propagation',
                             'q': 'query',
                             'st': 'showTiming',
                             'v': 'verbose'},
              'type': 'system'},
  'dgeval': {'flags': {'src': {'args': '<type __builtin__.bool>',
                               'longname': 'src',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'src'},
                       'verbose': {'args': '<type __builtin__.bool>',
                                   'longname': 'verbose',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'v'}},
             'shortFlags': {'src': 'src', 'v': 'verbose'},
             'type': 'system'},
  'dgfilter': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                       'longname': 'attribute',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'atr'},
                         'list': {'args': '<type __builtin__.bool>',
                                  'longname': 'list',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'l'},
                         'logicalAnd': {'args': ['<type __builtin__.str>',
                                                 '<type __builtin__.str>'],
                                        'longname': 'logicalAnd',
                                        'modes': ['create'],
                                        'numArgs': 2,
                                        'shortname': 'logicalAnd'},
                         'logicalNot': {'args': '<type __builtin__.str>',
                                        'longname': 'logicalNot',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'logicalNot'},
                         'logicalOr': {'args': ['<type __builtin__.str>',
                                                '<type __builtin__.str>'],
                                       'longname': 'logicalOr',
                                       'modes': ['create'],
                                       'numArgs': 2,
                                       'shortname': 'logicalOr'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'node': {'args': '<type __builtin__.str>',
                                  'longname': 'node',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'nd'},
                         'nodeType': {'args': '<type __builtin__.str>',
                                      'longname': 'nodeType',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'nt'},
                         'plug': {'args': '<type __builtin__.str>',
                                  'longname': 'plug',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'p'}},
               'shortFlags': {},
               'type': 'system'},
  'dgmodified': {'flags': {}, 'shortFlags': {}, 'type': 'system'},
  'dgstats': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'dgtimer': {'flags': {'combineType': {'args': '<type __builtin__.bool>',
                                        'longname': 'combineType',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'ct'},
                        'hide': {'args': '<type __builtin__.str>',
                                 'longname': 'hide',
                                 'modes': ['create', 'query', 'multiuse'],
                                 'numArgs': 1,
                                 'shortname': 'hi'},
                        'hierarchy': {'args': '<type __builtin__.bool>',
                                      'longname': 'hierarchy',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'h'},
                        'maxDisplay': {'args': '<type __builtin__.int>',
                                       'longname': 'maxDisplay',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['create', 'query'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'noHeader': {'args': '<type __builtin__.bool>',
                                     'longname': 'noHeader',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 'nh'},
                        'outputFile': {'args': '<type __builtin__.str>',
                                       'longname': 'outputFile',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                        'overhead': {'args': '<type __builtin__.bool>',
                                     'longname': 'overhead',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'oh'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'rangeLower': {'args': '<type __builtin__.float>',
                                       'longname': 'rangeLower',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'rgl'},
                        'rangeUpper': {'args': '<type __builtin__.float>',
                                       'longname': 'rangeUpper',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'rgu'},
                        'reset': {'args': '<type __builtin__.bool>',
                                  'longname': 'reset',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'r'},
                        'returnCode': {'args': '<type __builtin__.str>',
                                       'longname': 'returnCode',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'rc'},
                        'returnType': {'args': '<type __builtin__.str>',
                                       'longname': 'returnType',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'rt'},
                        'show': {'args': '<type __builtin__.str>',
                                 'longname': 'show',
                                 'modes': ['create', 'query', 'multiuse'],
                                 'numArgs': 1,
                                 'shortname': 'sh'},
                        'sortMetric': {'args': '<type __builtin__.str>',
                                       'longname': 'sortMetric',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'sm'},
                        'sortType': {'args': '<type __builtin__.str>',
                                     'longname': 'sortType',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'st'},
                        'threshold': {'args': '<type __builtin__.float>',
                                      'longname': 'threshold',
                                      'modes': ['query'],
                                      'numArgs': 1,
                                      'shortname': 'th'},
                        'timerOff': {'args': '<type __builtin__.bool>',
                                     'longname': 'timerOff',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 'off'},
                        'timerOn': {'args': '<type __builtin__.bool>',
                                    'longname': 'timerOn',
                                    'modes': ['create', 'query'],
                                    'numArgs': 0,
                                    'shortname': 'on'},
                        'trace': {'args': '<type __builtin__.bool>',
                                  'longname': 'trace',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'tr'},
                        'type': {'args': '<type __builtin__.str>',
                                 'longname': 'type',
                                 'modes': ['create', 'query'],
                                 'numArgs': 1,
                                 'shortname': 't'},
                        'uniqueName': {'args': '<type __builtin__.bool>',
                                       'longname': 'uniqueName',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'un'},
                        'updateHeatMap': {'args': '<type __builtin__.int>',
                                          'longname': 'updateHeatMap',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'uhm'}},
              'shortFlags': {'ct': 'combineType',
                             'h': 'hierarchy',
                             'hi': 'hide',
                             'm': 'maxDisplay',
                             'n': 'name',
                             'nh': 'noHeader',
                             'o': 'outputFile',
                             'off': 'timerOff',
                             'oh': 'overhead',
                             'on': 'timerOn',
                             'q': 'query',
                             'r': 'reset',
                             'rc': 'returnCode',
                             'rgl': 'rangeLower',
                             'rgu': 'rangeUpper',
                             'rt': 'returnType',
                             'sh': 'show',
                             'sm': 'sortMetric',
                             'st': 'sortType',
                             't': 'type',
                             'th': 'threshold',
                             'tr': 'trace',
                             'uhm': 'updateHeatMap',
                             'un': 'uniqueName'},
              'type': 'system'},
  'dimWhen': {'flags': {'clear': {'args': '<type __builtin__.bool>',
                                  'longname': 'clear',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'c'},
                        'false': {'args': '<type __builtin__.bool>',
                                  'longname': 'false',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'f'},
                        'true': {'args': '<type __builtin__.bool>',
                                 'longname': 'true',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 't'}},
              'shortFlags': {'c': 'clear', 'f': 'false', 't': 'true'},
              'type': 'language'},
  'directConnectPath': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'directKeyCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'history': {'args': '<type __builtin__.bool>',
                                         'longname': 'history',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ch'},
                             'image1': {'args': '<type __builtin__.str>',
                                        'longname': 'image1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i1'},
                             'image2': {'args': '<type __builtin__.str>',
                                        'longname': 'image2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i2'},
                             'image3': {'args': '<type __builtin__.str>',
                                        'longname': 'image3',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i3'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'option': {'args': '<type __builtin__.str>',
                                        'longname': 'option',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'selectedOnly': {'args': '<type __builtin__.bool>',
                                              'longname': 'selectedOnly',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'so'}},
                   'shortFlags': {'ch': 'history',
                                  'e': 'edit',
                                  'ex': 'exists',
                                  'i1': 'image1',
                                  'i2': 'image2',
                                  'i3': 'image3',
                                  'n': 'name',
                                  'o': 'option',
                                  'q': 'query',
                                  'so': 'selectedOnly'},
                   'type': 'context'},
  'directionalLight': {'flags': {'decayRate': {'args': '<type __builtin__.int>',
                                               'longname': 'decayRate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'd'},
                                 'discRadius': {'args': '<type __builtin__.float>',
                                                'longname': 'discRadius',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'drs'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'exclusive': {'args': '<type __builtin__.bool>',
                                               'longname': 'exclusive',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'exc'},
                                 'intensity': {'args': '<type __builtin__.float>',
                                               'longname': 'intensity',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'i'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'position': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'position',
                                              'modes': ['create', 'query'],
                                              'numArgs': 3,
                                              'shortname': 'pos'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'rgb': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'rgb',
                                         'modes': ['create', 'query'],
                                         'numArgs': 3,
                                         'shortname': 'rgb'},
                                 'rotation': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'rotation',
                                              'modes': ['create', 'query'],
                                              'numArgs': 3,
                                              'shortname': 'rot'},
                                 'shadowColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'shadowColor',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 3,
                                                 'shortname': 'sc'},
                                 'shadowDither': {'args': '<type __builtin__.float>',
                                                  'longname': 'shadowDither',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'sd'},
                                 'shadowSamples': {'args': '<type __builtin__.int>',
                                                   'longname': 'shadowSamples',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'sh'},
                                 'softShadow': {'args': '<type __builtin__.bool>',
                                                'longname': 'softShadow',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'ss'},
                                 'useRayTraceShadows': {'args': '<type __builtin__.bool>',
                                                        'longname': 'useRayTraceShadows',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'rs'}},
                       'shortFlags': {'d': 'decayRate',
                                      'drs': 'discRadius',
                                      'e': 'edit',
                                      'exc': 'exclusive',
                                      'i': 'intensity',
                                      'n': 'name',
                                      'pos': 'position',
                                      'q': 'query',
                                      'rgb': 'rgb',
                                      'rot': 'rotation',
                                      'rs': 'useRayTraceShadows',
                                      'sc': 'shadowColor',
                                      'sd': 'shadowDither',
                                      'sh': 'shadowSamples',
                                      'ss': 'softShadow'},
                       'type': 'rendering'},
  'dirmap': {'flags': {'convertDirectory': {'args': '<type __builtin__.str>',
                                            'longname': 'convertDirectory',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'cd'},
                       'enable': {'args': '<type __builtin__.bool>',
                                  'longname': 'enable',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'en'},
                       'getAllMappings': {'args': '<type __builtin__.bool>',
                                          'longname': 'getAllMappings',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'gam'},
                       'getMappedDirectory': {'args': '<type __builtin__.str>',
                                              'longname': 'getMappedDirectory',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'gmd'},
                       'mapDirectory': {'args': ['<type __builtin__.str>',
                                                 '<type __builtin__.str>'],
                                        'longname': 'mapDirectory',
                                        'modes': ['create'],
                                        'numArgs': 2,
                                        'shortname': 'm'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'unmapDirectory': {'args': '<type __builtin__.str>',
                                          'longname': 'unmapDirectory',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'um'}},
             'shortFlags': {'cd': 'convertDirectory',
                            'en': 'enable',
                            'gam': 'getAllMappings',
                            'gmd': 'getMappedDirectory',
                            'm': 'mapDirectory',
                            'q': 'query',
                            'um': 'unmapDirectory'},
             'type': 'system'},
  'disable': {'flags': {'value': {'args': '<type __builtin__.bool>',
                                  'longname': 'value',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'v'}},
              'shortFlags': {'v': 'value'},
              'type': 'language'},
  'disableIncorrectNameWarning': {'flags': {},
                                  'shortFlags': {},
                                  'type': 'windows'},
  'disconnectAttr': {'flags': {'nextAvailable': {'args': '<type __builtin__.bool>',
                                                 'longname': 'nextAvailable',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'na'}},
                     'shortFlags': {'na': 'nextAvailable'},
                     'type': 'general'},
  'disconnectJoint': {'flags': {'attachHandleMode': {'args': '<type __builtin__.bool>',
                                                     'longname': 'attachHandleMode',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'ahm'},
                                'deleteHandleMode': {'args': '<type __builtin__.bool>',
                                                     'longname': 'deleteHandleMode',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'dhm'}},
                      'shortFlags': {'ahm': 'attachHandleMode',
                                     'dhm': 'deleteHandleMode'},
                      'type': 'animation'},
  'diskCache': {'flags': {'append': {'args': '<type __builtin__.bool>',
                                     'longname': 'append',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 'a'},
                          'cacheType': {'args': '<type __builtin__.str>',
                                        'longname': 'cacheType',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'ct'},
                          'close': {'args': '<type __builtin__.str>',
                                    'longname': 'close',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'c'},
                          'closeAll': {'args': '<type __builtin__.bool>',
                                       'longname': 'closeAll',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'ca'},
                          'delete': {'args': '<type __builtin__.str>',
                                     'longname': 'delete',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'd'},
                          'deleteAll': {'args': '<type __builtin__.bool>',
                                        'longname': 'deleteAll',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'da'},
                          'empty': {'args': '<type __builtin__.str>',
                                    'longname': 'empty',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'e'},
                          'emptyAll': {'args': '<type __builtin__.bool>',
                                       'longname': 'emptyAll',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'ea'},
                          'enabledCachesOnly': {'args': '<type __builtin__.bool>',
                                                'longname': 'enabledCachesOnly',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 0,
                                                'shortname': 'eco'},
                          'endTime': {'args': 'time',
                                      'longname': 'endTime',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'et'},
                          'frameRangeType': {'args': '<type __builtin__.str>',
                                             'longname': 'frameRangeType',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'frt'},
                          'overSample': {'args': '<type __builtin__.bool>',
                                         'longname': 'overSample',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'os'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'samplingRate': {'args': '<type __builtin__.int>',
                                           'longname': 'samplingRate',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'sr'},
                          'startTime': {'args': 'time',
                                        'longname': 'startTime',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'st'},
                          'tempDir': {'args': '<type __builtin__.bool>',
                                      'longname': 'tempDir',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'tmp'}},
                'shortFlags': {'a': 'append',
                               'c': 'close',
                               'ca': 'closeAll',
                               'ct': 'cacheType',
                               'd': 'delete',
                               'da': 'deleteAll',
                               'e': 'empty',
                               'ea': 'emptyAll',
                               'eco': 'enabledCachesOnly',
                               'et': 'endTime',
                               'frt': 'frameRangeType',
                               'os': 'overSample',
                               'q': 'query',
                               'sr': 'samplingRate',
                               'st': 'startTime',
                               'tmp': 'tempDir'},
                'type': 'system'},
  'dispatchGenericCommand': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'displacementToPoly': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'findBboxOnly': {'args': '<type __builtin__.bool>',
                                                    'longname': 'findBboxOnly',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'fbb'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'}},
                         'shortFlags': {'e': 'edit',
                                        'fbb': 'findBboxOnly',
                                        'q': 'query'},
                         'type': 'rendering'},
  'displayAffected': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'}},
                      'shortFlags': {'q': 'query'},
                      'type': 'general'},
  'displayColor': {'flags': {'active': {'args': '<type __builtin__.bool>',
                                        'longname': 'active',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'a'},
                             'create': {'args': '<type __builtin__.bool>',
                                        'longname': 'create',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'c'},
                             'dormant': {'args': '<type __builtin__.bool>',
                                         'longname': 'dormant',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'd'},
                             'list': {'args': '<type __builtin__.bool>',
                                      'longname': 'list',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'l'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'queryIndex': {'args': '<type __builtin__.int>',
                                            'longname': 'queryIndex',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'qi'},
                             'resetToFactory': {'args': '<type __builtin__.bool>',
                                                'longname': 'resetToFactory',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'rf'},
                             'resetToSaved': {'args': '<type __builtin__.bool>',
                                              'longname': 'resetToSaved',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'rs'}},
                   'shortFlags': {'a': 'active',
                                  'c': 'create',
                                  'd': 'dormant',
                                  'l': 'list',
                                  'q': 'query',
                                  'qi': 'queryIndex',
                                  'rf': 'resetToFactory',
                                  'rs': 'resetToSaved'},
                   'type': 'general'},
  'displayCull': {'flags': {'backFaceCulling': {'args': '<type __builtin__.bool>',
                                                'longname': 'backFaceCulling',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'bfc'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'}},
                  'shortFlags': {'bfc': 'backFaceCulling', 'q': 'query'},
                  'type': 'general'},
  'displayLevelOfDetail': {'flags': {'levelOfDetail': {'args': '<type __builtin__.bool>',
                                                       'longname': 'levelOfDetail',
                                                       'modes': [],
                                                       'numArgs': 1,
                                                       'shortname': 'lod'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'}},
                           'shortFlags': {'lod': 'levelOfDetail',
                                          'q': 'query'},
                           'type': 'general'},
  'displayPref': {'flags': {'activeObjectPivots': {'args': '<type __builtin__.bool>',
                                                   'longname': 'activeObjectPivots',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'aop'},
                            'defaultFontSize': {'args': '<type __builtin__.int>',
                                                'longname': 'defaultFontSize',
                                                'numArgs': 1,
                                                'shortname': 'dfs'},
                            'displayAffected': {'args': '<type __builtin__.bool>',
                                                'longname': 'displayAffected',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'da'},
                            'displayGradient': {'args': '<type __builtin__.bool>',
                                                'longname': 'displayGradient',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'dgr'},
                            'fontSettingMode': {'args': '<type __builtin__.int>',
                                                'longname': 'fontSettingMode',
                                                'numArgs': 1,
                                                'shortname': 'fm'},
                            'ghostFrames': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'ghostFrames',
                                            'modes': ['create', 'query'],
                                            'numArgs': 3,
                                            'shortname': 'gf'},
                            'lineWidth': {'args': '<type __builtin__.float>',
                                          'longname': 'lineWidth',
                                          'numArgs': 1,
                                          'shortname': 'lw'},
                            'materialLoadingMode': {'args': '<type __builtin__.str>',
                                                    'longname': 'materialLoadingMode',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'mld'},
                            'maxHardwareTextureResolution': {'args': '<type __builtin__.bool>',
                                                             'longname': 'maxHardwareTextureResolution',
                                                             'modes': ['query'],
                                                             'numArgs': 0,
                                                             'shortname': 'mhr'},
                            'maxTextureResolution': {'args': '<type __builtin__.int>',
                                                     'longname': 'maxTextureResolution',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'mtr'},
                            'purgeExistingTextures': {'args': '<type __builtin__.bool>',
                                                      'longname': 'purgeExistingTextures',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'pet'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'regionOfEffect': {'args': '<type __builtin__.bool>',
                                               'longname': 'regionOfEffect',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'roe'},
                            'shadeTemplates': {'args': '<type __builtin__.bool>',
                                               'longname': 'shadeTemplates',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'st'},
                            'smallFontSize': {'args': '<type __builtin__.int>',
                                              'longname': 'smallFontSize',
                                              'numArgs': 1,
                                              'shortname': 'sfs'},
                            'textureDrawPixel': {'args': '<type __builtin__.bool>',
                                                 'longname': 'textureDrawPixel',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'tdp'},
                            'wireframeOnShadedActive': {'args': '<type __builtin__.str>',
                                                        'longname': 'wireframeOnShadedActive',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'wsa'}},
                  'shortFlags': {'aop': 'activeObjectPivots',
                                 'da': 'displayAffected',
                                 'dfs': 'defaultFontSize',
                                 'dgr': 'displayGradient',
                                 'fm': 'fontSettingMode',
                                 'gf': 'ghostFrames',
                                 'lw': 'lineWidth',
                                 'mhr': 'maxHardwareTextureResolution',
                                 'mlm': 'materialLoadingMode',
                                 'mtr': 'maxTextureResolution',
                                 'pet': 'purgeExistingTextures',
                                 'q': 'query',
                                 'roe': 'regionOfEffect',
                                 'sfs': 'smallFontSize',
                                 'st': 'shadeTemplates',
                                 'tdp': 'textureDrawPixel',
                                 'wsa': 'wireframeOnShadedActive'},
                  'type': 'general'},
  'displayRGBColor': {'flags': {'create': {'args': '<type __builtin__.bool>',
                                           'longname': 'create',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'c'},
                                'hueSaturationValue': {'args': '<type __builtin__.bool>',
                                                       'longname': 'hueSaturationValue',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 0,
                                                       'shortname': 'hsv'},
                                'list': {'args': '<type __builtin__.bool>',
                                         'longname': 'list',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'l'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'resetToFactory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'resetToFactory',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'rf'},
                                'resetToSaved': {'args': '<type __builtin__.bool>',
                                                 'longname': 'resetToSaved',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'rs'}},
                      'shortFlags': {'c': 'create',
                                     'hsv': 'hueSaturationValue',
                                     'l': 'list',
                                     'q': 'query',
                                     'rf': 'resetToFactory',
                                     'rs': 'resetToSaved'},
                      'type': 'general'},
  'displaySmoothness': {'flags': {'all': {'args': '<type __builtin__.bool>',
                                          'longname': 'all',
                                          'modes': ['create', 'query'],
                                          'numArgs': 0,
                                          'shortname': 'all'},
                                  'boundary': {'args': '<type __builtin__.bool>',
                                               'longname': 'boundary',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 0,
                                               'shortname': 'bn'},
                                  'defaultCreation': {'args': '<type __builtin__.bool>',
                                                      'longname': 'defaultCreation',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 0,
                                                      'shortname': 'dc'},
                                  'divisionsU': {'args': '<type __builtin__.int>',
                                                 'longname': 'divisionsU',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'du'},
                                  'divisionsV': {'args': '<type __builtin__.int>',
                                                 'longname': 'divisionsV',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'dv'},
                                  'full': {'args': '<type __builtin__.bool>',
                                           'longname': 'full',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'f'},
                                  'hull': {'args': '<type __builtin__.bool>',
                                           'longname': 'hull',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'hl'},
                                  'pointsShaded': {'args': '<type __builtin__.int>',
                                                   'longname': 'pointsShaded',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ps'},
                                  'pointsWire': {'args': '<type __builtin__.int>',
                                                 'longname': 'pointsWire',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'pw'},
                                  'polygonObject': {'args': '<type __builtin__.int>',
                                                    'longname': 'polygonObject',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'po'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'renderTessellation': {'args': '<type __builtin__.bool>',
                                                         'longname': 'renderTessellation',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'rt'},
                                  'simplifyU': {'args': '<type __builtin__.int>',
                                                'longname': 'simplifyU',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'su'},
                                  'simplifyV': {'args': '<type __builtin__.int>',
                                                'longname': 'simplifyV',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'sv'}},
                        'shortFlags': {'all': 'all',
                                       'bn': 'boundary',
                                       'dc': 'defaultCreation',
                                       'du': 'divisionsU',
                                       'dv': 'divisionsV',
                                       'f': 'full',
                                       'hl': 'hull',
                                       'po': 'polygonObject',
                                       'ps': 'pointsShaded',
                                       'pw': 'pointsWire',
                                       'q': 'query',
                                       'rt': 'renderTessellation',
                                       'su': 'simplifyU',
                                       'sv': 'simplifyV'},
                        'type': 'general'},
  'displayStats': {'flags': {'frameRate': {'args': '<type __builtin__.bool>',
                                           'longname': 'frameRate',
                                           'numArgs': 1,
                                           'shortname': 'fr'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'fr': 'frameRate', 'q': 'query'},
                   'type': 'general'},
  'displayString': {'flags': {'delete': {'args': '<type __builtin__.bool>',
                                         'longname': 'delete',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'd'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'keys': {'args': '<type __builtin__.bool>',
                                       'longname': 'keys',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'k'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'replace': {'args': '<type __builtin__.bool>',
                                          'longname': 'replace',
                                          'modes': ['create', 'query'],
                                          'numArgs': 0,
                                          'shortname': 'r'},
                              'value': {'args': '<type __builtin__.str>',
                                        'longname': 'value',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'v'}},
                    'shortFlags': {'d': 'delete',
                                   'ex': 'exists',
                                   'k': 'keys',
                                   'q': 'query',
                                   'r': 'replace',
                                   'v': 'value'},
                    'type': 'system'},
  'displaySurface': {'flags': {'flipNormals': {'args': '<type __builtin__.bool>',
                                               'longname': 'flipNormals',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'flp'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'twoSidedLighting': {'args': '<type __builtin__.bool>',
                                                    'longname': 'twoSidedLighting',
                                                    'modes': ['query'],
                                                    'numArgs': 1,
                                                    'shortname': 'two'},
                               'xRay': {'args': '<type __builtin__.bool>',
                                        'longname': 'xRay',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'x'}},
                     'shortFlags': {'flp': 'flipNormals',
                                    'q': 'query',
                                    'two': 'twoSidedLighting',
                                    'x': 'xRay'},
                     'type': 'general'},
  'distanceDimContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'history': {'args': '<type __builtin__.bool>',
                                               'longname': 'history',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'}},
                         'shortFlags': {'ch': 'history',
                                        'e': 'edit',
                                        'ex': 'exists',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'n': 'name',
                                        'q': 'query'},
                         'type': 'context'},
  'distanceDimension': {'flags': {'endPoint': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'endPoint',
                                               'modes': ['create'],
                                               'numArgs': 3,
                                               'shortname': 'ep'},
                                  'startPoint': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'startPoint',
                                                 'modes': ['create'],
                                                 'numArgs': 3,
                                                 'shortname': 'sp'}},
                        'shortFlags': {'ep': 'endPoint', 'sp': 'startPoint'},
                        'type': 'general'},
  'doBlur': {'flags': {'colorFile': {'args': '<type __builtin__.str>',
                                     'longname': 'colorFile',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'c'},
                       'length': {'args': '<type __builtin__.float>',
                                  'longname': 'length',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'l'},
                       'memCapSize': {'args': '<type __builtin__.float>',
                                      'longname': 'memCapSize',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                       'sharpness': {'args': '<type __builtin__.float>',
                                     'longname': 'sharpness',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 's'},
                       'smooth': {'args': '<type __builtin__.float>',
                                  'longname': 'smooth',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'm'},
                       'smoothColor': {'args': '<type __builtin__.bool>',
                                       'longname': 'smoothColor',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'r'},
                       'vectorFile': {'args': '<type __builtin__.str>',
                                      'longname': 'vectorFile',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'v'}},
             'shortFlags': {'c': 'colorFile',
                            'l': 'length',
                            'm': 'smooth',
                            'o': 'memCapSize',
                            'r': 'smoothColor',
                            's': 'sharpness',
                            'v': 'vectorFile'},
             'type': 'rendering'},
  'dockControl': {'callbackFlags': ['visibleChangeCommand',
                                    'vcc',
                                    'dragCallback',
                                    'dgc',
                                    'closeCommand',
                                    'cc',
                                    'floatChangeCommand',
                                    'fcc',
                                    'dropCallback',
                                    'dpc'],
                  'flags': {'allowedArea': {'args': '<type __builtin__.str>',
                                            'longname': 'allowedArea',
                                            'modes': ['create',
                                                      'query',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'aa'},
                            'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'area': {'args': '<type __builtin__.str>',
                                     'longname': 'area',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'a'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'closeCommand': {'args': 'script',
                                             'longname': 'closeCommand',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cc'},
                            'content': {'args': '<type __builtin__.str>',
                                        'longname': 'content',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'con'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dockStation': {'args': '<type __builtin__.str>',
                                            'longname': 'dockStation',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ds'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'enablePopupOption': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enablePopupOption',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'epo'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'fixedHeight': {'args': '<type __builtin__.bool>',
                                            'longname': 'fixedHeight',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'fh'},
                            'fixedWidth': {'args': '<type __builtin__.bool>',
                                           'longname': 'fixedWidth',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'fw'},
                            'floatChangeCommand': {'args': 'script',
                                                   'longname': 'floatChangeCommand',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'fcc'},
                            'floating': {'args': '<type __builtin__.bool>',
                                         'longname': 'floating',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'fl'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'label': {'args': '<type __builtin__.str>',
                                      'longname': 'label',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'l'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'moveable': {'args': '<type __builtin__.bool>',
                                         'longname': 'moveable',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'mov'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'r': {'args': '<type __builtin__.bool>',
                                  'longname': 'r',
                                  'modes': ['query', 'edit'],
                                  'numArgs': 0,
                                  'shortname': 'r'},
                            'retain': {'args': '<type __builtin__.bool>',
                                       'longname': 'retain',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ret'},
                            'sizeable': {'args': '<type __builtin__.bool>',
                                         'longname': 'sizeable',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 's'},
                            'splitLayout': {'args': '<type __builtin__.str>',
                                            'longname': 'splitLayout',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'sl'},
                            'state': {'args': '<type __builtin__.str>',
                                      'longname': 'state',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'st'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'removedFlags': {'raise': 'r'},
                  'shortFlags': {'a': 'area',
                                 'aa': 'allowedArea',
                                 'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'cc': 'closeCommand',
                                 'con': 'content',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'ds': 'dockStation',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'epo': 'enablePopupOption',
                                 'ex': 'exists',
                                 'fcc': 'floatChangeCommand',
                                 'fh': 'fixedHeight',
                                 'fl': 'floating',
                                 'fpn': 'fullPathName',
                                 'fw': 'fixedWidth',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'io': 'isObscured',
                                 'l': 'label',
                                 'm': 'manage',
                                 'mov': 'moveable',
                                 'nbg': 'noBackground',
                                 'npm': 'numberOfPopupMenus',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'r': 'r',
                                 'ret': 'retain',
                                 's': 'sizeable',
                                 'sbm': 'statusBarMessage',
                                 'sl': 'splitLayout',
                                 'st': 'state',
                                 'ut': 'useTemplate',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'dolly': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                   'longname': 'absolute',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'abs'},
                      'distance': {'args': '<type __builtin__.float>',
                                   'longname': 'distance',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'd'},
                      'dollyTowardsCenter': {'args': '<type __builtin__.bool>',
                                             'longname': 'dollyTowardsCenter',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dtc'},
                      'orthoScale': {'args': '<type __builtin__.float>',
                                     'longname': 'orthoScale',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'os'},
                      'relative': {'args': '<type __builtin__.bool>',
                                   'longname': 'relative',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'rel'}},
            'shortFlags': {'abs': 'absolute',
                           'd': 'distance',
                           'dtc': 'dollyTowardsCenter',
                           'os': 'orthoScale',
                           'rel': 'relative'},
            'type': 'rendering'},
  'dollyCtx': {'flags': {'alternateContext': {'args': '<type __builtin__.bool>',
                                              'longname': 'alternateContext',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'ac'},
                         'boxDollyType': {'args': '<type __builtin__.str>',
                                          'longname': 'boxDollyType',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'bdt'},
                         'centerOfInterestDolly': {'args': '<type __builtin__.bool>',
                                                   'longname': 'centerOfInterestDolly',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cd'},
                         'dollyTowardsCenter': {'args': '<type __builtin__.bool>',
                                                'longname': 'dollyTowardsCenter',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dtc'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'history': {'args': '<type __builtin__.bool>',
                                     'longname': 'history',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'ch'},
                         'image1': {'args': '<type __builtin__.str>',
                                    'longname': 'image1',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i1'},
                         'image2': {'args': '<type __builtin__.str>',
                                    'longname': 'image2',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i2'},
                         'image3': {'args': '<type __builtin__.str>',
                                    'longname': 'image3',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i3'},
                         'localDolly': {'args': '<type __builtin__.bool>',
                                        'longname': 'localDolly',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ld'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'orthoZoom': {'args': '<type __builtin__.bool>',
                                       'longname': 'orthoZoom',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'oz'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'scale': {'args': '<type __builtin__.float>',
                                   'longname': 'scale',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 's'},
                         'toolName': {'args': '<type __builtin__.str>',
                                      'longname': 'toolName',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'tn'}},
               'shortFlags': {'ac': 'alternateContext',
                              'bdt': 'boxDollyType',
                              'cd': 'centerOfInterestDolly',
                              'ch': 'history',
                              'dtc': 'dollyTowardsCenter',
                              'e': 'edit',
                              'ex': 'exists',
                              'i1': 'image1',
                              'i2': 'image2',
                              'i3': 'image3',
                              'ld': 'localDolly',
                              'n': 'name',
                              'oz': 'orthoZoom',
                              'q': 'query',
                              's': 'scale',
                              'tn': 'toolName'},
               'type': 'context'},
  'dopeSheetEditor': {'flags': {'autoFit': {'args': '<type __builtin__.str>',
                                            'longname': 'autoFit',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'af'},
                                'autoFitTime': {'args': '<type __builtin__.str>',
                                                'longname': 'autoFitTime',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'aft'},
                                'control': {'args': '<type __builtin__.bool>',
                                            'longname': 'control',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ctl'},
                                'defineTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'defineTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'dt'},
                                'displayActiveKeyTangents': {'args': '<type __builtin__.str>',
                                                             'longname': 'displayActiveKeyTangents',
                                                             'modes': ['edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'dat'},
                                'displayActiveKeys': {'args': '<type __builtin__.str>',
                                                      'longname': 'displayActiveKeys',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dak'},
                                'displayInfinities': {'args': '<type __builtin__.str>',
                                                      'longname': 'displayInfinities',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'di'},
                                'displayKeys': {'args': '<type __builtin__.str>',
                                                'longname': 'displayKeys',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'dk'},
                                'displayTangents': {'args': '<type __builtin__.str>',
                                                    'longname': 'displayTangents',
                                                    'modes': ['edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dtn'},
                                'displayValues': {'args': '<type __builtin__.str>',
                                                  'longname': 'displayValues',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dv'},
                                'docTag': {'args': '<type __builtin__.str>',
                                           'longname': 'docTag',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dtg'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'filter': {'args': '<type __builtin__.str>',
                                           'longname': 'filter',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'f'},
                                'forceMainConnection': {'args': '<type __builtin__.str>',
                                                        'longname': 'forceMainConnection',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'fmc'},
                                'hierarchyBelow': {'args': '<type __builtin__.bool>',
                                                   'longname': 'hierarchyBelow',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'hb'},
                                'highlightConnection': {'args': '<type __builtin__.str>',
                                                        'longname': 'highlightConnection',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'hlc'},
                                'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                       'longname': 'lockMainConnection',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'lck'},
                                'lookAt': {'args': '<type __builtin__.str>',
                                           'longname': 'lookAt',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'la'},
                                'mainListConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'mainListConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'mlc'},
                                'outliner': {'args': '<type __builtin__.str>',
                                             'longname': 'outliner',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'o'},
                                'panel': {'args': '<type __builtin__.str>',
                                          'longname': 'panel',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'pnl'},
                                'parent': {'args': '<type __builtin__.str>',
                                           'longname': 'parent',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'p'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'selectionConnection': {'args': '<type __builtin__.str>',
                                                        'longname': 'selectionConnection',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'slc'},
                                'selectionWindow': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'selectionWindow',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 4,
                                                    'shortname': 'sel'},
                                'showScene': {'args': '<type __builtin__.bool>',
                                              'longname': 'showScene',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sc'},
                                'showSummary': {'args': '<type __builtin__.bool>',
                                                'longname': 'showSummary',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ss'},
                                'showTicks': {'args': '<type __builtin__.bool>',
                                              'longname': 'showTicks',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'stk'},
                                'snapTime': {'args': '<type __builtin__.str>',
                                             'longname': 'snapTime',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'st'},
                                'snapValue': {'args': '<type __builtin__.str>',
                                              'longname': 'snapValue',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sv'},
                                'stateString': {'args': '<type __builtin__.bool>',
                                                'longname': 'stateString',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'sts'},
                                'unParent': {'args': '<type __builtin__.bool>',
                                             'longname': 'unParent',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'up'},
                                'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'unlockMainConnection',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'ulk'},
                                'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'updateMainConnection',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'upd'},
                                'useTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'useTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ut'}},
                      'shortFlags': {'af': 'autoFit',
                                     'aft': 'autoFitTime',
                                     'ctl': 'control',
                                     'dak': 'displayActiveKeys',
                                     'dat': 'displayActiveKeyTangents',
                                     'di': 'displayInfinities',
                                     'dk': 'displayKeys',
                                     'dt': 'defineTemplate',
                                     'dtg': 'docTag',
                                     'dtn': 'displayTangents',
                                     'dv': 'displayValues',
                                     'e': 'edit',
                                     'ex': 'exists',
                                     'f': 'filter',
                                     'fmc': 'forceMainConnection',
                                     'hb': 'hierarchyBelow',
                                     'hlc': 'highlightConnection',
                                     'la': 'lookAt',
                                     'lck': 'lockMainConnection',
                                     'mlc': 'mainListConnection',
                                     'o': 'outliner',
                                     'p': 'parent',
                                     'pnl': 'panel',
                                     'q': 'query',
                                     'sc': 'showScene',
                                     'sel': 'selectionWindow',
                                     'slc': 'selectionConnection',
                                     'ss': 'showSummary',
                                     'st': 'snapTime',
                                     'stk': 'showTicks',
                                     'sts': 'stateString',
                                     'sv': 'snapValue',
                                     'ulk': 'unlockMainConnection',
                                     'up': 'unParent',
                                     'upd': 'updateMainConnection',
                                     'ut': 'useTemplate'},
                      'type': 'animation'},
  'doubleProfileBirailSurface': {'flags': {'blendFactor': {'args': '<type __builtin__.float>',
                                                           'longname': 'blendFactor',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'bl'},
                                           'caching': {'args': '<type __builtin__.bool>',
                                                       'longname': 'caching',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'cch'},
                                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                                   'longname': 'constructionHistory',
                                                                   'modes': ['create'],
                                                                   'numArgs': 1,
                                                                   'shortname': 'ch'},
                                           'edit': {'args': '<type __builtin__.bool>',
                                                    'longname': 'edit',
                                                    'numArgs': 0,
                                                    'shortname': 'e'},
                                           'frozen': {'args': '<type __builtin__.bool>',
                                                      'longname': 'frozen',
                                                      'numArgs': 1,
                                                      'shortname': 'fzn'},
                                           'name': {'args': '<type __builtin__.str>',
                                                    'longname': 'name',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'n'},
                                           'nodeState': {'args': '<type __builtin__.int>',
                                                         'longname': 'nodeState',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'nds'},
                                           'object': {'args': '<type __builtin__.bool>',
                                                      'longname': 'object',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'o'},
                                           'polygon': {'args': '<type __builtin__.int>',
                                                       'longname': 'polygon',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'po'},
                                           'query': {'args': '<type __builtin__.bool>',
                                                     'longname': 'query',
                                                     'numArgs': 0,
                                                     'shortname': 'q'},
                                           'tangentContinuityProfile1': {'args': '<type __builtin__.bool>',
                                                                         'longname': 'tangentContinuityProfile1',
                                                                         'modes': ['create',
                                                                                   'query',
                                                                                   'edit'],
                                                                         'numArgs': 1,
                                                                         'shortname': 'tp1'},
                                           'tangentContinuityProfile2': {'args': '<type __builtin__.bool>',
                                                                         'longname': 'tangentContinuityProfile2',
                                                                         'modes': ['create',
                                                                                   'query',
                                                                                   'edit'],
                                                                         'numArgs': 1,
                                                                         'shortname': 'tp2'},
                                           'transformMode': {'args': '<type __builtin__.int>',
                                                             'longname': 'transformMode',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'tm'}},
                                 'shortFlags': {'bl': 'blendFactor',
                                                'cch': 'caching',
                                                'ch': 'constructionHistory',
                                                'e': 'edit',
                                                'fzn': 'frozen',
                                                'n': 'name',
                                                'nds': 'nodeState',
                                                'o': 'object',
                                                'po': 'polygon',
                                                'q': 'query',
                                                'tm': 'transformMode',
                                                'tp1': 'tangentContinuityProfile1',
                                                'tp2': 'tangentContinuityProfile2'},
                                 'type': 'modeling'},
  'dpBirailCtx': {'flags': {'activeNodes': {'args': '<type __builtin__.bool>',
                                            'longname': 'activeNodes',
                                            'numArgs': 0,
                                            'shortname': 'anq'},
                            'autoCreate': {'args': '<type __builtin__.bool>',
                                           'longname': 'autoCreate',
                                           'numArgs': 1,
                                           'shortname': 'ac'},
                            'bldProfileFirst': {'args': '<type __builtin__.bool>',
                                                'longname': 'bldProfileFirst',
                                                'numArgs': 1,
                                                'shortname': 'bpf'},
                            'bldProfileLast': {'args': '<type __builtin__.bool>',
                                               'longname': 'bldProfileLast',
                                               'numArgs': 1,
                                               'shortname': 'bpl'},
                            'bldProfiles': {'args': '<type __builtin__.bool>',
                                            'longname': 'bldProfiles',
                                            'numArgs': 1,
                                            'shortname': 'bp'},
                            'bldRailOne': {'args': '<type __builtin__.bool>',
                                           'longname': 'bldRailOne',
                                           'numArgs': 1,
                                           'shortname': 'br1'},
                            'bldRailTwo': {'args': '<type __builtin__.bool>',
                                           'longname': 'bldRailTwo',
                                           'numArgs': 1,
                                           'shortname': 'br2'},
                            'blendFactor': {'args': '<type __builtin__.float>',
                                            'longname': 'blendFactor',
                                            'numArgs': 1,
                                            'shortname': 'bl'},
                            'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'image1': {'args': '<type __builtin__.str>',
                                       'longname': 'image1',
                                       'numArgs': 1,
                                       'shortname': 'i1'},
                            'image2': {'args': '<type __builtin__.str>',
                                       'longname': 'image2',
                                       'numArgs': 1,
                                       'shortname': 'i2'},
                            'image3': {'args': '<type __builtin__.str>',
                                       'longname': 'image3',
                                       'numArgs': 1,
                                       'shortname': 'i3'},
                            'immediate': {'args': '<type __builtin__.bool>',
                                          'longname': 'immediate',
                                          'numArgs': 0,
                                          'shortname': 'im'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'polygon': {'args': '<type __builtin__.int>',
                                        'longname': 'polygon',
                                        'numArgs': 1,
                                        'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'reset': {'args': '<type __builtin__.bool>',
                                      'longname': 'reset',
                                      'numArgs': 0,
                                      'shortname': 'rs'},
                            'tangentContinuityProfile1': {'args': '<type __builtin__.bool>',
                                                          'longname': 'tangentContinuityProfile1',
                                                          'numArgs': 1,
                                                          'shortname': 'tp1'},
                            'tangentContinuityProfile2': {'args': '<type __builtin__.bool>',
                                                          'longname': 'tangentContinuityProfile2',
                                                          'numArgs': 1,
                                                          'shortname': 'tp2'},
                            'toolNode': {'args': '<type __builtin__.bool>',
                                         'longname': 'toolNode',
                                         'numArgs': 0,
                                         'shortname': 'tnq'},
                            'transformMode': {'args': '<type __builtin__.int>',
                                              'longname': 'transformMode',
                                              'numArgs': 1,
                                              'shortname': 'tm'}},
                  'shortFlags': {'ac': 'autoCreate',
                                 'anq': 'activeNodes',
                                 'bl': 'blendFactor',
                                 'bp': 'bldProfiles',
                                 'bpf': 'bldProfileFirst',
                                 'bpl': 'bldProfileLast',
                                 'br1': 'bldRailOne',
                                 'br2': 'bldRailTwo',
                                 'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'e': 'edit',
                                 'ex': 'exists',
                                 'fzn': 'frozen',
                                 'i1': 'image1',
                                 'i2': 'image2',
                                 'i3': 'image3',
                                 'im': 'immediate',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'o': 'object',
                                 'po': 'polygon',
                                 'q': 'query',
                                 'rs': 'reset',
                                 'tm': 'transformMode',
                                 'tnq': 'toolNode',
                                 'tp1': 'tangentContinuityProfile1',
                                 'tp2': 'tangentContinuityProfile2'},
                  'type': 'context'},
  'drag': {'flags': {'attenuation': {'args': '<type __builtin__.float>',
                                     'longname': 'attenuation',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'att'},
                     'directionX': {'args': '<type __builtin__.float>',
                                    'longname': 'directionX',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'dx'},
                     'directionY': {'args': '<type __builtin__.float>',
                                    'longname': 'directionY',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'dy'},
                     'directionZ': {'args': '<type __builtin__.float>',
                                    'longname': 'directionZ',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'dz'},
                     'edit': {'args': '<type __builtin__.bool>',
                              'longname': 'edit',
                              'numArgs': 0,
                              'shortname': 'e'},
                     'magnitude': {'args': '<type __builtin__.float>',
                                   'longname': 'magnitude',
                                   'modes': ['query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'm'},
                     'maxDistance': {'args': '<type __builtin__.float>',
                                     'longname': 'maxDistance',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'mxd'},
                     'name': {'args': '<type __builtin__.str>',
                              'longname': 'name',
                              'modes': ['query', 'edit'],
                              'numArgs': 1,
                              'shortname': 'n'},
                     'perVertex': {'args': '<type __builtin__.bool>',
                                   'longname': 'perVertex',
                                   'modes': ['query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'pv'},
                     'position': {'args': ['<type __builtin__.float>',
                                           '<type __builtin__.float>',
                                           '<type __builtin__.float>'],
                                  'longname': 'position',
                                  'modes': ['query', 'edit', 'multiuse'],
                                  'numArgs': 3,
                                  'shortname': 'pos'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'torusSectionRadius': {'args': '<type __builtin__.float>',
                                            'longname': 'torusSectionRadius',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'tsr'},
                     'useDirection': {'args': '<type __builtin__.bool>',
                                      'longname': 'useDirection',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'ud'},
                     'volumeExclusion': {'args': '<type __builtin__.bool>',
                                         'longname': 'volumeExclusion',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vex'},
                     'volumeOffset': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'volumeOffset',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 3,
                                      'shortname': 'vof'},
                     'volumeShape': {'args': '<type __builtin__.str>',
                                     'longname': 'volumeShape',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'vsh'},
                     'volumeSweep': {'args': '<type __builtin__.float>',
                                     'longname': 'volumeSweep',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'vsw'}},
           'resultNeedsUnpacking': True,
           'shortFlags': {'att': 'attenuation',
                          'dx': 'directionX',
                          'dy': 'directionY',
                          'dz': 'directionZ',
                          'e': 'edit',
                          'm': 'magnitude',
                          'mxd': 'maxDistance',
                          'n': 'name',
                          'pos': 'position',
                          'pv': 'perVertex',
                          'q': 'query',
                          'tsr': 'torusSectionRadius',
                          'ud': 'useDirection',
                          'vex': 'volumeExclusion',
                          'vof': 'volumeOffset',
                          'vsh': 'volumeShape',
                          'vsw': 'volumeSweep'},
           'type': 'effects'},
  'dragAttrContext': {'flags': {'connectTo': {'args': 'PyNode',
                                              'longname': 'connectTo',
                                              'modes': ['create',
                                                        'query',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'ct'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'history': {'args': '<type __builtin__.bool>',
                                            'longname': 'history',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ch'},
                                'image1': {'args': '<type __builtin__.str>',
                                           'longname': 'image1',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i1'},
                                'image2': {'args': '<type __builtin__.str>',
                                           'longname': 'image2',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i2'},
                                'image3': {'args': '<type __builtin__.str>',
                                           'longname': 'image3',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i3'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'reset': {'args': '<type __builtin__.bool>',
                                          'longname': 'reset',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'r'}},
                      'shortFlags': {'ch': 'history',
                                     'ct': 'connectTo',
                                     'e': 'edit',
                                     'ex': 'exists',
                                     'i1': 'image1',
                                     'i2': 'image2',
                                     'i3': 'image3',
                                     'n': 'name',
                                     'q': 'query',
                                     'r': 'reset'},
                      'type': 'context'},
  'draggerContext': {'callbackFlags': ['releaseCommand',
                                       'rc',
                                       'holdCommand',
                                       'hc',
                                       'dragCommand',
                                       'dc',
                                       'finalize',
                                       'fnz',
                                       'prePressCommand',
                                       'ppc',
                                       'initialize',
                                       'inz',
                                       'pressCommand',
                                       'pc'],
                     'flags': {'anchorPoint': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'anchorPoint',
                                               'modes': ['query'],
                                               'numArgs': 3,
                                               'shortname': 'ap'},
                               'button': {'args': '<type __builtin__.int>',
                                          'longname': 'button',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'bu'},
                               'currentStep': {'args': '<type __builtin__.int>',
                                               'longname': 'currentStep',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'cs'},
                               'cursor': {'args': '<type __builtin__.str>',
                                          'longname': 'cursor',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cur'},
                               'dragCommand': {'args': 'script',
                                               'longname': 'dragCommand',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dc'},
                               'dragPoint': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'dragPoint',
                                             'modes': ['query'],
                                             'numArgs': 3,
                                             'shortname': 'dp'},
                               'drawString': {'args': '<type __builtin__.str>',
                                              'longname': 'drawString',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ds'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'finalize': {'args': 'script',
                                            'longname': 'finalize',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'fnz'},
                               'helpString': {'args': '<type __builtin__.str>',
                                              'longname': 'helpString',
                                              'modes': ['query'],
                                              'numArgs': 1,
                                              'shortname': 'hs'},
                               'history': {'args': '<type __builtin__.bool>',
                                           'longname': 'history',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ch'},
                               'holdCommand': {'args': 'script',
                                               'longname': 'holdCommand',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'hc'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'initialize': {'args': 'script',
                                              'longname': 'initialize',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'inz'},
                               'modifier': {'args': '<type __builtin__.str>',
                                            'longname': 'modifier',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'mo'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'plane': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'plane',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 3,
                                         'shortname': 'pl'},
                               'prePressCommand': {'args': 'script',
                                                   'longname': 'prePressCommand',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ppc'},
                               'pressCommand': {'args': 'script',
                                                'longname': 'pressCommand',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'pc'},
                               'projection': {'args': '<type __builtin__.str>',
                                              'longname': 'projection',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pr'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'releaseCommand': {'args': 'script',
                                                  'longname': 'releaseCommand',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'rc'},
                               'snapping': {'args': '<type __builtin__.bool>',
                                            'longname': 'snapping',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'snp'},
                               'space': {'args': '<type __builtin__.str>',
                                         'longname': 'space',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'sp'},
                               'stepsCount': {'args': '<type __builtin__.int>',
                                              'longname': 'stepsCount',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sc'},
                               'undoMode': {'args': '<type __builtin__.str>',
                                            'longname': 'undoMode',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'um'}},
                     'shortFlags': {'ap': 'anchorPoint',
                                    'bu': 'button',
                                    'ch': 'history',
                                    'cs': 'currentStep',
                                    'cur': 'cursor',
                                    'dc': 'dragCommand',
                                    'dp': 'dragPoint',
                                    'ds': 'drawString',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'fnz': 'finalize',
                                    'hc': 'holdCommand',
                                    'hs': 'helpString',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'inz': 'initialize',
                                    'mo': 'modifier',
                                    'n': 'name',
                                    'pc': 'pressCommand',
                                    'pl': 'plane',
                                    'ppc': 'prePressCommand',
                                    'pr': 'projection',
                                    'q': 'query',
                                    'rc': 'releaseCommand',
                                    'sc': 'stepsCount',
                                    'snp': 'snapping',
                                    'sp': 'space',
                                    'um': 'undoMode'},
                     'type': 'context'},
  'drawExtrudeFacetCtx': {'flags': {'degree': {'args': '<type __builtin__.int>',
                                               'longname': 'degree',
                                               'numArgs': 1,
                                               'shortname': 'd'},
                                    'divisions': {'args': '<type __builtin__.int>',
                                                  'longname': 'divisions',
                                                  'numArgs': 1,
                                                  'shortname': 'div'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'facesTogether': {'args': '<type __builtin__.bool>',
                                                      'longname': 'facesTogether',
                                                      'numArgs': 1,
                                                      'shortname': 'ft'},
                                    'history': {'args': '<type __builtin__.bool>',
                                                'longname': 'history',
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'}},
                          'shortFlags': {'ch': 'history',
                                         'd': 'degree',
                                         'div': 'divisions',
                                         'e': 'edit',
                                         'ex': 'exists',
                                         'ft': 'facesTogether',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'n': 'name',
                                         'q': 'query'},
                          'type': 'context'},
  'dropoffLocator': {'flags': {}, 'shortFlags': {}, 'type': 'animation'},
  'duplicate': {'flags': {'inputConnections': {'args': '<type __builtin__.bool>',
                                               'longname': 'inputConnections',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ic'},
                          'instanceLeaf': {'args': '<type __builtin__.bool>',
                                           'longname': 'instanceLeaf',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ilf'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'parentOnly': {'args': '<type __builtin__.bool>',
                                         'longname': 'parentOnly',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'po'},
                          'renameChildren': {'args': '<type __builtin__.bool>',
                                             'longname': 'renameChildren',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'rc'},
                          'returnRootsOnly': {'args': '<type __builtin__.bool>',
                                              'longname': 'returnRootsOnly',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'rr'},
                          'smartTransform': {'args': '<type __builtin__.bool>',
                                             'longname': 'smartTransform',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'st'},
                          'transformsOnly': {'args': '<type __builtin__.bool>',
                                             'longname': 'transformsOnly',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'to'},
                          'upstreamNodes': {'args': '<type __builtin__.bool>',
                                            'longname': 'upstreamNodes',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'un'}},
                'shortFlags': {'ic': 'inputConnections',
                               'ilf': 'instanceLeaf',
                               'n': 'name',
                               'po': 'parentOnly',
                               'rc': 'renameChildren',
                               'rr': 'returnRootsOnly',
                               'st': 'smartTransform',
                               'to': 'transformsOnly',
                               'un': 'upstreamNodes'},
                'type': 'general'},
  'duplicateCurve': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'local': {'args': '<type __builtin__.bool>',
                                         'longname': 'local',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'l'},
                               'maxValue': {'args': '<type __builtin__.float>',
                                            'longname': 'maxValue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'max'},
                               'mergeItems': {'args': '<type __builtin__.bool>',
                                              'longname': 'mergeItems',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'mi'},
                               'minValue': {'args': '<type __builtin__.float>',
                                            'longname': 'minValue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'min'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'object': {'args': '<type __builtin__.bool>',
                                          'longname': 'object',
                                          'numArgs': 1,
                                          'shortname': 'o'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'range': {'args': '<type __builtin__.bool>',
                                         'longname': 'range',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'rn'},
                               'relative': {'args': '<type __builtin__.bool>',
                                            'longname': 'relative',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'r'}},
                     'shortFlags': {'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'e': 'edit',
                                    'fzn': 'frozen',
                                    'l': 'local',
                                    'max': 'maxValue',
                                    'mi': 'mergeItems',
                                    'min': 'minValue',
                                    'n': 'name',
                                    'nds': 'nodeState',
                                    'o': 'object',
                                    'q': 'query',
                                    'r': 'relative',
                                    'rn': 'range'},
                     'type': 'modeling'},
  'duplicateSurface': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                             'longname': 'caching',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cch'},
                                 'constructionHistory': {'args': '<type __builtin__.bool>',
                                                         'longname': 'constructionHistory',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'ch'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'faceCountU': {'args': '<type __builtin__.int>',
                                                'longname': 'faceCountU',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'fcu'},
                                 'faceCountV': {'args': '<type __builtin__.int>',
                                                'longname': 'faceCountV',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'fcv'},
                                 'firstFaceU': {'args': '<type __builtin__.int>',
                                                'longname': 'firstFaceU',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ffu'},
                                 'firstFaceV': {'args': '<type __builtin__.int>',
                                                'longname': 'firstFaceV',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ffv'},
                                 'frozen': {'args': '<type __builtin__.bool>',
                                            'longname': 'frozen',
                                            'numArgs': 1,
                                            'shortname': 'fzn'},
                                 'local': {'args': '<type __builtin__.bool>',
                                           'longname': 'local',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'l'},
                                 'mergeItems': {'args': '<type __builtin__.bool>',
                                                'longname': 'mergeItems',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'mi'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'nodeState': {'args': '<type __builtin__.int>',
                                               'longname': 'nodeState',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nds'},
                                 'object': {'args': '<type __builtin__.bool>',
                                            'longname': 'object',
                                            'numArgs': 1,
                                            'shortname': 'o'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'}},
                       'shortFlags': {'cch': 'caching',
                                      'ch': 'constructionHistory',
                                      'e': 'edit',
                                      'fcu': 'faceCountU',
                                      'fcv': 'faceCountV',
                                      'ffu': 'firstFaceU',
                                      'ffv': 'firstFaceV',
                                      'fzn': 'frozen',
                                      'l': 'local',
                                      'mi': 'mergeItems',
                                      'n': 'name',
                                      'nds': 'nodeState',
                                      'o': 'object',
                                      'q': 'query'},
                       'type': 'modeling'},
  'dynCache': {'flags': {}, 'shortFlags': {}, 'type': 'effects'},
  'dynControl': {'flags': {'autoCreate': {'args': '<type __builtin__.bool>',
                                          'longname': 'autoCreate',
                                          'numArgs': 1,
                                          'shortname': 'ac'},
                           'oversample': {'args': '<type __builtin__.int>',
                                          'longname': 'oversample',
                                          'numArgs': 1,
                                          'shortname': 'os'},
                           'particleCache': {'args': '<type __builtin__.bool>',
                                             'longname': 'particleCache',
                                             'numArgs': 1,
                                             'shortname': 'pc'},
                           'particleLOD': {'args': '<type __builtin__.float>',
                                           'longname': 'particleLOD',
                                           'numArgs': 1,
                                           'shortname': 'pld'},
                           'particlesOn': {'args': '<type __builtin__.bool>',
                                           'longname': 'particlesOn',
                                           'numArgs': 1,
                                           'shortname': 'po'},
                           'rigidOn': {'args': '<type __builtin__.bool>',
                                       'longname': 'rigidOn',
                                       'numArgs': 1,
                                       'shortname': 'ro'},
                           'seed': {'args': '<type __builtin__.int>',
                                    'longname': 'seed',
                                    'numArgs': 1,
                                    'shortname': 'sd'},
                           'startTime': {'args': u'time',
                                         'longname': 'startTime',
                                         'numArgs': 1,
                                         'shortname': 'st'},
                           'traceDepth': {'args': '<type __builtin__.int>',
                                          'longname': 'traceDepth',
                                          'numArgs': 1,
                                          'shortname': 'td'}},
                 'shortFlags': {'ac': 'autoCreate',
                                'os': 'oversample',
                                'pc': 'particleCache',
                                'pld': 'particleLOD',
                                'po': 'particlesOn',
                                'ro': 'rigidOn',
                                'sd': 'seed',
                                'st': 'startTime',
                                'td': 'traceDepth'},
                 'type': 'effects'},
  'dynExport': {'flags': {'allObjects': {'args': '<type __builtin__.bool>',
                                         'longname': 'allObjects',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'all'},
                          'attribute': {'args': '<type __builtin__.str>',
                                        'longname': 'attribute',
                                        'modes': ['create', 'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'atr'},
                          'format': {'args': '<type __builtin__.str>',
                                     'longname': 'format',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'f'},
                          'maxFrame': {'args': 'time',
                                       'longname': 'maxFrame',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'mxf'},
                          'minFrame': {'args': 'time',
                                       'longname': 'minFrame',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'mnf'},
                          'onlyUpdateParticles': {'args': '<type __builtin__.bool>',
                                                  'longname': 'onlyUpdateParticles',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'oup'},
                          'overSampling': {'args': '<type __builtin__.int>',
                                           'longname': 'overSampling',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'os'},
                          'path': {'args': '<type __builtin__.str>',
                                   'longname': 'path',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'p'}},
                'shortFlags': {'all': 'allObjects',
                               'atr': 'attribute',
                               'f': 'format',
                               'mnf': 'minFrame',
                               'mxf': 'maxFrame',
                               'os': 'overSampling',
                               'oup': 'onlyUpdateParticles',
                               'p': 'path'},
                'type': 'effects'},
  'dynExpression': {'flags': {'creation': {'args': '<type __builtin__.bool>',
                                           'longname': 'creation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 0,
                                           'shortname': 'c'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'runtime': {'args': '<type __builtin__.bool>',
                                          'longname': 'runtime',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'r'},
                              'runtimeAfterDynamics': {'args': '<type __builtin__.bool>',
                                                       'longname': 'runtimeAfterDynamics',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'rad'},
                              'runtimeBeforeDynamics': {'args': '<type __builtin__.bool>',
                                                        'longname': 'runtimeBeforeDynamics',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'rbd'},
                              'string': {'args': '<type __builtin__.str>',
                                         'longname': 'string',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 's'}},
                    'shortFlags': {'c': 'creation',
                                   'e': 'edit',
                                   'n': 'name',
                                   'q': 'query',
                                   'r': 'runtime',
                                   'rad': 'runtimeAfterDynamics',
                                   'rbd': 'runtimeBeforeDynamics',
                                   's': 'string'},
                    'type': 'effects'},
  'dynGlobals': {'flags': {'active': {'args': '<type __builtin__.bool>',
                                      'longname': 'active',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'a'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'listAll': {'args': '<type __builtin__.bool>',
                                       'longname': 'listAll',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'la'},
                           'overSampling': {'args': '<type __builtin__.int>',
                                            'longname': 'overSampling',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'os'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'a': 'active',
                                'e': 'edit',
                                'la': 'listAll',
                                'os': 'overSampling',
                                'q': 'query'},
                 'type': 'effects'},
  'dynPaintCtx': {'flags': {'brushDrag': {'args': '<type __builtin__.float>',
                                          'longname': 'brushDrag',
                                          'numArgs': 1,
                                          'shortname': 'bd'},
                            'brushMass': {'args': '<type __builtin__.float>',
                                          'longname': 'brushMass',
                                          'numArgs': 1,
                                          'shortname': 'bm'},
                            'displayQuality': {'args': '<type __builtin__.float>',
                                               'longname': 'displayQuality',
                                               'numArgs': 1,
                                               'shortname': 'dq'},
                            'doProject': {'args': '<type __builtin__.int>',
                                          'longname': 'doProject',
                                          'numArgs': 1,
                                          'shortname': 'dp'},
                            'dragBrushSize': {'args': '<type __builtin__.str>',
                                              'longname': 'dragBrushSize',
                                              'numArgs': 1,
                                              'shortname': 'dbs'},
                            'drawAsMesh': {'args': '<type __builtin__.bool>',
                                           'longname': 'drawAsMesh',
                                           'numArgs': 1,
                                           'shortname': 'dam'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'image1': {'args': '<type __builtin__.str>',
                                       'longname': 'image1',
                                       'numArgs': 1,
                                       'shortname': 'i1'},
                            'image2': {'args': '<type __builtin__.str>',
                                       'longname': 'image2',
                                       'numArgs': 1,
                                       'shortname': 'i2'},
                            'image3': {'args': '<type __builtin__.str>',
                                       'longname': 'image3',
                                       'numArgs': 1,
                                       'shortname': 'i3'},
                            'inputSamples': {'args': '<type __builtin__.int>',
                                             'longname': 'inputSamples',
                                             'numArgs': 1,
                                             'shortname': 'inputSamples'},
                            'paintAtDepth': {'args': '<type __builtin__.bool>',
                                             'longname': 'paintAtDepth',
                                             'numArgs': 1,
                                             'shortname': 'pd'},
                            'pixelMode': {'args': '<type __builtin__.int>',
                                          'longname': 'pixelMode',
                                          'numArgs': 1,
                                          'shortname': 'pxm'},
                            'pressureMapping1': {'args': '<type __builtin__.int>',
                                                 'longname': 'pressureMapping1',
                                                 'numArgs': 1,
                                                 'shortname': 'pm1'},
                            'pressureMapping2': {'args': '<type __builtin__.int>',
                                                 'longname': 'pressureMapping2',
                                                 'numArgs': 1,
                                                 'shortname': 'pm2'},
                            'pressureMapping3': {'args': '<type __builtin__.int>',
                                                 'longname': 'pressureMapping3',
                                                 'numArgs': 1,
                                                 'shortname': 'pm3'},
                            'pressureMax1': {'args': '<type __builtin__.float>',
                                             'longname': 'pressureMax1',
                                             'numArgs': 1,
                                             'shortname': 'px1'},
                            'pressureMax2': {'args': '<type __builtin__.float>',
                                             'longname': 'pressureMax2',
                                             'numArgs': 1,
                                             'shortname': 'px2'},
                            'pressureMax3': {'args': '<type __builtin__.float>',
                                             'longname': 'pressureMax3',
                                             'numArgs': 1,
                                             'shortname': 'px3'},
                            'pressureMin1': {'args': '<type __builtin__.float>',
                                             'longname': 'pressureMin1',
                                             'numArgs': 1,
                                             'shortname': 'ps1'},
                            'pressureMin2': {'args': '<type __builtin__.float>',
                                             'longname': 'pressureMin2',
                                             'numArgs': 1,
                                             'shortname': 'ps2'},
                            'pressureMin3': {'args': '<type __builtin__.float>',
                                             'longname': 'pressureMin3',
                                             'numArgs': 1,
                                             'shortname': 'ps3'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'sampleSeparation': {'args': '<type __builtin__.float>',
                                                 'longname': 'sampleSeparation',
                                                 'numArgs': 1,
                                                 'shortname': 'sp'},
                            'setSelection': {'args': '<type __builtin__.bool>',
                                             'longname': 'setSelection',
                                             'numArgs': 0,
                                             'shortname': 'ss'},
                            'surfaceOffset': {'args': '<type __builtin__.float>',
                                              'longname': 'surfaceOffset',
                                              'numArgs': 1,
                                              'shortname': 'sof'},
                            'usePressure': {'args': '<type __builtin__.bool>',
                                            'longname': 'usePressure',
                                            'numArgs': 1,
                                            'shortname': 'usp'}},
                  'removedFlags': {'is': 'inputSamples'},
                  'shortFlags': {'bd': 'brushDrag',
                                 'bm': 'brushMass',
                                 'dam': 'drawAsMesh',
                                 'dbs': 'dragBrushSize',
                                 'dp': 'doProject',
                                 'dq': 'displayQuality',
                                 'e': 'edit',
                                 'ex': 'exists',
                                 'i1': 'image1',
                                 'i2': 'image2',
                                 'i3': 'image3',
                                 'inputSamples': 'inputSamples',
                                 'pd': 'paintAtDepth',
                                 'pm1': 'pressureMapping1',
                                 'pm2': 'pressureMapping2',
                                 'pm3': 'pressureMapping3',
                                 'ps1': 'pressureMin1',
                                 'ps2': 'pressureMin2',
                                 'ps3': 'pressureMin3',
                                 'px1': 'pressureMax1',
                                 'px2': 'pressureMax2',
                                 'px3': 'pressureMax3',
                                 'pxm': 'pixelMode',
                                 'q': 'query',
                                 'sof': 'surfaceOffset',
                                 'sp': 'sampleSeparation',
                                 'ss': 'setSelection',
                                 'usp': 'usePressure'},
                  'type': 'context'},
  'dynPaintEditor': {'callbackFlags': ['changeCommand', 'cc'],
                     'flags': {'activeOnly': {'args': '<type __builtin__.bool>',
                                              'longname': 'activeOnly',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ao'},
                               'autoSave': {'args': '<type __builtin__.bool>',
                                            'longname': 'autoSave',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'autoSave'},
                               'camera': {'args': '<type __builtin__.str>',
                                          'longname': 'camera',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cam'},
                               'canvasMode': {'args': '<type __builtin__.bool>',
                                              'longname': 'canvasMode',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cm'},
                               'canvasUndo': {'args': '<type __builtin__.bool>',
                                              'longname': 'canvasUndo',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'cu'},
                               'changeCommand': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'changeCommand',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 4,
                                                 'shortname': 'cc'},
                               'clear': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'clear',
                                         'modes': ['edit'],
                                         'numArgs': 3,
                                         'shortname': 'cl'},
                               'control': {'args': '<type __builtin__.bool>',
                                           'longname': 'control',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'ctl'},
                               'currentCanvasSize': {'args': '<type __builtin__.bool>',
                                                     'longname': 'currentCanvasSize',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'ccs'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'displayAppearance': {'args': '<type __builtin__.str>',
                                                     'longname': 'displayAppearance',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dsa'},
                               'displayFog': {'args': '<type __builtin__.bool>',
                                              'longname': 'displayFog',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dfg'},
                               'displayImage': {'args': '<type __builtin__.int>',
                                                'longname': 'displayImage',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'di'},
                               'displayLights': {'args': '<type __builtin__.str>',
                                                 'longname': 'displayLights',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dsl'},
                               'displayStyle': {'args': '<type __builtin__.str>',
                                                'longname': 'displayStyle',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dst'},
                               'displayTextures': {'args': '<type __builtin__.bool>',
                                                   'longname': 'displayTextures',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dtx'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'doubleBuffer': {'args': '<type __builtin__.bool>',
                                                'longname': 'doubleBuffer',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'dbf'},
                               'drawAxis': {'args': '<type __builtin__.bool>',
                                            'longname': 'drawAxis',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'da'},
                               'drawContext': {'args': '<type __builtin__.bool>',
                                               'longname': 'drawContext',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'drc'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'fastUpdate': {'args': '<type __builtin__.int>',
                                              'longname': 'fastUpdate',
                                              'modes': [],
                                              'numArgs': 1,
                                              'shortname': 'fu'},
                               'fileName': {'args': '<type __builtin__.str>',
                                            'longname': 'fileName',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'fil'},
                               'filter': {'args': '<type __builtin__.str>',
                                          'longname': 'filter',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'f'},
                               'forceMainConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'forceMainConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'fmc'},
                               'highlightConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'highlightConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'hlc'},
                               'iconGrab': {'args': '<type __builtin__.bool>',
                                            'longname': 'iconGrab',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'ig'},
                               'loadImage': {'args': '<type __builtin__.str>',
                                             'longname': 'loadImage',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'li'},
                               'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                      'longname': 'lockMainConnection',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'lck'},
                               'mainListConnection': {'args': '<type __builtin__.str>',
                                                      'longname': 'mainListConnection',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'mlc'},
                               'menu': {'args': '<type __builtin__.str>',
                                        'longname': 'menu',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'mn'},
                               'nbImages': {'args': '<type __builtin__.bool>',
                                            'longname': 'nbImages',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'nim'},
                               'newImage': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'newImage',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 5,
                                            'shortname': 'ni'},
                               'paintAll': {'args': '<type __builtin__.float>',
                                            'longname': 'paintAll',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'pa'},
                               'panel': {'args': '<type __builtin__.str>',
                                         'longname': 'panel',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'pnl'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'redrawLast': {'args': '<type __builtin__.bool>',
                                              'longname': 'redrawLast',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'rl'},
                               'refresh': {'args': '<type __builtin__.bool>',
                                           'longname': 'refresh',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'ref'},
                               'refreshMode': {'args': '<type __builtin__.int>',
                                               'longname': 'refreshMode',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rmd'},
                               'removeAllImages': {'args': '<type __builtin__.bool>',
                                                   'longname': 'removeAllImages',
                                                   'modes': ['edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'ra'},
                               'removeImage': {'args': '<type __builtin__.bool>',
                                               'longname': 'removeImage',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'ri'},
                               'rollImage': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'rollImage',
                                             'modes': ['edit'],
                                             'numArgs': 2,
                                             'shortname': 'rig'},
                               'saveAlpha': {'args': '<type __builtin__.bool>',
                                             'longname': 'saveAlpha',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sa'},
                               'saveBumpmap': {'args': '<type __builtin__.str>',
                                               'longname': 'saveBumpmap',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sbm'},
                               'saveImage': {'args': '<type __builtin__.bool>',
                                             'longname': 'saveImage',
                                             'modes': ['edit'],
                                             'numArgs': 0,
                                             'shortname': 'si'},
                               'scaleBlue': {'args': '<type __builtin__.float>',
                                             'longname': 'scaleBlue',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sb'},
                               'scaleGreen': {'args': '<type __builtin__.float>',
                                              'longname': 'scaleGreen',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sg'},
                               'scaleRed': {'args': '<type __builtin__.float>',
                                            'longname': 'scaleRed',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sr'},
                               'selectionConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'selectionConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'slc'},
                               'singleBuffer': {'args': '<type __builtin__.bool>',
                                                'longname': 'singleBuffer',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'sbf'},
                               'snapShot': {'args': '<type __builtin__.bool>',
                                            'longname': 'snapShot',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'snp'},
                               'stateString': {'args': '<type __builtin__.bool>',
                                               'longname': 'stateString',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'sts'},
                               'swap': {'args': '<type __builtin__.int>',
                                        'longname': 'swap',
                                        'modes': [],
                                        'numArgs': 1,
                                        'shortname': 'swp'},
                               'tileSize': {'args': '<type __builtin__.int>',
                                            'longname': 'tileSize',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'ts'},
                               'unParent': {'args': '<type __builtin__.bool>',
                                            'longname': 'unParent',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'up'},
                               'undoCache': {'args': '<type __builtin__.bool>',
                                             'longname': 'undoCache',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'uc'},
                               'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                        'longname': 'unlockMainConnection',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'ulk'},
                               'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                        'longname': 'updateMainConnection',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'upd'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'},
                               'wrap': {'args': ['<type __builtin__.bool>',
                                                 '<type __builtin__.bool>'],
                                        'longname': 'wrap',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 2,
                                        'shortname': 'wr'},
                               'writeImage': {'args': '<type __builtin__.str>',
                                              'longname': 'writeImage',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'wi'},
                               'zoom': {'args': '<type __builtin__.float>',
                                        'longname': 'zoom',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'zm'}},
                     'removedFlags': {'as': 'autoSave'},
                     'shortFlags': {'ao': 'activeOnly',
                                    'autoSave': 'autoSave',
                                    'cam': 'camera',
                                    'cc': 'changeCommand',
                                    'ccs': 'currentCanvasSize',
                                    'cl': 'clear',
                                    'cm': 'canvasMode',
                                    'ctl': 'control',
                                    'cu': 'canvasUndo',
                                    'da': 'drawAxis',
                                    'dbf': 'doubleBuffer',
                                    'dfg': 'displayFog',
                                    'di': 'displayImage',
                                    'drc': 'drawContext',
                                    'dsa': 'displayAppearance',
                                    'dsl': 'displayLights',
                                    'dst': 'displayStyle',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'dtx': 'displayTextures',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'f': 'filter',
                                    'fil': 'fileName',
                                    'fmc': 'forceMainConnection',
                                    'fu': 'fastUpdate',
                                    'hlc': 'highlightConnection',
                                    'ig': 'iconGrab',
                                    'lck': 'lockMainConnection',
                                    'li': 'loadImage',
                                    'mlc': 'mainListConnection',
                                    'mn': 'menu',
                                    'ni': 'newImage',
                                    'nim': 'nbImages',
                                    'p': 'parent',
                                    'pa': 'paintAll',
                                    'pnl': 'panel',
                                    'q': 'query',
                                    'ra': 'removeAllImages',
                                    'ref': 'refresh',
                                    'ri': 'removeImage',
                                    'rig': 'rollImage',
                                    'rl': 'redrawLast',
                                    'rmd': 'refreshMode',
                                    'sa': 'saveAlpha',
                                    'sb': 'scaleBlue',
                                    'sbf': 'singleBuffer',
                                    'sbm': 'saveBumpmap',
                                    'sg': 'scaleGreen',
                                    'si': 'saveImage',
                                    'slc': 'selectionConnection',
                                    'snp': 'snapShot',
                                    'sr': 'scaleRed',
                                    'sts': 'stateString',
                                    'swp': 'swap',
                                    'ts': 'tileSize',
                                    'uc': 'undoCache',
                                    'ulk': 'unlockMainConnection',
                                    'up': 'unParent',
                                    'upd': 'updateMainConnection',
                                    'ut': 'useTemplate',
                                    'wi': 'writeImage',
                                    'wr': 'wrap',
                                    'zm': 'zoom'},
                     'type': 'effects'},
  'dynParticleCtx': {'flags': {'conserve': {'args': '<type __builtin__.float>',
                                            'longname': 'conserve',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'c'},
                               'cursorPlacement': {'args': '<type __builtin__.bool>',
                                                   'longname': 'cursorPlacement',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'cp'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'grid': {'args': '<type __builtin__.bool>',
                                        'longname': 'grid',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'gr'},
                               'gridSpacing': {'args': '<type __builtin__.float>',
                                               'longname': 'gridSpacing',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'grs'},
                               'history': {'args': '<type __builtin__.bool>',
                                           'longname': 'history',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ch'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'jitterRadius': {'args': '<type __builtin__.float>',
                                                'longname': 'jitterRadius',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'jr'},
                               'lowerLeftX': {'args': '<type __builtin__.float>',
                                              'longname': 'lowerLeftX',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'llx'},
                               'lowerLeftY': {'args': '<type __builtin__.float>',
                                              'longname': 'lowerLeftY',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lly'},
                               'lowerLeftZ': {'args': '<type __builtin__.float>',
                                              'longname': 'lowerLeftZ',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'llz'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'nucleus': {'args': '<type __builtin__.bool>',
                                           'longname': 'nucleus',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nc'},
                               'numJitters': {'args': '<type __builtin__.int>',
                                              'longname': 'numJitters',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nj'},
                               'particleName': {'args': '<type __builtin__.str>',
                                                'longname': 'particleName',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'pn'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'sketch': {'args': '<type __builtin__.bool>',
                                          'longname': 'sketch',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sk'},
                               'sketchInterval': {'args': '<type __builtin__.int>',
                                                  'longname': 'sketchInterval',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ski'},
                               'textPlacement': {'args': '<type __builtin__.bool>',
                                                 'longname': 'textPlacement',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'tp'},
                               'upperRightX': {'args': '<type __builtin__.float>',
                                               'longname': 'upperRightX',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'urx'},
                               'upperRightY': {'args': '<type __builtin__.float>',
                                               'longname': 'upperRightY',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ury'},
                               'upperZ': {'args': '<type __builtin__.float>',
                                          'longname': 'upperZ',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'urz'}},
                     'shortFlags': {'c': 'conserve',
                                    'ch': 'history',
                                    'cp': 'cursorPlacement',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'gr': 'grid',
                                    'grs': 'gridSpacing',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'jr': 'jitterRadius',
                                    'llx': 'lowerLeftX',
                                    'lly': 'lowerLeftY',
                                    'llz': 'lowerLeftZ',
                                    'n': 'name',
                                    'nc': 'nucleus',
                                    'nj': 'numJitters',
                                    'pn': 'particleName',
                                    'q': 'query',
                                    'sk': 'sketch',
                                    'ski': 'sketchInterval',
                                    'tp': 'textPlacement',
                                    'urx': 'upperRightX',
                                    'ury': 'upperRightY',
                                    'urz': 'upperZ'},
                     'type': 'context'},
  'dynPref': {'flags': {'autoCreate': {'args': '<type __builtin__.bool>',
                                       'longname': 'autoCreate',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'ac'},
                        'echoCollision': {'args': '<type __builtin__.bool>',
                                          'longname': 'echoCollision',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'ec'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'runupFrom': {'args': '<type __builtin__.int>',
                                      'longname': 'runupFrom',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'rf'},
                        'runupToCurrentTime': {'args': '<type __builtin__.bool>',
                                               'longname': 'runupToCurrentTime',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'rt'},
                        'saveOnQuit': {'args': '<type __builtin__.bool>',
                                       'longname': 'saveOnQuit',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'sq'},
                        'saveRuntimeState': {'args': '<type __builtin__.bool>',
                                             'longname': 'saveRuntimeState',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'sr'}},
              'shortFlags': {'ac': 'autoCreate',
                             'ec': 'echoCollision',
                             'q': 'query',
                             'rf': 'runupFrom',
                             'rt': 'runupToCurrentTime',
                             'sq': 'saveOnQuit',
                             'sr': 'saveRuntimeState'},
              'type': 'effects'},
  'dynSelectCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'numArgs': 1,
                                        'shortname': 'enb'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'image1': {'args': '<type __builtin__.str>',
                                        'longname': 'image1',
                                        'numArgs': 1,
                                        'shortname': 'i1'},
                             'image2': {'args': '<type __builtin__.str>',
                                        'longname': 'image2',
                                        'numArgs': 1,
                                        'shortname': 'i2'},
                             'image3': {'args': '<type __builtin__.str>',
                                        'longname': 'image3',
                                        'numArgs': 1,
                                        'shortname': 'i3'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'e': 'edit',
                                  'enb': 'enable',
                                  'ex': 'exists',
                                  'i1': 'image1',
                                  'i2': 'image2',
                                  'i3': 'image3',
                                  'q': 'query'},
                   'type': 'context'},
  'dynTestData': {'flags': {'arrayAttrs': {'args': '<type __builtin__.bool>',
                                           'longname': 'arrayAttrs',
                                           'numArgs': 1,
                                           'shortname': 'aa'},
                            'verbose': {'args': '<type __builtin__.bool>',
                                        'longname': 'verbose',
                                        'numArgs': 1,
                                        'shortname': 'v'}},
                  'shortFlags': {'aa': 'arrayAttrs', 'v': 'verbose'},
                  'type': 'other'},
  'dynWireCtx': {'flags': {'brushDrag': {'args': '<type __builtin__.float>',
                                         'longname': 'brushDrag',
                                         'numArgs': 1,
                                         'shortname': 'bd'},
                           'brushMass': {'args': '<type __builtin__.float>',
                                         'longname': 'brushMass',
                                         'numArgs': 1,
                                         'shortname': 'bm'},
                           'displayQuality': {'args': '<type __builtin__.float>',
                                              'longname': 'displayQuality',
                                              'numArgs': 1,
                                              'shortname': 'dq'},
                           'doProject': {'args': '<type __builtin__.int>',
                                         'longname': 'doProject',
                                         'numArgs': 1,
                                         'shortname': 'dp'},
                           'dragBrushSize': {'args': '<type __builtin__.str>',
                                             'longname': 'dragBrushSize',
                                             'numArgs': 1,
                                             'shortname': 'dbs'},
                           'drawAsMesh': {'args': '<type __builtin__.bool>',
                                          'longname': 'drawAsMesh',
                                          'numArgs': 1,
                                          'shortname': 'dam'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'inputSamples': {'args': '<type __builtin__.int>',
                                            'longname': 'inputSamples',
                                            'numArgs': 1,
                                            'shortname': 'inputSamples'},
                           'paintAtDepth': {'args': '<type __builtin__.bool>',
                                            'longname': 'paintAtDepth',
                                            'numArgs': 1,
                                            'shortname': 'pd'},
                           'pixelMode': {'args': '<type __builtin__.int>',
                                         'longname': 'pixelMode',
                                         'numArgs': 1,
                                         'shortname': 'pxm'},
                           'pressureMapping1': {'args': '<type __builtin__.int>',
                                                'longname': 'pressureMapping1',
                                                'numArgs': 1,
                                                'shortname': 'pm1'},
                           'pressureMapping2': {'args': '<type __builtin__.int>',
                                                'longname': 'pressureMapping2',
                                                'numArgs': 1,
                                                'shortname': 'pm2'},
                           'pressureMapping3': {'args': '<type __builtin__.int>',
                                                'longname': 'pressureMapping3',
                                                'numArgs': 1,
                                                'shortname': 'pm3'},
                           'pressureMax1': {'args': '<type __builtin__.float>',
                                            'longname': 'pressureMax1',
                                            'numArgs': 1,
                                            'shortname': 'px1'},
                           'pressureMax2': {'args': '<type __builtin__.float>',
                                            'longname': 'pressureMax2',
                                            'numArgs': 1,
                                            'shortname': 'px2'},
                           'pressureMax3': {'args': '<type __builtin__.float>',
                                            'longname': 'pressureMax3',
                                            'numArgs': 1,
                                            'shortname': 'px3'},
                           'pressureMin1': {'args': '<type __builtin__.float>',
                                            'longname': 'pressureMin1',
                                            'numArgs': 1,
                                            'shortname': 'ps1'},
                           'pressureMin2': {'args': '<type __builtin__.float>',
                                            'longname': 'pressureMin2',
                                            'numArgs': 1,
                                            'shortname': 'ps2'},
                           'pressureMin3': {'args': '<type __builtin__.float>',
                                            'longname': 'pressureMin3',
                                            'numArgs': 1,
                                            'shortname': 'ps3'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'sampleSeparation': {'args': '<type __builtin__.float>',
                                                'longname': 'sampleSeparation',
                                                'numArgs': 1,
                                                'shortname': 'sp'},
                           'setSelection': {'args': '<type __builtin__.bool>',
                                            'longname': 'setSelection',
                                            'numArgs': 0,
                                            'shortname': 'ss'},
                           'surfaceOffset': {'args': '<type __builtin__.float>',
                                             'longname': 'surfaceOffset',
                                             'numArgs': 1,
                                             'shortname': 'sof'},
                           'usePressure': {'args': '<type __builtin__.bool>',
                                           'longname': 'usePressure',
                                           'numArgs': 1,
                                           'shortname': 'usp'}},
                 'removedFlags': {'is': 'inputSamples'},
                 'shortFlags': {'bd': 'brushDrag',
                                'bm': 'brushMass',
                                'dam': 'drawAsMesh',
                                'dbs': 'dragBrushSize',
                                'dp': 'doProject',
                                'dq': 'displayQuality',
                                'e': 'edit',
                                'ex': 'exists',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'inputSamples': 'inputSamples',
                                'pd': 'paintAtDepth',
                                'pm1': 'pressureMapping1',
                                'pm2': 'pressureMapping2',
                                'pm3': 'pressureMapping3',
                                'ps1': 'pressureMin1',
                                'ps2': 'pressureMin2',
                                'ps3': 'pressureMin3',
                                'px1': 'pressureMax1',
                                'px2': 'pressureMax2',
                                'px3': 'pressureMax3',
                                'pxm': 'pixelMode',
                                'q': 'query',
                                'sof': 'surfaceOffset',
                                'sp': 'sampleSeparation',
                                'ss': 'setSelection',
                                'usp': 'usePressure'},
                 'type': 'context'},
  'dynamicConstraintRemove': {'type': 'runtime'},
  'dynamicLoad': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'}},
                  'shortFlags': {'q': 'query'},
                  'type': 'system'},
  'editDisplayLayerGlobals': {'flags': {'baseId': {'args': '<type __builtin__.int>',
                                                   'longname': 'baseId',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'bi'},
                                        'currentDisplayLayer': {'args': 'PyNode',
                                                                'longname': 'currentDisplayLayer',
                                                                'modes': ['create',
                                                                          'query'],
                                                                'numArgs': 1,
                                                                'shortname': 'cdl'},
                                        'mergeType': {'args': '<type __builtin__.int>',
                                                      'longname': 'mergeType',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'mt'},
                                        'query': {'args': '<type __builtin__.bool>',
                                                  'longname': 'query',
                                                  'numArgs': 0,
                                                  'shortname': 'q'},
                                        'useCurrent': {'args': '<type __builtin__.bool>',
                                                       'longname': 'useCurrent',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'uc'}},
                              'shortFlags': {'bi': 'baseId',
                                             'cdl': 'currentDisplayLayer',
                                             'mt': 'mergeType',
                                             'q': 'query',
                                             'uc': 'useCurrent'},
                              'type': 'general'},
  'editDisplayLayerMembers': {'flags': {'fullNames': {'args': '<type __builtin__.bool>',
                                                      'longname': 'fullNames',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'fn'},
                                        'noRecurse': {'args': '<type __builtin__.bool>',
                                                      'longname': 'noRecurse',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 0,
                                                      'shortname': 'nr'},
                                        'query': {'args': '<type __builtin__.bool>',
                                                  'longname': 'query',
                                                  'numArgs': 0,
                                                  'shortname': 'q'}},
                              'shortFlags': {'fn': 'fullNames',
                                             'nr': 'noRecurse',
                                             'q': 'query'},
                              'type': 'general'},
  'editMetadata': {'flags': {'channelName': {'args': '<type __builtin__.str>',
                                             'longname': 'channelName',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'cn'},
                             'channelType': {'args': '<type __builtin__.str>',
                                             'longname': 'channelType',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'cht'},
                             'endIndex': {'args': '<type __builtin__.str>',
                                          'longname': 'endIndex',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'eix'},
                             'index': {'args': '<type __builtin__.str>',
                                       'longname': 'index',
                                       'modes': ['create',
                                                 'query',
                                                 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'idx'},
                             'indexType': {'args': '<type __builtin__.str>',
                                           'longname': 'indexType',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'idt'},
                             'memberName': {'args': '<type __builtin__.str>',
                                            'longname': 'memberName',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'mn'},
                             'remove': {'args': '<type __builtin__.bool>',
                                        'longname': 'remove',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'rem'},
                             'scene': {'args': '<type __builtin__.bool>',
                                       'longname': 'scene',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'scn'},
                             'startIndex': {'args': '<type __builtin__.str>',
                                            'longname': 'startIndex',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'six'},
                             'streamName': {'args': '<type __builtin__.str>',
                                            'longname': 'streamName',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'stn'},
                             'stringValue': {'args': '<type __builtin__.str>',
                                             'longname': 'stringValue',
                                             'modes': ['create',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'sv'},
                             'value': {'args': '<type __builtin__.float>',
                                       'longname': 'value',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'v'}},
                   'shortFlags': {'cht': 'channelType',
                                  'cn': 'channelName',
                                  'eix': 'endIndex',
                                  'idt': 'indexType',
                                  'idx': 'index',
                                  'mn': 'memberName',
                                  'rem': 'remove',
                                  'scn': 'scene',
                                  'six': 'startIndex',
                                  'stn': 'streamName',
                                  'sv': 'stringValue',
                                  'v': 'value'},
                   'type': 'modeling'},
  'editRenderLayerAdjustment': {'flags': {'attributeLog': {'args': '<type __builtin__.bool>',
                                                           'longname': 'attributeLog',
                                                           'modes': ['query'],
                                                           'numArgs': 0,
                                                           'shortname': 'alg'},
                                          'layer': {'args': 'PyNode',
                                                    'longname': 'layer',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'lyr'},
                                          'nodeLog': {'args': '<type __builtin__.bool>',
                                                      'longname': 'nodeLog',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'nlg'},
                                          'query': {'args': '<type __builtin__.bool>',
                                                    'longname': 'query',
                                                    'numArgs': 0,
                                                    'shortname': 'q'},
                                          'remove': {'args': '<type __builtin__.bool>',
                                                     'longname': 'remove',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'r'}},
                                'shortFlags': {'alg': 'attributeLog',
                                               'lyr': 'layer',
                                               'nlg': 'nodeLog',
                                               'q': 'query',
                                               'r': 'remove'},
                                'type': 'rendering'},
  'editRenderLayerGlobals': {'flags': {'baseId': {'args': '<type __builtin__.int>',
                                                  'longname': 'baseId',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'bi'},
                                       'currentRenderLayer': {'args': 'PyNode',
                                                              'longname': 'currentRenderLayer',
                                                              'modes': ['create',
                                                                        'query'],
                                                              'numArgs': 1,
                                                              'shortname': 'crl'},
                                       'enableAutoAdjustments': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'enableAutoAdjustments',
                                                                 'modes': ['create',
                                                                           'query'],
                                                                 'numArgs': 1,
                                                                 'shortname': 'eaa'},
                                       'mergeType': {'args': '<type __builtin__.int>',
                                                     'longname': 'mergeType',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'mt'},
                                       'query': {'args': '<type __builtin__.bool>',
                                                 'longname': 'query',
                                                 'numArgs': 0,
                                                 'shortname': 'q'},
                                       'useCurrent': {'args': '<type __builtin__.bool>',
                                                      'longname': 'useCurrent',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'uc'}},
                             'shortFlags': {'bi': 'baseId',
                                            'crl': 'currentRenderLayer',
                                            'eaa': 'enableAutoAdjustments',
                                            'mt': 'mergeType',
                                            'q': 'query',
                                            'uc': 'useCurrent'},
                             'type': 'rendering'},
  'editRenderLayerMembers': {'flags': {'fullNames': {'args': '<type __builtin__.bool>',
                                                     'longname': 'fullNames',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'fn'},
                                       'noRecurse': {'args': '<type __builtin__.bool>',
                                                     'longname': 'noRecurse',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'nr'},
                                       'query': {'args': '<type __builtin__.bool>',
                                                 'longname': 'query',
                                                 'numArgs': 0,
                                                 'shortname': 'q'},
                                       'remove': {'args': '<type __builtin__.bool>',
                                                  'longname': 'remove',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'r'}},
                             'shortFlags': {'fn': 'fullNames',
                                            'nr': 'noRecurse',
                                            'q': 'query',
                                            'r': 'remove'},
                             'type': 'rendering'},
  'editor': {'flags': {'control': {'args': '<type __builtin__.bool>',
                                   'longname': 'control',
                                   'modes': ['query'],
                                   'numArgs': 0,
                                   'shortname': 'ctl'},
                       'defineTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'defineTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'dt'},
                       'docTag': {'args': '<type __builtin__.str>',
                                  'longname': 'docTag',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'dtg'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'exists': {'args': '<type __builtin__.bool>',
                                  'longname': 'exists',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'ex'},
                       'filter': {'args': '<type __builtin__.str>',
                                  'longname': 'filter',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'f'},
                       'forceMainConnection': {'args': '<type __builtin__.str>',
                                               'longname': 'forceMainConnection',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'fmc'},
                       'highlightConnection': {'args': '<type __builtin__.str>',
                                               'longname': 'highlightConnection',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'hlc'},
                       'lockMainConnection': {'args': '<type __builtin__.bool>',
                                              'longname': 'lockMainConnection',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'lck'},
                       'mainListConnection': {'args': '<type __builtin__.str>',
                                              'longname': 'mainListConnection',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'mlc'},
                       'panel': {'args': '<type __builtin__.str>',
                                 'longname': 'panel',
                                 'modes': ['create', 'query'],
                                 'numArgs': 1,
                                 'shortname': 'pnl'},
                       'parent': {'args': '<type __builtin__.str>',
                                  'longname': 'parent',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'p'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'selectionConnection': {'args': '<type __builtin__.str>',
                                               'longname': 'selectionConnection',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'slc'},
                       'stateString': {'args': '<type __builtin__.bool>',
                                       'longname': 'stateString',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'sts'},
                       'unParent': {'args': '<type __builtin__.bool>',
                                    'longname': 'unParent',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'up'},
                       'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                'longname': 'unlockMainConnection',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'ulk'},
                       'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                'longname': 'updateMainConnection',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'upd'},
                       'useTemplate': {'args': '<type __builtin__.str>',
                                       'longname': 'useTemplate',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ut'}},
             'shortFlags': {'ctl': 'control',
                            'dt': 'defineTemplate',
                            'dtg': 'docTag',
                            'e': 'edit',
                            'ex': 'exists',
                            'f': 'filter',
                            'fmc': 'forceMainConnection',
                            'hlc': 'highlightConnection',
                            'lck': 'lockMainConnection',
                            'mlc': 'mainListConnection',
                            'p': 'parent',
                            'pnl': 'panel',
                            'q': 'query',
                            'slc': 'selectionConnection',
                            'sts': 'stateString',
                            'ulk': 'unlockMainConnection',
                            'up': 'unParent',
                            'upd': 'updateMainConnection',
                            'ut': 'useTemplate'},
             'type': 'windows'},
  'editorTemplate': {'flags': {'addAdskAssetControls': {'args': '<type __builtin__.bool>',
                                                        'longname': 'addAdskAssetControls',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'aac'},
                               'addComponents': {'args': '<type __builtin__.bool>',
                                                 'longname': 'addComponents',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'acp'},
                               'addControl': {'args': '<type __builtin__.bool>',
                                              'longname': 'addControl',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ac'},
                               'addDynamicControl': {'args': '<type __builtin__.bool>',
                                                     'longname': 'addDynamicControl',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'adc'},
                               'addExtraControls': {'args': '<type __builtin__.bool>',
                                                    'longname': 'addExtraControls',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'aec'},
                               'addSeparator': {'args': '<type __builtin__.bool>',
                                                'longname': 'addSeparator',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'addSeparator'},
                               'annotateFieldOnly': {'args': '<type __builtin__.bool>',
                                                     'longname': 'annotateFieldOnly',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'afo'},
                               'annotation': {'args': '<type __builtin__.str>',
                                              'longname': 'annotation',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ann'},
                               'beginLayout': {'args': '<type __builtin__.str>',
                                               'longname': 'beginLayout',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'bl'},
                               'beginNoOptimize': {'args': '<type __builtin__.bool>',
                                                   'longname': 'beginNoOptimize',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'bn'},
                               'beginScrollLayout': {'args': '<type __builtin__.bool>',
                                                     'longname': 'beginScrollLayout',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'bsl'},
                               'callCustom': {'args': '<type __builtin__.bool>',
                                              'longname': 'callCustom',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ccu'},
                               'collapse': {'args': '<type __builtin__.bool>',
                                            'longname': 'collapse',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'cl'},
                               'debugMode': {'args': '<type __builtin__.bool>',
                                             'longname': 'debugMode',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dbm'},
                               'dimControl': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.bool>'],
                                              'longname': 'dimControl',
                                              'modes': ['create'],
                                              'numArgs': 3,
                                              'shortname': 'dc'},
                               'endLayout': {'args': '<type __builtin__.bool>',
                                             'longname': 'endLayout',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'el'},
                               'endNoOptimize': {'args': '<type __builtin__.bool>',
                                                 'longname': 'endNoOptimize',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'en'},
                               'endScrollLayout': {'args': '<type __builtin__.bool>',
                                                   'longname': 'endScrollLayout',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'esl'},
                               'extraControlsLabel': {'args': '<type __builtin__.str>',
                                                      'longname': 'extraControlsLabel',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ecl'},
                               'interruptOptimize': {'args': '<type __builtin__.bool>',
                                                     'longname': 'interruptOptimize',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'io'},
                               'label': {'args': '<type __builtin__.str>',
                                         'longname': 'label',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'l'},
                               'listExtraAttributes': {'args': '<type __builtin__.str>',
                                                       'longname': 'listExtraAttributes',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'lea'},
                               'preventOverride': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preventOverride',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'po'},
                               'queryControl': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'queryControl',
                                                'modes': ['create'],
                                                'numArgs': 2,
                                                'shortname': 'qc'},
                               'queryLabel': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'queryLabel',
                                              'modes': ['create'],
                                              'numArgs': 2,
                                              'shortname': 'ql'},
                               'queryName': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'queryName',
                                             'modes': ['create'],
                                             'numArgs': 2,
                                             'shortname': 'qn'},
                               'removeControl': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'removeControl',
                                                 'modes': ['create'],
                                                 'numArgs': 2,
                                                 'shortname': 'rc'},
                               'suppress': {'args': '<type __builtin__.str>',
                                            'longname': 'suppress',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 's'}},
                     'removedFlags': {'as': 'addSeparator'},
                     'shortFlags': {'aac': 'addAdskAssetControls',
                                    'ac': 'addControl',
                                    'acp': 'addComponents',
                                    'adc': 'addDynamicControl',
                                    'addSeparator': 'addSeparator',
                                    'aec': 'addExtraControls',
                                    'afo': 'annotateFieldOnly',
                                    'ann': 'annotation',
                                    'bl': 'beginLayout',
                                    'bn': 'beginNoOptimize',
                                    'bsl': 'beginScrollLayout',
                                    'ccu': 'callCustom',
                                    'cl': 'collapse',
                                    'dbm': 'debugMode',
                                    'dc': 'dimControl',
                                    'ecl': 'extraControlsLabel',
                                    'el': 'endLayout',
                                    'en': 'endNoOptimize',
                                    'esl': 'endScrollLayout',
                                    'io': 'interruptOptimize',
                                    'l': 'label',
                                    'lea': 'listExtraAttributes',
                                    'po': 'preventOverride',
                                    'qc': 'queryControl',
                                    'ql': 'queryLabel',
                                    'qn': 'queryName',
                                    'rc': 'removeControl',
                                    's': 'suppress'},
                     'type': 'windows'},
  'effector': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'hide': {'args': '<type __builtin__.bool>',
                                  'longname': 'hide',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'hi'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'}},
               'shortFlags': {'e': 'edit',
                              'hi': 'hide',
                              'n': 'name',
                              'q': 'query'},
               'type': 'animation'},
  'emit': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                   'longname': 'attribute',
                                   'modes': ['create', 'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 'at'},
                     'floatValue': {'args': '<type __builtin__.float>',
                                    'longname': 'floatValue',
                                    'modes': ['create', 'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'fv'},
                     'object': {'args': '<type __builtin__.str>',
                                'longname': 'object',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'o'},
                     'position': {'args': ['<type __builtin__.float>',
                                           '<type __builtin__.float>',
                                           '<type __builtin__.float>'],
                                  'longname': 'position',
                                  'modes': ['create', 'multiuse'],
                                  'numArgs': 3,
                                  'shortname': 'pos'},
                     'vectorValue': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'vectorValue',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 3,
                                     'shortname': 'vv'}},
           'shortFlags': {'at': 'attribute',
                          'fv': 'floatValue',
                          'o': 'object',
                          'pos': 'position',
                          'vv': 'vectorValue'},
           'type': 'effects'},
  'emitter': {'flags': {'alongAxis': {'args': '<type __builtin__.float>',
                                      'longname': 'alongAxis',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'alx'},
                        'aroundAxis': {'args': '<type __builtin__.float>',
                                       'longname': 'aroundAxis',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'arx'},
                        'awayFromAxis': {'args': '<type __builtin__.float>',
                                         'longname': 'awayFromAxis',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'afx'},
                        'awayFromCenter': {'args': '<type __builtin__.float>',
                                           'longname': 'awayFromCenter',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'afc'},
                        'cycleEmission': {'args': '<type __builtin__.str>',
                                          'longname': 'cycleEmission',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cye'},
                        'cycleInterval': {'args': '<type __builtin__.int>',
                                          'longname': 'cycleInterval',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cyi'},
                        'directionX': {'args': '<type __builtin__.float>',
                                       'longname': 'directionX',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dx'},
                        'directionY': {'args': '<type __builtin__.float>',
                                       'longname': 'directionY',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dy'},
                        'directionZ': {'args': '<type __builtin__.float>',
                                       'longname': 'directionZ',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dz'},
                        'directionalSpeed': {'args': '<type __builtin__.float>',
                                             'longname': 'directionalSpeed',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'drs'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'maxDistance': {'args': '<type __builtin__.float>',
                                        'longname': 'maxDistance',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'mxd'},
                        'minDistance': {'args': '<type __builtin__.float>',
                                        'longname': 'minDistance',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'mnd'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'needParentUV': {'args': '<type __builtin__.bool>',
                                         'longname': 'needParentUV',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nuv'},
                        'normalSpeed': {'args': '<type __builtin__.float>',
                                        'longname': 'normalSpeed',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'nsp'},
                        'position': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'position',
                                     'modes': ['create',
                                               'query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 3,
                                     'shortname': 'pos'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'randomDirection': {'args': '<type __builtin__.float>',
                                            'longname': 'randomDirection',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'rnd'},
                        'rate': {'args': '<type __builtin__.float>',
                                 'longname': 'rate',
                                 'modes': ['query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'r'},
                        'scaleRateByObjectSize': {'args': '<type __builtin__.bool>',
                                                  'longname': 'scaleRateByObjectSize',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sro'},
                        'scaleSpeedBySize': {'args': '<type __builtin__.bool>',
                                             'longname': 'scaleSpeedBySize',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ssz'},
                        'speed': {'args': '<type __builtin__.float>',
                                  'longname': 'speed',
                                  'modes': ['query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'spd'},
                        'speedRandom': {'args': '<type __builtin__.float>',
                                        'longname': 'speedRandom',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'srn'},
                        'spread': {'args': '<type __builtin__.float>',
                                   'longname': 'spread',
                                   'modes': ['query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'sp'},
                        'tangentSpeed': {'args': '<type __builtin__.float>',
                                         'longname': 'tangentSpeed',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'tsp'},
                        'torusSectionRadius': {'args': '<type __builtin__.float>',
                                               'longname': 'torusSectionRadius',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tsr'},
                        'type': {'args': '<type __builtin__.str>',
                                 'longname': 'type',
                                 'modes': ['query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'typ'},
                        'volumeOffset': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'volumeOffset',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 3,
                                         'shortname': 'vof'},
                        'volumeShape': {'args': '<type __builtin__.str>',
                                        'longname': 'volumeShape',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vsh'},
                        'volumeSweep': {'args': '<type __builtin__.float>',
                                        'longname': 'volumeSweep',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vsw'}},
              'resultNeedsUnpacking': True,
              'shortFlags': {'afc': 'awayFromCenter',
                             'afx': 'awayFromAxis',
                             'alx': 'alongAxis',
                             'arx': 'aroundAxis',
                             'cye': 'cycleEmission',
                             'cyi': 'cycleInterval',
                             'drs': 'directionalSpeed',
                             'dx': 'directionX',
                             'dy': 'directionY',
                             'dz': 'directionZ',
                             'e': 'edit',
                             'mnd': 'minDistance',
                             'mxd': 'maxDistance',
                             'n': 'name',
                             'nsp': 'normalSpeed',
                             'nuv': 'needParentUV',
                             'pos': 'position',
                             'q': 'query',
                             'r': 'rate',
                             'rnd': 'randomDirection',
                             'sp': 'spread',
                             'spd': 'speed',
                             'srn': 'speedRandom',
                             'sro': 'scaleRateByObjectSize',
                             'ssz': 'scaleSpeedBySize',
                             'tsp': 'tangentSpeed',
                             'tsr': 'torusSectionRadius',
                             'typ': 'type',
                             'vof': 'volumeOffset',
                             'vsh': 'volumeShape',
                             'vsw': 'volumeSweep'},
              'type': 'effects'},
  'enableDevice': {'flags': {'apply': {'args': '<type __builtin__.bool>',
                                       'longname': 'apply',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'a'},
                             'device': {'args': '<type __builtin__.str>',
                                        'longname': 'device',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'd'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'monitor': {'args': '<type __builtin__.bool>',
                                         'longname': 'monitor',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                             'record': {'args': '<type __builtin__.bool>',
                                        'longname': 'record',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'rec'}},
                   'shortFlags': {},
                   'type': 'animation'},
  'encodeString': {'flags': {}, 'shortFlags': {}, 'type': 'language'},
  'error': {'flags': {'noContext': {'args': '<type __builtin__.bool>',
                                    'longname': 'noContext',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'n'},
                      'showLineNumber': {'args': '<type __builtin__.bool>',
                                         'longname': 'showLineNumber',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'sl'}},
            'shortFlags': {'n': 'noContext', 'sl': 'showLineNumber'},
            'type': 'system'},
  'eval': {'flags': {'undoEnabled': {'args': '<type __builtin__.bool>',
                                     'longname': 'undoEnabled',
                                     'numArgs': 1,
                                     'shortname': 'ue'}},
           'shortFlags': {'ue': 'undoEnabled'},
           'type': None},
  'evalContinue': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'evalDeferred': {'flags': {'evaluateNext': {'args': '<type __builtin__.bool>',
                                              'longname': 'evaluateNext',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'en'},
                             'list': {'args': '<type __builtin__.bool>',
                                      'longname': 'list',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ls'},
                             'lowPriority': {'args': '<type __builtin__.bool>',
                                             'longname': 'lowPriority',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'low'},
                             'lowestPriority': {'args': '<type __builtin__.bool>',
                                                'longname': 'lowestPriority',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'lp'}},
                   'shortFlags': {'en': 'evaluateNext',
                                  'low': 'lowPriority',
                                  'lp': 'lowestPriority',
                                  'ls': 'list'},
                   'type': 'language'},
  'evalEcho': {'flags': {}, 'shortFlags': {}, 'type': 'language'},
  'evalNoSelectNotify': {'flags': {}, 'shortFlags': {}, 'type': 'language'},
  'evaluationManager': {'flags': {'cycleCluster': {'args': '<type __builtin__.str>',
                                                   'longname': 'cycleCluster',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ccl'},
                                  'disableInfo': {'args': '<type __builtin__.str>',
                                                  'longname': 'disableInfo',
                                                  'modes': ['query'],
                                                  'numArgs': 1,
                                                  'shortname': 'di'},
                                  'downstreamFrom': {'args': '<type __builtin__.str>',
                                                     'longname': 'downstreamFrom',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'dst'},
                                  'empty': {'args': '<type __builtin__.bool>',
                                            'longname': 'empty',
                                            'numArgs': 0,
                                            'shortname': 'mt'},
                                  'enabled': {'args': '<type __builtin__.bool>',
                                              'longname': 'enabled',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                  'idleAction': {'args': '<type __builtin__.int>',
                                                 'longname': 'idleAction',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ia'},
                                  'idleBuild': {'args': '<type __builtin__.bool>',
                                                'longname': 'idleBuild',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'ib'},
                                  'invalidate': {'args': '<type __builtin__.bool>',
                                                 'longname': 'invalidate',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'inv'},
                                  'manipulation': {'args': '<type __builtin__.bool>',
                                                   'longname': 'manipulation',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'man'},
                                  'mode': {'args': '<type __builtin__.str>',
                                           'longname': 'mode',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'm'},
                                  'nodeTypeGloballySerialize': {'args': '<type __builtin__.bool>',
                                                                'longname': 'nodeTypeGloballySerialize',
                                                                'modes': ['create',
                                                                          'query'],
                                                                'numArgs': 1,
                                                                'shortname': 'ntg'},
                                  'nodeTypeParallel': {'args': '<type __builtin__.bool>',
                                                       'longname': 'nodeTypeParallel',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'ntp'},
                                  'nodeTypeSerialize': {'args': '<type __builtin__.bool>',
                                                        'longname': 'nodeTypeSerialize',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'nts'},
                                  'nodeTypeUntrusted': {'args': '<type __builtin__.bool>',
                                                        'longname': 'nodeTypeUntrusted',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'ntu'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'safeMode': {'args': '<type __builtin__.bool>',
                                               'longname': 'safeMode',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'sfm'},
                                  'upstreamFrom': {'args': '<type __builtin__.str>',
                                                   'longname': 'upstreamFrom',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ust'}},
                        'shortFlags': {'ccl': 'cycleCluster',
                                       'di': 'disableInfo',
                                       'dst': 'downstreamFrom',
                                       'e': 'enabled',
                                       'ia': 'idleAction',
                                       'ib': 'idleBuild',
                                       'inv': 'invalidate',
                                       'm': 'mode',
                                       'man': 'manipulation',
                                       'mt': 'empty',
                                       'ntg': 'nodeTypeGloballySerialize',
                                       'ntp': 'nodeTypeParallel',
                                       'nts': 'nodeTypeSerialize',
                                       'ntu': 'nodeTypeUntrusted',
                                       'q': 'query',
                                       'sfm': 'safeMode',
                                       'ust': 'upstreamFrom'},
                        'type': 'animation'},
  'evaluator': {'flags': {'clusters': {'args': '<type __builtin__.bool>',
                                       'longname': 'clusters',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'cl'},
                          'configuration': {'args': '<type __builtin__.str>',
                                            'longname': 'configuration',
                                            'modes': ['create',
                                                      'query',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'c'},
                          'enable': {'args': '<type __builtin__.bool>',
                                     'longname': 'enable',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'en'},
                          'info': {'args': '<type __builtin__.bool>',
                                   'longname': 'info',
                                   'modes': ['query'],
                                   'numArgs': 1,
                                   'shortname': 'i'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nodeType': {'args': '<type __builtin__.str>',
                                       'longname': 'nodeType',
                                       'modes': ['create',
                                                 'query',
                                                 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'nt'},
                          'nodeTypeChildren': {'args': '<type __builtin__.bool>',
                                               'longname': 'nodeTypeChildren',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 0,
                                               'shortname': 'ntc'},
                          'priority': {'args': '<type __builtin__.int>',
                                       'longname': 'priority',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'valueName': {'args': '<type __builtin__.str>',
                                        'longname': 'valueName',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'vn'}},
                'shortFlags': {'c': 'configuration',
                               'cl': 'clusters',
                               'en': 'enable',
                               'i': 'info',
                               'n': 'name',
                               'nt': 'nodeType',
                               'ntc': 'nodeTypeChildren',
                               'p': 'priority',
                               'q': 'query',
                               'vn': 'valueName'},
                'type': 'animation'},
  'event': {'callbackFlags': ['proc', 'pr'],
            'flags': {'count': {'args': '<type __builtin__.int>',
                                'longname': 'count',
                                'modes': ['query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'ct'},
                      'delete': {'args': '<type __builtin__.bool>',
                                 'longname': 'delete',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'd'},
                      'dieAtCollision': {'args': '<type __builtin__.bool>',
                                         'longname': 'dieAtCollision',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'die'},
                      'edit': {'args': '<type __builtin__.bool>',
                               'longname': 'edit',
                               'numArgs': 0,
                               'shortname': 'e'},
                      'emit': {'args': '<type __builtin__.int>',
                               'longname': 'emit',
                               'modes': ['query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'em'},
                      'idNumber': {'args': '<type __builtin__.int>',
                                   'longname': 'idNumber',
                                   'numArgs': 1,
                                   'shortname': 'id'},
                      'list': {'args': '<type __builtin__.bool>',
                               'longname': 'list',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'ls'},
                      'name': {'args': '<type __builtin__.str>',
                               'longname': 'name',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'n'},
                      'proc': {'args': 'script',
                               'longname': 'proc',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'pr'},
                      'query': {'args': '<type __builtin__.bool>',
                                'longname': 'query',
                                'numArgs': 0,
                                'shortname': 'q'},
                      'random': {'args': '<type __builtin__.bool>',
                                 'longname': 'random',
                                 'modes': ['query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'r'},
                      'rename': {'args': '<type __builtin__.str>',
                                 'longname': 'rename',
                                 'modes': ['query'],
                                 'numArgs': 1,
                                 'shortname': 're'},
                      'select': {'args': '<type __builtin__.bool>',
                                 'longname': 'select',
                                 'modes': [],
                                 'numArgs': 0,
                                 'shortname': 's'},
                      'split': {'args': '<type __builtin__.int>',
                                'longname': 'split',
                                'modes': ['query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'spl'},
                      'spread': {'args': '<type __builtin__.float>',
                                 'longname': 'spread',
                                 'modes': ['query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'sp'},
                      'target': {'args': '<type __builtin__.str>',
                                 'longname': 'target',
                                 'modes': ['query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 't'}},
            'shortFlags': {'ct': 'count',
                           'd': 'delete',
                           'die': 'dieAtCollision',
                           'e': 'edit',
                           'em': 'emit',
                           'id': 'idNumber',
                           'ls': 'list',
                           'n': 'name',
                           'pr': 'proc',
                           'q': 'query',
                           'r': 'random',
                           're': 'rename',
                           's': 'select',
                           'sp': 'spread',
                           'spl': 'split',
                           't': 'target'},
            'type': 'effects'},
  'exactWorldBoundingBox': {'flags': {'calculateExactly': {'args': '<type __builtin__.bool>',
                                                           'longname': 'calculateExactly',
                                                           'modes': ['create'],
                                                           'numArgs': 1,
                                                           'shortname': 'ce'},
                                      'ignoreInvisible': {'args': '<type __builtin__.bool>',
                                                          'longname': 'ignoreInvisible',
                                                          'modes': ['create'],
                                                          'numArgs': 1,
                                                          'shortname': 'ii'}},
                            'shortFlags': {'ce': 'calculateExactly',
                                           'ii': 'ignoreInvisible'},
                            'type': 'general'},
  'exclusiveLightCheckBox': {'callbackFlags': ['visibleChangeCommand',
                                               'vcc',
                                               'dragCallback',
                                               'dgc',
                                               'dropCallback',
                                               'dpc'],
                             'flags': {'annotation': {'args': '<type __builtin__.str>',
                                                      'longname': 'annotation',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ann'},
                                       'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                    '<type __builtin__.float>',
                                                                    '<type __builtin__.float>'],
                                                           'longname': 'backgroundColor',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 3,
                                                           'shortname': 'bgc'},
                                       'defineTemplate': {'args': '<type __builtin__.str>',
                                                          'longname': 'defineTemplate',
                                                          'modes': ['create'],
                                                          'numArgs': 1,
                                                          'shortname': 'dt'},
                                       'docTag': {'args': '<type __builtin__.str>',
                                                  'longname': 'docTag',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dtg'},
                                       'dragCallback': {'args': 'script',
                                                        'longname': 'dragCallback',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'dgc'},
                                       'dropCallback': {'args': 'script',
                                                        'longname': 'dropCallback',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'dpc'},
                                       'edit': {'args': '<type __builtin__.bool>',
                                                'longname': 'edit',
                                                'numArgs': 0,
                                                'shortname': 'e'},
                                       'enable': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enable',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'en'},
                                       'enableBackground': {'args': '<type __builtin__.bool>',
                                                            'longname': 'enableBackground',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ebg'},
                                       'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                               'longname': 'enableKeyboardFocus',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'ekf'},
                                       'exists': {'args': '<type __builtin__.bool>',
                                                  'longname': 'exists',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'ex'},
                                       'fullPathName': {'args': '<type __builtin__.bool>',
                                                        'longname': 'fullPathName',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'fpn'},
                                       'height': {'args': '<type __builtin__.int>',
                                                  'longname': 'height',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'h'},
                                       'highlightColor': {'args': ['<type __builtin__.float>',
                                                                   '<type __builtin__.float>',
                                                                   '<type __builtin__.float>'],
                                                          'longname': 'highlightColor',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 3,
                                                          'shortname': 'hlc'},
                                       'isObscured': {'args': '<type __builtin__.bool>',
                                                      'longname': 'isObscured',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'io'},
                                       'label': {'args': '<type __builtin__.str>',
                                                 'longname': 'label',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'l'},
                                       'light': {'args': 'PyNode',
                                                 'longname': 'light',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'lt'},
                                       'manage': {'args': '<type __builtin__.bool>',
                                                  'longname': 'manage',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'm'},
                                       'noBackground': {'args': '<type __builtin__.bool>',
                                                        'longname': 'noBackground',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'nbg'},
                                       'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                              'longname': 'numberOfPopupMenus',
                                                              'modes': ['query'],
                                                              'numArgs': 0,
                                                              'shortname': 'npm'},
                                       'parent': {'args': '<type __builtin__.str>',
                                                  'longname': 'parent',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'p'},
                                       'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                          'longname': 'popupMenuArray',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'pma'},
                                       'preventOverride': {'args': '<type __builtin__.bool>',
                                                           'longname': 'preventOverride',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'po'},
                                       'query': {'args': '<type __builtin__.bool>',
                                                 'longname': 'query',
                                                 'numArgs': 0,
                                                 'shortname': 'q'},
                                       'statusBarMessage': {'args': '<type __builtin__.str>',
                                                            'longname': 'statusBarMessage',
                                                            'modes': ['create',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'sbm'},
                                       'useTemplate': {'args': '<type __builtin__.str>',
                                                       'longname': 'useTemplate',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ut'},
                                       'visible': {'args': '<type __builtin__.bool>',
                                                   'longname': 'visible',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'vis'},
                                       'visibleChangeCommand': {'args': 'script',
                                                                'longname': 'visibleChangeCommand',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'vcc'},
                                       'width': {'args': '<type __builtin__.int>',
                                                 'longname': 'width',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'w'}},
                             'shortFlags': {'ann': 'annotation',
                                            'bgc': 'backgroundColor',
                                            'dgc': 'dragCallback',
                                            'dpc': 'dropCallback',
                                            'dt': 'defineTemplate',
                                            'dtg': 'docTag',
                                            'e': 'edit',
                                            'ebg': 'enableBackground',
                                            'ekf': 'enableKeyboardFocus',
                                            'en': 'enable',
                                            'ex': 'exists',
                                            'fpn': 'fullPathName',
                                            'h': 'height',
                                            'hlc': 'highlightColor',
                                            'io': 'isObscured',
                                            'l': 'label',
                                            'lt': 'light',
                                            'm': 'manage',
                                            'nbg': 'noBackground',
                                            'npm': 'numberOfPopupMenus',
                                            'p': 'parent',
                                            'pma': 'popupMenuArray',
                                            'po': 'preventOverride',
                                            'q': 'query',
                                            'sbm': 'statusBarMessage',
                                            'ut': 'useTemplate',
                                            'vcc': 'visibleChangeCommand',
                                            'vis': 'visible',
                                            'w': 'width'},
                             'type': 'rendering'},
  'expandedSelection': {'flags': {'depth': {'args': '<type __builtin__.int>',
                                            'longname': 'depth',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'd'},
                                  'expansionType': {'args': '<type __builtin__.str>',
                                                    'longname': 'expansionType',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'et'}},
                        'shortFlags': {'d': 'depth', 'et': 'expansionType'},
                        'type': 'general'},
  'exportEdits': {'callbackFlags': ['editCommand', 'ec'],
                  'flags': {'editCommand': {'args': '<type __builtin__.str>',
                                            'longname': 'editCommand',
                                            'modes': ['create',
                                                      'query',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'ec'},
                            'excludeHierarchy': {'args': '<type __builtin__.bool>',
                                                 'longname': 'excludeHierarchy',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 0,
                                                 'shortname': 'ehr'},
                            'excludeNode': {'args': '<type __builtin__.str>',
                                            'longname': 'excludeNode',
                                            'modes': ['create',
                                                      'query',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'en'},
                            'exportSelected': {'args': '<type __builtin__.bool>',
                                               'longname': 'exportSelected',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 0,
                                               'shortname': 'exs'},
                            'force': {'args': '<type __builtin__.bool>',
                                      'longname': 'force',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'f'},
                            'includeAnimation': {'args': '<type __builtin__.bool>',
                                                 'longname': 'includeAnimation',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 0,
                                                 'shortname': 'ian'},
                            'includeConstraints': {'args': '<type __builtin__.bool>',
                                                   'longname': 'includeConstraints',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 0,
                                                   'shortname': 'ic'},
                            'includeDeformers': {'args': '<type __builtin__.bool>',
                                                 'longname': 'includeDeformers',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 0,
                                                 'shortname': 'idf'},
                            'includeNetwork': {'args': '<type __builtin__.bool>',
                                               'longname': 'includeNetwork',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 0,
                                               'shortname': 'inw'},
                            'includeNode': {'args': '<type __builtin__.str>',
                                            'longname': 'includeNode',
                                            'modes': ['create',
                                                      'query',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'includeNode'},
                            'includeSetAttrs': {'args': '<type __builtin__.bool>',
                                                'longname': 'includeSetAttrs',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'isa'},
                            'includeSetDrivenKeys': {'args': '<type __builtin__.bool>',
                                                     'longname': 'includeSetDrivenKeys',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 0,
                                                     'shortname': 'sdk'},
                            'includeShaders': {'args': '<type __builtin__.bool>',
                                               'longname': 'includeShaders',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 0,
                                               'shortname': 'ish'},
                            'onReferenceNode': {'args': '<type __builtin__.str>',
                                                'longname': 'onReferenceNode',
                                                'modes': ['create',
                                                          'query',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'orn'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'selected': {'args': '<type __builtin__.bool>',
                                         'longname': 'selected',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'sel'},
                            'type': {'args': '<type __builtin__.str>',
                                     'longname': 'type',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'typ'}},
                  'removedFlags': {'in': 'includeNode'},
                  'shortFlags': {'ec': 'editCommand',
                                 'ehr': 'excludeHierarchy',
                                 'en': 'excludeNode',
                                 'exs': 'exportSelected',
                                 'f': 'force',
                                 'ian': 'includeAnimation',
                                 'ic': 'includeConstraints',
                                 'idf': 'includeDeformers',
                                 'includeNode': 'includeNode',
                                 'inw': 'includeNetwork',
                                 'isa': 'includeSetAttrs',
                                 'ish': 'includeShaders',
                                 'orn': 'onReferenceNode',
                                 'q': 'query',
                                 'sdk': 'includeSetDrivenKeys',
                                 'sel': 'selected',
                                 'typ': 'type'},
                  'type': 'system'},
  'expression': {'flags': {'alwaysEvaluate': {'args': '<type __builtin__.int>',
                                              'longname': 'alwaysEvaluate',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ae'},
                           'animated': {'args': '<type __builtin__.int>',
                                        'longname': 'animated',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'an'},
                           'attribute': {'args': '<type __builtin__.str>',
                                         'longname': 'attribute',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'a'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'object': {'args': '<type __builtin__.str>',
                                      'longname': 'object',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'safe': {'args': '<type __builtin__.bool>',
                                    'longname': 'safe',
                                    'modes': ['query'],
                                    'numArgs': 0,
                                    'shortname': 'sf'},
                           'shortNames': {'args': '<type __builtin__.bool>',
                                          'longname': 'shortNames',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'sn'},
                           'string': {'args': '<type __builtin__.str>',
                                      'longname': 'string',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                           'timeDependent': {'args': '<type __builtin__.bool>',
                                             'longname': 'timeDependent',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'td'},
                           'unitConversion': {'args': '<type __builtin__.str>',
                                              'longname': 'unitConversion',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'uc'}},
                 'shortFlags': {'a': 'attribute',
                                'ae': 'alwaysEvaluate',
                                'an': 'animated',
                                'e': 'edit',
                                'n': 'name',
                                'o': 'object',
                                'q': 'query',
                                's': 'string',
                                'sf': 'safe',
                                'sn': 'shortNames',
                                'td': 'timeDependent',
                                'uc': 'unitConversion'},
                 'type': 'effects'},
  'expressionEditorListen': {'flags': {'listenFile': {'args': '<type __builtin__.str>',
                                                      'longname': 'listenFile',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'lf'},
                                       'listenForAttr': {'args': '<type __builtin__.str>',
                                                         'longname': 'listenForAttr',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'la'},
                                       'listenForExpression': {'args': '<type __builtin__.str>',
                                                               'longname': 'listenForExpression',
                                                               'modes': ['create'],
                                                               'numArgs': 1,
                                                               'shortname': 'le'},
                                       'listenForName': {'args': '<type __builtin__.str>',
                                                         'longname': 'listenForName',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'ln'},
                                       'stopListenForAttr': {'args': '<type __builtin__.str>',
                                                             'longname': 'stopListenForAttr',
                                                             'modes': ['create'],
                                                             'numArgs': 1,
                                                             'shortname': 'sla'},
                                       'stopListenForExpression': {'args': '<type __builtin__.str>',
                                                                   'longname': 'stopListenForExpression',
                                                                   'modes': ['create'],
                                                                   'numArgs': 1,
                                                                   'shortname': 'sle'},
                                       'stopListenForName': {'args': '<type __builtin__.str>',
                                                             'longname': 'stopListenForName',
                                                             'modes': ['create'],
                                                             'numArgs': 1,
                                                             'shortname': 'sln'}},
                             'shortFlags': {'la': 'listenForAttr',
                                            'le': 'listenForExpression',
                                            'lf': 'listenFile',
                                            'ln': 'listenForName',
                                            'sla': 'stopListenForAttr',
                                            'sle': 'stopListenForExpression',
                                            'sln': 'stopListenForName'},
                             'type': 'effects'},
  'extendCurve': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'curveOnSurface': {'args': '<type __builtin__.bool>',
                                               'longname': 'curveOnSurface',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'cos'},
                            'distance': {'args': '<type __builtin__.float>',
                                         'longname': 'distance',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'd'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'extendMethod': {'args': '<type __builtin__.int>',
                                             'longname': 'extendMethod',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'em'},
                            'extensionType': {'args': '<type __builtin__.int>',
                                              'longname': 'extensionType',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'et'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'inputPoint': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'inputPoint',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'ip'},
                            'join': {'args': '<type __builtin__.bool>',
                                     'longname': 'join',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'jn'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'noChanges': {'args': '<type __builtin__.bool>',
                                          'longname': 'noChanges',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nc'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'pointX': {'args': '<type __builtin__.float>',
                                       'longname': 'pointX',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'px'},
                            'pointY': {'args': '<type __builtin__.float>',
                                       'longname': 'pointY',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'py'},
                            'pointZ': {'args': '<type __builtin__.float>',
                                       'longname': 'pointZ',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'pz'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'range': {'args': '<type __builtin__.bool>',
                                      'longname': 'range',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'rn'},
                            'removeMultipleKnots': {'args': '<type __builtin__.bool>',
                                                    'longname': 'removeMultipleKnots',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'rmk'},
                            'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                'longname': 'replaceOriginal',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'rpo'},
                            'start': {'args': '<type __builtin__.int>',
                                      'longname': 'start',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 's'}},
                  'shortFlags': {'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'cos': 'curveOnSurface',
                                 'd': 'distance',
                                 'e': 'edit',
                                 'em': 'extendMethod',
                                 'et': 'extensionType',
                                 'fzn': 'frozen',
                                 'ip': 'inputPoint',
                                 'jn': 'join',
                                 'n': 'name',
                                 'nc': 'noChanges',
                                 'nds': 'nodeState',
                                 'o': 'object',
                                 'px': 'pointX',
                                 'py': 'pointY',
                                 'pz': 'pointZ',
                                 'q': 'query',
                                 'rmk': 'removeMultipleKnots',
                                 'rn': 'range',
                                 'rpo': 'replaceOriginal',
                                 's': 'start'},
                  'type': 'modeling'},
  'extendFluid': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'endD': {'args': '<type __builtin__.int>',
                                     'longname': 'endD',
                                     'numArgs': 1,
                                     'shortname': 'ed'},
                            'endH': {'args': '<type __builtin__.int>',
                                     'longname': 'endH',
                                     'numArgs': 1,
                                     'shortname': 'eh'},
                            'endW': {'args': '<type __builtin__.int>',
                                     'longname': 'endW',
                                     'numArgs': 1,
                                     'shortname': 'ew'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'startD': {'args': '<type __builtin__.int>',
                                       'longname': 'startD',
                                       'numArgs': 1,
                                       'shortname': 'sd'},
                            'startH': {'args': '<type __builtin__.int>',
                                       'longname': 'startH',
                                       'numArgs': 1,
                                       'shortname': 'sh'},
                            'startW': {'args': '<type __builtin__.int>',
                                       'longname': 'startW',
                                       'numArgs': 1,
                                       'shortname': 'sw'}},
                  'shortFlags': {'e': 'edit',
                                 'ed': 'endD',
                                 'eh': 'endH',
                                 'ew': 'endW',
                                 'q': 'query',
                                 'sd': 'startD',
                                 'sh': 'startH',
                                 'sw': 'startW'},
                  'type': 'other'},
  'extendSurface': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cch'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'distance': {'args': '<type __builtin__.float>',
                                           'longname': 'distance',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'd'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'extendDirection': {'args': '<type __builtin__.int>',
                                                  'longname': 'extendDirection',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ed'},
                              'extendMethod': {'args': '<type __builtin__.int>',
                                               'longname': 'extendMethod',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'em'},
                              'extendSide': {'args': '<type __builtin__.int>',
                                             'longname': 'extendSide',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'es'},
                              'extensionType': {'args': '<type __builtin__.int>',
                                                'longname': 'extensionType',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'et'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'join': {'args': '<type __builtin__.bool>',
                                       'longname': 'join',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'jn'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'noChanges': {'args': '<type __builtin__.bool>',
                                            'longname': 'noChanges',
                                            'numArgs': 1,
                                            'shortname': 'nc'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nds'},
                              'object': {'args': '<type __builtin__.bool>',
                                         'longname': 'object',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'o'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                  'longname': 'replaceOriginal',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'rpo'}},
                    'shortFlags': {'cch': 'caching',
                                   'ch': 'constructionHistory',
                                   'd': 'distance',
                                   'e': 'edit',
                                   'ed': 'extendDirection',
                                   'em': 'extendMethod',
                                   'es': 'extendSide',
                                   'et': 'extensionType',
                                   'fzn': 'frozen',
                                   'jn': 'join',
                                   'n': 'name',
                                   'nc': 'noChanges',
                                   'nds': 'nodeState',
                                   'o': 'object',
                                   'q': 'query',
                                   'rpo': 'replaceOriginal'},
                    'type': 'modeling'},
  'extrude': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                    'longname': 'caching',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'cch'},
                        'constructionHistory': {'args': '<type __builtin__.bool>',
                                                'longname': 'constructionHistory',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                        'degreeAlongLength': {'args': '<type __builtin__.int>',
                                              'longname': 'degreeAlongLength',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dl'},
                        'direction': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'direction',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 3,
                                      'shortname': 'd'},
                        'directionX': {'args': '<type __builtin__.float>',
                                       'longname': 'directionX',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dx'},
                        'directionY': {'args': '<type __builtin__.float>',
                                       'longname': 'directionY',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dy'},
                        'directionZ': {'args': '<type __builtin__.float>',
                                       'longname': 'directionZ',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dz'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'extrudeType': {'args': '<type __builtin__.int>',
                                        'longname': 'extrudeType',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'et'},
                        'fixedPath': {'args': '<type __builtin__.bool>',
                                      'longname': 'fixedPath',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'fpt'},
                        'frozen': {'args': '<type __builtin__.bool>',
                                   'longname': 'frozen',
                                   'numArgs': 1,
                                   'shortname': 'fzn'},
                        'length': {'args': '<type __builtin__.float>',
                                   'longname': 'length',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'l'},
                        'mergeItems': {'args': '<type __builtin__.bool>',
                                       'longname': 'mergeItems',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'mi'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'nodeState': {'args': '<type __builtin__.int>',
                                      'longname': 'nodeState',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'nds'},
                        'object': {'args': '<type __builtin__.bool>',
                                   'longname': 'object',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'o'},
                        'pivot': {'args': ['<type __builtin__.float>',
                                           '<type __builtin__.float>',
                                           '<type __builtin__.float>'],
                                  'longname': 'pivot',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 3,
                                  'shortname': 'p'},
                        'polygon': {'args': '<type __builtin__.int>',
                                    'longname': 'polygon',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'po'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'range': {'args': '<type __builtin__.bool>',
                                  'longname': 'range',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'rn'},
                        'rebuild': {'args': '<type __builtin__.bool>',
                                    'longname': 'rebuild',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'rb'},
                        'reverseSurfaceIfPathReversed': {'args': '<type __builtin__.bool>',
                                                         'longname': 'reverseSurfaceIfPathReversed',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'rsp'},
                        'rotation': {'args': '<type __builtin__.float>',
                                     'longname': 'rotation',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'ro'},
                        'scale': {'args': '<type __builtin__.float>',
                                  'longname': 'scale',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'sc'},
                        'subCurveSubSurface': {'args': '<type __builtin__.bool>',
                                               'longname': 'subCurveSubSurface',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'scs'},
                        'useComponentPivot': {'args': '<type __builtin__.int>',
                                              'longname': 'useComponentPivot',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ucp'},
                        'useProfileNormal': {'args': '<type __builtin__.bool>',
                                             'longname': 'useProfileNormal',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'upn'}},
              'shortFlags': {'cch': 'caching',
                             'ch': 'constructionHistory',
                             'd': 'direction',
                             'dl': 'degreeAlongLength',
                             'dx': 'directionX',
                             'dy': 'directionY',
                             'dz': 'directionZ',
                             'e': 'edit',
                             'et': 'extrudeType',
                             'fpt': 'fixedPath',
                             'fzn': 'frozen',
                             'l': 'length',
                             'mi': 'mergeItems',
                             'n': 'name',
                             'nds': 'nodeState',
                             'o': 'object',
                             'p': 'pivot',
                             'po': 'polygon',
                             'q': 'query',
                             'rb': 'rebuild',
                             'rn': 'range',
                             'ro': 'rotation',
                             'rsp': 'reverseSurfaceIfPathReversed',
                             'sc': 'scale',
                             'scs': 'subCurveSubSurface',
                             'ucp': 'useComponentPivot',
                             'upn': 'useProfileNormal'},
              'type': 'modeling'},
  'falloffCurve': {'callbackFlags': ['dropCallback',
                                     'dpc',
                                     'visibleChangeCommand',
                                     'vcc',
                                     'dragCallback',
                                     'dgc',
                                     'changeCommand',
                                     'cc'],
                   'flags': {'addControlVertex': {'args': '<type __builtin__.str>',
                                                  'longname': 'addControlVertex',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'acv'},
                             'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'asString': {'args': '<type __builtin__.str>',
                                          'longname': 'asString',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'asString'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'changeCommand': {'args': 'script',
                                               'longname': 'changeCommand',
                                               'modes': ['edit'],
                                               'numArgs': 1,
                                               'shortname': 'cc'},
                             'currentKey': {'args': '<type __builtin__.int>',
                                            'longname': 'currentKey',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ck'},
                             'currentKeyValue': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'currentKeyValue',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 2,
                                                 'shortname': 'ckv'},
                             'customCurveWidget': {'args': '<type __builtin__.bool>',
                                                   'longname': 'customCurveWidget',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ccw'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'deleteControlVertex': {'args': '<type __builtin__.int>',
                                                     'longname': 'deleteControlVertex',
                                                     'modes': ['edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dcv'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'optionVar': {'args': '<type __builtin__.str>',
                                           'longname': 'optionVar',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ov'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'readOnly': {'args': '<type __builtin__.bool>',
                                          'longname': 'readOnly',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ro'},
                             'snapToGrid': {'args': '<type __builtin__.bool>',
                                            'longname': 'snapToGrid',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'stg'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'removedFlags': {'as': 'asString'},
                   'shortFlags': {'acv': 'addControlVertex',
                                  'ann': 'annotation',
                                  'asString': 'asString',
                                  'bgc': 'backgroundColor',
                                  'cc': 'changeCommand',
                                  'ccw': 'customCurveWidget',
                                  'ck': 'currentKey',
                                  'ckv': 'currentKeyValue',
                                  'dcv': 'deleteControlVertex',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ebg': 'enableBackground',
                                  'ekf': 'enableKeyboardFocus',
                                  'en': 'enable',
                                  'ex': 'exists',
                                  'fpn': 'fullPathName',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'io': 'isObscured',
                                  'm': 'manage',
                                  'nbg': 'noBackground',
                                  'npm': 'numberOfPopupMenus',
                                  'ov': 'optionVar',
                                  'p': 'parent',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'q': 'query',
                                  'ro': 'readOnly',
                                  'sbm': 'statusBarMessage',
                                  'stg': 'snapToGrid',
                                  'ut': 'useTemplate',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'w': 'width'},
                   'type': 'windows'},
  'falloffCurveAttr': {'callbackFlags': ['dropCallback',
                                         'dpc',
                                         'visibleChangeCommand',
                                         'vcc',
                                         'dragCallback',
                                         'dgc',
                                         'changeCommand',
                                         'cc'],
                       'flags': {'addControlVertex': {'args': '<type __builtin__.str>',
                                                      'longname': 'addControlVertex',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'acv'},
                                 'annotation': {'args': '<type __builtin__.str>',
                                                'longname': 'annotation',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ann'},
                                 'asString': {'args': '<type __builtin__.str>',
                                              'longname': 'asString',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'asString'},
                                 'attribute': {'args': 'PyNode',
                                               'longname': 'attribute',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'at'},
                                 'backgroundColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'backgroundColor',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'bgc'},
                                 'changeCommand': {'args': 'script',
                                                   'longname': 'changeCommand',
                                                   'modes': ['edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cc'},
                                 'currentKey': {'args': '<type __builtin__.int>',
                                                'longname': 'currentKey',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ck'},
                                 'currentKeyValue': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'currentKeyValue',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'ckv'},
                                 'customCurveWidget': {'args': '<type __builtin__.int>',
                                                       'longname': 'customCurveWidget',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ccw'},
                                 'defineTemplate': {'args': '<type __builtin__.str>',
                                                    'longname': 'defineTemplate',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'dt'},
                                 'deleteControlVertex': {'args': '<type __builtin__.int>',
                                                         'longname': 'deleteControlVertex',
                                                         'modes': ['edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'dcv'},
                                 'docTag': {'args': '<type __builtin__.str>',
                                            'longname': 'docTag',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dtg'},
                                 'dragCallback': {'args': 'script',
                                                  'longname': 'dragCallback',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dgc'},
                                 'dropCallback': {'args': 'script',
                                                  'longname': 'dropCallback',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dpc'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'enable': {'args': '<type __builtin__.bool>',
                                            'longname': 'enable',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'en'},
                                 'enableBackground': {'args': '<type __builtin__.bool>',
                                                      'longname': 'enableBackground',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ebg'},
                                 'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                         'longname': 'enableKeyboardFocus',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ekf'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'fullPathName': {'args': '<type __builtin__.bool>',
                                                  'longname': 'fullPathName',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'fpn'},
                                 'height': {'args': '<type __builtin__.int>',
                                            'longname': 'height',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'h'},
                                 'highlightColor': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'highlightColor',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'hlc'},
                                 'isObscured': {'args': '<type __builtin__.bool>',
                                                'longname': 'isObscured',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'io'},
                                 'manage': {'args': '<type __builtin__.bool>',
                                            'longname': 'manage',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'm'},
                                 'noBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'noBackground',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'nbg'},
                                 'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                        'longname': 'numberOfPopupMenus',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'npm'},
                                 'parent': {'args': '<type __builtin__.str>',
                                            'longname': 'parent',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'p'},
                                 'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                    'longname': 'popupMenuArray',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'pma'},
                                 'preventOverride': {'args': '<type __builtin__.bool>',
                                                     'longname': 'preventOverride',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'po'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'readOnly': {'args': '<type __builtin__.int>',
                                              'longname': 'readOnly',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ro'},
                                 'selectedPositionControl': {'args': '<type __builtin__.str>',
                                                             'longname': 'selectedPositionControl',
                                                             'modes': ['create',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'spc'},
                                 'selectedValueControl': {'args': '<type __builtin__.str>',
                                                          'longname': 'selectedValueControl',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'svc'},
                                 'snapToGrid': {'args': '<type __builtin__.int>',
                                                'longname': 'snapToGrid',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'stg'},
                                 'statusBarMessage': {'args': '<type __builtin__.str>',
                                                      'longname': 'statusBarMessage',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'sbm'},
                                 'useTemplate': {'args': '<type __builtin__.str>',
                                                 'longname': 'useTemplate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ut'},
                                 'visible': {'args': '<type __builtin__.bool>',
                                             'longname': 'visible',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'vis'},
                                 'visibleChangeCommand': {'args': 'script',
                                                          'longname': 'visibleChangeCommand',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'vcc'},
                                 'width': {'args': '<type __builtin__.int>',
                                           'longname': 'width',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'w'}},
                       'removedFlags': {'as': 'asString'},
                       'shortFlags': {'acv': 'addControlVertex',
                                      'ann': 'annotation',
                                      'asString': 'asString',
                                      'at': 'attribute',
                                      'bgc': 'backgroundColor',
                                      'cc': 'changeCommand',
                                      'ccw': 'customCurveWidget',
                                      'ck': 'currentKey',
                                      'ckv': 'currentKeyValue',
                                      'dcv': 'deleteControlVertex',
                                      'dgc': 'dragCallback',
                                      'dpc': 'dropCallback',
                                      'dt': 'defineTemplate',
                                      'dtg': 'docTag',
                                      'e': 'edit',
                                      'ebg': 'enableBackground',
                                      'ekf': 'enableKeyboardFocus',
                                      'en': 'enable',
                                      'ex': 'exists',
                                      'fpn': 'fullPathName',
                                      'h': 'height',
                                      'hlc': 'highlightColor',
                                      'io': 'isObscured',
                                      'm': 'manage',
                                      'nbg': 'noBackground',
                                      'npm': 'numberOfPopupMenus',
                                      'p': 'parent',
                                      'pma': 'popupMenuArray',
                                      'po': 'preventOverride',
                                      'q': 'query',
                                      'ro': 'readOnly',
                                      'sbm': 'statusBarMessage',
                                      'spc': 'selectedPositionControl',
                                      'stg': 'snapToGrid',
                                      'svc': 'selectedValueControl',
                                      'ut': 'useTemplate',
                                      'vcc': 'visibleChangeCommand',
                                      'vis': 'visible',
                                      'w': 'width'},
                       'type': 'windows'},
  'fcheck': {'flags': {}, 'shortFlags': {}, 'type': 'system'},
  'file': {'callbackFlags': ['editCommand', 'ec', 'command', 'c'],
           'flags': {'activate': {'args': '<type __builtin__.bool>',
                                  'longname': 'activate',
                                  'modes': [],
                                  'numArgs': 0,
                                  'shortname': 'a'},
                     'activeProxy': {'args': '<type __builtin__.bool>',
                                     'longname': 'activeProxy',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ap'},
                     'add': {'args': '<type __builtin__.bool>',
                             'longname': 'add',
                             'modes': ['create'],
                             'numArgs': 0,
                             'shortname': 'add'},
                     'anyModified': {'args': '<type __builtin__.bool>',
                                     'longname': 'anyModified',
                                     'modes': ['query'],
                                     'numArgs': 0,
                                     'shortname': 'amf'},
                     'applyTo': {'args': '<type __builtin__.str>',
                                 'longname': 'applyTo',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'at'},
                     'buildLoadSettings': {'args': '<type __builtin__.bool>',
                                           'longname': 'buildLoadSettings',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'bls'},
                     'channels': {'args': '<type __builtin__.bool>',
                                  'longname': 'channels',
                                  'modes': ['create', 'query'],
                                  'modified': ['exportSelected'],
                                  'numArgs': 1,
                                  'shortname': 'chn'},
                     'cleanReference': {'args': '<type __builtin__.str>',
                                        'longname': 'cleanReference',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'secondaryFlags': ['editCommand'],
                                        'shortname': 'cr'},
                     'command': {'args': ['<type __builtin__.str>',
                                          '<type __builtin__.str>'],
                                 'longname': 'command',
                                 'modes': ['create', 'query'],
                                 'numArgs': 2,
                                 'shortname': 'c'},
                     'compress': {'args': '<type __builtin__.bool>',
                                  'longname': 'compress',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'cmp'},
                     'constraints': {'args': '<type __builtin__.bool>',
                                     'longname': 'constraints',
                                     'modes': ['create', 'query'],
                                     'modified': ['exportSelected'],
                                     'numArgs': 1,
                                     'shortname': 'con'},
                     'constructionHistory': {'args': '<type __builtin__.bool>',
                                             'longname': 'constructionHistory',
                                             'modes': ['create', 'query'],
                                             'modified': ['exportSelected'],
                                             'numArgs': 1,
                                             'shortname': 'ch'},
                     'copyNumberList': {'args': '<type __builtin__.bool>',
                                        'longname': 'copyNumberList',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'cnl'},
                     'defaultExtensions': {'args': '<type __builtin__.bool>',
                                           'longname': 'defaultExtensions',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'de'},
                     'defaultNamespace': {'args': '<type __builtin__.bool>',
                                          'longname': 'defaultNamespace',
                                          'modes': ['create'],
                                          'modified': ['reference', 'i'],
                                          'numArgs': 0,
                                          'shortname': 'dns'},
                     'deferReference': {'args': '<type __builtin__.bool>',
                                        'longname': 'deferReference',
                                        'modes': ['create', 'query'],
                                        'modified': ['reference', 'i'],
                                        'numArgs': 1,
                                        'shortname': 'dr'},
                     'edit': {'args': '<type __builtin__.bool>',
                              'longname': 'edit',
                              'numArgs': 0,
                              'shortname': 'e'},
                     'editCommand': {'args': '<type __builtin__.str>',
                                     'longname': 'editCommand',
                                     'modes': ['create'],
                                     'modified': ['cleanReference'],
                                     'numArgs': 1,
                                     'shortname': 'ec'},
                     'errorStatus': {'args': '<type __builtin__.bool>',
                                     'longname': 'errorStatus',
                                     'modes': ['query'],
                                     'numArgs': 0,
                                     'shortname': 'err'},
                     'executeScriptNodes': {'args': '<type __builtin__.bool>',
                                            'longname': 'executeScriptNodes',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'esn'},
                     'exists': {'args': '<type __builtin__.bool>',
                                'longname': 'exists',
                                'modes': ['query'],
                                'numArgs': 0,
                                'shortname': 'ex'},
                     'expandName': {'args': '<type __builtin__.bool>',
                                    'longname': 'expandName',
                                    'modes': ['query'],
                                    'numArgs': 0,
                                    'shortname': 'exn'},
                     'exportAll': {'args': '<type __builtin__.bool>',
                                   'longname': 'exportAll',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'secondaryFlags': ['force',
                                                      'preserveReferences',
                                                      'type'],
                                   'shortname': 'ea'},
                     'exportAnim': {'args': '<type __builtin__.bool>',
                                    'longname': 'exportAnim',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'secondaryFlags': ['force', 'type'],
                                    'shortname': 'ean'},
                     'exportAnimFromReference': {'args': '<type __builtin__.bool>',
                                                 'longname': 'exportAnimFromReference',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'secondaryFlags': ['force',
                                                                    'referenceNode',
                                                                    'type'],
                                                 'shortname': 'ear'},
                     'exportAsReference': {'args': '<type __builtin__.bool>',
                                           'longname': 'exportAsReference',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'secondaryFlags': ['namespace',
                                                              'renamingPrefix'],
                                           'shortname': 'er'},
                     'exportAsSegment': {'args': '<type __builtin__.bool>',
                                         'longname': 'exportAsSegment',
                                         'modes': [],
                                         'numArgs': 0,
                                         'shortname': 'exs'},
                     'exportSelected': {'args': '<type __builtin__.bool>',
                                        'longname': 'exportSelected',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'secondaryFlags': ['force',
                                                           'constructionHistory',
                                                           'channels',
                                                           'constraints',
                                                           'expressions',
                                                           'shader',
                                                           'preserveReferences',
                                                           'type'],
                                        'shortname': 'es'},
                     'exportSelectedAnim': {'args': '<type __builtin__.bool>',
                                            'longname': 'exportSelectedAnim',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'secondaryFlags': ['force',
                                                               'type'],
                                            'shortname': 'eas'},
                     'exportSelectedAnimFromReference': {'args': '<type __builtin__.bool>',
                                                         'longname': 'exportSelectedAnimFromReference',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'secondaryFlags': ['force',
                                                                            'referenceNode',
                                                                            'type'],
                                                         'shortname': 'esa'},
                     'exportSelectedNoReference': {'args': '<type __builtin__.bool>',
                                                   'longname': 'exportSelectedNoReference',
                                                   'modes': [],
                                                   'numArgs': 0,
                                                   'shortname': 'esr'},
                     'exportSelectedStrict': {'args': '<type __builtin__.bool>',
                                              'longname': 'exportSelectedStrict',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ess'},
                     'exportSnapshotCallback': {'args': ['script',
                                                         '<type __builtin__.str>'],
                                                'longname': 'exportSnapshotCallback',
                                                'modes': ['create'],
                                                'numArgs': 2,
                                                'shortname': 'esc'},
                     'exportUnloadedReferences': {'args': '<type __builtin__.bool>',
                                                  'longname': 'exportUnloadedReferences',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'eur'},
                     'expressions': {'args': '<type __builtin__.bool>',
                                     'longname': 'expressions',
                                     'modes': ['create', 'query'],
                                     'modified': ['exportSelected'],
                                     'numArgs': 1,
                                     'shortname': 'exp'},
                     'fileMetaData': {'args': '<type __builtin__.bool>',
                                      'longname': 'fileMetaData',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'fmd'},
                     'flushReference': {'args': '<type __builtin__.str>',
                                        'longname': 'flushReference',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'fr'},
                     'force': {'args': '<type __builtin__.bool>',
                               'longname': 'force',
                               'modes': ['create'],
                               'modified': ['open',
                                            'newFile',
                                            'save',
                                            'exportAll',
                                            'exportSelected',
                                            'exportAnim',
                                            'exportSelectedAnim',
                                            'exportAnimFromReference',
                                            'exportSelectedAnimFromReference'],
                               'numArgs': 0,
                               'shortname': 'f'},
                     'groupLocator': {'args': '<type __builtin__.bool>',
                                      'longname': 'groupLocator',
                                      'modes': ['create'],
                                      'modified': ['reference'],
                                      'numArgs': 0,
                                      'shortname': 'gl'},
                     'groupName': {'args': '<type __builtin__.str>',
                                   'longname': 'groupName',
                                   'modes': ['create'],
                                   'modified': ['reference', 'i'],
                                   'numArgs': 1,
                                   'shortname': 'gn'},
                     'groupReference': {'args': '<type __builtin__.bool>',
                                        'longname': 'groupReference',
                                        'modes': ['create'],
                                        'modified': ['reference', 'i'],
                                        'numArgs': 0,
                                        'shortname': 'gr'},
                     'i': {'args': '<type __builtin__.bool>',
                           'longname': 'i',
                           'modes': ['create'],
                           'numArgs': 0,
                           'secondaryFlags': ['loadNoReferences',
                                              'loadReferenceDepth',
                                              'defaultNamespace',
                                              'deferReference',
                                              'groupReference',
                                              'groupName',
                                              'renameAll',
                                              'renamingPrefix',
                                              'swapNamespace',
                                              'returnNewNodes',
                                              'preserveReferences'],
                           'shortname': 'i'},
                     'ignoreVersion': {'args': '<type __builtin__.bool>',
                                       'longname': 'ignoreVersion',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'iv'},
                     'importFrameRate': {'args': '<type __builtin__.bool>',
                                         'longname': 'importFrameRate',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ifr'},
                     'importReference': {'args': '<type __builtin__.bool>',
                                         'longname': 'importReference',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ir'},
                     'importTimeRange': {'args': '<type __builtin__.str>',
                                         'longname': 'importTimeRange',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'itr'},
                     'lastFileOption': {'args': '<type __builtin__.bool>',
                                        'longname': 'lastFileOption',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'lfo'},
                     'lastTempFile': {'args': '<type __builtin__.bool>',
                                      'longname': 'lastTempFile',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'ltf'},
                     'list': {'args': '<type __builtin__.bool>',
                              'longname': 'list',
                              'modes': ['query'],
                              'numArgs': 0,
                              'shortname': 'l'},
                     'loadAllDeferred': {'args': '<type __builtin__.bool>',
                                         'longname': 'loadAllDeferred',
                                         'modes': ['create'],
                                         'modified': ['open'],
                                         'numArgs': 1,
                                         'shortname': 'lad'},
                     'loadAllReferences': {'args': '<type __builtin__.bool>',
                                           'longname': 'loadAllReferences',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'lar'},
                     'loadNoReferences': {'args': '<type __builtin__.bool>',
                                          'longname': 'loadNoReferences',
                                          'modes': ['create'],
                                          'modified': ['open',
                                                       'i',
                                                       'reference',
                                                       'loadReference'],
                                          'numArgs': 0,
                                          'shortname': 'lnr'},
                     'loadReference': {'args': '<type __builtin__.str>',
                                       'longname': 'loadReference',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'secondaryFlags': ['loadNoReferences',
                                                          'loadReferenceDepth',
                                                          'returnNewNodes'],
                                       'shortname': 'lr'},
                     'loadReferenceDepth': {'args': '<type __builtin__.str>',
                                            'longname': 'loadReferenceDepth',
                                            'modes': ['create'],
                                            'modified': ['open',
                                                         'i',
                                                         'reference',
                                                         'loadReference'],
                                            'numArgs': 1,
                                            'shortname': 'lrd'},
                     'loadReferencePreview': {'args': '<type __builtin__.str>',
                                              'longname': 'loadReferencePreview',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'lrp'},
                     'loadSettings': {'args': '<type __builtin__.str>',
                                      'longname': 'loadSettings',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'ls'},
                     'location': {'args': '<type __builtin__.bool>',
                                  'longname': 'location',
                                  'modes': ['query'],
                                  'numArgs': 0,
                                  'shortname': 'loc'},
                     'lockContainerUnpublished': {'args': '<type __builtin__.bool>',
                                                  'longname': 'lockContainerUnpublished',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'lcu'},
                     'lockFile': {'args': '<type __builtin__.bool>',
                                  'longname': 'lockFile',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'lf'},
                     'lockReference': {'args': '<type __builtin__.bool>',
                                       'longname': 'lockReference',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'lck'},
                     'mapPlaceHolderNamespace': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'mapPlaceHolderNamespace',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit',
                                                           'multiuse'],
                                                 'numArgs': 2,
                                                 'shortname': 'mns'},
                     'mergeNamespaceWithParent': {'args': '<type __builtin__.bool>',
                                                  'longname': 'mergeNamespaceWithParent',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'mnp'},
                     'mergeNamespaceWithRoot': {'args': '<type __builtin__.bool>',
                                                'longname': 'mergeNamespaceWithRoot',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'mnr'},
                     'mergeNamespacesOnClash': {'args': '<type __builtin__.bool>',
                                                'longname': 'mergeNamespacesOnClash',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'mnc'},
                     'modified': {'args': '<type __builtin__.bool>',
                                  'longname': 'modified',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'mf'},
                     'moveSelected': {'args': '<type __builtin__.bool>',
                                      'longname': 'moveSelected',
                                      'modes': ['edit'],
                                      'numArgs': 0,
                                      'shortname': 'ms'},
                     'namespace': {'args': '<type __builtin__.str>',
                                   'longname': 'namespace',
                                   'modes': ['edit'],
                                   'modified': ['reference',
                                                'exportAsReference',
                                                'namespace'],
                                   'numArgs': 1,
                                   'secondaryFlags': ['namespace'],
                                   'shortname': 'ns'},
                     'newFile': {'args': '<type __builtin__.bool>',
                                 'longname': 'newFile',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'secondaryFlags': ['force', 'type'],
                                 'shortname': 'new'},
                     'open': {'args': '<type __builtin__.bool>',
                              'longname': 'open',
                              'modes': ['create'],
                              'numArgs': 0,
                              'secondaryFlags': ['loadAllDeferred',
                                                 'loadNoReferences',
                                                 'loadReferenceDepth',
                                                 'force',
                                                 'returnNewNodes',
                                                 'type'],
                              'shortname': 'o'},
                     'options': {'args': '<type __builtin__.str>',
                                 'longname': 'options',
                                 'modes': ['create', 'query'],
                                 'numArgs': 1,
                                 'shortname': 'op'},
                     'parentNamespace': {'args': '<type __builtin__.bool>',
                                         'longname': 'parentNamespace',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'pns'},
                     'postSaveScript': {'args': '<type __builtin__.str>',
                                        'longname': 'postSaveScript',
                                        'modes': ['create'],
                                        'modified': ['save'],
                                        'numArgs': 1,
                                        'shortname': 'pos'},
                     'preSaveScript': {'args': '<type __builtin__.str>',
                                       'longname': 'preSaveScript',
                                       'modes': ['create'],
                                       'modified': ['save'],
                                       'numArgs': 1,
                                       'shortname': 'prs'},
                     'preserveName': {'args': '<type __builtin__.bool>',
                                      'longname': 'preserveName',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'pn'},
                     'preserveReferences': {'args': '<type __builtin__.bool>',
                                            'longname': 'preserveReferences',
                                            'modes': ['create'],
                                            'modified': ['i',
                                                         'exportAll',
                                                         'exportSelected'],
                                            'numArgs': 0,
                                            'shortname': 'pr'},
                     'preserveUndo': {'args': '<type __builtin__.bool>',
                                      'longname': 'preserveUndo',
                                      'numArgs': 1,
                                      'shortname': 'pu'},
                     'preview': {'args': '<type __builtin__.bool>',
                                 'longname': 'preview',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'pv'},
                     'prompt': {'args': '<type __builtin__.bool>',
                                'longname': 'prompt',
                                'modes': ['create', 'query'],
                                'numArgs': 1,
                                'shortname': 'pmt'},
                     'proxyManager': {'args': '<type __builtin__.str>',
                                      'longname': 'proxyManager',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'pm'},
                     'proxyTag': {'args': '<type __builtin__.str>',
                                  'longname': 'proxyTag',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'pt'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'reference': {'args': '<type __builtin__.bool>',
                                   'longname': 'reference',
                                   'modes': ['query'],
                                   'numArgs': 0,
                                   'secondaryFlags': ['loadNoReferences',
                                                      'loadReferenceDepth',
                                                      'defaultNamespace',
                                                      'deferReference',
                                                      'groupReference',
                                                      'groupLocator',
                                                      'groupName',
                                                      'namespace',
                                                      'referenceNode',
                                                      'renamingPrefix',
                                                      'swapNamespace',
                                                      'sharedReferenceFile',
                                                      'sharedNodes',
                                                      'returnNewNodes'],
                                   'shortname': 'r'},
                     'referenceDepthInfo': {'args': '<type __builtin__.int>',
                                            'longname': 'referenceDepthInfo',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'rdi'},
                     'referenceNode': {'args': '<type __builtin__.str>',
                                       'longname': 'referenceNode',
                                       'modes': ['query'],
                                       'modified': ['reference',
                                                    'exportAnimFromReference',
                                                    'exportSelectedAnimFromReference'],
                                       'numArgs': 1,
                                       'shortname': 'rfn'},
                     'relativeNamespace': {'args': '<type __builtin__.str>',
                                           'longname': 'relativeNamespace',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'rns'},
                     'removeDuplicateNetworks': {'args': '<type __builtin__.bool>',
                                                 'longname': 'removeDuplicateNetworks',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'rdn'},
                     'removeReference': {'args': '<type __builtin__.bool>',
                                         'longname': 'removeReference',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'rr'},
                     'rename': {'args': '<type __builtin__.str>',
                                'longname': 'rename',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'rn'},
                     'renameAll': {'args': '<type __builtin__.bool>',
                                   'longname': 'renameAll',
                                   'modes': ['create'],
                                   'modified': ['i'],
                                   'numArgs': 1,
                                   'shortname': 'ra'},
                     'renameToSave': {'args': '<type __builtin__.bool>',
                                      'longname': 'renameToSave',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'rts'},
                     'renamingPrefix': {'args': '<type __builtin__.str>',
                                        'longname': 'renamingPrefix',
                                        'modes': ['create', 'query'],
                                        'modified': ['reference',
                                                     'i',
                                                     'exportAsReference'],
                                        'numArgs': 1,
                                        'shortname': 'rpr'},
                     'renamingPrefixList': {'args': '<type __builtin__.bool>',
                                            'longname': 'renamingPrefixList',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'rpl'},
                     'replaceName': {'args': ['<type __builtin__.str>',
                                              '<type __builtin__.str>'],
                                     'longname': 'replaceName',
                                     'modes': ['create',
                                               'query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 2,
                                     'shortname': 'rep'},
                     'resetError': {'args': '<type __builtin__.bool>',
                                    'longname': 'resetError',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'rer'},
                     'returnNewNodes': {'args': '<type __builtin__.bool>',
                                        'longname': 'returnNewNodes',
                                        'modes': ['create'],
                                        'modified': ['open',
                                                     'reference',
                                                     'i',
                                                     'loadReference'],
                                        'numArgs': 0,
                                        'shortname': 'rnn'},
                     'save': {'args': '<type __builtin__.bool>',
                              'longname': 'save',
                              'modes': ['create'],
                              'numArgs': 0,
                              'secondaryFlags': ['force',
                                                 'preSaveScript',
                                                 'postSaveScript',
                                                 'type'],
                              'shortname': 's'},
                     'saveDiskCache': {'args': '<type __builtin__.str>',
                                       'longname': 'saveDiskCache',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'sdc'},
                     'saveReference': {'args': '<type __builtin__.bool>',
                                       'longname': 'saveReference',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'sr'},
                     'saveReferencesUnloaded': {'args': '<type __builtin__.bool>',
                                                'longname': 'saveReferencesUnloaded',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'sru'},
                     'saveTextures': {'args': '<type __builtin__.str>',
                                      'longname': 'saveTextures',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'stx'},
                     'sceneName': {'args': '<type __builtin__.bool>',
                                   'longname': 'sceneName',
                                   'modes': ['query'],
                                   'numArgs': 0,
                                   'shortname': 'sn'},
                     'segment': {'args': '<type __builtin__.str>',
                                 'longname': 'segment',
                                 'modes': [],
                                 'numArgs': 1,
                                 'shortname': 'seg'},
                     'selectAll': {'args': '<type __builtin__.bool>',
                                   'longname': 'selectAll',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'sa'},
                     'shader': {'args': '<type __builtin__.bool>',
                                'longname': 'shader',
                                'modes': ['create', 'query'],
                                'modified': ['exportSelected'],
                                'numArgs': 1,
                                'shortname': 'sh'},
                     'sharedNodes': {'args': '<type __builtin__.str>',
                                     'longname': 'sharedNodes',
                                     'modes': ['create', 'multiuse'],
                                     'modified': ['reference'],
                                     'numArgs': 1,
                                     'shortname': 'shd'},
                     'sharedReferenceFile': {'args': '<type __builtin__.bool>',
                                             'longname': 'sharedReferenceFile',
                                             'modes': ['create'],
                                             'modified': ['reference'],
                                             'numArgs': 0,
                                             'shortname': 'srf'},
                     'shortName': {'args': '<type __builtin__.bool>',
                                   'longname': 'shortName',
                                   'modes': ['query'],
                                   'numArgs': 0,
                                   'shortname': 'shn'},
                     'strict': {'args': '<type __builtin__.bool>',
                                'longname': 'strict',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'str'},
                     'swapNamespace': {'args': ['<type __builtin__.str>',
                                                '<type __builtin__.str>'],
                                       'longname': 'swapNamespace',
                                       'modes': ['create', 'multiuse'],
                                       'modified': ['reference', 'i'],
                                       'numArgs': 2,
                                       'shortname': 'sns'},
                     'type': {'args': '<type __builtin__.str>',
                              'longname': 'type',
                              'modes': ['create', 'query'],
                              'modified': ['open',
                                           'newFile',
                                           'save',
                                           'exportAll',
                                           'exportSelected',
                                           'exportAnim',
                                           'exportSelectedAnim',
                                           'exportAnimFromReference',
                                           'exportSelectedAnimFromReference'],
                              'numArgs': 1,
                              'shortname': 'typ'},
                     'uiConfiguration': {'args': '<type __builtin__.bool>',
                                         'longname': 'uiConfiguration',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'uc'},
                     'unloadReference': {'args': '<type __builtin__.str>',
                                         'longname': 'unloadReference',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ur'},
                     'unresolvedName': {'args': '<type __builtin__.bool>',
                                        'longname': 'unresolvedName',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'un'},
                     'usingNamespaces': {'args': '<type __builtin__.bool>',
                                         'longname': 'usingNamespaces',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'uns'},
                     'withoutCopyNumber': {'args': '<type __builtin__.bool>',
                                           'longname': 'withoutCopyNumber',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'wcn'},
                     'writable': {'args': '<type __builtin__.bool>',
                                  'longname': 'writable',
                                  'modes': ['query'],
                                  'numArgs': 0,
                                  'shortname': 'w'}},
           'removedFlags': {'import': 'i'},
           'shortFlags': {'a': 'activate',
                          'add': 'add',
                          'amf': 'anyModified',
                          'ap': 'activeProxy',
                          'at': 'applyTo',
                          'bls': 'buildLoadSettings',
                          'c': 'command',
                          'ch': 'constructionHistory',
                          'chn': 'channels',
                          'cmp': 'compress',
                          'cnl': 'copyNumberList',
                          'con': 'constraints',
                          'cr': 'cleanReference',
                          'de': 'defaultExtensions',
                          'dns': 'defaultNamespace',
                          'dr': 'deferReference',
                          'e': 'edit',
                          'ea': 'exportAll',
                          'ean': 'exportAnim',
                          'ear': 'exportAnimFromReference',
                          'eas': 'exportSelectedAnim',
                          'ec': 'editCommand',
                          'er': 'exportAsReference',
                          'err': 'errorStatus',
                          'es': 'exportSelected',
                          'esa': 'exportSelectedAnimFromReference',
                          'esc': 'exportSnapshotCallback',
                          'esn': 'executeScriptNodes',
                          'esr': 'exportSelectedNoReference',
                          'ess': 'exportSelectedStrict',
                          'eur': 'exportUnloadedReferences',
                          'ex': 'exists',
                          'exn': 'expandName',
                          'exp': 'expressions',
                          'exs': 'exportAsSegment',
                          'f': 'force',
                          'fmd': 'fileMetaData',
                          'fr': 'flushReference',
                          'gl': 'groupLocator',
                          'gn': 'groupName',
                          'gr': 'groupReference',
                          'i': 'i',
                          'ifr': 'importFrameRate',
                          'ir': 'importReference',
                          'itr': 'importTimeRange',
                          'iv': 'ignoreVersion',
                          'l': 'list',
                          'lad': 'loadAllDeferred',
                          'lar': 'loadAllReferences',
                          'lck': 'lockReference',
                          'lcu': 'lockContainerUnpublished',
                          'lf': 'lockFile',
                          'lfo': 'lastFileOption',
                          'lnr': 'loadNoReferences',
                          'loc': 'location',
                          'lr': 'loadReference',
                          'lrd': 'loadReferenceDepth',
                          'lrp': 'loadReferencePreview',
                          'ls': 'loadSettings',
                          'ltf': 'lastTempFile',
                          'mf': 'modified',
                          'mnc': 'mergeNamespacesOnClash',
                          'mnp': 'mergeNamespaceWithParent',
                          'mnr': 'mergeNamespaceWithRoot',
                          'mns': 'mapPlaceHolderNamespace',
                          'ms': 'moveSelected',
                          'new': 'newFile',
                          'ns': 'namespace',
                          'o': 'open',
                          'op': 'options',
                          'pm': 'proxyManager',
                          'pmt': 'prompt',
                          'pn': 'preserveName',
                          'pns': 'parentNamespace',
                          'pos': 'postSaveScript',
                          'pr': 'preserveReferences',
                          'prs': 'preSaveScript',
                          'pt': 'proxyTag',
                          'pu': 'preserveUndo',
                          'pv': 'preview',
                          'q': 'query',
                          'r': 'reference',
                          'ra': 'renameAll',
                          'rdi': 'referenceDepthInfo',
                          'rdn': 'removeDuplicateNetworks',
                          'rep': 'replaceName',
                          'rer': 'resetError',
                          'rfn': 'referenceNode',
                          'rn': 'rename',
                          'rnn': 'returnNewNodes',
                          'rns': 'relativeNamespace',
                          'rpl': 'renamingPrefixList',
                          'rpr': 'renamingPrefix',
                          'rr': 'removeReference',
                          'rts': 'renameToSave',
                          's': 'save',
                          'sa': 'selectAll',
                          'sdc': 'saveDiskCache',
                          'seg': 'segment',
                          'sh': 'shader',
                          'shd': 'sharedNodes',
                          'shn': 'shortName',
                          'sn': 'sceneName',
                          'sns': 'swapNamespace',
                          'sr': 'saveReference',
                          'srf': 'sharedReferenceFile',
                          'sru': 'saveReferencesUnloaded',
                          'str': 'strict',
                          'stx': 'saveTextures',
                          'typ': 'type',
                          'uc': 'uiConfiguration',
                          'un': 'unresolvedName',
                          'uns': 'usingNamespaces',
                          'ur': 'unloadReference',
                          'w': 'writable',
                          'wcn': 'withoutCopyNumber'},
           'type': None},
  'fileBrowserDialog': {'callbackFlags': ['fileCommand', 'fc'],
                        'flags': {'actionName': {'args': '<type __builtin__.str>',
                                                 'longname': 'actionName',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'an'},
                                  'dialogStyle': {'args': '<type __builtin__.int>',
                                                  'longname': 'dialogStyle',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ds'},
                                  'fileCommand': {'args': 'script',
                                                  'longname': 'fileCommand',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'fc'},
                                  'fileType': {'args': '<type __builtin__.str>',
                                               'longname': 'fileType',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ft'},
                                  'filterList': {'args': '<type __builtin__.str>',
                                                 'longname': 'filterList',
                                                 'modes': ['create',
                                                           'multiuse'],
                                                 'numArgs': 1,
                                                 'shortname': 'fl'},
                                  'includeName': {'args': '<type __builtin__.str>',
                                                  'longname': 'includeName',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'includeName'},
                                  'mode': {'args': '<type __builtin__.int>',
                                           'longname': 'mode',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'm'},
                                  'operationMode': {'args': '<type __builtin__.str>',
                                                    'longname': 'operationMode',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'om'},
                                  'tipMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'tipMessage',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'tm'},
                                  'windowTitle': {'args': '<type __builtin__.str>',
                                                  'longname': 'windowTitle',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'wt'}},
                        'removedFlags': {'in': 'includeName'},
                        'shortFlags': {'an': 'actionName',
                                       'ds': 'dialogStyle',
                                       'fc': 'fileCommand',
                                       'fl': 'filterList',
                                       'ft': 'fileType',
                                       'includeName': 'includeName',
                                       'm': 'mode',
                                       'om': 'operationMode',
                                       'tm': 'tipMessage',
                                       'wt': 'windowTitle'},
                        'type': 'system'},
  'fileDialog': {'flags': {'application': {'args': '<type __builtin__.bool>',
                                           'longname': 'application',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'app'},
                           'defaultFileName': {'args': '<type __builtin__.str>',
                                               'longname': 'defaultFileName',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dfn'},
                           'directoryMask': {'args': '<type __builtin__.str>',
                                             'longname': 'directoryMask',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dm'},
                           'mode': {'args': '<type __builtin__.int>',
                                    'longname': 'mode',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'm'},
                           'title': {'args': '<type __builtin__.str>',
                                     'longname': 'title',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 't'}},
                 'shortFlags': {'app': 'application',
                                'dfn': 'defaultFileName',
                                'dm': 'directoryMask',
                                'm': 'mode',
                                't': 'title'},
                 'type': 'system'},
  'fileDialog2': {'callbackFlags': ['fileTypeChanged',
                                    'ftc',
                                    'optionsUICreate',
                                    'ocr',
                                    'selectionChanged',
                                    'sc',
                                    'optionsUIInit',
                                    'oin',
                                    'optionsUICommit',
                                    'ocm',
                                    'optionsUICancel',
                                    'oca',
                                    'optionsUICommit2',
                                    'oc2'],
                  'flags': {'buttonBoxOrientation': {'args': '<type __builtin__.int>',
                                                     'longname': 'buttonBoxOrientation',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'bbo'},
                            'cancelCaption': {'args': '<type __builtin__.str>',
                                              'longname': 'cancelCaption',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'cc'},
                            'caption': {'args': '<type __builtin__.str>',
                                        'longname': 'caption',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'cap'},
                            'dialogStyle': {'args': '<type __builtin__.int>',
                                            'longname': 'dialogStyle',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ds'},
                            'fileFilter': {'args': '<type __builtin__.str>',
                                           'longname': 'fileFilter',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ff'},
                            'fileMode': {'args': '<type __builtin__.int>',
                                         'longname': 'fileMode',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'fm'},
                            'fileTypeChanged': {'args': 'script',
                                                'longname': 'fileTypeChanged',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ftc'},
                            'hideFileExtensions': {'args': '<type __builtin__.bool>',
                                                   'longname': 'hideFileExtensions',
                                                   'numArgs': 1,
                                                   'shortname': 'hfe'},
                            'hideNameEdit': {'args': '<type __builtin__.bool>',
                                             'longname': 'hideNameEdit',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'hne'},
                            'okCaption': {'args': '<type __builtin__.str>',
                                          'longname': 'okCaption',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'okc'},
                            'optionsUICancel': {'args': 'script',
                                                'longname': 'optionsUICancel',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'oca'},
                            'optionsUICommit': {'args': 'script',
                                                'longname': 'optionsUICommit',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ocm'},
                            'optionsUICommit2': {'args': 'script',
                                                 'longname': 'optionsUICommit2',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'oc2'},
                            'optionsUICreate': {'args': 'script',
                                                'longname': 'optionsUICreate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ocr'},
                            'optionsUIInit': {'args': 'script',
                                              'longname': 'optionsUIInit',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'oin'},
                            'returnFilter': {'args': '<type __builtin__.bool>',
                                             'longname': 'returnFilter',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'rf'},
                            'selectFileFilter': {'args': '<type __builtin__.str>',
                                                 'longname': 'selectFileFilter',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'sff'},
                            'selectionChanged': {'args': 'script',
                                                 'longname': 'selectionChanged',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'sc'},
                            'setProjectBtnEnabled': {'args': '<type __builtin__.bool>',
                                                     'longname': 'setProjectBtnEnabled',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'spe'},
                            'startingDirectory': {'args': '<type __builtin__.str>',
                                                  'longname': 'startingDirectory',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dir'}},
                  'shortFlags': {'bbo': 'buttonBoxOrientation',
                                 'cap': 'caption',
                                 'cc': 'cancelCaption',
                                 'dir': 'startingDirectory',
                                 'ds': 'dialogStyle',
                                 'ff': 'fileFilter',
                                 'fm': 'fileMode',
                                 'ftc': 'fileTypeChanged',
                                 'hfe': 'hideFileExtensions',
                                 'hne': 'hideNameEdit',
                                 'oc2': 'optionsUICommit2',
                                 'oca': 'optionsUICancel',
                                 'ocm': 'optionsUICommit',
                                 'ocr': 'optionsUICreate',
                                 'oin': 'optionsUIInit',
                                 'okc': 'okCaption',
                                 'rf': 'returnFilter',
                                 'sc': 'selectionChanged',
                                 'sff': 'selectFileFilter',
                                 'spe': 'setProjectBtnEnabled'},
                  'type': 'system'},
  'fileInfo': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'referenceNode': {'args': '<type __builtin__.str>',
                                           'longname': 'referenceNode',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 'rfn'},
                         'remove': {'args': '<type __builtin__.str>',
                                    'longname': 'remove',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'rm'}},
               'shortFlags': {'q': 'query',
                              'rfn': 'referenceNode',
                              'rm': 'remove'},
               'type': 'system'},
  'filePathEditor': {'flags': {'attributeOnly': {'args': '<type __builtin__.bool>',
                                                 'longname': 'attributeOnly',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'ao'},
                               'attributeType': {'args': '<type __builtin__.str>',
                                                 'longname': 'attributeType',
                                                 'modes': ['query'],
                                                 'numArgs': 1,
                                                 'shortname': 'at'},
                               'byType': {'args': '<type __builtin__.str>',
                                          'longname': 'byType',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'bt'},
                               'copyAndRepath': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'copyAndRepath',
                                                 'modes': ['create'],
                                                 'numArgs': 2,
                                                 'shortname': 'cr'},
                               'deregisterType': {'args': '<type __builtin__.str>',
                                                  'longname': 'deregisterType',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'force': {'args': '<type __builtin__.bool>',
                                         'longname': 'force',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'f'},
                               'listDirectories': {'args': '<type __builtin__.str>',
                                                   'longname': 'listDirectories',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ld'},
                               'listFiles': {'args': '<type __builtin__.str>',
                                             'longname': 'listFiles',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'lf'},
                               'listRegisteredTypes': {'args': '<type __builtin__.bool>',
                                                       'longname': 'listRegisteredTypes',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'lrt'},
                               'preview': {'args': '<type __builtin__.bool>',
                                           'longname': 'preview',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'p'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'recursive': {'args': '<type __builtin__.bool>',
                                             'longname': 'recursive',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'rc'},
                               'refresh': {'args': '<type __builtin__.bool>',
                                           'longname': 'refresh',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'rf'},
                               'registerType': {'args': '<type __builtin__.str>',
                                                'longname': 'registerType',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'rt'},
                               'relativeNames': {'args': '<type __builtin__.bool>',
                                                 'longname': 'relativeNames',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'rel'},
                               'repath': {'args': '<type __builtin__.str>',
                                          'longname': 'repath',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'r'},
                               'replaceAll': {'args': '<type __builtin__.bool>',
                                              'longname': 'replaceAll',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ra'},
                               'replaceField': {'args': '<type __builtin__.str>',
                                                'longname': 'replaceField',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'rfd'},
                               'replaceString': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'replaceString',
                                                 'modes': ['create'],
                                                 'numArgs': 2,
                                                 'shortname': 'rs'},
                               'status': {'args': '<type __builtin__.bool>',
                                          'longname': 'status',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 's'},
                               'temporary': {'args': '<type __builtin__.bool>',
                                             'longname': 'temporary',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'tmp'},
                               'typeLabel': {'args': '<type __builtin__.str>',
                                             'longname': 'typeLabel',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'tl'},
                               'unresolved': {'args': '<type __builtin__.bool>',
                                              'longname': 'unresolved',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'u'},
                               'withAttribute': {'args': '<type __builtin__.bool>',
                                                 'longname': 'withAttribute',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'wa'}},
                     'shortFlags': {'ao': 'attributeOnly',
                                    'at': 'attributeType',
                                    'bt': 'byType',
                                    'cr': 'copyAndRepath',
                                    'dt': 'deregisterType',
                                    'f': 'force',
                                    'ld': 'listDirectories',
                                    'lf': 'listFiles',
                                    'lrt': 'listRegisteredTypes',
                                    'p': 'preview',
                                    'q': 'query',
                                    'r': 'repath',
                                    'ra': 'replaceAll',
                                    'rc': 'recursive',
                                    'rel': 'relativeNames',
                                    'rf': 'refresh',
                                    'rfd': 'replaceField',
                                    'rs': 'replaceString',
                                    'rt': 'registerType',
                                    's': 'status',
                                    'tl': 'typeLabel',
                                    'tmp': 'temporary',
                                    'u': 'unresolved',
                                    'wa': 'withAttribute'},
                     'type': 'system'},
  'filletCurve': {'flags': {'bias': {'args': '<type __builtin__.float>',
                                     'longname': 'bias',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'b'},
                            'blendControl': {'args': '<type __builtin__.bool>',
                                             'longname': 'blendControl',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'bc'},
                            'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'circular': {'args': '<type __builtin__.bool>',
                                         'longname': 'circular',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cir'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'curveParameter1': {'args': '<type __builtin__.float>',
                                                'longname': 'curveParameter1',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cp1'},
                            'curveParameter2': {'args': '<type __builtin__.float>',
                                                'longname': 'curveParameter2',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cp2'},
                            'depth': {'args': '<type __builtin__.float>',
                                      'longname': 'depth',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'd'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'freeformBlend': {'args': '<type __builtin__.bool>',
                                              'longname': 'freeformBlend',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fb'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'join': {'args': '<type __builtin__.bool>',
                                     'longname': 'join',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'jn'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'radius': {'args': '<type __builtin__.float>',
                                       'longname': 'radius',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'r'},
                            'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                'longname': 'replaceOriginal',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'rpo'},
                            'trim': {'args': '<type __builtin__.bool>',
                                     'longname': 'trim',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 't'}},
                  'shortFlags': {'b': 'bias',
                                 'bc': 'blendControl',
                                 'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'cir': 'circular',
                                 'cp1': 'curveParameter1',
                                 'cp2': 'curveParameter2',
                                 'd': 'depth',
                                 'e': 'edit',
                                 'fb': 'freeformBlend',
                                 'fzn': 'frozen',
                                 'jn': 'join',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'o': 'object',
                                 'q': 'query',
                                 'r': 'radius',
                                 'rpo': 'replaceOriginal',
                                 't': 'trim'},
                  'type': 'modeling'},
  'filter': {'flags': {'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'type': {'args': '<type __builtin__.str>',
                                'longname': 'type',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 't'}},
             'shortFlags': {},
             'type': None},
  'filterButterworthCtx': {'flags': {'apply': {'args': '<type __builtin__.bool>',
                                               'longname': 'apply',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'a'},
                                     'cutoffFrequency': {'args': '<type __builtin__.float>',
                                                         'longname': 'cutoffFrequency',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'cof'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'endTime': {'args': 'time',
                                                 'longname': 'endTime',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'e'},
                                     'exists': {'args': '<type __builtin__.bool>',
                                                'longname': 'exists',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ex'},
                                     'history': {'args': '<type __builtin__.bool>',
                                                 'longname': 'history',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                                     'image1': {'args': '<type __builtin__.str>',
                                                'longname': 'image1',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'i1'},
                                     'image2': {'args': '<type __builtin__.str>',
                                                'longname': 'image2',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'i2'},
                                     'image3': {'args': '<type __builtin__.str>',
                                                'longname': 'image3',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'i3'},
                                     'keepKeysOnFrame': {'args': '<type __builtin__.bool>',
                                                         'longname': 'keepKeysOnFrame',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'kof'},
                                     'name': {'args': '<type __builtin__.str>',
                                              'longname': 'name',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'n'},
                                     'preserveKeyTangent': {'args': '<type __builtin__.str>',
                                                            'longname': 'preserveKeyTangent',
                                                            'modes': ['multiuse'],
                                                            'numArgs': 1,
                                                            'shortname': 'pkt'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'samplingRate': {'args': '<type __builtin__.float>',
                                                      'longname': 'samplingRate',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'sr'},
                                     'selectedKeys': {'args': '<type __builtin__.bool>',
                                                      'longname': 'selectedKeys',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'sk'},
                                     'startTime': {'args': 'time',
                                                   'longname': 'startTime',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 's'}},
                           'shortFlags': {'a': 'apply',
                                          'ch': 'history',
                                          'cof': 'cutoffFrequency',
                                          'e': 'endTime',
                                          'ex': 'exists',
                                          'i1': 'image1',
                                          'i2': 'image2',
                                          'i3': 'image3',
                                          'kof': 'keepKeysOnFrame',
                                          'n': 'name',
                                          'pkt': 'preserveKeyTangent',
                                          'q': 'query',
                                          's': 'startTime',
                                          'sk': 'selectedKeys',
                                          'sr': 'samplingRate'},
                           'type': 'context'},
  'filterCurve': {'flags': {'cutoffFrequency': {'args': '<type __builtin__.float>',
                                                'longname': 'cutoffFrequency',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'cof'},
                            'endTime': {'args': 'time',
                                        'longname': 'endTime',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'e'},
                            'filter': {'args': '<type __builtin__.str>',
                                       'longname': 'filter',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'f'},
                            'keepKeysOnFrame': {'args': '<type __builtin__.bool>',
                                                'longname': 'keepKeysOnFrame',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'kof'},
                            'kernel': {'args': '<type __builtin__.str>',
                                       'longname': 'kernel',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ker'},
                            'keySync': {'args': '<type __builtin__.bool>',
                                        'longname': 'keySync',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ks'},
                            'maxTimeStep': {'args': '<type __builtin__.float>',
                                            'longname': 'maxTimeStep',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'mxs'},
                            'minTimeStep': {'args': '<type __builtin__.float>',
                                            'longname': 'minTimeStep',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'mns'},
                            'period': {'args': '<type __builtin__.float>',
                                       'longname': 'period',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'per'},
                            'precision': {'args': '<type __builtin__.float>',
                                          'longname': 'precision',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'pre'},
                            'precisionMode': {'args': '<type __builtin__.int>',
                                              'longname': 'precisionMode',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'pm'},
                            'preserveKeyTangent': {'args': '<type __builtin__.str>',
                                                   'longname': 'preserveKeyTangent',
                                                   'modes': ['create',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'pkt'},
                            'samplingRate': {'args': '<type __builtin__.float>',
                                             'longname': 'samplingRate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'sr'},
                            'selectedKeys': {'args': '<type __builtin__.bool>',
                                             'longname': 'selectedKeys',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'sk'},
                            'startTime': {'args': 'time',
                                          'longname': 'startTime',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 's'},
                            'timeTolerance': {'args': '<type __builtin__.float>',
                                              'longname': 'timeTolerance',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'tto'},
                            'tolerance': {'args': '<type __builtin__.float>',
                                          'longname': 'tolerance',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'tol'}},
                  'shortFlags': {'cof': 'cutoffFrequency',
                                 'e': 'endTime',
                                 'f': 'filter',
                                 'ker': 'kernel',
                                 'kof': 'keepKeysOnFrame',
                                 'ks': 'keySync',
                                 'mns': 'minTimeStep',
                                 'mxs': 'maxTimeStep',
                                 'per': 'period',
                                 'pkt': 'preserveKeyTangent',
                                 'pm': 'precisionMode',
                                 'pre': 'precision',
                                 's': 'startTime',
                                 'sk': 'selectedKeys',
                                 'sr': 'samplingRate',
                                 'tol': 'tolerance',
                                 'tto': 'timeTolerance'},
                  'type': 'animation'},
  'filterExpand': {'flags': {'expand': {'args': '<type __builtin__.bool>',
                                        'longname': 'expand',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ex'},
                             'fullPath': {'args': '<type __builtin__.bool>',
                                          'longname': 'fullPath',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'fp'},
                             'selectionMask': {'args': '<type __builtin__.int>',
                                               'longname': 'selectionMask',
                                               'modes': ['create',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'sm'},
                             'symActive': {'args': '<type __builtin__.bool>',
                                           'longname': 'symActive',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'sma'},
                             'symNegative': {'args': '<type __builtin__.bool>',
                                             'longname': 'symNegative',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'smn'},
                             'symPositive': {'args': '<type __builtin__.bool>',
                                             'longname': 'symPositive',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'smp'},
                             'symSeam': {'args': '<type __builtin__.bool>',
                                         'longname': 'symSeam',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'sms'}},
                   'shortFlags': {'ex': 'expand',
                                  'fp': 'fullPath',
                                  'sm': 'selectionMask',
                                  'sma': 'symActive',
                                  'smn': 'symNegative',
                                  'smp': 'symPositive',
                                  'sms': 'symSeam'},
                   'type': 'modeling'},
  'filterKeyReducerCtx': {'flags': {'apply': {'args': '<type __builtin__.bool>',
                                              'longname': 'apply',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'a'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'endTime': {'args': 'time',
                                                'longname': 'endTime',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'e'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'history': {'args': '<type __builtin__.bool>',
                                                'longname': 'history',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'keySync': {'args': '<type __builtin__.bool>',
                                                'longname': 'keySync',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ks'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'precision': {'args': '<type __builtin__.float>',
                                                  'longname': 'precision',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pre'},
                                    'precisionMode': {'args': '<type __builtin__.int>',
                                                      'longname': 'precisionMode',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'pm'},
                                    'preserveKeyTangent': {'args': '<type __builtin__.str>',
                                                           'longname': 'preserveKeyTangent',
                                                           'modes': ['query',
                                                                     'edit',
                                                                     'multiuse'],
                                                           'numArgs': 1,
                                                           'shortname': 'pkt'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'selectedKeys': {'args': '<type __builtin__.bool>',
                                                     'longname': 'selectedKeys',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sk'},
                                    'startTime': {'args': 'time',
                                                  'longname': 'startTime',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 's'}},
                          'shortFlags': {'a': 'apply',
                                         'ch': 'history',
                                         'e': 'endTime',
                                         'ex': 'exists',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'ks': 'keySync',
                                         'n': 'name',
                                         'pkt': 'preserveKeyTangent',
                                         'pm': 'precisionMode',
                                         'pre': 'precision',
                                         'q': 'query',
                                         's': 'startTime',
                                         'sk': 'selectedKeys'},
                          'type': 'context'},
  'findDeformers': {'flags': {}, 'shortFlags': {}, 'type': 'animation'},
  'findKeyframe': {'flags': {'animation': {'args': '<type __builtin__.str>',
                                           'longname': 'animation',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'an'},
                             'attribute': {'args': '<type __builtin__.str>',
                                           'longname': 'attribute',
                                           'modes': ['create', 'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'at'},
                             'controlPoints': {'args': '<type __builtin__.bool>',
                                               'longname': 'controlPoints',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'cp'},
                             'curve': {'args': '<type __builtin__.bool>',
                                       'longname': 'curve',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'c'},
                             'float': {'args': 'floatrange',
                                       'longname': 'float',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'f'},
                             'hierarchy': {'args': '<type __builtin__.str>',
                                           'longname': 'hierarchy',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'hi'},
                             'includeUpperBound': {'args': '<type __builtin__.bool>',
                                                   'longname': 'includeUpperBound',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'iub'},
                             'index': {'args': '<type __builtin__.int>',
                                       'longname': 'index',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'index'},
                             'shape': {'args': '<type __builtin__.bool>',
                                       'longname': 'shape',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 's'},
                             'time': {'args': 'timerange',
                                      'longname': 'time',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 't'},
                             'timeSlider': {'args': '<type __builtin__.bool>',
                                            'longname': 'timeSlider',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ts'},
                             'which': {'args': '<type __builtin__.str>',
                                       'longname': 'which',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'removedFlags': {'in': 'index'},
                   'shortFlags': {'an': 'animation',
                                  'at': 'attribute',
                                  'c': 'curve',
                                  'cp': 'controlPoints',
                                  'f': 'float',
                                  'hi': 'hierarchy',
                                  'index': 'index',
                                  'iub': 'includeUpperBound',
                                  's': 'shape',
                                  't': 'time',
                                  'ts': 'timeSlider',
                                  'w': 'which'},
                   'type': 'animation'},
  'findType': {'flags': {'deep': {'args': '<type __builtin__.bool>',
                                  'longname': 'deep',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'd'},
                         'exact': {'args': '<type __builtin__.bool>',
                                   'longname': 'exact',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'e'},
                         'forward': {'args': '<type __builtin__.bool>',
                                     'longname': 'forward',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'f'},
                         'type': {'args': '<type __builtin__.str>',
                                  'longname': 'type',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 't'}},
               'shortFlags': {'d': 'deep',
                              'e': 'exact',
                              'f': 'forward',
                              't': 'type'},
               'type': 'system'},
  'fitBspline': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'keepRange': {'args': '<type __builtin__.int>',
                                         'longname': 'keepRange',
                                         'numArgs': 1,
                                         'shortname': 'kr'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'object': {'args': '<type __builtin__.bool>',
                                      'longname': 'object',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'tolerance': {'args': '<type __builtin__.float>',
                                         'longname': 'tolerance',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'tol'}},
                 'shortFlags': {'cch': 'caching',
                                'ch': 'constructionHistory',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'kr': 'keepRange',
                                'n': 'name',
                                'nds': 'nodeState',
                                'o': 'object',
                                'q': 'query',
                                'tol': 'tolerance'},
                 'type': 'modeling'},
  'flagTest': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'floatRange': {'args': u'floatRange',
                                        'longname': 'floatRange',
                                        'modes': ['multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'fr'},
                         'indexRange': {'args': u'indexRange',
                                        'longname': 'indexRange',
                                        'modes': ['multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'ir'},
                         'multiUse': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.int>',
                                               '<type __builtin__.str>'],
                                      'longname': 'multiUse',
                                      'modes': ['multiuse'],
                                      'numArgs': 3,
                                      'shortname': 'mu'},
                         'noReport': {'args': '<type __builtin__.bool>',
                                      'longname': 'noReport',
                                      'numArgs': 0,
                                      'shortname': 'nr'},
                         'optionalQueryArgsFlag': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.int>',
                                                            '<type __builtin__.str>'],
                                                   'longname': 'optionalQueryArgsFlag',
                                                   'numArgs': 3,
                                                   'shortname': 'oqa'},
                         'pythonOptionalQueryArgsFlag': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.int>',
                                                                  '<type __builtin__.str>'],
                                                         'longname': 'pythonOptionalQueryArgsFlag',
                                                         'numArgs': 3,
                                                         'shortname': 'poq'},
                         'pythonQueryArgsFlag': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'pythonQueryArgsFlag',
                                                 'numArgs': 3,
                                                 'shortname': 'pq'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'queryArgsFlag': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.int>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'queryArgsFlag',
                                           'numArgs': 3,
                                           'shortname': 'qa'},
                         'simpleFlag': {'args': '<type __builtin__.bool>',
                                        'longname': 'simpleFlag',
                                        'numArgs': 0,
                                        'shortname': 's'},
                         'stringArrayFlag': {'args': u'string[...]',
                                             'longname': 'stringArrayFlag',
                                             'numArgs': 1,
                                             'shortname': 'saf'},
                         'stringFlag': {'args': '<type __builtin__.str>',
                                        'longname': 'stringFlag',
                                        'numArgs': 1,
                                        'shortname': 'sf'},
                         'timeRange': {'args': u'timeRange',
                                       'longname': 'timeRange',
                                       'modes': ['multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'tr'},
                         'tripleFloat': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'tripleFloat',
                                         'numArgs': 3,
                                         'shortname': 'tf'}},
               'shortFlags': {'e': 'edit',
                              'fr': 'floatRange',
                              'ir': 'indexRange',
                              'mu': 'multiUse',
                              'nr': 'noReport',
                              'oqa': 'optionalQueryArgsFlag',
                              'poq': 'pythonOptionalQueryArgsFlag',
                              'pq': 'pythonQueryArgsFlag',
                              'q': 'query',
                              'qa': 'queryArgsFlag',
                              's': 'simpleFlag',
                              'saf': 'stringArrayFlag',
                              'sf': 'stringFlag',
                              'tf': 'tripleFloat',
                              'tr': 'timeRange'},
               'type': 'other'},
  'flexor': {'callbackFlags': ['deformerCommand', 'dc'],
             'flags': {'atBones': {'args': '<type __builtin__.bool>',
                                   'longname': 'atBones',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'ab'},
                       'atJoints': {'args': '<type __builtin__.bool>',
                                    'longname': 'atJoints',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'aj'},
                       'deformerCommand': {'args': '<type __builtin__.str>',
                                           'longname': 'deformerCommand',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'dc'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'list': {'args': '<type __builtin__.bool>',
                                'longname': 'list',
                                'modes': ['query'],
                                'numArgs': 0,
                                'shortname': 'l'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'noScale': {'args': '<type __builtin__.bool>',
                                   'longname': 'noScale',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'ns'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'toSkeleton': {'args': '<type __builtin__.bool>',
                                      'longname': 'toSkeleton',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ts'},
                       'type': {'args': '<type __builtin__.str>',
                                'longname': 'type',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'typ'}},
             'shortFlags': {'ab': 'atBones',
                            'aj': 'atJoints',
                            'dc': 'deformerCommand',
                            'e': 'edit',
                            'l': 'list',
                            'n': 'name',
                            'ns': 'noScale',
                            'q': 'query',
                            'ts': 'toSkeleton',
                            'typ': 'type'},
             'type': 'animation'},
  'floatField': {'callbackFlags': ['dragCommand',
                                   'dc',
                                   'dropCallback',
                                   'dpc',
                                   'visibleChangeCommand',
                                   'vcc',
                                   'dragCallback',
                                   'dgc',
                                   'enterCommand',
                                   'ec',
                                   'receiveFocusCommand',
                                   'rfc',
                                   'changeCommand',
                                   'cc'],
                 'flags': {'annotation': {'args': '<type __builtin__.str>',
                                          'longname': 'annotation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ann'},
                           'backgroundColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'backgroundColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'bgc'},
                           'changeCommand': {'args': 'script',
                                             'longname': 'changeCommand',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cc'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'dragCallback': {'args': 'script',
                                            'longname': 'dragCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dgc'},
                           'dragCommand': {'args': 'script',
                                           'longname': 'dragCommand',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dc'},
                           'dropCallback': {'args': 'script',
                                            'longname': 'dropCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dpc'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'editable': {'args': '<type __builtin__.bool>',
                                        'longname': 'editable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ed'},
                           'enable': {'args': '<type __builtin__.bool>',
                                      'longname': 'enable',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'en'},
                           'enableBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableBackground',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ebg'},
                           'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableKeyboardFocus',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ekf'},
                           'enterCommand': {'args': 'script',
                                            'longname': 'enterCommand',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ec'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'fullPathName': {'args': '<type __builtin__.bool>',
                                            'longname': 'fullPathName',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fpn'},
                           'height': {'args': '<type __builtin__.int>',
                                      'longname': 'height',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                           'highlightColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'highlightColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'hlc'},
                           'isObscured': {'args': '<type __builtin__.bool>',
                                          'longname': 'isObscured',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'io'},
                           'manage': {'args': '<type __builtin__.bool>',
                                      'longname': 'manage',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                           'maxValue': {'args': '<type __builtin__.float>',
                                        'longname': 'maxValue',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'max'},
                           'minValue': {'args': '<type __builtin__.float>',
                                        'longname': 'minValue',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'min'},
                           'noBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'noBackground',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nbg'},
                           'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'numberOfPopupMenus',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'npm'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuArray': {'args': '<type __builtin__.bool>',
                                              'longname': 'popupMenuArray',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'pma'},
                           'precision': {'args': '<type __builtin__.int>',
                                         'longname': 'precision',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'pre'},
                           'preventOverride': {'args': '<type __builtin__.bool>',
                                               'longname': 'preventOverride',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'po'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'receiveFocusCommand': {'args': 'script',
                                                   'longname': 'receiveFocusCommand',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'rfc'},
                           'showTrailingZeros': {'args': '<type __builtin__.bool>',
                                                 'longname': 'showTrailingZeros',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'tze'},
                           'statusBarMessage': {'args': '<type __builtin__.str>',
                                                'longname': 'statusBarMessage',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sbm'},
                           'step': {'args': '<type __builtin__.float>',
                                    'longname': 'step',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 's'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'},
                           'value': {'args': '<type __builtin__.float>',
                                     'longname': 'value',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'v'},
                           'visible': {'args': '<type __builtin__.bool>',
                                       'longname': 'visible',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vis'},
                           'visibleChangeCommand': {'args': 'script',
                                                    'longname': 'visibleChangeCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'vcc'},
                           'width': {'args': '<type __builtin__.int>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'}},
                 'shortFlags': {'ann': 'annotation',
                                'bgc': 'backgroundColor',
                                'cc': 'changeCommand',
                                'dc': 'dragCommand',
                                'dgc': 'dragCallback',
                                'dpc': 'dropCallback',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'ebg': 'enableBackground',
                                'ec': 'enterCommand',
                                'ed': 'editable',
                                'ekf': 'enableKeyboardFocus',
                                'en': 'enable',
                                'ex': 'exists',
                                'fpn': 'fullPathName',
                                'h': 'height',
                                'hlc': 'highlightColor',
                                'io': 'isObscured',
                                'm': 'manage',
                                'max': 'maxValue',
                                'min': 'minValue',
                                'nbg': 'noBackground',
                                'npm': 'numberOfPopupMenus',
                                'p': 'parent',
                                'pma': 'popupMenuArray',
                                'po': 'preventOverride',
                                'pre': 'precision',
                                'q': 'query',
                                'rfc': 'receiveFocusCommand',
                                's': 'step',
                                'sbm': 'statusBarMessage',
                                'tze': 'showTrailingZeros',
                                'ut': 'useTemplate',
                                'v': 'value',
                                'vcc': 'visibleChangeCommand',
                                'vis': 'visible',
                                'w': 'width'},
                 'type': 'windows'},
  'floatFieldGrp': {'callbackFlags': ['dragCommand',
                                      'dc',
                                      'visibleChangeCommand',
                                      'vcc',
                                      'dragCallback',
                                      'dgc',
                                      'changeCommand',
                                      'cc',
                                      'dropCallback',
                                      'dpc'],
                    'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'adj'},
                              'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn2',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad2'},
                              'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn3',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad3'},
                              'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn4',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad4'},
                              'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn5',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad5'},
                              'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn6',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad6'},
                              'annotation': {'args': '<type __builtin__.str>',
                                             'longname': 'annotation',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ann'},
                              'backgroundColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'backgroundColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'bgc'},
                              'changeCommand': {'args': 'script',
                                                'longname': 'changeCommand',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cc'},
                              'columnAlign': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'cal'},
                              'columnAlign2': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'cl2'},
                              'columnAlign3': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'cl3'},
                              'columnAlign4': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'cl4'},
                              'columnAlign5': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'cl5'},
                              'columnAlign6': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'cl6'},
                              'columnAttach': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnAttach',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 3,
                                               'shortname': 'cat'},
                              'columnAttach2': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach2',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'ct2'},
                              'columnAttach3': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach3',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'ct3'},
                              'columnAttach4': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach4',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 4,
                                                'shortname': 'ct4'},
                              'columnAttach5': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach5',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 5,
                                                'shortname': 'ct5'},
                              'columnAttach6': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach6',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 6,
                                                'shortname': 'ct6'},
                              'columnOffset2': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset2',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'co2'},
                              'columnOffset3': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset3',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'co3'},
                              'columnOffset4': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset4',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 4,
                                                'shortname': 'co4'},
                              'columnOffset5': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset5',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 5,
                                                'shortname': 'co5'},
                              'columnOffset6': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset6',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 6,
                                                'shortname': 'co6'},
                              'columnWidth': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'cw'},
                              'columnWidth1': {'args': '<type __builtin__.int>',
                                               'longname': 'columnWidth1',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cw1'},
                              'columnWidth2': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'cw2'},
                              'columnWidth3': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'cw3'},
                              'columnWidth4': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'cw4'},
                              'columnWidth5': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'cw5'},
                              'columnWidth6': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'cw6'},
                              'defineTemplate': {'args': '<type __builtin__.str>',
                                                 'longname': 'defineTemplate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'dt'},
                              'docTag': {'args': '<type __builtin__.str>',
                                         'longname': 'docTag',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dtg'},
                              'dragCallback': {'args': 'script',
                                               'longname': 'dragCallback',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dgc'},
                              'dragCommand': {'args': 'script',
                                              'longname': 'dragCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dc'},
                              'dropCallback': {'args': 'script',
                                               'longname': 'dropCallback',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dpc'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'enable': {'args': '<type __builtin__.bool>',
                                         'longname': 'enable',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'en'},
                              'enable1': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en1'},
                              'enable2': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en2'},
                              'enable3': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en3'},
                              'enable4': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable4',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en4'},
                              'enableBackground': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableBackground',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ebg'},
                              'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'enableKeyboardFocus',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ekf'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'extraLabel': {'args': '<type __builtin__.str>',
                                             'longname': 'extraLabel',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'el'},
                              'fullPathName': {'args': '<type __builtin__.bool>',
                                               'longname': 'fullPathName',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'fpn'},
                              'height': {'args': '<type __builtin__.int>',
                                         'longname': 'height',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'h'},
                              'highlightColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'highlightColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'hlc'},
                              'isObscured': {'args': '<type __builtin__.bool>',
                                             'longname': 'isObscured',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'io'},
                              'label': {'args': '<type __builtin__.str>',
                                        'longname': 'label',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'l'},
                              'manage': {'args': '<type __builtin__.bool>',
                                         'longname': 'manage',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                              'noBackground': {'args': '<type __builtin__.bool>',
                                               'longname': 'noBackground',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nbg'},
                              'numberOfFields': {'args': '<type __builtin__.int>',
                                                 'longname': 'numberOfFields',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'nf'},
                              'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'numberOfPopupMenus',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'npm'},
                              'parent': {'args': '<type __builtin__.str>',
                                         'longname': 'parent',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'p'},
                              'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                 'longname': 'popupMenuArray',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'pma'},
                              'precision': {'args': '<type __builtin__.int>',
                                            'longname': 'precision',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pre'},
                              'preventOverride': {'args': '<type __builtin__.bool>',
                                                  'longname': 'preventOverride',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'po'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'rowAttach': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'rowAttach',
                                            'modes': ['create',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 3,
                                            'shortname': 'rat'},
                              'showTrailingZeros': {'args': '<type __builtin__.bool>',
                                                    'longname': 'showTrailingZeros',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'tze'},
                              'statusBarMessage': {'args': '<type __builtin__.str>',
                                                   'longname': 'statusBarMessage',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sbm'},
                              'step': {'args': '<type __builtin__.float>',
                                       'longname': 'step',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 's'},
                              'useTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'useTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ut'},
                              'value': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'value',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 4,
                                        'shortname': 'v'},
                              'value1': {'args': '<type __builtin__.float>',
                                         'longname': 'value1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'v1'},
                              'value2': {'args': '<type __builtin__.float>',
                                         'longname': 'value2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'v2'},
                              'value3': {'args': '<type __builtin__.float>',
                                         'longname': 'value3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'v3'},
                              'value4': {'args': '<type __builtin__.float>',
                                         'longname': 'value4',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'v4'},
                              'visible': {'args': '<type __builtin__.bool>',
                                          'longname': 'visible',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'vis'},
                              'visibleChangeCommand': {'args': 'script',
                                                       'longname': 'visibleChangeCommand',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'vcc'},
                              'width': {'args': '<type __builtin__.int>',
                                        'longname': 'width',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'w'}},
                    'shortFlags': {'ad2': 'adjustableColumn2',
                                   'ad3': 'adjustableColumn3',
                                   'ad4': 'adjustableColumn4',
                                   'ad5': 'adjustableColumn5',
                                   'ad6': 'adjustableColumn6',
                                   'adj': 'adjustableColumn',
                                   'ann': 'annotation',
                                   'bgc': 'backgroundColor',
                                   'cal': 'columnAlign',
                                   'cat': 'columnAttach',
                                   'cc': 'changeCommand',
                                   'cl2': 'columnAlign2',
                                   'cl3': 'columnAlign3',
                                   'cl4': 'columnAlign4',
                                   'cl5': 'columnAlign5',
                                   'cl6': 'columnAlign6',
                                   'co2': 'columnOffset2',
                                   'co3': 'columnOffset3',
                                   'co4': 'columnOffset4',
                                   'co5': 'columnOffset5',
                                   'co6': 'columnOffset6',
                                   'ct2': 'columnAttach2',
                                   'ct3': 'columnAttach3',
                                   'ct4': 'columnAttach4',
                                   'ct5': 'columnAttach5',
                                   'ct6': 'columnAttach6',
                                   'cw': 'columnWidth',
                                   'cw1': 'columnWidth1',
                                   'cw2': 'columnWidth2',
                                   'cw3': 'columnWidth3',
                                   'cw4': 'columnWidth4',
                                   'cw5': 'columnWidth5',
                                   'cw6': 'columnWidth6',
                                   'dc': 'dragCommand',
                                   'dgc': 'dragCallback',
                                   'dpc': 'dropCallback',
                                   'dt': 'defineTemplate',
                                   'dtg': 'docTag',
                                   'e': 'edit',
                                   'ebg': 'enableBackground',
                                   'ekf': 'enableKeyboardFocus',
                                   'el': 'extraLabel',
                                   'en': 'enable',
                                   'en1': 'enable1',
                                   'en2': 'enable2',
                                   'en3': 'enable3',
                                   'en4': 'enable4',
                                   'ex': 'exists',
                                   'fpn': 'fullPathName',
                                   'h': 'height',
                                   'hlc': 'highlightColor',
                                   'io': 'isObscured',
                                   'l': 'label',
                                   'm': 'manage',
                                   'nbg': 'noBackground',
                                   'nf': 'numberOfFields',
                                   'npm': 'numberOfPopupMenus',
                                   'p': 'parent',
                                   'pma': 'popupMenuArray',
                                   'po': 'preventOverride',
                                   'pre': 'precision',
                                   'q': 'query',
                                   'rat': 'rowAttach',
                                   's': 'step',
                                   'sbm': 'statusBarMessage',
                                   'tze': 'showTrailingZeros',
                                   'ut': 'useTemplate',
                                   'v': 'value',
                                   'v1': 'value1',
                                   'v2': 'value2',
                                   'v3': 'value3',
                                   'v4': 'value4',
                                   'vcc': 'visibleChangeCommand',
                                   'vis': 'visible',
                                   'w': 'width'},
                    'type': 'windows'},
  'floatScrollBar': {'callbackFlags': ['dragCommand',
                                       'dc',
                                       'dropCallback',
                                       'dpc',
                                       'visibleChangeCommand',
                                       'vcc',
                                       'dragCallback',
                                       'dgc',
                                       'changeCommand',
                                       'cc'],
                     'flags': {'annotation': {'args': '<type __builtin__.str>',
                                              'longname': 'annotation',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ann'},
                               'backgroundColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'backgroundColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'bgc'},
                               'changeCommand': {'args': 'script',
                                                 'longname': 'changeCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cc'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'dragCallback': {'args': 'script',
                                                'longname': 'dragCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dgc'},
                               'dragCommand': {'args': 'script',
                                               'longname': 'dragCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dc'},
                               'dropCallback': {'args': 'script',
                                                'longname': 'dropCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dpc'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'enable': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en'},
                               'enableBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableBackground',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ebg'},
                               'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                       'longname': 'enableKeyboardFocus',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ekf'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'fullPathName': {'args': '<type __builtin__.bool>',
                                                'longname': 'fullPathName',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'fpn'},
                               'height': {'args': '<type __builtin__.int>',
                                          'longname': 'height',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'h'},
                               'highlightColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'highlightColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'hlc'},
                               'horizontal': {'args': '<type __builtin__.bool>',
                                              'longname': 'horizontal',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'hr'},
                               'isObscured': {'args': '<type __builtin__.bool>',
                                              'longname': 'isObscured',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'io'},
                               'largeStep': {'args': '<type __builtin__.float>',
                                             'longname': 'largeStep',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ls'},
                               'manage': {'args': '<type __builtin__.bool>',
                                          'longname': 'manage',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'm'},
                               'maxValue': {'args': '<type __builtin__.float>',
                                            'longname': 'maxValue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'max'},
                               'minValue': {'args': '<type __builtin__.float>',
                                            'longname': 'minValue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'min'},
                               'noBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'noBackground',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nbg'},
                               'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'numberOfPopupMenus',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'npm'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                  'longname': 'popupMenuArray',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'pma'},
                               'preventOverride': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preventOverride',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'po'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'statusBarMessage': {'args': '<type __builtin__.str>',
                                                    'longname': 'statusBarMessage',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sbm'},
                               'step': {'args': '<type __builtin__.float>',
                                        'longname': 'step',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 's'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'},
                               'value': {'args': '<type __builtin__.float>',
                                         'longname': 'value',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'v'},
                               'visible': {'args': '<type __builtin__.bool>',
                                           'longname': 'visible',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vis'},
                               'visibleChangeCommand': {'args': 'script',
                                                        'longname': 'visibleChangeCommand',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'vcc'},
                               'width': {'args': '<type __builtin__.int>',
                                         'longname': 'width',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'w'}},
                     'shortFlags': {'ann': 'annotation',
                                    'bgc': 'backgroundColor',
                                    'cc': 'changeCommand',
                                    'dc': 'dragCommand',
                                    'dgc': 'dragCallback',
                                    'dpc': 'dropCallback',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'e': 'edit',
                                    'ebg': 'enableBackground',
                                    'ekf': 'enableKeyboardFocus',
                                    'en': 'enable',
                                    'ex': 'exists',
                                    'fpn': 'fullPathName',
                                    'h': 'height',
                                    'hlc': 'highlightColor',
                                    'hr': 'horizontal',
                                    'io': 'isObscured',
                                    'ls': 'largeStep',
                                    'm': 'manage',
                                    'max': 'maxValue',
                                    'min': 'minValue',
                                    'nbg': 'noBackground',
                                    'npm': 'numberOfPopupMenus',
                                    'p': 'parent',
                                    'pma': 'popupMenuArray',
                                    'po': 'preventOverride',
                                    'q': 'query',
                                    's': 'step',
                                    'sbm': 'statusBarMessage',
                                    'ut': 'useTemplate',
                                    'v': 'value',
                                    'vcc': 'visibleChangeCommand',
                                    'vis': 'visible',
                                    'w': 'width'},
                     'type': 'windows'},
  'floatSlider': {'callbackFlags': ['dragCommand',
                                    'dc',
                                    'dropCallback',
                                    'dpc',
                                    'visibleChangeCommand',
                                    'vcc',
                                    'dragCallback',
                                    'dgc',
                                    'changeCommand',
                                    'cc'],
                  'flags': {'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'changeCommand': {'args': 'script',
                                              'longname': 'changeCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cc'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dragCommand': {'args': 'script',
                                            'longname': 'dragCommand',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'horizontal': {'args': '<type __builtin__.bool>',
                                           'longname': 'horizontal',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'hr'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'maxValue': {'args': '<type __builtin__.float>',
                                         'longname': 'maxValue',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'max'},
                            'minValue': {'args': '<type __builtin__.float>',
                                         'longname': 'minValue',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'min'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'step': {'args': '<type __builtin__.float>',
                                     'longname': 'step',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 's'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'value': {'args': '<type __builtin__.float>',
                                      'longname': 'value',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'v'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'cc': 'changeCommand',
                                 'dc': 'dragCommand',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'ex': 'exists',
                                 'fpn': 'fullPathName',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'hr': 'horizontal',
                                 'io': 'isObscured',
                                 'm': 'manage',
                                 'max': 'maxValue',
                                 'min': 'minValue',
                                 'nbg': 'noBackground',
                                 'npm': 'numberOfPopupMenus',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 's': 'step',
                                 'sbm': 'statusBarMessage',
                                 'ut': 'useTemplate',
                                 'v': 'value',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'floatSlider2': {'callbackFlags': ['changeCommand2',
                                     'cc2',
                                     'changeCommand1',
                                     'cc1',
                                     'visibleChangeCommand',
                                     'vcc',
                                     'dragCallback',
                                     'dgc',
                                     'dropCallback',
                                     'dpc'],
                   'flags': {'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'changeCommand1': {'args': '<type __builtin__.str>',
                                                'longname': 'changeCommand1',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cc1'},
                             'changeCommand2': {'args': '<type __builtin__.str>',
                                                'longname': 'changeCommand2',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cc2'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'maximum': {'args': '<type __builtin__.float>',
                                         'longname': 'maximum',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'max'},
                             'minimum': {'args': '<type __builtin__.float>',
                                         'longname': 'minimum',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'min'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'polarity': {'args': '<type __builtin__.int>',
                                          'longname': 'polarity',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pol'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'positionControl1': {'args': '<type __builtin__.str>',
                                                  'longname': 'positionControl1',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pc1'},
                             'positionControl2': {'args': '<type __builtin__.str>',
                                                  'longname': 'positionControl2',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pc2'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'value1': {'args': '<type __builtin__.float>',
                                        'longname': 'value1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'v1'},
                             'value2': {'args': '<type __builtin__.float>',
                                        'longname': 'value2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'v2'},
                             'values': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'values',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 2,
                                        'shortname': 'vs'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'shortFlags': {'ann': 'annotation',
                                  'bgc': 'backgroundColor',
                                  'cc1': 'changeCommand1',
                                  'cc2': 'changeCommand2',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ebg': 'enableBackground',
                                  'ekf': 'enableKeyboardFocus',
                                  'en': 'enable',
                                  'ex': 'exists',
                                  'fpn': 'fullPathName',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'io': 'isObscured',
                                  'm': 'manage',
                                  'max': 'maximum',
                                  'min': 'minimum',
                                  'nbg': 'noBackground',
                                  'npm': 'numberOfPopupMenus',
                                  'p': 'parent',
                                  'pc1': 'positionControl1',
                                  'pc2': 'positionControl2',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'pol': 'polarity',
                                  'q': 'query',
                                  'sbm': 'statusBarMessage',
                                  'ut': 'useTemplate',
                                  'v1': 'value1',
                                  'v2': 'value2',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'vs': 'values',
                                  'w': 'width'},
                   'type': 'windows'},
  'floatSliderButtonGrp': {'callbackFlags': ['dragCommand',
                                             'dc',
                                             'visibleChangeCommand',
                                             'vcc',
                                             'dragCallback',
                                             'dgc',
                                             'buttonCommand',
                                             'bc',
                                             'symbolButtonCommand',
                                             'sbc',
                                             'changeCommand',
                                             'cc',
                                             'dropCallback',
                                             'dpc'],
                           'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                          'longname': 'adjustableColumn',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'adj'},
                                     'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                           'longname': 'adjustableColumn2',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ad2'},
                                     'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                           'longname': 'adjustableColumn3',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ad3'},
                                     'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                           'longname': 'adjustableColumn4',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ad4'},
                                     'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                           'longname': 'adjustableColumn5',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ad5'},
                                     'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                           'longname': 'adjustableColumn6',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ad6'},
                                     'annotation': {'args': '<type __builtin__.str>',
                                                    'longname': 'annotation',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ann'},
                                     'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'backgroundColor',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 3,
                                                         'shortname': 'bgc'},
                                     'buttonCommand': {'args': 'script',
                                                       'longname': 'buttonCommand',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'bc'},
                                     'buttonLabel': {'args': '<type __builtin__.str>',
                                                     'longname': 'buttonLabel',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'bl'},
                                     'changeCommand': {'args': 'script',
                                                       'longname': 'changeCommand',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'cc'},
                                     'columnAlign': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAlign',
                                                     'modes': ['create',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 2,
                                                     'shortname': 'cal'},
                                     'columnAlign2': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAlign2',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'cl2'},
                                     'columnAlign3': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAlign3',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'cl3'},
                                     'columnAlign4': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAlign4',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 4,
                                                      'shortname': 'cl4'},
                                     'columnAlign5': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAlign5',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 5,
                                                      'shortname': 'cl5'},
                                     'columnAlign6': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'columnAlign6',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 6,
                                                      'shortname': 'cl6'},
                                     'columnAttach': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.str>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnAttach',
                                                      'modes': ['create',
                                                                'edit',
                                                                'multiuse'],
                                                      'numArgs': 3,
                                                      'shortname': 'cat'},
                                     'columnAttach2': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAttach2',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 2,
                                                       'shortname': 'ct2'},
                                     'columnAttach3': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAttach3',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'ct3'},
                                     'columnAttach4': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAttach4',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 4,
                                                       'shortname': 'ct4'},
                                     'columnAttach5': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAttach5',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 5,
                                                       'shortname': 'ct5'},
                                     'columnAttach6': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'columnAttach6',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 6,
                                                       'shortname': 'ct6'},
                                     'columnOffset2': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnOffset2',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 2,
                                                       'shortname': 'co2'},
                                     'columnOffset3': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnOffset3',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'co3'},
                                     'columnOffset4': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnOffset4',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 4,
                                                       'shortname': 'co4'},
                                     'columnOffset5': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnOffset5',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 5,
                                                       'shortname': 'co5'},
                                     'columnOffset6': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'columnOffset6',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 6,
                                                       'shortname': 'co6'},
                                     'columnWidth': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnWidth',
                                                     'modes': ['create',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 2,
                                                     'shortname': 'cw'},
                                     'columnWidth1': {'args': '<type __builtin__.int>',
                                                      'longname': 'columnWidth1',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cw1'},
                                     'columnWidth2': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnWidth2',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'cw2'},
                                     'columnWidth3': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnWidth3',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'cw3'},
                                     'columnWidth4': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnWidth4',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 4,
                                                      'shortname': 'cw4'},
                                     'columnWidth5': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnWidth5',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 5,
                                                      'shortname': 'cw5'},
                                     'columnWidth6': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'columnWidth6',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 6,
                                                      'shortname': 'cw6'},
                                     'defineTemplate': {'args': '<type __builtin__.str>',
                                                        'longname': 'defineTemplate',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'dt'},
                                     'docTag': {'args': '<type __builtin__.str>',
                                                'longname': 'docTag',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dtg'},
                                     'dragCallback': {'args': 'script',
                                                      'longname': 'dragCallback',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dgc'},
                                     'dragCommand': {'args': 'script',
                                                     'longname': 'dragCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dc'},
                                     'dropCallback': {'args': 'script',
                                                      'longname': 'dropCallback',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dpc'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'enable': {'args': '<type __builtin__.bool>',
                                                'longname': 'enable',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'en'},
                                     'enableBackground': {'args': '<type __builtin__.bool>',
                                                          'longname': 'enableBackground',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ebg'},
                                     'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                             'longname': 'enableKeyboardFocus',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'ekf'},
                                     'exists': {'args': '<type __builtin__.bool>',
                                                'longname': 'exists',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ex'},
                                     'extraLabel': {'args': '<type __builtin__.str>',
                                                    'longname': 'extraLabel',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'el'},
                                     'field': {'args': '<type __builtin__.bool>',
                                               'longname': 'field',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'f'},
                                     'fieldMaxValue': {'args': '<type __builtin__.float>',
                                                       'longname': 'fieldMaxValue',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'fmx'},
                                     'fieldMinValue': {'args': '<type __builtin__.float>',
                                                       'longname': 'fieldMinValue',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'fmn'},
                                     'fieldStep': {'args': '<type __builtin__.float>',
                                                   'longname': 'fieldStep',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'fs'},
                                     'fullPathName': {'args': '<type __builtin__.bool>',
                                                      'longname': 'fullPathName',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'fpn'},
                                     'height': {'args': '<type __builtin__.int>',
                                                'longname': 'height',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'h'},
                                     'highlightColor': {'args': ['<type __builtin__.float>',
                                                                 '<type __builtin__.float>',
                                                                 '<type __builtin__.float>'],
                                                        'longname': 'highlightColor',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 3,
                                                        'shortname': 'hlc'},
                                     'image': {'args': '<type __builtin__.str>',
                                               'longname': 'image',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i'},
                                     'isObscured': {'args': '<type __builtin__.bool>',
                                                    'longname': 'isObscured',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'io'},
                                     'label': {'args': '<type __builtin__.str>',
                                               'longname': 'label',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'l'},
                                     'manage': {'args': '<type __builtin__.bool>',
                                                'longname': 'manage',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'm'},
                                     'maxValue': {'args': '<type __builtin__.float>',
                                                  'longname': 'maxValue',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'max'},
                                     'minValue': {'args': '<type __builtin__.float>',
                                                  'longname': 'minValue',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'min'},
                                     'noBackground': {'args': '<type __builtin__.bool>',
                                                      'longname': 'noBackground',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'nbg'},
                                     'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                            'longname': 'numberOfPopupMenus',
                                                            'modes': ['query'],
                                                            'numArgs': 0,
                                                            'shortname': 'npm'},
                                     'parent': {'args': '<type __builtin__.str>',
                                                'longname': 'parent',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'p'},
                                     'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                        'longname': 'popupMenuArray',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'pma'},
                                     'precision': {'args': '<type __builtin__.int>',
                                                   'longname': 'precision',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'pre'},
                                     'preventOverride': {'args': '<type __builtin__.bool>',
                                                         'longname': 'preventOverride',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'po'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'rowAttach': {'args': ['<type __builtin__.int>',
                                                            '<type __builtin__.str>',
                                                            '<type __builtin__.int>'],
                                                   'longname': 'rowAttach',
                                                   'modes': ['create',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 3,
                                                   'shortname': 'rat'},
                                     'sliderStep': {'args': '<type __builtin__.float>',
                                                    'longname': 'sliderStep',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ss'},
                                     'statusBarMessage': {'args': '<type __builtin__.str>',
                                                          'longname': 'statusBarMessage',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'sbm'},
                                     'step': {'args': '<type __builtin__.float>',
                                              'longname': 'step',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 's'},
                                     'symbolButtonCommand': {'args': 'script',
                                                             'longname': 'symbolButtonCommand',
                                                             'modes': ['create',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'sbc'},
                                     'symbolButtonDisplay': {'args': '<type __builtin__.bool>',
                                                             'longname': 'symbolButtonDisplay',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'sbd'},
                                     'useTemplate': {'args': '<type __builtin__.str>',
                                                     'longname': 'useTemplate',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ut'},
                                     'value': {'args': '<type __builtin__.float>',
                                               'longname': 'value',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'v'},
                                     'visible': {'args': '<type __builtin__.bool>',
                                                 'longname': 'visible',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'vis'},
                                     'visibleChangeCommand': {'args': 'script',
                                                              'longname': 'visibleChangeCommand',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'vcc'},
                                     'width': {'args': '<type __builtin__.int>',
                                               'longname': 'width',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'w'}},
                           'shortFlags': {'ad2': 'adjustableColumn2',
                                          'ad3': 'adjustableColumn3',
                                          'ad4': 'adjustableColumn4',
                                          'ad5': 'adjustableColumn5',
                                          'ad6': 'adjustableColumn6',
                                          'adj': 'adjustableColumn',
                                          'ann': 'annotation',
                                          'bc': 'buttonCommand',
                                          'bgc': 'backgroundColor',
                                          'bl': 'buttonLabel',
                                          'cal': 'columnAlign',
                                          'cat': 'columnAttach',
                                          'cc': 'changeCommand',
                                          'cl2': 'columnAlign2',
                                          'cl3': 'columnAlign3',
                                          'cl4': 'columnAlign4',
                                          'cl5': 'columnAlign5',
                                          'cl6': 'columnAlign6',
                                          'co2': 'columnOffset2',
                                          'co3': 'columnOffset3',
                                          'co4': 'columnOffset4',
                                          'co5': 'columnOffset5',
                                          'co6': 'columnOffset6',
                                          'ct2': 'columnAttach2',
                                          'ct3': 'columnAttach3',
                                          'ct4': 'columnAttach4',
                                          'ct5': 'columnAttach5',
                                          'ct6': 'columnAttach6',
                                          'cw': 'columnWidth',
                                          'cw1': 'columnWidth1',
                                          'cw2': 'columnWidth2',
                                          'cw3': 'columnWidth3',
                                          'cw4': 'columnWidth4',
                                          'cw5': 'columnWidth5',
                                          'cw6': 'columnWidth6',
                                          'dc': 'dragCommand',
                                          'dgc': 'dragCallback',
                                          'dpc': 'dropCallback',
                                          'dt': 'defineTemplate',
                                          'dtg': 'docTag',
                                          'e': 'edit',
                                          'ebg': 'enableBackground',
                                          'ekf': 'enableKeyboardFocus',
                                          'el': 'extraLabel',
                                          'en': 'enable',
                                          'ex': 'exists',
                                          'f': 'field',
                                          'fmn': 'fieldMinValue',
                                          'fmx': 'fieldMaxValue',
                                          'fpn': 'fullPathName',
                                          'fs': 'fieldStep',
                                          'h': 'height',
                                          'hlc': 'highlightColor',
                                          'i': 'image',
                                          'io': 'isObscured',
                                          'l': 'label',
                                          'm': 'manage',
                                          'max': 'maxValue',
                                          'min': 'minValue',
                                          'nbg': 'noBackground',
                                          'npm': 'numberOfPopupMenus',
                                          'p': 'parent',
                                          'pma': 'popupMenuArray',
                                          'po': 'preventOverride',
                                          'pre': 'precision',
                                          'q': 'query',
                                          'rat': 'rowAttach',
                                          's': 'step',
                                          'sbc': 'symbolButtonCommand',
                                          'sbd': 'symbolButtonDisplay',
                                          'sbm': 'statusBarMessage',
                                          'ss': 'sliderStep',
                                          'ut': 'useTemplate',
                                          'v': 'value',
                                          'vcc': 'visibleChangeCommand',
                                          'vis': 'visible',
                                          'w': 'width'},
                           'type': 'windows'},
  'floatSliderGrp': {'callbackFlags': ['dragCommand',
                                       'dc',
                                       'visibleChangeCommand',
                                       'vcc',
                                       'dragCallback',
                                       'dgc',
                                       'changeCommand',
                                       'cc',
                                       'dropCallback',
                                       'dpc'],
                     'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'adj'},
                               'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn2',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad2'},
                               'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn3',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad3'},
                               'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn4',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad4'},
                               'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn5',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad5'},
                               'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn6',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad6'},
                               'annotation': {'args': '<type __builtin__.str>',
                                              'longname': 'annotation',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ann'},
                               'backgroundColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'backgroundColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'bgc'},
                               'changeCommand': {'args': 'script',
                                                 'longname': 'changeCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cc'},
                               'columnAlign': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 2,
                                               'shortname': 'cal'},
                               'columnAlign2': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign2',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'cl2'},
                               'columnAlign3': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign3',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'cl3'},
                               'columnAlign4': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign4',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 4,
                                                'shortname': 'cl4'},
                               'columnAlign5': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign5',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 5,
                                                'shortname': 'cl5'},
                               'columnAlign6': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign6',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 6,
                                                'shortname': 'cl6'},
                               'columnAttach': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnAttach',
                                                'modes': ['create',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 3,
                                                'shortname': 'cat'},
                               'columnAttach2': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach2',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 2,
                                                 'shortname': 'ct2'},
                               'columnAttach3': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach3',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'ct3'},
                               'columnAttach4': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach4',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 4,
                                                 'shortname': 'ct4'},
                               'columnAttach5': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach5',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 5,
                                                 'shortname': 'ct5'},
                               'columnAttach6': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach6',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 6,
                                                 'shortname': 'ct6'},
                               'columnOffset2': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset2',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 2,
                                                 'shortname': 'co2'},
                               'columnOffset3': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset3',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'co3'},
                               'columnOffset4': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset4',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 4,
                                                 'shortname': 'co4'},
                               'columnOffset5': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset5',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 5,
                                                 'shortname': 'co5'},
                               'columnOffset6': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset6',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 6,
                                                 'shortname': 'co6'},
                               'columnWidth': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 2,
                                               'shortname': 'cw'},
                               'columnWidth1': {'args': '<type __builtin__.int>',
                                                'longname': 'columnWidth1',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cw1'},
                               'columnWidth2': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth2',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'cw2'},
                               'columnWidth3': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth3',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'cw3'},
                               'columnWidth4': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth4',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 4,
                                                'shortname': 'cw4'},
                               'columnWidth5': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth5',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 5,
                                                'shortname': 'cw5'},
                               'columnWidth6': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth6',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 6,
                                                'shortname': 'cw6'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'dragCallback': {'args': 'script',
                                                'longname': 'dragCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dgc'},
                               'dragCommand': {'args': 'script',
                                               'longname': 'dragCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dc'},
                               'dropCallback': {'args': 'script',
                                                'longname': 'dropCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dpc'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'enable': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en'},
                               'enableBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableBackground',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ebg'},
                               'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                       'longname': 'enableKeyboardFocus',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ekf'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'extraLabel': {'args': '<type __builtin__.str>',
                                              'longname': 'extraLabel',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'el'},
                               'field': {'args': '<type __builtin__.bool>',
                                         'longname': 'field',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'f'},
                               'fieldMaxValue': {'args': '<type __builtin__.float>',
                                                 'longname': 'fieldMaxValue',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'fmx'},
                               'fieldMinValue': {'args': '<type __builtin__.float>',
                                                 'longname': 'fieldMinValue',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'fmn'},
                               'fieldStep': {'args': '<type __builtin__.float>',
                                             'longname': 'fieldStep',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'fs'},
                               'fullPathName': {'args': '<type __builtin__.bool>',
                                                'longname': 'fullPathName',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'fpn'},
                               'height': {'args': '<type __builtin__.int>',
                                          'longname': 'height',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'h'},
                               'highlightColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'highlightColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'hlc'},
                               'isObscured': {'args': '<type __builtin__.bool>',
                                              'longname': 'isObscured',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'io'},
                               'label': {'args': '<type __builtin__.str>',
                                         'longname': 'label',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'l'},
                               'manage': {'args': '<type __builtin__.bool>',
                                          'longname': 'manage',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'm'},
                               'maxValue': {'args': '<type __builtin__.float>',
                                            'longname': 'maxValue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'max'},
                               'minValue': {'args': '<type __builtin__.float>',
                                            'longname': 'minValue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'min'},
                               'noBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'noBackground',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nbg'},
                               'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'numberOfPopupMenus',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'npm'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                  'longname': 'popupMenuArray',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'pma'},
                               'precision': {'args': '<type __builtin__.int>',
                                             'longname': 'precision',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'pre'},
                               'preventOverride': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preventOverride',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'po'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'rowAttach': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'rowAttach',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 3,
                                             'shortname': 'rat'},
                               'sliderStep': {'args': '<type __builtin__.float>',
                                              'longname': 'sliderStep',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ss'},
                               'statusBarMessage': {'args': '<type __builtin__.str>',
                                                    'longname': 'statusBarMessage',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sbm'},
                               'step': {'args': '<type __builtin__.float>',
                                        'longname': 'step',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 's'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'},
                               'value': {'args': '<type __builtin__.float>',
                                         'longname': 'value',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'v'},
                               'visible': {'args': '<type __builtin__.bool>',
                                           'longname': 'visible',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vis'},
                               'visibleChangeCommand': {'args': 'script',
                                                        'longname': 'visibleChangeCommand',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'vcc'},
                               'width': {'args': '<type __builtin__.int>',
                                         'longname': 'width',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'w'}},
                     'shortFlags': {'ad2': 'adjustableColumn2',
                                    'ad3': 'adjustableColumn3',
                                    'ad4': 'adjustableColumn4',
                                    'ad5': 'adjustableColumn5',
                                    'ad6': 'adjustableColumn6',
                                    'adj': 'adjustableColumn',
                                    'ann': 'annotation',
                                    'bgc': 'backgroundColor',
                                    'cal': 'columnAlign',
                                    'cat': 'columnAttach',
                                    'cc': 'changeCommand',
                                    'cl2': 'columnAlign2',
                                    'cl3': 'columnAlign3',
                                    'cl4': 'columnAlign4',
                                    'cl5': 'columnAlign5',
                                    'cl6': 'columnAlign6',
                                    'co2': 'columnOffset2',
                                    'co3': 'columnOffset3',
                                    'co4': 'columnOffset4',
                                    'co5': 'columnOffset5',
                                    'co6': 'columnOffset6',
                                    'ct2': 'columnAttach2',
                                    'ct3': 'columnAttach3',
                                    'ct4': 'columnAttach4',
                                    'ct5': 'columnAttach5',
                                    'ct6': 'columnAttach6',
                                    'cw': 'columnWidth',
                                    'cw1': 'columnWidth1',
                                    'cw2': 'columnWidth2',
                                    'cw3': 'columnWidth3',
                                    'cw4': 'columnWidth4',
                                    'cw5': 'columnWidth5',
                                    'cw6': 'columnWidth6',
                                    'dc': 'dragCommand',
                                    'dgc': 'dragCallback',
                                    'dpc': 'dropCallback',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'e': 'edit',
                                    'ebg': 'enableBackground',
                                    'ekf': 'enableKeyboardFocus',
                                    'el': 'extraLabel',
                                    'en': 'enable',
                                    'ex': 'exists',
                                    'f': 'field',
                                    'fmn': 'fieldMinValue',
                                    'fmx': 'fieldMaxValue',
                                    'fpn': 'fullPathName',
                                    'fs': 'fieldStep',
                                    'h': 'height',
                                    'hlc': 'highlightColor',
                                    'io': 'isObscured',
                                    'l': 'label',
                                    'm': 'manage',
                                    'max': 'maxValue',
                                    'min': 'minValue',
                                    'nbg': 'noBackground',
                                    'npm': 'numberOfPopupMenus',
                                    'p': 'parent',
                                    'pma': 'popupMenuArray',
                                    'po': 'preventOverride',
                                    'pre': 'precision',
                                    'q': 'query',
                                    'rat': 'rowAttach',
                                    's': 'step',
                                    'sbm': 'statusBarMessage',
                                    'ss': 'sliderStep',
                                    'ut': 'useTemplate',
                                    'v': 'value',
                                    'vcc': 'visibleChangeCommand',
                                    'vis': 'visible',
                                    'w': 'width'},
                     'type': 'windows'},
  'flow': {'flags': {'divisions': {'args': ['<type __builtin__.int>',
                                            '<type __builtin__.int>',
                                            '<type __builtin__.int>'],
                                   'longname': 'divisions',
                                   'modes': ['query'],
                                   'numArgs': 3,
                                   'shortname': 'dv'},
                     'edit': {'args': '<type __builtin__.bool>',
                              'longname': 'edit',
                              'numArgs': 0,
                              'shortname': 'e'},
                     'localCompute': {'args': '<type __builtin__.bool>',
                                      'longname': 'localCompute',
                                      'modes': ['query'],
                                      'numArgs': 1,
                                      'shortname': 'lc'},
                     'localDivisions': {'args': ['<type __builtin__.int>',
                                                 '<type __builtin__.int>',
                                                 '<type __builtin__.int>'],
                                        'longname': 'localDivisions',
                                        'modes': ['query'],
                                        'numArgs': 3,
                                        'shortname': 'ld'},
                     'objectCentered': {'args': '<type __builtin__.bool>',
                                        'longname': 'objectCentered',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'oc'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'}},
           'shortFlags': {'dv': 'divisions',
                          'e': 'edit',
                          'lc': 'localCompute',
                          'ld': 'localDivisions',
                          'oc': 'objectCentered',
                          'q': 'query'},
           'type': 'animation'},
  'flowLayout': {'callbackFlags': ['visibleChangeCommand',
                                   'vcc',
                                   'dragCallback',
                                   'dgc',
                                   'dropCallback',
                                   'dpc'],
                 'flags': {'annotation': {'args': '<type __builtin__.str>',
                                          'longname': 'annotation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ann'},
                           'backgroundColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'backgroundColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'bgc'},
                           'childArray': {'args': '<type __builtin__.bool>',
                                          'longname': 'childArray',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'ca'},
                           'columnSpacing': {'args': '<type __builtin__.int>',
                                             'longname': 'columnSpacing',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cs'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'dragCallback': {'args': 'script',
                                            'longname': 'dragCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dgc'},
                           'dropCallback': {'args': 'script',
                                            'longname': 'dropCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dpc'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'enable': {'args': '<type __builtin__.bool>',
                                      'longname': 'enable',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'en'},
                           'enableBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableBackground',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ebg'},
                           'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableKeyboardFocus',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ekf'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'fullPathName': {'args': '<type __builtin__.bool>',
                                            'longname': 'fullPathName',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fpn'},
                           'height': {'args': '<type __builtin__.int>',
                                      'longname': 'height',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                           'highlightColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'highlightColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'hlc'},
                           'horizontal': {'args': '<type __builtin__.bool>',
                                          'longname': 'horizontal',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'hr'},
                           'isObscured': {'args': '<type __builtin__.bool>',
                                          'longname': 'isObscured',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'io'},
                           'manage': {'args': '<type __builtin__.bool>',
                                      'longname': 'manage',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                           'noBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'noBackground',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nbg'},
                           'numberOfChildren': {'args': '<type __builtin__.bool>',
                                                'longname': 'numberOfChildren',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'nch'},
                           'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'numberOfPopupMenus',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'npm'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuArray': {'args': '<type __builtin__.bool>',
                                              'longname': 'popupMenuArray',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'pma'},
                           'preventOverride': {'args': '<type __builtin__.bool>',
                                               'longname': 'preventOverride',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'po'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'statusBarMessage': {'args': '<type __builtin__.str>',
                                                'longname': 'statusBarMessage',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sbm'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'},
                           'vertical': {'args': '<type __builtin__.bool>',
                                        'longname': 'vertical',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'v'},
                           'visible': {'args': '<type __builtin__.bool>',
                                       'longname': 'visible',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vis'},
                           'visibleChangeCommand': {'args': 'script',
                                                    'longname': 'visibleChangeCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'vcc'},
                           'width': {'args': '<type __builtin__.int>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'},
                           'wrap': {'args': '<type __builtin__.bool>',
                                    'longname': 'wrap',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'wr'}},
                 'shortFlags': {'ann': 'annotation',
                                'bgc': 'backgroundColor',
                                'ca': 'childArray',
                                'cs': 'columnSpacing',
                                'dgc': 'dragCallback',
                                'dpc': 'dropCallback',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'ebg': 'enableBackground',
                                'ekf': 'enableKeyboardFocus',
                                'en': 'enable',
                                'ex': 'exists',
                                'fpn': 'fullPathName',
                                'h': 'height',
                                'hlc': 'highlightColor',
                                'hr': 'horizontal',
                                'io': 'isObscured',
                                'm': 'manage',
                                'nbg': 'noBackground',
                                'nch': 'numberOfChildren',
                                'npm': 'numberOfPopupMenus',
                                'p': 'parent',
                                'pma': 'popupMenuArray',
                                'po': 'preventOverride',
                                'q': 'query',
                                'sbm': 'statusBarMessage',
                                'ut': 'useTemplate',
                                'v': 'vertical',
                                'vcc': 'visibleChangeCommand',
                                'vis': 'visible',
                                'w': 'width',
                                'wr': 'wrap'},
                 'type': 'windows'},
  'fluidAppend': {'type': 'runtime'},
  'fluidAppendOpt': {'type': 'runtime'},
  'fluidCacheInfo': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                             'longname': 'attribute',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'at'},
                               'cacheTime': {'args': 'time',
                                             'longname': 'cacheTime',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 't'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'endFrame': {'args': '<type __builtin__.bool>',
                                            'longname': 'endFrame',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 0,
                                            'shortname': 'ef'},
                               'hasCache': {'args': '<type __builtin__.bool>',
                                            'longname': 'hasCache',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 0,
                                            'shortname': 'hc'},
                               'hasData': {'args': '<type __builtin__.bool>',
                                           'longname': 'hasData',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 0,
                                           'shortname': 'hd'},
                               'initialConditions': {'args': '<type __builtin__.bool>',
                                                     'longname': 'initialConditions',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'ic'},
                               'playback': {'args': '<type __builtin__.bool>',
                                            'longname': 'playback',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 0,
                                            'shortname': 'pb'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'resolution': {'args': '<type __builtin__.bool>',
                                              'longname': 'resolution',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 0,
                                              'shortname': 're'},
                               'startFrame': {'args': '<type __builtin__.bool>',
                                              'longname': 'startFrame',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 0,
                                              'shortname': 'sf'}},
                     'shortFlags': {'at': 'attribute',
                                    'e': 'edit',
                                    'ef': 'endFrame',
                                    'hc': 'hasCache',
                                    'hd': 'hasData',
                                    'ic': 'initialConditions',
                                    'pb': 'playback',
                                    'q': 'query',
                                    're': 'resolution',
                                    'sf': 'startFrame',
                                    't': 'cacheTime'},
                     'type': 'effects'},
  'fluidDeleteCache': {'type': 'runtime'},
  'fluidDeleteCacheFrames': {'type': 'runtime'},
  'fluidDeleteCacheFramesOpt': {'type': 'runtime'},
  'fluidDeleteCacheOpt': {'type': 'runtime'},
  'fluidEmitter': {'flags': {'cycleEmission': {'args': '<type __builtin__.str>',
                                               'longname': 'cycleEmission',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cye'},
                             'cycleInterval': {'args': '<type __builtin__.int>',
                                               'longname': 'cycleInterval',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cyi'},
                             'densityEmissionRate': {'args': '<type __builtin__.float>',
                                                     'longname': 'densityEmissionRate',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'der'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'fluidDropoff': {'args': '<type __builtin__.float>',
                                              'longname': 'fluidDropoff',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fdr'},
                             'fuelEmissionRate': {'args': '<type __builtin__.float>',
                                                  'longname': 'fuelEmissionRate',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'fer'},
                             'heatEmissionRate': {'args': '<type __builtin__.float>',
                                                  'longname': 'heatEmissionRate',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'her'},
                             'maxDistance': {'args': '<type __builtin__.float>',
                                             'longname': 'maxDistance',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'mxd'},
                             'minDistance': {'args': '<type __builtin__.float>',
                                             'longname': 'minDistance',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'mnd'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'position': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'position',
                                          'modes': ['create',
                                                    'query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 3,
                                          'shortname': 'pos'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'rate': {'args': '<type __builtin__.float>',
                                      'longname': 'rate',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'r'},
                             'torusSectionRadius': {'args': '<type __builtin__.float>',
                                                    'longname': 'torusSectionRadius',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'tsr'},
                             'type': {'args': '<type __builtin__.str>',
                                      'longname': 'type',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'typ'},
                             'volumeOffset': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'volumeOffset',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'vof'},
                             'volumeShape': {'args': '<type __builtin__.str>',
                                             'longname': 'volumeShape',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'vsh'},
                             'volumeSweep': {'args': '<type __builtin__.float>',
                                             'longname': 'volumeSweep',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'vsw'}},
                   'resultNeedsUnpacking': True,
                   'shortFlags': {'cye': 'cycleEmission',
                                  'cyi': 'cycleInterval',
                                  'der': 'densityEmissionRate',
                                  'e': 'edit',
                                  'fdr': 'fluidDropoff',
                                  'fer': 'fuelEmissionRate',
                                  'her': 'heatEmissionRate',
                                  'mnd': 'minDistance',
                                  'mxd': 'maxDistance',
                                  'n': 'name',
                                  'pos': 'position',
                                  'q': 'query',
                                  'r': 'rate',
                                  'tsr': 'torusSectionRadius',
                                  'typ': 'type',
                                  'vof': 'volumeOffset',
                                  'vsh': 'volumeShape',
                                  'vsw': 'volumeSweep'},
                   'type': 'effects'},
  'fluidMergeCache': {'type': 'runtime'},
  'fluidMergeCacheOpt': {'type': 'runtime'},
  'fluidReplaceCache': {'type': 'runtime'},
  'fluidReplaceCacheOpt': {'type': 'runtime'},
  'fluidReplaceFrames': {'type': 'runtime'},
  'fluidReplaceFramesOpt': {'type': 'runtime'},
  'fluidVoxelInfo': {'flags': {'checkBounds': {'args': '<type __builtin__.bool>',
                                               'longname': 'checkBounds',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'cb'},
                               'inBounds': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'inBounds',
                                            'modes': ['create'],
                                            'numArgs': 3,
                                            'shortname': 'ib'},
                               'objectSpace': {'args': '<type __builtin__.bool>',
                                               'longname': 'objectSpace',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'os'},
                               'radius': {'args': '<type __builtin__.float>',
                                          'longname': 'radius',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'r'},
                               'voxel': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'voxel',
                                         'modes': ['create'],
                                         'numArgs': 3,
                                         'shortname': 'v'},
                               'voxelCenter': {'args': '<type __builtin__.bool>',
                                               'longname': 'voxelCenter',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'vc'},
                               'xIndex': {'args': '<type __builtin__.int>',
                                          'longname': 'xIndex',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'xi'},
                               'yIndex': {'args': '<type __builtin__.int>',
                                          'longname': 'yIndex',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'yi'},
                               'zIndex': {'args': '<type __builtin__.int>',
                                          'longname': 'zIndex',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'zi'}},
                     'shortFlags': {'cb': 'checkBounds',
                                    'ib': 'inBounds',
                                    'os': 'objectSpace',
                                    'r': 'radius',
                                    'v': 'voxel',
                                    'vc': 'voxelCenter',
                                    'xi': 'xIndex',
                                    'yi': 'yIndex',
                                    'zi': 'zIndex'},
                     'type': 'effects'},
  'flushIdleQueue': {'flags': {'resume': {'args': '<type __builtin__.bool>',
                                          'longname': 'resume',
                                          'numArgs': 0,
                                          'shortname': 'r'}},
                     'shortFlags': {'r': 'resume'},
                     'type': 'other'},
  'flushThumbnailCache': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'flushUndo': {'flags': {}, 'shortFlags': {}, 'type': 'system'},
  'fontAttributes': {'flags': {'faceName': {'args': '<type __builtin__.str>',
                                            'longname': 'faceName',
                                            'numArgs': 1,
                                            'shortname': 'fc'},
                               'font': {'args': '<type __builtin__.str>',
                                        'longname': 'font',
                                        'numArgs': 1,
                                        'shortname': 'fn'},
                               'pitch': {'args': '<type __builtin__.str>',
                                         'longname': 'pitch',
                                         'numArgs': 1,
                                         'shortname': 'p'},
                               'size': {'args': '<type __builtin__.int>',
                                        'longname': 'size',
                                        'numArgs': 1,
                                        'shortname': 'sz'},
                               'style': {'args': '<type __builtin__.str>',
                                         'longname': 'style',
                                         'numArgs': 1,
                                         'shortname': 'st'},
                               'weight': {'args': '<type __builtin__.str>',
                                          'longname': 'weight',
                                          'numArgs': 1,
                                          'shortname': 'wt'}},
                     'shortFlags': {'fc': 'faceName',
                                    'fn': 'font',
                                    'p': 'pitch',
                                    'st': 'style',
                                    'sz': 'size',
                                    'wt': 'weight'},
                     'type': 'other'},
  'fontDialog': {'flags': {'FontList': {'args': '<type __builtin__.bool>',
                                        'longname': 'FontList',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'fl'},
                           'scalable': {'args': '<type __builtin__.bool>',
                                        'longname': 'scalable',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'sc'}},
                 'shortFlags': {'fl': 'FontList', 'sc': 'scalable'},
                 'type': 'windows'},
  'formLayout': {'callbackFlags': ['visibleChangeCommand',
                                   'vcc',
                                   'dragCallback',
                                   'dgc',
                                   'dropCallback',
                                   'dpc'],
                 'flags': {'annotation': {'args': '<type __builtin__.str>',
                                          'longname': 'annotation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ann'},
                           'attachControl': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'attachControl',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 4,
                                             'shortname': 'ac'},
                           'attachForm': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.str>',
                                                   '<type __builtin__.int>'],
                                          'longname': 'attachForm',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 3,
                                          'shortname': 'af'},
                           'attachNone': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'attachNone',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'an'},
                           'attachOppositeControl': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'attachOppositeControl',
                                                     'modes': ['create',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 4,
                                                     'shortname': 'aoc'},
                           'attachOppositeForm': {'args': ['<type __builtin__.str>',
                                                           '<type __builtin__.str>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'attachOppositeForm',
                                                  'modes': ['create',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 3,
                                                  'shortname': 'aof'},
                           'attachPosition': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'attachPosition',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 4,
                                              'shortname': 'ap'},
                           'backgroundColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'backgroundColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'bgc'},
                           'childArray': {'args': '<type __builtin__.bool>',
                                          'longname': 'childArray',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'ca'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'dragCallback': {'args': 'script',
                                            'longname': 'dragCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dgc'},
                           'dropCallback': {'args': 'script',
                                            'longname': 'dropCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dpc'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'enable': {'args': '<type __builtin__.bool>',
                                      'longname': 'enable',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'en'},
                           'enableBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableBackground',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ebg'},
                           'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableKeyboardFocus',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ekf'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'fullPathName': {'args': '<type __builtin__.bool>',
                                            'longname': 'fullPathName',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fpn'},
                           'height': {'args': '<type __builtin__.int>',
                                      'longname': 'height',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                           'highlightColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'highlightColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'hlc'},
                           'isObscured': {'args': '<type __builtin__.bool>',
                                          'longname': 'isObscured',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'io'},
                           'manage': {'args': '<type __builtin__.bool>',
                                      'longname': 'manage',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                           'noBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'noBackground',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nbg'},
                           'numberOfChildren': {'args': '<type __builtin__.bool>',
                                                'longname': 'numberOfChildren',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'nch'},
                           'numberOfDivisions': {'args': '<type __builtin__.int>',
                                                 'longname': 'numberOfDivisions',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'nd'},
                           'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'numberOfPopupMenus',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'npm'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuArray': {'args': '<type __builtin__.bool>',
                                              'longname': 'popupMenuArray',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'pma'},
                           'preventOverride': {'args': '<type __builtin__.bool>',
                                               'longname': 'preventOverride',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'po'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'statusBarMessage': {'args': '<type __builtin__.str>',
                                                'longname': 'statusBarMessage',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sbm'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'},
                           'visible': {'args': '<type __builtin__.bool>',
                                       'longname': 'visible',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vis'},
                           'visibleChangeCommand': {'args': 'script',
                                                    'longname': 'visibleChangeCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'vcc'},
                           'width': {'args': '<type __builtin__.int>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'}},
                 'shortFlags': {'ac': 'attachControl',
                                'af': 'attachForm',
                                'an': 'attachNone',
                                'ann': 'annotation',
                                'aoc': 'attachOppositeControl',
                                'aof': 'attachOppositeForm',
                                'ap': 'attachPosition',
                                'bgc': 'backgroundColor',
                                'ca': 'childArray',
                                'dgc': 'dragCallback',
                                'dpc': 'dropCallback',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'ebg': 'enableBackground',
                                'ekf': 'enableKeyboardFocus',
                                'en': 'enable',
                                'ex': 'exists',
                                'fpn': 'fullPathName',
                                'h': 'height',
                                'hlc': 'highlightColor',
                                'io': 'isObscured',
                                'm': 'manage',
                                'nbg': 'noBackground',
                                'nch': 'numberOfChildren',
                                'nd': 'numberOfDivisions',
                                'npm': 'numberOfPopupMenus',
                                'p': 'parent',
                                'pma': 'popupMenuArray',
                                'po': 'preventOverride',
                                'q': 'query',
                                'sbm': 'statusBarMessage',
                                'ut': 'useTemplate',
                                'vcc': 'visibleChangeCommand',
                                'vis': 'visible',
                                'w': 'width'},
                 'type': 'windows'},
  'format': {'flags': {'stringArg': {'args': '<type __builtin__.str>',
                                     'longname': 'stringArg',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 's'}},
             'shortFlags': {'s': 'stringArg'},
             'type': 'language'},
  'frameBufferName': {'flags': {'autoTruncate': {'args': '<type __builtin__.bool>',
                                                 'longname': 'autoTruncate',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'a'},
                                'camera': {'args': '<type __builtin__.str>',
                                           'longname': 'camera',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'c'},
                                'renderLayer': {'args': '<type __builtin__.str>',
                                                'longname': 'renderLayer',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'l'},
                                'renderPass': {'args': '<type __builtin__.str>',
                                               'longname': 'renderPass',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'p'}},
                      'shortFlags': {'a': 'autoTruncate',
                                     'c': 'camera',
                                     'l': 'renderLayer',
                                     'p': 'renderPass'},
                      'type': 'rendering'},
  'frameLayout': {'callbackFlags': ['visibleChangeCommand',
                                    'vcc',
                                    'dragCallback',
                                    'dgc',
                                    'preExpandCommand',
                                    'pec',
                                    'preCollapseCommand',
                                    'pcc',
                                    'collapseCommand',
                                    'cc',
                                    'expandCommand',
                                    'ec',
                                    'dropCallback',
                                    'dpc'],
                  'flags': {'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'backgroundShade': {'args': '<type __builtin__.bool>',
                                                'longname': 'backgroundShade',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'bgs'},
                            'borderStyle': {'args': '<type __builtin__.str>',
                                            'longname': 'borderStyle',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'bs'},
                            'borderVisible': {'args': '<type __builtin__.bool>',
                                              'longname': 'borderVisible',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'bv'},
                            'childArray': {'args': '<type __builtin__.bool>',
                                           'longname': 'childArray',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'ca'},
                            'collapsable': {'args': '<type __builtin__.bool>',
                                            'longname': 'collapsable',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cll'},
                            'collapse': {'args': '<type __builtin__.bool>',
                                         'longname': 'collapse',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cl'},
                            'collapseCommand': {'args': 'script',
                                                'longname': 'collapseCommand',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cc'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'expandCommand': {'args': 'script',
                                              'longname': 'expandCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ec'},
                            'font': {'args': '<type __builtin__.str>',
                                     'longname': 'font',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'fn'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'label': {'args': '<type __builtin__.str>',
                                      'longname': 'label',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'l'},
                            'labelAlign': {'args': '<type __builtin__.str>',
                                           'longname': 'labelAlign',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'la'},
                            'labelIndent': {'args': '<type __builtin__.int>',
                                            'longname': 'labelIndent',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'li'},
                            'labelVisible': {'args': '<type __builtin__.bool>',
                                             'longname': 'labelVisible',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lv'},
                            'labelWidth': {'args': '<type __builtin__.int>',
                                           'longname': 'labelWidth',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'lw'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'marginHeight': {'args': '<type __builtin__.int>',
                                             'longname': 'marginHeight',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'mh'},
                            'marginWidth': {'args': '<type __builtin__.int>',
                                            'longname': 'marginWidth',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'mw'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfChildren': {'args': '<type __builtin__.bool>',
                                                 'longname': 'numberOfChildren',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'nch'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preCollapseCommand': {'args': 'script',
                                                   'longname': 'preCollapseCommand',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'pcc'},
                            'preExpandCommand': {'args': 'script',
                                                 'longname': 'preExpandCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'pec'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'bgs': 'backgroundShade',
                                 'bs': 'borderStyle',
                                 'bv': 'borderVisible',
                                 'ca': 'childArray',
                                 'cc': 'collapseCommand',
                                 'cl': 'collapse',
                                 'cll': 'collapsable',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ec': 'expandCommand',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'ex': 'exists',
                                 'fn': 'font',
                                 'fpn': 'fullPathName',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'io': 'isObscured',
                                 'l': 'label',
                                 'la': 'labelAlign',
                                 'li': 'labelIndent',
                                 'lv': 'labelVisible',
                                 'lw': 'labelWidth',
                                 'm': 'manage',
                                 'mh': 'marginHeight',
                                 'mw': 'marginWidth',
                                 'nbg': 'noBackground',
                                 'nch': 'numberOfChildren',
                                 'npm': 'numberOfPopupMenus',
                                 'p': 'parent',
                                 'pcc': 'preCollapseCommand',
                                 'pec': 'preExpandCommand',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'sbm': 'statusBarMessage',
                                 'ut': 'useTemplate',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'freeFormFillet': {'flags': {'bias': {'args': '<type __builtin__.float>',
                                        'longname': 'bias',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'b'},
                               'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'depth': {'args': '<type __builtin__.float>',
                                         'longname': 'depth',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'd'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'object': {'args': '<type __builtin__.bool>',
                                          'longname': 'object',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'o'},
                               'polygon': {'args': '<type __builtin__.int>',
                                           'longname': 'polygon',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'po'},
                               'positionTolerance': {'args': '<type __builtin__.float>',
                                                     'longname': 'positionTolerance',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'pt'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'range': {'args': '<type __builtin__.bool>',
                                         'longname': 'range',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'rn'},
                               'tangentTolerance': {'args': '<type __builtin__.float>',
                                                    'longname': 'tangentTolerance',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'tt'}},
                     'shortFlags': {'b': 'bias',
                                    'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'd': 'depth',
                                    'e': 'edit',
                                    'fzn': 'frozen',
                                    'n': 'name',
                                    'nds': 'nodeState',
                                    'o': 'object',
                                    'po': 'polygon',
                                    'pt': 'positionTolerance',
                                    'q': 'query',
                                    'rn': 'range',
                                    'tt': 'tangentTolerance'},
                     'type': 'modeling'},
  'freezeOptions': {'flags': {'displayLayers': {'args': '<type __builtin__.bool>',
                                                'longname': 'displayLayers',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'dl'},
                              'downstream': {'args': '<type __builtin__.str>',
                                             'longname': 'downstream',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'dn'},
                              'explicitPropagation': {'args': '<type __builtin__.bool>',
                                                      'longname': 'explicitPropagation',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'ep'},
                              'invisible': {'args': '<type __builtin__.bool>',
                                            'longname': 'invisible',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'inv'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'referencedNodes': {'args': '<type __builtin__.bool>',
                                                  'longname': 'referencedNodes',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'rn'},
                              'runtimePropagation': {'args': '<type __builtin__.bool>',
                                                     'longname': 'runtimePropagation',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'rt'},
                              'upstream': {'args': '<type __builtin__.str>',
                                           'longname': 'upstream',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'up'}},
                    'shortFlags': {'dl': 'displayLayers',
                                   'dn': 'downstream',
                                   'ep': 'explicitPropagation',
                                   'inv': 'invisible',
                                   'q': 'query',
                                   'rn': 'referencedNodes',
                                   'rt': 'runtimePropagation',
                                   'up': 'upstream'},
                    'type': 'animation'},
  'geomBind': {'flags': {'bindMethod': {'args': '<type __builtin__.int>',
                                        'longname': 'bindMethod',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'bm'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'falloff': {'args': '<type __builtin__.float>',
                                     'longname': 'falloff',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'fo'},
                         'geodesicVoxelParams': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.bool>'],
                                                 'longname': 'geodesicVoxelParams',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 2,
                                                 'shortname': 'gvp'},
                         'maxInfluences': {'args': '<type __builtin__.int>',
                                           'longname': 'maxInfluences',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'mi'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'}},
               'shortFlags': {'bm': 'bindMethod',
                              'e': 'edit',
                              'fo': 'falloff',
                              'gvp': 'geodesicVoxelParams',
                              'mi': 'maxInfluences',
                              'q': 'query'},
               'type': 'animation'},
  'geomToBBox': {'flags': {'bakeAnimation': {'args': '<type __builtin__.bool>',
                                             'longname': 'bakeAnimation',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ba'},
                           'combineMesh': {'args': '<type __builtin__.bool>',
                                           'longname': 'combineMesh',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'cm'},
                           'endTime': {'args': 'time',
                                       'longname': 'endTime',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'et'},
                           'keepOriginal': {'args': '<type __builtin__.bool>',
                                            'longname': 'keepOriginal',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ko'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nameSuffix': {'args': '<type __builtin__.str>',
                                          'longname': 'nameSuffix',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ns'},
                           'sampleBy': {'args': 'time',
                                        'longname': 'sampleBy',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'sb'},
                           'shaderColor': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'shaderColor',
                                           'modes': ['create'],
                                           'numArgs': 3,
                                           'shortname': 'sc'},
                           'single': {'args': '<type __builtin__.bool>',
                                      'longname': 'single',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 's'},
                           'startTime': {'args': 'time',
                                         'longname': 'startTime',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'st'}},
                 'shortFlags': {'ba': 'bakeAnimation',
                                'cm': 'combineMesh',
                                'et': 'endTime',
                                'ko': 'keepOriginal',
                                'n': 'name',
                                'ns': 'nameSuffix',
                                's': 'single',
                                'sb': 'sampleBy',
                                'sc': 'shaderColor',
                                'st': 'startTime'},
                 'type': 'modeling'},
  'geometryAppendCache': {'type': 'runtime'},
  'geometryAppendCacheOpt': {'type': 'runtime'},
  'geometryCache': {'type': 'runtime'},
  'geometryCacheOpt': {'type': 'runtime'},
  'geometryConstraint': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'layer': {'args': '<type __builtin__.str>',
                                             'longname': 'layer',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'l'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'remove': {'args': '<type __builtin__.bool>',
                                              'longname': 'remove',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'rm'},
                                   'targetList': {'args': '<type __builtin__.bool>',
                                                  'longname': 'targetList',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'tl'},
                                   'weight': {'args': '<type __builtin__.float>',
                                              'longname': 'weight',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'w'},
                                   'weightAliasList': {'args': '<type __builtin__.bool>',
                                                       'longname': 'weightAliasList',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'wal'}},
                         'resultNeedsUnpacking': True,
                         'shortFlags': {'e': 'edit',
                                        'n': 'name',
                                        'q': 'query',
                                        'rm': 'remove',
                                        'tl': 'targetList',
                                        'w': 'weight',
                                        'wal': 'weightAliasList'},
                         'type': 'animation'},
  'geometryDeleteCacheFrames': {'type': 'runtime'},
  'geometryDeleteCacheFramesOpt': {'type': 'runtime'},
  'geometryDeleteCacheOpt': {'type': 'runtime'},
  'geometryExportCache': {'type': 'runtime'},
  'geometryExportCacheOpt': {'type': 'runtime'},
  'geometryMergeCache': {'type': 'runtime'},
  'geometryMergeCacheOpt': {'type': 'runtime'},
  'geometryReplaceCache': {'type': 'runtime'},
  'geometryReplaceCacheFrames': {'type': 'runtime'},
  'geometryReplaceCacheFramesOpt': {'type': 'runtime'},
  'geometryReplaceCacheOpt': {'type': 'runtime'},
  'getAttr': {'flags': {'asString': {'args': '<type __builtin__.bool>',
                                     'longname': 'asString',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'asString'},
                        'caching': {'args': '<type __builtin__.bool>',
                                    'longname': 'caching',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ca'},
                        'channelBox': {'args': '<type __builtin__.bool>',
                                       'longname': 'channelBox',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'cb'},
                        'expandEnvironmentVariables': {'args': '<type __builtin__.bool>',
                                                       'longname': 'expandEnvironmentVariables',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'x'},
                        'keyable': {'args': '<type __builtin__.bool>',
                                    'longname': 'keyable',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'k'},
                        'lock': {'args': '<type __builtin__.bool>',
                                 'longname': 'lock',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'l'},
                        'multiIndices': {'args': '<type __builtin__.bool>',
                                         'longname': 'multiIndices',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'mi'},
                        'settable': {'args': '<type __builtin__.bool>',
                                     'longname': 'settable',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'se'},
                        'silent': {'args': '<type __builtin__.bool>',
                                   'longname': 'silent',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'sl'},
                        'size': {'args': '<type __builtin__.bool>',
                                 'longname': 'size',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 's'},
                        'time': {'args': 'time',
                                 'longname': 'time',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 't'},
                        'type': {'args': '<type __builtin__.bool>',
                                 'longname': 'type',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'typ'}},
              'removedFlags': {'as': 'asString'},
              'shortFlags': {'asString': 'asString',
                             'ca': 'caching',
                             'cb': 'channelBox',
                             'k': 'keyable',
                             'l': 'lock',
                             'mi': 'multiIndices',
                             's': 'size',
                             'se': 'settable',
                             'sl': 'silent',
                             't': 'time',
                             'typ': 'type',
                             'x': 'expandEnvironmentVariables'},
              'type': 'general'},
  'getClassification': {'flags': {'satisfies': {'args': '<type __builtin__.str>',
                                                'longname': 'satisfies',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'sat'}},
                        'shortFlags': {'sat': 'satisfies'},
                        'type': 'general'},
  'getDefaultBrush': {'flags': {}, 'shortFlags': {}, 'type': 'effects'},
  'getFileList': {'flags': {'filespec': {'args': '<type __builtin__.str>',
                                         'longname': 'filespec',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'fs'},
                            'folder': {'args': '<type __builtin__.str>',
                                       'longname': 'folder',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'fld'}},
                  'shortFlags': {'fld': 'folder', 'fs': 'filespec'},
                  'type': 'system'},
  'getFluidAttr': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                           'longname': 'attribute',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'at'},
                             'lowerFace': {'args': '<type __builtin__.bool>',
                                           'longname': 'lowerFace',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'lf'},
                             'xIndex': {'args': '<type __builtin__.int>',
                                        'longname': 'xIndex',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'xi'},
                             'xvalue': {'args': '<type __builtin__.bool>',
                                        'longname': 'xvalue',
                                        'modes': [],
                                        'numArgs': 0,
                                        'shortname': 'x'},
                             'yIndex': {'args': '<type __builtin__.int>',
                                        'longname': 'yIndex',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'yi'},
                             'yvalue': {'args': '<type __builtin__.bool>',
                                        'longname': 'yvalue',
                                        'modes': [],
                                        'numArgs': 0,
                                        'shortname': 'y'},
                             'zIndex': {'args': '<type __builtin__.int>',
                                        'longname': 'zIndex',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'zi'},
                             'zvalue': {'args': '<type __builtin__.bool>',
                                        'longname': 'zvalue',
                                        'modes': [],
                                        'numArgs': 0,
                                        'shortname': 'z'}},
                   'shortFlags': {'at': 'attribute',
                                  'lf': 'lowerFace',
                                  'x': 'xvalue',
                                  'xi': 'xIndex',
                                  'y': 'yvalue',
                                  'yi': 'yIndex',
                                  'z': 'zvalue',
                                  'zi': 'zIndex'},
                   'type': 'effects'},
  'getInputDeviceRange': {'flags': {'maxValue': {'args': '<type __builtin__.bool>',
                                                 'longname': 'maxValue',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'max'},
                                    'minValue': {'args': '<type __builtin__.bool>',
                                                 'longname': 'minValue',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'min'}},
                          'shortFlags': {},
                          'type': 'system'},
  'getLastError': {'flags': {}, 'shortFlags': {}, 'type': 'language'},
  'getMetadata': {'flags': {'channelName': {'args': '<type __builtin__.str>',
                                            'longname': 'channelName',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'cn'},
                            'channelType': {'args': '<type __builtin__.str>',
                                            'longname': 'channelType',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'cht'},
                            'dataType': {'args': '<type __builtin__.bool>',
                                         'longname': 'dataType',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'dt'},
                            'endIndex': {'args': '<type __builtin__.str>',
                                         'longname': 'endIndex',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'eix'},
                            'index': {'args': '<type __builtin__.str>',
                                      'longname': 'index',
                                      'modes': ['create',
                                                'query',
                                                'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'idx'},
                            'indexType': {'args': '<type __builtin__.str>',
                                          'longname': 'indexType',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'idt'},
                            'listChannelNames': {'args': '<type __builtin__.bool>',
                                                 'longname': 'listChannelNames',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'lcn'},
                            'listMemberNames': {'args': '<type __builtin__.bool>',
                                                'longname': 'listMemberNames',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'lmn'},
                            'listStreamNames': {'args': '<type __builtin__.bool>',
                                                'longname': 'listStreamNames',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'lsn'},
                            'memberName': {'args': '<type __builtin__.str>',
                                           'longname': 'memberName',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'mn'},
                            'scene': {'args': '<type __builtin__.bool>',
                                      'longname': 'scene',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'scn'},
                            'startIndex': {'args': '<type __builtin__.str>',
                                           'longname': 'startIndex',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'six'},
                            'streamName': {'args': '<type __builtin__.str>',
                                           'longname': 'streamName',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'stn'}},
                  'shortFlags': {'cht': 'channelType',
                                 'cn': 'channelName',
                                 'dt': 'dataType',
                                 'eix': 'endIndex',
                                 'idt': 'indexType',
                                 'idx': 'index',
                                 'lcn': 'listChannelNames',
                                 'lmn': 'listMemberNames',
                                 'lsn': 'listStreamNames',
                                 'mn': 'memberName',
                                 'scn': 'scene',
                                 'six': 'startIndex',
                                 'stn': 'streamName'},
                  'type': 'modeling'},
  'getModifiers': {'flags': {}, 'shortFlags': {}, 'type': 'system'},
  'getModulePath': {'flags': {'moduleName': {'args': '<type __builtin__.str>',
                                             'longname': 'moduleName',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'mn'}},
                    'shortFlags': {'mn': 'moduleName'},
                    'type': 'system'},
  'getPanel': {'flags': {'allConfigs': {'args': '<type __builtin__.bool>',
                                        'longname': 'allConfigs',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ac'},
                         'allPanels': {'args': '<type __builtin__.bool>',
                                       'longname': 'allPanels',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'all'},
                         'allScriptedTypes': {'args': '<type __builtin__.bool>',
                                              'longname': 'allScriptedTypes',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ast'},
                         'allTypes': {'args': '<type __builtin__.bool>',
                                      'longname': 'allTypes',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'at'},
                         'atPosition': {'args': ['<type __builtin__.int>',
                                                 '<type __builtin__.int>'],
                                        'longname': 'atPosition',
                                        'modes': ['create'],
                                        'numArgs': 2,
                                        'shortname': 'ap'},
                         'configWithLabel': {'args': '<type __builtin__.str>',
                                             'longname': 'configWithLabel',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'cwl'},
                         'containing': {'args': '<type __builtin__.str>',
                                        'longname': 'containing',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'c'},
                         'invisiblePanels': {'args': '<type __builtin__.bool>',
                                             'longname': 'invisiblePanels',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'inv'},
                         'scriptType': {'args': '<type __builtin__.str>',
                                        'longname': 'scriptType',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'sty'},
                         'type': {'args': '<type __builtin__.str>',
                                  'longname': 'type',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'typ'},
                         'typeOf': {'args': '<type __builtin__.str>',
                                    'longname': 'typeOf',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'to'},
                         'underPointer': {'args': '<type __builtin__.bool>',
                                          'longname': 'underPointer',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'up'},
                         'visiblePanels': {'args': '<type __builtin__.bool>',
                                           'longname': 'visiblePanels',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'vis'},
                         'withFocus': {'args': '<type __builtin__.bool>',
                                       'longname': 'withFocus',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'wf'},
                         'withLabel': {'args': '<type __builtin__.str>',
                                       'longname': 'withLabel',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'wl'}},
               'shortFlags': {'ac': 'allConfigs',
                              'all': 'allPanels',
                              'ap': 'atPosition',
                              'ast': 'allScriptedTypes',
                              'at': 'allTypes',
                              'c': 'containing',
                              'cwl': 'configWithLabel',
                              'inv': 'invisiblePanels',
                              'sty': 'scriptType',
                              'to': 'typeOf',
                              'typ': 'type',
                              'up': 'underPointer',
                              'vis': 'visiblePanels',
                              'wf': 'withFocus',
                              'wl': 'withLabel'},
               'type': 'windows'},
  'getParticleAttr': {'flags': {'array': {'args': '<type __builtin__.bool>',
                                          'longname': 'array',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'a'},
                                'attribute': {'args': '<type __builtin__.str>',
                                              'longname': 'attribute',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'at'},
                                'object': {'args': '<type __builtin__.str>',
                                           'longname': 'object',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'o'}},
                      'shortFlags': {'a': 'array',
                                     'at': 'attribute',
                                     'o': 'object'},
                      'type': 'effects'},
  'getProcArguments': {'flags': {}, 'shortFlags': {}, 'type': 'language'},
  'getRenderDependencies': {'flags': {},
                            'shortFlags': {},
                            'type': 'rendering'},
  'getRenderTasks': {'flags': {'camera': {'args': '<type __builtin__.str>',
                                          'longname': 'camera',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'c'},
                               'renderLayer': {'args': '<type __builtin__.str>',
                                               'longname': 'renderLayer',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'rl'}},
                     'shortFlags': {'c': 'camera', 'rl': 'renderLayer'},
                     'type': 'rendering'},
  'glRender': {'flags': {'accumBufferPasses': {'args': '<type __builtin__.int>',
                                               'longname': 'accumBufferPasses',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'abp'},
                         'alphaSource': {'args': '<type __builtin__.str>',
                                         'longname': 'alphaSource',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'alphaSource'},
                         'antiAliasMethod': {'args': '<type __builtin__.str>',
                                             'longname': 'antiAliasMethod',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'aam'},
                         'cameraIcons': {'args': '<type __builtin__.bool>',
                                         'longname': 'cameraIcons',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ci'},
                         'clearClr': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'clearClr',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 3,
                                      'shortname': 'cc'},
                         'collisionIcons': {'args': '<type __builtin__.bool>',
                                            'longname': 'collisionIcons',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'coi'},
                         'crossingEffect': {'args': '<type __builtin__.bool>',
                                            'longname': 'crossingEffect',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ce'},
                         'currentFrame': {'args': '<type __builtin__.bool>',
                                          'longname': 'currentFrame',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'cf'},
                         'drawStyle': {'args': '<type __builtin__.str>',
                                       'longname': 'drawStyle',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ds'},
                         'edgeSmoothness': {'args': '<type __builtin__.float>',
                                            'longname': 'edgeSmoothness',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'es'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'emitterIcons': {'args': '<type __builtin__.bool>',
                                          'longname': 'emitterIcons',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ei'},
                         'fieldIcons': {'args': '<type __builtin__.bool>',
                                        'longname': 'fieldIcons',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'fii'},
                         'flipbookCallback': {'args': '<type __builtin__.str>',
                                              'longname': 'flipbookCallback',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fc'},
                         'frameEnd': {'args': '<type __builtin__.int>',
                                      'longname': 'frameEnd',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'fe'},
                         'frameIncrement': {'args': '<type __builtin__.int>',
                                            'longname': 'frameIncrement',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'fi'},
                         'frameStart': {'args': '<type __builtin__.int>',
                                        'longname': 'frameStart',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'fs'},
                         'fullResolution': {'args': '<type __builtin__.bool>',
                                            'longname': 'fullResolution',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'fr'},
                         'grid': {'args': '<type __builtin__.bool>',
                                  'longname': 'grid',
                                  'modes': ['query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'gr'},
                         'imageDirectory': {'args': '<type __builtin__.str>',
                                            'longname': 'imageDirectory',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'id'},
                         'imageName': {'args': '<type __builtin__.str>',
                                       'longname': 'imageName',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'imageName'},
                         'imageSize': {'args': ['<type __builtin__.int>',
                                                '<type __builtin__.int>',
                                                '<type __builtin__.float>'],
                                       'longname': 'imageSize',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 3,
                                       'shortname': 'imageSize'},
                         'lightIcons': {'args': '<type __builtin__.bool>',
                                        'longname': 'lightIcons',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'li'},
                         'lightingMode': {'args': '<type __builtin__.str>',
                                          'longname': 'lightingMode',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'lm'},
                         'lineSmoothing': {'args': '<type __builtin__.bool>',
                                           'longname': 'lineSmoothing',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ls'},
                         'offScreen': {'args': '<type __builtin__.bool>',
                                       'longname': 'offScreen',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'os'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'renderFrame': {'args': '<type __builtin__.str>',
                                         'longname': 'renderFrame',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'rf'},
                         'renderSequence': {'args': '<type __builtin__.str>',
                                            'longname': 'renderSequence',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'rs'},
                         'sharpness': {'args': '<type __builtin__.float>',
                                       'longname': 'sharpness',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sh'},
                         'shutterAngle': {'args': '<type __builtin__.float>',
                                          'longname': 'shutterAngle',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sa'},
                         'textureDisplay': {'args': '<type __builtin__.bool>',
                                            'longname': 'textureDisplay',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'txd'},
                         'transformIcons': {'args': '<type __builtin__.bool>',
                                            'longname': 'transformIcons',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ti'},
                         'useAccumBuffer': {'args': '<type __builtin__.bool>',
                                            'longname': 'useAccumBuffer',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'uab'},
                         'viewport': {'args': ['<type __builtin__.int>',
                                               '<type __builtin__.int>',
                                               '<type __builtin__.float>'],
                                      'longname': 'viewport',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 3,
                                      'shortname': 'vp'},
                         'writeDepthMap': {'args': '<type __builtin__.bool>',
                                           'longname': 'writeDepthMap',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'wdm'}},
               'removedFlags': {'as': 'alphaSource',
                                'in': 'imageName',
                                'is': 'imageSize'},
               'shortFlags': {'aam': 'antiAliasMethod',
                              'abp': 'accumBufferPasses',
                              'alphaSource': 'alphaSource',
                              'cc': 'clearClr',
                              'ce': 'crossingEffect',
                              'cf': 'currentFrame',
                              'ci': 'cameraIcons',
                              'coi': 'collisionIcons',
                              'ds': 'drawStyle',
                              'e': 'edit',
                              'ei': 'emitterIcons',
                              'es': 'edgeSmoothness',
                              'fc': 'flipbookCallback',
                              'fe': 'frameEnd',
                              'fi': 'frameIncrement',
                              'fii': 'fieldIcons',
                              'fr': 'fullResolution',
                              'fs': 'frameStart',
                              'gr': 'grid',
                              'id': 'imageDirectory',
                              'imageName': 'imageName',
                              'imageSize': 'imageSize',
                              'li': 'lightIcons',
                              'lm': 'lightingMode',
                              'ls': 'lineSmoothing',
                              'os': 'offScreen',
                              'q': 'query',
                              'rf': 'renderFrame',
                              'rs': 'renderSequence',
                              'sa': 'shutterAngle',
                              'sh': 'sharpness',
                              'ti': 'transformIcons',
                              'txd': 'textureDisplay',
                              'uab': 'useAccumBuffer',
                              'vp': 'viewport',
                              'wdm': 'writeDepthMap'},
               'type': 'rendering'},
  'glRenderEditor': {'flags': {'control': {'args': '<type __builtin__.bool>',
                                           'longname': 'control',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'ctl'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'filter': {'args': '<type __builtin__.str>',
                                          'longname': 'filter',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'f'},
                               'forceMainConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'forceMainConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'fmc'},
                               'highlightConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'highlightConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'hlc'},
                               'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                      'longname': 'lockMainConnection',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'lck'},
                               'lookThru': {'args': '<type __builtin__.str>',
                                            'longname': 'lookThru',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'lt'},
                               'mainListConnection': {'args': '<type __builtin__.str>',
                                                      'longname': 'mainListConnection',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'mlc'},
                               'panel': {'args': '<type __builtin__.str>',
                                         'longname': 'panel',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'pnl'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'selectionConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'selectionConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'slc'},
                               'stateString': {'args': '<type __builtin__.bool>',
                                               'longname': 'stateString',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'sts'},
                               'unParent': {'args': '<type __builtin__.bool>',
                                            'longname': 'unParent',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'up'},
                               'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                        'longname': 'unlockMainConnection',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'ulk'},
                               'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                        'longname': 'updateMainConnection',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'upd'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'},
                               'viewCameraName': {'args': '<type __builtin__.bool>',
                                                  'longname': 'viewCameraName',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'vcn'}},
                     'shortFlags': {'ctl': 'control',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'f': 'filter',
                                    'fmc': 'forceMainConnection',
                                    'hlc': 'highlightConnection',
                                    'lck': 'lockMainConnection',
                                    'lt': 'lookThru',
                                    'mlc': 'mainListConnection',
                                    'p': 'parent',
                                    'pnl': 'panel',
                                    'q': 'query',
                                    'slc': 'selectionConnection',
                                    'sts': 'stateString',
                                    'ulk': 'unlockMainConnection',
                                    'up': 'unParent',
                                    'upd': 'updateMainConnection',
                                    'ut': 'useTemplate',
                                    'vcn': 'viewCameraName'},
                     'type': 'rendering'},
  'globalStitch': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'lockSurface': {'args': '<type __builtin__.bool>',
                                             'longname': 'lockSurface',
                                             'modes': ['create',
                                                       'query',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'lk'},
                             'maxSeparation': {'args': '<type __builtin__.float>',
                                               'longname': 'maxSeparation',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ms'},
                             'modificationResistance': {'args': '<type __builtin__.float>',
                                                        'longname': 'modificationResistance',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'mr'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'object': {'args': '<type __builtin__.bool>',
                                        'longname': 'object',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                 'longname': 'replaceOriginal',
                                                 'numArgs': 1,
                                                 'shortname': 'rpo'},
                             'sampling': {'args': '<type __builtin__.int>',
                                          'longname': 'sampling',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sam'},
                             'stitchCorners': {'args': '<type __builtin__.int>',
                                               'longname': 'stitchCorners',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sc'},
                             'stitchEdges': {'args': '<type __builtin__.int>',
                                             'longname': 'stitchEdges',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'se'},
                             'stitchPartialEdges': {'args': '<type __builtin__.bool>',
                                                    'longname': 'stitchPartialEdges',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'spe'},
                             'stitchSmoothness': {'args': '<type __builtin__.int>',
                                                  'longname': 'stitchSmoothness',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ss'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'lk': 'lockSurface',
                                  'mr': 'modificationResistance',
                                  'ms': 'maxSeparation',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'o': 'object',
                                  'q': 'query',
                                  'rpo': 'replaceOriginal',
                                  'sam': 'sampling',
                                  'sc': 'stitchCorners',
                                  'se': 'stitchEdges',
                                  'spe': 'stitchPartialEdges',
                                  'ss': 'stitchSmoothness'},
                   'type': 'modeling'},
  'goal': {'flags': {'goal': {'args': '<type __builtin__.str>',
                              'longname': 'goal',
                              'modes': ['create', 'query', 'multiuse'],
                              'numArgs': 1,
                              'shortname': 'g'},
                     'index': {'args': '<type __builtin__.bool>',
                               'longname': 'index',
                               'modes': ['query'],
                               'numArgs': 0,
                               'shortname': 'i'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'useTransformAsGoal': {'args': '<type __builtin__.bool>',
                                            'longname': 'useTransformAsGoal',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'utr'},
                     'weight': {'args': '<type __builtin__.float>',
                                'longname': 'weight',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'w'}},
           'shortFlags': {'g': 'goal',
                          'i': 'index',
                          'q': 'query',
                          'utr': 'useTransformAsGoal',
                          'w': 'weight'},
           'type': 'effects'},
  'grabColor': {'flags': {'hsvValue': {'args': '<type __builtin__.bool>',
                                       'longname': 'hsvValue',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'hsv'},
                          'rgbValue': {'args': '<type __builtin__.bool>',
                                       'longname': 'rgbValue',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'rgb'}},
                'shortFlags': {'hsv': 'hsvValue', 'rgb': 'rgbValue'},
                'type': 'windows'},
  'gradientControl': {'callbackFlags': ['visibleChangeCommand',
                                        'vcc',
                                        'dragCallback',
                                        'dgc',
                                        'dropCallback',
                                        'dpc'],
                      'flags': {'adaptiveScaling': {'args': '<type __builtin__.bool>',
                                                    'longname': 'adaptiveScaling',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'adaptiveScaling'},
                                'annotation': {'args': '<type __builtin__.str>',
                                               'longname': 'annotation',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ann'},
                                'attribute': {'args': 'PyNode',
                                              'longname': 'attribute',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'at'},
                                'backgroundColor': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'backgroundColor',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'bgc'},
                                'defineTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'defineTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'dt'},
                                'docTag': {'args': '<type __builtin__.str>',
                                           'longname': 'docTag',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dtg'},
                                'dragCallback': {'args': 'script',
                                                 'longname': 'dragCallback',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dgc'},
                                'dropCallback': {'args': 'script',
                                                 'longname': 'dropCallback',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dpc'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'enable': {'args': '<type __builtin__.bool>',
                                           'longname': 'enable',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'en'},
                                'enableBackground': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableBackground',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ebg'},
                                'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                        'longname': 'enableKeyboardFocus',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ekf'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'fullPathName': {'args': '<type __builtin__.bool>',
                                                 'longname': 'fullPathName',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'fpn'},
                                'height': {'args': '<type __builtin__.int>',
                                           'longname': 'height',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'h'},
                                'highlightColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'highlightColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'hlc'},
                                'isObscured': {'args': '<type __builtin__.bool>',
                                               'longname': 'isObscured',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'io'},
                                'manage': {'args': '<type __builtin__.bool>',
                                           'longname': 'manage',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'm'},
                                'noBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'noBackground',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'nbg'},
                                'numberOfControls': {'args': '<type __builtin__.int>',
                                                     'longname': 'numberOfControls',
                                                     'modes': ['query'],
                                                     'numArgs': 1,
                                                     'shortname': 'nc'},
                                'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                       'longname': 'numberOfPopupMenus',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'npm'},
                                'parent': {'args': '<type __builtin__.str>',
                                           'longname': 'parent',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'p'},
                                'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                   'longname': 'popupMenuArray',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'pma'},
                                'preventOverride': {'args': '<type __builtin__.bool>',
                                                    'longname': 'preventOverride',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'po'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'refreshOnRelease': {'args': '<type __builtin__.int>',
                                                     'longname': 'refreshOnRelease',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ror'},
                                'selectedColorControl': {'args': '<type __builtin__.str>',
                                                         'longname': 'selectedColorControl',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'scc'},
                                'selectedInterpControl': {'args': '<type __builtin__.str>',
                                                          'longname': 'selectedInterpControl',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'sic'},
                                'selectedPositionControl': {'args': '<type __builtin__.str>',
                                                            'longname': 'selectedPositionControl',
                                                            'modes': ['create',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'spc'},
                                'staticNumberOfControls': {'args': '<type __builtin__.bool>',
                                                           'longname': 'staticNumberOfControls',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'snc'},
                                'staticPositions': {'args': '<type __builtin__.bool>',
                                                    'longname': 'staticPositions',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sp'},
                                'statusBarMessage': {'args': '<type __builtin__.str>',
                                                     'longname': 'statusBarMessage',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sbm'},
                                'upperLimitControl': {'args': '<type __builtin__.str>',
                                                      'longname': 'upperLimitControl',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ulc'},
                                'useTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'useTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ut'},
                                'verticalLayout': {'args': '<type __builtin__.bool>',
                                                   'longname': 'verticalLayout',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'vl'},
                                'visible': {'args': '<type __builtin__.bool>',
                                            'longname': 'visible',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'vis'},
                                'visibleChangeCommand': {'args': 'script',
                                                         'longname': 'visibleChangeCommand',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'vcc'},
                                'width': {'args': '<type __builtin__.int>',
                                          'longname': 'width',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'w'}},
                      'removedFlags': {'as': 'adaptiveScaling'},
                      'shortFlags': {'adaptiveScaling': 'adaptiveScaling',
                                     'ann': 'annotation',
                                     'at': 'attribute',
                                     'bgc': 'backgroundColor',
                                     'dgc': 'dragCallback',
                                     'dpc': 'dropCallback',
                                     'dt': 'defineTemplate',
                                     'dtg': 'docTag',
                                     'e': 'edit',
                                     'ebg': 'enableBackground',
                                     'ekf': 'enableKeyboardFocus',
                                     'en': 'enable',
                                     'ex': 'exists',
                                     'fpn': 'fullPathName',
                                     'h': 'height',
                                     'hlc': 'highlightColor',
                                     'io': 'isObscured',
                                     'm': 'manage',
                                     'nbg': 'noBackground',
                                     'nc': 'numberOfControls',
                                     'npm': 'numberOfPopupMenus',
                                     'p': 'parent',
                                     'pma': 'popupMenuArray',
                                     'po': 'preventOverride',
                                     'q': 'query',
                                     'ror': 'refreshOnRelease',
                                     'sbm': 'statusBarMessage',
                                     'scc': 'selectedColorControl',
                                     'sic': 'selectedInterpControl',
                                     'snc': 'staticNumberOfControls',
                                     'sp': 'staticPositions',
                                     'spc': 'selectedPositionControl',
                                     'ulc': 'upperLimitControl',
                                     'ut': 'useTemplate',
                                     'vcc': 'visibleChangeCommand',
                                     'vis': 'visible',
                                     'vl': 'verticalLayout',
                                     'w': 'width'},
                      'type': 'windows'},
  'gradientControlNoAttr': {'callbackFlags': ['dragCommand',
                                              'dc',
                                              'dropCallback',
                                              'dpc',
                                              'visibleChangeCommand',
                                              'vcc',
                                              'dragCallback',
                                              'dgc',
                                              'changeCommand',
                                              'cc',
                                              'currentKeyChanged',
                                              'ckc'],
                            'flags': {'annotation': {'args': '<type __builtin__.str>',
                                                     'longname': 'annotation',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ann'},
                                      'asString': {'args': '<type __builtin__.str>',
                                                   'longname': 'asString',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'asString'},
                                      'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                   '<type __builtin__.float>',
                                                                   '<type __builtin__.float>'],
                                                          'longname': 'backgroundColor',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 3,
                                                          'shortname': 'bgc'},
                                      'changeCommand': {'args': 'script',
                                                        'longname': 'changeCommand',
                                                        'modes': ['edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'cc'},
                                      'currentKey': {'args': '<type __builtin__.int>',
                                                     'longname': 'currentKey',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ck'},
                                      'currentKeyChanged': {'args': 'script',
                                                            'longname': 'currentKeyChanged',
                                                            'modes': ['edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ckc'},
                                      'currentKeyColorValue': {'args': ['<type __builtin__.float>',
                                                                        '<type __builtin__.float>',
                                                                        '<type __builtin__.float>'],
                                                               'longname': 'currentKeyColorValue',
                                                               'modes': ['query',
                                                                         'edit'],
                                                               'numArgs': 3,
                                                               'shortname': 'clv'},
                                      'currentKeyCurveValue': {'args': '<type __builtin__.bool>',
                                                               'longname': 'currentKeyCurveValue',
                                                               'modes': ['query',
                                                                         'edit'],
                                                               'numArgs': 0,
                                                               'shortname': 'cvv'},
                                      'currentKeyInterpValue': {'args': '<type __builtin__.int>',
                                                                'longname': 'currentKeyInterpValue',
                                                                'modes': ['query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'civ'},
                                      'defineTemplate': {'args': '<type __builtin__.str>',
                                                         'longname': 'defineTemplate',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'dt'},
                                      'docTag': {'args': '<type __builtin__.str>',
                                                 'longname': 'docTag',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dtg'},
                                      'dragCallback': {'args': 'script',
                                                       'longname': 'dragCallback',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'dgc'},
                                      'dragCommand': {'args': 'script',
                                                      'longname': 'dragCommand',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dc'},
                                      'dropCallback': {'args': 'script',
                                                       'longname': 'dropCallback',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'dpc'},
                                      'edit': {'args': '<type __builtin__.bool>',
                                               'longname': 'edit',
                                               'numArgs': 0,
                                               'shortname': 'e'},
                                      'enable': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enable',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'en'},
                                      'enableBackground': {'args': '<type __builtin__.bool>',
                                                           'longname': 'enableBackground',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ebg'},
                                      'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                              'longname': 'enableKeyboardFocus',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'ekf'},
                                      'exists': {'args': '<type __builtin__.bool>',
                                                 'longname': 'exists',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ex'},
                                      'fullPathName': {'args': '<type __builtin__.bool>',
                                                       'longname': 'fullPathName',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'fpn'},
                                      'height': {'args': '<type __builtin__.int>',
                                                 'longname': 'height',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'h'},
                                      'highlightColor': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'highlightColor',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 3,
                                                         'shortname': 'hlc'},
                                      'isObscured': {'args': '<type __builtin__.bool>',
                                                     'longname': 'isObscured',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'io'},
                                      'manage': {'args': '<type __builtin__.bool>',
                                                 'longname': 'manage',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'm'},
                                      'noBackground': {'args': '<type __builtin__.bool>',
                                                       'longname': 'noBackground',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'nbg'},
                                      'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                             'longname': 'numberOfPopupMenus',
                                                             'modes': ['query'],
                                                             'numArgs': 0,
                                                             'shortname': 'npm'},
                                      'optionVar': {'args': '<type __builtin__.str>',
                                                    'longname': 'optionVar',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ov'},
                                      'parent': {'args': '<type __builtin__.str>',
                                                 'longname': 'parent',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'p'},
                                      'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                         'longname': 'popupMenuArray',
                                                         'modes': ['query'],
                                                         'numArgs': 0,
                                                         'shortname': 'pma'},
                                      'preventOverride': {'args': '<type __builtin__.bool>',
                                                          'longname': 'preventOverride',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'po'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'rampAsColor': {'args': '<type __builtin__.bool>',
                                                      'longname': 'rampAsColor',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'rac'},
                                      'statusBarMessage': {'args': '<type __builtin__.str>',
                                                           'longname': 'statusBarMessage',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'sbm'},
                                      'useTemplate': {'args': '<type __builtin__.str>',
                                                      'longname': 'useTemplate',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ut'},
                                      'valueAtPoint': {'args': '<type __builtin__.float>',
                                                       'longname': 'valueAtPoint',
                                                       'modes': ['query'],
                                                       'numArgs': 1,
                                                       'shortname': 'vap'},
                                      'visible': {'args': '<type __builtin__.bool>',
                                                  'longname': 'visible',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'vis'},
                                      'visibleChangeCommand': {'args': 'script',
                                                               'longname': 'visibleChangeCommand',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'vcc'},
                                      'width': {'args': '<type __builtin__.int>',
                                                'longname': 'width',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'w'}},
                            'removedFlags': {'as': 'asString'},
                            'shortFlags': {'ann': 'annotation',
                                           'asString': 'asString',
                                           'bgc': 'backgroundColor',
                                           'cc': 'changeCommand',
                                           'civ': 'currentKeyInterpValue',
                                           'ck': 'currentKey',
                                           'ckc': 'currentKeyChanged',
                                           'clv': 'currentKeyColorValue',
                                           'cvv': 'currentKeyCurveValue',
                                           'dc': 'dragCommand',
                                           'dgc': 'dragCallback',
                                           'dpc': 'dropCallback',
                                           'dt': 'defineTemplate',
                                           'dtg': 'docTag',
                                           'e': 'edit',
                                           'ebg': 'enableBackground',
                                           'ekf': 'enableKeyboardFocus',
                                           'en': 'enable',
                                           'ex': 'exists',
                                           'fpn': 'fullPathName',
                                           'h': 'height',
                                           'hlc': 'highlightColor',
                                           'io': 'isObscured',
                                           'm': 'manage',
                                           'nbg': 'noBackground',
                                           'npm': 'numberOfPopupMenus',
                                           'ov': 'optionVar',
                                           'p': 'parent',
                                           'pma': 'popupMenuArray',
                                           'po': 'preventOverride',
                                           'q': 'query',
                                           'rac': 'rampAsColor',
                                           'sbm': 'statusBarMessage',
                                           'ut': 'useTemplate',
                                           'vap': 'valueAtPoint',
                                           'vcc': 'visibleChangeCommand',
                                           'vis': 'visible',
                                           'w': 'width'},
                            'type': 'windows'},
  'graphDollyCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'history': {'args': '<type __builtin__.bool>',
                                          'longname': 'history',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ch'},
                              'image1': {'args': '<type __builtin__.str>',
                                         'longname': 'image1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i1'},
                              'image2': {'args': '<type __builtin__.str>',
                                         'longname': 'image2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i2'},
                              'image3': {'args': '<type __builtin__.str>',
                                         'longname': 'image3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i3'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'}},
                    'shortFlags': {'ch': 'history',
                                   'e': 'edit',
                                   'ex': 'exists',
                                   'i1': 'image1',
                                   'i2': 'image2',
                                   'i3': 'image3',
                                   'n': 'name',
                                   'q': 'query'},
                    'type': 'context'},
  'graphSelectContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'}},
                         'shortFlags': {'e': 'edit',
                                        'ex': 'exists',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'q': 'query'},
                         'type': 'context'},
  'graphTrackCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'history': {'args': '<type __builtin__.bool>',
                                          'longname': 'history',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ch'},
                              'image1': {'args': '<type __builtin__.str>',
                                         'longname': 'image1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i1'},
                              'image2': {'args': '<type __builtin__.str>',
                                         'longname': 'image2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i2'},
                              'image3': {'args': '<type __builtin__.str>',
                                         'longname': 'image3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i3'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'}},
                    'shortFlags': {'ch': 'history',
                                   'e': 'edit',
                                   'ex': 'exists',
                                   'i1': 'image1',
                                   'i2': 'image2',
                                   'i3': 'image3',
                                   'n': 'name',
                                   'q': 'query'},
                    'type': 'context'},
  'gravity': {'flags': {'attenuation': {'args': '<type __builtin__.float>',
                                        'longname': 'attenuation',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'att'},
                        'directionX': {'args': '<type __builtin__.float>',
                                       'longname': 'directionX',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dx'},
                        'directionY': {'args': '<type __builtin__.float>',
                                       'longname': 'directionY',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dy'},
                        'directionZ': {'args': '<type __builtin__.float>',
                                       'longname': 'directionZ',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dz'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'magnitude': {'args': '<type __builtin__.float>',
                                      'longname': 'magnitude',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                        'maxDistance': {'args': '<type __builtin__.float>',
                                        'longname': 'maxDistance',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'mxd'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'perVertex': {'args': '<type __builtin__.bool>',
                                      'longname': 'perVertex',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'pv'},
                        'position': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'position',
                                     'modes': ['query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 3,
                                     'shortname': 'pos'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'torusSectionRadius': {'args': '<type __builtin__.float>',
                                               'longname': 'torusSectionRadius',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tsr'},
                        'volumeExclusion': {'args': '<type __builtin__.bool>',
                                            'longname': 'volumeExclusion',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'vex'},
                        'volumeOffset': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'volumeOffset',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 3,
                                         'shortname': 'vof'},
                        'volumeShape': {'args': '<type __builtin__.str>',
                                        'longname': 'volumeShape',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vsh'},
                        'volumeSweep': {'args': '<type __builtin__.float>',
                                        'longname': 'volumeSweep',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vsw'}},
              'resultNeedsUnpacking': True,
              'shortFlags': {'att': 'attenuation',
                             'dx': 'directionX',
                             'dy': 'directionY',
                             'dz': 'directionZ',
                             'e': 'edit',
                             'm': 'magnitude',
                             'mxd': 'maxDistance',
                             'n': 'name',
                             'pos': 'position',
                             'pv': 'perVertex',
                             'q': 'query',
                             'tsr': 'torusSectionRadius',
                             'vex': 'volumeExclusion',
                             'vof': 'volumeOffset',
                             'vsh': 'volumeShape',
                             'vsw': 'volumeSweep'},
              'type': 'effects'},
  'greasePencil': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'greasePencilCtx': {'flags': {'autoCreateFrames': {'args': '<type __builtin__.bool>',
                                                     'longname': 'autoCreateFrames',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'acf'},
                                'canDraw': {'args': '<type __builtin__.bool>',
                                            'longname': 'canDraw',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'cd'},
                                'createOrEditFrame': {'args': '<type __builtin__.int>',
                                                      'longname': 'createOrEditFrame',
                                                      'modes': ['query'],
                                                      'numArgs': 1,
                                                      'shortname': 'cef'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'exportArchive': {'args': ['<type __builtin__.str>',
                                                           '<type __builtin__.str>'],
                                                  'longname': 'exportArchive',
                                                  'modes': ['edit'],
                                                  'numArgs': 2,
                                                  'shortname': 'eac'},
                                'fileTextureSize': {'args': '<type __builtin__.int>',
                                                    'longname': 'fileTextureSize',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'fts'},
                                'greasePencilType': {'args': '<type __builtin__.int>',
                                                     'longname': 'greasePencilType',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'gpt'},
                                'image1': {'args': '<type __builtin__.str>',
                                           'longname': 'image1',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i1'},
                                'image2': {'args': '<type __builtin__.str>',
                                           'longname': 'image2',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i2'},
                                'image3': {'args': '<type __builtin__.str>',
                                           'longname': 'image3',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i3'},
                                'importArchive': {'args': '<type __builtin__.str>',
                                                  'longname': 'importArchive',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'iac'},
                                'makeStroke': {'args': '<type __builtin__.int>',
                                               'longname': 'makeStroke',
                                               'modes': ['create',
                                                         'query',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'mst'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'removeFrame': {'args': '<type __builtin__.int>',
                                                'longname': 'removeFrame',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'rf'},
                                'resetBrushes': {'args': '<type __builtin__.bool>',
                                                 'longname': 'resetBrushes',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rb'},
                                'rgbcolor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'rgbcolor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'rgb'},
                                'sequenceNodeName': {'args': '<type __builtin__.str>',
                                                     'longname': 'sequenceNodeName',
                                                     'modes': ['query'],
                                                     'numArgs': 1,
                                                     'shortname': 'snn'}},
                      'shortFlags': {'acf': 'autoCreateFrames',
                                     'cd': 'canDraw',
                                     'cef': 'createOrEditFrame',
                                     'e': 'edit',
                                     'eac': 'exportArchive',
                                     'ex': 'exists',
                                     'fts': 'fileTextureSize',
                                     'gpt': 'greasePencilType',
                                     'i1': 'image1',
                                     'i2': 'image2',
                                     'i3': 'image3',
                                     'iac': 'importArchive',
                                     'mst': 'makeStroke',
                                     'q': 'query',
                                     'rb': 'resetBrushes',
                                     'rf': 'removeFrame',
                                     'rgb': 'rgbcolor',
                                     'snn': 'sequenceNodeName'},
                      'type': 'context'},
  'greasePencilHelper': {'flags': {'brushType': {'args': '<type __builtin__.int>',
                                                 'longname': 'brushType',
                                                 'numArgs': 1,
                                                 'shortname': 'bt'},
                                   'contextName': {'args': '<type __builtin__.str>',
                                                   'longname': 'contextName',
                                                   'numArgs': 1,
                                                   'shortname': 'cn'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'setColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'setColor',
                                                'numArgs': 3,
                                                'shortname': 'sc'},
                                   'updateVar': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'updateVar',
                                                 'numArgs': 2,
                                                 'shortname': 'uv'}},
                         'shortFlags': {'bt': 'brushType',
                                        'cn': 'contextName',
                                        'e': 'edit',
                                        'q': 'query',
                                        'sc': 'setColor',
                                        'uv': 'updateVar'},
                         'type': 'other'},
  'greaseRenderPlane': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'axis',
                                           'numArgs': 3,
                                           'shortname': 'ax'},
                                  'caching': {'args': '<type __builtin__.bool>',
                                              'longname': 'caching',
                                              'numArgs': 1,
                                              'shortname': 'cch'},
                                  'constructionHistory': {'args': '<type __builtin__.bool>',
                                                          'longname': 'constructionHistory',
                                                          'numArgs': 1,
                                                          'shortname': 'ch'},
                                  'createUVs': {'args': '<type __builtin__.int>',
                                                'longname': 'createUVs',
                                                'numArgs': 1,
                                                'shortname': 'cuv'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'frozen': {'args': '<type __builtin__.bool>',
                                             'longname': 'frozen',
                                             'numArgs': 1,
                                             'shortname': 'fzn'},
                                  'height': {'args': '<type __builtin__.float>',
                                             'longname': 'height',
                                             'numArgs': 1,
                                             'shortname': 'h'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'nodeState': {'args': '<type __builtin__.int>',
                                                'longname': 'nodeState',
                                                'numArgs': 1,
                                                'shortname': 'nds'},
                                  'object': {'args': '<type __builtin__.bool>',
                                             'longname': 'object',
                                             'numArgs': 1,
                                             'shortname': 'o'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                         'longname': 'subdivisionsHeight',
                                                         'numArgs': 1,
                                                         'shortname': 'sh'},
                                  'subdivisionsWidth': {'args': '<type __builtin__.int>',
                                                        'longname': 'subdivisionsWidth',
                                                        'numArgs': 1,
                                                        'shortname': 'sw'},
                                  'subdivisionsX': {'args': '<type __builtin__.int>',
                                                    'longname': 'subdivisionsX',
                                                    'numArgs': 1,
                                                    'shortname': 'sx'},
                                  'subdivisionsY': {'args': '<type __builtin__.int>',
                                                    'longname': 'subdivisionsY',
                                                    'numArgs': 1,
                                                    'shortname': 'sy'},
                                  'texture': {'args': '<type __builtin__.int>',
                                              'longname': 'texture',
                                              'numArgs': 1,
                                              'shortname': 'tx'},
                                  'width': {'args': '<type __builtin__.float>',
                                            'longname': 'width',
                                            'numArgs': 1,
                                            'shortname': 'w'}},
                        'shortFlags': {'ax': 'axis',
                                       'cch': 'caching',
                                       'ch': 'constructionHistory',
                                       'cuv': 'createUVs',
                                       'e': 'edit',
                                       'fzn': 'frozen',
                                       'h': 'height',
                                       'n': 'name',
                                       'nds': 'nodeState',
                                       'o': 'object',
                                       'q': 'query',
                                       'sh': 'subdivisionsHeight',
                                       'sw': 'subdivisionsWidth',
                                       'sx': 'subdivisionsX',
                                       'sy': 'subdivisionsY',
                                       'tx': 'texture',
                                       'w': 'width'},
                        'type': 'other'},
  'grid': {'flags': {'default': {'args': '<type __builtin__.bool>',
                                 'longname': 'default',
                                 'modes': ['query'],
                                 'numArgs': 0,
                                 'shortname': 'df'},
                     'displayAxes': {'args': '<type __builtin__.bool>',
                                     'longname': 'displayAxes',
                                     'modes': ['query'],
                                     'numArgs': 1,
                                     'shortname': 'da'},
                     'displayAxesBold': {'args': '<type __builtin__.bool>',
                                         'longname': 'displayAxesBold',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'dab'},
                     'displayDivisionLines': {'args': '<type __builtin__.bool>',
                                              'longname': 'displayDivisionLines',
                                              'modes': ['query'],
                                              'numArgs': 1,
                                              'shortname': 'ddl'},
                     'displayGridLines': {'args': '<type __builtin__.bool>',
                                          'longname': 'displayGridLines',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'dgl'},
                     'displayOrthographicLabels': {'args': '<type __builtin__.bool>',
                                                   'longname': 'displayOrthographicLabels',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'dol'},
                     'displayPerspectiveLabels': {'args': '<type __builtin__.bool>',
                                                  'longname': 'displayPerspectiveLabels',
                                                  'modes': ['query'],
                                                  'numArgs': 1,
                                                  'shortname': 'dpl'},
                     'divisions': {'args': '<type __builtin__.int>',
                                   'longname': 'divisions',
                                   'modes': ['query'],
                                   'numArgs': 1,
                                   'shortname': 'd'},
                     'orthographicLabelPosition': {'args': '<type __builtin__.str>',
                                                   'longname': 'orthographicLabelPosition',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'olp'},
                     'perspectiveLabelPosition': {'args': '<type __builtin__.str>',
                                                  'longname': 'perspectiveLabelPosition',
                                                  'modes': ['query'],
                                                  'numArgs': 1,
                                                  'shortname': 'plp'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'reset': {'args': '<type __builtin__.bool>',
                               'longname': 'reset',
                               'modes': [],
                               'numArgs': 0,
                               'shortname': 'r'},
                     'size': {'args': '<type __builtin__.float>',
                              'longname': 'size',
                              'modes': ['query'],
                              'numArgs': 1,
                              'shortname': 's'},
                     'spacing': {'args': '<type __builtin__.float>',
                                 'longname': 'spacing',
                                 'modes': ['query'],
                                 'numArgs': 1,
                                 'shortname': 'sp'},
                     'style': {'args': '<type __builtin__.int>',
                               'longname': 'style',
                               'modes': ['query'],
                               'numArgs': 1,
                               'shortname': 'st'},
                     'toggle': {'args': '<type __builtin__.bool>',
                                'longname': 'toggle',
                                'modes': ['query'],
                                'numArgs': 1,
                                'shortname': 'tgl'}},
           'shortFlags': {'d': 'divisions',
                          'da': 'displayAxes',
                          'dab': 'displayAxesBold',
                          'ddl': 'displayDivisionLines',
                          'df': 'default',
                          'dgl': 'displayGridLines',
                          'dol': 'displayOrthographicLabels',
                          'dpl': 'displayPerspectiveLabels',
                          'olp': 'orthographicLabelPosition',
                          'plp': 'perspectiveLabelPosition',
                          'q': 'query',
                          'r': 'reset',
                          's': 'size',
                          'sp': 'spacing',
                          'st': 'style',
                          'tgl': 'toggle'},
           'type': 'modeling'},
  'gridLayout': {'callbackFlags': ['visibleChangeCommand',
                                   'vcc',
                                   'dragCallback',
                                   'dgc',
                                   'dropCallback',
                                   'dpc'],
                 'flags': {'allowEmptyCells': {'args': '<type __builtin__.bool>',
                                               'longname': 'allowEmptyCells',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'aec'},
                           'annotation': {'args': '<type __builtin__.str>',
                                          'longname': 'annotation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ann'},
                           'autoGrow': {'args': '<type __builtin__.bool>',
                                        'longname': 'autoGrow',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'ag'},
                           'backgroundColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'backgroundColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'bgc'},
                           'cellHeight': {'args': '<type __builtin__.int>',
                                          'longname': 'cellHeight',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ch'},
                           'cellWidth': {'args': '<type __builtin__.int>',
                                         'longname': 'cellWidth',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cw'},
                           'cellWidthHeight': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'cellWidthHeight',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'cwh'},
                           'childArray': {'args': '<type __builtin__.bool>',
                                          'longname': 'childArray',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'ca'},
                           'columnsResizable': {'args': '<type __builtin__.bool>',
                                                'longname': 'columnsResizable',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'cr'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'dragCallback': {'args': 'script',
                                            'longname': 'dragCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dgc'},
                           'dropCallback': {'args': 'script',
                                            'longname': 'dropCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dpc'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'enable': {'args': '<type __builtin__.bool>',
                                      'longname': 'enable',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'en'},
                           'enableBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableBackground',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ebg'},
                           'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableKeyboardFocus',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ekf'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'fullPathName': {'args': '<type __builtin__.bool>',
                                            'longname': 'fullPathName',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fpn'},
                           'gridOrder': {'args': '<type __builtin__.bool>',
                                         'longname': 'gridOrder',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'go'},
                           'height': {'args': '<type __builtin__.int>',
                                      'longname': 'height',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                           'highlightColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'highlightColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'hlc'},
                           'isObscured': {'args': '<type __builtin__.bool>',
                                          'longname': 'isObscured',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'io'},
                           'manage': {'args': '<type __builtin__.bool>',
                                      'longname': 'manage',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                           'noBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'noBackground',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nbg'},
                           'numberOfChildren': {'args': '<type __builtin__.bool>',
                                                'longname': 'numberOfChildren',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'nch'},
                           'numberOfColumns': {'args': '<type __builtin__.int>',
                                               'longname': 'numberOfColumns',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nc'},
                           'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'numberOfPopupMenus',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'npm'},
                           'numberOfRows': {'args': '<type __builtin__.int>',
                                            'longname': 'numberOfRows',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nr'},
                           'numberOfRowsColumns': {'args': ['<type __builtin__.int>',
                                                            '<type __builtin__.int>'],
                                                   'longname': 'numberOfRowsColumns',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 2,
                                                   'shortname': 'nrc'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuArray': {'args': '<type __builtin__.bool>',
                                              'longname': 'popupMenuArray',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'pma'},
                           'position': {'args': ['<type __builtin__.str>',
                                                 '<type __builtin__.int>'],
                                        'longname': 'position',
                                        'modes': ['create',
                                                  'edit',
                                                  'multiuse'],
                                        'numArgs': 2,
                                        'shortname': 'pos'},
                           'preventOverride': {'args': '<type __builtin__.bool>',
                                               'longname': 'preventOverride',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'po'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'statusBarMessage': {'args': '<type __builtin__.str>',
                                                'longname': 'statusBarMessage',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sbm'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'},
                           'visible': {'args': '<type __builtin__.bool>',
                                       'longname': 'visible',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vis'},
                           'visibleChangeCommand': {'args': 'script',
                                                    'longname': 'visibleChangeCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'vcc'},
                           'width': {'args': '<type __builtin__.int>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'}},
                 'shortFlags': {'aec': 'allowEmptyCells',
                                'ag': 'autoGrow',
                                'ann': 'annotation',
                                'bgc': 'backgroundColor',
                                'ca': 'childArray',
                                'ch': 'cellHeight',
                                'cr': 'columnsResizable',
                                'cw': 'cellWidth',
                                'cwh': 'cellWidthHeight',
                                'dgc': 'dragCallback',
                                'dpc': 'dropCallback',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'ebg': 'enableBackground',
                                'ekf': 'enableKeyboardFocus',
                                'en': 'enable',
                                'ex': 'exists',
                                'fpn': 'fullPathName',
                                'go': 'gridOrder',
                                'h': 'height',
                                'hlc': 'highlightColor',
                                'io': 'isObscured',
                                'm': 'manage',
                                'nbg': 'noBackground',
                                'nc': 'numberOfColumns',
                                'nch': 'numberOfChildren',
                                'npm': 'numberOfPopupMenus',
                                'nr': 'numberOfRows',
                                'nrc': 'numberOfRowsColumns',
                                'p': 'parent',
                                'pma': 'popupMenuArray',
                                'po': 'preventOverride',
                                'pos': 'position',
                                'q': 'query',
                                'sbm': 'statusBarMessage',
                                'ut': 'useTemplate',
                                'vcc': 'visibleChangeCommand',
                                'vis': 'visible',
                                'w': 'width'},
                 'type': 'windows'},
  'group': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                   'longname': 'absolute',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'a'},
                      'empty': {'args': '<type __builtin__.bool>',
                                'longname': 'empty',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'em'},
                      'name': {'args': '<type __builtin__.str>',
                               'longname': 'name',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'n'},
                      'parent': {'args': '<type __builtin__.str>',
                                 'longname': 'parent',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'p'},
                      'relative': {'args': '<type __builtin__.bool>',
                                   'longname': 'relative',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'r'},
                      'useAsGroup': {'args': '<type __builtin__.str>',
                                     'longname': 'useAsGroup',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'uag'},
                      'world': {'args': '<type __builtin__.bool>',
                                'longname': 'world',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'w'}},
            'shortFlags': {'a': 'absolute',
                           'em': 'empty',
                           'n': 'name',
                           'p': 'parent',
                           'r': 'relative',
                           'uag': 'useAsGroup',
                           'w': 'world'},
            'type': 'general'},
  'groupParts': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'hardenPointCurve': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                             'longname': 'caching',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cch'},
                                 'constructionHistory': {'args': '<type __builtin__.bool>',
                                                         'longname': 'constructionHistory',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'ch'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'frozen': {'args': '<type __builtin__.bool>',
                                            'longname': 'frozen',
                                            'numArgs': 1,
                                            'shortname': 'fzn'},
                                 'multiplicity': {'args': '<type __builtin__.int>',
                                                  'longname': 'multiplicity',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'm'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'nodeState': {'args': '<type __builtin__.int>',
                                               'longname': 'nodeState',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nds'},
                                 'object': {'args': '<type __builtin__.bool>',
                                            'longname': 'object',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'o'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                     'longname': 'replaceOriginal',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'rpo'}},
                       'shortFlags': {'cch': 'caching',
                                      'ch': 'constructionHistory',
                                      'e': 'edit',
                                      'fzn': 'frozen',
                                      'm': 'multiplicity',
                                      'n': 'name',
                                      'nds': 'nodeState',
                                      'o': 'object',
                                      'q': 'query',
                                      'rpo': 'replaceOriginal'},
                       'type': 'modeling'},
  'hardware': {'flags': {'brdType': {'args': '<type __builtin__.bool>',
                                     'longname': 'brdType',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'brd'},
                         'cpuType': {'args': '<type __builtin__.bool>',
                                     'longname': 'cpuType',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'cpu'},
                         'graphicsType': {'args': '<type __builtin__.bool>',
                                          'longname': 'graphicsType',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'gfx'},
                         'megaHertz': {'args': '<type __builtin__.bool>',
                                       'longname': 'megaHertz',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'mhz'},
                         'numProcessors': {'args': '<type __builtin__.bool>',
                                           'longname': 'numProcessors',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'npr'}},
               'shortFlags': {'brd': 'brdType',
                              'cpu': 'cpuType',
                              'gfx': 'graphicsType',
                              'mhz': 'megaHertz',
                              'npr': 'numProcessors'},
               'type': 'system'},
  'hardwareRenderPanel': {'callbackFlags': ['popupMenuProcedure', 'pmp'],
                          'flags': {'camera': {'args': '<type __builtin__.str>',
                                               'longname': 'camera',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cam'},
                                    'control': {'args': '<type __builtin__.bool>',
                                                'longname': 'control',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'ctl'},
                                    'copy': {'args': '<type __builtin__.str>',
                                             'longname': 'copy',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'cp'},
                                    'createString': {'args': '<type __builtin__.bool>',
                                                     'longname': 'createString',
                                                     'modes': ['edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'cs'},
                                    'defineTemplate': {'args': '<type __builtin__.str>',
                                                       'longname': 'defineTemplate',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'dt'},
                                    'docTag': {'args': '<type __builtin__.str>',
                                               'longname': 'docTag',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dtg'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'editString': {'args': '<type __builtin__.bool>',
                                                   'longname': 'editString',
                                                   'modes': ['edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'es'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'glRenderEditor': {'args': '<type __builtin__.bool>',
                                                       'longname': 'glRenderEditor',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'gre'},
                                    'init': {'args': '<type __builtin__.bool>',
                                             'longname': 'init',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'init'},
                                    'isUnique': {'args': '<type __builtin__.bool>',
                                                 'longname': 'isUnique',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'iu'},
                                    'label': {'args': '<type __builtin__.str>',
                                              'longname': 'label',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'l'},
                                    'menuBarRepeatLast': {'args': '<type __builtin__.bool>',
                                                          'longname': 'menuBarRepeatLast',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'mrl'},
                                    'menuBarVisible': {'args': '<type __builtin__.bool>',
                                                       'longname': 'menuBarVisible',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'mbv'},
                                    'needsInit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'needsInit',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'ni'},
                                    'parent': {'args': '<type __builtin__.str>',
                                               'longname': 'parent',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'p'},
                                    'popupMenuProcedure': {'args': 'script',
                                                           'longname': 'popupMenuProcedure',
                                                           'modes': ['query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'pmp'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'replacePanel': {'args': '<type __builtin__.str>',
                                                     'longname': 'replacePanel',
                                                     'modes': ['edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'rp'},
                                    'tearOff': {'args': '<type __builtin__.bool>',
                                                'longname': 'tearOff',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 0,
                                                'shortname': 'to'},
                                    'tearOffCopy': {'args': '<type __builtin__.str>',
                                                    'longname': 'tearOffCopy',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'toc'},
                                    'tearOffRestore': {'args': '<type __builtin__.bool>',
                                                       'longname': 'tearOffRestore',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'tor'},
                                    'unParent': {'args': '<type __builtin__.bool>',
                                                 'longname': 'unParent',
                                                 'modes': ['edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'up'},
                                    'useTemplate': {'args': '<type __builtin__.str>',
                                                    'longname': 'useTemplate',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ut'}},
                          'removedFlags': {'in': 'init'},
                          'shortFlags': {'cam': 'camera',
                                         'cp': 'copy',
                                         'cs': 'createString',
                                         'ctl': 'control',
                                         'dt': 'defineTemplate',
                                         'dtg': 'docTag',
                                         'e': 'edit',
                                         'es': 'editString',
                                         'ex': 'exists',
                                         'gre': 'glRenderEditor',
                                         'init': 'init',
                                         'iu': 'isUnique',
                                         'l': 'label',
                                         'mbv': 'menuBarVisible',
                                         'mrl': 'menuBarRepeatLast',
                                         'ni': 'needsInit',
                                         'p': 'parent',
                                         'pmp': 'popupMenuProcedure',
                                         'q': 'query',
                                         'rp': 'replacePanel',
                                         'to': 'tearOff',
                                         'toc': 'tearOffCopy',
                                         'tor': 'tearOffRestore',
                                         'up': 'unParent',
                                         'ut': 'useTemplate'},
                          'type': 'windows'},
  'hasMetadata': {'flags': {'asList': {'args': '<type __builtin__.bool>',
                                       'longname': 'asList',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'al'},
                            'channelName': {'args': '<type __builtin__.str>',
                                            'longname': 'channelName',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'cn'},
                            'channelType': {'args': '<type __builtin__.str>',
                                            'longname': 'channelType',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'cht'},
                            'endIndex': {'args': '<type __builtin__.str>',
                                         'longname': 'endIndex',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'eix'},
                            'ignoreDefault': {'args': '<type __builtin__.bool>',
                                              'longname': 'ignoreDefault',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'id'},
                            'index': {'args': '<type __builtin__.str>',
                                      'longname': 'index',
                                      'modes': ['create',
                                                'query',
                                                'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'idx'},
                            'indexType': {'args': '<type __builtin__.str>',
                                          'longname': 'indexType',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'idt'},
                            'memberName': {'args': '<type __builtin__.str>',
                                           'longname': 'memberName',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'mn'},
                            'scene': {'args': '<type __builtin__.bool>',
                                      'longname': 'scene',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'scn'},
                            'startIndex': {'args': '<type __builtin__.str>',
                                           'longname': 'startIndex',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'six'},
                            'streamName': {'args': '<type __builtin__.str>',
                                           'longname': 'streamName',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'stn'}},
                  'shortFlags': {'al': 'asList',
                                 'cht': 'channelType',
                                 'cn': 'channelName',
                                 'eix': 'endIndex',
                                 'id': 'ignoreDefault',
                                 'idt': 'indexType',
                                 'idx': 'index',
                                 'mn': 'memberName',
                                 'scn': 'scene',
                                 'six': 'startIndex',
                                 'stn': 'streamName'},
                  'type': 'modeling'},
  'headsUpDisplay': {'callbackFlags': ['command', 'c'],
                     'flags': {'allDescendants': {'args': '<type __builtin__.bool>',
                                                  'longname': 'allDescendants',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'ad'},
                               'allowOverlap': {'args': '<type __builtin__.bool>',
                                                'longname': 'allowOverlap',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ao'},
                               'attachToRefresh': {'args': '<type __builtin__.bool>',
                                                   'longname': 'attachToRefresh',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'atr'},
                               'attributeChange': {'args': '<type __builtin__.str>',
                                                   'longname': 'attributeChange',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ac'},
                               'block': {'args': '<type __builtin__.int>',
                                         'longname': 'block',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'b'},
                               'blockAlignment': {'args': '<type __builtin__.str>',
                                                  'longname': 'blockAlignment',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ba'},
                               'blockSize': {'args': '<type __builtin__.str>',
                                             'longname': 'blockSize',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'bs'},
                               'command': {'args': 'script',
                                           'longname': 'command',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'c'},
                               'conditionChange': {'args': '<type __builtin__.str>',
                                                   'longname': 'conditionChange',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cc'},
                               'conditionFalse': {'args': '<type __builtin__.str>',
                                                  'longname': 'conditionFalse',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cf'},
                               'conditionTrue': {'args': '<type __builtin__.str>',
                                                 'longname': 'conditionTrue',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ct'},
                               'connectionChange': {'args': '<type __builtin__.str>',
                                                    'longname': 'connectionChange',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'con'},
                               'dataAlignment': {'args': '<type __builtin__.str>',
                                                 'longname': 'dataAlignment',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'da'},
                               'dataFontSize': {'args': '<type __builtin__.str>',
                                                'longname': 'dataFontSize',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dfs'},
                               'dataWidth': {'args': '<type __builtin__.int>',
                                             'longname': 'dataWidth',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dw'},
                               'decimalPrecision': {'args': '<type __builtin__.int>',
                                                    'longname': 'decimalPrecision',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dp'},
                               'disregardIndex': {'args': '<type __builtin__.bool>',
                                                  'longname': 'disregardIndex',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'di'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'event': {'args': '<type __builtin__.str>',
                                         'longname': 'event',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ev'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create', 'query'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'getOption': {'args': '<type __builtin__.str>',
                                             'longname': 'getOption',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'op'},
                               'gridColor': {'args': '<type __builtin__.int>',
                                             'longname': 'gridColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'gco'},
                               'label': {'args': '<type __builtin__.str>',
                                         'longname': 'label',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'l'},
                               'labelFontSize': {'args': '<type __builtin__.str>',
                                                 'longname': 'labelFontSize',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'lfs'},
                               'labelWidth': {'args': '<type __builtin__.int>',
                                              'longname': 'labelWidth',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lw'},
                               'lastOccupiedBlock': {'args': '<type __builtin__.int>',
                                                     'longname': 'lastOccupiedBlock',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'lob'},
                               'layoutVisibility': {'args': '<type __builtin__.bool>',
                                                    'longname': 'layoutVisibility',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'lv'},
                               'listConditions': {'args': '<type __builtin__.bool>',
                                                  'longname': 'listConditions',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 0,
                                                  'shortname': 'lc'},
                               'listEvents': {'args': '<type __builtin__.bool>',
                                              'longname': 'listEvents',
                                              'modes': ['create', 'query'],
                                              'numArgs': 0,
                                              'shortname': 'le'},
                               'listHeadsUpDisplays': {'args': '<type __builtin__.bool>',
                                                       'longname': 'listHeadsUpDisplays',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 0,
                                                       'shortname': 'lh'},
                               'listNodeChanges': {'args': '<type __builtin__.bool>',
                                                   'longname': 'listNodeChanges',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 0,
                                                   'shortname': 'lnc'},
                               'listPresets': {'args': '<type __builtin__.bool>',
                                               'longname': 'listPresets',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 0,
                                               'shortname': 'lp'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['edit'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'nextFreeBlock': {'args': '<type __builtin__.int>',
                                                 'longname': 'nextFreeBlock',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'nfb'},
                               'nodeChanges': {'args': '<type __builtin__.str>',
                                               'longname': 'nodeChanges',
                                               'modes': ['create',
                                                         'query',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'nc'},
                               'padding': {'args': '<type __builtin__.int>',
                                           'longname': 'padding',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'p'},
                               'preset': {'args': '<type __builtin__.str>',
                                          'longname': 'preset',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pre'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'refresh': {'args': '<type __builtin__.bool>',
                                           'longname': 'refresh',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'r'},
                               'remove': {'args': '<type __builtin__.bool>',
                                          'longname': 'remove',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'rem'},
                               'removeID': {'args': '<type __builtin__.int>',
                                            'longname': 'removeID',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'rid'},
                               'removePosition': {'args': ['<type __builtin__.int>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'removePosition',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 2,
                                                  'shortname': 'rp'},
                               'resetNodeChanges': {'args': '<type __builtin__.str>',
                                                    'longname': 'resetNodeChanges',
                                                    'modes': ['edit',
                                                              'multiuse'],
                                                    'numArgs': 1,
                                                    'shortname': 'rnc'},
                               'scriptResult': {'args': '<type __builtin__.bool>',
                                                'longname': 'scriptResult',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'sr'},
                               'section': {'args': '<type __builtin__.int>',
                                           'longname': 'section',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 's'},
                               'setOption': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'setOption',
                                             'modes': ['edit'],
                                             'numArgs': 2,
                                             'shortname': 'so'},
                               'showGrid': {'args': '<type __builtin__.bool>',
                                            'longname': 'showGrid',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 0,
                                            'shortname': 'sg'},
                               'visible': {'args': '<type __builtin__.bool>',
                                           'longname': 'visible',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vis'}},
                     'shortFlags': {'ac': 'attributeChange',
                                    'ad': 'allDescendants',
                                    'ao': 'allowOverlap',
                                    'atr': 'attachToRefresh',
                                    'b': 'block',
                                    'ba': 'blockAlignment',
                                    'bs': 'blockSize',
                                    'c': 'command',
                                    'cc': 'conditionChange',
                                    'cf': 'conditionFalse',
                                    'con': 'connectionChange',
                                    'ct': 'conditionTrue',
                                    'da': 'dataAlignment',
                                    'dfs': 'dataFontSize',
                                    'di': 'disregardIndex',
                                    'dp': 'decimalPrecision',
                                    'dw': 'dataWidth',
                                    'e': 'edit',
                                    'ev': 'event',
                                    'ex': 'exists',
                                    'gco': 'gridColor',
                                    'l': 'label',
                                    'lc': 'listConditions',
                                    'le': 'listEvents',
                                    'lfs': 'labelFontSize',
                                    'lh': 'listHeadsUpDisplays',
                                    'lnc': 'listNodeChanges',
                                    'lob': 'lastOccupiedBlock',
                                    'lp': 'listPresets',
                                    'lv': 'layoutVisibility',
                                    'lw': 'labelWidth',
                                    'n': 'name',
                                    'nc': 'nodeChanges',
                                    'nfb': 'nextFreeBlock',
                                    'op': 'getOption',
                                    'p': 'padding',
                                    'pre': 'preset',
                                    'q': 'query',
                                    'r': 'refresh',
                                    'rem': 'remove',
                                    'rid': 'removeID',
                                    'rnc': 'resetNodeChanges',
                                    'rp': 'removePosition',
                                    's': 'section',
                                    'sg': 'showGrid',
                                    'so': 'setOption',
                                    'sr': 'scriptResult',
                                    'vis': 'visible'},
                     'type': 'windows'},
  'headsUpMessage': {'flags': {'horizontalOffset': {'args': '<type __builtin__.int>',
                                                    'longname': 'horizontalOffset',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ho'},
                               'object': {'args': '<type __builtin__.str>',
                                          'longname': 'object',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'o'},
                               'selection': {'args': '<type __builtin__.bool>',
                                             'longname': 'selection',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 's'},
                               'time': {'args': '<type __builtin__.float>',
                                        'longname': 'time',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 't'},
                               'uvTextureEditor': {'args': '<type __builtin__.bool>',
                                                   'longname': 'uvTextureEditor',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'uve'},
                               'verticalOffset': {'args': '<type __builtin__.int>',
                                                  'longname': 'verticalOffset',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'vo'},
                               'viewport': {'args': '<type __builtin__.bool>',
                                            'longname': 'viewport',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'vp'}},
                     'shortFlags': {'ho': 'horizontalOffset',
                                    'o': 'object',
                                    's': 'selection',
                                    't': 'time',
                                    'uve': 'uvTextureEditor',
                                    'vo': 'verticalOffset',
                                    'vp': 'viewport'},
                     'type': 'windows'},
  'help': {'flags': {'documentation': {'args': '<type __builtin__.bool>',
                                       'longname': 'documentation',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'doc'},
                     'language': {'args': '<type __builtin__.str>',
                                  'longname': 'language',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'lng'},
                     'list': {'args': '<type __builtin__.bool>',
                              'longname': 'list',
                              'modes': ['create'],
                              'numArgs': 0,
                              'shortname': 'l'},
                     'popupDisplayTime': {'args': '<type __builtin__.int>',
                                          'longname': 'popupDisplayTime',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'pdt'},
                     'popupMode': {'args': '<type __builtin__.bool>',
                                   'longname': 'popupMode',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'pm'},
                     'popupPauseTime': {'args': '<type __builtin__.int>',
                                        'longname': 'popupPauseTime',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'ppt'},
                     'popupSimpleMode': {'args': '<type __builtin__.bool>',
                                         'longname': 'popupSimpleMode',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'psm'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'rolloverMode': {'args': '<type __builtin__.bool>',
                                      'longname': 'rolloverMode',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'rm'},
                     'syntaxOnly': {'args': '<type __builtin__.bool>',
                                    'longname': 'syntaxOnly',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'so'}},
           'shortFlags': {'doc': 'documentation',
                          'l': 'list',
                          'lng': 'language',
                          'pdt': 'popupDisplayTime',
                          'pm': 'popupMode',
                          'ppt': 'popupPauseTime',
                          'psm': 'popupSimpleMode',
                          'q': 'query',
                          'rm': 'rolloverMode',
                          'so': 'syntaxOnly'},
           'type': None},
  'helpLine': {'callbackFlags': ['visibleChangeCommand',
                                 'vcc',
                                 'dragCallback',
                                 'dgc',
                                 'dropCallback',
                                 'dpc'],
               'flags': {'annotation': {'args': '<type __builtin__.str>',
                                        'longname': 'annotation',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ann'},
                         'backgroundColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'backgroundColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'bgc'},
                         'defineTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'defineTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'dt'},
                         'docTag': {'args': '<type __builtin__.str>',
                                    'longname': 'docTag',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'dtg'},
                         'dragCallback': {'args': 'script',
                                          'longname': 'dragCallback',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dgc'},
                         'dropCallback': {'args': 'script',
                                          'longname': 'dropCallback',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dpc'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'enable': {'args': '<type __builtin__.bool>',
                                    'longname': 'enable',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'en'},
                         'enableBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'enableBackground',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ebg'},
                         'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableKeyboardFocus',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ekf'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'fullPathName': {'args': '<type __builtin__.bool>',
                                          'longname': 'fullPathName',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'fpn'},
                         'height': {'args': '<type __builtin__.int>',
                                    'longname': 'height',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'h'},
                         'highlightColor': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'highlightColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'hlc'},
                         'isObscured': {'args': '<type __builtin__.bool>',
                                        'longname': 'isObscured',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'io'},
                         'manage': {'args': '<type __builtin__.bool>',
                                    'longname': 'manage',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'm'},
                         'noBackground': {'args': '<type __builtin__.bool>',
                                          'longname': 'noBackground',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nbg'},
                         'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                'longname': 'numberOfPopupMenus',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'npm'},
                         'parent': {'args': '<type __builtin__.str>',
                                    'longname': 'parent',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'p'},
                         'popupMenuArray': {'args': '<type __builtin__.bool>',
                                            'longname': 'popupMenuArray',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'pma'},
                         'preventOverride': {'args': '<type __builtin__.bool>',
                                             'longname': 'preventOverride',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'po'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'statusBarMessage': {'args': '<type __builtin__.str>',
                                              'longname': 'statusBarMessage',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sbm'},
                         'useTemplate': {'args': '<type __builtin__.str>',
                                         'longname': 'useTemplate',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ut'},
                         'visible': {'args': '<type __builtin__.bool>',
                                     'longname': 'visible',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'vis'},
                         'visibleChangeCommand': {'args': 'script',
                                                  'longname': 'visibleChangeCommand',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'vcc'},
                         'width': {'args': '<type __builtin__.int>',
                                   'longname': 'width',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'w'}},
               'shortFlags': {'ann': 'annotation',
                              'bgc': 'backgroundColor',
                              'dgc': 'dragCallback',
                              'dpc': 'dropCallback',
                              'dt': 'defineTemplate',
                              'dtg': 'docTag',
                              'e': 'edit',
                              'ebg': 'enableBackground',
                              'ekf': 'enableKeyboardFocus',
                              'en': 'enable',
                              'ex': 'exists',
                              'fpn': 'fullPathName',
                              'h': 'height',
                              'hlc': 'highlightColor',
                              'io': 'isObscured',
                              'm': 'manage',
                              'nbg': 'noBackground',
                              'npm': 'numberOfPopupMenus',
                              'p': 'parent',
                              'pma': 'popupMenuArray',
                              'po': 'preventOverride',
                              'q': 'query',
                              'sbm': 'statusBarMessage',
                              'ut': 'useTemplate',
                              'vcc': 'visibleChangeCommand',
                              'vis': 'visible',
                              'w': 'width'},
               'type': 'windows'},
  'hide': {'flags': {'allObjects': {'args': '<type __builtin__.bool>',
                                    'longname': 'allObjects',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'all'},
                     'clearLastHidden': {'args': '<type __builtin__.bool>',
                                         'longname': 'clearLastHidden',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'clh'},
                     'clearSelection': {'args': '<type __builtin__.bool>',
                                        'longname': 'clearSelection',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'cs'},
                     'invertComponents': {'args': '<type __builtin__.bool>',
                                          'longname': 'invertComponents',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ic'},
                     'returnHidden': {'args': '<type __builtin__.bool>',
                                      'longname': 'returnHidden',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'rh'},
                     'testVisibility': {'args': '<type __builtin__.bool>',
                                        'longname': 'testVisibility',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'tv'}},
           'shortFlags': {'all': 'allObjects',
                          'clh': 'clearLastHidden',
                          'cs': 'clearSelection',
                          'ic': 'invertComponents',
                          'rh': 'returnHidden',
                          'tv': 'testVisibility'},
           'type': 'general'},
  'hikGlobals': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'releaseAllPinning': {'args': '<type __builtin__.bool>',
                                                 'longname': 'releaseAllPinning',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rap'}},
                 'shortFlags': {'e': 'edit',
                                'q': 'query',
                                'rap': 'releaseAllPinning'},
                 'type': 'animation'},
  'hilite': {'flags': {'replace': {'args': '<type __builtin__.bool>',
                                   'longname': 'replace',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'r'},
                       'toggle': {'args': '<type __builtin__.bool>',
                                  'longname': 'toggle',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'tgl'},
                       'unHilite': {'args': '<type __builtin__.bool>',
                                    'longname': 'unHilite',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'u'}},
             'shortFlags': {'r': 'replace',
                            'tgl': 'toggle',
                            'u': 'unHilite'},
             'type': 'general'},
  'hitTest': {'flags': {}, 'shortFlags': {}, 'type': 'system'},
  'hotBox': {'callbackFlags': ['noClickCommand', 'ncc'],
             'flags': {'PaneOnlyMenus': {'args': '<type __builtin__.bool>',
                                         'longname': 'PaneOnlyMenus',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'po'},
                       'PaneToggleMenus': {'args': '<type __builtin__.bool>',
                                           'longname': 'PaneToggleMenus',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'pt'},
                       'animationOnlyMenus': {'args': '<type __builtin__.bool>',
                                              'longname': 'animationOnlyMenus',
                                              'modes': ['create', 'query'],
                                              'numArgs': 0,
                                              'shortname': 'ao'},
                       'animationToggleMenus': {'args': '<type __builtin__.bool>',
                                                'longname': 'animationToggleMenus',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'at'},
                       'clothOnlyMenus': {'args': '<type __builtin__.bool>',
                                          'longname': 'clothOnlyMenus',
                                          'modes': ['create', 'query'],
                                          'numArgs': 0,
                                          'shortname': 'clo'},
                       'clothToggleMenus': {'args': '<type __builtin__.bool>',
                                            'longname': 'clothToggleMenus',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'clt'},
                       'commonOnlyMenus': {'args': '<type __builtin__.bool>',
                                           'longname': 'commonOnlyMenus',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'co'},
                       'commonToggleMenus': {'args': '<type __builtin__.bool>',
                                             'longname': 'commonToggleMenus',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'ct'},
                       'customMenuSetsToggleMenus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'customMenuSetsToggleMenus',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'cst'},
                       'displayCenterOnly': {'args': '<type __builtin__.bool>',
                                             'longname': 'displayCenterOnly',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'dco'},
                       'displayHotbox': {'args': '<type __builtin__.bool>',
                                         'longname': 'displayHotbox',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'dh'},
                       'displayStyle': {'args': '<type __builtin__.bool>',
                                        'longname': 'displayStyle',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'ds'},
                       'displayZonesOnly': {'args': '<type __builtin__.bool>',
                                            'longname': 'displayZonesOnly',
                                            'modes': ['create', 'query'],
                                            'numArgs': 0,
                                            'shortname': 'dzo'},
                       'dynamicsOnlyMenus': {'args': '<type __builtin__.bool>',
                                             'longname': 'dynamicsOnlyMenus',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'do'},
                       'dynamicsToggleMenus': {'args': '<type __builtin__.bool>',
                                               'longname': 'dynamicsToggleMenus',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                       'liveOnlyMenus': {'args': '<type __builtin__.bool>',
                                         'longname': 'liveOnlyMenus',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'lo'},
                       'liveToggleMenus': {'args': '<type __builtin__.bool>',
                                           'longname': 'liveToggleMenus',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'lt'},
                       'menuSetOnly': {'args': '<type __builtin__.str>',
                                       'longname': 'menuSetOnly',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'mso'},
                       'menuSetToggle': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.bool>'],
                                         'longname': 'menuSetToggle',
                                         'modes': ['create'],
                                         'numArgs': 2,
                                         'shortname': 'mst'},
                       'modelingOnlyMenus': {'args': '<type __builtin__.bool>',
                                             'longname': 'modelingOnlyMenus',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'mdo'},
                       'modelingToggleMenus': {'args': '<type __builtin__.bool>',
                                               'longname': 'modelingToggleMenus',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'mdt'},
                       'noClickCommand': {'args': 'script',
                                          'longname': 'noClickCommand',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ncc'},
                       'noClickDelay': {'args': '<type __builtin__.float>',
                                        'longname': 'noClickDelay',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ncd'},
                       'noClickPosition': {'args': '<type __builtin__.bool>',
                                           'longname': 'noClickPosition',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ncp'},
                       'noKeyPress': {'args': '<type __builtin__.bool>',
                                      'longname': 'noKeyPress',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'nkp'},
                       'polygonsOnlyMenus': {'args': '<type __builtin__.bool>',
                                             'longname': 'polygonsOnlyMenus',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'plo'},
                       'polygonsToggleMenus': {'args': '<type __builtin__.bool>',
                                               'longname': 'polygonsToggleMenus',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'plt'},
                       'position': {'args': ['<type __builtin__.int>',
                                             '<type __builtin__.int>'],
                                    'longname': 'position',
                                    'modes': ['create'],
                                    'numArgs': 2,
                                    'shortname': 'pos'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'release': {'args': '<type __builtin__.bool>',
                                   'longname': 'release',
                                   'modes': ['create', 'query'],
                                   'numArgs': 0,
                                   'shortname': 'rl'},
                       'renderingOnlyMenus': {'args': '<type __builtin__.bool>',
                                              'longname': 'renderingOnlyMenus',
                                              'modes': ['create', 'query'],
                                              'numArgs': 0,
                                              'shortname': 'ro'},
                       'renderingToggleMenus': {'args': '<type __builtin__.bool>',
                                                'longname': 'renderingToggleMenus',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'rt'},
                       'riggingOnlyMenus': {'args': '<type __builtin__.bool>',
                                            'longname': 'riggingOnlyMenus',
                                            'modes': ['create', 'query'],
                                            'numArgs': 0,
                                            'shortname': 'rgo'},
                       'riggingToggleMenus': {'args': '<type __builtin__.bool>',
                                              'longname': 'riggingToggleMenus',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'rgt'},
                       'rmbPopups': {'args': '<type __builtin__.bool>',
                                     'longname': 'rmbPopups',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'rmb'},
                       'showAllToggleMenus': {'args': '<type __builtin__.bool>',
                                              'longname': 'showAllToggleMenus',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'a'},
                       'surfacesOnlyMenus': {'args': '<type __builtin__.bool>',
                                             'longname': 'surfacesOnlyMenus',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'so'},
                       'surfacesToggleMenus': {'args': '<type __builtin__.bool>',
                                               'longname': 'surfacesToggleMenus',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'st'},
                       'transparenyLevel': {'args': '<type __builtin__.int>',
                                            'longname': 'transparenyLevel',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'tr'},
                       'updateMenus': {'args': '<type __builtin__.bool>',
                                       'longname': 'updateMenus',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'um'}},
             'shortFlags': {'a': 'showAllToggleMenus',
                            'ao': 'animationOnlyMenus',
                            'at': 'animationToggleMenus',
                            'clo': 'clothOnlyMenus',
                            'clt': 'clothToggleMenus',
                            'co': 'commonOnlyMenus',
                            'cst': 'customMenuSetsToggleMenus',
                            'ct': 'commonToggleMenus',
                            'dco': 'displayCenterOnly',
                            'dh': 'displayHotbox',
                            'do': 'dynamicsOnlyMenus',
                            'ds': 'displayStyle',
                            'dt': 'dynamicsToggleMenus',
                            'dzo': 'displayZonesOnly',
                            'lo': 'liveOnlyMenus',
                            'lt': 'liveToggleMenus',
                            'mdo': 'modelingOnlyMenus',
                            'mdt': 'modelingToggleMenus',
                            'mso': 'menuSetOnly',
                            'mst': 'menuSetToggle',
                            'ncc': 'noClickCommand',
                            'ncd': 'noClickDelay',
                            'ncp': 'noClickPosition',
                            'nkp': 'noKeyPress',
                            'plo': 'polygonsOnlyMenus',
                            'plt': 'polygonsToggleMenus',
                            'po': 'PaneOnlyMenus',
                            'pos': 'position',
                            'pt': 'PaneToggleMenus',
                            'q': 'query',
                            'rgo': 'riggingOnlyMenus',
                            'rgt': 'riggingToggleMenus',
                            'rl': 'release',
                            'rmb': 'rmbPopups',
                            'ro': 'renderingOnlyMenus',
                            'rt': 'renderingToggleMenus',
                            'so': 'surfacesOnlyMenus',
                            'st': 'surfacesToggleMenus',
                            'tr': 'transparenyLevel',
                            'um': 'updateMenus'},
             'type': 'windows'},
  'hotkey': {'callbackFlags': ['commandModifier',
                               'cmd',
                               'releaseCommandRepeat',
                               'rcr',
                               'pressCommandRepeat',
                               'pcr'],
             'flags': {'altModifier': {'args': '<type __builtin__.bool>',
                                       'longname': 'altModifier',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'alt'},
                       'autoSave': {'args': '<type __builtin__.bool>',
                                    'longname': 'autoSave',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'autoSave'},
                       'commandModifier': {'args': '<type __builtin__.bool>',
                                           'longname': 'commandModifier',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'cmd'},
                       'ctrlModifier': {'args': '<type __builtin__.bool>',
                                        'longname': 'ctrlModifier',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'ctl'},
                       'ctxClient': {'args': '<type __builtin__.str>',
                                     'longname': 'ctxClient',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'cc'},
                       'dragPress': {'args': '<type __builtin__.bool>',
                                     'longname': 'dragPress',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'dp'},
                       'factorySettings': {'args': '<type __builtin__.bool>',
                                           'longname': 'factorySettings',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'fs'},
                       'isModifier': {'args': '<type __builtin__.bool>',
                                      'longname': 'isModifier',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'mod'},
                       'keyShortcut': {'args': '<type __builtin__.str>',
                                       'longname': 'keyShortcut',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'k'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['create', 'query'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'pressCommandRepeat': {'args': '<type __builtin__.bool>',
                                              'longname': 'pressCommandRepeat',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'pcr'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'releaseCommandRepeat': {'args': '<type __builtin__.bool>',
                                                'longname': 'releaseCommandRepeat',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'rcr'},
                       'releaseName': {'args': '<type __builtin__.str>',
                                       'longname': 'releaseName',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'rn'},
                       'shiftModifier': {'args': '<type __builtin__.bool>',
                                         'longname': 'shiftModifier',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'sht'},
                       'sourceUserHotkeys': {'args': '<type __builtin__.bool>',
                                             'longname': 'sourceUserHotkeys',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'suh'}},
             'removedFlags': {'as': 'autoSave'},
             'shortFlags': {'alt': 'altModifier',
                            'autoSave': 'autoSave',
                            'cc': 'ctxClient',
                            'cmd': 'commandModifier',
                            'ctl': 'ctrlModifier',
                            'dp': 'dragPress',
                            'fs': 'factorySettings',
                            'k': 'keyShortcut',
                            'mod': 'isModifier',
                            'n': 'name',
                            'pcr': 'pressCommandRepeat',
                            'q': 'query',
                            'rcr': 'releaseCommandRepeat',
                            'rn': 'releaseName',
                            'sht': 'shiftModifier',
                            'suh': 'sourceUserHotkeys'},
             'type': 'windows'},
  'hotkeyCheck': {'callbackFlags': ['commandModifier', 'cmd'],
                  'flags': {'altModifier': {'args': '<type __builtin__.bool>',
                                            'longname': 'altModifier',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'alt'},
                            'commandModifier': {'args': '<type __builtin__.bool>',
                                                'longname': 'commandModifier',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'cmd'},
                            'ctrlModifier': {'args': '<type __builtin__.bool>',
                                             'longname': 'ctrlModifier',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ctl'},
                            'isRepeatable': {'args': '<type __builtin__.bool>',
                                             'longname': 'isRepeatable',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ir'},
                            'keyString': {'args': '<type __builtin__.str>',
                                          'longname': 'keyString',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'k'},
                            'keyUp': {'args': '<type __builtin__.bool>',
                                      'longname': 'keyUp',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'kup'},
                            'optionModifier': {'args': '<type __builtin__.bool>',
                                               'longname': 'optionModifier',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'opt'},
                            'toBeRemovedInFutureMayaRelease': {'args': '<type __builtin__.str>',
                                                               'longname': 'toBeRemovedInFutureMayaRelease',
                                                               'numArgs': 1,
                                                               'shortname': 'key'}},
                  'shortFlags': {'alt': 'altModifier',
                                 'cmd': 'commandModifier',
                                 'ctl': 'ctrlModifier',
                                 'ir': 'isRepeatable',
                                 'k': 'keyString',
                                 'key': 'toBeRemovedInFutureMayaRelease',
                                 'kup': 'keyUp',
                                 'opt': 'optionModifier'},
                  'type': 'windows'},
  'hotkeyCtx': {'flags': {'addClient': {'args': '<type __builtin__.str>',
                                        'longname': 'addClient',
                                        'modes': ['create', 'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'ac'},
                          'clientArray': {'args': '<type __builtin__.bool>',
                                          'longname': 'clientArray',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'ca'},
                          'currentClient': {'args': '<type __builtin__.str>',
                                            'longname': 'currentClient',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'cc'},
                          'insertTypeAt': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'insertTypeAt',
                                           'modes': ['create'],
                                           'numArgs': 2,
                                           'shortname': 'ita'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'removeAllClients': {'args': '<type __builtin__.bool>',
                                               'longname': 'removeAllClients',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'rac'},
                          'removeClient': {'args': '<type __builtin__.str>',
                                           'longname': 'removeClient',
                                           'modes': ['create', 'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'rc'},
                          'removeType': {'args': '<type __builtin__.str>',
                                         'longname': 'removeType',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'rt'},
                          'type': {'args': '<type __builtin__.str>',
                                   'longname': 'type',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 't'},
                          'typeArray': {'args': '<type __builtin__.bool>',
                                        'longname': 'typeArray',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'ta'},
                          'typeExists': {'args': '<type __builtin__.str>',
                                         'longname': 'typeExists',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'te'}},
                'shortFlags': {'ac': 'addClient',
                               'ca': 'clientArray',
                               'cc': 'currentClient',
                               'ita': 'insertTypeAt',
                               'q': 'query',
                               'rac': 'removeAllClients',
                               'rc': 'removeClient',
                               'rt': 'removeType',
                               't': 'type',
                               'ta': 'typeArray',
                               'te': 'typeExists'},
                'type': 'context'},
  'hotkeyEditor': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'hotkeyEditorPanel': {'callbackFlags': ['visibleChangeCommand',
                                          'vcc',
                                          'dragCallback',
                                          'dgc',
                                          'dropCallback',
                                          'dpc'],
                        'flags': {'annotation': {'args': '<type __builtin__.str>',
                                                 'longname': 'annotation',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ann'},
                                  'backgroundColor': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'backgroundColor',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'bgc'},
                                  'defineTemplate': {'args': '<type __builtin__.str>',
                                                     'longname': 'defineTemplate',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'dt'},
                                  'docTag': {'args': '<type __builtin__.str>',
                                             'longname': 'docTag',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dtg'},
                                  'dragCallback': {'args': 'script',
                                                   'longname': 'dragCallback',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dgc'},
                                  'dropCallback': {'args': 'script',
                                                   'longname': 'dropCallback',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dpc'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'enable': {'args': '<type __builtin__.bool>',
                                             'longname': 'enable',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'en'},
                                  'enableBackground': {'args': '<type __builtin__.bool>',
                                                       'longname': 'enableBackground',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ebg'},
                                  'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                          'longname': 'enableKeyboardFocus',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ekf'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'fullPathName': {'args': '<type __builtin__.bool>',
                                                   'longname': 'fullPathName',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'fpn'},
                                  'height': {'args': '<type __builtin__.int>',
                                             'longname': 'height',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'h'},
                                  'highlightColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'highlightColor',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'hlc'},
                                  'isObscured': {'args': '<type __builtin__.bool>',
                                                 'longname': 'isObscured',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'io'},
                                  'manage': {'args': '<type __builtin__.bool>',
                                             'longname': 'manage',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'm'},
                                  'noBackground': {'args': '<type __builtin__.bool>',
                                                   'longname': 'noBackground',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'nbg'},
                                  'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                         'longname': 'numberOfPopupMenus',
                                                         'modes': ['query'],
                                                         'numArgs': 0,
                                                         'shortname': 'npm'},
                                  'parent': {'args': '<type __builtin__.str>',
                                             'longname': 'parent',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'p'},
                                  'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                     'longname': 'popupMenuArray',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'pma'},
                                  'preventOverride': {'args': '<type __builtin__.bool>',
                                                      'longname': 'preventOverride',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'po'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'statusBarMessage': {'args': '<type __builtin__.str>',
                                                       'longname': 'statusBarMessage',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'sbm'},
                                  'useTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'useTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ut'},
                                  'visible': {'args': '<type __builtin__.bool>',
                                              'longname': 'visible',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'vis'},
                                  'visibleChangeCommand': {'args': 'script',
                                                           'longname': 'visibleChangeCommand',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'vcc'},
                                  'width': {'args': '<type __builtin__.int>',
                                            'longname': 'width',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'w'}},
                        'shortFlags': {'ann': 'annotation',
                                       'bgc': 'backgroundColor',
                                       'dgc': 'dragCallback',
                                       'dpc': 'dropCallback',
                                       'dt': 'defineTemplate',
                                       'dtg': 'docTag',
                                       'e': 'edit',
                                       'ebg': 'enableBackground',
                                       'ekf': 'enableKeyboardFocus',
                                       'en': 'enable',
                                       'ex': 'exists',
                                       'fpn': 'fullPathName',
                                       'h': 'height',
                                       'hlc': 'highlightColor',
                                       'io': 'isObscured',
                                       'm': 'manage',
                                       'nbg': 'noBackground',
                                       'npm': 'numberOfPopupMenus',
                                       'p': 'parent',
                                       'pma': 'popupMenuArray',
                                       'po': 'preventOverride',
                                       'q': 'query',
                                       'sbm': 'statusBarMessage',
                                       'ut': 'useTemplate',
                                       'vcc': 'visibleChangeCommand',
                                       'vis': 'visible',
                                       'w': 'width'},
                        'type': 'windows'},
  'hotkeyMapSet': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'hotkeySet': {'flags': {'current': {'args': '<type __builtin__.bool>',
                                      'longname': 'current',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'cu'},
                          'delete': {'args': '<type __builtin__.bool>',
                                     'longname': 'delete',
                                     'modes': ['edit'],
                                     'numArgs': 0,
                                     'shortname': 'delete'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'export': {'args': '<type __builtin__.str>',
                                     'longname': 'export',
                                     'modes': ['edit'],
                                     'numArgs': 1,
                                     'shortname': 'ep'},
                          'hotkeySetArray': {'args': '<type __builtin__.bool>',
                                             'longname': 'hotkeySetArray',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'hsa'},
                          'ip': {'args': '<type __builtin__.str>',
                                 'longname': 'ip',
                                 'modes': ['edit'],
                                 'numArgs': 1,
                                 'shortname': 'ip'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'rename': {'args': '<type __builtin__.str>',
                                     'longname': 'rename',
                                     'modes': ['edit'],
                                     'numArgs': 1,
                                     'shortname': 're'},
                          'source': {'args': '<type __builtin__.str>',
                                     'longname': 'source',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'src'}},
                'removedFlags': {'del': 'delete', 'import': 'ip'},
                'shortFlags': {'cu': 'current',
                               'delete': 'delete',
                               'e': 'edit',
                               'ep': 'export',
                               'ex': 'exists',
                               'hsa': 'hotkeySetArray',
                               'ip': 'ip',
                               'q': 'query',
                               're': 'rename',
                               'src': 'source'},
                'type': 'windows'},
  'hudButton': {'callbackFlags': ['releaseCommand',
                                  'rc',
                                  'pressCommand',
                                  'pc'],
                'flags': {'allowOverlap': {'args': '<type __builtin__.bool>',
                                           'longname': 'allowOverlap',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ao'},
                          'block': {'args': '<type __builtin__.int>',
                                    'longname': 'block',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'b'},
                          'blockAlignment': {'args': '<type __builtin__.str>',
                                             'longname': 'blockAlignment',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ba'},
                          'blockSize': {'args': '<type __builtin__.str>',
                                        'longname': 'blockSize',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'bs'},
                          'buttonShape': {'args': '<type __builtin__.str>',
                                          'longname': 'buttonShape',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'bsh'},
                          'buttonWidth': {'args': '<type __builtin__.int>',
                                          'longname': 'buttonWidth',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'bw'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'label': {'args': '<type __builtin__.str>',
                                    'longname': 'label',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'l'},
                          'labelFontSize': {'args': '<type __builtin__.str>',
                                            'longname': 'labelFontSize',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'lfs'},
                          'padding': {'args': '<type __builtin__.int>',
                                      'longname': 'padding',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                          'pressCommand': {'args': 'script',
                                           'longname': 'pressCommand',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pc'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'releaseCommand': {'args': 'script',
                                             'longname': 'releaseCommand',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'rc'},
                          'section': {'args': '<type __builtin__.int>',
                                      'longname': 'section',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                          'visible': {'args': '<type __builtin__.bool>',
                                      'longname': 'visible',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'vis'}},
                'shortFlags': {'ao': 'allowOverlap',
                               'b': 'block',
                               'ba': 'blockAlignment',
                               'bs': 'blockSize',
                               'bsh': 'buttonShape',
                               'bw': 'buttonWidth',
                               'e': 'edit',
                               'l': 'label',
                               'lfs': 'labelFontSize',
                               'p': 'padding',
                               'pc': 'pressCommand',
                               'q': 'query',
                               'rc': 'releaseCommand',
                               's': 'section',
                               'vis': 'visible'},
                'type': 'windows'},
  'hudSlider': {'callbackFlags': ['releaseCommand',
                                  'rc',
                                  'dragCommand',
                                  'dc',
                                  'pressCommand',
                                  'pc'],
                'flags': {'allowOverlap': {'args': '<type __builtin__.bool>',
                                           'longname': 'allowOverlap',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ao'},
                          'block': {'args': '<type __builtin__.int>',
                                    'longname': 'block',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'b'},
                          'blockAlignment': {'args': '<type __builtin__.str>',
                                             'longname': 'blockAlignment',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ba'},
                          'blockSize': {'args': '<type __builtin__.str>',
                                        'longname': 'blockSize',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'bs'},
                          'decimalPrecision': {'args': '<type __builtin__.int>',
                                               'longname': 'decimalPrecision',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dp'},
                          'dragCommand': {'args': 'script',
                                          'longname': 'dragCommand',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dc'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'internalPadding': {'args': '<type __builtin__.int>',
                                              'longname': 'internalPadding',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ip'},
                          'label': {'args': '<type __builtin__.str>',
                                    'longname': 'label',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'l'},
                          'labelFontSize': {'args': '<type __builtin__.str>',
                                            'longname': 'labelFontSize',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'lfs'},
                          'labelWidth': {'args': '<type __builtin__.int>',
                                         'longname': 'labelWidth',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'lw'},
                          'maxValue': {'args': '<type __builtin__.float>',
                                       'longname': 'maxValue',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'max'},
                          'minValue': {'args': '<type __builtin__.float>',
                                       'longname': 'minValue',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'min'},
                          'padding': {'args': '<type __builtin__.int>',
                                      'longname': 'padding',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                          'pressCommand': {'args': 'script',
                                           'longname': 'pressCommand',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pc'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'releaseCommand': {'args': 'script',
                                             'longname': 'releaseCommand',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'rc'},
                          'section': {'args': '<type __builtin__.int>',
                                      'longname': 'section',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                          'sliderIncrement': {'args': '<type __builtin__.float>',
                                              'longname': 'sliderIncrement',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'si'},
                          'sliderLength': {'args': '<type __builtin__.int>',
                                           'longname': 'sliderLength',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sl'},
                          'type': {'args': '<type __builtin__.str>',
                                   'longname': 'type',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 't'},
                          'value': {'args': '<type __builtin__.float>',
                                    'longname': 'value',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'v'},
                          'valueAlignment': {'args': '<type __builtin__.str>',
                                             'longname': 'valueAlignment',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'va'},
                          'valueFontSize': {'args': '<type __builtin__.str>',
                                            'longname': 'valueFontSize',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'vfs'},
                          'valueWidth': {'args': '<type __builtin__.int>',
                                         'longname': 'valueWidth',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vw'},
                          'visible': {'args': '<type __builtin__.bool>',
                                      'longname': 'visible',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'vis'}},
                'shortFlags': {'ao': 'allowOverlap',
                               'b': 'block',
                               'ba': 'blockAlignment',
                               'bs': 'blockSize',
                               'dc': 'dragCommand',
                               'dp': 'decimalPrecision',
                               'e': 'edit',
                               'ip': 'internalPadding',
                               'l': 'label',
                               'lfs': 'labelFontSize',
                               'lw': 'labelWidth',
                               'max': 'maxValue',
                               'min': 'minValue',
                               'p': 'padding',
                               'pc': 'pressCommand',
                               'q': 'query',
                               'rc': 'releaseCommand',
                               's': 'section',
                               'si': 'sliderIncrement',
                               'sl': 'sliderLength',
                               't': 'type',
                               'v': 'value',
                               'va': 'valueAlignment',
                               'vfs': 'valueFontSize',
                               'vis': 'visible',
                               'vw': 'valueWidth'},
                'type': 'windows'},
  'hudSliderButton': {'callbackFlags': ['sliderReleaseCommand',
                                        'src',
                                        'buttonPressCommand',
                                        'bpc',
                                        'buttonReleaseCommand',
                                        'brc',
                                        'sliderPressCommand',
                                        'spc',
                                        'sliderDragCommand',
                                        'sdc'],
                      'flags': {'allowOverlap': {'args': '<type __builtin__.bool>',
                                                 'longname': 'allowOverlap',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ao'},
                                'block': {'args': '<type __builtin__.int>',
                                          'longname': 'block',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'b'},
                                'blockAlignment': {'args': '<type __builtin__.str>',
                                                   'longname': 'blockAlignment',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ba'},
                                'blockSize': {'args': '<type __builtin__.str>',
                                              'longname': 'blockSize',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'bs'},
                                'buttonLabel': {'args': '<type __builtin__.str>',
                                                'longname': 'buttonLabel',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'bl'},
                                'buttonLabelFontSize': {'args': '<type __builtin__.str>',
                                                        'longname': 'buttonLabelFontSize',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'bfs'},
                                'buttonPressCommand': {'args': 'script',
                                                       'longname': 'buttonPressCommand',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'bpc'},
                                'buttonReleaseCommand': {'args': 'script',
                                                         'longname': 'buttonReleaseCommand',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'brc'},
                                'buttonShape': {'args': '<type __builtin__.str>',
                                                'longname': 'buttonShape',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'bsh'},
                                'buttonWidth': {'args': '<type __builtin__.int>',
                                                'longname': 'buttonWidth',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'bw'},
                                'decimalPrecision': {'args': '<type __builtin__.int>',
                                                     'longname': 'decimalPrecision',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dp'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'internalPadding': {'args': '<type __builtin__.int>',
                                                    'longname': 'internalPadding',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ip'},
                                'maxValue': {'args': '<type __builtin__.float>',
                                             'longname': 'maxValue',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'max'},
                                'minValue': {'args': '<type __builtin__.float>',
                                             'longname': 'minValue',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'min'},
                                'padding': {'args': '<type __builtin__.int>',
                                            'longname': 'padding',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'p'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'section': {'args': '<type __builtin__.int>',
                                            'longname': 'section',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 's'},
                                'sliderDragCommand': {'args': 'script',
                                                      'longname': 'sliderDragCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'sdc'},
                                'sliderIncrement': {'args': '<type __builtin__.float>',
                                                    'longname': 'sliderIncrement',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'si'},
                                'sliderLabel': {'args': '<type __builtin__.str>',
                                                'longname': 'sliderLabel',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sl'},
                                'sliderLabelFontSize': {'args': '<type __builtin__.str>',
                                                        'longname': 'sliderLabelFontSize',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'sfs'},
                                'sliderLabelWidth': {'args': '<type __builtin__.int>',
                                                     'longname': 'sliderLabelWidth',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'slw'},
                                'sliderLength': {'args': '<type __builtin__.int>',
                                                 'longname': 'sliderLength',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sln'},
                                'sliderPressCommand': {'args': 'script',
                                                       'longname': 'sliderPressCommand',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'spc'},
                                'sliderReleaseCommand': {'args': 'script',
                                                         'longname': 'sliderReleaseCommand',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'src'},
                                'type': {'args': '<type __builtin__.str>',
                                         'longname': 'type',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 't'},
                                'value': {'args': '<type __builtin__.float>',
                                          'longname': 'value',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'v'},
                                'valueAlignment': {'args': '<type __builtin__.str>',
                                                   'longname': 'valueAlignment',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'va'},
                                'valueFontSize': {'args': '<type __builtin__.str>',
                                                  'longname': 'valueFontSize',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'vfs'},
                                'valueWidth': {'args': '<type __builtin__.int>',
                                               'longname': 'valueWidth',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'vw'},
                                'visible': {'args': '<type __builtin__.bool>',
                                            'longname': 'visible',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'vis'}},
                      'shortFlags': {'ao': 'allowOverlap',
                                     'b': 'block',
                                     'ba': 'blockAlignment',
                                     'bfs': 'buttonLabelFontSize',
                                     'bl': 'buttonLabel',
                                     'bpc': 'buttonPressCommand',
                                     'brc': 'buttonReleaseCommand',
                                     'bs': 'blockSize',
                                     'bsh': 'buttonShape',
                                     'bw': 'buttonWidth',
                                     'dp': 'decimalPrecision',
                                     'e': 'edit',
                                     'ip': 'internalPadding',
                                     'max': 'maxValue',
                                     'min': 'minValue',
                                     'p': 'padding',
                                     'q': 'query',
                                     's': 'section',
                                     'sdc': 'sliderDragCommand',
                                     'sfs': 'sliderLabelFontSize',
                                     'si': 'sliderIncrement',
                                     'sl': 'sliderLabel',
                                     'sln': 'sliderLength',
                                     'slw': 'sliderLabelWidth',
                                     'spc': 'sliderPressCommand',
                                     'src': 'sliderReleaseCommand',
                                     't': 'type',
                                     'v': 'value',
                                     'va': 'valueAlignment',
                                     'vfs': 'valueFontSize',
                                     'vis': 'visible',
                                     'vw': 'valueWidth'},
                      'type': 'windows'},
  'hwReflectionMap': {'flags': {'backTextureName': {'args': '<type __builtin__.str>',
                                                    'longname': 'backTextureName',
                                                    'modes': ['query'],
                                                    'numArgs': 1,
                                                    'shortname': 'bkn'},
                                'bottomTextureName': {'args': '<type __builtin__.str>',
                                                      'longname': 'bottomTextureName',
                                                      'modes': ['query'],
                                                      'numArgs': 1,
                                                      'shortname': 'bmn'},
                                'cubeMap': {'args': '<type __builtin__.bool>',
                                            'longname': 'cubeMap',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'cm'},
                                'decalMode': {'args': '<type __builtin__.bool>',
                                              'longname': 'decalMode',
                                              'modes': ['query'],
                                              'numArgs': 1,
                                              'shortname': 'dm'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'enable': {'args': '<type __builtin__.bool>',
                                           'longname': 'enable',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 'en'},
                                'frontTextureName': {'args': '<type __builtin__.str>',
                                                     'longname': 'frontTextureName',
                                                     'modes': ['query'],
                                                     'numArgs': 1,
                                                     'shortname': 'ftn'},
                                'leftTextureName': {'args': '<type __builtin__.str>',
                                                    'longname': 'leftTextureName',
                                                    'modes': ['query'],
                                                    'numArgs': 1,
                                                    'shortname': 'ltn'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'rightTextureName': {'args': '<type __builtin__.str>',
                                                     'longname': 'rightTextureName',
                                                     'modes': ['query'],
                                                     'numArgs': 1,
                                                     'shortname': 'rtn'},
                                'sphereMapTextureName': {'args': '<type __builtin__.str>',
                                                         'longname': 'sphereMapTextureName',
                                                         'modes': ['query'],
                                                         'numArgs': 1,
                                                         'shortname': 'smn'},
                                'topTextureName': {'args': '<type __builtin__.str>',
                                                   'longname': 'topTextureName',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'tpn'}},
                      'shortFlags': {'bkn': 'backTextureName',
                                     'bmn': 'bottomTextureName',
                                     'cm': 'cubeMap',
                                     'dm': 'decalMode',
                                     'e': 'edit',
                                     'en': 'enable',
                                     'ftn': 'frontTextureName',
                                     'ltn': 'leftTextureName',
                                     'q': 'query',
                                     'rtn': 'rightTextureName',
                                     'smn': 'sphereMapTextureName',
                                     'tpn': 'topTextureName'},
                      'type': 'rendering'},
  'hwRender': {'flags': {'acceleratedMultiSampleSupport': {'args': '<type __builtin__.bool>',
                                                           'longname': 'acceleratedMultiSampleSupport',
                                                           'modes': ['query'],
                                                           'numArgs': 0,
                                                           'shortname': 'ams'},
                         'activeTextureCount': {'args': '<type __builtin__.bool>',
                                                'longname': 'activeTextureCount',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'atc'},
                         'camera': {'args': '<type __builtin__.str>',
                                    'longname': 'camera',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'cam'},
                         'currentFrame': {'args': '<type __builtin__.bool>',
                                          'longname': 'currentFrame',
                                          'modes': ['create', 'query'],
                                          'numArgs': 0,
                                          'shortname': 'cf'},
                         'currentView': {'args': '<type __builtin__.bool>',
                                         'longname': 'currentView',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'cv'},
                         'edgeAntiAliasing': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'edgeAntiAliasing',
                                              'modes': ['create', 'query'],
                                              'numArgs': 2,
                                              'shortname': 'eaa'},
                         'fixFileNameNumberPattern': {'args': '<type __builtin__.bool>',
                                                      'longname': 'fixFileNameNumberPattern',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 0,
                                                      'shortname': 'fnp'},
                         'frame': {'args': '<type __builtin__.float>',
                                   'longname': 'frame',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'f'},
                         'fullRenderSupport': {'args': '<type __builtin__.bool>',
                                               'longname': 'fullRenderSupport',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 0,
                                               'shortname': 'frs'},
                         'height': {'args': '<type __builtin__.int>',
                                    'longname': 'height',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'h'},
                         'imageFileName': {'args': '<type __builtin__.bool>',
                                           'longname': 'imageFileName',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'ifn'},
                         'layer': {'args': 'PyNode',
                                   'longname': 'layer',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'l'},
                         'limitedRenderSupport': {'args': '<type __builtin__.bool>',
                                                  'longname': 'limitedRenderSupport',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'lrs'},
                         'lowQualityLighting': {'args': '<type __builtin__.bool>',
                                                'longname': 'lowQualityLighting',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 0,
                                                'shortname': 'lql'},
                         'noRenderView': {'args': '<type __builtin__.bool>',
                                          'longname': 'noRenderView',
                                          'modes': ['create', 'query'],
                                          'numArgs': 0,
                                          'shortname': 'nrv'},
                         'notWriteToFile': {'args': '<type __builtin__.bool>',
                                            'longname': 'notWriteToFile',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'nwf'},
                         'printGeometry': {'args': '<type __builtin__.bool>',
                                           'longname': 'printGeometry',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'pg'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'renderHardwareName': {'args': '<type __builtin__.bool>',
                                                'longname': 'renderHardwareName',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'rhw'},
                         'renderRegion': {'args': ['<type __builtin__.int>',
                                                   '<type __builtin__.int>',
                                                   '<type __builtin__.int>',
                                                   '<type __builtin__.int>'],
                                          'longname': 'renderRegion',
                                          'modes': ['create', 'query'],
                                          'numArgs': 4,
                                          'shortname': 'reg'},
                         'renderSelected': {'args': '<type __builtin__.bool>',
                                            'longname': 'renderSelected',
                                            'modes': ['create', 'query'],
                                            'numArgs': 0,
                                            'shortname': 'rs'},
                         'textureResolution': {'args': '<type __builtin__.int>',
                                               'longname': 'textureResolution',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'res'},
                         'width': {'args': '<type __builtin__.int>',
                                   'longname': 'width',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'w'},
                         'writeAlpha': {'args': '<type __builtin__.bool>',
                                        'longname': 'writeAlpha',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'a'},
                         'writeDepth': {'args': '<type __builtin__.bool>',
                                        'longname': 'writeDepth',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'd'}},
               'shortFlags': {'a': 'writeAlpha',
                              'ams': 'acceleratedMultiSampleSupport',
                              'atc': 'activeTextureCount',
                              'cam': 'camera',
                              'cf': 'currentFrame',
                              'cv': 'currentView',
                              'd': 'writeDepth',
                              'eaa': 'edgeAntiAliasing',
                              'f': 'frame',
                              'fnp': 'fixFileNameNumberPattern',
                              'frs': 'fullRenderSupport',
                              'h': 'height',
                              'ifn': 'imageFileName',
                              'l': 'layer',
                              'lql': 'lowQualityLighting',
                              'lrs': 'limitedRenderSupport',
                              'nrv': 'noRenderView',
                              'nwf': 'notWriteToFile',
                              'pg': 'printGeometry',
                              'q': 'query',
                              'reg': 'renderRegion',
                              'res': 'textureResolution',
                              'rhw': 'renderHardwareName',
                              'rs': 'renderSelected',
                              'w': 'width'},
               'type': 'rendering'},
  'hwRenderLoad': {'flags': {}, 'shortFlags': {}, 'type': 'rendering'},
  'hyperGraph': {'callbackFlags': ['nodeReleaseCommand',
                                   'nr',
                                   'edgeDimmedDblClickCommand',
                                   'edd',
                                   'dragAndDropBehaviorCommand',
                                   'ddc',
                                   'edgeDblClickCommand',
                                   'edc',
                                   'edgeDropCommand',
                                   'edr',
                                   'nodeDblClickCommand',
                                   'ndc',
                                   'edgePressCommand',
                                   'ep',
                                   'directoryPressCommand',
                                   'dp',
                                   'nodeMenuCommand',
                                   'nm',
                                   'nodePressCommand',
                                   'np',
                                   'breakConnectionCommand',
                                   'bco',
                                   'focusCommand',
                                   'fc',
                                   'edgeReleaseCommand',
                                   'er',
                                   'nodeConnectCommand',
                                   'nco',
                                   'nodeDropCommand',
                                   'ndr'],
                 'flags': {'addBookmark': {'args': '<type __builtin__.bool>',
                                           'longname': 'addBookmark',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'abk'},
                           'addDependGraph': {'args': 'PyNode',
                                              'longname': 'addDependGraph',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'adg'},
                           'addDependNode': {'args': 'PyNode',
                                             'longname': 'addDependNode',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'adn'},
                           'animateTransition': {'args': '<type __builtin__.bool>',
                                                 'longname': 'animateTransition',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'atr'},
                           'attributeEditor': {'args': '<type __builtin__.str>',
                                               'longname': 'attributeEditor',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ae'},
                           'backward': {'args': '<type __builtin__.bool>',
                                        'longname': 'backward',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'bak'},
                           'bookmarkName': {'args': '<type __builtin__.bool>',
                                            'longname': 'bookmarkName',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'bn'},
                           'breakConnectionCommand': {'args': '<type __builtin__.str>',
                                                      'longname': 'breakConnectionCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'bco'},
                           'clear': {'args': '<type __builtin__.bool>',
                                     'longname': 'clear',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'clr'},
                           'collapseContainer': {'args': '<type __builtin__.bool>',
                                                 'longname': 'collapseContainer',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'cc'},
                           'connectionDrawStyle': {'args': '<type __builtin__.str>',
                                                   'longname': 'connectionDrawStyle',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cds'},
                           'control': {'args': '<type __builtin__.bool>',
                                       'longname': 'control',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'ctl'},
                           'currentEdge': {'args': '<type __builtin__.str>',
                                           'longname': 'currentEdge',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ced'},
                           'currentNode': {'args': '<type __builtin__.str>',
                                           'longname': 'currentNode',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cno'},
                           'debug': {'args': '<type __builtin__.str>',
                                     'longname': 'debug',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'deb'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'deleteBookmark': {'args': '<type __builtin__.str>',
                                              'longname': 'deleteBookmark',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dbk'},
                           'dependGraph': {'args': '<type __builtin__.bool>',
                                           'longname': 'dependGraph',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'dg'},
                           'dependNode': {'args': '<type __builtin__.str>',
                                          'longname': 'dependNode',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dn'},
                           'directoryPressCommand': {'args': '<type __builtin__.str>',
                                                     'longname': 'directoryPressCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dp'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'down': {'args': '<type __builtin__.bool>',
                                    'longname': 'down',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'do'},
                           'downstream': {'args': '<type __builtin__.bool>',
                                          'longname': 'downstream',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'ds'},
                           'dragAndDropBehaviorCommand': {'args': '<type __builtin__.str>',
                                                          'longname': 'dragAndDropBehaviorCommand',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ddc'},
                           'dropNode': {'args': '<type __builtin__.str>',
                                        'longname': 'dropNode',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'dr'},
                           'dropTargetNode': {'args': '<type __builtin__.str>',
                                              'longname': 'dropTargetNode',
                                              'modes': ['query'],
                                              'numArgs': 1,
                                              'shortname': 'drt'},
                           'edgeDblClickCommand': {'args': '<type __builtin__.str>',
                                                   'longname': 'edgeDblClickCommand',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'edc'},
                           'edgeDimmedDblClickCommand': {'args': '<type __builtin__.str>',
                                                         'longname': 'edgeDimmedDblClickCommand',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'edd'},
                           'edgeDropCommand': {'args': '<type __builtin__.str>',
                                               'longname': 'edgeDropCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'edr'},
                           'edgePressCommand': {'args': '<type __builtin__.str>',
                                                'longname': 'edgePressCommand',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ep'},
                           'edgeReleaseCommand': {'args': '<type __builtin__.str>',
                                                  'longname': 'edgeReleaseCommand',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'er'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'enableAutomaticLayout': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableAutomaticLayout',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'eal'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'expandContainer': {'args': '<type __builtin__.bool>',
                                               'longname': 'expandContainer',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'ec'},
                           'feedbackGadget': {'args': '<type __builtin__.str>',
                                              'longname': 'feedbackGadget',
                                              'modes': ['query'],
                                              'numArgs': 1,
                                              'shortname': 'fbg'},
                           'feedbackNode': {'args': '<type __builtin__.str>',
                                            'longname': 'feedbackNode',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'fbn'},
                           'filter': {'args': '<type __builtin__.str>',
                                      'longname': 'filter',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'f'},
                           'filterDetail': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.bool>'],
                                            'longname': 'filterDetail',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 2,
                                            'shortname': 'fd'},
                           'fitImageToHeight': {'args': '<type __builtin__.bool>',
                                                'longname': 'fitImageToHeight',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'fih'},
                           'fitImageToWidth': {'args': '<type __builtin__.bool>',
                                               'longname': 'fitImageToWidth',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'fiw'},
                           'focusCommand': {'args': '<type __builtin__.str>',
                                            'longname': 'focusCommand',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'fc'},
                           'fold': {'args': '<type __builtin__.bool>',
                                    'longname': 'fold',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'fo'},
                           'forceMainConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'forceMainConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'fmc'},
                           'forceRefresh': {'args': '<type __builtin__.bool>',
                                            'longname': 'forceRefresh',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'frf'},
                           'forward': {'args': '<type __builtin__.bool>',
                                       'longname': 'forward',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'fow'},
                           'frame': {'args': '<type __builtin__.bool>',
                                     'longname': 'frame',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'fr'},
                           'frameBranch': {'args': '<type __builtin__.bool>',
                                           'longname': 'frameBranch',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'frb'},
                           'frameGraph': {'args': '<type __builtin__.bool>',
                                          'longname': 'frameGraph',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'fg'},
                           'frameGraphNoRebuild': {'args': '<type __builtin__.bool>',
                                                   'longname': 'frameGraphNoRebuild',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'fgn'},
                           'frameHierarchy': {'args': '<type __builtin__.bool>',
                                              'longname': 'frameHierarchy',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'frh'},
                           'freeform': {'args': '<type __builtin__.bool>',
                                        'longname': 'freeform',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'fre'},
                           'fromAttr': {'args': '<type __builtin__.str>',
                                        'longname': 'fromAttr',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'fat'},
                           'fromNode': {'args': '<type __builtin__.str>',
                                        'longname': 'fromNode',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'frn'},
                           'getNodeList': {'args': '<type __builtin__.bool>',
                                           'longname': 'getNodeList',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'gnl'},
                           'getNodePosition': {'args': '<type __builtin__.str>',
                                               'longname': 'getNodePosition',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'gnp'},
                           'graphDescription': {'args': '<type __builtin__.bool>',
                                                'longname': 'graphDescription',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'gd'},
                           'graphLayoutStyle': {'args': '<type __builtin__.str>',
                                                'longname': 'graphLayoutStyle',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'gls'},
                           'graphType': {'args': '<type __builtin__.str>',
                                         'longname': 'graphType',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'gt'},
                           'heatMapDisplay': {'args': '<type __builtin__.bool>',
                                              'longname': 'heatMapDisplay',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'hmd'},
                           'highlightConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'highlightConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'hlc'},
                           'iconSize': {'args': '<type __builtin__.str>',
                                        'longname': 'iconSize',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ics'},
                           'image': {'args': '<type __builtin__.str>',
                                     'longname': 'image',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'img'},
                           'imageEnabled': {'args': '<type __builtin__.bool>',
                                            'longname': 'imageEnabled',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ime'},
                           'imageForContainer': {'args': '<type __builtin__.bool>',
                                                 'longname': 'imageForContainer',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'ifc'},
                           'imagePosition': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'imagePosition',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 2,
                                             'shortname': 'imp'},
                           'imageScale': {'args': '<type __builtin__.float>',
                                          'longname': 'imageScale',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ims'},
                           'initializeScript': {'args': '<type __builtin__.str>',
                                                'longname': 'initializeScript',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ini'},
                           'isHotkeyTarget': {'args': '<type __builtin__.bool>',
                                              'longname': 'isHotkeyTarget',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'iht'},
                           'layout': {'args': '<type __builtin__.bool>',
                                      'longname': 'layout',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'lay'},
                           'layoutSelected': {'args': '<type __builtin__.str>',
                                              'longname': 'layoutSelected',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lsl'},
                           'layoutSubtree': {'args': '<type __builtin__.bool>',
                                             'longname': 'layoutSubtree',
                                             'numArgs': 0,
                                             'shortname': 'lsb'},
                           'limitGraphTraversal': {'args': '<type __builtin__.int>',
                                                   'longname': 'limitGraphTraversal',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'lgt'},
                           'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                  'longname': 'lockMainConnection',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'lck'},
                           'look': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'look',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 2,
                                    'shortname': 'loo'},
                           'mainListConnection': {'args': '<type __builtin__.str>',
                                                  'longname': 'mainListConnection',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mlc'},
                           'mergeConnections': {'args': '<type __builtin__.bool>',
                                                'longname': 'mergeConnections',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'mc'},
                           'navigateHome': {'args': '<type __builtin__.bool>',
                                            'longname': 'navigateHome',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'hom'},
                           'navup': {'args': '<type __builtin__.bool>',
                                     'longname': 'navup',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'nu'},
                           'newInputConnection': {'args': '<type __builtin__.str>',
                                                  'longname': 'newInputConnection',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'nin'},
                           'newOutputConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'newOutputConnection',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'nou'},
                           'nextView': {'args': '<type __builtin__.bool>',
                                        'longname': 'nextView',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'nvw'},
                           'nodeConnectCommand': {'args': '<type __builtin__.str>',
                                                  'longname': 'nodeConnectCommand',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'nco'},
                           'nodeDblClickCommand': {'args': '<type __builtin__.str>',
                                                   'longname': 'nodeDblClickCommand',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ndc'},
                           'nodeDropCommand': {'args': '<type __builtin__.str>',
                                               'longname': 'nodeDropCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ndr'},
                           'nodeMenuCommand': {'args': '<type __builtin__.str>',
                                               'longname': 'nodeMenuCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nm'},
                           'nodePressCommand': {'args': '<type __builtin__.str>',
                                                'longname': 'nodePressCommand',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'np'},
                           'nodeReleaseCommand': {'args': '<type __builtin__.str>',
                                                  'longname': 'nodeReleaseCommand',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'nr'},
                           'opaqueContainers': {'args': '<type __builtin__.bool>',
                                                'longname': 'opaqueContainers',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'opc'},
                           'orientation': {'args': '<type __builtin__.str>',
                                           'longname': 'orientation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'orientation'},
                           'panView': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'panView',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 2,
                                       'shortname': 'pan'},
                           'panel': {'args': '<type __builtin__.str>',
                                     'longname': 'panel',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'pnl'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuScript': {'args': '<type __builtin__.str>',
                                               'longname': 'popupMenuScript',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'pms'},
                           'previousView': {'args': '<type __builtin__.bool>',
                                            'longname': 'previousView',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'pvw'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'range': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'range',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 2,
                                     'shortname': 'rg'},
                           'rebuild': {'args': '<type __builtin__.bool>',
                                       'longname': 'rebuild',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'rb'},
                           'removeNode': {'args': '<type __builtin__.str>',
                                          'longname': 'removeNode',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'rmn'},
                           'rename': {'args': '<type __builtin__.bool>',
                                      'longname': 'rename',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'rn'},
                           'resetFreeform': {'args': '<type __builtin__.bool>',
                                             'longname': 'resetFreeform',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'rf'},
                           'restoreBookmark': {'args': '<type __builtin__.str>',
                                               'longname': 'restoreBookmark',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rbk'},
                           'scrollUpDownNoZoom': {'args': '<type __builtin__.bool>',
                                                  'longname': 'scrollUpDownNoZoom',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'snz'},
                           'selectionConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'selectionConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'slc'},
                           'setNodePosition': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'setNodePosition',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'snp'},
                           'showCachedConnections': {'args': '<type __builtin__.bool>',
                                                     'longname': 'showCachedConnections',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'scc'},
                           'showConnectionFromSelected': {'args': '<type __builtin__.bool>',
                                                          'longname': 'showConnectionFromSelected',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'scf'},
                           'showConnectionToSelected': {'args': '<type __builtin__.bool>',
                                                        'longname': 'showConnectionToSelected',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'sct'},
                           'showConstraintLabels': {'args': '<type __builtin__.bool>',
                                                    'longname': 'showConstraintLabels',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'scl'},
                           'showConstraints': {'args': '<type __builtin__.bool>',
                                               'longname': 'showConstraints',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'shc'},
                           'showDeformers': {'args': '<type __builtin__.bool>',
                                             'longname': 'showDeformers',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'shd'},
                           'showExpressions': {'args': '<type __builtin__.bool>',
                                               'longname': 'showExpressions',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'shx'},
                           'showInvisible': {'args': '<type __builtin__.bool>',
                                             'longname': 'showInvisible',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'shi'},
                           'showRelationships': {'args': '<type __builtin__.bool>',
                                                 'longname': 'showRelationships',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'shr'},
                           'showShapes': {'args': '<type __builtin__.bool>',
                                          'longname': 'showShapes',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'shs'},
                           'showUnderworld': {'args': '<type __builtin__.bool>',
                                              'longname': 'showUnderworld',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'shu'},
                           'stateString': {'args': '<type __builtin__.bool>',
                                           'longname': 'stateString',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'sts'},
                           'toAttr': {'args': '<type __builtin__.str>',
                                      'longname': 'toAttr',
                                      'modes': ['query'],
                                      'numArgs': 1,
                                      'shortname': 'tat'},
                           'toNode': {'args': '<type __builtin__.str>',
                                      'longname': 'toNode',
                                      'modes': ['query'],
                                      'numArgs': 1,
                                      'shortname': 'ton'},
                           'transitionFrames': {'args': '<type __builtin__.int>',
                                                'longname': 'transitionFrames',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'tfr'},
                           'unParent': {'args': '<type __builtin__.bool>',
                                        'longname': 'unParent',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'up'},
                           'unfold': {'args': '<type __builtin__.bool>',
                                      'longname': 'unfold',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'uf'},
                           'unfoldAll': {'args': '<type __builtin__.bool>',
                                         'longname': 'unfoldAll',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'ua'},
                           'unfoldAllShapes': {'args': '<type __builtin__.bool>',
                                               'longname': 'unfoldAllShapes',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'uas'},
                           'unfoldHidden': {'args': '<type __builtin__.bool>',
                                            'longname': 'unfoldHidden',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'ufh'},
                           'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'unlockMainConnection',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'ulk'},
                           'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'updateMainConnection',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'upd'},
                           'updateNodeAdded': {'args': '<type __builtin__.bool>',
                                               'longname': 'updateNodeAdded',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'una'},
                           'updateSelection': {'args': '<type __builtin__.bool>',
                                               'longname': 'updateSelection',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'us'},
                           'upstream': {'args': '<type __builtin__.bool>',
                                        'longname': 'upstream',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'ups'},
                           'useDrawOverrideColor': {'args': '<type __builtin__.bool>',
                                                    'longname': 'useDrawOverrideColor',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'drc'},
                           'useFeedbackList': {'args': '<type __builtin__.bool>',
                                               'longname': 'useFeedbackList',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 0,
                                               'shortname': 'ufl'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'},
                           'viewOption': {'args': '<type __builtin__.str>',
                                          'longname': 'viewOption',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'vo'},
                           'visibility': {'args': '<type __builtin__.bool>',
                                          'longname': 'visibility',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'vis'},
                           'zoom': {'args': '<type __builtin__.float>',
                                    'longname': 'zoom',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'zm'}},
                 'removedFlags': {'or': 'orientation'},
                 'shortFlags': {'abk': 'addBookmark',
                                'adg': 'addDependGraph',
                                'adn': 'addDependNode',
                                'ae': 'attributeEditor',
                                'atr': 'animateTransition',
                                'bak': 'backward',
                                'bco': 'breakConnectionCommand',
                                'bn': 'bookmarkName',
                                'cc': 'collapseContainer',
                                'cds': 'connectionDrawStyle',
                                'ced': 'currentEdge',
                                'clr': 'clear',
                                'cno': 'currentNode',
                                'ctl': 'control',
                                'dbk': 'deleteBookmark',
                                'ddc': 'dragAndDropBehaviorCommand',
                                'deb': 'debug',
                                'dg': 'dependGraph',
                                'dn': 'dependNode',
                                'do': 'down',
                                'dp': 'directoryPressCommand',
                                'dr': 'dropNode',
                                'drc': 'useDrawOverrideColor',
                                'drt': 'dropTargetNode',
                                'ds': 'downstream',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'eal': 'enableAutomaticLayout',
                                'ec': 'expandContainer',
                                'edc': 'edgeDblClickCommand',
                                'edd': 'edgeDimmedDblClickCommand',
                                'edr': 'edgeDropCommand',
                                'ep': 'edgePressCommand',
                                'er': 'edgeReleaseCommand',
                                'ex': 'exists',
                                'f': 'filter',
                                'fat': 'fromAttr',
                                'fbg': 'feedbackGadget',
                                'fbn': 'feedbackNode',
                                'fc': 'focusCommand',
                                'fd': 'filterDetail',
                                'fg': 'frameGraph',
                                'fgn': 'frameGraphNoRebuild',
                                'fih': 'fitImageToHeight',
                                'fiw': 'fitImageToWidth',
                                'fmc': 'forceMainConnection',
                                'fo': 'fold',
                                'fow': 'forward',
                                'fr': 'frame',
                                'frb': 'frameBranch',
                                'fre': 'freeform',
                                'frf': 'forceRefresh',
                                'frh': 'frameHierarchy',
                                'frn': 'fromNode',
                                'gd': 'graphDescription',
                                'gls': 'graphLayoutStyle',
                                'gnl': 'getNodeList',
                                'gnp': 'getNodePosition',
                                'gt': 'graphType',
                                'hlc': 'highlightConnection',
                                'hmd': 'heatMapDisplay',
                                'hom': 'navigateHome',
                                'ics': 'iconSize',
                                'ifc': 'imageForContainer',
                                'iht': 'isHotkeyTarget',
                                'ime': 'imageEnabled',
                                'img': 'image',
                                'imp': 'imagePosition',
                                'ims': 'imageScale',
                                'ini': 'initializeScript',
                                'lay': 'layout',
                                'lck': 'lockMainConnection',
                                'lgt': 'limitGraphTraversal',
                                'loo': 'look',
                                'lsb': 'layoutSubtree',
                                'lsl': 'layoutSelected',
                                'mc': 'mergeConnections',
                                'mlc': 'mainListConnection',
                                'nco': 'nodeConnectCommand',
                                'ndc': 'nodeDblClickCommand',
                                'ndr': 'nodeDropCommand',
                                'nin': 'newInputConnection',
                                'nm': 'nodeMenuCommand',
                                'nou': 'newOutputConnection',
                                'np': 'nodePressCommand',
                                'nr': 'nodeReleaseCommand',
                                'nu': 'navup',
                                'nvw': 'nextView',
                                'opc': 'opaqueContainers',
                                'orientation': 'orientation',
                                'p': 'parent',
                                'pan': 'panView',
                                'pms': 'popupMenuScript',
                                'pnl': 'panel',
                                'pvw': 'previousView',
                                'q': 'query',
                                'rb': 'rebuild',
                                'rbk': 'restoreBookmark',
                                'rf': 'resetFreeform',
                                'rg': 'range',
                                'rmn': 'removeNode',
                                'rn': 'rename',
                                'scc': 'showCachedConnections',
                                'scf': 'showConnectionFromSelected',
                                'scl': 'showConstraintLabels',
                                'sct': 'showConnectionToSelected',
                                'shc': 'showConstraints',
                                'shd': 'showDeformers',
                                'shi': 'showInvisible',
                                'shr': 'showRelationships',
                                'shs': 'showShapes',
                                'shu': 'showUnderworld',
                                'shx': 'showExpressions',
                                'slc': 'selectionConnection',
                                'snp': 'setNodePosition',
                                'snz': 'scrollUpDownNoZoom',
                                'sts': 'stateString',
                                'tat': 'toAttr',
                                'tfr': 'transitionFrames',
                                'ton': 'toNode',
                                'ua': 'unfoldAll',
                                'uas': 'unfoldAllShapes',
                                'uf': 'unfold',
                                'ufh': 'unfoldHidden',
                                'ufl': 'useFeedbackList',
                                'ulk': 'unlockMainConnection',
                                'una': 'updateNodeAdded',
                                'up': 'unParent',
                                'upd': 'updateMainConnection',
                                'ups': 'upstream',
                                'us': 'updateSelection',
                                'ut': 'useTemplate',
                                'vis': 'visibility',
                                'vo': 'viewOption',
                                'zm': 'zoom'},
                 'type': 'windows'},
  'hyperPanel': {'callbackFlags': ['popupMenuProcedure', 'pmp'],
                 'flags': {'control': {'args': '<type __builtin__.bool>',
                                       'longname': 'control',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'ctl'},
                           'copy': {'args': '<type __builtin__.str>',
                                    'longname': 'copy',
                                    'modes': ['edit'],
                                    'numArgs': 1,
                                    'shortname': 'cp'},
                           'createString': {'args': '<type __builtin__.bool>',
                                            'longname': 'createString',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'cs'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'editString': {'args': '<type __builtin__.bool>',
                                          'longname': 'editString',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'es'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'hyperEditor': {'args': '<type __builtin__.bool>',
                                           'longname': 'hyperEditor',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'he'},
                           'init': {'args': '<type __builtin__.bool>',
                                    'longname': 'init',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'init'},
                           'isUnique': {'args': '<type __builtin__.bool>',
                                        'longname': 'isUnique',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'iu'},
                           'label': {'args': '<type __builtin__.str>',
                                     'longname': 'label',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'l'},
                           'menuBarRepeatLast': {'args': '<type __builtin__.bool>',
                                                 'longname': 'menuBarRepeatLast',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'mrl'},
                           'menuBarVisible': {'args': '<type __builtin__.bool>',
                                              'longname': 'menuBarVisible',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'mbv'},
                           'needsInit': {'args': '<type __builtin__.bool>',
                                         'longname': 'needsInit',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'ni'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuProcedure': {'args': 'script',
                                                  'longname': 'popupMenuProcedure',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pmp'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'replacePanel': {'args': '<type __builtin__.str>',
                                            'longname': 'replacePanel',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'rp'},
                           'tearOff': {'args': '<type __builtin__.bool>',
                                       'longname': 'tearOff',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'to'},
                           'tearOffCopy': {'args': '<type __builtin__.str>',
                                           'longname': 'tearOffCopy',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'toc'},
                           'tearOffRestore': {'args': '<type __builtin__.bool>',
                                              'longname': 'tearOffRestore',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'tor'},
                           'unParent': {'args': '<type __builtin__.bool>',
                                        'longname': 'unParent',
                                        'modes': ['edit'],
                                        'numArgs': 0,
                                        'shortname': 'up'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'}},
                 'removedFlags': {'in': 'init'},
                 'shortFlags': {'cp': 'copy',
                                'cs': 'createString',
                                'ctl': 'control',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'es': 'editString',
                                'ex': 'exists',
                                'he': 'hyperEditor',
                                'init': 'init',
                                'iu': 'isUnique',
                                'l': 'label',
                                'mbv': 'menuBarVisible',
                                'mrl': 'menuBarRepeatLast',
                                'ni': 'needsInit',
                                'p': 'parent',
                                'pmp': 'popupMenuProcedure',
                                'q': 'query',
                                'rp': 'replacePanel',
                                'to': 'tearOff',
                                'toc': 'tearOffCopy',
                                'tor': 'tearOffRestore',
                                'up': 'unParent',
                                'ut': 'useTemplate'},
                 'type': 'windows'},
  'hyperShade': {'flags': {'assign': {'args': '<type __builtin__.str>',
                                      'longname': 'assign',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'a'},
                           'clearWorkArea': {'args': '<type __builtin__.bool>',
                                             'longname': 'clearWorkArea',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'cwa'},
                           'collapse': {'args': '<type __builtin__.str>',
                                        'longname': 'collapse',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'clp'},
                           'createNode': {'args': '<type __builtin__.str>',
                                          'longname': 'createNode',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'rcn'},
                           'dependGraphArea': {'args': '<type __builtin__.bool>',
                                               'longname': 'dependGraphArea',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dg'},
                           'downStream': {'args': '<type __builtin__.bool>',
                                          'longname': 'downStream',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ds'},
                           'duplicate': {'args': '<type __builtin__.bool>',
                                         'longname': 'duplicate',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'dup'},
                           'fixRenderSize': {'args': '<type __builtin__.bool>',
                                             'longname': 'fixRenderSize',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'fix'},
                           'incremental': {'args': '<type __builtin__.bool>',
                                           'longname': 'incremental',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'inc'},
                           'listDownstreamNodes': {'args': 'PyNode',
                                                   'longname': 'listDownstreamNodes',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ldn'},
                           'listDownstreamShaderNodes': {'args': 'PyNode',
                                                         'longname': 'listDownstreamShaderNodes',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'lds'},
                           'listUpstreamNodes': {'args': 'PyNode',
                                                 'longname': 'listUpstreamNodes',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'lun'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'networks': {'args': '<type __builtin__.bool>',
                                        'longname': 'networks',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'net'},
                           'noSGShapes': {'args': '<type __builtin__.bool>',
                                          'longname': 'noSGShapes',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'nsg'},
                           'noShapes': {'args': '<type __builtin__.bool>',
                                        'longname': 'noShapes',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ns'},
                           'noTransforms': {'args': '<type __builtin__.bool>',
                                            'longname': 'noTransforms',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'nt'},
                           'objects': {'args': '<type __builtin__.str>',
                                       'longname': 'objects',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                           'renderCreateAndDrop': {'args': '<type __builtin__.str>',
                                                   'longname': 'renderCreateAndDrop',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'rcd'},
                           'reset': {'args': '<type __builtin__.bool>',
                                     'longname': 'reset',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'rst'},
                           'resetGraph': {'args': '<type __builtin__.bool>',
                                          'longname': 'resetGraph',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'rsg'},
                           'resetSwatch': {'args': '<type __builtin__.bool>',
                                           'longname': 'resetSwatch',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'rss'},
                           'setAllowsRegraphing': {'args': '<type __builtin__.bool>',
                                                   'longname': 'setAllowsRegraphing',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'sar'},
                           'setWorkArea': {'args': '<type __builtin__.str>',
                                           'longname': 'setWorkArea',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'swa'},
                           'shaderNetwork': {'args': '<type __builtin__.str>',
                                             'longname': 'shaderNetwork',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'sn'},
                           'shaderNetworks': {'args': '<type __builtin__.bool>',
                                              'longname': 'shaderNetworks',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'sns'},
                           'shaderNetworksSelectMaterialNodes': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'shaderNetworksSelectMaterialNodes',
                                                                 'modes': ['create'],
                                                                 'numArgs': 0,
                                                                 'shortname': 'smn'},
                           'snapShot': {'args': '<type __builtin__.bool>',
                                        'longname': 'snapShot',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'snp'},
                           'uncollapse': {'args': '<type __builtin__.str>',
                                          'longname': 'uncollapse',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ucl'},
                           'upStream': {'args': '<type __builtin__.bool>',
                                        'longname': 'upStream',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ups'},
                           'userDefinedLayout': {'args': '<type __builtin__.bool>',
                                                 'longname': 'userDefinedLayout',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'udl'},
                           'workAreaAddCmd': {'args': '<type __builtin__.str>',
                                              'longname': 'workAreaAddCmd',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'waa'},
                           'workAreaDeleteCmd': {'args': '<type __builtin__.str>',
                                                 'longname': 'workAreaDeleteCmd',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'wad'},
                           'workAreaSelectCmd': {'args': '<type __builtin__.str>',
                                                 'longname': 'workAreaSelectCmd',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'was'}},
                 'shortFlags': {'a': 'assign',
                                'clp': 'collapse',
                                'cwa': 'clearWorkArea',
                                'dg': 'dependGraphArea',
                                'ds': 'downStream',
                                'dup': 'duplicate',
                                'fix': 'fixRenderSize',
                                'inc': 'incremental',
                                'ldn': 'listDownstreamNodes',
                                'lds': 'listDownstreamShaderNodes',
                                'lun': 'listUpstreamNodes',
                                'n': 'name',
                                'net': 'networks',
                                'ns': 'noShapes',
                                'nsg': 'noSGShapes',
                                'nt': 'noTransforms',
                                'o': 'objects',
                                'rcd': 'renderCreateAndDrop',
                                'rcn': 'createNode',
                                'rsg': 'resetGraph',
                                'rss': 'resetSwatch',
                                'rst': 'reset',
                                'sar': 'setAllowsRegraphing',
                                'smn': 'shaderNetworksSelectMaterialNodes',
                                'sn': 'shaderNetwork',
                                'snp': 'snapShot',
                                'sns': 'shaderNetworks',
                                'swa': 'setWorkArea',
                                'ucl': 'uncollapse',
                                'udl': 'userDefinedLayout',
                                'ups': 'upStream',
                                'waa': 'workAreaAddCmd',
                                'wad': 'workAreaDeleteCmd',
                                'was': 'workAreaSelectCmd'},
                 'type': 'windows'},
  'iconTextButton': {'callbackFlags': ['doubleClickCommand',
                                       'dcc',
                                       'visibleChangeCommand',
                                       'vcc',
                                       'dragCallback',
                                       'dgc',
                                       'labelEditingCallback',
                                       'lec',
                                       'commandRepeatable',
                                       'rpt',
                                       'handleNodeDropCallback',
                                       'hnd',
                                       'command',
                                       'c',
                                       'dropCallback',
                                       'dpc'],
                     'flags': {'actionIsSubstitute': {'args': '<type __builtin__.bool>',
                                                      'longname': 'actionIsSubstitute',
                                                      'numArgs': 1,
                                                      'shortname': 'ais'},
                               'align': {'args': '<type __builtin__.str>',
                                         'longname': 'align',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'al'},
                               'annotation': {'args': '<type __builtin__.str>',
                                              'longname': 'annotation',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ann'},
                               'backgroundColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'backgroundColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'bgc'},
                               'command': {'args': 'script',
                                           'longname': 'command',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'c'},
                               'commandRepeatable': {'args': '<type __builtin__.bool>',
                                                     'longname': 'commandRepeatable',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'rpt'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'disabledImage': {'args': '<type __builtin__.str>',
                                                 'longname': 'disabledImage',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'di'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'doubleClickCommand': {'args': 'script',
                                                      'longname': 'doubleClickCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dcc'},
                               'dragCallback': {'args': 'script',
                                                'longname': 'dragCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dgc'},
                               'dropCallback': {'args': 'script',
                                                'longname': 'dropCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dpc'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'enable': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en'},
                               'enableBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableBackground',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ebg'},
                               'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                       'longname': 'enableKeyboardFocus',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ekf'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'flat': {'args': '<type __builtin__.bool>',
                                        'longname': 'flat',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'fla'},
                               'flipX': {'args': '<type __builtin__.bool>',
                                         'longname': 'flipX',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'fx'},
                               'flipY': {'args': '<type __builtin__.bool>',
                                         'longname': 'flipY',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'fy'},
                               'font': {'args': '<type __builtin__.str>',
                                        'longname': 'font',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'fn'},
                               'fullPathName': {'args': '<type __builtin__.bool>',
                                                'longname': 'fullPathName',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'fpn'},
                               'handleNodeDropCallback': {'args': 'script',
                                                          'longname': 'handleNodeDropCallback',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'hnd'},
                               'height': {'args': '<type __builtin__.int>',
                                          'longname': 'height',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'h'},
                               'highlightColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'highlightColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'hlc'},
                               'highlightImage': {'args': '<type __builtin__.str>',
                                                  'longname': 'highlightImage',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'hi'},
                               'image': {'args': '<type __builtin__.str>',
                                         'longname': 'image',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'imageOverlayLabel': {'args': '<type __builtin__.str>',
                                                     'longname': 'imageOverlayLabel',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'iol'},
                               'isObscured': {'args': '<type __builtin__.bool>',
                                              'longname': 'isObscured',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'io'},
                               'label': {'args': '<type __builtin__.str>',
                                         'longname': 'label',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'l'},
                               'labelEditingCallback': {'args': 'script',
                                                        'longname': 'labelEditingCallback',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'lec'},
                               'labelOffset': {'args': '<type __builtin__.int>',
                                               'longname': 'labelOffset',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'lo'},
                               'ltVersion': {'args': '<type __builtin__.str>',
                                             'longname': 'ltVersion',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lt'},
                               'manage': {'args': '<type __builtin__.bool>',
                                          'longname': 'manage',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'm'},
                               'marginHeight': {'args': '<type __builtin__.int>',
                                                'longname': 'marginHeight',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'mh'},
                               'marginWidth': {'args': '<type __builtin__.int>',
                                               'longname': 'marginWidth',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'mw'},
                               'noBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'noBackground',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nbg'},
                               'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'numberOfPopupMenus',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'npm'},
                               'overlayLabelBackColor': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'overlayLabelBackColor',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 4,
                                                         'shortname': 'olb'},
                               'overlayLabelColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'overlayLabelColor',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'olc'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                  'longname': 'popupMenuArray',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'pma'},
                               'preventOverride': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preventOverride',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'po'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'rotation': {'args': '<type __builtin__.float>',
                                            'longname': 'rotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'rot'},
                               'scaleIcon': {'args': '<type __builtin__.bool>',
                                             'longname': 'scaleIcon',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'sic'},
                               'selectionImage': {'args': '<type __builtin__.str>',
                                                  'longname': 'selectionImage',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'si'},
                               'sourceType': {'args': '<type __builtin__.str>',
                                              'longname': 'sourceType',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'stp'},
                               'statusBarMessage': {'args': '<type __builtin__.str>',
                                                    'longname': 'statusBarMessage',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sbm'},
                               'style': {'args': '<type __builtin__.str>',
                                         'longname': 'style',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'st'},
                               'useAlpha': {'args': '<type __builtin__.bool>',
                                            'longname': 'useAlpha',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ua'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'},
                               'version': {'args': '<type __builtin__.str>',
                                           'longname': 'version',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ver'},
                               'visible': {'args': '<type __builtin__.bool>',
                                           'longname': 'visible',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vis'},
                               'visibleChangeCommand': {'args': 'script',
                                                        'longname': 'visibleChangeCommand',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'vcc'},
                               'width': {'args': '<type __builtin__.int>',
                                         'longname': 'width',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'w'}},
                     'shortFlags': {'ais': 'actionIsSubstitute',
                                    'al': 'align',
                                    'ann': 'annotation',
                                    'bgc': 'backgroundColor',
                                    'c': 'command',
                                    'dcc': 'doubleClickCommand',
                                    'dgc': 'dragCallback',
                                    'di': 'disabledImage',
                                    'dpc': 'dropCallback',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'e': 'edit',
                                    'ebg': 'enableBackground',
                                    'ekf': 'enableKeyboardFocus',
                                    'en': 'enable',
                                    'ex': 'exists',
                                    'fla': 'flat',
                                    'fn': 'font',
                                    'fpn': 'fullPathName',
                                    'fx': 'flipX',
                                    'fy': 'flipY',
                                    'h': 'height',
                                    'hi': 'highlightImage',
                                    'hlc': 'highlightColor',
                                    'hnd': 'handleNodeDropCallback',
                                    'i': 'image',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'io': 'isObscured',
                                    'iol': 'imageOverlayLabel',
                                    'l': 'label',
                                    'lec': 'labelEditingCallback',
                                    'lo': 'labelOffset',
                                    'lt': 'ltVersion',
                                    'm': 'manage',
                                    'mh': 'marginHeight',
                                    'mw': 'marginWidth',
                                    'nbg': 'noBackground',
                                    'npm': 'numberOfPopupMenus',
                                    'olb': 'overlayLabelBackColor',
                                    'olc': 'overlayLabelColor',
                                    'p': 'parent',
                                    'pma': 'popupMenuArray',
                                    'po': 'preventOverride',
                                    'q': 'query',
                                    'rot': 'rotation',
                                    'rpt': 'commandRepeatable',
                                    'sbm': 'statusBarMessage',
                                    'si': 'selectionImage',
                                    'sic': 'scaleIcon',
                                    'st': 'style',
                                    'stp': 'sourceType',
                                    'ua': 'useAlpha',
                                    'ut': 'useTemplate',
                                    'vcc': 'visibleChangeCommand',
                                    'ver': 'version',
                                    'vis': 'visible',
                                    'w': 'width'},
                     'type': 'windows'},
  'iconTextCheckBox': {'callbackFlags': ['visibleChangeCommand',
                                         'vcc',
                                         'dragCallback',
                                         'dgc',
                                         'offCommand',
                                         'ofc',
                                         'changeCommand',
                                         'cc',
                                         'onCommand',
                                         'onc',
                                         'dropCallback',
                                         'dpc'],
                       'flags': {'align': {'args': '<type __builtin__.str>',
                                           'longname': 'align',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'al'},
                                 'annotation': {'args': '<type __builtin__.str>',
                                                'longname': 'annotation',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ann'},
                                 'backgroundColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'backgroundColor',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'bgc'},
                                 'changeCommand': {'args': 'script',
                                                   'longname': 'changeCommand',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cc'},
                                 'defineTemplate': {'args': '<type __builtin__.str>',
                                                    'longname': 'defineTemplate',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'dt'},
                                 'disabledImage': {'args': '<type __builtin__.str>',
                                                   'longname': 'disabledImage',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'di'},
                                 'docTag': {'args': '<type __builtin__.str>',
                                            'longname': 'docTag',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dtg'},
                                 'dragCallback': {'args': 'script',
                                                  'longname': 'dragCallback',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dgc'},
                                 'dropCallback': {'args': 'script',
                                                  'longname': 'dropCallback',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dpc'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'enable': {'args': '<type __builtin__.bool>',
                                            'longname': 'enable',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'en'},
                                 'enableBackground': {'args': '<type __builtin__.bool>',
                                                      'longname': 'enableBackground',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ebg'},
                                 'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                         'longname': 'enableKeyboardFocus',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ekf'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'flat': {'args': '<type __builtin__.bool>',
                                          'longname': 'flat',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'fla'},
                                 'flipX': {'args': '<type __builtin__.bool>',
                                           'longname': 'flipX',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'fx'},
                                 'flipY': {'args': '<type __builtin__.bool>',
                                           'longname': 'flipY',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'fy'},
                                 'font': {'args': '<type __builtin__.str>',
                                          'longname': 'font',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'fn'},
                                 'fullPathName': {'args': '<type __builtin__.bool>',
                                                  'longname': 'fullPathName',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'fpn'},
                                 'height': {'args': '<type __builtin__.int>',
                                            'longname': 'height',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'h'},
                                 'highlightColor': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'highlightColor',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'hlc'},
                                 'highlightImage': {'args': '<type __builtin__.str>',
                                                    'longname': 'highlightImage',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'hi'},
                                 'image': {'args': '<type __builtin__.str>',
                                           'longname': 'image',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i'},
                                 'image1': {'args': '<type __builtin__.str>',
                                            'longname': 'image1',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i1'},
                                 'image2': {'args': '<type __builtin__.str>',
                                            'longname': 'image2',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i2'},
                                 'image3': {'args': '<type __builtin__.str>',
                                            'longname': 'image3',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i3'},
                                 'imageOverlayLabel': {'args': '<type __builtin__.str>',
                                                       'longname': 'imageOverlayLabel',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'iol'},
                                 'isObscured': {'args': '<type __builtin__.bool>',
                                                'longname': 'isObscured',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'io'},
                                 'label': {'args': '<type __builtin__.str>',
                                           'longname': 'label',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'l'},
                                 'labelOffset': {'args': '<type __builtin__.int>',
                                                 'longname': 'labelOffset',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'lo'},
                                 'ltVersion': {'args': '<type __builtin__.str>',
                                               'longname': 'ltVersion',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'lt'},
                                 'manage': {'args': '<type __builtin__.bool>',
                                            'longname': 'manage',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'm'},
                                 'marginHeight': {'args': '<type __builtin__.int>',
                                                  'longname': 'marginHeight',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mh'},
                                 'marginWidth': {'args': '<type __builtin__.int>',
                                                 'longname': 'marginWidth',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'mw'},
                                 'noBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'noBackground',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'nbg'},
                                 'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                        'longname': 'numberOfPopupMenus',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'npm'},
                                 'offCommand': {'args': 'script',
                                                'longname': 'offCommand',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ofc'},
                                 'onCommand': {'args': 'script',
                                               'longname': 'onCommand',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'onc'},
                                 'overlayLabelBackColor': {'args': ['<type __builtin__.float>',
                                                                    '<type __builtin__.float>',
                                                                    '<type __builtin__.float>',
                                                                    '<type __builtin__.float>'],
                                                           'longname': 'overlayLabelBackColor',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 4,
                                                           'shortname': 'olb'},
                                 'overlayLabelColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'overlayLabelColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'olc'},
                                 'parent': {'args': '<type __builtin__.str>',
                                            'longname': 'parent',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'p'},
                                 'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                    'longname': 'popupMenuArray',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'pma'},
                                 'preventOverride': {'args': '<type __builtin__.bool>',
                                                     'longname': 'preventOverride',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'po'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'rotation': {'args': '<type __builtin__.float>',
                                              'longname': 'rotation',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'rot'},
                                 'selectionHighlightImage': {'args': '<type __builtin__.str>',
                                                             'longname': 'selectionHighlightImage',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'shi'},
                                 'selectionImage': {'args': '<type __builtin__.str>',
                                                    'longname': 'selectionImage',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'si'},
                                 'statusBarMessage': {'args': '<type __builtin__.str>',
                                                      'longname': 'statusBarMessage',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'sbm'},
                                 'style': {'args': '<type __builtin__.str>',
                                           'longname': 'style',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'st'},
                                 'useAlpha': {'args': '<type __builtin__.bool>',
                                              'longname': 'useAlpha',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ua'},
                                 'useTemplate': {'args': '<type __builtin__.str>',
                                                 'longname': 'useTemplate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ut'},
                                 'value': {'args': '<type __builtin__.bool>',
                                           'longname': 'value',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'v'},
                                 'version': {'args': '<type __builtin__.str>',
                                             'longname': 'version',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ver'},
                                 'visible': {'args': '<type __builtin__.bool>',
                                             'longname': 'visible',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'vis'},
                                 'visibleChangeCommand': {'args': 'script',
                                                          'longname': 'visibleChangeCommand',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'vcc'},
                                 'width': {'args': '<type __builtin__.int>',
                                           'longname': 'width',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'w'}},
                       'shortFlags': {'al': 'align',
                                      'ann': 'annotation',
                                      'bgc': 'backgroundColor',
                                      'cc': 'changeCommand',
                                      'dgc': 'dragCallback',
                                      'di': 'disabledImage',
                                      'dpc': 'dropCallback',
                                      'dt': 'defineTemplate',
                                      'dtg': 'docTag',
                                      'e': 'edit',
                                      'ebg': 'enableBackground',
                                      'ekf': 'enableKeyboardFocus',
                                      'en': 'enable',
                                      'ex': 'exists',
                                      'fla': 'flat',
                                      'fn': 'font',
                                      'fpn': 'fullPathName',
                                      'fx': 'flipX',
                                      'fy': 'flipY',
                                      'h': 'height',
                                      'hi': 'highlightImage',
                                      'hlc': 'highlightColor',
                                      'i': 'image',
                                      'i1': 'image1',
                                      'i2': 'image2',
                                      'i3': 'image3',
                                      'io': 'isObscured',
                                      'iol': 'imageOverlayLabel',
                                      'l': 'label',
                                      'lo': 'labelOffset',
                                      'lt': 'ltVersion',
                                      'm': 'manage',
                                      'mh': 'marginHeight',
                                      'mw': 'marginWidth',
                                      'nbg': 'noBackground',
                                      'npm': 'numberOfPopupMenus',
                                      'ofc': 'offCommand',
                                      'olb': 'overlayLabelBackColor',
                                      'olc': 'overlayLabelColor',
                                      'onc': 'onCommand',
                                      'p': 'parent',
                                      'pma': 'popupMenuArray',
                                      'po': 'preventOverride',
                                      'q': 'query',
                                      'rot': 'rotation',
                                      'sbm': 'statusBarMessage',
                                      'shi': 'selectionHighlightImage',
                                      'si': 'selectionImage',
                                      'st': 'style',
                                      'ua': 'useAlpha',
                                      'ut': 'useTemplate',
                                      'v': 'value',
                                      'vcc': 'visibleChangeCommand',
                                      'ver': 'version',
                                      'vis': 'visible',
                                      'w': 'width'},
                       'type': 'windows'},
  'iconTextRadioButton': {'callbackFlags': ['visibleChangeCommand',
                                            'vcc',
                                            'dragCallback',
                                            'dgc',
                                            'offCommand',
                                            'ofc',
                                            'changeCommand',
                                            'cc',
                                            'onCommand',
                                            'onc',
                                            'dropCallback',
                                            'dpc'],
                          'flags': {'align': {'args': '<type __builtin__.str>',
                                              'longname': 'align',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'al'},
                                    'annotation': {'args': '<type __builtin__.str>',
                                                   'longname': 'annotation',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ann'},
                                    'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                 '<type __builtin__.float>',
                                                                 '<type __builtin__.float>'],
                                                        'longname': 'backgroundColor',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 3,
                                                        'shortname': 'bgc'},
                                    'changeCommand': {'args': 'script',
                                                      'longname': 'changeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cc'},
                                    'collection': {'args': '<type __builtin__.str>',
                                                   'longname': 'collection',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'cl'},
                                    'defineTemplate': {'args': '<type __builtin__.str>',
                                                       'longname': 'defineTemplate',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'dt'},
                                    'disabledImage': {'args': '<type __builtin__.str>',
                                                      'longname': 'disabledImage',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'di'},
                                    'docTag': {'args': '<type __builtin__.str>',
                                               'longname': 'docTag',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dtg'},
                                    'dragCallback': {'args': 'script',
                                                     'longname': 'dragCallback',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dgc'},
                                    'dropCallback': {'args': 'script',
                                                     'longname': 'dropCallback',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dpc'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'enable': {'args': '<type __builtin__.bool>',
                                               'longname': 'enable',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'en'},
                                    'enableBackground': {'args': '<type __builtin__.bool>',
                                                         'longname': 'enableBackground',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ebg'},
                                    'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                            'longname': 'enableKeyboardFocus',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ekf'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'flat': {'args': '<type __builtin__.bool>',
                                             'longname': 'flat',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'fla'},
                                    'flipX': {'args': '<type __builtin__.bool>',
                                              'longname': 'flipX',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fx'},
                                    'flipY': {'args': '<type __builtin__.bool>',
                                              'longname': 'flipY',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fy'},
                                    'font': {'args': '<type __builtin__.str>',
                                             'longname': 'font',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'fn'},
                                    'fullPathName': {'args': '<type __builtin__.bool>',
                                                     'longname': 'fullPathName',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'fpn'},
                                    'height': {'args': '<type __builtin__.int>',
                                               'longname': 'height',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'h'},
                                    'highlightColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'highlightColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'hlc'},
                                    'highlightImage': {'args': '<type __builtin__.str>',
                                                       'longname': 'highlightImage',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'hi'},
                                    'image': {'args': '<type __builtin__.str>',
                                              'longname': 'image',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'imageOverlayLabel': {'args': '<type __builtin__.str>',
                                                          'longname': 'imageOverlayLabel',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'iol'},
                                    'isObscured': {'args': '<type __builtin__.bool>',
                                                   'longname': 'isObscured',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'io'},
                                    'label': {'args': '<type __builtin__.str>',
                                              'longname': 'label',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'l'},
                                    'labelOffset': {'args': '<type __builtin__.int>',
                                                    'longname': 'labelOffset',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'lo'},
                                    'ltVersion': {'args': '<type __builtin__.str>',
                                                  'longname': 'ltVersion',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'lt'},
                                    'manage': {'args': '<type __builtin__.bool>',
                                               'longname': 'manage',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'm'},
                                    'marginHeight': {'args': '<type __builtin__.int>',
                                                     'longname': 'marginHeight',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'mh'},
                                    'marginWidth': {'args': '<type __builtin__.int>',
                                                    'longname': 'marginWidth',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'mw'},
                                    'noBackground': {'args': '<type __builtin__.bool>',
                                                     'longname': 'noBackground',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'nbg'},
                                    'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                           'longname': 'numberOfPopupMenus',
                                                           'modes': ['query'],
                                                           'numArgs': 0,
                                                           'shortname': 'npm'},
                                    'offCommand': {'args': 'script',
                                                   'longname': 'offCommand',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ofc'},
                                    'onCommand': {'args': 'script',
                                                  'longname': 'onCommand',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'onc'},
                                    'overlayLabelBackColor': {'args': ['<type __builtin__.float>',
                                                                       '<type __builtin__.float>',
                                                                       '<type __builtin__.float>',
                                                                       '<type __builtin__.float>'],
                                                              'longname': 'overlayLabelBackColor',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 4,
                                                              'shortname': 'olb'},
                                    'overlayLabelColor': {'args': ['<type __builtin__.float>',
                                                                   '<type __builtin__.float>',
                                                                   '<type __builtin__.float>'],
                                                          'longname': 'overlayLabelColor',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 3,
                                                          'shortname': 'olc'},
                                    'parent': {'args': '<type __builtin__.str>',
                                               'longname': 'parent',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'p'},
                                    'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                       'longname': 'popupMenuArray',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'pma'},
                                    'preventOverride': {'args': '<type __builtin__.bool>',
                                                        'longname': 'preventOverride',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'po'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'rotation': {'args': '<type __builtin__.float>',
                                                 'longname': 'rotation',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rot'},
                                    'select': {'args': '<type __builtin__.bool>',
                                               'longname': 'select',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 0,
                                               'shortname': 'sl'},
                                    'selectionHighlightImage': {'args': '<type __builtin__.str>',
                                                                'longname': 'selectionHighlightImage',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'shi'},
                                    'selectionImage': {'args': '<type __builtin__.str>',
                                                       'longname': 'selectionImage',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'si'},
                                    'statusBarMessage': {'args': '<type __builtin__.str>',
                                                         'longname': 'statusBarMessage',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'sbm'},
                                    'style': {'args': '<type __builtin__.str>',
                                              'longname': 'style',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'st'},
                                    'useAlpha': {'args': '<type __builtin__.bool>',
                                                 'longname': 'useAlpha',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ua'},
                                    'useTemplate': {'args': '<type __builtin__.str>',
                                                    'longname': 'useTemplate',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ut'},
                                    'version': {'args': '<type __builtin__.str>',
                                                'longname': 'version',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ver'},
                                    'visible': {'args': '<type __builtin__.bool>',
                                                'longname': 'visible',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'vis'},
                                    'visibleChangeCommand': {'args': 'script',
                                                             'longname': 'visibleChangeCommand',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'vcc'},
                                    'width': {'args': '<type __builtin__.int>',
                                              'longname': 'width',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'w'}},
                          'shortFlags': {'al': 'align',
                                         'ann': 'annotation',
                                         'bgc': 'backgroundColor',
                                         'cc': 'changeCommand',
                                         'cl': 'collection',
                                         'dgc': 'dragCallback',
                                         'di': 'disabledImage',
                                         'dpc': 'dropCallback',
                                         'dt': 'defineTemplate',
                                         'dtg': 'docTag',
                                         'e': 'edit',
                                         'ebg': 'enableBackground',
                                         'ekf': 'enableKeyboardFocus',
                                         'en': 'enable',
                                         'ex': 'exists',
                                         'fla': 'flat',
                                         'fn': 'font',
                                         'fpn': 'fullPathName',
                                         'fx': 'flipX',
                                         'fy': 'flipY',
                                         'h': 'height',
                                         'hi': 'highlightImage',
                                         'hlc': 'highlightColor',
                                         'i': 'image',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'io': 'isObscured',
                                         'iol': 'imageOverlayLabel',
                                         'l': 'label',
                                         'lo': 'labelOffset',
                                         'lt': 'ltVersion',
                                         'm': 'manage',
                                         'mh': 'marginHeight',
                                         'mw': 'marginWidth',
                                         'nbg': 'noBackground',
                                         'npm': 'numberOfPopupMenus',
                                         'ofc': 'offCommand',
                                         'olb': 'overlayLabelBackColor',
                                         'olc': 'overlayLabelColor',
                                         'onc': 'onCommand',
                                         'p': 'parent',
                                         'pma': 'popupMenuArray',
                                         'po': 'preventOverride',
                                         'q': 'query',
                                         'rot': 'rotation',
                                         'sbm': 'statusBarMessage',
                                         'shi': 'selectionHighlightImage',
                                         'si': 'selectionImage',
                                         'sl': 'select',
                                         'st': 'style',
                                         'ua': 'useAlpha',
                                         'ut': 'useTemplate',
                                         'vcc': 'visibleChangeCommand',
                                         'ver': 'version',
                                         'vis': 'visible',
                                         'w': 'width'},
                          'type': 'windows'},
  'iconTextRadioCollection': {'callbackFlags': ['disableCommands', 'dcm'],
                              'flags': {'collectionItemArray': {'args': '<type __builtin__.bool>',
                                                                'longname': 'collectionItemArray',
                                                                'modes': ['query'],
                                                                'numArgs': 0,
                                                                'shortname': 'cia'},
                                        'defineTemplate': {'args': '<type __builtin__.str>',
                                                           'longname': 'defineTemplate',
                                                           'modes': ['create'],
                                                           'numArgs': 1,
                                                           'shortname': 'dt'},
                                        'disableCommands': {'args': '<type __builtin__.bool>',
                                                            'longname': 'disableCommands',
                                                            'modes': ['edit'],
                                                            'numArgs': 0,
                                                            'shortname': 'dcm'},
                                        'edit': {'args': '<type __builtin__.bool>',
                                                 'longname': 'edit',
                                                 'numArgs': 0,
                                                 'shortname': 'e'},
                                        'exists': {'args': '<type __builtin__.bool>',
                                                   'longname': 'exists',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'ex'},
                                        'gl': {'args': '<type __builtin__.bool>',
                                               'longname': 'gl',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'gl'},
                                        'numberOfCollectionItems': {'args': '<type __builtin__.bool>',
                                                                    'longname': 'numberOfCollectionItems',
                                                                    'modes': ['query'],
                                                                    'numArgs': 0,
                                                                    'shortname': 'nci'},
                                        'parent': {'args': '<type __builtin__.str>',
                                                   'longname': 'parent',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'p'},
                                        'query': {'args': '<type __builtin__.bool>',
                                                  'longname': 'query',
                                                  'numArgs': 0,
                                                  'shortname': 'q'},
                                        'select': {'args': '<type __builtin__.str>',
                                                   'longname': 'select',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sl'},
                                        'useTemplate': {'args': '<type __builtin__.str>',
                                                        'longname': 'useTemplate',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'ut'}},
                              'removedFlags': {'global': 'gl'},
                              'shortFlags': {'cia': 'collectionItemArray',
                                             'dcm': 'disableCommands',
                                             'dt': 'defineTemplate',
                                             'e': 'edit',
                                             'ex': 'exists',
                                             'gl': 'gl',
                                             'nci': 'numberOfCollectionItems',
                                             'p': 'parent',
                                             'q': 'query',
                                             'sl': 'select',
                                             'ut': 'useTemplate'},
                              'type': 'windows'},
  'iconTextScrollList': {'callbackFlags': ['selectCommand',
                                           'sc',
                                           'dropRectCallback',
                                           'drc',
                                           'doubleClickCommand',
                                           'dcc',
                                           'dragCallback',
                                           'dgc',
                                           'visibleChangeCommand',
                                           'vcc',
                                           'dropCallback',
                                           'dpc',
                                           'changeCommand',
                                           'cc'],
                         'flags': {'allowMultiSelection': {'args': '<type __builtin__.bool>',
                                                           'longname': 'allowMultiSelection',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ams'},
                                   'annotation': {'args': '<type __builtin__.str>',
                                                  'longname': 'annotation',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ann'},
                                   'append': {'args': '<type __builtin__.str>',
                                              'longname': 'append',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'a'},
                                   'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'backgroundColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'bgc'},
                                   'changeCommand': {'args': 'script',
                                                     'longname': 'changeCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'cc'},
                                   'defineTemplate': {'args': '<type __builtin__.str>',
                                                      'longname': 'defineTemplate',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'dt'},
                                   'deselectAll': {'args': '<type __builtin__.bool>',
                                                   'longname': 'deselectAll',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'da'},
                                   'docTag': {'args': '<type __builtin__.str>',
                                              'longname': 'docTag',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dtg'},
                                   'doubleClickCommand': {'args': 'script',
                                                          'longname': 'doubleClickCommand',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'dcc'},
                                   'dragCallback': {'args': 'script',
                                                    'longname': 'dragCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dgc'},
                                   'dragFeedbackVisible': {'args': '<type __builtin__.bool>',
                                                           'longname': 'dragFeedbackVisible',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'dfv'},
                                   'dropCallback': {'args': 'script',
                                                    'longname': 'dropCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dpc'},
                                   'dropRectCallback': {'args': 'script',
                                                        'longname': 'dropRectCallback',
                                                        'modes': ['edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'drc'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'editIndexed': {'args': '<type __builtin__.int>',
                                                   'longname': 'editIndexed',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ei'},
                                   'editable': {'args': '<type __builtin__.bool>',
                                                'longname': 'editable',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ed'},
                                   'enable': {'args': '<type __builtin__.bool>',
                                              'longname': 'enable',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'en'},
                                   'enableBackground': {'args': '<type __builtin__.bool>',
                                                        'longname': 'enableBackground',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ebg'},
                                   'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                           'longname': 'enableKeyboardFocus',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ekf'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'fullPathName': {'args': '<type __builtin__.bool>',
                                                    'longname': 'fullPathName',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'fpn'},
                                   'height': {'args': '<type __builtin__.int>',
                                              'longname': 'height',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'h'},
                                   'highlightColor': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'highlightColor',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'hlc'},
                                   'isObscured': {'args': '<type __builtin__.bool>',
                                                  'longname': 'isObscured',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'io'},
                                   'itemAt': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'itemAt',
                                              'modes': ['query'],
                                              'numArgs': 2,
                                              'shortname': 'ia'},
                                   'itemTextColor': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'itemTextColor',
                                                     'modes': ['create',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 4,
                                                     'shortname': 'itc'},
                                   'manage': {'args': '<type __builtin__.bool>',
                                              'longname': 'manage',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'm'},
                                   'noBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'noBackground',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'nbg'},
                                   'numberOfIcons': {'args': '<type __builtin__.int>',
                                                     'longname': 'numberOfIcons',
                                                     'modes': ['query'],
                                                     'numArgs': 1,
                                                     'shortname': 'nic'},
                                   'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                          'longname': 'numberOfPopupMenus',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'npm'},
                                   'numberOfRows': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfRows',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'nr'},
                                   'parent': {'args': '<type __builtin__.str>',
                                              'longname': 'parent',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'p'},
                                   'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                      'longname': 'popupMenuArray',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'pma'},
                                   'preventOverride': {'args': '<type __builtin__.bool>',
                                                       'longname': 'preventOverride',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'po'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'removeAll': {'args': '<type __builtin__.bool>',
                                                 'longname': 'removeAll',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'ra'},
                                   'selectCommand': {'args': 'script',
                                                     'longname': 'selectCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sc'},
                                   'selectIndexedItem': {'args': '<type __builtin__.int>',
                                                         'longname': 'selectIndexedItem',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit',
                                                                   'multiuse'],
                                                         'numArgs': 1,
                                                         'shortname': 'sii'},
                                   'selectItem': {'args': '<type __builtin__.str>',
                                                  'longname': 'selectItem',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'si'},
                                   'statusBarMessage': {'args': '<type __builtin__.str>',
                                                        'longname': 'statusBarMessage',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'sbm'},
                                   'useTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'useTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ut'},
                                   'visible': {'args': '<type __builtin__.bool>',
                                               'longname': 'visible',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'vis'},
                                   'visibleChangeCommand': {'args': 'script',
                                                            'longname': 'visibleChangeCommand',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'vcc'},
                                   'visualRectAt': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'visualRectAt',
                                                    'modes': ['query'],
                                                    'numArgs': 2,
                                                    'shortname': 'vra'},
                                   'width': {'args': '<type __builtin__.int>',
                                             'longname': 'width',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'w'}},
                         'shortFlags': {'a': 'append',
                                        'ams': 'allowMultiSelection',
                                        'ann': 'annotation',
                                        'bgc': 'backgroundColor',
                                        'cc': 'changeCommand',
                                        'da': 'deselectAll',
                                        'dcc': 'doubleClickCommand',
                                        'dfv': 'dragFeedbackVisible',
                                        'dgc': 'dragCallback',
                                        'dpc': 'dropCallback',
                                        'drc': 'dropRectCallback',
                                        'dt': 'defineTemplate',
                                        'dtg': 'docTag',
                                        'e': 'edit',
                                        'ebg': 'enableBackground',
                                        'ed': 'editable',
                                        'ei': 'editIndexed',
                                        'ekf': 'enableKeyboardFocus',
                                        'en': 'enable',
                                        'ex': 'exists',
                                        'fpn': 'fullPathName',
                                        'h': 'height',
                                        'hlc': 'highlightColor',
                                        'ia': 'itemAt',
                                        'io': 'isObscured',
                                        'itc': 'itemTextColor',
                                        'm': 'manage',
                                        'nbg': 'noBackground',
                                        'nic': 'numberOfIcons',
                                        'npm': 'numberOfPopupMenus',
                                        'nr': 'numberOfRows',
                                        'p': 'parent',
                                        'pma': 'popupMenuArray',
                                        'po': 'preventOverride',
                                        'q': 'query',
                                        'ra': 'removeAll',
                                        'sbm': 'statusBarMessage',
                                        'sc': 'selectCommand',
                                        'si': 'selectItem',
                                        'sii': 'selectIndexedItem',
                                        'ut': 'useTemplate',
                                        'vcc': 'visibleChangeCommand',
                                        'vis': 'visible',
                                        'vra': 'visualRectAt',
                                        'w': 'width'},
                         'type': 'windows'},
  'iconTextStaticLabel': {'callbackFlags': ['visibleChangeCommand',
                                            'vcc',
                                            'dragCallback',
                                            'dgc',
                                            'dropCallback',
                                            'dpc'],
                          'flags': {'align': {'args': '<type __builtin__.str>',
                                              'longname': 'align',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'al'},
                                    'annotation': {'args': '<type __builtin__.str>',
                                                   'longname': 'annotation',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ann'},
                                    'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                 '<type __builtin__.float>',
                                                                 '<type __builtin__.float>'],
                                                        'longname': 'backgroundColor',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 3,
                                                        'shortname': 'bgc'},
                                    'defineTemplate': {'args': '<type __builtin__.str>',
                                                       'longname': 'defineTemplate',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'dt'},
                                    'disabledImage': {'args': '<type __builtin__.str>',
                                                      'longname': 'disabledImage',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'di'},
                                    'docTag': {'args': '<type __builtin__.str>',
                                               'longname': 'docTag',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dtg'},
                                    'dragCallback': {'args': 'script',
                                                     'longname': 'dragCallback',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dgc'},
                                    'dropCallback': {'args': 'script',
                                                     'longname': 'dropCallback',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dpc'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'enable': {'args': '<type __builtin__.bool>',
                                               'longname': 'enable',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'en'},
                                    'enableBackground': {'args': '<type __builtin__.bool>',
                                                         'longname': 'enableBackground',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ebg'},
                                    'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                            'longname': 'enableKeyboardFocus',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ekf'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'flipX': {'args': '<type __builtin__.bool>',
                                              'longname': 'flipX',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fx'},
                                    'flipY': {'args': '<type __builtin__.bool>',
                                              'longname': 'flipY',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fy'},
                                    'font': {'args': '<type __builtin__.str>',
                                             'longname': 'font',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'fn'},
                                    'fullPathName': {'args': '<type __builtin__.bool>',
                                                     'longname': 'fullPathName',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'fpn'},
                                    'height': {'args': '<type __builtin__.int>',
                                               'longname': 'height',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'h'},
                                    'highlightColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'highlightColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'hlc'},
                                    'image': {'args': '<type __builtin__.str>',
                                              'longname': 'image',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'imageOverlayLabel': {'args': '<type __builtin__.str>',
                                                          'longname': 'imageOverlayLabel',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'iol'},
                                    'isObscured': {'args': '<type __builtin__.bool>',
                                                   'longname': 'isObscured',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'io'},
                                    'label': {'args': '<type __builtin__.str>',
                                              'longname': 'label',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'l'},
                                    'labelOffset': {'args': '<type __builtin__.int>',
                                                    'longname': 'labelOffset',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'lo'},
                                    'ltVersion': {'args': '<type __builtin__.str>',
                                                  'longname': 'ltVersion',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'lt'},
                                    'manage': {'args': '<type __builtin__.bool>',
                                               'longname': 'manage',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'm'},
                                    'marginHeight': {'args': '<type __builtin__.int>',
                                                     'longname': 'marginHeight',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'mh'},
                                    'marginWidth': {'args': '<type __builtin__.int>',
                                                    'longname': 'marginWidth',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'mw'},
                                    'noBackground': {'args': '<type __builtin__.bool>',
                                                     'longname': 'noBackground',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'nbg'},
                                    'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                           'longname': 'numberOfPopupMenus',
                                                           'modes': ['query'],
                                                           'numArgs': 0,
                                                           'shortname': 'npm'},
                                    'overlayLabelBackColor': {'args': ['<type __builtin__.float>',
                                                                       '<type __builtin__.float>',
                                                                       '<type __builtin__.float>',
                                                                       '<type __builtin__.float>'],
                                                              'longname': 'overlayLabelBackColor',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 4,
                                                              'shortname': 'olb'},
                                    'overlayLabelColor': {'args': ['<type __builtin__.float>',
                                                                   '<type __builtin__.float>',
                                                                   '<type __builtin__.float>'],
                                                          'longname': 'overlayLabelColor',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 3,
                                                          'shortname': 'olc'},
                                    'parent': {'args': '<type __builtin__.str>',
                                               'longname': 'parent',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'p'},
                                    'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                       'longname': 'popupMenuArray',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'pma'},
                                    'preventOverride': {'args': '<type __builtin__.bool>',
                                                        'longname': 'preventOverride',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'po'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'rotation': {'args': '<type __builtin__.float>',
                                                 'longname': 'rotation',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rot'},
                                    'statusBarMessage': {'args': '<type __builtin__.str>',
                                                         'longname': 'statusBarMessage',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'sbm'},
                                    'style': {'args': '<type __builtin__.str>',
                                              'longname': 'style',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'st'},
                                    'useAlpha': {'args': '<type __builtin__.bool>',
                                                 'longname': 'useAlpha',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ua'},
                                    'useTemplate': {'args': '<type __builtin__.str>',
                                                    'longname': 'useTemplate',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ut'},
                                    'version': {'args': '<type __builtin__.str>',
                                                'longname': 'version',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ver'},
                                    'visible': {'args': '<type __builtin__.bool>',
                                                'longname': 'visible',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'vis'},
                                    'visibleChangeCommand': {'args': 'script',
                                                             'longname': 'visibleChangeCommand',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'vcc'},
                                    'width': {'args': '<type __builtin__.int>',
                                              'longname': 'width',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'w'}},
                          'shortFlags': {'al': 'align',
                                         'ann': 'annotation',
                                         'bgc': 'backgroundColor',
                                         'dgc': 'dragCallback',
                                         'di': 'disabledImage',
                                         'dpc': 'dropCallback',
                                         'dt': 'defineTemplate',
                                         'dtg': 'docTag',
                                         'e': 'edit',
                                         'ebg': 'enableBackground',
                                         'ekf': 'enableKeyboardFocus',
                                         'en': 'enable',
                                         'ex': 'exists',
                                         'fn': 'font',
                                         'fpn': 'fullPathName',
                                         'fx': 'flipX',
                                         'fy': 'flipY',
                                         'h': 'height',
                                         'hlc': 'highlightColor',
                                         'i': 'image',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'io': 'isObscured',
                                         'iol': 'imageOverlayLabel',
                                         'l': 'label',
                                         'lo': 'labelOffset',
                                         'lt': 'ltVersion',
                                         'm': 'manage',
                                         'mh': 'marginHeight',
                                         'mw': 'marginWidth',
                                         'nbg': 'noBackground',
                                         'npm': 'numberOfPopupMenus',
                                         'olb': 'overlayLabelBackColor',
                                         'olc': 'overlayLabelColor',
                                         'p': 'parent',
                                         'pma': 'popupMenuArray',
                                         'po': 'preventOverride',
                                         'q': 'query',
                                         'rot': 'rotation',
                                         'sbm': 'statusBarMessage',
                                         'st': 'style',
                                         'ua': 'useAlpha',
                                         'ut': 'useTemplate',
                                         'vcc': 'visibleChangeCommand',
                                         'ver': 'version',
                                         'vis': 'visible',
                                         'w': 'width'},
                          'type': 'windows'},
  'ikHandle': {'flags': {'autoPriority': {'args': '<type __builtin__.bool>',
                                          'longname': 'autoPriority',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'ap'},
                         'connectEffector': {'args': '<type __builtin__.bool>',
                                             'longname': 'connectEffector',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ce'},
                         'createCurve': {'args': '<type __builtin__.bool>',
                                         'longname': 'createCurve',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ccv'},
                         'createRootAxis': {'args': '<type __builtin__.bool>',
                                            'longname': 'createRootAxis',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'cra'},
                         'curve': {'args': 'PyNode',
                                   'longname': 'curve',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'c'},
                         'disableHandles': {'args': '<type __builtin__.bool>',
                                            'longname': 'disableHandles',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'dh'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'enableHandles': {'args': '<type __builtin__.bool>',
                                           'longname': 'enableHandles',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'eh'},
                         'endEffector': {'args': '<type __builtin__.str>',
                                         'longname': 'endEffector',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ee'},
                         'exists': {'args': '<type __builtin__.str>',
                                    'longname': 'exists',
                                    'modes': ['edit'],
                                    'numArgs': 1,
                                    'shortname': 'ex'},
                         'forceSolver': {'args': '<type __builtin__.bool>',
                                         'longname': 'forceSolver',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'fs'},
                         'freezeJoints': {'args': '<type __builtin__.bool>',
                                          'longname': 'freezeJoints',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'fj'},
                         'jointList': {'args': '<type __builtin__.bool>',
                                       'longname': 'jointList',
                                       'modes': ['edit'],
                                       'numArgs': 0,
                                       'shortname': 'jl'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'numSpans': {'args': '<type __builtin__.int>',
                                      'longname': 'numSpans',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'ns'},
                         'parentCurve': {'args': '<type __builtin__.bool>',
                                         'longname': 'parentCurve',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'pcv'},
                         'positionWeight': {'args': '<type __builtin__.float>',
                                            'longname': 'positionWeight',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pw'},
                         'priority': {'args': '<type __builtin__.int>',
                                      'longname': 'priority',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'rootOnCurve': {'args': '<type __builtin__.bool>',
                                         'longname': 'rootOnCurve',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'roc'},
                         'rootTwistMode': {'args': '<type __builtin__.bool>',
                                           'longname': 'rootTwistMode',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'rtm'},
                         'setupForRPsolver': {'args': '<type __builtin__.bool>',
                                              'longname': 'setupForRPsolver',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'srp'},
                         'simplifyCurve': {'args': '<type __builtin__.bool>',
                                           'longname': 'simplifyCurve',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'scv'},
                         'snapCurve': {'args': '<type __builtin__.bool>',
                                       'longname': 'snapCurve',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'snc'},
                         'snapHandleFlagToggle': {'args': '<type __builtin__.bool>',
                                                  'longname': 'snapHandleFlagToggle',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'shf'},
                         'snapHandleToEffector': {'args': '<type __builtin__.bool>',
                                                  'longname': 'snapHandleToEffector',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'see'},
                         'solver': {'args': '<type __builtin__.str>',
                                    'longname': 'solver',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'sol'},
                         'startJoint': {'args': '<type __builtin__.str>',
                                        'longname': 'startJoint',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'sj'},
                         'sticky': {'args': '<type __builtin__.str>',
                                    'longname': 'sticky',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 's'},
                         'twistType': {'args': '<type __builtin__.str>',
                                       'longname': 'twistType',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'tws'},
                         'weight': {'args': '<type __builtin__.float>',
                                    'longname': 'weight',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'w'}},
               'shortFlags': {'ap': 'autoPriority',
                              'c': 'curve',
                              'ccv': 'createCurve',
                              'ce': 'connectEffector',
                              'cra': 'createRootAxis',
                              'dh': 'disableHandles',
                              'e': 'edit',
                              'ee': 'endEffector',
                              'eh': 'enableHandles',
                              'ex': 'exists',
                              'fj': 'freezeJoints',
                              'fs': 'forceSolver',
                              'jl': 'jointList',
                              'n': 'name',
                              'ns': 'numSpans',
                              'p': 'priority',
                              'pcv': 'parentCurve',
                              'pw': 'positionWeight',
                              'q': 'query',
                              'roc': 'rootOnCurve',
                              'rtm': 'rootTwistMode',
                              's': 'sticky',
                              'scv': 'simplifyCurve',
                              'see': 'snapHandleToEffector',
                              'shf': 'snapHandleFlagToggle',
                              'sj': 'startJoint',
                              'snc': 'snapCurve',
                              'sol': 'solver',
                              'srp': 'setupForRPsolver',
                              'tws': 'twistType',
                              'w': 'weight'},
               'type': 'animation'},
  'ikHandleCtx': {'flags': {'autoPriorityH': {'args': '<type __builtin__.bool>',
                                              'longname': 'autoPriorityH',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'apH'},
                            'createCurve': {'args': '<type __builtin__.bool>',
                                            'longname': 'createCurve',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ccv'},
                            'createRootAxis': {'args': '<type __builtin__.bool>',
                                               'longname': 'createRootAxis',
                                               'modes': ['edit'],
                                               'numArgs': 1,
                                               'shortname': 'cra'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'forceSolverH': {'args': '<type __builtin__.bool>',
                                             'longname': 'forceSolverH',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'fsH'},
                            'history': {'args': '<type __builtin__.bool>',
                                        'longname': 'history',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ch'},
                            'image1': {'args': '<type __builtin__.str>',
                                       'longname': 'image1',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i1'},
                            'image2': {'args': '<type __builtin__.str>',
                                       'longname': 'image2',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i2'},
                            'image3': {'args': '<type __builtin__.str>',
                                       'longname': 'image3',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i3'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'numSpans': {'args': '<type __builtin__.int>',
                                         'longname': 'numSpans',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 'ns'},
                            'parentCurve': {'args': '<type __builtin__.bool>',
                                            'longname': 'parentCurve',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'pcv'},
                            'poWeightH': {'args': '<type __builtin__.float>',
                                          'longname': 'poWeightH',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pwH'},
                            'priorityH': {'args': '<type __builtin__.int>',
                                          'longname': 'priorityH',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pH'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'rootOnCurve': {'args': '<type __builtin__.bool>',
                                            'longname': 'rootOnCurve',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'roc'},
                            'rootTwistMode': {'args': '<type __builtin__.bool>',
                                              'longname': 'rootTwistMode',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'rtm'},
                            'simplifyCurve': {'args': '<type __builtin__.bool>',
                                              'longname': 'simplifyCurve',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'scv'},
                            'snapCurve': {'args': '<type __builtin__.bool>',
                                          'longname': 'snapCurve',
                                          'modes': ['edit'],
                                          'numArgs': 1,
                                          'shortname': 'snc'},
                            'snapHandleH': {'args': '<type __builtin__.bool>',
                                            'longname': 'snapHandleH',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'snH'},
                            'solverTypeH': {'args': '<type __builtin__.str>',
                                            'longname': 'solverTypeH',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'stH'},
                            'stickyH': {'args': '<type __builtin__.str>',
                                        'longname': 'stickyH',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'sH'},
                            'twistType': {'args': '<type __builtin__.str>',
                                          'longname': 'twistType',
                                          'modes': ['edit'],
                                          'numArgs': 1,
                                          'shortname': 'tws'},
                            'weightH': {'args': '<type __builtin__.float>',
                                        'longname': 'weightH',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'wH'}},
                  'shortFlags': {'apH': 'autoPriorityH',
                                 'ccv': 'createCurve',
                                 'ch': 'history',
                                 'cra': 'createRootAxis',
                                 'e': 'edit',
                                 'ex': 'exists',
                                 'fsH': 'forceSolverH',
                                 'i1': 'image1',
                                 'i2': 'image2',
                                 'i3': 'image3',
                                 'n': 'name',
                                 'ns': 'numSpans',
                                 'pH': 'priorityH',
                                 'pcv': 'parentCurve',
                                 'pwH': 'poWeightH',
                                 'q': 'query',
                                 'roc': 'rootOnCurve',
                                 'rtm': 'rootTwistMode',
                                 'sH': 'stickyH',
                                 'scv': 'simplifyCurve',
                                 'snH': 'snapHandleH',
                                 'snc': 'snapCurve',
                                 'stH': 'solverTypeH',
                                 'tws': 'twistType',
                                 'wH': 'weightH'},
                  'type': 'context'},
  'ikHandleDisplayScale': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'}},
                           'shortFlags': {'q': 'query'},
                           'type': 'animation'},
  'ikSolver': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'epsilon': {'args': '<type __builtin__.float>',
                                     'longname': 'epsilon',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'ep'},
                         'maxIterations': {'args': '<type __builtin__.int>',
                                           'longname': 'maxIterations',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'mxi'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'solverType': {'args': '<type __builtin__.str>',
                                        'longname': 'solverType',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'st'}},
               'shortFlags': {'e': 'edit',
                              'ep': 'epsilon',
                              'mxi': 'maxIterations',
                              'n': 'name',
                              'q': 'query',
                              'st': 'solverType'},
               'type': 'animation'},
  'ikSplineHandleCtx': {'flags': {'autoPriorityH': {'args': '<type __builtin__.bool>',
                                                    'longname': 'autoPriorityH',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'apH'},
                                  'createCurve': {'args': '<type __builtin__.bool>',
                                                  'longname': 'createCurve',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ccv'},
                                  'createRootAxis': {'args': '<type __builtin__.bool>',
                                                     'longname': 'createRootAxis',
                                                     'modes': ['edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'cra'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'forceSolverH': {'args': '<type __builtin__.bool>',
                                                   'longname': 'forceSolverH',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'fsH'},
                                  'history': {'args': '<type __builtin__.bool>',
                                              'longname': 'history',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ch'},
                                  'image1': {'args': '<type __builtin__.str>',
                                             'longname': 'image1',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i1'},
                                  'image2': {'args': '<type __builtin__.str>',
                                             'longname': 'image2',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i2'},
                                  'image3': {'args': '<type __builtin__.str>',
                                             'longname': 'image3',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i3'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'numSpans': {'args': '<type __builtin__.int>',
                                               'longname': 'numSpans',
                                               'modes': ['edit'],
                                               'numArgs': 1,
                                               'shortname': 'ns'},
                                  'parentCurve': {'args': '<type __builtin__.bool>',
                                                  'longname': 'parentCurve',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pcv'},
                                  'poWeightH': {'args': '<type __builtin__.float>',
                                                'longname': 'poWeightH',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'pwH'},
                                  'priorityH': {'args': '<type __builtin__.int>',
                                                'longname': 'priorityH',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'pH'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'rootOnCurve': {'args': '<type __builtin__.bool>',
                                                  'longname': 'rootOnCurve',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'roc'},
                                  'rootTwistMode': {'args': '<type __builtin__.bool>',
                                                    'longname': 'rootTwistMode',
                                                    'modes': ['edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'rtm'},
                                  'simplifyCurve': {'args': '<type __builtin__.bool>',
                                                    'longname': 'simplifyCurve',
                                                    'modes': ['edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'scv'},
                                  'snapCurve': {'args': '<type __builtin__.bool>',
                                                'longname': 'snapCurve',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'snc'},
                                  'snapHandleH': {'args': '<type __builtin__.bool>',
                                                  'longname': 'snapHandleH',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'snH'},
                                  'solverTypeH': {'args': '<type __builtin__.str>',
                                                  'longname': 'solverTypeH',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'stH'},
                                  'stickyH': {'args': '<type __builtin__.str>',
                                              'longname': 'stickyH',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sH'},
                                  'twistType': {'args': '<type __builtin__.str>',
                                                'longname': 'twistType',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'tws'},
                                  'weightH': {'args': '<type __builtin__.float>',
                                              'longname': 'weightH',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'wH'}},
                        'shortFlags': {'apH': 'autoPriorityH',
                                       'ccv': 'createCurve',
                                       'ch': 'history',
                                       'cra': 'createRootAxis',
                                       'e': 'edit',
                                       'ex': 'exists',
                                       'fsH': 'forceSolverH',
                                       'i1': 'image1',
                                       'i2': 'image2',
                                       'i3': 'image3',
                                       'n': 'name',
                                       'ns': 'numSpans',
                                       'pH': 'priorityH',
                                       'pcv': 'parentCurve',
                                       'pwH': 'poWeightH',
                                       'q': 'query',
                                       'roc': 'rootOnCurve',
                                       'rtm': 'rootTwistMode',
                                       'sH': 'stickyH',
                                       'scv': 'simplifyCurve',
                                       'snH': 'snapHandleH',
                                       'snc': 'snapCurve',
                                       'stH': 'solverTypeH',
                                       'tws': 'twistType',
                                       'wH': 'weightH'},
                        'type': 'context'},
  'ikSystem': {'flags': {'allowRotation': {'args': '<type __builtin__.bool>',
                                           'longname': 'allowRotation',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ar'},
                         'autoPriority': {'args': '<type __builtin__.bool>',
                                          'longname': 'autoPriority',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'ap'},
                         'autoPriorityMC': {'args': '<type __builtin__.bool>',
                                            'longname': 'autoPriorityMC',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'apm'},
                         'autoPrioritySC': {'args': '<type __builtin__.bool>',
                                            'longname': 'autoPrioritySC',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'aps'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'list': {'args': ['<type __builtin__.int>',
                                           '<type __builtin__.int>'],
                                  'longname': 'list',
                                  'modes': ['query', 'edit'],
                                  'numArgs': 2,
                                  'shortname': 'ls'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'snap': {'args': '<type __builtin__.bool>',
                                  'longname': 'snap',
                                  'modes': ['query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'sn'},
                         'solve': {'args': '<type __builtin__.bool>',
                                   'longname': 'solve',
                                   'modes': ['query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'sol'},
                         'solverTypes': {'args': '<type __builtin__.bool>',
                                         'longname': 'solverTypes',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'st'}},
               'shortFlags': {'ap': 'autoPriority',
                              'apm': 'autoPriorityMC',
                              'aps': 'autoPrioritySC',
                              'ar': 'allowRotation',
                              'e': 'edit',
                              'ls': 'list',
                              'q': 'query',
                              'sn': 'snap',
                              'sol': 'solve',
                              'st': 'solverTypes'},
               'type': 'animation'},
  'ikSystemInfo': {'flags': {'globalSnapHandle': {'args': '<type __builtin__.bool>',
                                                  'longname': 'globalSnapHandle',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 0,
                                                  'shortname': 'gsh'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'gsh': 'globalSnapHandle', 'q': 'query'},
                   'type': 'animation'},
  'ikfkDisplayMethod': {'flags': {'display': {'args': '<type __builtin__.str>',
                                              'longname': 'display',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'd'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'}},
                        'shortFlags': {'d': 'display', 'q': 'query'},
                        'type': 'animation'},
  'illustratorCurves': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                              'longname': 'caching',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cch'},
                                  'constructionHistory': {'args': '<type __builtin__.bool>',
                                                          'longname': 'constructionHistory',
                                                          'modes': ['create'],
                                                          'numArgs': 1,
                                                          'shortname': 'ch'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'frozen': {'args': '<type __builtin__.bool>',
                                             'longname': 'frozen',
                                             'numArgs': 1,
                                             'shortname': 'fzn'},
                                  'illustratorFilename': {'args': '<type __builtin__.str>',
                                                          'longname': 'illustratorFilename',
                                                          'modes': ['create'],
                                                          'numArgs': 1,
                                                          'shortname': 'ifn'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'nodeState': {'args': '<type __builtin__.int>',
                                                'longname': 'nodeState',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nds'},
                                  'object': {'args': '<type __builtin__.bool>',
                                             'longname': 'object',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'o'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'scaleFactor': {'args': '<type __builtin__.float>',
                                                  'longname': 'scaleFactor',
                                                  'numArgs': 1,
                                                  'shortname': 'sf'},
                                  'tolerance': {'args': '<type __builtin__.float>',
                                                'longname': 'tolerance',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'tl'}},
                        'shortFlags': {'cch': 'caching',
                                       'ch': 'constructionHistory',
                                       'e': 'edit',
                                       'fzn': 'frozen',
                                       'ifn': 'illustratorFilename',
                                       'n': 'name',
                                       'nds': 'nodeState',
                                       'o': 'object',
                                       'q': 'query',
                                       'sf': 'scaleFactor',
                                       'tl': 'tolerance'},
                        'type': 'modeling'},
  'image': {'callbackFlags': ['visibleChangeCommand',
                              'vcc',
                              'dragCallback',
                              'dgc',
                              'dropCallback',
                              'dpc'],
            'flags': {'annotation': {'args': '<type __builtin__.str>',
                                     'longname': 'annotation',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'ann'},
                      'backgroundColor': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'backgroundColor',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 3,
                                          'shortname': 'bgc'},
                      'defineTemplate': {'args': '<type __builtin__.str>',
                                         'longname': 'defineTemplate',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'dt'},
                      'docTag': {'args': '<type __builtin__.str>',
                                 'longname': 'docTag',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'dtg'},
                      'dragCallback': {'args': 'script',
                                       'longname': 'dragCallback',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dgc'},
                      'dropCallback': {'args': 'script',
                                       'longname': 'dropCallback',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dpc'},
                      'edit': {'args': '<type __builtin__.bool>',
                               'longname': 'edit',
                               'numArgs': 0,
                               'shortname': 'e'},
                      'enable': {'args': '<type __builtin__.bool>',
                                 'longname': 'enable',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'en'},
                      'enableBackground': {'args': '<type __builtin__.bool>',
                                           'longname': 'enableBackground',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ebg'},
                      'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                              'longname': 'enableKeyboardFocus',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ekf'},
                      'exists': {'args': '<type __builtin__.bool>',
                                 'longname': 'exists',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'ex'},
                      'fullPathName': {'args': '<type __builtin__.bool>',
                                       'longname': 'fullPathName',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'fpn'},
                      'height': {'args': '<type __builtin__.int>',
                                 'longname': 'height',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'h'},
                      'highlightColor': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'highlightColor',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 3,
                                         'shortname': 'hlc'},
                      'image': {'args': '<type __builtin__.str>',
                                'longname': 'image',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'i'},
                      'isObscured': {'args': '<type __builtin__.bool>',
                                     'longname': 'isObscured',
                                     'modes': ['query'],
                                     'numArgs': 0,
                                     'shortname': 'io'},
                      'manage': {'args': '<type __builtin__.bool>',
                                 'longname': 'manage',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'm'},
                      'noBackground': {'args': '<type __builtin__.bool>',
                                       'longname': 'noBackground',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'nbg'},
                      'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                             'longname': 'numberOfPopupMenus',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'npm'},
                      'parent': {'args': '<type __builtin__.str>',
                                 'longname': 'parent',
                                 'modes': ['create', 'query'],
                                 'numArgs': 1,
                                 'shortname': 'p'},
                      'popupMenuArray': {'args': '<type __builtin__.bool>',
                                         'longname': 'popupMenuArray',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'pma'},
                      'preventOverride': {'args': '<type __builtin__.bool>',
                                          'longname': 'preventOverride',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'po'},
                      'query': {'args': '<type __builtin__.bool>',
                                'longname': 'query',
                                'numArgs': 0,
                                'shortname': 'q'},
                      'statusBarMessage': {'args': '<type __builtin__.str>',
                                           'longname': 'statusBarMessage',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sbm'},
                      'useTemplate': {'args': '<type __builtin__.str>',
                                      'longname': 'useTemplate',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'ut'},
                      'visible': {'args': '<type __builtin__.bool>',
                                  'longname': 'visible',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'vis'},
                      'visibleChangeCommand': {'args': 'script',
                                               'longname': 'visibleChangeCommand',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'vcc'},
                      'width': {'args': '<type __builtin__.int>',
                                'longname': 'width',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'w'}},
            'shortFlags': {'ann': 'annotation',
                           'bgc': 'backgroundColor',
                           'dgc': 'dragCallback',
                           'dpc': 'dropCallback',
                           'dt': 'defineTemplate',
                           'dtg': 'docTag',
                           'e': 'edit',
                           'ebg': 'enableBackground',
                           'ekf': 'enableKeyboardFocus',
                           'en': 'enable',
                           'ex': 'exists',
                           'fpn': 'fullPathName',
                           'h': 'height',
                           'hlc': 'highlightColor',
                           'i': 'image',
                           'io': 'isObscured',
                           'm': 'manage',
                           'nbg': 'noBackground',
                           'npm': 'numberOfPopupMenus',
                           'p': 'parent',
                           'pma': 'popupMenuArray',
                           'po': 'preventOverride',
                           'q': 'query',
                           'sbm': 'statusBarMessage',
                           'ut': 'useTemplate',
                           'vcc': 'visibleChangeCommand',
                           'vis': 'visible',
                           'w': 'width'},
            'type': 'windows'},
  'imagePlane': {'flags': {'camera': {'args': '<type __builtin__.str>',
                                      'longname': 'camera',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsUnpacking': True,
                                      'shortname': 'c'},
                           'counter': {'args': '<type __builtin__.bool>',
                                       'longname': 'counter',
                                       'modes': [],
                                       'numArgs': 1,
                                       'shortname': 'cn'},
                           'detach': {'args': '<type __builtin__.bool>',
                                      'longname': 'detach',
                                      'modes': ['edit'],
                                      'numArgs': 0,
                                      'shortname': 'd'},
                           'dropFrame': {'args': '<type __builtin__.bool>',
                                         'longname': 'dropFrame',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'df'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'fileName': {'args': '<type __builtin__.str>',
                                        'longname': 'fileName',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'fn'},
                           'frameDuration': {'args': '<type __builtin__.int>',
                                             'longname': 'frameDuration',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'fd'},
                           'height': {'args': '<type __builtin__.float>',
                                      'longname': 'height',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsUnpacking': True,
                                      'shortname': 'h'},
                           'imageSize': {'args': ['<type __builtin__.int>',
                                                  '<type __builtin__.int>'],
                                         'longname': 'imageSize',
                                         'modes': ['query'],
                                         'numArgs': 2,
                                         'shortname': 'iz'},
                           'lookThrough': {'args': '<type __builtin__.str>',
                                           'longname': 'lookThrough',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'resultNeedsUnpacking': True,
                                           'shortname': 'lt'},
                           'maintainRatio': {'args': '<type __builtin__.bool>',
                                             'longname': 'maintainRatio',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'resultNeedsUnpacking': True,
                                             'shortname': 'mr'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'negTimesOK': {'args': '<type __builtin__.bool>',
                                          'longname': 'negTimesOK',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'nt'},
                           'numFrames': {'args': '<type __builtin__.int>',
                                         'longname': 'numFrames',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'nf'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'quickTime': {'args': '<type __builtin__.bool>',
                                         'longname': 'quickTime',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'qt'},
                           'showInAllViews': {'args': '<type __builtin__.bool>',
                                              'longname': 'showInAllViews',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'resultNeedsUnpacking': True,
                                              'shortname': 'sia'},
                           'timeCode': {'args': '<type __builtin__.int>',
                                        'longname': 'timeCode',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'tc'},
                           'timeCodeTrack': {'args': '<type __builtin__.bool>',
                                             'longname': 'timeCodeTrack',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'tt'},
                           'timeScale': {'args': '<type __builtin__.int>',
                                         'longname': 'timeScale',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'ts'},
                           'twentyFourHourMax': {'args': '<type __builtin__.bool>',
                                                 'longname': 'twentyFourHourMax',
                                                 'modes': ['query'],
                                                 'numArgs': 1,
                                                 'shortname': 'tf'},
                           'width': {'args': '<type __builtin__.float>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsUnpacking': True,
                                     'shortname': 'w'}},
                 'shortFlags': {'c': 'camera',
                                'cn': 'counter',
                                'd': 'detach',
                                'df': 'dropFrame',
                                'e': 'edit',
                                'fd': 'frameDuration',
                                'fn': 'fileName',
                                'h': 'height',
                                'iz': 'imageSize',
                                'lt': 'lookThrough',
                                'mr': 'maintainRatio',
                                'n': 'name',
                                'nf': 'numFrames',
                                'nt': 'negTimesOK',
                                'q': 'query',
                                'qt': 'quickTime',
                                'sia': 'showInAllViews',
                                'tc': 'timeCode',
                                'tf': 'twentyFourHourMax',
                                'ts': 'timeScale',
                                'tt': 'timeCodeTrack',
                                'w': 'width'},
                 'type': 'rendering'},
  'imageWindowEditor': {'callbackFlags': ['changeCommand', 'cc'],
                        'flags': {'autoResize': {'args': '<type __builtin__.bool>',
                                                 'longname': 'autoResize',
                                                 'numArgs': 1,
                                                 'shortname': 'ar'},
                                  'changeCommand': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'changeCommand',
                                                    'numArgs': 4,
                                                    'shortname': 'cc'},
                                  'clear': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'clear',
                                            'numArgs': 5,
                                            'shortname': 'cl'},
                                  'control': {'args': '<type __builtin__.bool>',
                                              'longname': 'control',
                                              'numArgs': 0,
                                              'shortname': 'ctl'},
                                  'defineTemplate': {'args': '<type __builtin__.str>',
                                                     'longname': 'defineTemplate',
                                                     'numArgs': 1,
                                                     'shortname': 'dt'},
                                  'displayImage': {'args': '<type __builtin__.int>',
                                                   'longname': 'displayImage',
                                                   'numArgs': 1,
                                                   'shortname': 'di'},
                                  'displayStyle': {'args': '<type __builtin__.str>',
                                                   'longname': 'displayStyle',
                                                   'numArgs': 1,
                                                   'shortname': 'dst'},
                                  'docTag': {'args': '<type __builtin__.str>',
                                             'longname': 'docTag',
                                             'numArgs': 1,
                                             'shortname': 'dtg'},
                                  'doubleBuffer': {'args': '<type __builtin__.bool>',
                                                   'longname': 'doubleBuffer',
                                                   'numArgs': 0,
                                                   'shortname': 'dbf'},
                                  'drawAxis': {'args': '<type __builtin__.bool>',
                                               'longname': 'drawAxis',
                                               'numArgs': 1,
                                               'shortname': 'da'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'filter': {'args': '<type __builtin__.str>',
                                             'longname': 'filter',
                                             'numArgs': 1,
                                             'shortname': 'f'},
                                  'forceMainConnection': {'args': '<type __builtin__.str>',
                                                          'longname': 'forceMainConnection',
                                                          'numArgs': 1,
                                                          'shortname': 'fmc'},
                                  'frameImage': {'args': '<type __builtin__.bool>',
                                                 'longname': 'frameImage',
                                                 'numArgs': 0,
                                                 'shortname': 'fi'},
                                  'frameRegion': {'args': '<type __builtin__.bool>',
                                                  'longname': 'frameRegion',
                                                  'numArgs': 0,
                                                  'shortname': 'fr'},
                                  'highlightConnection': {'args': '<type __builtin__.str>',
                                                          'longname': 'highlightConnection',
                                                          'numArgs': 1,
                                                          'shortname': 'hlc'},
                                  'loadImage': {'args': '<type __builtin__.str>',
                                                'longname': 'loadImage',
                                                'numArgs': 1,
                                                'shortname': 'li'},
                                  'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'lockMainConnection',
                                                         'numArgs': 0,
                                                         'shortname': 'lck'},
                                  'mainListConnection': {'args': '<type __builtin__.str>',
                                                         'longname': 'mainListConnection',
                                                         'numArgs': 1,
                                                         'shortname': 'mlc'},
                                  'marquee': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'marquee',
                                              'numArgs': 4,
                                              'shortname': 'mq'},
                                  'nbImages': {'args': '<type __builtin__.bool>',
                                               'longname': 'nbImages',
                                               'numArgs': 0,
                                               'shortname': 'nim'},
                                  'panel': {'args': '<type __builtin__.str>',
                                            'longname': 'panel',
                                            'numArgs': 1,
                                            'shortname': 'pnl'},
                                  'parent': {'args': '<type __builtin__.str>',
                                             'longname': 'parent',
                                             'numArgs': 1,
                                             'shortname': 'p'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'realSize': {'args': '<type __builtin__.bool>',
                                               'longname': 'realSize',
                                               'numArgs': 0,
                                               'shortname': 'rs'},
                                  'refresh': {'args': '<type __builtin__.bool>',
                                              'longname': 'refresh',
                                              'numArgs': 0,
                                              'shortname': 'ref'},
                                  'removeAllImages': {'args': '<type __builtin__.bool>',
                                                      'longname': 'removeAllImages',
                                                      'numArgs': 0,
                                                      'shortname': 'ra'},
                                  'removeImage': {'args': '<type __builtin__.bool>',
                                                  'longname': 'removeImage',
                                                  'numArgs': 0,
                                                  'shortname': 'ri'},
                                  'saveImage': {'args': '<type __builtin__.bool>',
                                                'longname': 'saveImage',
                                                'numArgs': 0,
                                                'shortname': 'si'},
                                  'scaleBlue': {'args': '<type __builtin__.float>',
                                                'longname': 'scaleBlue',
                                                'numArgs': 1,
                                                'shortname': 'sb'},
                                  'scaleGreen': {'args': '<type __builtin__.float>',
                                                 'longname': 'scaleGreen',
                                                 'numArgs': 1,
                                                 'shortname': 'sg'},
                                  'scaleRed': {'args': '<type __builtin__.float>',
                                               'longname': 'scaleRed',
                                               'numArgs': 1,
                                               'shortname': 'sr'},
                                  'selectionConnection': {'args': '<type __builtin__.str>',
                                                          'longname': 'selectionConnection',
                                                          'numArgs': 1,
                                                          'shortname': 'slc'},
                                  'showRegion': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'showRegion',
                                                 'numArgs': 2,
                                                 'shortname': 'srg'},
                                  'singleBuffer': {'args': '<type __builtin__.bool>',
                                                   'longname': 'singleBuffer',
                                                   'numArgs': 0,
                                                   'shortname': 'sbf'},
                                  'stateString': {'args': '<type __builtin__.bool>',
                                                  'longname': 'stateString',
                                                  'numArgs': 0,
                                                  'shortname': 'sts'},
                                  'toggle': {'args': '<type __builtin__.bool>',
                                             'longname': 'toggle',
                                             'numArgs': 1,
                                             'shortname': 'tgl'},
                                  'unParent': {'args': '<type __builtin__.bool>',
                                               'longname': 'unParent',
                                               'numArgs': 0,
                                               'shortname': 'up'},
                                  'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                           'longname': 'unlockMainConnection',
                                                           'numArgs': 0,
                                                           'shortname': 'ulk'},
                                  'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                           'longname': 'updateMainConnection',
                                                           'numArgs': 0,
                                                           'shortname': 'upd'},
                                  'useTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'useTemplate',
                                                  'numArgs': 1,
                                                  'shortname': 'ut'},
                                  'writeImage': {'args': '<type __builtin__.str>',
                                                 'longname': 'writeImage',
                                                 'numArgs': 1,
                                                 'shortname': 'wi'}},
                        'shortFlags': {'ar': 'autoResize',
                                       'cc': 'changeCommand',
                                       'cl': 'clear',
                                       'ctl': 'control',
                                       'da': 'drawAxis',
                                       'dbf': 'doubleBuffer',
                                       'di': 'displayImage',
                                       'dst': 'displayStyle',
                                       'dt': 'defineTemplate',
                                       'dtg': 'docTag',
                                       'e': 'edit',
                                       'ex': 'exists',
                                       'f': 'filter',
                                       'fi': 'frameImage',
                                       'fmc': 'forceMainConnection',
                                       'fr': 'frameRegion',
                                       'hlc': 'highlightConnection',
                                       'lck': 'lockMainConnection',
                                       'li': 'loadImage',
                                       'mlc': 'mainListConnection',
                                       'mq': 'marquee',
                                       'nim': 'nbImages',
                                       'p': 'parent',
                                       'pnl': 'panel',
                                       'q': 'query',
                                       'ra': 'removeAllImages',
                                       'ref': 'refresh',
                                       'ri': 'removeImage',
                                       'rs': 'realSize',
                                       'sb': 'scaleBlue',
                                       'sbf': 'singleBuffer',
                                       'sg': 'scaleGreen',
                                       'si': 'saveImage',
                                       'slc': 'selectionConnection',
                                       'sr': 'scaleRed',
                                       'srg': 'showRegion',
                                       'sts': 'stateString',
                                       'tgl': 'toggle',
                                       'ulk': 'unlockMainConnection',
                                       'up': 'unParent',
                                       'upd': 'updateMainConnection',
                                       'ut': 'useTemplate',
                                       'wi': 'writeImage'},
                        'type': 'other'},
  'imfPlugins': {'flags': {'extension': {'args': '<type __builtin__.str>',
                                         'longname': 'extension',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'ext'},
                           'keyword': {'args': '<type __builtin__.str>',
                                       'longname': 'keyword',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'key'},
                           'multiFrameSupport': {'args': '<type __builtin__.str>',
                                                 'longname': 'multiFrameSupport',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'mfs'},
                           'pluginName': {'args': '<type __builtin__.str>',
                                          'longname': 'pluginName',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'pn'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'readSupport': {'args': '<type __builtin__.str>',
                                           'longname': 'readSupport',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'rs'},
                           'writeSupport': {'args': '<type __builtin__.str>',
                                            'longname': 'writeSupport',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'ws'}},
                 'shortFlags': {'ext': 'extension',
                                'key': 'keyword',
                                'mfs': 'multiFrameSupport',
                                'pn': 'pluginName',
                                'q': 'query',
                                'rs': 'readSupport',
                                'ws': 'writeSupport'},
                 'type': 'system'},
  'inViewEditor': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'v'}},
                   'shortFlags': {'q': 'query', 'v': 'visible'},
                   'type': 'windows'},
  'inViewMessage': {'flags': {'alpha': {'args': '<type __builtin__.float>',
                                        'longname': 'alpha',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'a'},
                              'assistMessage': {'args': '<type __builtin__.str>',
                                                'longname': 'assistMessage',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'amg'},
                              'backColor': {'args': '<type __builtin__.int>',
                                            'longname': 'backColor',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'bkc'},
                              'clear': {'args': '<type __builtin__.str>',
                                        'longname': 'clear',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'cl'},
                              'clickKill': {'args': '<type __builtin__.bool>',
                                            'longname': 'clickKill',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ck'},
                              'dragKill': {'args': '<type __builtin__.bool>',
                                           'longname': 'dragKill',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'dk'},
                              'fade': {'args': '<type __builtin__.bool>',
                                       'longname': 'fade',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'f'},
                              'fadeInTime': {'args': '<type __builtin__.int>',
                                             'longname': 'fadeInTime',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'fit'},
                              'fadeOutTime': {'args': '<type __builtin__.int>',
                                              'longname': 'fadeOutTime',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'fot'},
                              'fadeStayTime': {'args': '<type __builtin__.int>',
                                               'longname': 'fadeStayTime',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'fst'},
                              'font': {'args': '<type __builtin__.str>',
                                       'longname': 'font',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ft'},
                              'fontSize': {'args': '<type __builtin__.int>',
                                           'longname': 'fontSize',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'fts'},
                              'frameOffset': {'args': '<type __builtin__.int>',
                                              'longname': 'frameOffset',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'fof'},
                              'hide': {'args': '<type __builtin__.bool>',
                                       'longname': 'hide',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'hd'},
                              'message': {'args': '<type __builtin__.str>',
                                          'longname': 'message',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'msg'},
                              'minimize': {'args': '<type __builtin__.bool>',
                                           'longname': 'minimize',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'min'},
                              'position': {'args': '<type __builtin__.str>',
                                           'longname': 'position',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'pos'},
                              'restore': {'args': '<type __builtin__.bool>',
                                          'longname': 'restore',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'res'},
                              'show': {'args': '<type __builtin__.bool>',
                                       'longname': 'show',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'sh'},
                              'statusMessage': {'args': '<type __builtin__.str>',
                                                'longname': 'statusMessage',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'smg'},
                              'textAlpha': {'args': '<type __builtin__.float>',
                                            'longname': 'textAlpha',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ta'},
                              'textOffset': {'args': '<type __builtin__.int>',
                                             'longname': 'textOffset',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'tof'},
                              'uvEditor': {'args': '<type __builtin__.bool>',
                                           'longname': 'uvEditor',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'uv'}},
                    'shortFlags': {'a': 'alpha',
                                   'amg': 'assistMessage',
                                   'bkc': 'backColor',
                                   'ck': 'clickKill',
                                   'cl': 'clear',
                                   'dk': 'dragKill',
                                   'f': 'fade',
                                   'fit': 'fadeInTime',
                                   'fof': 'frameOffset',
                                   'fot': 'fadeOutTime',
                                   'fst': 'fadeStayTime',
                                   'ft': 'font',
                                   'fts': 'fontSize',
                                   'hd': 'hide',
                                   'min': 'minimize',
                                   'msg': 'message',
                                   'pos': 'position',
                                   'res': 'restore',
                                   'sh': 'show',
                                   'smg': 'statusMessage',
                                   'ta': 'textAlpha',
                                   'tof': 'textOffset',
                                   'uv': 'uvEditor'},
                    'type': 'windows'},
  'inheritTransform': {'flags': {'off': {'args': '<type __builtin__.bool>',
                                         'longname': 'off',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'off'},
                                 'on': {'args': '<type __builtin__.bool>',
                                        'longname': 'on',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'on'},
                                 'preserve': {'args': '<type __builtin__.bool>',
                                              'longname': 'preserve',
                                              'modes': ['create', 'query'],
                                              'numArgs': 0,
                                              'shortname': 'p'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'toggle': {'args': '<type __builtin__.bool>',
                                            'longname': 'toggle',
                                            'modes': ['create', 'query'],
                                            'numArgs': 0,
                                            'shortname': 'tgl'}},
                       'shortFlags': {'off': 'off',
                                      'on': 'on',
                                      'p': 'preserve',
                                      'q': 'query',
                                      'tgl': 'toggle'},
                       'type': 'general'},
  'insertJoint': {'flags': {}, 'shortFlags': {}, 'type': 'animation'},
  'insertJointCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'}},
                     'shortFlags': {'e': 'edit',
                                    'ex': 'exists',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'q': 'query'},
                     'type': 'context'},
  'insertKeyCtx': {'flags': {'breakdown': {'args': '<type __builtin__.bool>',
                                           'longname': 'breakdown',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'bd'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'history': {'args': '<type __builtin__.bool>',
                                         'longname': 'history',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ch'},
                             'image1': {'args': '<type __builtin__.str>',
                                        'longname': 'image1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i1'},
                             'image2': {'args': '<type __builtin__.str>',
                                        'longname': 'image2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i2'},
                             'image3': {'args': '<type __builtin__.str>',
                                        'longname': 'image3',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i3'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'bd': 'breakdown',
                                  'ch': 'history',
                                  'e': 'edit',
                                  'ex': 'exists',
                                  'i1': 'image1',
                                  'i2': 'image2',
                                  'i3': 'image3',
                                  'n': 'name',
                                  'q': 'query'},
                   'type': 'context'},
  'insertKnotCurve': {'flags': {'addKnots': {'args': '<type __builtin__.bool>',
                                             'longname': 'addKnots',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'add'},
                                'caching': {'args': '<type __builtin__.bool>',
                                            'longname': 'caching',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cch'},
                                'constructionHistory': {'args': '<type __builtin__.bool>',
                                                        'longname': 'constructionHistory',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'ch'},
                                'curveOnSurface': {'args': '<type __builtin__.bool>',
                                                   'longname': 'curveOnSurface',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'cos'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'frozen': {'args': '<type __builtin__.bool>',
                                           'longname': 'frozen',
                                           'numArgs': 1,
                                           'shortname': 'fzn'},
                                'insertBetween': {'args': '<type __builtin__.bool>',
                                                  'longname': 'insertBetween',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ib'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'nodeState': {'args': '<type __builtin__.int>',
                                              'longname': 'nodeState',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nds'},
                                'numberOfKnots': {'args': '<type __builtin__.int>',
                                                  'longname': 'numberOfKnots',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'nk'},
                                'object': {'args': '<type __builtin__.bool>',
                                           'longname': 'object',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'o'},
                                'parameter': {'args': '<type __builtin__.float>',
                                              'longname': 'parameter',
                                              'modes': ['create',
                                                        'query',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'p'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                    'longname': 'replaceOriginal',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'rpo'}},
                      'shortFlags': {'add': 'addKnots',
                                     'cch': 'caching',
                                     'ch': 'constructionHistory',
                                     'cos': 'curveOnSurface',
                                     'e': 'edit',
                                     'fzn': 'frozen',
                                     'ib': 'insertBetween',
                                     'n': 'name',
                                     'nds': 'nodeState',
                                     'nk': 'numberOfKnots',
                                     'o': 'object',
                                     'p': 'parameter',
                                     'q': 'query',
                                     'rpo': 'replaceOriginal'},
                      'type': 'modeling'},
  'insertKnotSurface': {'flags': {'addKnots': {'args': '<type __builtin__.bool>',
                                               'longname': 'addKnots',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'add'},
                                  'caching': {'args': '<type __builtin__.bool>',
                                              'longname': 'caching',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cch'},
                                  'constructionHistory': {'args': '<type __builtin__.bool>',
                                                          'longname': 'constructionHistory',
                                                          'modes': ['create'],
                                                          'numArgs': 1,
                                                          'shortname': 'ch'},
                                  'direction': {'args': '<type __builtin__.int>',
                                                'longname': 'direction',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'd'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'frozen': {'args': '<type __builtin__.bool>',
                                             'longname': 'frozen',
                                             'numArgs': 1,
                                             'shortname': 'fzn'},
                                  'insertBetween': {'args': '<type __builtin__.bool>',
                                                    'longname': 'insertBetween',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ib'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'nodeState': {'args': '<type __builtin__.int>',
                                                'longname': 'nodeState',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nds'},
                                  'numberOfKnots': {'args': '<type __builtin__.int>',
                                                    'longname': 'numberOfKnots',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 1,
                                                    'shortname': 'nk'},
                                  'object': {'args': '<type __builtin__.bool>',
                                             'longname': 'object',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'o'},
                                  'parameter': {'args': '<type __builtin__.float>',
                                                'longname': 'parameter',
                                                'modes': ['create',
                                                          'query',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'p'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                      'longname': 'replaceOriginal',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'rpo'}},
                        'shortFlags': {'add': 'addKnots',
                                       'cch': 'caching',
                                       'ch': 'constructionHistory',
                                       'd': 'direction',
                                       'e': 'edit',
                                       'fzn': 'frozen',
                                       'ib': 'insertBetween',
                                       'n': 'name',
                                       'nds': 'nodeState',
                                       'nk': 'numberOfKnots',
                                       'o': 'object',
                                       'p': 'parameter',
                                       'q': 'query',
                                       'rpo': 'replaceOriginal'},
                        'type': 'modeling'},
  'instance': {'flags': {'leaf': {'args': '<type __builtin__.bool>',
                                  'longname': 'leaf',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'lf'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'smartTransform': {'args': '<type __builtin__.bool>',
                                            'longname': 'smartTransform',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'st'}},
               'shortFlags': {'lf': 'leaf',
                              'n': 'name',
                              'st': 'smartTransform'},
               'type': 'general'},
  'instanceable': {'flags': {'allow': {'args': '<type __builtin__.bool>',
                                       'longname': 'allow',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'a'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'recursive': {'args': '<type __builtin__.bool>',
                                           'longname': 'recursive',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'r'},
                             'shape': {'args': '<type __builtin__.bool>',
                                       'longname': 'shape',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 's'}},
                   'shortFlags': {'a': 'allow',
                                  'q': 'query',
                                  'r': 'recursive',
                                  's': 'shape'},
                   'type': 'general'},
  'instancer': {'flags': {'addObject': {'args': '<type __builtin__.bool>',
                                        'longname': 'addObject',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'a'},
                          'cycle': {'args': '<type __builtin__.str>',
                                    'longname': 'cycle',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'c'},
                          'cycleStep': {'args': '<type __builtin__.float>',
                                        'longname': 'cycleStep',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cs'},
                          'cycleStepUnits': {'args': '<type __builtin__.str>',
                                             'longname': 'cycleStepUnits',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'csu'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'index': {'args': '<type __builtin__.int>',
                                    'longname': 'index',
                                    'modes': ['query'],
                                    'numArgs': 1,
                                    'shortname': 'i'},
                          'levelOfDetail': {'args': '<type __builtin__.str>',
                                            'longname': 'levelOfDetail',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'lod'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'object': {'args': '<type __builtin__.str>',
                                     'longname': 'object',
                                     'modes': ['create',
                                               'query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'obj'},
                          'objectPosition': {'args': '<type __builtin__.str>',
                                             'longname': 'objectPosition',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'op'},
                          'objectRotation': {'args': '<type __builtin__.str>',
                                             'longname': 'objectRotation',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'objectRotation'},
                          'objectScale': {'args': '<type __builtin__.str>',
                                          'longname': 'objectScale',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'os'},
                          'pointDataSource': {'args': '<type __builtin__.bool>',
                                              'longname': 'pointDataSource',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'pds'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'removeObject': {'args': '<type __builtin__.bool>',
                                           'longname': 'removeObject',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'rm'},
                          'rotationOrder': {'args': '<type __builtin__.str>',
                                            'longname': 'rotationOrder',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ro'},
                          'rotationUnits': {'args': '<type __builtin__.str>',
                                            'longname': 'rotationUnits',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ru'},
                          'valueName': {'args': '<type __builtin__.str>',
                                        'longname': 'valueName',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'vn'}},
                'removedFlags': {'or': 'objectRotation'},
                'shortFlags': {'a': 'addObject',
                               'c': 'cycle',
                               'cs': 'cycleStep',
                               'csu': 'cycleStepUnits',
                               'e': 'edit',
                               'i': 'index',
                               'lod': 'levelOfDetail',
                               'n': 'name',
                               'obj': 'object',
                               'objectRotation': 'objectRotation',
                               'op': 'objectPosition',
                               'os': 'objectScale',
                               'pds': 'pointDataSource',
                               'q': 'query',
                               'rm': 'removeObject',
                               'ro': 'rotationOrder',
                               'ru': 'rotationUnits',
                               'vn': 'valueName'},
                'type': 'general'},
  'intField': {'callbackFlags': ['dragCommand',
                                 'dc',
                                 'dropCallback',
                                 'dpc',
                                 'visibleChangeCommand',
                                 'vcc',
                                 'dragCallback',
                                 'dgc',
                                 'enterCommand',
                                 'ec',
                                 'receiveFocusCommand',
                                 'rfc',
                                 'changeCommand',
                                 'cc'],
               'flags': {'annotation': {'args': '<type __builtin__.str>',
                                        'longname': 'annotation',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ann'},
                         'backgroundColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'backgroundColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'bgc'},
                         'changeCommand': {'args': 'script',
                                           'longname': 'changeCommand',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cc'},
                         'defineTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'defineTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'dt'},
                         'docTag': {'args': '<type __builtin__.str>',
                                    'longname': 'docTag',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'dtg'},
                         'dragCallback': {'args': 'script',
                                          'longname': 'dragCallback',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dgc'},
                         'dragCommand': {'args': 'script',
                                         'longname': 'dragCommand',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dc'},
                         'dropCallback': {'args': 'script',
                                          'longname': 'dropCallback',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dpc'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'editable': {'args': '<type __builtin__.bool>',
                                      'longname': 'editable',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'ed'},
                         'enable': {'args': '<type __builtin__.bool>',
                                    'longname': 'enable',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'en'},
                         'enableBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'enableBackground',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ebg'},
                         'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableKeyboardFocus',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ekf'},
                         'enterCommand': {'args': 'script',
                                          'longname': 'enterCommand',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ec'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'fullPathName': {'args': '<type __builtin__.bool>',
                                          'longname': 'fullPathName',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'fpn'},
                         'height': {'args': '<type __builtin__.int>',
                                    'longname': 'height',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'h'},
                         'highlightColor': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'highlightColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'hlc'},
                         'isObscured': {'args': '<type __builtin__.bool>',
                                        'longname': 'isObscured',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'io'},
                         'manage': {'args': '<type __builtin__.bool>',
                                    'longname': 'manage',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'm'},
                         'maxValue': {'args': '<type __builtin__.int>',
                                      'longname': 'maxValue',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'max'},
                         'minValue': {'args': '<type __builtin__.int>',
                                      'longname': 'minValue',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'min'},
                         'noBackground': {'args': '<type __builtin__.bool>',
                                          'longname': 'noBackground',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nbg'},
                         'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                'longname': 'numberOfPopupMenus',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'npm'},
                         'parent': {'args': '<type __builtin__.str>',
                                    'longname': 'parent',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'p'},
                         'popupMenuArray': {'args': '<type __builtin__.bool>',
                                            'longname': 'popupMenuArray',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'pma'},
                         'preventOverride': {'args': '<type __builtin__.bool>',
                                             'longname': 'preventOverride',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'po'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'receiveFocusCommand': {'args': 'script',
                                                 'longname': 'receiveFocusCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rfc'},
                         'statusBarMessage': {'args': '<type __builtin__.str>',
                                              'longname': 'statusBarMessage',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sbm'},
                         'step': {'args': '<type __builtin__.int>',
                                  'longname': 'step',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 's'},
                         'useTemplate': {'args': '<type __builtin__.str>',
                                         'longname': 'useTemplate',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ut'},
                         'value': {'args': '<type __builtin__.int>',
                                   'longname': 'value',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'v'},
                         'visible': {'args': '<type __builtin__.bool>',
                                     'longname': 'visible',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'vis'},
                         'visibleChangeCommand': {'args': 'script',
                                                  'longname': 'visibleChangeCommand',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'vcc'},
                         'width': {'args': '<type __builtin__.int>',
                                   'longname': 'width',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'w'}},
               'shortFlags': {'ann': 'annotation',
                              'bgc': 'backgroundColor',
                              'cc': 'changeCommand',
                              'dc': 'dragCommand',
                              'dgc': 'dragCallback',
                              'dpc': 'dropCallback',
                              'dt': 'defineTemplate',
                              'dtg': 'docTag',
                              'e': 'edit',
                              'ebg': 'enableBackground',
                              'ec': 'enterCommand',
                              'ed': 'editable',
                              'ekf': 'enableKeyboardFocus',
                              'en': 'enable',
                              'ex': 'exists',
                              'fpn': 'fullPathName',
                              'h': 'height',
                              'hlc': 'highlightColor',
                              'io': 'isObscured',
                              'm': 'manage',
                              'max': 'maxValue',
                              'min': 'minValue',
                              'nbg': 'noBackground',
                              'npm': 'numberOfPopupMenus',
                              'p': 'parent',
                              'pma': 'popupMenuArray',
                              'po': 'preventOverride',
                              'q': 'query',
                              'rfc': 'receiveFocusCommand',
                              's': 'step',
                              'sbm': 'statusBarMessage',
                              'ut': 'useTemplate',
                              'v': 'value',
                              'vcc': 'visibleChangeCommand',
                              'vis': 'visible',
                              'w': 'width'},
               'type': 'windows'},
  'intFieldGrp': {'callbackFlags': ['dragCommand',
                                    'dc',
                                    'visibleChangeCommand',
                                    'vcc',
                                    'dragCallback',
                                    'dgc',
                                    'changeCommand',
                                    'cc',
                                    'dropCallback',
                                    'dpc'],
                  'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                 'longname': 'adjustableColumn',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'adj'},
                            'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn2',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ad2'},
                            'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn3',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ad3'},
                            'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn4',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ad4'},
                            'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn5',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ad5'},
                            'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn6',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ad6'},
                            'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'changeCommand': {'args': 'script',
                                              'longname': 'changeCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cc'},
                            'columnAlign': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'columnAlign',
                                            'modes': ['create',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 2,
                                            'shortname': 'cal'},
                            'columnAlign2': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign2',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 2,
                                             'shortname': 'cl2'},
                            'columnAlign3': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign3',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 3,
                                             'shortname': 'cl3'},
                            'columnAlign4': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign4',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 4,
                                             'shortname': 'cl4'},
                            'columnAlign5': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign5',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 5,
                                             'shortname': 'cl5'},
                            'columnAlign6': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign6',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 6,
                                             'shortname': 'cl6'},
                            'columnAttach': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnAttach',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 3,
                                             'shortname': 'cat'},
                            'columnAttach2': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAttach2',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'ct2'},
                            'columnAttach3': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAttach3',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'ct3'},
                            'columnAttach4': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAttach4',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 4,
                                              'shortname': 'ct4'},
                            'columnAttach5': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAttach5',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 5,
                                              'shortname': 'ct5'},
                            'columnAttach6': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAttach6',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 6,
                                              'shortname': 'ct6'},
                            'columnOffset2': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnOffset2',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'co2'},
                            'columnOffset3': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnOffset3',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'co3'},
                            'columnOffset4': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnOffset4',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 4,
                                              'shortname': 'co4'},
                            'columnOffset5': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnOffset5',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 5,
                                              'shortname': 'co5'},
                            'columnOffset6': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnOffset6',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 6,
                                              'shortname': 'co6'},
                            'columnWidth': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'columnWidth',
                                            'modes': ['create',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 2,
                                            'shortname': 'cw'},
                            'columnWidth1': {'args': '<type __builtin__.int>',
                                             'longname': 'columnWidth1',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cw1'},
                            'columnWidth2': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth2',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 2,
                                             'shortname': 'cw2'},
                            'columnWidth3': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth3',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 3,
                                             'shortname': 'cw3'},
                            'columnWidth4': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth4',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 4,
                                             'shortname': 'cw4'},
                            'columnWidth5': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth5',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 5,
                                             'shortname': 'cw5'},
                            'columnWidth6': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth6',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 6,
                                             'shortname': 'cw6'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dragCommand': {'args': 'script',
                                            'longname': 'dragCommand',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enable1': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en1'},
                            'enable2': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en2'},
                            'enable3': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable3',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en3'},
                            'enable4': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable4',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en4'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'extraLabel': {'args': '<type __builtin__.str>',
                                           'longname': 'extraLabel',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'el'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'label': {'args': '<type __builtin__.str>',
                                      'longname': 'label',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'l'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfFields': {'args': '<type __builtin__.int>',
                                               'longname': 'numberOfFields',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'nf'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'rowAttach': {'args': ['<type __builtin__.int>',
                                                   '<type __builtin__.str>',
                                                   '<type __builtin__.int>'],
                                          'longname': 'rowAttach',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 3,
                                          'shortname': 'rat'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'value': {'args': ['<type __builtin__.int>',
                                               '<type __builtin__.int>',
                                               '<type __builtin__.int>',
                                               '<type __builtin__.int>'],
                                      'longname': 'value',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 4,
                                      'shortname': 'v'},
                            'value1': {'args': '<type __builtin__.int>',
                                       'longname': 'value1',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'v1'},
                            'value2': {'args': '<type __builtin__.int>',
                                       'longname': 'value2',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'v2'},
                            'value3': {'args': '<type __builtin__.int>',
                                       'longname': 'value3',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'v3'},
                            'value4': {'args': '<type __builtin__.int>',
                                       'longname': 'value4',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'v4'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'ad2': 'adjustableColumn2',
                                 'ad3': 'adjustableColumn3',
                                 'ad4': 'adjustableColumn4',
                                 'ad5': 'adjustableColumn5',
                                 'ad6': 'adjustableColumn6',
                                 'adj': 'adjustableColumn',
                                 'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'cal': 'columnAlign',
                                 'cat': 'columnAttach',
                                 'cc': 'changeCommand',
                                 'cl2': 'columnAlign2',
                                 'cl3': 'columnAlign3',
                                 'cl4': 'columnAlign4',
                                 'cl5': 'columnAlign5',
                                 'cl6': 'columnAlign6',
                                 'co2': 'columnOffset2',
                                 'co3': 'columnOffset3',
                                 'co4': 'columnOffset4',
                                 'co5': 'columnOffset5',
                                 'co6': 'columnOffset6',
                                 'ct2': 'columnAttach2',
                                 'ct3': 'columnAttach3',
                                 'ct4': 'columnAttach4',
                                 'ct5': 'columnAttach5',
                                 'ct6': 'columnAttach6',
                                 'cw': 'columnWidth',
                                 'cw1': 'columnWidth1',
                                 'cw2': 'columnWidth2',
                                 'cw3': 'columnWidth3',
                                 'cw4': 'columnWidth4',
                                 'cw5': 'columnWidth5',
                                 'cw6': 'columnWidth6',
                                 'dc': 'dragCommand',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ekf': 'enableKeyboardFocus',
                                 'el': 'extraLabel',
                                 'en': 'enable',
                                 'en1': 'enable1',
                                 'en2': 'enable2',
                                 'en3': 'enable3',
                                 'en4': 'enable4',
                                 'ex': 'exists',
                                 'fpn': 'fullPathName',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'io': 'isObscured',
                                 'l': 'label',
                                 'm': 'manage',
                                 'nbg': 'noBackground',
                                 'nf': 'numberOfFields',
                                 'npm': 'numberOfPopupMenus',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'rat': 'rowAttach',
                                 'sbm': 'statusBarMessage',
                                 'ut': 'useTemplate',
                                 'v': 'value',
                                 'v1': 'value1',
                                 'v2': 'value2',
                                 'v3': 'value3',
                                 'v4': 'value4',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'intScrollBar': {'callbackFlags': ['dragCommand',
                                     'dc',
                                     'dropCallback',
                                     'dpc',
                                     'visibleChangeCommand',
                                     'vcc',
                                     'dragCallback',
                                     'dgc',
                                     'changeCommand',
                                     'cc'],
                   'flags': {'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'changeCommand': {'args': 'script',
                                               'longname': 'changeCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cc'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dragCommand': {'args': 'script',
                                             'longname': 'dragCommand',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'horizontal': {'args': '<type __builtin__.bool>',
                                            'longname': 'horizontal',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'hr'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'largeStep': {'args': '<type __builtin__.int>',
                                           'longname': 'largeStep',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ls'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'maxValue': {'args': '<type __builtin__.int>',
                                          'longname': 'maxValue',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'max'},
                             'minValue': {'args': '<type __builtin__.int>',
                                          'longname': 'minValue',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'min'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'step': {'args': '<type __builtin__.int>',
                                      'longname': 'step',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'value': {'args': '<type __builtin__.int>',
                                       'longname': 'value',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'v'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'shortFlags': {'ann': 'annotation',
                                  'bgc': 'backgroundColor',
                                  'cc': 'changeCommand',
                                  'dc': 'dragCommand',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ebg': 'enableBackground',
                                  'ekf': 'enableKeyboardFocus',
                                  'en': 'enable',
                                  'ex': 'exists',
                                  'fpn': 'fullPathName',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'hr': 'horizontal',
                                  'io': 'isObscured',
                                  'ls': 'largeStep',
                                  'm': 'manage',
                                  'max': 'maxValue',
                                  'min': 'minValue',
                                  'nbg': 'noBackground',
                                  'npm': 'numberOfPopupMenus',
                                  'p': 'parent',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'q': 'query',
                                  's': 'step',
                                  'sbm': 'statusBarMessage',
                                  'ut': 'useTemplate',
                                  'v': 'value',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'w': 'width'},
                   'type': 'windows'},
  'intSlider': {'callbackFlags': ['dragCommand',
                                  'dc',
                                  'dropCallback',
                                  'dpc',
                                  'visibleChangeCommand',
                                  'vcc',
                                  'dragCallback',
                                  'dgc',
                                  'changeCommand',
                                  'cc'],
                'flags': {'annotation': {'args': '<type __builtin__.str>',
                                         'longname': 'annotation',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ann'},
                          'backgroundColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'backgroundColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'bgc'},
                          'changeCommand': {'args': 'script',
                                            'longname': 'changeCommand',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cc'},
                          'defineTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'defineTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dt'},
                          'docTag': {'args': '<type __builtin__.str>',
                                     'longname': 'docTag',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'dtg'},
                          'dragCallback': {'args': 'script',
                                           'longname': 'dragCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dgc'},
                          'dragCommand': {'args': 'script',
                                          'longname': 'dragCommand',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dc'},
                          'dropCallback': {'args': 'script',
                                           'longname': 'dropCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dpc'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'enable': {'args': '<type __builtin__.bool>',
                                     'longname': 'enable',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'en'},
                          'enableBackground': {'args': '<type __builtin__.bool>',
                                               'longname': 'enableBackground',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ebg'},
                          'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableKeyboardFocus',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ekf'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'fullPathName': {'args': '<type __builtin__.bool>',
                                           'longname': 'fullPathName',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'fpn'},
                          'height': {'args': '<type __builtin__.int>',
                                     'longname': 'height',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'h'},
                          'highlightColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'highlightColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'hlc'},
                          'horizontal': {'args': '<type __builtin__.bool>',
                                         'longname': 'horizontal',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'hr'},
                          'isObscured': {'args': '<type __builtin__.bool>',
                                         'longname': 'isObscured',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'io'},
                          'manage': {'args': '<type __builtin__.bool>',
                                     'longname': 'manage',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'm'},
                          'maxValue': {'args': '<type __builtin__.int>',
                                       'longname': 'maxValue',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'max'},
                          'minValue': {'args': '<type __builtin__.int>',
                                       'longname': 'minValue',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'min'},
                          'noBackground': {'args': '<type __builtin__.bool>',
                                           'longname': 'noBackground',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nbg'},
                          'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'numberOfPopupMenus',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'npm'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                          'popupMenuArray': {'args': '<type __builtin__.bool>',
                                             'longname': 'popupMenuArray',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'pma'},
                          'preventOverride': {'args': '<type __builtin__.bool>',
                                              'longname': 'preventOverride',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'po'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'statusBarMessage': {'args': '<type __builtin__.str>',
                                               'longname': 'statusBarMessage',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sbm'},
                          'step': {'args': '<type __builtin__.int>',
                                   'longname': 'step',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 's'},
                          'useTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'useTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ut'},
                          'value': {'args': '<type __builtin__.int>',
                                    'longname': 'value',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'v'},
                          'visible': {'args': '<type __builtin__.bool>',
                                      'longname': 'visible',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'vis'},
                          'visibleChangeCommand': {'args': 'script',
                                                   'longname': 'visibleChangeCommand',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'vcc'},
                          'width': {'args': '<type __builtin__.int>',
                                    'longname': 'width',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'w'}},
                'shortFlags': {'ann': 'annotation',
                               'bgc': 'backgroundColor',
                               'cc': 'changeCommand',
                               'dc': 'dragCommand',
                               'dgc': 'dragCallback',
                               'dpc': 'dropCallback',
                               'dt': 'defineTemplate',
                               'dtg': 'docTag',
                               'e': 'edit',
                               'ebg': 'enableBackground',
                               'ekf': 'enableKeyboardFocus',
                               'en': 'enable',
                               'ex': 'exists',
                               'fpn': 'fullPathName',
                               'h': 'height',
                               'hlc': 'highlightColor',
                               'hr': 'horizontal',
                               'io': 'isObscured',
                               'm': 'manage',
                               'max': 'maxValue',
                               'min': 'minValue',
                               'nbg': 'noBackground',
                               'npm': 'numberOfPopupMenus',
                               'p': 'parent',
                               'pma': 'popupMenuArray',
                               'po': 'preventOverride',
                               'q': 'query',
                               's': 'step',
                               'sbm': 'statusBarMessage',
                               'ut': 'useTemplate',
                               'v': 'value',
                               'vcc': 'visibleChangeCommand',
                               'vis': 'visible',
                               'w': 'width'},
                'type': 'windows'},
  'intSliderGrp': {'callbackFlags': ['dragCommand',
                                     'dc',
                                     'visibleChangeCommand',
                                     'vcc',
                                     'dragCallback',
                                     'dgc',
                                     'changeCommand',
                                     'cc',
                                     'dropCallback',
                                     'dpc'],
                   'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'adj'},
                             'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn2',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad2'},
                             'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn3',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad3'},
                             'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn4',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad4'},
                             'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn5',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad5'},
                             'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn6',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad6'},
                             'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'changeCommand': {'args': 'script',
                                               'longname': 'changeCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cc'},
                             'columnAlign': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 2,
                                             'shortname': 'cal'},
                             'columnAlign2': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign2',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'cl2'},
                             'columnAlign3': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign3',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'cl3'},
                             'columnAlign4': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign4',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 4,
                                              'shortname': 'cl4'},
                             'columnAlign5': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign5',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 5,
                                              'shortname': 'cl5'},
                             'columnAlign6': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign6',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 6,
                                              'shortname': 'cl6'},
                             'columnAttach': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnAttach',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 3,
                                              'shortname': 'cat'},
                             'columnAttach2': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'ct2'},
                             'columnAttach3': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'ct3'},
                             'columnAttach4': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'ct4'},
                             'columnAttach5': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'ct5'},
                             'columnAttach6': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'ct6'},
                             'columnOffset2': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'co2'},
                             'columnOffset3': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'co3'},
                             'columnOffset4': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'co4'},
                             'columnOffset5': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'co5'},
                             'columnOffset6': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'co6'},
                             'columnWidth': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 2,
                                             'shortname': 'cw'},
                             'columnWidth1': {'args': '<type __builtin__.int>',
                                              'longname': 'columnWidth1',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cw1'},
                             'columnWidth2': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth2',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'cw2'},
                             'columnWidth3': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth3',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'cw3'},
                             'columnWidth4': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth4',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 4,
                                              'shortname': 'cw4'},
                             'columnWidth5': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth5',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 5,
                                              'shortname': 'cw5'},
                             'columnWidth6': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth6',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 6,
                                              'shortname': 'cw6'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dragCommand': {'args': 'script',
                                             'longname': 'dragCommand',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'extraLabel': {'args': '<type __builtin__.str>',
                                            'longname': 'extraLabel',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'el'},
                             'field': {'args': '<type __builtin__.bool>',
                                       'longname': 'field',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'f'},
                             'fieldMaxValue': {'args': '<type __builtin__.int>',
                                               'longname': 'fieldMaxValue',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'fmx'},
                             'fieldMinValue': {'args': '<type __builtin__.int>',
                                               'longname': 'fieldMinValue',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'fmn'},
                             'fieldStep': {'args': '<type __builtin__.int>',
                                           'longname': 'fieldStep',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'fs'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'label': {'args': '<type __builtin__.str>',
                                       'longname': 'label',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'l'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'maxValue': {'args': '<type __builtin__.int>',
                                          'longname': 'maxValue',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'max'},
                             'minValue': {'args': '<type __builtin__.int>',
                                          'longname': 'minValue',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'min'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'rowAttach': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'rowAttach',
                                           'modes': ['create',
                                                     'edit',
                                                     'multiuse'],
                                           'numArgs': 3,
                                           'shortname': 'rat'},
                             'sliderStep': {'args': '<type __builtin__.int>',
                                            'longname': 'sliderStep',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ss'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'step': {'args': '<type __builtin__.int>',
                                      'longname': 'step',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'value': {'args': '<type __builtin__.int>',
                                       'longname': 'value',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'v'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'shortFlags': {'ad2': 'adjustableColumn2',
                                  'ad3': 'adjustableColumn3',
                                  'ad4': 'adjustableColumn4',
                                  'ad5': 'adjustableColumn5',
                                  'ad6': 'adjustableColumn6',
                                  'adj': 'adjustableColumn',
                                  'ann': 'annotation',
                                  'bgc': 'backgroundColor',
                                  'cal': 'columnAlign',
                                  'cat': 'columnAttach',
                                  'cc': 'changeCommand',
                                  'cl2': 'columnAlign2',
                                  'cl3': 'columnAlign3',
                                  'cl4': 'columnAlign4',
                                  'cl5': 'columnAlign5',
                                  'cl6': 'columnAlign6',
                                  'co2': 'columnOffset2',
                                  'co3': 'columnOffset3',
                                  'co4': 'columnOffset4',
                                  'co5': 'columnOffset5',
                                  'co6': 'columnOffset6',
                                  'ct2': 'columnAttach2',
                                  'ct3': 'columnAttach3',
                                  'ct4': 'columnAttach4',
                                  'ct5': 'columnAttach5',
                                  'ct6': 'columnAttach6',
                                  'cw': 'columnWidth',
                                  'cw1': 'columnWidth1',
                                  'cw2': 'columnWidth2',
                                  'cw3': 'columnWidth3',
                                  'cw4': 'columnWidth4',
                                  'cw5': 'columnWidth5',
                                  'cw6': 'columnWidth6',
                                  'dc': 'dragCommand',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ebg': 'enableBackground',
                                  'ekf': 'enableKeyboardFocus',
                                  'el': 'extraLabel',
                                  'en': 'enable',
                                  'ex': 'exists',
                                  'f': 'field',
                                  'fmn': 'fieldMinValue',
                                  'fmx': 'fieldMaxValue',
                                  'fpn': 'fullPathName',
                                  'fs': 'fieldStep',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'io': 'isObscured',
                                  'l': 'label',
                                  'm': 'manage',
                                  'max': 'maxValue',
                                  'min': 'minValue',
                                  'nbg': 'noBackground',
                                  'npm': 'numberOfPopupMenus',
                                  'p': 'parent',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'q': 'query',
                                  'rat': 'rowAttach',
                                  's': 'step',
                                  'sbm': 'statusBarMessage',
                                  'ss': 'sliderStep',
                                  'ut': 'useTemplate',
                                  'v': 'value',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'w': 'width'},
                   'type': 'windows'},
  'interactionStyle': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'style': {'args': '<type __builtin__.str>',
                                           'longname': 'style',
                                           'numArgs': 1,
                                           'shortname': 's'}},
                       'shortFlags': {'q': 'query', 's': 'style'},
                       'type': 'other'},
  'internalVar': {'flags': {'userAppDir': {'args': '<type __builtin__.bool>',
                                           'longname': 'userAppDir',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'uad'},
                            'userBitmapsDir': {'args': '<type __builtin__.bool>',
                                               'longname': 'userBitmapsDir',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ubd'},
                            'userHotkeyDir': {'args': '<type __builtin__.bool>',
                                              'longname': 'userHotkeyDir',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'uhk'},
                            'userMarkingMenuDir': {'args': '<type __builtin__.bool>',
                                                   'longname': 'userMarkingMenuDir',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'umm'},
                            'userPrefDir': {'args': '<type __builtin__.bool>',
                                            'longname': 'userPrefDir',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'upd'},
                            'userPresetsDir': {'args': '<type __builtin__.bool>',
                                               'longname': 'userPresetsDir',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ups'},
                            'userScriptDir': {'args': '<type __builtin__.bool>',
                                              'longname': 'userScriptDir',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'usd'},
                            'userShelfDir': {'args': '<type __builtin__.bool>',
                                             'longname': 'userShelfDir',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ush'},
                            'userTmpDir': {'args': '<type __builtin__.bool>',
                                           'longname': 'userTmpDir',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'utd'},
                            'userWorkspaceDir': {'args': '<type __builtin__.bool>',
                                                 'longname': 'userWorkspaceDir',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'uwd'}},
                  'shortFlags': {'uad': 'userAppDir',
                                 'ubd': 'userBitmapsDir',
                                 'uhk': 'userHotkeyDir',
                                 'umm': 'userMarkingMenuDir',
                                 'upd': 'userPrefDir',
                                 'ups': 'userPresetsDir',
                                 'usd': 'userScriptDir',
                                 'ush': 'userShelfDir',
                                 'utd': 'userTmpDir',
                                 'uwd': 'userWorkspaceDir'},
                  'type': 'system'},
  'intersect': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                      'longname': 'caching',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'cch'},
                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'constructionHistory',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                          'curveOnSurface': {'args': '<type __builtin__.bool>',
                                             'longname': 'curveOnSurface',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'cos'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'firstSurface': {'args': '<type __builtin__.bool>',
                                           'longname': 'firstSurface',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'fs'},
                          'frozen': {'args': '<type __builtin__.bool>',
                                     'longname': 'frozen',
                                     'numArgs': 1,
                                     'shortname': 'fzn'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nodeState': {'args': '<type __builtin__.int>',
                                        'longname': 'nodeState',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'nds'},
                          'object': {'args': '<type __builtin__.bool>',
                                     'longname': 'object',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'o'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'tolerance': {'args': '<type __builtin__.float>',
                                        'longname': 'tolerance',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'tol'}},
                'shortFlags': {'cch': 'caching',
                               'ch': 'constructionHistory',
                               'cos': 'curveOnSurface',
                               'e': 'edit',
                               'fs': 'firstSurface',
                               'fzn': 'frozen',
                               'n': 'name',
                               'nds': 'nodeState',
                               'o': 'object',
                               'q': 'query',
                               'tol': 'tolerance'},
                'type': 'modeling'},
  'iprEngine': {'flags': {'copy': {'args': '<type __builtin__.str>',
                                   'longname': 'copy',
                                   'modes': ['edit'],
                                   'numArgs': 1,
                                   'shortname': 'cp'},
                          'defineTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'defineTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dt'},
                          'diagnostics': {'args': '<type __builtin__.bool>',
                                          'longname': 'diagnostics',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'dig'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'estimatedMemory': {'args': '<type __builtin__.bool>',
                                              'longname': 'estimatedMemory',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'mem'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'iprImage': {'args': '<type __builtin__.str>',
                                       'longname': 'iprImage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ipr'},
                          'motionVectorFile': {'args': '<type __builtin__.bool>',
                                               'longname': 'motionVectorFile',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'mvf'},
                          'object': {'args': 'PyNode',
                                     'longname': 'object',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'obj'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'region': {'args': ['<type __builtin__.int>',
                                              '<type __builtin__.int>',
                                              '<type __builtin__.int>',
                                              '<type __builtin__.int>'],
                                     'longname': 'region',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 4,
                                     'shortname': 'r'},
                          'relatedFiles': {'args': '<type __builtin__.bool>',
                                           'longname': 'relatedFiles',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'rel'},
                          'releaseIprImage': {'args': '<type __builtin__.bool>',
                                              'longname': 'releaseIprImage',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'rii'},
                          'resolution': {'args': '<type __builtin__.bool>',
                                         'longname': 'resolution',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'res'},
                          'scanlineIncrement': {'args': '<type __builtin__.str>',
                                                'longname': 'scanlineIncrement',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sli'},
                          'showProgressBar': {'args': '<type __builtin__.bool>',
                                              'longname': 'showProgressBar',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'spb'},
                          'startTuning': {'args': '<type __builtin__.bool>',
                                          'longname': 'startTuning',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'st'},
                          'stopTuning': {'args': '<type __builtin__.bool>',
                                         'longname': 'stopTuning',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 0,
                                         'shortname': 'spt'},
                          'underPixel': {'args': ['<type __builtin__.int>',
                                                  '<type __builtin__.int>'],
                                         'longname': 'underPixel',
                                         'modes': ['edit'],
                                         'numArgs': 2,
                                         'shortname': 'un'},
                          'update': {'args': '<type __builtin__.bool>',
                                     'longname': 'update',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'u'},
                          'updateDepthOfField': {'args': '<type __builtin__.bool>',
                                                 'longname': 'updateDepthOfField',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'udf'},
                          'updateLightGlow': {'args': '<type __builtin__.bool>',
                                              'longname': 'updateLightGlow',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ulg'},
                          'updateMotionBlur': {'args': '<type __builtin__.bool>',
                                               'longname': 'updateMotionBlur',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'umb'},
                          'updatePort': {'args': '<type __builtin__.str>',
                                         'longname': 'updatePort',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'up'},
                          'updateShaderGlow': {'args': '<type __builtin__.bool>',
                                               'longname': 'updateShaderGlow',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'usg'},
                          'updateShading': {'args': '<type __builtin__.bool>',
                                            'longname': 'updateShading',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'us'},
                          'updateShadowMaps': {'args': '<type __builtin__.bool>',
                                               'longname': 'updateShadowMaps',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'usm'},
                          'useTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'useTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ut'}},
                'shortFlags': {'cp': 'copy',
                               'dig': 'diagnostics',
                               'dt': 'defineTemplate',
                               'e': 'edit',
                               'ex': 'exists',
                               'ipr': 'iprImage',
                               'mem': 'estimatedMemory',
                               'mvf': 'motionVectorFile',
                               'obj': 'object',
                               'q': 'query',
                               'r': 'region',
                               'rel': 'relatedFiles',
                               'res': 'resolution',
                               'rii': 'releaseIprImage',
                               'sli': 'scanlineIncrement',
                               'spb': 'showProgressBar',
                               'spt': 'stopTuning',
                               'st': 'startTuning',
                               'u': 'update',
                               'udf': 'updateDepthOfField',
                               'ulg': 'updateLightGlow',
                               'umb': 'updateMotionBlur',
                               'un': 'underPixel',
                               'up': 'updatePort',
                               'us': 'updateShading',
                               'usg': 'updateShaderGlow',
                               'usm': 'updateShadowMaps',
                               'ut': 'useTemplate'},
                'type': 'rendering'},
  'isConnected': {'flags': {'ignoreUnitConversion': {'args': '<type __builtin__.bool>',
                                                     'longname': 'ignoreUnitConversion',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'iuc'}},
                  'shortFlags': {'iuc': 'ignoreUnitConversion'},
                  'type': 'general'},
  'isDirty': {'flags': {'connection': {'args': '<type __builtin__.bool>',
                                       'longname': 'connection',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'c'},
                        'datablock': {'args': '<type __builtin__.bool>',
                                      'longname': 'datablock',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'd'}},
              'shortFlags': {'c': 'connection', 'd': 'datablock'},
              'type': 'general'},
  'isTrue': {'flags': {}, 'shortFlags': {}, 'type': 'language'},
  'isolateSelect': {'flags': {'addDagObject': {'args': 'PyNode',
                                               'longname': 'addDagObject',
                                               'modes': [],
                                               'numArgs': 1,
                                               'shortname': 'ado'},
                              'addSelected': {'args': '<type __builtin__.bool>',
                                              'longname': 'addSelected',
                                              'modes': [],
                                              'numArgs': 0,
                                              'shortname': 'addSelected'},
                              'addSelectedObjects': {'args': '<type __builtin__.bool>',
                                                     'longname': 'addSelectedObjects',
                                                     'modes': [],
                                                     'numArgs': 0,
                                                     'shortname': 'aso'},
                              'loadSelected': {'args': '<type __builtin__.bool>',
                                               'longname': 'loadSelected',
                                               'modes': [],
                                               'numArgs': 0,
                                               'shortname': 'ls'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'removeDagObject': {'args': 'PyNode',
                                                  'longname': 'removeDagObject',
                                                  'modes': [],
                                                  'numArgs': 1,
                                                  'shortname': 'rdo'},
                              'removeSelected': {'args': '<type __builtin__.bool>',
                                                 'longname': 'removeSelected',
                                                 'modes': [],
                                                 'numArgs': 0,
                                                 'shortname': 'rs'},
                              'state': {'args': '<type __builtin__.bool>',
                                        'longname': 'state',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 's'},
                              'update': {'args': '<type __builtin__.bool>',
                                         'longname': 'update',
                                         'modes': [],
                                         'numArgs': 0,
                                         'shortname': 'u'},
                              'viewObjects': {'args': '<type __builtin__.bool>',
                                              'longname': 'viewObjects',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'vo'}},
                    'removedFlags': {'as': 'addSelected'},
                    'shortFlags': {'addSelected': 'addSelected',
                                   'ado': 'addDagObject',
                                   'aso': 'addSelectedObjects',
                                   'ls': 'loadSelected',
                                   'q': 'query',
                                   'rdo': 'removeDagObject',
                                   'rs': 'removeSelected',
                                   's': 'state',
                                   'u': 'update',
                                   'vo': 'viewObjects'},
                    'type': 'general'},
  'itemFilter': {'flags': {'byBin': {'args': '<type __builtin__.str>',
                                     'longname': 'byBin',
                                     'modes': ['create',
                                               'query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'bk'},
                           'byName': {'args': '<type __builtin__.str>',
                                      'longname': 'byName',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'bn'},
                           'byScript': {'args': '<type __builtin__.str>',
                                        'longname': 'byScript',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'bs'},
                           'byType': {'args': '<type __builtin__.str>',
                                      'longname': 'byType',
                                      'modes': ['create',
                                                'query',
                                                'edit',
                                                'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'bt'},
                           'category': {'args': '<type __builtin__.str>',
                                        'longname': 'category',
                                        'modes': ['create',
                                                  'query',
                                                  'edit',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'cat'},
                           'classification': {'args': '<type __builtin__.str>',
                                              'longname': 'classification',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cls'},
                           'clearByBin': {'args': '<type __builtin__.bool>',
                                          'longname': 'clearByBin',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'cbk'},
                           'clearByType': {'args': '<type __builtin__.bool>',
                                           'longname': 'clearByType',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'cbt'},
                           'difference': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'difference',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 2,
                                          'shortname': 'di'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'intersect': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'intersect',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 2,
                                         'shortname': 'intersect'},
                           'listBuiltInFilters': {'args': '<type __builtin__.bool>',
                                                  'longname': 'listBuiltInFilters',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'lbf'},
                           'listOtherFilters': {'args': '<type __builtin__.bool>',
                                                'longname': 'listOtherFilters',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'lof'},
                           'listUserFilters': {'args': '<type __builtin__.bool>',
                                               'longname': 'listUserFilters',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'luf'},
                           'negate': {'args': '<type __builtin__.bool>',
                                      'longname': 'negate',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'neg'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'pythonModule': {'args': '<type __builtin__.str>',
                                            'longname': 'pythonModule',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pym'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'secondScript': {'args': '<type __builtin__.str>',
                                            'longname': 'secondScript',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ss'},
                           'text': {'args': '<type __builtin__.str>',
                                    'longname': 'text',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 't'},
                           'union': {'args': ['<type __builtin__.str>',
                                              '<type __builtin__.str>'],
                                     'longname': 'union',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 2,
                                     'shortname': 'un'},
                           'uniqueNodeNames': {'args': '<type __builtin__.bool>',
                                               'longname': 'uniqueNodeNames',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'unn'}},
                 'removedFlags': {'in': 'intersect'},
                 'shortFlags': {'bk': 'byBin',
                                'bn': 'byName',
                                'bs': 'byScript',
                                'bt': 'byType',
                                'cat': 'category',
                                'cbk': 'clearByBin',
                                'cbt': 'clearByType',
                                'cls': 'classification',
                                'di': 'difference',
                                'e': 'edit',
                                'ex': 'exists',
                                'intersect': 'intersect',
                                'lbf': 'listBuiltInFilters',
                                'lof': 'listOtherFilters',
                                'luf': 'listUserFilters',
                                'neg': 'negate',
                                'p': 'parent',
                                'pym': 'pythonModule',
                                'q': 'query',
                                'ss': 'secondScript',
                                't': 'text',
                                'un': 'union',
                                'unn': 'uniqueNodeNames'},
                 'type': 'language'},
  'itemFilterAttr': {'flags': {'byName': {'args': '<type __builtin__.str>',
                                          'longname': 'byName',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'bn'},
                               'byNameString': {'args': '<type __builtin__.str>',
                                                'longname': 'byNameString',
                                                'modes': ['create',
                                                          'query',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'bns'},
                               'byScript': {'args': '<type __builtin__.str>',
                                            'longname': 'byScript',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'bs'},
                               'classification': {'args': '<type __builtin__.str>',
                                                  'longname': 'classification',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cls'},
                               'dynamic': {'args': '<type __builtin__.bool>',
                                           'longname': 'dynamic',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dy'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'hasCurve': {'args': '<type __builtin__.bool>',
                                            'longname': 'hasCurve',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'hc'},
                               'hasDrivenKey': {'args': '<type __builtin__.bool>',
                                                'longname': 'hasDrivenKey',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'hdk'},
                               'hasExpression': {'args': '<type __builtin__.bool>',
                                                 'longname': 'hasExpression',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'he'},
                               'hidden': {'args': '<type __builtin__.bool>',
                                          'longname': 'hidden',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'h'},
                               'intersect': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'intersect',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 2,
                                             'shortname': 'intersect'},
                               'keyable': {'args': '<type __builtin__.bool>',
                                           'longname': 'keyable',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'k'},
                               'listBuiltInFilters': {'args': '<type __builtin__.bool>',
                                                      'longname': 'listBuiltInFilters',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'lbf'},
                               'listOtherFilters': {'args': '<type __builtin__.bool>',
                                                    'longname': 'listOtherFilters',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'lof'},
                               'listUserFilters': {'args': '<type __builtin__.bool>',
                                                   'longname': 'listUserFilters',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'luf'},
                               'negate': {'args': '<type __builtin__.bool>',
                                          'longname': 'negate',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'neg'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': [],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'published': {'args': '<type __builtin__.bool>',
                                             'longname': 'published',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'pub'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'readable': {'args': '<type __builtin__.bool>',
                                            'longname': 'readable',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'r'},
                               'scaleRotateTranslate': {'args': '<type __builtin__.bool>',
                                                        'longname': 'scaleRotateTranslate',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'srt'},
                               'secondScript': {'args': '<type __builtin__.str>',
                                                'longname': 'secondScript',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ss'},
                               'text': {'args': '<type __builtin__.str>',
                                        'longname': 'text',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 't'},
                               'union': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'union',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 2,
                                         'shortname': 'un'},
                               'writable': {'args': '<type __builtin__.bool>',
                                            'longname': 'writable',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'w'}},
                     'removedFlags': {'in': 'intersect'},
                     'shortFlags': {'bn': 'byName',
                                    'bns': 'byNameString',
                                    'bs': 'byScript',
                                    'cls': 'classification',
                                    'dy': 'dynamic',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'h': 'hidden',
                                    'hc': 'hasCurve',
                                    'hdk': 'hasDrivenKey',
                                    'he': 'hasExpression',
                                    'intersect': 'intersect',
                                    'k': 'keyable',
                                    'lbf': 'listBuiltInFilters',
                                    'lof': 'listOtherFilters',
                                    'luf': 'listUserFilters',
                                    'neg': 'negate',
                                    'p': 'parent',
                                    'pub': 'published',
                                    'q': 'query',
                                    'r': 'readable',
                                    'srt': 'scaleRotateTranslate',
                                    'ss': 'secondScript',
                                    't': 'text',
                                    'un': 'union',
                                    'w': 'writable'},
                     'type': 'language'},
  'itemFilterRender': {'flags': {'anyTextures': {'args': '<type __builtin__.bool>',
                                                 'longname': 'anyTextures',
                                                 'numArgs': 1,
                                                 'shortname': 'at'},
                                 'category': {'args': '<type __builtin__.str>',
                                              'longname': 'category',
                                              'modes': ['multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'cat'},
                                 'classification': {'args': '<type __builtin__.str>',
                                                    'longname': 'classification',
                                                    'numArgs': 1,
                                                    'shortname': 'cls'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'exclusiveLights': {'args': '<type __builtin__.bool>',
                                                     'longname': 'exclusiveLights',
                                                     'numArgs': 1,
                                                     'shortname': 'exl'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'lightSets': {'args': '<type __builtin__.bool>',
                                               'longname': 'lightSets',
                                               'numArgs': 1,
                                               'shortname': 'ls'},
                                 'lights': {'args': '<type __builtin__.bool>',
                                            'longname': 'lights',
                                            'numArgs': 1,
                                            'shortname': 'l'},
                                 'linkedLights': {'args': '<type __builtin__.bool>',
                                                  'longname': 'linkedLights',
                                                  'numArgs': 1,
                                                  'shortname': 'll'},
                                 'listBuiltInFilters': {'args': '<type __builtin__.bool>',
                                                        'longname': 'listBuiltInFilters',
                                                        'numArgs': 0,
                                                        'shortname': 'lbf'},
                                 'listOtherFilters': {'args': '<type __builtin__.bool>',
                                                      'longname': 'listOtherFilters',
                                                      'numArgs': 0,
                                                      'shortname': 'lof'},
                                 'listUserFilters': {'args': '<type __builtin__.bool>',
                                                     'longname': 'listUserFilters',
                                                     'numArgs': 0,
                                                     'shortname': 'luf'},
                                 'negate': {'args': '<type __builtin__.bool>',
                                            'longname': 'negate',
                                            'numArgs': 1,
                                            'shortname': 'neg'},
                                 'nodeClassification': {'args': '<type __builtin__.str>',
                                                        'longname': 'nodeClassification',
                                                        'modes': ['multiuse'],
                                                        'numArgs': 1,
                                                        'shortname': 'nc'},
                                 'nonExclusiveLights': {'args': '<type __builtin__.bool>',
                                                        'longname': 'nonExclusiveLights',
                                                        'numArgs': 1,
                                                        'shortname': 'nxl'},
                                 'nonIlluminatingLights': {'args': '<type __builtin__.bool>',
                                                           'longname': 'nonIlluminatingLights',
                                                           'numArgs': 1,
                                                           'shortname': 'nil'},
                                 'parent': {'args': '<type __builtin__.str>',
                                            'longname': 'parent',
                                            'numArgs': 1,
                                            'shortname': 'p'},
                                 'postProcess': {'args': '<type __builtin__.bool>',
                                                 'longname': 'postProcess',
                                                 'numArgs': 1,
                                                 'shortname': 'pp'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'renderUtilityNode': {'args': '<type __builtin__.bool>',
                                                       'longname': 'renderUtilityNode',
                                                       'numArgs': 1,
                                                       'shortname': 'run'},
                                 'renderableObjectSets': {'args': '<type __builtin__.bool>',
                                                          'longname': 'renderableObjectSets',
                                                          'numArgs': 1,
                                                          'shortname': 'ros'},
                                 'renderingNode': {'args': '<type __builtin__.bool>',
                                                   'longname': 'renderingNode',
                                                   'numArgs': 1,
                                                   'shortname': 'rn'},
                                 'shaders': {'args': '<type __builtin__.bool>',
                                             'longname': 'shaders',
                                             'numArgs': 1,
                                             'shortname': 's'},
                                 'text': {'args': '<type __builtin__.str>',
                                          'longname': 'text',
                                          'numArgs': 1,
                                          'shortname': 't'},
                                 'textures2d': {'args': '<type __builtin__.bool>',
                                                'longname': 'textures2d',
                                                'numArgs': 1,
                                                'shortname': 't2d'},
                                 'textures3d': {'args': '<type __builtin__.bool>',
                                                'longname': 'textures3d',
                                                'numArgs': 1,
                                                'shortname': 't3d'},
                                 'texturesProcedural': {'args': '<type __builtin__.bool>',
                                                        'longname': 'texturesProcedural',
                                                        'numArgs': 1,
                                                        'shortname': 'tp'}},
                       'shortFlags': {'at': 'anyTextures',
                                      'cat': 'category',
                                      'cls': 'classification',
                                      'e': 'edit',
                                      'ex': 'exists',
                                      'exl': 'exclusiveLights',
                                      'l': 'lights',
                                      'lbf': 'listBuiltInFilters',
                                      'll': 'linkedLights',
                                      'lof': 'listOtherFilters',
                                      'ls': 'lightSets',
                                      'luf': 'listUserFilters',
                                      'nc': 'nodeClassification',
                                      'neg': 'negate',
                                      'nil': 'nonIlluminatingLights',
                                      'nxl': 'nonExclusiveLights',
                                      'p': 'parent',
                                      'pp': 'postProcess',
                                      'q': 'query',
                                      'rn': 'renderingNode',
                                      'ros': 'renderableObjectSets',
                                      'run': 'renderUtilityNode',
                                      's': 'shaders',
                                      't': 'text',
                                      't2d': 'textures2d',
                                      't3d': 'textures3d',
                                      'tp': 'texturesProcedural'},
                       'type': 'other'},
  'itemFilterType': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'text': {'args': '<type __builtin__.str>',
                                        'longname': 'text',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 't'},
                               'type': {'args': '<type __builtin__.bool>',
                                        'longname': 'type',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'typ'}},
                     'shortFlags': {'e': 'edit',
                                    'q': 'query',
                                    't': 'text',
                                    'typ': 'type'},
                     'type': 'language'},
  'iterOnNurbs': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'joint': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                   'longname': 'absolute',
                                   'modes': ['create', 'query', 'edit'],
                                   'modified': ['position'],
                                   'numArgs': 0,
                                   'shortname': 'a'},
                      'angleX': {'args': '<type __builtin__.float>',
                                 'longname': 'angleX',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'resultNeedsUnpacking': True,
                                 'shortname': 'ax'},
                      'angleY': {'args': '<type __builtin__.float>',
                                 'longname': 'angleY',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'resultNeedsUnpacking': True,
                                 'shortname': 'ay'},
                      'angleZ': {'args': '<type __builtin__.float>',
                                 'longname': 'angleZ',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'resultNeedsUnpacking': True,
                                 'shortname': 'az'},
                      'assumePreferredAngles': {'args': '<type __builtin__.bool>',
                                                'longname': 'assumePreferredAngles',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'apa'},
                      'automaticLimits': {'args': '<type __builtin__.bool>',
                                          'longname': 'automaticLimits',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'al'},
                      'children': {'args': '<type __builtin__.bool>',
                                   'longname': 'children',
                                   'modes': ['edit'],
                                   'numArgs': 0,
                                   'shortname': 'ch'},
                      'component': {'args': '<type __builtin__.bool>',
                                    'longname': 'component',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'co'},
                      'degreeOfFreedom': {'args': '<type __builtin__.str>',
                                          'longname': 'degreeOfFreedom',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dof'},
                      'edit': {'args': '<type __builtin__.bool>',
                               'longname': 'edit',
                               'numArgs': 0,
                               'shortname': 'e'},
                      'exists': {'args': '<type __builtin__.str>',
                                 'longname': 'exists',
                                 'modes': ['query'],
                                 'numArgs': 1,
                                 'shortname': 'ex'},
                      'limitSwitchX': {'args': '<type __builtin__.bool>',
                                       'longname': 'limitSwitchX',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'lsx'},
                      'limitSwitchY': {'args': '<type __builtin__.bool>',
                                       'longname': 'limitSwitchY',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'lsy'},
                      'limitSwitchZ': {'args': '<type __builtin__.bool>',
                                       'longname': 'limitSwitchZ',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'lsz'},
                      'limitX': {'args': ['<type __builtin__.float>',
                                          '<type __builtin__.float>'],
                                 'longname': 'limitX',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 2,
                                 'shortname': 'lx'},
                      'limitY': {'args': ['<type __builtin__.float>',
                                          '<type __builtin__.float>'],
                                 'longname': 'limitY',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 2,
                                 'shortname': 'ly'},
                      'limitZ': {'args': ['<type __builtin__.float>',
                                          '<type __builtin__.float>'],
                                 'longname': 'limitZ',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 2,
                                 'shortname': 'lz'},
                      'name': {'args': '<type __builtin__.str>',
                               'longname': 'name',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'n'},
                      'orientJoint': {'args': '<type __builtin__.str>',
                                      'longname': 'orientJoint',
                                      'modes': ['edit'],
                                      'numArgs': 1,
                                      'shortname': 'oj'},
                      'orientation': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'orientation',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 3,
                                      'shortname': 'o'},
                      'position': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'position',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 3,
                                   'secondaryFlags': ['absolute',
                                                      'relative'],
                                   'shortname': 'p'},
                      'query': {'args': '<type __builtin__.bool>',
                                'longname': 'query',
                                'numArgs': 0,
                                'shortname': 'q'},
                      'radius': {'args': '<type __builtin__.float>',
                                 'longname': 'radius',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'resultNeedsUnpacking': True,
                                 'shortname': 'rad'},
                      'relative': {'args': '<type __builtin__.bool>',
                                   'longname': 'relative',
                                   'modes': ['create', 'query', 'edit'],
                                   'modified': ['position'],
                                   'numArgs': 0,
                                   'shortname': 'r'},
                      'rotationOrder': {'args': '<type __builtin__.str>',
                                        'longname': 'rotationOrder',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'roo'},
                      'scale': {'args': ['<type __builtin__.float>',
                                         '<type __builtin__.float>',
                                         '<type __builtin__.float>'],
                                'longname': 'scale',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 3,
                                'shortname': 's'},
                      'scaleCompensate': {'args': '<type __builtin__.bool>',
                                          'longname': 'scaleCompensate',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sc'},
                      'scaleOrientation': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'scaleOrientation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'so'},
                      'secondaryAxisOrient': {'args': '<type __builtin__.str>',
                                              'longname': 'secondaryAxisOrient',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'sao'},
                      'setPreferredAngles': {'args': '<type __builtin__.bool>',
                                             'longname': 'setPreferredAngles',
                                             'modes': ['edit'],
                                             'numArgs': 0,
                                             'shortname': 'spa'},
                      'stiffnessX': {'args': '<type __builtin__.float>',
                                     'longname': 'stiffnessX',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsUnpacking': True,
                                     'shortname': 'stx'},
                      'stiffnessY': {'args': '<type __builtin__.float>',
                                     'longname': 'stiffnessY',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsUnpacking': True,
                                     'shortname': 'sty'},
                      'stiffnessZ': {'args': '<type __builtin__.float>',
                                     'longname': 'stiffnessZ',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsUnpacking': True,
                                     'shortname': 'stz'},
                      'symmetry': {'args': '<type __builtin__.bool>',
                                   'longname': 'symmetry',
                                   'modes': ['create', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'sym'},
                      'symmetryAxis': {'args': '<type __builtin__.str>',
                                       'longname': 'symmetryAxis',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sa'},
                      'zeroScaleOrient': {'args': '<type __builtin__.bool>',
                                          'longname': 'zeroScaleOrient',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'zso'}},
            'shortFlags': {'a': 'absolute',
                           'al': 'automaticLimits',
                           'apa': 'assumePreferredAngles',
                           'ax': 'angleX',
                           'ay': 'angleY',
                           'az': 'angleZ',
                           'ch': 'children',
                           'co': 'component',
                           'dof': 'degreeOfFreedom',
                           'e': 'edit',
                           'ex': 'exists',
                           'lsx': 'limitSwitchX',
                           'lsy': 'limitSwitchY',
                           'lsz': 'limitSwitchZ',
                           'lx': 'limitX',
                           'ly': 'limitY',
                           'lz': 'limitZ',
                           'n': 'name',
                           'o': 'orientation',
                           'oj': 'orientJoint',
                           'p': 'position',
                           'q': 'query',
                           'r': 'relative',
                           'rad': 'radius',
                           'roo': 'rotationOrder',
                           's': 'scale',
                           'sa': 'symmetryAxis',
                           'sao': 'secondaryAxisOrient',
                           'sc': 'scaleCompensate',
                           'so': 'scaleOrientation',
                           'spa': 'setPreferredAngles',
                           'stx': 'stiffnessX',
                           'sty': 'stiffnessY',
                           'stz': 'stiffnessZ',
                           'sym': 'symmetry',
                           'zso': 'zeroScaleOrient'},
            'type': 'animation'},
  'jointCluster': {'flags': {'aboveBound': {'args': '<type __builtin__.float>',
                                            'longname': 'aboveBound',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ab'},
                             'aboveCluster': {'args': '<type __builtin__.bool>',
                                              'longname': 'aboveCluster',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'ac'},
                             'aboveDropoffType': {'args': '<type __builtin__.str>',
                                                  'longname': 'aboveDropoffType',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'adt'},
                             'aboveValue': {'args': '<type __builtin__.float>',
                                            'longname': 'aboveValue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'av'},
                             'belowBound': {'args': '<type __builtin__.float>',
                                            'longname': 'belowBound',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'bb'},
                             'belowCluster': {'args': '<type __builtin__.bool>',
                                              'longname': 'belowCluster',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'bc'},
                             'belowDropoffType': {'args': '<type __builtin__.str>',
                                                  'longname': 'belowDropoffType',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'bdt'},
                             'belowValue': {'args': '<type __builtin__.float>',
                                            'longname': 'belowValue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'bv'},
                             'deformerTools': {'args': '<type __builtin__.bool>',
                                               'longname': 'deformerTools',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'dt'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'joint': {'args': '<type __builtin__.str>',
                                       'longname': 'joint',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'j'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'ab': 'aboveBound',
                                  'ac': 'aboveCluster',
                                  'adt': 'aboveDropoffType',
                                  'av': 'aboveValue',
                                  'bb': 'belowBound',
                                  'bc': 'belowCluster',
                                  'bdt': 'belowDropoffType',
                                  'bv': 'belowValue',
                                  'dt': 'deformerTools',
                                  'e': 'edit',
                                  'j': 'joint',
                                  'n': 'name',
                                  'q': 'query'},
                   'type': 'animation'},
  'jointCtx': {'flags': {'autoJointOrient': {'args': '<type __builtin__.str>',
                                             'longname': 'autoJointOrient',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ajo'},
                         'autoPriorityH': {'args': '<type __builtin__.bool>',
                                           'longname': 'autoPriorityH',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'apH'},
                         'createIKHandle': {'args': '<type __builtin__.bool>',
                                            'longname': 'createIKHandle',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ikh'},
                         'degreeOfFreedomJ': {'args': '<type __builtin__.str>',
                                              'longname': 'degreeOfFreedomJ',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dJ'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'forceSolverH': {'args': '<type __builtin__.bool>',
                                          'longname': 'forceSolverH',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'fsH'},
                         'image1': {'args': '<type __builtin__.str>',
                                    'longname': 'image1',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i1'},
                         'image2': {'args': '<type __builtin__.str>',
                                    'longname': 'image2',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i2'},
                         'image3': {'args': '<type __builtin__.str>',
                                    'longname': 'image3',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i3'},
                         'jointAutoLimits': {'args': '<type __builtin__.bool>',
                                             'longname': 'jointAutoLimits',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'jal'},
                         'jointOrientationJ': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'jointOrientationJ',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'joJ'},
                         'largeBoneLength': {'args': '<type __builtin__.float>',
                                             'longname': 'largeBoneLength',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lbl'},
                         'largeBoneRadius': {'args': '<type __builtin__.float>',
                                             'longname': 'largeBoneRadius',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lbr'},
                         'poWeightH': {'args': '<type __builtin__.float>',
                                       'longname': 'poWeightH',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'pwH'},
                         'priorityH': {'args': '<type __builtin__.int>',
                                       'longname': 'priorityH',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'pH'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'scaleCompensateJ': {'args': '<type __builtin__.bool>',
                                              'longname': 'scaleCompensateJ',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'scJ'},
                         'scaleJ': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'scaleJ',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 3,
                                    'shortname': 'sJ'},
                         'scaleOrientationJ': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'scaleOrientationJ',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'soJ'},
                         'secondaryAxisOrient': {'args': '<type __builtin__.str>',
                                                 'longname': 'secondaryAxisOrient',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sao'},
                         'smallBoneLength': {'args': '<type __builtin__.float>',
                                             'longname': 'smallBoneLength',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sbl'},
                         'smallBoneRadius': {'args': '<type __builtin__.float>',
                                             'longname': 'smallBoneRadius',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sbr'},
                         'snapHandleH': {'args': '<type __builtin__.bool>',
                                         'longname': 'snapHandleH',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'snH'},
                         'solverTypeH': {'args': '<type __builtin__.str>',
                                         'longname': 'solverTypeH',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'stH'},
                         'stickyH': {'args': '<type __builtin__.str>',
                                     'longname': 'stickyH',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'sH'},
                         'symmetry': {'args': '<type __builtin__.bool>',
                                      'longname': 'symmetry',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'sym'},
                         'symmetryAxis': {'args': '<type __builtin__.str>',
                                          'longname': 'symmetryAxis',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sa'},
                         'variableBoneSize': {'args': '<type __builtin__.bool>',
                                              'longname': 'variableBoneSize',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'vbs'},
                         'weightH': {'args': '<type __builtin__.float>',
                                     'longname': 'weightH',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'wH'}},
               'shortFlags': {'ajo': 'autoJointOrient',
                              'apH': 'autoPriorityH',
                              'dJ': 'degreeOfFreedomJ',
                              'e': 'edit',
                              'ex': 'exists',
                              'fsH': 'forceSolverH',
                              'i1': 'image1',
                              'i2': 'image2',
                              'i3': 'image3',
                              'ikh': 'createIKHandle',
                              'jal': 'jointAutoLimits',
                              'joJ': 'jointOrientationJ',
                              'lbl': 'largeBoneLength',
                              'lbr': 'largeBoneRadius',
                              'pH': 'priorityH',
                              'pwH': 'poWeightH',
                              'q': 'query',
                              'sH': 'stickyH',
                              'sJ': 'scaleJ',
                              'sa': 'symmetryAxis',
                              'sao': 'secondaryAxisOrient',
                              'sbl': 'smallBoneLength',
                              'sbr': 'smallBoneRadius',
                              'scJ': 'scaleCompensateJ',
                              'snH': 'snapHandleH',
                              'soJ': 'scaleOrientationJ',
                              'stH': 'solverTypeH',
                              'sym': 'symmetry',
                              'vbs': 'variableBoneSize',
                              'wH': 'weightH'},
               'type': 'context'},
  'jointDisplayScale': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                               'longname': 'absolute',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 0,
                                               'shortname': 'a'},
                                  'ikfk': {'args': '<type __builtin__.float>',
                                           'longname': 'ikfk',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ik'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'}},
                        'shortFlags': {'a': 'absolute',
                                       'ik': 'ikfk',
                                       'q': 'query'},
                        'type': 'animation'},
  'jointLattice': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                       'longname': 'after',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'af'},
                             'afterReference': {'args': '<type __builtin__.bool>',
                                                'longname': 'afterReference',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'ar'},
                             'before': {'args': '<type __builtin__.bool>',
                                        'longname': 'before',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'bf'},
                             'creasing': {'args': '<type __builtin__.float>',
                                          'longname': 'creasing',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cr'},
                             'deformerTools': {'args': '<type __builtin__.bool>',
                                               'longname': 'deformerTools',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'dt'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exclusive': {'args': '<type __builtin__.str>',
                                           'longname': 'exclusive',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'ex'},
                             'frontOfChain': {'args': '<type __builtin__.bool>',
                                              'longname': 'frontOfChain',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'foc'},
                             'geometry': {'args': '<type __builtin__.str>',
                                          'longname': 'geometry',
                                          'modes': ['query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'g'},
                             'geometryIndices': {'args': '<type __builtin__.bool>',
                                                 'longname': 'geometryIndices',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'gi'},
                             'ignoreSelected': {'args': '<type __builtin__.bool>',
                                                'longname': 'ignoreSelected',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ignoreSelected'},
                             'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                         'longname': 'includeHiddenSelections',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'ihs'},
                             'joint': {'args': '<type __builtin__.str>',
                                       'longname': 'joint',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'j'},
                             'lengthIn': {'args': '<type __builtin__.float>',
                                          'longname': 'lengthIn',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'li'},
                             'lengthOut': {'args': '<type __builtin__.float>',
                                           'longname': 'lengthOut',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'lo'},
                             'lowerBindSkin': {'args': '<type __builtin__.str>',
                                               'longname': 'lowerBindSkin',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'lb'},
                             'lowerTransform': {'args': '<type __builtin__.str>',
                                                'longname': 'lowerTransform',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'lt'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'parallel': {'args': '<type __builtin__.bool>',
                                          'longname': 'parallel',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'par'},
                             'prune': {'args': '<type __builtin__.bool>',
                                       'longname': 'prune',
                                       'modes': ['edit'],
                                       'numArgs': 0,
                                       'shortname': 'pr'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'remove': {'args': '<type __builtin__.bool>',
                                        'longname': 'remove',
                                        'modes': ['edit', 'multiuse'],
                                        'numArgs': 0,
                                        'shortname': 'rm'},
                             'rounding': {'args': '<type __builtin__.float>',
                                          'longname': 'rounding',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ro'},
                             'split': {'args': '<type __builtin__.bool>',
                                       'longname': 'split',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'sp'},
                             'upperBindSkin': {'args': '<type __builtin__.str>',
                                               'longname': 'upperBindSkin',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ub'},
                             'upperTransform': {'args': '<type __builtin__.str>',
                                                'longname': 'upperTransform',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ut'},
                             'widthLeft': {'args': '<type __builtin__.float>',
                                           'longname': 'widthLeft',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'wl'},
                             'widthRight': {'args': '<type __builtin__.float>',
                                            'longname': 'widthRight',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'wr'}},
                   'removedFlags': {'is': 'ignoreSelected'},
                   'shortFlags': {'af': 'after',
                                  'ar': 'afterReference',
                                  'bf': 'before',
                                  'cr': 'creasing',
                                  'dt': 'deformerTools',
                                  'e': 'edit',
                                  'ex': 'exclusive',
                                  'foc': 'frontOfChain',
                                  'g': 'geometry',
                                  'gi': 'geometryIndices',
                                  'ignoreSelected': 'ignoreSelected',
                                  'ihs': 'includeHiddenSelections',
                                  'j': 'joint',
                                  'lb': 'lowerBindSkin',
                                  'li': 'lengthIn',
                                  'lo': 'lengthOut',
                                  'lt': 'lowerTransform',
                                  'n': 'name',
                                  'par': 'parallel',
                                  'pr': 'prune',
                                  'q': 'query',
                                  'rm': 'remove',
                                  'ro': 'rounding',
                                  'sp': 'split',
                                  'ub': 'upperBindSkin',
                                  'ut': 'upperTransform',
                                  'wl': 'widthLeft',
                                  'wr': 'widthRight'},
                   'type': 'animation'},
  'journal': {'flags': {'comment': {'args': '<type __builtin__.str>',
                                    'longname': 'comment',
                                    'numArgs': 1,
                                    'shortname': 'c'},
                        'flush': {'args': '<type __builtin__.bool>',
                                  'longname': 'flush',
                                  'numArgs': 0,
                                  'shortname': 'fl'},
                        'highPrecision': {'args': '<type __builtin__.bool>',
                                          'longname': 'highPrecision',
                                          'numArgs': 1,
                                          'shortname': 'hp'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'state': {'args': '<type __builtin__.bool>',
                                  'longname': 'state',
                                  'numArgs': 1,
                                  'shortname': 'st'}},
              'shortFlags': {'c': 'comment',
                             'fl': 'flush',
                             'hp': 'highPrecision',
                             'q': 'query',
                             'st': 'state'},
              'type': 'other'},
  'keyTangent': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                        'longname': 'absolute',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'a'},
                           'animation': {'args': '<type __builtin__.str>',
                                         'longname': 'animation',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'an'},
                           'attribute': {'args': '<type __builtin__.str>',
                                         'longname': 'attribute',
                                         'modes': ['create', 'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'at'},
                           'controlPoints': {'args': '<type __builtin__.bool>',
                                             'longname': 'controlPoints',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'cp'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'float': {'args': 'floatrange',
                                     'longname': 'float',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'f'},
                           'g': {'args': '<type __builtin__.bool>',
                                 'longname': 'g',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'g'},
                           'hierarchy': {'args': '<type __builtin__.str>',
                                         'longname': 'hierarchy',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'hi'},
                           'inAngle': {'args': '<type __builtin__.float>',
                                       'longname': 'inAngle',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ia'},
                           'inTangentType': {'args': '<type __builtin__.str>',
                                             'longname': 'inTangentType',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'itt'},
                           'inWeight': {'args': '<type __builtin__.float>',
                                        'longname': 'inWeight',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'iw'},
                           'includeUpperBound': {'args': '<type __builtin__.bool>',
                                                 'longname': 'includeUpperBound',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'iub'},
                           'index': {'args': '<type __builtin__.int>',
                                     'longname': 'index',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'index'},
                           'ix': {'args': '<type __builtin__.float>',
                                  'longname': 'ix',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'ix'},
                           'iy': {'args': '<type __builtin__.float>',
                                  'longname': 'iy',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'iy'},
                           'lock': {'args': '<type __builtin__.bool>',
                                    'longname': 'lock',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'l'},
                           'outAngle': {'args': '<type __builtin__.float>',
                                        'longname': 'outAngle',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'oa'},
                           'outTangentType': {'args': '<type __builtin__.str>',
                                              'longname': 'outTangentType',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ott'},
                           'outWeight': {'args': '<type __builtin__.float>',
                                         'longname': 'outWeight',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ow'},
                           'ox': {'args': '<type __builtin__.float>',
                                  'longname': 'ox',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'ox'},
                           'oy': {'args': '<type __builtin__.float>',
                                  'longname': 'oy',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'oy'},
                           'pluginTangentTypes': {'args': '<type __builtin__.str>',
                                                  'longname': 'pluginTangentTypes',
                                                  'modes': ['query'],
                                                  'numArgs': 1,
                                                  'shortname': 'ptt'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'relative': {'args': '<type __builtin__.bool>',
                                        'longname': 'relative',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'r'},
                           'shape': {'args': '<type __builtin__.bool>',
                                     'longname': 'shape',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 's'},
                           'stepAttributes': {'args': '<type __builtin__.bool>',
                                              'longname': 'stepAttributes',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sa'},
                           'time': {'args': 'timerange',
                                    'longname': 'time',
                                    'modes': ['create', 'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 't'},
                           'unify': {'args': '<type __builtin__.bool>',
                                     'longname': 'unify',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'uni'},
                           'weightLock': {'args': '<type __builtin__.bool>',
                                          'longname': 'weightLock',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'wl'},
                           'weightedTangents': {'args': '<type __builtin__.bool>',
                                                'longname': 'weightedTangents',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'wt'}},
                 'removedFlags': {'global': 'g', 'in': 'index'},
                 'shortFlags': {'a': 'absolute',
                                'an': 'animation',
                                'at': 'attribute',
                                'cp': 'controlPoints',
                                'e': 'edit',
                                'f': 'float',
                                'g': 'g',
                                'hi': 'hierarchy',
                                'ia': 'inAngle',
                                'index': 'index',
                                'itt': 'inTangentType',
                                'iub': 'includeUpperBound',
                                'iw': 'inWeight',
                                'ix': 'ix',
                                'iy': 'iy',
                                'l': 'lock',
                                'oa': 'outAngle',
                                'ott': 'outTangentType',
                                'ow': 'outWeight',
                                'ox': 'ox',
                                'oy': 'oy',
                                'ptt': 'pluginTangentTypes',
                                'q': 'query',
                                'r': 'relative',
                                's': 'shape',
                                'sa': 'stepAttributes',
                                't': 'time',
                                'uni': 'unify',
                                'wl': 'weightLock',
                                'wt': 'weightedTangents'},
                 'type': 'animation'},
  'keyframe': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                      'longname': 'absolute',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'a'},
                         'adjustBreakdown': {'args': '<type __builtin__.bool>',
                                             'longname': 'adjustBreakdown',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'abd'},
                         'animation': {'args': '<type __builtin__.str>',
                                       'longname': 'animation',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'an'},
                         'attribute': {'args': '<type __builtin__.str>',
                                       'longname': 'attribute',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'at'},
                         'breakdown': {'args': '<type __builtin__.bool>',
                                       'longname': 'breakdown',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'bd'},
                         'clipTime': {'args': ['time',
                                               'time',
                                               '<type __builtin__.float>'],
                                      'longname': 'clipTime',
                                      'modes': ['create'],
                                      'numArgs': 3,
                                      'shortname': 'ct'},
                         'controlPoints': {'args': '<type __builtin__.bool>',
                                           'longname': 'controlPoints',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'cp'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'eval': {'args': '<type __builtin__.bool>',
                                  'longname': 'eval',
                                  'modes': ['create', 'query'],
                                  'numArgs': 0,
                                  'shortname': 'ev'},
                         'float': {'args': 'floatrange',
                                   'longname': 'float',
                                   'modes': ['create', 'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 'f'},
                         'floatChange': {'args': '<type __builtin__.float>',
                                         'longname': 'floatChange',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'fc'},
                         'hierarchy': {'args': '<type __builtin__.str>',
                                       'longname': 'hierarchy',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'hi'},
                         'includeUpperBound': {'args': '<type __builtin__.bool>',
                                               'longname': 'includeUpperBound',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'iub'},
                         'index': {'args': '<type __builtin__.int>',
                                   'longname': 'index',
                                   'modes': ['create', 'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 'index'},
                         'indexValue': {'args': '<type __builtin__.bool>',
                                        'longname': 'indexValue',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'iv'},
                         'keyframeCount': {'args': '<type __builtin__.bool>',
                                           'longname': 'keyframeCount',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'kc'},
                         'lastSelected': {'args': '<type __builtin__.bool>',
                                          'longname': 'lastSelected',
                                          'modes': ['create', 'query'],
                                          'numArgs': 0,
                                          'shortname': 'lsl'},
                         'name': {'args': '<type __builtin__.bool>',
                                  'longname': 'name',
                                  'modes': ['create', 'query'],
                                  'numArgs': 0,
                                  'shortname': 'n'},
                         'option': {'args': '<type __builtin__.str>',
                                    'longname': 'option',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'o'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'relative': {'args': '<type __builtin__.bool>',
                                      'longname': 'relative',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'r'},
                         'selected': {'args': '<type __builtin__.bool>',
                                      'longname': 'selected',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'sl'},
                         'shape': {'args': '<type __builtin__.bool>',
                                   'longname': 'shape',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 's'},
                         'tickDrawSpecial': {'args': '<type __builtin__.bool>',
                                             'longname': 'tickDrawSpecial',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'tds'},
                         'time': {'args': 'timerange',
                                  'longname': 'time',
                                  'modes': ['create', 'multiuse'],
                                  'numArgs': 1,
                                  'shortname': 't'},
                         'timeChange': {'args': 'time',
                                        'longname': 'timeChange',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'tc'},
                         'valueChange': {'args': '<type __builtin__.float>',
                                         'longname': 'valueChange',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vc'}},
               'removedFlags': {'in': 'index'},
               'shortFlags': {'a': 'absolute',
                              'abd': 'adjustBreakdown',
                              'an': 'animation',
                              'at': 'attribute',
                              'bd': 'breakdown',
                              'cp': 'controlPoints',
                              'ct': 'clipTime',
                              'e': 'edit',
                              'ev': 'eval',
                              'f': 'float',
                              'fc': 'floatChange',
                              'hi': 'hierarchy',
                              'index': 'index',
                              'iub': 'includeUpperBound',
                              'iv': 'indexValue',
                              'kc': 'keyframeCount',
                              'lsl': 'lastSelected',
                              'n': 'name',
                              'o': 'option',
                              'q': 'query',
                              'r': 'relative',
                              's': 'shape',
                              'sl': 'selected',
                              't': 'time',
                              'tc': 'timeChange',
                              'tds': 'tickDrawSpecial',
                              'vc': 'valueChange'},
               'type': 'animation'},
  'keyframeOutliner': {'callbackFlags': ['visibleChangeCommand',
                                         'vcc',
                                         'dragCallback',
                                         'dgc',
                                         'dropCallback',
                                         'dpc'],
                       'flags': {'animCurve': {'args': '<type __builtin__.str>',
                                               'longname': 'animCurve',
                                               'modes': ['edit'],
                                               'numArgs': 1,
                                               'shortname': 'ac'},
                                 'annotation': {'args': '<type __builtin__.str>',
                                                'longname': 'annotation',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ann'},
                                 'backgroundColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'backgroundColor',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'bgc'},
                                 'defineTemplate': {'args': '<type __builtin__.str>',
                                                    'longname': 'defineTemplate',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'dt'},
                                 'display': {'args': '<type __builtin__.str>',
                                             'longname': 'display',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dsp'},
                                 'docTag': {'args': '<type __builtin__.str>',
                                            'longname': 'docTag',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dtg'},
                                 'dragCallback': {'args': 'script',
                                                  'longname': 'dragCallback',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dgc'},
                                 'dropCallback': {'args': 'script',
                                                  'longname': 'dropCallback',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dpc'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'enable': {'args': '<type __builtin__.bool>',
                                            'longname': 'enable',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'en'},
                                 'enableBackground': {'args': '<type __builtin__.bool>',
                                                      'longname': 'enableBackground',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ebg'},
                                 'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                         'longname': 'enableKeyboardFocus',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ekf'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'fullPathName': {'args': '<type __builtin__.bool>',
                                                  'longname': 'fullPathName',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'fpn'},
                                 'height': {'args': '<type __builtin__.int>',
                                            'longname': 'height',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'h'},
                                 'highlightColor': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'highlightColor',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'hlc'},
                                 'isObscured': {'args': '<type __builtin__.bool>',
                                                'longname': 'isObscured',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'io'},
                                 'manage': {'args': '<type __builtin__.bool>',
                                            'longname': 'manage',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'm'},
                                 'noBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'noBackground',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'nbg'},
                                 'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                        'longname': 'numberOfPopupMenus',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'npm'},
                                 'parent': {'args': '<type __builtin__.str>',
                                            'longname': 'parent',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'p'},
                                 'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                    'longname': 'popupMenuArray',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'pma'},
                                 'preventOverride': {'args': '<type __builtin__.bool>',
                                                     'longname': 'preventOverride',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'po'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'statusBarMessage': {'args': '<type __builtin__.str>',
                                                      'longname': 'statusBarMessage',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'sbm'},
                                 'useTemplate': {'args': '<type __builtin__.str>',
                                                 'longname': 'useTemplate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ut'},
                                 'visible': {'args': '<type __builtin__.bool>',
                                             'longname': 'visible',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'vis'},
                                 'visibleChangeCommand': {'args': 'script',
                                                          'longname': 'visibleChangeCommand',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'vcc'},
                                 'width': {'args': '<type __builtin__.int>',
                                           'longname': 'width',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'w'}},
                       'shortFlags': {'ac': 'animCurve',
                                      'ann': 'annotation',
                                      'bgc': 'backgroundColor',
                                      'dgc': 'dragCallback',
                                      'dpc': 'dropCallback',
                                      'dsp': 'display',
                                      'dt': 'defineTemplate',
                                      'dtg': 'docTag',
                                      'e': 'edit',
                                      'ebg': 'enableBackground',
                                      'ekf': 'enableKeyboardFocus',
                                      'en': 'enable',
                                      'ex': 'exists',
                                      'fpn': 'fullPathName',
                                      'h': 'height',
                                      'hlc': 'highlightColor',
                                      'io': 'isObscured',
                                      'm': 'manage',
                                      'nbg': 'noBackground',
                                      'npm': 'numberOfPopupMenus',
                                      'p': 'parent',
                                      'pma': 'popupMenuArray',
                                      'po': 'preventOverride',
                                      'q': 'query',
                                      'sbm': 'statusBarMessage',
                                      'ut': 'useTemplate',
                                      'vcc': 'visibleChangeCommand',
                                      'vis': 'visible',
                                      'w': 'width'},
                       'type': 'animation'},
  'keyframeRegionCurrentTimeCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                      'longname': 'edit',
                                                      'numArgs': 0,
                                                      'shortname': 'e'},
                                             'exists': {'args': '<type __builtin__.bool>',
                                                        'longname': 'exists',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'ex'},
                                             'history': {'args': '<type __builtin__.bool>',
                                                         'longname': 'history',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'ch'},
                                             'image1': {'args': '<type __builtin__.str>',
                                                        'longname': 'image1',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'i1'},
                                             'image2': {'args': '<type __builtin__.str>',
                                                        'longname': 'image2',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'i2'},
                                             'image3': {'args': '<type __builtin__.str>',
                                                        'longname': 'image3',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'i3'},
                                             'name': {'args': '<type __builtin__.str>',
                                                      'longname': 'name',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'n'},
                                             'query': {'args': '<type __builtin__.bool>',
                                                       'longname': 'query',
                                                       'numArgs': 0,
                                                       'shortname': 'q'}},
                                   'shortFlags': {'ch': 'history',
                                                  'e': 'edit',
                                                  'ex': 'exists',
                                                  'i1': 'image1',
                                                  'i2': 'image2',
                                                  'i3': 'image3',
                                                  'n': 'name',
                                                  'q': 'query'},
                                   'type': 'context'},
  'keyframeRegionDirectKeyCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                    'longname': 'edit',
                                                    'numArgs': 0,
                                                    'shortname': 'e'},
                                           'exists': {'args': '<type __builtin__.bool>',
                                                      'longname': 'exists',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'ex'},
                                           'history': {'args': '<type __builtin__.bool>',
                                                       'longname': 'history',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                                           'image1': {'args': '<type __builtin__.str>',
                                                      'longname': 'image1',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'i1'},
                                           'image2': {'args': '<type __builtin__.str>',
                                                      'longname': 'image2',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'i2'},
                                           'image3': {'args': '<type __builtin__.str>',
                                                      'longname': 'image3',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'i3'},
                                           'name': {'args': '<type __builtin__.str>',
                                                    'longname': 'name',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'n'},
                                           'option': {'args': '<type __builtin__.str>',
                                                      'longname': 'option',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'o'},
                                           'query': {'args': '<type __builtin__.bool>',
                                                     'longname': 'query',
                                                     'numArgs': 0,
                                                     'shortname': 'q'}},
                                 'shortFlags': {'ch': 'history',
                                                'e': 'edit',
                                                'ex': 'exists',
                                                'i1': 'image1',
                                                'i2': 'image2',
                                                'i3': 'image3',
                                                'n': 'name',
                                                'o': 'option',
                                                'q': 'query'},
                                 'type': 'context'},
  'keyframeRegionDollyCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                'longname': 'edit',
                                                'numArgs': 0,
                                                'shortname': 'e'},
                                       'exists': {'args': '<type __builtin__.bool>',
                                                  'longname': 'exists',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'ex'},
                                       'history': {'args': '<type __builtin__.bool>',
                                                   'longname': 'history',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                                       'image1': {'args': '<type __builtin__.str>',
                                                  'longname': 'image1',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'i1'},
                                       'image2': {'args': '<type __builtin__.str>',
                                                  'longname': 'image2',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'i2'},
                                       'image3': {'args': '<type __builtin__.str>',
                                                  'longname': 'image3',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'i3'},
                                       'name': {'args': '<type __builtin__.str>',
                                                'longname': 'name',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'n'},
                                       'query': {'args': '<type __builtin__.bool>',
                                                 'longname': 'query',
                                                 'numArgs': 0,
                                                 'shortname': 'q'}},
                             'shortFlags': {'ch': 'history',
                                            'e': 'edit',
                                            'ex': 'exists',
                                            'i1': 'image1',
                                            'i2': 'image2',
                                            'i3': 'image3',
                                            'n': 'name',
                                            'q': 'query'},
                             'type': 'context'},
  'keyframeRegionInsertKeyCtx': {'flags': {'breakdown': {'args': '<type __builtin__.bool>',
                                                         'longname': 'breakdown',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'bd'},
                                           'edit': {'args': '<type __builtin__.bool>',
                                                    'longname': 'edit',
                                                    'numArgs': 0,
                                                    'shortname': 'e'},
                                           'exists': {'args': '<type __builtin__.bool>',
                                                      'longname': 'exists',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'ex'},
                                           'history': {'args': '<type __builtin__.bool>',
                                                       'longname': 'history',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                                           'image1': {'args': '<type __builtin__.str>',
                                                      'longname': 'image1',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'i1'},
                                           'image2': {'args': '<type __builtin__.str>',
                                                      'longname': 'image2',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'i2'},
                                           'image3': {'args': '<type __builtin__.str>',
                                                      'longname': 'image3',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'i3'},
                                           'name': {'args': '<type __builtin__.str>',
                                                    'longname': 'name',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'n'},
                                           'query': {'args': '<type __builtin__.bool>',
                                                     'longname': 'query',
                                                     'numArgs': 0,
                                                     'shortname': 'q'}},
                                 'shortFlags': {'bd': 'breakdown',
                                                'ch': 'history',
                                                'e': 'edit',
                                                'ex': 'exists',
                                                'i1': 'image1',
                                                'i2': 'image2',
                                                'i3': 'image3',
                                                'n': 'name',
                                                'q': 'query'},
                                 'type': 'context'},
  'keyframeRegionMoveKeyCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'edit',
                                                  'numArgs': 0,
                                                  'shortname': 'e'},
                                         'exists': {'args': '<type __builtin__.bool>',
                                                    'longname': 'exists',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ex'},
                                         'history': {'args': '<type __builtin__.bool>',
                                                     'longname': 'history',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                                         'image1': {'args': '<type __builtin__.str>',
                                                    'longname': 'image1',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'i1'},
                                         'image2': {'args': '<type __builtin__.str>',
                                                    'longname': 'image2',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'i2'},
                                         'image3': {'args': '<type __builtin__.str>',
                                                    'longname': 'image3',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'i3'},
                                         'name': {'args': '<type __builtin__.str>',
                                                  'longname': 'name',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'n'},
                                         'option': {'args': '<type __builtin__.str>',
                                                    'longname': 'option',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'o'},
                                         'query': {'args': '<type __builtin__.bool>',
                                                   'longname': 'query',
                                                   'numArgs': 0,
                                                   'shortname': 'q'}},
                               'shortFlags': {'ch': 'history',
                                              'e': 'edit',
                                              'ex': 'exists',
                                              'i1': 'image1',
                                              'i2': 'image2',
                                              'i3': 'image3',
                                              'n': 'name',
                                              'o': 'option',
                                              'q': 'query'},
                               'type': 'context'},
  'keyframeRegionScaleKeyCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                   'longname': 'edit',
                                                   'numArgs': 0,
                                                   'shortname': 'e'},
                                          'exists': {'args': '<type __builtin__.bool>',
                                                     'longname': 'exists',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'ex'},
                                          'history': {'args': '<type __builtin__.bool>',
                                                      'longname': 'history',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                                          'image1': {'args': '<type __builtin__.str>',
                                                     'longname': 'image1',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'i1'},
                                          'image2': {'args': '<type __builtin__.str>',
                                                     'longname': 'image2',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'i2'},
                                          'image3': {'args': '<type __builtin__.str>',
                                                     'longname': 'image3',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'i3'},
                                          'name': {'args': '<type __builtin__.str>',
                                                   'longname': 'name',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'n'},
                                          'query': {'args': '<type __builtin__.bool>',
                                                    'longname': 'query',
                                                    'numArgs': 0,
                                                    'shortname': 'q'},
                                          'scaleSpecifiedKeys': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'scaleSpecifiedKeys',
                                                                 'modes': ['query',
                                                                           'edit'],
                                                                 'numArgs': 1,
                                                                 'shortname': 'ssk'},
                                          'type': {'args': '<type __builtin__.str>',
                                                   'longname': 'type',
                                                   'modes': ['edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'typ'}},
                                'shortFlags': {'ch': 'history',
                                               'e': 'edit',
                                               'ex': 'exists',
                                               'i1': 'image1',
                                               'i2': 'image2',
                                               'i3': 'image3',
                                               'n': 'name',
                                               'q': 'query',
                                               'ssk': 'scaleSpecifiedKeys',
                                               'typ': 'type'},
                                'type': 'context'},
  'keyframeRegionSelectKeyCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                    'longname': 'edit',
                                                    'numArgs': 0,
                                                    'shortname': 'e'},
                                           'exists': {'args': '<type __builtin__.bool>',
                                                      'longname': 'exists',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'ex'},
                                           'history': {'args': '<type __builtin__.bool>',
                                                       'longname': 'history',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                                           'image1': {'args': '<type __builtin__.str>',
                                                      'longname': 'image1',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'i1'},
                                           'image2': {'args': '<type __builtin__.str>',
                                                      'longname': 'image2',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'i2'},
                                           'image3': {'args': '<type __builtin__.str>',
                                                      'longname': 'image3',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'i3'},
                                           'name': {'args': '<type __builtin__.str>',
                                                    'longname': 'name',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'n'},
                                           'query': {'args': '<type __builtin__.bool>',
                                                     'longname': 'query',
                                                     'numArgs': 0,
                                                     'shortname': 'q'}},
                                 'shortFlags': {'ch': 'history',
                                                'e': 'edit',
                                                'ex': 'exists',
                                                'i1': 'image1',
                                                'i2': 'image2',
                                                'i3': 'image3',
                                                'n': 'name',
                                                'q': 'query'},
                                 'type': 'context'},
  'keyframeRegionSetKeyCtx': {'flags': {'breakdown': {'args': '<type __builtin__.bool>',
                                                      'longname': 'breakdown',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'bd'},
                                        'edit': {'args': '<type __builtin__.bool>',
                                                 'longname': 'edit',
                                                 'numArgs': 0,
                                                 'shortname': 'e'},
                                        'exists': {'args': '<type __builtin__.bool>',
                                                   'longname': 'exists',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'ex'},
                                        'history': {'args': '<type __builtin__.bool>',
                                                    'longname': 'history',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                                        'image1': {'args': '<type __builtin__.str>',
                                                   'longname': 'image1',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'i1'},
                                        'image2': {'args': '<type __builtin__.str>',
                                                   'longname': 'image2',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'i2'},
                                        'image3': {'args': '<type __builtin__.str>',
                                                   'longname': 'image3',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'i3'},
                                        'name': {'args': '<type __builtin__.str>',
                                                 'longname': 'name',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'n'},
                                        'query': {'args': '<type __builtin__.bool>',
                                                  'longname': 'query',
                                                  'numArgs': 0,
                                                  'shortname': 'q'}},
                              'shortFlags': {'bd': 'breakdown',
                                             'ch': 'history',
                                             'e': 'edit',
                                             'ex': 'exists',
                                             'i1': 'image1',
                                             'i2': 'image2',
                                             'i3': 'image3',
                                             'n': 'name',
                                             'q': 'query'},
                              'type': 'context'},
  'keyframeRegionTrackCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                'longname': 'edit',
                                                'numArgs': 0,
                                                'shortname': 'e'},
                                       'exists': {'args': '<type __builtin__.bool>',
                                                  'longname': 'exists',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'ex'},
                                       'history': {'args': '<type __builtin__.bool>',
                                                   'longname': 'history',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                                       'image1': {'args': '<type __builtin__.str>',
                                                  'longname': 'image1',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'i1'},
                                       'image2': {'args': '<type __builtin__.str>',
                                                  'longname': 'image2',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'i2'},
                                       'image3': {'args': '<type __builtin__.str>',
                                                  'longname': 'image3',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'i3'},
                                       'name': {'args': '<type __builtin__.str>',
                                                'longname': 'name',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'n'},
                                       'query': {'args': '<type __builtin__.bool>',
                                                 'longname': 'query',
                                                 'numArgs': 0,
                                                 'shortname': 'q'}},
                             'shortFlags': {'ch': 'history',
                                            'e': 'edit',
                                            'ex': 'exists',
                                            'i1': 'image1',
                                            'i2': 'image2',
                                            'i3': 'image3',
                                            'n': 'name',
                                            'q': 'query'},
                             'type': 'context'},
  'keyframeStats': {'callbackFlags': ['visibleChangeCommand',
                                      'vcc',
                                      'dragCallback',
                                      'dgc',
                                      'dropCallback',
                                      'dpc'],
                    'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'adj'},
                              'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn2',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad2'},
                              'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn3',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad3'},
                              'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn4',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad4'},
                              'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn5',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad5'},
                              'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn6',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad6'},
                              'animEditor': {'args': '<type __builtin__.str>',
                                             'longname': 'animEditor',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ae'},
                              'annotation': {'args': '<type __builtin__.str>',
                                             'longname': 'annotation',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ann'},
                              'backgroundColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'backgroundColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'bgc'},
                              'classicMode': {'args': '<type __builtin__.bool>',
                                              'longname': 'classicMode',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cm'},
                              'columnAlign': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'cal'},
                              'columnAlign2': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'cl2'},
                              'columnAlign3': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'cl3'},
                              'columnAlign4': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'cl4'},
                              'columnAlign5': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'cl5'},
                              'columnAlign6': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'cl6'},
                              'columnAttach': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnAttach',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 3,
                                               'shortname': 'cat'},
                              'columnAttach2': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach2',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'ct2'},
                              'columnAttach3': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach3',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'ct3'},
                              'columnAttach4': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach4',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 4,
                                                'shortname': 'ct4'},
                              'columnAttach5': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach5',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 5,
                                                'shortname': 'ct5'},
                              'columnAttach6': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach6',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 6,
                                                'shortname': 'ct6'},
                              'columnOffset2': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset2',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'co2'},
                              'columnOffset3': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset3',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'co3'},
                              'columnOffset4': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset4',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 4,
                                                'shortname': 'co4'},
                              'columnOffset5': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset5',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 5,
                                                'shortname': 'co5'},
                              'columnOffset6': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset6',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 6,
                                                'shortname': 'co6'},
                              'columnWidth': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'cw'},
                              'columnWidth1': {'args': '<type __builtin__.int>',
                                               'longname': 'columnWidth1',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cw1'},
                              'columnWidth2': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'cw2'},
                              'columnWidth3': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'cw3'},
                              'columnWidth4': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'cw4'},
                              'columnWidth5': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'cw5'},
                              'columnWidth6': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'cw6'},
                              'defineTemplate': {'args': '<type __builtin__.str>',
                                                 'longname': 'defineTemplate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'dt'},
                              'docTag': {'args': '<type __builtin__.str>',
                                         'longname': 'docTag',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dtg'},
                              'dragCallback': {'args': 'script',
                                               'longname': 'dragCallback',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dgc'},
                              'dropCallback': {'args': 'script',
                                               'longname': 'dropCallback',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dpc'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'enable': {'args': '<type __builtin__.bool>',
                                         'longname': 'enable',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'en'},
                              'enableBackground': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableBackground',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ebg'},
                              'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'enableKeyboardFocus',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ekf'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'fullPathName': {'args': '<type __builtin__.bool>',
                                               'longname': 'fullPathName',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'fpn'},
                              'height': {'args': '<type __builtin__.int>',
                                         'longname': 'height',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'h'},
                              'highlightColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'highlightColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'hlc'},
                              'isObscured': {'args': '<type __builtin__.bool>',
                                             'longname': 'isObscured',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'io'},
                              'manage': {'args': '<type __builtin__.bool>',
                                         'longname': 'manage',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                              'noBackground': {'args': '<type __builtin__.bool>',
                                               'longname': 'noBackground',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nbg'},
                              'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'numberOfPopupMenus',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'npm'},
                              'parent': {'args': '<type __builtin__.str>',
                                         'longname': 'parent',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'p'},
                              'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                 'longname': 'popupMenuArray',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'pma'},
                              'precision': {'args': '<type __builtin__.int>',
                                            'longname': 'precision',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pre'},
                              'preventOverride': {'args': '<type __builtin__.bool>',
                                                  'longname': 'preventOverride',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'po'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'rowAttach': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'rowAttach',
                                            'modes': ['create',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 3,
                                            'shortname': 'rat'},
                              'statusBarMessage': {'args': '<type __builtin__.str>',
                                                   'longname': 'statusBarMessage',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sbm'},
                              'timeAnnotation': {'args': '<type __builtin__.str>',
                                                 'longname': 'timeAnnotation',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'tan'},
                              'useTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'useTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ut'},
                              'valueAnnotation': {'args': '<type __builtin__.str>',
                                                  'longname': 'valueAnnotation',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'van'},
                              'visible': {'args': '<type __builtin__.bool>',
                                          'longname': 'visible',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'vis'},
                              'visibleChangeCommand': {'args': 'script',
                                                       'longname': 'visibleChangeCommand',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'vcc'},
                              'width': {'args': '<type __builtin__.int>',
                                        'longname': 'width',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'w'}},
                    'shortFlags': {'ad2': 'adjustableColumn2',
                                   'ad3': 'adjustableColumn3',
                                   'ad4': 'adjustableColumn4',
                                   'ad5': 'adjustableColumn5',
                                   'ad6': 'adjustableColumn6',
                                   'adj': 'adjustableColumn',
                                   'ae': 'animEditor',
                                   'ann': 'annotation',
                                   'bgc': 'backgroundColor',
                                   'cal': 'columnAlign',
                                   'cat': 'columnAttach',
                                   'cl2': 'columnAlign2',
                                   'cl3': 'columnAlign3',
                                   'cl4': 'columnAlign4',
                                   'cl5': 'columnAlign5',
                                   'cl6': 'columnAlign6',
                                   'cm': 'classicMode',
                                   'co2': 'columnOffset2',
                                   'co3': 'columnOffset3',
                                   'co4': 'columnOffset4',
                                   'co5': 'columnOffset5',
                                   'co6': 'columnOffset6',
                                   'ct2': 'columnAttach2',
                                   'ct3': 'columnAttach3',
                                   'ct4': 'columnAttach4',
                                   'ct5': 'columnAttach5',
                                   'ct6': 'columnAttach6',
                                   'cw': 'columnWidth',
                                   'cw1': 'columnWidth1',
                                   'cw2': 'columnWidth2',
                                   'cw3': 'columnWidth3',
                                   'cw4': 'columnWidth4',
                                   'cw5': 'columnWidth5',
                                   'cw6': 'columnWidth6',
                                   'dgc': 'dragCallback',
                                   'dpc': 'dropCallback',
                                   'dt': 'defineTemplate',
                                   'dtg': 'docTag',
                                   'e': 'edit',
                                   'ebg': 'enableBackground',
                                   'ekf': 'enableKeyboardFocus',
                                   'en': 'enable',
                                   'ex': 'exists',
                                   'fpn': 'fullPathName',
                                   'h': 'height',
                                   'hlc': 'highlightColor',
                                   'io': 'isObscured',
                                   'm': 'manage',
                                   'nbg': 'noBackground',
                                   'npm': 'numberOfPopupMenus',
                                   'p': 'parent',
                                   'pma': 'popupMenuArray',
                                   'po': 'preventOverride',
                                   'pre': 'precision',
                                   'q': 'query',
                                   'rat': 'rowAttach',
                                   'sbm': 'statusBarMessage',
                                   'tan': 'timeAnnotation',
                                   'ut': 'useTemplate',
                                   'van': 'valueAnnotation',
                                   'vcc': 'visibleChangeCommand',
                                   'vis': 'visible',
                                   'w': 'width'},
                    'type': 'animation'},
  'keyingGroup': {'flags': {'activator': {'args': 'PyNode',
                                          'longname': 'activator',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'act'},
                            'addElement': {'args': 'PyNode',
                                           'longname': 'addElement',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'add'},
                            'afterFilters': {'args': '<type __builtin__.bool>',
                                             'longname': 'afterFilters',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'af'},
                            'category': {'args': '<type __builtin__.str>',
                                         'longname': 'category',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cat'},
                            'clear': {'args': 'PyNode',
                                      'longname': 'clear',
                                      'modes': ['edit'],
                                      'numArgs': 1,
                                      'shortname': 'cl'},
                            'color': {'args': '<type __builtin__.int>',
                                      'longname': 'color',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'co'},
                            'copy': {'args': 'PyNode',
                                     'longname': 'copy',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'cp'},
                            'edges': {'args': '<type __builtin__.bool>',
                                      'longname': 'edges',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'eg'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'editPoints': {'args': '<type __builtin__.bool>',
                                           'longname': 'editPoints',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'ep'},
                            'empty': {'args': '<type __builtin__.bool>',
                                      'longname': 'empty',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'em'},
                            'excludeDynamic': {'args': '<type __builtin__.bool>',
                                               'longname': 'excludeDynamic',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ed'},
                            'excludeRotate': {'args': '<type __builtin__.bool>',
                                              'longname': 'excludeRotate',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'er'},
                            'excludeScale': {'args': '<type __builtin__.bool>',
                                             'longname': 'excludeScale',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'es'},
                            'excludeTranslate': {'args': '<type __builtin__.bool>',
                                                 'longname': 'excludeTranslate',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'et'},
                            'excludeVisibility': {'args': '<type __builtin__.bool>',
                                                  'longname': 'excludeVisibility',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'ev'},
                            'facets': {'args': '<type __builtin__.bool>',
                                       'longname': 'facets',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'fc'},
                            'flatten': {'args': 'PyNode',
                                        'longname': 'flatten',
                                        'modes': ['edit'],
                                        'numArgs': 1,
                                        'shortname': 'fl'},
                            'forceElement': {'args': 'PyNode',
                                             'longname': 'forceElement',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'fe'},
                            'include': {'args': 'PyNode',
                                        'longname': 'include',
                                        'modes': ['edit'],
                                        'numArgs': 1,
                                        'shortname': 'include'},
                            'intersection': {'args': 'PyNode',
                                             'longname': 'intersection',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'int'},
                            'isIntersecting': {'args': 'PyNode',
                                               'longname': 'isIntersecting',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ii'},
                            'isMember': {'args': 'PyNode',
                                         'longname': 'isMember',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'im'},
                            'layer': {'args': '<type __builtin__.bool>',
                                      'longname': 'layer',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'l'},
                            'minimizeRotation': {'args': '<type __builtin__.bool>',
                                                 'longname': 'minimizeRotation',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'mr'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'noSurfaceShader': {'args': '<type __builtin__.bool>',
                                                'longname': 'noSurfaceShader',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'nss'},
                            'noWarnings': {'args': '<type __builtin__.bool>',
                                           'longname': 'noWarnings',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'nw'},
                            'nodesOnly': {'args': '<type __builtin__.bool>',
                                          'longname': 'nodesOnly',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'no'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'remove': {'args': 'PyNode',
                                       'longname': 'remove',
                                       'modes': ['edit'],
                                       'numArgs': 1,
                                       'shortname': 'rm'},
                            'removeActivator': {'args': 'PyNode',
                                                'longname': 'removeActivator',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'rac'},
                            'renderable': {'args': '<type __builtin__.bool>',
                                           'longname': 'renderable',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'r'},
                            'setActiveFilter': {'args': '<type __builtin__.str>',
                                                'longname': 'setActiveFilter',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'fil'},
                            'size': {'args': '<type __builtin__.bool>',
                                     'longname': 'size',
                                     'modes': ['query'],
                                     'numArgs': 0,
                                     'shortname': 's'},
                            'split': {'args': 'PyNode',
                                      'longname': 'split',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'sp'},
                            'subtract': {'args': 'PyNode',
                                         'longname': 'subtract',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'sub'},
                            'text': {'args': '<type __builtin__.str>',
                                     'longname': 'text',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 't'},
                            'union': {'args': 'PyNode',
                                      'longname': 'union',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'un'},
                            'vertices': {'args': '<type __builtin__.bool>',
                                         'longname': 'vertices',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'v'}},
                  'removedFlags': {'in': 'include'},
                  'shortFlags': {'act': 'activator',
                                 'add': 'addElement',
                                 'af': 'afterFilters',
                                 'cat': 'category',
                                 'cl': 'clear',
                                 'co': 'color',
                                 'cp': 'copy',
                                 'e': 'edit',
                                 'ed': 'excludeDynamic',
                                 'eg': 'edges',
                                 'em': 'empty',
                                 'ep': 'editPoints',
                                 'er': 'excludeRotate',
                                 'es': 'excludeScale',
                                 'et': 'excludeTranslate',
                                 'ev': 'excludeVisibility',
                                 'fc': 'facets',
                                 'fe': 'forceElement',
                                 'fil': 'setActiveFilter',
                                 'fl': 'flatten',
                                 'ii': 'isIntersecting',
                                 'im': 'isMember',
                                 'include': 'include',
                                 'int': 'intersection',
                                 'l': 'layer',
                                 'mr': 'minimizeRotation',
                                 'n': 'name',
                                 'no': 'nodesOnly',
                                 'nss': 'noSurfaceShader',
                                 'nw': 'noWarnings',
                                 'q': 'query',
                                 'r': 'renderable',
                                 'rac': 'removeActivator',
                                 'rm': 'remove',
                                 's': 'size',
                                 'sp': 'split',
                                 'sub': 'subtract',
                                 't': 'text',
                                 'un': 'union',
                                 'v': 'vertices'},
                  'type': 'animation'},
  'lassoContext': {'flags': {'drawClosed': {'args': '<type __builtin__.bool>',
                                            'longname': 'drawClosed',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fastComponents': {'args': '<type __builtin__.bool>',
                                                'longname': 'fastComponents',
                                                'numArgs': 1,
                                                'shortname': 'fc'},
                             'history': {'args': '<type __builtin__.bool>',
                                         'longname': 'history',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ch'},
                             'image1': {'args': '<type __builtin__.str>',
                                        'longname': 'image1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i1'},
                             'image2': {'args': '<type __builtin__.str>',
                                        'longname': 'image2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i2'},
                             'image3': {'args': '<type __builtin__.str>',
                                        'longname': 'image3',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i3'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'ch': 'history',
                                  'dc': 'drawClosed',
                                  'e': 'edit',
                                  'ex': 'exists',
                                  'fc': 'fastComponents',
                                  'i1': 'image1',
                                  'i2': 'image2',
                                  'i3': 'image3',
                                  'n': 'name',
                                  'q': 'query'},
                   'type': 'context'},
  'lattice': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                  'longname': 'after',
                                  'modes': ['create', 'edit'],
                                  'numArgs': 0,
                                  'shortname': 'af'},
                        'afterReference': {'args': '<type __builtin__.bool>',
                                           'longname': 'afterReference',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'ar'},
                        'before': {'args': '<type __builtin__.bool>',
                                   'longname': 'before',
                                   'modes': ['create', 'edit'],
                                   'numArgs': 0,
                                   'shortname': 'bf'},
                        'commonParent': {'args': '<type __builtin__.bool>',
                                         'longname': 'commonParent',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'cp'},
                        'deformerTools': {'args': '<type __builtin__.bool>',
                                          'longname': 'deformerTools',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'dt'},
                        'divisions': {'args': ['<type __builtin__.int>',
                                               '<type __builtin__.int>',
                                               '<type __builtin__.int>'],
                                      'longname': 'divisions',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 3,
                                      'shortname': 'dv'},
                        'dualBase': {'args': '<type __builtin__.bool>',
                                     'longname': 'dualBase',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'db'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'exclusive': {'args': '<type __builtin__.str>',
                                      'longname': 'exclusive',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'ex'},
                        'freezeMapping': {'args': '<type __builtin__.bool>',
                                          'longname': 'freezeMapping',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'fm'},
                        'frontOfChain': {'args': '<type __builtin__.bool>',
                                         'longname': 'frontOfChain',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'foc'},
                        'geometry': {'args': '<type __builtin__.str>',
                                     'longname': 'geometry',
                                     'modes': ['query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'g'},
                        'geometryIndices': {'args': '<type __builtin__.bool>',
                                            'longname': 'geometryIndices',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'gi'},
                        'ignoreSelected': {'args': '<type __builtin__.bool>',
                                           'longname': 'ignoreSelected',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ignoreSelected'},
                        'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                    'longname': 'includeHiddenSelections',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ihs'},
                        'latticeReset': {'args': '<type __builtin__.bool>',
                                         'longname': 'latticeReset',
                                         'modes': ['edit'],
                                         'numArgs': 0,
                                         'shortname': 'lr'},
                        'ldivisions': {'args': ['<type __builtin__.int>',
                                                '<type __builtin__.int>',
                                                '<type __builtin__.int>'],
                                       'longname': 'ldivisions',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 3,
                                       'shortname': 'ldv'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'objectCentered': {'args': '<type __builtin__.bool>',
                                           'longname': 'objectCentered',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'oc'},
                        'outsideFalloffDistance': {'args': '<type __builtin__.float>',
                                                   'longname': 'outsideFalloffDistance',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ofd'},
                        'outsideLattice': {'args': '<type __builtin__.int>',
                                           'longname': 'outsideLattice',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ol'},
                        'parallel': {'args': '<type __builtin__.bool>',
                                     'longname': 'parallel',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'par'},
                        'position': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'position',
                                     'modes': ['create'],
                                     'numArgs': 3,
                                     'shortname': 'pos'},
                        'prune': {'args': '<type __builtin__.bool>',
                                  'longname': 'prune',
                                  'modes': ['edit'],
                                  'numArgs': 0,
                                  'shortname': 'pr'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'remove': {'args': '<type __builtin__.bool>',
                                   'longname': 'remove',
                                   'modes': ['edit', 'multiuse'],
                                   'numArgs': 0,
                                   'shortname': 'rm'},
                        'removeTweaks': {'args': '<type __builtin__.bool>',
                                         'longname': 'removeTweaks',
                                         'modes': ['edit'],
                                         'numArgs': 0,
                                         'shortname': 'rt'},
                        'rotation': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'rotation',
                                     'modes': ['create'],
                                     'numArgs': 3,
                                     'shortname': 'ro'},
                        'scale': {'args': ['<type __builtin__.float>',
                                           '<type __builtin__.float>',
                                           '<type __builtin__.float>'],
                                  'longname': 'scale',
                                  'modes': ['create'],
                                  'numArgs': 3,
                                  'shortname': 's'},
                        'split': {'args': '<type __builtin__.bool>',
                                  'longname': 'split',
                                  'modes': ['create', 'edit'],
                                  'numArgs': 0,
                                  'shortname': 'sp'}},
              'removedFlags': {'is': 'ignoreSelected'},
              'shortFlags': {'af': 'after',
                             'ar': 'afterReference',
                             'bf': 'before',
                             'cp': 'commonParent',
                             'db': 'dualBase',
                             'dt': 'deformerTools',
                             'dv': 'divisions',
                             'e': 'edit',
                             'ex': 'exclusive',
                             'fm': 'freezeMapping',
                             'foc': 'frontOfChain',
                             'g': 'geometry',
                             'gi': 'geometryIndices',
                             'ignoreSelected': 'ignoreSelected',
                             'ihs': 'includeHiddenSelections',
                             'ldv': 'ldivisions',
                             'lr': 'latticeReset',
                             'n': 'name',
                             'oc': 'objectCentered',
                             'ofd': 'outsideFalloffDistance',
                             'ol': 'outsideLattice',
                             'par': 'parallel',
                             'pos': 'position',
                             'pr': 'prune',
                             'q': 'query',
                             'rm': 'remove',
                             'ro': 'rotation',
                             'rt': 'removeTweaks',
                             's': 'scale',
                             'sp': 'split'},
              'type': 'animation'},
  'latticeDeformKeyCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'envelope': {'args': '<type __builtin__.float>',
                                                 'longname': 'envelope',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ev'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'history': {'args': '<type __builtin__.bool>',
                                                'longname': 'history',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'latticeColumns': {'args': '<type __builtin__.int>',
                                                       'longname': 'latticeColumns',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'lc'},
                                    'latticeRows': {'args': '<type __builtin__.int>',
                                                    'longname': 'latticeRows',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'lr'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'scaleLatticePts': {'args': '<type __builtin__.bool>',
                                                        'longname': 'scaleLatticePts',
                                                        'modes': ['query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'slp'}},
                          'shortFlags': {'ch': 'history',
                                         'e': 'edit',
                                         'ev': 'envelope',
                                         'ex': 'exists',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'lc': 'latticeColumns',
                                         'lr': 'latticeRows',
                                         'n': 'name',
                                         'q': 'query',
                                         'slp': 'scaleLatticePts'},
                          'type': 'context'},
  'launch': {'flags': {'directory': {'args': '<type __builtin__.str>',
                                     'longname': 'directory',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'dir'},
                       'movie': {'args': '<type __builtin__.str>',
                                 'longname': 'movie',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'mov'},
                       'pdfFile': {'args': '<type __builtin__.str>',
                                   'longname': 'pdfFile',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'pdf'},
                       'webPage': {'args': '<type __builtin__.str>',
                                   'longname': 'webPage',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'web'}},
             'shortFlags': {'dir': 'directory',
                            'mov': 'movie',
                            'pdf': 'pdfFile',
                            'web': 'webPage'},
             'type': 'system'},
  'launchImageEditor': {'flags': {'editImageFile': {'args': '<type __builtin__.str>',
                                                    'longname': 'editImageFile',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'eif'},
                                  'viewImageFile': {'args': '<type __builtin__.str>',
                                                    'longname': 'viewImageFile',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'vif'}},
                        'shortFlags': {'eif': 'editImageFile',
                                       'vif': 'viewImageFile'},
                        'type': 'system'},
  'layerButton': {'callbackFlags': ['typeCommand',
                                    'tc',
                                    'doubleClickCommand',
                                    'dcc',
                                    'visibleChangeCommand',
                                    'vcc',
                                    'dragCallback',
                                    'dgc',
                                    'visibleCommand',
                                    'vc',
                                    'renameCommand',
                                    'rc',
                                    'hideOnPlaybackCommand',
                                    'hpc',
                                    'changeCommand',
                                    'cc',
                                    'command',
                                    'c',
                                    'dropCallback',
                                    'dpc'],
                  'flags': {'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'changeCommand': {'args': 'script',
                                              'longname': 'changeCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cc'},
                            'color': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'color',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 3,
                                      'shortname': 'cl'},
                            'command': {'args': 'script',
                                        'longname': 'command',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'c'},
                            'current': {'args': '<type __builtin__.bool>',
                                        'longname': 'current',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cr'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'doubleClickCommand': {'args': 'script',
                                                   'longname': 'doubleClickCommand',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dcc'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'hideOnPlaybackCommand': {'args': 'script',
                                                      'longname': 'hideOnPlaybackCommand',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'hpc'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'identification': {'args': '<type __builtin__.int>',
                                               'longname': 'identification',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'id'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'label': {'args': '<type __builtin__.str>',
                                      'longname': 'label',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'l'},
                            'labelWidth': {'args': '<type __builtin__.bool>',
                                           'longname': 'labelWidth',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'lw'},
                            'layerHideOnPlayback': {'args': '<type __builtin__.bool>',
                                                    'longname': 'layerHideOnPlayback',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'hp'},
                            'layerState': {'args': '<type __builtin__.str>',
                                           'longname': 'layerState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ls'},
                            'layerVisible': {'args': '<type __builtin__.bool>',
                                             'longname': 'layerVisible',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lv'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'renameCommand': {'args': '<type __builtin__.str>',
                                              'longname': 'renameCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'rc'},
                            'select': {'args': '<type __builtin__.bool>',
                                       'longname': 'select',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 's'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'transparent': {'args': '<type __builtin__.bool>',
                                            'longname': 'transparent',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 't'},
                            'typeCommand': {'args': 'script',
                                            'longname': 'typeCommand',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'tc'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'visibleCommand': {'args': 'script',
                                               'longname': 'visibleCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'vc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'c': 'command',
                                 'cc': 'changeCommand',
                                 'cl': 'color',
                                 'cr': 'current',
                                 'dcc': 'doubleClickCommand',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'ex': 'exists',
                                 'fpn': 'fullPathName',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'hp': 'layerHideOnPlayback',
                                 'hpc': 'hideOnPlaybackCommand',
                                 'id': 'identification',
                                 'io': 'isObscured',
                                 'l': 'label',
                                 'ls': 'layerState',
                                 'lv': 'layerVisible',
                                 'lw': 'labelWidth',
                                 'm': 'manage',
                                 'n': 'name',
                                 'nbg': 'noBackground',
                                 'npm': 'numberOfPopupMenus',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'rc': 'renameCommand',
                                 's': 'select',
                                 'sbm': 'statusBarMessage',
                                 't': 'transparent',
                                 'tc': 'typeCommand',
                                 'ut': 'useTemplate',
                                 'vc': 'visibleCommand',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'layeredShaderPort': {'callbackFlags': ['visibleChangeCommand',
                                          'vcc',
                                          'dragCallback',
                                          'dgc',
                                          'dropCallback',
                                          'dpc'],
                        'flags': {'annotation': {'args': '<type __builtin__.str>',
                                                 'longname': 'annotation',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ann'},
                                  'backgroundColor': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'backgroundColor',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'bgc'},
                                  'defineTemplate': {'args': '<type __builtin__.str>',
                                                     'longname': 'defineTemplate',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'dt'},
                                  'docTag': {'args': '<type __builtin__.str>',
                                             'longname': 'docTag',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dtg'},
                                  'dragCallback': {'args': 'script',
                                                   'longname': 'dragCallback',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dgc'},
                                  'dropCallback': {'args': 'script',
                                                   'longname': 'dropCallback',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dpc'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'enable': {'args': '<type __builtin__.bool>',
                                             'longname': 'enable',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'en'},
                                  'enableBackground': {'args': '<type __builtin__.bool>',
                                                       'longname': 'enableBackground',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ebg'},
                                  'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                          'longname': 'enableKeyboardFocus',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ekf'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'fullPathName': {'args': '<type __builtin__.bool>',
                                                   'longname': 'fullPathName',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'fpn'},
                                  'height': {'args': '<type __builtin__.int>',
                                             'longname': 'height',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'h'},
                                  'highlightColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'highlightColor',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'hlc'},
                                  'isObscured': {'args': '<type __builtin__.bool>',
                                                 'longname': 'isObscured',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'io'},
                                  'manage': {'args': '<type __builtin__.bool>',
                                             'longname': 'manage',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'm'},
                                  'noBackground': {'args': '<type __builtin__.bool>',
                                                   'longname': 'noBackground',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'nbg'},
                                  'node': {'args': 'PyNode',
                                           'longname': 'node',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                         'longname': 'numberOfPopupMenus',
                                                         'modes': ['query'],
                                                         'numArgs': 0,
                                                         'shortname': 'npm'},
                                  'parent': {'args': '<type __builtin__.str>',
                                             'longname': 'parent',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'p'},
                                  'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                     'longname': 'popupMenuArray',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'pma'},
                                  'preventOverride': {'args': '<type __builtin__.bool>',
                                                      'longname': 'preventOverride',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'po'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'selectedColorControl': {'args': '<type __builtin__.str>',
                                                           'longname': 'selectedColorControl',
                                                           'modes': ['create'],
                                                           'numArgs': 1,
                                                           'shortname': 'scc'},
                                  'selectedTransparencyControl': {'args': '<type __builtin__.str>',
                                                                  'longname': 'selectedTransparencyControl',
                                                                  'modes': ['create'],
                                                                  'numArgs': 1,
                                                                  'shortname': 'stc'},
                                  'statusBarMessage': {'args': '<type __builtin__.str>',
                                                       'longname': 'statusBarMessage',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'sbm'},
                                  'useTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'useTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ut'},
                                  'visible': {'args': '<type __builtin__.bool>',
                                              'longname': 'visible',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'vis'},
                                  'visibleChangeCommand': {'args': 'script',
                                                           'longname': 'visibleChangeCommand',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'vcc'},
                                  'width': {'args': '<type __builtin__.int>',
                                            'longname': 'width',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'w'}},
                        'shortFlags': {'ann': 'annotation',
                                       'bgc': 'backgroundColor',
                                       'dgc': 'dragCallback',
                                       'dpc': 'dropCallback',
                                       'dt': 'defineTemplate',
                                       'dtg': 'docTag',
                                       'e': 'edit',
                                       'ebg': 'enableBackground',
                                       'ekf': 'enableKeyboardFocus',
                                       'en': 'enable',
                                       'ex': 'exists',
                                       'fpn': 'fullPathName',
                                       'h': 'height',
                                       'hlc': 'highlightColor',
                                       'io': 'isObscured',
                                       'm': 'manage',
                                       'n': 'node',
                                       'nbg': 'noBackground',
                                       'npm': 'numberOfPopupMenus',
                                       'p': 'parent',
                                       'pma': 'popupMenuArray',
                                       'po': 'preventOverride',
                                       'q': 'query',
                                       'sbm': 'statusBarMessage',
                                       'scc': 'selectedColorControl',
                                       'stc': 'selectedTransparencyControl',
                                       'ut': 'useTemplate',
                                       'vcc': 'visibleChangeCommand',
                                       'vis': 'visible',
                                       'w': 'width'},
                        'type': 'rendering'},
  'layeredTexturePort': {'callbackFlags': ['visibleChangeCommand',
                                           'vcc',
                                           'dragCallback',
                                           'dgc',
                                           'dropCallback',
                                           'dpc'],
                         'flags': {'annotation': {'args': '<type __builtin__.str>',
                                                  'longname': 'annotation',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ann'},
                                   'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'backgroundColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'bgc'},
                                   'defineTemplate': {'args': '<type __builtin__.str>',
                                                      'longname': 'defineTemplate',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'dt'},
                                   'docTag': {'args': '<type __builtin__.str>',
                                              'longname': 'docTag',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dtg'},
                                   'dragCallback': {'args': 'script',
                                                    'longname': 'dragCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dgc'},
                                   'dropCallback': {'args': 'script',
                                                    'longname': 'dropCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dpc'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'enable': {'args': '<type __builtin__.bool>',
                                              'longname': 'enable',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'en'},
                                   'enableBackground': {'args': '<type __builtin__.bool>',
                                                        'longname': 'enableBackground',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ebg'},
                                   'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                           'longname': 'enableKeyboardFocus',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ekf'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'fullPathName': {'args': '<type __builtin__.bool>',
                                                    'longname': 'fullPathName',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'fpn'},
                                   'height': {'args': '<type __builtin__.int>',
                                              'longname': 'height',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'h'},
                                   'highlightColor': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'highlightColor',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'hlc'},
                                   'isObscured': {'args': '<type __builtin__.bool>',
                                                  'longname': 'isObscured',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'io'},
                                   'manage': {'args': '<type __builtin__.bool>',
                                              'longname': 'manage',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'm'},
                                   'noBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'noBackground',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'nbg'},
                                   'node': {'args': 'PyNode',
                                            'longname': 'node',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                          'longname': 'numberOfPopupMenus',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'npm'},
                                   'parent': {'args': '<type __builtin__.str>',
                                              'longname': 'parent',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'p'},
                                   'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                      'longname': 'popupMenuArray',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'pma'},
                                   'preventOverride': {'args': '<type __builtin__.bool>',
                                                       'longname': 'preventOverride',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'po'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'selectedAlphaControl': {'args': '<type __builtin__.str>',
                                                            'longname': 'selectedAlphaControl',
                                                            'modes': ['create'],
                                                            'numArgs': 1,
                                                            'shortname': 'sac'},
                                   'selectedBlendModeControl': {'args': '<type __builtin__.str>',
                                                                'longname': 'selectedBlendModeControl',
                                                                'modes': ['create'],
                                                                'numArgs': 1,
                                                                'shortname': 'sbc'},
                                   'selectedColorControl': {'args': '<type __builtin__.str>',
                                                            'longname': 'selectedColorControl',
                                                            'modes': ['create'],
                                                            'numArgs': 1,
                                                            'shortname': 'scc'},
                                   'selectedIsVisibleControl': {'args': '<type __builtin__.str>',
                                                                'longname': 'selectedIsVisibleControl',
                                                                'modes': ['create'],
                                                                'numArgs': 1,
                                                                'shortname': 'svc'},
                                   'statusBarMessage': {'args': '<type __builtin__.str>',
                                                        'longname': 'statusBarMessage',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'sbm'},
                                   'useTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'useTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ut'},
                                   'visible': {'args': '<type __builtin__.bool>',
                                               'longname': 'visible',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'vis'},
                                   'visibleChangeCommand': {'args': 'script',
                                                            'longname': 'visibleChangeCommand',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'vcc'},
                                   'width': {'args': '<type __builtin__.int>',
                                             'longname': 'width',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'w'}},
                         'shortFlags': {'ann': 'annotation',
                                        'bgc': 'backgroundColor',
                                        'dgc': 'dragCallback',
                                        'dpc': 'dropCallback',
                                        'dt': 'defineTemplate',
                                        'dtg': 'docTag',
                                        'e': 'edit',
                                        'ebg': 'enableBackground',
                                        'ekf': 'enableKeyboardFocus',
                                        'en': 'enable',
                                        'ex': 'exists',
                                        'fpn': 'fullPathName',
                                        'h': 'height',
                                        'hlc': 'highlightColor',
                                        'io': 'isObscured',
                                        'm': 'manage',
                                        'n': 'node',
                                        'nbg': 'noBackground',
                                        'npm': 'numberOfPopupMenus',
                                        'p': 'parent',
                                        'pma': 'popupMenuArray',
                                        'po': 'preventOverride',
                                        'q': 'query',
                                        'sac': 'selectedAlphaControl',
                                        'sbc': 'selectedBlendModeControl',
                                        'sbm': 'statusBarMessage',
                                        'scc': 'selectedColorControl',
                                        'svc': 'selectedIsVisibleControl',
                                        'ut': 'useTemplate',
                                        'vcc': 'visibleChangeCommand',
                                        'vis': 'visible',
                                        'w': 'width'},
                         'type': 'rendering'},
  'layout': {'callbackFlags': ['visibleChangeCommand',
                               'vcc',
                               'dragCallback',
                               'dgc',
                               'dropCallback',
                               'dpc'],
             'flags': {'annotation': {'args': '<type __builtin__.str>',
                                      'longname': 'annotation',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'ann'},
                       'backgroundColor': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'backgroundColor',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'bgc'},
                       'childArray': {'args': '<type __builtin__.bool>',
                                      'longname': 'childArray',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'ca'},
                       'defineTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'defineTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'dt'},
                       'docTag': {'args': '<type __builtin__.str>',
                                  'longname': 'docTag',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'dtg'},
                       'dragCallback': {'args': 'script',
                                        'longname': 'dragCallback',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dgc'},
                       'dropCallback': {'args': 'script',
                                        'longname': 'dropCallback',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dpc'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'enable': {'args': '<type __builtin__.bool>',
                                  'longname': 'enable',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'en'},
                       'enableBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'enableBackground',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ebg'},
                       'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                               'longname': 'enableKeyboardFocus',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ekf'},
                       'exists': {'args': '<type __builtin__.bool>',
                                  'longname': 'exists',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'ex'},
                       'fullPathName': {'args': '<type __builtin__.bool>',
                                        'longname': 'fullPathName',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'fpn'},
                       'height': {'args': '<type __builtin__.int>',
                                  'longname': 'height',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'h'},
                       'highlightColor': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'highlightColor',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 3,
                                          'shortname': 'hlc'},
                       'isObscured': {'args': '<type __builtin__.bool>',
                                      'longname': 'isObscured',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'io'},
                       'manage': {'args': '<type __builtin__.bool>',
                                  'longname': 'manage',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'm'},
                       'noBackground': {'args': '<type __builtin__.bool>',
                                        'longname': 'noBackground',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'nbg'},
                       'numberOfChildren': {'args': '<type __builtin__.bool>',
                                            'longname': 'numberOfChildren',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'nch'},
                       'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                              'longname': 'numberOfPopupMenus',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'npm'},
                       'parent': {'args': '<type __builtin__.str>',
                                  'longname': 'parent',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'p'},
                       'popupMenuArray': {'args': '<type __builtin__.bool>',
                                          'longname': 'popupMenuArray',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'pma'},
                       'preventOverride': {'args': '<type __builtin__.bool>',
                                           'longname': 'preventOverride',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'po'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'statusBarMessage': {'args': '<type __builtin__.str>',
                                            'longname': 'statusBarMessage',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sbm'},
                       'useTemplate': {'args': '<type __builtin__.str>',
                                       'longname': 'useTemplate',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ut'},
                       'visible': {'args': '<type __builtin__.bool>',
                                   'longname': 'visible',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'vis'},
                       'visibleChangeCommand': {'args': 'script',
                                                'longname': 'visibleChangeCommand',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'vcc'},
                       'width': {'args': '<type __builtin__.int>',
                                 'longname': 'width',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'w'}},
             'shortFlags': {'ann': 'annotation',
                            'bgc': 'backgroundColor',
                            'ca': 'childArray',
                            'dgc': 'dragCallback',
                            'dpc': 'dropCallback',
                            'dt': 'defineTemplate',
                            'dtg': 'docTag',
                            'e': 'edit',
                            'ebg': 'enableBackground',
                            'ekf': 'enableKeyboardFocus',
                            'en': 'enable',
                            'ex': 'exists',
                            'fpn': 'fullPathName',
                            'h': 'height',
                            'hlc': 'highlightColor',
                            'io': 'isObscured',
                            'm': 'manage',
                            'nbg': 'noBackground',
                            'nch': 'numberOfChildren',
                            'npm': 'numberOfPopupMenus',
                            'p': 'parent',
                            'pma': 'popupMenuArray',
                            'po': 'preventOverride',
                            'q': 'query',
                            'sbm': 'statusBarMessage',
                            'ut': 'useTemplate',
                            'vcc': 'visibleChangeCommand',
                            'vis': 'visible',
                            'w': 'width'},
             'type': 'windows'},
  'layoutDialog': {'callbackFlags': ['uiScript', 'ui'],
                   'flags': {'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'dismiss': {'args': '<type __builtin__.str>',
                                         'longname': 'dismiss',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'dis'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'title': {'args': '<type __builtin__.str>',
                                       'longname': 'title',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 't'},
                             'uiScript': {'args': 'script',
                                          'longname': 'uiScript',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ui'}},
                   'shortFlags': {'bgc': 'backgroundColor',
                                  'dis': 'dismiss',
                                  'p': 'parent',
                                  't': 'title',
                                  'ui': 'uiScript'},
                   'type': 'windows'},
  'license': {'flags': {'borrow': {'args': '<type __builtin__.bool>',
                                   'longname': 'borrow',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'b'},
                        'info': {'args': '<type __builtin__.bool>',
                                 'longname': 'info',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'i'},
                        'isBorrowed': {'args': '<type __builtin__.bool>',
                                       'longname': 'isBorrowed',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ib'},
                        'isExported': {'args': '<type __builtin__.bool>',
                                       'longname': 'isExported',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ie'},
                        'isTrial': {'args': '<type __builtin__.bool>',
                                    'longname': 'isTrial',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'it'},
                        'licenseMethod': {'args': '<type __builtin__.bool>',
                                          'longname': 'licenseMethod',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'lm'},
                        'productChoice': {'args': '<type __builtin__.bool>',
                                          'longname': 'productChoice',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'pc'},
                        'r': {'args': '<type __builtin__.bool>',
                              'longname': 'r',
                              'modes': ['create'],
                              'numArgs': 0,
                              'shortname': 'r'},
                        'showBorrowInfo': {'args': '<type __builtin__.bool>',
                                           'longname': 'showBorrowInfo',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'sbi'},
                        'showProductInfoDialog': {'args': '<type __builtin__.bool>',
                                                  'longname': 'showProductInfoDialog',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'spi'},
                        'status': {'args': '<type __builtin__.bool>',
                                   'longname': 'status',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 's'},
                        'usage': {'args': '<type __builtin__.bool>',
                                  'longname': 'usage',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'u'}},
              'removedFlags': {'return': 'r'},
              'shortFlags': {'b': 'borrow',
                             'i': 'info',
                             'ib': 'isBorrowed',
                             'ie': 'isExported',
                             'it': 'isTrial',
                             'lm': 'licenseMethod',
                             'pc': 'productChoice',
                             'r': 'r',
                             's': 'status',
                             'sbi': 'showBorrowInfo',
                             'spi': 'showProductInfoDialog',
                             'u': 'usage'},
              'type': 'general'},
  'licenseCheck': {'flags': {'mode': {'args': '<type __builtin__.str>',
                                      'longname': 'mode',
                                      'numArgs': 1,
                                      'shortname': 'm'},
                             'type': {'args': '<type __builtin__.str>',
                                      'longname': 'type',
                                      'numArgs': 1,
                                      'shortname': 'typ'}},
                   'shortFlags': {'m': 'mode', 'typ': 'type'},
                   'type': 'other'},
  'lightList': {'flags': {'add': {'args': 'PyNode',
                                  'longname': 'add',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'add'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'remove': {'args': 'PyNode',
                                     'longname': 'remove',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'rm'}},
                'shortFlags': {'add': 'add',
                               'e': 'edit',
                               'q': 'query',
                               'rm': 'remove'},
                'type': 'rendering'},
  'lightlink': {'flags': {'b': {'args': '<type __builtin__.bool>',
                                'longname': 'b',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'b'},
                          'hierarchy': {'args': '<type __builtin__.bool>',
                                        'longname': 'hierarchy',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                          'light': {'args': 'PyNode',
                                    'longname': 'light',
                                    'modes': ['create', 'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'l'},
                          'make': {'args': '<type __builtin__.bool>',
                                   'longname': 'make',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'm'},
                          'object': {'args': 'PyNode',
                                     'longname': 'object',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'o'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'sets': {'args': '<type __builtin__.bool>',
                                   'longname': 'sets',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'set'},
                          'shadow': {'args': '<type __builtin__.bool>',
                                     'longname': 'shadow',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'shd'},
                          'shapes': {'args': '<type __builtin__.bool>',
                                     'longname': 'shapes',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'shp'},
                          'transforms': {'args': '<type __builtin__.bool>',
                                         'longname': 'transforms',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 't'},
                          'useActiveLights': {'args': '<type __builtin__.bool>',
                                              'longname': 'useActiveLights',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ual'},
                          'useActiveObjects': {'args': '<type __builtin__.bool>',
                                               'longname': 'useActiveObjects',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'uao'}},
                'removedFlags': {'break': 'b'},
                'shortFlags': {'b': 'b',
                               'h': 'hierarchy',
                               'l': 'light',
                               'm': 'make',
                               'o': 'object',
                               'q': 'query',
                               'set': 'sets',
                               'shd': 'shadow',
                               'shp': 'shapes',
                               't': 'transforms',
                               'ual': 'useActiveLights',
                               'uao': 'useActiveObjects'},
                'type': 'rendering'},
  'linearPrecision': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'}},
                      'shortFlags': {'q': 'query'},
                      'type': 'windows'},
  'listAnimatable': {'flags': {'active': {'args': '<type __builtin__.bool>',
                                          'longname': 'active',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'act'},
                               'manip': {'args': '<type __builtin__.bool>',
                                         'longname': 'manip',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'm'},
                               'manipHandle': {'args': '<type __builtin__.bool>',
                                               'longname': 'manipHandle',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'mh'},
                               'shape': {'args': '<type __builtin__.bool>',
                                         'longname': 'shape',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 's'},
                               'type': {'args': '<type __builtin__.bool>',
                                        'longname': 'type',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'typ'}},
                     'shortFlags': {'act': 'active',
                                    'm': 'manip',
                                    'mh': 'manipHandle',
                                    's': 'shape',
                                    'typ': 'type'},
                     'type': 'animation'},
  'listAttr': {'flags': {'array': {'args': '<type __builtin__.bool>',
                                   'longname': 'array',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'a'},
                         'caching': {'args': '<type __builtin__.bool>',
                                     'longname': 'caching',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ca'},
                         'category': {'args': '<type __builtin__.str>',
                                      'longname': 'category',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'ct'},
                         'changedSinceFileOpen': {'args': '<type __builtin__.bool>',
                                                  'longname': 'changedSinceFileOpen',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'cfo'},
                         'channelBox': {'args': '<type __builtin__.bool>',
                                        'longname': 'channelBox',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'cb'},
                         'connectable': {'args': '<type __builtin__.bool>',
                                         'longname': 'connectable',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'c'},
                         'extension': {'args': '<type __builtin__.bool>',
                                       'longname': 'extension',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                         'fromPlugin': {'args': '<type __builtin__.bool>',
                                        'longname': 'fromPlugin',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'fp'},
                         'hasData': {'args': '<type __builtin__.bool>',
                                     'longname': 'hasData',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'hd'},
                         'hasNullData': {'args': '<type __builtin__.bool>',
                                         'longname': 'hasNullData',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'hnd'},
                         'inUse': {'args': '<type __builtin__.bool>',
                                   'longname': 'inUse',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'iu'},
                         'keyable': {'args': '<type __builtin__.bool>',
                                     'longname': 'keyable',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'k'},
                         'leaf': {'args': '<type __builtin__.bool>',
                                  'longname': 'leaf',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'lf'},
                         'locked': {'args': '<type __builtin__.bool>',
                                    'longname': 'locked',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'l'},
                         'multi': {'args': '<type __builtin__.bool>',
                                   'longname': 'multi',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'm'},
                         'output': {'args': '<type __builtin__.bool>',
                                    'longname': 'output',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'o'},
                         'ramp': {'args': '<type __builtin__.bool>',
                                  'longname': 'ramp',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'ra'},
                         'read': {'args': '<type __builtin__.bool>',
                                  'longname': 'read',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'r'},
                         'readOnly': {'args': '<type __builtin__.bool>',
                                      'longname': 'readOnly',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ro'},
                         'scalar': {'args': '<type __builtin__.bool>',
                                    'longname': 'scalar',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 's'},
                         'scalarAndArray': {'args': '<type __builtin__.bool>',
                                            'longname': 'scalarAndArray',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'sa'},
                         'settable': {'args': '<type __builtin__.bool>',
                                      'longname': 'settable',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'se'},
                         'shortNames': {'args': '<type __builtin__.bool>',
                                        'longname': 'shortNames',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'sn'},
                         'string': {'args': '<type __builtin__.str>',
                                    'longname': 'string',
                                    'modes': ['create', 'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'st'},
                         'unlocked': {'args': '<type __builtin__.bool>',
                                      'longname': 'unlocked',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'u'},
                         'usedAsFilename': {'args': '<type __builtin__.bool>',
                                            'longname': 'usedAsFilename',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'uf'},
                         'userDefined': {'args': '<type __builtin__.bool>',
                                         'longname': 'userDefined',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ud'},
                         'visible': {'args': '<type __builtin__.bool>',
                                     'longname': 'visible',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'v'},
                         'write': {'args': '<type __builtin__.bool>',
                                   'longname': 'write',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'w'}},
               'shortFlags': {'a': 'array',
                              'c': 'connectable',
                              'ca': 'caching',
                              'cb': 'channelBox',
                              'cfo': 'changedSinceFileOpen',
                              'ct': 'category',
                              'ex': 'extension',
                              'fp': 'fromPlugin',
                              'hd': 'hasData',
                              'hnd': 'hasNullData',
                              'iu': 'inUse',
                              'k': 'keyable',
                              'l': 'locked',
                              'lf': 'leaf',
                              'm': 'multi',
                              'o': 'output',
                              'r': 'read',
                              'ra': 'ramp',
                              'ro': 'readOnly',
                              's': 'scalar',
                              'sa': 'scalarAndArray',
                              'se': 'settable',
                              'sn': 'shortNames',
                              'st': 'string',
                              'u': 'unlocked',
                              'ud': 'userDefined',
                              'uf': 'usedAsFilename',
                              'v': 'visible',
                              'w': 'write'},
               'type': 'general'},
  'listAttrPatterns': {'flags': {'patternType': {'args': '<type __builtin__.bool>',
                                                 'longname': 'patternType',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'pt'},
                                 'verbose': {'args': '<type __builtin__.bool>',
                                             'longname': 'verbose',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'v'}},
                       'shortFlags': {'pt': 'patternType', 'v': 'verbose'},
                       'type': 'general'},
  'listCameras': {'flags': {'orthographic': {'args': '<type __builtin__.bool>',
                                             'longname': 'orthographic',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'o'},
                            'perspective': {'args': '<type __builtin__.bool>',
                                            'longname': 'perspective',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'p'}},
                  'shortFlags': {'o': 'orthographic', 'p': 'perspective'},
                  'type': 'rendering'},
  'listConnections': {'flags': {'connections': {'args': '<type __builtin__.bool>',
                                                'longname': 'connections',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'c'},
                                'destination': {'args': '<type __builtin__.bool>',
                                                'longname': 'destination',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'd'},
                                'exactType': {'args': '<type __builtin__.bool>',
                                              'longname': 'exactType',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'et'},
                                'plugs': {'args': '<type __builtin__.bool>',
                                          'longname': 'plugs',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                                'shapes': {'args': '<type __builtin__.bool>',
                                           'longname': 'shapes',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'sh'},
                                'skipConversionNodes': {'args': '<type __builtin__.bool>',
                                                        'longname': 'skipConversionNodes',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'scn'},
                                'source': {'args': '<type __builtin__.bool>',
                                           'longname': 'source',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 's'},
                                'type': {'args': '<type __builtin__.str>',
                                         'longname': 'type',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 't'}},
                      'shortFlags': {'c': 'connections',
                                     'd': 'destination',
                                     'et': 'exactType',
                                     'p': 'plugs',
                                     's': 'source',
                                     'scn': 'skipConversionNodes',
                                     'sh': 'shapes',
                                     't': 'type'},
                      'type': 'general'},
  'listDeviceAttachments': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                                    'longname': 'attribute',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'at'},
                                      'axis': {'args': '<type __builtin__.str>',
                                               'longname': 'axis',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ax'},
                                      'clutch': {'args': '<type __builtin__.str>',
                                                 'longname': 'clutch',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'c'},
                                      'device': {'args': '<type __builtin__.str>',
                                                 'longname': 'device',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'd'},
                                      'file': {'args': '<type __builtin__.str>',
                                               'longname': 'file',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'f'},
                                      'selection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'selection',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'sl'},
                                      'write': {'args': '<type __builtin__.bool>',
                                                'longname': 'write',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'w'}},
                            'shortFlags': {},
                            'type': 'system'},
  'listHistory': {'flags': {'allConnections': {'args': '<type __builtin__.bool>',
                                               'longname': 'allConnections',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ac'},
                            'allFuture': {'args': '<type __builtin__.bool>',
                                          'longname': 'allFuture',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'af'},
                            'allGraphs': {'args': '<type __builtin__.bool>',
                                          'longname': 'allGraphs',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ag'},
                            'breadthFirst': {'args': '<type __builtin__.bool>',
                                             'longname': 'breadthFirst',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'bf'},
                            'future': {'args': '<type __builtin__.bool>',
                                       'longname': 'future',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'f'},
                            'futureLocalAttr': {'args': '<type __builtin__.bool>',
                                                'longname': 'futureLocalAttr',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'fl'},
                            'futureWorldAttr': {'args': '<type __builtin__.bool>',
                                                'longname': 'futureWorldAttr',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'fw'},
                            'groupLevels': {'args': '<type __builtin__.bool>',
                                            'longname': 'groupLevels',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'gl'},
                            'historyAttr': {'args': '<type __builtin__.bool>',
                                            'longname': 'historyAttr',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ha'},
                            'interestLevel': {'args': '<type __builtin__.int>',
                                              'longname': 'interestLevel',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'il'},
                            'leaf': {'args': '<type __builtin__.bool>',
                                     'longname': 'leaf',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'lf'},
                            'levels': {'args': '<type __builtin__.int>',
                                       'longname': 'levels',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'lv'},
                            'pruneDagObjects': {'args': '<type __builtin__.bool>',
                                                'longname': 'pruneDagObjects',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'pdo'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'}},
                  'shortFlags': {'ac': 'allConnections',
                                 'af': 'allFuture',
                                 'ag': 'allGraphs',
                                 'bf': 'breadthFirst',
                                 'f': 'future',
                                 'fl': 'futureLocalAttr',
                                 'fw': 'futureWorldAttr',
                                 'gl': 'groupLevels',
                                 'ha': 'historyAttr',
                                 'il': 'interestLevel',
                                 'lf': 'leaf',
                                 'lv': 'levels',
                                 'pdo': 'pruneDagObjects',
                                 'q': 'query'},
                  'type': 'general'},
  'listInputDeviceAxes': {'flags': {}, 'shortFlags': {}, 'type': 'system'},
  'listInputDeviceButtons': {'flags': {}, 'shortFlags': {}, 'type': 'system'},
  'listInputDevices': {'flags': {'free': {'args': '<type __builtin__.bool>',
                                          'longname': 'free',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'f'},
                                 'primary': {'args': '<type __builtin__.bool>',
                                             'longname': 'primary',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'p'},
                                 'secondary': {'args': '<type __builtin__.bool>',
                                               'longname': 'secondary',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 's'}},
                       'shortFlags': {},
                       'type': 'system'},
  'listNodeTypes': {'flags': {'exclude': {'args': '<type __builtin__.str>',
                                          'longname': 'exclude',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ex'}},
                    'shortFlags': {'ex': 'exclude'},
                    'type': 'general'},
  'listNodesWithIncorrectNames': {'flags': {},
                                  'shortFlags': {},
                                  'type': 'general'},
  'listRelatives': {'flags': {'allDescendents': {'args': '<type __builtin__.bool>',
                                                 'longname': 'allDescendents',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ad'},
                              'allParents': {'args': '<type __builtin__.bool>',
                                             'longname': 'allParents',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ap'},
                              'children': {'args': '<type __builtin__.bool>',
                                           'longname': 'children',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'c'},
                              'fullPath': {'args': '<type __builtin__.bool>',
                                           'longname': 'fullPath',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'f'},
                              'noIntermediate': {'args': '<type __builtin__.bool>',
                                                 'longname': 'noIntermediate',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ni'},
                              'parent': {'args': '<type __builtin__.bool>',
                                         'longname': 'parent',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'p'},
                              'path': {'args': '<type __builtin__.bool>',
                                       'longname': 'path',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'pa'},
                              'shapes': {'args': '<type __builtin__.bool>',
                                         'longname': 'shapes',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 's'},
                              'type': {'args': '<type __builtin__.str>',
                                       'longname': 'type',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'typ'}},
                    'shortFlags': {'ad': 'allDescendents',
                                   'ap': 'allParents',
                                   'c': 'children',
                                   'f': 'fullPath',
                                   'ni': 'noIntermediate',
                                   'p': 'parent',
                                   'pa': 'path',
                                   's': 'shapes',
                                   'typ': 'type'},
                    'type': 'general'},
  'listSets': {'flags': {'allSets': {'args': '<type __builtin__.bool>',
                                     'longname': 'allSets',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'allSets'},
                         'extendToShape': {'args': '<type __builtin__.bool>',
                                           'longname': 'extendToShape',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ets'},
                         'object': {'args': 'PyNode',
                                    'longname': 'object',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'o'},
                         'type': {'args': '<type __builtin__.int>',
                                  'longname': 'type',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 't'}},
               'removedFlags': {'as': 'allSets'},
               'shortFlags': {'allSets': 'allSets',
                              'ets': 'extendToShape',
                              'o': 'object',
                              't': 'type'},
               'type': 'general'},
  'loadFluid': {'flags': {'currentTime': {'args': '<type __builtin__.bool>',
                                          'longname': 'currentTime',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'ct'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'frame': {'args': '<type __builtin__.float>',
                                    'longname': 'frame',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'f'},
                          'initialConditions': {'args': '<type __builtin__.bool>',
                                                'longname': 'initialConditions',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'ic'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'}},
                'shortFlags': {'ct': 'currentTime',
                               'e': 'edit',
                               'f': 'frame',
                               'ic': 'initialConditions',
                               'q': 'query'},
                'type': 'effects'},
  'loadModule': {'flags': {'allModules': {'args': '<type __builtin__.bool>',
                                          'longname': 'allModules',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'a'},
                           'load': {'args': '<type __builtin__.str>',
                                    'longname': 'load',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'ld'},
                           'scan': {'args': '<type __builtin__.bool>',
                                    'longname': 'scan',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'sc'}},
                 'shortFlags': {'a': 'allModules',
                                'ld': 'load',
                                'sc': 'scan'},
                 'type': 'system'},
  'loadPlugin': {'callbackFlags': ['addCallback',
                                   'ac',
                                   'removeCallback',
                                   'rc'],
                 'flags': {'addCallback': {'args': 'script',
                                           'longname': 'addCallback',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ac'},
                           'allPlugins': {'args': '<type __builtin__.bool>',
                                          'longname': 'allPlugins',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'a'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'qObsolete': {'args': '<type __builtin__.bool>',
                                         'longname': 'qObsolete',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                           'quiet': {'args': '<type __builtin__.bool>',
                                     'longname': 'quiet',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'qt'},
                           'removeCallback': {'args': 'script',
                                              'longname': 'removeCallback',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'rc'}},
                 'shortFlags': {'a': 'allPlugins',
                                'ac': 'addCallback',
                                'n': 'name',
                                'q': 'qObsolete',
                                'qt': 'quiet',
                                'rc': 'removeCallback'},
                 'type': 'system'},
  'loadPrefObjects': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'loadUI': {'flags': {'listTypes': {'args': '<type __builtin__.bool>',
                                     'longname': 'listTypes',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'lt'},
                       'uiFile': {'args': '<type __builtin__.str>',
                                  'longname': 'uiFile',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'f'},
                       'uiString': {'args': '<type __builtin__.str>',
                                    'longname': 'uiString',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 's'},
                       'verbose': {'args': '<type __builtin__.bool>',
                                   'longname': 'verbose',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'v'},
                       'workingDirectory': {'args': '<type __builtin__.str>',
                                            'longname': 'workingDirectory',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'wd'}},
             'shortFlags': {'f': 'uiFile',
                            'lt': 'listTypes',
                            's': 'uiString',
                            'v': 'verbose',
                            'wd': 'workingDirectory'},
             'type': 'windows'},
  'lockNode': {'flags': {'ignoreComponents': {'args': '<type __builtin__.bool>',
                                              'longname': 'ignoreComponents',
                                              'modes': ['create', 'query'],
                                              'numArgs': 0,
                                              'shortname': 'ic'},
                         'lock': {'args': '<type __builtin__.bool>',
                                  'longname': 'lock',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'l'},
                         'lockName': {'args': '<type __builtin__.bool>',
                                      'longname': 'lockName',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'ln'},
                         'lockUnpublished': {'args': '<type __builtin__.bool>',
                                             'longname': 'lockUnpublished',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'lu'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'}},
               'shortFlags': {'ic': 'ignoreComponents',
                              'l': 'lock',
                              'ln': 'lockName',
                              'lu': 'lockUnpublished',
                              'q': 'query'},
               'type': 'general'},
  'loft': {'flags': {'autoReverse': {'args': '<type __builtin__.bool>',
                                     'longname': 'autoReverse',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'ar'},
                     'caching': {'args': '<type __builtin__.bool>',
                                 'longname': 'caching',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'cch'},
                     'close': {'args': '<type __builtin__.bool>',
                               'longname': 'close',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'c'},
                     'constructionHistory': {'args': '<type __builtin__.bool>',
                                             'longname': 'constructionHistory',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ch'},
                     'createCusp': {'args': '<type __builtin__.bool>',
                                    'longname': 'createCusp',
                                    'modes': ['create',
                                              'query',
                                              'edit',
                                              'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'cc'},
                     'degree': {'args': '<type __builtin__.int>',
                                'longname': 'degree',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'd'},
                     'edit': {'args': '<type __builtin__.bool>',
                              'longname': 'edit',
                              'numArgs': 0,
                              'shortname': 'e'},
                     'frozen': {'args': '<type __builtin__.bool>',
                                'longname': 'frozen',
                                'numArgs': 1,
                                'shortname': 'fzn'},
                     'name': {'args': '<type __builtin__.str>',
                              'longname': 'name',
                              'modes': ['create'],
                              'numArgs': 1,
                              'shortname': 'n'},
                     'nodeState': {'args': '<type __builtin__.int>',
                                   'longname': 'nodeState',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'nds'},
                     'object': {'args': '<type __builtin__.bool>',
                                'longname': 'object',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'o'},
                     'polygon': {'args': '<type __builtin__.int>',
                                 'longname': 'polygon',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'po'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'range': {'args': '<type __builtin__.bool>',
                               'longname': 'range',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'rn'},
                     'rebuild': {'args': '<type __builtin__.bool>',
                                 'longname': 'rebuild',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'rb'},
                     'reverse': {'args': '<type __builtin__.bool>',
                                 'longname': 'reverse',
                                 'modes': ['create',
                                           'query',
                                           'edit',
                                           'multiuse'],
                                 'numArgs': 1,
                                 'shortname': 'r'},
                     'reverseSurfaceNormals': {'args': '<type __builtin__.bool>',
                                               'longname': 'reverseSurfaceNormals',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rsn'},
                     'sectionSpans': {'args': '<type __builtin__.int>',
                                      'longname': 'sectionSpans',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'ss'},
                     'uniform': {'args': '<type __builtin__.bool>',
                                 'longname': 'uniform',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'u'}},
           'shortFlags': {'ar': 'autoReverse',
                          'c': 'close',
                          'cc': 'createCusp',
                          'cch': 'caching',
                          'ch': 'constructionHistory',
                          'd': 'degree',
                          'e': 'edit',
                          'fzn': 'frozen',
                          'n': 'name',
                          'nds': 'nodeState',
                          'o': 'object',
                          'po': 'polygon',
                          'q': 'query',
                          'r': 'reverse',
                          'rb': 'rebuild',
                          'rn': 'range',
                          'rsn': 'reverseSurfaceNormals',
                          'ss': 'sectionSpans',
                          'u': 'uniform'},
           'type': 'modeling'},
  'lookThru': {'flags': {'farClip': {'args': '<type __builtin__.float>',
                                     'longname': 'farClip',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'fc'},
                         'nearClip': {'args': '<type __builtin__.float>',
                                      'longname': 'nearClip',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'nc'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'}},
               'shortFlags': {'fc': 'farClip',
                              'nc': 'nearClip',
                              'q': 'query'},
               'type': 'rendering'},
  'ls': {'flags': {'absoluteName': {'args': '<type __builtin__.bool>',
                                    'longname': 'absoluteName',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'an'},
                   'allPaths': {'args': '<type __builtin__.bool>',
                                'longname': 'allPaths',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'ap'},
                   'assemblies': {'args': '<type __builtin__.bool>',
                                  'longname': 'assemblies',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'assemblies'},
                   'cameras': {'args': '<type __builtin__.bool>',
                               'longname': 'cameras',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'ca'},
                   'containerType': {'args': '<type __builtin__.str>',
                                     'longname': 'containerType',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'ct'},
                   'containers': {'args': '<type __builtin__.bool>',
                                  'longname': 'containers',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'con'},
                   'dagObjects': {'args': '<type __builtin__.bool>',
                                  'longname': 'dagObjects',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'dag'},
                   'defaultNodes': {'args': '<type __builtin__.bool>',
                                    'longname': 'defaultNodes',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'dn'},
                   'dependencyNodes': {'args': '<type __builtin__.bool>',
                                       'longname': 'dependencyNodes',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'dep'},
                   'exactType': {'args': '<type __builtin__.str>',
                                 'longname': 'exactType',
                                 'modes': ['create', 'multiuse'],
                                 'numArgs': 1,
                                 'shortname': 'et'},
                   'excludeType': {'args': '<type __builtin__.str>',
                                   'longname': 'excludeType',
                                   'modes': ['create', 'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 'ext'},
                   'flatten': {'args': '<type __builtin__.bool>',
                               'longname': 'flatten',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'fl'},
                   'geometry': {'args': '<type __builtin__.bool>',
                                'longname': 'geometry',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'g'},
                   'ghost': {'args': '<type __builtin__.bool>',
                             'longname': 'ghost',
                             'modes': ['create'],
                             'numArgs': 0,
                             'shortname': 'gh'},
                   'head': {'args': '<type __builtin__.int>',
                            'longname': 'head',
                            'modes': ['create'],
                            'numArgs': 1,
                            'shortname': 'hd'},
                   'hilite': {'args': '<type __builtin__.bool>',
                              'longname': 'hilite',
                              'modes': ['create'],
                              'numArgs': 0,
                              'shortname': 'hl'},
                   'intermediateObjects': {'args': '<type __builtin__.bool>',
                                           'longname': 'intermediateObjects',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                   'invisible': {'args': '<type __builtin__.bool>',
                                 'longname': 'invisible',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'iv'},
                   'leaf': {'args': '<type __builtin__.bool>',
                            'longname': 'leaf',
                            'modes': ['create'],
                            'numArgs': 0,
                            'shortname': 'lf'},
                   'lights': {'args': '<type __builtin__.bool>',
                              'longname': 'lights',
                              'modes': ['create'],
                              'numArgs': 0,
                              'shortname': 'lt'},
                   'live': {'args': '<type __builtin__.bool>',
                            'longname': 'live',
                            'modes': ['create'],
                            'numArgs': 0,
                            'shortname': 'lv'},
                   'lockedNodes': {'args': '<type __builtin__.bool>',
                                   'longname': 'lockedNodes',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'ln'},
                   'long': {'args': '<type __builtin__.bool>',
                            'longname': 'long',
                            'modes': ['create'],
                            'numArgs': 0,
                            'shortname': 'l'},
                   'materials': {'args': '<type __builtin__.bool>',
                                 'longname': 'materials',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'mat'},
                   'modified': {'args': '<type __builtin__.bool>',
                                'longname': 'modified',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'mod'},
                   'noIntermediate': {'args': '<type __builtin__.bool>',
                                      'longname': 'noIntermediate',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ni'},
                   'nodeTypes': {'args': '<type __builtin__.bool>',
                                 'longname': 'nodeTypes',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'nt'},
                   'objectsOnly': {'args': '<type __builtin__.bool>',
                                   'longname': 'objectsOnly',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'o'},
                   'orderedSelection': {'args': '<type __builtin__.bool>',
                                        'longname': 'orderedSelection',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'os'},
                   'partitions': {'args': '<type __builtin__.bool>',
                                  'longname': 'partitions',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'pr'},
                   'persistentNodes': {'args': '<type __builtin__.bool>',
                                       'longname': 'persistentNodes',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'pn'},
                   'planes': {'args': '<type __builtin__.bool>',
                              'longname': 'planes',
                              'modes': ['create'],
                              'numArgs': 0,
                              'shortname': 'pl'},
                   'preSelectHilite': {'args': '<type __builtin__.bool>',
                                       'longname': 'preSelectHilite',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'psh'},
                   'readOnly': {'args': '<type __builtin__.bool>',
                                'longname': 'readOnly',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'ro'},
                   'recursive': {'args': '<type __builtin__.bool>',
                                 'longname': 'recursive',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'r'},
                   'referencedNodes': {'args': '<type __builtin__.bool>',
                                       'longname': 'referencedNodes',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'rn'},
                   'references': {'args': '<type __builtin__.bool>',
                                  'longname': 'references',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'rf'},
                   'renderGlobals': {'args': '<type __builtin__.bool>',
                                     'longname': 'renderGlobals',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'rg'},
                   'renderQualities': {'args': '<type __builtin__.bool>',
                                       'longname': 'renderQualities',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'rq'},
                   'renderResolutions': {'args': '<type __builtin__.bool>',
                                         'longname': 'renderResolutions',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'rr'},
                   'renderSetups': {'args': '<type __builtin__.bool>',
                                    'longname': 'renderSetups',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'rs'},
                   'selection': {'args': '<type __builtin__.bool>',
                                 'longname': 'selection',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'sl'},
                   'sets': {'args': '<type __builtin__.bool>',
                            'longname': 'sets',
                            'modes': ['create'],
                            'numArgs': 0,
                            'shortname': 'set'},
                   'shapes': {'args': '<type __builtin__.bool>',
                              'longname': 'shapes',
                              'modes': ['create'],
                              'numArgs': 0,
                              'shortname': 's'},
                   'shortNames': {'args': '<type __builtin__.bool>',
                                  'longname': 'shortNames',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'sn'},
                   'showNamespace': {'args': '<type __builtin__.bool>',
                                     'longname': 'showNamespace',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'sns'},
                   'showType': {'args': '<type __builtin__.bool>',
                                'longname': 'showType',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'st'},
                   'tail': {'args': '<type __builtin__.int>',
                            'longname': 'tail',
                            'modes': ['create'],
                            'numArgs': 1,
                            'shortname': 'tl'},
                   'templated': {'args': '<type __builtin__.bool>',
                                 'longname': 'templated',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'tm'},
                   'textures': {'args': '<type __builtin__.bool>',
                                'longname': 'textures',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'tex'},
                   'transforms': {'args': '<type __builtin__.bool>',
                                  'longname': 'transforms',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'tr'},
                   'type': {'args': '<type __builtin__.str>',
                            'longname': 'type',
                            'modes': ['create', 'multiuse'],
                            'numArgs': 1,
                            'shortname': 'typ'},
                   'undeletable': {'args': '<type __builtin__.bool>',
                                   'longname': 'undeletable',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'ud'},
                   'untemplated': {'args': '<type __builtin__.bool>',
                                   'longname': 'untemplated',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'ut'},
                   'uuid': {'args': '<type __builtin__.bool>',
                            'longname': 'uuid',
                            'modes': ['create'],
                            'numArgs': 0,
                            'shortname': 'uid'},
                   'visible': {'args': '<type __builtin__.bool>',
                               'longname': 'visible',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'v'}},
         'removedFlags': {'as': 'assemblies'},
         'shortFlags': {'an': 'absoluteName',
                        'ap': 'allPaths',
                        'assemblies': 'assemblies',
                        'ca': 'cameras',
                        'con': 'containers',
                        'ct': 'containerType',
                        'dag': 'dagObjects',
                        'dep': 'dependencyNodes',
                        'dn': 'defaultNodes',
                        'et': 'exactType',
                        'ext': 'excludeType',
                        'fl': 'flatten',
                        'g': 'geometry',
                        'gh': 'ghost',
                        'hd': 'head',
                        'hl': 'hilite',
                        'io': 'intermediateObjects',
                        'iv': 'invisible',
                        'l': 'long',
                        'lf': 'leaf',
                        'ln': 'lockedNodes',
                        'lt': 'lights',
                        'lv': 'live',
                        'mat': 'materials',
                        'mod': 'modified',
                        'ni': 'noIntermediate',
                        'nt': 'nodeTypes',
                        'o': 'objectsOnly',
                        'os': 'orderedSelection',
                        'pl': 'planes',
                        'pn': 'persistentNodes',
                        'pr': 'partitions',
                        'psh': 'preSelectHilite',
                        'r': 'recursive',
                        'rf': 'references',
                        'rg': 'renderGlobals',
                        'rn': 'referencedNodes',
                        'ro': 'readOnly',
                        'rq': 'renderQualities',
                        'rr': 'renderResolutions',
                        'rs': 'renderSetups',
                        's': 'shapes',
                        'set': 'sets',
                        'sl': 'selection',
                        'sn': 'shortNames',
                        'sns': 'showNamespace',
                        'st': 'showType',
                        'tex': 'textures',
                        'tl': 'tail',
                        'tm': 'templated',
                        'tr': 'transforms',
                        'typ': 'type',
                        'ud': 'undeletable',
                        'uid': 'uuid',
                        'ut': 'untemplated',
                        'v': 'visible'},
         'type': 'general'},
  'lsThroughFilter': {'flags': {'item': {'args': '<type __builtin__.str>',
                                         'longname': 'item',
                                         'modes': ['create', 'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'it'},
                                'nodeArray': {'args': '<type __builtin__.bool>',
                                              'longname': 'nodeArray',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'na'},
                                'reverse': {'args': '<type __builtin__.bool>',
                                            'longname': 'reverse',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'rv'},
                                'selection': {'args': '<type __builtin__.bool>',
                                              'longname': 'selection',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'sl'},
                                'sort': {'args': '<type __builtin__.str>',
                                         'longname': 'sort',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'so'}},
                      'shortFlags': {'it': 'item',
                                     'na': 'nodeArray',
                                     'rv': 'reverse',
                                     'sl': 'selection',
                                     'so': 'sort'},
                      'type': 'rendering'},
  'lsUI': {'flags': {'cmdTemplates': {'args': '<type __builtin__.bool>',
                                      'longname': 'cmdTemplates',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ct'},
                     'collection': {'args': '<type __builtin__.bool>',
                                    'longname': 'collection',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'col'},
                     'contexts': {'args': '<type __builtin__.bool>',
                                  'longname': 'contexts',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'ctx'},
                     'controlLayouts': {'args': '<type __builtin__.bool>',
                                        'longname': 'controlLayouts',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'cl'},
                     'controls': {'args': '<type __builtin__.bool>',
                                  'longname': 'controls',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'ctl'},
                     'dumpWidgets': {'args': '<type __builtin__.bool>',
                                     'longname': 'dumpWidgets',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'dw'},
                     'editors': {'args': '<type __builtin__.bool>',
                                 'longname': 'editors',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'ed'},
                     'head': {'args': '<type __builtin__.int>',
                              'longname': 'head',
                              'modes': ['create'],
                              'numArgs': 1,
                              'shortname': 'hd'},
                     'long': {'args': '<type __builtin__.bool>',
                              'longname': 'long',
                              'modes': ['create'],
                              'numArgs': 0,
                              'shortname': 'l'},
                     'menuItems': {'args': '<type __builtin__.bool>',
                                   'longname': 'menuItems',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'mi'},
                     'menus': {'args': '<type __builtin__.bool>',
                               'longname': 'menus',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'm'},
                     'numWidgets': {'args': '<type __builtin__.bool>',
                                    'longname': 'numWidgets',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'nw'},
                     'panels': {'args': '<type __builtin__.bool>',
                                'longname': 'panels',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'p'},
                     'radioMenuItemCollections': {'args': '<type __builtin__.bool>',
                                                  'longname': 'radioMenuItemCollections',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'rmc'},
                     'tail': {'args': '<type __builtin__.int>',
                              'longname': 'tail',
                              'modes': ['create'],
                              'numArgs': 1,
                              'shortname': 'tl'},
                     'type': {'args': '<type __builtin__.str>',
                              'longname': 'type',
                              'modes': ['create', 'multiuse'],
                              'numArgs': 1,
                              'shortname': 'typ'},
                     'windows': {'args': '<type __builtin__.bool>',
                                 'longname': 'windows',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'wnd'}},
           'shortFlags': {'cl': 'controlLayouts',
                          'col': 'collection',
                          'ct': 'cmdTemplates',
                          'ctl': 'controls',
                          'ctx': 'contexts',
                          'dw': 'dumpWidgets',
                          'ed': 'editors',
                          'hd': 'head',
                          'l': 'long',
                          'm': 'menus',
                          'mi': 'menuItems',
                          'nw': 'numWidgets',
                          'p': 'panels',
                          'rmc': 'radioMenuItemCollections',
                          'tl': 'tail',
                          'typ': 'type',
                          'wnd': 'windows'},
           'type': 'language'},
  'makeIdentity': {'flags': {'apply': {'args': '<type __builtin__.bool>',
                                       'longname': 'apply',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'a'},
                             'jointOrient': {'args': '<type __builtin__.bool>',
                                             'longname': 'jointOrient',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'jo'},
                             'normal': {'args': '<type __builtin__.int>',
                                        'longname': 'normal',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                             'preserveNormals': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preserveNormals',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'pn'},
                             'rotate': {'args': '<type __builtin__.bool>',
                                        'longname': 'rotate',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'r'},
                             'scale': {'args': '<type __builtin__.bool>',
                                       'longname': 'scale',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 's'},
                             'translate': {'args': '<type __builtin__.bool>',
                                           'longname': 'translate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 't'}},
                   'shortFlags': {'a': 'apply',
                                  'jo': 'jointOrient',
                                  'n': 'normal',
                                  'pn': 'preserveNormals',
                                  'r': 'rotate',
                                  's': 'scale',
                                  't': 'translate'},
                   'type': 'general'},
  'makeLive': {'flags': {'none': {'args': '<type __builtin__.bool>',
                                  'longname': 'none',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'n'}},
               'shortFlags': {'n': 'none'},
               'type': 'general'},
  'makePaintable': {'flags': {'activate': {'args': '<type __builtin__.bool>',
                                           'longname': 'activate',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'ac'},
                              'activateAll': {'args': '<type __builtin__.bool>',
                                              'longname': 'activateAll',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'aca'},
                              'altAttribute': {'args': '<type __builtin__.str>',
                                               'longname': 'altAttribute',
                                               'modes': ['create',
                                                         'query',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'aa'},
                              'attrType': {'args': '<type __builtin__.str>',
                                           'longname': 'attrType',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'at'},
                              'clearAll': {'args': '<type __builtin__.bool>',
                                           'longname': 'clearAll',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'ca'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'remove': {'args': '<type __builtin__.bool>',
                                         'longname': 'remove',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'rm'},
                              'shapeMode': {'args': '<type __builtin__.str>',
                                            'longname': 'shapeMode',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'sm'},
                              'uiName': {'args': '<type __builtin__.str>',
                                         'longname': 'uiName',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'ui'}},
                    'shortFlags': {'aa': 'altAttribute',
                                   'ac': 'activate',
                                   'aca': 'activateAll',
                                   'at': 'attrType',
                                   'ca': 'clearAll',
                                   'q': 'query',
                                   'rm': 'remove',
                                   'sm': 'shapeMode',
                                   'ui': 'uiName'},
                    'type': 'general'},
  'makeSingleSurface': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                              'longname': 'caching',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cch'},
                                  'chordHeight': {'args': '<type __builtin__.float>',
                                                  'longname': 'chordHeight',
                                                  'numArgs': 1,
                                                  'shortname': 'cht'},
                                  'chordHeightRatio': {'args': '<type __builtin__.float>',
                                                       'longname': 'chordHeightRatio',
                                                       'numArgs': 1,
                                                       'shortname': 'chr'},
                                  'constructionHistory': {'args': '<type __builtin__.bool>',
                                                          'longname': 'constructionHistory',
                                                          'modes': ['create'],
                                                          'numArgs': 1,
                                                          'shortname': 'ch'},
                                  'delta': {'args': '<type __builtin__.float>',
                                            'longname': 'delta',
                                            'numArgs': 1,
                                            'shortname': 'd'},
                                  'edgeSwap': {'args': '<type __builtin__.bool>',
                                               'longname': 'edgeSwap',
                                               'numArgs': 1,
                                               'shortname': 'es'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'format': {'args': '<type __builtin__.int>',
                                             'longname': 'format',
                                             'numArgs': 1,
                                             'shortname': 'f'},
                                  'fractionalTolerance': {'args': '<type __builtin__.float>',
                                                          'longname': 'fractionalTolerance',
                                                          'numArgs': 1,
                                                          'shortname': 'ft'},
                                  'frozen': {'args': '<type __builtin__.bool>',
                                             'longname': 'frozen',
                                             'numArgs': 1,
                                             'shortname': 'fzn'},
                                  'matchNormalDir': {'args': '<type __builtin__.bool>',
                                                     'longname': 'matchNormalDir',
                                                     'numArgs': 1,
                                                     'shortname': 'mnd'},
                                  'minEdgeLength': {'args': '<type __builtin__.float>',
                                                    'longname': 'minEdgeLength',
                                                    'numArgs': 1,
                                                    'shortname': 'mel'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'nodeState': {'args': '<type __builtin__.int>',
                                                'longname': 'nodeState',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nds'},
                                  'normalizeTrimmedUVRange': {'args': '<type __builtin__.bool>',
                                                              'longname': 'normalizeTrimmedUVRange',
                                                              'numArgs': 1,
                                                              'shortname': 'ntr'},
                                  'object': {'args': '<type __builtin__.bool>',
                                             'longname': 'object',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'o'},
                                  'polygonCount': {'args': '<type __builtin__.int>',
                                                   'longname': 'polygonCount',
                                                   'numArgs': 1,
                                                   'shortname': 'pc'},
                                  'polygonType': {'args': '<type __builtin__.int>',
                                                  'longname': 'polygonType',
                                                  'numArgs': 1,
                                                  'shortname': 'pt'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'stitchTolerance': {'args': '<type __builtin__.float>',
                                                      'longname': 'stitchTolerance',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'st'},
                                  'uNumber': {'args': '<type __builtin__.int>',
                                              'longname': 'uNumber',
                                              'numArgs': 1,
                                              'shortname': 'un'},
                                  'uType': {'args': '<type __builtin__.int>',
                                            'longname': 'uType',
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                                  'useChordHeight': {'args': '<type __builtin__.bool>',
                                                     'longname': 'useChordHeight',
                                                     'numArgs': 1,
                                                     'shortname': 'uch'},
                                  'useChordHeightRatio': {'args': '<type __builtin__.bool>',
                                                          'longname': 'useChordHeightRatio',
                                                          'numArgs': 1,
                                                          'shortname': 'ucr'},
                                  'vNumber': {'args': '<type __builtin__.int>',
                                              'longname': 'vNumber',
                                              'numArgs': 1,
                                              'shortname': 'vn'},
                                  'vType': {'args': '<type __builtin__.int>',
                                            'longname': 'vType',
                                            'numArgs': 1,
                                            'shortname': 'vt'}},
                        'shortFlags': {'cch': 'caching',
                                       'ch': 'constructionHistory',
                                       'chr': 'chordHeightRatio',
                                       'cht': 'chordHeight',
                                       'd': 'delta',
                                       'e': 'edit',
                                       'es': 'edgeSwap',
                                       'f': 'format',
                                       'ft': 'fractionalTolerance',
                                       'fzn': 'frozen',
                                       'mel': 'minEdgeLength',
                                       'mnd': 'matchNormalDir',
                                       'n': 'name',
                                       'nds': 'nodeState',
                                       'ntr': 'normalizeTrimmedUVRange',
                                       'o': 'object',
                                       'pc': 'polygonCount',
                                       'pt': 'polygonType',
                                       'q': 'query',
                                       'st': 'stitchTolerance',
                                       'uch': 'useChordHeight',
                                       'ucr': 'useChordHeightRatio',
                                       'un': 'uNumber',
                                       'ut': 'uType',
                                       'vn': 'vNumber',
                                       'vt': 'vType'},
                        'type': 'modeling'},
  'makebot': {'flags': {'checkdepends': {'args': '<type __builtin__.bool>',
                                         'longname': 'checkdepends',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'c'},
                        'checkres': {'args': '<type __builtin__.int>',
                                     'longname': 'checkres',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'r'},
                        'input': {'args': '<type __builtin__.str>',
                                  'longname': 'input',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'i'},
                        'nooverwrite': {'args': '<type __builtin__.bool>',
                                        'longname': 'nooverwrite',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'nov'},
                        'output': {'args': '<type __builtin__.str>',
                                   'longname': 'output',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'o'},
                        'verbose': {'args': '<type __builtin__.bool>',
                                    'longname': 'verbose',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'v'}},
              'shortFlags': {'c': 'checkdepends',
                             'i': 'input',
                             'nov': 'nooverwrite',
                             'o': 'output',
                             'r': 'checkres',
                             'v': 'verbose'},
              'type': 'rendering'},
  'manipComponentPivot': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'manipComponentUpdate': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'manipMoveContext': {'callbackFlags': ['preDragCommand',
                                         'prd',
                                         'postCommand',
                                         'psc',
                                         'preCommand',
                                         'prc',
                                         'postDragCommand',
                                         'pod'],
                       'flags': {'activeHandle': {'args': '<type __builtin__.int>',
                                                  'longname': 'activeHandle',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ah'},
                                 'activeHandleNormal': {'args': '<type __builtin__.int>',
                                                        'longname': 'activeHandleNormal',
                                                        'modes': ['query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ahn'},
                                 'alignAlong': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'alignAlong',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'aa'},
                                 'constrainAlongNormal': {'args': '<type __builtin__.bool>',
                                                          'longname': 'constrainAlongNormal',
                                                          'modes': ['query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'xn'},
                                 'currentActiveHandle': {'args': '<type __builtin__.int>',
                                                         'longname': 'currentActiveHandle',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'cah'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'editPivotMode': {'args': '<type __builtin__.bool>',
                                                   'longname': 'editPivotMode',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'epm'},
                                 'editPivotPosition': {'args': '<type __builtin__.bool>',
                                                       'longname': 'editPivotPosition',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'epp'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'image1': {'args': '<type __builtin__.str>',
                                            'longname': 'image1',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i1'},
                                 'image2': {'args': '<type __builtin__.str>',
                                            'longname': 'image2',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i2'},
                                 'image3': {'args': '<type __builtin__.str>',
                                            'longname': 'image3',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i3'},
                                 'interactiveUpdate': {'args': '<type __builtin__.bool>',
                                                       'longname': 'interactiveUpdate',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'iu'},
                                 'lastMode': {'args': '<type __builtin__.int>',
                                              'longname': 'lastMode',
                                              'modes': ['query'],
                                              'numArgs': 1,
                                              'shortname': 'lm'},
                                 'manipVisible': {'args': '<type __builtin__.bool>',
                                                  'longname': 'manipVisible',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'vis'},
                                 'mode': {'args': '<type __builtin__.int>',
                                          'longname': 'mode',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'm'},
                                 'orientAxes': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'orientAxes',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 3,
                                                'shortname': 'oa'},
                                 'orientJoint': {'args': '<type __builtin__.str>',
                                                 'longname': 'orientJoint',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'oj'},
                                 'orientJointEnabled': {'args': '<type __builtin__.bool>',
                                                        'longname': 'orientJointEnabled',
                                                        'modes': ['query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'oje'},
                                 'orientObject': {'args': '<type __builtin__.str>',
                                                  'longname': 'orientObject',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'oo'},
                                 'orientTowards': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'orientTowards',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'ot'},
                                 'pinPivot': {'args': '<type __builtin__.bool>',
                                              'longname': 'pinPivot',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pin'},
                                 'pivotOriHandle': {'args': '<type __builtin__.bool>',
                                                    'longname': 'pivotOriHandle',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'poh'},
                                 'position': {'args': '<type __builtin__.bool>',
                                              'longname': 'position',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'p'},
                                 'postCommand': {'args': 'script',
                                                 'longname': 'postCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'psc'},
                                 'postDragCommand': {'args': ['script',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'postDragCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'pod'},
                                 'preCommand': {'args': 'script',
                                                'longname': 'preCommand',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'prc'},
                                 'preDragCommand': {'args': ['script',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'preDragCommand',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 2,
                                                    'shortname': 'prd'},
                                 'preserveChildPosition': {'args': '<type __builtin__.bool>',
                                                           'longname': 'preserveChildPosition',
                                                           'modes': ['query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'pcp'},
                                 'preserveUV': {'args': '<type __builtin__.bool>',
                                                'longname': 'preserveUV',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'puv'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'reflection': {'args': '<type __builtin__.bool>',
                                                'longname': 'reflection',
                                                'modes': [],
                                                'numArgs': 1,
                                                'shortname': 'rfl'},
                                 'reflectionAbout': {'args': '<type __builtin__.int>',
                                                     'longname': 'reflectionAbout',
                                                     'modes': [],
                                                     'numArgs': 1,
                                                     'shortname': 'rab'},
                                 'reflectionAxis': {'args': '<type __builtin__.int>',
                                                    'longname': 'reflectionAxis',
                                                    'modes': [],
                                                    'numArgs': 1,
                                                    'shortname': 'rfa'},
                                 'reflectionTolerance': {'args': '<type __builtin__.float>',
                                                         'longname': 'reflectionTolerance',
                                                         'modes': [],
                                                         'numArgs': 1,
                                                         'shortname': 'rft'},
                                 'secondaryAxisOrient': {'args': '<type __builtin__.str>',
                                                         'longname': 'secondaryAxisOrient',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'sao'},
                                 'snap': {'args': '<type __builtin__.bool>',
                                          'longname': 'snap',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 's'},
                                 'snapComponentsRelative': {'args': '<type __builtin__.bool>',
                                                            'longname': 'snapComponentsRelative',
                                                            'modes': ['query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'scr'},
                                 'snapLiveFaceCenter': {'args': '<type __builtin__.bool>',
                                                        'longname': 'snapLiveFaceCenter',
                                                        'modes': ['query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'slf'},
                                 'snapLivePoint': {'args': '<type __builtin__.bool>',
                                                   'longname': 'snapLivePoint',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'slp'},
                                 'snapPivotOri': {'args': '<type __builtin__.bool>',
                                                  'longname': 'snapPivotOri',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'spo'},
                                 'snapPivotPos': {'args': '<type __builtin__.bool>',
                                                  'longname': 'snapPivotPos',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'spp'},
                                 'snapRelative': {'args': '<type __builtin__.bool>',
                                                  'longname': 'snapRelative',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sr'},
                                 'snapValue': {'args': '<type __builtin__.float>',
                                               'longname': 'snapValue',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sv'},
                                 'translate': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'translate',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'tr'},
                                 'tweakMode': {'args': '<type __builtin__.bool>',
                                               'longname': 'tweakMode',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'twk'},
                                 'xformConstraint': {'args': '<type __builtin__.str>',
                                                     'longname': 'xformConstraint',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'xc'}},
                       'shortFlags': {'aa': 'alignAlong',
                                      'ah': 'activeHandle',
                                      'ahn': 'activeHandleNormal',
                                      'cah': 'currentActiveHandle',
                                      'e': 'edit',
                                      'epm': 'editPivotMode',
                                      'epp': 'editPivotPosition',
                                      'ex': 'exists',
                                      'i1': 'image1',
                                      'i2': 'image2',
                                      'i3': 'image3',
                                      'iu': 'interactiveUpdate',
                                      'lm': 'lastMode',
                                      'm': 'mode',
                                      'oa': 'orientAxes',
                                      'oj': 'orientJoint',
                                      'oje': 'orientJointEnabled',
                                      'oo': 'orientObject',
                                      'ot': 'orientTowards',
                                      'p': 'position',
                                      'pcp': 'preserveChildPosition',
                                      'pin': 'pinPivot',
                                      'pod': 'postDragCommand',
                                      'poh': 'pivotOriHandle',
                                      'prc': 'preCommand',
                                      'prd': 'preDragCommand',
                                      'psc': 'postCommand',
                                      'puv': 'preserveUV',
                                      'q': 'query',
                                      'rab': 'reflectionAbout',
                                      'rfa': 'reflectionAxis',
                                      'rfl': 'reflection',
                                      'rft': 'reflectionTolerance',
                                      's': 'snap',
                                      'sao': 'secondaryAxisOrient',
                                      'scr': 'snapComponentsRelative',
                                      'slf': 'snapLiveFaceCenter',
                                      'slp': 'snapLivePoint',
                                      'spo': 'snapPivotOri',
                                      'spp': 'snapPivotPos',
                                      'sr': 'snapRelative',
                                      'sv': 'snapValue',
                                      'tr': 'translate',
                                      'twk': 'tweakMode',
                                      'vis': 'manipVisible',
                                      'xc': 'xformConstraint',
                                      'xn': 'constrainAlongNormal'},
                       'type': 'context'},
  'manipMoveLimitsCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'history': {'args': '<type __builtin__.bool>',
                                               'longname': 'history',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'}},
                         'shortFlags': {'ch': 'history',
                                        'e': 'edit',
                                        'ex': 'exists',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'n': 'name',
                                        'q': 'query'},
                         'type': 'context'},
  'manipOptions': {'flags': {'enableSmartDuplicate': {'args': '<type __builtin__.bool>',
                                                      'longname': 'enableSmartDuplicate',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'esd'},
                             'enableSmartExtrude': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableSmartExtrude',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'ese'},
                             'forceRefresh': {'args': '<type __builtin__.bool>',
                                              'longname': 'forceRefresh',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'fr'},
                             'handleSize': {'args': '<type __builtin__.float>',
                                            'longname': 'handleSize',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'hs'},
                             'hideManipOnCtrl': {'args': '<type __builtin__.bool>',
                                                 'longname': 'hideManipOnCtrl',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'hmc'},
                             'hideManipOnShift': {'args': '<type __builtin__.bool>',
                                                  'longname': 'hideManipOnShift',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'hms'},
                             'hideManipOnShiftCtrl': {'args': '<type __builtin__.bool>',
                                                      'longname': 'hideManipOnShiftCtrl',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'hsc'},
                             'linePick': {'args': '<type __builtin__.float>',
                                          'longname': 'linePick',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'lp'},
                             'lineSize': {'args': '<type __builtin__.float>',
                                          'longname': 'lineSize',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'ls'},
                             'middleMouseRepositioning': {'args': '<type __builtin__.bool>',
                                                          'longname': 'middleMouseRepositioning',
                                                          'modes': ['create',
                                                                    'query'],
                                                          'numArgs': 1,
                                                          'shortname': 'mm'},
                             'pivotRotateHandleOffset': {'args': '<type __builtin__.int>',
                                                         'longname': 'pivotRotateHandleOffset',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'pro'},
                             'planeHandleOffset': {'args': '<type __builtin__.int>',
                                                   'longname': 'planeHandleOffset',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'pho'},
                             'pointSize': {'args': '<type __builtin__.float>',
                                           'longname': 'pointSize',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'ps'},
                             'preselectHighlight': {'args': '<type __builtin__.bool>',
                                                    'longname': 'preselectHighlight',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'psh'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'refreshMode': {'args': '<type __builtin__.int>',
                                             'longname': 'refreshMode',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'rm'},
                             'relative': {'args': '<type __builtin__.bool>',
                                          'longname': 'relative',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'r'},
                             'rememberActiveHandle': {'args': '<type __builtin__.bool>',
                                                      'longname': 'rememberActiveHandle',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'rah'},
                             'rememberActiveHandleAfterToolSwitch': {'args': '<type __builtin__.bool>',
                                                                     'longname': 'rememberActiveHandleAfterToolSwitch',
                                                                     'modes': ['create',
                                                                               'query'],
                                                                     'numArgs': 1,
                                                                     'shortname': 'rhs'},
                             'scale': {'args': '<type __builtin__.float>',
                                       'longname': 'scale',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 's'},
                             'showExtrudeSliders': {'args': '<type __builtin__.bool>',
                                                    'longname': 'showExtrudeSliders',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'ses'},
                             'showPivotRotateHandle': {'args': '<type __builtin__.bool>',
                                                       'longname': 'showPivotRotateHandle',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'spr'},
                             'showPlaneHandles': {'args': '<type __builtin__.bool>',
                                                  'longname': 'showPlaneHandles',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'sph'},
                             'smartDuplicateType': {'args': '<type __builtin__.int>',
                                                    'longname': 'smartDuplicateType',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'sdt'}},
                   'shortFlags': {'esd': 'enableSmartDuplicate',
                                  'ese': 'enableSmartExtrude',
                                  'fr': 'forceRefresh',
                                  'hmc': 'hideManipOnCtrl',
                                  'hms': 'hideManipOnShift',
                                  'hs': 'handleSize',
                                  'hsc': 'hideManipOnShiftCtrl',
                                  'lp': 'linePick',
                                  'ls': 'lineSize',
                                  'mm': 'middleMouseRepositioning',
                                  'pho': 'planeHandleOffset',
                                  'pro': 'pivotRotateHandleOffset',
                                  'ps': 'pointSize',
                                  'psh': 'preselectHighlight',
                                  'q': 'query',
                                  'r': 'relative',
                                  'rah': 'rememberActiveHandle',
                                  'rhs': 'rememberActiveHandleAfterToolSwitch',
                                  'rm': 'refreshMode',
                                  's': 'scale',
                                  'sdt': 'smartDuplicateType',
                                  'ses': 'showExtrudeSliders',
                                  'sph': 'showPlaneHandles',
                                  'spr': 'showPivotRotateHandle'},
                   'type': 'modeling'},
  'manipPivot': {'flags': {'moveToolOri': {'args': '<type __builtin__.int>',
                                           'longname': 'moveToolOri',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'mto'},
                           'ori': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'ori',
                                   'modes': ['create', 'query'],
                                   'numArgs': 3,
                                   'shortname': 'o'},
                           'oriValid': {'args': '<type __builtin__.bool>',
                                        'longname': 'oriValid',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'ov'},
                           'orientation': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'orientation',
                                           'numArgs': 3,
                                           'shortname': 'o'},
                           'pinPivot': {'args': '<type __builtin__.bool>',
                                        'longname': 'pinPivot',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'pin'},
                           'pos': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'pos',
                                   'modes': ['create', 'query'],
                                   'numArgs': 3,
                                   'shortname': 'p'},
                           'posValid': {'args': '<type __builtin__.bool>',
                                        'longname': 'posValid',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'pv'},
                           'position': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'position',
                                        'numArgs': 3,
                                        'shortname': 'p'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'reset': {'args': '<type __builtin__.bool>',
                                     'longname': 'reset',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'r'},
                           'resetOri': {'args': '<type __builtin__.bool>',
                                        'longname': 'resetOri',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ro'},
                           'resetPos': {'args': '<type __builtin__.bool>',
                                        'longname': 'resetPos',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'rp'},
                           'rotateToolOri': {'args': '<type __builtin__.int>',
                                             'longname': 'rotateToolOri',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'rto'},
                           'scaleToolOri': {'args': '<type __builtin__.int>',
                                            'longname': 'scaleToolOri',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'sto'},
                           'snapOri': {'args': '<type __builtin__.bool>',
                                       'longname': 'snapOri',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'so'},
                           'snapPos': {'args': '<type __builtin__.bool>',
                                       'longname': 'snapPos',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'sp'},
                           'valid': {'args': '<type __builtin__.bool>',
                                     'longname': 'valid',
                                     'modes': ['query'],
                                     'numArgs': 1,
                                     'shortname': 'v'}},
                 'shortFlags': {'mto': 'moveToolOri',
                                'o': 'orientation',
                                'ov': 'oriValid',
                                'p': 'position',
                                'pin': 'pinPivot',
                                'pv': 'posValid',
                                'q': 'query',
                                'r': 'reset',
                                'ro': 'resetOri',
                                'rp': 'resetPos',
                                'rto': 'rotateToolOri',
                                'so': 'snapOri',
                                'sp': 'snapPos',
                                'sto': 'scaleToolOri',
                                'v': 'valid'},
                 'type': 'modeling'},
  'manipRotateContext': {'callbackFlags': ['preDragCommand',
                                           'prd',
                                           'postCommand',
                                           'psc',
                                           'preCommand',
                                           'prc',
                                           'postDragCommand',
                                           'pod'],
                         'flags': {'activeHandle': {'args': '<type __builtin__.int>',
                                                    'longname': 'activeHandle',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ah'},
                                   'alignAlong': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'alignAlong',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'aa'},
                                   'centerTrackball': {'args': '<type __builtin__.bool>',
                                                       'longname': 'centerTrackball',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ctb'},
                                   'constrainAlongNormal': {'args': '<type __builtin__.bool>',
                                                            'longname': 'constrainAlongNormal',
                                                            'modes': ['query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'xn'},
                                   'currentActiveHandle': {'args': '<type __builtin__.int>',
                                                           'longname': 'currentActiveHandle',
                                                           'modes': ['query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'cah'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'editPivotMode': {'args': '<type __builtin__.bool>',
                                                     'longname': 'editPivotMode',
                                                     'modes': ['query'],
                                                     'numArgs': 1,
                                                     'shortname': 'epm'},
                                   'editPivotPosition': {'args': '<type __builtin__.bool>',
                                                         'longname': 'editPivotPosition',
                                                         'modes': ['query'],
                                                         'numArgs': 0,
                                                         'shortname': 'epp'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'lastMode': {'args': '<type __builtin__.int>',
                                                'longname': 'lastMode',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'lm'},
                                   'manipVisible': {'args': '<type __builtin__.bool>',
                                                    'longname': 'manipVisible',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'vis'},
                                   'mode': {'args': '<type __builtin__.int>',
                                            'longname': 'mode',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'm'},
                                   'modifyTranslation': {'args': '<type __builtin__.bool>',
                                                         'longname': 'modifyTranslation',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'mt'},
                                   'orientAxes': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'orientAxes',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'oa'},
                                   'orientObject': {'args': '<type __builtin__.str>',
                                                    'longname': 'orientObject',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'oo'},
                                   'orientTowards': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'orientTowards',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'ot'},
                                   'pinPivot': {'args': '<type __builtin__.bool>',
                                                'longname': 'pinPivot',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'pin'},
                                   'pivotOriHandle': {'args': '<type __builtin__.bool>',
                                                      'longname': 'pivotOriHandle',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'poh'},
                                   'position': {'args': '<type __builtin__.bool>',
                                                'longname': 'position',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'p'},
                                   'postCommand': {'args': 'script',
                                                   'longname': 'postCommand',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'psc'},
                                   'postDragCommand': {'args': ['script',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'postDragCommand',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 2,
                                                       'shortname': 'pod'},
                                   'preCommand': {'args': 'script',
                                                  'longname': 'preCommand',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'prc'},
                                   'preDragCommand': {'args': ['script',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'preDragCommand',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'prd'},
                                   'preserveChildPosition': {'args': '<type __builtin__.bool>',
                                                             'longname': 'preserveChildPosition',
                                                             'modes': ['query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'pcp'},
                                   'preserveUV': {'args': '<type __builtin__.bool>',
                                                  'longname': 'preserveUV',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'puv'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'reflection': {'args': '<type __builtin__.bool>',
                                                  'longname': 'reflection',
                                                  'modes': [],
                                                  'numArgs': 1,
                                                  'shortname': 'rfl'},
                                   'reflectionAbout': {'args': '<type __builtin__.int>',
                                                       'longname': 'reflectionAbout',
                                                       'modes': [],
                                                       'numArgs': 1,
                                                       'shortname': 'rab'},
                                   'reflectionAxis': {'args': '<type __builtin__.int>',
                                                      'longname': 'reflectionAxis',
                                                      'modes': [],
                                                      'numArgs': 1,
                                                      'shortname': 'rfa'},
                                   'reflectionTolerance': {'args': '<type __builtin__.float>',
                                                           'longname': 'reflectionTolerance',
                                                           'modes': [],
                                                           'numArgs': 1,
                                                           'shortname': 'rft'},
                                   'rotate': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'rotate',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'ro'},
                                   'snap': {'args': '<type __builtin__.bool>',
                                            'longname': 'snap',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 's'},
                                   'snapPivotOri': {'args': '<type __builtin__.bool>',
                                                    'longname': 'snapPivotOri',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'spo'},
                                   'snapPivotPos': {'args': '<type __builtin__.bool>',
                                                    'longname': 'snapPivotPos',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'spp'},
                                   'snapRelative': {'args': '<type __builtin__.bool>',
                                                    'longname': 'snapRelative',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sr'},
                                   'snapValue': {'args': '<type __builtin__.float>',
                                                 'longname': 'snapValue',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sv'},
                                   'tweakMode': {'args': '<type __builtin__.bool>',
                                                 'longname': 'tweakMode',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'twk'},
                                   'useCenterPivot': {'args': '<type __builtin__.bool>',
                                                      'longname': 'useCenterPivot',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ucp'},
                                   'useManipPivot': {'args': '<type __builtin__.bool>',
                                                     'longname': 'useManipPivot',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ump'},
                                   'useObjectPivot': {'args': '<type __builtin__.bool>',
                                                      'longname': 'useObjectPivot',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'uop'},
                                   'xformConstraint': {'args': '<type __builtin__.str>',
                                                       'longname': 'xformConstraint',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'xc'}},
                         'shortFlags': {'aa': 'alignAlong',
                                        'ah': 'activeHandle',
                                        'cah': 'currentActiveHandle',
                                        'ctb': 'centerTrackball',
                                        'e': 'edit',
                                        'epm': 'editPivotMode',
                                        'epp': 'editPivotPosition',
                                        'ex': 'exists',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'lm': 'lastMode',
                                        'm': 'mode',
                                        'mt': 'modifyTranslation',
                                        'oa': 'orientAxes',
                                        'oo': 'orientObject',
                                        'ot': 'orientTowards',
                                        'p': 'position',
                                        'pcp': 'preserveChildPosition',
                                        'pin': 'pinPivot',
                                        'pod': 'postDragCommand',
                                        'poh': 'pivotOriHandle',
                                        'prc': 'preCommand',
                                        'prd': 'preDragCommand',
                                        'psc': 'postCommand',
                                        'puv': 'preserveUV',
                                        'q': 'query',
                                        'rab': 'reflectionAbout',
                                        'rfa': 'reflectionAxis',
                                        'rfl': 'reflection',
                                        'rft': 'reflectionTolerance',
                                        'ro': 'rotate',
                                        's': 'snap',
                                        'spo': 'snapPivotOri',
                                        'spp': 'snapPivotPos',
                                        'sr': 'snapRelative',
                                        'sv': 'snapValue',
                                        'twk': 'tweakMode',
                                        'ucp': 'useCenterPivot',
                                        'ump': 'useManipPivot',
                                        'uop': 'useObjectPivot',
                                        'vis': 'manipVisible',
                                        'xc': 'xformConstraint',
                                        'xn': 'constrainAlongNormal'},
                         'type': 'context'},
  'manipRotateLimitsCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'exists': {'args': '<type __builtin__.bool>',
                                                'longname': 'exists',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ex'},
                                     'history': {'args': '<type __builtin__.bool>',
                                                 'longname': 'history',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                                     'image1': {'args': '<type __builtin__.str>',
                                                'longname': 'image1',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'i1'},
                                     'image2': {'args': '<type __builtin__.str>',
                                                'longname': 'image2',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'i2'},
                                     'image3': {'args': '<type __builtin__.str>',
                                                'longname': 'image3',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'i3'},
                                     'name': {'args': '<type __builtin__.str>',
                                              'longname': 'name',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'n'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'}},
                           'shortFlags': {'ch': 'history',
                                          'e': 'edit',
                                          'ex': 'exists',
                                          'i1': 'image1',
                                          'i2': 'image2',
                                          'i3': 'image3',
                                          'n': 'name',
                                          'q': 'query'},
                           'type': 'context'},
  'manipScaleContext': {'callbackFlags': ['preDragCommand',
                                          'prd',
                                          'postDragCommand',
                                          'pod',
                                          'preCommand',
                                          'prc',
                                          'postCommand',
                                          'psc'],
                        'flags': {'activeHandle': {'args': '<type __builtin__.int>',
                                                   'longname': 'activeHandle',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ah'},
                                  'alignAlong': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'alignAlong',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'aa'},
                                  'constrainAlongNormal': {'args': '<type __builtin__.bool>',
                                                           'longname': 'constrainAlongNormal',
                                                           'modes': ['query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'xn'},
                                  'currentActiveHandle': {'args': '<type __builtin__.int>',
                                                          'longname': 'currentActiveHandle',
                                                          'modes': ['query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'cah'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'editPivotMode': {'args': '<type __builtin__.bool>',
                                                    'longname': 'editPivotMode',
                                                    'modes': ['query'],
                                                    'numArgs': 1,
                                                    'shortname': 'epm'},
                                  'editPivotPosition': {'args': '<type __builtin__.bool>',
                                                        'longname': 'editPivotPosition',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'epp'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'image1': {'args': '<type __builtin__.str>',
                                             'longname': 'image1',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i1'},
                                  'image2': {'args': '<type __builtin__.str>',
                                             'longname': 'image2',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i2'},
                                  'image3': {'args': '<type __builtin__.str>',
                                             'longname': 'image3',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i3'},
                                  'lastMode': {'args': '<type __builtin__.int>',
                                               'longname': 'lastMode',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'lm'},
                                  'manipVisible': {'args': '<type __builtin__.bool>',
                                                   'longname': 'manipVisible',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'vis'},
                                  'mode': {'args': '<type __builtin__.int>',
                                           'longname': 'mode',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'm'},
                                  'orientAxes': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'orientAxes',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'oa'},
                                  'orientObject': {'args': '<type __builtin__.str>',
                                                   'longname': 'orientObject',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'oo'},
                                  'orientTowards': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'orientTowards',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'ot'},
                                  'pinPivot': {'args': '<type __builtin__.bool>',
                                               'longname': 'pinPivot',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'pin'},
                                  'pivotOriHandle': {'args': '<type __builtin__.bool>',
                                                     'longname': 'pivotOriHandle',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'poh'},
                                  'position': {'args': '<type __builtin__.bool>',
                                               'longname': 'position',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'p'},
                                  'postCommand': {'args': 'script',
                                                  'longname': 'postCommand',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'psc'},
                                  'postDragCommand': {'args': ['script',
                                                               '<type __builtin__.str>'],
                                                      'longname': 'postDragCommand',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'pod'},
                                  'preCommand': {'args': 'script',
                                                 'longname': 'preCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'prc'},
                                  'preDragCommand': {'args': ['script',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'preDragCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'prd'},
                                  'preserveChildPosition': {'args': '<type __builtin__.bool>',
                                                            'longname': 'preserveChildPosition',
                                                            'modes': ['query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'pcp'},
                                  'preserveUV': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preserveUV',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'puv'},
                                  'preventNegativeScale': {'args': '<type __builtin__.bool>',
                                                           'longname': 'preventNegativeScale',
                                                           'modes': ['query'],
                                                           'numArgs': 1,
                                                           'shortname': 'pns'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'reflection': {'args': '<type __builtin__.bool>',
                                                 'longname': 'reflection',
                                                 'modes': [],
                                                 'numArgs': 1,
                                                 'shortname': 'rfl'},
                                  'reflectionAbout': {'args': '<type __builtin__.int>',
                                                      'longname': 'reflectionAbout',
                                                      'modes': [],
                                                      'numArgs': 1,
                                                      'shortname': 'rab'},
                                  'reflectionAxis': {'args': '<type __builtin__.int>',
                                                     'longname': 'reflectionAxis',
                                                     'modes': [],
                                                     'numArgs': 1,
                                                     'shortname': 'rfa'},
                                  'reflectionTolerance': {'args': '<type __builtin__.float>',
                                                          'longname': 'reflectionTolerance',
                                                          'modes': [],
                                                          'numArgs': 1,
                                                          'shortname': 'rft'},
                                  'scale': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'scale',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 3,
                                            'shortname': 'sc'},
                                  'snap': {'args': '<type __builtin__.bool>',
                                           'longname': 'snap',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 's'},
                                  'snapPivotOri': {'args': '<type __builtin__.bool>',
                                                   'longname': 'snapPivotOri',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'spo'},
                                  'snapPivotPos': {'args': '<type __builtin__.bool>',
                                                   'longname': 'snapPivotPos',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'spp'},
                                  'snapRelative': {'args': '<type __builtin__.bool>',
                                                   'longname': 'snapRelative',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sr'},
                                  'snapValue': {'args': '<type __builtin__.float>',
                                                'longname': 'snapValue',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sv'},
                                  'tweakMode': {'args': '<type __builtin__.bool>',
                                                'longname': 'tweakMode',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'twk'},
                                  'useManipPivot': {'args': '<type __builtin__.bool>',
                                                    'longname': 'useManipPivot',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ump'},
                                  'useObjectPivot': {'args': '<type __builtin__.bool>',
                                                     'longname': 'useObjectPivot',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'uop'},
                                  'xformConstraint': {'args': '<type __builtin__.str>',
                                                      'longname': 'xformConstraint',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'xc'}},
                        'shortFlags': {'aa': 'alignAlong',
                                       'ah': 'activeHandle',
                                       'cah': 'currentActiveHandle',
                                       'e': 'edit',
                                       'epm': 'editPivotMode',
                                       'epp': 'editPivotPosition',
                                       'ex': 'exists',
                                       'i1': 'image1',
                                       'i2': 'image2',
                                       'i3': 'image3',
                                       'lm': 'lastMode',
                                       'm': 'mode',
                                       'oa': 'orientAxes',
                                       'oo': 'orientObject',
                                       'ot': 'orientTowards',
                                       'p': 'position',
                                       'pcp': 'preserveChildPosition',
                                       'pin': 'pinPivot',
                                       'pns': 'preventNegativeScale',
                                       'pod': 'postDragCommand',
                                       'poh': 'pivotOriHandle',
                                       'prc': 'preCommand',
                                       'prd': 'preDragCommand',
                                       'psc': 'postCommand',
                                       'puv': 'preserveUV',
                                       'q': 'query',
                                       'rab': 'reflectionAbout',
                                       'rfa': 'reflectionAxis',
                                       'rfl': 'reflection',
                                       'rft': 'reflectionTolerance',
                                       's': 'snap',
                                       'sc': 'scale',
                                       'spo': 'snapPivotOri',
                                       'spp': 'snapPivotPos',
                                       'sr': 'snapRelative',
                                       'sv': 'snapValue',
                                       'twk': 'tweakMode',
                                       'ump': 'useManipPivot',
                                       'uop': 'useObjectPivot',
                                       'vis': 'manipVisible',
                                       'xc': 'xformConstraint',
                                       'xn': 'constrainAlongNormal'},
                        'type': 'context'},
  'manipScaleLimitsCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'history': {'args': '<type __builtin__.bool>',
                                                'longname': 'history',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'}},
                          'shortFlags': {'ch': 'history',
                                         'e': 'edit',
                                         'ex': 'exists',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'n': 'name',
                                         'q': 'query'},
                          'type': 'context'},
  'marker': {'flags': {'attach': {'args': '<type __builtin__.bool>',
                                  'longname': 'attach',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'a'},
                       'detach': {'args': '<type __builtin__.bool>',
                                  'longname': 'detach',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'd'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'frontTwist': {'args': '<type __builtin__.float>',
                                      'longname': 'frontTwist',
                                      'modes': ['query'],
                                      'numArgs': 1,
                                      'shortname': 'ft'},
                       'orientationMarker': {'args': '<type __builtin__.bool>',
                                             'longname': 'orientationMarker',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'om'},
                       'positionMarker': {'args': '<type __builtin__.bool>',
                                          'longname': 'positionMarker',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'pm'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'sideTwist': {'args': '<type __builtin__.float>',
                                     'longname': 'sideTwist',
                                     'modes': ['query'],
                                     'numArgs': 1,
                                     'shortname': 'st'},
                       'time': {'args': 'time',
                                'longname': 'time',
                                'modes': ['query'],
                                'numArgs': 1,
                                'shortname': 't'},
                       'upTwist': {'args': '<type __builtin__.float>',
                                   'longname': 'upTwist',
                                   'modes': ['query'],
                                   'numArgs': 1,
                                   'shortname': 'ut'},
                       'valueU': {'args': '<type __builtin__.float>',
                                  'longname': 'valueU',
                                  'modes': ['query'],
                                  'numArgs': 1,
                                  'shortname': 'u'}},
             'shortFlags': {'a': 'attach',
                            'd': 'detach',
                            'e': 'edit',
                            'ft': 'frontTwist',
                            'om': 'orientationMarker',
                            'pm': 'positionMarker',
                            'q': 'query',
                            'st': 'sideTwist',
                            't': 'time',
                            'u': 'valueU',
                            'ut': 'upTwist'},
             'type': 'animation'},
  'matchTransform': {'flags': {'pivots': {'args': '<type __builtin__.bool>',
                                          'longname': 'pivots',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'piv'},
                               'position': {'args': '<type __builtin__.bool>',
                                            'longname': 'position',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'pos'},
                               'rotation': {'args': '<type __builtin__.bool>',
                                            'longname': 'rotation',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'rot'},
                               'scale': {'args': '<type __builtin__.bool>',
                                         'longname': 'scale',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'scl'}},
                     'shortFlags': {'piv': 'pivots',
                                    'pos': 'position',
                                    'rot': 'rotation',
                                    'scl': 'scale'},
                     'type': 'general'},
  'mateCtx': {'flags': {}, 'shortFlags': {}, 'type': 'context'},
  'mayaDpiSetting': {'flags': {'mode': {'args': '<type __builtin__.int>',
                                        'longname': 'mode',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'realScaleValue': {'args': '<type __builtin__.bool>',
                                                  'longname': 'realScaleValue',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'rsv'},
                               'scaleValue': {'args': '<type __builtin__.float>',
                                              'longname': 'scaleValue',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'sv'},
                               'systemDpi': {'args': '<type __builtin__.bool>',
                                             'longname': 'systemDpi',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'sd'}},
                     'shortFlags': {'m': 'mode',
                                    'q': 'query',
                                    'rsv': 'realScaleValue',
                                    'sd': 'systemDpi',
                                    'sv': 'scaleValue'},
                     'type': 'windows'},
  'mayaDpiSettingAction': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'mayaHasRenderSetup': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'enableCurrentSession': {'args': '<type __builtin__.bool>',
                                                            'longname': 'enableCurrentSession',
                                                            'modes': ['query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ecs'},
                                   'enableDuringTests': {'args': '<type __builtin__.bool>',
                                                         'longname': 'enableDuringTests',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'edt'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'}},
                         'shortFlags': {'e': 'edit',
                                        'ecs': 'enableCurrentSession',
                                        'edt': 'enableDuringTests',
                                        'q': 'query'},
                         'type': 'rendering'},
  'mayaPreviewRenderIntoNewWindow': {'type': 'runtime'},
  'melInfo': {'flags': {}, 'shortFlags': {}, 'type': 'system'},
  'melOptions': {'flags': {'duplicateVariableWarnings': {'args': '<type __builtin__.bool>',
                                                         'longname': 'duplicateVariableWarnings',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'dvw'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'dvw': 'duplicateVariableWarnings',
                                'q': 'query'},
                 'type': 'language'},
  'memory': {'flags': {'adjustedVirtualMemory': {'args': '<type __builtin__.bool>',
                                                 'longname': 'adjustedVirtualMemory',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'av'},
                       'asFloat': {'args': '<type __builtin__.bool>',
                                   'longname': 'asFloat',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'af'},
                       'debug': {'args': '<type __builtin__.bool>',
                                 'longname': 'debug',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'dbg'},
                       'freeMemory': {'args': '<type __builtin__.bool>',
                                      'longname': 'freeMemory',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'fr'},
                       'gigaByte': {'args': '<type __builtin__.bool>',
                                    'longname': 'gigaByte',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'gb'},
                       'heapMemory': {'args': '<type __builtin__.bool>',
                                      'longname': 'heapMemory',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'he'},
                       'kiloByte': {'args': '<type __builtin__.bool>',
                                    'longname': 'kiloByte',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'kb'},
                       'megaByte': {'args': '<type __builtin__.bool>',
                                    'longname': 'megaByte',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'mb'},
                       'pageFaults': {'args': '<type __builtin__.bool>',
                                      'longname': 'pageFaults',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'pf'},
                       'pageReclaims': {'args': '<type __builtin__.bool>',
                                        'longname': 'pageReclaims',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'pr'},
                       'physicalMemory': {'args': '<type __builtin__.bool>',
                                          'longname': 'physicalMemory',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'phy'},
                       'processResidentMemory': {'args': '<type __builtin__.bool>',
                                                 'longname': 'processResidentMemory',
                                                 'numArgs': 0,
                                                 'shortname': 'prm'},
                       'processVirtualMemory': {'args': '<type __builtin__.bool>',
                                                'longname': 'processVirtualMemory',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'pv'},
                       'summary': {'args': '<type __builtin__.bool>',
                                   'longname': 'summary',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'sum'},
                       'swapFree': {'args': '<type __builtin__.bool>',
                                    'longname': 'swapFree',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'swf'},
                       'swapLogical': {'args': '<type __builtin__.bool>',
                                       'longname': 'swapLogical',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'swl'},
                       'swapMax': {'args': '<type __builtin__.bool>',
                                   'longname': 'swapMax',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'swm'},
                       'swapPhysical': {'args': '<type __builtin__.bool>',
                                        'longname': 'swapPhysical',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'swp'},
                       'swapReserved': {'args': '<type __builtin__.bool>',
                                        'longname': 'swapReserved',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'swr'},
                       'swapVirtual': {'args': '<type __builtin__.bool>',
                                       'longname': 'swapVirtual',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'swv'},
                       'swaps': {'args': '<type __builtin__.bool>',
                                 'longname': 'swaps',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'sw'}},
             'shortFlags': {'af': 'asFloat',
                            'av': 'adjustedVirtualMemory',
                            'dbg': 'debug',
                            'fr': 'freeMemory',
                            'gb': 'gigaByte',
                            'he': 'heapMemory',
                            'kb': 'kiloByte',
                            'mb': 'megaByte',
                            'pf': 'pageFaults',
                            'phy': 'physicalMemory',
                            'pr': 'pageReclaims',
                            'prm': 'processResidentMemory',
                            'pv': 'processVirtualMemory',
                            'sum': 'summary',
                            'sw': 'swaps',
                            'swf': 'swapFree',
                            'swl': 'swapLogical',
                            'swm': 'swapMax',
                            'swp': 'swapPhysical',
                            'swr': 'swapReserved',
                            'swv': 'swapVirtual'},
             'type': 'system'},
  'memoryDiag': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'menu': {'callbackFlags': ['postMenuCommandOnce',
                             'pmo',
                             'postMenuCommand',
                             'pmc'],
           'flags': {'allowOptionBoxes': {'args': '<type __builtin__.bool>',
                                          'longname': 'allowOptionBoxes',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'aob'},
                     'defineTemplate': {'args': '<type __builtin__.str>',
                                        'longname': 'defineTemplate',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'dt'},
                     'deleteAllItems': {'args': '<type __builtin__.bool>',
                                        'longname': 'deleteAllItems',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'dai'},
                     'docTag': {'args': '<type __builtin__.str>',
                                'longname': 'docTag',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'dtg'},
                     'edit': {'args': '<type __builtin__.bool>',
                              'longname': 'edit',
                              'numArgs': 0,
                              'shortname': 'e'},
                     'enable': {'args': '<type __builtin__.bool>',
                                'longname': 'enable',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'en'},
                     'exists': {'args': '<type __builtin__.bool>',
                                'longname': 'exists',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'ex'},
                     'familyImage': {'args': '<type __builtin__.str>',
                                     'longname': 'familyImage',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'fi'},
                     'helpMenu': {'args': '<type __builtin__.bool>',
                                  'longname': 'helpMenu',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'hm'},
                     'itemArray': {'args': '<type __builtin__.bool>',
                                   'longname': 'itemArray',
                                   'modes': ['query'],
                                   'numArgs': 0,
                                   'shortname': 'ia'},
                     'label': {'args': '<type __builtin__.str>',
                               'longname': 'label',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'l'},
                     'ltVersion': {'args': '<type __builtin__.str>',
                                   'longname': 'ltVersion',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'lt'},
                     'mnemonic': {'args': '<type __builtin__.str>',
                                  'longname': 'mnemonic',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'mn'},
                     'numberOfItems': {'args': '<type __builtin__.bool>',
                                       'longname': 'numberOfItems',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'ni'},
                     'parent': {'args': '<type __builtin__.str>',
                                'longname': 'parent',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'p'},
                     'postMenuCommand': {'args': 'script',
                                         'longname': 'postMenuCommand',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'pmc'},
                     'postMenuCommandOnce': {'args': '<type __builtin__.bool>',
                                             'longname': 'postMenuCommandOnce',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'pmo'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'scrollable': {'args': '<type __builtin__.bool>',
                                    'longname': 'scrollable',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'srb'},
                     'tearOff': {'args': '<type __builtin__.bool>',
                                 'longname': 'tearOff',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'to'},
                     'useTemplate': {'args': '<type __builtin__.str>',
                                     'longname': 'useTemplate',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'ut'},
                     'version': {'args': '<type __builtin__.str>',
                                 'longname': 'version',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'ver'},
                     'visible': {'args': '<type __builtin__.bool>',
                                 'longname': 'visible',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'vis'}},
           'shortFlags': {'aob': 'allowOptionBoxes',
                          'dai': 'deleteAllItems',
                          'dt': 'defineTemplate',
                          'dtg': 'docTag',
                          'e': 'edit',
                          'en': 'enable',
                          'ex': 'exists',
                          'fi': 'familyImage',
                          'hm': 'helpMenu',
                          'ia': 'itemArray',
                          'l': 'label',
                          'lt': 'ltVersion',
                          'mn': 'mnemonic',
                          'ni': 'numberOfItems',
                          'p': 'parent',
                          'pmc': 'postMenuCommand',
                          'pmo': 'postMenuCommandOnce',
                          'q': 'query',
                          'srb': 'scrollable',
                          'to': 'tearOff',
                          'ut': 'useTemplate',
                          'ver': 'version',
                          'vis': 'visible'},
           'type': 'windows'},
  'menuBarLayout': {'callbackFlags': ['visibleChangeCommand',
                                      'vcc',
                                      'dragCallback',
                                      'dgc',
                                      'dropCallback',
                                      'dpc'],
                    'flags': {'annotation': {'args': '<type __builtin__.str>',
                                             'longname': 'annotation',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ann'},
                              'backgroundColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'backgroundColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'bgc'},
                              'childArray': {'args': '<type __builtin__.bool>',
                                             'longname': 'childArray',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'ca'},
                              'defineTemplate': {'args': '<type __builtin__.str>',
                                                 'longname': 'defineTemplate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'dt'},
                              'docTag': {'args': '<type __builtin__.str>',
                                         'longname': 'docTag',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dtg'},
                              'dragCallback': {'args': 'script',
                                               'longname': 'dragCallback',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dgc'},
                              'dropCallback': {'args': 'script',
                                               'longname': 'dropCallback',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dpc'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'enable': {'args': '<type __builtin__.bool>',
                                         'longname': 'enable',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'en'},
                              'enableBackground': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableBackground',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ebg'},
                              'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'enableKeyboardFocus',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ekf'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'fullPathName': {'args': '<type __builtin__.bool>',
                                               'longname': 'fullPathName',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'fpn'},
                              'height': {'args': '<type __builtin__.int>',
                                         'longname': 'height',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'h'},
                              'highlightColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'highlightColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'hlc'},
                              'isObscured': {'args': '<type __builtin__.bool>',
                                             'longname': 'isObscured',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'io'},
                              'manage': {'args': '<type __builtin__.bool>',
                                         'longname': 'manage',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                              'menuArray': {'args': '<type __builtin__.bool>',
                                            'longname': 'menuArray',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ma'},
                              'menuBarVisible': {'args': '<type __builtin__.bool>',
                                                 'longname': 'menuBarVisible',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'mbv'},
                              'menuIndex': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'menuIndex',
                                            'modes': ['edit'],
                                            'numArgs': 2,
                                            'shortname': 'mi'},
                              'noBackground': {'args': '<type __builtin__.bool>',
                                               'longname': 'noBackground',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nbg'},
                              'numberOfChildren': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfChildren',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'nch'},
                              'numberOfMenus': {'args': '<type __builtin__.bool>',
                                                'longname': 'numberOfMenus',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'nm'},
                              'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'numberOfPopupMenus',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'npm'},
                              'parent': {'args': '<type __builtin__.str>',
                                         'longname': 'parent',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'p'},
                              'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                 'longname': 'popupMenuArray',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'pma'},
                              'preventOverride': {'args': '<type __builtin__.bool>',
                                                  'longname': 'preventOverride',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'po'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'statusBarMessage': {'args': '<type __builtin__.str>',
                                                   'longname': 'statusBarMessage',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sbm'},
                              'useTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'useTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ut'},
                              'visible': {'args': '<type __builtin__.bool>',
                                          'longname': 'visible',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'vis'},
                              'visibleChangeCommand': {'args': 'script',
                                                       'longname': 'visibleChangeCommand',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'vcc'},
                              'width': {'args': '<type __builtin__.int>',
                                        'longname': 'width',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'w'}},
                    'shortFlags': {'ann': 'annotation',
                                   'bgc': 'backgroundColor',
                                   'ca': 'childArray',
                                   'dgc': 'dragCallback',
                                   'dpc': 'dropCallback',
                                   'dt': 'defineTemplate',
                                   'dtg': 'docTag',
                                   'e': 'edit',
                                   'ebg': 'enableBackground',
                                   'ekf': 'enableKeyboardFocus',
                                   'en': 'enable',
                                   'ex': 'exists',
                                   'fpn': 'fullPathName',
                                   'h': 'height',
                                   'hlc': 'highlightColor',
                                   'io': 'isObscured',
                                   'm': 'manage',
                                   'ma': 'menuArray',
                                   'mbv': 'menuBarVisible',
                                   'mi': 'menuIndex',
                                   'nbg': 'noBackground',
                                   'nch': 'numberOfChildren',
                                   'nm': 'numberOfMenus',
                                   'npm': 'numberOfPopupMenus',
                                   'p': 'parent',
                                   'pma': 'popupMenuArray',
                                   'po': 'preventOverride',
                                   'q': 'query',
                                   'sbm': 'statusBarMessage',
                                   'ut': 'useTemplate',
                                   'vcc': 'visibleChangeCommand',
                                   'vis': 'visible',
                                   'w': 'width'},
                    'type': 'windows'},
  'menuEditor': {'callbackFlags': ['visibleChangeCommand',
                                   'vcc',
                                   'dragCallback',
                                   'dgc',
                                   'optionBoxCommand',
                                   'obc',
                                   'command',
                                   'c',
                                   'dropCallback',
                                   'dpc'],
                 'flags': {'annotation': {'args': '<type __builtin__.str>',
                                          'longname': 'annotation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ann'},
                           'backgroundColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'backgroundColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'bgc'},
                           'cellHeight': {'args': '<type __builtin__.int>',
                                          'longname': 'cellHeight',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ch'},
                           'cellWidth': {'args': '<type __builtin__.int>',
                                         'longname': 'cellWidth',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cw'},
                           'cellWidthHeight': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'cellWidthHeight',
                                               'modes': ['edit'],
                                               'numArgs': 2,
                                               'shortname': 'cwh'},
                           'checkBoxPresent': {'args': ['<type __builtin__.bool>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'checkBoxPresent',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'cbp'},
                           'checkBoxState': {'args': ['<type __builtin__.bool>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'checkBoxState',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 3,
                                             'shortname': 'cbs'},
                           'childArray': {'args': '<type __builtin__.bool>',
                                          'longname': 'childArray',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'ca'},
                           'command': {'args': ['<type __builtin__.str>',
                                                '<type __builtin__.str>',
                                                '<type __builtin__.int>'],
                                       'longname': 'command',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 3,
                                       'shortname': 'c'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'delete': {'args': ['<type __builtin__.str>',
                                               '<type __builtin__.int>'],
                                      'longname': 'delete',
                                      'modes': ['edit'],
                                      'numArgs': 2,
                                      'shortname': 'd'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'dragCallback': {'args': 'script',
                                            'longname': 'dragCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dgc'},
                           'dropCallback': {'args': 'script',
                                            'longname': 'dropCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dpc'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'enable': {'args': '<type __builtin__.bool>',
                                      'longname': 'enable',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'en'},
                           'enableBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableBackground',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ebg'},
                           'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableKeyboardFocus',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ekf'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'fullPathName': {'args': '<type __builtin__.bool>',
                                            'longname': 'fullPathName',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fpn'},
                           'height': {'args': '<type __builtin__.int>',
                                      'longname': 'height',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                           'highlightColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'highlightColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'hlc'},
                           'iconMenuCallback': {'args': '<type __builtin__.str>',
                                                'longname': 'iconMenuCallback',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'imc'},
                           'image': {'args': ['<type __builtin__.str>',
                                              '<type __builtin__.str>',
                                              '<type __builtin__.int>'],
                                     'longname': 'image',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 3,
                                     'shortname': 'i'},
                           'isObscured': {'args': '<type __builtin__.bool>',
                                          'longname': 'isObscured',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'io'},
                           'label': {'args': ['<type __builtin__.str>',
                                              '<type __builtin__.str>',
                                              '<type __builtin__.int>'],
                                     'longname': 'label',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 3,
                                     'shortname': 'l'},
                           'manage': {'args': '<type __builtin__.bool>',
                                      'longname': 'manage',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                           'menuItemTypes': {'args': '<type __builtin__.bool>',
                                             'longname': 'menuItemTypes',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'mit'},
                           'noBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'noBackground',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nbg'},
                           'numberOfChildren': {'args': '<type __builtin__.bool>',
                                                'longname': 'numberOfChildren',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'nch'},
                           'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'numberOfPopupMenus',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'npm'},
                           'optionBoxCommand': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'optionBoxCommand',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 3,
                                                'shortname': 'obc'},
                           'optionBoxPresent': {'args': ['<type __builtin__.bool>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'optionBoxPresent',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 3,
                                                'shortname': 'obp'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuArray': {'args': '<type __builtin__.bool>',
                                              'longname': 'popupMenuArray',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'pma'},
                           'preventOverride': {'args': '<type __builtin__.bool>',
                                               'longname': 'preventOverride',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'po'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'radioButtonPresent': {'args': ['<type __builtin__.bool>',
                                                           '<type __builtin__.str>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'radioButtonPresent',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'rbp'},
                           'radioButtonState': {'args': ['<type __builtin__.bool>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'radioButtonState',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 3,
                                                'shortname': 'rbs'},
                           'separator': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.int>'],
                                         'longname': 'separator',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 2,
                                         'shortname': 'sp'},
                           'statusBarMessage': {'args': '<type __builtin__.str>',
                                                'longname': 'statusBarMessage',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sbm'},
                           'style': {'args': '<type __builtin__.str>',
                                     'longname': 'style',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'st'},
                           'subMenuAt': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.int>'],
                                         'longname': 'subMenuAt',
                                         'modes': ['edit'],
                                         'numArgs': 2,
                                         'shortname': 'sma'},
                           'subMenuEditorWindow': {'args': '<type __builtin__.str>',
                                                   'longname': 'subMenuEditorWindow',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'sew'},
                           'subMenuEditorsOpen': {'args': '<type __builtin__.bool>',
                                                  'longname': 'subMenuEditorsOpen',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'seo'},
                           'subMenuOf': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.str>',
                                                  '<type __builtin__.int>'],
                                         'longname': 'subMenuOf',
                                         'modes': ['create'],
                                         'numArgs': 3,
                                         'shortname': 'smo'},
                           'topLevelMenu': {'args': '<type __builtin__.str>',
                                            'longname': 'topLevelMenu',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'tlm'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'},
                           'visible': {'args': '<type __builtin__.bool>',
                                       'longname': 'visible',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vis'},
                           'visibleChangeCommand': {'args': 'script',
                                                    'longname': 'visibleChangeCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'vcc'},
                           'width': {'args': '<type __builtin__.int>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'}},
                 'shortFlags': {'ann': 'annotation',
                                'bgc': 'backgroundColor',
                                'c': 'command',
                                'cbp': 'checkBoxPresent',
                                'cbs': 'checkBoxState',
                                'ch': 'cellHeight',
                                'cw': 'cellWidth',
                                'cwh': 'cellWidthHeight',
                                'd': 'delete',
                                'dgc': 'dragCallback',
                                'dpc': 'dropCallback',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'ebg': 'enableBackground',
                                'ekf': 'enableKeyboardFocus',
                                'en': 'enable',
                                'ex': 'exists',
                                'fpn': 'fullPathName',
                                'h': 'height',
                                'hlc': 'highlightColor',
                                'i': 'image',
                                'imc': 'iconMenuCallback',
                                'io': 'isObscured',
                                'l': 'label',
                                'm': 'manage',
                                'mit': 'menuItemTypes',
                                'nbg': 'noBackground',
                                'npm': 'numberOfPopupMenus',
                                'obc': 'optionBoxCommand',
                                'obp': 'optionBoxPresent',
                                'p': 'parent',
                                'pma': 'popupMenuArray',
                                'po': 'preventOverride',
                                'q': 'query',
                                'rbp': 'radioButtonPresent',
                                'rbs': 'radioButtonState',
                                'sbm': 'statusBarMessage',
                                'seo': 'subMenuEditorsOpen',
                                'sew': 'subMenuEditorWindow',
                                'sma': 'subMenuAt',
                                'smo': 'subMenuOf',
                                'sp': 'separator',
                                'st': 'style',
                                'tlm': 'topLevelMenu',
                                'ut': 'useTemplate',
                                'vcc': 'visibleChangeCommand',
                                'vis': 'visible',
                                'w': 'width'},
                 'type': 'windows'},
  'menuItem': {'callbackFlags': ['runTimeCommand',
                                 'rtc',
                                 'dragMenuCommand',
                                 'dmc',
                                 'dragDoubleClickCommand',
                                 'ddc',
                                 'postMenuCommandOnce',
                                 'pmo',
                                 'postMenuCommand',
                                 'pmc',
                                 'commandModifier',
                                 'cmd',
                                 'echoCommand',
                                 'ec',
                                 'enableCommandRepeat',
                                 'ecr',
                                 'command',
                                 'c'],
               'flags': {'allowOptionBoxes': {'args': '<type __builtin__.bool>',
                                              'longname': 'allowOptionBoxes',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'aob'},
                         'altModifier': {'args': '<type __builtin__.bool>',
                                         'longname': 'altModifier',
                                         'numArgs': 1,
                                         'shortname': 'alt'},
                         'annotation': {'args': '<type __builtin__.str>',
                                        'longname': 'annotation',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ann'},
                         'boldFont': {'args': '<type __builtin__.bool>',
                                      'longname': 'boldFont',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'bld'},
                         'checkBox': {'args': '<type __builtin__.bool>',
                                      'longname': 'checkBox',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'cb'},
                         'collection': {'args': '<type __builtin__.str>',
                                        'longname': 'collection',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'cl'},
                         'command': {'args': 'script',
                                     'longname': 'command',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'c'},
                         'commandModifier': {'args': '<type __builtin__.bool>',
                                             'longname': 'commandModifier',
                                             'numArgs': 1,
                                             'shortname': 'cmd'},
                         'ctrlModifier': {'args': '<type __builtin__.bool>',
                                          'longname': 'ctrlModifier',
                                          'numArgs': 1,
                                          'shortname': 'ctl'},
                         'data': {'args': '<type __builtin__.int>',
                                  'longname': 'data',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'da'},
                         'defineTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'defineTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'dt'},
                         'divider': {'args': '<type __builtin__.bool>',
                                     'longname': 'divider',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'd'},
                         'dividerLabel': {'args': '<type __builtin__.str>',
                                          'longname': 'dividerLabel',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dl'},
                         'docTag': {'args': '<type __builtin__.str>',
                                    'longname': 'docTag',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'dtg'},
                         'dragDoubleClickCommand': {'args': 'script',
                                                    'longname': 'dragDoubleClickCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ddc'},
                         'dragMenuCommand': {'args': 'script',
                                             'longname': 'dragMenuCommand',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dmc'},
                         'echoCommand': {'args': '<type __builtin__.bool>',
                                         'longname': 'echoCommand',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ec'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'enable': {'args': '<type __builtin__.bool>',
                                    'longname': 'enable',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'en'},
                         'enableCommandRepeat': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableCommandRepeat',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ecr'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'familyImage': {'args': '<type __builtin__.str>',
                                         'longname': 'familyImage',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'fi'},
                         'image': {'args': '<type __builtin__.str>',
                                   'longname': 'image',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'i'},
                         'imageOverlayLabel': {'args': '<type __builtin__.str>',
                                               'longname': 'imageOverlayLabel',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'iol'},
                         'insertAfter': {'args': '<type __builtin__.str>',
                                         'longname': 'insertAfter',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ia'},
                         'isCheckBox': {'args': '<type __builtin__.bool>',
                                        'longname': 'isCheckBox',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'icb'},
                         'isOptionBox': {'args': '<type __builtin__.bool>',
                                         'longname': 'isOptionBox',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'iob'},
                         'isRadioButton': {'args': '<type __builtin__.bool>',
                                           'longname': 'isRadioButton',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'irb'},
                         'italicized': {'args': '<type __builtin__.bool>',
                                        'longname': 'italicized',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'itl'},
                         'keyEquivalent': {'args': '<type __builtin__.str>',
                                           'longname': 'keyEquivalent',
                                           'numArgs': 1,
                                           'shortname': 'ke'},
                         'label': {'args': '<type __builtin__.str>',
                                   'longname': 'label',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'l'},
                         'longDivider': {'args': '<type __builtin__.bool>',
                                         'longname': 'longDivider',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ld'},
                         'ltVersion': {'args': '<type __builtin__.str>',
                                       'longname': 'ltVersion',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'lt'},
                         'mnemonic': {'args': '<type __builtin__.str>',
                                      'longname': 'mnemonic',
                                      'numArgs': 1,
                                      'shortname': 'mn'},
                         'optionBox': {'args': '<type __builtin__.bool>',
                                       'longname': 'optionBox',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'ob'},
                         'optionBoxIcon': {'args': '<type __builtin__.str>',
                                           'longname': 'optionBoxIcon',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'obi'},
                         'optionModifier': {'args': '<type __builtin__.bool>',
                                            'longname': 'optionModifier',
                                            'numArgs': 1,
                                            'shortname': 'opt'},
                         'parent': {'args': '<type __builtin__.str>',
                                    'longname': 'parent',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'p'},
                         'postMenuCommand': {'args': 'script',
                                             'longname': 'postMenuCommand',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'pmc'},
                         'postMenuCommandOnce': {'args': '<type __builtin__.bool>',
                                                 'longname': 'postMenuCommandOnce',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'pmo'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'radialPosition': {'args': '<type __builtin__.str>',
                                            'longname': 'radialPosition',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'rp'},
                         'radioButton': {'args': '<type __builtin__.bool>',
                                         'longname': 'radioButton',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'rb'},
                         'runTimeCommand': {'args': '<type __builtin__.str>',
                                            'longname': 'runTimeCommand',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'rtc'},
                         'shiftModifier': {'args': '<type __builtin__.bool>',
                                           'longname': 'shiftModifier',
                                           'numArgs': 1,
                                           'shortname': 'sh'},
                         'sourceType': {'args': '<type __builtin__.str>',
                                        'longname': 'sourceType',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'stp'},
                         'subMenu': {'args': '<type __builtin__.bool>',
                                     'longname': 'subMenu',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'sm'},
                         'tearOff': {'args': '<type __builtin__.bool>',
                                     'longname': 'tearOff',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'to'},
                         'useTemplate': {'args': '<type __builtin__.str>',
                                         'longname': 'useTemplate',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ut'},
                         'version': {'args': '<type __builtin__.str>',
                                     'longname': 'version',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'ver'}},
               'shortFlags': {'alt': 'altModifier',
                              'ann': 'annotation',
                              'aob': 'allowOptionBoxes',
                              'bld': 'boldFont',
                              'c': 'command',
                              'cb': 'checkBox',
                              'cl': 'collection',
                              'cmd': 'commandModifier',
                              'ctl': 'ctrlModifier',
                              'd': 'divider',
                              'da': 'data',
                              'ddc': 'dragDoubleClickCommand',
                              'dl': 'dividerLabel',
                              'dmc': 'dragMenuCommand',
                              'dt': 'defineTemplate',
                              'dtg': 'docTag',
                              'e': 'edit',
                              'ec': 'echoCommand',
                              'ecr': 'enableCommandRepeat',
                              'en': 'enable',
                              'ex': 'exists',
                              'fi': 'familyImage',
                              'i': 'image',
                              'ia': 'insertAfter',
                              'icb': 'isCheckBox',
                              'iob': 'isOptionBox',
                              'iol': 'imageOverlayLabel',
                              'irb': 'isRadioButton',
                              'itl': 'italicized',
                              'ke': 'keyEquivalent',
                              'l': 'label',
                              'ld': 'longDivider',
                              'lt': 'ltVersion',
                              'mn': 'mnemonic',
                              'ob': 'optionBox',
                              'obi': 'optionBoxIcon',
                              'opt': 'optionModifier',
                              'p': 'parent',
                              'pmc': 'postMenuCommand',
                              'pmo': 'postMenuCommandOnce',
                              'q': 'query',
                              'rb': 'radioButton',
                              'rp': 'radialPosition',
                              'rtc': 'runTimeCommand',
                              'sh': 'shiftModifier',
                              'sm': 'subMenu',
                              'stp': 'sourceType',
                              'to': 'tearOff',
                              'ut': 'useTemplate',
                              'ver': 'version'},
               'type': 'windows'},
  'menuSet': {'flags': {'addMenu': {'args': '<type __builtin__.str>',
                                    'longname': 'addMenu',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'am'},
                        'allMenuSets': {'args': '<type __builtin__.bool>',
                                        'longname': 'allMenuSets',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'ams'},
                        'currentMenuSet': {'args': '<type __builtin__.str>',
                                           'longname': 'currentMenuSet',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'cms'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'exists': {'args': '<type __builtin__.str>',
                                   'longname': 'exists',
                                   'modes': ['query'],
                                   'numArgs': 1,
                                   'shortname': 'ex'},
                        'hotBoxVisible': {'args': '<type __builtin__.bool>',
                                          'longname': 'hotBoxVisible',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'hbv'},
                        'insertMenu': {'args': ['<type __builtin__.str>',
                                                '<type __builtin__.int>'],
                                       'longname': 'insertMenu',
                                       'modes': ['create'],
                                       'numArgs': 2,
                                       'shortname': 'im'},
                        'label': {'args': '<type __builtin__.str>',
                                  'longname': 'label',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'l'},
                        'menuArray': {'args': '<type __builtin__.str>',
                                      'longname': 'menuArray',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'ma'},
                        'moveMenu': {'args': ['<type __builtin__.str>',
                                              '<type __builtin__.int>'],
                                     'longname': 'moveMenu',
                                     'modes': ['create'],
                                     'numArgs': 2,
                                     'shortname': 'mm'},
                        'moveMenuSet': {'args': ['<type __builtin__.str>',
                                                 '<type __builtin__.int>'],
                                        'longname': 'moveMenuSet',
                                        'modes': ['create'],
                                        'numArgs': 2,
                                        'shortname': 'mms'},
                        'numberOfMenuSets': {'args': '<type __builtin__.bool>',
                                             'longname': 'numberOfMenuSets',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'nms'},
                        'numberOfMenus': {'args': '<type __builtin__.bool>',
                                          'longname': 'numberOfMenus',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'nm'},
                        'permanent': {'args': '<type __builtin__.bool>',
                                      'longname': 'permanent',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'removeMenu': {'args': '<type __builtin__.str>',
                                       'longname': 'removeMenu',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'rm'},
                        'removeMenuSet': {'args': '<type __builtin__.str>',
                                          'longname': 'removeMenuSet',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'rms'}},
              'shortFlags': {'am': 'addMenu',
                             'ams': 'allMenuSets',
                             'cms': 'currentMenuSet',
                             'e': 'edit',
                             'ex': 'exists',
                             'hbv': 'hotBoxVisible',
                             'im': 'insertMenu',
                             'l': 'label',
                             'ma': 'menuArray',
                             'mm': 'moveMenu',
                             'mms': 'moveMenuSet',
                             'nm': 'numberOfMenus',
                             'nms': 'numberOfMenuSets',
                             'p': 'permanent',
                             'q': 'query',
                             'rm': 'removeMenu',
                             'rms': 'removeMenuSet'},
              'type': 'windows'},
  'menuSetPref': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'e'},
                            'force': {'args': '<type __builtin__.bool>',
                                      'longname': 'force',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'f'},
                            'loadAll': {'args': '<type __builtin__.bool>',
                                        'longname': 'loadAll',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'la'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'removeAll': {'args': '<type __builtin__.bool>',
                                          'longname': 'removeAll',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ra'},
                            'saveAll': {'args': '<type __builtin__.bool>',
                                        'longname': 'saveAll',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'sa'},
                            'saveBackup': {'args': '<type __builtin__.bool>',
                                           'longname': 'saveBackup',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'sb'},
                            'version': {'args': '<type __builtin__.bool>',
                                        'longname': 'version',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'v'}},
                  'shortFlags': {'e': 'exists',
                                 'f': 'force',
                                 'la': 'loadAll',
                                 'q': 'query',
                                 'ra': 'removeAll',
                                 'sa': 'saveAll',
                                 'sb': 'saveBackup',
                                 'v': 'version'},
                  'type': 'windows'},
  'meshIntersectTest': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'messageLine': {'callbackFlags': ['visibleChangeCommand',
                                    'vcc',
                                    'dragCallback',
                                    'dgc',
                                    'dropCallback',
                                    'dpc'],
                  'flags': {'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'ex': 'exists',
                                 'fpn': 'fullPathName',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'io': 'isObscured',
                                 'm': 'manage',
                                 'nbg': 'noBackground',
                                 'npm': 'numberOfPopupMenus',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'sbm': 'statusBarMessage',
                                 'ut': 'useTemplate',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'minimizeApp': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'mirrorJoint': {'flags': {'mirrorBehavior': {'args': '<type __builtin__.bool>',
                                               'longname': 'mirrorBehavior',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'mb'},
                            'mirrorXY': {'args': '<type __builtin__.bool>',
                                         'longname': 'mirrorXY',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'mxy'},
                            'mirrorXZ': {'args': '<type __builtin__.bool>',
                                         'longname': 'mirrorXZ',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'mxz'},
                            'mirrorYZ': {'args': '<type __builtin__.bool>',
                                         'longname': 'mirrorYZ',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'myz'},
                            'searchReplace': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'searchReplace',
                                              'modes': ['create'],
                                              'numArgs': 2,
                                              'shortname': 'sr'}},
                  'shortFlags': {'mb': 'mirrorBehavior',
                                 'mxy': 'mirrorXY',
                                 'mxz': 'mirrorXZ',
                                 'myz': 'mirrorYZ',
                                 'sr': 'searchReplace'},
                  'type': 'animation'},
  'modelCurrentTimeCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'history': {'args': '<type __builtin__.bool>',
                                                'longname': 'history',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'percent': {'args': '<type __builtin__.float>',
                                                'longname': 'percent',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'per'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'}},
                          'shortFlags': {'ch': 'history',
                                         'e': 'edit',
                                         'ex': 'exists',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'n': 'name',
                                         'per': 'percent',
                                         'q': 'query'},
                          'type': 'context'},
  'modelEditor': {'callbackFlags': ['objectFilterUI',
                                    'obu',
                                    'editorChanged',
                                    'ec',
                                    'objectFilterList',
                                    'ofl',
                                    'objectFilterListUI',
                                    'ofu',
                                    'objectFilter',
                                    'obf'],
                  'flags': {'activeComponentsXray': {'args': '<type __builtin__.bool>',
                                                     'longname': 'activeComponentsXray',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'acx'},
                            'activeCustomEnvironment': {'args': '<type __builtin__.str>',
                                                        'longname': 'activeCustomEnvironment',
                                                        'modes': ['edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ace'},
                            'activeCustomGeometry': {'args': '<type __builtin__.str>',
                                                     'longname': 'activeCustomGeometry',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'acg'},
                            'activeCustomLighSet': {'args': '<type __builtin__.str>',
                                                    'longname': 'activeCustomLighSet',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'acl'},
                            'activeCustomOverrideGeometry': {'args': '<type __builtin__.str>',
                                                             'longname': 'activeCustomOverrideGeometry',
                                                             'modes': ['query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'aog'},
                            'activeCustomRenderer': {'args': '<type __builtin__.str>',
                                                     'longname': 'activeCustomRenderer',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'acr'},
                            'activeOnly': {'args': '<type __builtin__.bool>',
                                           'longname': 'activeOnly',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ao'},
                            'activeShadingGraph': {'args': '<type __builtin__.str>',
                                                   'longname': 'activeShadingGraph',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'asg'},
                            'activeView': {'args': '<type __builtin__.bool>',
                                           'longname': 'activeView',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'av'},
                            'addObjects': {'args': '<type __builtin__.str>',
                                           'longname': 'addObjects',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'aob'},
                            'addSelected': {'args': '<type __builtin__.bool>',
                                            'longname': 'addSelected',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'addSelected'},
                            'addSelectedObjects': {'args': '<type __builtin__.bool>',
                                                   'longname': 'addSelectedObjects',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'aso'},
                            'allObjects': {'args': '<type __builtin__.bool>',
                                           'longname': 'allObjects',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'alo'},
                            'backfaceCulling': {'args': '<type __builtin__.bool>',
                                                'longname': 'backfaceCulling',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'bfc'},
                            'bufferMode': {'args': '<type __builtin__.str>',
                                           'longname': 'bufferMode',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'bm'},
                            'bumpResolution': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'bumpResolution',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'brz'},
                            'camera': {'args': '<type __builtin__.str>',
                                       'longname': 'camera',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cam'},
                            'cameraName': {'args': '<type __builtin__.str>',
                                           'longname': 'cameraName',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cn'},
                            'cameraSet': {'args': '<type __builtin__.str>',
                                          'longname': 'cameraSet',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cst'},
                            'cameraSetup': {'args': '<type __builtin__.bool>',
                                            'longname': 'cameraSetup',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'cs'},
                            'cameras': {'args': '<type __builtin__.bool>',
                                        'longname': 'cameras',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ca'},
                            'capture': {'args': '<type __builtin__.str>',
                                        'longname': 'capture',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cpt'},
                            'captureSequenceNumber': {'args': '<type __builtin__.int>',
                                                      'longname': 'captureSequenceNumber',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'csn'},
                            'clipGhosts': {'args': '<type __builtin__.bool>',
                                           'longname': 'clipGhosts',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cg'},
                            'cmEnabled': {'args': '<type __builtin__.bool>',
                                          'longname': 'cmEnabled',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cme'},
                            'colorMap': {'args': '<type __builtin__.bool>',
                                         'longname': 'colorMap',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'cm'},
                            'colorResolution': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'colorResolution',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 2,
                                                'shortname': 'crz'},
                            'control': {'args': '<type __builtin__.bool>',
                                        'longname': 'control',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'ctl'},
                            'controlVertices': {'args': '<type __builtin__.bool>',
                                                'longname': 'controlVertices',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cv'},
                            'controllers': {'args': '<type __builtin__.bool>',
                                            'longname': 'controllers',
                                            'numArgs': 1,
                                            'shortname': 'cts'},
                            'cullingOverride': {'args': '<type __builtin__.str>',
                                                'longname': 'cullingOverride',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cov'},
                            'default': {'args': '<type __builtin__.bool>',
                                        'longname': 'default',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'd'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'deformers': {'args': '<type __builtin__.bool>',
                                          'longname': 'deformers',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'df'},
                            'depthOfFieldPreview': {'args': '<type __builtin__.bool>',
                                                    'longname': 'depthOfFieldPreview',
                                                    'numArgs': 1,
                                                    'shortname': 'dof'},
                            'dimensions': {'args': '<type __builtin__.bool>',
                                           'longname': 'dimensions',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dim'},
                            'displayAppearance': {'args': '<type __builtin__.str>',
                                                  'longname': 'displayAppearance',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'da'},
                            'displayLights': {'args': '<type __builtin__.str>',
                                              'longname': 'displayLights',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dl'},
                            'displayTextures': {'args': '<type __builtin__.bool>',
                                                'longname': 'displayTextures',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dtx'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dynamicConstraints': {'args': '<type __builtin__.bool>',
                                                   'longname': 'dynamicConstraints',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dc'},
                            'dynamics': {'args': '<type __builtin__.bool>',
                                         'longname': 'dynamics',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dy'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'editorChanged': {'args': 'script',
                                              'longname': 'editorChanged',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ec'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'exposure': {'args': '<type __builtin__.float>',
                                         'longname': 'exposure',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'exp'},
                            'filter': {'args': '<type __builtin__.str>',
                                       'longname': 'filter',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'f'},
                            'filteredObjectList': {'args': '<type __builtin__.bool>',
                                                   'longname': 'filteredObjectList',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'fol'},
                            'fluids': {'args': '<type __builtin__.bool>',
                                       'longname': 'fluids',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'fl'},
                            'fogColor': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'fogColor',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 4,
                                         'shortname': 'fcl'},
                            'fogDensity': {'args': '<type __builtin__.float>',
                                           'longname': 'fogDensity',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'fdn'},
                            'fogEnd': {'args': '<type __builtin__.float>',
                                       'longname': 'fogEnd',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'fen'},
                            'fogMode': {'args': '<type __builtin__.str>',
                                        'longname': 'fogMode',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'fmd'},
                            'fogSource': {'args': '<type __builtin__.str>',
                                          'longname': 'fogSource',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'fsc'},
                            'fogStart': {'args': '<type __builtin__.float>',
                                         'longname': 'fogStart',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'fst'},
                            'fogging': {'args': '<type __builtin__.bool>',
                                        'longname': 'fogging',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'fg'},
                            'follicles': {'args': '<type __builtin__.bool>',
                                          'longname': 'follicles',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'fo'},
                            'forceMainConnection': {'args': '<type __builtin__.str>',
                                                    'longname': 'forceMainConnection',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'fmc'},
                            'gamma': {'args': '<type __builtin__.float>',
                                      'longname': 'gamma',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'ga'},
                            'greasePencils': {'args': '<type __builtin__.bool>',
                                              'longname': 'greasePencils',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'gp'},
                            'grid': {'args': '<type __builtin__.bool>',
                                     'longname': 'grid',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'gr'},
                            'hairSystems': {'args': '<type __builtin__.bool>',
                                            'longname': 'hairSystems',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'hs'},
                            'handles': {'args': '<type __builtin__.bool>',
                                        'longname': 'handles',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ha'},
                            'headsUpDisplay': {'args': '<type __builtin__.bool>',
                                               'longname': 'headsUpDisplay',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'hud'},
                            'height': {'args': '<type __builtin__.bool>',
                                       'longname': 'height',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightConnection': {'args': '<type __builtin__.str>',
                                                    'longname': 'highlightConnection',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'hlc'},
                            'holdOuts': {'args': '<type __builtin__.bool>',
                                         'longname': 'holdOuts',
                                         'numArgs': 1,
                                         'shortname': 'hos'},
                            'hulls': {'args': '<type __builtin__.bool>',
                                      'longname': 'hulls',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'hu'},
                            'ignorePanZoom': {'args': '<type __builtin__.bool>',
                                              'longname': 'ignorePanZoom',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ipz'},
                            'ikHandles': {'args': '<type __builtin__.bool>',
                                          'longname': 'ikHandles',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ikh'},
                            'imagePlane': {'args': '<type __builtin__.bool>',
                                           'longname': 'imagePlane',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'imp'},
                            'interactive': {'args': '<type __builtin__.bool>',
                                            'longname': 'interactive',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'i'},
                            'interactiveBackFaceCull': {'args': '<type __builtin__.bool>',
                                                        'longname': 'interactiveBackFaceCull',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ibc'},
                            'interactiveDisableShadows': {'args': '<type __builtin__.bool>',
                                                          'longname': 'interactiveDisableShadows',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'dis'},
                            'isFiltered': {'args': '<type __builtin__.bool>',
                                           'longname': 'isFiltered',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 'isFiltered'},
                            'jointXray': {'args': '<type __builtin__.bool>',
                                          'longname': 'jointXray',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'jx'},
                            'joints': {'args': '<type __builtin__.bool>',
                                       'longname': 'joints',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'j'},
                            'lights': {'args': '<type __builtin__.bool>',
                                       'longname': 'lights',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'lt'},
                            'lineWidth': {'args': '<type __builtin__.float>',
                                          'longname': 'lineWidth',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'lw'},
                            'locators': {'args': '<type __builtin__.bool>',
                                         'longname': 'locators',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'lc'},
                            'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                   'longname': 'lockMainConnection',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'lck'},
                            'lowQualityLighting': {'args': '<type __builtin__.bool>',
                                                   'longname': 'lowQualityLighting',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'lql'},
                            'mainListConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'mainListConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'mlc'},
                            'manipulators': {'args': '<type __builtin__.bool>',
                                             'longname': 'manipulators',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'm'},
                            'maxConstantTransparency': {'args': '<type __builtin__.float>',
                                                        'longname': 'maxConstantTransparency',
                                                        'modes': ['query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'mct'},
                            'maximumNumHardwareLights': {'args': '<type __builtin__.bool>',
                                                         'longname': 'maximumNumHardwareLights',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'mhl'},
                            'modelPanel': {'args': '<type __builtin__.str>',
                                           'longname': 'modelPanel',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'mp'},
                            'motionTrails': {'args': '<type __builtin__.bool>',
                                             'longname': 'motionTrails',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'mt'},
                            'nCloths': {'args': '<type __builtin__.bool>',
                                        'longname': 'nCloths',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ncl'},
                            'nParticles': {'args': '<type __builtin__.bool>',
                                           'longname': 'nParticles',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'npa'},
                            'nRigids': {'args': '<type __builtin__.bool>',
                                        'longname': 'nRigids',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'nr'},
                            'noUndo': {'args': '<type __builtin__.bool>',
                                       'longname': 'noUndo',
                                       'modes': ['edit'],
                                       'numArgs': 1,
                                       'shortname': 'nud'},
                            'nurbsCurves': {'args': '<type __builtin__.bool>',
                                            'longname': 'nurbsCurves',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nc'},
                            'nurbsSurfaces': {'args': '<type __builtin__.bool>',
                                              'longname': 'nurbsSurfaces',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ns'},
                            'objectFilter': {'args': 'script',
                                             'longname': 'objectFilter',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'obf'},
                            'objectFilterList': {'args': 'script',
                                                 'longname': 'objectFilterList',
                                                 'modes': ['query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ofl'},
                            'objectFilterListUI': {'args': 'script',
                                                   'longname': 'objectFilterListUI',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ofu'},
                            'objectFilterShowInHUD': {'args': '<type __builtin__.bool>',
                                                      'longname': 'objectFilterShowInHUD',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ofs'},
                            'objectFilterUI': {'args': 'script',
                                               'longname': 'objectFilterUI',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'obu'},
                            'occlusionCulling': {'args': '<type __builtin__.bool>',
                                                 'longname': 'occlusionCulling',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ocl'},
                            'panel': {'args': '<type __builtin__.str>',
                                      'longname': 'panel',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'pnl'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'particleInstancers': {'args': '<type __builtin__.bool>',
                                                   'longname': 'particleInstancers',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'pi'},
                            'pivots': {'args': '<type __builtin__.bool>',
                                       'longname': 'pivots',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'pv'},
                            'planes': {'args': '<type __builtin__.bool>',
                                       'longname': 'planes',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'pl'},
                            'pluginObjects': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.bool>'],
                                              'longname': 'pluginObjects',
                                              'modes': ['edit',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'po'},
                            'pluginShapes': {'args': '<type __builtin__.bool>',
                                             'longname': 'pluginShapes',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'ps'},
                            'polymeshes': {'args': '<type __builtin__.bool>',
                                           'longname': 'polymeshes',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pm'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'queryPluginObjects': {'args': '<type __builtin__.str>',
                                                   'longname': 'queryPluginObjects',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'qpo'},
                            'removeSelected': {'args': '<type __builtin__.bool>',
                                               'longname': 'removeSelected',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'rs'},
                            'rendererDeviceName': {'args': '<type __builtin__.bool>',
                                                   'longname': 'rendererDeviceName',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'rdn'},
                            'rendererList': {'args': '<type __builtin__.bool>',
                                             'longname': 'rendererList',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'rls'},
                            'rendererListUI': {'args': '<type __builtin__.bool>',
                                               'longname': 'rendererListUI',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'rlu'},
                            'rendererName': {'args': '<type __builtin__.str>',
                                             'longname': 'rendererName',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'rnm'},
                            'rendererOverrideList': {'args': '<type __builtin__.bool>',
                                                     'longname': 'rendererOverrideList',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'rol'},
                            'rendererOverrideListUI': {'args': '<type __builtin__.bool>',
                                                       'longname': 'rendererOverrideListUI',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'rou'},
                            'rendererOverrideName': {'args': '<type __builtin__.str>',
                                                     'longname': 'rendererOverrideName',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'rom'},
                            'resetCustomCamera': {'args': '<type __builtin__.bool>',
                                                  'longname': 'resetCustomCamera',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'rcc'},
                            'sceneRenderFilter': {'args': '<type __builtin__.str>',
                                                  'longname': 'sceneRenderFilter',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'srf'},
                            'selectionConnection': {'args': '<type __builtin__.str>',
                                                    'longname': 'selectionConnection',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'slc'},
                            'selectionHiliteDisplay': {'args': '<type __builtin__.bool>',
                                                       'longname': 'selectionHiliteDisplay',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'sel'},
                            'setSelected': {'args': '<type __builtin__.bool>',
                                            'longname': 'setSelected',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'ss'},
                            'shadingModel': {'args': '<type __builtin__.int>',
                                             'longname': 'shadingModel',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sml'},
                            'shadows': {'args': '<type __builtin__.bool>',
                                        'longname': 'shadows',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'sdw'},
                            'smallObjectCulling': {'args': '<type __builtin__.bool>',
                                                   'longname': 'smallObjectCulling',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'soc'},
                            'smallObjectThreshold': {'args': '<type __builtin__.float>',
                                                     'longname': 'smallObjectThreshold',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sot'},
                            'smoothWireframe': {'args': '<type __builtin__.bool>',
                                                'longname': 'smoothWireframe',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'swf'},
                            'sortTransparent': {'args': '<type __builtin__.bool>',
                                                'longname': 'sortTransparent',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'st'},
                            'stateString': {'args': '<type __builtin__.bool>',
                                            'longname': 'stateString',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'sts'},
                            'stereoDrawMode': {'args': '<type __builtin__.bool>',
                                               'longname': 'stereoDrawMode',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 0,
                                               'shortname': 'sdm'},
                            'strokes': {'args': '<type __builtin__.bool>',
                                        'longname': 'strokes',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'str'},
                            'subdivSurfaces': {'args': '<type __builtin__.bool>',
                                               'longname': 'subdivSurfaces',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sds'},
                            'textureAnisotropic': {'args': '<type __builtin__.bool>',
                                                   'longname': 'textureAnisotropic',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ta'},
                            'textureCompression': {'args': '<type __builtin__.bool>',
                                                   'longname': 'textureCompression',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'tcp'},
                            'textureDisplay': {'args': '<type __builtin__.str>',
                                               'longname': 'textureDisplay',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'td'},
                            'textureEnvironmentMap': {'args': '<type __builtin__.bool>',
                                                      'longname': 'textureEnvironmentMap',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'tem'},
                            'textureHilight': {'args': '<type __builtin__.bool>',
                                               'longname': 'textureHilight',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'th'},
                            'textureMaxSize': {'args': '<type __builtin__.int>',
                                               'longname': 'textureMaxSize',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tms'},
                            'textureMemoryUsed': {'args': '<type __builtin__.bool>',
                                                  'longname': 'textureMemoryUsed',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'tmu'},
                            'textureSampling': {'args': '<type __builtin__.int>',
                                                'longname': 'textureSampling',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ts'},
                            'textures': {'args': '<type __builtin__.bool>',
                                         'longname': 'textures',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'tx'},
                            'toggleExposure': {'args': '<type __builtin__.bool>',
                                               'longname': 'toggleExposure',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'tge'},
                            'toggleGamma': {'args': '<type __builtin__.bool>',
                                            'longname': 'toggleGamma',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'tgg'},
                            'transpInShadows': {'args': '<type __builtin__.bool>',
                                                'longname': 'transpInShadows',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'tis'},
                            'transparencyAlgorithm': {'args': '<type __builtin__.str>',
                                                      'longname': 'transparencyAlgorithm',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'tal'},
                            'twoSidedLighting': {'args': '<type __builtin__.bool>',
                                                 'longname': 'twoSidedLighting',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'tsl'},
                            'unParent': {'args': '<type __builtin__.bool>',
                                         'longname': 'unParent',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'up'},
                            'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                     'longname': 'unlockMainConnection',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'ulk'},
                            'updateColorMode': {'args': '<type __builtin__.bool>',
                                                'longname': 'updateColorMode',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'ucm'},
                            'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                     'longname': 'updateMainConnection',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'upd'},
                            'useBaseRenderer': {'args': '<type __builtin__.bool>',
                                                'longname': 'useBaseRenderer',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ubr'},
                            'useColorIndex': {'args': '<type __builtin__.bool>',
                                              'longname': 'useColorIndex',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'uci'},
                            'useDefaultMaterial': {'args': '<type __builtin__.bool>',
                                                   'longname': 'useDefaultMaterial',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'udm'},
                            'useInteractiveMode': {'args': '<type __builtin__.bool>',
                                                   'longname': 'useInteractiveMode',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ui'},
                            'useRGBImagePlane': {'args': '<type __builtin__.bool>',
                                                 'longname': 'useRGBImagePlane',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ip'},
                            'useReducedRenderer': {'args': '<type __builtin__.bool>',
                                                   'longname': 'useReducedRenderer',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'urr'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'userNode': {'args': '<type __builtin__.str>',
                                         'longname': 'userNode',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'un'},
                            'viewObjects': {'args': '<type __builtin__.bool>',
                                            'longname': 'viewObjects',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'vo'},
                            'viewSelected': {'args': '<type __builtin__.bool>',
                                             'longname': 'viewSelected',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'vs'},
                            'viewTransformName': {'args': '<type __builtin__.str>',
                                                  'longname': 'viewTransformName',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'vtn'},
                            'viewType': {'args': '<type __builtin__.bool>',
                                         'longname': 'viewType',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'vt'},
                            'width': {'args': '<type __builtin__.bool>',
                                      'longname': 'width',
                                      'modes': ['query'],
                                      'numArgs': 1,
                                      'shortname': 'w'},
                            'wireframeBackingStore': {'args': '<type __builtin__.bool>',
                                                      'longname': 'wireframeBackingStore',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'wbs'},
                            'wireframeOnShaded': {'args': '<type __builtin__.bool>',
                                                  'longname': 'wireframeOnShaded',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'wos'},
                            'xray': {'args': '<type __builtin__.bool>',
                                     'longname': 'xray',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'xr'}},
                  'removedFlags': {'as': 'addSelected', 'if': 'isFiltered'},
                  'shortFlags': {'ace': 'activeCustomEnvironment',
                                 'acg': 'activeCustomGeometry',
                                 'acl': 'activeCustomLighSet',
                                 'acr': 'activeCustomRenderer',
                                 'acx': 'activeComponentsXray',
                                 'addSelected': 'addSelected',
                                 'alo': 'allObjects',
                                 'ao': 'activeOnly',
                                 'aob': 'addObjects',
                                 'aog': 'activeCustomOverrideGeometry',
                                 'asg': 'activeShadingGraph',
                                 'aso': 'addSelectedObjects',
                                 'av': 'activeView',
                                 'bfc': 'backfaceCulling',
                                 'bm': 'bufferMode',
                                 'brz': 'bumpResolution',
                                 'ca': 'cameras',
                                 'cam': 'camera',
                                 'cg': 'clipGhosts',
                                 'cm': 'colorMap',
                                 'cme': 'cmEnabled',
                                 'cn': 'cameraName',
                                 'cov': 'cullingOverride',
                                 'cpt': 'capture',
                                 'crz': 'colorResolution',
                                 'cs': 'cameraSetup',
                                 'csn': 'captureSequenceNumber',
                                 'cst': 'cameraSet',
                                 'ctl': 'control',
                                 'cts': 'controllers',
                                 'cv': 'controlVertices',
                                 'd': 'default',
                                 'da': 'displayAppearance',
                                 'dc': 'dynamicConstraints',
                                 'df': 'deformers',
                                 'dim': 'dimensions',
                                 'dis': 'interactiveDisableShadows',
                                 'dl': 'displayLights',
                                 'dof': 'depthOfFieldPreview',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'dtx': 'displayTextures',
                                 'dy': 'dynamics',
                                 'e': 'edit',
                                 'ec': 'editorChanged',
                                 'ex': 'exists',
                                 'exp': 'exposure',
                                 'f': 'filter',
                                 'fcl': 'fogColor',
                                 'fdn': 'fogDensity',
                                 'fen': 'fogEnd',
                                 'fg': 'fogging',
                                 'fl': 'fluids',
                                 'fmc': 'forceMainConnection',
                                 'fmd': 'fogMode',
                                 'fo': 'follicles',
                                 'fol': 'filteredObjectList',
                                 'fsc': 'fogSource',
                                 'fst': 'fogStart',
                                 'ga': 'gamma',
                                 'gp': 'greasePencils',
                                 'gr': 'grid',
                                 'h': 'height',
                                 'ha': 'handles',
                                 'hlc': 'highlightConnection',
                                 'hos': 'holdOuts',
                                 'hs': 'hairSystems',
                                 'hu': 'hulls',
                                 'hud': 'headsUpDisplay',
                                 'i': 'interactive',
                                 'ibc': 'interactiveBackFaceCull',
                                 'ikh': 'ikHandles',
                                 'imp': 'imagePlane',
                                 'ip': 'useRGBImagePlane',
                                 'ipz': 'ignorePanZoom',
                                 'isFiltered': 'isFiltered',
                                 'j': 'joints',
                                 'jx': 'jointXray',
                                 'lc': 'locators',
                                 'lck': 'lockMainConnection',
                                 'lql': 'lowQualityLighting',
                                 'lt': 'lights',
                                 'lw': 'lineWidth',
                                 'm': 'manipulators',
                                 'mct': 'maxConstantTransparency',
                                 'mhl': 'maximumNumHardwareLights',
                                 'mlc': 'mainListConnection',
                                 'mp': 'modelPanel',
                                 'mt': 'motionTrails',
                                 'nc': 'nurbsCurves',
                                 'ncl': 'nCloths',
                                 'npa': 'nParticles',
                                 'nr': 'nRigids',
                                 'ns': 'nurbsSurfaces',
                                 'nud': 'noUndo',
                                 'obf': 'objectFilter',
                                 'obu': 'objectFilterUI',
                                 'ocl': 'occlusionCulling',
                                 'ofl': 'objectFilterList',
                                 'ofs': 'objectFilterShowInHUD',
                                 'ofu': 'objectFilterListUI',
                                 'p': 'parent',
                                 'pi': 'particleInstancers',
                                 'pl': 'planes',
                                 'pm': 'polymeshes',
                                 'pnl': 'panel',
                                 'po': 'pluginObjects',
                                 'ps': 'pluginShapes',
                                 'pv': 'pivots',
                                 'q': 'query',
                                 'qpo': 'queryPluginObjects',
                                 'rcc': 'resetCustomCamera',
                                 'rdn': 'rendererDeviceName',
                                 'rls': 'rendererList',
                                 'rlu': 'rendererListUI',
                                 'rnm': 'rendererName',
                                 'rol': 'rendererOverrideList',
                                 'rom': 'rendererOverrideName',
                                 'rou': 'rendererOverrideListUI',
                                 'rs': 'removeSelected',
                                 'sdm': 'stereoDrawMode',
                                 'sds': 'subdivSurfaces',
                                 'sdw': 'shadows',
                                 'sel': 'selectionHiliteDisplay',
                                 'slc': 'selectionConnection',
                                 'sml': 'shadingModel',
                                 'soc': 'smallObjectCulling',
                                 'sot': 'smallObjectThreshold',
                                 'srf': 'sceneRenderFilter',
                                 'ss': 'setSelected',
                                 'st': 'sortTransparent',
                                 'str': 'strokes',
                                 'sts': 'stateString',
                                 'swf': 'smoothWireframe',
                                 'ta': 'textureAnisotropic',
                                 'tal': 'transparencyAlgorithm',
                                 'tcp': 'textureCompression',
                                 'td': 'textureDisplay',
                                 'tem': 'textureEnvironmentMap',
                                 'tge': 'toggleExposure',
                                 'tgg': 'toggleGamma',
                                 'th': 'textureHilight',
                                 'tis': 'transpInShadows',
                                 'tms': 'textureMaxSize',
                                 'tmu': 'textureMemoryUsed',
                                 'ts': 'textureSampling',
                                 'tsl': 'twoSidedLighting',
                                 'tx': 'textures',
                                 'ubr': 'useBaseRenderer',
                                 'uci': 'useColorIndex',
                                 'ucm': 'updateColorMode',
                                 'udm': 'useDefaultMaterial',
                                 'ui': 'useInteractiveMode',
                                 'ulk': 'unlockMainConnection',
                                 'un': 'userNode',
                                 'up': 'unParent',
                                 'upd': 'updateMainConnection',
                                 'urr': 'useReducedRenderer',
                                 'ut': 'useTemplate',
                                 'vo': 'viewObjects',
                                 'vs': 'viewSelected',
                                 'vt': 'viewType',
                                 'vtn': 'viewTransformName',
                                 'w': 'width',
                                 'wbs': 'wireframeBackingStore',
                                 'wos': 'wireframeOnShaded',
                                 'xr': 'xray'},
                  'type': 'windows'},
  'modelPanel': {'callbackFlags': ['popupMenuProcedure',
                                   'pmp',
                                   'tabCommand',
                                   'tc'],
                 'flags': {'addTab': {'args': '<type __builtin__.str>',
                                      'longname': 'addTab',
                                      'numArgs': 1,
                                      'shortname': 'at'},
                           'barLayout': {'args': '<type __builtin__.bool>',
                                         'longname': 'barLayout',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'bl'},
                           'camera': {'args': '<type __builtin__.str>',
                                      'longname': 'camera',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'cam'},
                           'control': {'args': '<type __builtin__.bool>',
                                       'longname': 'control',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'ctl'},
                           'copy': {'args': '<type __builtin__.str>',
                                    'longname': 'copy',
                                    'modes': ['edit'],
                                    'numArgs': 1,
                                    'shortname': 'cp'},
                           'createString': {'args': '<type __builtin__.bool>',
                                            'longname': 'createString',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'cs'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'editString': {'args': '<type __builtin__.bool>',
                                          'longname': 'editString',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'es'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'init': {'args': '<type __builtin__.bool>',
                                    'longname': 'init',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'init'},
                           'isUnique': {'args': '<type __builtin__.bool>',
                                        'longname': 'isUnique',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'iu'},
                           'label': {'args': '<type __builtin__.str>',
                                     'longname': 'label',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'l'},
                           'menuBarRepeatLast': {'args': '<type __builtin__.bool>',
                                                 'longname': 'menuBarRepeatLast',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'mrl'},
                           'menuBarVisible': {'args': '<type __builtin__.bool>',
                                              'longname': 'menuBarVisible',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'mbv'},
                           'modelEditor': {'args': '<type __builtin__.bool>',
                                           'longname': 'modelEditor',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'me'},
                           'needsInit': {'args': '<type __builtin__.bool>',
                                         'longname': 'needsInit',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'ni'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuProcedure': {'args': 'script',
                                                  'longname': 'popupMenuProcedure',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pmp'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'removeTab': {'args': '<type __builtin__.str>',
                                         'longname': 'removeTab',
                                         'numArgs': 1,
                                         'shortname': 'rt'},
                           'replacePanel': {'args': '<type __builtin__.str>',
                                            'longname': 'replacePanel',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'rp'},
                           'selectTab': {'args': '<type __builtin__.str>',
                                         'longname': 'selectTab',
                                         'numArgs': 1,
                                         'shortname': 'st'},
                           'tabCommand': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'tabCommand',
                                          'numArgs': 2,
                                          'shortname': 'tc'},
                           'tearOff': {'args': '<type __builtin__.bool>',
                                       'longname': 'tearOff',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'to'},
                           'tearOffCopy': {'args': '<type __builtin__.str>',
                                           'longname': 'tearOffCopy',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'toc'},
                           'tearOffRestore': {'args': '<type __builtin__.bool>',
                                              'longname': 'tearOffRestore',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'tor'},
                           'unParent': {'args': '<type __builtin__.bool>',
                                        'longname': 'unParent',
                                        'modes': ['edit'],
                                        'numArgs': 0,
                                        'shortname': 'up'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'}},
                 'removedFlags': {'in': 'init'},
                 'shortFlags': {'at': 'addTab',
                                'bl': 'barLayout',
                                'cam': 'camera',
                                'cp': 'copy',
                                'cs': 'createString',
                                'ctl': 'control',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'es': 'editString',
                                'ex': 'exists',
                                'init': 'init',
                                'iu': 'isUnique',
                                'l': 'label',
                                'mbv': 'menuBarVisible',
                                'me': 'modelEditor',
                                'mrl': 'menuBarRepeatLast',
                                'ni': 'needsInit',
                                'p': 'parent',
                                'pmp': 'popupMenuProcedure',
                                'q': 'query',
                                'rp': 'replacePanel',
                                'rt': 'removeTab',
                                'st': 'selectTab',
                                'tc': 'tabCommand',
                                'to': 'tearOff',
                                'toc': 'tearOffCopy',
                                'tor': 'tearOffRestore',
                                'up': 'unParent',
                                'ut': 'useTemplate'},
                 'type': 'windows'},
  'modelingToolkitSuperCtx': {'flags': {'attach': {'args': '<type __builtin__.str>',
                                                   'longname': 'attach',
                                                   'modes': ['multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'a'},
                                        'edit': {'args': '<type __builtin__.bool>',
                                                 'longname': 'edit',
                                                 'numArgs': 0,
                                                 'shortname': 'e'},
                                        'exists': {'args': '<type __builtin__.bool>',
                                                   'longname': 'exists',
                                                   'numArgs': 0,
                                                   'shortname': 'ex'},
                                        'image1': {'args': '<type __builtin__.str>',
                                                   'longname': 'image1',
                                                   'numArgs': 1,
                                                   'shortname': 'i1'},
                                        'image2': {'args': '<type __builtin__.str>',
                                                   'longname': 'image2',
                                                   'numArgs': 1,
                                                   'shortname': 'i2'},
                                        'image3': {'args': '<type __builtin__.str>',
                                                   'longname': 'image3',
                                                   'numArgs': 1,
                                                   'shortname': 'i3'},
                                        'query': {'args': '<type __builtin__.bool>',
                                                  'longname': 'query',
                                                  'numArgs': 0,
                                                  'shortname': 'q'}},
                              'shortFlags': {'a': 'attach',
                                             'e': 'edit',
                                             'ex': 'exists',
                                             'i1': 'image1',
                                             'i2': 'image2',
                                             'i3': 'image3',
                                             'q': 'query'},
                              'type': 'context'},
  'moduleInfo': {'flags': {'definition': {'args': '<type __builtin__.bool>',
                                          'longname': 'definition',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'd'},
                           'listModules': {'args': '<type __builtin__.bool>',
                                           'longname': 'listModules',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'lm'},
                           'moduleName': {'args': '<type __builtin__.str>',
                                          'longname': 'moduleName',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'mn'},
                           'path': {'args': '<type __builtin__.bool>',
                                    'longname': 'path',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'p'},
                           'version': {'args': '<type __builtin__.bool>',
                                       'longname': 'version',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'v'}},
                 'shortFlags': {'d': 'definition',
                                'lm': 'listModules',
                                'mn': 'moduleName',
                                'p': 'path',
                                'v': 'version'},
                 'type': 'system'},
  'mouldMesh': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'mouldSrf': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'mouldSubdiv': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'mouse': {'flags': {'enableScrollWheel': {'args': '<type __builtin__.bool>',
                                            'longname': 'enableScrollWheel',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'esw'},
                      'mouseButtonTracking': {'args': '<type __builtin__.int>',
                                              'longname': 'mouseButtonTracking',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'mbt'},
                      'mouseButtonTrackingStatus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'mouseButtonTrackingStatus',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'mbs'},
                      'scrollWheelStatus': {'args': '<type __builtin__.bool>',
                                            'longname': 'scrollWheelStatus',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'sws'}},
            'shortFlags': {'esw': 'enableScrollWheel',
                           'mbs': 'mouseButtonTrackingStatus',
                           'mbt': 'mouseButtonTracking',
                           'sws': 'scrollWheelStatus'},
            'type': 'system'},
  'movIn': {'flags': {'file': {'args': '<type __builtin__.str>',
                               'longname': 'file',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'f'},
                      'startTime': {'args': 'time',
                                    'longname': 'startTime',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'st'}},
            'shortFlags': {'f': 'file', 'st': 'startTime'},
            'type': 'animation'},
  'movOut': {'flags': {'comment': {'args': '<type __builtin__.bool>',
                                   'longname': 'comment',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'c'},
                       'file': {'args': '<type __builtin__.str>',
                                'longname': 'file',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'f'},
                       'precision': {'args': '<type __builtin__.int>',
                                     'longname': 'precision',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'pre'},
                       'time': {'args': 'timerange',
                                'longname': 'time',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 't'}},
             'shortFlags': {'c': 'comment',
                            'f': 'file',
                            'pre': 'precision',
                            't': 'time'},
             'type': 'animation'},
  'move': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                  'longname': 'absolute',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'a'},
                     'componentOffset': {'args': '<type __builtin__.bool>',
                                         'longname': 'componentOffset',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'co'},
                     'componentSpace': {'args': '<type __builtin__.bool>',
                                        'longname': 'componentSpace',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'cs'},
                     'constrainAlongNormal': {'args': '<type __builtin__.bool>',
                                              'longname': 'constrainAlongNormal',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'xn'},
                     'deletePriorHistory': {'args': '<type __builtin__.bool>',
                                            'longname': 'deletePriorHistory',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'dph'},
                     'localSpace': {'args': '<type __builtin__.bool>',
                                    'longname': 'localSpace',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ls'},
                     'moveX': {'args': '<type __builtin__.bool>',
                               'longname': 'moveX',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'x'},
                     'moveXY': {'args': '<type __builtin__.bool>',
                                'longname': 'moveXY',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'xy'},
                     'moveXYZ': {'args': '<type __builtin__.bool>',
                                 'longname': 'moveXYZ',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'xyz'},
                     'moveXZ': {'args': '<type __builtin__.bool>',
                                'longname': 'moveXZ',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'xz'},
                     'moveY': {'args': '<type __builtin__.bool>',
                               'longname': 'moveY',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'y'},
                     'moveYZ': {'args': '<type __builtin__.bool>',
                                'longname': 'moveYZ',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'yz'},
                     'moveZ': {'args': '<type __builtin__.bool>',
                               'longname': 'moveZ',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'z'},
                     'objectSpace': {'args': '<type __builtin__.bool>',
                                     'longname': 'objectSpace',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'os'},
                     'orientJoint': {'args': '<type __builtin__.str>',
                                     'longname': 'orientJoint',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'oj'},
                     'parameter': {'args': '<type __builtin__.bool>',
                                   'longname': 'parameter',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'p'},
                     'preserveChildPosition': {'args': '<type __builtin__.bool>',
                                               'longname': 'preserveChildPosition',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'pcp'},
                     'preserveGeometryPosition': {'args': '<type __builtin__.bool>',
                                                  'longname': 'preserveGeometryPosition',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'pgp'},
                     'preserveUV': {'args': '<type __builtin__.bool>',
                                    'longname': 'preserveUV',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'puv'},
                     'reflection': {'args': '<type __builtin__.bool>',
                                    'longname': 'reflection',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'rfl'},
                     'reflectionAboutBBox': {'args': '<type __builtin__.bool>',
                                             'longname': 'reflectionAboutBBox',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'rab'},
                     'reflectionAboutOrigin': {'args': '<type __builtin__.bool>',
                                               'longname': 'reflectionAboutOrigin',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'rao'},
                     'reflectionAboutX': {'args': '<type __builtin__.bool>',
                                          'longname': 'reflectionAboutX',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'rax'},
                     'reflectionAboutY': {'args': '<type __builtin__.bool>',
                                          'longname': 'reflectionAboutY',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ray'},
                     'reflectionAboutZ': {'args': '<type __builtin__.bool>',
                                          'longname': 'reflectionAboutZ',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'raz'},
                     'reflectionTolerance': {'args': '<type __builtin__.float>',
                                             'longname': 'reflectionTolerance',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'rft'},
                     'relative': {'args': '<type __builtin__.bool>',
                                  'longname': 'relative',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'r'},
                     'rotatePivotRelative': {'args': '<type __builtin__.bool>',
                                             'longname': 'rotatePivotRelative',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'rpr'},
                     'scalePivotRelative': {'args': '<type __builtin__.bool>',
                                            'longname': 'scalePivotRelative',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'spr'},
                     'secondaryAxisOrient': {'args': '<type __builtin__.str>',
                                             'longname': 'secondaryAxisOrient',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'sao'},
                     'symNegative': {'args': '<type __builtin__.bool>',
                                     'longname': 'symNegative',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'smn'},
                     'worldSpace': {'args': '<type __builtin__.bool>',
                                    'longname': 'worldSpace',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ws'},
                     'worldSpaceDistance': {'args': '<type __builtin__.bool>',
                                            'longname': 'worldSpaceDistance',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'wd'},
                     'xformConstraint': {'args': '<type __builtin__.str>',
                                         'longname': 'xformConstraint',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'xc'}},
           'shortFlags': {'a': 'absolute',
                          'co': 'componentOffset',
                          'cs': 'componentSpace',
                          'dph': 'deletePriorHistory',
                          'ls': 'localSpace',
                          'oj': 'orientJoint',
                          'os': 'objectSpace',
                          'p': 'parameter',
                          'pcp': 'preserveChildPosition',
                          'pgp': 'preserveGeometryPosition',
                          'puv': 'preserveUV',
                          'r': 'relative',
                          'rab': 'reflectionAboutBBox',
                          'rao': 'reflectionAboutOrigin',
                          'rax': 'reflectionAboutX',
                          'ray': 'reflectionAboutY',
                          'raz': 'reflectionAboutZ',
                          'rfl': 'reflection',
                          'rft': 'reflectionTolerance',
                          'rpr': 'rotatePivotRelative',
                          'sao': 'secondaryAxisOrient',
                          'smn': 'symNegative',
                          'spr': 'scalePivotRelative',
                          'wd': 'worldSpaceDistance',
                          'ws': 'worldSpace',
                          'x': 'moveX',
                          'xc': 'xformConstraint',
                          'xn': 'constrainAlongNormal',
                          'xy': 'moveXY',
                          'xyz': 'moveXYZ',
                          'xz': 'moveXZ',
                          'y': 'moveY',
                          'yz': 'moveYZ',
                          'z': 'moveZ'},
           'type': 'general'},
  'moveKeyCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'history': {'args': '<type __builtin__.bool>',
                                       'longname': 'history',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ch'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'moveFunction': {'args': '<type __builtin__.str>',
                                            'longname': 'moveFunction',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'mf'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'option': {'args': '<type __builtin__.str>',
                                      'longname': 'option',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'ch': 'history',
                                'e': 'edit',
                                'ex': 'exists',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'mf': 'moveFunction',
                                'n': 'name',
                                'o': 'option',
                                'q': 'query'},
                 'type': 'context'},
  'moveVertexAlongDirection': {'flags': {'direction': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'direction',
                                                       'modes': ['create',
                                                                 'multiuse'],
                                                       'numArgs': 3,
                                                       'shortname': 'd'},
                                         'magnitude': {'args': '<type __builtin__.float>',
                                                       'longname': 'magnitude',
                                                       'modes': ['create',
                                                                 'multiuse'],
                                                       'numArgs': 1,
                                                       'shortname': 'm'},
                                         'normalDirection': {'args': '<type __builtin__.float>',
                                                             'longname': 'normalDirection',
                                                             'modes': ['create',
                                                                       'multiuse'],
                                                             'numArgs': 1,
                                                             'shortname': 'n'},
                                         'uDirection': {'args': '<type __builtin__.float>',
                                                        'longname': 'uDirection',
                                                        'modes': ['create',
                                                                  'multiuse'],
                                                        'numArgs': 1,
                                                        'shortname': 'u'},
                                         'uvNormalDirection': {'args': ['<type __builtin__.float>',
                                                                        '<type __builtin__.float>',
                                                                        '<type __builtin__.float>'],
                                                               'longname': 'uvNormalDirection',
                                                               'modes': ['create',
                                                                         'multiuse'],
                                                               'numArgs': 3,
                                                               'shortname': 'uvn'},
                                         'vDirection': {'args': '<type __builtin__.float>',
                                                        'longname': 'vDirection',
                                                        'modes': ['create',
                                                                  'multiuse'],
                                                        'numArgs': 1,
                                                        'shortname': 'v'}},
                               'shortFlags': {'d': 'direction',
                                              'm': 'magnitude',
                                              'n': 'normalDirection',
                                              'u': 'uDirection',
                                              'uvn': 'uvNormalDirection',
                                              'v': 'vDirection'},
                               'type': 'modeling'},
  'movieCompressor': {'flags': {'hardwareOptions': {'args': '<type __builtin__.bool>',
                                                    'longname': 'hardwareOptions',
                                                    'numArgs': 0,
                                                    'shortname': 'ho'},
                                'softwareOptions': {'args': '<type __builtin__.bool>',
                                                    'longname': 'softwareOptions',
                                                    'numArgs': 0,
                                                    'shortname': 'so'}},
                      'shortFlags': {'ho': 'hardwareOptions',
                                     'so': 'softwareOptions'},
                      'type': 'other'},
  'movieInfo': {'flags': {'counter': {'args': '<type __builtin__.bool>',
                                      'longname': 'counter',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'cn'},
                          'dropFrame': {'args': '<type __builtin__.bool>',
                                        'longname': 'dropFrame',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'df'},
                          'frameCount': {'args': '<type __builtin__.bool>',
                                         'longname': 'frameCount',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'f'},
                          'frameDuration': {'args': '<type __builtin__.bool>',
                                            'longname': 'frameDuration',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'fd'},
                          'height': {'args': '<type __builtin__.bool>',
                                     'longname': 'height',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'h'},
                          'movieTexture': {'args': '<type __builtin__.bool>',
                                           'longname': 'movieTexture',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'mt'},
                          'negTimesOK': {'args': '<type __builtin__.bool>',
                                         'longname': 'negTimesOK',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'nt'},
                          'numFrames': {'args': '<type __builtin__.bool>',
                                        'longname': 'numFrames',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'nf'},
                          'quickTime': {'args': '<type __builtin__.bool>',
                                        'longname': 'quickTime',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'qt'},
                          'timeCode': {'args': '<type __builtin__.bool>',
                                       'longname': 'timeCode',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'tc'},
                          'timeCodeTrack': {'args': '<type __builtin__.bool>',
                                            'longname': 'timeCodeTrack',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'tt'},
                          'timeScale': {'args': '<type __builtin__.bool>',
                                        'longname': 'timeScale',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ts'},
                          'twentyFourHourMax': {'args': '<type __builtin__.bool>',
                                                'longname': 'twentyFourHourMax',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'tf'},
                          'width': {'args': '<type __builtin__.bool>',
                                    'longname': 'width',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'w'}},
                'shortFlags': {'cn': 'counter',
                               'df': 'dropFrame',
                               'f': 'frameCount',
                               'fd': 'frameDuration',
                               'h': 'height',
                               'mt': 'movieTexture',
                               'nf': 'numFrames',
                               'nt': 'negTimesOK',
                               'qt': 'quickTime',
                               'tc': 'timeCode',
                               'tf': 'twentyFourHourMax',
                               'ts': 'timeScale',
                               'tt': 'timeCodeTrack',
                               'w': 'width'},
                'type': 'animation'},
  'mpBirailCtx': {'flags': {'activeNodes': {'args': '<type __builtin__.bool>',
                                            'longname': 'activeNodes',
                                            'numArgs': 0,
                                            'shortname': 'anq'},
                            'autoCreate': {'args': '<type __builtin__.bool>',
                                           'longname': 'autoCreate',
                                           'numArgs': 1,
                                           'shortname': 'ac'},
                            'bldProfileFirst': {'args': '<type __builtin__.bool>',
                                                'longname': 'bldProfileFirst',
                                                'numArgs': 1,
                                                'shortname': 'bpf'},
                            'bldProfileLast': {'args': '<type __builtin__.bool>',
                                               'longname': 'bldProfileLast',
                                               'numArgs': 1,
                                               'shortname': 'bpl'},
                            'bldProfiles': {'args': '<type __builtin__.bool>',
                                            'longname': 'bldProfiles',
                                            'numArgs': 1,
                                            'shortname': 'bp'},
                            'bldRailOne': {'args': '<type __builtin__.bool>',
                                           'longname': 'bldRailOne',
                                           'numArgs': 1,
                                           'shortname': 'br1'},
                            'bldRailTwo': {'args': '<type __builtin__.bool>',
                                           'longname': 'bldRailTwo',
                                           'numArgs': 1,
                                           'shortname': 'br2'},
                            'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'image1': {'args': '<type __builtin__.str>',
                                       'longname': 'image1',
                                       'numArgs': 1,
                                       'shortname': 'i1'},
                            'image2': {'args': '<type __builtin__.str>',
                                       'longname': 'image2',
                                       'numArgs': 1,
                                       'shortname': 'i2'},
                            'image3': {'args': '<type __builtin__.str>',
                                       'longname': 'image3',
                                       'numArgs': 1,
                                       'shortname': 'i3'},
                            'immediate': {'args': '<type __builtin__.bool>',
                                          'longname': 'immediate',
                                          'numArgs': 0,
                                          'shortname': 'im'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'polygon': {'args': '<type __builtin__.int>',
                                        'longname': 'polygon',
                                        'numArgs': 1,
                                        'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'reset': {'args': '<type __builtin__.bool>',
                                      'longname': 'reset',
                                      'numArgs': 0,
                                      'shortname': 'rs'},
                            'tangentContinuityProfile1': {'args': '<type __builtin__.bool>',
                                                          'longname': 'tangentContinuityProfile1',
                                                          'numArgs': 1,
                                                          'shortname': 'tp1'},
                            'tangentContinuityProfile2': {'args': '<type __builtin__.bool>',
                                                          'longname': 'tangentContinuityProfile2',
                                                          'numArgs': 1,
                                                          'shortname': 'tp2'},
                            'toolNode': {'args': '<type __builtin__.bool>',
                                         'longname': 'toolNode',
                                         'numArgs': 0,
                                         'shortname': 'tnq'},
                            'transformMode': {'args': '<type __builtin__.int>',
                                              'longname': 'transformMode',
                                              'numArgs': 1,
                                              'shortname': 'tm'}},
                  'shortFlags': {'ac': 'autoCreate',
                                 'anq': 'activeNodes',
                                 'bp': 'bldProfiles',
                                 'bpf': 'bldProfileFirst',
                                 'bpl': 'bldProfileLast',
                                 'br1': 'bldRailOne',
                                 'br2': 'bldRailTwo',
                                 'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'e': 'edit',
                                 'ex': 'exists',
                                 'fzn': 'frozen',
                                 'i1': 'image1',
                                 'i2': 'image2',
                                 'i3': 'image3',
                                 'im': 'immediate',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'o': 'object',
                                 'po': 'polygon',
                                 'q': 'query',
                                 'rs': 'reset',
                                 'tm': 'transformMode',
                                 'tnq': 'toolNode',
                                 'tp1': 'tangentContinuityProfile1',
                                 'tp2': 'tangentContinuityProfile2'},
                  'type': 'context'},
  'mrMapVisualizer': {'type': 'runtime'},
  'mrShaderManager': {'type': 'runtime'},
  'multiProfileBirailSurface': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                                      'longname': 'caching',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cch'},
                                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                                  'longname': 'constructionHistory',
                                                                  'modes': ['create'],
                                                                  'numArgs': 1,
                                                                  'shortname': 'ch'},
                                          'edit': {'args': '<type __builtin__.bool>',
                                                   'longname': 'edit',
                                                   'numArgs': 0,
                                                   'shortname': 'e'},
                                          'frozen': {'args': '<type __builtin__.bool>',
                                                     'longname': 'frozen',
                                                     'numArgs': 1,
                                                     'shortname': 'fzn'},
                                          'name': {'args': '<type __builtin__.str>',
                                                   'longname': 'name',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'n'},
                                          'nodeState': {'args': '<type __builtin__.int>',
                                                        'longname': 'nodeState',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'nds'},
                                          'object': {'args': '<type __builtin__.bool>',
                                                     'longname': 'object',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'o'},
                                          'polygon': {'args': '<type __builtin__.int>',
                                                      'longname': 'polygon',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'po'},
                                          'query': {'args': '<type __builtin__.bool>',
                                                    'longname': 'query',
                                                    'numArgs': 0,
                                                    'shortname': 'q'},
                                          'tangentContinuityProfile1': {'args': '<type __builtin__.bool>',
                                                                        'longname': 'tangentContinuityProfile1',
                                                                        'modes': ['create',
                                                                                  'query',
                                                                                  'edit'],
                                                                        'numArgs': 1,
                                                                        'shortname': 'tp1'},
                                          'tangentContinuityProfile2': {'args': '<type __builtin__.bool>',
                                                                        'longname': 'tangentContinuityProfile2',
                                                                        'modes': ['create',
                                                                                  'query',
                                                                                  'edit'],
                                                                        'numArgs': 1,
                                                                        'shortname': 'tp2'},
                                          'transformMode': {'args': '<type __builtin__.int>',
                                                            'longname': 'transformMode',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'tm'}},
                                'shortFlags': {'cch': 'caching',
                                               'ch': 'constructionHistory',
                                               'e': 'edit',
                                               'fzn': 'frozen',
                                               'n': 'name',
                                               'nds': 'nodeState',
                                               'o': 'object',
                                               'po': 'polygon',
                                               'q': 'query',
                                               'tm': 'transformMode',
                                               'tp1': 'tangentContinuityProfile1',
                                               'tp2': 'tangentContinuityProfile2'},
                                'type': 'modeling'},
  'multiTouch': {'flags': {'gestures': {'args': '<type __builtin__.bool>',
                                        'longname': 'gestures',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'g'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'trackpad': {'args': '<type __builtin__.int>',
                                        'longname': 'trackpad',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 't'}},
                 'shortFlags': {'g': 'gestures',
                                'q': 'query',
                                't': 'trackpad'},
                 'type': 'windows'},
  'mute': {'flags': {'disable': {'args': '<type __builtin__.bool>',
                                 'longname': 'disable',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'd'},
                     'force': {'args': '<type __builtin__.bool>',
                               'longname': 'force',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'f'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'}},
           'shortFlags': {'d': 'disable', 'f': 'force', 'q': 'query'},
           'type': 'animation'},
  'myTestCmd': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'nBase': {'flags': {'clearCachedTextureMap': {'args': '<type __builtin__.str>',
                                                'longname': 'clearCachedTextureMap',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cct'},
                      'clearStart': {'args': '<type __builtin__.bool>',
                                     'longname': 'clearStart',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'cs'},
                      'edit': {'args': '<type __builtin__.bool>',
                               'longname': 'edit',
                               'numArgs': 0,
                               'shortname': 'e'},
                      'query': {'args': '<type __builtin__.bool>',
                                'longname': 'query',
                                'numArgs': 0,
                                'shortname': 'q'},
                      'stuffStart': {'args': '<type __builtin__.bool>',
                                     'longname': 'stuffStart',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'ss'},
                      'textureToVertex': {'args': '<type __builtin__.str>',
                                          'longname': 'textureToVertex',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ttv'}},
            'shortFlags': {'cct': 'clearCachedTextureMap',
                           'cs': 'clearStart',
                           'e': 'edit',
                           'q': 'query',
                           'ss': 'stuffStart',
                           'ttv': 'textureToVertex'},
            'type': 'effects'},
  'nClothAppend': {'type': 'runtime'},
  'nClothAppendOpt': {'type': 'runtime'},
  'nClothCache': {'type': 'runtime'},
  'nClothCacheOpt': {'type': 'runtime'},
  'nClothCreate': {'type': 'runtime'},
  'nClothCreateOptions': {'type': 'runtime'},
  'nClothDeleteCacheFrames': {'type': 'runtime'},
  'nClothDeleteCacheFramesOpt': {'type': 'runtime'},
  'nClothDeleteCacheOpt': {'type': 'runtime'},
  'nClothDeleteHistory': {'type': 'runtime'},
  'nClothDeleteHistoryOpt': {'type': 'runtime'},
  'nClothDisplayCurrentMesh': {'type': 'runtime'},
  'nClothDisplayInputMesh': {'type': 'runtime'},
  'nClothLocalToWorld': {'type': 'runtime'},
  'nClothMakeCollide': {'type': 'runtime'},
  'nClothMakeCollideOptions': {'type': 'runtime'},
  'nClothMergeCache': {'type': 'runtime'},
  'nClothMergeCacheOpt': {'type': 'runtime'},
  'nClothRemove': {'type': 'runtime'},
  'nClothReplaceCache': {'type': 'runtime'},
  'nClothReplaceCacheOpt': {'type': 'runtime'},
  'nClothReplaceFrames': {'type': 'runtime'},
  'nClothReplaceFramesOpt': {'type': 'runtime'},
  'nClothRestToInput': {'type': 'runtime'},
  'nClothRestToInputStart': {'type': 'runtime'},
  'nClothRestToMesh': {'type': 'runtime'},
  'nClothWorldToLocal': {'type': 'runtime'},
  'nConstraintAddMembers': {'type': 'runtime'},
  'nConstraintAttractToMatch': {'type': 'runtime'},
  'nConstraintAttractToMatchOptions': {'type': 'runtime'},
  'nConstraintCollisionExclusion': {'type': 'runtime'},
  'nConstraintCollisionExclusionOptions': {'type': 'runtime'},
  'nConstraintComponent': {'type': 'runtime'},
  'nConstraintComponentOptions': {'type': 'runtime'},
  'nConstraintComponentToComponent': {'type': 'runtime'},
  'nConstraintComponentToComponentOptions': {'type': 'runtime'},
  'nConstraintConstraintMembershipTool': {'type': 'runtime'},
  'nConstraintDisableCollision': {'type': 'runtime'},
  'nConstraintDisableCollisionOptions': {'type': 'runtime'},
  'nConstraintForceField': {'type': 'runtime'},
  'nConstraintForceFieldOptions': {'type': 'runtime'},
  'nConstraintPointToSurface': {'type': 'runtime'},
  'nConstraintPointToSurfaceOptions': {'type': 'runtime'},
  'nConstraintRemoveMembers': {'type': 'runtime'},
  'nConstraintReplaceMembers': {'type': 'runtime'},
  'nConstraintSelectMembers': {'type': 'runtime'},
  'nConstraintSlideOnSurface': {'type': 'runtime'},
  'nConstraintSlideOnSurfaceOptions': {'type': 'runtime'},
  'nConstraintTearableSurface': {'type': 'runtime'},
  'nConstraintTearableSurfaceOptions': {'type': 'runtime'},
  'nConstraintTransform': {'type': 'runtime'},
  'nConstraintTransformOptions': {'type': 'runtime'},
  'nConstraintWeldBorders': {'type': 'runtime'},
  'nConstraintWeldBordersOptions': {'type': 'runtime'},
  'nParticle': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                        'longname': 'attribute',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'at'},
                          'cache': {'args': '<type __builtin__.bool>',
                                    'longname': 'cache',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'ch'},
                          'conserve': {'args': '<type __builtin__.float>',
                                       'longname': 'conserve',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'c'},
                          'count': {'args': '<type __builtin__.bool>',
                                    'longname': 'count',
                                    'modes': ['query'],
                                    'numArgs': 0,
                                    'shortname': 'ct'},
                          'deleteCache': {'args': '<type __builtin__.bool>',
                                          'longname': 'deleteCache',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'dc'},
                          'dynamicAttrList': {'args': '<type __builtin__.bool>',
                                              'longname': 'dynamicAttrList',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'dal'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'floatValue': {'args': '<type __builtin__.float>',
                                         'longname': 'floatValue',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 'fv'},
                          'gridSpacing': {'args': '<type __builtin__.float>',
                                          'longname': 'gridSpacing',
                                          'modes': ['create',
                                                    'query',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'grs'},
                          'inherit': {'args': '<type __builtin__.float>',
                                      'longname': 'inherit',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i'},
                          'jitterBasePoint': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'jitterBasePoint',
                                              'modes': ['create',
                                                        'query',
                                                        'multiuse'],
                                              'numArgs': 3,
                                              'shortname': 'jbp'},
                          'jitterRadius': {'args': '<type __builtin__.float>',
                                           'longname': 'jitterRadius',
                                           'modes': ['create',
                                                     'query',
                                                     'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'jr'},
                          'lowerLeft': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'lowerLeft',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 3,
                                        'shortname': 'll'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'numJitters': {'args': '<type __builtin__.int>',
                                         'longname': 'numJitters',
                                         'modes': ['create',
                                                   'query',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'nj'},
                          'order': {'args': '<type __builtin__.int>',
                                    'longname': 'order',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'order'},
                          'particleId': {'args': '<type __builtin__.int>',
                                         'longname': 'particleId',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'id'},
                          'perParticleDouble': {'args': '<type __builtin__.bool>',
                                                'longname': 'perParticleDouble',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'ppd'},
                          'perParticleVector': {'args': '<type __builtin__.bool>',
                                                'longname': 'perParticleVector',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'ppv'},
                          'position': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'position',
                                       'modes': ['multiuse'],
                                       'numArgs': 3,
                                       'shortname': 'p'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'shapeName': {'args': '<type __builtin__.str>',
                                        'longname': 'shapeName',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'sn'},
                          'upperRight': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'upperRight',
                                         'modes': ['create',
                                                   'query',
                                                   'multiuse'],
                                         'numArgs': 3,
                                         'shortname': 'ur'},
                          'vectorValue': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'vectorValue',
                                          'modes': ['edit'],
                                          'numArgs': 3,
                                          'shortname': 'vv'}},
                'removedFlags': {'or': 'order'},
                'shortFlags': {'at': 'attribute',
                               'c': 'conserve',
                               'ch': 'cache',
                               'ct': 'count',
                               'dal': 'dynamicAttrList',
                               'dc': 'deleteCache',
                               'e': 'edit',
                               'fv': 'floatValue',
                               'grs': 'gridSpacing',
                               'i': 'inherit',
                               'id': 'particleId',
                               'jbp': 'jitterBasePoint',
                               'jr': 'jitterRadius',
                               'll': 'lowerLeft',
                               'n': 'name',
                               'nj': 'numJitters',
                               'order': 'order',
                               'p': 'position',
                               'ppd': 'perParticleDouble',
                               'ppv': 'perParticleVector',
                               'q': 'query',
                               'sn': 'shapeName',
                               'ur': 'upperRight',
                               'vv': 'vectorValue'},
                'type': 'effects'},
  'nSoft': {'flags': {'convert': {'args': '<type __builtin__.bool>',
                                  'longname': 'convert',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'c'},
                      'duplicate': {'args': '<type __builtin__.bool>',
                                    'longname': 'duplicate',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'd'},
                      'duplicateHistory': {'args': '<type __builtin__.bool>',
                                           'longname': 'duplicateHistory',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'dh'},
                      'goal': {'args': '<type __builtin__.float>',
                               'longname': 'goal',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'g'},
                      'hideOriginal': {'args': '<type __builtin__.bool>',
                                       'longname': 'hideOriginal',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                      'name': {'args': '<type __builtin__.str>',
                               'longname': 'name',
                               'numArgs': 1,
                               'shortname': 'n'},
                      'query': {'args': '<type __builtin__.bool>',
                                'longname': 'query',
                                'numArgs': 0,
                                'shortname': 'q'}},
            'shortFlags': {'c': 'convert',
                           'd': 'duplicate',
                           'dh': 'duplicateHistory',
                           'g': 'goal',
                           'h': 'hideOriginal',
                           'n': 'name',
                           'q': 'query'},
            'type': 'effects'},
  'nameCommand': {'callbackFlags': ['command', 'c'],
                  'flags': {'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'command': {'args': 'script',
                                        'longname': 'command',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'c'},
                            'data1': {'args': '<type __builtin__.str>',
                                      'longname': 'data1',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'da1'},
                            'data2': {'args': '<type __builtin__.str>',
                                      'longname': 'data2',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'da2'},
                            'data3': {'args': '<type __builtin__.str>',
                                      'longname': 'data3',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'da3'},
                            'default': {'args': '<type __builtin__.bool>',
                                        'longname': 'default',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'd'},
                            'sourceType': {'args': '<type __builtin__.str>',
                                           'longname': 'sourceType',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'stp'}},
                  'shortFlags': {'ann': 'annotation',
                                 'c': 'command',
                                 'd': 'default',
                                 'da1': 'data1',
                                 'da2': 'data2',
                                 'da3': 'data3',
                                 'stp': 'sourceType'},
                  'type': 'windows'},
  'nameField': {'callbackFlags': ['dropCallback',
                                  'dpc',
                                  'visibleChangeCommand',
                                  'vcc',
                                  'dragCallback',
                                  'dgc',
                                  'receiveFocusCommand',
                                  'rfc',
                                  'changeCommand',
                                  'cc',
                                  'nameChangeCommand',
                                  'ncc'],
                'flags': {'annotation': {'args': '<type __builtin__.str>',
                                         'longname': 'annotation',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ann'},
                          'backgroundColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'backgroundColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'bgc'},
                          'changeCommand': {'args': 'script',
                                            'longname': 'changeCommand',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cc'},
                          'defineTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'defineTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dt'},
                          'docTag': {'args': '<type __builtin__.str>',
                                     'longname': 'docTag',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'dtg'},
                          'dragCallback': {'args': 'script',
                                           'longname': 'dragCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dgc'},
                          'drawInactiveFrame': {'args': '<type __builtin__.bool>',
                                                'longname': 'drawInactiveFrame',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dif'},
                          'dropCallback': {'args': 'script',
                                           'longname': 'dropCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dpc'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'enable': {'args': '<type __builtin__.bool>',
                                     'longname': 'enable',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'en'},
                          'enableBackground': {'args': '<type __builtin__.bool>',
                                               'longname': 'enableBackground',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ebg'},
                          'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableKeyboardFocus',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ekf'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'fullPathName': {'args': '<type __builtin__.bool>',
                                           'longname': 'fullPathName',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'fpn'},
                          'height': {'args': '<type __builtin__.int>',
                                     'longname': 'height',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'h'},
                          'highlightColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'highlightColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'hlc'},
                          'isObscured': {'args': '<type __builtin__.bool>',
                                         'longname': 'isObscured',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'io'},
                          'manage': {'args': '<type __builtin__.bool>',
                                     'longname': 'manage',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'm'},
                          'nameChangeCommand': {'args': 'script',
                                                'longname': 'nameChangeCommand',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ncc'},
                          'noBackground': {'args': '<type __builtin__.bool>',
                                           'longname': 'noBackground',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nbg'},
                          'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'numberOfPopupMenus',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'npm'},
                          'object': {'args': '<type __builtin__.str>',
                                     'longname': 'object',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'o'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                          'popupMenuArray': {'args': '<type __builtin__.bool>',
                                             'longname': 'popupMenuArray',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'pma'},
                          'preventOverride': {'args': '<type __builtin__.bool>',
                                              'longname': 'preventOverride',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'po'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'receiveFocusCommand': {'args': 'script',
                                                  'longname': 'receiveFocusCommand',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'rfc'},
                          'statusBarMessage': {'args': '<type __builtin__.str>',
                                               'longname': 'statusBarMessage',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sbm'},
                          'useTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'useTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ut'},
                          'visible': {'args': '<type __builtin__.bool>',
                                      'longname': 'visible',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'vis'},
                          'visibleChangeCommand': {'args': 'script',
                                                   'longname': 'visibleChangeCommand',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'vcc'},
                          'width': {'args': '<type __builtin__.int>',
                                    'longname': 'width',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'w'}},
                'shortFlags': {'ann': 'annotation',
                               'bgc': 'backgroundColor',
                               'cc': 'changeCommand',
                               'dgc': 'dragCallback',
                               'dif': 'drawInactiveFrame',
                               'dpc': 'dropCallback',
                               'dt': 'defineTemplate',
                               'dtg': 'docTag',
                               'e': 'edit',
                               'ebg': 'enableBackground',
                               'ekf': 'enableKeyboardFocus',
                               'en': 'enable',
                               'ex': 'exists',
                               'fpn': 'fullPathName',
                               'h': 'height',
                               'hlc': 'highlightColor',
                               'io': 'isObscured',
                               'm': 'manage',
                               'nbg': 'noBackground',
                               'ncc': 'nameChangeCommand',
                               'npm': 'numberOfPopupMenus',
                               'o': 'object',
                               'p': 'parent',
                               'pma': 'popupMenuArray',
                               'po': 'preventOverride',
                               'q': 'query',
                               'rfc': 'receiveFocusCommand',
                               'sbm': 'statusBarMessage',
                               'ut': 'useTemplate',
                               'vcc': 'visibleChangeCommand',
                               'vis': 'visible',
                               'w': 'width'},
                'type': 'windows'},
  'namespace': {'flags': {'absoluteName': {'args': '<type __builtin__.bool>',
                                           'longname': 'absoluteName',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'an'},
                          'addNamespace': {'args': '<type __builtin__.str>',
                                           'longname': 'addNamespace',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'add'},
                          'collapseAncestors': {'args': '<type __builtin__.str>',
                                                'longname': 'collapseAncestors',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                          'deleteNamespaceContent': {'args': '<type __builtin__.bool>',
                                                     'longname': 'deleteNamespaceContent',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'dnc'},
                          'exists': {'args': '<type __builtin__.str>',
                                     'longname': 'exists',
                                     'modes': ['query'],
                                     'numArgs': 1,
                                     'shortname': 'ex'},
                          'force': {'args': '<type __builtin__.bool>',
                                    'longname': 'force',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'f'},
                          'isRootNamespace': {'args': '<type __builtin__.str>',
                                              'longname': 'isRootNamespace',
                                              'modes': ['query'],
                                              'numArgs': 1,
                                              'shortname': 'ir'},
                          'mergeNamespaceWithParent': {'args': '<type __builtin__.bool>',
                                                       'longname': 'mergeNamespaceWithParent',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'mnp'},
                          'mergeNamespaceWithRoot': {'args': '<type __builtin__.bool>',
                                                     'longname': 'mergeNamespaceWithRoot',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'mnr'},
                          'moveNamespace': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'moveNamespace',
                                            'modes': ['create'],
                                            'numArgs': 2,
                                            'shortname': 'mv'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'recurse': {'args': '<type __builtin__.bool>',
                                      'longname': 'recurse',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'r'},
                          'relativeNames': {'args': '<type __builtin__.bool>',
                                            'longname': 'relativeNames',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'rel'},
                          'removeNamespace': {'args': '<type __builtin__.str>',
                                              'longname': 'removeNamespace',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'rm'},
                          'rename': {'args': ['<type __builtin__.str>',
                                              '<type __builtin__.str>'],
                                     'longname': 'rename',
                                     'modes': ['create'],
                                     'numArgs': 2,
                                     'shortname': 'ren'},
                          'setNamespace': {'args': '<type __builtin__.str>',
                                           'longname': 'setNamespace',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'set'},
                          'validateName': {'args': '<type __builtin__.str>',
                                           'longname': 'validateName',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'vn'}},
                'shortFlags': {'add': 'addNamespace',
                               'an': 'absoluteName',
                               'ca': 'collapseAncestors',
                               'dnc': 'deleteNamespaceContent',
                               'ex': 'exists',
                               'f': 'force',
                               'ir': 'isRootNamespace',
                               'mnp': 'mergeNamespaceWithParent',
                               'mnr': 'mergeNamespaceWithRoot',
                               'mv': 'moveNamespace',
                               'p': 'parent',
                               'q': 'query',
                               'r': 'recurse',
                               'rel': 'relativeNames',
                               'ren': 'rename',
                               'rm': 'removeNamespace',
                               'set': 'setNamespace',
                               'vn': 'validateName'},
                'type': 'system'},
  'namespaceInfo': {'flags': {'absoluteName': {'args': '<type __builtin__.bool>',
                                               'longname': 'absoluteName',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'an'},
                              'baseName': {'args': '<type __builtin__.bool>',
                                           'longname': 'baseName',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'bn'},
                              'currentNamespace': {'args': '<type __builtin__.bool>',
                                                   'longname': 'currentNamespace',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'cur'},
                              'dagPath': {'args': '<type __builtin__.bool>',
                                          'longname': 'dagPath',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'dp'},
                              'fullName': {'args': '<type __builtin__.bool>',
                                           'longname': 'fullName',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'fn'},
                              'internal': {'args': '<type __builtin__.bool>',
                                           'longname': 'internal',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'int'},
                              'isRootNamespace': {'args': '<type __builtin__.bool>',
                                                  'longname': 'isRootNamespace',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'ir'},
                              'listNamespace': {'args': '<type __builtin__.bool>',
                                                'longname': 'listNamespace',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ls'},
                              'listOnlyDependencyNodes': {'args': '<type __builtin__.bool>',
                                                          'longname': 'listOnlyDependencyNodes',
                                                          'modes': ['create'],
                                                          'numArgs': 0,
                                                          'shortname': 'lod'},
                              'listOnlyNamespaces': {'args': '<type __builtin__.bool>',
                                                     'longname': 'listOnlyNamespaces',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'lon'},
                              'parent': {'args': '<type __builtin__.bool>',
                                         'longname': 'parent',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'p'},
                              'recurse': {'args': '<type __builtin__.bool>',
                                          'longname': 'recurse',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'r'},
                              'shortName': {'args': '<type __builtin__.bool>',
                                            'longname': 'shortName',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'sn'}},
                    'shortFlags': {'an': 'absoluteName',
                                   'bn': 'baseName',
                                   'cur': 'currentNamespace',
                                   'dp': 'dagPath',
                                   'fn': 'fullName',
                                   'int': 'internal',
                                   'ir': 'isRootNamespace',
                                   'lod': 'listOnlyDependencyNodes',
                                   'lon': 'listOnlyNamespaces',
                                   'ls': 'listNamespace',
                                   'p': 'parent',
                                   'r': 'recurse',
                                   'sn': 'shortName'},
                    'type': 'system'},
  'newton': {'flags': {'attenuation': {'args': '<type __builtin__.float>',
                                       'longname': 'attenuation',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'att'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'magnitude': {'args': '<type __builtin__.float>',
                                     'longname': 'magnitude',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'm'},
                       'maxDistance': {'args': '<type __builtin__.float>',
                                       'longname': 'maxDistance',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'mxd'},
                       'minDistance': {'args': '<type __builtin__.float>',
                                       'longname': 'minDistance',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'mnd'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'perVertex': {'args': '<type __builtin__.bool>',
                                     'longname': 'perVertex',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'pv'},
                       'position': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'position',
                                    'modes': ['query', 'edit', 'multiuse'],
                                    'numArgs': 3,
                                    'shortname': 'pos'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'torusSectionRadius': {'args': '<type __builtin__.float>',
                                              'longname': 'torusSectionRadius',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'tsr'},
                       'volumeExclusion': {'args': '<type __builtin__.bool>',
                                           'longname': 'volumeExclusion',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vex'},
                       'volumeOffset': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'volumeOffset',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 3,
                                        'shortname': 'vof'},
                       'volumeShape': {'args': '<type __builtin__.str>',
                                       'longname': 'volumeShape',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vsh'},
                       'volumeSweep': {'args': '<type __builtin__.float>',
                                       'longname': 'volumeSweep',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vsw'}},
             'shortFlags': {'att': 'attenuation',
                            'e': 'edit',
                            'm': 'magnitude',
                            'mnd': 'minDistance',
                            'mxd': 'maxDistance',
                            'n': 'name',
                            'pos': 'position',
                            'pv': 'perVertex',
                            'q': 'query',
                            'tsr': 'torusSectionRadius',
                            'vex': 'volumeExclusion',
                            'vof': 'volumeOffset',
                            'vsh': 'volumeShape',
                            'vsw': 'volumeSweep'},
             'type': 'effects'},
  'nodeCast': {'flags': {'copyDynamicAttrs': {'args': '<type __builtin__.bool>',
                                              'longname': 'copyDynamicAttrs',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'cda'},
                         'disableAPICallbacks': {'args': '<type __builtin__.bool>',
                                                 'longname': 'disableAPICallbacks',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'dsa'},
                         'disableScriptJobCallbacks': {'args': '<type __builtin__.bool>',
                                                       'longname': 'disableScriptJobCallbacks',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'dsj'},
                         'disconnectUnmatchedAttrs': {'args': '<type __builtin__.bool>',
                                                      'longname': 'disconnectUnmatchedAttrs',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'dua'},
                         'force': {'args': '<type __builtin__.bool>',
                                   'longname': 'force',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'f'},
                         'swapNames': {'args': '<type __builtin__.bool>',
                                       'longname': 'swapNames',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'sn'},
                         'swapValues': {'args': '<type __builtin__.bool>',
                                        'longname': 'swapValues',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'sv'}},
               'shortFlags': {'cda': 'copyDynamicAttrs',
                              'dsa': 'disableAPICallbacks',
                              'dsj': 'disableScriptJobCallbacks',
                              'dua': 'disconnectUnmatchedAttrs',
                              'f': 'force',
                              'sn': 'swapNames',
                              'sv': 'swapValues'},
               'type': 'general'},
  'nodeEditor': {'callbackFlags': ['tabChangeCommand',
                                   'tcc',
                                   'createNodeCommand',
                                   'cnc',
                                   'filterCreateNodeTypes',
                                   'fcn',
                                   'toolTipCommand',
                                   'ttc',
                                   'layoutCommand',
                                   'lc',
                                   'popupMenuScript',
                                   'pms',
                                   'settingsChangedCallback',
                                   'scc',
                                   'contentsChangedCommand',
                                   'cc',
                                   'keyReleaseCommand',
                                   'krc',
                                   'focusCommand',
                                   'fc',
                                   'keyPressCommand',
                                   'kpc'],
                 'flags': {'activeTab': {'args': '<type __builtin__.int>',
                                         'longname': 'activeTab',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'at'},
                           'addNewNodes': {'args': '<type __builtin__.bool>',
                                           'longname': 'addNewNodes',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                           'addNode': {'args': '<type __builtin__.str>',
                                       'longname': 'addNode',
                                       'modes': ['edit', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'an'},
                           'additiveGraphingMode': {'args': '<type __builtin__.bool>',
                                                    'longname': 'additiveGraphingMode',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'agm'},
                           'allAttributes': {'args': '<type __builtin__.bool>',
                                             'longname': 'allAttributes',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ala'},
                           'allNodes': {'args': '<type __builtin__.bool>',
                                        'longname': 'allNodes',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'aln'},
                           'allowNewTabs': {'args': '<type __builtin__.bool>',
                                            'longname': 'allowNewTabs',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ant'},
                           'allowTabTearoff': {'args': '<type __builtin__.bool>',
                                               'longname': 'allowTabTearoff',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'att'},
                           'autoSizeNodes': {'args': '<type __builtin__.bool>',
                                             'longname': 'autoSizeNodes',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'asn'},
                           'backToParentView': {'args': '<type __builtin__.bool>',
                                                'longname': 'backToParentView',
                                                'numArgs': 0,
                                                'shortname': 'bpv'},
                           'beginCreateNode': {'args': '<type __builtin__.bool>',
                                               'longname': 'beginCreateNode',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'bcn'},
                           'beginNewConnection': {'args': '<type __builtin__.str>',
                                                  'longname': 'beginNewConnection',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'bnc'},
                           'breakSelectedConnections': {'args': '<type __builtin__.bool>',
                                                        'longname': 'breakSelectedConnections',
                                                        'modes': ['edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'bsc'},
                           'closeAllTabs': {'args': '<type __builtin__.bool>',
                                            'longname': 'closeAllTabs',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'cat'},
                           'closeTab': {'args': '<type __builtin__.int>',
                                        'longname': 'closeTab',
                                        'modes': ['edit'],
                                        'numArgs': 1,
                                        'shortname': 'clt'},
                           'connectNodeOnCreation': {'args': '<type __builtin__.bool>',
                                                     'longname': 'connectNodeOnCreation',
                                                     'numArgs': 1,
                                                     'shortname': 'cno'},
                           'connectOnDrop': {'args': '<type __builtin__.bool>',
                                             'longname': 'connectOnDrop',
                                             'numArgs': 1,
                                             'shortname': 'cod'},
                           'connectSelectedNodes': {'args': '<type __builtin__.bool>',
                                                    'longname': 'connectSelectedNodes',
                                                    'modes': ['edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'csn'},
                           'connectionMinSegment': {'args': '<type __builtin__.float>',
                                                    'longname': 'connectionMinSegment',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'csm'},
                           'connectionOffset': {'args': '<type __builtin__.float>',
                                                'longname': 'connectionOffset',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cso'},
                           'connectionRoundness': {'args': '<type __builtin__.float>',
                                                   'longname': 'connectionRoundness',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'csr'},
                           'connectionStyle': {'args': '<type __builtin__.str>',
                                               'longname': 'connectionStyle',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cs'},
                           'connectionTension': {'args': '<type __builtin__.int>',
                                                 'longname': 'connectionTension',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cst'},
                           'consistentNameSize': {'args': '<type __builtin__.bool>',
                                                  'longname': 'consistentNameSize',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cns'},
                           'contentsChangedCommand': {'args': 'script',
                                                      'longname': 'contentsChangedCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cc'},
                           'control': {'args': '<type __builtin__.bool>',
                                       'longname': 'control',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'ctl'},
                           'copyConnectionsOnPaste': {'args': '<type __builtin__.bool>',
                                                      'longname': 'copyConnectionsOnPaste',
                                                      'numArgs': 1,
                                                      'shortname': 'ccp'},
                           'createInfo': {'args': '<type __builtin__.str>',
                                          'longname': 'createInfo',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ci'},
                           'createNodeCommand': {'args': 'script',
                                                 'longname': 'createNodeCommand',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cnc'},
                           'createTab': {'args': ['<type __builtin__.int>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'createTab',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 2,
                                         'shortname': 'ct'},
                           'crosshairOnEdgeDragging': {'args': '<type __builtin__.bool>',
                                                       'longname': 'crosshairOnEdgeDragging',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ced'},
                           'customAttributeListEdit': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'customAttributeListEdit',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 2,
                                                       'shortname': 'cal'},
                           'cycleHUD': {'args': '<type __builtin__.bool>',
                                        'longname': 'cycleHUD',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'ch'},
                           'defaultPinnedState': {'args': '<type __builtin__.bool>',
                                                  'longname': 'defaultPinnedState',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dps'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'deleteSelected': {'args': '<type __builtin__.bool>',
                                              'longname': 'deleteSelected',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'deleteSelected'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'dotFormat': {'args': '<type __builtin__.str>',
                                         'longname': 'dotFormat',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dot'},
                           'downstream': {'args': '<type __builtin__.bool>',
                                          'longname': 'downstream',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'ds'},
                           'duplicateTab': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'duplicateTab',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 2,
                                            'shortname': 'dpt'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'editorMode': {'args': '<type __builtin__.str>',
                                          'longname': 'editorMode',
                                          'numArgs': 1,
                                          'shortname': 'em'},
                           'enableOpenGL': {'args': '<type __builtin__.bool>',
                                            'longname': 'enableOpenGL',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'egl'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'extendToShapes': {'args': '<type __builtin__.bool>',
                                              'longname': 'extendToShapes',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ets'},
                           'feedbackConnection': {'args': '<type __builtin__.bool>',
                                                  'longname': 'feedbackConnection',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'fbc'},
                           'feedbackNode': {'args': '<type __builtin__.bool>',
                                            'longname': 'feedbackNode',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fbn'},
                           'feedbackPlug': {'args': '<type __builtin__.bool>',
                                            'longname': 'feedbackPlug',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fbp'},
                           'feedbackTabIndex': {'args': '<type __builtin__.bool>',
                                                'longname': 'feedbackTabIndex',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'fbi'},
                           'feedbackType': {'args': '<type __builtin__.bool>',
                                            'longname': 'feedbackType',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fbt'},
                           'filter': {'args': '<type __builtin__.str>',
                                      'longname': 'filter',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'f'},
                           'filterCreateNodeTypes': {'args': 'script',
                                                     'longname': 'filterCreateNodeTypes',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'fcn'},
                           'focusCommand': {'args': 'script',
                                            'longname': 'focusCommand',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'fc'},
                           'forceMainConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'forceMainConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'fmc'},
                           'frameAll': {'args': '<type __builtin__.bool>',
                                        'longname': 'frameAll',
                                        'modes': ['edit'],
                                        'numArgs': 0,
                                        'shortname': 'fa'},
                           'frameModelSelection': {'args': '<type __builtin__.bool>',
                                                   'longname': 'frameModelSelection',
                                                   'modes': ['edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'fms'},
                           'frameSelected': {'args': '<type __builtin__.bool>',
                                             'longname': 'frameSelected',
                                             'modes': ['edit'],
                                             'numArgs': 0,
                                             'shortname': 'fs'},
                           'getNodeList': {'args': '<type __builtin__.bool>',
                                           'longname': 'getNodeList',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'gnl'},
                           'graphSelectedConnections': {'args': '<type __builtin__.bool>',
                                                        'longname': 'graphSelectedConnections',
                                                        'modes': ['edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'gsc'},
                           'graphSelection': {'args': '<type __builtin__.bool>',
                                              'longname': 'graphSelection',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'gsl'},
                           'gridSnap': {'args': '<type __builtin__.bool>',
                                        'longname': 'gridSnap',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'gs'},
                           'gridVisibility': {'args': '<type __builtin__.bool>',
                                              'longname': 'gridVisibility',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'gv'},
                           'hasWatchpoint': {'args': '<type __builtin__.bool>',
                                             'longname': 'hasWatchpoint',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'hw'},
                           'highlightConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'highlightConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'hlc'},
                           'highlightConnections': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.bool>'],
                                                    'longname': 'highlightConnections',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 2,
                                                    'shortname': 'hc'},
                           'hudMessage': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.int>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'hudMessage',
                                          'modes': ['edit'],
                                          'numArgs': 3,
                                          'shortname': 'hm'},
                           'ignoreAssets': {'args': '<type __builtin__.bool>',
                                            'longname': 'ignoreAssets',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ia'},
                           'inContainerView': {'args': '<type __builtin__.bool>',
                                               'longname': 'inContainerView',
                                               'numArgs': 0,
                                               'shortname': 'icv'},
                           'isContainerNode': {'args': '<type __builtin__.str>',
                                               'longname': 'isContainerNode',
                                               'numArgs': 1,
                                               'shortname': 'icn'},
                           'island': {'args': '<type __builtin__.bool>',
                                      'longname': 'island',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'isl'},
                           'keyPressCommand': {'args': 'script',
                                               'longname': 'keyPressCommand',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'kpc'},
                           'keyReleaseCommand': {'args': 'script',
                                                 'longname': 'keyReleaseCommand',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'krc'},
                           'layout': {'args': '<type __builtin__.bool>',
                                      'longname': 'layout',
                                      'modes': ['edit'],
                                      'numArgs': 0,
                                      'shortname': 'lay'},
                           'layoutCommand': {'args': 'script',
                                             'longname': 'layoutCommand',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lc'},
                           'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                  'longname': 'lockMainConnection',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'lck'},
                           'mainListConnection': {'args': '<type __builtin__.str>',
                                                  'longname': 'mainListConnection',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mlc'},
                           'nodeSwatchSize': {'args': '<type __builtin__.str>',
                                              'longname': 'nodeSwatchSize',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'nss'},
                           'nodeTitleMode': {'args': '<type __builtin__.str>',
                                             'longname': 'nodeTitleMode',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ntm'},
                           'nodeViewMode': {'args': '<type __builtin__.str>',
                                            'longname': 'nodeViewMode',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'nvm'},
                           'openContainerView': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.bool>'],
                                                 'longname': 'openContainerView',
                                                 'numArgs': 2,
                                                 'shortname': 'ocv'},
                           'overrideNodeDropPosition': {'args': ['<type __builtin__.float>',
                                                                 '<type __builtin__.float>'],
                                                        'longname': 'overrideNodeDropPosition',
                                                        'modes': ['edit'],
                                                        'numArgs': 2,
                                                        'shortname': 'onp'},
                           'panView': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'panView',
                                       'modes': ['edit'],
                                       'numArgs': 2,
                                       'shortname': 'pv'},
                           'panel': {'args': '<type __builtin__.str>',
                                     'longname': 'panel',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'pnl'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'pinSelectedNodes': {'args': '<type __builtin__.bool>',
                                                'longname': 'pinSelectedNodes',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'psn'},
                           'popupMenuScript': {'args': 'script',
                                               'longname': 'popupMenuScript',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'pms'},
                           'primary': {'args': '<type __builtin__.bool>',
                                       'longname': 'primary',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'pr'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'redockTab': {'args': '<type __builtin__.bool>',
                                         'longname': 'redockTab',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'rdt'},
                           'removeDownstream': {'args': '<type __builtin__.bool>',
                                                'longname': 'removeDownstream',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'rd'},
                           'removeNode': {'args': '<type __builtin__.str>',
                                          'longname': 'removeNode',
                                          'modes': ['edit', 'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'rem'},
                           'removeUnselected': {'args': '<type __builtin__.bool>',
                                                'longname': 'removeUnselected',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'run'},
                           'removeUpstream': {'args': '<type __builtin__.bool>',
                                              'longname': 'removeUpstream',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'ru'},
                           'renameNode': {'args': '<type __builtin__.str>',
                                          'longname': 'renameNode',
                                          'modes': ['edit'],
                                          'numArgs': 1,
                                          'shortname': 'ren'},
                           'renameTab': {'args': ['<type __builtin__.int>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'renameTab',
                                         'modes': ['edit'],
                                         'numArgs': 2,
                                         'shortname': 'rt'},
                           'restoreInfo': {'args': '<type __builtin__.str>',
                                           'longname': 'restoreInfo',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ri'},
                           'restoreLastClosedTab': {'args': '<type __builtin__.bool>',
                                                    'longname': 'restoreLastClosedTab',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'rlt'},
                           'rootNode': {'args': '<type __builtin__.str>',
                                        'longname': 'rootNode',
                                        'modes': ['create',
                                                  'edit',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'rn'},
                           'rootsFromSelection': {'args': '<type __builtin__.bool>',
                                                  'longname': 'rootsFromSelection',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'rfs'},
                           'scaleView': {'args': '<type __builtin__.float>',
                                         'longname': 'scaleView',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 'sv'},
                           'selectAll': {'args': '<type __builtin__.bool>',
                                         'longname': 'selectAll',
                                         'modes': ['edit'],
                                         'numArgs': 0,
                                         'shortname': 'sa'},
                           'selectConnectionNodes': {'args': '<type __builtin__.bool>',
                                                     'longname': 'selectConnectionNodes',
                                                     'modes': ['edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'scn'},
                           'selectDownstream': {'args': '<type __builtin__.bool>',
                                                'longname': 'selectDownstream',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'sd'},
                           'selectFeedbackConnection': {'args': '<type __builtin__.bool>',
                                                        'longname': 'selectFeedbackConnection',
                                                        'modes': ['edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'sfc'},
                           'selectNode': {'args': '<type __builtin__.str>',
                                          'longname': 'selectNode',
                                          'modes': ['query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'sln'},
                           'selectUpstream': {'args': '<type __builtin__.bool>',
                                              'longname': 'selectUpstream',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'su'},
                           'selectionConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'selectionConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'slc'},
                           'setWatchpoint': {'args': '<type __builtin__.bool>',
                                             'longname': 'setWatchpoint',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sw'},
                           'settingsChangedCallback': {'args': 'script',
                                                       'longname': 'settingsChangedCallback',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'scc'},
                           'shaderNetworks': {'args': '<type __builtin__.bool>',
                                              'longname': 'shaderNetworks',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'sns'},
                           'showAllNodeAttributes': {'args': '<type __builtin__.str>',
                                                     'longname': 'showAllNodeAttributes',
                                                     'modes': ['edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'saa'},
                           'showNamespace': {'args': '<type __builtin__.bool>',
                                             'longname': 'showNamespace',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sn'},
                           'showSGShapes': {'args': '<type __builtin__.bool>',
                                            'longname': 'showSGShapes',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ssg'},
                           'showShapes': {'args': '<type __builtin__.bool>',
                                          'longname': 'showShapes',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ss'},
                           'showTabs': {'args': '<type __builtin__.bool>',
                                        'longname': 'showTabs',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'tab'},
                           'showTransforms': {'args': '<type __builtin__.bool>',
                                              'longname': 'showTransforms',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'st'},
                           'stateString': {'args': '<type __builtin__.bool>',
                                           'longname': 'stateString',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'sts'},
                           'syncedSelection': {'args': '<type __builtin__.bool>',
                                               'longname': 'syncedSelection',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ssl'},
                           'tabChangeCommand': {'args': 'script',
                                                'longname': 'tabChangeCommand',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'tcc'},
                           'toggleAttrFilter': {'args': '<type __builtin__.bool>',
                                                'longname': 'toggleAttrFilter',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'taf'},
                           'toggleSelectedPins': {'args': '<type __builtin__.bool>',
                                                  'longname': 'toggleSelectedPins',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'tsp'},
                           'toggleSwatchSize': {'args': '<type __builtin__.str>',
                                                'longname': 'toggleSwatchSize',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'tss'},
                           'toolTipCommand': {'args': 'script',
                                              'longname': 'toolTipCommand',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ttc'},
                           'traversalDepthLimit': {'args': '<type __builtin__.int>',
                                                   'longname': 'traversalDepthLimit',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'tdl'},
                           'unParent': {'args': '<type __builtin__.bool>',
                                        'longname': 'unParent',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'up'},
                           'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'unlockMainConnection',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'ulk'},
                           'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'updateMainConnection',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'upd'},
                           'upstream': {'args': '<type __builtin__.bool>',
                                        'longname': 'upstream',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'ups'},
                           'useAssets': {'args': '<type __builtin__.bool>',
                                         'longname': 'useAssets',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ua'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'},
                           'vnnCompound': {'args': '<type __builtin__.bool>',
                                           'longname': 'vnnCompound',
                                           'numArgs': 0,
                                           'shortname': 'vc'},
                           'vnnDgContainer': {'args': '<type __builtin__.bool>',
                                              'longname': 'vnnDgContainer',
                                              'numArgs': 0,
                                              'shortname': 'vd'},
                           'vnnRuntime': {'args': '<type __builtin__.bool>',
                                          'longname': 'vnnRuntime',
                                          'numArgs': 0,
                                          'shortname': 'vr'}},
                 'removedFlags': {'del': 'deleteSelected'},
                 'shortFlags': {'agm': 'additiveGraphingMode',
                                'ala': 'allAttributes',
                                'aln': 'allNodes',
                                'an': 'addNode',
                                'ann': 'addNewNodes',
                                'ant': 'allowNewTabs',
                                'asn': 'autoSizeNodes',
                                'at': 'activeTab',
                                'att': 'allowTabTearoff',
                                'bcn': 'beginCreateNode',
                                'bnc': 'beginNewConnection',
                                'bpv': 'backToParentView',
                                'bsc': 'breakSelectedConnections',
                                'cal': 'customAttributeListEdit',
                                'cat': 'closeAllTabs',
                                'cc': 'contentsChangedCommand',
                                'ccp': 'copyConnectionsOnPaste',
                                'ced': 'crosshairOnEdgeDragging',
                                'ch': 'cycleHUD',
                                'ci': 'createInfo',
                                'clt': 'closeTab',
                                'cnc': 'createNodeCommand',
                                'cno': 'connectNodeOnCreation',
                                'cns': 'consistentNameSize',
                                'cod': 'connectOnDrop',
                                'cs': 'connectionStyle',
                                'csm': 'connectionMinSegment',
                                'csn': 'connectSelectedNodes',
                                'cso': 'connectionOffset',
                                'csr': 'connectionRoundness',
                                'cst': 'connectionTension',
                                'ct': 'createTab',
                                'ctl': 'control',
                                'deleteSelected': 'deleteSelected',
                                'dot': 'dotFormat',
                                'dps': 'defaultPinnedState',
                                'dpt': 'duplicateTab',
                                'ds': 'downstream',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'egl': 'enableOpenGL',
                                'em': 'editorMode',
                                'ets': 'extendToShapes',
                                'ex': 'exists',
                                'f': 'filter',
                                'fa': 'frameAll',
                                'fbc': 'feedbackConnection',
                                'fbi': 'feedbackTabIndex',
                                'fbn': 'feedbackNode',
                                'fbp': 'feedbackPlug',
                                'fbt': 'feedbackType',
                                'fc': 'focusCommand',
                                'fcn': 'filterCreateNodeTypes',
                                'fmc': 'forceMainConnection',
                                'fms': 'frameModelSelection',
                                'fs': 'frameSelected',
                                'gnl': 'getNodeList',
                                'gs': 'gridSnap',
                                'gsc': 'graphSelectedConnections',
                                'gsl': 'graphSelection',
                                'gv': 'gridVisibility',
                                'hc': 'highlightConnections',
                                'hlc': 'highlightConnection',
                                'hm': 'hudMessage',
                                'hw': 'hasWatchpoint',
                                'ia': 'ignoreAssets',
                                'icn': 'isContainerNode',
                                'icv': 'inContainerView',
                                'isl': 'island',
                                'kpc': 'keyPressCommand',
                                'krc': 'keyReleaseCommand',
                                'lay': 'layout',
                                'lc': 'layoutCommand',
                                'lck': 'lockMainConnection',
                                'mlc': 'mainListConnection',
                                'nss': 'nodeSwatchSize',
                                'ntm': 'nodeTitleMode',
                                'nvm': 'nodeViewMode',
                                'ocv': 'openContainerView',
                                'onp': 'overrideNodeDropPosition',
                                'p': 'parent',
                                'pms': 'popupMenuScript',
                                'pnl': 'panel',
                                'pr': 'primary',
                                'psn': 'pinSelectedNodes',
                                'pv': 'panView',
                                'q': 'query',
                                'rd': 'removeDownstream',
                                'rdt': 'redockTab',
                                'rem': 'removeNode',
                                'ren': 'renameNode',
                                'rfs': 'rootsFromSelection',
                                'ri': 'restoreInfo',
                                'rlt': 'restoreLastClosedTab',
                                'rn': 'rootNode',
                                'rt': 'renameTab',
                                'ru': 'removeUpstream',
                                'run': 'removeUnselected',
                                'sa': 'selectAll',
                                'saa': 'showAllNodeAttributes',
                                'scc': 'settingsChangedCallback',
                                'scn': 'selectConnectionNodes',
                                'sd': 'selectDownstream',
                                'sfc': 'selectFeedbackConnection',
                                'slc': 'selectionConnection',
                                'sln': 'selectNode',
                                'sn': 'showNamespace',
                                'sns': 'shaderNetworks',
                                'ss': 'showShapes',
                                'ssg': 'showSGShapes',
                                'ssl': 'syncedSelection',
                                'st': 'showTransforms',
                                'sts': 'stateString',
                                'su': 'selectUpstream',
                                'sv': 'scaleView',
                                'sw': 'setWatchpoint',
                                'tab': 'showTabs',
                                'taf': 'toggleAttrFilter',
                                'tcc': 'tabChangeCommand',
                                'tdl': 'traversalDepthLimit',
                                'tsp': 'toggleSelectedPins',
                                'tss': 'toggleSwatchSize',
                                'ttc': 'toolTipCommand',
                                'ua': 'useAssets',
                                'ulk': 'unlockMainConnection',
                                'up': 'unParent',
                                'upd': 'updateMainConnection',
                                'ups': 'upstream',
                                'ut': 'useTemplate',
                                'vc': 'vnnCompound',
                                'vd': 'vnnDgContainer',
                                'vr': 'vnnRuntime'},
                 'type': 'windows'},
  'nodeGrapher': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'nodeIconButton': {'callbackFlags': ['visibleChangeCommand',
                                       'vcc',
                                       'dragCallback',
                                       'dgc',
                                       'command',
                                       'c',
                                       'dropCallback',
                                       'dpc'],
                     'flags': {'align': {'args': '<type __builtin__.str>',
                                         'longname': 'align',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'al'},
                               'annotation': {'args': '<type __builtin__.str>',
                                              'longname': 'annotation',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ann'},
                               'backgroundColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'backgroundColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'bgc'},
                               'command': {'args': 'script',
                                           'longname': 'command',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'c'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'disabledImage': {'args': '<type __builtin__.str>',
                                                 'longname': 'disabledImage',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'di'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'dragCallback': {'args': 'script',
                                                'longname': 'dragCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dgc'},
                               'dropCallback': {'args': 'script',
                                                'longname': 'dropCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dpc'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'enable': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en'},
                               'enableBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableBackground',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ebg'},
                               'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                       'longname': 'enableKeyboardFocus',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ekf'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'flipX': {'args': '<type __builtin__.bool>',
                                         'longname': 'flipX',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'fx'},
                               'flipY': {'args': '<type __builtin__.bool>',
                                         'longname': 'flipY',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'fy'},
                               'font': {'args': '<type __builtin__.str>',
                                        'longname': 'font',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'fn'},
                               'fullPathName': {'args': '<type __builtin__.bool>',
                                                'longname': 'fullPathName',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'fpn'},
                               'height': {'args': '<type __builtin__.int>',
                                          'longname': 'height',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'h'},
                               'highlightColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'highlightColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'hlc'},
                               'image': {'args': '<type __builtin__.str>',
                                         'longname': 'image',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'imageOverlayLabel': {'args': '<type __builtin__.str>',
                                                     'longname': 'imageOverlayLabel',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'iol'},
                               'isObscured': {'args': '<type __builtin__.bool>',
                                              'longname': 'isObscured',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'io'},
                               'label': {'args': '<type __builtin__.str>',
                                         'longname': 'label',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'l'},
                               'labelOffset': {'args': '<type __builtin__.int>',
                                               'longname': 'labelOffset',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'lo'},
                               'ltVersion': {'args': '<type __builtin__.str>',
                                             'longname': 'ltVersion',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lt'},
                               'manage': {'args': '<type __builtin__.bool>',
                                          'longname': 'manage',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'm'},
                               'marginHeight': {'args': '<type __builtin__.int>',
                                                'longname': 'marginHeight',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'mh'},
                               'marginWidth': {'args': '<type __builtin__.int>',
                                               'longname': 'marginWidth',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'mw'},
                               'noBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'noBackground',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nbg'},
                               'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'numberOfPopupMenus',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'npm'},
                               'overlayLabelBackColor': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'overlayLabelBackColor',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 4,
                                                         'shortname': 'olb'},
                               'overlayLabelColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'overlayLabelColor',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'olc'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                  'longname': 'popupMenuArray',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'pma'},
                               'preventOverride': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preventOverride',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'po'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'rotation': {'args': '<type __builtin__.float>',
                                            'longname': 'rotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'rot'},
                               'statusBarMessage': {'args': '<type __builtin__.str>',
                                                    'longname': 'statusBarMessage',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sbm'},
                               'style': {'args': '<type __builtin__.str>',
                                         'longname': 'style',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'st'},
                               'useAlpha': {'args': '<type __builtin__.bool>',
                                            'longname': 'useAlpha',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ua'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'},
                               'version': {'args': '<type __builtin__.str>',
                                           'longname': 'version',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ver'},
                               'visible': {'args': '<type __builtin__.bool>',
                                           'longname': 'visible',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vis'},
                               'visibleChangeCommand': {'args': 'script',
                                                        'longname': 'visibleChangeCommand',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'vcc'},
                               'width': {'args': '<type __builtin__.int>',
                                         'longname': 'width',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'w'}},
                     'shortFlags': {'al': 'align',
                                    'ann': 'annotation',
                                    'bgc': 'backgroundColor',
                                    'c': 'command',
                                    'dgc': 'dragCallback',
                                    'di': 'disabledImage',
                                    'dpc': 'dropCallback',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'e': 'edit',
                                    'ebg': 'enableBackground',
                                    'ekf': 'enableKeyboardFocus',
                                    'en': 'enable',
                                    'ex': 'exists',
                                    'fn': 'font',
                                    'fpn': 'fullPathName',
                                    'fx': 'flipX',
                                    'fy': 'flipY',
                                    'h': 'height',
                                    'hlc': 'highlightColor',
                                    'i': 'image',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'io': 'isObscured',
                                    'iol': 'imageOverlayLabel',
                                    'l': 'label',
                                    'lo': 'labelOffset',
                                    'lt': 'ltVersion',
                                    'm': 'manage',
                                    'mh': 'marginHeight',
                                    'mw': 'marginWidth',
                                    'nbg': 'noBackground',
                                    'npm': 'numberOfPopupMenus',
                                    'olb': 'overlayLabelBackColor',
                                    'olc': 'overlayLabelColor',
                                    'p': 'parent',
                                    'pma': 'popupMenuArray',
                                    'po': 'preventOverride',
                                    'q': 'query',
                                    'rot': 'rotation',
                                    'sbm': 'statusBarMessage',
                                    'st': 'style',
                                    'ua': 'useAlpha',
                                    'ut': 'useTemplate',
                                    'vcc': 'visibleChangeCommand',
                                    'ver': 'version',
                                    'vis': 'visible',
                                    'w': 'width'},
                     'type': 'rendering'},
  'nodeOutliner': {'callbackFlags': ['selectCommand',
                                     'sc',
                                     'visibleChangeCommand',
                                     'vcc',
                                     'dc',
                                     'dc',
                                     'dragCallback',
                                     'dgc',
                                     'addCommand',
                                     'ac',
                                     'menuCommand',
                                     'mc',
                                     'dropCallback',
                                     'dpc'],
                   'flags': {'addCommand': {'args': 'script',
                                            'longname': 'addCommand',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ac'},
                             'addObject': {'args': 'PyNode',
                                           'longname': 'addObject',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'a'},
                             'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'attrAlphaOrder': {'args': '<type __builtin__.str>',
                                                'longname': 'attrAlphaOrder',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'aao'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'connectivity': {'args': 'PyNode',
                                              'longname': 'connectivity',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'c'},
                             'currentSelection': {'args': '<type __builtin__.bool>',
                                                  'longname': 'currentSelection',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'cs'},
                             'dc': {'args': '<type __builtin__.callable>',
                                    'longname': 'dc',
                                    'numArgs': 1,
                                    'shortname': 'dc'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'lastClickedNode': {'args': '<type __builtin__.bool>',
                                                 'longname': 'lastClickedNode',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'lcn'},
                             'lastMenuChoice': {'args': '<type __builtin__.str>',
                                                'longname': 'lastMenuChoice',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'lmc'},
                             'longNames': {'args': '<type __builtin__.bool>',
                                           'longname': 'longNames',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ln'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'menuCommand': {'args': 'script',
                                             'longname': 'menuCommand',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'mc'},
                             'menuMultiOption': {'args': '<type __builtin__.bool>',
                                                 'longname': 'menuMultiOption',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'mmo'},
                             'multiSelect': {'args': '<type __builtin__.bool>',
                                             'longname': 'multiSelect',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ms'},
                             'niceNames': {'args': '<type __builtin__.bool>',
                                           'longname': 'niceNames',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nn'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'noConnectivity': {'args': '<type __builtin__.bool>',
                                                'longname': 'noConnectivity',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'nc'},
                             'nodesDisplayed': {'args': '<type __builtin__.bool>',
                                                'longname': 'nodesDisplayed',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'nd'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'pressHighlightsUnconnected': {'args': '<type __builtin__.bool>',
                                                            'longname': 'pressHighlightsUnconnected',
                                                            'modes': ['query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'phu'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'redraw': {'args': '<type __builtin__.bool>',
                                        'longname': 'redraw',
                                        'modes': ['edit'],
                                        'numArgs': 0,
                                        'shortname': 'r'},
                             'redrawRow': {'args': '<type __builtin__.bool>',
                                           'longname': 'redrawRow',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'rr'},
                             'remove': {'args': '<type __builtin__.str>',
                                        'longname': 'remove',
                                        'modes': ['edit', 'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'rm'},
                             'removeAll': {'args': '<type __builtin__.bool>',
                                           'longname': 'removeAll',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'rma'},
                             'replace': {'args': 'PyNode',
                                         'longname': 'replace',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'rpl'},
                             'selectCommand': {'args': 'script',
                                               'longname': 'selectCommand',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sc'},
                             'showConnectedOnly': {'args': '<type __builtin__.bool>',
                                                   'longname': 'showConnectedOnly',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sco'},
                             'showHidden': {'args': '<type __builtin__.bool>',
                                            'longname': 'showHidden',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sh'},
                             'showInputs': {'args': '<type __builtin__.bool>',
                                            'longname': 'showInputs',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'si'},
                             'showNonConnectable': {'args': '<type __builtin__.bool>',
                                                    'longname': 'showNonConnectable',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'snc'},
                             'showNonKeyable': {'args': '<type __builtin__.bool>',
                                                'longname': 'showNonKeyable',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'snk'},
                             'showOutputs': {'args': '<type __builtin__.bool>',
                                             'longname': 'showOutputs',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'so'},
                             'showPublished': {'args': '<type __builtin__.bool>',
                                               'longname': 'showPublished',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sp'},
                             'showReadOnly': {'args': '<type __builtin__.bool>',
                                              'longname': 'showReadOnly',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sro'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'shortFlags': {'a': 'addObject',
                                  'aao': 'attrAlphaOrder',
                                  'ac': 'addCommand',
                                  'ann': 'annotation',
                                  'bgc': 'backgroundColor',
                                  'c': 'connectivity',
                                  'cs': 'currentSelection',
                                  'dc': 'dc',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ebg': 'enableBackground',
                                  'ekf': 'enableKeyboardFocus',
                                  'en': 'enable',
                                  'ex': 'exists',
                                  'fpn': 'fullPathName',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'io': 'isObscured',
                                  'lcn': 'lastClickedNode',
                                  'lmc': 'lastMenuChoice',
                                  'ln': 'longNames',
                                  'm': 'manage',
                                  'mc': 'menuCommand',
                                  'mmo': 'menuMultiOption',
                                  'ms': 'multiSelect',
                                  'nbg': 'noBackground',
                                  'nc': 'noConnectivity',
                                  'nd': 'nodesDisplayed',
                                  'nn': 'niceNames',
                                  'npm': 'numberOfPopupMenus',
                                  'p': 'parent',
                                  'phu': 'pressHighlightsUnconnected',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'q': 'query',
                                  'r': 'redraw',
                                  'rm': 'remove',
                                  'rma': 'removeAll',
                                  'rpl': 'replace',
                                  'rr': 'redrawRow',
                                  'sbm': 'statusBarMessage',
                                  'sc': 'selectCommand',
                                  'sco': 'showConnectedOnly',
                                  'sh': 'showHidden',
                                  'si': 'showInputs',
                                  'snc': 'showNonConnectable',
                                  'snk': 'showNonKeyable',
                                  'so': 'showOutputs',
                                  'sp': 'showPublished',
                                  'sro': 'showReadOnly',
                                  'ut': 'useTemplate',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'w': 'width'},
                   'type': 'windows'},
  'nodePreset': {'flags': {'attributes': {'args': '<type __builtin__.str>',
                                          'longname': 'attributes',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'atr'},
                           'custom': {'args': '<type __builtin__.str>',
                                      'longname': 'custom',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'ctm'},
                           'delete': {'args': ['PyNode',
                                               '<type __builtin__.str>'],
                                      'longname': 'delete',
                                      'modes': ['create'],
                                      'numArgs': 2,
                                      'shortname': 'delete'},
                           'exists': {'args': ['PyNode',
                                               '<type __builtin__.str>'],
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 2,
                                      'shortname': 'ex'},
                           'isValidName': {'args': '<type __builtin__.str>',
                                           'longname': 'isValidName',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ivn'},
                           'list': {'args': 'PyNode',
                                    'longname': 'list',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'ls'},
                           'load': {'args': ['PyNode',
                                             '<type __builtin__.str>'],
                                    'longname': 'load',
                                    'modes': ['create'],
                                    'numArgs': 2,
                                    'shortname': 'ld'},
                           'save': {'args': ['PyNode',
                                             '<type __builtin__.str>'],
                                    'longname': 'save',
                                    'modes': ['create'],
                                    'numArgs': 2,
                                    'shortname': 'sv'}},
                 'removedFlags': {'del': 'delete'},
                 'shortFlags': {'atr': 'attributes',
                                'ctm': 'custom',
                                'delete': 'delete',
                                'ex': 'exists',
                                'ivn': 'isValidName',
                                'ld': 'load',
                                'ls': 'list',
                                'sv': 'save'},
                 'type': 'rendering'},
  'nodeTreeLister': {'callbackFlags': ['refreshCommand',
                                       'rc',
                                       'visibleChangeCommand',
                                       'vcc',
                                       'dragCallback',
                                       'dgc',
                                       'dropCallback',
                                       'dpc',
                                       'favoritesCallback',
                                       'fcb'],
                     'flags': {'addFavorite': {'args': '<type __builtin__.str>',
                                               'longname': 'addFavorite',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'af'},
                               'addItem': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    'script'],
                                           'longname': 'addItem',
                                           'modes': ['create',
                                                     'edit',
                                                     'multiuse'],
                                           'numArgs': 3,
                                           'shortname': 'add'},
                               'addVnnItem': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'addVnnItem',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 4,
                                              'shortname': 'avi'},
                               'annotation': {'args': '<type __builtin__.str>',
                                              'longname': 'annotation',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ann'},
                               'backgroundColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'backgroundColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'bgc'},
                               'clearContents': {'args': '<type __builtin__.bool>',
                                                 'longname': 'clearContents',
                                                 'modes': ['edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'clr'},
                               'collapsePath': {'args': '<type __builtin__.str>',
                                                'longname': 'collapsePath',
                                                'modes': ['edit',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'cp'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'dragCallback': {'args': 'script',
                                                'longname': 'dragCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dgc'},
                               'dropCallback': {'args': 'script',
                                                'longname': 'dropCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dpc'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'enable': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en'},
                               'enableBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableBackground',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ebg'},
                               'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                       'longname': 'enableKeyboardFocus',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ekf'},
                               'executeItem': {'args': '<type __builtin__.str>',
                                               'longname': 'executeItem',
                                               'modes': ['edit'],
                                               'numArgs': 1,
                                               'shortname': 'ei'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'expandPath': {'args': '<type __builtin__.str>',
                                              'longname': 'expandPath',
                                              'modes': ['edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'ep'},
                               'expandToDepth': {'args': '<type __builtin__.int>',
                                                 'longname': 'expandToDepth',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'etd'},
                               'favoritesCallback': {'args': 'script',
                                                     'longname': 'favoritesCallback',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'fcb'},
                               'favoritesList': {'args': '<type __builtin__.bool>',
                                                 'longname': 'favoritesList',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'fl'},
                               'fullPathName': {'args': '<type __builtin__.bool>',
                                                'longname': 'fullPathName',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'fpn'},
                               'height': {'args': '<type __builtin__.int>',
                                          'longname': 'height',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'h'},
                               'highlightColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'highlightColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'hlc'},
                               'isObscured': {'args': '<type __builtin__.bool>',
                                              'longname': 'isObscured',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'io'},
                               'itemScript': {'args': '<type __builtin__.str>',
                                              'longname': 'itemScript',
                                              'modes': ['query'],
                                              'numArgs': 1,
                                              'shortname': 'isc'},
                               'manage': {'args': '<type __builtin__.bool>',
                                          'longname': 'manage',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'm'},
                               'noBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'noBackground',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nbg'},
                               'nodeLibrary': {'args': '<type __builtin__.str>',
                                               'longname': 'nodeLibrary',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nl'},
                               'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'numberOfPopupMenus',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'npm'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                  'longname': 'popupMenuArray',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'pma'},
                               'preventOverride': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preventOverride',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'po'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'refreshCommand': {'args': 'script',
                                                  'longname': 'refreshCommand',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'rc'},
                               'removeFavorite': {'args': '<type __builtin__.str>',
                                                  'longname': 'removeFavorite',
                                                  'modes': ['edit',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'rf'},
                               'removeItem': {'args': '<type __builtin__.str>',
                                              'longname': 'removeItem',
                                              'modes': ['edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'rem'},
                               'resultsPathUnderCursor': {'args': '<type __builtin__.bool>',
                                                          'longname': 'resultsPathUnderCursor',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'ruc'},
                               'selectPath': {'args': '<type __builtin__.str>',
                                              'longname': 'selectPath',
                                              'modes': ['edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'sp'},
                               'statusBarMessage': {'args': '<type __builtin__.str>',
                                                    'longname': 'statusBarMessage',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sbm'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'},
                               'visible': {'args': '<type __builtin__.bool>',
                                           'longname': 'visible',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vis'},
                               'visibleChangeCommand': {'args': 'script',
                                                        'longname': 'visibleChangeCommand',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'vcc'},
                               'vnnString': {'args': '<type __builtin__.bool>',
                                             'longname': 'vnnString',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'vnn'},
                               'width': {'args': '<type __builtin__.int>',
                                         'longname': 'width',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'w'}},
                     'shortFlags': {'add': 'addItem',
                                    'af': 'addFavorite',
                                    'ann': 'annotation',
                                    'avi': 'addVnnItem',
                                    'bgc': 'backgroundColor',
                                    'clr': 'clearContents',
                                    'cp': 'collapsePath',
                                    'dgc': 'dragCallback',
                                    'dpc': 'dropCallback',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'e': 'edit',
                                    'ebg': 'enableBackground',
                                    'ei': 'executeItem',
                                    'ekf': 'enableKeyboardFocus',
                                    'en': 'enable',
                                    'ep': 'expandPath',
                                    'etd': 'expandToDepth',
                                    'ex': 'exists',
                                    'fcb': 'favoritesCallback',
                                    'fl': 'favoritesList',
                                    'fpn': 'fullPathName',
                                    'h': 'height',
                                    'hlc': 'highlightColor',
                                    'io': 'isObscured',
                                    'isc': 'itemScript',
                                    'm': 'manage',
                                    'nbg': 'noBackground',
                                    'nl': 'nodeLibrary',
                                    'npm': 'numberOfPopupMenus',
                                    'p': 'parent',
                                    'pma': 'popupMenuArray',
                                    'po': 'preventOverride',
                                    'q': 'query',
                                    'rc': 'refreshCommand',
                                    'rem': 'removeItem',
                                    'rf': 'removeFavorite',
                                    'ruc': 'resultsPathUnderCursor',
                                    'sbm': 'statusBarMessage',
                                    'sp': 'selectPath',
                                    'ut': 'useTemplate',
                                    'vcc': 'visibleChangeCommand',
                                    'vis': 'visible',
                                    'vnn': 'vnnString',
                                    'w': 'width'},
                     'type': 'windows'},
  'nodeType': {'flags': {'apiType': {'args': '<type __builtin__.bool>',
                                     'longname': 'apiType',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'api'},
                         'derived': {'args': '<type __builtin__.bool>',
                                     'longname': 'derived',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'd'},
                         'inherited': {'args': '<type __builtin__.bool>',
                                       'longname': 'inherited',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'i'},
                         'isTypeName': {'args': '<type __builtin__.bool>',
                                        'longname': 'isTypeName',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'itn'}},
               'shortFlags': {'api': 'apiType',
                              'd': 'derived',
                              'i': 'inherited',
                              'itn': 'isTypeName'},
               'type': 'general'},
  'nonLinear': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                    'longname': 'after',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'af'},
                          'afterReference': {'args': '<type __builtin__.bool>',
                                             'longname': 'afterReference',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'ar'},
                          'autoParent': {'args': '<type __builtin__.bool>',
                                         'longname': 'autoParent',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ap'},
                          'before': {'args': '<type __builtin__.bool>',
                                     'longname': 'before',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'bf'},
                          'commonParent': {'args': '<type __builtin__.bool>',
                                           'longname': 'commonParent',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'cp'},
                          'defaultScale': {'args': '<type __builtin__.bool>',
                                           'longname': 'defaultScale',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ds'},
                          'deformerTools': {'args': '<type __builtin__.bool>',
                                            'longname': 'deformerTools',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'dt'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'exclusive': {'args': '<type __builtin__.str>',
                                        'longname': 'exclusive',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'ex'},
                          'frontOfChain': {'args': '<type __builtin__.bool>',
                                           'longname': 'frontOfChain',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'foc'},
                          'geometry': {'args': '<type __builtin__.str>',
                                       'longname': 'geometry',
                                       'modes': ['query',
                                                 'edit',
                                                 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'g'},
                          'geometryIndices': {'args': '<type __builtin__.bool>',
                                              'longname': 'geometryIndices',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'gi'},
                          'ignoreSelected': {'args': '<type __builtin__.bool>',
                                             'longname': 'ignoreSelected',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ignoreSelected'},
                          'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                      'longname': 'includeHiddenSelections',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'ihs'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'parallel': {'args': '<type __builtin__.bool>',
                                       'longname': 'parallel',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'par'},
                          'prune': {'args': '<type __builtin__.bool>',
                                    'longname': 'prune',
                                    'modes': ['edit'],
                                    'numArgs': 0,
                                    'shortname': 'pr'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'remove': {'args': '<type __builtin__.bool>',
                                     'longname': 'remove',
                                     'modes': ['edit', 'multiuse'],
                                     'numArgs': 0,
                                     'shortname': 'rm'},
                          'split': {'args': '<type __builtin__.bool>',
                                    'longname': 'split',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'sp'},
                          'type': {'args': '<type __builtin__.str>',
                                   'longname': 'type',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'typ'}},
                'removedFlags': {'is': 'ignoreSelected'},
                'shortFlags': {'af': 'after',
                               'ap': 'autoParent',
                               'ar': 'afterReference',
                               'bf': 'before',
                               'cp': 'commonParent',
                               'ds': 'defaultScale',
                               'dt': 'deformerTools',
                               'e': 'edit',
                               'ex': 'exclusive',
                               'foc': 'frontOfChain',
                               'g': 'geometry',
                               'gi': 'geometryIndices',
                               'ignoreSelected': 'ignoreSelected',
                               'ihs': 'includeHiddenSelections',
                               'n': 'name',
                               'par': 'parallel',
                               'pr': 'prune',
                               'q': 'query',
                               'rm': 'remove',
                               'sp': 'split',
                               'typ': 'type'},
                'type': 'animation'},
  'nop': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'normalConstraint': {'flags': {'aimVector': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'aimVector',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'aim'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'layer': {'args': '<type __builtin__.str>',
                                           'longname': 'layer',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'l'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'remove': {'args': '<type __builtin__.bool>',
                                            'longname': 'remove',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'rm'},
                                 'targetList': {'args': '<type __builtin__.bool>',
                                                'longname': 'targetList',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'tl'},
                                 'upVector': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'upVector',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'u'},
                                 'weight': {'args': '<type __builtin__.float>',
                                            'longname': 'weight',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'w'},
                                 'weightAliasList': {'args': '<type __builtin__.bool>',
                                                     'longname': 'weightAliasList',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'wal'},
                                 'worldUpObject': {'args': 'PyNode',
                                                   'longname': 'worldUpObject',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'wuo'},
                                 'worldUpType': {'args': '<type __builtin__.str>',
                                                 'longname': 'worldUpType',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'wut'},
                                 'worldUpVector': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'worldUpVector',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'wu'}},
                       'resultNeedsUnpacking': True,
                       'shortFlags': {'aim': 'aimVector',
                                      'e': 'edit',
                                      'n': 'name',
                                      'q': 'query',
                                      'rm': 'remove',
                                      'tl': 'targetList',
                                      'u': 'upVector',
                                      'w': 'weight',
                                      'wal': 'weightAliasList',
                                      'wu': 'worldUpVector',
                                      'wuo': 'worldUpObject',
                                      'wut': 'worldUpType'},
                       'type': 'animation'},
  'nucleusDisplayDynamicConstraintNodes': {'type': 'runtime'},
  'nucleusDisplayMaterialNodes': {'type': 'runtime'},
  'nucleusDisplayNComponentNodes': {'type': 'runtime'},
  'nucleusDisplayOtherNodes': {'type': 'runtime'},
  'nucleusDisplayTextureNodes': {'type': 'runtime'},
  'nucleusDisplayTransformNodes': {'type': 'runtime'},
  'nucleusGetEffectsAsset': {'type': 'runtime'},
  'nucleusGetnClothExample': {'type': 'runtime'},
  'nucleusGetnParticleExample': {'type': 'runtime'},
  'nurbsBoolean': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'nsrfsInFirstShell': {'args': '<type __builtin__.int>',
                                                   'longname': 'nsrfsInFirstShell',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'nsf'},
                             'object': {'args': '<type __builtin__.bool>',
                                        'longname': 'object',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                             'operation': {'args': '<type __builtin__.int>',
                                           'longname': 'operation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'op'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'smartConnection': {'args': '<type __builtin__.bool>',
                                                 'longname': 'smartConnection',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'sc'},
                             'tolerance': {'args': '<type __builtin__.float>',
                                           'longname': 'tolerance',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'tlb'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'nsf': 'nsrfsInFirstShell',
                                  'o': 'object',
                                  'op': 'operation',
                                  'q': 'query',
                                  'sc': 'smartConnection',
                                  'tlb': 'tolerance'},
                   'type': 'modeling'},
  'nurbsCopyUVSet': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'}},
                     'shortFlags': {'e': 'edit', 'q': 'query'},
                     'type': 'modeling'},
  'nurbsCube': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'axis',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 3,
                                   'shortname': 'ax'},
                          'caching': {'args': '<type __builtin__.bool>',
                                      'longname': 'caching',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsCasting': True,
                                      'shortname': 'cch'},
                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'constructionHistory',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                          'degree': {'args': '<type __builtin__.int>',
                                     'longname': 'degree',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsCasting': True,
                                     'shortname': 'd'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'frozen': {'args': '<type __builtin__.bool>',
                                     'longname': 'frozen',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'fzn'},
                          'heightRatio': {'args': '<type __builtin__.float>',
                                          'longname': 'heightRatio',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'hr'},
                          'lengthRatio': {'args': '<type __builtin__.float>',
                                          'longname': 'lengthRatio',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'lr'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nodeState': {'args': '<type __builtin__.int>',
                                        'longname': 'nodeState',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'nds'},
                          'object': {'args': '<type __builtin__.bool>',
                                     'longname': 'object',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'o'},
                          'patchesU': {'args': '<type __builtin__.int>',
                                       'longname': 'patchesU',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'u'},
                          'patchesV': {'args': '<type __builtin__.int>',
                                       'longname': 'patchesV',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'v'},
                          'pivot': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'pivot',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 3,
                                    'shortname': 'p'},
                          'polygon': {'args': '<type __builtin__.int>',
                                      'longname': 'polygon',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'po'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'width': {'args': '<type __builtin__.float>',
                                    'longname': 'width',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'w'}},
                'shortFlags': {'ax': 'axis',
                               'cch': 'caching',
                               'ch': 'constructionHistory',
                               'd': 'degree',
                               'e': 'edit',
                               'fzn': 'frozen',
                               'hr': 'heightRatio',
                               'lr': 'lengthRatio',
                               'n': 'name',
                               'nds': 'nodeState',
                               'o': 'object',
                               'p': 'pivot',
                               'po': 'polygon',
                               'q': 'query',
                               'u': 'patchesU',
                               'v': 'patchesV',
                               'w': 'width'},
                'type': 'modeling'},
  'nurbsCurveRebuildPref': {'flags': {'degree': {'args': '<type __builtin__.int>',
                                                 'longname': 'degree',
                                                 'numArgs': 1,
                                                 'shortname': 'd'},
                                      'endKnots': {'args': '<type __builtin__.int>',
                                                   'longname': 'endKnots',
                                                   'numArgs': 1,
                                                   'shortname': 'end'},
                                      'fitRebuild': {'args': '<type __builtin__.int>',
                                                     'longname': 'fitRebuild',
                                                     'numArgs': 1,
                                                     'shortname': 'fr'},
                                      'keepControlPoints': {'args': '<type __builtin__.bool>',
                                                            'longname': 'keepControlPoints',
                                                            'numArgs': 1,
                                                            'shortname': 'kcp'},
                                      'keepEndPoints': {'args': '<type __builtin__.bool>',
                                                        'longname': 'keepEndPoints',
                                                        'numArgs': 1,
                                                        'shortname': 'kep'},
                                      'keepRange': {'args': '<type __builtin__.int>',
                                                    'longname': 'keepRange',
                                                    'numArgs': 1,
                                                    'shortname': 'kr'},
                                      'keepTangents': {'args': '<type __builtin__.bool>',
                                                       'longname': 'keepTangents',
                                                       'numArgs': 1,
                                                       'shortname': 'kt'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'rebuildType': {'args': '<type __builtin__.int>',
                                                      'longname': 'rebuildType',
                                                      'numArgs': 1,
                                                      'shortname': 'rt'},
                                      'smartSurfaceCurve': {'args': '<type __builtin__.bool>',
                                                            'longname': 'smartSurfaceCurve',
                                                            'numArgs': 1,
                                                            'shortname': 'scr'},
                                      'spans': {'args': '<type __builtin__.int>',
                                                'longname': 'spans',
                                                'numArgs': 1,
                                                'shortname': 's'},
                                      'tolerance': {'args': '<type __builtin__.float>',
                                                    'longname': 'tolerance',
                                                    'numArgs': 1,
                                                    'shortname': 'tol'}},
                            'shortFlags': {'d': 'degree',
                                           'end': 'endKnots',
                                           'fr': 'fitRebuild',
                                           'kcp': 'keepControlPoints',
                                           'kep': 'keepEndPoints',
                                           'kr': 'keepRange',
                                           'kt': 'keepTangents',
                                           'q': 'query',
                                           'rt': 'rebuildType',
                                           's': 'spans',
                                           'scr': 'smartSurfaceCurve',
                                           'tol': 'tolerance'},
                            'type': 'other'},
  'nurbsCurveToBezier': {'flags': {}, 'shortFlags': {}, 'type': 'modeling'},
  'nurbsEditUV': {'flags': {'angle': {'args': '<type __builtin__.float>',
                                      'longname': 'angle',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'a'},
                            'pivotU': {'args': '<type __builtin__.float>',
                                       'longname': 'pivotU',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'pu'},
                            'pivotV': {'args': '<type __builtin__.float>',
                                       'longname': 'pivotV',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'pv'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'relative': {'args': '<type __builtin__.bool>',
                                         'longname': 'relative',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'r'},
                            'rotateRatio': {'args': '<type __builtin__.float>',
                                            'longname': 'rotateRatio',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'rr'},
                            'rotation': {'args': '<type __builtin__.bool>',
                                         'longname': 'rotation',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'rot'},
                            'scale': {'args': '<type __builtin__.bool>',
                                      'longname': 'scale',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                            'scaleU': {'args': '<type __builtin__.float>',
                                       'longname': 'scaleU',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'su'},
                            'scaleV': {'args': '<type __builtin__.float>',
                                       'longname': 'scaleV',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'sv'},
                            'uValue': {'args': '<type __builtin__.float>',
                                       'longname': 'uValue',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'u'},
                            'vValue': {'args': '<type __builtin__.float>',
                                       'longname': 'vValue',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'v'}},
                  'shortFlags': {'a': 'angle',
                                 'pu': 'pivotU',
                                 'pv': 'pivotV',
                                 'q': 'query',
                                 'r': 'relative',
                                 'rot': 'rotation',
                                 'rr': 'rotateRatio',
                                 's': 'scale',
                                 'su': 'scaleU',
                                 'sv': 'scaleV',
                                 'u': 'uValue',
                                 'v': 'vValue'},
                  'type': 'modeling'},
  'nurbsPlane': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'axis',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 3,
                                    'shortname': 'ax'},
                           'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'degree': {'args': '<type __builtin__.int>',
                                      'longname': 'degree',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsCasting': True,
                                      'shortname': 'd'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'lengthRatio': {'args': '<type __builtin__.float>',
                                           'longname': 'lengthRatio',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'lr'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'resultNeedsCasting': True,
                                         'shortname': 'nds'},
                           'object': {'args': '<type __builtin__.bool>',
                                      'longname': 'object',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'patchesU': {'args': '<type __builtin__.int>',
                                        'longname': 'patchesU',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'u'},
                           'patchesV': {'args': '<type __builtin__.int>',
                                        'longname': 'patchesV',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'v'},
                           'pivot': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'pivot',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 3,
                                     'shortname': 'p'},
                           'polygon': {'args': '<type __builtin__.int>',
                                       'longname': 'polygon',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'po'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'width': {'args': '<type __builtin__.float>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'}},
                 'shortFlags': {'ax': 'axis',
                                'cch': 'caching',
                                'ch': 'constructionHistory',
                                'd': 'degree',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'lr': 'lengthRatio',
                                'n': 'name',
                                'nds': 'nodeState',
                                'o': 'object',
                                'p': 'pivot',
                                'po': 'polygon',
                                'q': 'query',
                                'u': 'patchesU',
                                'v': 'patchesV',
                                'w': 'width'},
                 'type': 'modeling'},
  'nurbsSelect': {'flags': {'borderSelection': {'args': '<type __builtin__.bool>',
                                                'longname': 'borderSelection',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'bs'},
                            'bottomBorder': {'args': '<type __builtin__.bool>',
                                             'longname': 'bottomBorder',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'bb'},
                            'growSelection': {'args': '<type __builtin__.int>',
                                              'longname': 'growSelection',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'gs'},
                            'leftBorder': {'args': '<type __builtin__.bool>',
                                           'longname': 'leftBorder',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'lb'},
                            'rightBorder': {'args': '<type __builtin__.bool>',
                                            'longname': 'rightBorder',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'rb'},
                            'shrinkSelection': {'args': '<type __builtin__.int>',
                                                'longname': 'shrinkSelection',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ss'},
                            'topBorder': {'args': '<type __builtin__.bool>',
                                          'longname': 'topBorder',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'tb'}},
                  'shortFlags': {'bb': 'bottomBorder',
                                 'bs': 'borderSelection',
                                 'gs': 'growSelection',
                                 'lb': 'leftBorder',
                                 'rb': 'rightBorder',
                                 'ss': 'shrinkSelection',
                                 'tb': 'topBorder'},
                  'type': 'modeling'},
  'nurbsSquare': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'cch'},
                            'center': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'center',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 3,
                                       'shortname': 'c'},
                            'centerX': {'args': '<type __builtin__.float>',
                                        'longname': 'centerX',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cx'},
                            'centerY': {'args': '<type __builtin__.float>',
                                        'longname': 'centerY',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cy'},
                            'centerZ': {'args': '<type __builtin__.float>',
                                        'longname': 'centerZ',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cz'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'degree': {'args': '<type __builtin__.int>',
                                       'longname': 'degree',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'd'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'resultNeedsCasting': True,
                                          'shortname': 'nds'},
                            'normal': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'normal',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 3,
                                       'shortname': 'nr'},
                            'normalX': {'args': '<type __builtin__.float>',
                                        'longname': 'normalX',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'nrx'},
                            'normalY': {'args': '<type __builtin__.float>',
                                        'longname': 'normalY',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'nry'},
                            'normalZ': {'args': '<type __builtin__.float>',
                                        'longname': 'normalZ',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'nrz'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'sideLength1': {'args': '<type __builtin__.float>',
                                            'longname': 'sideLength1',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sl1'},
                            'sideLength2': {'args': '<type __builtin__.float>',
                                            'longname': 'sideLength2',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sl2'},
                            'spansPerSide': {'args': '<type __builtin__.int>',
                                             'longname': 'spansPerSide',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'resultNeedsCasting': True,
                                             'shortname': 'sps'}},
                  'shortFlags': {'c': 'center',
                                 'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'cx': 'centerX',
                                 'cy': 'centerY',
                                 'cz': 'centerZ',
                                 'd': 'degree',
                                 'e': 'edit',
                                 'fzn': 'frozen',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'nr': 'normal',
                                 'nrx': 'normalX',
                                 'nry': 'normalY',
                                 'nrz': 'normalZ',
                                 'o': 'object',
                                 'q': 'query',
                                 'sl1': 'sideLength1',
                                 'sl2': 'sideLength2',
                                 'sps': 'spansPerSide'},
                  'type': 'modeling'},
  'nurbsToPoly': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'chordHeight': {'args': '<type __builtin__.float>',
                                            'longname': 'chordHeight',
                                            'numArgs': 1,
                                            'shortname': 'cht'},
                            'chordHeightRatio': {'args': '<type __builtin__.float>',
                                                 'longname': 'chordHeightRatio',
                                                 'numArgs': 1,
                                                 'shortname': 'chr'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'curvatureTolerance': {'args': '<type __builtin__.int>',
                                                   'longname': 'curvatureTolerance',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cvt'},
                            'delta': {'args': '<type __builtin__.float>',
                                      'longname': 'delta',
                                      'numArgs': 1,
                                      'shortname': 'd'},
                            'edgeSwap': {'args': '<type __builtin__.bool>',
                                         'longname': 'edgeSwap',
                                         'numArgs': 1,
                                         'shortname': 'es'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'explicitTessellationAttributes': {'args': '<type __builtin__.bool>',
                                                               'longname': 'explicitTessellationAttributes',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'eta'},
                            'format': {'args': '<type __builtin__.int>',
                                       'longname': 'format',
                                       'numArgs': 1,
                                       'shortname': 'f'},
                            'fractionalTolerance': {'args': '<type __builtin__.float>',
                                                    'longname': 'fractionalTolerance',
                                                    'numArgs': 1,
                                                    'shortname': 'ft'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'matchNormalDir': {'args': '<type __builtin__.bool>',
                                               'longname': 'matchNormalDir',
                                               'numArgs': 1,
                                               'shortname': 'mnd'},
                            'matchRenderTessellation': {'args': '<type __builtin__.bool>',
                                                        'longname': 'matchRenderTessellation',
                                                        'numArgs': 1,
                                                        'shortname': 'mrt'},
                            'minEdgeLength': {'args': '<type __builtin__.float>',
                                              'longname': 'minEdgeLength',
                                              'numArgs': 1,
                                              'shortname': 'mel'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'normalizeTrimmedUVRange': {'args': '<type __builtin__.bool>',
                                                        'longname': 'normalizeTrimmedUVRange',
                                                        'numArgs': 1,
                                                        'shortname': 'ntr'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'polygonCount': {'args': '<type __builtin__.int>',
                                             'longname': 'polygonCount',
                                             'numArgs': 1,
                                             'shortname': 'pc'},
                            'polygonType': {'args': '<type __builtin__.int>',
                                            'longname': 'polygonType',
                                            'numArgs': 1,
                                            'shortname': 'pt'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'smoothEdge': {'args': '<type __builtin__.bool>',
                                           'longname': 'smoothEdge',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ues'},
                            'smoothEdgeRatio': {'args': '<type __builtin__.float>',
                                                'longname': 'smoothEdgeRatio',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'esr'},
                            'uDivisionsFactor': {'args': '<type __builtin__.float>',
                                                 'longname': 'uDivisionsFactor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'nuf'},
                            'uNumber': {'args': '<type __builtin__.int>',
                                        'longname': 'uNumber',
                                        'numArgs': 1,
                                        'shortname': 'un'},
                            'uType': {'args': '<type __builtin__.int>',
                                      'longname': 'uType',
                                      'numArgs': 1,
                                      'shortname': 'ut'},
                            'useChordHeight': {'args': '<type __builtin__.bool>',
                                               'longname': 'useChordHeight',
                                               'numArgs': 1,
                                               'shortname': 'uch'},
                            'useChordHeightRatio': {'args': '<type __builtin__.bool>',
                                                    'longname': 'useChordHeightRatio',
                                                    'numArgs': 1,
                                                    'shortname': 'ucr'},
                            'useSurfaceShader': {'args': '<type __builtin__.bool>',
                                                 'longname': 'useSurfaceShader',
                                                 'numArgs': 1,
                                                 'shortname': 'uss'},
                            'vDivisionsFactor': {'args': '<type __builtin__.float>',
                                                 'longname': 'vDivisionsFactor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'nvf'},
                            'vNumber': {'args': '<type __builtin__.int>',
                                        'longname': 'vNumber',
                                        'numArgs': 1,
                                        'shortname': 'vn'},
                            'vType': {'args': '<type __builtin__.int>',
                                      'longname': 'vType',
                                      'numArgs': 1,
                                      'shortname': 'vt'}},
                  'shortFlags': {'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'chr': 'chordHeightRatio',
                                 'cht': 'chordHeight',
                                 'cvt': 'curvatureTolerance',
                                 'd': 'delta',
                                 'e': 'edit',
                                 'es': 'edgeSwap',
                                 'esr': 'smoothEdgeRatio',
                                 'eta': 'explicitTessellationAttributes',
                                 'f': 'format',
                                 'ft': 'fractionalTolerance',
                                 'fzn': 'frozen',
                                 'mel': 'minEdgeLength',
                                 'mnd': 'matchNormalDir',
                                 'mrt': 'matchRenderTessellation',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'ntr': 'normalizeTrimmedUVRange',
                                 'nuf': 'uDivisionsFactor',
                                 'nvf': 'vDivisionsFactor',
                                 'o': 'object',
                                 'pc': 'polygonCount',
                                 'pt': 'polygonType',
                                 'q': 'query',
                                 'uch': 'useChordHeight',
                                 'ucr': 'useChordHeightRatio',
                                 'ues': 'smoothEdge',
                                 'un': 'uNumber',
                                 'uss': 'useSurfaceShader',
                                 'ut': 'uType',
                                 'vn': 'vNumber',
                                 'vt': 'vType'},
                  'type': 'modeling'},
  'nurbsToPolygonsPref': {'flags': {'chordHeight': {'args': '<type __builtin__.float>',
                                                    'longname': 'chordHeight',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'cht'},
                                    'chordHeightRatio': {'args': '<type __builtin__.float>',
                                                         'longname': 'chordHeightRatio',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'chr'},
                                    'delta3D': {'args': '<type __builtin__.float>',
                                                'longname': 'delta3D',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'd'},
                                    'edgeSwap': {'args': '<type __builtin__.bool>',
                                                 'longname': 'edgeSwap',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'es'},
                                    'format': {'args': '<type __builtin__.int>',
                                               'longname': 'format',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'f'},
                                    'fraction': {'args': '<type __builtin__.float>',
                                                 'longname': 'fraction',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ft'},
                                    'matchRenderTessellation': {'args': '<type __builtin__.int>',
                                                                'longname': 'matchRenderTessellation',
                                                                'modes': ['create',
                                                                          'query'],
                                                                'numArgs': 1,
                                                                'shortname': 'mrt'},
                                    'merge': {'args': '<type __builtin__.int>',
                                              'longname': 'merge',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'm'},
                                    'mergeTolerance': {'args': '<type __builtin__.float>',
                                                       'longname': 'mergeTolerance',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'mt'},
                                    'minEdgeLen': {'args': '<type __builtin__.float>',
                                                   'longname': 'minEdgeLen',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'mel'},
                                    'polyCount': {'args': '<type __builtin__.int>',
                                                  'longname': 'polyCount',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'pc'},
                                    'polyType': {'args': '<type __builtin__.int>',
                                                 'longname': 'polyType',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'pt'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'uNumber': {'args': '<type __builtin__.int>',
                                                'longname': 'uNumber',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'un'},
                                    'uType': {'args': '<type __builtin__.int>',
                                              'longname': 'uType',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'ut'},
                                    'useChordHeight': {'args': '<type __builtin__.bool>',
                                                       'longname': 'useChordHeight',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'uch'},
                                    'useChordHeightRatio': {'args': '<type __builtin__.bool>',
                                                            'longname': 'useChordHeightRatio',
                                                            'modes': ['create',
                                                                      'query'],
                                                            'numArgs': 1,
                                                            'shortname': 'ucr'},
                                    'vNumber': {'args': '<type __builtin__.int>',
                                                'longname': 'vNumber',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'vn'},
                                    'vType': {'args': '<type __builtin__.int>',
                                              'longname': 'vType',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'vt'}},
                          'shortFlags': {'chr': 'chordHeightRatio',
                                         'cht': 'chordHeight',
                                         'd': 'delta3D',
                                         'es': 'edgeSwap',
                                         'f': 'format',
                                         'ft': 'fraction',
                                         'm': 'merge',
                                         'mel': 'minEdgeLen',
                                         'mrt': 'matchRenderTessellation',
                                         'mt': 'mergeTolerance',
                                         'pc': 'polyCount',
                                         'pt': 'polyType',
                                         'q': 'query',
                                         'uch': 'useChordHeight',
                                         'ucr': 'useChordHeightRatio',
                                         'un': 'uNumber',
                                         'ut': 'uType',
                                         'vn': 'vNumber',
                                         'vt': 'vType'},
                          'type': 'modeling'},
  'nurbsToSubdiv': {'flags': {'addUnderTransform': {'args': '<type __builtin__.bool>',
                                                    'longname': 'addUnderTransform',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'aut'},
                              'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cch'},
                              'collapsePoles': {'args': '<type __builtin__.bool>',
                                                'longname': 'collapsePoles',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cp'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'matchPeriodic': {'args': '<type __builtin__.bool>',
                                                'longname': 'matchPeriodic',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'mp'},
                              'maxPolyCount': {'args': '<type __builtin__.int>',
                                               'longname': 'maxPolyCount',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'mpc'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nds'},
                              'object': {'args': '<type __builtin__.bool>',
                                         'longname': 'object',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'o'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'reverseNormal': {'args': '<type __builtin__.bool>',
                                                'longname': 'reverseNormal',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'rn'}},
                    'shortFlags': {'aut': 'addUnderTransform',
                                   'cch': 'caching',
                                   'ch': 'constructionHistory',
                                   'cp': 'collapsePoles',
                                   'e': 'edit',
                                   'fzn': 'frozen',
                                   'mp': 'matchPeriodic',
                                   'mpc': 'maxPolyCount',
                                   'n': 'name',
                                   'nds': 'nodeState',
                                   'o': 'object',
                                   'q': 'query',
                                   'rn': 'reverseNormal'},
                    'type': 'modeling'},
  'nurbsToSubdivPref': {'flags': {'bridge': {'args': '<type __builtin__.int>',
                                             'longname': 'bridge',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'br'},
                                  'capType': {'args': '<type __builtin__.int>',
                                              'longname': 'capType',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'ct'},
                                  'collapsePoles': {'args': '<type __builtin__.bool>',
                                                    'longname': 'collapsePoles',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'cp'},
                                  'matchPeriodic': {'args': '<type __builtin__.bool>',
                                                    'longname': 'matchPeriodic',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'mp'},
                                  'maxPolyCount': {'args': '<type __builtin__.int>',
                                                   'longname': 'maxPolyCount',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'mpc'},
                                  'offset': {'args': '<type __builtin__.float>',
                                             'longname': 'offset',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'o'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'reverseNormal': {'args': '<type __builtin__.bool>',
                                                    'longname': 'reverseNormal',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'rn'},
                                  'solidType': {'args': '<type __builtin__.int>',
                                                'longname': 'solidType',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'st'},
                                  'trans00': {'args': '<type __builtin__.float>',
                                              'longname': 'trans00',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 't00'},
                                  'trans01': {'args': '<type __builtin__.float>',
                                              'longname': 'trans01',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 't01'},
                                  'trans02': {'args': '<type __builtin__.float>',
                                              'longname': 'trans02',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 't02'},
                                  'trans10': {'args': '<type __builtin__.float>',
                                              'longname': 'trans10',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 't10'},
                                  'trans11': {'args': '<type __builtin__.float>',
                                              'longname': 'trans11',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 't11'},
                                  'trans12': {'args': '<type __builtin__.float>',
                                              'longname': 'trans12',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 't12'},
                                  'trans20': {'args': '<type __builtin__.float>',
                                              'longname': 'trans20',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 't20'},
                                  'trans21': {'args': '<type __builtin__.float>',
                                              'longname': 'trans21',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 't21'},
                                  'trans22': {'args': '<type __builtin__.float>',
                                              'longname': 'trans22',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 't22'},
                                  'trans30': {'args': '<type __builtin__.float>',
                                              'longname': 'trans30',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 't30'},
                                  'trans31': {'args': '<type __builtin__.float>',
                                              'longname': 'trans31',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 't31'},
                                  'trans32': {'args': '<type __builtin__.float>',
                                              'longname': 'trans32',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 't32'}},
                        'shortFlags': {'br': 'bridge',
                                       'cp': 'collapsePoles',
                                       'ct': 'capType',
                                       'mp': 'matchPeriodic',
                                       'mpc': 'maxPolyCount',
                                       'o': 'offset',
                                       'q': 'query',
                                       'rn': 'reverseNormal',
                                       'st': 'solidType',
                                       't00': 'trans00',
                                       't01': 'trans01',
                                       't02': 'trans02',
                                       't10': 'trans10',
                                       't11': 'trans11',
                                       't12': 'trans12',
                                       't20': 'trans20',
                                       't21': 'trans21',
                                       't22': 'trans22',
                                       't30': 'trans30',
                                       't31': 'trans31',
                                       't32': 'trans32'},
                        'type': 'modeling'},
  'nurbsUVSet': {'flags': {'create': {'args': '<type __builtin__.bool>',
                                      'longname': 'create',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'c'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'useExplicit': {'args': '<type __builtin__.bool>',
                                           'longname': 'useExplicit',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ue'}},
                 'shortFlags': {'c': 'create',
                                'e': 'edit',
                                'q': 'query',
                                'ue': 'useExplicit'},
                 'type': 'modeling'},
  'objExists': {'flags': {}, 'shortFlags': {}, 'type': 'general'},
  'objectCenter': {'flags': {'gl': {'args': '<type __builtin__.bool>',
                                    'longname': 'gl',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'gl'},
                             'local': {'args': '<type __builtin__.bool>',
                                       'longname': 'local',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'l'},
                             'x': {'args': '<type __builtin__.bool>',
                                   'longname': 'x',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'x'},
                             'y': {'args': '<type __builtin__.bool>',
                                   'longname': 'y',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'y'},
                             'z': {'args': '<type __builtin__.bool>',
                                   'longname': 'z',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'z'}},
                   'removedFlags': {'global': 'gl'},
                   'shortFlags': {'gl': 'gl',
                                  'l': 'local',
                                  'x': 'x',
                                  'y': 'y',
                                  'z': 'z'},
                   'type': 'general'},
  'objectType': {'flags': {'convertTag': {'args': '<type __builtin__.str>',
                                          'longname': 'convertTag',
                                          'numArgs': 1,
                                          'shortname': 'ct'},
                           'isAType': {'args': '<type __builtin__.str>',
                                       'longname': 'isAType',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'isa'},
                           'isType': {'args': '<type __builtin__.str>',
                                      'longname': 'isType',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'i'},
                           'tagFromType': {'args': '<type __builtin__.str>',
                                           'longname': 'tagFromType',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'tgt'},
                           'typeFromTag': {'args': '<type __builtin__.int>',
                                           'longname': 'typeFromTag',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'tpt'},
                           'typeTag': {'args': '<type __builtin__.bool>',
                                       'longname': 'typeTag',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'tt'}},
                 'shortFlags': {'ct': 'convertTag',
                                'i': 'isType',
                                'isa': 'isAType',
                                'tgt': 'tagFromType',
                                'tpt': 'typeFromTag',
                                'tt': 'typeTag'},
                 'type': 'general'},
  'objectTypeUI': {'flags': {'isType': {'args': '<type __builtin__.str>',
                                        'longname': 'isType',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'i'},
                             'listAll': {'args': '<type __builtin__.bool>',
                                         'longname': 'listAll',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'la'},
                             'superClasses': {'args': '<type __builtin__.bool>',
                                              'longname': 'superClasses',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'sc'}},
                   'shortFlags': {'i': 'isType',
                                  'la': 'listAll',
                                  'sc': 'superClasses'},
                   'type': 'language'},
  'offsetCurve': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'connectBreaks': {'args': '<type __builtin__.int>',
                                              'longname': 'connectBreaks',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cb'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'cutLoop': {'args': '<type __builtin__.bool>',
                                        'longname': 'cutLoop',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cl'},
                            'cutRadius': {'args': '<type __builtin__.float>',
                                          'longname': 'cutRadius',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cr'},
                            'distance': {'args': '<type __builtin__.float>',
                                         'longname': 'distance',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'd'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'normal': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'normal',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 3,
                                       'shortname': 'nr'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'range': {'args': '<type __builtin__.bool>',
                                      'longname': 'range',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'rn'},
                            'reparameterize': {'args': '<type __builtin__.bool>',
                                               'longname': 'reparameterize',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rp'},
                            'stitch': {'args': '<type __builtin__.bool>',
                                       'longname': 'stitch',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'st'},
                            'subdivisionDensity': {'args': '<type __builtin__.int>',
                                                   'longname': 'subdivisionDensity',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sd'},
                            'tolerance': {'args': '<type __builtin__.float>',
                                          'longname': 'tolerance',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'tol'},
                            'useGivenNormal': {'args': '<type __builtin__.bool>',
                                               'longname': 'useGivenNormal',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ugn'}},
                  'shortFlags': {'cb': 'connectBreaks',
                                 'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'cl': 'cutLoop',
                                 'cr': 'cutRadius',
                                 'd': 'distance',
                                 'e': 'edit',
                                 'fzn': 'frozen',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'nr': 'normal',
                                 'o': 'object',
                                 'q': 'query',
                                 'rn': 'range',
                                 'rp': 'reparameterize',
                                 'sd': 'subdivisionDensity',
                                 'st': 'stitch',
                                 'tol': 'tolerance',
                                 'ugn': 'useGivenNormal'},
                  'type': 'modeling'},
  'offsetCurveOnSurface': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                                 'longname': 'caching',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cch'},
                                     'checkPoints': {'args': '<type __builtin__.int>',
                                                     'longname': 'checkPoints',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'cp'},
                                     'connectBreaks': {'args': '<type __builtin__.int>',
                                                       'longname': 'connectBreaks',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'cb'},
                                     'constructionHistory': {'args': '<type __builtin__.bool>',
                                                             'longname': 'constructionHistory',
                                                             'modes': ['create'],
                                                             'numArgs': 1,
                                                             'shortname': 'ch'},
                                     'cutLoop': {'args': '<type __builtin__.bool>',
                                                 'longname': 'cutLoop',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cl'},
                                     'distance': {'args': '<type __builtin__.float>',
                                                  'longname': 'distance',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'd'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'frozen': {'args': '<type __builtin__.bool>',
                                                'longname': 'frozen',
                                                'numArgs': 1,
                                                'shortname': 'fzn'},
                                     'name': {'args': '<type __builtin__.str>',
                                              'longname': 'name',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'n'},
                                     'nodeState': {'args': '<type __builtin__.int>',
                                                   'longname': 'nodeState',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'nds'},
                                     'object': {'args': '<type __builtin__.bool>',
                                                'longname': 'object',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'o'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'range': {'args': '<type __builtin__.bool>',
                                               'longname': 'range',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'rn'},
                                     'stitch': {'args': '<type __builtin__.bool>',
                                                'longname': 'stitch',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'st'},
                                     'subdivisionDensity': {'args': '<type __builtin__.int>',
                                                            'longname': 'subdivisionDensity',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'sd'},
                                     'tolerance': {'args': '<type __builtin__.float>',
                                                   'longname': 'tolerance',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'tol'}},
                           'shortFlags': {'cb': 'connectBreaks',
                                          'cch': 'caching',
                                          'ch': 'constructionHistory',
                                          'cl': 'cutLoop',
                                          'cp': 'checkPoints',
                                          'd': 'distance',
                                          'e': 'edit',
                                          'fzn': 'frozen',
                                          'n': 'name',
                                          'nds': 'nodeState',
                                          'o': 'object',
                                          'q': 'query',
                                          'rn': 'range',
                                          'sd': 'subdivisionDensity',
                                          'st': 'stitch',
                                          'tol': 'tolerance'},
                           'type': 'modeling'},
  'offsetSurface': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cch'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'distance': {'args': '<type __builtin__.float>',
                                           'longname': 'distance',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'd'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'method': {'args': '<type __builtin__.int>',
                                         'longname': 'method',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nds'},
                              'object': {'args': '<type __builtin__.bool>',
                                         'longname': 'object',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'o'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'}},
                    'shortFlags': {'cch': 'caching',
                                   'ch': 'constructionHistory',
                                   'd': 'distance',
                                   'e': 'edit',
                                   'fzn': 'frozen',
                                   'm': 'method',
                                   'n': 'name',
                                   'nds': 'nodeState',
                                   'o': 'object',
                                   'q': 'query'},
                    'type': 'modeling'},
  'ogs': {'flags': {'deviceInformation': {'args': '<type __builtin__.bool>',
                                          'longname': 'deviceInformation',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'di'},
                    'disposeReleasableTextures': {'args': '<type __builtin__.bool>',
                                                  'longname': 'disposeReleasableTextures',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'drt'},
                    'dumpTexture': {'args': '<type __builtin__.str>',
                                    'longname': 'dumpTexture',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'dt'},
                    'enableHardwareInstancing': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableHardwareInstancing',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'hwi'},
                    'fragmentEditor': {'args': '<type __builtin__.str>',
                                       'longname': 'fragmentEditor',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'fe'},
                    'fragmentXML': {'args': '<type __builtin__.str>',
                                    'longname': 'fragmentXML',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'xml'},
                    'gpuMemoryTotal': {'args': '<type __builtin__.int>',
                                       'longname': 'gpuMemoryTotal',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'gmt'},
                    'gpuMemoryUsed': {'args': '<type __builtin__.bool>',
                                      'longname': 'gpuMemoryUsed',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'gpu'},
                    'isLegacyViewportEnabled': {'args': '<type __builtin__.bool>',
                                                'longname': 'isLegacyViewportEnabled',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'lve'},
                    'isWinRemoteSession': {'args': '<type __builtin__.bool>',
                                           'longname': 'isWinRemoteSession',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'irs'},
                    'pause': {'args': '<type __builtin__.bool>',
                              'longname': 'pause',
                              'modes': ['create', 'query'],
                              'numArgs': 0,
                              'shortname': 'p'},
                    'query': {'args': '<type __builtin__.bool>',
                              'longname': 'query',
                              'numArgs': 0,
                              'shortname': 'q'},
                    'rebakeTextures': {'args': '<type __builtin__.bool>',
                                       'longname': 'rebakeTextures',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'rbt'},
                    'regenerateUVTilePreview': {'args': '<type __builtin__.str>',
                                                'longname': 'regenerateUVTilePreview',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'rup'},
                    'reloadTextures': {'args': '<type __builtin__.bool>',
                                       'longname': 'reloadTextures',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'rlt'},
                    'reset': {'args': '<type __builtin__.bool>',
                              'longname': 'reset',
                              'modes': ['create', 'query'],
                              'numArgs': 0,
                              'shortname': 'r'},
                    'shaderSource': {'args': '<type __builtin__.str>',
                                     'longname': 'shaderSource',
                                     'modes': ['query'],
                                     'numArgs': 1,
                                     'shortname': 'ss'},
                    'toggleTexturePaging': {'args': '<type __builtin__.bool>',
                                            'longname': 'toggleTexturePaging',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ttp'},
                    'traceRenderPipeline': {'args': '<type __builtin__.bool>',
                                            'longname': 'traceRenderPipeline',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'trp'}},
          'shortFlags': {'di': 'deviceInformation',
                         'drt': 'disposeReleasableTextures',
                         'dt': 'dumpTexture',
                         'fe': 'fragmentEditor',
                         'gmt': 'gpuMemoryTotal',
                         'gpu': 'gpuMemoryUsed',
                         'hwi': 'enableHardwareInstancing',
                         'irs': 'isWinRemoteSession',
                         'lve': 'isLegacyViewportEnabled',
                         'p': 'pause',
                         'q': 'query',
                         'r': 'reset',
                         'rbt': 'rebakeTextures',
                         'rlt': 'reloadTextures',
                         'rup': 'regenerateUVTilePreview',
                         'ss': 'shaderSource',
                         'trp': 'traceRenderPipeline',
                         'ttp': 'toggleTexturePaging',
                         'xml': 'fragmentXML'},
          'type': 'system'},
  'ogsRender': {'flags': {'activeMultisampleType': {'args': '<type __builtin__.str>',
                                                    'longname': 'activeMultisampleType',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'mst'},
                          'activeRenderOverride': {'args': '<type __builtin__.str>',
                                                   'longname': 'activeRenderOverride',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cro'},
                          'activeRenderTargetFormat': {'args': '<type __builtin__.str>',
                                                       'longname': 'activeRenderTargetFormat',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'fpt'},
                          'availableFloatingPointTargetFormat': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'availableFloatingPointTargetFormat',
                                                                 'modes': ['query',
                                                                           'edit'],
                                                                 'numArgs': 0,
                                                                 'shortname': 'afp'},
                          'availableMultisampleType': {'args': '<type __builtin__.bool>',
                                                       'longname': 'availableMultisampleType',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'amt'},
                          'availableRenderOverrides': {'args': '<type __builtin__.bool>',
                                                       'longname': 'availableRenderOverrides',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'aro'},
                          'camera': {'args': '<type __builtin__.str>',
                                     'longname': 'camera',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'cam'},
                          'currentFrame': {'args': '<type __builtin__.bool>',
                                           'longname': 'currentFrame',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 0,
                                           'shortname': 'cf'},
                          'currentView': {'args': '<type __builtin__.bool>',
                                          'longname': 'currentView',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'cv'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'enableFloatingPointRenderTarget': {'args': '<type __builtin__.bool>',
                                                              'longname': 'enableFloatingPointRenderTarget',
                                                              'modes': ['query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'efp'},
                          'enableMultisample': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableMultisample',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ems'},
                          'frame': {'args': '<type __builtin__.float>',
                                    'longname': 'frame',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'f'},
                          'height': {'args': '<type __builtin__.int>',
                                     'longname': 'height',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'h'},
                          'layer': {'args': 'PyNode',
                                    'longname': 'layer',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'l'},
                          'noRenderView': {'args': '<type __builtin__.bool>',
                                           'longname': 'noRenderView',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 0,
                                           'shortname': 'nrv'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'width': {'args': '<type __builtin__.int>',
                                    'longname': 'width',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'w'}},
                'shortFlags': {'afp': 'availableFloatingPointTargetFormat',
                               'amt': 'availableMultisampleType',
                               'aro': 'availableRenderOverrides',
                               'cam': 'camera',
                               'cf': 'currentFrame',
                               'cro': 'activeRenderOverride',
                               'cv': 'currentView',
                               'e': 'edit',
                               'efp': 'enableFloatingPointRenderTarget',
                               'ems': 'enableMultisample',
                               'f': 'frame',
                               'fpt': 'activeRenderTargetFormat',
                               'h': 'height',
                               'l': 'layer',
                               'mst': 'activeMultisampleType',
                               'nrv': 'noRenderView',
                               'q': 'query',
                               'w': 'width'},
                'type': 'rendering'},
  'ogsdebug': {'flags': {'count': {'args': '<type __builtin__.int>',
                                   'longname': 'count',
                                   'numArgs': 1,
                                   'shortname': 'c'},
                         'debug': {'args': '<type __builtin__.str>',
                                   'longname': 'debug',
                                   'numArgs': 1,
                                   'shortname': 'd'},
                         'timing': {'args': '<type __builtin__.str>',
                                    'longname': 'timing',
                                    'numArgs': 1,
                                    'shortname': 't'},
                         'verbose': {'args': '<type __builtin__.bool>',
                                     'longname': 'verbose',
                                     'numArgs': 0,
                                     'shortname': 'v'}},
               'shortFlags': {'c': 'count',
                              'd': 'debug',
                              't': 'timing',
                              'v': 'verbose'},
               'type': 'other'},
  'openGLExtension': {'flags': {'extension': {'args': '<type __builtin__.str>',
                                              'longname': 'extension',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ext'},
                                'renderer': {'args': '<type __builtin__.bool>',
                                             'longname': 'renderer',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'rnd'},
                                'vendor': {'args': '<type __builtin__.bool>',
                                           'longname': 'vendor',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'vnd'},
                                'version': {'args': '<type __builtin__.bool>',
                                            'longname': 'version',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ver'}},
                      'shortFlags': {'ext': 'extension',
                                     'rnd': 'renderer',
                                     'ver': 'version',
                                     'vnd': 'vendor'},
                      'type': 'system'},
  'openMayaPref': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'errlog': {'args': '<type __builtin__.bool>',
                                        'longname': 'errlog',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'el'},
                             'lazyLoad': {'args': '<type __builtin__.bool>',
                                          'longname': 'lazyLoad',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'lz'},
                             'oldPluginWarning': {'args': '<type __builtin__.bool>',
                                                  'longname': 'oldPluginWarning',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ow'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'e': 'edit',
                                  'el': 'errlog',
                                  'lz': 'lazyLoad',
                                  'ow': 'oldPluginWarning',
                                  'q': 'query'},
                   'type': 'system'},
  'optionMenu': {'callbackFlags': ['beforeShowPopup',
                                   'bsp',
                                   'dropCallback',
                                   'dpc',
                                   'visibleChangeCommand',
                                   'vcc',
                                   'dragCallback',
                                   'dgc',
                                   'alwaysCallChangeCommand',
                                   'acc',
                                   'changeCommand',
                                   'cc'],
                 'flags': {'alwaysCallChangeCommand': {'args': '<type __builtin__.bool>',
                                                       'longname': 'alwaysCallChangeCommand',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 0,
                                                       'shortname': 'acc'},
                           'annotation': {'args': '<type __builtin__.str>',
                                          'longname': 'annotation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ann'},
                           'backgroundColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'backgroundColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'bgc'},
                           'beforeShowPopup': {'args': 'script',
                                               'longname': 'beforeShowPopup',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'bsp'},
                           'changeCommand': {'args': 'script',
                                             'longname': 'changeCommand',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cc'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'deleteAllItems': {'args': '<type __builtin__.bool>',
                                              'longname': 'deleteAllItems',
                                              'numArgs': 0,
                                              'shortname': 'dai'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'dragCallback': {'args': 'script',
                                            'longname': 'dragCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dgc'},
                           'dropCallback': {'args': 'script',
                                            'longname': 'dropCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dpc'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'enable': {'args': '<type __builtin__.bool>',
                                      'longname': 'enable',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'en'},
                           'enableBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableBackground',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ebg'},
                           'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableKeyboardFocus',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ekf'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'fullPathName': {'args': '<type __builtin__.bool>',
                                            'longname': 'fullPathName',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fpn'},
                           'height': {'args': '<type __builtin__.int>',
                                      'longname': 'height',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                           'highlightColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'highlightColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'hlc'},
                           'isObscured': {'args': '<type __builtin__.bool>',
                                          'longname': 'isObscured',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'io'},
                           'itemListLong': {'args': '<type __builtin__.bool>',
                                            'longname': 'itemListLong',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ill'},
                           'itemListShort': {'args': '<type __builtin__.bool>',
                                             'longname': 'itemListShort',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'ils'},
                           'label': {'args': '<type __builtin__.str>',
                                     'longname': 'label',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'l'},
                           'manage': {'args': '<type __builtin__.bool>',
                                      'longname': 'manage',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                           'maxVisibleItems': {'args': '<type __builtin__.int>',
                                               'longname': 'maxVisibleItems',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'mvi'},
                           'noBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'noBackground',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nbg'},
                           'numberOfItems': {'args': '<type __builtin__.bool>',
                                             'longname': 'numberOfItems',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'ni'},
                           'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'numberOfPopupMenus',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'npm'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuArray': {'args': '<type __builtin__.bool>',
                                              'longname': 'popupMenuArray',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'pma'},
                           'preventOverride': {'args': '<type __builtin__.bool>',
                                               'longname': 'preventOverride',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'po'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'select': {'args': '<type __builtin__.int>',
                                      'longname': 'select',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'sl'},
                           'statusBarMessage': {'args': '<type __builtin__.str>',
                                                'longname': 'statusBarMessage',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sbm'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'},
                           'value': {'args': '<type __builtin__.str>',
                                     'longname': 'value',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'v'},
                           'visible': {'args': '<type __builtin__.bool>',
                                       'longname': 'visible',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vis'},
                           'visibleChangeCommand': {'args': 'script',
                                                    'longname': 'visibleChangeCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'vcc'},
                           'width': {'args': '<type __builtin__.int>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'}},
                 'shortFlags': {'acc': 'alwaysCallChangeCommand',
                                'ann': 'annotation',
                                'bgc': 'backgroundColor',
                                'bsp': 'beforeShowPopup',
                                'cc': 'changeCommand',
                                'dai': 'deleteAllItems',
                                'dgc': 'dragCallback',
                                'dpc': 'dropCallback',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'ebg': 'enableBackground',
                                'ekf': 'enableKeyboardFocus',
                                'en': 'enable',
                                'ex': 'exists',
                                'fpn': 'fullPathName',
                                'h': 'height',
                                'hlc': 'highlightColor',
                                'ill': 'itemListLong',
                                'ils': 'itemListShort',
                                'io': 'isObscured',
                                'l': 'label',
                                'm': 'manage',
                                'mvi': 'maxVisibleItems',
                                'nbg': 'noBackground',
                                'ni': 'numberOfItems',
                                'npm': 'numberOfPopupMenus',
                                'p': 'parent',
                                'pma': 'popupMenuArray',
                                'po': 'preventOverride',
                                'q': 'query',
                                'sbm': 'statusBarMessage',
                                'sl': 'select',
                                'ut': 'useTemplate',
                                'v': 'value',
                                'vcc': 'visibleChangeCommand',
                                'vis': 'visible',
                                'w': 'width'},
                 'type': 'windows'},
  'optionMenuGrp': {'callbackFlags': ['visibleChangeCommand',
                                      'vcc',
                                      'dragCallback',
                                      'dgc',
                                      'changeCommand',
                                      'cc',
                                      'dropCallback',
                                      'dpc'],
                    'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'adj'},
                              'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn2',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad2'},
                              'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn3',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad3'},
                              'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn4',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad4'},
                              'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn5',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad5'},
                              'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn6',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ad6'},
                              'annotation': {'args': '<type __builtin__.str>',
                                             'longname': 'annotation',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ann'},
                              'backgroundColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'backgroundColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'bgc'},
                              'changeCommand': {'args': 'script',
                                                'longname': 'changeCommand',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cc'},
                              'columnAlign': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'cal'},
                              'columnAlign2': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'cl2'},
                              'columnAlign3': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'cl3'},
                              'columnAlign4': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'cl4'},
                              'columnAlign5': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'cl5'},
                              'columnAlign6': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'cl6'},
                              'columnAttach': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnAttach',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 3,
                                               'shortname': 'cat'},
                              'columnAttach2': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach2',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'ct2'},
                              'columnAttach3': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach3',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'ct3'},
                              'columnAttach4': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach4',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 4,
                                                'shortname': 'ct4'},
                              'columnAttach5': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach5',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 5,
                                                'shortname': 'ct5'},
                              'columnAttach6': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAttach6',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 6,
                                                'shortname': 'ct6'},
                              'columnOffset2': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset2',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'co2'},
                              'columnOffset3': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset3',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'co3'},
                              'columnOffset4': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset4',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 4,
                                                'shortname': 'co4'},
                              'columnOffset5': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset5',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 5,
                                                'shortname': 'co5'},
                              'columnOffset6': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnOffset6',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 6,
                                                'shortname': 'co6'},
                              'columnWidth': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'cw'},
                              'columnWidth1': {'args': '<type __builtin__.int>',
                                               'longname': 'columnWidth1',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cw1'},
                              'columnWidth2': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'cw2'},
                              'columnWidth3': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'cw3'},
                              'columnWidth4': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'cw4'},
                              'columnWidth5': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'cw5'},
                              'columnWidth6': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'cw6'},
                              'defineTemplate': {'args': '<type __builtin__.str>',
                                                 'longname': 'defineTemplate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'dt'},
                              'docTag': {'args': '<type __builtin__.str>',
                                         'longname': 'docTag',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dtg'},
                              'dragCallback': {'args': 'script',
                                               'longname': 'dragCallback',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dgc'},
                              'dropCallback': {'args': 'script',
                                               'longname': 'dropCallback',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dpc'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'enable': {'args': '<type __builtin__.bool>',
                                         'longname': 'enable',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'en'},
                              'enableBackground': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableBackground',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ebg'},
                              'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'enableKeyboardFocus',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ekf'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'extraLabel': {'args': '<type __builtin__.str>',
                                             'longname': 'extraLabel',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'el'},
                              'fullPathName': {'args': '<type __builtin__.bool>',
                                               'longname': 'fullPathName',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'fpn'},
                              'height': {'args': '<type __builtin__.int>',
                                         'longname': 'height',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'h'},
                              'highlightColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'highlightColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'hlc'},
                              'isObscured': {'args': '<type __builtin__.bool>',
                                             'longname': 'isObscured',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'io'},
                              'itemListLong': {'args': '<type __builtin__.bool>',
                                               'longname': 'itemListLong',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'ill'},
                              'itemListShort': {'args': '<type __builtin__.bool>',
                                                'longname': 'itemListShort',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'ils'},
                              'label': {'args': '<type __builtin__.str>',
                                        'longname': 'label',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'l'},
                              'manage': {'args': '<type __builtin__.bool>',
                                         'longname': 'manage',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                              'noBackground': {'args': '<type __builtin__.bool>',
                                               'longname': 'noBackground',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nbg'},
                              'numberOfItems': {'args': '<type __builtin__.bool>',
                                                'longname': 'numberOfItems',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'ni'},
                              'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'numberOfPopupMenus',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'npm'},
                              'parent': {'args': '<type __builtin__.str>',
                                         'longname': 'parent',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'p'},
                              'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                 'longname': 'popupMenuArray',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'pma'},
                              'preventOverride': {'args': '<type __builtin__.bool>',
                                                  'longname': 'preventOverride',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'po'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'rowAttach': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'rowAttach',
                                            'modes': ['create',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 3,
                                            'shortname': 'rat'},
                              'select': {'args': '<type __builtin__.int>',
                                         'longname': 'select',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'sl'},
                              'statusBarMessage': {'args': '<type __builtin__.str>',
                                                   'longname': 'statusBarMessage',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sbm'},
                              'useTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'useTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ut'},
                              'value': {'args': '<type __builtin__.str>',
                                        'longname': 'value',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'v'},
                              'visible': {'args': '<type __builtin__.bool>',
                                          'longname': 'visible',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'vis'},
                              'visibleChangeCommand': {'args': 'script',
                                                       'longname': 'visibleChangeCommand',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'vcc'},
                              'width': {'args': '<type __builtin__.int>',
                                        'longname': 'width',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'w'}},
                    'shortFlags': {'ad2': 'adjustableColumn2',
                                   'ad3': 'adjustableColumn3',
                                   'ad4': 'adjustableColumn4',
                                   'ad5': 'adjustableColumn5',
                                   'ad6': 'adjustableColumn6',
                                   'adj': 'adjustableColumn',
                                   'ann': 'annotation',
                                   'bgc': 'backgroundColor',
                                   'cal': 'columnAlign',
                                   'cat': 'columnAttach',
                                   'cc': 'changeCommand',
                                   'cl2': 'columnAlign2',
                                   'cl3': 'columnAlign3',
                                   'cl4': 'columnAlign4',
                                   'cl5': 'columnAlign5',
                                   'cl6': 'columnAlign6',
                                   'co2': 'columnOffset2',
                                   'co3': 'columnOffset3',
                                   'co4': 'columnOffset4',
                                   'co5': 'columnOffset5',
                                   'co6': 'columnOffset6',
                                   'ct2': 'columnAttach2',
                                   'ct3': 'columnAttach3',
                                   'ct4': 'columnAttach4',
                                   'ct5': 'columnAttach5',
                                   'ct6': 'columnAttach6',
                                   'cw': 'columnWidth',
                                   'cw1': 'columnWidth1',
                                   'cw2': 'columnWidth2',
                                   'cw3': 'columnWidth3',
                                   'cw4': 'columnWidth4',
                                   'cw5': 'columnWidth5',
                                   'cw6': 'columnWidth6',
                                   'dgc': 'dragCallback',
                                   'dpc': 'dropCallback',
                                   'dt': 'defineTemplate',
                                   'dtg': 'docTag',
                                   'e': 'edit',
                                   'ebg': 'enableBackground',
                                   'ekf': 'enableKeyboardFocus',
                                   'el': 'extraLabel',
                                   'en': 'enable',
                                   'ex': 'exists',
                                   'fpn': 'fullPathName',
                                   'h': 'height',
                                   'hlc': 'highlightColor',
                                   'ill': 'itemListLong',
                                   'ils': 'itemListShort',
                                   'io': 'isObscured',
                                   'l': 'label',
                                   'm': 'manage',
                                   'nbg': 'noBackground',
                                   'ni': 'numberOfItems',
                                   'npm': 'numberOfPopupMenus',
                                   'p': 'parent',
                                   'pma': 'popupMenuArray',
                                   'po': 'preventOverride',
                                   'q': 'query',
                                   'rat': 'rowAttach',
                                   'sbm': 'statusBarMessage',
                                   'sl': 'select',
                                   'ut': 'useTemplate',
                                   'v': 'value',
                                   'vcc': 'visibleChangeCommand',
                                   'vis': 'visible',
                                   'w': 'width'},
                    'type': 'windows'},
  'optionVar': {'flags': {'arraySize': {'args': '<type __builtin__.str>',
                                        'longname': 'arraySize',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'arraySize'},
                          'clearArray': {'args': '<type __builtin__.str>',
                                         'longname': 'clearArray',
                                         'modes': ['create', 'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'ca'},
                          'exists': {'args': '<type __builtin__.str>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'ex'},
                          'floatValue': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'floatValue',
                                         'modes': ['create', 'multiuse'],
                                         'numArgs': 2,
                                         'shortname': 'fv'},
                          'floatValueAppend': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'floatValueAppend',
                                               'modes': ['create',
                                                         'multiuse'],
                                               'numArgs': 2,
                                               'shortname': 'fva'},
                          'intValue': {'args': ['<type __builtin__.str>',
                                                '<type __builtin__.int>'],
                                       'longname': 'intValue',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 2,
                                       'shortname': 'iv'},
                          'intValueAppend': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'intValueAppend',
                                             'modes': ['create',
                                                       'multiuse'],
                                             'numArgs': 2,
                                             'shortname': 'iva'},
                          'list': {'args': '<type __builtin__.bool>',
                                   'longname': 'list',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'l'},
                          'query': {'args': '<type __builtin__.str>',
                                    'longname': 'query',
                                    'numArgs': 1,
                                    'shortname': 'q'},
                          'remove': {'args': '<type __builtin__.str>',
                                     'longname': 'remove',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'rm'},
                          'removeFromArray': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'removeFromArray',
                                              'modes': ['create',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'rfa'},
                          'stringValue': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'stringValue',
                                          'modes': ['create', 'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'sv'},
                          'stringValueAppend': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'stringValueAppend',
                                                'modes': ['create',
                                                          'multiuse'],
                                                'numArgs': 2,
                                                'shortname': 'sva'},
                          'version': {'args': '<type __builtin__.int>',
                                      'longname': 'version',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'v'}},
                'removedFlags': {'as': 'arraySize'},
                'shortFlags': {'arraySize': 'arraySize',
                               'ca': 'clearArray',
                               'ex': 'exists',
                               'fv': 'floatValue',
                               'fva': 'floatValueAppend',
                               'iv': 'intValue',
                               'iva': 'intValueAppend',
                               'l': 'list',
                               'q': 'query',
                               'rfa': 'removeFromArray',
                               'rm': 'remove',
                               'sv': 'stringValue',
                               'sva': 'stringValueAppend',
                               'v': 'version'},
                'type': 'language'},
  'orbit': {'flags': {'horizontalAngle': {'args': '<type __builtin__.float>',
                                          'longname': 'horizontalAngle',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ha'},
                      'pivotPoint': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'pivotPoint',
                                     'modes': ['create'],
                                     'numArgs': 3,
                                     'shortname': 'pp'},
                      'rotationAngles': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'rotationAngles',
                                         'modes': ['create'],
                                         'numArgs': 2,
                                         'shortname': 'ra'},
                      'verticalAngle': {'args': '<type __builtin__.float>',
                                        'longname': 'verticalAngle',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'va'}},
            'shortFlags': {'ha': 'horizontalAngle',
                           'pp': 'pivotPoint',
                           'ra': 'rotationAngles',
                           'va': 'verticalAngle'},
            'type': 'rendering'},
  'orbitCtx': {'flags': {'alternateContext': {'args': '<type __builtin__.bool>',
                                              'longname': 'alternateContext',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'ac'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'history': {'args': '<type __builtin__.bool>',
                                     'longname': 'history',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'ch'},
                         'image1': {'args': '<type __builtin__.str>',
                                    'longname': 'image1',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i1'},
                         'image2': {'args': '<type __builtin__.str>',
                                    'longname': 'image2',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i2'},
                         'image3': {'args': '<type __builtin__.str>',
                                    'longname': 'image3',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i3'},
                         'localOrbit': {'args': '<type __builtin__.bool>',
                                        'longname': 'localOrbit',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'lo'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'orbitScale': {'args': '<type __builtin__.float>',
                                        'longname': 'orbitScale',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'os'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'toolName': {'args': '<type __builtin__.str>',
                                      'longname': 'toolName',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'tn'}},
               'shortFlags': {'ac': 'alternateContext',
                              'ch': 'history',
                              'e': 'edit',
                              'ex': 'exists',
                              'i1': 'image1',
                              'i2': 'image2',
                              'i3': 'image3',
                              'lo': 'localOrbit',
                              'n': 'name',
                              'os': 'orbitScale',
                              'q': 'query',
                              'tn': 'toolName'},
               'type': 'context'},
  'orientConstraint': {'flags': {'createCache': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'createCache',
                                                 'modes': ['edit'],
                                                 'numArgs': 2,
                                                 'shortname': 'cc'},
                                 'deleteCache': {'args': '<type __builtin__.bool>',
                                                 'longname': 'deleteCache',
                                                 'modes': ['edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'dc'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'layer': {'args': '<type __builtin__.str>',
                                           'longname': 'layer',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'l'},
                                 'maintainOffset': {'args': '<type __builtin__.bool>',
                                                    'longname': 'maintainOffset',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'mo'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'offset': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'offset',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'o'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'remove': {'args': '<type __builtin__.bool>',
                                            'longname': 'remove',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'rm'},
                                 'skip': {'args': '<type __builtin__.str>',
                                          'longname': 'skip',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'sk'},
                                 'targetList': {'args': '<type __builtin__.bool>',
                                                'longname': 'targetList',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'tl'},
                                 'weight': {'args': '<type __builtin__.float>',
                                            'longname': 'weight',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'w'},
                                 'weightAliasList': {'args': '<type __builtin__.bool>',
                                                     'longname': 'weightAliasList',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'wal'}},
                       'resultNeedsUnpacking': True,
                       'shortFlags': {'cc': 'createCache',
                                      'dc': 'deleteCache',
                                      'e': 'edit',
                                      'l': 'layer',
                                      'mo': 'maintainOffset',
                                      'n': 'name',
                                      'o': 'offset',
                                      'q': 'query',
                                      'rm': 'remove',
                                      'sk': 'skip',
                                      'tl': 'targetList',
                                      'w': 'weight',
                                      'wal': 'weightAliasList'},
                       'type': 'animation'},
  'outlinerEditor': {'callbackFlags': ['selectCommand',
                                       'sec',
                                       'sortCommand',
                                       'soc'],
                     'flags': {'allowMultiSelection': {'args': '<type __builtin__.bool>',
                                                       'longname': 'allowMultiSelection',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ams'},
                               'alwaysToggleSelect': {'args': '<type __builtin__.bool>',
                                                      'longname': 'alwaysToggleSelect',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ats'},
                               'animLayerFilterOptions': {'args': '<type __builtin__.str>',
                                                          'longname': 'animLayerFilterOptions',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'alf'},
                               'attrAlphaOrder': {'args': '<type __builtin__.str>',
                                                  'longname': 'attrAlphaOrder',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'aao'},
                               'attrFilter': {'args': '<type __builtin__.str>',
                                              'longname': 'attrFilter',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'af'},
                               'autoExpand': {'args': '<type __builtin__.bool>',
                                              'longname': 'autoExpand',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'xpd'},
                               'autoExpandLayers': {'args': '<type __builtin__.bool>',
                                                    'longname': 'autoExpandLayers',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ael'},
                               'autoSelectNewObjects': {'args': '<type __builtin__.bool>',
                                                        'longname': 'autoSelectNewObjects',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'autoSelectNewObjects'},
                               'containersIgnoreFilters': {'args': '<type __builtin__.bool>',
                                                           'longname': 'containersIgnoreFilters',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'cif'},
                               'control': {'args': '<type __builtin__.bool>',
                                           'longname': 'control',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'ctl'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'directSelect': {'args': '<type __builtin__.bool>',
                                                'longname': 'directSelect',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ds'},
                               'displayMode': {'args': '<type __builtin__.str>',
                                               'longname': 'displayMode',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dm'},
                               'doNotSelectNewObjects': {'args': '<type __builtin__.bool>',
                                                         'longname': 'doNotSelectNewObjects',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'dns'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'dropIsParent': {'args': '<type __builtin__.bool>',
                                                'longname': 'dropIsParent',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dip'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'editAttrName': {'args': '<type __builtin__.bool>',
                                                'longname': 'editAttrName',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ean'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'expandAllItems': {'args': '<type __builtin__.bool>',
                                                  'longname': 'expandAllItems',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'eai'},
                               'expandAllSelectedItems': {'args': '<type __builtin__.bool>',
                                                          'longname': 'expandAllSelectedItems',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'eas'},
                               'expandAttribute': {'args': '<type __builtin__.bool>',
                                                   'longname': 'expandAttribute',
                                                   'modes': ['edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'att'},
                               'expandConnections': {'args': '<type __builtin__.bool>',
                                                     'longname': 'expandConnections',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'xc'},
                               'expandObjects': {'args': '<type __builtin__.bool>',
                                                 'longname': 'expandObjects',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'eo'},
                               'feedbackItemName': {'args': '<type __builtin__.bool>',
                                                    'longname': 'feedbackItemName',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'fbn'},
                               'feedbackRowNumber': {'args': '<type __builtin__.bool>',
                                                     'longname': 'feedbackRowNumber',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'fbr'},
                               'filter': {'args': '<type __builtin__.str>',
                                          'longname': 'filter',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'f'},
                               'forceMainConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'forceMainConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'fmc'},
                               'getCurrentSetOfItem': {'args': '<type __builtin__.int>',
                                                       'longname': 'getCurrentSetOfItem',
                                                       'modes': ['query'],
                                                       'numArgs': 1,
                                                       'shortname': 'gcs'},
                               'highlightActive': {'args': '<type __builtin__.bool>',
                                                   'longname': 'highlightActive',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ha'},
                               'highlightConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'highlightConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'hlc'},
                               'highlightSecondary': {'args': '<type __builtin__.bool>',
                                                      'longname': 'highlightSecondary',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'hs'},
                               'ignoreDagHierarchy': {'args': '<type __builtin__.bool>',
                                                      'longname': 'ignoreDagHierarchy',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'hir'},
                               'ignoreHiddenAttribute': {'args': '<type __builtin__.bool>',
                                                         'longname': 'ignoreHiddenAttribute',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'iha'},
                               'ignoreOutlinerColor': {'args': '<type __builtin__.bool>',
                                                       'longname': 'ignoreOutlinerColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ioc'},
                               'isChildSelected': {'args': 'PyNode',
                                                   'longname': 'isChildSelected',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ics'},
                               'isSet': {'args': '<type __builtin__.int>',
                                         'longname': 'isSet',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'isSet'},
                               'isSetMember': {'args': '<type __builtin__.int>',
                                               'longname': 'isSetMember',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'ism'},
                               'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                      'longname': 'lockMainConnection',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'lck'},
                               'longNames': {'args': '<type __builtin__.bool>',
                                             'longname': 'longNames',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ln'},
                               'mainListConnection': {'args': '<type __builtin__.str>',
                                                      'longname': 'mainListConnection',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'mlc'},
                               'mapMotionTrails': {'args': '<type __builtin__.bool>',
                                                   'longname': 'mapMotionTrails',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'mmt'},
                               'masterOutliner': {'args': '<type __builtin__.str>',
                                                  'longname': 'masterOutliner',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mst'},
                               'niceNames': {'args': '<type __builtin__.bool>',
                                             'longname': 'niceNames',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nn'},
                               'object': {'args': 'PyNode',
                                          'longname': 'object',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'obj'},
                               'organizeByClip': {'args': '<type __builtin__.bool>',
                                                  'longname': 'organizeByClip',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'obc'},
                               'organizeByLayer': {'args': '<type __builtin__.bool>',
                                                   'longname': 'organizeByLayer',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'obl'},
                               'panel': {'args': '<type __builtin__.str>',
                                         'longname': 'panel',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'pnl'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'parentObject': {'args': '<type __builtin__.bool>',
                                                'longname': 'parentObject',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'po'},
                               'pinPlug': {'args': 'PyNode',
                                           'longname': 'pinPlug',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pin'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'refresh': {'args': '<type __builtin__.bool>',
                                           'longname': 'refresh',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'rfs'},
                               'removeFromCurrentSet': {'args': '<type __builtin__.int>',
                                                        'longname': 'removeFromCurrentSet',
                                                        'modes': ['edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'rcs'},
                               'renameItem': {'args': '<type __builtin__.int>',
                                              'longname': 'renameItem',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'rni'},
                               'renameSelectedItem': {'args': '<type __builtin__.bool>',
                                                      'longname': 'renameSelectedItem',
                                                      'modes': ['edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'rsi'},
                               'renderFilterActive': {'args': '<type __builtin__.bool>',
                                                      'longname': 'renderFilterActive',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'rfa'},
                               'renderFilterIndex': {'args': '<type __builtin__.int>',
                                                     'longname': 'renderFilterIndex',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'rfi'},
                               'renderFilterVisible': {'args': '<type __builtin__.bool>',
                                                       'longname': 'renderFilterVisible',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'rfv'},
                               'selectCommand': {'args': 'script',
                                                 'longname': 'selectCommand',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sec'},
                               'selectionConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'selectionConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'slc'},
                               'selectionOrder': {'args': '<type __builtin__.str>',
                                                  'longname': 'selectionOrder',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sod'},
                               'setFilter': {'args': '<type __builtin__.str>',
                                             'longname': 'setFilter',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sf'},
                               'setsIgnoreFilters': {'args': '<type __builtin__.bool>',
                                                     'longname': 'setsIgnoreFilters',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sif'},
                               'showAnimCurvesOnly': {'args': '<type __builtin__.bool>',
                                                      'longname': 'showAnimCurvesOnly',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'aco'},
                               'showAnimLayerWeight': {'args': '<type __builtin__.bool>',
                                                       'longname': 'showAnimLayerWeight',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'saw'},
                               'showAssets': {'args': '<type __builtin__.bool>',
                                              'longname': 'showAssets',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'a'},
                               'showAssignedMaterials': {'args': '<type __builtin__.bool>',
                                                         'longname': 'showAssignedMaterials',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'sam'},
                               'showAttrValues': {'args': '<type __builtin__.bool>',
                                                  'longname': 'showAttrValues',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'av'},
                               'showAttributes': {'args': '<type __builtin__.bool>',
                                                  'longname': 'showAttributes',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'atr'},
                               'showCompounds': {'args': '<type __builtin__.bool>',
                                                 'longname': 'showCompounds',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cmp'},
                               'showConnected': {'args': '<type __builtin__.bool>',
                                                 'longname': 'showConnected',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'con'},
                               'showContainedOnly': {'args': '<type __builtin__.bool>',
                                                     'longname': 'showContainedOnly',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sco'},
                               'showContainerContents': {'args': '<type __builtin__.bool>',
                                                         'longname': 'showContainerContents',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'scc'},
                               'showDagOnly': {'args': '<type __builtin__.bool>',
                                               'longname': 'showDagOnly',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dag'},
                               'showLeafs': {'args': '<type __builtin__.bool>',
                                             'longname': 'showLeafs',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'laf'},
                               'showMuteInfo': {'args': '<type __builtin__.bool>',
                                                'longname': 'showMuteInfo',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'smi'},
                               'showNamespace': {'args': '<type __builtin__.bool>',
                                                 'longname': 'showNamespace',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sn'},
                               'showNumericAttrsOnly': {'args': '<type __builtin__.bool>',
                                                        'longname': 'showNumericAttrsOnly',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'num'},
                               'showParentContainers': {'args': '<type __builtin__.bool>',
                                                        'longname': 'showParentContainers',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'spa'},
                               'showPinIcons': {'args': '<type __builtin__.bool>',
                                                'longname': 'showPinIcons',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'spi'},
                               'showPublishedAsConnected': {'args': '<type __builtin__.bool>',
                                                            'longname': 'showPublishedAsConnected',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'spc'},
                               'showReferenceMembers': {'args': '<type __builtin__.bool>',
                                                        'longname': 'showReferenceMembers',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'rm'},
                               'showReferenceNodes': {'args': '<type __builtin__.bool>',
                                                      'longname': 'showReferenceNodes',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'rn'},
                               'showSelected': {'args': '<type __builtin__.bool>',
                                                'longname': 'showSelected',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sc'},
                               'showSetMembers': {'args': '<type __builtin__.bool>',
                                                  'longname': 'showSetMembers',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ssm'},
                               'showShapes': {'args': '<type __builtin__.bool>',
                                              'longname': 'showShapes',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'shp'},
                               'showTextureNodesOnly': {'args': '<type __builtin__.bool>',
                                                        'longname': 'showTextureNodesOnly',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'tno'},
                               'showTimeEditor': {'args': '<type __builtin__.bool>',
                                                  'longname': 'showTimeEditor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ste'},
                               'showUVAttrsOnly': {'args': '<type __builtin__.bool>',
                                                   'longname': 'showUVAttrsOnly',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'uv'},
                               'showUnitlessCurves': {'args': '<type __builtin__.bool>',
                                                      'longname': 'showUnitlessCurves',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'su'},
                               'showUpstreamCurves': {'args': '<type __builtin__.bool>',
                                                      'longname': 'showUpstreamCurves',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'suc'},
                               'sortCommand': {'args': '<type __builtin__.callable>',
                                               'longname': 'sortCommand',
                                               'numArgs': 1,
                                               'shortname': 'soc'},
                               'sortOrder': {'args': '<type __builtin__.str>',
                                             'longname': 'sortOrder',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'so'},
                               'stateString': {'args': '<type __builtin__.bool>',
                                               'longname': 'stateString',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'sts'},
                               'transmitFilters': {'args': '<type __builtin__.bool>',
                                                   'longname': 'transmitFilters',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'tf'},
                               'unParent': {'args': '<type __builtin__.bool>',
                                            'longname': 'unParent',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'up'},
                               'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                        'longname': 'unlockMainConnection',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'ulk'},
                               'unpinPlug': {'args': 'PyNode',
                                             'longname': 'unpinPlug',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'unp'},
                               'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                        'longname': 'updateMainConnection',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'upd'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'}},
                     'removedFlags': {'as': 'autoSelectNewObjects',
                                      'is': 'isSet'},
                     'shortFlags': {'a': 'showAssets',
                                    'aao': 'attrAlphaOrder',
                                    'aco': 'showAnimCurvesOnly',
                                    'ael': 'autoExpandLayers',
                                    'af': 'attrFilter',
                                    'alf': 'animLayerFilterOptions',
                                    'ams': 'allowMultiSelection',
                                    'atr': 'showAttributes',
                                    'ats': 'alwaysToggleSelect',
                                    'autoSelectNewObjects': 'autoSelectNewObjects',
                                    'av': 'showAttrValues',
                                    'cif': 'containersIgnoreFilters',
                                    'cmp': 'showCompounds',
                                    'con': 'showConnected',
                                    'ctl': 'control',
                                    'dag': 'showDagOnly',
                                    'dip': 'dropIsParent',
                                    'dm': 'displayMode',
                                    'dns': 'doNotSelectNewObjects',
                                    'ds': 'directSelect',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'e': 'edit',
                                    'eai': 'expandAllItems',
                                    'ean': 'editAttrName',
                                    'eas': 'expandAllSelectedItems',
                                    'eat': 'expandAttribute',
                                    'eo': 'expandObjects',
                                    'ex': 'exists',
                                    'f': 'filter',
                                    'fbn': 'feedbackItemName',
                                    'fbr': 'feedbackRowNumber',
                                    'fmc': 'forceMainConnection',
                                    'gcs': 'getCurrentSetOfItem',
                                    'ha': 'highlightActive',
                                    'hir': 'ignoreDagHierarchy',
                                    'hlc': 'highlightConnection',
                                    'hs': 'highlightSecondary',
                                    'ics': 'isChildSelected',
                                    'iha': 'ignoreHiddenAttribute',
                                    'ioc': 'ignoreOutlinerColor',
                                    'isSet': 'isSet',
                                    'ism': 'isSetMember',
                                    'laf': 'showLeafs',
                                    'lck': 'lockMainConnection',
                                    'ln': 'longNames',
                                    'mlc': 'mainListConnection',
                                    'mmt': 'mapMotionTrails',
                                    'mst': 'masterOutliner',
                                    'nn': 'niceNames',
                                    'num': 'showNumericAttrsOnly',
                                    'obc': 'organizeByClip',
                                    'obj': 'object',
                                    'obl': 'organizeByLayer',
                                    'p': 'parent',
                                    'pin': 'pinPlug',
                                    'pnl': 'panel',
                                    'po': 'parentObject',
                                    'q': 'query',
                                    'rcs': 'removeFromCurrentSet',
                                    'rfa': 'renderFilterActive',
                                    'rfi': 'renderFilterIndex',
                                    'rfs': 'refresh',
                                    'rfv': 'renderFilterVisible',
                                    'rm': 'showReferenceMembers',
                                    'rn': 'showReferenceNodes',
                                    'rni': 'renameItem',
                                    'rsi': 'renameSelectedItem',
                                    'sam': 'showAssignedMaterials',
                                    'saw': 'showAnimLayerWeight',
                                    'sc': 'showSelected',
                                    'scc': 'showContainerContents',
                                    'sco': 'showContainedOnly',
                                    'sec': 'selectCommand',
                                    'sf': 'setFilter',
                                    'shp': 'showShapes',
                                    'sif': 'setsIgnoreFilters',
                                    'slc': 'selectionConnection',
                                    'smi': 'showMuteInfo',
                                    'sn': 'showNamespace',
                                    'so': 'sortOrder',
                                    'soc': 'sortCommand',
                                    'sod': 'selectionOrder',
                                    'spa': 'showParentContainers',
                                    'spc': 'showPublishedAsConnected',
                                    'spi': 'showPinIcons',
                                    'ssm': 'showSetMembers',
                                    'ste': 'showTimeEditor',
                                    'sts': 'stateString',
                                    'su': 'showUnitlessCurves',
                                    'suc': 'showUpstreamCurves',
                                    'tf': 'transmitFilters',
                                    'tno': 'showTextureNodesOnly',
                                    'ulk': 'unlockMainConnection',
                                    'unp': 'unpinPlug',
                                    'up': 'unParent',
                                    'upd': 'updateMainConnection',
                                    'ut': 'useTemplate',
                                    'uv': 'showUVAttrsOnly',
                                    'xc': 'expandConnections',
                                    'xpd': 'autoExpand'},
                     'type': 'windows'},
  'outlinerPanel': {'callbackFlags': ['popupMenuProcedure', 'pmp'],
                    'flags': {'control': {'args': '<type __builtin__.bool>',
                                          'longname': 'control',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'ctl'},
                              'copy': {'args': '<type __builtin__.str>',
                                       'longname': 'copy',
                                       'modes': ['edit'],
                                       'numArgs': 1,
                                       'shortname': 'cp'},
                              'createString': {'args': '<type __builtin__.bool>',
                                               'longname': 'createString',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'cs'},
                              'defineTemplate': {'args': '<type __builtin__.str>',
                                                 'longname': 'defineTemplate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'dt'},
                              'docTag': {'args': '<type __builtin__.str>',
                                         'longname': 'docTag',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dtg'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'editString': {'args': '<type __builtin__.bool>',
                                             'longname': 'editString',
                                             'modes': ['edit'],
                                             'numArgs': 0,
                                             'shortname': 'es'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'init': {'args': '<type __builtin__.bool>',
                                       'longname': 'init',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'init'},
                              'isUnique': {'args': '<type __builtin__.bool>',
                                           'longname': 'isUnique',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'iu'},
                              'label': {'args': '<type __builtin__.str>',
                                        'longname': 'label',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'l'},
                              'menuBarRepeatLast': {'args': '<type __builtin__.bool>',
                                                    'longname': 'menuBarRepeatLast',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'mrl'},
                              'menuBarVisible': {'args': '<type __builtin__.bool>',
                                                 'longname': 'menuBarVisible',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'mbv'},
                              'needsInit': {'args': '<type __builtin__.bool>',
                                            'longname': 'needsInit',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'ni'},
                              'outlinerEditor': {'args': '<type __builtin__.bool>',
                                                 'longname': 'outlinerEditor',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'oe'},
                              'parent': {'args': '<type __builtin__.str>',
                                         'longname': 'parent',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'p'},
                              'popupMenuProcedure': {'args': 'script',
                                                     'longname': 'popupMenuProcedure',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'pmp'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'replacePanel': {'args': '<type __builtin__.str>',
                                               'longname': 'replacePanel',
                                               'modes': ['edit'],
                                               'numArgs': 1,
                                               'shortname': 'rp'},
                              'tearOff': {'args': '<type __builtin__.bool>',
                                          'longname': 'tearOff',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'to'},
                              'tearOffCopy': {'args': '<type __builtin__.str>',
                                              'longname': 'tearOffCopy',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'toc'},
                              'tearOffRestore': {'args': '<type __builtin__.bool>',
                                                 'longname': 'tearOffRestore',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'tor'},
                              'unParent': {'args': '<type __builtin__.bool>',
                                           'longname': 'unParent',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'up'},
                              'useTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'useTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ut'}},
                    'removedFlags': {'in': 'init'},
                    'shortFlags': {'cp': 'copy',
                                   'cs': 'createString',
                                   'ctl': 'control',
                                   'dt': 'defineTemplate',
                                   'dtg': 'docTag',
                                   'e': 'edit',
                                   'es': 'editString',
                                   'ex': 'exists',
                                   'init': 'init',
                                   'iu': 'isUnique',
                                   'l': 'label',
                                   'mbv': 'menuBarVisible',
                                   'mrl': 'menuBarRepeatLast',
                                   'ni': 'needsInit',
                                   'oe': 'outlinerEditor',
                                   'p': 'parent',
                                   'pmp': 'popupMenuProcedure',
                                   'q': 'query',
                                   'rp': 'replacePanel',
                                   'to': 'tearOff',
                                   'toc': 'tearOffCopy',
                                   'tor': 'tearOffRestore',
                                   'up': 'unParent',
                                   'ut': 'useTemplate'},
                    'type': 'windows'},
  'overrideModifier': {'flags': {'clear': {'args': '<type __builtin__.bool>',
                                           'longname': 'clear',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'cl'},
                                 'press': {'args': '<type __builtin__.str>',
                                           'longname': 'press',
                                           'modes': ['create', 'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'p'},
                                 'release': {'args': '<type __builtin__.str>',
                                             'longname': 'release',
                                             'modes': ['create',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'r'}},
                       'shortFlags': {},
                       'type': 'windows'},
  'paint3d': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'paintEffectsDisplay': {'flags': {'meshDrawEnable': {'args': '<type __builtin__.bool>',
                                                       'longname': 'meshDrawEnable',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'me'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'}},
                          'shortFlags': {'me': 'meshDrawEnable',
                                         'q': 'query'},
                          'type': 'effects'},
  'pairBlend': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                        'longname': 'attribute',
                                        'modes': ['create', 'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'at'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'input1': {'args': '<type __builtin__.bool>',
                                     'longname': 'input1',
                                     'modes': ['query'],
                                     'numArgs': 0,
                                     'shortname': 'i1'},
                          'input2': {'args': '<type __builtin__.bool>',
                                     'longname': 'input2',
                                     'modes': ['query'],
                                     'numArgs': 0,
                                     'shortname': 'i2'},
                          'node': {'args': '<type __builtin__.str>',
                                   'longname': 'node',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'nd'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'}},
                'shortFlags': {'at': 'attribute',
                               'e': 'edit',
                               'i1': 'input1',
                               'i2': 'input2',
                               'nd': 'node',
                               'q': 'query'},
                'type': 'animation'},
  'palettePort': {'callbackFlags': ['colorEdited',
                                    'ce',
                                    'dropCallback',
                                    'dpc',
                                    'visibleChangeCommand',
                                    'vcc',
                                    'dragCallback',
                                    'dgc',
                                    'changeCommand',
                                    'cc'],
                  'flags': {'actualTotal': {'args': '<type __builtin__.int>',
                                            'longname': 'actualTotal',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'at'},
                            'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'changeCommand': {'args': 'script',
                                              'longname': 'changeCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cc'},
                            'colorEditable': {'args': '<type __builtin__.bool>',
                                              'longname': 'colorEditable',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ced'},
                            'colorEdited': {'args': 'script',
                                            'longname': 'colorEdited',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ce'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'dimensions': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'dimensions',
                                           'modes': ['create', 'query'],
                                           'numArgs': 2,
                                           'shortname': 'dim'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'editable': {'args': '<type __builtin__.bool>',
                                         'longname': 'editable',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ed'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'hsvValue': {'args': ['<type __builtin__.int>',
                                                  '<type __builtin__.int>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'hsvValue',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 5,
                                         'shortname': 'hsv'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'redraw': {'args': '<type __builtin__.bool>',
                                       'longname': 'redraw',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'r'},
                            'rgbValue': {'args': ['<type __builtin__.int>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'rgbValue',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 4,
                                         'shortname': 'rgb'},
                            'setCurCell': {'args': '<type __builtin__.int>',
                                           'longname': 'setCurCell',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'scc'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'topDown': {'args': '<type __builtin__.bool>',
                                        'longname': 'topDown',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'td'},
                            'transparent': {'args': '<type __builtin__.int>',
                                            'longname': 'transparent',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 't'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'ann': 'annotation',
                                 'at': 'actualTotal',
                                 'bgc': 'backgroundColor',
                                 'cc': 'changeCommand',
                                 'ce': 'colorEdited',
                                 'ced': 'colorEditable',
                                 'dgc': 'dragCallback',
                                 'dim': 'dimensions',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ed': 'editable',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'ex': 'exists',
                                 'fpn': 'fullPathName',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'hsv': 'hsvValue',
                                 'io': 'isObscured',
                                 'm': 'manage',
                                 'nbg': 'noBackground',
                                 'npm': 'numberOfPopupMenus',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'r': 'redraw',
                                 'rgb': 'rgbValue',
                                 'sbm': 'statusBarMessage',
                                 'scc': 'setCurCell',
                                 't': 'transparent',
                                 'td': 'topDown',
                                 'ut': 'useTemplate',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'panZoom': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                     'longname': 'absolute',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'abs'},
                        'downDistance': {'args': '<type __builtin__.float>',
                                         'longname': 'downDistance',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'd'},
                        'leftDistance': {'args': '<type __builtin__.float>',
                                         'longname': 'leftDistance',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'l'},
                        'relative': {'args': '<type __builtin__.bool>',
                                     'longname': 'relative',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'rel'},
                        'rightDistance': {'args': '<type __builtin__.float>',
                                          'longname': 'rightDistance',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'r'},
                        'upDistance': {'args': '<type __builtin__.float>',
                                       'longname': 'upDistance',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'u'},
                        'zoomRatio': {'args': '<type __builtin__.float>',
                                      'longname': 'zoomRatio',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'z'}},
              'shortFlags': {'abs': 'absolute',
                             'd': 'downDistance',
                             'l': 'leftDistance',
                             'r': 'rightDistance',
                             'rel': 'relative',
                             'u': 'upDistance',
                             'z': 'zoomRatio'},
              'type': 'rendering'},
  'panZoomCtx': {'flags': {'alternateContext': {'args': '<type __builtin__.bool>',
                                                'longname': 'alternateContext',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'ac'},
                           'buttonDown': {'args': '<type __builtin__.bool>',
                                          'longname': 'buttonDown',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'btd'},
                           'buttonUp': {'args': '<type __builtin__.bool>',
                                        'longname': 'buttonUp',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'btu'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'history': {'args': '<type __builtin__.bool>',
                                       'longname': 'history',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ch'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'panMode': {'args': '<type __builtin__.bool>',
                                       'longname': 'panMode',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'pm'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'toolName': {'args': '<type __builtin__.str>',
                                        'longname': 'toolName',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'tn'},
                           'zoomMode': {'args': '<type __builtin__.bool>',
                                        'longname': 'zoomMode',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'zm'},
                           'zoomScale': {'args': '<type __builtin__.float>',
                                         'longname': 'zoomScale',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'zs'}},
                 'shortFlags': {'ac': 'alternateContext',
                                'btd': 'buttonDown',
                                'btu': 'buttonUp',
                                'ch': 'history',
                                'e': 'edit',
                                'ex': 'exists',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'n': 'name',
                                'pm': 'panMode',
                                'q': 'query',
                                'tn': 'toolName',
                                'zm': 'zoomMode',
                                'zs': 'zoomScale'},
                 'type': 'context'},
  'paneLayout': {'callbackFlags': ['visibleChangeCommand',
                                   'vcc',
                                   'dragCallback',
                                   'dgc',
                                   'separatorMovedCommand',
                                   'smc',
                                   'dropCallback',
                                   'dpc'],
                 'flags': {'activeFrameThickness': {'args': '<type __builtin__.int>',
                                                    'longname': 'activeFrameThickness',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'aft'},
                           'activePane': {'args': '<type __builtin__.str>',
                                          'longname': 'activePane',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ap'},
                           'activePaneIndex': {'args': '<type __builtin__.int>',
                                               'longname': 'activePaneIndex',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'api'},
                           'annotation': {'args': '<type __builtin__.str>',
                                          'longname': 'annotation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ann'},
                           'backgroundColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'backgroundColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'bgc'},
                           'childArray': {'args': '<type __builtin__.bool>',
                                          'longname': 'childArray',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'ca'},
                           'configuration': {'args': '<type __builtin__.str>',
                                             'longname': 'configuration',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cn'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'dragCallback': {'args': 'script',
                                            'longname': 'dragCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dgc'},
                           'dropCallback': {'args': 'script',
                                            'longname': 'dropCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dpc'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'enable': {'args': '<type __builtin__.bool>',
                                      'longname': 'enable',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'en'},
                           'enableBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableBackground',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ebg'},
                           'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableKeyboardFocus',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ekf'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'fullPathName': {'args': '<type __builtin__.bool>',
                                            'longname': 'fullPathName',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fpn'},
                           'height': {'args': '<type __builtin__.int>',
                                      'longname': 'height',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                           'highlightColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'highlightColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'hlc'},
                           'isObscured': {'args': '<type __builtin__.bool>',
                                          'longname': 'isObscured',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'io'},
                           'manage': {'args': '<type __builtin__.bool>',
                                      'longname': 'manage',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                           'noBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'noBackground',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nbg'},
                           'numberOfChildren': {'args': '<type __builtin__.bool>',
                                                'longname': 'numberOfChildren',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'nch'},
                           'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'numberOfPopupMenus',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'npm'},
                           'numberOfVisiblePanes': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfVisiblePanes',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'nvp'},
                           'pane1': {'args': '<type __builtin__.bool>',
                                     'longname': 'pane1',
                                     'modes': ['query'],
                                     'numArgs': 0,
                                     'shortname': 'p1'},
                           'pane2': {'args': '<type __builtin__.bool>',
                                     'longname': 'pane2',
                                     'modes': ['query'],
                                     'numArgs': 0,
                                     'shortname': 'p2'},
                           'pane3': {'args': '<type __builtin__.bool>',
                                     'longname': 'pane3',
                                     'modes': ['query'],
                                     'numArgs': 0,
                                     'shortname': 'p3'},
                           'pane4': {'args': '<type __builtin__.bool>',
                                     'longname': 'pane4',
                                     'modes': ['query'],
                                     'numArgs': 0,
                                     'shortname': 'p4'},
                           'paneSize': {'args': ['<type __builtin__.int>',
                                                 '<type __builtin__.int>',
                                                 '<type __builtin__.int>'],
                                        'longname': 'paneSize',
                                        'modes': ['create',
                                                  'query',
                                                  'edit',
                                                  'multiuse'],
                                        'numArgs': 3,
                                        'shortname': 'ps'},
                           'paneUnderPointer': {'args': '<type __builtin__.bool>',
                                                'longname': 'paneUnderPointer',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pup'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuArray': {'args': '<type __builtin__.bool>',
                                              'longname': 'popupMenuArray',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'pma'},
                           'preventOverride': {'args': '<type __builtin__.bool>',
                                               'longname': 'preventOverride',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'po'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'separatorMovedCommand': {'args': 'script',
                                                     'longname': 'separatorMovedCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'smc'},
                           'separatorThickness': {'args': '<type __builtin__.int>',
                                                  'longname': 'separatorThickness',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'st'},
                           'setPane': {'args': ['<type __builtin__.str>',
                                                '<type __builtin__.int>'],
                                       'longname': 'setPane',
                                       'modes': ['create',
                                                 'edit',
                                                 'multiuse'],
                                       'numArgs': 2,
                                       'shortname': 'sp'},
                           'staticHeightPane': {'args': '<type __builtin__.int>',
                                                'longname': 'staticHeightPane',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'shp'},
                           'staticWidthPane': {'args': '<type __builtin__.int>',
                                               'longname': 'staticWidthPane',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'swp'},
                           'statusBarMessage': {'args': '<type __builtin__.str>',
                                                'longname': 'statusBarMessage',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sbm'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'},
                           'visible': {'args': '<type __builtin__.bool>',
                                       'longname': 'visible',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vis'},
                           'visibleChangeCommand': {'args': 'script',
                                                    'longname': 'visibleChangeCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'vcc'},
                           'width': {'args': '<type __builtin__.int>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'}},
                 'shortFlags': {'aft': 'activeFrameThickness',
                                'ann': 'annotation',
                                'ap': 'activePane',
                                'api': 'activePaneIndex',
                                'bgc': 'backgroundColor',
                                'ca': 'childArray',
                                'cn': 'configuration',
                                'dgc': 'dragCallback',
                                'dpc': 'dropCallback',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'ebg': 'enableBackground',
                                'ekf': 'enableKeyboardFocus',
                                'en': 'enable',
                                'ex': 'exists',
                                'fpn': 'fullPathName',
                                'h': 'height',
                                'hlc': 'highlightColor',
                                'io': 'isObscured',
                                'm': 'manage',
                                'nbg': 'noBackground',
                                'nch': 'numberOfChildren',
                                'npm': 'numberOfPopupMenus',
                                'nvp': 'numberOfVisiblePanes',
                                'p': 'parent',
                                'p1': 'pane1',
                                'p2': 'pane2',
                                'p3': 'pane3',
                                'p4': 'pane4',
                                'pma': 'popupMenuArray',
                                'po': 'preventOverride',
                                'ps': 'paneSize',
                                'pup': 'paneUnderPointer',
                                'q': 'query',
                                'sbm': 'statusBarMessage',
                                'shp': 'staticHeightPane',
                                'smc': 'separatorMovedCommand',
                                'sp': 'setPane',
                                'st': 'separatorThickness',
                                'swp': 'staticWidthPane',
                                'ut': 'useTemplate',
                                'vcc': 'visibleChangeCommand',
                                'vis': 'visible',
                                'w': 'width'},
                 'type': 'windows'},
  'panel': {'callbackFlags': ['popupMenuProcedure', 'pmp'],
            'flags': {'control': {'args': '<type __builtin__.bool>',
                                  'longname': 'control',
                                  'modes': ['query'],
                                  'numArgs': 0,
                                  'shortname': 'ctl'},
                      'copy': {'args': '<type __builtin__.str>',
                               'longname': 'copy',
                               'modes': ['edit'],
                               'numArgs': 1,
                               'shortname': 'cp'},
                      'createString': {'args': '<type __builtin__.bool>',
                                       'longname': 'createString',
                                       'modes': ['edit'],
                                       'numArgs': 0,
                                       'shortname': 'cs'},
                      'defineTemplate': {'args': '<type __builtin__.str>',
                                         'longname': 'defineTemplate',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'dt'},
                      'docTag': {'args': '<type __builtin__.str>',
                                 'longname': 'docTag',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'dtg'},
                      'edit': {'args': '<type __builtin__.bool>',
                               'longname': 'edit',
                               'numArgs': 0,
                               'shortname': 'e'},
                      'editString': {'args': '<type __builtin__.bool>',
                                     'longname': 'editString',
                                     'modes': ['edit'],
                                     'numArgs': 0,
                                     'shortname': 'es'},
                      'exists': {'args': '<type __builtin__.bool>',
                                 'longname': 'exists',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'ex'},
                      'init': {'args': '<type __builtin__.bool>',
                               'longname': 'init',
                               'modes': ['create', 'edit'],
                               'numArgs': 0,
                               'shortname': 'init'},
                      'isUnique': {'args': '<type __builtin__.bool>',
                                   'longname': 'isUnique',
                                   'modes': ['query'],
                                   'numArgs': 0,
                                   'shortname': 'iu'},
                      'label': {'args': '<type __builtin__.str>',
                                'longname': 'label',
                                'modes': ['query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'l'},
                      'menuBarRepeatLast': {'args': '<type __builtin__.bool>',
                                            'longname': 'menuBarRepeatLast',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'mrl'},
                      'menuBarVisible': {'args': '<type __builtin__.bool>',
                                         'longname': 'menuBarVisible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'mbv'},
                      'needsInit': {'args': '<type __builtin__.bool>',
                                    'longname': 'needsInit',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'ni'},
                      'parent': {'args': '<type __builtin__.str>',
                                 'longname': 'parent',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'p'},
                      'popupMenuProcedure': {'args': 'script',
                                             'longname': 'popupMenuProcedure',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'pmp'},
                      'query': {'args': '<type __builtin__.bool>',
                                'longname': 'query',
                                'numArgs': 0,
                                'shortname': 'q'},
                      'replacePanel': {'args': '<type __builtin__.str>',
                                       'longname': 'replacePanel',
                                       'modes': ['edit'],
                                       'numArgs': 1,
                                       'shortname': 'rp'},
                      'tearOff': {'args': '<type __builtin__.bool>',
                                  'longname': 'tearOff',
                                  'modes': ['query', 'edit'],
                                  'numArgs': 0,
                                  'shortname': 'to'},
                      'tearOffCopy': {'args': '<type __builtin__.str>',
                                      'longname': 'tearOffCopy',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'toc'},
                      'tearOffRestore': {'args': '<type __builtin__.bool>',
                                         'longname': 'tearOffRestore',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'tor'},
                      'unParent': {'args': '<type __builtin__.bool>',
                                   'longname': 'unParent',
                                   'modes': ['edit'],
                                   'numArgs': 0,
                                   'shortname': 'up'},
                      'useTemplate': {'args': '<type __builtin__.str>',
                                      'longname': 'useTemplate',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'ut'}},
            'removedFlags': {'in': 'init'},
            'shortFlags': {'cp': 'copy',
                           'cs': 'createString',
                           'ctl': 'control',
                           'dt': 'defineTemplate',
                           'dtg': 'docTag',
                           'e': 'edit',
                           'es': 'editString',
                           'ex': 'exists',
                           'init': 'init',
                           'iu': 'isUnique',
                           'l': 'label',
                           'mbv': 'menuBarVisible',
                           'mrl': 'menuBarRepeatLast',
                           'ni': 'needsInit',
                           'p': 'parent',
                           'pmp': 'popupMenuProcedure',
                           'q': 'query',
                           'rp': 'replacePanel',
                           'to': 'tearOff',
                           'toc': 'tearOffCopy',
                           'tor': 'tearOffRestore',
                           'up': 'unParent',
                           'ut': 'useTemplate'},
            'type': 'windows'},
  'panelConfiguration': {'flags': {'addPanel': {'args': ['<type __builtin__.bool>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'addPanel',
                                                'modes': ['create',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 5,
                                                'shortname': 'ap'},
                                   'configString': {'args': '<type __builtin__.str>',
                                                    'longname': 'configString',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'cfs'},
                                   'createStrings': {'args': '<type __builtin__.bool>',
                                                     'longname': 'createStrings',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'cs'},
                                   'defaultImage': {'args': '<type __builtin__.str>',
                                                    'longname': 'defaultImage',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'di'},
                                   'defineTemplate': {'args': '<type __builtin__.str>',
                                                      'longname': 'defineTemplate',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'dt'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'editStrings': {'args': '<type __builtin__.bool>',
                                                   'longname': 'editStrings',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'es'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'image': {'args': '<type __builtin__.str>',
                                             'longname': 'image',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i'},
                                   'isFixedState': {'args': '<type __builtin__.bool>',
                                                    'longname': 'isFixedState',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'isFixedState'},
                                   'label': {'args': '<type __builtin__.str>',
                                             'longname': 'label',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'l'},
                                   'labelStrings': {'args': '<type __builtin__.bool>',
                                                    'longname': 'labelStrings',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'ls'},
                                   'numberOfPanels': {'args': '<type __builtin__.bool>',
                                                      'longname': 'numberOfPanels',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'np'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'removeAllPanels': {'args': '<type __builtin__.bool>',
                                                       'longname': 'removeAllPanels',
                                                       'modes': ['edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'rap'},
                                   'removeLastPanel': {'args': '<type __builtin__.bool>',
                                                       'longname': 'removeLastPanel',
                                                       'modes': ['edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'rlp'},
                                   'replaceCreateString': {'args': ['<type __builtin__.int>',
                                                                    '<type __builtin__.str>'],
                                                           'longname': 'replaceCreateString',
                                                           'modes': ['edit'],
                                                           'numArgs': 2,
                                                           'shortname': 'rcs'},
                                   'replaceEditString': {'args': ['<type __builtin__.int>',
                                                                  '<type __builtin__.str>'],
                                                         'longname': 'replaceEditString',
                                                         'modes': ['edit'],
                                                         'numArgs': 2,
                                                         'shortname': 'res'},
                                   'replaceFixedState': {'args': ['<type __builtin__.int>',
                                                                  '<type __builtin__.bool>'],
                                                         'longname': 'replaceFixedState',
                                                         'modes': ['edit'],
                                                         'numArgs': 2,
                                                         'shortname': 'rfs'},
                                   'replaceLabel': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'replaceLabel',
                                                    'modes': ['edit'],
                                                    'numArgs': 2,
                                                    'shortname': 'rl'},
                                   'replacePanel': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.bool>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'replacePanel',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 6,
                                                    'shortname': 'rp'},
                                   'replaceTypeString': {'args': ['<type __builtin__.int>',
                                                                  '<type __builtin__.str>'],
                                                         'longname': 'replaceTypeString',
                                                         'modes': ['edit'],
                                                         'numArgs': 2,
                                                         'shortname': 'rts'},
                                   'sceneConfig': {'args': '<type __builtin__.bool>',
                                                   'longname': 'sceneConfig',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sc'},
                                   'typeStrings': {'args': '<type __builtin__.bool>',
                                                   'longname': 'typeStrings',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'ts'},
                                   'useTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'useTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ut'},
                                   'userCreated': {'args': '<type __builtin__.bool>',
                                                   'longname': 'userCreated',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'uc'}},
                         'removedFlags': {'if': 'isFixedState'},
                         'shortFlags': {'ap': 'addPanel',
                                        'cfs': 'configString',
                                        'cs': 'createStrings',
                                        'di': 'defaultImage',
                                        'dt': 'defineTemplate',
                                        'e': 'edit',
                                        'es': 'editStrings',
                                        'ex': 'exists',
                                        'i': 'image',
                                        'isFixedState': 'isFixedState',
                                        'l': 'label',
                                        'ls': 'labelStrings',
                                        'np': 'numberOfPanels',
                                        'q': 'query',
                                        'rap': 'removeAllPanels',
                                        'rcs': 'replaceCreateString',
                                        'res': 'replaceEditString',
                                        'rfs': 'replaceFixedState',
                                        'rl': 'replaceLabel',
                                        'rlp': 'removeLastPanel',
                                        'rp': 'replacePanel',
                                        'rts': 'replaceTypeString',
                                        'sc': 'sceneConfig',
                                        'ts': 'typeStrings',
                                        'uc': 'userCreated',
                                        'ut': 'useTemplate'},
                         'type': 'windows'},
  'panelHistory': {'flags': {'back': {'args': '<type __builtin__.bool>',
                                      'longname': 'back',
                                      'modes': ['edit'],
                                      'numArgs': 0,
                                      'shortname': 'b'},
                             'clear': {'args': '<type __builtin__.bool>',
                                       'longname': 'clear',
                                       'modes': ['edit'],
                                       'numArgs': 0,
                                       'shortname': 'cl'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'forward': {'args': '<type __builtin__.bool>',
                                         'longname': 'forward',
                                         'modes': ['edit'],
                                         'numArgs': 0,
                                         'shortname': 'f'},
                             'historyDepth': {'args': '<type __builtin__.int>',
                                              'longname': 'historyDepth',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'hd'},
                             'isEmpty': {'args': '<type __builtin__.bool>',
                                         'longname': 'isEmpty',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'ie'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'suspend': {'args': '<type __builtin__.bool>',
                                         'longname': 'suspend',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 's'},
                             'targetPane': {'args': '<type __builtin__.str>',
                                            'longname': 'targetPane',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'tp'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'wrap': {'args': '<type __builtin__.bool>',
                                      'longname': 'wrap',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                   'shortFlags': {'b': 'back',
                                  'cl': 'clear',
                                  'dt': 'defineTemplate',
                                  'e': 'edit',
                                  'ex': 'exists',
                                  'f': 'forward',
                                  'hd': 'historyDepth',
                                  'ie': 'isEmpty',
                                  'q': 'query',
                                  's': 'suspend',
                                  'tp': 'targetPane',
                                  'ut': 'useTemplate',
                                  'w': 'wrap'},
                   'type': 'windows'},
  'paramDimContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'history': {'args': '<type __builtin__.bool>',
                                            'longname': 'history',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ch'},
                                'image1': {'args': '<type __builtin__.str>',
                                           'longname': 'image1',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i1'},
                                'image2': {'args': '<type __builtin__.str>',
                                           'longname': 'image2',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i2'},
                                'image3': {'args': '<type __builtin__.str>',
                                           'longname': 'image3',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i3'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'}},
                      'shortFlags': {'ch': 'history',
                                     'e': 'edit',
                                     'ex': 'exists',
                                     'i1': 'image1',
                                     'i2': 'image2',
                                     'i3': 'image3',
                                     'n': 'name',
                                     'q': 'query'},
                      'type': 'context'},
  'paramDimension': {'flags': {}, 'shortFlags': {}, 'type': 'general'},
  'paramLocator': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'position': {'args': '<type __builtin__.bool>',
                                          'longname': 'position',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'p'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'e': 'edit',
                                  'p': 'position',
                                  'q': 'query'},
                   'type': 'general'},
  'parent': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                    'longname': 'absolute',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'a'},
                       'addObject': {'args': '<type __builtin__.bool>',
                                     'longname': 'addObject',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'add'},
                       'noConnections': {'args': '<type __builtin__.bool>',
                                         'longname': 'noConnections',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'nc'},
                       'noInvScale': {'args': '<type __builtin__.bool>',
                                      'longname': 'noInvScale',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'nis'},
                       'relative': {'args': '<type __builtin__.bool>',
                                    'longname': 'relative',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'r'},
                       'removeObject': {'args': '<type __builtin__.bool>',
                                        'longname': 'removeObject',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'rm'},
                       'shape': {'args': '<type __builtin__.bool>',
                                 'longname': 'shape',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 's'},
                       'world': {'args': '<type __builtin__.bool>',
                                 'longname': 'world',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'w'}},
             'shortFlags': {'a': 'absolute',
                            'add': 'addObject',
                            'nc': 'noConnections',
                            'nis': 'noInvScale',
                            'r': 'relative',
                            'rm': 'removeObject',
                            's': 'shape',
                            'w': 'world'},
             'type': 'general'},
  'parentConstraint': {'flags': {'createCache': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'createCache',
                                                 'modes': ['edit'],
                                                 'numArgs': 2,
                                                 'shortname': 'cc'},
                                 'decompRotationToChild': {'args': '<type __builtin__.bool>',
                                                           'longname': 'decompRotationToChild',
                                                           'modes': ['create'],
                                                           'numArgs': 0,
                                                           'shortname': 'dr'},
                                 'deleteCache': {'args': '<type __builtin__.bool>',
                                                 'longname': 'deleteCache',
                                                 'modes': ['edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'dc'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'layer': {'args': '<type __builtin__.str>',
                                           'longname': 'layer',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'l'},
                                 'maintainOffset': {'args': '<type __builtin__.bool>',
                                                    'longname': 'maintainOffset',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'mo'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'remove': {'args': '<type __builtin__.bool>',
                                            'longname': 'remove',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'rm'},
                                 'skipRotate': {'args': '<type __builtin__.str>',
                                                'longname': 'skipRotate',
                                                'modes': ['create',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'sr'},
                                 'skipTranslate': {'args': '<type __builtin__.str>',
                                                   'longname': 'skipTranslate',
                                                   'modes': ['create',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'st'},
                                 'targetList': {'args': '<type __builtin__.bool>',
                                                'longname': 'targetList',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'tl'},
                                 'weight': {'args': '<type __builtin__.float>',
                                            'longname': 'weight',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'w'},
                                 'weightAliasList': {'args': '<type __builtin__.bool>',
                                                     'longname': 'weightAliasList',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'wal'}},
                       'resultNeedsUnpacking': True,
                       'shortFlags': {'cc': 'createCache',
                                      'dc': 'deleteCache',
                                      'dr': 'decompRotationToChild',
                                      'e': 'edit',
                                      'l': 'layer',
                                      'mo': 'maintainOffset',
                                      'n': 'name',
                                      'q': 'query',
                                      'rm': 'remove',
                                      'sr': 'skipRotate',
                                      'st': 'skipTranslate',
                                      'tl': 'targetList',
                                      'w': 'weight',
                                      'wal': 'weightAliasList'},
                       'type': 'animation'},
  'particle': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                       'longname': 'attribute',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'at'},
                         'cache': {'args': '<type __builtin__.bool>',
                                   'longname': 'cache',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'ch'},
                         'conserve': {'args': '<type __builtin__.float>',
                                      'longname': 'conserve',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'c'},
                         'count': {'args': '<type __builtin__.bool>',
                                   'longname': 'count',
                                   'modes': ['query'],
                                   'numArgs': 0,
                                   'shortname': 'ct'},
                         'deleteCache': {'args': '<type __builtin__.bool>',
                                         'longname': 'deleteCache',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'dc'},
                         'dynamicAttrList': {'args': '<type __builtin__.bool>',
                                             'longname': 'dynamicAttrList',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'dal'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'floatValue': {'args': '<type __builtin__.float>',
                                        'longname': 'floatValue',
                                        'modes': ['edit'],
                                        'numArgs': 1,
                                        'shortname': 'fv'},
                         'gridSpacing': {'args': '<type __builtin__.float>',
                                         'longname': 'gridSpacing',
                                         'modes': ['create',
                                                   'query',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'grs'},
                         'inherit': {'args': '<type __builtin__.float>',
                                     'longname': 'inherit',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'i'},
                         'jitterBasePoint': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'jitterBasePoint',
                                             'modes': ['create',
                                                       'query',
                                                       'multiuse'],
                                             'numArgs': 3,
                                             'shortname': 'jbp'},
                         'jitterRadius': {'args': '<type __builtin__.float>',
                                          'longname': 'jitterRadius',
                                          'modes': ['create',
                                                    'query',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'jr'},
                         'lowerLeft': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'lowerLeft',
                                       'modes': ['create',
                                                 'query',
                                                 'multiuse'],
                                       'numArgs': 3,
                                       'shortname': 'll'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'numJitters': {'args': '<type __builtin__.int>',
                                        'longname': 'numJitters',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'nj'},
                         'order': {'args': '<type __builtin__.int>',
                                   'longname': 'order',
                                   'modes': ['query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'order'},
                         'particleId': {'args': '<type __builtin__.int>',
                                        'longname': 'particleId',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'id'},
                         'perParticleDouble': {'args': '<type __builtin__.bool>',
                                               'longname': 'perParticleDouble',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'ppd'},
                         'perParticleVector': {'args': '<type __builtin__.bool>',
                                               'longname': 'perParticleVector',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'ppv'},
                         'position': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'position',
                                      'modes': ['multiuse'],
                                      'numArgs': 3,
                                      'shortname': 'p'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'shapeName': {'args': '<type __builtin__.str>',
                                       'longname': 'shapeName',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sn'},
                         'upperRight': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'upperRight',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 3,
                                        'shortname': 'ur'},
                         'vectorValue': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'vectorValue',
                                         'modes': ['edit'],
                                         'numArgs': 3,
                                         'shortname': 'vv'}},
               'removedFlags': {'or': 'order'},
               'shortFlags': {'at': 'attribute',
                              'c': 'conserve',
                              'ch': 'cache',
                              'ct': 'count',
                              'dal': 'dynamicAttrList',
                              'dc': 'deleteCache',
                              'e': 'edit',
                              'fv': 'floatValue',
                              'grs': 'gridSpacing',
                              'i': 'inherit',
                              'id': 'particleId',
                              'jbp': 'jitterBasePoint',
                              'jr': 'jitterRadius',
                              'll': 'lowerLeft',
                              'n': 'name',
                              'nj': 'numJitters',
                              'order': 'order',
                              'p': 'position',
                              'ppd': 'perParticleDouble',
                              'ppv': 'perParticleVector',
                              'q': 'query',
                              'sn': 'shapeName',
                              'ur': 'upperRight',
                              'vv': 'vectorValue'},
               'type': 'effects'},
  'particleExists': {'flags': {}, 'shortFlags': {}, 'type': 'effects'},
  'particleFill': {'flags': {'closePacking': {'args': '<type __builtin__.bool>',
                                              'longname': 'closePacking',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'cp'},
                             'doubleWalled': {'args': '<type __builtin__.bool>',
                                              'longname': 'doubleWalled',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'dw'},
                             'maxX': {'args': '<type __builtin__.float>',
                                      'longname': 'maxX',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'mxx'},
                             'maxY': {'args': '<type __builtin__.float>',
                                      'longname': 'maxY',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'mxy'},
                             'maxZ': {'args': '<type __builtin__.float>',
                                      'longname': 'maxZ',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'mxz'},
                             'minX': {'args': '<type __builtin__.float>',
                                      'longname': 'minX',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'mnx'},
                             'minY': {'args': '<type __builtin__.float>',
                                      'longname': 'minY',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'mny'},
                             'minZ': {'args': '<type __builtin__.float>',
                                      'longname': 'minZ',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'mnz'},
                             'particleDensity': {'args': '<type __builtin__.float>',
                                                 'longname': 'particleDensity',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'pd'},
                             'resolution': {'args': '<type __builtin__.int>',
                                            'longname': 'resolution',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'rs'}},
                   'shortFlags': {'cp': 'closePacking',
                                  'dw': 'doubleWalled',
                                  'mnx': 'minX',
                                  'mny': 'minY',
                                  'mnz': 'minZ',
                                  'mxx': 'maxX',
                                  'mxy': 'maxY',
                                  'mxz': 'maxZ',
                                  'pd': 'particleDensity',
                                  'rs': 'resolution'},
                   'type': 'effects'},
  'particleInstancer': {'flags': {'addObject': {'args': '<type __builtin__.bool>',
                                                'longname': 'addObject',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'a'},
                                  'aimAxis': {'args': '<type __builtin__.str>',
                                              'longname': 'aimAxis',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'aa'},
                                  'aimDirection': {'args': '<type __builtin__.str>',
                                                   'longname': 'aimDirection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad'},
                                  'aimPosition': {'args': '<type __builtin__.str>',
                                                  'longname': 'aimPosition',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ap'},
                                  'aimUpAxis': {'args': '<type __builtin__.str>',
                                                'longname': 'aimUpAxis',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'aua'},
                                  'aimWorldUp': {'args': '<type __builtin__.str>',
                                                 'longname': 'aimWorldUp',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'awu'},
                                  'attributeMapping': {'args': '<type __builtin__.bool>',
                                                       'longname': 'attributeMapping',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'am'},
                                  'cycle': {'args': '<type __builtin__.str>',
                                            'longname': 'cycle',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'c'},
                                  'cycleStartObject': {'args': '<type __builtin__.str>',
                                                       'longname': 'cycleStartObject',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'sto'},
                                  'cycleStep': {'args': '<type __builtin__.float>',
                                                'longname': 'cycleStep',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cs'},
                                  'cycleStepUnits': {'args': '<type __builtin__.str>',
                                                     'longname': 'cycleStepUnits',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'csu'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'index': {'args': '<type __builtin__.int>',
                                            'longname': 'index',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'i'},
                                  'instanceId': {'args': '<type __builtin__.str>',
                                                 'longname': 'instanceId',
                                                 'modes': ['query'],
                                                 'numArgs': 1,
                                                 'shortname': 'id'},
                                  'levelOfDetail': {'args': '<type __builtin__.str>',
                                                    'longname': 'levelOfDetail',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'lod'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'object': {'args': '<type __builtin__.str>',
                                             'longname': 'object',
                                             'modes': ['create',
                                                       'query',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'obj'},
                                  'objectIndex': {'args': '<type __builtin__.str>',
                                                  'longname': 'objectIndex',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'oi'},
                                  'particleAge': {'args': '<type __builtin__.str>',
                                                  'longname': 'particleAge',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'age'},
                                  'position': {'args': '<type __builtin__.str>',
                                               'longname': 'position',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'p'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'removeObject': {'args': '<type __builtin__.bool>',
                                                   'longname': 'removeObject',
                                                   'modes': ['edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'rm'},
                                  'rotation': {'args': '<type __builtin__.str>',
                                               'longname': 'rotation',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'r'},
                                  'rotationOrder': {'args': '<type __builtin__.str>',
                                                    'longname': 'rotationOrder',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ro'},
                                  'rotationType': {'args': '<type __builtin__.str>',
                                                   'longname': 'rotationType',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'rt'},
                                  'rotationUnits': {'args': '<type __builtin__.str>',
                                                    'longname': 'rotationUnits',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ru'},
                                  'scale': {'args': '<type __builtin__.str>',
                                            'longname': 'scale',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sc'},
                                  'shear': {'args': '<type __builtin__.str>',
                                            'longname': 'shear',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sh'},
                                  'visibility': {'args': '<type __builtin__.str>',
                                                 'longname': 'visibility',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'vis'}},
                        'shortFlags': {'a': 'addObject',
                                       'aa': 'aimAxis',
                                       'ad': 'aimDirection',
                                       'age': 'particleAge',
                                       'am': 'attributeMapping',
                                       'ap': 'aimPosition',
                                       'aua': 'aimUpAxis',
                                       'awu': 'aimWorldUp',
                                       'c': 'cycle',
                                       'cs': 'cycleStep',
                                       'csu': 'cycleStepUnits',
                                       'e': 'edit',
                                       'i': 'index',
                                       'id': 'instanceId',
                                       'lod': 'levelOfDetail',
                                       'n': 'name',
                                       'obj': 'object',
                                       'oi': 'objectIndex',
                                       'p': 'position',
                                       'q': 'query',
                                       'r': 'rotation',
                                       'rm': 'removeObject',
                                       'ro': 'rotationOrder',
                                       'rt': 'rotationType',
                                       'ru': 'rotationUnits',
                                       'sc': 'scale',
                                       'sh': 'shear',
                                       'sto': 'cycleStartObject',
                                       'vis': 'visibility'},
                        'type': 'effects'},
  'particleRenderInfo': {'flags': {'attrList': {'args': '<type __builtin__.int>',
                                                'longname': 'attrList',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'al'},
                                   'attrListAll': {'args': '<type __builtin__.bool>',
                                                   'longname': 'attrListAll',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'ala'},
                                   'name': {'args': '<type __builtin__.int>',
                                            'longname': 'name',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'renderTypeCount': {'args': '<type __builtin__.bool>',
                                                       'longname': 'renderTypeCount',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'rtc'}},
                         'shortFlags': {'al': 'attrList',
                                        'ala': 'attrListAll',
                                        'n': 'name',
                                        'q': 'query',
                                        'rtc': 'renderTypeCount'},
                         'type': 'effects'},
  'partition': {'flags': {'addSet': {'args': 'PyNode',
                                     'longname': 'addSet',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'add'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'removeSet': {'args': 'PyNode',
                                        'longname': 'removeSet',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'rm'},
                          'render': {'args': '<type __builtin__.bool>',
                                     'longname': 'render',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 're'}},
                'shortFlags': {'add': 'addSet',
                               'e': 'edit',
                               'n': 'name',
                               'q': 'query',
                               're': 'render',
                               'rm': 'removeSet'},
                'type': 'general'},
  'pasteKey': {'flags': {'animLayer': {'args': '<type __builtin__.str>',
                                       'longname': 'animLayer',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'al'},
                         'animation': {'args': '<type __builtin__.str>',
                                       'longname': 'animation',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'an'},
                         'attribute': {'args': '<type __builtin__.str>',
                                       'longname': 'attribute',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'at'},
                         'clipboard': {'args': '<type __builtin__.str>',
                                       'longname': 'clipboard',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'cb'},
                         'connect': {'args': '<type __builtin__.bool>',
                                     'longname': 'connect',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'c'},
                         'copies': {'args': '<type __builtin__.int>',
                                    'longname': 'copies',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'cp'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'float': {'args': 'floatrange',
                                   'longname': 'float',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'f'},
                         'floatOffset': {'args': '<type __builtin__.float>',
                                         'longname': 'floatOffset',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'fo'},
                         'includeUpperBound': {'args': '<type __builtin__.bool>',
                                               'longname': 'includeUpperBound',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'iub'},
                         'index': {'args': '<type __builtin__.int>',
                                   'longname': 'index',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'index'},
                         'matchByName': {'args': '<type __builtin__.bool>',
                                         'longname': 'matchByName',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'mn'},
                         'option': {'args': '<type __builtin__.str>',
                                    'longname': 'option',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'o'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'time': {'args': 'timerange',
                                  'longname': 'time',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 't'},
                         'timeOffset': {'args': 'time',
                                        'longname': 'timeOffset',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'to'},
                         'valueOffset': {'args': '<type __builtin__.float>',
                                         'longname': 'valueOffset',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'vo'}},
               'removedFlags': {'in': 'index'},
               'shortFlags': {'al': 'animLayer',
                              'an': 'animation',
                              'at': 'attribute',
                              'c': 'connect',
                              'cb': 'clipboard',
                              'cp': 'copies',
                              'e': 'edit',
                              'f': 'float',
                              'fo': 'floatOffset',
                              'index': 'index',
                              'iub': 'includeUpperBound',
                              'mn': 'matchByName',
                              'o': 'option',
                              'q': 'query',
                              't': 'time',
                              'to': 'timeOffset',
                              'vo': 'valueOffset'},
               'type': 'animation'},
  'pathAnimation': {'flags': {'bank': {'args': '<type __builtin__.bool>',
                                       'longname': 'bank',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'b'},
                              'bankScale': {'args': '<type __builtin__.float>',
                                            'longname': 'bankScale',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'bs'},
                              'bankThreshold': {'args': '<type __builtin__.float>',
                                                'longname': 'bankThreshold',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'bt'},
                              'curve': {'args': '<type __builtin__.str>',
                                        'longname': 'curve',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'c'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'endTimeU': {'args': 'time',
                                           'longname': 'endTimeU',
                                           'modes': ['query', 'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'etu'},
                              'endU': {'args': '<type __builtin__.float>',
                                       'longname': 'endU',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'eu'},
                              'follow': {'args': '<type __builtin__.bool>',
                                         'longname': 'follow',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'f'},
                              'followAxis': {'args': '<type __builtin__.str>',
                                             'longname': 'followAxis',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'fa'},
                              'fractionMode': {'args': '<type __builtin__.bool>',
                                               'longname': 'fractionMode',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'fm'},
                              'inverseFront': {'args': '<type __builtin__.bool>',
                                               'longname': 'inverseFront',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'inverseFront'},
                              'inverseUp': {'args': '<type __builtin__.bool>',
                                            'longname': 'inverseUp',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'iu'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['query'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'startTimeU': {'args': 'time',
                                             'longname': 'startTimeU',
                                             'modes': ['query',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'stu'},
                              'startU': {'args': '<type __builtin__.float>',
                                         'longname': 'startU',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'su'},
                              'upAxis': {'args': '<type __builtin__.str>',
                                         'longname': 'upAxis',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'ua'},
                              'useNormal': {'args': '<type __builtin__.bool>',
                                            'longname': 'useNormal',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'un'},
                              'worldUpObject': {'args': 'PyNode',
                                                'longname': 'worldUpObject',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'wuo'},
                              'worldUpType': {'args': '<type __builtin__.str>',
                                              'longname': 'worldUpType',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'wut'},
                              'worldUpVector': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'worldUpVector',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'wu'}},
                    'removedFlags': {'if': 'inverseFront'},
                    'shortFlags': {'b': 'bank',
                                   'bs': 'bankScale',
                                   'bt': 'bankThreshold',
                                   'c': 'curve',
                                   'e': 'edit',
                                   'etu': 'endTimeU',
                                   'eu': 'endU',
                                   'f': 'follow',
                                   'fa': 'followAxis',
                                   'fm': 'fractionMode',
                                   'inverseFront': 'inverseFront',
                                   'iu': 'inverseUp',
                                   'n': 'name',
                                   'q': 'query',
                                   'stu': 'startTimeU',
                                   'su': 'startU',
                                   'ua': 'upAxis',
                                   'un': 'useNormal',
                                   'wu': 'worldUpVector',
                                   'wuo': 'worldUpObject',
                                   'wut': 'worldUpType'},
                    'type': 'animation'},
  'pause': {'flags': {'seconds': {'args': '<type __builtin__.int>',
                                  'longname': 'seconds',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'sec'}},
            'shortFlags': {'sec': 'seconds'},
            'type': 'language'},
  'perCameraVisibility': {'flags': {'camera': {'args': 'PyNode',
                                               'longname': 'camera',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'c'},
                                    'exclusive': {'args': '<type __builtin__.bool>',
                                                  'longname': 'exclusive',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 0,
                                                  'shortname': 'ex'},
                                    'hide': {'args': '<type __builtin__.bool>',
                                             'longname': 'hide',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'hi'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'remove': {'args': '<type __builtin__.bool>',
                                               'longname': 'remove',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'rm'},
                                    'removeAll': {'args': '<type __builtin__.bool>',
                                                  'longname': 'removeAll',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'ra'},
                                    'removeCamera': {'args': '<type __builtin__.bool>',
                                                     'longname': 'removeCamera',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'rc'}},
                          'shortFlags': {'c': 'camera',
                                         'ex': 'exclusive',
                                         'hi': 'hide',
                                         'q': 'query',
                                         'ra': 'removeAll',
                                         'rc': 'removeCamera',
                                         'rm': 'remove'},
                          'type': 'rendering'},
  'percent': {'flags': {'addPercent': {'args': '<type __builtin__.bool>',
                                       'longname': 'addPercent',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ap'},
                        'dropoffAxis': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'dropoffAxis',
                                        'modes': ['create'],
                                        'numArgs': 3,
                                        'shortname': 'dax'},
                        'dropoffCurve': {'args': '<type __builtin__.str>',
                                         'longname': 'dropoffCurve',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'dc'},
                        'dropoffDistance': {'args': '<type __builtin__.float>',
                                            'longname': 'dropoffDistance',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'dds'},
                        'dropoffPosition': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'dropoffPosition',
                                            'modes': ['create'],
                                            'numArgs': 3,
                                            'shortname': 'dp'},
                        'dropoffType': {'args': '<type __builtin__.str>',
                                        'longname': 'dropoffType',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'dt'},
                        'multiplyPercent': {'args': '<type __builtin__.bool>',
                                            'longname': 'multiplyPercent',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'mp'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'value': {'args': '<type __builtin__.float>',
                                  'longname': 'value',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'v'}},
              'shortFlags': {'ap': 'addPercent',
                             'dax': 'dropoffAxis',
                             'dc': 'dropoffCurve',
                             'dds': 'dropoffDistance',
                             'dp': 'dropoffPosition',
                             'dt': 'dropoffType',
                             'mp': 'multiplyPercent',
                             'q': 'query',
                             'v': 'value'},
              'type': 'animation'},
  'performanceOptions': {'flags': {'clusterResolution': {'args': '<type __builtin__.float>',
                                                         'longname': 'clusterResolution',
                                                         'modes': ['query'],
                                                         'numArgs': 1,
                                                         'shortname': 'cr'},
                                   'disableStitch': {'args': '<type __builtin__.str>',
                                                     'longname': 'disableStitch',
                                                     'modes': ['query'],
                                                     'numArgs': 1,
                                                     'shortname': 'ds'},
                                   'disableTrimBoundaryDisplay': {'args': '<type __builtin__.str>',
                                                                  'longname': 'disableTrimBoundaryDisplay',
                                                                  'modes': ['query'],
                                                                  'numArgs': 1,
                                                                  'shortname': 'dtb'},
                                   'disableTrimDisplay': {'args': '<type __builtin__.str>',
                                                          'longname': 'disableTrimDisplay',
                                                          'modes': ['query'],
                                                          'numArgs': 1,
                                                          'shortname': 'dt'},
                                   'latticeResolution': {'args': '<type __builtin__.float>',
                                                         'longname': 'latticeResolution',
                                                         'modes': ['query'],
                                                         'numArgs': 1,
                                                         'shortname': 'lr'},
                                   'passThroughBindSkinAndFlexors': {'args': '<type __builtin__.str>',
                                                                     'longname': 'passThroughBindSkinAndFlexors',
                                                                     'modes': ['query'],
                                                                     'numArgs': 1,
                                                                     'shortname': 'pbf'},
                                   'passThroughBlendShape': {'args': '<type __builtin__.str>',
                                                             'longname': 'passThroughBlendShape',
                                                             'modes': ['query'],
                                                             'numArgs': 1,
                                                             'shortname': 'pbs'},
                                   'passThroughCluster': {'args': '<type __builtin__.str>',
                                                          'longname': 'passThroughCluster',
                                                          'modes': ['query'],
                                                          'numArgs': 1,
                                                          'shortname': 'pc'},
                                   'passThroughDeltaMush': {'args': '<type __builtin__.str>',
                                                            'longname': 'passThroughDeltaMush',
                                                            'modes': ['query'],
                                                            'numArgs': 1,
                                                            'shortname': 'pdm'},
                                   'passThroughFlexors': {'args': '<type __builtin__.str>',
                                                          'longname': 'passThroughFlexors',
                                                          'modes': ['query'],
                                                          'numArgs': 1,
                                                          'shortname': 'pf'},
                                   'passThroughLattice': {'args': '<type __builtin__.str>',
                                                          'longname': 'passThroughLattice',
                                                          'modes': ['query'],
                                                          'numArgs': 1,
                                                          'shortname': 'pl'},
                                   'passThroughPaintEffects': {'args': '<type __builtin__.str>',
                                                               'longname': 'passThroughPaintEffects',
                                                               'modes': ['query'],
                                                               'numArgs': 1,
                                                               'shortname': 'pp'},
                                   'passThroughSculpt': {'args': '<type __builtin__.str>',
                                                         'longname': 'passThroughSculpt',
                                                         'modes': ['query'],
                                                         'numArgs': 1,
                                                         'shortname': 'ps'},
                                   'passThroughWire': {'args': '<type __builtin__.str>',
                                                       'longname': 'passThroughWire',
                                                       'modes': ['query'],
                                                       'numArgs': 1,
                                                       'shortname': 'pw'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'skipHierarchyTraversal': {'args': '<type __builtin__.bool>',
                                                              'longname': 'skipHierarchyTraversal',
                                                              'modes': ['query'],
                                                              'numArgs': 1,
                                                              'shortname': 'sht'},
                                   'useClusterResolution': {'args': '<type __builtin__.str>',
                                                            'longname': 'useClusterResolution',
                                                            'modes': ['query'],
                                                            'numArgs': 1,
                                                            'shortname': 'ucr'},
                                   'useLatticeResolution': {'args': '<type __builtin__.str>',
                                                            'longname': 'useLatticeResolution',
                                                            'modes': ['query'],
                                                            'numArgs': 1,
                                                            'shortname': 'ulr'}},
                         'shortFlags': {'cr': 'clusterResolution',
                                        'ds': 'disableStitch',
                                        'dt': 'disableTrimDisplay',
                                        'dtb': 'disableTrimBoundaryDisplay',
                                        'lr': 'latticeResolution',
                                        'pbf': 'passThroughBindSkinAndFlexors',
                                        'pbs': 'passThroughBlendShape',
                                        'pc': 'passThroughCluster',
                                        'pdm': 'passThroughDeltaMush',
                                        'pf': 'passThroughFlexors',
                                        'pl': 'passThroughLattice',
                                        'pp': 'passThroughPaintEffects',
                                        'ps': 'passThroughSculpt',
                                        'pw': 'passThroughWire',
                                        'q': 'query',
                                        'sht': 'skipHierarchyTraversal',
                                        'ucr': 'useClusterResolution',
                                        'ulr': 'useLatticeResolution'},
                         'type': 'general'},
  'pfxstrokes': {'flags': {'filename': {'args': '<type __builtin__.str>',
                                        'longname': 'filename',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'fn'},
                           'postCallback': {'args': '<type __builtin__.bool>',
                                            'longname': 'postCallback',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'pc'},
                           'selected': {'args': '<type __builtin__.bool>',
                                        'longname': 'selected',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'sl'}},
                 'shortFlags': {'fn': 'filename',
                                'pc': 'postCallback',
                                'sl': 'selected'},
                 'type': 'effects'},
  'pickWalk': {'flags': {'direction': {'args': '<type __builtin__.str>',
                                       'longname': 'direction',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'd'},
                         'recurse': {'args': '<type __builtin__.bool>',
                                     'longname': 'recurse',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'r'},
                         'type': {'args': '<type __builtin__.str>',
                                  'longname': 'type',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'typ'}},
               'shortFlags': {'d': 'direction',
                              'r': 'recurse',
                              'typ': 'type'},
               'type': 'general'},
  'picture': {'callbackFlags': ['visibleChangeCommand',
                                'vcc',
                                'dragCallback',
                                'dgc',
                                'dropCallback',
                                'dpc'],
              'flags': {'annotation': {'args': '<type __builtin__.str>',
                                       'longname': 'annotation',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ann'},
                        'backgroundColor': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'backgroundColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'bgc'},
                        'defineTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'defineTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'dt'},
                        'docTag': {'args': '<type __builtin__.str>',
                                   'longname': 'docTag',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'dtg'},
                        'dragCallback': {'args': 'script',
                                         'longname': 'dragCallback',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dgc'},
                        'dropCallback': {'args': 'script',
                                         'longname': 'dropCallback',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dpc'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'enable': {'args': '<type __builtin__.bool>',
                                   'longname': 'enable',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'en'},
                        'enableBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'enableBackground',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ebg'},
                        'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableKeyboardFocus',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ekf'},
                        'exists': {'args': '<type __builtin__.bool>',
                                   'longname': 'exists',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'ex'},
                        'fullPathName': {'args': '<type __builtin__.bool>',
                                         'longname': 'fullPathName',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'fpn'},
                        'height': {'args': '<type __builtin__.int>',
                                   'longname': 'height',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'h'},
                        'highlightColor': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'highlightColor',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'hlc'},
                        'image': {'args': '<type __builtin__.str>',
                                  'longname': 'image',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'i'},
                        'isObscured': {'args': '<type __builtin__.bool>',
                                       'longname': 'isObscured',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'io'},
                        'manage': {'args': '<type __builtin__.bool>',
                                   'longname': 'manage',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'm'},
                        'noBackground': {'args': '<type __builtin__.bool>',
                                         'longname': 'noBackground',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nbg'},
                        'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                               'longname': 'numberOfPopupMenus',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'npm'},
                        'parent': {'args': '<type __builtin__.str>',
                                   'longname': 'parent',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'p'},
                        'popupMenuArray': {'args': '<type __builtin__.bool>',
                                           'longname': 'popupMenuArray',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'pma'},
                        'preventOverride': {'args': '<type __builtin__.bool>',
                                            'longname': 'preventOverride',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'po'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'statusBarMessage': {'args': '<type __builtin__.str>',
                                             'longname': 'statusBarMessage',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sbm'},
                        'tile': {'args': '<type __builtin__.bool>',
                                 'longname': 'tile',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'tl'},
                        'useTemplate': {'args': '<type __builtin__.str>',
                                        'longname': 'useTemplate',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ut'},
                        'visible': {'args': '<type __builtin__.bool>',
                                    'longname': 'visible',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'vis'},
                        'visibleChangeCommand': {'args': 'script',
                                                 'longname': 'visibleChangeCommand',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'vcc'},
                        'width': {'args': '<type __builtin__.int>',
                                  'longname': 'width',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'w'}},
              'shortFlags': {'ann': 'annotation',
                             'bgc': 'backgroundColor',
                             'dgc': 'dragCallback',
                             'dpc': 'dropCallback',
                             'dt': 'defineTemplate',
                             'dtg': 'docTag',
                             'e': 'edit',
                             'ebg': 'enableBackground',
                             'ekf': 'enableKeyboardFocus',
                             'en': 'enable',
                             'ex': 'exists',
                             'fpn': 'fullPathName',
                             'h': 'height',
                             'hlc': 'highlightColor',
                             'i': 'image',
                             'io': 'isObscured',
                             'm': 'manage',
                             'nbg': 'noBackground',
                             'npm': 'numberOfPopupMenus',
                             'p': 'parent',
                             'pma': 'popupMenuArray',
                             'po': 'preventOverride',
                             'q': 'query',
                             'sbm': 'statusBarMessage',
                             'tl': 'tile',
                             'ut': 'useTemplate',
                             'vcc': 'visibleChangeCommand',
                             'vis': 'visible',
                             'w': 'width'},
              'type': 'windows'},
  'pixelMove': {'flags': {}, 'shortFlags': {}, 'type': 'general'},
  'planarSrf': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                      'longname': 'caching',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'cch'},
                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'constructionHistory',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                          'degree': {'args': '<type __builtin__.int>',
                                     'longname': 'degree',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'd'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'frozen': {'args': '<type __builtin__.bool>',
                                     'longname': 'frozen',
                                     'numArgs': 1,
                                     'shortname': 'fzn'},
                          'keepOutside': {'args': '<type __builtin__.bool>',
                                          'longname': 'keepOutside',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ko'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nodeState': {'args': '<type __builtin__.int>',
                                        'longname': 'nodeState',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'nds'},
                          'object': {'args': '<type __builtin__.bool>',
                                     'longname': 'object',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'o'},
                          'polygon': {'args': '<type __builtin__.int>',
                                      'longname': 'polygon',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'po'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'range': {'args': '<type __builtin__.bool>',
                                    'longname': 'range',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'rn'},
                          'tolerance': {'args': '<type __builtin__.float>',
                                        'longname': 'tolerance',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'tol'}},
                'shortFlags': {'cch': 'caching',
                               'ch': 'constructionHistory',
                               'd': 'degree',
                               'e': 'edit',
                               'fzn': 'frozen',
                               'ko': 'keepOutside',
                               'n': 'name',
                               'nds': 'nodeState',
                               'o': 'object',
                               'po': 'polygon',
                               'q': 'query',
                               'rn': 'range',
                               'tol': 'tolerance'},
                'type': 'modeling'},
  'plane': {'flags': {'length': {'args': '<type __builtin__.float>',
                                 'longname': 'length',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'l'},
                      'name': {'args': '<type __builtin__.str>',
                               'longname': 'name',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'n'},
                      'position': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'position',
                                   'modes': ['create'],
                                   'numArgs': 3,
                                   'shortname': 'p'},
                      'rotation': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'rotation',
                                   'modes': ['create'],
                                   'numArgs': 3,
                                   'shortname': 'r'},
                      'size': {'args': '<type __builtin__.float>',
                               'longname': 'size',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 's'},
                      'width': {'args': '<type __builtin__.float>',
                                'longname': 'width',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'w'}},
            'shortFlags': {'l': 'length',
                           'n': 'name',
                           'p': 'position',
                           'r': 'rotation',
                           's': 'size',
                           'w': 'width'},
            'type': 'modeling'},
  'play': {'flags': {'forward': {'args': '<type __builtin__.bool>',
                                 'longname': 'forward',
                                 'modes': ['create', 'query'],
                                 'numArgs': 1,
                                 'shortname': 'f'},
                     'playSound': {'args': '<type __builtin__.bool>',
                                   'longname': 'playSound',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'ps'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'record': {'args': '<type __builtin__.bool>',
                                'longname': 'record',
                                'modes': ['create', 'query'],
                                'numArgs': 0,
                                'shortname': 'rec'},
                     'sound': {'args': '<type __builtin__.str>',
                               'longname': 'sound',
                               'modes': ['create', 'query'],
                               'numArgs': 1,
                               'shortname': 's'},
                     'state': {'args': '<type __builtin__.bool>',
                               'longname': 'state',
                               'modes': ['create', 'query'],
                               'numArgs': 1,
                               'shortname': 'st'},
                     'wait': {'args': '<type __builtin__.bool>',
                              'longname': 'wait',
                              'modes': ['create'],
                              'numArgs': 0,
                              'shortname': 'w'}},
           'shortFlags': {'f': 'forward',
                          'ps': 'playSound',
                          'q': 'query',
                          'rec': 'record',
                          's': 'sound',
                          'st': 'state',
                          'w': 'wait'},
           'type': 'animation'},
  'playbackOptions': {'flags': {'animationEndTime': {'args': 'time',
                                                     'longname': 'animationEndTime',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'aet'},
                                'animationStartTime': {'args': 'time',
                                                       'longname': 'animationStartTime',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ast'},
                                'blockingAnim': {'args': '<type __builtin__.bool>',
                                                 'longname': 'blockingAnim',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ba'},
                                'by': {'args': '<type __builtin__.float>',
                                       'longname': 'by',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'by'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'framesPerSecond': {'args': '<type __builtin__.bool>',
                                                    'longname': 'framesPerSecond',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 0,
                                                    'shortname': 'fps'},
                                'loop': {'args': '<type __builtin__.str>',
                                         'longname': 'loop',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'l'},
                                'maxPlaybackSpeed': {'args': '<type __builtin__.float>',
                                                     'longname': 'maxPlaybackSpeed',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'mps'},
                                'maxTime': {'args': 'time',
                                            'longname': 'maxTime',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'max'},
                                'minTime': {'args': 'time',
                                            'longname': 'minTime',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'min'},
                                'playbackSpeed': {'args': '<type __builtin__.float>',
                                                  'longname': 'playbackSpeed',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ps'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'view': {'args': '<type __builtin__.str>',
                                         'longname': 'view',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'v'}},
                      'shortFlags': {'aet': 'animationEndTime',
                                     'ast': 'animationStartTime',
                                     'ba': 'blockingAnim',
                                     'by': 'by',
                                     'e': 'edit',
                                     'fps': 'framesPerSecond',
                                     'l': 'loop',
                                     'max': 'maxTime',
                                     'min': 'minTime',
                                     'mps': 'maxPlaybackSpeed',
                                     'ps': 'playbackSpeed',
                                     'q': 'query',
                                     'v': 'view'},
                      'type': 'animation'},
  'playblast': {'flags': {'activeEditor': {'args': '<type __builtin__.bool>',
                                           'longname': 'activeEditor',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ae'},
                          'cameraSetup': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'cameraSetup',
                                          'modes': ['create', 'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'cs'},
                          'clearCache': {'args': '<type __builtin__.bool>',
                                         'longname': 'clearCache',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'cc'},
                          'codecOptions': {'args': '<type __builtin__.bool>',
                                           'longname': 'codecOptions',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'co'},
                          'combineSound': {'args': '<type __builtin__.bool>',
                                           'longname': 'combineSound',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'csd'},
                          'completeFilename': {'args': '<type __builtin__.str>',
                                               'longname': 'completeFilename',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'cf'},
                          'compression': {'args': '<type __builtin__.str>',
                                          'longname': 'compression',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'c'},
                          'editorPanelName': {'args': '<type __builtin__.str>',
                                              'longname': 'editorPanelName',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'epn'},
                          'endTime': {'args': 'time',
                                      'longname': 'endTime',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'et'},
                          'filename': {'args': '<type __builtin__.str>',
                                       'longname': 'filename',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'f'},
                          'forceOverwrite': {'args': '<type __builtin__.bool>',
                                             'longname': 'forceOverwrite',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'fo'},
                          'format': {'args': '<type __builtin__.str>',
                                     'longname': 'format',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'fmt'},
                          'frame': {'args': 'time',
                                    'longname': 'frame',
                                    'modes': ['create', 'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'fr'},
                          'framePadding': {'args': '<type __builtin__.int>',
                                           'longname': 'framePadding',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'fp'},
                          'height': {'args': '<type __builtin__.int>',
                                     'longname': 'height',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'h'},
                          'indexFromZero': {'args': '<type __builtin__.bool>',
                                            'longname': 'indexFromZero',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ifz'},
                          'offScreen': {'args': '<type __builtin__.bool>',
                                        'longname': 'offScreen',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'os'},
                          'offScreenViewportUpdate': {'args': '<type __builtin__.bool>',
                                                      'longname': 'offScreenViewportUpdate',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'osv'},
                          'options': {'args': '<type __builtin__.bool>',
                                      'longname': 'options',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'o'},
                          'percent': {'args': '<type __builtin__.int>',
                                      'longname': 'percent',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                          'quality': {'args': '<type __builtin__.int>',
                                      'longname': 'quality',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'qlt'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'rawFrameNumbers': {'args': '<type __builtin__.bool>',
                                              'longname': 'rawFrameNumbers',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'rfn'},
                          'replaceAudioOnly': {'args': '<type __builtin__.bool>',
                                               'longname': 'replaceAudioOnly',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'rao'},
                          'replaceEndTime': {'args': 'time',
                                             'longname': 'replaceEndTime',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ret'},
                          'replaceFilename': {'args': '<type __builtin__.str>',
                                              'longname': 'replaceFilename',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'rf'},
                          'replaceStartTime': {'args': 'time',
                                               'longname': 'replaceStartTime',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'rst'},
                          'saveDepth': {'args': '<type __builtin__.bool>',
                                        'longname': 'saveDepth',
                                        'numArgs': 0,
                                        'shortname': 'sd'},
                          'sequenceTime': {'args': '<type __builtin__.bool>',
                                           'longname': 'sequenceTime',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'sqt'},
                          'showOrnaments': {'args': '<type __builtin__.bool>',
                                            'longname': 'showOrnaments',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'orn'},
                          'sound': {'args': '<type __builtin__.str>',
                                    'longname': 'sound',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 's'},
                          'startTime': {'args': 'time',
                                        'longname': 'startTime',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'st'},
                          'throwOnError': {'args': '<type __builtin__.bool>',
                                           'longname': 'throwOnError',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'toe'},
                          'useTraxSounds': {'args': '<type __builtin__.bool>',
                                            'longname': 'useTraxSounds',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'uts'},
                          'viewer': {'args': '<type __builtin__.bool>',
                                     'longname': 'viewer',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'v'},
                          'width': {'args': '<type __builtin__.int>',
                                    'longname': 'width',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'w'},
                          'widthHeight': {'args': ['<type __builtin__.int>',
                                                   '<type __builtin__.int>'],
                                          'longname': 'widthHeight',
                                          'modes': ['create'],
                                          'numArgs': 2,
                                          'shortname': 'wh'}},
                'shortFlags': {'ae': 'activeEditor',
                               'c': 'compression',
                               'cc': 'clearCache',
                               'cf': 'completeFilename',
                               'co': 'codecOptions',
                               'cs': 'cameraSetup',
                               'csd': 'combineSound',
                               'epn': 'editorPanelName',
                               'et': 'endTime',
                               'f': 'filename',
                               'fmt': 'format',
                               'fo': 'forceOverwrite',
                               'fp': 'framePadding',
                               'fr': 'frame',
                               'h': 'height',
                               'ifz': 'indexFromZero',
                               'o': 'options',
                               'orn': 'showOrnaments',
                               'os': 'offScreen',
                               'osv': 'offScreenViewportUpdate',
                               'p': 'percent',
                               'q': 'query',
                               'qlt': 'quality',
                               'rao': 'replaceAudioOnly',
                               'ret': 'replaceEndTime',
                               'rf': 'replaceFilename',
                               'rfn': 'rawFrameNumbers',
                               'rst': 'replaceStartTime',
                               's': 'sound',
                               'sd': 'saveDepth',
                               'sqt': 'sequenceTime',
                               'st': 'startTime',
                               'toe': 'throwOnError',
                               'uts': 'useTraxSounds',
                               'v': 'viewer',
                               'w': 'width',
                               'wh': 'widthHeight'},
                'type': 'animation'},
  'pluginDisplayFilter': {'flags': {'classification': {'args': '<type __builtin__.str>',
                                                       'longname': 'classification',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'cls'},
                                    'deregister': {'args': '<type __builtin__.bool>',
                                                   'longname': 'deregister',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'dr'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'label': {'args': '<type __builtin__.str>',
                                              'longname': 'label',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'l'},
                                    'listFilters': {'args': '<type __builtin__.bool>',
                                                    'longname': 'listFilters',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'lf'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'register': {'args': '<type __builtin__.bool>',
                                                 'longname': 'register',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'r'}},
                          'shortFlags': {'cls': 'classification',
                                         'dr': 'deregister',
                                         'ex': 'exists',
                                         'l': 'label',
                                         'lf': 'listFilters',
                                         'q': 'query',
                                         'r': 'register'},
                          'type': 'system'},
  'pluginInfo': {'callbackFlags': ['controlCommand',
                                   'ctc',
                                   'modelEditorCommand',
                                   'mec',
                                   'constraintCommand',
                                   'cnc',
                                   'changedCommand',
                                   'cc',
                                   'command',
                                   'c'],
                 'flags': {'activeFile': {'args': '<type __builtin__.bool>',
                                          'longname': 'activeFile',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'af'},
                           'animCurveInterp': {'args': '<type __builtin__.str>',
                                               'longname': 'animCurveInterp',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'aci'},
                           'apiVersion': {'args': '<type __builtin__.bool>',
                                          'longname': 'apiVersion',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'av'},
                           'autoload': {'args': '<type __builtin__.bool>',
                                        'longname': 'autoload',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'a'},
                           'cacheFormat': {'args': '<type __builtin__.bool>',
                                           'longname': 'cacheFormat',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'cf'},
                           'changedCommand': {'args': 'script',
                                              'longname': 'changedCommand',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'cc'},
                           'command': {'args': '<type __builtin__.str>',
                                       'longname': 'command',
                                       'modes': ['query', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'c'},
                           'constraintCommand': {'args': '<type __builtin__.bool>',
                                                 'longname': 'constraintCommand',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'cnc'},
                           'controlCommand': {'args': '<type __builtin__.bool>',
                                              'longname': 'controlCommand',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'ctc'},
                           'data': {'args': ['<type __builtin__.str>',
                                             '<type __builtin__.str>'],
                                    'longname': 'data',
                                    'modes': ['query', 'multiuse'],
                                    'numArgs': 2,
                                    'shortname': 'd'},
                           'dependNode': {'args': '<type __builtin__.bool>',
                                          'longname': 'dependNode',
                                          'modes': ['query', 'multiuse'],
                                          'numArgs': 0,
                                          'shortname': 'dn'},
                           'dependNodeByType': {'args': '<type __builtin__.str>',
                                                'longname': 'dependNodeByType',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'dnt'},
                           'dependNodeId': {'args': '<type __builtin__.str>',
                                            'longname': 'dependNodeId',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'dni'},
                           'device': {'args': '<type __builtin__.bool>',
                                      'longname': 'device',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'dv'},
                           'dragAndDropBehavior': {'args': '<type __builtin__.bool>',
                                                   'longname': 'dragAndDropBehavior',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'ddb'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'iksolver': {'args': '<type __builtin__.bool>',
                                        'longname': 'iksolver',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'ik'},
                           'listPlugins': {'args': '<type __builtin__.bool>',
                                           'longname': 'listPlugins',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'ls'},
                           'listPluginsPath': {'args': '<type __builtin__.bool>',
                                               'longname': 'listPluginsPath',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'lsp'},
                           'loadPluginPrefs': {'args': '<type __builtin__.bool>',
                                               'longname': 'loadPluginPrefs',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'lpp'},
                           'loaded': {'args': '<type __builtin__.bool>',
                                      'longname': 'loaded',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'l'},
                           'modelEditorCommand': {'args': '<type __builtin__.bool>',
                                                  'longname': 'modelEditorCommand',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'mec'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['query'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'path': {'args': '<type __builtin__.str>',
                                    'longname': 'path',
                                    'modes': ['query'],
                                    'numArgs': 1,
                                    'shortname': 'p'},
                           'pluginsInUse': {'args': '<type __builtin__.bool>',
                                            'longname': 'pluginsInUse',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'pu'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'referenceTranslators': {'args': '<type __builtin__.bool>',
                                                    'longname': 'referenceTranslators',
                                                    'modes': ['query'],
                                                    'numArgs': 1,
                                                    'shortname': 'rtr'},
                           'registered': {'args': '<type __builtin__.bool>',
                                          'longname': 'registered',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'r'},
                           'remove': {'args': '<type __builtin__.bool>',
                                      'longname': 'remove',
                                      'modes': ['edit'],
                                      'numArgs': 0,
                                      'shortname': 'rm'},
                           'renderer': {'args': '<type __builtin__.bool>',
                                        'longname': 'renderer',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'rdr'},
                           'savePluginPrefs': {'args': '<type __builtin__.bool>',
                                               'longname': 'savePluginPrefs',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'spp'},
                           'serviceDescriptions': {'args': '<type __builtin__.bool>',
                                                   'longname': 'serviceDescriptions',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'sd'},
                           'settings': {'args': '<type __builtin__.bool>',
                                        'longname': 'settings',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'set'},
                           'tool': {'args': '<type __builtin__.str>',
                                    'longname': 'tool',
                                    'modes': ['query', 'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 't'},
                           'translator': {'args': '<type __builtin__.bool>',
                                          'longname': 'translator',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'tr'},
                           'unloadOk': {'args': '<type __builtin__.bool>',
                                        'longname': 'unloadOk',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'uo'},
                           'userNamed': {'args': '<type __builtin__.bool>',
                                         'longname': 'userNamed',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'u'},
                           'vendor': {'args': '<type __builtin__.str>',
                                      'longname': 'vendor',
                                      'modes': ['query'],
                                      'numArgs': 1,
                                      'shortname': 'vd'},
                           'version': {'args': '<type __builtin__.bool>',
                                       'longname': 'version',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'v'},
                           'writeRequires': {'args': '<type __builtin__.bool>',
                                             'longname': 'writeRequires',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'wr'}},
                 'shortFlags': {'a': 'autoload',
                                'aci': 'animCurveInterp',
                                'af': 'activeFile',
                                'av': 'apiVersion',
                                'c': 'command',
                                'cc': 'changedCommand',
                                'cf': 'cacheFormat',
                                'cnc': 'constraintCommand',
                                'ctc': 'controlCommand',
                                'd': 'data',
                                'ddb': 'dragAndDropBehavior',
                                'dn': 'dependNode',
                                'dni': 'dependNodeId',
                                'dnt': 'dependNodeByType',
                                'dv': 'device',
                                'e': 'edit',
                                'ik': 'iksolver',
                                'l': 'loaded',
                                'lpp': 'loadPluginPrefs',
                                'ls': 'listPlugins',
                                'lsp': 'listPluginsPath',
                                'mec': 'modelEditorCommand',
                                'n': 'name',
                                'p': 'path',
                                'pu': 'pluginsInUse',
                                'q': 'query',
                                'r': 'registered',
                                'rdr': 'renderer',
                                'rm': 'remove',
                                'rtr': 'referenceTranslators',
                                'sd': 'serviceDescriptions',
                                'set': 'settings',
                                'spp': 'savePluginPrefs',
                                't': 'tool',
                                'tr': 'translator',
                                'u': 'userNamed',
                                'uo': 'unloadOk',
                                'v': 'version',
                                'vd': 'vendor',
                                'wr': 'writeRequires'},
                 'type': 'system'},
  'pointConstraint': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'layer': {'args': '<type __builtin__.str>',
                                          'longname': 'layer',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'l'},
                                'maintainOffset': {'args': '<type __builtin__.bool>',
                                                   'longname': 'maintainOffset',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'mo'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'offset': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'offset',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'o'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'remove': {'args': '<type __builtin__.bool>',
                                           'longname': 'remove',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'rm'},
                                'skip': {'args': '<type __builtin__.str>',
                                         'longname': 'skip',
                                         'modes': ['create',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'sk'},
                                'targetList': {'args': '<type __builtin__.bool>',
                                               'longname': 'targetList',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'tl'},
                                'weight': {'args': '<type __builtin__.float>',
                                           'longname': 'weight',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'w'},
                                'weightAliasList': {'args': '<type __builtin__.bool>',
                                                    'longname': 'weightAliasList',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'wal'}},
                      'resultNeedsUnpacking': True,
                      'shortFlags': {'e': 'edit',
                                     'l': 'layer',
                                     'mo': 'maintainOffset',
                                     'n': 'name',
                                     'o': 'offset',
                                     'q': 'query',
                                     'rm': 'remove',
                                     'sk': 'skip',
                                     'tl': 'targetList',
                                     'w': 'weight',
                                     'wal': 'weightAliasList'},
                      'type': 'animation'},
  'pointCurveConstraint': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                                 'longname': 'caching',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cch'},
                                     'constructionHistory': {'args': '<type __builtin__.bool>',
                                                             'longname': 'constructionHistory',
                                                             'modes': ['create'],
                                                             'numArgs': 1,
                                                             'shortname': 'ch'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'frozen': {'args': '<type __builtin__.bool>',
                                                'longname': 'frozen',
                                                'numArgs': 1,
                                                'shortname': 'fzn'},
                                     'name': {'args': '<type __builtin__.str>',
                                              'longname': 'name',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'n'},
                                     'nodeState': {'args': '<type __builtin__.int>',
                                                   'longname': 'nodeState',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'nds'},
                                     'object': {'args': '<type __builtin__.bool>',
                                                'longname': 'object',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'o'},
                                     'pointConstraintUVW': {'args': ['<type __builtin__.float>',
                                                                     '<type __builtin__.float>',
                                                                     '<type __builtin__.float>'],
                                                            'longname': 'pointConstraintUVW',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 3,
                                                            'shortname': 'puv'},
                                     'pointWeight': {'args': '<type __builtin__.float>',
                                                     'longname': 'pointWeight',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'pw'},
                                     'position': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'position',
                                                  'modes': ['create'],
                                                  'numArgs': 3,
                                                  'shortname': 'p'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                         'longname': 'replaceOriginal',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'rpo'},
                                     'weight': {'args': '<type __builtin__.float>',
                                                'longname': 'weight',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'w'}},
                           'shortFlags': {'cch': 'caching',
                                          'ch': 'constructionHistory',
                                          'e': 'edit',
                                          'fzn': 'frozen',
                                          'n': 'name',
                                          'nds': 'nodeState',
                                          'o': 'object',
                                          'p': 'position',
                                          'puv': 'pointConstraintUVW',
                                          'pw': 'pointWeight',
                                          'q': 'query',
                                          'rpo': 'replaceOriginal',
                                          'w': 'weight'},
                           'type': 'modeling'},
  'pointLight': {'flags': {'decayRate': {'args': '<type __builtin__.int>',
                                         'longname': 'decayRate',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'd'},
                           'discRadius': {'args': '<type __builtin__.float>',
                                          'longname': 'discRadius',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'drs'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exclusive': {'args': '<type __builtin__.bool>',
                                         'longname': 'exclusive',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'exc'},
                           'intensity': {'args': '<type __builtin__.float>',
                                         'longname': 'intensity',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'position': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'position',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 3,
                                        'shortname': 'pos'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'rgb': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'rgb',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 3,
                                   'shortname': 'rgb'},
                           'rotation': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'rotation',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 3,
                                        'shortname': 'rot'},
                           'shadowColor': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'shadowColor',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'sc'},
                           'shadowDither': {'args': '<type __builtin__.float>',
                                            'longname': 'shadowDither',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sd'},
                           'shadowSamples': {'args': '<type __builtin__.int>',
                                             'longname': 'shadowSamples',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sh'},
                           'softShadow': {'args': '<type __builtin__.bool>',
                                          'longname': 'softShadow',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ss'},
                           'useRayTraceShadows': {'args': '<type __builtin__.bool>',
                                                  'longname': 'useRayTraceShadows',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'rs'}},
                 'shortFlags': {'d': 'decayRate',
                                'drs': 'discRadius',
                                'e': 'edit',
                                'exc': 'exclusive',
                                'i': 'intensity',
                                'n': 'name',
                                'pos': 'position',
                                'q': 'query',
                                'rgb': 'rgb',
                                'rot': 'rotation',
                                'rs': 'useRayTraceShadows',
                                'sc': 'shadowColor',
                                'sd': 'shadowDither',
                                'sh': 'shadowSamples',
                                'ss': 'softShadow'},
                 'type': 'rendering'},
  'pointOnCurve': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'curvatureCenter': {'args': '<type __builtin__.bool>',
                                                 'longname': 'curvatureCenter',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'cc'},
                             'curvatureRadius': {'args': '<type __builtin__.bool>',
                                                 'longname': 'curvatureRadius',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'cr'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'normal': {'args': '<type __builtin__.bool>',
                                        'longname': 'normal',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'no'},
                             'normalizedNormal': {'args': '<type __builtin__.bool>',
                                                  'longname': 'normalizedNormal',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'nn'},
                             'normalizedTangent': {'args': '<type __builtin__.bool>',
                                                   'longname': 'normalizedTangent',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'nt'},
                             'parameter': {'args': '<type __builtin__.float>',
                                           'longname': 'parameter',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pr'},
                             'position': {'args': '<type __builtin__.bool>',
                                          'longname': 'position',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'p'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'tangent': {'args': '<type __builtin__.bool>',
                                         'longname': 'tangent',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 't'},
                             'turnOnPercentage': {'args': '<type __builtin__.bool>',
                                                  'longname': 'turnOnPercentage',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'top'}},
                   'shortFlags': {'cc': 'curvatureCenter',
                                  'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'cr': 'curvatureRadius',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'nds': 'nodeState',
                                  'nn': 'normalizedNormal',
                                  'no': 'normal',
                                  'nt': 'normalizedTangent',
                                  'p': 'position',
                                  'pr': 'parameter',
                                  'q': 'query',
                                  't': 'tangent',
                                  'top': 'turnOnPercentage'},
                   'type': 'modeling'},
  'pointOnPolyConstraint': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                               'longname': 'edit',
                                               'numArgs': 0,
                                               'shortname': 'e'},
                                      'layer': {'args': '<type __builtin__.str>',
                                                'longname': 'layer',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'l'},
                                      'maintainOffset': {'args': '<type __builtin__.bool>',
                                                         'longname': 'maintainOffset',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'mo'},
                                      'name': {'args': '<type __builtin__.str>',
                                               'longname': 'name',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'n'},
                                      'offset': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'offset',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'o'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'remove': {'args': '<type __builtin__.bool>',
                                                 'longname': 'remove',
                                                 'modes': ['edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'rm'},
                                      'skip': {'args': '<type __builtin__.str>',
                                               'longname': 'skip',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'sk'},
                                      'targetList': {'args': '<type __builtin__.bool>',
                                                     'longname': 'targetList',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'tl'},
                                      'weight': {'args': '<type __builtin__.float>',
                                                 'longname': 'weight',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'w'},
                                      'weightAliasList': {'args': '<type __builtin__.bool>',
                                                          'longname': 'weightAliasList',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'wal'}},
                            'shortFlags': {'e': 'edit',
                                           'l': 'layer',
                                           'mo': 'maintainOffset',
                                           'n': 'name',
                                           'o': 'offset',
                                           'q': 'query',
                                           'rm': 'remove',
                                           'sk': 'skip',
                                           'tl': 'targetList',
                                           'w': 'weight',
                                           'wal': 'weightAliasList'},
                            'type': 'animation'},
  'pointOnSurface': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'normal': {'args': '<type __builtin__.bool>',
                                          'longname': 'normal',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'no'},
                               'normalizedNormal': {'args': '<type __builtin__.bool>',
                                                    'longname': 'normalizedNormal',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'nn'},
                               'normalizedTangentU': {'args': '<type __builtin__.bool>',
                                                      'longname': 'normalizedTangentU',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'ntu'},
                               'normalizedTangentV': {'args': '<type __builtin__.bool>',
                                                      'longname': 'normalizedTangentV',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'ntv'},
                               'parameterU': {'args': '<type __builtin__.float>',
                                              'longname': 'parameterU',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'u'},
                               'parameterV': {'args': '<type __builtin__.float>',
                                              'longname': 'parameterV',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'v'},
                               'position': {'args': '<type __builtin__.bool>',
                                            'longname': 'position',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 0,
                                            'shortname': 'p'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'tangentU': {'args': '<type __builtin__.bool>',
                                            'longname': 'tangentU',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 0,
                                            'shortname': 'tu'},
                               'tangentV': {'args': '<type __builtin__.bool>',
                                            'longname': 'tangentV',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 0,
                                            'shortname': 'tv'},
                               'turnOnPercentage': {'args': '<type __builtin__.bool>',
                                                    'longname': 'turnOnPercentage',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'top'}},
                     'shortFlags': {'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'e': 'edit',
                                    'fzn': 'frozen',
                                    'nds': 'nodeState',
                                    'nn': 'normalizedNormal',
                                    'no': 'normal',
                                    'ntu': 'normalizedTangentU',
                                    'ntv': 'normalizedTangentV',
                                    'p': 'position',
                                    'q': 'query',
                                    'top': 'turnOnPercentage',
                                    'tu': 'tangentU',
                                    'tv': 'tangentV',
                                    'u': 'parameterU',
                                    'v': 'parameterV'},
                     'type': 'modeling'},
  'pointPosition': {'flags': {'local': {'args': '<type __builtin__.bool>',
                                        'longname': 'local',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'l'},
                              'world': {'args': '<type __builtin__.bool>',
                                        'longname': 'world',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'w'}},
                    'shortFlags': {'l': 'local', 'w': 'world'},
                    'type': 'modeling'},
  'poleVectorConstraint': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'layer': {'args': '<type __builtin__.str>',
                                               'longname': 'layer',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'l'},
                                     'name': {'args': '<type __builtin__.str>',
                                              'longname': 'name',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'n'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'remove': {'args': '<type __builtin__.bool>',
                                                'longname': 'remove',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'rm'},
                                     'targetList': {'args': '<type __builtin__.bool>',
                                                    'longname': 'targetList',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'tl'},
                                     'weight': {'args': '<type __builtin__.float>',
                                                'longname': 'weight',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'w'},
                                     'weightAliasList': {'args': '<type __builtin__.bool>',
                                                         'longname': 'weightAliasList',
                                                         'modes': ['query'],
                                                         'numArgs': 0,
                                                         'shortname': 'wal'}},
                           'resultNeedsUnpacking': True,
                           'shortFlags': {'e': 'edit',
                                          'n': 'name',
                                          'q': 'query',
                                          'rm': 'remove',
                                          'tl': 'targetList',
                                          'w': 'weight',
                                          'wal': 'weightAliasList'},
                           'type': 'animation'},
  'polyAppend': {'flags': {'append': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'append',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 3,
                                      'shortname': 'a'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'edge': {'args': '<type __builtin__.int>',
                                    'longname': 'edge',
                                    'modes': ['create', 'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'ed'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'hole': {'args': '<type __builtin__.bool>',
                                    'longname': 'hole',
                                    'modes': ['create', 'multiuse'],
                                    'numArgs': 0,
                                    'shortname': 'hl'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'point': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'point',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 3,
                                     'shortname': 'p'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'subdivision': {'args': '<type __builtin__.int>',
                                           'longname': 'subdivision',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 's'},
                           'texture': {'args': '<type __builtin__.int>',
                                       'longname': 'texture',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'tx'}},
                 'shortFlags': {'a': 'append',
                                'ch': 'constructionHistory',
                                'e': 'edit',
                                'ed': 'edge',
                                'hl': 'hole',
                                'n': 'name',
                                'p': 'point',
                                'q': 'query',
                                's': 'subdivision',
                                'tx': 'texture'},
                 'type': 'modeling'},
  'polyAppendFacetCtx': {'flags': {'append': {'args': '<type __builtin__.bool>',
                                              'longname': 'append',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ap'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'isRotateAvailable': {'args': '<type __builtin__.bool>',
                                                         'longname': 'isRotateAvailable',
                                                         'modes': ['query'],
                                                         'numArgs': 1,
                                                         'shortname': 'isr'},
                                   'maximumNumberOfPoints': {'args': '<type __builtin__.int>',
                                                             'longname': 'maximumNumberOfPoints',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'mp'},
                                   'planarConstraint': {'args': '<type __builtin__.bool>',
                                                        'longname': 'planarConstraint',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'pc'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'rotate': {'args': '<type __builtin__.float>',
                                              'longname': 'rotate',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'r'},
                                   'subdivision': {'args': '<type __builtin__.int>',
                                                   'longname': 'subdivision',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 's'},
                                   'texture': {'args': '<type __builtin__.int>',
                                               'longname': 'texture',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tx'}},
                         'shortFlags': {'ap': 'append',
                                        'e': 'edit',
                                        'ex': 'exists',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'isr': 'isRotateAvailable',
                                        'mp': 'maximumNumberOfPoints',
                                        'pc': 'planarConstraint',
                                        'q': 'query',
                                        'r': 'rotate',
                                        's': 'subdivision',
                                        'tx': 'texture'},
                         'type': 'context'},
  'polyAppendVertex': {'flags': {'append': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'append',
                                            'modes': ['create',
                                                      'multiuse'],
                                            'numArgs': 3,
                                            'shortname': 'a'},
                                 'constructionHistory': {'args': '<type __builtin__.bool>',
                                                         'longname': 'constructionHistory',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'ch'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'hole': {'args': '<type __builtin__.bool>',
                                          'longname': 'hole',
                                          'modes': ['create', 'multiuse'],
                                          'numArgs': 0,
                                          'shortname': 'h'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'point': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'point',
                                           'modes': ['create', 'multiuse'],
                                           'numArgs': 3,
                                           'shortname': 'p'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'texture': {'args': '<type __builtin__.int>',
                                             'longname': 'texture',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'tx'},
                                 'vertex': {'args': '<type __builtin__.int>',
                                            'longname': 'vertex',
                                            'modes': ['create',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'v'}},
                       'shortFlags': {'a': 'append',
                                      'ch': 'constructionHistory',
                                      'e': 'edit',
                                      'h': 'hole',
                                      'n': 'name',
                                      'p': 'point',
                                      'q': 'query',
                                      'tx': 'texture',
                                      'v': 'vertex'},
                       'type': 'modeling'},
  'polyAutoProjection': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                               'longname': 'caching',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cch'},
                                   'constructionHistory': {'args': '<type __builtin__.bool>',
                                                           'longname': 'constructionHistory',
                                                           'modes': ['create',
                                                                     'query'],
                                                           'numArgs': 1,
                                                           'shortname': 'ch'},
                                   'createNewMap': {'args': '<type __builtin__.bool>',
                                                    'longname': 'createNewMap',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'cm'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'frozen': {'args': '<type __builtin__.bool>',
                                              'longname': 'frozen',
                                              'numArgs': 1,
                                              'shortname': 'fzn'},
                                   'insertBeforeDeformers': {'args': '<type __builtin__.bool>',
                                                             'longname': 'insertBeforeDeformers',
                                                             'modes': ['create'],
                                                             'numArgs': 1,
                                                             'shortname': 'ibd'},
                                   'layout': {'args': '<type __builtin__.int>',
                                              'longname': 'layout',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'l'},
                                   'layoutMethod': {'args': '<type __builtin__.int>',
                                                    'longname': 'layoutMethod',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'lm'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'nodeState': {'args': '<type __builtin__.int>',
                                                 'longname': 'nodeState',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'nds'},
                                   'optimize': {'args': '<type __builtin__.int>',
                                                'longname': 'optimize',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'o'},
                                   'percentageSpace': {'args': '<type __builtin__.float>',
                                                       'longname': 'percentageSpace',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ps'},
                                   'pivot': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'pivot',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'pvt'},
                                   'pivotX': {'args': '<type __builtin__.float>',
                                              'longname': 'pivotX',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pvx'},
                                   'pivotY': {'args': '<type __builtin__.float>',
                                              'longname': 'pivotY',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pvy'},
                                   'pivotZ': {'args': '<type __builtin__.float>',
                                              'longname': 'pivotZ',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pvz'},
                                   'planes': {'args': '<type __builtin__.int>',
                                              'longname': 'planes',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'p'},
                                   'projectBothDirections': {'args': '<type __builtin__.bool>',
                                                             'longname': 'projectBothDirections',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'pb'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'rotate': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'rotate',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'ro'},
                                   'rotateX': {'args': '<type __builtin__.float>',
                                               'longname': 'rotateX',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rx'},
                                   'rotateY': {'args': '<type __builtin__.float>',
                                               'longname': 'rotateY',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ry'},
                                   'rotateZ': {'args': '<type __builtin__.float>',
                                               'longname': 'rotateZ',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rz'},
                                   'scale': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'scale',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 's'},
                                   'scaleMode': {'args': '<type __builtin__.int>',
                                                 'longname': 'scaleMode',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sc'},
                                   'scaleX': {'args': '<type __builtin__.float>',
                                              'longname': 'scaleX',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sx'},
                                   'scaleY': {'args': '<type __builtin__.float>',
                                              'longname': 'scaleY',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sy'},
                                   'scaleZ': {'args': '<type __builtin__.float>',
                                              'longname': 'scaleZ',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sz'},
                                   'skipIntersect': {'args': '<type __builtin__.bool>',
                                                     'longname': 'skipIntersect',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'si'},
                                   'translate': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'translate',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 't'},
                                   'translateX': {'args': '<type __builtin__.float>',
                                                  'longname': 'translateX',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'tx'},
                                   'translateY': {'args': '<type __builtin__.float>',
                                                  'longname': 'translateY',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ty'},
                                   'translateZ': {'args': '<type __builtin__.float>',
                                                  'longname': 'translateZ',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'tz'},
                                   'uvSetName': {'args': '<type __builtin__.str>',
                                                 'longname': 'uvSetName',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'uvs'},
                                   'worldSpace': {'args': '<type __builtin__.bool>',
                                                  'longname': 'worldSpace',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ws'}},
                         'shortFlags': {'cch': 'caching',
                                        'ch': 'constructionHistory',
                                        'cm': 'createNewMap',
                                        'e': 'edit',
                                        'fzn': 'frozen',
                                        'ibd': 'insertBeforeDeformers',
                                        'l': 'layout',
                                        'lm': 'layoutMethod',
                                        'n': 'name',
                                        'nds': 'nodeState',
                                        'o': 'optimize',
                                        'p': 'planes',
                                        'pb': 'projectBothDirections',
                                        'ps': 'percentageSpace',
                                        'pvt': 'pivot',
                                        'pvx': 'pivotX',
                                        'pvy': 'pivotY',
                                        'pvz': 'pivotZ',
                                        'q': 'query',
                                        'ro': 'rotate',
                                        'rx': 'rotateX',
                                        'ry': 'rotateY',
                                        'rz': 'rotateZ',
                                        's': 'scale',
                                        'sc': 'scaleMode',
                                        'si': 'skipIntersect',
                                        'sx': 'scaleX',
                                        'sy': 'scaleY',
                                        'sz': 'scaleZ',
                                        't': 'translate',
                                        'tx': 'translateX',
                                        'ty': 'translateY',
                                        'tz': 'translateZ',
                                        'uvs': 'uvSetName',
                                        'ws': 'worldSpace'},
                         'type': 'modeling'},
  'polyAverageNormal': {'flags': {'allowZeroNormal': {'args': '<type __builtin__.bool>',
                                                      'longname': 'allowZeroNormal',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'azn'},
                                  'distance': {'args': '<type __builtin__.float>',
                                               'longname': 'distance',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'd'},
                                  'postnormalize': {'args': '<type __builtin__.bool>',
                                                    'longname': 'postnormalize',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'pon'},
                                  'prenormalize': {'args': '<type __builtin__.bool>',
                                                   'longname': 'prenormalize',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'prn'},
                                  'replaceNormalXYZ': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'replaceNormalXYZ',
                                                       'modes': ['create'],
                                                       'numArgs': 3,
                                                       'shortname': 'xyz'}},
                        'shortFlags': {'azn': 'allowZeroNormal',
                                       'd': 'distance',
                                       'pon': 'postnormalize',
                                       'prn': 'prenormalize',
                                       'xyz': 'replaceNormalXYZ'},
                        'type': 'modeling'},
  'polyAverageVertex': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                              'longname': 'caching',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cch'},
                                  'constructionHistory': {'args': '<type __builtin__.bool>',
                                                          'longname': 'constructionHistory',
                                                          'modes': ['create',
                                                                    'query'],
                                                          'numArgs': 1,
                                                          'shortname': 'ch'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'frozen': {'args': '<type __builtin__.bool>',
                                             'longname': 'frozen',
                                             'numArgs': 1,
                                             'shortname': 'fzn'},
                                  'iterations': {'args': '<type __builtin__.int>',
                                                 'longname': 'iterations',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'i'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'nodeState': {'args': '<type __builtin__.int>',
                                                'longname': 'nodeState',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nds'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'worldSpace': {'args': '<type __builtin__.bool>',
                                                 'longname': 'worldSpace',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ws'}},
                        'shortFlags': {'cch': 'caching',
                                       'ch': 'constructionHistory',
                                       'e': 'edit',
                                       'fzn': 'frozen',
                                       'i': 'iterations',
                                       'n': 'name',
                                       'nds': 'nodeState',
                                       'q': 'query',
                                       'ws': 'worldSpace'},
                        'type': 'modeling'},
  'polyBevel': {'flags': {'angleTolerance': {'args': '<type __builtin__.float>',
                                             'longname': 'angleTolerance',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'at'},
                          'autoFit': {'args': '<type __builtin__.bool>',
                                      'longname': 'autoFit',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'af'},
                          'caching': {'args': '<type __builtin__.bool>',
                                      'longname': 'caching',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'cch'},
                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'constructionHistory',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'fillNgons': {'args': '<type __builtin__.bool>',
                                        'longname': 'fillNgons',
                                        'numArgs': 1,
                                        'shortname': 'fn'},
                          'fraction': {'args': '<type __builtin__.float>',
                                       'longname': 'fraction',
                                       'numArgs': 1,
                                       'shortname': 'f'},
                          'frozen': {'args': '<type __builtin__.bool>',
                                     'longname': 'frozen',
                                     'numArgs': 1,
                                     'shortname': 'fzn'},
                          'maya2015': {'args': '<type __builtin__.bool>',
                                       'longname': 'maya2015',
                                       'numArgs': 1,
                                       'shortname': 'm15'},
                          'mergeVertexTolerance': {'args': '<type __builtin__.float>',
                                                   'longname': 'mergeVertexTolerance',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'mvt'},
                          'mergeVertices': {'args': '<type __builtin__.bool>',
                                            'longname': 'mergeVertices',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'mv'},
                          'miteringAngle': {'args': '<type __builtin__.float>',
                                            'longname': 'miteringAngle',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ma'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nodeState': {'args': '<type __builtin__.int>',
                                        'longname': 'nodeState',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'nds'},
                          'offset': {'args': '<type __builtin__.float>',
                                     'longname': 'offset',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'o'},
                          'offsetAsFraction': {'args': '<type __builtin__.bool>',
                                               'longname': 'offsetAsFraction',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'oaf'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'roundness': {'args': '<type __builtin__.float>',
                                        'longname': 'roundness',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'r'},
                          'segments': {'args': '<type __builtin__.int>',
                                       'longname': 'segments',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sg'},
                          'smoothingAngle': {'args': '<type __builtin__.float>',
                                             'longname': 'smoothingAngle',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sa'},
                          'subdivideNgons': {'args': '<type __builtin__.bool>',
                                             'longname': 'subdivideNgons',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sn'},
                          'useLegacyBevelAlgorithm': {'args': '<type __builtin__.bool>',
                                                      'longname': 'useLegacyBevelAlgorithm',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'com'},
                          'uvAssignment': {'args': '<type __builtin__.int>',
                                           'longname': 'uvAssignment',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ua'},
                          'worldSpace': {'args': '<type __builtin__.bool>',
                                         'longname': 'worldSpace',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ws'}},
                'shortFlags': {'af': 'autoFit',
                               'at': 'angleTolerance',
                               'cch': 'caching',
                               'ch': 'constructionHistory',
                               'com': 'useLegacyBevelAlgorithm',
                               'e': 'edit',
                               'f': 'fraction',
                               'fn': 'fillNgons',
                               'fzn': 'frozen',
                               'm15': 'maya2015',
                               'ma': 'miteringAngle',
                               'mv': 'mergeVertices',
                               'mvt': 'mergeVertexTolerance',
                               'n': 'name',
                               'nds': 'nodeState',
                               'o': 'offset',
                               'oaf': 'offsetAsFraction',
                               'q': 'query',
                               'r': 'roundness',
                               'sa': 'smoothingAngle',
                               'sg': 'segments',
                               'ua': 'uvAssignment',
                               'ws': 'worldSpace'},
                'type': 'modeling'},
  'polyBevel3': {'flags': {'angleTolerance': {'args': '<type __builtin__.float>',
                                              'longname': 'angleTolerance',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'at'},
                           'autoFit': {'args': '<type __builtin__.bool>',
                                       'longname': 'autoFit',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'af'},
                           'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'chamfer': {'args': '<type __builtin__.bool>',
                                       'longname': 'chamfer',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'c'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'depth': {'args': '<type __builtin__.float>',
                                     'longname': 'depth',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'd'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'fillNgons': {'args': '<type __builtin__.bool>',
                                         'longname': 'fillNgons',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'fn'},
                           'forceParallel': {'args': '<type __builtin__.bool>',
                                             'longname': 'forceParallel',
                                             'numArgs': 1,
                                             'shortname': 'fp'},
                           'fraction': {'args': '<type __builtin__.float>',
                                        'longname': 'fraction',
                                        'numArgs': 1,
                                        'shortname': 'f'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'maya2015': {'args': '<type __builtin__.bool>',
                                        'longname': 'maya2015',
                                        'numArgs': 1,
                                        'shortname': 'm15'},
                           'maya2016SP3': {'args': '<type __builtin__.bool>',
                                           'longname': 'maya2016SP3',
                                           'numArgs': 1,
                                           'shortname': 'm16'},
                           'maya2017Update1': {'args': '<type __builtin__.bool>',
                                               'longname': 'maya2017Update1',
                                               'numArgs': 1,
                                               'shortname': 'm17'},
                           'mergeVertexTolerance': {'args': '<type __builtin__.float>',
                                                    'longname': 'mergeVertexTolerance',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'mvt'},
                           'mergeVertices': {'args': '<type __builtin__.bool>',
                                             'longname': 'mergeVertices',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'mv'},
                           'miterAlong': {'args': '<type __builtin__.int>',
                                          'longname': 'miterAlong',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'mia'},
                           'mitering': {'args': '<type __builtin__.int>',
                                        'longname': 'mitering',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                           'miteringAngle': {'args': '<type __builtin__.float>',
                                             'longname': 'miteringAngle',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ma'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'offset': {'args': '<type __builtin__.float>',
                                      'longname': 'offset',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'offsetAsFraction': {'args': '<type __builtin__.bool>',
                                                'longname': 'offsetAsFraction',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'oaf'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'roundness': {'args': '<type __builtin__.float>',
                                         'longname': 'roundness',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'r'},
                           'segments': {'args': '<type __builtin__.int>',
                                        'longname': 'segments',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'sg'},
                           'smoothingAngle': {'args': '<type __builtin__.float>',
                                              'longname': 'smoothingAngle',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sa'},
                           'subdivideNgons': {'args': '<type __builtin__.bool>',
                                              'longname': 'subdivideNgons',
                                              'numArgs': 1,
                                              'shortname': 'sn'},
                           'uvAssignment': {'args': '<type __builtin__.int>',
                                            'longname': 'uvAssignment',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ua'},
                           'worldSpace': {'args': '<type __builtin__.bool>',
                                          'longname': 'worldSpace',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ws'}},
                 'shortFlags': {'af': 'autoFit',
                                'at': 'angleTolerance',
                                'c': 'chamfer',
                                'cch': 'caching',
                                'ch': 'constructionHistory',
                                'd': 'depth',
                                'e': 'edit',
                                'f': 'fraction',
                                'fn': 'fillNgons',
                                'fp': 'forceParallel',
                                'fzn': 'frozen',
                                'm': 'mitering',
                                'm15': 'maya2015',
                                'm16': 'maya2016SP3',
                                'm17': 'maya2017Update1',
                                'ma': 'miteringAngle',
                                'mia': 'miterAlong',
                                'mv': 'mergeVertices',
                                'mvt': 'mergeVertexTolerance',
                                'n': 'name',
                                'nds': 'nodeState',
                                'o': 'offset',
                                'oaf': 'offsetAsFraction',
                                'q': 'query',
                                'r': 'roundness',
                                'sa': 'smoothingAngle',
                                'sg': 'segments',
                                'sn': 'subdivideNgons',
                                'ws': 'worldSpace'},
                 'type': 'modeling'},
  'polyBlendColor': {'flags': {'baseColorName': {'args': '<type __builtin__.str>',
                                                 'longname': 'baseColorName',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'bcn'},
                               'blendFunc': {'args': '<type __builtin__.int>',
                                             'longname': 'blendFunc',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'bfn'},
                               'blendWeightA': {'args': '<type __builtin__.float>',
                                                'longname': 'blendWeightA',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'bwa'},
                               'blendWeightB': {'args': '<type __builtin__.float>',
                                                'longname': 'blendWeightB',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'bwb'},
                               'blendWeightC': {'args': '<type __builtin__.float>',
                                                'longname': 'blendWeightC',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'bwc'},
                               'blendWeightD': {'args': '<type __builtin__.float>',
                                                'longname': 'blendWeightD',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'bwd'},
                               'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'dstColorName': {'args': '<type __builtin__.str>',
                                                'longname': 'dstColorName',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dst'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'srcColorName': {'args': '<type __builtin__.str>',
                                                'longname': 'srcColorName',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'src'}},
                     'shortFlags': {'bcn': 'baseColorName',
                                    'bfn': 'blendFunc',
                                    'bwa': 'blendWeightA',
                                    'bwb': 'blendWeightB',
                                    'bwc': 'blendWeightC',
                                    'bwd': 'blendWeightD',
                                    'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'dst': 'dstColorName',
                                    'e': 'edit',
                                    'fzn': 'frozen',
                                    'n': 'name',
                                    'nds': 'nodeState',
                                    'q': 'query',
                                    'src': 'srcColorName'},
                     'type': 'modeling'},
  'polyBlindData': {'flags': {'associationType': {'args': '<type __builtin__.str>',
                                                  'longname': 'associationType',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'at'},
                              'binaryData': {'args': '<type __builtin__.str>',
                                             'longname': 'binaryData',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'bnd'},
                              'booleanData': {'args': '<type __builtin__.bool>',
                                              'longname': 'booleanData',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'bd'},
                              'delete': {'args': '<type __builtin__.bool>',
                                         'longname': 'delete',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'delete'},
                              'doubleData': {'args': '<type __builtin__.float>',
                                             'longname': 'doubleData',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'dbd'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'int64Data': {'args': '<type __builtin__.int>',
                                            'longname': 'int64Data',
                                            'modes': ['create',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'lid'},
                              'intData': {'args': '<type __builtin__.int>',
                                          'longname': 'intData',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'ind'},
                              'longDataName': {'args': '<type __builtin__.str>',
                                               'longname': 'longDataName',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'ldn'},
                              'rescan': {'args': '<type __builtin__.bool>',
                                         'longname': 'rescan',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'res'},
                              'reset': {'args': '<type __builtin__.bool>',
                                        'longname': 'reset',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'rst'},
                              'shape': {'args': '<type __builtin__.bool>',
                                        'longname': 'shape',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'sh'},
                              'shortDataName': {'args': '<type __builtin__.str>',
                                                'longname': 'shortDataName',
                                                'modes': ['create',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'sdn'},
                              'stringData': {'args': '<type __builtin__.str>',
                                             'longname': 'stringData',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'sd'},
                              'typeId': {'args': '<type __builtin__.int>',
                                         'longname': 'typeId',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'id'}},
                    'removedFlags': {'del': 'delete'},
                    'shortFlags': {'at': 'associationType',
                                   'bd': 'booleanData',
                                   'bnd': 'binaryData',
                                   'dbd': 'doubleData',
                                   'delete': 'delete',
                                   'e': 'edit',
                                   'id': 'typeId',
                                   'ind': 'intData',
                                   'ldn': 'longDataName',
                                   'lid': 'int64Data',
                                   'res': 'rescan',
                                   'rst': 'reset',
                                   'sd': 'stringData',
                                   'sdn': 'shortDataName',
                                   'sh': 'shape'},
                    'type': 'modeling'},
  'polyBoolOp': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'classification': {'args': '<type __builtin__.int>',
                                              'longname': 'classification',
                                              'numArgs': 1,
                                              'shortname': 'cls'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'faceAreaThreshold': {'args': '<type __builtin__.float>',
                                                 'longname': 'faceAreaThreshold',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'fat'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'mergeUVSets': {'args': '<type __builtin__.int>',
                                           'longname': 'mergeUVSets',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'muv'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'object': {'args': '<type __builtin__.bool>',
                                      'longname': 'object',
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'operation': {'args': '<type __builtin__.int>',
                                         'longname': 'operation',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'op'},
                           'preserveColor': {'args': '<type __builtin__.bool>',
                                             'longname': 'preserveColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'pcr'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'useThresholds': {'args': '<type __builtin__.bool>',
                                             'longname': 'useThresholds',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'uth'},
                           'vertexDistanceThreshold': {'args': '<type __builtin__.float>',
                                                       'longname': 'vertexDistanceThreshold',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'vdt'}},
                 'shortFlags': {'cch': 'caching',
                                'ch': 'constructionHistory',
                                'cls': 'classification',
                                'e': 'edit',
                                'fat': 'faceAreaThreshold',
                                'fzn': 'frozen',
                                'muv': 'mergeUVSets',
                                'n': 'name',
                                'nds': 'nodeState',
                                'o': 'object',
                                'op': 'operation',
                                'pcr': 'preserveColor',
                                'q': 'query',
                                'uth': 'useThresholds',
                                'vdt': 'vertexDistanceThreshold'},
                 'type': 'modeling'},
  'polyBridgeEdge': {'flags': {'bridgeOffset': {'args': '<type __builtin__.int>',
                                                'longname': 'bridgeOffset',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'bo'},
                               'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'curveType': {'args': '<type __builtin__.int>',
                                             'longname': 'curveType',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ctp'},
                               'direction': {'args': '<type __builtin__.int>',
                                             'longname': 'direction',
                                             'numArgs': 1,
                                             'shortname': 'd'},
                               'divisions': {'args': '<type __builtin__.int>',
                                             'longname': 'divisions',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dv'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'inputCurve': {'args': 'PyNode',
                                              'longname': 'inputCurve',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'inc'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'reverse': {'args': '<type __builtin__.bool>',
                                           'longname': 'reverse',
                                           'numArgs': 1,
                                           'shortname': 'rev'},
                               'smoothingAngle': {'args': '<type __builtin__.float>',
                                                  'longname': 'smoothingAngle',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sma'},
                               'sourceDirection': {'args': '<type __builtin__.int>',
                                                   'longname': 'sourceDirection',
                                                   'numArgs': 1,
                                                   'shortname': 'sd'},
                               'startVert1': {'args': '<type __builtin__.int>',
                                              'longname': 'startVert1',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sv1'},
                               'startVert2': {'args': '<type __builtin__.int>',
                                              'longname': 'startVert2',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sv2'},
                               'taper': {'args': '<type __builtin__.float>',
                                         'longname': 'taper',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'tp'},
                               'taperCurve_FloatValue': {'args': '<type __builtin__.float>',
                                                         'longname': 'taperCurve_FloatValue',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'cfv'},
                               'taperCurve_Interp': {'args': '<type __builtin__.int>',
                                                     'longname': 'taperCurve_Interp',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ci'},
                               'taperCurve_Position': {'args': '<type __builtin__.float>',
                                                       'longname': 'taperCurve_Position',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'cp'},
                               'targetDirection': {'args': '<type __builtin__.int>',
                                                   'longname': 'targetDirection',
                                                   'numArgs': 1,
                                                   'shortname': 'td'},
                               'twist': {'args': '<type __builtin__.float>',
                                         'longname': 'twist',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'twt'},
                               'worldSpace': {'args': '<type __builtin__.bool>',
                                              'longname': 'worldSpace',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ws'}},
                     'shortFlags': {'bo': 'bridgeOffset',
                                    'cch': 'caching',
                                    'cfv': 'taperCurve_FloatValue',
                                    'ch': 'constructionHistory',
                                    'ci': 'taperCurve_Interp',
                                    'cp': 'taperCurve_Position',
                                    'ctp': 'curveType',
                                    'd': 'direction',
                                    'dv': 'divisions',
                                    'e': 'edit',
                                    'fzn': 'frozen',
                                    'inc': 'inputCurve',
                                    'n': 'name',
                                    'nds': 'nodeState',
                                    'q': 'query',
                                    'rev': 'reverse',
                                    'sd': 'sourceDirection',
                                    'sma': 'smoothingAngle',
                                    'sv1': 'startVert1',
                                    'sv2': 'startVert2',
                                    'td': 'targetDirection',
                                    'tp': 'taper',
                                    'twt': 'twist',
                                    'ws': 'worldSpace'},
                     'type': 'modeling'},
  'polyCBoolOp': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'classification': {'args': '<type __builtin__.int>',
                                               'longname': 'classification',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cls'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'edgeInterpolation': {'args': '<type __builtin__.bool>',
                                                  'longname': 'edgeInterpolation',
                                                  'numArgs': 1,
                                                  'shortname': 'eit'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'faceAreaThreshold': {'args': '<type __builtin__.float>',
                                                  'longname': 'faceAreaThreshold',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'fat'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'mergeUVSets': {'args': '<type __builtin__.int>',
                                            'longname': 'mergeUVSets',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'muv'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'operation': {'args': '<type __builtin__.int>',
                                          'longname': 'operation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'op'},
                            'planarTolerance': {'args': '<type __builtin__.float>',
                                                'longname': 'planarTolerance',
                                                'numArgs': 1,
                                                'shortname': 'ptl'},
                            'preserveColor': {'args': '<type __builtin__.bool>',
                                              'longname': 'preserveColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pcr'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'sortOutput': {'args': '<type __builtin__.bool>',
                                           'longname': 'sortOutput',
                                           'numArgs': 1,
                                           'shortname': 'sop'},
                            'useCarveBooleans': {'args': '<type __builtin__.bool>',
                                                 'longname': 'useCarveBooleans',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ucb'},
                            'useThresholds': {'args': '<type __builtin__.bool>',
                                              'longname': 'useThresholds',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'uth'},
                            'vertexDistanceThreshold': {'args': '<type __builtin__.float>',
                                                        'longname': 'vertexDistanceThreshold',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'vdt'}},
                  'shortFlags': {'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'cls': 'classification',
                                 'e': 'edit',
                                 'eit': 'edgeInterpolation',
                                 'fat': 'faceAreaThreshold',
                                 'fzn': 'frozen',
                                 'muv': 'mergeUVSets',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'o': 'object',
                                 'op': 'operation',
                                 'pcr': 'preserveColor',
                                 'ptl': 'planarTolerance',
                                 'q': 'query',
                                 'sop': 'sortOutput',
                                 'ucb': 'useCarveBooleans',
                                 'uth': 'useThresholds',
                                 'vdt': 'vertexDistanceThreshold'},
                  'type': 'modeling'},
  'polyCacheMonitor': {'flags': {'cacheValue': {'args': '<type __builtin__.bool>',
                                                'longname': 'cacheValue',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'chv'},
                                 'nodeName': {'args': '<type __builtin__.str>',
                                              'longname': 'nodeName',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'nm'}},
                       'shortFlags': {'chv': 'cacheValue', 'nm': 'nodeName'},
                       'type': 'modeling'},
  'polyCanBridgeEdge': {'flags': {}, 'shortFlags': {}, 'type': 'modeling'},
  'polyCheck': {'flags': {'edge': {'args': '<type __builtin__.bool>',
                                   'longname': 'edge',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'face': {'args': '<type __builtin__.bool>',
                                   'longname': 'face',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'f'},
                          'faceOffset': {'args': '<type __builtin__.bool>',
                                         'longname': 'faceOffset',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'fo'},
                          'openFile': {'args': '<type __builtin__.str>',
                                       'longname': 'openFile',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'of'}},
                'shortFlags': {'e': 'edge',
                               'f': 'face',
                               'fo': 'faceOffset',
                               'of': 'openFile'},
                'type': 'modeling'},
  'polyChipOff': {'flags': {'attraction': {'args': '<type __builtin__.float>',
                                           'longname': 'attraction',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'att'},
                            'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'duplicate': {'args': '<type __builtin__.bool>',
                                          'longname': 'duplicate',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dup'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'gain': {'args': '<type __builtin__.float>',
                                     'longname': 'gain',
                                     'modes': ['create',
                                               'query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'ga'},
                            'gravity': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'gravity',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 3,
                                        'shortname': 'g'},
                            'gravityX': {'args': '<type __builtin__.float>',
                                         'longname': 'gravityX',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'gx'},
                            'gravityY': {'args': '<type __builtin__.float>',
                                         'longname': 'gravityY',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'gy'},
                            'gravityZ': {'args': '<type __builtin__.float>',
                                         'longname': 'gravityZ',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'gz'},
                            'keepFacesTogether': {'args': '<type __builtin__.bool>',
                                                  'longname': 'keepFacesTogether',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'kft'},
                            'keepFacetTogether': {'args': '<type __builtin__.bool>',
                                                  'longname': 'keepFacetTogether',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'xft'},
                            'localCenter': {'args': '<type __builtin__.int>',
                                            'longname': 'localCenter',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'lc'},
                            'localDirection': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'localDirection',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'ld'},
                            'localDirectionX': {'args': '<type __builtin__.float>',
                                                'longname': 'localDirectionX',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ldx'},
                            'localDirectionY': {'args': '<type __builtin__.float>',
                                                'longname': 'localDirectionY',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ldy'},
                            'localDirectionZ': {'args': '<type __builtin__.float>',
                                                'longname': 'localDirectionZ',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ldz'},
                            'localRotate': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'localRotate',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'lr'},
                            'localRotateX': {'args': '<type __builtin__.float>',
                                             'longname': 'localRotateX',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lrx'},
                            'localRotateY': {'args': '<type __builtin__.float>',
                                             'longname': 'localRotateY',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lry'},
                            'localRotateZ': {'args': '<type __builtin__.float>',
                                             'longname': 'localRotateZ',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lrz'},
                            'localScale': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'localScale',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'ls'},
                            'localScaleX': {'args': '<type __builtin__.float>',
                                            'longname': 'localScaleX',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'lsx'},
                            'localScaleY': {'args': '<type __builtin__.float>',
                                            'longname': 'localScaleY',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'lsy'},
                            'localScaleZ': {'args': '<type __builtin__.float>',
                                            'longname': 'localScaleZ',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'lsz'},
                            'localTranslate': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'localTranslate',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'lt'},
                            'localTranslateX': {'args': '<type __builtin__.float>',
                                                'longname': 'localTranslateX',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ltx'},
                            'localTranslateY': {'args': '<type __builtin__.float>',
                                                'longname': 'localTranslateY',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'lty'},
                            'localTranslateZ': {'args': '<type __builtin__.float>',
                                                'longname': 'localTranslateZ',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ltz'},
                            'magnX': {'args': '<type __builtin__.float>',
                                      'longname': 'magnX',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'mx'},
                            'magnY': {'args': '<type __builtin__.float>',
                                      'longname': 'magnY',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'my'},
                            'magnZ': {'args': '<type __builtin__.float>',
                                      'longname': 'magnZ',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'mz'},
                            'magnet': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'magnet',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 3,
                                       'shortname': 'm'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'offset': {'args': '<type __builtin__.float>',
                                       'longname': 'offset',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'off'},
                            'pivot': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'pivot',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 3,
                                      'shortname': 'pvt'},
                            'pivotX': {'args': '<type __builtin__.float>',
                                       'longname': 'pivotX',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'pvx'},
                            'pivotY': {'args': '<type __builtin__.float>',
                                       'longname': 'pivotY',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'pvy'},
                            'pivotZ': {'args': '<type __builtin__.float>',
                                       'longname': 'pivotZ',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'pvz'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'random': {'args': '<type __builtin__.float>',
                                       'longname': 'random',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ran'},
                            'rotate': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'rotate',
                                       'numArgs': 3,
                                       'shortname': 'ro'},
                            'rotateX': {'args': '<type __builtin__.float>',
                                        'longname': 'rotateX',
                                        'numArgs': 1,
                                        'shortname': 'rx'},
                            'rotateY': {'args': '<type __builtin__.float>',
                                        'longname': 'rotateY',
                                        'numArgs': 1,
                                        'shortname': 'ry'},
                            'rotateZ': {'args': '<type __builtin__.float>',
                                        'longname': 'rotateZ',
                                        'numArgs': 1,
                                        'shortname': 'rz'},
                            'scale': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'scale',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 3,
                                      'shortname': 's'},
                            'scaleX': {'args': '<type __builtin__.float>',
                                       'longname': 'scaleX',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sx'},
                            'scaleY': {'args': '<type __builtin__.float>',
                                       'longname': 'scaleY',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sy'},
                            'scaleZ': {'args': '<type __builtin__.float>',
                                       'longname': 'scaleZ',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sz'},
                            'translate': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'translate',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 3,
                                          'shortname': 't'},
                            'translateX': {'args': '<type __builtin__.float>',
                                           'longname': 'translateX',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'tx'},
                            'translateY': {'args': '<type __builtin__.float>',
                                           'longname': 'translateY',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ty'},
                            'translateZ': {'args': '<type __builtin__.float>',
                                           'longname': 'translateZ',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'tz'},
                            'weight': {'args': '<type __builtin__.float>',
                                       'longname': 'weight',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'},
                            'worldSpace': {'args': '<type __builtin__.bool>',
                                           'longname': 'worldSpace',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ws'}},
                  'shortFlags': {'att': 'attraction',
                                 'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'dup': 'duplicate',
                                 'e': 'edit',
                                 'fzn': 'frozen',
                                 'g': 'gravity',
                                 'ga': 'gain',
                                 'gx': 'gravityX',
                                 'gy': 'gravityY',
                                 'gz': 'gravityZ',
                                 'kft': 'keepFacesTogether',
                                 'lc': 'localCenter',
                                 'ld': 'localDirection',
                                 'ldx': 'localDirectionX',
                                 'ldy': 'localDirectionY',
                                 'ldz': 'localDirectionZ',
                                 'lr': 'localRotate',
                                 'lrx': 'localRotateX',
                                 'lry': 'localRotateY',
                                 'lrz': 'localRotateZ',
                                 'ls': 'localScale',
                                 'lsx': 'localScaleX',
                                 'lsy': 'localScaleY',
                                 'lsz': 'localScaleZ',
                                 'lt': 'localTranslate',
                                 'ltx': 'localTranslateX',
                                 'lty': 'localTranslateY',
                                 'ltz': 'localTranslateZ',
                                 'm': 'magnet',
                                 'mx': 'magnX',
                                 'my': 'magnY',
                                 'mz': 'magnZ',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'off': 'offset',
                                 'pvt': 'pivot',
                                 'pvx': 'pivotX',
                                 'pvy': 'pivotY',
                                 'pvz': 'pivotZ',
                                 'q': 'query',
                                 'ran': 'random',
                                 'ro': 'rotate',
                                 'rx': 'rotateX',
                                 'ry': 'rotateY',
                                 'rz': 'rotateZ',
                                 's': 'scale',
                                 'sx': 'scaleX',
                                 'sy': 'scaleY',
                                 'sz': 'scaleZ',
                                 't': 'translate',
                                 'tx': 'translateX',
                                 'ty': 'translateY',
                                 'tz': 'translateZ',
                                 'w': 'weight',
                                 'ws': 'worldSpace',
                                 'xft': 'keepFacetTogether'},
                  'type': 'modeling'},
  'polyCircularize': {'flags': {'alignment': {'args': '<type __builtin__.int>',
                                              'longname': 'alignment',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'al'},
                                'caching': {'args': '<type __builtin__.bool>',
                                            'longname': 'caching',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cch'},
                                'constructionHistory': {'args': '<type __builtin__.bool>',
                                                        'longname': 'constructionHistory',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'ch'},
                                'createCurve': {'args': '<type __builtin__.bool>',
                                                'longname': 'createCurve',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'cc'},
                                'divisions': {'args': '<type __builtin__.int>',
                                              'longname': 'divisions',
                                              'numArgs': 1,
                                              'shortname': 'd'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'evenlyDistribute': {'args': '<type __builtin__.bool>',
                                                     'longname': 'evenlyDistribute',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ed'},
                                'frozen': {'args': '<type __builtin__.bool>',
                                           'longname': 'frozen',
                                           'numArgs': 1,
                                           'shortname': 'fzn'},
                                'inputCurve': {'args': 'PyNode',
                                               'longname': 'inputCurve',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'inc'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'nodeState': {'args': '<type __builtin__.int>',
                                              'longname': 'nodeState',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nds'},
                                'normalOffset': {'args': '<type __builtin__.float>',
                                                 'longname': 'normalOffset',
                                                 'numArgs': 1,
                                                 'shortname': 'no'},
                                'normalOrientation': {'args': '<type __builtin__.int>',
                                                      'longname': 'normalOrientation',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'nor'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'radialOffset': {'args': '<type __builtin__.float>',
                                                 'longname': 'radialOffset',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ro'},
                                'relaxInterior': {'args': '<type __builtin__.float>',
                                                  'longname': 'relaxInterior',
                                                  'numArgs': 1,
                                                  'shortname': 'ri'},
                                'smoothingAngle': {'args': '<type __builtin__.float>',
                                                   'longname': 'smoothingAngle',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sa'},
                                'supportingEdges': {'args': '<type __builtin__.int>',
                                                    'longname': 'supportingEdges',
                                                    'numArgs': 1,
                                                    'shortname': 'se'},
                                'twist': {'args': '<type __builtin__.float>',
                                          'longname': 'twist',
                                          'numArgs': 1,
                                          'shortname': 't'},
                                'worldSpace': {'args': '<type __builtin__.bool>',
                                               'longname': 'worldSpace',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ws'}},
                      'shortFlags': {'al': 'alignment',
                                     'cch': 'caching',
                                     'ch': 'constructionHistory',
                                     'd': 'divisions',
                                     'e': 'edit',
                                     'ed': 'evenlyDistribute',
                                     'fzn': 'frozen',
                                     'n': 'name',
                                     'nds': 'nodeState',
                                     'no': 'normalOffset',
                                     'nor': 'normalOrientation',
                                     'q': 'query',
                                     'ri': 'relaxInterior',
                                     'ro': 'radialOffset',
                                     'sa': 'smoothingAngle',
                                     'se': 'supportingEdges',
                                     't': 'twist',
                                     'ws': 'worldSpace'},
                      'type': 'modeling'},
  'polyCircularizeEdge': {'flags': {'alignment': {'args': '<type __builtin__.int>',
                                                  'longname': 'alignment',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'al'},
                                    'caching': {'args': '<type __builtin__.bool>',
                                                'longname': 'caching',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cch'},
                                    'constructionHistory': {'args': '<type __builtin__.bool>',
                                                            'longname': 'constructionHistory',
                                                            'modes': ['create',
                                                                      'query'],
                                                            'numArgs': 1,
                                                            'shortname': 'ch'},
                                    'createCurve': {'args': '<type __builtin__.bool>',
                                                    'longname': 'createCurve',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'cc'},
                                    'divisions': {'args': '<type __builtin__.int>',
                                                  'longname': 'divisions',
                                                  'numArgs': 1,
                                                  'shortname': 'd'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'evenlyDistribute': {'args': '<type __builtin__.bool>',
                                                         'longname': 'evenlyDistribute',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ed'},
                                    'frozen': {'args': '<type __builtin__.bool>',
                                               'longname': 'frozen',
                                               'numArgs': 1,
                                               'shortname': 'fzn'},
                                    'inputCurve': {'args': 'PyNode',
                                                   'longname': 'inputCurve',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'inc'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'nodeState': {'args': '<type __builtin__.int>',
                                                  'longname': 'nodeState',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'nds'},
                                    'normalOffset': {'args': '<type __builtin__.float>',
                                                     'longname': 'normalOffset',
                                                     'numArgs': 1,
                                                     'shortname': 'no'},
                                    'normalOrientation': {'args': '<type __builtin__.int>',
                                                          'longname': 'normalOrientation',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'nor'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'radialOffset': {'args': '<type __builtin__.float>',
                                                     'longname': 'radialOffset',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ro'},
                                    'relaxInterior': {'args': '<type __builtin__.float>',
                                                      'longname': 'relaxInterior',
                                                      'numArgs': 1,
                                                      'shortname': 'ri'},
                                    'smoothingAngle': {'args': '<type __builtin__.float>',
                                                       'longname': 'smoothingAngle',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'sa'},
                                    'supportingEdges': {'args': '<type __builtin__.int>',
                                                        'longname': 'supportingEdges',
                                                        'numArgs': 1,
                                                        'shortname': 'se'},
                                    'twist': {'args': '<type __builtin__.float>',
                                              'longname': 'twist',
                                              'numArgs': 1,
                                              'shortname': 't'},
                                    'worldSpace': {'args': '<type __builtin__.bool>',
                                                   'longname': 'worldSpace',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ws'}},
                          'shortFlags': {'al': 'alignment',
                                         'cch': 'caching',
                                         'ch': 'constructionHistory',
                                         'd': 'divisions',
                                         'e': 'edit',
                                         'ed': 'evenlyDistribute',
                                         'fzn': 'frozen',
                                         'n': 'name',
                                         'nds': 'nodeState',
                                         'no': 'normalOffset',
                                         'nor': 'normalOrientation',
                                         'q': 'query',
                                         'ri': 'relaxInterior',
                                         'ro': 'radialOffset',
                                         'sa': 'smoothingAngle',
                                         'se': 'supportingEdges',
                                         't': 'twist',
                                         'ws': 'worldSpace'},
                          'type': 'modeling'},
  'polyCircularizeFace': {'flags': {'alignment': {'args': '<type __builtin__.int>',
                                                  'longname': 'alignment',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'al'},
                                    'caching': {'args': '<type __builtin__.bool>',
                                                'longname': 'caching',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cch'},
                                    'constructionHistory': {'args': '<type __builtin__.bool>',
                                                            'longname': 'constructionHistory',
                                                            'modes': ['create',
                                                                      'query'],
                                                            'numArgs': 1,
                                                            'shortname': 'ch'},
                                    'createCurve': {'args': '<type __builtin__.bool>',
                                                    'longname': 'createCurve',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'cc'},
                                    'divisions': {'args': '<type __builtin__.int>',
                                                  'longname': 'divisions',
                                                  'numArgs': 1,
                                                  'shortname': 'd'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'evenlyDistribute': {'args': '<type __builtin__.bool>',
                                                         'longname': 'evenlyDistribute',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ed'},
                                    'frozen': {'args': '<type __builtin__.bool>',
                                               'longname': 'frozen',
                                               'numArgs': 1,
                                               'shortname': 'fzn'},
                                    'inputCurve': {'args': 'PyNode',
                                                   'longname': 'inputCurve',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'inc'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'nodeState': {'args': '<type __builtin__.int>',
                                                  'longname': 'nodeState',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'nds'},
                                    'normalOffset': {'args': '<type __builtin__.float>',
                                                     'longname': 'normalOffset',
                                                     'numArgs': 1,
                                                     'shortname': 'no'},
                                    'normalOrientation': {'args': '<type __builtin__.int>',
                                                          'longname': 'normalOrientation',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'nor'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'radialOffset': {'args': '<type __builtin__.float>',
                                                     'longname': 'radialOffset',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ro'},
                                    'relaxInterior': {'args': '<type __builtin__.float>',
                                                      'longname': 'relaxInterior',
                                                      'numArgs': 1,
                                                      'shortname': 'ri'},
                                    'smoothingAngle': {'args': '<type __builtin__.float>',
                                                       'longname': 'smoothingAngle',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'sa'},
                                    'supportingEdges': {'args': '<type __builtin__.int>',
                                                        'longname': 'supportingEdges',
                                                        'numArgs': 1,
                                                        'shortname': 'se'},
                                    'twist': {'args': '<type __builtin__.float>',
                                              'longname': 'twist',
                                              'numArgs': 1,
                                              'shortname': 't'},
                                    'worldSpace': {'args': '<type __builtin__.bool>',
                                                   'longname': 'worldSpace',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ws'}},
                          'shortFlags': {'al': 'alignment',
                                         'cch': 'caching',
                                         'ch': 'constructionHistory',
                                         'd': 'divisions',
                                         'e': 'edit',
                                         'ed': 'evenlyDistribute',
                                         'fzn': 'frozen',
                                         'n': 'name',
                                         'nds': 'nodeState',
                                         'no': 'normalOffset',
                                         'nor': 'normalOrientation',
                                         'q': 'query',
                                         'ri': 'relaxInterior',
                                         'ro': 'radialOffset',
                                         'sa': 'smoothingAngle',
                                         'se': 'supportingEdges',
                                         't': 'twist',
                                         'ws': 'worldSpace'},
                          'type': 'modeling'},
  'polyClean': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                      'longname': 'caching',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'cch'},
                          'cleanEdges': {'args': '<type __builtin__.bool>',
                                         'longname': 'cleanEdges',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ce'},
                          'cleanPartialUVMapping': {'args': '<type __builtin__.bool>',
                                                    'longname': 'cleanPartialUVMapping',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'cpm'},
                          'cleanUVs': {'args': '<type __builtin__.bool>',
                                       'longname': 'cleanUVs',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cuv'},
                          'cleanVertices': {'args': '<type __builtin__.bool>',
                                            'longname': 'cleanVertices',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cv'},
                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'constructionHistory',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'frozen': {'args': '<type __builtin__.bool>',
                                     'longname': 'frozen',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'fzn'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nodeState': {'args': '<type __builtin__.int>',
                                        'longname': 'nodeState',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'nds'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'}},
                'shortFlags': {'cch': 'caching',
                               'ce': 'cleanEdges',
                               'ch': 'constructionHistory',
                               'cpm': 'cleanPartialUVMapping',
                               'cuv': 'cleanUVs',
                               'cv': 'cleanVertices',
                               'e': 'edit',
                               'fzn': 'frozen',
                               'n': 'name',
                               'nds': 'nodeState',
                               'q': 'query'},
                'type': 'modeling'},
  'polyClipboard': {'flags': {'clear': {'args': '<type __builtin__.bool>',
                                        'longname': 'clear',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'cl'},
                              'color': {'args': '<type __builtin__.bool>',
                                        'longname': 'color',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'clr'},
                              'copy': {'args': '<type __builtin__.bool>',
                                       'longname': 'copy',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'cp'},
                              'paste': {'args': '<type __builtin__.bool>',
                                        'longname': 'paste',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ps'},
                              'shader': {'args': '<type __builtin__.bool>',
                                         'longname': 'shader',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'sh'},
                              'uvCoordinates': {'args': '<type __builtin__.bool>',
                                                'longname': 'uvCoordinates',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'uv'}},
                    'shortFlags': {'cl': 'clear',
                                   'clr': 'color',
                                   'cp': 'copy',
                                   'ps': 'paste',
                                   'sh': 'shader',
                                   'uv': 'uvCoordinates'},
                    'type': 'modeling'},
  'polyCloseBorder': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                            'longname': 'caching',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cch'},
                                'constructionHistory': {'args': '<type __builtin__.bool>',
                                                        'longname': 'constructionHistory',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'ch'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'frozen': {'args': '<type __builtin__.bool>',
                                           'longname': 'frozen',
                                           'numArgs': 1,
                                           'shortname': 'fzn'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'nodeState': {'args': '<type __builtin__.int>',
                                              'longname': 'nodeState',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nds'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'}},
                      'shortFlags': {'cch': 'caching',
                                     'ch': 'constructionHistory',
                                     'e': 'edit',
                                     'fzn': 'frozen',
                                     'n': 'name',
                                     'nds': 'nodeState',
                                     'q': 'query'},
                      'type': 'modeling'},
  'polyCollapseEdge': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                             'longname': 'caching',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cch'},
                                 'constructionHistory': {'args': '<type __builtin__.bool>',
                                                         'longname': 'constructionHistory',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'ch'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'frozen': {'args': '<type __builtin__.bool>',
                                            'longname': 'frozen',
                                            'numArgs': 1,
                                            'shortname': 'fzn'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'nodeState': {'args': '<type __builtin__.int>',
                                               'longname': 'nodeState',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nds'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'}},
                       'shortFlags': {'cch': 'caching',
                                      'ch': 'constructionHistory',
                                      'e': 'edit',
                                      'fzn': 'frozen',
                                      'n': 'name',
                                      'nds': 'nodeState',
                                      'q': 'query'},
                       'type': 'modeling'},
  'polyCollapseFacet': {'flags': {'areaThreshold': {'args': '<type __builtin__.float>',
                                                    'longname': 'areaThreshold',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'at'},
                                  'caching': {'args': '<type __builtin__.bool>',
                                              'longname': 'caching',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cch'},
                                  'constructionHistory': {'args': '<type __builtin__.bool>',
                                                          'longname': 'constructionHistory',
                                                          'modes': ['create',
                                                                    'query'],
                                                          'numArgs': 1,
                                                          'shortname': 'ch'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'frozen': {'args': '<type __builtin__.bool>',
                                             'longname': 'frozen',
                                             'numArgs': 1,
                                             'shortname': 'fzn'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'nodeState': {'args': '<type __builtin__.int>',
                                                'longname': 'nodeState',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nds'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'useAreaThreshold': {'args': '<type __builtin__.bool>',
                                                       'longname': 'useAreaThreshold',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'uat'}},
                        'shortFlags': {'at': 'areaThreshold',
                                       'cch': 'caching',
                                       'ch': 'constructionHistory',
                                       'e': 'edit',
                                       'fzn': 'frozen',
                                       'n': 'name',
                                       'nds': 'nodeState',
                                       'q': 'query',
                                       'uat': 'useAreaThreshold'},
                        'type': 'modeling'},
  'polyCollapseTweaks': {'flags': {'hasVertexTweaks': {'args': '<type __builtin__.bool>',
                                                       'longname': 'hasVertexTweaks',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 0,
                                                       'shortname': 'hvt'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'}},
                         'shortFlags': {'hvt': 'hasVertexTweaks',
                                        'q': 'query'},
                         'type': 'modeling'},
  'polyColorBlindData': {'flags': {'aboveMaxColorBlue': {'args': '<type __builtin__.float>',
                                                         'longname': 'aboveMaxColorBlue',
                                                         'modes': ['create',
                                                                   'multiuse'],
                                                         'numArgs': 1,
                                                         'shortname': 'amb'},
                                   'aboveMaxColorGreen': {'args': '<type __builtin__.float>',
                                                          'longname': 'aboveMaxColorGreen',
                                                          'modes': ['create',
                                                                    'multiuse'],
                                                          'numArgs': 1,
                                                          'shortname': 'amg'},
                                   'aboveMaxColorRed': {'args': '<type __builtin__.float>',
                                                        'longname': 'aboveMaxColorRed',
                                                        'modes': ['create',
                                                                  'multiuse'],
                                                        'numArgs': 1,
                                                        'shortname': 'amr'},
                                   'attrName': {'args': '<type __builtin__.str>',
                                                'longname': 'attrName',
                                                'modes': ['create',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'n'},
                                   'belowMinColorBlue': {'args': '<type __builtin__.float>',
                                                         'longname': 'belowMinColorBlue',
                                                         'modes': ['create',
                                                                   'multiuse'],
                                                         'numArgs': 1,
                                                         'shortname': 'bmb'},
                                   'belowMinColorGreen': {'args': '<type __builtin__.float>',
                                                          'longname': 'belowMinColorGreen',
                                                          'modes': ['create',
                                                                    'multiuse'],
                                                          'numArgs': 1,
                                                          'shortname': 'bmg'},
                                   'belowMinColorRed': {'args': '<type __builtin__.float>',
                                                        'longname': 'belowMinColorRed',
                                                        'modes': ['create',
                                                                  'multiuse'],
                                                        'numArgs': 1,
                                                        'shortname': 'bmr'},
                                   'clashColorBlue': {'args': '<type __builtin__.float>',
                                                      'longname': 'clashColorBlue',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ccb'},
                                   'clashColorGreen': {'args': '<type __builtin__.float>',
                                                       'longname': 'clashColorGreen',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ccg'},
                                   'clashColorRed': {'args': '<type __builtin__.float>',
                                                     'longname': 'clashColorRed',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ccr'},
                                   'colorBlue': {'args': '<type __builtin__.float>',
                                                 'longname': 'colorBlue',
                                                 'modes': ['create',
                                                           'multiuse'],
                                                 'numArgs': 1,
                                                 'shortname': 'cb'},
                                   'colorGreen': {'args': '<type __builtin__.float>',
                                                  'longname': 'colorGreen',
                                                  'modes': ['create',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'cg'},
                                   'colorRed': {'args': '<type __builtin__.float>',
                                                'longname': 'colorRed',
                                                'modes': ['create',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'cr'},
                                   'dataType': {'args': '<type __builtin__.str>',
                                                'longname': 'dataType',
                                                'modes': ['create',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                                   'enableFalseColor': {'args': '<type __builtin__.bool>',
                                                        'longname': 'enableFalseColor',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'efc'},
                                   'maxColorBlue': {'args': '<type __builtin__.float>',
                                                    'longname': 'maxColorBlue',
                                                    'modes': ['create',
                                                              'multiuse'],
                                                    'numArgs': 1,
                                                    'shortname': 'mxb'},
                                   'maxColorGreen': {'args': '<type __builtin__.float>',
                                                     'longname': 'maxColorGreen',
                                                     'modes': ['create',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'mxg'},
                                   'maxColorRed': {'args': '<type __builtin__.float>',
                                                   'longname': 'maxColorRed',
                                                   'modes': ['create',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'mxr'},
                                   'maxValue': {'args': '<type __builtin__.float>',
                                                'longname': 'maxValue',
                                                'modes': ['create',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'mxv'},
                                   'minColorBlue': {'args': '<type __builtin__.float>',
                                                    'longname': 'minColorBlue',
                                                    'modes': ['create',
                                                              'multiuse'],
                                                    'numArgs': 1,
                                                    'shortname': 'mnb'},
                                   'minColorGreen': {'args': '<type __builtin__.float>',
                                                     'longname': 'minColorGreen',
                                                     'modes': ['create',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'mng'},
                                   'minColorRed': {'args': '<type __builtin__.float>',
                                                   'longname': 'minColorRed',
                                                   'modes': ['create',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'mnr'},
                                   'minValue': {'args': '<type __builtin__.float>',
                                                'longname': 'minValue',
                                                'modes': ['create',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'mnv'},
                                   'mode': {'args': '<type __builtin__.int>',
                                            'longname': 'mode',
                                            'modes': ['create',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'm'},
                                   'noColorBlue': {'args': '<type __builtin__.float>',
                                                   'longname': 'noColorBlue',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ncb'},
                                   'noColorGreen': {'args': '<type __builtin__.float>',
                                                    'longname': 'noColorGreen',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ncg'},
                                   'noColorRed': {'args': '<type __builtin__.float>',
                                                  'longname': 'noColorRed',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ncr'},
                                   'numIdTypes': {'args': '<type __builtin__.int>',
                                                  'longname': 'numIdTypes',
                                                  'modes': ['create',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'num'},
                                   'queryMode': {'args': '<type __builtin__.bool>',
                                                 'longname': 'queryMode',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'q'},
                                   'typeId': {'args': '<type __builtin__.int>',
                                              'longname': 'typeId',
                                              'modes': ['create',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'id'},
                                   'useMax': {'args': '<type __builtin__.bool>',
                                              'longname': 'useMax',
                                              'modes': ['create',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'umx'},
                                   'useMin': {'args': '<type __builtin__.bool>',
                                              'longname': 'useMin',
                                              'modes': ['create',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'umn'},
                                   'value': {'args': '<type __builtin__.str>',
                                             'longname': 'value',
                                             'modes': ['create',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'v'}},
                         'shortFlags': {'amb': 'aboveMaxColorBlue',
                                        'amg': 'aboveMaxColorGreen',
                                        'amr': 'aboveMaxColorRed',
                                        'bmb': 'belowMinColorBlue',
                                        'bmg': 'belowMinColorGreen',
                                        'bmr': 'belowMinColorRed',
                                        'cb': 'colorBlue',
                                        'ccb': 'clashColorBlue',
                                        'ccg': 'clashColorGreen',
                                        'ccr': 'clashColorRed',
                                        'cg': 'colorGreen',
                                        'cr': 'colorRed',
                                        'dt': 'dataType',
                                        'efc': 'enableFalseColor',
                                        'id': 'typeId',
                                        'm': 'mode',
                                        'mnb': 'minColorBlue',
                                        'mng': 'minColorGreen',
                                        'mnr': 'minColorRed',
                                        'mnv': 'minValue',
                                        'mxb': 'maxColorBlue',
                                        'mxg': 'maxColorGreen',
                                        'mxr': 'maxColorRed',
                                        'mxv': 'maxValue',
                                        'n': 'attrName',
                                        'ncb': 'noColorBlue',
                                        'ncg': 'noColorGreen',
                                        'ncr': 'noColorRed',
                                        'num': 'numIdTypes',
                                        'q': 'queryMode',
                                        'umn': 'useMin',
                                        'umx': 'useMax',
                                        'v': 'value'},
                         'type': 'modeling'},
  'polyColorDel': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'colorSetName': {'args': '<type __builtin__.str>',
                                              'longname': 'colorSetName',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cls'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'cls': 'colorSetName',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'q': 'query'},
                   'type': 'modeling'},
  'polyColorMod': {'flags': {'alphaScale_FloatValue': {'args': '<type __builtin__.float>',
                                                       'longname': 'alphaScale_FloatValue',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'afv'},
                             'alphaScale_Interp': {'args': '<type __builtin__.int>',
                                                   'longname': 'alphaScale_Interp',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ai'},
                             'alphaScale_Position': {'args': '<type __builtin__.float>',
                                                     'longname': 'alphaScale_Position',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ap'},
                             'baseColorName': {'args': '<type __builtin__.str>',
                                               'longname': 'baseColorName',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'bcn'},
                             'blueScale_FloatValue': {'args': '<type __builtin__.float>',
                                                      'longname': 'blueScale_FloatValue',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'bfv'},
                             'blueScale_Interp': {'args': '<type __builtin__.int>',
                                                  'longname': 'blueScale_Interp',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'bi'},
                             'blueScale_Position': {'args': '<type __builtin__.float>',
                                                    'longname': 'blueScale_Position',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'bp'},
                             'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'greenScale_FloatValue': {'args': '<type __builtin__.float>',
                                                       'longname': 'greenScale_FloatValue',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'gfv'},
                             'greenScale_Interp': {'args': '<type __builtin__.int>',
                                                   'longname': 'greenScale_Interp',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'gi'},
                             'greenScale_Position': {'args': '<type __builtin__.float>',
                                                     'longname': 'greenScale_Position',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'gp'},
                             'huev': {'args': '<type __builtin__.float>',
                                      'longname': 'huev',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                             'intensityScale_FloatValue': {'args': '<type __builtin__.float>',
                                                           'longname': 'intensityScale_FloatValue',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'nfv'},
                             'intensityScale_Interp': {'args': '<type __builtin__.int>',
                                                       'longname': 'intensityScale_Interp',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ni'},
                             'intensityScale_Position': {'args': '<type __builtin__.float>',
                                                         'longname': 'intensityScale_Position',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'np'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'redScale_FloatValue': {'args': '<type __builtin__.float>',
                                                     'longname': 'redScale_FloatValue',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'rfv'},
                             'redScale_Interp': {'args': '<type __builtin__.int>',
                                                 'longname': 'redScale_Interp',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ri'},
                             'redScale_Position': {'args': '<type __builtin__.float>',
                                                   'longname': 'redScale_Position',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'rp'},
                             'satv': {'args': '<type __builtin__.float>',
                                      'longname': 'satv',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                             'value': {'args': '<type __builtin__.float>',
                                       'longname': 'value',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'v'}},
                   'shortFlags': {'afv': 'alphaScale_FloatValue',
                                  'ai': 'alphaScale_Interp',
                                  'ap': 'alphaScale_Position',
                                  'bcn': 'baseColorName',
                                  'bfv': 'blueScale_FloatValue',
                                  'bi': 'blueScale_Interp',
                                  'bp': 'blueScale_Position',
                                  'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'gfv': 'greenScale_FloatValue',
                                  'gi': 'greenScale_Interp',
                                  'gp': 'greenScale_Position',
                                  'h': 'huev',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'nfv': 'intensityScale_FloatValue',
                                  'ni': 'intensityScale_Interp',
                                  'np': 'intensityScale_Position',
                                  'q': 'query',
                                  'rfv': 'redScale_FloatValue',
                                  'ri': 'redScale_Interp',
                                  'rp': 'redScale_Position',
                                  's': 'satv',
                                  'v': 'value'},
                   'type': 'modeling'},
  'polyColorPerVertex': {'flags': {'alpha': {'args': '<type __builtin__.float>',
                                             'longname': 'alpha',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'a'},
                                   'clamped': {'args': '<type __builtin__.bool>',
                                               'longname': 'clamped',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cla'},
                                   'colorB': {'args': '<type __builtin__.float>',
                                              'longname': 'colorB',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'b'},
                                   'colorDisplayOption': {'args': '<type __builtin__.bool>',
                                                          'longname': 'colorDisplayOption',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 0,
                                                          'shortname': 'cdo'},
                                   'colorG': {'args': '<type __builtin__.float>',
                                              'longname': 'colorG',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'g'},
                                   'colorR': {'args': '<type __builtin__.float>',
                                              'longname': 'colorR',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'r'},
                                   'colorRGB': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'colorRGB',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'rgb'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'notUndoable': {'args': '<type __builtin__.bool>',
                                                   'longname': 'notUndoable',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'nun'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'relative': {'args': '<type __builtin__.bool>',
                                                'longname': 'relative',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'rel'},
                                   'remove': {'args': '<type __builtin__.bool>',
                                              'longname': 'remove',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 0,
                                              'shortname': 'rem'},
                                   'representation': {'args': '<type __builtin__.int>',
                                                      'longname': 'representation',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'rpt'}},
                         'shortFlags': {'a': 'alpha',
                                        'b': 'colorB',
                                        'cdo': 'colorDisplayOption',
                                        'cla': 'clamped',
                                        'e': 'edit',
                                        'g': 'colorG',
                                        'nun': 'notUndoable',
                                        'q': 'query',
                                        'r': 'colorR',
                                        'rel': 'relative',
                                        'rem': 'remove',
                                        'rgb': 'colorRGB',
                                        'rpt': 'representation'},
                         'type': 'modeling'},
  'polyColorSet': {'flags': {'allColorSets': {'args': '<type __builtin__.bool>',
                                              'longname': 'allColorSets',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 0,
                                              'shortname': 'acs'},
                             'clamped': {'args': '<type __builtin__.bool>',
                                         'longname': 'clamped',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cla'},
                             'colorSet': {'args': '<type __builtin__.str>',
                                          'longname': 'colorSet',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cs'},
                             'copy': {'args': '<type __builtin__.bool>',
                                      'longname': 'copy',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'cp'},
                             'create': {'args': '<type __builtin__.bool>',
                                        'longname': 'create',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 0,
                                        'shortname': 'cr'},
                             'currentColorSet': {'args': '<type __builtin__.bool>',
                                                 'longname': 'currentColorSet',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'ccs'},
                             'currentPerInstanceSet': {'args': '<type __builtin__.bool>',
                                                       'longname': 'currentPerInstanceSet',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'cpi'},
                             'delete': {'args': '<type __builtin__.bool>',
                                        'longname': 'delete',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 0,
                                        'shortname': 'd'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'newColorSet': {'args': '<type __builtin__.str>',
                                             'longname': 'newColorSet',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nc'},
                             'perInstance': {'args': '<type __builtin__.bool>',
                                             'longname': 'perInstance',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'pi'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'rename': {'args': '<type __builtin__.bool>',
                                        'longname': 'rename',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 0,
                                        'shortname': 'rn'},
                             'representation': {'args': '<type __builtin__.str>',
                                                'longname': 'representation',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'rpt'},
                             'shareInstances': {'args': '<type __builtin__.bool>',
                                                'longname': 'shareInstances',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'si'},
                             'unshared': {'args': '<type __builtin__.bool>',
                                          'longname': 'unshared',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'us'}},
                   'shortFlags': {'acs': 'allColorSets',
                                  'ccs': 'currentColorSet',
                                  'cla': 'clamped',
                                  'cp': 'copy',
                                  'cpi': 'currentPerInstanceSet',
                                  'cr': 'create',
                                  'cs': 'colorSet',
                                  'd': 'delete',
                                  'e': 'edit',
                                  'nc': 'newColorSet',
                                  'pi': 'perInstance',
                                  'q': 'query',
                                  'rn': 'rename',
                                  'rpt': 'representation',
                                  'si': 'shareInstances',
                                  'us': 'unshared'},
                   'type': 'modeling'},
  'polyColorSetCmdWrapper': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'polyCompare': {'flags': {'colorSetIndices': {'args': '<type __builtin__.bool>',
                                                'longname': 'colorSetIndices',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ic'},
                            'colorSets': {'args': '<type __builtin__.bool>',
                                          'longname': 'colorSets',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'c'},
                            'edges': {'args': '<type __builtin__.bool>',
                                      'longname': 'edges',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'e'},
                            'faceDesc': {'args': '<type __builtin__.bool>',
                                         'longname': 'faceDesc',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'fd'},
                            'userNormals': {'args': '<type __builtin__.bool>',
                                            'longname': 'userNormals',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'un'},
                            'uvSetIndices': {'args': '<type __builtin__.bool>',
                                             'longname': 'uvSetIndices',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'iuv'},
                            'uvSets': {'args': '<type __builtin__.bool>',
                                       'longname': 'uvSets',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'uv'},
                            'vertices': {'args': '<type __builtin__.bool>',
                                         'longname': 'vertices',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'v'}},
                  'shortFlags': {'c': 'colorSets',
                                 'e': 'edges',
                                 'fd': 'faceDesc',
                                 'ic': 'colorSetIndices',
                                 'iuv': 'uvSetIndices',
                                 'un': 'userNormals',
                                 'uv': 'uvSets',
                                 'v': 'vertices'},
                  'type': 'modeling'},
  'polyCone': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                           '<type __builtin__.float>',
                                           '<type __builtin__.float>'],
                                  'longname': 'axis',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 3,
                                  'shortname': 'ax'},
                         'caching': {'args': '<type __builtin__.bool>',
                                     'longname': 'caching',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsCasting': True,
                                     'shortname': 'cch'},
                         'constructionHistory': {'args': '<type __builtin__.bool>',
                                                 'longname': 'constructionHistory',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                         'createUVs': {'args': '<type __builtin__.int>',
                                       'longname': 'createUVs',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'cuv'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'frozen': {'args': '<type __builtin__.bool>',
                                    'longname': 'frozen',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'fzn'},
                         'height': {'args': '<type __builtin__.float>',
                                    'longname': 'height',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'h'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'nodeState': {'args': '<type __builtin__.int>',
                                       'longname': 'nodeState',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'nds'},
                         'object': {'args': '<type __builtin__.bool>',
                                    'longname': 'object',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'o'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'radius': {'args': '<type __builtin__.float>',
                                    'longname': 'radius',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'r'},
                         'roundCap': {'args': '<type __builtin__.bool>',
                                      'longname': 'roundCap',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsCasting': True,
                                      'shortname': 'rcp'},
                         'subdivisionsAxis': {'args': '<type __builtin__.int>',
                                              'longname': 'subdivisionsAxis',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'resultNeedsCasting': True,
                                              'shortname': 'sa'},
                         'subdivisionsCap': {'args': '<type __builtin__.int>',
                                             'longname': 'subdivisionsCap',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'resultNeedsCasting': True,
                                             'shortname': 'sc'},
                         'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                'longname': 'subdivisionsHeight',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'resultNeedsCasting': True,
                                                'shortname': 'sh'},
                         'subdivisionsX': {'args': '<type __builtin__.int>',
                                           'longname': 'subdivisionsX',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'resultNeedsCasting': True,
                                           'shortname': 'sx'},
                         'subdivisionsY': {'args': '<type __builtin__.int>',
                                           'longname': 'subdivisionsY',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'resultNeedsCasting': True,
                                           'shortname': 'sy'},
                         'subdivisionsZ': {'args': '<type __builtin__.int>',
                                           'longname': 'subdivisionsZ',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'resultNeedsCasting': True,
                                           'shortname': 'sz'},
                         'texture': {'args': '<type __builtin__.bool>',
                                     'longname': 'texture',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsCasting': True,
                                     'shortname': 'tx'}},
               'shortFlags': {'ax': 'axis',
                              'cch': 'caching',
                              'ch': 'constructionHistory',
                              'cuv': 'createUVs',
                              'e': 'edit',
                              'fzn': 'frozen',
                              'h': 'height',
                              'n': 'name',
                              'nds': 'nodeState',
                              'o': 'object',
                              'q': 'query',
                              'r': 'radius',
                              'rcp': 'roundCap',
                              'sa': 'subdivisionsAxis',
                              'sc': 'subdivisionsCap',
                              'sh': 'subdivisionsHeight',
                              'sx': 'subdivisionsX',
                              'sy': 'subdivisionsY',
                              'sz': 'subdivisionsZ',
                              'tx': 'texture'},
               'type': 'modeling'},
  'polyConnectComponents': {'flags': {'adjustEdgeFlow': {'args': '<type __builtin__.float>',
                                                         'longname': 'adjustEdgeFlow',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'aef'},
                                      'caching': {'args': '<type __builtin__.bool>',
                                                  'longname': 'caching',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cch'},
                                      'constructionHistory': {'args': '<type __builtin__.bool>',
                                                              'longname': 'constructionHistory',
                                                              'modes': ['create',
                                                                        'query'],
                                                              'numArgs': 1,
                                                              'shortname': 'ch'},
                                      'edit': {'args': '<type __builtin__.bool>',
                                               'longname': 'edit',
                                               'numArgs': 0,
                                               'shortname': 'e'},
                                      'frozen': {'args': '<type __builtin__.bool>',
                                                 'longname': 'frozen',
                                                 'numArgs': 1,
                                                 'shortname': 'fzn'},
                                      'insertWithEdgeFlow': {'args': '<type __builtin__.bool>',
                                                             'longname': 'insertWithEdgeFlow',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'ief'},
                                      'name': {'args': '<type __builtin__.str>',
                                               'longname': 'name',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'n'},
                                      'nodeState': {'args': '<type __builtin__.int>',
                                                    'longname': 'nodeState',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'nds'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'}},
                            'shortFlags': {'aef': 'adjustEdgeFlow',
                                           'cch': 'caching',
                                           'ch': 'constructionHistory',
                                           'e': 'edit',
                                           'fzn': 'frozen',
                                           'ief': 'insertWithEdgeFlow',
                                           'n': 'name',
                                           'nds': 'nodeState',
                                           'q': 'query'},
                            'type': 'modeling'},
  'polyContourProjection': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                                  'longname': 'caching',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cch'},
                                      'constructionHistory': {'args': '<type __builtin__.bool>',
                                                              'longname': 'constructionHistory',
                                                              'modes': ['create',
                                                                        'query'],
                                                              'numArgs': 1,
                                                              'shortname': 'ch'},
                                      'createNewMap': {'args': '<type __builtin__.bool>',
                                                       'longname': 'createNewMap',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'cm'},
                                      'edit': {'args': '<type __builtin__.bool>',
                                               'longname': 'edit',
                                               'numArgs': 0,
                                               'shortname': 'e'},
                                      'flipRails': {'args': '<type __builtin__.bool>',
                                                    'longname': 'flipRails',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'fr'},
                                      'frozen': {'args': '<type __builtin__.bool>',
                                                 'longname': 'frozen',
                                                 'numArgs': 1,
                                                 'shortname': 'fzn'},
                                      'insertBeforeDeformers': {'args': '<type __builtin__.bool>',
                                                                'longname': 'insertBeforeDeformers',
                                                                'modes': ['create'],
                                                                'numArgs': 1,
                                                                'shortname': 'ibd'},
                                      'method': {'args': '<type __builtin__.int>',
                                                 'longname': 'method',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'm'},
                                      'name': {'args': '<type __builtin__.str>',
                                               'longname': 'name',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'n'},
                                      'nodeState': {'args': '<type __builtin__.int>',
                                                    'longname': 'nodeState',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'nds'},
                                      'offset0': {'args': '<type __builtin__.float>',
                                                  'longname': 'offset0',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'o0'},
                                      'offset1': {'args': '<type __builtin__.float>',
                                                  'longname': 'offset1',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'o1'},
                                      'offset2': {'args': '<type __builtin__.float>',
                                                  'longname': 'offset2',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'o2'},
                                      'offset3': {'args': '<type __builtin__.float>',
                                                  'longname': 'offset3',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'o3'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'reduceShear': {'args': '<type __builtin__.float>',
                                                      'longname': 'reduceShear',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'rs'},
                                      'smoothness0': {'args': '<type __builtin__.float>',
                                                      'longname': 'smoothness0',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 's0'},
                                      'smoothness1': {'args': '<type __builtin__.float>',
                                                      'longname': 'smoothness1',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 's1'},
                                      'smoothness2': {'args': '<type __builtin__.float>',
                                                      'longname': 'smoothness2',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 's2'},
                                      'smoothness3': {'args': '<type __builtin__.float>',
                                                      'longname': 'smoothness3',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 's3'},
                                      'userDefinedCorners': {'args': '<type __builtin__.bool>',
                                                             'longname': 'userDefinedCorners',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'udc'},
                                      'uvSetName': {'args': '<type __builtin__.str>',
                                                    'longname': 'uvSetName',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'uvs'},
                                      'worldSpace': {'args': '<type __builtin__.bool>',
                                                     'longname': 'worldSpace',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ws'}},
                            'shortFlags': {'cch': 'caching',
                                           'ch': 'constructionHistory',
                                           'cm': 'createNewMap',
                                           'e': 'edit',
                                           'fr': 'flipRails',
                                           'fzn': 'frozen',
                                           'ibd': 'insertBeforeDeformers',
                                           'm': 'method',
                                           'n': 'name',
                                           'nds': 'nodeState',
                                           'o0': 'offset0',
                                           'o1': 'offset1',
                                           'o2': 'offset2',
                                           'o3': 'offset3',
                                           'q': 'query',
                                           'rs': 'reduceShear',
                                           's0': 'smoothness0',
                                           's1': 'smoothness1',
                                           's2': 'smoothness2',
                                           's3': 'smoothness3',
                                           'udc': 'userDefinedCorners',
                                           'uvs': 'uvSetName',
                                           'ws': 'worldSpace'},
                            'type': 'modeling'},
  'polyCopyUV': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'createNewMap': {'args': '<type __builtin__.bool>',
                                            'longname': 'createNewMap',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'cm'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'uvSetName': {'args': '<type __builtin__.str>',
                                         'longname': 'uvSetName',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'uvs'},
                           'uvSetNameInput': {'args': '<type __builtin__.str>',
                                              'longname': 'uvSetNameInput',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'uvi'},
                           'worldSpace': {'args': '<type __builtin__.bool>',
                                          'longname': 'worldSpace',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ws'}},
                 'shortFlags': {'cch': 'caching',
                                'ch': 'constructionHistory',
                                'cm': 'createNewMap',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'n': 'name',
                                'nds': 'nodeState',
                                'q': 'query',
                                'uvi': 'uvSetNameInput',
                                'uvs': 'uvSetName',
                                'ws': 'worldSpace'},
                 'type': 'modeling'},
  'polyCrease': {'flags': {'createHistory': {'args': '<type __builtin__.bool>',
                                             'longname': 'createHistory',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ch'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'operation': {'args': '<type __builtin__.int>',
                                         'longname': 'operation',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'op'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'relativeValue': {'args': '<type __builtin__.float>',
                                             'longname': 'relativeValue',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'rv'},
                           'value': {'args': '<type __builtin__.float>',
                                     'longname': 'value',
                                     'modes': ['create',
                                               'query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'v'},
                           'vertexValue': {'args': '<type __builtin__.float>',
                                           'longname': 'vertexValue',
                                           'modes': ['create',
                                                     'query',
                                                     'edit',
                                                     'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'vv'}},
                 'shortFlags': {'ch': 'createHistory',
                                'e': 'edit',
                                'op': 'operation',
                                'q': 'query',
                                'rv': 'relativeValue',
                                'v': 'value',
                                'vv': 'vertexValue'},
                 'type': 'modeling'},
  'polyCreaseCtx': {'flags': {'createSet': {'args': '<type __builtin__.str>',
                                            'longname': 'createSet',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'cs'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'extendSelection': {'args': '<type __builtin__.bool>',
                                                  'longname': 'extendSelection',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'es'},
                              'image1': {'args': '<type __builtin__.str>',
                                         'longname': 'image1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i1'},
                              'image2': {'args': '<type __builtin__.str>',
                                         'longname': 'image2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i2'},
                              'image3': {'args': '<type __builtin__.str>',
                                         'longname': 'image3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i3'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'relative': {'args': '<type __builtin__.bool>',
                                           'longname': 'relative',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'r'}},
                    'shortFlags': {'cs': 'createSet',
                                   'e': 'edit',
                                   'es': 'extendSelection',
                                   'ex': 'exists',
                                   'i1': 'image1',
                                   'i2': 'image2',
                                   'i3': 'image3',
                                   'q': 'query',
                                   'r': 'relative'},
                    'type': 'context'},
  'polyCreateFacet': {'flags': {'constructionHistory': {'args': '<type __builtin__.bool>',
                                                        'longname': 'constructionHistory',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'ch'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'hole': {'args': '<type __builtin__.bool>',
                                         'longname': 'hole',
                                         'modes': ['create', 'multiuse'],
                                         'numArgs': 0,
                                         'shortname': 'hl'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'point': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'point',
                                          'modes': ['create', 'multiuse'],
                                          'numArgs': 3,
                                          'shortname': 'p'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'subdivision': {'args': '<type __builtin__.int>',
                                                'longname': 'subdivision',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 's'},
                                'texture': {'args': '<type __builtin__.int>',
                                            'longname': 'texture',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'tx'}},
                      'shortFlags': {'ch': 'constructionHistory',
                                     'e': 'edit',
                                     'hl': 'hole',
                                     'n': 'name',
                                     'p': 'point',
                                     'q': 'query',
                                     's': 'subdivision',
                                     'tx': 'texture'},
                      'type': 'modeling'},
  'polyCreateFacetCtx': {'flags': {'append': {'args': '<type __builtin__.bool>',
                                              'longname': 'append',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ap'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'maximumNumberOfPoints': {'args': '<type __builtin__.int>',
                                                             'longname': 'maximumNumberOfPoints',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'mp'},
                                   'planarConstraint': {'args': '<type __builtin__.bool>',
                                                        'longname': 'planarConstraint',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'pc'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'subdivision': {'args': '<type __builtin__.int>',
                                                   'longname': 'subdivision',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 's'},
                                   'texture': {'args': '<type __builtin__.int>',
                                               'longname': 'texture',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tx'}},
                         'shortFlags': {'ap': 'append',
                                        'e': 'edit',
                                        'ex': 'exists',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'mp': 'maximumNumberOfPoints',
                                        'pc': 'planarConstraint',
                                        'q': 'query',
                                        's': 'subdivision',
                                        'tx': 'texture'},
                         'type': 'context'},
  'polyCube': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                           '<type __builtin__.float>',
                                           '<type __builtin__.float>'],
                                  'longname': 'axis',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 3,
                                  'shortname': 'ax'},
                         'caching': {'args': '<type __builtin__.bool>',
                                     'longname': 'caching',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsCasting': True,
                                     'shortname': 'cch'},
                         'constructionHistory': {'args': '<type __builtin__.bool>',
                                                 'longname': 'constructionHistory',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                         'createUVs': {'args': '<type __builtin__.int>',
                                       'longname': 'createUVs',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'cuv'},
                         'depth': {'args': '<type __builtin__.float>',
                                   'longname': 'depth',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'd'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'frozen': {'args': '<type __builtin__.bool>',
                                    'longname': 'frozen',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'fzn'},
                         'height': {'args': '<type __builtin__.float>',
                                    'longname': 'height',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'h'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'nodeState': {'args': '<type __builtin__.int>',
                                       'longname': 'nodeState',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'nds'},
                         'object': {'args': '<type __builtin__.bool>',
                                    'longname': 'object',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'o'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'subdivisionsDepth': {'args': '<type __builtin__.int>',
                                               'longname': 'subdivisionsDepth',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'resultNeedsCasting': True,
                                               'shortname': 'sd'},
                         'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                'longname': 'subdivisionsHeight',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'resultNeedsCasting': True,
                                                'shortname': 'sh'},
                         'subdivisionsWidth': {'args': '<type __builtin__.int>',
                                               'longname': 'subdivisionsWidth',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'resultNeedsCasting': True,
                                               'shortname': 'sw'},
                         'subdivisionsX': {'args': '<type __builtin__.int>',
                                           'longname': 'subdivisionsX',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'resultNeedsCasting': True,
                                           'shortname': 'sx'},
                         'subdivisionsY': {'args': '<type __builtin__.int>',
                                           'longname': 'subdivisionsY',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'resultNeedsCasting': True,
                                           'shortname': 'sy'},
                         'subdivisionsZ': {'args': '<type __builtin__.int>',
                                           'longname': 'subdivisionsZ',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'resultNeedsCasting': True,
                                           'shortname': 'sz'},
                         'texture': {'args': '<type __builtin__.int>',
                                     'longname': 'texture',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsCasting': True,
                                     'shortname': 'tx'},
                         'width': {'args': '<type __builtin__.float>',
                                   'longname': 'width',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'w'}},
               'shortFlags': {'ax': 'axis',
                              'cch': 'caching',
                              'ch': 'constructionHistory',
                              'cuv': 'createUVs',
                              'd': 'depth',
                              'e': 'edit',
                              'fzn': 'frozen',
                              'h': 'height',
                              'n': 'name',
                              'nds': 'nodeState',
                              'o': 'object',
                              'q': 'query',
                              'sd': 'subdivisionsDepth',
                              'sh': 'subdivisionsHeight',
                              'sw': 'subdivisionsWidth',
                              'sx': 'subdivisionsX',
                              'sy': 'subdivisionsY',
                              'sz': 'subdivisionsZ',
                              'tx': 'texture',
                              'w': 'width'},
               'type': 'modeling'},
  'polyCut': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                    'longname': 'caching',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'cch'},
                        'constructionHistory': {'args': '<type __builtin__.bool>',
                                                'longname': 'constructionHistory',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                        'cutPlaneCenter': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'cutPlaneCenter',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'pc'},
                        'cutPlaneCenterX': {'args': '<type __builtin__.float>',
                                            'longname': 'cutPlaneCenterX',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pcx'},
                        'cutPlaneCenterY': {'args': '<type __builtin__.float>',
                                            'longname': 'cutPlaneCenterY',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pcy'},
                        'cutPlaneCenterZ': {'args': '<type __builtin__.float>',
                                            'longname': 'cutPlaneCenterZ',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pcz'},
                        'cutPlaneHeight': {'args': '<type __builtin__.float>',
                                           'longname': 'cutPlaneHeight',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ph'},
                        'cutPlaneRotate': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'cutPlaneRotate',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'ro'},
                        'cutPlaneRotateX': {'args': '<type __builtin__.float>',
                                            'longname': 'cutPlaneRotateX',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'rx'},
                        'cutPlaneRotateY': {'args': '<type __builtin__.float>',
                                            'longname': 'cutPlaneRotateY',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ry'},
                        'cutPlaneRotateZ': {'args': '<type __builtin__.float>',
                                            'longname': 'cutPlaneRotateZ',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'rz'},
                        'cutPlaneSize': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'cutPlaneSize',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 2,
                                         'shortname': 'ps'},
                        'cutPlaneWidth': {'args': '<type __builtin__.float>',
                                          'longname': 'cutPlaneWidth',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pw'},
                        'cuttingDirection': {'args': '<type __builtin__.str>',
                                             'longname': 'cuttingDirection',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'cd'},
                        'deleteFaces': {'args': '<type __builtin__.bool>',
                                        'longname': 'deleteFaces',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'df'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'extractFaces': {'args': '<type __builtin__.bool>',
                                         'longname': 'extractFaces',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ef'},
                        'extractOffset': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'extractOffset',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 3,
                                          'shortname': 'eo'},
                        'extractOffsetX': {'args': '<type __builtin__.float>',
                                           'longname': 'extractOffsetX',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'eox'},
                        'extractOffsetY': {'args': '<type __builtin__.float>',
                                           'longname': 'extractOffsetY',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'eoy'},
                        'extractOffsetZ': {'args': '<type __builtin__.float>',
                                           'longname': 'extractOffsetZ',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'eoz'},
                        'frozen': {'args': '<type __builtin__.bool>',
                                   'longname': 'frozen',
                                   'numArgs': 1,
                                   'shortname': 'fzn'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'nodeState': {'args': '<type __builtin__.int>',
                                      'longname': 'nodeState',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'nds'},
                        'onObject': {'args': '<type __builtin__.bool>',
                                     'longname': 'onObject',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'oo'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'worldSpace': {'args': '<type __builtin__.bool>',
                                       'longname': 'worldSpace',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ws'}},
              'shortFlags': {'cch': 'caching',
                             'cd': 'cuttingDirection',
                             'ch': 'constructionHistory',
                             'df': 'deleteFaces',
                             'e': 'edit',
                             'ef': 'extractFaces',
                             'eo': 'extractOffset',
                             'eox': 'extractOffsetX',
                             'eoy': 'extractOffsetY',
                             'eoz': 'extractOffsetZ',
                             'fzn': 'frozen',
                             'n': 'name',
                             'nds': 'nodeState',
                             'oo': 'onObject',
                             'pc': 'cutPlaneCenter',
                             'pcx': 'cutPlaneCenterX',
                             'pcy': 'cutPlaneCenterY',
                             'pcz': 'cutPlaneCenterZ',
                             'ph': 'cutPlaneHeight',
                             'ps': 'cutPlaneSize',
                             'pw': 'cutPlaneWidth',
                             'q': 'query',
                             'ro': 'cutPlaneRotate',
                             'rx': 'cutPlaneRotateX',
                             'ry': 'cutPlaneRotateY',
                             'rz': 'cutPlaneRotateZ',
                             'ws': 'worldSpace'},
              'type': 'modeling'},
  'polyCutCtx': {'flags': {'deleteFaces': {'args': '<type __builtin__.bool>',
                                           'longname': 'deleteFaces',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'df'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'extractFaces': {'args': '<type __builtin__.bool>',
                                            'longname': 'extractFaces',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ef'},
                           'extractOffset': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'extractOffset',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'eo'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'df': 'deleteFaces',
                                'e': 'edit',
                                'ef': 'extractFaces',
                                'eo': 'extractOffset',
                                'ex': 'exists',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'q': 'query'},
                 'type': 'context'},
  'polyCutUVCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'history': {'args': '<type __builtin__.bool>',
                                         'longname': 'history',
                                         'numArgs': 1,
                                         'shortname': 'ch'},
                             'image1': {'args': '<type __builtin__.str>',
                                        'longname': 'image1',
                                        'numArgs': 1,
                                        'shortname': 'i1'},
                             'image2': {'args': '<type __builtin__.str>',
                                        'longname': 'image2',
                                        'numArgs': 1,
                                        'shortname': 'i2'},
                             'image3': {'args': '<type __builtin__.str>',
                                        'longname': 'image3',
                                        'numArgs': 1,
                                        'shortname': 'i3'},
                             'loopSpeed': {'args': '<type __builtin__.int>',
                                           'longname': 'loopSpeed',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ls'},
                             'mapBordersColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'mapBordersColor',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'mbc'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'showCheckerMap': {'args': '<type __builtin__.bool>',
                                                'longname': 'showCheckerMap',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'scm'},
                             'showTextureBorders': {'args': '<type __builtin__.bool>',
                                                    'longname': 'showTextureBorders',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'stb'},
                             'showUVShellColoring': {'args': '<type __builtin__.bool>',
                                                     'longname': 'showUVShellColoring',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ssc'},
                             'steadyStroke': {'args': '<type __builtin__.bool>',
                                              'longname': 'steadyStroke',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ss'},
                             'steadyStrokeDistance': {'args': '<type __builtin__.float>',
                                                      'longname': 'steadyStrokeDistance',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ssd'},
                             'symmetry': {'args': '<type __builtin__.int>',
                                          'longname': 'symmetry',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sym'}},
                   'shortFlags': {'ch': 'history',
                                  'e': 'edit',
                                  'ex': 'exists',
                                  'i1': 'image1',
                                  'i2': 'image2',
                                  'i3': 'image3',
                                  'ls': 'loopSpeed',
                                  'mbc': 'mapBordersColor',
                                  'n': 'name',
                                  'q': 'query',
                                  'scm': 'showCheckerMap',
                                  'ss': 'steadyStroke',
                                  'ssc': 'showUVShellColoring',
                                  'ssd': 'steadyStrokeDistance',
                                  'stb': 'showTextureBorders',
                                  'sym': 'symmetry'},
                   'type': 'context'},
  'polyCylinder': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'axis',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 3,
                                      'shortname': 'ax'},
                             'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'resultNeedsCasting': True,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'createUVs': {'args': '<type __builtin__.int>',
                                           'longname': 'createUVs',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'resultNeedsCasting': True,
                                           'shortname': 'cuv'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'height': {'args': '<type __builtin__.float>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'resultNeedsCasting': True,
                                           'shortname': 'nds'},
                             'object': {'args': '<type __builtin__.bool>',
                                        'longname': 'object',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'radius': {'args': '<type __builtin__.float>',
                                        'longname': 'radius',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'r'},
                             'roundCap': {'args': '<type __builtin__.bool>',
                                          'longname': 'roundCap',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'resultNeedsCasting': True,
                                          'shortname': 'rcp'},
                             'subdivisionsAxis': {'args': '<type __builtin__.int>',
                                                  'longname': 'subdivisionsAxis',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'resultNeedsCasting': True,
                                                  'shortname': 'sa'},
                             'subdivisionsCaps': {'args': '<type __builtin__.int>',
                                                  'longname': 'subdivisionsCaps',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'resultNeedsCasting': True,
                                                  'shortname': 'sc'},
                             'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                    'longname': 'subdivisionsHeight',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'resultNeedsCasting': True,
                                                    'shortname': 'sh'},
                             'subdivisionsX': {'args': '<type __builtin__.int>',
                                               'longname': 'subdivisionsX',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'resultNeedsCasting': True,
                                               'shortname': 'sx'},
                             'subdivisionsY': {'args': '<type __builtin__.int>',
                                               'longname': 'subdivisionsY',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'resultNeedsCasting': True,
                                               'shortname': 'sy'},
                             'subdivisionsZ': {'args': '<type __builtin__.int>',
                                               'longname': 'subdivisionsZ',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'resultNeedsCasting': True,
                                               'shortname': 'sz'},
                             'texture': {'args': '<type __builtin__.int>',
                                         'longname': 'texture',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'resultNeedsCasting': True,
                                         'shortname': 'tx'}},
                   'shortFlags': {'ax': 'axis',
                                  'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'cuv': 'createUVs',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'h': 'height',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'o': 'object',
                                  'q': 'query',
                                  'r': 'radius',
                                  'rcp': 'roundCap',
                                  'sa': 'subdivisionsAxis',
                                  'sc': 'subdivisionsCaps',
                                  'sh': 'subdivisionsHeight',
                                  'sx': 'subdivisionsX',
                                  'sy': 'subdivisionsY',
                                  'sz': 'subdivisionsZ',
                                  'tx': 'texture'},
                   'type': 'modeling'},
  'polyCylindricalProjection': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                                      'longname': 'caching',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cch'},
                                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                                  'longname': 'constructionHistory',
                                                                  'modes': ['create',
                                                                            'query'],
                                                                  'numArgs': 1,
                                                                  'shortname': 'ch'},
                                          'createNewMap': {'args': '<type __builtin__.bool>',
                                                           'longname': 'createNewMap',
                                                           'modes': ['create',
                                                                     'query'],
                                                           'numArgs': 1,
                                                           'shortname': 'cm'},
                                          'edit': {'args': '<type __builtin__.bool>',
                                                   'longname': 'edit',
                                                   'numArgs': 0,
                                                   'shortname': 'e'},
                                          'frozen': {'args': '<type __builtin__.bool>',
                                                     'longname': 'frozen',
                                                     'numArgs': 1,
                                                     'shortname': 'fzn'},
                                          'imageCenter': {'args': ['<type __builtin__.float>',
                                                                   '<type __builtin__.float>'],
                                                          'longname': 'imageCenter',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 2,
                                                          'shortname': 'ic'},
                                          'imageCenterX': {'args': '<type __builtin__.float>',
                                                           'longname': 'imageCenterX',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'icx'},
                                          'imageCenterY': {'args': '<type __builtin__.float>',
                                                           'longname': 'imageCenterY',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'icy'},
                                          'imageScale': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'imageScale',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 2,
                                                         'shortname': 'imageScale'},
                                          'imageScaleU': {'args': '<type __builtin__.float>',
                                                          'longname': 'imageScaleU',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'isu'},
                                          'imageScaleV': {'args': '<type __builtin__.float>',
                                                          'longname': 'imageScaleV',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'isv'},
                                          'insertBeforeDeformers': {'args': '<type __builtin__.bool>',
                                                                    'longname': 'insertBeforeDeformers',
                                                                    'modes': ['create'],
                                                                    'numArgs': 1,
                                                                    'shortname': 'ibd'},
                                          'keepImageRatio': {'args': '<type __builtin__.bool>',
                                                             'longname': 'keepImageRatio',
                                                             'modes': ['create'],
                                                             'numArgs': 1,
                                                             'shortname': 'kir'},
                                          'mapDirection': {'args': '<type __builtin__.str>',
                                                           'longname': 'mapDirection',
                                                           'modes': ['create'],
                                                           'numArgs': 1,
                                                           'shortname': 'md'},
                                          'name': {'args': '<type __builtin__.str>',
                                                   'longname': 'name',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'n'},
                                          'nodeState': {'args': '<type __builtin__.int>',
                                                        'longname': 'nodeState',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'nds'},
                                          'perInstance': {'args': '<type __builtin__.bool>',
                                                          'longname': 'perInstance',
                                                          'modes': ['create'],
                                                          'numArgs': 1,
                                                          'shortname': 'pi'},
                                          'projectionCenter': {'args': ['<type __builtin__.float>',
                                                                        '<type __builtin__.float>',
                                                                        '<type __builtin__.float>'],
                                                               'longname': 'projectionCenter',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 3,
                                                               'shortname': 'pc'},
                                          'projectionCenterX': {'args': '<type __builtin__.float>',
                                                                'longname': 'projectionCenterX',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'pcx'},
                                          'projectionCenterY': {'args': '<type __builtin__.float>',
                                                                'longname': 'projectionCenterY',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'pcy'},
                                          'projectionCenterZ': {'args': '<type __builtin__.float>',
                                                                'longname': 'projectionCenterZ',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'pcz'},
                                          'projectionHeight': {'args': '<type __builtin__.float>',
                                                               'longname': 'projectionHeight',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'ph'},
                                          'projectionHorizontalSweep': {'args': '<type __builtin__.float>',
                                                                        'longname': 'projectionHorizontalSweep',
                                                                        'modes': ['create',
                                                                                  'query',
                                                                                  'edit'],
                                                                        'numArgs': 1,
                                                                        'shortname': 'phs'},
                                          'projectionScale': {'args': ['<type __builtin__.float>',
                                                                       '<type __builtin__.float>'],
                                                              'longname': 'projectionScale',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 2,
                                                              'shortname': 'ps'},
                                          'projectionScaleU': {'args': '<type __builtin__.float>',
                                                               'longname': 'projectionScaleU',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'psu'},
                                          'projectionScaleV': {'args': '<type __builtin__.float>',
                                                               'longname': 'projectionScaleV',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'psv'},
                                          'query': {'args': '<type __builtin__.bool>',
                                                    'longname': 'query',
                                                    'numArgs': 0,
                                                    'shortname': 'q'},
                                          'radius': {'args': '<type __builtin__.float>',
                                                     'longname': 'radius',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'r'},
                                          'rotate': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'rotate',
                                                     'numArgs': 3,
                                                     'shortname': 'ro'},
                                          'rotateX': {'args': '<type __builtin__.float>',
                                                      'longname': 'rotateX',
                                                      'numArgs': 1,
                                                      'shortname': 'rx'},
                                          'rotateY': {'args': '<type __builtin__.float>',
                                                      'longname': 'rotateY',
                                                      'numArgs': 1,
                                                      'shortname': 'ry'},
                                          'rotateZ': {'args': '<type __builtin__.float>',
                                                      'longname': 'rotateZ',
                                                      'numArgs': 1,
                                                      'shortname': 'rz'},
                                          'rotationAngle': {'args': '<type __builtin__.float>',
                                                            'longname': 'rotationAngle',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ra'},
                                          'seamCorrect': {'args': '<type __builtin__.bool>',
                                                          'longname': 'seamCorrect',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'sc'},
                                          'smartFit': {'args': '<type __builtin__.bool>',
                                                       'longname': 'smartFit',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'sf'},
                                          'worldSpace': {'args': '<type __builtin__.bool>',
                                                         'longname': 'worldSpace',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ws'}},
                                'removedFlags': {'is': 'imageScale'},
                                'shortFlags': {'cch': 'caching',
                                               'ch': 'constructionHistory',
                                               'cm': 'createNewMap',
                                               'e': 'edit',
                                               'fzn': 'frozen',
                                               'ibd': 'insertBeforeDeformers',
                                               'ic': 'imageCenter',
                                               'icx': 'imageCenterX',
                                               'icy': 'imageCenterY',
                                               'imageScale': 'imageScale',
                                               'isu': 'imageScaleU',
                                               'isv': 'imageScaleV',
                                               'kir': 'keepImageRatio',
                                               'md': 'mapDirection',
                                               'n': 'name',
                                               'nds': 'nodeState',
                                               'pc': 'projectionCenter',
                                               'pcx': 'projectionCenterX',
                                               'pcy': 'projectionCenterY',
                                               'pcz': 'projectionCenterZ',
                                               'ph': 'projectionHeight',
                                               'phs': 'projectionHorizontalSweep',
                                               'pi': 'perInstance',
                                               'ps': 'projectionScale',
                                               'psu': 'projectionScaleU',
                                               'psv': 'projectionScaleV',
                                               'q': 'query',
                                               'r': 'radius',
                                               'ra': 'rotationAngle',
                                               'ro': 'rotate',
                                               'rx': 'rotateX',
                                               'ry': 'rotateY',
                                               'rz': 'rotateZ',
                                               'sc': 'seamCorrect',
                                               'sf': 'smartFit',
                                               'ws': 'worldSpace'},
                                'type': 'modeling'},
  'polyDelEdge': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'cleanVertices': {'args': '<type __builtin__.bool>',
                                              'longname': 'cleanVertices',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cv'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'}},
                  'shortFlags': {'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'cv': 'cleanVertices',
                                 'e': 'edit',
                                 'fzn': 'frozen',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'q': 'query'},
                  'type': 'modeling'},
  'polyDelFacet': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'q': 'query'},
                   'type': 'modeling'},
  'polyDelVertex': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cch'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nds'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'}},
                    'shortFlags': {'cch': 'caching',
                                   'ch': 'constructionHistory',
                                   'e': 'edit',
                                   'fzn': 'frozen',
                                   'n': 'name',
                                   'nds': 'nodeState',
                                   'q': 'query'},
                    'type': 'modeling'},
  'polyDuplicateAndConnect': {'flags': {'removeOriginalFromShaders': {'args': '<type __builtin__.bool>',
                                                                      'longname': 'removeOriginalFromShaders',
                                                                      'modes': ['create'],
                                                                      'numArgs': 0,
                                                                      'shortname': 'ros'},
                                        'renameChildren': {'args': '<type __builtin__.bool>',
                                                           'longname': 'renameChildren',
                                                           'modes': ['create'],
                                                           'numArgs': 0,
                                                           'shortname': 'rc'}},
                              'shortFlags': {'rc': 'renameChildren',
                                             'ros': 'removeOriginalFromShaders'},
                              'type': 'modeling'},
  'polyDuplicateEdge': {'flags': {'adjustEdgeFlow': {'args': '<type __builtin__.float>',
                                                     'longname': 'adjustEdgeFlow',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'aef'},
                                  'caching': {'args': '<type __builtin__.bool>',
                                              'longname': 'caching',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cch'},
                                  'constructionHistory': {'args': '<type __builtin__.bool>',
                                                          'longname': 'constructionHistory',
                                                          'modes': ['create',
                                                                    'query'],
                                                          'numArgs': 1,
                                                          'shortname': 'ch'},
                                  'deleteEdge': {'args': '<type __builtin__.bool>',
                                                 'longname': 'deleteEdge',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'de'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'endVertexOffset': {'args': '<type __builtin__.float>',
                                                      'longname': 'endVertexOffset',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'evo'},
                                  'frozen': {'args': '<type __builtin__.bool>',
                                             'longname': 'frozen',
                                             'numArgs': 1,
                                             'shortname': 'fzn'},
                                  'insertWithEdgeFlow': {'args': '<type __builtin__.bool>',
                                                         'longname': 'insertWithEdgeFlow',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ief'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'nodeState': {'args': '<type __builtin__.int>',
                                                'longname': 'nodeState',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nds'},
                                  'offset': {'args': '<type __builtin__.float>',
                                             'longname': 'offset',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'of'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'smoothingAngle': {'args': '<type __builtin__.float>',
                                                     'longname': 'smoothingAngle',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sma'},
                                  'splitType': {'args': '<type __builtin__.int>',
                                                'longname': 'splitType',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'stp'},
                                  'startVertexOffset': {'args': '<type __builtin__.float>',
                                                        'longname': 'startVertexOffset',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'svo'}},
                        'shortFlags': {'aef': 'adjustEdgeFlow',
                                       'cch': 'caching',
                                       'ch': 'constructionHistory',
                                       'de': 'deleteEdge',
                                       'e': 'edit',
                                       'evo': 'endVertexOffset',
                                       'fzn': 'frozen',
                                       'ief': 'insertWithEdgeFlow',
                                       'n': 'name',
                                       'nds': 'nodeState',
                                       'of': 'offset',
                                       'q': 'query',
                                       'sma': 'smoothingAngle',
                                       'stp': 'splitType',
                                       'svo': 'startVertexOffset'},
                        'type': 'modeling'},
  'polyEditEdgeFlow': {'flags': {'adjustEdgeFlow': {'args': '<type __builtin__.float>',
                                                    'longname': 'adjustEdgeFlow',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'aef'},
                                 'caching': {'args': '<type __builtin__.bool>',
                                             'longname': 'caching',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cch'},
                                 'constructionHistory': {'args': '<type __builtin__.bool>',
                                                         'longname': 'constructionHistory',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'ch'},
                                 'edgeFlow': {'args': '<type __builtin__.bool>',
                                              'longname': 'edgeFlow',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ef'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'frozen': {'args': '<type __builtin__.bool>',
                                            'longname': 'frozen',
                                            'numArgs': 1,
                                            'shortname': 'fzn'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'nodeState': {'args': '<type __builtin__.int>',
                                               'longname': 'nodeState',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nds'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'}},
                       'shortFlags': {'aef': 'adjustEdgeFlow',
                                      'cch': 'caching',
                                      'ch': 'constructionHistory',
                                      'e': 'edit',
                                      'ef': 'edgeFlow',
                                      'fzn': 'frozen',
                                      'n': 'name',
                                      'nds': 'nodeState',
                                      'q': 'query'},
                       'type': 'modeling'},
  'polyEditUV': {'flags': {'angle': {'args': '<type __builtin__.float>',
                                     'longname': 'angle',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'a'},
                           'pivotU': {'args': '<type __builtin__.float>',
                                      'longname': 'pivotU',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'pu'},
                           'pivotV': {'args': '<type __builtin__.float>',
                                      'longname': 'pivotV',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'pv'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'relative': {'args': '<type __builtin__.bool>',
                                        'longname': 'relative',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'r'},
                           'rotateRatio': {'args': '<type __builtin__.float>',
                                           'longname': 'rotateRatio',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'rr'},
                           'rotation': {'args': '<type __builtin__.bool>',
                                        'longname': 'rotation',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'rot'},
                           'scale': {'args': '<type __builtin__.bool>',
                                     'longname': 'scale',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 's'},
                           'scaleU': {'args': '<type __builtin__.float>',
                                      'longname': 'scaleU',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'su'},
                           'scaleV': {'args': '<type __builtin__.float>',
                                      'longname': 'scaleV',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'sv'},
                           'uValue': {'args': '<type __builtin__.float>',
                                      'longname': 'uValue',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'u'},
                           'uvSetName': {'args': '<type __builtin__.str>',
                                         'longname': 'uvSetName',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'uvs'},
                           'vValue': {'args': '<type __builtin__.float>',
                                      'longname': 'vValue',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'v'}},
                 'shortFlags': {'a': 'angle',
                                'pu': 'pivotU',
                                'pv': 'pivotV',
                                'q': 'query',
                                'r': 'relative',
                                'rot': 'rotation',
                                'rr': 'rotateRatio',
                                's': 'scale',
                                'su': 'scaleU',
                                'sv': 'scaleV',
                                'u': 'uValue',
                                'uvs': 'uvSetName',
                                'v': 'vValue'},
                 'type': 'modeling'},
  'polyEditUVShell': {'flags': {'angle': {'args': '<type __builtin__.float>',
                                          'longname': 'angle',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'a'},
                                'pivotU': {'args': '<type __builtin__.float>',
                                           'longname': 'pivotU',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'pu'},
                                'pivotV': {'args': '<type __builtin__.float>',
                                           'longname': 'pivotV',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'pv'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'relative': {'args': '<type __builtin__.bool>',
                                             'longname': 'relative',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'r'},
                                'rotateRatio': {'args': '<type __builtin__.float>',
                                                'longname': 'rotateRatio',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'rr'},
                                'rotation': {'args': '<type __builtin__.bool>',
                                             'longname': 'rotation',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'rot'},
                                'scale': {'args': '<type __builtin__.bool>',
                                          'longname': 'scale',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 's'},
                                'scaleU': {'args': '<type __builtin__.float>',
                                           'longname': 'scaleU',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'su'},
                                'scaleV': {'args': '<type __builtin__.float>',
                                           'longname': 'scaleV',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'sv'},
                                'uValue': {'args': '<type __builtin__.float>',
                                           'longname': 'uValue',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'u'},
                                'uvSetName': {'args': '<type __builtin__.str>',
                                              'longname': 'uvSetName',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'uvs'},
                                'vValue': {'args': '<type __builtin__.float>',
                                           'longname': 'vValue',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'v'}},
                      'shortFlags': {'a': 'angle',
                                     'pu': 'pivotU',
                                     'pv': 'pivotV',
                                     'q': 'query',
                                     'r': 'relative',
                                     'rot': 'rotation',
                                     'rr': 'rotateRatio',
                                     's': 'scale',
                                     'su': 'scaleU',
                                     'sv': 'scaleV',
                                     'u': 'uValue',
                                     'uvs': 'uvSetName',
                                     'v': 'vValue'},
                      'type': 'modeling'},
  'polyEvaluate': {'flags': {'accurateEvaluation': {'args': '<type __builtin__.bool>',
                                                    'longname': 'accurateEvaluation',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ae'},
                             'activeShells': {'args': '<type __builtin__.bool>',
                                              'longname': 'activeShells',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'activeShells'},
                             'activeUVShells': {'args': '<type __builtin__.bool>',
                                                'longname': 'activeUVShells',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'aus'},
                             'area': {'args': '<type __builtin__.bool>',
                                      'longname': 'area',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'a'},
                             'boundingBox': {'args': '<type __builtin__.bool>',
                                             'longname': 'boundingBox',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'b'},
                             'boundingBox2d': {'args': '<type __builtin__.bool>',
                                               'longname': 'boundingBox2d',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'b2'},
                             'boundingBoxComponent': {'args': '<type __builtin__.bool>',
                                                      'longname': 'boundingBoxComponent',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'bc'},
                             'boundingBoxComponent2d': {'args': '<type __builtin__.bool>',
                                                        'longname': 'boundingBoxComponent2d',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'bc2'},
                             'displayStats': {'args': '<type __builtin__.bool>',
                                              'longname': 'displayStats',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ds'},
                             'edge': {'args': '<type __builtin__.bool>',
                                      'longname': 'edge',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'edgeComponent': {'args': '<type __builtin__.bool>',
                                               'longname': 'edgeComponent',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ec'},
                             'face': {'args': '<type __builtin__.bool>',
                                      'longname': 'face',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'f'},
                             'faceArea': {'args': '<type __builtin__.bool>',
                                          'longname': 'faceArea',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'fa'},
                             'faceComponent': {'args': '<type __builtin__.bool>',
                                               'longname': 'faceComponent',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'fc'},
                             'format': {'args': '<type __builtin__.bool>',
                                        'longname': 'format',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'fmt'},
                             'shell': {'args': '<type __builtin__.bool>',
                                       'longname': 'shell',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 's'},
                             'triangle': {'args': '<type __builtin__.bool>',
                                          'longname': 'triangle',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 't'},
                             'triangleComponent': {'args': '<type __builtin__.bool>',
                                                   'longname': 'triangleComponent',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'tc'},
                             'uvArea': {'args': '<type __builtin__.bool>',
                                        'longname': 'uvArea',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'uva'},
                             'uvComponent': {'args': '<type __builtin__.bool>',
                                             'longname': 'uvComponent',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'uvc'},
                             'uvEdgePairs': {'args': '<type __builtin__.bool>',
                                             'longname': 'uvEdgePairs',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'uep'},
                             'uvFaceArea': {'args': '<type __builtin__.bool>',
                                            'longname': 'uvFaceArea',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ufa'},
                             'uvSetName': {'args': '<type __builtin__.str>',
                                           'longname': 'uvSetName',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'uvs'},
                             'uvShell': {'args': '<type __builtin__.bool>',
                                         'longname': 'uvShell',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'us'},
                             'uvShellIds': {'args': '<type __builtin__.bool>',
                                            'longname': 'uvShellIds',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'usi'},
                             'uvcoord': {'args': '<type __builtin__.bool>',
                                         'longname': 'uvcoord',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'uv'},
                             'uvsInShell': {'args': '<type __builtin__.int>',
                                            'longname': 'uvsInShell',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'uis'},
                             'vertex': {'args': '<type __builtin__.bool>',
                                        'longname': 'vertex',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'v'},
                             'vertexComponent': {'args': '<type __builtin__.bool>',
                                                 'longname': 'vertexComponent',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'vc'},
                             'worldArea': {'args': '<type __builtin__.bool>',
                                           'longname': 'worldArea',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'wa'},
                             'worldFaceArea': {'args': '<type __builtin__.bool>',
                                               'longname': 'worldFaceArea',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'wfa'}},
                   'removedFlags': {'as': 'activeShells'},
                   'shortFlags': {'a': 'area',
                                  'activeShells': 'activeShells',
                                  'ae': 'accurateEvaluation',
                                  'aus': 'activeUVShells',
                                  'b': 'boundingBox',
                                  'b2': 'boundingBox2d',
                                  'bc': 'boundingBoxComponent',
                                  'bc2': 'boundingBoxComponent2d',
                                  'ds': 'displayStats',
                                  'e': 'edge',
                                  'ec': 'edgeComponent',
                                  'f': 'face',
                                  'fa': 'faceArea',
                                  'fc': 'faceComponent',
                                  'fmt': 'format',
                                  's': 'shell',
                                  't': 'triangle',
                                  'tc': 'triangleComponent',
                                  'uep': 'uvEdgePairs',
                                  'ufa': 'uvFaceArea',
                                  'uis': 'uvsInShell',
                                  'us': 'uvShell',
                                  'usi': 'uvShellIds',
                                  'uv': 'uvcoord',
                                  'uva': 'uvArea',
                                  'uvc': 'uvComponent',
                                  'uvs': 'uvSetName',
                                  'v': 'vertex',
                                  'vc': 'vertexComponent',
                                  'wa': 'worldArea',
                                  'wfa': 'worldFaceArea'},
                   'type': 'modeling'},
  'polyExtrudeEdge': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                            'longname': 'caching',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cch'},
                                'constructionHistory': {'args': '<type __builtin__.bool>',
                                                        'longname': 'constructionHistory',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'ch'},
                                'createCurve': {'args': '<type __builtin__.bool>',
                                                'longname': 'createCurve',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'cc'},
                                'divisions': {'args': '<type __builtin__.int>',
                                              'longname': 'divisions',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'd'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'frozen': {'args': '<type __builtin__.bool>',
                                           'longname': 'frozen',
                                           'numArgs': 1,
                                           'shortname': 'fzn'},
                                'gain': {'args': '<type __builtin__.float>',
                                         'longname': 'gain',
                                         'modes': ['create',
                                                   'query',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'ga'},
                                'inputCurve': {'args': 'PyNode',
                                               'longname': 'inputCurve',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'inc'},
                                'keepFacesTogether': {'args': '<type __builtin__.bool>',
                                                      'longname': 'keepFacesTogether',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'kft'},
                                'localCenter': {'args': '<type __builtin__.int>',
                                                'longname': 'localCenter',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'lc'},
                                'localDirection': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'localDirection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'ld'},
                                'localDirectionX': {'args': '<type __builtin__.float>',
                                                    'longname': 'localDirectionX',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ldx'},
                                'localDirectionY': {'args': '<type __builtin__.float>',
                                                    'longname': 'localDirectionY',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ldy'},
                                'localDirectionZ': {'args': '<type __builtin__.float>',
                                                    'longname': 'localDirectionZ',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ldz'},
                                'localRotate': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'localRotate',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'lr'},
                                'localRotateX': {'args': '<type __builtin__.float>',
                                                 'longname': 'localRotateX',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'lrx'},
                                'localRotateY': {'args': '<type __builtin__.float>',
                                                 'longname': 'localRotateY',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'lry'},
                                'localRotateZ': {'args': '<type __builtin__.float>',
                                                 'longname': 'localRotateZ',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'lrz'},
                                'localScale': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'localScale',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'ls'},
                                'localScaleX': {'args': '<type __builtin__.float>',
                                                'longname': 'localScaleX',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'lsx'},
                                'localScaleY': {'args': '<type __builtin__.float>',
                                                'longname': 'localScaleY',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'lsy'},
                                'localScaleZ': {'args': '<type __builtin__.float>',
                                                'longname': 'localScaleZ',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'lsz'},
                                'localTranslate': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'localTranslate',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'lt'},
                                'localTranslateX': {'args': '<type __builtin__.float>',
                                                    'longname': 'localTranslateX',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ltx'},
                                'localTranslateY': {'args': '<type __builtin__.float>',
                                                    'longname': 'localTranslateY',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'lty'},
                                'localTranslateZ': {'args': '<type __builtin__.float>',
                                                    'longname': 'localTranslateZ',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ltz'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'nodeState': {'args': '<type __builtin__.int>',
                                              'longname': 'nodeState',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nds'},
                                'offset': {'args': '<type __builtin__.float>',
                                           'longname': 'offset',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'off'},
                                'pivot': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'pivot',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 3,
                                          'shortname': 'pvt'},
                                'pivotX': {'args': '<type __builtin__.float>',
                                           'longname': 'pivotX',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pvx'},
                                'pivotY': {'args': '<type __builtin__.float>',
                                           'longname': 'pivotY',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pvy'},
                                'pivotZ': {'args': '<type __builtin__.float>',
                                           'longname': 'pivotZ',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pvz'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'random': {'args': '<type __builtin__.float>',
                                           'longname': 'random',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ran'},
                                'rotate': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'rotate',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'ro'},
                                'rotateX': {'args': '<type __builtin__.float>',
                                            'longname': 'rotateX',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'rx'},
                                'rotateY': {'args': '<type __builtin__.float>',
                                            'longname': 'rotateY',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ry'},
                                'rotateZ': {'args': '<type __builtin__.float>',
                                            'longname': 'rotateZ',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'rz'},
                                'scale': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'scale',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 3,
                                          'shortname': 's'},
                                'scaleX': {'args': '<type __builtin__.float>',
                                           'longname': 'scaleX',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sx'},
                                'scaleY': {'args': '<type __builtin__.float>',
                                           'longname': 'scaleY',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sy'},
                                'scaleZ': {'args': '<type __builtin__.float>',
                                           'longname': 'scaleZ',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sz'},
                                'smoothingAngle': {'args': '<type __builtin__.float>',
                                                   'longname': 'smoothingAngle',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sma'},
                                'taper': {'args': '<type __builtin__.float>',
                                          'longname': 'taper',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'tp'},
                                'taperCurve_FloatValue': {'args': '<type __builtin__.float>',
                                                          'longname': 'taperCurve_FloatValue',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'cfv'},
                                'taperCurve_Interp': {'args': '<type __builtin__.int>',
                                                      'longname': 'taperCurve_Interp',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ci'},
                                'taperCurve_Position': {'args': '<type __builtin__.float>',
                                                        'longname': 'taperCurve_Position',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'cp'},
                                'thickness': {'args': '<type __builtin__.float>',
                                              'longname': 'thickness',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'tk'},
                                'translate': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'translate',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 't'},
                                'translateX': {'args': '<type __builtin__.float>',
                                               'longname': 'translateX',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tx'},
                                'translateY': {'args': '<type __builtin__.float>',
                                               'longname': 'translateY',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ty'},
                                'translateZ': {'args': '<type __builtin__.float>',
                                               'longname': 'translateZ',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tz'},
                                'twist': {'args': '<type __builtin__.float>',
                                          'longname': 'twist',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'twt'},
                                'worldSpace': {'args': '<type __builtin__.bool>',
                                               'longname': 'worldSpace',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ws'}},
                      'shortFlags': {'cc': 'createCurve',
                                     'cch': 'caching',
                                     'cfv': 'taperCurve_FloatValue',
                                     'ch': 'constructionHistory',
                                     'ci': 'taperCurve_Interp',
                                     'cp': 'taperCurve_Position',
                                     'd': 'divisions',
                                     'e': 'edit',
                                     'fzn': 'frozen',
                                     'ga': 'gain',
                                     'inc': 'inputCurve',
                                     'kft': 'keepFacesTogether',
                                     'lc': 'localCenter',
                                     'ld': 'localDirection',
                                     'ldx': 'localDirectionX',
                                     'ldy': 'localDirectionY',
                                     'ldz': 'localDirectionZ',
                                     'lr': 'localRotate',
                                     'lrx': 'localRotateX',
                                     'lry': 'localRotateY',
                                     'lrz': 'localRotateZ',
                                     'ls': 'localScale',
                                     'lsx': 'localScaleX',
                                     'lsy': 'localScaleY',
                                     'lsz': 'localScaleZ',
                                     'lt': 'localTranslate',
                                     'ltx': 'localTranslateX',
                                     'lty': 'localTranslateY',
                                     'ltz': 'localTranslateZ',
                                     'n': 'name',
                                     'nds': 'nodeState',
                                     'off': 'offset',
                                     'pvt': 'pivot',
                                     'pvx': 'pivotX',
                                     'pvy': 'pivotY',
                                     'pvz': 'pivotZ',
                                     'q': 'query',
                                     'ran': 'random',
                                     'ro': 'rotate',
                                     'rx': 'rotateX',
                                     'ry': 'rotateY',
                                     'rz': 'rotateZ',
                                     's': 'scale',
                                     'sma': 'smoothingAngle',
                                     'sx': 'scaleX',
                                     'sy': 'scaleY',
                                     'sz': 'scaleZ',
                                     't': 'translate',
                                     'tk': 'thickness',
                                     'tp': 'taper',
                                     'twt': 'twist',
                                     'tx': 'translateX',
                                     'ty': 'translateY',
                                     'tz': 'translateZ',
                                     'ws': 'worldSpace'},
                      'type': 'modeling'},
  'polyExtrudeFacet': {'flags': {'attraction': {'args': '<type __builtin__.float>',
                                                'longname': 'attraction',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'att'},
                                 'caching': {'args': '<type __builtin__.bool>',
                                             'longname': 'caching',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cch'},
                                 'constructionHistory': {'args': '<type __builtin__.bool>',
                                                         'longname': 'constructionHistory',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'ch'},
                                 'createCurve': {'args': '<type __builtin__.bool>',
                                                 'longname': 'createCurve',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'cc'},
                                 'divisions': {'args': '<type __builtin__.int>',
                                               'longname': 'divisions',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'd'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'frozen': {'args': '<type __builtin__.bool>',
                                            'longname': 'frozen',
                                            'numArgs': 1,
                                            'shortname': 'fzn'},
                                 'gain': {'args': '<type __builtin__.float>',
                                          'longname': 'gain',
                                          'modes': ['create',
                                                    'query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'ga'},
                                 'gravity': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'gravity',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'g'},
                                 'gravityX': {'args': '<type __builtin__.float>',
                                              'longname': 'gravityX',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'gx'},
                                 'gravityY': {'args': '<type __builtin__.float>',
                                              'longname': 'gravityY',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'gy'},
                                 'gravityZ': {'args': '<type __builtin__.float>',
                                              'longname': 'gravityZ',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'gz'},
                                 'inputCurve': {'args': 'PyNode',
                                                'longname': 'inputCurve',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'inc'},
                                 'keepFacesTogether': {'args': '<type __builtin__.bool>',
                                                       'longname': 'keepFacesTogether',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'kft'},
                                 'keepFacetTogether': {'args': '<type __builtin__.bool>',
                                                       'longname': 'keepFacetTogether',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'xft'},
                                 'localCenter': {'args': '<type __builtin__.int>',
                                                 'longname': 'localCenter',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'lc'},
                                 'localDirection': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'localDirection',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'ld'},
                                 'localDirectionX': {'args': '<type __builtin__.float>',
                                                     'longname': 'localDirectionX',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ldx'},
                                 'localDirectionY': {'args': '<type __builtin__.float>',
                                                     'longname': 'localDirectionY',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ldy'},
                                 'localDirectionZ': {'args': '<type __builtin__.float>',
                                                     'longname': 'localDirectionZ',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ldz'},
                                 'localRotate': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'localRotate',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'lr'},
                                 'localRotateX': {'args': '<type __builtin__.float>',
                                                  'longname': 'localRotateX',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'lrx'},
                                 'localRotateY': {'args': '<type __builtin__.float>',
                                                  'longname': 'localRotateY',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'lry'},
                                 'localRotateZ': {'args': '<type __builtin__.float>',
                                                  'longname': 'localRotateZ',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'lrz'},
                                 'localScale': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'localScale',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'ls'},
                                 'localScaleX': {'args': '<type __builtin__.float>',
                                                 'longname': 'localScaleX',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'lsx'},
                                 'localScaleY': {'args': '<type __builtin__.float>',
                                                 'longname': 'localScaleY',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'lsy'},
                                 'localScaleZ': {'args': '<type __builtin__.float>',
                                                 'longname': 'localScaleZ',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'lsz'},
                                 'localTranslate': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'localTranslate',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'lt'},
                                 'localTranslateX': {'args': '<type __builtin__.float>',
                                                     'longname': 'localTranslateX',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ltx'},
                                 'localTranslateY': {'args': '<type __builtin__.float>',
                                                     'longname': 'localTranslateY',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'lty'},
                                 'localTranslateZ': {'args': '<type __builtin__.float>',
                                                     'longname': 'localTranslateZ',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ltz'},
                                 'magnX': {'args': '<type __builtin__.float>',
                                           'longname': 'magnX',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'mx'},
                                 'magnY': {'args': '<type __builtin__.float>',
                                           'longname': 'magnY',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'my'},
                                 'magnZ': {'args': '<type __builtin__.float>',
                                           'longname': 'magnZ',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'mz'},
                                 'magnet': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'magnet',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'm'},
                                 'maya2012': {'args': '<type __builtin__.bool>',
                                              'longname': 'maya2012',
                                              'numArgs': 1,
                                              'shortname': 'm12'},
                                 'maya2018': {'args': '<type __builtin__.bool>',
                                              'longname': 'maya2018',
                                              'numArgs': 1,
                                              'shortname': 'm18'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'nodeState': {'args': '<type __builtin__.int>',
                                               'longname': 'nodeState',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nds'},
                                 'offset': {'args': '<type __builtin__.float>',
                                            'longname': 'offset',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'off'},
                                 'pivot': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'pivot',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'pvt'},
                                 'pivotX': {'args': '<type __builtin__.float>',
                                            'longname': 'pivotX',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pvx'},
                                 'pivotY': {'args': '<type __builtin__.float>',
                                            'longname': 'pivotY',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pvy'},
                                 'pivotZ': {'args': '<type __builtin__.float>',
                                            'longname': 'pivotZ',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pvz'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'random': {'args': '<type __builtin__.float>',
                                            'longname': 'random',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ran'},
                                 'reverseAllFaces': {'args': '<type __builtin__.bool>',
                                                     'longname': 'reverseAllFaces',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'raf'},
                                 'rotate': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'rotate',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'ro'},
                                 'rotateX': {'args': '<type __builtin__.float>',
                                             'longname': 'rotateX',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'rx'},
                                 'rotateY': {'args': '<type __builtin__.float>',
                                             'longname': 'rotateY',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ry'},
                                 'rotateZ': {'args': '<type __builtin__.float>',
                                             'longname': 'rotateZ',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'rz'},
                                 'scale': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'scale',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 's'},
                                 'scaleX': {'args': '<type __builtin__.float>',
                                            'longname': 'scaleX',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sx'},
                                 'scaleY': {'args': '<type __builtin__.float>',
                                            'longname': 'scaleY',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sy'},
                                 'scaleZ': {'args': '<type __builtin__.float>',
                                            'longname': 'scaleZ',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sz'},
                                 'smoothingAngle': {'args': '<type __builtin__.float>',
                                                    'longname': 'smoothingAngle',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sma'},
                                 'taper': {'args': '<type __builtin__.float>',
                                           'longname': 'taper',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'tp'},
                                 'taperCurve_FloatValue': {'args': '<type __builtin__.float>',
                                                           'longname': 'taperCurve_FloatValue',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'cfv'},
                                 'taperCurve_Interp': {'args': '<type __builtin__.int>',
                                                       'longname': 'taperCurve_Interp',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ci'},
                                 'taperCurve_Position': {'args': '<type __builtin__.float>',
                                                         'longname': 'taperCurve_Position',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'cp'},
                                 'thickness': {'args': '<type __builtin__.float>',
                                               'longname': 'thickness',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tk'},
                                 'translate': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'translate',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 't'},
                                 'translateX': {'args': '<type __builtin__.float>',
                                                'longname': 'translateX',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'tx'},
                                 'translateY': {'args': '<type __builtin__.float>',
                                                'longname': 'translateY',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ty'},
                                 'translateZ': {'args': '<type __builtin__.float>',
                                                'longname': 'translateZ',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'tz'},
                                 'twist': {'args': '<type __builtin__.float>',
                                           'longname': 'twist',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'twt'},
                                 'weight': {'args': '<type __builtin__.float>',
                                            'longname': 'weight',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'w'},
                                 'worldSpace': {'args': '<type __builtin__.bool>',
                                                'longname': 'worldSpace',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ws'}},
                       'shortFlags': {'att': 'attraction',
                                      'cc': 'createCurve',
                                      'cch': 'caching',
                                      'cfv': 'taperCurve_FloatValue',
                                      'ch': 'constructionHistory',
                                      'ci': 'taperCurve_Interp',
                                      'cp': 'taperCurve_Position',
                                      'd': 'divisions',
                                      'e': 'edit',
                                      'fzn': 'frozen',
                                      'g': 'gravity',
                                      'ga': 'gain',
                                      'gx': 'gravityX',
                                      'gy': 'gravityY',
                                      'gz': 'gravityZ',
                                      'inc': 'inputCurve',
                                      'kft': 'keepFacesTogether',
                                      'lc': 'localCenter',
                                      'ld': 'localDirection',
                                      'ldx': 'localDirectionX',
                                      'ldy': 'localDirectionY',
                                      'ldz': 'localDirectionZ',
                                      'lr': 'localRotate',
                                      'lrx': 'localRotateX',
                                      'lry': 'localRotateY',
                                      'lrz': 'localRotateZ',
                                      'ls': 'localScale',
                                      'lsx': 'localScaleX',
                                      'lsy': 'localScaleY',
                                      'lsz': 'localScaleZ',
                                      'lt': 'localTranslate',
                                      'ltx': 'localTranslateX',
                                      'lty': 'localTranslateY',
                                      'ltz': 'localTranslateZ',
                                      'm': 'magnet',
                                      'm12': 'maya2012',
                                      'm18': 'maya2018',
                                      'mx': 'magnX',
                                      'my': 'magnY',
                                      'mz': 'magnZ',
                                      'n': 'name',
                                      'nds': 'nodeState',
                                      'off': 'offset',
                                      'pvt': 'pivot',
                                      'pvx': 'pivotX',
                                      'pvy': 'pivotY',
                                      'pvz': 'pivotZ',
                                      'q': 'query',
                                      'raf': 'reverseAllFaces',
                                      'ran': 'random',
                                      'ro': 'rotate',
                                      'rx': 'rotateX',
                                      'ry': 'rotateY',
                                      'rz': 'rotateZ',
                                      's': 'scale',
                                      'sma': 'smoothingAngle',
                                      'sx': 'scaleX',
                                      'sy': 'scaleY',
                                      'sz': 'scaleZ',
                                      't': 'translate',
                                      'tk': 'thickness',
                                      'tp': 'taper',
                                      'twt': 'twist',
                                      'tx': 'translateX',
                                      'ty': 'translateY',
                                      'tz': 'translateZ',
                                      'w': 'weight',
                                      'ws': 'worldSpace',
                                      'xft': 'keepFacetTogether'},
                       'type': 'modeling'},
  'polyExtrudeVertex': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                              'longname': 'caching',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cch'},
                                  'constructionHistory': {'args': '<type __builtin__.bool>',
                                                          'longname': 'constructionHistory',
                                                          'modes': ['create',
                                                                    'query'],
                                                          'numArgs': 1,
                                                          'shortname': 'ch'},
                                  'divisions': {'args': '<type __builtin__.int>',
                                                'longname': 'divisions',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'd'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'frozen': {'args': '<type __builtin__.bool>',
                                             'longname': 'frozen',
                                             'numArgs': 1,
                                             'shortname': 'fzn'},
                                  'length': {'args': '<type __builtin__.float>',
                                             'longname': 'length',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'l'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'nodeState': {'args': '<type __builtin__.int>',
                                                'longname': 'nodeState',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nds'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'width': {'args': '<type __builtin__.float>',
                                            'longname': 'width',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'w'},
                                  'worldSpace': {'args': '<type __builtin__.bool>',
                                                 'longname': 'worldSpace',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ws'}},
                        'shortFlags': {'cch': 'caching',
                                       'ch': 'constructionHistory',
                                       'd': 'divisions',
                                       'e': 'edit',
                                       'fzn': 'frozen',
                                       'l': 'length',
                                       'n': 'name',
                                       'nds': 'nodeState',
                                       'q': 'query',
                                       'w': 'width',
                                       'ws': 'worldSpace'},
                        'type': 'modeling'},
  'polyFlipEdge': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'e': 'edit', 'q': 'query'},
                   'type': 'modeling'},
  'polyFlipUV': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'createNewMap': {'args': '<type __builtin__.bool>',
                                            'longname': 'createNewMap',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'cm'},
                           'cutUV': {'args': '<type __builtin__.bool>',
                                     'longname': 'cutUV',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'cut'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'flipType': {'args': '<type __builtin__.int>',
                                        'longname': 'flipType',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ft'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'insertBeforeDeformers': {'args': '<type __builtin__.bool>',
                                                     'longname': 'insertBeforeDeformers',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ibd'},
                           'local': {'args': '<type __builtin__.bool>',
                                     'longname': 'local',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'l'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'pivotU': {'args': '<type __builtin__.float>',
                                      'longname': 'pivotU',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'pu'},
                           'pivotV': {'args': '<type __builtin__.float>',
                                      'longname': 'pivotV',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'pv'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'usePivot': {'args': '<type __builtin__.bool>',
                                        'longname': 'usePivot',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'up'},
                           'uvSetName': {'args': '<type __builtin__.str>',
                                         'longname': 'uvSetName',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'uvs'},
                           'worldSpace': {'args': '<type __builtin__.bool>',
                                          'longname': 'worldSpace',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ws'}},
                 'shortFlags': {'cch': 'caching',
                                'ch': 'constructionHistory',
                                'cm': 'createNewMap',
                                'cut': 'cutUV',
                                'e': 'edit',
                                'ft': 'flipType',
                                'fzn': 'frozen',
                                'ibd': 'insertBeforeDeformers',
                                'l': 'local',
                                'n': 'name',
                                'nds': 'nodeState',
                                'pu': 'pivotU',
                                'pv': 'pivotV',
                                'q': 'query',
                                'up': 'usePivot',
                                'uvs': 'uvSetName',
                                'ws': 'worldSpace'},
                 'type': 'modeling'},
  'polyForceUV': {'flags': {'cameraProjection': {'args': '<type __builtin__.bool>',
                                                 'longname': 'cameraProjection',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'cp'},
                            'createNewMap': {'args': '<type __builtin__.bool>',
                                             'longname': 'createNewMap',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'cm'},
                            'flipHorizontal': {'args': '<type __builtin__.bool>',
                                               'longname': 'flipHorizontal',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'fh'},
                            'flipVertical': {'args': '<type __builtin__.bool>',
                                             'longname': 'flipVertical',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'fv'},
                            'g': {'args': '<type __builtin__.bool>',
                                  'longname': 'g',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'g'},
                            'local': {'args': '<type __builtin__.bool>',
                                      'longname': 'local',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'l'},
                            'normalize': {'args': '<type __builtin__.str>',
                                          'longname': 'normalize',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'nor'},
                            'numItems': {'args': '<type __builtin__.int>',
                                         'longname': 'numItems',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ni'},
                            'preserveAspectRatio': {'args': '<type __builtin__.bool>',
                                                    'longname': 'preserveAspectRatio',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'par'},
                            'unitize': {'args': '<type __builtin__.bool>',
                                        'longname': 'unitize',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'uni'},
                            'unshare': {'args': '<type __builtin__.bool>',
                                        'longname': 'unshare',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'u'},
                            'uvSetName': {'args': '<type __builtin__.str>',
                                          'longname': 'uvSetName',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'uvs'}},
                  'removedFlags': {'global': 'g'},
                  'shortFlags': {'cm': 'createNewMap',
                                 'cp': 'cameraProjection',
                                 'fh': 'flipHorizontal',
                                 'fv': 'flipVertical',
                                 'g': 'g',
                                 'l': 'local',
                                 'ni': 'numItems',
                                 'nor': 'normalize',
                                 'par': 'preserveAspectRatio',
                                 'u': 'unshare',
                                 'uni': 'unitize',
                                 'uvs': 'uvSetName'},
                  'type': 'modeling'},
  'polyGeoSampler': {'flags': {'alphaBlend': {'args': '<type __builtin__.str>',
                                              'longname': 'alphaBlend',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'abl'},
                               'averageColor': {'args': '<type __builtin__.bool>',
                                                'longname': 'averageColor',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'ac'},
                               'clampAlphaMax': {'args': '<type __builtin__.float>',
                                                 'longname': 'clampAlphaMax',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'amx'},
                               'clampAlphaMin': {'args': '<type __builtin__.float>',
                                                 'longname': 'clampAlphaMin',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'amn'},
                               'clampRGBMax': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'clampRGBMax',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'cmx'},
                               'clampRGBMin': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'clampRGBMin',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'cmn'},
                               'colorBlend': {'args': '<type __builtin__.str>',
                                              'longname': 'colorBlend',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cbl'},
                               'colorDisplayOption': {'args': '<type __builtin__.bool>',
                                                      'longname': 'colorDisplayOption',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'cdo'},
                               'computeShadows': {'args': '<type __builtin__.bool>',
                                                  'longname': 'computeShadows',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'cs'},
                               'displaceGeometry': {'args': '<type __builtin__.bool>',
                                                    'longname': 'displaceGeometry',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'dg'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'flatShading': {'args': '<type __builtin__.bool>',
                                               'longname': 'flatShading',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'fs'},
                               'ignoreDoubleSided': {'args': '<type __builtin__.bool>',
                                                     'longname': 'ignoreDoubleSided',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'ids'},
                               'lightingOnly': {'args': '<type __builtin__.bool>',
                                                'longname': 'lightingOnly',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'lo'},
                               'reuseShadows': {'args': '<type __builtin__.bool>',
                                                'longname': 'reuseShadows',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'rs'},
                               'sampleByFace': {'args': '<type __builtin__.bool>',
                                                'longname': 'sampleByFace',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'bf'},
                               'scaleFactor': {'args': '<type __builtin__.float>',
                                               'longname': 'scaleFactor',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sf'},
                               'shareUV': {'args': '<type __builtin__.bool>',
                                           'longname': 'shareUV',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'su'},
                               'useLightShadows': {'args': '<type __builtin__.bool>',
                                                   'longname': 'useLightShadows',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'ul'}},
                     'shortFlags': {'abl': 'alphaBlend',
                                    'ac': 'averageColor',
                                    'amn': 'clampAlphaMin',
                                    'amx': 'clampAlphaMax',
                                    'bf': 'sampleByFace',
                                    'cbl': 'colorBlend',
                                    'cdo': 'colorDisplayOption',
                                    'cmn': 'clampRGBMin',
                                    'cmx': 'clampRGBMax',
                                    'cs': 'computeShadows',
                                    'dg': 'displaceGeometry',
                                    'e': 'edit',
                                    'fs': 'flatShading',
                                    'ids': 'ignoreDoubleSided',
                                    'lo': 'lightingOnly',
                                    'rs': 'reuseShadows',
                                    'sf': 'scaleFactor',
                                    'su': 'shareUV',
                                    'ul': 'useLightShadows'},
                     'type': 'modeling'},
  'polyHelix': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'axis',
                                   'modes': ['query', 'edit'],
                                   'numArgs': 3,
                                   'shortname': 'ax'},
                          'caching': {'args': '<type __builtin__.bool>',
                                      'longname': 'caching',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsCasting': True,
                                      'shortname': 'cch'},
                          'coils': {'args': '<type __builtin__.float>',
                                    'longname': 'coils',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'c'},
                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'constructionHistory',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                          'createUVs': {'args': '<type __builtin__.int>',
                                        'longname': 'createUVs',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'cuv'},
                          'direction': {'args': '<type __builtin__.int>',
                                        'longname': 'direction',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'd'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'frozen': {'args': '<type __builtin__.bool>',
                                     'longname': 'frozen',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'fzn'},
                          'height': {'args': '<type __builtin__.float>',
                                     'longname': 'height',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'h'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nodeState': {'args': '<type __builtin__.int>',
                                        'longname': 'nodeState',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'nds'},
                          'object': {'args': '<type __builtin__.bool>',
                                     'longname': 'object',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'o'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'radius': {'args': '<type __builtin__.float>',
                                     'longname': 'radius',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'r'},
                          'roundCap': {'args': '<type __builtin__.bool>',
                                       'longname': 'roundCap',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'rcp'},
                          'subdivisionsAxis': {'args': '<type __builtin__.int>',
                                               'longname': 'subdivisionsAxis',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'resultNeedsCasting': True,
                                               'shortname': 'sa'},
                          'subdivisionsCaps': {'args': '<type __builtin__.int>',
                                               'longname': 'subdivisionsCaps',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'resultNeedsCasting': True,
                                               'shortname': 'sc'},
                          'subdivisionsCoil': {'args': '<type __builtin__.int>',
                                               'longname': 'subdivisionsCoil',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'resultNeedsCasting': True,
                                               'shortname': 'sco'},
                          'texture': {'args': '<type __builtin__.int>',
                                      'longname': 'texture',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsCasting': True,
                                      'shortname': 'tx'},
                          'useOldInitBehaviour': {'args': '<type __builtin__.bool>',
                                                  'longname': 'useOldInitBehaviour',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'resultNeedsCasting': True,
                                                  'shortname': 'oib'},
                          'width': {'args': '<type __builtin__.float>',
                                    'longname': 'width',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'w'}},
                'shortFlags': {'ax': 'axis',
                               'c': 'coils',
                               'cch': 'caching',
                               'ch': 'constructionHistory',
                               'cuv': 'createUVs',
                               'd': 'direction',
                               'e': 'edit',
                               'fzn': 'frozen',
                               'h': 'height',
                               'n': 'name',
                               'nds': 'nodeState',
                               'o': 'object',
                               'oib': 'useOldInitBehaviour',
                               'q': 'query',
                               'r': 'radius',
                               'rcp': 'roundCap',
                               'sa': 'subdivisionsAxis',
                               'sc': 'subdivisionsCaps',
                               'sco': 'subdivisionsCoil',
                               'tx': 'texture',
                               'w': 'width'},
                'type': 'modeling'},
  'polyHole': {'flags': {'assignHole': {'args': '<type __builtin__.bool>',
                                        'longname': 'assignHole',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ah'},
                         'createHistory': {'args': '<type __builtin__.bool>',
                                           'longname': 'createHistory',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ch'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'}},
               'shortFlags': {'ah': 'assignHole',
                              'ch': 'createHistory',
                              'e': 'edit',
                              'q': 'query'},
               'type': 'modeling'},
  'polyInfo': {'flags': {'edgeToFace': {'args': '<type __builtin__.bool>',
                                        'longname': 'edgeToFace',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ef'},
                         'edgeToVertex': {'args': '<type __builtin__.bool>',
                                          'longname': 'edgeToVertex',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ev'},
                         'faceNormals': {'args': '<type __builtin__.bool>',
                                         'longname': 'faceNormals',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'fn'},
                         'faceToEdge': {'args': '<type __builtin__.bool>',
                                        'longname': 'faceToEdge',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'fe'},
                         'faceToVertex': {'args': '<type __builtin__.bool>',
                                          'longname': 'faceToVertex',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'fv'},
                         'invalidEdges': {'args': '<type __builtin__.bool>',
                                          'longname': 'invalidEdges',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ie'},
                         'invalidVertices': {'args': '<type __builtin__.bool>',
                                             'longname': 'invalidVertices',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'iv'},
                         'laminaFaces': {'args': '<type __builtin__.bool>',
                                         'longname': 'laminaFaces',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'lf'},
                         'nonManifoldEdges': {'args': '<type __builtin__.bool>',
                                              'longname': 'nonManifoldEdges',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'nme'},
                         'nonManifoldUVEdges': {'args': '<type __builtin__.bool>',
                                                'longname': 'nonManifoldUVEdges',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'nue'},
                         'nonManifoldUVs': {'args': '<type __builtin__.bool>',
                                            'longname': 'nonManifoldUVs',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'nuv'},
                         'nonManifoldVertices': {'args': '<type __builtin__.bool>',
                                                 'longname': 'nonManifoldVertices',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'nmv'},
                         'vertexToEdge': {'args': '<type __builtin__.bool>',
                                          'longname': 'vertexToEdge',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 've'},
                         'vertexToFace': {'args': '<type __builtin__.bool>',
                                          'longname': 'vertexToFace',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'vf'}},
               'shortFlags': {'ef': 'edgeToFace',
                              'ev': 'edgeToVertex',
                              'fe': 'faceToEdge',
                              'fn': 'faceNormals',
                              'fv': 'faceToVertex',
                              'ie': 'invalidEdges',
                              'iv': 'invalidVertices',
                              'lf': 'laminaFaces',
                              'nme': 'nonManifoldEdges',
                              'nmv': 'nonManifoldVertices',
                              'nue': 'nonManifoldUVEdges',
                              'nuv': 'nonManifoldUVs',
                              've': 'vertexToEdge',
                              'vf': 'vertexToFace'},
               'type': 'modeling'},
  'polyInstallAction': {'callbackFlags': ['commandName', 'cn'],
                        'flags': {'commandName': {'args': '<type __builtin__.bool>',
                                                  'longname': 'commandName',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'cn'},
                                  'convertSelection': {'args': '<type __builtin__.bool>',
                                                       'longname': 'convertSelection',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'cs'},
                                  'installConstraint': {'args': '<type __builtin__.bool>',
                                                        'longname': 'installConstraint',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 0,
                                                        'shortname': 'ic'},
                                  'installDisplay': {'args': '<type __builtin__.bool>',
                                                     'longname': 'installDisplay',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 0,
                                                     'shortname': 'id'},
                                  'keepInstances': {'args': '<type __builtin__.bool>',
                                                    'longname': 'keepInstances',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ki'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'uninstallConstraint': {'args': '<type __builtin__.bool>',
                                                          'longname': 'uninstallConstraint',
                                                          'modes': ['create'],
                                                          'numArgs': 0,
                                                          'shortname': 'uc'},
                                  'uninstallDisplay': {'args': '<type __builtin__.bool>',
                                                       'longname': 'uninstallDisplay',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'ud'}},
                        'shortFlags': {'cn': 'commandName',
                                       'cs': 'convertSelection',
                                       'ic': 'installConstraint',
                                       'id': 'installDisplay',
                                       'ki': 'keepInstances',
                                       'q': 'query',
                                       'uc': 'uninstallConstraint',
                                       'ud': 'uninstallDisplay'},
                        'type': 'modeling'},
  'polyIterOnPoly': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'polyLayoutUV': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'flipReversed': {'args': '<type __builtin__.bool>',
                                              'longname': 'flipReversed',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fr'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'gridU': {'args': '<type __builtin__.int>',
                                       'longname': 'gridU',
                                       'numArgs': 1,
                                       'shortname': 'gu'},
                             'gridV': {'args': '<type __builtin__.int>',
                                       'longname': 'gridV',
                                       'numArgs': 1,
                                       'shortname': 'gv'},
                             'layout': {'args': '<type __builtin__.int>',
                                        'longname': 'layout',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'l'},
                             'layoutMethod': {'args': '<type __builtin__.int>',
                                              'longname': 'layoutMethod',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lm'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'percentageSpace': {'args': '<type __builtin__.float>',
                                                 'longname': 'percentageSpace',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ps'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'rotateForBestFit': {'args': '<type __builtin__.int>',
                                                  'longname': 'rotateForBestFit',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'rbf'},
                             'scale': {'args': '<type __builtin__.int>',
                                       'longname': 'scale',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sc'},
                             'separate': {'args': '<type __builtin__.int>',
                                          'longname': 'separate',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'se'},
                             'uvSetName': {'args': '<type __builtin__.str>',
                                           'longname': 'uvSetName',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'uvs'},
                             'worldSpace': {'args': '<type __builtin__.bool>',
                                            'longname': 'worldSpace',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ws'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'e': 'edit',
                                  'fr': 'flipReversed',
                                  'fzn': 'frozen',
                                  'gu': 'gridU',
                                  'gv': 'gridV',
                                  'l': 'layout',
                                  'lm': 'layoutMethod',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'ps': 'percentageSpace',
                                  'q': 'query',
                                  'rbf': 'rotateForBestFit',
                                  'sc': 'scale',
                                  'se': 'separate',
                                  'uvs': 'uvSetName',
                                  'ws': 'worldSpace'},
                   'type': 'modeling'},
  'polyListComponentConversion': {'flags': {'border': {'args': '<type __builtin__.bool>',
                                                       'longname': 'border',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'bo'},
                                            'fromEdge': {'args': '<type __builtin__.bool>',
                                                         'longname': 'fromEdge',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'fe'},
                                            'fromFace': {'args': '<type __builtin__.bool>',
                                                         'longname': 'fromFace',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'ff'},
                                            'fromUV': {'args': '<type __builtin__.bool>',
                                                       'longname': 'fromUV',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'fuv'},
                                            'fromVertex': {'args': '<type __builtin__.bool>',
                                                           'longname': 'fromVertex',
                                                           'modes': ['create'],
                                                           'numArgs': 0,
                                                           'shortname': 'fv'},
                                            'fromVertexFace': {'args': '<type __builtin__.bool>',
                                                               'longname': 'fromVertexFace',
                                                               'modes': ['create'],
                                                               'numArgs': 0,
                                                               'shortname': 'fvf'},
                                            'internal': {'args': '<type __builtin__.bool>',
                                                         'longname': 'internal',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'internal'},
                                            'toEdge': {'args': '<type __builtin__.bool>',
                                                       'longname': 'toEdge',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'te'},
                                            'toFace': {'args': '<type __builtin__.bool>',
                                                       'longname': 'toFace',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'tf'},
                                            'toUV': {'args': '<type __builtin__.bool>',
                                                     'longname': 'toUV',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'tuv'},
                                            'toVertex': {'args': '<type __builtin__.bool>',
                                                         'longname': 'toVertex',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'tv'},
                                            'toVertexFace': {'args': '<type __builtin__.bool>',
                                                             'longname': 'toVertexFace',
                                                             'modes': ['create'],
                                                             'numArgs': 0,
                                                             'shortname': 'tvf'},
                                            'uvShell': {'args': '<type __builtin__.bool>',
                                                        'longname': 'uvShell',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'uvs'},
                                            'vertexFaceAllEdges': {'args': '<type __builtin__.bool>',
                                                                   'longname': 'vertexFaceAllEdges',
                                                                   'modes': ['create'],
                                                                   'numArgs': 0,
                                                                   'shortname': 'vfa'}},
                                  'removedFlags': {'in': 'internal'},
                                  'shortFlags': {'bo': 'border',
                                                 'fe': 'fromEdge',
                                                 'ff': 'fromFace',
                                                 'fuv': 'fromUV',
                                                 'fv': 'fromVertex',
                                                 'fvf': 'fromVertexFace',
                                                 'internal': 'internal',
                                                 'te': 'toEdge',
                                                 'tf': 'toFace',
                                                 'tuv': 'toUV',
                                                 'tv': 'toVertex',
                                                 'tvf': 'toVertexFace',
                                                 'uvs': 'uvShell',
                                                 'vfa': 'vertexFaceAllEdges'},
                                  'type': 'modeling'},
  'polyMapCut': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'moveRatio': {'args': '<type __builtin__.float>',
                                         'longname': 'moveRatio',
                                         'numArgs': 1,
                                         'shortname': 'mr'},
                           'moveratio': {'args': '<type __builtin__.float>',
                                         'longname': 'moveratio',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'mvr'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'cch': 'caching',
                                'ch': 'constructionHistory',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'mr': 'moveRatio',
                                'mvr': 'moveratio',
                                'n': 'name',
                                'nds': 'nodeState',
                                'q': 'query'},
                 'type': 'modeling'},
  'polyMapDel': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'cch': 'caching',
                                'ch': 'constructionHistory',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'n': 'name',
                                'nds': 'nodeState',
                                'q': 'query'},
                 'type': 'modeling'},
  'polyMapSew': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'cch': 'caching',
                                'ch': 'constructionHistory',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'n': 'name',
                                'nds': 'nodeState',
                                'q': 'query'},
                 'type': 'modeling'},
  'polyMapSewMove': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'limitPieceSize': {'args': '<type __builtin__.bool>',
                                                  'longname': 'limitPieceSize',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'lps'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'numberFaces': {'args': '<type __builtin__.int>',
                                               'longname': 'numberFaces',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nf'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'uvSetName': {'args': '<type __builtin__.str>',
                                             'longname': 'uvSetName',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'uvs'},
                               'worldSpace': {'args': '<type __builtin__.bool>',
                                              'longname': 'worldSpace',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ws'}},
                     'shortFlags': {'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'e': 'edit',
                                    'fzn': 'frozen',
                                    'lps': 'limitPieceSize',
                                    'n': 'name',
                                    'nds': 'nodeState',
                                    'nf': 'numberFaces',
                                    'q': 'query',
                                    'uvs': 'uvSetName',
                                    'ws': 'worldSpace'},
                     'type': 'modeling'},
  'polyMergeEdge': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cch'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'firstEdge': {'args': '<type __builtin__.int>',
                                            'longname': 'firstEdge',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'fe'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'mergeMode': {'args': '<type __builtin__.int>',
                                            'longname': 'mergeMode',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'mm'},
                              'mergeTexture': {'args': '<type __builtin__.bool>',
                                               'longname': 'mergeTexture',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'mt'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nds'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'secondEdge': {'args': '<type __builtin__.int>',
                                             'longname': 'secondEdge',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'se'}},
                    'shortFlags': {'cch': 'caching',
                                   'ch': 'constructionHistory',
                                   'e': 'edit',
                                   'fe': 'firstEdge',
                                   'fzn': 'frozen',
                                   'mm': 'mergeMode',
                                   'mt': 'mergeTexture',
                                   'n': 'name',
                                   'nds': 'nodeState',
                                   'q': 'query',
                                   'se': 'secondEdge'},
                    'type': 'modeling'},
  'polyMergeEdgeCtx': {'flags': {'activeNodes': {'args': '<type __builtin__.bool>',
                                                 'longname': 'activeNodes',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'anq'},
                                 'caching': {'args': '<type __builtin__.bool>',
                                             'longname': 'caching',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cch'},
                                 'constructionHistory': {'args': '<type __builtin__.bool>',
                                                         'longname': 'constructionHistory',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'ch'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'firstEdge': {'args': '<type __builtin__.int>',
                                               'longname': 'firstEdge',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'fe'},
                                 'frozen': {'args': '<type __builtin__.bool>',
                                            'longname': 'frozen',
                                            'numArgs': 1,
                                            'shortname': 'fzn'},
                                 'image1': {'args': '<type __builtin__.str>',
                                            'longname': 'image1',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i1'},
                                 'image2': {'args': '<type __builtin__.str>',
                                            'longname': 'image2',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i2'},
                                 'image3': {'args': '<type __builtin__.str>',
                                            'longname': 'image3',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i3'},
                                 'immediate': {'args': '<type __builtin__.bool>',
                                               'longname': 'immediate',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'im'},
                                 'mergeMode': {'args': '<type __builtin__.int>',
                                               'longname': 'mergeMode',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'mm'},
                                 'mergeTexture': {'args': '<type __builtin__.bool>',
                                                  'longname': 'mergeTexture',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mt'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'nodeState': {'args': '<type __builtin__.int>',
                                               'longname': 'nodeState',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nds'},
                                 'previous': {'args': '<type __builtin__.bool>',
                                              'longname': 'previous',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'pv'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'reset': {'args': '<type __builtin__.bool>',
                                           'longname': 'reset',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'rs'},
                                 'secondEdge': {'args': '<type __builtin__.int>',
                                                'longname': 'secondEdge',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'se'},
                                 'toolNode': {'args': '<type __builtin__.bool>',
                                              'longname': 'toolNode',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'tnq'}},
                       'shortFlags': {'anq': 'activeNodes',
                                      'cch': 'caching',
                                      'ch': 'constructionHistory',
                                      'e': 'edit',
                                      'ex': 'exists',
                                      'fe': 'firstEdge',
                                      'fzn': 'frozen',
                                      'i1': 'image1',
                                      'i2': 'image2',
                                      'i3': 'image3',
                                      'im': 'immediate',
                                      'mm': 'mergeMode',
                                      'mt': 'mergeTexture',
                                      'n': 'name',
                                      'nds': 'nodeState',
                                      'pv': 'previous',
                                      'q': 'query',
                                      'rs': 'reset',
                                      'se': 'secondEdge',
                                      'tnq': 'toolNode'},
                       'type': 'context'},
  'polyMergeFacet': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'firstFacet': {'args': '<type __builtin__.int>',
                                              'longname': 'firstFacet',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ff'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'mergeMode': {'args': '<type __builtin__.int>',
                                             'longname': 'mergeMode',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'mm'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'secondFacet': {'args': '<type __builtin__.int>',
                                               'longname': 'secondFacet',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sf'}},
                     'shortFlags': {'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'e': 'edit',
                                    'ff': 'firstFacet',
                                    'fzn': 'frozen',
                                    'mm': 'mergeMode',
                                    'n': 'name',
                                    'nds': 'nodeState',
                                    'q': 'query',
                                    'sf': 'secondFacet'},
                     'type': 'modeling'},
  'polyMergeFacetCtx': {'flags': {'activeNodes': {'args': '<type __builtin__.bool>',
                                                  'longname': 'activeNodes',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'anq'},
                                  'caching': {'args': '<type __builtin__.bool>',
                                              'longname': 'caching',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cch'},
                                  'constructionHistory': {'args': '<type __builtin__.bool>',
                                                          'longname': 'constructionHistory',
                                                          'modes': ['create',
                                                                    'query'],
                                                          'numArgs': 1,
                                                          'shortname': 'ch'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'firstFacet': {'args': '<type __builtin__.int>',
                                                 'longname': 'firstFacet',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ff'},
                                  'frozen': {'args': '<type __builtin__.bool>',
                                             'longname': 'frozen',
                                             'numArgs': 1,
                                             'shortname': 'fzn'},
                                  'image1': {'args': '<type __builtin__.str>',
                                             'longname': 'image1',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i1'},
                                  'image2': {'args': '<type __builtin__.str>',
                                             'longname': 'image2',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i2'},
                                  'image3': {'args': '<type __builtin__.str>',
                                             'longname': 'image3',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i3'},
                                  'immediate': {'args': '<type __builtin__.bool>',
                                                'longname': 'immediate',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'im'},
                                  'mergeMode': {'args': '<type __builtin__.int>',
                                                'longname': 'mergeMode',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'mm'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'nodeState': {'args': '<type __builtin__.int>',
                                                'longname': 'nodeState',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nds'},
                                  'previous': {'args': '<type __builtin__.bool>',
                                               'longname': 'previous',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'pv'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'reset': {'args': '<type __builtin__.bool>',
                                            'longname': 'reset',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'rs'},
                                  'secondFacet': {'args': '<type __builtin__.int>',
                                                  'longname': 'secondFacet',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sf'},
                                  'toolNode': {'args': '<type __builtin__.bool>',
                                               'longname': 'toolNode',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'tnq'}},
                        'shortFlags': {'anq': 'activeNodes',
                                       'cch': 'caching',
                                       'ch': 'constructionHistory',
                                       'e': 'edit',
                                       'ex': 'exists',
                                       'ff': 'firstFacet',
                                       'fzn': 'frozen',
                                       'i1': 'image1',
                                       'i2': 'image2',
                                       'i3': 'image3',
                                       'im': 'immediate',
                                       'mm': 'mergeMode',
                                       'n': 'name',
                                       'nds': 'nodeState',
                                       'pv': 'previous',
                                       'q': 'query',
                                       'rs': 'reset',
                                       'sf': 'secondFacet',
                                       'tnq': 'toolNode'},
                        'type': 'context'},
  'polyMergeUV': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'distance': {'args': '<type __builtin__.float>',
                                         'longname': 'distance',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'd'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'uvSetName': {'args': '<type __builtin__.str>',
                                          'longname': 'uvSetName',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'uvs'},
                            'worldSpace': {'args': '<type __builtin__.bool>',
                                           'longname': 'worldSpace',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ws'}},
                  'shortFlags': {'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'd': 'distance',
                                 'e': 'edit',
                                 'fzn': 'frozen',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'q': 'query',
                                 'uvs': 'uvSetName',
                                 'ws': 'worldSpace'},
                  'type': 'modeling'},
  'polyMergeVertex': {'flags': {'alwaysMergeTwoVertices': {'args': '<type __builtin__.bool>',
                                                           'longname': 'alwaysMergeTwoVertices',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'am'},
                                'caching': {'args': '<type __builtin__.bool>',
                                            'longname': 'caching',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cch'},
                                'constructionHistory': {'args': '<type __builtin__.bool>',
                                                        'longname': 'constructionHistory',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'ch'},
                                'distance': {'args': '<type __builtin__.float>',
                                             'longname': 'distance',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'd'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'frozen': {'args': '<type __builtin__.bool>',
                                           'longname': 'frozen',
                                           'numArgs': 1,
                                           'shortname': 'fzn'},
                                'mergeToComponents': {'args': '<type __builtin__.str>',
                                                      'longname': 'mergeToComponents',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'mtc'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'nodeState': {'args': '<type __builtin__.int>',
                                              'longname': 'nodeState',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nds'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'texture': {'args': '<type __builtin__.bool>',
                                            'longname': 'texture',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'tx'},
                                'worldSpace': {'args': '<type __builtin__.bool>',
                                               'longname': 'worldSpace',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ws'}},
                      'shortFlags': {'am': 'alwaysMergeTwoVertices',
                                     'cch': 'caching',
                                     'ch': 'constructionHistory',
                                     'd': 'distance',
                                     'e': 'edit',
                                     'fzn': 'frozen',
                                     'mtc': 'mergeToComponents',
                                     'n': 'name',
                                     'nds': 'nodeState',
                                     'q': 'query',
                                     'tx': 'texture',
                                     'ws': 'worldSpace'},
                      'type': 'modeling'},
  'polyMirrorFace': {'flags': {'axis': {'args': '<type __builtin__.int>',
                                        'longname': 'axis',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'a'},
                               'axisDirection': {'args': '<type __builtin__.int>',
                                                 'longname': 'axisDirection',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ad'},
                               'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'cutMesh': {'args': '<type __builtin__.bool>',
                                           'longname': 'cutMesh',
                                           'numArgs': 1,
                                           'shortname': 'cm'},
                               'direction': {'args': '<type __builtin__.int>',
                                             'longname': 'direction',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'd'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'firstNewFace': {'args': '<type __builtin__.int>',
                                                'longname': 'firstNewFace',
                                                'numArgs': 1,
                                                'shortname': 'fnf'},
                               'flipUVs': {'args': '<type __builtin__.int>',
                                           'longname': 'flipUVs',
                                           'numArgs': 1,
                                           'shortname': 'fuv'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'keepVertexIDs': {'args': '<type __builtin__.bool>',
                                                 'longname': 'keepVertexIDs',
                                                 'numArgs': 1,
                                                 'shortname': 'kv'},
                               'lastNewFace': {'args': '<type __builtin__.int>',
                                               'longname': 'lastNewFace',
                                               'numArgs': 1,
                                               'shortname': 'lnf'},
                               'mergeMode': {'args': '<type __builtin__.int>',
                                             'longname': 'mergeMode',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'mm'},
                               'mergeThreshold': {'args': '<type __builtin__.float>',
                                                  'longname': 'mergeThreshold',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mt'},
                               'mergeThresholdType': {'args': '<type __builtin__.int>',
                                                      'longname': 'mergeThresholdType',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'mtt'},
                               'mirrorAxis': {'args': '<type __builtin__.int>',
                                              'longname': 'mirrorAxis',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ma'},
                               'mirrorPlaneCenter': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'mirrorPlaneCenter',
                                                     'numArgs': 3,
                                                     'shortname': 'pc'},
                               'mirrorPlaneCenterX': {'args': '<type __builtin__.float>',
                                                      'longname': 'mirrorPlaneCenterX',
                                                      'numArgs': 1,
                                                      'shortname': 'pcx'},
                               'mirrorPlaneCenterY': {'args': '<type __builtin__.float>',
                                                      'longname': 'mirrorPlaneCenterY',
                                                      'numArgs': 1,
                                                      'shortname': 'pcy'},
                               'mirrorPlaneCenterZ': {'args': '<type __builtin__.float>',
                                                      'longname': 'mirrorPlaneCenterZ',
                                                      'numArgs': 1,
                                                      'shortname': 'pcz'},
                               'mirrorPlaneRotate': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'mirrorPlaneRotate',
                                                     'numArgs': 3,
                                                     'shortname': 'ro'},
                               'mirrorPlaneRotateX': {'args': '<type __builtin__.float>',
                                                      'longname': 'mirrorPlaneRotateX',
                                                      'numArgs': 1,
                                                      'shortname': 'rx'},
                               'mirrorPlaneRotateY': {'args': '<type __builtin__.float>',
                                                      'longname': 'mirrorPlaneRotateY',
                                                      'numArgs': 1,
                                                      'shortname': 'ry'},
                               'mirrorPlaneRotateZ': {'args': '<type __builtin__.float>',
                                                      'longname': 'mirrorPlaneRotateZ',
                                                      'numArgs': 1,
                                                      'shortname': 'rz'},
                               'mirrorPosition': {'args': '<type __builtin__.float>',
                                                  'longname': 'mirrorPosition',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mps'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'pivot': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'pivot',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 3,
                                         'shortname': 'p'},
                               'pivotX': {'args': '<type __builtin__.float>',
                                          'longname': 'pivotX',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'px'},
                               'pivotY': {'args': '<type __builtin__.float>',
                                          'longname': 'pivotY',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'py'},
                               'pivotZ': {'args': '<type __builtin__.float>',
                                          'longname': 'pivotZ',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pz'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'scalePivotX': {'args': '<type __builtin__.float>',
                                               'longname': 'scalePivotX',
                                               'numArgs': 1,
                                               'shortname': 'spx'},
                               'scalePivotY': {'args': '<type __builtin__.float>',
                                               'longname': 'scalePivotY',
                                               'numArgs': 1,
                                               'shortname': 'spy'},
                               'scalePivotZ': {'args': '<type __builtin__.float>',
                                               'longname': 'scalePivotZ',
                                               'numArgs': 1,
                                               'shortname': 'spz'},
                               'smoothingAngle': {'args': '<type __builtin__.float>',
                                                  'longname': 'smoothingAngle',
                                                  'numArgs': 1,
                                                  'shortname': 'sa'},
                               'userSpecifiedPivot': {'args': '<type __builtin__.bool>',
                                                      'longname': 'userSpecifiedPivot',
                                                      'numArgs': 1,
                                                      'shortname': 'pu'},
                               'worldSpace': {'args': '<type __builtin__.bool>',
                                              'longname': 'worldSpace',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ws'}},
                     'shortFlags': {'a': 'axis',
                                    'ad': 'axisDirection',
                                    'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'cm': 'cutMesh',
                                    'd': 'direction',
                                    'e': 'edit',
                                    'fnf': 'firstNewFace',
                                    'fuv': 'flipUVs',
                                    'fzn': 'frozen',
                                    'kv': 'keepVertexIDs',
                                    'lnf': 'lastNewFace',
                                    'ma': 'mirrorAxis',
                                    'mm': 'mergeMode',
                                    'mps': 'mirrorPosition',
                                    'mt': 'mergeThreshold',
                                    'mtt': 'mergeThresholdType',
                                    'n': 'name',
                                    'nds': 'nodeState',
                                    'p': 'pivot',
                                    'pc': 'mirrorPlaneCenter',
                                    'pcx': 'mirrorPlaneCenterX',
                                    'pcy': 'mirrorPlaneCenterY',
                                    'pcz': 'mirrorPlaneCenterZ',
                                    'pu': 'userSpecifiedPivot',
                                    'px': 'pivotX',
                                    'py': 'pivotY',
                                    'pz': 'pivotZ',
                                    'q': 'query',
                                    'ro': 'mirrorPlaneRotate',
                                    'rx': 'mirrorPlaneRotateX',
                                    'ry': 'mirrorPlaneRotateY',
                                    'rz': 'mirrorPlaneRotateZ',
                                    'sa': 'smoothingAngle',
                                    'spx': 'scalePivotX',
                                    'spy': 'scalePivotY',
                                    'spz': 'scalePivotZ',
                                    'ws': 'worldSpace'},
                     'type': 'modeling'},
  'polyMoveEdge': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'gain': {'args': '<type __builtin__.float>',
                                      'longname': 'gain',
                                      'modes': ['create',
                                                'query',
                                                'edit',
                                                'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'ga'},
                             'localCenter': {'args': '<type __builtin__.int>',
                                             'longname': 'localCenter',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lc'},
                             'localDirection': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'localDirection',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'ld'},
                             'localDirectionX': {'args': '<type __builtin__.float>',
                                                 'longname': 'localDirectionX',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ldx'},
                             'localDirectionY': {'args': '<type __builtin__.float>',
                                                 'longname': 'localDirectionY',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ldy'},
                             'localDirectionZ': {'args': '<type __builtin__.float>',
                                                 'longname': 'localDirectionZ',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ldz'},
                             'localRotate': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'localRotate',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'lr'},
                             'localRotateX': {'args': '<type __builtin__.float>',
                                              'longname': 'localRotateX',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lrx'},
                             'localRotateY': {'args': '<type __builtin__.float>',
                                              'longname': 'localRotateY',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lry'},
                             'localRotateZ': {'args': '<type __builtin__.float>',
                                              'longname': 'localRotateZ',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lrz'},
                             'localScale': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'localScale',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'ls'},
                             'localScaleX': {'args': '<type __builtin__.float>',
                                             'longname': 'localScaleX',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lsx'},
                             'localScaleY': {'args': '<type __builtin__.float>',
                                             'longname': 'localScaleY',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lsy'},
                             'localScaleZ': {'args': '<type __builtin__.float>',
                                             'longname': 'localScaleZ',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lsz'},
                             'localTranslate': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'localTranslate',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'lt'},
                             'localTranslateX': {'args': '<type __builtin__.float>',
                                                 'longname': 'localTranslateX',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ltx'},
                             'localTranslateY': {'args': '<type __builtin__.float>',
                                                 'longname': 'localTranslateY',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'lty'},
                             'localTranslateZ': {'args': '<type __builtin__.float>',
                                                 'longname': 'localTranslateZ',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ltz'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'pivot': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'pivot',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 3,
                                       'shortname': 'pvt'},
                             'pivotX': {'args': '<type __builtin__.float>',
                                        'longname': 'pivotX',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'pvx'},
                             'pivotY': {'args': '<type __builtin__.float>',
                                        'longname': 'pivotY',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'pvy'},
                             'pivotZ': {'args': '<type __builtin__.float>',
                                        'longname': 'pivotZ',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'pvz'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'random': {'args': '<type __builtin__.float>',
                                        'longname': 'random',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ran'},
                             'rotate': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'rotate',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 3,
                                        'shortname': 'ro'},
                             'rotateX': {'args': '<type __builtin__.float>',
                                         'longname': 'rotateX',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'rx'},
                             'rotateY': {'args': '<type __builtin__.float>',
                                         'longname': 'rotateY',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ry'},
                             'rotateZ': {'args': '<type __builtin__.float>',
                                         'longname': 'rotateZ',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'rz'},
                             'scale': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'scale',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 3,
                                       'shortname': 's'},
                             'scaleX': {'args': '<type __builtin__.float>',
                                        'longname': 'scaleX',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'sx'},
                             'scaleY': {'args': '<type __builtin__.float>',
                                        'longname': 'scaleY',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'sy'},
                             'scaleZ': {'args': '<type __builtin__.float>',
                                        'longname': 'scaleZ',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'sz'},
                             'translate': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'translate',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 't'},
                             'translateX': {'args': '<type __builtin__.float>',
                                            'longname': 'translateX',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'tx'},
                             'translateY': {'args': '<type __builtin__.float>',
                                            'longname': 'translateY',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ty'},
                             'translateZ': {'args': '<type __builtin__.float>',
                                            'longname': 'translateZ',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'tz'},
                             'worldSpace': {'args': '<type __builtin__.bool>',
                                            'longname': 'worldSpace',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ws'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'ga': 'gain',
                                  'lc': 'localCenter',
                                  'ld': 'localDirection',
                                  'ldx': 'localDirectionX',
                                  'ldy': 'localDirectionY',
                                  'ldz': 'localDirectionZ',
                                  'lr': 'localRotate',
                                  'lrx': 'localRotateX',
                                  'lry': 'localRotateY',
                                  'lrz': 'localRotateZ',
                                  'ls': 'localScale',
                                  'lsx': 'localScaleX',
                                  'lsy': 'localScaleY',
                                  'lsz': 'localScaleZ',
                                  'lt': 'localTranslate',
                                  'ltx': 'localTranslateX',
                                  'lty': 'localTranslateY',
                                  'ltz': 'localTranslateZ',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'pvt': 'pivot',
                                  'pvx': 'pivotX',
                                  'pvy': 'pivotY',
                                  'pvz': 'pivotZ',
                                  'q': 'query',
                                  'ran': 'random',
                                  'ro': 'rotate',
                                  'rx': 'rotateX',
                                  'ry': 'rotateY',
                                  'rz': 'rotateZ',
                                  's': 'scale',
                                  'sx': 'scaleX',
                                  'sy': 'scaleY',
                                  'sz': 'scaleZ',
                                  't': 'translate',
                                  'tx': 'translateX',
                                  'ty': 'translateY',
                                  'tz': 'translateZ',
                                  'ws': 'worldSpace'},
                   'type': 'modeling'},
  'polyMoveFacet': {'flags': {'attraction': {'args': '<type __builtin__.float>',
                                             'longname': 'attraction',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'att'},
                              'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cch'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'gain': {'args': '<type __builtin__.float>',
                                       'longname': 'gain',
                                       'modes': ['create',
                                                 'query',
                                                 'edit',
                                                 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'ga'},
                              'gravity': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'gravity',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 3,
                                          'shortname': 'g'},
                              'gravityX': {'args': '<type __builtin__.float>',
                                           'longname': 'gravityX',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'gx'},
                              'gravityY': {'args': '<type __builtin__.float>',
                                           'longname': 'gravityY',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'gy'},
                              'gravityZ': {'args': '<type __builtin__.float>',
                                           'longname': 'gravityZ',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'gz'},
                              'localCenter': {'args': '<type __builtin__.int>',
                                              'longname': 'localCenter',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lc'},
                              'localDirection': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'localDirection',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'ld'},
                              'localDirectionX': {'args': '<type __builtin__.float>',
                                                  'longname': 'localDirectionX',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ldx'},
                              'localDirectionY': {'args': '<type __builtin__.float>',
                                                  'longname': 'localDirectionY',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ldy'},
                              'localDirectionZ': {'args': '<type __builtin__.float>',
                                                  'longname': 'localDirectionZ',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ldz'},
                              'localRotate': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'localRotate',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'lr'},
                              'localRotateX': {'args': '<type __builtin__.float>',
                                               'longname': 'localRotateX',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'lrx'},
                              'localRotateY': {'args': '<type __builtin__.float>',
                                               'longname': 'localRotateY',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'lry'},
                              'localRotateZ': {'args': '<type __builtin__.float>',
                                               'longname': 'localRotateZ',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'lrz'},
                              'localScale': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'localScale',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'ls'},
                              'localScaleX': {'args': '<type __builtin__.float>',
                                              'longname': 'localScaleX',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lsx'},
                              'localScaleY': {'args': '<type __builtin__.float>',
                                              'longname': 'localScaleY',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lsy'},
                              'localScaleZ': {'args': '<type __builtin__.float>',
                                              'longname': 'localScaleZ',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lsz'},
                              'localTranslate': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'localTranslate',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'lt'},
                              'localTranslateX': {'args': '<type __builtin__.float>',
                                                  'longname': 'localTranslateX',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ltx'},
                              'localTranslateY': {'args': '<type __builtin__.float>',
                                                  'longname': 'localTranslateY',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'lty'},
                              'localTranslateZ': {'args': '<type __builtin__.float>',
                                                  'longname': 'localTranslateZ',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ltz'},
                              'magnX': {'args': '<type __builtin__.float>',
                                        'longname': 'magnX',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'mx'},
                              'magnY': {'args': '<type __builtin__.float>',
                                        'longname': 'magnY',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'my'},
                              'magnZ': {'args': '<type __builtin__.float>',
                                        'longname': 'magnZ',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'mz'},
                              'magnet': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'magnet',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 3,
                                         'shortname': 'm'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nds'},
                              'offset': {'args': '<type __builtin__.float>',
                                         'longname': 'offset',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'off'},
                              'pivot': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'pivot',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 3,
                                        'shortname': 'pvt'},
                              'pivotX': {'args': '<type __builtin__.float>',
                                         'longname': 'pivotX',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'pvx'},
                              'pivotY': {'args': '<type __builtin__.float>',
                                         'longname': 'pivotY',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'pvy'},
                              'pivotZ': {'args': '<type __builtin__.float>',
                                         'longname': 'pivotZ',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'pvz'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'random': {'args': '<type __builtin__.float>',
                                         'longname': 'random',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ran'},
                              'rotate': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'rotate',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 3,
                                         'shortname': 'ro'},
                              'rotateX': {'args': '<type __builtin__.float>',
                                          'longname': 'rotateX',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'rx'},
                              'rotateY': {'args': '<type __builtin__.float>',
                                          'longname': 'rotateY',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ry'},
                              'rotateZ': {'args': '<type __builtin__.float>',
                                          'longname': 'rotateZ',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'rz'},
                              'scale': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'scale',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 3,
                                        'shortname': 's'},
                              'scaleX': {'args': '<type __builtin__.float>',
                                         'longname': 'scaleX',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'sx'},
                              'scaleY': {'args': '<type __builtin__.float>',
                                         'longname': 'scaleY',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'sy'},
                              'scaleZ': {'args': '<type __builtin__.float>',
                                         'longname': 'scaleZ',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'sz'},
                              'translate': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'translate',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 't'},
                              'translateX': {'args': '<type __builtin__.float>',
                                             'longname': 'translateX',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'tx'},
                              'translateY': {'args': '<type __builtin__.float>',
                                             'longname': 'translateY',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ty'},
                              'translateZ': {'args': '<type __builtin__.float>',
                                             'longname': 'translateZ',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'tz'},
                              'weight': {'args': '<type __builtin__.float>',
                                         'longname': 'weight',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'w'},
                              'worldSpace': {'args': '<type __builtin__.bool>',
                                             'longname': 'worldSpace',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ws'}},
                    'shortFlags': {'att': 'attraction',
                                   'cch': 'caching',
                                   'ch': 'constructionHistory',
                                   'e': 'edit',
                                   'fzn': 'frozen',
                                   'g': 'gravity',
                                   'ga': 'gain',
                                   'gx': 'gravityX',
                                   'gy': 'gravityY',
                                   'gz': 'gravityZ',
                                   'lc': 'localCenter',
                                   'ld': 'localDirection',
                                   'ldx': 'localDirectionX',
                                   'ldy': 'localDirectionY',
                                   'ldz': 'localDirectionZ',
                                   'lr': 'localRotate',
                                   'lrx': 'localRotateX',
                                   'lry': 'localRotateY',
                                   'lrz': 'localRotateZ',
                                   'ls': 'localScale',
                                   'lsx': 'localScaleX',
                                   'lsy': 'localScaleY',
                                   'lsz': 'localScaleZ',
                                   'lt': 'localTranslate',
                                   'ltx': 'localTranslateX',
                                   'lty': 'localTranslateY',
                                   'ltz': 'localTranslateZ',
                                   'm': 'magnet',
                                   'mx': 'magnX',
                                   'my': 'magnY',
                                   'mz': 'magnZ',
                                   'n': 'name',
                                   'nds': 'nodeState',
                                   'off': 'offset',
                                   'pvt': 'pivot',
                                   'pvx': 'pivotX',
                                   'pvy': 'pivotY',
                                   'pvz': 'pivotZ',
                                   'q': 'query',
                                   'ran': 'random',
                                   'ro': 'rotate',
                                   'rx': 'rotateX',
                                   'ry': 'rotateY',
                                   'rz': 'rotateZ',
                                   's': 'scale',
                                   'sx': 'scaleX',
                                   'sy': 'scaleY',
                                   'sz': 'scaleZ',
                                   't': 'translate',
                                   'tx': 'translateX',
                                   'ty': 'translateY',
                                   'tz': 'translateZ',
                                   'w': 'weight',
                                   'ws': 'worldSpace'},
                    'type': 'modeling'},
  'polyMoveFacetUV': {'flags': {'axisLen': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'axisLen',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 2,
                                            'shortname': 'l'},
                                'axisLenX': {'args': '<type __builtin__.float>',
                                             'longname': 'axisLenX',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lx'},
                                'axisLenY': {'args': '<type __builtin__.float>',
                                             'longname': 'axisLenY',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ly'},
                                'caching': {'args': '<type __builtin__.bool>',
                                            'longname': 'caching',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cch'},
                                'constructionHistory': {'args': '<type __builtin__.bool>',
                                                        'longname': 'constructionHistory',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'ch'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'frozen': {'args': '<type __builtin__.bool>',
                                           'longname': 'frozen',
                                           'numArgs': 1,
                                           'shortname': 'fzn'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'nodeState': {'args': '<type __builtin__.int>',
                                              'longname': 'nodeState',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nds'},
                                'pivot': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'pivot',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 2,
                                          'shortname': 'pvt'},
                                'pivotU': {'args': '<type __builtin__.float>',
                                           'longname': 'pivotU',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pvu'},
                                'pivotV': {'args': '<type __builtin__.float>',
                                           'longname': 'pivotV',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pvv'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'random': {'args': '<type __builtin__.float>',
                                           'longname': 'random',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ran'},
                                'rotationAngle': {'args': '<type __builtin__.float>',
                                                  'longname': 'rotationAngle',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ra'},
                                'scale': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'scale',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 2,
                                          'shortname': 's'},
                                'scaleU': {'args': '<type __builtin__.float>',
                                           'longname': 'scaleU',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'su'},
                                'scaleV': {'args': '<type __builtin__.float>',
                                           'longname': 'scaleV',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sv'},
                                'translate': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'translate',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 2,
                                              'shortname': 't'},
                                'translateU': {'args': '<type __builtin__.float>',
                                               'longname': 'translateU',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tu'},
                                'translateV': {'args': '<type __builtin__.float>',
                                               'longname': 'translateV',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tv'}},
                      'shortFlags': {'cch': 'caching',
                                     'ch': 'constructionHistory',
                                     'e': 'edit',
                                     'fzn': 'frozen',
                                     'l': 'axisLen',
                                     'lx': 'axisLenX',
                                     'ly': 'axisLenY',
                                     'n': 'name',
                                     'nds': 'nodeState',
                                     'pvt': 'pivot',
                                     'pvu': 'pivotU',
                                     'pvv': 'pivotV',
                                     'q': 'query',
                                     'ra': 'rotationAngle',
                                     'ran': 'random',
                                     's': 'scale',
                                     'su': 'scaleU',
                                     'sv': 'scaleV',
                                     't': 'translate',
                                     'tu': 'translateU',
                                     'tv': 'translateV'},
                      'type': 'modeling'},
  'polyMoveUV': {'flags': {'axisLen': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'axisLen',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 2,
                                       'shortname': 'l'},
                           'axisLenX': {'args': '<type __builtin__.float>',
                                        'longname': 'axisLenX',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'lx'},
                           'axisLenY': {'args': '<type __builtin__.float>',
                                        'longname': 'axisLenY',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ly'},
                           'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'pivot': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'pivot',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 2,
                                     'shortname': 'pvt'},
                           'pivotU': {'args': '<type __builtin__.float>',
                                      'longname': 'pivotU',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'pvu'},
                           'pivotV': {'args': '<type __builtin__.float>',
                                      'longname': 'pivotV',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'pvv'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'random': {'args': '<type __builtin__.float>',
                                      'longname': 'random',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'ran'},
                           'rotationAngle': {'args': '<type __builtin__.float>',
                                             'longname': 'rotationAngle',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ra'},
                           'scale': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'scale',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 2,
                                     'shortname': 's'},
                           'scaleU': {'args': '<type __builtin__.float>',
                                      'longname': 'scaleU',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'su'},
                           'scaleV': {'args': '<type __builtin__.float>',
                                      'longname': 'scaleV',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'sv'},
                           'translate': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'translate',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 2,
                                         'shortname': 't'},
                           'translateU': {'args': '<type __builtin__.float>',
                                          'longname': 'translateU',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'tu'},
                           'translateV': {'args': '<type __builtin__.float>',
                                          'longname': 'translateV',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'tv'}},
                 'shortFlags': {'cch': 'caching',
                                'ch': 'constructionHistory',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'l': 'axisLen',
                                'lx': 'axisLenX',
                                'ly': 'axisLenY',
                                'n': 'name',
                                'nds': 'nodeState',
                                'pvt': 'pivot',
                                'pvu': 'pivotU',
                                'pvv': 'pivotV',
                                'q': 'query',
                                'ra': 'rotationAngle',
                                'ran': 'random',
                                's': 'scale',
                                'su': 'scaleU',
                                'sv': 'scaleV',
                                't': 'translate',
                                'tu': 'translateU',
                                'tv': 'translateV'},
                 'type': 'modeling'},
  'polyMoveVertex': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'gain': {'args': '<type __builtin__.float>',
                                        'longname': 'gain',
                                        'modes': ['create',
                                                  'query',
                                                  'edit',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'ga'},
                               'localDirection': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'localDirection',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'ld'},
                               'localDirectionX': {'args': '<type __builtin__.float>',
                                                   'longname': 'localDirectionX',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ldx'},
                               'localDirectionY': {'args': '<type __builtin__.float>',
                                                   'longname': 'localDirectionY',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ldy'},
                               'localDirectionZ': {'args': '<type __builtin__.float>',
                                                   'longname': 'localDirectionZ',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ldz'},
                               'localTranslate': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'localTranslate',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'lt'},
                               'localTranslateX': {'args': '<type __builtin__.float>',
                                                   'longname': 'localTranslateX',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ltx'},
                               'localTranslateY': {'args': '<type __builtin__.float>',
                                                   'longname': 'localTranslateY',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'lty'},
                               'localTranslateZ': {'args': '<type __builtin__.float>',
                                                   'longname': 'localTranslateZ',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ltz'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'pivot': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'pivot',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 3,
                                         'shortname': 'pvt'},
                               'pivotX': {'args': '<type __builtin__.float>',
                                          'longname': 'pivotX',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pvx'},
                               'pivotY': {'args': '<type __builtin__.float>',
                                          'longname': 'pivotY',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pvy'},
                               'pivotZ': {'args': '<type __builtin__.float>',
                                          'longname': 'pivotZ',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pvz'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'random': {'args': '<type __builtin__.float>',
                                          'longname': 'random',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ran'},
                               'rotate': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'rotate',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 3,
                                          'shortname': 'ro'},
                               'rotateX': {'args': '<type __builtin__.float>',
                                           'longname': 'rotateX',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'rx'},
                               'rotateY': {'args': '<type __builtin__.float>',
                                           'longname': 'rotateY',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ry'},
                               'rotateZ': {'args': '<type __builtin__.float>',
                                           'longname': 'rotateZ',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'rz'},
                               'scale': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'scale',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 3,
                                         'shortname': 's'},
                               'scaleX': {'args': '<type __builtin__.float>',
                                          'longname': 'scaleX',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sx'},
                               'scaleY': {'args': '<type __builtin__.float>',
                                          'longname': 'scaleY',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sy'},
                               'scaleZ': {'args': '<type __builtin__.float>',
                                          'longname': 'scaleZ',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sz'},
                               'translate': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'translate',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 't'},
                               'translateX': {'args': '<type __builtin__.float>',
                                              'longname': 'translateX',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'tx'},
                               'translateY': {'args': '<type __builtin__.float>',
                                              'longname': 'translateY',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ty'},
                               'translateZ': {'args': '<type __builtin__.float>',
                                              'longname': 'translateZ',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'tz'},
                               'worldSpace': {'args': '<type __builtin__.bool>',
                                              'longname': 'worldSpace',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ws'}},
                     'shortFlags': {'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'e': 'edit',
                                    'fzn': 'frozen',
                                    'ga': 'gain',
                                    'ld': 'localDirection',
                                    'ldx': 'localDirectionX',
                                    'ldy': 'localDirectionY',
                                    'ldz': 'localDirectionZ',
                                    'lt': 'localTranslate',
                                    'ltx': 'localTranslateX',
                                    'lty': 'localTranslateY',
                                    'ltz': 'localTranslateZ',
                                    'n': 'name',
                                    'nds': 'nodeState',
                                    'pvt': 'pivot',
                                    'pvx': 'pivotX',
                                    'pvy': 'pivotY',
                                    'pvz': 'pivotZ',
                                    'q': 'query',
                                    'ran': 'random',
                                    'ro': 'rotate',
                                    'rx': 'rotateX',
                                    'ry': 'rotateY',
                                    'rz': 'rotateZ',
                                    's': 'scale',
                                    'sx': 'scaleX',
                                    'sy': 'scaleY',
                                    'sz': 'scaleZ',
                                    't': 'translate',
                                    'tx': 'translateX',
                                    'ty': 'translateY',
                                    'tz': 'translateZ',
                                    'ws': 'worldSpace'},
                     'type': 'modeling'},
  'polyMultiLayoutUV': {'flags': {'flipReversed': {'args': '<type __builtin__.bool>',
                                                   'longname': 'flipReversed',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'fr'},
                                  'gridU': {'args': '<type __builtin__.int>',
                                            'longname': 'gridU',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'gu'},
                                  'gridV': {'args': '<type __builtin__.int>',
                                            'longname': 'gridV',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'gv'},
                                  'layout': {'args': '<type __builtin__.int>',
                                             'longname': 'layout',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'l'},
                                  'layoutMethod': {'args': '<type __builtin__.int>',
                                                   'longname': 'layoutMethod',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'lm'},
                                  'offsetU': {'args': '<type __builtin__.float>',
                                              'longname': 'offsetU',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ou'},
                                  'offsetV': {'args': '<type __builtin__.float>',
                                              'longname': 'offsetV',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ov'},
                                  'percentageSpace': {'args': '<type __builtin__.float>',
                                                      'longname': 'percentageSpace',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ps'},
                                  'prescale': {'args': '<type __builtin__.int>',
                                               'longname': 'prescale',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'psc'},
                                  'rotateForBestFit': {'args': '<type __builtin__.int>',
                                                       'longname': 'rotateForBestFit',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'rbf'},
                                  'scale': {'args': '<type __builtin__.int>',
                                            'longname': 'scale',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'sc'},
                                  'sizeU': {'args': '<type __builtin__.float>',
                                            'longname': 'sizeU',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'su'},
                                  'sizeV': {'args': '<type __builtin__.float>',
                                            'longname': 'sizeV',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'sv'},
                                  'uvSetName': {'args': '<type __builtin__.str>',
                                                'longname': 'uvSetName',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'uvs'}},
                        'shortFlags': {'fr': 'flipReversed',
                                       'gu': 'gridU',
                                       'gv': 'gridV',
                                       'l': 'layout',
                                       'lm': 'layoutMethod',
                                       'ou': 'offsetU',
                                       'ov': 'offsetV',
                                       'ps': 'percentageSpace',
                                       'psc': 'prescale',
                                       'rbf': 'rotateForBestFit',
                                       'sc': 'scale',
                                       'su': 'sizeU',
                                       'sv': 'sizeV',
                                       'uvs': 'uvSetName'},
                        'type': 'modeling'},
  'polyNormal': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'normalMode': {'args': '<type __builtin__.int>',
                                          'longname': 'normalMode',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nm'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'userNormalMode': {'args': '<type __builtin__.bool>',
                                              'longname': 'userNormalMode',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'unm'}},
                 'shortFlags': {'cch': 'caching',
                                'ch': 'constructionHistory',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'n': 'name',
                                'nds': 'nodeState',
                                'nm': 'normalMode',
                                'q': 'query',
                                'unm': 'userNormalMode'},
                 'type': 'modeling'},
  'polyNormalPerVertex': {'flags': {'allLocked': {'args': '<type __builtin__.bool>',
                                                  'longname': 'allLocked',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'al'},
                                    'deformable': {'args': '<type __builtin__.bool>',
                                                   'longname': 'deformable',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'deformable'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'freezeNormal': {'args': '<type __builtin__.bool>',
                                                     'longname': 'freezeNormal',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'fn'},
                                    'normalX': {'args': '<type __builtin__.float>',
                                                'longname': 'normalX',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'x'},
                                    'normalXYZ': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'normalXYZ',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 3,
                                                  'shortname': 'xyz'},
                                    'normalY': {'args': '<type __builtin__.float>',
                                                'longname': 'normalY',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'y'},
                                    'normalZ': {'args': '<type __builtin__.float>',
                                                'longname': 'normalZ',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'z'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'relative': {'args': '<type __builtin__.bool>',
                                                 'longname': 'relative',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'rel'},
                                    'unFreezeNormal': {'args': '<type __builtin__.bool>',
                                                       'longname': 'unFreezeNormal',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ufn'}},
                          'removedFlags': {'def': 'deformable'},
                          'shortFlags': {'al': 'allLocked',
                                         'deformable': 'deformable',
                                         'e': 'edit',
                                         'fn': 'freezeNormal',
                                         'q': 'query',
                                         'rel': 'relative',
                                         'ufn': 'unFreezeNormal',
                                         'x': 'normalX',
                                         'xyz': 'normalXYZ',
                                         'y': 'normalY',
                                         'z': 'normalZ'},
                          'type': 'modeling'},
  'polyNormalizeUV': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                            'longname': 'caching',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cch'},
                                'centerOnTile': {'args': '<type __builtin__.bool>',
                                                 'longname': 'centerOnTile',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cot'},
                                'constructionHistory': {'args': '<type __builtin__.bool>',
                                                        'longname': 'constructionHistory',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'ch'},
                                'createNewMap': {'args': '<type __builtin__.bool>',
                                                 'longname': 'createNewMap',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'cm'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'frozen': {'args': '<type __builtin__.bool>',
                                           'longname': 'frozen',
                                           'numArgs': 1,
                                           'shortname': 'fzn'},
                                'insertBeforeDeformers': {'args': '<type __builtin__.bool>',
                                                          'longname': 'insertBeforeDeformers',
                                                          'modes': ['create'],
                                                          'numArgs': 1,
                                                          'shortname': 'ibd'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'nodeState': {'args': '<type __builtin__.int>',
                                              'longname': 'nodeState',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nds'},
                                'normalizeDirection': {'args': '<type __builtin__.int>',
                                                       'longname': 'normalizeDirection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'nd'},
                                'normalizeType': {'args': '<type __builtin__.int>',
                                                  'longname': 'normalizeType',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'nt'},
                                'preserveAspectRatio': {'args': '<type __builtin__.bool>',
                                                        'longname': 'preserveAspectRatio',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'pa'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'uvSetName': {'args': '<type __builtin__.str>',
                                              'longname': 'uvSetName',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'uvs'},
                                'worldSpace': {'args': '<type __builtin__.bool>',
                                               'longname': 'worldSpace',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ws'}},
                      'shortFlags': {'cch': 'caching',
                                     'ch': 'constructionHistory',
                                     'cm': 'createNewMap',
                                     'cot': 'centerOnTile',
                                     'e': 'edit',
                                     'fzn': 'frozen',
                                     'ibd': 'insertBeforeDeformers',
                                     'n': 'name',
                                     'nd': 'normalizeDirection',
                                     'nds': 'nodeState',
                                     'nt': 'normalizeType',
                                     'pa': 'preserveAspectRatio',
                                     'q': 'query',
                                     'uvs': 'uvSetName',
                                     'ws': 'worldSpace'},
                      'type': 'modeling'},
  'polyOptUvs': {'flags': {'applyToShell': {'args': '<type __builtin__.bool>',
                                            'longname': 'applyToShell',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'applyToShell'},
                           'areaWeight': {'args': '<type __builtin__.float>',
                                          'longname': 'areaWeight',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'aw'},
                           'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'globalBlend': {'args': '<type __builtin__.float>',
                                           'longname': 'globalBlend',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'gb'},
                           'globalMethodBlend': {'args': '<type __builtin__.float>',
                                                 'longname': 'globalMethodBlend',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'gmb'},
                           'iterations': {'args': '<type __builtin__.int>',
                                          'longname': 'iterations',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'i'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'optimizeAxis': {'args': '<type __builtin__.int>',
                                            'longname': 'optimizeAxis',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'oa'},
                           'pinSelected': {'args': '<type __builtin__.bool>',
                                           'longname': 'pinSelected',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ps'},
                           'pinUvBorder': {'args': '<type __builtin__.bool>',
                                           'longname': 'pinUvBorder',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'pub'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'scale': {'args': '<type __builtin__.float>',
                                     'longname': 'scale',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 's'},
                           'stoppingThreshold': {'args': '<type __builtin__.float>',
                                                 'longname': 'stoppingThreshold',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ss'},
                           'useScale': {'args': '<type __builtin__.bool>',
                                        'longname': 'useScale',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'us'},
                           'uvSetName': {'args': '<type __builtin__.str>',
                                         'longname': 'uvSetName',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'uvs'},
                           'worldSpace': {'args': '<type __builtin__.bool>',
                                          'longname': 'worldSpace',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ws'}},
                 'removedFlags': {'as': 'applyToShell'},
                 'shortFlags': {'applyToShell': 'applyToShell',
                                'aw': 'areaWeight',
                                'cch': 'caching',
                                'ch': 'constructionHistory',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'gb': 'globalBlend',
                                'gmb': 'globalMethodBlend',
                                'i': 'iterations',
                                'n': 'name',
                                'nds': 'nodeState',
                                'oa': 'optimizeAxis',
                                'ps': 'pinSelected',
                                'pub': 'pinUvBorder',
                                'q': 'query',
                                's': 'scale',
                                'ss': 'stoppingThreshold',
                                'us': 'useScale',
                                'uvs': 'uvSetName',
                                'ws': 'worldSpace'},
                 'type': 'modeling'},
  'polyOptions': {'flags': {'activeObjects': {'args': '<type __builtin__.bool>',
                                              'longname': 'activeObjects',
                                              'modes': ['create', 'query'],
                                              'numArgs': 0,
                                              'shortname': 'ao'},
                            'allEdges': {'args': '<type __builtin__.bool>',
                                         'longname': 'allEdges',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'ae'},
                            'backCullVertex': {'args': '<type __builtin__.bool>',
                                               'longname': 'backCullVertex',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'bcv'},
                            'backCulling': {'args': '<type __builtin__.bool>',
                                            'longname': 'backCulling',
                                            'modes': ['create', 'query'],
                                            'numArgs': 0,
                                            'shortname': 'bc'},
                            'colorMaterialChannel': {'args': '<type __builtin__.str>',
                                                     'longname': 'colorMaterialChannel',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'cm'},
                            'colorShadedDisplay': {'args': '<type __builtin__.bool>',
                                                   'longname': 'colorShadedDisplay',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'cs'},
                            'displayAlphaAsGreyScale': {'args': '<type __builtin__.bool>',
                                                        'longname': 'displayAlphaAsGreyScale',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'dal'},
                            'displayBorder': {'args': '<type __builtin__.bool>',
                                              'longname': 'displayBorder',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'db'},
                            'displayCenter': {'args': '<type __builtin__.bool>',
                                              'longname': 'displayCenter',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'dc'},
                            'displayCreaseEdge': {'args': '<type __builtin__.bool>',
                                                  'longname': 'displayCreaseEdge',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'dce'},
                            'displayCreaseVertex': {'args': '<type __builtin__.bool>',
                                                    'longname': 'displayCreaseVertex',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'dcv'},
                            'displayGeometry': {'args': '<type __builtin__.bool>',
                                                'longname': 'displayGeometry',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'dg'},
                            'displayInvisibleFaces': {'args': '<type __builtin__.bool>',
                                                      'longname': 'displayInvisibleFaces',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'dif'},
                            'displayItemNumbers': {'args': ['<type __builtin__.bool>',
                                                            '<type __builtin__.bool>',
                                                            '<type __builtin__.bool>',
                                                            '<type __builtin__.bool>'],
                                                   'longname': 'displayItemNumbers',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 4,
                                                   'shortname': 'din'},
                            'displayMapBorder': {'args': '<type __builtin__.bool>',
                                                 'longname': 'displayMapBorder',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'dmb'},
                            'displayMetadata': {'args': ['<type __builtin__.bool>',
                                                         '<type __builtin__.bool>',
                                                         '<type __builtin__.bool>'],
                                                'longname': 'displayMetadata',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 3,
                                                'shortname': 'dmt'},
                            'displayNormal': {'args': '<type __builtin__.bool>',
                                              'longname': 'displayNormal',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'dn'},
                            'displayShellBorder': {'args': '<type __builtin__.bool>',
                                                   'longname': 'displayShellBorder',
                                                   'numArgs': 1,
                                                   'shortname': 'dsb'},
                            'displaySubdComps': {'args': '<type __builtin__.bool>',
                                                 'longname': 'displaySubdComps',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'dsc'},
                            'displayTangent': {'args': '<type __builtin__.bool>',
                                               'longname': 'displayTangent',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'dtn'},
                            'displayTriangle': {'args': '<type __builtin__.bool>',
                                                'longname': 'displayTriangle',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                            'displayUVTopology': {'args': '<type __builtin__.bool>',
                                                  'longname': 'displayUVTopology',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'uvt'},
                            'displayUVs': {'args': '<type __builtin__.bool>',
                                           'longname': 'displayUVs',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'duv'},
                            'displayVertex': {'args': '<type __builtin__.bool>',
                                              'longname': 'displayVertex',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'dv'},
                            'displayWarp': {'args': '<type __builtin__.bool>',
                                            'longname': 'displayWarp',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'dw'},
                            'facet': {'args': '<type __builtin__.bool>',
                                      'longname': 'facet',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'f'},
                            'fullBack': {'args': '<type __builtin__.bool>',
                                         'longname': 'fullBack',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'fb'},
                            'gl': {'args': '<type __builtin__.bool>',
                                   'longname': 'gl',
                                   'modes': ['create', 'query'],
                                   'numArgs': 0,
                                   'shortname': 'gl'},
                            'hardBack': {'args': '<type __builtin__.bool>',
                                         'longname': 'hardBack',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'hb'},
                            'hardEdge': {'args': '<type __builtin__.bool>',
                                         'longname': 'hardEdge',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'he'},
                            'hardEdgeColor': {'args': '<type __builtin__.bool>',
                                              'longname': 'hardEdgeColor',
                                              'modes': ['create', 'query'],
                                              'numArgs': 0,
                                              'shortname': 'hec'},
                            'materialBlend': {'args': '<type __builtin__.str>',
                                              'longname': 'materialBlend',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'mb'},
                            'newPolymesh': {'args': '<type __builtin__.bool>',
                                            'longname': 'newPolymesh',
                                            'modes': ['create', 'query'],
                                            'numArgs': 0,
                                            'shortname': 'np'},
                            'point': {'args': '<type __builtin__.bool>',
                                      'longname': 'point',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'pt'},
                            'pointFacet': {'args': '<type __builtin__.bool>',
                                           'longname': 'pointFacet',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'pf'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'relative': {'args': '<type __builtin__.bool>',
                                         'longname': 'relative',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'r'},
                            'reuseTriangles': {'args': '<type __builtin__.bool>',
                                               'longname': 'reuseTriangles',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'rt'},
                            'sizeBorder': {'args': '<type __builtin__.float>',
                                           'longname': 'sizeBorder',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'sb'},
                            'sizeNormal': {'args': '<type __builtin__.float>',
                                           'longname': 'sizeNormal',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'sn'},
                            'sizeUV': {'args': '<type __builtin__.float>',
                                       'longname': 'sizeUV',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'suv'},
                            'sizeVertex': {'args': '<type __builtin__.float>',
                                           'longname': 'sizeVertex',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'sv'},
                            'smoothDrawType': {'args': '<type __builtin__.int>',
                                               'longname': 'smoothDrawType',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'sdt'},
                            'softEdge': {'args': '<type __builtin__.bool>',
                                         'longname': 'softEdge',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'se'},
                            'vertexNormalMethod': {'args': '<type __builtin__.int>',
                                                   'longname': 'vertexNormalMethod',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'vnm'},
                            'wireBackCulling': {'args': '<type __builtin__.bool>',
                                                'longname': 'wireBackCulling',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 0,
                                                'shortname': 'wbc'}},
                  'removedFlags': {'global': 'gl'},
                  'shortFlags': {'ae': 'allEdges',
                                 'ao': 'activeObjects',
                                 'bc': 'backCulling',
                                 'bcv': 'backCullVertex',
                                 'cm': 'colorMaterialChannel',
                                 'cs': 'colorShadedDisplay',
                                 'dal': 'displayAlphaAsGreyScale',
                                 'db': 'displayBorder',
                                 'dc': 'displayCenter',
                                 'dce': 'displayCreaseEdge',
                                 'dcv': 'displayCreaseVertex',
                                 'dg': 'displayGeometry',
                                 'dif': 'displayInvisibleFaces',
                                 'din': 'displayItemNumbers',
                                 'dmb': 'displayMapBorder',
                                 'dmt': 'displayMetadata',
                                 'dn': 'displayNormal',
                                 'dsb': 'displayShellBorder',
                                 'dsc': 'displaySubdComps',
                                 'dt': 'displayTriangle',
                                 'dtn': 'displayTangent',
                                 'duv': 'displayUVs',
                                 'dv': 'displayVertex',
                                 'dw': 'displayWarp',
                                 'f': 'facet',
                                 'fb': 'fullBack',
                                 'gl': 'gl',
                                 'hb': 'hardBack',
                                 'he': 'hardEdge',
                                 'hec': 'hardEdgeColor',
                                 'mb': 'materialBlend',
                                 'np': 'newPolymesh',
                                 'pf': 'pointFacet',
                                 'pt': 'point',
                                 'q': 'query',
                                 'r': 'relative',
                                 'rt': 'reuseTriangles',
                                 'sb': 'sizeBorder',
                                 'sdt': 'smoothDrawType',
                                 'se': 'softEdge',
                                 'sn': 'sizeNormal',
                                 'suv': 'sizeUV',
                                 'sv': 'sizeVertex',
                                 'uvt': 'displayUVTopology',
                                 'vnm': 'vertexNormalMethod',
                                 'wbc': 'wireBackCulling'},
                  'type': 'modeling'},
  'polyOutput': {'flags': {'allValues': {'args': '<type __builtin__.bool>',
                                         'longname': 'allValues',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'a'},
                           'color': {'args': '<type __builtin__.bool>',
                                     'longname': 'color',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'c'},
                           'colorDesc': {'args': '<type __builtin__.bool>',
                                         'longname': 'colorDesc',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'cd'},
                           'edge': {'args': '<type __builtin__.bool>',
                                    'longname': 'edge',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'edgeFace': {'args': '<type __builtin__.bool>',
                                        'longname': 'edgeFace',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ef'},
                           'face': {'args': '<type __builtin__.bool>',
                                    'longname': 'face',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'f'},
                           'faceNorm': {'args': '<type __builtin__.bool>',
                                        'longname': 'faceNorm',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'fn'},
                           'force': {'args': '<type __builtin__.bool>',
                                     'longname': 'force',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'fo'},
                           'group': {'args': '<type __builtin__.bool>',
                                     'longname': 'group',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'g'},
                           'noOutput': {'args': '<type __builtin__.bool>',
                                        'longname': 'noOutput',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'no'},
                           'normDesc': {'args': '<type __builtin__.bool>',
                                        'longname': 'normDesc',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'nd'},
                           'outputFile': {'args': '<type __builtin__.str>',
                                          'longname': 'outputFile',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'of'},
                           'triangle': {'args': '<type __builtin__.bool>',
                                        'longname': 'triangle',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 't'},
                           'uvDesc': {'args': '<type __builtin__.bool>',
                                      'longname': 'uvDesc',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'uvd'},
                           'uvValue': {'args': '<type __builtin__.bool>',
                                       'longname': 'uvValue',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'uv'},
                           'vert': {'args': '<type __builtin__.bool>',
                                    'longname': 'vert',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'v'},
                           'vertEdge': {'args': '<type __builtin__.bool>',
                                        'longname': 'vertEdge',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 've'},
                           'vertNorm': {'args': '<type __builtin__.bool>',
                                        'longname': 'vertNorm',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'vn'}},
                 'shortFlags': {'a': 'allValues',
                                'c': 'color',
                                'cd': 'colorDesc',
                                'e': 'edge',
                                'ef': 'edgeFace',
                                'f': 'face',
                                'fn': 'faceNorm',
                                'fo': 'force',
                                'g': 'group',
                                'nd': 'normDesc',
                                'no': 'noOutput',
                                'of': 'outputFile',
                                't': 'triangle',
                                'uv': 'uvValue',
                                'uvd': 'uvDesc',
                                'v': 'vert',
                                've': 'vertEdge',
                                'vn': 'vertNorm'},
                 'type': 'modeling'},
  'polyPinUV': {'flags': {'createHistory': {'args': '<type __builtin__.bool>',
                                            'longname': 'createHistory',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ch'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'operation': {'args': '<type __builtin__.int>',
                                        'longname': 'operation',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'op'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'unpinned': {'args': '<type __builtin__.bool>',
                                       'longname': 'unpinned',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'unp'},
                          'uvSetName': {'args': '<type __builtin__.str>',
                                        'longname': 'uvSetName',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'uvs'},
                          'value': {'args': '<type __builtin__.float>',
                                    'longname': 'value',
                                    'modes': ['create',
                                              'query',
                                              'edit',
                                              'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'v'}},
                'shortFlags': {'ch': 'createHistory',
                               'e': 'edit',
                               'op': 'operation',
                               'q': 'query',
                               'unp': 'unpinned',
                               'uvs': 'uvSetName',
                               'v': 'value'},
                'type': 'modeling'},
  'polyPipe': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                           '<type __builtin__.float>',
                                           '<type __builtin__.float>'],
                                  'longname': 'axis',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 3,
                                  'shortname': 'ax'},
                         'caching': {'args': '<type __builtin__.bool>',
                                     'longname': 'caching',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsCasting': True,
                                     'shortname': 'cch'},
                         'constructionHistory': {'args': '<type __builtin__.bool>',
                                                 'longname': 'constructionHistory',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                         'createUVs': {'args': '<type __builtin__.bool>',
                                       'longname': 'createUVs',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'cuv'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'frozen': {'args': '<type __builtin__.bool>',
                                    'longname': 'frozen',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'fzn'},
                         'height': {'args': '<type __builtin__.float>',
                                    'longname': 'height',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'h'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'nodeState': {'args': '<type __builtin__.int>',
                                       'longname': 'nodeState',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'nds'},
                         'object': {'args': '<type __builtin__.bool>',
                                    'longname': 'object',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'o'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'radius': {'args': '<type __builtin__.float>',
                                    'longname': 'radius',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'r'},
                         'roundCap': {'args': '<type __builtin__.bool>',
                                      'longname': 'roundCap',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsCasting': True,
                                      'shortname': 'rcp'},
                         'subdivisionsAxis': {'args': '<type __builtin__.int>',
                                              'longname': 'subdivisionsAxis',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'resultNeedsCasting': True,
                                              'shortname': 'sa'},
                         'subdivisionsCaps': {'args': '<type __builtin__.int>',
                                              'longname': 'subdivisionsCaps',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'resultNeedsCasting': True,
                                              'shortname': 'sc'},
                         'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                'longname': 'subdivisionsHeight',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'resultNeedsCasting': True,
                                                'shortname': 'sh'},
                         'texture': {'args': '<type __builtin__.bool>',
                                     'longname': 'texture',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsCasting': True,
                                     'shortname': 'tx'},
                         'thickness': {'args': '<type __builtin__.float>',
                                       'longname': 'thickness',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 't'}},
               'shortFlags': {'ax': 'axis',
                              'cch': 'caching',
                              'ch': 'constructionHistory',
                              'cuv': 'createUVs',
                              'e': 'edit',
                              'fzn': 'frozen',
                              'h': 'height',
                              'n': 'name',
                              'nds': 'nodeState',
                              'o': 'object',
                              'q': 'query',
                              'r': 'radius',
                              'rcp': 'roundCap',
                              'sa': 'subdivisionsAxis',
                              'sc': 'subdivisionsCaps',
                              'sh': 'subdivisionsHeight',
                              't': 'thickness',
                              'tx': 'texture'},
               'type': 'modeling'},
  'polyPlanarProjection': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                                 'longname': 'caching',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cch'},
                                     'constructionHistory': {'args': '<type __builtin__.bool>',
                                                             'longname': 'constructionHistory',
                                                             'modes': ['create',
                                                                       'query'],
                                                             'numArgs': 1,
                                                             'shortname': 'ch'},
                                     'createNewMap': {'args': '<type __builtin__.bool>',
                                                      'longname': 'createNewMap',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'cm'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'frozen': {'args': '<type __builtin__.bool>',
                                                'longname': 'frozen',
                                                'numArgs': 1,
                                                'shortname': 'fzn'},
                                     'imageCenter': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'imageCenter',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'ic'},
                                     'imageCenterX': {'args': '<type __builtin__.float>',
                                                      'longname': 'imageCenterX',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'icx'},
                                     'imageCenterY': {'args': '<type __builtin__.float>',
                                                      'longname': 'imageCenterY',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'icy'},
                                     'imageScale': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'imageScale',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 2,
                                                    'shortname': 'imageScale'},
                                     'imageScaleU': {'args': '<type __builtin__.float>',
                                                     'longname': 'imageScaleU',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'isu'},
                                     'imageScaleV': {'args': '<type __builtin__.float>',
                                                     'longname': 'imageScaleV',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'isv'},
                                     'insertBeforeDeformers': {'args': '<type __builtin__.bool>',
                                                               'longname': 'insertBeforeDeformers',
                                                               'modes': ['create'],
                                                               'numArgs': 1,
                                                               'shortname': 'ibd'},
                                     'keepImageRatio': {'args': '<type __builtin__.bool>',
                                                        'longname': 'keepImageRatio',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'kir'},
                                     'mapDirection': {'args': '<type __builtin__.str>',
                                                      'longname': 'mapDirection',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'md'},
                                     'name': {'args': '<type __builtin__.str>',
                                              'longname': 'name',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'n'},
                                     'nodeState': {'args': '<type __builtin__.int>',
                                                   'longname': 'nodeState',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'nds'},
                                     'perInstance': {'args': '<type __builtin__.bool>',
                                                     'longname': 'perInstance',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'pi'},
                                     'projectionCenter': {'args': ['<type __builtin__.float>',
                                                                   '<type __builtin__.float>',
                                                                   '<type __builtin__.float>'],
                                                          'longname': 'projectionCenter',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 3,
                                                          'shortname': 'pc'},
                                     'projectionCenterX': {'args': '<type __builtin__.float>',
                                                           'longname': 'projectionCenterX',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'pcx'},
                                     'projectionCenterY': {'args': '<type __builtin__.float>',
                                                           'longname': 'projectionCenterY',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'pcy'},
                                     'projectionCenterZ': {'args': '<type __builtin__.float>',
                                                           'longname': 'projectionCenterZ',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'pcz'},
                                     'projectionHeight': {'args': '<type __builtin__.float>',
                                                          'longname': 'projectionHeight',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ph'},
                                     'projectionHorizontalSweep': {'args': '<type __builtin__.float>',
                                                                   'longname': 'projectionHorizontalSweep',
                                                                   'modes': ['create',
                                                                             'query',
                                                                             'edit'],
                                                                   'numArgs': 1,
                                                                   'shortname': 'phs'},
                                     'projectionScale': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'projectionScale',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 2,
                                                         'shortname': 'ps'},
                                     'projectionScaleU': {'args': '<type __builtin__.float>',
                                                          'longname': 'projectionScaleU',
                                                          'numArgs': 1,
                                                          'shortname': 'psu'},
                                     'projectionScaleV': {'args': '<type __builtin__.float>',
                                                          'longname': 'projectionScaleV',
                                                          'numArgs': 1,
                                                          'shortname': 'psv'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'radius': {'args': '<type __builtin__.float>',
                                                'longname': 'radius',
                                                'numArgs': 1,
                                                'shortname': 'r'},
                                     'rotate': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'rotate',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'ro'},
                                     'rotateX': {'args': '<type __builtin__.float>',
                                                 'longname': 'rotateX',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rx'},
                                     'rotateY': {'args': '<type __builtin__.float>',
                                                 'longname': 'rotateY',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ry'},
                                     'rotateZ': {'args': '<type __builtin__.float>',
                                                 'longname': 'rotateZ',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rz'},
                                     'rotationAngle': {'args': '<type __builtin__.float>',
                                                       'longname': 'rotationAngle',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ra'},
                                     'seamCorrect': {'args': '<type __builtin__.bool>',
                                                     'longname': 'seamCorrect',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sc'},
                                     'smartFit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'smartFit',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'sf'},
                                     'worldSpace': {'args': '<type __builtin__.bool>',
                                                    'longname': 'worldSpace',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ws'}},
                           'removedFlags': {'is': 'imageScale'},
                           'shortFlags': {'cch': 'caching',
                                          'ch': 'constructionHistory',
                                          'cm': 'createNewMap',
                                          'e': 'edit',
                                          'fzn': 'frozen',
                                          'ibd': 'insertBeforeDeformers',
                                          'ic': 'imageCenter',
                                          'icx': 'imageCenterX',
                                          'icy': 'imageCenterY',
                                          'imageScale': 'imageScale',
                                          'isu': 'imageScaleU',
                                          'isv': 'imageScaleV',
                                          'kir': 'keepImageRatio',
                                          'md': 'mapDirection',
                                          'n': 'name',
                                          'nds': 'nodeState',
                                          'pc': 'projectionCenter',
                                          'pcx': 'projectionCenterX',
                                          'pcy': 'projectionCenterY',
                                          'pcz': 'projectionCenterZ',
                                          'ph': 'projectionHeight',
                                          'phs': 'projectionHorizontalSweep',
                                          'pi': 'perInstance',
                                          'ps': 'projectionScale',
                                          'psu': 'projectionScaleU',
                                          'psv': 'projectionScaleV',
                                          'q': 'query',
                                          'r': 'radius',
                                          'ra': 'rotationAngle',
                                          'ro': 'rotate',
                                          'rx': 'rotateX',
                                          'ry': 'rotateY',
                                          'rz': 'rotateZ',
                                          'sc': 'seamCorrect',
                                          'sf': 'smartFit',
                                          'ws': 'worldSpace'},
                           'type': 'modeling'},
  'polyPlane': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'axis',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 3,
                                   'shortname': 'ax'},
                          'caching': {'args': '<type __builtin__.bool>',
                                      'longname': 'caching',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsCasting': True,
                                      'shortname': 'cch'},
                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'constructionHistory',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                          'createUVs': {'args': '<type __builtin__.int>',
                                        'longname': 'createUVs',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'cuv'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'frozen': {'args': '<type __builtin__.bool>',
                                     'longname': 'frozen',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'fzn'},
                          'height': {'args': '<type __builtin__.float>',
                                     'longname': 'height',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'h'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nodeState': {'args': '<type __builtin__.int>',
                                        'longname': 'nodeState',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'nds'},
                          'object': {'args': '<type __builtin__.bool>',
                                     'longname': 'object',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'o'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                 'longname': 'subdivisionsHeight',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'resultNeedsCasting': True,
                                                 'shortname': 'sh'},
                          'subdivisionsWidth': {'args': '<type __builtin__.int>',
                                                'longname': 'subdivisionsWidth',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'resultNeedsCasting': True,
                                                'shortname': 'sw'},
                          'subdivisionsX': {'args': '<type __builtin__.int>',
                                            'longname': 'subdivisionsX',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'resultNeedsCasting': True,
                                            'shortname': 'sx'},
                          'subdivisionsY': {'args': '<type __builtin__.int>',
                                            'longname': 'subdivisionsY',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'resultNeedsCasting': True,
                                            'shortname': 'sy'},
                          'texture': {'args': '<type __builtin__.int>',
                                      'longname': 'texture',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsCasting': True,
                                      'shortname': 'tx'},
                          'width': {'args': '<type __builtin__.float>',
                                    'longname': 'width',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'w'}},
                'shortFlags': {'ax': 'axis',
                               'cch': 'caching',
                               'ch': 'constructionHistory',
                               'cuv': 'createUVs',
                               'e': 'edit',
                               'fzn': 'frozen',
                               'h': 'height',
                               'n': 'name',
                               'nds': 'nodeState',
                               'o': 'object',
                               'q': 'query',
                               'sh': 'subdivisionsHeight',
                               'sw': 'subdivisionsWidth',
                               'sx': 'subdivisionsX',
                               'sy': 'subdivisionsY',
                               'tx': 'texture',
                               'w': 'width'},
                'type': 'modeling'},
  'polyPlatonicSolid': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'axis',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'ax'},
                                  'caching': {'args': '<type __builtin__.bool>',
                                              'longname': 'caching',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'resultNeedsCasting': True,
                                              'shortname': 'cch'},
                                  'constructionHistory': {'args': '<type __builtin__.bool>',
                                                          'longname': 'constructionHistory',
                                                          'modes': ['create',
                                                                    'query'],
                                                          'numArgs': 1,
                                                          'shortname': 'ch'},
                                  'createUVs': {'args': '<type __builtin__.int>',
                                                'longname': 'createUVs',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'cuv'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'frozen': {'args': '<type __builtin__.bool>',
                                             'longname': 'frozen',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'fzn'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'nodeState': {'args': '<type __builtin__.int>',
                                                'longname': 'nodeState',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'resultNeedsCasting': True,
                                                'shortname': 'nds'},
                                  'object': {'args': '<type __builtin__.bool>',
                                             'longname': 'object',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'o'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'radius': {'args': '<type __builtin__.float>',
                                             'longname': 'radius',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'r'},
                                  'sideLength': {'args': '<type __builtin__.float>',
                                                 'longname': 'sideLength',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'l'},
                                  'solidType': {'args': '<type __builtin__.int>',
                                                'longname': 'solidType',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'st'},
                                  'texture': {'args': '<type __builtin__.int>',
                                              'longname': 'texture',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'tx'}},
                        'shortFlags': {'ax': 'axis',
                                       'cch': 'caching',
                                       'ch': 'constructionHistory',
                                       'cuv': 'createUVs',
                                       'e': 'edit',
                                       'fzn': 'frozen',
                                       'l': 'sideLength',
                                       'n': 'name',
                                       'nds': 'nodeState',
                                       'o': 'object',
                                       'q': 'query',
                                       'r': 'radius',
                                       'st': 'solidType',
                                       'tx': 'texture'},
                        'type': 'modeling'},
  'polyPoke': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                     'longname': 'caching',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'cch'},
                         'constructionHistory': {'args': '<type __builtin__.bool>',
                                                 'longname': 'constructionHistory',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'frozen': {'args': '<type __builtin__.bool>',
                                    'longname': 'frozen',
                                    'numArgs': 1,
                                    'shortname': 'fzn'},
                         'localTranslate': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'localTranslate',
                                            'modes': ['create'],
                                            'numArgs': 3,
                                            'shortname': 'lt'},
                         'localTranslateX': {'args': '<type __builtin__.float>',
                                             'longname': 'localTranslateX',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ltx'},
                         'localTranslateY': {'args': '<type __builtin__.float>',
                                             'longname': 'localTranslateY',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'lty'},
                         'localTranslateZ': {'args': '<type __builtin__.float>',
                                             'longname': 'localTranslateZ',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ltz'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'nodeState': {'args': '<type __builtin__.int>',
                                       'longname': 'nodeState',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'nds'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'translate': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'translate',
                                       'modes': ['create'],
                                       'numArgs': 3,
                                       'shortname': 't'},
                         'translateX': {'args': '<type __builtin__.float>',
                                        'longname': 'translateX',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'tx'},
                         'translateY': {'args': '<type __builtin__.float>',
                                        'longname': 'translateY',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ty'},
                         'translateZ': {'args': '<type __builtin__.float>',
                                        'longname': 'translateZ',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'tz'},
                         'worldSpace': {'args': '<type __builtin__.bool>',
                                        'longname': 'worldSpace',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ws'}},
               'shortFlags': {'cch': 'caching',
                              'ch': 'constructionHistory',
                              'e': 'edit',
                              'fzn': 'frozen',
                              'lt': 'localTranslate',
                              'ltx': 'localTranslateX',
                              'lty': 'localTranslateY',
                              'ltz': 'localTranslateZ',
                              'n': 'name',
                              'nds': 'nodeState',
                              'q': 'query',
                              't': 'translate',
                              'tx': 'translateX',
                              'ty': 'translateY',
                              'tz': 'translateZ',
                              'ws': 'worldSpace'},
               'type': 'modeling'},
  'polyPrimitive': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'axis',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 3,
                                       'shortname': 'ax'},
                              'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'resultNeedsCasting': True,
                                          'shortname': 'cch'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'createUVs': {'args': '<type __builtin__.int>',
                                            'longname': 'createUVs',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'resultNeedsCasting': True,
                                            'shortname': 'cuv'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'resultNeedsCasting': True,
                                            'shortname': 'nds'},
                              'object': {'args': '<type __builtin__.bool>',
                                         'longname': 'object',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'o'},
                              'polyType': {'args': '<type __builtin__.int>',
                                           'longname': 'polyType',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'pt'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'radius': {'args': '<type __builtin__.float>',
                                         'longname': 'radius',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'r'},
                              'sideLength': {'args': '<type __builtin__.float>',
                                             'longname': 'sideLength',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'l'},
                              'texture': {'args': '<type __builtin__.int>',
                                          'longname': 'texture',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'resultNeedsCasting': True,
                                          'shortname': 'tx'}},
                    'shortFlags': {'ax': 'axis',
                                   'cch': 'caching',
                                   'ch': 'constructionHistory',
                                   'cuv': 'createUVs',
                                   'e': 'edit',
                                   'fzn': 'frozen',
                                   'l': 'sideLength',
                                   'n': 'name',
                                   'nds': 'nodeState',
                                   'o': 'object',
                                   'pt': 'polyType',
                                   'q': 'query',
                                   'r': 'radius',
                                   'tx': 'texture'},
                    'type': 'modeling'},
  'polyPrimitiveMisc': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'polyPrism': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'axis',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 3,
                                   'shortname': 'ax'},
                          'caching': {'args': '<type __builtin__.bool>',
                                      'longname': 'caching',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsCasting': True,
                                      'shortname': 'cch'},
                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'constructionHistory',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                          'createUVs': {'args': '<type __builtin__.int>',
                                        'longname': 'createUVs',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'cuv'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'frozen': {'args': '<type __builtin__.bool>',
                                     'longname': 'frozen',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'fzn'},
                          'length': {'args': '<type __builtin__.float>',
                                     'longname': 'length',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'l'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nodeState': {'args': '<type __builtin__.int>',
                                        'longname': 'nodeState',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'nds'},
                          'numberOfSides': {'args': '<type __builtin__.int>',
                                            'longname': 'numberOfSides',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'resultNeedsCasting': True,
                                            'shortname': 'ns'},
                          'numderOfSides': {'args': '<type __builtin__.int>',
                                            'longname': 'numderOfSides',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'resultNeedsCasting': True,
                                            'shortname': 'nsi'},
                          'object': {'args': '<type __builtin__.bool>',
                                     'longname': 'object',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'o'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'sideLength': {'args': '<type __builtin__.float>',
                                         'longname': 'sideLength',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'w'},
                          'subdivisionsCaps': {'args': '<type __builtin__.int>',
                                               'longname': 'subdivisionsCaps',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'resultNeedsCasting': True,
                                               'shortname': 'sc'},
                          'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                 'longname': 'subdivisionsHeight',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'resultNeedsCasting': True,
                                                 'shortname': 'sh'},
                          'texture': {'args': '<type __builtin__.int>',
                                      'longname': 'texture',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'tx'}},
                'shortFlags': {'ax': 'axis',
                               'cch': 'caching',
                               'ch': 'constructionHistory',
                               'cuv': 'createUVs',
                               'e': 'edit',
                               'fzn': 'frozen',
                               'l': 'length',
                               'n': 'name',
                               'nds': 'nodeState',
                               'ns': 'numberOfSides',
                               'nsi': 'numderOfSides',
                               'o': 'object',
                               'q': 'query',
                               'sc': 'subdivisionsCaps',
                               'sh': 'subdivisionsHeight',
                               'tx': 'texture',
                               'w': 'sideLength'},
                'type': 'modeling'},
  'polyProjectCurve': {'flags': {'addUnderTransform': {'args': '<type __builtin__.bool>',
                                                       'longname': 'addUnderTransform',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'aut'},
                                 'automatic': {'args': '<type __builtin__.bool>',
                                               'longname': 'automatic',
                                               'numArgs': 1,
                                               'shortname': 'automatic'},
                                 'baryCoord': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'baryCoord',
                                               'numArgs': 3,
                                               'shortname': 'bc'},
                                 'baryCoord1': {'args': '<type __builtin__.float>',
                                                'longname': 'baryCoord1',
                                                'numArgs': 1,
                                                'shortname': 'bc1'},
                                 'baryCoord2': {'args': '<type __builtin__.float>',
                                                'longname': 'baryCoord2',
                                                'numArgs': 1,
                                                'shortname': 'bc2'},
                                 'baryCoord3': {'args': '<type __builtin__.float>',
                                                'longname': 'baryCoord3',
                                                'numArgs': 1,
                                                'shortname': 'bc3'},
                                 'caching': {'args': '<type __builtin__.bool>',
                                             'longname': 'caching',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cch'},
                                 'constructionHistory': {'args': '<type __builtin__.bool>',
                                                         'longname': 'constructionHistory',
                                                         'numArgs': 1,
                                                         'shortname': 'ch'},
                                 'curveSamples': {'args': '<type __builtin__.int>',
                                                  'longname': 'curveSamples',
                                                  'numArgs': 1,
                                                  'shortname': 'cs'},
                                 'direction': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'direction',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'd'},
                                 'directionX': {'args': '<type __builtin__.float>',
                                                'longname': 'directionX',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dx'},
                                 'directionY': {'args': '<type __builtin__.float>',
                                                'longname': 'directionY',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dy'},
                                 'directionZ': {'args': '<type __builtin__.float>',
                                                'longname': 'directionZ',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dz'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'face': {'args': '<type __builtin__.int>',
                                          'longname': 'face',
                                          'numArgs': 1,
                                          'shortname': 'f'},
                                 'frozen': {'args': '<type __builtin__.bool>',
                                            'longname': 'frozen',
                                            'numArgs': 1,
                                            'shortname': 'fzn'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'nodeState': {'args': '<type __builtin__.int>',
                                               'longname': 'nodeState',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nds'},
                                 'object': {'args': '<type __builtin__.bool>',
                                            'longname': 'object',
                                            'numArgs': 1,
                                            'shortname': 'o'},
                                 'pointsOnEdges': {'args': '<type __builtin__.bool>',
                                                   'longname': 'pointsOnEdges',
                                                   'numArgs': 1,
                                                   'shortname': 'poe'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'tolerance': {'args': '<type __builtin__.float>',
                                               'longname': 'tolerance',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tol'},
                                 'triangle': {'args': '<type __builtin__.int>',
                                              'longname': 'triangle',
                                              'numArgs': 1,
                                              'shortname': 't'}},
                       'removedFlags': {'as': 'automatic'},
                       'shortFlags': {'aut': 'addUnderTransform',
                                      'automatic': 'automatic',
                                      'bc': 'baryCoord',
                                      'bc1': 'baryCoord1',
                                      'bc2': 'baryCoord2',
                                      'bc3': 'baryCoord3',
                                      'cch': 'caching',
                                      'ch': 'constructionHistory',
                                      'cs': 'curveSamples',
                                      'd': 'direction',
                                      'dx': 'directionX',
                                      'dy': 'directionY',
                                      'dz': 'directionZ',
                                      'e': 'edit',
                                      'f': 'face',
                                      'fzn': 'frozen',
                                      'n': 'name',
                                      'nds': 'nodeState',
                                      'o': 'object',
                                      'poe': 'pointsOnEdges',
                                      'q': 'query',
                                      't': 'triangle',
                                      'tol': 'tolerance'},
                       'type': 'modeling'},
  'polyProjection': {'flags': {'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'createNewMap': {'args': '<type __builtin__.bool>',
                                                'longname': 'createNewMap',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'cm'},
                               'imageCenterX': {'args': '<type __builtin__.float>',
                                                'longname': 'imageCenterX',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'icx'},
                               'imageCenterY': {'args': '<type __builtin__.float>',
                                                'longname': 'imageCenterY',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'icy'},
                               'imageScaleU': {'args': '<type __builtin__.float>',
                                               'longname': 'imageScaleU',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'isu'},
                               'imageScaleV': {'args': '<type __builtin__.float>',
                                               'longname': 'imageScaleV',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'isv'},
                               'insertBeforeDeformers': {'args': '<type __builtin__.bool>',
                                                         'longname': 'insertBeforeDeformers',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'ibd'},
                               'keepImageRatio': {'args': '<type __builtin__.bool>',
                                                  'longname': 'keepImageRatio',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'kir'},
                               'mapDirection': {'args': '<type __builtin__.str>',
                                                'longname': 'mapDirection',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'md'},
                               'projectionCenterX': {'args': '<type __builtin__.float>',
                                                     'longname': 'projectionCenterX',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'pcx'},
                               'projectionCenterY': {'args': '<type __builtin__.float>',
                                                     'longname': 'projectionCenterY',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'pcy'},
                               'projectionCenterZ': {'args': '<type __builtin__.float>',
                                                     'longname': 'projectionCenterZ',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'pcz'},
                               'projectionScaleU': {'args': '<type __builtin__.float>',
                                                    'longname': 'projectionScaleU',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'psu'},
                               'projectionScaleV': {'args': '<type __builtin__.float>',
                                                    'longname': 'projectionScaleV',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'psv'},
                               'rotateX': {'args': '<type __builtin__.float>',
                                           'longname': 'rotateX',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'rx'},
                               'rotateY': {'args': '<type __builtin__.float>',
                                           'longname': 'rotateY',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ry'},
                               'rotateZ': {'args': '<type __builtin__.float>',
                                           'longname': 'rotateZ',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'rz'},
                               'rotationAngle': {'args': '<type __builtin__.float>',
                                                 'longname': 'rotationAngle',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ra'},
                               'seamCorrect': {'args': '<type __builtin__.bool>',
                                               'longname': 'seamCorrect',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'sc'},
                               'smartFit': {'args': '<type __builtin__.bool>',
                                            'longname': 'smartFit',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'sf'},
                               'type': {'args': '<type __builtin__.str>',
                                        'longname': 'type',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 't'},
                               'uvSetName': {'args': '<type __builtin__.str>',
                                             'longname': 'uvSetName',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'uvs'}},
                     'shortFlags': {'ch': 'constructionHistory',
                                    'cm': 'createNewMap',
                                    'ibd': 'insertBeforeDeformers',
                                    'icx': 'imageCenterX',
                                    'icy': 'imageCenterY',
                                    'isu': 'imageScaleU',
                                    'isv': 'imageScaleV',
                                    'kir': 'keepImageRatio',
                                    'md': 'mapDirection',
                                    'pcx': 'projectionCenterX',
                                    'pcy': 'projectionCenterY',
                                    'pcz': 'projectionCenterZ',
                                    'psu': 'projectionScaleU',
                                    'psv': 'projectionScaleV',
                                    'ra': 'rotationAngle',
                                    'rx': 'rotateX',
                                    'ry': 'rotateY',
                                    'rz': 'rotateZ',
                                    'sc': 'seamCorrect',
                                    'sf': 'smartFit',
                                    't': 'type',
                                    'uvs': 'uvSetName'},
                     'type': 'modeling'},
  'polyPyramid': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'axis',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 3,
                                     'shortname': 'ax'},
                            'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'cch'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'createUVs': {'args': '<type __builtin__.int>',
                                          'longname': 'createUVs',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'resultNeedsCasting': True,
                                          'shortname': 'cuv'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'resultNeedsCasting': True,
                                          'shortname': 'nds'},
                            'numberOfSides': {'args': '<type __builtin__.int>',
                                              'longname': 'numberOfSides',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'resultNeedsCasting': True,
                                              'shortname': 'ns'},
                            'numderOfSides': {'args': '<type __builtin__.int>',
                                              'longname': 'numderOfSides',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'resultNeedsCasting': True,
                                              'shortname': 'nsi'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'sideLength': {'args': '<type __builtin__.float>',
                                           'longname': 'sideLength',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'w'},
                            'subdivisionsCaps': {'args': '<type __builtin__.int>',
                                                 'longname': 'subdivisionsCaps',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'resultNeedsCasting': True,
                                                 'shortname': 'sc'},
                            'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                   'longname': 'subdivisionsHeight',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'resultNeedsCasting': True,
                                                   'shortname': 'sh'},
                            'texture': {'args': '<type __builtin__.bool>',
                                        'longname': 'texture',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'tx'}},
                  'shortFlags': {'ax': 'axis',
                                 'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'cuv': 'createUVs',
                                 'e': 'edit',
                                 'fzn': 'frozen',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'ns': 'numberOfSides',
                                 'nsi': 'numderOfSides',
                                 'o': 'object',
                                 'q': 'query',
                                 'sc': 'subdivisionsCaps',
                                 'sh': 'subdivisionsHeight',
                                 'tx': 'texture',
                                 'w': 'sideLength'},
                  'type': 'modeling'},
  'polyQuad': {'flags': {'angle': {'args': '<type __builtin__.float>',
                                   'longname': 'angle',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'a'},
                         'caching': {'args': '<type __builtin__.bool>',
                                     'longname': 'caching',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'cch'},
                         'constructionHistory': {'args': '<type __builtin__.bool>',
                                                 'longname': 'constructionHistory',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'frozen': {'args': '<type __builtin__.bool>',
                                    'longname': 'frozen',
                                    'numArgs': 1,
                                    'shortname': 'fzn'},
                         'keepGroupBorder': {'args': '<type __builtin__.bool>',
                                             'longname': 'keepGroupBorder',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'kgb'},
                         'keepHardEdges': {'args': '<type __builtin__.bool>',
                                           'longname': 'keepHardEdges',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'khe'},
                         'keepTextureBorders': {'args': '<type __builtin__.bool>',
                                                'longname': 'keepTextureBorders',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ktb'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'nodeState': {'args': '<type __builtin__.int>',
                                       'longname': 'nodeState',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'nds'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'worldSpace': {'args': '<type __builtin__.bool>',
                                        'longname': 'worldSpace',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ws'}},
               'shortFlags': {'a': 'angle',
                              'cch': 'caching',
                              'ch': 'constructionHistory',
                              'e': 'edit',
                              'fzn': 'frozen',
                              'kgb': 'keepGroupBorder',
                              'khe': 'keepHardEdges',
                              'ktb': 'keepTextureBorders',
                              'n': 'name',
                              'nds': 'nodeState',
                              'q': 'query',
                              'ws': 'worldSpace'},
               'type': 'modeling'},
  'polyQueryBlindData': {'flags': {'associationType': {'args': '<type __builtin__.str>',
                                                       'longname': 'associationType',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'at'},
                                   'binaryData': {'args': '<type __builtin__.str>',
                                                  'longname': 'binaryData',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'bnd'},
                                   'booleanData': {'args': '<type __builtin__.bool>',
                                                   'longname': 'booleanData',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'bd'},
                                   'doubleData': {'args': '<type __builtin__.float>',
                                                  'longname': 'doubleData',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dbd'},
                                   'intData': {'args': '<type __builtin__.int>',
                                               'longname': 'intData',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ind'},
                                   'longDataName': {'args': '<type __builtin__.str>',
                                                    'longname': 'longDataName',
                                                    'modes': ['create',
                                                              'multiuse'],
                                                    'numArgs': 1,
                                                    'shortname': 'ldn'},
                                   'maxValue': {'args': '<type __builtin__.float>',
                                                'longname': 'maxValue',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'max'},
                                   'minValue': {'args': '<type __builtin__.float>',
                                                'longname': 'minValue',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'min'},
                                   'shortDataName': {'args': '<type __builtin__.str>',
                                                     'longname': 'shortDataName',
                                                     'modes': ['create',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'sdn'},
                                   'showComp': {'args': '<type __builtin__.bool>',
                                                'longname': 'showComp',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'sc'},
                                   'stringData': {'args': '<type __builtin__.str>',
                                                  'longname': 'stringData',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'sd'},
                                   'subString': {'args': '<type __builtin__.str>',
                                                 'longname': 'subString',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ss'},
                                   'typeId': {'args': '<type __builtin__.int>',
                                              'longname': 'typeId',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'id'}},
                         'shortFlags': {'at': 'associationType',
                                        'bd': 'booleanData',
                                        'bnd': 'binaryData',
                                        'dbd': 'doubleData',
                                        'id': 'typeId',
                                        'ind': 'intData',
                                        'ldn': 'longDataName',
                                        'max': 'maxValue',
                                        'min': 'minValue',
                                        'sc': 'showComp',
                                        'sd': 'stringData',
                                        'sdn': 'shortDataName',
                                        'ss': 'subString'},
                         'type': 'modeling'},
  'polyReduce': {'flags': {'border': {'args': '<type __builtin__.float>',
                                      'longname': 'border',
                                      'numArgs': 1,
                                      'shortname': 'b'},
                           'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'cachingReduce': {'args': '<type __builtin__.bool>',
                                             'longname': 'cachingReduce',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cr'},
                           'colorWeights': {'args': '<type __builtin__.float>',
                                            'longname': 'colorWeights',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cwt'},
                           'compactness': {'args': '<type __builtin__.float>',
                                           'longname': 'compactness',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'com'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'detail': {'args': '<type __builtin__.float>',
                                      'longname': 'detail',
                                      'numArgs': 1,
                                      'shortname': 'd'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'geomWeights': {'args': '<type __builtin__.float>',
                                           'longname': 'geomWeights',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'gwt'},
                           'invertVertexWeights': {'args': '<type __builtin__.bool>',
                                                   'longname': 'invertVertexWeights',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'iwt'},
                           'keepBorder': {'args': '<type __builtin__.bool>',
                                          'longname': 'keepBorder',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'kb'},
                           'keepBorderWeight': {'args': '<type __builtin__.float>',
                                                'longname': 'keepBorderWeight',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'kbw'},
                           'keepColorBorder': {'args': '<type __builtin__.bool>',
                                               'longname': 'keepColorBorder',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'kcb'},
                           'keepColorBorderWeight': {'args': '<type __builtin__.float>',
                                                     'longname': 'keepColorBorderWeight',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'kcw'},
                           'keepCreaseEdge': {'args': '<type __builtin__.bool>',
                                              'longname': 'keepCreaseEdge',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'kce'},
                           'keepCreaseEdgeWeight': {'args': '<type __builtin__.float>',
                                                    'longname': 'keepCreaseEdgeWeight',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'cew'},
                           'keepFaceGroupBorder': {'args': '<type __builtin__.bool>',
                                                   'longname': 'keepFaceGroupBorder',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'kfb'},
                           'keepFaceGroupBorderWeight': {'args': '<type __builtin__.float>',
                                                         'longname': 'keepFaceGroupBorderWeight',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'kfw'},
                           'keepHardEdge': {'args': '<type __builtin__.bool>',
                                            'longname': 'keepHardEdge',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'khe'},
                           'keepHardEdgeWeight': {'args': '<type __builtin__.float>',
                                                  'longname': 'keepHardEdgeWeight',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'khw'},
                           'keepMapBorder': {'args': '<type __builtin__.bool>',
                                             'longname': 'keepMapBorder',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'kmb'},
                           'keepMapBorderWeight': {'args': '<type __builtin__.float>',
                                                   'longname': 'keepMapBorderWeight',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'kmw'},
                           'keepOriginalVertices': {'args': '<type __builtin__.bool>',
                                                    'longname': 'keepOriginalVertices',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'kev'},
                           'keepQuadsWeight': {'args': '<type __builtin__.float>',
                                               'longname': 'keepQuadsWeight',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'kqw'},
                           'line': {'args': '<type __builtin__.float>',
                                    'longname': 'line',
                                    'numArgs': 1,
                                    'shortname': 'l'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'percentage': {'args': '<type __builtin__.float>',
                                          'longname': 'percentage',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                           'preserveLocation': {'args': '<type __builtin__.bool>',
                                                'longname': 'preserveLocation',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'pl'},
                           'preserveTopology': {'args': '<type __builtin__.bool>',
                                                'longname': 'preserveTopology',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'top'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'replaceOriginal': {'args': '<type __builtin__.bool>',
                                               'longname': 'replaceOriginal',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'rpo'},
                           'sharpness': {'args': '<type __builtin__.float>',
                                         'longname': 'sharpness',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'shp'},
                           'symmetryPlane': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'symmetryPlane',
                                             'numArgs': 4,
                                             'shortname': 'sym'},
                           'symmetryPlaneW': {'args': '<type __builtin__.float>',
                                              'longname': 'symmetryPlaneW',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sw'},
                           'symmetryPlaneX': {'args': '<type __builtin__.float>',
                                              'longname': 'symmetryPlaneX',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sx'},
                           'symmetryPlaneY': {'args': '<type __builtin__.float>',
                                              'longname': 'symmetryPlaneY',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sy'},
                           'symmetryPlaneZ': {'args': '<type __builtin__.float>',
                                              'longname': 'symmetryPlaneZ',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sz'},
                           'symmetryTolerance': {'args': '<type __builtin__.float>',
                                                 'longname': 'symmetryTolerance',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'stl'},
                           'termination': {'args': '<type __builtin__.int>',
                                           'longname': 'termination',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'trm'},
                           'triangleCount': {'args': '<type __builtin__.int>',
                                             'longname': 'triangleCount',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'tct'},
                           'triangulate': {'args': '<type __builtin__.bool>',
                                           'longname': 'triangulate',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 't'},
                           'useVirtualSymmetry': {'args': '<type __builtin__.int>',
                                                  'longname': 'useVirtualSymmetry',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'uvs'},
                           'uvWeights': {'args': '<type __builtin__.float>',
                                         'longname': 'uvWeights',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'uwt'},
                           'version': {'args': '<type __builtin__.int>',
                                       'longname': 'version',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ver'},
                           'vertexCount': {'args': '<type __builtin__.int>',
                                           'longname': 'vertexCount',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vct'},
                           'vertexMapName': {'args': '<type __builtin__.str>',
                                             'longname': 'vertexMapName',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'vmp'},
                           'vertexWeightCoefficient': {'args': '<type __builtin__.float>',
                                                       'longname': 'vertexWeightCoefficient',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'vwc'},
                           'weightCoefficient': {'args': '<type __builtin__.float>',
                                                 'longname': 'weightCoefficient',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'wc'}},
                 'shortFlags': {'b': 'border',
                                'cch': 'caching',
                                'cew': 'keepCreaseEdgeWeight',
                                'ch': 'constructionHistory',
                                'com': 'compactness',
                                'cr': 'cachingReduce',
                                'cwt': 'colorWeights',
                                'd': 'detail',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'gwt': 'geomWeights',
                                'iwt': 'invertVertexWeights',
                                'kb': 'keepBorder',
                                'kbw': 'keepBorderWeight',
                                'kcb': 'keepColorBorder',
                                'kce': 'keepCreaseEdge',
                                'kcw': 'keepColorBorderWeight',
                                'kev': 'keepOriginalVertices',
                                'kfb': 'keepFaceGroupBorder',
                                'kfw': 'keepFaceGroupBorderWeight',
                                'khe': 'keepHardEdge',
                                'khw': 'keepHardEdgeWeight',
                                'kmb': 'keepMapBorder',
                                'kmw': 'keepMapBorderWeight',
                                'kqw': 'keepQuadsWeight',
                                'l': 'line',
                                'n': 'name',
                                'nds': 'nodeState',
                                'p': 'percentage',
                                'pl': 'preserveLocation',
                                'q': 'query',
                                'rpo': 'replaceOriginal',
                                'shp': 'sharpness',
                                'stl': 'symmetryTolerance',
                                'sw': 'symmetryPlaneW',
                                'sx': 'symmetryPlaneX',
                                'sy': 'symmetryPlaneY',
                                'sym': 'symmetryPlane',
                                'sz': 'symmetryPlaneZ',
                                't': 'triangulate',
                                'tct': 'triangleCount',
                                'top': 'preserveTopology',
                                'trm': 'termination',
                                'uvs': 'useVirtualSymmetry',
                                'uwt': 'uvWeights',
                                'vct': 'vertexCount',
                                'ver': 'version',
                                'vmp': 'vertexMapName',
                                'vwc': 'vertexWeightCoefficient',
                                'wc': 'weightCoefficient'},
                 'type': 'modeling'},
  'polyRemesh': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'interpolationType': {'args': '<type __builtin__.int>',
                                                 'longname': 'interpolationType',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ipt'},
                           'maxTriangleCount': {'args': '<type __builtin__.int>',
                                                'longname': 'maxTriangleCount',
                                                'numArgs': 1,
                                                'shortname': 'mtc'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'reduceThreshold': {'args': '<type __builtin__.float>',
                                               'longname': 'reduceThreshold',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rdt'},
                           'refineThreshold': {'args': '<type __builtin__.float>',
                                               'longname': 'refineThreshold',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rft'},
                           'smoothStrength': {'args': '<type __builtin__.float>',
                                              'longname': 'smoothStrength',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'smt'},
                           'tessellateBorders': {'args': '<type __builtin__.bool>',
                                                 'longname': 'tessellateBorders',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'tsb'},
                           'worldSpace': {'args': '<type __builtin__.bool>',
                                          'longname': 'worldSpace',
                                          'numArgs': 1,
                                          'shortname': 'ws'}},
                 'shortFlags': {'cch': 'caching',
                                'ch': 'constructionHistory',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'ipt': 'interpolationType',
                                'mtc': 'maxTriangleCount',
                                'n': 'name',
                                'nds': 'nodeState',
                                'q': 'query',
                                'rdt': 'reduceThreshold',
                                'rft': 'refineThreshold',
                                'smt': 'smoothStrength',
                                'tsb': 'tessellateBorders',
                                'ws': 'worldSpace'},
                 'type': 'modeling'},
  'polyRetopo': {'flags': {'anisotropy': {'args': '<type __builtin__.float>',
                                          'longname': 'anisotropy',
                                          'numArgs': 1,
                                          'shortname': 'a'},
                           'anisotropyFilter': {'args': '<type __builtin__.float>',
                                                'longname': 'anisotropyFilter',
                                                'numArgs': 1,
                                                'shortname': 'af'},
                           'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'curveInfluenceDirection': {'args': '<type __builtin__.float>',
                                                       'longname': 'curveInfluenceDirection',
                                                       'numArgs': 1,
                                                       'shortname': 'cid'},
                           'curveSingularitySeparation': {'args': '<type __builtin__.float>',
                                                          'longname': 'curveSingularitySeparation',
                                                          'numArgs': 1,
                                                          'shortname': 'css'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'geodesicRadius': {'args': '<type __builtin__.float>',
                                              'longname': 'geodesicRadius',
                                              'numArgs': 1,
                                              'shortname': 'gr'},
                           'magnitudeMin': {'args': '<type __builtin__.float>',
                                            'longname': 'magnitudeMin',
                                            'numArgs': 1,
                                            'shortname': 'mm'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'resultRefinementLevel': {'args': '<type __builtin__.int>',
                                                     'longname': 'resultRefinementLevel',
                                                     'numArgs': 1,
                                                     'shortname': 'rrl'},
                           'resultSmoothing': {'args': '<type __builtin__.bool>',
                                               'longname': 'resultSmoothing',
                                               'numArgs': 1,
                                               'shortname': 'rs'},
                           'targetEdgeLengthMax': {'args': '<type __builtin__.float>',
                                                   'longname': 'targetEdgeLengthMax',
                                                   'numArgs': 1,
                                                   'shortname': 'tea'},
                           'targetEdgeLengthMin': {'args': '<type __builtin__.float>',
                                                   'longname': 'targetEdgeLengthMin',
                                                   'numArgs': 1,
                                                   'shortname': 'tem'},
                           'targetFaceCount': {'args': '<type __builtin__.int>',
                                               'longname': 'targetFaceCount',
                                               'numArgs': 1,
                                               'shortname': 'tfc'}},
                 'shortFlags': {'a': 'anisotropy',
                                'af': 'anisotropyFilter',
                                'cch': 'caching',
                                'ch': 'constructionHistory',
                                'cid': 'curveInfluenceDirection',
                                'css': 'curveSingularitySeparation',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'gr': 'geodesicRadius',
                                'mm': 'magnitudeMin',
                                'n': 'name',
                                'nds': 'nodeState',
                                'q': 'query',
                                'rrl': 'resultRefinementLevel',
                                'rs': 'resultSmoothing',
                                'tea': 'targetEdgeLengthMax',
                                'tem': 'targetEdgeLengthMin',
                                'tfc': 'targetFaceCount'},
                 'type': 'modeling'},
  'polyRetopoCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'history': {'args': '<type __builtin__.bool>',
                                          'longname': 'history',
                                          'numArgs': 1,
                                          'shortname': 'ch'},
                              'image1': {'args': '<type __builtin__.str>',
                                         'longname': 'image1',
                                         'numArgs': 1,
                                         'shortname': 'i1'},
                              'image2': {'args': '<type __builtin__.str>',
                                         'longname': 'image2',
                                         'numArgs': 1,
                                         'shortname': 'i2'},
                              'image3': {'args': '<type __builtin__.str>',
                                         'longname': 'image3',
                                         'numArgs': 1,
                                         'shortname': 'i3'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'}},
                    'shortFlags': {'ch': 'history',
                                   'e': 'edit',
                                   'ex': 'exists',
                                   'i1': 'image1',
                                   'i2': 'image2',
                                   'i3': 'image3',
                                   'n': 'name',
                                   'q': 'query'},
                    'type': 'context'},
  'polySelect': {'flags': {'add': {'args': '<type __builtin__.bool>',
                                   'longname': 'add',
                                   'modes': ['create', 'query'],
                                   'numArgs': 0,
                                   'shortname': 'add'},
                           'addFirst': {'args': '<type __builtin__.bool>',
                                        'longname': 'addFirst',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'af'},
                           'asSelectString': {'args': '<type __builtin__.bool>',
                                              'longname': 'asSelectString',
                                              'modes': ['create', 'query'],
                                              'numArgs': 0,
                                              'shortname': 'ass'},
                           'deselect': {'args': '<type __builtin__.bool>',
                                        'longname': 'deselect',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'd'},
                           'edgeBorder': {'args': '<type __builtin__.int>',
                                          'longname': 'edgeBorder',
                                          'modes': ['create',
                                                    'query',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'eb'},
                           'edgeBorderPath': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'edgeBorderPath',
                                              'modes': ['create',
                                                        'query',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'ebp'},
                           'edgeBorderPattern': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'edgeBorderPattern',
                                                 'modes': ['create',
                                                           'query',
                                                           'multiuse'],
                                                 'numArgs': 2,
                                                 'shortname': 'bpt'},
                           'edgeLoop': {'args': '<type __builtin__.int>',
                                        'longname': 'edgeLoop',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'el'},
                           'edgeLoopOrBorder': {'args': '<type __builtin__.int>',
                                                'longname': 'edgeLoopOrBorder',
                                                'modes': ['create',
                                                          'query',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'elb'},
                           'edgeLoopOrBorderPattern': {'args': ['<type __builtin__.int>',
                                                                '<type __builtin__.int>'],
                                                       'longname': 'edgeLoopOrBorderPattern',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'multiuse'],
                                                       'numArgs': 2,
                                                       'shortname': 'lbp'},
                           'edgeLoopPath': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'edgeLoopPath',
                                            'modes': ['create',
                                                      'query',
                                                      'multiuse'],
                                            'numArgs': 2,
                                            'shortname': 'elp'},
                           'edgeLoopPattern': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'edgeLoopPattern',
                                               'modes': ['create',
                                                         'query',
                                                         'multiuse'],
                                               'numArgs': 2,
                                               'shortname': 'lpt'},
                           'edgeRing': {'args': '<type __builtin__.int>',
                                        'longname': 'edgeRing',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'er'},
                           'edgeRingPath': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'edgeRingPath',
                                            'modes': ['create',
                                                      'query',
                                                      'multiuse'],
                                            'numArgs': 2,
                                            'shortname': 'erp'},
                           'edgeRingPattern': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'edgeRingPattern',
                                               'modes': ['create',
                                                         'query',
                                                         'multiuse'],
                                               'numArgs': 2,
                                               'shortname': 'rpt'},
                           'edgeUVLoopOrBorder': {'args': '<type __builtin__.int>',
                                                  'longname': 'edgeUVLoopOrBorder',
                                                  'modes': ['create',
                                                            'query',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'euv'},
                           'everyN': {'args': '<type __builtin__.int>',
                                      'longname': 'everyN',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'en'},
                           'extendToShell': {'args': '<type __builtin__.int>',
                                             'longname': 'extendToShell',
                                             'modes': ['create',
                                                       'query',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'ets'},
                           'noSelection': {'args': '<type __builtin__.bool>',
                                           'longname': 'noSelection',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'ns'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'replace': {'args': '<type __builtin__.bool>',
                                       'longname': 'replace',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'r'},
                           'shortestEdgePath': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'shortestEdgePath',
                                                'modes': ['create',
                                                          'query',
                                                          'multiuse'],
                                                'numArgs': 2,
                                                'shortname': 'sep'},
                           'shortestEdgePathUV': {'args': ['<type __builtin__.int>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'shortestEdgePathUV',
                                                  'modes': ['create',
                                                            'query',
                                                            'multiuse'],
                                                  'numArgs': 2,
                                                  'shortname': 'spu'},
                           'shortestFacePath': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'shortestFacePath',
                                                'modes': ['create',
                                                          'query',
                                                          'multiuse'],
                                                'numArgs': 2,
                                                'shortname': 'sfp'},
                           'toggle': {'args': '<type __builtin__.bool>',
                                      'longname': 'toggle',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'tgl'}},
                 'shortFlags': {'add': 'add',
                                'af': 'addFirst',
                                'ass': 'asSelectString',
                                'bpt': 'edgeBorderPattern',
                                'd': 'deselect',
                                'eb': 'edgeBorder',
                                'ebp': 'edgeBorderPath',
                                'el': 'edgeLoop',
                                'elb': 'edgeLoopOrBorder',
                                'elp': 'edgeLoopPath',
                                'en': 'everyN',
                                'er': 'edgeRing',
                                'erp': 'edgeRingPath',
                                'ets': 'extendToShell',
                                'euv': 'edgeUVLoopOrBorder',
                                'lbp': 'edgeLoopOrBorderPattern',
                                'lpt': 'edgeLoopPattern',
                                'ns': 'noSelection',
                                'q': 'query',
                                'r': 'replace',
                                'rpt': 'edgeRingPattern',
                                'sep': 'shortestEdgePath',
                                'sfp': 'shortestFacePath',
                                'spu': 'shortestEdgePathUV',
                                'tgl': 'toggle'},
                 'type': 'modeling'},
  'polySelectConstraint': {'flags': {'angle': {'args': '<type __builtin__.int>',
                                               'longname': 'angle',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'a'},
                                     'anglePropagation': {'args': '<type __builtin__.bool>',
                                                          'longname': 'anglePropagation',
                                                          'modes': ['create',
                                                                    'query'],
                                                          'numArgs': 1,
                                                          'shortname': 'ap'},
                                     'angleTolerance': {'args': '<type __builtin__.float>',
                                                        'longname': 'angleTolerance',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'at'},
                                     'anglebound': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'anglebound',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 2,
                                                    'shortname': 'ab'},
                                     'border': {'args': '<type __builtin__.bool>',
                                                'longname': 'border',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'bo'},
                                     'borderPropagation': {'args': '<type __builtin__.bool>',
                                                           'longname': 'borderPropagation',
                                                           'modes': ['create',
                                                                     'query'],
                                                           'numArgs': 1,
                                                           'shortname': 'bp'},
                                     'convexity': {'args': '<type __builtin__.int>',
                                                   'longname': 'convexity',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'c'},
                                     'crease': {'args': '<type __builtin__.bool>',
                                                'longname': 'crease',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'cr'},
                                     'disable': {'args': '<type __builtin__.bool>',
                                                 'longname': 'disable',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'dis'},
                                     'dist': {'args': '<type __builtin__.int>',
                                              'longname': 'dist',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'd'},
                                     'distaxis': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'distaxis',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 3,
                                                  'shortname': 'da'},
                                     'distbound': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'distbound',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 2,
                                                   'shortname': 'db'},
                                     'distpoint': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'distpoint',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 3,
                                                   'shortname': 'dp'},
                                     'edgeDistance': {'args': '<type __builtin__.int>',
                                                      'longname': 'edgeDistance',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ed'},
                                     'geometricarea': {'args': '<type __builtin__.int>',
                                                       'longname': 'geometricarea',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'ga'},
                                     'geometricareabound': {'args': ['<type __builtin__.float>',
                                                                     '<type __builtin__.float>'],
                                                            'longname': 'geometricareabound',
                                                            'modes': ['create',
                                                                      'query'],
                                                            'numArgs': 2,
                                                            'shortname': 'gab'},
                                     'holes': {'args': '<type __builtin__.int>',
                                               'longname': 'holes',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'h'},
                                     'length': {'args': '<type __builtin__.int>',
                                                'longname': 'length',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'l'},
                                     'lengthbound': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'lengthbound',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 2,
                                                     'shortname': 'lb'},
                                     'loopPropagation': {'args': '<type __builtin__.bool>',
                                                         'longname': 'loopPropagation',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'lp'},
                                     'max2dAngle': {'args': '<type __builtin__.float>',
                                                    'longname': 'max2dAngle',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'm2a'},
                                     'max3dAngle': {'args': '<type __builtin__.float>',
                                                    'longname': 'max3dAngle',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'm3a'},
                                     'mode': {'args': '<type __builtin__.int>',
                                              'longname': 'mode',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'm'},
                                     'nonmanifold': {'args': '<type __builtin__.int>',
                                                     'longname': 'nonmanifold',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'nm'},
                                     'oppositeEdges': {'args': '<type __builtin__.bool>',
                                                       'longname': 'oppositeEdges',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'oe'},
                                     'order': {'args': '<type __builtin__.int>',
                                               'longname': 'order',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'order'},
                                     'orderbound': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'orderbound',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 2,
                                                    'shortname': 'orb'},
                                     'orient': {'args': '<type __builtin__.int>',
                                                'longname': 'orient',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'o'},
                                     'orientaxis': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'orientaxis',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 3,
                                                    'shortname': 'oa'},
                                     'orientbound': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'orientbound',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 2,
                                                     'shortname': 'ob'},
                                     'planarity': {'args': '<type __builtin__.int>',
                                                   'longname': 'planarity',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'p'},
                                     'propagate': {'args': '<type __builtin__.int>',
                                                   'longname': 'propagate',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'pp'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'random': {'args': '<type __builtin__.int>',
                                                'longname': 'random',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'r'},
                                     'randomratio': {'args': '<type __builtin__.float>',
                                                     'longname': 'randomratio',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'rr'},
                                     'returnSelection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'returnSelection',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'rs'},
                                     'ringPropagation': {'args': '<type __builtin__.bool>',
                                                         'longname': 'ringPropagation',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'rp'},
                                     'shell': {'args': '<type __builtin__.bool>',
                                               'longname': 'shell',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'sh'},
                                     'size': {'args': '<type __builtin__.int>',
                                              'longname': 'size',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'sz'},
                                     'smoothness': {'args': '<type __builtin__.int>',
                                                    'longname': 'smoothness',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'sm'},
                                     'stateString': {'args': '<type __builtin__.bool>',
                                                     'longname': 'stateString',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'sts'},
                                     'textured': {'args': '<type __builtin__.int>',
                                                  'longname': 'textured',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'tx'},
                                     'texturedarea': {'args': '<type __builtin__.int>',
                                                      'longname': 'texturedarea',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'ta'},
                                     'texturedareabound': {'args': ['<type __builtin__.float>',
                                                                    '<type __builtin__.float>'],
                                                           'longname': 'texturedareabound',
                                                           'modes': ['create',
                                                                     'query'],
                                                           'numArgs': 2,
                                                           'shortname': 'tab'},
                                     'textureshared': {'args': '<type __builtin__.int>',
                                                       'longname': 'textureshared',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'ts'},
                                     'topology': {'args': '<type __builtin__.int>',
                                                  'longname': 'topology',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'tp'},
                                     'type': {'args': '<type __builtin__.int>',
                                              'longname': 'type',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 't'},
                                     'uvBorderSelection': {'args': '<type __builtin__.bool>',
                                                           'longname': 'uvBorderSelection',
                                                           'modes': ['create',
                                                                     'query'],
                                                           'numArgs': 1,
                                                           'shortname': 'ubs'},
                                     'uvConstraint': {'args': '<type __builtin__.bool>',
                                                      'longname': 'uvConstraint',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'uvc'},
                                     'uvEdgeLoopPropagation': {'args': '<type __builtin__.bool>',
                                                               'longname': 'uvEdgeLoopPropagation',
                                                               'modes': ['create',
                                                                         'query'],
                                                               'numArgs': 1,
                                                               'shortname': 'ulp'},
                                     'uvEdgeRingPropagation': {'args': '<type __builtin__.bool>',
                                                               'longname': 'uvEdgeRingPropagation',
                                                               'modes': ['create',
                                                                         'query'],
                                                               'numArgs': 1,
                                                               'shortname': 'urp'},
                                     'uvFaceOrientation': {'args': '<type __builtin__.int>',
                                                           'longname': 'uvFaceOrientation',
                                                           'modes': ['create',
                                                                     'query'],
                                                           'numArgs': 1,
                                                           'shortname': 'ufo'},
                                     'uvShell': {'args': '<type __builtin__.bool>',
                                                 'longname': 'uvShell',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'uv'},
                                     'visibility': {'args': '<type __builtin__.int>',
                                                    'longname': 'visibility',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'v'},
                                     'visibilityangle': {'args': '<type __builtin__.float>',
                                                         'longname': 'visibilityangle',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'va'},
                                     'visibilitypoint': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'visibilitypoint',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 3,
                                                         'shortname': 'vp'},
                                     'where': {'args': '<type __builtin__.int>',
                                               'longname': 'where',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'w'},
                                     'wholeSensitive': {'args': '<type __builtin__.bool>',
                                                        'longname': 'wholeSensitive',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'ws'}},
                           'removedFlags': {'or': 'order'},
                           'shortFlags': {'a': 'angle',
                                          'ab': 'anglebound',
                                          'ap': 'anglePropagation',
                                          'at': 'angleTolerance',
                                          'bo': 'border',
                                          'bp': 'borderPropagation',
                                          'c': 'convexity',
                                          'cr': 'crease',
                                          'd': 'dist',
                                          'da': 'distaxis',
                                          'db': 'distbound',
                                          'dis': 'disable',
                                          'dp': 'distpoint',
                                          'ed': 'edgeDistance',
                                          'ga': 'geometricarea',
                                          'gab': 'geometricareabound',
                                          'h': 'holes',
                                          'l': 'length',
                                          'lb': 'lengthbound',
                                          'lp': 'loopPropagation',
                                          'm': 'mode',
                                          'm2a': 'max2dAngle',
                                          'm3a': 'max3dAngle',
                                          'nm': 'nonmanifold',
                                          'o': 'orient',
                                          'oa': 'orientaxis',
                                          'ob': 'orientbound',
                                          'oe': 'oppositeEdges',
                                          'orb': 'orderbound',
                                          'order': 'order',
                                          'p': 'planarity',
                                          'pp': 'propagate',
                                          'q': 'query',
                                          'r': 'random',
                                          'rp': 'ringPropagation',
                                          'rr': 'randomratio',
                                          'rs': 'returnSelection',
                                          'sh': 'shell',
                                          'sm': 'smoothness',
                                          'sts': 'stateString',
                                          'sz': 'size',
                                          't': 'type',
                                          'ta': 'texturedarea',
                                          'tab': 'texturedareabound',
                                          'tp': 'topology',
                                          'ts': 'textureshared',
                                          'tx': 'textured',
                                          'ubs': 'uvBorderSelection',
                                          'ufo': 'uvFaceOrientation',
                                          'ulp': 'uvEdgeLoopPropagation',
                                          'urp': 'uvEdgeRingPropagation',
                                          'uv': 'uvShell',
                                          'uvc': 'uvConstraint',
                                          'v': 'visibility',
                                          'va': 'visibilityangle',
                                          'vp': 'visibilitypoint',
                                          'w': 'where',
                                          'ws': 'wholeSensitive'},
                           'type': 'modeling'},
  'polySelectConstraintMonitor': {'callbackFlags': ['changeCommand', 'cc'],
                                  'flags': {'changeCommand': {'args': ['<type __builtin__.str>',
                                                                       '<type __builtin__.str>'],
                                                              'longname': 'changeCommand',
                                                              'modes': ['create'],
                                                              'numArgs': 2,
                                                              'shortname': 'cc'},
                                            'create': {'args': '<type __builtin__.bool>',
                                                       'longname': 'create',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'c'},
                                            'delete': {'args': '<type __builtin__.bool>',
                                                       'longname': 'delete',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'd'}},
                                  'shortFlags': {'c': 'create',
                                                 'cc': 'changeCommand',
                                                 'd': 'delete'},
                                  'type': 'modeling'},
  'polySelectCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'image1': {'args': '<type __builtin__.str>',
                                         'longname': 'image1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i1'},
                              'image2': {'args': '<type __builtin__.str>',
                                         'longname': 'image2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i2'},
                              'image3': {'args': '<type __builtin__.str>',
                                         'longname': 'image3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i3'},
                              'mode': {'args': '<type __builtin__.int>',
                                       'longname': 'mode',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'}},
                    'shortFlags': {'e': 'edit',
                                   'ex': 'exists',
                                   'i1': 'image1',
                                   'i2': 'image2',
                                   'i3': 'image3',
                                   'm': 'mode',
                                   'q': 'query'},
                    'type': 'context'},
  'polySelectEditCtx': {'flags': {'absoluteOffset': {'args': '<type __builtin__.bool>',
                                                     'longname': 'absoluteOffset',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'abo'},
                                  'adjustEdgeFlow': {'args': '<type __builtin__.float>',
                                                     'longname': 'adjustEdgeFlow',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'aef'},
                                  'autoComplete': {'args': '<type __builtin__.bool>',
                                                   'longname': 'autoComplete',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ac'},
                                  'deleteEdge': {'args': '<type __builtin__.bool>',
                                                 'longname': 'deleteEdge',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'de'},
                                  'divisions': {'args': '<type __builtin__.int>',
                                                'longname': 'divisions',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'div'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'endVertexOffset': {'args': '<type __builtin__.float>',
                                                      'longname': 'endVertexOffset',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'evo'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'fixQuads': {'args': '<type __builtin__.bool>',
                                               'longname': 'fixQuads',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'fq'},
                                  'image1': {'args': '<type __builtin__.str>',
                                             'longname': 'image1',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i1'},
                                  'image2': {'args': '<type __builtin__.str>',
                                             'longname': 'image2',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i2'},
                                  'image3': {'args': '<type __builtin__.str>',
                                             'longname': 'image3',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i3'},
                                  'insertWithEdgeFlow': {'args': '<type __builtin__.bool>',
                                                         'longname': 'insertWithEdgeFlow',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ief'},
                                  'mode': {'args': '<type __builtin__.int>',
                                           'longname': 'mode',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'm'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'smoothingAngle': {'args': '<type __builtin__.float>',
                                                     'longname': 'smoothingAngle',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sma'},
                                  'splitType': {'args': '<type __builtin__.int>',
                                                'longname': 'splitType',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'stp'},
                                  'startVertexOffset': {'args': '<type __builtin__.float>',
                                                        'longname': 'startVertexOffset',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'svo'},
                                  'useEqualMultiplier': {'args': '<type __builtin__.bool>',
                                                         'longname': 'useEqualMultiplier',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'uem'}},
                        'shortFlags': {'abo': 'absoluteOffset',
                                       'ac': 'autoComplete',
                                       'aef': 'adjustEdgeFlow',
                                       'de': 'deleteEdge',
                                       'div': 'divisions',
                                       'e': 'edit',
                                       'evo': 'endVertexOffset',
                                       'ex': 'exists',
                                       'fq': 'fixQuads',
                                       'i1': 'image1',
                                       'i2': 'image2',
                                       'i3': 'image3',
                                       'ief': 'insertWithEdgeFlow',
                                       'm': 'mode',
                                       'q': 'query',
                                       'sma': 'smoothingAngle',
                                       'spt': 'splitType',
                                       'svo': 'startVertexOffset',
                                       'uem': 'useEqualMultiplier'},
                        'type': 'context'},
  'polySelectEditCtxDataCmd': {'flags': {},
                               'shortFlags': {},
                               'type': 'other'},
  'polySelectSp': {'flags': {'loop': {'args': '<type __builtin__.bool>',
                                      'longname': 'loop',
                                      'numArgs': 0,
                                      'shortname': 'l'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'ring': {'args': '<type __builtin__.bool>',
                                      'longname': 'ring',
                                      'numArgs': 0,
                                      'shortname': 'r'}},
                   'shortFlags': {'l': 'loop', 'q': 'query', 'r': 'ring'},
                   'type': 'other'},
  'polySeparate': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'endFace': {'args': '<type __builtin__.int>',
                                         'longname': 'endFace',
                                         'numArgs': 1,
                                         'shortname': 'ef'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'inPlace': {'args': '<type __builtin__.bool>',
                                         'longname': 'inPlace',
                                         'numArgs': 1,
                                         'shortname': 'inp'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'object': {'args': '<type __builtin__.bool>',
                                        'longname': 'object',
                                        'numArgs': 1,
                                        'shortname': 'o'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'removeShells': {'args': '<type __builtin__.bool>',
                                              'longname': 'removeShells',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'rs'},
                             'separateSpecificShell': {'args': '<type __builtin__.int>',
                                                       'longname': 'separateSpecificShell',
                                                       'modes': ['create',
                                                                 'multiuse'],
                                                       'numArgs': 1,
                                                       'shortname': 'sss'},
                             'startFace': {'args': '<type __builtin__.int>',
                                           'longname': 'startFace',
                                           'numArgs': 1,
                                           'shortname': 'sf'},
                             'userSpecifiedShells': {'args': '<type __builtin__.bool>',
                                                     'longname': 'userSpecifiedShells',
                                                     'numArgs': 1,
                                                     'shortname': 'uss'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'e': 'edit',
                                  'ef': 'endFace',
                                  'fzn': 'frozen',
                                  'inp': 'inPlace',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'o': 'object',
                                  'q': 'query',
                                  'rs': 'removeShells',
                                  'sf': 'startFace',
                                  'sss': 'separateSpecificShell',
                                  'uss': 'userSpecifiedShells'},
                   'type': 'modeling'},
  'polySetToFaceNormal': {'flags': {'setUserNormal': {'args': '<type __builtin__.bool>',
                                                      'longname': 'setUserNormal',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'su'}},
                          'shortFlags': {'su': 'setUserNormal'},
                          'type': 'modeling'},
  'polySetVertices': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'polySewEdge': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'texture': {'args': '<type __builtin__.bool>',
                                        'longname': 'texture',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'tx'},
                            'tolerance': {'args': '<type __builtin__.float>',
                                          'longname': 'tolerance',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 't'},
                            'worldSpace': {'args': '<type __builtin__.bool>',
                                           'longname': 'worldSpace',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ws'}},
                  'shortFlags': {'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'e': 'edit',
                                 'fzn': 'frozen',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'q': 'query',
                                 't': 'tolerance',
                                 'tx': 'texture',
                                 'ws': 'worldSpace'},
                  'type': 'modeling'},
  'polyShortestPathCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'}},
                          'shortFlags': {'e': 'edit',
                                         'ex': 'exists',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'q': 'query'},
                          'type': 'context'},
  'polySlideEdge': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                           'longname': 'absolute',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'a'},
                              'direction': {'args': '<type __builtin__.int>',
                                            'longname': 'direction',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'd'},
                              'edgeDirection': {'args': '<type __builtin__.float>',
                                                'longname': 'edgeDirection',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ed'},
                              'symmetry': {'args': '<type __builtin__.bool>',
                                           'longname': 'symmetry',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'sym'}},
                    'shortFlags': {'a': 'absolute',
                                   'd': 'direction',
                                   'ed': 'edgeDirection',
                                   'sym': 'symmetry'},
                    'type': 'modeling'},
  'polySlideEdgeCtx': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                              'longname': 'absolute',
                                              'numArgs': 1,
                                              'shortname': 'a'},
                                 'direction': {'args': '<type __builtin__.int>',
                                               'longname': 'direction',
                                               'numArgs': 1,
                                               'shortname': 'd'},
                                 'edgeDirection': {'args': '<type __builtin__.float>',
                                                   'longname': 'edgeDirection',
                                                   'numArgs': 1,
                                                   'shortname': 'ed'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'history': {'args': '<type __builtin__.bool>',
                                             'longname': 'history',
                                             'numArgs': 1,
                                             'shortname': 'ch'},
                                 'image1': {'args': '<type __builtin__.str>',
                                            'longname': 'image1',
                                            'numArgs': 1,
                                            'shortname': 'i1'},
                                 'image2': {'args': '<type __builtin__.str>',
                                            'longname': 'image2',
                                            'numArgs': 1,
                                            'shortname': 'i2'},
                                 'image3': {'args': '<type __builtin__.str>',
                                            'longname': 'image3',
                                            'numArgs': 1,
                                            'shortname': 'i3'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'snappingPoints': {'args': '<type __builtin__.int>',
                                                    'longname': 'snappingPoints',
                                                    'numArgs': 1,
                                                    'shortname': 'sp'},
                                 'snappingTolerance': {'args': '<type __builtin__.float>',
                                                       'longname': 'snappingTolerance',
                                                       'numArgs': 1,
                                                       'shortname': 'st'},
                                 'useSnapping': {'args': '<type __builtin__.bool>',
                                                 'longname': 'useSnapping',
                                                 'numArgs': 1,
                                                 'shortname': 'us'}},
                       'shortFlags': {'a': 'absolute',
                                      'ch': 'history',
                                      'd': 'direction',
                                      'e': 'edit',
                                      'ed': 'edgeDirection',
                                      'ex': 'exists',
                                      'i1': 'image1',
                                      'i2': 'image2',
                                      'i3': 'image3',
                                      'n': 'name',
                                      'q': 'query',
                                      'sp': 'snappingPoints',
                                      'st': 'snappingTolerance',
                                      'us': 'useSnapping'},
                       'type': 'context'},
  'polySmooth': {'flags': {'boundaryRule': {'args': '<type __builtin__.int>',
                                            'longname': 'boundaryRule',
                                            'numArgs': 1,
                                            'shortname': 'bnr'},
                           'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'continuity': {'args': '<type __builtin__.float>',
                                          'longname': 'continuity',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'c'},
                           'degree': {'args': '<type __builtin__.int>',
                                      'longname': 'degree',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'deg'},
                           'divisions': {'args': '<type __builtin__.int>',
                                         'longname': 'divisions',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dv'},
                           'divisionsPerEdge': {'args': '<type __builtin__.int>',
                                                'longname': 'divisionsPerEdge',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dpe'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'keepBorder': {'args': '<type __builtin__.bool>',
                                          'longname': 'keepBorder',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'kb'},
                           'keepHardEdge': {'args': '<type __builtin__.bool>',
                                            'longname': 'keepHardEdge',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'khe'},
                           'keepMapBorders': {'args': '<type __builtin__.int>',
                                              'longname': 'keepMapBorders',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'kmb'},
                           'keepSelectionBorder': {'args': '<type __builtin__.bool>',
                                                   'longname': 'keepSelectionBorder',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ksb'},
                           'keepTesselation': {'args': '<type __builtin__.bool>',
                                               'longname': 'keepTesselation',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'xkt'},
                           'keepTessellation': {'args': '<type __builtin__.bool>',
                                                'longname': 'keepTessellation',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'kt'},
                           'method': {'args': '<type __builtin__.int>',
                                      'longname': 'method',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'mth'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'osdCreaseMethod': {'args': '<type __builtin__.int>',
                                               'longname': 'osdCreaseMethod',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ocr'},
                           'osdFvarBoundary': {'args': '<type __builtin__.int>',
                                               'longname': 'osdFvarBoundary',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ofb'},
                           'osdFvarPropagateCorners': {'args': '<type __builtin__.bool>',
                                                       'longname': 'osdFvarPropagateCorners',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ofc'},
                           'osdSmoothTriangles': {'args': '<type __builtin__.bool>',
                                                  'longname': 'osdSmoothTriangles',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ost'},
                           'osdVertBoundary': {'args': '<type __builtin__.int>',
                                               'longname': 'osdVertBoundary',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ovb'},
                           'propagateEdgeHardness': {'args': '<type __builtin__.bool>',
                                                     'longname': 'propagateEdgeHardness',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'peh'},
                           'pushStrength': {'args': '<type __builtin__.float>',
                                            'longname': 'pushStrength',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ps'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'roundness': {'args': '<type __builtin__.float>',
                                         'longname': 'roundness',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ro'},
                           'smoothUVs': {'args': '<type __builtin__.bool>',
                                         'longname': 'smoothUVs',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'suv'},
                           'subdivisionLevels': {'args': '<type __builtin__.int>',
                                                 'longname': 'subdivisionLevels',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'sl'},
                           'subdivisionType': {'args': '<type __builtin__.int>',
                                               'longname': 'subdivisionType',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sdt'}},
                 'shortFlags': {'bnr': 'boundaryRule',
                                'c': 'continuity',
                                'cch': 'caching',
                                'ch': 'constructionHistory',
                                'deg': 'degree',
                                'dpe': 'divisionsPerEdge',
                                'dv': 'divisions',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'kb': 'keepBorder',
                                'khe': 'keepHardEdge',
                                'kmb': 'keepMapBorders',
                                'ksb': 'keepSelectionBorder',
                                'kt': 'keepTessellation',
                                'mth': 'method',
                                'n': 'name',
                                'nds': 'nodeState',
                                'ocr': 'osdCreaseMethod',
                                'ofb': 'osdFvarBoundary',
                                'ofc': 'osdFvarPropagateCorners',
                                'ost': 'osdSmoothTriangles',
                                'ovb': 'osdVertBoundary',
                                'peh': 'propagateEdgeHardness',
                                'ps': 'pushStrength',
                                'q': 'query',
                                'ro': 'roundness',
                                'sdt': 'subdivisionType',
                                'sl': 'subdivisionLevels',
                                'suv': 'smoothUVs',
                                'xkt': 'keepTesselation'},
                 'type': 'modeling'},
  'polySoftEdge': {'flags': {'angle': {'args': '<type __builtin__.float>',
                                       'longname': 'angle',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'a'},
                             'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'worldSpace': {'args': '<type __builtin__.bool>',
                                            'longname': 'worldSpace',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ws'}},
                   'shortFlags': {'a': 'angle',
                                  'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'q': 'query',
                                  'ws': 'worldSpace'},
                   'type': 'modeling'},
  'polySphere': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'axis',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 3,
                                    'shortname': 'ax'},
                           'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'createUVs': {'args': '<type __builtin__.int>',
                                         'longname': 'createUVs',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'cuv'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'resultNeedsCasting': True,
                                         'shortname': 'nds'},
                           'object': {'args': '<type __builtin__.bool>',
                                      'longname': 'object',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'radius': {'args': '<type __builtin__.float>',
                                      'longname': 'radius',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'r'},
                           'subdivisionsAxis': {'args': '<type __builtin__.int>',
                                                'longname': 'subdivisionsAxis',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'resultNeedsCasting': True,
                                                'shortname': 'sa'},
                           'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                  'longname': 'subdivisionsHeight',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'resultNeedsCasting': True,
                                                  'shortname': 'sh'},
                           'subdivisionsX': {'args': '<type __builtin__.int>',
                                             'longname': 'subdivisionsX',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'resultNeedsCasting': True,
                                             'shortname': 'sx'},
                           'subdivisionsY': {'args': '<type __builtin__.int>',
                                             'longname': 'subdivisionsY',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'resultNeedsCasting': True,
                                             'shortname': 'sy'},
                           'texture': {'args': '<type __builtin__.int>',
                                       'longname': 'texture',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'tx'}},
                 'shortFlags': {'ax': 'axis',
                                'cch': 'caching',
                                'ch': 'constructionHistory',
                                'cuv': 'createUVs',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'n': 'name',
                                'nds': 'nodeState',
                                'o': 'object',
                                'q': 'query',
                                'r': 'radius',
                                'sa': 'subdivisionsAxis',
                                'sh': 'subdivisionsHeight',
                                'sx': 'subdivisionsX',
                                'sy': 'subdivisionsY',
                                'tx': 'texture'},
                 'type': 'modeling'},
  'polySphericalProjection': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                                    'longname': 'caching',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'cch'},
                                        'constructionHistory': {'args': '<type __builtin__.bool>',
                                                                'longname': 'constructionHistory',
                                                                'modes': ['create',
                                                                          'query'],
                                                                'numArgs': 1,
                                                                'shortname': 'ch'},
                                        'createNewMap': {'args': '<type __builtin__.bool>',
                                                         'longname': 'createNewMap',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'cm'},
                                        'edit': {'args': '<type __builtin__.bool>',
                                                 'longname': 'edit',
                                                 'numArgs': 0,
                                                 'shortname': 'e'},
                                        'frozen': {'args': '<type __builtin__.bool>',
                                                   'longname': 'frozen',
                                                   'numArgs': 1,
                                                   'shortname': 'fzn'},
                                        'imageCenter': {'args': ['<type __builtin__.float>',
                                                                 '<type __builtin__.float>'],
                                                        'longname': 'imageCenter',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 2,
                                                        'shortname': 'ic'},
                                        'imageCenterX': {'args': '<type __builtin__.float>',
                                                         'longname': 'imageCenterX',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'icx'},
                                        'imageCenterY': {'args': '<type __builtin__.float>',
                                                         'longname': 'imageCenterY',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'icy'},
                                        'imageScale': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'imageScale',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 2,
                                                       'shortname': 'imageScale'},
                                        'imageScaleU': {'args': '<type __builtin__.float>',
                                                        'longname': 'imageScaleU',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'isu'},
                                        'imageScaleV': {'args': '<type __builtin__.float>',
                                                        'longname': 'imageScaleV',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'isv'},
                                        'insertBeforeDeformers': {'args': '<type __builtin__.bool>',
                                                                  'longname': 'insertBeforeDeformers',
                                                                  'modes': ['create'],
                                                                  'numArgs': 1,
                                                                  'shortname': 'ibd'},
                                        'keepImageRatio': {'args': '<type __builtin__.bool>',
                                                           'longname': 'keepImageRatio',
                                                           'modes': ['create'],
                                                           'numArgs': 1,
                                                           'shortname': 'kir'},
                                        'mapDirection': {'args': '<type __builtin__.str>',
                                                         'longname': 'mapDirection',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'md'},
                                        'name': {'args': '<type __builtin__.str>',
                                                 'longname': 'name',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'n'},
                                        'nodeState': {'args': '<type __builtin__.int>',
                                                      'longname': 'nodeState',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'nds'},
                                        'perInstance': {'args': '<type __builtin__.bool>',
                                                        'longname': 'perInstance',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'pi'},
                                        'projectionCenter': {'args': ['<type __builtin__.float>',
                                                                      '<type __builtin__.float>',
                                                                      '<type __builtin__.float>'],
                                                             'longname': 'projectionCenter',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 3,
                                                             'shortname': 'pc'},
                                        'projectionCenterX': {'args': '<type __builtin__.float>',
                                                              'longname': 'projectionCenterX',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'pcx'},
                                        'projectionCenterY': {'args': '<type __builtin__.float>',
                                                              'longname': 'projectionCenterY',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'pcy'},
                                        'projectionCenterZ': {'args': '<type __builtin__.float>',
                                                              'longname': 'projectionCenterZ',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'pcz'},
                                        'projectionHeight': {'args': '<type __builtin__.float>',
                                                             'longname': 'projectionHeight',
                                                             'numArgs': 1,
                                                             'shortname': 'ph'},
                                        'projectionHorizontalSweep': {'args': '<type __builtin__.float>',
                                                                      'longname': 'projectionHorizontalSweep',
                                                                      'modes': ['create',
                                                                                'query',
                                                                                'edit'],
                                                                      'numArgs': 1,
                                                                      'shortname': 'phs'},
                                        'projectionScale': {'args': ['<type __builtin__.float>',
                                                                     '<type __builtin__.float>'],
                                                            'longname': 'projectionScale',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 2,
                                                            'shortname': 'ps'},
                                        'projectionScaleU': {'args': '<type __builtin__.float>',
                                                             'longname': 'projectionScaleU',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'psu'},
                                        'projectionScaleV': {'args': '<type __builtin__.float>',
                                                             'longname': 'projectionScaleV',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'psv'},
                                        'query': {'args': '<type __builtin__.bool>',
                                                  'longname': 'query',
                                                  'numArgs': 0,
                                                  'shortname': 'q'},
                                        'radius': {'args': '<type __builtin__.float>',
                                                   'longname': 'radius',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'r'},
                                        'rotate': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'rotate',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'ro'},
                                        'rotateX': {'args': '<type __builtin__.float>',
                                                    'longname': 'rotateX',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'rx'},
                                        'rotateY': {'args': '<type __builtin__.float>',
                                                    'longname': 'rotateY',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ry'},
                                        'rotateZ': {'args': '<type __builtin__.float>',
                                                    'longname': 'rotateZ',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'rz'},
                                        'rotationAngle': {'args': '<type __builtin__.float>',
                                                          'longname': 'rotationAngle',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ra'},
                                        'seamCorrect': {'args': '<type __builtin__.bool>',
                                                        'longname': 'seamCorrect',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'sc'},
                                        'smartFit': {'args': '<type __builtin__.bool>',
                                                     'longname': 'smartFit',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'sf'},
                                        'worldSpace': {'args': '<type __builtin__.bool>',
                                                       'longname': 'worldSpace',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ws'}},
                              'removedFlags': {'is': 'imageScale'},
                              'shortFlags': {'cch': 'caching',
                                             'ch': 'constructionHistory',
                                             'cm': 'createNewMap',
                                             'e': 'edit',
                                             'fzn': 'frozen',
                                             'ibd': 'insertBeforeDeformers',
                                             'ic': 'imageCenter',
                                             'icx': 'imageCenterX',
                                             'icy': 'imageCenterY',
                                             'imageScale': 'imageScale',
                                             'isu': 'imageScaleU',
                                             'isv': 'imageScaleV',
                                             'kir': 'keepImageRatio',
                                             'md': 'mapDirection',
                                             'n': 'name',
                                             'nds': 'nodeState',
                                             'pc': 'projectionCenter',
                                             'pcx': 'projectionCenterX',
                                             'pcy': 'projectionCenterY',
                                             'pcz': 'projectionCenterZ',
                                             'ph': 'projectionHeight',
                                             'phs': 'projectionHorizontalSweep',
                                             'pi': 'perInstance',
                                             'ps': 'projectionScale',
                                             'psu': 'projectionScaleU',
                                             'psv': 'projectionScaleV',
                                             'q': 'query',
                                             'r': 'radius',
                                             'ra': 'rotationAngle',
                                             'ro': 'rotate',
                                             'rx': 'rotateX',
                                             'ry': 'rotateY',
                                             'rz': 'rotateZ',
                                             'sc': 'seamCorrect',
                                             'sf': 'smartFit',
                                             'ws': 'worldSpace'},
                              'type': 'modeling'},
  'polySpinEdge': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'offset': {'args': '<type __builtin__.int>',
                                        'longname': 'offset',
                                        'numArgs': 1,
                                        'shortname': 'off'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'reverse': {'args': '<type __builtin__.bool>',
                                         'longname': 'reverse',
                                         'numArgs': 1,
                                         'shortname': 'rev'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'off': 'offset',
                                  'q': 'query',
                                  'rev': 'reverse'},
                   'type': 'other'},
  'polySplit': {'flags': {'adjustEdgeFlow': {'args': '<type __builtin__.float>',
                                             'longname': 'adjustEdgeFlow',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'aef'},
                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'constructionHistory',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                          'detachEdges': {'args': '<type __builtin__.bool>',
                                          'longname': 'detachEdges',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'de'},
                          'edgepoint': {'args': ['<type __builtin__.int>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'edgepoint',
                                        'modes': ['create', 'multiuse'],
                                        'numArgs': 2,
                                        'shortname': 'ep'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'facepoint': {'args': ['<type __builtin__.int>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'facepoint',
                                        'modes': ['create', 'multiuse'],
                                        'numArgs': 4,
                                        'shortname': 'fp'},
                          'insertWithEdgeFlow': {'args': '<type __builtin__.bool>',
                                                 'longname': 'insertWithEdgeFlow',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ief'},
                          'insertpoint': {'args': ['<type __builtin__.int>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'insertpoint',
                                          'modes': ['create', 'multiuse'],
                                          'numArgs': 4,
                                          'shortname': 'ip'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'projectedCurve': {'args': 'PyNode',
                                             'longname': 'projectedCurve',
                                             'modes': ['create',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'pc'},
                          'projectedCurveTolerance': {'args': '<type __builtin__.float>',
                                                      'longname': 'projectedCurveTolerance',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'pct'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'smoothingangle': {'args': '<type __builtin__.float>',
                                             'longname': 'smoothingangle',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'sma'},
                          'subdivision': {'args': '<type __builtin__.int>',
                                          'longname': 'subdivision',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 's'}},
                'shortFlags': {'aef': 'adjustEdgeFlow',
                               'ch': 'constructionHistory',
                               'de': 'detachEdges',
                               'e': 'edit',
                               'ep': 'edgepoint',
                               'fp': 'facepoint',
                               'ief': 'insertWithEdgeFlow',
                               'ip': 'insertpoint',
                               'n': 'name',
                               'pc': 'projectedCurve',
                               'pct': 'projectedCurveTolerance',
                               'q': 'query',
                               's': 'subdivision',
                               'sma': 'smoothingangle'},
                'type': 'modeling'},
  'polySplitCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enablesnap': {'args': '<type __builtin__.bool>',
                                            'longname': 'enablesnap',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'es'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'image1': {'args': '<type __builtin__.str>',
                                        'longname': 'image1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i1'},
                             'image2': {'args': '<type __builtin__.str>',
                                        'longname': 'image2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i2'},
                             'image3': {'args': '<type __builtin__.str>',
                                        'longname': 'image3',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i3'},
                             'magnetsnap': {'args': '<type __builtin__.int>',
                                            'longname': 'magnetsnap',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ms'},
                             'precsnap': {'args': '<type __builtin__.float>',
                                          'longname': 'precsnap',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ps'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'smoothingangle': {'args': '<type __builtin__.float>',
                                                'longname': 'smoothingangle',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sma'},
                             'snaptoedge': {'args': '<type __builtin__.bool>',
                                            'longname': 'snaptoedge',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ste'},
                             'subdivision': {'args': '<type __builtin__.int>',
                                             'longname': 'subdivision',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 's'}},
                   'shortFlags': {'e': 'edit',
                                  'es': 'enablesnap',
                                  'ex': 'exists',
                                  'i1': 'image1',
                                  'i2': 'image2',
                                  'i3': 'image3',
                                  'ms': 'magnetsnap',
                                  'ps': 'precsnap',
                                  'q': 'query',
                                  's': 'subdivision',
                                  'sma': 'smoothingangle',
                                  'ste': 'snaptoedge'},
                   'type': 'context'},
  'polySplitCtx2': {'flags': {'adjustEdgeFlow': {'args': '<type __builtin__.float>',
                                                 'longname': 'adjustEdgeFlow',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'aef'},
                              'constrainToEdges': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constrainToEdges',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cte'},
                              'detachEdges': {'args': '<type __builtin__.bool>',
                                              'longname': 'detachEdges',
                                              'numArgs': 1,
                                              'shortname': 'de'},
                              'edgeMagnets': {'args': '<type __builtin__.int>',
                                              'longname': 'edgeMagnets',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'em'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'highlightPointColor': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'highlightPointColor',
                                                      'numArgs': 3,
                                                      'shortname': 'hpc'},
                              'image1': {'args': '<type __builtin__.str>',
                                         'longname': 'image1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i1'},
                              'image2': {'args': '<type __builtin__.str>',
                                         'longname': 'image2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i2'},
                              'image3': {'args': '<type __builtin__.str>',
                                         'longname': 'image3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i3'},
                              'insertWithEdgeFlow': {'args': '<type __builtin__.bool>',
                                                     'longname': 'insertWithEdgeFlow',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ief'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'snapTolerance': {'args': '<type __builtin__.float>',
                                                'longname': 'snapTolerance',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'st'},
                              'snappedToEdgeColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'snappedToEdgeColor',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'sec'},
                              'snappedToFaceColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'snappedToFaceColor',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'sfc'},
                              'snappedToMagnetColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'snappedToMagnetColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'smc'},
                              'snappedToVertexColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'snappedToVertexColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'svc'},
                              'snappingTolerance': {'args': '<type __builtin__.float>',
                                                    'longname': 'snappingTolerance',
                                                    'numArgs': 1,
                                                    'shortname': 'st'},
                              'splitLineColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'splitLineColor',
                                                 'numArgs': 3,
                                                 'shortname': 'slc'}},
                    'shortFlags': {'aef': 'adjustEdgeFlow',
                                   'cte': 'constrainToEdges',
                                   'de': 'detachEdges',
                                   'e': 'edit',
                                   'em': 'edgeMagnets',
                                   'ex': 'exists',
                                   'hpc': 'highlightPointColor',
                                   'i1': 'image1',
                                   'i2': 'image2',
                                   'i3': 'image3',
                                   'ief': 'insertWithEdgeFlow',
                                   'q': 'query',
                                   'sec': 'snappedToEdgeColor',
                                   'sfc': 'snappedToFaceColor',
                                   'slc': 'splitLineColor',
                                   'smc': 'snappedToMagnetColor',
                                   'st': 'snappingTolerance',
                                   'svc': 'snappedToVertexColor'},
                    'type': 'general'},
  'polySplitEdge': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cch'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nds'},
                              'operation': {'args': '<type __builtin__.int>',
                                            'longname': 'operation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'op'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'}},
                    'shortFlags': {'cch': 'caching',
                                   'ch': 'constructionHistory',
                                   'e': 'edit',
                                   'fzn': 'frozen',
                                   'n': 'name',
                                   'nds': 'nodeState',
                                   'op': 'operation',
                                   'q': 'query'},
                    'type': 'modeling'},
  'polySplitRing': {'flags': {'adjustEdgeFlow': {'args': '<type __builtin__.float>',
                                                 'longname': 'adjustEdgeFlow',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'aef'},
                              'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cch'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'direction': {'args': '<type __builtin__.bool>',
                                            'longname': 'direction',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dr'},
                              'divisions': {'args': '<type __builtin__.int>',
                                            'longname': 'divisions',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'div'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'enableProfileCurve': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableProfileCurve',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'epc'},
                              'fixQuads': {'args': '<type __builtin__.bool>',
                                           'longname': 'fixQuads',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'fq'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'insertWithEdgeFlow': {'args': '<type __builtin__.bool>',
                                                     'longname': 'insertWithEdgeFlow',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ief'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nds'},
                              'profileCurveInputOffset': {'args': '<type __builtin__.float>',
                                                          'longname': 'profileCurveInputOffset',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'pio'},
                              'profileCurveInputScale': {'args': '<type __builtin__.float>',
                                                         'longname': 'profileCurveInputScale',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'pis'},
                              'profileCurve_FloatValue': {'args': '<type __builtin__.float>',
                                                          'longname': 'profileCurve_FloatValue',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'pfv'},
                              'profileCurve_Interp': {'args': '<type __builtin__.int>',
                                                      'longname': 'profileCurve_Interp',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'pi'},
                              'profileCurve_Position': {'args': '<type __builtin__.float>',
                                                        'longname': 'profileCurve_Position',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'pp'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'rootEdge': {'args': '<type __builtin__.int>',
                                           'longname': 'rootEdge',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 're'},
                              'smoothingAngle': {'args': '<type __builtin__.float>',
                                                 'longname': 'smoothingAngle',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sma'},
                              'splitType': {'args': '<type __builtin__.int>',
                                            'longname': 'splitType',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'stp'},
                              'useEqualMultiplier': {'args': '<type __builtin__.bool>',
                                                     'longname': 'useEqualMultiplier',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'uem'},
                              'useFaceNormalsAtEnds': {'args': '<type __builtin__.bool>',
                                                       'longname': 'useFaceNormalsAtEnds',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'fne'},
                              'weight': {'args': '<type __builtin__.float>',
                                         'longname': 'weight',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'wt'},
                              'worldSpace': {'args': '<type __builtin__.bool>',
                                             'longname': 'worldSpace',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ws'}},
                    'shortFlags': {'aef': 'adjustEdgeFlow',
                                   'cch': 'caching',
                                   'ch': 'constructionHistory',
                                   'div': 'divisions',
                                   'dr': 'direction',
                                   'e': 'edit',
                                   'epc': 'enableProfileCurve',
                                   'fne': 'useFaceNormalsAtEnds',
                                   'fq': 'fixQuads',
                                   'fzn': 'frozen',
                                   'ief': 'insertWithEdgeFlow',
                                   'n': 'name',
                                   'nds': 'nodeState',
                                   'pfv': 'profileCurve_FloatValue',
                                   'pi': 'profileCurve_Interp',
                                   'pio': 'profileCurveInputOffset',
                                   'pis': 'profileCurveInputScale',
                                   'pp': 'profileCurve_Position',
                                   'q': 'query',
                                   're': 'rootEdge',
                                   'sma': 'smoothingAngle',
                                   'stp': 'splitType',
                                   'uem': 'useEqualMultiplier',
                                   'ws': 'worldSpace',
                                   'wt': 'weight'},
                    'type': 'modeling'},
  'polySplitVertex': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                            'longname': 'caching',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cch'},
                                'constructionHistory': {'args': '<type __builtin__.bool>',
                                                        'longname': 'constructionHistory',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'ch'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'frozen': {'args': '<type __builtin__.bool>',
                                           'longname': 'frozen',
                                           'numArgs': 1,
                                           'shortname': 'fzn'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'nodeState': {'args': '<type __builtin__.int>',
                                              'longname': 'nodeState',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nds'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'worldSpace': {'args': '<type __builtin__.bool>',
                                               'longname': 'worldSpace',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ws'}},
                      'shortFlags': {'cch': 'caching',
                                     'ch': 'constructionHistory',
                                     'e': 'edit',
                                     'fzn': 'frozen',
                                     'n': 'name',
                                     'nds': 'nodeState',
                                     'q': 'query',
                                     'ws': 'worldSpace'},
                      'type': 'modeling'},
  'polyStraightenUVBorder': {'flags': {'blendOriginal': {'args': '<type __builtin__.float>',
                                                         'longname': 'blendOriginal',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'bo'},
                                       'caching': {'args': '<type __builtin__.bool>',
                                                   'longname': 'caching',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cch'},
                                       'constructionHistory': {'args': '<type __builtin__.bool>',
                                                               'longname': 'constructionHistory',
                                                               'modes': ['create',
                                                                         'query'],
                                                               'numArgs': 1,
                                                               'shortname': 'ch'},
                                       'curvature': {'args': '<type __builtin__.float>',
                                                     'longname': 'curvature',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'c'},
                                       'edit': {'args': '<type __builtin__.bool>',
                                                'longname': 'edit',
                                                'numArgs': 0,
                                                'shortname': 'e'},
                                       'frozen': {'args': '<type __builtin__.bool>',
                                                  'longname': 'frozen',
                                                  'numArgs': 1,
                                                  'shortname': 'fzn'},
                                       'gapTolerance': {'args': '<type __builtin__.int>',
                                                        'longname': 'gapTolerance',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'gt'},
                                       'name': {'args': '<type __builtin__.str>',
                                                'longname': 'name',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'n'},
                                       'nodeState': {'args': '<type __builtin__.int>',
                                                     'longname': 'nodeState',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'nds'},
                                       'preserveLength': {'args': '<type __builtin__.float>',
                                                          'longname': 'preserveLength',
                                                          'modes': ['create',
                                                                    'query'],
                                                          'numArgs': 1,
                                                          'shortname': 'pl'},
                                       'query': {'args': '<type __builtin__.bool>',
                                                 'longname': 'query',
                                                 'numArgs': 0,
                                                 'shortname': 'q'},
                                       'uvSetName': {'args': '<type __builtin__.str>',
                                                     'longname': 'uvSetName',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'uvs'},
                                       'worldSpace': {'args': '<type __builtin__.bool>',
                                                      'longname': 'worldSpace',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ws'}},
                             'shortFlags': {'bo': 'blendOriginal',
                                            'c': 'curvature',
                                            'cch': 'caching',
                                            'ch': 'constructionHistory',
                                            'e': 'edit',
                                            'fzn': 'frozen',
                                            'gt': 'gapTolerance',
                                            'n': 'name',
                                            'nds': 'nodeState',
                                            'pl': 'preserveLength',
                                            'q': 'query',
                                            'uvs': 'uvSetName',
                                            'ws': 'worldSpace'},
                             'type': 'modeling'},
  'polySubdivideEdge': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                              'longname': 'caching',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cch'},
                                  'constructionHistory': {'args': '<type __builtin__.bool>',
                                                          'longname': 'constructionHistory',
                                                          'modes': ['create',
                                                                    'query'],
                                                          'numArgs': 1,
                                                          'shortname': 'ch'},
                                  'divisions': {'args': '<type __builtin__.int>',
                                                'longname': 'divisions',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dv'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'frozen': {'args': '<type __builtin__.bool>',
                                             'longname': 'frozen',
                                             'numArgs': 1,
                                             'shortname': 'fzn'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'nodeState': {'args': '<type __builtin__.int>',
                                                'longname': 'nodeState',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nds'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'size': {'args': '<type __builtin__.float>',
                                           'longname': 'size',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 's'},
                                  'worldSpace': {'args': '<type __builtin__.bool>',
                                                 'longname': 'worldSpace',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ws'}},
                        'shortFlags': {'cch': 'caching',
                                       'ch': 'constructionHistory',
                                       'dv': 'divisions',
                                       'e': 'edit',
                                       'fzn': 'frozen',
                                       'n': 'name',
                                       'nds': 'nodeState',
                                       'q': 'query',
                                       's': 'size',
                                       'ws': 'worldSpace'},
                        'type': 'modeling'},
  'polySubdivideFacet': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                               'longname': 'caching',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cch'},
                                   'constructionHistory': {'args': '<type __builtin__.bool>',
                                                           'longname': 'constructionHistory',
                                                           'modes': ['create',
                                                                     'query'],
                                                           'numArgs': 1,
                                                           'shortname': 'ch'},
                                   'divisions': {'args': '<type __builtin__.int>',
                                                 'longname': 'divisions',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dv'},
                                   'divisionsU': {'args': '<type __builtin__.int>',
                                                  'longname': 'divisionsU',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'duv'},
                                   'divisionsV': {'args': '<type __builtin__.int>',
                                                  'longname': 'divisionsV',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dvv'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'frozen': {'args': '<type __builtin__.bool>',
                                              'longname': 'frozen',
                                              'numArgs': 1,
                                              'shortname': 'fzn'},
                                   'mode': {'args': '<type __builtin__.int>',
                                            'longname': 'mode',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'm'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'nodeState': {'args': '<type __builtin__.int>',
                                                 'longname': 'nodeState',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'nds'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'subdMethod': {'args': '<type __builtin__.int>',
                                                  'longname': 'subdMethod',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'}},
                         'shortFlags': {'cch': 'caching',
                                        'ch': 'constructionHistory',
                                        'duv': 'divisionsU',
                                        'dv': 'divisions',
                                        'dvv': 'divisionsV',
                                        'e': 'edit',
                                        'fzn': 'frozen',
                                        'm': 'mode',
                                        'n': 'name',
                                        'nds': 'nodeState',
                                        'q': 'query',
                                        'sbm': 'subdMethod'},
                         'type': 'modeling'},
  'polySuperCtx': {'flags': {'attach': {'args': '<type __builtin__.str>',
                                        'longname': 'attach',
                                        'modes': ['multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'a'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'image1': {'args': '<type __builtin__.str>',
                                        'longname': 'image1',
                                        'numArgs': 1,
                                        'shortname': 'i1'},
                             'image2': {'args': '<type __builtin__.str>',
                                        'longname': 'image2',
                                        'numArgs': 1,
                                        'shortname': 'i2'},
                             'image3': {'args': '<type __builtin__.str>',
                                        'longname': 'image3',
                                        'numArgs': 1,
                                        'shortname': 'i3'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'a': 'attach',
                                  'e': 'edit',
                                  'ex': 'exists',
                                  'i1': 'image1',
                                  'i2': 'image2',
                                  'i3': 'image3',
                                  'q': 'query'},
                   'type': 'context'},
  'polyTestPop': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'polyToCurve': {'flags': {'addUnderTransform': {'args': '<type __builtin__.bool>',
                                                  'longname': 'addUnderTransform',
                                                  'numArgs': 1,
                                                  'shortname': 'aut'},
                            'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'conformToSmoothMeshPreview': {'args': '<type __builtin__.bool>',
                                                           'longname': 'conformToSmoothMeshPreview',
                                                           'numArgs': 1,
                                                           'shortname': 'usm'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'degree': {'args': '<type __builtin__.int>',
                                       'longname': 'degree',
                                       'numArgs': 1,
                                       'shortname': 'dg'},
                            'displaySmoothMesh': {'args': '<type __builtin__.int>',
                                                  'longname': 'displaySmoothMesh',
                                                  'numArgs': 1,
                                                  'shortname': 'dsm'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'form': {'args': '<type __builtin__.int>',
                                     'longname': 'form',
                                     'numArgs': 1,
                                     'shortname': 'f'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'}},
                  'shortFlags': {'aut': 'addUnderTransform',
                                 'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'dg': 'degree',
                                 'dsm': 'displaySmoothMesh',
                                 'e': 'edit',
                                 'f': 'form',
                                 'fzn': 'frozen',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'o': 'object',
                                 'q': 'query',
                                 'usm': 'conformToSmoothMeshPreview'},
                  'type': 'other'},
  'polyToSubdiv': {'flags': {'absolutePosition': {'args': '<type __builtin__.bool>',
                                                  'longname': 'absolutePosition',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ap'},
                             'addUnderTransform': {'args': '<type __builtin__.bool>',
                                                   'longname': 'addUnderTransform',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'aut'},
                             'applyMatrixToResult': {'args': '<type __builtin__.bool>',
                                                     'longname': 'applyMatrixToResult',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'amr'},
                             'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'maxEdgesPerVert': {'args': '<type __builtin__.int>',
                                                 'longname': 'maxEdgesPerVert',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'me'},
                             'maxPolyCount': {'args': '<type __builtin__.int>',
                                              'longname': 'maxPolyCount',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'mpc'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'object': {'args': '<type __builtin__.bool>',
                                        'longname': 'object',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                             'preserveVertexOrdering': {'args': '<type __builtin__.bool>',
                                                        'longname': 'preserveVertexOrdering',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'pvo'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'quickConvert': {'args': '<type __builtin__.bool>',
                                              'longname': 'quickConvert',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'qc'},
                             'uvPoints': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'uvPoints',
                                          'modes': ['create',
                                                    'query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'uvp'},
                             'uvPointsU': {'args': '<type __builtin__.float>',
                                           'longname': 'uvPointsU',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'uvu'},
                             'uvPointsV': {'args': '<type __builtin__.float>',
                                           'longname': 'uvPointsV',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'uvv'},
                             'uvTreatment': {'args': '<type __builtin__.int>',
                                             'longname': 'uvTreatment',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'uvt'}},
                   'shortFlags': {'amr': 'applyMatrixToResult',
                                  'ap': 'absolutePosition',
                                  'aut': 'addUnderTransform',
                                  'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'me': 'maxEdgesPerVert',
                                  'mpc': 'maxPolyCount',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'o': 'object',
                                  'pvo': 'preserveVertexOrdering',
                                  'q': 'query',
                                  'qc': 'quickConvert',
                                  'uvp': 'uvPoints',
                                  'uvt': 'uvTreatment',
                                  'uvu': 'uvPointsU',
                                  'uvv': 'uvPointsV'},
                   'type': 'modeling'},
  'polyTorus': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'axis',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 3,
                                   'shortname': 'ax'},
                          'caching': {'args': '<type __builtin__.bool>',
                                      'longname': 'caching',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsCasting': True,
                                      'shortname': 'cch'},
                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'constructionHistory',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                          'createUVs': {'args': '<type __builtin__.bool>',
                                        'longname': 'createUVs',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'cuv'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'frozen': {'args': '<type __builtin__.bool>',
                                     'longname': 'frozen',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'fzn'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nodeState': {'args': '<type __builtin__.int>',
                                        'longname': 'nodeState',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'nds'},
                          'object': {'args': '<type __builtin__.bool>',
                                     'longname': 'object',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'o'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'radius': {'args': '<type __builtin__.float>',
                                     'longname': 'radius',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'r'},
                          'sectionRadius': {'args': '<type __builtin__.float>',
                                            'longname': 'sectionRadius',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sr'},
                          'subdivisionsAxis': {'args': '<type __builtin__.int>',
                                               'longname': 'subdivisionsAxis',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'resultNeedsCasting': True,
                                               'shortname': 'sa'},
                          'subdivisionsHeight': {'args': '<type __builtin__.int>',
                                                 'longname': 'subdivisionsHeight',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'resultNeedsCasting': True,
                                                 'shortname': 'sh'},
                          'subdivisionsX': {'args': '<type __builtin__.int>',
                                            'longname': 'subdivisionsX',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'resultNeedsCasting': True,
                                            'shortname': 'sx'},
                          'subdivisionsY': {'args': '<type __builtin__.int>',
                                            'longname': 'subdivisionsY',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'resultNeedsCasting': True,
                                            'shortname': 'sy'},
                          'texture': {'args': '<type __builtin__.bool>',
                                      'longname': 'texture',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'resultNeedsCasting': True,
                                      'shortname': 'tx'},
                          'twist': {'args': '<type __builtin__.float>',
                                    'longname': 'twist',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'tw'}},
                'shortFlags': {'ax': 'axis',
                               'cch': 'caching',
                               'ch': 'constructionHistory',
                               'cuv': 'createUVs',
                               'e': 'edit',
                               'fzn': 'frozen',
                               'n': 'name',
                               'nds': 'nodeState',
                               'o': 'object',
                               'q': 'query',
                               'r': 'radius',
                               'sa': 'subdivisionsAxis',
                               'sh': 'subdivisionsHeight',
                               'sr': 'sectionRadius',
                               'sx': 'subdivisionsX',
                               'sy': 'subdivisionsY',
                               'tw': 'twist',
                               'tx': 'texture'},
                'type': 'modeling'},
  'polyTransfer': {'flags': {'alternateObject': {'args': '<type __builtin__.str>',
                                                 'longname': 'alternateObject',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ao'},
                             'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'uvSets': {'args': '<type __builtin__.bool>',
                                        'longname': 'uvSets',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'uv'},
                             'vertexColor': {'args': '<type __builtin__.bool>',
                                             'longname': 'vertexColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'vc'},
                             'vertices': {'args': '<type __builtin__.bool>',
                                          'longname': 'vertices',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'v'}},
                   'shortFlags': {'ao': 'alternateObject',
                                  'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'q': 'query',
                                  'uv': 'uvSets',
                                  'v': 'vertices',
                                  'vc': 'vertexColor'},
                   'type': 'modeling'},
  'polyTriangulate': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                            'longname': 'caching',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cch'},
                                'constructionHistory': {'args': '<type __builtin__.bool>',
                                                        'longname': 'constructionHistory',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'ch'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'frozen': {'args': '<type __builtin__.bool>',
                                           'longname': 'frozen',
                                           'numArgs': 1,
                                           'shortname': 'fzn'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'nodeState': {'args': '<type __builtin__.int>',
                                              'longname': 'nodeState',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nds'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'}},
                      'shortFlags': {'cch': 'caching',
                                     'ch': 'constructionHistory',
                                     'e': 'edit',
                                     'fzn': 'frozen',
                                     'n': 'name',
                                     'nds': 'nodeState',
                                     'q': 'query'},
                      'type': 'modeling'},
  'polyUVCoverage': {'flags': {'uvRange': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'uvRange',
                                           'modes': ['create'],
                                           'numArgs': 4,
                                           'shortname': 'ur'}},
                     'shortFlags': {'ur': 'uvRange'},
                     'type': 'modeling'},
  'polyUVOverlap': {'flags': {'nonOverlappingComponents': {'args': '<type __builtin__.bool>',
                                                           'longname': 'nonOverlappingComponents',
                                                           'modes': ['create'],
                                                           'numArgs': 0,
                                                           'shortname': 'noc'},
                              'overlappingComponents': {'args': '<type __builtin__.bool>',
                                                        'longname': 'overlappingComponents',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'oc'}},
                    'shortFlags': {'noc': 'nonOverlappingComponents',
                                   'oc': 'overlappingComponents'},
                    'type': 'modeling'},
  'polyUVRectangle': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                            'longname': 'caching',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cch'},
                                'constructionHistory': {'args': '<type __builtin__.bool>',
                                                        'longname': 'constructionHistory',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'ch'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'frozen': {'args': '<type __builtin__.bool>',
                                           'longname': 'frozen',
                                           'numArgs': 1,
                                           'shortname': 'fzn'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'nodeState': {'args': '<type __builtin__.int>',
                                              'longname': 'nodeState',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nds'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'}},
                      'shortFlags': {'cch': 'caching',
                                     'ch': 'constructionHistory',
                                     'e': 'edit',
                                     'fzn': 'frozen',
                                     'n': 'name',
                                     'nds': 'nodeState',
                                     'q': 'query'},
                      'type': 'modeling'},
  'polyUVSet': {'flags': {'allUVSets': {'args': '<type __builtin__.bool>',
                                        'longname': 'allUVSets',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'auv'},
                          'allUVSetsIndices': {'args': '<type __builtin__.bool>',
                                               'longname': 'allUVSetsIndices',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'uvn'},
                          'allUVSetsWithCount': {'args': '<type __builtin__.bool>',
                                                 'longname': 'allUVSetsWithCount',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'awc'},
                          'copy': {'args': '<type __builtin__.bool>',
                                   'longname': 'copy',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 0,
                                   'shortname': 'cp'},
                          'create': {'args': '<type __builtin__.bool>',
                                     'longname': 'create',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'cr'},
                          'currentLastUVSet': {'args': '<type __builtin__.bool>',
                                               'longname': 'currentLastUVSet',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 0,
                                               'shortname': 'luv'},
                          'currentPerInstanceUVSet': {'args': '<type __builtin__.bool>',
                                                      'longname': 'currentPerInstanceUVSet',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'cpi'},
                          'currentUVSet': {'args': '<type __builtin__.bool>',
                                           'longname': 'currentUVSet',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 0,
                                           'shortname': 'cuv'},
                          'delete': {'args': '<type __builtin__.bool>',
                                     'longname': 'delete',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'd'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'genNewUVSet': {'args': '<type __builtin__.bool>',
                                          'longname': 'genNewUVSet',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'gen'},
                          'newUVSet': {'args': '<type __builtin__.str>',
                                       'longname': 'newUVSet',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'nuv'},
                          'perInstance': {'args': '<type __builtin__.bool>',
                                          'longname': 'perInstance',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'pi'},
                          'projections': {'args': '<type __builtin__.bool>',
                                          'longname': 'projections',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'pr'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'rename': {'args': '<type __builtin__.bool>',
                                     'longname': 'rename',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'rn'},
                          'reorder': {'args': '<type __builtin__.bool>',
                                      'longname': 'reorder',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'ro'},
                          'shareInstances': {'args': '<type __builtin__.bool>',
                                             'longname': 'shareInstances',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 0,
                                             'shortname': 'si'},
                          'unshared': {'args': '<type __builtin__.bool>',
                                       'longname': 'unshared',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'us'},
                          'uvSet': {'args': '<type __builtin__.str>',
                                    'longname': 'uvSet',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'uvs'}},
                'shortFlags': {'auv': 'allUVSets',
                               'awc': 'allUVSetsWithCount',
                               'cp': 'copy',
                               'cpi': 'currentPerInstanceUVSet',
                               'cr': 'create',
                               'cuv': 'currentUVSet',
                               'd': 'delete',
                               'e': 'edit',
                               'gen': 'genNewUVSet',
                               'luv': 'currentLastUVSet',
                               'nuv': 'newUVSet',
                               'pi': 'perInstance',
                               'pr': 'projections',
                               'q': 'query',
                               'rn': 'rename',
                               'ro': 'reorder',
                               'si': 'shareInstances',
                               'us': 'unshared',
                               'uvn': 'allUVSetsIndices',
                               'uvs': 'uvSet'},
                'type': 'modeling'},
  'polyUVStackSimilarShells': {'flags': {'onlyMatch': {'args': '<type __builtin__.bool>',
                                                       'longname': 'onlyMatch',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'om'},
                                         'tolerance': {'args': '<type __builtin__.float>',
                                                       'longname': 'tolerance',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'to'}},
                               'shortFlags': {'om': 'onlyMatch',
                                              'to': 'tolerance'},
                               'type': 'modeling'},
  'polyUVStackSimilarShellsCmd': {'flags': {},
                                  'shortFlags': {},
                                  'type': 'other'},
  'polyUnite': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                      'longname': 'caching',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'cch'},
                          'centerPivot': {'args': '<type __builtin__.bool>',
                                          'longname': 'centerPivot',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'cp'},
                          'constructionHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'constructionHistory',
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'frozen': {'args': '<type __builtin__.bool>',
                                     'longname': 'frozen',
                                     'numArgs': 1,
                                     'shortname': 'fzn'},
                          'mergeUVSets': {'args': '<type __builtin__.int>',
                                          'longname': 'mergeUVSets',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'muv'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nodeState': {'args': '<type __builtin__.int>',
                                        'longname': 'nodeState',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'nds'},
                          'object': {'args': '<type __builtin__.bool>',
                                     'longname': 'object',
                                     'numArgs': 1,
                                     'shortname': 'o'},
                          'objectPivot': {'args': '<type __builtin__.bool>',
                                          'longname': 'objectPivot',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'op'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'}},
                'shortFlags': {'cch': 'caching',
                               'ch': 'constructionHistory',
                               'cp': 'centerPivot',
                               'e': 'edit',
                               'fzn': 'frozen',
                               'muv': 'mergeUVSets',
                               'n': 'name',
                               'nds': 'nodeState',
                               'o': 'object',
                               'op': 'objectPivot',
                               'q': 'query'},
                'type': 'modeling'},
  'polyUniteSkinned': {'flags': {'centerPivot': {'args': '<type __builtin__.bool>',
                                                 'longname': 'centerPivot',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'cp'},
                                 'constructionHistory': {'args': '<type __builtin__.bool>',
                                                         'longname': 'constructionHistory',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ch'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'mergeUVSets': {'args': '<type __builtin__.int>',
                                                 'longname': 'mergeUVSets',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'muv'},
                                 'objectPivot': {'args': '<type __builtin__.bool>',
                                                 'longname': 'objectPivot',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'op'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'}},
                       'shortFlags': {'ch': 'constructionHistory',
                                      'cp': 'centerPivot',
                                      'e': 'edit',
                                      'muv': 'mergeUVSets',
                                      'op': 'objectPivot',
                                      'q': 'query'},
                       'type': 'animation'},
  'polyVertexNormalCtx': {'flags': {'deformmode': {'args': '<type __builtin__.bool>',
                                                   'longname': 'deformmode',
                                                   'numArgs': 1,
                                                   'shortname': 'dm'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'lockmode': {'args': '<type __builtin__.int>',
                                                 'longname': 'lockmode',
                                                 'numArgs': 1,
                                                 'shortname': 'lm'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'relativemode': {'args': '<type __builtin__.int>',
                                                     'longname': 'relativemode',
                                                     'numArgs': 1,
                                                     'shortname': 'rm'},
                                    'rotatemode': {'args': '<type __builtin__.int>',
                                                   'longname': 'rotatemode',
                                                   'numArgs': 1,
                                                   'shortname': 'rot'}},
                          'shortFlags': {'dm': 'deformmode',
                                         'e': 'edit',
                                         'ex': 'exists',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'lm': 'lockmode',
                                         'q': 'query',
                                         'rm': 'relativemode',
                                         'rot': 'rotatemode'},
                          'type': 'context'},
  'polyWarpImage': {'flags': {'background': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'background',
                                             'numArgs': 3,
                                             'shortname': 'bg'},
                              'bilinear': {'args': '<type __builtin__.bool>',
                                           'longname': 'bilinear',
                                           'numArgs': 0,
                                           'shortname': 'b'},
                              'fileFormat': {'args': '<type __builtin__.str>',
                                             'longname': 'fileFormat',
                                             'numArgs': 1,
                                             'shortname': 'ff'},
                              'inputName': {'args': '<type __builtin__.str>',
                                            'longname': 'inputName',
                                            'numArgs': 1,
                                            'shortname': 'inputName'},
                              'inputUvSetName': {'args': '<type __builtin__.str>',
                                                 'longname': 'inputUvSetName',
                                                 'numArgs': 1,
                                                 'shortname': 'iuv'},
                              'noAlpha': {'args': '<type __builtin__.bool>',
                                          'longname': 'noAlpha',
                                          'numArgs': 0,
                                          'shortname': 'na'},
                              'outputName': {'args': '<type __builtin__.str>',
                                             'longname': 'outputName',
                                             'numArgs': 1,
                                             'shortname': 'on'},
                              'outputUvSetName': {'args': '<type __builtin__.str>',
                                                  'longname': 'outputUvSetName',
                                                  'numArgs': 1,
                                                  'shortname': 'ouv'},
                              'overwrite': {'args': '<type __builtin__.bool>',
                                            'longname': 'overwrite',
                                            'numArgs': 0,
                                            'shortname': 'o'},
                              'tiled': {'args': '<type __builtin__.bool>',
                                        'longname': 'tiled',
                                        'numArgs': 0,
                                        'shortname': 't'},
                              'xResolution': {'args': '<type __builtin__.int>',
                                              'longname': 'xResolution',
                                              'numArgs': 1,
                                              'shortname': 'xr'},
                              'yResolution': {'args': '<type __builtin__.int>',
                                              'longname': 'yResolution',
                                              'numArgs': 1,
                                              'shortname': 'yr'}},
                    'removedFlags': {'in': 'inputName'},
                    'shortFlags': {'b': 'bilinear',
                                   'bg': 'background',
                                   'ff': 'fileFormat',
                                   'inputName': 'inputName',
                                   'iuv': 'inputUvSetName',
                                   'na': 'noAlpha',
                                   'o': 'overwrite',
                                   'on': 'outputName',
                                   'ouv': 'outputUvSetName',
                                   't': 'tiled',
                                   'xr': 'xResolution',
                                   'yr': 'yResolution'},
                    'type': 'other'},
  'polyWedgeFace': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'axis',
                                       'modes': ['create'],
                                       'numArgs': 3,
                                       'shortname': 'ax'},
                              'axisX': {'args': '<type __builtin__.float>',
                                        'longname': 'axisX',
                                        'numArgs': 1,
                                        'shortname': 'asx'},
                              'axisY': {'args': '<type __builtin__.float>',
                                        'longname': 'axisY',
                                        'numArgs': 1,
                                        'shortname': 'asy'},
                              'axisZ': {'args': '<type __builtin__.float>',
                                        'longname': 'axisZ',
                                        'numArgs': 1,
                                        'shortname': 'asz'},
                              'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cch'},
                              'center': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'center',
                                         'modes': ['create'],
                                         'numArgs': 3,
                                         'shortname': 'cen'},
                              'centerX': {'args': '<type __builtin__.float>',
                                          'longname': 'centerX',
                                          'numArgs': 1,
                                          'shortname': 'ctx'},
                              'centerY': {'args': '<type __builtin__.float>',
                                          'longname': 'centerY',
                                          'numArgs': 1,
                                          'shortname': 'cty'},
                              'centerZ': {'args': '<type __builtin__.float>',
                                          'longname': 'centerZ',
                                          'numArgs': 1,
                                          'shortname': 'ctz'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'divisions': {'args': '<type __builtin__.int>',
                                            'longname': 'divisions',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'd'},
                              'edge': {'args': '<type __builtin__.int>',
                                       'longname': 'edge',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'ed'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nds'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'wedgeAngle': {'args': '<type __builtin__.float>',
                                             'longname': 'wedgeAngle',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'wa'},
                              'worldSpace': {'args': '<type __builtin__.bool>',
                                             'longname': 'worldSpace',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ws'}},
                    'removedFlags': {'as': 'axis'},
                    'shortFlags': {'asx': 'axisX',
                                   'asy': 'axisY',
                                   'asz': 'axisZ',
                                   'axis': 'axis',
                                   'cch': 'caching',
                                   'ch': 'constructionHistory',
                                   'ct': 'center',
                                   'ctx': 'centerX',
                                   'cty': 'centerY',
                                   'ctz': 'centerZ',
                                   'd': 'divisions',
                                   'e': 'edit',
                                   'ed': 'edge',
                                   'fzn': 'frozen',
                                   'n': 'name',
                                   'nds': 'nodeState',
                                   'q': 'query',
                                   'wa': 'wedgeAngle',
                                   'ws': 'worldSpace'},
                    'type': 'modeling'},
  'popupMenu': {'callbackFlags': ['postMenuCommandOnce',
                                  'pmo',
                                  'postMenuCommand',
                                  'pmc'],
                'flags': {'allowOptionBoxes': {'args': '<type __builtin__.bool>',
                                               'longname': 'allowOptionBoxes',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'aob'},
                          'altModifier': {'args': '<type __builtin__.bool>',
                                          'longname': 'altModifier',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'alt'},
                          'button': {'args': '<type __builtin__.int>',
                                     'longname': 'button',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'b'},
                          'ctrlModifier': {'args': '<type __builtin__.bool>',
                                           'longname': 'ctrlModifier',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ctl'},
                          'defineTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'defineTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dt'},
                          'deleteAllItems': {'args': '<type __builtin__.bool>',
                                             'longname': 'deleteAllItems',
                                             'modes': ['edit'],
                                             'numArgs': 0,
                                             'shortname': 'dai'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'itemArray': {'args': '<type __builtin__.bool>',
                                        'longname': 'itemArray',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'ia'},
                          'markingMenu': {'args': '<type __builtin__.bool>',
                                          'longname': 'markingMenu',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'mm'},
                          'numberOfItems': {'args': '<type __builtin__.bool>',
                                            'longname': 'numberOfItems',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ni'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                          'postMenuCommand': {'args': 'script',
                                              'longname': 'postMenuCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pmc'},
                          'postMenuCommandOnce': {'args': '<type __builtin__.bool>',
                                                  'longname': 'postMenuCommandOnce',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pmo'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'shiftModifier': {'args': '<type __builtin__.bool>',
                                            'longname': 'shiftModifier',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sh'},
                          'useTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'useTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ut'}},
                'shortFlags': {'alt': 'altModifier',
                               'aob': 'allowOptionBoxes',
                               'b': 'button',
                               'ctl': 'ctrlModifier',
                               'dai': 'deleteAllItems',
                               'dt': 'defineTemplate',
                               'e': 'edit',
                               'ex': 'exists',
                               'ia': 'itemArray',
                               'mm': 'markingMenu',
                               'ni': 'numberOfItems',
                               'p': 'parent',
                               'pmc': 'postMenuCommand',
                               'pmo': 'postMenuCommandOnce',
                               'q': 'query',
                               'sh': 'shiftModifier',
                               'ut': 'useTemplate'},
                'type': 'windows'},
  'pose': {'flags': {'allPoses': {'args': '<type __builtin__.bool>',
                                  'longname': 'allPoses',
                                  'modes': ['query'],
                                  'numArgs': 0,
                                  'shortname': 'ap'},
                     'apply': {'args': '<type __builtin__.bool>',
                               'longname': 'apply',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'a'},
                     'edit': {'args': '<type __builtin__.bool>',
                              'longname': 'edit',
                              'numArgs': 0,
                              'shortname': 'e'},
                     'name': {'args': '<type __builtin__.str>',
                              'longname': 'name',
                              'modes': ['create', 'query'],
                              'numArgs': 1,
                              'shortname': 'n'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'}},
           'shortFlags': {'a': 'apply',
                          'ap': 'allPoses',
                          'e': 'edit',
                          'n': 'name',
                          'q': 'query'},
           'type': 'animation'},
  'poseEditor': {'flags': {'control': {'args': '<type __builtin__.bool>',
                                       'longname': 'control',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'ctl'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'filter': {'args': '<type __builtin__.str>',
                                      'longname': 'filter',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'f'},
                           'forceMainConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'forceMainConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'fmc'},
                           'highlightConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'highlightConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'hlc'},
                           'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                  'longname': 'lockMainConnection',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'lck'},
                           'mainListConnection': {'args': '<type __builtin__.str>',
                                                  'longname': 'mainListConnection',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mlc'},
                           'panel': {'args': '<type __builtin__.str>',
                                     'longname': 'panel',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'pnl'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'selectionConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'selectionConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'slc'},
                           'stateString': {'args': '<type __builtin__.bool>',
                                           'longname': 'stateString',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'sts'},
                           'targetControlList': {'args': '<type __builtin__.bool>',
                                                 'longname': 'targetControlList',
                                                 'numArgs': 0,
                                                 'shortname': 'tcl'},
                           'targetList': {'args': '<type __builtin__.bool>',
                                          'longname': 'targetList',
                                          'numArgs': 0,
                                          'shortname': 'tl'},
                           'unParent': {'args': '<type __builtin__.bool>',
                                        'longname': 'unParent',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'up'},
                           'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'unlockMainConnection',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'ulk'},
                           'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'updateMainConnection',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'upd'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'}},
                 'shortFlags': {'ctl': 'control',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'ex': 'exists',
                                'f': 'filter',
                                'fmc': 'forceMainConnection',
                                'hlc': 'highlightConnection',
                                'lck': 'lockMainConnection',
                                'mlc': 'mainListConnection',
                                'p': 'parent',
                                'pnl': 'panel',
                                'q': 'query',
                                'slc': 'selectionConnection',
                                'sts': 'stateString',
                                'tcl': 'targetControlList',
                                'tl': 'targetList',
                                'ulk': 'unlockMainConnection',
                                'up': 'unParent',
                                'upd': 'updateMainConnection',
                                'ut': 'useTemplate'},
                 'type': 'animation'},
  'posePanel': {'callbackFlags': ['popupMenuProcedure', 'pmp'],
                'flags': {'control': {'args': '<type __builtin__.bool>',
                                      'longname': 'control',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'ctl'},
                          'copy': {'args': '<type __builtin__.str>',
                                   'longname': 'copy',
                                   'modes': ['edit'],
                                   'numArgs': 1,
                                   'shortname': 'cp'},
                          'createString': {'args': '<type __builtin__.bool>',
                                           'longname': 'createString',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'cs'},
                          'defineTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'defineTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dt'},
                          'docTag': {'args': '<type __builtin__.str>',
                                     'longname': 'docTag',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'dtg'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'editString': {'args': '<type __builtin__.bool>',
                                         'longname': 'editString',
                                         'modes': ['edit'],
                                         'numArgs': 0,
                                         'shortname': 'es'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'init': {'args': '<type __builtin__.bool>',
                                   'longname': 'init',
                                   'modes': ['create', 'edit'],
                                   'numArgs': 0,
                                   'shortname': 'init'},
                          'isUnique': {'args': '<type __builtin__.bool>',
                                       'longname': 'isUnique',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'iu'},
                          'label': {'args': '<type __builtin__.str>',
                                    'longname': 'label',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'l'},
                          'menuBarRepeatLast': {'args': '<type __builtin__.bool>',
                                                'longname': 'menuBarRepeatLast',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'mrl'},
                          'menuBarVisible': {'args': '<type __builtin__.bool>',
                                             'longname': 'menuBarVisible',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'mbv'},
                          'needsInit': {'args': '<type __builtin__.bool>',
                                        'longname': 'needsInit',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'ni'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                          'popupMenuProcedure': {'args': 'script',
                                                 'longname': 'popupMenuProcedure',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'pmp'},
                          'poseEditor': {'args': '<type __builtin__.bool>',
                                         'longname': 'poseEditor',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'pe'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'replacePanel': {'args': '<type __builtin__.str>',
                                           'longname': 'replacePanel',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'rp'},
                          'tearOff': {'args': '<type __builtin__.bool>',
                                      'longname': 'tearOff',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'to'},
                          'tearOffCopy': {'args': '<type __builtin__.str>',
                                          'longname': 'tearOffCopy',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'toc'},
                          'tearOffRestore': {'args': '<type __builtin__.bool>',
                                             'longname': 'tearOffRestore',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'tor'},
                          'unParent': {'args': '<type __builtin__.bool>',
                                       'longname': 'unParent',
                                       'modes': ['edit'],
                                       'numArgs': 0,
                                       'shortname': 'up'},
                          'useTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'useTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ut'}},
                'removedFlags': {'in': 'init'},
                'shortFlags': {'cp': 'copy',
                               'cs': 'createString',
                               'ctl': 'control',
                               'dt': 'defineTemplate',
                               'dtg': 'docTag',
                               'e': 'edit',
                               'es': 'editString',
                               'ex': 'exists',
                               'init': 'init',
                               'iu': 'isUnique',
                               'l': 'label',
                               'mbv': 'menuBarVisible',
                               'mrl': 'menuBarRepeatLast',
                               'ni': 'needsInit',
                               'p': 'parent',
                               'pe': 'poseEditor',
                               'pmp': 'popupMenuProcedure',
                               'q': 'query',
                               'rp': 'replacePanel',
                               'to': 'tearOff',
                               'toc': 'tearOffCopy',
                               'tor': 'tearOffRestore',
                               'up': 'unParent',
                               'ut': 'useTemplate'},
                'type': 'animation'},
  'preferredRenderer': {'flags': {'fallback': {'args': '<type __builtin__.str>',
                                               'longname': 'fallback',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'f'},
                                  'makeCurrent': {'args': '<type __builtin__.bool>',
                                                  'longname': 'makeCurrent',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'mc'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'}},
                        'shortFlags': {'f': 'fallback',
                                       'mc': 'makeCurrent',
                                       'q': 'query'},
                        'type': 'rendering'},
  'preloadRefEd': {'callbackFlags': ['selectCommand', 'sc'],
                   'flags': {'control': {'args': '<type __builtin__.bool>',
                                         'longname': 'control',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'ctl'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'filter': {'args': '<type __builtin__.str>',
                                        'longname': 'filter',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'f'},
                             'forceMainConnection': {'args': '<type __builtin__.str>',
                                                     'longname': 'forceMainConnection',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'fmc'},
                             'highlightConnection': {'args': '<type __builtin__.str>',
                                                     'longname': 'highlightConnection',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'hlc'},
                             'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'lockMainConnection',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'lck'},
                             'mainListConnection': {'args': '<type __builtin__.str>',
                                                    'longname': 'mainListConnection',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'mlc'},
                             'panel': {'args': '<type __builtin__.str>',
                                       'longname': 'panel',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'pnl'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'selectCommand': {'args': 'script',
                                               'longname': 'selectCommand',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sc'},
                             'selectFileNode': {'args': '<type __builtin__.bool>',
                                                'longname': 'selectFileNode',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'sf'},
                             'selectionConnection': {'args': '<type __builtin__.str>',
                                                     'longname': 'selectionConnection',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'slc'},
                             'stateString': {'args': '<type __builtin__.bool>',
                                             'longname': 'stateString',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'sts'},
                             'unParent': {'args': '<type __builtin__.bool>',
                                          'longname': 'unParent',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'up'},
                             'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                      'longname': 'unlockMainConnection',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'ulk'},
                             'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                      'longname': 'updateMainConnection',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'upd'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'}},
                   'shortFlags': {'ctl': 'control',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ex': 'exists',
                                  'f': 'filter',
                                  'fmc': 'forceMainConnection',
                                  'hlc': 'highlightConnection',
                                  'lck': 'lockMainConnection',
                                  'mlc': 'mainListConnection',
                                  'p': 'parent',
                                  'pnl': 'panel',
                                  'q': 'query',
                                  'sc': 'selectCommand',
                                  'sf': 'selectFileNode',
                                  'slc': 'selectionConnection',
                                  'sts': 'stateString',
                                  'ulk': 'unlockMainConnection',
                                  'up': 'unParent',
                                  'upd': 'updateMainConnection',
                                  'ut': 'useTemplate'},
                   'type': 'system'},
  'prepareRender': {'callbackFlags': ['preRenderFrame',
                                      'prf',
                                      'preRender',
                                      'prr',
                                      'postRenderFrame',
                                      'pof',
                                      'postRender',
                                      'por',
                                      'settingsUI',
                                      'sui',
                                      'postRenderLayer',
                                      'pol',
                                      'traversalSetInit',
                                      'tsi',
                                      'preRenderLayer',
                                      'prl'],
                    'flags': {'defaultTraversalSet': {'args': '<type __builtin__.str>',
                                                      'longname': 'defaultTraversalSet',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dt'},
                              'deregister': {'args': '<type __builtin__.str>',
                                             'longname': 'deregister',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'd'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'invokePostRender': {'args': '<type __builtin__.bool>',
                                                   'longname': 'invokePostRender',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'ior'},
                              'invokePostRenderFrame': {'args': '<type __builtin__.bool>',
                                                        'longname': 'invokePostRenderFrame',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'iof'},
                              'invokePostRenderLayer': {'args': '<type __builtin__.bool>',
                                                        'longname': 'invokePostRenderLayer',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'iol'},
                              'invokePreRender': {'args': '<type __builtin__.bool>',
                                                  'longname': 'invokePreRender',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'irr'},
                              'invokePreRenderFrame': {'args': '<type __builtin__.bool>',
                                                       'longname': 'invokePreRenderFrame',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'irf'},
                              'invokePreRenderLayer': {'args': '<type __builtin__.bool>',
                                                       'longname': 'invokePreRenderLayer',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'irl'},
                              'invokeSettingsUI': {'args': '<type __builtin__.bool>',
                                                   'longname': 'invokeSettingsUI',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'isu'},
                              'label': {'args': '<type __builtin__.str>',
                                        'longname': 'label',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'lbl'},
                              'listTraversalSets': {'args': '<type __builtin__.bool>',
                                                    'longname': 'listTraversalSets',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'lt'},
                              'postRender': {'args': 'script',
                                             'longname': 'postRender',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'por'},
                              'postRenderFrame': {'args': 'script',
                                                  'longname': 'postRenderFrame',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pof'},
                              'postRenderLayer': {'args': 'script',
                                                  'longname': 'postRenderLayer',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pol'},
                              'preRender': {'args': 'script',
                                            'longname': 'preRender',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'prr'},
                              'preRenderFrame': {'args': 'script',
                                                 'longname': 'preRenderFrame',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'prf'},
                              'preRenderLayer': {'args': 'script',
                                                 'longname': 'preRenderLayer',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'prl'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'restore': {'args': '<type __builtin__.bool>',
                                          'longname': 'restore',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'rtr'},
                              'saveAssemblyConfig': {'args': '<type __builtin__.bool>',
                                                     'longname': 'saveAssemblyConfig',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sac'},
                              'settingsUI': {'args': 'script',
                                             'longname': 'settingsUI',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sui'},
                              'setup': {'args': '<type __builtin__.bool>',
                                        'longname': 'setup',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'stp'},
                              'traversalSet': {'args': '<type __builtin__.str>',
                                               'longname': 'traversalSet',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ts'},
                              'traversalSetInit': {'args': 'script',
                                                   'longname': 'traversalSetInit',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'tsi'}},
                    'shortFlags': {'d': 'deregister',
                                   'dt': 'defaultTraversalSet',
                                   'e': 'edit',
                                   'iof': 'invokePostRenderFrame',
                                   'iol': 'invokePostRenderLayer',
                                   'ior': 'invokePostRender',
                                   'irf': 'invokePreRenderFrame',
                                   'irl': 'invokePreRenderLayer',
                                   'irr': 'invokePreRender',
                                   'isu': 'invokeSettingsUI',
                                   'lbl': 'label',
                                   'lt': 'listTraversalSets',
                                   'pof': 'postRenderFrame',
                                   'pol': 'postRenderLayer',
                                   'por': 'postRender',
                                   'prf': 'preRenderFrame',
                                   'prl': 'preRenderLayer',
                                   'prr': 'preRender',
                                   'q': 'query',
                                   'rtr': 'restore',
                                   'sac': 'saveAssemblyConfig',
                                   'stp': 'setup',
                                   'sui': 'settingsUI',
                                   'ts': 'traversalSet',
                                   'tsi': 'traversalSetInit'},
                    'type': 'rendering'},
  'profiler': {'flags': {'addCategory': {'args': '<type __builtin__.str>',
                                         'longname': 'addCategory',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'a'},
                         'allCategories': {'args': '<type __builtin__.bool>',
                                           'longname': 'allCategories',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'ac'},
                         'bufferSize': {'args': '<type __builtin__.int>',
                                        'longname': 'bufferSize',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'b'},
                         'categoryIndex': {'args': '<type __builtin__.int>',
                                           'longname': 'categoryIndex',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'ci'},
                         'categoryIndexToName': {'args': '<type __builtin__.int>',
                                                 'longname': 'categoryIndexToName',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'cin'},
                         'categoryInfo': {'args': '<type __builtin__.str>',
                                          'longname': 'categoryInfo',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'cf'},
                         'categoryName': {'args': '<type __builtin__.str>',
                                          'longname': 'categoryName',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'cn'},
                         'categoryNameToIndex': {'args': '<type __builtin__.str>',
                                                 'longname': 'categoryNameToIndex',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'cni'},
                         'categoryRecording': {'args': '<type __builtin__.bool>',
                                               'longname': 'categoryRecording',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'cr'},
                         'clearAllMelInstrumentation': {'args': '<type __builtin__.bool>',
                                                        'longname': 'clearAllMelInstrumentation',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'cam'},
                         'colorIndex': {'args': '<type __builtin__.int>',
                                        'longname': 'colorIndex',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'coi'},
                         'eventCPUId': {'args': '<type __builtin__.bool>',
                                        'longname': 'eventCPUId',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'eci'},
                         'eventCategory': {'args': '<type __builtin__.bool>',
                                           'longname': 'eventCategory',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'eca'},
                         'eventColor': {'args': '<type __builtin__.bool>',
                                        'longname': 'eventColor',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'eco'},
                         'eventCount': {'args': '<type __builtin__.bool>',
                                        'longname': 'eventCount',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'ec'},
                         'eventDescription': {'args': '<type __builtin__.bool>',
                                              'longname': 'eventDescription',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'ed'},
                         'eventDuration': {'args': '<type __builtin__.bool>',
                                           'longname': 'eventDuration',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'edu'},
                         'eventIndex': {'args': '<type __builtin__.int>',
                                        'longname': 'eventIndex',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'ei'},
                         'eventName': {'args': '<type __builtin__.bool>',
                                       'longname': 'eventName',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'en'},
                         'eventStartTime': {'args': '<type __builtin__.bool>',
                                            'longname': 'eventStartTime',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'et'},
                         'eventThreadId': {'args': '<type __builtin__.bool>',
                                           'longname': 'eventThreadId',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'eti'},
                         'instrumentMel': {'args': '<type __builtin__.bool>',
                                           'longname': 'instrumentMel',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'instrumentMel'},
                         'load': {'args': '<type __builtin__.str>',
                                  'longname': 'load',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'l'},
                         'output': {'args': '<type __builtin__.str>',
                                    'longname': 'output',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'o'},
                         'procedureDescription': {'args': '<type __builtin__.str>',
                                                  'longname': 'procedureDescription',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'pd'},
                         'procedureName': {'args': '<type __builtin__.str>',
                                           'longname': 'procedureName',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'pn'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'removeCategory': {'args': '<type __builtin__.str>',
                                            'longname': 'removeCategory',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'rc'},
                         'reset': {'args': '<type __builtin__.bool>',
                                   'longname': 'reset',
                                   'modes': ['create', 'query'],
                                   'numArgs': 0,
                                   'shortname': 'r'},
                         'sampling': {'args': '<type __builtin__.bool>',
                                      'longname': 'sampling',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                         'signalEvent': {'args': '<type __builtin__.bool>',
                                         'longname': 'signalEvent',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'sig'},
                         'signalMelEvent': {'args': '<type __builtin__.bool>',
                                            'longname': 'signalMelEvent',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'sim'}},
               'removedFlags': {'in': 'instrumentMel'},
               'shortFlags': {'a': 'addCategory',
                              'ac': 'allCategories',
                              'b': 'bufferSize',
                              'cam': 'clearAllMelInstrumentation',
                              'cf': 'categoryInfo',
                              'ci': 'categoryIndex',
                              'cin': 'categoryIndexToName',
                              'cn': 'categoryName',
                              'cni': 'categoryNameToIndex',
                              'coi': 'colorIndex',
                              'cr': 'categoryRecording',
                              'ec': 'eventCount',
                              'eca': 'eventCategory',
                              'eci': 'eventCPUId',
                              'eco': 'eventColor',
                              'ed': 'eventDescription',
                              'edu': 'eventDuration',
                              'ei': 'eventIndex',
                              'en': 'eventName',
                              'et': 'eventStartTime',
                              'eti': 'eventThreadId',
                              'instrumentMel': 'instrumentMel',
                              'l': 'load',
                              'o': 'output',
                              'pd': 'procedureDescription',
                              'pn': 'procedureName',
                              'q': 'query',
                              'r': 'reset',
                              'rc': 'removeCategory',
                              's': 'sampling',
                              'sig': 'signalEvent',
                              'sim': 'signalMelEvent'},
               'type': 'system'},
  'profilerTool': {'flags': {'categoryView': {'args': '<type __builtin__.bool>',
                                              'longname': 'categoryView',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'cat'},
                             'collapseSelectedEvents': {'args': '<type __builtin__.bool>',
                                                        'longname': 'collapseSelectedEvents',
                                                        'modes': ['edit'],
                                                        'numArgs': 0,
                                                        'shortname': 'cs'},
                             'collapseSelectedEventsRepetition': {'args': '<type __builtin__.bool>',
                                                                  'longname': 'collapseSelectedEventsRepetition',
                                                                  'modes': ['edit'],
                                                                  'numArgs': 0,
                                                                  'shortname': 'csr'},
                             'cpuView': {'args': '<type __builtin__.bool>',
                                         'longname': 'cpuView',
                                         'modes': ['edit'],
                                         'numArgs': 0,
                                         'shortname': 'cpu'},
                             'destroy': {'args': '<type __builtin__.bool>',
                                         'longname': 'destroy',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'dtr'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'eventTypes': {'args': '<type __builtin__.bool>',
                                            'longname': 'eventTypes',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'et'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'expandSelectedEvents': {'args': '<type __builtin__.bool>',
                                                      'longname': 'expandSelectedEvents',
                                                      'modes': ['edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'es'},
                             'expandSelectedEventsRepetition': {'args': '<type __builtin__.bool>',
                                                                'longname': 'expandSelectedEventsRepetition',
                                                                'modes': ['edit'],
                                                                'numArgs': 0,
                                                                'shortname': 'esr'},
                             'findNext': {'args': '<type __builtin__.bool>',
                                          'longname': 'findNext',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'fn'},
                             'findPrevious': {'args': '<type __builtin__.bool>',
                                              'longname': 'findPrevious',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fp'},
                             'frameAll': {'args': '<type __builtin__.bool>',
                                          'longname': 'frameAll',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'fa'},
                             'frameSelected': {'args': '<type __builtin__.bool>',
                                               'longname': 'frameSelected',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'fs'},
                             'isolateSegment': {'args': '<type __builtin__.int>',
                                                'longname': 'isolateSegment',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'isolateSegment'},
                             'make': {'args': '<type __builtin__.bool>',
                                      'longname': 'make',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'mk'},
                             'matchWholeWord': {'args': '<type __builtin__.bool>',
                                                'longname': 'matchWholeWord',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'mww'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'searchEvent': {'args': '<type __builtin__.str>',
                                             'longname': 'searchEvent',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'se'},
                             'segmentCount': {'args': '<type __builtin__.bool>',
                                              'longname': 'segmentCount',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'sc'},
                             'showAllEvent': {'args': '<type __builtin__.bool>',
                                              'longname': 'showAllEvent',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'sa'},
                             'showCriticalPath': {'args': '<type __builtin__.bool>',
                                                  'longname': 'showCriticalPath',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'scp'},
                             'showSelectedEvents': {'args': '<type __builtin__.bool>',
                                                    'longname': 'showSelectedEvents',
                                                    'modes': ['edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ss'},
                             'showSelectedEventsRepetition': {'args': '<type __builtin__.bool>',
                                                              'longname': 'showSelectedEventsRepetition',
                                                              'modes': ['edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'ssr'},
                             'threadView': {'args': '<type __builtin__.bool>',
                                            'longname': 'threadView',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'thd'},
                             'unisolateSegment': {'args': '<type __builtin__.bool>',
                                                  'longname': 'unisolateSegment',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'uis'}},
                   'removedFlags': {'is': 'isolateSegment'},
                   'shortFlags': {'cat': 'categoryView',
                                  'cpu': 'cpuView',
                                  'cs': 'collapseSelectedEvents',
                                  'csr': 'collapseSelectedEventsRepetition',
                                  'dtr': 'destroy',
                                  'e': 'edit',
                                  'es': 'expandSelectedEvents',
                                  'esr': 'expandSelectedEventsRepetition',
                                  'et': 'eventTypes',
                                  'ex': 'exists',
                                  'fa': 'frameAll',
                                  'fn': 'findNext',
                                  'fp': 'findPrevious',
                                  'fs': 'frameSelected',
                                  'isolateSegment': 'isolateSegment',
                                  'mk': 'make',
                                  'mww': 'matchWholeWord',
                                  'q': 'query',
                                  'sa': 'showAllEvent',
                                  'sc': 'segmentCount',
                                  'scp': 'showCriticalPath',
                                  'se': 'searchEvent',
                                  'ss': 'showSelectedEvents',
                                  'ssr': 'showSelectedEventsRepetition',
                                  'thd': 'threadView',
                                  'uis': 'unisolateSegment'},
                   'type': 'system'},
  'progressBar': {'callbackFlags': ['visibleChangeCommand',
                                    'vcc',
                                    'dragCallback',
                                    'dgc',
                                    'dropCallback',
                                    'dpc'],
                  'flags': {'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'beginProgress': {'args': '<type __builtin__.bool>',
                                              'longname': 'beginProgress',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'bp'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'endProgress': {'args': '<type __builtin__.bool>',
                                            'longname': 'endProgress',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'ep'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'isCancelled': {'args': '<type __builtin__.bool>',
                                            'longname': 'isCancelled',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'ic'},
                            'isInterruptable': {'args': '<type __builtin__.bool>',
                                                'longname': 'isInterruptable',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ii'},
                            'isMainProgressBar': {'args': '<type __builtin__.bool>',
                                                  'longname': 'isMainProgressBar',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'imp'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'maxValue': {'args': '<type __builtin__.int>',
                                         'longname': 'maxValue',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'max'},
                            'minValue': {'args': '<type __builtin__.int>',
                                         'longname': 'minValue',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'min'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'progress': {'args': '<type __builtin__.int>',
                                         'longname': 'progress',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'pr'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'status': {'args': '<type __builtin__.str>',
                                       'longname': 'status',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'st'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'step': {'args': '<type __builtin__.int>',
                                     'longname': 'step',
                                     'modes': ['edit'],
                                     'numArgs': 1,
                                     'shortname': 's'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'bp': 'beginProgress',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'ep': 'endProgress',
                                 'ex': 'exists',
                                 'fpn': 'fullPathName',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'ic': 'isCancelled',
                                 'ii': 'isInterruptable',
                                 'imp': 'isMainProgressBar',
                                 'io': 'isObscured',
                                 'm': 'manage',
                                 'max': 'maxValue',
                                 'min': 'minValue',
                                 'nbg': 'noBackground',
                                 'npm': 'numberOfPopupMenus',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'pr': 'progress',
                                 'q': 'query',
                                 's': 'step',
                                 'sbm': 'statusBarMessage',
                                 'st': 'status',
                                 'ut': 'useTemplate',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'progressWindow': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'endProgress': {'args': '<type __builtin__.bool>',
                                               'longname': 'endProgress',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ep'},
                               'isCancelled': {'args': '<type __builtin__.bool>',
                                               'longname': 'isCancelled',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'ic'},
                               'isInterruptable': {'args': '<type __builtin__.bool>',
                                                   'longname': 'isInterruptable',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ii'},
                               'maxValue': {'args': '<type __builtin__.int>',
                                            'longname': 'maxValue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'max'},
                               'minValue': {'args': '<type __builtin__.int>',
                                            'longname': 'minValue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'min'},
                               'progress': {'args': '<type __builtin__.int>',
                                            'longname': 'progress',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pr'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'status': {'args': '<type __builtin__.str>',
                                          'longname': 'status',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'st'},
                               'step': {'args': '<type __builtin__.int>',
                                        'longname': 'step',
                                        'modes': ['edit'],
                                        'numArgs': 1,
                                        'shortname': 's'},
                               'title': {'args': '<type __builtin__.str>',
                                         'longname': 'title',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 't'}},
                     'shortFlags': {'e': 'edit',
                                    'ep': 'endProgress',
                                    'ic': 'isCancelled',
                                    'ii': 'isInterruptable',
                                    'max': 'maxValue',
                                    'min': 'minValue',
                                    'pr': 'progress',
                                    'q': 'query',
                                    's': 'step',
                                    'st': 'status',
                                    't': 'title'},
                     'type': 'windows'},
  'projectCurve': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'direction': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'direction',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'd'},
                             'directionX': {'args': '<type __builtin__.float>',
                                            'longname': 'directionX',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dx'},
                             'directionY': {'args': '<type __builtin__.float>',
                                            'longname': 'directionY',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dy'},
                             'directionZ': {'args': '<type __builtin__.float>',
                                            'longname': 'directionZ',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dz'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'object': {'args': '<type __builtin__.bool>',
                                        'longname': 'object',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'range': {'args': '<type __builtin__.bool>',
                                       'longname': 'range',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'rn'},
                             'tolerance': {'args': '<type __builtin__.float>',
                                           'longname': 'tolerance',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'tol'},
                             'useNormal': {'args': '<type __builtin__.bool>',
                                           'longname': 'useNormal',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'un'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'd': 'direction',
                                  'dx': 'directionX',
                                  'dy': 'directionY',
                                  'dz': 'directionZ',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'o': 'object',
                                  'q': 'query',
                                  'rn': 'range',
                                  'tol': 'tolerance',
                                  'un': 'useNormal'},
                   'type': 'modeling'},
  'projectTangent': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'curvature': {'args': '<type __builtin__.bool>',
                                             'longname': 'curvature',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'c'},
                               'curvatureScale': {'args': '<type __builtin__.float>',
                                                  'longname': 'curvatureScale',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cs'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'ignoreEdges': {'args': '<type __builtin__.bool>',
                                               'longname': 'ignoreEdges',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ie'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'object': {'args': '<type __builtin__.bool>',
                                          'longname': 'object',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'o'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                   'longname': 'replaceOriginal',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'rpo'},
                               'reverseTangent': {'args': '<type __builtin__.bool>',
                                                  'longname': 'reverseTangent',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'rt'},
                               'rotate': {'args': '<type __builtin__.float>',
                                          'longname': 'rotate',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ro'},
                               'tangentDirection': {'args': '<type __builtin__.int>',
                                                    'longname': 'tangentDirection',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'td'},
                               'tangentScale': {'args': '<type __builtin__.float>',
                                                'longname': 'tangentScale',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ts'}},
                     'shortFlags': {'c': 'curvature',
                                    'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'cs': 'curvatureScale',
                                    'e': 'edit',
                                    'fzn': 'frozen',
                                    'ie': 'ignoreEdges',
                                    'n': 'name',
                                    'nds': 'nodeState',
                                    'o': 'object',
                                    'q': 'query',
                                    'ro': 'rotate',
                                    'rpo': 'replaceOriginal',
                                    'rt': 'reverseTangent',
                                    'td': 'tangentDirection',
                                    'ts': 'tangentScale'},
                     'type': 'modeling'},
  'projectionContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'history': {'args': '<type __builtin__.bool>',
                                              'longname': 'history',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ch'},
                                  'image1': {'args': '<type __builtin__.str>',
                                             'longname': 'image1',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i1'},
                                  'image2': {'args': '<type __builtin__.str>',
                                             'longname': 'image2',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i2'},
                                  'image3': {'args': '<type __builtin__.str>',
                                             'longname': 'image3',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i3'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'}},
                        'shortFlags': {'ch': 'history',
                                       'e': 'edit',
                                       'ex': 'exists',
                                       'i1': 'image1',
                                       'i2': 'image2',
                                       'i3': 'image3',
                                       'n': 'name',
                                       'q': 'query'},
                        'type': 'context'},
  'projectionManip': {'flags': {'fitBBox': {'args': '<type __builtin__.bool>',
                                            'longname': 'fitBBox',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'fb'},
                                'projType': {'args': '<type __builtin__.int>',
                                             'longname': 'projType',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'pt'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'switchType': {'args': '<type __builtin__.bool>',
                                               'longname': 'switchType',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'st'}},
                      'shortFlags': {'fb': 'fitBBox',
                                     'pt': 'projType',
                                     'q': 'query',
                                     'st': 'switchType'},
                      'type': 'rendering'},
  'promptDialog': {'flags': {'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'button': {'args': '<type __builtin__.str>',
                                        'longname': 'button',
                                        'modes': ['create', 'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'b'},
                             'cancelButton': {'args': '<type __builtin__.str>',
                                              'longname': 'cancelButton',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'cb'},
                             'defaultButton': {'args': '<type __builtin__.str>',
                                               'longname': 'defaultButton',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'db'},
                             'dismissString': {'args': '<type __builtin__.str>',
                                               'longname': 'dismissString',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ds'},
                             'message': {'args': '<type __builtin__.str>',
                                         'longname': 'message',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                             'messageAlign': {'args': '<type __builtin__.str>',
                                              'longname': 'messageAlign',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ma'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'scrollableField': {'args': '<type __builtin__.bool>',
                                                 'longname': 'scrollableField',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'sf'},
                             'style': {'args': '<type __builtin__.str>',
                                       'longname': 'style',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'st'},
                             'text': {'args': '<type __builtin__.str>',
                                      'longname': 'text',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'tx'},
                             'title': {'args': '<type __builtin__.str>',
                                       'longname': 'title',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 't'}},
                   'shortFlags': {'b': 'button',
                                  'bgc': 'backgroundColor',
                                  'cb': 'cancelButton',
                                  'db': 'defaultButton',
                                  'ds': 'dismissString',
                                  'm': 'message',
                                  'ma': 'messageAlign',
                                  'p': 'parent',
                                  'q': 'query',
                                  'sf': 'scrollableField',
                                  'st': 'style',
                                  't': 'title',
                                  'tx': 'text'},
                   'type': 'windows'},
  'propModCtx': {'flags': {'animCurve': {'args': '<type __builtin__.str>',
                                         'longname': 'animCurve',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ac'},
                           'animCurveFalloff': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'animCurveFalloff',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'acf'},
                           'animCurveParam': {'args': '<type __builtin__.str>',
                                              'longname': 'animCurveParam',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'acp'},
                           'direction': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'direction',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 3,
                                         'shortname': 'd'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'linear': {'args': '<type __builtin__.float>',
                                      'longname': 'linear',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'l'},
                           'linearParam': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'linearParam',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 2,
                                           'shortname': 'lp'},
                           'nurbsCurve': {'args': '<type __builtin__.str>',
                                          'longname': 'nurbsCurve',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nc'},
                           'powerCutoff': {'args': '<type __builtin__.float>',
                                           'longname': 'powerCutoff',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pc'},
                           'powerCutoffParam': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'powerCutoffParam',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'pcp'},
                           'powerDegree': {'args': '<type __builtin__.float>',
                                           'longname': 'powerDegree',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pd'},
                           'powerDegreeParam': {'args': '<type __builtin__.float>',
                                                'longname': 'powerDegreeParam',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'pdp'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'script': {'args': '<type __builtin__.str>',
                                      'longname': 'script',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                           'scriptParam': {'args': '<type __builtin__.str>',
                                           'longname': 'scriptParam',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sp'},
                           'type': {'args': '<type __builtin__.int>',
                                    'longname': 'type',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 't'},
                           'worldspace': {'args': '<type __builtin__.bool>',
                                          'longname': 'worldspace',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ws'}},
                 'shortFlags': {'ac': 'animCurve',
                                'acf': 'animCurveFalloff',
                                'acp': 'animCurveParam',
                                'd': 'direction',
                                'e': 'edit',
                                'ex': 'exists',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'l': 'linear',
                                'lp': 'linearParam',
                                'nc': 'nurbsCurve',
                                'pc': 'powerCutoff',
                                'pcp': 'powerCutoffParam',
                                'pd': 'powerDegree',
                                'pdp': 'powerDegreeParam',
                                'q': 'query',
                                's': 'script',
                                'sp': 'scriptParam',
                                't': 'type',
                                'ws': 'worldspace'},
                 'type': 'context'},
  'propMove': {'flags': {'percent': {'args': '<type __builtin__.float>',
                                     'longname': 'percent',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                         'percentX': {'args': '<type __builtin__.float>',
                                      'longname': 'percentX',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'px'},
                         'percentY': {'args': '<type __builtin__.float>',
                                      'longname': 'percentY',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'py'},
                         'percentZ': {'args': '<type __builtin__.float>',
                                      'longname': 'percentZ',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'pz'},
                         'pivot': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'pivot',
                                   'modes': ['create'],
                                   'numArgs': 3,
                                   'shortname': 'pi'},
                         'rotate': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'rotate',
                                    'modes': ['create'],
                                    'numArgs': 3,
                                    'shortname': 'r'},
                         'scale': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'scale',
                                   'modes': ['create'],
                                   'numArgs': 3,
                                   'shortname': 's'},
                         'translate': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'translate',
                                       'modes': ['create'],
                                       'numArgs': 3,
                                       'shortname': 't'},
                         'worldSpace': {'args': '<type __builtin__.bool>',
                                        'longname': 'worldSpace',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ws'}},
               'shortFlags': {'p': 'percent',
                              'pi': 'pivot',
                              'px': 'percentX',
                              'py': 'percentY',
                              'pz': 'percentZ',
                              'r': 'rotate',
                              's': 'scale',
                              't': 'translate',
                              'ws': 'worldSpace'},
               'type': 'modeling'},
  'psdChannelOutliner': {'callbackFlags': ['doubleClickCommand',
                                           'dcc',
                                           'selectCommand',
                                           'sc',
                                           'visibleChangeCommand',
                                           'vcc',
                                           'dragCallback',
                                           'dgc',
                                           'dropCallback',
                                           'dpc'],
                         'flags': {'addChild': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'addChild',
                                                'modes': ['edit',
                                                          'multiuse'],
                                                'numArgs': 2,
                                                'shortname': 'ach'},
                                   'allItems': {'args': '<type __builtin__.bool>',
                                                'longname': 'allItems',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'all'},
                                   'annotation': {'args': '<type __builtin__.str>',
                                                  'longname': 'annotation',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ann'},
                                   'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'backgroundColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'bgc'},
                                   'defineTemplate': {'args': '<type __builtin__.str>',
                                                      'longname': 'defineTemplate',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'dt'},
                                   'docTag': {'args': '<type __builtin__.str>',
                                              'longname': 'docTag',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dtg'},
                                   'doubleClickCommand': {'args': '<type __builtin__.str>',
                                                          'longname': 'doubleClickCommand',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'dcc'},
                                   'dragCallback': {'args': 'script',
                                                    'longname': 'dragCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dgc'},
                                   'dropCallback': {'args': 'script',
                                                    'longname': 'dropCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dpc'},
                                   'enable': {'args': '<type __builtin__.bool>',
                                              'longname': 'enable',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'en'},
                                   'enableBackground': {'args': '<type __builtin__.bool>',
                                                        'longname': 'enableBackground',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ebg'},
                                   'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                           'longname': 'enableKeyboardFocus',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ekf'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ex'},
                                   'fullPathName': {'args': '<type __builtin__.bool>',
                                                    'longname': 'fullPathName',
                                                    'modes': ['query'],
                                                    'numArgs': 1,
                                                    'shortname': 'fpn'},
                                   'height': {'args': '<type __builtin__.int>',
                                              'longname': 'height',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'h'},
                                   'highlightColor': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'highlightColor',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'hlc'},
                                   'isObscured': {'args': '<type __builtin__.bool>',
                                                  'longname': 'isObscured',
                                                  'modes': ['query'],
                                                  'numArgs': 1,
                                                  'shortname': 'io'},
                                   'manage': {'args': '<type __builtin__.bool>',
                                              'longname': 'manage',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'm'},
                                   'noBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'noBackground',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'nbg'},
                                   'numberOfItems': {'args': '<type __builtin__.bool>',
                                                     'longname': 'numberOfItems',
                                                     'modes': ['query'],
                                                     'numArgs': 1,
                                                     'shortname': 'ni'},
                                   'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                          'longname': 'numberOfPopupMenus',
                                                          'modes': ['query'],
                                                          'numArgs': 1,
                                                          'shortname': 'npm'},
                                   'parent': {'args': '<type __builtin__.str>',
                                              'longname': 'parent',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'p'},
                                   'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                      'longname': 'popupMenuArray',
                                                      'modes': ['query'],
                                                      'numArgs': 1,
                                                      'shortname': 'pma'},
                                   'preventOverride': {'args': '<type __builtin__.bool>',
                                                       'longname': 'preventOverride',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'po'},
                                   'psdParent': {'args': '<type __builtin__.str>',
                                                 'longname': 'psdParent',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ppa'},
                                   'removeAll': {'args': '<type __builtin__.bool>',
                                                 'longname': 'removeAll',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ra'},
                                   'removeChild': {'args': '<type __builtin__.str>',
                                                   'longname': 'removeChild',
                                                   'modes': ['edit',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'rc'},
                                   'select': {'args': '<type __builtin__.str>',
                                              'longname': 'select',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'sel'},
                                   'selectCommand': {'args': '<type __builtin__.str>',
                                                     'longname': 'selectCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sc'},
                                   'selectItem': {'args': '<type __builtin__.bool>',
                                                  'longname': 'selectItem',
                                                  'modes': ['query'],
                                                  'numArgs': 1,
                                                  'shortname': 'si'},
                                   'statusBarMessage': {'args': '<type __builtin__.str>',
                                                        'longname': 'statusBarMessage',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'sbm'},
                                   'useTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'useTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ut'},
                                   'visible': {'args': '<type __builtin__.bool>',
                                               'longname': 'visible',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'vis'},
                                   'visibleChangeCommand': {'args': 'script',
                                                            'longname': 'visibleChangeCommand',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'vcc'},
                                   'width': {'args': '<type __builtin__.int>',
                                             'longname': 'width',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'w'}},
                         'shortFlags': {},
                         'type': 'rendering'},
  'psdConvSolidTxOptions': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'psdEditTextureFile': {'flags': {'addChannel': {'args': '<type __builtin__.str>',
                                                  'longname': 'addChannel',
                                                  'modes': ['create',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'adc'},
                                   'addChannelColor': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'addChannelColor',
                                                       'modes': ['create',
                                                                 'multiuse'],
                                                       'numArgs': 4,
                                                       'shortname': 'acc'},
                                   'addChannelImage': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.str>'],
                                                       'longname': 'addChannelImage',
                                                       'modes': ['create',
                                                                 'multiuse'],
                                                       'numArgs': 2,
                                                       'shortname': 'aci'},
                                   'deleteChannel': {'args': '<type __builtin__.str>',
                                                     'longname': 'deleteChannel',
                                                     'modes': ['create',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'deleteChannel'},
                                   'psdFileName': {'args': '<type __builtin__.str>',
                                                   'longname': 'psdFileName',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'psf'},
                                   'snapShotImage': {'args': '<type __builtin__.str>',
                                                     'longname': 'snapShotImage',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ssi'},
                                   'uvSnapPostionTop': {'args': '<type __builtin__.bool>',
                                                        'longname': 'uvSnapPostionTop',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'uvt'}},
                         'shortFlags': {},
                         'type': 'rendering'},
  'psdExport': {'flags': {'alphaChannelIdx': {'args': '<type __builtin__.int>',
                                              'longname': 'alphaChannelIdx',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'aci'},
                          'bytesPerChannel': {'args': '<type __builtin__.int>',
                                              'longname': 'bytesPerChannel',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'bpc'},
                          'emptyLayerSet': {'args': '<type __builtin__.bool>',
                                            'longname': 'emptyLayerSet',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'els'},
                          'format': {'args': '<type __builtin__.str>',
                                     'longname': 'format',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'format'},
                          'layerName': {'args': '<type __builtin__.str>',
                                        'longname': 'layerName',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'lyn'},
                          'layerSetName': {'args': '<type __builtin__.str>',
                                           'longname': 'layerSetName',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'lsn'},
                          'outFileName': {'args': '<type __builtin__.str>',
                                          'longname': 'outFileName',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'ofn'},
                          'preMultiplyAlpha': {'args': '<type __builtin__.bool>',
                                               'longname': 'preMultiplyAlpha',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'pma'},
                          'psdFileName': {'args': '<type __builtin__.str>',
                                          'longname': 'psdFileName',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'ifn'}},
                'shortFlags': {},
                'type': 'rendering'},
  'psdTextureFile': {'flags': {'channelRGB': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'channelRGB',
                                              'modes': ['create',
                                                        'multiuse'],
                                              'numArgs': 5,
                                              'shortname': 'chc'},
                               'channels': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.bool>'],
                                            'longname': 'channels',
                                            'modes': ['create',
                                                      'multiuse'],
                                            'numArgs': 3,
                                            'shortname': 'chs'},
                               'imageFileName': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'imageFileName',
                                                 'modes': ['create',
                                                           'multiuse'],
                                                 'numArgs': 3,
                                                 'shortname': 'ifn'},
                               'psdFileName': {'args': '<type __builtin__.str>',
                                               'longname': 'psdFileName',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'psf'},
                               'snapShotImageName': {'args': '<type __builtin__.str>',
                                                     'longname': 'snapShotImageName',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ssi'},
                               'uvSnapPostionTop': {'args': '<type __builtin__.bool>',
                                                    'longname': 'uvSnapPostionTop',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'uvt'},
                               'xResolution': {'args': '<type __builtin__.int>',
                                               'longname': 'xResolution',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'xr'},
                               'yResolution': {'args': '<type __builtin__.int>',
                                               'longname': 'yResolution',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'yr'}},
                     'shortFlags': {},
                     'type': 'rendering'},
  'python': {'flags': {}, 'shortFlags': {}, 'type': 'language'},
  'querySubdiv': {'flags': {'action': {'args': '<type __builtin__.int>',
                                       'longname': 'action',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'a'},
                            'level': {'args': '<type __builtin__.int>',
                                      'longname': 'level',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'l'},
                            'relative': {'args': '<type __builtin__.bool>',
                                         'longname': 'relative',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'r'}},
                  'shortFlags': {'a': 'action',
                                 'l': 'level',
                                 'r': 'relative'},
                  'type': 'modeling'},
  'quit': {'flags': {'abort': {'args': '<type __builtin__.bool>',
                               'longname': 'abort',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'a'},
                     'exitCode': {'args': '<type __builtin__.int>',
                                  'longname': 'exitCode',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'ec'},
                     'force': {'args': '<type __builtin__.bool>',
                               'longname': 'force',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'f'}},
           'shortFlags': {'a': 'abort', 'ec': 'exitCode', 'f': 'force'},
           'type': None},
  'radial': {'flags': {'attenuation': {'args': '<type __builtin__.float>',
                                       'longname': 'attenuation',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'att'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'magnitude': {'args': '<type __builtin__.float>',
                                     'longname': 'magnitude',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'm'},
                       'maxDistance': {'args': '<type __builtin__.float>',
                                       'longname': 'maxDistance',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'mxd'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'perVertex': {'args': '<type __builtin__.bool>',
                                     'longname': 'perVertex',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'pv'},
                       'position': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'position',
                                    'modes': ['query', 'edit', 'multiuse'],
                                    'numArgs': 3,
                                    'shortname': 'pos'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'torusSectionRadius': {'args': '<type __builtin__.float>',
                                              'longname': 'torusSectionRadius',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'tsr'},
                       'type': {'args': '<type __builtin__.float>',
                                'longname': 'type',
                                'modes': ['query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'typ'},
                       'volumeExclusion': {'args': '<type __builtin__.bool>',
                                           'longname': 'volumeExclusion',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vex'},
                       'volumeOffset': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'volumeOffset',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 3,
                                        'shortname': 'vof'},
                       'volumeShape': {'args': '<type __builtin__.str>',
                                       'longname': 'volumeShape',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vsh'},
                       'volumeSweep': {'args': '<type __builtin__.float>',
                                       'longname': 'volumeSweep',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vsw'}},
             'shortFlags': {'att': 'attenuation',
                            'e': 'edit',
                            'm': 'magnitude',
                            'mxd': 'maxDistance',
                            'n': 'name',
                            'pos': 'position',
                            'pv': 'perVertex',
                            'q': 'query',
                            'tsr': 'torusSectionRadius',
                            'typ': 'type',
                            'vex': 'volumeExclusion',
                            'vof': 'volumeOffset',
                            'vsh': 'volumeShape',
                            'vsw': 'volumeSweep'},
             'type': 'effects'},
  'radioButton': {'callbackFlags': ['dropCallback',
                                    'dpc',
                                    'visibleChangeCommand',
                                    'vcc',
                                    'onCommand',
                                    'onc',
                                    'dragCallback',
                                    'dgc',
                                    'offCommand',
                                    'ofc',
                                    'changeCommand',
                                    'cc'],
                  'flags': {'align': {'args': '<type __builtin__.str>',
                                      'longname': 'align',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'al'},
                            'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'changeCommand': {'args': 'script',
                                              'longname': 'changeCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cc'},
                            'collection': {'args': '<type __builtin__.str>',
                                           'longname': 'collection',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'cl'},
                            'data': {'args': '<type __builtin__.int>',
                                     'longname': 'data',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'da'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'editable': {'args': '<type __builtin__.bool>',
                                         'longname': 'editable',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ed'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'label': {'args': '<type __builtin__.str>',
                                      'longname': 'label',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'l'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'offCommand': {'args': 'script',
                                           'longname': 'offCommand',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ofc'},
                            'onCommand': {'args': 'script',
                                          'longname': 'onCommand',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'onc'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'recomputeSize': {'args': '<type __builtin__.bool>',
                                              'longname': 'recomputeSize',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'rs'},
                            'select': {'args': '<type __builtin__.bool>',
                                       'longname': 'select',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'sl'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'al': 'align',
                                 'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'cc': 'changeCommand',
                                 'cl': 'collection',
                                 'da': 'data',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ed': 'editable',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'ex': 'exists',
                                 'fpn': 'fullPathName',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'io': 'isObscured',
                                 'l': 'label',
                                 'm': 'manage',
                                 'nbg': 'noBackground',
                                 'npm': 'numberOfPopupMenus',
                                 'ofc': 'offCommand',
                                 'onc': 'onCommand',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'rs': 'recomputeSize',
                                 'sbm': 'statusBarMessage',
                                 'sl': 'select',
                                 'ut': 'useTemplate',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'radioButtonGrp': {'callbackFlags': ['changeCommand4',
                                       'cc4',
                                       'dragCallback',
                                       'dgc',
                                       'changeCommand3',
                                       'cc3',
                                       'changeCommand2',
                                       'cc2',
                                       'changeCommand1',
                                       'cc1',
                                       'offCommand',
                                       'ofc',
                                       'visibleChangeCommand',
                                       'vcc',
                                       'offCommand1',
                                       'of1',
                                       'offCommand3',
                                       'of3',
                                       'offCommand2',
                                       'of2',
                                       'offCommand4',
                                       'of4',
                                       'dropCallback',
                                       'dpc',
                                       'onCommand',
                                       'onc',
                                       'changeCommand',
                                       'cc',
                                       'onCommand4',
                                       'on4',
                                       'onCommand1',
                                       'on1',
                                       'onCommand2',
                                       'on2',
                                       'onCommand3',
                                       'on3'],
                     'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                    'longname': 'adjustableColumn',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'adj'},
                               'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn2',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad2'},
                               'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn3',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad3'},
                               'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn4',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad4'},
                               'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn5',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad5'},
                               'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn6',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ad6'},
                               'annotation': {'args': '<type __builtin__.str>',
                                              'longname': 'annotation',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ann'},
                               'annotation1': {'args': '<type __builtin__.str>',
                                               'longname': 'annotation1',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'an1'},
                               'annotation2': {'args': '<type __builtin__.str>',
                                               'longname': 'annotation2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'an2'},
                               'annotation3': {'args': '<type __builtin__.str>',
                                               'longname': 'annotation3',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'an3'},
                               'annotation4': {'args': '<type __builtin__.str>',
                                               'longname': 'annotation4',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'an4'},
                               'backgroundColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'backgroundColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'bgc'},
                               'changeCommand': {'args': 'script',
                                                 'longname': 'changeCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cc'},
                               'changeCommand1': {'args': 'script',
                                                  'longname': 'changeCommand1',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cc1'},
                               'changeCommand2': {'args': 'script',
                                                  'longname': 'changeCommand2',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cc2'},
                               'changeCommand3': {'args': 'script',
                                                  'longname': 'changeCommand3',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cc3'},
                               'changeCommand4': {'args': 'script',
                                                  'longname': 'changeCommand4',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cc4'},
                               'columnAlign': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAlign',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 2,
                                               'shortname': 'cal'},
                               'columnAlign2': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign2',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'cl2'},
                               'columnAlign3': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign3',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'cl3'},
                               'columnAlign4': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign4',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 4,
                                                'shortname': 'cl4'},
                               'columnAlign5': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign5',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 5,
                                                'shortname': 'cl5'},
                               'columnAlign6': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign6',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 6,
                                                'shortname': 'cl6'},
                               'columnAttach': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnAttach',
                                                'modes': ['create',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 3,
                                                'shortname': 'cat'},
                               'columnAttach2': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach2',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 2,
                                                 'shortname': 'ct2'},
                               'columnAttach3': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach3',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'ct3'},
                               'columnAttach4': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach4',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 4,
                                                 'shortname': 'ct4'},
                               'columnAttach5': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach5',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 5,
                                                 'shortname': 'ct5'},
                               'columnAttach6': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'columnAttach6',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 6,
                                                 'shortname': 'ct6'},
                               'columnOffset2': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset2',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 2,
                                                 'shortname': 'co2'},
                               'columnOffset3': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset3',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'co3'},
                               'columnOffset4': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset4',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 4,
                                                 'shortname': 'co4'},
                               'columnOffset5': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset5',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 5,
                                                 'shortname': 'co5'},
                               'columnOffset6': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset6',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 6,
                                                 'shortname': 'co6'},
                               'columnWidth': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnWidth',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 2,
                                               'shortname': 'cw'},
                               'columnWidth1': {'args': '<type __builtin__.int>',
                                                'longname': 'columnWidth1',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cw1'},
                               'columnWidth2': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth2',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'cw2'},
                               'columnWidth3': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth3',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'cw3'},
                               'columnWidth4': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth4',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 4,
                                                'shortname': 'cw4'},
                               'columnWidth5': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth5',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 5,
                                                'shortname': 'cw5'},
                               'columnWidth6': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth6',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 6,
                                                'shortname': 'cw6'},
                               'data1': {'args': '<type __builtin__.int>',
                                         'longname': 'data1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'da1'},
                               'data2': {'args': '<type __builtin__.int>',
                                         'longname': 'data2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'da2'},
                               'data3': {'args': '<type __builtin__.int>',
                                         'longname': 'data3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'da3'},
                               'data4': {'args': '<type __builtin__.int>',
                                         'longname': 'data4',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'da4'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'dragCallback': {'args': 'script',
                                                'longname': 'dragCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dgc'},
                               'dropCallback': {'args': 'script',
                                                'longname': 'dropCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dpc'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'editable': {'args': '<type __builtin__.bool>',
                                            'longname': 'editable',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ed'},
                               'enable': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en'},
                               'enable1': {'args': '<type __builtin__.bool>',
                                           'longname': 'enable1',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'en1'},
                               'enable2': {'args': '<type __builtin__.bool>',
                                           'longname': 'enable2',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'en2'},
                               'enable3': {'args': '<type __builtin__.bool>',
                                           'longname': 'enable3',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'en3'},
                               'enable4': {'args': '<type __builtin__.bool>',
                                           'longname': 'enable4',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'en4'},
                               'enableBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableBackground',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ebg'},
                               'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                       'longname': 'enableKeyboardFocus',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ekf'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'fullPathName': {'args': '<type __builtin__.bool>',
                                                'longname': 'fullPathName',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'fpn'},
                               'height': {'args': '<type __builtin__.int>',
                                          'longname': 'height',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'h'},
                               'highlightColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'highlightColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'hlc'},
                               'isObscured': {'args': '<type __builtin__.bool>',
                                              'longname': 'isObscured',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'io'},
                               'label': {'args': '<type __builtin__.str>',
                                         'longname': 'label',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'l'},
                               'label1': {'args': '<type __builtin__.str>',
                                          'longname': 'label1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'l1'},
                               'label2': {'args': '<type __builtin__.str>',
                                          'longname': 'label2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'l2'},
                               'label3': {'args': '<type __builtin__.str>',
                                          'longname': 'label3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'l3'},
                               'label4': {'args': '<type __builtin__.str>',
                                          'longname': 'label4',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'l4'},
                               'labelAnnotation': {'args': '<type __builtin__.str>',
                                                   'longname': 'labelAnnotation',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'la'},
                               'labelArray2': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'labelArray2',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 2,
                                               'shortname': 'la2'},
                               'labelArray3': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'labelArray3',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'la3'},
                               'labelArray4': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'labelArray4',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 4,
                                               'shortname': 'la4'},
                               'manage': {'args': '<type __builtin__.bool>',
                                          'longname': 'manage',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'm'},
                               'noBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'noBackground',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nbg'},
                               'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'numberOfPopupMenus',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'npm'},
                               'numberOfRadioButtons': {'args': '<type __builtin__.int>',
                                                        'longname': 'numberOfRadioButtons',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'nrb'},
                               'offCommand': {'args': 'script',
                                              'longname': 'offCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ofc'},
                               'offCommand1': {'args': 'script',
                                               'longname': 'offCommand1',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'of1'},
                               'offCommand2': {'args': 'script',
                                               'longname': 'offCommand2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'of2'},
                               'offCommand3': {'args': 'script',
                                               'longname': 'offCommand3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'of3'},
                               'offCommand4': {'args': 'script',
                                               'longname': 'offCommand4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'of4'},
                               'onCommand': {'args': 'script',
                                             'longname': 'onCommand',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'onc'},
                               'onCommand1': {'args': 'script',
                                              'longname': 'onCommand1',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'on1'},
                               'onCommand2': {'args': 'script',
                                              'longname': 'onCommand2',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'on2'},
                               'onCommand3': {'args': 'script',
                                              'longname': 'onCommand3',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'on3'},
                               'onCommand4': {'args': 'script',
                                              'longname': 'onCommand4',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'on4'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                  'longname': 'popupMenuArray',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'pma'},
                               'preventOverride': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preventOverride',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'po'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'rowAttach': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.str>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'rowAttach',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 3,
                                             'shortname': 'rat'},
                               'select': {'args': '<type __builtin__.int>',
                                          'longname': 'select',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sl'},
                               'shareCollection': {'args': '<type __builtin__.str>',
                                                   'longname': 'shareCollection',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'scl'},
                               'statusBarMessage': {'args': '<type __builtin__.str>',
                                                    'longname': 'statusBarMessage',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sbm'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'},
                               'vertical': {'args': '<type __builtin__.bool>',
                                            'longname': 'vertical',
                                            'modes': ['create', 'query'],
                                            'numArgs': 0,
                                            'shortname': 'vr'},
                               'visible': {'args': '<type __builtin__.bool>',
                                           'longname': 'visible',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vis'},
                               'visibleChangeCommand': {'args': 'script',
                                                        'longname': 'visibleChangeCommand',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'vcc'},
                               'width': {'args': '<type __builtin__.int>',
                                         'longname': 'width',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'w'}},
                     'shortFlags': {'ad2': 'adjustableColumn2',
                                    'ad3': 'adjustableColumn3',
                                    'ad4': 'adjustableColumn4',
                                    'ad5': 'adjustableColumn5',
                                    'ad6': 'adjustableColumn6',
                                    'adj': 'adjustableColumn',
                                    'an1': 'annotation1',
                                    'an2': 'annotation2',
                                    'an3': 'annotation3',
                                    'an4': 'annotation4',
                                    'ann': 'annotation',
                                    'bgc': 'backgroundColor',
                                    'cal': 'columnAlign',
                                    'cat': 'columnAttach',
                                    'cc': 'changeCommand',
                                    'cc1': 'changeCommand1',
                                    'cc2': 'changeCommand2',
                                    'cc3': 'changeCommand3',
                                    'cc4': 'changeCommand4',
                                    'cl2': 'columnAlign2',
                                    'cl3': 'columnAlign3',
                                    'cl4': 'columnAlign4',
                                    'cl5': 'columnAlign5',
                                    'cl6': 'columnAlign6',
                                    'co2': 'columnOffset2',
                                    'co3': 'columnOffset3',
                                    'co4': 'columnOffset4',
                                    'co5': 'columnOffset5',
                                    'co6': 'columnOffset6',
                                    'ct2': 'columnAttach2',
                                    'ct3': 'columnAttach3',
                                    'ct4': 'columnAttach4',
                                    'ct5': 'columnAttach5',
                                    'ct6': 'columnAttach6',
                                    'cw': 'columnWidth',
                                    'cw1': 'columnWidth1',
                                    'cw2': 'columnWidth2',
                                    'cw3': 'columnWidth3',
                                    'cw4': 'columnWidth4',
                                    'cw5': 'columnWidth5',
                                    'cw6': 'columnWidth6',
                                    'da1': 'data1',
                                    'da2': 'data2',
                                    'da3': 'data3',
                                    'da4': 'data4',
                                    'dgc': 'dragCallback',
                                    'dpc': 'dropCallback',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'e': 'edit',
                                    'ebg': 'enableBackground',
                                    'ed': 'editable',
                                    'ekf': 'enableKeyboardFocus',
                                    'en': 'enable',
                                    'en1': 'enable1',
                                    'en2': 'enable2',
                                    'en3': 'enable3',
                                    'en4': 'enable4',
                                    'ex': 'exists',
                                    'fpn': 'fullPathName',
                                    'h': 'height',
                                    'hlc': 'highlightColor',
                                    'io': 'isObscured',
                                    'l': 'label',
                                    'l1': 'label1',
                                    'l2': 'label2',
                                    'l3': 'label3',
                                    'l4': 'label4',
                                    'la': 'labelAnnotation',
                                    'la2': 'labelArray2',
                                    'la3': 'labelArray3',
                                    'la4': 'labelArray4',
                                    'm': 'manage',
                                    'nbg': 'noBackground',
                                    'npm': 'numberOfPopupMenus',
                                    'nrb': 'numberOfRadioButtons',
                                    'of1': 'offCommand1',
                                    'of2': 'offCommand2',
                                    'of3': 'offCommand3',
                                    'of4': 'offCommand4',
                                    'ofc': 'offCommand',
                                    'on1': 'onCommand1',
                                    'on2': 'onCommand2',
                                    'on3': 'onCommand3',
                                    'on4': 'onCommand4',
                                    'onc': 'onCommand',
                                    'p': 'parent',
                                    'pma': 'popupMenuArray',
                                    'po': 'preventOverride',
                                    'q': 'query',
                                    'rat': 'rowAttach',
                                    'sbm': 'statusBarMessage',
                                    'scl': 'shareCollection',
                                    'sl': 'select',
                                    'ut': 'useTemplate',
                                    'vcc': 'visibleChangeCommand',
                                    'vis': 'visible',
                                    'vr': 'vertical',
                                    'w': 'width'},
                     'type': 'windows'},
  'radioCollection': {'flags': {'collectionItemArray': {'args': '<type __builtin__.bool>',
                                                        'longname': 'collectionItemArray',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'cia'},
                                'defineTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'defineTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'dt'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'gl': {'args': '<type __builtin__.bool>',
                                       'longname': 'gl',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'gl'},
                                'numberOfCollectionItems': {'args': '<type __builtin__.bool>',
                                                            'longname': 'numberOfCollectionItems',
                                                            'modes': ['query'],
                                                            'numArgs': 0,
                                                            'shortname': 'nci'},
                                'parent': {'args': '<type __builtin__.str>',
                                           'longname': 'parent',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'p'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'select': {'args': '<type __builtin__.str>',
                                           'longname': 'select',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sl'},
                                'useTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'useTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ut'}},
                      'removedFlags': {'global': 'gl'},
                      'shortFlags': {'cia': 'collectionItemArray',
                                     'dt': 'defineTemplate',
                                     'e': 'edit',
                                     'ex': 'exists',
                                     'gl': 'gl',
                                     'nci': 'numberOfCollectionItems',
                                     'p': 'parent',
                                     'q': 'query',
                                     'sl': 'select',
                                     'ut': 'useTemplate'},
                      'type': 'windows'},
  'radioMenuItemCollection': {'flags': {'defineTemplate': {'args': '<type __builtin__.str>',
                                                           'longname': 'defineTemplate',
                                                           'modes': ['create'],
                                                           'numArgs': 1,
                                                           'shortname': 'dt'},
                                        'edit': {'args': '<type __builtin__.bool>',
                                                 'longname': 'edit',
                                                 'numArgs': 0,
                                                 'shortname': 'e'},
                                        'exists': {'args': '<type __builtin__.bool>',
                                                   'longname': 'exists',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'ex'},
                                        'gl': {'args': '<type __builtin__.bool>',
                                               'longname': 'gl',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'gl'},
                                        'parent': {'args': '<type __builtin__.str>',
                                                   'longname': 'parent',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'p'},
                                        'query': {'args': '<type __builtin__.bool>',
                                                  'longname': 'query',
                                                  'numArgs': 0,
                                                  'shortname': 'q'},
                                        'useTemplate': {'args': '<type __builtin__.str>',
                                                        'longname': 'useTemplate',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'ut'}},
                              'removedFlags': {'global': 'gl'},
                              'shortFlags': {'dt': 'defineTemplate',
                                             'e': 'edit',
                                             'ex': 'exists',
                                             'gl': 'gl',
                                             'p': 'parent',
                                             'q': 'query',
                                             'ut': 'useTemplate'},
                              'type': 'windows'},
  'rampColorPort': {'callbackFlags': ['visibleChangeCommand',
                                      'vcc',
                                      'dragCallback',
                                      'dgc',
                                      'dropCallback',
                                      'dpc'],
                    'flags': {'annotation': {'args': '<type __builtin__.str>',
                                             'longname': 'annotation',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ann'},
                              'backgroundColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'backgroundColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'bgc'},
                              'defineTemplate': {'args': '<type __builtin__.str>',
                                                 'longname': 'defineTemplate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'dt'},
                              'docTag': {'args': '<type __builtin__.str>',
                                         'longname': 'docTag',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dtg'},
                              'dragCallback': {'args': 'script',
                                               'longname': 'dragCallback',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dgc'},
                              'dropCallback': {'args': 'script',
                                               'longname': 'dropCallback',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dpc'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'enable': {'args': '<type __builtin__.bool>',
                                         'longname': 'enable',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'en'},
                              'enableBackground': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableBackground',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ebg'},
                              'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'enableKeyboardFocus',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ekf'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'fullPathName': {'args': '<type __builtin__.bool>',
                                               'longname': 'fullPathName',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'fpn'},
                              'height': {'args': '<type __builtin__.int>',
                                         'longname': 'height',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'h'},
                              'highlightColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'highlightColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'hlc'},
                              'isObscured': {'args': '<type __builtin__.bool>',
                                             'longname': 'isObscured',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'io'},
                              'manage': {'args': '<type __builtin__.bool>',
                                         'longname': 'manage',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                              'noBackground': {'args': '<type __builtin__.bool>',
                                               'longname': 'noBackground',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nbg'},
                              'node': {'args': 'PyNode',
                                       'longname': 'node',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'numberOfPopupMenus',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'npm'},
                              'parent': {'args': '<type __builtin__.str>',
                                         'longname': 'parent',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'p'},
                              'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                 'longname': 'popupMenuArray',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'pma'},
                              'preventOverride': {'args': '<type __builtin__.bool>',
                                                  'longname': 'preventOverride',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'po'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'selectedColorControl': {'args': '<type __builtin__.str>',
                                                       'longname': 'selectedColorControl',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'sc'},
                              'selectedInterpControl': {'args': '<type __builtin__.str>',
                                                        'longname': 'selectedInterpControl',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'si'},
                              'selectedPositionControl': {'args': '<type __builtin__.str>',
                                                          'longname': 'selectedPositionControl',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'sp'},
                              'statusBarMessage': {'args': '<type __builtin__.str>',
                                                   'longname': 'statusBarMessage',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sbm'},
                              'useTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'useTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ut'},
                              'verticalLayout': {'args': '<type __builtin__.bool>',
                                                 'longname': 'verticalLayout',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'vl'},
                              'visible': {'args': '<type __builtin__.bool>',
                                          'longname': 'visible',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'vis'},
                              'visibleChangeCommand': {'args': 'script',
                                                       'longname': 'visibleChangeCommand',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'vcc'},
                              'width': {'args': '<type __builtin__.int>',
                                        'longname': 'width',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'w'}},
                    'shortFlags': {'ann': 'annotation',
                                   'bgc': 'backgroundColor',
                                   'dgc': 'dragCallback',
                                   'dpc': 'dropCallback',
                                   'dt': 'defineTemplate',
                                   'dtg': 'docTag',
                                   'e': 'edit',
                                   'ebg': 'enableBackground',
                                   'ekf': 'enableKeyboardFocus',
                                   'en': 'enable',
                                   'ex': 'exists',
                                   'fpn': 'fullPathName',
                                   'h': 'height',
                                   'hlc': 'highlightColor',
                                   'io': 'isObscured',
                                   'm': 'manage',
                                   'n': 'node',
                                   'nbg': 'noBackground',
                                   'npm': 'numberOfPopupMenus',
                                   'p': 'parent',
                                   'pma': 'popupMenuArray',
                                   'po': 'preventOverride',
                                   'q': 'query',
                                   'sbm': 'statusBarMessage',
                                   'sc': 'selectedColorControl',
                                   'si': 'selectedInterpControl',
                                   'sp': 'selectedPositionControl',
                                   'ut': 'useTemplate',
                                   'vcc': 'visibleChangeCommand',
                                   'vis': 'visible',
                                   'vl': 'verticalLayout',
                                   'w': 'width'},
                    'type': 'rendering'},
  'rampWidget': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'rampWidgetAttrless': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'rangeControl': {'callbackFlags': ['visibleChangeCommand',
                                     'vcc',
                                     'dragCallback',
                                     'dgc',
                                     'changedCommand',
                                     'cc',
                                     'dropCallback',
                                     'dpc'],
                   'flags': {'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'changedCommand': {'args': 'script',
                                                'longname': 'changedCommand',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cc'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'maxRange': {'args': 'time',
                                          'longname': 'maxRange',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'max'},
                             'minRange': {'args': 'time',
                                          'longname': 'minRange',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'min'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'},
                             'widthHeight': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'widthHeight',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 2,
                                             'shortname': 'wh'}},
                   'shortFlags': {'ann': 'annotation',
                                  'bgc': 'backgroundColor',
                                  'cc': 'changedCommand',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ebg': 'enableBackground',
                                  'ekf': 'enableKeyboardFocus',
                                  'en': 'enable',
                                  'ex': 'exists',
                                  'fpn': 'fullPathName',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'io': 'isObscured',
                                  'm': 'manage',
                                  'max': 'maxRange',
                                  'min': 'minRange',
                                  'nbg': 'noBackground',
                                  'npm': 'numberOfPopupMenus',
                                  'p': 'parent',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'q': 'query',
                                  'sbm': 'statusBarMessage',
                                  'ut': 'useTemplate',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'w': 'width',
                                  'wh': 'widthHeight'},
                   'type': 'windows'},
  'readPDC': {'flags': {'file': {'args': '<type __builtin__.str>',
                                 'longname': 'file',
                                 'numArgs': 1,
                                 'shortname': 'f'},
                        'test': {'args': '<type __builtin__.bool>',
                                 'longname': 'test',
                                 'numArgs': 0,
                                 'shortname': 't'}},
              'shortFlags': {'f': 'file', 't': 'test'},
              'type': 'other'},
  'readTake': {'flags': {'angle': {'args': '<type __builtin__.str>',
                                   'longname': 'angle',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'a'},
                         'device': {'args': '<type __builtin__.str>',
                                    'longname': 'device',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'd'},
                         'frequency': {'args': '<type __builtin__.float>',
                                       'longname': 'frequency',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'f'},
                         'linear': {'args': '<type __builtin__.str>',
                                    'longname': 'linear',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'l'},
                         'noTime': {'args': '<type __builtin__.bool>',
                                    'longname': 'noTime',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'nt'},
                         'take': {'args': '<type __builtin__.str>',
                                  'longname': 'take',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 't'}},
               'shortFlags': {},
               'type': 'animation'},
  'rebuildCurve': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'degree': {'args': '<type __builtin__.int>',
                                        'longname': 'degree',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'd'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'endKnots': {'args': '<type __builtin__.int>',
                                          'longname': 'endKnots',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'end'},
                             'fitRebuild': {'args': '<type __builtin__.bool>',
                                            'longname': 'fitRebuild',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'fr'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'keepControlPoints': {'args': '<type __builtin__.bool>',
                                                   'longname': 'keepControlPoints',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'kcp'},
                             'keepEndPoints': {'args': '<type __builtin__.bool>',
                                               'longname': 'keepEndPoints',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'kep'},
                             'keepRange': {'args': '<type __builtin__.int>',
                                           'longname': 'keepRange',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'kr'},
                             'keepTangents': {'args': '<type __builtin__.bool>',
                                              'longname': 'keepTangents',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'kt'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'noChanges': {'args': '<type __builtin__.bool>',
                                           'longname': 'noChanges',
                                           'numArgs': 1,
                                           'shortname': 'nc'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'object': {'args': '<type __builtin__.bool>',
                                        'longname': 'object',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'range': {'args': '<type __builtin__.bool>',
                                       'longname': 'range',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'rn'},
                             'rebuildType': {'args': '<type __builtin__.int>',
                                             'longname': 'rebuildType',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'rt'},
                             'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                 'longname': 'replaceOriginal',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'rpo'},
                             'smartSurfaceCurveRebuild': {'args': '<type __builtin__.bool>',
                                                          'longname': 'smartSurfaceCurveRebuild',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'scr'},
                             'smooth': {'args': '<type __builtin__.float>',
                                        'longname': 'smooth',
                                        'numArgs': 1,
                                        'shortname': 'sm'},
                             'spans': {'args': '<type __builtin__.int>',
                                       'longname': 'spans',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 's'},
                             'tolerance': {'args': '<type __builtin__.float>',
                                           'longname': 'tolerance',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'tol'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'd': 'degree',
                                  'e': 'edit',
                                  'end': 'endKnots',
                                  'fr': 'fitRebuild',
                                  'fzn': 'frozen',
                                  'kcp': 'keepControlPoints',
                                  'kep': 'keepEndPoints',
                                  'kr': 'keepRange',
                                  'kt': 'keepTangents',
                                  'n': 'name',
                                  'nc': 'noChanges',
                                  'nds': 'nodeState',
                                  'o': 'object',
                                  'q': 'query',
                                  'rn': 'range',
                                  'rpo': 'replaceOriginal',
                                  'rt': 'rebuildType',
                                  's': 'spans',
                                  'scr': 'smartSurfaceCurveRebuild',
                                  'sm': 'smooth',
                                  'tol': 'tolerance'},
                   'type': 'modeling'},
  'rebuildSurface': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'degreeU': {'args': '<type __builtin__.int>',
                                           'longname': 'degreeU',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'du'},
                               'degreeV': {'args': '<type __builtin__.int>',
                                           'longname': 'degreeV',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dv'},
                               'direction': {'args': '<type __builtin__.int>',
                                             'longname': 'direction',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dir'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'endKnots': {'args': '<type __builtin__.int>',
                                            'longname': 'endKnots',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'end'},
                               'fitRebuild': {'args': '<type __builtin__.int>',
                                              'longname': 'fitRebuild',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fr'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'keepControlPoints': {'args': '<type __builtin__.bool>',
                                                     'longname': 'keepControlPoints',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'kcp'},
                               'keepCorners': {'args': '<type __builtin__.bool>',
                                               'longname': 'keepCorners',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'kc'},
                               'keepRange': {'args': '<type __builtin__.int>',
                                             'longname': 'keepRange',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'kr'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'noChanges': {'args': '<type __builtin__.bool>',
                                             'longname': 'noChanges',
                                             'numArgs': 1,
                                             'shortname': 'nc'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'object': {'args': '<type __builtin__.bool>',
                                          'longname': 'object',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'o'},
                               'polygon': {'args': '<type __builtin__.int>',
                                           'longname': 'polygon',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'po'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'rebuildType': {'args': '<type __builtin__.int>',
                                               'longname': 'rebuildType',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rt'},
                               'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                   'longname': 'replaceOriginal',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'rpo'},
                               'spansU': {'args': '<type __builtin__.int>',
                                          'longname': 'spansU',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'su'},
                               'spansV': {'args': '<type __builtin__.int>',
                                          'longname': 'spansV',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sv'},
                               'tolerance': {'args': '<type __builtin__.float>',
                                             'longname': 'tolerance',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'tol'}},
                     'shortFlags': {'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'dir': 'direction',
                                    'du': 'degreeU',
                                    'dv': 'degreeV',
                                    'e': 'edit',
                                    'end': 'endKnots',
                                    'fr': 'fitRebuild',
                                    'fzn': 'frozen',
                                    'kc': 'keepCorners',
                                    'kcp': 'keepControlPoints',
                                    'kr': 'keepRange',
                                    'n': 'name',
                                    'nc': 'noChanges',
                                    'nds': 'nodeState',
                                    'o': 'object',
                                    'po': 'polygon',
                                    'q': 'query',
                                    'rpo': 'replaceOriginal',
                                    'rt': 'rebuildType',
                                    'su': 'spansU',
                                    'sv': 'spansV',
                                    'tol': 'tolerance'},
                     'type': 'modeling'},
  'recordAttr': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                         'longname': 'attribute',
                                         'modes': ['create', 'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'at'},
                           'delete': {'args': '<type __builtin__.bool>',
                                      'longname': 'delete',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'd'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'at': 'attribute',
                                'd': 'delete',
                                'e': 'edit',
                                'q': 'query'},
                 'type': 'system'},
  'recordDevice': {'flags': {'cleanup': {'args': '<type __builtin__.bool>',
                                         'longname': 'cleanup',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'c'},
                             'data': {'args': '<type __builtin__.bool>',
                                      'longname': 'data',
                                      'modes': ['query'],
                                      'numArgs': 1,
                                      'shortname': 'da'},
                             'device': {'args': '<type __builtin__.str>',
                                        'longname': 'device',
                                        'modes': ['create', 'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'd'},
                             'duration': {'args': '<type __builtin__.int>',
                                          'longname': 'duration',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'dr'},
                             'playback': {'args': '<type __builtin__.bool>',
                                          'longname': 'playback',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                             'state': {'args': '<type __builtin__.bool>',
                                       'longname': 'state',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'st'},
                             'wait': {'args': '<type __builtin__.bool>',
                                      'longname': 'wait',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                   'shortFlags': {},
                   'type': 'animation'},
  'redo': {'flags': {}, 'shortFlags': {}, 'type': 'system'},
  'reference': {'callbackFlags': ['editCommand', 'ec'],
                'flags': {'connectionsBroken': {'args': '<type __builtin__.bool>',
                                                'longname': 'connectionsBroken',
                                                'numArgs': 0,
                                                'shortname': 'cb'},
                          'connectionsMade': {'args': '<type __builtin__.bool>',
                                              'longname': 'connectionsMade',
                                              'numArgs': 0,
                                              'shortname': 'cm'},
                          'dagPath': {'args': '<type __builtin__.bool>',
                                      'longname': 'dagPath',
                                      'numArgs': 0,
                                      'shortname': 'dp'},
                          'editCommand': {'args': '<type __builtin__.str>',
                                          'longname': 'editCommand',
                                          'numArgs': 1,
                                          'shortname': 'ec'},
                          'filename': {'args': '<type __builtin__.str>',
                                       'longname': 'filename',
                                       'numArgs': 1,
                                       'shortname': 'f'},
                          'isNodeReferenced': {'args': '<type __builtin__.bool>',
                                               'longname': 'isNodeReferenced',
                                               'numArgs': 0,
                                               'shortname': 'inr'},
                          'longName': {'args': '<type __builtin__.bool>',
                                       'longname': 'longName',
                                       'numArgs': 0,
                                       'shortname': 'ln'},
                          'node': {'args': '<type __builtin__.str>',
                                   'longname': 'node',
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'referenceNode': {'args': '<type __builtin__.str>',
                                            'longname': 'referenceNode',
                                            'numArgs': 1,
                                            'shortname': 'rfn'},
                          'shortName': {'args': '<type __builtin__.bool>',
                                        'longname': 'shortName',
                                        'numArgs': 0,
                                        'shortname': 'sn'}},
                'shortFlags': {'cb': 'connectionsBroken',
                               'cm': 'connectionsMade',
                               'dp': 'dagPath',
                               'ec': 'editCommand',
                               'f': 'filename',
                               'inr': 'isNodeReferenced',
                               'ln': 'longName',
                               'n': 'node',
                               'q': 'query',
                               'rfn': 'referenceNode',
                               'sn': 'shortName'},
                'type': 'system'},
  'referenceEdit': {'callbackFlags': ['editCommand', 'ec'],
                    'flags': {'applyFailedEdits': {'args': '<type __builtin__.bool>',
                                                   'longname': 'applyFailedEdits',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'afe'},
                              'changeEditTarget': {'args': ['<type __builtin__.str>',
                                                            '<type __builtin__.str>'],
                                                   'longname': 'changeEditTarget',
                                                   'modes': ['create'],
                                                   'numArgs': 2,
                                                   'shortname': 'cet'},
                              'editCommand': {'args': '<type __builtin__.str>',
                                              'longname': 'editCommand',
                                              'modes': ['create',
                                                        'query',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'ec'},
                              'failedEdits': {'args': '<type __builtin__.bool>',
                                              'longname': 'failedEdits',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'fld'},
                              'onReferenceNode': {'args': '<type __builtin__.str>',
                                                  'longname': 'onReferenceNode',
                                                  'modes': ['create',
                                                            'query',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'orn'},
                              'removeEdits': {'args': '<type __builtin__.bool>',
                                              'longname': 'removeEdits',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'r'},
                              'successfulEdits': {'args': '<type __builtin__.bool>',
                                                  'longname': 'successfulEdits',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'scs'}},
                    'shortFlags': {'afe': 'applyFailedEdits',
                                   'cet': 'changeEditTarget',
                                   'ec': 'editCommand',
                                   'fld': 'failedEdits',
                                   'orn': 'onReferenceNode',
                                   'r': 'removeEdits',
                                   'scs': 'successfulEdits'},
                    'type': 'system'},
  'referenceQuery': {'callbackFlags': ['editCommand', 'ec'],
                     'flags': {'child': {'args': '<type __builtin__.bool>',
                                         'longname': 'child',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ch'},
                               'dagPath': {'args': '<type __builtin__.bool>',
                                           'longname': 'dagPath',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'dp'},
                               'editAttrs': {'args': '<type __builtin__.bool>',
                                             'longname': 'editAttrs',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ea'},
                               'editCommand': {'args': '<type __builtin__.str>',
                                               'longname': 'editCommand',
                                               'modes': ['create',
                                                         'query',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'ec'},
                               'editNodes': {'args': '<type __builtin__.bool>',
                                             'longname': 'editNodes',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'en'},
                               'editStrings': {'args': '<type __builtin__.bool>',
                                               'longname': 'editStrings',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'es'},
                               'failedEdits': {'args': '<type __builtin__.bool>',
                                               'longname': 'failedEdits',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'fld'},
                               'filename': {'args': '<type __builtin__.bool>',
                                            'longname': 'filename',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'f'},
                               'isExportEdits': {'args': '<type __builtin__.bool>',
                                                 'longname': 'isExportEdits',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'iee'},
                               'isLoaded': {'args': '<type __builtin__.bool>',
                                            'longname': 'isLoaded',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'il'},
                               'isNodeReferenced': {'args': '<type __builtin__.bool>',
                                                    'longname': 'isNodeReferenced',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'inr'},
                               'isPreviewOnly': {'args': '<type __builtin__.bool>',
                                                 'longname': 'isPreviewOnly',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ipo'},
                               'liveEdits': {'args': '<type __builtin__.bool>',
                                             'longname': 'liveEdits',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'le'},
                               'namespace': {'args': '<type __builtin__.bool>',
                                             'longname': 'namespace',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ns'},
                               'nodes': {'args': '<type __builtin__.bool>',
                                         'longname': 'nodes',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'n'},
                               'onReferenceNode': {'args': '<type __builtin__.str>',
                                                   'longname': 'onReferenceNode',
                                                   'modes': ['create',
                                                             'query',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'orn'},
                               'parent': {'args': '<type __builtin__.bool>',
                                          'longname': 'parent',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'p'},
                               'parentNamespace': {'args': '<type __builtin__.bool>',
                                                   'longname': 'parentNamespace',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'pns'},
                               'referenceNode': {'args': '<type __builtin__.bool>',
                                                 'longname': 'referenceNode',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'rfn'},
                               'shortName': {'args': '<type __builtin__.bool>',
                                             'longname': 'shortName',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'shn'},
                               'showDagPath': {'args': '<type __builtin__.bool>',
                                               'longname': 'showDagPath',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'sdp'},
                               'showNamespace': {'args': '<type __builtin__.bool>',
                                                 'longname': 'showNamespace',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'sns'},
                               'successfulEdits': {'args': '<type __builtin__.bool>',
                                                   'longname': 'successfulEdits',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'scs'},
                               'topReference': {'args': '<type __builtin__.bool>',
                                                'longname': 'topReference',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'tr'},
                               'unresolvedName': {'args': '<type __builtin__.bool>',
                                                  'longname': 'unresolvedName',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'un'},
                               'withoutCopyNumber': {'args': '<type __builtin__.bool>',
                                                     'longname': 'withoutCopyNumber',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'wcn'}},
                     'shortFlags': {'ch': 'child',
                                    'dp': 'dagPath',
                                    'ea': 'editAttrs',
                                    'ec': 'editCommand',
                                    'en': 'editNodes',
                                    'es': 'editStrings',
                                    'f': 'filename',
                                    'fld': 'failedEdits',
                                    'iee': 'isExportEdits',
                                    'il': 'isLoaded',
                                    'inr': 'isNodeReferenced',
                                    'ipo': 'isPreviewOnly',
                                    'le': 'liveEdits',
                                    'n': 'nodes',
                                    'ns': 'namespace',
                                    'orn': 'onReferenceNode',
                                    'p': 'parent',
                                    'pns': 'parentNamespace',
                                    'rfn': 'referenceNode',
                                    'scs': 'successfulEdits',
                                    'sdp': 'showDagPath',
                                    'shn': 'shortName',
                                    'sns': 'showNamespace',
                                    'tr': 'topReference',
                                    'un': 'unresolvedName',
                                    'wcn': 'withoutCopyNumber'},
                     'type': 'system'},
  'refineSubdivSelectionList': {'flags': {},
                                'shortFlags': {},
                                'type': 'modeling'},
  'refresh': {'flags': {'currentView': {'args': '<type __builtin__.bool>',
                                        'longname': 'currentView',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'cv'},
                        'fileExtension': {'args': '<type __builtin__.str>',
                                          'longname': 'fileExtension',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'fe'},
                        'filename': {'args': '<type __builtin__.str>',
                                     'longname': 'filename',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'fn'},
                        'force': {'args': '<type __builtin__.bool>',
                                  'longname': 'force',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'f'},
                        'suspend': {'args': '<type __builtin__.bool>',
                                    'longname': 'suspend',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'su'}},
              'shortFlags': {'cv': 'currentView',
                             'f': 'force',
                             'fe': 'fileExtension',
                             'fn': 'filename',
                             'su': 'suspend'},
              'type': 'language'},
  'refreshEditorTemplates': {'flags': {},
                             'shortFlags': {},
                             'type': 'windows'},
  'regionSelectKeyCtx': {'flags': {'bottomManip': {'args': '<type __builtin__.float>',
                                                   'longname': 'bottomManip',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'bot'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'history': {'args': '<type __builtin__.bool>',
                                               'longname': 'history',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'leftManip': {'args': '<type __builtin__.float>',
                                                 'longname': 'leftManip',
                                                 'modes': ['query'],
                                                 'numArgs': 1,
                                                 'shortname': 'lft'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'rightManip': {'args': '<type __builtin__.float>',
                                                  'longname': 'rightManip',
                                                  'modes': ['query'],
                                                  'numArgs': 1,
                                                  'shortname': 'rgt'},
                                   'topManip': {'args': '<type __builtin__.float>',
                                                'longname': 'topManip',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'top'}},
                         'shortFlags': {'bot': 'bottomManip',
                                        'ch': 'history',
                                        'e': 'edit',
                                        'ex': 'exists',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'lft': 'leftManip',
                                        'n': 'name',
                                        'q': 'query',
                                        'rgt': 'rightManip',
                                        'top': 'topManip'},
                         'type': 'context'},
  'rehash': {'flags': {}, 'shortFlags': {}, 'type': 'system'},
  'relationship': {'flags': {'b': {'args': '<type __builtin__.bool>',
                                   'longname': 'b',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 0,
                                   'shortname': 'b'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'relationshipData': {'args': '<type __builtin__.str>',
                                                  'longname': 'relationshipData',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'rd'}},
                   'removedFlags': {'break': 'b'},
                   'shortFlags': {'b': 'b',
                                  'e': 'edit',
                                  'q': 'query',
                                  'rd': 'relationshipData'},
                   'type': 'general'},
  'reloadImage': {'flags': {}, 'shortFlags': {}, 'type': 'system'},
  'removeJoint': {'flags': {}, 'shortFlags': {}, 'type': 'animation'},
  'removeMultiInstance': {'flags': {'allChildren': {'args': '<type __builtin__.bool>',
                                                    'longname': 'allChildren',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'all'},
                                    'b': {'args': '<type __builtin__.bool>',
                                          'longname': 'b',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'b'}},
                          'removedFlags': {'break': 'b'},
                          'shortFlags': {'all': 'allChildren', 'b': 'b'},
                          'type': 'general'},
  'rename': {'flags': {'ignoreShape': {'args': '<type __builtin__.bool>',
                                       'longname': 'ignoreShape',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ignoreShape'},
                       'uuid': {'args': '<type __builtin__.bool>',
                                'longname': 'uuid',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'uid'}},
             'removedFlags': {'is': 'ignoreShape'},
             'shortFlags': {'ignoreShape': 'ignoreShape', 'uid': 'uuid'},
             'type': 'general'},
  'renameAttr': {'flags': {}, 'shortFlags': {}, 'type': 'general'},
  'renameUI': {'flags': {}, 'shortFlags': {}, 'type': 'language'},
  'render': {'flags': {'abortMissingTexture': {'args': '<type __builtin__.bool>',
                                               'longname': 'abortMissingTexture',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'amt'},
                       'batch': {'args': '<type __builtin__.bool>',
                                 'longname': 'batch',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'b'},
                       'keepPreImage': {'args': '<type __builtin__.bool>',
                                        'longname': 'keepPreImage',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'kpi'},
                       'layer': {'args': '<type __builtin__.str>',
                                 'longname': 'layer',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'l'},
                       'nglowpass': {'args': '<type __builtin__.bool>',
                                     'longname': 'nglowpass',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'ngl'},
                       'nshadows': {'args': '<type __builtin__.bool>',
                                    'longname': 'nshadows',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'nsh'},
                       'replace': {'args': '<type __builtin__.bool>',
                                   'longname': 'replace',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'rep'},
                       'xresolution': {'args': '<type __builtin__.int>',
                                       'longname': 'xresolution',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'x'},
                       'yresolution': {'args': '<type __builtin__.int>',
                                       'longname': 'yresolution',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'y'}},
             'shortFlags': {'amt': 'abortMissingTexture',
                            'b': 'batch',
                            'kpi': 'keepPreImage',
                            'l': 'layer',
                            'ngl': 'nglowpass',
                            'nsh': 'nshadows',
                            'rep': 'replace',
                            'x': 'xresolution',
                            'y': 'yresolution'},
             'type': 'rendering'},
  'renderGlobalsNode': {'flags': {'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'parent': {'args': '<type __builtin__.str>',
                                             'longname': 'parent',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'p'},
                                  'renderQuality': {'args': '<type __builtin__.str>',
                                                    'longname': 'renderQuality',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'rq'},
                                  'renderResolution': {'args': '<type __builtin__.str>',
                                                       'longname': 'renderResolution',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'rr'},
                                  'shared': {'args': '<type __builtin__.bool>',
                                             'longname': 'shared',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 's'},
                                  'skipSelect': {'args': '<type __builtin__.bool>',
                                                 'longname': 'skipSelect',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ss'}},
                        'shortFlags': {'n': 'name',
                                       'p': 'parent',
                                       'rq': 'renderQuality',
                                       'rr': 'renderResolution',
                                       's': 'shared',
                                       'ss': 'skipSelect'},
                        'type': 'rendering'},
  'renderInfo': {'flags': {'castShadows': {'args': '<type __builtin__.bool>',
                                           'longname': 'castShadows',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'cs'},
                           'chordHeight': {'args': '<type __builtin__.float>',
                                           'longname': 'chordHeight',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ch'},
                           'chordHeightRatio': {'args': '<type __builtin__.float>',
                                                'longname': 'chordHeightRatio',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'chr'},
                           'doubleSided': {'args': '<type __builtin__.bool>',
                                           'longname': 'doubleSided',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ds'},
                           'edgeSwap': {'args': '<type __builtin__.bool>',
                                        'longname': 'edgeSwap',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'es'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'minScreen': {'args': '<type __builtin__.float>',
                                         'longname': 'minScreen',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ms'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'opposite': {'args': '<type __builtin__.bool>',
                                        'longname': 'opposite',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'smoothShading': {'args': '<type __builtin__.bool>',
                                             'longname': 'smoothShading',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ss'},
                           'unum': {'args': '<type __builtin__.int>',
                                    'longname': 'unum',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'un'},
                           'useChordHeight': {'args': '<type __builtin__.bool>',
                                              'longname': 'useChordHeight',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'uch'},
                           'useChordHeightRatio': {'args': '<type __builtin__.bool>',
                                                   'longname': 'useChordHeightRatio',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ucr'},
                           'useDefaultLights': {'args': '<type __builtin__.bool>',
                                                'longname': 'useDefaultLights',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'udl'},
                           'useMinScreen': {'args': '<type __builtin__.bool>',
                                            'longname': 'useMinScreen',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ums'},
                           'utype': {'args': '<type __builtin__.int>',
                                     'longname': 'utype',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'ut'},
                           'vnum': {'args': '<type __builtin__.int>',
                                    'longname': 'vnum',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'vn'},
                           'vtype': {'args': '<type __builtin__.int>',
                                     'longname': 'vtype',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'vt'}},
                 'shortFlags': {'ch': 'chordHeight',
                                'chr': 'chordHeightRatio',
                                'cs': 'castShadows',
                                'ds': 'doubleSided',
                                'e': 'edit',
                                'es': 'edgeSwap',
                                'ms': 'minScreen',
                                'n': 'name',
                                'o': 'opposite',
                                'q': 'query',
                                'ss': 'smoothShading',
                                'uch': 'useChordHeight',
                                'ucr': 'useChordHeightRatio',
                                'udl': 'useDefaultLights',
                                'ums': 'useMinScreen',
                                'un': 'unum',
                                'ut': 'utype',
                                'vn': 'vnum',
                                'vt': 'vtype'},
                 'type': 'rendering'},
  'renderLayerPostProcess': {'flags': {'keepImages': {'args': '<type __builtin__.bool>',
                                                      'longname': 'keepImages',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'ki'},
                                       'query': {'args': '<type __builtin__.bool>',
                                                 'longname': 'query',
                                                 'numArgs': 0,
                                                 'shortname': 'q'},
                                       'sceneName': {'args': '<type __builtin__.str>',
                                                     'longname': 'sceneName',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'sn'}},
                             'shortFlags': {'ki': 'keepImages',
                                            'q': 'query',
                                            'sn': 'sceneName'},
                             'type': 'rendering'},
  'renderManip': {'flags': {'camera': {'args': ['<type __builtin__.bool>',
                                                '<type __builtin__.bool>',
                                                '<type __builtin__.bool>',
                                                '<type __builtin__.bool>',
                                                '<type __builtin__.bool>'],
                                       'longname': 'camera',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 5,
                                       'shortname': 'cam'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'light': {'args': ['<type __builtin__.bool>',
                                               '<type __builtin__.bool>',
                                               '<type __builtin__.bool>'],
                                      'longname': 'light',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 3,
                                      'shortname': 'lt'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'spotLight': {'args': ['<type __builtin__.bool>',
                                                   '<type __builtin__.bool>',
                                                   '<type __builtin__.bool>',
                                                   '<type __builtin__.bool>',
                                                   '<type __builtin__.bool>',
                                                   '<type __builtin__.bool>',
                                                   '<type __builtin__.bool>'],
                                          'longname': 'spotLight',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 7,
                                          'shortname': 'slt'},
                            'state': {'args': '<type __builtin__.bool>',
                                      'longname': 'state',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'st'}},
                  'shortFlags': {'cam': 'camera',
                                 'e': 'edit',
                                 'lt': 'light',
                                 'q': 'query',
                                 'slt': 'spotLight',
                                 'st': 'state'},
                  'type': 'rendering'},
  'renderPartition': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'}},
                      'shortFlags': {'q': 'query'},
                      'type': 'rendering'},
  'renderPassRegistry': {'flags': {'channels': {'args': '<type __builtin__.int>',
                                                'longname': 'channels',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                                   'isPassSupported': {'args': '<type __builtin__.bool>',
                                                       'longname': 'isPassSupported',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'ips'},
                                   'passID': {'args': '<type __builtin__.str>',
                                              'longname': 'passID',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'pi'},
                                   'passName': {'args': '<type __builtin__.bool>',
                                                'longname': 'passName',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'pn'},
                                   'renderer': {'args': '<type __builtin__.str>',
                                                'longname': 'renderer',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'r'},
                                   'supportedChannelCounts': {'args': '<type __builtin__.bool>',
                                                              'longname': 'supportedChannelCounts',
                                                              'modes': ['create'],
                                                              'numArgs': 0,
                                                              'shortname': 'scc'},
                                   'supportedDataTypes': {'args': '<type __builtin__.bool>',
                                                          'longname': 'supportedDataTypes',
                                                          'modes': ['create'],
                                                          'numArgs': 0,
                                                          'shortname': 'sdt'},
                                   'supportedPassSemantics': {'args': '<type __builtin__.bool>',
                                                              'longname': 'supportedPassSemantics',
                                                              'modes': ['create'],
                                                              'numArgs': 0,
                                                              'shortname': 'ps'},
                                   'supportedRenderPassNames': {'args': '<type __builtin__.bool>',
                                                                'longname': 'supportedRenderPassNames',
                                                                'modes': ['create'],
                                                                'numArgs': 0,
                                                                'shortname': 'spn'},
                                   'supportedRenderPasses': {'args': '<type __builtin__.bool>',
                                                             'longname': 'supportedRenderPasses',
                                                             'modes': ['create'],
                                                             'numArgs': 0,
                                                             'shortname': 'srp'}},
                         'shortFlags': {'ch': 'channels',
                                        'ips': 'isPassSupported',
                                        'pi': 'passID',
                                        'pn': 'passName',
                                        'ps': 'supportedPassSemantics',
                                        'r': 'renderer',
                                        'scc': 'supportedChannelCounts',
                                        'sdt': 'supportedDataTypes',
                                        'spn': 'supportedRenderPassNames',
                                        'srp': 'supportedRenderPasses'},
                         'type': 'rendering'},
  'renderQualityNode': {'flags': {'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'parent': {'args': '<type __builtin__.str>',
                                             'longname': 'parent',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'p'},
                                  'shared': {'args': '<type __builtin__.bool>',
                                             'longname': 'shared',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 's'},
                                  'skipSelect': {'args': '<type __builtin__.bool>',
                                                 'longname': 'skipSelect',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ss'}},
                        'shortFlags': {'n': 'name',
                                       'p': 'parent',
                                       's': 'shared',
                                       'ss': 'skipSelect'},
                        'type': 'rendering'},
  'renderSettings': {'flags': {'camera': {'args': '<type __builtin__.str>',
                                          'longname': 'camera',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'cam'},
                               'customTokenString': {'args': '<type __builtin__.str>',
                                                     'longname': 'customTokenString',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'cts'},
                               'firstImageName': {'args': '<type __builtin__.bool>',
                                                  'longname': 'firstImageName',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'fin'},
                               'fullPath': {'args': '<type __builtin__.bool>',
                                            'longname': 'fullPath',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'fp'},
                               'fullPathTemp': {'args': '<type __builtin__.bool>',
                                                'longname': 'fullPathTemp',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'fpt'},
                               'genericFrameImageName': {'args': '<type __builtin__.str>',
                                                         'longname': 'genericFrameImageName',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'gin'},
                               'imageGenericName': {'args': '<type __builtin__.bool>',
                                                    'longname': 'imageGenericName',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ign'},
                               'lastImageName': {'args': '<type __builtin__.bool>',
                                                 'longname': 'lastImageName',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'lin'},
                               'layer': {'args': '<type __builtin__.str>',
                                         'longname': 'layer',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'lyr'},
                               'leaveUnmatchedTokens': {'args': '<type __builtin__.bool>',
                                                        'longname': 'leaveUnmatchedTokens',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'lut'}},
                     'shortFlags': {'cam': 'camera',
                                    'cts': 'customTokenString',
                                    'fin': 'firstImageName',
                                    'fp': 'fullPath',
                                    'fpt': 'fullPathTemp',
                                    'gin': 'genericFrameImageName',
                                    'ign': 'imageGenericName',
                                    'lin': 'lastImageName',
                                    'lut': 'leaveUnmatchedTokens',
                                    'lyr': 'layer'},
                     'type': 'rendering'},
  'renderThumbnailUpdate': {'flags': {'forceUpdate': {'args': '<type __builtin__.str>',
                                                      'longname': 'forceUpdate',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'fu'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'}},
                            'shortFlags': {'fu': 'forceUpdate',
                                           'q': 'query'},
                            'type': 'rendering'},
  'renderWindowEditor': {'callbackFlags': ['changeCommand', 'cc'],
                         'flags': {'autoResize': {'args': '<type __builtin__.bool>',
                                                  'longname': 'autoResize',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ar'},
                                   'blendMode': {'args': '<type __builtin__.int>',
                                                 'longname': 'blendMode',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'blm'},
                                   'caption': {'args': '<type __builtin__.str>',
                                               'longname': 'caption',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cap'},
                                   'changeCommand': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'changeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 4,
                                                     'shortname': 'cc'},
                                   'clear': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'clear',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 5,
                                             'shortname': 'cl'},
                                   'cmEnabled': {'args': '<type __builtin__.bool>',
                                                 'longname': 'cmEnabled',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cme'},
                                   'colorManage': {'args': '<type __builtin__.bool>',
                                                   'longname': 'colorManage',
                                                   'modes': ['edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'com'},
                                   'compDisplay': {'args': '<type __builtin__.int>',
                                                   'longname': 'compDisplay',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cd'},
                                   'compImageFile': {'args': '<type __builtin__.str>',
                                                     'longname': 'compImageFile',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'cif'},
                                   'control': {'args': '<type __builtin__.bool>',
                                               'longname': 'control',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'ctl'},
                                   'currentCamera': {'args': '<type __builtin__.str>',
                                                     'longname': 'currentCamera',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'crc'},
                                   'currentCameraRig': {'args': '<type __builtin__.str>',
                                                        'longname': 'currentCameraRig',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'crg'},
                                   'defineTemplate': {'args': '<type __builtin__.str>',
                                                      'longname': 'defineTemplate',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'dt'},
                                   'displayImage': {'args': '<type __builtin__.int>',
                                                    'longname': 'displayImage',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'di'},
                                   'displayImageViewCount': {'args': '<type __builtin__.int>',
                                                             'longname': 'displayImageViewCount',
                                                             'modes': ['query'],
                                                             'numArgs': 1,
                                                             'shortname': 'dvc'},
                                   'displayStyle': {'args': '<type __builtin__.str>',
                                                    'longname': 'displayStyle',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dst'},
                                   'docTag': {'args': '<type __builtin__.str>',
                                              'longname': 'docTag',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dtg'},
                                   'doubleBuffer': {'args': '<type __builtin__.bool>',
                                                    'longname': 'doubleBuffer',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'dbf'},
                                   'drawAxis': {'args': '<type __builtin__.bool>',
                                                'longname': 'drawAxis',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'da'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'editorName': {'args': '<type __builtin__.bool>',
                                                  'longname': 'editorName',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'en'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'exposure': {'args': '<type __builtin__.float>',
                                                'longname': 'exposure',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'exp'},
                                   'filter': {'args': '<type __builtin__.str>',
                                              'longname': 'filter',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'f'},
                                   'forceMainConnection': {'args': '<type __builtin__.str>',
                                                           'longname': 'forceMainConnection',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'fmc'},
                                   'frameImage': {'args': '<type __builtin__.bool>',
                                                  'longname': 'frameImage',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'fi'},
                                   'frameRegion': {'args': '<type __builtin__.bool>',
                                                   'longname': 'frameRegion',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'fr'},
                                   'gamma': {'args': '<type __builtin__.float>',
                                             'longname': 'gamma',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ga'},
                                   'highlightConnection': {'args': '<type __builtin__.str>',
                                                           'longname': 'highlightConnection',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'hlc'},
                                   'loadImage': {'args': '<type __builtin__.str>',
                                                 'longname': 'loadImage',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'li'},
                                   'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                          'longname': 'lockMainConnection',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 0,
                                                          'shortname': 'lck'},
                                   'mainListConnection': {'args': '<type __builtin__.str>',
                                                          'longname': 'mainListConnection',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'mlc'},
                                   'marquee': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'marquee',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 4,
                                               'shortname': 'mq'},
                                   'nbImages': {'args': '<type __builtin__.bool>',
                                                'longname': 'nbImages',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'nim'},
                                   'nextViewImage': {'args': '<type __builtin__.bool>',
                                                     'longname': 'nextViewImage',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'nvi'},
                                   'outputColorManage': {'args': '<type __builtin__.bool>',
                                                         'longname': 'outputColorManage',
                                                         'modes': ['edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'ocm'},
                                   'panel': {'args': '<type __builtin__.str>',
                                             'longname': 'panel',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'pnl'},
                                   'parent': {'args': '<type __builtin__.str>',
                                              'longname': 'parent',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'p'},
                                   'pcaption': {'args': '<type __builtin__.str>',
                                                'longname': 'pcaption',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'pca'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'realSize': {'args': '<type __builtin__.bool>',
                                                'longname': 'realSize',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'rs'},
                                   'refresh': {'args': '<type __builtin__.bool>',
                                               'longname': 'refresh',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'ref'},
                                   'removeAllImages': {'args': '<type __builtin__.bool>',
                                                       'longname': 'removeAllImages',
                                                       'modes': ['edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'ra'},
                                   'removeImage': {'args': '<type __builtin__.bool>',
                                                   'longname': 'removeImage',
                                                   'modes': ['edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'ri'},
                                   'resetRegion': {'args': '<type __builtin__.bool>',
                                                   'longname': 'resetRegion',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'rr'},
                                   'resetViewImage': {'args': '<type __builtin__.bool>',
                                                      'longname': 'resetViewImage',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'rvi'},
                                   'saveImage': {'args': '<type __builtin__.bool>',
                                                 'longname': 'saveImage',
                                                 'modes': ['edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'si'},
                                   'scaleBlue': {'args': '<type __builtin__.float>',
                                                 'longname': 'scaleBlue',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sb'},
                                   'scaleGreen': {'args': '<type __builtin__.float>',
                                                  'longname': 'scaleGreen',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sg'},
                                   'scaleRed': {'args': '<type __builtin__.float>',
                                                'longname': 'scaleRed',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sr'},
                                   'selectionConnection': {'args': '<type __builtin__.str>',
                                                           'longname': 'selectionConnection',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'slc'},
                                   'showRegion': {'args': ['<type __builtin__.int>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'showRegion',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 2,
                                                  'shortname': 'srg'},
                                   'singleBuffer': {'args': '<type __builtin__.bool>',
                                                    'longname': 'singleBuffer',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'sbf'},
                                   'snapshot': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'snapshot',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'snp'},
                                   'snapshotMode': {'args': '<type __builtin__.bool>',
                                                    'longname': 'snapshotMode',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'snm'},
                                   'stateString': {'args': '<type __builtin__.bool>',
                                                   'longname': 'stateString',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'sts'},
                                   'stereo': {'args': '<type __builtin__.int>',
                                              'longname': 'stereo',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 's'},
                                   'stereoImageOrientation': {'args': ['<type __builtin__.str>',
                                                                       '<type __builtin__.str>'],
                                                              'longname': 'stereoImageOrientation',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 2,
                                                              'shortname': 'sio'},
                                   'stereoMode': {'args': '<type __builtin__.str>',
                                                  'longname': 'stereoMode',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sm'},
                                   'toggle': {'args': '<type __builtin__.bool>',
                                              'longname': 'toggle',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'tgl'},
                                   'unParent': {'args': '<type __builtin__.bool>',
                                                'longname': 'unParent',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'up'},
                                   'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                            'longname': 'unlockMainConnection',
                                                            'modes': ['create',
                                                                      'edit'],
                                                            'numArgs': 0,
                                                            'shortname': 'ulk'},
                                   'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                            'longname': 'updateMainConnection',
                                                            'modes': ['create',
                                                                      'edit'],
                                                            'numArgs': 0,
                                                            'shortname': 'upd'},
                                   'useTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'useTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ut'},
                                   'viewImageCount': {'args': '<type __builtin__.int>',
                                                      'longname': 'viewImageCount',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vic'},
                                   'viewTransformName': {'args': '<type __builtin__.str>',
                                                         'longname': 'viewTransformName',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'vtn'},
                                   'writeImage': {'args': '<type __builtin__.str>',
                                                  'longname': 'writeImage',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'wi'}},
                         'shortFlags': {'ar': 'autoResize',
                                        'blm': 'blendMode',
                                        'cap': 'caption',
                                        'cc': 'changeCommand',
                                        'cd': 'compDisplay',
                                        'cif': 'compImageFile',
                                        'cl': 'clear',
                                        'cme': 'cmEnabled',
                                        'com': 'colorManage',
                                        'crc': 'currentCamera',
                                        'crg': 'currentCameraRig',
                                        'ctl': 'control',
                                        'da': 'drawAxis',
                                        'dbf': 'doubleBuffer',
                                        'di': 'displayImage',
                                        'dst': 'displayStyle',
                                        'dt': 'defineTemplate',
                                        'dtg': 'docTag',
                                        'dvc': 'displayImageViewCount',
                                        'e': 'edit',
                                        'en': 'editorName',
                                        'ex': 'exists',
                                        'exp': 'exposure',
                                        'f': 'filter',
                                        'fi': 'frameImage',
                                        'fmc': 'forceMainConnection',
                                        'fr': 'frameRegion',
                                        'ga': 'gamma',
                                        'hlc': 'highlightConnection',
                                        'lck': 'lockMainConnection',
                                        'li': 'loadImage',
                                        'mlc': 'mainListConnection',
                                        'mq': 'marquee',
                                        'nim': 'nbImages',
                                        'nvi': 'nextViewImage',
                                        'ocm': 'outputColorManage',
                                        'p': 'parent',
                                        'pca': 'pcaption',
                                        'pnl': 'panel',
                                        'q': 'query',
                                        'ra': 'removeAllImages',
                                        'ref': 'refresh',
                                        'ri': 'removeImage',
                                        'rr': 'resetRegion',
                                        'rs': 'realSize',
                                        'rvi': 'resetViewImage',
                                        's': 'stereo',
                                        'sb': 'scaleBlue',
                                        'sbf': 'singleBuffer',
                                        'sg': 'scaleGreen',
                                        'si': 'saveImage',
                                        'sio': 'stereoImageOrientation',
                                        'slc': 'selectionConnection',
                                        'sm': 'stereoMode',
                                        'snm': 'snapshotMode',
                                        'snp': 'snapshot',
                                        'sr': 'scaleRed',
                                        'srg': 'showRegion',
                                        'sts': 'stateString',
                                        'tgl': 'toggle',
                                        'ulk': 'unlockMainConnection',
                                        'up': 'unParent',
                                        'upd': 'updateMainConnection',
                                        'ut': 'useTemplate',
                                        'vic': 'viewImageCount',
                                        'vtn': 'viewTransformName',
                                        'wi': 'writeImage'},
                         'type': 'rendering'},
  'renderWindowSelectContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                   'longname': 'edit',
                                                   'numArgs': 0,
                                                   'shortname': 'e'},
                                          'exists': {'args': '<type __builtin__.bool>',
                                                     'longname': 'exists',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'ex'},
                                          'image1': {'args': '<type __builtin__.str>',
                                                     'longname': 'image1',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'i1'},
                                          'image2': {'args': '<type __builtin__.str>',
                                                     'longname': 'image2',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'i2'},
                                          'image3': {'args': '<type __builtin__.str>',
                                                     'longname': 'image3',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'i3'},
                                          'query': {'args': '<type __builtin__.bool>',
                                                    'longname': 'query',
                                                    'numArgs': 0,
                                                    'shortname': 'q'}},
                                'shortFlags': {'e': 'edit',
                                               'ex': 'exists',
                                               'i1': 'image1',
                                               'i2': 'image2',
                                               'i3': 'image3',
                                               'q': 'query'},
                                'type': 'context'},
  'renderer': {'callbackFlags': ['commandRenderProcedure', 'cr'],
               'flags': {'addGlobalsNode': {'args': '<type __builtin__.str>',
                                            'longname': 'addGlobalsNode',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'agn'},
                         'addGlobalsTab': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'addGlobalsTab',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 3,
                                           'shortname': 'agt'},
                         'batchRenderOptionsProcedure': {'args': '<type __builtin__.str>',
                                                         'longname': 'batchRenderOptionsProcedure',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'bro'},
                         'batchRenderOptionsStringProcedure': {'args': '<type __builtin__.str>',
                                                               'longname': 'batchRenderOptionsStringProcedure',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'bso'},
                         'batchRenderProcedure': {'args': '<type __builtin__.str>',
                                                  'longname': 'batchRenderProcedure',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'br'},
                         'cancelBatchRenderProcedure': {'args': '<type __builtin__.str>',
                                                        'longname': 'cancelBatchRenderProcedure',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'cbr'},
                         'changeIprRegionProcedure': {'args': '<type __builtin__.str>',
                                                      'longname': 'changeIprRegionProcedure',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cir'},
                         'commandRenderProcedure': {'args': '<type __builtin__.str>',
                                                    'longname': 'commandRenderProcedure',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'cr'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'globalsNodes': {'args': '<type __builtin__.bool>',
                                          'longname': 'globalsNodes',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'gn'},
                         'globalsTabCreateProcNames': {'args': '<type __builtin__.bool>',
                                                       'longname': 'globalsTabCreateProcNames',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'gtc'},
                         'globalsTabLabels': {'args': '<type __builtin__.bool>',
                                              'longname': 'globalsTabLabels',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 0,
                                              'shortname': 'gtl'},
                         'globalsTabUpdateProcNames': {'args': '<type __builtin__.bool>',
                                                       'longname': 'globalsTabUpdateProcNames',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'gtu'},
                         'iprOptionsMenuLabel': {'args': '<type __builtin__.str>',
                                                 'longname': 'iprOptionsMenuLabel',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ipm'},
                         'iprOptionsProcedure': {'args': '<type __builtin__.str>',
                                                 'longname': 'iprOptionsProcedure',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'io'},
                         'iprOptionsSubMenuProcedure': {'args': '<type __builtin__.str>',
                                                        'longname': 'iprOptionsSubMenuProcedure',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ips'},
                         'iprRenderProcedure': {'args': '<type __builtin__.str>',
                                                'longname': 'iprRenderProcedure',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ipr'},
                         'iprRenderSubMenuProcedure': {'args': '<type __builtin__.str>',
                                                       'longname': 'iprRenderSubMenuProcedure',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'irs'},
                         'isRunningIprProcedure': {'args': '<type __builtin__.str>',
                                                   'longname': 'isRunningIprProcedure',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'isr'},
                         'logoCallbackProcedure': {'args': '<type __builtin__.str>',
                                                   'longname': 'logoCallbackProcedure',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'lgc'},
                         'logoImageName': {'args': '<type __builtin__.str>',
                                           'longname': 'logoImageName',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'log'},
                         'materialViewRendererList': {'args': '<type __builtin__.bool>',
                                                      'longname': 'materialViewRendererList',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'mvl'},
                         'materialViewRendererPause': {'args': '<type __builtin__.bool>',
                                                       'longname': 'materialViewRendererPause',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'mvp'},
                         'materialViewRendererSuspend': {'args': '<type __builtin__.bool>',
                                                         'longname': 'materialViewRendererSuspend',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'mvs'},
                         'namesOfAvailableRenderers': {'args': '<type __builtin__.bool>',
                                                       'longname': 'namesOfAvailableRenderers',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'ava'},
                         'pauseIprRenderProcedure': {'args': '<type __builtin__.str>',
                                                     'longname': 'pauseIprRenderProcedure',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'psi'},
                         'polyPrelightProcedure': {'args': '<type __builtin__.str>',
                                                   'longname': 'polyPrelightProcedure',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'pp'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'refreshIprRenderProcedure': {'args': '<type __builtin__.str>',
                                                       'longname': 'refreshIprRenderProcedure',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'rfi'},
                         'renderDiagnosticsProcedure': {'args': '<type __builtin__.str>',
                                                        'longname': 'renderDiagnosticsProcedure',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'rd'},
                         'renderGlobalsProcedure': {'args': '<type __builtin__.str>',
                                                    'longname': 'renderGlobalsProcedure',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'rg'},
                         'renderMenuProcedure': {'args': '<type __builtin__.str>',
                                                 'longname': 'renderMenuProcedure',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rm'},
                         'renderOptionsProcedure': {'args': '<type __builtin__.str>',
                                                    'longname': 'renderOptionsProcedure',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ro'},
                         'renderProcedure': {'args': '<type __builtin__.str>',
                                             'longname': 'renderProcedure',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'r'},
                         'renderRegionProcedure': {'args': '<type __builtin__.str>',
                                                   'longname': 'renderRegionProcedure',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'rr'},
                         'renderSequenceProcedure': {'args': '<type __builtin__.str>',
                                                     'longname': 'renderSequenceProcedure',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'rs'},
                         'rendererUIName': {'args': '<type __builtin__.str>',
                                            'longname': 'rendererUIName',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ui'},
                         'renderingEditorsSubMenuProcedure': {'args': '<type __builtin__.str>',
                                                              'longname': 'renderingEditorsSubMenuProcedure',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'res'},
                         'showBatchRenderLogProcedure': {'args': '<type __builtin__.str>',
                                                         'longname': 'showBatchRenderLogProcedure',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'brl'},
                         'showBatchRenderProcedure': {'args': '<type __builtin__.str>',
                                                      'longname': 'showBatchRenderProcedure',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'sbr'},
                         'showRenderLogProcedure': {'args': '<type __builtin__.str>',
                                                    'longname': 'showRenderLogProcedure',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'srl'},
                         'startIprRenderProcedure': {'args': '<type __builtin__.str>',
                                                     'longname': 'startIprRenderProcedure',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sti'},
                         'stopIprRenderProcedure': {'args': '<type __builtin__.str>',
                                                    'longname': 'stopIprRenderProcedure',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'spi'},
                         'supportColorManagement': {'args': '<type __builtin__.bool>',
                                                    'longname': 'supportColorManagement',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'scm'},
                         'textureBakingProcedure': {'args': '<type __builtin__.str>',
                                                    'longname': 'textureBakingProcedure',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'tb'},
                         'unregisterRenderer': {'args': '<type __builtin__.bool>',
                                                'longname': 'unregisterRenderer',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 0,
                                                'shortname': 'unr'}},
               'shortFlags': {'agn': 'addGlobalsNode',
                              'agt': 'addGlobalsTab',
                              'ava': 'namesOfAvailableRenderers',
                              'br': 'batchRenderProcedure',
                              'brl': 'showBatchRenderLogProcedure',
                              'bro': 'batchRenderOptionsProcedure',
                              'bso': 'batchRenderOptionsStringProcedure',
                              'cbr': 'cancelBatchRenderProcedure',
                              'cir': 'changeIprRegionProcedure',
                              'cr': 'commandRenderProcedure',
                              'e': 'edit',
                              'ex': 'exists',
                              'gn': 'globalsNodes',
                              'gtc': 'globalsTabCreateProcNames',
                              'gtl': 'globalsTabLabels',
                              'gtu': 'globalsTabUpdateProcNames',
                              'io': 'iprOptionsProcedure',
                              'ipm': 'iprOptionsMenuLabel',
                              'ipr': 'iprRenderProcedure',
                              'ips': 'iprOptionsSubMenuProcedure',
                              'irs': 'iprRenderSubMenuProcedure',
                              'isr': 'isRunningIprProcedure',
                              'lgc': 'logoCallbackProcedure',
                              'log': 'logoImageName',
                              'mvl': 'materialViewRendererList',
                              'mvp': 'materialViewRendererPause',
                              'mvs': 'materialViewRendererSuspend',
                              'pp': 'polyPrelightProcedure',
                              'psi': 'pauseIprRenderProcedure',
                              'q': 'query',
                              'r': 'renderProcedure',
                              'rd': 'renderDiagnosticsProcedure',
                              'res': 'renderingEditorsSubMenuProcedure',
                              'rfi': 'refreshIprRenderProcedure',
                              'rg': 'renderGlobalsProcedure',
                              'rm': 'renderMenuProcedure',
                              'ro': 'renderOptionsProcedure',
                              'rr': 'renderRegionProcedure',
                              'rs': 'renderSequenceProcedure',
                              'sbr': 'showBatchRenderProcedure',
                              'scm': 'supportColorManagement',
                              'spi': 'stopIprRenderProcedure',
                              'srl': 'showRenderLogProcedure',
                              'sti': 'startIprRenderProcedure',
                              'tb': 'textureBakingProcedure',
                              'ui': 'rendererUIName',
                              'unr': 'unregisterRenderer'},
               'type': 'rendering'},
  'reorder': {'flags': {'back': {'args': '<type __builtin__.bool>',
                                 'longname': 'back',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'b'},
                        'front': {'args': '<type __builtin__.bool>',
                                  'longname': 'front',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'f'},
                        'relative': {'args': '<type __builtin__.int>',
                                     'longname': 'relative',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'r'}},
              'shortFlags': {'b': 'back', 'f': 'front', 'r': 'relative'},
              'type': 'general'},
  'reorderContainer': {'flags': {'back': {'args': '<type __builtin__.bool>',
                                          'longname': 'back',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'b'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'front': {'args': '<type __builtin__.bool>',
                                           'longname': 'front',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 0,
                                           'shortname': 'f'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'relative': {'args': '<type __builtin__.int>',
                                              'longname': 'relative',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'r'}},
                       'shortFlags': {'b': 'back',
                                      'e': 'edit',
                                      'f': 'front',
                                      'q': 'query',
                                      'r': 'relative'},
                       'type': 'general'},
  'reorderDeformers': {'flags': {'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'n'}},
                       'shortFlags': {'n': 'name'},
                       'type': 'animation'},
  'repeatLast': {'callbackFlags': ['addCommandLabel',
                                   'acl',
                                   'addCommand',
                                   'ac',
                                   'commandList',
                                   'cl',
                                   'commandNameList',
                                   'cnl'],
                 'flags': {'addCommand': {'args': '<type __builtin__.str>',
                                          'longname': 'addCommand',
                                          'numArgs': 1,
                                          'shortname': 'ac'},
                           'addCommandLabel': {'args': '<type __builtin__.str>',
                                               'longname': 'addCommandLabel',
                                               'numArgs': 1,
                                               'shortname': 'acl'},
                           'commandList': {'args': '<type __builtin__.int>',
                                           'longname': 'commandList',
                                           'numArgs': 1,
                                           'shortname': 'cl'},
                           'commandNameList': {'args': '<type __builtin__.int>',
                                               'longname': 'commandNameList',
                                               'numArgs': 1,
                                               'shortname': 'cnl'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'historyLimit': {'args': '<type __builtin__.int>',
                                            'longname': 'historyLimit',
                                            'numArgs': 1,
                                            'shortname': 'hl'},
                           'item': {'args': '<type __builtin__.int>',
                                    'longname': 'item',
                                    'numArgs': 1,
                                    'shortname': 'i'},
                           'numberOfHistoryItems': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfHistoryItems',
                                                    'numArgs': 0,
                                                    'shortname': 'nhi'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'ac': 'addCommand',
                                'acl': 'addCommandLabel',
                                'cl': 'commandList',
                                'cnl': 'commandNameList',
                                'e': 'edit',
                                'hl': 'historyLimit',
                                'i': 'item',
                                'nhi': 'numberOfHistoryItems',
                                'q': 'query'},
                 'type': 'other'},
  'replaceCacheFrames': {'type': 'runtime'},
  'replaceCacheFramesOpt': {'type': 'runtime'},
  'requires': {'flags': {'dataType': {'args': '<type __builtin__.str>',
                                      'longname': 'dataType',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'dt'},
                         'nodeType': {'args': '<type __builtin__.str>',
                                      'longname': 'nodeType',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'nt'}},
               'shortFlags': {'dt': 'dataType', 'nt': 'nodeType'},
               'type': 'system'},
  'reroot': {'flags': {}, 'shortFlags': {}, 'type': 'animation'},
  'resampleFluid': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'resampleDepth': {'args': '<type __builtin__.int>',
                                                'longname': 'resampleDepth',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'rd'},
                              'resampleHeight': {'args': '<type __builtin__.int>',
                                                 'longname': 'resampleHeight',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rh'},
                              'resampleWidth': {'args': '<type __builtin__.int>',
                                                'longname': 'resampleWidth',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'rw'}},
                    'shortFlags': {'e': 'edit',
                                   'q': 'query',
                                   'rd': 'resampleDepth',
                                   'rh': 'resampleHeight',
                                   'rw': 'resampleWidth'},
                    'type': 'effects'},
  'resetTool': {'flags': {}, 'shortFlags': {}, 'type': 'general'},
  'resolutionNode': {'flags': {'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'shared': {'args': '<type __builtin__.bool>',
                                          'longname': 'shared',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 's'},
                               'skipSelect': {'args': '<type __builtin__.bool>',
                                              'longname': 'skipSelect',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ss'}},
                     'shortFlags': {'n': 'name',
                                    'p': 'parent',
                                    's': 'shared',
                                    'ss': 'skipSelect'},
                     'type': 'rendering'},
  'resourceManager': {'flags': {'nameFilter': {'args': '<type __builtin__.str>',
                                               'longname': 'nameFilter',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'nf'},
                                'saveAs': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'saveAs',
                                           'modes': ['create'],
                                           'numArgs': 2,
                                           'shortname': 's'}},
                      'shortFlags': {'nf': 'nameFilter', 's': 'saveAs'},
                      'type': 'language'},
  'retimeHelper': {'flags': {'deleteFrame': {'args': '<type __builtin__.int>',
                                             'longname': 'deleteFrame',
                                             'numArgs': 1,
                                             'shortname': 'df'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frame': {'args': '<type __builtin__.float>',
                                       'longname': 'frame',
                                       'numArgs': 1,
                                       'shortname': 'f'},
                             'lockBar': {'args': ['<type __builtin__.int>',
                                                  '<type __builtin__.int>'],
                                         'longname': 'lockBar',
                                         'numArgs': 2,
                                         'shortname': 'lb'},
                             'locks': {'args': '<type __builtin__.int>',
                                       'longname': 'locks',
                                       'numArgs': 1,
                                       'shortname': 'lk'},
                             'mouseOver': {'args': '<type __builtin__.bool>',
                                           'longname': 'mouseOver',
                                           'numArgs': 0,
                                           'shortname': 'mo'},
                             'moveFrame': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'moveFrame',
                                           'numArgs': 2,
                                           'shortname': 'mf'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'df': 'deleteFrame',
                                  'e': 'edit',
                                  'f': 'frame',
                                  'lb': 'lockBar',
                                  'lk': 'locks',
                                  'mf': 'moveFrame',
                                  'mo': 'mouseOver',
                                  'q': 'query'},
                   'type': 'other'},
  'retimeKeyCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'history': {'args': '<type __builtin__.bool>',
                                         'longname': 'history',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ch'},
                             'image1': {'args': '<type __builtin__.str>',
                                        'longname': 'image1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i1'},
                             'image2': {'args': '<type __builtin__.str>',
                                        'longname': 'image2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i2'},
                             'image3': {'args': '<type __builtin__.str>',
                                        'longname': 'image3',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i3'},
                             'moveByFrame': {'args': '<type __builtin__.int>',
                                             'longname': 'moveByFrame',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'mbf'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'snapOnFrame': {'args': '<type __builtin__.bool>',
                                             'longname': 'snapOnFrame',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sof'}},
                   'shortFlags': {'ch': 'history',
                                  'e': 'edit',
                                  'ex': 'exists',
                                  'i1': 'image1',
                                  'i2': 'image2',
                                  'i3': 'image3',
                                  'mbf': 'moveByFrame',
                                  'n': 'name',
                                  'q': 'query',
                                  'sof': 'snapOnFrame'},
                   'type': 'context'},
  'reverseCurve': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'curveOnSurface': {'args': '<type __builtin__.bool>',
                                                'longname': 'curveOnSurface',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'cos'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'noChanges': {'args': '<type __builtin__.bool>',
                                           'longname': 'noChanges',
                                           'numArgs': 1,
                                           'shortname': 'nc'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'object': {'args': '<type __builtin__.bool>',
                                        'longname': 'object',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'range': {'args': '<type __builtin__.bool>',
                                       'longname': 'range',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'rn'},
                             'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                 'longname': 'replaceOriginal',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'rpo'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'cos': 'curveOnSurface',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'n': 'name',
                                  'nc': 'noChanges',
                                  'nds': 'nodeState',
                                  'o': 'object',
                                  'q': 'query',
                                  'rn': 'range',
                                  'rpo': 'replaceOriginal'},
                   'type': 'modeling'},
  'reverseSurface': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'direction': {'args': '<type __builtin__.int>',
                                             'longname': 'direction',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'd'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'noChanges': {'args': '<type __builtin__.bool>',
                                             'longname': 'noChanges',
                                             'numArgs': 1,
                                             'shortname': 'nc'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'object': {'args': '<type __builtin__.bool>',
                                          'longname': 'object',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'o'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                   'longname': 'replaceOriginal',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'rpo'}},
                     'shortFlags': {'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'd': 'direction',
                                    'e': 'edit',
                                    'fzn': 'frozen',
                                    'n': 'name',
                                    'nc': 'noChanges',
                                    'nds': 'nodeState',
                                    'o': 'object',
                                    'q': 'query',
                                    'rpo': 'replaceOriginal'},
                     'type': 'modeling'},
  'revolve': {'flags': {'autoCorrectNormal': {'args': '<type __builtin__.bool>',
                                              'longname': 'autoCorrectNormal',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'acn'},
                        'axis': {'args': ['<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>'],
                                 'longname': 'axis',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 3,
                                 'shortname': 'ax'},
                        'axisChoice': {'args': '<type __builtin__.int>',
                                       'longname': 'axisChoice',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'aco'},
                        'axisX': {'args': '<type __builtin__.float>',
                                  'longname': 'axisX',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'axx'},
                        'axisY': {'args': '<type __builtin__.float>',
                                  'longname': 'axisY',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'axy'},
                        'axisZ': {'args': '<type __builtin__.float>',
                                  'longname': 'axisZ',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'axz'},
                        'bridge': {'args': '<type __builtin__.bool>',
                                   'longname': 'bridge',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'br'},
                        'caching': {'args': '<type __builtin__.bool>',
                                    'longname': 'caching',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'cch'},
                        'computePivotAndAxis': {'args': '<type __builtin__.int>',
                                                'longname': 'computePivotAndAxis',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cpa'},
                        'constructionHistory': {'args': '<type __builtin__.bool>',
                                                'longname': 'constructionHistory',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                        'degree': {'args': '<type __builtin__.int>',
                                   'longname': 'degree',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'd'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'endSweep': {'args': '<type __builtin__.float>',
                                     'longname': 'endSweep',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'esw'},
                        'frozen': {'args': '<type __builtin__.bool>',
                                   'longname': 'frozen',
                                   'numArgs': 1,
                                   'shortname': 'fzn'},
                        'mergeItems': {'args': '<type __builtin__.bool>',
                                       'longname': 'mergeItems',
                                       'numArgs': 1,
                                       'shortname': 'mi'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'nodeState': {'args': '<type __builtin__.int>',
                                      'longname': 'nodeState',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'nds'},
                        'object': {'args': '<type __builtin__.bool>',
                                   'longname': 'object',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'o'},
                        'pivot': {'args': ['<type __builtin__.float>',
                                           '<type __builtin__.float>',
                                           '<type __builtin__.float>'],
                                  'longname': 'pivot',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 3,
                                  'shortname': 'p'},
                        'pivotX': {'args': '<type __builtin__.float>',
                                   'longname': 'pivotX',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'px'},
                        'pivotY': {'args': '<type __builtin__.float>',
                                   'longname': 'pivotY',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'py'},
                        'pivotZ': {'args': '<type __builtin__.float>',
                                   'longname': 'pivotZ',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'pz'},
                        'polygon': {'args': '<type __builtin__.int>',
                                    'longname': 'polygon',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'po'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'radius': {'args': '<type __builtin__.float>',
                                   'longname': 'radius',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'r'},
                        'radiusAnchor': {'args': '<type __builtin__.float>',
                                         'longname': 'radiusAnchor',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ra'},
                        'range': {'args': '<type __builtin__.bool>',
                                  'longname': 'range',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'rn'},
                        'rebuild': {'args': '<type __builtin__.bool>',
                                    'longname': 'rebuild',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'rb'},
                        'sections': {'args': '<type __builtin__.int>',
                                     'longname': 'sections',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 's'},
                        'startSweep': {'args': '<type __builtin__.float>',
                                       'longname': 'startSweep',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ssw'},
                        'tolerance': {'args': '<type __builtin__.float>',
                                      'longname': 'tolerance',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'tol'},
                        'useLocalPivot': {'args': '<type __builtin__.bool>',
                                          'longname': 'useLocalPivot',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ulp'},
                        'useTolerance': {'args': '<type __builtin__.bool>',
                                         'longname': 'useTolerance',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ut'}},
              'shortFlags': {'acn': 'autoCorrectNormal',
                             'aco': 'axisChoice',
                             'ax': 'axis',
                             'axx': 'axisX',
                             'axy': 'axisY',
                             'axz': 'axisZ',
                             'br': 'bridge',
                             'cch': 'caching',
                             'ch': 'constructionHistory',
                             'cpa': 'computePivotAndAxis',
                             'd': 'degree',
                             'e': 'edit',
                             'esw': 'endSweep',
                             'fzn': 'frozen',
                             'mi': 'mergeItems',
                             'n': 'name',
                             'nds': 'nodeState',
                             'o': 'object',
                             'p': 'pivot',
                             'po': 'polygon',
                             'px': 'pivotX',
                             'py': 'pivotY',
                             'pz': 'pivotZ',
                             'q': 'query',
                             'r': 'radius',
                             'ra': 'radiusAnchor',
                             'rb': 'rebuild',
                             'rn': 'range',
                             's': 'sections',
                             'ssw': 'startSweep',
                             'tol': 'tolerance',
                             'ulp': 'useLocalPivot',
                             'ut': 'useTolerance'},
              'type': 'modeling'},
  'rigidBody': {'flags': {'active': {'args': '<type __builtin__.bool>',
                                     'longname': 'active',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'act'},
                          'angularVelocity': {'args': '<type __builtin__.bool>',
                                              'longname': 'angularVelocity',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'av'},
                          'applyForceAt': {'args': '<type __builtin__.str>',
                                           'longname': 'applyForceAt',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'afa'},
                          'bounciness': {'args': '<type __builtin__.float>',
                                         'longname': 'bounciness',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'b'},
                          'cache': {'args': '<type __builtin__.bool>',
                                    'longname': 'cache',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'c'},
                          'centerOfMass': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'centerOfMass',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'com'},
                          'collisions': {'args': '<type __builtin__.bool>',
                                         'longname': 'collisions',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cl'},
                          'contactCount': {'args': '<type __builtin__.bool>',
                                           'longname': 'contactCount',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'cc'},
                          'contactName': {'args': '<type __builtin__.bool>',
                                          'longname': 'contactName',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'cn'},
                          'contactPosition': {'args': '<type __builtin__.bool>',
                                              'longname': 'contactPosition',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'cp'},
                          'damping': {'args': '<type __builtin__.float>',
                                      'longname': 'damping',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dp'},
                          'deleteCache': {'args': '<type __builtin__.bool>',
                                          'longname': 'deleteCache',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'dc'},
                          'dynamicFriction': {'args': '<type __builtin__.float>',
                                              'longname': 'dynamicFriction',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'df'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'force': {'args': '<type __builtin__.bool>',
                                    'longname': 'force',
                                    'modes': ['query'],
                                    'numArgs': 0,
                                    'shortname': 'f'},
                          'ignore': {'args': '<type __builtin__.bool>',
                                     'longname': 'ignore',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'ig'},
                          'impulse': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'impulse',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 3,
                                      'shortname': 'i'},
                          'impulsePosition': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'impulsePosition',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'imp'},
                          'initialAngularVelocity': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'initialAngularVelocity',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'iav'},
                          'initialVelocity': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'initialVelocity',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'iv'},
                          'layer': {'args': '<type __builtin__.int>',
                                    'longname': 'layer',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'l'},
                          'lockCenterOfMass': {'args': '<type __builtin__.bool>',
                                               'longname': 'lockCenterOfMass',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'lcm'},
                          'mass': {'args': '<type __builtin__.float>',
                                   'longname': 'mass',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'm'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'orientation': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'orientation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 3,
                                          'shortname': 'o'},
                          'particleCollision': {'args': '<type __builtin__.bool>',
                                                'longname': 'particleCollision',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'pc'},
                          'passive': {'args': '<type __builtin__.bool>',
                                      'longname': 'passive',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'pas'},
                          'position': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'position',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 3,
                                       'shortname': 'p'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'removeShape': {'args': '<type __builtin__.str>',
                                          'longname': 'removeShape',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'rs'},
                          'solver': {'args': '<type __builtin__.str>',
                                     'longname': 'solver',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'slv'},
                          'spinImpulse': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'spinImpulse',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 3,
                                          'shortname': 'si'},
                          'standInObject': {'args': '<type __builtin__.str>',
                                            'longname': 'standInObject',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sio'},
                          'staticFriction': {'args': '<type __builtin__.float>',
                                             'longname': 'staticFriction',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sf'},
                          'tesselationFactor': {'args': '<type __builtin__.int>',
                                                'longname': 'tesselationFactor',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'tf'},
                          'velocity': {'args': '<type __builtin__.bool>',
                                       'longname': 'velocity',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'vel'}},
                'shortFlags': {'act': 'active',
                               'afa': 'applyForceAt',
                               'av': 'angularVelocity',
                               'b': 'bounciness',
                               'c': 'cache',
                               'cc': 'contactCount',
                               'cl': 'collisions',
                               'cn': 'contactName',
                               'com': 'centerOfMass',
                               'cp': 'contactPosition',
                               'dc': 'deleteCache',
                               'df': 'dynamicFriction',
                               'dp': 'damping',
                               'e': 'edit',
                               'f': 'force',
                               'i': 'impulse',
                               'iav': 'initialAngularVelocity',
                               'ig': 'ignore',
                               'imp': 'impulsePosition',
                               'iv': 'initialVelocity',
                               'l': 'layer',
                               'lcm': 'lockCenterOfMass',
                               'm': 'mass',
                               'n': 'name',
                               'o': 'orientation',
                               'p': 'position',
                               'pas': 'passive',
                               'pc': 'particleCollision',
                               'q': 'query',
                               'rs': 'removeShape',
                               'sf': 'staticFriction',
                               'si': 'spinImpulse',
                               'sio': 'standInObject',
                               'slv': 'solver',
                               'tf': 'tesselationFactor',
                               'vel': 'velocity'},
                'type': 'effects'},
  'rigidSolver': {'flags': {'autoTolerances': {'args': '<type __builtin__.bool>',
                                               'longname': 'autoTolerances',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'at'},
                            'bounciness': {'args': '<type __builtin__.bool>',
                                           'longname': 'bounciness',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'b'},
                            'cacheData': {'args': '<type __builtin__.bool>',
                                          'longname': 'cacheData',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cd'},
                            'collide': {'args': '<type __builtin__.bool>',
                                        'longname': 'collide',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'c'},
                            'collisionTolerance': {'args': '<type __builtin__.float>',
                                                   'longname': 'collisionTolerance',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ct'},
                            'contactData': {'args': '<type __builtin__.bool>',
                                            'longname': 'contactData',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ctd'},
                            'create': {'args': '<type __builtin__.bool>',
                                       'longname': 'create',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'cr'},
                            'current': {'args': '<type __builtin__.bool>',
                                        'longname': 'current',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'cu'},
                            'deleteCache': {'args': '<type __builtin__.bool>',
                                            'longname': 'deleteCache',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'deleteCache'},
                            'displayCenterOfMass': {'args': '<type __builtin__.bool>',
                                                    'longname': 'displayCenterOfMass',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dcm'},
                            'displayConstraint': {'args': '<type __builtin__.bool>',
                                                  'longname': 'displayConstraint',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dc'},
                            'displayVelocity': {'args': '<type __builtin__.bool>',
                                                'longname': 'displayVelocity',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dv'},
                            'dynamics': {'args': '<type __builtin__.bool>',
                                         'longname': 'dynamics',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'd'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'friction': {'args': '<type __builtin__.bool>',
                                         'longname': 'friction',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'f'},
                            'interpenetrate': {'args': '<type __builtin__.bool>',
                                               'longname': 'interpenetrate',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'i'},
                            'interpenetrationCheck': {'args': '<type __builtin__.bool>',
                                                      'longname': 'interpenetrationCheck',
                                                      'modes': ['edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'ic'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'rigidBodies': {'args': '<type __builtin__.bool>',
                                            'longname': 'rigidBodies',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'rb'},
                            'rigidBodyCount': {'args': '<type __builtin__.bool>',
                                               'longname': 'rigidBodyCount',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'rbc'},
                            'showCollision': {'args': '<type __builtin__.bool>',
                                              'longname': 'showCollision',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'sc'},
                            'showInterpenetration': {'args': '<type __builtin__.bool>',
                                                     'longname': 'showInterpenetration',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'si'},
                            'solverMethod': {'args': '<type __builtin__.int>',
                                             'longname': 'solverMethod',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sm'},
                            'startTime': {'args': '<type __builtin__.float>',
                                          'longname': 'startTime',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'stt'},
                            'state': {'args': '<type __builtin__.bool>',
                                      'longname': 'state',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'st'},
                            'statistics': {'args': '<type __builtin__.bool>',
                                           'longname': 'statistics',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sta'},
                            'stepSize': {'args': '<type __builtin__.float>',
                                         'longname': 'stepSize',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 's'},
                            'velocityVectorScale': {'args': '<type __builtin__.float>',
                                                    'longname': 'velocityVectorScale',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'vs'}},
                  'removedFlags': {'del': 'deleteCache'},
                  'shortFlags': {'at': 'autoTolerances',
                                 'b': 'bounciness',
                                 'c': 'collide',
                                 'cd': 'cacheData',
                                 'cr': 'create',
                                 'ct': 'collisionTolerance',
                                 'ctd': 'contactData',
                                 'cu': 'current',
                                 'd': 'dynamics',
                                 'dc': 'displayConstraint',
                                 'dcm': 'displayCenterOfMass',
                                 'deleteCache': 'deleteCache',
                                 'dv': 'displayVelocity',
                                 'e': 'edit',
                                 'f': 'friction',
                                 'i': 'interpenetrate',
                                 'ic': 'interpenetrationCheck',
                                 'n': 'name',
                                 'q': 'query',
                                 'rb': 'rigidBodies',
                                 'rbc': 'rigidBodyCount',
                                 's': 'stepSize',
                                 'sc': 'showCollision',
                                 'si': 'showInterpenetration',
                                 'sm': 'solverMethod',
                                 'st': 'state',
                                 'sta': 'statistics',
                                 'stt': 'startTime',
                                 'vs': 'velocityVectorScale'},
                  'type': 'effects'},
  'roll': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                  'longname': 'absolute',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'abs'},
                     'degree': {'args': '<type __builtin__.float>',
                                'longname': 'degree',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'd'},
                     'relative': {'args': '<type __builtin__.bool>',
                                  'longname': 'relative',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'rel'}},
           'shortFlags': {'abs': 'absolute',
                          'd': 'degree',
                          'rel': 'relative'},
           'type': 'rendering'},
  'rollCtx': {'flags': {'alternateContext': {'args': '<type __builtin__.bool>',
                                             'longname': 'alternateContext',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'ac'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'exists': {'args': '<type __builtin__.bool>',
                                   'longname': 'exists',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'ex'},
                        'history': {'args': '<type __builtin__.bool>',
                                    'longname': 'history',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'ch'},
                        'image1': {'args': '<type __builtin__.str>',
                                   'longname': 'image1',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'i1'},
                        'image2': {'args': '<type __builtin__.str>',
                                   'longname': 'image2',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'i2'},
                        'image3': {'args': '<type __builtin__.str>',
                                   'longname': 'image3',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'i3'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'rollScale': {'args': '<type __builtin__.float>',
                                      'longname': 'rollScale',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'rs'},
                        'toolName': {'args': '<type __builtin__.str>',
                                     'longname': 'toolName',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'tn'}},
              'shortFlags': {'ac': 'alternateContext',
                             'ch': 'history',
                             'e': 'edit',
                             'ex': 'exists',
                             'i1': 'image1',
                             'i2': 'image2',
                             'i3': 'image3',
                             'n': 'name',
                             'q': 'query',
                             'rs': 'rollScale',
                             'tn': 'toolName'},
              'type': 'context'},
  'rotate': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                    'longname': 'absolute',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'a'},
                       'centerPivot': {'args': '<type __builtin__.bool>',
                                       'longname': 'centerPivot',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'cp'},
                       'componentSpace': {'args': '<type __builtin__.bool>',
                                          'longname': 'componentSpace',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'cs'},
                       'constrainAlongNormal': {'args': '<type __builtin__.bool>',
                                                'longname': 'constrainAlongNormal',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'xn'},
                       'deletePriorHistory': {'args': '<type __builtin__.bool>',
                                              'longname': 'deletePriorHistory',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dph'},
                       'euler': {'args': '<type __builtin__.bool>',
                                 'longname': 'euler',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'eu'},
                       'forceOrderXYZ': {'args': '<type __builtin__.bool>',
                                         'longname': 'forceOrderXYZ',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'fo'},
                       'objectCenterPivot': {'args': '<type __builtin__.bool>',
                                             'longname': 'objectCenterPivot',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ocp'},
                       'objectSpace': {'args': '<type __builtin__.bool>',
                                       'longname': 'objectSpace',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'os'},
                       'orientAxes': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'orientAxes',
                                      'modes': ['create'],
                                      'numArgs': 3,
                                      'shortname': 'oa'},
                       'pivot': {'args': ['<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>'],
                                 'longname': 'pivot',
                                 'modes': ['create'],
                                 'numArgs': 3,
                                 'shortname': 'p'},
                       'preserveChildPosition': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preserveChildPosition',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'pcp'},
                       'preserveGeometryPosition': {'args': '<type __builtin__.bool>',
                                                    'longname': 'preserveGeometryPosition',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'pgp'},
                       'preserveUV': {'args': '<type __builtin__.bool>',
                                      'longname': 'preserveUV',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'puv'},
                       'reflection': {'args': '<type __builtin__.bool>',
                                      'longname': 'reflection',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'rfl'},
                       'reflectionAboutBBox': {'args': '<type __builtin__.bool>',
                                               'longname': 'reflectionAboutBBox',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'rab'},
                       'reflectionAboutOrigin': {'args': '<type __builtin__.bool>',
                                                 'longname': 'reflectionAboutOrigin',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'rao'},
                       'reflectionAboutX': {'args': '<type __builtin__.bool>',
                                            'longname': 'reflectionAboutX',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'rax'},
                       'reflectionAboutY': {'args': '<type __builtin__.bool>',
                                            'longname': 'reflectionAboutY',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ray'},
                       'reflectionAboutZ': {'args': '<type __builtin__.bool>',
                                            'longname': 'reflectionAboutZ',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'raz'},
                       'reflectionTolerance': {'args': '<type __builtin__.float>',
                                               'longname': 'reflectionTolerance',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'rft'},
                       'relative': {'args': '<type __builtin__.bool>',
                                    'longname': 'relative',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'r'},
                       'rotateX': {'args': '<type __builtin__.bool>',
                                   'longname': 'rotateX',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'x'},
                       'rotateXY': {'args': '<type __builtin__.bool>',
                                    'longname': 'rotateXY',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'xy'},
                       'rotateXYZ': {'args': '<type __builtin__.bool>',
                                     'longname': 'rotateXYZ',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'xyz'},
                       'rotateXZ': {'args': '<type __builtin__.bool>',
                                    'longname': 'rotateXZ',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'xz'},
                       'rotateY': {'args': '<type __builtin__.bool>',
                                   'longname': 'rotateY',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'y'},
                       'rotateYZ': {'args': '<type __builtin__.bool>',
                                    'longname': 'rotateYZ',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'yz'},
                       'rotateZ': {'args': '<type __builtin__.bool>',
                                   'longname': 'rotateZ',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'z'},
                       'symNegative': {'args': '<type __builtin__.bool>',
                                       'longname': 'symNegative',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'smn'},
                       'translate': {'args': '<type __builtin__.bool>',
                                     'longname': 'translate',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 't'},
                       'worldSpace': {'args': '<type __builtin__.bool>',
                                      'longname': 'worldSpace',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ws'},
                       'xformConstraint': {'args': '<type __builtin__.str>',
                                           'longname': 'xformConstraint',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'xc'}},
             'shortFlags': {'a': 'absolute',
                            'cp': 'centerPivot',
                            'cs': 'componentSpace',
                            'dph': 'deletePriorHistory',
                            'eu': 'euler',
                            'fo': 'forceOrderXYZ',
                            'oa': 'orientAxes',
                            'ocp': 'objectCenterPivot',
                            'os': 'objectSpace',
                            'p': 'pivot',
                            'pcp': 'preserveChildPosition',
                            'pgp': 'preserveGeometryPosition',
                            'puv': 'preserveUV',
                            'r': 'relative',
                            'rab': 'reflectionAboutBBox',
                            'rao': 'reflectionAboutOrigin',
                            'rax': 'reflectionAboutX',
                            'ray': 'reflectionAboutY',
                            'raz': 'reflectionAboutZ',
                            'rfl': 'reflection',
                            'rft': 'reflectionTolerance',
                            'smn': 'symNegative',
                            't': 'translate',
                            'ws': 'worldSpace',
                            'x': 'rotateX',
                            'xc': 'xformConstraint',
                            'xn': 'constrainAlongNormal',
                            'xy': 'rotateXY',
                            'xyz': 'rotateXYZ',
                            'xz': 'rotateXZ',
                            'y': 'rotateY',
                            'yz': 'rotateYZ',
                            'z': 'rotateZ'},
             'type': 'general'},
  'rotationInterpolation': {'flags': {'convert': {'args': '<type __builtin__.str>',
                                                  'longname': 'convert',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'c'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'}},
                            'shortFlags': {'c': 'convert', 'q': 'query'},
                            'type': 'animation'},
  'roundCRCtx': {'flags': {'activeNodes': {'args': '<type __builtin__.bool>',
                                           'longname': 'activeNodes',
                                           'numArgs': 0,
                                           'shortname': 'anq'},
                           'append': {'args': '<type __builtin__.bool>',
                                      'longname': 'append',
                                      'numArgs': 1,
                                      'shortname': 'a'},
                           'autoCreate': {'args': '<type __builtin__.bool>',
                                          'longname': 'autoCreate',
                                          'numArgs': 1,
                                          'shortname': 'ac'},
                           'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'currentEdge': {'args': '<type __builtin__.bool>',
                                           'longname': 'currentEdge',
                                           'numArgs': 0,
                                           'shortname': 'ce'},
                           'currentEdgeRadius': {'args': '<type __builtin__.bool>',
                                                 'longname': 'currentEdgeRadius',
                                                 'numArgs': 0,
                                                 'shortname': 'cer'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'immediate': {'args': '<type __builtin__.bool>',
                                         'longname': 'immediate',
                                         'numArgs': 0,
                                         'shortname': 'im'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'object': {'args': '<type __builtin__.bool>',
                                      'longname': 'object',
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'radius': {'args': '<type __builtin__.float>',
                                      'longname': 'radius',
                                      'modes': ['multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'r'},
                           'radiusToUse': {'args': '<type __builtin__.float>',
                                           'longname': 'radiusToUse',
                                           'numArgs': 1,
                                           'shortname': 'rtu'},
                           'reset': {'args': '<type __builtin__.bool>',
                                     'longname': 'reset',
                                     'numArgs': 0,
                                     'shortname': 'rs'},
                           'tolerance': {'args': '<type __builtin__.float>',
                                         'longname': 'tolerance',
                                         'numArgs': 1,
                                         'shortname': 'tol'},
                           'toolNode': {'args': '<type __builtin__.bool>',
                                        'longname': 'toolNode',
                                        'numArgs': 0,
                                        'shortname': 'tnq'},
                           'useGlobalTolerance': {'args': '<type __builtin__.bool>',
                                                  'longname': 'useGlobalTolerance',
                                                  'numArgs': 1,
                                                  'shortname': 'ugt'}},
                 'shortFlags': {'a': 'append',
                                'ac': 'autoCreate',
                                'anq': 'activeNodes',
                                'cch': 'caching',
                                'ce': 'currentEdge',
                                'cer': 'currentEdgeRadius',
                                'ch': 'constructionHistory',
                                'e': 'edit',
                                'ex': 'exists',
                                'fzn': 'frozen',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'im': 'immediate',
                                'n': 'name',
                                'nds': 'nodeState',
                                'o': 'object',
                                'q': 'query',
                                'r': 'radius',
                                'rs': 'reset',
                                'rtu': 'radiusToUse',
                                'tnq': 'toolNode',
                                'tol': 'tolerance',
                                'ugt': 'useGlobalTolerance'},
                 'type': 'context'},
  'roundConstantRadius': {'flags': {'append': {'args': '<type __builtin__.bool>',
                                               'longname': 'append',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'a'},
                                    'caching': {'args': '<type __builtin__.bool>',
                                                'longname': 'caching',
                                                'numArgs': 1,
                                                'shortname': 'cch'},
                                    'constructionHistory': {'args': '<type __builtin__.bool>',
                                                            'longname': 'constructionHistory',
                                                            'modes': ['create'],
                                                            'numArgs': 1,
                                                            'shortname': 'ch'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'frozen': {'args': '<type __builtin__.bool>',
                                               'longname': 'frozen',
                                               'numArgs': 1,
                                               'shortname': 'fzn'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'nodeState': {'args': '<type __builtin__.int>',
                                                  'longname': 'nodeState',
                                                  'numArgs': 1,
                                                  'shortname': 'nds'},
                                    'object': {'args': '<type __builtin__.bool>',
                                               'longname': 'object',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'o'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'radius': {'args': '<type __builtin__.float>',
                                               'longname': 'radius',
                                               'modes': ['multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'r'},
                                    'radiuss': {'args': '<type __builtin__.float>',
                                                'longname': 'radiuss',
                                                'modes': ['create',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'rad'},
                                    'side': {'args': ['<type __builtin__.str>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'side',
                                             'modes': ['create',
                                                       'multiuse'],
                                             'numArgs': 2,
                                             'shortname': 's'},
                                    'sidea': {'args': '<type __builtin__.int>',
                                              'longname': 'sidea',
                                              'modes': ['create',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'sa'},
                                    'sideb': {'args': '<type __builtin__.int>',
                                              'longname': 'sideb',
                                              'modes': ['create',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'sb'},
                                    'tolerance': {'args': '<type __builtin__.float>',
                                                  'longname': 'tolerance',
                                                  'numArgs': 1,
                                                  'shortname': 'tol'}},
                          'shortFlags': {'a': 'append',
                                         'cch': 'caching',
                                         'ch': 'constructionHistory',
                                         'e': 'edit',
                                         'fzn': 'frozen',
                                         'n': 'name',
                                         'nds': 'nodeState',
                                         'o': 'object',
                                         'q': 'query',
                                         'r': 'radius',
                                         'rad': 'radiuss',
                                         's': 'side',
                                         'sa': 'sidea',
                                         'sb': 'sideb',
                                         'tol': 'tolerance'},
                          'type': 'modeling'},
  'rowColumnLayout': {'callbackFlags': ['visibleChangeCommand',
                                        'vcc',
                                        'dragCallback',
                                        'dgc',
                                        'dropCallback',
                                        'dpc'],
                      'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                     'longname': 'adjustableColumn',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'adj'},
                                'annotation': {'args': '<type __builtin__.str>',
                                               'longname': 'annotation',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ann'},
                                'backgroundColor': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'backgroundColor',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'bgc'},
                                'childArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'childArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'ca'},
                                'columnAlign': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'columnAlign',
                                                'modes': ['create',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 2,
                                                'shortname': 'cal'},
                                'columnAttach': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnAttach',
                                                 'modes': ['create',
                                                           'edit',
                                                           'multiuse'],
                                                 'numArgs': 3,
                                                 'shortname': 'cat'},
                                'columnOffset': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'columnOffset',
                                                 'modes': ['create',
                                                           'edit',
                                                           'multiuse'],
                                                 'numArgs': 3,
                                                 'shortname': 'co'},
                                'columnSpacing': {'args': ['<type __builtin__.int>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'columnSpacing',
                                                  'modes': ['create',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 2,
                                                  'shortname': 'cs'},
                                'columnWidth': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'columnWidth',
                                                'modes': ['create',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 2,
                                                'shortname': 'cw'},
                                'defineTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'defineTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'dt'},
                                'docTag': {'args': '<type __builtin__.str>',
                                           'longname': 'docTag',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dtg'},
                                'dragCallback': {'args': 'script',
                                                 'longname': 'dragCallback',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dgc'},
                                'dropCallback': {'args': 'script',
                                                 'longname': 'dropCallback',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dpc'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'enable': {'args': '<type __builtin__.bool>',
                                           'longname': 'enable',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'en'},
                                'enableBackground': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableBackground',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ebg'},
                                'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                        'longname': 'enableKeyboardFocus',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ekf'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'fullPathName': {'args': '<type __builtin__.bool>',
                                                 'longname': 'fullPathName',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'fpn'},
                                'height': {'args': '<type __builtin__.int>',
                                           'longname': 'height',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'h'},
                                'highlightColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'highlightColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'hlc'},
                                'isObscured': {'args': '<type __builtin__.bool>',
                                               'longname': 'isObscured',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'io'},
                                'manage': {'args': '<type __builtin__.bool>',
                                           'longname': 'manage',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'm'},
                                'noBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'noBackground',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'nbg'},
                                'numberOfChildren': {'args': '<type __builtin__.bool>',
                                                     'longname': 'numberOfChildren',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'nch'},
                                'numberOfColumns': {'args': '<type __builtin__.int>',
                                                    'longname': 'numberOfColumns',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'nc'},
                                'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                       'longname': 'numberOfPopupMenus',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'npm'},
                                'numberOfRows': {'args': '<type __builtin__.int>',
                                                 'longname': 'numberOfRows',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'nr'},
                                'parent': {'args': '<type __builtin__.str>',
                                           'longname': 'parent',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'p'},
                                'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                   'longname': 'popupMenuArray',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'pma'},
                                'preventOverride': {'args': '<type __builtin__.bool>',
                                                    'longname': 'preventOverride',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'po'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'rowAlign': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'rowAlign',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 2,
                                             'shortname': 'ral'},
                                'rowAttach': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'rowAttach',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 3,
                                              'shortname': 'rat'},
                                'rowHeight': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'rowHeight',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'rh'},
                                'rowOffset': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'rowOffset',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 3,
                                              'shortname': 'ro'},
                                'rowSpacing': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'rowSpacing',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 2,
                                               'shortname': 'rs'},
                                'statusBarMessage': {'args': '<type __builtin__.str>',
                                                     'longname': 'statusBarMessage',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sbm'},
                                'useTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'useTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ut'},
                                'visible': {'args': '<type __builtin__.bool>',
                                            'longname': 'visible',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'vis'},
                                'visibleChangeCommand': {'args': 'script',
                                                         'longname': 'visibleChangeCommand',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'vcc'},
                                'width': {'args': '<type __builtin__.int>',
                                          'longname': 'width',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'w'}},
                      'shortFlags': {'adj': 'adjustableColumn',
                                     'ann': 'annotation',
                                     'bgc': 'backgroundColor',
                                     'ca': 'childArray',
                                     'cal': 'columnAlign',
                                     'cat': 'columnAttach',
                                     'co': 'columnOffset',
                                     'cs': 'columnSpacing',
                                     'cw': 'columnWidth',
                                     'dgc': 'dragCallback',
                                     'dpc': 'dropCallback',
                                     'dt': 'defineTemplate',
                                     'dtg': 'docTag',
                                     'e': 'edit',
                                     'ebg': 'enableBackground',
                                     'ekf': 'enableKeyboardFocus',
                                     'en': 'enable',
                                     'ex': 'exists',
                                     'fpn': 'fullPathName',
                                     'h': 'height',
                                     'hlc': 'highlightColor',
                                     'io': 'isObscured',
                                     'm': 'manage',
                                     'nbg': 'noBackground',
                                     'nc': 'numberOfColumns',
                                     'nch': 'numberOfChildren',
                                     'npm': 'numberOfPopupMenus',
                                     'nr': 'numberOfRows',
                                     'p': 'parent',
                                     'pma': 'popupMenuArray',
                                     'po': 'preventOverride',
                                     'q': 'query',
                                     'ral': 'rowAlign',
                                     'rat': 'rowAttach',
                                     'rh': 'rowHeight',
                                     'ro': 'rowOffset',
                                     'rs': 'rowSpacing',
                                     'sbm': 'statusBarMessage',
                                     'ut': 'useTemplate',
                                     'vcc': 'visibleChangeCommand',
                                     'vis': 'visible',
                                     'w': 'width'},
                      'type': 'windows'},
  'rowLayout': {'callbackFlags': ['visibleChangeCommand',
                                  'vcc',
                                  'dragCallback',
                                  'dgc',
                                  'dropCallback',
                                  'dpc'],
                'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                               'longname': 'adjustableColumn',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'adj'},
                          'adjustableColumn1': {'args': '<type __builtin__.int>',
                                                'longname': 'adjustableColumn1',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ad1'},
                          'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                'longname': 'adjustableColumn2',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ad2'},
                          'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                'longname': 'adjustableColumn3',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ad3'},
                          'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                'longname': 'adjustableColumn4',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ad4'},
                          'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                'longname': 'adjustableColumn5',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ad5'},
                          'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                'longname': 'adjustableColumn6',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ad6'},
                          'annotation': {'args': '<type __builtin__.str>',
                                         'longname': 'annotation',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ann'},
                          'backgroundColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'backgroundColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'bgc'},
                          'childArray': {'args': '<type __builtin__.bool>',
                                         'longname': 'childArray',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'ca'},
                          'columnAlign': {'args': ['<type __builtin__.int>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'columnAlign',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'cal'},
                          'columnAlign1': {'args': '<type __builtin__.str>',
                                           'longname': 'columnAlign1',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'cl1'},
                          'columnAlign2': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'columnAlign2',
                                           'modes': ['create'],
                                           'numArgs': 2,
                                           'shortname': 'cl2'},
                          'columnAlign3': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'columnAlign3',
                                           'modes': ['create'],
                                           'numArgs': 3,
                                           'shortname': 'cl3'},
                          'columnAlign4': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'columnAlign4',
                                           'modes': ['create'],
                                           'numArgs': 4,
                                           'shortname': 'cl4'},
                          'columnAlign5': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'columnAlign5',
                                           'modes': ['create'],
                                           'numArgs': 5,
                                           'shortname': 'cl5'},
                          'columnAlign6': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'columnAlign6',
                                           'modes': ['create'],
                                           'numArgs': 6,
                                           'shortname': 'cl6'},
                          'columnAttach': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'columnAttach',
                                           'modes': ['create',
                                                     'edit',
                                                     'multiuse'],
                                           'numArgs': 3,
                                           'shortname': 'cat'},
                          'columnAttach1': {'args': '<type __builtin__.str>',
                                            'longname': 'columnAttach1',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ct1'},
                          'columnAttach2': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'columnAttach2',
                                            'modes': ['create'],
                                            'numArgs': 2,
                                            'shortname': 'ct2'},
                          'columnAttach3': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'columnAttach3',
                                            'modes': ['create'],
                                            'numArgs': 3,
                                            'shortname': 'ct3'},
                          'columnAttach4': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'columnAttach4',
                                            'modes': ['create'],
                                            'numArgs': 4,
                                            'shortname': 'ct4'},
                          'columnAttach5': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'columnAttach5',
                                            'modes': ['create'],
                                            'numArgs': 5,
                                            'shortname': 'ct5'},
                          'columnAttach6': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.str>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'columnAttach6',
                                            'modes': ['create'],
                                            'numArgs': 6,
                                            'shortname': 'ct6'},
                          'columnOffset1': {'args': '<type __builtin__.int>',
                                            'longname': 'columnOffset1',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'co1'},
                          'columnOffset2': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'columnOffset2',
                                            'modes': ['create'],
                                            'numArgs': 2,
                                            'shortname': 'co2'},
                          'columnOffset3': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'columnOffset3',
                                            'modes': ['create'],
                                            'numArgs': 3,
                                            'shortname': 'co3'},
                          'columnOffset4': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'columnOffset4',
                                            'modes': ['create'],
                                            'numArgs': 4,
                                            'shortname': 'co4'},
                          'columnOffset5': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'columnOffset5',
                                            'modes': ['create'],
                                            'numArgs': 5,
                                            'shortname': 'co5'},
                          'columnOffset6': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'columnOffset6',
                                            'modes': ['create'],
                                            'numArgs': 6,
                                            'shortname': 'co6'},
                          'columnWidth': {'args': ['<type __builtin__.int>',
                                                   '<type __builtin__.int>'],
                                          'longname': 'columnWidth',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'cw'},
                          'columnWidth1': {'args': '<type __builtin__.int>',
                                           'longname': 'columnWidth1',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'cw1'},
                          'columnWidth2': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'columnWidth2',
                                           'modes': ['create'],
                                           'numArgs': 2,
                                           'shortname': 'cw2'},
                          'columnWidth3': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.int>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'columnWidth3',
                                           'modes': ['create'],
                                           'numArgs': 3,
                                           'shortname': 'cw3'},
                          'columnWidth4': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.int>',
                                                    '<type __builtin__.int>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'columnWidth4',
                                           'modes': ['create'],
                                           'numArgs': 4,
                                           'shortname': 'cw4'},
                          'columnWidth5': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.int>',
                                                    '<type __builtin__.int>',
                                                    '<type __builtin__.int>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'columnWidth5',
                                           'modes': ['create'],
                                           'numArgs': 5,
                                           'shortname': 'cw5'},
                          'columnWidth6': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.int>',
                                                    '<type __builtin__.int>',
                                                    '<type __builtin__.int>',
                                                    '<type __builtin__.int>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'columnWidth6',
                                           'modes': ['create'],
                                           'numArgs': 6,
                                           'shortname': 'cw6'},
                          'defineTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'defineTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dt'},
                          'docTag': {'args': '<type __builtin__.str>',
                                     'longname': 'docTag',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'dtg'},
                          'dragCallback': {'args': 'script',
                                           'longname': 'dragCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dgc'},
                          'dropCallback': {'args': 'script',
                                           'longname': 'dropCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dpc'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'enable': {'args': '<type __builtin__.bool>',
                                     'longname': 'enable',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'en'},
                          'enableBackground': {'args': '<type __builtin__.bool>',
                                               'longname': 'enableBackground',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ebg'},
                          'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableKeyboardFocus',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ekf'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'fullPathName': {'args': '<type __builtin__.bool>',
                                           'longname': 'fullPathName',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'fpn'},
                          'height': {'args': '<type __builtin__.int>',
                                     'longname': 'height',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'h'},
                          'highlightColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'highlightColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'hlc'},
                          'isObscured': {'args': '<type __builtin__.bool>',
                                         'longname': 'isObscured',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'io'},
                          'manage': {'args': '<type __builtin__.bool>',
                                     'longname': 'manage',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'm'},
                          'noBackground': {'args': '<type __builtin__.bool>',
                                           'longname': 'noBackground',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nbg'},
                          'numberOfChildren': {'args': '<type __builtin__.bool>',
                                               'longname': 'numberOfChildren',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'nch'},
                          'numberOfColumns': {'args': '<type __builtin__.int>',
                                              'longname': 'numberOfColumns',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'nc'},
                          'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'numberOfPopupMenus',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'npm'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                          'popupMenuArray': {'args': '<type __builtin__.bool>',
                                             'longname': 'popupMenuArray',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'pma'},
                          'preventOverride': {'args': '<type __builtin__.bool>',
                                              'longname': 'preventOverride',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'po'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'rowAttach': {'args': ['<type __builtin__.int>',
                                                 '<type __builtin__.str>',
                                                 '<type __builtin__.int>'],
                                        'longname': 'rowAttach',
                                        'modes': ['create',
                                                  'edit',
                                                  'multiuse'],
                                        'numArgs': 3,
                                        'shortname': 'rat'},
                          'statusBarMessage': {'args': '<type __builtin__.str>',
                                               'longname': 'statusBarMessage',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sbm'},
                          'useTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'useTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ut'},
                          'visible': {'args': '<type __builtin__.bool>',
                                      'longname': 'visible',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'vis'},
                          'visibleChangeCommand': {'args': 'script',
                                                   'longname': 'visibleChangeCommand',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'vcc'},
                          'width': {'args': '<type __builtin__.int>',
                                    'longname': 'width',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'w'}},
                'shortFlags': {'ad1': 'adjustableColumn1',
                               'ad2': 'adjustableColumn2',
                               'ad3': 'adjustableColumn3',
                               'ad4': 'adjustableColumn4',
                               'ad5': 'adjustableColumn5',
                               'ad6': 'adjustableColumn6',
                               'adj': 'adjustableColumn',
                               'ann': 'annotation',
                               'bgc': 'backgroundColor',
                               'ca': 'childArray',
                               'cal': 'columnAlign',
                               'cat': 'columnAttach',
                               'cl1': 'columnAlign1',
                               'cl2': 'columnAlign2',
                               'cl3': 'columnAlign3',
                               'cl4': 'columnAlign4',
                               'cl5': 'columnAlign5',
                               'cl6': 'columnAlign6',
                               'co1': 'columnOffset1',
                               'co2': 'columnOffset2',
                               'co3': 'columnOffset3',
                               'co4': 'columnOffset4',
                               'co5': 'columnOffset5',
                               'co6': 'columnOffset6',
                               'ct1': 'columnAttach1',
                               'ct2': 'columnAttach2',
                               'ct3': 'columnAttach3',
                               'ct4': 'columnAttach4',
                               'ct5': 'columnAttach5',
                               'ct6': 'columnAttach6',
                               'cw': 'columnWidth',
                               'cw1': 'columnWidth1',
                               'cw2': 'columnWidth2',
                               'cw3': 'columnWidth3',
                               'cw4': 'columnWidth4',
                               'cw5': 'columnWidth5',
                               'cw6': 'columnWidth6',
                               'dgc': 'dragCallback',
                               'dpc': 'dropCallback',
                               'dt': 'defineTemplate',
                               'dtg': 'docTag',
                               'e': 'edit',
                               'ebg': 'enableBackground',
                               'ekf': 'enableKeyboardFocus',
                               'en': 'enable',
                               'ex': 'exists',
                               'fpn': 'fullPathName',
                               'h': 'height',
                               'hlc': 'highlightColor',
                               'io': 'isObscured',
                               'm': 'manage',
                               'nbg': 'noBackground',
                               'nc': 'numberOfColumns',
                               'nch': 'numberOfChildren',
                               'npm': 'numberOfPopupMenus',
                               'p': 'parent',
                               'pma': 'popupMenuArray',
                               'po': 'preventOverride',
                               'q': 'query',
                               'rat': 'rowAttach',
                               'sbm': 'statusBarMessage',
                               'ut': 'useTemplate',
                               'vcc': 'visibleChangeCommand',
                               'vis': 'visible',
                               'w': 'width'},
                'type': 'windows'},
  'runTimeCommand': {'callbackFlags': ['numberOfDefaultCommands',
                                       'ndc',
                                       'commandArray',
                                       'ca',
                                       'commandLanguage',
                                       'cl',
                                       'numberOfCommands',
                                       'nc',
                                       'numberOfUserCommands',
                                       'nuc',
                                       'defaultCommandArray',
                                       'dca',
                                       'userCommandArray',
                                       'uca',
                                       'command',
                                       'c'],
                     'flags': {'annotation': {'args': '<type __builtin__.str>',
                                              'longname': 'annotation',
                                              'numArgs': 1,
                                              'shortname': 'ann'},
                               'category': {'args': '<type __builtin__.str>',
                                            'longname': 'category',
                                            'numArgs': 1,
                                            'shortname': 'cat'},
                               'categoryArray': {'args': '<type __builtin__.bool>',
                                                 'longname': 'categoryArray',
                                                 'numArgs': 0,
                                                 'shortname': 'caa'},
                               'command': {'args': '<type __builtin__.callable>',
                                           'longname': 'command',
                                           'numArgs': 1,
                                           'shortname': 'c'},
                               'commandArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'commandArray',
                                                'numArgs': 0,
                                                'shortname': 'ca'},
                               'commandLanguage': {'args': '<type __builtin__.str>',
                                                   'longname': 'commandLanguage',
                                                   'numArgs': 1,
                                                   'shortname': 'cl'},
                               'default': {'args': '<type __builtin__.bool>',
                                           'longname': 'default',
                                           'numArgs': 1,
                                           'shortname': 'd'},
                               'defaultCommandArray': {'args': '<type __builtin__.bool>',
                                                       'longname': 'defaultCommandArray',
                                                       'numArgs': 0,
                                                       'shortname': 'dca'},
                               'delete': {'args': '<type __builtin__.bool>',
                                          'longname': 'delete',
                                          'numArgs': 0,
                                          'shortname': 'delete'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'hotkeyCtx': {'args': '<type __builtin__.str>',
                                             'longname': 'hotkeyCtx',
                                             'numArgs': 1,
                                             'shortname': 'hc'},
                               'image': {'args': '<type __builtin__.str>',
                                         'longname': 'image',
                                         'numArgs': 1,
                                         'shortname': 'i'},
                               'keywords': {'args': '<type __builtin__.str>',
                                            'longname': 'keywords',
                                            'numArgs': 1,
                                            'shortname': 'k'},
                               'label': {'args': '<type __builtin__.str>',
                                         'longname': 'label',
                                         'numArgs': 1,
                                         'shortname': 'l'},
                               'longAnnotation': {'args': '<type __builtin__.str>',
                                                  'longname': 'longAnnotation',
                                                  'numArgs': 1,
                                                  'shortname': 'la'},
                               'numberOfCommands': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfCommands',
                                                    'numArgs': 0,
                                                    'shortname': 'nc'},
                               'numberOfDefaultCommands': {'args': '<type __builtin__.bool>',
                                                           'longname': 'numberOfDefaultCommands',
                                                           'numArgs': 0,
                                                           'shortname': 'ndc'},
                               'numberOfUserCommands': {'args': '<type __builtin__.bool>',
                                                        'longname': 'numberOfUserCommands',
                                                        'numArgs': 0,
                                                        'shortname': 'nuc'},
                               'plugin': {'args': '<type __builtin__.str>',
                                          'longname': 'plugin',
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'save': {'args': '<type __builtin__.bool>',
                                        'longname': 'save',
                                        'numArgs': 0,
                                        'shortname': 's'},
                               'showInHotkeyEditor': {'args': '<type __builtin__.bool>',
                                                      'longname': 'showInHotkeyEditor',
                                                      'numArgs': 1,
                                                      'shortname': 'she'},
                               'tags': {'args': '<type __builtin__.str>',
                                        'longname': 'tags',
                                        'numArgs': 1,
                                        'shortname': 't'},
                               'userCommandArray': {'args': '<type __builtin__.bool>',
                                                    'longname': 'userCommandArray',
                                                    'numArgs': 0,
                                                    'shortname': 'uca'}},
                     'removedFlags': {'del': 'delete'},
                     'shortFlags': {'ann': 'annotation',
                                    'c': 'command',
                                    'ca': 'commandArray',
                                    'caa': 'categoryArray',
                                    'cat': 'category',
                                    'cl': 'commandLanguage',
                                    'd': 'default',
                                    'dca': 'defaultCommandArray',
                                    'delete': 'delete',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'hc': 'hotkeyCtx',
                                    'i': 'image',
                                    'k': 'keywords',
                                    'l': 'label',
                                    'la': 'longAnnotation',
                                    'nc': 'numberOfCommands',
                                    'ndc': 'numberOfDefaultCommands',
                                    'nuc': 'numberOfUserCommands',
                                    'p': 'plugin',
                                    'q': 'query',
                                    's': 'save',
                                    'she': 'showInHotkeyEditor',
                                    't': 'tags',
                                    'uca': 'userCommandArray'},
                     'type': 'windows'},
  'runup': {'flags': {'cache': {'args': '<type __builtin__.bool>',
                                'longname': 'cache',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'cch'},
                      'fromPreviousFrame': {'args': '<type __builtin__.bool>',
                                            'longname': 'fromPreviousFrame',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'fpf'},
                      'fromStartFrame': {'args': '<type __builtin__.bool>',
                                         'longname': 'fromStartFrame',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'fsf'},
                      'maxFrame': {'args': 'time',
                                   'longname': 'maxFrame',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'mxf'},
                      'state': {'args': '<type __builtin__.bool>',
                                'longname': 'state',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'st'}},
            'shortFlags': {'cch': 'cache',
                           'fpf': 'fromPreviousFrame',
                           'fsf': 'fromStartFrame',
                           'mxf': 'maxFrame',
                           'st': 'state'},
            'type': 'effects'},
  'sampleImage': {'flags': {'fastSample': {'args': '<type __builtin__.bool>',
                                           'longname': 'fastSample',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'f'},
                            'resolution': {'args': ['<type __builtin__.int>',
                                                    'PyNode'],
                                           'longname': 'resolution',
                                           'modes': ['create'],
                                           'numArgs': 2,
                                           'shortname': 'r'}},
                  'shortFlags': {'f': 'fastSample', 'r': 'resolution'},
                  'type': 'rendering'},
  'saveAllShelves': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'saveFluid': {'flags': {'currentTime': {'args': '<type __builtin__.int>',
                                          'longname': 'currentTime',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ct'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'endTime': {'args': '<type __builtin__.int>',
                                      'longname': 'endTime',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'et'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'startTime': {'args': '<type __builtin__.int>',
                                        'longname': 'startTime',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'st'}},
                'shortFlags': {'ct': 'currentTime',
                               'e': 'edit',
                               'et': 'endTime',
                               'q': 'query',
                               'st': 'startTime'},
                'type': 'effects'},
  'saveImage': {'callbackFlags': ['visibleChangeCommand',
                                  'vcc',
                                  'dragCallback',
                                  'dgc',
                                  'dropCallback',
                                  'dpc'],
                'flags': {'annotation': {'args': '<type __builtin__.str>',
                                         'longname': 'annotation',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ann'},
                          'backgroundColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'backgroundColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'bgc'},
                          'currentView': {'args': '<type __builtin__.bool>',
                                          'longname': 'currentView',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'cv'},
                          'defineTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'defineTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dt'},
                          'docTag': {'args': '<type __builtin__.str>',
                                     'longname': 'docTag',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'dtg'},
                          'dragCallback': {'args': 'script',
                                           'longname': 'dragCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dgc'},
                          'dropCallback': {'args': 'script',
                                           'longname': 'dropCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dpc'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'enable': {'args': '<type __builtin__.bool>',
                                     'longname': 'enable',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'en'},
                          'enableBackground': {'args': '<type __builtin__.bool>',
                                               'longname': 'enableBackground',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ebg'},
                          'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableKeyboardFocus',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ekf'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'fullPathName': {'args': '<type __builtin__.bool>',
                                           'longname': 'fullPathName',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'fpn'},
                          'height': {'args': '<type __builtin__.int>',
                                     'longname': 'height',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'h'},
                          'highlightColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'highlightColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'hlc'},
                          'image': {'args': '<type __builtin__.str>',
                                    'longname': 'image',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i'},
                          'isObscured': {'args': '<type __builtin__.bool>',
                                         'longname': 'isObscured',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'io'},
                          'manage': {'args': '<type __builtin__.bool>',
                                     'longname': 'manage',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'm'},
                          'noBackground': {'args': '<type __builtin__.bool>',
                                           'longname': 'noBackground',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nbg'},
                          'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'numberOfPopupMenus',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'npm'},
                          'objectThumbnail': {'args': '<type __builtin__.str>',
                                              'longname': 'objectThumbnail',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'ot'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                          'popupMenuArray': {'args': '<type __builtin__.bool>',
                                             'longname': 'popupMenuArray',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'pma'},
                          'preventOverride': {'args': '<type __builtin__.bool>',
                                              'longname': 'preventOverride',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'po'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'sceneFile': {'args': '<type __builtin__.str>',
                                        'longname': 'sceneFile',
                                        'modes': ['edit'],
                                        'numArgs': 1,
                                        'shortname': 'sf'},
                          'statusBarMessage': {'args': '<type __builtin__.str>',
                                               'longname': 'statusBarMessage',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sbm'},
                          'useTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'useTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ut'},
                          'visible': {'args': '<type __builtin__.bool>',
                                      'longname': 'visible',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'vis'},
                          'visibleChangeCommand': {'args': 'script',
                                                   'longname': 'visibleChangeCommand',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'vcc'},
                          'width': {'args': '<type __builtin__.int>',
                                    'longname': 'width',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'w'}},
                'shortFlags': {'ann': 'annotation',
                               'bgc': 'backgroundColor',
                               'cv': 'currentView',
                               'dgc': 'dragCallback',
                               'dpc': 'dropCallback',
                               'dt': 'defineTemplate',
                               'dtg': 'docTag',
                               'e': 'edit',
                               'ebg': 'enableBackground',
                               'ekf': 'enableKeyboardFocus',
                               'en': 'enable',
                               'ex': 'exists',
                               'fpn': 'fullPathName',
                               'h': 'height',
                               'hlc': 'highlightColor',
                               'i': 'image',
                               'io': 'isObscured',
                               'm': 'manage',
                               'nbg': 'noBackground',
                               'npm': 'numberOfPopupMenus',
                               'ot': 'objectThumbnail',
                               'p': 'parent',
                               'pma': 'popupMenuArray',
                               'po': 'preventOverride',
                               'q': 'query',
                               'sbm': 'statusBarMessage',
                               'sf': 'sceneFile',
                               'ut': 'useTemplate',
                               'vcc': 'visibleChangeCommand',
                               'vis': 'visible',
                               'w': 'width'},
                'type': 'system'},
  'saveInitialState': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                               'longname': 'attribute',
                                               'modes': ['create',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'atr'},
                                 'saveall': {'args': '<type __builtin__.bool>',
                                             'longname': 'saveall',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'all'}},
                       'shortFlags': {'all': 'saveall', 'atr': 'attribute'},
                       'type': 'effects'},
  'saveMenu': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'savePrefObjects': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'savePrefs': {'flags': {'colors': {'args': '<type __builtin__.bool>',
                                     'longname': 'colors',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'c'},
                          'general': {'args': '<type __builtin__.bool>',
                                      'longname': 'general',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'g'},
                          'hotkeys': {'args': '<type __builtin__.bool>',
                                      'longname': 'hotkeys',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'hk'},
                          'menuSets': {'args': '<type __builtin__.bool>',
                                       'longname': 'menuSets',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ms'},
                          'plugins': {'args': '<type __builtin__.bool>',
                                      'longname': 'plugins',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'pl'},
                          'uiLayout': {'args': '<type __builtin__.bool>',
                                       'longname': 'uiLayout',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ui'}},
                'shortFlags': {'c': 'colors',
                               'g': 'general',
                               'hk': 'hotkeys',
                               'ms': 'menuSets',
                               'pl': 'plugins',
                               'ui': 'uiLayout'},
                'type': 'windows'},
  'saveShelf': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'saveToolSettings': {'flags': {}, 'shortFlags': {}, 'type': 'general'},
  'saveViewportSettings': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'scale': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                   'longname': 'absolute',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'a'},
                      'centerPivot': {'args': '<type __builtin__.bool>',
                                      'longname': 'centerPivot',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'cp'},
                      'componentSpace': {'args': '<type __builtin__.bool>',
                                         'longname': 'componentSpace',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'cs'},
                      'constrainAlongNormal': {'args': '<type __builtin__.bool>',
                                               'longname': 'constrainAlongNormal',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'xn'},
                      'deletePriorHistory': {'args': '<type __builtin__.bool>',
                                             'longname': 'deletePriorHistory',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dph'},
                      'distanceOnly': {'args': '<type __builtin__.bool>',
                                       'longname': 'distanceOnly',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'dso'},
                      'localSpace': {'args': '<type __builtin__.bool>',
                                     'longname': 'localSpace',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ls'},
                      'objectCenterPivot': {'args': '<type __builtin__.bool>',
                                            'longname': 'objectCenterPivot',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ocp'},
                      'objectSpace': {'args': '<type __builtin__.bool>',
                                      'longname': 'objectSpace',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'os'},
                      'orientAxes': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'orientAxes',
                                     'modes': ['create'],
                                     'numArgs': 3,
                                     'shortname': 'oa'},
                      'pivot': {'args': ['<type __builtin__.float>',
                                         '<type __builtin__.float>',
                                         '<type __builtin__.float>'],
                                'longname': 'pivot',
                                'modes': ['create'],
                                'numArgs': 3,
                                'shortname': 'p'},
                      'preserveChildPosition': {'args': '<type __builtin__.bool>',
                                                'longname': 'preserveChildPosition',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'pcp'},
                      'preserveGeometryPosition': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preserveGeometryPosition',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'pgp'},
                      'preserveUV': {'args': '<type __builtin__.bool>',
                                     'longname': 'preserveUV',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'puv'},
                      'reflection': {'args': '<type __builtin__.bool>',
                                     'longname': 'reflection',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'rfl'},
                      'reflectionAboutBBox': {'args': '<type __builtin__.bool>',
                                              'longname': 'reflectionAboutBBox',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'rab'},
                      'reflectionAboutOrigin': {'args': '<type __builtin__.bool>',
                                                'longname': 'reflectionAboutOrigin',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'rao'},
                      'reflectionAboutX': {'args': '<type __builtin__.bool>',
                                           'longname': 'reflectionAboutX',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'rax'},
                      'reflectionAboutY': {'args': '<type __builtin__.bool>',
                                           'longname': 'reflectionAboutY',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ray'},
                      'reflectionAboutZ': {'args': '<type __builtin__.bool>',
                                           'longname': 'reflectionAboutZ',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'raz'},
                      'reflectionTolerance': {'args': '<type __builtin__.float>',
                                              'longname': 'reflectionTolerance',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'rft'},
                      'relative': {'args': '<type __builtin__.bool>',
                                   'longname': 'relative',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'r'},
                      'scaleX': {'args': '<type __builtin__.bool>',
                                 'longname': 'scaleX',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'x'},
                      'scaleXY': {'args': '<type __builtin__.bool>',
                                  'longname': 'scaleXY',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'xy'},
                      'scaleXYZ': {'args': '<type __builtin__.bool>',
                                   'longname': 'scaleXYZ',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'xyz'},
                      'scaleXZ': {'args': '<type __builtin__.bool>',
                                  'longname': 'scaleXZ',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'xz'},
                      'scaleY': {'args': '<type __builtin__.bool>',
                                 'longname': 'scaleY',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'y'},
                      'scaleYZ': {'args': '<type __builtin__.bool>',
                                  'longname': 'scaleYZ',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'yz'},
                      'scaleZ': {'args': '<type __builtin__.bool>',
                                 'longname': 'scaleZ',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'z'},
                      'symNegative': {'args': '<type __builtin__.bool>',
                                      'longname': 'symNegative',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'smn'},
                      'worldSpace': {'args': '<type __builtin__.bool>',
                                     'longname': 'worldSpace',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ws'},
                      'xformConstraint': {'args': '<type __builtin__.str>',
                                          'longname': 'xformConstraint',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'xc'}},
            'shortFlags': {'a': 'absolute',
                           'cp': 'centerPivot',
                           'cs': 'componentSpace',
                           'dph': 'deletePriorHistory',
                           'dso': 'distanceOnly',
                           'ls': 'localSpace',
                           'oa': 'orientAxes',
                           'ocp': 'objectCenterPivot',
                           'os': 'objectSpace',
                           'p': 'pivot',
                           'pcp': 'preserveChildPosition',
                           'pgp': 'preserveGeometryPosition',
                           'puv': 'preserveUV',
                           'r': 'relative',
                           'rab': 'reflectionAboutBBox',
                           'rao': 'reflectionAboutOrigin',
                           'rax': 'reflectionAboutX',
                           'ray': 'reflectionAboutY',
                           'raz': 'reflectionAboutZ',
                           'rfl': 'reflection',
                           'rft': 'reflectionTolerance',
                           'smn': 'symNegative',
                           'ws': 'worldSpace',
                           'x': 'scaleX',
                           'xc': 'xformConstraint',
                           'xn': 'constrainAlongNormal',
                           'xy': 'scaleXY',
                           'xyz': 'scaleXYZ',
                           'xz': 'scaleXZ',
                           'y': 'scaleY',
                           'yz': 'scaleYZ',
                           'z': 'scaleZ'},
            'type': 'general'},
  'scaleComponents': {'flags': {'pivot': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'pivot',
                                          'modes': ['create'],
                                          'numArgs': 3,
                                          'shortname': 'p'},
                                'rotation': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'rotation',
                                             'modes': ['create'],
                                             'numArgs': 3,
                                             'shortname': 'ro'}},
                      'shortFlags': {'p': 'pivot', 'ro': 'rotation'},
                      'type': 'general'},
  'scaleConstraint': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'layer': {'args': '<type __builtin__.str>',
                                          'longname': 'layer',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'l'},
                                'maintainOffset': {'args': '<type __builtin__.bool>',
                                                   'longname': 'maintainOffset',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'mo'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'offset': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'offset',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'o'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'remove': {'args': '<type __builtin__.bool>',
                                           'longname': 'remove',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'rm'},
                                'scaleCompensate': {'args': '<type __builtin__.bool>',
                                                    'longname': 'scaleCompensate',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sc'},
                                'skip': {'args': '<type __builtin__.str>',
                                         'longname': 'skip',
                                         'modes': ['create',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'sk'},
                                'targetList': {'args': '<type __builtin__.bool>',
                                               'longname': 'targetList',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'tl'},
                                'weight': {'args': '<type __builtin__.float>',
                                           'longname': 'weight',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'w'},
                                'weightAliasList': {'args': '<type __builtin__.bool>',
                                                    'longname': 'weightAliasList',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'wal'}},
                      'resultNeedsUnpacking': True,
                      'shortFlags': {'e': 'edit',
                                     'l': 'layer',
                                     'mo': 'maintainOffset',
                                     'n': 'name',
                                     'o': 'offset',
                                     'q': 'query',
                                     'rm': 'remove',
                                     'sc': 'scaleCompensate',
                                     'sk': 'skip',
                                     'tl': 'targetList',
                                     'w': 'weight',
                                     'wal': 'weightAliasList'},
                      'type': 'animation'},
  'scaleKey': {'flags': {'animation': {'args': '<type __builtin__.str>',
                                       'longname': 'animation',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'an'},
                         'attribute': {'args': '<type __builtin__.str>',
                                       'longname': 'attribute',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'at'},
                         'controlPoints': {'args': '<type __builtin__.bool>',
                                           'longname': 'controlPoints',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'cp'},
                         'float': {'args': 'floatrange',
                                   'longname': 'float',
                                   'modes': ['create', 'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 'f'},
                         'floatPivot': {'args': '<type __builtin__.float>',
                                        'longname': 'floatPivot',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'fp'},
                         'floatScale': {'args': '<type __builtin__.float>',
                                        'longname': 'floatScale',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'fs'},
                         'hierarchy': {'args': '<type __builtin__.str>',
                                       'longname': 'hierarchy',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'hi'},
                         'includeUpperBound': {'args': '<type __builtin__.bool>',
                                               'longname': 'includeUpperBound',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'iub'},
                         'index': {'args': '<type __builtin__.int>',
                                   'longname': 'index',
                                   'modes': ['create', 'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 'index'},
                         'newEndFloat': {'args': '<type __builtin__.float>',
                                         'longname': 'newEndFloat',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'nef'},
                         'newEndTime': {'args': 'time',
                                        'longname': 'newEndTime',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'net'},
                         'newStartFloat': {'args': '<type __builtin__.float>',
                                           'longname': 'newStartFloat',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'nsf'},
                         'newStartTime': {'args': 'time',
                                          'longname': 'newStartTime',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'nst'},
                         'scaleSpecifiedKeys': {'args': '<type __builtin__.bool>',
                                                'longname': 'scaleSpecifiedKeys',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ssk'},
                         'shape': {'args': '<type __builtin__.bool>',
                                   'longname': 'shape',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 's'},
                         'time': {'args': 'timerange',
                                  'longname': 'time',
                                  'modes': ['create', 'multiuse'],
                                  'numArgs': 1,
                                  'shortname': 't'},
                         'timePivot': {'args': 'time',
                                       'longname': 'timePivot',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'tp'},
                         'timeScale': {'args': '<type __builtin__.float>',
                                       'longname': 'timeScale',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ts'},
                         'valuePivot': {'args': '<type __builtin__.float>',
                                        'longname': 'valuePivot',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'vp'},
                         'valueScale': {'args': '<type __builtin__.float>',
                                        'longname': 'valueScale',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'vs'}},
               'removedFlags': {'in': 'index'},
               'shortFlags': {'an': 'animation',
                              'at': 'attribute',
                              'cp': 'controlPoints',
                              'f': 'float',
                              'fp': 'floatPivot',
                              'fs': 'floatScale',
                              'hi': 'hierarchy',
                              'index': 'index',
                              'iub': 'includeUpperBound',
                              'nef': 'newEndFloat',
                              'net': 'newEndTime',
                              'nsf': 'newStartFloat',
                              'nst': 'newStartTime',
                              's': 'shape',
                              'ssk': 'scaleSpecifiedKeys',
                              't': 'time',
                              'tp': 'timePivot',
                              'ts': 'timeScale',
                              'vp': 'valuePivot',
                              'vs': 'valueScale'},
               'type': 'animation'},
  'scaleKeyCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'history': {'args': '<type __builtin__.bool>',
                                        'longname': 'history',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ch'},
                            'image1': {'args': '<type __builtin__.str>',
                                       'longname': 'image1',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i1'},
                            'image2': {'args': '<type __builtin__.str>',
                                       'longname': 'image2',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i2'},
                            'image3': {'args': '<type __builtin__.str>',
                                       'longname': 'image3',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i3'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'scaleSpecifiedKeys': {'args': '<type __builtin__.bool>',
                                                   'longname': 'scaleSpecifiedKeys',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ssk'},
                            'type': {'args': '<type __builtin__.str>',
                                     'longname': 'type',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'typ'}},
                  'shortFlags': {'ch': 'history',
                                 'e': 'edit',
                                 'ex': 'exists',
                                 'i1': 'image1',
                                 'i2': 'image2',
                                 'i3': 'image3',
                                 'n': 'name',
                                 'q': 'query',
                                 'ssk': 'scaleSpecifiedKeys',
                                 'typ': 'type'},
                  'type': 'context'},
  'sceneEditor': {'callbackFlags': ['selectCommand', 'sc'],
                  'flags': {'control': {'args': '<type __builtin__.bool>',
                                        'longname': 'control',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'ctl'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'filter': {'args': '<type __builtin__.str>',
                                       'longname': 'filter',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'f'},
                            'forceMainConnection': {'args': '<type __builtin__.str>',
                                                    'longname': 'forceMainConnection',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'fmc'},
                            'highlightConnection': {'args': '<type __builtin__.str>',
                                                    'longname': 'highlightConnection',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'hlc'},
                            'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                   'longname': 'lockMainConnection',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'lck'},
                            'mainListConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'mainListConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'mlc'},
                            'onlyParents': {'args': '<type __builtin__.bool>',
                                            'longname': 'onlyParents',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'op'},
                            'panel': {'args': '<type __builtin__.str>',
                                      'longname': 'panel',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'pnl'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'refreshReferences': {'args': '<type __builtin__.bool>',
                                                  'longname': 'refreshReferences',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'rr'},
                            'selectCommand': {'args': 'script',
                                              'longname': 'selectCommand',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sc'},
                            'selectItem': {'args': '<type __builtin__.int>',
                                           'longname': 'selectItem',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'si'},
                            'selectReference': {'args': '<type __builtin__.str>',
                                                'longname': 'selectReference',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'sr'},
                            'selectionConnection': {'args': '<type __builtin__.str>',
                                                    'longname': 'selectionConnection',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'slc'},
                            'shortName': {'args': '<type __builtin__.bool>',
                                          'longname': 'shortName',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'shn'},
                            'stateString': {'args': '<type __builtin__.bool>',
                                            'longname': 'stateString',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'sts'},
                            'unParent': {'args': '<type __builtin__.bool>',
                                         'longname': 'unParent',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'up'},
                            'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                     'longname': 'unlockMainConnection',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'ulk'},
                            'unresolvedName': {'args': '<type __builtin__.bool>',
                                               'longname': 'unresolvedName',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'un'},
                            'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                     'longname': 'updateMainConnection',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'upd'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'withoutCopyNumber': {'args': '<type __builtin__.bool>',
                                                  'longname': 'withoutCopyNumber',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'wcn'}},
                  'shortFlags': {'ctl': 'control',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ex': 'exists',
                                 'f': 'filter',
                                 'fmc': 'forceMainConnection',
                                 'hlc': 'highlightConnection',
                                 'lck': 'lockMainConnection',
                                 'mlc': 'mainListConnection',
                                 'op': 'onlyParents',
                                 'p': 'parent',
                                 'pnl': 'panel',
                                 'q': 'query',
                                 'rr': 'refreshReferences',
                                 'sc': 'selectCommand',
                                 'shn': 'shortName',
                                 'si': 'selectItem',
                                 'slc': 'selectionConnection',
                                 'sr': 'selectReference',
                                 'sts': 'stateString',
                                 'ulk': 'unlockMainConnection',
                                 'un': 'unresolvedName',
                                 'up': 'unParent',
                                 'upd': 'updateMainConnection',
                                 'ut': 'useTemplate',
                                 'wcn': 'withoutCopyNumber'},
                  'type': 'system'},
  'sceneLint': {'flags': {'issueType': {'args': '<type __builtin__.str>',
                                        'longname': 'issueType',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'i'},
                          'verbose': {'args': '<type __builtin__.bool>',
                                      'longname': 'verbose',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'v'}},
                'shortFlags': {},
                'type': 'general'},
  'sceneUIReplacement': {'flags': {'clear': {'args': '<type __builtin__.bool>',
                                             'longname': 'clear',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'cl'},
                                   'deleteRemaining': {'args': '<type __builtin__.bool>',
                                                       'longname': 'deleteRemaining',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'dr'},
                                   'getNextFilter': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'getNextFilter',
                                                     'modes': ['create'],
                                                     'numArgs': 2,
                                                     'shortname': 'gf'},
                                   'getNextPanel': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'getNextPanel',
                                                    'modes': ['create'],
                                                    'numArgs': 2,
                                                    'shortname': 'gp'},
                                   'getNextScriptedPanel': {'args': ['<type __builtin__.str>',
                                                                     '<type __builtin__.str>'],
                                                            'longname': 'getNextScriptedPanel',
                                                            'modes': ['create'],
                                                            'numArgs': 2,
                                                            'shortname': 'gsp'},
                                   'update': {'args': '<type __builtin__.str>',
                                              'longname': 'update',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'u'}},
                         'shortFlags': {'cl': 'clear',
                                        'dr': 'deleteRemaining',
                                        'gf': 'getNextFilter',
                                        'gp': 'getNextPanel',
                                        'gsp': 'getNextScriptedPanel',
                                        'u': 'update'},
                         'type': 'system'},
  'scmh': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                  'longname': 'absolute',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'a'},
                     'ignore': {'args': '<type __builtin__.int>',
                                'longname': 'ignore',
                                'modes': ['create', 'multiuse'],
                                'numArgs': 1,
                                'shortname': 'i'},
                     'quiet': {'args': '<type __builtin__.bool>',
                               'longname': 'quiet',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'q'},
                     'relative': {'args': '<type __builtin__.bool>',
                                  'longname': 'relative',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'r'}},
           'shortFlags': {'a': 'absolute',
                          'i': 'ignore',
                          'q': 'quiet',
                          'r': 'relative'},
           'type': 'windows'},
  'scriptCtx': {'callbackFlags': ['toolStart',
                                  'ts',
                                  'finalCommandScript',
                                  'fcs',
                                  'toolFinish',
                                  'tf'],
                'flags': {'allComponents': {'args': '<type __builtin__.bool>',
                                            'longname': 'allComponents',
                                            'modes': ['create',
                                                      'query',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'alc'},
                          'allObjects': {'args': '<type __builtin__.bool>',
                                         'longname': 'allObjects',
                                         'modes': ['create',
                                                   'query',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'alo'},
                          'animBreakdown': {'args': '<type __builtin__.bool>',
                                            'longname': 'animBreakdown',
                                            'modes': ['create',
                                                      'query',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'abd'},
                          'animCurve': {'args': '<type __builtin__.bool>',
                                        'longname': 'animCurve',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'ac'},
                          'animInTangent': {'args': '<type __builtin__.bool>',
                                            'longname': 'animInTangent',
                                            'modes': ['create',
                                                      'query',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'ait'},
                          'animKeyframe': {'args': '<type __builtin__.bool>',
                                           'longname': 'animKeyframe',
                                           'modes': ['create',
                                                     'query',
                                                     'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'ak'},
                          'animOutTangent': {'args': '<type __builtin__.bool>',
                                             'longname': 'animOutTangent',
                                             'modes': ['create',
                                                       'query',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'aot'},
                          'baseClassName': {'args': '<type __builtin__.str>',
                                            'longname': 'baseClassName',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'bcn'},
                          'byName': {'args': ['<type __builtin__.str>',
                                              '<type __builtin__.int>'],
                                     'longname': 'byName',
                                     'modes': ['multiuse'],
                                     'numArgs': 2,
                                     'shortname': 'bn'},
                          'camera': {'args': '<type __builtin__.bool>',
                                     'longname': 'camera',
                                     'modes': ['create',
                                               'query',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'ca'},
                          'cluster': {'args': '<type __builtin__.bool>',
                                      'longname': 'cluster',
                                      'modes': ['create',
                                                'query',
                                                'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'cl'},
                          'collisionModel': {'args': '<type __builtin__.bool>',
                                             'longname': 'collisionModel',
                                             'modes': ['create',
                                                       'query',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'clm'},
                          'controlVertex': {'args': '<type __builtin__.bool>',
                                            'longname': 'controlVertex',
                                            'modes': ['create',
                                                      'query',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'cv'},
                          'cumulativeLists': {'args': '<type __builtin__.bool>',
                                              'longname': 'cumulativeLists',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cls'},
                          'curve': {'args': '<type __builtin__.bool>',
                                    'longname': 'curve',
                                    'modes': ['create',
                                              'query',
                                              'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'c'},
                          'curveKnot': {'args': '<type __builtin__.bool>',
                                        'longname': 'curveKnot',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'ck'},
                          'curveOnSurface': {'args': '<type __builtin__.bool>',
                                             'longname': 'curveOnSurface',
                                             'modes': ['create',
                                                       'query',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'cos'},
                          'curveParameterPoint': {'args': '<type __builtin__.bool>',
                                                  'longname': 'curveParameterPoint',
                                                  'modes': ['create',
                                                            'query',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'cpp'},
                          'dimension': {'args': '<type __builtin__.bool>',
                                        'longname': 'dimension',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'dim'},
                          'dynamicConstraint': {'args': '<type __builtin__.bool>',
                                                'longname': 'dynamicConstraint',
                                                'modes': ['create',
                                                          'query',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'dc'},
                          'edge': {'args': '<type __builtin__.bool>',
                                   'longname': 'edge',
                                   'modes': ['create',
                                             'query',
                                             'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 'eg'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'editPoint': {'args': '<type __builtin__.bool>',
                                        'longname': 'editPoint',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'ep'},
                          'emitter': {'args': '<type __builtin__.bool>',
                                      'longname': 'emitter',
                                      'modes': ['create',
                                                'query',
                                                'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'em'},
                          'enableRootSelection': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableRootSelection',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ers'},
                          'escToQuit': {'args': '<type __builtin__.bool>',
                                        'longname': 'escToQuit',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'esc'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'exitUponCompletion': {'args': '<type __builtin__.bool>',
                                                 'longname': 'exitUponCompletion',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'euc'},
                          'expandSelectionList': {'args': '<type __builtin__.bool>',
                                                  'longname': 'expandSelectionList',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'esl'},
                          'facet': {'args': '<type __builtin__.bool>',
                                    'longname': 'facet',
                                    'modes': ['create',
                                              'query',
                                              'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'fc'},
                          'field': {'args': '<type __builtin__.bool>',
                                    'longname': 'field',
                                    'modes': ['create',
                                              'query',
                                              'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'fi'},
                          'finalCommandScript': {'args': 'script',
                                                 'longname': 'finalCommandScript',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'fcs'},
                          'fluid': {'args': '<type __builtin__.bool>',
                                    'longname': 'fluid',
                                    'modes': ['create',
                                              'query',
                                              'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'fl'},
                          'follicle': {'args': '<type __builtin__.bool>',
                                       'longname': 'follicle',
                                       'modes': ['create',
                                                 'query',
                                                 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'fo'},
                          'forceAddSelect': {'args': '<type __builtin__.bool>',
                                             'longname': 'forceAddSelect',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'fas'},
                          'hairSystem': {'args': '<type __builtin__.bool>',
                                         'longname': 'hairSystem',
                                         'modes': ['create',
                                                   'query',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'hs'},
                          'handle': {'args': '<type __builtin__.bool>',
                                     'longname': 'handle',
                                     'modes': ['create',
                                               'query',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'ha'},
                          'history': {'args': '<type __builtin__.bool>',
                                      'longname': 'history',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'ch'},
                          'hull': {'args': '<type __builtin__.bool>',
                                   'longname': 'hull',
                                   'modes': ['create',
                                             'query',
                                             'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 'hl'},
                          'ignoreInvalidItems': {'args': '<type __builtin__.bool>',
                                                 'longname': 'ignoreInvalidItems',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'iii'},
                          'ikEndEffector': {'args': '<type __builtin__.bool>',
                                            'longname': 'ikEndEffector',
                                            'modes': ['create',
                                                      'query',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'iee'},
                          'ikHandle': {'args': '<type __builtin__.bool>',
                                       'longname': 'ikHandle',
                                       'modes': ['create',
                                                 'query',
                                                 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'ikh'},
                          'image1': {'args': '<type __builtin__.str>',
                                     'longname': 'image1',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'i1'},
                          'image2': {'args': '<type __builtin__.str>',
                                     'longname': 'image2',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'i2'},
                          'image3': {'args': '<type __builtin__.str>',
                                     'longname': 'image3',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'i3'},
                          'imagePlane': {'args': '<type __builtin__.bool>',
                                         'longname': 'imagePlane',
                                         'modes': ['create',
                                                   'query',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'ip'},
                          'implicitGeometry': {'args': '<type __builtin__.bool>',
                                               'longname': 'implicitGeometry',
                                               'modes': ['create',
                                                         'query',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'ig'},
                          'isoparm': {'args': '<type __builtin__.bool>',
                                      'longname': 'isoparm',
                                      'modes': ['create',
                                                'query',
                                                'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'iso'},
                          'joint': {'args': '<type __builtin__.bool>',
                                    'longname': 'joint',
                                    'modes': ['create',
                                              'query',
                                              'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'j'},
                          'jointPivot': {'args': '<type __builtin__.bool>',
                                         'longname': 'jointPivot',
                                         'modes': ['create',
                                                   'query',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'jp'},
                          'lastAutoComplete': {'args': '<type __builtin__.bool>',
                                               'longname': 'lastAutoComplete',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'lac'},
                          'lattice': {'args': '<type __builtin__.bool>',
                                      'longname': 'lattice',
                                      'modes': ['create',
                                                'query',
                                                'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'la'},
                          'latticePoint': {'args': '<type __builtin__.bool>',
                                           'longname': 'latticePoint',
                                           'modes': ['create',
                                                     'query',
                                                     'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'lp'},
                          'light': {'args': '<type __builtin__.bool>',
                                    'longname': 'light',
                                    'modes': ['create',
                                              'query',
                                              'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'lt'},
                          'localRotationAxis': {'args': '<type __builtin__.bool>',
                                                'longname': 'localRotationAxis',
                                                'modes': ['create',
                                                          'query',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'ra'},
                          'locator': {'args': '<type __builtin__.bool>',
                                      'longname': 'locator',
                                      'modes': ['create',
                                                'query',
                                                'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'lc'},
                          'locatorUV': {'args': '<type __builtin__.bool>',
                                        'longname': 'locatorUV',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'luv'},
                          'locatorXYZ': {'args': '<type __builtin__.bool>',
                                         'longname': 'locatorXYZ',
                                         'modes': ['create',
                                                   'query',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'xyz'},
                          'meshComponents': {'args': '<type __builtin__.bool>',
                                             'longname': 'meshComponents',
                                             'modes': ['multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'mc'},
                          'meshUVShell': {'args': '<type __builtin__.bool>',
                                          'longname': 'meshUVShell',
                                          'modes': ['multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'msh'},
                          'motionTrailPoint': {'args': '<type __builtin__.bool>',
                                               'longname': 'motionTrailPoint',
                                               'modes': ['multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'mtp'},
                          'motionTrailTangent': {'args': '<type __builtin__.bool>',
                                                 'longname': 'motionTrailTangent',
                                                 'modes': ['multiuse'],
                                                 'numArgs': 1,
                                                 'shortname': 'mtt'},
                          'nCloth': {'args': '<type __builtin__.bool>',
                                     'longname': 'nCloth',
                                     'modes': ['create',
                                               'query',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'ncl'},
                          'nParticle': {'args': '<type __builtin__.bool>',
                                        'longname': 'nParticle',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'npr'},
                          'nParticleShape': {'args': '<type __builtin__.bool>',
                                             'longname': 'nParticleShape',
                                             'modes': ['create',
                                                       'query',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'nps'},
                          'nRigid': {'args': '<type __builtin__.bool>',
                                     'longname': 'nRigid',
                                     'modes': ['create',
                                               'query',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'nr'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'nonlinear': {'args': '<type __builtin__.bool>',
                                        'longname': 'nonlinear',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'nl'},
                          'nurbsCurve': {'args': '<type __builtin__.bool>',
                                         'longname': 'nurbsCurve',
                                         'modes': ['create',
                                                   'query',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'nc'},
                          'nurbsSurface': {'args': '<type __builtin__.bool>',
                                           'longname': 'nurbsSurface',
                                           'modes': ['create',
                                                     'query',
                                                     'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'ns'},
                          'objectComponent': {'args': '<type __builtin__.bool>',
                                              'longname': 'objectComponent',
                                              'modes': ['create', 'query'],
                                              'numArgs': 0,
                                              'shortname': 'ocm'},
                          'orientationLocator': {'args': '<type __builtin__.bool>',
                                                 'longname': 'orientationLocator',
                                                 'modes': ['create',
                                                           'query',
                                                           'multiuse'],
                                                 'numArgs': 1,
                                                 'shortname': 'ol'},
                          'particle': {'args': '<type __builtin__.bool>',
                                       'longname': 'particle',
                                       'modes': ['create',
                                                 'query',
                                                 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'pr'},
                          'particleShape': {'args': '<type __builtin__.bool>',
                                            'longname': 'particleShape',
                                            'modes': ['create',
                                                      'query',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'ps'},
                          'plane': {'args': '<type __builtin__.bool>',
                                    'longname': 'plane',
                                    'modes': ['create',
                                              'query',
                                              'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'pl'},
                          'polymesh': {'args': '<type __builtin__.bool>',
                                       'longname': 'polymesh',
                                       'modes': ['create',
                                                 'query',
                                                 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                          'polymeshEdge': {'args': '<type __builtin__.bool>',
                                           'longname': 'polymeshEdge',
                                           'modes': ['create',
                                                     'query',
                                                     'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'pe'},
                          'polymeshFace': {'args': '<type __builtin__.bool>',
                                           'longname': 'polymeshFace',
                                           'modes': ['create',
                                                     'query',
                                                     'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'pf'},
                          'polymeshFreeEdge': {'args': '<type __builtin__.bool>',
                                               'longname': 'polymeshFreeEdge',
                                               'modes': ['create',
                                                         'query',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'pfe'},
                          'polymeshUV': {'args': '<type __builtin__.bool>',
                                         'longname': 'polymeshUV',
                                         'modes': ['create',
                                                   'query',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'puv'},
                          'polymeshVertex': {'args': '<type __builtin__.bool>',
                                             'longname': 'polymeshVertex',
                                             'modes': ['create',
                                                       'query',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'pv'},
                          'polymeshVtxFace': {'args': '<type __builtin__.bool>',
                                              'longname': 'polymeshVtxFace',
                                              'modes': ['create',
                                                        'query',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'pvf'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'queryByName': {'args': '<type __builtin__.str>',
                                          'longname': 'queryByName',
                                          'numArgs': 1,
                                          'shortname': 'qbn'},
                          'rigidBody': {'args': '<type __builtin__.bool>',
                                        'longname': 'rigidBody',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'rb'},
                          'rigidConstraint': {'args': '<type __builtin__.bool>',
                                              'longname': 'rigidConstraint',
                                              'modes': ['create',
                                                        'query',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'rc'},
                          'rotatePivot': {'args': '<type __builtin__.bool>',
                                          'longname': 'rotatePivot',
                                          'modes': ['create',
                                                    'query',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'rp'},
                          'scalePivot': {'args': '<type __builtin__.bool>',
                                         'longname': 'scalePivot',
                                         'modes': ['create',
                                                   'query',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'sp'},
                          'sculpt': {'args': '<type __builtin__.bool>',
                                     'longname': 'sculpt',
                                     'modes': ['create',
                                               'query',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'sc'},
                          'selectHandle': {'args': '<type __builtin__.bool>',
                                           'longname': 'selectHandle',
                                           'modes': ['create',
                                                     'query',
                                                     'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'sh'},
                          'setAllowExcessCount': {'args': '<type __builtin__.bool>',
                                                  'longname': 'setAllowExcessCount',
                                                  'modes': ['create',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'sae'},
                          'setAutoComplete': {'args': '<type __builtin__.bool>',
                                              'longname': 'setAutoComplete',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'sac'},
                          'setAutoToggleSelection': {'args': '<type __builtin__.bool>',
                                                     'longname': 'setAutoToggleSelection',
                                                     'modes': ['create',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'sat'},
                          'setDoneSelectionPrompt': {'args': '<type __builtin__.str>',
                                                     'longname': 'setDoneSelectionPrompt',
                                                     'modes': ['create',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'dsp'},
                          'setNoSelectionHeadsUp': {'args': '<type __builtin__.str>',
                                                    'longname': 'setNoSelectionHeadsUp',
                                                    'modes': ['create',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 1,
                                                    'shortname': 'snh'},
                          'setNoSelectionPrompt': {'args': '<type __builtin__.str>',
                                                   'longname': 'setNoSelectionPrompt',
                                                   'modes': ['create',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'snp'},
                          'setSelectionCount': {'args': '<type __builtin__.int>',
                                                'longname': 'setSelectionCount',
                                                'modes': ['create',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'ssc'},
                          'setSelectionHeadsUp': {'args': '<type __builtin__.str>',
                                                  'longname': 'setSelectionHeadsUp',
                                                  'modes': ['create',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'ssh'},
                          'setSelectionPrompt': {'args': '<type __builtin__.str>',
                                                 'longname': 'setSelectionPrompt',
                                                 'modes': ['create',
                                                           'edit',
                                                           'multiuse'],
                                                 'numArgs': 1,
                                                 'shortname': 'ssp'},
                          'showManipulators': {'args': '<type __builtin__.bool>',
                                               'longname': 'showManipulators',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sm'},
                          'spring': {'args': '<type __builtin__.bool>',
                                     'longname': 'spring',
                                     'modes': ['create',
                                               'query',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'spr'},
                          'springComponent': {'args': '<type __builtin__.bool>',
                                              'longname': 'springComponent',
                                              'modes': ['create',
                                                        'query',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'spc'},
                          'stroke': {'args': '<type __builtin__.bool>',
                                     'longname': 'stroke',
                                     'modes': ['create',
                                               'query',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'str'},
                          'subdiv': {'args': '<type __builtin__.bool>',
                                     'longname': 'subdiv',
                                     'modes': ['create',
                                               'query',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'sd'},
                          'subdivMeshEdge': {'args': '<type __builtin__.bool>',
                                             'longname': 'subdivMeshEdge',
                                             'modes': ['create',
                                                       'query',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'sme'},
                          'subdivMeshFace': {'args': '<type __builtin__.bool>',
                                             'longname': 'subdivMeshFace',
                                             'modes': ['create',
                                                       'query',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'smf'},
                          'subdivMeshPoint': {'args': '<type __builtin__.bool>',
                                              'longname': 'subdivMeshPoint',
                                              'modes': ['create',
                                                        'query',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'smp'},
                          'subdivMeshUV': {'args': '<type __builtin__.bool>',
                                           'longname': 'subdivMeshUV',
                                           'modes': ['create',
                                                     'query',
                                                     'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'smu'},
                          'surfaceEdge': {'args': '<type __builtin__.bool>',
                                          'longname': 'surfaceEdge',
                                          'modes': ['create',
                                                    'query',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'se'},
                          'surfaceFace': {'args': '<type __builtin__.bool>',
                                          'longname': 'surfaceFace',
                                          'modes': ['create',
                                                    'query',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'sf'},
                          'surfaceKnot': {'args': '<type __builtin__.bool>',
                                          'longname': 'surfaceKnot',
                                          'modes': ['create',
                                                    'query',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'sk'},
                          'surfaceParameterPoint': {'args': '<type __builtin__.bool>',
                                                    'longname': 'surfaceParameterPoint',
                                                    'modes': ['create',
                                                              'query',
                                                              'multiuse'],
                                                    'numArgs': 1,
                                                    'shortname': 'spp'},
                          'surfaceRange': {'args': '<type __builtin__.bool>',
                                           'longname': 'surfaceRange',
                                           'modes': ['create',
                                                     'query',
                                                     'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'sr'},
                          'surfaceUV': {'args': '<type __builtin__.bool>',
                                        'longname': 'surfaceUV',
                                        'modes': ['create',
                                                  'query',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'suv'},
                          'texture': {'args': '<type __builtin__.bool>',
                                      'longname': 'texture',
                                      'modes': ['create',
                                                'query',
                                                'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'tx'},
                          'title': {'args': '<type __builtin__.str>',
                                    'longname': 'title',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 't'},
                          'toolCursorType': {'args': '<type __builtin__.str>',
                                             'longname': 'toolCursorType',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'tct'},
                          'toolFinish': {'args': 'script',
                                         'longname': 'toolFinish',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'tf'},
                          'toolStart': {'args': 'script',
                                        'longname': 'toolStart',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ts'},
                          'totalSelectionSets': {'args': '<type __builtin__.int>',
                                                 'longname': 'totalSelectionSets',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'tss'},
                          'vertex': {'args': '<type __builtin__.bool>',
                                     'longname': 'vertex',
                                     'modes': ['create',
                                               'query',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'v'}},
                'shortFlags': {'abd': 'animBreakdown',
                               'ac': 'animCurve',
                               'ait': 'animInTangent',
                               'ak': 'animKeyframe',
                               'alc': 'allComponents',
                               'alo': 'allObjects',
                               'aot': 'animOutTangent',
                               'bcn': 'baseClassName',
                               'bn': 'byName',
                               'c': 'curve',
                               'ca': 'camera',
                               'ch': 'history',
                               'ck': 'curveKnot',
                               'cl': 'cluster',
                               'clm': 'collisionModel',
                               'cls': 'cumulativeLists',
                               'cos': 'curveOnSurface',
                               'cpp': 'curveParameterPoint',
                               'cv': 'controlVertex',
                               'dc': 'dynamicConstraint',
                               'dim': 'dimension',
                               'dsp': 'setDoneSelectionPrompt',
                               'e': 'edit',
                               'eg': 'edge',
                               'em': 'emitter',
                               'ep': 'editPoint',
                               'ers': 'enableRootSelection',
                               'esc': 'escToQuit',
                               'esl': 'expandSelectionList',
                               'euc': 'exitUponCompletion',
                               'ex': 'exists',
                               'fas': 'forceAddSelect',
                               'fc': 'facet',
                               'fcs': 'finalCommandScript',
                               'fi': 'field',
                               'fl': 'fluid',
                               'fo': 'follicle',
                               'ha': 'handle',
                               'hl': 'hull',
                               'hs': 'hairSystem',
                               'i1': 'image1',
                               'i2': 'image2',
                               'i3': 'image3',
                               'iee': 'ikEndEffector',
                               'ig': 'implicitGeometry',
                               'iii': 'ignoreInvalidItems',
                               'ikh': 'ikHandle',
                               'ip': 'imagePlane',
                               'iso': 'isoparm',
                               'j': 'joint',
                               'jp': 'jointPivot',
                               'la': 'lattice',
                               'lac': 'lastAutoComplete',
                               'lc': 'locator',
                               'lp': 'latticePoint',
                               'lt': 'light',
                               'luv': 'locatorUV',
                               'mc': 'meshComponents',
                               'msh': 'meshUVShell',
                               'mtp': 'motionTrailPoint',
                               'mtt': 'motionTrailTangent',
                               'n': 'name',
                               'nc': 'nurbsCurve',
                               'ncl': 'nCloth',
                               'nl': 'nonlinear',
                               'npr': 'nParticle',
                               'nps': 'nParticleShape',
                               'nr': 'nRigid',
                               'ns': 'nurbsSurface',
                               'ocm': 'objectComponent',
                               'ol': 'orientationLocator',
                               'p': 'polymesh',
                               'pe': 'polymeshEdge',
                               'pf': 'polymeshFace',
                               'pfe': 'polymeshFreeEdge',
                               'pl': 'plane',
                               'pr': 'particle',
                               'ps': 'particleShape',
                               'puv': 'polymeshUV',
                               'pv': 'polymeshVertex',
                               'pvf': 'polymeshVtxFace',
                               'q': 'query',
                               'qbn': 'queryByName',
                               'ra': 'localRotationAxis',
                               'rb': 'rigidBody',
                               'rc': 'rigidConstraint',
                               'rp': 'rotatePivot',
                               'sac': 'setAutoComplete',
                               'sae': 'setAllowExcessCount',
                               'sat': 'setAutoToggleSelection',
                               'sc': 'sculpt',
                               'sd': 'subdiv',
                               'se': 'surfaceEdge',
                               'sf': 'surfaceFace',
                               'sh': 'selectHandle',
                               'sk': 'surfaceKnot',
                               'sm': 'showManipulators',
                               'sme': 'subdivMeshEdge',
                               'smf': 'subdivMeshFace',
                               'smp': 'subdivMeshPoint',
                               'smu': 'subdivMeshUV',
                               'snh': 'setNoSelectionHeadsUp',
                               'snp': 'setNoSelectionPrompt',
                               'sp': 'scalePivot',
                               'spc': 'springComponent',
                               'spp': 'surfaceParameterPoint',
                               'spr': 'spring',
                               'sr': 'surfaceRange',
                               'ssc': 'setSelectionCount',
                               'ssh': 'setSelectionHeadsUp',
                               'ssp': 'setSelectionPrompt',
                               'str': 'stroke',
                               'suv': 'surfaceUV',
                               't': 'title',
                               'tct': 'toolCursorType',
                               'tf': 'toolFinish',
                               'ts': 'toolStart',
                               'tss': 'totalSelectionSets',
                               'tx': 'texture',
                               'v': 'vertex',
                               'xyz': 'locatorXYZ'},
                'type': 'context'},
  'scriptEditorInfo': {'flags': {'clearHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'clearHistory',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'ch'},
                                 'clearHistoryFile': {'args': '<type __builtin__.bool>',
                                                      'longname': 'clearHistoryFile',
                                                      'modes': ['edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'chf'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'historyFilename': {'args': '<type __builtin__.str>',
                                                     'longname': 'historyFilename',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'hfn'},
                                 'input': {'args': '<type __builtin__.str>',
                                           'longname': 'input',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'i'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'suppressErrors': {'args': '<type __builtin__.bool>',
                                                    'longname': 'suppressErrors',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'se'},
                                 'suppressInfo': {'args': '<type __builtin__.bool>',
                                                  'longname': 'suppressInfo',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'si'},
                                 'suppressResults': {'args': '<type __builtin__.bool>',
                                                     'longname': 'suppressResults',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sr'},
                                 'suppressStackWindow': {'args': '<type __builtin__.bool>',
                                                         'longname': 'suppressStackWindow',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ssw'},
                                 'suppressWarnings': {'args': '<type __builtin__.bool>',
                                                      'longname': 'suppressWarnings',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'sw'},
                                 'writeHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'writeHistory',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'wh'}},
                       'shortFlags': {'ch': 'clearHistory',
                                      'chf': 'clearHistoryFile',
                                      'e': 'edit',
                                      'hfn': 'historyFilename',
                                      'i': 'input',
                                      'q': 'query',
                                      'se': 'suppressErrors',
                                      'si': 'suppressInfo',
                                      'sr': 'suppressResults',
                                      'ssw': 'suppressStackWindow',
                                      'sw': 'suppressWarnings',
                                      'wh': 'writeHistory'},
                       'type': 'windows'},
  'scriptJob': {'callbackFlags': ['idleEvent', 'ie', 'timeChange', 'tc'],
                'flags': {'allChildren': {'args': '<type __builtin__.bool>',
                                          'longname': 'allChildren',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'alc'},
                          'attributeAdded': {'args': ['<type __builtin__.str>',
                                                      'script'],
                                             'longname': 'attributeAdded',
                                             'modes': ['create'],
                                             'numArgs': 2,
                                             'shortname': 'aa'},
                          'attributeChange': {'args': ['<type __builtin__.str>',
                                                       'script'],
                                              'longname': 'attributeChange',
                                              'modes': ['create'],
                                              'numArgs': 2,
                                              'shortname': 'ac'},
                          'attributeDeleted': {'args': ['<type __builtin__.str>',
                                                        'script'],
                                               'longname': 'attributeDeleted',
                                               'modes': ['create'],
                                               'numArgs': 2,
                                               'shortname': 'ad'},
                          'compressUndo': {'args': '<type __builtin__.bool>',
                                           'longname': 'compressUndo',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'cu'},
                          'conditionChange': {'args': ['<type __builtin__.str>',
                                                       'script'],
                                              'longname': 'conditionChange',
                                              'modes': ['create'],
                                              'numArgs': 2,
                                              'shortname': 'cc'},
                          'conditionFalse': {'args': ['<type __builtin__.str>',
                                                      'script'],
                                             'longname': 'conditionFalse',
                                             'modes': ['create'],
                                             'numArgs': 2,
                                             'shortname': 'cf'},
                          'conditionTrue': {'args': ['<type __builtin__.str>',
                                                     'script'],
                                            'longname': 'conditionTrue',
                                            'modes': ['create'],
                                            'numArgs': 2,
                                            'shortname': 'ct'},
                          'connectionChange': {'args': ['<type __builtin__.str>',
                                                        'script'],
                                               'longname': 'connectionChange',
                                               'modes': ['create'],
                                               'numArgs': 2,
                                               'shortname': 'con'},
                          'disregardIndex': {'args': '<type __builtin__.bool>',
                                             'longname': 'disregardIndex',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'dri'},
                          'event': {'args': ['<type __builtin__.str>',
                                             'script'],
                                    'longname': 'event',
                                    'modes': ['create'],
                                    'numArgs': 2,
                                    'shortname': 'e'},
                          'exists': {'args': '<type __builtin__.int>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'ex'},
                          'force': {'args': '<type __builtin__.bool>',
                                    'longname': 'force',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'f'},
                          'idleEvent': {'args': 'script',
                                        'longname': 'idleEvent',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ie'},
                          'kill': {'args': '<type __builtin__.int>',
                                   'longname': 'kill',
                                   'modes': ['create', 'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 'k'},
                          'killAll': {'args': '<type __builtin__.bool>',
                                      'longname': 'killAll',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ka'},
                          'killWithScene': {'args': '<type __builtin__.bool>',
                                            'longname': 'killWithScene',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'kws'},
                          'listConditions': {'args': '<type __builtin__.bool>',
                                             'longname': 'listConditions',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'lc'},
                          'listEvents': {'args': '<type __builtin__.bool>',
                                         'longname': 'listEvents',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'le'},
                          'listJobs': {'args': '<type __builtin__.bool>',
                                       'longname': 'listJobs',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'lj'},
                          'nodeDeleted': {'args': ['<type __builtin__.str>',
                                                   'script'],
                                          'longname': 'nodeDeleted',
                                          'modes': ['create'],
                                          'numArgs': 2,
                                          'shortname': 'nd'},
                          'nodeNameChanged': {'args': ['<type __builtin__.str>',
                                                       'script'],
                                              'longname': 'nodeNameChanged',
                                              'modes': ['create'],
                                              'numArgs': 2,
                                              'shortname': 'nnc'},
                          'optionVarChange': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.callable>'],
                                              'longname': 'optionVarChange',
                                              'numArgs': 2,
                                              'shortname': 'ovc'},
                          'optionVarChanged': {'args': ['<type __builtin__.str>',
                                                        'script'],
                                               'longname': 'optionVarChanged',
                                               'modes': ['create'],
                                               'numArgs': 2,
                                               'shortname': 'ovc'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                          'permanent': {'args': '<type __builtin__.bool>',
                                        'longname': 'permanent',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'per'},
                          'protected': {'args': '<type __builtin__.bool>',
                                        'longname': 'protected',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'pro'},
                          'replacePrevious': {'args': '<type __builtin__.bool>',
                                              'longname': 'replacePrevious',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'rp'},
                          'runOnce': {'args': '<type __builtin__.bool>',
                                      'longname': 'runOnce',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'ro'},
                          'timeChange': {'args': 'script',
                                         'longname': 'timeChange',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'tc'},
                          'uiDeleted': {'args': ['<type __builtin__.str>',
                                                 'script'],
                                        'longname': 'uiDeleted',
                                        'modes': ['create'],
                                        'numArgs': 2,
                                        'shortname': 'uid'}},
                'shortFlags': {'aa': 'attributeAdded',
                               'ac': 'attributeChange',
                               'ad': 'attributeDeleted',
                               'alc': 'allChildren',
                               'cc': 'conditionChange',
                               'cf': 'conditionFalse',
                               'con': 'connectionChange',
                               'ct': 'conditionTrue',
                               'cu': 'compressUndo',
                               'dri': 'disregardIndex',
                               'e': 'event',
                               'ex': 'exists',
                               'f': 'force',
                               'ie': 'idleEvent',
                               'k': 'kill',
                               'ka': 'killAll',
                               'kws': 'killWithScene',
                               'lc': 'listConditions',
                               'le': 'listEvents',
                               'lj': 'listJobs',
                               'nd': 'nodeDeleted',
                               'nnc': 'nodeNameChanged',
                               'ovc': 'optionVarChange',
                               'p': 'parent',
                               'per': 'permanent',
                               'pro': 'protected',
                               'ro': 'runOnce',
                               'rp': 'replacePrevious',
                               'tc': 'timeChange',
                               'uid': 'uiDeleted'},
                'type': 'language'},
  'scriptNode': {'flags': {'afterScript': {'args': '<type __builtin__.str>',
                                           'longname': 'afterScript',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'afterScript'},
                           'beforeScript': {'args': '<type __builtin__.str>',
                                            'longname': 'beforeScript',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'bs'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'executeAfter': {'args': '<type __builtin__.bool>',
                                            'longname': 'executeAfter',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ea'},
                           'executeBefore': {'args': '<type __builtin__.bool>',
                                             'longname': 'executeBefore',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'eb'},
                           'ignoreReferenceEdits': {'args': '<type __builtin__.bool>',
                                                    'longname': 'ignoreReferenceEdits',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ire'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'scriptType': {'args': '<type __builtin__.int>',
                                          'longname': 'scriptType',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'st'},
                           'sourceType': {'args': '<type __builtin__.str>',
                                          'longname': 'sourceType',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'stp'}},
                 'removedFlags': {'as': 'afterScript'},
                 'shortFlags': {'afterScript': 'afterScript',
                                'bs': 'beforeScript',
                                'e': 'edit',
                                'ea': 'executeAfter',
                                'eb': 'executeBefore',
                                'ire': 'ignoreReferenceEdits',
                                'n': 'name',
                                'q': 'query',
                                'st': 'scriptType',
                                'stp': 'sourceType'},
                 'type': 'system'},
  'scriptTable': {'callbackFlags': ['cellChangedCmd',
                                    'ccc',
                                    'cellBackgroundColorCommand',
                                    'cbc',
                                    'cellForegroundColorCommand',
                                    'cfc',
                                    'dragCallback',
                                    'dgc',
                                    'rowsRemovedCmd',
                                    'rrc',
                                    'rowsToBeRemovedCmd',
                                    'rtc',
                                    'getCellCmd',
                                    'gcc',
                                    'visibleChangeCommand',
                                    'vcc',
                                    'afterCellChangedCmd',
                                    'acc',
                                    'selectionChangedCmd',
                                    'scc',
                                    'dropCallback',
                                    'dpc'],
                  'flags': {'afterCellChangedCmd': {'args': 'script',
                                                    'longname': 'afterCellChangedCmd',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'acc'},
                            'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'cellBackgroundColorCommand': {'args': 'script',
                                                           'longname': 'cellBackgroundColorCommand',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'cbc'},
                            'cellChangedCmd': {'args': 'script',
                                               'longname': 'cellChangedCmd',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ccc'},
                            'cellForegroundColorCommand': {'args': 'script',
                                                           'longname': 'cellForegroundColorCommand',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'cfc'},
                            'cellIndex': {'args': ['<type __builtin__.int>',
                                                   '<type __builtin__.int>'],
                                          'longname': 'cellIndex',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 2,
                                          'shortname': 'ci'},
                            'cellValue': {'args': '<type __builtin__.str>',
                                          'longname': 'cellValue',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cv'},
                            'clearRow': {'args': '<type __builtin__.int>',
                                         'longname': 'clearRow',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 'cr'},
                            'clearTable': {'args': '<type __builtin__.bool>',
                                           'longname': 'clearTable',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'ct'},
                            'columnFilter': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnFilter',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 2,
                                             'shortname': 'cf'},
                            'columnWidth': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.int>'],
                                            'longname': 'columnWidth',
                                            'modes': ['create',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 2,
                                            'shortname': 'cw'},
                            'columns': {'args': '<type __builtin__.int>',
                                        'longname': 'columns',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'c'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'deleteRow': {'args': '<type __builtin__.int>',
                                          'longname': 'deleteRow',
                                          'modes': ['edit'],
                                          'numArgs': 1,
                                          'shortname': 'dr'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'editable': {'args': '<type __builtin__.bool>',
                                         'longname': 'editable',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ed'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'excludingHeaders': {'args': '<type __builtin__.bool>',
                                                 'longname': 'excludingHeaders',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'eh'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'getCellCmd': {'args': 'script',
                                           'longname': 'getCellCmd',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'gcc'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'insertRow': {'args': '<type __builtin__.int>',
                                          'longname': 'insertRow',
                                          'modes': ['edit'],
                                          'numArgs': 1,
                                          'shortname': 'ir'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'label': {'args': ['<type __builtin__.int>',
                                               '<type __builtin__.str>'],
                                      'longname': 'label',
                                      'modes': ['create',
                                                'edit',
                                                'multiuse'],
                                      'numArgs': 2,
                                      'shortname': 'l'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'multiEditEnabled': {'args': '<type __builtin__.bool>',
                                                 'longname': 'multiEditEnabled',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'mee'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'rowHeight': {'args': '<type __builtin__.int>',
                                          'longname': 'rowHeight',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'rh'},
                            'rows': {'args': '<type __builtin__.int>',
                                     'longname': 'rows',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'r'},
                            'rowsRemovedCmd': {'args': 'script',
                                               'longname': 'rowsRemovedCmd',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rrc'},
                            'rowsToBeRemovedCmd': {'args': 'script',
                                                   'longname': 'rowsToBeRemovedCmd',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'rtc'},
                            'selectedCells': {'args': '<type __builtin__.int>',
                                              'longname': 'selectedCells',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sc'},
                            'selectedColumns': {'args': '<type __builtin__.int>',
                                                'longname': 'selectedColumns',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'scs'},
                            'selectedRow': {'args': '<type __builtin__.bool>',
                                            'longname': 'selectedRow',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'sr'},
                            'selectedRows': {'args': '<type __builtin__.int>',
                                             'longname': 'selectedRows',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'srs'},
                            'selectionBehavior': {'args': '<type __builtin__.int>',
                                                  'longname': 'selectionBehavior',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sb'},
                            'selectionChangedCmd': {'args': 'script',
                                                    'longname': 'selectionChangedCmd',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'scc'},
                            'selectionMode': {'args': '<type __builtin__.int>',
                                              'longname': 'selectionMode',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sm'},
                            'sortEnabled': {'args': '<type __builtin__.bool>',
                                            'longname': 'sortEnabled',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'se'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'underPointerColumn': {'args': '<type __builtin__.bool>',
                                                   'longname': 'underPointerColumn',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'upc'},
                            'underPointerRow': {'args': '<type __builtin__.bool>',
                                                'longname': 'underPointerRow',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'upr'},
                            'useDoubleClickEdit': {'args': '<type __builtin__.bool>',
                                                   'longname': 'useDoubleClickEdit',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'udc'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'acc': 'afterCellChangedCmd',
                                 'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'c': 'columns',
                                 'cbc': 'cellBackgroundColorCommand',
                                 'ccc': 'cellChangedCmd',
                                 'cf': 'columnFilter',
                                 'cfc': 'cellForegroundColorCommand',
                                 'ci': 'cellIndex',
                                 'cr': 'clearRow',
                                 'ct': 'clearTable',
                                 'cv': 'cellValue',
                                 'cw': 'columnWidth',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'dr': 'deleteRow',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ed': 'editable',
                                 'eh': 'excludingHeaders',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'ex': 'exists',
                                 'fpn': 'fullPathName',
                                 'gcc': 'getCellCmd',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'io': 'isObscured',
                                 'ir': 'insertRow',
                                 'l': 'label',
                                 'm': 'manage',
                                 'mee': 'multiEditEnabled',
                                 'nbg': 'noBackground',
                                 'npm': 'numberOfPopupMenus',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'r': 'rows',
                                 'rh': 'rowHeight',
                                 'rrc': 'rowsRemovedCmd',
                                 'rtc': 'rowsToBeRemovedCmd',
                                 'sb': 'selectionBehavior',
                                 'sbm': 'statusBarMessage',
                                 'sc': 'selectedCells',
                                 'scc': 'selectionChangedCmd',
                                 'scs': 'selectedColumns',
                                 'se': 'sortEnabled',
                                 'sm': 'selectionMode',
                                 'sr': 'selectedRow',
                                 'srs': 'selectedRows',
                                 'udc': 'useDoubleClickEdit',
                                 'upc': 'underPointerColumn',
                                 'upr': 'underPointerRow',
                                 'ut': 'useTemplate',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'scriptedPanel': {'callbackFlags': ['popupMenuProcedure', 'pmp'],
                    'flags': {'control': {'args': '<type __builtin__.bool>',
                                          'longname': 'control',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'ctl'},
                              'copy': {'args': '<type __builtin__.str>',
                                       'longname': 'copy',
                                       'modes': ['edit'],
                                       'numArgs': 1,
                                       'shortname': 'cp'},
                              'createString': {'args': '<type __builtin__.bool>',
                                               'longname': 'createString',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'cs'},
                              'defineTemplate': {'args': '<type __builtin__.str>',
                                                 'longname': 'defineTemplate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'dt'},
                              'docTag': {'args': '<type __builtin__.str>',
                                         'longname': 'docTag',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dtg'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'editString': {'args': '<type __builtin__.bool>',
                                             'longname': 'editString',
                                             'modes': ['edit'],
                                             'numArgs': 0,
                                             'shortname': 'es'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'init': {'args': '<type __builtin__.bool>',
                                       'longname': 'init',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'init'},
                              'isUnique': {'args': '<type __builtin__.bool>',
                                           'longname': 'isUnique',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'iu'},
                              'label': {'args': '<type __builtin__.str>',
                                        'longname': 'label',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'l'},
                              'menuBarRepeatLast': {'args': '<type __builtin__.bool>',
                                                    'longname': 'menuBarRepeatLast',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'mrl'},
                              'menuBarVisible': {'args': '<type __builtin__.bool>',
                                                 'longname': 'menuBarVisible',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'mbv'},
                              'needsInit': {'args': '<type __builtin__.bool>',
                                            'longname': 'needsInit',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'ni'},
                              'parent': {'args': '<type __builtin__.str>',
                                         'longname': 'parent',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'p'},
                              'popupMenuProcedure': {'args': 'script',
                                                     'longname': 'popupMenuProcedure',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'pmp'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'replacePanel': {'args': '<type __builtin__.str>',
                                               'longname': 'replacePanel',
                                               'modes': ['edit'],
                                               'numArgs': 1,
                                               'shortname': 'rp'},
                              'tearOff': {'args': '<type __builtin__.bool>',
                                          'longname': 'tearOff',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'to'},
                              'tearOffCopy': {'args': '<type __builtin__.str>',
                                              'longname': 'tearOffCopy',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'toc'},
                              'tearOffRestore': {'args': '<type __builtin__.bool>',
                                                 'longname': 'tearOffRestore',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'tor'},
                              'type': {'args': '<type __builtin__.str>',
                                       'longname': 'type',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'typ'},
                              'unParent': {'args': '<type __builtin__.bool>',
                                           'longname': 'unParent',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'up'},
                              'useTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'useTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ut'}},
                    'removedFlags': {'in': 'init'},
                    'shortFlags': {'cp': 'copy',
                                   'cs': 'createString',
                                   'ctl': 'control',
                                   'dt': 'defineTemplate',
                                   'dtg': 'docTag',
                                   'e': 'edit',
                                   'es': 'editString',
                                   'ex': 'exists',
                                   'init': 'init',
                                   'iu': 'isUnique',
                                   'l': 'label',
                                   'mbv': 'menuBarVisible',
                                   'mrl': 'menuBarRepeatLast',
                                   'ni': 'needsInit',
                                   'p': 'parent',
                                   'pmp': 'popupMenuProcedure',
                                   'q': 'query',
                                   'rp': 'replacePanel',
                                   'to': 'tearOff',
                                   'toc': 'tearOffCopy',
                                   'tor': 'tearOffRestore',
                                   'typ': 'type',
                                   'up': 'unParent',
                                   'ut': 'useTemplate'},
                    'type': 'windows'},
  'scriptedPanelType': {'flags': {'addCallback': {'args': '<type __builtin__.str>',
                                                  'longname': 'addCallback',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'acb'},
                                  'copyStateCallback': {'args': '<type __builtin__.str>',
                                                        'longname': 'copyStateCallback',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ocb'},
                                  'createCallback': {'args': '<type __builtin__.str>',
                                                     'longname': 'createCallback',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ccb'},
                                  'customView': {'args': '<type __builtin__.bool>',
                                                 'longname': 'customView',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cv'},
                                  'defineTemplate': {'args': '<type __builtin__.str>',
                                                     'longname': 'defineTemplate',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'dt'},
                                  'deleteCallback': {'args': '<type __builtin__.str>',
                                                     'longname': 'deleteCallback',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'dcb'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'hotkeyCtxClient': {'args': '<type __builtin__.str>',
                                                      'longname': 'hotkeyCtxClient',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'hkc'},
                                  'initCallback': {'args': '<type __builtin__.str>',
                                                   'longname': 'initCallback',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'icb'},
                                  'label': {'args': '<type __builtin__.str>',
                                            'longname': 'label',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'l'},
                                  'obsolete': {'args': '<type __builtin__.bool>',
                                               'longname': 'obsolete',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'o'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'removeCallback': {'args': '<type __builtin__.str>',
                                                     'longname': 'removeCallback',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'rcb'},
                                  'retainOnFileOpen': {'args': '<type __builtin__.bool>',
                                                       'longname': 'retainOnFileOpen',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'rfo'},
                                  'saveStateCallback': {'args': '<type __builtin__.str>',
                                                        'longname': 'saveStateCallback',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'scb'},
                                  'unique': {'args': '<type __builtin__.bool>',
                                             'longname': 'unique',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'u'},
                                  'useTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'useTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ut'}},
                        'shortFlags': {'acb': 'addCallback',
                                       'ccb': 'createCallback',
                                       'cv': 'customView',
                                       'dcb': 'deleteCallback',
                                       'dt': 'defineTemplate',
                                       'e': 'edit',
                                       'ex': 'exists',
                                       'hkc': 'hotkeyCtxClient',
                                       'icb': 'initCallback',
                                       'l': 'label',
                                       'o': 'obsolete',
                                       'ocb': 'copyStateCallback',
                                       'q': 'query',
                                       'rcb': 'removeCallback',
                                       'rfo': 'retainOnFileOpen',
                                       'scb': 'saveStateCallback',
                                       'u': 'unique',
                                       'ut': 'useTemplate'},
                        'type': 'windows'},
  'scrollField': {'callbackFlags': ['dropCallback',
                                    'dpc',
                                    'visibleChangeCommand',
                                    'vcc',
                                    'keyPressCommand',
                                    'kpc',
                                    'dragCallback',
                                    'dgc',
                                    'enterCommand',
                                    'ec',
                                    'command',
                                    'c',
                                    'changeCommand',
                                    'cc'],
                  'flags': {'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'changeCommand': {'args': 'script',
                                              'longname': 'changeCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cc'},
                            'clear': {'args': '<type __builtin__.bool>',
                                      'longname': 'clear',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'cl'},
                            'command': {'args': '<type __builtin__.str>',
                                        'longname': 'command',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'c'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'editable': {'args': '<type __builtin__.bool>',
                                         'longname': 'editable',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ed'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'enterCommand': {'args': 'script',
                                             'longname': 'enterCommand',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ec'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'font': {'args': '<type __builtin__.str>',
                                     'longname': 'font',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'fn'},
                            'fontPointSize': {'args': '<type __builtin__.int>',
                                              'longname': 'fontPointSize',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fns'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'insertText': {'args': '<type __builtin__.str>',
                                           'longname': 'insertText',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'it'},
                            'insertionPosition': {'args': '<type __builtin__.int>',
                                                  'longname': 'insertionPosition',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ip'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'keyPressCommand': {'args': 'script',
                                                'longname': 'keyPressCommand',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'kpc'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfLines': {'args': '<type __builtin__.int>',
                                              'longname': 'numberOfLines',
                                              'modes': ['query'],
                                              'numArgs': 1,
                                              'shortname': 'nl'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'qtFont': {'args': '<type __builtin__.str>',
                                       'longname': 'qtFont',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'qfn'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'selection': {'args': '<type __builtin__.bool>',
                                          'longname': 'selection',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'sl'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'text': {'args': '<type __builtin__.str>',
                                     'longname': 'text',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'tx'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'},
                            'wordWrap': {'args': '<type __builtin__.bool>',
                                         'longname': 'wordWrap',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ww'}},
                  'shortFlags': {'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'c': 'command',
                                 'cc': 'changeCommand',
                                 'cl': 'clear',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ec': 'enterCommand',
                                 'ed': 'editable',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'ex': 'exists',
                                 'fn': 'font',
                                 'fns': 'fontPointSize',
                                 'fpn': 'fullPathName',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'io': 'isObscured',
                                 'ip': 'insertionPosition',
                                 'it': 'insertText',
                                 'kpc': 'keyPressCommand',
                                 'm': 'manage',
                                 'nbg': 'noBackground',
                                 'nl': 'numberOfLines',
                                 'npm': 'numberOfPopupMenus',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'qfn': 'qtFont',
                                 'sbm': 'statusBarMessage',
                                 'sl': 'selection',
                                 'tx': 'text',
                                 'ut': 'useTemplate',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width',
                                 'ww': 'wordWrap'},
                  'type': 'windows'},
  'scrollLayout': {'callbackFlags': ['dropCallback',
                                     'dpc',
                                     'visibleChangeCommand',
                                     'vcc',
                                     'dragCallback',
                                     'dgc',
                                     'resizeCommand',
                                     'rc'],
                   'flags': {'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'borderVisible': {'args': '<type __builtin__.bool>',
                                               'longname': 'borderVisible',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'bv'},
                             'childArray': {'args': '<type __builtin__.bool>',
                                            'longname': 'childArray',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ca'},
                             'childResizable': {'args': '<type __builtin__.bool>',
                                                'longname': 'childResizable',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'cr'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'horizontalScrollBarThickness': {'args': '<type __builtin__.int>',
                                                              'longname': 'horizontalScrollBarThickness',
                                                              'modes': ['create',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'hst'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'minChildWidth': {'args': '<type __builtin__.int>',
                                               'longname': 'minChildWidth',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'mcw'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'numberOfChildren': {'args': '<type __builtin__.bool>',
                                                  'longname': 'numberOfChildren',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'nch'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'panEnabled': {'args': '<type __builtin__.bool>',
                                            'longname': 'panEnabled',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'pe'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'resizeCommand': {'args': 'script',
                                               'longname': 'resizeCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rc'},
                             'scrollAreaHeight': {'args': '<type __builtin__.bool>',
                                                  'longname': 'scrollAreaHeight',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'sah'},
                             'scrollAreaValue': {'args': '<type __builtin__.bool>',
                                                 'longname': 'scrollAreaValue',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'sav'},
                             'scrollAreaWidth': {'args': '<type __builtin__.bool>',
                                                 'longname': 'scrollAreaWidth',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'saw'},
                             'scrollByPixel': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'scrollByPixel',
                                               'modes': ['edit'],
                                               'numArgs': 2,
                                               'shortname': 'sbp'},
                             'scrollPage': {'args': '<type __builtin__.str>',
                                            'longname': 'scrollPage',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'sp'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'verticalScrollBarAlwaysVisible': {'args': '<type __builtin__.bool>',
                                                                'longname': 'verticalScrollBarAlwaysVisible',
                                                                'modes': ['create'],
                                                                'numArgs': 1,
                                                                'shortname': 'vsb'},
                             'verticalScrollBarThickness': {'args': '<type __builtin__.int>',
                                                            'longname': 'verticalScrollBarThickness',
                                                            'modes': ['create',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'vst'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'shortFlags': {'ann': 'annotation',
                                  'bgc': 'backgroundColor',
                                  'bv': 'borderVisible',
                                  'ca': 'childArray',
                                  'cr': 'childResizable',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ebg': 'enableBackground',
                                  'ekf': 'enableKeyboardFocus',
                                  'en': 'enable',
                                  'ex': 'exists',
                                  'fpn': 'fullPathName',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'hst': 'horizontalScrollBarThickness',
                                  'io': 'isObscured',
                                  'm': 'manage',
                                  'mcw': 'minChildWidth',
                                  'nbg': 'noBackground',
                                  'nch': 'numberOfChildren',
                                  'npm': 'numberOfPopupMenus',
                                  'p': 'parent',
                                  'pe': 'panEnabled',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'q': 'query',
                                  'rc': 'resizeCommand',
                                  'sah': 'scrollAreaHeight',
                                  'sav': 'scrollAreaValue',
                                  'saw': 'scrollAreaWidth',
                                  'sbm': 'statusBarMessage',
                                  'sbp': 'scrollByPixel',
                                  'sp': 'scrollPage',
                                  'ut': 'useTemplate',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'vsb': 'verticalScrollBarAlwaysVisible',
                                  'vst': 'verticalScrollBarThickness',
                                  'w': 'width'},
                   'type': 'windows'},
  'sculpt': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                 'longname': 'after',
                                 'modes': ['create', 'edit'],
                                 'numArgs': 0,
                                 'shortname': 'af'},
                       'afterReference': {'args': '<type __builtin__.bool>',
                                          'longname': 'afterReference',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'ar'},
                       'before': {'args': '<type __builtin__.bool>',
                                  'longname': 'before',
                                  'modes': ['create', 'edit'],
                                  'numArgs': 0,
                                  'shortname': 'bf'},
                       'deformerTools': {'args': '<type __builtin__.bool>',
                                         'longname': 'deformerTools',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'dt'},
                       'dropoffDistance': {'args': '<type __builtin__.float>',
                                           'longname': 'dropoffDistance',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dds'},
                       'dropoffType': {'args': '<type __builtin__.str>',
                                       'longname': 'dropoffType',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'drt'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'exclusive': {'args': '<type __builtin__.str>',
                                     'longname': 'exclusive',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'ex'},
                       'frontOfChain': {'args': '<type __builtin__.bool>',
                                        'longname': 'frontOfChain',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'foc'},
                       'geometry': {'args': '<type __builtin__.str>',
                                    'longname': 'geometry',
                                    'modes': ['query', 'edit', 'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'g'},
                       'geometryIndices': {'args': '<type __builtin__.bool>',
                                           'longname': 'geometryIndices',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'gi'},
                       'groupWithLocator': {'args': '<type __builtin__.bool>',
                                            'longname': 'groupWithLocator',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'gwl'},
                       'ignoreSelected': {'args': '<type __builtin__.bool>',
                                          'longname': 'ignoreSelected',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ignoreSelected'},
                       'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                   'longname': 'includeHiddenSelections',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'ihs'},
                       'insideMode': {'args': '<type __builtin__.str>',
                                      'longname': 'insideMode',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'im'},
                       'maxDisplacement': {'args': '<type __builtin__.float>',
                                           'longname': 'maxDisplacement',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'mxd'},
                       'mode': {'args': '<type __builtin__.str>',
                                'longname': 'mode',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'm'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'objectCentered': {'args': '<type __builtin__.bool>',
                                          'longname': 'objectCentered',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'oc'},
                       'parallel': {'args': '<type __builtin__.bool>',
                                    'longname': 'parallel',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'par'},
                       'prune': {'args': '<type __builtin__.bool>',
                                 'longname': 'prune',
                                 'modes': ['edit'],
                                 'numArgs': 0,
                                 'shortname': 'pr'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'remove': {'args': '<type __builtin__.bool>',
                                  'longname': 'remove',
                                  'modes': ['edit', 'multiuse'],
                                  'numArgs': 0,
                                  'shortname': 'rm'},
                       'sculptTool': {'args': '<type __builtin__.str>',
                                      'longname': 'sculptTool',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'st'},
                       'split': {'args': '<type __builtin__.bool>',
                                 'longname': 'split',
                                 'modes': ['create', 'edit'],
                                 'numArgs': 0,
                                 'shortname': 'sp'}},
             'removedFlags': {'is': 'ignoreSelected'},
             'shortFlags': {'af': 'after',
                            'ar': 'afterReference',
                            'bf': 'before',
                            'dds': 'dropoffDistance',
                            'drt': 'dropoffType',
                            'dt': 'deformerTools',
                            'e': 'edit',
                            'ex': 'exclusive',
                            'foc': 'frontOfChain',
                            'g': 'geometry',
                            'gi': 'geometryIndices',
                            'gwl': 'groupWithLocator',
                            'ignoreSelected': 'ignoreSelected',
                            'ihs': 'includeHiddenSelections',
                            'im': 'insideMode',
                            'm': 'mode',
                            'mxd': 'maxDisplacement',
                            'n': 'name',
                            'oc': 'objectCentered',
                            'par': 'parallel',
                            'pr': 'prune',
                            'q': 'query',
                            'rm': 'remove',
                            'sp': 'split',
                            'st': 'sculptTool'},
             'type': 'animation'},
  'sculptMeshCacheChangeCloneSource': {'flags': {'blendShape': {'args': '<type __builtin__.str>',
                                                                'longname': 'blendShape',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'bs'},
                                                 'edit': {'args': '<type __builtin__.bool>',
                                                          'longname': 'edit',
                                                          'numArgs': 0,
                                                          'shortname': 'e'},
                                                 'query': {'args': '<type __builtin__.bool>',
                                                           'longname': 'query',
                                                           'numArgs': 0,
                                                           'shortname': 'q'},
                                                 'target': {'args': '<type __builtin__.str>',
                                                            'longname': 'target',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 't'}},
                                       'shortFlags': {'bs': 'blendShape',
                                                      'e': 'edit',
                                                      'q': 'query',
                                                      't': 'target'},
                                       'type': 'general'},
  'sculptMeshCacheCtx': {'flags': {'adjustSize': {'args': '<type __builtin__.bool>',
                                                  'longname': 'adjustSize',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'asz'},
                                   'adjustStrength': {'args': '<type __builtin__.bool>',
                                                      'longname': 'adjustStrength',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ast'},
                                   'affectAllLayers': {'args': '<type __builtin__.bool>',
                                                       'longname': 'affectAllLayers',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'aal'},
                                   'brushDirection': {'args': '<type __builtin__.int>',
                                                      'longname': 'brushDirection',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'bd'},
                                   'brushSize': {'args': '<type __builtin__.float>',
                                                 'longname': 'brushSize',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'bsz'},
                                   'brushStrength': {'args': '<type __builtin__.float>',
                                                     'longname': 'brushStrength',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'bst'},
                                   'buildUpRate': {'args': '<type __builtin__.float>',
                                                   'longname': 'buildUpRate',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'bur'},
                                   'cloneHideSource': {'args': '<type __builtin__.bool>',
                                                       'longname': 'cloneHideSource',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'chs'},
                                   'cloneMethod': {'args': '<type __builtin__.int>',
                                                   'longname': 'cloneMethod',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cm'},
                                   'cloneShapeSource': {'args': '<type __builtin__.str>',
                                                        'longname': 'cloneShapeSource',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'css'},
                                   'cloneTargetSource': {'args': '<type __builtin__.str>',
                                                         'longname': 'cloneTargetSource',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'cas'},
                                   'constrainToSurface': {'args': '<type __builtin__.bool>',
                                                          'longname': 'constrainToSurface',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'cts'},
                                   'direction': {'args': '<type __builtin__.int>',
                                                 'longname': 'direction',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'd'},
                                   'displayFrozen': {'args': '<type __builtin__.bool>',
                                                     'longname': 'displayFrozen',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'df'},
                                   'displayMask': {'args': '<type __builtin__.bool>',
                                                   'longname': 'displayMask',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dm'},
                                   'displayWireframe': {'args': '<type __builtin__.bool>',
                                                        'longname': 'displayWireframe',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'dw'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'falloffType': {'args': '<type __builtin__.int>',
                                                   'longname': 'falloffType',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ft'},
                                   'flood': {'args': '<type __builtin__.float>',
                                             'longname': 'flood',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'fl'},
                                   'floodFreeze': {'args': '<type __builtin__.float>',
                                                   'longname': 'floodFreeze',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ff'},
                                   'frame': {'args': '<type __builtin__.bool>',
                                             'longname': 'frame',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'frm'},
                                   'freezeSelection': {'args': '<type __builtin__.bool>',
                                                       'longname': 'freezeSelection',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'fsl'},
                                   'freezesmooth': {'args': '<type __builtin__.bool>',
                                                    'longname': 'freezesmooth',
                                                    'numArgs': 1,
                                                    'shortname': 'fs'},
                                   'freezesmoothandpaint': {'args': '<type __builtin__.bool>',
                                                            'longname': 'freezesmoothandpaint',
                                                            'numArgs': 1,
                                                            'shortname': 'fsp'},
                                   'grabFollowPath': {'args': '<type __builtin__.bool>',
                                                      'longname': 'grabFollowPath',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'gfp'},
                                   'grabSilhouette': {'args': '<type __builtin__.bool>',
                                                      'longname': 'grabSilhouette',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'gs'},
                                   'grabTwist': {'args': '<type __builtin__.bool>',
                                                 'longname': 'grabTwist',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'gtw'},
                                   'history': {'args': '<type __builtin__.bool>',
                                               'longname': 'history',
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'inverted': {'args': '<type __builtin__.bool>',
                                                'longname': 'inverted',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'inv'},
                                   'lastMode': {'args': '<type __builtin__.str>',
                                                'longname': 'lastMode',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'lm'},
                                   'lockShellBorder': {'args': '<type __builtin__.bool>',
                                                       'longname': 'lockShellBorder',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'lsb'},
                                   'makeStroke': {'args': ['<type __builtin__.int>',
                                                           '<type __builtin__.int>',
                                                           '<type __builtin__.int>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'makeStroke',
                                                  'modes': ['edit',
                                                            'multiuse'],
                                                  'numArgs': 5,
                                                  'shortname': 'mt'},
                                   'minSize': {'args': '<type __builtin__.float>',
                                               'longname': 'minSize',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'msz'},
                                   'minStrength': {'args': '<type __builtin__.float>',
                                                   'longname': 'minStrength',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'mst'},
                                   'mirror': {'args': '<type __builtin__.int>',
                                              'longname': 'mirror',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'mr'},
                                   'mode': {'args': '<type __builtin__.str>',
                                            'longname': 'mode',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'm'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'orientToSurface': {'args': '<type __builtin__.bool>',
                                                       'longname': 'orientToSurface',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ots'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'recordStroke': {'args': '<type __builtin__.bool>',
                                                    'longname': 'recordStroke',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'rcs'},
                                   'sculptFalloffCurve': {'args': '<type __builtin__.str>',
                                                          'longname': 'sculptFalloffCurve',
                                                          'modes': ['query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'sfc'},
                                   'size': {'args': '<type __builtin__.float>',
                                            'longname': 'size',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sz'},
                                   'stampDistance': {'args': '<type __builtin__.float>',
                                                     'longname': 'stampDistance',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 's'},
                                   'stampFile': {'args': '<type __builtin__.str>',
                                                 'longname': 'stampFile',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'stp'},
                                   'stampFlipX': {'args': '<type __builtin__.bool>',
                                                  'longname': 'stampFlipX',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sfx'},
                                   'stampFlipY': {'args': '<type __builtin__.bool>',
                                                  'longname': 'stampFlipY',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sfy'},
                                   'stampOrientToStroke': {'args': '<type __builtin__.bool>',
                                                           'longname': 'stampOrientToStroke',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'sos'},
                                   'stampPlacement': {'args': '<type __builtin__.int>',
                                                      'longname': 'stampPlacement',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'sp'},
                                   'stampRandomization': {'args': '<type __builtin__.bool>',
                                                          'longname': 'stampRandomization',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'srd'},
                                   'stampRandomizationSeed': {'args': '<type __builtin__.int>',
                                                              'longname': 'stampRandomizationSeed',
                                                              'modes': ['edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'sre'},
                                   'stampRandomizeFlipX': {'args': '<type __builtin__.bool>',
                                                           'longname': 'stampRandomizeFlipX',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'srx'},
                                   'stampRandomizeFlipY': {'args': '<type __builtin__.bool>',
                                                           'longname': 'stampRandomizeFlipY',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'sry'},
                                   'stampRandomizePosX': {'args': '<type __builtin__.float>',
                                                          'longname': 'stampRandomizePosX',
                                                          'modes': ['query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'spx'},
                                   'stampRandomizePosY': {'args': '<type __builtin__.float>',
                                                          'longname': 'stampRandomizePosY',
                                                          'modes': ['query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'spy'},
                                   'stampRandomizeRotation': {'args': '<type __builtin__.float>',
                                                              'longname': 'stampRandomizeRotation',
                                                              'modes': ['query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'srr'},
                                   'stampRandomizeScale': {'args': '<type __builtin__.float>',
                                                           'longname': 'stampRandomizeScale',
                                                           'modes': ['query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'src'},
                                   'stampRandomizeStrength': {'args': '<type __builtin__.float>',
                                                              'longname': 'stampRandomizeStrength',
                                                              'modes': ['query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'srs'},
                                   'stampRotation': {'args': '<type __builtin__.float>',
                                                     'longname': 'stampRotation',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sr'},
                                   'steadyStrokeDistance': {'args': '<type __builtin__.float>',
                                                            'longname': 'steadyStrokeDistance',
                                                            'modes': ['query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ssd'},
                                   'strength': {'args': '<type __builtin__.float>',
                                                'longname': 'strength',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'st'},
                                   'updatePlane': {'args': '<type __builtin__.bool>',
                                                   'longname': 'updatePlane',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'upl'},
                                   'useGlobalSize': {'args': '<type __builtin__.bool>',
                                                     'longname': 'useGlobalSize',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ugs'},
                                   'useScreenSpace': {'args': '<type __builtin__.bool>',
                                                      'longname': 'useScreenSpace',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ssp'},
                                   'useStampDistance': {'args': '<type __builtin__.bool>',
                                                        'longname': 'useStampDistance',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'usd'},
                                   'useStampImage': {'args': '<type __builtin__.bool>',
                                                     'longname': 'useStampImage',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'usi'},
                                   'useSteadyStroke': {'args': '<type __builtin__.bool>',
                                                       'longname': 'useSteadyStroke',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'uss'},
                                   'wholeStroke': {'args': '<type __builtin__.bool>',
                                                   'longname': 'wholeStroke',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'wst'},
                                   'wireframeAlpha': {'args': '<type __builtin__.float>',
                                                      'longname': 'wireframeAlpha',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'wa'},
                                   'wireframeColor': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'wireframeColor',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'wc'}},
                         'shortFlags': {'aal': 'affectAllLayers',
                                        'ast': 'adjustStrength',
                                        'asz': 'adjustSize',
                                        'bd': 'brushDirection',
                                        'bst': 'brushStrength',
                                        'bsz': 'brushSize',
                                        'bur': 'buildUpRate',
                                        'cas': 'cloneTargetSource',
                                        'ch': 'history',
                                        'chs': 'cloneHideSource',
                                        'cm': 'cloneMethod',
                                        'css': 'cloneShapeSource',
                                        'cts': 'constrainToSurface',
                                        'd': 'direction',
                                        'df': 'displayFrozen',
                                        'dm': 'displayMask',
                                        'dw': 'displayWireframe',
                                        'e': 'edit',
                                        'ex': 'exists',
                                        'ff': 'floodFreeze',
                                        'fl': 'flood',
                                        'frm': 'frame',
                                        'fs': 'freezesmooth',
                                        'fsl': 'freezeSelection',
                                        'fsp': 'freezesmoothandpaint',
                                        'ft': 'falloffType',
                                        'gfp': 'grabFollowPath',
                                        'gs': 'grabSilhouette',
                                        'gtw': 'grabTwist',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'inv': 'inverted',
                                        'lm': 'lastMode',
                                        'lsb': 'lockShellBorder',
                                        'm': 'mode',
                                        'mr': 'mirror',
                                        'mst': 'minStrength',
                                        'msz': 'minSize',
                                        'mt': 'makeStroke',
                                        'n': 'name',
                                        'ots': 'orientToSurface',
                                        'q': 'query',
                                        'rcs': 'recordStroke',
                                        's': 'stampDistance',
                                        'sfc': 'sculptFalloffCurve',
                                        'sfx': 'stampFlipX',
                                        'sfy': 'stampFlipY',
                                        'sos': 'stampOrientToStroke',
                                        'sp': 'stampPlacement',
                                        'spx': 'stampRandomizePosX',
                                        'spy': 'stampRandomizePosY',
                                        'sr': 'stampRotation',
                                        'src': 'stampRandomizeScale',
                                        'srd': 'stampRandomization',
                                        'sre': 'stampRandomizationSeed',
                                        'srr': 'stampRandomizeRotation',
                                        'srs': 'stampRandomizeStrength',
                                        'srx': 'stampRandomizeFlipX',
                                        'sry': 'stampRandomizeFlipY',
                                        'ssd': 'steadyStrokeDistance',
                                        'ssp': 'useScreenSpace',
                                        'st': 'strength',
                                        'stp': 'stampFile',
                                        'sz': 'size',
                                        'ugs': 'useGlobalSize',
                                        'upl': 'updatePlane',
                                        'usd': 'useStampDistance',
                                        'usi': 'useStampImage',
                                        'uss': 'useSteadyStroke',
                                        'wa': 'wireframeAlpha',
                                        'wc': 'wireframeColor',
                                        'wst': 'wholeStroke'},
                         'type': 'context'},
  'sculptTarget': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                       'longname': 'after',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'af'},
                             'afterReference': {'args': '<type __builtin__.bool>',
                                                'longname': 'afterReference',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'ar'},
                             'before': {'args': '<type __builtin__.bool>',
                                        'longname': 'before',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'bf'},
                             'deformerTools': {'args': '<type __builtin__.bool>',
                                               'longname': 'deformerTools',
                                               'modes': [],
                                               'numArgs': 0,
                                               'shortname': 'dt'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exclusive': {'args': '<type __builtin__.str>',
                                           'longname': 'exclusive',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ex'},
                             'frontOfChain': {'args': '<type __builtin__.bool>',
                                              'longname': 'frontOfChain',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'foc'},
                             'geometry': {'args': '<type __builtin__.str>',
                                          'longname': 'geometry',
                                          'modes': ['edit', 'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'g'},
                             'geometryIndices': {'args': '<type __builtin__.bool>',
                                                 'longname': 'geometryIndices',
                                                 'modes': [],
                                                 'numArgs': 0,
                                                 'shortname': 'gi'},
                             'ignoreSelected': {'args': '<type __builtin__.bool>',
                                                'longname': 'ignoreSelected',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ignoreSelected'},
                             'inbetweenWeight': {'args': '<type __builtin__.float>',
                                                 'longname': 'inbetweenWeight',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ibw'},
                             'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                         'longname': 'includeHiddenSelections',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'ihs'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'parallel': {'args': '<type __builtin__.bool>',
                                          'longname': 'parallel',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'par'},
                             'prune': {'args': '<type __builtin__.bool>',
                                       'longname': 'prune',
                                       'modes': ['edit'],
                                       'numArgs': 0,
                                       'shortname': 'pr'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'regenerate': {'args': '<type __builtin__.bool>',
                                            'longname': 'regenerate',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'r'},
                             'remove': {'args': '<type __builtin__.bool>',
                                        'longname': 'remove',
                                        'modes': ['edit', 'multiuse'],
                                        'numArgs': 0,
                                        'shortname': 'rm'},
                             'snapshot': {'args': '<type __builtin__.int>',
                                          'longname': 'snapshot',
                                          'modes': ['edit'],
                                          'numArgs': 1,
                                          'shortname': 's'},
                             'split': {'args': '<type __builtin__.bool>',
                                       'longname': 'split',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'sp'},
                             'target': {'args': '<type __builtin__.int>',
                                        'longname': 'target',
                                        'modes': ['edit'],
                                        'numArgs': 1,
                                        'shortname': 't'}},
                   'removedFlags': {'is': 'ignoreSelected'},
                   'shortFlags': {'af': 'after',
                                  'ar': 'afterReference',
                                  'bf': 'before',
                                  'dt': 'deformerTools',
                                  'e': 'edit',
                                  'ex': 'exclusive',
                                  'foc': 'frontOfChain',
                                  'g': 'geometry',
                                  'gi': 'geometryIndices',
                                  'ibw': 'inbetweenWeight',
                                  'ignoreSelected': 'ignoreSelected',
                                  'ihs': 'includeHiddenSelections',
                                  'n': 'name',
                                  'par': 'parallel',
                                  'pr': 'prune',
                                  'q': 'query',
                                  'r': 'regenerate',
                                  'rm': 'remove',
                                  's': 'snapshot',
                                  'sp': 'split',
                                  't': 'target'},
                   'type': 'animation'},
  'selLoadSettings': {'flags': {'activeProxy': {'args': '<type __builtin__.str>',
                                                'longname': 'activeProxy',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ap'},
                                'deferReference': {'args': '<type __builtin__.bool>',
                                                   'longname': 'deferReference',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dr'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'fileName': {'args': '<type __builtin__.str>',
                                             'longname': 'fileName',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'fn'},
                                'numSettings': {'args': '<type __builtin__.int>',
                                                'longname': 'numSettings',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'ns'},
                                'proxyManager': {'args': '<type __builtin__.str>',
                                                 'longname': 'proxyManager',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'pm'},
                                'proxySetFiles': {'args': '<type __builtin__.str>',
                                                  'longname': 'proxySetFiles',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'psf'},
                                'proxySetTags': {'args': '<type __builtin__.str>',
                                                 'longname': 'proxySetTags',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'pst'},
                                'proxyTag': {'args': '<type __builtin__.str>',
                                             'longname': 'proxyTag',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'pt'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'referenceNode': {'args': '<type __builtin__.str>',
                                                  'longname': 'referenceNode',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'rfn'},
                                'shortName': {'args': '<type __builtin__.bool>',
                                              'longname': 'shortName',
                                              'modes': ['create', 'query'],
                                              'numArgs': 0,
                                              'shortname': 'shn'},
                                'unresolvedName': {'args': '<type __builtin__.bool>',
                                                   'longname': 'unresolvedName',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 0,
                                                   'shortname': 'un'}},
                      'shortFlags': {'ap': 'activeProxy',
                                     'dr': 'deferReference',
                                     'e': 'edit',
                                     'fn': 'fileName',
                                     'ns': 'numSettings',
                                     'pm': 'proxyManager',
                                     'psf': 'proxySetFiles',
                                     'pst': 'proxySetTags',
                                     'pt': 'proxyTag',
                                     'q': 'query',
                                     'rfn': 'referenceNode',
                                     'shn': 'shortName',
                                     'un': 'unresolvedName'},
                      'type': 'system'},
  'select': {'flags': {'add': {'args': '<type __builtin__.bool>',
                               'longname': 'add',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'add'},
                       'addFirst': {'args': '<type __builtin__.bool>',
                                    'longname': 'addFirst',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'af'},
                       'all': {'args': '<type __builtin__.bool>',
                               'longname': 'all',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'all'},
                       'allDagObjects': {'args': '<type __builtin__.bool>',
                                         'longname': 'allDagObjects',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ado'},
                       'allDependencyNodes': {'args': '<type __builtin__.bool>',
                                              'longname': 'allDependencyNodes',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'adn'},
                       'clear': {'args': '<type __builtin__.bool>',
                                 'longname': 'clear',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'cl'},
                       'containerCentric': {'args': '<type __builtin__.bool>',
                                            'longname': 'containerCentric',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'cc'},
                       'deselect': {'args': '<type __builtin__.bool>',
                                    'longname': 'deselect',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'd'},
                       'hierarchy': {'args': '<type __builtin__.bool>',
                                     'longname': 'hierarchy',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'hi'},
                       'noExpand': {'args': '<type __builtin__.bool>',
                                    'longname': 'noExpand',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ne'},
                       'replace': {'args': '<type __builtin__.bool>',
                                   'longname': 'replace',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'r'},
                       'symmetry': {'args': '<type __builtin__.bool>',
                                    'longname': 'symmetry',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'sym'},
                       'symmetrySide': {'args': '<type __builtin__.int>',
                                        'longname': 'symmetrySide',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'sys'},
                       'toggle': {'args': '<type __builtin__.bool>',
                                  'longname': 'toggle',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'tgl'},
                       'visible': {'args': '<type __builtin__.bool>',
                                   'longname': 'visible',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'vis'}},
             'shortFlags': {'add': 'add',
                            'adn': 'allDependencyNodes',
                            'ado': 'allDagObjects',
                            'af': 'addFirst',
                            'all': 'all',
                            'cc': 'containerCentric',
                            'cl': 'clear',
                            'd': 'deselect',
                            'hi': 'hierarchy',
                            'ne': 'noExpand',
                            'r': 'replace',
                            'sym': 'symmetry',
                            'sys': 'symmetrySide',
                            'tgl': 'toggle',
                            'vis': 'visible'},
             'type': 'general'},
  'selectContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'history': {'args': '<type __builtin__.bool>',
                                          'longname': 'history',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ch'},
                              'image1': {'args': '<type __builtin__.str>',
                                         'longname': 'image1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i1'},
                              'image2': {'args': '<type __builtin__.str>',
                                         'longname': 'image2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i2'},
                              'image3': {'args': '<type __builtin__.str>',
                                         'longname': 'image3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i3'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'}},
                    'shortFlags': {'ch': 'history',
                                   'e': 'edit',
                                   'ex': 'exists',
                                   'i1': 'image1',
                                   'i2': 'image2',
                                   'i3': 'image3',
                                   'n': 'name',
                                   'q': 'query'},
                    'type': 'context'},
  'selectKey': {'flags': {'addTo': {'args': '<type __builtin__.bool>',
                                    'longname': 'addTo',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'add'},
                          'animation': {'args': '<type __builtin__.str>',
                                        'longname': 'animation',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'an'},
                          'attribute': {'args': '<type __builtin__.str>',
                                        'longname': 'attribute',
                                        'modes': ['create', 'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'at'},
                          'clear': {'args': '<type __builtin__.bool>',
                                    'longname': 'clear',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'cl'},
                          'controlPoints': {'args': '<type __builtin__.bool>',
                                            'longname': 'controlPoints',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'cp'},
                          'float': {'args': 'floatrange',
                                    'longname': 'float',
                                    'modes': ['create', 'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'f'},
                          'hierarchy': {'args': '<type __builtin__.str>',
                                        'longname': 'hierarchy',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'hi'},
                          'inTangent': {'args': '<type __builtin__.bool>',
                                        'longname': 'inTangent',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'it'},
                          'includeUpperBound': {'args': '<type __builtin__.bool>',
                                                'longname': 'includeUpperBound',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'iub'},
                          'index': {'args': '<type __builtin__.int>',
                                    'longname': 'index',
                                    'modes': ['create', 'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'index'},
                          'keyframe': {'args': '<type __builtin__.bool>',
                                       'longname': 'keyframe',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'k'},
                          'outTangent': {'args': '<type __builtin__.bool>',
                                         'longname': 'outTangent',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ot'},
                          'remove': {'args': '<type __builtin__.bool>',
                                     'longname': 'remove',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'rm'},
                          'replace': {'args': '<type __builtin__.bool>',
                                      'longname': 'replace',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'r'},
                          'shape': {'args': '<type __builtin__.bool>',
                                    'longname': 'shape',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 's'},
                          'time': {'args': 'timerange',
                                   'longname': 'time',
                                   'modes': ['create', 'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 't'},
                          'toggle': {'args': '<type __builtin__.bool>',
                                     'longname': 'toggle',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'tgl'},
                          'unsnappedKeys': {'args': '<type __builtin__.float>',
                                            'longname': 'unsnappedKeys',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'uk'}},
                'removedFlags': {'in': 'index'},
                'shortFlags': {'add': 'addTo',
                               'an': 'animation',
                               'at': 'attribute',
                               'cl': 'clear',
                               'cp': 'controlPoints',
                               'f': 'float',
                               'hi': 'hierarchy',
                               'index': 'index',
                               'it': 'inTangent',
                               'iub': 'includeUpperBound',
                               'k': 'keyframe',
                               'ot': 'outTangent',
                               'r': 'replace',
                               'rm': 'remove',
                               's': 'shape',
                               't': 'time',
                               'tgl': 'toggle',
                               'uk': 'unsnappedKeys'},
                'type': 'general'},
  'selectKeyCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'history': {'args': '<type __builtin__.bool>',
                                         'longname': 'history',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ch'},
                             'image1': {'args': '<type __builtin__.str>',
                                        'longname': 'image1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i1'},
                             'image2': {'args': '<type __builtin__.str>',
                                        'longname': 'image2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i2'},
                             'image3': {'args': '<type __builtin__.str>',
                                        'longname': 'image3',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i3'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'ch': 'history',
                                  'e': 'edit',
                                  'ex': 'exists',
                                  'i1': 'image1',
                                  'i2': 'image2',
                                  'i3': 'image3',
                                  'n': 'name',
                                  'q': 'query'},
                   'type': 'context'},
  'selectKeyframe': {'flags': {'animation': {'args': '<type __builtin__.str>',
                                             'longname': 'animation',
                                             'numArgs': 1,
                                             'shortname': 'an'},
                               'attribute': {'args': '<type __builtin__.str>',
                                             'longname': 'attribute',
                                             'modes': ['multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'at'},
                               'controlPoints': {'args': '<type __builtin__.bool>',
                                                 'longname': 'controlPoints',
                                                 'numArgs': 1,
                                                 'shortname': 'cp'},
                               'float': {'args': u'floatRange',
                                         'longname': 'float',
                                         'modes': ['multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'f'},
                               'hierarchy': {'args': '<type __builtin__.str>',
                                             'longname': 'hierarchy',
                                             'numArgs': 1,
                                             'shortname': 'hi'},
                               'includeUpperBound': {'args': '<type __builtin__.bool>',
                                                     'longname': 'includeUpperBound',
                                                     'numArgs': 1,
                                                     'shortname': 'iub'},
                               'index': {'args': u'indexRange',
                                         'longname': 'index',
                                         'modes': ['multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'index'},
                               'selectionWindow': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'selectionWindow',
                                                   'numArgs': 4,
                                                   'shortname': 'sel'},
                               'shape': {'args': '<type __builtin__.bool>',
                                         'longname': 'shape',
                                         'numArgs': 1,
                                         'shortname': 's'},
                               'time': {'args': u'timeRange',
                                        'longname': 'time',
                                        'modes': ['multiuse'],
                                        'numArgs': 1,
                                        'shortname': 't'}},
                     'removedFlags': {'in': 'index'},
                     'shortFlags': {'an': 'animation',
                                    'at': 'attribute',
                                    'cp': 'controlPoints',
                                    'f': 'float',
                                    'hi': 'hierarchy',
                                    'index': 'index',
                                    'iub': 'includeUpperBound',
                                    's': 'shape',
                                    'sel': 'selectionWindow',
                                    't': 'time'},
                     'type': 'other'},
  'selectKeyframeRegionCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                 'longname': 'edit',
                                                 'numArgs': 0,
                                                 'shortname': 'e'},
                                        'exists': {'args': '<type __builtin__.bool>',
                                                   'longname': 'exists',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'ex'},
                                        'history': {'args': '<type __builtin__.bool>',
                                                    'longname': 'history',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                                        'image1': {'args': '<type __builtin__.str>',
                                                   'longname': 'image1',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'i1'},
                                        'image2': {'args': '<type __builtin__.str>',
                                                   'longname': 'image2',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'i2'},
                                        'image3': {'args': '<type __builtin__.str>',
                                                   'longname': 'image3',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'i3'},
                                        'name': {'args': '<type __builtin__.str>',
                                                 'longname': 'name',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'n'},
                                        'query': {'args': '<type __builtin__.bool>',
                                                  'longname': 'query',
                                                  'numArgs': 0,
                                                  'shortname': 'q'}},
                              'shortFlags': {'ch': 'history',
                                             'e': 'edit',
                                             'ex': 'exists',
                                             'i1': 'image1',
                                             'i2': 'image2',
                                             'i3': 'image3',
                                             'n': 'name',
                                             'q': 'query'},
                              'type': 'context'},
  'selectMode': {'flags': {'component': {'args': '<type __builtin__.bool>',
                                         'longname': 'component',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'co'},
                           'hierarchical': {'args': '<type __builtin__.bool>',
                                            'longname': 'hierarchical',
                                            'modes': ['create', 'query'],
                                            'numArgs': 0,
                                            'shortname': 'h'},
                           'leaf': {'args': '<type __builtin__.bool>',
                                    'longname': 'leaf',
                                    'modes': ['create', 'query'],
                                    'numArgs': 0,
                                    'shortname': 'l'},
                           'object': {'args': '<type __builtin__.bool>',
                                      'longname': 'object',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'o'},
                           'preset': {'args': '<type __builtin__.bool>',
                                      'longname': 'preset',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'p'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'root': {'args': '<type __builtin__.bool>',
                                    'longname': 'root',
                                    'modes': ['create', 'query'],
                                    'numArgs': 0,
                                    'shortname': 'r'},
                           'template': {'args': '<type __builtin__.bool>',
                                        'longname': 'template',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 't'}},
                 'shortFlags': {'co': 'component',
                                'h': 'hierarchical',
                                'l': 'leaf',
                                'o': 'object',
                                'p': 'preset',
                                'q': 'query',
                                'r': 'root',
                                't': 'template'},
                 'type': 'general'},
  'selectPref': {'flags': {'affectsActive': {'args': '<type __builtin__.bool>',
                                             'longname': 'affectsActive',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'aa'},
                           'allowHiliteSelection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'allowHiliteSelection',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'ahs'},
                           'autoSelectContainer': {'args': '<type __builtin__.bool>',
                                                   'longname': 'autoSelectContainer',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'asc'},
                           'autoUseDepth': {'args': '<type __builtin__.bool>',
                                            'longname': 'autoUseDepth',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'aud'},
                           'clickBoxSize': {'args': '<type __builtin__.int>',
                                            'longname': 'clickBoxSize',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'cbs'},
                           'clickDrag': {'args': '<type __builtin__.bool>',
                                         'longname': 'clickDrag',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'cld'},
                           'containerCentricSelection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'containerCentricSelection',
                                                         'modes': ['query'],
                                                         'numArgs': 1,
                                                         'shortname': 'ccs'},
                           'disableComponentPopups': {'args': '<type __builtin__.bool>',
                                                      'longname': 'disableComponentPopups',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'dcp'},
                           'expandPopupList': {'args': '<type __builtin__.bool>',
                                               'longname': 'expandPopupList',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'epl'},
                           'ignoreSelectionPriority': {'args': '<type __builtin__.bool>',
                                                       'longname': 'ignoreSelectionPriority',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'isp'},
                           'manipClickBoxSize': {'args': '<type __builtin__.int>',
                                                 'longname': 'manipClickBoxSize',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'mcb'},
                           'paintSelect': {'args': '<type __builtin__.bool>',
                                           'longname': 'paintSelect',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 'ps'},
                           'paintSelectRefine': {'args': '<type __builtin__.bool>',
                                                 'longname': 'paintSelectRefine',
                                                 'numArgs': 1,
                                                 'shortname': 'psf'},
                           'paintSelectWithDepth': {'args': '<type __builtin__.bool>',
                                                    'longname': 'paintSelectWithDepth',
                                                    'modes': ['query'],
                                                    'numArgs': 1,
                                                    'shortname': 'psd'},
                           'popupMenuSelection': {'args': '<type __builtin__.bool>',
                                                  'longname': 'popupMenuSelection',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'pms'},
                           'preSelectBackfacing': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preSelectBackfacing',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'psb'},
                           'preSelectClosest': {'args': '<type __builtin__.bool>',
                                                'longname': 'preSelectClosest',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'psc'},
                           'preSelectDeadSpace': {'args': '<type __builtin__.int>',
                                                  'longname': 'preSelectDeadSpace',
                                                  'modes': ['query'],
                                                  'numArgs': 1,
                                                  'shortname': 'pds'},
                           'preSelectHilite': {'args': '<type __builtin__.bool>',
                                               'longname': 'preSelectHilite',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'psh'},
                           'preSelectHiliteSize': {'args': '<type __builtin__.float>',
                                                   'longname': 'preSelectHiliteSize',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'phs'},
                           'preSelectSize': {'args': '<type __builtin__.int>',
                                             'longname': 'preSelectSize',
                                             'numArgs': 1,
                                             'shortname': 'pss'},
                           'preSelectTweakDeadSpace': {'args': '<type __builtin__.int>',
                                                       'longname': 'preSelectTweakDeadSpace',
                                                       'modes': ['query'],
                                                       'numArgs': 1,
                                                       'shortname': 'pdt'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'selectTypeChangeAffectsActive': {'args': '<type __builtin__.bool>',
                                                             'longname': 'selectTypeChangeAffectsActive',
                                                             'modes': ['query'],
                                                             'numArgs': 1,
                                                             'shortname': 'stc'},
                           'selectionChildHighlightMode': {'args': '<type __builtin__.int>',
                                                           'longname': 'selectionChildHighlightMode',
                                                           'modes': ['create',
                                                                     'query'],
                                                           'numArgs': 1,
                                                           'shortname': 'sch'},
                           'singleBoxSelection': {'args': '<type __builtin__.bool>',
                                                  'longname': 'singleBoxSelection',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbs'},
                           'straightLineDistance': {'args': '<type __builtin__.bool>',
                                                    'longname': 'straightLineDistance',
                                                    'modes': ['query'],
                                                    'numArgs': 1,
                                                    'shortname': 'sld'},
                           'trackSelectionOrder': {'args': '<type __builtin__.bool>',
                                                   'longname': 'trackSelectionOrder',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'tso'},
                           'useDepth': {'args': '<type __builtin__.bool>',
                                        'longname': 'useDepth',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'ud'},
                           'xformNoSelect': {'args': '<type __builtin__.bool>',
                                             'longname': 'xformNoSelect',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'xns'}},
                 'shortFlags': {'aa': 'affectsActive',
                                'ahs': 'allowHiliteSelection',
                                'asc': 'autoSelectContainer',
                                'aud': 'autoUseDepth',
                                'cbs': 'clickBoxSize',
                                'ccs': 'containerCentricSelection',
                                'cld': 'clickDrag',
                                'dcp': 'disableComponentPopups',
                                'epl': 'expandPopupList',
                                'isp': 'ignoreSelectionPriority',
                                'mcb': 'manipClickBoxSize',
                                'pds': 'preSelectDeadSpace',
                                'pdt': 'preSelectTweakDeadSpace',
                                'phs': 'preSelectHiliteSize',
                                'pms': 'popupMenuSelection',
                                'ps': 'paintSelect',
                                'psb': 'preSelectBackfacing',
                                'psc': 'preSelectClosest',
                                'psd': 'paintSelectWithDepth',
                                'psf': 'paintSelectRefine',
                                'psh': 'preSelectHilite',
                                'pss': 'preSelectSize',
                                'q': 'query',
                                'sbs': 'singleBoxSelection',
                                'sch': 'selectionChildHighlightMode',
                                'sld': 'straightLineDistance',
                                'stc': 'selectTypeChangeAffectsActive',
                                'tso': 'trackSelectionOrder',
                                'ud': 'useDepth',
                                'xns': 'xformNoSelect'},
                 'type': 'general'},
  'selectPriority': {'flags': {'allComponents': {'args': '<type __builtin__.int>',
                                                 'longname': 'allComponents',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'alc'},
                               'allObjects': {'args': '<type __builtin__.int>',
                                              'longname': 'allObjects',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'alo'},
                               'animBreakdown': {'args': '<type __builtin__.int>',
                                                 'longname': 'animBreakdown',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'abd'},
                               'animCurve': {'args': '<type __builtin__.int>',
                                             'longname': 'animCurve',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'ac'},
                               'animInTangent': {'args': '<type __builtin__.int>',
                                                 'longname': 'animInTangent',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ait'},
                               'animKeyframe': {'args': '<type __builtin__.int>',
                                                'longname': 'animKeyframe',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'ak'},
                               'animOutTangent': {'args': '<type __builtin__.int>',
                                                  'longname': 'animOutTangent',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'aot'},
                               'byName': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.bool>'],
                                          'longname': 'byName',
                                          'modes': ['create', 'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'bn'},
                               'camera': {'args': '<type __builtin__.int>',
                                          'longname': 'camera',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'ca'},
                               'cluster': {'args': '<type __builtin__.int>',
                                           'longname': 'cluster',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'cl'},
                               'collisionModel': {'args': '<type __builtin__.int>',
                                                  'longname': 'collisionModel',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'clm'},
                               'controlVertex': {'args': '<type __builtin__.int>',
                                                 'longname': 'controlVertex',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'cv'},
                               'curve': {'args': '<type __builtin__.int>',
                                         'longname': 'curve',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'c'},
                               'curveKnot': {'args': '<type __builtin__.int>',
                                             'longname': 'curveKnot',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'ck'},
                               'curveOnSurface': {'args': '<type __builtin__.int>',
                                                  'longname': 'curveOnSurface',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'cos'},
                               'curveParameterPoint': {'args': '<type __builtin__.int>',
                                                       'longname': 'curveParameterPoint',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'cpp'},
                               'dimension': {'args': '<type __builtin__.int>',
                                             'longname': 'dimension',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'dim'},
                               'dynamicConstraint': {'args': '<type __builtin__.int>',
                                                     'longname': 'dynamicConstraint',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'dc'},
                               'edge': {'args': '<type __builtin__.int>',
                                        'longname': 'edge',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'eg'},
                               'editPoint': {'args': '<type __builtin__.int>',
                                             'longname': 'editPoint',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'ep'},
                               'emitter': {'args': '<type __builtin__.int>',
                                           'longname': 'emitter',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'em'},
                               'facet': {'args': '<type __builtin__.int>',
                                         'longname': 'facet',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'fc'},
                               'field': {'args': '<type __builtin__.int>',
                                         'longname': 'field',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'fi'},
                               'fluid': {'args': '<type __builtin__.int>',
                                         'longname': 'fluid',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'fl'},
                               'follicle': {'args': '<type __builtin__.int>',
                                            'longname': 'follicle',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'fo'},
                               'hairSystem': {'args': '<type __builtin__.int>',
                                              'longname': 'hairSystem',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'hs'},
                               'handle': {'args': '<type __builtin__.int>',
                                          'longname': 'handle',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'ha'},
                               'hull': {'args': '<type __builtin__.int>',
                                        'longname': 'hull',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'hl'},
                               'ikEndEffector': {'args': '<type __builtin__.int>',
                                                 'longname': 'ikEndEffector',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'iee'},
                               'ikHandle': {'args': '<type __builtin__.int>',
                                            'longname': 'ikHandle',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'ikh'},
                               'imagePlane': {'args': '<type __builtin__.int>',
                                              'longname': 'imagePlane',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'ip'},
                               'implicitGeometry': {'args': '<type __builtin__.int>',
                                                    'longname': 'implicitGeometry',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'ig'},
                               'isoparm': {'args': '<type __builtin__.int>',
                                           'longname': 'isoparm',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'iso'},
                               'joint': {'args': '<type __builtin__.int>',
                                         'longname': 'joint',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'j'},
                               'jointPivot': {'args': '<type __builtin__.int>',
                                              'longname': 'jointPivot',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'jp'},
                               'lattice': {'args': '<type __builtin__.int>',
                                           'longname': 'lattice',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'la'},
                               'latticePoint': {'args': '<type __builtin__.int>',
                                                'longname': 'latticePoint',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'lp'},
                               'light': {'args': '<type __builtin__.int>',
                                         'longname': 'light',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'lt'},
                               'localRotationAxis': {'args': '<type __builtin__.int>',
                                                     'longname': 'localRotationAxis',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'ra'},
                               'locator': {'args': '<type __builtin__.int>',
                                           'longname': 'locator',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'lc'},
                               'locatorUV': {'args': '<type __builtin__.int>',
                                             'longname': 'locatorUV',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'luv'},
                               'locatorXYZ': {'args': '<type __builtin__.int>',
                                              'longname': 'locatorXYZ',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'xyz'},
                               'meshUVShell': {'args': '<type __builtin__.int>',
                                               'longname': 'meshUVShell',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'msh'},
                               'motionTrailPoint': {'args': '<type __builtin__.int>',
                                                    'longname': 'motionTrailPoint',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'mtp'},
                               'motionTrailTangent': {'args': '<type __builtin__.int>',
                                                      'longname': 'motionTrailTangent',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'mtt'},
                               'nCloth': {'args': '<type __builtin__.int>',
                                          'longname': 'nCloth',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'ncl'},
                               'nParticle': {'args': '<type __builtin__.int>',
                                             'longname': 'nParticle',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'npr'},
                               'nParticleShape': {'args': '<type __builtin__.int>',
                                                  'longname': 'nParticleShape',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'nps'},
                               'nRigid': {'args': '<type __builtin__.int>',
                                          'longname': 'nRigid',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'nr'},
                               'nonlinear': {'args': '<type __builtin__.int>',
                                             'longname': 'nonlinear',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'nl'},
                               'nurbsCurve': {'args': '<type __builtin__.int>',
                                              'longname': 'nurbsCurve',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'nc'},
                               'nurbsSurface': {'args': '<type __builtin__.int>',
                                                'longname': 'nurbsSurface',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'ns'},
                               'orientationLocator': {'args': '<type __builtin__.int>',
                                                      'longname': 'orientationLocator',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'ol'},
                               'particle': {'args': '<type __builtin__.int>',
                                            'longname': 'particle',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'pr'},
                               'particleShape': {'args': '<type __builtin__.int>',
                                                 'longname': 'particleShape',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ps'},
                               'plane': {'args': '<type __builtin__.int>',
                                         'longname': 'plane',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'pl'},
                               'polymesh': {'args': '<type __builtin__.int>',
                                            'longname': 'polymesh',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'p'},
                               'polymeshEdge': {'args': '<type __builtin__.int>',
                                                'longname': 'polymeshEdge',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'pe'},
                               'polymeshFace': {'args': '<type __builtin__.int>',
                                                'longname': 'polymeshFace',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'pf'},
                               'polymeshFreeEdge': {'args': '<type __builtin__.int>',
                                                    'longname': 'polymeshFreeEdge',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'pfe'},
                               'polymeshUV': {'args': '<type __builtin__.int>',
                                              'longname': 'polymeshUV',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'puv'},
                               'polymeshVertex': {'args': '<type __builtin__.int>',
                                                  'longname': 'polymeshVertex',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'pv'},
                               'polymeshVtxFace': {'args': '<type __builtin__.int>',
                                                   'longname': 'polymeshVtxFace',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'pvf'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'queryByName': {'args': '<type __builtin__.str>',
                                               'longname': 'queryByName',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'qbn'},
                               'rigidBody': {'args': '<type __builtin__.int>',
                                             'longname': 'rigidBody',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'rb'},
                               'rigidConstraint': {'args': '<type __builtin__.int>',
                                                   'longname': 'rigidConstraint',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'rc'},
                               'rotatePivot': {'args': '<type __builtin__.int>',
                                               'longname': 'rotatePivot',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'rp'},
                               'scalePivot': {'args': '<type __builtin__.int>',
                                              'longname': 'scalePivot',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'sp'},
                               'sculpt': {'args': '<type __builtin__.int>',
                                          'longname': 'sculpt',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'sc'},
                               'selectHandle': {'args': '<type __builtin__.int>',
                                                'longname': 'selectHandle',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'sh'},
                               'spring': {'args': '<type __builtin__.int>',
                                          'longname': 'spring',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'spr'},
                               'springComponent': {'args': '<type __builtin__.int>',
                                                   'longname': 'springComponent',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'spc'},
                               'stroke': {'args': '<type __builtin__.int>',
                                          'longname': 'stroke',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'str'},
                               'subdiv': {'args': '<type __builtin__.int>',
                                          'longname': 'subdiv',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'sd'},
                               'subdivMeshEdge': {'args': '<type __builtin__.int>',
                                                  'longname': 'subdivMeshEdge',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'sme'},
                               'subdivMeshFace': {'args': '<type __builtin__.int>',
                                                  'longname': 'subdivMeshFace',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'smf'},
                               'subdivMeshPoint': {'args': '<type __builtin__.int>',
                                                   'longname': 'subdivMeshPoint',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'smp'},
                               'subdivMeshUV': {'args': '<type __builtin__.int>',
                                                'longname': 'subdivMeshUV',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'smu'},
                               'surfaceEdge': {'args': '<type __builtin__.int>',
                                               'longname': 'surfaceEdge',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'se'},
                               'surfaceFace': {'args': '<type __builtin__.int>',
                                               'longname': 'surfaceFace',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'sf'},
                               'surfaceKnot': {'args': '<type __builtin__.int>',
                                               'longname': 'surfaceKnot',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'sk'},
                               'surfaceParameterPoint': {'args': '<type __builtin__.int>',
                                                         'longname': 'surfaceParameterPoint',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'spp'},
                               'surfaceRange': {'args': '<type __builtin__.int>',
                                                'longname': 'surfaceRange',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'sr'},
                               'texture': {'args': '<type __builtin__.int>',
                                           'longname': 'texture',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'tx'},
                               'vertex': {'args': '<type __builtin__.int>',
                                          'longname': 'vertex',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'v'}},
                     'shortFlags': {'abd': 'animBreakdown',
                                    'ac': 'animCurve',
                                    'ait': 'animInTangent',
                                    'ak': 'animKeyframe',
                                    'alc': 'allComponents',
                                    'alo': 'allObjects',
                                    'aot': 'animOutTangent',
                                    'bn': 'byName',
                                    'c': 'curve',
                                    'ca': 'camera',
                                    'ck': 'curveKnot',
                                    'cl': 'cluster',
                                    'clm': 'collisionModel',
                                    'cos': 'curveOnSurface',
                                    'cpp': 'curveParameterPoint',
                                    'cv': 'controlVertex',
                                    'dc': 'dynamicConstraint',
                                    'dim': 'dimension',
                                    'eg': 'edge',
                                    'em': 'emitter',
                                    'ep': 'editPoint',
                                    'fc': 'facet',
                                    'fi': 'field',
                                    'fl': 'fluid',
                                    'fo': 'follicle',
                                    'ha': 'handle',
                                    'hl': 'hull',
                                    'hs': 'hairSystem',
                                    'iee': 'ikEndEffector',
                                    'ig': 'implicitGeometry',
                                    'ikh': 'ikHandle',
                                    'ip': 'imagePlane',
                                    'iso': 'isoparm',
                                    'j': 'joint',
                                    'jp': 'jointPivot',
                                    'la': 'lattice',
                                    'lc': 'locator',
                                    'lp': 'latticePoint',
                                    'lt': 'light',
                                    'luv': 'locatorUV',
                                    'msh': 'meshUVShell',
                                    'mtp': 'motionTrailPoint',
                                    'mtt': 'motionTrailTangent',
                                    'nc': 'nurbsCurve',
                                    'ncl': 'nCloth',
                                    'nl': 'nonlinear',
                                    'npr': 'nParticle',
                                    'nps': 'nParticleShape',
                                    'nr': 'nRigid',
                                    'ns': 'nurbsSurface',
                                    'ol': 'orientationLocator',
                                    'p': 'polymesh',
                                    'pe': 'polymeshEdge',
                                    'pf': 'polymeshFace',
                                    'pfe': 'polymeshFreeEdge',
                                    'pl': 'plane',
                                    'pr': 'particle',
                                    'ps': 'particleShape',
                                    'puv': 'polymeshUV',
                                    'pv': 'polymeshVertex',
                                    'pvf': 'polymeshVtxFace',
                                    'q': 'query',
                                    'qbn': 'queryByName',
                                    'ra': 'localRotationAxis',
                                    'rb': 'rigidBody',
                                    'rc': 'rigidConstraint',
                                    'rp': 'rotatePivot',
                                    'sc': 'sculpt',
                                    'sd': 'subdiv',
                                    'se': 'surfaceEdge',
                                    'sf': 'surfaceFace',
                                    'sh': 'selectHandle',
                                    'sk': 'surfaceKnot',
                                    'sme': 'subdivMeshEdge',
                                    'smf': 'subdivMeshFace',
                                    'smp': 'subdivMeshPoint',
                                    'smu': 'subdivMeshUV',
                                    'sp': 'scalePivot',
                                    'spc': 'springComponent',
                                    'spp': 'surfaceParameterPoint',
                                    'spr': 'spring',
                                    'sr': 'surfaceRange',
                                    'str': 'stroke',
                                    'tx': 'texture',
                                    'v': 'vertex',
                                    'xyz': 'locatorXYZ'},
                     'type': 'general'},
  'selectType': {'flags': {'allComponents': {'args': '<type __builtin__.bool>',
                                             'longname': 'allComponents',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'alc'},
                           'allObjects': {'args': '<type __builtin__.bool>',
                                          'longname': 'allObjects',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'alo'},
                           'animBreakdown': {'args': '<type __builtin__.bool>',
                                             'longname': 'animBreakdown',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'abd'},
                           'animCurve': {'args': '<type __builtin__.bool>',
                                         'longname': 'animCurve',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'ac'},
                           'animInTangent': {'args': '<type __builtin__.bool>',
                                             'longname': 'animInTangent',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'ait'},
                           'animKeyframe': {'args': '<type __builtin__.bool>',
                                            'longname': 'animKeyframe',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'ak'},
                           'animOutTangent': {'args': '<type __builtin__.bool>',
                                              'longname': 'animOutTangent',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'aot'},
                           'byName': {'args': ['<type __builtin__.str>',
                                               '<type __builtin__.bool>'],
                                      'longname': 'byName',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 2,
                                      'shortname': 'bn'},
                           'camera': {'args': '<type __builtin__.bool>',
                                      'longname': 'camera',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'ca'},
                           'cluster': {'args': '<type __builtin__.bool>',
                                       'longname': 'cluster',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'cl'},
                           'collisionModel': {'args': '<type __builtin__.bool>',
                                              'longname': 'collisionModel',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'clm'},
                           'controlVertex': {'args': '<type __builtin__.bool>',
                                             'longname': 'controlVertex',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'cv'},
                           'curve': {'args': '<type __builtin__.bool>',
                                     'longname': 'curve',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'c'},
                           'curveKnot': {'args': '<type __builtin__.bool>',
                                         'longname': 'curveKnot',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'ck'},
                           'curveOnSurface': {'args': '<type __builtin__.bool>',
                                              'longname': 'curveOnSurface',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'cos'},
                           'curveParameterPoint': {'args': '<type __builtin__.bool>',
                                                   'longname': 'curveParameterPoint',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'cpp'},
                           'dimension': {'args': '<type __builtin__.bool>',
                                         'longname': 'dimension',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'dim'},
                           'dynamicConstraint': {'args': '<type __builtin__.bool>',
                                                 'longname': 'dynamicConstraint',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'dc'},
                           'edge': {'args': '<type __builtin__.bool>',
                                    'longname': 'edge',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'eg'},
                           'editPoint': {'args': '<type __builtin__.bool>',
                                         'longname': 'editPoint',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'ep'},
                           'emitter': {'args': '<type __builtin__.bool>',
                                       'longname': 'emitter',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'em'},
                           'facet': {'args': '<type __builtin__.bool>',
                                     'longname': 'facet',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'fc'},
                           'field': {'args': '<type __builtin__.bool>',
                                     'longname': 'field',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'fi'},
                           'fluid': {'args': '<type __builtin__.bool>',
                                     'longname': 'fluid',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'fl'},
                           'follicle': {'args': '<type __builtin__.bool>',
                                        'longname': 'follicle',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'fo'},
                           'hairSystem': {'args': '<type __builtin__.bool>',
                                          'longname': 'hairSystem',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'hs'},
                           'handle': {'args': '<type __builtin__.bool>',
                                      'longname': 'handle',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'ha'},
                           'hull': {'args': '<type __builtin__.bool>',
                                    'longname': 'hull',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'hl'},
                           'ikEndEffector': {'args': '<type __builtin__.bool>',
                                             'longname': 'ikEndEffector',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'iee'},
                           'ikHandle': {'args': '<type __builtin__.bool>',
                                        'longname': 'ikHandle',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'ikh'},
                           'imagePlane': {'args': '<type __builtin__.bool>',
                                          'longname': 'imagePlane',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'ip'},
                           'implicitGeometry': {'args': '<type __builtin__.bool>',
                                                'longname': 'implicitGeometry',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'ig'},
                           'isoparm': {'args': '<type __builtin__.bool>',
                                       'longname': 'isoparm',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'iso'},
                           'joint': {'args': '<type __builtin__.bool>',
                                     'longname': 'joint',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'j'},
                           'jointPivot': {'args': '<type __builtin__.bool>',
                                          'longname': 'jointPivot',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'jp'},
                           'lattice': {'args': '<type __builtin__.bool>',
                                       'longname': 'lattice',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'la'},
                           'latticePoint': {'args': '<type __builtin__.bool>',
                                            'longname': 'latticePoint',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'lp'},
                           'light': {'args': '<type __builtin__.bool>',
                                     'longname': 'light',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'lt'},
                           'localRotationAxis': {'args': '<type __builtin__.bool>',
                                                 'longname': 'localRotationAxis',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ra'},
                           'locator': {'args': '<type __builtin__.bool>',
                                       'longname': 'locator',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'lc'},
                           'locatorUV': {'args': '<type __builtin__.bool>',
                                         'longname': 'locatorUV',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'luv'},
                           'locatorXYZ': {'args': '<type __builtin__.bool>',
                                          'longname': 'locatorXYZ',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'xyz'},
                           'meshComponents': {'args': '<type __builtin__.bool>',
                                              'longname': 'meshComponents',
                                              'numArgs': 1,
                                              'shortname': 'mc'},
                           'meshUVShell': {'args': '<type __builtin__.bool>',
                                           'longname': 'meshUVShell',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'msh'},
                           'motionTrailPoint': {'args': '<type __builtin__.bool>',
                                                'longname': 'motionTrailPoint',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'mtp'},
                           'motionTrailTangent': {'args': '<type __builtin__.bool>',
                                                  'longname': 'motionTrailTangent',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'mtt'},
                           'nCloth': {'args': '<type __builtin__.bool>',
                                      'longname': 'nCloth',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'ncl'},
                           'nParticle': {'args': '<type __builtin__.bool>',
                                         'longname': 'nParticle',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'npr'},
                           'nParticleShape': {'args': '<type __builtin__.bool>',
                                              'longname': 'nParticleShape',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'nps'},
                           'nRigid': {'args': '<type __builtin__.bool>',
                                      'longname': 'nRigid',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'nr'},
                           'nonlinear': {'args': '<type __builtin__.bool>',
                                         'longname': 'nonlinear',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'nl'},
                           'nurbsCurve': {'args': '<type __builtin__.bool>',
                                          'longname': 'nurbsCurve',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'nc'},
                           'nurbsSurface': {'args': '<type __builtin__.bool>',
                                            'longname': 'nurbsSurface',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'ns'},
                           'objectComponent': {'args': '<type __builtin__.bool>',
                                               'longname': 'objectComponent',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 0,
                                               'shortname': 'ocm'},
                           'orientationLocator': {'args': '<type __builtin__.bool>',
                                                  'longname': 'orientationLocator',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'ol'},
                           'particle': {'args': '<type __builtin__.bool>',
                                        'longname': 'particle',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'pr'},
                           'particleShape': {'args': '<type __builtin__.bool>',
                                             'longname': 'particleShape',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'ps'},
                           'plane': {'args': '<type __builtin__.bool>',
                                     'longname': 'plane',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'pl'},
                           'polymesh': {'args': '<type __builtin__.bool>',
                                        'longname': 'polymesh',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                           'polymeshEdge': {'args': '<type __builtin__.bool>',
                                            'longname': 'polymeshEdge',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'pe'},
                           'polymeshFace': {'args': '<type __builtin__.bool>',
                                            'longname': 'polymeshFace',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'pf'},
                           'polymeshFreeEdge': {'args': '<type __builtin__.bool>',
                                                'longname': 'polymeshFreeEdge',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'pfe'},
                           'polymeshUV': {'args': '<type __builtin__.bool>',
                                          'longname': 'polymeshUV',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'puv'},
                           'polymeshVertex': {'args': '<type __builtin__.bool>',
                                              'longname': 'polymeshVertex',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'pv'},
                           'polymeshVtxFace': {'args': '<type __builtin__.bool>',
                                               'longname': 'polymeshVtxFace',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'pvf'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'queryByName': {'args': '<type __builtin__.str>',
                                           'longname': 'queryByName',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 'qbn'},
                           'rigidBody': {'args': '<type __builtin__.bool>',
                                         'longname': 'rigidBody',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'rb'},
                           'rigidConstraint': {'args': '<type __builtin__.bool>',
                                               'longname': 'rigidConstraint',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'rc'},
                           'rotatePivot': {'args': '<type __builtin__.bool>',
                                           'longname': 'rotatePivot',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'rp'},
                           'scalePivot': {'args': '<type __builtin__.bool>',
                                          'longname': 'scalePivot',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'sp'},
                           'sculpt': {'args': '<type __builtin__.bool>',
                                      'longname': 'sculpt',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'sc'},
                           'selectHandle': {'args': '<type __builtin__.bool>',
                                            'longname': 'selectHandle',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'sh'},
                           'spring': {'args': '<type __builtin__.bool>',
                                      'longname': 'spring',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'spr'},
                           'springComponent': {'args': '<type __builtin__.bool>',
                                               'longname': 'springComponent',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'spc'},
                           'stroke': {'args': '<type __builtin__.bool>',
                                      'longname': 'stroke',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'str'},
                           'subdiv': {'args': '<type __builtin__.bool>',
                                      'longname': 'subdiv',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'sd'},
                           'subdivMeshEdge': {'args': '<type __builtin__.bool>',
                                              'longname': 'subdivMeshEdge',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'sme'},
                           'subdivMeshFace': {'args': '<type __builtin__.bool>',
                                              'longname': 'subdivMeshFace',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'smf'},
                           'subdivMeshPoint': {'args': '<type __builtin__.bool>',
                                               'longname': 'subdivMeshPoint',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'smp'},
                           'subdivMeshUV': {'args': '<type __builtin__.bool>',
                                            'longname': 'subdivMeshUV',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'smu'},
                           'surfaceEdge': {'args': '<type __builtin__.bool>',
                                           'longname': 'surfaceEdge',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'se'},
                           'surfaceFace': {'args': '<type __builtin__.bool>',
                                           'longname': 'surfaceFace',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'sf'},
                           'surfaceKnot': {'args': '<type __builtin__.bool>',
                                           'longname': 'surfaceKnot',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'sk'},
                           'surfaceParameterPoint': {'args': '<type __builtin__.bool>',
                                                     'longname': 'surfaceParameterPoint',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'spp'},
                           'surfaceRange': {'args': '<type __builtin__.bool>',
                                            'longname': 'surfaceRange',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'sr'},
                           'surfaceUV': {'args': '<type __builtin__.bool>',
                                         'longname': 'surfaceUV',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'suv'},
                           'texture': {'args': '<type __builtin__.bool>',
                                       'longname': 'texture',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'tx'},
                           'vertex': {'args': '<type __builtin__.bool>',
                                      'longname': 'vertex',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'v'}},
                 'shortFlags': {'abd': 'animBreakdown',
                                'ac': 'animCurve',
                                'ait': 'animInTangent',
                                'ak': 'animKeyframe',
                                'alc': 'allComponents',
                                'alo': 'allObjects',
                                'aot': 'animOutTangent',
                                'bn': 'byName',
                                'c': 'curve',
                                'ca': 'camera',
                                'ck': 'curveKnot',
                                'cl': 'cluster',
                                'clm': 'collisionModel',
                                'cos': 'curveOnSurface',
                                'cpp': 'curveParameterPoint',
                                'cv': 'controlVertex',
                                'dc': 'dynamicConstraint',
                                'dim': 'dimension',
                                'eg': 'edge',
                                'em': 'emitter',
                                'ep': 'editPoint',
                                'fc': 'facet',
                                'fi': 'field',
                                'fl': 'fluid',
                                'fo': 'follicle',
                                'ha': 'handle',
                                'hl': 'hull',
                                'hs': 'hairSystem',
                                'iee': 'ikEndEffector',
                                'ig': 'implicitGeometry',
                                'ikh': 'ikHandle',
                                'ip': 'imagePlane',
                                'iso': 'isoparm',
                                'j': 'joint',
                                'jp': 'jointPivot',
                                'la': 'lattice',
                                'lc': 'locator',
                                'lp': 'latticePoint',
                                'lt': 'light',
                                'luv': 'locatorUV',
                                'mc': 'meshComponents',
                                'msh': 'meshUVShell',
                                'mtp': 'motionTrailPoint',
                                'mtt': 'motionTrailTangent',
                                'nc': 'nurbsCurve',
                                'ncl': 'nCloth',
                                'nl': 'nonlinear',
                                'npr': 'nParticle',
                                'nps': 'nParticleShape',
                                'nr': 'nRigid',
                                'ns': 'nurbsSurface',
                                'ocm': 'objectComponent',
                                'ol': 'orientationLocator',
                                'p': 'polymesh',
                                'pe': 'polymeshEdge',
                                'pf': 'polymeshFace',
                                'pfe': 'polymeshFreeEdge',
                                'pl': 'plane',
                                'pr': 'particle',
                                'ps': 'particleShape',
                                'puv': 'polymeshUV',
                                'pv': 'polymeshVertex',
                                'pvf': 'polymeshVtxFace',
                                'q': 'query',
                                'qbn': 'queryByName',
                                'ra': 'localRotationAxis',
                                'rb': 'rigidBody',
                                'rc': 'rigidConstraint',
                                'rp': 'rotatePivot',
                                'sc': 'sculpt',
                                'sd': 'subdiv',
                                'se': 'surfaceEdge',
                                'sf': 'surfaceFace',
                                'sh': 'selectHandle',
                                'sk': 'surfaceKnot',
                                'sme': 'subdivMeshEdge',
                                'smf': 'subdivMeshFace',
                                'smp': 'subdivMeshPoint',
                                'smu': 'subdivMeshUV',
                                'sp': 'scalePivot',
                                'spc': 'springComponent',
                                'spp': 'surfaceParameterPoint',
                                'spr': 'spring',
                                'sr': 'surfaceRange',
                                'str': 'stroke',
                                'suv': 'surfaceUV',
                                'tx': 'texture',
                                'v': 'vertex',
                                'xyz': 'locatorXYZ'},
                 'type': 'general'},
  'selectedNodes': {'flags': {'dagObjects': {'args': '<type __builtin__.bool>',
                                             'longname': 'dagObjects',
                                             'numArgs': 0,
                                             'shortname': 'do'}},
                    'shortFlags': {'do': 'dagObjects'},
                    'type': 'general'},
  'selectionConnection': {'callbackFlags': ['removeScript',
                                            'rs',
                                            'addScript',
                                            'addScript'],
                          'flags': {'activeCacheList': {'args': '<type __builtin__.bool>',
                                                        'longname': 'activeCacheList',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'atc'},
                                    'activeCharacterList': {'args': '<type __builtin__.bool>',
                                                            'longname': 'activeCharacterList',
                                                            'modes': ['create'],
                                                            'numArgs': 0,
                                                            'shortname': 'acl'},
                                    'activeList': {'args': '<type __builtin__.bool>',
                                                   'longname': 'activeList',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'act'},
                                    'addScript': {'args': 'script',
                                                  'longname': 'addScript',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'addScript'},
                                    'addTo': {'args': '<type __builtin__.str>',
                                              'longname': 'addTo',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'add'},
                                    'characterList': {'args': '<type __builtin__.bool>',
                                                      'longname': 'characterList',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'cl'},
                                    'clear': {'args': '<type __builtin__.bool>',
                                              'longname': 'clear',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'clr'},
                                    'connectionList': {'args': '<type __builtin__.bool>',
                                                       'longname': 'connectionList',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 0,
                                                       'shortname': 'lst'},
                                    'defineTemplate': {'args': '<type __builtin__.str>',
                                                       'longname': 'defineTemplate',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'dt'},
                                    'deselect': {'args': 'PyNode',
                                                 'longname': 'deselect',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'd'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'editor': {'args': '<type __builtin__.str>',
                                               'longname': 'editor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ed'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'filter': {'args': '<type __builtin__.str>',
                                               'longname': 'filter',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'f'},
                                    'findObject': {'args': 'PyNode',
                                                   'longname': 'findObject',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'fo'},
                                    'g': {'args': '<type __builtin__.bool>',
                                          'longname': 'g',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'g'},
                                    'highlightList': {'args': '<type __builtin__.bool>',
                                                      'longname': 'highlightList',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'hl'},
                                    'identify': {'args': '<type __builtin__.bool>',
                                                 'longname': 'identify',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'id'},
                                    'keyframeList': {'args': '<type __builtin__.bool>',
                                                     'longname': 'keyframeList',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'key'},
                                    'lock': {'args': '<type __builtin__.bool>',
                                             'longname': 'lock',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lck'},
                                    'modelList': {'args': '<type __builtin__.bool>',
                                                  'longname': 'modelList',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'mdl'},
                                    'object': {'args': 'PyNode',
                                               'longname': 'object',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'obj'},
                                    'parent': {'args': '<type __builtin__.str>',
                                               'longname': 'parent',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'p'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'remove': {'args': '<type __builtin__.str>',
                                               'longname': 'remove',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rm'},
                                    'removeScript': {'args': 'script',
                                                     'longname': 'removeScript',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'rs'},
                                    'select': {'args': 'PyNode',
                                               'longname': 'select',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 's'},
                                    'setList': {'args': '<type __builtin__.bool>',
                                                'longname': 'setList',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'sl'},
                                    'switch': {'args': '<type __builtin__.bool>',
                                               'longname': 'switch',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 0,
                                               'shortname': 'sw'},
                                    'useTemplate': {'args': '<type __builtin__.str>',
                                                    'longname': 'useTemplate',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ut'},
                                    'worldList': {'args': '<type __builtin__.bool>',
                                                  'longname': 'worldList',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'wl'}},
                          'removedFlags': {'as': 'addScript', 'global': 'g'},
                          'shortFlags': {'acl': 'activeCharacterList',
                                         'act': 'activeList',
                                         'add': 'addTo',
                                         'addScript': 'addScript',
                                         'atc': 'activeCacheList',
                                         'cl': 'characterList',
                                         'clr': 'clear',
                                         'd': 'deselect',
                                         'dt': 'defineTemplate',
                                         'e': 'edit',
                                         'ed': 'editor',
                                         'ex': 'exists',
                                         'f': 'filter',
                                         'fo': 'findObject',
                                         'g': 'g',
                                         'hl': 'highlightList',
                                         'id': 'identify',
                                         'key': 'keyframeList',
                                         'lck': 'lock',
                                         'lst': 'connectionList',
                                         'mdl': 'modelList',
                                         'obj': 'object',
                                         'p': 'parent',
                                         'q': 'query',
                                         'rm': 'remove',
                                         'rs': 'removeScript',
                                         's': 'select',
                                         'sl': 'setList',
                                         'sw': 'switch',
                                         'ut': 'useTemplate',
                                         'wl': 'worldList'},
                          'type': 'language'},
  'separator': {'callbackFlags': ['visibleChangeCommand',
                                  'vcc',
                                  'dragCallback',
                                  'dgc',
                                  'dropCallback',
                                  'dpc'],
                'flags': {'annotation': {'args': '<type __builtin__.str>',
                                         'longname': 'annotation',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ann'},
                          'backgroundColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'backgroundColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'bgc'},
                          'defineTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'defineTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dt'},
                          'docTag': {'args': '<type __builtin__.str>',
                                     'longname': 'docTag',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'dtg'},
                          'dragCallback': {'args': 'script',
                                           'longname': 'dragCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dgc'},
                          'dropCallback': {'args': 'script',
                                           'longname': 'dropCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dpc'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'enable': {'args': '<type __builtin__.bool>',
                                     'longname': 'enable',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'en'},
                          'enableBackground': {'args': '<type __builtin__.bool>',
                                               'longname': 'enableBackground',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ebg'},
                          'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableKeyboardFocus',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ekf'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'fullPathName': {'args': '<type __builtin__.bool>',
                                           'longname': 'fullPathName',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'fpn'},
                          'height': {'args': '<type __builtin__.int>',
                                     'longname': 'height',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'h'},
                          'highlightColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'highlightColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'hlc'},
                          'horizontal': {'args': '<type __builtin__.bool>',
                                         'longname': 'horizontal',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'hr'},
                          'isObscured': {'args': '<type __builtin__.bool>',
                                         'longname': 'isObscured',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'io'},
                          'manage': {'args': '<type __builtin__.bool>',
                                     'longname': 'manage',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'm'},
                          'noBackground': {'args': '<type __builtin__.bool>',
                                           'longname': 'noBackground',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nbg'},
                          'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'numberOfPopupMenus',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'npm'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                          'popupMenuArray': {'args': '<type __builtin__.bool>',
                                             'longname': 'popupMenuArray',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'pma'},
                          'preventOverride': {'args': '<type __builtin__.bool>',
                                              'longname': 'preventOverride',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'po'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'statusBarMessage': {'args': '<type __builtin__.str>',
                                               'longname': 'statusBarMessage',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sbm'},
                          'style': {'args': '<type __builtin__.str>',
                                    'longname': 'style',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'st'},
                          'useTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'useTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ut'},
                          'visible': {'args': '<type __builtin__.bool>',
                                      'longname': 'visible',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'vis'},
                          'visibleChangeCommand': {'args': 'script',
                                                   'longname': 'visibleChangeCommand',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'vcc'},
                          'width': {'args': '<type __builtin__.int>',
                                    'longname': 'width',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'w'}},
                'shortFlags': {'ann': 'annotation',
                               'bgc': 'backgroundColor',
                               'dgc': 'dragCallback',
                               'dpc': 'dropCallback',
                               'dt': 'defineTemplate',
                               'dtg': 'docTag',
                               'e': 'edit',
                               'ebg': 'enableBackground',
                               'ekf': 'enableKeyboardFocus',
                               'en': 'enable',
                               'ex': 'exists',
                               'fpn': 'fullPathName',
                               'h': 'height',
                               'hlc': 'highlightColor',
                               'hr': 'horizontal',
                               'io': 'isObscured',
                               'm': 'manage',
                               'nbg': 'noBackground',
                               'npm': 'numberOfPopupMenus',
                               'p': 'parent',
                               'pma': 'popupMenuArray',
                               'po': 'preventOverride',
                               'q': 'query',
                               'sbm': 'statusBarMessage',
                               'st': 'style',
                               'ut': 'useTemplate',
                               'vcc': 'visibleChangeCommand',
                               'vis': 'visible',
                               'w': 'width'},
                'type': 'windows'},
  'sequenceManager': {'flags': {'addSequencerAudio': {'args': '<type __builtin__.str>',
                                                      'longname': 'addSequencerAudio',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'asa'},
                                'attachSequencerAudio': {'args': '<type __builtin__.str>',
                                                         'longname': 'attachSequencerAudio',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'ata'},
                                'currentShot': {'args': '<type __builtin__.str>',
                                                'longname': 'currentShot',
                                                'modes': ['query'],
                                                'numArgs': 1,
                                                'shortname': 'cs'},
                                'currentTime': {'args': 'time',
                                                'longname': 'currentTime',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'ct'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'listSequencerAudio': {'args': '<type __builtin__.str>',
                                                       'longname': 'listSequencerAudio',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'lsa'},
                                'listShots': {'args': '<type __builtin__.bool>',
                                              'longname': 'listShots',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'lsh'},
                                'modelPanel': {'args': '<type __builtin__.str>',
                                               'longname': 'modelPanel',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'mp'},
                                'node': {'args': '<type __builtin__.str>',
                                         'longname': 'node',
                                         'modes': ['query'],
                                         'numArgs': 1,
                                         'shortname': 'nd'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'writableSequencer': {'args': '<type __builtin__.str>',
                                                      'longname': 'writableSequencer',
                                                      'modes': ['query'],
                                                      'numArgs': 1,
                                                      'shortname': 'ws'}},
                      'shortFlags': {'asa': 'addSequencerAudio',
                                     'ata': 'attachSequencerAudio',
                                     'cs': 'currentShot',
                                     'ct': 'currentTime',
                                     'e': 'edit',
                                     'lsa': 'listSequencerAudio',
                                     'lsh': 'listShots',
                                     'mp': 'modelPanel',
                                     'nd': 'node',
                                     'q': 'query',
                                     'ws': 'writableSequencer'},
                      'type': 'animation'},
  'setAttr': {'flags': {'alteredValue': {'args': '<type __builtin__.bool>',
                                         'longname': 'alteredValue',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'av'},
                        'caching': {'args': '<type __builtin__.bool>',
                                    'longname': 'caching',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'ca'},
                        'capacityHint': {'args': '<type __builtin__.int>',
                                         'longname': 'capacityHint',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ch'},
                        'channelBox': {'args': '<type __builtin__.bool>',
                                       'longname': 'channelBox',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'cb'},
                        'clamp': {'args': '<type __builtin__.bool>',
                                  'longname': 'clamp',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'c'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'keyable': {'args': '<type __builtin__.bool>',
                                    'longname': 'keyable',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'k'},
                        'lock': {'args': '<type __builtin__.bool>',
                                 'longname': 'lock',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'l'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'size': {'args': '<type __builtin__.int>',
                                 'longname': 'size',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 's'},
                        'type': {'args': '<type __builtin__.str>',
                                 'longname': 'type',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'typ'}},
              'shortFlags': {'av': 'alteredValue',
                             'c': 'clamp',
                             'ca': 'caching',
                             'cb': 'channelBox',
                             'ch': 'capacityHint',
                             'e': 'edit',
                             'k': 'keyable',
                             'l': 'lock',
                             'q': 'query',
                             's': 'size',
                             'typ': 'type'},
              'type': 'general'},
  'setAttrMapping': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                            'longname': 'absolute',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'a'},
                               'attribute': {'args': '<type __builtin__.str>',
                                             'longname': 'attribute',
                                             'modes': ['create',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'at'},
                               'axis': {'args': '<type __builtin__.str>',
                                        'longname': 'axis',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ax'},
                               'clutch': {'args': '<type __builtin__.str>',
                                          'longname': 'clutch',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'c'},
                               'device': {'args': '<type __builtin__.str>',
                                          'longname': 'device',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'd'},
                               'offset': {'args': '<type __builtin__.float>',
                                          'longname': 'offset',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'o'},
                               'relative': {'args': '<type __builtin__.bool>',
                                            'longname': 'relative',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'r'},
                               'scale': {'args': '<type __builtin__.float>',
                                         'longname': 'scale',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 's'},
                               'selection': {'args': '<type __builtin__.bool>',
                                             'longname': 'selection',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'sl'}},
                     'shortFlags': {},
                     'type': 'system'},
  'setDefaultShadingGroup': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                                 'longname': 'query',
                                                 'numArgs': 0,
                                                 'shortname': 'q'}},
                             'shortFlags': {'q': 'query'},
                             'type': 'rendering'},
  'setDrivenKeyframe': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                                'longname': 'attribute',
                                                'modes': ['create',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'at'},
                                  'controlPoints': {'args': '<type __builtin__.bool>',
                                                    'longname': 'controlPoints',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'cp'},
                                  'currentDriver': {'args': '<type __builtin__.str>',
                                                    'longname': 'currentDriver',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'cd'},
                                  'driven': {'args': '<type __builtin__.bool>',
                                             'longname': 'driven',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'dn'},
                                  'driver': {'args': '<type __builtin__.bool>',
                                             'longname': 'driver',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'dr'},
                                  'driverValue': {'args': '<type __builtin__.float>',
                                                  'longname': 'driverValue',
                                                  'modes': ['create',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'dv'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'hierarchy': {'args': '<type __builtin__.str>',
                                                'longname': 'hierarchy',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'hi'},
                                  'inTangentType': {'args': '<type __builtin__.str>',
                                                    'longname': 'inTangentType',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'itt'},
                                  'insert': {'args': '<type __builtin__.bool>',
                                             'longname': 'insert',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'i'},
                                  'insertBlend': {'args': '<type __builtin__.bool>',
                                                  'longname': 'insertBlend',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ib'},
                                  'outTangentType': {'args': '<type __builtin__.str>',
                                                     'longname': 'outTangentType',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ott'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'shape': {'args': '<type __builtin__.bool>',
                                            'longname': 'shape',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 's'},
                                  'value': {'args': '<type __builtin__.float>',
                                            'longname': 'value',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'v'}},
                        'shortFlags': {'at': 'attribute',
                                       'cd': 'currentDriver',
                                       'cp': 'controlPoints',
                                       'dn': 'driven',
                                       'dr': 'driver',
                                       'dv': 'driverValue',
                                       'e': 'edit',
                                       'hi': 'hierarchy',
                                       'i': 'insert',
                                       'ib': 'insertBlend',
                                       'itt': 'inTangentType',
                                       'ott': 'outTangentType',
                                       'q': 'query',
                                       's': 'shape',
                                       'v': 'value'},
                        'type': 'animation'},
  'setDynStartState': {'type': 'runtime'},
  'setDynamic': {'flags': {'allOnWhenRun': {'args': '<type __builtin__.bool>',
                                            'longname': 'allOnWhenRun',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'awr'},
                           'disableAllOnWhenRun': {'args': '<type __builtin__.bool>',
                                                   'longname': 'disableAllOnWhenRun',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'dwr'},
                           'setAll': {'args': '<type __builtin__.bool>',
                                      'longname': 'setAll',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'all'},
                           'setOff': {'args': '<type __builtin__.bool>',
                                      'longname': 'setOff',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'off'},
                           'setOn': {'args': '<type __builtin__.bool>',
                                     'longname': 'setOn',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'on'}},
                 'shortFlags': {'all': 'setAll',
                                'awr': 'allOnWhenRun',
                                'dwr': 'disableAllOnWhenRun',
                                'off': 'setOff',
                                'on': 'setOn'},
                 'type': 'effects'},
  'setEditCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'history': {'args': '<type __builtin__.bool>',
                                       'longname': 'history',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ch'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'ch': 'history',
                                'e': 'edit',
                                'ex': 'exists',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'n': 'name',
                                'q': 'query'},
                 'type': 'context'},
  'setFluidAttr': {'flags': {'addValue': {'args': '<type __builtin__.bool>',
                                          'longname': 'addValue',
                                          'modes': [],
                                          'numArgs': 0,
                                          'shortname': 'ad'},
                             'attribute': {'args': '<type __builtin__.str>',
                                           'longname': 'attribute',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'at'},
                             'clear': {'args': '<type __builtin__.bool>',
                                       'longname': 'clear',
                                       'modes': [],
                                       'numArgs': 0,
                                       'shortname': 'cl'},
                             'floatRandom': {'args': '<type __builtin__.float>',
                                             'longname': 'floatRandom',
                                             'modes': [],
                                             'numArgs': 1,
                                             'shortname': 'fr'},
                             'floatValue': {'args': '<type __builtin__.float>',
                                            'longname': 'floatValue',
                                            'modes': [],
                                            'numArgs': 1,
                                            'shortname': 'fv'},
                             'lowerFace': {'args': '<type __builtin__.bool>',
                                           'longname': 'lowerFace',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'lf'},
                             'reset': {'args': '<type __builtin__.bool>',
                                       'longname': 'reset',
                                       'modes': [],
                                       'numArgs': 0,
                                       'shortname': 're'},
                             'vectorRandom': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'vectorRandom',
                                              'modes': [],
                                              'numArgs': 3,
                                              'shortname': 'vr'},
                             'vectorValue': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'vectorValue',
                                             'modes': [],
                                             'numArgs': 3,
                                             'shortname': 'vv'},
                             'xIndex': {'args': '<type __builtin__.int>',
                                        'longname': 'xIndex',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'xi'},
                             'xvalue': {'args': '<type __builtin__.bool>',
                                        'longname': 'xvalue',
                                        'modes': [],
                                        'numArgs': 0,
                                        'shortname': 'x'},
                             'yIndex': {'args': '<type __builtin__.int>',
                                        'longname': 'yIndex',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'yi'},
                             'yvalue': {'args': '<type __builtin__.bool>',
                                        'longname': 'yvalue',
                                        'modes': [],
                                        'numArgs': 0,
                                        'shortname': 'y'},
                             'zIndex': {'args': '<type __builtin__.int>',
                                        'longname': 'zIndex',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'zi'},
                             'zvalue': {'args': '<type __builtin__.bool>',
                                        'longname': 'zvalue',
                                        'modes': [],
                                        'numArgs': 0,
                                        'shortname': 'z'}},
                   'shortFlags': {'ad': 'addValue',
                                  'at': 'attribute',
                                  'cl': 'clear',
                                  'fr': 'floatRandom',
                                  'fv': 'floatValue',
                                  'lf': 'lowerFace',
                                  're': 'reset',
                                  'vr': 'vectorRandom',
                                  'vv': 'vectorValue',
                                  'x': 'xvalue',
                                  'xi': 'xIndex',
                                  'y': 'yvalue',
                                  'yi': 'yIndex',
                                  'z': 'zvalue',
                                  'zi': 'zIndex'},
                   'type': 'effects'},
  'setFocus': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'setInfinity': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                          'longname': 'attribute',
                                          'modes': ['create', 'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'at'},
                            'controlPoints': {'args': '<type __builtin__.bool>',
                                              'longname': 'controlPoints',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'cp'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'hierarchy': {'args': '<type __builtin__.str>',
                                          'longname': 'hierarchy',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'hi'},
                            'postInfinite': {'args': '<type __builtin__.str>',
                                             'longname': 'postInfinite',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'poi'},
                            'preInfinite': {'args': '<type __builtin__.str>',
                                            'longname': 'preInfinite',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'pri'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'shape': {'args': '<type __builtin__.bool>',
                                      'longname': 'shape',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 's'}},
                  'shortFlags': {'at': 'attribute',
                                 'cp': 'controlPoints',
                                 'e': 'edit',
                                 'hi': 'hierarchy',
                                 'poi': 'postInfinite',
                                 'pri': 'preInfinite',
                                 'q': 'query',
                                 's': 'shape'},
                  'type': 'animation'},
  'setInputDeviceMapping': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                                   'longname': 'absolute',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'a'},
                                      'axis': {'args': '<type __builtin__.str>',
                                               'longname': 'axis',
                                               'modes': ['create',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'ax'},
                                      'device': {'args': '<type __builtin__.str>',
                                                 'longname': 'device',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'd'},
                                      'offset': {'args': '<type __builtin__.float>',
                                                 'longname': 'offset',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'o'},
                                      'relative': {'args': '<type __builtin__.bool>',
                                                   'longname': 'relative',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'r'},
                                      'scale': {'args': '<type __builtin__.float>',
                                                'longname': 'scale',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 's'},
                                      'view': {'args': '<type __builtin__.bool>',
                                               'longname': 'view',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'v'},
                                      'world': {'args': '<type __builtin__.bool>',
                                                'longname': 'world',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'w'}},
                            'shortFlags': {},
                            'type': 'system'},
  'setKeyCtx': {'flags': {'breakdown': {'args': '<type __builtin__.bool>',
                                        'longname': 'breakdown',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'bd'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'history': {'args': '<type __builtin__.bool>',
                                      'longname': 'history',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'ch'},
                          'image1': {'args': '<type __builtin__.str>',
                                     'longname': 'image1',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'i1'},
                          'image2': {'args': '<type __builtin__.str>',
                                     'longname': 'image2',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'i2'},
                          'image3': {'args': '<type __builtin__.str>',
                                     'longname': 'image3',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'i3'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'}},
                'shortFlags': {'bd': 'breakdown',
                               'ch': 'history',
                               'e': 'edit',
                               'ex': 'exists',
                               'i1': 'image1',
                               'i2': 'image2',
                               'i3': 'image3',
                               'n': 'name',
                               'q': 'query'},
                'type': 'context'},
  'setKeyPath': {'flags': {}, 'shortFlags': {}, 'type': 'animation'},
  'setKeyframe': {'flags': {'animLayer': {'args': '<type __builtin__.str>',
                                          'longname': 'animLayer',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'al'},
                            'animated': {'args': '<type __builtin__.bool>',
                                         'longname': 'animated',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'an'},
                            'attribute': {'args': '<type __builtin__.str>',
                                          'longname': 'attribute',
                                          'modes': ['create', 'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'at'},
                            'breakdown': {'args': '<type __builtin__.bool>',
                                          'longname': 'breakdown',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'bd'},
                            'clip': {'args': '<type __builtin__.str>',
                                     'longname': 'clip',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'c'},
                            'controlPoints': {'args': '<type __builtin__.bool>',
                                              'longname': 'controlPoints',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'cp'},
                            'dirtyDG': {'args': '<type __builtin__.bool>',
                                        'longname': 'dirtyDG',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'dd'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'float': {'args': '<type __builtin__.float>',
                                      'longname': 'float',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'f'},
                            'hierarchy': {'args': '<type __builtin__.str>',
                                          'longname': 'hierarchy',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'hi'},
                            'identity': {'args': '<type __builtin__.bool>',
                                         'longname': 'identity',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'id'},
                            'inTangentType': {'args': '<type __builtin__.str>',
                                              'longname': 'inTangentType',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'itt'},
                            'insert': {'args': '<type __builtin__.bool>',
                                       'longname': 'insert',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'i'},
                            'insertBlend': {'args': '<type __builtin__.bool>',
                                            'longname': 'insertBlend',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ib'},
                            'minimizeRotation': {'args': '<type __builtin__.bool>',
                                                 'longname': 'minimizeRotation',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'mr'},
                            'noResolve': {'args': '<type __builtin__.bool>',
                                          'longname': 'noResolve',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'nr'},
                            'outTangentType': {'args': '<type __builtin__.str>',
                                               'longname': 'outTangentType',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ott'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'respectKeyable': {'args': '<type __builtin__.bool>',
                                               'longname': 'respectKeyable',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'rk'},
                            'shape': {'args': '<type __builtin__.bool>',
                                      'longname': 'shape',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                            'time': {'args': 'time',
                                     'longname': 'time',
                                     'modes': ['create', 'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 't'},
                            'useCurrentLockedWeights': {'args': '<type __builtin__.bool>',
                                                        'longname': 'useCurrentLockedWeights',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'lw'},
                            'value': {'args': '<type __builtin__.float>',
                                      'longname': 'value',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'v'}},
                  'shortFlags': {'al': 'animLayer',
                                 'an': 'animated',
                                 'at': 'attribute',
                                 'bd': 'breakdown',
                                 'c': 'clip',
                                 'cp': 'controlPoints',
                                 'dd': 'dirtyDG',
                                 'e': 'edit',
                                 'f': 'float',
                                 'hi': 'hierarchy',
                                 'i': 'insert',
                                 'ib': 'insertBlend',
                                 'id': 'identity',
                                 'itt': 'inTangentType',
                                 'lw': 'useCurrentLockedWeights',
                                 'mr': 'minimizeRotation',
                                 'nr': 'noResolve',
                                 'ott': 'outTangentType',
                                 'q': 'query',
                                 'rk': 'respectKeyable',
                                 's': 'shape',
                                 't': 'time',
                                 'v': 'value'},
                  'type': 'animation'},
  'setKeyframeBlendshapeTargetWts': {'flags': {},
                                     'shortFlags': {},
                                     'type': 'animation'},
  'setMenuMode': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'setNClothStartState': {'type': 'runtime'},
  'setNodeTypeFlag': {'flags': {'display': {'args': '<type __builtin__.bool>',
                                            'longname': 'display',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'dsp'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'threadSafe': {'args': '<type __builtin__.bool>',
                                               'longname': 'threadSafe',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'ts'}},
                      'shortFlags': {'dsp': 'display',
                                     'q': 'query',
                                     'ts': 'threadSafe'},
                      'type': 'windows'},
  'setParent': {'flags': {'defineTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'defineTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dt'},
                          'menu': {'args': '<type __builtin__.bool>',
                                   'longname': 'menu',
                                   'modes': ['create', 'query'],
                                   'numArgs': 0,
                                   'shortname': 'm'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'topLevel': {'args': '<type __builtin__.bool>',
                                       'longname': 'topLevel',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'top'},
                          'upLevel': {'args': '<type __builtin__.bool>',
                                      'longname': 'upLevel',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'u'},
                          'useTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'useTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ut'}},
                'shortFlags': {'dt': 'defineTemplate',
                               'm': 'menu',
                               'q': 'query',
                               'top': 'topLevel',
                               'u': 'upLevel',
                               'ut': 'useTemplate'},
                'type': 'language'},
  'setParticleAttr': {'flags': {'attribute': {'args': '<type __builtin__.str>',
                                              'longname': 'attribute',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'at'},
                                'floatValue': {'args': '<type __builtin__.float>',
                                               'longname': 'floatValue',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'fv'},
                                'object': {'args': '<type __builtin__.str>',
                                           'longname': 'object',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'o'},
                                'randomFloat': {'args': '<type __builtin__.float>',
                                                'longname': 'randomFloat',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'rf'},
                                'randomVector': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'randomVector',
                                                 'modes': ['create'],
                                                 'numArgs': 3,
                                                 'shortname': 'rv'},
                                'relative': {'args': '<type __builtin__.bool>',
                                             'longname': 'relative',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'r'},
                                'vectorValue': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'vectorValue',
                                                'modes': ['create'],
                                                'numArgs': 3,
                                                'shortname': 'vv'}},
                      'shortFlags': {'at': 'attribute',
                                     'fv': 'floatValue',
                                     'o': 'object',
                                     'r': 'relative',
                                     'rf': 'randomFloat',
                                     'rv': 'randomVector',
                                     'vv': 'vectorValue'},
                      'type': 'effects'},
  'setRenderPassType': {'flags': {'defaultDataType': {'args': '<type __builtin__.bool>',
                                                      'longname': 'defaultDataType',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'd'},
                                  'numChannels': {'args': '<type __builtin__.int>',
                                                  'longname': 'numChannels',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'n'},
                                  'type': {'args': '<type __builtin__.str>',
                                           'longname': 'type',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 't'}},
                        'shortFlags': {'d': 'defaultDataType',
                                       'n': 'numChannels',
                                       't': 'type'},
                        'type': 'rendering'},
  'setStartupMessage': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'setToolTo': {'flags': {}, 'shortFlags': {}, 'type': 'general'},
  'setUITemplate': {'flags': {'popTemplate': {'args': '<type __builtin__.bool>',
                                              'longname': 'popTemplate',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ppt'},
                              'pushTemplate': {'args': '<type __builtin__.bool>',
                                               'longname': 'pushTemplate',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'pst'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'}},
                    'shortFlags': {'ppt': 'popTemplate',
                                   'pst': 'pushTemplate',
                                   'q': 'query'},
                    'type': 'language'},
  'setXformManip': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'showUnits': {'args': '<type __builtin__.bool>',
                                            'longname': 'showUnits',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'su'},
                              'suppress': {'args': '<type __builtin__.bool>',
                                           'longname': 'suppress',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 's'},
                              'useRotatePivot': {'args': '<type __builtin__.bool>',
                                                 'longname': 'useRotatePivot',
                                                 'modes': ['query'],
                                                 'numArgs': 1,
                                                 'shortname': 'urp'},
                              'worldSpace': {'args': '<type __builtin__.bool>',
                                             'longname': 'worldSpace',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'ws'}},
                    'shortFlags': {'q': 'query',
                                   's': 'suppress',
                                   'su': 'showUnits',
                                   'urp': 'useRotatePivot',
                                   'ws': 'worldSpace'},
                    'type': 'modeling'},
  'sets': {'flags': {'addElement': {'args': 'PyNode',
                                    'longname': 'addElement',
                                    'modes': ['edit'],
                                    'numArgs': 1,
                                    'shortname': 'add'},
                     'afterFilters': {'args': '<type __builtin__.bool>',
                                      'longname': 'afterFilters',
                                      'modes': ['edit'],
                                      'numArgs': 1,
                                      'shortname': 'af'},
                     'clear': {'args': 'PyNode',
                               'longname': 'clear',
                               'modes': ['edit'],
                               'numArgs': 1,
                               'shortname': 'cl'},
                     'color': {'args': '<type __builtin__.int>',
                               'longname': 'color',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'co'},
                     'copy': {'args': 'PyNode',
                              'longname': 'copy',
                              'modes': ['create'],
                              'numArgs': 1,
                              'shortname': 'cp'},
                     'edges': {'args': '<type __builtin__.bool>',
                               'longname': 'edges',
                               'modes': ['create', 'query'],
                               'numArgs': 1,
                               'shortname': 'eg'},
                     'edit': {'args': '<type __builtin__.bool>',
                              'longname': 'edit',
                              'numArgs': 0,
                              'shortname': 'e'},
                     'editPoints': {'args': '<type __builtin__.bool>',
                                    'longname': 'editPoints',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'ep'},
                     'empty': {'args': '<type __builtin__.bool>',
                               'longname': 'empty',
                               'modes': ['create'],
                               'numArgs': 0,
                               'shortname': 'em'},
                     'facets': {'args': '<type __builtin__.bool>',
                                'longname': 'facets',
                                'modes': ['create', 'query'],
                                'numArgs': 1,
                                'shortname': 'fc'},
                     'flatten': {'args': 'PyNode',
                                 'longname': 'flatten',
                                 'modes': ['edit'],
                                 'numArgs': 1,
                                 'shortname': 'fl'},
                     'forceElement': {'args': 'PyNode',
                                      'longname': 'forceElement',
                                      'modes': ['edit'],
                                      'numArgs': 1,
                                      'shortname': 'fe'},
                     'include': {'args': 'PyNode',
                                 'longname': 'include',
                                 'modes': ['edit'],
                                 'numArgs': 1,
                                 'shortname': 'include'},
                     'intersection': {'args': 'PyNode',
                                      'longname': 'intersection',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'int'},
                     'isIntersecting': {'args': 'PyNode',
                                        'longname': 'isIntersecting',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ii'},
                     'isMember': {'args': 'PyNode',
                                  'longname': 'isMember',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'im'},
                     'layer': {'args': '<type __builtin__.bool>',
                               'longname': 'layer',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'l'},
                     'name': {'args': '<type __builtin__.str>',
                              'longname': 'name',
                              'modes': ['create'],
                              'numArgs': 1,
                              'shortname': 'n'},
                     'noSurfaceShader': {'args': '<type __builtin__.bool>',
                                         'longname': 'noSurfaceShader',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'nss'},
                     'noWarnings': {'args': '<type __builtin__.bool>',
                                    'longname': 'noWarnings',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'nw'},
                     'nodesOnly': {'args': '<type __builtin__.bool>',
                                   'longname': 'nodesOnly',
                                   'modes': ['query'],
                                   'numArgs': 0,
                                   'shortname': 'no'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'remove': {'args': 'PyNode',
                                'longname': 'remove',
                                'modes': ['edit'],
                                'numArgs': 1,
                                'shortname': 'rm'},
                     'renderable': {'args': '<type __builtin__.bool>',
                                    'longname': 'renderable',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'r'},
                     'size': {'args': '<type __builtin__.bool>',
                              'longname': 'size',
                              'modes': ['query'],
                              'numArgs': 0,
                              'shortname': 's'},
                     'split': {'args': 'PyNode',
                               'longname': 'split',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'sp'},
                     'subtract': {'args': 'PyNode',
                                  'longname': 'subtract',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'sub'},
                     'text': {'args': '<type __builtin__.str>',
                              'longname': 'text',
                              'modes': ['create', 'query', 'edit'],
                              'numArgs': 1,
                              'shortname': 't'},
                     'union': {'args': 'PyNode',
                               'longname': 'union',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'un'},
                     'vertices': {'args': '<type __builtin__.bool>',
                                  'longname': 'vertices',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'v'}},
           'removedFlags': {'in': 'include'},
           'shortFlags': {'add': 'addElement',
                          'af': 'afterFilters',
                          'cl': 'clear',
                          'co': 'color',
                          'cp': 'copy',
                          'e': 'edit',
                          'eg': 'edges',
                          'em': 'empty',
                          'ep': 'editPoints',
                          'fc': 'facets',
                          'fe': 'forceElement',
                          'fl': 'flatten',
                          'ii': 'isIntersecting',
                          'im': 'isMember',
                          'include': 'include',
                          'int': 'intersection',
                          'l': 'layer',
                          'n': 'name',
                          'no': 'nodesOnly',
                          'nss': 'noSurfaceShader',
                          'nw': 'noWarnings',
                          'q': 'query',
                          'r': 'renderable',
                          'rm': 'remove',
                          's': 'size',
                          'sp': 'split',
                          'sub': 'subtract',
                          't': 'text',
                          'un': 'union',
                          'v': 'vertices'},
           'type': 'general'},
  'shadingConnection': {'flags': {'connectionState': {'args': '<type __builtin__.bool>',
                                                      'longname': 'connectionState',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cs'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'}},
                        'shortFlags': {'cs': 'connectionState',
                                       'e': 'edit',
                                       'q': 'query'},
                        'type': 'rendering'},
  'shadingGeometryRelCtx': {'callbackFlags': ['onCommand',
                                              'onc',
                                              'offCommand',
                                              'ofc'],
                            'flags': {'edit': {'args': '<type __builtin__.bool>',
                                               'longname': 'edit',
                                               'numArgs': 0,
                                               'shortname': 'e'},
                                      'exists': {'args': '<type __builtin__.bool>',
                                                 'longname': 'exists',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ex'},
                                      'history': {'args': '<type __builtin__.bool>',
                                                  'longname': 'history',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                                      'image1': {'args': '<type __builtin__.str>',
                                                 'longname': 'image1',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'i1'},
                                      'image2': {'args': '<type __builtin__.str>',
                                                 'longname': 'image2',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'i2'},
                                      'image3': {'args': '<type __builtin__.str>',
                                                 'longname': 'image3',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'i3'},
                                      'name': {'args': '<type __builtin__.str>',
                                               'longname': 'name',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'n'},
                                      'offCommand': {'args': '<type __builtin__.str>',
                                                     'longname': 'offCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ofc'},
                                      'onCommand': {'args': '<type __builtin__.str>',
                                                    'longname': 'onCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'onc'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'shadingCentric': {'args': '<type __builtin__.bool>',
                                                         'longname': 'shadingCentric',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 's'}},
                            'shortFlags': {'ch': 'history',
                                           'e': 'edit',
                                           'ex': 'exists',
                                           'i1': 'image1',
                                           'i2': 'image2',
                                           'i3': 'image3',
                                           'n': 'name',
                                           'ofc': 'offCommand',
                                           'onc': 'onCommand',
                                           'q': 'query',
                                           's': 'shadingCentric'},
                            'type': 'context'},
  'shadingLightRelCtx': {'callbackFlags': ['onCommand',
                                           'onc',
                                           'offCommand',
                                           'ofc'],
                         'flags': {'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'history': {'args': '<type __builtin__.bool>',
                                               'longname': 'history',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'offCommand': {'args': '<type __builtin__.str>',
                                                  'longname': 'offCommand',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ofc'},
                                   'onCommand': {'args': '<type __builtin__.str>',
                                                 'longname': 'onCommand',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'onc'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'shadingCentric': {'args': '<type __builtin__.bool>',
                                                      'longname': 'shadingCentric',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 's'}},
                         'shortFlags': {'ch': 'history',
                                        'e': 'edit',
                                        'ex': 'exists',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'n': 'name',
                                        'ofc': 'offCommand',
                                        'onc': 'onCommand',
                                        'q': 'query',
                                        's': 'shadingCentric'},
                         'type': 'context'},
  'shadingNetworkCompare': {'flags': {'byName': {'args': '<type __builtin__.bool>',
                                                 'longname': 'byName',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'nam'},
                                      'byValue': {'args': '<type __builtin__.bool>',
                                                  'longname': 'byValue',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'val'},
                                      'delete': {'args': '<type __builtin__.bool>',
                                                 'longname': 'delete',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'delete'},
                                      'equivalent': {'args': '<type __builtin__.bool>',
                                                     'longname': 'equivalent',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'eq'},
                                      'network1': {'args': '<type __builtin__.bool>',
                                                   'longname': 'network1',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'n1'},
                                      'network2': {'args': '<type __builtin__.bool>',
                                                   'longname': 'network2',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'n2'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'upstreamOnly': {'args': '<type __builtin__.bool>',
                                                       'longname': 'upstreamOnly',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'up'}},
                            'removedFlags': {'del': 'delete'},
                            'shortFlags': {'delete': 'delete',
                                           'eq': 'equivalent',
                                           'n1': 'network1',
                                           'n2': 'network2',
                                           'nam': 'byName',
                                           'q': 'query',
                                           'up': 'upstreamOnly',
                                           'val': 'byValue'},
                            'type': 'rendering'},
  'shadingNode': {'flags': {'asLight': {'args': '<type __builtin__.bool>',
                                        'longname': 'asLight',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'al'},
                            'asPostProcess': {'args': '<type __builtin__.bool>',
                                              'longname': 'asPostProcess',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'app'},
                            'asRendering': {'args': '<type __builtin__.bool>',
                                            'longname': 'asRendering',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ar'},
                            'asShader': {'args': '<type __builtin__.bool>',
                                         'longname': 'asShader',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'asShader'},
                            'asTexture': {'args': '<type __builtin__.bool>',
                                          'longname': 'asTexture',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'at'},
                            'asUtility': {'args': '<type __builtin__.bool>',
                                          'longname': 'asUtility',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'au'},
                            'isColorManaged': {'args': '<type __builtin__.bool>',
                                               'longname': 'isColorManaged',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'icm'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'shared': {'args': '<type __builtin__.bool>',
                                       'longname': 'shared',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 's'},
                            'skipSelect': {'args': '<type __builtin__.bool>',
                                           'longname': 'skipSelect',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ss'}},
                  'removedFlags': {'as': 'asShader'},
                  'shortFlags': {'al': 'asLight',
                                 'app': 'asPostProcess',
                                 'ar': 'asRendering',
                                 'asShader': 'asShader',
                                 'at': 'asTexture',
                                 'au': 'asUtility',
                                 'icm': 'isColorManaged',
                                 'n': 'name',
                                 'p': 'parent',
                                 's': 'shared',
                                 'ss': 'skipSelect'},
                  'type': 'rendering'},
  'shapeCompare': {'flags': {}, 'shortFlags': {}, 'type': 'general'},
  'shapeEditor': {'flags': {'clearSelection': {'args': '<type __builtin__.bool>',
                                               'longname': 'clearSelection',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'cs'},
                            'control': {'args': '<type __builtin__.bool>',
                                        'longname': 'control',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'ctl'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'filter': {'args': '<type __builtin__.str>',
                                       'longname': 'filter',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'f'},
                            'forceMainConnection': {'args': '<type __builtin__.str>',
                                                    'longname': 'forceMainConnection',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'fmc'},
                            'highlightConnection': {'args': '<type __builtin__.str>',
                                                    'longname': 'highlightConnection',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'hlc'},
                            'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                   'longname': 'lockMainConnection',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'lck'},
                            'lowestSelection': {'args': '<type __builtin__.bool>',
                                                'longname': 'lowestSelection',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'ls'},
                            'mainListConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'mainListConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'mlc'},
                            'panel': {'args': '<type __builtin__.str>',
                                      'longname': 'panel',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'pnl'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'selectionConnection': {'args': '<type __builtin__.str>',
                                                    'longname': 'selectionConnection',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'slc'},
                            'stateString': {'args': '<type __builtin__.bool>',
                                            'longname': 'stateString',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'sts'},
                            'targetControlList': {'args': '<type __builtin__.bool>',
                                                  'longname': 'targetControlList',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'tcl'},
                            'targetList': {'args': '<type __builtin__.bool>',
                                           'longname': 'targetList',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'tl'},
                            'unParent': {'args': '<type __builtin__.bool>',
                                         'longname': 'unParent',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'up'},
                            'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                     'longname': 'unlockMainConnection',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'ulk'},
                            'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                     'longname': 'updateMainConnection',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'upd'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'verticalSliders': {'args': '<type __builtin__.bool>',
                                                'longname': 'verticalSliders',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'vs'}},
                  'shortFlags': {'cs': 'clearSelection',
                                 'ctl': 'control',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ex': 'exists',
                                 'f': 'filter',
                                 'fmc': 'forceMainConnection',
                                 'hlc': 'highlightConnection',
                                 'lck': 'lockMainConnection',
                                 'ls': 'lowestSelection',
                                 'mlc': 'mainListConnection',
                                 'p': 'parent',
                                 'pnl': 'panel',
                                 'q': 'query',
                                 'slc': 'selectionConnection',
                                 'sts': 'stateString',
                                 'tcl': 'targetControlList',
                                 'tl': 'targetList',
                                 'ulk': 'unlockMainConnection',
                                 'up': 'unParent',
                                 'upd': 'updateMainConnection',
                                 'ut': 'useTemplate'},
                  'type': 'animation'},
  'shapePanel': {'callbackFlags': ['popupMenuProcedure', 'pmp'],
                 'flags': {'control': {'args': '<type __builtin__.bool>',
                                       'longname': 'control',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'ctl'},
                           'copy': {'args': '<type __builtin__.str>',
                                    'longname': 'copy',
                                    'modes': ['edit'],
                                    'numArgs': 1,
                                    'shortname': 'cp'},
                           'createString': {'args': '<type __builtin__.bool>',
                                            'longname': 'createString',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'cs'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'editString': {'args': '<type __builtin__.bool>',
                                          'longname': 'editString',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'es'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'init': {'args': '<type __builtin__.bool>',
                                    'longname': 'init',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 0,
                                    'shortname': 'init'},
                           'isUnique': {'args': '<type __builtin__.bool>',
                                        'longname': 'isUnique',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'iu'},
                           'label': {'args': '<type __builtin__.str>',
                                     'longname': 'label',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'l'},
                           'menuBarRepeatLast': {'args': '<type __builtin__.bool>',
                                                 'longname': 'menuBarRepeatLast',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'mrl'},
                           'menuBarVisible': {'args': '<type __builtin__.bool>',
                                              'longname': 'menuBarVisible',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'mbv'},
                           'needsInit': {'args': '<type __builtin__.bool>',
                                         'longname': 'needsInit',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'ni'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuProcedure': {'args': 'script',
                                                  'longname': 'popupMenuProcedure',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'pmp'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'replacePanel': {'args': '<type __builtin__.str>',
                                            'longname': 'replacePanel',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'rp'},
                           'shapeEditor': {'args': '<type __builtin__.bool>',
                                           'longname': 'shapeEditor',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'se'},
                           'tearOff': {'args': '<type __builtin__.bool>',
                                       'longname': 'tearOff',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'to'},
                           'tearOffCopy': {'args': '<type __builtin__.str>',
                                           'longname': 'tearOffCopy',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'toc'},
                           'tearOffRestore': {'args': '<type __builtin__.bool>',
                                              'longname': 'tearOffRestore',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'tor'},
                           'unParent': {'args': '<type __builtin__.bool>',
                                        'longname': 'unParent',
                                        'modes': ['edit'],
                                        'numArgs': 0,
                                        'shortname': 'up'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'}},
                 'removedFlags': {'in': 'init'},
                 'shortFlags': {'cp': 'copy',
                                'cs': 'createString',
                                'ctl': 'control',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'es': 'editString',
                                'ex': 'exists',
                                'init': 'init',
                                'iu': 'isUnique',
                                'l': 'label',
                                'mbv': 'menuBarVisible',
                                'mrl': 'menuBarRepeatLast',
                                'ni': 'needsInit',
                                'p': 'parent',
                                'pmp': 'popupMenuProcedure',
                                'q': 'query',
                                'rp': 'replacePanel',
                                'se': 'shapeEditor',
                                'to': 'tearOff',
                                'toc': 'tearOffCopy',
                                'tor': 'tearOffRestore',
                                'up': 'unParent',
                                'ut': 'useTemplate'},
                 'type': 'animation'},
  'shelfButton': {'callbackFlags': ['dragCallback',
                                    'dgc',
                                    'doubleClickCommand',
                                    'dcc',
                                    'visibleChangeCommand',
                                    'vcc',
                                    'commandRepeatable',
                                    'rpt',
                                    'labelEditingCallback',
                                    'lec',
                                    'handleNodeDropCallback',
                                    'hnd',
                                    'enableCommandRepeat',
                                    'ecr',
                                    'command',
                                    'c',
                                    'dropCallback',
                                    'dpc'],
                  'flags': {'actionIsSubstitute': {'args': '<type __builtin__.bool>',
                                                   'longname': 'actionIsSubstitute',
                                                   'numArgs': 1,
                                                   'shortname': 'ais'},
                            'align': {'args': '<type __builtin__.str>',
                                      'longname': 'align',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'al'},
                            'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'command': {'args': 'script',
                                        'longname': 'command',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'c'},
                            'commandRepeatable': {'args': '<type __builtin__.bool>',
                                                  'longname': 'commandRepeatable',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'rpt'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'disabledImage': {'args': '<type __builtin__.str>',
                                              'longname': 'disabledImage',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'di'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'doubleClickCommand': {'args': 'script',
                                                   'longname': 'doubleClickCommand',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dcc'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableCommandRepeat': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableCommandRepeat',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ecr'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'flat': {'args': '<type __builtin__.bool>',
                                     'longname': 'flat',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'fla'},
                            'flexibleWidthType': {'args': '<type __builtin__.int>',
                                                  'longname': 'flexibleWidthType',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'fwt'},
                            'flexibleWidthValue': {'args': '<type __builtin__.int>',
                                                   'longname': 'flexibleWidthValue',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'fwv'},
                            'flipX': {'args': '<type __builtin__.bool>',
                                      'longname': 'flipX',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'fx'},
                            'flipY': {'args': '<type __builtin__.bool>',
                                      'longname': 'flipY',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'fy'},
                            'font': {'args': '<type __builtin__.str>',
                                     'longname': 'font',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'fn'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'handleNodeDropCallback': {'args': 'script',
                                                       'longname': 'handleNodeDropCallback',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'hnd'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'highlightImage': {'args': '<type __builtin__.str>',
                                               'longname': 'highlightImage',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'hi'},
                            'image': {'args': '<type __builtin__.str>',
                                      'longname': 'image',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i'},
                            'image1': {'args': '<type __builtin__.str>',
                                       'longname': 'image1',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i1'},
                            'image2': {'args': '<type __builtin__.str>',
                                       'longname': 'image2',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i2'},
                            'image3': {'args': '<type __builtin__.str>',
                                       'longname': 'image3',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i3'},
                            'imageOverlayLabel': {'args': '<type __builtin__.str>',
                                                  'longname': 'imageOverlayLabel',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'iol'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'label': {'args': '<type __builtin__.str>',
                                      'longname': 'label',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'l'},
                            'labelEditingCallback': {'args': 'script',
                                                     'longname': 'labelEditingCallback',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'lec'},
                            'labelOffset': {'args': '<type __builtin__.int>',
                                            'longname': 'labelOffset',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'lo'},
                            'ltVersion': {'args': '<type __builtin__.str>',
                                          'longname': 'ltVersion',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'lt'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'marginHeight': {'args': '<type __builtin__.int>',
                                             'longname': 'marginHeight',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'mh'},
                            'marginWidth': {'args': '<type __builtin__.int>',
                                            'longname': 'marginWidth',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'mw'},
                            'menuItem': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'menuItem',
                                         'modes': ['create',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 2,
                                         'shortname': 'mi'},
                            'menuItemPython': {'args': '<type __builtin__.int>',
                                               'longname': 'menuItemPython',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'mip'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'noDefaultPopup': {'args': '<type __builtin__.bool>',
                                               'longname': 'noDefaultPopup',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ndp'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'overlayLabelBackColor': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'overlayLabelBackColor',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 4,
                                                      'shortname': 'olb'},
                            'overlayLabelColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'overlayLabelColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'olc'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'rotation': {'args': '<type __builtin__.float>',
                                         'longname': 'rotation',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'rot'},
                            'scaleIcon': {'args': '<type __builtin__.bool>',
                                          'longname': 'scaleIcon',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'sic'},
                            'selectionImage': {'args': '<type __builtin__.str>',
                                               'longname': 'selectionImage',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'si'},
                            'sourceType': {'args': '<type __builtin__.str>',
                                           'longname': 'sourceType',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'stp'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'style': {'args': '<type __builtin__.str>',
                                      'longname': 'style',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'st'},
                            'useAlpha': {'args': '<type __builtin__.bool>',
                                         'longname': 'useAlpha',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ua'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'version': {'args': '<type __builtin__.str>',
                                        'longname': 'version',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ver'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'ais': 'actionIsSubstitute',
                                 'al': 'align',
                                 'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'c': 'command',
                                 'dcc': 'doubleClickCommand',
                                 'dgc': 'dragCallback',
                                 'di': 'disabledImage',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ecr': 'enableCommandRepeat',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'ex': 'exists',
                                 'fla': 'flat',
                                 'fn': 'font',
                                 'fpn': 'fullPathName',
                                 'fwt': 'flexibleWidthType',
                                 'fwv': 'flexibleWidthValue',
                                 'fx': 'flipX',
                                 'fy': 'flipY',
                                 'h': 'height',
                                 'hi': 'highlightImage',
                                 'hlc': 'highlightColor',
                                 'hnd': 'handleNodeDropCallback',
                                 'i': 'image',
                                 'i1': 'image1',
                                 'i2': 'image2',
                                 'i3': 'image3',
                                 'io': 'isObscured',
                                 'iol': 'imageOverlayLabel',
                                 'l': 'label',
                                 'lec': 'labelEditingCallback',
                                 'lo': 'labelOffset',
                                 'lt': 'ltVersion',
                                 'm': 'manage',
                                 'mh': 'marginHeight',
                                 'mi': 'menuItem',
                                 'mip': 'menuItemPython',
                                 'mw': 'marginWidth',
                                 'nbg': 'noBackground',
                                 'ndp': 'noDefaultPopup',
                                 'npm': 'numberOfPopupMenus',
                                 'olb': 'overlayLabelBackColor',
                                 'olc': 'overlayLabelColor',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'rot': 'rotation',
                                 'rpt': 'commandRepeatable',
                                 'sbm': 'statusBarMessage',
                                 'si': 'selectionImage',
                                 'sic': 'scaleIcon',
                                 'st': 'style',
                                 'stp': 'sourceType',
                                 'ua': 'useAlpha',
                                 'ut': 'useTemplate',
                                 'vcc': 'visibleChangeCommand',
                                 'ver': 'version',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'shelfLayout': {'callbackFlags': ['visibleChangeCommand',
                                    'vcc',
                                    'dragCallback',
                                    'dgc',
                                    'dropCallback',
                                    'dpc'],
                  'flags': {'alignment': {'args': '<type __builtin__.str>',
                                          'longname': 'alignment',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'aln'},
                            'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'cellHeight': {'args': '<type __builtin__.int>',
                                           'longname': 'cellHeight',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ch'},
                            'cellWidth': {'args': '<type __builtin__.int>',
                                          'longname': 'cellWidth',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cw'},
                            'cellWidthHeight': {'args': ['<type __builtin__.int>',
                                                         '<type __builtin__.int>'],
                                                'longname': 'cellWidthHeight',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 2,
                                                'shortname': 'cwh'},
                            'childArray': {'args': '<type __builtin__.bool>',
                                           'longname': 'childArray',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'ca'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'horizontal': {'args': '<type __builtin__.bool>',
                                           'longname': 'horizontal',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'hr'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'ltVersion': {'args': '<type __builtin__.str>',
                                          'longname': 'ltVersion',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'lt'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfChildren': {'args': '<type __builtin__.bool>',
                                                 'longname': 'numberOfChildren',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'nch'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'position': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.int>'],
                                         'longname': 'position',
                                         'modes': ['create',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 2,
                                         'shortname': 'pos'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'spacing': {'args': '<type __builtin__.int>',
                                        'longname': 'spacing',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'spa'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'style': {'args': '<type __builtin__.str>',
                                      'longname': 'style',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'st'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'version': {'args': '<type __builtin__.str>',
                                        'longname': 'version',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ver'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'aln': 'alignment',
                                 'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'ca': 'childArray',
                                 'ch': 'cellHeight',
                                 'cw': 'cellWidth',
                                 'cwh': 'cellWidthHeight',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'ex': 'exists',
                                 'fpn': 'fullPathName',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'hr': 'horizontal',
                                 'io': 'isObscured',
                                 'lt': 'ltVersion',
                                 'm': 'manage',
                                 'nbg': 'noBackground',
                                 'nch': 'numberOfChildren',
                                 'npm': 'numberOfPopupMenus',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'pos': 'position',
                                 'q': 'query',
                                 'sbm': 'statusBarMessage',
                                 'spa': 'spacing',
                                 'st': 'style',
                                 'ut': 'useTemplate',
                                 'vcc': 'visibleChangeCommand',
                                 'ver': 'version',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'shelfTabLayout': {'callbackFlags': ['selectCommand',
                                       'sc',
                                       'doubleClickCommand',
                                       'dcc',
                                       'visibleChangeCommand',
                                       'vcc',
                                       'preSelectCommand',
                                       'psc',
                                       'dragCallback',
                                       'dgc',
                                       'newTabCommand',
                                       'ntc',
                                       'changeCommand',
                                       'cc',
                                       'dropCallback',
                                       'dpc'],
                     'flags': {'annotation': {'args': '<type __builtin__.str>',
                                              'longname': 'annotation',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ann'},
                               'backgroundColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'backgroundColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'bgc'},
                               'borderStyle': {'args': '<type __builtin__.str>',
                                               'longname': 'borderStyle',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'bs'},
                               'changeCommand': {'args': 'script',
                                                 'longname': 'changeCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cc'},
                               'childArray': {'args': '<type __builtin__.bool>',
                                              'longname': 'childArray',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'ca'},
                               'childResizable': {'args': '<type __builtin__.bool>',
                                                  'longname': 'childResizable',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'cr'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'doubleClickCommand': {'args': 'script',
                                                      'longname': 'doubleClickCommand',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dcc'},
                               'dragCallback': {'args': 'script',
                                                'longname': 'dragCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dgc'},
                               'dropCallback': {'args': 'script',
                                                'longname': 'dropCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dpc'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'enable': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en'},
                               'enableBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableBackground',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ebg'},
                               'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                       'longname': 'enableKeyboardFocus',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ekf'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'fullPathName': {'args': '<type __builtin__.bool>',
                                                'longname': 'fullPathName',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'fpn'},
                               'height': {'args': '<type __builtin__.int>',
                                          'longname': 'height',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'h'},
                               'highlightColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'highlightColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'hlc'},
                               'horizontalScrollBarThickness': {'args': '<type __builtin__.int>',
                                                                'longname': 'horizontalScrollBarThickness',
                                                                'modes': ['create',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'hst'},
                               'image': {'args': '<type __builtin__.str>',
                                         'longname': 'image',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i'},
                               'imageVisible': {'args': '<type __builtin__.bool>',
                                                'longname': 'imageVisible',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'iv'},
                               'innerMarginHeight': {'args': '<type __builtin__.int>',
                                                     'longname': 'innerMarginHeight',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 1,
                                                     'shortname': 'imh'},
                               'innerMarginWidth': {'args': '<type __builtin__.int>',
                                                    'longname': 'innerMarginWidth',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 1,
                                                    'shortname': 'imw'},
                               'isObscured': {'args': '<type __builtin__.bool>',
                                              'longname': 'isObscured',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'io'},
                               'manage': {'args': '<type __builtin__.bool>',
                                          'longname': 'manage',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'm'},
                               'minChildWidth': {'args': '<type __builtin__.int>',
                                                 'longname': 'minChildWidth',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'mcw'},
                               'moveTab': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'moveTab',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 2,
                                           'shortname': 'mt'},
                               'newTabCommand': {'args': 'script',
                                                 'longname': 'newTabCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ntc'},
                               'noBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'noBackground',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nbg'},
                               'numberOfChildren': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfChildren',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'nch'},
                               'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'numberOfPopupMenus',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'npm'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                  'longname': 'popupMenuArray',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'pma'},
                               'preSelectCommand': {'args': 'script',
                                                    'longname': 'preSelectCommand',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'psc'},
                               'preventOverride': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preventOverride',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'po'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'scrollable': {'args': '<type __builtin__.bool>',
                                              'longname': 'scrollable',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'scr'},
                               'scrollableTabs': {'args': '<type __builtin__.bool>',
                                                  'longname': 'scrollableTabs',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'stb'},
                               'selectCommand': {'args': 'script',
                                                 'longname': 'selectCommand',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sc'},
                               'selectTab': {'args': '<type __builtin__.str>',
                                             'longname': 'selectTab',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'st'},
                               'selectTabIndex': {'args': '<type __builtin__.int>',
                                                  'longname': 'selectTabIndex',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sti'},
                               'showNewTab': {'args': '<type __builtin__.bool>',
                                              'longname': 'showNewTab',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'snt'},
                               'statusBarMessage': {'args': '<type __builtin__.str>',
                                                    'longname': 'statusBarMessage',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sbm'},
                               'tabLabel': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'tabLabel',
                                            'modes': ['create',
                                                      'query',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 2,
                                            'shortname': 'tl'},
                               'tabLabelIndex': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'tabLabelIndex',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit',
                                                           'multiuse'],
                                                 'numArgs': 2,
                                                 'shortname': 'tli'},
                               'tabPosition': {'args': '<type __builtin__.str>',
                                               'longname': 'tabPosition',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tp'},
                               'tabTooltip': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'tabTooltip',
                                              'modes': ['create',
                                                        'query',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'tt'},
                               'tabTooltipIndex': {'args': ['<type __builtin__.int>',
                                                            '<type __builtin__.str>'],
                                                   'longname': 'tabTooltipIndex',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 2,
                                                   'shortname': 'tti'},
                               'tabsClosable': {'args': '<type __builtin__.bool>',
                                                'longname': 'tabsClosable',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'tc'},
                               'tabsVisible': {'args': '<type __builtin__.bool>',
                                               'longname': 'tabsVisible',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tv'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'},
                               'verticalScrollBarThickness': {'args': '<type __builtin__.int>',
                                                              'longname': 'verticalScrollBarThickness',
                                                              'modes': ['create',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'vst'},
                               'visible': {'args': '<type __builtin__.bool>',
                                           'longname': 'visible',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vis'},
                               'visibleChangeCommand': {'args': 'script',
                                                        'longname': 'visibleChangeCommand',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'vcc'},
                               'width': {'args': '<type __builtin__.int>',
                                         'longname': 'width',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'w'}},
                     'shortFlags': {'ann': 'annotation',
                                    'bgc': 'backgroundColor',
                                    'bs': 'borderStyle',
                                    'ca': 'childArray',
                                    'cc': 'changeCommand',
                                    'cr': 'childResizable',
                                    'dcc': 'doubleClickCommand',
                                    'dgc': 'dragCallback',
                                    'dpc': 'dropCallback',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'e': 'edit',
                                    'ebg': 'enableBackground',
                                    'ekf': 'enableKeyboardFocus',
                                    'en': 'enable',
                                    'ex': 'exists',
                                    'fpn': 'fullPathName',
                                    'h': 'height',
                                    'hlc': 'highlightColor',
                                    'hst': 'horizontalScrollBarThickness',
                                    'i': 'image',
                                    'imh': 'innerMarginHeight',
                                    'imw': 'innerMarginWidth',
                                    'io': 'isObscured',
                                    'iv': 'imageVisible',
                                    'm': 'manage',
                                    'mcw': 'minChildWidth',
                                    'mt': 'moveTab',
                                    'nbg': 'noBackground',
                                    'nch': 'numberOfChildren',
                                    'npm': 'numberOfPopupMenus',
                                    'ntc': 'newTabCommand',
                                    'p': 'parent',
                                    'pma': 'popupMenuArray',
                                    'po': 'preventOverride',
                                    'psc': 'preSelectCommand',
                                    'q': 'query',
                                    'sbm': 'statusBarMessage',
                                    'sc': 'selectCommand',
                                    'scr': 'scrollable',
                                    'snt': 'showNewTab',
                                    'st': 'selectTab',
                                    'stb': 'scrollableTabs',
                                    'sti': 'selectTabIndex',
                                    'tc': 'tabsClosable',
                                    'tl': 'tabLabel',
                                    'tli': 'tabLabelIndex',
                                    'tp': 'tabPosition',
                                    'tt': 'tabTooltip',
                                    'tti': 'tabTooltipIndex',
                                    'tv': 'tabsVisible',
                                    'ut': 'useTemplate',
                                    'vcc': 'visibleChangeCommand',
                                    'vis': 'visible',
                                    'vst': 'verticalScrollBarThickness',
                                    'w': 'width'},
                     'type': 'windows'},
  'shot': {'flags': {'audio': {'args': '<type __builtin__.str>',
                               'longname': 'audio',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'aud'},
                     'clip': {'args': '<type __builtin__.str>',
                              'longname': 'clip',
                              'modes': ['create', 'query', 'edit'],
                              'numArgs': 1,
                              'shortname': 'cl'},
                     'clipDuration': {'args': 'time',
                                      'longname': 'clipDuration',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'cd'},
                     'clipOpacity': {'args': '<type __builtin__.float>',
                                     'longname': 'clipOpacity',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'co'},
                     'clipSyncState': {'args': '<type __builtin__.bool>',
                                       'longname': 'clipSyncState',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 0,
                                       'resultNeedsCasting': True,
                                       'shortname': 'css'},
                     'clipZeroOffset': {'args': 'time',
                                        'longname': 'clipZeroOffset',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'czo'},
                     'copy': {'args': '<type __builtin__.bool>',
                              'longname': 'copy',
                              'modes': ['create', 'query', 'edit'],
                              'numArgs': 0,
                              'shortname': 'c'},
                     'createCustomAnim': {'args': '<type __builtin__.bool>',
                                          'longname': 'createCustomAnim',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'cca'},
                     'currentCamera': {'args': '<type __builtin__.str>',
                                       'longname': 'currentCamera',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cc'},
                     'customAnim': {'args': '<type __builtin__.bool>',
                                    'longname': 'customAnim',
                                    'modes': ['query'],
                                    'numArgs': 0,
                                    'shortname': 'ca'},
                     'deleteCustomAnim': {'args': '<type __builtin__.bool>',
                                          'longname': 'deleteCustomAnim',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'dca'},
                     'determineTrack': {'args': '<type __builtin__.bool>',
                                        'longname': 'determineTrack',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'dt'},
                     'edit': {'args': '<type __builtin__.bool>',
                              'longname': 'edit',
                              'numArgs': 0,
                              'shortname': 'e'},
                     'endTime': {'args': 'time',
                                 'longname': 'endTime',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'et'},
                     'favorite': {'args': '<type __builtin__.bool>',
                                  'longname': 'favorite',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'fav'},
                     'flag1': {'args': '<type __builtin__.bool>',
                               'longname': 'flag1',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'f1'},
                     'flag10': {'args': '<type __builtin__.bool>',
                                'longname': 'flag10',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'f10'},
                     'flag11': {'args': '<type __builtin__.bool>',
                                'longname': 'flag11',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'f11'},
                     'flag12': {'args': '<type __builtin__.bool>',
                                'longname': 'flag12',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'f12'},
                     'flag2': {'args': '<type __builtin__.bool>',
                               'longname': 'flag2',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'f2'},
                     'flag3': {'args': '<type __builtin__.bool>',
                               'longname': 'flag3',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'f3'},
                     'flag4': {'args': '<type __builtin__.bool>',
                               'longname': 'flag4',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'f4'},
                     'flag5': {'args': '<type __builtin__.bool>',
                               'longname': 'flag5',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'f5'},
                     'flag6': {'args': '<type __builtin__.bool>',
                               'longname': 'flag6',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'f6'},
                     'flag7': {'args': '<type __builtin__.bool>',
                               'longname': 'flag7',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'f7'},
                     'flag8': {'args': '<type __builtin__.bool>',
                               'longname': 'flag8',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'f8'},
                     'flag9': {'args': '<type __builtin__.bool>',
                               'longname': 'flag9',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'f9'},
                     'hasCameraSet': {'args': '<type __builtin__.bool>',
                                      'longname': 'hasCameraSet',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'hcs'},
                     'hasStereoCamera': {'args': '<type __builtin__.bool>',
                                         'longname': 'hasStereoCamera',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'hsc'},
                     'imagePlaneVisibility': {'args': '<type __builtin__.bool>',
                                              'longname': 'imagePlaneVisibility',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ipv'},
                     'linkAudio': {'args': '<type __builtin__.str>',
                                   'longname': 'linkAudio',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'la'},
                     'lock': {'args': '<type __builtin__.bool>',
                              'longname': 'lock',
                              'modes': ['create', 'query', 'edit'],
                              'numArgs': 1,
                              'shortname': 'lck'},
                     'mute': {'args': '<type __builtin__.bool>',
                              'longname': 'mute',
                              'modes': ['create', 'query', 'edit'],
                              'numArgs': 1,
                              'shortname': 'm'},
                     'paste': {'args': '<type __builtin__.bool>',
                               'longname': 'paste',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 0,
                               'shortname': 'p'},
                     'pasteInstance': {'args': '<type __builtin__.bool>',
                                       'longname': 'pasteInstance',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'pi'},
                     'postHoldTime': {'args': 'time',
                                      'longname': 'postHoldTime',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'pst'},
                     'preHoldTime': {'args': 'time',
                                     'longname': 'preHoldTime',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'prt'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'scale': {'args': '<type __builtin__.float>',
                               'longname': 'scale',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 's'},
                     'selfmute': {'args': '<type __builtin__.bool>',
                                  'longname': 'selfmute',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'sm'},
                     'sequenceDuration': {'args': 'time',
                                          'longname': 'sequenceDuration',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sqd'},
                     'sequenceEndTime': {'args': 'time',
                                         'longname': 'sequenceEndTime',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'set'},
                     'sequenceStartTime': {'args': 'time',
                                           'longname': 'sequenceStartTime',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sst'},
                     'shotName': {'args': '<type __builtin__.str>',
                                  'longname': 'shotName',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'sn'},
                     'sourceDuration': {'args': 'time',
                                        'longname': 'sourceDuration',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'sd'},
                     'startTime': {'args': 'time',
                                   'longname': 'startTime',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'st'},
                     'track': {'args': '<type __builtin__.int>',
                               'longname': 'track',
                               'modes': ['query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'trk'},
                     'transitionInLength': {'args': 'time',
                                            'longname': 'transitionInLength',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'til'},
                     'transitionInType': {'args': '<type __builtin__.int>',
                                          'longname': 'transitionInType',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'tit'},
                     'transitionOutLength': {'args': 'time',
                                             'longname': 'transitionOutLength',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'tol'},
                     'transitionOutType': {'args': '<type __builtin__.int>',
                                           'longname': 'transitionOutType',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'tot'},
                     'unlinkAudio': {'args': '<type __builtin__.bool>',
                                     'longname': 'unlinkAudio',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'ula'}},
           'shortFlags': {'aud': 'audio',
                          'c': 'copy',
                          'ca': 'customAnim',
                          'cc': 'currentCamera',
                          'cca': 'createCustomAnim',
                          'cd': 'clipDuration',
                          'cl': 'clip',
                          'co': 'clipOpacity',
                          'css': 'clipSyncState',
                          'czo': 'clipZeroOffset',
                          'dca': 'deleteCustomAnim',
                          'dt': 'determineTrack',
                          'e': 'edit',
                          'et': 'endTime',
                          'f1': 'flag1',
                          'f10': 'flag10',
                          'f11': 'flag11',
                          'f12': 'flag12',
                          'f2': 'flag2',
                          'f3': 'flag3',
                          'f4': 'flag4',
                          'f5': 'flag5',
                          'f6': 'flag6',
                          'f7': 'flag7',
                          'f8': 'flag8',
                          'f9': 'flag9',
                          'fav': 'favorite',
                          'hcs': 'hasCameraSet',
                          'hsc': 'hasStereoCamera',
                          'ipv': 'imagePlaneVisibility',
                          'la': 'linkAudio',
                          'lck': 'lock',
                          'm': 'mute',
                          'p': 'paste',
                          'pi': 'pasteInstance',
                          'prt': 'preHoldTime',
                          'pst': 'postHoldTime',
                          'q': 'query',
                          's': 'scale',
                          'sd': 'sourceDuration',
                          'set': 'sequenceEndTime',
                          'sm': 'selfmute',
                          'sn': 'shotName',
                          'sqd': 'sequenceDuration',
                          'sst': 'sequenceStartTime',
                          'st': 'startTime',
                          'til': 'transitionInLength',
                          'tit': 'transitionInType',
                          'tol': 'transitionOutLength',
                          'tot': 'transitionOutType',
                          'trk': 'track',
                          'ula': 'unlinkAudio'},
           'type': 'animation'},
  'shotRipple': {'flags': {'deleted': {'args': '<type __builtin__.bool>',
                                       'longname': 'deleted',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'd'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'endDelta': {'args': 'time',
                                        'longname': 'endDelta',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ed'},
                           'endTime': {'args': 'time',
                                       'longname': 'endTime',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'et'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'startDelta': {'args': 'time',
                                          'longname': 'startDelta',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sd'},
                           'startTime': {'args': 'time',
                                         'longname': 'startTime',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'st'}},
                 'shortFlags': {'d': 'deleted',
                                'e': 'edit',
                                'ed': 'endDelta',
                                'et': 'endTime',
                                'q': 'query',
                                'sd': 'startDelta',
                                'st': 'startTime'},
                 'type': 'animation'},
  'shotTrack': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'insertTrack': {'args': '<type __builtin__.int>',
                                          'longname': 'insertTrack',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'it'},
                          'lock': {'args': '<type __builtin__.bool>',
                                   'longname': 'lock',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'l'},
                          'mute': {'args': '<type __builtin__.bool>',
                                   'longname': 'mute',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'm'},
                          'numTracks': {'args': '<type __builtin__.int>',
                                        'longname': 'numTracks',
                                        'modes': ['query'],
                                        'numArgs': 1,
                                        'shortname': 'nt'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'removeEmptyTracks': {'args': '<type __builtin__.bool>',
                                                'longname': 'removeEmptyTracks',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ret'},
                          'removeTrack': {'args': '<type __builtin__.int>',
                                          'longname': 'removeTrack',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'rt'},
                          'selfmute': {'args': '<type __builtin__.bool>',
                                       'longname': 'selfmute',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sm'},
                          'solo': {'args': '<type __builtin__.bool>',
                                   'longname': 'solo',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'so'},
                          'swapTracks': {'args': ['<type __builtin__.int>',
                                                  '<type __builtin__.int>'],
                                         'longname': 'swapTracks',
                                         'modes': ['create'],
                                         'numArgs': 2,
                                         'shortname': 'st'},
                          'title': {'args': '<type __builtin__.str>',
                                    'longname': 'title',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 't'},
                          'track': {'args': '<type __builtin__.int>',
                                    'longname': 'track',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'tr'},
                          'unsolo': {'args': '<type __builtin__.bool>',
                                     'longname': 'unsolo',
                                     'modes': ['query'],
                                     'numArgs': 1,
                                     'shortname': 'uso'}},
                'shortFlags': {'e': 'edit',
                               'it': 'insertTrack',
                               'l': 'lock',
                               'm': 'mute',
                               'nt': 'numTracks',
                               'q': 'query',
                               'ret': 'removeEmptyTracks',
                               'rt': 'removeTrack',
                               'sm': 'selfmute',
                               'so': 'solo',
                               'st': 'swapTracks',
                               't': 'title',
                               'tr': 'track',
                               'uso': 'unsolo'},
                'type': 'system'},
  'showHelp': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                      'longname': 'absolute',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'a'},
                         'docs': {'args': '<type __builtin__.bool>',
                                  'longname': 'docs',
                                  'modes': ['create', 'query'],
                                  'numArgs': 0,
                                  'shortname': 'd'},
                         'helpTable': {'args': '<type __builtin__.bool>',
                                       'longname': 'helpTable',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'ht'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'version': {'args': '<type __builtin__.bool>',
                                     'longname': 'version',
                                     'modes': ['query'],
                                     'numArgs': 0,
                                     'shortname': 'v'}},
               'shortFlags': {'a': 'absolute',
                              'd': 'docs',
                              'ht': 'helpTable',
                              'q': 'query',
                              'v': 'version'},
               'type': 'system'},
  'showHidden': {'flags': {'above': {'args': '<type __builtin__.bool>',
                                     'longname': 'above',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'a'},
                           'allObjects': {'args': '<type __builtin__.bool>',
                                          'longname': 'allObjects',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'all'},
                           'below': {'args': '<type __builtin__.bool>',
                                     'longname': 'below',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'b'},
                           'lastHidden': {'args': '<type __builtin__.bool>',
                                          'longname': 'lastHidden',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'lh'}},
                 'shortFlags': {'a': 'above',
                                'all': 'allObjects',
                                'b': 'below',
                                'lh': 'lastHidden'},
                 'type': 'general'},
  'showManipCtx': {'callbackFlags': ['toolFinish', 'tf', 'toolStart', 'ts'],
                   'flags': {'currentNodeName': {'args': '<type __builtin__.bool>',
                                                 'longname': 'currentNodeName',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'cnn'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'history': {'args': '<type __builtin__.bool>',
                                         'longname': 'history',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ch'},
                             'image1': {'args': '<type __builtin__.str>',
                                        'longname': 'image1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i1'},
                             'image2': {'args': '<type __builtin__.str>',
                                        'longname': 'image2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i2'},
                             'image3': {'args': '<type __builtin__.str>',
                                        'longname': 'image3',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'i3'},
                             'incSnap': {'args': ['<type __builtin__.int>',
                                                  '<type __builtin__.bool>'],
                                         'longname': 'incSnap',
                                         'modes': ['create',
                                                   'query',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 2,
                                         'shortname': 'incSnap'},
                             'incSnapRelative': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.bool>'],
                                                 'longname': 'incSnapRelative',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit',
                                                           'multiuse'],
                                                 'numArgs': 2,
                                                 'shortname': 'isr'},
                             'incSnapUI': {'args': '<type __builtin__.bool>',
                                           'longname': 'incSnapUI',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'isu'},
                             'incSnapValue': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'incSnapValue',
                                              'modes': ['create',
                                                        'query',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'isv'},
                             'lockSelection': {'args': '<type __builtin__.bool>',
                                               'longname': 'lockSelection',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ls'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'toggleIncSnap': {'args': '<type __builtin__.bool>',
                                               'longname': 'toggleIncSnap',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'tis'},
                             'toolFinish': {'args': 'script',
                                            'longname': 'toolFinish',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'tf'},
                             'toolStart': {'args': 'script',
                                           'longname': 'toolStart',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ts'}},
                   'removedFlags': {'is': 'incSnap'},
                   'shortFlags': {'ch': 'history',
                                  'cnn': 'currentNodeName',
                                  'e': 'edit',
                                  'ex': 'exists',
                                  'i1': 'image1',
                                  'i2': 'image2',
                                  'i3': 'image3',
                                  'incSnap': 'incSnap',
                                  'isr': 'incSnapRelative',
                                  'isu': 'incSnapUI',
                                  'isv': 'incSnapValue',
                                  'ls': 'lockSelection',
                                  'n': 'name',
                                  'q': 'query',
                                  'tf': 'toolFinish',
                                  'tis': 'toggleIncSnap',
                                  'ts': 'toolStart'},
                   'type': 'context'},
  'showMetadata': {'flags': {'auto': {'args': '<type __builtin__.bool>',
                                      'longname': 'auto',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'a'},
                             'dataType': {'args': '<type __builtin__.str>',
                                          'longname': 'dataType',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'dt'},
                             'interpolation': {'args': '<type __builtin__.bool>',
                                               'longname': 'interpolation',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'i'},
                             'isActivated': {'args': '<type __builtin__.bool>',
                                             'longname': 'isActivated',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'ia'},
                             'listAllStreams': {'args': '<type __builtin__.bool>',
                                                'longname': 'listAllStreams',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 0,
                                                'shortname': 'las'},
                             'listMembers': {'args': '<type __builtin__.bool>',
                                             'longname': 'listMembers',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'lm'},
                             'listValidMethods': {'args': '<type __builtin__.bool>',
                                                  'longname': 'listValidMethods',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 0,
                                                  'shortname': 'lvm'},
                             'listVisibleStreams': {'args': '<type __builtin__.bool>',
                                                    'longname': 'listVisibleStreams',
                                                    'modes': ['create',
                                                              'query'],
                                                    'numArgs': 0,
                                                    'shortname': 'lvs'},
                             'member': {'args': '<type __builtin__.str>',
                                        'longname': 'member',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'mb'},
                             'method': {'args': '<type __builtin__.str>',
                                        'longname': 'method',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'off': {'args': ['In create mode',
                                              'turn off the member which is specified\nby the flags'],
                                     'longname': 'off',
                                     'modes': ['create', 'query'],
                                     'numArgs': 2,
                                     'shortname': 'boolean'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'range': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'range',
                                       'modes': ['create', 'query'],
                                       'numArgs': 2,
                                       'shortname': 'r'},
                             'rayScale': {'args': '<type __builtin__.float>',
                                          'longname': 'rayScale',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'rs'},
                             'stream': {'args': '<type __builtin__.str>',
                                        'longname': 'stream',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 's'}},
                   'shortFlags': {'a': 'auto',
                                  'dt': 'dataType',
                                  'i': 'interpolation',
                                  'ia': 'isActivated',
                                  'las': 'listAllStreams',
                                  'lm': 'listMembers',
                                  'lvm': 'listValidMethods',
                                  'lvs': 'listVisibleStreams',
                                  'm': 'method',
                                  'mb': 'member',
                                  'off': 'off',
                                  'q': 'query',
                                  'r': 'range',
                                  'rs': 'rayScale',
                                  's': 'stream'},
                   'type': 'modeling'},
  'showSelectionInTitle': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'showShadingGroupAttrEditor': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                                     'longname': 'query',
                                                     'numArgs': 0,
                                                     'shortname': 'q'}},
                                 'shortFlags': {'q': 'query'},
                                 'type': 'rendering'},
  'showWindow': {'flags': {}, 'shortFlags': {}, 'type': 'windows'},
  'simplify': {'flags': {'animation': {'args': '<type __builtin__.str>',
                                       'longname': 'animation',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'an'},
                         'attribute': {'args': '<type __builtin__.str>',
                                       'longname': 'attribute',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'at'},
                         'controlPoints': {'args': '<type __builtin__.bool>',
                                           'longname': 'controlPoints',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'cp'},
                         'float': {'args': 'floatrange',
                                   'longname': 'float',
                                   'modes': ['create', 'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 'f'},
                         'floatTolerance': {'args': '<type __builtin__.float>',
                                            'longname': 'floatTolerance',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ft'},
                         'hierarchy': {'args': '<type __builtin__.str>',
                                       'longname': 'hierarchy',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'hi'},
                         'includeUpperBound': {'args': '<type __builtin__.bool>',
                                               'longname': 'includeUpperBound',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'iub'},
                         'index': {'args': '<type __builtin__.int>',
                                   'longname': 'index',
                                   'modes': ['create', 'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 'index'},
                         'shape': {'args': '<type __builtin__.bool>',
                                   'longname': 'shape',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 's'},
                         'time': {'args': 'timerange',
                                  'longname': 'time',
                                  'modes': ['create', 'multiuse'],
                                  'numArgs': 1,
                                  'shortname': 't'},
                         'timeTolerance': {'args': 'time',
                                           'longname': 'timeTolerance',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'tt'},
                         'valueTolerance': {'args': '<type __builtin__.float>',
                                            'longname': 'valueTolerance',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'vt'}},
               'removedFlags': {'in': 'index'},
               'shortFlags': {'an': 'animation',
                              'at': 'attribute',
                              'cp': 'controlPoints',
                              'f': 'float',
                              'ft': 'floatTolerance',
                              'hi': 'hierarchy',
                              'index': 'index',
                              'iub': 'includeUpperBound',
                              's': 'shape',
                              't': 'time',
                              'tt': 'timeTolerance',
                              'vt': 'valueTolerance'},
               'type': 'animation'},
  'singleProfileBirailSurface': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                                       'longname': 'caching',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'cch'},
                                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                                   'longname': 'constructionHistory',
                                                                   'modes': ['create'],
                                                                   'numArgs': 1,
                                                                   'shortname': 'ch'},
                                           'edit': {'args': '<type __builtin__.bool>',
                                                    'longname': 'edit',
                                                    'numArgs': 0,
                                                    'shortname': 'e'},
                                           'frozen': {'args': '<type __builtin__.bool>',
                                                      'longname': 'frozen',
                                                      'numArgs': 1,
                                                      'shortname': 'fzn'},
                                           'name': {'args': '<type __builtin__.str>',
                                                    'longname': 'name',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'n'},
                                           'nodeState': {'args': '<type __builtin__.int>',
                                                         'longname': 'nodeState',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'nds'},
                                           'object': {'args': '<type __builtin__.bool>',
                                                      'longname': 'object',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'o'},
                                           'polygon': {'args': '<type __builtin__.int>',
                                                       'longname': 'polygon',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'po'},
                                           'query': {'args': '<type __builtin__.bool>',
                                                     'longname': 'query',
                                                     'numArgs': 0,
                                                     'shortname': 'q'},
                                           'tangentContinuityProfile1': {'args': '<type __builtin__.bool>',
                                                                         'longname': 'tangentContinuityProfile1',
                                                                         'modes': ['create',
                                                                                   'query',
                                                                                   'edit'],
                                                                         'numArgs': 1,
                                                                         'shortname': 'tp1'},
                                           'transformMode': {'args': '<type __builtin__.int>',
                                                             'longname': 'transformMode',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'tm'}},
                                 'shortFlags': {'cch': 'caching',
                                                'ch': 'constructionHistory',
                                                'e': 'edit',
                                                'fzn': 'frozen',
                                                'n': 'name',
                                                'nds': 'nodeState',
                                                'o': 'object',
                                                'po': 'polygon',
                                                'q': 'query',
                                                'tm': 'transformMode',
                                                'tp1': 'tangentContinuityProfile1'},
                                 'type': 'modeling'},
  'skeletonEmbed': {'flags': {'mergedMesh': {'args': '<type __builtin__.bool>',
                                             'longname': 'mergedMesh',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'mm'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'segmentationMethod': {'args': '<type __builtin__.int>',
                                                     'longname': 'segmentationMethod',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'sm'},
                              'segmentationResolution': {'args': '<type __builtin__.int>',
                                                         'longname': 'segmentationResolution',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'sr'}},
                    'shortFlags': {'mm': 'mergedMesh',
                                   'q': 'query',
                                   'sm': 'segmentationMethod',
                                   'sr': 'segmentationResolution'},
                    'type': 'animation'},
  'skinBindCtx': {'flags': {'about': {'args': '<type __builtin__.str>',
                                      'longname': 'about',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'a'},
                            'axis': {'args': '<type __builtin__.str>',
                                     'longname': 'axis',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'ax'},
                            'colorRamp': {'args': '<type __builtin__.str>',
                                          'longname': 'colorRamp',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cr'},
                            'currentInfluence': {'args': '<type __builtin__.str>',
                                                 'longname': 'currentInfluence',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ci'},
                            'displayInactiveMode': {'args': '<type __builtin__.int>',
                                                    'longname': 'displayInactiveMode',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'di'},
                            'displayNormalized': {'args': '<type __builtin__.bool>',
                                                  'longname': 'displayNormalized',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dn'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'falloffCurve': {'args': '<type __builtin__.str>',
                                             'longname': 'falloffCurve',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'fc'},
                            'history': {'args': '<type __builtin__.bool>',
                                        'longname': 'history',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ch'},
                            'image1': {'args': '<type __builtin__.str>',
                                       'longname': 'image1',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i1'},
                            'image2': {'args': '<type __builtin__.str>',
                                       'longname': 'image2',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i2'},
                            'image3': {'args': '<type __builtin__.str>',
                                       'longname': 'image3',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i3'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'symmetry': {'args': '<type __builtin__.bool>',
                                         'longname': 'symmetry',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 's'},
                            'tolerance': {'args': '<type __builtin__.float>',
                                          'longname': 'tolerance',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 't'}},
                  'shortFlags': {'a': 'about',
                                 'ax': 'axis',
                                 'ch': 'history',
                                 'ci': 'currentInfluence',
                                 'cr': 'colorRamp',
                                 'di': 'displayInactiveMode',
                                 'dn': 'displayNormalized',
                                 'e': 'edit',
                                 'ex': 'exists',
                                 'fc': 'falloffCurve',
                                 'i1': 'image1',
                                 'i2': 'image2',
                                 'i3': 'image3',
                                 'n': 'name',
                                 'q': 'query',
                                 's': 'symmetry',
                                 't': 'tolerance'},
                  'type': 'context'},
  'skinCluster': {'flags': {'addInfluence': {'args': '<type __builtin__.str>',
                                             'longname': 'addInfluence',
                                             'modes': ['edit', 'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'ai'},
                            'addToSelection': {'args': '<type __builtin__.bool>',
                                               'longname': 'addToSelection',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'ats'},
                            'after': {'args': '<type __builtin__.bool>',
                                      'longname': 'after',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'af'},
                            'afterReference': {'args': '<type __builtin__.bool>',
                                               'longname': 'afterReference',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'ar'},
                            'baseShape': {'args': '<type __builtin__.str>',
                                          'longname': 'baseShape',
                                          'modes': ['edit', 'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'bsh'},
                            'before': {'args': '<type __builtin__.bool>',
                                       'longname': 'before',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'bf'},
                            'bindMethod': {'args': '<type __builtin__.int>',
                                           'longname': 'bindMethod',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'bm'},
                            'deformerTools': {'args': '<type __builtin__.bool>',
                                              'longname': 'deformerTools',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'dt'},
                            'dropoffRate': {'args': '<type __builtin__.float>',
                                            'longname': 'dropoffRate',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dr'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'exclusive': {'args': '<type __builtin__.str>',
                                          'longname': 'exclusive',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'ex'},
                            'forceNormalizeWeights': {'args': '<type __builtin__.bool>',
                                                      'longname': 'forceNormalizeWeights',
                                                      'modes': ['edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'fnw'},
                            'frontOfChain': {'args': '<type __builtin__.bool>',
                                             'longname': 'frontOfChain',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'foc'},
                            'geometry': {'args': '<type __builtin__.str>',
                                         'longname': 'geometry',
                                         'modes': ['query',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'g'},
                            'geometryIndices': {'args': '<type __builtin__.bool>',
                                                'longname': 'geometryIndices',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'gi'},
                            'heatmapFalloff': {'args': '<type __builtin__.float>',
                                               'longname': 'heatmapFalloff',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'hmf'},
                            'ignoreBindPose': {'args': '<type __builtin__.bool>',
                                               'longname': 'ignoreBindPose',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'ibp'},
                            'ignoreHierarchy': {'args': '<type __builtin__.bool>',
                                                'longname': 'ignoreHierarchy',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 0,
                                                'shortname': 'ih'},
                            'ignoreSelected': {'args': '<type __builtin__.bool>',
                                               'longname': 'ignoreSelected',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ignoreSelected'},
                            'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                        'longname': 'includeHiddenSelections',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'ihs'},
                            'influence': {'args': '<type __builtin__.str>',
                                          'longname': 'influence',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'inf'},
                            'lockWeights': {'args': '<type __builtin__.bool>',
                                            'longname': 'lockWeights',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'lw'},
                            'maximumInfluences': {'args': '<type __builtin__.int>',
                                                  'longname': 'maximumInfluences',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mi'},
                            'moveJointsMode': {'args': '<type __builtin__.bool>',
                                               'longname': 'moveJointsMode',
                                               'modes': ['edit'],
                                               'numArgs': 1,
                                               'shortname': 'mjm'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'normalizeWeights': {'args': '<type __builtin__.int>',
                                                 'longname': 'normalizeWeights',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'nw'},
                            'nurbsSamples': {'args': '<type __builtin__.int>',
                                             'longname': 'nurbsSamples',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ns'},
                            'obeyMaxInfluences': {'args': '<type __builtin__.bool>',
                                                  'longname': 'obeyMaxInfluences',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'omi'},
                            'parallel': {'args': '<type __builtin__.bool>',
                                         'longname': 'parallel',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'par'},
                            'polySmoothness': {'args': '<type __builtin__.float>',
                                               'longname': 'polySmoothness',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ps'},
                            'prune': {'args': '<type __builtin__.bool>',
                                      'longname': 'prune',
                                      'modes': ['edit'],
                                      'numArgs': 0,
                                      'shortname': 'pr'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'recacheBindMatrices': {'args': '<type __builtin__.bool>',
                                                    'longname': 'recacheBindMatrices',
                                                    'modes': ['edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'rbm'},
                            'remove': {'args': '<type __builtin__.bool>',
                                       'longname': 'remove',
                                       'modes': ['edit', 'multiuse'],
                                       'numArgs': 0,
                                       'shortname': 'rm'},
                            'removeFromSelection': {'args': '<type __builtin__.bool>',
                                                    'longname': 'removeFromSelection',
                                                    'modes': ['edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'rfs'},
                            'removeInfluence': {'args': '<type __builtin__.str>',
                                                'longname': 'removeInfluence',
                                                'modes': ['edit',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'ri'},
                            'removeUnusedInfluence': {'args': '<type __builtin__.bool>',
                                                      'longname': 'removeUnusedInfluence',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'rui'},
                            'selectInfluenceVerts': {'args': '<type __builtin__.str>',
                                                     'longname': 'selectInfluenceVerts',
                                                     'modes': ['edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'siv'},
                            'skinMethod': {'args': '<type __builtin__.int>',
                                           'longname': 'skinMethod',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sm'},
                            'smoothWeights': {'args': '<type __builtin__.float>',
                                              'longname': 'smoothWeights',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'sw'},
                            'smoothWeightsMaxIterations': {'args': '<type __builtin__.int>',
                                                           'longname': 'smoothWeightsMaxIterations',
                                                           'modes': ['edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'swi'},
                            'split': {'args': '<type __builtin__.bool>',
                                      'longname': 'split',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'sp'},
                            'toSelectedBones': {'args': '<type __builtin__.bool>',
                                                'longname': 'toSelectedBones',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'tsb'},
                            'toSkeletonAndTransforms': {'args': '<type __builtin__.bool>',
                                                        'longname': 'toSkeletonAndTransforms',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'tst'},
                            'unbind': {'args': '<type __builtin__.bool>',
                                       'longname': 'unbind',
                                       'modes': ['edit'],
                                       'numArgs': 0,
                                       'shortname': 'ub'},
                            'unbindKeepHistory': {'args': '<type __builtin__.bool>',
                                                  'longname': 'unbindKeepHistory',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'ubk'},
                            'useGeometry': {'args': '<type __builtin__.bool>',
                                            'longname': 'useGeometry',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'ug'},
                            'volumeBind': {'args': '<type __builtin__.float>',
                                           'longname': 'volumeBind',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'vb'},
                            'volumeType': {'args': '<type __builtin__.int>',
                                           'longname': 'volumeType',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'vt'},
                            'weight': {'args': '<type __builtin__.float>',
                                       'longname': 'weight',
                                       'modes': ['edit'],
                                       'numArgs': 1,
                                       'shortname': 'wt'},
                            'weightDistribution': {'args': '<type __builtin__.int>',
                                                   'longname': 'weightDistribution',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'wd'},
                            'weightedInfluence': {'args': '<type __builtin__.bool>',
                                                  'longname': 'weightedInfluence',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'wi'}},
                  'removedFlags': {'is': 'ignoreSelected'},
                  'resultNeedsUnpacking': True,
                  'shortFlags': {'af': 'after',
                                 'ai': 'addInfluence',
                                 'ar': 'afterReference',
                                 'ats': 'addToSelection',
                                 'bf': 'before',
                                 'bm': 'bindMethod',
                                 'bsh': 'baseShape',
                                 'dr': 'dropoffRate',
                                 'dt': 'deformerTools',
                                 'e': 'edit',
                                 'ex': 'exclusive',
                                 'fnw': 'forceNormalizeWeights',
                                 'foc': 'frontOfChain',
                                 'g': 'geometry',
                                 'gi': 'geometryIndices',
                                 'hmf': 'heatmapFalloff',
                                 'ibp': 'ignoreBindPose',
                                 'ignoreSelected': 'ignoreSelected',
                                 'ih': 'ignoreHierarchy',
                                 'ihs': 'includeHiddenSelections',
                                 'inf': 'influence',
                                 'lw': 'lockWeights',
                                 'mi': 'maximumInfluences',
                                 'mjm': 'moveJointsMode',
                                 'n': 'name',
                                 'ns': 'nurbsSamples',
                                 'nw': 'normalizeWeights',
                                 'omi': 'obeyMaxInfluences',
                                 'par': 'parallel',
                                 'pr': 'prune',
                                 'ps': 'polySmoothness',
                                 'q': 'query',
                                 'rbm': 'recacheBindMatrices',
                                 'rfs': 'removeFromSelection',
                                 'ri': 'removeInfluence',
                                 'rm': 'remove',
                                 'rui': 'removeUnusedInfluence',
                                 'siv': 'selectInfluenceVerts',
                                 'sm': 'skinMethod',
                                 'sp': 'split',
                                 'sw': 'smoothWeights',
                                 'swi': 'smoothWeightsMaxIterations',
                                 'tsb': 'toSelectedBones',
                                 'tst': 'toSkeletonAndTransforms',
                                 'ub': 'unbind',
                                 'ubk': 'unbindKeepHistory',
                                 'ug': 'useGeometry',
                                 'vb': 'volumeBind',
                                 'vt': 'volumeType',
                                 'wd': 'weightDistribution',
                                 'wi': 'weightedInfluence',
                                 'wt': 'weight'},
                  'type': 'animation'},
  'skinPercent': {'flags': {'ignoreBelow': {'args': '<type __builtin__.float>',
                                            'longname': 'ignoreBelow',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'ib'},
                            'normalize': {'args': '<type __builtin__.bool>',
                                          'longname': 'normalize',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'nrm'},
                            'pruneWeights': {'args': '<type __builtin__.float>',
                                             'longname': 'pruneWeights',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'prw'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'relative': {'args': '<type __builtin__.bool>',
                                         'longname': 'relative',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'r'},
                            'resetToDefault': {'args': '<type __builtin__.bool>',
                                               'longname': 'resetToDefault',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'rtd'},
                            'transform': {'args': '<type __builtin__.str>',
                                          'longname': 'transform',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 't'},
                            'transformMoveWeights': {'args': '<type __builtin__.str>',
                                                     'longname': 'transformMoveWeights',
                                                     'modes': ['create',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'tmw'},
                            'transformValue': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'transformValue',
                                               'modes': ['create',
                                                         'multiuse'],
                                               'numArgs': 2,
                                               'shortname': 'tv'},
                            'value': {'args': '<type __builtin__.bool>',
                                      'longname': 'value',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'v'},
                            'zeroRemainingInfluences': {'args': '<type __builtin__.bool>',
                                                        'longname': 'zeroRemainingInfluences',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'zri'}},
                  'shortFlags': {'ib': 'ignoreBelow',
                                 'nrm': 'normalize',
                                 'prw': 'pruneWeights',
                                 'q': 'query',
                                 'r': 'relative',
                                 'rtd': 'resetToDefault',
                                 't': 'transform',
                                 'tmw': 'transformMoveWeights',
                                 'tv': 'transformValue',
                                 'v': 'value',
                                 'zri': 'zeroRemainingInfluences'},
                  'type': 'animation'},
  'smoothCurve': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                'longname': 'replaceOriginal',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'rpo'},
                            'smoothness': {'args': '<type __builtin__.float>',
                                           'longname': 'smoothness',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 's'}},
                  'shortFlags': {'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'e': 'edit',
                                 'fzn': 'frozen',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'o': 'object',
                                 'q': 'query',
                                 'rpo': 'replaceOriginal',
                                 's': 'smoothness'},
                  'type': 'modeling'},
  'smoothTangentSurface': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                                 'longname': 'caching',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cch'},
                                     'constructionHistory': {'args': '<type __builtin__.bool>',
                                                             'longname': 'constructionHistory',
                                                             'modes': ['create'],
                                                             'numArgs': 1,
                                                             'shortname': 'ch'},
                                     'direction': {'args': '<type __builtin__.int>',
                                                   'longname': 'direction',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'd'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'frozen': {'args': '<type __builtin__.bool>',
                                                'longname': 'frozen',
                                                'numArgs': 1,
                                                'shortname': 'fzn'},
                                     'name': {'args': '<type __builtin__.str>',
                                              'longname': 'name',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'n'},
                                     'nodeState': {'args': '<type __builtin__.int>',
                                                   'longname': 'nodeState',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'nds'},
                                     'object': {'args': '<type __builtin__.bool>',
                                                'longname': 'object',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'o'},
                                     'parameter': {'args': '<type __builtin__.float>',
                                                   'longname': 'parameter',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'p'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                         'longname': 'replaceOriginal',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'rpo'},
                                     'smoothness': {'args': '<type __builtin__.int>',
                                                    'longname': 'smoothness',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 's'}},
                           'shortFlags': {'cch': 'caching',
                                          'ch': 'constructionHistory',
                                          'd': 'direction',
                                          'e': 'edit',
                                          'fzn': 'frozen',
                                          'n': 'name',
                                          'nds': 'nodeState',
                                          'o': 'object',
                                          'p': 'parameter',
                                          'q': 'query',
                                          'rpo': 'replaceOriginal',
                                          's': 'smoothness'},
                           'type': 'modeling'},
  'snapKey': {'flags': {'animation': {'args': '<type __builtin__.str>',
                                      'longname': 'animation',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'an'},
                        'attribute': {'args': '<type __builtin__.str>',
                                      'longname': 'attribute',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'at'},
                        'controlPoints': {'args': '<type __builtin__.bool>',
                                          'longname': 'controlPoints',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'cp'},
                        'float': {'args': 'floatrange',
                                  'longname': 'float',
                                  'modes': ['create', 'multiuse'],
                                  'numArgs': 1,
                                  'shortname': 'f'},
                        'hierarchy': {'args': '<type __builtin__.str>',
                                      'longname': 'hierarchy',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'hi'},
                        'includeUpperBound': {'args': '<type __builtin__.bool>',
                                              'longname': 'includeUpperBound',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'iub'},
                        'index': {'args': '<type __builtin__.int>',
                                  'longname': 'index',
                                  'modes': ['create', 'multiuse'],
                                  'numArgs': 1,
                                  'shortname': 'index'},
                        'shape': {'args': '<type __builtin__.bool>',
                                  'longname': 'shape',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 's'},
                        'time': {'args': 'timerange',
                                 'longname': 'time',
                                 'modes': ['create', 'multiuse'],
                                 'numArgs': 1,
                                 'shortname': 't'},
                        'timeMultiple': {'args': '<type __builtin__.float>',
                                         'longname': 'timeMultiple',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'tm'},
                        'valueMultiple': {'args': '<type __builtin__.float>',
                                          'longname': 'valueMultiple',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'vm'}},
              'removedFlags': {'in': 'index'},
              'shortFlags': {'an': 'animation',
                             'at': 'attribute',
                             'cp': 'controlPoints',
                             'f': 'float',
                             'hi': 'hierarchy',
                             'index': 'index',
                             'iub': 'includeUpperBound',
                             's': 'shape',
                             't': 'time',
                             'tm': 'timeMultiple',
                             'vm': 'valueMultiple'},
              'type': 'animation'},
  'snapMode': {'flags': {'curve': {'args': '<type __builtin__.bool>',
                                   'longname': 'curve',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'c'},
                         'distanceIncrement': {'args': '<type __builtin__.float>',
                                               'longname': 'distanceIncrement',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'dsi'},
                         'edgeMagnet': {'args': '<type __builtin__.int>',
                                        'longname': 'edgeMagnet',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'em'},
                         'edgeMagnetTolerance': {'args': '<type __builtin__.float>',
                                                 'longname': 'edgeMagnetTolerance',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'emt'},
                         'grid': {'args': '<type __builtin__.bool>',
                                  'longname': 'grid',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'gr'},
                         'liveFaceCenter': {'args': '<type __builtin__.bool>',
                                            'longname': 'liveFaceCenter',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'lfc'},
                         'livePoint': {'args': '<type __builtin__.bool>',
                                       'longname': 'livePoint',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'lp'},
                         'meshCenter': {'args': '<type __builtin__.bool>',
                                        'longname': 'meshCenter',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'mc'},
                         'pixelCenter': {'args': '<type __builtin__.bool>',
                                         'longname': 'pixelCenter',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'pc'},
                         'pixelSnap': {'args': '<type __builtin__.bool>',
                                       'longname': 'pixelSnap',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'ps'},
                         'point': {'args': '<type __builtin__.bool>',
                                   'longname': 'point',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'p'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'tolerance': {'args': '<type __builtin__.int>',
                                       'longname': 'tolerance',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 't'},
                         'useTolerance': {'args': '<type __builtin__.bool>',
                                          'longname': 'useTolerance',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'ut'},
                         'uvTolerance': {'args': '<type __builtin__.int>',
                                         'longname': 'uvTolerance',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'uvt'},
                         'viewPlane': {'args': '<type __builtin__.bool>',
                                       'longname': 'viewPlane',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'vp'}},
               'shortFlags': {'c': 'curve',
                              'dsi': 'distanceIncrement',
                              'em': 'edgeMagnet',
                              'emt': 'edgeMagnetTolerance',
                              'gr': 'grid',
                              'lfc': 'liveFaceCenter',
                              'lp': 'livePoint',
                              'mc': 'meshCenter',
                              'p': 'point',
                              'pc': 'pixelCenter',
                              'ps': 'pixelSnap',
                              'q': 'query',
                              't': 'tolerance',
                              'ut': 'useTolerance',
                              'uvt': 'uvTolerance',
                              'vp': 'viewPlane'},
               'type': 'general'},
  'snapTogetherCtx': {'flags': {'clearSelection': {'args': '<type __builtin__.bool>',
                                                   'longname': 'clearSelection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cs'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'history': {'args': '<type __builtin__.bool>',
                                            'longname': 'history',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ch'},
                                'image1': {'args': '<type __builtin__.str>',
                                           'longname': 'image1',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i1'},
                                'image2': {'args': '<type __builtin__.str>',
                                           'longname': 'image2',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i2'},
                                'image3': {'args': '<type __builtin__.str>',
                                           'longname': 'image3',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i3'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'setOrientation': {'args': '<type __builtin__.bool>',
                                                   'longname': 'setOrientation',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'so'},
                                'snapPolygonFace': {'args': '<type __builtin__.bool>',
                                                    'longname': 'snapPolygonFace',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'spf'}},
                      'shortFlags': {'ch': 'history',
                                     'cs': 'clearSelection',
                                     'e': 'edit',
                                     'ex': 'exists',
                                     'i1': 'image1',
                                     'i2': 'image2',
                                     'i3': 'image3',
                                     'n': 'name',
                                     'q': 'query',
                                     'so': 'setOrientation',
                                     'spf': 'snapPolygonFace'},
                      'type': 'context'},
  'snapshot': {'flags': {'anchorTransform': {'args': '<type __builtin__.str>',
                                             'longname': 'anchorTransform',
                                             'numArgs': 1,
                                             'shortname': 'at'},
                         'constructionHistory': {'args': '<type __builtin__.bool>',
                                                 'longname': 'constructionHistory',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'endTime': {'args': 'time',
                                     'longname': 'endTime',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'et'},
                         'increment': {'args': 'time',
                                       'longname': 'increment',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i'},
                         'motionTrail': {'args': '<type __builtin__.bool>',
                                         'longname': 'motionTrail',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'mt'},
                         'motionTrailName': {'args': '<type __builtin__.str>',
                                             'longname': 'motionTrailName',
                                             'numArgs': 1,
                                             'shortname': 'mtn'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'removeAnchorTransform': {'args': '<type __builtin__.str>',
                                                   'longname': 'removeAnchorTransform',
                                                   'numArgs': 1,
                                                   'shortname': 'rat'},
                         'startTime': {'args': 'time',
                                       'longname': 'startTime',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'st'},
                         'update': {'args': '<type __builtin__.str>',
                                    'longname': 'update',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'u'}},
               'shortFlags': {'at': 'anchorTransform',
                              'ch': 'constructionHistory',
                              'e': 'edit',
                              'et': 'endTime',
                              'i': 'increment',
                              'mt': 'motionTrail',
                              'mtn': 'motionTrailName',
                              'n': 'name',
                              'q': 'query',
                              'rat': 'removeAnchorTransform',
                              'st': 'startTime',
                              'u': 'update'},
               'type': 'animation'},
  'snapshotBeadContext': {'flags': {}, 'shortFlags': {}, 'type': 'context'},
  'snapshotBeadCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'history': {'args': '<type __builtin__.bool>',
                                            'longname': 'history',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ch'},
                                'image1': {'args': '<type __builtin__.str>',
                                           'longname': 'image1',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i1'},
                                'image2': {'args': '<type __builtin__.str>',
                                           'longname': 'image2',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i2'},
                                'image3': {'args': '<type __builtin__.str>',
                                           'longname': 'image3',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i3'},
                                'inTangent': {'args': '<type __builtin__.bool>',
                                              'longname': 'inTangent',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'i'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'outTangent': {'args': '<type __builtin__.bool>',
                                               'longname': 'outTangent',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'o'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'}},
                      'shortFlags': {'ch': 'history',
                                     'e': 'edit',
                                     'ex': 'exists',
                                     'i': 'inTangent',
                                     'i1': 'image1',
                                     'i2': 'image2',
                                     'i3': 'image3',
                                     'n': 'name',
                                     'o': 'outTangent',
                                     'q': 'query'},
                      'type': 'context'},
  'snapshotModifyKeyCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'exists': {'args': '<type __builtin__.bool>',
                                                'longname': 'exists',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ex'},
                                     'history': {'args': '<type __builtin__.bool>',
                                                 'longname': 'history',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ch'},
                                     'image1': {'args': '<type __builtin__.str>',
                                                'longname': 'image1',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'i1'},
                                     'image2': {'args': '<type __builtin__.str>',
                                                'longname': 'image2',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'i2'},
                                     'image3': {'args': '<type __builtin__.str>',
                                                'longname': 'image3',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'i3'},
                                     'name': {'args': '<type __builtin__.str>',
                                              'longname': 'name',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'n'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'}},
                           'shortFlags': {'ch': 'history',
                                          'e': 'edit',
                                          'ex': 'exists',
                                          'i1': 'image1',
                                          'i2': 'image2',
                                          'i3': 'image3',
                                          'n': 'name',
                                          'q': 'query'},
                           'type': 'context'},
  'soft': {'flags': {'convert': {'args': '<type __builtin__.bool>',
                                 'longname': 'convert',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'c'},
                     'duplicate': {'args': '<type __builtin__.bool>',
                                   'longname': 'duplicate',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'd'},
                     'duplicateHistory': {'args': '<type __builtin__.bool>',
                                          'longname': 'duplicateHistory',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'dh'},
                     'goal': {'args': '<type __builtin__.float>',
                              'longname': 'goal',
                              'modes': ['create'],
                              'numArgs': 1,
                              'shortname': 'g'},
                     'hideOriginal': {'args': '<type __builtin__.bool>',
                                      'longname': 'hideOriginal',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                     'name': {'args': '<type __builtin__.str>',
                              'longname': 'name',
                              'modes': [],
                              'numArgs': 1,
                              'shortname': 'n'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'}},
           'shortFlags': {'c': 'convert',
                          'd': 'duplicate',
                          'dh': 'duplicateHistory',
                          'g': 'goal',
                          'h': 'hideOriginal',
                          'n': 'name',
                          'q': 'query'},
           'type': 'effects'},
  'softMod': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                  'longname': 'after',
                                  'modes': ['create', 'edit'],
                                  'numArgs': 0,
                                  'shortname': 'af'},
                        'afterReference': {'args': '<type __builtin__.bool>',
                                           'longname': 'afterReference',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'ar'},
                        'before': {'args': '<type __builtin__.bool>',
                                   'longname': 'before',
                                   'modes': ['create', 'edit'],
                                   'numArgs': 0,
                                   'shortname': 'bf'},
                        'bindState': {'args': '<type __builtin__.bool>',
                                      'longname': 'bindState',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'bs'},
                        'curveInterpolation': {'args': '<type __builtin__.int>',
                                               'longname': 'curveInterpolation',
                                               'modes': ['create',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'ci'},
                        'curvePoint': {'args': '<type __builtin__.float>',
                                       'longname': 'curvePoint',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'cp'},
                        'curveValue': {'args': '<type __builtin__.float>',
                                       'longname': 'curveValue',
                                       'modes': ['create', 'multiuse'],
                                       'numArgs': 1,
                                       'shortname': 'cv'},
                        'deformerTools': {'args': '<type __builtin__.bool>',
                                          'longname': 'deformerTools',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'dt'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'envelope': {'args': '<type __builtin__.float>',
                                     'longname': 'envelope',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'en'},
                        'exclusive': {'args': '<type __builtin__.str>',
                                      'longname': 'exclusive',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'ex'},
                        'falloffAroundSelection': {'args': '<type __builtin__.bool>',
                                                   'longname': 'falloffAroundSelection',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'fas'},
                        'falloffBasedOnX': {'args': '<type __builtin__.bool>',
                                            'longname': 'falloffBasedOnX',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'fbx'},
                        'falloffBasedOnY': {'args': '<type __builtin__.bool>',
                                            'longname': 'falloffBasedOnY',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'fby'},
                        'falloffBasedOnZ': {'args': '<type __builtin__.bool>',
                                            'longname': 'falloffBasedOnZ',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'fbz'},
                        'falloffCenter': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'falloffCenter',
                                          'modes': ['create'],
                                          'numArgs': 3,
                                          'shortname': 'fc'},
                        'falloffMasking': {'args': '<type __builtin__.bool>',
                                           'longname': 'falloffMasking',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'fm'},
                        'falloffMode': {'args': '<type __builtin__.int>',
                                        'longname': 'falloffMode',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'fom'},
                        'falloffRadius': {'args': '<type __builtin__.float>',
                                          'longname': 'falloffRadius',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'fr'},
                        'frontOfChain': {'args': '<type __builtin__.bool>',
                                         'longname': 'frontOfChain',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'foc'},
                        'geometry': {'args': '<type __builtin__.str>',
                                     'longname': 'geometry',
                                     'modes': ['query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'g'},
                        'geometryIndices': {'args': '<type __builtin__.bool>',
                                            'longname': 'geometryIndices',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'gi'},
                        'ignoreSelected': {'args': '<type __builtin__.bool>',
                                           'longname': 'ignoreSelected',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ignoreSelected'},
                        'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                    'longname': 'includeHiddenSelections',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ihs'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'parallel': {'args': '<type __builtin__.bool>',
                                     'longname': 'parallel',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'par'},
                        'prune': {'args': '<type __builtin__.bool>',
                                  'longname': 'prune',
                                  'modes': ['edit'],
                                  'numArgs': 0,
                                  'shortname': 'pr'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'relative': {'args': '<type __builtin__.bool>',
                                     'longname': 'relative',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'rel'},
                        'remove': {'args': '<type __builtin__.bool>',
                                   'longname': 'remove',
                                   'modes': ['edit', 'multiuse'],
                                   'numArgs': 0,
                                   'shortname': 'rm'},
                        'resetGeometry': {'args': '<type __builtin__.bool>',
                                          'longname': 'resetGeometry',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'rg'},
                        'split': {'args': '<type __builtin__.bool>',
                                  'longname': 'split',
                                  'modes': ['create', 'edit'],
                                  'numArgs': 0,
                                  'shortname': 'sp'},
                        'weightedNode': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'weightedNode',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 2,
                                         'shortname': 'wn'}},
              'removedFlags': {'is': 'ignoreSelected'},
              'shortFlags': {'af': 'after',
                             'ar': 'afterReference',
                             'bf': 'before',
                             'bs': 'bindState',
                             'ci': 'curveInterpolation',
                             'cp': 'curvePoint',
                             'cv': 'curveValue',
                             'dt': 'deformerTools',
                             'e': 'edit',
                             'en': 'envelope',
                             'ex': 'exclusive',
                             'fas': 'falloffAroundSelection',
                             'fbx': 'falloffBasedOnX',
                             'fby': 'falloffBasedOnY',
                             'fbz': 'falloffBasedOnZ',
                             'fc': 'falloffCenter',
                             'fm': 'falloffMasking',
                             'foc': 'frontOfChain',
                             'fom': 'falloffMode',
                             'fr': 'falloffRadius',
                             'g': 'geometry',
                             'gi': 'geometryIndices',
                             'ignoreSelected': 'ignoreSelected',
                             'ihs': 'includeHiddenSelections',
                             'n': 'name',
                             'par': 'parallel',
                             'pr': 'prune',
                             'q': 'query',
                             'rel': 'relative',
                             'rg': 'resetGeometry',
                             'rm': 'remove',
                             'sp': 'split',
                             'wn': 'weightedNode'},
              'type': 'animation'},
  'softModContext': {'flags': {}, 'shortFlags': {}, 'type': 'context'},
  'softModCtx': {'flags': {'dragSlider': {'args': '<type __builtin__.str>',
                                          'longname': 'dragSlider',
                                          'modes': ['edit'],
                                          'numArgs': 1,
                                          'shortname': 'ds'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'falseColor': {'args': '<type __builtin__.bool>',
                                          'longname': 'falseColor',
                                          'modes': ['edit'],
                                          'numArgs': 1,
                                          'shortname': 'fc'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'reset': {'args': '<type __builtin__.bool>',
                                     'longname': 'reset',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'rst'}},
                 'shortFlags': {'ds': 'dragSlider',
                                'e': 'edit',
                                'ex': 'exists',
                                'fc': 'falseColor',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'q': 'query',
                                'rst': 'reset'},
                 'type': 'context'},
  'softSelect': {'flags': {'compressUndo': {'args': '<type __builtin__.int>',
                                            'longname': 'compressUndo',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cu'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'enableFalseColor': {'args': '<type __builtin__.int>',
                                                'longname': 'enableFalseColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'efc'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'softSelectColorCurve': {'args': '<type __builtin__.str>',
                                                    'longname': 'softSelectColorCurve',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'scc'},
                           'softSelectCurve': {'args': '<type __builtin__.str>',
                                               'longname': 'softSelectCurve',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ssc'},
                           'softSelectDistance': {'args': '<type __builtin__.float>',
                                                  'longname': 'softSelectDistance',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ssd'},
                           'softSelectEnabled': {'args': '<type __builtin__.int>',
                                                 'longname': 'softSelectEnabled',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sse'},
                           'softSelectFalloff': {'args': '<type __builtin__.int>',
                                                 'longname': 'softSelectFalloff',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ssf'},
                           'softSelectReset': {'args': '<type __builtin__.bool>',
                                               'longname': 'softSelectReset',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'ssr'},
                           'softSelectUVDistance': {'args': '<type __builtin__.float>',
                                                    'longname': 'softSelectUVDistance',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sud'}},
                 'shortFlags': {'cu': 'compressUndo',
                                'e': 'edit',
                                'efc': 'enableFalseColor',
                                'q': 'query',
                                'scc': 'softSelectColorCurve',
                                'ssc': 'softSelectCurve',
                                'ssd': 'softSelectDistance',
                                'sse': 'softSelectEnabled',
                                'ssf': 'softSelectFalloff',
                                'ssr': 'softSelectReset',
                                'sud': 'softSelectUVDistance'},
                 'type': 'general'},
  'softSelectOptionsCtx': {'flags': {'buttonDown': {'args': '<type __builtin__.bool>',
                                                    'longname': 'buttonDown',
                                                    'numArgs': 0,
                                                    'shortname': 'btd'},
                                     'buttonUp': {'args': '<type __builtin__.bool>',
                                                  'longname': 'buttonUp',
                                                  'numArgs': 0,
                                                  'shortname': 'btu'},
                                     'colorCurve': {'args': '<type __builtin__.str>',
                                                    'longname': 'colorCurve',
                                                    'numArgs': 1,
                                                    'shortname': 'cc'},
                                     'condition': {'args': '<type __builtin__.bool>',
                                                   'longname': 'condition',
                                                   'numArgs': 0,
                                                   'shortname': 'cdn'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'enableFalseColor': {'args': '<type __builtin__.int>',
                                                          'longname': 'enableFalseColor',
                                                          'numArgs': 1,
                                                          'shortname': 'efc'},
                                     'enabled': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enabled',
                                                 'numArgs': 1,
                                                 'shortname': 'en'},
                                     'exists': {'args': '<type __builtin__.bool>',
                                                'longname': 'exists',
                                                'numArgs': 0,
                                                'shortname': 'ex'},
                                     'falloffCurve': {'args': '<type __builtin__.str>',
                                                      'longname': 'falloffCurve',
                                                      'numArgs': 1,
                                                      'shortname': 'fc'},
                                     'falloffMode': {'args': '<type __builtin__.int>',
                                                     'longname': 'falloffMode',
                                                     'numArgs': 1,
                                                     'shortname': 'fm'},
                                     'image1': {'args': '<type __builtin__.str>',
                                                'longname': 'image1',
                                                'numArgs': 1,
                                                'shortname': 'i1'},
                                     'image2': {'args': '<type __builtin__.str>',
                                                'longname': 'image2',
                                                'numArgs': 1,
                                                'shortname': 'i2'},
                                     'image3': {'args': '<type __builtin__.str>',
                                                'longname': 'image3',
                                                'numArgs': 1,
                                                'shortname': 'i3'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'size': {'args': '<type __builtin__.float>',
                                              'longname': 'size',
                                              'numArgs': 1,
                                              'shortname': 'sz'},
                                     'uvSize': {'args': '<type __builtin__.float>',
                                                'longname': 'uvSize',
                                                'numArgs': 1,
                                                'shortname': 'uvs'}},
                           'shortFlags': {'btd': 'buttonDown',
                                          'btu': 'buttonUp',
                                          'cc': 'colorCurve',
                                          'cdn': 'condition',
                                          'e': 'edit',
                                          'efc': 'enableFalseColor',
                                          'en': 'enabled',
                                          'ex': 'exists',
                                          'fc': 'falloffCurve',
                                          'fm': 'falloffMode',
                                          'i1': 'image1',
                                          'i2': 'image2',
                                          'i3': 'image3',
                                          'q': 'query',
                                          'sz': 'size',
                                          'uvs': 'uvSize'},
                           'type': 'context'},
  'soloMaterial': {'flags': {'attr': {'args': '<type __builtin__.str>',
                                      'longname': 'attr',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'a'},
                             'last': {'args': '<type __builtin__.bool>',
                                      'longname': 'last',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'l'},
                             'node': {'args': '<type __builtin__.str>',
                                      'longname': 'node',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'unsolo': {'args': '<type __builtin__.bool>',
                                        'longname': 'unsolo',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'us'}},
                   'shortFlags': {'a': 'attr',
                                  'l': 'last',
                                  'n': 'node',
                                  'q': 'query',
                                  'us': 'unsolo'},
                   'type': 'rendering'},
  'sortCaseInsensitive': {'flags': {}, 'shortFlags': {}, 'type': 'language'},
  'sound': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                               'longname': 'edit',
                               'numArgs': 0,
                               'shortname': 'e'},
                      'endTime': {'args': 'time',
                                  'longname': 'endTime',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'et'},
                      'file': {'args': '<type __builtin__.str>',
                               'longname': 'file',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'f'},
                      'length': {'args': '<type __builtin__.bool>',
                                 'longname': 'length',
                                 'modes': ['query'],
                                 'numArgs': 0,
                                 'shortname': 'l'},
                      'mute': {'args': '<type __builtin__.bool>',
                               'longname': 'mute',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'm'},
                      'name': {'args': '<type __builtin__.str>',
                               'longname': 'name',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'n'},
                      'offset': {'args': 'time',
                                 'longname': 'offset',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'o'},
                      'query': {'args': '<type __builtin__.bool>',
                                'longname': 'query',
                                'numArgs': 0,
                                'shortname': 'q'},
                      'sourceEnd': {'args': 'time',
                                    'longname': 'sourceEnd',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'se'},
                      'sourceStart': {'args': 'time',
                                      'longname': 'sourceStart',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'ss'}},
            'shortFlags': {'e': 'edit',
                           'et': 'endTime',
                           'f': 'file',
                           'l': 'length',
                           'm': 'mute',
                           'n': 'name',
                           'o': 'offset',
                           'q': 'query',
                           'se': 'sourceEnd',
                           'st': 'sourceStart'},
            'type': 'animation'},
  'soundControl': {'callbackFlags': ['releaseCommand',
                                     'rc',
                                     'pressCommand',
                                     'pc',
                                     'visibleChangeCommand',
                                     'vcc',
                                     'dragCallback',
                                     'dgc',
                                     'dropCallback',
                                     'dpc'],
                   'flags': {'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'beginScrub': {'args': '<type __builtin__.bool>',
                                            'longname': 'beginScrub',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'bs'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'displaySound': {'args': '<type __builtin__.bool>',
                                              'longname': 'displaySound',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ds'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'endScrub': {'args': '<type __builtin__.bool>',
                                          'longname': 'endScrub',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'es'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'maxTime': {'args': 'time',
                                         'longname': 'maxTime',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'max'},
                             'minTime': {'args': 'time',
                                         'longname': 'minTime',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'min'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'pressCommand': {'args': '<type __builtin__.str>',
                                              'longname': 'pressCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pc'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'releaseCommand': {'args': '<type __builtin__.str>',
                                                'longname': 'releaseCommand',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'rc'},
                             'repeatChunkSize': {'args': '<type __builtin__.float>',
                                                 'longname': 'repeatChunkSize',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rcs'},
                             'repeatOnHold': {'args': '<type __builtin__.bool>',
                                              'longname': 'repeatOnHold',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'roh'},
                             'resample': {'args': '<type __builtin__.bool>',
                                          'longname': 'resample',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'r'},
                             'sound': {'args': '<type __builtin__.str>',
                                       'longname': 'sound',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 's'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'waveform': {'args': '<type __builtin__.str>',
                                          'longname': 'waveform',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'wf'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'shortFlags': {'ann': 'annotation',
                                  'bgc': 'backgroundColor',
                                  'bs': 'beginScrub',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'ds': 'displaySound',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ebg': 'enableBackground',
                                  'ekf': 'enableKeyboardFocus',
                                  'en': 'enable',
                                  'es': 'endScrub',
                                  'ex': 'exists',
                                  'fpn': 'fullPathName',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'io': 'isObscured',
                                  'm': 'manage',
                                  'max': 'maxTime',
                                  'min': 'minTime',
                                  'nbg': 'noBackground',
                                  'npm': 'numberOfPopupMenus',
                                  'p': 'parent',
                                  'pc': 'pressCommand',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'q': 'query',
                                  'r': 'resample',
                                  'rc': 'releaseCommand',
                                  'rcs': 'repeatChunkSize',
                                  'roh': 'repeatOnHold',
                                  's': 'sound',
                                  'sbm': 'statusBarMessage',
                                  'ut': 'useTemplate',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'w': 'width',
                                  'wf': 'waveform'},
                   'type': 'windows'},
  'spBirailCtx': {'flags': {'activeNodes': {'args': '<type __builtin__.bool>',
                                            'longname': 'activeNodes',
                                            'numArgs': 0,
                                            'shortname': 'anq'},
                            'autoCreate': {'args': '<type __builtin__.bool>',
                                           'longname': 'autoCreate',
                                           'numArgs': 1,
                                           'shortname': 'ac'},
                            'bldProfileFirst': {'args': '<type __builtin__.bool>',
                                                'longname': 'bldProfileFirst',
                                                'numArgs': 1,
                                                'shortname': 'bpf'},
                            'bldProfileLast': {'args': '<type __builtin__.bool>',
                                               'longname': 'bldProfileLast',
                                               'numArgs': 1,
                                               'shortname': 'bpl'},
                            'bldProfiles': {'args': '<type __builtin__.bool>',
                                            'longname': 'bldProfiles',
                                            'numArgs': 1,
                                            'shortname': 'bp'},
                            'bldRailOne': {'args': '<type __builtin__.bool>',
                                           'longname': 'bldRailOne',
                                           'numArgs': 1,
                                           'shortname': 'br1'},
                            'bldRailTwo': {'args': '<type __builtin__.bool>',
                                           'longname': 'bldRailTwo',
                                           'numArgs': 1,
                                           'shortname': 'br2'},
                            'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'image1': {'args': '<type __builtin__.str>',
                                       'longname': 'image1',
                                       'numArgs': 1,
                                       'shortname': 'i1'},
                            'image2': {'args': '<type __builtin__.str>',
                                       'longname': 'image2',
                                       'numArgs': 1,
                                       'shortname': 'i2'},
                            'image3': {'args': '<type __builtin__.str>',
                                       'longname': 'image3',
                                       'numArgs': 1,
                                       'shortname': 'i3'},
                            'immediate': {'args': '<type __builtin__.bool>',
                                          'longname': 'immediate',
                                          'numArgs': 0,
                                          'shortname': 'im'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'polygon': {'args': '<type __builtin__.int>',
                                        'longname': 'polygon',
                                        'numArgs': 1,
                                        'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'reset': {'args': '<type __builtin__.bool>',
                                      'longname': 'reset',
                                      'numArgs': 0,
                                      'shortname': 'rs'},
                            'tangentContinuityProfile1': {'args': '<type __builtin__.bool>',
                                                          'longname': 'tangentContinuityProfile1',
                                                          'numArgs': 1,
                                                          'shortname': 'tp1'},
                            'toolNode': {'args': '<type __builtin__.bool>',
                                         'longname': 'toolNode',
                                         'numArgs': 0,
                                         'shortname': 'tnq'},
                            'transformMode': {'args': '<type __builtin__.int>',
                                              'longname': 'transformMode',
                                              'numArgs': 1,
                                              'shortname': 'tm'}},
                  'shortFlags': {'ac': 'autoCreate',
                                 'anq': 'activeNodes',
                                 'bp': 'bldProfiles',
                                 'bpf': 'bldProfileFirst',
                                 'bpl': 'bldProfileLast',
                                 'br1': 'bldRailOne',
                                 'br2': 'bldRailTwo',
                                 'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'e': 'edit',
                                 'ex': 'exists',
                                 'fzn': 'frozen',
                                 'i1': 'image1',
                                 'i2': 'image2',
                                 'i3': 'image3',
                                 'im': 'immediate',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'o': 'object',
                                 'po': 'polygon',
                                 'q': 'query',
                                 'rs': 'reset',
                                 'tm': 'transformMode',
                                 'tnq': 'toolNode',
                                 'tp1': 'tangentContinuityProfile1'},
                  'type': 'context'},
  'spaceLocator': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                          'longname': 'absolute',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'a'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'position': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'position',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 3,
                                          'shortname': 'p'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'relative': {'args': '<type __builtin__.bool>',
                                          'longname': 'relative',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'r'}},
                   'resultNeedsUnpacking': True,
                   'shortFlags': {'a': 'absolute',
                                  'e': 'edit',
                                  'n': 'name',
                                  'p': 'position',
                                  'q': 'query',
                                  'r': 'relative'},
                   'type': 'general'},
  'sphere': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                         '<type __builtin__.float>',
                                         '<type __builtin__.float>'],
                                'longname': 'axis',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 3,
                                'shortname': 'ax'},
                       'caching': {'args': '<type __builtin__.bool>',
                                   'longname': 'caching',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'resultNeedsCasting': True,
                                   'shortname': 'cch'},
                       'constructionHistory': {'args': '<type __builtin__.bool>',
                                               'longname': 'constructionHistory',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                       'degree': {'args': '<type __builtin__.int>',
                                  'longname': 'degree',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'resultNeedsCasting': True,
                                  'shortname': 'd'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'endSweep': {'args': '<type __builtin__.float>',
                                    'longname': 'endSweep',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'esw'},
                       'frozen': {'args': '<type __builtin__.bool>',
                                  'longname': 'frozen',
                                  'modes': ['query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'fzn'},
                       'heightRatio': {'args': '<type __builtin__.float>',
                                       'longname': 'heightRatio',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'hr'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'nodeState': {'args': '<type __builtin__.int>',
                                     'longname': 'nodeState',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsCasting': True,
                                     'shortname': 'nds'},
                       'object': {'args': '<type __builtin__.bool>',
                                  'longname': 'object',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'o'},
                       'pivot': {'args': ['<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>'],
                                 'longname': 'pivot',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 3,
                                 'shortname': 'p'},
                       'polygon': {'args': '<type __builtin__.int>',
                                   'longname': 'polygon',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'po'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'radius': {'args': '<type __builtin__.float>',
                                  'longname': 'radius',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'r'},
                       'sections': {'args': '<type __builtin__.int>',
                                    'longname': 'sections',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'resultNeedsCasting': True,
                                    'shortname': 's'},
                       'spans': {'args': '<type __builtin__.int>',
                                 'longname': 'spans',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'resultNeedsCasting': True,
                                 'shortname': 'nsp'},
                       'startSweep': {'args': '<type __builtin__.float>',
                                      'longname': 'startSweep',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'ssw'},
                       'tolerance': {'args': '<type __builtin__.float>',
                                     'longname': 'tolerance',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'tol'},
                       'useTolerance': {'args': '<type __builtin__.bool>',
                                        'longname': 'useTolerance',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'resultNeedsCasting': True,
                                        'shortname': 'ut'}},
             'shortFlags': {'ax': 'axis',
                            'cch': 'caching',
                            'ch': 'constructionHistory',
                            'd': 'degree',
                            'e': 'edit',
                            'esw': 'endSweep',
                            'fzn': 'frozen',
                            'hr': 'heightRatio',
                            'n': 'name',
                            'nds': 'nodeState',
                            'nsp': 'spans',
                            'o': 'object',
                            'p': 'pivot',
                            'po': 'polygon',
                            'q': 'query',
                            'r': 'radius',
                            's': 'sections',
                            'ssw': 'startSweep',
                            'tol': 'tolerance',
                            'ut': 'useTolerance'},
             'type': 'modeling'},
  'spotLight': {'flags': {'barnDoors': {'args': '<type __builtin__.bool>',
                                        'longname': 'barnDoors',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'bd'},
                          'bottomBarnDoorAngle': {'args': '<type __builtin__.float>',
                                                  'longname': 'bottomBarnDoorAngle',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'bbd'},
                          'coneAngle': {'args': '<type __builtin__.float>',
                                        'longname': 'coneAngle',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ca'},
                          'decayRate': {'args': '<type __builtin__.int>',
                                        'longname': 'decayRate',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'd'},
                          'discRadius': {'args': '<type __builtin__.float>',
                                         'longname': 'discRadius',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'drs'},
                          'dropOff': {'args': '<type __builtin__.float>',
                                      'longname': 'dropOff',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'do'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'exclusive': {'args': '<type __builtin__.bool>',
                                        'longname': 'exclusive',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'exc'},
                          'intensity': {'args': '<type __builtin__.float>',
                                        'longname': 'intensity',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'i'},
                          'leftBarnDoorAngle': {'args': '<type __builtin__.float>',
                                                'longname': 'leftBarnDoorAngle',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'lbd'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'penumbra': {'args': '<type __builtin__.float>',
                                       'longname': 'penumbra',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                          'position': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'position',
                                       'modes': ['create', 'query'],
                                       'numArgs': 3,
                                       'shortname': 'pos'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'rgb': {'args': ['<type __builtin__.float>',
                                           '<type __builtin__.float>',
                                           '<type __builtin__.float>'],
                                  'longname': 'rgb',
                                  'modes': ['create', 'query'],
                                  'numArgs': 3,
                                  'shortname': 'rgb'},
                          'rightBarnDoorAngle': {'args': '<type __builtin__.float>',
                                                 'longname': 'rightBarnDoorAngle',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rbd'},
                          'rotation': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'rotation',
                                       'modes': ['create', 'query'],
                                       'numArgs': 3,
                                       'shortname': 'rot'},
                          'shadowColor': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'shadowColor',
                                          'modes': ['create', 'query'],
                                          'numArgs': 3,
                                          'shortname': 'sc'},
                          'shadowDither': {'args': '<type __builtin__.float>',
                                           'longname': 'shadowDither',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'sd'},
                          'shadowSamples': {'args': '<type __builtin__.int>',
                                            'longname': 'shadowSamples',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'sh'},
                          'softShadow': {'args': '<type __builtin__.bool>',
                                         'longname': 'softShadow',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'ss'},
                          'topBarnDoorAngle': {'args': '<type __builtin__.float>',
                                               'longname': 'topBarnDoorAngle',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tbd'},
                          'useRayTraceShadows': {'args': '<type __builtin__.bool>',
                                                 'longname': 'useRayTraceShadows',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'rs'}},
                'shortFlags': {'bbd': 'bottomBarnDoorAngle',
                               'bd': 'barnDoors',
                               'ca': 'coneAngle',
                               'd': 'decayRate',
                               'do': 'dropOff',
                               'drs': 'discRadius',
                               'e': 'edit',
                               'exc': 'exclusive',
                               'i': 'intensity',
                               'lbd': 'leftBarnDoorAngle',
                               'n': 'name',
                               'p': 'penumbra',
                               'pos': 'position',
                               'q': 'query',
                               'rbd': 'rightBarnDoorAngle',
                               'rgb': 'rgb',
                               'rot': 'rotation',
                               'rs': 'useRayTraceShadows',
                               'sc': 'shadowColor',
                               'sd': 'shadowDither',
                               'sh': 'shadowSamples',
                               'ss': 'softShadow',
                               'tbd': 'topBarnDoorAngle'},
                'type': 'rendering'},
  'spotLightPreviewPort': {'callbackFlags': ['visibleChangeCommand',
                                             'vcc',
                                             'dragCallback',
                                             'dgc',
                                             'dropCallback',
                                             'dpc'],
                           'flags': {'annotation': {'args': '<type __builtin__.str>',
                                                    'longname': 'annotation',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ann'},
                                     'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'backgroundColor',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 3,
                                                         'shortname': 'bgc'},
                                     'defineTemplate': {'args': '<type __builtin__.str>',
                                                        'longname': 'defineTemplate',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'dt'},
                                     'docTag': {'args': '<type __builtin__.str>',
                                                'longname': 'docTag',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dtg'},
                                     'dragCallback': {'args': 'script',
                                                      'longname': 'dragCallback',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dgc'},
                                     'dropCallback': {'args': 'script',
                                                      'longname': 'dropCallback',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dpc'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'enable': {'args': '<type __builtin__.bool>',
                                                'longname': 'enable',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'en'},
                                     'enableBackground': {'args': '<type __builtin__.bool>',
                                                          'longname': 'enableBackground',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ebg'},
                                     'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                             'longname': 'enableKeyboardFocus',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'ekf'},
                                     'exists': {'args': '<type __builtin__.bool>',
                                                'longname': 'exists',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ex'},
                                     'fullPathName': {'args': '<type __builtin__.bool>',
                                                      'longname': 'fullPathName',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'fpn'},
                                     'height': {'args': '<type __builtin__.int>',
                                                'longname': 'height',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'h'},
                                     'highlightColor': {'args': ['<type __builtin__.float>',
                                                                 '<type __builtin__.float>',
                                                                 '<type __builtin__.float>'],
                                                        'longname': 'highlightColor',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 3,
                                                        'shortname': 'hlc'},
                                     'isObscured': {'args': '<type __builtin__.bool>',
                                                    'longname': 'isObscured',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'io'},
                                     'manage': {'args': '<type __builtin__.bool>',
                                                'longname': 'manage',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'm'},
                                     'noBackground': {'args': '<type __builtin__.bool>',
                                                      'longname': 'noBackground',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'nbg'},
                                     'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                            'longname': 'numberOfPopupMenus',
                                                            'modes': ['query'],
                                                            'numArgs': 0,
                                                            'shortname': 'npm'},
                                     'parent': {'args': '<type __builtin__.str>',
                                                'longname': 'parent',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'p'},
                                     'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                        'longname': 'popupMenuArray',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'pma'},
                                     'preventOverride': {'args': '<type __builtin__.bool>',
                                                         'longname': 'preventOverride',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'po'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'spotLight': {'args': 'PyNode',
                                                   'longname': 'spotLight',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'sl'},
                                     'statusBarMessage': {'args': '<type __builtin__.str>',
                                                          'longname': 'statusBarMessage',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'sbm'},
                                     'useTemplate': {'args': '<type __builtin__.str>',
                                                     'longname': 'useTemplate',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ut'},
                                     'visible': {'args': '<type __builtin__.bool>',
                                                 'longname': 'visible',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'vis'},
                                     'visibleChangeCommand': {'args': 'script',
                                                              'longname': 'visibleChangeCommand',
                                                              'modes': ['create',
                                                                        'query',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'vcc'},
                                     'width': {'args': '<type __builtin__.int>',
                                               'longname': 'width',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'w'},
                                     'widthHeight': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'widthHeight',
                                                     'modes': ['create'],
                                                     'numArgs': 2,
                                                     'shortname': 'wh'}},
                           'shortFlags': {'ann': 'annotation',
                                          'bgc': 'backgroundColor',
                                          'dgc': 'dragCallback',
                                          'dpc': 'dropCallback',
                                          'dt': 'defineTemplate',
                                          'dtg': 'docTag',
                                          'e': 'edit',
                                          'ebg': 'enableBackground',
                                          'ekf': 'enableKeyboardFocus',
                                          'en': 'enable',
                                          'ex': 'exists',
                                          'fpn': 'fullPathName',
                                          'h': 'height',
                                          'hlc': 'highlightColor',
                                          'io': 'isObscured',
                                          'm': 'manage',
                                          'nbg': 'noBackground',
                                          'npm': 'numberOfPopupMenus',
                                          'p': 'parent',
                                          'pma': 'popupMenuArray',
                                          'po': 'preventOverride',
                                          'q': 'query',
                                          'sbm': 'statusBarMessage',
                                          'sl': 'spotLight',
                                          'ut': 'useTemplate',
                                          'vcc': 'visibleChangeCommand',
                                          'vis': 'visible',
                                          'w': 'width',
                                          'wh': 'widthHeight'},
                           'type': 'rendering'},
  'spreadSheetEditor': {'flags': {'allAttr': {'args': '<type __builtin__.bool>',
                                              'longname': 'allAttr',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'aa'},
                                  'attrRegExp': {'args': '<type __builtin__.str>',
                                                 'longname': 'attrRegExp',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'are'},
                                  'control': {'args': '<type __builtin__.bool>',
                                              'longname': 'control',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'ctl'},
                                  'defineTemplate': {'args': '<type __builtin__.str>',
                                                     'longname': 'defineTemplate',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'dt'},
                                  'docTag': {'args': '<type __builtin__.str>',
                                             'longname': 'docTag',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dtg'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'execute': {'args': '<type __builtin__.str>',
                                              'longname': 'execute',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'exe'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'filter': {'args': '<type __builtin__.str>',
                                             'longname': 'filter',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'f'},
                                  'fixedAttrList': {'args': '<type __builtin__.str>',
                                                    'longname': 'fixedAttrList',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'fal'},
                                  'forceMainConnection': {'args': '<type __builtin__.str>',
                                                          'longname': 'forceMainConnection',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'fmc'},
                                  'highlightConnection': {'args': '<type __builtin__.str>',
                                                          'longname': 'highlightConnection',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'hlc'},
                                  'keyableOnly': {'args': '<type __builtin__.bool>',
                                                  'longname': 'keyableOnly',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ko'},
                                  'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'lockMainConnection',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'lck'},
                                  'longNames': {'args': '<type __builtin__.bool>',
                                                'longname': 'longNames',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ln'},
                                  'mainListConnection': {'args': '<type __builtin__.str>',
                                                         'longname': 'mainListConnection',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'mlc'},
                                  'niceNames': {'args': '<type __builtin__.bool>',
                                                'longname': 'niceNames',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nn'},
                                  'panel': {'args': '<type __builtin__.str>',
                                            'longname': 'panel',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'pnl'},
                                  'parent': {'args': '<type __builtin__.str>',
                                             'longname': 'parent',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'p'},
                                  'precision': {'args': '<type __builtin__.int>',
                                                'longname': 'precision',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'pre'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'selectedAttr': {'args': '<type __builtin__.bool>',
                                                   'longname': 'selectedAttr',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'sla'},
                                  'selectionConnection': {'args': '<type __builtin__.str>',
                                                          'longname': 'selectionConnection',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'slc'},
                                  'showShapes': {'args': '<type __builtin__.bool>',
                                                 'longname': 'showShapes',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ss'},
                                  'stateString': {'args': '<type __builtin__.bool>',
                                                  'longname': 'stateString',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'sts'},
                                  'unParent': {'args': '<type __builtin__.bool>',
                                               'longname': 'unParent',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'up'},
                                  'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                           'longname': 'unlockMainConnection',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 0,
                                                           'shortname': 'ulk'},
                                  'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                           'longname': 'updateMainConnection',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 0,
                                                           'shortname': 'upd'},
                                  'useTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'useTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ut'}},
                        'shortFlags': {'aa': 'allAttr',
                                       'are': 'attrRegExp',
                                       'ctl': 'control',
                                       'dt': 'defineTemplate',
                                       'dtg': 'docTag',
                                       'e': 'edit',
                                       'ex': 'exists',
                                       'exe': 'execute',
                                       'f': 'filter',
                                       'fal': 'fixedAttrList',
                                       'fmc': 'forceMainConnection',
                                       'hlc': 'highlightConnection',
                                       'ko': 'keyableOnly',
                                       'lck': 'lockMainConnection',
                                       'ln': 'longNames',
                                       'mlc': 'mainListConnection',
                                       'nn': 'niceNames',
                                       'p': 'parent',
                                       'pnl': 'panel',
                                       'pre': 'precision',
                                       'q': 'query',
                                       'sla': 'selectedAttr',
                                       'slc': 'selectionConnection',
                                       'ss': 'showShapes',
                                       'sts': 'stateString',
                                       'ulk': 'unlockMainConnection',
                                       'up': 'unParent',
                                       'upd': 'updateMainConnection',
                                       'ut': 'useTemplate'},
                        'type': 'windows'},
  'spring': {'flags': {'addSprings': {'args': '<type __builtin__.bool>',
                                      'longname': 'addSprings',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'add'},
                       'allPoints': {'args': '<type __builtin__.bool>',
                                     'longname': 'allPoints',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'all'},
                       'count': {'args': '<type __builtin__.bool>',
                                 'longname': 'count',
                                 'modes': ['query'],
                                 'numArgs': 0,
                                 'shortname': 'ct'},
                       'damp': {'args': '<type __builtin__.float>',
                                'longname': 'damp',
                                'numArgs': 1,
                                'shortname': 'dmp'},
                       'damping': {'args': '<type __builtin__.float>',
                                   'longname': 'damping',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'd'},
                       'dampingPS': {'args': '<type __builtin__.float>',
                                     'longname': 'dampingPS',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'dPS'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'endForceWeight': {'args': '<type __builtin__.float>',
                                          'longname': 'endForceWeight',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'efw'},
                       'exclusive': {'args': '<type __builtin__.bool>',
                                     'longname': 'exclusive',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'exc'},
                       'length': {'args': '<type __builtin__.float>',
                                  'longname': 'length',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'l'},
                       'maxDistance': {'args': '<type __builtin__.float>',
                                       'longname': 'maxDistance',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'mxd'},
                       'minDistance': {'args': '<type __builtin__.float>',
                                       'longname': 'minDistance',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'mnd'},
                       'minMax': {'args': '<type __builtin__.bool>',
                                  'longname': 'minMax',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'mm'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['create', 'query'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'noDuplicate': {'args': '<type __builtin__.bool>',
                                       'longname': 'noDuplicate',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'nd'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'restLength': {'args': '<type __builtin__.float>',
                                      'longname': 'restLength',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'rl'},
                       'restLengthPS': {'args': '<type __builtin__.float>',
                                        'longname': 'restLengthPS',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'rPS'},
                       'startForceWeight': {'args': '<type __builtin__.float>',
                                            'longname': 'startForceWeight',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sfw'},
                       'stiffness': {'args': '<type __builtin__.float>',
                                     'longname': 'stiffness',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 's'},
                       'stiffnessPS': {'args': '<type __builtin__.float>',
                                       'longname': 'stiffnessPS',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sPS'},
                       'strength': {'args': '<type __builtin__.float>',
                                    'longname': 'strength',
                                    'numArgs': 1,
                                    'shortname': 'str'},
                       'useDampingPS': {'args': '<type __builtin__.bool>',
                                        'longname': 'useDampingPS',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'udp'},
                       'useRestLengthPS': {'args': '<type __builtin__.bool>',
                                           'longname': 'useRestLengthPS',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'urp'},
                       'useStiffnessPS': {'args': '<type __builtin__.bool>',
                                          'longname': 'useStiffnessPS',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'usp'},
                       'walkLength': {'args': '<type __builtin__.int>',
                                      'longname': 'walkLength',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'wl'},
                       'wireframe': {'args': '<type __builtin__.bool>',
                                     'longname': 'wireframe',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'wf'}},
             'shortFlags': {'add': 'addSprings',
                            'all': 'allPoints',
                            'ct': 'count',
                            'd': 'damping',
                            'dPS': 'dampingPS',
                            'dmp': 'damp',
                            'e': 'edit',
                            'efw': 'endForceWeight',
                            'exc': 'exclusive',
                            'l': 'length',
                            'mm': 'minMax',
                            'mnd': 'minDistance',
                            'mxd': 'maxDistance',
                            'n': 'name',
                            'nd': 'noDuplicate',
                            'q': 'query',
                            'rPS': 'restLengthPS',
                            'rl': 'restLength',
                            's': 'stiffness',
                            'sPS': 'stiffnessPS',
                            'sfw': 'startForceWeight',
                            'str': 'strength',
                            'udp': 'useDampingPS',
                            'urp': 'useRestLengthPS',
                            'usp': 'useStiffnessPS',
                            'wf': 'wireframe',
                            'wl': 'walkLength'},
             'type': 'effects'},
  'squareSurface': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cch'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'continuityType1': {'args': '<type __builtin__.int>',
                                                  'longname': 'continuityType1',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ct1'},
                              'continuityType2': {'args': '<type __builtin__.int>',
                                                  'longname': 'continuityType2',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ct2'},
                              'continuityType3': {'args': '<type __builtin__.int>',
                                                  'longname': 'continuityType3',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ct3'},
                              'continuityType4': {'args': '<type __builtin__.int>',
                                                  'longname': 'continuityType4',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ct4'},
                              'curveFitCheckpoints': {'args': '<type __builtin__.int>',
                                                      'longname': 'curveFitCheckpoints',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cfc'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'endPointTolerance': {'args': '<type __builtin__.float>',
                                                    'longname': 'endPointTolerance',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ept'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nds'},
                              'object': {'args': '<type __builtin__.bool>',
                                         'longname': 'object',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'o'},
                              'polygon': {'args': '<type __builtin__.int>',
                                          'longname': 'polygon',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'po'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'rebuildCurve1': {'args': '<type __builtin__.bool>',
                                                'longname': 'rebuildCurve1',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'rc1'},
                              'rebuildCurve2': {'args': '<type __builtin__.bool>',
                                                'longname': 'rebuildCurve2',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'rc2'},
                              'rebuildCurve3': {'args': '<type __builtin__.bool>',
                                                'longname': 'rebuildCurve3',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'rc3'},
                              'rebuildCurve4': {'args': '<type __builtin__.bool>',
                                                'longname': 'rebuildCurve4',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'rc4'}},
                    'shortFlags': {'cch': 'caching',
                                   'cfc': 'curveFitCheckpoints',
                                   'ch': 'constructionHistory',
                                   'ct1': 'continuityType1',
                                   'ct2': 'continuityType2',
                                   'ct3': 'continuityType3',
                                   'ct4': 'continuityType4',
                                   'e': 'edit',
                                   'ept': 'endPointTolerance',
                                   'fzn': 'frozen',
                                   'n': 'name',
                                   'nds': 'nodeState',
                                   'o': 'object',
                                   'po': 'polygon',
                                   'q': 'query',
                                   'rc1': 'rebuildCurve1',
                                   'rc2': 'rebuildCurve2',
                                   'rc3': 'rebuildCurve3',
                                   'rc4': 'rebuildCurve4'},
                    'type': 'modeling'},
  'srtContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'history': {'args': '<type __builtin__.bool>',
                                       'longname': 'history',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ch'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'ch': 'history',
                                'e': 'edit',
                                'ex': 'exists',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'n': 'name',
                                'q': 'query'},
                 'type': 'context'},
  'stackTrace': {'flags': {'dump': {'args': '<type __builtin__.bool>',
                                    'longname': 'dump',
                                    'numArgs': 0,
                                    'shortname': 'd'},
                           'parameterCount': {'args': '<type __builtin__.int>',
                                              'longname': 'parameterCount',
                                              'numArgs': 1,
                                              'shortname': 'pc'},
                           'parameterType': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'parameterType',
                                             'numArgs': 2,
                                             'shortname': 'pt'},
                           'parameterValue': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'parameterValue',
                                              'numArgs': 2,
                                              'shortname': 'pv'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'state': {'args': '<type __builtin__.bool>',
                                     'longname': 'state',
                                     'numArgs': 1,
                                     'shortname': 's'}},
                 'shortFlags': {'d': 'dump',
                                'pc': 'parameterCount',
                                'pt': 'parameterType',
                                'pv': 'parameterValue',
                                'q': 'query',
                                's': 'state'},
                 'type': 'language'},
  'stitchSurface': {'flags': {'bias': {'args': '<type __builtin__.float>',
                                       'longname': 'bias',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'b'},
                              'caching': {'args': '<type __builtin__.bool>',
                                          'longname': 'caching',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cch'},
                              'cascade': {'args': '<type __builtin__.bool>',
                                          'longname': 'cascade',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'c'},
                              'constructionHistory': {'args': '<type __builtin__.bool>',
                                                      'longname': 'constructionHistory',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ch'},
                              'cvIthIndex': {'args': '<type __builtin__.int>',
                                             'longname': 'cvIthIndex',
                                             'modes': ['create',
                                                       'query',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'ci'},
                              'cvJthIndex': {'args': '<type __builtin__.int>',
                                             'longname': 'cvJthIndex',
                                             'modes': ['create',
                                                       'query',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'cj'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'fixBoundary': {'args': '<type __builtin__.bool>',
                                              'longname': 'fixBoundary',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fb'},
                              'frozen': {'args': '<type __builtin__.bool>',
                                         'longname': 'frozen',
                                         'numArgs': 1,
                                         'shortname': 'fzn'},
                              'keepG0Continuity': {'args': '<type __builtin__.bool>',
                                                   'longname': 'keepG0Continuity',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'kg0'},
                              'keepG1Continuity': {'args': '<type __builtin__.bool>',
                                                   'longname': 'keepG1Continuity',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'kg1'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'nodeState': {'args': '<type __builtin__.int>',
                                            'longname': 'nodeState',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nds'},
                              'numberOfSamples': {'args': '<type __builtin__.int>',
                                                  'longname': 'numberOfSamples',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ns'},
                              'object': {'args': '<type __builtin__.bool>',
                                         'longname': 'object',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'o'},
                              'parameterU': {'args': '<type __builtin__.float>',
                                             'longname': 'parameterU',
                                             'modes': ['create',
                                                       'query',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'u'},
                              'parameterV': {'args': '<type __builtin__.float>',
                                             'longname': 'parameterV',
                                             'modes': ['create',
                                                       'query',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'v'},
                              'positionalContinuity': {'args': '<type __builtin__.bool>',
                                                       'longname': 'positionalContinuity',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit',
                                                                 'multiuse'],
                                                       'numArgs': 1,
                                                       'shortname': 'pc'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                  'longname': 'replaceOriginal',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'rpo'},
                              'stepCount': {'args': '<type __builtin__.int>',
                                            'longname': 'stepCount',
                                            'modes': ['create',
                                                      'query',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'sc'},
                              'tangentialContinuity': {'args': '<type __builtin__.bool>',
                                                       'longname': 'tangentialContinuity',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit',
                                                                 'multiuse'],
                                                       'numArgs': 1,
                                                       'shortname': 'tc'},
                              'togglePointNormals': {'args': '<type __builtin__.bool>',
                                                     'longname': 'togglePointNormals',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'tpn'},
                              'togglePointPosition': {'args': '<type __builtin__.bool>',
                                                      'longname': 'togglePointPosition',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'tpp'},
                              'toggleTolerance': {'args': '<type __builtin__.bool>',
                                                  'longname': 'toggleTolerance',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'tt'},
                              'tolerance': {'args': '<type __builtin__.float>',
                                            'longname': 'tolerance',
                                            'modes': ['create',
                                                      'query',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'tol'},
                              'weight0': {'args': '<type __builtin__.float>',
                                          'longname': 'weight0',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'wt0'},
                              'weight1': {'args': '<type __builtin__.float>',
                                          'longname': 'weight1',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'wt1'}},
                    'shortFlags': {'b': 'bias',
                                   'c': 'cascade',
                                   'cch': 'caching',
                                   'ch': 'constructionHistory',
                                   'ci': 'cvIthIndex',
                                   'cj': 'cvJthIndex',
                                   'e': 'edit',
                                   'fb': 'fixBoundary',
                                   'fzn': 'frozen',
                                   'kg0': 'keepG0Continuity',
                                   'kg1': 'keepG1Continuity',
                                   'n': 'name',
                                   'nds': 'nodeState',
                                   'ns': 'numberOfSamples',
                                   'o': 'object',
                                   'pc': 'positionalContinuity',
                                   'q': 'query',
                                   'rpo': 'replaceOriginal',
                                   'sc': 'stepCount',
                                   'tc': 'tangentialContinuity',
                                   'tol': 'tolerance',
                                   'tpn': 'togglePointNormals',
                                   'tpp': 'togglePointPosition',
                                   'tt': 'toggleTolerance',
                                   'u': 'parameterU',
                                   'v': 'parameterV',
                                   'wt0': 'weight0',
                                   'wt1': 'weight1'},
                    'type': 'modeling'},
  'stitchSurfaceCtx': {'flags': {'activeNodes': {'args': '<type __builtin__.bool>',
                                                 'longname': 'activeNodes',
                                                 'numArgs': 0,
                                                 'shortname': 'anq'},
                                 'autoCreate': {'args': '<type __builtin__.bool>',
                                                'longname': 'autoCreate',
                                                'numArgs': 1,
                                                'shortname': 'ac'},
                                 'bias': {'args': '<type __builtin__.float>',
                                          'longname': 'bias',
                                          'numArgs': 1,
                                          'shortname': 'b'},
                                 'caching': {'args': '<type __builtin__.bool>',
                                             'longname': 'caching',
                                             'numArgs': 1,
                                             'shortname': 'cch'},
                                 'cascade': {'args': '<type __builtin__.bool>',
                                             'longname': 'cascade',
                                             'numArgs': 1,
                                             'shortname': 'c'},
                                 'constructionHistory': {'args': '<type __builtin__.bool>',
                                                         'longname': 'constructionHistory',
                                                         'numArgs': 1,
                                                         'shortname': 'ch'},
                                 'cvIthIndex': {'args': '<type __builtin__.int>',
                                                'longname': 'cvIthIndex',
                                                'modes': ['multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'ci'},
                                 'cvJthIndex': {'args': '<type __builtin__.int>',
                                                'longname': 'cvJthIndex',
                                                'modes': ['multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'cj'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'fixBoundary': {'args': '<type __builtin__.bool>',
                                                 'longname': 'fixBoundary',
                                                 'numArgs': 1,
                                                 'shortname': 'fb'},
                                 'frozen': {'args': '<type __builtin__.bool>',
                                            'longname': 'frozen',
                                            'numArgs': 1,
                                            'shortname': 'fzn'},
                                 'image1': {'args': '<type __builtin__.str>',
                                            'longname': 'image1',
                                            'numArgs': 1,
                                            'shortname': 'i1'},
                                 'image2': {'args': '<type __builtin__.str>',
                                            'longname': 'image2',
                                            'numArgs': 1,
                                            'shortname': 'i2'},
                                 'image3': {'args': '<type __builtin__.str>',
                                            'longname': 'image3',
                                            'numArgs': 1,
                                            'shortname': 'i3'},
                                 'immediate': {'args': '<type __builtin__.bool>',
                                               'longname': 'immediate',
                                               'numArgs': 0,
                                               'shortname': 'im'},
                                 'keepG0Continuity': {'args': '<type __builtin__.bool>',
                                                      'longname': 'keepG0Continuity',
                                                      'numArgs': 1,
                                                      'shortname': 'kg0'},
                                 'keepG1Continuity': {'args': '<type __builtin__.bool>',
                                                      'longname': 'keepG1Continuity',
                                                      'numArgs': 1,
                                                      'shortname': 'kg1'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'nodeState': {'args': '<type __builtin__.int>',
                                               'longname': 'nodeState',
                                               'numArgs': 1,
                                               'shortname': 'nds'},
                                 'numberOfSamples': {'args': '<type __builtin__.int>',
                                                     'longname': 'numberOfSamples',
                                                     'numArgs': 1,
                                                     'shortname': 'ns'},
                                 'object': {'args': '<type __builtin__.bool>',
                                            'longname': 'object',
                                            'numArgs': 1,
                                            'shortname': 'o'},
                                 'parameterU': {'args': '<type __builtin__.float>',
                                                'longname': 'parameterU',
                                                'modes': ['multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'u'},
                                 'parameterV': {'args': '<type __builtin__.float>',
                                                'longname': 'parameterV',
                                                'modes': ['multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'v'},
                                 'positionalContinuity': {'args': '<type __builtin__.bool>',
                                                          'longname': 'positionalContinuity',
                                                          'modes': ['multiuse'],
                                                          'numArgs': 1,
                                                          'shortname': 'pc'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                     'longname': 'replaceOriginal',
                                                     'numArgs': 1,
                                                     'shortname': 'rpo'},
                                 'reset': {'args': '<type __builtin__.bool>',
                                           'longname': 'reset',
                                           'numArgs': 0,
                                           'shortname': 'rs'},
                                 'stepCount': {'args': '<type __builtin__.int>',
                                               'longname': 'stepCount',
                                               'modes': ['multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'sc'},
                                 'tangentialContinuity': {'args': '<type __builtin__.bool>',
                                                          'longname': 'tangentialContinuity',
                                                          'modes': ['multiuse'],
                                                          'numArgs': 1,
                                                          'shortname': 'tc'},
                                 'togglePointNormals': {'args': '<type __builtin__.bool>',
                                                        'longname': 'togglePointNormals',
                                                        'numArgs': 1,
                                                        'shortname': 'tpn'},
                                 'togglePointPosition': {'args': '<type __builtin__.bool>',
                                                         'longname': 'togglePointPosition',
                                                         'numArgs': 1,
                                                         'shortname': 'tpp'},
                                 'toggleTolerance': {'args': '<type __builtin__.bool>',
                                                     'longname': 'toggleTolerance',
                                                     'modes': ['multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'tt'},
                                 'tolerance': {'args': '<type __builtin__.float>',
                                               'longname': 'tolerance',
                                               'modes': ['multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'tol'},
                                 'toolNode': {'args': '<type __builtin__.bool>',
                                              'longname': 'toolNode',
                                              'numArgs': 0,
                                              'shortname': 'tnq'},
                                 'weight0': {'args': '<type __builtin__.float>',
                                             'longname': 'weight0',
                                             'numArgs': 1,
                                             'shortname': 'wt0'},
                                 'weight1': {'args': '<type __builtin__.float>',
                                             'longname': 'weight1',
                                             'numArgs': 1,
                                             'shortname': 'wt1'}},
                       'shortFlags': {'ac': 'autoCreate',
                                      'anq': 'activeNodes',
                                      'b': 'bias',
                                      'c': 'cascade',
                                      'cch': 'caching',
                                      'ch': 'constructionHistory',
                                      'ci': 'cvIthIndex',
                                      'cj': 'cvJthIndex',
                                      'e': 'edit',
                                      'ex': 'exists',
                                      'fb': 'fixBoundary',
                                      'fzn': 'frozen',
                                      'i1': 'image1',
                                      'i2': 'image2',
                                      'i3': 'image3',
                                      'im': 'immediate',
                                      'kg0': 'keepG0Continuity',
                                      'kg1': 'keepG1Continuity',
                                      'n': 'name',
                                      'nds': 'nodeState',
                                      'ns': 'numberOfSamples',
                                      'o': 'object',
                                      'pc': 'positionalContinuity',
                                      'q': 'query',
                                      'rpo': 'replaceOriginal',
                                      'rs': 'reset',
                                      'sc': 'stepCount',
                                      'tc': 'tangentialContinuity',
                                      'tnq': 'toolNode',
                                      'tol': 'tolerance',
                                      'tpn': 'togglePointNormals',
                                      'tpp': 'togglePointPosition',
                                      'tt': 'toggleTolerance',
                                      'u': 'parameterU',
                                      'v': 'parameterV',
                                      'wt0': 'weight0',
                                      'wt1': 'weight1'},
                       'type': 'context'},
  'stitchSurfacePoints': {'flags': {'bias': {'args': '<type __builtin__.float>',
                                             'longname': 'bias',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'b'},
                                    'caching': {'args': '<type __builtin__.bool>',
                                                'longname': 'caching',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cch'},
                                    'cascade': {'args': '<type __builtin__.bool>',
                                                'longname': 'cascade',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'c'},
                                    'constructionHistory': {'args': '<type __builtin__.bool>',
                                                            'longname': 'constructionHistory',
                                                            'modes': ['create'],
                                                            'numArgs': 1,
                                                            'shortname': 'ch'},
                                    'cvIthIndex': {'args': '<type __builtin__.int>',
                                                   'longname': 'cvIthIndex',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'ci'},
                                    'cvJthIndex': {'args': '<type __builtin__.int>',
                                                   'longname': 'cvJthIndex',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'cj'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'equalWeight': {'args': '<type __builtin__.bool>',
                                                    'longname': 'equalWeight',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ewt'},
                                    'fixBoundary': {'args': '<type __builtin__.bool>',
                                                    'longname': 'fixBoundary',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'fb'},
                                    'frozen': {'args': '<type __builtin__.bool>',
                                               'longname': 'frozen',
                                               'numArgs': 1,
                                               'shortname': 'fzn'},
                                    'keepG0Continuity': {'args': '<type __builtin__.bool>',
                                                         'longname': 'keepG0Continuity',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'kg0'},
                                    'keepG1Continuity': {'args': '<type __builtin__.bool>',
                                                         'longname': 'keepG1Continuity',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'kg1'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'nodeState': {'args': '<type __builtin__.int>',
                                                  'longname': 'nodeState',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'nds'},
                                    'object': {'args': '<type __builtin__.bool>',
                                               'longname': 'object',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'o'},
                                    'parameterU': {'args': '<type __builtin__.float>',
                                                   'longname': 'parameterU',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'u'},
                                    'parameterV': {'args': '<type __builtin__.float>',
                                                   'longname': 'parameterV',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'v'},
                                    'positionalContinuity': {'args': '<type __builtin__.bool>',
                                                             'longname': 'positionalContinuity',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit',
                                                                       'multiuse'],
                                                             'numArgs': 1,
                                                             'shortname': 'pc'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'replaceOriginal': {'args': '<type __builtin__.bool>',
                                                        'longname': 'replaceOriginal',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'rpo'},
                                    'stepCount': {'args': '<type __builtin__.int>',
                                                  'longname': 'stepCount',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'sc'},
                                    'tangentialContinuity': {'args': '<type __builtin__.bool>',
                                                             'longname': 'tangentialContinuity',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit',
                                                                       'multiuse'],
                                                             'numArgs': 1,
                                                             'shortname': 'tc'},
                                    'togglePointNormals': {'args': '<type __builtin__.bool>',
                                                           'longname': 'togglePointNormals',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'tpn'},
                                    'togglePointPosition': {'args': '<type __builtin__.bool>',
                                                            'longname': 'togglePointPosition',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'tpp'},
                                    'toggleTolerance': {'args': '<type __builtin__.bool>',
                                                        'longname': 'toggleTolerance',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit',
                                                                  'multiuse'],
                                                        'numArgs': 1,
                                                        'shortname': 'tt'},
                                    'tolerance': {'args': '<type __builtin__.float>',
                                                  'longname': 'tolerance',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'tol'}},
                          'shortFlags': {'b': 'bias',
                                         'c': 'cascade',
                                         'cch': 'caching',
                                         'ch': 'constructionHistory',
                                         'ci': 'cvIthIndex',
                                         'cj': 'cvJthIndex',
                                         'e': 'edit',
                                         'ewt': 'equalWeight',
                                         'fb': 'fixBoundary',
                                         'fzn': 'frozen',
                                         'kg0': 'keepG0Continuity',
                                         'kg1': 'keepG1Continuity',
                                         'n': 'name',
                                         'nds': 'nodeState',
                                         'o': 'object',
                                         'pc': 'positionalContinuity',
                                         'q': 'query',
                                         'rpo': 'replaceOriginal',
                                         'sc': 'stepCount',
                                         'tc': 'tangentialContinuity',
                                         'tol': 'tolerance',
                                         'tpn': 'togglePointNormals',
                                         'tpp': 'togglePointPosition',
                                         'tt': 'toggleTolerance',
                                         'u': 'parameterU',
                                         'v': 'parameterV'},
                          'type': 'modeling'},
  'stringArrayIntersector': {'flags': {'allowDuplicates': {'args': '<type __builtin__.bool>',
                                                           'longname': 'allowDuplicates',
                                                           'modes': ['create'],
                                                           'numArgs': 1,
                                                           'shortname': 'ad'},
                                       'defineTemplate': {'args': '<type __builtin__.str>',
                                                          'longname': 'defineTemplate',
                                                          'modes': ['create'],
                                                          'numArgs': 1,
                                                          'shortname': 'dt'},
                                       'edit': {'args': '<type __builtin__.bool>',
                                                'longname': 'edit',
                                                'numArgs': 0,
                                                'shortname': 'e'},
                                       'exists': {'args': '<type __builtin__.bool>',
                                                  'longname': 'exists',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'ex'},
                                       'intersect': {'args': '<type __builtin__.str>',
                                                     'longname': 'intersect',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'i'},
                                       'query': {'args': '<type __builtin__.bool>',
                                                 'longname': 'query',
                                                 'numArgs': 0,
                                                 'shortname': 'q'},
                                       'reset': {'args': '<type __builtin__.bool>',
                                                 'longname': 'reset',
                                                 'modes': ['edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'r'},
                                       'useTemplate': {'args': '<type __builtin__.str>',
                                                       'longname': 'useTemplate',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ut'}},
                             'shortFlags': {'ad': 'allowDuplicates',
                                            'dt': 'defineTemplate',
                                            'e': 'edit',
                                            'ex': 'exists',
                                            'i': 'intersect',
                                            'q': 'query',
                                            'r': 'reset',
                                            'ut': 'useTemplate'},
                             'type': 'language'},
  'stroke': {'flags': {'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'pressure': {'args': '<type __builtin__.bool>',
                                    'longname': 'pressure',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'pr'},
                       'seed': {'args': '<type __builtin__.int>',
                                'longname': 'seed',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 's'}},
             'shortFlags': {'n': 'name', 'pr': 'pressure', 's': 'seed'},
             'type': 'effects'},
  'subdAutoProjection': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                               'longname': 'caching',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cch'},
                                   'constructionHistory': {'args': '<type __builtin__.bool>',
                                                           'longname': 'constructionHistory',
                                                           'modes': ['create'],
                                                           'numArgs': 1,
                                                           'shortname': 'ch'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'frozen': {'args': '<type __builtin__.bool>',
                                              'longname': 'frozen',
                                              'numArgs': 1,
                                              'shortname': 'fzn'},
                                   'layout': {'args': '<type __builtin__.int>',
                                              'longname': 'layout',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'l'},
                                   'layoutMethod': {'args': '<type __builtin__.int>',
                                                    'longname': 'layoutMethod',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'lm'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'nodeState': {'args': '<type __builtin__.int>',
                                                 'longname': 'nodeState',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'nds'},
                                   'optimize': {'args': '<type __builtin__.int>',
                                                'longname': 'optimize',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'o'},
                                   'percentageSpace': {'args': '<type __builtin__.float>',
                                                       'longname': 'percentageSpace',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ps'},
                                   'planes': {'args': '<type __builtin__.int>',
                                              'longname': 'planes',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'p'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'scale': {'args': '<type __builtin__.int>',
                                             'longname': 'scale',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sc'},
                                   'skipIntersect': {'args': '<type __builtin__.bool>',
                                                     'longname': 'skipIntersect',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'si'},
                                   'worldSpace': {'args': '<type __builtin__.bool>',
                                                  'longname': 'worldSpace',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ws'}},
                         'shortFlags': {'cch': 'caching',
                                        'ch': 'constructionHistory',
                                        'e': 'edit',
                                        'fzn': 'frozen',
                                        'l': 'layout',
                                        'lm': 'layoutMethod',
                                        'n': 'name',
                                        'nds': 'nodeState',
                                        'o': 'optimize',
                                        'p': 'planes',
                                        'ps': 'percentageSpace',
                                        'q': 'query',
                                        'sc': 'scale',
                                        'si': 'skipIntersect',
                                        'ws': 'worldSpace'},
                         'type': 'modeling'},
  'subdCleanTopology': {'flags': {}, 'shortFlags': {}, 'type': 'modeling'},
  'subdCollapse': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'level': {'args': '<type __builtin__.int>',
                                       'longname': 'level',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'l'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'object': {'args': '<type __builtin__.bool>',
                                        'longname': 'object',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'o'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'e': 'edit',
                                  'fzn': 'frozen',
                                  'l': 'level',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'o': 'object',
                                  'q': 'query'},
                   'type': 'modeling'},
  'subdDisplayMode': {'flags': {'hideFaceGadgets': {'args': '<type __builtin__.bool>',
                                                    'longname': 'hideFaceGadgets',
                                                    'numArgs': 1,
                                                    'shortname': 'hfg'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'showComponentsAsNumerals': {'args': '<type __builtin__.bool>',
                                                             'longname': 'showComponentsAsNumerals',
                                                             'numArgs': 1,
                                                             'shortname': 'scn'},
                                'showFaceRegions': {'args': '<type __builtin__.bool>',
                                                    'longname': 'showFaceRegions',
                                                    'numArgs': 1,
                                                    'shortname': 'sfr'},
                                'showVisualEdgeVertices': {'args': '<type __builtin__.bool>',
                                                           'longname': 'showVisualEdgeVertices',
                                                           'numArgs': 1,
                                                           'shortname': 'svv'},
                                'showVisualEdges': {'args': '<type __builtin__.bool>',
                                                    'longname': 'showVisualEdges',
                                                    'numArgs': 1,
                                                    'shortname': 'sve'},
                                'subdivEdgeMask': {'args': '<type __builtin__.int>',
                                                   'longname': 'subdivEdgeMask',
                                                   'numArgs': 1,
                                                   'shortname': 'sem'}},
                      'shortFlags': {'hfg': 'hideFaceGadgets',
                                     'q': 'query',
                                     'scn': 'showComponentsAsNumerals',
                                     'sem': 'subdivEdgeMask',
                                     'sfr': 'showFaceRegions',
                                     'sve': 'showVisualEdges',
                                     'svv': 'showVisualEdgeVertices'},
                      'type': 'other'},
  'subdDuplicateAndConnect': {'flags': {},
                              'shortFlags': {},
                              'type': 'modeling'},
  'subdEditUV': {'flags': {'angle': {'args': '<type __builtin__.float>',
                                     'longname': 'angle',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'a'},
                           'pivotU': {'args': '<type __builtin__.float>',
                                      'longname': 'pivotU',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'pu'},
                           'pivotV': {'args': '<type __builtin__.float>',
                                      'longname': 'pivotV',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'pv'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'relative': {'args': '<type __builtin__.bool>',
                                        'longname': 'relative',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'r'},
                           'rotateRatio': {'args': '<type __builtin__.float>',
                                           'longname': 'rotateRatio',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'rr'},
                           'rotation': {'args': '<type __builtin__.bool>',
                                        'longname': 'rotation',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'rot'},
                           'scale': {'args': '<type __builtin__.bool>',
                                     'longname': 'scale',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 's'},
                           'scaleU': {'args': '<type __builtin__.float>',
                                      'longname': 'scaleU',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'su'},
                           'scaleV': {'args': '<type __builtin__.float>',
                                      'longname': 'scaleV',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'sv'},
                           'uValue': {'args': '<type __builtin__.float>',
                                      'longname': 'uValue',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'u'},
                           'uvSetName': {'args': '<type __builtin__.str>',
                                         'longname': 'uvSetName',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'uvs'},
                           'vValue': {'args': '<type __builtin__.float>',
                                      'longname': 'vValue',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'v'}},
                 'shortFlags': {'a': 'angle',
                                'pu': 'pivotU',
                                'pv': 'pivotV',
                                'q': 'query',
                                'r': 'relative',
                                'rot': 'rotation',
                                'rr': 'rotateRatio',
                                's': 'scale',
                                'su': 'scaleU',
                                'sv': 'scaleV',
                                'u': 'uValue',
                                'uvs': 'uvSetName',
                                'v': 'vValue'},
                 'type': 'modeling'},
  'subdLayoutUV': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                         'longname': 'caching',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'cch'},
                             'constructionHistory': {'args': '<type __builtin__.bool>',
                                                     'longname': 'constructionHistory',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'ch'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'flipReversed': {'args': '<type __builtin__.bool>',
                                              'longname': 'flipReversed',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fr'},
                             'frozen': {'args': '<type __builtin__.bool>',
                                        'longname': 'frozen',
                                        'numArgs': 1,
                                        'shortname': 'fzn'},
                             'layout': {'args': '<type __builtin__.int>',
                                        'longname': 'layout',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'l'},
                             'layoutMethod': {'args': '<type __builtin__.int>',
                                              'longname': 'layoutMethod',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'lm'},
                             'name': {'args': '<type __builtin__.str>',
                                      'longname': 'name',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'n'},
                             'nodeState': {'args': '<type __builtin__.int>',
                                           'longname': 'nodeState',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nds'},
                             'percentageSpace': {'args': '<type __builtin__.float>',
                                                 'longname': 'percentageSpace',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ps'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'rotateForBestFit': {'args': '<type __builtin__.int>',
                                                  'longname': 'rotateForBestFit',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'rbf'},
                             'scale': {'args': '<type __builtin__.int>',
                                       'longname': 'scale',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sc'},
                             'separate': {'args': '<type __builtin__.int>',
                                          'longname': 'separate',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'se'},
                             'worldSpace': {'args': '<type __builtin__.bool>',
                                            'longname': 'worldSpace',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ws'}},
                   'shortFlags': {'cch': 'caching',
                                  'ch': 'constructionHistory',
                                  'e': 'edit',
                                  'fr': 'flipReversed',
                                  'fzn': 'frozen',
                                  'l': 'layout',
                                  'lm': 'layoutMethod',
                                  'n': 'name',
                                  'nds': 'nodeState',
                                  'ps': 'percentageSpace',
                                  'q': 'query',
                                  'rbf': 'rotateForBestFit',
                                  'sc': 'scale',
                                  'se': 'separate',
                                  'ws': 'worldSpace'},
                   'type': 'modeling'},
  'subdListComponentConversion': {'flags': {'border': {'args': '<type __builtin__.bool>',
                                                       'longname': 'border',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'bo'},
                                            'fromEdge': {'args': '<type __builtin__.bool>',
                                                         'longname': 'fromEdge',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'fe'},
                                            'fromFace': {'args': '<type __builtin__.bool>',
                                                         'longname': 'fromFace',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'ff'},
                                            'fromUV': {'args': '<type __builtin__.bool>',
                                                       'longname': 'fromUV',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'fuv'},
                                            'fromVertex': {'args': '<type __builtin__.bool>',
                                                           'longname': 'fromVertex',
                                                           'modes': ['create'],
                                                           'numArgs': 0,
                                                           'shortname': 'fv'},
                                            'internal': {'args': '<type __builtin__.bool>',
                                                         'longname': 'internal',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'internal'},
                                            'toEdge': {'args': '<type __builtin__.bool>',
                                                       'longname': 'toEdge',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'te'},
                                            'toFace': {'args': '<type __builtin__.bool>',
                                                       'longname': 'toFace',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'tf'},
                                            'toUV': {'args': '<type __builtin__.bool>',
                                                     'longname': 'toUV',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'tuv'},
                                            'toVertex': {'args': '<type __builtin__.bool>',
                                                         'longname': 'toVertex',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'tv'},
                                            'uvShell': {'args': '<type __builtin__.bool>',
                                                        'longname': 'uvShell',
                                                        'modes': ['create'],
                                                        'numArgs': 0,
                                                        'shortname': 'uvs'},
                                            'uvShellBorder': {'args': '<type __builtin__.bool>',
                                                              'longname': 'uvShellBorder',
                                                              'modes': ['create'],
                                                              'numArgs': 0,
                                                              'shortname': 'uvb'}},
                                  'removedFlags': {'in': 'internal'},
                                  'shortFlags': {'bo': 'border',
                                                 'fe': 'fromEdge',
                                                 'ff': 'fromFace',
                                                 'fuv': 'fromUV',
                                                 'fv': 'fromVertex',
                                                 'internal': 'internal',
                                                 'te': 'toEdge',
                                                 'tf': 'toFace',
                                                 'tuv': 'toUV',
                                                 'tv': 'toVertex',
                                                 'uvb': 'uvShellBorder',
                                                 'uvs': 'uvShell'},
                                  'type': 'modeling'},
  'subdMapCut': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'cch': 'caching',
                                'ch': 'constructionHistory',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'n': 'name',
                                'nds': 'nodeState',
                                'q': 'query'},
                 'type': 'modeling'},
  'subdMapSewMove': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                           'longname': 'caching',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'cch'},
                               'constructionHistory': {'args': '<type __builtin__.bool>',
                                                       'longname': 'constructionHistory',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'ch'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'frozen': {'args': '<type __builtin__.bool>',
                                          'longname': 'frozen',
                                          'numArgs': 1,
                                          'shortname': 'fzn'},
                               'limitPieceSize': {'args': '<type __builtin__.bool>',
                                                  'longname': 'limitPieceSize',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'lps'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'nodeState': {'args': '<type __builtin__.int>',
                                             'longname': 'nodeState',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nds'},
                               'numberFaces': {'args': '<type __builtin__.int>',
                                               'longname': 'numberFaces',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'nf'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'worldSpace': {'args': '<type __builtin__.bool>',
                                              'longname': 'worldSpace',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ws'}},
                     'shortFlags': {'cch': 'caching',
                                    'ch': 'constructionHistory',
                                    'e': 'edit',
                                    'fzn': 'frozen',
                                    'lps': 'limitPieceSize',
                                    'n': 'name',
                                    'nds': 'nodeState',
                                    'nf': 'numberFaces',
                                    'q': 'query',
                                    'ws': 'worldSpace'},
                     'type': 'modeling'},
  'subdMatchTopology': {'flags': {'frontOfChain': {'args': '<type __builtin__.bool>',
                                                   'longname': 'frontOfChain',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'foc'}},
                        'shortFlags': {'foc': 'frontOfChain'},
                        'type': 'modeling'},
  'subdMirror': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'object': {'args': '<type __builtin__.bool>',
                                      'longname': 'object',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'xMirror': {'args': '<type __builtin__.bool>',
                                       'longname': 'xMirror',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'xm'},
                           'yMirror': {'args': '<type __builtin__.bool>',
                                       'longname': 'yMirror',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ym'},
                           'zMirror': {'args': '<type __builtin__.bool>',
                                       'longname': 'zMirror',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'zm'}},
                 'shortFlags': {'cch': 'caching',
                                'ch': 'constructionHistory',
                                'e': 'edit',
                                'fzn': 'frozen',
                                'n': 'name',
                                'nds': 'nodeState',
                                'o': 'object',
                                'q': 'query',
                                'xm': 'xMirror',
                                'ym': 'yMirror',
                                'zm': 'zMirror'},
                 'type': 'modeling'},
  'subdPlanarProjection': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                                 'longname': 'caching',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cch'},
                                     'constructionHistory': {'args': '<type __builtin__.bool>',
                                                             'longname': 'constructionHistory',
                                                             'modes': ['create'],
                                                             'numArgs': 1,
                                                             'shortname': 'ch'},
                                     'createNewMap': {'args': '<type __builtin__.bool>',
                                                      'longname': 'createNewMap',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'cm'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'frozen': {'args': '<type __builtin__.bool>',
                                                'longname': 'frozen',
                                                'numArgs': 1,
                                                'shortname': 'fzn'},
                                     'imageCenter': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'imageCenter',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'ic2'},
                                     'imageCenterX': {'args': '<type __builtin__.float>',
                                                      'longname': 'imageCenterX',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'icx'},
                                     'imageCenterY': {'args': '<type __builtin__.float>',
                                                      'longname': 'imageCenterY',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'icy'},
                                     'imageScale': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'imageScale',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 2,
                                                    'shortname': 'is2'},
                                     'imageScaleU': {'args': '<type __builtin__.float>',
                                                     'longname': 'imageScaleU',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'isu'},
                                     'imageScaleV': {'args': '<type __builtin__.float>',
                                                     'longname': 'imageScaleV',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'isv'},
                                     'insertBeforeDeformers': {'args': '<type __builtin__.bool>',
                                                               'longname': 'insertBeforeDeformers',
                                                               'modes': ['create'],
                                                               'numArgs': 1,
                                                               'shortname': 'ibd'},
                                     'keepImageRatio': {'args': '<type __builtin__.bool>',
                                                        'longname': 'keepImageRatio',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'kir'},
                                     'mapDirection': {'args': '<type __builtin__.str>',
                                                      'longname': 'mapDirection',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'md'},
                                     'name': {'args': '<type __builtin__.str>',
                                              'longname': 'name',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'n'},
                                     'nodeState': {'args': '<type __builtin__.int>',
                                                   'longname': 'nodeState',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'nds'},
                                     'projectionCenter': {'args': ['<type __builtin__.float>',
                                                                   '<type __builtin__.float>',
                                                                   '<type __builtin__.float>'],
                                                          'longname': 'projectionCenter',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 3,
                                                          'shortname': 'pc'},
                                     'projectionCenterX': {'args': '<type __builtin__.float>',
                                                           'longname': 'projectionCenterX',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'pcx'},
                                     'projectionCenterY': {'args': '<type __builtin__.float>',
                                                           'longname': 'projectionCenterY',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'pcy'},
                                     'projectionCenterZ': {'args': '<type __builtin__.float>',
                                                           'longname': 'projectionCenterZ',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'pcz'},
                                     'projectionHeight': {'args': '<type __builtin__.float>',
                                                          'longname': 'projectionHeight',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ph'},
                                     'projectionScale': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'projectionScale',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 2,
                                                         'shortname': 'ps'},
                                     'projectionWidth': {'args': '<type __builtin__.float>',
                                                         'longname': 'projectionWidth',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'pw'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'rotate': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'rotate',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'ro'},
                                     'rotateX': {'args': '<type __builtin__.float>',
                                                 'longname': 'rotateX',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rx'},
                                     'rotateY': {'args': '<type __builtin__.float>',
                                                 'longname': 'rotateY',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ry'},
                                     'rotateZ': {'args': '<type __builtin__.float>',
                                                 'longname': 'rotateZ',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rz'},
                                     'rotationAngle': {'args': '<type __builtin__.float>',
                                                       'longname': 'rotationAngle',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ra'},
                                     'smartFit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'smartFit',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'sf'},
                                     'worldSpace': {'args': '<type __builtin__.bool>',
                                                    'longname': 'worldSpace',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ws'}},
                           'shortFlags': {'cch': 'caching',
                                          'ch': 'constructionHistory',
                                          'cm': 'createNewMap',
                                          'e': 'edit',
                                          'fzn': 'frozen',
                                          'ibd': 'insertBeforeDeformers',
                                          'ic2': 'imageCenter',
                                          'icx': 'imageCenterX',
                                          'icy': 'imageCenterY',
                                          'is2': 'imageScale',
                                          'isu': 'imageScaleU',
                                          'isv': 'imageScaleV',
                                          'kir': 'keepImageRatio',
                                          'md': 'mapDirection',
                                          'n': 'name',
                                          'nds': 'nodeState',
                                          'pc': 'projectionCenter',
                                          'pcx': 'projectionCenterX',
                                          'pcy': 'projectionCenterY',
                                          'pcz': 'projectionCenterZ',
                                          'ph': 'projectionHeight',
                                          'ps': 'projectionScale',
                                          'pw': 'projectionWidth',
                                          'q': 'query',
                                          'ra': 'rotationAngle',
                                          'ro': 'rotate',
                                          'rx': 'rotateX',
                                          'ry': 'rotateY',
                                          'rz': 'rotateZ',
                                          'sf': 'smartFit',
                                          'ws': 'worldSpace'},
                           'type': 'modeling'},
  'subdToBlind': {'flags': {'absolutePosition': {'args': '<type __builtin__.bool>',
                                                 'longname': 'absolutePosition',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ap'},
                            'includeCreases': {'args': '<type __builtin__.bool>',
                                               'longname': 'includeCreases',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ic'},
                            'includeZeroOffsets': {'args': '<type __builtin__.bool>',
                                                   'longname': 'includeZeroOffsets',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'izo'}},
                  'shortFlags': {'ap': 'absolutePosition',
                                 'ic': 'includeCreases',
                                 'izo': 'includeZeroOffsets'},
                  'type': 'modeling'},
  'subdToNurbs': {'flags': {'addUnderTransform': {'args': '<type __builtin__.bool>',
                                                  'longname': 'addUnderTransform',
                                                  'numArgs': 1,
                                                  'shortname': 'aut'},
                            'applyMatrixToResult': {'args': '<type __builtin__.bool>',
                                                    'longname': 'applyMatrixToResult',
                                                    'numArgs': 1,
                                                    'shortname': 'amr'},
                            'caching': {'args': '<type __builtin__.bool>',
                                        'longname': 'caching',
                                        'numArgs': 1,
                                        'shortname': 'cch'},
                            'constructionHistory': {'args': '<type __builtin__.bool>',
                                                    'longname': 'constructionHistory',
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'frozen': {'args': '<type __builtin__.bool>',
                                       'longname': 'frozen',
                                       'numArgs': 1,
                                       'shortname': 'fzn'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'nodeState': {'args': '<type __builtin__.int>',
                                          'longname': 'nodeState',
                                          'numArgs': 1,
                                          'shortname': 'nds'},
                            'object': {'args': '<type __builtin__.bool>',
                                       'longname': 'object',
                                       'numArgs': 1,
                                       'shortname': 'o'},
                            'outputType': {'args': '<type __builtin__.int>',
                                           'longname': 'outputType',
                                           'numArgs': 1,
                                           'shortname': 'ot'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'}},
                  'shortFlags': {'amr': 'applyMatrixToResult',
                                 'aut': 'addUnderTransform',
                                 'cch': 'caching',
                                 'ch': 'constructionHistory',
                                 'e': 'edit',
                                 'fzn': 'frozen',
                                 'n': 'name',
                                 'nds': 'nodeState',
                                 'o': 'object',
                                 'ot': 'outputType',
                                 'q': 'query'},
                  'type': 'other'},
  'subdToPoly': {'flags': {'addUnderTransform': {'args': '<type __builtin__.bool>',
                                                 'longname': 'addUnderTransform',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'aut'},
                           'applyMatrixToResult': {'args': '<type __builtin__.bool>',
                                                   'longname': 'applyMatrixToResult',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'amr'},
                           'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'connectShaders': {'args': '<type __builtin__.bool>',
                                              'longname': 'connectShaders',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'cs'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'copyUVTopology': {'args': '<type __builtin__.bool>',
                                              'longname': 'copyUVTopology',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cut'},
                           'depth': {'args': '<type __builtin__.int>',
                                     'longname': 'depth',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'd'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'extractPointPosition': {'args': '<type __builtin__.bool>',
                                                    'longname': 'extractPointPosition',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'epp'},
                           'format': {'args': '<type __builtin__.int>',
                                      'longname': 'format',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'f'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'inSubdCVId': {'args': ['<type __builtin__.int>',
                                                   '<type __builtin__.int>'],
                                          'longname': 'inSubdCVId',
                                          'modes': ['create',
                                                    'query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'inSubdCVId'},
                           'inSubdCVIdLeft': {'args': '<type __builtin__.int>',
                                              'longname': 'inSubdCVIdLeft',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'isl'},
                           'inSubdCVIdRight': {'args': '<type __builtin__.int>',
                                               'longname': 'inSubdCVIdRight',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'isr'},
                           'maxPolys': {'args': '<type __builtin__.int>',
                                        'longname': 'maxPolys',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'mp'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'object': {'args': '<type __builtin__.bool>',
                                      'longname': 'object',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'outSubdCVId': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'outSubdCVId',
                                           'modes': ['create',
                                                     'query',
                                                     'edit',
                                                     'multiuse'],
                                           'numArgs': 2,
                                           'shortname': 'os'},
                           'outSubdCVIdLeft': {'args': '<type __builtin__.int>',
                                               'longname': 'outSubdCVIdLeft',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'osl'},
                           'outSubdCVIdRight': {'args': '<type __builtin__.int>',
                                                'longname': 'outSubdCVIdRight',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'osr'},
                           'outv': {'args': '<type __builtin__.int>',
                                    'longname': 'outv',
                                    'modes': ['create',
                                              'query',
                                              'edit',
                                              'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'ov'},
                           'preserveVertexOrdering': {'args': '<type __builtin__.bool>',
                                                      'longname': 'preserveVertexOrdering',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'pvo'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'sampleCount': {'args': '<type __builtin__.int>',
                                           'longname': 'sampleCount',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sc'},
                           'shareUVs': {'args': '<type __builtin__.bool>',
                                        'longname': 'shareUVs',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'suv'},
                           'subdNormals': {'args': '<type __builtin__.bool>',
                                           'longname': 'subdNormals',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'un'}},
                 'removedFlags': {'is': 'inSubdCVId'},
                 'shortFlags': {'amr': 'applyMatrixToResult',
                                'aut': 'addUnderTransform',
                                'cch': 'caching',
                                'ch': 'constructionHistory',
                                'cs': 'connectShaders',
                                'cut': 'copyUVTopology',
                                'd': 'depth',
                                'e': 'edit',
                                'epp': 'extractPointPosition',
                                'f': 'format',
                                'fzn': 'frozen',
                                'inSubdCVId': 'inSubdCVId',
                                'isl': 'inSubdCVIdLeft',
                                'isr': 'inSubdCVIdRight',
                                'mp': 'maxPolys',
                                'n': 'name',
                                'nds': 'nodeState',
                                'o': 'object',
                                'os': 'outSubdCVId',
                                'osl': 'outSubdCVIdLeft',
                                'osr': 'outSubdCVIdRight',
                                'ov': 'outv',
                                'pvo': 'preserveVertexOrdering',
                                'q': 'query',
                                'sc': 'sampleCount',
                                'suv': 'shareUVs',
                                'un': 'subdNormals'},
                 'type': 'modeling'},
  'subdTransferUVsToCache': {'flags': {},
                             'shortFlags': {},
                             'type': 'modeling'},
  'subdiv': {'flags': {'currentLevel': {'args': '<type __builtin__.bool>',
                                        'longname': 'currentLevel',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'cl'},
                       'currentSubdLevel': {'args': '<type __builtin__.bool>',
                                            'longname': 'currentSubdLevel',
                                            'modes': ['create', 'query'],
                                            'numArgs': 0,
                                            'shortname': 'csl'},
                       'deepestLevel': {'args': '<type __builtin__.int>',
                                        'longname': 'deepestLevel',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'dl'},
                       'displayLoad': {'args': '<type __builtin__.bool>',
                                       'longname': 'displayLoad',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'dsl'},
                       'edgeStats': {'args': '<type __builtin__.bool>',
                                     'longname': 'edgeStats',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 'est'},
                       'faceStats': {'args': '<type __builtin__.bool>',
                                     'longname': 'faceStats',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 'fst'},
                       'maxPossibleLevel': {'args': '<type __builtin__.int>',
                                            'longname': 'maxPossibleLevel',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'mpl'},
                       'proxyMode': {'args': '<type __builtin__.int>',
                                     'longname': 'proxyMode',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'pm'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'smallOffsets': {'args': '<type __builtin__.bool>',
                                        'longname': 'smallOffsets',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'so'}},
             'shortFlags': {'cl': 'currentLevel',
                            'csl': 'currentSubdLevel',
                            'dl': 'deepestLevel',
                            'dsl': 'displayLoad',
                            'est': 'edgeStats',
                            'fst': 'faceStats',
                            'mpl': 'maxPossibleLevel',
                            'pm': 'proxyMode',
                            'q': 'query',
                            'so': 'smallOffsets'},
             'type': 'modeling'},
  'subdivCrease': {'flags': {'sharpness': {'args': '<type __builtin__.bool>',
                                           'longname': 'sharpness',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'sh'}},
                   'shortFlags': {'sh': 'sharpness'},
                   'type': 'modeling'},
  'subdivDisplaySmoothness': {'flags': {'all': {'args': ['If set',
                                                         'change smoothness for all subdivision surfaces'],
                                                'longname': 'all',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 2,
                                                'shortname': 'boolean'},
                                        'query': {'args': '<type __builtin__.bool>',
                                                  'longname': 'query',
                                                  'numArgs': 0,
                                                  'shortname': 'q'},
                                        'smoothness': {'args': '<type __builtin__.int>',
                                                       'longname': 'smoothness',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 's'}},
                              'shortFlags': {'all': 'all',
                                             'q': 'query',
                                             's': 'smoothness'},
                              'type': 'modeling'},
  'subgraph': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'substituteGeometry': {'flags': {'disableNonSkinDeformers': {'args': '<type __builtin__.bool>',
                                                               'longname': 'disableNonSkinDeformers',
                                                               'modes': ['create'],
                                                               'numArgs': 1,
                                                               'shortname': 'dnd'},
                                   'newGeometryToLayer': {'args': '<type __builtin__.bool>',
                                                          'longname': 'newGeometryToLayer',
                                                          'modes': ['create'],
                                                          'numArgs': 0,
                                                          'shortname': 'ngl'},
                                   'oldGeometryToLayer': {'args': '<type __builtin__.bool>',
                                                          'longname': 'oldGeometryToLayer',
                                                          'modes': ['create'],
                                                          'numArgs': 0,
                                                          'shortname': 'ogl'},
                                   'reWeightDistTolerance': {'args': '<type __builtin__.float>',
                                                             'longname': 'reWeightDistTolerance',
                                                             'modes': ['create'],
                                                             'numArgs': 1,
                                                             'shortname': 'wdt'},
                                   'retainOldGeometry': {'args': '<type __builtin__.bool>',
                                                         'longname': 'retainOldGeometry',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'rog'}},
                         'shortFlags': {'dnd': 'disableNonSkinDeformers',
                                        'ngl': 'newGeometryToLayer',
                                        'ogl': 'oldGeometryToLayer',
                                        'rog': 'retainOldGeometry',
                                        'wdt': 'reWeightDistTolerance'},
                         'type': 'animation'},
  'suitePrefs': {'flags': {'applyToSuite': {'args': '<type __builtin__.str>',
                                            'longname': 'applyToSuite',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ats'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'installedAsSuite': {'args': '<type __builtin__.bool>',
                                                'longname': 'installedAsSuite',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ias'},
                           'isCompleteSuite': {'args': '<type __builtin__.bool>',
                                               'longname': 'isCompleteSuite',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ics'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'ats': 'applyToSuite',
                                'e': 'edit',
                                'ias': 'installedAsSuite',
                                'ics': 'isCompleteSuite',
                                'q': 'query'},
                 'type': 'general'},
  'superCtx': {'flags': {'attach': {'args': '<type __builtin__.str>',
                                    'longname': 'attach',
                                    'modes': ['multiuse'],
                                    'numArgs': 1,
                                    'shortname': 'a'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'image1': {'args': '<type __builtin__.str>',
                                    'longname': 'image1',
                                    'numArgs': 1,
                                    'shortname': 'i1'},
                         'image2': {'args': '<type __builtin__.str>',
                                    'longname': 'image2',
                                    'numArgs': 1,
                                    'shortname': 'i2'},
                         'image3': {'args': '<type __builtin__.str>',
                                    'longname': 'image3',
                                    'numArgs': 1,
                                    'shortname': 'i3'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'}},
               'shortFlags': {'a': 'attach',
                              'e': 'edit',
                              'ex': 'exists',
                              'i1': 'image1',
                              'i2': 'image2',
                              'i3': 'image3',
                              'q': 'query'},
               'type': 'context'},
  'surface': {'flags': {'degreeU': {'args': '<type __builtin__.int>',
                                    'longname': 'degreeU',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'du'},
                        'degreeV': {'args': '<type __builtin__.int>',
                                    'longname': 'degreeV',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'dv'},
                        'formU': {'args': '<type __builtin__.str>',
                                  'longname': 'formU',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'fu'},
                        'formV': {'args': '<type __builtin__.str>',
                                  'longname': 'formV',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'fv'},
                        'knotU': {'args': '<type __builtin__.float>',
                                  'longname': 'knotU',
                                  'modes': ['create', 'multiuse'],
                                  'numArgs': 1,
                                  'shortname': 'ku'},
                        'knotV': {'args': '<type __builtin__.float>',
                                  'longname': 'knotV',
                                  'modes': ['create', 'multiuse'],
                                  'numArgs': 1,
                                  'shortname': 'kv'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'objectSpace': {'args': '<type __builtin__.bool>',
                                        'longname': 'objectSpace',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ob'},
                        'point': {'args': ['<type __builtin__.float>',
                                           '<type __builtin__.float>',
                                           '<type __builtin__.float>'],
                                  'longname': 'point',
                                  'modes': ['create', 'multiuse'],
                                  'numArgs': 3,
                                  'shortname': 'p'},
                        'pointWeight': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'pointWeight',
                                        'modes': ['create', 'multiuse'],
                                        'numArgs': 4,
                                        'shortname': 'pw'},
                        'worldSpace': {'args': '<type __builtin__.bool>',
                                       'longname': 'worldSpace',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ws'}},
              'shortFlags': {'du': 'degreeU',
                             'dv': 'degreeV',
                             'fu': 'formU',
                             'fv': 'formV',
                             'ku': 'knotU',
                             'kv': 'knotV',
                             'n': 'name',
                             'ob': 'objectSpace',
                             'p': 'point',
                             'pw': 'pointWeight',
                             'ws': 'worldSpace'},
              'type': 'modeling'},
  'surfaceSampler': {'flags': {'camera': {'args': 'PyNode',
                                          'longname': 'camera',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'cam'},
                               'fileFormat': {'args': '<type __builtin__.str>',
                                              'longname': 'fileFormat',
                                              'modes': ['create',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'ff'},
                               'filename': {'args': '<type __builtin__.str>',
                                            'longname': 'filename',
                                            'modes': ['create',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'fn'},
                               'filterSize': {'args': '<type __builtin__.float>',
                                              'longname': 'filterSize',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'fs'},
                               'filterType': {'args': '<type __builtin__.int>',
                                              'longname': 'filterType',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ft'},
                               'flipU': {'args': '<type __builtin__.bool>',
                                         'longname': 'flipU',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'fu'},
                               'flipV': {'args': '<type __builtin__.bool>',
                                         'longname': 'flipV',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'fv'},
                               'ignoreMirroredFaces': {'args': '<type __builtin__.bool>',
                                                       'longname': 'ignoreMirroredFaces',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'imf'},
                               'ignoreTransforms': {'args': '<type __builtin__.bool>',
                                                    'longname': 'ignoreTransforms',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'it'},
                               'mapHeight': {'args': '<type __builtin__.int>',
                                             'longname': 'mapHeight',
                                             'modes': ['create',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'mh'},
                               'mapMaterials': {'args': '<type __builtin__.bool>',
                                                'longname': 'mapMaterials',
                                                'modes': ['create',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'mm'},
                               'mapOutput': {'args': '<type __builtin__.str>',
                                             'longname': 'mapOutput',
                                             'modes': ['create',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'mo'},
                               'mapSpace': {'args': '<type __builtin__.str>',
                                            'longname': 'mapSpace',
                                            'modes': ['create',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'sp'},
                               'mapWidth': {'args': '<type __builtin__.int>',
                                            'longname': 'mapWidth',
                                            'modes': ['create',
                                                      'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'mw'},
                               'maxSearchDistance': {'args': '<type __builtin__.float>',
                                                     'longname': 'maxSearchDistance',
                                                     'modes': ['create',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'msd'},
                               'maximumValue': {'args': '<type __builtin__.float>',
                                                'longname': 'maximumValue',
                                                'modes': ['create',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'max'},
                               'overscan': {'args': '<type __builtin__.int>',
                                            'longname': 'overscan',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'os'},
                               'searchCage': {'args': '<type __builtin__.str>',
                                              'longname': 'searchCage',
                                              'modes': ['create',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'sc'},
                               'searchMethod': {'args': '<type __builtin__.int>',
                                                'longname': 'searchMethod',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'sm'},
                               'searchOffset': {'args': '<type __builtin__.float>',
                                                'longname': 'searchOffset',
                                                'modes': ['create',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'so'},
                               'shadows': {'args': '<type __builtin__.bool>',
                                           'longname': 'shadows',
                                           'modes': ['create', 'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'sh'},
                               'source': {'args': '<type __builtin__.str>',
                                          'longname': 'source',
                                          'modes': ['create', 'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 's'},
                               'sourceUVSpace': {'args': '<type __builtin__.str>',
                                                 'longname': 'sourceUVSpace',
                                                 'modes': ['create',
                                                           'multiuse'],
                                                 'numArgs': 1,
                                                 'shortname': 'sus'},
                               'superSampling': {'args': '<type __builtin__.int>',
                                                 'longname': 'superSampling',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ss'},
                               'target': {'args': '<type __builtin__.str>',
                                          'longname': 'target',
                                          'modes': ['create', 'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 't'},
                               'targetUVSpace': {'args': '<type __builtin__.str>',
                                                 'longname': 'targetUVSpace',
                                                 'modes': ['create',
                                                           'multiuse'],
                                                 'numArgs': 1,
                                                 'shortname': 'tus'},
                               'useGeometryNormals': {'args': '<type __builtin__.bool>',
                                                      'longname': 'useGeometryNormals',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'ugn'},
                               'uvSet': {'args': '<type __builtin__.str>',
                                         'longname': 'uvSet',
                                         'modes': ['create', 'multiuse'],
                                         'numArgs': 1,
                                         'shortname': 'uv'}},
                     'shortFlags': {'cam': 'camera',
                                    'ff': 'fileFormat',
                                    'fn': 'filename',
                                    'fs': 'filterSize',
                                    'ft': 'filterType',
                                    'fu': 'flipU',
                                    'fv': 'flipV',
                                    'imf': 'ignoreMirroredFaces',
                                    'it': 'ignoreTransforms',
                                    'max': 'maximumValue',
                                    'mh': 'mapHeight',
                                    'mm': 'mapMaterials',
                                    'mo': 'mapOutput',
                                    'msd': 'maxSearchDistance',
                                    'mw': 'mapWidth',
                                    'os': 'overscan',
                                    's': 'source',
                                    'sc': 'searchCage',
                                    'sh': 'shadows',
                                    'sm': 'searchMethod',
                                    'so': 'searchOffset',
                                    'sp': 'mapSpace',
                                    'ss': 'superSampling',
                                    'sus': 'sourceUVSpace',
                                    't': 'target',
                                    'tus': 'targetUVSpace',
                                    'ugn': 'useGeometryNormals',
                                    'uv': 'uvSet'},
                     'type': 'rendering'},
  'surfaceShaderList': {'flags': {'add': {'args': 'PyNode',
                                          'longname': 'add',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'add'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'remove': {'args': 'PyNode',
                                             'longname': 'remove',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'rm'}},
                        'shortFlags': {'add': 'add',
                                       'e': 'edit',
                                       'q': 'query',
                                       'rm': 'remove'},
                        'type': 'rendering'},
  'swatchDisplayPort': {'callbackFlags': ['pressCommand',
                                          'pc',
                                          'visibleChangeCommand',
                                          'vcc',
                                          'dragCallback',
                                          'dgc',
                                          'dropCallback',
                                          'dpc'],
                        'flags': {'annotation': {'args': '<type __builtin__.str>',
                                                 'longname': 'annotation',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ann'},
                                  'backgroundColor': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'backgroundColor',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'bgc'},
                                  'borderColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'borderColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'bc'},
                                  'borderWidth': {'args': '<type __builtin__.int>',
                                                  'longname': 'borderWidth',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'bw'},
                                  'defineTemplate': {'args': '<type __builtin__.str>',
                                                     'longname': 'defineTemplate',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'dt'},
                                  'docTag': {'args': '<type __builtin__.str>',
                                             'longname': 'docTag',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dtg'},
                                  'dragCallback': {'args': 'script',
                                                   'longname': 'dragCallback',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dgc'},
                                  'dropCallback': {'args': 'script',
                                                   'longname': 'dropCallback',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dpc'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'enable': {'args': '<type __builtin__.bool>',
                                             'longname': 'enable',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'en'},
                                  'enableBackground': {'args': '<type __builtin__.bool>',
                                                       'longname': 'enableBackground',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ebg'},
                                  'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                          'longname': 'enableKeyboardFocus',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'ekf'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'fullPathName': {'args': '<type __builtin__.bool>',
                                                   'longname': 'fullPathName',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'fpn'},
                                  'height': {'args': '<type __builtin__.int>',
                                             'longname': 'height',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'h'},
                                  'highlightColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'highlightColor',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'hlc'},
                                  'isObscured': {'args': '<type __builtin__.bool>',
                                                 'longname': 'isObscured',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'io'},
                                  'manage': {'args': '<type __builtin__.bool>',
                                             'longname': 'manage',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'm'},
                                  'noBackground': {'args': '<type __builtin__.bool>',
                                                   'longname': 'noBackground',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'nbg'},
                                  'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                         'longname': 'numberOfPopupMenus',
                                                         'modes': ['query'],
                                                         'numArgs': 0,
                                                         'shortname': 'npm'},
                                  'parent': {'args': '<type __builtin__.str>',
                                             'longname': 'parent',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'p'},
                                  'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                     'longname': 'popupMenuArray',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'pma'},
                                  'pressCommand': {'args': 'script',
                                                   'longname': 'pressCommand',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'pc'},
                                  'preventOverride': {'args': '<type __builtin__.bool>',
                                                      'longname': 'preventOverride',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'po'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'renderPriority': {'args': '<type __builtin__.int>',
                                                     'longname': 'renderPriority',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'rp'},
                                  'renderSize': {'args': '<type __builtin__.int>',
                                                 'longname': 'renderSize',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rs'},
                                  'shadingNode': {'args': 'PyNode',
                                                  'longname': 'shadingNode',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sn'},
                                  'statusBarMessage': {'args': '<type __builtin__.str>',
                                                       'longname': 'statusBarMessage',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'sbm'},
                                  'useTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'useTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ut'},
                                  'visible': {'args': '<type __builtin__.bool>',
                                              'longname': 'visible',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'vis'},
                                  'visibleChangeCommand': {'args': 'script',
                                                           'longname': 'visibleChangeCommand',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'vcc'},
                                  'width': {'args': '<type __builtin__.int>',
                                            'longname': 'width',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'w'},
                                  'widthHeight': {'args': ['<type __builtin__.int>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'widthHeight',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 2,
                                                  'shortname': 'wh'}},
                        'shortFlags': {'ann': 'annotation',
                                       'bc': 'borderColor',
                                       'bgc': 'backgroundColor',
                                       'bw': 'borderWidth',
                                       'dgc': 'dragCallback',
                                       'dpc': 'dropCallback',
                                       'dt': 'defineTemplate',
                                       'dtg': 'docTag',
                                       'e': 'edit',
                                       'ebg': 'enableBackground',
                                       'ekf': 'enableKeyboardFocus',
                                       'en': 'enable',
                                       'ex': 'exists',
                                       'fpn': 'fullPathName',
                                       'h': 'height',
                                       'hlc': 'highlightColor',
                                       'io': 'isObscured',
                                       'm': 'manage',
                                       'nbg': 'noBackground',
                                       'npm': 'numberOfPopupMenus',
                                       'p': 'parent',
                                       'pc': 'pressCommand',
                                       'pma': 'popupMenuArray',
                                       'po': 'preventOverride',
                                       'q': 'query',
                                       'rp': 'renderPriority',
                                       'rs': 'renderSize',
                                       'sbm': 'statusBarMessage',
                                       'sn': 'shadingNode',
                                       'ut': 'useTemplate',
                                       'vcc': 'visibleChangeCommand',
                                       'vis': 'visible',
                                       'w': 'width',
                                       'wh': 'widthHeight'},
                        'type': 'windows'},
  'swatchRefresh': {'flags': {}, 'shortFlags': {}, 'type': 'rendering'},
  'switchTable': {'callbackFlags': ['visibleChangeCommand',
                                    'vcc',
                                    'dragCallback',
                                    'dgc',
                                    'dropCallback',
                                    'dpc'],
                  'flags': {'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'label1': {'args': '<type __builtin__.str>',
                                       'longname': 'label1',
                                       'modes': ['edit'],
                                       'numArgs': 1,
                                       'shortname': 'l1'},
                            'label2': {'args': '<type __builtin__.str>',
                                       'longname': 'label2',
                                       'modes': ['edit'],
                                       'numArgs': 1,
                                       'shortname': 'l2'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'selectedRow': {'args': '<type __builtin__.bool>',
                                            'longname': 'selectedRow',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'sr'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'switchNode': {'args': 'PyNode',
                                           'longname': 'switchNode',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sn'},
                            'underPointerRow': {'args': '<type __builtin__.bool>',
                                                'longname': 'underPointerRow',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'upr'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'ann': 'annotation',
                                 'bgc': 'backgroundColor',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'ex': 'exists',
                                 'fpn': 'fullPathName',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'io': 'isObscured',
                                 'l1': 'label1',
                                 'l2': 'label2',
                                 'm': 'manage',
                                 'nbg': 'noBackground',
                                 'npm': 'numberOfPopupMenus',
                                 'p': 'parent',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'sbm': 'statusBarMessage',
                                 'sn': 'switchNode',
                                 'sr': 'selectedRow',
                                 'upr': 'underPointerRow',
                                 'ut': 'useTemplate',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width'},
                  'type': 'windows'},
  'symbolButton': {'callbackFlags': ['visibleChangeCommand',
                                     'vcc',
                                     'dragCallback',
                                     'dgc',
                                     'command',
                                     'c',
                                     'dropCallback',
                                     'dpc'],
                   'flags': {'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'command': {'args': 'script',
                                         'longname': 'command',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'c'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'image': {'args': '<type __builtin__.str>',
                                       'longname': 'image',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'shortFlags': {'ann': 'annotation',
                                  'bgc': 'backgroundColor',
                                  'c': 'command',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ebg': 'enableBackground',
                                  'ekf': 'enableKeyboardFocus',
                                  'en': 'enable',
                                  'ex': 'exists',
                                  'fpn': 'fullPathName',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'i': 'image',
                                  'io': 'isObscured',
                                  'm': 'manage',
                                  'nbg': 'noBackground',
                                  'npm': 'numberOfPopupMenus',
                                  'p': 'parent',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'q': 'query',
                                  'sbm': 'statusBarMessage',
                                  'ut': 'useTemplate',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'w': 'width'},
                   'type': 'windows'},
  'symbolCheckBox': {'callbackFlags': ['dropCallback',
                                       'dpc',
                                       'visibleChangeCommand',
                                       'vcc',
                                       'onCommand',
                                       'onc',
                                       'dragCallback',
                                       'dgc',
                                       'offCommand',
                                       'ofc',
                                       'changeCommand',
                                       'cc'],
                     'flags': {'annotation': {'args': '<type __builtin__.str>',
                                              'longname': 'annotation',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ann'},
                               'backgroundColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'backgroundColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'bgc'},
                               'changeCommand': {'args': 'script',
                                                 'longname': 'changeCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cc'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'disableOffImage': {'args': '<type __builtin__.str>',
                                                   'longname': 'disableOffImage',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dfi'},
                               'disableOnImage': {'args': '<type __builtin__.str>',
                                                  'longname': 'disableOnImage',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dni'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'dragCallback': {'args': 'script',
                                                'longname': 'dragCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dgc'},
                               'dropCallback': {'args': 'script',
                                                'longname': 'dropCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dpc'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'enable': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en'},
                               'enableBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableBackground',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ebg'},
                               'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                       'longname': 'enableKeyboardFocus',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ekf'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'fullPathName': {'args': '<type __builtin__.bool>',
                                                'longname': 'fullPathName',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'fpn'},
                               'height': {'args': '<type __builtin__.int>',
                                          'longname': 'height',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'h'},
                               'highlightColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'highlightColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'hlc'},
                               'image': {'args': '<type __builtin__.str>',
                                         'longname': 'image',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i'},
                               'innerMargin': {'args': '<type __builtin__.bool>',
                                               'longname': 'innerMargin',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 0,
                                               'shortname': 'im'},
                               'isObscured': {'args': '<type __builtin__.bool>',
                                              'longname': 'isObscured',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'io'},
                               'ltVersion': {'args': '<type __builtin__.str>',
                                             'longname': 'ltVersion',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'lt'},
                               'manage': {'args': '<type __builtin__.bool>',
                                          'longname': 'manage',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'm'},
                               'noBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'noBackground',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nbg'},
                               'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'numberOfPopupMenus',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'npm'},
                               'offCommand': {'args': 'script',
                                              'longname': 'offCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ofc'},
                               'offImage': {'args': '<type __builtin__.str>',
                                            'longname': 'offImage',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ofi'},
                               'onCommand': {'args': 'script',
                                             'longname': 'onCommand',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'onc'},
                               'onImage': {'args': '<type __builtin__.str>',
                                           'longname': 'onImage',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'oni'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                  'longname': 'popupMenuArray',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'pma'},
                               'preventOverride': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preventOverride',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'po'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'statusBarMessage': {'args': '<type __builtin__.str>',
                                                    'longname': 'statusBarMessage',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sbm'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'},
                               'value': {'args': '<type __builtin__.bool>',
                                         'longname': 'value',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'v'},
                               'version': {'args': '<type __builtin__.str>',
                                           'longname': 'version',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ver'},
                               'visible': {'args': '<type __builtin__.bool>',
                                           'longname': 'visible',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vis'},
                               'visibleChangeCommand': {'args': 'script',
                                                        'longname': 'visibleChangeCommand',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'vcc'},
                               'width': {'args': '<type __builtin__.int>',
                                         'longname': 'width',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'w'}},
                     'shortFlags': {'ann': 'annotation',
                                    'bgc': 'backgroundColor',
                                    'cc': 'changeCommand',
                                    'dfi': 'disableOffImage',
                                    'dgc': 'dragCallback',
                                    'dni': 'disableOnImage',
                                    'dpc': 'dropCallback',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'e': 'edit',
                                    'ebg': 'enableBackground',
                                    'ekf': 'enableKeyboardFocus',
                                    'en': 'enable',
                                    'ex': 'exists',
                                    'fpn': 'fullPathName',
                                    'h': 'height',
                                    'hlc': 'highlightColor',
                                    'i': 'image',
                                    'im': 'innerMargin',
                                    'io': 'isObscured',
                                    'lt': 'ltVersion',
                                    'm': 'manage',
                                    'nbg': 'noBackground',
                                    'npm': 'numberOfPopupMenus',
                                    'ofc': 'offCommand',
                                    'ofi': 'offImage',
                                    'onc': 'onCommand',
                                    'oni': 'onImage',
                                    'p': 'parent',
                                    'pma': 'popupMenuArray',
                                    'po': 'preventOverride',
                                    'q': 'query',
                                    'sbm': 'statusBarMessage',
                                    'ut': 'useTemplate',
                                    'v': 'value',
                                    'vcc': 'visibleChangeCommand',
                                    'ver': 'version',
                                    'vis': 'visible',
                                    'w': 'width'},
                     'type': 'windows'},
  'symmetricModelling': {'flags': {'about': {'args': '<type __builtin__.str>',
                                             'longname': 'about',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'a'},
                                   'allowPartial': {'args': '<type __builtin__.bool>',
                                                    'longname': 'allowPartial',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ap'},
                                   'axis': {'args': '<type __builtin__.str>',
                                            'longname': 'axis',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ax'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'preserveSeam': {'args': '<type __builtin__.int>',
                                                    'longname': 'preserveSeam',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ps'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'reset': {'args': '<type __builtin__.bool>',
                                             'longname': 'reset',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'r'},
                                   'seamFalloffCurve': {'args': '<type __builtin__.str>',
                                                        'longname': 'seamFalloffCurve',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'sf'},
                                   'seamTolerance': {'args': '<type __builtin__.float>',
                                                     'longname': 'seamTolerance',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'st'},
                                   'symmetry': {'args': '<type __builtin__.int>',
                                                'longname': 'symmetry',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 's'},
                                   'tolerance': {'args': '<type __builtin__.float>',
                                                 'longname': 'tolerance',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 't'},
                                   'topoSymmetry': {'args': '<type __builtin__.bool>',
                                                    'longname': 'topoSymmetry',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ts'}},
                         'shortFlags': {'a': 'about',
                                        'ap': 'allowPartial',
                                        'ax': 'axis',
                                        'e': 'edit',
                                        'ps': 'preserveSeam',
                                        'q': 'query',
                                        'r': 'reset',
                                        's': 'symmetry',
                                        'sf': 'seamFalloffCurve',
                                        'st': 'seamTolerance',
                                        't': 'tolerance',
                                        'ts': 'topoSymmetry'},
                         'type': 'general'},
  'syncSculptCache': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'sysFile': {'flags': {'copy': {'args': '<type __builtin__.str>',
                                 'longname': 'copy',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'cp'},
                        'delete': {'args': '<type __builtin__.bool>',
                                   'longname': 'delete',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'delete'},
                        'makeDir': {'args': '<type __builtin__.bool>',
                                    'longname': 'makeDir',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'md'},
                        'move': {'args': '<type __builtin__.str>',
                                 'longname': 'move',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'mov'},
                        'removeEmptyDir': {'args': '<type __builtin__.bool>',
                                           'longname': 'removeEmptyDir',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'red'},
                        'rename': {'args': '<type __builtin__.str>',
                                   'longname': 'rename',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'ren'}},
              'removedFlags': {'del': 'delete'},
              'shortFlags': {'cp': 'copy',
                             'delete': 'delete',
                             'md': 'makeDir',
                             'mov': 'move',
                             'red': 'removeEmptyDir',
                             'ren': 'rename'},
              'type': 'system'},
  'tabLayout': {'callbackFlags': ['selectCommand',
                                  'sc',
                                  'doubleClickCommand',
                                  'dcc',
                                  'visibleChangeCommand',
                                  'vcc',
                                  'preSelectCommand',
                                  'psc',
                                  'dragCallback',
                                  'dgc',
                                  'newTabCommand',
                                  'ntc',
                                  'changeCommand',
                                  'cc',
                                  'dropCallback',
                                  'dpc'],
                'flags': {'annotation': {'args': '<type __builtin__.str>',
                                         'longname': 'annotation',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ann'},
                          'backgroundColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'backgroundColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'bgc'},
                          'borderStyle': {'args': '<type __builtin__.str>',
                                          'longname': 'borderStyle',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'bs'},
                          'changeCommand': {'args': 'script',
                                            'longname': 'changeCommand',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cc'},
                          'childArray': {'args': '<type __builtin__.bool>',
                                         'longname': 'childArray',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'ca'},
                          'childResizable': {'args': '<type __builtin__.bool>',
                                             'longname': 'childResizable',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'cr'},
                          'defineTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'defineTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dt'},
                          'docTag': {'args': '<type __builtin__.str>',
                                     'longname': 'docTag',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'dtg'},
                          'doubleClickCommand': {'args': 'script',
                                                 'longname': 'doubleClickCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dcc'},
                          'dragCallback': {'args': 'script',
                                           'longname': 'dragCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dgc'},
                          'dropCallback': {'args': 'script',
                                           'longname': 'dropCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dpc'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'enable': {'args': '<type __builtin__.bool>',
                                     'longname': 'enable',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'en'},
                          'enableBackground': {'args': '<type __builtin__.bool>',
                                               'longname': 'enableBackground',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ebg'},
                          'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableKeyboardFocus',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ekf'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'fullPathName': {'args': '<type __builtin__.bool>',
                                           'longname': 'fullPathName',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'fpn'},
                          'height': {'args': '<type __builtin__.int>',
                                     'longname': 'height',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'h'},
                          'highlightColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'highlightColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'hlc'},
                          'horizontalScrollBarThickness': {'args': '<type __builtin__.int>',
                                                           'longname': 'horizontalScrollBarThickness',
                                                           'modes': ['create',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'hst'},
                          'image': {'args': '<type __builtin__.str>',
                                    'longname': 'image',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i'},
                          'imageVisible': {'args': '<type __builtin__.bool>',
                                           'longname': 'imageVisible',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'iv'},
                          'innerMarginHeight': {'args': '<type __builtin__.int>',
                                                'longname': 'innerMarginHeight',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'imh'},
                          'innerMarginWidth': {'args': '<type __builtin__.int>',
                                               'longname': 'innerMarginWidth',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'imw'},
                          'isObscured': {'args': '<type __builtin__.bool>',
                                         'longname': 'isObscured',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'io'},
                          'manage': {'args': '<type __builtin__.bool>',
                                     'longname': 'manage',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'm'},
                          'minChildWidth': {'args': '<type __builtin__.int>',
                                            'longname': 'minChildWidth',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'mcw'},
                          'moveTab': {'args': ['<type __builtin__.int>',
                                               '<type __builtin__.int>'],
                                      'longname': 'moveTab',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 2,
                                      'shortname': 'mt'},
                          'newTabCommand': {'args': 'script',
                                            'longname': 'newTabCommand',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ntc'},
                          'noBackground': {'args': '<type __builtin__.bool>',
                                           'longname': 'noBackground',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nbg'},
                          'numberOfChildren': {'args': '<type __builtin__.bool>',
                                               'longname': 'numberOfChildren',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'nch'},
                          'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'numberOfPopupMenus',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'npm'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                          'popupMenuArray': {'args': '<type __builtin__.bool>',
                                             'longname': 'popupMenuArray',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'pma'},
                          'preSelectCommand': {'args': 'script',
                                               'longname': 'preSelectCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'psc'},
                          'preventOverride': {'args': '<type __builtin__.bool>',
                                              'longname': 'preventOverride',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'po'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'scrollable': {'args': '<type __builtin__.bool>',
                                         'longname': 'scrollable',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'scr'},
                          'scrollableTabs': {'args': '<type __builtin__.bool>',
                                             'longname': 'scrollableTabs',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'stb'},
                          'selectCommand': {'args': 'script',
                                            'longname': 'selectCommand',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sc'},
                          'selectTab': {'args': '<type __builtin__.str>',
                                        'longname': 'selectTab',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'st'},
                          'selectTabIndex': {'args': '<type __builtin__.int>',
                                             'longname': 'selectTabIndex',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sti'},
                          'showNewTab': {'args': '<type __builtin__.bool>',
                                         'longname': 'showNewTab',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'snt'},
                          'statusBarMessage': {'args': '<type __builtin__.str>',
                                               'longname': 'statusBarMessage',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sbm'},
                          'tabLabel': {'args': ['<type __builtin__.str>',
                                                '<type __builtin__.str>'],
                                       'longname': 'tabLabel',
                                       'modes': ['create',
                                                 'query',
                                                 'edit',
                                                 'multiuse'],
                                       'numArgs': 2,
                                       'shortname': 'tl'},
                          'tabLabelIndex': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'tabLabelIndex',
                                            'modes': ['create',
                                                      'query',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 2,
                                            'shortname': 'tli'},
                          'tabPosition': {'args': '<type __builtin__.str>',
                                          'longname': 'tabPosition',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'tp'},
                          'tabTooltip': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'tabTooltip',
                                         'modes': ['create',
                                                   'query',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 2,
                                         'shortname': 'tt'},
                          'tabTooltipIndex': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'tabTooltipIndex',
                                              'modes': ['create',
                                                        'query',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'tti'},
                          'tabsClosable': {'args': '<type __builtin__.bool>',
                                           'longname': 'tabsClosable',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'tc'},
                          'tabsVisible': {'args': '<type __builtin__.bool>',
                                          'longname': 'tabsVisible',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'tv'},
                          'useTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'useTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ut'},
                          'verticalScrollBarThickness': {'args': '<type __builtin__.int>',
                                                         'longname': 'verticalScrollBarThickness',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'vst'},
                          'visible': {'args': '<type __builtin__.bool>',
                                      'longname': 'visible',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'vis'},
                          'visibleChangeCommand': {'args': 'script',
                                                   'longname': 'visibleChangeCommand',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'vcc'},
                          'width': {'args': '<type __builtin__.int>',
                                    'longname': 'width',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'w'}},
                'shortFlags': {'ann': 'annotation',
                               'bgc': 'backgroundColor',
                               'bs': 'borderStyle',
                               'ca': 'childArray',
                               'cc': 'changeCommand',
                               'cr': 'childResizable',
                               'dcc': 'doubleClickCommand',
                               'dgc': 'dragCallback',
                               'dpc': 'dropCallback',
                               'dt': 'defineTemplate',
                               'dtg': 'docTag',
                               'e': 'edit',
                               'ebg': 'enableBackground',
                               'ekf': 'enableKeyboardFocus',
                               'en': 'enable',
                               'ex': 'exists',
                               'fpn': 'fullPathName',
                               'h': 'height',
                               'hlc': 'highlightColor',
                               'hst': 'horizontalScrollBarThickness',
                               'i': 'image',
                               'imh': 'innerMarginHeight',
                               'imw': 'innerMarginWidth',
                               'io': 'isObscured',
                               'iv': 'imageVisible',
                               'm': 'manage',
                               'mcw': 'minChildWidth',
                               'mt': 'moveTab',
                               'nbg': 'noBackground',
                               'nch': 'numberOfChildren',
                               'npm': 'numberOfPopupMenus',
                               'ntc': 'newTabCommand',
                               'p': 'parent',
                               'pma': 'popupMenuArray',
                               'po': 'preventOverride',
                               'psc': 'preSelectCommand',
                               'q': 'query',
                               'sbm': 'statusBarMessage',
                               'sc': 'selectCommand',
                               'scr': 'scrollable',
                               'snt': 'showNewTab',
                               'st': 'selectTab',
                               'stb': 'scrollableTabs',
                               'sti': 'selectTabIndex',
                               'tc': 'tabsClosable',
                               'tl': 'tabLabel',
                               'tli': 'tabLabelIndex',
                               'tp': 'tabPosition',
                               'tt': 'tabTooltip',
                               'tti': 'tabTooltipIndex',
                               'tv': 'tabsVisible',
                               'ut': 'useTemplate',
                               'vcc': 'visibleChangeCommand',
                               'vis': 'visible',
                               'vst': 'verticalScrollBarThickness',
                               'w': 'width'},
                'type': 'windows'},
  'tangentConstraint': {'flags': {'aimVector': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'aimVector',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'aim'},
                                  'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'layer': {'args': '<type __builtin__.str>',
                                            'longname': 'layer',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'l'},
                                  'name': {'args': '<type __builtin__.str>',
                                           'longname': 'name',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'n'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'remove': {'args': '<type __builtin__.bool>',
                                             'longname': 'remove',
                                             'modes': ['edit'],
                                             'numArgs': 0,
                                             'shortname': 'rm'},
                                  'targetList': {'args': '<type __builtin__.bool>',
                                                 'longname': 'targetList',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'tl'},
                                  'upVector': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'upVector',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'u'},
                                  'weight': {'args': '<type __builtin__.float>',
                                             'longname': 'weight',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'w'},
                                  'weightAliasList': {'args': '<type __builtin__.bool>',
                                                      'longname': 'weightAliasList',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'wal'},
                                  'worldUpObject': {'args': 'PyNode',
                                                    'longname': 'worldUpObject',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'wuo'},
                                  'worldUpType': {'args': '<type __builtin__.str>',
                                                  'longname': 'worldUpType',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'wut'},
                                  'worldUpVector': {'args': ['<type __builtin__.float>',
                                                             '<type __builtin__.float>',
                                                             '<type __builtin__.float>'],
                                                    'longname': 'worldUpVector',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'wu'}},
                        'resultNeedsUnpacking': True,
                        'shortFlags': {'aim': 'aimVector',
                                       'e': 'edit',
                                       'n': 'name',
                                       'q': 'query',
                                       'rm': 'remove',
                                       'tl': 'targetList',
                                       'u': 'upVector',
                                       'w': 'weight',
                                       'wal': 'weightAliasList',
                                       'wu': 'worldUpVector',
                                       'wuo': 'worldUpObject',
                                       'wut': 'worldUpType'},
                        'type': 'animation'},
  'targetWeldCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'image1': {'args': '<type __builtin__.str>',
                                         'longname': 'image1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i1'},
                              'image2': {'args': '<type __builtin__.str>',
                                         'longname': 'image2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i2'},
                              'image3': {'args': '<type __builtin__.str>',
                                         'longname': 'image3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i3'},
                              'mergeToCenter': {'args': '<type __builtin__.bool>',
                                                'longname': 'mergeToCenter',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'mtc'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'}},
                    'shortFlags': {'e': 'edit',
                                   'ex': 'exists',
                                   'i1': 'image1',
                                   'i2': 'image2',
                                   'i3': 'image3',
                                   'mtc': 'mergeToCenter',
                                   'q': 'query'},
                    'type': 'context'},
  'tension': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                  'longname': 'after',
                                  'modes': ['create', 'edit'],
                                  'numArgs': 0,
                                  'shortname': 'af'},
                        'afterReference': {'args': '<type __builtin__.bool>',
                                           'longname': 'afterReference',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'ar'},
                        'before': {'args': '<type __builtin__.bool>',
                                   'longname': 'before',
                                   'modes': ['create', 'edit'],
                                   'numArgs': 0,
                                   'shortname': 'bf'},
                        'deformerTools': {'args': '<type __builtin__.bool>',
                                          'longname': 'deformerTools',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'dt'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'envelope': {'args': '<type __builtin__.float>',
                                     'longname': 'envelope',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'resultNeedsUnpacking': True,
                                     'shortname': 'en'},
                        'exclusive': {'args': '<type __builtin__.str>',
                                      'longname': 'exclusive',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'ex'},
                        'frontOfChain': {'args': '<type __builtin__.bool>',
                                         'longname': 'frontOfChain',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'foc'},
                        'geometry': {'args': '<type __builtin__.str>',
                                     'longname': 'geometry',
                                     'modes': ['query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'g'},
                        'geometryIndices': {'args': '<type __builtin__.bool>',
                                            'longname': 'geometryIndices',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'gi'},
                        'ignoreSelected': {'args': '<type __builtin__.bool>',
                                           'longname': 'ignoreSelected',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ignoreSelected'},
                        'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                    'longname': 'includeHiddenSelections',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ihs'},
                        'inwardConstraint': {'args': '<type __builtin__.float>',
                                             'longname': 'inwardConstraint',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'resultNeedsUnpacking': True,
                                             'shortname': 'iwc'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'outwardConstraint': {'args': '<type __builtin__.float>',
                                              'longname': 'outwardConstraint',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'resultNeedsUnpacking': True,
                                              'shortname': 'owc'},
                        'parallel': {'args': '<type __builtin__.bool>',
                                     'longname': 'parallel',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 0,
                                     'shortname': 'par'},
                        'pinBorderVertices': {'args': '<type __builtin__.bool>',
                                              'longname': 'pinBorderVertices',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'resultNeedsUnpacking': True,
                                              'shortname': 'pbv'},
                        'prune': {'args': '<type __builtin__.bool>',
                                  'longname': 'prune',
                                  'modes': ['edit'],
                                  'numArgs': 0,
                                  'shortname': 'pr'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'remove': {'args': '<type __builtin__.bool>',
                                   'longname': 'remove',
                                   'modes': ['edit', 'multiuse'],
                                   'numArgs': 0,
                                   'shortname': 'rm'},
                        'smoothingIterations': {'args': '<type __builtin__.int>',
                                                'longname': 'smoothingIterations',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'si'},
                        'smoothingStep': {'args': '<type __builtin__.float>',
                                          'longname': 'smoothingStep',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'resultNeedsUnpacking': True,
                                          'shortname': 'ss'},
                        'split': {'args': '<type __builtin__.bool>',
                                  'longname': 'split',
                                  'modes': ['create', 'edit'],
                                  'numArgs': 0,
                                  'shortname': 'sp'}},
              'removedFlags': {'is': 'ignoreSelected'},
              'resultNeedsUnpacking': True,
              'shortFlags': {'af': 'after',
                             'ar': 'afterReference',
                             'bf': 'before',
                             'dt': 'deformerTools',
                             'e': 'edit',
                             'en': 'envelope',
                             'ex': 'exclusive',
                             'foc': 'frontOfChain',
                             'g': 'geometry',
                             'gi': 'geometryIndices',
                             'ignoreSelected': 'ignoreSelected',
                             'ihs': 'includeHiddenSelections',
                             'iwc': 'inwardConstraint',
                             'n': 'name',
                             'owc': 'outwardConstraint',
                             'par': 'parallel',
                             'pbv': 'pinBorderVertices',
                             'pr': 'prune',
                             'q': 'query',
                             'rm': 'remove',
                             'si': 'smoothingIterations',
                             'sp': 'split',
                             'ss': 'smoothingStep'},
              'type': 'animation'},
  'testPa': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'testPassContribution': {'flags': {'renderLayer': {'args': '<type __builtin__.str>',
                                                     'longname': 'renderLayer',
                                                     'numArgs': 1,
                                                     'shortname': 'rl'},
                                     'renderPass': {'args': '<type __builtin__.str>',
                                                    'longname': 'renderPass',
                                                    'numArgs': 1,
                                                    'shortname': 'rp'}},
                           'shortFlags': {'rl': 'renderLayer',
                                          'rp': 'renderPass'},
                           'type': 'other'},
  'texCutContext': {'flags': {'adjustSize': {'args': '<type __builtin__.bool>',
                                             'longname': 'adjustSize',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'asz'},
                              'displayShellBorders': {'args': '<type __builtin__.bool>',
                                                      'longname': 'displayShellBorders',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dsb'},
                              'edgeSelectSensitive': {'args': '<type __builtin__.float>',
                                                      'longname': 'edgeSelectSensitive',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ess'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'history': {'args': '<type __builtin__.bool>',
                                          'longname': 'history',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ch'},
                              'image1': {'args': '<type __builtin__.str>',
                                         'longname': 'image1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i1'},
                              'image2': {'args': '<type __builtin__.str>',
                                         'longname': 'image2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i2'},
                              'image3': {'args': '<type __builtin__.str>',
                                         'longname': 'image3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i3'},
                              'mode': {'args': '<type __builtin__.str>',
                                       'longname': 'mode',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                              'moveRatio': {'args': '<type __builtin__.float>',
                                            'longname': 'moveRatio',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'mvr'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'size': {'args': '<type __builtin__.float>',
                                       'longname': 'size',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'sz'},
                              'steadyStroke': {'args': '<type __builtin__.bool>',
                                               'longname': 'steadyStroke',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ss'},
                              'steadyStrokeDistance': {'args': '<type __builtin__.float>',
                                                       'longname': 'steadyStrokeDistance',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ssd'},
                              'touchToSew': {'args': '<type __builtin__.bool>',
                                             'longname': 'touchToSew',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'tts'}},
                    'shortFlags': {'asz': 'adjustSize',
                                   'ch': 'history',
                                   'dsb': 'displayShellBorders',
                                   'e': 'edit',
                                   'ess': 'edgeSelectSensitive',
                                   'ex': 'exists',
                                   'i1': 'image1',
                                   'i2': 'image2',
                                   'i3': 'image3',
                                   'm': 'mode',
                                   'mvr': 'moveRatio',
                                   'n': 'name',
                                   'q': 'query',
                                   'ss': 'steadyStroke',
                                   'ssd': 'steadyStrokeDistance',
                                   'sz': 'size',
                                   'tts': 'touchToSew'},
                    'type': 'context'},
  'texLatticeDeformContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                 'longname': 'edit',
                                                 'numArgs': 0,
                                                 'shortname': 'e'},
                                        'envelope': {'args': '<type __builtin__.float>',
                                                     'longname': 'envelope',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ev'},
                                        'exists': {'args': '<type __builtin__.bool>',
                                                   'longname': 'exists',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'ex'},
                                        'history': {'args': '<type __builtin__.bool>',
                                                    'longname': 'history',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                                        'image1': {'args': '<type __builtin__.str>',
                                                   'longname': 'image1',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'i1'},
                                        'image2': {'args': '<type __builtin__.str>',
                                                   'longname': 'image2',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'i2'},
                                        'image3': {'args': '<type __builtin__.str>',
                                                   'longname': 'image3',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'i3'},
                                        'latticeColumns': {'args': '<type __builtin__.int>',
                                                           'longname': 'latticeColumns',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'lc'},
                                        'latticeRows': {'args': '<type __builtin__.int>',
                                                        'longname': 'latticeRows',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'lr'},
                                        'name': {'args': '<type __builtin__.str>',
                                                 'longname': 'name',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'n'},
                                        'query': {'args': '<type __builtin__.bool>',
                                                  'longname': 'query',
                                                  'numArgs': 0,
                                                  'shortname': 'q'},
                                        'showMoveManipulator': {'args': '<type __builtin__.bool>',
                                                                'longname': 'showMoveManipulator',
                                                                'modes': ['create',
                                                                          'query',
                                                                          'edit'],
                                                                'numArgs': 1,
                                                                'shortname': 'smm'},
                                        'snapPixelMode': {'args': '<type __builtin__.bool>',
                                                          'longname': 'snapPixelMode',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'spm'},
                                        'useBoundingRect': {'args': '<type __builtin__.bool>',
                                                            'longname': 'useBoundingRect',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'ubr'}},
                              'shortFlags': {'ch': 'history',
                                             'e': 'edit',
                                             'ev': 'envelope',
                                             'ex': 'exists',
                                             'i1': 'image1',
                                             'i2': 'image2',
                                             'i3': 'image3',
                                             'lc': 'latticeColumns',
                                             'lr': 'latticeRows',
                                             'n': 'name',
                                             'q': 'query',
                                             'smm': 'showMoveManipulator',
                                             'spm': 'snapPixelMode',
                                             'ubr': 'useBoundingRect'},
                              'type': 'context'},
  'texManipContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'image1': {'args': '<type __builtin__.str>',
                                           'longname': 'image1',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i1'},
                                'image2': {'args': '<type __builtin__.str>',
                                           'longname': 'image2',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i2'},
                                'image3': {'args': '<type __builtin__.str>',
                                           'longname': 'image3',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i3'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'}},
                      'shortFlags': {'e': 'edit',
                                     'ex': 'exists',
                                     'i1': 'image1',
                                     'i2': 'image2',
                                     'i3': 'image3',
                                     'q': 'query'},
                      'type': 'context'},
  'texMoveContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'editPivotMode': {'args': '<type __builtin__.bool>',
                                                 'longname': 'editPivotMode',
                                                 'modes': ['query'],
                                                 'numArgs': 1,
                                                 'shortname': 'epm'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'position': {'args': '<type __builtin__.bool>',
                                            'longname': 'position',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'p'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'snap': {'args': '<type __builtin__.bool>',
                                        'longname': 'snap',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 's'},
                               'snapComponentsRelative': {'args': '<type __builtin__.bool>',
                                                          'longname': 'snapComponentsRelative',
                                                          'modes': ['query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'scr'},
                               'snapPixelMode': {'args': '<type __builtin__.int>',
                                                 'longname': 'snapPixelMode',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'spm'},
                               'snapValue': {'args': '<type __builtin__.float>',
                                             'longname': 'snapValue',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sv'},
                               'tweakMode': {'args': '<type __builtin__.bool>',
                                             'longname': 'tweakMode',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'twk'}},
                     'shortFlags': {'e': 'edit',
                                    'epm': 'editPivotMode',
                                    'ex': 'exists',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'p': 'position',
                                    'q': 'query',
                                    's': 'snap',
                                    'scr': 'snapComponentsRelative',
                                    'spm': 'snapPixelMode',
                                    'sv': 'snapValue',
                                    'twk': 'tweakMode'},
                     'type': 'context'},
  'texMoveUVShellContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                               'longname': 'edit',
                                               'numArgs': 0,
                                               'shortname': 'e'},
                                      'exists': {'args': '<type __builtin__.bool>',
                                                 'longname': 'exists',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ex'},
                                      'image1': {'args': '<type __builtin__.str>',
                                                 'longname': 'image1',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'i1'},
                                      'image2': {'args': '<type __builtin__.str>',
                                                 'longname': 'image2',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'i2'},
                                      'image3': {'args': '<type __builtin__.str>',
                                                 'longname': 'image3',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'i3'},
                                      'iterations': {'args': '<type __builtin__.int>',
                                                     'longname': 'iterations',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'it'},
                                      'mask': {'args': '<type __builtin__.bool>',
                                               'longname': 'mask',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'm'},
                                      'position': {'args': '<type __builtin__.bool>',
                                                   'longname': 'position',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'p'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'shellBorder': {'args': '<type __builtin__.float>',
                                                      'longname': 'shellBorder',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'sb'}},
                            'shortFlags': {'e': 'edit',
                                           'ex': 'exists',
                                           'i1': 'image1',
                                           'i2': 'image2',
                                           'i3': 'image3',
                                           'it': 'iterations',
                                           'm': 'mask',
                                           'p': 'position',
                                           'q': 'query',
                                           'sb': 'shellBorder'},
                            'type': 'context'},
  'texRotateContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'editPivotMode': {'args': '<type __builtin__.bool>',
                                                   'longname': 'editPivotMode',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'epm'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'image1': {'args': '<type __builtin__.str>',
                                            'longname': 'image1',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i1'},
                                 'image2': {'args': '<type __builtin__.str>',
                                            'longname': 'image2',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i2'},
                                 'image3': {'args': '<type __builtin__.str>',
                                            'longname': 'image3',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i3'},
                                 'position': {'args': '<type __builtin__.bool>',
                                              'longname': 'position',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'p'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'snap': {'args': '<type __builtin__.bool>',
                                          'longname': 'snap',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 's'},
                                 'snapRelative': {'args': '<type __builtin__.bool>',
                                                  'longname': 'snapRelative',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sr'},
                                 'snapValue': {'args': '<type __builtin__.float>',
                                               'longname': 'snapValue',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sv'},
                                 'tweakMode': {'args': '<type __builtin__.bool>',
                                               'longname': 'tweakMode',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'twk'}},
                       'shortFlags': {'e': 'edit',
                                      'epm': 'editPivotMode',
                                      'ex': 'exists',
                                      'i1': 'image1',
                                      'i2': 'image2',
                                      'i3': 'image3',
                                      'p': 'position',
                                      'q': 'query',
                                      's': 'snap',
                                      'sr': 'snapRelative',
                                      'sv': 'snapValue',
                                      'twk': 'tweakMode'},
                       'type': 'context'},
  'texScaleContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'editPivotMode': {'args': '<type __builtin__.bool>',
                                                  'longname': 'editPivotMode',
                                                  'modes': ['query'],
                                                  'numArgs': 1,
                                                  'shortname': 'epm'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'image1': {'args': '<type __builtin__.str>',
                                           'longname': 'image1',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i1'},
                                'image2': {'args': '<type __builtin__.str>',
                                           'longname': 'image2',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i2'},
                                'image3': {'args': '<type __builtin__.str>',
                                           'longname': 'image3',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'i3'},
                                'position': {'args': '<type __builtin__.bool>',
                                             'longname': 'position',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'p'},
                                'preventNegativeScale': {'args': '<type __builtin__.bool>',
                                                         'longname': 'preventNegativeScale',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'pns'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'snap': {'args': '<type __builtin__.bool>',
                                         'longname': 'snap',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 's'},
                                'snapRelative': {'args': '<type __builtin__.bool>',
                                                 'longname': 'snapRelative',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sr'},
                                'snapValue': {'args': '<type __builtin__.float>',
                                              'longname': 'snapValue',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sv'},
                                'tweakMode': {'args': '<type __builtin__.bool>',
                                              'longname': 'tweakMode',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'twk'}},
                      'shortFlags': {'e': 'edit',
                                     'epm': 'editPivotMode',
                                     'ex': 'exists',
                                     'i1': 'image1',
                                     'i2': 'image2',
                                     'i3': 'image3',
                                     'p': 'position',
                                     'pns': 'preventNegativeScale',
                                     'q': 'query',
                                     's': 'snap',
                                     'sr': 'snapRelative',
                                     'sv': 'snapValue',
                                     'twk': 'tweakMode'},
                      'type': 'context'},
  'texSculptCacheContext': {'flags': {'adjustSize': {'args': '<type __builtin__.bool>',
                                                     'longname': 'adjustSize',
                                                     'modes': ['edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'asz'},
                                      'adjustStrength': {'args': '<type __builtin__.bool>',
                                                         'longname': 'adjustStrength',
                                                         'modes': ['edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ast'},
                                      'direction': {'args': '<type __builtin__.int>',
                                                    'longname': 'direction',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'd'},
                                      'edit': {'args': '<type __builtin__.bool>',
                                               'longname': 'edit',
                                               'numArgs': 0,
                                               'shortname': 'e'},
                                      'exists': {'args': '<type __builtin__.bool>',
                                                 'longname': 'exists',
                                                 'numArgs': 0,
                                                 'shortname': 'ex'},
                                      'falloffType': {'args': '<type __builtin__.int>',
                                                      'longname': 'falloffType',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'ft'},
                                      'floodPin': {'args': '<type __builtin__.float>',
                                                   'longname': 'floodPin',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'fp'},
                                      'grabTwist': {'args': '<type __builtin__.bool>',
                                                    'longname': 'grabTwist',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'gtw'},
                                      'history': {'args': '<type __builtin__.bool>',
                                                  'longname': 'history',
                                                  'numArgs': 1,
                                                  'shortname': 'ch'},
                                      'image1': {'args': '<type __builtin__.str>',
                                                 'longname': 'image1',
                                                 'numArgs': 1,
                                                 'shortname': 'i1'},
                                      'image2': {'args': '<type __builtin__.str>',
                                                 'longname': 'image2',
                                                 'numArgs': 1,
                                                 'shortname': 'i2'},
                                      'image3': {'args': '<type __builtin__.str>',
                                                 'longname': 'image3',
                                                 'numArgs': 1,
                                                 'shortname': 'i3'},
                                      'inverted': {'args': '<type __builtin__.bool>',
                                                   'longname': 'inverted',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'inv'},
                                      'mode': {'args': '<type __builtin__.str>',
                                               'longname': 'mode',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'm'},
                                      'name': {'args': '<type __builtin__.str>',
                                               'longname': 'name',
                                               'numArgs': 1,
                                               'shortname': 'n'},
                                      'pinsmooth': {'args': '<type __builtin__.bool>',
                                                    'longname': 'pinsmooth',
                                                    'numArgs': 1,
                                                    'shortname': 'ps'},
                                      'pinsmoothandpaint': {'args': '<type __builtin__.bool>',
                                                            'longname': 'pinsmoothandpaint',
                                                            'numArgs': 1,
                                                            'shortname': 'psp'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'sculptFalloffCurve': {'args': '<type __builtin__.str>',
                                                             'longname': 'sculptFalloffCurve',
                                                             'modes': ['query',
                                                                       'edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'sfc'},
                                      'showBrushRingDuringStroke': {'args': '<type __builtin__.bool>',
                                                                    'longname': 'showBrushRingDuringStroke',
                                                                    'modes': ['query',
                                                                              'edit'],
                                                                    'numArgs': 1,
                                                                    'shortname': 'sbr'},
                                      'size': {'args': '<type __builtin__.float>',
                                               'longname': 'size',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sz'},
                                      'strength': {'args': '<type __builtin__.float>',
                                                   'longname': 'strength',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'st'}},
                            'shortFlags': {'ast': 'adjustStrength',
                                           'asz': 'adjustSize',
                                           'ch': 'history',
                                           'd': 'direction',
                                           'e': 'edit',
                                           'ex': 'exists',
                                           'fp': 'floodPin',
                                           'ft': 'falloffType',
                                           'gtw': 'grabTwist',
                                           'i1': 'image1',
                                           'i2': 'image2',
                                           'i3': 'image3',
                                           'inv': 'inverted',
                                           'm': 'mode',
                                           'n': 'name',
                                           'ps': 'pinsmooth',
                                           'psp': 'pinsmoothandpaint',
                                           'q': 'query',
                                           'sbr': 'showBrushRingDuringStroke',
                                           'sfc': 'sculptFalloffCurve',
                                           'st': 'strength',
                                           'sz': 'size'},
                            'type': 'context'},
  'texSculptCacheSync': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'texSelectContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'image1': {'args': '<type __builtin__.str>',
                                            'longname': 'image1',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i1'},
                                 'image2': {'args': '<type __builtin__.str>',
                                            'longname': 'image2',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i2'},
                                 'image3': {'args': '<type __builtin__.str>',
                                            'longname': 'image3',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i3'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'}},
                       'shortFlags': {'e': 'edit',
                                      'ex': 'exists',
                                      'i1': 'image1',
                                      'i2': 'image2',
                                      'i3': 'image3',
                                      'q': 'query'},
                       'type': 'context'},
  'texSelectShortestPathCtx': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                  'longname': 'edit',
                                                  'numArgs': 0,
                                                  'shortname': 'e'},
                                         'exists': {'args': '<type __builtin__.bool>',
                                                    'longname': 'exists',
                                                    'modes': ['create'],
                                                    'numArgs': 0,
                                                    'shortname': 'ex'},
                                         'image1': {'args': '<type __builtin__.str>',
                                                    'longname': 'image1',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'i1'},
                                         'image2': {'args': '<type __builtin__.str>',
                                                    'longname': 'image2',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'i2'},
                                         'image3': {'args': '<type __builtin__.str>',
                                                    'longname': 'image3',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'i3'},
                                         'query': {'args': '<type __builtin__.bool>',
                                                   'longname': 'query',
                                                   'numArgs': 0,
                                                   'shortname': 'q'}},
                               'shortFlags': {'e': 'edit',
                                              'ex': 'exists',
                                              'i1': 'image1',
                                              'i2': 'image2',
                                              'i3': 'image3',
                                              'q': 'query'},
                               'type': 'context'},
  'texSmoothContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'history': {'args': '<type __builtin__.bool>',
                                             'longname': 'history',
                                             'numArgs': 1,
                                             'shortname': 'ch'},
                                 'image1': {'args': '<type __builtin__.str>',
                                            'longname': 'image1',
                                            'numArgs': 1,
                                            'shortname': 'i1'},
                                 'image2': {'args': '<type __builtin__.str>',
                                            'longname': 'image2',
                                            'numArgs': 1,
                                            'shortname': 'i2'},
                                 'image3': {'args': '<type __builtin__.str>',
                                            'longname': 'image3',
                                            'numArgs': 1,
                                            'shortname': 'i3'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'pinBorder': {'args': '<type __builtin__.bool>',
                                               'longname': 'pinBorder',
                                               'numArgs': 1,
                                               'shortname': 'pb'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'space': {'args': '<type __builtin__.int>',
                                           'longname': 'space',
                                           'numArgs': 1,
                                           'shortname': 'sp'}},
                       'shortFlags': {'ch': 'history',
                                      'e': 'edit',
                                      'ex': 'exists',
                                      'i1': 'image1',
                                      'i2': 'image2',
                                      'i3': 'image3',
                                      'n': 'name',
                                      'pb': 'pinBorder',
                                      'q': 'query',
                                      'sp': 'space'},
                       'type': 'context'},
  'texSmudgeUVContext': {'flags': {'dragSlider': {'args': '<type __builtin__.str>',
                                                  'longname': 'dragSlider',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ds'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'effectType': {'args': '<type __builtin__.str>',
                                                  'longname': 'effectType',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'et'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'functionType': {'args': '<type __builtin__.str>',
                                                    'longname': 'functionType',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ft'},
                                   'history': {'args': '<type __builtin__.bool>',
                                               'longname': 'history',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                                   'image1': {'args': '<type __builtin__.str>',
                                              'longname': 'image1',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i1'},
                                   'image2': {'args': '<type __builtin__.str>',
                                              'longname': 'image2',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i2'},
                                   'image3': {'args': '<type __builtin__.str>',
                                              'longname': 'image3',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'i3'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'pressure': {'args': '<type __builtin__.float>',
                                                'longname': 'pressure',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'prs'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'radius': {'args': '<type __builtin__.float>',
                                              'longname': 'radius',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'r'},
                                   'smudgeIsMiddle': {'args': '<type __builtin__.bool>',
                                                      'longname': 'smudgeIsMiddle',
                                                      'modes': ['query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'sim'}},
                         'shortFlags': {'ch': 'history',
                                        'ds': 'dragSlider',
                                        'e': 'edit',
                                        'et': 'effectType',
                                        'ex': 'exists',
                                        'ft': 'functionType',
                                        'i1': 'image1',
                                        'i2': 'image2',
                                        'i3': 'image3',
                                        'n': 'name',
                                        'prs': 'pressure',
                                        'q': 'query',
                                        'r': 'radius',
                                        'sim': 'smudgeIsMiddle'},
                         'type': 'context'},
  'texTweakUVContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'exists': {'args': '<type __builtin__.bool>',
                                             'longname': 'exists',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ex'},
                                  'image1': {'args': '<type __builtin__.str>',
                                             'longname': 'image1',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i1'},
                                  'image2': {'args': '<type __builtin__.str>',
                                             'longname': 'image2',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i2'},
                                  'image3': {'args': '<type __builtin__.str>',
                                             'longname': 'image3',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'i3'},
                                  'position': {'args': '<type __builtin__.bool>',
                                               'longname': 'position',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'p'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'tolerance': {'args': '<type __builtin__.float>',
                                                'longname': 'tolerance',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 't'}},
                        'shortFlags': {'e': 'edit',
                                       'ex': 'exists',
                                       'i1': 'image1',
                                       'i2': 'image2',
                                       'i3': 'image3',
                                       'p': 'position',
                                       'q': 'query',
                                       't': 'tolerance'},
                        'type': 'context'},
  'texWinToolCtx': {'flags': {'alternateContext': {'args': '<type __builtin__.bool>',
                                                   'longname': 'alternateContext',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ac'},
                              'boxzoom': {'args': '<type __builtin__.bool>',
                                          'longname': 'boxzoom',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 0,
                                          'shortname': 'bz'},
                              'dolly': {'args': '<type __builtin__.bool>',
                                        'longname': 'dolly',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 0,
                                        'shortname': 'do'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'history': {'args': '<type __builtin__.bool>',
                                          'longname': 'history',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ch'},
                              'image1': {'args': '<type __builtin__.str>',
                                         'longname': 'image1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i1'},
                              'image2': {'args': '<type __builtin__.str>',
                                         'longname': 'image2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i2'},
                              'image3': {'args': '<type __builtin__.str>',
                                         'longname': 'image3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i3'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'toolName': {'args': '<type __builtin__.str>',
                                           'longname': 'toolName',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'tn'},
                              'track': {'args': '<type __builtin__.bool>',
                                        'longname': 'track',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 0,
                                        'shortname': 'tr'}},
                    'shortFlags': {'ac': 'alternateContext',
                                   'bz': 'boxzoom',
                                   'ch': 'history',
                                   'do': 'dolly',
                                   'e': 'edit',
                                   'ex': 'exists',
                                   'i1': 'image1',
                                   'i2': 'image2',
                                   'i3': 'image3',
                                   'n': 'name',
                                   'q': 'query',
                                   'tn': 'toolName',
                                   'tr': 'track'},
                    'type': 'context'},
  'text': {'callbackFlags': ['dropRectCallback',
                             'drc',
                             'visibleChangeCommand',
                             'vcc',
                             'dragCallback',
                             'dgc',
                             'dropCallback',
                             'dpc'],
           'flags': {'align': {'args': '<type __builtin__.str>',
                               'longname': 'align',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'al'},
                     'annotation': {'args': '<type __builtin__.str>',
                                    'longname': 'annotation',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'ann'},
                     'backgroundColor': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'backgroundColor',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 3,
                                         'shortname': 'bgc'},
                     'defineTemplate': {'args': '<type __builtin__.str>',
                                        'longname': 'defineTemplate',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'dt'},
                     'docTag': {'args': '<type __builtin__.str>',
                                'longname': 'docTag',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'dtg'},
                     'dragCallback': {'args': 'script',
                                      'longname': 'dragCallback',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dgc'},
                     'dropCallback': {'args': 'script',
                                      'longname': 'dropCallback',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dpc'},
                     'dropRectCallback': {'args': 'script',
                                          'longname': 'dropRectCallback',
                                          'modes': ['edit'],
                                          'numArgs': 1,
                                          'shortname': 'drc'},
                     'edit': {'args': '<type __builtin__.bool>',
                              'longname': 'edit',
                              'numArgs': 0,
                              'shortname': 'e'},
                     'enable': {'args': '<type __builtin__.bool>',
                                'longname': 'enable',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'en'},
                     'enableBackground': {'args': '<type __builtin__.bool>',
                                          'longname': 'enableBackground',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ebg'},
                     'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                             'longname': 'enableKeyboardFocus',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ekf'},
                     'exists': {'args': '<type __builtin__.bool>',
                                'longname': 'exists',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'ex'},
                     'font': {'args': '<type __builtin__.str>',
                              'longname': 'font',
                              'modes': ['create', 'query', 'edit'],
                              'numArgs': 1,
                              'shortname': 'fn'},
                     'fullPathName': {'args': '<type __builtin__.bool>',
                                      'longname': 'fullPathName',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'fpn'},
                     'height': {'args': '<type __builtin__.int>',
                                'longname': 'height',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'h'},
                     'highlightColor': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'highlightColor',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 3,
                                        'shortname': 'hlc'},
                     'hyperlink': {'args': '<type __builtin__.bool>',
                                   'longname': 'hyperlink',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'hl'},
                     'isObscured': {'args': '<type __builtin__.bool>',
                                    'longname': 'isObscured',
                                    'modes': ['query'],
                                    'numArgs': 0,
                                    'shortname': 'io'},
                     'label': {'args': '<type __builtin__.str>',
                               'longname': 'label',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'l'},
                     'manage': {'args': '<type __builtin__.bool>',
                                'longname': 'manage',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'm'},
                     'noBackground': {'args': '<type __builtin__.bool>',
                                      'longname': 'noBackground',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'nbg'},
                     'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                            'longname': 'numberOfPopupMenus',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'npm'},
                     'parent': {'args': '<type __builtin__.str>',
                                'longname': 'parent',
                                'modes': ['create', 'query'],
                                'numArgs': 1,
                                'shortname': 'p'},
                     'popupMenuArray': {'args': '<type __builtin__.bool>',
                                        'longname': 'popupMenuArray',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'pma'},
                     'preventOverride': {'args': '<type __builtin__.bool>',
                                         'longname': 'preventOverride',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'po'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'recomputeSize': {'args': '<type __builtin__.bool>',
                                       'longname': 'recomputeSize',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'rs'},
                     'statusBarMessage': {'args': '<type __builtin__.str>',
                                          'longname': 'statusBarMessage',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sbm'},
                     'useTemplate': {'args': '<type __builtin__.str>',
                                     'longname': 'useTemplate',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'ut'},
                     'visible': {'args': '<type __builtin__.bool>',
                                 'longname': 'visible',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'vis'},
                     'visibleChangeCommand': {'args': 'script',
                                              'longname': 'visibleChangeCommand',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'vcc'},
                     'width': {'args': '<type __builtin__.int>',
                               'longname': 'width',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 1,
                               'shortname': 'w'},
                     'wordWrap': {'args': '<type __builtin__.bool>',
                                  'longname': 'wordWrap',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'ww'}},
           'shortFlags': {'al': 'align',
                          'ann': 'annotation',
                          'bgc': 'backgroundColor',
                          'dgc': 'dragCallback',
                          'dpc': 'dropCallback',
                          'drc': 'dropRectCallback',
                          'dt': 'defineTemplate',
                          'dtg': 'docTag',
                          'e': 'edit',
                          'ebg': 'enableBackground',
                          'ekf': 'enableKeyboardFocus',
                          'en': 'enable',
                          'ex': 'exists',
                          'fn': 'font',
                          'fpn': 'fullPathName',
                          'h': 'height',
                          'hl': 'hyperlink',
                          'hlc': 'highlightColor',
                          'io': 'isObscured',
                          'l': 'label',
                          'm': 'manage',
                          'nbg': 'noBackground',
                          'npm': 'numberOfPopupMenus',
                          'p': 'parent',
                          'pma': 'popupMenuArray',
                          'po': 'preventOverride',
                          'q': 'query',
                          'rs': 'recomputeSize',
                          'sbm': 'statusBarMessage',
                          'ut': 'useTemplate',
                          'vcc': 'visibleChangeCommand',
                          'vis': 'visible',
                          'w': 'width',
                          'ww': 'wordWrap'},
           'type': 'windows'},
  'textCurves': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                       'longname': 'caching',
                                       'numArgs': 1,
                                       'shortname': 'cch'},
                           'constructionHistory': {'args': '<type __builtin__.bool>',
                                                   'longname': 'constructionHistory',
                                                   'numArgs': 1,
                                                   'shortname': 'ch'},
                           'deprecatedFontName': {'args': '<type __builtin__.bool>',
                                                  'longname': 'deprecatedFontName',
                                                  'numArgs': 1,
                                                  'shortname': 'dfn'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'font': {'args': '<type __builtin__.str>',
                                    'longname': 'font',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'f'},
                           'frozen': {'args': '<type __builtin__.bool>',
                                      'longname': 'frozen',
                                      'numArgs': 1,
                                      'shortname': 'fzn'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'nodeState': {'args': '<type __builtin__.int>',
                                         'longname': 'nodeState',
                                         'numArgs': 1,
                                         'shortname': 'nds'},
                           'object': {'args': '<type __builtin__.bool>',
                                      'longname': 'object',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'text': {'args': '<type __builtin__.str>',
                                    'longname': 'text',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 't'}},
                 'shortFlags': {'cch': 'caching',
                                'ch': 'constructionHistory',
                                'dfn': 'deprecatedFontName',
                                'e': 'edit',
                                'f': 'font',
                                'fzn': 'frozen',
                                'n': 'name',
                                'nds': 'nodeState',
                                'o': 'object',
                                'q': 'query',
                                't': 'text'},
                 'type': 'modeling'},
  'textField': {'callbackFlags': ['visibleChangeCommand',
                                  'vcc',
                                  'dragCallback',
                                  'dgc',
                                  'enterCommand',
                                  'ec',
                                  'receiveFocusCommand',
                                  'rfc',
                                  'textChangedCommand',
                                  'tcc',
                                  'changeCommand',
                                  'cc',
                                  'dropCallback',
                                  'dpc',
                                  'alwaysInvokeEnterCommandOnReturn',
                                  'aie'],
                'flags': {'alwaysInvokeEnterCommandOnReturn': {'args': '<type __builtin__.bool>',
                                                               'longname': 'alwaysInvokeEnterCommandOnReturn',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'aie'},
                          'annotation': {'args': '<type __builtin__.str>',
                                         'longname': 'annotation',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ann'},
                          'backgroundColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'backgroundColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'bgc'},
                          'changeCommand': {'args': 'script',
                                            'longname': 'changeCommand',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cc'},
                          'defineTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'defineTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dt'},
                          'disableButtons': {'args': '<type __builtin__.bool>',
                                             'longname': 'disableButtons',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'db'},
                          'disableClearButton': {'args': '<type __builtin__.bool>',
                                                 'longname': 'disableClearButton',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dcb'},
                          'disableHistoryButton': {'args': '<type __builtin__.bool>',
                                                   'longname': 'disableHistoryButton',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dhb'},
                          'docTag': {'args': '<type __builtin__.str>',
                                     'longname': 'docTag',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'dtg'},
                          'dragCallback': {'args': 'script',
                                           'longname': 'dragCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dgc'},
                          'drawInactiveFrame': {'args': '<type __builtin__.bool>',
                                                'longname': 'drawInactiveFrame',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dif'},
                          'dropCallback': {'args': 'script',
                                           'longname': 'dropCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dpc'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'editable': {'args': '<type __builtin__.bool>',
                                       'longname': 'editable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ed'},
                          'enable': {'args': '<type __builtin__.bool>',
                                     'longname': 'enable',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'en'},
                          'enableBackground': {'args': '<type __builtin__.bool>',
                                               'longname': 'enableBackground',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ebg'},
                          'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableKeyboardFocus',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ekf'},
                          'enterCommand': {'args': 'script',
                                           'longname': 'enterCommand',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ec'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'fileName': {'args': '<type __builtin__.str>',
                                       'longname': 'fileName',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'fi'},
                          'font': {'args': '<type __builtin__.str>',
                                   'longname': 'font',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'fn'},
                          'fullPathName': {'args': '<type __builtin__.bool>',
                                           'longname': 'fullPathName',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'fpn'},
                          'height': {'args': '<type __builtin__.int>',
                                     'longname': 'height',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'h'},
                          'highlightColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'highlightColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'hlc'},
                          'insertText': {'args': '<type __builtin__.str>',
                                         'longname': 'insertText',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'it'},
                          'insertionPosition': {'args': '<type __builtin__.int>',
                                                'longname': 'insertionPosition',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ip'},
                          'isObscured': {'args': '<type __builtin__.bool>',
                                         'longname': 'isObscured',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'io'},
                          'manage': {'args': '<type __builtin__.bool>',
                                     'longname': 'manage',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'm'},
                          'noBackground': {'args': '<type __builtin__.bool>',
                                           'longname': 'noBackground',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nbg'},
                          'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'numberOfPopupMenus',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'npm'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                          'placeholderText': {'args': '<type __builtin__.str>',
                                              'longname': 'placeholderText',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'pht'},
                          'popupMenuArray': {'args': '<type __builtin__.bool>',
                                             'longname': 'popupMenuArray',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'pma'},
                          'preventOverride': {'args': '<type __builtin__.bool>',
                                              'longname': 'preventOverride',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'po'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'receiveFocusCommand': {'args': 'script',
                                                  'longname': 'receiveFocusCommand',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'rfc'},
                          'searchField': {'args': '<type __builtin__.bool>',
                                          'longname': 'searchField',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'sf'},
                          'statusBarMessage': {'args': '<type __builtin__.str>',
                                               'longname': 'statusBarMessage',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sbm'},
                          'text': {'args': '<type __builtin__.str>',
                                   'longname': 'text',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'tx'},
                          'textChangedCommand': {'args': 'script',
                                                 'longname': 'textChangedCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'tcc'},
                          'useTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'useTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ut'},
                          'visible': {'args': '<type __builtin__.bool>',
                                      'longname': 'visible',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'vis'},
                          'visibleChangeCommand': {'args': 'script',
                                                   'longname': 'visibleChangeCommand',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'vcc'},
                          'width': {'args': '<type __builtin__.int>',
                                    'longname': 'width',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'w'}},
                'shortFlags': {'aie': 'alwaysInvokeEnterCommandOnReturn',
                               'ann': 'annotation',
                               'bgc': 'backgroundColor',
                               'cc': 'changeCommand',
                               'db': 'disableButtons',
                               'dcb': 'disableClearButton',
                               'dgc': 'dragCallback',
                               'dhb': 'disableHistoryButton',
                               'dif': 'drawInactiveFrame',
                               'dpc': 'dropCallback',
                               'dt': 'defineTemplate',
                               'dtg': 'docTag',
                               'e': 'edit',
                               'ebg': 'enableBackground',
                               'ec': 'enterCommand',
                               'ed': 'editable',
                               'ekf': 'enableKeyboardFocus',
                               'en': 'enable',
                               'ex': 'exists',
                               'fi': 'fileName',
                               'fn': 'font',
                               'fpn': 'fullPathName',
                               'h': 'height',
                               'hlc': 'highlightColor',
                               'io': 'isObscured',
                               'ip': 'insertionPosition',
                               'it': 'insertText',
                               'm': 'manage',
                               'nbg': 'noBackground',
                               'npm': 'numberOfPopupMenus',
                               'p': 'parent',
                               'pht': 'placeholderText',
                               'pma': 'popupMenuArray',
                               'po': 'preventOverride',
                               'q': 'query',
                               'rfc': 'receiveFocusCommand',
                               'sbm': 'statusBarMessage',
                               'sf': 'searchField',
                               'tcc': 'textChangedCommand',
                               'tx': 'text',
                               'ut': 'useTemplate',
                               'vcc': 'visibleChangeCommand',
                               'vis': 'visible',
                               'w': 'width'},
                'type': 'windows'},
  'textFieldButtonGrp': {'callbackFlags': ['visibleChangeCommand',
                                           'vcc',
                                           'dragCallback',
                                           'dgc',
                                           'buttonCommand',
                                           'bc',
                                           'textChangedCommand',
                                           'tcc',
                                           'forceChangeCommand',
                                           'fcc',
                                           'changeCommand',
                                           'cc',
                                           'dropCallback',
                                           'dpc'],
                         'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                        'longname': 'adjustableColumn',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'adj'},
                                   'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn2',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad2'},
                                   'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn3',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad3'},
                                   'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn4',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad4'},
                                   'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn5',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad5'},
                                   'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                         'longname': 'adjustableColumn6',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ad6'},
                                   'annotation': {'args': '<type __builtin__.str>',
                                                  'longname': 'annotation',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ann'},
                                   'backgroundColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'backgroundColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 3,
                                                       'shortname': 'bgc'},
                                   'buttonCommand': {'args': 'script',
                                                     'longname': 'buttonCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'bc'},
                                   'buttonLabel': {'args': '<type __builtin__.str>',
                                                   'longname': 'buttonLabel',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'bl'},
                                   'changeCommand': {'args': 'script',
                                                     'longname': 'changeCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'cc'},
                                   'columnAlign': {'args': ['<type __builtin__.int>',
                                                            '<type __builtin__.str>'],
                                                   'longname': 'columnAlign',
                                                   'modes': ['create',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 2,
                                                   'shortname': 'cal'},
                                   'columnAlign2': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign2',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 2,
                                                    'shortname': 'cl2'},
                                   'columnAlign3': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign3',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'cl3'},
                                   'columnAlign4': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign4',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 4,
                                                    'shortname': 'cl4'},
                                   'columnAlign5': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign5',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 5,
                                                    'shortname': 'cl5'},
                                   'columnAlign6': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.str>'],
                                                    'longname': 'columnAlign6',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 6,
                                                    'shortname': 'cl6'},
                                   'columnAttach': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.str>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnAttach',
                                                    'modes': ['create',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 3,
                                                    'shortname': 'cat'},
                                   'columnAttach2': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach2',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'ct2'},
                                   'columnAttach3': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach3',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'ct3'},
                                   'columnAttach4': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach4',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 4,
                                                     'shortname': 'ct4'},
                                   'columnAttach5': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach5',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 5,
                                                     'shortname': 'ct5'},
                                   'columnAttach6': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>',
                                                              '<type __builtin__.str>'],
                                                     'longname': 'columnAttach6',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 6,
                                                     'shortname': 'ct6'},
                                   'columnOffset2': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset2',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 2,
                                                     'shortname': 'co2'},
                                   'columnOffset3': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset3',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'co3'},
                                   'columnOffset4': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset4',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 4,
                                                     'shortname': 'co4'},
                                   'columnOffset5': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset5',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 5,
                                                     'shortname': 'co5'},
                                   'columnOffset6': {'args': ['<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.int>'],
                                                     'longname': 'columnOffset6',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 6,
                                                     'shortname': 'co6'},
                                   'columnWidth': {'args': ['<type __builtin__.int>',
                                                            '<type __builtin__.int>'],
                                                   'longname': 'columnWidth',
                                                   'modes': ['create',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 2,
                                                   'shortname': 'cw'},
                                   'columnWidth1': {'args': '<type __builtin__.int>',
                                                    'longname': 'columnWidth1',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'cw1'},
                                   'columnWidth2': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth2',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 2,
                                                    'shortname': 'cw2'},
                                   'columnWidth3': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth3',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 3,
                                                    'shortname': 'cw3'},
                                   'columnWidth4': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth4',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 4,
                                                    'shortname': 'cw4'},
                                   'columnWidth5': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth5',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 5,
                                                    'shortname': 'cw5'},
                                   'columnWidth6': {'args': ['<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'columnWidth6',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 6,
                                                    'shortname': 'cw6'},
                                   'defineTemplate': {'args': '<type __builtin__.str>',
                                                      'longname': 'defineTemplate',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'dt'},
                                   'docTag': {'args': '<type __builtin__.str>',
                                              'longname': 'docTag',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dtg'},
                                   'dragCallback': {'args': 'script',
                                                    'longname': 'dragCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dgc'},
                                   'dropCallback': {'args': 'script',
                                                    'longname': 'dropCallback',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dpc'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'editable': {'args': '<type __builtin__.bool>',
                                                'longname': 'editable',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ed'},
                                   'enable': {'args': '<type __builtin__.bool>',
                                              'longname': 'enable',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'en'},
                                   'enableBackground': {'args': '<type __builtin__.bool>',
                                                        'longname': 'enableBackground',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ebg'},
                                   'enableButton': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableButton',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'eb'},
                                   'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                           'longname': 'enableKeyboardFocus',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'ekf'},
                                   'exists': {'args': '<type __builtin__.bool>',
                                              'longname': 'exists',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ex'},
                                   'fileName': {'args': '<type __builtin__.str>',
                                                'longname': 'fileName',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'fi'},
                                   'forceChangeCommand': {'args': '<type __builtin__.bool>',
                                                          'longname': 'forceChangeCommand',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 0,
                                                          'shortname': 'fcc'},
                                   'fullPathName': {'args': '<type __builtin__.bool>',
                                                    'longname': 'fullPathName',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'fpn'},
                                   'height': {'args': '<type __builtin__.int>',
                                              'longname': 'height',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'h'},
                                   'highlightColor': {'args': ['<type __builtin__.float>',
                                                               '<type __builtin__.float>',
                                                               '<type __builtin__.float>'],
                                                      'longname': 'highlightColor',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 3,
                                                      'shortname': 'hlc'},
                                   'insertText': {'args': '<type __builtin__.str>',
                                                  'longname': 'insertText',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'it'},
                                   'insertionPosition': {'args': '<type __builtin__.int>',
                                                         'longname': 'insertionPosition',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'ip'},
                                   'isObscured': {'args': '<type __builtin__.bool>',
                                                  'longname': 'isObscured',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'io'},
                                   'label': {'args': '<type __builtin__.str>',
                                             'longname': 'label',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'l'},
                                   'manage': {'args': '<type __builtin__.bool>',
                                              'longname': 'manage',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'm'},
                                   'noBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'noBackground',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'nbg'},
                                   'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                          'longname': 'numberOfPopupMenus',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'npm'},
                                   'parent': {'args': '<type __builtin__.str>',
                                              'longname': 'parent',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'p'},
                                   'placeholderText': {'args': '<type __builtin__.str>',
                                                       'longname': 'placeholderText',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'pht'},
                                   'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                      'longname': 'popupMenuArray',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'pma'},
                                   'preventOverride': {'args': '<type __builtin__.bool>',
                                                       'longname': 'preventOverride',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'po'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'rowAttach': {'args': ['<type __builtin__.int>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.int>'],
                                                 'longname': 'rowAttach',
                                                 'modes': ['create',
                                                           'edit',
                                                           'multiuse'],
                                                 'numArgs': 3,
                                                 'shortname': 'rat'},
                                   'statusBarMessage': {'args': '<type __builtin__.str>',
                                                        'longname': 'statusBarMessage',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'sbm'},
                                   'text': {'args': '<type __builtin__.str>',
                                            'longname': 'text',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'tx'},
                                   'textChangedCommand': {'args': 'script',
                                                          'longname': 'textChangedCommand',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'tcc'},
                                   'useTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'useTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'ut'},
                                   'visible': {'args': '<type __builtin__.bool>',
                                               'longname': 'visible',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'vis'},
                                   'visibleChangeCommand': {'args': 'script',
                                                            'longname': 'visibleChangeCommand',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'vcc'},
                                   'width': {'args': '<type __builtin__.int>',
                                             'longname': 'width',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'w'}},
                         'shortFlags': {'ad2': 'adjustableColumn2',
                                        'ad3': 'adjustableColumn3',
                                        'ad4': 'adjustableColumn4',
                                        'ad5': 'adjustableColumn5',
                                        'ad6': 'adjustableColumn6',
                                        'adj': 'adjustableColumn',
                                        'ann': 'annotation',
                                        'bc': 'buttonCommand',
                                        'bgc': 'backgroundColor',
                                        'bl': 'buttonLabel',
                                        'cal': 'columnAlign',
                                        'cat': 'columnAttach',
                                        'cc': 'changeCommand',
                                        'cl2': 'columnAlign2',
                                        'cl3': 'columnAlign3',
                                        'cl4': 'columnAlign4',
                                        'cl5': 'columnAlign5',
                                        'cl6': 'columnAlign6',
                                        'co2': 'columnOffset2',
                                        'co3': 'columnOffset3',
                                        'co4': 'columnOffset4',
                                        'co5': 'columnOffset5',
                                        'co6': 'columnOffset6',
                                        'ct2': 'columnAttach2',
                                        'ct3': 'columnAttach3',
                                        'ct4': 'columnAttach4',
                                        'ct5': 'columnAttach5',
                                        'ct6': 'columnAttach6',
                                        'cw': 'columnWidth',
                                        'cw1': 'columnWidth1',
                                        'cw2': 'columnWidth2',
                                        'cw3': 'columnWidth3',
                                        'cw4': 'columnWidth4',
                                        'cw5': 'columnWidth5',
                                        'cw6': 'columnWidth6',
                                        'dgc': 'dragCallback',
                                        'dpc': 'dropCallback',
                                        'dt': 'defineTemplate',
                                        'dtg': 'docTag',
                                        'e': 'edit',
                                        'eb': 'enableButton',
                                        'ebg': 'enableBackground',
                                        'ed': 'editable',
                                        'ekf': 'enableKeyboardFocus',
                                        'en': 'enable',
                                        'ex': 'exists',
                                        'fcc': 'forceChangeCommand',
                                        'fi': 'fileName',
                                        'fpn': 'fullPathName',
                                        'h': 'height',
                                        'hlc': 'highlightColor',
                                        'io': 'isObscured',
                                        'ip': 'insertionPosition',
                                        'it': 'insertText',
                                        'l': 'label',
                                        'm': 'manage',
                                        'nbg': 'noBackground',
                                        'npm': 'numberOfPopupMenus',
                                        'p': 'parent',
                                        'pht': 'placeholderText',
                                        'pma': 'popupMenuArray',
                                        'po': 'preventOverride',
                                        'q': 'query',
                                        'rat': 'rowAttach',
                                        'sbm': 'statusBarMessage',
                                        'tcc': 'textChangedCommand',
                                        'tx': 'text',
                                        'ut': 'useTemplate',
                                        'vcc': 'visibleChangeCommand',
                                        'vis': 'visible',
                                        'w': 'width'},
                         'type': 'windows'},
  'textFieldGrp': {'callbackFlags': ['visibleChangeCommand',
                                     'vcc',
                                     'dragCallback',
                                     'dgc',
                                     'textChangedCommand',
                                     'tcc',
                                     'forceChangeCommand',
                                     'fcc',
                                     'changeCommand',
                                     'cc',
                                     'dropCallback',
                                     'dpc'],
                   'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'adj'},
                             'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn2',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad2'},
                             'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn3',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad3'},
                             'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn4',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad4'},
                             'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn5',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad5'},
                             'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn6',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad6'},
                             'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'changeCommand': {'args': 'script',
                                               'longname': 'changeCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cc'},
                             'columnAlign': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 2,
                                             'shortname': 'cal'},
                             'columnAlign2': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign2',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'cl2'},
                             'columnAlign3': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign3',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'cl3'},
                             'columnAlign4': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign4',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 4,
                                              'shortname': 'cl4'},
                             'columnAlign5': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign5',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 5,
                                              'shortname': 'cl5'},
                             'columnAlign6': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign6',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 6,
                                              'shortname': 'cl6'},
                             'columnAttach': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnAttach',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 3,
                                              'shortname': 'cat'},
                             'columnAttach2': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'ct2'},
                             'columnAttach3': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'ct3'},
                             'columnAttach4': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'ct4'},
                             'columnAttach5': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'ct5'},
                             'columnAttach6': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'ct6'},
                             'columnOffset2': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'co2'},
                             'columnOffset3': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'co3'},
                             'columnOffset4': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'co4'},
                             'columnOffset5': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'co5'},
                             'columnOffset6': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'co6'},
                             'columnWidth': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 2,
                                             'shortname': 'cw'},
                             'columnWidth1': {'args': '<type __builtin__.int>',
                                              'longname': 'columnWidth1',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cw1'},
                             'columnWidth2': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth2',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'cw2'},
                             'columnWidth3': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth3',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'cw3'},
                             'columnWidth4': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth4',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 4,
                                              'shortname': 'cw4'},
                             'columnWidth5': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth5',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 5,
                                              'shortname': 'cw5'},
                             'columnWidth6': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth6',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 6,
                                              'shortname': 'cw6'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'editable': {'args': '<type __builtin__.bool>',
                                          'longname': 'editable',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ed'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'fileName': {'args': '<type __builtin__.str>',
                                          'longname': 'fileName',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'fi'},
                             'forceChangeCommand': {'args': '<type __builtin__.bool>',
                                                    'longname': 'forceChangeCommand',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'fcc'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'insertText': {'args': '<type __builtin__.str>',
                                            'longname': 'insertText',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'it'},
                             'insertionPosition': {'args': '<type __builtin__.int>',
                                                   'longname': 'insertionPosition',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ip'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'label': {'args': '<type __builtin__.str>',
                                       'longname': 'label',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'l'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'placeholderText': {'args': '<type __builtin__.str>',
                                                 'longname': 'placeholderText',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'pht'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'rowAttach': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'rowAttach',
                                           'modes': ['create',
                                                     'edit',
                                                     'multiuse'],
                                           'numArgs': 3,
                                           'shortname': 'rat'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'text': {'args': '<type __builtin__.str>',
                                      'longname': 'text',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'tx'},
                             'textChangedCommand': {'args': 'script',
                                                    'longname': 'textChangedCommand',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'tcc'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'shortFlags': {'ad2': 'adjustableColumn2',
                                  'ad3': 'adjustableColumn3',
                                  'ad4': 'adjustableColumn4',
                                  'ad5': 'adjustableColumn5',
                                  'ad6': 'adjustableColumn6',
                                  'adj': 'adjustableColumn',
                                  'ann': 'annotation',
                                  'bgc': 'backgroundColor',
                                  'cal': 'columnAlign',
                                  'cat': 'columnAttach',
                                  'cc': 'changeCommand',
                                  'cl2': 'columnAlign2',
                                  'cl3': 'columnAlign3',
                                  'cl4': 'columnAlign4',
                                  'cl5': 'columnAlign5',
                                  'cl6': 'columnAlign6',
                                  'co2': 'columnOffset2',
                                  'co3': 'columnOffset3',
                                  'co4': 'columnOffset4',
                                  'co5': 'columnOffset5',
                                  'co6': 'columnOffset6',
                                  'ct2': 'columnAttach2',
                                  'ct3': 'columnAttach3',
                                  'ct4': 'columnAttach4',
                                  'ct5': 'columnAttach5',
                                  'ct6': 'columnAttach6',
                                  'cw': 'columnWidth',
                                  'cw1': 'columnWidth1',
                                  'cw2': 'columnWidth2',
                                  'cw3': 'columnWidth3',
                                  'cw4': 'columnWidth4',
                                  'cw5': 'columnWidth5',
                                  'cw6': 'columnWidth6',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ebg': 'enableBackground',
                                  'ed': 'editable',
                                  'ekf': 'enableKeyboardFocus',
                                  'en': 'enable',
                                  'ex': 'exists',
                                  'fcc': 'forceChangeCommand',
                                  'fi': 'fileName',
                                  'fpn': 'fullPathName',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'io': 'isObscured',
                                  'ip': 'insertionPosition',
                                  'it': 'insertText',
                                  'l': 'label',
                                  'm': 'manage',
                                  'nbg': 'noBackground',
                                  'npm': 'numberOfPopupMenus',
                                  'p': 'parent',
                                  'pht': 'placeholderText',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'q': 'query',
                                  'rat': 'rowAttach',
                                  'sbm': 'statusBarMessage',
                                  'tcc': 'textChangedCommand',
                                  'tx': 'text',
                                  'ut': 'useTemplate',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'w': 'width'},
                   'type': 'windows'},
  'textManip': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'visible': {'args': '<type __builtin__.bool>',
                                      'longname': 'visible',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'v'}},
                'shortFlags': {'q': 'query', 'v': 'visible'},
                'type': 'windows'},
  'textScrollList': {'callbackFlags': ['deleteKeyCommand',
                                       'dkc',
                                       'selectCommand',
                                       'sc',
                                       'doubleClickCommand',
                                       'dcc',
                                       'visibleChangeCommand',
                                       'vcc',
                                       'dragCallback',
                                       'dgc',
                                       'dropCallback',
                                       'dpc'],
                     'flags': {'allItems': {'args': '<type __builtin__.bool>',
                                            'longname': 'allItems',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ai'},
                               'allowAutomaticSelection': {'args': '<type __builtin__.bool>',
                                                           'longname': 'allowAutomaticSelection',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'aas'},
                               'allowMultiSelection': {'args': '<type __builtin__.bool>',
                                                       'longname': 'allowMultiSelection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ams'},
                               'annotation': {'args': '<type __builtin__.str>',
                                              'longname': 'annotation',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ann'},
                               'append': {'args': '<type __builtin__.str>',
                                          'longname': 'append',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'a'},
                               'appendPosition': {'args': ['<type __builtin__.int>',
                                                           '<type __builtin__.str>'],
                                                  'longname': 'appendPosition',
                                                  'modes': ['create',
                                                            'edit',
                                                            'multiuse'],
                                                  'numArgs': 2,
                                                  'shortname': 'ap'},
                               'backgroundColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'backgroundColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'bgc'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'deleteKeyCommand': {'args': 'script',
                                                    'longname': 'deleteKeyCommand',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dkc'},
                               'deselectAll': {'args': '<type __builtin__.bool>',
                                               'longname': 'deselectAll',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'da'},
                               'deselectIndexedItem': {'args': '<type __builtin__.int>',
                                                       'longname': 'deselectIndexedItem',
                                                       'modes': ['create',
                                                                 'edit',
                                                                 'multiuse'],
                                                       'numArgs': 1,
                                                       'shortname': 'dii'},
                               'deselectItem': {'args': '<type __builtin__.str>',
                                                'longname': 'deselectItem',
                                                'modes': ['create',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'di'},
                               'docTag': {'args': '<type __builtin__.str>',
                                          'longname': 'docTag',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dtg'},
                               'doubleClickCommand': {'args': 'script',
                                                      'longname': 'doubleClickCommand',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dcc'},
                               'dragCallback': {'args': 'script',
                                                'longname': 'dragCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dgc'},
                               'dropCallback': {'args': 'script',
                                                'longname': 'dropCallback',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dpc'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'enable': {'args': '<type __builtin__.bool>',
                                          'longname': 'enable',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'en'},
                               'enableBackground': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableBackground',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ebg'},
                               'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                       'longname': 'enableKeyboardFocus',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ekf'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'font': {'args': '<type __builtin__.str>',
                                        'longname': 'font',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'fn'},
                               'fullPathName': {'args': '<type __builtin__.bool>',
                                                'longname': 'fullPathName',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'fpn'},
                               'height': {'args': '<type __builtin__.int>',
                                          'longname': 'height',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'h'},
                               'highlightColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'highlightColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 3,
                                                  'shortname': 'hlc'},
                               'isObscured': {'args': '<type __builtin__.bool>',
                                              'longname': 'isObscured',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'io'},
                               'lineFont': {'args': ['<type __builtin__.int>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'lineFont',
                                            'modes': ['create',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 2,
                                            'shortname': 'lf'},
                               'manage': {'args': '<type __builtin__.bool>',
                                          'longname': 'manage',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'm'},
                               'noBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'noBackground',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nbg'},
                               'numberOfItems': {'args': '<type __builtin__.bool>',
                                                 'longname': 'numberOfItems',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'ni'},
                               'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                      'longname': 'numberOfPopupMenus',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'npm'},
                               'numberOfRows': {'args': '<type __builtin__.int>',
                                                'longname': 'numberOfRows',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'nr'},
                               'numberOfSelectedItems': {'args': '<type __builtin__.bool>',
                                                         'longname': 'numberOfSelectedItems',
                                                         'modes': ['query'],
                                                         'numArgs': 0,
                                                         'shortname': 'nsi'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                  'longname': 'popupMenuArray',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'pma'},
                               'preventOverride': {'args': '<type __builtin__.bool>',
                                                   'longname': 'preventOverride',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'po'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'removeAll': {'args': '<type __builtin__.bool>',
                                             'longname': 'removeAll',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'ra'},
                               'removeIndexedItem': {'args': '<type __builtin__.int>',
                                                     'longname': 'removeIndexedItem',
                                                     'modes': ['create',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'rii'},
                               'removeItem': {'args': '<type __builtin__.str>',
                                              'longname': 'removeItem',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'ri'},
                               'selectCommand': {'args': 'script',
                                                 'longname': 'selectCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sc'},
                               'selectIndexedItem': {'args': '<type __builtin__.int>',
                                                     'longname': 'selectIndexedItem',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'sii'},
                               'selectItem': {'args': '<type __builtin__.str>',
                                              'longname': 'selectItem',
                                              'modes': ['create',
                                                        'query',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'si'},
                               'selectUniqueTagItem': {'args': '<type __builtin__.str>',
                                                       'longname': 'selectUniqueTagItem',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit',
                                                                 'multiuse'],
                                                       'numArgs': 1,
                                                       'shortname': 'sut'},
                               'showIndexedItem': {'args': '<type __builtin__.int>',
                                                   'longname': 'showIndexedItem',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'shi'},
                               'statusBarMessage': {'args': '<type __builtin__.str>',
                                                    'longname': 'statusBarMessage',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sbm'},
                               'uniqueTag': {'args': '<type __builtin__.str>',
                                             'longname': 'uniqueTag',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'utg'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'},
                               'visible': {'args': '<type __builtin__.bool>',
                                           'longname': 'visible',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vis'},
                               'visibleChangeCommand': {'args': 'script',
                                                        'longname': 'visibleChangeCommand',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'vcc'},
                               'width': {'args': '<type __builtin__.int>',
                                         'longname': 'width',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'w'}},
                     'shortFlags': {'a': 'append',
                                    'aas': 'allowAutomaticSelection',
                                    'ai': 'allItems',
                                    'ams': 'allowMultiSelection',
                                    'ann': 'annotation',
                                    'ap': 'appendPosition',
                                    'bgc': 'backgroundColor',
                                    'da': 'deselectAll',
                                    'dcc': 'doubleClickCommand',
                                    'dgc': 'dragCallback',
                                    'di': 'deselectItem',
                                    'dii': 'deselectIndexedItem',
                                    'dkc': 'deleteKeyCommand',
                                    'dpc': 'dropCallback',
                                    'dt': 'defineTemplate',
                                    'dtg': 'docTag',
                                    'e': 'edit',
                                    'ebg': 'enableBackground',
                                    'ekf': 'enableKeyboardFocus',
                                    'en': 'enable',
                                    'ex': 'exists',
                                    'fn': 'font',
                                    'fpn': 'fullPathName',
                                    'h': 'height',
                                    'hlc': 'highlightColor',
                                    'io': 'isObscured',
                                    'lf': 'lineFont',
                                    'm': 'manage',
                                    'nbg': 'noBackground',
                                    'ni': 'numberOfItems',
                                    'npm': 'numberOfPopupMenus',
                                    'nr': 'numberOfRows',
                                    'nsi': 'numberOfSelectedItems',
                                    'p': 'parent',
                                    'pma': 'popupMenuArray',
                                    'po': 'preventOverride',
                                    'q': 'query',
                                    'ra': 'removeAll',
                                    'ri': 'removeItem',
                                    'rii': 'removeIndexedItem',
                                    'sbm': 'statusBarMessage',
                                    'sc': 'selectCommand',
                                    'shi': 'showIndexedItem',
                                    'si': 'selectItem',
                                    'sii': 'selectIndexedItem',
                                    'sut': 'selectUniqueTagItem',
                                    'ut': 'useTemplate',
                                    'utg': 'uniqueTag',
                                    'vcc': 'visibleChangeCommand',
                                    'vis': 'visible',
                                    'w': 'width'},
                     'type': 'windows'},
  'textureDeformer': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                          'longname': 'after',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'af'},
                                'afterReference': {'args': '<type __builtin__.bool>',
                                                   'longname': 'afterReference',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'ar'},
                                'before': {'args': '<type __builtin__.bool>',
                                           'longname': 'before',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'bf'},
                                'deformerTools': {'args': '<type __builtin__.bool>',
                                                  'longname': 'deformerTools',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'dt'},
                                'direction': {'args': '<type __builtin__.str>',
                                              'longname': 'direction',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'd'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'envelope': {'args': '<type __builtin__.float>',
                                             'longname': 'envelope',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'en'},
                                'exclusive': {'args': '<type __builtin__.str>',
                                              'longname': 'exclusive',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'ex'},
                                'frontOfChain': {'args': '<type __builtin__.bool>',
                                                 'longname': 'frontOfChain',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'foc'},
                                'geometry': {'args': '<type __builtin__.str>',
                                             'longname': 'geometry',
                                             'modes': ['query',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'g'},
                                'geometryIndices': {'args': '<type __builtin__.bool>',
                                                    'longname': 'geometryIndices',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'gi'},
                                'ignoreSelected': {'args': '<type __builtin__.bool>',
                                                   'longname': 'ignoreSelected',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'ignoreSelected'},
                                'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                            'longname': 'includeHiddenSelections',
                                                            'modes': ['create'],
                                                            'numArgs': 0,
                                                            'shortname': 'ihs'},
                                'name': {'args': '<type __builtin__.str>',
                                         'longname': 'name',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'n'},
                                'offset': {'args': '<type __builtin__.float>',
                                           'longname': 'offset',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'o'},
                                'parallel': {'args': '<type __builtin__.bool>',
                                             'longname': 'parallel',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'par'},
                                'pointSpace': {'args': '<type __builtin__.str>',
                                               'longname': 'pointSpace',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ps'},
                                'prune': {'args': '<type __builtin__.bool>',
                                          'longname': 'prune',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'pr'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'remove': {'args': '<type __builtin__.bool>',
                                           'longname': 'remove',
                                           'modes': ['edit', 'multiuse'],
                                           'numArgs': 0,
                                           'shortname': 'rm'},
                                'split': {'args': '<type __builtin__.bool>',
                                          'longname': 'split',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 0,
                                          'shortname': 'sp'},
                                'strength': {'args': '<type __builtin__.float>',
                                             'longname': 'strength',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 's'},
                                'vectorOffset': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'vectorOffset',
                                                 'modes': ['create'],
                                                 'numArgs': 3,
                                                 'shortname': 'vo'},
                                'vectorSpace': {'args': '<type __builtin__.str>',
                                                'longname': 'vectorSpace',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'vsp'},
                                'vectorStrength': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'vectorStrength',
                                                   'modes': ['create'],
                                                   'numArgs': 3,
                                                   'shortname': 'vs'}},
                      'removedFlags': {'is': 'ignoreSelected'},
                      'shortFlags': {'af': 'after',
                                     'ar': 'afterReference',
                                     'bf': 'before',
                                     'd': 'direction',
                                     'dt': 'deformerTools',
                                     'e': 'edit',
                                     'en': 'envelope',
                                     'ex': 'exclusive',
                                     'foc': 'frontOfChain',
                                     'g': 'geometry',
                                     'gi': 'geometryIndices',
                                     'ignoreSelected': 'ignoreSelected',
                                     'ihs': 'includeHiddenSelections',
                                     'n': 'name',
                                     'o': 'offset',
                                     'par': 'parallel',
                                     'pr': 'prune',
                                     'ps': 'pointSpace',
                                     'q': 'query',
                                     'rm': 'remove',
                                     's': 'strength',
                                     'sp': 'split',
                                     'vo': 'vectorOffset',
                                     'vs': 'vectorStrength',
                                     'vsp': 'vectorSpace'},
                      'type': 'animation'},
  'textureLassoContext': {'flags': {'drawClosed': {'args': '<type __builtin__.bool>',
                                                   'longname': 'drawClosed',
                                                   'numArgs': 1,
                                                   'shortname': 'dc'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'exists': {'args': '<type __builtin__.bool>',
                                               'longname': 'exists',
                                               'numArgs': 0,
                                               'shortname': 'ex'},
                                    'history': {'args': '<type __builtin__.bool>',
                                                'longname': 'history',
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                                    'image1': {'args': '<type __builtin__.str>',
                                               'longname': 'image1',
                                               'numArgs': 1,
                                               'shortname': 'i1'},
                                    'image2': {'args': '<type __builtin__.str>',
                                               'longname': 'image2',
                                               'numArgs': 1,
                                               'shortname': 'i2'},
                                    'image3': {'args': '<type __builtin__.str>',
                                               'longname': 'image3',
                                               'numArgs': 1,
                                               'shortname': 'i3'},
                                    'name': {'args': '<type __builtin__.str>',
                                             'longname': 'name',
                                             'numArgs': 1,
                                             'shortname': 'n'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'}},
                          'shortFlags': {'ch': 'history',
                                         'dc': 'drawClosed',
                                         'e': 'edit',
                                         'ex': 'exists',
                                         'i1': 'image1',
                                         'i2': 'image2',
                                         'i3': 'image3',
                                         'n': 'name',
                                         'q': 'query'},
                          'type': 'context'},
  'texturePlacementContext': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                                 'longname': 'edit',
                                                 'numArgs': 0,
                                                 'shortname': 'e'},
                                        'exists': {'args': '<type __builtin__.bool>',
                                                   'longname': 'exists',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'ex'},
                                        'history': {'args': '<type __builtin__.bool>',
                                                    'longname': 'history',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'ch'},
                                        'image1': {'args': '<type __builtin__.str>',
                                                   'longname': 'image1',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'i1'},
                                        'image2': {'args': '<type __builtin__.str>',
                                                   'longname': 'image2',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'i2'},
                                        'image3': {'args': '<type __builtin__.str>',
                                                   'longname': 'image3',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'i3'},
                                        'labelMapping': {'args': '<type __builtin__.bool>',
                                                         'longname': 'labelMapping',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'lm'},
                                        'name': {'args': '<type __builtin__.str>',
                                                 'longname': 'name',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'n'},
                                        'query': {'args': '<type __builtin__.bool>',
                                                  'longname': 'query',
                                                  'numArgs': 0,
                                                  'shortname': 'q'}},
                              'shortFlags': {'ch': 'history',
                                             'e': 'edit',
                                             'ex': 'exists',
                                             'i1': 'image1',
                                             'i2': 'image2',
                                             'i3': 'image3',
                                             'lm': 'labelMapping',
                                             'n': 'name',
                                             'q': 'query'},
                              'type': 'context'},
  'textureWindow': {'callbackFlags': ['changeCommand', 'cc'],
                    'flags': {'activeSelectionOnTop': {'args': '<type __builtin__.bool>',
                                                       'longname': 'activeSelectionOnTop',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ast'},
                              'axesColor': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'axesColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'axc'},
                              'backFacingColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'backFacingColor',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 4,
                                                  'shortname': 'bfc'},
                              'capture': {'args': '<type __builtin__.str>',
                                          'longname': 'capture',
                                          'modes': ['edit'],
                                          'numArgs': 1,
                                          'shortname': 'cpt'},
                              'captureSequenceNumber': {'args': '<type __builtin__.int>',
                                                        'longname': 'captureSequenceNumber',
                                                        'modes': ['edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'csn'},
                              'changeCommand': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'changeCommand',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 4,
                                                'shortname': 'cc'},
                              'checkerColor1': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'checkerColor1',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'cc1'},
                              'checkerColor2': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'checkerColor2',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'cc2'},
                              'checkerColorMode': {'args': '<type __builtin__.int>',
                                                   'longname': 'checkerColorMode',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ccm'},
                              'checkerDensity': {'args': '<type __builtin__.int>',
                                                 'longname': 'checkerDensity',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cd'},
                              'checkerDrawTileLabels': {'args': '<type __builtin__.bool>',
                                                        'longname': 'checkerDrawTileLabels',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'cdt'},
                              'checkerGradient1': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'checkerGradient1',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'cg1'},
                              'checkerGradient2': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'checkerGradient2',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'cg2'},
                              'checkerGradientOverlay': {'args': '<type __builtin__.bool>',
                                                         'longname': 'checkerGradientOverlay',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'cgo'},
                              'checkerTileLabelColor': {'args': ['<type __builtin__.float>',
                                                                 '<type __builtin__.float>',
                                                                 '<type __builtin__.float>'],
                                                        'longname': 'checkerTileLabelColor',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 3,
                                                        'shortname': 'ctc'},
                              'clearImage': {'args': '<type __builtin__.bool>',
                                             'longname': 'clearImage',
                                             'modes': ['edit'],
                                             'numArgs': 0,
                                             'shortname': 'ci'},
                              'cmEnabled': {'args': '<type __builtin__.bool>',
                                            'longname': 'cmEnabled',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cme'},
                              'control': {'args': '<type __builtin__.bool>',
                                          'longname': 'control',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'ctl'},
                              'defineTemplate': {'args': '<type __builtin__.str>',
                                                 'longname': 'defineTemplate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'dt'},
                              'displayAxes': {'args': '<type __builtin__.bool>',
                                              'longname': 'displayAxes',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dax'},
                              'displayCheckered': {'args': '<type __builtin__.bool>',
                                                   'longname': 'displayCheckered',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dct'},
                              'displayDistortion': {'args': '<type __builtin__.bool>',
                                                    'longname': 'displayDistortion',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ddt'},
                              'displayDivisionLines': {'args': '<type __builtin__.bool>',
                                                       'longname': 'displayDivisionLines',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ddl'},
                              'displayGridLines': {'args': '<type __builtin__.bool>',
                                                   'longname': 'displayGridLines',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'dgl'},
                              'displayImage': {'args': '<type __builtin__.int>',
                                               'longname': 'displayImage',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'di'},
                              'displayIsolateSelectHUD': {'args': '<type __builtin__.bool>',
                                                          'longname': 'displayIsolateSelectHUD',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'dih'},
                              'displayLabels': {'args': '<type __builtin__.bool>',
                                                'longname': 'displayLabels',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dl'},
                              'displayOverlappingUVCountHUD': {'args': '<type __builtin__.bool>',
                                                               'longname': 'displayOverlappingUVCountHUD',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 1,
                                                               'shortname': 'doh'},
                              'displayPreselection': {'args': '<type __builtin__.bool>',
                                                      'longname': 'displayPreselection',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dps'},
                              'displayReversedUVCountHUD': {'args': '<type __builtin__.bool>',
                                                            'longname': 'displayReversedUVCountHUD',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 1,
                                                            'shortname': 'drh'},
                              'displaySolidMap': {'args': '<type __builtin__.bool>',
                                                  'longname': 'displaySolidMap',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dsm'},
                              'displayStyle': {'args': '<type __builtin__.str>',
                                               'longname': 'displayStyle',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'dst'},
                              'displayTextureBorder': {'args': '<type __builtin__.bool>',
                                                       'longname': 'displayTextureBorder',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'dtb'},
                              'displayUVShellCountHUD': {'args': '<type __builtin__.bool>',
                                                         'longname': 'displayUVShellCountHUD',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'dsh'},
                              'displayUVStatisticsHUD': {'args': '<type __builtin__.bool>',
                                                         'longname': 'displayUVStatisticsHUD',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'duh'},
                              'displayUsedPercentageHUD': {'args': '<type __builtin__.bool>',
                                                           'longname': 'displayUsedPercentageHUD',
                                                           'modes': ['create',
                                                                     'query',
                                                                     'edit'],
                                                           'numArgs': 1,
                                                           'shortname': 'dph'},
                              'distortionAlpha': {'args': '<type __builtin__.float>',
                                                  'longname': 'distortionAlpha',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dta'},
                              'distortionPerObject': {'args': '<type __builtin__.bool>',
                                                      'longname': 'distortionPerObject',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dpo'},
                              'divisions': {'args': '<type __builtin__.int>',
                                            'longname': 'divisions',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'd'},
                              'docTag': {'args': '<type __builtin__.str>',
                                         'longname': 'docTag',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dtg'},
                              'doubleBuffer': {'args': '<type __builtin__.bool>',
                                               'longname': 'doubleBuffer',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 0,
                                               'shortname': 'dbf'},
                              'drawAxis': {'args': '<type __builtin__.bool>',
                                           'longname': 'drawAxis',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'da'},
                              'drawSubregions': {'args': '<type __builtin__.bool>',
                                                 'longname': 'drawSubregions',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'dsr'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'exposure': {'args': '<type __builtin__.float>',
                                           'longname': 'exposure',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'exp'},
                              'filter': {'args': '<type __builtin__.str>',
                                         'longname': 'filter',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'f'},
                              'forceMainConnection': {'args': '<type __builtin__.str>',
                                                      'longname': 'forceMainConnection',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'fmc'},
                              'forceRebake': {'args': '<type __builtin__.bool>',
                                              'longname': 'forceRebake',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'frb'},
                              'frameAll': {'args': '<type __builtin__.bool>',
                                           'longname': 'frameAll',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'fa'},
                              'frameSelected': {'args': '<type __builtin__.bool>',
                                                'longname': 'frameSelected',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'fs'},
                              'frontFacingColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'frontFacingColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 4,
                                                   'shortname': 'ffc'},
                              'gamma': {'args': '<type __builtin__.float>',
                                        'longname': 'gamma',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ga'},
                              'gridLinesColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'gridLinesColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'glc'},
                              'gridNumbersColor': {'args': ['<type __builtin__.float>',
                                                            '<type __builtin__.float>',
                                                            '<type __builtin__.float>'],
                                                   'longname': 'gridNumbersColor',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 3,
                                                   'shortname': 'gnc'},
                              'highlightConnection': {'args': '<type __builtin__.str>',
                                                      'longname': 'highlightConnection',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'hlc'},
                              'imageBaseColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'imageBaseColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'ibc'},
                              'imageDim': {'args': '<type __builtin__.bool>',
                                           'longname': 'imageDim',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'idm'},
                              'imageDisplay': {'args': '<type __builtin__.bool>',
                                               'longname': 'imageDisplay',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'id'},
                              'imageNames': {'args': '<type __builtin__.bool>',
                                             'longname': 'imageNames',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'imn'},
                              'imageNumber': {'args': '<type __builtin__.int>',
                                              'longname': 'imageNumber',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'imageNumber'},
                              'imagePixelSnap': {'args': '<type __builtin__.bool>',
                                                 'longname': 'imagePixelSnap',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ip'},
                              'imageRatio': {'args': '<type __builtin__.bool>',
                                             'longname': 'imageRatio',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'imr'},
                              'imageRatioValue': {'args': '<type __builtin__.float>',
                                                  'longname': 'imageRatioValue',
                                                  'modes': ['query'],
                                                  'numArgs': 1,
                                                  'shortname': 'irv'},
                              'imageSize': {'args': '<type __builtin__.bool>',
                                            'longname': 'imageSize',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'imageSize'},
                              'imageTileRange': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'imageTileRange',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 4,
                                                 'shortname': 'itr'},
                              'imageToTextureNumber': {'args': '<type __builtin__.int>',
                                                       'longname': 'imageToTextureNumber',
                                                       'numArgs': 1,
                                                       'shortname': 'itn'},
                              'imageUnfiltered': {'args': '<type __builtin__.bool>',
                                                  'longname': 'imageUnfiltered',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'iuf'},
                              'internalFaces': {'args': '<type __builtin__.bool>',
                                                'longname': 'internalFaces',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'internalFaces'},
                              'isolateSelectSetUpdated': {'args': '<type __builtin__.bool>',
                                                          'longname': 'isolateSelectSetUpdated',
                                                          'numArgs': 0,
                                                          'shortname': 'isu'},
                              'labelPosition': {'args': '<type __builtin__.str>',
                                                'longname': 'labelPosition',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'lp'},
                              'loadImage': {'args': '<type __builtin__.str>',
                                            'longname': 'loadImage',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'li'},
                              'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                     'longname': 'lockMainConnection',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'lck'},
                              'mainListConnection': {'args': '<type __builtin__.str>',
                                                     'longname': 'mainListConnection',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'mlc'},
                              'maxResolution': {'args': '<type __builtin__.int>',
                                                'longname': 'maxResolution',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'mrs'},
                              'multiColorAlpha': {'args': '<type __builtin__.float>',
                                                  'longname': 'multiColorAlpha',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mca'},
                              'nbImages': {'args': '<type __builtin__.bool>',
                                           'longname': 'nbImages',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'nim'},
                              'nextView': {'args': '<type __builtin__.bool>',
                                           'longname': 'nextView',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'nv'},
                              'numUvSets': {'args': '<type __builtin__.bool>',
                                            'longname': 'numUvSets',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 0,
                                            'shortname': 'nuv'},
                              'numberOfImages': {'args': '<type __builtin__.int>',
                                                 'longname': 'numberOfImages',
                                                 'modes': ['query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ni'},
                              'numberOfTextures': {'args': '<type __builtin__.int>',
                                                   'longname': 'numberOfTextures',
                                                   'modes': ['query'],
                                                   'numArgs': 1,
                                                   'shortname': 'nt'},
                              'panel': {'args': '<type __builtin__.str>',
                                        'longname': 'panel',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'pnl'},
                              'parent': {'args': '<type __builtin__.str>',
                                         'longname': 'parent',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'p'},
                              'previousView': {'args': '<type __builtin__.bool>',
                                               'longname': 'previousView',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'pv'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'realSize': {'args': '<type __builtin__.bool>',
                                           'longname': 'realSize',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'rs'},
                              'refresh': {'args': '<type __builtin__.bool>',
                                          'longname': 'refresh',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'ref'},
                              'relatedFaces': {'args': '<type __builtin__.bool>',
                                               'longname': 'relatedFaces',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rf'},
                              'removeAllImages': {'args': '<type __builtin__.bool>',
                                                  'longname': 'removeAllImages',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'ra'},
                              'removeImage': {'args': '<type __builtin__.bool>',
                                              'longname': 'removeImage',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'ri'},
                              'rendererString': {'args': '<type __builtin__.str>',
                                                 'longname': 'rendererString',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rds'},
                              'reset': {'args': '<type __builtin__.bool>',
                                        'longname': 'reset',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'r'},
                              'saveImage': {'args': '<type __builtin__.bool>',
                                            'longname': 'saveImage',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'si'},
                              'scaleBlue': {'args': '<type __builtin__.float>',
                                            'longname': 'scaleBlue',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'sb'},
                              'scaleGreen': {'args': '<type __builtin__.float>',
                                             'longname': 'scaleGreen',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sg'},
                              'scaleRed': {'args': '<type __builtin__.float>',
                                           'longname': 'scaleRed',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sr'},
                              'selectInternalFaces': {'args': '<type __builtin__.bool>',
                                                      'longname': 'selectInternalFaces',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'sif'},
                              'selectRelatedFaces': {'args': '<type __builtin__.bool>',
                                                     'longname': 'selectRelatedFaces',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'srf'},
                              'selectionConnection': {'args': '<type __builtin__.str>',
                                                      'longname': 'selectionConnection',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'slc'},
                              'setUvSet': {'args': '<type __builtin__.int>',
                                           'longname': 'setUvSet',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'suv'},
                              'singleBuffer': {'args': '<type __builtin__.bool>',
                                               'longname': 'singleBuffer',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 0,
                                               'shortname': 'sbf'},
                              'size': {'args': '<type __builtin__.float>',
                                       'longname': 'size',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 's'},
                              'solidMap3dView': {'args': '<type __builtin__.bool>',
                                                 'longname': 'solidMap3dView',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 's3d'},
                              'solidMapColorSeed': {'args': '<type __builtin__.int>',
                                                    'longname': 'solidMapColorSeed',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'scs'},
                              'solidMapPerShell': {'args': '<type __builtin__.bool>',
                                                   'longname': 'solidMapPerShell',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'sps'},
                              'spacing': {'args': '<type __builtin__.float>',
                                          'longname': 'spacing',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'sp'},
                              'stateString': {'args': '<type __builtin__.bool>',
                                              'longname': 'stateString',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'sts'},
                              'style': {'args': '<type __builtin__.int>',
                                        'longname': 'style',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'st'},
                              'subdivisionLinesColor': {'args': ['<type __builtin__.float>',
                                                                 '<type __builtin__.float>',
                                                                 '<type __builtin__.float>'],
                                                        'longname': 'subdivisionLinesColor',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 3,
                                                        'shortname': 'sdc'},
                              'textureBorder3dView': {'args': '<type __builtin__.bool>',
                                                      'longname': 'textureBorder3dView',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 't3d'},
                              'textureBorderColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'textureBorderColor',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 3,
                                                     'shortname': 'tbc'},
                              'textureBorderWidth': {'args': '<type __builtin__.int>',
                                                     'longname': 'textureBorderWidth',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'tbw'},
                              'textureNames': {'args': '<type __builtin__.bool>',
                                               'longname': 'textureNames',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'txn'},
                              'textureNumber': {'args': '<type __builtin__.int>',
                                                'longname': 'textureNumber',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'tn'},
                              'textureToImageNumber': {'args': '<type __builtin__.int>',
                                                       'longname': 'textureToImageNumber',
                                                       'numArgs': 1,
                                                       'shortname': 'tin'},
                              'tileLabels': {'args': '<type __builtin__.bool>',
                                             'longname': 'tileLabels',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'tlb'},
                              'tileLinesColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'tileLinesColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'tlc'},
                              'toggle': {'args': '<type __builtin__.bool>',
                                         'longname': 'toggle',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'tgl'},
                              'toggleExposure': {'args': '<type __builtin__.bool>',
                                                 'longname': 'toggleExposure',
                                                 'modes': ['edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'tge'},
                              'toggleGamma': {'args': '<type __builtin__.bool>',
                                              'longname': 'toggleGamma',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'tgg'},
                              'unParent': {'args': '<type __builtin__.bool>',
                                           'longname': 'unParent',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 0,
                                           'shortname': 'up'},
                              'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                       'longname': 'unlockMainConnection',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'ulk'},
                              'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                       'longname': 'updateMainConnection',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'upd'},
                              'useFaceGroup': {'args': '<type __builtin__.bool>',
                                               'longname': 'useFaceGroup',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'uf'},
                              'useTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'useTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ut'},
                              'usedPercentageHUDRange': {'args': ['<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>',
                                                                  '<type __builtin__.float>'],
                                                         'longname': 'usedPercentageHUDRange',
                                                         'modes': ['create',
                                                                   'query',
                                                                   'edit'],
                                                         'numArgs': 4,
                                                         'shortname': 'upr'},
                              'uvSets': {'args': '<type __builtin__.bool>',
                                         'longname': 'uvSets',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 0,
                                         'shortname': 'uvs'},
                              'viewPortImage': {'args': '<type __builtin__.bool>',
                                                'longname': 'viewPortImage',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'vpi'},
                              'viewTransformName': {'args': '<type __builtin__.str>',
                                                    'longname': 'viewTransformName',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'vtn'},
                              'wireframeComponentColor': {'args': ['<type __builtin__.float>',
                                                                   '<type __builtin__.float>',
                                                                   '<type __builtin__.float>',
                                                                   '<type __builtin__.float>'],
                                                          'longname': 'wireframeComponentColor',
                                                          'modes': ['create',
                                                                    'query',
                                                                    'edit'],
                                                          'numArgs': 4,
                                                          'shortname': 'wcc'},
                              'wireframeObjectColor': {'args': ['<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'wireframeObjectColor',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 4,
                                                       'shortname': 'woc'},
                              'writeImage': {'args': '<type __builtin__.str>',
                                             'longname': 'writeImage',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'wi'}},
                    'removedFlags': {'if': 'internalFaces',
                                     'in': 'imageNumber',
                                     'is': 'imageSize'},
                    'shortFlags': {'ast': 'activeSelectionOnTop',
                                   'axc': 'axesColor',
                                   'bfc': 'backFacingColor',
                                   'cc': 'changeCommand',
                                   'cc1': 'checkerColor1',
                                   'cc2': 'checkerColor2',
                                   'ccm': 'checkerColorMode',
                                   'cd': 'checkerDensity',
                                   'cdt': 'checkerDrawTileLabels',
                                   'cg1': 'checkerGradient1',
                                   'cg2': 'checkerGradient2',
                                   'cgo': 'checkerGradientOverlay',
                                   'ci': 'clearImage',
                                   'cme': 'cmEnabled',
                                   'cpt': 'capture',
                                   'csn': 'captureSequenceNumber',
                                   'ctc': 'checkerTileLabelColor',
                                   'ctl': 'control',
                                   'd': 'divisions',
                                   'da': 'drawAxis',
                                   'dax': 'displayAxes',
                                   'dbf': 'doubleBuffer',
                                   'dct': 'displayCheckered',
                                   'ddl': 'displayDivisionLines',
                                   'ddt': 'displayDistortion',
                                   'dgl': 'displayGridLines',
                                   'di': 'displayImage',
                                   'dih': 'displayIsolateSelectHUD',
                                   'dl': 'displayLabels',
                                   'doh': 'displayOverlappingUVCountHUD',
                                   'dph': 'displayUsedPercentageHUD',
                                   'dpo': 'distortionPerObject',
                                   'dps': 'displayPreselection',
                                   'drh': 'displayReversedUVCountHUD',
                                   'dsh': 'displayUVShellCountHUD',
                                   'dsm': 'displaySolidMap',
                                   'dsr': 'drawSubregions',
                                   'dst': 'displayStyle',
                                   'dt': 'defineTemplate',
                                   'dta': 'distortionAlpha',
                                   'dtb': 'displayTextureBorder',
                                   'dtg': 'docTag',
                                   'duh': 'displayUVStatisticsHUD',
                                   'e': 'edit',
                                   'ex': 'exists',
                                   'exp': 'exposure',
                                   'f': 'filter',
                                   'fa': 'frameAll',
                                   'ffc': 'frontFacingColor',
                                   'fmc': 'forceMainConnection',
                                   'frb': 'forceRebake',
                                   'fs': 'frameSelected',
                                   'ga': 'gamma',
                                   'glc': 'gridLinesColor',
                                   'gnc': 'gridNumbersColor',
                                   'hlc': 'highlightConnection',
                                   'ibc': 'imageBaseColor',
                                   'id': 'imageDisplay',
                                   'idm': 'imageDim',
                                   'imageNumber': 'imageNumber',
                                   'imageSize': 'imageSize',
                                   'imn': 'imageNames',
                                   'imr': 'imageRatio',
                                   'internalFaces': 'internalFaces',
                                   'ip': 'imagePixelSnap',
                                   'irv': 'imageRatioValue',
                                   'isu': 'isolateSelectSetUpdated',
                                   'itn': 'imageToTextureNumber',
                                   'itr': 'imageTileRange',
                                   'iuf': 'imageUnfiltered',
                                   'lck': 'lockMainConnection',
                                   'li': 'loadImage',
                                   'lp': 'labelPosition',
                                   'mca': 'multiColorAlpha',
                                   'mlc': 'mainListConnection',
                                   'mrs': 'maxResolution',
                                   'ni': 'numberOfImages',
                                   'nim': 'nbImages',
                                   'nt': 'numberOfTextures',
                                   'nuv': 'numUvSets',
                                   'nv': 'nextView',
                                   'p': 'parent',
                                   'pnl': 'panel',
                                   'pv': 'previousView',
                                   'q': 'query',
                                   'r': 'reset',
                                   'ra': 'removeAllImages',
                                   'rds': 'rendererString',
                                   'ref': 'refresh',
                                   'rf': 'relatedFaces',
                                   'ri': 'removeImage',
                                   'rs': 'realSize',
                                   's': 'size',
                                   's3d': 'solidMap3dView',
                                   'sb': 'scaleBlue',
                                   'sbf': 'singleBuffer',
                                   'scs': 'solidMapColorSeed',
                                   'sdc': 'subdivisionLinesColor',
                                   'sg': 'scaleGreen',
                                   'si': 'saveImage',
                                   'sif': 'selectInternalFaces',
                                   'slc': 'selectionConnection',
                                   'sp': 'spacing',
                                   'sps': 'solidMapPerShell',
                                   'sr': 'scaleRed',
                                   'srf': 'selectRelatedFaces',
                                   'st': 'style',
                                   'sts': 'stateString',
                                   'suv': 'setUvSet',
                                   't3d': 'textureBorder3dView',
                                   'tbc': 'textureBorderColor',
                                   'tbw': 'textureBorderWidth',
                                   'tge': 'toggleExposure',
                                   'tgg': 'toggleGamma',
                                   'tgl': 'toggle',
                                   'tin': 'textureToImageNumber',
                                   'tlb': 'tileLabels',
                                   'tlc': 'tileLinesColor',
                                   'tn': 'textureNumber',
                                   'txn': 'textureNames',
                                   'uf': 'useFaceGroup',
                                   'ulk': 'unlockMainConnection',
                                   'up': 'unParent',
                                   'upd': 'updateMainConnection',
                                   'upr': 'usedPercentageHUDRange',
                                   'ut': 'useTemplate',
                                   'uvs': 'uvSets',
                                   'vpi': 'viewPortImage',
                                   'vtn': 'viewTransformName',
                                   'wcc': 'wireframeComponentColor',
                                   'wi': 'writeImage',
                                   'woc': 'wireframeObjectColor'},
                    'type': 'rendering'},
  'threadCount': {'flags': {'numberOfThreads': {'args': '<type __builtin__.int>',
                                                'longname': 'numberOfThreads',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'n'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'}},
                  'shortFlags': {'n': 'numberOfThreads', 'q': 'query'},
                  'type': 'general'},
  'threePointArcCtx': {'flags': {'degree': {'args': '<type __builtin__.int>',
                                            'longname': 'degree',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'd'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'history': {'args': '<type __builtin__.bool>',
                                             'longname': 'history',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ch'},
                                 'image1': {'args': '<type __builtin__.str>',
                                            'longname': 'image1',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i1'},
                                 'image2': {'args': '<type __builtin__.str>',
                                            'longname': 'image2',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i2'},
                                 'image3': {'args': '<type __builtin__.str>',
                                            'longname': 'image3',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'i3'},
                                 'name': {'args': '<type __builtin__.str>',
                                          'longname': 'name',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'n'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'spans': {'args': '<type __builtin__.int>',
                                           'longname': 'spans',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 's'}},
                       'shortFlags': {'ch': 'history',
                                      'd': 'degree',
                                      'e': 'edit',
                                      'ex': 'exists',
                                      'i1': 'image1',
                                      'i2': 'image2',
                                      'i3': 'image3',
                                      'n': 'name',
                                      'q': 'query',
                                      's': 'spans'},
                       'type': 'context'},
  'thumbnailCaptureComponent': {'flags': {'capture': {'args': '<type __builtin__.bool>',
                                                      'longname': 'capture',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'c'},
                                          'capturedFrameCount': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'capturedFrameCount',
                                                                 'modes': ['query'],
                                                                 'numArgs': 0,
                                                                 'shortname': 'cfc'},
                                          'closeCurrentSession': {'args': '<type __builtin__.bool>',
                                                                  'longname': 'closeCurrentSession',
                                                                  'modes': ['create'],
                                                                  'numArgs': 0,
                                                                  'shortname': 'ccs'},
                                          'delete': {'args': '<type __builtin__.bool>',
                                                     'longname': 'delete',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'd'},
                                          'endFrame': {'args': '<type __builtin__.int>',
                                                       'longname': 'endFrame',
                                                       'modes': ['create',
                                                                 'query'],
                                                       'numArgs': 1,
                                                       'shortname': 'ef'},
                                          'fileDialogCallback': {'args': '<type __builtin__.str>',
                                                                 'longname': 'fileDialogCallback',
                                                                 'modes': ['create'],
                                                                 'numArgs': 1,
                                                                 'shortname': 'fdc'},
                                          'fileDialogProcessing': {'args': '<type __builtin__.bool>',
                                                                   'longname': 'fileDialogProcessing',
                                                                   'numArgs': 1,
                                                                   'shortname': 'fdp'},
                                          'isSessionOpened': {'args': '<type __builtin__.bool>',
                                                              'longname': 'isSessionOpened',
                                                              'modes': ['query'],
                                                              'numArgs': 0,
                                                              'shortname': 'iso'},
                                          'launchedFromOptionsBox': {'args': '<type __builtin__.bool>',
                                                                     'longname': 'launchedFromOptionsBox',
                                                                     'modes': ['query'],
                                                                     'numArgs': 0,
                                                                     'shortname': 'lfo'},
                                          'previewPath': {'args': '<type __builtin__.bool>',
                                                          'longname': 'previewPath',
                                                          'modes': ['query'],
                                                          'numArgs': 1,
                                                          'shortname': 'pp'},
                                          'query': {'args': '<type __builtin__.bool>',
                                                    'longname': 'query',
                                                    'numArgs': 0,
                                                    'shortname': 'q'},
                                          'removeProjectThumbnail': {'args': '<type __builtin__.str>',
                                                                     'longname': 'removeProjectThumbnail',
                                                                     'modes': ['create'],
                                                                     'numArgs': 1,
                                                                     'shortname': 'rpt'},
                                          'save': {'args': '<type __builtin__.str>',
                                                   'longname': 'save',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 's'},
                                          'selectedFileName': {'args': '<type __builtin__.bool>',
                                                               'longname': 'selectedFileName',
                                                               'numArgs': 0,
                                                               'shortname': 'sfn'},
                                          'startFrame': {'args': '<type __builtin__.int>',
                                                         'longname': 'startFrame',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 1,
                                                         'shortname': 'sf'}},
                                'shortFlags': {'c': 'capture',
                                               'ccs': 'closeCurrentSession',
                                               'cfc': 'capturedFrameCount',
                                               'd': 'delete',
                                               'ef': 'endFrame',
                                               'fdc': 'fileDialogCallback',
                                               'fdp': 'fileDialogProcessing',
                                               'iso': 'isSessionOpened',
                                               'lfo': 'launchedFromOptionsBox',
                                               'pp': 'previewPath',
                                               'q': 'query',
                                               'rpt': 'removeProjectThumbnail',
                                               's': 'save',
                                               'sf': 'startFrame',
                                               'sfn': 'selectedFileName'},
                                'type': 'windows'},
  'timeCode': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'mayaStartFrame': {'args': '<type __builtin__.float>',
                                            'longname': 'mayaStartFrame',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'msf'},
                         'productionStartFrame': {'args': '<type __builtin__.float>',
                                                  'longname': 'productionStartFrame',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'psf'},
                         'productionStartHour': {'args': '<type __builtin__.float>',
                                                 'longname': 'productionStartHour',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'psh'},
                         'productionStartMinute': {'args': '<type __builtin__.float>',
                                                   'longname': 'productionStartMinute',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'psm'},
                         'productionStartSecond': {'args': '<type __builtin__.float>',
                                                   'longname': 'productionStartSecond',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'pss'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'}},
               'shortFlags': {'e': 'edit',
                              'msf': 'mayaStartFrame',
                              'psf': 'productionStartFrame',
                              'psh': 'productionStartHour',
                              'psm': 'productionStartMinute',
                              'pss': 'productionStartSecond',
                              'q': 'query'},
               'type': 'general'},
  'timeControl': {'callbackFlags': ['dragCallback',
                                    'dgc',
                                    'releaseCommand',
                                    'rc',
                                    'visibleChangeCommand',
                                    'vcc',
                                    'pressCommand',
                                    'pc',
                                    'dropCallback',
                                    'dpc'],
                  'flags': {'animCurveNames': {'args': '<type __builtin__.bool>',
                                               'longname': 'animCurveNames',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 0,
                                               'shortname': 'acn'},
                            'animLayerFilterOptions': {'args': '<type __builtin__.str>',
                                                       'longname': 'animLayerFilterOptions',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'alf'},
                            'animLayerShowWeight': {'args': '<type __builtin__.bool>',
                                                    'longname': 'animLayerShowWeight',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'asw'},
                            'annotation': {'args': '<type __builtin__.str>',
                                           'longname': 'annotation',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ann'},
                            'backgroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'backgroundColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'bgc'},
                            'beginScrub': {'args': '<type __builtin__.bool>',
                                           'longname': 'beginScrub',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'bs'},
                            'currentFrameColor': {'args': ['<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'currentFrameColor',
                                                  'modes': ['edit'],
                                                  'numArgs': 4,
                                                  'shortname': 'cfc'},
                            'defineTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'defineTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'dt'},
                            'displaySound': {'args': '<type __builtin__.bool>',
                                             'longname': 'displaySound',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ds'},
                            'docTag': {'args': '<type __builtin__.str>',
                                       'longname': 'docTag',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dtg'},
                            'dragCallback': {'args': 'script',
                                             'longname': 'dragCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dgc'},
                            'dropCallback': {'args': 'script',
                                             'longname': 'dropCallback',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dpc'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'enable': {'args': '<type __builtin__.bool>',
                                       'longname': 'enable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'en'},
                            'enableBackground': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableBackground',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ebg'},
                            'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'enableKeyboardFocus',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ekf'},
                            'endScrub': {'args': '<type __builtin__.bool>',
                                         'longname': 'endScrub',
                                         'modes': ['edit'],
                                         'numArgs': 0,
                                         'shortname': 'es'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'forceRedraw': {'args': '<type __builtin__.bool>',
                                            'longname': 'forceRedraw',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'fd'},
                            'forceRefresh': {'args': '<type __builtin__.bool>',
                                             'longname': 'forceRefresh',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'fr'},
                            'foregroundColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'foregroundColor',
                                                'modes': ['edit'],
                                                'numArgs': 3,
                                                'shortname': 'fgc'},
                            'fullPathName': {'args': '<type __builtin__.bool>',
                                             'longname': 'fullPathName',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fpn'},
                            'globalTime': {'args': '<type __builtin__.bool>',
                                           'longname': 'globalTime',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'gt'},
                            'greasePencilSequenceNames': {'args': '<type __builtin__.bool>',
                                                          'longname': 'greasePencilSequenceNames',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'gn'},
                            'height': {'args': '<type __builtin__.int>',
                                       'longname': 'height',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'h'},
                            'highlightColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'highlightColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'hlc'},
                            'isObscured': {'args': '<type __builtin__.bool>',
                                           'longname': 'isObscured',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'io'},
                            'mainListConnection': {'args': '<type __builtin__.str>',
                                                   'longname': 'mainListConnection',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'mlc'},
                            'manage': {'args': '<type __builtin__.bool>',
                                       'longname': 'manage',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'm'},
                            'noBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'noBackground',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nbg'},
                            'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'numberOfPopupMenus',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'npm'},
                            'outsideSpacing': {'args': '<type __builtin__.int>',
                                               'longname': 'outsideSpacing',
                                               'numArgs': 1,
                                               'shortname': 'os'},
                            'parent': {'args': '<type __builtin__.str>',
                                       'longname': 'parent',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'p'},
                            'popupMenuArray': {'args': '<type __builtin__.bool>',
                                               'longname': 'popupMenuArray',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'pma'},
                            'pressCommand': {'args': 'script',
                                             'longname': 'pressCommand',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'pc'},
                            'preventOverride': {'args': '<type __builtin__.bool>',
                                                'longname': 'preventOverride',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'po'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'range': {'args': '<type __builtin__.bool>',
                                      'longname': 'range',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'rng'},
                            'rangeArray': {'args': '<type __builtin__.bool>',
                                           'longname': 'rangeArray',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'ra'},
                            'rangeVisible': {'args': '<type __builtin__.bool>',
                                             'longname': 'rangeVisible',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'rv'},
                            'releaseCommand': {'args': 'script',
                                               'longname': 'releaseCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rc'},
                            'repeatChunkSize': {'args': '<type __builtin__.float>',
                                                'longname': 'repeatChunkSize',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'rcs'},
                            'repeatOnHold': {'args': '<type __builtin__.bool>',
                                             'longname': 'repeatOnHold',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'roh'},
                            'resample': {'args': '<type __builtin__.bool>',
                                         'longname': 'resample',
                                         'modes': ['edit'],
                                         'numArgs': 0,
                                         'shortname': 'r'},
                            'showGreaseFrames': {'args': '<type __builtin__.str>',
                                                 'longname': 'showGreaseFrames',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sg'},
                            'showKeys': {'args': '<type __builtin__.str>',
                                         'longname': 'showKeys',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'sk'},
                            'showKeysCombined': {'args': '<type __builtin__.bool>',
                                                 'longname': 'showKeysCombined',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'skc'},
                            'snap': {'args': '<type __builtin__.bool>',
                                     'longname': 'snap',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'sn'},
                            'sound': {'args': '<type __builtin__.str>',
                                      'longname': 'sound',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                            'statusBarMessage': {'args': '<type __builtin__.str>',
                                                 'longname': 'statusBarMessage',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'sbm'},
                            'tickSize': {'args': '<type __builtin__.int>',
                                         'longname': 'tickSize',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ts'},
                            'tickSpan': {'args': '<type __builtin__.int>',
                                         'longname': 'tickSpan',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'tsp'},
                            'useTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'useTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ut'},
                            'visible': {'args': '<type __builtin__.bool>',
                                        'longname': 'visible',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vis'},
                            'visibleChangeCommand': {'args': 'script',
                                                     'longname': 'visibleChangeCommand',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'vcc'},
                            'waveform': {'args': '<type __builtin__.str>',
                                         'longname': 'waveform',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'wf'},
                            'width': {'args': '<type __builtin__.int>',
                                      'longname': 'width',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'w'}},
                  'shortFlags': {'acn': 'animCurveNames',
                                 'alf': 'animLayerFilterOptions',
                                 'ann': 'annotation',
                                 'asw': 'animLayerShowWeight',
                                 'bgc': 'backgroundColor',
                                 'bs': 'beginScrub',
                                 'cfc': 'currentFrameColor',
                                 'dgc': 'dragCallback',
                                 'dpc': 'dropCallback',
                                 'ds': 'displaySound',
                                 'dt': 'defineTemplate',
                                 'dtg': 'docTag',
                                 'e': 'edit',
                                 'ebg': 'enableBackground',
                                 'ekf': 'enableKeyboardFocus',
                                 'en': 'enable',
                                 'es': 'endScrub',
                                 'ex': 'exists',
                                 'fd': 'forceRedraw',
                                 'fgc': 'foregroundColor',
                                 'fpn': 'fullPathName',
                                 'fr': 'forceRefresh',
                                 'gn': 'greasePencilSequenceNames',
                                 'gt': 'globalTime',
                                 'h': 'height',
                                 'hlc': 'highlightColor',
                                 'io': 'isObscured',
                                 'm': 'manage',
                                 'mlc': 'mainListConnection',
                                 'nbg': 'noBackground',
                                 'npm': 'numberOfPopupMenus',
                                 'os': 'outsideSpacing',
                                 'p': 'parent',
                                 'pc': 'pressCommand',
                                 'pma': 'popupMenuArray',
                                 'po': 'preventOverride',
                                 'q': 'query',
                                 'r': 'resample',
                                 'ra': 'rangeArray',
                                 'rc': 'releaseCommand',
                                 'rcs': 'repeatChunkSize',
                                 'rng': 'range',
                                 'roh': 'repeatOnHold',
                                 'rv': 'rangeVisible',
                                 's': 'sound',
                                 'sbm': 'statusBarMessage',
                                 'sg': 'showGreaseFrames',
                                 'sk': 'showKeys',
                                 'skc': 'showKeysCombined',
                                 'sn': 'snap',
                                 'ts': 'tickSize',
                                 'tsp': 'tickSpan',
                                 'ut': 'useTemplate',
                                 'vcc': 'visibleChangeCommand',
                                 'vis': 'visible',
                                 'w': 'width',
                                 'wf': 'waveform'},
                  'type': 'windows'},
  'timeEditor': {'flags': {'allClips': {'args': '<type __builtin__.str>',
                                        'longname': 'allClips',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'alc'},
                           'clipId': {'args': '<type __builtin__.int>',
                                      'longname': 'clipId',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 1,
                                      'shortname': 'id'},
                           'commonParentTrack': {'args': '<type __builtin__.bool>',
                                                 'longname': 'commonParentTrack',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'cpt'},
                           'composition': {'args': '<type __builtin__.str>',
                                           'longname': 'composition',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'cp'},
                           'drivingClipsForAttr': {'args': '<type __builtin__.str>',
                                                   'longname': 'drivingClipsForAttr',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'dca'},
                           'drivingClipsForObj': {'args': ['<type __builtin__.str>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'drivingClipsForObj',
                                                  'modes': ['create'],
                                                  'numArgs': 2,
                                                  'shortname': 'dco'},
                           'includeParent': {'args': '<type __builtin__.bool>',
                                             'longname': 'includeParent',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'ip'},
                           'mute': {'args': '<type __builtin__.bool>',
                                    'longname': 'mute',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'm'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'selectedClips': {'args': '<type __builtin__.str>',
                                             'longname': 'selectedClips',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'sc'},
                           'topLevelClips': {'args': '<type __builtin__.str>',
                                             'longname': 'topLevelClips',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'tlc'}},
                 'shortFlags': {'alc': 'allClips',
                                'cp': 'composition',
                                'cpt': 'commonParentTrack',
                                'dca': 'drivingClipsForAttr',
                                'dco': 'drivingClipsForObj',
                                'id': 'clipId',
                                'ip': 'includeParent',
                                'm': 'mute',
                                'q': 'query',
                                'sc': 'selectedClips',
                                'tlc': 'topLevelClips'},
                 'type': 'animation'},
  'timeEditorAnimSource': {'flags': {'addObjects': {'args': '<type __builtin__.str>',
                                                    'longname': 'addObjects',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ao'},
                                     'addRelatedKG': {'args': '<type __builtin__.bool>',
                                                      'longname': 'addRelatedKG',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'akg'},
                                     'addSelectedObjects': {'args': '<type __builtin__.bool>',
                                                            'longname': 'addSelectedObjects',
                                                            'modes': ['create',
                                                                      'query',
                                                                      'edit'],
                                                            'numArgs': 0,
                                                            'shortname': 'aso'},
                                     'addSource': {'args': '<type __builtin__.str>',
                                                   'longname': 'addSource',
                                                   'modes': ['edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'asc'},
                                     'apply': {'args': '<type __builtin__.bool>',
                                               'longname': 'apply',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'ap'},
                                     'attribute': {'args': '<type __builtin__.str>',
                                                   'longname': 'attribute',
                                                   'modes': ['create',
                                                             'edit',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'at'},
                                     'bakeToAnimSource': {'args': '<type __builtin__.str>',
                                                          'longname': 'bakeToAnimSource',
                                                          'modes': ['edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'bas'},
                                     'calculateTiming': {'args': '<type __builtin__.bool>',
                                                         'longname': 'calculateTiming',
                                                         'modes': ['query',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'ct'},
                                     'copyAnimation': {'args': '<type __builtin__.bool>',
                                                       'longname': 'copyAnimation',
                                                       'modes': ['edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'cp'},
                                     'drivenClips': {'args': '<type __builtin__.bool>',
                                                     'longname': 'drivenClips',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'dc'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'exclusive': {'args': '<type __builtin__.bool>',
                                                   'longname': 'exclusive',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'exc'},
                                     'export': {'args': '<type __builtin__.str>',
                                                'longname': 'export',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'ex'},
                                     'importAllFbxTakes': {'args': '<type __builtin__.bool>',
                                                           'longname': 'importAllFbxTakes',
                                                           'modes': ['create'],
                                                           'numArgs': 0,
                                                           'shortname': 'aft'},
                                     'importFbx': {'args': '<type __builtin__.str>',
                                                   'longname': 'importFbx',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'fbx'},
                                     'importFbxTakes': {'args': '<type __builtin__.str>',
                                                        'longname': 'importFbxTakes',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'ft'},
                                     'importMayaFile': {'args': '<type __builtin__.str>',
                                                        'longname': 'importMayaFile',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'mf'},
                                     'importOption': {'args': '<type __builtin__.str>',
                                                      'longname': 'importOption',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'io'},
                                     'importPopulateOption': {'args': '<type __builtin__.str>',
                                                              'longname': 'importPopulateOption',
                                                              'modes': ['edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'ipo'},
                                     'importedContainerNames': {'args': '<type __builtin__.str>',
                                                                'longname': 'importedContainerNames',
                                                                'modes': ['create'],
                                                                'numArgs': 1,
                                                                'shortname': 'icn'},
                                     'includeRoot': {'args': '<type __builtin__.bool>',
                                                     'longname': 'includeRoot',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'irt'},
                                     'isUnique': {'args': '<type __builtin__.bool>',
                                                  'longname': 'isUnique',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'iu'},
                                     'populateImportedAnimSources': {'args': '<type __builtin__.str>',
                                                                     'longname': 'populateImportedAnimSources',
                                                                     'modes': ['create'],
                                                                     'numArgs': 1,
                                                                     'shortname': 'pia'},
                                     'poseClip': {'args': '<type __builtin__.bool>',
                                                  'longname': 'poseClip',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'poc'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'recursively': {'args': '<type __builtin__.bool>',
                                                     'longname': 'recursively',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'rec'},
                                     'removeSceneAnimation': {'args': '<type __builtin__.bool>',
                                                              'longname': 'removeSceneAnimation',
                                                              'modes': ['create',
                                                                        'edit'],
                                                              'numArgs': 1,
                                                              'shortname': 'rsa'},
                                     'removeSource': {'args': '<type __builtin__.str>',
                                                      'longname': 'removeSource',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'rs'},
                                     'showAnimSourceRemapping': {'args': '<type __builtin__.bool>',
                                                                 'longname': 'showAnimSourceRemapping',
                                                                 'modes': ['create'],
                                                                 'numArgs': 0,
                                                                 'shortname': 'sar'},
                                     'takeList': {'args': '<type __builtin__.str>',
                                                  'longname': 'takeList',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'tl'},
                                     'takesToImport': {'args': '<type __builtin__.str>',
                                                       'longname': 'takesToImport',
                                                       'modes': ['create'],
                                                       'numArgs': 1,
                                                       'shortname': 'toi'},
                                     'targetIndex': {'args': '<type __builtin__.str>',
                                                     'longname': 'targetIndex',
                                                     'modes': ['query'],
                                                     'numArgs': 1,
                                                     'shortname': 'ti'},
                                     'targets': {'args': '<type __builtin__.bool>',
                                                 'longname': 'targets',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'trg'},
                                     'type': {'args': '<type __builtin__.str>',
                                              'longname': 'type',
                                              'modes': ['create',
                                                        'query',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'typ'}},
                           'shortFlags': {'aft': 'importAllFbxTakes',
                                          'akg': 'addRelatedKG',
                                          'ao': 'addObjects',
                                          'ap': 'apply',
                                          'asc': 'addSource',
                                          'aso': 'addSelectedObjects',
                                          'at': 'attribute',
                                          'bas': 'bakeToAnimSource',
                                          'cp': 'copyAnimation',
                                          'ct': 'calculateTiming',
                                          'dc': 'drivenClips',
                                          'e': 'edit',
                                          'ex': 'export',
                                          'exc': 'exclusive',
                                          'fbx': 'importFbx',
                                          'ft': 'importFbxTakes',
                                          'icn': 'importedContainerNames',
                                          'io': 'importOption',
                                          'ipo': 'importPopulateOption',
                                          'irt': 'includeRoot',
                                          'iu': 'isUnique',
                                          'mf': 'importMayaFile',
                                          'pia': 'populateImportedAnimSources',
                                          'poc': 'poseClip',
                                          'q': 'query',
                                          'rec': 'recursively',
                                          'rs': 'removeSource',
                                          'rsa': 'removeSceneAnimation',
                                          'sar': 'showAnimSourceRemapping',
                                          'ti': 'targetIndex',
                                          'tl': 'takeList',
                                          'toi': 'takesToImport',
                                          'trg': 'targets',
                                          'typ': 'type'},
                           'type': 'animation'},
  'timeEditorBakeClips': {'flags': {'bakeToAnimSource': {'args': '<type __builtin__.str>',
                                                         'longname': 'bakeToAnimSource',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'bas'},
                                    'bakeToClip': {'args': '<type __builtin__.str>',
                                                   'longname': 'bakeToClip',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'btc'},
                                    'clipId': {'args': '<type __builtin__.int>',
                                               'longname': 'clipId',
                                               'modes': ['create',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'id'},
                                    'combineLayers': {'args': '<type __builtin__.bool>',
                                                      'longname': 'combineLayers',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'cl'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'forceSampling': {'args': '<type __builtin__.bool>',
                                                      'longname': 'forceSampling',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'fs'},
                                    'keepOriginalClip': {'args': '<type __builtin__.bool>',
                                                         'longname': 'keepOriginalClip',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'koc'},
                                    'path': {'args': '<type __builtin__.str>',
                                             'longname': 'path',
                                             'modes': ['create',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'pt'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'sampleBy': {'args': 'time',
                                                 'longname': 'sampleBy',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'sb'},
                                    'targetTrackIndex': {'args': '<type __builtin__.int>',
                                                         'longname': 'targetTrackIndex',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'tti'},
                                    'targetTracksNode': {'args': '<type __builtin__.str>',
                                                         'longname': 'targetTracksNode',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'ttn'}},
                          'shortFlags': {'bas': 'bakeToAnimSource',
                                         'btc': 'bakeToClip',
                                         'cl': 'combineLayers',
                                         'e': 'edit',
                                         'fs': 'forceSampling',
                                         'id': 'clipId',
                                         'koc': 'keepOriginalClip',
                                         'pt': 'path',
                                         'q': 'query',
                                         'sb': 'sampleBy',
                                         'tti': 'targetTrackIndex',
                                         'ttn': 'targetTracksNode'},
                          'type': 'animation'},
  'timeEditorClip': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                            'longname': 'absolute',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'abs'},
                               'addAttribute': {'args': '<type __builtin__.str>',
                                                'longname': 'addAttribute',
                                                'modes': ['edit',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'aa'},
                               'addObjects': {'args': '<type __builtin__.str>',
                                              'longname': 'addObjects',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ao'},
                               'addRelatedKG': {'args': '<type __builtin__.bool>',
                                                'longname': 'addRelatedKG',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'akg'},
                               'addSelectedObjects': {'args': '<type __builtin__.bool>',
                                                      'longname': 'addSelectedObjects',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'aso'},
                               'allowShrinking': {'args': '<type __builtin__.bool>',
                                                  'longname': 'allowShrinking',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'eas'},
                               'animSource': {'args': '<type __builtin__.str>',
                                              'longname': 'animSource',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'asr'},
                               'attribute': {'args': '<type __builtin__.str>',
                                             'longname': 'attribute',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 1,
                                             'shortname': 'at'},
                               'audio': {'args': '<type __builtin__.str>',
                                         'longname': 'audio',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'au'},
                               'children': {'args': '<type __builtin__.int>',
                                            'longname': 'children',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'chl'},
                               'clipAfter': {'args': '<type __builtin__.bool>',
                                             'longname': 'clipAfter',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'ca'},
                               'clipBefore': {'args': '<type __builtin__.bool>',
                                              'longname': 'clipBefore',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'cb'},
                               'clipDataType': {'args': '<type __builtin__.bool>',
                                                'longname': 'clipDataType',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'cdt'},
                               'clipId': {'args': '<type __builtin__.int>',
                                          'longname': 'clipId',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'id'},
                               'clipIdFromNodeName': {'args': '<type __builtin__.int>',
                                                      'longname': 'clipIdFromNodeName',
                                                      'modes': ['query'],
                                                      'numArgs': 1,
                                                      'shortname': 'idn'},
                               'clipIdFromPath': {'args': '<type __builtin__.bool>',
                                                  'longname': 'clipIdFromPath',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'idp'},
                               'clipNode': {'args': '<type __builtin__.bool>',
                                            'longname': 'clipNode',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'cln'},
                               'clipPath': {'args': '<type __builtin__.bool>',
                                            'longname': 'clipPath',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'clp'},
                               'copyClip': {'args': '<type __builtin__.bool>',
                                            'longname': 'copyClip',
                                            'modes': ['edit'],
                                            'numArgs': 0,
                                            'shortname': 'ccl'},
                               'crossfadeMode': {'args': '<type __builtin__.int>',
                                                 'longname': 'crossfadeMode',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'cfm'},
                               'crossfadePlug': {'args': '<type __builtin__.bool>',
                                                 'longname': 'crossfadePlug',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'cfp'},
                               'curveTime': {'args': 'time',
                                             'longname': 'curveTime',
                                             'modes': ['query'],
                                             'numArgs': 1,
                                             'shortname': 'cvt'},
                               'defaultGhostRoot': {'args': '<type __builtin__.bool>',
                                                    'longname': 'defaultGhostRoot',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dgr'},
                               'drivenAttributes': {'args': '<type __builtin__.bool>',
                                                    'longname': 'drivenAttributes',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'dat'},
                               'drivenClipsBySource': {'args': '<type __builtin__.str>',
                                                       'longname': 'drivenClipsBySource',
                                                       'modes': ['query'],
                                                       'numArgs': 1,
                                                       'shortname': 'dcs'},
                               'drivenObjects': {'args': '<type __builtin__.bool>',
                                                 'longname': 'drivenObjects',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'dos'},
                               'drivenRootObjects': {'args': '<type __builtin__.bool>',
                                                     'longname': 'drivenRootObjects',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'dro'},
                               'drivingSources': {'args': '<type __builtin__.str>',
                                                  'longname': 'drivingSources',
                                                  'modes': ['query'],
                                                  'numArgs': 1,
                                                  'shortname': 'dsc'},
                               'duplicateClip': {'args': '<type __builtin__.bool>',
                                                 'longname': 'duplicateClip',
                                                 'modes': ['edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'dcl'},
                               'duration': {'args': 'time',
                                            'longname': 'duration',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'd'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'emptySource': {'args': '<type __builtin__.bool>',
                                               'longname': 'emptySource',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'ems'},
                               'endTime': {'args': 'time',
                                           'longname': 'endTime',
                                           'modes': ['query'],
                                           'numArgs': 1,
                                           'shortname': 'et'},
                               'exclusive': {'args': '<type __builtin__.bool>',
                                             'longname': 'exclusive',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'exc'},
                               'existingOnly': {'args': '<type __builtin__.bool>',
                                                'longname': 'existingOnly',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'exo'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'exs'},
                               'explode': {'args': '<type __builtin__.int>',
                                           'longname': 'explode',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'epl'},
                               'exportAllClips': {'args': '<type __builtin__.bool>',
                                                  'longname': 'exportAllClips',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'eac'},
                               'exportFbx': {'args': '<type __builtin__.str>',
                                             'longname': 'exportFbx',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'ef'},
                               'extend': {'args': '<type __builtin__.bool>',
                                          'longname': 'extend',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'extendParent': {'args': '<type __builtin__.bool>',
                                                'longname': 'extendParent',
                                                'modes': ['edit'],
                                                'numArgs': 0,
                                                'shortname': 'exp'},
                               'ghost': {'args': '<type __builtin__.bool>',
                                         'longname': 'ghost',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'gh'},
                               'ghostRootAdd': {'args': '<type __builtin__.str>',
                                                'longname': 'ghostRootAdd',
                                                'modes': ['edit',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'gra'},
                               'ghostRootRemove': {'args': '<type __builtin__.str>',
                                                   'longname': 'ghostRootRemove',
                                                   'modes': ['edit',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'grr'},
                               'group': {'args': '<type __builtin__.bool>',
                                         'longname': 'group',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'grp'},
                               'holdEnd': {'args': 'time',
                                           'longname': 'holdEnd',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'he'},
                               'holdStart': {'args': 'time',
                                             'longname': 'holdStart',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'hs'},
                               'importAllFbxTakes': {'args': '<type __builtin__.bool>',
                                                     'longname': 'importAllFbxTakes',
                                                     'modes': ['create'],
                                                     'numArgs': 0,
                                                     'shortname': 'aft'},
                               'importFbx': {'args': '<type __builtin__.str>',
                                             'longname': 'importFbx',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'fbx'},
                               'importFbxTakes': {'args': '<type __builtin__.str>',
                                                  'longname': 'importFbxTakes',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'ft'},
                               'importMayaFile': {'args': '<type __builtin__.str>',
                                                  'longname': 'importMayaFile',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'mf'},
                               'importOption': {'args': '<type __builtin__.str>',
                                                'longname': 'importOption',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'io'},
                               'importPopulateOption': {'args': '<type __builtin__.str>',
                                                        'longname': 'importPopulateOption',
                                                        'modes': ['edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ipo'},
                               'importTakeDestination': {'args': '<type __builtin__.int>',
                                                         'longname': 'importTakeDestination',
                                                         'modes': ['create'],
                                                         'numArgs': 1,
                                                         'shortname': 'itd'},
                               'importedContainerNames': {'args': '<type __builtin__.str>',
                                                          'longname': 'importedContainerNames',
                                                          'modes': ['create'],
                                                          'numArgs': 1,
                                                          'shortname': 'icn'},
                               'includeRoot': {'args': '<type __builtin__.bool>',
                                               'longname': 'includeRoot',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'irt'},
                               'isContainer': {'args': '<type __builtin__.bool>',
                                               'longname': 'isContainer',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'ict'},
                               'listUserGhostRoot': {'args': '<type __builtin__.bool>',
                                                     'longname': 'listUserGhostRoot',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'lug'},
                               'loopEnd': {'args': 'time',
                                           'longname': 'loopEnd',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'le'},
                               'loopStart': {'args': 'time',
                                             'longname': 'loopStart',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ls'},
                               'minClipDuration': {'args': '<type __builtin__.bool>',
                                                   'longname': 'minClipDuration',
                                                   'modes': ['query'],
                                                   'numArgs': 0,
                                                   'shortname': 'mcd'},
                               'modifyAnimSource': {'args': '<type __builtin__.bool>',
                                                    'longname': 'modifyAnimSource',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'mas'},
                               'moveClip': {'args': 'time',
                                            'longname': 'moveClip',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'mcl'},
                               'mute': {'args': '<type __builtin__.bool>',
                                        'longname': 'mute',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'parent': {'args': '<type __builtin__.int>',
                                          'longname': 'parent',
                                          'modes': ['edit'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'parentClipId': {'args': '<type __builtin__.int>',
                                                'longname': 'parentClipId',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'pid'},
                               'parentGroupId': {'args': '<type __builtin__.bool>',
                                                 'longname': 'parentGroupId',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'pgd'},
                               'pasteClip': {'args': 'time',
                                             'longname': 'pasteClip',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'pcl'},
                               'path': {'args': '<type __builtin__.str>',
                                        'longname': 'path',
                                        'modes': ['edit', 'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'pt'},
                               'populateImportedAnimSources': {'args': '<type __builtin__.str>',
                                                               'longname': 'populateImportedAnimSources',
                                                               'modes': ['create'],
                                                               'numArgs': 1,
                                                               'shortname': 'pia'},
                               'poseClip': {'args': '<type __builtin__.bool>',
                                            'longname': 'poseClip',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'poc'},
                               'preserveAnimationTiming': {'args': '<type __builtin__.bool>',
                                                           'longname': 'preserveAnimationTiming',
                                                           'modes': ['create'],
                                                           'numArgs': 0,
                                                           'shortname': 'pat'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'razorClip': {'args': 'time',
                                             'longname': 'razorClip',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'rcl'},
                               'recursively': {'args': '<type __builtin__.bool>',
                                               'longname': 'recursively',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 0,
                                               'shortname': 'rec'},
                               'remap': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'remap',
                                         'modes': ['edit'],
                                         'numArgs': 2,
                                         'shortname': 'rmp'},
                               'remapSource': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'remapSource',
                                               'modes': ['edit'],
                                               'numArgs': 2,
                                               'shortname': 'rs'},
                               'remappedSourceAttrs': {'args': '<type __builtin__.bool>',
                                                       'longname': 'remappedSourceAttrs',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'rms'},
                               'remappedTargetAttrs': {'args': '<type __builtin__.bool>',
                                                       'longname': 'remappedTargetAttrs',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'rmt'},
                               'removeAttribute': {'args': '<type __builtin__.str>',
                                                   'longname': 'removeAttribute',
                                                   'modes': ['edit',
                                                             'multiuse'],
                                                   'numArgs': 1,
                                                   'shortname': 'ra'},
                               'removeClip': {'args': '<type __builtin__.bool>',
                                              'longname': 'removeClip',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'rmc'},
                               'removeCrossfade': {'args': '<type __builtin__.bool>',
                                                   'longname': 'removeCrossfade',
                                                   'modes': ['edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'rcf'},
                               'removeSceneAnimation': {'args': '<type __builtin__.bool>',
                                                        'longname': 'removeSceneAnimation',
                                                        'modes': ['create',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'rsa'},
                               'removeWeightCurve': {'args': '<type __builtin__.bool>',
                                                     'longname': 'removeWeightCurve',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 0,
                                                     'shortname': 'rwc'},
                               'resetTiming': {'args': '<type __builtin__.bool>',
                                               'longname': 'resetTiming',
                                               'modes': ['edit'],
                                               'numArgs': 0,
                                               'shortname': 'rt'},
                               'resetTransition': {'args': '<type __builtin__.bool>',
                                                   'longname': 'resetTransition',
                                                   'modes': ['edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'rtr'},
                               'ripple': {'args': '<type __builtin__.bool>',
                                          'longname': 'ripple',
                                          'modes': ['edit'],
                                          'numArgs': 0,
                                          'shortname': 'rpl'},
                               'rootClipId': {'args': '<type __builtin__.int>',
                                              'longname': 'rootClipId',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'rti'},
                               'rootPath': {'args': '<type __builtin__.str>',
                                            'longname': 'rootPath',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'rpt'},
                               'scaleEnd': {'args': 'time',
                                            'longname': 'scaleEnd',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'sce'},
                               'scalePivot': {'args': 'time',
                                              'longname': 'scalePivot',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'scp'},
                               'scaleStart': {'args': 'time',
                                              'longname': 'scaleStart',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'scs'},
                               'setKeyframe': {'args': '<type __builtin__.str>',
                                               'longname': 'setKeyframe',
                                               'modes': ['edit',
                                                         'multiuse'],
                                               'numArgs': 1,
                                               'shortname': 'k'},
                               'showAnimSourceRemapping': {'args': '<type __builtin__.bool>',
                                                           'longname': 'showAnimSourceRemapping',
                                                           'modes': ['create'],
                                                           'numArgs': 0,
                                                           'shortname': 'sar'},
                               'speedRamping': {'args': '<type __builtin__.int>',
                                                'longname': 'speedRamping',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'src'},
                               'startTime': {'args': 'time',
                                             'longname': 'startTime',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 's'},
                               'takeList': {'args': '<type __builtin__.str>',
                                            'longname': 'takeList',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'tl'},
                               'takesToImport': {'args': '<type __builtin__.str>',
                                                 'longname': 'takesToImport',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'toi'},
                               'timeWarp': {'args': '<type __builtin__.bool>',
                                            'longname': 'timeWarp',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'tw'},
                               'timeWarpCurve': {'args': '<type __builtin__.bool>',
                                                 'longname': 'timeWarpCurve',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'twc'},
                               'timeWarpType': {'args': '<type __builtin__.int>',
                                                'longname': 'timeWarpType',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'twt'},
                               'track': {'args': '<type __builtin__.str>',
                                         'longname': 'track',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'trk'},
                               'tracksNode': {'args': '<type __builtin__.bool>',
                                              'longname': 'tracksNode',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'trn'},
                               'transition': {'args': '<type __builtin__.bool>',
                                              'longname': 'transition',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'tra'},
                               'trimEnd': {'args': 'time',
                                           'longname': 'trimEnd',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'tre'},
                               'trimStart': {'args': 'time',
                                             'longname': 'trimStart',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'trs'},
                               'truncated': {'args': '<type __builtin__.bool>',
                                             'longname': 'truncated',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'trc'},
                               'type': {'args': '<type __builtin__.str>',
                                        'longname': 'type',
                                        'modes': ['create',
                                                  'query',
                                                  'edit',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'typ'},
                               'uniqueAnimSource': {'args': '<type __builtin__.bool>',
                                                    'longname': 'uniqueAnimSource',
                                                    'modes': ['edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'uas'},
                               'userGhostRoot': {'args': '<type __builtin__.bool>',
                                                 'longname': 'userGhostRoot',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ugr'},
                               'weightCurve': {'args': '<type __builtin__.bool>',
                                               'longname': 'weightCurve',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 0,
                                               'shortname': 'wc'},
                               'zeroKeying': {'args': '<type __builtin__.bool>',
                                              'longname': 'zeroKeying',
                                              'modes': ['edit'],
                                              'numArgs': 0,
                                              'shortname': 'zk'}},
                     'shortFlags': {'aa': 'addAttribute',
                                    'abs': 'absolute',
                                    'aft': 'importAllFbxTakes',
                                    'akg': 'addRelatedKG',
                                    'ao': 'addObjects',
                                    'aso': 'addSelectedObjects',
                                    'asr': 'animSource',
                                    'at': 'attribute',
                                    'au': 'audio',
                                    'ca': 'clipAfter',
                                    'cb': 'clipBefore',
                                    'ccl': 'copyClip',
                                    'cdt': 'clipDataType',
                                    'cfm': 'crossfadeMode',
                                    'cfp': 'crossfadePlug',
                                    'chl': 'children',
                                    'cln': 'clipNode',
                                    'clp': 'clipPath',
                                    'cvt': 'curveTime',
                                    'd': 'duration',
                                    'dat': 'drivenAttributes',
                                    'dcl': 'duplicateClip',
                                    'dcs': 'drivenClipsBySource',
                                    'dgr': 'defaultGhostRoot',
                                    'dos': 'drivenObjects',
                                    'dro': 'drivenRootObjects',
                                    'dsc': 'drivingSources',
                                    'e': 'edit',
                                    'eac': 'exportAllClips',
                                    'eas': 'allowShrinking',
                                    'ef': 'exportFbx',
                                    'ems': 'emptySource',
                                    'epl': 'explode',
                                    'et': 'endTime',
                                    'ex': 'extend',
                                    'exc': 'exclusive',
                                    'exo': 'existingOnly',
                                    'exp': 'extendParent',
                                    'exs': 'exists',
                                    'fbx': 'importFbx',
                                    'ft': 'importFbxTakes',
                                    'gh': 'ghost',
                                    'gra': 'ghostRootAdd',
                                    'grp': 'group',
                                    'grr': 'ghostRootRemove',
                                    'he': 'holdEnd',
                                    'hs': 'holdStart',
                                    'icn': 'importedContainerNames',
                                    'ict': 'isContainer',
                                    'id': 'clipId',
                                    'idn': 'clipIdFromNodeName',
                                    'idp': 'clipIdFromPath',
                                    'io': 'importOption',
                                    'ipo': 'importPopulateOption',
                                    'irt': 'includeRoot',
                                    'itd': 'importTakeDestination',
                                    'k': 'setKeyframe',
                                    'le': 'loopEnd',
                                    'ls': 'loopStart',
                                    'lug': 'listUserGhostRoot',
                                    'm': 'mute',
                                    'mas': 'modifyAnimSource',
                                    'mcd': 'minClipDuration',
                                    'mcl': 'moveClip',
                                    'mf': 'importMayaFile',
                                    'n': 'name',
                                    'p': 'parent',
                                    'pat': 'preserveAnimationTiming',
                                    'pcl': 'pasteClip',
                                    'pgd': 'parentGroupId',
                                    'pia': 'populateImportedAnimSources',
                                    'pid': 'parentClipId',
                                    'poc': 'poseClip',
                                    'pt': 'path',
                                    'q': 'query',
                                    'ra': 'removeAttribute',
                                    'rcf': 'removeCrossfade',
                                    'rcl': 'razorClip',
                                    'rec': 'recursively',
                                    'rmc': 'removeClip',
                                    'rmp': 'remap',
                                    'rms': 'remappedSourceAttrs',
                                    'rmt': 'remappedTargetAttrs',
                                    'rpl': 'ripple',
                                    'rpt': 'rootPath',
                                    'rs': 'remapSource',
                                    'rsa': 'removeSceneAnimation',
                                    'rt': 'resetTiming',
                                    'rti': 'rootClipId',
                                    'rtr': 'resetTransition',
                                    'rwc': 'removeWeightCurve',
                                    's': 'startTime',
                                    'sar': 'showAnimSourceRemapping',
                                    'sce': 'scaleEnd',
                                    'scp': 'scalePivot',
                                    'scs': 'scaleStart',
                                    'src': 'speedRamping',
                                    'tl': 'takeList',
                                    'toi': 'takesToImport',
                                    'tra': 'transition',
                                    'trc': 'truncated',
                                    'tre': 'trimEnd',
                                    'trk': 'track',
                                    'trn': 'tracksNode',
                                    'trs': 'trimStart',
                                    'tw': 'timeWarp',
                                    'twc': 'timeWarpCurve',
                                    'twt': 'timeWarpType',
                                    'typ': 'type',
                                    'uas': 'uniqueAnimSource',
                                    'ugr': 'userGhostRoot',
                                    'wc': 'weightCurve',
                                    'zk': 'zeroKeying'},
                     'type': 'animation'},
  'timeEditorClipLayer': {'flags': {'addAttribute': {'args': '<type __builtin__.str>',
                                                     'longname': 'addAttribute',
                                                     'modes': ['edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'aa'},
                                    'addLayer': {'args': '<type __builtin__.str>',
                                                 'longname': 'addLayer',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'al'},
                                    'addObject': {'args': '<type __builtin__.str>',
                                                  'longname': 'addObject',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ao'},
                                    'allLayers': {'args': '<type __builtin__.bool>',
                                                  'longname': 'allLayers',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'all'},
                                    'attribute': {'args': '<type __builtin__.str>',
                                                  'longname': 'attribute',
                                                  'modes': ['edit',
                                                            'multiuse'],
                                                  'numArgs': 1,
                                                  'shortname': 'a'},
                                    'attributeKeyable': {'args': '<type __builtin__.str>',
                                                         'longname': 'attributeKeyable',
                                                         'modes': ['query'],
                                                         'numArgs': 1,
                                                         'shortname': 'ak'},
                                    'clipId': {'args': '<type __builtin__.int>',
                                               'longname': 'clipId',
                                               'modes': ['edit'],
                                               'numArgs': 1,
                                               'shortname': 'cid'},
                                    'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'index': {'args': '<type __builtin__.int>',
                                              'longname': 'index',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'idx'},
                                    'keySiblings': {'args': '<type __builtin__.bool>',
                                                    'longname': 'keySiblings',
                                                    'modes': ['edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'ks'},
                                    'layerId': {'args': '<type __builtin__.int>',
                                                'longname': 'layerId',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'lid'},
                                    'layerName': {'args': '<type __builtin__.str>',
                                                  'longname': 'layerName',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ln'},
                                    'mode': {'args': '<type __builtin__.int>',
                                             'longname': 'mode',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'm'},
                                    'mute': {'args': '<type __builtin__.bool>',
                                             'longname': 'mute',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'mu'},
                                    'name': {'args': '<type __builtin__.bool>',
                                             'longname': 'name',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'n'},
                                    'path': {'args': '<type __builtin__.str>',
                                             'longname': 'path',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'pt'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'removeAttribute': {'args': '<type __builtin__.str>',
                                                        'longname': 'removeAttribute',
                                                        'modes': ['edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'ra'},
                                    'removeLayer': {'args': '<type __builtin__.bool>',
                                                    'longname': 'removeLayer',
                                                    'modes': ['edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'rl'},
                                    'removeObject': {'args': '<type __builtin__.str>',
                                                     'longname': 'removeObject',
                                                     'modes': ['edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ro'},
                                    'resetSolo': {'args': '<type __builtin__.bool>',
                                                  'longname': 'resetSolo',
                                                  'modes': ['edit'],
                                                  'numArgs': 0,
                                                  'shortname': 'rs'},
                                    'setKeyframe': {'args': '<type __builtin__.bool>',
                                                    'longname': 'setKeyframe',
                                                    'modes': ['edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'k'},
                                    'solo': {'args': '<type __builtin__.bool>',
                                             'longname': 'solo',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'sl'},
                                    'zeroKeying': {'args': '<type __builtin__.bool>',
                                                   'longname': 'zeroKeying',
                                                   'modes': ['edit'],
                                                   'numArgs': 0,
                                                   'shortname': 'zk'}},
                          'shortFlags': {'a': 'attribute',
                                         'aa': 'addAttribute',
                                         'ak': 'attributeKeyable',
                                         'al': 'addLayer',
                                         'all': 'allLayers',
                                         'ao': 'addObject',
                                         'cid': 'clipId',
                                         'e': 'edit',
                                         'idx': 'index',
                                         'k': 'setKeyframe',
                                         'ks': 'keySiblings',
                                         'lid': 'layerId',
                                         'ln': 'layerName',
                                         'm': 'mode',
                                         'mu': 'mute',
                                         'n': 'name',
                                         'pt': 'path',
                                         'q': 'query',
                                         'ra': 'removeAttribute',
                                         'rl': 'removeLayer',
                                         'ro': 'removeObject',
                                         'rs': 'resetSolo',
                                         'sl': 'solo',
                                         'zk': 'zeroKeying'},
                          'type': 'animation'},
  'timeEditorClipOffset': {'flags': {'applyToAllRoots': {'args': '<type __builtin__.bool>',
                                                         'longname': 'applyToAllRoots',
                                                         'modes': ['create'],
                                                         'numArgs': 0,
                                                         'shortname': 'atr'},
                                     'clipId': {'args': '<type __builtin__.int>',
                                                'longname': 'clipId',
                                                'modes': ['create',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'id'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'matchClipId': {'args': '<type __builtin__.int>',
                                                     'longname': 'matchClipId',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'mci'},
                                     'matchDstTime': {'args': 'time',
                                                      'longname': 'matchDstTime',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'mdt'},
                                     'matchObj': {'args': 'PyNode',
                                                  'longname': 'matchObj',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'mob'},
                                     'matchOffsetRot': {'args': '<type __builtin__.bool>',
                                                        'longname': 'matchOffsetRot',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'mor'},
                                     'matchOffsetTrans': {'args': '<type __builtin__.bool>',
                                                          'longname': 'matchOffsetTrans',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'mot'},
                                     'matchPath': {'args': '<type __builtin__.str>',
                                                   'longname': 'matchPath',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'mpt'},
                                     'matchRotOp': {'args': '<type __builtin__.int>',
                                                    'longname': 'matchRotOp',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'mro'},
                                     'matchSrcTime': {'args': 'time',
                                                      'longname': 'matchSrcTime',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'mst'},
                                     'matchTransOp': {'args': '<type __builtin__.int>',
                                                      'longname': 'matchTransOp',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'mto'},
                                     'offsetTransform': {'args': '<type __builtin__.bool>',
                                                         'longname': 'offsetTransform',
                                                         'modes': ['create',
                                                                   'query'],
                                                         'numArgs': 0,
                                                         'shortname': 'oft'},
                                     'path': {'args': '<type __builtin__.str>',
                                              'longname': 'path',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'pt'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'resetMatch': {'args': '<type __builtin__.int>',
                                                    'longname': 'resetMatch',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'rsm'},
                                     'resetMatchPath': {'args': '<type __builtin__.str>',
                                                        'longname': 'resetMatchPath',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'rmp'},
                                     'rootObj': {'args': '<type __builtin__.str>',
                                                 'longname': 'rootObj',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit',
                                                           'multiuse'],
                                                 'numArgs': 1,
                                                 'shortname': 'rob'},
                                     'upVectorX': {'args': '<type __builtin__.float>',
                                                   'longname': 'upVectorX',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'upx'},
                                     'upVectorY': {'args': '<type __builtin__.float>',
                                                   'longname': 'upVectorY',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'upy'},
                                     'upVectorZ': {'args': '<type __builtin__.float>',
                                                   'longname': 'upVectorZ',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'upz'}},
                           'shortFlags': {'atr': 'applyToAllRoots',
                                          'e': 'edit',
                                          'id': 'clipId',
                                          'mci': 'matchClipId',
                                          'mdt': 'matchDstTime',
                                          'mob': 'matchObj',
                                          'mor': 'matchOffsetRot',
                                          'mot': 'matchOffsetTrans',
                                          'mpt': 'matchPath',
                                          'mro': 'matchRotOp',
                                          'mst': 'matchSrcTime',
                                          'mto': 'matchTransOp',
                                          'oft': 'offsetTransform',
                                          'pt': 'path',
                                          'q': 'query',
                                          'rmp': 'resetMatchPath',
                                          'rob': 'rootObj',
                                          'rsm': 'resetMatch',
                                          'upx': 'upVectorX',
                                          'upy': 'upVectorY',
                                          'upz': 'upVectorZ'},
                           'type': 'animation'},
  'timeEditorComposition': {'flags': {'active': {'args': '<type __builtin__.bool>',
                                                 'longname': 'active',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'act'},
                                      'allCompositions': {'args': '<type __builtin__.bool>',
                                                          'longname': 'allCompositions',
                                                          'modes': ['query'],
                                                          'numArgs': 0,
                                                          'shortname': 'acp'},
                                      'createTrack': {'args': '<type __builtin__.bool>',
                                                      'longname': 'createTrack',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'ct'},
                                      'delete': {'args': '<type __builtin__.bool>',
                                                 'longname': 'delete',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 0,
                                                 'shortname': 'delete'},
                                      'duplicateFrom': {'args': '<type __builtin__.str>',
                                                        'longname': 'duplicateFrom',
                                                        'modes': ['create'],
                                                        'numArgs': 1,
                                                        'shortname': 'df'},
                                      'edit': {'args': '<type __builtin__.bool>',
                                               'longname': 'edit',
                                               'numArgs': 0,
                                               'shortname': 'e'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'rename': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>'],
                                                 'longname': 'rename',
                                                 'modes': ['edit'],
                                                 'numArgs': 2,
                                                 'shortname': 'ren'},
                                      'tracksNode': {'args': '<type __builtin__.bool>',
                                                     'longname': 'tracksNode',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'tn'}},
                            'removedFlags': {'del': 'delete'},
                            'shortFlags': {'acp': 'allCompositions',
                                           'act': 'active',
                                           'ct': 'createTrack',
                                           'delete': 'delete',
                                           'df': 'duplicateFrom',
                                           'e': 'edit',
                                           'q': 'query',
                                           'ren': 'rename',
                                           'tn': 'tracksNode'},
                            'type': 'animation'},
  'timeEditorPanel': {'callbackFlags': ['menu', 'm'],
                      'flags': {'activeClipEditMode': {'args': '<type __builtin__.int>',
                                                       'longname': 'activeClipEditMode',
                                                       'modes': ['query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'ace'},
                                'activeTabRootClipId': {'args': '<type __builtin__.bool>',
                                                        'longname': 'activeTabRootClipId',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'atr'},
                                'activeTabTime': {'args': '<type __builtin__.bool>',
                                                  'longname': 'activeTabTime',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'att'},
                                'activeTabView': {'args': '<type __builtin__.int>',
                                                  'longname': 'activeTabView',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'atv'},
                                'autoFit': {'args': '<type __builtin__.str>',
                                            'longname': 'autoFit',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'af'},
                                'autoFitTime': {'args': '<type __builtin__.str>',
                                                'longname': 'autoFitTime',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'aft'},
                                'contextMenu': {'args': '<type __builtin__.bool>',
                                                'longname': 'contextMenu',
                                                'numArgs': 0,
                                                'shortname': 'cm'},
                                'control': {'args': '<type __builtin__.bool>',
                                            'longname': 'control',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ctl'},
                                'defineTemplate': {'args': '<type __builtin__.str>',
                                                   'longname': 'defineTemplate',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'dt'},
                                'displayActiveKeyTangents': {'args': '<type __builtin__.str>',
                                                             'longname': 'displayActiveKeyTangents',
                                                             'modes': ['edit'],
                                                             'numArgs': 1,
                                                             'shortname': 'dat'},
                                'displayActiveKeys': {'args': '<type __builtin__.str>',
                                                      'longname': 'displayActiveKeys',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'dak'},
                                'displayInfinities': {'args': '<type __builtin__.str>',
                                                      'longname': 'displayInfinities',
                                                      'modes': ['edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'di'},
                                'displayKeys': {'args': '<type __builtin__.str>',
                                                'longname': 'displayKeys',
                                                'modes': ['edit'],
                                                'numArgs': 1,
                                                'shortname': 'dk'},
                                'displayTangents': {'args': '<type __builtin__.str>',
                                                    'longname': 'displayTangents',
                                                    'modes': ['edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'dtn'},
                                'displayValues': {'args': '<type __builtin__.str>',
                                                  'longname': 'displayValues',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dv'},
                                'docTag': {'args': '<type __builtin__.str>',
                                           'longname': 'docTag',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dtg'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'exists': {'args': '<type __builtin__.bool>',
                                           'longname': 'exists',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ex'},
                                'filter': {'args': '<type __builtin__.str>',
                                           'longname': 'filter',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'f'},
                                'focusTrack': {'args': '<type __builtin__.str>',
                                               'longname': 'focusTrack',
                                               'numArgs': 1,
                                               'shortname': 'ft'},
                                'forceMainConnection': {'args': '<type __builtin__.str>',
                                                        'longname': 'forceMainConnection',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'fmc'},
                                'groupIdForTabView': {'args': '<type __builtin__.int>',
                                                      'longname': 'groupIdForTabView',
                                                      'modes': ['query'],
                                                      'numArgs': 1,
                                                      'shortname': 'gtv'},
                                'highlightConnection': {'args': '<type __builtin__.str>',
                                                        'longname': 'highlightConnection',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'hlc'},
                                'keyingTarget': {'args': '<type __builtin__.int>',
                                                 'longname': 'keyingTarget',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'kt'},
                                'layerId': {'args': '<type __builtin__.int>',
                                            'longname': 'layerId',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'l'},
                                'lockMainConnection': {'args': '<type __builtin__.bool>',
                                                       'longname': 'lockMainConnection',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 0,
                                                       'shortname': 'lck'},
                                'lookAt': {'args': '<type __builtin__.str>',
                                           'longname': 'lookAt',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'la'},
                                'mainListConnection': {'args': '<type __builtin__.str>',
                                                       'longname': 'mainListConnection',
                                                       'modes': ['create',
                                                                 'query',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'mlc'},
                                'menu': {'args': 'script',
                                         'longname': 'menu',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                                'minClipWidth': {'args': '<type __builtin__.int>',
                                                 'longname': 'minClipWidth',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'mcw'},
                                'panel': {'args': '<type __builtin__.str>',
                                          'longname': 'panel',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'pnl'},
                                'parent': {'args': '<type __builtin__.str>',
                                           'longname': 'parent',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'p'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'selectionConnection': {'args': '<type __builtin__.str>',
                                                        'longname': 'selectionConnection',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 1,
                                                        'shortname': 'slc'},
                                'setToPrevClipEditMode': {'args': '<type __builtin__.bool>',
                                                          'longname': 'setToPrevClipEditMode',
                                                          'modes': ['edit'],
                                                          'numArgs': 0,
                                                          'shortname': 'spe'},
                                'snapTime': {'args': '<type __builtin__.str>',
                                             'longname': 'snapTime',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'st'},
                                'snapToClip': {'args': '<type __builtin__.bool>',
                                               'longname': 'snapToClip',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'stc'},
                                'snapToFrame': {'args': '<type __builtin__.bool>',
                                                'longname': 'snapToFrame',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'stf'},
                                'snapTolerance': {'args': '<type __builtin__.int>',
                                                  'longname': 'snapTolerance',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sto'},
                                'snapValue': {'args': '<type __builtin__.str>',
                                              'longname': 'snapValue',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sv'},
                                'stateString': {'args': '<type __builtin__.bool>',
                                                'longname': 'stateString',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'sts'},
                                'tabView': {'args': '<type __builtin__.int>',
                                            'longname': 'tabView',
                                            'modes': ['edit'],
                                            'numArgs': 1,
                                            'shortname': 'tv'},
                                'timeCursor': {'args': '<type __builtin__.bool>',
                                               'longname': 'timeCursor',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tc'},
                                'togglekeyview': {'args': '<type __builtin__.bool>',
                                                  'longname': 'togglekeyview',
                                                  'numArgs': 0,
                                                  'shortname': 'tkv'},
                                'unParent': {'args': '<type __builtin__.bool>',
                                             'longname': 'unParent',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'up'},
                                'unlockMainConnection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'unlockMainConnection',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'ulk'},
                                'updateMainConnection': {'args': '<type __builtin__.bool>',
                                                         'longname': 'updateMainConnection',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 0,
                                                         'shortname': 'upd'},
                                'useTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'useTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'ut'},
                                'viewLeft': {'args': '<type __builtin__.bool>',
                                             'longname': 'viewLeft',
                                             'numArgs': 0,
                                             'shortname': 'vl'},
                                'viewRight': {'args': '<type __builtin__.bool>',
                                              'longname': 'viewRight',
                                              'numArgs': 0,
                                              'shortname': 'vr'}},
                      'shortFlags': {'ace': 'activeClipEditMode',
                                     'af': 'autoFit',
                                     'aft': 'autoFitTime',
                                     'atr': 'activeTabRootClipId',
                                     'att': 'activeTabTime',
                                     'atv': 'activeTabView',
                                     'cm': 'contextMenu',
                                     'ctl': 'control',
                                     'dak': 'displayActiveKeys',
                                     'dat': 'displayActiveKeyTangents',
                                     'di': 'displayInfinities',
                                     'dk': 'displayKeys',
                                     'dt': 'defineTemplate',
                                     'dtg': 'docTag',
                                     'dtn': 'displayTangents',
                                     'dv': 'displayValues',
                                     'e': 'edit',
                                     'ex': 'exists',
                                     'f': 'filter',
                                     'fmc': 'forceMainConnection',
                                     'ft': 'focusTrack',
                                     'gtv': 'groupIdForTabView',
                                     'hlc': 'highlightConnection',
                                     'kt': 'keyingTarget',
                                     'l': 'layerId',
                                     'la': 'lookAt',
                                     'lck': 'lockMainConnection',
                                     'm': 'menu',
                                     'mcw': 'minClipWidth',
                                     'mlc': 'mainListConnection',
                                     'p': 'parent',
                                     'pnl': 'panel',
                                     'q': 'query',
                                     'slc': 'selectionConnection',
                                     'spe': 'setToPrevClipEditMode',
                                     'st': 'snapTime',
                                     'stc': 'snapToClip',
                                     'stf': 'snapToFrame',
                                     'sto': 'snapTolerance',
                                     'sts': 'stateString',
                                     'sv': 'snapValue',
                                     'tc': 'timeCursor',
                                     'tkv': 'togglekeyview',
                                     'tv': 'tabView',
                                     'ulk': 'unlockMainConnection',
                                     'up': 'unParent',
                                     'upd': 'updateMainConnection',
                                     'ut': 'useTemplate',
                                     'vl': 'viewLeft',
                                     'vr': 'viewRight'},
                      'type': 'animation'},
  'timeEditorTracks': {'flags': {'activeClipWeight': {'args': 'time',
                                                      'longname': 'activeClipWeight',
                                                      'modes': ['query'],
                                                      'numArgs': 1,
                                                      'shortname': 'acw'},
                                 'activeClipWeightId': {'args': 'time',
                                                        'longname': 'activeClipWeightId',
                                                        'modes': ['query'],
                                                        'numArgs': 1,
                                                        'shortname': 'aci'},
                                 'addTrack': {'args': '<type __builtin__.int>',
                                              'longname': 'addTrack',
                                              'modes': ['edit'],
                                              'numArgs': 1,
                                              'shortname': 'at'},
                                 'allClips': {'args': '<type __builtin__.bool>',
                                              'longname': 'allClips',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'ac'},
                                 'allTracks': {'args': '<type __builtin__.bool>',
                                               'longname': 'allTracks',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'atc'},
                                 'allTracksRecursive': {'args': '<type __builtin__.bool>',
                                                        'longname': 'allTracksRecursive',
                                                        'modes': ['query'],
                                                        'numArgs': 0,
                                                        'shortname': 'atr'},
                                 'composition': {'args': '<type __builtin__.bool>',
                                                 'longname': 'composition',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'cp'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'path': {'args': '<type __builtin__.str>',
                                          'longname': 'path',
                                          'modes': ['edit'],
                                          'numArgs': 1,
                                          'shortname': 'pt'},
                                 'plugIndex': {'args': '<type __builtin__.int>',
                                               'longname': 'plugIndex',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'pi'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'removeTrack': {'args': '<type __builtin__.int>',
                                                 'longname': 'removeTrack',
                                                 'modes': ['edit',
                                                           'multiuse'],
                                                 'numArgs': 1,
                                                 'shortname': 'rt'},
                                 'removeTrackByPath': {'args': '<type __builtin__.str>',
                                                       'longname': 'removeTrackByPath',
                                                       'modes': ['edit',
                                                                 'multiuse'],
                                                       'numArgs': 1,
                                                       'shortname': 'rtp'},
                                 'reorderTrack': {'args': ['<type __builtin__.int>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'reorderTrack',
                                                  'modes': ['edit'],
                                                  'numArgs': 2,
                                                  'shortname': 'rot'},
                                 'resetMute': {'args': '<type __builtin__.bool>',
                                               'longname': 'resetMute',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'rm'},
                                 'resetSolo': {'args': '<type __builtin__.bool>',
                                               'longname': 'resetSolo',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'rs'},
                                 'selectedTracks': {'args': '<type __builtin__.bool>',
                                                    'longname': 'selectedTracks',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'st'},
                                 'trackGhost': {'args': '<type __builtin__.bool>',
                                                'longname': 'trackGhost',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'tgh'},
                                 'trackIndex': {'args': '<type __builtin__.int>',
                                                'longname': 'trackIndex',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ti'},
                                 'trackMuted': {'args': '<type __builtin__.bool>',
                                                'longname': 'trackMuted',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'tm'},
                                 'trackName': {'args': '<type __builtin__.str>',
                                               'longname': 'trackName',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tn'},
                                 'trackSolo': {'args': '<type __builtin__.bool>',
                                               'longname': 'trackSolo',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ts'},
                                 'trackType': {'args': '<type __builtin__.int>',
                                               'longname': 'trackType',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tt'}},
                       'shortFlags': {'ac': 'allClips',
                                      'aci': 'activeClipWeightId',
                                      'acw': 'activeClipWeight',
                                      'at': 'addTrack',
                                      'atc': 'allTracks',
                                      'atr': 'allTracksRecursive',
                                      'cp': 'composition',
                                      'e': 'edit',
                                      'pi': 'plugIndex',
                                      'pt': 'path',
                                      'q': 'query',
                                      'rm': 'resetMute',
                                      'rot': 'reorderTrack',
                                      'rs': 'resetSolo',
                                      'rt': 'removeTrack',
                                      'rtp': 'removeTrackByPath',
                                      'st': 'selectedTracks',
                                      'tgh': 'trackGhost',
                                      'ti': 'trackIndex',
                                      'tm': 'trackMuted',
                                      'tn': 'trackName',
                                      'ts': 'trackSolo',
                                      'tt': 'trackType'},
                       'type': 'animation'},
  'timeField': {'callbackFlags': ['dragCommand',
                                  'dc',
                                  'dropCallback',
                                  'dpc',
                                  'visibleChangeCommand',
                                  'vcc',
                                  'dragCallback',
                                  'dgc',
                                  'enterCommand',
                                  'ec',
                                  'receiveFocusCommand',
                                  'rfc',
                                  'changeCommand',
                                  'cc'],
                'flags': {'annotation': {'args': '<type __builtin__.str>',
                                         'longname': 'annotation',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'ann'},
                          'backgroundColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'backgroundColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'bgc'},
                          'changeCommand': {'args': 'script',
                                            'longname': 'changeCommand',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'cc'},
                          'defineTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'defineTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dt'},
                          'docTag': {'args': '<type __builtin__.str>',
                                     'longname': 'docTag',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'dtg'},
                          'dragCallback': {'args': 'script',
                                           'longname': 'dragCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dgc'},
                          'dragCommand': {'args': 'script',
                                          'longname': 'dragCommand',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dc'},
                          'dropCallback': {'args': 'script',
                                           'longname': 'dropCallback',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'dpc'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'editable': {'args': '<type __builtin__.bool>',
                                       'longname': 'editable',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ed'},
                          'enable': {'args': '<type __builtin__.bool>',
                                     'longname': 'enable',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'en'},
                          'enableBackground': {'args': '<type __builtin__.bool>',
                                               'longname': 'enableBackground',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ebg'},
                          'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableKeyboardFocus',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ekf'},
                          'enterCommand': {'args': 'script',
                                           'longname': 'enterCommand',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ec'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'fullPathName': {'args': '<type __builtin__.bool>',
                                           'longname': 'fullPathName',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'fpn'},
                          'height': {'args': '<type __builtin__.int>',
                                     'longname': 'height',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'h'},
                          'highlightColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'highlightColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'hlc'},
                          'isObscured': {'args': '<type __builtin__.bool>',
                                         'longname': 'isObscured',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'io'},
                          'manage': {'args': '<type __builtin__.bool>',
                                     'longname': 'manage',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'm'},
                          'noBackground': {'args': '<type __builtin__.bool>',
                                           'longname': 'noBackground',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'nbg'},
                          'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'numberOfPopupMenus',
                                                 'modes': ['query'],
                                                 'numArgs': 0,
                                                 'shortname': 'npm'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                          'popupMenuArray': {'args': '<type __builtin__.bool>',
                                             'longname': 'popupMenuArray',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'pma'},
                          'precision': {'args': '<type __builtin__.int>',
                                        'longname': 'precision',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'pre'},
                          'preventOverride': {'args': '<type __builtin__.bool>',
                                              'longname': 'preventOverride',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'po'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'receiveFocusCommand': {'args': 'script',
                                                  'longname': 'receiveFocusCommand',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'rfc'},
                          'statusBarMessage': {'args': '<type __builtin__.str>',
                                               'longname': 'statusBarMessage',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'sbm'},
                          'step': {'args': 'time',
                                   'longname': 'step',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 's'},
                          'useTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'useTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ut'},
                          'value': {'args': 'time',
                                    'longname': 'value',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'v'},
                          'visible': {'args': '<type __builtin__.bool>',
                                      'longname': 'visible',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'vis'},
                          'visibleChangeCommand': {'args': 'script',
                                                   'longname': 'visibleChangeCommand',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'vcc'},
                          'width': {'args': '<type __builtin__.int>',
                                    'longname': 'width',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'w'}},
                'shortFlags': {'ann': 'annotation',
                               'bgc': 'backgroundColor',
                               'cc': 'changeCommand',
                               'dc': 'dragCommand',
                               'dgc': 'dragCallback',
                               'dpc': 'dropCallback',
                               'dt': 'defineTemplate',
                               'dtg': 'docTag',
                               'e': 'edit',
                               'ebg': 'enableBackground',
                               'ec': 'enterCommand',
                               'ed': 'editable',
                               'ekf': 'enableKeyboardFocus',
                               'en': 'enable',
                               'ex': 'exists',
                               'fpn': 'fullPathName',
                               'h': 'height',
                               'hlc': 'highlightColor',
                               'io': 'isObscured',
                               'm': 'manage',
                               'nbg': 'noBackground',
                               'npm': 'numberOfPopupMenus',
                               'p': 'parent',
                               'pma': 'popupMenuArray',
                               'po': 'preventOverride',
                               'pre': 'precision',
                               'q': 'query',
                               'rfc': 'receiveFocusCommand',
                               's': 'step',
                               'sbm': 'statusBarMessage',
                               'ut': 'useTemplate',
                               'v': 'value',
                               'vcc': 'visibleChangeCommand',
                               'vis': 'visible',
                               'w': 'width'},
                'type': 'windows'},
  'timeFieldGrp': {'callbackFlags': ['dragCommand',
                                     'dc',
                                     'visibleChangeCommand',
                                     'vcc',
                                     'dragCallback',
                                     'dgc',
                                     'changeCommand',
                                     'cc',
                                     'dropCallback',
                                     'dpc'],
                   'flags': {'adjustableColumn': {'args': '<type __builtin__.int>',
                                                  'longname': 'adjustableColumn',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'adj'},
                             'adjustableColumn2': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn2',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad2'},
                             'adjustableColumn3': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn3',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad3'},
                             'adjustableColumn4': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn4',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad4'},
                             'adjustableColumn5': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn5',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad5'},
                             'adjustableColumn6': {'args': '<type __builtin__.int>',
                                                   'longname': 'adjustableColumn6',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ad6'},
                             'annotation': {'args': '<type __builtin__.str>',
                                            'longname': 'annotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'ann'},
                             'backgroundColor': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'backgroundColor',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'bgc'},
                             'changeCommand': {'args': 'script',
                                               'longname': 'changeCommand',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'cc'},
                             'columnAlign': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.str>'],
                                             'longname': 'columnAlign',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 2,
                                             'shortname': 'cal'},
                             'columnAlign2': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign2',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'cl2'},
                             'columnAlign3': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign3',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'cl3'},
                             'columnAlign4': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign4',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 4,
                                              'shortname': 'cl4'},
                             'columnAlign5': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign5',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 5,
                                              'shortname': 'cl5'},
                             'columnAlign6': {'args': ['<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'columnAlign6',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 6,
                                              'shortname': 'cl6'},
                             'columnAttach': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.str>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnAttach',
                                              'modes': ['create',
                                                        'edit',
                                                        'multiuse'],
                                              'numArgs': 3,
                                              'shortname': 'cat'},
                             'columnAttach2': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'ct2'},
                             'columnAttach3': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'ct3'},
                             'columnAttach4': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'ct4'},
                             'columnAttach5': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'ct5'},
                             'columnAttach6': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'columnAttach6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'ct6'},
                             'columnOffset2': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset2',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'co2'},
                             'columnOffset3': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset3',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 3,
                                               'shortname': 'co3'},
                             'columnOffset4': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset4',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 4,
                                               'shortname': 'co4'},
                             'columnOffset5': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset5',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 5,
                                               'shortname': 'co5'},
                             'columnOffset6': {'args': ['<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'columnOffset6',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 6,
                                               'shortname': 'co6'},
                             'columnWidth': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'columnWidth',
                                             'modes': ['create',
                                                       'edit',
                                                       'multiuse'],
                                             'numArgs': 2,
                                             'shortname': 'cw'},
                             'columnWidth1': {'args': '<type __builtin__.int>',
                                              'longname': 'columnWidth1',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'cw1'},
                             'columnWidth2': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth2',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 2,
                                              'shortname': 'cw2'},
                             'columnWidth3': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth3',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 3,
                                              'shortname': 'cw3'},
                             'columnWidth4': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth4',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 4,
                                              'shortname': 'cw4'},
                             'columnWidth5': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth5',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 5,
                                              'shortname': 'cw5'},
                             'columnWidth6': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>',
                                                       '<type __builtin__.int>'],
                                              'longname': 'columnWidth6',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 6,
                                              'shortname': 'cw6'},
                             'defineTemplate': {'args': '<type __builtin__.str>',
                                                'longname': 'defineTemplate',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'dt'},
                             'docTag': {'args': '<type __builtin__.str>',
                                        'longname': 'docTag',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'dtg'},
                             'dragCallback': {'args': 'script',
                                              'longname': 'dragCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dgc'},
                             'dragCommand': {'args': 'script',
                                             'longname': 'dragCommand',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'dc'},
                             'dropCallback': {'args': 'script',
                                              'longname': 'dropCallback',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'dpc'},
                             'edit': {'args': '<type __builtin__.bool>',
                                      'longname': 'edit',
                                      'numArgs': 0,
                                      'shortname': 'e'},
                             'enable': {'args': '<type __builtin__.bool>',
                                        'longname': 'enable',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'en'},
                             'enable1': {'args': '<type __builtin__.bool>',
                                         'longname': 'enable1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'en1'},
                             'enable2': {'args': '<type __builtin__.bool>',
                                         'longname': 'enable2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'en2'},
                             'enable3': {'args': '<type __builtin__.bool>',
                                         'longname': 'enable3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'en3'},
                             'enable4': {'args': '<type __builtin__.bool>',
                                         'longname': 'enable4',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'en4'},
                             'enableBackground': {'args': '<type __builtin__.bool>',
                                                  'longname': 'enableBackground',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ebg'},
                             'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                     'longname': 'enableKeyboardFocus',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'ekf'},
                             'exists': {'args': '<type __builtin__.bool>',
                                        'longname': 'exists',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ex'},
                             'extraLabel': {'args': '<type __builtin__.str>',
                                            'longname': 'extraLabel',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'el'},
                             'fullPathName': {'args': '<type __builtin__.bool>',
                                              'longname': 'fullPathName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'fpn'},
                             'height': {'args': '<type __builtin__.int>',
                                        'longname': 'height',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'h'},
                             'highlightColor': {'args': ['<type __builtin__.float>',
                                                         '<type __builtin__.float>',
                                                         '<type __builtin__.float>'],
                                                'longname': 'highlightColor',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 3,
                                                'shortname': 'hlc'},
                             'isObscured': {'args': '<type __builtin__.bool>',
                                            'longname': 'isObscured',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'io'},
                             'label': {'args': '<type __builtin__.str>',
                                       'longname': 'label',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'l'},
                             'manage': {'args': '<type __builtin__.bool>',
                                        'longname': 'manage',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'm'},
                             'noBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'noBackground',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'nbg'},
                             'numberOfFields': {'args': '<type __builtin__.int>',
                                                'longname': 'numberOfFields',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'nf'},
                             'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                    'longname': 'numberOfPopupMenus',
                                                    'modes': ['query'],
                                                    'numArgs': 0,
                                                    'shortname': 'npm'},
                             'parent': {'args': '<type __builtin__.str>',
                                        'longname': 'parent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'p'},
                             'popupMenuArray': {'args': '<type __builtin__.bool>',
                                                'longname': 'popupMenuArray',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'pma'},
                             'precision': {'args': '<type __builtin__.int>',
                                           'longname': 'precision',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'pre'},
                             'preventOverride': {'args': '<type __builtin__.bool>',
                                                 'longname': 'preventOverride',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'po'},
                             'query': {'args': '<type __builtin__.bool>',
                                       'longname': 'query',
                                       'numArgs': 0,
                                       'shortname': 'q'},
                             'rowAttach': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.str>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'rowAttach',
                                           'modes': ['create',
                                                     'edit',
                                                     'multiuse'],
                                           'numArgs': 3,
                                           'shortname': 'rat'},
                             'statusBarMessage': {'args': '<type __builtin__.str>',
                                                  'longname': 'statusBarMessage',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'sbm'},
                             'step': {'args': 'time',
                                      'longname': 'step',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 's'},
                             'useTemplate': {'args': '<type __builtin__.str>',
                                             'longname': 'useTemplate',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ut'},
                             'value': {'args': ['time',
                                                'time',
                                                'time',
                                                'time'],
                                       'longname': 'value',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 4,
                                       'shortname': 'v'},
                             'value1': {'args': 'time',
                                        'longname': 'value1',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'v1'},
                             'value2': {'args': 'time',
                                        'longname': 'value2',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'v2'},
                             'value3': {'args': 'time',
                                        'longname': 'value3',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'v3'},
                             'value4': {'args': 'time',
                                        'longname': 'value4',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'v4'},
                             'visible': {'args': '<type __builtin__.bool>',
                                         'longname': 'visible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'vis'},
                             'visibleChangeCommand': {'args': 'script',
                                                      'longname': 'visibleChangeCommand',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'vcc'},
                             'width': {'args': '<type __builtin__.int>',
                                       'longname': 'width',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'w'}},
                   'shortFlags': {'ad2': 'adjustableColumn2',
                                  'ad3': 'adjustableColumn3',
                                  'ad4': 'adjustableColumn4',
                                  'ad5': 'adjustableColumn5',
                                  'ad6': 'adjustableColumn6',
                                  'adj': 'adjustableColumn',
                                  'ann': 'annotation',
                                  'bgc': 'backgroundColor',
                                  'cal': 'columnAlign',
                                  'cat': 'columnAttach',
                                  'cc': 'changeCommand',
                                  'cl2': 'columnAlign2',
                                  'cl3': 'columnAlign3',
                                  'cl4': 'columnAlign4',
                                  'cl5': 'columnAlign5',
                                  'cl6': 'columnAlign6',
                                  'co2': 'columnOffset2',
                                  'co3': 'columnOffset3',
                                  'co4': 'columnOffset4',
                                  'co5': 'columnOffset5',
                                  'co6': 'columnOffset6',
                                  'ct2': 'columnAttach2',
                                  'ct3': 'columnAttach3',
                                  'ct4': 'columnAttach4',
                                  'ct5': 'columnAttach5',
                                  'ct6': 'columnAttach6',
                                  'cw': 'columnWidth',
                                  'cw1': 'columnWidth1',
                                  'cw2': 'columnWidth2',
                                  'cw3': 'columnWidth3',
                                  'cw4': 'columnWidth4',
                                  'cw5': 'columnWidth5',
                                  'cw6': 'columnWidth6',
                                  'dc': 'dragCommand',
                                  'dgc': 'dragCallback',
                                  'dpc': 'dropCallback',
                                  'dt': 'defineTemplate',
                                  'dtg': 'docTag',
                                  'e': 'edit',
                                  'ebg': 'enableBackground',
                                  'ekf': 'enableKeyboardFocus',
                                  'el': 'extraLabel',
                                  'en': 'enable',
                                  'en1': 'enable1',
                                  'en2': 'enable2',
                                  'en3': 'enable3',
                                  'en4': 'enable4',
                                  'ex': 'exists',
                                  'fpn': 'fullPathName',
                                  'h': 'height',
                                  'hlc': 'highlightColor',
                                  'io': 'isObscured',
                                  'l': 'label',
                                  'm': 'manage',
                                  'nbg': 'noBackground',
                                  'nf': 'numberOfFields',
                                  'npm': 'numberOfPopupMenus',
                                  'p': 'parent',
                                  'pma': 'popupMenuArray',
                                  'po': 'preventOverride',
                                  'pre': 'precision',
                                  'q': 'query',
                                  'rat': 'rowAttach',
                                  's': 'step',
                                  'sbm': 'statusBarMessage',
                                  'ut': 'useTemplate',
                                  'v': 'value',
                                  'v1': 'value1',
                                  'v2': 'value2',
                                  'v3': 'value3',
                                  'v4': 'value4',
                                  'vcc': 'visibleChangeCommand',
                                  'vis': 'visible',
                                  'w': 'width'},
                   'type': 'windows'},
  'timePort': {'callbackFlags': ['visibleChangeCommand',
                                 'vcc',
                                 'dragCallback',
                                 'dgc',
                                 'dropCallback',
                                 'dpc'],
               'flags': {'annotation': {'args': '<type __builtin__.str>',
                                        'longname': 'annotation',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ann'},
                         'backgroundColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'backgroundColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'bgc'},
                         'defineTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'defineTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'dt'},
                         'docTag': {'args': '<type __builtin__.str>',
                                    'longname': 'docTag',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'dtg'},
                         'dragCallback': {'args': 'script',
                                          'longname': 'dragCallback',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dgc'},
                         'dropCallback': {'args': 'script',
                                          'longname': 'dropCallback',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dpc'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'enable': {'args': '<type __builtin__.bool>',
                                    'longname': 'enable',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'en'},
                         'enableBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'enableBackground',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ebg'},
                         'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableKeyboardFocus',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ekf'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'fullPathName': {'args': '<type __builtin__.bool>',
                                          'longname': 'fullPathName',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'fpn'},
                         'globalTime': {'args': '<type __builtin__.bool>',
                                        'longname': 'globalTime',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'gt'},
                         'height': {'args': '<type __builtin__.int>',
                                    'longname': 'height',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'h'},
                         'highlightColor': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'highlightColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'hlc'},
                         'isObscured': {'args': '<type __builtin__.bool>',
                                        'longname': 'isObscured',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'io'},
                         'manage': {'args': '<type __builtin__.bool>',
                                    'longname': 'manage',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'm'},
                         'noBackground': {'args': '<type __builtin__.bool>',
                                          'longname': 'noBackground',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nbg'},
                         'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                'longname': 'numberOfPopupMenus',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'npm'},
                         'parent': {'args': '<type __builtin__.str>',
                                    'longname': 'parent',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'p'},
                         'popupMenuArray': {'args': '<type __builtin__.bool>',
                                            'longname': 'popupMenuArray',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'pma'},
                         'preventOverride': {'args': '<type __builtin__.bool>',
                                             'longname': 'preventOverride',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'po'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'snap': {'args': '<type __builtin__.bool>',
                                  'longname': 'snap',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'sn'},
                         'statusBarMessage': {'args': '<type __builtin__.str>',
                                              'longname': 'statusBarMessage',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sbm'},
                         'useTemplate': {'args': '<type __builtin__.str>',
                                         'longname': 'useTemplate',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ut'},
                         'visible': {'args': '<type __builtin__.bool>',
                                     'longname': 'visible',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'vis'},
                         'visibleChangeCommand': {'args': 'script',
                                                  'longname': 'visibleChangeCommand',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'vcc'},
                         'width': {'args': '<type __builtin__.int>',
                                   'longname': 'width',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'w'}},
               'shortFlags': {'ann': 'annotation',
                              'bgc': 'backgroundColor',
                              'dgc': 'dragCallback',
                              'dpc': 'dropCallback',
                              'dt': 'defineTemplate',
                              'dtg': 'docTag',
                              'e': 'edit',
                              'ebg': 'enableBackground',
                              'ekf': 'enableKeyboardFocus',
                              'en': 'enable',
                              'ex': 'exists',
                              'fpn': 'fullPathName',
                              'gt': 'globalTime',
                              'h': 'height',
                              'hlc': 'highlightColor',
                              'io': 'isObscured',
                              'm': 'manage',
                              'nbg': 'noBackground',
                              'npm': 'numberOfPopupMenus',
                              'p': 'parent',
                              'pma': 'popupMenuArray',
                              'po': 'preventOverride',
                              'q': 'query',
                              'sbm': 'statusBarMessage',
                              'sn': 'snap',
                              'ut': 'useTemplate',
                              'vcc': 'visibleChangeCommand',
                              'vis': 'visible',
                              'w': 'width'},
               'type': 'windows'},
  'timeSliderCustomDraw': {'flags': {'clearPrimitives': {'args': '<type __builtin__.bool>',
                                                         'longname': 'clearPrimitives',
                                                         'numArgs': 0,
                                                         'shortname': 'cp'},
                                     'color': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'color',
                                               'numArgs': 4,
                                               'shortname': 'clr'},
                                     'deregister': {'args': '<type __builtin__.int>',
                                                    'longname': 'deregister',
                                                    'numArgs': 1,
                                                    'shortname': 'der'},
                                     'edit': {'args': '<type __builtin__.bool>',
                                              'longname': 'edit',
                                              'numArgs': 0,
                                              'shortname': 'e'},
                                     'height': {'args': '<type __builtin__.int>',
                                                'longname': 'height',
                                                'numArgs': 1,
                                                'shortname': 'h'},
                                     'layer': {'args': '<type __builtin__.int>',
                                               'longname': 'layer',
                                               'numArgs': 1,
                                               'shortname': 'l'},
                                     'location': {'args': '<type __builtin__.int>',
                                                  'longname': 'location',
                                                  'numArgs': 1,
                                                  'shortname': 'loc'},
                                     'priority': {'args': '<type __builtin__.int>',
                                                  'longname': 'priority',
                                                  'numArgs': 1,
                                                  'shortname': 'p'},
                                     'query': {'args': '<type __builtin__.bool>',
                                               'longname': 'query',
                                               'numArgs': 0,
                                               'shortname': 'q'},
                                     'registerAbove': {'args': '<type __builtin__.str>',
                                                       'longname': 'registerAbove',
                                                       'numArgs': 1,
                                                       'shortname': 'rab'},
                                     'registerBelow': {'args': '<type __builtin__.str>',
                                                       'longname': 'registerBelow',
                                                       'numArgs': 1,
                                                       'shortname': 'rbe'},
                                     'registerOn': {'args': ['<type __builtin__.str>',
                                                             '<type __builtin__.int>'],
                                                    'longname': 'registerOn',
                                                    'numArgs': 2,
                                                    'shortname': 'ron'},
                                     'setPrimitives': {'args': ['<type __builtin__.str>',
                                                                '<type __builtin__.float>',
                                                                '<type __builtin__.float>'],
                                                       'longname': 'setPrimitives',
                                                       'numArgs': 3,
                                                       'shortname': 'sp'},
                                     'visible': {'args': '<type __builtin__.bool>',
                                                 'longname': 'visible',
                                                 'numArgs': 1,
                                                 'shortname': 'v'}},
                           'shortFlags': {'clr': 'color',
                                          'cp': 'clearPrimitives',
                                          'der': 'deregister',
                                          'e': 'edit',
                                          'h': 'height',
                                          'l': 'layer',
                                          'loc': 'location',
                                          'p': 'priority',
                                          'q': 'query',
                                          'rab': 'registerAbove',
                                          'rbe': 'registerBelow',
                                          'ron': 'registerOn',
                                          'sp': 'setPrimitives',
                                          'v': 'visible'},
                           'type': 'other'},
  'timeWarp': {'flags': {'deleteFrame': {'args': '<type __builtin__.int>',
                                         'longname': 'deleteFrame',
                                         'modes': ['edit'],
                                         'numArgs': 1,
                                         'shortname': 'df'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'frame': {'args': '<type __builtin__.float>',
                                   'longname': 'frame',
                                   'modes': ['create',
                                             'query',
                                             'edit',
                                             'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 'f'},
                         'g': {'args': '<type __builtin__.bool>',
                               'longname': 'g',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 0,
                               'shortname': 'g'},
                         'interpType': {'args': ['<type __builtin__.int>',
                                                 '<type __builtin__.str>'],
                                        'longname': 'interpType',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 2,
                                        'shortname': 'it'},
                         'moveFrame': {'args': ['<type __builtin__.int>',
                                                '<type __builtin__.float>'],
                                       'longname': 'moveFrame',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 2,
                                       'shortname': 'mf'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'}},
               'removedFlags': {'global': 'g'},
               'shortFlags': {'df': 'deleteFrame',
                              'e': 'edit',
                              'f': 'frame',
                              'g': 'g',
                              'it': 'interpType',
                              'mf': 'moveFrame',
                              'q': 'query'},
               'type': 'animation'},
  'timer': {'flags': {'endTimer': {'args': '<type __builtin__.bool>',
                                   'longname': 'endTimer',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'e'},
                      'lapTime': {'args': '<type __builtin__.bool>',
                                  'longname': 'lapTime',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'lap'},
                      'name': {'args': '<type __builtin__.str>',
                               'longname': 'name',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'n'},
                      'startTimer': {'args': '<type __builtin__.bool>',
                                     'longname': 'startTimer',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 's'}},
            'shortFlags': {'e': 'endTimer',
                           'lap': 'lapTime',
                           'n': 'name',
                           's': 'startTimer'},
            'type': 'system'},
  'timerX': {'flags': {'startTime': {'args': '<type __builtin__.float>',
                                     'longname': 'startTime',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'st'}},
             'shortFlags': {'st': 'startTime'},
             'type': 'system'},
  'toggle': {'flags': {'above': {'args': '<type __builtin__.bool>',
                                 'longname': 'above',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'a'},
                       'below': {'args': '<type __builtin__.bool>',
                                 'longname': 'below',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'b'},
                       'boundary': {'args': '<type __builtin__.bool>',
                                    'longname': 'boundary',
                                    'modes': ['create', 'query'],
                                    'numArgs': 0,
                                    'shortname': 'bn'},
                       'boundingBox': {'args': '<type __builtin__.bool>',
                                       'longname': 'boundingBox',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'box'},
                       'controlVertex': {'args': '<type __builtin__.bool>',
                                         'longname': 'controlVertex',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'cv'},
                       'doNotWrite': {'args': '<type __builtin__.bool>',
                                      'longname': 'doNotWrite',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'dnw'},
                       'editPoint': {'args': '<type __builtin__.bool>',
                                     'longname': 'editPoint',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 'ep'},
                       'extent': {'args': '<type __builtin__.bool>',
                                  'longname': 'extent',
                                  'modes': ['create', 'query'],
                                  'numArgs': 0,
                                  'shortname': 'et'},
                       'facet': {'args': '<type __builtin__.bool>',
                                 'longname': 'facet',
                                 'modes': ['create', 'query'],
                                 'numArgs': 0,
                                 'shortname': 'f'},
                       'geometry': {'args': '<type __builtin__.bool>',
                                    'longname': 'geometry',
                                    'modes': ['create', 'query'],
                                    'numArgs': 0,
                                    'shortname': 'g'},
                       'gl': {'args': '<type __builtin__.bool>',
                              'longname': 'gl',
                              'modes': ['create'],
                              'numArgs': 0,
                              'shortname': 'gl'},
                       'highPrecisionNurbs': {'args': '<type __builtin__.bool>',
                                              'longname': 'highPrecisionNurbs',
                                              'modes': ['create', 'query'],
                                              'numArgs': 0,
                                              'shortname': 'hpn'},
                       'hull': {'args': '<type __builtin__.bool>',
                                'longname': 'hull',
                                'modes': ['create', 'query'],
                                'numArgs': 0,
                                'shortname': 'hl'},
                       'latticePoint': {'args': '<type __builtin__.bool>',
                                        'longname': 'latticePoint',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'lp'},
                       'latticeShape': {'args': '<type __builtin__.bool>',
                                        'longname': 'latticeShape',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'ls'},
                       'localAxis': {'args': '<type __builtin__.bool>',
                                     'longname': 'localAxis',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 'la'},
                       'newCurve': {'args': '<type __builtin__.bool>',
                                    'longname': 'newCurve',
                                    'modes': ['create', 'query'],
                                    'numArgs': 0,
                                    'shortname': 'nc'},
                       'newPolymesh': {'args': '<type __builtin__.bool>',
                                       'longname': 'newPolymesh',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'np'},
                       'newSurface': {'args': '<type __builtin__.bool>',
                                      'longname': 'newSurface',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'ns'},
                       'normal': {'args': '<type __builtin__.bool>',
                                  'longname': 'normal',
                                  'modes': ['create', 'query'],
                                  'numArgs': 0,
                                  'shortname': 'nr'},
                       'origin': {'args': '<type __builtin__.bool>',
                                  'longname': 'origin',
                                  'modes': ['create', 'query'],
                                  'numArgs': 0,
                                  'shortname': 'o'},
                       'point': {'args': '<type __builtin__.bool>',
                                 'longname': 'point',
                                 'modes': ['create', 'query'],
                                 'numArgs': 0,
                                 'shortname': 'pt'},
                       'pointDisplay': {'args': '<type __builtin__.bool>',
                                        'longname': 'pointDisplay',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'pd'},
                       'pointFacet': {'args': '<type __builtin__.bool>',
                                      'longname': 'pointFacet',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'pf'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'rotatePivot': {'args': '<type __builtin__.bool>',
                                       'longname': 'rotatePivot',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'rp'},
                       'scalePivot': {'args': '<type __builtin__.bool>',
                                      'longname': 'scalePivot',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'sp'},
                       'selectHandle': {'args': '<type __builtin__.bool>',
                                        'longname': 'selectHandle',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'sh'},
                       'state': {'args': '<type __builtin__.bool>',
                                 'longname': 'state',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'st'},
                       'surfaceFace': {'args': '<type __builtin__.bool>',
                                       'longname': 'surfaceFace',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'sf'},
                       'template': {'args': '<type __builtin__.bool>',
                                    'longname': 'template',
                                    'modes': ['create', 'query'],
                                    'numArgs': 0,
                                    'shortname': 'te'},
                       'uvCoords': {'args': '<type __builtin__.bool>',
                                    'longname': 'uvCoords',
                                    'modes': ['create', 'query'],
                                    'numArgs': 0,
                                    'shortname': 'uv'},
                       'vertex': {'args': '<type __builtin__.bool>',
                                  'longname': 'vertex',
                                  'modes': ['create', 'query'],
                                  'numArgs': 0,
                                  'shortname': 'vt'}},
             'removedFlags': {'global': 'gl'},
             'shortFlags': {'a': 'above',
                            'b': 'below',
                            'bn': 'boundary',
                            'box': 'boundingBox',
                            'cv': 'controlVertex',
                            'dnw': 'doNotWrite',
                            'ep': 'editPoint',
                            'et': 'extent',
                            'f': 'facet',
                            'g': 'geometry',
                            'gl': 'gl',
                            'hl': 'hull',
                            'hpn': 'highPrecisionNurbs',
                            'la': 'localAxis',
                            'lp': 'latticePoint',
                            'ls': 'latticeShape',
                            'nc': 'newCurve',
                            'np': 'newPolymesh',
                            'nr': 'normal',
                            'ns': 'newSurface',
                            'o': 'origin',
                            'pd': 'pointDisplay',
                            'pf': 'pointFacet',
                            'pt': 'point',
                            'q': 'query',
                            'rp': 'rotatePivot',
                            'sf': 'surfaceFace',
                            'sh': 'selectHandle',
                            'sp': 'scalePivot',
                            'st': 'state',
                            'te': 'template',
                            'uv': 'uvCoords',
                            'vt': 'vertex'},
             'type': 'general'},
  'toggleAxis': {'flags': {'origin': {'args': '<type __builtin__.bool>',
                                      'longname': 'origin',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'o'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'view': {'args': '<type __builtin__.bool>',
                                    'longname': 'view',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'v'}},
                 'shortFlags': {'o': 'origin', 'q': 'query', 'v': 'view'},
                 'type': 'general'},
  'toggleDisplacement': {'flags': {}, 'shortFlags': {}, 'type': 'general'},
  'toggleWindowVisibility': {'flags': {},
                             'shortFlags': {},
                             'type': 'windows'},
  'tolerance': {'flags': {'angular': {'args': '<type __builtin__.float>',
                                      'longname': 'angular',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'a'},
                          'linear': {'args': '<type __builtin__.float>',
                                     'longname': 'linear',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'l'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'}},
                'shortFlags': {'a': 'angular', 'l': 'linear', 'q': 'query'},
                'type': 'modeling'},
  'toolBar': {'callbackFlags': ['visibleChangeCommand',
                                'vcc',
                                'dragCallback',
                                'dgc',
                                'dropCallback',
                                'dpc'],
              'flags': {'allowedArea': {'args': '<type __builtin__.str>',
                                        'longname': 'allowedArea',
                                        'modes': ['create',
                                                  'query',
                                                  'edit',
                                                  'multiuse'],
                                        'numArgs': 1,
                                        'shortname': 'aa'},
                        'annotation': {'args': '<type __builtin__.str>',
                                       'longname': 'annotation',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ann'},
                        'area': {'args': '<type __builtin__.str>',
                                 'longname': 'area',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'a'},
                        'backgroundColor': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'backgroundColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'bgc'},
                        'content': {'args': '<type __builtin__.str>',
                                    'longname': 'content',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'con'},
                        'defineTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'defineTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'dt'},
                        'docTag': {'args': '<type __builtin__.str>',
                                   'longname': 'docTag',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'dtg'},
                        'dragCallback': {'args': 'script',
                                         'longname': 'dragCallback',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dgc'},
                        'dropCallback': {'args': 'script',
                                         'longname': 'dropCallback',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dpc'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'enable': {'args': '<type __builtin__.bool>',
                                   'longname': 'enable',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'en'},
                        'enableBackground': {'args': '<type __builtin__.bool>',
                                             'longname': 'enableBackground',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'ebg'},
                        'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableKeyboardFocus',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ekf'},
                        'exists': {'args': '<type __builtin__.bool>',
                                   'longname': 'exists',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'ex'},
                        'fullPathName': {'args': '<type __builtin__.bool>',
                                         'longname': 'fullPathName',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'fpn'},
                        'height': {'args': '<type __builtin__.int>',
                                   'longname': 'height',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'h'},
                        'highlightColor': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'highlightColor',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 3,
                                           'shortname': 'hlc'},
                        'isObscured': {'args': '<type __builtin__.bool>',
                                       'longname': 'isObscured',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'io'},
                        'label': {'args': '<type __builtin__.str>',
                                  'longname': 'label',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'l'},
                        'manage': {'args': '<type __builtin__.bool>',
                                   'longname': 'manage',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'm'},
                        'noBackground': {'args': '<type __builtin__.bool>',
                                         'longname': 'noBackground',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'nbg'},
                        'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                               'longname': 'numberOfPopupMenus',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'npm'},
                        'parent': {'args': '<type __builtin__.str>',
                                   'longname': 'parent',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'p'},
                        'popupMenuArray': {'args': '<type __builtin__.bool>',
                                           'longname': 'popupMenuArray',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'pma'},
                        'preventOverride': {'args': '<type __builtin__.bool>',
                                            'longname': 'preventOverride',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'po'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'statusBarMessage': {'args': '<type __builtin__.str>',
                                             'longname': 'statusBarMessage',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'sbm'},
                        'useTemplate': {'args': '<type __builtin__.str>',
                                        'longname': 'useTemplate',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ut'},
                        'visible': {'args': '<type __builtin__.bool>',
                                    'longname': 'visible',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'vis'},
                        'visibleChangeCommand': {'args': 'script',
                                                 'longname': 'visibleChangeCommand',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'vcc'},
                        'width': {'args': '<type __builtin__.int>',
                                  'longname': 'width',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'w'}},
              'shortFlags': {'a': 'area',
                             'aa': 'allowedArea',
                             'ann': 'annotation',
                             'bgc': 'backgroundColor',
                             'con': 'content',
                             'dgc': 'dragCallback',
                             'dpc': 'dropCallback',
                             'dt': 'defineTemplate',
                             'dtg': 'docTag',
                             'e': 'edit',
                             'ebg': 'enableBackground',
                             'ekf': 'enableKeyboardFocus',
                             'en': 'enable',
                             'ex': 'exists',
                             'fpn': 'fullPathName',
                             'h': 'height',
                             'hlc': 'highlightColor',
                             'io': 'isObscured',
                             'l': 'label',
                             'm': 'manage',
                             'nbg': 'noBackground',
                             'npm': 'numberOfPopupMenus',
                             'p': 'parent',
                             'pma': 'popupMenuArray',
                             'po': 'preventOverride',
                             'q': 'query',
                             'sbm': 'statusBarMessage',
                             'ut': 'useTemplate',
                             'vcc': 'visibleChangeCommand',
                             'vis': 'visible',
                             'w': 'width'},
              'type': 'windows'},
  'toolButton': {'callbackFlags': ['doubleClickCommand',
                                   'dcc',
                                   'visibleChangeCommand',
                                   'vcc',
                                   'dragCallback',
                                   'dgc',
                                   'offCommand',
                                   'ofc',
                                   'changeCommand',
                                   'cc',
                                   'onCommand',
                                   'onc',
                                   'dropCallback',
                                   'dpc'],
                 'flags': {'allowMultipleTools': {'args': '<type __builtin__.bool>',
                                                  'longname': 'allowMultipleTools',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'amt'},
                           'annotation': {'args': '<type __builtin__.str>',
                                          'longname': 'annotation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ann'},
                           'backgroundColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'backgroundColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'bgc'},
                           'changeCommand': {'args': 'script',
                                             'longname': 'changeCommand',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'cc'},
                           'collection': {'args': '<type __builtin__.str>',
                                          'longname': 'collection',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'cl'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'doubleClickCommand': {'args': 'script',
                                                  'longname': 'doubleClickCommand',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dcc'},
                           'dragCallback': {'args': 'script',
                                            'longname': 'dragCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dgc'},
                           'dropCallback': {'args': 'script',
                                            'longname': 'dropCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dpc'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'enable': {'args': '<type __builtin__.bool>',
                                      'longname': 'enable',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'en'},
                           'enableBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableBackground',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ebg'},
                           'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableKeyboardFocus',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ekf'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'fullPathName': {'args': '<type __builtin__.bool>',
                                            'longname': 'fullPathName',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fpn'},
                           'height': {'args': '<type __builtin__.int>',
                                      'longname': 'height',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                           'highlightColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'highlightColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'hlc'},
                           'image1': {'args': '<type __builtin__.str>',
                                      'longname': 'image1',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i1'},
                           'image2': {'args': '<type __builtin__.str>',
                                      'longname': 'image2',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i2'},
                           'image3': {'args': '<type __builtin__.str>',
                                      'longname': 'image3',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'i3'},
                           'imageOverlayLabel': {'args': '<type __builtin__.str>',
                                                 'longname': 'imageOverlayLabel',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'iol'},
                           'isObscured': {'args': '<type __builtin__.bool>',
                                          'longname': 'isObscured',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'io'},
                           'ltVersion': {'args': '<type __builtin__.str>',
                                         'longname': 'ltVersion',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'lt'},
                           'manage': {'args': '<type __builtin__.bool>',
                                      'longname': 'manage',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                           'noBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'noBackground',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nbg'},
                           'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'numberOfPopupMenus',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'npm'},
                           'offCommand': {'args': 'script',
                                          'longname': 'offCommand',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ofc'},
                           'onCommand': {'args': 'script',
                                         'longname': 'onCommand',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'onc'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupIndicatorVisible': {'args': '<type __builtin__.bool>',
                                                     'longname': 'popupIndicatorVisible',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'piv'},
                           'popupMenuArray': {'args': '<type __builtin__.bool>',
                                              'longname': 'popupMenuArray',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'pma'},
                           'preventOverride': {'args': '<type __builtin__.bool>',
                                               'longname': 'preventOverride',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'po'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'select': {'args': '<type __builtin__.bool>',
                                      'longname': 'select',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'sl'},
                           'statusBarMessage': {'args': '<type __builtin__.str>',
                                                'longname': 'statusBarMessage',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sbm'},
                           'style': {'args': '<type __builtin__.str>',
                                     'longname': 'style',
                                     'modes': ['create', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'st'},
                           'tool': {'args': '<type __builtin__.str>',
                                    'longname': 'tool',
                                    'modes': ['create',
                                              'query',
                                              'edit',
                                              'multiuse'],
                                    'numArgs': 1,
                                    'shortname': 't'},
                           'toolArray': {'args': '<type __builtin__.bool>',
                                         'longname': 'toolArray',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'ta'},
                           'toolCount': {'args': '<type __builtin__.bool>',
                                         'longname': 'toolCount',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'tc'},
                           'toolImage1': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'toolImage1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'ti1'},
                           'toolImage2': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'toolImage2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'ti2'},
                           'toolImage3': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'toolImage3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'ti3'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'},
                           'version': {'args': '<type __builtin__.str>',
                                       'longname': 'version',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ver'},
                           'visible': {'args': '<type __builtin__.bool>',
                                       'longname': 'visible',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vis'},
                           'visibleChangeCommand': {'args': 'script',
                                                    'longname': 'visibleChangeCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'vcc'},
                           'width': {'args': '<type __builtin__.int>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'}},
                 'shortFlags': {'amt': 'allowMultipleTools',
                                'ann': 'annotation',
                                'bgc': 'backgroundColor',
                                'cc': 'changeCommand',
                                'cl': 'collection',
                                'dcc': 'doubleClickCommand',
                                'dgc': 'dragCallback',
                                'dpc': 'dropCallback',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'ebg': 'enableBackground',
                                'ekf': 'enableKeyboardFocus',
                                'en': 'enable',
                                'ex': 'exists',
                                'fpn': 'fullPathName',
                                'h': 'height',
                                'hlc': 'highlightColor',
                                'i1': 'image1',
                                'i2': 'image2',
                                'i3': 'image3',
                                'io': 'isObscured',
                                'iol': 'imageOverlayLabel',
                                'lt': 'ltVersion',
                                'm': 'manage',
                                'nbg': 'noBackground',
                                'npm': 'numberOfPopupMenus',
                                'ofc': 'offCommand',
                                'onc': 'onCommand',
                                'p': 'parent',
                                'piv': 'popupIndicatorVisible',
                                'pma': 'popupMenuArray',
                                'po': 'preventOverride',
                                'q': 'query',
                                'sbm': 'statusBarMessage',
                                'sl': 'select',
                                'st': 'style',
                                't': 'tool',
                                'ta': 'toolArray',
                                'tc': 'toolCount',
                                'ti1': 'toolImage1',
                                'ti2': 'toolImage2',
                                'ti3': 'toolImage3',
                                'ut': 'useTemplate',
                                'vcc': 'visibleChangeCommand',
                                'ver': 'version',
                                'vis': 'visible',
                                'w': 'width'},
                 'type': 'windows'},
  'toolCollection': {'flags': {'collectionItemArray': {'args': '<type __builtin__.bool>',
                                                       'longname': 'collectionItemArray',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'cia'},
                               'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'gl': {'args': '<type __builtin__.bool>',
                                      'longname': 'gl',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'gl'},
                               'numberOfCollectionItems': {'args': '<type __builtin__.bool>',
                                                           'longname': 'numberOfCollectionItems',
                                                           'modes': ['query'],
                                                           'numArgs': 0,
                                                           'shortname': 'nci'},
                               'parent': {'args': '<type __builtin__.str>',
                                          'longname': 'parent',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'select': {'args': '<type __builtin__.str>',
                                          'longname': 'select',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'sl'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'}},
                     'removedFlags': {'global': 'gl'},
                     'shortFlags': {'cia': 'collectionItemArray',
                                    'dt': 'defineTemplate',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'gl': 'gl',
                                    'nci': 'numberOfCollectionItems',
                                    'p': 'parent',
                                    'q': 'query',
                                    'sl': 'select',
                                    'ut': 'useTemplate'},
                     'type': 'windows'},
  'toolDropped': {'flags': {}, 'shortFlags': {}, 'type': 'general'},
  'toolHasOptions': {'flags': {}, 'shortFlags': {}, 'type': 'general'},
  'toolPropertyWindow': {'callbackFlags': ['showCommand',
                                           'shw',
                                           'selectCommand',
                                           'sel'],
                         'flags': {'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'field': {'args': '<type __builtin__.str>',
                                             'longname': 'field',
                                             'modes': ['query', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'fld'},
                                   'helpButton': {'args': '<type __builtin__.str>',
                                                  'longname': 'helpButton',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'hb'},
                                   'icon': {'args': '<type __builtin__.str>',
                                            'longname': 'icon',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'icn'},
                                   'inMainWindow': {'args': '<type __builtin__.bool>',
                                                    'longname': 'inMainWindow',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'imw'},
                                   'location': {'args': '<type __builtin__.str>',
                                                'longname': 'location',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'loc'},
                                   'noviceMode': {'args': '<type __builtin__.bool>',
                                                  'longname': 'noviceMode',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'nm'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'refresh': {'args': '<type __builtin__.bool>',
                                               'longname': 'refresh',
                                               'numArgs': 0,
                                               'shortname': 'rf'},
                                   'resetButton': {'args': '<type __builtin__.str>',
                                                   'longname': 'resetButton',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'rb'},
                                   'restore': {'args': '<type __builtin__.bool>',
                                               'longname': 'restore',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'rs'},
                                   'selectCommand': {'args': '<type __builtin__.str>',
                                                     'longname': 'selectCommand',
                                                     'modes': ['query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sel'},
                                   'showCommand': {'args': '<type __builtin__.str>',
                                                   'longname': 'showCommand',
                                                   'modes': ['query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'shw'}},
                         'shortFlags': {'e': 'edit',
                                        'fld': 'field',
                                        'hb': 'helpButton',
                                        'icn': 'icon',
                                        'imw': 'inMainWindow',
                                        'loc': 'location',
                                        'nm': 'noviceMode',
                                        'q': 'query',
                                        'rb': 'resetButton',
                                        'rf': 'refresh',
                                        'rs': 'restore',
                                        'sel': 'selectCommand',
                                        'shw': 'showCommand'},
                         'type': 'general'},
  'torus': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                        '<type __builtin__.float>',
                                        '<type __builtin__.float>'],
                               'longname': 'axis',
                               'modes': ['create', 'query', 'edit'],
                               'numArgs': 3,
                               'shortname': 'ax'},
                      'caching': {'args': '<type __builtin__.bool>',
                                  'longname': 'caching',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'resultNeedsCasting': True,
                                  'shortname': 'cch'},
                      'constructionHistory': {'args': '<type __builtin__.bool>',
                                              'longname': 'constructionHistory',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'ch'},
                      'degree': {'args': '<type __builtin__.int>',
                                 'longname': 'degree',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'resultNeedsCasting': True,
                                 'shortname': 'd'},
                      'edit': {'args': '<type __builtin__.bool>',
                               'longname': 'edit',
                               'numArgs': 0,
                               'shortname': 'e'},
                      'endSweep': {'args': '<type __builtin__.float>',
                                   'longname': 'endSweep',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'esw'},
                      'frozen': {'args': '<type __builtin__.bool>',
                                 'longname': 'frozen',
                                 'modes': ['query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'fzn'},
                      'heightRatio': {'args': '<type __builtin__.float>',
                                      'longname': 'heightRatio',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'hr'},
                      'minorSweep': {'args': '<type __builtin__.float>',
                                     'longname': 'minorSweep',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'msw'},
                      'name': {'args': '<type __builtin__.str>',
                               'longname': 'name',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'n'},
                      'nodeState': {'args': '<type __builtin__.int>',
                                    'longname': 'nodeState',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'resultNeedsCasting': True,
                                    'shortname': 'nds'},
                      'object': {'args': '<type __builtin__.bool>',
                                 'longname': 'object',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'o'},
                      'pivot': {'args': ['<type __builtin__.float>',
                                         '<type __builtin__.float>',
                                         '<type __builtin__.float>'],
                                'longname': 'pivot',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 3,
                                'shortname': 'p'},
                      'polygon': {'args': '<type __builtin__.int>',
                                  'longname': 'polygon',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'po'},
                      'query': {'args': '<type __builtin__.bool>',
                                'longname': 'query',
                                'numArgs': 0,
                                'shortname': 'q'},
                      'radius': {'args': '<type __builtin__.float>',
                                 'longname': 'radius',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'r'},
                      'sections': {'args': '<type __builtin__.int>',
                                   'longname': 'sections',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'resultNeedsCasting': True,
                                   'shortname': 's'},
                      'spans': {'args': '<type __builtin__.int>',
                                'longname': 'spans',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'resultNeedsCasting': True,
                                'shortname': 'nsp'},
                      'startSweep': {'args': '<type __builtin__.float>',
                                     'longname': 'startSweep',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'ssw'},
                      'tolerance': {'args': '<type __builtin__.float>',
                                    'longname': 'tolerance',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'tol'},
                      'useTolerance': {'args': '<type __builtin__.bool>',
                                       'longname': 'useTolerance',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'resultNeedsCasting': True,
                                       'shortname': 'ut'}},
            'shortFlags': {'ax': 'axis',
                           'cch': 'caching',
                           'ch': 'constructionHistory',
                           'd': 'degree',
                           'e': 'edit',
                           'esw': 'endSweep',
                           'fzn': 'frozen',
                           'hr': 'heightRatio',
                           'msw': 'minorSweep',
                           'n': 'name',
                           'nds': 'nodeState',
                           'nsp': 'spans',
                           'o': 'object',
                           'p': 'pivot',
                           'po': 'polygon',
                           'q': 'query',
                           'r': 'radius',
                           's': 'sections',
                           'ssw': 'startSweep',
                           'tol': 'tolerance',
                           'ut': 'useTolerance'},
            'type': 'modeling'},
  'track': {'flags': {'down': {'args': '<type __builtin__.float>',
                               'longname': 'down',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'd'},
                      'left': {'args': '<type __builtin__.float>',
                               'longname': 'left',
                               'modes': ['create'],
                               'numArgs': 1,
                               'shortname': 'l'},
                      'right': {'args': '<type __builtin__.float>',
                                'longname': 'right',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'r'},
                      'upDistance01': {'args': '<type __builtin__.float>',
                                       'longname': 'upDistance01',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'u'},
                      'upDistance02': {'args': '<type __builtin__.float>',
                                       'longname': 'upDistance02',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'up'}},
            'shortFlags': {'d': 'down',
                           'l': 'left',
                           'r': 'right',
                           'u': 'upDistance01',
                           'up': 'upDistance02'},
            'type': 'rendering'},
  'trackCtx': {'flags': {'alternateContext': {'args': '<type __builtin__.bool>',
                                              'longname': 'alternateContext',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'ac'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'history': {'args': '<type __builtin__.bool>',
                                     'longname': 'history',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'ch'},
                         'image1': {'args': '<type __builtin__.str>',
                                    'longname': 'image1',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i1'},
                         'image2': {'args': '<type __builtin__.str>',
                                    'longname': 'image2',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i2'},
                         'image3': {'args': '<type __builtin__.str>',
                                    'longname': 'image3',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'i3'},
                         'name': {'args': '<type __builtin__.str>',
                                  'longname': 'name',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'n'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'toolName': {'args': '<type __builtin__.str>',
                                      'longname': 'toolName',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'tn'},
                         'trackGeometry': {'args': '<type __builtin__.bool>',
                                           'longname': 'trackGeometry',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'tg'},
                         'trackScale': {'args': '<type __builtin__.float>',
                                        'longname': 'trackScale',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ts'}},
               'shortFlags': {'ac': 'alternateContext',
                              'ch': 'history',
                              'e': 'edit',
                              'ex': 'exists',
                              'i1': 'image1',
                              'i2': 'image2',
                              'i3': 'image3',
                              'n': 'name',
                              'q': 'query',
                              'tg': 'trackGeometry',
                              'tn': 'toolName',
                              'ts': 'trackScale'},
               'type': 'context'},
  'transferAttributes': {'flags': {'after': {'args': '<type __builtin__.bool>',
                                             'longname': 'after',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'af'},
                                   'afterReference': {'args': '<type __builtin__.bool>',
                                                      'longname': 'afterReference',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 0,
                                                      'shortname': 'ar'},
                                   'before': {'args': '<type __builtin__.bool>',
                                              'longname': 'before',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 0,
                                              'shortname': 'bf'},
                                   'colorBorders': {'args': '<type __builtin__.int>',
                                                    'longname': 'colorBorders',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'clb'},
                                   'deformerTools': {'args': '<type __builtin__.bool>',
                                                     'longname': 'deformerTools',
                                                     'modes': ['query'],
                                                     'numArgs': 0,
                                                     'shortname': 'dt'},
                                   'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'exclusive': {'args': '<type __builtin__.str>',
                                                 'longname': 'exclusive',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 1,
                                                 'shortname': 'ex'},
                                   'flipUVs': {'args': '<type __builtin__.int>',
                                               'longname': 'flipUVs',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'fuv'},
                                   'frontOfChain': {'args': '<type __builtin__.bool>',
                                                    'longname': 'frontOfChain',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 0,
                                                    'shortname': 'foc'},
                                   'geometry': {'args': '<type __builtin__.str>',
                                                'longname': 'geometry',
                                                'modes': ['query',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 1,
                                                'shortname': 'g'},
                                   'geometryIndices': {'args': '<type __builtin__.bool>',
                                                       'longname': 'geometryIndices',
                                                       'modes': ['query'],
                                                       'numArgs': 0,
                                                       'shortname': 'gi'},
                                   'ignoreSelected': {'args': '<type __builtin__.bool>',
                                                      'longname': 'ignoreSelected',
                                                      'modes': ['create'],
                                                      'numArgs': 0,
                                                      'shortname': 'ignoreSelected'},
                                   'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                               'longname': 'includeHiddenSelections',
                                                               'modes': ['create'],
                                                               'numArgs': 0,
                                                               'shortname': 'ihs'},
                                   'matchChoice': {'args': '<type __builtin__.int>',
                                                   'longname': 'matchChoice',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'mch'},
                                   'name': {'args': '<type __builtin__.str>',
                                            'longname': 'name',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'n'},
                                   'parallel': {'args': '<type __builtin__.bool>',
                                                'longname': 'parallel',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 0,
                                                'shortname': 'par'},
                                   'prune': {'args': '<type __builtin__.bool>',
                                             'longname': 'prune',
                                             'modes': ['edit'],
                                             'numArgs': 0,
                                             'shortname': 'pr'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'},
                                   'remove': {'args': '<type __builtin__.bool>',
                                              'longname': 'remove',
                                              'modes': ['edit',
                                                        'multiuse'],
                                              'numArgs': 0,
                                              'shortname': 'rm'},
                                   'sampleSpace': {'args': '<type __builtin__.int>',
                                                   'longname': 'sampleSpace',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'spa'},
                                   'searchMethod': {'args': '<type __builtin__.int>',
                                                    'longname': 'searchMethod',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'sm'},
                                   'searchScaleX': {'args': '<type __builtin__.float>',
                                                    'longname': 'searchScaleX',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ssx'},
                                   'searchScaleY': {'args': '<type __builtin__.float>',
                                                    'longname': 'searchScaleY',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ssy'},
                                   'searchScaleZ': {'args': '<type __builtin__.float>',
                                                    'longname': 'searchScaleZ',
                                                    'modes': ['create',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'ssz'},
                                   'sourceColorSet': {'args': '<type __builtin__.str>',
                                                      'longname': 'sourceColorSet',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'scs'},
                                   'sourceUvSet': {'args': '<type __builtin__.str>',
                                                   'longname': 'sourceUvSet',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'suv'},
                                   'sourceUvSpace': {'args': '<type __builtin__.str>',
                                                     'longname': 'sourceUvSpace',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'sus'},
                                   'split': {'args': '<type __builtin__.bool>',
                                             'longname': 'split',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'sp'},
                                   'targetColorSet': {'args': '<type __builtin__.str>',
                                                      'longname': 'targetColorSet',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'tcs'},
                                   'targetUvSet': {'args': '<type __builtin__.str>',
                                                   'longname': 'targetUvSet',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'tuv'},
                                   'targetUvSpace': {'args': '<type __builtin__.str>',
                                                     'longname': 'targetUvSpace',
                                                     'modes': ['create'],
                                                     'numArgs': 1,
                                                     'shortname': 'tus'},
                                   'transferColors': {'args': '<type __builtin__.int>',
                                                      'longname': 'transferColors',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 1,
                                                      'shortname': 'col'},
                                   'transferNormals': {'args': '<type __builtin__.int>',
                                                       'longname': 'transferNormals',
                                                       'modes': ['create',
                                                                 'edit'],
                                                       'numArgs': 1,
                                                       'shortname': 'nml'},
                                   'transferPositions': {'args': '<type __builtin__.int>',
                                                         'longname': 'transferPositions',
                                                         'modes': ['create',
                                                                   'edit'],
                                                         'numArgs': 1,
                                                         'shortname': 'pos'},
                                   'transferUVs': {'args': '<type __builtin__.int>',
                                                   'longname': 'transferUVs',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'uvs'}},
                         'removedFlags': {'is': 'ignoreSelected'},
                         'shortFlags': {'af': 'after',
                                        'ar': 'afterReference',
                                        'bf': 'before',
                                        'clb': 'colorBorders',
                                        'col': 'transferColors',
                                        'dt': 'deformerTools',
                                        'e': 'edit',
                                        'ex': 'exclusive',
                                        'foc': 'frontOfChain',
                                        'fuv': 'flipUVs',
                                        'g': 'geometry',
                                        'gi': 'geometryIndices',
                                        'ignoreSelected': 'ignoreSelected',
                                        'ihs': 'includeHiddenSelections',
                                        'mch': 'matchChoice',
                                        'n': 'name',
                                        'nml': 'transferNormals',
                                        'par': 'parallel',
                                        'pos': 'transferPositions',
                                        'pr': 'prune',
                                        'q': 'query',
                                        'rm': 'remove',
                                        'scs': 'sourceColorSet',
                                        'sm': 'searchMethod',
                                        'sp': 'split',
                                        'spa': 'sampleSpace',
                                        'ssx': 'searchScaleX',
                                        'ssy': 'searchScaleY',
                                        'ssz': 'searchScaleZ',
                                        'sus': 'sourceUvSpace',
                                        'suv': 'sourceUvSet',
                                        'tcs': 'targetColorSet',
                                        'tus': 'targetUvSpace',
                                        'tuv': 'targetUvSet',
                                        'uvs': 'transferUVs'},
                         'type': 'modeling'},
  'transferShadingSets': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                             'longname': 'edit',
                                             'numArgs': 0,
                                             'shortname': 'e'},
                                    'query': {'args': '<type __builtin__.bool>',
                                              'longname': 'query',
                                              'numArgs': 0,
                                              'shortname': 'q'},
                                    'sampleSpace': {'args': '<type __builtin__.int>',
                                                    'longname': 'sampleSpace',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'spa'},
                                    'searchMethod': {'args': '<type __builtin__.int>',
                                                     'longname': 'searchMethod',
                                                     'modes': ['create',
                                                               'query',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'sm'}},
                          'shortFlags': {'e': 'edit',
                                         'q': 'query',
                                         'sm': 'searchMethod',
                                         'spa': 'sampleSpace'},
                          'type': 'modeling'},
  'transformCompare': {'flags': {'root': {'args': '<type __builtin__.bool>',
                                          'longname': 'root',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'r'}},
                       'shortFlags': {'r': 'root'},
                       'type': 'general'},
  'transformLimits': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'enableRotationX': {'args': ['<type __builtin__.bool>',
                                                             '<type __builtin__.bool>'],
                                                    'longname': 'enableRotationX',
                                                    'modes': ['query'],
                                                    'numArgs': 2,
                                                    'shortname': 'erx'},
                                'enableRotationY': {'args': ['<type __builtin__.bool>',
                                                             '<type __builtin__.bool>'],
                                                    'longname': 'enableRotationY',
                                                    'modes': ['query'],
                                                    'numArgs': 2,
                                                    'shortname': 'ery'},
                                'enableRotationZ': {'args': ['<type __builtin__.bool>',
                                                             '<type __builtin__.bool>'],
                                                    'longname': 'enableRotationZ',
                                                    'modes': ['query'],
                                                    'numArgs': 2,
                                                    'shortname': 'erz'},
                                'enableScaleX': {'args': ['<type __builtin__.bool>',
                                                          '<type __builtin__.bool>'],
                                                 'longname': 'enableScaleX',
                                                 'modes': ['query'],
                                                 'numArgs': 2,
                                                 'shortname': 'esx'},
                                'enableScaleY': {'args': ['<type __builtin__.bool>',
                                                          '<type __builtin__.bool>'],
                                                 'longname': 'enableScaleY',
                                                 'modes': ['query'],
                                                 'numArgs': 2,
                                                 'shortname': 'esy'},
                                'enableScaleZ': {'args': ['<type __builtin__.bool>',
                                                          '<type __builtin__.bool>'],
                                                 'longname': 'enableScaleZ',
                                                 'modes': ['query'],
                                                 'numArgs': 2,
                                                 'shortname': 'esz'},
                                'enableTranslationX': {'args': ['<type __builtin__.bool>',
                                                                '<type __builtin__.bool>'],
                                                       'longname': 'enableTranslationX',
                                                       'modes': ['query'],
                                                       'numArgs': 2,
                                                       'shortname': 'etx'},
                                'enableTranslationY': {'args': ['<type __builtin__.bool>',
                                                                '<type __builtin__.bool>'],
                                                       'longname': 'enableTranslationY',
                                                       'modes': ['query'],
                                                       'numArgs': 2,
                                                       'shortname': 'ety'},
                                'enableTranslationZ': {'args': ['<type __builtin__.bool>',
                                                                '<type __builtin__.bool>'],
                                                       'longname': 'enableTranslationZ',
                                                       'modes': ['query'],
                                                       'numArgs': 2,
                                                       'shortname': 'etz'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'remove': {'args': '<type __builtin__.bool>',
                                           'longname': 'remove',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'rm'},
                                'rotationX': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'rotationX',
                                              'modes': ['query'],
                                              'numArgs': 2,
                                              'shortname': 'rx'},
                                'rotationY': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'rotationY',
                                              'modes': ['query'],
                                              'numArgs': 2,
                                              'shortname': 'ry'},
                                'rotationZ': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'rotationZ',
                                              'modes': ['query'],
                                              'numArgs': 2,
                                              'shortname': 'rz'},
                                'scaleX': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'scaleX',
                                           'modes': ['query'],
                                           'numArgs': 2,
                                           'shortname': 'sx'},
                                'scaleY': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'scaleY',
                                           'modes': ['query'],
                                           'numArgs': 2,
                                           'shortname': 'sy'},
                                'scaleZ': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'scaleZ',
                                           'modes': ['query'],
                                           'numArgs': 2,
                                           'shortname': 'sz'},
                                'translationX': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'translationX',
                                                 'modes': ['query'],
                                                 'numArgs': 2,
                                                 'shortname': 'tx'},
                                'translationY': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'translationY',
                                                 'modes': ['query'],
                                                 'numArgs': 2,
                                                 'shortname': 'ty'},
                                'translationZ': {'args': ['<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'translationZ',
                                                 'modes': ['query'],
                                                 'numArgs': 2,
                                                 'shortname': 'tz'}},
                      'shortFlags': {'e': 'edit',
                                     'erx': 'enableRotationX',
                                     'ery': 'enableRotationY',
                                     'erz': 'enableRotationZ',
                                     'esx': 'enableScaleX',
                                     'esy': 'enableScaleY',
                                     'esz': 'enableScaleZ',
                                     'etx': 'enableTranslationX',
                                     'ety': 'enableTranslationY',
                                     'etz': 'enableTranslationZ',
                                     'q': 'query',
                                     'rm': 'remove',
                                     'rx': 'rotationX',
                                     'ry': 'rotationY',
                                     'rz': 'rotationZ',
                                     'sx': 'scaleX',
                                     'sy': 'scaleY',
                                     'sz': 'scaleZ',
                                     'tx': 'translationX',
                                     'ty': 'translationY',
                                     'tz': 'translationZ'},
                      'type': 'general'},
  'translator': {'flags': {'defaultFileRule': {'args': '<type __builtin__.bool>',
                                               'longname': 'defaultFileRule',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'shortname': 'dfr'},
                           'defaultOptions': {'args': '<type __builtin__.str>',
                                              'longname': 'defaultOptions',
                                              'modes': ['query'],
                                              'numArgs': 1,
                                              'shortname': 'do'},
                           'extension': {'args': '<type __builtin__.bool>',
                                         'longname': 'extension',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'ext'},
                           'fileCompression': {'args': '<type __builtin__.str>',
                                               'longname': 'fileCompression',
                                               'modes': ['query'],
                                               'numArgs': 1,
                                               'shortname': 'cmp'},
                           'filter': {'args': '<type __builtin__.bool>',
                                      'longname': 'filter',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'f'},
                           'list': {'args': '<type __builtin__.bool>',
                                    'longname': 'list',
                                    'modes': ['query'],
                                    'numArgs': 0,
                                    'shortname': 'l'},
                           'loaded': {'args': '<type __builtin__.bool>',
                                      'longname': 'loaded',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'shortname': 'ld'},
                           'objectType': {'args': '<type __builtin__.bool>',
                                          'longname': 'objectType',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'ot'},
                           'optionsScript': {'args': '<type __builtin__.bool>',
                                             'longname': 'optionsScript',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'os'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'readSupport': {'args': '<type __builtin__.bool>',
                                           'longname': 'readSupport',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'rs'},
                           'writeSupport': {'args': '<type __builtin__.bool>',
                                            'longname': 'writeSupport',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'ws'}},
                 'shortFlags': {'cmp': 'fileCompression',
                                'dfr': 'defaultFileRule',
                                'do': 'defaultOptions',
                                'ext': 'extension',
                                'f': 'filter',
                                'l': 'list',
                                'ld': 'loaded',
                                'os': 'optionsScript',
                                'ot': 'objectType',
                                'q': 'query',
                                'rs': 'readSupport',
                                'ws': 'writeSupport'},
                 'type': 'system'},
  'treeLister': {'callbackFlags': ['refreshCommand',
                                   'rc',
                                   'visibleChangeCommand',
                                   'vcc',
                                   'dragCallback',
                                   'dgc',
                                   'dropCallback',
                                   'dpc',
                                   'favoritesCallback',
                                   'fcb'],
                 'flags': {'addFavorite': {'args': '<type __builtin__.str>',
                                           'longname': 'addFavorite',
                                           'modes': ['create',
                                                     'edit',
                                                     'multiuse'],
                                           'numArgs': 1,
                                           'shortname': 'af'},
                           'addItem': {'args': ['<type __builtin__.str>',
                                                '<type __builtin__.str>',
                                                'script'],
                                       'longname': 'addItem',
                                       'modes': ['create',
                                                 'edit',
                                                 'multiuse'],
                                       'numArgs': 3,
                                       'shortname': 'add'},
                           'addVnnItem': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.str>',
                                                   '<type __builtin__.str>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'addVnnItem',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 4,
                                          'shortname': 'avi'},
                           'annotation': {'args': '<type __builtin__.str>',
                                          'longname': 'annotation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ann'},
                           'backgroundColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'backgroundColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'bgc'},
                           'clearContents': {'args': '<type __builtin__.bool>',
                                             'longname': 'clearContents',
                                             'modes': ['edit'],
                                             'numArgs': 0,
                                             'shortname': 'clr'},
                           'collapsePath': {'args': '<type __builtin__.str>',
                                            'longname': 'collapsePath',
                                            'modes': ['edit', 'multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'cp'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'dragCallback': {'args': 'script',
                                            'longname': 'dragCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dgc'},
                           'dropCallback': {'args': 'script',
                                            'longname': 'dropCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dpc'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'enable': {'args': '<type __builtin__.bool>',
                                      'longname': 'enable',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'en'},
                           'enableBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableBackground',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ebg'},
                           'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableKeyboardFocus',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ekf'},
                           'executeItem': {'args': '<type __builtin__.str>',
                                           'longname': 'executeItem',
                                           'modes': ['edit'],
                                           'numArgs': 1,
                                           'shortname': 'ei'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'expandPath': {'args': '<type __builtin__.str>',
                                          'longname': 'expandPath',
                                          'modes': ['edit', 'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'ep'},
                           'expandToDepth': {'args': '<type __builtin__.int>',
                                             'longname': 'expandToDepth',
                                             'modes': ['edit'],
                                             'numArgs': 1,
                                             'shortname': 'etd'},
                           'favoritesCallback': {'args': 'script',
                                                 'longname': 'favoritesCallback',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'fcb'},
                           'favoritesList': {'args': '<type __builtin__.bool>',
                                             'longname': 'favoritesList',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'fl'},
                           'fullPathName': {'args': '<type __builtin__.bool>',
                                            'longname': 'fullPathName',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fpn'},
                           'height': {'args': '<type __builtin__.int>',
                                      'longname': 'height',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                           'highlightColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'highlightColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'hlc'},
                           'isObscured': {'args': '<type __builtin__.bool>',
                                          'longname': 'isObscured',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'io'},
                           'itemScript': {'args': '<type __builtin__.str>',
                                          'longname': 'itemScript',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'isc'},
                           'manage': {'args': '<type __builtin__.bool>',
                                      'longname': 'manage',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                           'noBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'noBackground',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nbg'},
                           'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'numberOfPopupMenus',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'npm'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuArray': {'args': '<type __builtin__.bool>',
                                              'longname': 'popupMenuArray',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'pma'},
                           'preventOverride': {'args': '<type __builtin__.bool>',
                                               'longname': 'preventOverride',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'po'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'refreshCommand': {'args': 'script',
                                              'longname': 'refreshCommand',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'rc'},
                           'removeFavorite': {'args': '<type __builtin__.str>',
                                              'longname': 'removeFavorite',
                                              'modes': ['edit',
                                                        'multiuse'],
                                              'numArgs': 1,
                                              'shortname': 'rf'},
                           'removeItem': {'args': '<type __builtin__.str>',
                                          'longname': 'removeItem',
                                          'modes': ['edit', 'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'rem'},
                           'resultsPathUnderCursor': {'args': '<type __builtin__.bool>',
                                                      'longname': 'resultsPathUnderCursor',
                                                      'modes': ['query'],
                                                      'numArgs': 0,
                                                      'shortname': 'ruc'},
                           'selectPath': {'args': '<type __builtin__.str>',
                                          'longname': 'selectPath',
                                          'modes': ['edit', 'multiuse'],
                                          'numArgs': 1,
                                          'shortname': 'sp'},
                           'statusBarMessage': {'args': '<type __builtin__.str>',
                                                'longname': 'statusBarMessage',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sbm'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'},
                           'visible': {'args': '<type __builtin__.bool>',
                                       'longname': 'visible',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vis'},
                           'visibleChangeCommand': {'args': 'script',
                                                    'longname': 'visibleChangeCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'vcc'},
                           'vnnString': {'args': '<type __builtin__.bool>',
                                         'longname': 'vnnString',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'vnn'},
                           'width': {'args': '<type __builtin__.int>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'}},
                 'shortFlags': {'add': 'addItem',
                                'af': 'addFavorite',
                                'ann': 'annotation',
                                'avi': 'addVnnItem',
                                'bgc': 'backgroundColor',
                                'clr': 'clearContents',
                                'cp': 'collapsePath',
                                'dgc': 'dragCallback',
                                'dpc': 'dropCallback',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'ebg': 'enableBackground',
                                'ei': 'executeItem',
                                'ekf': 'enableKeyboardFocus',
                                'en': 'enable',
                                'ep': 'expandPath',
                                'etd': 'expandToDepth',
                                'ex': 'exists',
                                'fcb': 'favoritesCallback',
                                'fl': 'favoritesList',
                                'fpn': 'fullPathName',
                                'h': 'height',
                                'hlc': 'highlightColor',
                                'io': 'isObscured',
                                'isc': 'itemScript',
                                'm': 'manage',
                                'nbg': 'noBackground',
                                'npm': 'numberOfPopupMenus',
                                'p': 'parent',
                                'pma': 'popupMenuArray',
                                'po': 'preventOverride',
                                'q': 'query',
                                'rc': 'refreshCommand',
                                'rem': 'removeItem',
                                'rf': 'removeFavorite',
                                'ruc': 'resultsPathUnderCursor',
                                'sbm': 'statusBarMessage',
                                'sp': 'selectPath',
                                'ut': 'useTemplate',
                                'vcc': 'visibleChangeCommand',
                                'vis': 'visible',
                                'vnn': 'vnnString',
                                'w': 'width'},
                 'type': 'windows'},
  'treeView': {'callbackFlags': ['itemDblClickCommand2',
                                 'dc2',
                                 'selectCommand',
                                 'sc',
                                 'dragCallback',
                                 'dgc',
                                 'editLabelCommand',
                                 'elc',
                                 'dragAndDropCommand',
                                 'dad',
                                 'itemRenamedCommand',
                                 'irc',
                                 'dropCallback',
                                 'dpc',
                                 'expandCollapseCommand',
                                 'ecc',
                                 'contextMenuCommand',
                                 'cmc',
                                 'rightPressCommand',
                                 'rpc',
                                 'itemDblClickCommand',
                                 'idc',
                                 'visibleChangeCommand',
                                 'vcc',
                                 'selectionChangedCommand',
                                 'scc',
                                 'pressCommand',
                                 'pc'],
               'flags': {'addItem': {'args': ['<type __builtin__.str>',
                                              '<type __builtin__.str>'],
                                     'longname': 'addItem',
                                     'modes': ['create',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 2,
                                     'shortname': 'ai'},
                         'allowDragAndDrop': {'args': '<type __builtin__.bool>',
                                              'longname': 'allowDragAndDrop',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'adr'},
                         'allowHiddenParents': {'args': '<type __builtin__.bool>',
                                                'longname': 'allowHiddenParents',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ahp'},
                         'allowMultiSelection': {'args': '<type __builtin__.bool>',
                                                 'longname': 'allowMultiSelection',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ams'},
                         'allowReparenting': {'args': '<type __builtin__.bool>',
                                              'longname': 'allowReparenting',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'arp'},
                         'annotation': {'args': '<type __builtin__.str>',
                                        'longname': 'annotation',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ann'},
                         'attachButtonRight': {'args': '<type __builtin__.int>',
                                               'longname': 'attachButtonRight',
                                               'modes': ['create', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'abr'},
                         'backgroundColor': {'args': ['<type __builtin__.float>',
                                                      '<type __builtin__.float>',
                                                      '<type __builtin__.float>'],
                                             'longname': 'backgroundColor',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 3,
                                             'shortname': 'bgc'},
                         'borderHighlite': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.bool>'],
                                            'longname': 'borderHighlite',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 2,
                                            'shortname': 'bh'},
                         'borderHighliteColor': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>',
                                                          '<type __builtin__.float>'],
                                                 'longname': 'borderHighliteColor',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 4,
                                                 'shortname': 'bcl'},
                         'buttonErase': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.bool>'],
                                         'longname': 'buttonErase',
                                         'modes': ['create',
                                                   'query',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 2,
                                         'shortname': 'bef'},
                         'buttonState': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.int>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'buttonState',
                                         'modes': ['create',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 3,
                                         'shortname': 'bst'},
                         'buttonStyle': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.int>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'buttonStyle',
                                         'modes': ['create',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 3,
                                         'shortname': 'bs'},
                         'buttonTextIcon': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.int>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'buttonTextIcon',
                                            'modes': ['create',
                                                      'edit',
                                                      'multiuse'],
                                            'numArgs': 3,
                                            'shortname': 'bti'},
                         'buttonTooltip': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.int>',
                                                    '<type __builtin__.str>'],
                                           'longname': 'buttonTooltip',
                                           'modes': ['create',
                                                     'edit',
                                                     'multiuse'],
                                           'numArgs': 3,
                                           'shortname': 'btp'},
                         'buttonTransparencyColor': {'args': ['<type __builtin__.str>',
                                                              '<type __builtin__.int>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'buttonTransparencyColor',
                                                     'modes': ['create',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 5,
                                                     'shortname': 'btc'},
                         'buttonTransparencyOverride': {'args': ['<type __builtin__.str>',
                                                                 '<type __builtin__.int>',
                                                                 '<type __builtin__.bool>'],
                                                        'longname': 'buttonTransparencyOverride',
                                                        'modes': ['create',
                                                                  'edit',
                                                                  'multiuse'],
                                                        'numArgs': 3,
                                                        'shortname': 'bto'},
                         'buttonVisible': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.int>',
                                                    '<type __builtin__.bool>'],
                                           'longname': 'buttonVisible',
                                           'modes': ['create', 'multiuse'],
                                           'numArgs': 3,
                                           'shortname': 'bvf'},
                         'children': {'args': '<type __builtin__.str>',
                                      'longname': 'children',
                                      'modes': ['query'],
                                      'numArgs': 1,
                                      'shortname': 'ch'},
                         'clearSelection': {'args': '<type __builtin__.bool>',
                                            'longname': 'clearSelection',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 0,
                                            'shortname': 'cs'},
                         'contextMenuCommand': {'args': 'script',
                                                'longname': 'contextMenuCommand',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'cmc'},
                         'defineTemplate': {'args': '<type __builtin__.str>',
                                            'longname': 'defineTemplate',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'dt'},
                         'displayLabel': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'displayLabel',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'dl'},
                         'displayLabelSuffix': {'args': ['<type __builtin__.str>',
                                                         '<type __builtin__.str>'],
                                                'longname': 'displayLabelSuffix',
                                                'modes': ['create',
                                                          'edit',
                                                          'multiuse'],
                                                'numArgs': 2,
                                                'shortname': 'dls'},
                         'docTag': {'args': '<type __builtin__.str>',
                                    'longname': 'docTag',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'dtg'},
                         'dragAndDropCommand': {'args': 'script',
                                                'longname': 'dragAndDropCommand',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dad'},
                         'dragCallback': {'args': 'script',
                                          'longname': 'dragCallback',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dgc'},
                         'dropCallback': {'args': 'script',
                                          'longname': 'dropCallback',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dpc'},
                         'edit': {'args': '<type __builtin__.bool>',
                                  'longname': 'edit',
                                  'numArgs': 0,
                                  'shortname': 'e'},
                         'editLabelCommand': {'args': 'script',
                                              'longname': 'editLabelCommand',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'elc'},
                         'enable': {'args': '<type __builtin__.bool>',
                                    'longname': 'enable',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'en'},
                         'enableBackground': {'args': '<type __builtin__.bool>',
                                              'longname': 'enableBackground',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ebg'},
                         'enableButton': {'args': ['<type __builtin__.str>',
                                                   '<type __builtin__.int>',
                                                   '<type __builtin__.int>'],
                                          'longname': 'enableButton',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 3,
                                          'shortname': 'eb'},
                         'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                 'longname': 'enableKeyboardFocus',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ekf'},
                         'enableKeys': {'args': '<type __builtin__.bool>',
                                        'longname': 'enableKeys',
                                        'modes': ['edit'],
                                        'numArgs': 1,
                                        'shortname': 'enk'},
                         'enableLabel': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.int>'],
                                         'longname': 'enableLabel',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 2,
                                         'shortname': 'enl'},
                         'exists': {'args': '<type __builtin__.bool>',
                                    'longname': 'exists',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'ex'},
                         'expandCollapseCommand': {'args': 'script',
                                                   'longname': 'expandCollapseCommand',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ecc'},
                         'expandItem': {'args': ['<type __builtin__.str>',
                                                 '<type __builtin__.bool>'],
                                        'longname': 'expandItem',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 2,
                                        'shortname': 'ei'},
                         'flatButton': {'args': '<type __builtin__.int>',
                                        'longname': 'flatButton',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'fb'},
                         'font': {'args': ['<type __builtin__.str>',
                                           '<type __builtin__.str>'],
                                  'longname': 'font',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 2,
                                  'shortname': 'fn'},
                         'fontFace': {'args': ['<type __builtin__.str>',
                                               '<type __builtin__.int>'],
                                      'longname': 'fontFace',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 2,
                                      'shortname': 'ff'},
                         'fullPathName': {'args': '<type __builtin__.bool>',
                                          'longname': 'fullPathName',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'fpn'},
                         'height': {'args': '<type __builtin__.int>',
                                    'longname': 'height',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'h'},
                         'hideButtons': {'args': '<type __builtin__.bool>',
                                         'longname': 'hideButtons',
                                         'modes': ['create', 'edit'],
                                         'numArgs': 0,
                                         'shortname': 'hb'},
                         'highlightColor': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'highlightColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 3,
                                            'shortname': 'hlc'},
                         'highlite': {'args': ['<type __builtin__.str>',
                                               '<type __builtin__.bool>'],
                                      'longname': 'highlite',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 2,
                                      'shortname': 'hl'},
                         'highliteColor': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'highliteColor',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 4,
                                           'shortname': 'hc'},
                         'ignoreButtonClick': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.int>',
                                                        '<type __builtin__.int>'],
                                               'longname': 'ignoreButtonClick',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 3,
                                               'shortname': 'ibc'},
                         'image': {'args': ['<type __builtin__.str>',
                                            '<type __builtin__.int>',
                                            '<type __builtin__.str>'],
                                   'longname': 'image',
                                   'modes': ['create', 'edit', 'multiuse'],
                                   'numArgs': 3,
                                   'shortname': 'i'},
                         'insertItem': {'args': ['<type __builtin__.str>',
                                                 '<type __builtin__.str>',
                                                 '<type __builtin__.int>'],
                                        'longname': 'insertItem',
                                        'modes': ['create',
                                                  'edit',
                                                  'multiuse'],
                                        'numArgs': 3,
                                        'shortname': 'ii'},
                         'isItemExpanded': {'args': '<type __builtin__.str>',
                                            'longname': 'isItemExpanded',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'iie'},
                         'isLeaf': {'args': '<type __builtin__.str>',
                                    'longname': 'isLeaf',
                                    'modes': ['query'],
                                    'numArgs': 1,
                                    'shortname': 'il'},
                         'isObscured': {'args': '<type __builtin__.bool>',
                                        'longname': 'isObscured',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'io'},
                         'item': {'args': '<type __builtin__.str>',
                                  'longname': 'item',
                                  'modes': ['query'],
                                  'numArgs': 1,
                                  'shortname': 'it'},
                         'itemAnnotation': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.str>'],
                                            'longname': 'itemAnnotation',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 2,
                                            'shortname': 'ia'},
                         'itemDblClickCommand': {'args': 'script',
                                                 'longname': 'itemDblClickCommand',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'idc'},
                         'itemDblClickCommand2': {'args': 'script',
                                                  'longname': 'itemDblClickCommand2',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dc2'},
                         'itemExists': {'args': '<type __builtin__.str>',
                                        'longname': 'itemExists',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'iex'},
                         'itemIndex': {'args': '<type __builtin__.str>',
                                       'longname': 'itemIndex',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'idx'},
                         'itemParent': {'args': '<type __builtin__.str>',
                                        'longname': 'itemParent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'ip'},
                         'itemRenamedCommand': {'args': 'script',
                                                'longname': 'itemRenamedCommand',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'irc'},
                         'itemSelected': {'args': '<type __builtin__.str>',
                                          'longname': 'itemSelected',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'isl'},
                         'itemVisible': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.bool>'],
                                         'longname': 'itemVisible',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 2,
                                         'shortname': 'iv'},
                         'labelBackgroundColor': {'args': ['<type __builtin__.str>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>',
                                                           '<type __builtin__.float>'],
                                                  'longname': 'labelBackgroundColor',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 4,
                                                  'shortname': 'lbc'},
                         'manage': {'args': '<type __builtin__.bool>',
                                    'longname': 'manage',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'm'},
                         'noBackground': {'args': '<type __builtin__.bool>',
                                          'longname': 'noBackground',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nbg'},
                         'numberOfButtons': {'args': '<type __builtin__.int>',
                                             'longname': 'numberOfButtons',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 1,
                                             'shortname': 'nb'},
                         'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                'longname': 'numberOfPopupMenus',
                                                'modes': ['query'],
                                                'numArgs': 0,
                                                'shortname': 'npm'},
                         'ornament': {'args': ['<type __builtin__.str>',
                                               '<type __builtin__.int>',
                                               '<type __builtin__.int>',
                                               '<type __builtin__.int>'],
                                      'longname': 'ornament',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 4,
                                      'shortname': 'ornament'},
                         'ornamentColor': {'args': ['<type __builtin__.str>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'ornamentColor',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 4,
                                           'shortname': 'orc'},
                         'parent': {'args': '<type __builtin__.str>',
                                    'longname': 'parent',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'p'},
                         'popupMenuArray': {'args': '<type __builtin__.bool>',
                                            'longname': 'popupMenuArray',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'pma'},
                         'pressCommand': {'args': ['<type __builtin__.int>',
                                                   'script'],
                                          'longname': 'pressCommand',
                                          'modes': ['create',
                                                    'edit',
                                                    'multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'pc'},
                         'preventOverride': {'args': '<type __builtin__.bool>',
                                             'longname': 'preventOverride',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'po'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'removeAll': {'args': '<type __builtin__.bool>',
                                       'longname': 'removeAll',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'ra'},
                         'removeItem': {'args': '<type __builtin__.str>',
                                        'longname': 'removeItem',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ri'},
                         'reverseTreeOrder': {'args': '<type __builtin__.bool>',
                                              'longname': 'reverseTreeOrder',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'rto'},
                         'rightPressCommand': {'args': ['<type __builtin__.int>',
                                                        'script'],
                                               'longname': 'rightPressCommand',
                                               'modes': ['create',
                                                         'edit',
                                                         'multiuse'],
                                               'numArgs': 2,
                                               'shortname': 'rpc'},
                         'select': {'args': ['<type __builtin__.str>',
                                             '<type __builtin__.int>'],
                                    'longname': 'select',
                                    'modes': ['create', 'edit'],
                                    'numArgs': 2,
                                    'shortname': 'sl'},
                         'selectCommand': {'args': 'script',
                                           'longname': 'selectCommand',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'sc'},
                         'selectItem': {'args': ['<type __builtin__.str>',
                                                 '<type __builtin__.bool>'],
                                        'longname': 'selectItem',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 2,
                                        'shortname': 'si'},
                         'selectionChangedCommand': {'args': 'script',
                                                     'longname': 'selectionChangedCommand',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'scc'},
                         'selectionColor': {'args': ['<type __builtin__.str>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'selectionColor',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 4,
                                            'shortname': 'scl'},
                         'showItem': {'args': '<type __builtin__.str>',
                                      'longname': 'showItem',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'shi'},
                         'statusBarMessage': {'args': '<type __builtin__.str>',
                                              'longname': 'statusBarMessage',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'sbm'},
                         'textColor': {'args': ['<type __builtin__.str>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'textColor',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 4,
                                       'shortname': 'tc'},
                         'useTemplate': {'args': '<type __builtin__.str>',
                                         'longname': 'useTemplate',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'ut'},
                         'visible': {'args': '<type __builtin__.bool>',
                                     'longname': 'visible',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'vis'},
                         'visibleChangeCommand': {'args': 'script',
                                                  'longname': 'visibleChangeCommand',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'vcc'},
                         'width': {'args': '<type __builtin__.int>',
                                   'longname': 'width',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'w'}},
               'removedFlags': {'or': 'ornament'},
               'shortFlags': {'abr': 'attachButtonRight',
                              'adr': 'allowDragAndDrop',
                              'ahp': 'allowHiddenParents',
                              'ai': 'addItem',
                              'ams': 'allowMultiSelection',
                              'ann': 'annotation',
                              'arp': 'allowReparenting',
                              'bcl': 'borderHighliteColor',
                              'bef': 'buttonErase',
                              'bgc': 'backgroundColor',
                              'bh': 'borderHighlite',
                              'bs': 'buttonStyle',
                              'bst': 'buttonState',
                              'btc': 'buttonTransparencyColor',
                              'bti': 'buttonTextIcon',
                              'bto': 'buttonTransparencyOverride',
                              'btp': 'buttonTooltip',
                              'bvf': 'buttonVisible',
                              'ch': 'children',
                              'cmc': 'contextMenuCommand',
                              'cs': 'clearSelection',
                              'dad': 'dragAndDropCommand',
                              'dc2': 'itemDblClickCommand2',
                              'dgc': 'dragCallback',
                              'dl': 'displayLabel',
                              'dls': 'displayLabelSuffix',
                              'dpc': 'dropCallback',
                              'dt': 'defineTemplate',
                              'dtg': 'docTag',
                              'e': 'edit',
                              'eb': 'enableButton',
                              'ebg': 'enableBackground',
                              'ecc': 'expandCollapseCommand',
                              'ei': 'expandItem',
                              'ekf': 'enableKeyboardFocus',
                              'elc': 'editLabelCommand',
                              'en': 'enable',
                              'enk': 'enableKeys',
                              'enl': 'enableLabel',
                              'ex': 'exists',
                              'fb': 'flatButton',
                              'ff': 'fontFace',
                              'fn': 'font',
                              'fpn': 'fullPathName',
                              'h': 'height',
                              'hb': 'hideButtons',
                              'hc': 'highliteColor',
                              'hl': 'highlite',
                              'hlc': 'highlightColor',
                              'i': 'image',
                              'ia': 'itemAnnotation',
                              'ibc': 'ignoreButtonClick',
                              'idc': 'itemDblClickCommand',
                              'idx': 'itemIndex',
                              'iex': 'itemExists',
                              'ii': 'insertItem',
                              'iie': 'isItemExpanded',
                              'il': 'isLeaf',
                              'io': 'isObscured',
                              'ip': 'itemParent',
                              'irc': 'itemRenamedCommand',
                              'isl': 'itemSelected',
                              'it': 'item',
                              'iv': 'itemVisible',
                              'lbc': 'labelBackgroundColor',
                              'm': 'manage',
                              'nb': 'numberOfButtons',
                              'nbg': 'noBackground',
                              'npm': 'numberOfPopupMenus',
                              'orc': 'ornamentColor',
                              'ornament': 'ornament',
                              'p': 'parent',
                              'pc': 'pressCommand',
                              'pma': 'popupMenuArray',
                              'po': 'preventOverride',
                              'q': 'query',
                              'ra': 'removeAll',
                              'ri': 'removeItem',
                              'rpc': 'rightPressCommand',
                              'rto': 'reverseTreeOrder',
                              'sbm': 'statusBarMessage',
                              'sc': 'selectCommand',
                              'scc': 'selectionChangedCommand',
                              'scl': 'selectionColor',
                              'shi': 'showItem',
                              'si': 'selectItem',
                              'sl': 'select',
                              'tc': 'textColor',
                              'ut': 'useTemplate',
                              'vcc': 'visibleChangeCommand',
                              'vis': 'visible',
                              'w': 'width'},
               'type': 'windows'},
  'trim': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                 'longname': 'caching',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'cch'},
                     'constructionHistory': {'args': '<type __builtin__.bool>',
                                             'longname': 'constructionHistory',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ch'},
                     'edit': {'args': '<type __builtin__.bool>',
                              'longname': 'edit',
                              'numArgs': 0,
                              'shortname': 'e'},
                     'frozen': {'args': '<type __builtin__.bool>',
                                'longname': 'frozen',
                                'numArgs': 1,
                                'shortname': 'fzn'},
                     'locatorU': {'args': '<type __builtin__.float>',
                                  'longname': 'locatorU',
                                  'modes': ['create',
                                            'query',
                                            'edit',
                                            'multiuse'],
                                  'numArgs': 1,
                                  'shortname': 'lu'},
                     'locatorV': {'args': '<type __builtin__.float>',
                                  'longname': 'locatorV',
                                  'modes': ['create',
                                            'query',
                                            'edit',
                                            'multiuse'],
                                  'numArgs': 1,
                                  'shortname': 'lv'},
                     'name': {'args': '<type __builtin__.str>',
                              'longname': 'name',
                              'modes': ['create'],
                              'numArgs': 1,
                              'shortname': 'n'},
                     'nodeState': {'args': '<type __builtin__.int>',
                                   'longname': 'nodeState',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'nds'},
                     'object': {'args': '<type __builtin__.bool>',
                                'longname': 'object',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'o'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'replaceOriginal': {'args': '<type __builtin__.bool>',
                                         'longname': 'replaceOriginal',
                                         'numArgs': 1,
                                         'shortname': 'rpo'},
                     'selected': {'args': '<type __builtin__.int>',
                                  'longname': 'selected',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'sl'},
                     'shrink': {'args': '<type __builtin__.bool>',
                                'longname': 'shrink',
                                'modes': ['create', 'query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'sh'},
                     'tolerance': {'args': '<type __builtin__.float>',
                                   'longname': 'tolerance',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'tol'}},
           'shortFlags': {'cch': 'caching',
                          'ch': 'constructionHistory',
                          'e': 'edit',
                          'fzn': 'frozen',
                          'lu': 'locatorU',
                          'lv': 'locatorV',
                          'n': 'name',
                          'nds': 'nodeState',
                          'o': 'object',
                          'q': 'query',
                          'rpo': 'replaceOriginal',
                          'sh': 'shrink',
                          'sl': 'selected',
                          'tol': 'tolerance'},
           'type': 'modeling'},
  'trimCtx': {'flags': {'activeNodes': {'args': '<type __builtin__.bool>',
                                        'longname': 'activeNodes',
                                        'numArgs': 0,
                                        'shortname': 'anq'},
                        'autoCreate': {'args': '<type __builtin__.bool>',
                                       'longname': 'autoCreate',
                                       'numArgs': 1,
                                       'shortname': 'ac'},
                        'caching': {'args': '<type __builtin__.bool>',
                                    'longname': 'caching',
                                    'numArgs': 1,
                                    'shortname': 'cch'},
                        'constructionHistory': {'args': '<type __builtin__.bool>',
                                                'longname': 'constructionHistory',
                                                'numArgs': 1,
                                                'shortname': 'ch'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'exists': {'args': '<type __builtin__.bool>',
                                   'longname': 'exists',
                                   'numArgs': 0,
                                   'shortname': 'ex'},
                        'frozen': {'args': '<type __builtin__.bool>',
                                   'longname': 'frozen',
                                   'numArgs': 1,
                                   'shortname': 'fzn'},
                        'image1': {'args': '<type __builtin__.str>',
                                   'longname': 'image1',
                                   'numArgs': 1,
                                   'shortname': 'i1'},
                        'image2': {'args': '<type __builtin__.str>',
                                   'longname': 'image2',
                                   'numArgs': 1,
                                   'shortname': 'i2'},
                        'image3': {'args': '<type __builtin__.str>',
                                   'longname': 'image3',
                                   'numArgs': 1,
                                   'shortname': 'i3'},
                        'immediate': {'args': '<type __builtin__.bool>',
                                      'longname': 'immediate',
                                      'numArgs': 0,
                                      'shortname': 'im'},
                        'locatorU': {'args': '<type __builtin__.float>',
                                     'longname': 'locatorU',
                                     'modes': ['multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'lu'},
                        'locatorV': {'args': '<type __builtin__.float>',
                                     'longname': 'locatorV',
                                     'modes': ['multiuse'],
                                     'numArgs': 1,
                                     'shortname': 'lv'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'nodeState': {'args': '<type __builtin__.int>',
                                      'longname': 'nodeState',
                                      'numArgs': 1,
                                      'shortname': 'nds'},
                        'object': {'args': '<type __builtin__.bool>',
                                   'longname': 'object',
                                   'numArgs': 1,
                                   'shortname': 'o'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'replaceOriginal': {'args': '<type __builtin__.bool>',
                                            'longname': 'replaceOriginal',
                                            'numArgs': 1,
                                            'shortname': 'rpo'},
                        'reset': {'args': '<type __builtin__.bool>',
                                  'longname': 'reset',
                                  'numArgs': 0,
                                  'shortname': 'rs'},
                        'selected': {'args': '<type __builtin__.int>',
                                     'longname': 'selected',
                                     'numArgs': 1,
                                     'shortname': 'sl'},
                        'shrink': {'args': '<type __builtin__.bool>',
                                   'longname': 'shrink',
                                   'numArgs': 1,
                                   'shortname': 'sh'},
                        'tolerance': {'args': '<type __builtin__.float>',
                                      'longname': 'tolerance',
                                      'numArgs': 1,
                                      'shortname': 'tol'},
                        'toolNode': {'args': '<type __builtin__.bool>',
                                     'longname': 'toolNode',
                                     'numArgs': 0,
                                     'shortname': 'tnq'}},
              'shortFlags': {'ac': 'autoCreate',
                             'anq': 'activeNodes',
                             'cch': 'caching',
                             'ch': 'constructionHistory',
                             'e': 'edit',
                             'ex': 'exists',
                             'fzn': 'frozen',
                             'i1': 'image1',
                             'i2': 'image2',
                             'i3': 'image3',
                             'im': 'immediate',
                             'lu': 'locatorU',
                             'lv': 'locatorV',
                             'n': 'name',
                             'nds': 'nodeState',
                             'o': 'object',
                             'q': 'query',
                             'rpo': 'replaceOriginal',
                             'rs': 'reset',
                             'sh': 'shrink',
                             'sl': 'selected',
                             'tnq': 'toolNode',
                             'tol': 'tolerance'},
              'type': 'context'},
  'truncateFluidCache': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                            'longname': 'edit',
                                            'numArgs': 0,
                                            'shortname': 'e'},
                                   'query': {'args': '<type __builtin__.bool>',
                                             'longname': 'query',
                                             'numArgs': 0,
                                             'shortname': 'q'}},
                         'shortFlags': {'e': 'edit', 'q': 'query'},
                         'type': 'effects'},
  'truncateHairCache': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                           'longname': 'edit',
                                           'numArgs': 0,
                                           'shortname': 'e'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'}},
                        'shortFlags': {'e': 'edit', 'q': 'query'},
                        'type': 'effects'},
  'tumble': {'flags': {'azimuthAngle': {'args': '<type __builtin__.float>',
                                        'longname': 'azimuthAngle',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'aa'},
                       'elevationAngle': {'args': '<type __builtin__.float>',
                                          'longname': 'elevationAngle',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ea'},
                       'localTumble': {'args': '<type __builtin__.int>',
                                       'longname': 'localTumble',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'lt'},
                       'pivotPoint': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'pivotPoint',
                                      'modes': ['create'],
                                      'numArgs': 3,
                                      'shortname': 'pp'},
                       'rotationAngles': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'rotationAngles',
                                          'modes': ['create'],
                                          'numArgs': 2,
                                          'shortname': 'ra'}},
             'shortFlags': {'aa': 'azimuthAngle',
                            'ea': 'elevationAngle',
                            'lt': 'localTumble',
                            'pp': 'pivotPoint',
                            'ra': 'rotationAngles'},
             'type': 'rendering'},
  'tumbleCtx': {'flags': {'alternateContext': {'args': '<type __builtin__.bool>',
                                               'longname': 'alternateContext',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'ac'},
                          'autoOrthoConstrain': {'args': '<type __builtin__.bool>',
                                                 'longname': 'autoOrthoConstrain',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'aoc'},
                          'autoSetPivot': {'args': '<type __builtin__.bool>',
                                           'longname': 'autoSetPivot',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'asp'},
                          'edit': {'args': '<type __builtin__.bool>',
                                   'longname': 'edit',
                                   'numArgs': 0,
                                   'shortname': 'e'},
                          'exists': {'args': '<type __builtin__.bool>',
                                     'longname': 'exists',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ex'},
                          'history': {'args': '<type __builtin__.bool>',
                                      'longname': 'history',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'ch'},
                          'image1': {'args': '<type __builtin__.str>',
                                     'longname': 'image1',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'i1'},
                          'image2': {'args': '<type __builtin__.str>',
                                     'longname': 'image2',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'i2'},
                          'image3': {'args': '<type __builtin__.str>',
                                     'longname': 'image3',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'i3'},
                          'localTumble': {'args': '<type __builtin__.int>',
                                          'longname': 'localTumble',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'lt'},
                          'name': {'args': '<type __builtin__.str>',
                                   'longname': 'name',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'n'},
                          'objectTumble': {'args': '<type __builtin__.bool>',
                                           'longname': 'objectTumble',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'ot'},
                          'orthoLock': {'args': '<type __builtin__.bool>',
                                        'longname': 'orthoLock',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ol'},
                          'orthoStep': {'args': '<type __builtin__.float>',
                                        'longname': 'orthoStep',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'os'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'toolName': {'args': '<type __builtin__.str>',
                                       'longname': 'toolName',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'tn'},
                          'tumbleScale': {'args': '<type __builtin__.float>',
                                          'longname': 'tumbleScale',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ts'}},
                'shortFlags': {'ac': 'alternateContext',
                               'aoc': 'autoOrthoConstrain',
                               'asp': 'autoSetPivot',
                               'ch': 'history',
                               'e': 'edit',
                               'ex': 'exists',
                               'i1': 'image1',
                               'i2': 'image2',
                               'i3': 'image3',
                               'lt': 'localTumble',
                               'n': 'name',
                               'ol': 'orthoLock',
                               'os': 'orthoStep',
                               'ot': 'objectTumble',
                               'q': 'query',
                               'tn': 'toolName',
                               'ts': 'tumbleScale'},
                'type': 'context'},
  'turbulence': {'flags': {'attenuation': {'args': '<type __builtin__.float>',
                                           'longname': 'attenuation',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'att'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'frequency': {'args': '<type __builtin__.float>',
                                         'longname': 'frequency',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'f'},
                           'magnitude': {'args': '<type __builtin__.float>',
                                         'longname': 'magnitude',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                           'maxDistance': {'args': '<type __builtin__.float>',
                                           'longname': 'maxDistance',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'mxd'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'noiseLevel': {'args': '<type __builtin__.int>',
                                          'longname': 'noiseLevel',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'resultNeedsCasting': True,
                                          'shortname': 'nsl'},
                           'noiseRatio': {'args': '<type __builtin__.float>',
                                          'longname': 'noiseRatio',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'nsr'},
                           'perVertex': {'args': '<type __builtin__.bool>',
                                         'longname': 'perVertex',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'pv'},
                           'phase': {'args': '<type __builtin__.float>',
                                     'longname': 'phase',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'p'},
                           'phaseX': {'args': '<type __builtin__.float>',
                                      'longname': 'phaseX',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'px'},
                           'phaseY': {'args': '<type __builtin__.float>',
                                      'longname': 'phaseY',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'py'},
                           'phaseZ': {'args': '<type __builtin__.float>',
                                      'longname': 'phaseZ',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'pz'},
                           'position': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'position',
                                        'modes': ['query',
                                                  'edit',
                                                  'multiuse'],
                                        'numArgs': 3,
                                        'shortname': 'pos'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'torusSectionRadius': {'args': '<type __builtin__.float>',
                                                  'longname': 'torusSectionRadius',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'tsr'},
                           'volumeExclusion': {'args': '<type __builtin__.bool>',
                                               'longname': 'volumeExclusion',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'vex'},
                           'volumeOffset': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'volumeOffset',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 3,
                                            'shortname': 'vof'},
                           'volumeShape': {'args': '<type __builtin__.str>',
                                           'longname': 'volumeShape',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vsh'},
                           'volumeSweep': {'args': '<type __builtin__.float>',
                                           'longname': 'volumeSweep',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vsw'}},
                 'resultNeedsUnpacking': True,
                 'shortFlags': {'att': 'attenuation',
                                'e': 'edit',
                                'f': 'frequency',
                                'm': 'magnitude',
                                'mxd': 'maxDistance',
                                'n': 'name',
                                'nsl': 'noiseLevel',
                                'nsr': 'noiseRatio',
                                'p': 'phase',
                                'pos': 'position',
                                'pv': 'perVertex',
                                'px': 'phaseX',
                                'py': 'phaseY',
                                'pz': 'phaseZ',
                                'q': 'query',
                                'tsr': 'torusSectionRadius',
                                'vex': 'volumeExclusion',
                                'vof': 'volumeOffset',
                                'vsh': 'volumeShape',
                                'vsw': 'volumeSweep'},
                 'type': 'effects'},
  'twoPointArcCtx': {'flags': {'degree': {'args': '<type __builtin__.int>',
                                          'longname': 'degree',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'd'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'history': {'args': '<type __builtin__.bool>',
                                           'longname': 'history',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ch'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'spans': {'args': '<type __builtin__.int>',
                                         'longname': 'spans',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 's'}},
                     'shortFlags': {'ch': 'history',
                                    'd': 'degree',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'n': 'name',
                                    'q': 'query',
                                    's': 'spans'},
                     'type': 'context'},
  'ubercam': {'flags': {}, 'shortFlags': {}, 'type': 'animation'},
  'uiTemplate': {'flags': {'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'}},
                 'shortFlags': {'dt': 'defineTemplate',
                                'e': 'edit',
                                'ex': 'exists',
                                'q': 'query',
                                'ut': 'useTemplate'},
                 'type': 'language'},
  'unassignInputDevice': {'flags': {'clutch': {'args': '<type __builtin__.str>',
                                               'longname': 'clutch',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'c'},
                                    'device': {'args': '<type __builtin__.str>',
                                               'longname': 'device',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'd'}},
                          'shortFlags': {},
                          'type': 'system'},
  'undo': {'flags': {}, 'shortFlags': {}, 'type': 'system'},
  'undoInfo': {'flags': {'chunkName': {'args': '<type __builtin__.str>',
                                       'longname': 'chunkName',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'cn'},
                         'closeChunk': {'args': '<type __builtin__.bool>',
                                        'longname': 'closeChunk',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'cck'},
                         'infinity': {'args': '<type __builtin__.bool>',
                                      'longname': 'infinity',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'infinity'},
                         'length': {'args': '<type __builtin__.int>',
                                    'longname': 'length',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'l'},
                         'openChunk': {'args': '<type __builtin__.bool>',
                                       'longname': 'openChunk',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ock'},
                         'printQueue': {'args': '<type __builtin__.bool>',
                                        'longname': 'printQueue',
                                        'modes': ['query'],
                                        'numArgs': 0,
                                        'shortname': 'pq'},
                         'printRedoQueue': {'args': '<type __builtin__.bool>',
                                            'longname': 'printRedoQueue',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'prq'},
                         'query': {'args': '<type __builtin__.bool>',
                                   'longname': 'query',
                                   'numArgs': 0,
                                   'shortname': 'q'},
                         'redoName': {'args': '<type __builtin__.str>',
                                      'longname': 'redoName',
                                      'modes': ['query'],
                                      'numArgs': 1,
                                      'shortname': 'rn'},
                         'redoQueueEmpty': {'args': '<type __builtin__.bool>',
                                            'longname': 'redoQueueEmpty',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'rqe'},
                         'state': {'args': '<type __builtin__.bool>',
                                   'longname': 'state',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'st'},
                         'stateWithoutFlush': {'args': '<type __builtin__.bool>',
                                               'longname': 'stateWithoutFlush',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'swf'},
                         'undoName': {'args': '<type __builtin__.str>',
                                      'longname': 'undoName',
                                      'modes': ['query'],
                                      'numArgs': 1,
                                      'shortname': 'un'},
                         'undoQueueEmpty': {'args': '<type __builtin__.bool>',
                                            'longname': 'undoQueueEmpty',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'uqe'}},
               'removedFlags': {'in': 'infinity'},
               'shortFlags': {'cck': 'closeChunk',
                              'cn': 'chunkName',
                              'infinity': 'infinity',
                              'l': 'length',
                              'ock': 'openChunk',
                              'pq': 'printQueue',
                              'prq': 'printRedoQueue',
                              'q': 'query',
                              'rn': 'redoName',
                              'rqe': 'redoQueueEmpty',
                              'st': 'state',
                              'swf': 'stateWithoutFlush',
                              'un': 'undoName',
                              'uqe': 'undoQueueEmpty'},
               'type': 'system'},
  'unfold': {'flags': {'applyToShell': {'args': '<type __builtin__.bool>',
                                        'longname': 'applyToShell',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'applyToShell'},
                       'areaWeight': {'args': '<type __builtin__.float>',
                                      'longname': 'areaWeight',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'aw'},
                       'globalBlend': {'args': '<type __builtin__.float>',
                                       'longname': 'globalBlend',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'gb'},
                       'globalMethodBlend': {'args': '<type __builtin__.float>',
                                             'longname': 'globalMethodBlend',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'gmb'},
                       'iterations': {'args': '<type __builtin__.int>',
                                      'longname': 'iterations',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'i'},
                       'optimizeAxis': {'args': '<type __builtin__.int>',
                                        'longname': 'optimizeAxis',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'oa'},
                       'pinSelected': {'args': '<type __builtin__.bool>',
                                       'longname': 'pinSelected',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ps'},
                       'pinUvBorder': {'args': '<type __builtin__.bool>',
                                       'longname': 'pinUvBorder',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'pub'},
                       'scale': {'args': '<type __builtin__.float>',
                                 'longname': 'scale',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 's'},
                       'stoppingThreshold': {'args': '<type __builtin__.float>',
                                             'longname': 'stoppingThreshold',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ss'},
                       'useScale': {'args': '<type __builtin__.bool>',
                                    'longname': 'useScale',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'us'}},
             'removedFlags': {'as': 'applyToShell'},
             'shortFlags': {'applyToShell': 'applyToShell',
                            'aw': 'areaWeight',
                            'gb': 'globalBlend',
                            'gmb': 'globalMethodBlend',
                            'i': 'iterations',
                            'oa': 'optimizeAxis',
                            'ps': 'pinSelected',
                            'pub': 'pinUvBorder',
                            's': 'scale',
                            'ss': 'stoppingThreshold',
                            'us': 'useScale'},
             'type': 'modeling'},
  'ungroup': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                     'longname': 'absolute',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'a'},
                        'parent': {'args': '<type __builtin__.str>',
                                   'longname': 'parent',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 'p'},
                        'relative': {'args': '<type __builtin__.bool>',
                                     'longname': 'relative',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'r'},
                        'world': {'args': '<type __builtin__.bool>',
                                  'longname': 'world',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'w'}},
              'shortFlags': {'a': 'absolute',
                             'p': 'parent',
                             'r': 'relative',
                             'w': 'world'},
              'type': 'general'},
  'uniform': {'flags': {'attenuation': {'args': '<type __builtin__.float>',
                                        'longname': 'attenuation',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'att'},
                        'directionX': {'args': '<type __builtin__.float>',
                                       'longname': 'directionX',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dx'},
                        'directionY': {'args': '<type __builtin__.float>',
                                       'longname': 'directionY',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dy'},
                        'directionZ': {'args': '<type __builtin__.float>',
                                       'longname': 'directionZ',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'dz'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'magnitude': {'args': '<type __builtin__.float>',
                                      'longname': 'magnitude',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                        'maxDistance': {'args': '<type __builtin__.float>',
                                        'longname': 'maxDistance',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'mxd'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'perVertex': {'args': '<type __builtin__.bool>',
                                      'longname': 'perVertex',
                                      'modes': ['query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'pv'},
                        'position': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'position',
                                     'modes': ['query',
                                               'edit',
                                               'multiuse'],
                                     'numArgs': 3,
                                     'shortname': 'pos'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'torusSectionRadius': {'args': '<type __builtin__.float>',
                                               'longname': 'torusSectionRadius',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'tsr'},
                        'volumeExclusion': {'args': '<type __builtin__.bool>',
                                            'longname': 'volumeExclusion',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'vex'},
                        'volumeOffset': {'args': ['<type __builtin__.float>',
                                                  '<type __builtin__.float>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'volumeOffset',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 3,
                                         'shortname': 'vof'},
                        'volumeShape': {'args': '<type __builtin__.str>',
                                        'longname': 'volumeShape',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vsh'},
                        'volumeSweep': {'args': '<type __builtin__.float>',
                                        'longname': 'volumeSweep',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'vsw'}},
              'resultNeedsUnpacking': True,
              'shortFlags': {'att': 'attenuation',
                             'dx': 'directionX',
                             'dy': 'directionY',
                             'dz': 'directionZ',
                             'e': 'edit',
                             'm': 'magnitude',
                             'mxd': 'maxDistance',
                             'n': 'name',
                             'pos': 'position',
                             'pv': 'perVertex',
                             'q': 'query',
                             'tsr': 'torusSectionRadius',
                             'vex': 'volumeExclusion',
                             'vof': 'volumeOffset',
                             'vsh': 'volumeShape',
                             'vsw': 'volumeSweep'},
              'type': 'effects'},
  'unknownNode': {'flags': {'plugin': {'args': '<type __builtin__.bool>',
                                       'longname': 'plugin',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'p'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'},
                            'realClassName': {'args': '<type __builtin__.bool>',
                                              'longname': 'realClassName',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'rcn'},
                            'realClassTag': {'args': '<type __builtin__.bool>',
                                             'longname': 'realClassTag',
                                             'modes': ['query'],
                                             'numArgs': 0,
                                             'shortname': 'rct'}},
                  'shortFlags': {'p': 'plugin',
                                 'q': 'query',
                                 'rcn': 'realClassName',
                                 'rct': 'realClassTag'},
                  'type': 'system'},
  'unknownPlugin': {'flags': {'dataTypes': {'args': '<type __builtin__.bool>',
                                            'longname': 'dataTypes',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'dt'},
                              'list': {'args': '<type __builtin__.bool>',
                                       'longname': 'list',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'l'},
                              'nodeTypes': {'args': '<type __builtin__.bool>',
                                            'longname': 'nodeTypes',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'nt'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'remove': {'args': '<type __builtin__.bool>',
                                         'longname': 'remove',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'r'},
                              'version': {'args': '<type __builtin__.bool>',
                                          'longname': 'version',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'v'}},
                    'shortFlags': {'dt': 'dataTypes',
                                   'l': 'list',
                                   'nt': 'nodeTypes',
                                   'q': 'query',
                                   'r': 'remove',
                                   'v': 'version'},
                    'type': 'system'},
  'unloadPlugin': {'callbackFlags': ['removeCallback',
                                     'rc',
                                     'addCallback',
                                     'ac'],
                   'flags': {'addCallback': {'args': 'script',
                                             'longname': 'addCallback',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'ac'},
                             'force': {'args': '<type __builtin__.bool>',
                                       'longname': 'force',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'f'},
                             'removeCallback': {'args': 'script',
                                                'longname': 'removeCallback',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'rc'}},
                   'shortFlags': {'ac': 'addCallback',
                                  'f': 'force',
                                  'rc': 'removeCallback'},
                   'type': 'system'},
  'untangleUV': {'flags': {'mapBorder': {'args': '<type __builtin__.str>',
                                         'longname': 'mapBorder',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'mb'},
                           'maxRelaxIterations': {'args': '<type __builtin__.int>',
                                                  'longname': 'maxRelaxIterations',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'mri'},
                           'pinBorder': {'args': '<type __builtin__.bool>',
                                         'longname': 'pinBorder',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'pb'},
                           'pinSelected': {'args': '<type __builtin__.bool>',
                                           'longname': 'pinSelected',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ps'},
                           'pinUnselected': {'args': '<type __builtin__.bool>',
                                             'longname': 'pinUnselected',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'pu'},
                           'relax': {'args': '<type __builtin__.str>',
                                     'longname': 'relax',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'r'},
                           'relaxTolerance': {'args': '<type __builtin__.float>',
                                              'longname': 'relaxTolerance',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'rt'},
                           'shapeDetail': {'args': '<type __builtin__.float>',
                                           'longname': 'shapeDetail',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'sd'}},
                 'shortFlags': {'mb': 'mapBorder',
                                'mri': 'maxRelaxIterations',
                                'pb': 'pinBorder',
                                'ps': 'pinSelected',
                                'pu': 'pinUnselected',
                                'r': 'relax',
                                'rt': 'relaxTolerance',
                                'sd': 'shapeDetail'},
                 'type': 'modeling'},
  'untrim': {'flags': {'caching': {'args': '<type __builtin__.bool>',
                                   'longname': 'caching',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'cch'},
                       'constructionHistory': {'args': '<type __builtin__.bool>',
                                               'longname': 'constructionHistory',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ch'},
                       'curveOnSurface': {'args': '<type __builtin__.bool>',
                                          'longname': 'curveOnSurface',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'cos'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'frozen': {'args': '<type __builtin__.bool>',
                                  'longname': 'frozen',
                                  'numArgs': 1,
                                  'shortname': 'fzn'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['create'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'noChanges': {'args': '<type __builtin__.bool>',
                                     'longname': 'noChanges',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'nc'},
                       'nodeState': {'args': '<type __builtin__.int>',
                                     'longname': 'nodeState',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'nds'},
                       'object': {'args': '<type __builtin__.bool>',
                                  'longname': 'object',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'o'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'replaceOriginal': {'args': '<type __builtin__.bool>',
                                           'longname': 'replaceOriginal',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'rpo'},
                       'untrimAll': {'args': '<type __builtin__.bool>',
                                     'longname': 'untrimAll',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'all'}},
             'shortFlags': {'all': 'untrimAll',
                            'cch': 'caching',
                            'ch': 'constructionHistory',
                            'cos': 'curveOnSurface',
                            'e': 'edit',
                            'fzn': 'frozen',
                            'n': 'name',
                            'nc': 'noChanges',
                            'nds': 'nodeState',
                            'o': 'object',
                            'q': 'query',
                            'rpo': 'replaceOriginal'},
             'type': 'modeling'},
  'upAxis': {'flags': {'axis': {'args': '<type __builtin__.str>',
                                'longname': 'axis',
                                'modes': ['query'],
                                'numArgs': 1,
                                'shortname': 'ax'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'rotateView': {'args': '<type __builtin__.bool>',
                                      'longname': 'rotateView',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'rv'}},
             'shortFlags': {'ax': 'axis', 'q': 'query', 'rv': 'rotateView'},
             'type': 'general'},
  'userCtx': {'callbackFlags': ['finalCommand', 'fc', 'editCommand', 'ec'],
              'flags': {'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'editCommand': {'args': '<type __builtin__.callable>',
                                        'longname': 'editCommand',
                                        'numArgs': 1,
                                        'shortname': 'ec'},
                        'editPrompt': {'args': '<type __builtin__.str>',
                                       'longname': 'editPrompt',
                                       'numArgs': 1,
                                       'shortname': 'ep'},
                        'exists': {'args': '<type __builtin__.bool>',
                                   'longname': 'exists',
                                   'numArgs': 0,
                                   'shortname': 'ex'},
                        'finalCommand': {'args': '<type __builtin__.str>',
                                         'longname': 'finalCommand',
                                         'numArgs': 1,
                                         'shortname': 'fc'},
                        'history': {'args': '<type __builtin__.bool>',
                                    'longname': 'history',
                                    'numArgs': 1,
                                    'shortname': 'ch'},
                        'image1': {'args': '<type __builtin__.str>',
                                   'longname': 'image1',
                                   'numArgs': 1,
                                   'shortname': 'i1'},
                        'image2': {'args': '<type __builtin__.str>',
                                   'longname': 'image2',
                                   'numArgs': 1,
                                   'shortname': 'i2'},
                        'image3': {'args': '<type __builtin__.str>',
                                   'longname': 'image3',
                                   'numArgs': 1,
                                   'shortname': 'i3'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'noSelectionPrompt': {'args': ['<type __builtin__.int>',
                                                       '<type __builtin__.str>'],
                                              'longname': 'noSelectionPrompt',
                                              'modes': ['multiuse'],
                                              'numArgs': 2,
                                              'shortname': 'nsp'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'selectionCount': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'selectionCount',
                                           'modes': ['multiuse'],
                                           'numArgs': 2,
                                           'shortname': 'sc'},
                        'selectionFlag': {'args': ['<type __builtin__.int>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'selectionFlag',
                                          'modes': ['multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'flg'},
                        'selectionMask': {'args': ['<type __builtin__.int>',
                                                   '<type __builtin__.str>'],
                                          'longname': 'selectionMask',
                                          'modes': ['multiuse'],
                                          'numArgs': 2,
                                          'shortname': 'sm'},
                        'selectionPrompt': {'args': '<type __builtin__.str>',
                                            'longname': 'selectionPrompt',
                                            'modes': ['multiuse'],
                                            'numArgs': 1,
                                            'shortname': 'sp'}},
              'shortFlags': {'ch': 'history',
                             'e': 'edit',
                             'ec': 'editCommand',
                             'ep': 'editPrompt',
                             'ex': 'exists',
                             'fc': 'finalCommand',
                             'flg': 'selectionFlag',
                             'i1': 'image1',
                             'i2': 'image2',
                             'i3': 'image3',
                             'n': 'name',
                             'nsp': 'noSelectionPrompt',
                             'q': 'query',
                             'sc': 'selectionCount',
                             'sm': 'selectionMask',
                             'sp': 'selectionPrompt'},
              'type': 'context'},
  'uvLink': {'flags': {'b': {'args': '<type __builtin__.bool>',
                             'longname': 'b',
                             'modes': ['create'],
                             'numArgs': 0,
                             'shortname': 'b'},
                       'isValid': {'args': '<type __builtin__.bool>',
                                   'longname': 'isValid',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'iv'},
                       'make': {'args': '<type __builtin__.bool>',
                                'longname': 'make',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 'm'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'queryObject': {'args': 'PyNode',
                                       'longname': 'queryObject',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'qo'},
                       'texture': {'args': 'PyNode',
                                   'longname': 'texture',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 't'},
                       'uvSet': {'args': 'PyNode',
                                 'longname': 'uvSet',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'uvs'}},
             'removedFlags': {'break': 'b'},
             'shortFlags': {'b': 'b',
                            'iv': 'isValid',
                            'm': 'make',
                            'q': 'query',
                            'qo': 'queryObject',
                            't': 'texture',
                            'uvs': 'uvSet'},
             'type': 'rendering'},
  'uvSnapshot': {'flags': {'antiAliased': {'args': '<type __builtin__.bool>',
                                           'longname': 'antiAliased',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'aa'},
                           'blueColor': {'args': '<type __builtin__.int>',
                                         'longname': 'blueColor',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'b'},
                           'entireUVRange': {'args': '<type __builtin__.bool>',
                                             'longname': 'entireUVRange',
                                             'modes': ['create'],
                                             'numArgs': 0,
                                             'shortname': 'euv'},
                           'fileFormat': {'args': '<type __builtin__.str>',
                                          'longname': 'fileFormat',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ff'},
                           'greenColor': {'args': '<type __builtin__.int>',
                                          'longname': 'greenColor',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'g'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'overwrite': {'args': '<type __builtin__.bool>',
                                         'longname': 'overwrite',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'o'},
                           'redColor': {'args': '<type __builtin__.int>',
                                        'longname': 'redColor',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'r'},
                           'uMax': {'args': '<type __builtin__.float>',
                                    'longname': 'uMax',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'umx'},
                           'uMin': {'args': '<type __builtin__.float>',
                                    'longname': 'uMin',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'umn'},
                           'uvSetName': {'args': '<type __builtin__.str>',
                                         'longname': 'uvSetName',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'uvs'},
                           'vMax': {'args': '<type __builtin__.float>',
                                    'longname': 'vMax',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'vmx'},
                           'vMin': {'args': '<type __builtin__.float>',
                                    'longname': 'vMin',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'vmn'},
                           'xResolution': {'args': '<type __builtin__.int>',
                                           'longname': 'xResolution',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'xr'},
                           'yResolution': {'args': '<type __builtin__.int>',
                                           'longname': 'yResolution',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'yr'}},
                 'shortFlags': {'aa': 'antiAliased',
                                'b': 'blueColor',
                                'euv': 'entireUVRange',
                                'ff': 'fileFormat',
                                'g': 'greenColor',
                                'n': 'name',
                                'o': 'overwrite',
                                'r': 'redColor',
                                'umn': 'uMin',
                                'umx': 'uMax',
                                'uvs': 'uvSetName',
                                'vmn': 'vMin',
                                'vmx': 'vMax',
                                'xr': 'xResolution',
                                'yr': 'yResolution'},
                 'type': 'modeling'},
  'view2dToolCtx': {'flags': {'alternateContext': {'args': '<type __builtin__.bool>',
                                                   'longname': 'alternateContext',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 1,
                                                   'shortname': 'ac'},
                              'boxzoom': {'args': '<type __builtin__.bool>',
                                          'longname': 'boxzoom',
                                          'modes': ['create', 'query'],
                                          'numArgs': 0,
                                          'shortname': 'bz'},
                              'dolly': {'args': '<type __builtin__.bool>',
                                        'longname': 'dolly',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'do'},
                              'edit': {'args': '<type __builtin__.bool>',
                                       'longname': 'edit',
                                       'numArgs': 0,
                                       'shortname': 'e'},
                              'exists': {'args': '<type __builtin__.bool>',
                                         'longname': 'exists',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ex'},
                              'history': {'args': '<type __builtin__.bool>',
                                          'longname': 'history',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'ch'},
                              'image1': {'args': '<type __builtin__.str>',
                                         'longname': 'image1',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i1'},
                              'image2': {'args': '<type __builtin__.str>',
                                         'longname': 'image2',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i2'},
                              'image3': {'args': '<type __builtin__.str>',
                                         'longname': 'image3',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'i3'},
                              'name': {'args': '<type __builtin__.str>',
                                       'longname': 'name',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'n'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'toolName': {'args': '<type __builtin__.str>',
                                           'longname': 'toolName',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'tn'},
                              'track': {'args': '<type __builtin__.bool>',
                                        'longname': 'track',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'tr'}},
                    'shortFlags': {'ac': 'alternateContext',
                                   'bz': 'boxzoom',
                                   'ch': 'history',
                                   'do': 'dolly',
                                   'e': 'edit',
                                   'ex': 'exists',
                                   'i1': 'image1',
                                   'i2': 'image2',
                                   'i3': 'image3',
                                   'n': 'name',
                                   'q': 'query',
                                   'tn': 'toolName',
                                   'tr': 'track'},
                    'type': 'context'},
  'viewCamera': {'flags': {'move': {'args': 'PyNode',
                                    'longname': 'move',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'm'},
                           'sideView': {'args': '<type __builtin__.bool>',
                                        'longname': 'sideView',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 's'},
                           'topView': {'args': '<type __builtin__.bool>',
                                       'longname': 'topView',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 't'}},
                 'shortFlags': {'m': 'move',
                                's': 'sideView',
                                't': 'topView'},
                 'type': 'rendering'},
  'viewClipPlane': {'flags': {'autoClipPlane': {'args': '<type __builtin__.bool>',
                                                'longname': 'autoClipPlane',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 0,
                                                'shortname': 'acp'},
                              'farClipPlane': {'args': '<type __builtin__.float>',
                                               'longname': 'farClipPlane',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'fcp'},
                              'nearClipPlane': {'args': '<type __builtin__.float>',
                                                'longname': 'nearClipPlane',
                                                'modes': ['create',
                                                          'query'],
                                                'numArgs': 1,
                                                'shortname': 'ncp'},
                              'query': {'args': '<type __builtin__.bool>',
                                        'longname': 'query',
                                        'numArgs': 0,
                                        'shortname': 'q'},
                              'surfacesOnly': {'args': '<type __builtin__.bool>',
                                               'longname': 'surfacesOnly',
                                               'modes': ['create'],
                                               'numArgs': 0,
                                               'shortname': 'so'}},
                    'shortFlags': {'acp': 'autoClipPlane',
                                   'fcp': 'farClipPlane',
                                   'ncp': 'nearClipPlane',
                                   'q': 'query',
                                   'so': 'surfacesOnly'},
                    'type': 'rendering'},
  'viewFit': {'flags': {'allObjects': {'args': '<type __builtin__.bool>',
                                       'longname': 'allObjects',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'all'},
                        'animate': {'args': '<type __builtin__.bool>',
                                    'longname': 'animate',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'an'},
                        'center': {'args': '<type __builtin__.bool>',
                                   'longname': 'center',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'c'},
                        'fitFactor': {'args': '<type __builtin__.float>',
                                      'longname': 'fitFactor',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'f'},
                        'namespace': {'args': '<type __builtin__.str>',
                                      'longname': 'namespace',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'ns'},
                        'panel': {'args': '<type __builtin__.str>',
                                  'longname': 'panel',
                                  'numArgs': 1,
                                  'shortname': 'p'}},
              'shortFlags': {'all': 'allObjects',
                             'an': 'animate',
                             'c': 'center',
                             'f': 'fitFactor',
                             'ns': 'namespace',
                             'p': 'panel'},
              'type': 'rendering'},
  'viewHeadOn': {'flags': {}, 'shortFlags': {}, 'type': 'rendering'},
  'viewLookAt': {'flags': {'position': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'position',
                                        'modes': ['create'],
                                        'numArgs': 3,
                                        'shortname': 'pos'}},
                 'shortFlags': {'pos': 'position'},
                 'type': 'rendering'},
  'viewManip': {'callbackFlags': ['preCommand', 'pr', 'postCommand', 'p'],
                'flags': {'bottomLeft': {'args': '<type __builtin__.bool>',
                                         'longname': 'bottomLeft',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'bl'},
                          'bottomRight': {'args': '<type __builtin__.bool>',
                                          'longname': 'bottomRight',
                                          'modes': ['create', 'query'],
                                          'numArgs': 0,
                                          'shortname': 'br'},
                          'compassAngle': {'args': '<type __builtin__.float>',
                                           'longname': 'compassAngle',
                                           'modes': ['create', 'query'],
                                           'numArgs': 1,
                                           'shortname': 'ca'},
                          'dragSnap': {'args': '<type __builtin__.bool>',
                                       'longname': 'dragSnap',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'ds'},
                          'drawCompass': {'args': '<type __builtin__.bool>',
                                          'longname': 'drawCompass',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'dc'},
                          'fitToView': {'args': '<type __builtin__.bool>',
                                        'longname': 'fitToView',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ftv'},
                          'frontParameters': {'args': '<type __builtin__.str>',
                                              'longname': 'frontParameters',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'fp'},
                          'goDefault': {'args': '<type __builtin__.bool>',
                                        'longname': 'goDefault',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'gd'},
                          'goHome': {'args': '<type __builtin__.bool>',
                                     'longname': 'goHome',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 'gh'},
                          'homeParameters': {'args': '<type __builtin__.str>',
                                             'longname': 'homeParameters',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'hp'},
                          'levelCamera': {'args': '<type __builtin__.bool>',
                                          'longname': 'levelCamera',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'lc'},
                          'minOpacity': {'args': '<type __builtin__.float>',
                                         'longname': 'minOpacity',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'mo'},
                          'namespace': {'args': '<type __builtin__.str>',
                                        'longname': 'namespace',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'ns'},
                          'postCommand': {'args': '<type __builtin__.str>',
                                          'longname': 'postCommand',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'p'},
                          'preCommand': {'args': '<type __builtin__.str>',
                                         'longname': 'preCommand',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'pr'},
                          'preserveSceneUp': {'args': '<type __builtin__.bool>',
                                              'longname': 'preserveSceneUp',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'psu'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'resetFront': {'args': '<type __builtin__.bool>',
                                         'longname': 'resetFront',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'rf'},
                          'resetHome': {'args': '<type __builtin__.bool>',
                                        'longname': 'resetHome',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'rh'},
                          'restoreCenter': {'args': '<type __builtin__.bool>',
                                            'longname': 'restoreCenter',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'rc'},
                          'selectionLockParameters': {'args': '<type __builtin__.str>',
                                                      'longname': 'selectionLockParameters',
                                                      'modes': ['create',
                                                                'query'],
                                                      'numArgs': 1,
                                                      'shortname': 'slp'},
                          'setFront': {'args': '<type __builtin__.bool>',
                                       'longname': 'setFront',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'sf'},
                          'setHome': {'args': '<type __builtin__.bool>',
                                      'longname': 'setHome',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'sh'},
                          'size': {'args': '<type __builtin__.str>',
                                   'longname': 'size',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 's'},
                          'toggleSelectionLock': {'args': '<type __builtin__.bool>',
                                                  'longname': 'toggleSelectionLock',
                                                  'modes': ['create'],
                                                  'numArgs': 0,
                                                  'shortname': 'tsl'},
                          'topLeft': {'args': '<type __builtin__.bool>',
                                      'longname': 'topLeft',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'tl'},
                          'topRight': {'args': '<type __builtin__.bool>',
                                       'longname': 'topRight',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'tr'},
                          'visible': {'args': '<type __builtin__.bool>',
                                      'longname': 'visible',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'v'},
                          'zoomToFitScene': {'args': '<type __builtin__.bool>',
                                             'longname': 'zoomToFitScene',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'zf'}},
                'shortFlags': {'bl': 'bottomLeft',
                               'br': 'bottomRight',
                               'ca': 'compassAngle',
                               'dc': 'drawCompass',
                               'ds': 'dragSnap',
                               'fp': 'frontParameters',
                               'ftv': 'fitToView',
                               'gd': 'goDefault',
                               'gh': 'goHome',
                               'hp': 'homeParameters',
                               'lc': 'levelCamera',
                               'mo': 'minOpacity',
                               'ns': 'namespace',
                               'p': 'postCommand',
                               'pr': 'preCommand',
                               'psu': 'preserveSceneUp',
                               'q': 'query',
                               'rc': 'restoreCenter',
                               'rf': 'resetFront',
                               'rh': 'resetHome',
                               's': 'size',
                               'sf': 'setFront',
                               'sh': 'setHome',
                               'slp': 'selectionLockParameters',
                               'tl': 'topLeft',
                               'tr': 'topRight',
                               'tsl': 'toggleSelectionLock',
                               'v': 'visible',
                               'zf': 'zoomToFitScene'},
                'type': 'windows'},
  'viewPlace': {'flags': {'animate': {'args': '<type __builtin__.bool>',
                                      'longname': 'animate',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'an'},
                          'eyePoint': {'args': ['<type __builtin__.float>',
                                                '<type __builtin__.float>',
                                                '<type __builtin__.float>'],
                                       'longname': 'eyePoint',
                                       'modes': ['create'],
                                       'numArgs': 3,
                                       'shortname': 'eye'},
                          'fieldOfView': {'args': '<type __builtin__.float>',
                                          'longname': 'fieldOfView',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'fov'},
                          'lookAt': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'lookAt',
                                     'modes': ['create'],
                                     'numArgs': 3,
                                     'shortname': 'la'},
                          'ortho': {'args': '<type __builtin__.bool>',
                                    'longname': 'ortho',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'o'},
                          'perspective': {'args': '<type __builtin__.bool>',
                                          'longname': 'perspective',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'p'},
                          'upDirection': {'args': ['<type __builtin__.float>',
                                                   '<type __builtin__.float>',
                                                   '<type __builtin__.float>'],
                                          'longname': 'upDirection',
                                          'modes': ['create'],
                                          'numArgs': 3,
                                          'shortname': 'up'},
                          'viewDirection': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'viewDirection',
                                            'modes': ['create'],
                                            'numArgs': 3,
                                            'shortname': 'vd'}},
                'shortFlags': {'an': 'animate',
                               'eye': 'eyePoint',
                               'fov': 'fieldOfView',
                               'la': 'lookAt',
                               'o': 'ortho',
                               'p': 'perspective',
                               'up': 'upDirection',
                               'vd': 'viewDirection'},
                'type': 'rendering'},
  'viewSet': {'flags': {'animate': {'args': '<type __builtin__.bool>',
                                    'longname': 'animate',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'an'},
                        'back': {'args': '<type __builtin__.bool>',
                                 'longname': 'back',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'b'},
                        'bottom': {'args': '<type __builtin__.bool>',
                                   'longname': 'bottom',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'bo'},
                        'fit': {'args': '<type __builtin__.bool>',
                                'longname': 'fit',
                                'modes': ['create', 'query'],
                                'numArgs': 0,
                                'shortname': 'fit'},
                        'fitFactor': {'args': '<type __builtin__.float>',
                                      'longname': 'fitFactor',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'ff'},
                        'front': {'args': '<type __builtin__.bool>',
                                  'longname': 'front',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'f'},
                        'home': {'args': '<type __builtin__.bool>',
                                 'longname': 'home',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 'h'},
                        'keepRenderSettings': {'args': '<type __builtin__.bool>',
                                               'longname': 'keepRenderSettings',
                                               'modes': ['create',
                                                         'query'],
                                               'numArgs': 1,
                                               'shortname': 'krs'},
                        'leftSide': {'args': '<type __builtin__.bool>',
                                     'longname': 'leftSide',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'ls'},
                        'namespace': {'args': '<type __builtin__.str>',
                                      'longname': 'namespace',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'ns'},
                        'nextView': {'args': '<type __builtin__.bool>',
                                     'longname': 'nextView',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 'nv'},
                        'persp': {'args': '<type __builtin__.bool>',
                                  'longname': 'persp',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'p'},
                        'previousView': {'args': '<type __builtin__.bool>',
                                         'longname': 'previousView',
                                         'modes': ['create', 'query'],
                                         'numArgs': 0,
                                         'shortname': 'pv'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'rightSide': {'args': '<type __builtin__.bool>',
                                      'longname': 'rightSide',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'rs'},
                        'side': {'args': '<type __builtin__.bool>',
                                 'longname': 'side',
                                 'modes': ['create'],
                                 'numArgs': 0,
                                 'shortname': 's'},
                        't': {'args': '<type __builtin__.bool>',
                              'longname': 't',
                              'numArgs': 0,
                              'shortname': 't'},
                        'top': {'args': '<type __builtin__.bool>',
                                'longname': 'top',
                                'modes': ['create'],
                                'numArgs': 0,
                                'shortname': 't'},
                        'viewNegativeX': {'args': '<type __builtin__.bool>',
                                          'longname': 'viewNegativeX',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'vnx'},
                        'viewNegativeY': {'args': '<type __builtin__.bool>',
                                          'longname': 'viewNegativeY',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'vny'},
                        'viewNegativeZ': {'args': '<type __builtin__.bool>',
                                          'longname': 'viewNegativeZ',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'vnz'},
                        'viewX': {'args': '<type __builtin__.bool>',
                                  'longname': 'viewX',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'vx'},
                        'viewY': {'args': '<type __builtin__.bool>',
                                  'longname': 'viewY',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'vy'},
                        'viewZ': {'args': '<type __builtin__.bool>',
                                  'longname': 'viewZ',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'vz'}},
              'shortFlags': {'an': 'animate',
                             'b': 'back',
                             'bo': 'bottom',
                             'f': 'front',
                             'ff': 'fitFactor',
                             'fit': 'fit',
                             'h': 'home',
                             'krs': 'keepRenderSettings',
                             'ls': 'leftSide',
                             'ns': 'namespace',
                             'nv': 'nextView',
                             'p': 'persp',
                             'pv': 'previousView',
                             'q': 'query',
                             'rs': 'rightSide',
                             's': 'side',
                             't': 't',
                             'top': 'top',
                             'vnx': 'viewNegativeX',
                             'vny': 'viewNegativeY',
                             'vnz': 'viewNegativeZ',
                             'vx': 'viewX',
                             'vy': 'viewY',
                             'vz': 'viewZ'},
              'type': 'rendering'},
  'visor': {'callbackFlags': ['command', 'cmd'],
            'flags': {'addFolder': {'args': '<type __builtin__.bool>',
                                    'longname': 'addFolder',
                                    'modes': ['create', 'query'],
                                    'numArgs': 0,
                                    'shortname': 'add'},
                      'addNodes': {'args': '<type __builtin__.str>',
                                   'longname': 'addNodes',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'adn'},
                      'allowPanningInX': {'args': '<type __builtin__.bool>',
                                          'longname': 'allowPanningInX',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'apx'},
                      'allowPanningInY': {'args': '<type __builtin__.bool>',
                                          'longname': 'allowPanningInY',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'apy'},
                      'allowZooming': {'args': '<type __builtin__.bool>',
                                       'longname': 'allowZooming',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'az'},
                      'command': {'args': '<type __builtin__.str>',
                                  'longname': 'command',
                                  'modes': ['create', 'query'],
                                  'numArgs': 1,
                                  'shortname': 'cmd'},
                      'deleteFolder': {'args': '<type __builtin__.str>',
                                       'longname': 'deleteFolder',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'deleteFolder'},
                      'editFolder': {'args': '<type __builtin__.str>',
                                     'longname': 'editFolder',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'edf'},
                      'folderList': {'args': '<type __builtin__.str>',
                                     'longname': 'folderList',
                                     'modes': ['query'],
                                     'numArgs': 1,
                                     'shortname': 'fl'},
                      'menu': {'args': '<type __builtin__.str>',
                               'longname': 'menu',
                               'modes': ['create', 'query'],
                               'numArgs': 1,
                               'shortname': 'mn'},
                      'name': {'args': '<type __builtin__.str>',
                               'longname': 'name',
                               'modes': ['create', 'query'],
                               'numArgs': 1,
                               'shortname': 'n'},
                      'nodeType': {'args': '<type __builtin__.str>',
                                   'longname': 'nodeType',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'ntp'},
                      'openDirectories': {'args': '<type __builtin__.bool>',
                                          'longname': 'openDirectories',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'opd'},
                      'openFolder': {'args': '<type __builtin__.bool>',
                                     'longname': 'openFolder',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'opf'},
                      'parent': {'args': '<type __builtin__.str>',
                                 'longname': 'parent',
                                 'modes': ['create', 'query'],
                                 'numArgs': 1,
                                 'shortname': 'p'},
                      'path': {'args': '<type __builtin__.str>',
                               'longname': 'path',
                               'modes': ['create', 'query'],
                               'numArgs': 1,
                               'shortname': 'pth'},
                      'popupMenuScript': {'args': '<type __builtin__.str>',
                                          'longname': 'popupMenuScript',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'pms'},
                      'query': {'args': '<type __builtin__.bool>',
                                'longname': 'query',
                                'numArgs': 0,
                                'shortname': 'q'},
                      'rebuild': {'args': '<type __builtin__.bool>',
                                  'longname': 'rebuild',
                                  'modes': ['create', 'query'],
                                  'numArgs': 0,
                                  'shortname': 're'},
                      'refreshAllSwatches': {'args': '<type __builtin__.bool>',
                                             'longname': 'refreshAllSwatches',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'ras'},
                      'refreshSelectedSwatches': {'args': '<type __builtin__.bool>',
                                                  'longname': 'refreshSelectedSwatches',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 0,
                                                  'shortname': 'rss'},
                      'refreshSwatch': {'args': '<type __builtin__.str>',
                                        'longname': 'refreshSwatch',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'rs'},
                      'reset': {'args': '<type __builtin__.bool>',
                                'longname': 'reset',
                                'modes': ['create', 'query'],
                                'numArgs': 0,
                                'shortname': 'rst'},
                      'restrictPanAndZoom': {'args': '<type __builtin__.bool>',
                                             'longname': 'restrictPanAndZoom',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'rpz'},
                      'saveSwatches': {'args': '<type __builtin__.bool>',
                                       'longname': 'saveSwatches',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'ss'},
                      'scrollBar': {'args': '<type __builtin__.str>',
                                    'longname': 'scrollBar',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'sb'},
                      'scrollPercent': {'args': '<type __builtin__.float>',
                                        'longname': 'scrollPercent',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'sp'},
                      'selectedGadgets': {'args': '<type __builtin__.str>',
                                          'longname': 'selectedGadgets',
                                          'modes': ['query'],
                                          'numArgs': 1,
                                          'shortname': 'sg'},
                      'showDividers': {'args': '<type __builtin__.bool>',
                                       'longname': 'showDividers',
                                       'modes': ['create', 'query'],
                                       'numArgs': 1,
                                       'shortname': 'sd'},
                      'showFiles': {'args': '<type __builtin__.bool>',
                                    'longname': 'showFiles',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'sfi'},
                      'showFolders': {'args': '<type __builtin__.bool>',
                                      'longname': 'showFolders',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'sfo'},
                      'showNodes': {'args': '<type __builtin__.bool>',
                                    'longname': 'showNodes',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'sn'},
                      'stateString': {'args': '<type __builtin__.bool>',
                                      'longname': 'stateString',
                                      'modes': ['create', 'query'],
                                      'numArgs': 0,
                                      'shortname': 'sts'},
                      'style': {'args': '<type __builtin__.str>',
                                'longname': 'style',
                                'modes': ['create', 'query'],
                                'numArgs': 1,
                                'shortname': 'stl'},
                      'transform': {'args': '<type __builtin__.str>',
                                    'longname': 'transform',
                                    'modes': ['create', 'query'],
                                    'numArgs': 1,
                                    'shortname': 'trn'},
                      'type': {'args': '<type __builtin__.str>',
                               'longname': 'type',
                               'modes': ['create', 'query'],
                               'numArgs': 1,
                               'shortname': 'typ'}},
            'removedFlags': {'del': 'deleteFolder'},
            'shortFlags': {'add': 'addFolder',
                           'adn': 'addNodes',
                           'apx': 'allowPanningInX',
                           'apy': 'allowPanningInY',
                           'az': 'allowZooming',
                           'cmd': 'command',
                           'deleteFolder': 'deleteFolder',
                           'edf': 'editFolder',
                           'fl': 'folderList',
                           'mn': 'menu',
                           'n': 'name',
                           'ntp': 'nodeType',
                           'opd': 'openDirectories',
                           'opf': 'openFolder',
                           'p': 'parent',
                           'pms': 'popupMenuScript',
                           'pth': 'path',
                           'q': 'query',
                           'ras': 'refreshAllSwatches',
                           're': 'rebuild',
                           'rpz': 'restrictPanAndZoom',
                           'rs': 'refreshSwatch',
                           'rss': 'refreshSelectedSwatches',
                           'rst': 'reset',
                           'sb': 'scrollBar',
                           'sd': 'showDividers',
                           'sfi': 'showFiles',
                           'sfo': 'showFolders',
                           'sg': 'selectedGadgets',
                           'sn': 'showNodes',
                           'sp': 'scrollPercent',
                           'ss': 'saveSwatches',
                           'stl': 'style',
                           'sts': 'stateString',
                           'trn': 'transform',
                           'typ': 'type'},
            'type': 'windows'},
  'volumeAxis': {'flags': {'alongAxis': {'args': '<type __builtin__.float>',
                                         'longname': 'alongAxis',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'alx'},
                           'aroundAxis': {'args': '<type __builtin__.float>',
                                          'longname': 'aroundAxis',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'arx'},
                           'attenuation': {'args': '<type __builtin__.float>',
                                           'longname': 'attenuation',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'att'},
                           'awayFromAxis': {'args': '<type __builtin__.float>',
                                            'longname': 'awayFromAxis',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'afx'},
                           'awayFromCenter': {'args': '<type __builtin__.float>',
                                              'longname': 'awayFromCenter',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'afc'},
                           'detailTurbulence': {'args': '<type __builtin__.float>',
                                                'longname': 'detailTurbulence',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dtr'},
                           'directionX': {'args': '<type __builtin__.float>',
                                          'longname': 'directionX',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dx'},
                           'directionY': {'args': '<type __builtin__.float>',
                                          'longname': 'directionY',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dy'},
                           'directionZ': {'args': '<type __builtin__.float>',
                                          'longname': 'directionZ',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'dz'},
                           'directionalSpeed': {'args': '<type __builtin__.float>',
                                                'longname': 'directionalSpeed',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'drs'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'invertAttenuation': {'args': '<type __builtin__.bool>',
                                                 'longname': 'invertAttenuation',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ia'},
                           'magnitude': {'args': '<type __builtin__.float>',
                                         'longname': 'magnitude',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'm'},
                           'maxDistance': {'args': '<type __builtin__.float>',
                                           'longname': 'maxDistance',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'mxd'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'perVertex': {'args': '<type __builtin__.bool>',
                                         'longname': 'perVertex',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'pv'},
                           'position': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'position',
                                        'modes': ['query',
                                                  'edit',
                                                  'multiuse'],
                                        'numArgs': 3,
                                        'shortname': 'pos'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'torusSectionRadius': {'args': '<type __builtin__.float>',
                                                  'longname': 'torusSectionRadius',
                                                  'modes': ['query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'tsr'},
                           'turbulence': {'args': '<type __builtin__.float>',
                                          'longname': 'turbulence',
                                          'modes': ['query', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'trb'},
                           'turbulenceFrequencyX': {'args': '<type __builtin__.float>',
                                                    'longname': 'turbulenceFrequencyX',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'tfx'},
                           'turbulenceFrequencyY': {'args': '<type __builtin__.float>',
                                                    'longname': 'turbulenceFrequencyY',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'tfy'},
                           'turbulenceFrequencyZ': {'args': '<type __builtin__.float>',
                                                    'longname': 'turbulenceFrequencyZ',
                                                    'modes': ['query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'tfz'},
                           'turbulenceOffsetX': {'args': '<type __builtin__.float>',
                                                 'longname': 'turbulenceOffsetX',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'tox'},
                           'turbulenceOffsetY': {'args': '<type __builtin__.float>',
                                                 'longname': 'turbulenceOffsetY',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'toy'},
                           'turbulenceOffsetZ': {'args': '<type __builtin__.float>',
                                                 'longname': 'turbulenceOffsetZ',
                                                 'modes': ['query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'toz'},
                           'turbulenceSpeed': {'args': '<type __builtin__.float>',
                                               'longname': 'turbulenceSpeed',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'trs'},
                           'volumeExclusion': {'args': '<type __builtin__.bool>',
                                               'longname': 'volumeExclusion',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 1,
                                               'shortname': 'vex'},
                           'volumeOffset': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'volumeOffset',
                                            'modes': ['query', 'edit'],
                                            'numArgs': 3,
                                            'shortname': 'vof'},
                           'volumeShape': {'args': '<type __builtin__.str>',
                                           'longname': 'volumeShape',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vsh'},
                           'volumeSweep': {'args': '<type __builtin__.float>',
                                           'longname': 'volumeSweep',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vsw'}},
                 'resultNeedsUnpacking': True,
                 'shortFlags': {'afc': 'awayFromCenter',
                                'afx': 'awayFromAxis',
                                'alx': 'alongAxis',
                                'arx': 'aroundAxis',
                                'att': 'attenuation',
                                'drs': 'directionalSpeed',
                                'dtr': 'detailTurbulence',
                                'dx': 'directionX',
                                'dy': 'directionY',
                                'dz': 'directionZ',
                                'e': 'edit',
                                'ia': 'invertAttenuation',
                                'm': 'magnitude',
                                'mxd': 'maxDistance',
                                'n': 'name',
                                'pos': 'position',
                                'pv': 'perVertex',
                                'q': 'query',
                                'tfx': 'turbulenceFrequencyX',
                                'tfy': 'turbulenceFrequencyY',
                                'tfz': 'turbulenceFrequencyZ',
                                'tox': 'turbulenceOffsetX',
                                'toy': 'turbulenceOffsetY',
                                'toz': 'turbulenceOffsetZ',
                                'trb': 'turbulence',
                                'trs': 'turbulenceSpeed',
                                'tsr': 'torusSectionRadius',
                                'vex': 'volumeExclusion',
                                'vof': 'volumeOffset',
                                'vsh': 'volumeShape',
                                'vsw': 'volumeSweep'},
                 'type': 'effects'},
  'volumeBind': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'influence': {'args': '<type __builtin__.str>',
                                         'longname': 'influence',
                                         'modes': ['query', 'edit'],
                                         'numArgs': 1,
                                         'shortname': 'inf'},
                           'name': {'args': '<type __builtin__.str>',
                                    'longname': 'name',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'n'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'}},
                 'shortFlags': {'e': 'edit',
                                'inf': 'influence',
                                'n': 'name',
                                'q': 'query'},
                 'type': 'animation'},
  'vortex': {'flags': {'attenuation': {'args': '<type __builtin__.float>',
                                       'longname': 'attenuation',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'att'},
                       'axisX': {'args': '<type __builtin__.float>',
                                 'longname': 'axisX',
                                 'modes': ['query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'ax'},
                       'axisY': {'args': '<type __builtin__.float>',
                                 'longname': 'axisY',
                                 'modes': ['query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'ay'},
                       'axisZ': {'args': '<type __builtin__.float>',
                                 'longname': 'axisZ',
                                 'modes': ['query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'az'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'magnitude': {'args': '<type __builtin__.float>',
                                     'longname': 'magnitude',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'm'},
                       'maxDistance': {'args': '<type __builtin__.float>',
                                       'longname': 'maxDistance',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'mxd'},
                       'name': {'args': '<type __builtin__.str>',
                                'longname': 'name',
                                'modes': ['query', 'edit'],
                                'numArgs': 1,
                                'shortname': 'n'},
                       'perVertex': {'args': '<type __builtin__.bool>',
                                     'longname': 'perVertex',
                                     'modes': ['query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'pv'},
                       'position': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'position',
                                    'modes': ['query', 'edit', 'multiuse'],
                                    'numArgs': 3,
                                    'shortname': 'pos'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'torusSectionRadius': {'args': '<type __builtin__.float>',
                                              'longname': 'torusSectionRadius',
                                              'modes': ['query', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'tsr'},
                       'volumeExclusion': {'args': '<type __builtin__.bool>',
                                           'longname': 'volumeExclusion',
                                           'modes': ['query', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'vex'},
                       'volumeOffset': {'args': ['<type __builtin__.float>',
                                                 '<type __builtin__.float>',
                                                 '<type __builtin__.float>'],
                                        'longname': 'volumeOffset',
                                        'modes': ['query', 'edit'],
                                        'numArgs': 3,
                                        'shortname': 'vof'},
                       'volumeShape': {'args': '<type __builtin__.str>',
                                       'longname': 'volumeShape',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vsh'},
                       'volumeSweep': {'args': '<type __builtin__.float>',
                                       'longname': 'volumeSweep',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vsw'}},
             'resultNeedsUnpacking': True,
             'shortFlags': {'att': 'attenuation',
                            'ax': 'axisX',
                            'ay': 'axisY',
                            'az': 'axisZ',
                            'e': 'edit',
                            'm': 'magnitude',
                            'mxd': 'maxDistance',
                            'n': 'name',
                            'pos': 'position',
                            'pv': 'perVertex',
                            'q': 'query',
                            'tsr': 'torusSectionRadius',
                            'vex': 'volumeExclusion',
                            'vof': 'volumeOffset',
                            'vsh': 'volumeShape',
                            'vsw': 'volumeSweep'},
             'type': 'effects'},
  'waitCursor': {'flags': {'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'state': {'args': '<type __builtin__.bool>',
                                     'longname': 'state',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'st'}},
                 'shortFlags': {'q': 'query', 'st': 'state'},
                 'type': 'language'},
  'walkCtx': {'flags': {'alternateContext': {'args': '<type __builtin__.bool>',
                                             'longname': 'alternateContext',
                                             'modes': ['create', 'query'],
                                             'numArgs': 1,
                                             'shortname': 'ac'},
                        'crouchCount': {'args': '<type __builtin__.float>',
                                        'longname': 'crouchCount',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'wcc'},
                        'edit': {'args': '<type __builtin__.bool>',
                                 'longname': 'edit',
                                 'numArgs': 0,
                                 'shortname': 'e'},
                        'exists': {'args': '<type __builtin__.bool>',
                                   'longname': 'exists',
                                   'modes': ['create'],
                                   'numArgs': 0,
                                   'shortname': 'ex'},
                        'history': {'args': '<type __builtin__.bool>',
                                    'longname': 'history',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'ch'},
                        'image1': {'args': '<type __builtin__.str>',
                                   'longname': 'image1',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'i1'},
                        'image2': {'args': '<type __builtin__.str>',
                                   'longname': 'image2',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'i2'},
                        'image3': {'args': '<type __builtin__.str>',
                                   'longname': 'image3',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'i3'},
                        'name': {'args': '<type __builtin__.str>',
                                 'longname': 'name',
                                 'modes': ['create'],
                                 'numArgs': 1,
                                 'shortname': 'n'},
                        'query': {'args': '<type __builtin__.bool>',
                                  'longname': 'query',
                                  'numArgs': 0,
                                  'shortname': 'q'},
                        'toolName': {'args': '<type __builtin__.str>',
                                     'longname': 'toolName',
                                     'modes': ['create', 'query'],
                                     'numArgs': 1,
                                     'shortname': 'tn'},
                        'walkHeight': {'args': '<type __builtin__.float>',
                                       'longname': 'walkHeight',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'wh'},
                        'walkSensitivity': {'args': '<type __builtin__.float>',
                                            'longname': 'walkSensitivity',
                                            'modes': ['create',
                                                      'query',
                                                      'edit'],
                                            'numArgs': 1,
                                            'shortname': 'wsv'},
                        'walkSpeed': {'args': '<type __builtin__.float>',
                                      'longname': 'walkSpeed',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'ws'},
                        'walkToolHud': {'args': '<type __builtin__.bool>',
                                        'longname': 'walkToolHud',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'wth'}},
              'shortFlags': {'ac': 'alternateContext',
                             'ch': 'history',
                             'e': 'edit',
                             'ex': 'exists',
                             'i1': 'image1',
                             'i2': 'image2',
                             'i3': 'image3',
                             'n': 'name',
                             'q': 'query',
                             'tn': 'toolName',
                             'wcc': 'crouchCount',
                             'wh': 'walkHeight',
                             'ws': 'walkSpeed',
                             'wsv': 'walkSensitivity',
                             'wth': 'walkToolHud'},
              'type': 'context'},
  'warning': {'flags': {'noContext': {'args': '<type __builtin__.bool>',
                                      'longname': 'noContext',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'n'},
                        'showLineNumber': {'args': '<type __builtin__.bool>',
                                           'longname': 'showLineNumber',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'sl'}},
              'shortFlags': {'n': 'noContext', 'sl': 'showLineNumber'},
              'type': 'system'},
  'webBrowser': {'callbackFlags': ['visibleChangeCommand',
                                   'vcc',
                                   'dragCallback',
                                   'dgc',
                                   'command',
                                   'c',
                                   'dropCallback',
                                   'dpc'],
                 'flags': {'annotation': {'args': '<type __builtin__.str>',
                                          'longname': 'annotation',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'ann'},
                           'back': {'args': '<type __builtin__.bool>',
                                    'longname': 'back',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'bk'},
                           'backgroundColor': {'args': ['<type __builtin__.float>',
                                                        '<type __builtin__.float>',
                                                        '<type __builtin__.float>'],
                                               'longname': 'backgroundColor',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 3,
                                               'shortname': 'bgc'},
                           'command': {'args': '<type __builtin__.str>',
                                       'longname': 'command',
                                       'numArgs': 1,
                                       'shortname': 'c'},
                           'defineTemplate': {'args': '<type __builtin__.str>',
                                              'longname': 'defineTemplate',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'dt'},
                           'docTag': {'args': '<type __builtin__.str>',
                                      'longname': 'docTag',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'dtg'},
                           'dragCallback': {'args': 'script',
                                            'longname': 'dragCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dgc'},
                           'dropCallback': {'args': 'script',
                                            'longname': 'dropCallback',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'dpc'},
                           'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'enable': {'args': '<type __builtin__.bool>',
                                      'longname': 'enable',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'en'},
                           'enableBackground': {'args': '<type __builtin__.bool>',
                                                'longname': 'enableBackground',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'ebg'},
                           'enableKeyboardFocus': {'args': '<type __builtin__.bool>',
                                                   'longname': 'enableKeyboardFocus',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ekf'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'find': {'args': '<type __builtin__.str>',
                                    'longname': 'find',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'f'},
                           'forward': {'args': '<type __builtin__.bool>',
                                       'longname': 'forward',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'fwd'},
                           'fullPathName': {'args': '<type __builtin__.bool>',
                                            'longname': 'fullPathName',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'fpn'},
                           'height': {'args': '<type __builtin__.int>',
                                      'longname': 'height',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                           'highlightColor': {'args': ['<type __builtin__.float>',
                                                       '<type __builtin__.float>',
                                                       '<type __builtin__.float>'],
                                              'longname': 'highlightColor',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 3,
                                              'shortname': 'hlc'},
                           'home': {'args': '<type __builtin__.bool>',
                                    'longname': 'home',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'hm'},
                           'isObscured': {'args': '<type __builtin__.bool>',
                                          'longname': 'isObscured',
                                          'modes': ['query'],
                                          'numArgs': 0,
                                          'shortname': 'io'},
                           'manage': {'args': '<type __builtin__.bool>',
                                      'longname': 'manage',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'm'},
                           'matchCase': {'args': '<type __builtin__.bool>',
                                         'longname': 'matchCase',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'mc'},
                           'matchWholeWord': {'args': '<type __builtin__.bool>',
                                              'longname': 'matchWholeWord',
                                              'numArgs': 1,
                                              'shortname': 'mww'},
                           'matchWholeWorld': {'args': '<type __builtin__.bool>',
                                               'longname': 'matchWholeWorld',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'mww'},
                           'noBackground': {'args': '<type __builtin__.bool>',
                                            'longname': 'noBackground',
                                            'modes': ['create', 'edit'],
                                            'numArgs': 1,
                                            'shortname': 'nbg'},
                           'numberOfPopupMenus': {'args': '<type __builtin__.bool>',
                                                  'longname': 'numberOfPopupMenus',
                                                  'modes': ['query'],
                                                  'numArgs': 0,
                                                  'shortname': 'npm'},
                           'openURL': {'args': '<type __builtin__.str>',
                                       'longname': 'openURL',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'url'},
                           'parent': {'args': '<type __builtin__.str>',
                                      'longname': 'parent',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'shortname': 'p'},
                           'popupMenuArray': {'args': '<type __builtin__.bool>',
                                              'longname': 'popupMenuArray',
                                              'modes': ['query'],
                                              'numArgs': 0,
                                              'shortname': 'pma'},
                           'preventOverride': {'args': '<type __builtin__.bool>',
                                               'longname': 'preventOverride',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'po'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'reload': {'args': '<type __builtin__.bool>',
                                      'longname': 'reload',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'rl'},
                           'searchForward': {'args': '<type __builtin__.bool>',
                                             'longname': 'searchForward',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'sf'},
                           'statusBarMessage': {'args': '<type __builtin__.str>',
                                                'longname': 'statusBarMessage',
                                                'modes': ['create',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'sbm'},
                           'stop': {'args': '<type __builtin__.bool>',
                                    'longname': 'stop',
                                    'modes': ['create'],
                                    'numArgs': 0,
                                    'shortname': 'st'},
                           'urlChangedCb': {'args': '<type __builtin__.str>',
                                            'longname': 'urlChangedCb',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'ucc'},
                           'useTemplate': {'args': '<type __builtin__.str>',
                                           'longname': 'useTemplate',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ut'},
                           'visible': {'args': '<type __builtin__.bool>',
                                       'longname': 'visible',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'vis'},
                           'visibleChangeCommand': {'args': 'script',
                                                    'longname': 'visibleChangeCommand',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'vcc'},
                           'width': {'args': '<type __builtin__.int>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'},
                           'wrap': {'args': '<type __builtin__.bool>',
                                    'longname': 'wrap',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'wr'}},
                 'shortFlags': {'ann': 'annotation',
                                'bgc': 'backgroundColor',
                                'bk': 'back',
                                'c': 'command',
                                'dgc': 'dragCallback',
                                'dpc': 'dropCallback',
                                'dt': 'defineTemplate',
                                'dtg': 'docTag',
                                'e': 'edit',
                                'ebg': 'enableBackground',
                                'ekf': 'enableKeyboardFocus',
                                'en': 'enable',
                                'ex': 'exists',
                                'f': 'find',
                                'fpn': 'fullPathName',
                                'fwd': 'forward',
                                'h': 'height',
                                'hlc': 'highlightColor',
                                'hm': 'home',
                                'io': 'isObscured',
                                'm': 'manage',
                                'mc': 'matchCase',
                                'mww': 'matchWholeWord',
                                'nbg': 'noBackground',
                                'npm': 'numberOfPopupMenus',
                                'p': 'parent',
                                'pma': 'popupMenuArray',
                                'po': 'preventOverride',
                                'q': 'query',
                                'rl': 'reload',
                                'sbm': 'statusBarMessage',
                                'sf': 'searchForward',
                                'st': 'stop',
                                'ucc': 'urlChangedCb',
                                'url': 'openURL',
                                'ut': 'useTemplate',
                                'vcc': 'visibleChangeCommand',
                                'vis': 'visible',
                                'w': 'width',
                                'wr': 'wrap'},
                 'type': 'windows'},
  'webBrowserPrefs': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'preference': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'preference',
                                               'numArgs': 2,
                                               'shortname': 'prf'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'}},
                      'shortFlags': {'e': 'edit',
                                     'prf': 'preference',
                                     'q': 'query'},
                      'type': 'windows'},
  'webView': {'flags': {'urlAddress': {'args': '<type __builtin__.str>',
                                       'longname': 'urlAddress',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'url'},
                        'windowHeight': {'args': '<type __builtin__.int>',
                                         'longname': 'windowHeight',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'wh'},
                        'windowWidth': {'args': '<type __builtin__.int>',
                                        'longname': 'windowWidth',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ww'}},
              'shortFlags': {'url': 'urlAddress',
                             'wh': 'windowHeight',
                             'ww': 'windowWidth'},
              'type': 'general'},
  'webViewCmd': {'flags': {}, 'shortFlags': {}, 'type': 'other'},
  'whatsNewHighlight': {'flags': {'highlightColor': {'args': ['<type __builtin__.float>',
                                                              '<type __builtin__.float>',
                                                              '<type __builtin__.float>'],
                                                     'longname': 'highlightColor',
                                                     'modes': ['create',
                                                               'query'],
                                                     'numArgs': 3,
                                                     'shortname': 'hc'},
                                  'highlightOn': {'args': '<type __builtin__.bool>',
                                                  'longname': 'highlightOn',
                                                  'modes': ['create',
                                                            'query'],
                                                  'numArgs': 1,
                                                  'shortname': 'ho'},
                                  'query': {'args': '<type __builtin__.bool>',
                                            'longname': 'query',
                                            'numArgs': 0,
                                            'shortname': 'q'},
                                  'showStartupDialog': {'args': '<type __builtin__.bool>',
                                                        'longname': 'showStartupDialog',
                                                        'modes': ['create',
                                                                  'query'],
                                                        'numArgs': 1,
                                                        'shortname': 'ssd'}},
                        'shortFlags': {'hc': 'highlightColor',
                                       'ho': 'highlightOn',
                                       'q': 'query',
                                       'ssd': 'showStartupDialog'},
                        'type': 'system'},
  'window': {'callbackFlags': ['closeCommand',
                               'cc',
                               'restoreCommand',
                               'rc',
                               'minimizeCommand',
                               'mnc'],
             'flags': {'backgroundColor': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'backgroundColor',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 3,
                                           'shortname': 'bgc'},
                       'closeCommand': {'args': 'script',
                                        'longname': 'closeCommand',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 1,
                                        'shortname': 'cc'},
                       'defineTemplate': {'args': '<type __builtin__.str>',
                                          'longname': 'defineTemplate',
                                          'modes': ['create'],
                                          'numArgs': 1,
                                          'shortname': 'dt'},
                       'docTag': {'args': '<type __builtin__.str>',
                                  'longname': 'docTag',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'dtg'},
                       'dockCorner': {'args': ['<type __builtin__.str>',
                                               '<type __builtin__.str>'],
                                      'longname': 'dockCorner',
                                      'modes': ['create', 'multiuse'],
                                      'numArgs': 2,
                                      'shortname': 'dc'},
                       'dockStation': {'args': '<type __builtin__.bool>',
                                       'longname': 'dockStation',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ds'},
                       'dockingLayout': {'args': '<type __builtin__.str>',
                                         'longname': 'dockingLayout',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'dl'},
                       'edit': {'args': '<type __builtin__.bool>',
                                'longname': 'edit',
                                'numArgs': 0,
                                'shortname': 'e'},
                       'exists': {'args': '<type __builtin__.bool>',
                                  'longname': 'exists',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'ex'},
                       'frontWindow': {'args': '<type __builtin__.bool>',
                                       'longname': 'frontWindow',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'fw'},
                       'height': {'args': '<type __builtin__.int>',
                                  'longname': 'height',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'h'},
                       'iconName': {'args': '<type __builtin__.str>',
                                    'longname': 'iconName',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'iconName'},
                       'iconify': {'args': '<type __builtin__.bool>',
                                   'longname': 'iconify',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'i'},
                       'interactivePlacement': {'args': '<type __builtin__.bool>',
                                                'longname': 'interactivePlacement',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'ip'},
                       'leftEdge': {'args': '<type __builtin__.int>',
                                    'longname': 'leftEdge',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'le'},
                       'mainMenuBar': {'args': '<type __builtin__.bool>',
                                       'longname': 'mainMenuBar',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'mm'},
                       'mainWindow': {'args': '<type __builtin__.bool>',
                                      'longname': 'mainWindow',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'mw'},
                       'maximizeButton': {'args': '<type __builtin__.bool>',
                                          'longname': 'maximizeButton',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'mxb'},
                       'menuArray': {'args': '<type __builtin__.bool>',
                                     'longname': 'menuArray',
                                     'modes': ['query'],
                                     'numArgs': 0,
                                     'shortname': 'ma'},
                       'menuBar': {'args': '<type __builtin__.bool>',
                                   'longname': 'menuBar',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'mb'},
                       'menuBarCornerWidget': {'args': ['<type __builtin__.str>',
                                                        '<type __builtin__.str>'],
                                               'longname': 'menuBarCornerWidget',
                                               'modes': ['query', 'edit'],
                                               'numArgs': 2,
                                               'shortname': 'mcw'},
                       'menuBarResize': {'args': '<type __builtin__.bool>',
                                         'longname': 'menuBarResize',
                                         'modes': ['edit'],
                                         'numArgs': 0,
                                         'shortname': 'mbr'},
                       'menuBarVisible': {'args': '<type __builtin__.bool>',
                                          'longname': 'menuBarVisible',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'mbv'},
                       'menuIndex': {'args': ['<type __builtin__.str>',
                                              '<type __builtin__.int>'],
                                     'longname': 'menuIndex',
                                     'modes': ['edit'],
                                     'numArgs': 2,
                                     'shortname': 'mi'},
                       'minimizeButton': {'args': '<type __builtin__.bool>',
                                          'longname': 'minimizeButton',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'mnb'},
                       'minimizeCommand': {'args': 'script',
                                           'longname': 'minimizeCommand',
                                           'modes': ['create', 'edit'],
                                           'numArgs': 1,
                                           'shortname': 'mnc'},
                       'nestedDockingEnabled': {'args': '<type __builtin__.bool>',
                                                'longname': 'nestedDockingEnabled',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 'nde'},
                       'numberOfMenus': {'args': '<type __builtin__.bool>',
                                         'longname': 'numberOfMenus',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'nm'},
                       'parent': {'args': '<type __builtin__.str>',
                                  'longname': 'parent',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'p'},
                       'query': {'args': '<type __builtin__.bool>',
                                 'longname': 'query',
                                 'numArgs': 0,
                                 'shortname': 'q'},
                       'resizeToFitChildren': {'args': '<type __builtin__.bool>',
                                               'longname': 'resizeToFitChildren',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'rtf'},
                       'restoreCommand': {'args': 'script',
                                          'longname': 'restoreCommand',
                                          'modes': ['create', 'edit'],
                                          'numArgs': 1,
                                          'shortname': 'rc'},
                       'retain': {'args': '<type __builtin__.bool>',
                                  'longname': 'retain',
                                  'modes': ['create'],
                                  'numArgs': 0,
                                  'shortname': 'ret'},
                       'sizeable': {'args': '<type __builtin__.bool>',
                                    'longname': 'sizeable',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 's'},
                       'state': {'args': '<type __builtin__.str>',
                                 'longname': 'state',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'st'},
                       'title': {'args': '<type __builtin__.str>',
                                 'longname': 'title',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 't'},
                       'titleBar': {'args': '<type __builtin__.bool>',
                                    'longname': 'titleBar',
                                    'modes': ['create', 'query', 'edit'],
                                    'numArgs': 1,
                                    'shortname': 'tb'},
                       'titleBarMenu': {'args': '<type __builtin__.bool>',
                                        'longname': 'titleBarMenu',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'tbm'},
                       'toolbox': {'args': '<type __builtin__.bool>',
                                   'longname': 'toolbox',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'tlb'},
                       'topEdge': {'args': '<type __builtin__.int>',
                                   'longname': 'topEdge',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'te'},
                       'topLeftCorner': {'args': ['<type __builtin__.int>',
                                                  '<type __builtin__.int>'],
                                         'longname': 'topLeftCorner',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 2,
                                         'shortname': 'tlc'},
                       'useTemplate': {'args': '<type __builtin__.str>',
                                       'longname': 'useTemplate',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'ut'},
                       'visible': {'args': '<type __builtin__.bool>',
                                   'longname': 'visible',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'vis'},
                       'width': {'args': '<type __builtin__.int>',
                                 'longname': 'width',
                                 'modes': ['create', 'query', 'edit'],
                                 'numArgs': 1,
                                 'shortname': 'w'},
                       'widthHeight': {'args': ['<type __builtin__.int>',
                                                '<type __builtin__.int>'],
                                       'longname': 'widthHeight',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 2,
                                       'shortname': 'wh'}},
             'removedFlags': {'in': 'iconName'},
             'shortFlags': {'bgc': 'backgroundColor',
                            'cc': 'closeCommand',
                            'dc': 'dockCorner',
                            'dl': 'dockingLayout',
                            'ds': 'dockStation',
                            'dt': 'defineTemplate',
                            'dtg': 'docTag',
                            'e': 'edit',
                            'ex': 'exists',
                            'fw': 'frontWindow',
                            'h': 'height',
                            'i': 'iconify',
                            'iconName': 'iconName',
                            'ip': 'interactivePlacement',
                            'le': 'leftEdge',
                            'ma': 'menuArray',
                            'mb': 'menuBar',
                            'mbr': 'menuBarResize',
                            'mbv': 'menuBarVisible',
                            'mcw': 'menuBarCornerWidget',
                            'mi': 'menuIndex',
                            'mm': 'mainMenuBar',
                            'mnb': 'minimizeButton',
                            'mnc': 'minimizeCommand',
                            'mw': 'mainWindow',
                            'mxb': 'maximizeButton',
                            'nde': 'nestedDockingEnabled',
                            'nm': 'numberOfMenus',
                            'p': 'parent',
                            'q': 'query',
                            'rc': 'restoreCommand',
                            'ret': 'retain',
                            'rtf': 'resizeToFitChildren',
                            's': 'sizeable',
                            'st': 'state',
                            't': 'title',
                            'tb': 'titleBar',
                            'tbm': 'titleBarMenu',
                            'te': 'topEdge',
                            'tlb': 'toolbox',
                            'tlc': 'topLeftCorner',
                            'ut': 'useTemplate',
                            'vis': 'visible',
                            'w': 'width',
                            'wh': 'widthHeight'},
             'type': 'windows'},
  'windowPref': {'flags': {'edit': {'args': '<type __builtin__.bool>',
                                    'longname': 'edit',
                                    'numArgs': 0,
                                    'shortname': 'e'},
                           'enableAll': {'args': '<type __builtin__.bool>',
                                         'longname': 'enableAll',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'ea'},
                           'exists': {'args': '<type __builtin__.bool>',
                                      'longname': 'exists',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'ex'},
                           'height': {'args': '<type __builtin__.int>',
                                      'longname': 'height',
                                      'modes': ['create', 'query', 'edit'],
                                      'numArgs': 1,
                                      'shortname': 'h'},
                           'leftEdge': {'args': '<type __builtin__.int>',
                                        'longname': 'leftEdge',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'le'},
                           'loadAll': {'args': '<type __builtin__.bool>',
                                       'longname': 'loadAll',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'la'},
                           'maximized': {'args': '<type __builtin__.bool>',
                                         'longname': 'maximized',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'max'},
                           'parentMain': {'args': '<type __builtin__.bool>',
                                          'longname': 'parentMain',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'pm'},
                           'query': {'args': '<type __builtin__.bool>',
                                     'longname': 'query',
                                     'numArgs': 0,
                                     'shortname': 'q'},
                           'remove': {'args': '<type __builtin__.bool>',
                                      'longname': 'remove',
                                      'modes': ['create'],
                                      'numArgs': 0,
                                      'shortname': 'r'},
                           'removeAll': {'args': '<type __builtin__.bool>',
                                         'longname': 'removeAll',
                                         'modes': ['create'],
                                         'numArgs': 0,
                                         'shortname': 'ra'},
                           'restoreMainWindowState': {'args': '<type __builtin__.str>',
                                                      'longname': 'restoreMainWindowState',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'rms'},
                           'saveAll': {'args': '<type __builtin__.bool>',
                                       'longname': 'saveAll',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'sa'},
                           'saveMainWindowState': {'args': '<type __builtin__.str>',
                                                   'longname': 'saveMainWindowState',
                                                   'modes': ['create'],
                                                   'numArgs': 1,
                                                   'shortname': 'sms'},
                           'topEdge': {'args': '<type __builtin__.int>',
                                       'longname': 'topEdge',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'te'},
                           'topLeftCorner': {'args': ['<type __builtin__.int>',
                                                      '<type __builtin__.int>'],
                                             'longname': 'topLeftCorner',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 2,
                                             'shortname': 'tlc'},
                           'width': {'args': '<type __builtin__.int>',
                                     'longname': 'width',
                                     'modes': ['create', 'query', 'edit'],
                                     'numArgs': 1,
                                     'shortname': 'w'},
                           'widthHeight': {'args': ['<type __builtin__.int>',
                                                    '<type __builtin__.int>'],
                                           'longname': 'widthHeight',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 2,
                                           'shortname': 'wh'}},
                 'shortFlags': {'e': 'edit',
                                'ea': 'enableAll',
                                'ex': 'exists',
                                'h': 'height',
                                'la': 'loadAll',
                                'le': 'leftEdge',
                                'max': 'maximized',
                                'pm': 'parentMain',
                                'q': 'query',
                                'r': 'remove',
                                'ra': 'removeAll',
                                'rms': 'restoreMainWindowState',
                                'sa': 'saveAll',
                                'sms': 'saveMainWindowState',
                                'te': 'topEdge',
                                'tlc': 'topLeftCorner',
                                'w': 'width',
                                'wh': 'widthHeight'},
                 'type': 'windows'},
  'wire': {'flags': {'after': {'args': '<type __builtin__.bool>',
                               'longname': 'after',
                               'modes': ['create', 'edit'],
                               'numArgs': 0,
                               'shortname': 'af'},
                     'afterReference': {'args': '<type __builtin__.bool>',
                                        'longname': 'afterReference',
                                        'modes': ['create', 'edit'],
                                        'numArgs': 0,
                                        'shortname': 'ar'},
                     'before': {'args': '<type __builtin__.bool>',
                                'longname': 'before',
                                'modes': ['create', 'edit'],
                                'numArgs': 0,
                                'shortname': 'bf'},
                     'crossingEffect': {'args': '<type __builtin__.float>',
                                        'longname': 'crossingEffect',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'ce'},
                     'deformerTools': {'args': '<type __builtin__.bool>',
                                       'longname': 'deformerTools',
                                       'modes': ['query'],
                                       'numArgs': 0,
                                       'shortname': 'dt'},
                     'dropoffDistance': {'args': ['<type __builtin__.int>',
                                                  '<type __builtin__.float>'],
                                         'longname': 'dropoffDistance',
                                         'modes': ['create',
                                                   'query',
                                                   'edit',
                                                   'multiuse'],
                                         'numArgs': 2,
                                         'shortname': 'dds'},
                     'edit': {'args': '<type __builtin__.bool>',
                              'longname': 'edit',
                              'numArgs': 0,
                              'shortname': 'e'},
                     'envelope': {'args': '<type __builtin__.float>',
                                  'longname': 'envelope',
                                  'modes': ['create', 'query', 'edit'],
                                  'numArgs': 1,
                                  'shortname': 'en'},
                     'exclusive': {'args': '<type __builtin__.str>',
                                   'longname': 'exclusive',
                                   'modes': ['create', 'query'],
                                   'numArgs': 1,
                                   'shortname': 'ex'},
                     'frontOfChain': {'args': '<type __builtin__.bool>',
                                      'longname': 'frontOfChain',
                                      'modes': ['create', 'edit'],
                                      'numArgs': 0,
                                      'shortname': 'foc'},
                     'geometry': {'args': '<type __builtin__.str>',
                                  'longname': 'geometry',
                                  'modes': ['query', 'edit', 'multiuse'],
                                  'numArgs': 1,
                                  'shortname': 'g'},
                     'geometryIndices': {'args': '<type __builtin__.bool>',
                                         'longname': 'geometryIndices',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'gi'},
                     'groupWithBase': {'args': '<type __builtin__.bool>',
                                       'longname': 'groupWithBase',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'gw'},
                     'holder': {'args': ['<type __builtin__.int>',
                                         '<type __builtin__.str>'],
                                'longname': 'holder',
                                'modes': ['create',
                                          'query',
                                          'edit',
                                          'multiuse'],
                                'numArgs': 2,
                                'shortname': 'ho'},
                     'ignoreSelected': {'args': '<type __builtin__.bool>',
                                        'longname': 'ignoreSelected',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ignoreSelected'},
                     'includeHiddenSelections': {'args': '<type __builtin__.bool>',
                                                 'longname': 'includeHiddenSelections',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ihs'},
                     'localInfluence': {'args': '<type __builtin__.float>',
                                        'longname': 'localInfluence',
                                        'modes': ['create',
                                                  'query',
                                                  'edit'],
                                        'numArgs': 1,
                                        'shortname': 'li'},
                     'name': {'args': '<type __builtin__.str>',
                              'longname': 'name',
                              'modes': ['create'],
                              'numArgs': 1,
                              'shortname': 'n'},
                     'parallel': {'args': '<type __builtin__.bool>',
                                  'longname': 'parallel',
                                  'modes': ['create', 'edit'],
                                  'numArgs': 0,
                                  'shortname': 'par'},
                     'prune': {'args': '<type __builtin__.bool>',
                               'longname': 'prune',
                               'modes': ['edit'],
                               'numArgs': 0,
                               'shortname': 'pr'},
                     'query': {'args': '<type __builtin__.bool>',
                               'longname': 'query',
                               'numArgs': 0,
                               'shortname': 'q'},
                     'remove': {'args': '<type __builtin__.bool>',
                                'longname': 'remove',
                                'modes': ['edit', 'multiuse'],
                                'numArgs': 0,
                                'shortname': 'rm'},
                     'split': {'args': '<type __builtin__.bool>',
                               'longname': 'split',
                               'modes': ['create', 'edit'],
                               'numArgs': 0,
                               'shortname': 'sp'},
                     'wire': {'args': '<type __builtin__.str>',
                              'longname': 'wire',
                              'modes': ['create',
                                        'query',
                                        'edit',
                                        'multiuse'],
                              'numArgs': 1,
                              'shortname': 'w'},
                     'wireCount': {'args': '<type __builtin__.int>',
                                   'longname': 'wireCount',
                                   'modes': ['create', 'query', 'edit'],
                                   'numArgs': 1,
                                   'shortname': 'wc'}},
           'removedFlags': {'is': 'ignoreSelected'},
           'shortFlags': {'af': 'after',
                          'ar': 'afterReference',
                          'bf': 'before',
                          'ce': 'crossingEffect',
                          'dds': 'dropoffDistance',
                          'dt': 'deformerTools',
                          'e': 'edit',
                          'en': 'envelope',
                          'ex': 'exclusive',
                          'foc': 'frontOfChain',
                          'g': 'geometry',
                          'gi': 'geometryIndices',
                          'gw': 'groupWithBase',
                          'ho': 'holder',
                          'ignoreSelected': 'ignoreSelected',
                          'ihs': 'includeHiddenSelections',
                          'li': 'localInfluence',
                          'n': 'name',
                          'par': 'parallel',
                          'pr': 'prune',
                          'q': 'query',
                          'rm': 'remove',
                          'sp': 'split',
                          'w': 'wire',
                          'wc': 'wireCount'},
           'type': 'animation'},
  'wireContext': {'flags': {'crossingEffect': {'args': '<type __builtin__.float>',
                                               'longname': 'crossingEffect',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'ce'},
                            'deformationOrder': {'args': '<type __builtin__.str>',
                                                 'longname': 'deformationOrder',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'do'},
                            'dropoffDistance': {'args': '<type __builtin__.float>',
                                                'longname': 'dropoffDistance',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'dds'},
                            'edit': {'args': '<type __builtin__.bool>',
                                     'longname': 'edit',
                                     'numArgs': 0,
                                     'shortname': 'e'},
                            'envelope': {'args': '<type __builtin__.float>',
                                         'longname': 'envelope',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'en'},
                            'exclusive': {'args': '<type __builtin__.bool>',
                                          'longname': 'exclusive',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'exc'},
                            'exclusivePartition': {'args': '<type __builtin__.str>',
                                                   'longname': 'exclusivePartition',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ep'},
                            'exists': {'args': '<type __builtin__.bool>',
                                       'longname': 'exists',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'shortname': 'ex'},
                            'groupWithBase': {'args': '<type __builtin__.bool>',
                                              'longname': 'groupWithBase',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'gw'},
                            'history': {'args': '<type __builtin__.bool>',
                                        'longname': 'history',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'ch'},
                            'holder': {'args': '<type __builtin__.bool>',
                                       'longname': 'holder',
                                       'modes': ['create', 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'ho'},
                            'image1': {'args': '<type __builtin__.str>',
                                       'longname': 'image1',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i1'},
                            'image2': {'args': '<type __builtin__.str>',
                                       'longname': 'image2',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i2'},
                            'image3': {'args': '<type __builtin__.str>',
                                       'longname': 'image3',
                                       'modes': ['create',
                                                 'query',
                                                 'edit'],
                                       'numArgs': 1,
                                       'shortname': 'i3'},
                            'localInfluence': {'args': '<type __builtin__.float>',
                                               'longname': 'localInfluence',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'li'},
                            'name': {'args': '<type __builtin__.str>',
                                     'longname': 'name',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'n'},
                            'query': {'args': '<type __builtin__.bool>',
                                      'longname': 'query',
                                      'numArgs': 0,
                                      'shortname': 'q'}},
                  'shortFlags': {'ce': 'crossingEffect',
                                 'ch': 'history',
                                 'dds': 'dropoffDistance',
                                 'do': 'deformationOrder',
                                 'e': 'edit',
                                 'en': 'envelope',
                                 'ep': 'exclusivePartition',
                                 'ex': 'exists',
                                 'exc': 'exclusive',
                                 'gw': 'groupWithBase',
                                 'ho': 'holder',
                                 'i1': 'image1',
                                 'i2': 'image2',
                                 'i3': 'image3',
                                 'li': 'localInfluence',
                                 'n': 'name',
                                 'q': 'query'},
                  'type': 'context'},
  'workspace': {'flags': {'active': {'args': '<type __builtin__.bool>',
                                     'longname': 'active',
                                     'modes': ['create', 'query'],
                                     'numArgs': 0,
                                     'shortname': 'act'},
                          'baseWorkspace': {'args': '<type __builtin__.str>',
                                            'longname': 'baseWorkspace',
                                            'modes': ['query'],
                                            'numArgs': 1,
                                            'shortname': 'bw'},
                          'create': {'args': '<type __builtin__.str>',
                                     'longname': 'create',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'cr'},
                          'directory': {'args': '<type __builtin__.str>',
                                        'longname': 'directory',
                                        'modes': ['create', 'query'],
                                        'numArgs': 1,
                                        'shortname': 'dir'},
                          'expandName': {'args': '<type __builtin__.str>',
                                         'longname': 'expandName',
                                         'modes': ['create', 'query'],
                                         'numArgs': 1,
                                         'shortname': 'en'},
                          'fileRule': {'args': ['<type __builtin__.str>',
                                                '<type __builtin__.str>'],
                                       'longname': 'fileRule',
                                       'modes': ['create', 'query'],
                                       'numArgs': 2,
                                       'shortname': 'fr'},
                          'fileRuleEntry': {'args': '<type __builtin__.str>',
                                            'longname': 'fileRuleEntry',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 'fre'},
                          'fileRuleList': {'args': '<type __builtin__.bool>',
                                           'longname': 'fileRuleList',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'frl'},
                          'filter': {'args': '<type __builtin__.bool>',
                                     'longname': 'filter',
                                     'modes': [],
                                     'numArgs': 0,
                                     'shortname': 'f'},
                          'fullName': {'args': '<type __builtin__.bool>',
                                       'longname': 'fullName',
                                       'modes': ['create', 'query'],
                                       'numArgs': 0,
                                       'shortname': 'fn'},
                          'list': {'args': '<type __builtin__.bool>',
                                   'longname': 'list',
                                   'modes': ['create', 'query'],
                                   'numArgs': 0,
                                   'shortname': 'l'},
                          'listFullWorkspaces': {'args': '<type __builtin__.bool>',
                                                 'longname': 'listFullWorkspaces',
                                                 'modes': ['create',
                                                           'query'],
                                                 'numArgs': 0,
                                                 'shortname': 'lfw'},
                          'listWorkspaces': {'args': '<type __builtin__.bool>',
                                             'longname': 'listWorkspaces',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'lw'},
                          'newWorkspace': {'args': '<type __builtin__.bool>',
                                           'longname': 'newWorkspace',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'n'},
                          'objectType': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'objectType',
                                         'modes': ['create', 'query'],
                                         'numArgs': 2,
                                         'shortname': 'ot'},
                          'objectTypeEntry': {'args': '<type __builtin__.str>',
                                              'longname': 'objectTypeEntry',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'ote'},
                          'objectTypeList': {'args': '<type __builtin__.bool>',
                                             'longname': 'objectTypeList',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'otl'},
                          'openWorkspace': {'args': '<type __builtin__.bool>',
                                            'longname': 'openWorkspace',
                                            'modes': ['create', 'query'],
                                            'numArgs': 0,
                                            'shortname': 'o'},
                          'projectPath': {'args': '<type __builtin__.str>',
                                          'longname': 'projectPath',
                                          'modes': ['create', 'query'],
                                          'numArgs': 1,
                                          'shortname': 'pp'},
                          'query': {'args': '<type __builtin__.bool>',
                                    'longname': 'query',
                                    'numArgs': 0,
                                    'shortname': 'q'},
                          'removeFileRuleEntry': {'args': '<type __builtin__.str>',
                                                  'longname': 'removeFileRuleEntry',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'rfr'},
                          'removeVariableEntry': {'args': '<type __builtin__.str>',
                                                  'longname': 'removeVariableEntry',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'rve'},
                          'renderType': {'args': ['<type __builtin__.str>',
                                                  '<type __builtin__.str>'],
                                         'longname': 'renderType',
                                         'modes': ['create', 'query'],
                                         'numArgs': 2,
                                         'shortname': 'rt'},
                          'renderTypeEntry': {'args': '<type __builtin__.str>',
                                              'longname': 'renderTypeEntry',
                                              'modes': ['create', 'query'],
                                              'numArgs': 1,
                                              'shortname': 'rte'},
                          'renderTypeList': {'args': '<type __builtin__.bool>',
                                             'longname': 'renderTypeList',
                                             'modes': ['create', 'query'],
                                             'numArgs': 0,
                                             'shortname': 'rtl'},
                          'rootDirectory': {'args': '<type __builtin__.bool>',
                                            'longname': 'rootDirectory',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'rd'},
                          'saveWorkspace': {'args': '<type __builtin__.bool>',
                                            'longname': 'saveWorkspace',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 's'},
                          'shortName': {'args': '<type __builtin__.bool>',
                                        'longname': 'shortName',
                                        'modes': ['create', 'query'],
                                        'numArgs': 0,
                                        'shortname': 'sn'},
                          'update': {'args': '<type __builtin__.bool>',
                                     'longname': 'update',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'u'},
                          'updateAll': {'args': '<type __builtin__.bool>',
                                        'longname': 'updateAll',
                                        'modes': ['create'],
                                        'numArgs': 0,
                                        'shortname': 'ua'},
                          'variable': {'args': ['<type __builtin__.str>',
                                                '<type __builtin__.str>'],
                                       'longname': 'variable',
                                       'modes': ['create', 'query'],
                                       'numArgs': 2,
                                       'shortname': 'v'},
                          'variableEntry': {'args': '<type __builtin__.str>',
                                            'longname': 'variableEntry',
                                            'modes': ['create', 'query'],
                                            'numArgs': 1,
                                            'shortname': 've'},
                          'variableList': {'args': '<type __builtin__.bool>',
                                           'longname': 'variableList',
                                           'modes': ['create', 'query'],
                                           'numArgs': 0,
                                           'shortname': 'vl'}},
                'shortFlags': {'act': 'active',
                               'bw': 'baseWorkspace',
                               'cr': 'create',
                               'dir': 'directory',
                               'en': 'expandName',
                               'f': 'filter',
                               'fn': 'fullName',
                               'fr': 'fileRule',
                               'fre': 'fileRuleEntry',
                               'frl': 'fileRuleList',
                               'l': 'list',
                               'lfw': 'listFullWorkspaces',
                               'lw': 'listWorkspaces',
                               'n': 'newWorkspace',
                               'o': 'openWorkspace',
                               'ot': 'objectType',
                               'ote': 'objectTypeEntry',
                               'otl': 'objectTypeList',
                               'pp': 'projectPath',
                               'q': 'query',
                               'rd': 'rootDirectory',
                               'rfr': 'removeFileRuleEntry',
                               'rt': 'renderType',
                               'rte': 'renderTypeEntry',
                               'rtl': 'renderTypeList',
                               'rve': 'removeVariableEntry',
                               's': 'saveWorkspace',
                               'sn': 'shortName',
                               'u': 'update',
                               'ua': 'updateAll',
                               'v': 'variable',
                               've': 'variableEntry',
                               'vl': 'variableList'},
                'type': 'system'},
  'workspaceControl': {'callbackFlags': ['uiScript',
                                         'ui',
                                         'closeCommand',
                                         'cc',
                                         'visibleChangeCommand',
                                         'vcc'],
                       'flags': {'checksPlugins': {'args': '<type __builtin__.bool>',
                                                   'longname': 'checksPlugins',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'cp'},
                                 'close': {'args': '<type __builtin__.bool>',
                                           'longname': 'close',
                                           'modes': ['edit'],
                                           'numArgs': 0,
                                           'shortname': 'cl'},
                                 'closeCommand': {'args': 'script',
                                                  'longname': 'closeCommand',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'cc'},
                                 'collapse': {'args': '<type __builtin__.bool>',
                                              'longname': 'collapse',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'clp'},
                                 'defineTemplate': {'args': '<type __builtin__.str>',
                                                    'longname': 'defineTemplate',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'dt'},
                                 'deleteLater': {'args': '<type __builtin__.bool>',
                                                 'longname': 'deleteLater',
                                                 'numArgs': 1,
                                                 'shortname': 'deleteLater'},
                                 'dockToControl': {'args': ['<type __builtin__.str>',
                                                            '<type __builtin__.str>'],
                                                   'longname': 'dockToControl',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 2,
                                                   'shortname': 'dtc'},
                                 'dockToMainWindow': {'args': ['<type __builtin__.str>',
                                                               '<type __builtin__.bool>'],
                                                      'longname': 'dockToMainWindow',
                                                      'modes': ['create',
                                                                'edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'dtm'},
                                 'dockToPanel': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.str>',
                                                          '<type __builtin__.bool>'],
                                                 'longname': 'dockToPanel',
                                                 'modes': ['create',
                                                           'edit'],
                                                 'numArgs': 3,
                                                 'shortname': 'dtp'},
                                 'duplicatable': {'args': '<type __builtin__.bool>',
                                                  'longname': 'duplicatable',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'dup'},
                                 'edit': {'args': '<type __builtin__.bool>',
                                          'longname': 'edit',
                                          'numArgs': 0,
                                          'shortname': 'e'},
                                 'exists': {'args': '<type __builtin__.bool>',
                                            'longname': 'exists',
                                            'modes': ['create'],
                                            'numArgs': 0,
                                            'shortname': 'ex'},
                                 'floating': {'args': '<type __builtin__.bool>',
                                              'longname': 'floating',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'fl'},
                                 'height': {'args': '<type __builtin__.bool>',
                                            'longname': 'height',
                                            'modes': ['query'],
                                            'numArgs': 0,
                                            'shortname': 'h'},
                                 'heightProperty': {'args': '<type __builtin__.str>',
                                                    'longname': 'heightProperty',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'hp'},
                                 'horizontal': {'args': '<type __builtin__.bool>',
                                                'longname': 'horizontal',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'hr'},
                                 'initCallback': {'args': '<type __builtin__.str>',
                                                  'longname': 'initCallback',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'ic'},
                                 'initialHeight': {'args': '<type __builtin__.int>',
                                                   'longname': 'initialHeight',
                                                   'modes': ['create',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'ih'},
                                 'initialWidth': {'args': '<type __builtin__.int>',
                                                  'longname': 'initialWidth',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'iw'},
                                 'label': {'args': '<type __builtin__.str>',
                                           'longname': 'label',
                                           'modes': ['create',
                                                     'query',
                                                     'edit'],
                                           'numArgs': 1,
                                           'shortname': 'l'},
                                 'loadImmediately': {'args': '<type __builtin__.bool>',
                                                     'longname': 'loadImmediately',
                                                     'modes': ['create',
                                                               'edit'],
                                                     'numArgs': 1,
                                                     'shortname': 'li'},
                                 'minimumHeight': {'args': '<type __builtin__.int>',
                                                   'longname': 'minimumHeight',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'mh'},
                                 'minimumWidth': {'args': '<type __builtin__.int>',
                                                  'longname': 'minimumWidth',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'mw'},
                                 'query': {'args': '<type __builtin__.bool>',
                                           'longname': 'query',
                                           'numArgs': 0,
                                           'shortname': 'q'},
                                 'r': {'args': '<type __builtin__.bool>',
                                       'longname': 'r',
                                       'modes': ['query', 'edit'],
                                       'numArgs': 0,
                                       'shortname': 'r'},
                                 'requiredControl': {'args': '<type __builtin__.str>',
                                                     'longname': 'requiredControl',
                                                     'modes': ['create',
                                                               'edit',
                                                               'multiuse'],
                                                     'numArgs': 1,
                                                     'shortname': 'rc'},
                                 'requiredPlugin': {'args': '<type __builtin__.str>',
                                                    'longname': 'requiredPlugin',
                                                    'modes': ['create',
                                                              'edit',
                                                              'multiuse'],
                                                    'numArgs': 1,
                                                    'shortname': 'rp'},
                                 'resizeHeight': {'args': '<type __builtin__.int>',
                                                  'longname': 'resizeHeight',
                                                  'modes': ['edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'rsh'},
                                 'resizeWidth': {'args': '<type __builtin__.int>',
                                                 'longname': 'resizeWidth',
                                                 'modes': ['edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'rsw'},
                                 'restore': {'args': '<type __builtin__.bool>',
                                             'longname': 'restore',
                                             'modes': ['create', 'edit'],
                                             'numArgs': 0,
                                             'shortname': 'rs'},
                                 'retain': {'args': '<type __builtin__.bool>',
                                            'longname': 'retain',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'rt'},
                                 'stateString': {'args': '<type __builtin__.str>',
                                                 'longname': 'stateString',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'ss'},
                                 'tabPosition': {'args': ['<type __builtin__.str>',
                                                          '<type __builtin__.bool>'],
                                                 'longname': 'tabPosition',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 2,
                                                 'shortname': 'tp'},
                                 'tabToControl': {'args': ['<type __builtin__.str>',
                                                           '<type __builtin__.int>'],
                                                  'longname': 'tabToControl',
                                                  'modes': ['create',
                                                            'edit'],
                                                  'numArgs': 2,
                                                  'shortname': 'ttc'},
                                 'uiScript': {'args': 'script',
                                              'longname': 'uiScript',
                                              'modes': ['create', 'edit'],
                                              'numArgs': 1,
                                              'shortname': 'ui'},
                                 'useTemplate': {'args': '<type __builtin__.str>',
                                                 'longname': 'useTemplate',
                                                 'modes': ['create'],
                                                 'numArgs': 1,
                                                 'shortname': 'ut'},
                                 'visible': {'args': '<type __builtin__.bool>',
                                             'longname': 'visible',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'vis'},
                                 'visibleChangeCommand': {'args': 'script',
                                                          'longname': 'visibleChangeCommand',
                                                          'modes': ['create',
                                                                    'edit'],
                                                          'numArgs': 1,
                                                          'shortname': 'vcc'},
                                 'width': {'args': '<type __builtin__.bool>',
                                           'longname': 'width',
                                           'modes': ['query'],
                                           'numArgs': 0,
                                           'shortname': 'w'},
                                 'widthProperty': {'args': '<type __builtin__.str>',
                                                   'longname': 'widthProperty',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'wp'}},
                       'removedFlags': {'del': 'deleteLater', 'raise': 'r'},
                       'shortFlags': {'cc': 'closeCommand',
                                      'cl': 'close',
                                      'clp': 'collapse',
                                      'cp': 'checksPlugins',
                                      'deleteLater': 'deleteLater',
                                      'dt': 'defineTemplate',
                                      'dtc': 'dockToControl',
                                      'dtm': 'dockToMainWindow',
                                      'dtp': 'dockToPanel',
                                      'dup': 'duplicatable',
                                      'e': 'edit',
                                      'ex': 'exists',
                                      'fl': 'floating',
                                      'h': 'height',
                                      'hp': 'heightProperty',
                                      'hr': 'horizontal',
                                      'ic': 'initCallback',
                                      'ih': 'initialHeight',
                                      'iw': 'initialWidth',
                                      'l': 'label',
                                      'li': 'loadImmediately',
                                      'mh': 'minimumHeight',
                                      'mw': 'minimumWidth',
                                      'q': 'query',
                                      'r': 'r',
                                      'rc': 'requiredControl',
                                      'rp': 'requiredPlugin',
                                      'rs': 'restore',
                                      'rsh': 'resizeHeight',
                                      'rsw': 'resizeWidth',
                                      'rt': 'retain',
                                      'ss': 'stateString',
                                      'tp': 'tabPosition',
                                      'ttc': 'tabToControl',
                                      'ui': 'uiScript',
                                      'ut': 'useTemplate',
                                      'vcc': 'visibleChangeCommand',
                                      'vis': 'visible',
                                      'w': 'width',
                                      'wp': 'widthProperty'},
                       'type': 'windows'},
  'workspaceControlState': {'flags': {'defaultTopLeftCorner': {'args': ['<type __builtin__.int>',
                                                                        '<type __builtin__.int>'],
                                                               'longname': 'defaultTopLeftCorner',
                                                               'modes': ['create',
                                                                         'query',
                                                                         'edit'],
                                                               'numArgs': 2,
                                                               'shortname': 'dc'},
                                      'defaultWidthHeight': {'args': ['<type __builtin__.int>',
                                                                      '<type __builtin__.int>'],
                                                             'longname': 'defaultWidthHeight',
                                                             'modes': ['create',
                                                                       'query',
                                                                       'edit'],
                                                             'numArgs': 2,
                                                             'shortname': 'dwh'},
                                      'edit': {'args': '<type __builtin__.bool>',
                                               'longname': 'edit',
                                               'numArgs': 0,
                                               'shortname': 'e'},
                                      'exists': {'args': '<type __builtin__.bool>',
                                                 'longname': 'exists',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'ex'},
                                      'height': {'args': '<type __builtin__.int>',
                                                 'longname': 'height',
                                                 'modes': ['create',
                                                           'query',
                                                           'edit'],
                                                 'numArgs': 1,
                                                 'shortname': 'h'},
                                      'leftEdge': {'args': '<type __builtin__.int>',
                                                   'longname': 'leftEdge',
                                                   'modes': ['create',
                                                             'query',
                                                             'edit'],
                                                   'numArgs': 1,
                                                   'shortname': 'le'},
                                      'maximized': {'args': '<type __builtin__.bool>',
                                                    'longname': 'maximized',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'max'},
                                      'query': {'args': '<type __builtin__.bool>',
                                                'longname': 'query',
                                                'numArgs': 0,
                                                'shortname': 'q'},
                                      'remove': {'args': '<type __builtin__.bool>',
                                                 'longname': 'remove',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'r'},
                                      'topEdge': {'args': '<type __builtin__.int>',
                                                  'longname': 'topEdge',
                                                  'modes': ['create',
                                                            'query',
                                                            'edit'],
                                                  'numArgs': 1,
                                                  'shortname': 'te'},
                                      'topLeftCorner': {'args': ['<type __builtin__.int>',
                                                                 '<type __builtin__.int>'],
                                                        'longname': 'topLeftCorner',
                                                        'modes': ['create',
                                                                  'query',
                                                                  'edit'],
                                                        'numArgs': 2,
                                                        'shortname': 'tlc'},
                                      'width': {'args': '<type __builtin__.int>',
                                                'longname': 'width',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'w'},
                                      'widthHeight': {'args': ['<type __builtin__.int>',
                                                               '<type __builtin__.int>'],
                                                      'longname': 'widthHeight',
                                                      'modes': ['create',
                                                                'query',
                                                                'edit'],
                                                      'numArgs': 2,
                                                      'shortname': 'wh'}},
                            'shortFlags': {'dc': 'defaultTopLeftCorner',
                                           'dwh': 'defaultWidthHeight',
                                           'e': 'edit',
                                           'ex': 'exists',
                                           'h': 'height',
                                           'le': 'leftEdge',
                                           'max': 'maximized',
                                           'q': 'query',
                                           'r': 'remove',
                                           'te': 'topEdge',
                                           'tlc': 'topLeftCorner',
                                           'w': 'width',
                                           'wh': 'widthHeight'},
                            'type': 'windows'},
  'workspaceLayoutManager': {'flags': {'collapseMainWindowControls': {'args': ['<type __builtin__.str>',
                                                                               '<type __builtin__.bool>'],
                                                                      'longname': 'collapseMainWindowControls',
                                                                      'modes': ['create'],
                                                                      'numArgs': 2,
                                                                      'shortname': 'cmw'},
                                       'current': {'args': '<type __builtin__.bool>',
                                                   'longname': 'current',
                                                   'modes': ['create',
                                                             'query'],
                                                   'numArgs': 0,
                                                   'shortname': 'cu'},
                                       'delete': {'args': '<type __builtin__.str>',
                                                  'longname': 'delete',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'd'},
                                       'edit': {'args': '<type __builtin__.bool>',
                                                'longname': 'edit',
                                                'numArgs': 0,
                                                'shortname': 'e'},
                                       'i': {'args': '<type __builtin__.str>',
                                             'longname': 'i',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'i'},
                                       'listLayouts': {'args': '<type __builtin__.bool>',
                                                       'longname': 'listLayouts',
                                                       'modes': ['create'],
                                                       'numArgs': 0,
                                                       'shortname': 'll'},
                                       'listModuleLayouts': {'args': '<type __builtin__.bool>',
                                                             'longname': 'listModuleLayouts',
                                                             'modes': ['create'],
                                                             'numArgs': 0,
                                                             'shortname': 'lml'},
                                       'listUserLayouts': {'args': '<type __builtin__.bool>',
                                                           'longname': 'listUserLayouts',
                                                           'modes': ['create'],
                                                           'numArgs': 0,
                                                           'shortname': 'lul'},
                                       'modified': {'args': '<type __builtin__.str>',
                                                    'longname': 'modified',
                                                    'modes': ['create'],
                                                    'numArgs': 1,
                                                    'shortname': 'm'},
                                       'parentWorkspaceControl': {'args': '<type __builtin__.str>',
                                                                  'longname': 'parentWorkspaceControl',
                                                                  'modes': ['create'],
                                                                  'numArgs': 1,
                                                                  'shortname': 'pwc'},
                                       'query': {'args': '<type __builtin__.bool>',
                                                 'longname': 'query',
                                                 'numArgs': 0,
                                                 'shortname': 'q'},
                                       'reset': {'args': '<type __builtin__.bool>',
                                                 'longname': 'reset',
                                                 'modes': ['create'],
                                                 'numArgs': 0,
                                                 'shortname': 'rs'},
                                       'restoreMainWindowControls': {'args': '<type __builtin__.bool>',
                                                                     'longname': 'restoreMainWindowControls',
                                                                     'modes': ['create'],
                                                                     'numArgs': 0,
                                                                     'shortname': 'rmw'},
                                       'save': {'args': '<type __builtin__.bool>',
                                                'longname': 'save',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 's'},
                                       'saveAs': {'args': '<type __builtin__.str>',
                                                  'longname': 'saveAs',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'sa'},
                                       'setCurrent': {'args': '<type __builtin__.str>',
                                                      'longname': 'setCurrent',
                                                      'modes': ['create'],
                                                      'numArgs': 1,
                                                      'shortname': 'sc'},
                                       'setCurrentCallback': {'args': '<type __builtin__.str>',
                                                              'longname': 'setCurrentCallback',
                                                              'modes': ['create'],
                                                              'numArgs': 1,
                                                              'shortname': 'scc'},
                                       'setModifiedCallback': {'args': '<type __builtin__.str>',
                                                               'longname': 'setModifiedCallback',
                                                               'modes': ['create'],
                                                               'numArgs': 1,
                                                               'shortname': 'smc'},
                                       'type': {'args': '<type __builtin__.str>',
                                                'longname': 'type',
                                                'modes': ['create'],
                                                'numArgs': 1,
                                                'shortname': 't'}},
                             'removedFlags': {'import': 'i'},
                             'shortFlags': {'cmw': 'collapseMainWindowControls',
                                            'cu': 'current',
                                            'd': 'delete',
                                            'e': 'edit',
                                            'i': 'i',
                                            'll': 'listLayouts',
                                            'lml': 'listModuleLayouts',
                                            'lul': 'listUserLayouts',
                                            'm': 'modified',
                                            'pwc': 'parentWorkspaceControl',
                                            'q': 'query',
                                            'rmw': 'restoreMainWindowControls',
                                            'rs': 'reset',
                                            's': 'save',
                                            'sa': 'saveAs',
                                            'sc': 'setCurrent',
                                            'scc': 'setCurrentCallback',
                                            'smc': 'setModifiedCallback',
                                            't': 'type'},
                             'type': 'windows'},
  'workspacePanel': {'flags': {'defineTemplate': {'args': '<type __builtin__.str>',
                                                  'longname': 'defineTemplate',
                                                  'modes': ['create'],
                                                  'numArgs': 1,
                                                  'shortname': 'dt'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'mainWindow': {'args': '<type __builtin__.bool>',
                                              'longname': 'mainWindow',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 0,
                                              'shortname': 'mw'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'useTemplate': {'args': '<type __builtin__.str>',
                                               'longname': 'useTemplate',
                                               'modes': ['create'],
                                               'numArgs': 1,
                                               'shortname': 'ut'}},
                     'shortFlags': {'dt': 'defineTemplate',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'mw': 'mainWindow',
                                    'q': 'query',
                                    'ut': 'useTemplate'},
                     'type': 'windows'},
  'wrinkle': {'flags': {'axis': {'args': ['<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>'],
                                 'longname': 'axis',
                                 'modes': ['create'],
                                 'numArgs': 3,
                                 'shortname': 'ax'},
                        'branchCount': {'args': '<type __builtin__.int>',
                                        'longname': 'branchCount',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'brc'},
                        'branchDepth': {'args': '<type __builtin__.int>',
                                        'longname': 'branchDepth',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'bd'},
                        'center': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'center',
                                   'modes': ['create'],
                                   'numArgs': 3,
                                   'shortname': 'ct'},
                        'crease': {'args': '<type __builtin__.str>',
                                   'longname': 'crease',
                                   'modes': ['create', 'multiuse'],
                                   'numArgs': 1,
                                   'shortname': 'cr'},
                        'dropoffDistance': {'args': '<type __builtin__.float>',
                                            'longname': 'dropoffDistance',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'dds'},
                        'envelope': {'args': '<type __builtin__.float>',
                                     'longname': 'envelope',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'en'},
                        'randomness': {'args': '<type __builtin__.float>',
                                       'longname': 'randomness',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'rnd'},
                        'style': {'args': '<type __builtin__.str>',
                                  'longname': 'style',
                                  'modes': ['create'],
                                  'numArgs': 1,
                                  'shortname': 'st'},
                        'thickness': {'args': '<type __builtin__.float>',
                                      'longname': 'thickness',
                                      'modes': ['create'],
                                      'numArgs': 1,
                                      'shortname': 'th'},
                        'uvSpace': {'args': ['<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>',
                                             '<type __builtin__.float>'],
                                    'longname': 'uvSpace',
                                    'modes': ['create'],
                                    'numArgs': 5,
                                    'shortname': 'uv'},
                        'wrinkleCount': {'args': '<type __builtin__.int>',
                                         'longname': 'wrinkleCount',
                                         'modes': ['create'],
                                         'numArgs': 1,
                                         'shortname': 'wc'},
                        'wrinkleIntensity': {'args': '<type __builtin__.float>',
                                             'longname': 'wrinkleIntensity',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'wi'}},
              'shortFlags': {'ax': 'axis',
                             'bd': 'branchDepth',
                             'brc': 'branchCount',
                             'cr': 'crease',
                             'ct': 'center',
                             'dds': 'dropoffDistance',
                             'en': 'envelope',
                             'rnd': 'randomness',
                             'st': 'style',
                             'th': 'thickness',
                             'uv': 'uvSpace',
                             'wc': 'wrinkleCount',
                             'wi': 'wrinkleIntensity'},
              'type': 'animation'},
  'wrinkleContext': {'flags': {'branchCount': {'args': '<type __builtin__.int>',
                                               'longname': 'branchCount',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'brc'},
                               'branchDepth': {'args': '<type __builtin__.int>',
                                               'longname': 'branchDepth',
                                               'modes': ['create',
                                                         'query',
                                                         'edit'],
                                               'numArgs': 1,
                                               'shortname': 'bd'},
                               'edit': {'args': '<type __builtin__.bool>',
                                        'longname': 'edit',
                                        'numArgs': 0,
                                        'shortname': 'e'},
                               'exists': {'args': '<type __builtin__.bool>',
                                          'longname': 'exists',
                                          'modes': ['create'],
                                          'numArgs': 0,
                                          'shortname': 'ex'},
                               'history': {'args': '<type __builtin__.bool>',
                                           'longname': 'history',
                                           'modes': ['create'],
                                           'numArgs': 1,
                                           'shortname': 'ch'},
                               'image1': {'args': '<type __builtin__.str>',
                                          'longname': 'image1',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i1'},
                               'image2': {'args': '<type __builtin__.str>',
                                          'longname': 'image2',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i2'},
                               'image3': {'args': '<type __builtin__.str>',
                                          'longname': 'image3',
                                          'modes': ['create',
                                                    'query',
                                                    'edit'],
                                          'numArgs': 1,
                                          'shortname': 'i3'},
                               'name': {'args': '<type __builtin__.str>',
                                        'longname': 'name',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'n'},
                               'query': {'args': '<type __builtin__.bool>',
                                         'longname': 'query',
                                         'numArgs': 0,
                                         'shortname': 'q'},
                               'randomness': {'args': '<type __builtin__.float>',
                                              'longname': 'randomness',
                                              'modes': ['create',
                                                        'query',
                                                        'edit'],
                                              'numArgs': 1,
                                              'shortname': 'rnd'},
                               'style': {'args': '<type __builtin__.str>',
                                         'longname': 'style',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 'st'},
                               'thickness': {'args': '<type __builtin__.float>',
                                             'longname': 'thickness',
                                             'modes': ['create',
                                                       'query',
                                                       'edit'],
                                             'numArgs': 1,
                                             'shortname': 'th'},
                               'wrinkleCount': {'args': '<type __builtin__.int>',
                                                'longname': 'wrinkleCount',
                                                'modes': ['create',
                                                          'query',
                                                          'edit'],
                                                'numArgs': 1,
                                                'shortname': 'wc'},
                               'wrinkleIntensity': {'args': '<type __builtin__.float>',
                                                    'longname': 'wrinkleIntensity',
                                                    'modes': ['create',
                                                              'query',
                                                              'edit'],
                                                    'numArgs': 1,
                                                    'shortname': 'wi'}},
                     'shortFlags': {'bd': 'branchDepth',
                                    'brc': 'branchCount',
                                    'ch': 'history',
                                    'e': 'edit',
                                    'ex': 'exists',
                                    'i1': 'image1',
                                    'i2': 'image2',
                                    'i3': 'image3',
                                    'n': 'name',
                                    'q': 'query',
                                    'rnd': 'randomness',
                                    'st': 'style',
                                    'th': 'thickness',
                                    'wc': 'wrinkleCount',
                                    'wi': 'wrinkleIntensity'},
                     'type': 'context'},
  'writeTake': {'flags': {'angle': {'args': '<type __builtin__.str>',
                                    'longname': 'angle',
                                    'modes': ['create'],
                                    'numArgs': 1,
                                    'shortname': 'a'},
                          'device': {'args': '<type __builtin__.str>',
                                     'longname': 'device',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'd'},
                          'linear': {'args': '<type __builtin__.str>',
                                     'longname': 'linear',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'l'},
                          'noTime': {'args': '<type __builtin__.bool>',
                                     'longname': 'noTime',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'nt'},
                          'precision': {'args': '<type __builtin__.int>',
                                        'longname': 'precision',
                                        'modes': ['create'],
                                        'numArgs': 1,
                                        'shortname': 'pre'},
                          'take': {'args': '<type __builtin__.str>',
                                   'longname': 'take',
                                   'modes': ['create'],
                                   'numArgs': 1,
                                   'shortname': 't'},
                          'virtualDevice': {'args': '<type __builtin__.str>',
                                            'longname': 'virtualDevice',
                                            'modes': ['create'],
                                            'numArgs': 1,
                                            'shortname': 'vd'}},
                'shortFlags': {},
                'type': 'animation'},
  'xform': {'flags': {'absolute': {'args': '<type __builtin__.bool>',
                                   'longname': 'absolute',
                                   'modes': ['create'],
                                   'modified': [],
                                   'numArgs': 0,
                                   'shortname': 'a'},
                      'boundingBox': {'args': '<type __builtin__.bool>',
                                      'longname': 'boundingBox',
                                      'modes': ['query'],
                                      'numArgs': 0,
                                      'secondaryFlags': ['objectSpace',
                                                         'worldSpace'],
                                      'shortname': 'bb'},
                      'boundingBoxInvisible': {'args': '<type __builtin__.bool>',
                                               'longname': 'boundingBoxInvisible',
                                               'modes': ['query'],
                                               'numArgs': 0,
                                               'secondaryFlags': ['objectSpace',
                                                                  'worldSpace'],
                                               'shortname': 'bbi'},
                      'centerPivots': {'args': '<type __builtin__.bool>',
                                       'longname': 'centerPivots',
                                       'modes': ['create'],
                                       'numArgs': 0,
                                       'secondaryFlags': ['preserve'],
                                       'shortname': 'cp'},
                      'centerPivotsOnComponents': {'args': '<type __builtin__.bool>',
                                                   'longname': 'centerPivotsOnComponents',
                                                   'modes': ['create'],
                                                   'numArgs': 0,
                                                   'shortname': 'cpc'},
                      'deletePriorHistory': {'args': '<type __builtin__.bool>',
                                             'longname': 'deletePriorHistory',
                                             'modes': ['create'],
                                             'numArgs': 1,
                                             'shortname': 'dph'},
                      'euler': {'args': '<type __builtin__.bool>',
                                'longname': 'euler',
                                'modes': ['create'],
                                'modified': ['relative'],
                                'numArgs': 0,
                                'shortname': 'eu'},
                      'matrix': {'args': ['<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>'],
                                 'longname': 'matrix',
                                 'modes': ['create', 'query'],
                                 'numArgs': 16,
                                 'secondaryFlags': ['objectSpace',
                                                    'worldSpace'],
                                 'shortname': 'm'},
                      'objectSpace': {'args': '<type __builtin__.bool>',
                                      'longname': 'objectSpace',
                                      'modes': ['create', 'query'],
                                      'modified': ['scalePivot',
                                                   'rotatePivot',
                                                   'rotateAxis',
                                                   'rotation',
                                                   'rotateTranslation',
                                                   'translation',
                                                   'matrix',
                                                   'boundingBox',
                                                   'boundingBoxInvisible',
                                                   'pivots'],
                                      'numArgs': 0,
                                      'shortname': 'os'},
                      'pivots': {'args': ['<type __builtin__.float>',
                                          '<type __builtin__.float>',
                                          '<type __builtin__.float>'],
                                 'longname': 'pivots',
                                 'modes': ['create', 'query'],
                                 'numArgs': 3,
                                 'secondaryFlags': ['objectSpace',
                                                    'worldSpace',
                                                    'worldSpaceDistance'],
                                 'shortname': 'piv'},
                      'preserve': {'args': '<type __builtin__.bool>',
                                   'longname': 'preserve',
                                   'modes': ['create'],
                                   'modified': ['scalePivot',
                                                'rotatePivot',
                                                'rotateOrder',
                                                'rotateAxis',
                                                'centerPivots'],
                                   'numArgs': 1,
                                   'shortname': 'p'},
                      'preserveUV': {'args': '<type __builtin__.bool>',
                                     'longname': 'preserveUV',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'puv'},
                      'query': {'args': '<type __builtin__.bool>',
                                'longname': 'query',
                                'numArgs': 0,
                                'shortname': 'q'},
                      'reflection': {'args': '<type __builtin__.bool>',
                                     'longname': 'reflection',
                                     'modes': ['create'],
                                     'numArgs': 0,
                                     'shortname': 'rfl'},
                      'reflectionAboutBBox': {'args': '<type __builtin__.bool>',
                                              'longname': 'reflectionAboutBBox',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'rab'},
                      'reflectionAboutOrigin': {'args': '<type __builtin__.bool>',
                                                'longname': 'reflectionAboutOrigin',
                                                'modes': ['create'],
                                                'numArgs': 0,
                                                'shortname': 'rao'},
                      'reflectionAboutX': {'args': '<type __builtin__.bool>',
                                           'longname': 'reflectionAboutX',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'rax'},
                      'reflectionAboutY': {'args': '<type __builtin__.bool>',
                                           'longname': 'reflectionAboutY',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'ray'},
                      'reflectionAboutZ': {'args': '<type __builtin__.bool>',
                                           'longname': 'reflectionAboutZ',
                                           'modes': ['create'],
                                           'numArgs': 0,
                                           'shortname': 'raz'},
                      'reflectionTolerance': {'args': '<type __builtin__.float>',
                                              'longname': 'reflectionTolerance',
                                              'modes': ['create'],
                                              'numArgs': 1,
                                              'shortname': 'rft'},
                      'relative': {'args': '<type __builtin__.bool>',
                                   'longname': 'relative',
                                   'modes': ['create'],
                                   'modified': [],
                                   'numArgs': 0,
                                   'secondaryFlags': ['euler'],
                                   'shortname': 'r'},
                      'rotateAxis': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'rotateAxis',
                                     'modes': ['create', 'query'],
                                     'numArgs': 3,
                                     'secondaryFlags': ['objectSpace',
                                                        'worldSpace',
                                                        'preserve'],
                                     'shortname': 'ra'},
                      'rotateOrder': {'args': '<type __builtin__.str>',
                                      'longname': 'rotateOrder',
                                      'modes': ['create', 'query'],
                                      'numArgs': 1,
                                      'secondaryFlags': ['preserve'],
                                      'shortname': 'roo'},
                      'rotatePivot': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'rotatePivot',
                                      'modes': ['create', 'query'],
                                      'numArgs': 3,
                                      'secondaryFlags': ['objectSpace',
                                                         'worldSpace',
                                                         'preserve',
                                                         'worldSpaceDistance'],
                                      'shortname': 'rp'},
                      'rotateTranslation': {'args': ['<type __builtin__.float>',
                                                     '<type __builtin__.float>',
                                                     '<type __builtin__.float>'],
                                            'longname': 'rotateTranslation',
                                            'modes': ['create', 'query'],
                                            'numArgs': 3,
                                            'secondaryFlags': ['objectSpace',
                                                               'worldSpace',
                                                               'worldSpaceDistance'],
                                            'shortname': 'rt'},
                      'rotation': {'args': ['<type __builtin__.float>',
                                            '<type __builtin__.float>',
                                            '<type __builtin__.float>'],
                                   'longname': 'rotation',
                                   'modes': ['create', 'query'],
                                   'numArgs': 3,
                                   'secondaryFlags': ['objectSpace',
                                                      'worldSpace'],
                                   'shortname': 'ro'},
                      'scale': {'args': ['<type __builtin__.float>',
                                         '<type __builtin__.float>',
                                         '<type __builtin__.float>'],
                                'longname': 'scale',
                                'modes': ['create', 'query'],
                                'numArgs': 3,
                                'shortname': 's'},
                      'scalePivot': {'args': ['<type __builtin__.float>',
                                              '<type __builtin__.float>',
                                              '<type __builtin__.float>'],
                                     'longname': 'scalePivot',
                                     'modes': ['create', 'query'],
                                     'numArgs': 3,
                                     'secondaryFlags': ['objectSpace',
                                                        'worldSpace',
                                                        'preserve',
                                                        'worldSpaceDistance'],
                                     'shortname': 'sp'},
                      'scaleTranslation': {'args': ['<type __builtin__.float>',
                                                    '<type __builtin__.float>',
                                                    '<type __builtin__.float>'],
                                           'longname': 'scaleTranslation',
                                           'modes': ['create', 'query'],
                                           'numArgs': 3,
                                           'secondaryFlags': ['worldSpaceDistance'],
                                           'shortname': 'st'},
                      'shear': {'args': ['<type __builtin__.float>',
                                         '<type __builtin__.float>',
                                         '<type __builtin__.float>'],
                                'longname': 'shear',
                                'modes': ['create', 'query'],
                                'numArgs': 3,
                                'shortname': 'sh'},
                      'translation': {'args': ['<type __builtin__.float>',
                                               '<type __builtin__.float>',
                                               '<type __builtin__.float>'],
                                      'longname': 'translation',
                                      'modes': ['create', 'query'],
                                      'numArgs': 3,
                                      'secondaryFlags': ['objectSpace',
                                                         'worldSpace',
                                                         'worldSpaceDistance'],
                                      'shortname': 't'},
                      'worldSpace': {'args': '<type __builtin__.bool>',
                                     'longname': 'worldSpace',
                                     'modes': ['create', 'query'],
                                     'modified': ['scalePivot',
                                                  'rotatePivot',
                                                  'rotateAxis',
                                                  'rotation',
                                                  'rotateTranslation',
                                                  'translation',
                                                  'matrix',
                                                  'boundingBox',
                                                  'boundingBoxInvisible',
                                                  'pivots'],
                                     'numArgs': 0,
                                     'shortname': 'ws'},
                      'worldSpaceDistance': {'args': '<type __builtin__.bool>',
                                             'longname': 'worldSpaceDistance',
                                             'modes': ['create', 'query'],
                                             'modified': ['scalePivot',
                                                          'rotatePivot',
                                                          'scaleTranslation',
                                                          'rotateTranslation',
                                                          'translation',
                                                          'pivots'],
                                             'numArgs': 0,
                                             'shortname': 'wd'},
                      'zeroTransformPivots': {'args': '<type __builtin__.bool>',
                                              'longname': 'zeroTransformPivots',
                                              'modes': ['create'],
                                              'numArgs': 0,
                                              'shortname': 'ztp'}},
            'shortFlags': {'a': 'absolute',
                           'bb': 'boundingBox',
                           'bbi': 'boundingBoxInvisible',
                           'cp': 'centerPivots',
                           'cpc': 'centerPivotsOnComponents',
                           'dph': 'deletePriorHistory',
                           'eu': 'euler',
                           'm': 'matrix',
                           'os': 'objectSpace',
                           'p': 'preserve',
                           'piv': 'pivots',
                           'puv': 'preserveUV',
                           'q': 'query',
                           'r': 'relative',
                           'ra': 'rotateAxis',
                           'rab': 'reflectionAboutBBox',
                           'rao': 'reflectionAboutOrigin',
                           'rax': 'reflectionAboutX',
                           'ray': 'reflectionAboutY',
                           'raz': 'reflectionAboutZ',
                           'rfl': 'reflection',
                           'rft': 'reflectionTolerance',
                           'ro': 'rotation',
                           'roo': 'rotateOrder',
                           'rp': 'rotatePivot',
                           'rt': 'rotateTranslation',
                           's': 'scale',
                           'sh': 'shear',
                           'sp': 'scalePivot',
                           'st': 'scaleTranslation',
                           't': 'translation',
                           'wd': 'worldSpaceDistance',
                           'ws': 'worldSpace',
                           'ztp': 'zeroTransformPivots'},
            'type': 'general'},
  'xformConstraint': {'flags': {'alongNormal': {'args': '<type __builtin__.int>',
                                                'longname': 'alongNormal',
                                                'modes': ['query', 'edit'],
                                                'numArgs': 1,
                                                'shortname': 'n'},
                                'edit': {'args': '<type __builtin__.bool>',
                                         'longname': 'edit',
                                         'numArgs': 0,
                                         'shortname': 'e'},
                                'live': {'args': '<type __builtin__.bool>',
                                         'longname': 'live',
                                         'modes': ['query'],
                                         'numArgs': 0,
                                         'shortname': 'l'},
                                'query': {'args': '<type __builtin__.bool>',
                                          'longname': 'query',
                                          'numArgs': 0,
                                          'shortname': 'q'},
                                'type': {'args': '<type __builtin__.str>',
                                         'longname': 'type',
                                         'modes': ['create',
                                                   'query',
                                                   'edit'],
                                         'numArgs': 1,
                                         'shortname': 't'}},
                      'shortFlags': {'e': 'edit',
                                     'l': 'live',
                                     'n': 'alongNormal',
                                     'q': 'query',
                                     't': 'type'},
                      'type': 'general'},
  'xpmPicker': {'flags': {'fileName': {'args': '<type __builtin__.str>',
                                       'longname': 'fileName',
                                       'modes': ['create'],
                                       'numArgs': 1,
                                       'shortname': 'fn'},
                          'parent': {'args': '<type __builtin__.str>',
                                     'longname': 'parent',
                                     'modes': ['create'],
                                     'numArgs': 1,
                                     'shortname': 'p'}},
                'shortFlags': {'fn': 'fileName', 'p': 'parent'},
                'type': 'system'}},
 [('dependNode',
   (),
   (u'THdependNode',
    u'TadskAssetInstanceNode_TdependNode',
    u'aboutToSetValueTestNode',
    u'abstractBaseCreate',
    u'addDoubleLinear',
    u'addMatrix',
    u'angleBetween',
    u'animBlend',
    u'animBlendNodeBase',
    u'animClip',
    u'animCurve',
    u'arrayMapper',
    u'attrHierarchyTest',
    u'audio',
    u'blend',
    u'blendColors',
    u'blindDataTemplate',
    u'brush',
    u'bump3d',
    u'cacheBase',
    u'cameraSet',
    u'cameraView',
    u'characterMap',
    u'characterOffset',
    u'choice',
    u'chooser',
    u'clamp',
    u'clipLibrary',
    u'clipScheduler',
    u'clipToGhostData',
    u'closestPointOnMesh',
    u'colorManagementGlobals',
    u'colorProfile',
    u'combinationShape',
    u'compactPlugArrayTest',
    u'condition',
    u'containerBase',
    u'contrast',
    u'controller',
    u'copyColorSet',
    u'copyUVSet',
    u'createColorSet',
    u'createUVSet',
    u'curveFromMesh',
    u'curveNormalizer',
    u'dagPose',
    u'dataBlockTest',
    u'defaultLightList',
    u'defaultRenderUtilityList',
    u'defaultRenderingList',
    u'defaultShaderList',
    u'defaultTextureList',
    u'deleteColorSet',
    u'deleteComponent',
    u'deleteUVSet',
    u'diskCache',
    u'displacementShader',
    u'displayLayer',
    u'displayLayerManager',
    u'distanceBetween',
    u'dof',
    u'dynController',
    u'dynGlobals',
    u'editMetadata',
    u'editsManager',
    u'expression',
    u'facade',
    u'flow',
    u'fourByFourMatrix',
    u'frameCache',
    u'gammaCorrect',
    u'geoConnector',
    u'geomBind',
    u'geometryFilter',
    u'globalCacheControl',
    u'greasePencilSequence',
    u'groupId',
    u'groupParts',
    u'guide',
    u'hardwareRenderGlobals',
    u'hardwareRenderingGlobals',
    u'hierarchyTestNode1',
    u'holdMatrix',
    u'hsvToRgb',
    u'hwReflectionMap',
    u'hwRenderGlobals',
    u'hyperGraphInfo',
    u'hyperLayout',
    u'hyperView',
    u'ikSolver',
    u'ikSystem',
    u'imageSource',
    u'lightInfo',
    u'lightLinker',
    u'lightList',
    u'lodThresholds',
    u'luminance',
    u'makeGroup',
    u'materialInfo',
    u'membrane',
    u'motionPath',
    u'multDoubleLinear',
    u'multMatrix',
    u'multilisterLight',
    u'mute',
    u'nComponent',
    u'network',
    u'nodeGraphEditorBookmarkInfo',
    u'nodeGraphEditorBookmarks',
    u'nodeGraphEditorInfo',
    u'objectFilter',
    u'oldBlindDataBase',
    u'oldGeometryConstraint',
    u'opticalFX',
    u'pairBlend',
    u'particleAgeMapper',
    u'particleColorMapper',
    u'particleIncandMapper',
    u'particleTranspMapper',
    u'passContributionMap',
    u'passMatrix',
    u'pointMatrixMult',
    u'polyBase',
    u'polyBlindData',
    u'polySeparate',
    u'polyToSubdiv',
    u'poseInterpolatorManager',
    u'postProcessList',
    u'precompExport',
    u'proxyManager',
    u'record',
    u'reference',
    u'remapColor',
    u'remapHsv',
    u'remapValue',
    u'renderGlobals',
    u'renderGlobalsList',
    u'renderLayerManager',
    u'renderPassSet',
    u'renderQuality',
    u'reorderUVSet',
    u'resolution',
    u'reverse',
    u'rgbToHsv',
    u'rigidSolver',
    u'sampler',
    u'script',
    u'selectionListOperator',
    u'sequenceManager',
    u'sequencer',
    u'setRange',
    u'shaderGlow',
    u'shadingDependNode',
    u'shapeEditorManager',
    u'shot',
    u'simpleTestNode',
    u'simpleVolumeShader',
    u'skinBinding',
    u'snapshot',
    u'strokeGlobals',
    u'subdBase',
    u'subdivCollapse',
    u'subdivComponentId',
    u'subdivReverseFaces',
    u'subdivToPoly',
    u'surfaceLuminance',
    u'surfaceShader',
    u'texLattice',
    u'textureToGeom',
    u'threadedDevice',
    u'time',
    u'timeEditor',
    u'timeEditorAnimSource',
    u'timeEditorClipBase',
    u'timeEditorClipEvaluator',
    u'timeEditorInterpolator',
    u'timeEditorTracks',
    u'timeFunction',
    u'timeToUnitConversion',
    u'timeWarp',
    u'toonLineAttributes',
    u'trackInfoManager',
    u'unitConversion',
    u'unitToTimeConversion',
    u'unknown',
    u'uvChooser',
    u'viewColorManager',
    u'volumeShader',
    u'wtAddMatrix')),
  (u'THdependNode', ('dependNode',), ()),
  (u'TadskAssetInstanceNode_TdependNode',
   ('dependNode',),
   (u'adskMaterial',)),
  (u'adskMaterial',
   (u'TadskAssetInstanceNode_TdependNode', 'dependNode'),
   ()),
  (u'aboutToSetValueTestNode', ('dependNode',), ()),
  (u'abstractBaseCreate',
   ('dependNode',),
   (u'abstractBaseNurbsConversion',
    u'alignCurve',
    u'alignSurface',
    u'arubaTessellate',
    u'attachCurve',
    u'attachSurface',
    u'avgCurves',
    u'avgNurbsSurfacePoints',
    u'avgSurfacePoints',
    u'bevel',
    u'bevelPlus',
    u'bezierCurveToNurbs',
    u'birailSrf',
    u'boolean',
    u'boundaryBase',
    u'closeCurve',
    u'closeSurface',
    u'closestPointOnSurface',
    u'curveInfo',
    u'curveIntersect',
    u'curveRange',
    u'detachCurve',
    u'detachSurface',
    u'explodeNurbsShell',
    u'extendCurve',
    u'extendSurface',
    u'extrude',
    u'ffBlendSrf',
    u'ffBlendSrfObsolete',
    u'ffFilletSrf',
    u'filletCurve',
    u'fitBspline',
    u'globalStitch',
    u'hardenPoint',
    u'insertKnotCurve',
    u'insertKnotSurface',
    u'intersectSurface',
    u'leastSquaresModifier',
    u'loft',
    u'makeCircularArc',
    u'makeIllustratorCurves',
    u'makeNurbCircle',
    u'makeNurbsSquare',
    u'makeTextCurves',
    u'nearestPointOnCurve',
    u'nurbsCurveToBezier',
    u'offsetCos',
    u'offsetCurve',
    u'offsetSurface',
    u'planarTrimSurface',
    u'pointOnCurveInfo',
    u'pointOnSurfaceInfo',
    u'polyEdgeToCurve',
    u'polyProjectCurve',
    u'primitive',
    u'projectCurve',
    u'projectTangent',
    u'rbfSrf',
    u'rebuildCurve',
    u'reverseCurve',
    u'reverseSurface',
    u'revolve',
    u'roundConstantRadius',
    u'smoothCurve',
    u'smoothTangentSrf',
    u'stitchAsNurbsShell',
    u'stitchSrf',
    u'styleCurve',
    u'subSurface',
    u'subdivToNurbs',
    u'surfaceInfo',
    u'transformGeometry',
    u'trim',
    u'trimWithBoundaries',
    u'untrim')),
  (u'abstractBaseNurbsConversion',
   (u'abstractBaseCreate', 'dependNode'),
   (u'nurbsToSubdiv', u'parentTessellate', u'rebuildSurface')),
  (u'nurbsToSubdiv',
   (u'abstractBaseNurbsConversion', u'abstractBaseCreate', 'dependNode'),
   (u'nurbsToSubdivProc',)),
  (u'nurbsToSubdivProc',
   (u'nurbsToSubdiv',
    u'abstractBaseNurbsConversion',
    u'abstractBaseCreate',
    'dependNode'),
   ()),
  (u'parentTessellate',
   (u'abstractBaseNurbsConversion', u'abstractBaseCreate', 'dependNode'),
   (u'nurbsTessellate', u'shellTessellate')),
  (u'nurbsTessellate',
   (u'parentTessellate',
    u'abstractBaseNurbsConversion',
    u'abstractBaseCreate',
    'dependNode'),
   ()),
  (u'shellTessellate',
   (u'parentTessellate',
    u'abstractBaseNurbsConversion',
    u'abstractBaseCreate',
    'dependNode'),
   ()),
  (u'rebuildSurface',
   (u'abstractBaseNurbsConversion', u'abstractBaseCreate', 'dependNode'),
   ()),
  (u'alignCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'alignSurface', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'arubaTessellate', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'attachCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'attachSurface', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'avgCurves', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'avgNurbsSurfacePoints', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'avgSurfacePoints', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'bevel', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'bevelPlus', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'bezierCurveToNurbs', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'birailSrf',
   (u'abstractBaseCreate', 'dependNode'),
   (u'dpBirailSrf', u'mpBirailSrf', u'spBirailSrf')),
  (u'dpBirailSrf', (u'birailSrf', u'abstractBaseCreate', 'dependNode'), ()),
  (u'mpBirailSrf', (u'birailSrf', u'abstractBaseCreate', 'dependNode'), ()),
  (u'spBirailSrf', (u'birailSrf', u'abstractBaseCreate', 'dependNode'), ()),
  (u'boolean', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'boundaryBase',
   (u'abstractBaseCreate', 'dependNode'),
   (u'boundary', u'squareSrf')),
  (u'boundary', (u'boundaryBase', u'abstractBaseCreate', 'dependNode'), ()),
  (u'squareSrf', (u'boundaryBase', u'abstractBaseCreate', 'dependNode'), ()),
  (u'closeCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'closeSurface', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'closestPointOnSurface', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'curveInfo', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'curveIntersect', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'curveRange',
   (u'abstractBaseCreate', 'dependNode'),
   (u'curveFromSubdiv', u'curveFromSurface', u'subCurve')),
  (u'curveFromSubdiv',
   (u'curveRange', u'abstractBaseCreate', 'dependNode'),
   (u'curveFromSubdivEdge', u'curveFromSubdivFace')),
  (u'curveFromSubdivEdge',
   (u'curveFromSubdiv', u'curveRange', u'abstractBaseCreate', 'dependNode'),
   ()),
  (u'curveFromSubdivFace',
   (u'curveFromSubdiv', u'curveRange', u'abstractBaseCreate', 'dependNode'),
   ()),
  (u'curveFromSurface',
   (u'curveRange', u'abstractBaseCreate', 'dependNode'),
   (u'curveFromSurfaceBnd', u'curveFromSurfaceCoS', u'curveFromSurfaceIso')),
  (u'curveFromSurfaceBnd',
   (u'curveFromSurface', u'curveRange', u'abstractBaseCreate', 'dependNode'),
   ()),
  (u'curveFromSurfaceCoS',
   (u'curveFromSurface', u'curveRange', u'abstractBaseCreate', 'dependNode'),
   ()),
  (u'curveFromSurfaceIso',
   (u'curveFromSurface', u'curveRange', u'abstractBaseCreate', 'dependNode'),
   ()),
  (u'subCurve', (u'curveRange', u'abstractBaseCreate', 'dependNode'), ()),
  (u'detachCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'detachSurface', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'explodeNurbsShell', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'extendCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'extendSurface', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'extrude', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'ffBlendSrf', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'ffBlendSrfObsolete', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'ffFilletSrf', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'filletCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'fitBspline', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'globalStitch', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'hardenPoint', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'insertKnotCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'insertKnotSurface', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'intersectSurface', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'leastSquaresModifier', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'loft', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'makeCircularArc',
   (u'abstractBaseCreate', 'dependNode'),
   (u'makeThreePointCircularArc', u'makeTwoPointCircularArc')),
  (u'makeThreePointCircularArc',
   (u'makeCircularArc', u'abstractBaseCreate', 'dependNode'),
   ()),
  (u'makeTwoPointCircularArc',
   (u'makeCircularArc', u'abstractBaseCreate', 'dependNode'),
   ()),
  (u'makeIllustratorCurves', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'makeNurbCircle', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'makeNurbsSquare', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'makeTextCurves', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'nearestPointOnCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'nurbsCurveToBezier', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'offsetCos', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'offsetCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'offsetSurface', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'planarTrimSurface', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'pointOnCurveInfo', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'pointOnSurfaceInfo', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'polyEdgeToCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'polyProjectCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'primitive',
   (u'abstractBaseCreate', 'dependNode'),
   (u'makeNurbCube', u'makeNurbPlane', u'revolvedPrimitive')),
  (u'makeNurbCube', (u'primitive', u'abstractBaseCreate', 'dependNode'), ()),
  (u'makeNurbPlane', (u'primitive', u'abstractBaseCreate', 'dependNode'), ()),
  (u'revolvedPrimitive',
   (u'primitive', u'abstractBaseCreate', 'dependNode'),
   (u'makeNurbCone',
    u'makeNurbCylinder',
    u'makeNurbSphere',
    u'makeNurbTorus')),
  (u'makeNurbCone',
   (u'revolvedPrimitive', u'primitive', u'abstractBaseCreate', 'dependNode'),
   ()),
  (u'makeNurbCylinder',
   (u'revolvedPrimitive', u'primitive', u'abstractBaseCreate', 'dependNode'),
   ()),
  (u'makeNurbSphere',
   (u'revolvedPrimitive', u'primitive', u'abstractBaseCreate', 'dependNode'),
   ()),
  (u'makeNurbTorus',
   (u'revolvedPrimitive', u'primitive', u'abstractBaseCreate', 'dependNode'),
   ()),
  (u'projectCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'projectTangent', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'rbfSrf', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'rebuildCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'reverseCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'reverseSurface', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'revolve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'roundConstantRadius', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'smoothCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'smoothTangentSrf', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'stitchAsNurbsShell', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'stitchSrf', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'styleCurve', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'subSurface', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'subdivToNurbs', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'surfaceInfo', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'transformGeometry', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'trim', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'trimWithBoundaries', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'untrim', (u'abstractBaseCreate', 'dependNode'), ()),
  (u'addDoubleLinear', ('dependNode',), ()),
  (u'addMatrix', ('dependNode',), ()),
  (u'angleBetween', ('dependNode',), ()),
  (u'animBlend', ('dependNode',), (u'animBlendInOut',)),
  (u'animBlendInOut', (u'animBlend', 'dependNode'), ()),
  (u'animBlendNodeBase',
   ('dependNode',),
   (u'animBlendNodeAdditive',
    u'animBlendNodeAdditiveDA',
    u'animBlendNodeAdditiveDL',
    u'animBlendNodeAdditiveF',
    u'animBlendNodeAdditiveFA',
    u'animBlendNodeAdditiveFL',
    u'animBlendNodeAdditiveI16',
    u'animBlendNodeAdditiveI32',
    u'animBlendNodeAdditiveRotation',
    u'animBlendNodeAdditiveScale',
    u'animBlendNodeBoolean',
    u'animBlendNodeEnum',
    u'animBlendNodeTime')),
  (u'animBlendNodeAdditive', (u'animBlendNodeBase', 'dependNode'), ()),
  (u'animBlendNodeAdditiveDA', (u'animBlendNodeBase', 'dependNode'), ()),
  (u'animBlendNodeAdditiveDL', (u'animBlendNodeBase', 'dependNode'), ()),
  (u'animBlendNodeAdditiveF', (u'animBlendNodeBase', 'dependNode'), ()),
  (u'animBlendNodeAdditiveFA', (u'animBlendNodeBase', 'dependNode'), ()),
  (u'animBlendNodeAdditiveFL', (u'animBlendNodeBase', 'dependNode'), ()),
  (u'animBlendNodeAdditiveI16', (u'animBlendNodeBase', 'dependNode'), ()),
  (u'animBlendNodeAdditiveI32', (u'animBlendNodeBase', 'dependNode'), ()),
  (u'animBlendNodeAdditiveRotation',
   (u'animBlendNodeBase', 'dependNode'),
   ()),
  (u'animBlendNodeAdditiveScale', (u'animBlendNodeBase', 'dependNode'), ()),
  (u'animBlendNodeBoolean', (u'animBlendNodeBase', 'dependNode'), ()),
  (u'animBlendNodeEnum', (u'animBlendNodeBase', 'dependNode'), ()),
  (u'animBlendNodeTime', (u'animBlendNodeBase', 'dependNode'), ()),
  (u'animClip', ('dependNode',), ()),
  (u'animCurve',
   ('dependNode',),
   (u'animCurveTA',
    u'animCurveTL',
    u'animCurveTT',
    u'animCurveTU',
    u'animCurveUA',
    u'animCurveUL',
    u'animCurveUT',
    u'animCurveUU',
    u'resultCurve')),
  (u'animCurveTA', (u'animCurve', 'dependNode'), ()),
  (u'animCurveTL', (u'animCurve', 'dependNode'), ()),
  (u'animCurveTT', (u'animCurve', 'dependNode'), ()),
  (u'animCurveTU', (u'animCurve', 'dependNode'), ()),
  (u'animCurveUA', (u'animCurve', 'dependNode'), ()),
  (u'animCurveUL', (u'animCurve', 'dependNode'), ()),
  (u'animCurveUT', (u'animCurve', 'dependNode'), ()),
  (u'animCurveUU', (u'animCurve', 'dependNode'), ()),
  (u'resultCurve',
   (u'animCurve', 'dependNode'),
   (u'resultCurveTimeToAngular',
    u'resultCurveTimeToLinear',
    u'resultCurveTimeToTime',
    u'resultCurveTimeToUnitless')),
  (u'resultCurveTimeToAngular',
   (u'resultCurve', u'animCurve', 'dependNode'),
   ()),
  (u'resultCurveTimeToLinear',
   (u'resultCurve', u'animCurve', 'dependNode'),
   ()),
  (u'resultCurveTimeToTime',
   (u'resultCurve', u'animCurve', 'dependNode'),
   ()),
  (u'resultCurveTimeToUnitless',
   (u'resultCurve', u'animCurve', 'dependNode'),
   ()),
  (u'arrayMapper', ('dependNode',), (u'THarrayMapper',)),
  (u'THarrayMapper', (u'arrayMapper', 'dependNode'), ()),
  (u'attrHierarchyTest', ('dependNode',), ()),
  (u'audio', ('dependNode',), ()),
  (u'blend',
   ('dependNode',),
   (u'blendDevice', u'blendTwoAttr', u'blendWeighted')),
  (u'blendDevice', (u'blend', 'dependNode'), ()),
  (u'blendTwoAttr', (u'blend', 'dependNode'), ()),
  (u'blendWeighted', (u'blend', 'dependNode'), ()),
  (u'blendColors', ('dependNode',), ()),
  (u'blindDataTemplate', ('dependNode',), (u'subdHierBlind',)),
  (u'subdHierBlind', (u'blindDataTemplate', 'dependNode'), ()),
  (u'brush', ('dependNode',), ()),
  (u'bump3d', ('dependNode',), ()),
  (u'cacheBase', ('dependNode',), (u'cacheBlend', u'cacheFile')),
  (u'cacheBlend', (u'cacheBase', 'dependNode'), ()),
  (u'cacheFile', (u'cacheBase', 'dependNode'), ()),
  (u'cameraSet', ('dependNode',), (u'THcameraSet',)),
  (u'THcameraSet', (u'cameraSet', 'dependNode'), ()),
  (u'cameraView', ('dependNode',), ()),
  (u'characterMap', ('dependNode',), ()),
  (u'characterOffset', ('dependNode',), ()),
  (u'choice', ('dependNode',), ()),
  (u'chooser', ('dependNode',), ()),
  (u'clamp', ('dependNode',), ()),
  (u'clipLibrary', ('dependNode',), ()),
  (u'clipScheduler', ('dependNode',), ()),
  (u'clipToGhostData', ('dependNode',), ()),
  (u'closestPointOnMesh', ('dependNode',), ()),
  (u'colorManagementGlobals', ('dependNode',), ()),
  (u'colorProfile', ('dependNode',), ()),
  (u'combinationShape', ('dependNode',), ()),
  (u'compactPlugArrayTest', ('dependNode',), ()),
  (u'condition', ('dependNode',), ()),
  (u'containerBase', ('dependNode',), (u'container', u'entity')),
  (u'container', (u'containerBase', 'dependNode'), ()),
  (u'entity',
   (u'containerBase', 'dependNode'),
   (u'dagNode', u'objectSet', u'partition')),
  (u'dagNode',
   (u'entity', u'containerBase', 'dependNode'),
   (u'shape', u'transform', u'unknownDag')),
  (u'shape',
   (u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'baseLattice',
    u'camera',
    u'clusterHandle',
    u'deformFunc',
    u'dimensionShape',
    u'dynHolder',
    u'dynamicConstraint',
    u'flexorShape',
    u'follicle',
    u'geoConnectable',
    u'geometryShape',
    u'hairConstraint',
    u'hairSystem',
    u'imagePlane',
    u'light',
    u'lineModifier',
    u'pfxGeometry',
    u'rigidBody',
    u'softModHandle',
    u'spring')),
  (u'baseLattice',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'camera',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'stereoRigCamera',)),
  (u'stereoRigCamera',
   (u'camera',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'clusterHandle',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'deformFunc',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'deformBend',
    u'deformFlare',
    u'deformSine',
    u'deformSquash',
    u'deformTwist',
    u'deformWave')),
  (u'deformBend',
   (u'deformFunc',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'deformFlare',
   (u'deformFunc',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'deformSine',
   (u'deformFunc',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'deformSquash',
   (u'deformFunc',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'deformTwist',
   (u'deformFunc',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'deformWave',
   (u'deformFunc',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'dimensionShape',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'angleDimension',
    u'annotationShape',
    u'distanceDimShape',
    u'nurbsDimShape')),
  (u'angleDimension',
   (u'dimensionShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'annotationShape',
   (u'dimensionShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'distanceDimShape',
   (u'dimensionShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'nurbsDimShape',
   (u'dimensionShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'arcLengthDimension', u'paramDimension')),
  (u'arcLengthDimension',
   (u'nurbsDimShape',
    u'dimensionShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'paramDimension',
   (u'nurbsDimShape',
    u'dimensionShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'dynHolder',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'dynamicConstraint',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'flexorShape',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'clusterFlexorShape',)),
  (u'clusterFlexorShape',
   (u'flexorShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'follicle',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'geoConnectable',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'geometryShape',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'deformableShape',
    u'directedDisc',
    u'environmentFog',
    u'implicitBox',
    u'implicitCone',
    u'implicitSphere',
    u'locator',
    u'plane',
    u'renderRect')),
  (u'deformableShape',
   (u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'controlPoint', u'particle')),
  (u'controlPoint',
   (u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'curveShape', u'lattice', u'snapshotShape', u'surfaceShape')),
  (u'curveShape',
   (u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'nurbsCurve',)),
  (u'nurbsCurve',
   (u'curveShape',
    u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'bezierCurve',)),
  (u'bezierCurve',
   (u'nurbsCurve',
    u'curveShape',
    u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'lattice',
   (u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'snapshotShape',
   (u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'motionTrailShape',)),
  (u'motionTrailShape',
   (u'snapshotShape',
    u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'surfaceShape',
   (u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'THsurfaceShape',
    u'fluidShape',
    u'heightField',
    u'mesh',
    u'nurbsSurface',
    u'subdiv')),
  (u'THsurfaceShape',
   (u'surfaceShape',
    u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'fluidShape',
   (u'surfaceShape',
    u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'fluidTexture2D', u'fluidTexture3D')),
  (u'fluidTexture2D',
   (u'fluidShape',
    u'surfaceShape',
    u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'fluidTexture3D',
   (u'fluidShape',
    u'surfaceShape',
    u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'heightField',
   (u'surfaceShape',
    u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'mesh',
   (u'surfaceShape',
    u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'greasePlaneRenderShape',)),
  (u'greasePlaneRenderShape',
   (u'mesh',
    u'surfaceShape',
    u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'nurbsSurface',
   (u'surfaceShape',
    u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'subdiv',
   (u'surfaceShape',
    u'controlPoint',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'particle',
   (u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'nBase',)),
  (u'nBase',
   (u'particle',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'nCloth', u'nParticle', u'nRigid')),
  (u'nCloth',
   (u'nBase',
    u'particle',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'nParticle',
   (u'nBase',
    u'particle',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'nRigid',
   (u'nBase',
    u'particle',
    u'deformableShape',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'directedDisc',
   (u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'environmentFog',
   (u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'implicitBox',
   (u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'renderBox',)),
  (u'renderBox',
   (u'implicitBox',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'implicitCone',
   (u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'renderCone',)),
  (u'renderCone',
   (u'implicitCone',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'implicitSphere',
   (u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'renderSphere',)),
  (u'renderSphere',
   (u'implicitSphere',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'locator',
   (u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'THlocatorShape',
    u'dropoffLocator',
    u'hikFloorContactMarker',
    u'positionMarker')),
  (u'THlocatorShape',
   (u'locator',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'dropoffLocator',
   (u'locator',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'hikFloorContactMarker',
   (u'locator',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'positionMarker',
   (u'locator',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'orientationMarker',)),
  (u'orientationMarker',
   (u'positionMarker',
    u'locator',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'plane',
   (u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'sketchPlane',)),
  (u'sketchPlane',
   (u'plane',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'groundPlane', u'orthoGrid')),
  (u'groundPlane',
   (u'sketchPlane',
    u'plane',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'orthoGrid',
   (u'sketchPlane',
    u'plane',
    u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'renderRect',
   (u'geometryShape',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'hairConstraint',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'hairSystem',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'imagePlane',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'THimagePlane', u'greasePlane')),
  (u'THimagePlane',
   (u'imagePlane',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'greasePlane',
   (u'imagePlane',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'light',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'TadskAssetInstanceNode_TlightShape', u'renderLight')),
  (u'TadskAssetInstanceNode_TlightShape',
   (u'light',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'renderLight',
   (u'light',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'ambientLight', u'nonAmbientLightShapeNode')),
  (u'ambientLight',
   (u'renderLight',
    u'light',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'nonAmbientLightShapeNode',
   (u'renderLight',
    u'light',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'nonExtendedLightShapeNode',)),
  (u'nonExtendedLightShapeNode',
   (u'nonAmbientLightShapeNode',
    u'renderLight',
    u'light',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'areaLight', u'directionalLight', u'pointLight', u'spotLight')),
  (u'areaLight',
   (u'nonExtendedLightShapeNode',
    u'nonAmbientLightShapeNode',
    u'renderLight',
    u'light',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'directionalLight',
   (u'nonExtendedLightShapeNode',
    u'nonAmbientLightShapeNode',
    u'renderLight',
    u'light',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'pointLight',
   (u'nonExtendedLightShapeNode',
    u'nonAmbientLightShapeNode',
    u'renderLight',
    u'light',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'volumeLight',)),
  (u'volumeLight',
   (u'pointLight',
    u'nonExtendedLightShapeNode',
    u'nonAmbientLightShapeNode',
    u'renderLight',
    u'light',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'spotLight',
   (u'nonExtendedLightShapeNode',
    u'nonAmbientLightShapeNode',
    u'renderLight',
    u'light',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'lineModifier',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'pfxGeometry',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'pfxHair', u'pfxToon', u'stroke')),
  (u'pfxHair',
   (u'pfxGeometry',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'pfxToon',
   (u'pfxGeometry',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'stroke',
   (u'pfxGeometry',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'rigidBody',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'softModHandle',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'spring',
   (u'shape', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'THdynSpring',)),
  (u'THdynSpring',
   (u'spring',
    u'shape',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'transform',
   (u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'THcustomTransform',
    u'baseGeometryVarGroup',
    u'clipGhostShape',
    u'constraint',
    u'dagContainer',
    u'dynBase',
    u'fosterParent',
    u'hikEffector',
    u'hikGroundPlane',
    u'hikIKEffector',
    u'ikEffector',
    u'ikHandle',
    u'instancer',
    u'joint',
    u'lodGroup',
    u'nucleus',
    u'place3dTexture',
    u'rigidConstraint',
    u'textureDeformerHandle',
    u'ufeProxyTransform',
    u'unknownTransform')),
  (u'THcustomTransform',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'baseGeometryVarGroup',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'curveVarGroup',
    u'geometryVarGroup',
    u'meshVarGroup',
    u'subdivSurfaceVarGroup',
    u'surfaceVarGroup')),
  (u'curveVarGroup',
   (u'baseGeometryVarGroup',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'geometryVarGroup',
   (u'baseGeometryVarGroup',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'meshVarGroup',
   (u'baseGeometryVarGroup',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'subdivSurfaceVarGroup',
   (u'baseGeometryVarGroup',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'surfaceVarGroup',
   (u'baseGeometryVarGroup',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'clipGhostShape',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'constraint',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'THconstraint',
    u'aimConstraint',
    u'geometryConstraint',
    u'normalConstraint',
    u'oldNormalConstraint',
    u'oldTangentConstraint',
    u'orientConstraint',
    u'parentConstraint',
    u'pointConstraint',
    u'pointOnPolyConstraint',
    u'scaleConstraint',
    u'symmetryConstraint',
    u'tangentConstraint')),
  (u'THconstraint',
   (u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'aimConstraint',
   (u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'lookAt',)),
  (u'lookAt',
   (u'aimConstraint',
    u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'geometryConstraint',
   (u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'normalConstraint',
   (u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'oldNormalConstraint',
   (u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'oldTangentConstraint',
   (u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'orientConstraint',
   (u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'parentConstraint',
   (u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'pointConstraint',
   (u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'poleVectorConstraint',)),
  (u'poleVectorConstraint',
   (u'pointConstraint',
    u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'pointOnPolyConstraint',
   (u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'scaleConstraint',
   (u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'symmetryConstraint',
   (u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'tangentConstraint',
   (u'constraint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'dagContainer',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'assembly',)),
  (u'assembly',
   (u'dagContainer',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'THassembly',)),
  (u'THassembly',
   (u'assembly',
    u'dagContainer',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'dynBase',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'collisionModel', u'field', u'pointEmitter')),
  (u'collisionModel',
   (u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'field',
   (u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'THdynField',
    u'airField',
    u'dragField',
    u'gravityField',
    u'newtonField',
    u'radialField',
    u'turbulenceField',
    u'uniformField',
    u'volumeAxisField',
    u'vortexField')),
  (u'THdynField',
   (u'field',
    u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'airField',
   (u'field',
    u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'dragField',
   (u'field',
    u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'gravityField',
   (u'field',
    u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'newtonField',
   (u'field',
    u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'radialField',
   (u'field',
    u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'turbulenceField',
   (u'field',
    u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'uniformField',
   (u'field',
    u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'volumeAxisField',
   (u'field',
    u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'vortexField',
   (u'field',
    u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'pointEmitter',
   (u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'THdynEmitter', u'fluidEmitter')),
  (u'THdynEmitter',
   (u'pointEmitter',
    u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'fluidEmitter',
   (u'pointEmitter',
    u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   (u'THfluidEmitter',)),
  (u'THfluidEmitter',
   (u'fluidEmitter',
    u'pointEmitter',
    u'dynBase',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'fosterParent',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'hikEffector',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'hikGroundPlane',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'hikIKEffector',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'ikEffector',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'ikHandle',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'hikHandle',)),
  (u'hikHandle',
   (u'ikHandle',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'instancer',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'joint',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   (u'hikFKJoint',)),
  (u'hikFKJoint',
   (u'joint',
    u'transform',
    u'dagNode',
    u'entity',
    u'containerBase',
    'dependNode'),
   ()),
  (u'lodGroup',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'nucleus',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'place3dTexture',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'rigidConstraint',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'textureDeformerHandle',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'ufeProxyTransform',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'unknownTransform',
   (u'transform', u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'unknownDag',
   (u'dagNode', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'objectSet',
   (u'entity', u'containerBase', 'dependNode'),
   (u'THobjectSet',
    u'animLayer',
    u'bakeSet',
    u'character',
    u'creaseSet',
    u'keyingGroup',
    u'shadingEngine')),
  (u'THobjectSet',
   (u'objectSet', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'animLayer',
   (u'objectSet', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'bakeSet',
   (u'objectSet', u'entity', u'containerBase', 'dependNode'),
   (u'textureBakeSet', u'vertexBakeSet')),
  (u'textureBakeSet',
   (u'bakeSet', u'objectSet', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'vertexBakeSet',
   (u'bakeSet', u'objectSet', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'character',
   (u'objectSet', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'creaseSet',
   (u'objectSet', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'keyingGroup',
   (u'objectSet', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'shadingEngine',
   (u'objectSet', u'entity', u'containerBase', 'dependNode'),
   ()),
  (u'partition', (u'entity', u'containerBase', 'dependNode'), ()),
  (u'contrast', ('dependNode',), ()),
  (u'controller', ('dependNode',), ()),
  (u'copyColorSet', ('dependNode',), ()),
  (u'copyUVSet', ('dependNode',), ()),
  (u'createColorSet', ('dependNode',), ()),
  (u'createUVSet', ('dependNode',), ()),
  (u'curveFromMesh',
   ('dependNode',),
   (u'curveFromMeshCoM', u'curveFromMeshEdge')),
  (u'curveFromMeshCoM', (u'curveFromMesh', 'dependNode'), ()),
  (u'curveFromMeshEdge', (u'curveFromMesh', 'dependNode'), ()),
  (u'curveNormalizer',
   ('dependNode',),
   (u'curveNormalizerAngle', u'curveNormalizerLinear')),
  (u'curveNormalizerAngle', (u'curveNormalizer', 'dependNode'), ()),
  (u'curveNormalizerLinear', (u'curveNormalizer', 'dependNode'), ()),
  (u'dagPose', ('dependNode',), ()),
  (u'dataBlockTest', ('dependNode',), ()),
  (u'defaultLightList', ('dependNode',), ()),
  (u'defaultRenderUtilityList', ('dependNode',), ()),
  (u'defaultRenderingList', ('dependNode',), ()),
  (u'defaultShaderList', ('dependNode',), ()),
  (u'defaultTextureList', ('dependNode',), ()),
  (u'deleteColorSet', ('dependNode',), ()),
  (u'deleteComponent', ('dependNode',), ()),
  (u'deleteUVSet', ('dependNode',), ()),
  (u'diskCache', ('dependNode',), ()),
  (u'displacementShader', ('dependNode',), ()),
  (u'displayLayer', ('dependNode',), ()),
  (u'displayLayerManager', ('dependNode',), ()),
  (u'distanceBetween', ('dependNode',), ()),
  (u'dof', ('dependNode',), ()),
  (u'dynController', ('dependNode',), ()),
  (u'dynGlobals', ('dependNode',), ()),
  (u'editMetadata', ('dependNode',), ()),
  (u'editsManager', ('dependNode',), ()),
  (u'expression', ('dependNode',), ()),
  (u'facade', ('dependNode',), (u'envFacade', u'materialFacade')),
  (u'envFacade', (u'facade', 'dependNode'), (u'AISEnvFacade',)),
  (u'AISEnvFacade', (u'envFacade', u'facade', 'dependNode'), ()),
  (u'materialFacade', (u'facade', 'dependNode'), ()),
  (u'flow', ('dependNode',), ()),
  (u'fourByFourMatrix', ('dependNode',), ()),
  (u'frameCache', ('dependNode',), ()),
  (u'gammaCorrect', ('dependNode',), ()),
  (u'geoConnector', ('dependNode',), ()),
  (u'geomBind', ('dependNode',), ()),
  (u'geometryFilter',
   ('dependNode',),
   (u'THgeometryFilter',
    u'blendShape',
    u'boneLattice',
    u'historySwitch',
    u'jointLattice',
    u'sculpt',
    u'skinCluster',
    u'tweak',
    u'weightGeometryFilter',
    u'wrap')),
  (u'THgeometryFilter', (u'geometryFilter', 'dependNode'), ()),
  (u'blendShape', (u'geometryFilter', 'dependNode'), (u'THblendShape',)),
  (u'THblendShape', (u'blendShape', u'geometryFilter', 'dependNode'), ()),
  (u'boneLattice', (u'geometryFilter', 'dependNode'), ()),
  (u'historySwitch', (u'geometryFilter', 'dependNode'), ()),
  (u'jointLattice', (u'geometryFilter', 'dependNode'), ()),
  (u'sculpt', (u'geometryFilter', 'dependNode'), ()),
  (u'skinCluster', (u'geometryFilter', 'dependNode'), (u'THskinCluster',)),
  (u'THskinCluster', (u'skinCluster', u'geometryFilter', 'dependNode'), ()),
  (u'tweak', (u'geometryFilter', 'dependNode'), ()),
  (u'weightGeometryFilter',
   (u'geometryFilter', 'dependNode'),
   (u'THdeformer',
    u'cluster',
    u'deltaMush',
    u'ffd',
    u'jiggle',
    u'nonLinear',
    u'shrinkWrap',
    u'softMod',
    u'tension',
    u'textureDeformer',
    u'transferAttributes',
    u'wire')),
  (u'THdeformer',
   (u'weightGeometryFilter', u'geometryFilter', 'dependNode'),
   ()),
  (u'cluster',
   (u'weightGeometryFilter', u'geometryFilter', 'dependNode'),
   (u'jointCluster',)),
  (u'jointCluster',
   (u'cluster', u'weightGeometryFilter', u'geometryFilter', 'dependNode'),
   ()),
  (u'deltaMush',
   (u'weightGeometryFilter', u'geometryFilter', 'dependNode'),
   ()),
  (u'ffd',
   (u'weightGeometryFilter', u'geometryFilter', 'dependNode'),
   (u'jointFfd',)),
  (u'jointFfd',
   (u'ffd', u'weightGeometryFilter', u'geometryFilter', 'dependNode'),
   ()),
  (u'jiggle', (u'weightGeometryFilter', u'geometryFilter', 'dependNode'), ()),
  (u'nonLinear',
   (u'weightGeometryFilter', u'geometryFilter', 'dependNode'),
   ()),
  (u'shrinkWrap',
   (u'weightGeometryFilter', u'geometryFilter', 'dependNode'),
   ()),
  (u'softMod',
   (u'weightGeometryFilter', u'geometryFilter', 'dependNode'),
   ()),
  (u'tension',
   (u'weightGeometryFilter', u'geometryFilter', 'dependNode'),
   ()),
  (u'textureDeformer',
   (u'weightGeometryFilter', u'geometryFilter', 'dependNode'),
   ()),
  (u'transferAttributes',
   (u'weightGeometryFilter', u'geometryFilter', 'dependNode'),
   ()),
  (u'wire', (u'weightGeometryFilter', u'geometryFilter', 'dependNode'), ()),
  (u'wrap', (u'geometryFilter', 'dependNode'), ()),
  (u'globalCacheControl', ('dependNode',), ()),
  (u'greasePencilSequence', ('dependNode',), ()),
  (u'groupId', ('dependNode',), ()),
  (u'groupParts', ('dependNode',), ()),
  (u'guide', ('dependNode',), ()),
  (u'hardwareRenderGlobals', ('dependNode',), ()),
  (u'hardwareRenderingGlobals', ('dependNode',), ()),
  (u'hierarchyTestNode1', ('dependNode',), (u'hierarchyTestNode2',)),
  (u'hierarchyTestNode2',
   (u'hierarchyTestNode1', 'dependNode'),
   (u'hierarchyTestNode3',)),
  (u'hierarchyTestNode3',
   (u'hierarchyTestNode2', u'hierarchyTestNode1', 'dependNode'),
   ()),
  (u'holdMatrix', ('dependNode',), ()),
  (u'hsvToRgb', ('dependNode',), ()),
  (u'hwReflectionMap', ('dependNode',), ()),
  (u'hwRenderGlobals', ('dependNode',), ()),
  (u'hyperGraphInfo', ('dependNode',), ()),
  (u'hyperLayout', ('dependNode',), ()),
  (u'hyperView', ('dependNode',), ()),
  (u'ikSolver',
   ('dependNode',),
   (u'THikSolverNode',
    u'hikSolver',
    u'ikMCsolver',
    u'ikPASolver',
    u'ikSCsolver',
    u'ikSplineSolver')),
  (u'THikSolverNode', (u'ikSolver', 'dependNode'), ()),
  (u'hikSolver', (u'ikSolver', 'dependNode'), ()),
  (u'ikMCsolver', (u'ikSolver', 'dependNode'), ()),
  (u'ikPASolver', (u'ikSolver', 'dependNode'), ()),
  (u'ikSCsolver', (u'ikSolver', 'dependNode'), (u'ikRPsolver',)),
  (u'ikRPsolver', (u'ikSCsolver', u'ikSolver', 'dependNode'), ()),
  (u'ikSplineSolver', (u'ikSolver', 'dependNode'), ()),
  (u'ikSystem', ('dependNode',), ()),
  (u'imageSource',
   ('dependNode',),
   (u'renderLayer', u'renderPass', u'renderTarget', u'renderedImageSource')),
  (u'renderLayer', (u'imageSource', 'dependNode'), ()),
  (u'renderPass', (u'imageSource', 'dependNode'), ()),
  (u'renderTarget', (u'imageSource', 'dependNode'), ()),
  (u'renderedImageSource', (u'imageSource', 'dependNode'), ()),
  (u'lightInfo', ('dependNode',), ()),
  (u'lightLinker', ('dependNode',), ()),
  (u'lightList', ('dependNode',), ()),
  (u'lodThresholds', ('dependNode',), ()),
  (u'luminance', ('dependNode',), ()),
  (u'makeGroup', ('dependNode',), ()),
  (u'materialInfo', ('dependNode',), ()),
  (u'membrane', ('dependNode',), ()),
  (u'motionPath', ('dependNode',), (u'THmotionPath',)),
  (u'THmotionPath', (u'motionPath', 'dependNode'), ()),
  (u'multDoubleLinear', ('dependNode',), ()),
  (u'multMatrix', ('dependNode',), ()),
  (u'multilisterLight', ('dependNode',), ()),
  (u'mute', ('dependNode',), ()),
  (u'nComponent', ('dependNode',), ()),
  (u'network', ('dependNode',), ()),
  (u'nodeGraphEditorBookmarkInfo', ('dependNode',), ()),
  (u'nodeGraphEditorBookmarks', ('dependNode',), ()),
  (u'nodeGraphEditorInfo', ('dependNode',), ()),
  (u'objectFilter',
   ('dependNode',),
   (u'objectAttrFilter',
    u'objectBinFilter',
    u'objectMultiFilter',
    u'objectNameFilter',
    u'objectRenderFilter',
    u'objectScriptFilter',
    u'objectTypeFilter')),
  (u'objectAttrFilter', (u'objectFilter', 'dependNode'), ()),
  (u'objectBinFilter', (u'objectFilter', 'dependNode'), ()),
  (u'objectMultiFilter', (u'objectFilter', 'dependNode'), ()),
  (u'objectNameFilter', (u'objectFilter', 'dependNode'), ()),
  (u'objectRenderFilter', (u'objectFilter', 'dependNode'), ()),
  (u'objectScriptFilter', (u'objectFilter', 'dependNode'), ()),
  (u'objectTypeFilter', (u'objectFilter', 'dependNode'), ()),
  (u'oldBlindDataBase', ('dependNode',), ()),
  (u'oldGeometryConstraint', ('dependNode',), ()),
  (u'opticalFX', ('dependNode',), ()),
  (u'pairBlend', ('dependNode',), ()),
  (u'particleAgeMapper', ('dependNode',), ()),
  (u'particleColorMapper', ('dependNode',), ()),
  (u'particleIncandMapper', ('dependNode',), ()),
  (u'particleTranspMapper', ('dependNode',), ()),
  (u'passContributionMap', ('dependNode',), ()),
  (u'passMatrix', ('dependNode',), ()),
  (u'pointMatrixMult', ('dependNode',), ()),
  (u'polyBase', ('dependNode',), (u'polyCreator', u'polyModifier')),
  (u'polyCreator',
   (u'polyBase', 'dependNode'),
   (u'polyCreateFace', u'polyPrimitive', u'polyUnite')),
  (u'polyCreateFace', (u'polyCreator', u'polyBase', 'dependNode'), ()),
  (u'polyPrimitive',
   (u'polyCreator', u'polyBase', 'dependNode'),
   (u'polyCone',
    u'polyCube',
    u'polyCylinder',
    u'polyHelix',
    u'polyPipe',
    u'polyPlane',
    u'polyPlatonicSolid',
    u'polyPrimitiveMisc',
    u'polyPrism',
    u'polyPyramid',
    u'polySphere',
    u'polyTorus')),
  (u'polyCone',
   (u'polyPrimitive', u'polyCreator', u'polyBase', 'dependNode'),
   ()),
  (u'polyCube',
   (u'polyPrimitive', u'polyCreator', u'polyBase', 'dependNode'),
   ()),
  (u'polyCylinder',
   (u'polyPrimitive', u'polyCreator', u'polyBase', 'dependNode'),
   ()),
  (u'polyHelix',
   (u'polyPrimitive', u'polyCreator', u'polyBase', 'dependNode'),
   ()),
  (u'polyPipe',
   (u'polyPrimitive', u'polyCreator', u'polyBase', 'dependNode'),
   ()),
  (u'polyPlane',
   (u'polyPrimitive', u'polyCreator', u'polyBase', 'dependNode'),
   ()),
  (u'polyPlatonicSolid',
   (u'polyPrimitive', u'polyCreator', u'polyBase', 'dependNode'),
   ()),
  (u'polyPrimitiveMisc',
   (u'polyPrimitive', u'polyCreator', u'polyBase', 'dependNode'),
   ()),
  (u'polyPrism',
   (u'polyPrimitive', u'polyCreator', u'polyBase', 'dependNode'),
   ()),
  (u'polyPyramid',
   (u'polyPrimitive', u'polyCreator', u'polyBase', 'dependNode'),
   ()),
  (u'polySphere',
   (u'polyPrimitive', u'polyCreator', u'polyBase', 'dependNode'),
   ()),
  (u'polyTorus',
   (u'polyPrimitive', u'polyCreator', u'polyBase', 'dependNode'),
   ()),
  (u'polyUnite',
   (u'polyCreator', u'polyBase', 'dependNode'),
   (u'polyBoolOp',)),
  (u'polyBoolOp',
   (u'polyUnite', u'polyCreator', u'polyBase', 'dependNode'),
   (u'polyCBoolOp',)),
  (u'polyCBoolOp',
   (u'polyBoolOp', u'polyUnite', u'polyCreator', u'polyBase', 'dependNode'),
   ()),
  (u'polyModifier',
   (u'polyBase', 'dependNode'),
   (u'blendColorSets',
    u'polyAppend',
    u'polyAppendVertex',
    u'polyClean',
    u'polyCloseBorder',
    u'polyCollapseEdge',
    u'polyCollapseF',
    u'polyColorDel',
    u'polyColorMod',
    u'polyColorPerVertex',
    u'polyConnectComponents',
    u'polyDelEdge',
    u'polyDelFacet',
    u'polyDelVertex',
    u'polyDuplicateEdge',
    u'polyEditEdgeFlow',
    u'polyFlipEdge',
    u'polyMapCut',
    u'polyMapDel',
    u'polyMapSew',
    u'polyMergeEdge',
    u'polyMergeFace',
    u'polyModifierWorld',
    u'polyMoveUV',
    u'polyNormal',
    u'polyNormalPerVertex',
    u'polyReduce',
    u'polyRetopo',
    u'polySmooth',
    u'polySmoothFace',
    u'polySmoothProxy',
    u'polySpinEdge',
    u'polySplit',
    u'polySplitEdge',
    u'polySubdFace',
    u'polyTransfer',
    u'polyTriangulate',
    u'polyTweak',
    u'polyTweakUV',
    u'polyUVRectangle')),
  (u'blendColorSets', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyAppend', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyAppendVertex', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyClean', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyCloseBorder', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyCollapseEdge', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyCollapseF', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyColorDel', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyColorMod', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyColorPerVertex', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyConnectComponents',
   (u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyDelEdge', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyDelFacet', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyDelVertex', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyDuplicateEdge', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyEditEdgeFlow', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyFlipEdge', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyMapCut', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyMapDel', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyMapSew', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyMergeEdge', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyMergeFace', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyModifierWorld',
   (u'polyModifier', u'polyBase', 'dependNode'),
   (u'polyAverageVertex',
    u'polyBevel',
    u'polyBevel3',
    u'polyBridgeEdge',
    u'polyCircularize',
    u'polyCrease',
    u'polyCut',
    u'polyCylProj',
    u'polyExtrudeVertex',
    u'polyHoleFace',
    u'polyMergeVert',
    u'polyMirror',
    u'polyModifierUV',
    u'polyMoveVertex',
    u'polyPassThru',
    u'polyPlanarProj',
    u'polyPoke',
    u'polyProj',
    u'polyQuad',
    u'polyRemesh',
    u'polySewEdge',
    u'polySoftEdge',
    u'polySphProj',
    u'polySplitRing',
    u'polySplitVert',
    u'polySubdEdge',
    u'polyWedgeFace')),
  (u'polyAverageVertex',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyBevel',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   (u'polyBevel2',)),
  (u'polyBevel2',
   (u'polyBevel',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyBevel3',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyBridgeEdge',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyCircularize',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyCrease',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   (u'polyCreaseEdge',)),
  (u'polyCreaseEdge',
   (u'polyCrease',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyCut',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyCylProj',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyExtrudeVertex',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyHoleFace',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyMergeVert',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyMirror',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyModifierUV',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   (u'polyAutoProj',
    u'polyContourProj',
    u'polyCopyUV',
    u'polyFlipUV',
    u'polyLayoutUV',
    u'polyMapSewMove',
    u'polyMergeUV',
    u'polyNormalizeUV',
    u'polyOptUvs',
    u'polyPinUV',
    u'polyStraightenUVBorder')),
  (u'polyAutoProj',
   (u'polyModifierUV',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyContourProj',
   (u'polyModifierUV',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyCopyUV',
   (u'polyModifierUV',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyFlipUV',
   (u'polyModifierUV',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyLayoutUV',
   (u'polyModifierUV',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyMapSewMove',
   (u'polyModifierUV',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyMergeUV',
   (u'polyModifierUV',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyNormalizeUV',
   (u'polyModifierUV',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyOptUvs',
   (u'polyModifierUV',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyPinUV',
   (u'polyModifierUV',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyStraightenUVBorder',
   (u'polyModifierUV',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyMoveVertex',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   (u'polyMoveEdge',)),
  (u'polyMoveEdge',
   (u'polyMoveVertex',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   (u'polyExtrudeEdge', u'polyMoveFace')),
  (u'polyExtrudeEdge',
   (u'polyMoveEdge',
    u'polyMoveVertex',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyMoveFace',
   (u'polyMoveEdge',
    u'polyMoveVertex',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   (u'polyChipOff', u'polyExtrudeFace')),
  (u'polyChipOff',
   (u'polyMoveFace',
    u'polyMoveEdge',
    u'polyMoveVertex',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyExtrudeFace',
   (u'polyMoveFace',
    u'polyMoveEdge',
    u'polyMoveVertex',
    u'polyModifierWorld',
    u'polyModifier',
    u'polyBase',
    'dependNode'),
   ()),
  (u'polyPassThru',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyPlanarProj',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyPoke',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyProj',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyQuad',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyRemesh',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polySewEdge',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polySoftEdge',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polySphProj',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polySplitRing',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polySplitVert',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polySubdEdge',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyWedgeFace',
   (u'polyModifierWorld', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyMoveUV',
   (u'polyModifier', u'polyBase', 'dependNode'),
   (u'polyMoveFacetUV',)),
  (u'polyMoveFacetUV',
   (u'polyMoveUV', u'polyModifier', u'polyBase', 'dependNode'),
   ()),
  (u'polyNormal', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyNormalPerVertex', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyReduce', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyRetopo', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polySmooth', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polySmoothFace', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polySmoothProxy', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polySpinEdge', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polySplit', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polySplitEdge', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polySubdFace', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyTransfer', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyTriangulate', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyTweak', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyTweakUV', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyUVRectangle', (u'polyModifier', u'polyBase', 'dependNode'), ()),
  (u'polyBlindData', ('dependNode',), (u'subdBlindData',)),
  (u'subdBlindData', (u'polyBlindData', 'dependNode'), ()),
  (u'polySeparate', ('dependNode',), ()),
  (u'polyToSubdiv', ('dependNode',), ()),
  (u'poseInterpolatorManager', ('dependNode',), ()),
  (u'postProcessList', ('dependNode',), ()),
  (u'precompExport', ('dependNode',), ()),
  (u'proxyManager', ('dependNode',), ()),
  (u'record', ('dependNode',), ()),
  (u'reference', ('dependNode',), ()),
  (u'remapColor', ('dependNode',), ()),
  (u'remapHsv', ('dependNode',), ()),
  (u'remapValue', ('dependNode',), ()),
  (u'renderGlobals', ('dependNode',), ()),
  (u'renderGlobalsList', ('dependNode',), ()),
  (u'renderLayerManager', ('dependNode',), ()),
  (u'renderPassSet', ('dependNode',), ()),
  (u'renderQuality', ('dependNode',), ()),
  (u'reorderUVSet', ('dependNode',), ()),
  (u'resolution', ('dependNode',), ()),
  (u'reverse', ('dependNode',), ()),
  (u'rgbToHsv', ('dependNode',), ()),
  (u'rigidSolver', ('dependNode',), ()),
  (u'sampler', ('dependNode',), ()),
  (u'script', ('dependNode',), ()),
  (u'selectionListOperator', ('dependNode',), ()),
  (u'sequenceManager', ('dependNode',), ()),
  (u'sequencer', ('dependNode',), ()),
  (u'setRange', ('dependNode',), ()),
  (u'shaderGlow', ('dependNode',), ()),
  (u'shadingDependNode',
   ('dependNode',),
   (u'baseShadingSwitch',
    u'bump2d',
    u'hwShader',
    u'lambert',
    u'layeredShader',
    u'layeredTexture',
    u'lightFog',
    u'multiplyDivide',
    u'oceanShader',
    u'particleCloud',
    u'particleSamplerInfo',
    u'place2dTexture',
    u'plusMinusAverage',
    u'rampShader',
    u'samplerInfo',
    u'shadingMap',
    u'texture2d',
    u'texture3d',
    u'textureEnv',
    u'useBackground',
    u'vectorProduct',
    u'volumeFog')),
  (u'baseShadingSwitch',
   (u'shadingDependNode', 'dependNode'),
   (u'doubleShadingSwitch',
    u'quadShadingSwitch',
    u'singleShadingSwitch',
    u'tripleShadingSwitch')),
  (u'doubleShadingSwitch',
   (u'baseShadingSwitch', u'shadingDependNode', 'dependNode'),
   ()),
  (u'quadShadingSwitch',
   (u'baseShadingSwitch', u'shadingDependNode', 'dependNode'),
   ()),
  (u'singleShadingSwitch',
   (u'baseShadingSwitch', u'shadingDependNode', 'dependNode'),
   ()),
  (u'tripleShadingSwitch',
   (u'baseShadingSwitch', u'shadingDependNode', 'dependNode'),
   ()),
  (u'bump2d', (u'shadingDependNode', 'dependNode'), ()),
  (u'hwShader',
   (u'shadingDependNode', 'dependNode'),
   (u'THhardwareShader', u'THhwShader')),
  (u'THhardwareShader',
   (u'hwShader', u'shadingDependNode', 'dependNode'),
   ()),
  (u'THhwShader', (u'hwShader', u'shadingDependNode', 'dependNode'), ()),
  (u'lambert', (u'shadingDependNode', 'dependNode'), (u'reflect',)),
  (u'reflect',
   (u'lambert', u'shadingDependNode', 'dependNode'),
   (u'anisotropic', u'blinn', u'hairTubeShader', u'phong', u'phongE')),
  (u'anisotropic',
   (u'reflect', u'lambert', u'shadingDependNode', 'dependNode'),
   ()),
  (u'blinn',
   (u'reflect', u'lambert', u'shadingDependNode', 'dependNode'),
   ()),
  (u'hairTubeShader',
   (u'reflect', u'lambert', u'shadingDependNode', 'dependNode'),
   ()),
  (u'phong',
   (u'reflect', u'lambert', u'shadingDependNode', 'dependNode'),
   ()),
  (u'phongE',
   (u'reflect', u'lambert', u'shadingDependNode', 'dependNode'),
   ()),
  (u'layeredShader', (u'shadingDependNode', 'dependNode'), ()),
  (u'layeredTexture', (u'shadingDependNode', 'dependNode'), ()),
  (u'lightFog', (u'shadingDependNode', 'dependNode'), (u'envFog',)),
  (u'envFog', (u'lightFog', u'shadingDependNode', 'dependNode'), ()),
  (u'multiplyDivide', (u'shadingDependNode', 'dependNode'), ()),
  (u'oceanShader', (u'shadingDependNode', 'dependNode'), ()),
  (u'particleCloud', (u'shadingDependNode', 'dependNode'), ()),
  (u'particleSamplerInfo', (u'shadingDependNode', 'dependNode'), ()),
  (u'place2dTexture', (u'shadingDependNode', 'dependNode'), ()),
  (u'plusMinusAverage', (u'shadingDependNode', 'dependNode'), ()),
  (u'rampShader', (u'shadingDependNode', 'dependNode'), ()),
  (u'samplerInfo', (u'shadingDependNode', 'dependNode'), ()),
  (u'shadingMap', (u'shadingDependNode', 'dependNode'), ()),
  (u'texture2d',
   (u'shadingDependNode', 'dependNode'),
   (u'TadskAssetInstanceNode_TdnTx2D',
    u'bulge',
    u'checker',
    u'cloth',
    u'file',
    u'fractal',
    u'grid',
    u'mandelbrot',
    u'mountain',
    u'noise',
    u'ocean',
    u'ramp',
    u'stencil',
    u'water')),
  (u'TadskAssetInstanceNode_TdnTx2D',
   (u'texture2d', u'shadingDependNode', 'dependNode'),
   ()),
  (u'bulge', (u'texture2d', u'shadingDependNode', 'dependNode'), ()),
  (u'checker', (u'texture2d', u'shadingDependNode', 'dependNode'), ()),
  (u'cloth', (u'texture2d', u'shadingDependNode', 'dependNode'), ()),
  (u'file',
   (u'texture2d', u'shadingDependNode', 'dependNode'),
   (u'movie', u'psdFileTex')),
  (u'movie', (u'file', u'texture2d', u'shadingDependNode', 'dependNode'), ()),
  (u'psdFileTex',
   (u'file', u'texture2d', u'shadingDependNode', 'dependNode'),
   ()),
  (u'fractal', (u'texture2d', u'shadingDependNode', 'dependNode'), ()),
  (u'grid', (u'texture2d', u'shadingDependNode', 'dependNode'), ()),
  (u'mandelbrot', (u'texture2d', u'shadingDependNode', 'dependNode'), ()),
  (u'mountain', (u'texture2d', u'shadingDependNode', 'dependNode'), ()),
  (u'noise', (u'texture2d', u'shadingDependNode', 'dependNode'), ()),
  (u'ocean', (u'texture2d', u'shadingDependNode', 'dependNode'), ()),
  (u'ramp', (u'texture2d', u'shadingDependNode', 'dependNode'), ()),
  (u'stencil', (u'texture2d', u'shadingDependNode', 'dependNode'), ()),
  (u'water', (u'texture2d', u'shadingDependNode', 'dependNode'), ()),
  (u'texture3d',
   (u'shadingDependNode', 'dependNode'),
   (u'brownian',
    u'cloud',
    u'crater',
    u'granite',
    u'leather',
    u'mandelbrot3D',
    u'marble',
    u'projection',
    u'rock',
    u'snow',
    u'solidFractal',
    u'stucco',
    u'volumeNoise',
    u'wood')),
  (u'brownian', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'cloud', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'crater', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'granite', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'leather', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'mandelbrot3D', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'marble', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'projection', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'rock', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'snow', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'solidFractal', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'stucco', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'volumeNoise', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'wood', (u'texture3d', u'shadingDependNode', 'dependNode'), ()),
  (u'textureEnv',
   (u'shadingDependNode', 'dependNode'),
   (u'envBall', u'envChrome', u'envCube', u'envSky', u'envSphere')),
  (u'envBall', (u'textureEnv', u'shadingDependNode', 'dependNode'), ()),
  (u'envChrome', (u'textureEnv', u'shadingDependNode', 'dependNode'), ()),
  (u'envCube', (u'textureEnv', u'shadingDependNode', 'dependNode'), ()),
  (u'envSky', (u'textureEnv', u'shadingDependNode', 'dependNode'), ()),
  (u'envSphere', (u'textureEnv', u'shadingDependNode', 'dependNode'), ()),
  (u'useBackground', (u'shadingDependNode', 'dependNode'), ()),
  (u'vectorProduct', (u'shadingDependNode', 'dependNode'), ()),
  (u'volumeFog', (u'shadingDependNode', 'dependNode'), ()),
  (u'shapeEditorManager', ('dependNode',), ()),
  (u'shot', ('dependNode',), ()),
  (u'simpleTestNode', ('dependNode',), ()),
  (u'simpleVolumeShader', ('dependNode',), ()),
  (u'skinBinding', ('dependNode',), ()),
  (u'snapshot', ('dependNode',), (u'motionTrail',)),
  (u'motionTrail', (u'snapshot', 'dependNode'), ()),
  (u'strokeGlobals', ('dependNode',), ()),
  (u'subdBase', ('dependNode',), (u'subdModifier',)),
  (u'subdModifier',
   (u'subdBase', 'dependNode'),
   (u'subdAddTopology',
    u'subdCleanTopology',
    u'subdMapCut',
    u'subdModifierWorld',
    u'subdTweak',
    u'subdTweakUV')),
  (u'subdAddTopology', (u'subdModifier', u'subdBase', 'dependNode'), ()),
  (u'subdCleanTopology', (u'subdModifier', u'subdBase', 'dependNode'), ()),
  (u'subdMapCut', (u'subdModifier', u'subdBase', 'dependNode'), ()),
  (u'subdModifierWorld',
   (u'subdModifier', u'subdBase', 'dependNode'),
   (u'subdModifierUV', u'subdPlanarProj')),
  (u'subdModifierUV',
   (u'subdModifierWorld', u'subdModifier', u'subdBase', 'dependNode'),
   (u'subdAutoProj', u'subdLayoutUV', u'subdMapSewMove')),
  (u'subdAutoProj',
   (u'subdModifierUV',
    u'subdModifierWorld',
    u'subdModifier',
    u'subdBase',
    'dependNode'),
   ()),
  (u'subdLayoutUV',
   (u'subdModifierUV',
    u'subdModifierWorld',
    u'subdModifier',
    u'subdBase',
    'dependNode'),
   ()),
  (u'subdMapSewMove',
   (u'subdModifierUV',
    u'subdModifierWorld',
    u'subdModifier',
    u'subdBase',
    'dependNode'),
   ()),
  (u'subdPlanarProj',
   (u'subdModifierWorld', u'subdModifier', u'subdBase', 'dependNode'),
   ()),
  (u'subdTweak', (u'subdModifier', u'subdBase', 'dependNode'), ()),
  (u'subdTweakUV', (u'subdModifier', u'subdBase', 'dependNode'), ()),
  (u'subdivCollapse', ('dependNode',), ()),
  (u'subdivComponentId', ('dependNode',), ()),
  (u'subdivReverseFaces', ('dependNode',), ()),
  (u'subdivToPoly', ('dependNode',), ()),
  (u'surfaceLuminance', ('dependNode',), ()),
  (u'surfaceShader', ('dependNode',), ()),
  (u'texLattice', ('dependNode',), ()),
  (u'textureToGeom', ('dependNode',), ()),
  (u'threadedDevice', ('dependNode',), (u'clientDevice',)),
  (u'clientDevice', (u'threadedDevice', 'dependNode'), ()),
  (u'time', ('dependNode',), ()),
  (u'timeEditor', ('dependNode',), ()),
  (u'timeEditorAnimSource', ('dependNode',), ()),
  (u'timeEditorClipBase', ('dependNode',), (u'timeEditorClip',)),
  (u'timeEditorClip', (u'timeEditorClipBase', 'dependNode'), ()),
  (u'timeEditorClipEvaluator', ('dependNode',), ()),
  (u'timeEditorInterpolator', ('dependNode',), ()),
  (u'timeEditorTracks', ('dependNode',), ()),
  (u'timeFunction', ('dependNode',), ()),
  (u'timeToUnitConversion', ('dependNode',), ()),
  (u'timeWarp', ('dependNode',), ()),
  (u'toonLineAttributes', ('dependNode',), ()),
  (u'trackInfoManager', ('dependNode',), ()),
  (u'unitConversion', ('dependNode',), ()),
  (u'unitToTimeConversion', ('dependNode',), ()),
  (u'unknown', ('dependNode',), ()),
  (u'uvChooser', ('dependNode',), ()),
  (u'viewColorManager', ('dependNode',), ()),
  (u'volumeShader', ('dependNode',), ()),
  (u'wtAddMatrix', ('dependNode',), ())],
 ['attrColorSliderGrp',
  'attrControlGrp',
  'attrEnumOptionMenu',
  'attrEnumOptionMenuGrp',
  'attrFieldGrp',
  'attrFieldSliderGrp',
  'attrNavigationControlGrp',
  'attributeMenu',
  'colorIndexSliderGrp',
  'colorSliderButtonGrp',
  'colorSliderGrp',
  'columnLayout',
  'colorEditor',
  'floatField',
  'floatFieldGrp',
  'floatScrollBar',
  'floatSlider',
  'floatSlider2',
  'floatSliderButtonGrp',
  'floatSliderGrp',
  'frameLayout',
  'iconTextButton',
  'iconTextCheckBox',
  'iconTextRadioButton',
  'iconTextRadioCollection',
  'iconTextScrollList',
  'iconTextStaticLabel',
  'intField',
  'intFieldGrp',
  'intScrollBar',
  'intSlider',
  'intSliderGrp',
  'paneLayout',
  'panel',
  'radioButton',
  'radioButtonGrp',
  'radioCollection',
  'radioMenuItemCollection',
  'symbolButton',
  'symbolCheckBox',
  'textCurves',
  'textField',
  'textFieldButtonGrp',
  'textFieldGrp',
  'text',
  'textScrollList',
  'toolButton',
  'toolCollection',
  'window',
  'blendShapeEditor',
  'blendShapePanel',
  'button',
  'checkBox',
  'checkBoxGrp',
  'confirmDialog',
  'fontDialog',
  'formLayout',
  'menu',
  'menuBarLayout',
  'menuEditor',
  'menuItem',
  'menuSet',
  'promptDialog',
  'scrollField',
  'scrollLayout',
  'scriptedPanel',
  'scriptedPanelType',
  'shelfButton',
  'shelfLayout',
  'shelfTabLayout',
  'tabLayout',
  'outlinerEditor',
  'optionMenu',
  'outlinerPanel',
  'optionMenuGrp',
  'animCurveEditor',
  'animDisplay',
  'separator',
  'visor',
  'layout',
  'layoutDialog',
  'layerButton',
  'hyperGraph',
  'hyperPanel',
  'hyperShade',
  'rowColumnLayout',
  'rowLayout',
  'renderLayerButton',
  'renderWindowEditor',
  'glRenderEditor',
  'scriptTable',
  'keyframeStats',
  'keyframeOutliner',
  'canvas',
  'channelBox',
  'gradientControl',
  'gradientControlNoAttr',
  'gridLayout',
  'messageLine',
  'popupMenu',
  'modelEditor',
  'modelPanel',
  'helpLine',
  'hardwareRenderPanel',
  'image',
  'nodeIconButton',
  'commandLine',
  'progressBar',
  'defaultLightListCheckBox',
  'exclusiveLightCheckBox',
  'shellField',
  'clipSchedulerOutliner',
  'clipEditor',
  'deviceEditor',
  'devicePanel',
  'dynRelEdPanel',
  'dynRelEditor',
  'dynPaintEditor',
  'nameField',
  'cmdScrollFieldExecuter',
  'cmdScrollFieldReporter',
  'cmdShell',
  'nameField',
  'palettePort'],
 ['aimConstraint',
  'air',
  'alignCurve',
  'alignSurface',
  'ambientLight',
  'angleBetween',
  'animLayer',
  'arcLengthDimension',
  'arrayMapper',
  'assembly',
  'attachCurve',
  'attachSurface',
  'bevel',
  'bevelPlus',
  'bezierCurveToNurbs',
  'blend',
  'blendShape',
  'blendTwoAttr',
  'boneLattice',
  'boundary',
  'cacheFile',
  'camera',
  'cameraSet',
  'cameraView',
  'character',
  'characterMap',
  'choice',
  'circle',
  'closeCurve',
  'closeSurface',
  'cluster',
  'combinationShape',
  'condition',
  'cone',
  'container',
  'controller',
  'curve',
  'curveIntersect',
  'cylinder',
  'dagPose',
  'deltaMush',
  'detachCurve',
  'detachSurface',
  'directionalLight',
  'diskCache',
  'drag',
  'dropoffLocator',
  'dynGlobals',
  'editMetadata',
  'emitter',
  'expression',
  'extendCurve',
  'extendSurface',
  'extrude',
  'file',
  'filletCurve',
  'fitBspline',
  'flow',
  'fluidEmitter',
  'geomBind',
  'geometryConstraint',
  'globalStitch',
  'gravity',
  'grid',
  'groupParts',
  'hwReflectionMap',
  'ikHandle',
  'ikSolver',
  'ikSystem',
  'imagePlane',
  'insertKnotCurve',
  'insertKnotSurface',
  'instancer',
  'joint',
  'jointCluster',
  'jointLattice',
  'keyingGroup',
  'lattice',
  'lightList',
  'loft',
  'mute',
  'nBase',
  'nParticle',
  'nonLinear',
  'normalConstraint',
  'nurbsCube',
  'nurbsCurveToBezier',
  'nurbsPlane',
  'nurbsSquare',
  'nurbsToSubdiv',
  'offsetCurve',
  'offsetSurface',
  'orientConstraint',
  'pairBlend',
  'paramDimension',
  'parentConstraint',
  'particle',
  'partition',
  'plane',
  'pointConstraint',
  'pointLight',
  'pointOnPolyConstraint',
  'poleVectorConstraint',
  'polyAppend',
  'polyAppendVertex',
  'polyAverageVertex',
  'polyBevel',
  'polyBevel3',
  'polyBlindData',
  'polyBoolOp',
  'polyBridgeEdge',
  'polyCBoolOp',
  'polyChipOff',
  'polyCircularize',
  'polyClean',
  'polyCloseBorder',
  'polyCollapseEdge',
  'polyColorDel',
  'polyColorMod',
  'polyColorPerVertex',
  'polyCone',
  'polyConnectComponents',
  'polyCopyUV',
  'polyCrease',
  'polyCube',
  'polyCut',
  'polyCylinder',
  'polyDelEdge',
  'polyDelFacet',
  'polyDelVertex',
  'polyDuplicateEdge',
  'polyEditEdgeFlow',
  'polyExtrudeEdge',
  'polyExtrudeVertex',
  'polyFlipEdge',
  'polyFlipUV',
  'polyHelix',
  'polyLayoutUV',
  'polyMapCut',
  'polyMapDel',
  'polyMapSew',
  'polyMapSewMove',
  'polyMergeEdge',
  'polyMergeUV',
  'polyMoveEdge',
  'polyMoveFacetUV',
  'polyMoveUV',
  'polyMoveVertex',
  'polyNormal',
  'polyNormalPerVertex',
  'polyNormalizeUV',
  'polyOptUvs',
  'polyPinUV',
  'polyPipe',
  'polyPlane',
  'polyPlatonicSolid',
  'polyPoke',
  'polyPrimitive',
  'polyPrimitiveMisc',
  'polyPrism',
  'polyProjectCurve',
  'polyPyramid',
  'polyQuad',
  'polyReduce',
  'polyRemesh',
  'polyRetopo',
  'polySeparate',
  'polySewEdge',
  'polySmooth',
  'polySoftEdge',
  'polySphere',
  'polySpinEdge',
  'polySplit',
  'polySplitEdge',
  'polySplitRing',
  'polyStraightenUVBorder',
  'polyToSubdiv',
  'polyTorus',
  'polyTransfer',
  'polyTriangulate',
  'polyUVRectangle',
  'polyUnite',
  'polyWedgeFace',
  'projectCurve',
  'projectTangent',
  'rebuildCurve',
  'rebuildSurface',
  'reference',
  'reverseCurve',
  'reverseSurface',
  'revolve',
  'rigidBody',
  'rigidSolver',
  'roundConstantRadius',
  'scaleConstraint',
  'sculpt',
  'sequenceManager',
  'shot',
  'skinCluster',
  'smoothCurve',
  'snapshot',
  'softMod',
  'spaceLocator',
  'sphere',
  'spotLight',
  'spring',
  'stroke',
  'subdCleanTopology',
  'subdLayoutUV',
  'subdMapCut',
  'subdMapSewMove',
  'subdiv',
  'surface',
  'tangentConstraint',
  'tension',
  'textureDeformer',
  'timeEditor',
  'timeEditorAnimSource',
  'timeEditorClip',
  'timeEditorTracks',
  'timeWarp',
  'torus',
  'transferAttributes',
  'trim',
  'turbulence',
  'uniform',
  'untrim',
  'volumeAxis',
  'vortex',
  'wire'],
 {'animation': ['aimConstraint',
                'animCurveEditor',
                'animDisplay',
                'animLayer',
                'animView',
                'applyTake',
                'autoKeyframe',
                'backgroundEvaluationManager',
                'bakeClip',
                'bakeDeformer',
                'bakeResults',
                'bakeSimulation',
                'bindSkin',
                'blendShape',
                'blendShapeEditor',
                'blendShapePanel',
                'blendTwoAttr',
                'boneLattice',
                'bufferCurve',
                'buildBookmarkMenu',
                'buildKeyframeMenu',
                'character',
                'characterMap',
                'characterize',
                'choice',
                'clip',
                'clipEditor',
                'clipMatching',
                'clipSchedule',
                'clipSchedulerOutliner',
                'cluster',
                'combinationShape',
                'connectJoint',
                'controller',
                'copyDeformerWeights',
                'copyFlexor',
                'copyKey',
                'copySkinWeights',
                'currentTime',
                'cutKey',
                'dagPose',
                'defineDataServer',
                'defineVirtualDevice',
                'deformer',
                'deformerWeights',
                'deltaMush',
                'deviceManager',
                'disconnectJoint',
                'dopeSheetEditor',
                'dropoffLocator',
                'effector',
                'enableDevice',
                'evaluationManager',
                'evaluator',
                'filterCurve',
                'findDeformers',
                'findKeyframe',
                'flexor',
                'flow',
                'freezeOptions',
                'geomBind',
                'geometryConstraint',
                'hikGlobals',
                'ikHandle',
                'ikHandleDisplayScale',
                'ikSolver',
                'ikSystem',
                'ikSystemInfo',
                'ikfkDisplayMethod',
                'insertJoint',
                'joint',
                'jointCluster',
                'jointDisplayScale',
                'jointLattice',
                'keyTangent',
                'keyframe',
                'keyframeOutliner',
                'keyframeStats',
                'keyingGroup',
                'lattice',
                'listAnimatable',
                'marker',
                'mirrorJoint',
                'movIn',
                'movOut',
                'movieInfo',
                'mute',
                'nonLinear',
                'normalConstraint',
                'orientConstraint',
                'pairBlend',
                'parentConstraint',
                'pasteKey',
                'pathAnimation',
                'percent',
                'play',
                'playbackOptions',
                'playblast',
                'pointConstraint',
                'pointOnPolyConstraint',
                'poleVectorConstraint',
                'polyUniteSkinned',
                'pose',
                'poseEditor',
                'posePanel',
                'readTake',
                'recordDevice',
                'removeJoint',
                'reorderDeformers',
                'reroot',
                'rotationInterpolation',
                'scaleConstraint',
                'scaleKey',
                'sculpt',
                'sculptTarget',
                'sequenceManager',
                'setDrivenKeyframe',
                'setInfinity',
                'setKeyPath',
                'setKeyframe',
                'setKeyframeBlendshapeTargetWts',
                'shapeEditor',
                'shapePanel',
                'shot',
                'shotRipple',
                'simplify',
                'skeletonEmbed',
                'skinCluster',
                'skinPercent',
                'snapKey',
                'snapshot',
                'softMod',
                'sound',
                'substituteGeometry',
                'tangentConstraint',
                'tension',
                'textureDeformer',
                'timeEditor',
                'timeEditorAnimSource',
                'timeEditorBakeClips',
                'timeEditorClip',
                'timeEditorClipLayer',
                'timeEditorClipOffset',
                'timeEditorComposition',
                'timeEditorPanel',
                'timeEditorTracks',
                'timeWarp',
                'ubercam',
                'volumeBind',
                'wire',
                'wrinkle',
                'writeTake'],
  'context': ['alignCtx',
              'arcLenDimContext',
              'art3dPaintCtx',
              'artAttrCtx',
              'artAttrPaintVertexCtx',
              'artAttrSkinPaintCtx',
              'artBaseCtx',
              'artFluidAttrCtx',
              'artPuttyCtx',
              'artSelectCtx',
              'artSetPaintCtx',
              'artUserPaintCtx',
              'blendCtx',
              'boxDollyCtx',
              'boxZoomCtx',
              'clipEditorCurrentTimeCtx',
              'createNurbsCircleCtx',
              'createNurbsConeCtx',
              'createNurbsCubeCtx',
              'createNurbsCylinderCtx',
              'createNurbsPlaneCtx',
              'createNurbsSphereCtx',
              'createNurbsSquareCtx',
              'createNurbsTorusCtx',
              'createPolyConeCtx',
              'createPolyCubeCtx',
              'createPolyCylinderCtx',
              'createPolyHelixCtx',
              'createPolyPipeCtx',
              'createPolyPlaneCtx',
              'createPolyPlatonicSolidCtx',
              'createPolyPrismCtx',
              'createPolyPyramidCtx',
              'createPolySoccerBallCtx',
              'createPolySphereCtx',
              'createPolyTorusCtx',
              'ctxAbort',
              'ctxCompletion',
              'ctxData',
              'ctxEditMode',
              'ctxTraverse',
              'currentCtx',
              'currentTimeCtx',
              'curveAddPtCtx',
              'curveBezierCtx',
              'curveCVCtx',
              'curveEPCtx',
              'curveEditorCtx',
              'curveMoveEPCtx',
              'curveSketchCtx',
              'directKeyCtx',
              'distanceDimContext',
              'dollyCtx',
              'dpBirailCtx',
              'dragAttrContext',
              'draggerContext',
              'drawExtrudeFacetCtx',
              'dynPaintCtx',
              'dynParticleCtx',
              'dynSelectCtx',
              'dynWireCtx',
              'filterButterworthCtx',
              'filterKeyReducerCtx',
              'graphDollyCtx',
              'graphSelectContext',
              'graphTrackCtx',
              'greasePencilCtx',
              'hotkeyCtx',
              'ikHandleCtx',
              'ikSplineHandleCtx',
              'insertJointCtx',
              'insertKeyCtx',
              'jointCtx',
              'keyframeRegionCurrentTimeCtx',
              'keyframeRegionDirectKeyCtx',
              'keyframeRegionDollyCtx',
              'keyframeRegionInsertKeyCtx',
              'keyframeRegionMoveKeyCtx',
              'keyframeRegionScaleKeyCtx',
              'keyframeRegionSelectKeyCtx',
              'keyframeRegionSetKeyCtx',
              'keyframeRegionTrackCtx',
              'lassoContext',
              'latticeDeformKeyCtx',
              'manipMoveContext',
              'manipMoveLimitsCtx',
              'manipRotateContext',
              'manipRotateLimitsCtx',
              'manipScaleContext',
              'manipScaleLimitsCtx',
              'mateCtx',
              'modelCurrentTimeCtx',
              'modelingToolkitSuperCtx',
              'moveKeyCtx',
              'mpBirailCtx',
              'orbitCtx',
              'panZoomCtx',
              'paramDimContext',
              'polyAppendFacetCtx',
              'polyCreaseCtx',
              'polyCreateFacetCtx',
              'polyCutCtx',
              'polyCutUVCtx',
              'polyMergeEdgeCtx',
              'polyMergeFacetCtx',
              'polyRetopoCtx',
              'polySelectCtx',
              'polySelectEditCtx',
              'polyShortestPathCtx',
              'polySlideEdgeCtx',
              'polySplitCtx',
              'polySuperCtx',
              'polyVertexNormalCtx',
              'projectionContext',
              'propModCtx',
              'regionSelectKeyCtx',
              'renderWindowSelectContext',
              'retimeKeyCtx',
              'rollCtx',
              'roundCRCtx',
              'scaleKeyCtx',
              'scriptCtx',
              'sculptMeshCacheCtx',
              'selectContext',
              'selectKeyCtx',
              'selectKeyframeRegionCtx',
              'setEditCtx',
              'setKeyCtx',
              'shadingGeometryRelCtx',
              'shadingLightRelCtx',
              'showManipCtx',
              'skinBindCtx',
              'snapTogetherCtx',
              'snapshotBeadContext',
              'snapshotBeadCtx',
              'snapshotModifyKeyCtx',
              'softModContext',
              'softModCtx',
              'softSelectOptionsCtx',
              'spBirailCtx',
              'srtContext',
              'stitchSurfaceCtx',
              'superCtx',
              'targetWeldCtx',
              'texCutContext',
              'texLatticeDeformContext',
              'texManipContext',
              'texMoveContext',
              'texMoveUVShellContext',
              'texRotateContext',
              'texScaleContext',
              'texSculptCacheContext',
              'texSelectContext',
              'texSelectShortestPathCtx',
              'texSmoothContext',
              'texSmudgeUVContext',
              'texTweakUVContext',
              'texWinToolCtx',
              'textureLassoContext',
              'texturePlacementContext',
              'threePointArcCtx',
              'trackCtx',
              'trimCtx',
              'tumbleCtx',
              'twoPointArcCtx',
              'userCtx',
              'view2dToolCtx',
              'walkCtx',
              'wireContext',
              'wrinkleContext'],
  'effects': ['addDynamic',
              'addPP',
              'air',
              'arrayMapper',
              'collision',
              'colorAtPoint',
              'connectDynamic',
              'constrain',
              'drag',
              'dynCache',
              'dynControl',
              'dynExport',
              'dynExpression',
              'dynGlobals',
              'dynPaintEditor',
              'dynPref',
              'emit',
              'emitter',
              'event',
              'expression',
              'expressionEditorListen',
              'fluidCacheInfo',
              'fluidEmitter',
              'fluidVoxelInfo',
              'getDefaultBrush',
              'getFluidAttr',
              'getParticleAttr',
              'goal',
              'gravity',
              'loadFluid',
              'nBase',
              'nParticle',
              'nSoft',
              'newton',
              'paintEffectsDisplay',
              'particle',
              'particleExists',
              'particleFill',
              'particleInstancer',
              'particleRenderInfo',
              'pfxstrokes',
              'radial',
              'resampleFluid',
              'rigidBody',
              'rigidSolver',
              'runup',
              'saveFluid',
              'saveInitialState',
              'setDynamic',
              'setFluidAttr',
              'setParticleAttr',
              'soft',
              'spring',
              'stroke',
              'truncateFluidCache',
              'truncateHairCache',
              'turbulence',
              'uniform',
              'volumeAxis',
              'vortex'],
  'general': ['about',
              'addAttr',
              'addExtension',
              'affectedNet',
              'affects',
              'aliasAttr',
              'align',
              'applyAttrPattern',
              'artAttrTool',
              'assembly',
              'attributeInfo',
              'attributeName',
              'attributeQuery',
              'bakePartialHistory',
              'baseTemplate',
              'baseView',
              'color',
              'colorIndex',
              'colorManagementCatalog',
              'colorManagementConvert',
              'colorManagementFileRules',
              'colorManagementPrefs',
              'commandLogging',
              'commandPort',
              'connectAttr',
              'connectionInfo',
              'container',
              'containerBind',
              'containerProxy',
              'containerPublish',
              'containerTemplate',
              'containerView',
              'contextInfo',
              'copyAttr',
              'createAttrPatterns',
              'createDisplayLayer',
              'createNode',
              'currentUnit',
              'curveRGBColor',
              'cycleCheck',
              'delete',
              'deleteAttr',
              'deleteAttrPattern',
              'deleteExtension',
              'disconnectAttr',
              'displayAffected',
              'displayColor',
              'displayCull',
              'displayLevelOfDetail',
              'displayPref',
              'displayRGBColor',
              'displaySmoothness',
              'displayStats',
              'displaySurface',
              'distanceDimension',
              'duplicate',
              'editDisplayLayerGlobals',
              'editDisplayLayerMembers',
              'exactWorldBoundingBox',
              'expandedSelection',
              'getAttr',
              'getClassification',
              'group',
              'hide',
              'hilite',
              'inheritTransform',
              'instance',
              'instanceable',
              'instancer',
              'isConnected',
              'isDirty',
              'isolateSelect',
              'license',
              'listAttr',
              'listAttrPatterns',
              'listConnections',
              'listHistory',
              'listNodeTypes',
              'listNodesWithIncorrectNames',
              'listRelatives',
              'listSets',
              'lockNode',
              'ls',
              'makeIdentity',
              'makeLive',
              'makePaintable',
              'matchTransform',
              'move',
              'nodeCast',
              'nodeType',
              'objExists',
              'objectCenter',
              'objectType',
              'paramDimension',
              'paramLocator',
              'parent',
              'partition',
              'performanceOptions',
              'pickWalk',
              'pixelMove',
              'polySplitCtx2',
              'relationship',
              'removeMultiInstance',
              'rename',
              'renameAttr',
              'reorder',
              'reorderContainer',
              'resetTool',
              'rotate',
              'saveToolSettings',
              'scale',
              'scaleComponents',
              'sceneLint',
              'sculptMeshCacheChangeCloneSource',
              'select',
              'selectKey',
              'selectMode',
              'selectPref',
              'selectPriority',
              'selectType',
              'selectedNodes',
              'setAttr',
              'setToolTo',
              'sets',
              'shapeCompare',
              'showHidden',
              'snapMode',
              'softSelect',
              'spaceLocator',
              'suitePrefs',
              'symmetricModelling',
              'threadCount',
              'timeCode',
              'toggle',
              'toggleAxis',
              'toggleDisplacement',
              'toolDropped',
              'toolHasOptions',
              'toolPropertyWindow',
              'transformCompare',
              'transformLimits',
              'ungroup',
              'upAxis',
              'webView',
              'xform',
              'xformConstraint'],
  'language': ['assignCommand',
               'callbacks',
               'commandEcho',
               'condition',
               'connectControl',
               'deleteUI',
               'dimWhen',
               'disable',
               'encodeString',
               'evalDeferred',
               'evalEcho',
               'evalNoSelectNotify',
               'format',
               'getLastError',
               'getProcArguments',
               'isTrue',
               'itemFilter',
               'itemFilterAttr',
               'itemFilterType',
               'lsUI',
               'melOptions',
               'objectTypeUI',
               'optionVar',
               'pause',
               'python',
               'refresh',
               'renameUI',
               'resourceManager',
               'scriptJob',
               'selectionConnection',
               'setParent',
               'setUITemplate',
               'sortCaseInsensitive',
               'stackTrace',
               'stringArrayIntersector',
               'uiTemplate',
               'waitCursor'],
  'modeling': ['addMetadata',
               'alignCurve',
               'alignSurface',
               'angleBetween',
               'applyMetadata',
               'arcLengthDimension',
               'arclen',
               'arubaNurbsToPoly',
               'attachCurve',
               'attachSurface',
               'bevel',
               'bevelPlus',
               'bezierAnchorPreset',
               'bezierAnchorState',
               'bezierCurveToNurbs',
               'bezierInfo',
               'blend2',
               'blindDataType',
               'boundary',
               'canCreateManip',
               'changeSubdivComponentDisplayLevel',
               'changeSubdivRegion',
               'circle',
               'circularFillet',
               'closeCurve',
               'closeSurface',
               'coarsenSubdivSelectionList',
               'cone',
               'constructionHistory',
               'createSubdivRegion',
               'curve',
               'curveIntersect',
               'curveOnSurface',
               'cylinder',
               'dataStructure',
               'detachCurve',
               'detachSurface',
               'doubleProfileBirailSurface',
               'duplicateCurve',
               'duplicateSurface',
               'editMetadata',
               'extendCurve',
               'extendSurface',
               'extrude',
               'filletCurve',
               'filterExpand',
               'fitBspline',
               'freeFormFillet',
               'geomToBBox',
               'getMetadata',
               'globalStitch',
               'grid',
               'hardenPointCurve',
               'hasMetadata',
               'illustratorCurves',
               'insertKnotCurve',
               'insertKnotSurface',
               'intersect',
               'loft',
               'makeSingleSurface',
               'manipOptions',
               'manipPivot',
               'moveVertexAlongDirection',
               'multiProfileBirailSurface',
               'nurbsBoolean',
               'nurbsCopyUVSet',
               'nurbsCube',
               'nurbsCurveToBezier',
               'nurbsEditUV',
               'nurbsPlane',
               'nurbsSelect',
               'nurbsSquare',
               'nurbsToPoly',
               'nurbsToPolygonsPref',
               'nurbsToSubdiv',
               'nurbsToSubdivPref',
               'nurbsUVSet',
               'offsetCurve',
               'offsetCurveOnSurface',
               'offsetSurface',
               'planarSrf',
               'plane',
               'pointCurveConstraint',
               'pointOnCurve',
               'pointOnSurface',
               'pointPosition',
               'polyAppend',
               'polyAppendVertex',
               'polyAutoProjection',
               'polyAverageNormal',
               'polyAverageVertex',
               'polyBevel',
               'polyBevel3',
               'polyBlendColor',
               'polyBlindData',
               'polyBoolOp',
               'polyBridgeEdge',
               'polyCBoolOp',
               'polyCacheMonitor',
               'polyCanBridgeEdge',
               'polyCheck',
               'polyChipOff',
               'polyCircularize',
               'polyCircularizeEdge',
               'polyCircularizeFace',
               'polyClean',
               'polyClipboard',
               'polyCloseBorder',
               'polyCollapseEdge',
               'polyCollapseFacet',
               'polyCollapseTweaks',
               'polyColorBlindData',
               'polyColorDel',
               'polyColorMod',
               'polyColorPerVertex',
               'polyColorSet',
               'polyCompare',
               'polyCone',
               'polyConnectComponents',
               'polyContourProjection',
               'polyCopyUV',
               'polyCrease',
               'polyCreateFacet',
               'polyCube',
               'polyCut',
               'polyCylinder',
               'polyCylindricalProjection',
               'polyDelEdge',
               'polyDelFacet',
               'polyDelVertex',
               'polyDuplicateAndConnect',
               'polyDuplicateEdge',
               'polyEditEdgeFlow',
               'polyEditUV',
               'polyEditUVShell',
               'polyEvaluate',
               'polyExtrudeEdge',
               'polyExtrudeFacet',
               'polyExtrudeVertex',
               'polyFlipEdge',
               'polyFlipUV',
               'polyForceUV',
               'polyGeoSampler',
               'polyHelix',
               'polyHole',
               'polyInfo',
               'polyInstallAction',
               'polyLayoutUV',
               'polyListComponentConversion',
               'polyMapCut',
               'polyMapDel',
               'polyMapSew',
               'polyMapSewMove',
               'polyMergeEdge',
               'polyMergeFacet',
               'polyMergeUV',
               'polyMergeVertex',
               'polyMirrorFace',
               'polyMoveEdge',
               'polyMoveFacet',
               'polyMoveFacetUV',
               'polyMoveUV',
               'polyMoveVertex',
               'polyMultiLayoutUV',
               'polyNormal',
               'polyNormalPerVertex',
               'polyNormalizeUV',
               'polyOptUvs',
               'polyOptions',
               'polyOutput',
               'polyPinUV',
               'polyPipe',
               'polyPlanarProjection',
               'polyPlane',
               'polyPlatonicSolid',
               'polyPoke',
               'polyPrimitive',
               'polyPrism',
               'polyProjectCurve',
               'polyProjection',
               'polyPyramid',
               'polyQuad',
               'polyQueryBlindData',
               'polyReduce',
               'polyRemesh',
               'polyRetopo',
               'polySelect',
               'polySelectConstraint',
               'polySelectConstraintMonitor',
               'polySeparate',
               'polySetToFaceNormal',
               'polySewEdge',
               'polySlideEdge',
               'polySmooth',
               'polySoftEdge',
               'polySphere',
               'polySphericalProjection',
               'polySplit',
               'polySplitEdge',
               'polySplitRing',
               'polySplitVertex',
               'polyStraightenUVBorder',
               'polySubdivideEdge',
               'polySubdivideFacet',
               'polyToSubdiv',
               'polyTorus',
               'polyTransfer',
               'polyTriangulate',
               'polyUVCoverage',
               'polyUVOverlap',
               'polyUVRectangle',
               'polyUVSet',
               'polyUVStackSimilarShells',
               'polyUnite',
               'polyWedgeFace',
               'projectCurve',
               'projectTangent',
               'propMove',
               'querySubdiv',
               'rebuildCurve',
               'rebuildSurface',
               'refineSubdivSelectionList',
               'reverseCurve',
               'reverseSurface',
               'revolve',
               'roundConstantRadius',
               'setXformManip',
               'showMetadata',
               'singleProfileBirailSurface',
               'smoothCurve',
               'smoothTangentSurface',
               'sphere',
               'squareSurface',
               'stitchSurface',
               'stitchSurfacePoints',
               'subdAutoProjection',
               'subdCleanTopology',
               'subdCollapse',
               'subdDuplicateAndConnect',
               'subdEditUV',
               'subdLayoutUV',
               'subdListComponentConversion',
               'subdMapCut',
               'subdMapSewMove',
               'subdMatchTopology',
               'subdMirror',
               'subdPlanarProjection',
               'subdToBlind',
               'subdToPoly',
               'subdTransferUVsToCache',
               'subdiv',
               'subdivCrease',
               'subdivDisplaySmoothness',
               'surface',
               'textCurves',
               'tolerance',
               'torus',
               'transferAttributes',
               'transferShadingSets',
               'trim',
               'unfold',
               'untangleUV',
               'untrim',
               'uvSnapshot'],
  'other': ['TanimLayer',
            'TrenderSetupStates',
            'adpAnalyticsDialog',
            'adskAsset',
            'adskAssetLibrary',
            'adskAssetList',
            'adskAssetListUI',
            'agFormatIn',
            'agFormatOut',
            'artAttr',
            'artAttrSkinPaint',
            'artAttrSkinPaintCmd',
            'artFluidAttr',
            'artSelect',
            'artSetPaint',
            'blend',
            'caddyManip',
            'clearShear',
            'copyNode',
            'crashInfoCmd',
            'dagCommandWrapper',
            'dagObjectHit',
            'debug',
            'debugNamespace',
            'debugVar',
            'dgControl',
            'dgPerformance',
            'dgcontrol',
            'dgdebug',
            'dgstats',
            'directConnectPath',
            'dispatchGenericCommand',
            'dynTestData',
            'evalContinue',
            'extendFluid',
            'flagTest',
            'flushIdleQueue',
            'flushThumbnailCache',
            'fontAttributes',
            'greasePencil',
            'greasePencilHelper',
            'greaseRenderPlane',
            'groupParts',
            'hotkeyEditor',
            'hotkeyMapSet',
            'imageWindowEditor',
            'interactionStyle',
            'itemFilterRender',
            'iterOnNurbs',
            'journal',
            'licenseCheck',
            'manipComponentPivot',
            'manipComponentUpdate',
            'mayaDpiSettingAction',
            'memoryDiag',
            'meshIntersectTest',
            'mouldMesh',
            'mouldSrf',
            'mouldSubdiv',
            'movieCompressor',
            'myTestCmd',
            'nodeGrapher',
            'nop',
            'nurbsCurveRebuildPref',
            'ogsdebug',
            'paint3d',
            'polyColorSetCmdWrapper',
            'polyIterOnPoly',
            'polyPrimitiveMisc',
            'polySelectEditCtxDataCmd',
            'polySelectSp',
            'polySetVertices',
            'polySpinEdge',
            'polyTestPop',
            'polyToCurve',
            'polyUVStackSimilarShellsCmd',
            'polyWarpImage',
            'psdConvSolidTxOptions',
            'rampWidget',
            'rampWidgetAttrless',
            'readPDC',
            'repeatLast',
            'retimeHelper',
            'selectKeyframe',
            'subdDisplayMode',
            'subdToNurbs',
            'subgraph',
            'syncSculptCache',
            'testPa',
            'testPassContribution',
            'texSculptCacheSync',
            'timeSliderCustomDraw',
            'webViewCmd'],
  'rendering': ['ambientLight',
                'assignViewportFactories',
                'batchRender',
                'binMembership',
                'camera',
                'cameraSet',
                'cameraView',
                'checkDefaultRenderGlobals',
                'convertIffToPsd',
                'convertSolidTx',
                'convertTessellation',
                'createLayeredPsdFile',
                'createRenderLayer',
                'defaultLightListCheckBox',
                'directionalLight',
                'displacementToPoly',
                'doBlur',
                'dolly',
                'editRenderLayerAdjustment',
                'editRenderLayerGlobals',
                'editRenderLayerMembers',
                'exclusiveLightCheckBox',
                'frameBufferName',
                'getRenderDependencies',
                'getRenderTasks',
                'glRender',
                'glRenderEditor',
                'hwReflectionMap',
                'hwRender',
                'hwRenderLoad',
                'imagePlane',
                'iprEngine',
                'layeredShaderPort',
                'layeredTexturePort',
                'lightList',
                'lightlink',
                'listCameras',
                'lookThru',
                'lsThroughFilter',
                'makebot',
                'mayaHasRenderSetup',
                'nodeIconButton',
                'nodePreset',
                'ogsRender',
                'orbit',
                'panZoom',
                'perCameraVisibility',
                'pointLight',
                'preferredRenderer',
                'prepareRender',
                'projectionManip',
                'psdChannelOutliner',
                'psdEditTextureFile',
                'psdExport',
                'psdTextureFile',
                'rampColorPort',
                'render',
                'renderGlobalsNode',
                'renderInfo',
                'renderLayerPostProcess',
                'renderManip',
                'renderPartition',
                'renderPassRegistry',
                'renderQualityNode',
                'renderSettings',
                'renderThumbnailUpdate',
                'renderWindowEditor',
                'renderer',
                'resolutionNode',
                'roll',
                'sampleImage',
                'setDefaultShadingGroup',
                'setRenderPassType',
                'shadingConnection',
                'shadingNetworkCompare',
                'shadingNode',
                'showShadingGroupAttrEditor',
                'soloMaterial',
                'spotLight',
                'spotLightPreviewPort',
                'surfaceSampler',
                'surfaceShaderList',
                'swatchRefresh',
                'textureWindow',
                'track',
                'tumble',
                'uvLink',
                'viewCamera',
                'viewClipPlane',
                'viewFit',
                'viewHeadOn',
                'viewLookAt',
                'viewPlace',
                'viewSet'],
  'runtime': ['ATOMTemplate',
              'AbortCurrentTool',
              'ActivateGlobalScreenSlider',
              'ActivateGlobalScreenSliderModeMarkingMenu',
              'ActivateViewport20',
              'AddAnimationOffset',
              'AddAnimationOffsetOptions',
              'AddAttribute',
              'AddBlendShape',
              'AddBlendShapeOptions',
              'AddBoatLocator',
              'AddBoatLocatorOptions',
              'AddCombinationTarget',
              'AddCombinationTargetOptions',
              'AddCurvesToHairSystem',
              'AddDivisions',
              'AddDivisionsOptions',
              'AddDynamicBuoy',
              'AddDynamicBuoyOptions',
              'AddEdgeDivisions',
              'AddEdgeDivisionsOptions',
              'AddFaceDivisions',
              'AddFaceDivisionsOptions',
              'AddFloorContactPlane',
              'AddHolder',
              'AddHolderOptions',
              'AddInBetweenTargetShape',
              'AddInBetweenTargetShapeOptions',
              'AddInbetween',
              'AddInfluence',
              'AddInfluenceOptions',
              'AddKeyToolActivate',
              'AddKeyToolDeactivate',
              'AddKeysTool',
              'AddKeysToolOptions',
              'AddOceanDynamicLocator',
              'AddOceanDynamicLocatorOptions',
              'AddOceanPreviewPlane',
              'AddOceanSurfaceLocator',
              'AddPfxToHairSystem',
              'AddPointsTool',
              'AddPondBoatLocator',
              'AddPondBoatLocatorOptions',
              'AddPondDynamicBuoy',
              'AddPondDynamicBuoyOptions',
              'AddPondDynamicLocator',
              'AddPondDynamicLocatorOptions',
              'AddPondSurfaceLocator',
              'AddSelectionAsCombinationTarget',
              'AddSelectionAsCombinationTargetOptions',
              'AddSelectionAsInBetweenTargetShape',
              'AddSelectionAsInBetweenTargetShapeOptions',
              'AddSelectionAsTargetShape',
              'AddSelectionAsTargetShapeOptions',
              'AddShrinkWrapSurfaces',
              'AddTargetShape',
              'AddTargetShapeOptions',
              'AddTimeWarp',
              'AddToCharacterSet',
              'AddToContainer',
              'AddToContainerOptions',
              'AddToCurrentScene3dsMax',
              'AddToCurrentSceneMotionBuilder',
              'AddToCurrentSceneMudbox',
              'AddWire',
              'AddWireOptions',
              'AddWrapInfluence',
              'AffectSelectedObject',
              'AimConstraint',
              'AimConstraintOptions',
              'Air',
              'AirOptions',
              'AlignCameraToPolygon',
              'AlignCurve',
              'AlignCurveOptions',
              'AlignObjects',
              'AlignSurfaces',
              'AlignSurfacesOptions',
              'AlignUV',
              'AlignUVOptions',
              'AnimLayerRelationshipEditor',
              'AnimationSnapshot',
              'AnimationSnapshotOptions',
              'AnimationSweep',
              'AnimationSweepOptions',
              'AnimationTurntable',
              'AnimationTurntableOptions',
              'AppendToHairCache',
              'AppendToHairCacheOptions',
              'AppendToPolygonTool',
              'AppendToPolygonToolOptions',
              'ApplySettingsToLastStroke',
              'ApplySettingsToSelectedStroke',
              'ArcLengthTool',
              'ArchiveScene',
              'ArchiveSceneOptions',
              'Art3dPaintTool',
              'Art3dPaintToolOptions',
              'ArtPaintAttrTool',
              'ArtPaintAttrToolOptions',
              'ArtPaintBlendShapeWeightsTool',
              'ArtPaintBlendShapeWeightsToolOptions',
              'ArtPaintSelectTool',
              'ArtPaintSelectToolOptions',
              'ArtPaintSkinWeightsTool',
              'ArtPaintSkinWeightsToolOptions',
              'AssetEditor',
              'AssignBrushToHairSystem',
              'AssignBrushToPfxToon',
              'AssignHairConstraint',
              'AssignHairConstraintOptions',
              'AssignNewMaterial',
              'AssignNewPfxToon',
              'AssignNewSet',
              'AssignOfflineFile',
              'AssignOfflineFileFromRefEd',
              'AssignOfflineFileFromRefEdOptions',
              'AssignOfflineFileOptions',
              'AssignTemplate',
              'AssignTemplateOptions',
              'AssignToonShaderCircleHighlight',
              'AssignToonShaderDarkProfile',
              'AssignToonShaderLightAngle',
              'AssignToonShaderRimLight',
              'AssignToonShaderShadedBrightness',
              'AssignToonShaderSolid',
              'AssignToonShaderThreeToneBrightness',
              'AssumePreferredAngle',
              'AssumePreferredAngleOptions',
              'AttachBrushToCurves',
              'AttachCurve',
              'AttachCurveOptions',
              'AttachSelectedAsSourceField',
              'AttachSubdivSurface',
              'AttachSubdivSurfaceOptions',
              'AttachSurfaceWithoutMoving',
              'AttachSurfaces',
              'AttachSurfacesOptions',
              'AttachToPath',
              'AttachToPathOptions',
              'AttributeEditor',
              'AutoPaintMarkingMenu',
              'AutoPaintMarkingMenuPopDown',
              'AutoProjection',
              'AutoProjectionOptions',
              'AutoSeamUVs',
              'AutoSeamUVsOptions',
              'AutobindContainer',
              'AutobindContainerOptions',
              'AveragePolygonNormals',
              'AveragePolygonNormalsOptions',
              'AverageVertex',
              'BakeAllNonDefHistory',
              'BakeChannel',
              'BakeChannelOptions',
              'BakeCustomPivot',
              'BakeCustomPivotOptions',
              'BakeDeformerTool',
              'BakeNonDefHistory',
              'BakeNonDefHistoryOptions',
              'BakeSimulation',
              'BakeSimulationOptions',
              'BakeSpringAnimation',
              'BakeSpringAnimationOptions',
              'BakeSurfaceToTexture',
              'BakeTopologyToTargets',
              'BaseLevelComponentDisplay',
              'BatchBake',
              'BatchBakeOptions',
              'BatchRender',
              'BatchRenderOptions',
              'Bend',
              'BendCurves',
              'BendCurvesOptions',
              'BendOptions',
              'BestPlaneTexturingTool',
              'Bevel',
              'BevelOptions',
              'BevelPlus',
              'BevelPlusOptions',
              'BevelPolygon',
              'BevelPolygonOptions',
              'BezierCurveToNurbs',
              'BezierPresetBezier',
              'BezierPresetBezierCorner',
              'BezierPresetCorner',
              'BezierSetAnchorBroken',
              'BezierSetAnchorEven',
              'BezierSetAnchorSmooth',
              'BezierSetAnchorUneven',
              'Birail1',
              'Birail1Options',
              'Birail2',
              'Birail2Options',
              'Birail3',
              'Birail3Options',
              'BlendShapeEditor',
              'BlindDataEditor',
              'BothProxySubdivDisplay',
              'Boundary',
              'BoundaryOptions',
              'BreakLightLinks',
              'BreakRigidBodyConnection',
              'BreakShadowLinks',
              'BreakStereoRigs',
              'BreakTangent',
              'BreakTangents',
              'BridgeEdge',
              'BridgeEdgeOptions',
              'BridgeOrFill',
              'BrushAnimationMarkingMenu',
              'BrushAnimationMarkingMenuPopDown',
              'BrushPresetBlend',
              'BrushPresetBlendOff',
              'BrushPresetBlendShading',
              'BrushPresetBlendShadingOff',
              'BrushPresetBlendShape',
              'BrushPresetBlendShapeOff',
              'BrushPresetReplaceShading',
              'BrushPresetReplaceShadingOff',
              'BufferCurveSnapshot',
              'CVCurveTool',
              'CVCurveToolOptions',
              'CVHardness',
              'CVHardnessOptions',
              'CameraModeOrthographic',
              'CameraModePerspective',
              'CameraModeToggle',
              'CameraRemoveAll',
              'CameraRemoveAllForAll',
              'CameraRemoveFromExclusive',
              'CameraRemoveFromHidden',
              'CameraSetEditor',
              'CancelBatchRender',
              'CenterPivot',
              'CenterViewOfSelection',
              'ChamferVertex',
              'ChamferVertexOptions',
              'ChangeAnimPrefs',
              'ChangeColorPrefs',
              'ChangeEdgeWidth',
              'ChangeNormalSize',
              'ChangeUIPrefs',
              'ChangeUVSize',
              'ChangeVertexSize',
              'ChannelControlEditor',
              'CharacterAnimationEditor',
              'CharacterMapper',
              'CharacterSetEditor',
              'CircularFillet',
              'CircularFilletOptions',
              'CleanupPolygon',
              'CleanupPolygonOptions',
              'ClearCurrentCharacterList',
              'ClearCurrentContainer',
              'ClearInitialState',
              'ClearPaintEffectsView',
              'CloseFrontWindow',
              'ClosestPointOn',
              'ClosestPointOnOptions',
              'ClusterCurve',
              'CoarseLevelComponentDisplay',
              'CoarsenSelectedComponents',
              'CoarserSubdivLevel',
              'CollapseSubdivSurfaceHierarchy',
              'CollapseSubdivSurfaceHierarchyOptions',
              'ColorPreferencesWindow',
              'CombinePolygons',
              'CombinePolygonsOptions',
              'CommandShell',
              'CommandWindow',
              'CompleteCurrentTool',
              'ComponentEditor',
              'ConformPolygon',
              'ConformPolygonNormals',
              'ConformPolygonOptions',
              'ConnectComponents',
              'ConnectComponentsOptions',
              'ConnectJoint',
              'ConnectJointOptions',
              'ConnectNodeToIKFK',
              'ConnectToTime',
              'ConnectionEditor',
              'ContentBrowserLayout',
              'ContentBrowserWindow',
              'ConvertHairSelectionToConstraints',
              'ConvertHairSelectionToCurrentPositions',
              'ConvertHairSelectionToFollicles',
              'ConvertHairSelectionToHairSystems',
              'ConvertHairSelectionToRestCurveEndCVs',
              'ConvertHairSelectionToRestCurves',
              'ConvertHairSelectionToStartAndRestCurveEndCVs',
              'ConvertHairSelectionToStartCurveEndCVs',
              'ConvertHairSelectionToStartCurves',
              'ConvertInstanceToObject',
              'ConvertPaintEffectsToPoly',
              'ConvertPaintEffectsToPolyOptions',
              'ConvertSelectionToContainedEdges',
              'ConvertSelectionToContainedFaces',
              'ConvertSelectionToEdgePerimeter',
              'ConvertSelectionToEdges',
              'ConvertSelectionToFacePerimeter',
              'ConvertSelectionToFaces',
              'ConvertSelectionToShell',
              'ConvertSelectionToShellBorder',
              'ConvertSelectionToUVBorder',
              'ConvertSelectionToUVEdgeLoop',
              'ConvertSelectionToUVPerimeter',
              'ConvertSelectionToUVShell',
              'ConvertSelectionToUVShellBorder',
              'ConvertSelectionToUVs',
              'ConvertSelectionToVertexFaces',
              'ConvertSelectionToVertexPerimeter',
              'ConvertSelectionToVertices',
              'ConvertToBreakdown',
              'ConvertToFrozen',
              'ConvertToKey',
              'ConvertTypeCapsToCurves',
              'CopyFlexor',
              'CopyKeys',
              'CopyKeysOptions',
              'CopyMeshAttributes',
              'CopySelected',
              'CopySkinWeights',
              'CopySkinWeightsOptions',
              'CopyUVs',
              'CopyUVsToUVSet',
              'CopyUVsToUVSetOptions',
              'CopyVertexSkinWeights',
              'CopyVertexWeights',
              'CreaseProxyEdgeTool',
              'CreaseProxyEdgeToolOptions',
              'Create2DContainer',
              'Create2DContainerEmitter',
              'Create2DContainerEmitterOptions',
              'Create2DContainerOptions',
              'Create3DContainer',
              'Create3DContainerEmitter',
              'Create3DContainerEmitterOptions',
              'Create3DContainerOptions',
              'CreateActiveRigidBody',
              'CreateActiveRigidBodyOptions',
              'CreateAmbientLight',
              'CreateAmbientLightOptions',
              'CreateAnnotateNode',
              'CreateAreaLight',
              'CreateAreaLightOptions',
              'CreateBarrierConstraint',
              'CreateBarrierConstraintOptions',
              'CreateBezierCurveTool',
              'CreateBezierCurveToolOptions',
              'CreateBindingSet',
              'CreateBlendShape',
              'CreateBlendShapeOptions',
              'CreateCameraAim',
              'CreateCameraAimOptions',
              'CreateCameraAimUp',
              'CreateCameraAimUpOptions',
              'CreateCameraFromView',
              'CreateCameraOnly',
              'CreateCameraOnlyOptions',
              'CreateCharacter',
              'CreateCharacterOptions',
              'CreateClip',
              'CreateClipOptions',
              'CreateCluster',
              'CreateClusterOptions',
              'CreateConstraint',
              'CreateConstraintClip',
              'CreateConstraintClipOptions',
              'CreateConstraintOptions',
              'CreateConstructionPlane',
              'CreateConstructionPlaneOptions',
              'CreateContainer',
              'CreateContainerOptions',
              'CreateControlRig',
              'CreateCreaseSet',
              'CreateCreaseSetOptions',
              'CreateCurveField',
              'CreateCurveFromPoly',
              'CreateCurveFromPolyOptions',
              'CreateDagContainer',
              'CreateDagContainerOptions',
              'CreateDirectionalLight',
              'CreateDirectionalLightOptions',
              'CreateDiskCache',
              'CreateDiskCacheOptions',
              'CreateEmitter',
              'CreateEmitterOptions',
              'CreateEmptyGroup',
              'CreateEmptySet',
              'CreateEmptySetOptions',
              'CreateEmptyUVSet',
              'CreateEmptyUVSetOptions',
              'CreateExpressionClip',
              'CreateExpressionClipOptions',
              'CreateFlexorWindow',
              'CreateFluidCache',
              'CreateFluidCacheOptions',
              'CreateGhost',
              'CreateGhostOptions',
              'CreateHair',
              'CreateHairCache',
              'CreateHairCacheOptions',
              'CreateHairOptions',
              'CreateHingeConstraint',
              'CreateHingeConstraintOptions',
              'CreateIllustratorCurves',
              'CreateIllustratorCurvesOptions',
              'CreateImagePlane',
              'CreateImagePlaneOptions',
              'CreateJiggleDeformer',
              'CreateJiggleOptions',
              'CreateLattice',
              'CreateLatticeOptions',
              'CreateLineModifier',
              'CreateLocator',
              'CreateMotionTrail',
              'CreateMotionTrailOptions',
              'CreateMultiStereoRig',
              'CreateNSoftBody',
              'CreateNSoftBodyOptions',
              'CreateNURBSCircle',
              'CreateNURBSCircleOptions',
              'CreateNURBSCone',
              'CreateNURBSConeOptions',
              'CreateNURBSCube',
              'CreateNURBSCubeOptions',
              'CreateNURBSCylinder',
              'CreateNURBSCylinderOptions',
              'CreateNURBSPlane',
              'CreateNURBSPlaneOptions',
              'CreateNURBSSphere',
              'CreateNURBSSphereOptions',
              'CreateNURBSSquare',
              'CreateNURBSSquareOptions',
              'CreateNURBSTorus',
              'CreateNURBSTorusOptions',
              'CreateNailConstraint',
              'CreateNailConstraintOptions',
              'CreateNodeWindow',
              'CreateOcean',
              'CreateOceanOptions',
              'CreateOceanWake',
              'CreateOceanWakeOptions',
              'CreatePSDTextureItem',
              'CreateParticleDiskCache',
              'CreateParticleDiskCacheOptions',
              'CreatePartition',
              'CreatePartitionOptions',
              'CreatePassiveRigidBody',
              'CreatePassiveRigidBodyOptions',
              'CreatePinConstraint',
              'CreatePinConstraintOptions',
              'CreatePlatonicSolid',
              'CreatePlatonicSolidOptions',
              'CreatePointLight',
              'CreatePointLightOptions',
              'CreatePolyFromPreview',
              'CreatePolygonCone',
              'CreatePolygonConeOptions',
              'CreatePolygonCube',
              'CreatePolygonCubeOptions',
              'CreatePolygonCylinder',
              'CreatePolygonCylinderOptions',
              'CreatePolygonDisc',
              'CreatePolygonDiscOptions',
              'CreatePolygonGear',
              'CreatePolygonGearOptions',
              'CreatePolygonHelix',
              'CreatePolygonHelixOptions',
              'CreatePolygonPipe',
              'CreatePolygonPipeOptions',
              'CreatePolygonPlane',
              'CreatePolygonPlaneOptions',
              'CreatePolygonPlatonic',
              'CreatePolygonPlatonicOptions',
              'CreatePolygonPrism',
              'CreatePolygonPrismOptions',
              'CreatePolygonPyramid',
              'CreatePolygonPyramidOptions',
              'CreatePolygonSVG',
              'CreatePolygonSoccerBall',
              'CreatePolygonSoccerBallOptions',
              'CreatePolygonSphere',
              'CreatePolygonSphereOptions',
              'CreatePolygonSphericalHarmonics',
              'CreatePolygonSphericalHarmonicsOptions',
              'CreatePolygonSuperEllipse',
              'CreatePolygonSuperEllipseOptions',
              'CreatePolygonTool',
              'CreatePolygonToolOptions',
              'CreatePolygonTorus',
              'CreatePolygonTorusOptions',
              'CreatePolygonType',
              'CreatePolygonUltraShape',
              'CreatePolygonUltraShapeOptions',
              'CreatePond',
              'CreatePondOptions',
              'CreatePose',
              'CreatePoseInterpolator',
              'CreatePoseInterpolatorOptions',
              'CreatePoseOptions',
              'CreateQuickSelectSet',
              'CreateReference',
              'CreateReferenceOptions',
              'CreateRigidBodySolver',
              'CreateSculptDeformer',
              'CreateSculptDeformerOptions',
              'CreateSet',
              'CreateSetOptions',
              'CreateShot',
              'CreateShotOptions',
              'CreateShrinkWrap',
              'CreateShrinkWrapOptions',
              'CreateSoftBody',
              'CreateSoftBodyOptions',
              'CreateSpotLight',
              'CreateSpotLightOptions',
              'CreateSpring',
              'CreateSpringConstraint',
              'CreateSpringConstraintOptions',
              'CreateSpringOptions',
              'CreateStereoRig',
              'CreateSubCharacter',
              'CreateSubCharacterOptions',
              'CreateSubdivCone',
              'CreateSubdivCube',
              'CreateSubdivCylinder',
              'CreateSubdivPlane',
              'CreateSubdivRegion',
              'CreateSubdivSphere',
              'CreateSubdivSurface',
              'CreateSubdivSurfaceOptions',
              'CreateSubdivSurfacePoly',
              'CreateSubdivSurfacePolyOptions',
              'CreateSubdivTorus',
              'CreateText',
              'CreateTextOptions',
              'CreateTextureDeformer',
              'CreateTextureDeformerOptions',
              'CreateTextureReferenceObject',
              'CreateUVShellAlongBorder',
              'CreateUVsBasedOnCamera',
              'CreateUVsBasedOnCameraOptions',
              'CreateVolumeCone',
              'CreateVolumeCube',
              'CreateVolumeLight',
              'CreateVolumeLightOptions',
              'CreateVolumeSphere',
              'CreateWake',
              'CreateWakeOptions',
              'CreateWrap',
              'CreateWrapOptions',
              'CurlCurves',
              'CurlCurvesOptions',
              'CurveEditTool',
              'CurveFillet',
              'CurveFilletOptions',
              'CurveFlow',
              'CurveFlowOptions',
              'CurveSmoothnessCoarse',
              'CurveSmoothnessFine',
              'CurveSmoothnessMedium',
              'CurveSmoothnessRough',
              'CurveUtilitiesMarkingMenu',
              'CurveUtilitiesMarkingMenuPopDown',
              'CurveWarp',
              'CustomNURBSComponentsOptions',
              'CustomNURBSSmoothness',
              'CustomNURBSSmoothnessOptions',
              'CustomPolygonDisplay',
              'CustomPolygonDisplayOptions',
              'CutCurve',
              'CutCurveOptions',
              'CutKeys',
              'CutKeysOptions',
              'CutPolygon',
              'CutPolygonOptions',
              'CutSelected',
              'CutUVs',
              'CutUVs3D',
              'CutUVsWithoutHotkey',
              'CycleBackgroundColor',
              'CycleIKHandleStickyState',
              'CycleThroughCameras',
              'DeactivateGlobalScreenSlider',
              'DeactivateGlobalScreenSliderModeMarkingMenu',
              'DecreaseCheckerDensity',
              'DecreaseExposureCoarse',
              'DecreaseExposureFine',
              'DecreaseGammaCoarse',
              'DecreaseGammaFine',
              'DecreaseManipulatorSize',
              'DecrementFluidCenter',
              'DefaultQualityDisplay',
              'DeformerSetEditor',
              'Delete',
              'DeleteAllCameras',
              'DeleteAllChannels',
              'DeleteAllClips',
              'DeleteAllClusters',
              'DeleteAllConstraints',
              'DeleteAllContainers',
              'DeleteAllControllers',
              'DeleteAllDynamicConstraints',
              'DeleteAllExpressions',
              'DeleteAllFluids',
              'DeleteAllFurs',
              'DeleteAllHistory',
              'DeleteAllIKHandles',
              'DeleteAllImagePlanes',
              'DeleteAllJoints',
              'DeleteAllLattices',
              'DeleteAllLights',
              'DeleteAllMotionPaths',
              'DeleteAllNCloths',
              'DeleteAllNParticles',
              'DeleteAllNRigids',
              'DeleteAllNonLinearDeformers',
              'DeleteAllParticles',
              'DeleteAllPoses',
              'DeleteAllRigidBodies',
              'DeleteAllRigidConstraints',
              'DeleteAllSculptObjects',
              'DeleteAllShadingGroupsAndMaterials',
              'DeleteAllSounds',
              'DeleteAllStaticChannels',
              'DeleteAllStrokes',
              'DeleteAllWires',
              'DeleteAttribute',
              'DeleteChannels',
              'DeleteChannelsOptions',
              'DeleteConstraints',
              'DeleteCurrentColorSet',
              'DeleteCurrentSet',
              'DeleteCurrentUVSet',
              'DeleteCurrentWorkspace',
              'DeleteEdge',
              'DeleteEntireHairSystem',
              'DeleteExpressions',
              'DeleteExpressionsOptions',
              'DeleteHair',
              'DeleteHairCache',
              'DeleteHistory',
              'DeleteKeys',
              'DeleteKeysOptions',
              'DeleteMemoryCaching',
              'DeleteMotionPaths',
              'DeletePolyElements',
              'DeleteRigidBodies',
              'DeleteSelectedContainers',
              'DeleteStaticChannels',
              'DeleteStaticChannelsOptions',
              'DeleteSurfaceFlow',
              'DeleteSurfaceFlowOptions',
              'DeleteTextureReferenceObject',
              'DeleteTimeWarp',
              'DeleteUVs',
              'DeleteUVsWithoutHotkey',
              'DeleteUnknownNodes',
              'DeleteVertex',
              'DeltaMush',
              'DeltaMushOptions',
              'DetachComponent',
              'DetachCurve',
              'DetachCurveOptions',
              'DetachEdgeComponent',
              'DetachSkeleton',
              'DetachSkeletonJoints',
              'DetachSkin',
              'DetachSkinOptions',
              'DetachSurfaces',
              'DetachSurfacesOptions',
              'DetachVertexComponent',
              'DeviceEditor',
              'DisableAll',
              'DisableAllCaches',
              'DisableAllGeometryCache',
              'DisableConstraints',
              'DisableExpressions',
              'DisableFluids',
              'DisableGlobalStitch',
              'DisableIKSolvers',
              'DisableMemoryCaching',
              'DisableParticles',
              'DisableRigidBodies',
              'DisableSelectedIKHandles',
              'DisableSnapshots',
              'DisableTimeChangeUndoConsolidation',
              'DisableWeightNrm',
              'DisconnectJoint',
              'DisplacementToPolygon',
              'DisplayCurrentHairCurves',
              'DisplayHairCurves',
              'DisplayHairCurvesCurrentAndRest',
              'DisplayHairCurvesCurrentAndStart',
              'DisplayHairCurvesRestPosition',
              'DisplayHairCurvesStart',
              'DisplayIntermediateObjects',
              'DisplayLayerEditorWindow',
              'DisplayLight',
              'DisplayShaded',
              'DisplayShadedAndTextured',
              'DisplayShadingMarkingMenu',
              'DisplayShadingMarkingMenuPopDown',
              'DisplaySmoothShaded',
              'DisplayUVShaded',
              'DisplayUVWireframe',
              'DisplayViewport',
              'DisplayWireframe',
              'DistanceTool',
              'DistributeShells',
              'DistributeShellsOptions',
              'DistributeUVs',
              'DistributeUVsOptions',
              'DoUnghost',
              'DoUnghostOptions',
              'DollyTool',
              'DopeSheetEditor',
              'DownloadBonusTools',
              'Drag',
              'DragOptions',
              'Duplicate',
              'DuplicateCurve',
              'DuplicateCurveOptions',
              'DuplicateEdges',
              'DuplicateEdgesOptions',
              'DuplicateFace',
              'DuplicateFaceOptions',
              'DuplicateNURBSPatches',
              'DuplicateNURBSPatchesOptions',
              'DuplicateSpecial',
              'DuplicateSpecialOptions',
              'DuplicateWithTransform',
              'DynamicRelationshipEditor',
              'EPCurveTool',
              'EPCurveToolOptions',
              'EditAssignedSet',
              'EditCharacterAttributes',
              'EditFluidResolution',
              'EditFluidResolutionOptions',
              'EditMembershipTool',
              'EditNormalizationGroups',
              'EditOversamplingForCacheSettings',
              'EditPSDTextureItem',
              'EditPolygonType',
              'EditTexture',
              'EmitFluidFromObject',
              'EmitFluidFromObjectOptions',
              'EmitFromObject',
              'EmitFromObjectOptions',
              'EmptyAnimLayer',
              'EnableAll',
              'EnableAllCaches',
              'EnableAllGeometryCache',
              'EnableConstraints',
              'EnableDynamicConstraints',
              'EnableExpressions',
              'EnableFluids',
              'EnableGlobalStitch',
              'EnableIKSolvers',
              'EnableMemoryCaching',
              'EnableNCloths',
              'EnableNParticles',
              'EnableNRigids',
              'EnableNucleuses',
              'EnableParticles',
              'EnableRigidBodies',
              'EnableSelectTool',
              'EnableSelectedIKHandles',
              'EnableSnapshots',
              'EnableTimeChangeUndoConsolidation',
              'EnableTimeWarp',
              'EnableWeightNrm',
              'EnableWeightPostNrm',
              'EnterConnectTool',
              'EnterEditMode',
              'EnterEditModePress',
              'EnterEditModeRelease',
              'EvaluationToolkit',
              'ExpandSelectedComponents',
              'Export',
              'ExportAnim',
              'ExportDeformerWeights',
              'ExportDeformerWeightsOptions',
              'ExportOfflineFile',
              'ExportOfflineFileFromRefEd',
              'ExportOfflineFileFromRefEdOptions',
              'ExportOfflineFileOptions',
              'ExportOptions',
              'ExportProxyContainer',
              'ExportProxyContainerOptions',
              'ExportSelection',
              'ExportSelectionOptions',
              'ExportSkinWeightMaps',
              'ExportSkinWeightMapsOptions',
              'ExpressionEditor',
              'ExtendCurve',
              'ExtendCurveOnSurface',
              'ExtendCurveOnSurfaceOptions',
              'ExtendCurveOptions',
              'ExtendFluid',
              'ExtendFluidOptions',
              'ExtendSurfaces',
              'ExtendSurfacesOptions',
              'ExtractFace',
              'ExtractFaceOptions',
              'ExtractSubdivSurfaceVertices',
              'ExtractSubdivSurfaceVerticesOptions',
              'Extrude',
              'ExtrudeEdge',
              'ExtrudeEdgeOptions',
              'ExtrudeFace',
              'ExtrudeFaceOptions',
              'ExtrudeOptions',
              'ExtrudeVertex',
              'ExtrudeVertexOptions',
              'FBIKLabelArm',
              'FBIKLabelBigToe',
              'FBIKLabelCenter',
              'FBIKLabelCollar',
              'FBIKLabelElbow',
              'FBIKLabelExtraFinger',
              'FBIKLabelFingerBase',
              'FBIKLabelFoot',
              'FBIKLabelFootThumb',
              'FBIKLabelHand',
              'FBIKLabelHead',
              'FBIKLabelHip',
              'FBIKLabelIndex',
              'FBIKLabelIndexToe',
              'FBIKLabelKnee',
              'FBIKLabelLeft',
              'FBIKLabelLeg',
              'FBIKLabelMiddleFinger',
              'FBIKLabelMiddleToe',
              'FBIKLabelNeck',
              'FBIKLabelNone',
              'FBIKLabelPinky',
              'FBIKLabelPinkyToe',
              'FBIKLabelRight',
              'FBIKLabelRingFinger',
              'FBIKLabelRingToe',
              'FBIKLabelRoot',
              'FBIKLabelShoulder',
              'FBIKLabelSpine',
              'FBIKLabelThumb',
              'FBIKLabelToeBase',
              'FilePathEditor',
              'FillHole',
              'FilletBlendTool',
              'FilletBlendToolOptions',
              'FineLevelComponentDisplay',
              'Fire',
              'FireOptions',
              'Fireworks',
              'FireworksOptions',
              'FitBSpline',
              'FitBSplineOptions',
              'Flare',
              'FlareOptions',
              'FlipMesh',
              'FlipTriangleEdge',
              'FlipTubeDirection',
              'FlipUVs',
              'FlipUVsOptions',
              'FloatSelectedObjects',
              'FloatSelectedObjectsOptions',
              'FloatSelectedPondObjects',
              'FloatSelectedPondObjectsOptions',
              'FloodSurfaces',
              'FlowPathObject',
              'FlowPathObjectOptions',
              'FluidEmitter',
              'FluidEmitterOptions',
              'FluidGradients',
              'FluidGradientsOptions',
              'FluidsToPolygons',
              'FourViewArrangement',
              'FourViewLayout',
              'FrameAll',
              'FrameAllInAllViews',
              'FrameSelected',
              'FrameSelected2',
              'FrameSelectedInAllViews',
              'FrameSelectedWithoutChildren',
              'FrameSelectedWithoutChildrenInAllViews',
              'FreeTangentWeight',
              'FreeformFillet',
              'FreeformFilletOptions',
              'FreezeTransformations',
              'FreezeTransformationsOptions',
              'FrontPerspViewLayout',
              'FullCreaseSubdivSurface',
              'FullHotboxDisplay',
              'GameExporterWnd',
              'GeometryConstraint',
              'GeometryConstraintOptions',
              'GeometryToBoundingBox',
              'GeometryToBoundingBoxOptions',
              'GetCartoonExample',
              'GetFluidExample',
              'GetHairExample',
              'GetOceanPondExample',
              'GetSettingsFromSelectedStroke',
              'GetToonExample',
              'GhostObject',
              'GlobalDiskCacheControl',
              'GlobalStitch',
              'GlobalStitchOptions',
              'GoToBindPose',
              'GoToDefaultView',
              'GoToMaxFrame',
              'GoToMinFrame',
              'GoToNextDrivenKey',
              'GoToPreviousDrivenKey',
              'GoToWorkingFrame',
              'Goal',
              'GoalOptions',
              'GraphCopy',
              'GraphCopyOptions',
              'GraphCut',
              'GraphCutOptions',
              'GraphDelete',
              'GraphDeleteOptions',
              'GraphEditor',
              'GraphEditorAbsoluteView',
              'GraphEditorAlwaysDisplayTangents',
              'GraphEditorDisableCurveSelection',
              'GraphEditorDisplayTangentActive',
              'GraphEditorDisplayValues',
              'GraphEditorEnableCurveSelection',
              'GraphEditorFrameAll',
              'GraphEditorFrameCenterView',
              'GraphEditorFramePlaybackRange',
              'GraphEditorFrameSelected',
              'GraphEditorLockChannel',
              'GraphEditorNeverDisplayTangents',
              'GraphEditorNormalizedView',
              'GraphEditorStackedView',
              'GraphEditorUnlockChannel',
              'GraphEditorValueLinesToggle',
              'GraphPaste',
              'GraphPasteOptions',
              'GraphSnap',
              'GraphSnapOptions',
              'Gravity',
              'GravityOptions',
              'GreasePencilTool',
              'GridOptions',
              'GridUV',
              'GridUVOptions',
              'Group',
              'GroupOptions',
              'GrowLoopPolygonSelectionRegion',
              'GrowPolygonSelectionRegion',
              'HIKBodyPartMode',
              'HIKCharacterControlsTool',
              'HIKCycleMode',
              'HIKFullBodyMode',
              'HIKLiveConnectionTool',
              'HIKPinRotate',
              'HIKPinTranslate',
              'HIKSelectedMode',
              'HIKSetBodyPartKey',
              'HIKSetFullBodyKey',
              'HIKSetSelectionKey',
              'HIKToggleReleasePinning',
              'HairScaleTool',
              'HairUVSetLinkingEditor',
              'HardwareRenderBuffer',
              'Help',
              'HideAll',
              'HideAllLabels',
              'HideBoundingBox',
              'HideCameraManipulators',
              'HideCameras',
              'HideClusters',
              'HideControllers',
              'HideDeformers',
              'HideDeformingGeometry',
              'HideDynamicConstraints',
              'HideFluids',
              'HideFollicles',
              'HideFur',
              'HideGeometry',
              'HideHairSystems',
              'HideHotbox',
              'HideIKHandles',
              'HideIntermediateObjects',
              'HideJoints',
              'HideKinematics',
              'HideLattices',
              'HideLightManipulators',
              'HideLights',
              'HideManipulators',
              'HideMarkers',
              'HideNCloths',
              'HideNParticles',
              'HideNRigids',
              'HideNURBSCurves',
              'HideNURBSSurfaces',
              'HideNonlinears',
              'HideObjectGeometry',
              'HidePlanes',
              'HidePolygonSurfaces',
              'HideSculptObjects',
              'HideSelectedObjects',
              'HideSmoothSkinInfluences',
              'HideStrokeControlCurves',
              'HideStrokePathCurves',
              'HideStrokes',
              'HideSubdivSurfaces',
              'HideTexturePlacements',
              'HideUIElements',
              'HideUnselectedCVs',
              'HideUnselectedObjects',
              'HideWrapInfluences',
              'HighQualityDisplay',
              'HighlightWhatsNew',
              'HoldCurrentKeys',
              'HotkeyPreferencesWindow',
              'HyperGraphPanelRedoViewChange',
              'HyperGraphPanelUndoViewChange',
              'HypergraphDGWindow',
              'HypergraphDecreaseDepth',
              'HypergraphHierarchyWindow',
              'HypergraphIncreaseDepth',
              'HypergraphWindow',
              'HypershadeAddOnNodeCreate',
              'HypershadeAdditiveGraphingMode',
              'HypershadeAutoSizeNodes',
              'HypershadeCloseActiveTab',
              'HypershadeCloseAllTabs',
              'HypershadeCollapseAsset',
              'HypershadeConnectSelected',
              'HypershadeConvertPSDToFileTexture',
              'HypershadeConvertPSDToLayeredTexture',
              'HypershadeConvertToFileTexture',
              'HypershadeConvertToFileTextureOptionBox',
              'HypershadeCreateAsset',
              'HypershadeCreateContainerOptions',
              'HypershadeCreateNewTab',
              'HypershadeCreatePSDFile',
              'HypershadeCreateTab',
              'HypershadeDeleteAllBakeSets',
              'HypershadeDeleteAllCamerasAndImagePlanes',
              'HypershadeDeleteAllLights',
              'HypershadeDeleteAllShadingGroupsAndMaterials',
              'HypershadeDeleteAllTextures',
              'HypershadeDeleteAllUtilities',
              'HypershadeDeleteDuplicateShadingNetworks',
              'HypershadeDeleteNodes',
              'HypershadeDeleteSelected',
              'HypershadeDeleteUnusedNodes',
              'HypershadeDisplayAllShapes',
              'HypershadeDisplayAsExtraLargeSwatches',
              'HypershadeDisplayAsIcons',
              'HypershadeDisplayAsLargeSwatches',
              'HypershadeDisplayAsList',
              'HypershadeDisplayAsMediumSwatches',
              'HypershadeDisplayAsSmallSwatches',
              'HypershadeDisplayInterestingShapes',
              'HypershadeDisplayNoShapes',
              'HypershadeDuplicateShadingNetwork',
              'HypershadeDuplicateWithConnections',
              'HypershadeDuplicateWithoutNetwork',
              'HypershadeEditPSDFile',
              'HypershadeEditTexture',
              'HypershadeExpandAsset',
              'HypershadeExportSelectedNetwork',
              'HypershadeFrameAll',
              'HypershadeFrameSelected',
              'HypershadeGraphAddSelected',
              'HypershadeGraphClearGraph',
              'HypershadeGraphDownstream',
              'HypershadeGraphMaterialsOnSelectedObjects',
              'HypershadeGraphRearrange',
              'HypershadeGraphRemoveDownstream',
              'HypershadeGraphRemoveSelected',
              'HypershadeGraphRemoveUnselected',
              'HypershadeGraphRemoveUpstream',
              'HypershadeGraphUpDownstream',
              'HypershadeGraphUpstream',
              'HypershadeGridToggleSnap',
              'HypershadeGridToggleVisibility',
              'HypershadeHideAttributes',
              'HypershadeImport',
              'HypershadeIncreaseTraversalDepth',
              'HypershadeMoveTabDown',
              'HypershadeMoveTabLeft',
              'HypershadeMoveTabRight',
              'HypershadeMoveTabUp',
              'HypershadeOpenBinsWindow',
              'HypershadeOpenBrowserWindow',
              'HypershadeOpenConnectWindow',
              'HypershadeOpenCreateWindow',
              'HypershadeOpenGraphEditorWindow',
              'HypershadeOpenMaterialViewerWindow',
              'HypershadeOpenModelEditorWindow',
              'HypershadeOpenOutlinerWindow',
              'HypershadeOpenPropertyEditorWindow',
              'HypershadeOpenRenderViewWindow',
              'HypershadeOpenSpreadSheetWindow',
              'HypershadeOpenUVEditorWindow',
              'HypershadeOutlinerPerspLayout',
              'HypershadePerspLayout',
              'HypershadePickWalkDown',
              'HypershadePickWalkLeft',
              'HypershadePickWalkRight',
              'HypershadePickWalkUp',
              'HypershadePinByDefault',
              'HypershadePinSelected',
              'HypershadePublishConnections',
              'HypershadeReduceTraversalDepth',
              'HypershadeRefreshAllSwatchesOnDisk',
              'HypershadeRefreshFileListing',
              'HypershadeRefreshSelectedSwatches',
              'HypershadeRefreshSelectedSwatchesOnDisk',
              'HypershadeRemoveAsset',
              'HypershadeRemoveTab',
              'HypershadeRenameActiveTab',
              'HypershadeRenameTab',
              'HypershadeRenderPerspLayout',
              'HypershadeRenderTextureRange',
              'HypershadeRenderTextureRangeOptions',
              'HypershadeRestoreLastClosedTab',
              'HypershadeRevertSelectedSwatches',
              'HypershadeRevertToDefaultTabs',
              'HypershadeSaveSwatchesToDisk',
              'HypershadeSelectBakeSets',
              'HypershadeSelectCamerasAndImagePlanes',
              'HypershadeSelectConnected',
              'HypershadeSelectDownStream',
              'HypershadeSelectLights',
              'HypershadeSelectMaterialsFromObjects',
              'HypershadeSelectObjectsWithMaterials',
              'HypershadeSelectShadingGroupsAndMaterials',
              'HypershadeSelectTextures',
              'HypershadeSelectUpStream',
              'HypershadeSelectUtilities',
              'HypershadeSetLargeNodeSwatchSize',
              'HypershadeSetSmallNodeSwatchSize',
              'HypershadeSetTraversalDepthUnlim',
              'HypershadeSetTraversalDepthZero',
              'HypershadeShapeMenuStateAll',
              'HypershadeShapeMenuStateAllExceptShadingGroupMembers',
              'HypershadeShapeMenuStateNoShapes',
              'HypershadeShowAllAttrs',
              'HypershadeShowConnectedAttrs',
              'HypershadeShowCustomAttrs',
              'HypershadeShowDirectoriesAndFiles',
              'HypershadeShowDirectoriesOnly',
              'HypershadeSortByName',
              'HypershadeSortByTime',
              'HypershadeSortByType',
              'HypershadeSortReverseOrder',
              'HypershadeTestTexture',
              'HypershadeTestTextureOptions',
              'HypershadeToggleAttrFilter',
              'HypershadeToggleCrosshairOnEdgeDragging',
              'HypershadeToggleNodeSwatchSize',
              'HypershadeToggleNodeTitleMode',
              'HypershadeToggleShowNamespace',
              'HypershadeToggleTransformDisplay',
              'HypershadeToggleUseAssetsAndPublishedAttributes',
              'HypershadeToggleZoomIn',
              'HypershadeToggleZoomOut',
              'HypershadeTransferAttributeValues',
              'HypershadeTransferAttributeValuesOptions',
              'HypershadeUnpinSelected',
              'HypershadeUpdatePSDNetworks',
              'HypershadeWindow',
              'IKHandleTool',
              'IKHandleToolOptions',
              'IKSplineHandleTool',
              'IKSplineHandleToolOptions',
              'IPROptions',
              'IPRRenderIntoNewWindow',
              'IgnoreHardwareShader',
              'IkHdsWin',
              'IkfkjdsWin',
              'Import',
              'ImportAnim',
              'ImportDeformerWeights',
              'ImportDeformerWeightsOptions',
              'ImportOptions',
              'ImportSkinWeightMaps',
              'ImportWorkspaceFiles',
              'InTangentAuto',
              'InTangentClamped',
              'InTangentFixed',
              'InTangentFlat',
              'InTangentLinear',
              'InTangentPlateau',
              'InTangentSpline',
              'IncreaseCheckerDensity',
              'IncreaseExposureCoarse',
              'IncreaseExposureFine',
              'IncreaseGammaCoarse',
              'IncreaseGammaFine',
              'IncreaseManipulatorSize',
              'IncrementAndSave',
              'IncrementFluidCenter',
              'InitialFluidStates',
              'InitialFluidStatesOptions',
              'InsertEdgeLoopTool',
              'InsertEdgeLoopToolOptions',
              'InsertIsoparms',
              'InsertIsoparmsOptions',
              'InsertJointTool',
              'InsertKeyToolActivate',
              'InsertKeyToolDeactivate',
              'InsertKeysTool',
              'InsertKeysToolOptions',
              'InsertKnot',
              'InsertKnotOptions',
              'InteractiveBindSkin',
              'InteractiveBindSkinOptions',
              'InteractivePlayback',
              'InteractiveSplitTool',
              'InteractiveSplitToolOptions',
              'IntersectCurve',
              'IntersectCurveOptions',
              'IntersectSurfaces',
              'IntersectSurfacesOptions',
              'InvertSelection',
              'JdsWin',
              'JointTool',
              'JointToolOptions',
              'KeyBlendShapeTargetsWeight',
              'KeyframeTangentMarkingMenu',
              'KeyframeTangentMarkingMenuPopDown',
              'LODGenerateMeshes',
              'LODGenerateMeshesOptions',
              'LabelBasedOnJointNames',
              'LassoTool',
              'LastActionTool',
              'LatticeDeformKeysTool',
              'LatticeDeformKeysToolOptions',
              'LatticeUVTool',
              'LatticeUVToolOptions',
              'LayerRelationshipEditor',
              'LayoutUV',
              'LayoutUVAlong',
              'LayoutUVAlongOptions',
              'LayoutUVOptions',
              'LayoutUVRectangle',
              'LevelOfDetailGroup',
              'LevelOfDetailGroupOptions',
              'LevelOfDetailUngroup',
              'LightCentricLightLinkingEditor',
              'Lightning',
              'LightningOptions',
              'LockCamera',
              'LockContainer',
              'LockCurveLength',
              'LockNormals',
              'LockTangentWeight',
              'Loft',
              'LoftOptions',
              'LongPolygonNormals',
              'LookAtSelection',
              'LoopBrushAnimation',
              'LoopBrushAnimationOptions',
              'LowQualityDisplay',
              'MakeBoats',
              'MakeBoatsOptions',
              'MakeBrushSpring',
              'MakeBrushSpringOptions',
              'MakeCollide',
              'MakeCollideHair',
              'MakeCollideOptions',
              'MakeCurvesDynamic',
              'MakeCurvesDynamicOptions',
              'MakeFluidCollide',
              'MakeFluidCollideOptions',
              'MakeHoleTool',
              'MakeHoleToolOptions',
              'MakeLightLinks',
              'MakeLive',
              'MakeMotionField',
              'MakeMotorBoats',
              'MakeMotorBoatsOptions',
              'MakePaintable',
              'MakePondBoats',
              'MakePondBoatsOptions',
              'MakePondMotorBoats',
              'MakePondMotorBoatsOptions',
              'MakePressureCurve',
              'MakePressureCurveOptions',
              'MakeShadowLinks',
              'MakeStereoLinks',
              'MakeUVInstanceCurrent',
              'MapUVBorder',
              'MapUVBorderOptions',
              'MarkingMenuPopDown',
              'MarkingMenuPreferencesWindow',
              'MatchPivots',
              'MatchRotation',
              'MatchScaling',
              'MatchTransform',
              'MatchTranslation',
              'MatchUVs',
              'MatchUVsOptions',
              'MediumPolygonNormals',
              'MediumQualityDisplay',
              'MergeCharacterSet',
              'MergeEdgeTool',
              'MergeEdgeToolOptions',
              'MergeMultipleEdges',
              'MergeMultipleEdgesOptions',
              'MergeToCenter',
              'MergeUV',
              'MergeUVOptions',
              'MergeVertexTool',
              'MergeVertexToolOptions',
              'MergeVertices',
              'MergeVerticesOptions',
              'MinimizeApplication',
              'MirrorCutPolygonGeometry',
              'MirrorCutPolygonGeometryOptions',
              'MirrorDeformerWeights',
              'MirrorDeformerWeightsOptions',
              'MirrorJoint',
              'MirrorJointOptions',
              'MirrorPolygonGeometry',
              'MirrorPolygonGeometryOptions',
              'MirrorSkinWeights',
              'MirrorSkinWeightsOptions',
              'MirrorSubdivSurface',
              'MirrorSubdivSurfaceOptions',
              'ModelingPanelRedoViewChange',
              'ModelingPanelUndoViewChange',
              'ModifyConstraintAxis',
              'ModifyConstraintAxisOptions',
              'ModifyCurrentSet',
              'ModifyDisplacementPress',
              'ModifyDisplacementRelease',
              'ModifyLowerRadiusPress',
              'ModifyLowerRadiusRelease',
              'ModifyOpacityPress',
              'ModifyOpacityRelease',
              'ModifyPaintValuePress',
              'ModifyPaintValueRelease',
              'ModifyStampDepthPress',
              'ModifyStampDepthRelease',
              'ModifyUVVectorPress',
              'ModifyUVVectorRelease',
              'ModifyUpperRadiusPress',
              'ModifyUpperRadiusRelease',
              'MoveCacheToInput',
              'MoveCurveSeam',
              'MoveDown',
              'MoveIKtoFK',
              'MoveInfluence',
              'MoveLeft',
              'MoveNearestPickedKeyToolActivate',
              'MoveNearestPickedKeyToolDeactivate',
              'MoveNormalTool',
              'MoveNormalToolOptions',
              'MovePolygonComponent',
              'MovePolygonComponentOptions',
              'MoveRight',
              'MoveRotateScaleTool',
              'MoveRotateScaleToolToggleSnapMode',
              'MoveRotateScaleToolToggleSnapRelativeMode',
              'MoveSewUVs',
              'MoveSkinJointsTool',
              'MoveSkinJointsToolOptions',
              'MoveSurfaceSeam',
              'MoveTool',
              'MoveToolOptions',
              'MoveUVTool',
              'MoveUp',
              'MultiCutTool',
              'NCreateEmitter',
              'NCreateEmitterOptions',
              'NEmitFromObject',
              'NEmitFromObjectOptions',
              'NParticleStyleBalls',
              'NParticleStyleCloud',
              'NParticleStylePoints',
              'NParticleStyleThickCloud',
              'NParticleStyleWater',
              'NParticleToPolygons',
              'NParticleTool',
              'NParticleToolOptions',
              'NURBSSmoothnessFine',
              'NURBSSmoothnessFineOptions',
              'NURBSSmoothnessHull',
              'NURBSSmoothnessHullOptions',
              'NURBSSmoothnessMedium',
              'NURBSSmoothnessMediumOptions',
              'NURBSSmoothnessRough',
              'NURBSSmoothnessRoughOptions',
              'NURBSTexturePlacementTool',
              'NURBSTexturePlacementToolOptions',
              'NURBSToPolygons',
              'NURBSToPolygonsOptions',
              'NamespaceEditor',
              'NewScene',
              'NewSceneOptions',
              'Newton',
              'NewtonOptions',
              'NextFrame',
              'NextGreasePencilFrame',
              'NextKey',
              'NextManipulatorHandle',
              'NextSkinPaintMode',
              'NextViewArrangement',
              'NodeEditorAddIterationStatePorts',
              'NodeEditorAddOnNodeCreate',
              'NodeEditorAdditiveGraphingMode',
              'NodeEditorAutoSizeNodes',
              'NodeEditorBackToParent',
              'NodeEditorCloseActiveTab',
              'NodeEditorCloseAllTabs',
              'NodeEditorConnectNodeOnCreation',
              'NodeEditorConnectOnDrop',
              'NodeEditorConnectSelectedNodes',
              'NodeEditorConnectionStyleBezier',
              'NodeEditorConnectionStyleCorner',
              'NodeEditorConnectionStyleSShape',
              'NodeEditorConnectionStyleStraight',
              'NodeEditorCopy',
              'NodeEditorCopyConnectionsOnPaste',
              'NodeEditorCreateCompound',
              'NodeEditorCreateDoWhileCompound',
              'NodeEditorCreateForEachCompound',
              'NodeEditorCreateIterateCompound',
              'NodeEditorCreateNodePopup',
              'NodeEditorCreateTab',
              'NodeEditorCut',
              'NodeEditorDeleteNodes',
              'NodeEditorDiveIntoCompound',
              'NodeEditorExplodeCompound',
              'NodeEditorExtendToShapes',
              'NodeEditorGraphAddSelected',
              'NodeEditorGraphAllShapes',
              'NodeEditorGraphAllShapesExceptShading',
              'NodeEditorGraphClearGraph',
              'NodeEditorGraphDownstream',
              'NodeEditorGraphNoShapes',
              'NodeEditorGraphRearrange',
              'NodeEditorGraphRemoveDownstream',
              'NodeEditorGraphRemoveSelected',
              'NodeEditorGraphRemoveUnselected',
              'NodeEditorGraphRemoveUpstream',
              'NodeEditorGraphUpDownstream',
              'NodeEditorGraphUpstream',
              'NodeEditorGridToggleCrosshairOnEdgeDragging',
              'NodeEditorGridToggleSnap',
              'NodeEditorGridToggleVisibility',
              'NodeEditorHideAttributes',
              'NodeEditorHighlightConnectionsOnNodeSelection',
              'NodeEditorIncreaseTraversalDepth',
              'NodeEditorLayout',
              'NodeEditorPaste',
              'NodeEditorPickWalkDown',
              'NodeEditorPickWalkLeft',
              'NodeEditorPickWalkRight',
              'NodeEditorPickWalkUp',
              'NodeEditorPinByDefault',
              'NodeEditorPinSelected',
              'NodeEditorPublishCompound',
              'NodeEditorRedockTornOffTab',
              'NodeEditorReduceTraversalDepth',
              'NodeEditorRenameActiveTab',
              'NodeEditorRenderSwatches',
              'NodeEditorRestoreLastClosedTab',
              'NodeEditorSelectConnected',
              'NodeEditorSelectDownStream',
              'NodeEditorSelectUpStream',
              'NodeEditorSetLargeNodeSwatchSize',
              'NodeEditorSetSmallNodeSwatchSize',
              'NodeEditorSetTraversalDepthUnlim',
              'NodeEditorSetTraversalDepthZero',
              'NodeEditorShapeMenuStateAll',
              'NodeEditorShapeMenuStateAllExceptShadingGroupMembers',
              'NodeEditorShapeMenuStateNoShapes',
              'NodeEditorShowAllAttrs',
              'NodeEditorShowAllCustomAttrs',
              'NodeEditorShowConnectedAttrs',
              'NodeEditorShowCustomAttrs',
              'NodeEditorToggleAttrFilter',
              'NodeEditorToggleConsistentNodeNameSize',
              'NodeEditorToggleCreateNodePane',
              'NodeEditorToggleLockUnlock',
              'NodeEditorToggleNodeSelectedPins',
              'NodeEditorToggleNodeSwatchSize',
              'NodeEditorToggleNodeTitleMode',
              'NodeEditorToggleShowNamespace',
              'NodeEditorToggleSyncedSelection',
              'NodeEditorToggleUseAssetsAndPublishedAttributes',
              'NodeEditorToggleZoomIn',
              'NodeEditorToggleZoomOut',
              'NodeEditorTransforms',
              'NodeEditorUnpinSelected',
              'NodeEditorWindow',
              'NonSacredTool',
              'NonWeightedTangents',
              'NormalConstraint',
              'NormalConstraintOptions',
              'NormalizeUVs',
              'NormalizeUVsOptions',
              'NormalizeWeights',
              'NurbsCurveToBezier',
              'ObjectCentricLightLinkingEditor',
              'OffsetCurve',
              'OffsetCurveOnSurface',
              'OffsetCurveOnSurfaceOptions',
              'OffsetCurveOptions',
              'OffsetEdgeLoopTool',
              'OffsetEdgeLoopToolOptions',
              'OffsetSurfaces',
              'OffsetSurfacesOptions',
              'OpenAELiveLink',
              'OpenAREACommunity',
              'OpenAREAForums',
              'OpenAutodeskAccount',
              'OpenAutodeskExchange',
              'OpenAutodeskStore',
              'OpenBifContentBrowser',
              'OpenBrowserSetupAssistant',
              'OpenBugReport',
              'OpenChannelBox',
              'OpenChannelsLayers',
              'OpenCharacterGenerator',
              'OpenCloseCurve',
              'OpenCloseCurveOptions',
              'OpenCloseSurfaces',
              'OpenCloseSurfacesOptions',
              'OpenColorSetEditor',
              'OpenContentBrowser',
              'OpenCreaseEditor',
              'OpenDGProfiler',
              'OpenDesktopAnalytics',
              'OpenFBXReview',
              'OpenFacebook',
              'OpenFeatureRequest',
              'OpenHomePage',
              'OpenLayerEditor',
              'OpenLearningChannel',
              'OpenLearningPath',
              'OpenLightEditor',
              'OpenMASHContentBrowser',
              'OpenMelCmdRef',
              'OpenMenuFinder',
              'OpenModelingToolkit',
              'OpenNodeAttrRef',
              'OpenProductResearch',
              'OpenPyCmdRef',
              'OpenReleaseNotes',
              'OpenScene',
              'OpenSceneOptions',
              'OpenStartupMovies',
              'OpenStereoRigManager',
              'OpenSupportCenter',
              'OpenTimeSliderPrefs',
              'OpenTinkercad',
              'OpenTrialTutorials',
              'OpenTutorials',
              'OpenViewportPrefs',
              'OpenVisorForMeshes',
              'OpenXGenEditor',
              'OptimizeScene',
              'OptimizeSceneOptions',
              'OptimzeUVs',
              'OptimzeUVsOptions',
              'OrientConstraint',
              'OrientConstraintOptions',
              'OrientJoint',
              'OrientJointOptions',
              'OutTangentAuto',
              'OutTangentClamped',
              'OutTangentFixed',
              'OutTangentFlat',
              'OutTangentLinear',
              'OutTangentPlateau',
              'OutTangentSpline',
              'OutlinerCollapseAllItems',
              'OutlinerCollapseAllSelectedItems',
              'OutlinerDoHide',
              'OutlinerExpandAllItems',
              'OutlinerExpandAllSelectedItems',
              'OutlinerRenameSelectedItem',
              'OutlinerRevealSelected',
              'OutlinerToggleAssignedMaterials',
              'OutlinerToggleAttributes',
              'OutlinerToggleAutoExpandLayers',
              'OutlinerToggleConnected',
              'OutlinerToggleDAGOnly',
              'OutlinerToggleIgnoreHidden',
              'OutlinerToggleIgnoreUseColor',
              'OutlinerToggleNamespace',
              'OutlinerToggleOrganizeByClip',
              'OutlinerToggleOrganizeByLayer',
              'OutlinerToggleReferenceMembers',
              'OutlinerToggleReferenceNodes',
              'OutlinerToggleSetMembers',
              'OutlinerToggleShapes',
              'OutlinerToggleShowMuteInformation',
              'OutlinerToggleTimeEditor',
              'OutlinerUnhide',
              'OutlinerWindow',
              'PFXUVSetLinkingEditor',
              'PaintCacheTool',
              'PaintCacheToolOptions',
              'PaintClusterWeightsTool',
              'PaintClusterWeightsToolOptions',
              'PaintDeltaMushWeightsTool',
              'PaintDeltaMushWeightsToolOptions',
              'PaintEffectPanelActivate',
              'PaintEffectPanelDeactivate',
              'PaintEffectsGlobalSettings',
              'PaintEffectsMeshQuality',
              'PaintEffectsPanel',
              'PaintEffectsToCurve',
              'PaintEffectsToCurveOptions',
              'PaintEffectsToNurbs',
              'PaintEffectsToNurbsOptions',
              'PaintEffectsToPoly',
              'PaintEffectsToPolyOptions',
              'PaintEffectsTool',
              'PaintEffectsToolOptions',
              'PaintEffectsWindow',
              'PaintFluidsTool',
              'PaintFluidsToolOptions',
              'PaintGeomCacheTool',
              'PaintGeomCacheToolOptions',
              'PaintGrid',
              'PaintGridOptions',
              'PaintHairBaldness',
              'PaintHairColor',
              'PaintHairFollicles',
              'PaintHairFolliclesOptions',
              'PaintHairSpecularColor',
              'PaintJiggleWeightsTool',
              'PaintJiggleWeightsToolOptions',
              'PaintLatticeWeightsTool',
              'PaintLatticeWeightsToolOptions',
              'PaintNonlinearWeightsTool',
              'PaintNonlinearWeightsToolOptions',
              'PaintOnPaintableObjects',
              'PaintOnViewPlane',
              'PaintOperationMarkingMenuPress',
              'PaintOperationMarkingMenuRelease',
              'PaintRandom',
              'PaintRandomOptions',
              'PaintReduceWeightsTool',
              'PaintReduceWeightsToolOptions',
              'PaintSetMembershipTool',
              'PaintSetMembershipToolOptions',
              'PaintShrinkWrapWeightsTool',
              'PaintShrinkWrapWeightsToolOptions',
              'PaintSoftWeights',
              'PaintSoftWeightsOptions',
              'PaintTensionWeightsTool',
              'PaintTensionWeightsToolOptions',
              'PaintTextureDeformerWeightsTool',
              'PaintTextureDeformerWeightsToolOptions',
              'PaintToonBorderColor',
              'PaintToonCreaseColor',
              'PaintToonLineOffset',
              'PaintToonLineOpacity',
              'PaintToonLineWidth',
              'PaintToonProfileColor',
              'PaintTransferAttributes',
              'PaintVertexColorTool',
              'PaintVertexColorToolOptions',
              'PaintWireWeightsTool',
              'PaintWireWeightsToolOptions',
              'PanZoomTool',
              'PanePop',
              'PanelPreferencesWindow',
              'ParameterTool',
              'Parent',
              'ParentBaseWire',
              'ParentBaseWireOptions',
              'ParentConstraint',
              'ParentConstraintOptions',
              'ParentOptions',
              'PartSpriteWizard',
              'PartialCreaseSubdivSurface',
              'ParticleCollisionEvents',
              'ParticleFill',
              'ParticleFillOptions',
              'ParticleInstancer',
              'ParticleInstancerOptions',
              'ParticleTool',
              'ParticleToolOptions',
              'PartitionEditor',
              'PasteKeys',
              'PasteKeysOptions',
              'PasteSelected',
              'PasteUVs',
              'PasteVertexSkinWeights',
              'PasteVertexWeights',
              'PauseViewportEval',
              'PencilCurveTool',
              'PencilCurveToolOptions',
              'PerPointEmissionRates',
              'PerformExportToBackburner',
              'PerformPrecompExport',
              'PerformPrecompExportOptions',
              'PerformTessellationSetup',
              'PerformTessellationSetupOptions',
              'PerformanceSettingsWindow',
              'PerspGraphHypergraphLayout',
              'PerspGraphLayout',
              'PerspGraphOutlinerLayout',
              'PerspRelationshipEditorLayout',
              'PerspTextureLayout',
              'PfxBrushTransfer',
              'PfxFlipTubeDir',
              'PfxGetBrush',
              'PfxMakeCollide',
              'PfxPresetBlend',
              'PfxSetLineModifierObject',
              'PickColorActivate',
              'PickColorDeactivate',
              'PickWalkDown',
              'PickWalkDownSelect',
              'PickWalkIn',
              'PickWalkLeft',
              'PickWalkLeftSelect',
              'PickWalkOut',
              'PickWalkRight',
              'PickWalkRightSelect',
              'PickWalkStopAtTransform',
              'PickWalkUp',
              'PickWalkUpSelect',
              'PickWalkUseController',
              'PinSelection',
              'PinSelectionOptions',
              'PixelMoveDown',
              'PixelMoveLeft',
              'PixelMoveRight',
              'PixelMoveUp',
              'Planar',
              'PlanarOptions',
              'PlaybackBackward',
              'PlaybackForward',
              'PlaybackFree',
              'PlaybackLoopContinuous',
              'PlaybackLoopOnce',
              'PlaybackLoopOscillate',
              'PlaybackPefRealtime',
              'PlaybackRangeEnabledClips',
              'PlaybackRangeHighlight',
              'PlaybackRangeMinMax',
              'PlaybackRangePrefs',
              'PlaybackRangeSound',
              'PlaybackRangeStartEnd',
              'PlaybackRealtime',
              'PlaybackSteppedPreview',
              'PlaybackStop',
              'PlaybackToggle',
              'PlayblastOptions',
              'PlayblastWindow',
              'PlayblastWindowOptions',
              'PluginManager',
              'PointConstraint',
              'PointConstraintOptions',
              'PointOnCurve',
              'PointOnCurveOptions',
              'PointOnPolyConstraint',
              'PointOnPolyConstraintOptions',
              'PokePolygon',
              'PokePolygonOptions',
              'PoleVectorConstraint',
              'PoleVectorConstraintOptions',
              'PolyAssignSubdivHole',
              'PolyAssignSubdivHoleOptions',
              'PolyBrushMarkingMenu',
              'PolyBrushMarkingMenuPopDown',
              'PolyCircularize',
              'PolyCircularizeOptions',
              'PolyConvertToLoopAndDelete',
              'PolyConvertToLoopAndDuplicate',
              'PolyConvertToRingAndCollapse',
              'PolyConvertToRingAndSplit',
              'PolyCreaseTool',
              'PolyCreaseToolOptions',
              'PolyDisplayReset',
              'PolyEditEdgeFlow',
              'PolyEditEdgeFlowOptions',
              'PolyExtrude',
              'PolyExtrudeEdges',
              'PolyExtrudeEdgesOptions',
              'PolyExtrudeFaces',
              'PolyExtrudeFacesOptions',
              'PolyExtrudeOptions',
              'PolyExtrudeVertices',
              'PolyExtrudeVerticesOptions',
              'PolyMerge',
              'PolyMergeEdges',
              'PolyMergeEdgesOptions',
              'PolyMergeOptions',
              'PolyMergeVertices',
              'PolyMergeVerticesOptions',
              'PolyRemoveAllCrease',
              'PolyRemoveCrease',
              'PolySelectTool',
              'PolySelectToolOptions',
              'PolySpinEdgeBackward',
              'PolySpinEdgeForward',
              'PolygonApplyColor',
              'PolygonApplyColorOptions',
              'PolygonBooleanDifference',
              'PolygonBooleanDifferenceOptions',
              'PolygonBooleanIntersection',
              'PolygonBooleanIntersectionOptions',
              'PolygonBooleanUnion',
              'PolygonBooleanUnionOptions',
              'PolygonClearClipboard',
              'PolygonClearClipboardOptions',
              'PolygonCollapse',
              'PolygonCollapseEdges',
              'PolygonCollapseFaces',
              'PolygonCopy',
              'PolygonCopyOptions',
              'PolygonHardenEdge',
              'PolygonNormalEditTool',
              'PolygonPaste',
              'PolygonPasteOptions',
              'PolygonSelectionConstraints',
              'PolygonSoftenEdge',
              'PolygonSoftenHarden',
              'PolygonSoftenHardenOptions',
              'PoseEditor',
              'PoseInterpolatorNewGroup',
              'PositionAlongCurve',
              'PostInfinityConstant',
              'PostInfinityCycle',
              'PostInfinityCycleOffset',
              'PostInfinityLinear',
              'PostInfinityOscillate',
              'PreInfinityConstant',
              'PreInfinityCycle',
              'PreInfinityCycleOffset',
              'PreInfinityLinear',
              'PreInfinityOscillate',
              'PreferencesWindow',
              'PrefixHierarchyNames',
              'PreflightPolygon',
              'PreflightPolygonOptions',
              'PrelightPolygon',
              'PrelightPolygonOptions',
              'PreloadReferenceEditor',
              'PresetBlendingWindow',
              'PrevSkinPaintMode',
              'PreviousFrame',
              'PreviousGreasePencilFrame',
              'PreviousKey',
              'PreviousManipulatorHandle',
              'PreviousViewArrangement',
              'ProductInformation',
              'ProfilerTool',
              'ProfilerToolCategoryView',
              'ProfilerToolCpuView',
              'ProfilerToolHideSelected',
              'ProfilerToolHideSelectedRepetition',
              'ProfilerToolReset',
              'ProfilerToolShowAll',
              'ProfilerToolShowSelected',
              'ProfilerToolShowSelectedRepetition',
              'ProfilerToolThreadView',
              'ProfilerToolToggleRecording',
              'ProjectCurveOnMesh',
              'ProjectCurveOnMeshOptions',
              'ProjectCurveOnSurface',
              'ProjectCurveOnSurfaceOptions',
              'ProjectTangent',
              'ProjectTangentOptions',
              'ProjectWindow',
              'ProportionalModificationTool',
              'PruneCluster',
              'PruneLattice',
              'PruneSculpt',
              'PruneSmallWeights',
              'PruneSmallWeightsOptions',
              'PruneWire',
              'PublishAttributes',
              'PublishAttributesOptions',
              'PublishChildAnchor',
              'PublishChildAnchorOptions',
              'PublishConnections',
              'PublishConnectionsOptions',
              'PublishNode',
              'PublishParentAnchor',
              'PublishParentAnchorOptions',
              'PublishRootTransform',
              'PublishRootTransformOptions',
              'QuadDrawTool',
              'Quadrangulate',
              'QuadrangulateOptions',
              'QualityDisplayMarkingMenu',
              'QualityDisplayMarkingMenuPopDown',
              'QuickRigEditor',
              'Quit',
              'Radial',
              'RadialOptions',
              'RaiseApplicationWindows',
              'RaiseMainWindow',
              'RandomizeFollicles',
              'RandomizeFolliclesOptions',
              'RandomizeShells',
              'RandomizeShellsOptions',
              'ReassignBoneLatticeJoint',
              'ReattachSkeleton',
              'ReattachSkeletonJoints',
              'RebuildCurve',
              'RebuildCurveOptions',
              'RebuildSurfaces',
              'RebuildSurfacesOptions',
              'RecentCommandsWindow',
              'Redo',
              'RedoPreviousIPRRender',
              'RedoPreviousRender',
              'RedoViewChange',
              'ReducePolygon',
              'ReducePolygonOptions',
              'ReferenceEditor',
              'RefineSelectedComponents',
              'RegionKeysTool',
              'RelaxInitialState',
              'RelaxInitialStateOptions',
              'RelaxUVShell',
              'RelaxUVShellOptions',
              'RemoveBindingSet',
              'RemoveBlendShape',
              'RemoveBlendShapeOptions',
              'RemoveBrushSharing',
              'RemoveConstraintTarget',
              'RemoveConstraintTargetOptions',
              'RemoveFromCharacterSet',
              'RemoveFromContainer',
              'RemoveFromContainerOptions',
              'RemoveInbetween',
              'RemoveInfluence',
              'RemoveJoint',
              'RemoveLatticeTweaks',
              'RemoveMaterialSoloing',
              'RemoveNewPfxToon',
              'RemoveShrinkWrapInnerObject',
              'RemoveShrinkWrapSurfaces',
              'RemoveShrinkWrapTarget',
              'RemoveSubdivProxyMirror',
              'RemoveSubdivProxyMirrorOptions',
              'RemoveUnusedInfluences',
              'RemoveWire',
              'RemoveWireOptions',
              'RemoveWrapInfluence',
              'RenameAttribute',
              'RenameCurrentColorSet',
              'RenameCurrentSet',
              'RenameCurrentUVSet',
              'RenameJointsFromLabels',
              'RenderDiagnostics',
              'RenderFlagsWindow',
              'RenderGlobalsWindow',
              'RenderIntoNewWindow',
              'RenderLayerEditorWindow',
              'RenderLayerRelationshipEditor',
              'RenderOptions',
              'RenderPassSetEditor',
              'RenderSequence',
              'RenderSequenceOptions',
              'RenderSetupWindow',
              'RenderTextureRange',
              'RenderTextureRangeOptions',
              'RenderViewNextImage',
              'RenderViewPrevImage',
              'RenderViewWindow',
              'ReorderVertex',
              'RepeatLast',
              'RepeatLastActionAtMousePosition',
              'ReplaceObjects',
              'ReplaceObjectsOptions',
              'RerootSkeleton',
              'ResampleCurve',
              'ResampleCurveOptions',
              'ResetCurrentWorkspace',
              'ResetDisplay',
              'ResetLattice',
              'ResetReflectionOptions',
              'ResetSoftSelectOptions',
              'ResetTemplateBrush',
              'ResetTransformations',
              'ResetTransformationsOptions',
              'ResetViewport',
              'ResetWeightsToDefault',
              'ResetWire',
              'ResetWireOptions',
              'ResolveInterpenetration',
              'ResolveInterpenetrationOptions',
              'RestoreUIElements',
              'RetimeKeysTool',
              'RetimeKeysToolOptions',
              'ReverseCurve',
              'ReverseCurveOptions',
              'ReversePolygonNormals',
              'ReversePolygonNormalsOptions',
              'ReverseSurfaceDirection',
              'ReverseSurfaceDirectionOptions',
              'ReverseToonObjects',
              'Revolve',
              'RevolveOptions',
              'RigidBindSkin',
              'RigidBindSkinOptions',
              'RigidBodySolver',
              'RotateTool',
              'RotateToolMarkingMenu',
              'RotateToolMarkingMenuPopDown',
              'RotateToolOptions',
              'RotateToolWithSnapMarkingMenu',
              'RotateToolWithSnapMarkingMenuPopDown',
              'RotateUVTool',
              'RotateUVs',
              'RotateUVsOptions',
              'RoundTool',
              'RoundToolOptions',
              'STRSTweakModeOff',
              'STRSTweakModeOn',
              'STRSTweakModeToggle',
              'SaveBrushPreset',
              'SaveCurrentLayout',
              'SaveCurrentWorkspace',
              'SaveFluidStateAs',
              'SaveInitialStateAll',
              'SavePreferences',
              'SaveScene',
              'SaveSceneAs',
              'SaveSceneAsOptions',
              'SaveSceneOptions',
              'ScaleConstraint',
              'ScaleConstraintOptions',
              'ScaleCurvature',
              'ScaleCurvatureOptions',
              'ScaleKeys',
              'ScaleKeysOptions',
              'ScaleTool',
              'ScaleToolMarkingMenu',
              'ScaleToolMarkingMenuPopDown',
              'ScaleToolOptions',
              'ScaleToolWithSnapMarkingMenu',
              'ScaleToolWithSnapMarkingMenuPopDown',
              'ScaleUVTool',
              'ScriptEditor',
              'ScriptPaintTool',
              'ScriptPaintToolOptions',
              'SculptGeometryTool',
              'SculptGeometryToolOptions',
              'SculptMeshActivateBrushSize',
              'SculptMeshActivateBrushStrength',
              'SculptMeshDeactivateBrushSize',
              'SculptMeshDeactivateBrushStrength',
              'SculptMeshFrame',
              'SculptMeshInvertFreeze',
              'SculptMeshUnfreezeAll',
              'SculptPolygonsTool',
              'SculptPolygonsToolOptions',
              'SculptReferenceVectorMarkingMenuPress',
              'SculptReferenceVectorMarkingMenuRelease',
              'SculptSubdivsTool',
              'SculptSubdivsToolOptions',
              'SculptSurfacesTool',
              'SculptSurfacesToolOptions',
              'SearchAndReplaceNames',
              'SelectAll',
              'SelectAllAssets',
              'SelectAllBrushes',
              'SelectAllCameras',
              'SelectAllClusters',
              'SelectAllDynamicConstraints',
              'SelectAllFluids',
              'SelectAllFollicles',
              'SelectAllFurs',
              'SelectAllGeometry',
              'SelectAllHairSystem',
              'SelectAllIKHandles',
              'SelectAllImagePlanes',
              'SelectAllInput',
              'SelectAllJoints',
              'SelectAllLattices',
              'SelectAllLights',
              'SelectAllMarkingMenu',
              'SelectAllMarkingMenuPopDown',
              'SelectAllNCloths',
              'SelectAllNParticles',
              'SelectAllNRigids',
              'SelectAllNURBSCurves',
              'SelectAllNURBSSurfaces',
              'SelectAllOutput',
              'SelectAllParticles',
              'SelectAllPolygonGeometry',
              'SelectAllRigidBodies',
              'SelectAllRigidConstraints',
              'SelectAllSculptObjects',
              'SelectAllStrokes',
              'SelectAllSubdivGeometry',
              'SelectAllTransforms',
              'SelectAllWires',
              'SelectBorderEdgeTool',
              'SelectBrushNames',
              'SelectCVSelectionBoundary',
              'SelectCVsMask',
              'SelectComponentToolMarkingMenu',
              'SelectComponentToolMarkingMenuPopDown',
              'SelectContainerContents',
              'SelectContiguousEdges',
              'SelectContiguousEdgesOptions',
              'SelectCurveCVsAll',
              'SelectCurveCVsFirst',
              'SelectCurveCVsLast',
              'SelectCurvePointsMask',
              'SelectEdgeLoop',
              'SelectEdgeLoopSp',
              'SelectEdgeMask',
              'SelectEdgeRing',
              'SelectEdgeRingSp',
              'SelectFacePath',
              'SelectFacetMask',
              'SelectHierarchy',
              'SelectHullsMask',
              'SelectIsolate',
              'SelectLightsIlluminatingObject',
              'SelectLightsShadowingObject',
              'SelectLinesMask',
              'SelectMaskToolMarkingMenu',
              'SelectMaskToolMarkingMenuPopDown',
              'SelectMeshUVShell',
              'SelectMultiComponentMask',
              'SelectNextIntermediatObject',
              'SelectNone',
              'SelectObjectsIlluminatedByLight',
              'SelectObjectsShadowedByLight',
              'SelectPointsMask',
              'SelectPolygonSelectionBoundary',
              'SelectPolygonToolMarkingMenu',
              'SelectPolygonToolMarkingMenuPopDown',
              'SelectPreviousObjects3dsMax',
              'SelectPreviousObjectsMotionBuilder',
              'SelectPreviousObjectsMudbox',
              'SelectSharedColorInstances',
              'SelectSharedUVInstances',
              'SelectShortestEdgePathTool',
              'SelectSimilar',
              'SelectSimilarOptions',
              'SelectSurfaceBorder',
              'SelectSurfaceBorderOptions',
              'SelectSurfacePointsMask',
              'SelectTextureReferenceObject',
              'SelectTimeWarp',
              'SelectToggleMode',
              'SelectTool',
              'SelectToolMarkingMenu',
              'SelectToolMarkingMenuPopDown',
              'SelectToolOptionsMarkingMenu',
              'SelectToolOptionsMarkingMenuPopDown',
              'SelectUVBackFacingComponents',
              'SelectUVBorder',
              'SelectUVBorderComponents',
              'SelectUVFrontFacingComponents',
              'SelectUVMask',
              'SelectUVNonOverlappingComponents',
              'SelectUVNonOverlappingComponentsPerObject',
              'SelectUVOverlappingComponents',
              'SelectUVOverlappingComponentsPerObject',
              'SelectUVShell',
              'SelectUVTool',
              'SelectUnmappedFaces',
              'SelectVertexFaceMask',
              'SelectVertexMask',
              'SelectedAnimLayer',
              'SendAsNewScene3dsMax',
              'SendAsNewSceneMotionBuilder',
              'SendAsNewSceneMudbox',
              'SeparatePolygon',
              'SequenceEditor',
              'SetActiveKey',
              'SetAlignTool',
              'SetAsCombinationTarget',
              'SetAsCombinationTargetOptions',
              'SetBreakdownKey',
              'SetBreakdownKeyOptions',
              'SetCMCAmbient',
              'SetCMCAmbientDiffuse',
              'SetCMCDiffuse',
              'SetCMCEmission',
              'SetCMCNone',
              'SetCMCSpecular',
              'SetCurrentColorSet',
              'SetCurrentUVSet',
              'SetCutSewUVTool',
              'SetDefaultManipMove',
              'SetDefaultManipNone',
              'SetDefaultManipRotate',
              'SetDefaultManipScale',
              'SetDefaultManipTransform',
              'SetDrivenKey',
              'SetDrivenKeyOptions',
              'SetEditor',
              'SetExclusiveToCamera',
              'SetFluidAttrFromCurve',
              'SetFluidAttrFromCurveOptions',
              'SetFocusToCommandLine',
              'SetFocusToNumericInputLine',
              'SetFullBodyIKKeys',
              'SetFullBodyIKKeysAll',
              'SetFullBodyIKKeysBodyPart',
              'SetFullBodyIKKeysKeyToPin',
              'SetFullBodyIKKeysOptions',
              'SetFullBodyIKKeysSelected',
              'SetHairRestPositionFromCurrent',
              'SetHairRestPositionFromStart',
              'SetHairStartPositionFromCurrent',
              'SetHairStartPositionFromRest',
              'SetHiddenFromCamera',
              'SetIKFKKeyframe',
              'SetInitialState',
              'SetInitialStateOptions',
              'SetKey',
              'SetKeyAnimated',
              'SetKeyOptions',
              'SetKeyPath',
              'SetKeyRotate',
              'SetKeyScale',
              'SetKeyTranslate',
              'SetKeyVertexColor',
              'SetKeyframeForVertexColor',
              'SetMBSAdd',
              'SetMBSAverage',
              'SetMBSDivide',
              'SetMBSModulate2',
              'SetMBSMultiply',
              'SetMBSOverwrite',
              'SetMBSSubtract',
              'SetMaxInfluences',
              'SetMeshAmplifyTool',
              'SetMeshBulgeTool',
              'SetMeshCloneTargetTool',
              'SetMeshEraseTool',
              'SetMeshFillTool',
              'SetMeshFlattenTool',
              'SetMeshFoamyTool',
              'SetMeshFreezeTool',
              'SetMeshGrabTool',
              'SetMeshGrabUVTool',
              'SetMeshImprintTool',
              'SetMeshKnifeTool',
              'SetMeshMaskTool',
              'SetMeshPinchTool',
              'SetMeshRelaxTool',
              'SetMeshRepeatTool',
              'SetMeshScrapeTool',
              'SetMeshSculptTool',
              'SetMeshSmearTool',
              'SetMeshSmoothTargetTool',
              'SetMeshSmoothTool',
              'SetMeshSprayTool',
              'SetMeshWaxTool',
              'SetNClothStartFromMesh',
              'SetNormalAngle',
              'SetPassiveKey',
              'SetPreferredAngle',
              'SetPreferredAngleOptions',
              'SetProject',
              'SetReFormTool',
              'SetRestPosition',
              'SetRigidBodyCollision',
              'SetRigidBodyInterpenetration',
              'SetShrinkWrapInnerObject',
              'SetShrinkWrapTarget',
              'SetSnapTogetherTool',
              'SetSnapTogetherToolOptions',
              'SetStrokeControlCurves',
              'SetTimecode',
              'SetToFaceNormals',
              'SetToFaceNormalsOptions',
              'SetVertexNormal',
              'SetVertexNormalOptions',
              'SetWireframeColor',
              'SetWorkingFrame',
              'SetupAnimatedDisplacement',
              'SewUVs',
              'SewUVs3D',
              'SewUVsWithoutHotkey',
              'ShadingGroupAttributeEditor',
              'ShapeEditor',
              'ShapeEditorDuplicateTarget',
              'ShapeEditorNewGroup',
              'ShapeEditorSelectNone',
              'ShareColorInstances',
              'ShareOneBrush',
              'ShareUVInstances',
              'Shatter',
              'ShatterOptions',
              'ShelfPreferencesWindow',
              'ShortPolygonNormals',
              'ShotPlaylistEditor',
              'ShowAll',
              'ShowAllComponents',
              'ShowAllEditedComponents',
              'ShowAllLabels',
              'ShowAllPolyComponents',
              'ShowAllUI',
              'ShowAnimationUI',
              'ShowAttributeEditorOrChannelBox',
              'ShowBaseWire',
              'ShowBatchRender',
              'ShowBoundingBox',
              'ShowCameraManipulators',
              'ShowCameras',
              'ShowClusters',
              'ShowControllers',
              'ShowDeformers',
              'ShowDeformingGeometry',
              'ShowDynamicConstraints',
              'ShowDynamicsUI',
              'ShowFluids',
              'ShowFollicles',
              'ShowFur',
              'ShowGeometry',
              'ShowHairSystems',
              'ShowHotbox',
              'ShowIKHandles',
              'ShowJoints',
              'ShowKinematics',
              'ShowLastHidden',
              'ShowLattices',
              'ShowLightManipulators',
              'ShowLights',
              'ShowManipulatorTool',
              'ShowManipulators',
              'ShowMarkers',
              'ShowMeshAmplifyToolOptions',
              'ShowMeshBulgeToolOptions',
              'ShowMeshCloneTargetToolOptions',
              'ShowMeshEraseToolOptions',
              'ShowMeshFillToolOptions',
              'ShowMeshFlattenToolOptions',
              'ShowMeshFoamyToolOptions',
              'ShowMeshFreezeToolOptions',
              'ShowMeshGrabToolOptions',
              'ShowMeshGrabUVToolOptions',
              'ShowMeshImprintToolOptions',
              'ShowMeshKnifeToolOptions',
              'ShowMeshMaskToolOptions',
              'ShowMeshPinchToolOptions',
              'ShowMeshRelaxToolOptions',
              'ShowMeshRepeatToolOptions',
              'ShowMeshScrapeToolOptions',
              'ShowMeshSculptToolOptions',
              'ShowMeshSmearToolOptions',
              'ShowMeshSmoothTargetToolOptions',
              'ShowMeshSmoothToolOptions',
              'ShowMeshSprayToolOptions',
              'ShowMeshWaxToolOptions',
              'ShowModelingUI',
              'ShowNCloths',
              'ShowNParticles',
              'ShowNRigids',
              'ShowNURBSCurves',
              'ShowNURBSSurfaces',
              'ShowNonlinears',
              'ShowObjectGeometry',
              'ShowPlanes',
              'ShowPolygonSurfaces',
              'ShowRenderingUI',
              'ShowResultsOptions',
              'ShowRiggingUI',
              'ShowSculptObjects',
              'ShowSelectedObjects',
              'ShowShadingGroupAttributeEditor',
              'ShowSmoothSkinInfluences',
              'ShowStrokeControlCurves',
              'ShowStrokePathCurves',
              'ShowStrokes',
              'ShowSubdivSurfaces',
              'ShowSurfaceCVs',
              'ShowTexturePlacements',
              'ShowUIElements',
              'ShowWhatsNew',
              'ShowWrapInfluences',
              'ShrinkLoopPolygonSelectionRegion',
              'ShrinkPolygonSelectionRegion',
              'SimplifyCurve',
              'SimplifyCurveOptions',
              'SimplifyStrokePathCurves',
              'Sine',
              'SineOptions',
              'SinglePerspectiveViewLayout',
              'SingleViewArrangement',
              'SlideEdgeTool',
              'SlideEdgeToolOptions',
              'Smoke',
              'SmokeOptions',
              'SmoothBindSkin',
              'SmoothBindSkinOptions',
              'SmoothCurve',
              'SmoothCurveOptions',
              'SmoothHairCurves',
              'SmoothHairCurvesOptions',
              'SmoothPolygon',
              'SmoothPolygonOptions',
              'SmoothProxy',
              'SmoothProxyOptions',
              'SmoothSkinWeights',
              'SmoothSkinWeightsOptions',
              'SmoothTangent',
              'SmoothingDisplayShowBoth',
              'SmoothingDisplayToggle',
              'SmoothingLevelDecrease',
              'SmoothingLevelIncrease',
              'Snap2PointsTo2Points',
              'Snap2PointsTo2PointsOptions',
              'Snap3PointsTo3Points',
              'Snap3PointsTo3PointsOptions',
              'SnapKeys',
              'SnapKeysOptions',
              'SnapPointToPoint',
              'SnapPointToPointOptions',
              'SnapRotation',
              'SnapToCurve',
              'SnapToCurvePress',
              'SnapToCurveRelease',
              'SnapToGrid',
              'SnapToGridPress',
              'SnapToGridRelease',
              'SnapToMeshCenter',
              'SnapToMeshCenterPress',
              'SnapToMeshCenterRelease',
              'SnapToPixel',
              'SnapToPoint',
              'SnapToPointPress',
              'SnapToPointRelease',
              'SoftModDeformer',
              'SoftModDeformerOptions',
              'SoftModTool',
              'SoftModToolOptions',
              'SoloLastOutput',
              'SoloMaterial',
              'SplitEdge',
              'SplitEdgeRingTool',
              'SplitEdgeRingToolOptions',
              'SplitMeshWithProjectedCurve',
              'SplitMeshWithProjectedCurveOptions',
              'SplitPolygonTool',
              'SplitPolygonToolOptions',
              'SplitUV',
              'SplitVertex',
              'SpreadSheetEditor',
              'SquareSurface',
              'SquareSurfaceOptions',
              'Squash',
              'SquashOptions',
              'StitchEdgesTool',
              'StitchEdgesToolOptions',
              'StitchSurfacePoints',
              'StitchSurfacePointsOptions',
              'StitchTogether',
              'StitchTogetherOptions',
              'StraightenCurves',
              'StraightenCurvesOptions',
              'StraightenUVBorder',
              'StraightenUVBorderOptions',
              'SubdCutUVs',
              'SubdivProxy',
              'SubdivProxyOptions',
              'SubdivSmoothnessFine',
              'SubdivSmoothnessFineOptions',
              'SubdivSmoothnessHull',
              'SubdivSmoothnessHullOptions',
              'SubdivSmoothnessMedium',
              'SubdivSmoothnessMediumOptions',
              'SubdivSmoothnessRough',
              'SubdivSmoothnessRoughOptions',
              'SubdivSurfaceCleanTopology',
              'SubdivSurfaceHierarchyMode',
              'SubdivSurfaceMatchTopology',
              'SubdivSurfacePolygonProxyMode',
              'SubdivToNURBS',
              'SubdivToNURBSOptions',
              'SubdividePolygon',
              'SubdividePolygonOptions',
              'SubstituteGeometry',
              'SubstituteGeometryOptions',
              'SurfaceBooleanIntersectTool',
              'SurfaceBooleanIntersectToolOptions',
              'SurfaceBooleanSubtractTool',
              'SurfaceBooleanSubtractToolOptions',
              'SurfaceBooleanUnionTool',
              'SurfaceBooleanUnionToolOptions',
              'SurfaceEditingTool',
              'SurfaceEditingToolOptions',
              'SurfaceFlow',
              'SurfaceFlowOptions',
              'SwapBlendShape',
              'SwapBlendShapeOptions',
              'SwapBufferCurve',
              'Symmetrize',
              'SymmetrizeSelection',
              'SymmetrizeUV',
              'SymmetrizeUVBrushSizeOff',
              'SymmetrizeUVBrushSizeOn',
              'SymmetrizeUVOptions',
              'TagAsController',
              'TagAsControllerParent',
              'TangentConstraint',
              'TangentConstraintOptions',
              'TangentsAuto',
              'TangentsClamped',
              'TangentsFixed',
              'TangentsFlat',
              'TangentsLinear',
              'TangentsPlateau',
              'TangentsSpline',
              'TangentsStepped',
              'TemplateBrushSettings',
              'TemplateObject',
              'Tension',
              'TensionOptions',
              'TesselateSubdivSurface',
              'TesselateSubdivSurfaceOptions',
              'TestTexture',
              'TestTextureOptions',
              'TexSculptActivateBrushSize',
              'TexSculptActivateBrushStrength',
              'TexSculptDeactivateBrushSize',
              'TexSculptDeactivateBrushStrength',
              'TexSculptInvertPin',
              'TexSculptUnpinAll',
              'TexSewActivateBrushSize',
              'TexSewDeactivateBrushSize',
              'TextureCentricUVLinkingEditor',
              'TextureToGeometry',
              'TextureToGeometryOptions',
              'TextureViewWindow',
              'ThreeBottomSplitViewArrangement',
              'ThreeLeftSplitViewArrangement',
              'ThreePointArcTool',
              'ThreePointArcToolOptions',
              'ThreeRightSplitViewArrangement',
              'ThreeTopSplitViewArrangement',
              'TimeDraggerToolActivate',
              'TimeDraggerToolDeactivate',
              'TimeEditorAddToSoloSelectedTracks',
              'TimeEditorClipHoldToggle',
              'TimeEditorClipLoopToggle',
              'TimeEditorClipRazor',
              'TimeEditorClipResetTiming',
              'TimeEditorClipScaleEnd',
              'TimeEditorClipScaleStart',
              'TimeEditorClipScaleToggle',
              'TimeEditorClipTrimEnd',
              'TimeEditorClipTrimStart',
              'TimeEditorClipTrimToggle',
              'TimeEditorCopyClips',
              'TimeEditorCreateAdditiveLayer',
              'TimeEditorCreateAnimTracksAtEnd',
              'TimeEditorCreateAudioClip',
              'TimeEditorCreateAudioTracksAtEnd',
              'TimeEditorCreateClip',
              'TimeEditorCreateClipOptions',
              'TimeEditorCreateGroupFromSelection',
              'TimeEditorCreateOverrideLayer',
              'TimeEditorCreatePoseClip',
              'TimeEditorCreatePoseClipOptions',
              'TimeEditorCutClips',
              'TimeEditorDeleteClips',
              'TimeEditorDeleteSelectedTracks',
              'TimeEditorExplodeGroup',
              'TimeEditorExportSelection',
              'TimeEditorExportSelectionOpt',
              'TimeEditorFbxExportAll',
              'TimeEditorFrameAll',
              'TimeEditorFrameCenterView',
              'TimeEditorFramePlaybackRange',
              'TimeEditorFrameSelected',
              'TimeEditorGhostTrackToggle',
              'TimeEditorImportAnimation',
              'TimeEditorKeepTransitionsTogglePress',
              'TimeEditorKeepTransitionsToggleRelease',
              'TimeEditorMuteSelectedTracks',
              'TimeEditorOpenContentBrowser',
              'TimeEditorPasteClips',
              'TimeEditorRealTimeRefreshToggle',
              'TimeEditorRippleEditTogglePress',
              'TimeEditorRippleEditToggleRelease',
              'TimeEditorSceneAuthoringToggle',
              'TimeEditorSetKey',
              'TimeEditorSetZeroKey',
              'TimeEditorSoloSelectedTracks',
              'TimeEditorToggleMuteSelectedTracks',
              'TimeEditorToggleSnapToClipPress',
              'TimeEditorToggleSnapToClipRelease',
              'TimeEditorToggleSoloSelectedTracks',
              'TimeEditorToggleTimeCursorPress',
              'TimeEditorToggleTimeCursorRelease',
              'TimeEditorUnmuteAllTracks',
              'TimeEditorUnmuteSelectedTracks',
              'TimeEditorUnsoloAllTracks',
              'TimeEditorWindow',
              'ToggleAnimationDetails',
              'ToggleAttributeEditor',
              'ToggleAutoActivateBodyPart',
              'ToggleAutoFrame',
              'ToggleAutoFrameTime',
              'ToggleAutoSmooth',
              'ToggleBackfaceCulling',
              'ToggleBackfaceGeometry',
              'ToggleBorderEdges',
              'ToggleCVs',
              'ToggleCacheVisibility',
              'ToggleCameraNames',
              'ToggleCapsLockDisplay',
              'ToggleChannelBox',
              'ToggleChannelsLayers',
              'ToggleCharacterControls',
              'ToggleColorFeedback',
              'ToggleCommandLine',
              'ToggleCompIDs',
              'ToggleContainerCentric',
              'ToggleCreaseEdges',
              'ToggleCreaseVertices',
              'ToggleCreateNurbsPrimitivesAsTool',
              'ToggleCreatePolyPrimitivesAsTool',
              'ToggleCullingVertices',
              'ToggleCurrentContainerHud',
              'ToggleCurrentFrame',
              'ToggleCustomNURBSComponents',
              'ToggleDisplacement',
              'ToggleDisplayColorsAttr',
              'ToggleDisplayGradient',
              'ToggleEdgeIDs',
              'ToggleEdgeMetadata',
              'ToggleEditPivot',
              'ToggleEditPoints',
              'ToggleEffectsMeshDisplay',
              'ToggleEvaluationManagerVisibility',
              'ToggleFaceIDs',
              'ToggleFaceMetadata',
              'ToggleFaceNormalDisplay',
              'ToggleFaceNormals',
              'ToggleFastInteraction',
              'ToggleFkIk',
              'ToggleFocalLength',
              'ToggleFrameRate',
              'ToggleFullScreenMode',
              'ToggleGrid',
              'ToggleHelpLine',
              'ToggleHikDetails',
              'ToggleHoleFaces',
              'ToggleHulls',
              'ToggleIKAllowRotation',
              'ToggleIKHandleSnap',
              'ToggleIKSolvers',
              'ToggleInViewEditor',
              'ToggleInViewMessage',
              'ToggleIsolateSelect',
              'ToggleJointLabels',
              'ToggleKeepHardEdgeCulling',
              'ToggleKeepWireCulling',
              'ToggleLatticePoints',
              'ToggleLatticeShape',
              'ToggleLayerBar',
              'ToggleLocalRotationAxes',
              'ToggleMainMenubar',
              'ToggleMaterialLoadingDetailsVisibility',
              'ToggleMeshEdges',
              'ToggleMeshFaces',
              'ToggleMeshMaps',
              'ToggleMeshPoints',
              'ToggleMeshUVBorders',
              'ToggleMetadata',
              'ToggleModelEditorBars',
              'ToggleModelingToolkit',
              'ToggleMultiColorFeedback',
              'ToggleNormals',
              'ToggleNurbsPrimitivesAsToolExitOnComplete',
              'ToggleObjectDetails',
              'ToggleOppositeFlagOfSelectedShapes',
              'ToggleOriginAxis',
              'ToggleOutliner',
              'TogglePaintAtDepth',
              'TogglePaintOnPaintableObjects',
              'TogglePanZoomPress',
              'TogglePanZoomRelease',
              'TogglePanelMenubar',
              'ToggleParticleCount',
              'TogglePolyCount',
              'TogglePolyDisplayEdges',
              'TogglePolyDisplayHardEdges',
              'TogglePolyDisplayHardEdgesColor',
              'TogglePolyDisplayLimitToSelected',
              'TogglePolyDisplaySoftEdges',
              'TogglePolyNonPlanarFaceDisplay',
              'TogglePolyPrimitivesAsToolExitOnComplete',
              'TogglePolyUVsCreateShader',
              'TogglePolygonFaceCenters',
              'TogglePolygonFaceTriangles',
              'TogglePolygonFaceTrianglesDisplay',
              'ToggleProxyDisplay',
              'ToggleRangeSlider',
              'ToggleReflection',
              'ToggleRotationPivots',
              'ToggleScalePivots',
              'ToggleSceneTimecode',
              'ToggleSelectDetails',
              'ToggleSelectedLabels',
              'ToggleSelectionHandles',
              'ToggleShelf',
              'ToggleShowBufferCurves',
              'ToggleShowResults',
              'ToggleSoftEdges',
              'ToggleStatusLine',
              'ToggleSubdDetails',
              'ToggleSurfaceFaceCenters',
              'ToggleSurfaceOrigin',
              'ToggleSymmetryDisplay',
              'ToggleTangentDisplay',
              'ToggleTextureBorder',
              'ToggleTextureBorderEdges',
              'ToggleTimeSlider',
              'ToggleToolMessage',
              'ToggleToolSettings',
              'ToggleToolbox',
              'ToggleUIElements',
              'ToggleUVDistortion',
              'ToggleUVEditorIsolateSelectHUD',
              'ToggleUVEditorUVStatisticsHUD',
              'ToggleUVEditorUVStatisticsHUDOptions',
              'ToggleUVIsolateViewSelected',
              'ToggleUVShellBorder',
              'ToggleUVTextureImage',
              'ToggleUVs',
              'ToggleUnsharedUVs',
              'ToggleUseDefaultMaterial',
              'ToggleVertIDs',
              'ToggleVertMetadata',
              'ToggleVertexNormalDisplay',
              'ToggleVertices',
              'ToggleViewAxis',
              'ToggleViewCube',
              'ToggleViewportRenderer',
              'ToggleVisibilityAndKeepSelection',
              'ToggleVisibilityAndKeepSelectionOptions',
              'ToggleWireframeInArtisan',
              'ToggleZoomInMode',
              'ToolSettingsWindow',
              'TrackTool',
              'TransferAttributeValues',
              'TransferAttributeValuesOptions',
              'TransferAttributes',
              'TransferShadingSets',
              'TransferVertexOrder',
              'TransformNoSelectOffTool',
              'TransformNoSelectOnTool',
              'TransformPolygonComponent',
              'TransformPolygonComponentOptions',
              'TranslateToolMarkingMenu',
              'TranslateToolMarkingMenuPopDown',
              'TranslateToolWithSnapMarkingMenu',
              'TranslateToolWithSnapMarkingMenuPopDown',
              'TransplantHair',
              'TransplantHairOptions',
              'Triangulate',
              'TrimTool',
              'TrimToolOptions',
              'TruncateHairCache',
              'TumbleTool',
              'Turbulence',
              'TurbulenceOptions',
              'Twist',
              'TwistOptions',
              'TwoPointArcTool',
              'TwoPointArcToolOptions',
              'TwoSideBySideViewArrangement',
              'TwoStackedViewArrangement',
              'U3DBrushPressureOff',
              'U3DBrushPressureOn',
              'U3DBrushSizeOff',
              'U3DBrushSizeOn',
              'UIModeMarkingMenu',
              'UIModeMarkingMenuPopDown',
              'UVAutomaticProjection',
              'UVAutomaticProjectionOptions',
              'UVCameraBasedProjection',
              'UVCameraBasedProjectionOptions',
              'UVCentricUVLinkingEditor',
              'UVContourStretchProjection',
              'UVContourStretchProjectionOptions',
              'UVCreateSnapshot',
              'UVCylindricProjection',
              'UVCylindricProjectionOptions',
              'UVEditorFrameAll',
              'UVEditorFrameSelected',
              'UVEditorInvertPin',
              'UVEditorResetAllToDefault',
              'UVEditorToggleTextureBorderDisplay',
              'UVEditorUnpinAll',
              'UVGatherShells',
              'UVIsolateLoadSet',
              'UVNormalBasedProjection',
              'UVNormalBasedProjectionOptions',
              'UVOrientShells',
              'UVPlanarProjection',
              'UVPlanarProjectionOptions',
              'UVSetEditor',
              'UVSnapTogether',
              'UVSnapTogetherOptions',
              'UVSphericalProjection',
              'UVSphericalProjectionOptions',
              'UVStackSimilarShells',
              'UVStackSimilarShellsOptions',
              'UVStraighten',
              'UVStraightenOptions',
              'UVUnstackShells',
              'UVUnstackShellsOptions',
              'UncreaseSubdivSurface',
              'Undo',
              'UndoCanvas',
              'UndoViewChange',
              'UnfoldPackUVs3DInCurrentTile',
              'UnfoldPackUVs3DInEmptyTile',
              'UnfoldUV',
              'UnfoldUVOptions',
              'UnghostAll',
              'UnghostObject',
              'Ungroup',
              'UngroupOptions',
              'Uniform',
              'UniformOptions',
              'UnifyTangents',
              'UnitizeUVs',
              'UnitizeUVsOptions',
              'UniversalManip',
              'UniversalManipOptions',
              'UnlockContainer',
              'UnlockCurveLength',
              'UnlockNormals',
              'UnmirrorSmoothProxy',
              'UnmirrorSmoothProxyOptions',
              'Unparent',
              'UnparentOptions',
              'UnpinSelection',
              'UnpublishAttributes',
              'UnpublishChildAnchor',
              'UnpublishNode',
              'UnpublishParentAnchor',
              'UnpublishRootTransform',
              'UntemplateObject',
              'UntrimSurfaces',
              'UntrimSurfacesOptions',
              'UpdateBindingSet',
              'UpdateBindingSetOptions',
              'UpdateCurrentScene3dsMax',
              'UpdateCurrentSceneMotionBuilder',
              'UpdateCurrentSceneMudbox',
              'UpdateEraseSurface',
              'UpdatePSDTextureItem',
              'UpdateReferenceSurface',
              'UpdateSnapshot',
              'UseHardwareShader',
              'UseSelectedEmitter',
              'VertexNormalEditTool',
              'ViewAlongAxisNegativeX',
              'ViewAlongAxisNegativeY',
              'ViewAlongAxisNegativeZ',
              'ViewAlongAxisX',
              'ViewAlongAxisY',
              'ViewAlongAxisZ',
              'ViewImage',
              'ViewSequence',
              'ViewportEnableSmoothing',
              'VisorWindow',
              'VisualizeMetadataOptions',
              'VolumeAxis',
              'VolumeAxisOptions',
              'VolumeSkinBinding',
              'VolumeSkinBindingOptions',
              'Vortex',
              'VortexOptions',
              'WalkTool',
              'WarpImage',
              'WarpImageOptions',
              'Wave',
              'WaveOptions',
              'WedgePolygon',
              'WedgePolygonOptions',
              'WeightHammer',
              'WeightedTangents',
              'WhatsNewHighlightingOff',
              'WhatsNewHighlightingOn',
              'WhatsNewStartupDialogOff',
              'WhatsNewStartupDialogOn',
              'WireDropoffLocator',
              'WireDropoffLocatorOptions',
              'WireTool',
              'WireToolOptions',
              'WrinkleTool',
              'WrinkleToolOptions',
              'XgCreateIgSplineEditor',
              'XgmCreateInteractiveGroomSplines',
              'XgmCreateInteractiveGroomSplinesOption',
              'XgmSetClumpBrushTool',
              'XgmSetClumpBrushToolOption',
              'XgmSetCombBrushTool',
              'XgmSetCombBrushToolOption',
              'XgmSetCutBrushTool',
              'XgmSetCutBrushToolOption',
              'XgmSetDensityBrushTool',
              'XgmSetDensityBrushToolOption',
              'XgmSetDirectionBrushTool',
              'XgmSetDirectionBrushToolOption',
              'XgmSetFreezeBrushTool',
              'XgmSetFreezeBrushToolOption',
              'XgmSetGrabBrushTool',
              'XgmSetGrabBrushToolOption',
              'XgmSetLengthBrushTool',
              'XgmSetLengthBrushToolOption',
              'XgmSetNoiseBrushTool',
              'XgmSetNoiseBrushToolOption',
              'XgmSetPartBrushTool',
              'XgmSetPartBrushToolOption',
              'XgmSetPlaceBrushTool',
              'XgmSetPlaceBrushToolOption',
              'XgmSetSelectBrushTool',
              'XgmSetSelectBrushToolOption',
              'XgmSetSmoothBrushTool',
              'XgmSetSmoothBrushToolOption',
              'XgmSetWidthBrushTool',
              'XgmSetWidthBrushToolOption',
              'XgmSplineCacheCreate',
              'XgmSplineCacheCreateOptions',
              'XgmSplineCacheDelete',
              'XgmSplineCacheDeleteNodesAhead',
              'XgmSplineCacheDeleteOptions',
              'XgmSplineCacheDisableSelectedCache',
              'XgmSplineCacheEnableSelectedCache',
              'XgmSplineCacheExport',
              'XgmSplineCacheExportOptions',
              'XgmSplineCacheImport',
              'XgmSplineCacheImportOptions',
              'XgmSplineCacheReplace',
              'XgmSplineCacheReplaceOptions',
              'XgmSplineGeometryConvert',
              'XgmSplinePresetExport',
              'XgmSplinePresetImport',
              'XgmSplineSelectConvertToFreeze',
              'XgmSplineSelectReplaceBySelectedFaces',
              'ZoomTool',
              'attachCache',
              'attachFluidCache',
              'attachGeometryCache',
              'attachNclothCache',
              'buildSendToBackburnerDialog',
              'cacheAppend',
              'cacheAppendOpt',
              'clearDynStartState',
              'clearNClothStartState',
              'deleteGeometryCache',
              'deleteHistoryAheadOfGeomCache',
              'deleteNclothCache',
              'dynamicConstraintRemove',
              'fluidAppend',
              'fluidAppendOpt',
              'fluidDeleteCache',
              'fluidDeleteCacheFrames',
              'fluidDeleteCacheFramesOpt',
              'fluidDeleteCacheOpt',
              'fluidMergeCache',
              'fluidMergeCacheOpt',
              'fluidReplaceCache',
              'fluidReplaceCacheOpt',
              'fluidReplaceFrames',
              'fluidReplaceFramesOpt',
              'geometryAppendCache',
              'geometryAppendCacheOpt',
              'geometryCache',
              'geometryCacheOpt',
              'geometryDeleteCacheFrames',
              'geometryDeleteCacheFramesOpt',
              'geometryDeleteCacheOpt',
              'geometryExportCache',
              'geometryExportCacheOpt',
              'geometryMergeCache',
              'geometryMergeCacheOpt',
              'geometryReplaceCache',
              'geometryReplaceCacheFrames',
              'geometryReplaceCacheFramesOpt',
              'geometryReplaceCacheOpt',
              'mayaPreviewRenderIntoNewWindow',
              'mrMapVisualizer',
              'mrShaderManager',
              'nClothAppend',
              'nClothAppendOpt',
              'nClothCache',
              'nClothCacheOpt',
              'nClothCreate',
              'nClothCreateOptions',
              'nClothDeleteCacheFrames',
              'nClothDeleteCacheFramesOpt',
              'nClothDeleteCacheOpt',
              'nClothDeleteHistory',
              'nClothDeleteHistoryOpt',
              'nClothDisplayCurrentMesh',
              'nClothDisplayInputMesh',
              'nClothLocalToWorld',
              'nClothMakeCollide',
              'nClothMakeCollideOptions',
              'nClothMergeCache',
              'nClothMergeCacheOpt',
              'nClothRemove',
              'nClothReplaceCache',
              'nClothReplaceCacheOpt',
              'nClothReplaceFrames',
              'nClothReplaceFramesOpt',
              'nClothRestToInput',
              'nClothRestToInputStart',
              'nClothRestToMesh',
              'nClothWorldToLocal',
              'nConstraintAddMembers',
              'nConstraintAttractToMatch',
              'nConstraintAttractToMatchOptions',
              'nConstraintCollisionExclusion',
              'nConstraintCollisionExclusionOptions',
              'nConstraintComponent',
              'nConstraintComponentOptions',
              'nConstraintComponentToComponent',
              'nConstraintComponentToComponentOptions',
              'nConstraintConstraintMembershipTool',
              'nConstraintDisableCollision',
              'nConstraintDisableCollisionOptions',
              'nConstraintForceField',
              'nConstraintForceFieldOptions',
              'nConstraintPointToSurface',
              'nConstraintPointToSurfaceOptions',
              'nConstraintRemoveMembers',
              'nConstraintReplaceMembers',
              'nConstraintSelectMembers',
              'nConstraintSlideOnSurface',
              'nConstraintSlideOnSurfaceOptions',
              'nConstraintTearableSurface',
              'nConstraintTearableSurfaceOptions',
              'nConstraintTransform',
              'nConstraintTransformOptions',
              'nConstraintWeldBorders',
              'nConstraintWeldBordersOptions',
              'nucleusDisplayDynamicConstraintNodes',
              'nucleusDisplayMaterialNodes',
              'nucleusDisplayNComponentNodes',
              'nucleusDisplayOtherNodes',
              'nucleusDisplayTextureNodes',
              'nucleusDisplayTransformNodes',
              'nucleusGetEffectsAsset',
              'nucleusGetnClothExample',
              'nucleusGetnParticleExample',
              'replaceCacheFrames',
              'replaceCacheFramesOpt',
              'setDynStartState',
              'setNClothStartState'],
  'system': ['aaf2fcp',
             'allNodeTypes',
             'assignInputDevice',
             'attachDeviceAttr',
             'attrCompatibility',
             'audioTrack',
             'autoSave',
             'cacheFile',
             'cacheFileCombine',
             'cacheFileMerge',
             'cacheFileTrack',
             'clearCache',
             'cmdFileOutput',
             'convertUnit',
             'crashInfo',
             'dagObjectCompare',
             'date',
             'dbcount',
             'dbfootprint',
             'dbmessage',
             'dbpeek',
             'dbtrace',
             'detachDeviceAttr',
             'deviceEditor',
             'devicePanel',
             'dgInfo',
             'dgValidateCurve',
             'dgdirty',
             'dgeval',
             'dgfilter',
             'dgmodified',
             'dgtimer',
             'dirmap',
             'diskCache',
             'displayString',
             'dynamicLoad',
             'error',
             'exportEdits',
             'fcheck',
             'fileBrowserDialog',
             'fileDialog',
             'fileDialog2',
             'fileInfo',
             'filePathEditor',
             'findType',
             'flushUndo',
             'getFileList',
             'getInputDeviceRange',
             'getModifiers',
             'getModulePath',
             'hardware',
             'hitTest',
             'imfPlugins',
             'internalVar',
             'launch',
             'launchImageEditor',
             'listDeviceAttachments',
             'listInputDeviceAxes',
             'listInputDeviceButtons',
             'listInputDevices',
             'loadModule',
             'loadPlugin',
             'melInfo',
             'memory',
             'moduleInfo',
             'mouse',
             'namespace',
             'namespaceInfo',
             'ogs',
             'openGLExtension',
             'openMayaPref',
             'pluginDisplayFilter',
             'pluginInfo',
             'preloadRefEd',
             'profiler',
             'profilerTool',
             'recordAttr',
             'redo',
             'reference',
             'referenceEdit',
             'referenceQuery',
             'rehash',
             'reloadImage',
             'requires',
             'saveImage',
             'sceneEditor',
             'sceneUIReplacement',
             'scriptNode',
             'selLoadSettings',
             'setAttrMapping',
             'setInputDeviceMapping',
             'shotTrack',
             'showHelp',
             'sysFile',
             'timer',
             'timerX',
             'translator',
             'unassignInputDevice',
             'undo',
             'undoInfo',
             'unknownNode',
             'unknownPlugin',
             'unloadPlugin',
             'warning',
             'whatsNewHighlight',
             'workspace',
             'xpmPicker'],
  'uiClass': [],
  'windows': ['annotate',
              'artBuildPaintMenu',
              'attrColorSliderGrp',
              'attrControlGrp',
              'attrEnumOptionMenu',
              'attrEnumOptionMenuGrp',
              'attrFieldGrp',
              'attrFieldSliderGrp',
              'attrNavigationControlGrp',
              'attributeMenu',
              'autoPlace',
              'button',
              'buttonManip',
              'canCreateCaddyManip',
              'canvas',
              'channelBox',
              'checkBox',
              'checkBoxGrp',
              'cmdScrollFieldExecuter',
              'cmdScrollFieldReporter',
              'cmdShell',
              'colorEditor',
              'colorIndexSliderGrp',
              'colorInputWidgetGrp',
              'colorSliderButtonGrp',
              'colorSliderGrp',
              'columnLayout',
              'commandLine',
              'componentBox',
              'componentEditor',
              'confirmDialog',
              'contentBrowser',
              'control',
              'createEditor',
              'defaultNavigation',
              'disableIncorrectNameWarning',
              'dockControl',
              'editor',
              'editorTemplate',
              'falloffCurve',
              'falloffCurveAttr',
              'floatField',
              'floatFieldGrp',
              'floatScrollBar',
              'floatSlider',
              'floatSlider2',
              'floatSliderButtonGrp',
              'floatSliderGrp',
              'flowLayout',
              'fontDialog',
              'formLayout',
              'frameLayout',
              'getPanel',
              'grabColor',
              'gradientControl',
              'gradientControlNoAttr',
              'gridLayout',
              'hardwareRenderPanel',
              'headsUpDisplay',
              'headsUpMessage',
              'helpLine',
              'hotBox',
              'hotkey',
              'hotkeyCheck',
              'hotkeyEditorPanel',
              'hotkeySet',
              'hudButton',
              'hudSlider',
              'hudSliderButton',
              'hyperGraph',
              'hyperPanel',
              'hyperShade',
              'iconTextButton',
              'iconTextCheckBox',
              'iconTextRadioButton',
              'iconTextRadioCollection',
              'iconTextScrollList',
              'iconTextStaticLabel',
              'image',
              'inViewEditor',
              'inViewMessage',
              'intField',
              'intFieldGrp',
              'intScrollBar',
              'intSlider',
              'intSliderGrp',
              'layerButton',
              'layout',
              'layoutDialog',
              'linearPrecision',
              'loadPrefObjects',
              'loadUI',
              'mayaDpiSetting',
              'menu',
              'menuBarLayout',
              'menuEditor',
              'menuItem',
              'menuSet',
              'menuSetPref',
              'messageLine',
              'minimizeApp',
              'modelEditor',
              'modelPanel',
              'multiTouch',
              'nameCommand',
              'nameField',
              'nodeEditor',
              'nodeOutliner',
              'nodeTreeLister',
              'optionMenu',
              'optionMenuGrp',
              'outlinerEditor',
              'outlinerPanel',
              'overrideModifier',
              'palettePort',
              'paneLayout',
              'panel',
              'panelConfiguration',
              'panelHistory',
              'picture',
              'popupMenu',
              'progressBar',
              'progressWindow',
              'promptDialog',
              'radioButton',
              'radioButtonGrp',
              'radioCollection',
              'radioMenuItemCollection',
              'rangeControl',
              'refreshEditorTemplates',
              'rowColumnLayout',
              'rowLayout',
              'runTimeCommand',
              'saveAllShelves',
              'saveMenu',
              'savePrefObjects',
              'savePrefs',
              'saveShelf',
              'saveViewportSettings',
              'scmh',
              'scriptEditorInfo',
              'scriptTable',
              'scriptedPanel',
              'scriptedPanelType',
              'scrollField',
              'scrollLayout',
              'separator',
              'setFocus',
              'setMenuMode',
              'setNodeTypeFlag',
              'setStartupMessage',
              'shelfButton',
              'shelfLayout',
              'shelfTabLayout',
              'showSelectionInTitle',
              'showWindow',
              'soundControl',
              'spreadSheetEditor',
              'swatchDisplayPort',
              'switchTable',
              'symbolButton',
              'symbolCheckBox',
              'tabLayout',
              'text',
              'textField',
              'textFieldButtonGrp',
              'textFieldGrp',
              'textManip',
              'textScrollList',
              'thumbnailCaptureComponent',
              'timeControl',
              'timeField',
              'timeFieldGrp',
              'timePort',
              'toggleWindowVisibility',
              'toolBar',
              'toolButton',
              'toolCollection',
              'treeLister',
              'treeView',
              'viewManip',
              'visor',
              'webBrowser',
              'webBrowserPrefs',
              'window',
              'windowPref',
              'workspaceControl',
              'workspaceControlState',
              'workspaceLayoutManager',
              'workspacePanel']})